# -*- coding: latin-1 -*-
###############################################################################
# Copyright © 2018 NetApp, Inc. All Rights Reserved.
#
# CONFIDENTIALITY NOTICE:  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION OF
# NETAPP, INC. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE PRIOR
# EXPRESS WRITTEN PERMISSION OF NETAPP, INC.
###############################################################################

from unittest import TestCase
from unittest.mock import patch, mock_open, MagicMock
from requests.exceptions import HTTPError
from tenacity import RetryError
from authtools.config import Config


class TestConfig(TestCase):
    def setUp(self):
        self.testConfig = {'token_url': 'http://localhost:5555/connect/token',
                           'jwks_url': 'http://localhost:5555/.well-known/openid-configuration/jwks',
                           'config_url': 'http://localhost:5555/auth/api/1/configuration',
                           'testing_mode': True}

    @patch("builtins.open", new_callable=mock_open, read_data="testSecret")
    def test_get_client_secret(self, mock_file):
        expectedSecret = 'testSecret'
        secret = Config.getClientSecret()

        self.assertEqual(secret, expectedSecret)
        mock_file.assert_called_with('/sf/etc/auth/unknown.secret', 'r', encoding='utf8')

    @patch('authtools.config._read_json')
    def test_get_token_url(self, mock_config):
        mock_config.return_value = self.testConfig

        expected_url = 'http://localhost:5555/connect/token'

        token_url = Config.getTokenUrl()

        self.assertEqual(token_url, expected_url)
        mock_config.assert_called_with('/sf/etc/auth/config.json')

    @patch('authtools.config._read_json')
    def test_get_configuration_url(self, mock_config):
        mock_config.return_value = self.testConfig

        expected_url = 'http://localhost:5555/auth/api/1/configuration'

        token_url = Config.getConfigurationUrl()

        self.assertEqual(token_url, expected_url)
        mock_config.assert_called_with('/sf/etc/auth/config.json')

    @patch('authtools.config._read_json')
    def test_get_jwks_url(self, mock_config):
        mock_config.return_value = self.testConfig

        expected_url = 'http://localhost:5555/.well-known/openid-configuration/jwks'

        token_url = Config.getJwksUrl()

        self.assertEqual(token_url, expected_url)
        mock_config.assert_called_with('/sf/etc/auth/config.json')

    @patch('authtools.config.Config.getJwksUrl')
    @patch('authtools.config.requests.get')
    def test_get_jwks_data_succeeds(self, mock_get, mock_url):
        mock_url.return_value = 'http://localhost:5555/.well-known/openid-configuration/jwks'
        mock_get.return_value.text = 'testData'

        data = Config.getJwksData()

        self.assertEqual(data, 'testData')
        mock_url.assert_called()

    @patch('authtools.config.Config.getJwksUrl')
    @patch('authtools.config.requests.get')
    def test_get_jwks_data_fails(self, mock_get, mock_url):
        mock_url.return_value = 'http://localhost:5555/.well-known/openid-configuration/jwks'
        mock_get.side_effect = HTTPError()

        with self.assertRaises(RetryError):
            Config.getJwksData()

        mock_url.assert_called()

    @patch('authtools.config.Config.getJwksUrl')
    @patch('authtools.config.requests.get')
    def test_get_jwks_data_retry_succeds(self, mock_get, mock_url):
        mock_url.return_value = 'http://localhost:5555/.well-known/openid-configuration/jwks'
        successMock = MagicMock()
        successMock.text = 'testData'
        mock_get.side_effect = [HTTPError(), HTTPError(), successMock]

        data = Config.getJwksData()

        self.assertEqual(data, 'testData')
        mock_url.assert_called()

    @patch('authtools.config._read_json')
    def test_get_testing_mode_from_file_with_true_value(self, mock_config):
        mock_config.return_value = self.testConfig

        testing_mode = Config.getTestingMode()

        self.assertTrue(testing_mode)

    @patch('authtools.config._read_json')
    def test_get_testing_mode_from_file_with_false_value(self, mock_config):
        mock_config.return_value = self.testConfig
        self.testConfig['testing_mode'] = False

        testing_mode = Config.getTestingMode()

        self.assertFalse(testing_mode)

    @patch('authtools.config.Config.getTokenUrl')
    def test_get_about_url(self, mock_token_url):
        mock_token_url.return_value = "https://192.168.0.1/auth/connect/token"
        expected = "https://192.168.0.1/auth/about"

        actual = Config.getAboutUrl()
        self.assertEqual(expected, actual)
