from unittest import TestCase
from unittest.mock import patch, Mock

import requests
from requests import HTTPError
from requests.auth import HTTPBasicAuth

from authtools import get_client_credentials_session, get_resource_owner_password_session, get_mnode_service_session
from authtools.token_requests.request import check_auth_service_up, AuthToolsException

unreachable_url = "https://foobar.shouldnt.exist.foo.foobar.gov"


class TestRequest(TestCase):
    @patch("requests_oauthlib.oauth2_session.OAuth2Session.fetch_token")
    @patch("authtools.token_requests.request.check_auth_service_up")
    def test_get_client_credentials_session(self, mock_auth_service_up, mock_fetch_token):
        mock_auth_service_up = Mock()
        mock_fetch_token.return_value = Mock(requests)

        get_client_credentials_session("https://foo", "foo", "foo")

        mock_fetch_token.assert_called_with(token_url="https://foo", auth=HTTPBasicAuth("foo", "foo"), verify=True)

    @patch('authtools.config.Config.getAboutUrl')
    @patch('requests.get')
    def test_get_client_credentials_session_throws_when_about_is_unreachable(self, mock_get, mock_about_url):
        mock_request = Mock()
        mock_request.raise_for_status.side_effect = HTTPError()
        mock_get.return_value = mock_request
        mock_about_url.return_value = unreachable_url

        with self.assertRaises(AuthToolsException):
            get_client_credentials_session("https://foo", "foo", "foo")

    @patch("requests_oauthlib.oauth2_session.OAuth2Session.fetch_token")
    @patch("authtools.token_requests.request.check_auth_service_up")
    def test_get_resource_owner_password_session(self, mock_auth_service_up, mock_fetch_token):
        mock_auth_service_up = Mock()
        mock_fetch_token.return_value = Mock(requests)

        get_resource_owner_password_session("https://foo", "foo", "foo", "bar")

        mock_fetch_token.assert_called_with(token_url="https://foo",
                                            auth=HTTPBasicAuth("foo", None),
                                            verify=True,
                                            username="foo",
                                            password="bar")

    @patch('authtools.config.Config.getAboutUrl')
    @patch('requests.get')
    def test_get_resource_owner_password_session_throws_when_about_is_unreachable(self, mock_get, mock_about_url):
        mock_request = Mock()
        mock_request.raise_for_status.side_effect = HTTPError()
        mock_get.return_value = mock_request
        mock_about_url.return_value = unreachable_url

        with self.assertRaises(AuthToolsException):
            get_resource_owner_password_session("https://foo", "foo", "foo", "bar")

    @patch("authtools.token_requests.request.get_client_credentials_session")
    @patch("authtools.Config.getTokenUrl")
    @patch("authtools.Config.getClientSecret")
    def test_get_mnode_service_session(self, mock_get_client_secret, mock_get_token_url,
                                       mock_get_client_credentials_session):
        mock_get_client_secret.return_value = "secret"
        mock_get_token_url.return_value = "token_url"
        mock_get_client_credentials_session.return_value = Mock(requests)

        get_mnode_service_session()

        mock_get_client_credentials_session.assert_called_with(token_endpoint="token_url",
                                                               client_id="unknown",
                                                               client_secret="secret",
                                                               scope=None,
                                                               verify=True)

    @patch('authtools.config.Config.getAboutUrl')
    @patch("authtools.Config.getTokenUrl")
    @patch("authtools.Config.getClientSecret")
    @patch('requests.get')
    def test_get_mnode_service_session_throws_when_about_is_unreachable(self, mock_get, mock_get_client_secret,
                                                                        mock_get_token_url, mock_about_url):
        mock_request = Mock()
        mock_request.raise_for_status.side_effect = HTTPError()
        mock_get.return_value = mock_request
        mock_get_client_secret.return_value = "secret"
        mock_get_token_url.return_value = "token_url"
        mock_about_url.return_value = unreachable_url

        with self.assertRaises(AuthToolsException):
            get_mnode_service_session()

    @patch('authtools.config.Config.getAboutUrl')
    @patch('requests.get')
    def test_check_auth_service_up_no_exceptions(self, mock_get, mock_about_url):
        mock_about_url.return_value = "https://foo.bar"
        mock_get.return_value = Mock()

        try:
            check_auth_service_up()
        except:
            self.fail()

    @patch('authtools.config.Config.getAboutUrl')
    @patch('requests.get')
    def test_check_auth_service_up_should_throw_exception(self, mock_get, mock_about_url):
        mock_request = Mock()
        mock_request.raise_for_status.side_effect = HTTPError()
        mock_get.return_value = mock_request
        mock_about_url.return_value = "https://foo.bar"

        with self.assertRaises(AuthToolsException):
            check_auth_service_up()
