# -*- coding: latin-1 -*-
###############################################################################
# Copyright © 2018 NetApp, Inc. All Rights Reserved.
#
# CONFIDENTIALITY NOTICE:  THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION OF
# NETAPP, INC. USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED WITHOUT THE PRIOR
# EXPRESS WRITTEN PERMISSION OF NETAPP, INC.
###############################################################################
import requests
from requests import RequestException
from requests.auth import HTTPBasicAuth
from oauthlib.oauth2 import BackendApplicationClient, LegacyApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session

from authtools.config import Config


def get_client_credentials_session(token_endpoint: str, client_id: str, client_secret: str, scope=None, verify=True):
    """
    Get a request session with a bearer token requested using the client credentials grant type.
    :param token_endpoint: The token endpoint of the auth service
    :param client_id: The ID of the client registered with the authentication server
    :param client_secret: The secret of the client registered with the authentication server.
    :param scope: List of scopes whcih should be included in the session's access token
    :param verify: True to perform SSL verification for requests from this session, otherwise false
    :except requests.HTTPError: will throw when the auth server's about endpoint returns a 4xx or 5xx response.
    :except requests.ConnectionError: will throw when the auth server's about endpoint is unreachable.
    :return: A Requests-Session instance extended to provide bearer authentication
    """
    check_auth_service_up()
    auth = HTTPBasicAuth(client_id, client_secret)
    client = BackendApplicationClient(client_id=client_id, scope=scope)
    session = OAuth2Session(client=client)
    session.fetch_token(token_url=token_endpoint, auth=auth, verify=verify)
    return session


def get_resource_owner_password_session(token_endpoint: str, client_id: str, username: str, password: str,
                                        client_secret: str = None, scope=None, verify=True):
    """
    Get a request session with a bearer token requested using the resource owner grant type.
    :param token_endpoint: The token endpoint of the auth service
    :param client_id:  The ID of the client registered with the authentication server
    :param username: The username of the resource owner to authenticate with
    :param password: The password of the resource owner to authenticate with
    :param client_secret: The secret of the client registered with the authentication server. May be optional
    :param scope: List of scopes whcih should be included in the session's access token
    :param verify: True to perform SSL verification for requests from this session, otherwise false
    :except requests.HTTPError: will throw when the auth server's about endpoint returns a 4xx or 5xx response.
    :except requests.ConnectionError: will throw when the auth server's about endpoint is unreachable.
    :return: A Requests-Session instance extended to provide bearer authentication
    """
    check_auth_service_up()
    auth = HTTPBasicAuth(client_id, client_secret)
    client = LegacyApplicationClient(client_id=client_id, scope=scope)
    session = OAuth2Session(client=client)
    session.fetch_token(token_url=token_endpoint, auth=auth, username=username, password=password, verify=verify)
    return session


def get_mnode_service_session(scope=None, verify=True) -> OAuth2Session:
    """
    Get a request session for a standard MNode service with default client configuration.
    :param scope: List of scopes whcih should be included in the session's access token
    :param verify: True to perform SSL verification for requests from this session, otherwise false
    :return: A Requests-Session instance extended to provide bearer authentication
    """
    secret = Config.getClientSecret()
    token_url = Config.getTokenUrl()

    session = get_client_credentials_session(token_endpoint=token_url,
                                             client_id=Config.AUTH_CLIENT_ID, client_secret=secret,
                                             scope=scope, verify=verify)
    return session


class AuthToolsException(Exception):
    """Generic exception for authtools"""
    pass


def check_auth_service_up():
    """
    Checks the auth service about endpoint.
    :except AuthToolsException: when about endpoint returns a 4xx or 5xx response, or if
                                the Element Auth instance is unreachable.
    """
    about_url = Config.getAboutUrl()
    r = requests.get(about_url, verify=False)
    try:
        r.raise_for_status()
    except RequestException:
        raise AuthToolsException(
            "Unable to reach the Element Auth instance. "
            "Please ensure that the storage cluster is version 11.3 or higher. "
            "If the storage cluster version is correct, "
            "check that the auth server is online and responding on '{}'.".format(about_url))
