"""Camelot Test Case classes.

All Test Case definitions will inherit from camelot.TestCase, either directly or indirectly.

.. note::

    Some method names do appear a little redundant. They're defined as such to allow for a simpler
    refactor in the future, as well as a clear developer api.

Usage:

An actual Test Case can define the following 6 methods.

    #. __init__ [Required]
    #. _preconditions [Optional]
    #. _setup [Optional]
    #. _test [Required]
    #. _post_test_assertions [Optional]
    #. _teardown [Optional]

Test cases will accept the specific resources that they require in their ``__init__``, and must
call ``super(TestCaseName, self).__init__()``.
"""

import abc
import datetime
import sys

import autologger
import pytz
import six

from camelot import (
    _case_result,
    _enums,
    _errors,
    _formatter,
    _history,
    _record,
)

_log = autologger.get_sflogging(__name__)


class TestStep(object):
    """Camelot step object.

    TestStep objects are generated dynamically by a test case every time the
    :py:meth:`~camelot._case.TestCase._log_step` method is called.

    It's recommended the ``expected_result`` is a simple "Pass" or "Fail".

    :param int index: Index of the step in the test case
    :param str action: Step action
    :param str expected_result: The expected result of the action
    """

    def __init__(self, index, action, expected_result):
        self.index = index
        self.action = action
        self.expected_result = expected_result

    def __repr__(self):
        return 'Step {}: {} '.format(self.index, self.action)


@six.add_metaclass(abc.ABCMeta)
class TestCase(object):
    """Object which represents a single test case.

    A test case is a collection of steps and assertions which form a complete or partial test

    .. document private functions
    .. automethod:: _preconditions
    .. automethod:: _setup
    .. automethod:: _skip
    .. automethod:: _test
    .. automethod:: _post_test_assertions
    .. automethod:: _teardown
    .. automethod:: _log_step
    """

    statuses = _enums.ResultStatus

    def __init__(self):
        self.name = self.__class__.__name__
        self.history = _history.History(self.name)
        self._preconditions_ran = False
        self.status = _enums.ResultStatus.pending
        self._step_index = -1  # we want the first step to be index 0
        self._current_step = None
        self._steps_ran = []
        self.current_result = None  # the result is provided upon execute

    def _preconditions(self):
        """[Optional] Child test cases should override this method.

        Use this to create any component objects or assert the test case can run. This method is meant to
        catch problems that would prevent a test case from running, e.g. invalid hardware, bad
        components. If any test case precondition fails, the whole suite fails. No operations or API
        calls against resources.
        """

    def preconditions(self, index=None):
        """Setup and verify the test case preconditions have been met.

        Calls self._preconditions().

        :param int index: Index of the test case within a suite. If run outside of a suite, this
            will be None.
        """
        # The suite will call preconditions before calling execute on any test case.
        # However, execute will still call preconditions to verify that it has been run in order
        # to support running a test case outside of a suite.
        if not self._preconditions_ran:
            _log.info(_formatter.FMT_TEST_CASE_PRECONDITIONS.format(
                index=index, name=self.name))
            self._preconditions()
            self._preconditions_ran = True

    def _setup(self):
        """[Optional] Child classes should override this method to define your own setup.

        Setups may interact with the cluster or node.
        """

    def _skip(self):
        """[Optional] Child classes should override this to conditionally skip the test case.

        Return True to skip the test case. This method is run right before setup, and api calls
        are allowed.

        Mainly this applies to healthchecks and upgrades where we have evolving functionality that
        does not have versioned API that follows Element versions.

        :rtype: bool
        :returns: True, if the test case should be skipped
        """
        return False

    @abc.abstractmethod
    def _test(self):
        """[Required] Child classes must override this method to define your test case logic."""

    def _post_test_assertions(self):
        """[Optional] Child classes should override this method to verify any assertions.

        Collecting support bundles or checking for core files should occur here. Only getter API calls
        should be used against resources.
        """

    def _teardown(self):
        """[Optional] Child classes should override this method to define your own teardown.

        Teardowns may interact with resources.
        """

    def _set_final_result(self, result, status, transition, exc_info=None):
        """Set final status on self and result; also sets result.end_time as current time.

        :param camelot.CaseResult result: result object to modify
        :param camelot.ResultStatus status: status to set on self and result
        :param camelot._enums.CaseTransition transition: "setup", "test", "teardown", define
            where the exception info should go.
        :param camelot._record.ExcInfo exc_info: Exception info object
        """
        result.status = self.status = status
        if transition is _enums.CaseTransition.setup:
            result.setup_exc_info = exc_info
        if transition is _enums.CaseTransition.test or _enums.CaseTransition.post_test_assertions:
            result.test_exc_info = exc_info
        if transition is _enums.CaseTransition.teardown:
            result.teardown_exc_info = exc_info
        if exc_info:
            result.exc_info = exc_info
        result.end_time = datetime.datetime.now(pytz.utc)

    def execute(self, stop_on_failure=False, result=None):  # pylint: disable=too-complex
        """Execute the test case in the order: setup, test, teardown.

        :param bool stop_on_failure: If true, exception in _test will be raised immediately and teardown
            is skipped.
        :param camelot._case_result.CaseResult result: Optional test result object to update.
                                If not provided, one will be created and returned.
        :returns: Result object
        :rtype: camelot._case_result.CaseResult
        """
        if result is None:
            result = _case_result.CaseResult(self)
        elif result.status is not _enums.ResultStatus.pending:
            raise _errors.TestRuntimeError('Result object not pending. Status is {}. Result '
                                           'should be pending'.format(result.status))
        self.current_result = result
        result.steps = self._steps_ran
        result.status = self.status = _enums.ResultStatus.running
        result.start_time = datetime.datetime.now(pytz.utc)

        self.preconditions()
        if self._skip():
            _log.info(_formatter.banner(_formatter.FMT_TEST_CASE_SKIP.format(
                index=self.current_result.index, name=self.name)))
            self.status = _enums.ResultStatus.skipped
            self._set_final_result(result, self.status, _enums.CaseTransition.setup)
            return result
        # Run the setup.
        result = self._internal_setup(result, stop_on_failure)
        if self.status is _enums.ResultStatus.error:
            # if we already errored in the setup, run the teardown
            return self._internal_teardown(result)
        elif self.status is _enums.ResultStatus.cancelled:
            # if we cancelled in the setup, return
            return result

        # If preconditions or _internal_setup have signified that test case should be skipped, do so.
        if self.status is not _enums.ResultStatus.skipped:
            # Run the test and post_test_assertions
            result = self._internal_test(result, stop_on_failure)

        # If the test is cancelled or setup said we need to skip it, return the result
        if result.status in [_enums.ResultStatus.cancelled, _enums.ResultStatus.skipped]:
            return result

        # try out the teardown and return
        result = self._internal_teardown(result)
        if result.status is _enums.ResultStatus.running:
            self._set_final_result(result, _enums.ResultStatus.passed, _enums.CaseTransition.teardown)
        return result

    def __repr__(self):
        """Override repr to format the test case name when logging."""
        return self.name

    @staticmethod
    def _update_terminal(text):
        """Update the terminal title.

        This is useful to easily tell what test case and step a suite is on by looking at the terminal
        title. This has been tested with the Mint terminal, but does not work with Mac.

        :param str text: The text to set the terminal title to
        """
        if sys.platform.startswith('linux') and getattr(sys.stdout, 'isatty', lambda: False)():
            sys.stdout.write('\x1b]2;{}\x07'.format(text))  # pragma: no cover

    def _log_step(self, action, expected_result=''):
        """Log the step info with an optional expected result.

        :param str action: The step action. E.g. 'Power off node'. What action it takes to get there.
        :param str expected_result: What this test step intends to accomplish. E.g. Induce bin sync
        """
        self._step_index += 1
        self._current_step = TestStep(self._step_index, action, expected_result)
        self._update_terminal('{} - {}'.format(self.name, self._current_step))
        self._steps_ran.append(self._current_step)
        log_string = _formatter.step_banner(
            _formatter.FMT_TEST_STEP.format(index=self._current_step.index,
                                            name=self.name,
                                            action=self._current_step.action,
                                            expect=self._current_step.expected_result
                                            ))
        _log.info(log_string)

    def _get_exception_record(self, status, transition):
        """Log and set the exception record.

        Should only be called by camelot.

        :param camelot._enums.ResultStatus status: status to log
        :param camelot._enums.CaseTransition transition: the transition of the test case to log.
        :rtype: camelot._record.ExcInfo
        """
        exc = _record.ExcInfo(sys.exc_info())
        _log.error(_formatter.FMT_TEST_CASE_TRANSITION.format(
                status=status.value,
                name=self.name,
                transition=transition.value,
                exc=exc.pretty_fmt())
        )
        return exc

    def _internal_determine_status(self, status_if_not_cancelled, exc):
        """If exception is a KeyboardInterrupt, return status as cancelled, else the status arg.

        :param camelot._enums.ResultStatus status_if_not_cancelled: status to return if not
            cancelled
        :param six.moves.builtins.Exception exc: exception to check
        :rtype: camelot._enums.ResultStatus
        """
        if isinstance(exc, KeyboardInterrupt):
            return _enums.ResultStatus.cancelled
        else:
            return status_if_not_cancelled

    def _internal_setup(self, result, stop_on_failure):
        """Execute the setup.

        If the setup raises an Exception, set status to error and return. If cancelled,
        set to cancelled and return. ``stop_on_failure`` only honored for errored, not cancelled.

        :param camelot._case_result.CaseResult result: Result object running in the test
        :param bool stop_on_failure: whether we stop on failure
        :returns: Result object
        :rtype: camelot._case_result.CaseResult
        """
        try:
            self._update_terminal('{} - Setup'.format(self.name))
            _log.info(_formatter.banner(_formatter.FMT_TEST_CASE_SETUP.format(
                    index=result.index, name=self.name)))
            self._setup()
        except (Exception, KeyboardInterrupt) as e:  # pylint: disable=broad-except
            status = self._internal_determine_status(_enums.ResultStatus.error, e)
            exc = self._get_exception_record(status, _enums.CaseTransition.setup)
            self._set_final_result(result, status, _enums.CaseTransition.setup, exc_info=exc)
            if stop_on_failure and status is not _enums.ResultStatus.cancelled:
                raise
        return result

    def _internal_test(self, result, stop_on_failure):
        """Execute the test.

        :param camelot._case_result.CaseResult result: Result object running in the test
        :param bool stop_on_failure: whether we stop on failure
        :returns: Result object
        :rtype: camelot._case_result.CaseResult
        """
        try:
            _log.info(_formatter.banner(_formatter.FMT_TEST_CASE_TEST.format(
                    index=result.index, name=self.name)))
            self._test()
        except (Exception, KeyboardInterrupt) as e:  # pylint: disable=broad-except
            status = self._internal_determine_status(_enums.ResultStatus.failed, e)
            exc = self._get_exception_record(status, _enums.CaseTransition.test)
            self._set_final_result(result, status, _enums.CaseTransition.test, exc_info=exc)
            if stop_on_failure and status is _enums.ResultStatus.failed:
                raise
        if result.status is _enums.ResultStatus.cancelled:
            return result
        return self._internal_post_test_assertions(result, stop_on_failure)

    def _internal_post_test_assertions(self, result, stop_on_failure):
        """Execute the post_test_assertions.

        :param camelot._case_result.CaseResult result: Result object running in the test
        :param bool stop_on_failure: whether we stop on failure
        :returns: Result object
        :rtype: camelot._case_result.CaseResult
        """
        try:
            _log.info(_formatter.FMT_TEST_CASE_POST_TEST_ASSERTIONS.format(
                    index=result.index, name=self.name))
            self._post_test_assertions()
        # If we get here there was a error in the post_test_assertions.
        # First, if we get here when status is running, that means we need to mark it
        # failed or cancelled based on the exception.
        # If we get here when status has already been marked as failed/cancelled or whatever,
        # we want that to take precedence, so we'll either return or raise from the above
        # except block.
        except (Exception, KeyboardInterrupt) as e:  # pylint: disable=broad-except
            if result.status is _enums.ResultStatus.running:
                status = self._internal_determine_status(_enums.ResultStatus.failed, e)
            else:
                status = result.status
            exc = self._get_exception_record(status, _enums.CaseTransition.post_test_assertions)
            self._set_final_result(result, status, _enums.CaseTransition.post_test_assertions,
                                   exc_info=exc)
            if stop_on_failure and status is _enums.ResultStatus.failed:
                raise
        return result  # pylint: disable=lost-exception

    def _internal_teardown(self, result):
        """Execute the teardown.

        If the teardown raises an exception, who cares.

        :param camelot._case_result.CaseResult result: Result object running in the test
        :returns: Result object
        :rtype: camelot._case_result.CaseResult
        """
        # try out the teardown
        try:
            self._update_terminal('{} - Teardown'.format(self.name))
            _log.info(_formatter.banner(_formatter.FMT_TEST_CASE_TEARDOWN.format(
                    index=result.index, name=self.name)))
            self._teardown()
        except (Exception, KeyboardInterrupt) as e:  # pylint: disable=broad-except
            if result.status in (_enums.ResultStatus.running, _enums.ResultStatus.failed):
                status = self._internal_determine_status(_enums.ResultStatus.error, e)
            exc = self._get_exception_record(status, _enums.CaseTransition.teardown)
            self._set_final_result(result, status, _enums.CaseTransition.teardown, exc_info=exc)
        finally:
            return result  # pylint: disable=lost-exception
