"""Test Result object in use for camelot."""
import datetime
import uuid

from camelot import _enums


class CaseResult(object):
    """Test Result object.

    :param camelot.TestCase case: Actual Test Case
    :param int index: Index of test case within a suite. Will be None if test case
        is executed outside of a suite.
    :- camelot.enums.ResultStatus status: The result of the test.
    :- camelot.record.ExcInfo exc_info: The exception info from the last exception.
    :- camelot.record.ExcInfo test_exc_info: The exception info from _test and
        _post_test_assertions, if any.
    :- camelot.record.ExcInfo setup_exc_info: The exception info from _setup, if any.
    :- camelot.record.ExcInfo teardown_exc_info: The exception info from _teardown, if any.
    :- datetime.datetime start_time: The time the test started
    :- datetime.datetime stop_time: The time the test stopped
    :- uuid.UUID instance_id: A unique instance id
    :- list[camelot.case.TestStep] steps: The steps from the test
    """
    def __init__(self, case, index=None):
        self._case = case
        self.case_name = case.name
        self.index = index
        self.status = _enums.ResultStatus.pending
        self.exc_info = None
        self.test_exc_info = None
        self.setup_exc_info = None
        self.teardown_exc_info = None
        self.start_time = None
        self.end_time = None
        self.instance_id = uuid.uuid1()
        self.steps = None

    @property
    def case_path(self):
        """Fully qualified class name of test case.

        :rtype: str
        """
        return self._case.__module__ + '.' + self._case.__class__.__name__

    @property
    def exc_pretty_fmt(self):
        return '{}{}{}'.format(
            'test_exc_info: ' + self.test_exc_info.pretty_fmt() + '\n' if self.test_exc_info else '',
            'setup_exc_info: ' + self.setup_exc_info.pretty_fmt() + '\n' if self.setup_exc_info else '',
            'teardown_exc_info: ' + self.teardown_exc_info.pretty_fmt() if self.teardown_exc_info
            else ''
        )

    @property
    def execution_time(self):
        """Get the execution time based on start and end time.

        :return: Execution time of the test, end_time - start_time.
        :rtype: datetime.timedelta
        """
        if self.start_time and self.end_time:
            return self.end_time - self.start_time
        else:
            return datetime.timedelta(0)

    def __repr__(self):
        return 'CaseResult<index={}, name={}, status={}, execution_time={}>'.format(
            self.index,
            self.case_name,
            self.status.value.upper(),
            self.execution_time)
