"""Class with a context manager that is used to delay exceptions.

.. note:: The context manager behaves similarly to a try..except block. This means any code in the "with"
    block that is after a line that triggers an exception will not be executed.


Usage::

    from camelot import delay_exception
    exception_delayer = delay_exception.ExceptionDelayer('Name')

    # Delay any exceptions
    with exception_delayer.delay_exception():
        x = an_undefined_variable

    # Delay a specific exception
    with exception_delayer.delay_exception(exception_classes=(ZeroDivisionError, )):
        y = 2
        _ = 1/0
        x = 1

    # NOTE: x will not be assigned at this point since the exception was raised before that statement
    # was executed. However, y will be set.

    # Check if any errors occurred
    assert exception_delayer.has_errors()

    # Print the exception info
    print(exception_delayer.pretty_format())
"""

import contextlib

from camelot import _history


class ExceptionDelayer(object):
    """Delay exceptions and store the stack traces for later.

    :param str name: Name to use when reporting history
    """

    def __init__(self, name):
        self._name = name
        self._history = _history.History(name)
        self._exceptions = []

    @contextlib.contextmanager
    def delay_exception(self, exception_classes=(Exception, )):
        """Context manager for delaying exceptions.

        :param tuple[Any] exception_classes: A tuple of exceptions to catch. By default, all
            exceptions are caught.
        """
        try:
            yield
        except exception_classes:
            self._history.record('Critical Failure', exc_info=True)

    def pretty_format(self):
        """Pretty format the exception info.

        :rtype: str
        """
        return self._history.pretty_fmt()

    def has_errors(self):
        """Return True if there are any errors.

        :rtype: bool
        """
        return bool(self._history)

    def clear_errors(self, name=None):
        """Clear any accumulated errors.

        :param str name: Name to use when reporting history. If not given, use the one that was passed
            to the constructor.
        """
        self._history = _history.History(name or self._name)
