"""History module."""

import inspect

import autologger

from camelot import (
    _helpers,
    _record,
)

_log = autologger.get_sflogging(__name__)


class History(list, _helpers.PrettyLines):
    """History object.

    This class stores timestamped records for pretty printing information in the case of failure.

    :param str name: it is recommended to assign a name to history object.
    :param list stack: Stack trace from inspect.stack()
    """
    def __init__(self, name=None, stack=None):
        list.__init__(self)
        self.name = name
        stack = stack or inspect.stack()
        self.created_at = _helpers.fmt_called_from(stack)

    def record(self, msg, history=None, exc_info=None, stack=None, ts=None, log=False):
        """Record an event.

        :param str msg: the message to record
        :param camelot._history.History history: a History object to embed in the record
        :param tuple|bool exc_info: set to True to record sys.exc_info, or pass in sys.exc_info
            or ExcInfo object.
        :param float ts: leave None to record timestamp as time.time() or pass in float
        :param list stack: optional inspect.stack() for logging location of call
        :param bool log: If True, log debug information
        """
        stack = stack or inspect.stack()
        r = _record.Record(msg, history=history, exc_info=exc_info, stack=stack, ts=ts)
        self.append(r)
        if log:
            _log.debug('recorded w/stack: {} || {}'.format(stack[1][3], r.pretty_fmt()))

    def pretty_lines(self, indent=''):
        """Get the format in a list of lines.

        :param str indent: String used to indent lines
        :rtype: list[str]
        """
        name = self.name or ''
        if not self:
            return [indent + '# NO HISTORY FOR {}[{}]#'.format(name, self.created_at)]
        lines = [indent + '# History of {}[{}]'.format(name, self.created_at)]
        next_indent = _helpers.add_indent(indent)
        # noinspection PyTypeChecker
        for r in self:
            lines.extend(r.pretty_lines(next_indent))
        return lines
