"""Record classes for history object."""
import inspect
import sys
import time
import traceback
import types
from datetime import datetime

import six

from camelot import _helpers

tb_type = types.TracebackType


# TODO: VL-2052: this should be in logging
class ExcInfo(_helpers.PrettyLines):
    """Wrapper to store the exception info and format it more easily.

    :param tuple sys_exc_info: Exception info object from sys.exc_info()
    :param camelot._history.History history: History object
    """

    __hash__ = None

    def __init__(self, sys_exc_info=None, history=None):
        if sys_exc_info is None:
            sys_exc_info = sys.exc_info()
        if len(sys_exc_info) != 3:
            raise TypeError('Invalid exception info, must be of length 3: {}'.format(sys_exc_info))
        if any(n is None for n in sys_exc_info):
            raise TypeError('cannot have None values in sys_exc_info: {}'.format(sys_exc_info))
        if not isinstance(sys_exc_info[2], tb_type):
            raise TypeError('3rd index must be traceback type: {}'.format(sys_exc_info))
        self.etype, self.value, self.tb = sys_exc_info  # pylint: disable=unpacking-non-sequence
        if history is not None:
            self.value.history = history

    def reraise(self):
        """Reraise the stored exception."""
        six.reraise(*self.exc_info())

    def set_history(self, history):
        """Set a history object onto the exception.

        :param camelot._history.History history: History object
        """
        if hasattr(self.value, 'history'):
            exc = ValueError('history already exists on ExcInfo {}'.format(self))
            exc.history = self.value.history
            raise exc
        self.value.history = history

    def exc_info(self):
        """Return exc_info tuple.

        :rtype: tuple
        """
        return self.etype, self.value, self.tb

    def pretty_lines(self, indent=''):
        """Pretty format lines.

        :param str indent: String used to indent lines
        :rtype: list[str]
        """
        lines = [indent + '!!! Exception Occurred:']
        if hasattr(self.value, 'history'):
            lines.extend(self.value.history.pretty_lines(indent + '!'))
        for line in traceback.format_exception(self.etype, self.value, self.tb):
            # there can be newlines in the traceback message
            lines.extend(indent + '! ' + l.rstrip() for l in line.rstrip().split('\n'))
        lines.append(indent + '!!!')
        return lines

    def __eq__(self, other):
        """Equality for ExcInfo.

        :param camelot._record.ExcInfo other: Object to compare to
        :rtype: bool
        """
        if self.etype == other.etype and self.value == other.value and self.tb == other.tb:
            return True
        else:
            return False

    def __ne__(self, other):
        """The inverse of __eq__.

        :param camelot._record.ExcInfo other: Object to compare to
        :rtype: bool
        """
        return not self == other


class Record(_helpers.PrettyLines):
    """Record event for the history.

    :param str msg: the message to include
    :param camelot._history.History history: history object
    :param camelot._record.ExcInfo|bool|tuple exc_info: (if there was an exception) the exception
        info from sys.exc_info. Set to True to automatically retrieve from sys.exc_info.
    :param list stack: stack trace object from inspect.stack()
    :param float ts: the timestamp of the occurrence from time.time()
    """
    def __init__(self, msg, history=None, exc_info=None, stack=None, ts=None):
        if ts is None:
            ts = time.time()
        if not isinstance(ts, float):
            raise TypeError('ts must be type float: {}'.format(type(ts)))

        if isinstance(exc_info, ExcInfo):
            pass
        elif exc_info is True:
            exc_info = ExcInfo()
        elif exc_info:
            exc_info = ExcInfo(exc_info)
        else:
            assert exc_info in {None, False}

        self.msg = str(msg)
        self.history = history
        stack = stack or inspect.stack()
        self.called_from = _helpers.fmt_called_from(stack)
        self.exc_info = exc_info
        self.ts = ts

    def pretty_lines(self, indent=''):
        """Pretty formatted lines.

        :param str indent: String used to indent lines
        :rtype: list[str]
        """
        lines = []
        lines.append(indent + '> UTC[{}] LOC[{}] [{}]: {}{}'.format(
            datetime.utcfromtimestamp(self.ts),
            datetime.fromtimestamp(self.ts),
            self.called_from,
            '<ERROR> ' if self.exc_info else '',
            self.msg))
        next_indent = _helpers.add_indent(indent)
        if self.history:
            lines.extend(self.history.pretty_lines(next_indent))
        if self.exc_info:
            lines.extend(self.exc_info.pretty_lines(indent))
        return lines
