"""Camelot test suite definition: Use as the base class for all test suites."""

import abc
import collections
import datetime
import inspect
import json
import multiprocessing
import os
import sys
import threading

import autologger
import logbook
import pytz
import sf_artifacts
import sf_batch
import six
from _pytest import outcomes

from camelot import (
    _case_result,
    _enums,
    _errors,
    _formatter,
    _history,
    _suite_result,
)

_log = autologger.get_sflogging(__name__)

TestRun = collections.namedtuple('TestRun', ['case', 'result'])

DEFAULT_ARTIFACTS_DIR = os.path.expanduser('~/vl_artifacts')
# Environment variable that overrides the artifacts dir
ARTIFACTS_DIR_ENV_VAR = 'ARTIFACTS_PATH'
REPORT_FILE = 'report.json'
SUMMARY_FILE = 'summary.log'
TEST_LOG_FILE = 'test.log'


@six.add_metaclass(abc.ABCMeta)
class TestSuite(object):
    """A test suite containing setup and teardown methods as well as the test cases to be run.

    Child class must override the __init__ method and all test cases should be added via the
    add_tests method.

    :param camelot.models.SuiteResources resources: Resources data model for all test cases
    :param camelot.models.SuiteConfig config: Suite configuration data model

    **Instance variables:**

        * list[camelot.case.TestCase] tests: The test cases to run
        * camelot.enums.ResultStatus status: the status of the suite.
        * history.History history: The suite history object
        * str name: name of the suite
        * list[camelot.TestRun] test_runs: list of namedtuple[TestCase, CaseResult]
        * camelot.SuiteResult result: Suite result object

    .. automethod:: _post_test_case_execute
    """
    _skip_artifact_dir = False  # this is used for unit tests to skip the artifact directory creation
    # this is used for unittests or tests that don't ever want to push to groot/qtest
    _skip_remote_reporting = False

    ##############################################
    # Developer API

    statuses = _enums.ResultStatus

    def __init__(self, resources, config):
        self.resources = resources
        self.config = config
        self.tests = []
        self.status = _enums.ResultStatus.pending
        self.history = None
        self.name = self.__class__.__name__
        self.full_name = self.__module__ + '.' + self.name
        self.test_runs = []
        self.result = None
        self.summary_file = None
        # Validate resources and config
        self.resources.validate()
        self.config.validate()

        self._initialize_artifact_credentials()

        # Where to store artifacts locally. Will be None if nothing should be stored.
        self.local_artifacts_dir = self._get_local_artifacts_dir()

        if self.local_artifacts_dir is not None:
            self._create_local_artifacts_dir()
            self.summary_file = os.path.join(self.local_artifacts_dir, SUMMARY_FILE)
            self._init_logging_file_handler()
            _log.info('Local artifacts will be stored at: {}'.format(self.local_artifacts_dir))

        if not self._skip_artifact_dir:
            # dump the resources and config as an artifact for reference in triage
            artifact_manager = sf_artifacts.get_and_set_global_manager(
                    self.config.artifact_credentials.serialize())
            artifact_manager.store_text(json.dumps(self.resources.serialize(), indent=4),
                                        prefix='suite_resources-', suffix='.json')
            artifact_manager.store_text(json.dumps(self.config.serialize(), indent=4),
                                        prefix='suite_config-', suffix='.json')

    @abc.abstractmethod
    def setup(self):
        """Suite setup."""

    def teardown(self):
        """Optional suite teardown."""

    ##############################################
    # User API

    def execute(self):  # pylint: disable=too-complex
        """Execute the suite.

        Each time the suite is executed the test_runs will be reset with new results.
        The result will also be recreated each time.

        :returns: A list of test results
        :rtype: tuple[camelot.TestRun]
        """
        self.history = _history.History(self.name)
        self.create_test_runs()  # resets self.test_runs if executing again
        self._init_result()
        # Run _preconditions before setting status to "running"
        self._verify_runtime()
        self.result.start_time = datetime.datetime.now(pytz.utc)
        self.result.status = self.status = _enums.ResultStatus.running
        self._record(_formatter.FMT_SUITE_SETUP.format(name=self.name))

        # create summary file with suite name at top
        if self.summary_file is not None:
            with open(self.summary_file, 'w') as summary_file_handle:
                summary_file_handle.write('Suite Name: {}\n\n'.format(self.name))
        try:
            self.setup()
            self.run_tests()
        except (KeyboardInterrupt, SystemExit):
            self.result.status = self.status = _enums.ResultStatus.cancelled
            self._check_execute_failure(sys.exc_info())
        except (Exception, outcomes.Failed):  # pylint: disable=broad-except
            if self.status is _enums.ResultStatus.running:
                # self.status will either be failed, error, or running. If failed, keep failed. If
                # running, force to error. If config.stop_on_failure is true, and an exception is raised,
                # suite status will always be error.
                self.result.status = self.status = _enums.ResultStatus.error
            self._check_execute_failure(sys.exc_info())
        self._record(_formatter.FMT_SUITE_TEARDOWN.format(name=self.name))
        try:
            if self.config.run_suite_teardown:
                self.teardown()
            else:
                _log.info('configured to not execute suite-teardown')
        except (KeyboardInterrupt, SystemExit):
            self.result.status = self.status = _enums.ResultStatus.cancelled
            raise
        except (Exception, outcomes.Failed):  # pylint: disable=broad-except
            self.result.status = self.status = _enums.ResultStatus.error
            self._record('teardown FAILED', error=True)
        finally:
            self.report_results()

        return self.test_runs

    def report_results(self):
        """Report all suite results."""
        self.create_final_report()
        self.list_failures()
        if self.local_artifacts_dir is not None:
            _log.info('Output files dir: {}'.format(self.local_artifacts_dir))
        if not self._skip_artifact_dir and self.config.store_report:
            self._write_report()
            self._write_to_summary_file(str(self.result))

    def create_final_report(self):
        """Generate the final report."""

        # Ensure there are no hanging threads and join check for errors.
        self._log_threads(check_for_errors=True)

        # Set end time and final status
        self.result.end_time = datetime.datetime.now(pytz.utc)
        if self.status is _enums.ResultStatus.running:
            # we didn't hit an error, didn't fail a test case and suite teardown succeeded
            self.result.status = self.status = _enums.ResultStatus.passed

        _log.info(_formatter.banner('Final Report'))
        _log.info(self.result)

        # Print a summary of failed test cases without the stack trace
        failures = [r.result for r in self.result.failed_test_runs if r.result is not None]
        if failures:
            _log.info('Failed test cases')
            for f in failures:
                _log.info(f)

        _log.info('\n\n' + self.history.pretty_fmt())

    def list_failures(self):
        """List any test case failures with stack traces."""
        failures = [r.result for r in self.result.failed_test_runs if r.result is not None]
        if failures:
            failure_strings = ['\n{}\n{}\n'.format(f, f.exc_pretty_fmt) for f in failures]
            # Print all failures with stack traces
            _log.info(_formatter.banner('The following tests failed'))
            _log.info('\n{}'.format((''.join(failure_strings))))

    def run_tests(self):
        """Run all test cases."""
        for run in self.test_runs:
            if run.result.status is _enums.ResultStatus.skipped:
                msg = _formatter.FMT_TEST_CASE_SKIP.format(index=run.result.index, name=run.case)
                _log.info(_formatter.banner(msg))
                self._record_result_in_summary_file(run)
                continue
            self._pre_test_case_execute(run)
            # note: if config.stop_on_failure is set, this will always raise an exception on failure
            run.case.execute(stop_on_failure=self.config.stop_on_failure, result=run.result)

            _log.info(_formatter.FMT_TEST_CASE_STATUS.format(
                    index=run.result.index, name=run.case, status=run.result.status.value))
            if run.result.status in {_enums.ResultStatus.failed, _enums.ResultStatus.error,
                                     _enums.ResultStatus.cancelled}:
                self.result.status = self.status = run.result.status
                # If the status is error, the suite will stop executing, leaving the status error.
                if run.result.status in {_enums.ResultStatus.error, _enums.ResultStatus.cancelled}:
                    self._record_result_in_summary_file(run)
                    run.result.exc_info.reraise()
            self._log_threads()
            self._record_result_in_summary_file(run)
            msg = _formatter.FMT_TEST_CASE_POST_TEST_EXECUTION.format(
                    index=run.result.index, name=run.case)
            _log.info(_formatter.banner(msg))
            self._post_test_case_execute(run)

    def _post_test_case_execute(self, run):
        """Optional method that is run after a test case has executed.

        Override this if you need to perform some operation after every test case.

        :param camelot.TestRun run: TestRun object of executed test case
        """

    def _pre_test_case_execute(self, run):
        """Optional method that is run before a test case has executed.

        Override this if you need to perform some operation before every test case.

        :param camelot.TestRun run: TestCase object
        """

    def add_tests(self, tests):
        """Add the tests to the list of tests. Tests should be initialized.

        :param typing.Iterable[camelot._case.TestCase] tests: list or tuple of tests you
            would like to add to the suite.
        """
        self.tests.extend(tests)

    def get_final_report(self):
        """Get the suite report.

        :rtype: dict
        """
        return self.result.generate_report()

    ##############################################
    # Internal API

    def _check_execute_failure(self, exc_info):
        """Checks the status and raises if necessary.

        :param tuple exc_info: Tuple from sys.exc_info()
        """
        if self.status is _enums.ResultStatus.cancelled or self.config.stop_on_failure:
            # we totally blew up
            if self.status is not _enums.ResultStatus.cancelled:
                self.history.record('Critical Failure', exc_info=exc_info)
                _log.info(_formatter.banner('Critical Failure, here is the entire history:'))
                _log.error('\n' + self.history.pretty_fmt())
            self.report_results()
            six.reraise(*exc_info)
        else:
            self.history.record('Critical Failure', exc_info=exc_info)
            _log.info(_formatter.banner('Critical Failure, here is the entire history:'))
            _log.error('\n' + self.history.pretty_fmt())

    def _record_result_in_summary_file(self, run):
        """Record a descriptive line to the summary file for each run in run_tests.

        :param camelot.TestRun run: TestRun to record in the summary file
        """
        if self.summary_file is not None:
            msg = '{}'.format(run.result)
            self._write_to_summary_file(msg)

    def _write_to_summary_file(self, msg):
        """Writes a line to the file.

        :param str msg: the line to write to self.summary_file
        """
        with open(self.summary_file, 'a') as summary_file_handle:
            summary_file_handle.write(msg + '\n')

    def _write_report(self):
        """Write a report file."""
        report = self.get_final_report()
        report_file = os.path.join(self.local_artifacts_dir, REPORT_FILE)
        with open(report_file, 'w') as report_file_handle:
            json.dump(report, report_file_handle, indent=4)

    def _record(self, stage, error=False):
        """Record event in suite history.

        :param str stage: Description of the the stage in the suit at which the event occurred.
                          e.g. Teardown/Setup/Execution
        :param bool error: Whether there was an error
        """
        if error:
            msg = 'ERROR in ' + stage
        else:
            msg = stage
        self.history.record(msg, exc_info=error, stack=inspect.stack(), log=True)
        _log.info(_formatter.banner(msg))

    def create_test_runs(self):
        """Build the test run list based on the indices in the config."""
        if self.test_runs:
            # if you want to run a suite again
            self.test_runs = []
        # pylint: disable=range-builtin-not-iterating
        skip_indices = self.config.skip_indices
        run_indices = self.config.run_indices if self.config.run_indices else range(len(self.tests))
        start_index = self.config.start_index or 0
        # have to check 'is not None', since it could be 0
        stop_index = self.config.stop_index if self.config.stop_index is not None else len(self.tests)
        skip_indices.extend(range(start_index))  # skip tests up to the start index
        skip_indices.extend(range(stop_index, len(self.tests)))  # skip tests from the stop to the end
        for i in run_indices:
            try:
                run = TestRun(self.tests[i], _case_result.CaseResult(self.tests[i], i))
            except IndexError:
                raise _errors.TestRuntimeError('Run index: {} is not valid'.format(i))
            if i in skip_indices:
                run.result.status = _enums.ResultStatus.skipped
            self.test_runs.append(run)
        # pylint: enable=range-builtin-not-iterating

    def _verify_runtime(self):
        """Verify that there are tests to run and that those tests can run by calling _preconditions."""

        pending_runs = [run for run in self.test_runs if run.result.status is
                        _enums.ResultStatus.pending]
        pending_cases = [run.case for run in pending_runs]
        if not pending_runs:
            self.status = _enums.ResultStatus.error
            raise ValueError('No tests to run!')

        _log.info('{} will run the following tests: {}'.format(self.name, pending_cases))
        self._record('All Test Case Preconditions')
        for run in pending_runs:
            run.case.preconditions(run.result.index)

    def _init_result(self):
        """Initialize a new result object for the suite."""
        self.result = _suite_result.SuiteResult(self.name, self.full_name, self.resources, self.config,
                                                self.test_runs)

    @staticmethod
    def _log_threads(check_for_errors=False):
        """Log threads and processes active.

        :param bool check_for_errors: whether to check for hanging threads and join on them.
        """
        _log.debug('Suite threads info (should be len=1)  : {}'.format(
                threading.enumerate()))
        # pylint gets a false positive not-callable error in python 3
        _log.debug('Suite processes info (should be len=0): {}'.format(
                multiprocessing.active_children()))  # pylint: disable=not-callable

        if check_for_errors:
            for thread in threading.enumerate():
                if isinstance(thread, sf_batch.SFThread):
                    thread.join(timeout=60)
                    if thread.is_alive():
                        raise _errors.TestRuntimeError('Hanging thread: {}'.format(thread.ident))

                    sf_batch.check_for_error_and_raise_exception([thread])

    def _initialize_artifact_credentials(self):
        """Update artifact credentials on the config object."""
        # If we have artifact creds from AT2, we don't need to update
        # artifact_credentials with a base dir, and local will be None.
        # If running from the shell or driver, the default will have all artifact_cred attrs set
        # to None (the AT2 attrs specifically). In this case, set local to True and set base_dir.
        if self.config.artifact_credentials.username is None:
            self.config.artifact_credentials.local = True
            if os.environ.get(ARTIFACTS_DIR_ENV_VAR):
                base_dir = os.getenv(ARTIFACTS_DIR_ENV_VAR)
            else:
                base_dir = self.config.artifact_credentials.base_dir
                base_dir = base_dir if base_dir else DEFAULT_ARTIFACTS_DIR
            run_dir = '{}-{}'.format(self.name,
                                     datetime.datetime.now().strftime('%Y-%m-%dT%H.%M.%S'))
            our_base = os.path.join(base_dir, run_dir)
            self.config.artifact_credentials.base_dir = our_base

    def _get_local_artifacts_dir(self):
        """Get directory where local artifacts are written.

        Return None if artifacts should not be written locally.

        :rtype: str|None
        """
        if not self._skip_artifact_dir and self.config.artifact_credentials.local is True:
            # Running outside of At2
            return self.config.artifact_credentials.base_dir
        else:
            return None

    def _create_local_artifacts_dir(self):
        """Create tmp directory in which to store artifacts."""
        os.makedirs(self.local_artifacts_dir)

    def _init_logging_file_handler(self):
        """Initialize a file handler for logging."""
        test_log_file = os.path.join(self.local_artifacts_dir, TEST_LOG_FILE)
        file_handler = logbook.FileHandler(
                test_log_file, bubble=True,
                format_string=autologger.FORMAT_STRING
        )
        file_handler.push_application()
