"""Test Suite Result object."""
import datetime
import uuid

from camelot import (
    _enums,
    _formatter,
)


class SuiteResult(object):
    """Test Suite Result object.

    :param str suite_name: Name of the suite
    :param str suite_path: Full python path to the suite
    :param camelot.models.SuiteResources resources: Resources used by the suite
    :param camelot.models.SuiteConfig config: Config used by the suite
    :param list[camelot.TestRun] test_runs: Test runs are a list of
        namedtuple[TestCase, CaseResult]
    :- datetime.datetime start_time: Time the suite started
    :- datetime.datetime end_time: Time the suite finished
    """
    def __init__(self, suite_name, suite_path, resources, config, test_runs):
        self.suite_name = suite_name
        self.suite_path = suite_path
        self.config = config
        self.resources = resources
        self.test_runs = test_runs
        self.status = _enums.ResultStatus.pending
        self.instance_id = uuid.uuid1()
        self.start_time = None
        self.end_time = None
        self.monitor_failures = []

    def __repr__(self):
        return 'SuiteResult<name={}, status={}, execution_time={}>'.format(self.suite_name,
                                                                           self.status.value.upper(),
                                                                           self.execution_time)

    def _filter_runs(self, status):
        """Filter the runs from the status.

        :param camelot._enums.ResultStatus status: Status to filter on
        :return: list of matching test_runs
        :rtype: list[camelot.TestRun]
        """
        return [run for run in self.test_runs if run.result.status is status]

    @property
    def failed_test_runs(self):
        """Failed test runs.

        :returns: list of failed test runs
        :rtype: list[camelot.TestRun]
        """
        return self._filter_runs(_enums.ResultStatus.failed)

    @property
    def errored_test_runs(self):
        """Errored test runs.

        :returns: list of errored test runs
        :rtype: list[camelot.TestRun]
        """
        return self._filter_runs(_enums.ResultStatus.error)

    @property
    def skipped_test_runs(self):
        """Skipped test runs.

        :returns: list of skipped test runs
        :rtype: list[camelot.TestRun]
        """
        return self._filter_runs(_enums.ResultStatus.skipped)

    @property
    def passed_test_runs(self):
        """Passed test runs.

        :returns: list of passed test runs
        :rtype: list[camelot.TestRun]
        """
        return self._filter_runs(_enums.ResultStatus.passed)

    @property
    def pending_test_runs(self):
        """Pending test runs.

        :returns: list of pending test runs
        :rtype: list[camelot.TestRun]
        """
        return self._filter_runs(_enums.ResultStatus.pending)

    @property
    def build(self):
        """Build version used in the suite.

        :returns: build as the version string, e.g. '9.0.0.1544'
        :rtype: str
        """
        try:
            # VL-370 will break this.
            return self.resources.clusters[0].config.build.version
        except (IndexError, AttributeError):
            # its possible they don't have clusters
            # I think we should treat builds as either a config or a resource
            return None  # maybe '0.0.0.0'

    @property
    def artifact_credentials(self):
        """Artifact credentials dict.

        :returns: artifact credentials dict
        :rtype: dict
        """
        return self.config.artifact_credentials.serialize()

    @property
    def execution_time(self):
        """Execution time of the test.

        :return: Execution time of the test, end_time - start_time.
        :rtype: datetime.timedelta
        """
        if self.start_time and self.end_time:
            return self.end_time - self.start_time
        else:
            return datetime.timedelta(0)

    def generate_report(self):
        """Generate the standard suite report.

        :rtype: dict
        """
        return {
            # NOTE: Increment report_version each time this data changes
            'report_version': 5,
            'build': self.build,
            'suite_path': self.suite_path,
            'end_time': self.end_time.isoformat() if self.end_time else None,
            'failed_testcases': len(self.failed_test_runs)+len(self.errored_test_runs),
            'num_suites': 1,
            'passed_testcases': len(self.passed_test_runs),
            'queued_testcases': len(self.pending_test_runs),
            'skipped_testcases': len(self.skipped_test_runs),
            'start_time': self.start_time.isoformat() if self.start_time else None,
            'status': self.status.value,
            'suite_name': self.suite_name,
            'execution_time': _formatter.format_time_delta(self.execution_time),
            'total_testcases': len(self.test_runs),
            'artifact_credentials': self.artifact_credentials,
            'test_runs': [{'name': run.result.case_name,
                           'status': run.result.status.value,
                           'exception_info': run.result.exc_pretty_fmt,
                           'execution_time': _formatter.format_time_delta(run.result.execution_time),
                           'start_time':
                               run.result.start_time.isoformat() if run.result.start_time else None,
                           'end_time': run.result.end_time.isoformat() if run.result.end_time else None,
                           'case_path': run.result.case_path,
                           'instance_id': str(run.result.instance_id),
                           'steps': [{'index': step.index,
                                      'action': step.action,
                                      'expected_result': step.expected_result}
                                     for step in run.result.steps] if run.result.steps else None}
                          for run in self.test_runs],
            'changesets': [c.serialize() for c in self.config.changesets],
            'resources': self.resources.serialize(),
            'config': self.config.serialize(),
            'instance_id': str(self.instance_id),
            'monitor_failures': [m.serialize() for m in self.monitor_failures],
        }
