"""Filter log helpers for the Valence driver."""
import os
import re
import shutil
from tempfile import NamedTemporaryFile

import autologger

import camelot

_log = autologger.get_sflogging(__name__)


def filter_and_create_log(log_filter, test_log_path):
    """Apply a single filter and create the new log file.

    :param camelot.models.CamelotLogFilter log_filter: A single camelot log filter
    :param str test_log_path: Path to the full test.log to filter
    """
    if not log_filter.create_file:
        return
    match_regexes = log_filter.match_regexes
    exclude_regexes = log_filter.exclude_regexes
    if log_filter.output_log == camelot.TEST_LOG_FILE:
        new_output_log = 'new.' + camelot.TEST_LOG_FILE
        log_filter.output_log = new_output_log
        _log.warn('You requested a filtered log to be named the same as the main test log {}. Please '
                  'go to your suite_config.json and change the name. It is being force changed to {}.'
                  .format(camelot.TEST_LOG_FILE, new_output_log))

    filtered_log = os.path.join(os.path.dirname(test_log_path), log_filter.output_log)
    with open(test_log_path) as input_file:
        with NamedTemporaryFile(mode='w', dir=os.path.dirname(filtered_log)) as outfile:
            for line in input_file:
                filtered_line = process_input_line(line, match_regexes, exclude_regexes)
                if filtered_line:
                    outfile.write(filtered_line)
            outfile.flush()
            shutil.copyfile(outfile.name, filtered_log)


def process_input_line(input_line, match_regexes, exclude_regexes):
    """Apply filters to a single line.

    :param str input_line: The line to apply the filters on
    :param list[str] match_regexes: Regex filters to display the line if it matches
    :param list[str] exclude_regexes: Regex filters to exclude if it matches, this takes precedence
    :rtype: str|None
    """
    match = None
    exclude = None
    exclude_patterns = '|'.join(exclude_regexes)
    match_patterns = '|'.join(match_regexes)
    if match_patterns:
        match = re.search(match_patterns, input_line)
    if exclude_patterns:
        exclude = re.search(exclude_patterns, input_line)
    if not match and match_patterns:
        input_line = None
    if exclude:
        input_line = None
    return input_line
