"""Logic for list command.

This command lists test cases, suites, and modules.
"""
from __future__ import print_function

import importlib
import inspect
import json

import camelot
from camelot import models
from camelot.driver import (
    _constants,
    _main,
)
from camelot.models import at2_enums


def get_object_name(obj):
    """Get name of a python object.

    :param types.ModuleType|six.class_types obj: A python class or module
    :return: Name of object
    :rtype: str
    """
    if inspect.isclass(obj):
        return obj.__module__ + '.' + obj.__name__
    else:
        return obj.__name__


def get_doc(obj):
    """Get docstring of object.

    :param types.ModuleType|six.class_types obj: A python class or module
    :return: Docstring of object
    :rtype: str
    """
    result = ''
    doc = inspect.getdoc(obj)
    if doc:
        for line in doc.split('\n'):
            result += '    {}\n'.format(line)
    return result


def get_mock_resources_dm():
    """Get the most accurate at2 based resources.json.

    Contains:
        * 2 cluster
        * 2 clients
        * 5 fdvas
        * 5 nodes
        * 5 hypervisors
        * 5 compute nodes
        * 0 virtual networks
        * 5 ontaps
        * 5 networks
        * 0 ip ranges
        * None virtual ips

    :return: Fake resources dm.
    :rtype: camelot.models.SuiteResources
    """

    cluster = models.ClusterSuiteResource.get_mock_object(include_optional=True)
    cluster.config.virtual_ips = None  # pylint: disable=no-member
    cluster.config.options = None  # pylint: disable=no-member

    # Clients can't be the same because of checks in the ManageConnectionsComponent
    client = models.ClientSuiteResource.get_mock_object(include_optional=True)
    client.model.resource_type_name = None  # pylint: disable=no-member
    client.model.os = at2_enums.ClientOS.ubuntu_14_04  # pylint: disable=no-member
    client2 = models.ClientSuiteResource.get_mock_object(include_optional=True)
    client2.model.resource_type_name = None  # pylint: disable=no-member
    client2.model.os = at2_enums.ClientOS.ubuntu_14_04  # pylint: disable=no-member

    fdva = models.ManagementNodeSuiteResource.get_mock_object(include_optional=True)
    fdva.model.resource_type_name = None  # pylint: disable=no-member
    node = models.NodeSuiteResource.get_mock_object(include_optional=True)
    node.model.location_id = 1  # pylint: disable=no-member
    ontap = models.OntapSuiteResource.get_mock_object(include_optional=True)
    compute_node = models.ComputeNodeSuiteResource.get_mock_object(include_optional=True)
    network = models.NetworkSuiteResource.get_mock_object(include_optional=True)
    ip = models.IpSuiteResource.get_mock_object(include_optional=True)
    hypervisor = models.HypervisorSuiteResource.get_mock_object(include_optional=True)
    return models.SuiteResources.create(
            clients=[client, client2],
            clusters=[cluster for _ in range(5)],
            fdvas=[fdva for _ in range(5)],
            management_nodes=[fdva for _ in range(5)],
            hypervisors=[hypervisor for _ in range(5)],
            ip_ranges=[],
            ips=[ip for _ in range(5)],
            networks=[network for _ in range(5)],
            nodes=[node for _ in range(5)],
            ontaps=[ontap for _ in range(5)],
            virtual_networks=[],
            compute_nodes=[compute_node for _ in range(5)],
            virtual_ips=None,
    )


def get_resources_dm():
    """Get default suite resources data model.

    :rtype: camelot.models.SuiteResources
    """
    resources_file = _main.get_config_file(_constants.DEFAULT_SUITE_RESOURCES_FILE)
    if resources_file is None:
        resources = get_mock_resources_dm()
    else:
        with open(resources_file) as r:
            resources = models.SuiteResources(json.load(r))
    return resources


class ListCommand(object):
    """Lists modules, suites and test cases.

    :param str path: Path to module, package or suite
    :param bool verbose: If True, print out docstrings of members
    :param camelot.models.SuiteConfig config: Suite config data model
    """

    def __init__(self, path, verbose=False, config=None):
        self.path = path
        self.verbose = verbose

        # self.get_members sets the following variables
        self._members_populated = False
        self.modules = []
        self.suites = []
        self.cases = []
        self.steps = []

        if config is None:
            self.config = models.SuiteConfig()
        else:
            self.config = config
        self.resources = get_resources_dm()

        # self.set_object sets the following variables
        self.is_suite = False
        self.is_case = False
        self.object = None
        self.set_object()

    def _print_invalid_module_message(self):
        """Print a message declaring that self.path is invalid."""
        print('"{}" is not a valid module or class'.format(self.path))

    def set_object(self):
        """Attempt to import the object at self.path and set it to self.object."""
        try:
            self.object = importlib.import_module(self.path)  # type: type
        except ImportError:
            # Path may be a class. Try to get the class object
            if '.' not in self.path:
                self._print_invalid_module_message()
                return
            try:
                self.object = _main.get_class_from_path(self.path)  # type: type
                if issubclass(self.object, camelot.TestSuite):
                    self.is_suite = True
                if issubclass(self.object, camelot.TestCase):
                    self.is_case = True
            except AttributeError:
                self._print_invalid_module_message()

    def _get_cases_from_suite(self, suite):
        """In order to get a list of test cases, we need to instantiate the suite.

        :param six.class_types suite: Reference to a suite class
        """
        # Prevents vl_artifact directory from being created, restore flag when done
        save_skip_artifacts = suite._skip_artifact_dir  # pylint: disable=protected-access
        suite._skip_artifact_dir = True   # pylint: disable=protected-access
        suite_instance = suite(self.resources, self.config)
        suite._skip_artifact_dir = save_skip_artifacts  # pylint: disable=protected-access

        if not suite_instance.tests:
            print('No tests found for suite')
        for test in suite_instance.tests:
            self.cases.append(type(test))

    def get_members(self):  # pylint: disable=too-complex
        """Gets the members of self.object and sets relevant instance variables.

        This method sets self.cases, self.suites and self.modules.
        """
        if self._members_populated:
            return
        if self.is_suite:
            self._get_cases_from_suite(self.object)
        elif self.is_case:
            self.steps = camelot.get_steps_from_case(self.object)
        else:
            members = inspect.getmembers(self.object)
            for member in members:
                m_type = member[1]
                if inspect.isclass(m_type):
                    if issubclass(m_type, camelot.TestSuite):
                        self.suites.append(member[1])
                    elif issubclass(m_type, camelot.TestCase):
                        self.cases.append(member[1])
                elif inspect.ismodule(m_type):
                    if not member[0].startswith('_'):
                        self.modules.append(member[1])
        self._members_populated = True

    def print_members(self):  # pylint: disable=too-complex
        """Prints out members of self.object."""
        self.get_members()
        if self.suites:
            print(' Suites '.center(60, '='))
            for suite in self.suites:
                print(get_object_name(suite))
                if self.verbose:
                    print(get_doc(suite))
            print('')
        if self.cases:
            if self.is_suite:
                print(' Cases in {} '.format(self.object.__name__).center(60, '='))
            else:
                print(' Cases '.center(60, '='))
            for i, case in enumerate(self.cases):
                output = ''
                if self.is_suite:
                    # When listing a suite, print the index of each test case
                    output += '{} '.format(i)
                output += get_object_name(case)
                print(output)
                if self.verbose:
                    print(get_doc(case))
            print('')
        if self.steps:
            print(' Steps in {} '.format(self.object.__name__).center(60, '='))
            for step in self.steps:
                print(step)
        elif self.is_case:
            print(' No Steps in {} '.format(self.object.__name__).center(60, '='))
        if self.modules:
            print(' Modules '.center(60, '='))
            for module in self.modules:
                print(get_object_name(module))
                if self.verbose:
                    print(get_doc(module))
            print('')
