"""Main logic for driver script."""
from __future__ import print_function

import copy
import csv
import importlib
import itertools
import json
import multiprocessing
import os
import pdb
import sys
import traceback

import autologger
import repo_utils
import requests
import sf_models
import six  # pylint: disable=unused-import
from at2_sdk.at2 import LegacyAt2Api
from requests import auth

import camelot
from camelot import (
    groot,
    models,
)
from camelot.driver import (
    _constants,
    _filter_log,
)
from camelot.models import at2_constants

_log = autologger.get_sflogging(logger=__name__)

BUILD_TEMPLATE = sf_models.Build(dict(build_id=0,
                                      release='RELEASE-PATCHx',
                                      version='VERSION.LIKE.THIS.ONE',
                                      element_id=0,
                                      repository_name='RELEASE-PATCHx'))
CONFIG = 'config'
HEALTHTOOLS_VERSION = 'healthtools_version'
HEALTHTOOLS_DOWN_VERSION = 'healthtools_down_versions'
UPGRADE_BUILDS = 'upgrade_builds'
UPDATE_REPO = 'update_repo'


def get_changesets(changeset_modules):
    """Get changesets associated with each relevant repository.

    :param list[str] changeset_modules: A list of modules that are in each repository for which
        the changeset should be stored.
    :rtype: list[camelot.models.RepoChangeset]
    """
    changesets = []

    for module_name in changeset_modules:
        try:
            mod = importlib.import_module(module_name)
        except ImportError:
            _log.debug('Cannot import module "{}".'
                       'Changeset will not be stored'.format(module_name))
            continue
        path = os.path.dirname(mod.__file__)
        repo = repo_utils.get_repo(path)
        if isinstance(repo, repo_utils.NoRepo):
            _log.debug('"{}" module is not in a repo.'
                       'Changeset will not be stored'.format(module_name))
            continue
        changeset = models.RepoChangeset.create(
            changeset_id=repo.get_changeset_id(),
            repo_name=repo.get_name())
        changesets.append(changeset)

    return changesets


def get_ip_resources(steps):
    """Get the ip information from a ReserveIp and SetupCluster step.

    :param list[dict] steps: steps to get resources from
    :return: list of resources for the requested steps
    :rtype: list[dict]
    """
    ips = []
    for step in steps:
        # Get ip resources from ReserveIp step
        if step[at2_constants.STEP_NAME] == at2_constants.RESERVE_IP:
            ips.append(step[at2_constants.OUTPUT_VALUE])
        # Get ip resources from SetupCluster step
        elif step[at2_constants.STEP_NAME] == at2_constants.SETUP_CLUSTER:
            setup_ips = []
            setup_inputs = step[at2_constants.INPUTS]
            setup_ips.extend([i[at2_constants.VALUE] for i in setup_inputs
                              if i[at2_constants.DATA_TYPE_NAME] == at2_constants.IP])

            # _at2_constants.VALUE for SetupCluster inputs are missing the dataTypeName key,
            # so add it here
            for ip in setup_ips:
                ip[at2_constants.DATA_TYPE_NAME] = at2_constants.IP

            ips.extend(setup_ips)

    return ips


def get_raw_resources(steps, step_name):
    """Get the resource ids from a ReserveResources step (i.e. nodes, fdvas, clients, etc).

    :param list[dict] steps: steps to get resources from
    :param str step_name: step name to filter by
    :return: list of resources for the requested steps
    :rtype: list[dict]
    :raises RuntimeError: If a step has an output value of None
    """
    output = [s[at2_constants.OUTPUT_VALUE] for s in steps if s[at2_constants.STEP_NAME] == step_name]
    if None in output:
        raise RuntimeError('Step: {} failed.'.format(step_name))
    if step_name in [at2_constants.SETUP_CLUSTER, at2_constants.CREATE_CLUSTER]:
        output = [output]
    # pull out info and flatten
    return list(n for n in itertools.chain.from_iterable(output))


def get_task_resources(task_id):
    """Get the resources from an AT2 task.

    :param int task_id: task ID to gather resources from
    :return: Suite resources data model
    :rtype: camelot.models.SuiteResources
    """
    at2_api = LegacyAt2Api(username=at2_constants.AUTO_USER, password=at2_constants.AUTO_PASS)
    task = at2_api.get_task_instance_by_id(task_id)
    return _get_task_resources(task)


def get_docker_resources(mvip):
    """Get the resources from a docker cluster.

    :param str mvip: Mvip of the docker cluster, including the port
    :return: Suite resources data model
    :rtype: camelot.models.SuiteResources
    """
    # Import inside this function so vl can be used without needed to depend on the full valence package
    from valence import functional_models
    cluster_data = {
        'mvip': mvip,
        'username': 'admin',
        'password': 'admin',
        'api_endpoint': '10.0',
    }
    initial_cluster_dm = sf_models.ResourceCluster(cluster_data)
    cluster_fm = functional_models.Cluster(initial_cluster_dm)
    cluster_fm._validate_responses = False  # pylint: disable=protected-access

    # Populate cluster data model with fields from cluster info
    cluster_info = cluster_fm.get_cluster_info().serialize()

    # The mvip from get_cluster_info is not accurate, so use the provided one.
    del cluster_info['mvip']
    cluster_data.update(cluster_info)
    cluster_dm = sf_models.ResourceCluster(cluster_data)
    cluster_dm.api_endpoint = cluster_fm.get_current_version()

    # Create node data models
    node_dms = cluster_fm.list_all_nodes()
    suite_resources = models.SuiteResources()
    cluster_config = models.ClusterSuiteConfig.create(build=sf_models.Build(), upgrade_builds=[])
    suite_resources.clusters = [models.ClusterSuiteResource.create(model=cluster_dm,
                                                                   config=cluster_config)]
    rs_node_dms = [sf_models.ResourceNode(n.serialize()) for n in node_dms]
    for node in rs_node_dms:
        node.model = sf_models.NodeType.STANDARD_TYPES.SFDOCKER
    node_config = models.NodeSuiteConfig.create(build=sf_models.Build())
    suite_rs_node_dms = [models.NodeSuiteResource.create(model=n, config=node_config) for
                         n in rs_node_dms]
    suite_resources.nodes = suite_rs_node_dms

    # Create client data model.
    # This is hardcoded based on the Vagrant configuration in the sanity-hackathon repo.
    client_dm = sf_models.ResourceClient()
    client = models.ClientSuiteResource.create(model=client_dm)
    client.model.mip = '192.168.0.42'
    client.model.username = 'root'
    client.model.password = 'solidfire'
    client.model.client_name = 'ubuntuminion'
    suite_resources.clients = [client]
    return suite_resources


def get_task_logs(task_id):
    """Downloads the logs from an AT2 task.

    :param int task_id: task ID to gather resources from
    """
    at2_api = LegacyAt2Api(username=at2_constants.AUTO_USER, password=at2_constants.AUTO_PASS)
    task = at2_api.get_task_instance_by_id(task_id)
    _download_task_logs(task)


def _generate_filtered_logs(suite_inst):
    """Created filtered logs.

    :param camelot.TestSuite suite_inst: Already run suite instance
    """
    local_artifacts_dir = suite_inst.local_artifacts_dir
    if local_artifacts_dir is None:
        return
    try:
        test_log_file = os.path.join(local_artifacts_dir, camelot.TEST_LOG_FILE)
        for log_filter in suite_inst.config.log_filters:
            _filter_log.filter_and_create_log(log_filter, test_log_file)
    except Exception as e:  # pylint: disable=broad-except
        _log.warn('Generating filtered test logs failed! Exception details: {}'.format(e))


def _post_to_groot(suite_inst):
    """Post the suite result to GROOT, if we want to.

    :param camelot.TestSuite suite_inst: Already run suite instance
    """
    if not suite_inst._skip_remote_reporting:  # pylint: disable=protected-access
        suite_report = suite_inst.result.generate_report()
        _log.debug('Posting to groot')
        groot.post_to_groot(groot.REPORT_ENDPOINT, suite_report)


def _post_to_hipchat(suite_inst):
    """Post message to HipChat group if the suite fails to run using HipChatAPI.

    :param camelot.TestSuite suite_inst: Already run suite instance
    """
    if suite_inst is None or suite_inst.status is not suite_inst.statuses.passed:
        if suite_inst.config.hipchat.token and suite_inst.config.hipchat.room_id_or_name is not None:
            try:
                _hipchat_notify(token=suite_inst.config.hipchat.token,
                                room_id_or_name=suite_inst.config.hipchat.room_id_or_name,
                                message='{} Suite Failed'.format(suite_inst.name))
            except Exception as e:  # pylint: disable=broad-except
                if isinstance(e, SystemExit) or isinstance(e, KeyboardInterrupt):
                    raise  # pragma: no cover
                else:
                    _log.debug('Got Exception with hipchat notify.', exc_info=True)


def _get_builds(steps):  # pylint: disable=too-complex
    """Given at2 steps, get the node and fdva builds.

    :param list[dict] steps: steps to get builds from
    :return: (node_build, fdva_build)
    :rtype: tuple[sf_models.Build, sf_models.Build]
    """
    def get_build_from_input(inp):
        """Gets the build from input."""
        if inp[at2_constants.INPUT_NAME] == at2_constants.INPUT_BUILD_VERSION:
            value = inp[at2_constants.VALUE]
            if value:
                return sf_models.Build(value)

    def get_build(steps, step_name):
        """Gets the build of step_name from steps."""
        for s in steps:
            if s[at2_constants.STEP_NAME] == step_name:
                for i in s[at2_constants.INPUTS]:
                    build = get_build_from_input(i)
                    if build:
                        return build

    node_build = get_build(steps, at2_constants.RESERVE_NODES)
    fdva_build = get_build(steps, at2_constants.RESERVE_FDVA)

    if not node_build:
        node_build = BUILD_TEMPLATE
    if not fdva_build:
        fdva_build = BUILD_TEMPLATE
    return node_build, fdva_build


def _get_networks(node_dms):
    """Get all networks from the nodes without repetition.

    :param list[sf_models.ResourceNode] node_dms: node data models
    :return: list of networks suite resources
    :rtype: list[camelot.models.NetworkSuiteResource]
    """
    gotten_groups = set()
    network_dms = []
    for n in (n.network_profile for n in node_dms):
        if n.network_name not in gotten_groups:
            gotten_groups.add(n.network_name)
            network_dms.append(n)

    networks = [models.NetworkSuiteResource(dict(model=n, config=None)) for n in network_dms]
    return networks


def _download_task_logs(task):
    """Get logs for run_camelot_suite step in an at2 task and write them to a txt file.

    :param dict task: task object from AT2
    """
    task_id = task[at2_constants.TASK_INSTANCE_ID]
    steps = task[at2_constants.TASK_INSTANCE_STEPS]
    output = [s[at2_constants.OUTPUT_VALUE] for s in steps
              if s[at2_constants.STEP_NAME] == at2_constants.RUN_CAMELOT_SUITE_STEP]
    file_path = '{}/task_log_{}.txt'.format(os.getcwd(), task_id)

    if output:
        try:
            instance_id = output[0].get(at2_constants.ARTIFACT_CREDENTIALS,
                                        {}).get(at2_constants.TASK_INSTANCE_STEP_ID)
            if instance_id:
                url = 'https://autotest2.solidfire.net/stream/stdout/{}/'.format(instance_id)
                AUTH = auth.HTTPBasicAuth(at2_constants.AUTO_USER, at2_constants.AUTO_PASS)
                session = requests.Session()
                session.auth = AUTH

                resp = session.get(url, stream=True)
                with open(file_path, 'w') as f:
                    for l in resp.iter_lines():
                        f.write(l+'\n')
                print('Task {} logs have been written to file {}'.format(task_id, file_path))
        except AttributeError:
            print('Looks like task {} did not run camelot suite. No logs to generate.'.format(task_id))
    else:
        print("Task {} does not contain 'run camelot suite' step we can pull logs for.".format(task_id))


def _get_task_resources(task):
    """Parse the task resources out of the task object.

    :param dict task: task object from AT2
    :return: Suite resources data model
    :rtype: camelot.models.SuiteResources
    """

    steps = task[at2_constants.TASK_INSTANCE_STEPS]

    # Process steps to get cluster and node data models
    cluster_dms = _get_cluster_dms(steps)
    node_dms = _get_node_dms(steps)

    # Process steps to get the full suite resources for IPs and Clients.
    ips = _get_ip_suite_resources(steps)
    clients = _get_client_suite_resources(steps)

    task_virtual_networks = get_raw_resources(steps, at2_constants.RESERVE_VIRTUAL_NETWORKS)
    camelot_cluster_resources = get_raw_resources(steps, at2_constants.CREATE_CAMELOT_CLUSTER_RESOURCES)

    node_build, fdva_build = _get_builds(steps)

    # If we fail to get the upgrade_builds from the task, then use old behavior where upgrade builds is
    # the node build.
    upgrade_builds = None
    if camelot_cluster_resources and CONFIG in camelot_cluster_resources[0]:
        upgrade_builds = camelot_cluster_resources[0][CONFIG].get(UPGRADE_BUILDS, None)

    if not upgrade_builds:
        upgrade_builds = [copy.copy(node_build)]

    # Process steps to get the full suites resources for the FDVAs.
    fdvas = _get_fdva_suite_resources(steps, fdva_build)

    # note: config.upgrade_builds is the "upgrade path" and config.build
    # is the "version under test" which means the up_build
    # for non-upgrade tests these builds are equal.
    cluster_configs = [models.ClusterSuiteConfig.create(
        build=copy.copy(node_build),
        upgrade_builds=upgrade_builds,
        options=[],
        rtfi_flags=[],
        sfinstall_allow_flags=[],
        sfinstall_method='apt',  # TODO: use constant from upgrades module
        sfinstall_options='',
        unique_id=c.unique_id,
        virtual_ips=[])
                       for c in cluster_dms]

    node_configs = [models.NodeSuiteConfig(dict(build=copy.copy(node_build))) for _ in node_dms]

    # process network and remove duplicates
    networks = _get_networks(node_dms)

    assert not task_virtual_networks, 'virtual network models do not yet exist'

    clusters = None
    if not cluster_dms and camelot_cluster_resources:
        clusters = [models.ClusterSuiteResource(c) for c in camelot_cluster_resources]
    resource_dm = models.SuiteResources.create(
        clients=clients,
        fdvas=fdvas,
        management_nodes=fdvas,
        clusters=clusters or [models.ClusterSuiteResource(dict(model=c, config=cfg))
                              for (c, cfg) in zip(cluster_dms, cluster_configs)],
        hypervisors=[],  # TODO: implement hypervisors
        ip_ranges=[],  # TODO: implement ip_ranges
        ips=ips,
        networks=networks,
        nodes=[models.NodeSuiteResource(dict(model=n, config=cfg))
               for (n, cfg) in zip(node_dms, node_configs)],
        virtual_networks=[],  # TODO: implement virtual networks
    )
    # resource_dm.validate() # FIXME: after fixing VL-387
    return resource_dm


def _get_fdva_suite_resources(steps, fdva_build):
    """Process the step information and return corresponding Resource Cluster.

    :param list[dict] steps: steps to get resources from
    :param sf_models.Build fdva_build: FDVA build to use for config create
    :rtype: list[camelot.models.ManagementNodeSuiteResource]
    :return: list of FDVA Suite Resources
    """

    camelot_fdva_resources = get_raw_resources(steps, at2_constants.CREATE_CAMELOT_FDVA_RESOURCES)
    if camelot_fdva_resources and CONFIG in camelot_fdva_resources[0]:
        healthtools_version = camelot_fdva_resources[0][CONFIG].get(HEALTHTOOLS_VERSION, None)
        healthtools_down_versions = camelot_fdva_resources[0][CONFIG].get(HEALTHTOOLS_DOWN_VERSION, None)
        update_repo = camelot_fdva_resources[0][CONFIG].get(UPDATE_REPO, None)
    else:
        healthtools_version = None
        healthtools_down_versions = None
        update_repo = None

    task_fdvas = [s[at2_constants.OUTPUT_VALUE] for s in steps
                  if s[at2_constants.STEP_NAME] == at2_constants.RESERVE_FDVA]
    fdva_dms = [sf_models.ResourceManagementNode(f) for f in task_fdvas]
    fdva_configs = [models.ManagementNodeSuiteConfig.create(
        build=fdva_build, update_repo=update_repo, build_install_sequence=[fdva_build],
        healthtools_version=healthtools_version,
        healthtools_down_versions=healthtools_down_versions)
        for _ in fdva_dms]

    fdvas = [models.ManagementNodeSuiteResource(dict(model=f, config=cfg))
             for (f, cfg) in zip(fdva_dms, fdva_configs)]
    return fdvas


def _get_cluster_dms(steps):
    """Process the step information and return corresponding Resource Cluster.

    :param list[dict] steps: steps to get resources from
    :rtype: list[sf_models.ResourceCluster]
    :return: list of ResourceCluster Data Models
    """

    task_clusters = get_raw_resources(steps, at2_constants.SETUP_CLUSTER)
    task_clusters += get_raw_resources(steps, at2_constants.SETUP_CLUSTERS)
    task_clusters += get_raw_resources(steps, at2_constants.CREATE_CLUSTER)
    cluster_dms = [sf_models.ResourceCluster(c) for c in task_clusters]
    return cluster_dms


def _get_node_dms(steps):
    """Process the step information and return corresponding Resource Nodes.

    :param list[dict] steps: steps to get resources from
    :rtype: list[sf_models.ResourceNode]
    :return: list of ResourceNode Data Models
    """
    task_nodes = get_raw_resources(steps, at2_constants.CREATE_CAMELOT_NODE_RESOURCES)
    if not task_nodes:
        task_nodes = get_raw_resources(steps, at2_constants.RESERVE_NODES)
    else:
        task_nodes = [node['model'] for node in task_nodes]
    node_dms = [sf_models.ResourceNode(n) for n in task_nodes]
    return node_dms


def _get_ip_suite_resources(steps):
    """Process the step information and return corresponding IP Suite Resources.

    :param list[dict] steps: steps to get resources from
    :rtype: list[camelot.models.IpSuiteResource]
    :return: list of IP Suite Resources
    """
    task_ips = get_ip_resources(steps)
    ip_dms = [sf_models.ResourceIp(c) for c in task_ips]
    ips = [models.IpSuiteResource(dict(model=i)) for i in ip_dms]
    return ips


def _get_client_suite_resources(steps):
    """Process the step information and return corresponding Client Suite Resources.

    :param list[dict] steps: steps to get resources from
    :rtype: list[camelot.models.ClientSuiteResource]
    :return: list of ClientSuiteResource Data Models
    """
    task_clients = get_raw_resources(steps, at2_constants.RESERVE_CLIENTS)
    client_dms = [sf_models.ResourceClient(c) for c in task_clients]
    clients = [models.ClientSuiteResource(dict(model=c, config=None)) for c in client_dms]
    return clients


def set_log_level(config_dm):
    """Method to set the log level.

    :param camelot.models.SuiteConfig config_dm: Config data model
    """
    log_level = config_dm.logging_level.value
    autologger.set_global_loglevel(log_level)


def get_class_from_path(path):
    """Get the class object from a string module path.

    :param str path: Path to the class
    :return: The suite object
    :rtype: six.class_types
    """
    module_path, _, class_name = path.rpartition('.')
    module = importlib.import_module(module_path)
    class_obj = getattr(module, class_name)
    return class_obj


def _hipchat_notify(token, room_id_or_name, message, color='gray', notify=False, format_style='text'):
    """Send notification to a HipChat room via API version 2.

    :param str token: The token can be used to make REST calls to HipChat
    :param str room_id_or_name: The id or url encoded name of the room
    :param str message: The message to post
    :param str color: The background color of the message
                      ("yellow", "red", "green", "purple", "gray" or "random")
    :param bool notify: Whether this message should trigger a user notification
    :param str format_style: The format that can be html or text
    :rtype: int

    """
    host = os.getenv('HIPCHAT_SERVER', 'hipchat.ngage.netapp.com')
    url = 'https://{0}/v2/room/{1}/notification'.format(host, room_id_or_name)
    headers = {'Content-type': 'application/json'}
    headers['Authorization'] = 'Bearer ' + token
    payload = {
        'message': message,
        'notify': notify,
        'message_format': format_style,
        'color': color
    }
    response = requests.post(url, data=json.dumps(payload), headers=headers)
    response.raise_for_status()

    return response.status_code


def run_suite_from_module(suite_path, resources_dm, config_dm,  # pylint: disable=too-complex
                          open_pdb=False):
    """Actually run the suite.

    :param str suite_path: Path to suite
    :param camelot.models.SuiteResources resources_dm: Suite resources data model
    :param camelot.models.SuiteConfig config_dm: Suite config data model
    :param bool open_pdb: If True, open a pdb debugger on failure. config_dm.stop_on_failure
        must be True for this to work.
    :return: The suite object
    :rtype: camelot.TestSuite
    :raises ValueError: If suite_path is not a valid path to a suite
    """
    try:
        suite = get_class_from_path(suite_path)
    except (AttributeError, ImportError, ValueError):
        _log.debug('Invalid suite path: {}'.format(suite_path))
        raise

    if config_dm.changeset_modules:
        # Modify config_dm with changeset info. This must be determined at runtime.
        changesets = get_changesets(config_dm.changeset_modules)
        data = config_dm.serialize()
        data['changesets'] = [c.serialize() for c in changesets]
        config_dm = models.SuiteConfig(data)

    try:
        suite_instance = suite(resources_dm, config_dm)
    except TypeError:
        _log.debug('Invalid suite path: {}'.format(suite_path))
        raise

    try:
        suite_instance.execute()
    except (Exception, KeyboardInterrupt):  # pylint: disable=broad-except
        if suite_instance.status is suite_instance.statuses.pending and not open_pdb:
            # This means preconditions failed, or an exception happened in __init__
            raise
        # KeyboardInterrupt is caught so we can go into pdb after a CTRL-C if needed
        traceback.print_exc()
        if open_pdb:
            # exc_info returns a tuple containing (type, value, traceback)
            _, _, tb = sys.exc_info()
            # The no member error is disabled here in order to support the pdbpp package.
            # Pylint gets a false positive when that package is installed.
            pdb.post_mortem(tb)  # pylint: disable=no-member
    finally:
        _post_to_groot(suite_instance)
        _generate_filtered_logs(suite_instance)
        _post_to_hipchat(suite_instance)
    return suite_instance


def _instantiate_model(file_name, model_class, overrides=None):
    """Parse given file as json, validate it, and return the data model representation.

    :param str file_name: Full path of the file
    :param six.class_types model_class: The class of the model that the file represents
    :param dict overrides: A dictionary of values to override
    :rtype: sf_models.Base
    :return: The associated data model
    """
    with open(file_name) as json_file:
        data = json.load(json_file)
    if overrides is not None:
        data.update(overrides)
    dm = model_class(data)
    # FIXME: uncomment once client iDRAC ip is working
    # dm.validate()
    return dm


def get_config_dm(config_file, stop_on_failure=None, store_report=None, start_index=None,
                  stop_index=None, log_level=None):
    """Returns the driver config data model.

    :param str|None config_file: Path to config file. If None, look for suite_config.json in default
        data directory.  If that file does not exist, instantiate a default config.
    :param bool stop_on_failure: This will override the stop_on_failure value in config file.
    :param bool store_report: This will override the store_report value in the config file.
    :param int start_index: Test case start index in Suite, inclusive
    :param int stop_index: Test case stop index in Suite, exclusive
    :param camelot.LoggingLevel log_level: Global Log Level Override
    :rtype: camelot.models.SuiteConfig
    """
    overrides = {}
    if stop_on_failure is not None:
        overrides[_constants.STOP_ON_FAILURE] = stop_on_failure
    if store_report is not None:
        overrides[_constants.STORE_REPORT] = store_report
    if start_index is not None:
        overrides[_constants.START_INDEX] = start_index
    if stop_index is not None:
        overrides[_constants.STOP_INDEX] = stop_index
    if log_level is not None:
        overrides[_constants.LOGGING_LEVEL] = log_level.value

    # Instantiate default config
    default_config_dm = models.SuiteConfig(overrides)

    if config_file is None:
        # Look for a config file in default locations
        default_config_file = get_config_file(_constants.DEFAULT_SUITE_CONFIG_FILE)
        if default_config_file:
            _log.debug('Loading suite config from: {}'.format(default_config_file))
            suite_model = models.SuiteConfig  # type: six.class_types
            default_config_dm = _instantiate_model(default_config_file, suite_model, overrides)
        else:
            _log.debug('No suite config file was found. Using default values.')
    else:
        _log.debug('Loading suite config from: {}'.format(config_file))
        suite_model = models.SuiteConfig  # type: six.class_types
        default_config_dm = _instantiate_model(config_file, suite_model, overrides)

    return default_config_dm


def get_config_file(file_name):
    """Get the path to a config file.

    Look for data directory in the project root directory.
    Returns a valid path to a file or None. Used for suite_config.json or
    resource.json.

    :param str file_name: Name of config file
    :return: Path to resources file if it exists, None otherwise
    :rtype: str|None
    """
    # Look for config file in default location
    config_file_path = os.path.join(_constants.DEFAULT_DATA_DIR, file_name)
    if os.path.isfile(config_file_path):
        return config_file_path
    else:
        # Look for resources file in the project root directory
        project_dir = get_project_dir()
        if project_dir:
            config_file_path = os.path.join(project_dir, _constants.DEFAULT_DATA_DIR, file_name)
            if os.path.isfile(config_file_path):
                return config_file_path

            else:
                return None
        else:
            return None


def get_project_dir():
    """Get root project dir.

    Only works when in a virtual environment.

    :rtype: str
    """
    project_file = os.path.expandvars('$VIRTUAL_ENV/$VIRTUALENVWRAPPER_PROJECT_FILENAME')
    if not os.path.isfile(project_file):
        return None
    with open(project_file) as p:
        project_dir = p.read().strip()
    return project_dir


def get_resources_dm(resources_file=None, resources_class_path=None):
    """Returns the suite resources data model.

    If resources_file is not given, look for the file in <project_root>/vl_data/resources.json.

    :param str resources_file: Path to resources file. If None, look for resources.json in data
        directory. If that file does not exist, return an empty resources data model.

    :param str resources_class_path: Class to instantiate as the resources for the test run. If
        nothing is specified camelot.models.SuiteResources is used.
    :rtype: sf_models.Base
    """
    # Look up the class to override with, if any
    if resources_class_path is None:
        resources_class = models.SuiteResources
    else:
        resources_class = get_class_from_path(resources_class_path)

    # Instantiate empty resources model as a default
    result = resources_class()

    if resources_file is None:
        resources_file = get_config_file(_constants.DEFAULT_SUITE_RESOURCES_FILE)
        if resources_file is not None:
            _log.debug('Loading suite resources from {}'.format(resources_file))
            result = _instantiate_model(resources_file, resources_class)
        else:
            _log.debug('WARNING: No resources file found. An empty resources object will be used.')
    else:
        _log.debug('Loading suite resources from {}'.format(resources_file))
        result = _instantiate_model(resources_file, resources_class)

    return result


def make_data_dir(data_directory=_constants.DEFAULT_DATA_DIR):
    """Makes the data directory if it does not already exist.

    :param str data_directory: directory name where the suite config json is
    """
    if not os.path.exists(data_directory):
        os.makedirs(data_directory)


def run(suite_path, resources_file=None, config_file=None, open_pdb=None,
        store_report=None, resources_class_path=None, start_index=None, stop_index=None,
        log_level=None, log_resource_utilization=False):
    """Run the driver. This runs the suite.

    :param str suite_path: Path to suite (e.g. tests.featureA.suites.sunny_day.TestA)
    :param str resources_file: resource file name
    :param str config_file: suite config file name
    :param bool open_pdb: If True, ensure that a post mortem debugger is opened upon failure. This
        will also set the suite config stop_on_failure variable to True.
    :param bool store_report: If True, write a report.json file to the data directory upon completion
    :param str resources_class_path: Path to the class to use for suite resources. Defaults to
        camelot.models.SuiteResources
    :param int start_index: Test case start index in Suite, inclusive
    :param int stop_index: Test case stop index in Suite, exclusive
    :param camelot.LoggingLevel log_level: Global Log Level Override
    :param bool log_resource_utilization: If True, log memory/cpu utilization to a csv file
    :return: The executed suite object
    :rtype: camelot.TestSuite|None
    """
    config_dm = get_config_dm(config_file, open_pdb, store_report, start_index, stop_index, log_level)
    set_log_level(config_dm)
    resources_dm = get_resources_dm(resources_file, resources_class_path)  # type: models.SuiteResources
    if not log_resource_utilization:
        return run_suite_from_module(suite_path, resources_dm, config_dm, open_pdb)
    else:  # pragma: no cover
        import psutil
        suite_process = multiprocessing.Process(target=run_suite_from_module,
                                                args=(suite_path, resources_dm, config_dm, open_pdb,))
        suite_process.start()
        p = psutil.Process(suite_process.pid)
        with open('resource_utilization.csv', 'w') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(['CPU Percent', 'Memory', 'Virtual Memory'])
            while suite_process.is_alive():
                writer.writerow([p.cpu_percent(), p.memory_percent(memtype='uss'),
                                p.memory_percent(memtype='vms')])
                suite_process.join(10)
