"""Logic for provision command.

This command creates a cluster from raw resources, deploying any ephemeral resources passed in.
"""
import uuid

from sf_models import at2_enums

from sf_api import api_errors
from valence import (
    enums,
    functional_models,
    models,
    vl_helpers,
)


class ProvisionCommand(object):
    """Deploy ephemeral resources and create a cluster.

    :param valence.models.SuiteResources resources: Resources to provision
    :param bool cicd: Whether this is being run from cicd or not, default False
    :param int witness_num: Number of witness nodes to deploy
    :param int cluster_num: Number of clusters to create. Max 1 is supported.

    """
    def __init__(self, resources, cicd=False, witness_num=0, cluster_num=1):
        self.suite_resources = resources
        self.witness_num = witness_num
        self.cluster_num = cluster_num
        self.cicd = cicd
        self.one_gig_ips = [ip for ip in self.suite_resources.ips if ip.model.network is
                            at2_enums.NetworkSpeed.one_gig]
        self.ten_gig_ips = [ip for ip in self.suite_resources.ips if ip.model.network is
                            at2_enums.NetworkSpeed.ten_gig]
        self.witness_node_dms = [node for node in self.suite_resources.node_dms if node.is_witness()]
        self.build = None

    def validate_input_resources(self):
        """Validate the correct number of IP pairs and cluster doesn't exist."""
        assert self.suite_resources.nodes, 'No nodes found for provisioning'
        # Raise an error if a cluster already is in the resources.
        msg = 'Cluster found in resources. Has provision run already?'
        assert not self.suite_resources.clusters, msg

        # Assert if there aren't enough IP addresses
        msg = 'Not enough {} IP addresses'
        assert len(self.one_gig_ips) >= self.witness_num + self.cluster_num, msg.format('Management')
        assert len(self.ten_gig_ips) >= self.witness_num + self.cluster_num, msg.format('Storage')

        # Creation of one cluster is supported
        assert self.cluster_num < 2, 'cluster_num can only be 0 or 1'

    def transform_resources(self):
        """Update the SuiteResources object to include witness resource dms and cluster dms."""
        self.build = self.suite_resources.nodes[0].config.build
        for _ in range(self.witness_num):
            node_name = 'CICD-Witness-Node-{}'.format(str(uuid.uuid4())[:4]) if self.cicd else None
            witness_resource_dm = vl_helpers.create_witness_model(
                one_gig=self.one_gig_ips.pop().model,
                ten_gig=self.ten_gig_ips.pop().model,
                node_name=node_name
            )
            witness_config = models.NodeSuiteConfig.create(build=self.build)
            self.witness_node_dms.append(witness_resource_dm)
            node_suite_resource = models.NodeSuiteResource.create(model=witness_resource_dm,
                                                                  config=witness_config)
            self.suite_resources.nodes.append(node_suite_resource)

        for _ in range(self.cluster_num):
            unique_id = str(uuid.uuid4())[:4]
            mvip = self.one_gig_ips.pop().model.address
            svip = self.ten_gig_ips.pop().model.address
            rs_cluster_dm = models.ResourceCluster.create(
                mvip=mvip,
                rep_count=2,
                svip=svip,
                username='admin',
                password='admin',
                api_endpoint=self.build.sf_version.api_endpoint,
                name='Cluster-{}'.format(unique_id),
                unique_id=unique_id
            )
            cluster_model = models.ClusterSuiteConfig.create(
                upgrade_builds=[self.build],
                build=self.build,
                options=[],
                rtfi_flags=[],
                sfinstall_allow_flags=[],
                sfinstall_method='apt',
                sfinstall_options='',
                unique_id=rs_cluster_dm.unique_id,
                virtual_ips=[]
            )
            cluster_resource = models.ClusterSuiteResource.create(model=rs_cluster_dm,
                                                                  config=cluster_model)
            self.suite_resources.clusters.append(cluster_resource)

        self.suite_resources.ips = self.one_gig_ips + self.ten_gig_ips

    def deploy_witnesses(self):
        """Deploy any witness nodes."""
        if self.witness_node_dms:
            print('Deploying {} witness node(s)'.format(self.witness_num))
            vl_helpers.create_witness_nodes(self.witness_node_dms, self.build)

    def create_cluster(self):
        """Create the cluster."""
        nodes = [node.model for node in self.suite_resources.nodes]
        node_fm = functional_models.Node(nodes[0], None)
        for cluster in self.suite_resources.clusters:
            print('Calling CreateCluster against node {}'.format(node_fm))
            mvip = cluster.model.mvip
            svip = cluster.model.svip
            vl_helpers.create_cluster(node_fm, mvip, svip, nodes)

    def validate_teardown(self):
        """Validate either a cluster_dm or witness_dms exist for teardown to run."""
        msg = 'No valid resources to tear down, has provision been run yet?'
        assert self.suite_resources.clusters or self.witness_node_dms, msg

    def teardown(self):
        """Delete witness nodes, free ips, and update suite resource."""
        storage_dms = [node for node in self.suite_resources.node_dms if not node.is_witness()]
        if self.witness_node_dms:
            print('Deleting witness nodes.')
            vl_helpers.delete_witness_vms(self.witness_node_dms)
        if self.suite_resources.clusters:
            print('Freeing cluster IPs')
            for node in storage_dms:
                node_fm = functional_models.Node(node, self.suite_resources.clusters[0].model)
                # Forcing the cluster name to an empty string will always return an error
                try:
                    node_fm.set_cluster_config(cluster='', use_the_force_luke=True)
                except api_errors.SfApiException:
                    node_fm.restart_binary(enums.NodeBinary.solidfire)
