"""Valence driver qTest push subcommand."""
import json

import qtestapi
from qtestapi import constants as qt_constants

UNCOMMITTED = 'Uncommitted'


class QPushCommand(object):
    """Takes path to valence test result (report.json), pushes results to qTest.

    :param str report_path: full path to report.json
    :param str username: username for qTest project (email address)
    :param str password: password for this account
    :param int project_id: ID of project to push results to
    """
    def __init__(self, report_path, username, password, project_id):
        self.report_path = report_path
        self.api = qtestapi.QTestAPI(username, password, project_id)

        # String indentifiers parsed from report fields
        self.feature = None
        self.build = None
        self.release = None

        # qTest project objects
        self.qt_release = None
        self.qt_build = None
        self.qt_cycle = None
        self.qt_suite = None
        self.qt_test_cases = {}
        self.qt_test_runs = {}

        with open(report_path) as report:
            self.report = json.load(report)

    def post_to_qtest(self):
        """Posts report.json results found in the provided path to SET qTest project.

        Creates all necessary project structures prior to submitting results.
        """
        # Make sure test cases have been scraped to qTest with scribe
        self._verify_qtest_cases_exist()

        # Gather release, build, feature info from report
        self._parse_release_build_from_report()
        self._parse_feature_from_suite_path()

        # Find the appropriate qTest Release or create it
        self._get_qtest_release()

        # Find the appropriate qTest Release or create it
        self._get_qtest_build()

        # Find the appropriate qTest Release or create it
        self._get_qtest_test_cycle()

        # Find the appropriate qTest Release or create it
        self._get_qtest_suite()

        # Find the appropriate qTest Release or create it
        self._create_qtest_test_runs()

        # Finally, submit the results of our test run to qTest
        self._submit_qtest_test_run_logs()

    def _parse_feature_from_suite_path(self):
        """Attempts to parse feature name from test-suite path."""

        split_path = self.report['suite_path'].split('.')
        try:
            # Attempt to find 'tests' in suite_path.  Use the directory below to deduce feature
            #   if tests is not not found in path, will use default value.
            feature_index = split_path.index('tests') + 1
            self.feature = split_path[feature_index].upper()
        except ValueError:
            self.feature = UNCOMMITTED

    def _parse_release_build_from_report(self):
        """Attempts to parse release and build information from the report.json file."""

        # For now, grab build info from first node in resources.
        # TODO: update to 'build' field when VL-1618 is done
        try:
            first_node = self.report['resources']['nodes'][0]['config']['build']
            self.release = first_node['release'].split('-')[0]
            self.build = '-'.join([first_node['repositoryName'], first_node['version']])
        except (KeyError, IndexError):
            self.release = UNCOMMITTED
            self.build = UNCOMMITTED

    def _get_qtest_build(self):
        """Gets or creates qTest Test Build data model under which results will be reported."""

        matching_builds = [bld for bld in self.api.get_builds_under_release(self.qt_release)
                           if bld.name == self.build]

        # Check if the build already exists
        if matching_builds:
            self.qt_build = matching_builds[0]
        else:
            # If it doesn't , we need to create it.
            props = self.api.get_ns_default_properties(qt_constants.BUILDS)
            self.qt_build = self.api.create_build(self.qt_release.id, name=self.build,
                                                  properties=[prop.to_dict() for prop in props])

    def _get_qtest_release(self):
        """Gets or creates qTest Release data model under which results will be reported."""

        matching_releases = self.api.search_releases_by_name(self.release)

        # See if this release already exists
        if matching_releases:
            self.qt_release = matching_releases[0]
        else:
            # If it doesn't exist, we need to create it.
            props = self.api.get_ns_default_properties(qt_constants.RELEASES)
            self.qt_release = self.api.create_release(
                name=self.release, properties=[prop.to_dict() for prop in props])

    def _get_qtest_test_cycle(self):
        """Gets or creates qTest Test Cycle data model under which results will be reported."""

        matching_cycles = [tc for tc in self.api.get_test_cycles_under_release(self.qt_release) if
                           tc.name == self.feature]

        if matching_cycles:
            self.qt_cycle = matching_cycles[0]
        else:
            self.qt_cycle = self.api.create_test_cycle_under_release(self.qt_release.id, self.feature)

    def _get_build_field_id(self, field):
        """Get field_id for the build this report ran with.

        :param qtestapi.models.Field field:  Target Release/Build field
        :rtype: int
        """
        for av in field.allowed_values:
            if self.build in av.label:
                return av.value

    def _get_qtest_suite(self):
        """Gets or creates qTest Test Suite data model under which results will be reported."""

        suite_name = self.report['suite_name']
        matching_suites = [st for st in self.api.get_test_suites_under_cycle(self.qt_cycle) if
                           st.name == suite_name]

        # If we created new build/releases we need to update the allowed values
        self.api.refresh_field_cache()

        if matching_suites:
            # make sure release/build is correct
            suite = matching_suites[0]
            tgt_build_prop = suite.properties['Target Release/Build']
            if self.build not in tgt_build_prop.hr_value:
                tgt_build_id = self._get_build_field_id(tgt_build_prop.field)
                if tgt_build_id:
                    tgt_build_prop.update(tgt_build_id)
                    suite = self.api.update_resource(suite)
            self.qt_suite = suite
        else:
            props = self.api.get_ns_default_properties(qt_constants.TEST_SUITES)

            # set the target release/build, add to default properties
            self.api.refresh_field_cache()
            build_field = self.api.get_ns_field_by_label(qt_constants.TEST_SUITES,
                                                         'Target Release/Build')
            tgt_build_id = self._get_build_field_id(build_field)
            if tgt_build_id:
                props.append(build_field.to_property(tgt_build_id))

            self.qt_suite = self.api.create_test_suite_under_cycle(
                self.qt_cycle.id, suite_name, properties=[prop.to_dict() for prop in props])

    def _create_qtest_test_runs(self):
        """Creates test runs for any test cases found to have run in the report.json."""

        test_runs = self.api.get_test_runs_under_suite(self.qt_suite)

        for test_case, test_case_dm in self.qt_test_cases.items():
            matching_runs = [tr for tr in test_runs if tr.name == test_case]

            if matching_runs:
                self.qt_test_runs[test_case] = matching_runs[0]
            else:
                props = self.api.get_ns_default_properties(qt_constants.TEST_RUNS)
                self.qt_test_runs[test_case] = self.api.create_test_run_under_suite(
                    self.qt_suite.id, test_case_dm.name, test_case_dm.id,
                    properties=[prop.to_dict() for prop in props])

    def _submit_qtest_test_run_logs(self):
        """Submits test_run information found in the report.json file."""

        for test_run, test_run_dm in self.qt_test_runs.items():
            report_tr = [tr for tr in self.report['test_runs'] if tr['name'] == test_run][0]
            self.api.submit_test_run_log(test_run_dm.id,
                                         report_tr['status'],
                                         report_tr['start_time'],
                                         report_tr['end_time'])

    def _verify_qtest_cases_exist(self):
        """Queries qTest project for Test Cases matching python path of TestCases we ran.

        tc_scribe needs to be run to scrape the test-cases to qTest prior to attempting to post.
        """
        for test_case in self.report['test_runs']:
            matching_cases = self.api.search_test_cases_by_python_path(test_case['case_path'])

            if not matching_cases:
                raise RuntimeError('Test Case {} does not exist in qTest.  Use Scribe to create it.'
                                   .format(test_case['name']))
            else:
                self.qt_test_cases[test_case['name']] = matching_cases[0]
