"""Helper methods to issue a remote execution on AI's unnamed containerized execution environment."""

import json

import autologger
import requests
from future.moves.urllib.parse import urljoin

log = autologger.get_sflogging(__name__)

ASYNC_ID = 'async_id'
EXECUTION_SERVICE_URL = 'http://ai-test.solidfire.net/api/executions/'
STAGING_SERVICE_URL = 'http://ai-test-stg.solidfire.net/api/executions/'
PROVISION_SERVICE_URL = 'http://ai-test.solidfire.net/api/provision/'
PROVISION_STAGING_SERVICE_URL = 'http://ai-test-stg.solidfire.net/api/provision/'


def _make_request(url, payload):
    """Make a POST request to supplied URL.

    :param str url: target URL
    :param dict payload: request payload
    :return: response from server
    :rtype: dict
    """

    session = requests.Session()
    log.info('Making request on {} with args:\n{}'.format(url, json.dumps(payload, indent=4)))
    response = session.post(url=url, json=payload, timeout=(30, 30))
    json_response = response.json()
    log.debug('Response: \n{}'.format(json.dumps(json_response, indent=4)))
    response.raise_for_status()
    return json_response


def run_remote_suite(cmd_str, docker_registry, image_repository, tag,
                     resources=None, config=None, staging=False):
    """Run a single suite against the execution server.

    :param str cmd_str: Full path to the suite that will run and all other flags that are specified
    :param str docker_registry: The docker registry for the image
    :param str image_repository: The image repository
    :param str tag: The image tag
    :param dict resources: The resources for the tests
    :param dict config: The config for the tests
    :param bool staging: Flag to run suite in staging
    :returns: The execution ID
    :rtype: dict
    """

    internal_payload = {
        'resources': resources or {},
        'config': config or {}
    }
    payload_params = {
        'registry': docker_registry,
        'repository': image_repository,
        'tag': tag,
        'command': 'vl run ' + cmd_str,
        'payload': internal_payload
    }

    endpoint_url = EXECUTION_SERVICE_URL

    if staging:
        endpoint_url = STAGING_SERVICE_URL

    json_response = _make_request(endpoint_url, payload_params)

    if ASYNC_ID in json_response:
        log.info('URL: \n{}'.format(urljoin(endpoint_url, json_response[ASYNC_ID])))
    return json_response


def run_remote_node_rtfi(node, build, rtfi_params, timeouts, traditional_rtfi=False,
                         skip_post_rtfi_network_config=False, staging=False):
    """Issue request to the remote RTFI service to RTFI a single node.

    :param dict node: Node to RTFI
    :param dict build: Build to RTFI Eg {"element": "sodium", "version": "11.7.0.78",
                                         "repository": "sodium-patch7-dev"}
    :param dict rtfi_params: RTFI parameters to be passed to node
    :param dict timeouts: Custom timeout to set for each stage. Defaults will be used if None
    :param bool traditional_rtfi: If true, will perform tRTFI, else iRTFI
    :param bool skip_post_rtfi_network_config: If true, will skip issuing SetNetworkConfig/
                SetClusterConfig calls, leave the node in DHCP
    :param bool staging: If true, issue call to staging env
    :return: The execution ID
    :rtype: dict
    """
    # TODO: Add some param checking here before calling out to the service

    rtfi_payload = {
        'node': node,
        'build': build,
        'skip_post_rtfi_network': skip_post_rtfi_network_config,
        'rtfi_params': {} if rtfi_params is None else rtfi_params,
    }

    if traditional_rtfi:
        rtfi_payload['rtfi_params'].update({'sf_inplace': False})
    else:
        rtfi_payload['rtfi_params'].update({'sf_inplace': True})

    if timeouts:
        rtfi_payload['timeouts'] = timeouts

    endpoint_url = PROVISION_SERVICE_URL
    status_endpoint_url = EXECUTION_SERVICE_URL

    if staging:
        endpoint_url = PROVISION_STAGING_SERVICE_URL
        status_endpoint_url = STAGING_SERVICE_URL

    endpoint_url = urljoin(endpoint_url, 'node')

    json_response = _make_request(endpoint_url, rtfi_payload)

    if ASYNC_ID in json_response:
        log.info('URL: \n{}'.format(urljoin(status_endpoint_url, json_response[ASYNC_ID])))
    return json_response
