"""Tests the vl_docker utility"""
import json
import os
import sys
import unittest

import docker
import mock
import requests
from requests.exceptions import HTTPError

from camelot.driver import vl_docker


class TestVlDocker(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with mock.patch.object(docker, 'from_env', mock.Mock()):
            cls.client = vl_docker.DockerClient('fake_repo')

    def test_push_image(self):
        self.client.api.push.return_value = ['line1', 'line2']
        self.client.push_image('repo_name', 'tag')

    def test_build_image(self):
        expected_call_args = {'TESTS_REPO': 'tests_repo'}
        self.client.api.build.return_value = ['line1', 'line2']
        self.client.build_image('repo_name', 'tag', 'tests_repo', 'Dockerfile', None)
        assert self.client.api.build.called
        assert self.client.api.build.call_args[1]['buildargs'] == expected_call_args

    def test_build_image_with_build_arg(self):
        mock_build_args = {'<varname>': '<value>'}
        expected_call_args = {'<varname>': '<value>', 'TESTS_REPO': 'tests_repo'}
        self.client.api.build.return_value = ['line1', 'line2']
        self.client.build_image('repo_name', 'tag', 'tests_repo', 'Dockerfile', mock_build_args)
        assert self.client.api.build.called
        assert self.client.api.build.call_args[1]['buildargs'] == expected_call_args

    @mock.patch('subprocess.check_output')
    def test_get_tag(self, mock_check_output):
        mock_check_output.return_value = 'whatsinaname'
        tag = vl_docker.get_tag()
        assert 'whatsinaname' in tag

    @mock.patch('requests.get')
    def test_get_tags(self, mock_get):
        response_dict = {
            'name': 'test',
            'tags': ['latest']
        }

        mock_response = mock.Mock(spec=requests.Response)
        mock_response.content = json.dumps(response_dict).encode()

        mock_get.return_value = mock_response

        tags = vl_docker.get_tags('valence-docker-local-den',
                                  'test',
                                  'jenkins',
                                  'solidfire')
        assert tags == ['latest']

        expected_url = ('http://sf-artifactory.solidfire.net/artifactory/api/docker/'
                        'valence-docker-local-den/v2/test/tags/list')
        mock_get.assert_called_with(expected_url, auth=('jenkins', 'solidfire'))

    @mock.patch('requests.get')
    def test_get_tags_request_failed(self, mock_get):
        """test case where artifactory is down"""

        mock_resp = mock.Mock(spec=requests.Response)
        mock_resp.raise_for_status.side_effect = HTTPError('Server is down')
        mock_resp.status_code = 500
        mock_get.return_value = mock_resp
        self.assertRaises(HTTPError, lambda: vl_docker.get_tags('artifactory_repository_name',
                                                                'image_name',
                                                                'username',
                                                                'password'))

    def test_get_latest_tag_version(self):
        tags = ['v1', 'v.100', 'va', 'v2', 'v2.1']
        latest = vl_docker.get_latest_tag_version(tags)
        assert latest == 'v2.1'
        tags = ['bad_tag']
        latest = vl_docker.get_latest_tag_version(tags)
        assert latest == 'v0'

    def test_bump_version(self):
        version = 'v1'
        new_version = vl_docker.bump_version(version)
        assert new_version == 'v2'
        version = 'v1.2.3'
        new_version = vl_docker.bump_version(version)
        assert new_version == 'v1.2.4'
        version = 'v1.2.3.10'
        new_version = vl_docker.bump_version(version)
        assert new_version == 'v1.2.3.11'
        version = 'v1.2.3.10'
        new_version = vl_docker.bump_version(version, version_level=2)
        assert new_version == 'v1.2.4.10'
        version = 'v1.2.3.10'
        with self.assertRaises(IndexError):
            vl_docker.bump_version(version, version_level=5)

    def test_get_default_dockerfile(self):
        dockerfile = vl_docker.get_default_dockerfile()
        assert os.path.exists(dockerfile)


class TestVlDockerArgparse(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with mock.patch('camelot.driver.vl_docker.get_tag') as mock_tag:
            mock_tag.return_value = 'mock_tag'
            with mock.patch('camelot.driver.vl_docker.get_default_dockerfile') as mock_dockerfile:
                mock_dockerfile.return_value = 'mock/path/to/dockerfile'
                cls.parser = vl_docker.create_parser()

    def test_create_parser_upload_parser(self):
        input_args = ['push']
        arguments = self.parser.parse_args(input_args)
        assert arguments.subparser_name == 'push'

    def test_create_parser_build_parser(self):
        input_args = ['build']
        with self.assertRaises(SystemExit) as _:
            _ = self.parser.parse_args(input_args)

    def test_create_parser_build_parser_with_path(self):
        input_args = ['build', '.']
        arguments = self.parser.parse_args(input_args)
        assert arguments.subparser_name == 'build'

    def test_create_parser_build_parser_with_build_arg(self):
        input_args = ['build', '.', '--build-args', '<varname0>=<value0>',
                      '--build-args', '<varname1>=<value1>', '<varname2>=<value2>']
        expected_output = {'<varname0>': '<value0>', '<varname1>': '<value1>',
                           '<varname2>': '<value2>'}
        arguments = self.parser.parse_args(input_args)
        assert arguments.subparser_name == 'build'
        assert arguments.build_args == expected_output

    @mock.patch('camelot.driver.vl_docker.get_tag')
    @mock.patch('camelot.driver.vl_docker.get_tags')
    @mock.patch('camelot.driver.vl_docker.DockerClient.build_image')
    def test_main_driver_build(self, mock_build, mock_get_tags, mock_get_tag):
        mock_get_tag.return_value = ['v1']
        mock_get_tags.return_value = ['v0']
        mock_build.return_value = None
        input_args = ['prog', 'build', '/path', '-i']
        with mock.patch.object(sys, 'argv', input_args):
            vl_docker.main()

        mock_build.assert_called_with(
            build_args=None,
            dockerfile_path=mock.ANY,
            repo_name='valence-local',
            tag='v1',
            tests_repo_path='/path'
        )

    @mock.patch('camelot.driver.vl_docker.get_tag')
    @mock.patch('camelot.driver.vl_docker.get_tags')
    @mock.patch('camelot.driver.vl_docker.DockerClient.push_image')
    def test_main_driver_push(self, mock_push, mock_get_tags, mock_get_tag):
        mock_get_tag.return_value = ['v1']
        mock_get_tags.return_value = ['v0']
        mock_push.return_value = None
        input_args = ['prog', 'push', '-i']
        with mock.patch.object(sys, 'argv', input_args):
            vl_docker.main()

        mock_push.assert_called_with(repo_name='valence-local', tag='v1')
