import json
import os
import pdb
import shutil
import tempfile
import unittest
import uuid

import at2_sdk.at2
import autologger
import logbook
import mock
import pytest
import repo_utils
import requests

from camelot import (
    _enums,
    driver,
    models,
)
from camelot.driver import (
    _main,
    _run_remote,
)
from camelot.models import at2_constants
from valence import unittest_utils

DATA_PATH = unittest_utils.get_data_dir(__file__)

# Python paths to dummy suites and cases
DUMMY_CASES = 'valence.unittest_utils.cases.'
DUMMY_SUITES = 'valence.unittest_utils.suites.'


def log_level_to_int(log_level):
    """Converts a string representation of logging level to its integer representation.

    :param str log_level: A string defining the log level
    :return: The autologger module's integer representation of the logging level
    :rtype: int
    """
    return getattr(logbook, log_level)


@mock.patch('repo_utils.GitRepo.get_name')
@mock.patch('repo_utils.GitRepo.get_changeset_id')
class TestGetChangesets(unittest.TestCase):
    """Test camelot.driver._main.get_changesets."""

    def test_get_changesets(self, mock_git_get_changeset_id, mock_get_name):
        mock_git_get_changeset_id.return_value = '7b8f1692e74cc27af692978bfc76363226f77f35'
        mock_get_name.return_value = 'valence'
        changesets = _main.get_changesets(['valence'])
        ids = [c.changeset_id for c in changesets]
        assert '7b8f1692e74cc27af692978bfc76363226f77f35' in ids
        assert changesets[0].repo_name == 'valence'

    def test_get_changesets_import_error(self, mock_git_get_changeset_id, mock_get_name):
        changesets = _main.get_changesets(['not_a_module'])
        assert not changesets

    @mock.patch('repo_utils.get_repo')
    def test_get_changesets_not_a_repo(self, mock_get_repo, mock_git_get_changeset_id, mock_get_name):
        mock_get_repo.return_value = repo_utils.NoRepo('/etc')
        changesets = _main.get_changesets(['valence'])
        assert not changesets


@mock.patch('requests.Session.post')
class TestRunRemote(unittest.TestCase):

    def test_run_remote(self, mock_post):
        mock_response = mock.Mock()
        mock_post.return_value = mock_response
        mock_response.json.return_value = {_run_remote.ASYNC_ID: '1234'}
        driver.run_remote_suite('suite', 'registry', 'repo', 'tag')
        driver.run_remote_suite('suite', 'registry', 'repo', 'tag', staging=True)


@mock.patch('requests.Session.post')
class TestRunRemoteRtfi(unittest.TestCase):

    def test_run_remote_rtfi(self, mocked_post):
        mock_response = mock.Mock()
        mocked_post.return_value = mock_response
        mock_response.json.return_value = {_run_remote.ASYNC_ID: '1234'}
        resp = driver.run_remote_node_rtfi({'node_name': 'node_name'},
                                           {'repository': 'sodium',
                                            'build_version': '11.0.0.1',
                                            'element': 'sodium'},
                                           {'sf_netcfg': 'some value'},
                                           {},
                                           traditional_rtfi=False,
                                           skip_post_rtfi_network_config=False,
                                           staging=False)
        self.assertIn(_run_remote.ASYNC_ID, resp)

        resp = driver.run_remote_node_rtfi({'node_name': 'node_name'},
                                           {'repository': 'sodium',
                                            'build_version': '11.0.0.1',
                                            'element': 'sodium'},
                                           {'sf_netcfg': 'some value'},
                                           {'OS_UP_TIMEOUT': 100},
                                           traditional_rtfi=True,
                                           skip_post_rtfi_network_config=True,
                                           staging=True)
        self.assertIn(_run_remote.ASYNC_ID, resp)


class TestDriver(unittest_utils.MockHistoryUnitTest):
    def setUp(self):
        super(TestDriver, self).setUp()
        self.resources = models.SuiteResources.get_mock_object()
        self.config = models.SuiteConfig({'store_report': False,
                                          'artifact_credentials': {'local': False},
                                          'changeset_modules': [],
                                          'log_filters': []}
                                         )
        self.config_file = os.path.join(DATA_PATH, 'suite_config.json')
        self.resources_file = os.path.join(DATA_PATH, 'resources.json')
        self.failing_suite = DUMMY_SUITES + 'failing_suite.FailingSuite'
        self.passing_suite = DUMMY_SUITES + 'passing_suite.PassingSuite'
        self.exception_suite = DUMMY_SUITES + 'exception_in_setup.ExceptionSuite'

        # Mock get_changesets for all tests in this class since it's tested in another class
        self._get_changesets = _main.get_changesets
        _main.get_changesets = mock.Mock(return_value=models.RepoChangeset.get_mock_object_list(2))

    def tearDown(self):
        super(TestDriver, self).tearDown()
        _main.get_changesets = self._get_changesets

    def test_get_config_dm_stop_on_failure(self):
        config_dm = driver.get_config_dm(self.config_file, stop_on_failure=True, store_report=False)
        assert config_dm.stop_on_failure is True

    def test_get_config_dm_start_index(self):
        config_dm = driver.get_config_dm(self.config_file, start_index=99)
        assert config_dm.start_index == 99

    def test_get_config_dm_stop_index(self):
        config_dm = driver.get_config_dm(self.config_file, stop_index=99)
        assert config_dm.stop_index == 99

    def test_get_config_dm_log_level(self):
        config_dm = driver.get_config_dm(self.config_file, log_level=_enums.LoggingLevel.critical)
        assert config_dm.logging_level == _enums.LoggingLevel.critical

    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_config_dm_file_is_none_default_file_exists(self, mock_is_file,
                                                            mock_instantiate_model,
                                                            mock_mkdirs):
        mock_is_file.return_value = True
        mock_instantiate_model.return_value = self.config
        driver.get_config_dm(None)
        assert mock_instantiate_model.called is True

    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_config_dm_file_is_none_default_file_does_not_exist(self, mock_is_file,
                                                                    mock_instantiate_model,
                                                                    mock_mkdirs):
        mock_is_file.return_value = False
        config_dm = driver.get_config_dm(None)
        assert isinstance(config_dm, models.SuiteConfig)
        assert mock_instantiate_model.called is False

    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_resources_dm_file_is_none_default_file_exists(self, mock_is_file,
                                                               mock_instantiate_model,
                                                               mock_mkdirs):
        mock_is_file.return_value = True
        mock_instantiate_model.return_value = self.config
        resources_dm = driver.get_resources_dm()
        assert mock_instantiate_model.called is True

    @mock.patch('os.path.join')
    @mock.patch('camelot.driver._main.open')
    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_resources_dm_file_is_none_project_file_exists(self, mock_is_file,
                                                               mock_instantiate_model,
                                                               mock_mkdirs,
                                                               mock_open,
                                                               mock_join):
        # TODO: fails on jenkins. get_project_dir?
        mock_is_file.side_effect = [False, True, True]
        mock_instantiate_model.return_value = self.config
        mock_open(read_data='DATA_DIR').return_value = 'DATA_DIR'
        mock_join.return_value = 'DATA_DIR'
        driver.get_resources_dm()
        assert mock_instantiate_model.called is True

    @mock.patch('os.path.join')
    @mock.patch('camelot.driver._main.open')
    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_resources_dm_file_is_none_project_file_does_not_exist(self, mock_is_file,
                                                                       mock_instantiate_model,
                                                                       mock_mkdirs,
                                                                       mock_open,
                                                                       mock_join):
        mock_is_file.side_effect = [False, True, False]
        mock_instantiate_model.return_value = self.config
        mock_open(read_data='DATA_DIR').return_value = 'DATA_DIR'
        mock_join.return_value = 'DATA_DIR'
        driver.get_resources_dm()
        assert mock_instantiate_model.called is False

    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_resources_dm_file_is_none_default_file_does_not_exist(self, mock_is_file,
                                                                       mock_instantiate_model,
                                                                       mock_mkdirs):
        mock_is_file.return_value = False
        resources_dm = driver.get_resources_dm()
        assert isinstance(resources_dm, models.SuiteResources)
        assert mock_instantiate_model.called is False

    @mock.patch('os.makedirs')
    @mock.patch('camelot.driver._main._instantiate_model')
    @mock.patch('os.path.isfile')
    def test_get_resources_dm_override_default_model_path(self, mock_is_file,
                                                          mock_instantiate_model,
                                                          mock_mkdirs):
        mock_is_file.return_value = True
        mock_instantiate_model.return_value = self.resources
        resources_dm = driver.get_resources_dm(resources_class_path='camelot.models.SuiteResources')
        assert isinstance(resources_dm, models.SuiteResources)
        assert mock_instantiate_model.called is True

    @mock.patch.object(at2_sdk.at2.LegacyAt2Api, 'authenticate')
    @mock.patch.object(at2_sdk.at2.LegacyAt2Api, 'get_task_instance_by_id')
    def test_get_task_resources(self, mock_get_task_instance_by_id, mock_auth):
        mock_auth.return_value = True  # have to mock the auth or it calls AT2
        task_instance_data = unittest_utils.generate_at2_task_instance_dict()
        mock_get_task_instance_by_id.return_value = task_instance_data
        driver.get_task_resources(1)

    @mock.patch.object(at2_sdk.at2.LegacyAt2Api, 'authenticate')
    @mock.patch.object(at2_sdk.at2.LegacyAt2Api, 'get_task_instance_by_id')
    def test_get_task_resources_cluster_failed(self, mock_get_task_instance_by_id, mock_auth):
        mock_auth.return_value = True  # have to mock the auth or it calls AT2
        task_instance_data = unittest_utils.generate_at2_task_instance_dict()
        # Change the output value of cluster creation to none to simulate a step failure
        for t in task_instance_data[at2_constants.TASK_INSTANCE_STEPS]:
            if t[at2_constants.STEP_NAME] == at2_constants.SETUP_CLUSTER:
                t[at2_constants.OUTPUT_VALUE] = None
        mock_get_task_instance_by_id.return_value = task_instance_data
        with pytest.raises(RuntimeError):
            driver.get_task_resources(1)

    @mock.patch.object(at2_sdk.at2.LegacyAt2Api, 'authenticate')
    @mock.patch.object(at2_sdk.at2.LegacyAt2Api, 'get_task_instance_by_id')
    def test_get_task_resources_reserve_nodes_failed(self, mock_get_task_instance_by_id, mock_auth):
        mock_auth.return_value = True  # have to mock the auth or it calls AT2
        task_instance_data = unittest_utils.generate_at2_task_instance_dict()
        # Change the output value of reserve nodes to none to simulate a step failure
        for t in task_instance_data[at2_constants.TASK_INSTANCE_STEPS]:
            if t[at2_constants.STEP_NAME] == at2_constants.RESERVE_NODES:
                t[at2_constants.OUTPUT_VALUE] = None
        mock_get_task_instance_by_id.return_value = task_instance_data
        with pytest.raises(RuntimeError):
            driver.get_task_resources(1)

    @mock.patch('os.makedirs')
    def test_make_data_dir(self, mkdir):
        temp_dir = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        driver.make_data_dir(temp_dir)
        mkdir.assert_called()

    def test_set_log_level(self):
        config_dm = models.SuiteConfig.create(
            logging_level='DEBUG'
        )
        driver.set_log_level(config_dm)
        assert autologger.main.get_global_loglevel() == log_level_to_int('DEBUG')

    def test_run_suite_from_module_pass(self):
        result = driver.run_suite_from_module(self.passing_suite, self.resources, self.config)
        repr(result)
        assert result.status is result.statuses.passed

    @mock.patch('camelot.driver._main.requests.post')
    def test_run_suite_from_module_fail(self, mock_hipchat):
        result = driver.run_suite_from_module(self.failing_suite, self.resources, self.config)
        repr(result)
        assert result.status is result.statuses.failed

    @mock.patch('camelot.driver._main.requests.post')
    def test_run_suite_from_module_fail_in_preconditions(self, mock_hipchat):
        """When a test fails in preconditions, it should raise the error."""
        suite = DUMMY_SUITES + 'FailInPreconditions'
        with pytest.raises(AssertionError):
            driver.run_suite_from_module(suite, self.resources, self.config)

    @mock.patch('camelot.driver._main.requests.post')
    def test_run_suite_from_module_with_report(self, mock_hipchat):
        report_location = tempfile.mkdtemp()
        config_dm = models.SuiteConfig.create(
            data_directory=report_location,
            store_report=True
        )

        driver.run_suite_from_module(self.failing_suite, self.resources, config_dm)

        shutil.rmtree(report_location)

    def test_run_suite_from_module_with_module_only_in_path(self):
        invalid_suite_name = DUMMY_SUITES + 'failing_suite'
        with pytest.raises(TypeError):
            driver.run_suite_from_module(invalid_suite_name, self.resources, self.config)

    def test_run_suite_from_module_with_empty_path(self):
        empty_path = ''
        with pytest.raises(ValueError):
            driver.run_suite_from_module(empty_path, self.resources, self.config)

    def test_run_suite_from_module_with_misspelled_class_name(self):
        invalid_suite_name = DUMMY_SUITES + 'passing_suite.PassingSute'
        with pytest.raises(AttributeError):
            driver.run_suite_from_module(invalid_suite_name, self.resources, self.config)

    def test_run_suite_from_module_with_misspelled_module_name(self):
        invalid_suite_name = DUMMY_SUITES + 'passing_suie.PassingSuite'
        with pytest.raises(ImportError):
            driver.run_suite_from_module(invalid_suite_name, self.resources, self.config)

    @mock.patch.object(driver._main, 'run_suite_from_module')
    def test_run(self, mock_run_suite_from_module):
        passing_suite = unittest_utils.suites.PassingSuite(self.resources, self.config)
        passing_suite.status = passing_suite.statuses.passed
        mock_run_suite_from_module.return_value = passing_suite
        driver.run(self.passing_suite, self.resources_file, self.config_file)
        with open(self.config_file) as cf:
            config_data = json.load(cf)
        log_level = log_level_to_int(config_data.get('logging_level'))
        assert log_level == autologger.main.get_global_loglevel()

    @mock.patch('camelot.driver._main.requests.post')
    @mock.patch.object(pdb, 'post_mortem')
    def test_debugger_opens_pdb(self, mock_post_mortem, mock_hipchat):
        suite_path = self.failing_suite
        config = models.SuiteConfig.create(
            stop_on_failure=True,
            store_report=False
        )
        driver.run_suite_from_module(suite_path, self.resources, config, open_pdb=True)
        self.assertTrue(mock_post_mortem.called)

    @mock.patch('valence.camelot.SuiteResult')
    @mock.patch.object(requests, 'post')
    def test_push_to_groot(self, mock_post, mock_result):
        mock_post.side_effect = [None, requests.HTTPError]
        passing_suite = unittest_utils.suites.PassingSuite(self.resources, self.config)
        passing_suite._skip_remote_reporting = False
        mock_result.generate_report = mock.MagicMock(return_value=dict())
        passing_suite.result = mock_result
        _main._post_to_groot(passing_suite)
        _main._post_to_groot(passing_suite)
        cc = mock_post.call_count
        passing_suite._skip_remote_reporting = True
        _main._post_to_groot(passing_suite)
        assert cc == mock_post.call_count

    @mock.patch('camelot.driver._filter_log.filter_and_create_log')
    def test_generate_filtered_logs(self, mock_filter_and_create_log):
        passing_suite = unittest_utils.suites.PassingSuite(self.resources, self.config)
        mock_camelot_filters = models.CamelotLogFilter.get_mock_object_list(1)
        passing_suite.local_artifacts_dir = ''
        passing_suite.config.log_filters = mock_camelot_filters
        _main._generate_filtered_logs(passing_suite)

    @mock.patch('camelot.driver._filter_log.filter_and_create_log')
    def test_generate_filtered_logs_fails(self, mock_filter_and_create_log):
        passing_suite = unittest_utils.suites.PassingSuite(self.resources, self.config)
        passing_suite.local_artifacts_dir = ''
        passing_suite.config.log_filters = 12
        _main._generate_filtered_logs(passing_suite)
