try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import os

import mock
import repo_utils

from camelot.driver import (
    _constants,
    _list_command,
)
from valence import (
    models,
    unittest_utils,
)

DATA_DIR = unittest_utils.get_data_dir(__file__)
UT_UTILS_PATH = os.path.split(unittest_utils.__file__)[0]
VCS_ROOT = os.path.split(os.path.split(UT_UTILS_PATH)[0])[0]


module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock get_vcs_root to speed up tests
    module_mock.mock(repo_utils, 'get_vcs_root', mock.Mock(return_value=VCS_ROOT))


def tearDownModule():
    module_mock.restore()


class TestListCommand(unittest_utils.MockHistoryUnitTest):
    """Test driver list command."""
    @classmethod
    def setUpClass(cls):
        with mock.patch('camelot.driver._main.get_config_file') as get_rs_file:
            get_rs_file.return_value = None
            # This call is very expensive, so I'm caching the result and mocking it in the tests
            cls.resources_dm = _list_command.get_resources_dm()

    def setUp(self):
        super(TestListCommand, self).setUp()
        self.valence_path = 'valence.unittest_utils'
        self.cases_path = self.valence_path + '.cases'
        self.suites_path = self.valence_path + '.suites'
        self.passing_suite_path = self.suites_path + '.PassingSuite'
        self.passing_case_path = self.cases_path + '.PassingCase'
        self.empty_suite_path = self.suites_path + '.EmptySuite'

    @staticmethod
    def assert_modules_in_output(output):
        assert '== Modules ==' in output

    @staticmethod
    def assert_cases_in_output(output):
        assert '== Cases ==' in output

    @staticmethod
    def assert_suites_in_output(output):
        assert '== Suites ==' in output

    @mock.patch('camelot.driver._main.get_config_file')
    def test_get_resources_dm_no_file(self, mock_get_config_file):
        mock_get_config_file.return_value = None
        resources_dm = _list_command.get_resources_dm()
        # Nodes should not be an empty list because it should load the dummy resources
        assert resources_dm.nodes

    @mock.patch('camelot.driver._main.get_config_file')
    def test_get_resources_with_file(self, mock_get_config):
        mock_get_config.return_value = DATA_DIR + '/' + _constants.DEFAULT_SUITE_RESOURCES_FILE
        resources_dm = _list_command.get_resources_dm()
        assert resources_dm.node_dms
        assert resources_dm.cluster_dms

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_modules(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.valence_path)
        cmd.get_members()
        assert unittest_utils.cases in cmd.modules
        assert unittest_utils.suites in cmd.modules

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_modules_with_config(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.valence_path, config=models.SuiteConfig())
        cmd.get_members()
        assert unittest_utils.cases in cmd.modules
        assert unittest_utils.suites in cmd.modules

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_invalid_module(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.valence_path + '.invalid')
        cmd.get_members()
        assert cmd.object is None

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_invalid_module_root(self, mock_get_dm):
        """Test an invalid module at the root namespace, i.e. there's no period in the path."""
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand('someinvalid_module')
        cmd.get_members()
        assert cmd.object is None

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_invalid_class(self, mock_get_dm):
        """The module is valid, but the class is not."""
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.suites_path + '.NonExistentSuite')
        cmd.get_members()
        assert cmd.object is None

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_suites(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.suites_path)
        cmd.get_members()
        assert unittest_utils.suites.ExceptionSuite in cmd.suites

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_call_get_members_twice(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.valence_path)
        cmd.get_members()
        num_modules = len(cmd.modules)
        cmd.get_members()
        assert num_modules == len(cmd.modules)

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_single_case(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.cases_path + '.PassingCase')
        cmd.get_members()
        assert len(cmd.steps) == 2
        assert not cmd.modules
        assert not cmd.suites
        assert not cmd.cases

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_cases(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.passing_suite_path)
        cmd.get_members()
        assert unittest_utils.cases.PassingCase in cmd.cases

    @mock.patch.object(_list_command, 'get_resources_dm')
    def test_list_cases_empty_suite(self, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.empty_suite_path)
        cmd.get_members()
        assert not cmd.cases

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_modules(self, mock_stdout, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.valence_path)
        cmd.print_members()
        self.assert_modules_in_output(mock_stdout.getvalue())

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_cases(self, mock_stdout, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.cases_path)
        cmd.print_members()
        output = mock_stdout.getvalue()
        self.assert_cases_in_output(output)
        self.assert_modules_in_output(output)
        assert 'PassingCase' in output

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_steps(self, mock_stdout, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.passing_case_path)
        cmd.print_members()
        output = mock_stdout.getvalue()
        assert 'PassingCase' in output
        assert 'Step 0' in output

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_case_without_steps(self, mock_stdout, mock_get_dm):
        case_path = self.cases_path + '.PassingCaseNoSteps'
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(case_path)
        cmd.print_members()
        output = mock_stdout.getvalue()
        assert 'PassingCaseNoSteps' in output

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_suites_verbose(self, mock_stdout, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.suites_path, verbose=True)
        cmd.print_members()
        output = mock_stdout.getvalue()
        self.assert_suites_in_output(output)
        self.assert_modules_in_output(output)
        assert 'PassingSuite' in output
        assert 'Contains a single test case' in output

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_suites(self, mock_stdout, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.suites_path)
        cmd.print_members()
        output = mock_stdout.getvalue()
        self.assert_suites_in_output(output)
        self.assert_modules_in_output(output)
        assert 'PassingSuite' in output

    @mock.patch.object(_list_command, 'get_resources_dm')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_print_cases_from_suite(self, mock_stdout, mock_get_dm):
        mock_get_dm.return_value = self.resources_dm
        cmd = _list_command.ListCommand(self.passing_suite_path, verbose=True)
        cmd.print_members()
        output = mock_stdout.getvalue()
        assert '== Cases in PassingSuite ==' in output
        assert 'PassingCase' in output
        assert 'Test case that passes' in output
