"""Unit tests for vl provision"""
import unittest

import mock
import pytest
import sf_models

from camelot.driver import _provision_command
from sf_api import api_errors
from valence import (
    functional_models,
    models,
)


class TestProvisionCommand(unittest.TestCase):
    """Test driver provision command."""

    def setUp(self):
        mgmt_resource_ips = [models.IpSuiteResource.get_mock_data(overrides=dict(
            model=dict(network='1g'))) for _ in range(3)]
        strg_resource_ips = [models.IpSuiteResource.get_mock_data(overrides=dict(
            model=dict(network='10g'))) for _ in range(3)]
        nodes = [models.NodeSuiteResource.get_mock_data(force_mock=dict(config=True)) for _ in range(2)]
        for node in nodes:
            node['config']['build']['version'] = '12.0.0.1'
        resources = models.SuiteResources.create(ips=mgmt_resource_ips + strg_resource_ips,
                                                 nodes=nodes)
        self.provisioner = _provision_command.ProvisionCommand(resources,
                                                               cicd=False,
                                                               witness_num=1)

        resources = models.SuiteResources.create(ips=mgmt_resource_ips + strg_resource_ips,
                                                 clusters=[
                                                     models.ClusterSuiteResource.get_mock_object()],
                                                 nodes=nodes)
        witness_node = models.NodeSuiteResource.get_mock_object(overrides=dict(model=dict(
            model=sf_models.NodeType(sf_models.NodeType.STANDARD_TYPES.WITNESS))))
        resources.nodes.append(witness_node)
        self.transformed_provisioner = _provision_command.ProvisionCommand(resources,
                                                                           cicd=False,
                                                                           witness_num=1)
        self.transformed_provisioner.witness_node_dms.append(witness_node)
        self.transformed_provisioner.build = resources.nodes[0].config.build

    def test_validate_input(self):
        self.provisioner.validate_input_resources()

    def test_invalid_input(self):
        with pytest.raises(AssertionError):
            self.transformed_provisioner.validate_input_resources()
        self.provisioner.cluster_num = 2
        with pytest.raises(AssertionError):
            self.provisioner.validate_input_resources()

    def test_transform_resources(self):
        assert not self.provisioner.suite_resources.clusters
        assert len(self.provisioner.suite_resources.nodes) == 2
        assert len(self.provisioner.suite_resources.ips) == 6
        self.provisioner.transform_resources()
        assert len(self.provisioner.suite_resources.cluster_dms) == 1
        assert len(self.provisioner.suite_resources.nodes) == 3
        witness_dms = [node.model for node in self.provisioner.suite_resources.node_dms
                       if node.is_witness()]
        assert witness_dms
        assert len(self.provisioner.suite_resources.ips) == 2

    @mock.patch('valence.vl_helpers.create_witness_nodes')
    def test_deploy_witness(self, mock_deploy):
        self.transformed_provisioner.deploy_witnesses()
        mock_deploy.assert_called_with(self.transformed_provisioner.witness_node_dms,
                                       self.transformed_provisioner.build)

    @mock.patch('valence.vl_helpers.create_cluster')
    def test_create_cluster(self, mock_create_cluster):
        self.transformed_provisioner.create_cluster()
        mock_create_cluster.assert_called()

    def test_validate_teardown(self):
        self.transformed_provisioner.validate_teardown()

    def test_teardown_invalid(self):
        with pytest.raises(AssertionError):
            self.provisioner.validate_teardown()

    @mock.patch('valence.vl_helpers.delete_witness_vms')
    @mock.patch.object(functional_models.Node, 'set_cluster_config')
    @mock.patch.object(functional_models.Node, 'restart_binary')
    def test_teardown(self, mock_restart_services, mock_set_config, mock_delete):
        response = models.ClusterConfig.get_mock_object()
        mock_set_config.return_value = response
        mock_set_config.side_effect = api_errors.SfApiException('Error', response)
        self.transformed_provisioner.teardown()
        mock_delete.assert_called()
        mock_set_config.assert_called()
        mock_restart_services.assert_called()

    def test_no_witnesses(self):
        node = models.ResourceNode.create(model='SF3010')
        nodes = [models.NodeSuiteResource.get_mock_object(overrides=dict(model=node))]
        resources = models.SuiteResources.create(nodes=nodes)
        provisioner = _provision_command.ProvisionCommand(resources,
                                                          cicd=False,
                                                          witness_num=0)
        provisioner.deploy_witnesses()
        assert not provisioner.witness_node_dms
