import os
import unittest

import mock
import pytest
from qtestapi import models
from qtestapi.tests import utils

from camelot import driver
from camelot.driver import _qpush_command
from valence import unittest_utils

DATA_DIR = unittest_utils.get_data_dir(__file__)
REPORT_PATH = os.path.join(DATA_DIR, 'report.json')
REPORT_BAD_BUILD = os.path.join(DATA_DIR, 'report_missing_build.json')
REPORT_BAD_FEATURE = os.path.join(DATA_DIR, 'report_missing_feature.json')

MOCK_API = utils.get_mock_project()
USERNAME = 'username'
PASSWORD = 'password'
PROJECT_ID = 12345

BUILD_INFO = {
    u'allowed_values': [
       {u'is_active': True,
        u'is_default': False,
        u'label': u'Uncommitted',
        u'links': [],
        u'order': 1,
        u'value': -279290},
       {u'is_active': True,
        u'is_default': False,
        u'label': u'neon',
        u'links': [],
        u'order': 2,
        u'value': -298814},
       {u'is_active': True,
        u'is_default': False,
        u'label': u'    ember-sds-phase1-10.3.0.73',
        u'links': [],
        u'order': 3,
        u'value': 319404},
       {u'is_active': True,
        u'is_default': False,
        u'label': u'    ember-sds-phase1-10.3.0.74',
        u'links': [],
        u'order': 4,
        u'value': 319405}],
    u'attribute_type': u'Number',
    u'constrained': True,
    u'data_type': 3,
    u'free_text_search': False,
    u'id': 5236805,
    u'is_active': True,
    u'label': u'Target Release/Build',
    u'links': [{u'href': u'https://solidfire.qtestnet.com/api/v3/projects/73162/5236805',
                u'rel': u'self'}],
    u'multiple': False,
    u'order': 3,
    u'original_name': u'TargetBuildTestSuite',
    u'required': False,
    u'searchable': False,
    u'system_field': True}


class TestQPush(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.build_field = models.Field.from_dict(BUILD_INFO)

        cls.report_path = REPORT_PATH
        cls.report_bad_build = REPORT_BAD_BUILD
        cls.report_bad_feat = REPORT_BAD_FEATURE

        cls.mock_case = utils.get_ns_mock_model('test-cases', api=MOCK_API)
        cls.mock_case.name = 'SDSAddSfDemoNodeToCluster'

        cls.mock_release = utils.get_ns_mock_model('releases', api=MOCK_API)
        cls.mock_release.name = 'neon'

        cls.mock_build = utils.get_ns_mock_model('builds', api=MOCK_API)
        cls.mock_build.name = 'ember-sds-phase1-10.3.0.73'

        cls.mock_cycle = utils.get_ns_mock_model('test-cycles', api=MOCK_API)
        cls.mock_cycle.name = 'SDS'

        cls.mock_suite = utils.get_ns_mock_model('test-suites', api=MOCK_API)
        cls.mock_suite.name = 'ClusterCompSuite'
        cls.mock_suite.properties.append(cls.build_field.to_property(319404))

        cls.mock_run = utils.get_ns_mock_model('test-runs', api=MOCK_API)
        cls.mock_run.name = 'SDSAddSfDemoNodeToCluster'

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_all_resources_exist(self, mock_api):
        """Assume all external resources are already created for you."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[self.mock_release])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[self.mock_build])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[self.mock_cycle])
        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[self.mock_suite])
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[self.mock_run])

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_path, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_submit.assert_called_once()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_unknown_build(self, mock_api):
        """Build/Release information missing in report, check it goes to uncomitted."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[self.mock_cycle])
        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[self.mock_suite])
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[self.mock_run])

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_bad_build, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_submit.assert_called_once()
        mock_client.search_releases_by_name.assert_called_with('Uncommitted')
        mock_client.create_release.assert_called_once()
        mock_client.create_build.assert_called_once()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_unknown_feature(self, mock_api):
        """Feature information missing from report.json.  Check that it goes to uncomitted folder."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[self.mock_release])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[self.mock_build])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[])
        mock_create_cycle = mock.MagicMock()
        mock_client.create_test_cycle_under_release = mock_create_cycle
        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[self.mock_suite])
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[self.mock_run])

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_bad_feat, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_create_cycle.assert_called_with(self.mock_release.id, 'Uncommitted')
        mock_submit.assert_called_once()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_missing_test_case(self, mock_api):
        """Missing test case definition on qTest, should cause exception."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[])

        cmd = _qpush_command.QPushCommand(self.report_path, USERNAME, PASSWORD, PROJECT_ID)
        with pytest.raises(RuntimeError):
            cmd.post_to_qtest()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_create_new_suite(self, mock_api):
        """Assume all external resources are already created for you."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[self.mock_release])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[self.mock_build])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[self.mock_cycle])
        mock_create_suite = mock.MagicMock()
        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[])
        mock_client.get_ns_field_by_label.return_value = self.build_field
        mock_client.create_test_suite_under_cycle = mock_create_suite
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[self.mock_run])

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_path, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_submit.assert_called_once()
        mock_create_suite.assert_called_once()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_update_suite_new_build(self, mock_api):
        """Assume all external resources are already created for you."""

        mock_suite = utils.get_ns_mock_model('test-suites', api=MOCK_API)
        mock_suite.name = 'ClusterCompSuite'
        mock_suite.properties.append(self.build_field.to_property(319405))

        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[self.mock_release])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[self.mock_build])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[self.mock_cycle])

        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[mock_suite])
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[self.mock_run])

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_path, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_client.update_resource.assert_called_once_with(mock_suite)
        mock_submit.assert_called_once()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_update_suite_existing_build(self, mock_api):
        """Assume all external resources are already created for you."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[self.mock_release])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[self.mock_build])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[self.mock_cycle])

        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[self.mock_suite])
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[self.mock_run])

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_path, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_submit.assert_called_once()

    @mock.patch('qtestapi.QTestAPI', autospec=False)
    def test_create_new_run(self, mock_api):
        """Assume all external resources are already created for you."""
        mock_client = mock.MagicMock()
        mock_api.return_value = mock_client
        mock_client.search_test_cases_by_python_path = mock.MagicMock(return_value=[self.mock_case])
        mock_client.search_releases_by_name = mock.MagicMock(return_value=[self.mock_release])
        mock_client.get_builds_under_release = mock.MagicMock(return_value=[self.mock_build])
        mock_client.get_test_cycles_under_release = mock.MagicMock(return_value=[self.mock_cycle])
        mock_client.get_test_suites_under_cycle = mock.MagicMock(return_value=[self.mock_suite])
        mock_create_run = mock.MagicMock()
        mock_client.get_test_runs_under_suite = mock.MagicMock(return_value=[])
        mock_client.create_test_run_under_suite = mock_create_run

        mock_submit = mock.MagicMock()
        mock_client.submit_test_run_log = mock_submit

        cmd = _qpush_command.QPushCommand(self.report_path, USERNAME, PASSWORD, PROJECT_ID)
        cmd.post_to_qtest()

        mock_submit.assert_called_once()
        mock_create_run.assert_called_once()
