#!/usr/bin/env python
"""Tool to aid with docker interface.

Usage::
    vl_docker build ./staging-autotest-content
    vl_docker push
"""
import argparse
import json
import os
import re
import subprocess
from datetime import datetime

import requests

# this is the default docker registry for valence
BASE_ARTIFACTORY_DOCKER_API = 'http://sf-artifactory.solidfire.net/artifactory/api/docker'
DEFAULT_ARTIFACTORY_REPOSITORY = 'valence-docker-virt'
REGISTRY_ADDRESS = 'sf-artifactory.solidfire.net:9003'
DEFAULT_DOCKER_REPO = 'valence-local'
VERSION_REGEX_PATTERN = r"^v\d+(\.\d+)*$"


class DockerClient(object):
    """Aids with docker builds for containerized execution.

    :param str registry_address: Address of the docker registry
    """

    def __init__(self, registry_address):
        # Import docker here so this file can be imported from AT2
        import docker
        self.registry_address = registry_address
        self.client = docker.from_env()

    @property
    def api(self):
        return self.client.api

    def push_image(self, repo_name, tag):
        """Push a tagged image to docker registry.

        :param str repo_name: Name of the docker repo for the image
        :param str tag: Tag for the image
        """
        repo = '{}/{}'.format(self.registry_address, repo_name)
        image_name = '{}:{}'.format(repo, tag)
        print('Pushing image: {}'.format(image_name))
        for line in self.api.push(repo, tag=tag, stream=True):
            print(line.strip())

    def build_image(self, repo_name, tag, tests_repo_path, dockerfile_path, build_args):
        """Build a docker image.

        :param str repo_name: Name of the docker repo for the image
        :param str tag: Tag for the image
        :param str tests_repo_path: Path to the tests repo
        :param str dockerfile_path: Path to the Dockerfile
        :param dict|None build_args: Build time variables
        :rtype: docker.models.images.Image
        """
        tests_repo_absolute_path = os.path.realpath(tests_repo_path)
        build_context = get_build_context(tests_repo_absolute_path)
        os.chdir(build_context)
        # Get the tests repo path related to the build context
        tests_repo_path = os.path.relpath(tests_repo_absolute_path, start=build_context)

        build_args = build_args or {}
        build_args['TESTS_REPO'] = tests_repo_path

        image_name = '{}/{}:{}'.format(self.registry_address, repo_name, tag)
        print('Building image: {}'.format(image_name))
        print('Dockerfile: {}'.format(dockerfile_path))
        print('Tests Repo: {}'.format(tests_repo_absolute_path))
        for line in self.api.build(path='.', dockerfile=dockerfile_path,
                                   tag=image_name, buildargs=build_args):
            print(line.strip())


def get_build_context(tests_repo_path):
    """Get the docker build context.

    The build context needs to be the parent directory of the tests_repo to avoid the "Forbidden path
    outside the build context" Docker error. This allows the user to run vl_docker build from any
    directory.

    :param str tests_repo_path: Path to the tests repo
    :rtype: str
    """
    return os.path.dirname(tests_repo_path.rstrip('/'))


def get_default_dockerfile():
    """Gets the path to Dockerfile.local in the valence repo.

    :rtype: str
    """
    script_path = os.path.realpath(__file__)
    root_path = os.path.dirname(os.path.dirname(os.path.dirname(script_path)))
    dockerfile = os.path.join(root_path, 'cicd', 'Dockerfile.local')
    return dockerfile


def get_tag():
    """Get a tag based on the username and date.

    :rtype: str
    """
    username = subprocess.check_output(['whoami']).strip()
    date = datetime.now().strftime('%m-%d-%y')
    return '{}-{}'.format(username, date)


def get_tags(artifactory_repository, image_name, username, password):
    """Grabs the list of tags from artifactory.

    :param str artifactory_repository: Docker repository. EX. Valence-docker-local-den
    :param str image_name: Image name in repository. EX Valence
    :param str username: username to access artifactory
    :param str password: password to access artifactory
    :rtype: list[str]
    """
    # Grab the list of tags from artifactory
    url = '{}/{}/v2/{}/tags/list'.format(
        BASE_ARTIFACTORY_DOCKER_API, artifactory_repository, image_name
    )
    try:
        response = requests.get(url, auth=(username, password))
        # If the response was successful, no Exception will be raised
        response.raise_for_status()
    except Exception as err:
        print('HTTP error occurred: {}'.format(err))
        raise

    return json.loads(response.content.decode('utf-8')).get('tags')


def get_latest_tag_version(tags):
    """Gets the highest alphanumerically sorted tag number.

    :param list tags: list of tags returned from artifactory
    :rtype: str
    """
    # Regex pattern: EX: v1, v1.1, v10.1.1.1, etc.
    tags = [tag for tag in tags if re.match(VERSION_REGEX_PATTERN, tag)]
    if tags:
        tags.sort(key=lambda s: list(map(int, s.lstrip('v').split('.'))))
        return tags[-1]
    else:
        return 'v0'


def bump_version(version, version_level=None):
    """Bumps up version. Will bump up minor version unless version level is specified.

    The version level goes from left to right. So version_level = 0 bumps the left most.
    Version_level = length(version_components) - 1 would bump the right most.

    :param str version: starts with v followed by an int and any number of ints and periods
    :param int version_level: Level of version to bump. EX version_level=1: v1.1.1 - > v1.2.1

    :rtype: str
    """
    # Might as well check that the version still matches the right format
    if re.match(VERSION_REGEX_PATTERN, version):
        version = version.lstrip('v')
        if '.' in version:
            version_components = version.split('.')
            if isinstance(version_level, int):
                try:
                    version_components[version_level] = str(int(version_components[version_level]) + 1)
                except IndexError:
                    print('Version number is invalid. Use version format v<num>[<.><num>]')
                    raise
            else:
                version_components[-1] = str(int(version_components[-1]) + 1)
            return 'v{}'.format('.'.join(version_components))
        else:
            return 'v{}'.format(int(version) + 1)


class StoreDictKeyPair(argparse.Action):
    """Custom Action class for argparse.

    This action will parse a list of strings '<varname>=<value>' to {<varname>: <value>}
    in a dictionary.

    For example:

    '--build-arg <varname0>=<value0> <varname1>=<value1> --build-arg <varname2>=<value2>'

    will result in

    {'<varname0>': '<value0>', '<varname1>': '<value1>', '<varname2>': '<value2>'}

    :rtype: dict
    """

    def __call__(self, parser, namespace, values, option_string=None):
        build_args = getattr(namespace, self.dest) or {}
        for key_val in values:
            key, val = key_val.split('=')
            build_args[key] = val
        setattr(namespace, self.dest, build_args)


def create_parser():
    """Factory for creating an ArgumentParser object.

    :rtype: argparse.ArgumentParser
    """
    default_tag = get_tag()
    default_dockerfile = get_default_dockerfile()
    parser = argparse.ArgumentParser(description='Helps valence users build and deploy docker images',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(title='commands', dest='subparser_name')
    upload_parser = subparsers.add_parser('push',
                                          help='This feature pushes images to a docker registry',
                                          formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    upload_parser.add_argument('-r', '--registry-address',
                               help='URL of the docker registry (do not prepend with http/www)',
                               dest='registry_address', default=REGISTRY_ADDRESS)
    upload_parser.add_argument('--repository', help='Name of the docker repository',
                               default=DEFAULT_DOCKER_REPO)
    upload_parser.add_argument('-t', '--tag', help='The tag of the image', default=default_tag)
    upload_parser.add_argument('-i', '--increment-package-version',
                               help='Bump and use the latest package version',
                               default=False, action='store_true')

    build_parser = subparsers.add_parser('build',
                                         help='Build a docker image from a local tests repo',
                                         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    build_parser.add_argument('tests_repository', help='Path to a tests repo that contains the camelot '
                                                       'tests to run and a requirements.txt')
    build_parser.add_argument('--repository', help='Name of the docker repository',
                              default=DEFAULT_DOCKER_REPO)
    build_parser.add_argument('-d', '--dockerfile', help='Path to the Dockerfile to use',
                              default=default_dockerfile)
    build_parser.add_argument('-i', '--increment-package-version',
                              help='Bump and use the latest package version',
                              default=False, action='store_true')
    build_parser.add_argument('-t', '--tag', help='The tag of the image', default=default_tag)
    build_parser.add_argument('-r', '--registry-address',
                              help='URL of the docker registry (do not prepend with http/www)',
                              dest='registry_address', default=REGISTRY_ADDRESS)
    build_parser.add_argument('--build-args', help='Set build time variables', nargs='*',
                              action=StoreDictKeyPair, metavar='<varname>=<value>')
    return parser


def main():
    """Driver function."""
    parser = create_parser()
    arguments = parser.parse_args()

    client = DockerClient(registry_address=arguments.registry_address)
    tag_to_use = arguments.tag
    if arguments.increment_package_version:
        # Grab the current artifactory tags for image
        # Artifactory repositories are mirrored so it should be fine to use default
        # Going to want to get credentials out of code probably
        artifactory_tags = get_tags(DEFAULT_ARTIFACTORY_REPOSITORY,
                                    arguments.repository,
                                    'jenkins',
                                    'solidfire')
        # Grab the latest based on versioning scheme
        current_latest_tag = get_latest_tag_version(artifactory_tags)
        # Set to tag to use to the bumped up version
        tag_to_use = bump_version(current_latest_tag)
    if arguments.subparser_name == 'push':
        client.push_image(repo_name=arguments.repository, tag=tag_to_use)
    if arguments.subparser_name == 'build':
        client.build_image(repo_name=arguments.repository, tag=tag_to_use,
                           tests_repo_path=arguments.tests_repository,
                           dockerfile_path=arguments.dockerfile,
                           build_args=arguments.build_args)


if __name__ == '__main__':
    main()
