"""Code for inspecting test cases and suites.

Usage Examples
==============
Get list of suites in a module:

>>> import camelot
>>> from valence import suites
>>> suite_classes = camelot.get_suites_from_module(suites)

Get list of cases in a module:

>>> from valence import cases
>>> case_classes = camelot.get_cases_from_module(cases)

Get a list of steps in a test case:

>>> steps = camelot.get_steps_from_case(cases.PassingCase)

Print steps in a test case along with their source code:

>>> camelot.print_steps(cases.PassingCase, print_source=True)
"""

from __future__ import print_function

import ast
import hashlib
import inspect
import pydoc

import autologger
import six  # pylint: disable=unused-import

from camelot import (
    _case,
    _errors,
    _suite,
)

_log = autologger.get_sflogging(__name__)

LOG_STEP_METHOD = '_log_step'
TEST_METHOD = '_test'
RETURN_INSIDE_TEST_ERROR = 'Do not call return inside _test()'
SUPER_TEST_CALLED_IN_TEST = 'Do not call super._test() inside _test().'


class StepInfo(_case.TestStep):
    """Extends TestStep with extra info.

    :param int index: Index of the step in the test case
    :param str action: Step action
    :param str expected_result: The expected result of the action
    :param list[str] source_lines: Lines of source code associated with step
    :param int starting_line: Starting line relative to test case that implements _test
    :param int ending_line: Ending line relative to test case that implements _test
    """
    def __init__(self, index, action, expected_result, source_lines=None,
                 starting_line=None, ending_line=None):
        super(StepInfo, self).__init__(index, action, expected_result)
        if source_lines is None:
            source_lines = []
        self.source_lines = source_lines
        self.starting_line = starting_line
        self.ending_line = ending_line

    @property
    def source(self):
        """Return source code associated with step.

        :rtype: str
        """
        return ''.join(self.source_lines)


class CaseInspectorProblem(object):
    """Error class for test case error in CaseInspector.

    :param str module_name: The module name of the test case
    :param str case_name: The test case name
    :param str error_msg: Test case error message
    :param int line_number: Line number of error message
    :param str case_file: The location of test module
    :param str source: The source code around error message
    """
    def __init__(self, module_name, case_name, error_msg, line_number=None, case_file=None,
                 source=None):
        self.module_name = module_name
        self.case_name = case_name
        self.error_msg = error_msg
        self.line_number = line_number or 'N/A'
        self.case_file = case_file
        self.source = source

    def __str__(self):
        """A string representation of the CaseInspectorProblem object.

        :rtype: str
        """
        return '{}:{}:{} --- {}'.format(self.module_name, self.case_name, self.line_number,
                                        self.error_msg)


class CaseInspector(object):
    """Inspect a camelot test case.

    :param Any test_class: Class of test case to analyze
    """
    def __init__(self, test_class):
        if not inspect.isclass(test_class) or not issubclass(test_class, _case.TestCase):
            raise TypeError('{} is not a subclass of camelot.TestCase'.format(test_class))
        self.test_case_class = test_class
        self.test_case_node = self._get_case_node(self.test_case_class)

        # Get parent classes of test_case_class to search for inherited test cases.
        self.all_cases = self._get_all_cases()
        # Find node for _test method
        self._find_test_method_node()

        # Default arguments for _log_step
        log_step_method = getattr(self.test_case_class, LOG_STEP_METHOD)
        try:
            defaults = (inspect.signature(log_step_method).parameters['expected_result'].default, )
        except AttributeError:  # pragma: no cover
            # getargspec was deprecated in Python 3 and removed in 3.6
            # pylint: disable=deprecated-method
            defaults = inspect.getargspec(log_step_method).defaults
        self._log_step_defaults = defaults
        self.test_case_errors = []

        # Cache the source lines so we don't re-read them over and over
        self.cached_source_lines = []
        for case_class in self.all_cases:
            self.cached_source_lines.append(inspect.getsourcelines(case_class))

    def get_steps(self):  # pylint: disable=too-complex
        """Get steps associated with test case.

        This method walks through the inheritance tree collecting logsteps.
        The valence standard is not to call super._test() therefore we're only going to scribe the
        first occurrence of test().

        :rtype: list[camelot.inspection.StepInfo]
        """
        steps = []
        for case_index, case_class in enumerate(self.all_cases):
            case_node = self._get_case_node(case_class)
            step_nodes, test_implemented = self._get_log_step_nodes(case_node)

            test_super_error = self._validate_super_test(case_node, case_index)
            if test_super_error:
                self.test_case_errors.append(test_super_error)

            test_return_error = self._validate_return_node(case_node, case_index)
            if test_return_error:
                self.test_case_errors.append(test_return_error)

            # Get sorted list of ast.Expr objects for each _log_step node
            for step_index, node in enumerate(step_nodes):
                step = self._get_step_object_from_node(node, step_index)
                test_step_error = self._validate_log_step_node(node, case_index)
                if test_step_error:
                    self.test_case_errors.append(test_step_error)

                starting_line = node.lineno
                if step_index == len(step_nodes) - 1:
                    # The ending line of the last step is the final line of _test()
                    ending_line = len(self.cached_source_lines[case_index][0])
                else:
                    # Else, ending line is right before the next step
                    ending_line = step_nodes[step_index + 1].lineno - 1
                source_lines = self._get_source_lines(starting_line, ending_line, case_index)
                step.source_lines = source_lines
                step.starting_line = starting_line
                step.ending_line = ending_line
                steps.append(step)

            # If this node has implemented test() then we're done scribing.
            if test_implemented:
                break

        if self.test_case_errors:
            raise _errors.InspectionError
        return steps

    def get_source_hash(self):
        """Attempts to calculate hash of source lines for all related test_cases.

        :rtype: str
        """
        source_lines = list()
        for case_index, _ in enumerate(self.all_cases):
            source_lines.extend(self.cached_source_lines[case_index][0])
        source_blob = '\n'.join(source_lines)

        return hashlib.md5(source_blob.encode()).hexdigest()

    def _get_source_lines(self, starting_line, ending_line, case_index):
        """Get a range of source code lines.

        :param int starting_line: Starting line number relative to the test case
        :param int ending_line: Ending line number relative to the test case
        :param int case_index: The current test case that is being inspected
        :rtype: list[str]
        """
        lines = self.cached_source_lines[case_index][0][starting_line: ending_line]
        lines = [line.partition('    ')[2] for line in lines]
        return lines

    @staticmethod
    def _node_is_try_py2_and_py3(node):
        """Checks whether node is an instance of ast.Try, TryFinally or tryExcept.

        The actual attribute in ast depends on python version.

        :param ast.Expr node: The ast node being checked.
        :rtype: bool
        :returns: whether ast node is a Try node.
        """
        # This returns whether node is an ast.Try node or not.
        # ast.Try indicates that this particular statement is try:
        # Unfortunately, in py2.7 they had ast.TryFinally and ast.TryExcept, but in 3.6 they changed
        # to ast.Try only.
        # So pylint will now fail if I try to use version-specific options, hence the use of hasattr
        # and pydoc.locate().
        # The reason for the giant logical statement is b/c unit test coverage will be 99%
        # if we split this line into py2 and py3 specific statements, since there's no way to write
        # unit tests in 2.7 that utilize 3.6 specific features.
        # This way pytest is being duped into believing that this line of code is always covered.
        return ((hasattr(ast, 'Try') and isinstance(node, pydoc.locate('ast.Try'))) or
                (hasattr(ast, 'TryFinally') and isinstance(node, pydoc.locate('ast.TryFinally'))) or
                (hasattr(ast, 'TryExcept') and isinstance(node, pydoc.locate('ast.TryExcept'))))

    def _make_error(self, node, case_index, error_msg):
        """Makes an CaseInspectorProblem object for given node.

        :param ast.Expr|ast.Return node: The ast node representing the _log_step() or return expression
        :param int case_index: The current test case that is being inspected
        :param str error_msg: Error message text.
        :rtype: camelot.inspection.CaseInspectorProblem|None
        :returns: Returns CaseInspectorProblem if there is an error; returns None otherwise
        """
        parent = node.parent
        source = ''.join(self._get_source_lines(parent.lineno - 1, node.lineno, case_index))
        lineno = self.cached_source_lines[case_index][1] + node.lineno - 1
        case_file = inspect.getfile(self.all_cases[case_index])
        return CaseInspectorProblem(module_name=self.test_case_class.__module__,
                                    case_name=self.test_case_class.__name__, error_msg=error_msg,
                                    line_number=lineno, case_file=case_file, source=source)

    def _validate_log_step_node(self, node, case_index):
        """Validates that _log_step was called correctly.

        It is not valid to call it outside of _test(), in a loop, or in a conditional.

        :param ast.Expr node: The ast node representing the _log_step() expression
        :param int case_index: The current test case that is being inspected
        :rtype: camelot.inspection.CaseInspectorProblem|None
        :returns: Returns CaseInspectorProblem if there is an error; returns None otherwise
        """
        parent = node.parent
        if isinstance(parent, ast.FunctionDef) and parent.name == TEST_METHOD:
            # The parent of the _log_step() call is _test()
            return None
        # We allow _log_step inside try, or with, or combination of both, inside of _test().
        elif isinstance(parent, ast.With) and self._get_antecedent_name(parent) == TEST_METHOD:
            # If the parent is not the _test() node, it can be inside a with statement
            return None
        elif self._node_is_try_py2_and_py3(parent) and self._get_antecedent_name(parent) == TEST_METHOD:
            return None
        elif isinstance(parent, ast.For) or isinstance(parent, ast.While):
            error_msg = 'Do not call _log_step inside of a loop'
        elif isinstance(parent, ast.If) or isinstance(parent, ast.IfExp):
            error_msg = 'Do not call _log_step inside of a conditional'
        else:
            error_msg = 'Do not call _log_step outside of _test method'
        return self._make_error(node, case_index, error_msg)

    def _validate_return_node(self, case_node, case_index):
        """Validates that _test() doesn't have returns in them.

        It is not valid to call return inside of _test() since this can create conditional logic.

        :param ast.ClassDef case_node: Node used to search for return calls
        :param int case_index: The current test case that is being inspected
        :rtype: camelot.inspection.CaseInspectorProblem|None
        :returns: Returns CaseInspectorProblem if there is an error; returns None otherwise
        """

        # ast.walk yields all descendant nodes recursively, with no context such as the parent
        for node in ast.walk(case_node):
            # Set parent node for each child
            for child in ast.iter_child_nodes(node):
                child.parent = node
            # Find nodes that are an expression. _log_step() calls will be contained in an expression.
            if isinstance(node, ast.Return):
                if self._get_antecedent_name(node) == TEST_METHOD:
                    return self._make_error(node, case_index, RETURN_INSIDE_TEST_ERROR)

    def _validate_super_test(self, case_node, case_index):
        """Validates that _test() doesn't call super._test().

        :param ast.ClassDef case_node: Node used to search for return calls
        :param int case_index: The current test case that is being inspected
        :rtype: camelot.inspection.CaseInspectorProblem|None
        :returns: Returns CaseInspectorProblem if there is an error; returns None otherwise
        """

        # ast.walk yields all descendant nodes recursively, with no context such as the parent
        for node in ast.walk(case_node):
            for child in ast.iter_child_nodes(node):
                child.parent = node

            if (isinstance(node, ast.Expr) and isinstance(node.value, ast.Call) and
                    getattr(node.value.func, 'attr', None)):
                if node.value.func.attr == TEST_METHOD:
                    return self._make_error(node, case_index, SUPER_TEST_CALLED_IN_TEST)

    def _get_antecedent_name(self, node):
        """Returns the name of nodes' first parent/grandparent/antecedent that is an ast.FunctionDef.

        This will recursively search the node tree up until it encounters a first ast.FunctionDef node
        and return its' name.

        :param ast.Expr|ast.Return|ast.With node: The ast node representing the _log_step() or return
            expression
        :rtype: str
        """
        parent = getattr(node, 'parent', None)
        if parent:
            if isinstance(parent, ast.FunctionDef):
                return parent.name
            else:
                return self._get_antecedent_name(parent)

    @staticmethod
    def _get_log_step_nodes(case_node):
        """Get the ast nodes for each _log_step call and set their parent node.

        :param ast.ClassDef case_node: Node used to search for _log_step calls
        :rtype: tuple[list[ast.Expr],bool]
        """
        result = []
        test_case_implements_test = False

        # ast.walk yields all descendant nodes recursively, with no context such as the parent
        for node in ast.walk(case_node):
            # Set parent node for each child
            for child in ast.iter_child_nodes(node):
                child.parent = node
            # Does this method actually implement _test()?
            if isinstance(node, ast.FunctionDef) and node.name == TEST_METHOD:
                test_case_implements_test = True
            # Find nodes that are an expression. _log_step() calls will be contained in an expression.
            if isinstance(node, ast.Expr):
                # Get the value of the expression
                call = node.value
                # Check if it's a _log_step() method call, which indicates the start of a step.
                if isinstance(call, ast.Call) and getattr(call.func, 'attr', None):
                    if call.func.attr == LOG_STEP_METHOD:
                        result.append(node)

        return sorted(result, key=lambda x: x.lineno), test_case_implements_test

    def print_steps(self, print_source=False):
        """Print steps.

        :param bool print_source: If True, print the source code of each step
        """
        for step in self.get_steps():
            print(step)
            if print_source:
                print(step.source)
        print('')

    def _parse_format_arg(self, arg, default='{}'):
        """Parse an argument to str.format and attempt to resolve to a string value.

        :param Any arg: Argument that was passed to str.format()
        :param str default: Default value to use if argument cannot be resolved to a string
        :rtype: str
        """
        if isinstance(arg, ast.Attribute):
            # Argument should be a static variable. Get the arg from the test case.
            try:
                arg = getattr(self.test_case_class, arg.attr)
            except AttributeError:
                # It's not possible to get the arg statically, so restore the original string
                arg = default
        elif isinstance(arg, ast.Str):
            arg = arg.s
        else:
            # Arg is of unsupported type, so restore the original string
            arg = default
        return arg

    def _parse_format_call(self, call):
        """Parse a str.format call and attempt to return the full formatted string.

        :param ast.Call call: Ast object of of a str.format call
        :rtype: str
        """
        # Positional args to .format()
        args = call.args[:]

        # Keyword args
        keywords = call.keywords

        # The raw string that's being formatted
        template_str = call.func.value.s

        # String values to pass to .format()
        format_args = []
        format_kwargs = {}

        for keyword in keywords:
            value = keyword.value
            format_kwargs[keyword.arg] = self._parse_format_arg(
                value, default='{' + keyword.arg + '}')
        for arg in args:
            format_args.append(self._parse_format_arg(arg))
        return template_str.format(*format_args, **format_kwargs)

    def _get_val_from_arg(self, arg):
        """Get value of argument that was passed to _log_step.

        :param Any arg: Argument that was passed in
        :rtype: str
        """
        if isinstance(arg, ast.Str):
            # Argument was passed as raw string
            return arg.s

        elif isinstance(arg, ast.BinOp):
            # Argument is of type BinOp
            error_msg = 'Use .format() instead for string concatenation'
            self.test_case_errors.append(CaseInspectorProblem(
                module_name=self.test_case_class.__module__,
                case_name=self.test_case_class.__name__,
                error_msg=error_msg,
                line_number=arg.lineno))
            raise _errors.InspectionError(error_msg)

        elif isinstance(arg, ast.Call):
            # Argument was a str.format() call
            return self._parse_format_call(arg)

        elif isinstance(arg, ast.Attribute):
            return self._parse_format_arg(arg)

        else:
            error_msg = 'Something is wrong with your _log_step args is:{}, should be one of: [' \
                        'ast.Call, ast.Attribute...'.format(arg)
            raise _errors.InspectionError(error_msg)

    def _get_kwarg_val(self, call, keyword):
        """Get value of keyword argument that was passed to _log_step.

        :param ast.Call call: Ast node call for _log_step
        :param str keyword: keyword of kw/arg pair passed to function
        :rtype: str
        """
        kwarg = [kw for kw in call.keywords if kw.arg == keyword][0]
        return self._get_val_from_arg(kwarg.value)

    def _get_action_from_call(self, call):
        """Gets action string from _log_step call.

        :param ast.Call call: Ast node call for _log_step
        :rtype: str
        """
        if call.args:
            return self._get_val_from_arg(call.args[0])
        else:
            return self._get_kwarg_val(call, 'action')

    def _get_expected_result_from_call(self, call):
        """Gets expected_value string from _log_step call.

        :param ast.Call call: Ast node call for _log_step
        :rtype: str
        """
        # Could call _log_step('action', 'expected_result') (len == 2)
        # OR _log_step('action', expected_result='expected_result') (len == 1)
        # OR _log_step(action='action', expected_result='expected_result') (len == 0)
        # OR _log_step(action='action') (len == 0, implicit expected_result == 'Pass')
        if len(call.args) == 2:
            return self._get_val_from_arg(call.args[1])
        else:
            try:
                return self._get_kwarg_val(call, 'expected_result')
            except IndexError:
                return self._log_step_defaults[0]

    def _get_step_object_from_node(self, expr, index):
        """Get StepInfo object from ast call node.

        :param ast.Expr expr: Ast expr node for _log_step
        :param int index: index of step
        :rtype: camelot.inspection.StepInfo
        """
        call = expr.value
        action = self._get_action_from_call(call)
        expected_result = self._get_expected_result_from_call(call)

        return StepInfo(index, action, expected_result)

    @staticmethod
    def _get_case_node(case_class):
        """Get ast node for test case.

        :param case_class: Class of the test case
        :type case_class: (Any) -> camelot.TestCase
        :rtype: ast.ClassDef
        """
        return ast.parse(inspect.getsource(case_class)).body[0]

    def _get_all_cases(self):
        """Get ast nodes for all test case classes inherited by test_case_class including itself.

        :rtype: list[(Any) -> camelot.TestCase]
        """
        parent_cases = []
        parents = inspect.getmro(self.test_case_class)
        for parent in parents:
            if issubclass(parent, _case.TestCase) and parent is not _case.TestCase:
                parent_cases.append(parent)
        return parent_cases

    def _find_test_method_node(self):
        """Find if _test method is implemented.

        :rtype: None
        """

        def get_test_method_node(case_node):
            for node in case_node.body:
                if isinstance(node, ast.FunctionDef) and node.name == TEST_METHOD:
                    return node
        result = get_test_method_node(self.test_case_node)
        if result is not None:
            return
        # Inspect parent classes for _test method
        parents = inspect.getmro(self.test_case_class)
        for parent in parents:
            if issubclass(parent, _case.TestCase) and parent is not _case.TestCase:
                # Find the _test method
                # This needs the 'typing' module to fix properly, ignore for now
                # noinspection PyTypeChecker
                result = get_test_method_node(self._get_case_node(parent))
                if result is not None:
                    return
        raise TypeError('Unable to find _test method on {}'.format(self.test_case_class))


def print_steps(case_class, print_source=False):
    """Print steps in a test case.

    :param Any case_class: Class of test case
    :param bool print_source: If True, print the source code of each step
    """
    CaseInspector(case_class).print_steps(print_source)


def get_steps_from_case(case_class):
    """Get steps in given case.

    :param Any case_class: Class of test case
    :return: StepInfo objects
    :rtype: list[camelot.inspection.StepInfo]
    """
    return CaseInspector(case_class).get_steps()


def get_cases_from_module(mod):
    """Get test case classes under namespace of given module.

    :param Any mod: Module to inspect
    :rtype: list[six.class_types]
    """
    cases = []
    for member in inspect.getmembers(mod):
        if not inspect.isclass(member[1]) or inspect.isabstract(member[1]):
            continue
        if issubclass(member[1], _case.TestCase):
            cases.append(member[1])
    return cases


def get_suites_from_module(mod):
    """Get test suite classes under namespace of given module.

    :param Any mod: Module to inspect
    :rtype: list[six.class_types]
    """
    suites = []
    for member in inspect.getmembers(mod):
        if not inspect.isclass(member[1]) or inspect.isabstract(member[1]):
            continue
        if issubclass(member[1], _suite.TestSuite):
            suites.append(member[1])
    return suites
