"""Data model that defines suite-wide configuration."""

import sf_models
from sf_models import types

from camelot import _enums
from camelot.models import at2_constants


class ArtifactCredentials(sf_models.Base):
    """Artifact Credentials.

    This object is mutable, because base_dir may need to be updated by the suite based on the way it's
    executed.
    """
    # valence specific
    using_valence = types.BooleanType(serialized_name='usingValence')

    # local settings
    local = types.BooleanType()
    base_dir = types.StringType()

    # At2 Settings
    username = types.StringType(serialized_name=at2_constants.USERNAME)
    queue_daemon_uuid = types.StringType(serialized_name=at2_constants.QUEUE_DAEMON_UUID)
    location_name = types.StringType(serialized_name=at2_constants.LOCATION_NAME)
    autotest_host = types.StringType(serialized_name=at2_constants.AUTOTEST_HOST)
    password = types.StringType(serialized_name=at2_constants.PASSWORD)
    task_instance_step_id = types.IntType(serialized_name=at2_constants.TASK_INSTANCE_STEP_ID)

    def __repr__(self):
        return '{}<task_instance_step_id={}, base_dir={}, local={}>'.format(
            self.__class__.__name__, self.task_instance_step_id, self.base_dir, self.local)


class HipChatAPI(sf_models.Base):
    """Configure Parameters for HipChat API."""
    room_id_or_name = types.StringType()
    token = types.StringType()

    def __repr__(self):
        return '{}<room_id_or_name={}, token={}>'.format(self.__class__.__name__, self.room_id_or_name,
                                                         self.token)


class RepoChangeset(sf_models.Base):
    """A changeset of a repository."""
    changeset_id = types.StringType(required=True)
    repo_name = types.StringType(required=True)

    def __repr__(self):
        return '{}<changeset_id={}, repo_name={}>'.format(
            self.__class__.__name__, self.changeset_id, self.repo_name)


class CamelotLogFilter(sf_models.Base):
    """Defines a log filter."""
    create_file = types.BooleanType(default=False)
    exclude_regexes = types.ListType(types.StringType(), default=['sf_api.base.call_method_raw'])
    match_regexes = types.ListType(types.StringType(), default=list)
    output_log = types.StringType(default='filtered.test.log')

    def __repr__(self):
        return '{}<output_log={}>'.format(
            self.__class__.__name__, self.output_log)


class SuiteConfig(sf_models.Base):
    """Camelot Suite Configuration.

    Indices are zero based, where start_index is inclusive, and stop_index is exclusive. e.g.
    [start_index, stop_index), where [4,5) would only run test 4.

    artifact_credentials
        The AT2 artifact credentials object. If empty, artifacts are stored locally in ~/tmp/artifacts.
    run_suite_teardown
        Enable/disable the suite teardown. This allows you to check your resources after tests complete.
    stop_on_failure
        If True, this will cause any exception in a test case to be raised so the suite will stop
        execution.
    start_index
        The test case to start at in a suite. Start index is 0 and inclusive.
    stop_index
        The test to stop at in the suite. Stop index is exclusive. Meaning test at this index will
        not run.
    run_indices
        Run specific tests. Allows you to run the same test case sequentially, or in a different order
        than the suite was originally constructed. Overrides start_index and stop_index. However,
        skip_indices is still honored.
    skip_indices
        Indices of test cases to skip.
    logging_level
        The default log level to provide to logbook. No logs under this level will be logged.
        Valid levels are: INFO, DEBUG, WARNING, ERROR, CRITICAL.
    store_artifacts
        Used to specify under what conditions artifacts from tests should be saves. Options include:
        on_fail - only store artifacts if there was a failure.
        always - always store artifacts.
        never - never store artifacts.
    random_seed
        Not yet implemented.
    store_report
        If true, write out report and summary files when suite finishes.
    changesets
        This field is automatically generated when a suite is run based on changeset_modules.
    changeset_modules
        A list of python modules that are in local repositories. The changesets field will be set
        at runtime based on this field. The changeset of each repository that contains these python
        modules will be saved as part of the report file.
    hipchat
        The Parameters that are required for the HipChatAPI.
    hci_config
        An entry for HCI-specific suite config info. Overridden in the sfprime/system_test
        base test suite.
    mnode2x_config
        Dict of config options for suites that utilize mnode2x.
    upgrade_package_url
        URL of a package to be applied to a cluster.
    """
    artifact_credentials = types.ModelType(ArtifactCredentials, default=ArtifactCredentials,
                                           required=True)
    run_suite_teardown = types.BooleanType(serialized_name='run_suite_teardown',
                                           deserialize_from=['teardown'], default=True)
    stop_on_failure = types.BooleanType(default=False)
    start_index = types.IntType()
    stop_index = types.IntType()
    run_indices = types.ListType(types.IntType(), default=list)
    skip_indices = types.ListType(types.IntType(), default=list)
    logging_level = types.EnumType(_enums.LoggingLevel, default=_enums.LoggingLevel.debug)
    store_artifacts = types.EnumType(_enums.SaveState, default=_enums.SaveState.on_fail)
    random_seed = types.IntType()

    # Driver specific fields
    store_report = types.BooleanType(default=True)
    changesets = types.ListType(types.ModelType(RepoChangeset), default=list)
    # These are modules that are inside repos where changesets should saved
    changeset_modules = types.ListType(types.StringType, default=['valence', 'sf_platform', 'steps'])
    log_filters = types.ListType(types.ModelType(CamelotLogFilter, default=dict),
                                 default=[CamelotLogFilter()])  # type: list
    hipchat = types.ModelType(HipChatAPI, required=False, default=HipChatAPI)

    # HCI-specific suite config options. Overridden in sfprime/system_test code.
    hci_config = types.DictType(types.BaseType(), required=False)
    # MNode2x-specific config options.
    mnode2x_config = types.DictType(types.BaseType(), required=False)
    # Link to a package to apply to a cluster.
    upgrade_package_url = types.URLType(fqdn=True, verify_exists=False, required=False)

    def __repr__(self):
        return '{}'.format(
            self.__class__.__name__)
