"""Resource data models that are used by camelot suites."""
import pprint
import typing  # pylint: disable=unused-import

import sf_models
from sf_models import types

from camelot.models import (
    _ontap,
    _witness_resources,
    at2_enums,
)


class BaseSuiteResource(sf_models.Base):
    """A suite resource consists of a data model and a config model.

    The model is always required, but some child classes do not require a config.
    """
    model = None
    config = None


class ClusterSuiteConfig(sf_models.Base):
    """Cluster Configuration."""
    upgrade_builds = types.ListType(types.ModelType(sf_models.Build))
    build = types.ModelType(sf_models.Build)
    options = types.ListType(types.StringType)
    rtfi_flags = types.ListType(types.StringType)
    sfinstall_allow_flags = types.ListType(types.StringType)
    sfinstall_method = types.StringType()
    sfinstall_options = types.StringType()
    unique_id = types.StringType()
    virtual_ips = types.ListType(types.IPAddressType)


class ClusterSuiteResource(BaseSuiteResource):
    """Cluster Suite Resource."""
    model = types.ModelType(sf_models.ResourceCluster, required=True)
    config = types.ModelType(ClusterSuiteConfig)


class ManagementNodeSuiteConfig(sf_models.Base):
    """Fdva Configuration."""
    update_repo = types.StringType()
    healthtools_version = types.StringType()
    healthtools_down_versions = types.ListType(types.StringType)
    build_install_sequence = types.ListType(types.ModelType(sf_models.Build))

    # IN PROCESS OF DEPRECATING IN FAVOR OF BUILD_INSTALL_SEQUENCE
    build = types.ModelType(sf_models.Build)


class ManagementNodeSuiteResource(BaseSuiteResource):
    """Fdva Suite Resource."""
    model = types.ModelType(sf_models.ResourceManagementNode, required=True)
    config = types.ModelType(ManagementNodeSuiteConfig)


class NodeSuiteConfig(sf_models.Base):
    """Node configuration."""
    build = types.ModelType(sf_models.Build)
    irtfi_params = types.StringType()
    one_gig_bond_mode = types.EnumType(at2_enums.BondMode)
    ten_gig_bond_mode = types.EnumType(at2_enums.BondMode)
    enable_root_access = types.BooleanType()
    secure_erase = types.BooleanType()
    skip_network_checks = types.BooleanType()


class NodeSuiteResource(BaseSuiteResource):
    """Node Suite resource."""
    model = types.ModelType(sf_models.ResourceNode, required=True)
    config = types.ModelType(NodeSuiteConfig)


class ClientSuiteResource(BaseSuiteResource):
    """Client Suite Resource."""
    model = types.ModelType(sf_models.ResourceClient, required=True)


class NetworkSuiteResource(BaseSuiteResource):
    """Network Suite Resource."""
    model = types.ModelType(sf_models.NetworkProfile, required=True)


# TODO: finish IPRangeSuiteResource, and rtype for property below
class IPRangeSuiteResource(BaseSuiteResource):
    """IpRange Suite Resource."""
    model = types.DictType(types.StringType)


class IpSuiteResource(BaseSuiteResource):
    """IP Suite Resource."""
    model = types.ModelType(sf_models.ResourceIp, required=True)


class VirtualNetworkSuiteResource(BaseSuiteResource):
    """Virtual Network Suite Resource."""
    model = types.ModelType(sf_models.VirtualNetwork, required=True)


# TODO: finish HypervisorSuiteResource, and rtype for property below
class HypervisorSuiteResource(BaseSuiteResource):
    """Hypervisor Suite Resource."""
    model = types.DictType(types.StringType)


class OntapSuiteResource(BaseSuiteResource):
    """ONTAP Suite Resource."""
    model = types.ModelType(_ontap.ResourceOntap, required=True)


class ComputeNodeSuiteResource(sf_models.Base):
    """ComputeNode Suite Resource."""
    model = types.ModelType(sf_models.ResourceComputeNode, required=True)


class WitnessVirtualizationResource(sf_models.Base):
    """ComputeNode Suite Resource."""
    model = types.ModelType(_witness_resources.WitnessResource, required=True)


# pylint: disable=not-an-iterable
class SuiteResources(sf_models.Base):
    """Contains a list of each Suite Resource that may be associated with a suite."""
    clients = types.ListType(types.ModelType(ClientSuiteResource),
                             default=list)  # type: typing.Iterable
    clusters = types.ListType(types.ModelType(ClusterSuiteResource),
                              default=list)  # type: typing.Iterable
    management_nodes = types.ListType(types.ModelType(ManagementNodeSuiteResource),
                                      deserialize_from=['management_nodes', 'fdvas'],
                                      default=list)  # type: typing.Iterable
    hypervisors = types.ListType(types.ModelType(HypervisorSuiteResource),
                                 default=list)  # type: typing.Iterable
    ip_ranges = types.ListType(types.ModelType(IPRangeSuiteResource),
                               default=list)  # type: typing.Iterable
    ips = types.ListType(types.ModelType(IpSuiteResource), default=list)  # type: typing.Iterable
    networks = types.ListType(types.ModelType(NetworkSuiteResource),
                              default=list)  # type: typing.Iterable
    nodes = types.ListType(types.ModelType(NodeSuiteResource),
                           default=list)  # type: typing.Iterable
    ontaps = types.ListType(types.ModelType(OntapSuiteResource),
                            default=list)  # type: typing.Iterable
    virtual_networks = types.ListType(types.ModelType(VirtualNetworkSuiteResource),
                                      default=list)  # type: typing.Iterable
    compute_nodes = types.ListType(types.ModelType(ComputeNodeSuiteResource),
                                   default=list)  # type: typing.Iterable

    @property
    def cluster_dms(self):
        """The list of cluster data models associated with the suite resources.

        :rtype: list[sf_models.ResourceCluster]
        """
        return [c.model for c in self.clusters]

    @property
    def node_dms(self):
        """The list of node data models associated with the suite resources.

        :rtype: list[sf_models.ResourceNode]
        """
        return [n.model for n in self.nodes]

    @property
    def management_node_dms(self):
        """The list of mnode data models associated with the suite resources.

        :rtype: list[sf_models.ResourceManagementNode]
        """
        return [f.model for f in self.management_nodes]

    @property
    def client_dms(self):
        """The list of client data models associated with the suite resources.

        :rtype: list[sf_models.ResourceClient]
        """
        return [c.model for c in self.clients]

    @property
    def network_dms(self):
        """The list of network data models associated with the suite resources.

        :rtype: list[sf_models.NetworkProfile]
        """
        return [n.model for n in self.networks]

    @property
    def virtual_network_dms(self):
        """The list of virtual network data models associated with the suite resources.

        :rtype: list
        """
        return [v.model for v in self.virtual_networks]

    @property
    def hypervisor_dms(self):
        """The list of hypervisor data models associated with the suite resources.

        :rtype: list
        """
        return [h.model for h in self.hypervisors]

    @property
    def ip_range_dms(self):
        """The list of ip range data models associated with the suite resources.

        :rtype: list
        """
        return [i.model for i in self.ip_ranges]

    @property
    def ip_dms(self):
        """The list of ip data models associated with the suite resources.

        :rtype: list[camelot.models.IpSuiteResource]
        """
        return [i.model for i in self.ips]

    @property
    def ontap_dms(self):
        """The list of ontap cluster models associated with the suite resources.

        :rtype: list[camelot.models.ResourceOntap]
        """
        return [o.model for o in self.ontaps]

    @property
    def compute_node_dms(self):
        """The list of compute nodes models associated with the suite resources.

        :rtype: list[sf_models.ResourceComputeNode]
        """
        return [c.model for c in self.compute_nodes]

    def __repr__(self):
        """Represent the suite resources a pretty printed dict."""
        return pprint.pformat(dict(
            clusters=self.cluster_dms,
            nodes=self.node_dms,
            clients=self.client_dms,
            management_nodes=self.management_node_dms,
            compute_nodes=self.compute_node_dms,
        ))
