"""qTest project config utilities."""
from __future__ import print_function

import configparser
import os

INI_FILE_DEFAULT_NAME = '.qtest_creds.ini'
INI_FILE_DEFAULT_PATH = os.path.join(os.path.expanduser('~/'), INI_FILE_DEFAULT_NAME)


class QTestConfig(object):
    """Attempts to locate qTest config in the user's home directory.

    :param str cfg_file_path: optionally specified path to qtest credentials file
    """
    def __init__(self, cfg_file_path=None):
        if cfg_file_path:
            self.ini_path = cfg_file_path
        else:
            self.ini_path = INI_FILE_DEFAULT_PATH

        if not os.path.isfile(self.ini_path):
            self.print_help()
            raise IOError('qTest credential INI file not found at {}'.format(self.ini_path))

        self.username = None
        self.password = None

    def get_qtest_credentials(self):
        """Attempts to read qTest credentials from settings file.

        :returns: tuple containing username, password, project_id
        :rtype: tuple
        :raises ValueError: if credentials cannot be found or parsed from INI file
        """
        config = self._get_parser()
        config.read(self.ini_path)

        try:
            self.username = config['QTEST']['username']
            self.password = config['QTEST']['password']
        except (KeyError, ValueError, configparser.ParsingError):
            self.print_help()
            raise ValueError('Valid qTest credentials not found in {}'.format(self.ini_path))

        return self.username, self.password

    @staticmethod
    def _get_parser():
        """Gets a configparser object.  Seperated for unit testing.

        :rtype: configparser.ConfigParser
        """
        return configparser.ConfigParser()

    @staticmethod
    def print_help():
        """Prints sample config INI for defining qTest credentials."""
        msg = """
        ATTENTION:  No valid qTest Project INI File found.

        In order to use qTest related functionality on the local workspace, a settings file is needed.
        An example template is provided below.

        Replace the credentials with your own and save this file to {}.

        [QTEST]
        username=first.last@netapp.com
        password=SuperSecretPassword
        """.format(INI_FILE_DEFAULT_PATH)
        print(msg)
