"""Unit tests for base test case definition."""

import mock
import pytest

import camelot
from camelot import (
    _case_result,
    _errors,
)
from valence import (
    functional_models,
    models,
)
from valence.unittest_utils import (
    cases,
    helpers,
)


class TestCaseKeyboardInterrupt(helpers.MockHistoryUnitTest):
    """Test that a KeyboardInterrupt is properly handled."""

    def test_interrupt_in_setup(self):
        case = cases.ExceptionInSetup(exception_type=KeyboardInterrupt)
        result = case.execute()
        assert case.status is camelot.ResultStatus.cancelled
        assert result.status is camelot.ResultStatus.cancelled
        assert result.end_time is not None

    def test_interrupt_in_test(self):
        case = cases.ExceptionInTest(exception_type=KeyboardInterrupt)
        result = case.execute()
        assert case.status is camelot.ResultStatus.cancelled
        assert result.status is camelot.ResultStatus.cancelled
        assert result.end_time is not None

    def test_interrupt_in_teardown(self):
        case = cases.ExceptionInTeardown(exception_type=KeyboardInterrupt)
        result = case.execute()
        assert case.status is camelot.ResultStatus.cancelled
        assert result.status is camelot.ResultStatus.cancelled
        assert result.end_time is not None

    def test_interrupt_in_post_test_assertions(self):
        case = cases.ExceptionInPostTestAssertions(exception_type=KeyboardInterrupt)
        result = case.execute()
        assert case.status is camelot.ResultStatus.cancelled
        assert result.status is camelot.ResultStatus.cancelled
        assert result.end_time is not None


class TestBaseCase(helpers.MockHistoryUnitTest):

    def setUp(self):
        super(TestBaseCase, self).setUp()

    def test_case_fails(self):
        case = cases.FailingCase()
        res = case.execute()
        assert res.status is camelot.ResultStatus.failed

    def test_instantiate_base_case(self):
        with pytest.raises(TypeError):
            camelot.TestCase()

    def test_case_passes(self):
        case = cases.PassingCase()
        res = case.execute()
        assert res.status is camelot.ResultStatus.passed

    def test_case_result_attributes(self):
        case = cases.PassingCase()
        res = case.execute()
        assert res.status is camelot.ResultStatus.passed
        assert case.current_result is res
        case.execute()
        assert case.current_result is not res

    def test_case_passes_with_assertions(self):
        case = cases.PassingCaseWithAssertions()
        res = case.execute()
        assert res.status is camelot.ResultStatus.passed

    @mock.patch.object(cases.ExceptionInSetup, '_teardown')
    @mock.patch.object(cases.ExceptionInSetup, '_test')
    def test_case_fails_in_setup(self, mock_test, mock_td):
        case = cases.ExceptionInSetup()
        res = case.execute()
        assert res.status is camelot.ResultStatus.error
        assert mock_td.called
        assert not mock_test.called

    def test_case_fails_in_teardown(self):
        case = cases.ExceptionInTeardown()
        res = case.execute()
        assert res.status is camelot.ResultStatus.error

    @mock.patch.object(cases.ExceptionInSetup, '_teardown')
    @mock.patch.object(cases.ExceptionInSetup, '_post_test_assertions')
    def test_case_stop_on_failure_is_true(self, mock_pta, mock_td):
        case = cases.FailingCase()
        with pytest.raises(Exception):
            case.execute(True)
        assert not mock_td.called
        assert not mock_pta.called

    @mock.patch.object(cases.ExceptionInSetup, '_teardown')
    @mock.patch.object(cases.ExceptionInSetup, '_post_test_assertions')
    def test_case_stop_on_failure_in_post_test_assertions_and_test(self, mock_pta, mock_td):
        case = cases.ExceptionInTestAndPostTestAssertions()
        with pytest.raises(Exception):
            case.execute(True)
        assert not mock_td.called
        assert not mock_pta.called

    @mock.patch.object(cases.ExceptionInSetup, '_teardown')
    def test_case_stop_on_failure_in_post_test_assertions(self, mock_td):
        case = cases.ExceptionInPostTestAssertions()
        with pytest.raises(Exception):
            case.execute(True)
        assert not mock_td.called

    def test_set_skipped(self):
        case = cases.PassingCase()
        with pytest.raises(Exception):
            case.execute(result=self.skip_result)

    def test_skip_test_case(self):
        case = cases.SkippedCase()
        res = case.execute()
        assert res.status is camelot.ResultStatus.skipped

    def test_case_fails_in_test_and_teardown(self):
        case = cases.ExceptionInTestAndTeardown()
        res = case.execute()
        assert res.status is camelot.ResultStatus.error

    def test_case_fails_in_post_test_assertions(self):
        case = cases.ExceptionInPostTestAssertions()
        res = case.execute()
        assert res.status is camelot.ResultStatus.failed

    def test_case_fails_in_test_and_post_test_assertions(self):
        case = cases.ExceptionInTestAndPostTestAssertions()
        res = case.execute()
        assert res.status is camelot.ResultStatus.failed

    def test_log_step_method(self):
        case = cases.PassingCase()
        cluster_resource = models.ClusterSuiteResource.get_mock_object()
        resources = models.SuiteResources.create(clusters=[cluster_resource])
        assert case._step_index == -1
        case._log_step('action')
        assert case._step_index == 0
        assert case._current_step is case._steps_ran[0]
        case._log_step('action')
        assert case._step_index == 1
        assert case._current_step is case._steps_ran[1]

    def test_log_step_passing_case(self):
        case = cases.PassingCase()
        res = case.execute()
        assert len(res.steps) == 2
        assert case._current_step is res.steps[-1]
        assert case._steps_ran[0].index == 0
        assert case._steps_ran[0].action == 'Assert True'
        assert case._steps_ran[0].expected_result == ''
        assert case._steps_ran[1].index == 1
        assert case._steps_ran[1].action == 'Assert 1 is an int'
        assert case._steps_ran[1].expected_result == 'Pass'

    def test_log_step_failing_case(self):
        case = cases.FailingCase()
        res = case.execute()
        assert len(res.steps) == 2
        assert case._current_step is res.steps[-1]
        assert case._steps_ran[1].index == 1
        assert case._steps_ran[1].action == 'Assert False'
        assert case._steps_ran[1].expected_result == 'Pass'

    def test_case_execute_result_not_pending(self):
        case = cases.PassingCase()
        res = _case_result.CaseResult(case)
        res.status = camelot.ResultStatus.running
        with pytest.raises(_errors.TestRuntimeError):
            case.execute(result=res)
