"""Test the camelot.inspection module."""

import abc
import inspect
import tempfile
import unittest

import pytest
import six

import camelot
from camelot import (
    _errors,
    inspection,
)
from valence.unittest_utils import (
    cases,
    suites,
)


class NoTestMethodCase(camelot.TestCase):
    """A test case without a _test method."""


class LogStepForLoopCase(camelot.TestCase):
    """A test case where log_step is called in a loop."""
    def _test(self):
        for _ in range(4):
            self._log_step('Wrong things')


class LogStepIfStatementCase(camelot.TestCase):
    """A test case where log_step is called in a conditional."""
    def _test(self):
        if True:
            self._log_step('Stranger things')


class LogStepTryInsideFinallyStatementCase(camelot.TestCase):
    """A test case where log_step is called in a finally block."""
    def _test(self):
        try:
            pass
        finally:
            self._log_step('Final things')


class LogStepTryInsideExceptStatementCase(camelot.TestCase):
    """A test case where log_step is called in an except block."""
    def _test(self):
        try:
            pass
        except Exception:
            self._log_step('Excepting things')


class LogStepWithStatementCase(camelot.TestCase):
    """A test case where log_step is called in a with."""
    def _test(self):
        with True:
            self._log_step('Needful things')


class LogStepTryExceptStatementCase(camelot.TestCase):
    """A test case where log_step is called in a try except."""
    def _test(self):
        try:
            self._log_step('Exceptional things')
        except Exception:
            pass


class LogStepTryFinallyStatementCase(camelot.TestCase):
    """A test case where log_step is called in a try finally."""
    def _test(self):
        try:
            self._log_step('Mechanical things')
        finally:
            pass


class LogStepWithTryFinallyStatementCase(camelot.TestCase):
    """A test case where log_step is called in a try finally."""
    def _test(self):
        with True:
            try:
                self._log_step('Deep things')
            finally:
                pass


class LogStepOutsideTestMethodCase(camelot.TestCase):
    """A test case where log_step is called outside of the _test method."""
    def _secret(self):
        self._log_step('Secret things')

    def _test(self):
        self._log_step('The thing things')
        self._secret()


class LogStepToHandleBinOp(camelot.TestCase):
    def _test(self):
        endpoint_count = '10'
        self._log_step('Create this many SnapMirror endpoints: ' + endpoint_count)


class LogStepSourceLines(camelot.TestCase):
    """A test case with source lines between _log_step calls."""
    def _test(self):
        self._log_step('Stranger Things Season 2')
        x = 10
        y = 10
        z = 10
        self._log_step('Thingamajig')
        n = 10


class LogStepMultipleInheritance(cases.PassingCase, cases.PassingCaseWithAssertions):
    """A test case that inherits from two passing cases"""


class PassingCaseContextManager(camelot.TestCase):
    """The whole test is inside a context manager."""
    def _test(self):
        with tempfile.TemporaryFile() as _:
            self._log_step('Assert {}'.format('True'))
            assert True
            self._log_step('Assert 1 is an int', 'Pass')
            assert isinstance(1, int)


class TestInspection(unittest.TestCase):
    """Test the camelot.inspection module."""
    def test_log_step_in_for_loop(self):
        with pytest.raises(camelot.InspectionError):
            camelot.get_steps_from_case(LogStepForLoopCase)

    def test_log_step_in_if_statement(self):
        with pytest.raises(camelot.InspectionError):
            camelot.get_steps_from_case(LogStepIfStatementCase)

    def test_log_step_in_with_statement(self):
        steps = camelot.get_steps_from_case(LogStepWithStatementCase)
        assert len(steps) == 1

    def test_log_step_in_try_except_statement(self):
        steps = camelot.get_steps_from_case(LogStepTryExceptStatementCase)
        assert len(steps) == 1

    def test_log_step_in_try_finally_statement(self):
        steps = camelot.get_steps_from_case(LogStepTryFinallyStatementCase)
        assert len(steps) == 1

    def test_log_step_in_with_try_statement(self):
        steps = camelot.get_steps_from_case(LogStepWithTryFinallyStatementCase)
        assert len(steps) == 1

    def test_log_step_in_except_statement(self):
        with pytest.raises(camelot.InspectionError):
            camelot.get_steps_from_case(LogStepTryInsideExceptStatementCase)

    def test_log_step_in_finally_statement(self):
        steps = camelot.get_steps_from_case(LogStepTryInsideFinallyStatementCase)
        assert len(steps) == 1

    def test_get_steps_from_case(self):
        steps = camelot.get_steps_from_case(cases.PassingCase)
        assert len(steps) == 2
        assert len(steps[0].source_lines) > 0
        assert steps[0].source.strip() == 'assert True'
        assert steps[1].source.strip() == 'assert isinstance(1, int)'
        assert steps[0].expected_result == ''

    def test_get_steps_from_case_with_context_manager(self):
        steps = camelot.get_steps_from_case(PassingCaseContextManager)
        assert len(steps) == 2

    def test_get_steps_from_case_with_module(self):
        with pytest.raises(TypeError):
            camelot.get_steps_from_case(cases)

    def test_get_steps_from_case_with_wrong_class(self):
        with pytest.raises(TypeError):
            camelot.get_steps_from_case(camelot.CaseResult)

    def test_get_steps_from_case_with_abstract_class(self):
        cases_list = camelot.get_cases_from_module(cases.abstract_child_case)
        assert len(cases_list) == 1

    def test_get_steps_from_case_with_log_step_outside_test_method(self):
        with pytest.raises(_errors.InspectionError):
            camelot.get_steps_from_case(LogStepOutsideTestMethodCase)

    def test_get_val_from_arg(self):
        with pytest.raises(_errors.InspectionError):
            camelot.get_steps_from_case(LogStepToHandleBinOp)

    def test_get_steps_from_case_source_lines(self):
        # Test if get_steps_from_case returns the appropriate source lines
        steps = camelot.get_steps_from_case(LogStepSourceLines)
        source_lines = inspect.getsourcelines(LogStepSourceLines)[0]
        source_1 = [line.partition('    ')[2] for line in source_lines[4:7]]
        source_2 = source_lines[8].partition('    ')[2]
        assert steps[0].source == ''.join(source_1)
        assert steps[1].source == ''.join(source_2)

    def test_get_steps_static_variables(self):
        steps = camelot.get_steps_from_case(cases.PassingLogStepCase)
        assert steps[0].action == 'Assert True'
        assert cases.PassingLogStepCase.REASON in steps[1].action
        assert str(cases.PassingLogStepCase.NUMBER) in steps[1].action
        assert '{unknown}' in steps[3].action

    def test_get_cases_from_module(self):
        cases_list = camelot.get_cases_from_module(cases)
        assert len(cases_list) > 0
        assert all([issubclass(c, camelot.TestCase) for c in cases_list])

    def test_get_suites_from_module(self):
        suites_list = camelot.get_suites_from_module(suites)
        assert len(suites_list) > 0
        assert all([issubclass(s, camelot.TestSuite) for s in suites_list])

    def test_get_suites_from_module_with_abstract_class(self):
        suites_list = camelot.get_suites_from_module(suites.abstract_child_suite)
        assert len(suites_list) == 1

    def test_print_steps(self):
        camelot.print_steps(cases.PassingCase, print_source=True)

    def test_case_without_test_method(self):
        with pytest.raises(TypeError):
            inspection.CaseInspector(NoTestMethodCase)

    def test_inherited_case_get_steps(self):
        """Regression test for VL-686.

        Tests that steps can be found for a test case that inherits _test() from a parent.
        """
        inspector = inspection.CaseInspector(cases.InheritedCase)
        inspector2 = inspection.CaseInspector(cases.PassingCase)
        # Steps should be same as parent test case
        assert repr(inspector.get_steps()) == repr(inspector2.get_steps())

    def test_inherited_case_get_steps_source(self):
        """Source code lines should be the same."""
        inspector1 = inspection.CaseInspector(cases.InheritedCase)
        inspector2 = inspection.CaseInspector(cases.PassingCase)
        steps1 = inspector1.get_steps()
        steps2 = inspector2.get_steps()
        # Source code should be the same
        assert len(steps1[0].source) > 0
        assert steps1[0].source == steps2[0].source

    def test_inherited_case_with_mixin(self):
        """Tests inspection with a case that uses multiple inheritance."""
        inspector1 = inspection.CaseInspector(cases.InheritedCaseWithMixin)
        inspector2 = inspection.CaseInspector(cases.PassingCase)
        assert repr(inspector1.get_steps()) == repr(inspector2.get_steps())

    def test_multiple_inherited_case_get_steps(self):
        """Tests inspection with a case that uses multiple inheritance with _log_steps calls.

        In this case, we only logstep to the first _test() call we find. PassingCaseWithAssertions
        will be ignored.
        """
        inspector1 = inspection.CaseInspector(LogStepMultipleInheritance)
        inspector2 = inspection.CaseInspector(cases.PassingCase)
        steps = []
        steps.extend(inspector2.get_steps())
        assert repr(steps) == repr(inspector1.get_steps())
