"""Test valence.utils.qtest."""
from __future__ import unicode_literals

import configparser
import unittest

import mock
import pytest

from camelot import qtest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


GOOD_CFG = configparser.ConfigParser()
GOOD_CFG.add_section('QTEST')
GOOD_CFG.set('QTEST', 'username', 'username')
GOOD_CFG.set('QTEST', 'password', 'password')

BAD_CFG = configparser.ConfigParser()
BAD_CFG.add_section('QTEST')
BAD_CFG.set('QTEST', 'username', 'username')
BAD_CFG.set('QTEST', 'pissword', 'password')


class TestQTestConfig(unittest.TestCase):
    """Test QTestConfig object."""
    @mock.patch('os.path.isfile')
    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_missing_cfg_file(self, mock_stdout, mock_is_file):
        """Default location assumed, but cfg file missing."""
        mock_is_file.return_value = False

        with pytest.raises(IOError):
            cfg = qtest.QTestConfig()
        assert 'ATTENTION' in mock_stdout.getvalue()

    @mock.patch('os.path.isfile')
    @mock.patch.object(configparser.ConfigParser, 'read')
    @mock.patch.object(qtest.QTestConfig, '_get_parser')
    def test_valid_default_config(self, mock_parser, mock_read_cfg, mock_is_file):
        mock_is_file.return_value = True
        cfg = qtest.QTestConfig()

        mock_parser.return_value = GOOD_CFG

        username, password = cfg.get_qtest_credentials()
        assert username == 'username'
        assert password == 'password'

    @mock.patch('os.path.isfile')
    @mock.patch.object(configparser.ConfigParser, 'read')
    @mock.patch.object(qtest.QTestConfig, '_get_parser')
    @mock.patch.object(qtest.QTestConfig, 'print_help')
    def test_bad_default_config(self, mock_print_help, mock_parser, mock_read_cfg, mock_is_file):
        mock_is_file.return_value = True
        cfg = qtest.QTestConfig()

        mock_parser.return_value = BAD_CFG

        with pytest.raises(ValueError):
            cfg.get_qtest_credentials()
        mock_print_help.assert_called_once()

    @mock.patch('os.path.isfile')
    @mock.patch.object(configparser.ConfigParser, 'read')
    @mock.patch.object(qtest.QTestConfig, '_get_parser')
    def test_valid_specified_config(self, mock_parser, mock_read_cfg, mock_is_file):
        mock_is_file.return_value = True
        cfg_path = '/path/to/something/tasty'
        cfg = qtest.QTestConfig(cfg_path)

        mock_parser.return_value = GOOD_CFG

        username, password = cfg.get_qtest_credentials()
        assert username == 'username'
        assert password == 'password'
        mock_read_cfg.assert_called_with(cfg_path)

    @mock.patch('os.path.isfile')
    def test_get_parser(self, mock_is_file):
        mock_is_file.return_value = True
        cfg = qtest.QTestConfig()

        assert isinstance(cfg._get_parser(), configparser.ConfigParser)
