import re
import sys
import time
import unittest

import pytest

from camelot import (
    _history,
    _record,
)

try:
    1/0
except ZeroDivisionError:
    valid_exec_info = sys.exc_info()

dummy_history = _history.History(stack=[])


class TestExecInfo(unittest.TestCase):

    def test_invalid_tuple(self):
        """Pass in a tuple of invalid length."""
        with pytest.raises(TypeError):
            _record.ExcInfo((None, None))

    def test_none_values(self):
        exc_info = list(valid_exec_info)
        exc_info[0] = None
        exc_info = tuple(exc_info)
        with pytest.raises(TypeError):
            _record.ExcInfo(exc_info)

    def test_not_traceback_type(self):
        exc_info = list(valid_exec_info)
        exc_info[2] = 'not a tb type'
        exc_info = tuple(exc_info)
        with pytest.raises(TypeError):
            _record.ExcInfo(exc_info)

    def test_set_history(self):
        """Try to set history when it's already set"""
        exc_info = _record.ExcInfo(valid_exec_info)
        exc_info.set_history(dummy_history)

    def test_set_history_error(self):
        """Try to set history when it's already set"""
        exc_info = _record.ExcInfo(valid_exec_info)
        exc_info.value.history = dummy_history
        with pytest.raises(ValueError):
            exc_info.set_history(dummy_history)

    def test_equality(self):
        exec_info_obj = _record.ExcInfo(valid_exec_info)
        exec_info_obj2 = _record.ExcInfo(valid_exec_info)
        assert exec_info_obj == exec_info_obj2
        exec_info_obj2.etype = ValueError
        assert exec_info_obj != exec_info_obj2
        assert not exec_info_obj == exec_info_obj2


class TestRecord(unittest.TestCase):
    def test_create(self):
        r = _record.Record('test')
        assert r.msg == 'test'
        assert r.ts < time.time()
        assert r.exc_info is None
        ts = 1000.
        r = _record.Record('test2', ts=ts)
        assert r.msg == 'test2'
        assert r.ts == ts
        assert r.exc_info is None
        expected_pat = (
            r'> UTC\[1970-01-01 00:16:40\] LOC\[\d{4}-\d{1,2}-\d{1,2} \d{2}:16:40\] \[.*\]: test2')
        assert re.match(expected_pat, r.pretty_fmt())

    def test_ts_invalid(self):
        with pytest.raises(TypeError):
            _record.Record('msg', ts='not a float')

    def test_exc_info_object(self):
        r = _record.Record('msg', exc_info=_record.ExcInfo(valid_exec_info))
        assert isinstance(r.exc_info, _record.ExcInfo)

    def test_exc_info_tuple(self):
        r = _record.Record('msg', exc_info=valid_exec_info)
        assert isinstance(r.exc_info, _record.ExcInfo)

    def test_exception(self):
        ts = 1000.
        try:
            1 / 0
        except ZeroDivisionError:
            exc_info = sys.exc_info()
            exc1 = _record.ExcInfo()
            r = _record.Record('test_exc', exc_info=True, ts=ts)
        exc2 = _record.ExcInfo(exc_info)
        assert exc2 == exc1
        assert exc1 == r.exc_info
        result = r.pretty_fmt()
        expected_pat = r'''\
> UTC\[1970-01-01 00:16:40\] LOC\[1969-12-31 17:16:40\] \[.*\]: test_exc
|!!! Exception Occurred:
|Traceback \(most recent call last\):
|  File ".*", line 28, in test_exception
|    1 / 0
|ZeroDivisionError: integer division or modulo by zero
|!!!'''
        assert re.search(expected_pat, result)
