"""Unit tests for base test suite definition."""
import os

import mock
import pytest
import sf_artifacts
import sf_batch

import camelot
from camelot import (
    _enums,
    _errors,
)
from valence import (
    models,
    unittest_utils,
)
from valence.unittest_utils import (
    helpers,
    suites,
)

DATA_DIR = unittest_utils.get_data_dir(__file__)


class TestSuiteKeyboardInterrupt(helpers.MockHistoryUnitTest):
    """Test that suite handles a KeyboardInterrupts correctly."""

    def setUp(self):
        super(TestSuiteKeyboardInterrupt, self).setUp()
        self.config = models.SuiteConfig()
        self.resources = models.SuiteResources()

    @mock.patch.object(suites.PassingSuite, 'teardown')
    def test_interrupt_in_setup(self, mock_teardown):
        suite = suites.KeyboardInterruptInSetupSuite(self.resources, self.config)
        with pytest.raises(KeyboardInterrupt):
            suite.execute()
        assert suite.status is camelot.ResultStatus.cancelled
        assert suite.result.end_time is not None
        assert not mock_teardown.called

    def test_interrupt_in_teardown(self):
        suite = suites.KeyboardInterruptInTeardownSuite(self.resources, self.config)
        with pytest.raises(KeyboardInterrupt):
            suite.execute()
        assert suite.status is camelot.ResultStatus.cancelled
        assert suite.test_runs[0].result.status is camelot.ResultStatus.passed
        assert suite.result.end_time is not None

    @mock.patch.object(suites.PassingSuite, 'teardown')
    def test_interrupt_in_case_setup(self, mock_teardown):
        suite = suites.KeyboardInterruptInCaseSetupSuite(self.resources, self.config)
        with pytest.raises(KeyboardInterrupt):
            suite.execute()
        assert suite.status is camelot.ResultStatus.cancelled
        assert suite.test_runs[0].result.status is camelot.ResultStatus.passed
        assert suite.test_runs[1].result.status is camelot.ResultStatus.cancelled
        assert suite.result.end_time is not None
        assert not mock_teardown.called

    @mock.patch.object(suites.PassingSuite, 'teardown')
    def test_interrupt_in_case_test(self, mock_teardown):
        suite = suites.KeyboardInterruptInCaseTestSuite(self.resources, self.config)
        with pytest.raises(KeyboardInterrupt):
            suite.execute()
        assert suite.status is camelot.ResultStatus.cancelled
        assert suite.test_runs[0].result.status is camelot.ResultStatus.cancelled
        assert suite.test_runs[1].result.status is camelot.ResultStatus.pending
        assert suite.result.end_time is not None
        assert not mock_teardown.called

    def test_interrupt_in_case_teardown(self):
        suite = suites.KeyboardInterruptInCaseTeardownSuite(self.resources, self.config)
        with pytest.raises(KeyboardInterrupt):
            suite.execute()
        assert suite.status is camelot.ResultStatus.cancelled
        assert suite.test_runs[0].result.status is camelot.ResultStatus.cancelled
        assert suite.result.end_time is not None


class TestBaseSuite(helpers.MockHistoryUnitTest):
    """Test default test suite parameter values."""

    @classmethod
    def setUpClass(cls):
        cls.config = models.SuiteConfig()
        cls.resources = models.SuiteResources()
        cls.resources_file = os.path.join(DATA_DIR, 'resources.json')
        cls.config_file = os.path.join(DATA_DIR, 'suite_config.json')
        cls.run_indices_config = models.SuiteConfig({'run_indices': [0]})
        cls.get_mock_manager()

    @staticmethod
    def get_mock_manager():
        sf_artifacts.get_manager = mock.MagicMock(return_value=mock.MagicMock(
                spec=sf_artifacts.ArtifactManager))

    def test_failing_suite(self):
        failing_suite = suites.FailingSuite(self.resources, self.config)
        failing_suite.execute()
        assert failing_suite.status is camelot.ResultStatus.failed

    def test_suite_without_tests(self):
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite.tests = []
        with pytest.raises(ValueError):
            passing_suite.execute()
        assert passing_suite.status is camelot.ResultStatus.error

    def test_end_time_is_set_for_passing_suite(self):
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite.execute()
        assert passing_suite.status is camelot.ResultStatus.passed
        assert passing_suite.test_runs[0].result.end_time is not None

    def test_index_is_set_for_passing_suite(self):
        passing_two_suite = suites.PassingTwoTestSuite(self.resources, self.config)
        assert passing_two_suite.execute()
        assert passing_two_suite.test_runs[0].result.index == 0
        assert passing_two_suite.test_runs[1].result.index == 1

    def test_exception_in_preconditions(self):
        suite = suites.FailInPreconditions(self.resources, self.config)
        with pytest.raises(AssertionError):
            suite.execute()

    def test_end_time_is_set_for_error_suite(self):
        failing_suite_tc_error = suites.FailingSuiteTCError(self.resources, self.config)
        failing_suite_tc_error.execute()
        assert failing_suite_tc_error.result.end_time is not None
        assert failing_suite_tc_error.test_runs[0].result.end_time is not None

    def test_things_set_for_assertion_failure(self):
        failing_suite_tc_double = suites.FailingSuitePostTestAssertionsAndTest(self.resources,
                                                                               self.config)
        failing_suite_tc_double.execute()
        assert failing_suite_tc_double.result.end_time is not None
        assert failing_suite_tc_double.test_runs[0].result.end_time is not None

    @mock.patch.object(suites.FailingSuite, 'teardown')
    def test_stop_on_failure_is_true_failing_test_case(self, mock_teardown):
        """Exception should be raised."""
        config = models.SuiteConfig.create(stop_on_failure=True)
        failing_suite = suites.FailingSuite(self.resources, self.config)
        failing_suite.config = config
        with pytest.raises(AssertionError):
            failing_suite.execute()
        assert failing_suite.status is camelot.ResultStatus.error
        assert not mock_teardown.called

    @mock.patch.object(suites.FailingSuitePostTestAssertions, 'teardown')
    def test_stop_on_failure_is_true_failing_test_case_assertions(self, mock_teardown):
        """Exception should be raised."""
        config = models.SuiteConfig.create(stop_on_failure=True)
        failing_suite_tc_assertions = suites.FailingSuitePostTestAssertions(self.resources, self.config)
        failing_suite_tc_assertions.config = config
        with pytest.raises(Exception):
            failing_suite_tc_assertions.execute()
        assert failing_suite_tc_assertions.status is camelot.ResultStatus.error
        assert not mock_teardown.called

    @mock.patch.object(suites.FailingSuitePostTestAssertionsAndTest, 'teardown')
    def test_stop_on_failure_is_true_failing_test_case_assertions_and_test(self, mock_teardown):
        """Exception should be raised."""
        config = models.SuiteConfig.create(stop_on_failure=True)
        failing_suite_tc_double = suites.FailingSuitePostTestAssertionsAndTest(self.resources,
                                                                               self.config)
        failing_suite_tc_double.config = config
        failing_suite_tc_double.tests[0]._post_test_assertions = mock.MagicMock()
        with pytest.raises(Exception):
            failing_suite_tc_double.execute()
        assert failing_suite_tc_double.status is camelot.ResultStatus.error
        # post test assertions won't be called since we're stop on failure
        assert not failing_suite_tc_double.tests[0]._post_test_assertions.called
        assert not mock_teardown.called

    @mock.patch.object(suites.FailingSuiteTCError, 'teardown')
    def test_stop_on_failure_is_true_error_test_case(self, mock_teardown):
        """Exception should be raised."""
        config = models.SuiteConfig.create(stop_on_failure=True)
        failing_suite_tc_error = suites.FailingSuiteTCError(self.resources, self.config)
        failing_suite_tc_error.config = config
        with pytest.raises(Exception):
            failing_suite_tc_error.execute()
        assert failing_suite_tc_error.status is camelot.ResultStatus.error
        assert not mock_teardown.called

    @mock.patch.object(suites.FailingSuite, 'teardown')
    def test_stop_on_failure_is_false(self, mock_teardown):
        """Exception should not be raised, but suite should fail."""
        config = models.SuiteConfig.create(stop_on_failure=False)
        failing_suite = suites.FailingSuite(self.resources, self.config)
        failing_suite.config = config
        failing_suite.execute()
        assert failing_suite.status is camelot.ResultStatus.failed
        assert mock_teardown.called

    def test_stop_on_failure_is_false_failing_assertion(self):
        """Exception should not be raised, but suite should fail."""
        config = models.SuiteConfig.create(stop_on_failure=False)
        failing_suite_tc_assertions = suites.FailingSuitePostTestAssertions(self.resources, self.config)
        failing_suite_tc_assertions.config = config
        failing_suite_tc_assertions.execute()
        assert failing_suite_tc_assertions.status is camelot.ResultStatus.failed

    def test_stop_on_failure_is_false_failing_assertion_and_test(self):
        """Exception should not be raised, but suite should fail."""
        config = models.SuiteConfig.create(stop_on_failure=False)
        failing_suite_tc_double = suites.FailingSuitePostTestAssertionsAndTest(self.resources,
                                                                               self.config)
        failing_suite_tc_double.tests[0]._post_test_assertions = mock.MagicMock(
            side_effect=AssertionError('Mocking to make sure it gets called'))
        failing_suite_tc_double.config = config
        failing_suite_tc_double.execute()
        assert failing_suite_tc_double.status is camelot.ResultStatus.failed
        # post test assertions WILL be called since we're stop on failure
        assert failing_suite_tc_double.tests[0]._post_test_assertions.called

    @mock.patch.object(camelot.TestSuite, 'run_tests')
    def test_exception_in_run_tests(self, mock_run_tests):
        passing_suite = suites.PassingSuite(self.resources, self.config)
        mock_run_tests.side_effect = RuntimeError
        passing_suite.execute()
        assert passing_suite.status is camelot.ResultStatus.error

    @mock.patch.object(suites.PassingSuite, 'teardown')
    def test_teardown_disabled(self, mock_teardown):
        config = models.SuiteConfig.create(run_suite_teardown=False)
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite.config = config
        passing_suite.execute()
        assert mock_teardown.called is False
        assert passing_suite.status is camelot.ResultStatus.passed

    @mock.patch.object(suites.PassingSuite, 'teardown')
    def test_exception_in_teardown(self, mock_teardown):
        passing_suite = suites.PassingSuite(self.resources, self.config)
        mock_teardown.side_effect = RuntimeError
        trs = passing_suite.execute()
        assert passing_suite.status is camelot.ResultStatus.error
        assert trs is passing_suite.test_runs

    @mock.patch.object(suites.PassingSuite, 'teardown')
    def test_teardown_is_false_and_stop_on_failure_is_true(self, mock_teardown):
        """Suite-teardown should not be executed."""
        config = models.SuiteConfig.create(run_suite_teardown=False, stop_on_failure=True)
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite.config = config
        passing_suite.execute()
        assert mock_teardown.called is False

    @mock.patch.object(suites.FailingSuite, 'teardown')
    def test_teardown_is_true_and_stop_on_failure_is_true(self, mock_teardown):
        """Suite-teardown should be executed."""
        config = models.SuiteConfig.create(run_suite_teardown=True, stop_on_failure=True)
        failing_suite = suites.FailingSuite(self.resources, self.config)
        failing_suite.config = config
        with pytest.raises(AssertionError):
            failing_suite.execute()
        assert failing_suite.status is camelot.ResultStatus.error

    @mock.patch.object(suites.FailingSuite, 'teardown')
    def test_teardown_is_false_and_stop_on_failure_is_false(self, mock_teardown):
        """Suite-teardown should be executed."""
        config = models.SuiteConfig.create(run_suite_teardown=False, stop_on_failure=False)
        passing_suite = suites.FailingSuite(self.resources, self.config)
        passing_suite.config = config
        passing_suite.execute()
        assert mock_teardown.called is False

    def test_skipping_test(self):
        skip_config = models.SuiteConfig({'stop_index': 1})
        skipping_suite = suites.PassingTwoTestSuite(self.resources, skip_config)
        skipping_suite.execute()
        assert skipping_suite.test_runs[1][1].status is camelot.ResultStatus.skipped

    def test_skipping_no_tests(self):
        two_test_suite = suites.PassingTwoTestSuite(self.resources, self.config)
        two_test_suite.execute()
        assert all([r.status is camelot.ResultStatus.passed for t, r in two_test_suite.test_runs])

    def test_skip_all_tests(self):
        skip_all_config = models.SuiteConfig({'stop_index': 0})
        skip_all_suite = suites.PassingTwoTestSuite(self.resources, skip_all_config)
        with pytest.raises(Exception):
            skip_all_suite.execute()

    def test_run_indices_valid(self):
        run_indices_suite = suites.PassingTwoTestSuite(self.resources, self.run_indices_config)
        run_indices_suite.execute()
        assert len(run_indices_suite.test_runs) == 1
        assert len(run_indices_suite.tests) == 2

    def test_run_indices_invalid(self):
        inv_run_indices_config = models.SuiteConfig({'run_indices': [10]})
        inv_run_indices_suite = suites.PassingTwoTestSuite(self.resources,
                                                           inv_run_indices_config)
        with pytest.raises(Exception):
            inv_run_indices_suite.execute()

    def test_execute_twice(self):
        run_indices_suite = suites.PassingTwoTestSuite(self.resources, self.run_indices_config)
        run_indices_suite.execute()
        assert len(run_indices_suite.test_runs) == 1
        assert len(run_indices_suite.tests) == 2
        run_indices_suite.execute()
        assert len(run_indices_suite.test_runs) == 1
        assert len(run_indices_suite.tests) == 2

    @mock.patch.object(camelot.TestSuite, '_write_to_summary_file')
    @mock.patch('sf_artifacts.LocalArtifactManager.store_text')
    @mock.patch('camelot._suite.open', create=True)
    @mock.patch('logbook.FileHandler')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('os.makedirs')
    def test_initialize_artifact_credentials_local_provided_and_base_dir(self, mock_mkdirs, mock_temp,
                                                                         mock_file_handler, mock_open,
                                                                         mock_write_to_summary,
                                                                         mock_store):
        artifacts_env_var = camelot._suite.ARTIFACTS_DIR_ENV_VAR
        base_dir = '/tmp/abcd'
        mock_os_environ = mock.patch.dict(os.environ,
                                          {artifacts_env_var: '/tmp/abcd'})
        mock_os_environ.start()
        config_dm = models.SuiteConfig(dict(artifact_credentials={'local': True,
                                                                  'base_dir': base_dir}))
        suite = suites.PassingSuiteArtifacts(self.resources, config_dm)
        assert mock_mkdirs.called
        assert base_dir in mock_mkdirs.call_args[0][0]
        assert mock_file_handler.called
        suite.execute()
        assert mock_open.called
        assert mock_write_to_summary.called == 1
        assert mock_store.call_count == 2
        mock_os_environ.stop()
        if artifacts_env_var in os.environ:
            del os.environ[artifacts_env_var]
        suite.execute()

    @mock.patch('sf_artifacts.LocalArtifactManager.store_text')
    @mock.patch('logbook.FileHandler')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('os.makedirs')
    def test_initialize_artifact_credentials_no_credentials(self, mock_mkdirs,
                                                            mock_temp,
                                                            mock_file_handler,
                                                            mock_store_text):
        config = models.SuiteConfig()
        suite = suites.PassingSuiteArtifacts(self.resources, config)
        assert mock_mkdirs.called
        mock_mkdirs.assert_called_with(suite.local_artifacts_dir)
        assert mock_file_handler.called
        assert mock_store_text.call_count == 2

    @mock.patch('sf_artifacts.At2ArtifactManager.store_text')
    @mock.patch('at2_sdk.at2.BaseAT2Api.call')
    @mock.patch('at2_sdk.at2.BaseAT2Api.authenticate')
    @mock.patch('logbook.FileHandler')
    @mock.patch('os.makedirs')
    def test_with_at2_credentials(self, mock_mkdirs, mock_file_handler, mock_auth, mock_call,
                                  mock_store):
        config = models.SuiteConfig(dict(artifact_credentials=dict(username='abcded',
                                                                   password='garbage',
                                                                   task_instance_step_id=10)))
        suites.PassingSuiteArtifacts(self.resources, config)
        # Local files should not be written when run in At2
        assert not mock_mkdirs.called
        assert not mock_file_handler.called
        assert mock_store.call_count == 2

    @mock.patch.object(suites.ExceptionSuite, 'teardown')
    def test_suite_setup_exception(self, mock_teardown):
        exception_suite = suites.ExceptionSuite(self.resources, self.config)
        exception_suite.execute()
        assert exception_suite.status is camelot.ResultStatus.error
        assert mock_teardown.called
        assert all(r.result.status is camelot.ResultStatus.pending for r in
                   exception_suite.test_runs)

    @mock.patch('sf_artifacts.LocalArtifactManager.store_text')
    @mock.patch('camelot._suite.open', create=True)
    @mock.patch('logbook.FileHandler')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('os.makedirs')
    def test_write_to_summary_file(self, mock_mkdirs, mock_temp, mock_file_handler, mock_open,
                                   mock_store):
        """Test for coverage on _write_to_summary_file.

        Mock makedirs and Filehandler so file system is not modified when the UT runs.
        """
        config = models.SuiteConfig()
        suite = suites.PassingSuiteArtifacts(self.resources, config)
        suite.execute()
        assert mock_open.called

    @mock.patch('sf_artifacts.LocalArtifactManager.store_text')
    @mock.patch.object(camelot.TestSuite, '_write_to_summary_file')
    @mock.patch('camelot._suite.open', create=True)
    @mock.patch('logbook.FileHandler')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('os.makedirs')
    def test_record_to_summary_file_skipped(self, mock_mkdirs, mock_temp, mock_file_handler, mock_open,
                                            mock_write_to_summary, mock_store):
        """Test that skipped test cases still record to the summary file.

        Mock makedirs and Filehandler so file system is not modified when the UT runs.
        """
        skip_config = models.SuiteConfig({'stop_index': 1})
        suite = suites.MultiErrorFailingSuiteArtifacts(self.resources, skip_config)
        suite.execute()
        assert mock_write_to_summary.call_count == 3

        # Check that the expected TestCase result enum's value is inside the summary string
        # Parsing into mock's call_args_list; a list of calls, a list of data, and the str is in a tuple
        assert _enums.ResultStatus.passed.value.upper() in mock_write_to_summary.call_args_list[0][0][0]
        assert _enums.ResultStatus.skipped.value.upper() in mock_write_to_summary.call_args_list[1][0][0]

    @mock.patch('sf_artifacts.LocalArtifactManager.store_text')
    @mock.patch.object(camelot.TestSuite, '_write_to_summary_file')
    @mock.patch('camelot._suite.open', create=True)
    @mock.patch('logbook.FileHandler')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('os.makedirs')
    def test_record_to_summary_file_stop_on_failed(self, mock_mkdirs, mock_temp, mock_file_handler,
                                                   mock_open,
                                                   mock_write_to_summary, mock_store):
        """Test that failed test cases still record to the summary file.

        Mock makedirs and Filehandler so file system is not modified when the UT runs.
        """
        config = models.SuiteConfig.create(stop_on_failure=True)
        suite = suites.MultiErrorFailingSuiteArtifacts(self.resources, config)
        with pytest.raises(RuntimeError):
            suite.execute()
        assert mock_write_to_summary.call_count == 2

        # Check that the expected TestCase result enum's value is inside the summary string
        # Parsing into mock's call_args_list; a list of calls, a list of data, and the str is in a tuple
        assert _enums.ResultStatus.passed.value.upper() in mock_write_to_summary.call_args_list[0][0][0]
        assert _enums.ResultStatus.error.value.upper() in mock_write_to_summary.call_args_list[1][0][0]

    @mock.patch('threading.enumerate')
    @mock.patch('sf_batch.SFThread.join')
    def test_log_threads_join_fail(self, mock_join, mock_enumerate):
        mock_enumerate.return_value = [sf_batch.SFThread(lambda x: None)]
        mock_join.side_effect = ValueError
        with pytest.raises(ValueError):
            camelot.TestSuite._log_threads(check_for_errors=True)
        mock_join.assert_called()

    @mock.patch('threading.enumerate')
    @mock.patch('sf_batch.SFThread.join')
    @mock.patch('sf_batch.check_for_error_and_raise_exception')
    def test_log_threads_join_no_check(self, mock_check, mock_join, mock_enumerate):
        mock_enumerate.return_value = [sf_batch.SFThread(lambda x: None)]
        camelot.TestSuite._log_threads(check_for_errors=False)
        mock_check.assert_not_called()
        mock_join.assert_not_called()

    @mock.patch('threading.enumerate')
    @mock.patch('sf_batch.SFThread.join')
    @mock.patch('sf_batch.check_for_error_and_raise_exception')
    @mock.patch('sf_batch.SFThread.is_alive')
    def test_log_threads_join(self, mock_alive, mock_check, mock_join, mock_enumerate):
        mock_alive.return_value = False
        mock_enumerate.return_value = [sf_batch.SFThread(lambda x: None)]
        camelot.TestSuite._log_threads(check_for_errors=True)
        mock_check.assert_called()
        mock_join.assert_called()
        mock_alive.assert_called()

    @mock.patch('threading.enumerate')
    @mock.patch('sf_batch.SFThread.join')
    @mock.patch('sf_batch.check_for_error_and_raise_exception')
    @mock.patch('sf_batch.SFThread.is_alive')
    def test_log_threads_join_still_alive(self, mock_alive, mock_check, mock_join, mock_enumerate):
        mock_alive.return_value = True
        mock_enumerate.return_value = [sf_batch.SFThread(lambda x: None)]

        with pytest.raises(_errors.TestRuntimeError):
            camelot.TestSuite._log_threads(check_for_errors=True)

        mock_join.assert_called()
        mock_check.assert_not_called()
