"""Utility functions to make it easier to interact with Valence in a Python Shell.

This code should only be used during development, and never in a production environment.
"""
import subprocess

import autologger
import requests

from camelot import driver
from valence import (
    functional_models,
    models,
)

COLOR_DEFAULT = '\033[0m'
COLOR_ERROR = '\033[91m'


def print_red(message):
    """Print a provided message in red.

    :param str message: string message to print
    """
    print(COLOR_ERROR + message + COLOR_DEFAULT)


def get_suite_resources(resources_file=None):
    """Returns a SuiteResources data model.

    If resources_file is not given, look for the file in <project_root>/vl_data/resources.json.

    .. note:: This just wraps camelot.driver.get_resources_dm()

    :param str resources_file: Path to resources.json file. If None, look for resources.json in data
         directory. If that file does not exist, return an empty SuiteResources data model.
    :rtype: valence.models.SuiteResources
    """
    # note - this will always use a default SuiteConfig
    return driver.get_resources_dm(resources_file)


def docker_sfnode_running():
    """Determine if docker-sfnode is running locally.

    :rtype: bool
    :return: True, if docker-sfnode is running
    """
    output = subprocess.check_output(['ps', 'auxww'])
    return 'docker-sfnode' in output.decode('utf-8', 'ignore')


class ShellInitializer(object):
    """Class that initializes commonly used objects.

    :param str resources_file: Path to resources.json file. If None, look under the vl_data directory
         in the project directory (requires use of virtualenvwrapper).
    :param bool attempt_docker_sf_node: If True, use a docker-sfnode instance if available
    """

    def __init__(self, resources_file=None, attempt_docker_sf_node=True):
        self.suite_resources = get_suite_resources(resources_file)
        self.cluster_fms = []
        self.node_fms = []
        self.client_fms = []
        if attempt_docker_sf_node and docker_sfnode_running():
            self.init_with_docker_sf_node()
        elif self.suite_resources:
            self.init_with_cluster()

    def init_with_cluster(self):
        """Initialize with actual cluster."""
        # Initialize clusters
        for cluster_dm in self.suite_resources.cluster_dms:
            self.cluster_fms.append(functional_models.Cluster(cluster_dm, authenticate=False))
            try:
                self.cluster_fms[-1].authenticate()
            except requests.RequestException:
                print_red('WARNING: Unable to authenticate cluster: {}'.format(cluster_dm))

        # Initialize nodes
        models.set_node_resources(self.suite_resources.node_dms)
        models.set_cluster_resources(self.suite_resources.cluster_dms)
        autologger.set_global_loglevel(autologger.CRITICAL)
        for cluster_fm in self.cluster_fms:
            self.node_fms.extend(cluster_fm.get_node_fms(authenticate=False))
        autologger.set_global_loglevel(autologger.DEBUG)
        clustered_node_names = [fm.node_dm.node_name for fm in self.node_fms]
        for node_dm in self.suite_resources.node_dms:
            if node_dm.node_name not in clustered_node_names:
                self.node_fms.append(functional_models.Node(
                    node_dm=node_dm,
                    cluster_dm=None,
                    authenticate=False
                ))

        # Initialize clients
        if self.suite_resources.client_dms:
            self.client_fms.extend(
                functional_models.get_client_from_dm(dm) for dm in self.suite_resources.client_dms
            )

    def init_with_docker_sf_node(self):
        """Initialize with docker cluster."""
        print('Using docker-sfnode cluster')
        sf_node_ip = '0.0.0.0'
        sf_node_port = '10000'
        resource_cluster = models.ResourceCluster.create(
            mvip='{}:{}'.format(sf_node_ip, sf_node_port),
            username='admin',
            password='admin',
            api_endpoint='10.0'
        )  # type: models.ResourceCluster

        self.cluster_fms = [functional_models.Cluster(resource_cluster)]
        node_dms = self.cluster_fms[0].list_active_nodes()
        models.set_node_resources(node_dms)
        models.set_cluster_resources([resource_cluster])
        self.node_fms = [functional_models.Node(n, cluster_dm=resource_cluster) for n in node_dms]
