"""Base class for interacting with SolidFire API."""

import abc
import itertools
import json
import logging
import threading
import time

import aenum
import requests
import six
from requests import auth as requests_auth
from requests.packages import urllib3

from sf_api import (
    _retry_sf_exception_constants,
    api_constants,
    api_errors,
    enums,
    sf_version,
)
from sf_api.models import types

# shut up urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

_log = logging.getLogger(__name__)
_JSON_RPC_ID = itertools.count()
_LOCK = threading.Lock()


def get_json_rpc_id():
    """Method to lock the thread & update global _JSON_RPC_ID for requests.

    next() is atomic, so it should be thread safe, but the Google Python style guide advises against
    relying on atomicity of built-in types.

    :rtype: int
    """
    with _LOCK:
        return next(_JSON_RPC_ID)


class SfErrorResponse(object):
    """Object representing an error returned from the SF API.

    :param dict raw_data: The dictionary of the error object in a SolidFire response.
    """

    def __init__(self, raw_data):
        self.raw_data = raw_data
        self.code = raw_data.get(api_constants.CODE)
        self.name = raw_data.get(api_constants.NAME)
        self.message = raw_data.get(api_constants.MESSAGE)

    def __repr__(self):
        """Pretty format with error name."""
        return '{}<name={}>'.format(self.__class__.__name__, self.name)


class SfResponse(object):
    """Object representing a response from the SF API.

    :param dict raw_data: The full dictionary response from SolidFire.
    :param sf_api.ClusterMethod|sf_api.NodeMethod method: Name of the method that the response came from
    :param result_dm_class: The API Response class the result is modeled after.
    :type result_dm_class: (dict) -> sf_api.models.Base
    """

    def __init__(self, raw_data, method, result_dm_class=None):
        self.raw_data = raw_data
        self.method = method
        self._result_dm_class = result_dm_class
        self._result_dm = None
        self.id = raw_data.get(api_constants.ID)
        self.result = raw_data.get(api_constants.RESULT)
        error = raw_data.get(api_constants.ERROR)
        if error is not None:
            error = SfErrorResponse(error)
        self.error = error
        self.warning = raw_data.get(api_constants.WARNING)

    @property
    def result_dm(self):
        """Get the result in data model form.

        :rtype: sf_api.result_models.BaseApiResult
        """
        if self._result_dm is None and self._result_dm_class is not None:
            self._result_dm = self._result_dm_class(self.result)
        return self._result_dm

    def __repr__(self):
        return '{}<method={}, id={}>'.format(self.__class__.__name__, self.method.value, self.id)


class SfAuth(requests_auth.AuthBase):
    """Holds authentication information and URI for interacting with the SolidFire API.

    :param str ip: The ip of the resource. Can be IPv4 or IPv6.
    :param int port: The port of the resource (optional)
    :param str username: Username for node or cluster. Not required for orphaned nodes.
    :param str password: Password for node of cluster. Not required for orphaned nodes.
    :param sf_api.SfVersion version: Version of the product.
    """

    def __init__(self, ip, port=None, username=None, password=None, version=None):
        self.ip = ip
        self.port = port
        self.username = username
        self.password = password
        self._version = version

    @property
    def version(self):
        """Return the version for this object instance.

        If version is None, the cluster will be queried for the current version.

        :rtype: sf_api.SfVersion
        """
        if self._version is None:
            self._version = self.get_version()
        return self._version

    @version.setter
    def version(self, value):
        """Sets the version for this object instance.

        :param sf_api.SfVersion value: value of the API version number to use, e.g. '11.0'
        """
        self._version = value

    @property
    def root_uri(self):
        """Root URI for node or cluster.

        Only includes ip and (if it is a node) the port.

        :rtype: str
        """
        ip = '[{}]'.format(self.ip) if types.IPv6Type.valid_ip(self.ip) else self.ip
        uri = 'https://{}'.format(ip)
        if self.port is not None:
            uri = '{}:{}'.format(uri, self.port)
        return uri

    @property
    def uri(self):
        """URI for a node or cluster API.

        This uri defines the JSON RPC endpoint with the version.

        :rtype: str
        """
        return '{}/json-rpc/{}'.format(self.root_uri, self.version)

    def get_version(self):
        """Get current version via GetAPI.

        :rtype: sf_api.SfVersion
        """
        _log.debug('Querying cluster or node for supported API versions')
        method = enums.ClusterMethod.get_api.value
        json_request = {'method': method, 'id': get_json_rpc_id()}

        uri = '{}/json-rpc/{}'.format(self.root_uri, api_constants.DEFAULT_API_VERSION)
        requests_response = requests.post(uri, json=json_request, timeout=(30.5, 60), auth=self,
                                          verify=False)
        requests_response.raise_for_status()
        response_json = requests_response.json()

        return sf_version.SfVersion(response_json['result'].get(api_constants.CURRENT_VERSION))

    def __call__(self, request):
        # Use basic http auth
        http_auth = requests_auth.HTTPBasicAuth(username=self.username, password=self.password)
        http_auth(request)
        return request


@six.add_metaclass(abc.ABCMeta)
class BaseJsonRpcClient(object):
    """Abstract base class for interfacing with a SolidFire JSON RPC API.

    :param str ip: The ip of the resource
    :param str username: Username for node or cluster. Not required for orphaned nodes.
    :param str password: Password for node of cluster. Not required for orphaned nodes.
    :param sf_api.SfVersion version: Version of the product. If not provided, query the cluster
        with GetApi.
    :param bool authenticate: If true, authenticate resource in constructor (default=True)
    :param typing.Iterable[Any] retry_sf_exceptions_override: SolidFire exceptions that all
        methods will retry on. If provided, this will override the retry faults for all methods.
    :param int retry_count: Number of times to retry before raising an exception
    :param int port: Optional port number
    """

    def __init__(self, ip, username=None, password=None, version=None, authenticate=True,
                 retry_sf_exceptions_override=None, retry_count=2, port=None):
        self.ip = ip
        self.retry_count = retry_count
        self.retry_sf_exceptions_override = retry_sf_exceptions_override
        self.username = username
        self.password = password
        self.port = port
        # Timeout for http calls. (connection timeout, read timeout)
        self.timeout = (30.5, 300)
        self.min_retry_delay = 1
        self._session = requests.session()
        self._session.verify = False
        self._auth = SfAuth(ip=ip, username=username, password=password, version=version, port=port)

        if authenticate:
            self.authenticate()

    # PUBLIC METHODS

    def get_auth(self):
        """Return a new auth object.

        Fields are populated based on the current state of the API object.

        :rtype: sf_api.SfAuth
        """
        return SfAuth(ip=self.ip, username=self.username, password=self.password,
                      version=self.version, port=self.port)

    def authenticate(self):
        """Authenticates a resource using HTTP basic authentication.

        :rtype: requests.Response
        :return: requests Response object
        """
        try:
            _log.debug('Authenticating to url: %s', self.uri)
            response = self._session.get(self.uri, auth=self._auth, timeout=self.timeout)
        except requests.exceptions.RequestException as re:
            _log.debug('Connection error: %s', re)
            raise

        if response.ok is False:
            _log.debug('HTTP error: %d', response.status_code)
            # Raises stored HTTPError
            response.raise_for_status()
        else:
            return response

    def http_get(self, path, retry_delay=1):
        """Perform an http get call.

        :param str path: Path relative to root_uri, e.g. "reports" will resolve to
            "https://<ip>/reports"
        :param int retry_delay: Amount of time (in seconds) to delay between retry attempts
        :raises requests.HTTPError: If an error status code is returned
        :rtype: requests.Response
        """
        num_retries = 0
        request_response = None
        while num_retries <= self.retry_count:
            try:
                request_response = self._session.get('{}/{}'.format(self.root_uri, path),
                                                     timeout=self.timeout, auth=self._auth)
                request_response.raise_for_status()
                break
            except requests.RequestException:
                if num_retries == self.retry_count:
                    raise
                num_retries += 1
                time.sleep(retry_delay)
        return request_response

    @staticmethod
    def get_method_retries(method, include_global=True):
        """Get exceptions to retry on for the given method, including global retries.

        :param sf_api.ClusterMethod|sf_api.NodeMethod method: method enum you want the model for.
        :param bool include_global: If True, include retries that are global to all methods.
        :return: A tuple of sf_api.SfApiException classes
        :rtype: tuple[six.class_types]
        """
        faults = _retry_sf_exception_constants.METHOD_RETRY_SF_EXCEPTIONS.get(method, ())
        if include_global:
            faults += _retry_sf_exception_constants.SF_XDB_RETRIES
        return faults

    @staticmethod
    def _log_retries(method, num_retries, exceptions):
        """Log retry info, if necessary.

        :param sf_api.ClusterMethod|sf_api.NodeMethod method: method that was called
        :param int num_retries: number of times call was retried
        :param list[sf_api.SfApiException] exceptions: exceptions that were encountered
        """
        if num_retries:
            if num_retries == 1:
                log_string = 'Retried %s call %d time with exceptions: %s'
            else:
                log_string = 'Retried %s call %d times with exceptions: %s'
            _log.debug(log_string, method.value, num_retries, exceptions)

    def call_method(self, method, params=None, result_dm_class=None, log_response=True,
                    auth=None):
        """Call the API, converts native params to primitives.

        :param sf_api.ClusterMethod|sf_api.NodeMethod method: method to call
        :param dict params: the params for the method. Enums will be converted to strings
        :param Any result_dm_class: The API Response class the result is modeled after. If there is no
            result data model associated with the call, this can be None.
        :param bool log_response: If true, log the full JSON response from the API
        :param sf_api.SfAuth auth: Override the authentication object. This can be use to modify fields
            such as the password for a single api call like this::

                auth = api.get_auth()
                auth.password = 'new password'
                api.call_method(method, params, auth=auth)

        :rtype: sf_api.SfResponse
        """
        retry_faults = (
            requests.ConnectionError,
            requests.HTTPError,
            requests.Timeout
        )
        raised_exceptions = []
        num_retries = 0
        if self.retry_sf_exceptions_override is not None:
            retry_faults += tuple(self.retry_sf_exceptions_override)
        else:
            retry_faults += self.get_method_retries(method)
        while num_retries <= self.retry_count:
            try:
                response_dict = self.call_method_raw(
                    method=method.value,
                    params=self._construct_params(params) if params else None,
                    auth=auth
                )
                sf_response = self._process_response(response_dict, method, result_dm_class)
                self._log_retries(method, num_retries, raised_exceptions)
                if log_response:
                    _log.debug('%s response: %s', method.value, json.dumps(sf_response.raw_data))
                return sf_response

            except retry_faults as e:
                raised_exceptions.append(e)
                if num_retries == self.retry_count:
                    self._log_retries(method, num_retries, raised_exceptions)
                    raise
                num_retries += 1
                retry_delay = max(self.min_retry_delay, getattr(e, 'retry_delay', 1))
                time.sleep(retry_delay)

    def call_method_raw(self, method, params=None, auth=None):
        """Call the API with raw params dictionary.

        Send Json RPC request to uri, return the SfResponse object.

        If self.version is not set, the cluster will be queried to get the current version.

        The retry behavior is determined by the values of constants in _retry_sf_exception_constants.py,
        self.retry_count, and the retry_delay property on the SfApiException objects.
        self.retry_faults_override will override what faults this method retries on.

        :param str method: method to call
        :param dict params: the params for the method.
        :param sf_api.SfAuth auth: Override the SfAuth object
        :rtype: dict
        """
        auth = auth or self._auth
        json_request = {'method': method, 'id': get_json_rpc_id()}
        if params is not None:
            json_request['params'] = params
        _log.debug('Sending request to %s: %s', auth.uri, json.dumps(json_request))
        requests_response = self._session.post(auth.uri, json=json_request, timeout=self.timeout,
                                               auth=auth)
        requests_response.raise_for_status()
        return requests_response.json()

    # PROPERTIES
    @property
    def uri(self):
        """URI for a node or cluster API.

        This uri defines the JSON RPC endpoint with the version.

        :rtype: str
        """
        return self._auth.uri

    @property
    def root_uri(self):
        """Root URI for node or cluster.

        Only includes ip and (if it is a node) the port.

        :rtype: str
        """
        return self._auth.root_uri

    @property
    def version(self):
        """Return the version for this object instance.

        :rtype: sf_api.SfVersion
        """
        return self._auth.version

    @version.setter
    def version(self, value):
        """Sets the version for this object instance.

        :param sf_api.SfVersion value: value of the API version number to use, e.g. '11.0'
        """
        self._auth.version = value

    # PRIVATE METHODS
    @classmethod
    def _construct_params(cls, params):
        """Construct the params to be sent to call_method.

        Enums are converted to primitives. If the first element of a list or tuple is an enum, it is
        assumed that the list is homogeneous, and all elements are converted to primitives.

        None values will not be part the params dict.

        Usage::
            params = self._construct_params({'accountIDs': account_ids, 'force': force})

        :param dict params: The params dict to be be properly formatted for the api call
        :rtype: dict
        :returns: The given params dict with none values stripped out and enums converted to primitives
        """
        new_params = {}
        for key, value in six.iteritems(params):
            if value is None:
                continue
            if cls._is_list_of_enums(value):
                new_params[key] = [v.value for v in value]
            elif isinstance(value, aenum.Enum):
                new_params[key] = value.value
            else:
                new_params[key] = value
        return new_params

    @staticmethod
    def _is_list_of_enums(value):
        """Determine if given value is a list or tuple of enums.

        :param Any value: The object to check
        :rtype: bool
        """
        if value and type(value) in (list, tuple) and isinstance(value[0], aenum.Enum):
            return True
        else:
            return False

    @staticmethod
    def _process_response(requests_response, method, result_dm_class=None):
        """Validate response and convert to custom Response object.

        Check for error or no result, and raise exceptions if needed. Return SfResponse object.

        :param requests.Response|dict requests_response: response object.
        :param sf_api.ClusterMethod|sf_api.NodeMethod method: The name of the method that was called.
        :param Any result_dm_class: The API Response class the result is modeled after. If there is no
            result data model associated with the call, this can be None.
        :rtype: sf_api.SfResponse
        :raises sf_api.errors.MalformedResponseError: If the response is not valid JSON.
        :raises sf_api.errors.SfApiException: If the response contains an error object.
        """
        if isinstance(requests_response, requests.Response):
            requests_response = requests_response.json()
        response = SfResponse(requests_response, method, result_dm_class=result_dm_class)
        if response.error is not None:
            error_name = response.error.name
            exception_class = api_errors.get_exception_class_from_string(error_name)
            _log.debug('SfApiException received from %s: %s', method, requests_response)
            raise exception_class(
                'Server returned error: {}<{}> for {}'.format(
                    error_name, response.error.message, response),
                response=response
            )
        elif response.result is None:
            raise api_errors.MalformedResponseError('No result key found in response',
                                                    response=response)

        return response
