"""The Base REST client and Oauth2 Authentication class live here."""
import logging
import time

import requests
from requests import auth as requests_auth

from sf_api import api_constants

_log = logging.getLogger(__name__)


class Oauth2Auth(requests_auth.AuthBase):
    """Holds authentication information and URI for interacting with a REST API.

    :param sf_models.ResourceCluster cluster_dm: the cluster data model
    :param str client_id: the ID of the client requesting a token (get from Hydra)
    :param str base_url: the url of the service to authenticate to
    """
    def __init__(self, cluster_dm, client_id, base_url):
        self.cluster_dm = cluster_dm
        self.client_id = client_id
        self.base_url = base_url
        self._bearer_token = None
        self._token_expires_time = None

    def __repr__(self):
        """Pretty format with token value."""
        return '{}<token={}>'.format(self.__class__.__name__, self._bearer_token)

    @property
    def is_expired(self):
        current_time = time.time()
        if self._token_expires_time is None or current_time > self._token_expires_time:
            return True
        return False

    @property
    def bearer_token(self):
        """Return the version for this object instance.

        If version is None, the cluster will be queried for the current version.

        :rtype: str
        :returns bearer_token: token from Element auth service
        """
        if self._bearer_token is None or self.is_expired is True:
            self._bearer_token = self._get_bearer_token()
        return self._bearer_token

    def _get_bearer_token(self):
        token_url = 'https://{}/auth/connect/token'.format(self.cluster_dm.mvip)
        token_request = {
            'client_id': self.client_id,
            'grant_type': 'password',
            'username': self.cluster_dm.username,
            'password': self.cluster_dm.password
        }
        current_time = time.time()
        session = requests.session()
        session.verify = False
        requests_response = session.post(url=token_url, data=token_request)
        requests_response.raise_for_status()
        rr = RestResponse(requests_response)
        seconds_to_expiration = rr.result['expires_in']
        self._token_expires_time = current_time + seconds_to_expiration
        return rr.result['access_token']

    def __call__(self, request):
        request.headers['Authorization'] = 'Bearer {}'.format(self.bearer_token)
        return request


class BaseRestClient(object):
    """Abstract base class for interfacing with a SolidFire REST API.

    :param sf_models.ResourceCluster cluster_dm: cluster data model
    :param str client_id: the ID of the auth client
    :param str base_url: the base URL (mnode MIP, generally) we're using for all calls
    :param bool authenticate: whether or not to add a token header to calls
    """
    def __init__(self, cluster_dm, client_id, base_url, authenticate=True):
        self.base_url = base_url.rstrip('/') if base_url.endswith('/') else base_url
        self.authenticate = authenticate
        self._auth = Oauth2Auth(cluster_dm=cluster_dm, client_id=client_id, base_url=base_url) \
            if authenticate else None
        self._headers = {'accept': 'application/json', 'Content-Type': 'application/json'}
        # Timeout for http calls. (connection timeout, read timeout)
        self.timeout = (300, 300)

    def get(self, url='', service='', path='', params=None, result_dm_class=None):
        """Process an http GET request to a url, return results.

        :param str url: URL to GET.
        :param str service: REST service.
        :param str path: the path to the REST resources (beyond base_url)
        :param dict params: a dictionary of params to pass
        :param result_dm_class: The API Response class the result is modeled after.
        :type result_dm_class: (dict) -> sf_api.models.Base
        :return: The result of the Get API call.
        :rtype: sf_api.RestResponse
        """
        assert (url and not service) or (not url and service), 'Either URL or service must be used'
        url = url if url else self._construct_url(service, path)
        session = requests.session()
        session.verify = False
        response = session.get(url, params=params, auth=self._auth, timeout=self.timeout)
        return self._process_response(response, result_dm_class)

    def post(self, service, path='', request_json=None, result_dm_class=None, params=None):
        """Process an http POST request to a url, return results.

        :param str service: REST service.
        :param str path: the path to the REST resources (beyond base_url)
        :param dict request_json: a dictionary to pass to request json parameter.
        :param dict|None params: Dictionary of URL params to add (not data!)
        :param result_dm_class: The API Response class the result is modeled after.
        :type result_dm_class: (dict) -> sf_api.models.Base
        :return: The response from the post API call
        :rtype: sf_api.RestResponse
        """
        url = self._construct_url(service, path)
        session = requests.session()
        session.verify = False
        response = session.post(url, json=request_json, headers=self._headers, auth=self._auth,
                                timeout=self.timeout, params=params)
        return self._process_response(response, result_dm_class)

    def put(self, service, data=None, path='', result_dm_class=None, params=None, files=None,
            headers=None):
        """Process an http PUT method on a REST resource, return results.

        :param str service: REST service.
        :param str path: the path to the REST resources (beyond base_url)
        :param dict|None data: a dictionary of params to pass
        :param dict|None files: dict for file requests.
        :param dict|None headers: dict for headers.
        :param result_dm_class: The API Response class the result is modeled after.
        :param dict|None params: Dictionary of URL params to add (not data!)
        :type result_dm_class: (dict) -> sf_api.models.Base
        :return: The response from the Put API call
        :rtype: sf_api.RestResponse
        """
        url = self._construct_url(service, path)
        session = requests.session()
        session.verify = False
        headers = headers if headers else self._headers
        response = session.put(url, data=data, headers=headers, auth=self._auth,
                               timeout=self.timeout, params=params, files=files)
        return self._process_response(response, result_dm_class)

    def delete(self, service, path='', result_dm_class=None):
        """Process an http DELETE request on a REST resource, return results.

        :param str service: REST service.
        :param str path: the path to the REST resources (beyond base_url)
        :param result_dm_class: The API Response class the result is modeled after.
        :type result_dm_class: (dict) -> sf_api.models.Base
        :return: response from delete API call
        :rtype: sf_api.RestResponse
        """
        url = self._construct_url(service, path)
        session = requests.session()
        session.verify = False
        response = session.delete(url, headers=self._headers, auth=self._auth, timeout=self.timeout)
        return self._process_response(response, result_dm_class)

    def _construct_url(self, service, path):
        """Construct the url to use in the requests call.

        :param str service: REST service.
        :param str path: path to REST resource
        :return url
        :rtype: str
        """
        path = path.lstrip('/') if path.startswith('/') else path
        url = 'https://{}/{}/{}'.format(self.base_url, service, path)
        return url

    @staticmethod
    def _process_response(requests_response, result_dm_class=None):
        """Validate response and convert to custom Response object.

        Check for error or no result, and raise exceptions if needed.

        :param requests.Response requests_response: response object.
        :param Any result_dm_class: The API Response class the result is modeled after. If there is no
            result data model associated with the call, this can be None.
        :rtype: sf_api.RestResponse
        """
        requests_response.raise_for_status()
        rr = RestResponse(requests_response, result_dm_class=result_dm_class)
        if rr and result_dm_class:
            rr.result_dm.validate()
        return rr


class RestResponse(object):
    """Object representing a response from the SF API.

    :param requests.Response raw_data: The full dictionary response from REST.
    :param result_dm_class: The API Response class the result is modeled after.
    :type result_dm_class: (dict) -> sf_api.models.Base
    """
    def __init__(self, raw_data, result_dm_class=None):
        self.raw_data = raw_data
        self._result_dm_class = result_dm_class
        self._result_dm = None
        self.status_code = 0
        if self.raw_data:
            if self.raw_data.text:
                self.result = self.raw_data.json()
            if hasattr(self.raw_data, api_constants.STATUS_CODE):
                self.status_code = self.raw_data.status_code

    @property
    def result_dm(self):
        """Get the result in data model form.

        :rtype: sf_api.result_models.BaseApiResult
        """
        if self._result_dm is None and self._result_dm_class is not None:
            self._result_dm = self._result_dm_class(self.result)

        return self._result_dm
