"""Cluster API module."""

from sf_api import (
    _base,
    enums,
    result_models,
)


class ClusterApi(_base.BaseJsonRpcClient):
    """SolidFire Cluster API class.

    Class to interact with a SoldFire Cluster over the JSON-RPC API. Username and password must be
    provided the first time this class in instantiated. Version, if not provided, will be determined
    upon instantiation. If username, password or version is provided again, the stored auth object will
    be updated.

    :param str mvip: The management virtual IP of cluster
    :param str username: username of a cluster admin, e.g. "admin"
    :param str password: password of a cluster admin e.g. "admin"
    :param sf_api.SfVersion version: version of the cluster, e.g "9.0"
    :param bool authenticate: If true, authenticate resource in constructor (default=True)
    :param typing.Iterable[Any] retry_sf_exceptions_override: SolidFire exceptions that all
        methods will retry on. If provided, this will override the retry faults for all methods.
    :param int retry_count: Number of times to retry before raising an exception
    :param int port: Optional port number
    """

    def __init__(self, mvip, username=None, password=None, version=None, authenticate=True,
                 retry_sf_exceptions_override=None, retry_count=2, port=None):
        super(ClusterApi, self).__init__(
            ip=mvip,
            username=username,
            password=password,
            version=version,
            authenticate=authenticate,
            retry_sf_exceptions_override=retry_sf_exceptions_override,
            retry_count=retry_count,
            port=port
        )

    @property
    def mvip(self):
        return self.ip

    def get_api(self, show_internal=True):
        """Call GetAPI.

        Return a list of all customer facing API calls with the option to include internal calls.

        :param bool show_internal: If true, show internal API calls
        :rtype: sf_api.SfResponse
        """
        return self.call_method(enums.ClusterMethod.get_api,
                                dict(__includeSecretMethods=show_internal), result_dm_class=None)

    def get_cluster_stats(self):
        """Call GetClusterStats.

        SfResponse result data model is :py:class:`~sf_api.result_models.cluster.GetClusterStats`

        :rtype: sf_api.SfResponse
        """
        return self.call_method(enums.ClusterMethod.get_cluster_stats,
                                result_dm_class=result_models.GetClusterStats)
