"""Errors for the SF Api package.

All the exceptions subclassed off of SfApiException must have a member in sf_api.enums.SfExceptions
Enum. Make sure that the name of the exception matches the value of the enum (case sensitive).

e.g.
sf_api.enums.SfException.api_not_permitted has the value 'xAPINotPermitted'
This corresponds to errors.APINotPermitted defined here. Note the exception of the leading character 'x',
all the rest of the string of the enum is the same as the name of the exception. This is mandatory.
"""


class Error(Exception):
    """Base Exception for SF Api Package."""


class ResponseError(Error):
    """Error that includes a response.

    :param str msg: Exception message
    :param sf_api.SfResponse response: the response object
    """
    def __init__(self, msg, response):
        super(ResponseError, self).__init__(msg)
        self.response = response


class MalformedResponseError(ResponseError):
    """The response is invalid."""


class SfApiException(ResponseError):
    """The request returned a response with an error code."""

    @property
    def retry_delay(self):
        """Time in seconds to wait before retrying.

        Child classes can override this to change the interval.

        :rtype: int
        """
        return 1


class AccountIDDoesNotExist(SfApiException):
    pass


class AddNonEncryptableNodes(SfApiException):
    pass


class AddNonFipsDrives(SfApiException):
    pass


class AddNonFipsNodes(SfApiException):
    pass


class AdminIDDoesNotExist(SfApiException):
    pass


class AlreadyInVolumeAccessGroup(SfApiException):
    pass


class APINotPermitted(SfApiException):
    pass


class AssertionFailure(SfApiException):
    pass


class AsyncHandleDoesNotExist(SfApiException):
    pass


class BackupTargetIDDoesNotExist(SfApiException):
    pass


class BinSyncInProgress(SfApiException):
    pass


class CheckFailure(SfApiException):
    pass


class ClusterAlreadyCreated(SfApiException):
    pass


class ClusterPairIDDoesNotExist(SfApiException):
    pass


class ConflictingListItems(SfApiException):
    pass


class ConflictingParameters(SfApiException):
    pass


class ConflictingParameterValues(SfApiException):
    pass


class CsrGenerationFailed(SfApiException):
    pass


class DaemonStatusChangeFailed(SfApiException):
    pass


class DaemonUnknownType(SfApiException):
    pass


class DBBadArguments(SfApiException):
    pass


class DBClosing(SfApiException):
    pass


class DBConnectionLoss(SfApiException):

    @property
    def retry_delay(self):
        """Time in seconds to wait before retrying.

        :rtype: int
        """
        return 5


class DBInvalidState(SfApiException):
    pass


class DBMultiCommitFailed(SfApiException):
    pass


class DBOperationTimeout(SfApiException):
    pass


class DBPathExists(SfApiException):
    pass


class DBPathNotEmpty(SfApiException):
    pass


class DBNoServerResponse(SfApiException):
    pass


class DBNoSuchPath(SfApiException):
    pass


class DBRuntimeInconsistency(SfApiException):
    pass


class DBSessionExpired(SfApiException):
    pass


class DBSessionMoved(SfApiException):
    pass


class DBVersionMismatch(SfApiException):
    pass


class DriveIDDoesNotExist(SfApiException):
    pass


class DuplicateUsername(SfApiException):
    pass


class EkmGenericError(SfApiException):
    pass


class EnsembleInvalidSize(SfApiException):
    pass


class ExceededLimit(SfApiException):
    pass


class EmptyBroadcast(SfApiException):
    pass


class EncryptionAtRestDisabled(SfApiException):
    pass


class FileNotFound(SfApiException):
    pass


class FileTransferFailed(SfApiException):
    pass


class FreezeTimeout(SfApiException):
    pass


class GCAlreadyRunning(SfApiException):
    pass


class GCNotStartedDueTocSkipGCSet(SfApiException):
    pass


class IdpAuthenticationEnabled(SfApiException):
    pass


class IdpAuthenticationNotEnabled(SfApiException):
    pass


class InitiatorDoesNotExist(SfApiException):
    pass


class InvalidAccountName(SfApiException):
    pass


class InvalidAPIParameter(SfApiException):
    pass


class InvalidCertificate(SfApiException):
    pass


class InvalidClusterState(SfApiException):
    pass


class InvalidDriveID(SfApiException):
    pass


class InvalidEncryptionAtRestStateTransition(SfApiException):
    pass


class InvalidFipsDriveConfiguration(SfApiException):
    pass


class InvalidLun(SfApiException):
    pass


class InvalidNetworkInterface(SfApiException):
    pass


class InvalidPairingKey(SfApiException):
    pass


class InvalidParameter(SfApiException):
    pass


class InvalidParameterType(SfApiException):
    pass


class InvalidRegex(SfApiException):
    pass


class InvalidScheduleParameters(SfApiException):
    pass


class InvalidScheduleType(SfApiException):
    pass


class InvalidSnapshotSchedule(SfApiException):
    pass


class InvalidValue(SfApiException):
    pass


class InvalidVolumeAccessGroupName(SfApiException):
    pass


class IsNotPrimary(SfApiException):
    pass


class KeyPairGenerationFailed(SfApiException):
    pass


class KeyPairInUse(SfApiException):
    pass


class KeyProviderActive(SfApiException):
    pass


class KeyProviderConfigurationError(SfApiException):
    pass


class KeyProviderFailed(SfApiException):
    pass


class KeyProviderUnavailable(SfApiException):
    pass


class KeyServerLastForActiveProvider(SfApiException):
    pass


class KeyServerUnavailable(SfApiException):
    pass


class KeyServerFailed(SfApiException):
    pass


class KeyServerConflict(SfApiException):
    pass


class KeyServerAlreadyExists(SfApiException):
    pass


class KeyServerHasProviderAssigned(SfApiException):
    pass


class KeyServerInvalidAuth(SfApiException):
    pass


class KeyServerAuthFailed(SfApiException):
    pass


class KeyServerSslError(SfApiException):
    pass


class MaxVolumesPerGroupSnapshotExceeded(SfApiException):
    pass


class LunInUse(SfApiException):
    pass


class MismatchedCertificateAndKey(SfApiException):
    pass


class MissingListItem(SfApiException):
    pass


class MissingParameter(SfApiException):
    pass


class NetworkConfigCommitFailed(SfApiException):
    pass


class NodeIDDoesNotExist(SfApiException):
    pass


class NoHandler(SfApiException):
    pass


class NonFipsDrivesPresent(SfApiException):
    pass


class NonFipsNodesPresent(SfApiException):
    pass


class NotInVolumeAccessGroup(SfApiException):
    pass


class NotPrimary(SfApiException):
    pass


class NotReadyForIo(SfApiException):
    pass


class ObjectDoesNotExist(SfApiException):
    pass


class PackageActive(SfApiException):
    pass


class PackageAlreadyInstalled(SfApiException):
    pass


class PackageNotFound(SfApiException):
    pass


class PackageRemovalNotAllowed(SfApiException):
    pass


class PackageValidationFailed(SfApiException):
    pass


class PackageFormatUnsupported(SfApiException):
    pass


class ParameterNotValid(SfApiException):
    pass


class PermissionDenied(SfApiException):
    pass


class PrimaryClusterAdmin(SfApiException):
    pass


class ProviderVerificationFailed(SfApiException):
    pass


class QoSPolicyDoesNotExist(SfApiException):
    pass


class QoSPolicyInvalidName(SfApiException):
    pass


class RecvTimeout(SfApiException):
    pass


class Shutdown(SfApiException):
    pass


class SliceServiceSelectionFailed(SfApiException):
    pass


class SnapshotFailed(SfApiException):
    pass


class SnapshotIDDoesNotExist(SfApiException):
    pass


class SoftwareInstallPackageNotValid(SfApiException):
    pass


class SoftwareInstallNotInProgress(SfApiException):
    pass


class SslKeysDoNotExist(SfApiException):
    pass


class SystemCommandFailure(SfApiException):
    pass


class TooFewDrives(SfApiException):
    pass


class TooManyAuthSessions(SfApiException):
    pass


class TooManyAuthSessionsForUser(SfApiException):
    pass


class UnbalancedCluster(SfApiException):
    pass


class UnknownAccount(SfApiException):
    pass


class UnknownAPIMethod(SfApiException):
    pass


class UnknownAPIConstant(SfApiException):
    pass


class UnknownAuthSession(SfApiException):
    pass


class UnknownBlockID(SfApiException):
    pass


class UnknownEnumType(SfApiException):
    pass


class UnknownException(SfApiException):
    pass


class UnknownReport(SfApiException):
    pass


class UnknownServiceID(SfApiException):
    pass


class UnknownUsername(SfApiException):
    pass


class UnrecognizedEnumString(SfApiException):
    pass


class ValueLimitExceeded(SfApiException):
    pass


class VolumeAccessGroupIDDoesNotExist(SfApiException):
    pass


class VolumeIDDoesNotExist(SfApiException):
    pass


class WaitForNetworkInterfaceStatusFailed(SfApiException):
    pass


def get_exception_class_from_string(exception_string):
    """Get the SfApiException object for the given the SolidFire API error string.

    If a corresponding exception is not found, the generic SfApiException class will be returned.

    :param str exception_string: SolidFire exception string, e.g. xAccountIDDoesNotExist
    :rtype: typing.Callable
    :return: The corresponding SfApiException class
    """
    exception_class_name = exception_string.lstrip('x')
    exception_class = globals().get(exception_class_name) or SfApiException
    return exception_class
