"""Account Data Models."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import types


class Account(sf_models.Base):
    """Account data model.

    Corresponds to the elements of the accounts object in the ListAccounts response.
    """
    account_id = types.IntIdType(required=True, serialized_name=api_constants.ACCOUNT_ID)
    attributes = types.DictType(types.StringType(), default=dict)
    initiator_secret = types.StringType(serialized_name=api_constants.INITIATOR_SECRET)
    status = types.EnumType(enums.AccountStatus, required=True)
    target_secret = types.StringType(serialized_name=api_constants.TARGET_SECRET)
    username = types.StringType(required=True)
    volumes = types.ListType(types.IntType())

    # Version 9.0 Additions

    # This field only exists up when the account acts as a VVOL's storage container
    account_type = types.EnumType(
        enum_class=enums.AccountType,
        serialized_name=api_constants.ACCOUNT_TYPE,
        version_added=sf_version.SfVersion('9.0')
    )
    storage_container_id = types.UUIDType(
        required=True,
        serialized_name=api_constants.STORAGE_CONTAINER_ID,
        version_added=sf_version.SfVersion('9.0')
    )

    @classmethod
    def create(cls, username=None, initiator_secret=None, target_secret=None, attributes=None,
               status=None, volumes=None, account_id=None, account_type=None, storage_container_id=None):
        """Define an account profile to pass to create functions/methods.

        :param str username: Account username, must be unique.
        :param str initiator_secret: Account iSCSI initiator secret.
        :param str target_secret: Account iSCSI target secret.
        :param dict attributes: Any additional key value pairs to be added to the Account.
        :param sf_api.enums.VolumeStatus|str status: Volume status enum
        :param list[int] volumes: List of volume ids
        :param int account_id: Unique ID of account
        :param sf_api.enums.AccountType|str account_type: Type of account
        :param str storage_container_id: UUID of storage container
        :return: Partially defined account data model to be used for account creation.
        :rtype: sf_api.models.Account
        """
        return super(Account, cls).create(
            username=username,
            initiator_secret=initiator_secret,
            target_secret=target_secret,
            attributes=attributes,
            status=status,
            volumes=volumes,
            account_id=account_id,
            account_type=account_type,
            storage_container_id=storage_container_id
        )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'account_id', self.account_id,
            'username', self.username
        )


class AccountEfficiency(sf_models.Base):
    """Account efficiency data model."""
    compression = types.DecimalType(required=True)
    deduplication = types.DecimalType(required=True)
    missing_volumes = types.ListType(types.IntIdType(), required=True, serialized_name='missingVolumes')
    thin_provisioning = types.DecimalType(required=True, serialized_name='thinProvisioning')
    timestamp = types.DateTimeType(required=True)
