"""Async Data Models."""
import sf_models
import six

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import (
    _cluster,
    types,
)


class AddDrivesDetails(sf_models.Base):
    """AddDrivesDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is DriveAdd.
    """
    aborted_drives = types.ListType(
        field=types.IntType(),
        required=True,
        default=list,
        serialized_name=api_constants.ABORTED_DRIVES
    )
    added_drives = types.ListType(
        field=types.IntType(),
        required=True,
        default=list,
        serialized_name=api_constants.ADDED_DRIVES
    )
    pending_drives = types.ListType(
        field=types.IntType(),
        required=True,
        default=list,
        serialized_name=api_constants.PENDING_DRIVES
    )

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to identify it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if (data.get(api_constants.ABORTED_DRIVES) is not None
                and data.get(api_constants.ADDED_DRIVES) is not None
                and data.get(api_constants.PENDING_DRIVES) is not None):
            return AddDrivesDetails

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'added_drives', self.added_drives,
            'pending_drives', self.pending_drives,
            'aborted_drives', self.aborted_drives
        )


class RemoveDrivesDetails(sf_models.Base):
    """RemoveDrivesDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is
    DriveRemoval.
    """
    bs_disks = types.ListType(types.IntType(), default=list, serialized_name=api_constants.BS_DISKS)
    ss_disks = types.ListType(types.IntType(), default=list, serialized_name=api_constants.SS_DISKS)

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to identify it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.BS_DISKS) is not None and data.get(api_constants.SS_DISKS) is not None:
            return RemoveDrivesDetails

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'bs_disks', self.bs_disks,
            'ss_disks', self.ss_disks
        )


class SecureEraseDrivesDetails(sf_models.Base):
    """SecureEraseDrivesDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is
    DriveSecureErase.
    """
    drives = types.ListType(
        field=types.IntType(),
        required=True,
        default=list,
        serialized_name=api_constants.DRIVES
    )
    successful_completion = types.BooleanType(required=True, serialized_name='successfulCompletion')

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to identify it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.DRIVES) is not None and data.get('successfulCompletion') is not None:
            return SecureEraseDrivesDetails

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'drives', self.drives
        )


class CloneVolumeDetails(sf_models.Base):
    """CloneVolumeDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is Clone.
    """
    clone_id = types.IntType(required=True, serialized_name=api_constants.CLONE_ID)
    message = types.StringType(required=True)
    volume_id = types.IntType(required=True, serialized_name=api_constants.VOLUME_ID)

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.CLONE_ID) is not None \
                and data.get(api_constants.VOLUME_ID) is not None:
            return CloneVolumeDetails

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'clone_id', self.clone_id,
            'volume_id', self.volume_id,
            'message', self.message
        )


class GroupCloneVolumesDetails(sf_models.Base):
    """GroupCloneVolumesDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is GroupClone.
    """
    group_clone_id = types.IntType(required=True, serialized_name='groupCloneID')
    message = types.StringType(required=True)
    volumes = types.ListType(types.IntIdType(), required=True, serialized_name=api_constants.VOLUMES)

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get('groupCloneID') is not None and data.get(api_constants.VOLUMES) is not None:
            return GroupCloneVolumesDetails

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'group_clone_id', self.group_clone_id,
            'volumes', self.volumes,
            'message', self.message
        )


class AddNodesDetails(sf_models.Base):
    """AddNodes data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is
    rtfi_pending_node.
    """
    assigned_node_id = types.IntType(required=True, serialized_name=api_constants.ASSIGNED_NODE_ID)
    percent = types.IntType(required=True)
    state = types.StringType(required=True)
    status_url_all = types.URLType(required=True, serialized_name='statusUrlAll')
    status_url_current = types.URLType(required=True, serialized_name='statusUrlCurrent')
    status_url_log_file = types.URLType(required=True, serialized_name='statusUrlLogfile')

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.ASSIGNED_NODE_ID) is not None and data.get('statusUrlAll') is not None:
            return AddNodesDetails

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'assigned_node_id', self.assigned_node_id,
            'percent', self.percent,
            'state', self.state,
        )


class FillSliceDetails(sf_models.Base):
    """FillSliceDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is
    FillSlice.
    """
    block_services = types.DictType(types.StringType(), serialized_name='blockServices')
    service_details = types.BaseType(serialized_name='serviceDetails')
    slice_id = types.IntType(required=True, serialized_name=api_constants.SLICE_ID)
    slice_services = types.DictType(types.StringType(), required=True, serialized_name='sliceServices')

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.SLICE_ID) is not None and data.get('sliceServices') is not None:
            return FillSliceDetails

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'slice_id', self.slice_id,
            'slice_services', self.slice_services
        )


class GarbageCollectionDetailsPreMg(sf_models.Base):
    """GarbageCollectionDetails data model.

    Corresponds to the details object in the GetAsyncResult response when the result_type is
    GarbageCollection.
    """
    name = types.StringType()
    message = types.StringType()
    block_services = types.DictType(types.StringType(), required=True, serialized_name='blockServices')
    service_details = types.BaseType(required=True, serialized_name='serviceDetails')
    slice_services = types.DictType(types.StringType(), required=True, serialized_name='sliceServices')

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.SLICE_ID) is None and data.get('sliceServices') is not None and (
                    data.get('blockServices') is not None):
            return GarbageCollectionDetailsPreMg

    def __repr__(self):
        return '{}<{}; {}={}, {}={}>'.format(
            self.__class__.__name__,
            self.name,
            'block_services', self.block_services,
            'slice_services', self.slice_services
        )


class GarbageCollectionDetails(sf_models.Base):
    """This represents the output of GetGCStatus, and the details attribute of GetAsyncResult for GC.

    Prior to 12.0, GetGCStatus and GetAsyncResult for GC were separate objects (See
    GarbageCollectionDetailsPreMg for the old format). 12.0 and forward they
    are equivalent.
    """
    completed_replica_count = types.IntType(serialized_name='completedReplicaCount',
                                            version_added=sf_version.SfVersion('12.0'))
    double_helix_discarded_blocks_count = types.IntType(
        serialized_name='doubleHelixDiscardedBlocksCount',
        version_added=sf_version.SfVersion('12.0')
    )
    double_helix_retained_blocks_count = types.IntType(
        serialized_name='doubleHelixRetainedBlocksCount',
        version_added=sf_version.SfVersion('12.0')
    )
    failed_replica_count = types.IntType(serialized_name='failedReplicaCount',
                                         version_added=sf_version.SfVersion('12.0'))
    generation = types.IntType(required=True)
    progress_phase = types.EnumType(enum_class=enums.GcProgressPhase,
                                    serialized_name='progressPhase',
                                    version_added=sf_version.SfVersion('12.0'))
    services = types.ListType(types.ModelType(_cluster.ServiceGarbageCollectionStatus), required=True)
    single_helix_discarded_blocks_count = types.IntType(
        serialized_name='singleHelixDiscardedBlocksCount',
        version_added=sf_version.SfVersion('12.0')
    )
    single_helix_retained_blocks_count = types.IntType(
        serialized_name='singleHelixRetainedBlocksCount',
        version_added=sf_version.SfVersion('12.0')
    )
    status = types.EnumType(required=True, enum_class=enums.GcStatus)
    total_replicas = types.IntType(serialized_name='totalReplicas',
                                   version_added=sf_version.SfVersion('12.0'))

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get('completedReplicaCount') is not None:
            return GarbageCollectionDetails

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'status', self.status,
            'generation', self.generation
        )


class ErrorResponse(sf_models.Base):
    """GetAsyncResult error data model.

    Corresponds to the error field in the GetAsyncResult response when the error field is returned.
    """
    message = types.StringType()
    name = types.StringType()

    @staticmethod
    def _claim_polymorphic(data):
        if data.get('name') is not None and data.get('message') is not None and len(data) == 2:
            return ErrorResponse


class DetailsString(sf_models.Base):
    """DetailsString data model.

    Corresponds to the data/details/result field in SetClusterStructure.
    """
    details_string = types.StringType()

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get('details_string') is not None and len(data) == 1:
            return DetailsString


class BSCheckDetails(sf_models.Base):
    """BSCheck data model.

    This represents the details attribute of GetAsyncResult for BSCheck.
    """
    service_details = types.BaseType(serialized_name='serviceDetails')
    slice_services = types.DictType(types.StringType(), required=True, serialized_name='sliceServices')

    @staticmethod
    def _claim_polymorphic(data):
        """Member of a PolyType, know how to select it.

        :param dict data: A dictionary with the data to match to a model.
        :rtype: schematics.models.ModelMeta
        """
        if data.get(api_constants.SLICE_ID) is None and data.get('sliceServices') is not None and (
                data.get('blockServices') is None):
            return BSCheckDetails


# Tuple of all possible GetAsyncResult details models
DETAILS_MODELS = (
    AddDrivesDetails,
    RemoveDrivesDetails,
    SecureEraseDrivesDetails,
    CloneVolumeDetails,
    GroupCloneVolumesDetails,
    AddNodesDetails,
    FillSliceDetails,
    GarbageCollectionDetailsPreMg,
    GarbageCollectionDetails,
    ErrorResponse,
    DetailsString,
    BSCheckDetails,
)


class AsyncResult(sf_models.Base):
    """Async result data model.

    Corresponds to the response of GetAsyncResult and the list nested in ListAsyncResults.

    Prior to 12.0 the AsyncResult fields change depending on the state of the asynchronous process.

        * If the method GetAsyncResult fails, there is no 'result' in the response only a SfErrorResponse
        * If the async method is running, there will be a 'details' dict in the SFResult
        * If the async method succeeded, the 'details' dict becomes a 'result' dict in the SFResult
        * If the async method failed, the 'details' dict becomes an 'errors' dict in the SFResult
        * If the async result is retrieved via ListAsyncResults, 'details' becomes 'data'

    :param raw_data: The data used to initialize the model
    :type raw_data: dict[str, Any]
    """
    # Product is inconsistent about what key to use for details
    _DETAILS_KEYS = [api_constants.DETAILS, api_constants.RESULT, api_constants.DATA,
                     api_constants.ERROR]

    create_time = types.DateTimeType(required=True, serialized_name=api_constants.CREATE_TIME)
    details = types.PolyModelType(
        model_spec=DETAILS_MODELS,
        required=True,
        deserialize_from=_DETAILS_KEYS
    )
    last_update_time = types.DateTimeType(required=True, serialized_name=api_constants.LAST_UPDATE_TIME)
    result_type = types.EnumType(
        enum_class=enums.AsyncResultType,
        required=True,
        serialized_name=api_constants.RESULT_TYPE
    )

    status = types.EnumType(enums.AsyncResultStatus, required=True)

    # ListAsyncResult specific fields.
    async_result_id = types.IntType(required=False, serialized_name='asyncResultID')
    completed = types.BooleanType(required=False)
    success = types.BooleanType(required=False)

    def __init__(self, raw_data=None, **kwargs):
        # Sometimes the product returns a string in place of a json object.
        # Intercept it if it does and turn the field into an object so it can be a PolyModelType.
        details_key = api_constants.DETAILS
        if isinstance(raw_data, dict):
            for k in self._DETAILS_KEYS:
                if k in raw_data:
                    details_key = k
                    if isinstance(raw_data[k], six.string_types):
                        raw_data[k] = dict(details_string=raw_data[k])

        super(AsyncResult, self).__init__(raw_data=raw_data, **kwargs)

        # The product sometimes marks failed results as complete, but we can tell it failed by the
        # existence of the error key. If the error key exists, set status to failed.
        if raw_data:
            if details_key == api_constants.ERROR:
                self.status = enums.AsyncResultStatus.failed

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'async_result_id', self.async_result_id,
            'status', self.status
        )
