"""Cluster Data Model."""
import logging

import sf_models
from sf_models import types

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import (
    _account,
    _drive,
    _general,
    _network,
    _node,
    _schedule,
    _snapmirror,
    _volume,
)

_log = logging.getLogger(__name__)


class Cluster(sf_models.Base):
    """Cluster data model.

    Corresponds to the cluster_info object in the GetClusterInfo response.
    """
    attributes = types.DictType(types.StringType(), required=True)
    encryption_at_rest_state = types.EnumType(
        enum_class=enums.EarState,
        required=True,
        serialized_name=api_constants.EAR_STATE
    )
    ensemble = types.ListType(types.IPAddressType(required=True), required=True)
    mvip = types.IPAddressType(required=True)
    mvip_node_id = types.IntIdType(required=True, serialized_name=api_constants.MVIP_NODE_ID)
    name = types.StringType(required=True)
    rep_count = types.IntType(required=True, serialized_name=api_constants.REP_COUNT)
    svip = types.IPAddressType(required=True)
    svip_node_id = types.IntIdType(required=True, serialized_name=api_constants.SVIP_NODE_ID)
    unique_id = types.StringType(required=True, serialized_name=api_constants.UNIQUE_ID)

    # Version 7.0 Additions
    uuid = types.UUIDType(required=True, version_added=sf_version.SfVersion('7.0'))

    # Version 9.0 Additions
    mvip_interface = types.NetworkInterfaceNameType(
        required=True,
        serialized_name=api_constants.MVIP_INTERFACE,
        version_added=sf_version.SfVersion('9.0')
    )
    mvip_vlan_tag = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name=api_constants.MVIP_VLAN_TAG,
        version_added=sf_version.SfVersion('9.0')
    )
    svip_interface = types.NetworkInterfaceNameType(
        required=True,
        serialized_name=api_constants.SVIP_INTERFACE,
        version_added=sf_version.SfVersion('9.0')
    )
    svip_vlan_tag = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name=api_constants.SVIP_VLAN_TAG,
        version_added=sf_version.SfVersion('9.0')
    )

    # Version 11.0 Additions
    default_protection_scheme = types.EnumType(
        enums.ProtectionScheme,
        required=True,
        serialized_name='defaultProtectionScheme',
        version_added=sf_version.SfVersion('11.0')
    )
    enabled_protection_schemes = types.ListType(
        types.EnumType(enums.ProtectionScheme),
        required=True,
        serialized_name='enabledProtectionSchemes',
        version_added=sf_version.SfVersion('11.0')
    )
    supported_protection_schemes = types.ListType(
        types.EnumType(enums.ProtectionScheme),
        required=True,
        serialized_name='supportedProtectionSchemes',
        version_added=sf_version.SfVersion('11.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'mvip', self.mvip,
            'svip', self.svip
        )


class ClusterAdmin(sf_models.Base):
    """ClusterAdmin data model.

    Corresponds to the elements of the cluster_admins object in the ListClusterAdmins response.
    """
    access = types.ListType(types.EnumType(enums.ClusterAccessLevel), required=True)
    attributes = types.DictType(types.StringType())
    auth_method = types.EnumType(enums.ClusterAuthMethod, required=True,
                                 serialized_name='authMethod')
    cluster_admin_id = types.IntIdType(required=True, serialized_name='clusterAdminID')
    username = types.StringType(required=True, min_length=1, max_length=1024)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'cluster_admin_id', self.cluster_admin_id,
            'username', self.username
        )


class ClusterCapacity(sf_models.Base):
    """ClusterCapacity data model.

    Corresponds to the cluster_capacity object in the GetClusterCapacity response.
    """
    active_sessions = types.IntType(required=True, serialized_name=api_constants.ACTIVE_SESSIONS)
    average_iops = types.IntType(required=True, serialized_name='averageIOPS')
    cluster_recent_io_size = types.IntType(required=True, serialized_name='clusterRecentIOSize')
    compression_percent = types.IntType(serialized_name='compressionPercent')
    current_iops = types.IntType(required=True, serialized_name='currentIOPS')
    date = types.StringType()
    deduplication_percent = types.IntType(serialized_name='deDuplicationPercent')
    efficiency_percent = types.IntType(serialized_name='efficiencyPercent')
    max_iops = types.IntType(required=True, serialized_name='maxIOPS')
    max_provisioned_space = types.IntType(required=True, serialized_name='maxProvisionedSpace')
    max_used_space = types.IntType(required=True, serialized_name='maxUsedSpace')
    non_zero_blocks = types.IntType(required=True, serialized_name='nonZeroBlocks')
    peak_active_sessions = types.IntType(required=True, serialized_name='peakActiveSessions')
    peak_iops = types.IntType(required=True, serialized_name='peakIOPS')
    provisioned_space = types.IntType(required=True, serialized_name='provisionedSpace')
    thin_provisioning_percent = types.IntType(serialized_name='thinProvisioningPercent')
    total_ops = types.IntType(required=True, serialized_name='totalOps')
    unique_blocks = types.IntType(required=True, serialized_name=api_constants.UNIQUE_BLOCKS)
    unique_blocks_used_space = types.IntType(required=True,
                                             serialized_name=api_constants.UNIQUE_BLOCK_USED_SPACE)
    used_space = types.IntType(required=True, serialized_name='usedSpace')
    zero_blocks = types.IntType(required=True, serialized_name='zeroBlocks')

    # Version 6.0 Additions
    active_block_space = types.IntType(
        required=True,
        serialized_name='activeBlockSpace',
        version_added=sf_version.SfVersion('6.0')
    )
    max_over_provisionable_space = types.IntType(
        required=True,
        serialized_name='maxOverProvisionableSpace',
        version_added=sf_version.SfVersion('6.0')
    )
    max_used_metadata_space = types.IntType(
        required=True,
        serialized_name='maxUsedMetadataSpace',
        version_added=sf_version.SfVersion('6.0')
    )
    timestamp = types.DateTimeType(required=True, version_added=sf_version.SfVersion('6.0'))
    used_metadata_space = types.IntType(
        required=True,
        serialized_name='usedMetadataSpace',
        version_added=sf_version.SfVersion('6.0')
    )
    used_metadata_space_in_snapshots = types.IntType(
        required=True,
        serialized_name='usedMetadataSpaceInSnapshots',
        version_added=sf_version.SfVersion('6.0')
    )

    # Version 7.4 Additions
    # TODO : VL-1795
    # Fix for version 8.0 when the field was removed before being added in 8.1
    snapshot_non_zero_blocks = types.IntType(
        required=True,
        serialized_name='snapshotNonZeroBlocks',
        version_added=sf_version.SfVersion('7.4')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'used_space', self.used_space,
            'provisioned_space', self.provisioned_space,
            'max_provisioned_space', self.max_provisioned_space
        )


class ClusterConfig(sf_models.Base):
    """ClusterConfig data model.

    Corresponds to the cluster object in the GetClusterConfig response.
    """
    cipi = types.NetworkInterfaceNameType(required=True)
    cluster = types.StringType(required=True)
    encryption_capable = types.BooleanType(
        required=True,
        serialized_name='encryptionCapable',
        version_added=sf_version.SfVersion('8.7')
    )
    ensemble = types.ListType(
        types.StringType(regex=r'(^\d:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$)|^N/A$'),
        required=True
    )
    fips_drive_configuration = types.BooleanType(
        required=False,
        serialized_name='fipsDriveConfiguration',
        version_added=sf_version.SfVersion('11.5')
    )
    mipi = types.NetworkInterfaceNameType(required=True)
    name = types.StringType(required=True)
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    pending_node_id = types.IntIdType(required=True, serialized_name=api_constants.PENDING_NODE_ID)
    role = types.EnumType(enums.NodeRole, required=True)
    sipi = types.NetworkInterfaceNameType(required=True)
    state = types.EnumType(enums.NodeState, required=True)
    version = types.SfBuildVersionType()

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'cluster', self.cluster,
            'role', self.role,
            'state', self.state
        )

    class Options:
        serialize_when_none = False


class ClusterVersionInfo(sf_models.Base):
    cluster_api_version = types.SfApiVersionType(required=True, serialized_name='clusterAPIVersion')
    cluster_version = types.SfBuildVersionType(required=True, serialized_name='clusterVersion')
    cluster_version_info = types.ListType(
        field=types.ModelType(_node.NodeVersionInfo),
        required=True,
        serialized_name='clusterVersionInfo'
    )
    software_version_info = types.ModelType(
        model_spec=_general.SoftwareVersionInfo,
        required=True,
        serialized_name='softwareVersionInfo'
    )


class ClusterEnsembleNode(sf_models.Base):
    """ClusterEnsembleNode data model.

    Corresponds to the elements of the nodes object in the ClusterEnsemble data model.
    Part of the TestLocateCluster response.
    """
    ip = types.IPAddressType(required=True, serialized_name='IP')
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'ip', self.ip
        )


class ClusterEnsemble(sf_models.Base):
    """ClusterEnsemble data model.

    Corresponds to the ensemble object in the ClusterLocationDetails data model.
    Part of the TestLocateCluster response.
    """
    mode = types.EnumType(enums.EnsembleMode, required=True)
    nodes = types.ListType(types.ModelType(ClusterEnsembleNode), required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'nodes', self.nodes)


class ClusterEvent(sf_models.Base):
    """ClusterEvent data model.

    Corresponds to the elements of the events object in the ListEvents response.
    """
    details = types.BaseType()
    drive_id = types.IntIdType(serialized_name=api_constants.DRIVE_ID, required=True)
    drive_ids = types.ListType(types.IntIdType(), serialized_name=api_constants.DRIVE_IDS, required=True)
    event_info_type = types.EnumType(enums.ClusterEventType,
                                     serialized_name=api_constants.EVENT_INFO_TYPE,
                                     required=True)
    message = types.StringType(required=True)
    node_id = types.IntIdType(serialized_name=api_constants.NODE_ID, required=True)
    service_id = types.IntIdType(serialized_name=api_constants.SERVICE_ID, required=True)
    severity = types.IntType(required=True)
    time_of_report = types.DateTimeType(serialized_name='timeOfReport', required=True)

    # Version 6.0 Additions
    event_id = types.IntIdType(
        serialized_name='eventID',
        version_added=sf_version.SfVersion('6.0'),
        required=True
    )
    time_of_publish = types.DateTimeType(
        serialized_name='timeOfPublish',
        version_added=sf_version.SfVersion('6.0'),
        required=True
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'event_id', self.event_id,
            'event_info_type', self.event_info_type,
            'message', self.message,
            'details', self.details
        )


class ClusterFault(sf_models.Base):
    """ClusterFault data model.

    Corresponds to the elements of the faults object in the ListClusterFaults response.
    """
    checker = types.StringType(required=True, version_removed=sf_version.SfVersion('6.0'))
    code = types.EnumType(enums.ClusterFaultCode, required=True)
    data = types.BaseType()
    date = types.DateTimeType(required=True)
    details = types.StringType(required=True)
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    network_id = types.IntIdType(required=True, serialized_name=api_constants.NETWORK_ID,
                                 version_removed=sf_version.SfVersion('9.0'))
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    resolved = types.BooleanType(required=True)
    resolved_date = types.OptionalDateTimeType(serialized_name='resolvedDate')
    service_id = types.IntIdType(required=True, serialized_name=api_constants.SERVICE_ID)
    severity = types.EnumType(enums.ClusterFaultSeverity, required=True)
    type = types.EnumType(enums.ClusterFaultType, required=True)

    # Version 6.0 Additions
    cluster_fault_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.CLUSTER_FAULT_ID,
        version_added=sf_version.SfVersion('6.0')
    )

    # Version 8.0 Additions
    node_hardware_fault_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.NODE_HARDWARE_FAULT_ID,
        version_added=sf_version.SfVersion('8.0')
    )
    request_id = types.IntIdType(
        required=True,
        serialized_name='requestID',
        version_removed=sf_version.SfVersion('8.0')
    )

    # Version 9.0 additions
    drive_ids = types.ListType(types.IntIdType, required=True, serialized_name=api_constants.DRIVE_IDS,
                               version_added=sf_version.SfVersion('9.0'))
    # replaces network_id
    network_interface = types.StringType(required=True,
                                         version_added=sf_version.SfVersion('9.0'),
                                         serialized_name=api_constants.NETWORK_INTERFACE)

    # Version 11.0 additions
    blocks_upgrade = types.BooleanType(required=True,
                                       serialized_name='blocksUpgrade',
                                       version_added=sf_version.SfVersion('11.0'))
    external_source = types.StringType(required=True,
                                       serialized_name='externalSource',
                                       version_added=sf_version.SfVersion('11.0'))

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'cluster_fault_id', self.cluster_fault_id,
            'code', self.code,
            'details', self.details,
            'resolved', self.resolved,
            'severity', self.severity
        )


class ClusterHardwareInfo(sf_models.Base):
    """ClusterHardwareInfo data model.

    Corresponds to the cluster_hardware_info object in the GetClusterHardwareInfo response.
    """
    drives = types.DictType(types.ModelType(_drive.DriveHardwareInfo))
    nodes = types.DictType(types.ModelType(_node.NodeHardwareInfo))

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'drives', self.drives,
            'nodes', self.nodes
        )


class ClusterLocationDetails(sf_models.Base):
    """ClusterLocationDetails data model.

    Corresponds to the details object in the TestLocateCluster response.
    """
    complete = types.BooleanType(required=True)
    ensemble = types.ModelType(ClusterEnsemble, required=True)
    version = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'version', self.version,
            'complete', self.complete
        )


class ClusterNodeConfig(sf_models.Base):
    """ClusterNodeConfig data model.

    Corresponds to the elements of the nodes object in the GetConfig response.
    """
    cluster = types.ModelType(ClusterConfig, required=True)
    network = types.DictType(types.ModelType(
        _network.NetworkInterfaceDetails), required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'cluster', self.cluster,
            'network', self.network
        )


class ClusterNodeConfigExtended(sf_models.Base):
    """ClusterNodeConfig extended data model.

    Corresponds to the elements of the nodes object in the GetConfigExtended response.
    """
    cluster = types.ModelType(ClusterConfig, required=True)
    network = types.AttributeKeyedModelListType(
        attr_for_key=_network.NetworkInterfaceDetailsExtended.name,
        model_class=_network.NetworkInterfaceDetailsExtended,
        required=True
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'cluster', self.cluster,
            'network', self.network
        )


class ClusterNodeStats(sf_models.Base):
    """ClusterNodeStats data model.

    Corresponds to the node_stats object in the ListNodeStats response.
    """
    nodes = types.ListType(types.ModelType(_node.NodeStats), required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'nodes', self.nodes)


class ZkEntryStat(sf_models.Base):
    """Zookeeper children stat data model."""
    create_time = types.StringType(required=True, serialized_name='createTime')
    data_version = types.IntType(required=True, serialized_name='dataVersion')
    num_children = types.IntType(required=True, serialized_name='numChildren')
    size = types.DataSizeType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'create_time', self.create_time,
            'data_version', self.data_version,
            'num_children', self.num_children,
            'size', self.size
        )


class ZkDatabaseEntry(sf_models.Base):
    """Zookeeper ChildrenData data model.

    Corresponds to response from ListDatabaseChildrenData.
    """
    data = types.DictType(types.BaseType)
    name = types.StringType(required=True)
    stat = types.ModelType(ZkEntryStat, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'data', self.data,
            'stat', self.stat
        )


class ClusterPair(sf_models.Base):
    """ClusterPair data model.

    Corresponds to the elements of the cluster_pairs object in the ListClusterPairs response.
    """
    cluster_name = types.StringType(required=True, serialized_name='clusterName')
    cluster_pair_id = types.IntIdType(required=True, serialized_name='clusterPairID')
    cluster_pair_uuid = types.StringType(
        required=True,
        regex=r'^[0-9a-z-]+$',
        serialized_name='clusterPairUUID'
    )
    uuid = types.IntType(required=True)
    latency = types.IntType(required=True)
    mvip = types.IPAddressType(required=True)
    status = types.EnumType(enums.ClusterPairStatus, required=True)
    version = types.StringType(required=True, regex=r'^[0-9.]+$')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'cluster_pair_id', self.cluster_pair_id,
            'mvip', self.mvip,
            'status', self.status
        )


class ClusterReport(sf_models.Base):
    """ClusterReport data model.

    Corresponds to the elements of the reports object in the GetReport response.
    """
    format = types.StringType(required=True)
    name = types.StringType(required=True)
    path = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'path', self.path
        )


class ClusterService(sf_models.Base):
    """ClusterService data model.

    Corresponds to the service object in the ClusterServiceInfo data model.
    Part of the ListServices response.
    """
    associated_bv = types.IntType(required=True, serialized_name='associatedBV')
    associated_ts = types.IntType(required=True, serialized_name='associatedTS')
    associated_vs = types.IntType(required=True, serialized_name='associatedVS')
    async_result_ids = types.ListType(types.IntIdType(), required=True, serialized_name='asyncResultIDs')
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    first_time_startup = types.BooleanType(required=True, serialized_name='firstTimeStartup')
    ipc_port = types.IntType(required=True, serialized_name='ipcPort')
    iscsi_port = types.IntType(required=True, serialized_name='iscsiPort')
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    service_id = types.IntIdType(required=True, serialized_name=api_constants.SERVICE_ID)
    service_format = types.EnumType(enums.ServiceFormat, required=True,
                                    serialized_name='serviceFormat',
                                    version_added=sf_version.SfVersion('11.0'))
    service_type = types.EnumType(enums.ServiceType, required=True, serialized_name='serviceType')
    started_drive_ids = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name='startedDriveIDs'
    )
    status = types.EnumType(enums.ServiceStatus, required=True)

    # Version 8.6 Additions
    drive_ids = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name=api_constants.DRIVE_IDS,
        version_added=sf_version.SfVersion('8.6')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'service_type', self.service_type,
            'status', self.status
        )


class ClusterServiceInfo(sf_models.Base):
    """ClusterServiceInfo data model.

    Corresponds to the elements of the services object in the ListServices response.
    """
    drive = types.ModelType(_drive.ClusterServiceDrive)
    drives = types.ListType(types.ModelType(_drive.ClusterServiceDrive))
    node = types.ModelType(_node.ActiveNode, required=True)
    service = types.ModelType(ClusterService, required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'service', self.service)


class ClusterStats(sf_models.Base):
    """ClusterStats data model.

    Corresponds to the cluster_stats object in the GetClusterStats response.
    """
    cluster_utilization = types.DecimalType(required=True, serialized_name='clusterUtilization')
    read_bytes = types.IntType(required=True, serialized_name='readBytes')
    read_ops = types.IntType(required=True, serialized_name='readOps')
    timestamp = types.DateTimeType(required=True)
    write_bytes = types.IntType(required=True, serialized_name='writeBytes')
    write_ops = types.IntType(required=True, serialized_name='writeOps')

    # Version 7.0 Additions
    client_queue_depth = types.IntType(
        required=True,
        serialized_name=api_constants.CLIENT_QUEUE_DEPTH,
        version_added=sf_version.SfVersion('7.0')
    )

    # Version 8.1 Additions
    actual_iops = types.IntType(
        required=True,
        serialized_name=api_constants.ACTUAL_IOPS,
        version_added=sf_version.SfVersion('8.1')
    )
    average_iop_size = types.IntType(
        required=True,
        serialized_name=api_constants.AVERAGE_IOP_SIZE,
        version_added=sf_version.SfVersion('8.1')
    )
    latency_usec = types.IntType(
        required=True,
        serialized_name=api_constants.LATENCY_USEC,
        version_added=sf_version.SfVersion('8.1')
    )
    read_latency_usec = types.IntType(
        required=True,
        serialized_name=api_constants.READ_LATENCY_USEC,
        version_added=sf_version.SfVersion('8.1')
    )
    unaligned_reads = types.IntType(
        required=True,
        serialized_name=api_constants.UNALIGNED_READS,
        version_added=sf_version.SfVersion('8.1')
    )
    unaligned_writes = types.IntType(
        required=True,
        serialized_name=api_constants.UNALIGNED_WRITES,
        version_added=sf_version.SfVersion('8.1')
    )
    write_latency_usec = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_LATENCY_USEC,
        version_added=sf_version.SfVersion('8.1')
    )

    # Version 9.0 Additions
    read_bytes_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.READ_BYTES_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )
    read_ops_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.READ_OPS_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )
    sample_period_msec = types.IntType(
        required=True,
        serialized_name=api_constants.SAMPLE_PERIOD_MSEC,
        version_added=sf_version.SfVersion('9.0')
    )
    write_bytes_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_BYTES_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )
    write_ops_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_OPS_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )

    # Version 10.0 Additions
    normalized_iops = types.IntType(
        required=True,
        serialized_name=api_constants.NORMALIZED_IOPS,
        version_added=sf_version.SfVersion('10.0')
    )
    read_latency_usec_total = types.IntType(
        required=True,
        serialized_name=api_constants.READ_LATENCY_USEC_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )
    services_count = types.IntType(
        required=True,
        serialized_name='servicesCount',
        version_added=sf_version.SfVersion('10.0')
    )
    services_total = types.IntType(
        required=True,
        serialized_name='servicesTotal',
        version_added=sf_version.SfVersion('10.0')
    )
    write_latency_usec_total = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_LATENCY_USEC_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'read_ops', self.read_ops,
            'read_bytes', self.read_bytes,
            'write_ops', self.write_ops,
            'write_bytes', self.write_bytes
        )


class EncryptionAtRestInfo(sf_models.Base):
    """Encryption at rest information data model.

    Corresponds to the GetEncryptionAtRestInfo response.
    """
    generation_number = types.IntType(required=True, serialized_name='generationNumber')
    next_generation_to_store = types.IntType(required=True, serialized_name='nextGenerationToStore')
    password_created_timestamp = types.IntType(required=True, serialized_name='passwordCreatedTimestamp')
    redistribution_count = types.IntType(required=True, serialized_name='redistributionCount')
    state = types.StringType(required=True)
    secure_enable_required = types.BooleanType(required=True, serialized_name='secureEnableRequired',
                                               version_added=sf_version.SfVersion('11.7'))
    key_id = types.StringType(required=True, serialized_name='keyID',
                              version_added=sf_version.SfVersion('11.7'))
    key_provider_id = types.IntType(required=True, serialized_name='keyProviderID',
                                    version_added=sf_version.SfVersion('11.7'))


class PrivateProtectionDomain(sf_models.Base):
    """ProtectionDomain data model for private methods.

    Corresponds to the 11.0 implementation of a protection domains response.
    """
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    protection_domain_name = types.StringType(required=True, serialized_name='protectionDomainName')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'protection_domain_name', self.protection_domain_name
        )


class ProtectionDomain(sf_models.Base):
    """ProtectionDomain data model.

    Extensible protection domain data type.
    """
    protection_domain_type = types.EnumType(
        enums.ProtectionDomainType, required=True, serialized_name=api_constants.PROTECTION_DOMAIN_TYPE)
    protection_domain_name = types.StringType(
        required=True, serialized_name=api_constants.PROTECTION_DOMAIN_NAME)

    def __repr__(self):
        return '{}<{}={}><{}={}>'.format(
            self.__class__.__name__,
            'protection_domain_type', self.protection_domain_type,
            'protection_domain_name', self.protection_domain_name
        )


class NodeProtectionDomainLayout(sf_models.Base):
    """NodeProtectionDomainLayout data model.

    Provides a model for the protection domain layout on a specific node.
    """
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    protection_domains = types.ListType(
        types.ModelType(ProtectionDomain), required=True,
        serialized_name=api_constants.PROTECTION_DOMAINS)

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
        )


class TestConnectEnsembleDetails(sf_models.Base):
    """TestConnectEnsembleDetails data model.

    Corresponds to the details object in the TestConnectEnsemble response.
    """
    nodes = types.DictType(types.EnumType(enums.ApiTestResult), required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'nodes', self.nodes)


class HardwareInfoOverrides(sf_models.Base):
    """HardwareInfoOverride data model."""
    nodes = types.DictType(types.DictType(types.StringType), required=True)

    def __repr__(self):
        return '{}'.format(
            self.__class__.__name__,
        )


class ProposedClusterError(sf_models.Base):
    """Proposed cluster error data model."""
    error_code = types.EnumType(
        enum_class=enums.ProposedClusterErrorCode,
        serialized_name='errorCode'
    )
    description = types.StringType(serialized_name='description')
    node_ips = types.ListType(types.IPAddressType(), serialized_name='nodeIPs')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'error_code', self.error_code,
            'description', self.description,
            'node_ips', self.node_ips
        )


class ServiceGarbageCollectionStatus(sf_models.Base):
    """Garbage collection status for a given service.

    Corresponds to the elements of the services object in the GetGCStatus response.
    """
    drive_id = types.IntIdType(required=True, serialized_name='driveID')
    drive_ids = types.ListType(types.IntIdType(), required=True, serialized_name='driveIDs')
    generation = types.IntType(required=True, serialized_name='generation')
    node_id = types.IntIdType(required=True, serialized_name='nodeID')
    service_id = types.IntIdType(required=True, serialized_name='serviceID')
    status = types.EnumType(required=True, enum_class=enums.GcStatus)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive_id', self.drive_id,
            'drive_ids', self.drive_ids,
            'generation', self.generation,
            'node_id', self.node_id,
            'service_id', self.service_id,
            'status', self.status
        )


class ClusterFullThreshold(sf_models.Base):
    fullness = types.EnumType(enums.ClusterStage, required=True)
    slice_reserve_used_threshold_pct = types.IntType(
        required=True,
        serialized_name='sliceReserveUsedThresholdPct'
    )
    stage2_aware_threshold = types.IntType(
        required=True,
        serialized_name='stage2AwareThreshold'
    )
    stage3_low_threshold = types.IntType(
        required=True,
        serialized_name='stage3LowThreshold'
    )
    stage4_critical_threshold = types.IntType(
        required=True,
        serialized_name='stage4CriticalThreshold'
    )
    sum_total_cluster_bytes = types.IntType(
        required=True,
        serialized_name='sumTotalClusterBytes'
    )
    sum_used_cluster_bytes = types.IntType(
        required=True,
        serialized_name='sumUsedClusterBytes'
    )

    # Version 6.0 Additions
    block_fullness = types.EnumType(
        enum_class=enums.ClusterStage,
        required=True,
        serialized_name='blockFullness',
        version_added=sf_version.SfVersion('6.0')
    )
    max_metadata_over_provision_factor = types.IntType(
        required=True,
        serialized_name='maxMetadataOverProvisionFactor',
        version_added=sf_version.SfVersion('6.0')
    )
    metadata_fullness = types.EnumType(
        enum_class=enums.ClusterStage,
        required=True,
        serialized_name='metadataFullness',
        version_added=sf_version.SfVersion('6.0')
    )
    sum_total_metadata_cluster_bytes = types.IntType(
        required=True,
        serialized_name='sumTotalMetadataClusterBytes',
        version_added=sf_version.SfVersion('6.0')
    )
    sum_used_metadata_cluster_bytes = types.IntType(
        required=True,
        serialized_name='sumUsedMetadataClusterBytes',
        version_added=sf_version.SfVersion('6.0')
    )

    # Version 8.0 Additions
    stage2_block_threshold_bytes = types.IntType(
        required=True,
        serialized_name='stage2BlockThresholdBytes',
        version_added=sf_version.SfVersion('8.0')
    )
    stage3_block_threshold_bytes = types.IntType(
        required=True,
        serialized_name='stage3BlockThresholdBytes',
        version_added=sf_version.SfVersion('8.0')
    )
    stage3_block_threshold_percent = types.IntType(
        required=True,
        serialized_name='stage3BlockThresholdPercent',
        version_added=sf_version.SfVersion('8.0')
    )
    stage4_block_threshold_bytes = types.IntType(
        required=True,
        serialized_name='stage4BlockThresholdBytes',
        version_added=sf_version.SfVersion('8.0')
    )
    stage5_block_threshold_bytes = types.IntType(
        required=True,
        serialized_name='stage5BlockThresholdBytes',
        version_added=sf_version.SfVersion('8.0')
    )

    # Version 12.0 Additions
    stage2_metadata_threshold_bytes = types.IntType(
        required=False,
        serialized_name='stage2MetadataThresholdBytes',
        version_added=sf_version.SfVersion('12.0')
    )
    stage3_metadata_threshold_bytes = types.IntType(
        required=False,
        serialized_name='stage3MetadataThresholdBytes',
        version_added=sf_version.SfVersion('12.0')
    )
    stage3_metadata_threshold_percent = types.IntType(
        required=False,
        serialized_name='stage3MetadataThresholdPercent',
        version_added=sf_version.SfVersion('12.0')
    )
    stage4_metadata_threshold_bytes = types.IntType(
        required=False,
        serialized_name='stage4MetadataThresholdBytes',
        version_added=sf_version.SfVersion('12.0')
    )
    stage5_metadata_threshold_bytes = types.IntType(
        required=False,
        serialized_name='stage5MetadataThresholdBytes',
        version_added=sf_version.SfVersion('12.0')
    )


class ClusterLimits(sf_models.Base):
    account_count_max = types.IntType(required=True, serialized_name='accountCountMax')
    account_name_length_max = types.IntType(required=True, serialized_name='accountNameLengthMax')
    account_name_length_min = types.IntType(required=True, serialized_name='accountNameLengthMin')
    bulk_volume_jobs_per_node_max = types.IntType(
        required=True,
        serialized_name='bulkVolumeJobsPerNodeMax'
    )
    bulk_volume_jobs_per_volume_max = types.IntType(
        required=True,
        serialized_name='bulkVolumeJobsPerVolumeMax'
    )
    clone_jobs_per_volume_max = types.IntType(required=True, serialized_name='cloneJobsPerVolumeMax')
    cluster_admin_account_max = types.IntType(required=True, serialized_name='clusterAdminAccountMax')
    cluster_pairs_count_max = types.IntType(required=True, serialized_name='clusterPairsCountMax')
    fibre_channel_volume_access_max = types.IntType(
        required=True,
        serialized_name='fibreChannelVolumeAccessMax'
    )
    initiator_alias_length_max = types.IntType(required=True, serialized_name='initiatorAliasLengthMax')
    initiator_count_max = types.IntType(required=True, serialized_name='initiatorCountMax')
    initiator_name_length_max = types.IntType(required=True, serialized_name='initiatorNameLengthMax')
    initiators_per_volume_access_group_count_max = types.IntType(
        required=True,
        serialized_name='initiatorsPerVolumeAccessGroupCountMax'
    )
    iscsi_sessions_from_fibre_channel_nodes_max = types.IntType(
        required=True,
        serialized_name='iscsiSessionsFromFibreChannelNodesMax'
    )
    secret_length_max = types.IntType(required=True, serialized_name='secretLengthMax')
    secret_length_min = types.IntType(required=True, serialized_name='secretLengthMin')
    snapshot_name_length_max = types.IntType(required=True, serialized_name='snapshotNameLengthMax')
    snapshots_per_volume_max = types.IntType(required=True, serialized_name='snapshotsPerVolumeMax')
    virtual_volume_count_max = types.IntType(required=True, serialized_name='virtualVolumeCountMax')
    virtual_volumes_per_account_count_max = types.IntType(
        required=True,
        serialized_name='virtualVolumesPerAccountCountMax'
    )
    volume_access_group_count_max = types.IntType(
        required=True,
        serialized_name='volumeAccessGroupCountMax'
    )
    volume_access_group_lun_max = types.IntType(required=True, serialized_name='volumeAccessGroupLunMax')
    volume_access_group_name_length_max = types.IntType(
        required=True,
        serialized_name='volumeAccessGroupNameLengthMax'
    )
    volume_access_group_name_length_min = types.IntType(
        required=True,
        serialized_name='volumeAccessGroupNameLengthMin'
    )
    volume_access_groups_per_initiator_count_max = types.IntType(
        required=True,
        serialized_name='volumeAccessGroupsPerInitiatorCountMax'
    )
    volume_access_groups_per_volume_count_max = types.IntType(
        required=True,
        serialized_name='volumeAccessGroupsPerVolumeCountMax'
    )
    volume_burst_iops_max = types.IntType(required=True, serialized_name='volumeBurstIOPSMax')
    volume_burst_iops_min = types.IntType(required=True, serialized_name='volumeBurstIOPSMin')
    volume_count_max = types.IntType(required=True, serialized_name='volumeCountMax')
    volume_max_iops_max = types.IntType(required=True, serialized_name='volumeMaxIOPSMax')
    volume_max_iops_min = types.IntType(required=True, serialized_name='volumeMaxIOPSMin')
    volume_min_iops_max = types.IntType(required=True, serialized_name='volumeMinIOPSMax')
    volume_min_iops_min = types.IntType(required=True, serialized_name='volumeMinIOPSMin')
    volume_name_length_max = types.IntType(required=True, serialized_name='volumeNameLengthMax')
    volume_name_length_min = types.IntType(required=True, serialized_name='volumeNameLengthMin')
    volume_size_max = types.IntType(required=True, serialized_name='volumeSizeMax')
    volume_size_min = types.IntType(required=True, serialized_name='volumeSizeMin')
    volumes_per_account_count_max = types.IntType(
        required=True,
        serialized_name='volumesPerAccountCountMax'
    )
    volumes_per_group_snapshot_max = types.IntType(
        required=True,
        serialized_name='volumesPerGroupSnapshotMax'
    )
    volumes_per_volume_access_group_count_max = types.IntType(
        required=True,
        serialized_name='volumesPerVolumeAccessGroupCountMax'
    )


class NtpInfo(sf_models.Base):
    servers = types.ListType(types.StringType(), required=True)
    broadcast_client = types.BooleanType(required=True, serialized_name='broadcastclient')


class SnmpInfo(sf_models.Base):
    networks = types.ListType(types.ModelType(_general.SnmpNetworkInfo, required=False), required=True)
    enabled = types.BooleanType(required=True)
    snvmp_v3_enabled = types.BooleanType(required=True, serialized_name='snmpV3Enabled')
    usm_users = types.ListType(types.ModelType(_general.SnmpUserInfo), required=False,
                               serialized_name='usmUsers')


class CiphersInfo(sf_models.Base):
    mandatory_ciphers = types.ListType(
        field=types.StringType(),
        serialized_name='mandatoryCiphers',
        version_added=sf_version.SfVersion('11.3'),
        required=True)
    supplemental_ciphers = types.ListType(
        field=types.StringType(),
        serialized_name='supplementalCiphers',
        version_added=sf_version.SfVersion('11.3'),
        required=True)


class ClusterStructure(sf_models.Base):
    """ClusterStructure data model."""
    accounts = types.ListType(types.ModelType(_account.Account), required=True)
    cluster_admins = types.ListType(types.ModelType(ClusterAdmin), required=True,
                                    serialized_name='clusterAdmins')
    cluster_info = types.ModelType(Cluster, required=True, serialized_name='clusterInfo')
    default_qos = types.ModelType(_volume.Qos, required=True, serialized_name='defaultQoS')
    features = types.ListType(types.ModelType(_general.FeatureStatus), required=True, )
    initiators = types.ListType(types.ModelType(_volume.Initiator), required=True, )
    ldap_configuration = types.ModelType(_general.LdapConfig, required=True,
                                         serialized_name='ldapConfiguration')
    ntp = types.ModelType(NtpInfo, required=True)
    qos_policies = types.ListType(types.ModelType(_volume.QosPolicy), required=True,
                                  serialized_name='qosPolicies')
    remote_hosts = types.ListType(types.ModelType(_general.LoggingServer), required=True,
                                  serialized_name='remoteHosts')
    snapmirror_endpoints = types.ListType(types.ModelType(_snapmirror.SnapMirrorEndpoint),
                                          required=False, serialized_name='snapMirrorEndpoints')
    storage_containers = types.ListType(types.ModelType(_node.StorageInfo), required=False,
                                        serialized_name='storageContainers')
    schedules = types.ListType(types.ModelType(_schedule.Schedule), required=True)
    snmp = types.ModelType(SnmpInfo, required=True)
    tls_ciphers = types.ModelType(CiphersInfo, required=True, serialized_name='tlsCiphers',
                                  version_added=sf_version.SfVersion('11.3'))
    virtual_networks = types.ListType(types.ModelType(_network.VirtualNetwork), required=True,
                                      serialized_name='virtualNetworks')
    vag_lun_assignments = types.ListType(types.ModelType(_volume.VagLunAssignments), required=True,
                                         serialized_name='volumeAccessGroupLunAssignments')
    vags = types.ListType(types.ModelType(_volume.VolumeAccessGroup), required=True,
                          serialized_name='volumeAccessGroups')
    volumes = types.ListType(types.ModelType(_volume.Volume), required=True)


class PerServiceStatus(sf_models.Base):
    """PerServiceStatus data model."""
    success = types.DictType(types.BooleanType())
    failed = types.DictType(types.StringType())


class ClusterBlockInfo(sf_models.Base):
    block_data = types.StringType(required=True, serialized_name='blockData')
    service_id = types.IntIdType(serialized_name=api_constants.SERVICE_ID, required=True)
    status = types.EnumType(enums.BinStatus, required=True)
    replica = types.IntIdType(required=True, version_added=sf_version.SfVersion('12.0'))


class ClusterBlock(sf_models.Base):
    """ClusterBlock data model."""
    bin = types.IntType(required=True)
    hash = types.StringType(required=True)
    services = types.ListType(types.ModelType(ClusterBlockInfo), required=True)


class ZoneLayout(sf_models.Base):
    """Zone Layout data model."""
    node_ids = types.ListType(types.IntType(required=True), required=True, serialized_name=u'nodeIDs')
    zone_name = types.StringType(required=True, serialized_name=u'zoneName')
    zone_level = types.StringType(required=True, serialized_name=u'zoneLevel')

    def __repr__(self):
        return '{}<{}={}, {}={}><{}={}>'.format(
            self.__class__.__name__,
            'nodeIDs', self.node_ids,
            'zoneName', self.zone_name,
            'zoneLevel', self.zone_level
        )


class FipsNodeReport(sf_models.Base):
    """FipsReport data model.

    Corresponds to the elements of the nodes object in the GetFipsReport response.
    """
    node_id = types.IntIdType(required=True, serialized_name='nodeID')
    https_enabled = types.BooleanType(required=True, serialized_name='httpsEnabled')
    fips_drives = types.EnumType(
        enums.FipsDrivesStatus,
        required=True,
        serialized_name='fipsDrives',
        version_added=sf_version.SfVersion('11.5')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'https_enabled', self.https_enabled,
            'fips_drives', self.fips_drives
        )


class FipsErrorNodeReport(sf_models.Base):
    """FipsReport data model.

    Corresponds to the elements of the error nodes object in the GetFipsReport response.
    """
    error = types.StringType(required=True)
    node_id = types.IntIdType(required=True, serialized_name='nodeID')

    def __repr__(self):
        return '{}<{}={}, {}={}'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'error', self.error
        )


class RawStats(sf_models.Base):
    """RawStats data model.

    Corresponds to the elements in the GetRawStats response.
    """
    def __repr__(self):
        return '{}<>'.format(
            self.__class__.__name__,
        )
