"""Drive Data Models."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import (
    _general,
    types,
)


class ClusterServiceDrive(sf_models.Base):
    """ClusterServiceDrive data model from ListServices API call.

    Corresponds to the elements of the drives object in the ClusterServiceInfo data model.
    Part of the ListServices response. This is called Drive in the behavior spec.
    """
    assigned_service = types.IntType(serialized_name='assignedService')
    async_result_ids = types.ListType(types.IntIdType(), required=True, serialized_name='asyncResultIDs')
    attributes = types.DictType(types.StringType(), required=True, default=dict)
    capacity = types.IntType(required=True)
    customer_slice_file_capacity = types.IntType(serialized_name='customerSliceFileCapacity')
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    drive_failure_detail = types.StringType(required=True, serialized_name='driveFailureDetail',
                                            version_added=sf_version.SfVersion('11.7'))
    drive_security_fault_reason = types.StringType(
        required=True,
        serialized_name='driveSecurityFaultReason',
        version_added=sf_version.SfVersion('11.7'))
    drive_status = types.EnumType(enums.DriveStatus, required=True, serialized_name='driveStatus')
    drive_type = types.EnumType(enums.DriveType, required=True, serialized_name='driveType',
                                lowercase=True)
    key_id = types.UUIDType(required=True, serialized_name='keyID',
                            version_added=sf_version.SfVersion('11.7'))
    key_provider_id = types.IntType(required=True, serialized_name='keyProviderID',
                                    version_added=sf_version.SfVersion('11.7'))
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    reserved_slice_file_capacity = types.IntType(serialized_name='reservedSliceFileCapacity')
    serial = types.StringType(required=True)
    slot = types.IntIdType()
    smart_ssd_write_capable = types.BooleanType(serialized_name='smartSsdWriteCapable')

    # Version 10.0 Additions
    fail_count = types.IntType(serialized_name='failCount', version_added=sf_version.SfVersion('10.0'))

    # Version 11.0 Additions
    usable_capacity = types.IntType(
        required=True,
        serialized_name='usableCapacity',
        version_added=sf_version.SfVersion('11.0')
    )
    segment_file_size = types.IntType(
        required=True,
        serialized_name='segmentFileSize',
        version_added=sf_version.SfVersion('11.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive_id', self.drive_id,
            'node_id', self.node_id,
            'slot', self.slot,
            'status', self.drive_status,
            'type', self.drive_type
        )


class Drive(sf_models.Base):
    """Drive data model.

    Corresponds to the elements of the drives object in the ListDrives response. This is called
    DriveInfo in the behavior spec.
    """
    attributes = types.DictType(types.BaseType(), default=dict)
    capacity = types.IntType(required=True)
    drive_failure_detail = types.StringType(
        required=True,
        serialized_name='driveFailureDetail',
        version_added=sf_version.SfVersion('11.7'))
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    drive_security_fault_reason = types.StringType(
        required=True,
        serialized_name='driveSecurityFaultReason',
        version_added=sf_version.SfVersion('11.7'))
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    key_id = types.UUIDType(
        required=True,
        serialized_name='keyID',
        version_added=sf_version.SfVersion('11.7'))
    key_provider_id = types.IntType(
        required=True,
        serialized_name='keyProviderID',
        version_added=sf_version.SfVersion('11.7'))
    serial = types.StringType(required=True)
    slot = types.IntIdType(required=True)
    status = types.EnumType(enums.DriveStatus, required=True)
    type = types.EnumType(enums.DriveType, required=True)

    # Version 10.0 Additions
    chassis_slot = types.StringType(
        required=True,
        regex=r'^[A-D]?-?[0-9]{1,2}$',
        serialized_name='chassisSlot',
        version_added=sf_version.SfVersion('10.0')
    )

    # Version 11.0 Additions
    segment_file_size = types.DataSizeType(
        required=True,
        serialized_name=api_constants.SEGMENT_FILE_SIZE,
        version_added=sf_version.SfVersion('11.0')
    )
    internal_usable_capacity = types.DataSizeType(
        required=True,
        serialized_name=api_constants.USABLE_CAPACITY,
        version_added=sf_version.SfVersion('11.0')
    )

    @property
    def usable_capacity(self):
        """Usable capacity of the drive.

        Backwards compatible to older versions without the usable_capacity field.

        :rtype: valence.models.DataSize
        """
        if self.internal_usable_capacity:
            return self.internal_usable_capacity
        elif self.capacity:
            return sf_models.DataSize(self.capacity)
        else:
            return None

    @usable_capacity.setter
    def usable_capacity(self, capacity):
        """Set the usable capacity of the drive.

        :param valence.models.DataSize capacity: Drive capacity.
        """
        self.internal_usable_capacity = capacity

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive_id', self.drive_id,
            'node_id', self.node_id,
            'slot', self.slot,
            'status', self.status,
            'type', self.type
        )


class DriveError(sf_models.Base):
    """DriveError data model.

    Corresponds to the elements of the errors object in the ListDriveStats response.
    """
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    exception = types.ModelType(_general.SolidFireException, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive_id', self.drive_id,
            'exception', self.exception
        )


class DriveHardwareInfo(sf_models.Base):
    """DriveHardwareInfo data model.

    Corresponds to the drive_hardware_info object in the GetDriveHardwareInfo response.
    """
    description = types.StringType(required=True)
    dev = types.StringType(required=True, regex=r'^(\d+)?(:\d+)?$')
    devpath = types.StringType(required=True)
    drive_security_at_maximum = types.BooleanType(
        required=True,
        serialized_name='driveSecurityAtMaximum'
    )
    drive_security_frozen = types.BooleanType(required=True, serialized_name='driveSecurityFrozen')
    drive_security_locked = types.BooleanType(required=True, serialized_name='driveSecurityLocked')
    logicalname = types.StringType(required=True)
    product = types.StringType(required=True)
    scsi_compat_id = types.StringType(required=True, serialized_name='scsiCompatID')
    security_feature_enabled = types.BooleanType(required=True, serialized_name='securityFeatureEnabled')
    security_feature_supported = types.BooleanType(
        required=True,
        serialized_name='securityFeatureSupported'
    )
    serial = types.StringType(required=True)
    size = types.IntType(required=True)
    uuid = types.StringType(required=True, regex=r'^[0-9a-z-]+$')
    version = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'logicalname', self.logicalname,
            'description', self.description
        )


class DriveInfo(sf_models.Base):
    """DriveInfo data model.

    Corresponds to the elements of the drives object in the DriveConfig data model.
    Part of the GetDriveConfig response. This is called DriveHardware in the behavior spec.
    """
    canonical_name = types.StringType(required=True, serialized_name='canonicalName')
    connected = types.BooleanType(required=True)
    dev = types.IntType(required=True)
    dev_path = types.StringType(required=True, serialized_name='devPath')
    drive_encryption_capability = types.EnumType(
        enums.DriveEncryptionCapability,
        required=False,
        serialized_name='driveEncryptionCapability',
        version_added=sf_version.SfVersion('11.5')
    )
    drive_type = types.EnumType(enums.DriveType, required=True, lowercase=True,
                                serialized_name='driveType')
    fs_type = types.StringType(serialized_name='fsType')
    is_mounted = types.BooleanType(serialized_name='isMounted')
    life_remaining_percent = types.IntType(required=True, serialized_name='lifeRemainingPercent')
    lifetime_read_bytes = types.IntType(required=True, serialized_name='lifetimeReadBytes')
    lifetime_write_bytes = types.IntType(required=True, serialized_name='lifetimeWriteBytes')
    mount_point = types.StringType(serialized_name='mountPoint')
    name = types.StringType(required=True)
    path = types.StringType(required=True)
    path_link = types.StringType(required=True, serialized_name='pathLink')
    power_on_hours = types.IntType(required=True, serialized_name='powerOnHours')
    product = types.StringType(required=True)
    reallocated_sectors = types.IntType(required=True, serialized_name='reallocatedSectors')
    reserve_capacity_percent = types.IntType(required=True, serialized_name='reserveCapacityPercent')
    scsi_compat_id = types.StringType(required=True, serialized_name='scsiCompatId')
    scsi_state = types.EnumType(enums.ScsiState, required=True, serialized_name='scsiState')
    security_at_maximum = types.BooleanType(required=True, serialized_name='securityAtMaximum')
    security_enabled = types.BooleanType(required=True, serialized_name='securityEnabled')
    security_frozen = types.BooleanType(required=True, serialized_name='securityFrozen')
    security_locked = types.BooleanType(required=True, serialized_name='securityLocked')
    security_supported = types.BooleanType(required=True, serialized_name='securitySupported')
    serial = types.StringType(required=True)
    size = types.IntType(required=True)
    slot = types.IntType(required=True)
    uncorrectable_errors = types.IntType(serialized_name='uncorrectableErrors')
    uuid = types.UUIDType(required=True)
    vendor = types.StringType(required=True)
    version = types.StringType(required=True, regex=r'^[0-9a-zA-Z]+$')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive_type', self.drive_type,
            'slot', self.slot,
            'mount_point', self.mount_point
        )


class DriveStats(sf_models.Base):
    """DriveStats data model.

    NOTE: Slice drives do not populate the following fields:
    - read_bytes
    - read_ops
    - used_capacity
    - write_bytes
    - write_ops

    Corresponds to the drive_stats object in the GetDriveStats response.
    """
    active_sessions = types.IntType(serialized_name=api_constants.ACTIVE_SESSIONS)
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    failed_die_count = types.IntType(required=True, serialized_name='failedDieCount')
    life_remaining_percent = types.IntType(required=True, serialized_name='lifeRemainingPercent')
    lifetime_read_bytes = types.IntType(required=True, serialized_name='lifetimeReadBytes')
    lifetime_write_bytes = types.IntType(required=True, serialized_name='lifetimeWriteBytes')
    power_on_hours = types.IntType(required=True, serialized_name='powerOnHours')
    read_bytes = types.IntType(serialized_name='readBytes')
    read_ops = types.IntType(serialized_name='readOps')
    reallocated_sectors = types.IntType(required=True, serialized_name='reallocatedSectors')
    reserve_capacity_percent = types.IntType(required=True, serialized_name='reserveCapacityPercent')
    timestamp = types.DateTimeType(required=True)
    total_capacity = types.IntType(required=True, serialized_name='totalCapacity')
    used_capacity = types.IntType(serialized_name='usedCapacity')
    used_memory = types.IntType(required=True, serialized_name='usedMemory')
    write_bytes = types.IntType(serialized_name='writeBytes')
    write_ops = types.IntType(serialized_name='writeOps')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive_id', self.drive_id,
            'used_capacity', self.used_capacity,
            'total_capacity', self.total_capacity,
            'life_remaining_percent', self.life_remaining_percent
        )


class DriveConfig(sf_models.Base):
    """DriveConfig data model.

    Corresponds to the drive_config object in the NodeDriveConfigResult data model.
    Part of the GetDriveConfig response.
    """
    drives = types.ListType(types.ModelType(DriveInfo), required=True)
    num_block_actual = types.IntType(required=True, serialized_name='numBlockActual')
    num_block_expected = types.IntType(required=True, serialized_name='numBlockExpected')
    num_slice_actual = types.IntType(required=True, serialized_name='numSliceActual')
    num_slice_expected = types.IntType(required=True, serialized_name='numSliceExpected')
    num_total_actual = types.IntType(required=True, serialized_name='numTotalActual')
    num_total_expected = types.IntType(required=True, serialized_name='numTotalExpected')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'num_total_expected', self.num_total_expected,
            'num_total_actual', self.num_total_actual
        )


class NodeDriveConfigResult(sf_models.Base):
    """NodeDriveConfigResult data model.

    Corresponds to the result object in the NodeDriveConfig data model.
    Part of the GetDriveConfig response.
    """
    drive_config = types.ModelType(DriveConfig, required=True, serialized_name='driveConfig')

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'drive_config', self.drive_config)


class NodeDriveConfig(sf_models.Base):
    """NodeDriveConfig data model.

    Corresponds to the elements of the nodes object in the GetDriveConfig response.
    """
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    result = types.ModelType(NodeDriveConfigResult, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'result', self.result
        )


class NodeDriveHardwareResult(sf_models.Base):
    """NodeDriveHardwareResult data model.

    Corresponds to the result object in the NodeDriveHardware data model.
    Part of the ListDriveHardware response.
    """
    drive_hardware = types.ListType(
        field=types.ModelType(DriveInfo),
        required=True,
        serialized_name='driveHardware'
    )

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'drive_hardware', self.drive_hardware)


class NodeDriveHardware(sf_models.Base):
    """NodeDriveHardware data model.

    Corresponds to the elements of the nodes object in the ListDriveHardware response.
    """
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    result = types.ModelType(NodeDriveHardwareResult, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'result', self.result
        )


class ResetDriveStatus(sf_models.Base):
    """ResetDriveStatus data model.

    Corresponds to the elements of the drives object in the ResetDriveList data model.
    Part of the ResetDrives response.
    """
    drive = types.StringType(required=True)
    return_code = types.IntType(required=True, serialized_name='returnCode')
    stderr = types.StringType(required=True)
    stdout = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'drive', self.drive,
            'stdout', self.stdout,
            'stderr', self.stderr
        )


class ResetDriveList(sf_models.Base):
    """ResetDriveList data model.

    Corresponds to the details object in the ResetDrives response.
    """
    drives = types.ListType(types.ModelType(ResetDriveStatus), required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'drives', self.drives)
