"""Fibre Channel Data Models."""
import sf_models

from sf_api import (
    api_constants,
    enums,
)
from sf_api.models import types


class FibreChannelPort(sf_models.Base):
    """FibreChannelPort data model.

    Corresponds to the elements of the fibre_channel_ports object in the ListNodeFibreChannelPortInfo
    response.
    """
    firmware = types.StringType(required=True)
    hba_port = types.IntType(required=True, serialized_name='hbaPort')
    model = types.StringType(required=True, regex=r'^[0-9A-Z]+$')
    n_port_id = types.StringType(required=True, regex=r'^0x[0-9a-z]+$', serialized_name='nPortID')
    pci_slot = types.IntType(required=True, serialized_name='pciSlot')
    serial = types.StringType(required=True, regex=r'^[0-9A-Z]+$')
    speed = types.StringType(required=True, regex=r'^\d+ Gbit$')
    state = types.EnumType(enums.FibreChannelPortState, required=True)
    switch_wwn = types.StringType(
        required=True,
        regex=api_constants.REGEX_WORLD_WIDE_NAME,
        serialized_name='switchWwn'
    )
    wwnn = types.StringType(required=True, regex=api_constants.REGEX_WORLD_WIDE_NAME)
    wwpn = types.StringType(required=True, regex=api_constants.REGEX_WORLD_WIDE_NAME)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'n_port_id', self.n_port_id,
            'switch_wwn', self.switch_wwn
        )


class FibreChannelPortInfo(sf_models.Base):
    """FibreChannelPortInfo data model.

    Corresponds to the elements of the fibre_channel_port_info object in the ListFibreChannelPortInfo
    response.
    """
    fibre_channel_ports = types.ListType(
        field=types.ModelType(FibreChannelPort),
        required=True,
        serialized_name='fibreChannelPorts'
    )

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'fibre_channel_ports', self.fibre_channel_ports
        )


class NodeFibreChannelPortInfo(sf_models.Base):
    """Data model for a node entry returned by the cluster.

    Corresponds to the elements of the nodes object in the ListNodeFibreChannelPortInfo response.
    Data format is two dictionary entries nodeID and result. The result contains the data defined by the
    class FibreChannelPortInfo below.
    """
    node_id = types.IntType(required=True, serialized_name=api_constants.NODE_ID)
    result = types.ModelType(
        model_spec=FibreChannelPortInfo,
        required=True,
        serialized_name=api_constants.RESULT
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'result', self.result
        )


class FibreChannelSession(sf_models.Base):
    """FibreChannelSession data model.

    Corresponds to the elements of the sessions object in the ListFibreChannelSessions response.
    """
    initiator_wwpn = types.StringType(
        required=True,
        regex=api_constants.REGEX_WORLD_WIDE_NAME,
        serialized_name='initiatorWWPN'
    )
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    service_id = types.IntIdType(required=True, serialized_name=api_constants.SERVICE_ID)
    target_wwpn = types.StringType(
        required=True,
        regex=api_constants.REGEX_WORLD_WIDE_NAME,
        serialized_name='targetWWPN'
    )
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_access_group_id', self.volume_access_group_id,
            'initiator_wwpn', self.initiator_wwpn
        )
