"""General Data Models."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import types


class FeatureStatus(sf_models.Base):
    """FeatureStatus data model.

    Corresponds to the elements of the features object in the GetFeatureStatus response.
    """
    enabled = types.BooleanType(required=True)
    feature = types.EnumType(enums.SfFeature, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'feature', self.feature,
            'enabled', self.enabled
        )


class Histogram(sf_models.Base):
    """Histogram data model.

    Corresponds to the ssLoadHistogram object in the NodeStats data model.
    Part of the GetNodeStats response.
    """
    bucket_0 = types.IntType(serialized_name='Bucket0')
    bucket_1_to_19 = types.IntType(serialized_name='Bucket1To19')
    bucket_20_to_39 = types.IntType(serialized_name='Bucket20To39')
    bucket_40_to_59 = types.IntType(serialized_name='Bucket40To59')
    bucket_60_to_79 = types.IntType(serialized_name='Bucket60To79')
    bucket_80_to_100 = types.IntType(serialized_name='Bucket80To100')


class IscsiSession(sf_models.Base):
    """IscsiSession data model.

    Corresponds to the elements of the sessions object in the ListISCSISessions response.
    """
    account_id = types.IntIdType(required=True, serialized_name=api_constants.ACCOUNT_ID)
    account_name = types.StringType(required=True, serialized_name='accountName')
    drive_id = types.IntIdType(required=True, serialized_name=api_constants.DRIVE_ID)
    initiator_ip = types.StringType(
        required=True,
        regex=r'^\d+\.\d+\.\d+\.\d+\:\d+$',
        serialized_name='initiatorIP'
    )
    initiator_name = types.StringType(required=True, serialized_name='initiatorName')
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    service_id = types.IntIdType(required=True, serialized_name=api_constants.SERVICE_ID)
    session_id = types.IntIdType(required=True, serialized_name='sessionID')
    target_name = types.StringType(required=True, serialized_name='targetName')
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)

    # Version 7.0 Additions
    target_ip = types.StringType(
        required=True,
        regex=r'^\d+\.\d+\.\d+\.\d+\:\d+$',
        serialized_name='targetIP',
        version_added=sf_version.SfVersion('7.0')
    )
    virtual_network_id = types.IntIdType(
        required=True,
        serialized_name='virtualNetworkID',
        version_added=sf_version.SfVersion('7.0')
    )

    # Version 8.2 Additions
    create_time = types.DateTimeType(
        required=True,
        serialized_name='createTime',
        version_added=sf_version.SfVersion('8.2')
    )

    # Version 8.6 Additions
    drive_ids = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name=api_constants.DRIVE_IDS,
        version_added=sf_version.SfVersion('8.6')
    )
    initiator = types.StringType(
        version_added=sf_version.SfVersion('8.6')
    )
    initiator_port_name = types.StringType(
        required=True,
        serialized_name='initiatorPortName',
        version_added=sf_version.SfVersion('8.6')
    )
    initiator_session_id = types.IntIdType(
        required=True,
        serialized_name='initiatorSessionID',
        version_added=sf_version.SfVersion('8.6')
    )
    target_port_name = types.StringType(
        required=True,
        serialized_name='targetPortName',
        version_added=sf_version.SfVersion('8.6')
    )
    volume_instance = types.IntType(
        required=True,
        serialized_name='volumeInstance',
        version_added=sf_version.SfVersion('8.6')
    )

    # Version 10.0 Additions
    ms_since_last_iscsi_pdu = types.IntType(
        required=True,
        serialized_name='msSinceLastIscsiPDU',
        version_added=sf_version.SfVersion('10.0')
    )
    ms_since_last_scsi_command = types.IntType(
        required=True,
        serialized_name='msSinceLastScsiCommand',
        version_added=sf_version.SfVersion('10.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'session_id', self.session_id,
            'volume_id', self.volume_id,
            'target_ip', self.target_ip
        )


class LdapConfig(sf_models.Base):
    """LdapConfig data model.

    Corresponds to the ldap_configuration object in the GetLdapConfiguration response.
    """
    auth_type = types.EnumType(enums.LdapAuthType, required=True, serialized_name='authType')
    enabled = types.BooleanType(required=True)
    group_searchsf_models_dn = types.StringType(required=True, serialized_name='groupSearchBaseDN')
    group_search_custom_filter = types.StringType(
        required=True,
        serialized_name='groupSearchCustomFilter'
    )
    group_search_type = types.EnumType(
        enum_class=enums.LdapSearchType,
        required=True,
        serialized_name='groupSearchType'
    )
    search_bind_dn = types.StringType(required=True, serialized_name='searchBindDN')
    server_uris = types.ListType(
        field=types.StringType(regex=r'^ldaps?://\d+.\d+.\d+.\d+$'),
        required=True,
        serialized_name='serverURIs'
    )
    user_dn_template = types.StringType(required=True, serialized_name='userDNTemplate')
    user_searchsf_models_dn = types.StringType(required=True, serialized_name='userSearchBaseDN')
    user_search_filter = types.StringType(required=True, serialized_name='userSearchFilter')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'auth_type', self.auth_type,
            'enabled', self.enabled
        )


class LoggingServer(sf_models.Base):
    """LoggingServer data model.

    Corresponds to the elements of the remote_hosts object in the GetRemoteLoggingHosts response.
    """
    host = types.IPv4Type(required=True)
    port = types.IntType()

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'host', self.host,
            'port', self.port
        )


class LoginBanner(sf_models.Base):
    """LoginBanner data model.

    Corresponds to the login_banner object in the GetLoginBanner response.
    """
    banner = types.StringType(required=True)
    enabled = types.BooleanType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'banner', self.banner,
            'enabled', self.enabled
        )


class LoginSessionInfo(sf_models.Base):
    """LoginSessionInfo data model.

    Corresponds to the login_session_info object in the GetLoginSessionInfo response.
    """
    # from spec: All leading zeros and colons are removed regardless of the format the
    # timeout was entered.
    timeout = types.StringType(required=True, regex=r'^\d+(:[0-5]\d)*(\.\d+)?$')

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'timeout', self.timeout)


class PendingOperation(sf_models.Base):
    """PendingOperation data model.

    Corresponds to the pending_operation object in the GetPendingOperation response.
    """
    operation = types.StringType(required=True)
    pending = types.BooleanType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'operation', self.operation,
            'pending', self.pending
        )


class SnmpNetworkInfo(sf_models.Base):
    """SnmpNetworkInfo data model.

    Corresponds to the elements of the networks object in the GetSnmpInfo response.
    """
    access = types.StringType(required=True)
    cidr = types.IntType(required=True)
    community = types.EnumType(enums.SnmpCommunity, required=True)
    network = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'community', self.community,
            'network', self.network,
            'access', self.access
        )


class SnmpTrapRecipient(sf_models.Base):
    """SnmpTrapRecipient data model.

    Corresponds to the elements of the remote_hosts object in the GetRemoteLoggingHosts response.
    """
    community = types.EnumType(enums.SnmpCommunity, required=True)
    host = types.StringType(required=True)
    port = types.IntType()

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'community', self.community,
            'host', self.host,
            'port', self.port
        )


class SnmpUserInfo(sf_models.Base):
    """SnmpUserInfo data model.

    Corresponds to the elements of the usm_users object in the GetSnmpACL response.
    """
    access = types.StringType(required=True)
    name = types.StringType(required=True)
    passphrase = types.StringType(required=True)
    password = types.StringType(required=True)
    sec_level = types.StringType(required=True, serialized_name='secLevel')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'access', self.access
        )


class SoftwareVersionInfo(sf_models.Base):
    """SoftwareVersionInfo data model.

    Corresponds to the software_version_info object in the GetClusterVersionInfo response.
    """
    current_version = types.SfBuildVersionType(required=True, serialized_name='currentVersion')
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    package_name = types.StringType(required=True, serialized_name='packageName')
    pending_version = types.SfBuildVersionType(required=True, serialized_name='pendingVersion')
    start_time = types.StringType(required=True, serialized_name='startTime')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'current_version', self.current_version,
            'pending_version', self.pending_version
        )


class SolidFireException(sf_models.Base):
    """SolidFireException data model.

    Corresponds to the exception object in the DriveError data model.
    Part of the ListDriveStats response.
    """
    message = types.StringType(required=True)
    name = types.EnumType(enums.SfException, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'message', self.message
        )


class CertificateDetails(sf_models.Base):
    """Certificate Details data model."""
    issuer = types.StringType(required=True)
    modulus = types.StringType(required=True)
    not_after = types.DateTimeType(required=True, serialized_name='notAfter')
    not_before = types.DateTimeType(required=True, serialized_name='notBefore')
    serial = types.StringType(required=True)
    sha1_fingerprint = types.StringType(required=True, serialized_name='sha1Fingerprint')
    subject = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'mod', self.modulus)
