"""Network Data Model."""
import sf_models
from schematics import common

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import (
    _route,
    types,
)

# Backward Compatibility
VirtualNetwork = sf_models.VirtualNetwork
AddressBlocks = sf_models.AddressBlocks


class NetworkInterface(sf_models.Base):
    """NetworkInterface data model.

    Corresponds to the elements of the interfaces object in the List/Get NetworkInterfaces response.
    """
    address = types.IPv4Type(required=True)
    broadcast = types.IPv4Type(required=True)
    mac_address = types.StringType(required=True, serialized_name='macAddress')
    mtu = types.IntType()
    name = types.NetworkInterfaceNameType(required=True)
    namespace = types.BooleanType()
    netmask = types.StringType(required=True, regex=api_constants.REGEX_IPV4_OR_NA)
    status = types.EnumType(enums.NetworkStatus, required=True)
    type = types.EnumType(enums.NetworkType, required=True)
    virtual_network_tag = types.IntType(serialized_name='virtualNetworkTag')

    # 11.0 Added
    address_v6 = types.IPv6Type(version_added=sf_version.SfVersion('11.0'), serialized_name='addressV6')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'type', self.type,
            'address', self.address,
            'status', self.status
        )


class NodeHardwareInfoInterfaceStatus(sf_models.Base):
    """NodeHardwareInfoInterfaceStatus data model.

    Corresponds to the elements of the network_interfaces object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    is_configured = types.BooleanType(required=True, serialized_name='isConfigured')
    is_up = types.BooleanType(required=True, serialized_name='isUp')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'is_configured', self.is_configured,
            'is_up', self.is_up
        )


class PhysicalInterfaceDetails(sf_models.Base):
    """PhysicalInterfaceDetails data model.

    Corresponds to the physical object in the NetworkInterfaceDetails data model.
    Part of the GetNetworkConfig response.
    """
    address = types.StringType(required=True, regex=api_constants.REGEX_IPV4_OR_EMPTY)

    mac_address = types.StringType(
        required=True,
        regex=api_constants.REGEX_MAC_ADDRESS,
        serialized_name='macAddress'
    )
    mac_address_permanent = types.StringType(
        required=True,
        regex=api_constants.REGEX_MAC_ADDRESS,
        serialized_name='macAddressPermanent'
    )
    mtu = types.StringType(regex=r'^\d+$')
    netmask = types.StringType(required=True, regex=api_constants.REGEX_IPV4_OR_EMPTY)
    network = types.StringType(required=True, regex=api_constants.REGEX_IPV4_OR_EMPTY)
    up_and_running = types.BooleanType(required=True, serialized_name='upAndRunning')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'address', self.address,
            'up_and_running', self.up_and_running
        )


class NetworkInterfaceDetails(sf_models.Base):
    """NetworkInterfaceDetails data model.

    Corresponds to the elements of the NetworkConfig data model.
    Part of the GetNetworkConfig response.
    """
    address = types.IPAddressType()
    auto = types.BooleanType(required=True)
    bond_down_delay = types.StringType(serialized_name='bond-downdelay', regex=r'^\d+$')
    bond_fail_over_mac = types.EnumType(enums.BondFailOverMac, serialized_name='bond-fail_over_mac')
    bond_master = types.NetworkInterfaceNameType(serialized_name='bond-master')
    bond_miimon = types.StringType(serialized_name='bond-miimon', regex=r'^\d+$')
    bond_mode = types.EnumType(enums.BondMode, serialized_name='bond-mode')
    bond_primary_reselect = types.EnumType(
        enum_class=enums.BondReselectStatus,
        serialized_name='bond-primary_reselect'
    )
    bond_slaves = types.StringType(serialized_name='bond-slaves')
    bond_up_delay = types.StringType(serialized_name='bond-updelay', regex=r'^\d+$')
    default = types.BooleanType(serialized_name='#default')
    dns_name_servers = types.StringListType(types.IPAddressType, serialized_name='dns-nameservers')
    dns_search = types.StringType(serialized_name='dns-search')
    family = types.StringType(required=True)
    gateway = types.IPAddressType()
    mac_address = types.StringType(
        required=True,
        regex=api_constants.REGEX_MAC_ADDRESS,
        serialized_name='macAddress'
    )
    mac_address_permanent = types.StringType(
        required=True,
        regex=api_constants.REGEX_MAC_ADDRESS,
        serialized_name='macAddressPermanent'
    )
    method = types.StringType(required=True)
    mtu = types.StringType(regex=r'^\d+$')
    # Name is not actually present in the API response. This is a hack to model the outer structure.
    name = types.NetworkInterfaceNameType(serialize_when_none=False)
    # TODO The API will return either an empty string or an ipv6 address,
    # This breaks our framework so we don't support it and just match on any string.
    netmask = types.StringType()
    network = types.StringType()
    physical = types.ModelType(PhysicalInterfaceDetails, required=True)
    routes = types.ListType(types.ModelType(_route.Route))
    status = types.EnumType(enums.NetworkStatus, required=True)
    symmetric_route_rules = types.ListType(types.StringType(), serialized_name='symmetricRouteRules')
    up_and_running = types.BooleanType(required=True, serialized_name='upAndRunning')
    virtual_network_tag = types.StringType(serialized_name='virtualNetworkTag', regex=r'^\d+$')

    # 11.0 Added
    address_v6 = types.IPv6Type(version_added=sf_version.SfVersion('11.0'), serialized_name='addressV6')
    gateway_v6 = types.IPv6Type(version_added=sf_version.SfVersion('11.0'), serialized_name='gatewayV6')

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'address', self.address
        )

    class Options:
        export_level = common.NOT_NONE


class NetworkInterfaceDetailsExtended(sf_models.Base):
    """NetworkInterfaceDetailsExtended data model.

    Corresponds to the elements of the extended NetworkConfig data model.
    Part of the GetConfigExtended response.
    """
    address = types.IPAddressType()
    auto = types.BooleanType(required=False)
    base_interface_name = types.EnumType(enums.NetworkInterface, serialized_name='baseInterfaceName',
                                         version_added=sf_version.SfVersion('12.0'))
    bond_down_delay = types.StringType(serialized_name='bond-downdelay', regex=r'^\d+$')
    bond_fail_over_mac = types.EnumType(enums.BondFailOverMac, serialized_name='bond-fail_over_mac')
    bond_master = types.NetworkInterfaceNameType(serialized_name='bond-master')
    bond_miimon = types.StringType(serialized_name='bond-miimon', regex=r'^\d+$')
    bond_mode = types.EnumType(enums.BondMode, serialized_name='bond-mode')
    bond_primary_reselect = types.EnumType(
        enum_class=enums.BondReselectStatus,
        serialized_name='bond-primary_reselect'
    )
    bond_slaves = types.StringType(serialized_name='bond-slaves')
    bond_up_delay = types.StringType(serialized_name='bond-updelay', regex=r'^\d+$')
    default = types.BooleanType(serialized_name='#default')
    dns_name_servers = types.StringListType(types.IPAddressType, serialized_name='dns-nameservers')
    dns_search = types.StringType(serialized_name='dns-search')
    family = types.StringType(required=True)
    gateway = types.IPAddressType()
    mac_address = types.StringType(
        required=True,
        regex=api_constants.REGEX_MAC_ADDRESS,
        serialized_name='macAddress'
    )
    mac_address_permanent = types.StringType(
        required=True,
        regex=api_constants.REGEX_MAC_ADDRESS,
        serialized_name='macAddressPermanent'
    )
    method = types.StringType(required=True)
    mtu = types.IntType()
    # Name is not actually present in the API response. This is a hack to model the outer structure.
    name = types.NetworkInterfaceNameType(serialize_when_none=False, required=True)
    netmask = types.IPAddressType()
    network = types.IPAddressType()
    physical = types.ModelType(PhysicalInterfaceDetails, required=False)
    routes = types.ListType(types.ModelType(_route.Route))
    status = types.EnumType(enums.NetworkStatus, required=True)
    symmetric_route_rules = types.ListType(types.StringType(), serialized_name='symmetricRouteRules')
    up_and_running = types.BooleanType(required=False, serialized_name='upAndRunning')
    upper_interfaces = types.ListType(types.StringType(), version_added=sf_version.SfVersion('12.0'),
                                      serialized_name='upperInterfaces')
    virtual_network_tag = types.IntType(serialized_name='virtualNetworkTag')

    # 11.0 Added
    address_v6 = types.IPv6Type(version_added=sf_version.SfVersion('11.0'), serialized_name='addressV6')
    gateway_v6 = types.IPv6Type(version_added=sf_version.SfVersion('11.0'), serialized_name='gatewayV6')

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'address', self.address
        )

    class Options:
        serialize_when_none = False


class NetworkConfig(sf_models.Base):
    """NetworkConfig data model.

    Corresponds to the network object in the GetNetworkConfig response.
    """
    bond_1g = types.ModelType(
        model_spec=NetworkInterfaceDetails,
        serialized_name=api_constants.BOND_1G
    )
    bond_10g = types.ModelType(
        model_spec=NetworkInterfaceDetails,
        serialized_name=api_constants.BOND_10G
    )
    eth0 = types.ModelType(NetworkInterfaceDetails)
    eth1 = types.ModelType(NetworkInterfaceDetails)
    eth2 = types.ModelType(NetworkInterfaceDetails)
    eth3 = types.ModelType(NetworkInterfaceDetails)
    eth4 = types.ModelType(NetworkInterfaceDetails)
    eth5 = types.ModelType(NetworkInterfaceDetails)
    lo = types.ModelType(NetworkInterfaceDetails, required=True)
    net0 = types.ModelType(model_spec=NetworkInterfaceDetails)
    net1 = types.ModelType(model_spec=NetworkInterfaceDetails)

    def __repr__(self):
        if self.bond_1g:
            return '{}<{}={}, {}={}>'.format(
                self.__class__.__name__,
                'bond_1g', self.bond_1g,
                'bond_10g', self.bond_10g,
            )
        else:
            return '{}<{}={}, {}={}>'.format(
                self.__class__.__name__,
                'net0', self.net0,
                'net1', self.net1
            )

    class Options:
        export_level = common.NOT_NONE


class NetworkInterfaceStats(sf_models.Base):
    """NetworkInterfaceStats data model.

    Corresponds to the result object in the NodeNetworkInterfaceStats data model.
    Part of the GetNetworkInterfaceStats response.
    """
    collisions = types.DictType(types.BaseType(), required=True)
    rx_bytes = types.DictType(types.IntType(), required=True, serialized_name='rxBytes')
    rx_crc_errors = types.DictType(types.IntType(), required=True, serialized_name='rxCrcErrors')
    rx_dropped = types.DictType(types.IntType(), required=True, serialized_name='rxDropped')
    rx_errors = types.DictType(types.IntType(), required=True, serialized_name='rxErrors')
    rx_fifo_errors = types.DictType(types.IntType(), required=True, serialized_name='rxFifoErrors')
    rx_frame_errors = types.DictType(types.IntType(), required=True, serialized_name='rxFrameErrors')
    rx_length_errors = types.DictType(types.IntType(), required=True, serialized_name='rxLengthErrors')
    rx_missed_errors = types.DictType(types.IntType(), required=True, serialized_name='rxMissedErrors')
    rx_over_errors = types.DictType(types.IntType(), required=True, serialized_name='rxOverErrors')
    rx_packets = types.DictType(types.IntType(), required=True, serialized_name='rxPackets')
    tx_bytes = types.DictType(types.IntType(), required=True, serialized_name='txBytes')
    tx_carrier_errors = types.DictType(types.IntType(), required=True, serialized_name='txCarrierErrors')
    tx_dropped = types.DictType(types.IntType(), required=True, serialized_name='txDropped')
    tx_errors = types.DictType(types.IntType(), required=True, serialized_name='txErrors')
    tx_fifo_errors = types.DictType(types.IntType(), required=True, serialized_name='txFifoErrors')
    tx_packets = types.DictType(types.IntType(), required=True, serialized_name='txPackets')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'rx_packets', self.rx_packets,
            'rx_errors', self.rx_errors,
            'tx_packets', self.tx_packets,
            'tx_errors', self.tx_errors
        )


class NodeHardwareNetworkInterface(sf_models.Base):
    """NodeHardwareNetworkInterface data model.

    Corresponds to the elements of the networks object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    businfo = types.StringType()
    capacity = types.StringType(regex=r'^\d+$')
    clock = types.StringType(regex=r'^\d+$')
    description = types.StringType()
    logicalname = types.StringType()
    physid = types.StringType(required=True, regex=r'^[0-9a-z]+$|^[0-9a-z]+\.[0-9a-z]$')
    product = types.StringType()
    serial = types.StringType(regex=r'^[0-9a-zA-Z]{2}(:[0-9a-zA-Z]{2}){5}$')
    size = types.StringType(regex=r'^\d+$')
    vendor = types.StringType()
    version = types.StringType()
    width = types.StringType(regex=r'^\d+$')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'description', self.description,
            'logicalname', self.logicalname
        )


class PingResponse(sf_models.Base):
    """PingResponse data model.

    Corresponds to the elements of the ping_bytes object in the NodeMvipDetails, NodeSvipDetails
    data models.
    Part of the TestConnectMvip, TestConnectSvip responses.
    """
    individual_response_codes = types.ListType(
        field=types.EnumType(enums.ApiTestResult),
        required=True,
        serialized_name='individualResponseCodes'
    )
    individual_response_times = types.ListType(
        field=types.StringType(regex=api_constants.REGEX_TIMESTAMP),
        required=True,
        serialized_name='individualResponseTimes'
    )
    individual_status = types.ListType(
        field=types.BooleanType(),
        required=True,
        serialized_name='individualStatus'
    )
    response_time = types.StringType(
        required=True,
        regex=api_constants.REGEX_TIMESTAMP,
        serialized_name='responseTime'
    )
    successful = types.BooleanType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'response_time', self.response_time,
            'successful', self.successful
        )


class MinimalVirtualNetwork(sf_models.Base):
    """MinimalVirtualNetwork data model.

    Corresponds to the virtual networks listed on active nodes. This is different from the
    VirtualNetwork object in the SF API documentation (VirtualNetwork, above this class) :(
    """
    address = types.IPAddressType()
    virtual_network_id = types.IntIdType(serialized_name='virtualNetworkID')
