"""Node Data Models."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import _drive as drive_models
from sf_api.models import _general as general_models
from sf_api.models import _network as network_models
from sf_api.models import _nvram as nvram_models
from sf_api.models import types


class PlatformInfo(sf_models.Base):
    """PlatformInfo data model.

    Corresponds to the platform_info object in the ActiveNode, PendingActiveNode, PendingNode
    data models.
    Part of the ListActiveNodes, ListPendingActiveNodes, ListPendingNodes responses.
    """
    chassis_type = types.EnumType(
        enum_class=enums.ChassisType,
        required=True,
        serialized_name=api_constants.CHASSIS_TYPE
    )
    cpu_model = types.StringType(required=True, serialized_name=api_constants.CPU_MODEL)
    # TODO: use data size type for node_memory
    node_memory = types.IntType(required=True, serialized_name=api_constants.NODE_MEMORY_GB)
    node_type = types.NodeTypeType(required=True, serialized_name=api_constants.NODE_TYPE)

    # Version 9.1 Additions
    platform_config_version = types.StringType(
        required=True,
        serialized_name=api_constants.PLATFORM_CONFIG_VERSION,
        version_added=sf_version.SfVersion('9.1')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_type', self.node_type,
            'platform_config_version', self.platform_config_version
        )


class SfNodeBase(sf_models.Base, sf_models.ResourceInterface):
    """Common fields for SF nodes across Active, PendingActive, and Pending."""
    cip = types.IPAddressType(required=True)
    mip = types.IPAddressType(required=True)
    platform_info = types.ModelType(
        model_spec=PlatformInfo,
        required=True,
        serialized_name=api_constants.PLATFORM_INFO
    )
    sip = types.IPAddressType(required=True)
    software_version = types.SfBuildVersionType(
        required=True,
        serialized_name=api_constants.SOFTWARE_VERSION
    )
    uuid = types.UUIDType(required=True)

    # Version 12.0 Additions
    internal_role = types.EnumType(
        enum_class=enums.NodeRole,
        serialized_name='role',
        version_added=sf_version.SfVersion('12.0')
    )
    name = types.StringType(required=True)

    @property
    def static_id(self):
        """Unique identifier for the node.

        :rtype: str
        """
        return self.mip

    @static_id.setter
    def static_id(self, id_):
        """Set the unique identifier for the node.

        :param str id_: new id
        """
        self.mip = id_

    @property
    def role(self):
        """Role of the node.

        Backwards compatible to older versions without the role field.

        :rtype: sf_api.enums.NodeRole
        """
        # pylint: disable=no-member
        if self.platform_info:
            if sf_models.NodeType(self.platform_info.node_type.value).is_fc_node_model:
                return enums.NodeRole.fibre_channel
        if not self.internal_role:
            return enums.NodeRole.storage
        else:
            return self.internal_role

    @role.setter
    def role(self, role):
        """Set the role of the node.

        :param sf_api.enums.NodeRole role: Node role.
        """
        self.internal_role = role

    def is_witness(self):
        """Returns whether witness node or not.

        :rtype: bool
        """
        return self.internal_role is enums.NodeRole.witness


class ActiveNode(SfNodeBase):
    """ActiveNode data model.

    Corresponds to the nodes object in the ListActiveNodes response.
    """
    associated_fservice_id = types.IntType(
        required=True,
        serialized_name=api_constants.ASSOCIATED_FSERVICE_ID
    )
    associated_master_service_id = types.IntType(
        required=True,
        serialized_name=api_constants.ASSOCIATED_MASTER_SERVICE_ID
    )
    custom_protection_domain_name = types.StringType(
        version_added=sf_version.SfVersion('11.3'),
        serialized_name=api_constants.CUSTOM_PROTECTION_DOMAIN_NAME)
    attributes = types.DictType(types.StringType(), required=True, default=dict)
    chassis_name = types.StringType(version_added=sf_version.SfVersion('11.0'),
                                    serialized_name=api_constants.CHASSIS_NAME)
    chassis_name_override = types.StringType(required=False,
                                             serialized_name=api_constants.CHASSIS_NAME_OVERRIDE)
    cipi = types.StringType(required=True)
    fibre_channel_target_port_group = types.IntType(
        serialized_name=api_constants.FIBRE_CHANNEL_TARGET_PORT_GROUP
    )
    mip_v6 = types.IPAddressType(version_added=sf_version.SfVersion('11.0'), serialized_name='mipV6')
    mipi = types.StringType(required=True)
    name = types.StringType(required=True)
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    node_slot = types.StringType(version_added=sf_version.SfVersion('10.0'), serialized_name='nodeSlot')
    sipi = types.StringType(required=True)
    virtual_networks = types.ListType(
        field=types.ModelType(network_models.MinimalVirtualNetwork),
        required=True,
        serialized_name=api_constants.VIRTUAL_NETWORKS
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'name', self.name,
            'mip', self.mip,
            'sip', self.sip
        )


class AvailableNode(sf_models.Base):
    """AvailableNode data model.

    Corresponds to the nodes object in the ListAvailableNodes response.
    """
    cip = types.IPAddressType(required=True)
    compatible = types.BooleanType(required=True)
    hostname = types.StringType(required=True)
    mip = types.IPv4Type(required=True)
    mip_v6 = types.IPv6Type(version_added=sf_version.SfVersion('11.0'), serialized_name='mipV6')
    node_type = types.NodeTypeType(required=True, serialized_name='nodeType')

    # Version 12.0 Additions
    internal_role = types.EnumType(
        enum_class=enums.NodeRole,
        serialized_name='role',
        version_added=sf_version.SfVersion('12.0')
    )
    version = types.SfBuildVersionType(required=True)
    chassis_type = types.EnumType(enums.ChassisType, required=True, serialized_name='chassisType')

    @property
    def role(self):
        """Role of the node.

        Backwards compatible to older versions without the role field.

        :rtype: sf_api.enums.NodeRole
        """
        return self.internal_role or enums.NodeRole.storage


class CoreInfo(sf_models.Base):
    """CoreInfo data model.

    Corresponds to the elements of the bus object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    description = types.StringType(required=True)
    physid = types.StringType(required=True, regex=r'^\d+$')
    product = types.StringType(required=True)
    serial = types.StringType()
    vendor = types.StringType(required=True)
    version = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'physid', self.physid,
            'product', self.product,
            'version', self.version,
            'description', self.description
        )


class IpmiConfig(sf_models.Base):
    """IpmiConfig data model.

    Corresponds to the ipmi_config object in the GetIpmiConfig response.
    """
    sensor_name = types.StringType(required=True, serialized_name='sensorName')
    unique_sensor_id = types.StringType(required=True, serialized_name='uniqueSensorID')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'unique_sensor_id', self.unique_sensor_id,
            'sensor_name', self.sensor_name
        )


class IpmiInfo(sf_models.Base):
    """IpmiInfo data model.

    Corresponds to the ipmi_info object in the GetIpmiInfo response.
    """
    sensors = types.ListType(types.DictType(types.StringType), required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'sensors', self.sensors
        )


class NodeHistoryEntry(sf_models.Base):
    """NodeHistoryEntry data model.

    Corresponds to entries in GetHardwareConfig's nodeTypeHistoryList.
    """
    node_type = types.NodeTypeType(required=True, serialized_name='nodeType')
    revision = types.SfApiVersionType(required=True)


class NodeHardwareConfig(sf_models.Base):
    """NodeHardwareConfig data model.

    Corresponds to the elements of the nodes object in the GetHardwareConfig response.
    """
    bios_revision = types.StringType(regex=r'^\d+\.\d+$', serialized_name='biosRevision')
    bios_vendor = types.ListType(
        field=types.EnumType(enums.BiosVendor),
        serialized_name='biosVendor'
    )
    bios_version = types.StringType(serialized_name='biosVersion')
    block_drive_size_bytes = types.IntType(required=True, serialized_name='blockDriveSizeBytes')
    bmc_firmware_revision = types.StringType(serialized_name='bmcFirmwareRevision')
    bmc_ipmi_version = types.StringType(serialized_name='bmcIpmiVersion')
    chassis_type = types.EnumType(enums.ChassisType, required=True, serialized_name='chassisType')
    cpu_cores = types.IntType(required=True, serialized_name='cpuCores')
    cpu_cores_enabled = types.IntType(serialized_name='cpuCoresEnabled')
    cpu_model = types.StringType(required=True, serialized_name='cpuModel')
    cpu_threads = types.IntType(serialized_name='cpuThreads')
    drive_size_bytes_internal = types.IntType(serialized_name='driveSizeBytesInternal')
    fibre_channel_firmware_revision = types.StringType(
        required=True,
        serialized_name='fibreChannelFirmwareRevision'
    )
    fibre_channel_model = types.StringType(required=True, serialized_name='fibreChannelModel')
    idrac_version = types.StringType(serialized_name='idracVersion')
    lifecycle_version = types.StringType(required=False, serialized_name='lifecycleVersion')
    memory_gb = types.IntType(required=True, serialized_name='memoryGB')
    memory_mhz = types.IntType(serialized_name='memoryMhz')
    network_driver = types.ListType(types.StringType(), required=True, serialized_name='networkDriver')
    node_type = types.NodeTypeType(required=True, serialized_name='nodeType')
    node_type_history = types.ListType(types.ModelType(NodeHistoryEntry),
                                       serialized_name='nodeTypeHistory',
                                       version_added=sf_version.SfVersion('11.3'))
    num_cpu = types.IntType(required=True, serialized_name='numCpu')
    num_drives = types.IntType(required=True, serialized_name='numDrives')
    num_drives_internal = types.IntType(required=True, serialized_name='numDrivesInternal')
    nvram_vendor = types.EnumType(enums.NvramVendor, serialized_name='nvramVendor')
    nvids = types.ListType(types.StringType(), serialized_name='nvids')
    root_drive = types.StringType(required=True, serialized_name='rootDrive')
    scsi_bus_external_driver = types.StringType(required=True, serialized_name='scsiBusExternalDriver')
    scsi_bus_Internal_driver = types.StringType(required=True, serialized_name='scsiBusInternalDriver')
    slice_drive_size_bytes = types.IntType(required=True, serialized_name='sliceDriveSizeBytes')
    slot_offset = types.IntType(required=True, serialized_name='slotOffset')
    solidfire_defaults = types.DictType(types.IntType(), serialized_name='solidfireDefaults')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_type', self.node_type,
            'chassis_type', self.chassis_type
        )


class StorageInfo(sf_models.Base):
    """StorageInfo data model.

    Corresponds to the elements of the storage object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    businfo = types.StringType(required=True)
    clock = types.StringType(required=True, regex=r'^\d+$')
    description = types.StringType(required=True)
    physid = types.StringType(required=True)
    product = types.StringType()
    vendor = types.StringType(required=True)
    version = types.StringType(required=True, regex=r'^\d+$')
    width = types.StringType(required=True, regex=r'^\d+$')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'physid', self.physid,
            'product', self.product,
            'version', self.version,
            'description', self.description
        )


class SupportBundle(sf_models.Base):
    """SupportBundle data model.

    Corresponds to the elements of the details object in the CreateSupportBundle result data model.
    """
    bundle_name = types.StringType(required=True, serialized_name=api_constants.BUNDLE_NAME)
    extra_args = types.StringType(required=True, serialized_name=api_constants.EXTRA_ARGS)
    files = types.ListType(types.StringType(), required=True)
    output = types.StringType(required=True)
    timeout_sec = types.IntType(required=True, serialized_name=api_constants.TIMEOUT_SEC)
    url = types.ListType(types.StringType(), required=True)


class SystemInfo(sf_models.Base):
    """SystemInfo data model.

    Corresponds to the system object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    description = types.StringType(required=True)
    product = types.StringType(required=True)
    serial = types.StringType(required=True)
    vendor = types.StringType(required=True)
    width = types.StringType(required=True, regex=r'^\d+$')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'product', self.product,
            'description', self.description
        )


class SystemMemory(sf_models.Base):
    """SystemMemory data model.

    Corresponds to the system_memory object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    free = types.IntType(required=True)
    severity = types.EnumType(enums.SystemMemorySeverity)
    total = types.IntType(required=True)
    used = types.IntType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'total', self.total,
            'used', self.used,
            'severity', self.severity
        )


class NodeHardwareNvramRms(sf_models.Base):
    """NodeHardwareNvramRms data model.

    This is the model for the RMS-200 card type.

    Corresponds to the nvram object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    errors = types.ModelType(nvram_models.NvramErrors, required=True)
    extended = types.ModelType(nvram_models.NvramExtended, required=True)
    firmware = types.ModelType(nvram_models.NvramFirmware, required=True)
    identify = types.ModelType(nvram_models.NvramIdentify, required=True)
    smart = types.ModelType(nvram_models.NvramSmart, required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'identify', self.identify)


class NodeHardwareNvramOther(sf_models.Base):
    """NodeHardwareNvramOther data model.

    This model is used occasionally in virtual setups.

    Corresponds to the nvram object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    info = types.StringType(required=True)
    raw = types.StringType(required=True)
    version = types.StringType(required=True, regex=r'^\d+\.\d+$')
    status = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'status', self.status)


class NodeHardwareTest(sf_models.Base):
    """NodeHardwareTest data model.

    Corresponds to the elements of the details object in the TestHardwareConfig response.
    """
    actual = types.StringType(required=True)
    comparator = types.StringType(required=True)
    expected = types.StringType(required=True)
    passed = types.BooleanType(required=True, serialized_name='Passed')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'expected', self.expected,
            'actual', self.actual,
            'passed', self.passed
        )


class NodeHardwareInfo(sf_models.Base):
    """NodeHardwareInfo data model.

    Corresponds to the hardware_info object in the GetHardwareInfo response.
    """
    bus = types.DictType(types.ModelType(CoreInfo))
    drive_hardware = types.ListType(
        field=types.ModelType(drive_models.DriveInfo),
        required=True,
        serialized_name='driveHardware'
    )
    fibre_channel_ports = types.ListType(
        field=types.IntType(),
        required=True,
        serialized_name='fibreChannelPorts'
    )
    file_system_usage = types.DictType(
        field=types.StringType(),
        required=True,
        serialized_name='fileSystemUsage'
    )
    hardware_config = types.DictType(
        field=types.ModelType(NodeHardwareTest),
        serialized_name='hardwareConfig'
    )
    kernel_crash_dump_state = types.StringType(required=True, serialized_name='kernelCrashDumpState')
    memory = types.DictType(types.DictType(types.StringType()))
    network = types.DictType(types.ModelType(network_models.NodeHardwareNetworkInterface))
    network_interfaces = types.DictType(
        field=types.ModelType(network_models.NodeHardwareInfoInterfaceStatus),
        required=True,
        serialized_name='networkInterfaces'
    )
    nvram = types.PolyModelType([NodeHardwareNvramRms, NodeHardwareNvramOther])
    origin = types.BaseType()
    platform = types.ModelType(PlatformInfo, required=True)
    serial = types.StringType(required=True, regex=r'^[0-9a-zA-Z\-\s]*$')
    storage = types.DictType(types.ModelType(StorageInfo))
    system = types.DictType(types.ModelType(SystemInfo))
    system_memory = types.ModelType(SystemMemory, required=True, serialized_name='systemMemory')
    uuid = types.UUIDType(required=True)

    # Version 10.0 Additions
    board_serial = types.StringType(
        required=False,
        serialized_name='boardSerial',
        version_added=sf_version.SfVersion('10.0')
    )
    chassis_serial = types.StringType(
        required=True,
        serialized_name='chassisSerial',
        version_added=sf_version.SfVersion('10.0')
    )
    node_slot = types.StringType(
        regex=r'^[A-D]$',
        serialized_name='nodeSlot',
        version_added=sf_version.SfVersion('10.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'uuid', self.uuid,
            'drive_hardware', self.drive_hardware,
            'file_system_usage', self.file_system_usage
        )


class NodeMvipDetails(sf_models.Base):
    """NodeMvipDetails data model.

    Corresponds to the details object in the TestConnectMvip response.
    """
    connected = types.BooleanType(required=True)
    mvip = types.IPAddressType(required=True)
    ping_bytes = types.DictType(
        field=types.ModelType(network_models.PingResponse),
        required=True,
        serialized_name='pingBytes'
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'mvip', self.mvip,
            'connected', self.connected
        )


class NodeNetworkInterfaceStats(sf_models.Base):
    """NodeNetworkInterfaceStats data model.

    Corresponds to the elements of the nodes object in the GetNetworkInterfaceStats response.
    """
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    result = types.ModelType(network_models.NetworkInterfaceStats, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'result', self.result
        )


class NodeNetworkInterfaces(sf_models.Base):
    """Data model for the result field in the NodeNetworkInterfaceStatus data model.

    Part of the cluster API GetNetworkInterfaceStatus response.
    """
    interfaces = types.DictType(types.EnumType(enums.NetworkStatus), required=True)


class NodeNetworkInterfaceStatus(sf_models.Base):
    """Data model for each node in the cluster api GetNetworkInterfaceStatus method."""
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    result = types.ModelType(NodeNetworkInterfaces, required=True)


class NodeRtfiInfo(sf_models.Base):
    """NodeRtfiInfo data model.

    Corresponds to the rtfi_info object in the ResetNode response.
    """
    build = types.StringType(required=True)
    generation = types.IntType(required=True)
    generation_next = types.IntType(required=True, serialized_name='generationNext')
    mip = types.StringType(required=True)
    mipi = types.StringType(required=True)
    status_url_all = types.StringType(required=True, serialized_name='statusUrlAll')
    status_url_current = types.StringType(required=True, serialized_name='statusUrlCurrent')
    status_url_logfile = types.StringType(required=True, serialized_name='statusUrlLogfile')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'build', self.build,
            'status_url_all', self.status_url_all
        )


class NodeRtfiDetails(sf_models.Base):
    """NodeRtfi details."""
    rtfi_info = types.ModelType(NodeRtfiInfo, serialized_name='rtfiInfo')


class NodeStats(sf_models.Base):
    """NodeStats data model.

    Corresponds to the node_stats object in the GetNodeStats response.
    """
    c_bytes_in = types.IntType(required=True, serialized_name=api_constants.C_BYTES_IN)
    c_bytes_out = types.IntType(required=True, serialized_name=api_constants.C_BYTES_OUT)
    cpu = types.IntType(required=True)
    m_bytes_in = types.IntType(required=True, serialized_name=api_constants.M_BYTES_IN)
    m_bytes_out = types.IntType(required=True, serialized_name=api_constants.M_BYTES_OUT)
    network_utilization_cluster = types.IntType(
        required=True,
        serialized_name=api_constants.NETWORK_UTILIZATION_CLUSTER
    )
    network_utilization_storage = types.IntType(
        required=True,
        serialized_name=api_constants.NETWORK_UTILIZATION_STORAGE
    )
    node_id = types.IntIdType(serialized_name=api_constants.NODE_ID)
    s_bytes_in = types.IntType(required=True, serialized_name=api_constants.S_BYTES_IN)
    s_bytes_out = types.IntType(required=True, serialized_name=api_constants.S_BYTES_OUT)
    timestamp = types.DateTimeType(required=True)
    used_memory = types.IntType(required=True, serialized_name='usedMemory')

    # Version 10.0 Additions
    count = types.IntType(version_added=sf_version.SfVersion('10.0'))
    cpu_total = types.IntType(
        serialized_name=api_constants.CPU_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )
    read_latency_usec_total = types.IntType(
        required=True,
        serialized_name=api_constants.READ_LATENCY_USEC_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )
    read_ops = types.IntType(
        required=True,
        serialized_name='readOps',
        version_added=sf_version.SfVersion('10.0')
    )
    write_latency_usec_total = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_LATENCY_USEC_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )
    write_ops = types.IntType(
        required=True,
        serialized_name='writeOps',
        version_added=sf_version.SfVersion('10.0')
    )

    # Version 11.1 Additions
    ssload_histogram = types.ModelType(
        model_spec=general_models.Histogram,
        serialized_name='ssLoadHistogram',
        version_added=sf_version.SfVersion('11.1')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'used_memory', self.used_memory,
            'cpu', self.cpu
        )


class NodeSvipDetails(sf_models.Base):
    """NodeSvipDetails data model.

    Corresponds to the details object in the TestConnectSvip response.
    """
    connected = types.BooleanType(required=True)
    ping_bytes = types.DictType(
        field=types.ModelType(network_models.PingResponse),
        required=True,
        serialized_name='pingBytes'
    )
    svip = types.IPAddressType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'svip', self.svip,
            'connected', self.connected
        )


class NodeVersionInfo(sf_models.Base):
    """NodeVersionInfo data model.

    Corresponds to the elements of the cluster_version_info object in the GetClusterVersionInfo response.
    """
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    node_internal_revision = types.StringType(required=True, serialized_name='nodeInternalRevision')
    node_version = types.SfBuildVersionType(required=True, serialized_name='nodeVersion')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'node_version', self.node_version
        )


class PendingActiveNode(SfNodeBase):
    """PendingActiveNode data model.

    Corresponds to the elements of the pending_active_nodes object in the ListPendingActiveNodes
    response.
    """
    pending_node_id = types.IntIdType(required=True, serialized_name=api_constants.PENDING_NODE_ID)

    # Version 9.0 Additions
    active_node_key = types.StringType(
        required=True,
        serialized_name=api_constants.ACTIVE_NODE_KEY,
        version_added=sf_version.SfVersion('9.0')
    )
    assigned_node_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ASSIGNED_NODE_ID,
        version_added=sf_version.SfVersion('9.0')
    )
    async_handle = types.IntIdType(
        required=True,
        min_value=1,
        serialized_name=api_constants.ASYNC_HANDLE,
        version_added=sf_version.SfVersion('9.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'assigned_node_id', self.assigned_node_id,
            'mip', self.mip,
            'sip', self.sip
        )


class PendingNode(SfNodeBase):
    """PendingNode data model.

    Corresponds to the elements of the pending_nodes object in the ListPendingNodes response.
    """
    active_node_key = types.StringType(serialized_name=api_constants.ACTIVE_NODE_KEY)
    assigned_node_id = types.IntIdType(required=True, serialized_name=api_constants.ASSIGNED_NODE_ID)
    async_handle = types.IntIdType(serialized_name=api_constants.ASYNC_HANDLE)
    cipi = types.StringType(required=True)
    mipi = types.StringType(required=True)
    name = types.StringType(required=True)
    pending_node_id = types.IntIdType(required=True, serialized_name=api_constants.PENDING_NODE_ID)
    sipi = types.StringType(required=True)
    # 11.0 Availability zones
    chassis_name = types.StringType(version_added=sf_version.SfVersion('11.0'),
                                    serialized_name=api_constants.CHASSIS_NAME)
    node_slot = types.StringType(version_added=sf_version.SfVersion('11.0'), serialized_name='nodeSlot')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'pending_node_id', self.pending_node_id,
            'name', self.name,
            'mip', self.mip,
            'sip', self.sip
        )


class EnsembleNode(sf_models.Base):
    """Zookeeper node."""
    ip = types.IPAddressType(required=True, serialized_name='IP')
    node_id = types.IntType(required=True, serialized_name=api_constants.NODE_ID)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'ip', self.ip,
            'node_id', self.node_id
        )


class Ensemble(sf_models.Base):
    """Zookeeper ensemble data model."""
    mode = types.StringType(required=True)
    nodes = types.ListType(types.ModelType(EnsembleNode), required=True)

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'mode', self.mode
        )


class ZookeeperView(sf_models.Base):
    """Zookeeper view data model."""
    active_ensemble = types.ModelType(Ensemble, required=True, serialized_name='activeEnsemble')
    active_ensemble_peer = types.BooleanType(
        required=True, serialized_name='activeEnsemblePeer')
    dynamic_config = types.StringType(serialized_name='dynamicConfig')
    my_id = types.IntType(serialized_name='myID')
    proposed_ensemble = types.ModelType(Ensemble)
    proposed_ensemble_peer = types.BooleanType(
        required=True, serialized_name='proposedEnsemblePeer')
    proposed_equals_active = types.BooleanType(
        required=True, serialized_name='proposedEqualsActive')
    reconfigure_pending = types.BooleanType(
        required=True, serialized_name='reconfigurePending')
    static_config = types.StringType(serialized_name='staticConfig')

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'active_ensemble', self.active_ensemble
        )


class EnsembleNodeInfo(sf_models.Base):
    """ActiveNode data model.

    Corresponds to the nodes object in the ListActiveNodes response.
    """
    associated_fservice_id = types.IntType(
        required=True,
        serialized_name=api_constants.ASSOCIATED_FSERVICE_ID
    )
    cip = types.IPv4Type(required=True)
    cipi = types.StringType(required=True)
    fibre_channel_target_port_group = types.IntType(
        serialized_name=api_constants.FIBRE_CHANNEL_TARGET_PORT_GROUP
    )
    mip = types.IPv4Type(required=True)
    mipi = types.StringType(required=True)
    name = types.StringType(required=True)
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    node_slot = types.StringType(required=False, serialized_name=api_constants.NODE_SLOT)
    platform_info = types.ModelType(
        model_spec=PlatformInfo,
        required=True,
        serialized_name=api_constants.PLATFORM_INFO
    )
    sip = types.IPv4Type(required=True)
    sipi = types.StringType(required=True)
    software_version = types.SfBuildVersionType(
        required=True,
        serialized_name=api_constants.SOFTWARE_VERSION
    )
    uuid = types.UUIDType(required=True)
    virtual_networks = types.ListType(
        field=types.ModelType(network_models.MinimalVirtualNetwork),
        required=True,
        serialized_name=api_constants.VIRTUAL_NETWORKS
    )

    # 11.0 Added
    mip_v6 = types.IPAddressType(version_added=sf_version.SfVersion('11.0'),
                                 serialized_name='mipV6')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_id', self.node_id,
            'name', self.name,
            'mip', self.mip,
            'sip', self.sip
        )


class ServerStats(sf_models.Base):
    """Zookeeper server stats."""
    connections = types.IntType(required=True, serialized_name='Connections')
    latency_min_avg_max = types.StringType(
        required=True, serialized_name='Latency min/avg/max')
    mode = types.EnumType(enums.ZookeeperType, serialized_name='Mode')
    node_count = types.IntType(required=True, serialized_name='Node count')
    outstanding = types.IntType(required=True, serialized_name='Outstanding')
    received = types.IntType(required=True, serialized_name='Received')
    sent = types.IntType(required=True, serialized_name='Sent')
    zookeeper_version = types.StringType(required=True, serialized_name='Zookeeper version')
    zxid = types.StringType(required=True, serialized_name='Zxid')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'mode', self.mode,
            'node_count', self.node_count
        )


class Package(sf_models.Base):
    """Package data model.

    Corresponds to the packages object in the ListPackageInfo response.
    """
    active = types.BooleanType(required=True)
    activation_status = types.DictType(
        types.UnionType([
            types.StringType,
            types.ListType(types.DictType(types.StringType))
        ]),
        required=False,
        version_added=sf_version.SfVersion('12.0'),
        serialized_name=api_constants.ACTIVATION_STATUS
    )
    category = types.StringType()
    comment = types.StringType()
    description = types.StringType()
    display_name = types.StringType(serialized_name=api_constants.DISPLAY_NAME)
    factory_installed = types.BooleanType(required=True, serialized_name=api_constants.FACTORY_INSTALLED)
    homepage = types.StringType()
    license = types.StringType()
    name = types.StringType(required=True)
    package = types.StringType()
    repository = types.StringType()
    revision = types.StringType()
    show_in_ui = types.BooleanType(serialized_name=api_constants.SHOW_IN_UI)
    useflags = types.StringType()
    version = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'active', self.active,
            'factory_installed', self.factory_installed,
            'show_in_ui', self.show_in_ui,
            'version', self.version
        )


class BootstrapInfo(sf_models.Base):
    """Node Models for bootstrap info (Type=NodeWaitingToJoin).

    Description for each node that is actively waiting to join this cluster: compatible - Indicates
    if the listed node is compatible with the node the API call was executed against. name - IP
    address of each node. version - version of SolidFire Element software currently installed
    on the node.
    """
    chassis_type = types.EnumType(
        enum_class=enums.ChassisType,
        serialized_name=api_constants.CHASSIS_TYPE
    )
    cip = types.IPAddressType()
    compatible = types.BooleanType(required=True)
    hostname = types.StringType()
    mip = types.IPAddressType()
    mipV6 = types.IPAddressType(version_added=sf_version.SfVersion('11.0'))
    name = types.StringType(version_added=sf_version.SfVersion('12.0'))
    node_id = types.IntIdType(serialized_name=api_constants.NODE_ID)
    node_type = types.NodeTypeType(serialized_name=api_constants.NODE_TYPE)
    pending_node_id = types.IntIdType(serialized_name=api_constants.PENDING_NODE_ID,
                                      version_added=sf_version.SfVersion('12.0'))
    sip = types.NetworkInterfaceNameType(version_added=sf_version.SfVersion('12.0'))
    role = types.EnumType(enums.NodeRole)
    version = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'node_type', self.node_type,
            'node_id', self.node_id,
            'role', self.role
        )


class BootstrapConfig(sf_models.Base):
    """Cluster and node information from the bootstrap configuration file."""

    cluster_name = types.StringType(
        serialized_name=api_constants.CLUSTER_NAME,
        required=True
    )
    mvip = types.IPAddressType(required=True,
                               version_added=sf_version.SfVersion('12.0'),
                               serialized_name=api_constants.MVIP)
    node_name = types.StringType(
        serialized_name=api_constants.NODE_NAME,
        required=True
    )
    nodes = types.ListType(types.ModelType(BootstrapInfo, required=True), required=True,
                           serialized_name=api_constants.NODES)
    svip = types.IPAddressType(required=True,
                               version_added=sf_version.SfVersion('12.0'),
                               serialized_name=api_constants.SVIP)
    version = types.StringType(required=True, serialized_name=api_constants.VERSION)


class NodeStateInfo(sf_models.Base):
    """Node state info."""

    cluster = types.StringType(  # name of cluster
        serialized_name=api_constants.CLUSTER,
        required=True
    )
    state = types.EnumType(
        enums.NodeState,
        serialized_name=api_constants.STATE,
        required=True
    )


class NodeState(sf_models.Base):
    """Objects for each node in the cluster (Type=NodeStateResult)."""

    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    result = types.ModelType(NodeStateInfo, serialized_name=api_constants.RESULT)


class NodeBinaryVersionInfoDetails(sf_models.Base):
    """Objects for each binary enumerated in NodeBinaryVersionInfo."""

    build_date = types.StringType(required=True, serialized_name='BuildDate')
    build_type = types.StringType(required=True, serialized_name='BuildType')
    element = types.StringType(required=True, serialized_name='Element')
    ember_version = types.StringType(required=True, serialized_name='EmberVersion')
    md5 = types.StringType(required=True)
    options = types.StringType(required=True, serialized_name='Options')
    release = types.StringType(required=True, serialized_name='Release')
    release_short = types.StringType(required=True, serialized_name='ReleaseShort')
    repository = types.StringType(required=True, serialized_name='Repository')
    vcs = types.StringType(required=True, serialized_name='VCS')
    version = types.SfBuildVersionType(required=True, serialized_name='Version')


class NodeBinaryVersionInfo(sf_models.Base):
    """Object inside of sf_version in GetVersionInfo."""

    sfapp = types.ModelType(NodeBinaryVersionInfoDetails, required=True)
    sfbasiciocheck = types.ModelType(NodeBinaryVersionInfoDetails, required=True)
    sfconfig = types.ModelType(NodeBinaryVersionInfoDetails, required=True)
    sfnetwd = types.ModelType(NodeBinaryVersionInfoDetails, required=True)
    sfsvcmgr = types.ModelType(NodeBinaryVersionInfoDetails, required=True)


class ClusterState(sf_models.Base):
    """Enables you to indicate if a node is part of a cluster or not.

    The three states are:
    Available: Node has not been configured with a cluster name.
    Pending: Node is pending for a specific named cluster and can be added.
    Active: Node is an active member of a cluster and may not be added to another cluster.
    """

    cluster = types.StringType(serialized_name=api_constants.CLUSTER)
    state = types.EnumType(enums.NodeState, serialized_name=api_constants.STATE)
    nodes = types.ListType(types.ModelType(NodeState),
                           version_added=sf_version.SfVersion('12.0'),
                           serialized_name=api_constants.NODES
                           )


class MountedFileSystem(sf_models.Base):
    """Contains information about file system mounted on a node."""
    available = types.DataSizeType(required=True)
    capacity = types.DataSizeType(required=True)
    dev_path = types.StringType(required=True, serialized_name='devPath')
    device_id = types.IntIdType(required=True, serialized_name='deviceId')
    directory = types.StringType(required=True)
    free = types.DataSizeType(required=True)
    path = types.StringType(required=True, serialized_name='path')
    path_link = types.StringType(required=True, serialized_name='pathLink')
    type = types.StringType(required=True)
