"""NVRAM Data Models."""
import sf_models

from sf_api import api_constants
from sf_api.models import types


class NvramErrors(sf_models.Base):
    """NvramErrors data model.

    Corresponds to the errors object in the NvramInfo data model.
    Part of the GetHardwareInfo response.
    """
    num_of_error_log_entries = types.StringType(
        required=True,
        regex=r'^\d+',
        serialized_name='numOfErrorLogEntries'
    )

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'num_of_error_log_entries', self.num_of_error_log_entries
        )


class NvramIdentify(sf_models.Base):
    """NvramIdentify data model.

    Corresponds to the identify object in the NvramInfo data model.
    Part of the GetHardwareInfo response.
    """
    firmware_version = types.StringType(required=True, serialized_name='firmwareVersion')
    hardware_revision = types.StringType(
        required=True,
        regex=r'^[0-9a-zA-Z]+',
        serialized_name='hardwareRevision'
    )
    model_number = types.StringType(
        required=True,
        regex=r'^[0-9a-zA-Z-]+',
        serialized_name='modelNumber'
    )
    serial_number = types.StringType(required=True, regex=r'^\d+', serialized_name='serialNumber')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'model_number', self.model_number,
            'hardware_revision', self.hardware_revision
        )


class NvramFirmware(sf_models.Base):
    """NvramFirmware data model.

    Corresponds to the firmware object in the NvramInfo data model.
    Part of the GetHardwareInfo response.
    """
    active_slot_number = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='activeSlotNumber'
    )
    slot1_version = types.StringType(
        required=True,
        regex=r'^[0-9a-zA-Z]+',
        serialized_name='slot1Version'
    )
    slot2_version = types.StringType(
        required=True,
        regex=r'^[0-9a-zA-Z]+',
        serialized_name='slot2Version'
    )
    slot3_version = types.StringType(
        required=True,
        regex=r'^[0-9a-zA-Z]+',
        serialized_name='slot3Version'
    )
    slot4_version = types.StringType(
        required=True,
        regex=r'^[0-9a-zA-Z]+',
        serialized_name='slot4Version'
    )

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'active_slot_number', self.active_slot_number
        )


class NvramEvent(sf_models.Base):
    """NvramEvent data model.

    Corresponds to the elements of the event object in the NvramExtended data model.
    Part of the GetHardwareInfo response.
    """
    name = types.StringType(required=True)
    time = types.StringType(
        required=True,
        regex=api_constants.REGEX_DATETIMESTAMP
    )
    value = types.StringType(required=True, regex=r'^\d+$')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'value', self.value
        )


class NvramEventOccurrence(sf_models.Base):
    """NvramEventOccurrence data model.

    Corresponds to the elements of the event_occurrences object in the NvramExtended data model.
    Part of the GetHardwareInfo response.
    """
    count = types.StringType(required=True, regex=r'^\d+$')
    name = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'count', self.count
        )


class NvramSmartCounter(sf_models.Base):
    """NvramSmartCounter data model.

    Corresponds to the elements of the smart_counters object in the NvramExtended data model.
    Part of the GetHardwareInfo response.
    """
    name = types.StringType(required=True)
    value = types.StringType(required=True, regex=r'^\d+$')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'value', self.value
        )


class NvramExtended(sf_models.Base):
    """NvramExtended data model.

    Corresponds to the extended object in the NvramInfo data model.
    Part of the GetHardwareInfo response.
    """
    dialog_version = types.StringType(required=True, regex=r'^\d+$', serialized_name='dialogVersion')
    event = types.ListType(
        field=types.ModelType(NvramEvent),
        required=True
    )
    event_occurrences = types.ListType(
        field=types.ModelType(NvramEventOccurrence),
        required=True,
        serialized_name='eventOccurrences'
    )
    initial_capacitance = types.StringType(
        required=True,
        regex=r'^\d+\.\d+ F',
        serialized_name='initialCapacitance'
    )
    initial_esr = types.StringType(required=True, regex=r'^\d+\.\d+ Ohm', serialized_name='initialEsr')
    measurement = types.ListType(types.DictType(types.BaseType), required=True)
    smart_counters = types.ListType(
        field=types.ModelType(NvramSmartCounter),
        required=True,
        serialized_name='smartCounters'
    )
    snapshot_time = types.StringType(
        required=True,
        regex=api_constants.REGEX_DATETIMESTAMP,
        serialized_name='snapshotTime'
    )

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'event', self.event
        )


class NvramSmart(sf_models.Base):
    """NvramSmart data model.

    Corresponds to the smart object in the NvramInfo data model.
    Part of the GetHardwareInfo response.
    """
    available_space = types.StringType(
        required=True,
        regex=api_constants.REGEX_PERCENT,
        serialized_name='availableSpace'
    )
    available_space_threshold = types.StringType(
        required=True,
        regex=api_constants.REGEX_PERCENT,
        serialized_name='availableSpaceThreshold'
    )
    controller_busy_time_minutes = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='controllerBusyTimeMinutes'
    )
    critical_error_vector = types.StringType(
        required=True,
        regex=r'^\d+x\d+',
        serialized_name='criticalErrorVector'
    )
    media_errors = types.StringType(required=True, regex=r'^\d+$', serialized_name='mediaErrors')
    number_of_512_byte_blocks_read = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='numberOf512ByteBlocksRead'
    )
    number_of_512_byte_blocks_written = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='numberOf512ByteBlocksWritten'
    )
    number_of_error_info_logs = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='numberOfErrorInfoLogs'
    )
    number_of_host_read_commands = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='numberOfHostReadCommands'
    )
    number_of_host_write_commands = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='numberOfHostWriteCommands'
    )
    number_of_power_cycles = types.StringType(
        required=True,
        regex=r'^\d+$',
        serialized_name='numberOfPowerCycles'
    )
    power_on_hours = types.StringType(required=True, regex=r'^\d+$', serialized_name='powerOnHours')
    temperature = types.StringType(required=True, regex=r'^\d+ Kelvin')
    unsafe_shutdowns = types.StringType(required=True, regex=r'^\d+$', serialized_name='unsafeShutdowns')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'available_space', self.available_space,
            'available_space_threshold', self.available_space_threshold
        )


class NvramInfo(sf_models.Base):
    """NvramInfo data model.

    Corresponds to the nvram object in the NodeHardwareInfo data model.
    Part of the GetHardwareInfo response.
    """
    errors = types.ModelType(NvramErrors, required=True)
    extended = types.ModelType(NvramExtended, required=True)
    firmware = types.ModelType(NvramFirmware, required=True)
    identify = types.ModelType(NvramIdentify, required=True)
    smart = types.ModelType(NvramSmart, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'identify', self.identify,
            'firmware', self.firmware
        )
