"""SolidFire Schedule Data Models."""

import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import types


class Weekday(sf_models.Base):
    """Weekday data model for a schedule."""
    day = types.IntType(required=True, min_value=0, max_value=6)
    offset = types.IntType(min_value=0, max_value=6)

    @classmethod
    def create(cls, day=None, offset=None):
        """Get a weekday object to be used in a Schedule.

        :param int day: day to run schedule; valid is 0-6, 0 is Sunday. 6 is Saturday.
        :param int offset: occurrence in the month; valid is 0-4. If offset is 0, the schedule will
            occur every occurrence of the day. Otherwise, it will run on the 1st, 2nd, 3rd,
            or 4th occurrence of the day in the month.
        :return: weekday object
        :rtype: sf_api.models.Weekday
        """

        return super(Weekday, cls).create(
            day=day,
            offset=offset
        )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__, 'day', self.day, 'offset', self.offset)


class ScheduleInfo(sf_models.Base):
    """SolidFire Schedule info data model."""
    # volume_id and volumes are technically mutually exclusive, if one appears, the other will not
    volume_id = types.IntType(serialized_name=api_constants.VOLUME_ID, serialize_when_none=False)
    volumes = types.ListType(types.IntIdType, serialize_when_none=False)
    # SolidFire only returns these keys if you set them.
    name = types.StringType(serialize_when_none=False)
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION,
        serialize_when_none=False
    )
    retention = types.StringType(serialize_when_none=False)

    @classmethod
    def create(cls, volume_id=None, volumes=None, name=None,
               enable_remote_replication=None, retention=None):
        """Get a ScheduleInfo object.

        volume_id and volumes should be mutually exclusive. Other params are optional.

        :param int volume_id: The volume id to snapshot
        :param list[int] volumes: the volume_ids to group snapshot
        :param str name: name of the resulting snapshot
        :param bool enable_remote_replication: remote rep the snapshot
        :param str retention: length of time to retain
        :return: schedule info object
        :rtype: sf_api.models.ScheduleInfo
        """
        return super(ScheduleInfo, cls).create(
            volume_id=volume_id,
            volumes=volumes,
            name=name,
            enable_remote_replication=enable_remote_replication,
            retention=retention
        )

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__, 'name', self.name)


class Schedule(sf_models.Base):
    """SolidFire Schedule Data Model."""
    attributes = types.OptionalDictType(types.BaseType(), default=dict)
    has_error = types.BooleanType(required=True, serialized_name=api_constants.HAS_ERROR)
    hours = types.IntType(required=True, min_value=0, max_value=23)
    last_run_status = types.EnumType(enums.ScheduleStatus, required=True,
                                     serialized_name=api_constants.LAST_RUN_STATUS)
    last_run_time_started = types.DateTimeType(serialized_name=api_constants.LAST_RUN_TIME_STARTED)
    minutes = types.IntType(required=True, min_value=0, max_value=59)
    monthdays = types.ListType(types.IntType(min_value=1, max_value=31), required=True, default=list)
    paused = types.BooleanType(required=True)
    recurring = types.BooleanType(required=True)
    run_next_interval = types.BooleanType(required=True, serialized_name=api_constants.RUN_NEXT_INTERVAL)
    schedule_id = types.IntIdType(required=True, serialized_name=api_constants.SCHEDULE_ID)
    schedule_info = types.ModelType(ScheduleInfo, required=True,
                                    serialized_name=api_constants.SCHEDULE_INFO)
    schedule_name = types.StringType(required=True, serialized_name=api_constants.SCHEDULE_NAME)
    schedule_type = types.EnumType(enums.ScheduleType, required=True,
                                   serialized_name=api_constants.SCHEDULE_TYPE)
    snapmirror_label = types.StringType(serialized_name=api_constants.SNAPMIRROR_LABEL,
                                        version_added=sf_version.SfVersion('10.0'))
    starting_date = types.DateTimeType(serialized_name=api_constants.STARTING_DATE)
    to_be_deleted = types.BooleanType(serialized_name=api_constants.TO_BE_DELETED)
    weekdays = types.ListType(types.ModelType(Weekday), default=list)

    @classmethod
    def create(cls, schedule_name=None, schedule_info=None,
               schedule_type=enums.ScheduleType.snapshot, monthdays=None, weekdays=None,
               recurring=None, paused=None, hours=None,
               minutes=None, attributes=None, starting_date=None, snapmirror_label=None,
               has_error=None, last_run_status=None, last_run_time_started=None, run_next_interval=None,
               schedule_id=None, to_be_deleted=None):
        """Get a schedule object.

        :param str schedule_name: The name of the schedule
        :param sf_api.models.ScheduleInfo schedule_info: The schedule info, essentially what it's
            taking a snapshot of.
        :param sf_api.enums.ScheduleType schedule_type: The schedule type, only snapshot is supported
        :param list[int] monthdays: The days of the month to schedule a snapshot
        :param list[sf_api.models.Weekday] weekdays: List of weekday objects
        :param bool recurring: If the schedule is recurring
        :param bool paused: If the schedule should already be paused
        :param int hours: Hours between snapshots
        :param int minutes: Minutes between snapshots
        :param dict attributes: Generic attributes dictionary
        :param str starting_date: The ISO 8601 formatted timestamp to start. if none, starts now.
        :param str snapmirror_label: Label used by snapmirror to specify snapshot retention on
            snapmirror endpoints.
        :param bool has_error: Indicates if the schedule has errors
        :param sf_api.enums.ScheduleStatus|str last_run_status: Status of last scheduled snapshot
        :param datetime.datetime|str last_run_time_started: Last time the schedule started
        :param bool run_next_interval: Indicates if the schedule will run the next time the scheduler
            is active.
        :param int schedule_id: Unique ID of the schedule
        :param bool to_be_deleted: Indicates if the schedule is marked for deletion
        :return: Partial schedule object
        :rtype: sf_api.models.Schedule
        """
        return super(Schedule, cls).create(
            schedule_name=schedule_name,
            schedule_type=schedule_type,
            schedule_info=schedule_info,
            weekdays=weekdays,
            monthdays=monthdays,
            recurring=recurring,
            paused=paused,
            hours=hours,
            minutes=minutes,
            attributes=attributes,
            starting_date=starting_date,
            snapmirror_label=snapmirror_label,
            has_error=has_error,
            last_run_status=last_run_status,
            last_run_time_started=last_run_time_started,
            run_next_interval=run_next_interval,
            schedule_id=schedule_id,
            to_be_deleted=to_be_deleted
        )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'schedule_id', self.schedule_id,
            'schedule_name', self.schedule_name,
            'schedule_type', self.schedule_type)
