"""Snapshot data model."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import types


class SnapshotRemoteStatus(sf_models.Base):
    """Snapshot remote status data model.

    Corresponds to SF object returned by ListSnapshots.
    """
    volume_pair_uuid = types.UUIDType(required=True, serialized_name=api_constants.VOLUME_PAIR_UUID,
                                      version_added=sf_version.SfVersion('11.1'))
    remote_status = types.EnumType(enums.RemoteClusterSnapshotStatus, required=True,
                                   serialized_name=api_constants.REMOTE_STATUS,
                                   version_added=sf_version.SfVersion('11.1'))


class Snapshot(sf_models.Base):
    """Snapshot Data Model.

    Corresponds to SF object returned by ListSnapshots.
    """
    attributes = types.DictType(types.StringType)
    checksum = types.StringType(required=True)
    create_time = types.DateTimeType(required=True, serialized_name=api_constants.CREATE_TIME)
    enable_remote_replication = types.BooleanType(
        required=True,
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION
    )
    expiration_reason = types.StringType(required=True, serialized_name=api_constants.EXPIRATION_REASON)
    expiration_time = types.DateTimeType(serialized_name=api_constants.EXPIRATION_TIME)
    group_id = types.IntType(required=True, serialized_name=api_constants.GROUP_ID)
    group_snapshot_uuid = types.UUIDType(
        required=True,
        serialized_name=api_constants.GROUP_SNAPSHOT_UUID
    )
    name = types.StringType(required=True)
    snapmirror_label = types.StringType(serialized_name=api_constants.SNAPMIRROR_LABEL,
                                        version_added=sf_version.SfVersion('10.0'))
    snapmirror_attributes = types.DictType(types.StringType, default=dict,
                                           serialized_name=api_constants.SNAPMIRROR_ATTRIBUTES,
                                           version_added=sf_version.SfVersion('10.0'))
    snapshot_id = types.IntIdType(required=True, serialized_name=api_constants.SNAPSHOT_ID)
    snapshot_uuid = types.UUIDType(required=True, serialized_name=api_constants.SNAPSHOT_UUID)
    status = types.EnumType(enums.SnapshotStatus, required=True)
    total_size = types.DataSizeType(required=True, serialized_name=api_constants.TOTAL_SIZE)
    virtual_volume_id = types.UUIDType(required=False, serialized_name=api_constants.VIRTUAL_VOLUME_ID)
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)

    # Remote Statuses is a required field on the source side, but is not displayed on the target side.
    remote_statuses = types.ListType(types.ModelType(SnapshotRemoteStatus),
                                     serialized_name=api_constants.REMOTE_STATUSES,
                                     version_added=sf_version.SfVersion('11.1'))

    @classmethod
    def create(cls, volume_id=None, name=None, enable_remote_replication=None, attributes=None,
               checksum=None, create_time=None, expiration_reason=None, expiration_time=None,
               group_id=None, group_snapshot_uuid=None, snapmirror_label=None,
               snapmirror_attributes=None, snapshot_id=None, snapshot_uuid=None, status=None,
               total_size=None, virtual_volume_id=None):
        """Define a snapshot profile to pass to create functions/methods.

        :param int volume_id: ID of source volume.
        :param str name: Name of the snapshot.
        :param bool enable_remote_replication: If true snapshot will replicate to remote storage.
        :param dict attributes: Any additional key value pairs to be added to the account.
        :param str checksum: Checksum of data stored in snapshot
        :param datetime.datetime|str create_time: Time snapshot was created
        :param str expiration_reason: Indicates how the expiration was set
        :param datetime.datetime|str expiration_time: Time when snapshot will expire and be purged
        :param int group_id: Group ID if snapshot is member of a group
        :param str group_snapshot_uuid: Each member of a group has this UUID
        :param str snapmirror_label: Snapmirror label
        :param dict snapmirror_attributes: Snapmirror attributes object
        :param str snapshot_id: Unique ID of an existing snapshot
        :param str snapshot_uuid: Universal Unique ID of an existing snapshot
        :param str|sf_api.enums.SnapshotStatus status: Current status of snapshot
        :param int|sf_api.data_size.DataSize total_size: Total size of snapshot
        :param str virtual_volume_id: Unique ID of virtual volume associated with snapshot
        :return: Partially defined snapshot data model to use for snapshot creation.
        :rtype: sf_api.models.Snapshot
        """
        return super(Snapshot, cls).create(
            volume_id=volume_id,
            name=name,
            enable_remote_replication=enable_remote_replication,
            attributes=attributes,
            checksum=checksum,
            create_time=create_time,
            expiration_reason=expiration_reason,
            expiration_time=expiration_time,
            group_id=group_id,
            group_snapshot_uuid=group_snapshot_uuid,
            snapmirror_label=snapmirror_label,
            snapmirror_attributes=snapmirror_attributes,
            snapshot_id=snapshot_id,
            snapshot_uuid=snapshot_uuid,
            status=status,
            total_size=total_size,
            virtual_volume_id=virtual_volume_id
        )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__, 'name', self.name, 'snapshot_id', self.snapshot_id)


class MemberSnapshot(sf_models.Base):
    """Member data model.

    Corresponds to SF object returned by CreateGroupSnapshots.
    """
    checksum = types.StringType(required=True)
    snapshot_id = types.IntIdType(required=True, serialized_name=api_constants.SNAPSHOT_ID)
    snapshot_uuid = types.UUIDType(required=True, serialized_name=api_constants.SNAPSHOT_UUID)
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__, 'snapshot_id', self.snapshot_id, 'volume_id', self.volume_id)


class RemoteClusterSnapshotStatus(sf_models.Base):
    """Remote Cluster Snapshot Status Data model.

    Corresponds to SF object returned by ListGroupSnapshots.
    """
    remote_status = types.EnumType(enums.RemoteClusterSnapshotStatus, required=True,
                                   serialized_name=api_constants.REMOTE_STATUS,
                                   version_added=sf_version.SfVersion('11.1'))


class GroupSnapshot(sf_models.Base):
    """GroupSnapshot Data Model.

    Corresponds to SF object returned by ListGroupSnapshots.
    """
    attributes = types.DictType(types.StringType)
    create_time = types.DateTimeType(required=True, serialized_name=api_constants.CREATE_TIME)
    group_snapshot_id = types.IntIdType(required=True, serialized_name=api_constants.GROUP_SNAPSHOT_ID)
    group_snapshot_uuid = types.UUIDType(required=True,
                                         serialized_name=api_constants.GROUP_SNAPSHOT_UUID)
    members = types.ListType(types.ModelType(Snapshot), required=True, default=list)
    name = types.StringType(required=True)
    status = types.EnumType(enums.GroupSnapshotStatus, required=True)

    # TODO (ELEM-2127): Once all production 11.3 builds support it, we can make this required.
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION,
        version_added=sf_version.SfVersion('11.3')
    )

    # Remote Statuses is a required field on the source side, but is not displayed on the target side.
    remote_statuses = types.ListType(types.ModelType(RemoteClusterSnapshotStatus),
                                     serialized_name=api_constants.REMOTE_STATUSES,
                                     version_added=sf_version.SfVersion('11.3'))

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__, 'name', self.name,
            'group_snapshot_id', self.group_snapshot_id)
