"""Volume Data Models."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    sf_version,
)
from sf_api.models import types


class BackupTarget(sf_models.Base):
    """BackupTarget data model.

    Corresponds to the backup_target object in the GetBackupTarget response.
    """
    attributes = types.DictType(types.BaseType(), required=True)
    backup_target_id = types.IntIdType(required=True, serialized_name='backupTargetID')
    name = types.StringType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'backup_target_id', self.backup_target_id
        )


class BulkVolumeJobAttributes(sf_models.Base):
    """BulkVolumeJobAttributes data model.

    Corresponds to the attributes object in the BulkVolumeJob data model.
    Part of the ListBulkVolumeJobs response.
    """
    blocks_per_transfer = types.IntType(required=True, serialized_name='blocksPerTransfer')
    first_pending_lba = types.IntType(required=True, serialized_name='firstPendingLba')
    n_lbas = types.IntType(required=True, serialized_name='nLbas')
    next_lba = types.IntType(required=True, serialized_name='nextLba')
    pending_lbas = types.ListType(types.IntType(), required=True, serialized_name='pendingLbas')
    percent_complete = types.IntType(required=True, serialized_name='percentComplete')
    start_lba = types.IntType(required=True, serialized_name='startLba')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'n_lbas', self.n_lbas,
            'percent_complete', self.percent_complete
        )


class BulkVolumeJob(sf_models.Base):
    """BulkVolumeJob data model.

    Corresponds to the elements of the bulk_volume_jobs object in the ListBulkVolumeJobs response.
    """
    attributes = types.ModelType(BulkVolumeJobAttributes, required=True)
    bulk_volume_id = types.IntIdType(required=True, serialized_name='bulkVolumeID')
    create_time = types.DateTimeType(required=True, serialized_name='createTime')
    elapsed_time = types.IntType(required=True, serialized_name='elapsedTime')
    format = types.StringType(required=True)
    key = types.StringType(required=True, regex=r'^[0-9a-z]+$')
    percent_complete = types.IntType(required=True, serialized_name='percentComplete')
    remaining_time = types.IntType(required=True, serialized_name='remainingTime')
    script = types.StringType(required=True)
    snapshot_id = types.IntIdType(required=True, serialized_name='snapshotID')
    src_volume_id = types.IntIdType(required=True, serialized_name='srcVolumeID')
    status = types.EnumType(enums.AsyncResultStatus, required=True)
    type = types.EnumType(enums.BulkVolumeJobType, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'bulk_volume_id', self.bulk_volume_id,
            'status', self.status,
            'percent_complete', self.percent_complete
        )


class CloneJob(sf_models.Base):
    """CloneJob data model.

    Corresponds to the elements of the clone_jobs object in the ListCloneJobs response.
    """
    bytes_per_second = types.FloatType(required=True, serialized_name='bytesPerSecond')
    current_bytes = types.IntType(required=True, serialized_name='currentBytes')
    destination_volume_id = types.IntIdType(required=True, serialized_name='dstVolumeID')
    elapsed_time = types.FloatType(required=True, serialized_name='elapsedTime')
    node_id = types.IntIdType(required=True, serialized_name=api_constants.NODE_ID)
    percent_complete = types.FloatType(required=True, serialized_name='percentComplete')
    remaining_time = types.FloatType(required=True, serialized_name='remainingTime')
    snapshot_id = types.IntIdType(required=True, serialized_name=api_constants.SNAPSHOT_ID)
    source_volume_id = types.IntIdType(required=True, serialized_name='srcVolumeID')
    total_bytes = types.IntType(required=True, serialized_name='totalBytes')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'source_volume_id', self.source_volume_id,
            'destination_volume_id', self.destination_volume_id,
            'percent_complete', self.percent_complete,
            'remaining_time', self.remaining_time
        )


class ClonedVolumeInfo(sf_models.Base):
    """ClonedVolumeInfo data model.

    Corresponds to the elements of the members object in the CloneMultipleVolumes response.
    """
    src_volume_id = types.IntIdType(required=True, serialized_name='srcVolumeID')
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_id', self.volume_id,
            'src_volume_id', self.src_volume_id
        )


class Initiator(sf_models.Base):
    """Initiator data model.

    Corresponds to elements of the initiators object in the ListInitiators response.
    """
    alias = types.StringType(required=True)
    attributes = types.DictType(types.BaseType(), required=True)
    chap_username = types.StringType(required=False, serialized_name=api_constants.CHAP_USERNAME,
                                     version_added=sf_version.SfVersion('11.3'))
    initiator_secret = types.StringType(required=False, serialized_name=api_constants.INITIATOR_SECRET,
                                        version_added=sf_version.SfVersion('11.3'))
    initiator_id = types.IntIdType(required=True, serialized_name=api_constants.INITIATOR_ID)
    initiator_name = types.StringType(required=True, serialized_name=api_constants.INITIATOR_NAME)
    require_chap = types.BooleanType(required=True, serialized_name=api_constants.REQUIRE_CHAP,
                                     version_added=sf_version.SfVersion('11.3'))
    target_secret = types.StringType(required=False, serialized_name=api_constants.TARGET_SECRET,
                                     version_added=sf_version.SfVersion('11.3'))
    volume_access_groups = types.ListType(
        field=types.IntType(),
        required=True,
        default=list,
        serialized_name=api_constants.VOLUME_ACCESS_GROUPS
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'initiator_name', self.initiator_name,
            'initiator_id', self.initiator_id
        )

    @classmethod
    def create(cls, initiator_name, alias=None, attributes=None, initiator_id=None,
               volume_access_groups=None, chap_username=None, initiator_secret=None,
               require_chap=None, target_secret=None):
        """Define an initiator to pass to create functions/methods.

        :param str initiator_name: the name of the initiator
        :param str alias: alias for the initiator
        :param dict attributes: attributes to be applied to initiator
        :param str chap_username: The unique CHAP username for this initiator
        :param str initiator_secret: The CHAP secret used to authenticate the initiator
        :param bool require_chap: True if CHAP is required for this initiator
        :param str target_secret: The CHAP secret used to authenticate the target using mutual CHAP
        :param int initiator_id: id number for the new initiator
        :param list[int] volume_access_groups: the volume access group which the initiator is a member
        :rtype: valence.models.Initiator
        """
        return super(Initiator, cls).create(alias=alias, attributes=attributes,
                                            chap_username=chap_username,
                                            initiator_secret=initiator_secret,
                                            initiator_id=initiator_id,
                                            initiator_name=initiator_name,
                                            require_chap=require_chap,
                                            target_secret=target_secret,
                                            volume_access_groups=volume_access_groups)


class LunAssignment(sf_models.Base):
    """LunAssignment data model.

    Corresponds to the elements of the lun_assignments object in the VAGLunAssignments data model.
    Part of the ModifyVolumeAccessGroupLunAssignments response.
    """
    lun = types.IntIdType(required=True)
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'lun', self.lun,
            'volume_id', self.volume_id
        )


class MetadataHosts(sf_models.Base):
    """MetadataHosts data model.

    Corresponds to the metadata_hosts object in the VolumeStats data model.
    Part of the GetVolumeStats response.
    """
    dead_secondaries = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name='deadSecondaries'
    )
    live_secondaries = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name='liveSecondaries'
    )
    primary = types.IntIdType(required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'primary', self.primary,
            'live_secondaries', self.live_secondaries,
            'dead_secondaries', self.dead_secondaries
        )


class QosCurve(sf_models.Base):
    """QoSCurve data model.

    Corresponds to the curve object in the CreateVolume response.
    """
    b_4096 = types.IntType(required=True, default=100, serialized_name='4096')
    b_8192 = types.IntType(required=True, default=160, serialized_name='8192')
    b_16384 = types.IntType(required=True, default=270, serialized_name='16384')
    b_32768 = types.IntType(required=True, default=500, serialized_name='32768')
    b_65536 = types.IntType(required=True, default=1000, serialized_name='65536')
    b_131072 = types.IntType(required=True, default=1900, serialized_name='131072')
    b_262144 = types.IntType(required=True, default=3000, serialized_name='262144')
    b_524288 = types.IntType(required=True, default=7500, serialized_name='524288')
    b_1048576 = types.IntType(required=True, default=15000, serialized_name='1048576')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}, {}={}, {}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            '4096', self.b_4096,
            '8192', self.b_8192,
            '16384', self.b_16384,
            '32768', self.b_32768,
            '65536', self.b_65536,
            '131072', self.b_131072,
            '262144', self.b_262144,
            '524288', self.b_524288,
            '1048576', self.b_1048576
        )


class Qos(sf_models.Base):
    """QoS data model.

    Corresponds to the result object in the GetDefaultQoS response.
    """
    min_iops = types.IntType(required=True, serialized_name=api_constants.MIN_IOPS)
    max_iops = types.IntType(required=True, serialized_name=api_constants.MAX_IOPS)
    burst_iops = types.IntType(required=True, serialized_name=api_constants.BURST_IOPS)
    burst_time = types.IntType(required=True, serialized_name=api_constants.BURST_TIME)
    curve = types.ModelType(QosCurve, required=True)

    @classmethod
    def create(cls, min_iops=None, max_iops=None, burst_iops=None, burst_time=None, curve=None):
        """Define a QoS profile to pass to create functions/methods.

        :param int min_iops: Minimum IOPS
        :param int max_iops: Maximum IOPS
        :param int burst_iops: Burst IOPS
        :param int burst_time: Duration in seconds to sustain burst_iops
        :param sf_api.models.QosCurve|dict curve: Qos curve model
        :returns: QoS data model
        :rtype: sf_api.models.Qos
        """
        return super(Qos, cls).create(
            min_iops=min_iops,
            max_iops=max_iops,
            burst_iops=burst_iops,
            burst_time=burst_time,
            curve=curve
        )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'min_iops', self.min_iops,
            'max_iops', self.max_iops
        )


class QosPolicy(sf_models.Base):
    """QoSPolicy data model.

    Corresponds to the qos_policy object in the GetQoSPolicy response.
    """
    name = types.StringType(required=True)
    qos = types.ModelType(Qos, required=True)
    qos_policy_id = types.IntType(required=True, serialized_name='qosPolicyID')
    volume_ids = types.ListType(types.IntIdType(), required=True, serialized_name='volumeIDs')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'qos_policy_id', self.qos_policy_id,
            'qos', self.qos
        )


class SnapshotReplication(sf_models.Base):
    """SnapshotReplication data model.

    Corresponds to the snapshot_replication object in the VolumePairRemoteReplication data model.
    Part of the ListVolumes response.
    """
    state = types.EnumType(enums.RemoteRepState, required=True)
    state_details = types.StringType(required=False, serialized_name=api_constants.STATE_DETAILS)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'state', self.state,
            'state_details', self.state_details
        )


class SyncJob(sf_models.Base):
    """SyncJob data model.

    Corresponds to the elements of the sync_jobs object in the ListSyncJobs response.
    """
    # Fields required if type is remote
    blocks_per_second = types.IntType(serialized_name='blocksPerSecond')
    branch_type = types.EnumType(enums.SyncJobBranchType, serialized_name='branchType')
    dst_service_id = types.IntIdType(serialized_name='dstServiceID')

    # Fields required if type is remote or clone
    dst_volume_id = types.IntIdType(serialized_name='dstVolumeID')
    stage = types.EnumType(enums.SyncJobStage)

    # Fields required if type is clone
    clone_id = types.IntIdType(serialized_name=api_constants.CLONE_ID)
    node_id = types.IntIdType(serialized_name=api_constants.NODE_ID)
    snapshot_id = types.IntIdType(serialized_name=api_constants.SNAPSHOT_ID)
    group_clone_id = types.IntIdType(serialized_name='groupCloneID')

    # Fields required if type is clone or slice
    bytes_per_second = types.FloatType(serialized_name='bytesPerSecond')
    current_bytes = types.IntType(serialized_name='currentBytes')
    total_bytes = types.IntType(serialized_name='totalBytes')

    # Fields required if type is slice
    slice_id = types.IntIdType(serialized_name='sliceID')

    # Fields required for all types (some can be null)
    elapsed_time = types.FloatType(serialized_name='elapsedTime')
    percent_complete = types.FloatType(serialized_name='percentComplete')
    remaining_time = types.FloatType(serialized_name='remainingTime')
    src_service_id = types.IntIdType(serialized_name='srcServiceID')
    src_volume_id = types.IntIdType(serialized_name='srcVolumeID')
    type = types.EnumType(enums.SyncJobType, required=True)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'type', self.type,
            'slice_id', self.slice_id,
            'percent_complete', self.percent_complete,
            'remaining_time', self.remaining_time
        )


class VagLunAssignments(sf_models.Base):
    """VagLunAssignments data model.

    Corresponds to the volume_access_group_lun_assignments object in the
    ModifyVolumeAccessGroupLunAssignments response.
    """
    deleted_lun_assignments = types.ListType(
        field=types.ModelType(LunAssignment),
        required=True,
        serialized_name='deletedLunAssignments'
    )
    lun_assignments = types.ListType(
        field=types.ModelType(LunAssignment),
        required=True,
        serialized_name='lunAssignments'
    )
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_access_group_id', self.volume_access_group_id,
            'lun_assignments', self.lun_assignments
        )


class VagVolume(sf_models.Base):
    """VagVolume data model (valid in v6.1 - v7.0).

    Corresponds to the elements of the volumes_6 object in the VolumeAccessGroup data model.
    Part of the ListVolumeAccessGroups response.
    """
    logical_unit_number = types.IntType(required=True, serialized_name='logicalUnitNumber')
    volume_id = types.IntIdType(required=True, serialized_name='volumeID')

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'volume_id', self.volume_id
        )


class VolumePairRemoteReplication(sf_models.Base):
    """VolumePairRemoteReplication data model.

    Corresponds to the remote_replication object in the VolumePair data model.
    Part of the ListVolumes response.
    """
    mode = types.EnumType(enums.RemoteRepMode, required=True)
    pause_limit = types.IntType(required=True, serialized_name=api_constants.PAUSE_LIMIT)
    remote_service_id = types.IntType(required=True, serialized_name=api_constants.REMOTE_SERVICE_ID)
    resume_details = types.StringType(required=False, default='',
                                      serialized_name=api_constants.RESUME_DETAILS)
    snapshot_replication = types.ModelType(SnapshotReplication, required=True,
                                           serialized_name=api_constants.SNAPSHOT_REPLICATION)
    state = types.EnumType(enums.RemoteRepState, required=True)
    state_details = types.StringType(required=False, serialized_name=api_constants.STATE_DETAILS)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'state', self.state)


class VolumePair(sf_models.Base):
    """VolumePair data model.

    Corresponds to the elements of the volume_pairs object in the Volume data model.
    Part of the ListVolumes response.
    """
    cluster_pair_id = types.IntType(required=True, serialized_name=api_constants.CLUSTER_PAIR_ID)
    remote_replication = types.ModelType(VolumePairRemoteReplication, required=True,
                                         serialized_name=api_constants.REMOTE_REPLICATION)
    remote_slice_id = types.IntType(required=True, serialized_name=api_constants.REMOTE_SLICE_ID)
    remote_volume_id = types.IntType(required=True, serialized_name=api_constants.REMOTE_VOLUME_ID)
    remote_volume_name = types.StringType(required=True,
                                          serialized_name=api_constants.REMOTE_VOLUME_NAME)
    volume_pair_uuid = types.UUIDType(required=True, serialized_name=api_constants.VOLUME_PAIR_UUID)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'cluster_pair_id', self.cluster_pair_id,
            'remote_volume_name', self.remote_volume_name
        )


class Volume(sf_models.Base):
    """Volume data model.

    Corresponds to the elements of the volumes object in the ListVolumes response.
    """
    access = types.EnumType(enums.VolumeAccess, required=True)
    account_id = types.IntType(required=True, serialized_name=api_constants.ACCOUNT_ID)
    attributes = types.DictType(types.BaseType())
    block_size = types.IntType(required=True, serialized_name=api_constants.BLOCK_SIZE)
    create_time = types.DateTimeType(required=True, serialized_name=api_constants.CREATE_TIME)
    delete_time = types.OptionalDateTimeType(serialized_name=api_constants.DELETE_TIME)
    enable512e = types.BooleanType(required=True)
    iqn = types.StringType(required=True)
    name = types.StringType(required=True)
    purge_time = types.OptionalDateTimeType(serialized_name=api_constants.PURGE_TIME)
    qos = types.ModelType(Qos, required=True)
    scsi_eui_device_id = types.StringType(required=True,
                                          serialized_name=api_constants.SCSI_EUI_DEVICE_ID)
    scsi_naa_device_id = types.StringType(required=True,
                                          serialized_name=api_constants.SCSI_NAA_DEVICE_ID)
    slice_count = types.IntType(required=True, serialized_name=api_constants.SLICE_COUNT)
    status = types.EnumType(enums.VolumeStatus, required=True)
    total_size = types.DataSizeType(required=True, serialized_name=api_constants.TOTAL_SIZE)
    virtual_volume_id = types.UUIDType(required=False, serialized_name=api_constants.VIRTUAL_VOLUME_ID)
    volume_access_groups = types.ListType(
        field=types.IntIdType(),
        default=list,
        serialized_name=api_constants.VOLUME_ACCESS_GROUPS
    )
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)
    volume_pairs = types.ListType(types.ModelType(VolumePair), default=list,
                                  serialized_name=api_constants.VOLUME_PAIRS)

    # Version 10.0 Additions
    enable_snapmirror_replication = types.BooleanType(
        required=False,
        serialized_name=api_constants.ENABLE_SNAPMIRROR_REPLICATION,
        version_added=sf_version.SfVersion('10.0')
    )
    last_access_time = types.DateTimeType(
        serialized_name=api_constants.LAST_ACCESS_TIME,
        version_added=sf_version.SfVersion('10.0')
    )
    last_access_time_io = types.DateTimeType(
        serialized_name=api_constants.LAST_ACCESS_TIME_IO,
        version_added=sf_version.SfVersion('10.0')
    )
    qos_policy_id = types.IntIdType(
        serialized_name=api_constants.QOS_POLICY_ID,
        version_added=sf_version.SfVersion('10.0')
    )
    current_protection_scheme = types.EnumType(
        enums.ProtectionScheme,
        serialized_name=api_constants.CURRENT_PROTECTION_SCHEME,
        version_added=sf_version.SfVersion('11.0')
    )
    previous_protection_scheme = types.EnumType(
        enums.ProtectionScheme,
        serialized_name=api_constants.PREVIOUS_PROTECTION_SCHEME,
        version_added=sf_version.SfVersion('11.0')
    )
    volume_consistency_group_uuid = types.UUIDType(required=True,
                                                   serialized_name='volumeConsistencyGroupUUID',
                                                   version_added=sf_version.SfVersion('10.0'))
    volume_uuid = types.UUIDType(required=True,
                                 serialized_name='volumeUUID',
                                 version_added=sf_version.SfVersion('10.0'))

    @classmethod
    def create(cls, name=None, account_id=None, total_size=None, enable512e=False, qos=None,
               attributes=None, qos_policy_id=None, access=None, block_size=None, create_time=None,
               delete_time=None, iqn=None, purge_time=None, scsi_eui_device_id=None,
               scsi_naa_device_id=None, slice_count=None, status=None, virtual_volume_id=None,
               volume_access_groups=None, volume_id=None, volume_pairs=None,
               enable_snapmirror_replication=None, last_access_time=None, last_access_time_io=None,
               current_protection_scheme=None, previous_protection_scheme=None):
        """Define a volume profile to pass to create functions/methods.

        :param str name: Name of volume.
        :param int account_id: AccountID volume will belong to.
        :param sf_api.data_size.DataSize total_size: Size of volume.
        :param bool enable512e: If true enable 512 byte emulation.
        :param sf_api.models.Qos qos: QoS data model.
        :param dict attributes: Any additional key value pairs to be added to the Account.
        :param int qos_policy_id: QoSPolicyID from which to define QoS.
        :param sf_api.enums.VolumeAccess|str access: Access allowed for the volume
        :param int block_size: Size of blocks on the volume
        :param datetime.datetime|str create_time: Time volume was created
        :param datetime.datetime|str delete_time: Time volume was deleted
        :param str iqn: Volume iSCSI qualified name
        :param datetime.datetime|str purge_time: Time volume was purged
        :param str scsi_eui_device_id: Globally unique SCSI device ID in EUI-64 based 16 byte format
        :param str scsi_naa_device_id: Globally unique SCSI device ID in NAA IEEE format
        :param int slice_count: Number of slices on the volume
        :param sf_api.enums.VolumeStatus|str status: Current status of the volume
        :param str virtual_volume_id: Unique virtual volume ID associated with volume
        :param list[sf_api.models.VolumeAccessGroup] volume_access_groups: List of volume access groups
        :param int volume_id: Unique id for the volume
        :param list[sf_api.models.VolumePair] volume_pairs: Info about a paired volume
        :param bool enable_snapmirror_replication: Indicates whether to enable snapmirror replication
        :param datetime.datetime|str last_access_time: Last access time
        :param datetime.datetime|str last_access_time_io: Last access time IO
        :param sf_api.enums.ProtectionScheme|str current_protection_scheme: Current Protection Scheme
        :param sf_api.enums.ProtectionScheme|str previous_protection_scheme: Previous Protection Scheme
        :return: Partially defined volume data model to be used for volume creation.
        :rtype: sf_api.models.Volume
        """
        return super(Volume, cls).create(
            name=name,
            account_id=account_id,
            total_size=total_size,
            enable512e=enable512e,
            qos=qos,
            attributes=attributes,
            qos_policy_id=qos_policy_id,
            access=access,
            block_size=block_size,
            create_time=create_time,
            delete_time=delete_time,
            iqn=iqn,
            purge_time=purge_time,
            scsi_eui_device_id=scsi_eui_device_id,
            scsi_naa_device_id=scsi_naa_device_id,
            slice_count=slice_count,
            status=status,
            virtual_volume_id=virtual_volume_id,
            volume_access_groups=volume_access_groups,
            volume_id=volume_id,
            volume_pairs=volume_pairs,
            enable_snapmirror_replication=enable_snapmirror_replication,
            last_access_time=last_access_time,
            last_access_time_io=last_access_time_io,
            current_protection_scheme=current_protection_scheme,
            previous_protection_scheme=previous_protection_scheme
        )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_id', self.volume_id,
            'name', self.name
        )


# TODO: Uncomment old fields when model versioning is fixed. Jira Issue: VL-1062
class VolumeAccessGroup(sf_models.Base):
    """VolumeAccessGroup data model.

    Corresponds to the elements of the volume_access_groups object in the ListVolumeAccessGroups
    response.
    """

    attributes = types.DictType(types.BaseType())
    initiator_ids = types.ListType(types.IntIdType(), required=True, serialized_name='initiatorIDs')
    # initiators_5 = types.ListType(
    #     field=types.StringType(min_length=1, max_length=224),
    #     required=True,
    #     version_removed=sf_version.SfVersion('6.1')
    # )
    name = types.StringType(required=True)
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID
    )

    # Version 6.1 Additions
    fibre_channel_initiators = types.ListType(
        field=types.StringType(),
        required=True,
        serialized_name='fibreChannelInitiators',
        version_added=sf_version.SfVersion('6.1'),
        version_removed=sf_version.SfVersion('7.0')
    )
    iscsi_initiators = types.ListType(
        field=types.StringType(),
        required=True,
        serialized_name='iscsiInitiators',
        version_added=sf_version.SfVersion('6.1'),
        version_removed=sf_version.SfVersion('7.0')
    )

    # Version 7.0 Modifications

    initiators = types.ListType(
        field=types.StringType(min_length=1, max_length=224),
        required=True,
        serialized_name='initiators',
        version_added=sf_version.SfVersion('7.0')
    )

    # volumes and deleted_volumes are fields that have changed types a couple times.  This current API
    # uses a standard IntId.  So for the sake of keeping the API consistent with those alternative
    # inputs, we will provide the key that can pull the equivalent IntId data from that
    # non-compatible input source.
    volumes = types.ListType(
        field=types.IntIdTypeWithDictKeyConvert(convert_key=api_constants.VOLUME_ID),
        required=True,
        serialized_name='volumes',
    )
    deleted_volumes = types.ListType(
        field=types.IntIdTypeWithDictKeyConvert(convert_key=api_constants.VOLUME_ID),
        required=True,
        serialized_name='deletedVolumes',
    )

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_access_group_id', self.volume_access_group_id,
            'name', self.name
        )


class QuintileHistogram(sf_models.Base):
    """QuintileHistogram data model.

    Corresponds to the Histograms object in the ListVolumeQoSHistograms response.
    """
    bucket_0 = types.IntType(required=False, serialized_name=api_constants.BUCKET_0)
    bucket_1_to_19 = types.IntType(required=True, serialized_name=api_constants.BUCKET_1_TO_19)
    bucket_20_to_39 = types.IntType(required=True, serialized_name=api_constants.BUCKET_20_TO_39)
    bucket_40_to_59 = types.IntType(required=True, serialized_name=api_constants.BUCKET_40_TO_59)
    bucket_60_to_79 = types.IntType(required=True, serialized_name=api_constants.BUCKET_60_TO_79)
    bucket_80_to_100 = types.IntType(required=True, serialized_name=api_constants.BUCKET_80_TO_100)
    bucket_101_plus = types.IntType(required=False, serialized_name=api_constants.BUCKET_101_PLUS)

    def __repr__(self):
        return ('{}<{}={}, {}={}, {}={}, '
                '{}={}, {}={}, {}={}, {}={}>').format(
            self.__class__.__name__,
            api_constants.BUCKET_0, self.bucket_0,
            api_constants.BUCKET_1_TO_19, self.bucket_1_to_19,
            api_constants.BUCKET_20_TO_39, self.bucket_20_to_39,
            api_constants.BUCKET_40_TO_59, self.bucket_40_to_59,
            api_constants.BUCKET_60_TO_79, self.bucket_60_to_79,
            api_constants.BUCKET_80_TO_100, self.bucket_80_to_100,
            api_constants.BUCKET_101_PLUS, self.bucket_101_plus
        )


class BlockSizeHistogram(sf_models.Base):
    """BlockSizeHistogram data model.

    Corresponds to the Histograms object in the ListVolumeQoSHistograms response.
    """
    bucket_512_to_4095 = types.IntType(required=True,
                                       serialized_name=api_constants.BUCKET_512_TO_4095)
    bucket_4096_to_8191 = types.IntType(required=True,
                                        serialized_name=api_constants.BUCKET_4096_TO_8191)
    bucket_8192_to_16383 = types.IntType(required=True,
                                         serialized_name=api_constants.BUCKET_8192_TO_16383)
    bucket_16384_to_32767 = types.IntType(required=True,
                                          serialized_name=api_constants.BUCKET_16384_TO_32767)
    bucket_32768_to_65535 = types.IntType(required=True,
                                          serialized_name=api_constants.BUCKET_32768_TO_65535)
    bucket_65536_to_131071 = types.IntType(required=True,
                                           serialized_name=api_constants.BUCKET_65536_TO_131071)
    bucket_131072_plus = types.IntType(required=True,
                                       serialized_name=api_constants.BUCKET_131072_PLUS)

    def __repr__(self):
        return ('{}<{}={}, {}={}, {}={}, '
                '{}={}, {}={}, {}={}, {}={}>').format(
            self.__class__.__name__,
            api_constants.BUCKET_512_TO_4095, self.bucket_512_to_4095,
            api_constants.BUCKET_4096_TO_8191, self.bucket_4096_to_8191,
            api_constants.BUCKET_8192_TO_16383, self.bucket_8192_to_16383,
            api_constants.BUCKET_16384_TO_32767, self.bucket_16384_to_32767,
            api_constants.BUCKET_32768_TO_65535, self.bucket_32768_to_65535,
            api_constants.BUCKET_65536_TO_131071, self.bucket_65536_to_131071,
            api_constants.BUCKET_131072_PLUS, self.bucket_131072_plus
        )


class Histograms(sf_models.Base):
    """Histograms data model.

    Corresponds to the Histograms object in the ListVolumeQoSHistograms response.
    """
    below_min_iops_percentages = types.ModelType(
        QuintileHistogram,
        required=True,
        serialized_name=api_constants.BELOW_MIN_IOPS_PERCENTAGES)
    min_to_max_iops_percentages = types.ModelType(
        QuintileHistogram,
        required=True,
        serialized_name=api_constants.MIN_TO_MAX_IOPS_PERCENTAGES)
    target_utilization_percentages = types.ModelType(
        QuintileHistogram,
        required=True,
        serialized_name=api_constants.TARGET_UTILIZATION_PERCENTAGES)
    throttle_percentages = types.ModelType(
        QuintileHistogram,
        required=True,
        serialized_name=api_constants.THROTTLE_PERCENTAGES)
    write_block_sizes = types.ModelType(
        BlockSizeHistogram,
        required=True,
        serialized_name=api_constants.WRITE_BLOCK_SIZES)
    read_block_sizes = types.ModelType(
        BlockSizeHistogram,
        required=True,
        serialized_name=api_constants.READ_BLOCK_SIZES)

    def __repr__(self):
        return ('{}<{}={}, {}={}, {}={}, '
                '{}={}, {}={}, {}={}>').format(
            self.__class__.__name__,
            api_constants.BELOW_MIN_IOPS_PERCENTAGES, self.below_min_iops_percentages,
            api_constants.MIN_TO_MAX_IOPS_PERCENTAGES, self.min_to_max_iops_percentages,
            api_constants.TARGET_UTILIZATION_PERCENTAGES, self.target_utilization_percentages,
            api_constants.THROTTLE_PERCENTAGES, self.throttle_percentages,
            api_constants.WRITE_BLOCK_SIZES, self.write_block_sizes,
            api_constants.READ_BLOCK_SIZES, self.read_block_sizes
        )


class VolumeQosHistograms(sf_models.Base):
    """VolumeQosHistograms data model.

    Corresponds to the VolumeQosHistograms object in the ListVolumeQoSHistograms response.
    """
    timestamp = types.DateTimeType(required=True, serialized_name=api_constants.TIMESTAMP)
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)
    histograms = types.ModelType(Histograms, required=True,
                                 serialized_name=api_constants.HISTOGRAMS)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            api_constants.TIMESTAMP, self.timestamp,
            api_constants.VOLUME_ID, self.volume_id,
            api_constants.HISTOGRAMS, self.histograms
        )


class VolumeStats(sf_models.Base):
    """VolumeStats data model.

    Corresponds to the volume_stats object in the GetVolumeStats response.
    """
    account_id = types.IntIdType(required=True, serialized_name='accountID')
    actual_iops = types.IntType(required=False, serialized_name=api_constants.ACTUAL_IOPS)
    average_iop_size = types.IntType(required=False, serialized_name=api_constants.AVERAGE_IOP_SIZE)
    burst_iops_credit = types.IntType(required=False, serialized_name=api_constants.BURST_IOPS_CREDIT)
    client_queue_depth = types.IntType(required=False, serialized_name=api_constants.CLIENT_QUEUE_DEPTH)
    latency_usec = types.IntType(required=False, serialized_name=api_constants.LATENCY_USEC)
    non_zero_blocks = types.IntType(required=True, serialized_name='nonZeroBlocks')
    read_bytes = types.IntType(required=True, serialized_name='readBytes')
    read_latency_usec = types.IntType(required=False, serialized_name=api_constants.READ_LATENCY_USEC)
    read_ops = types.IntType(required=True, serialized_name='readOps')
    throttle = types.IntType(required=False)
    timestamp = types.DateTimeType(required=True)
    unaligned_reads = types.IntType(required=True, serialized_name=api_constants.UNALIGNED_READS)
    unaligned_writes = types.IntType(required=True, serialized_name=api_constants.UNALIGNED_WRITES)
    volume_access_groups = types.ListType(
        field=types.IntType(),
        required=True,
        serialized_name='volumeAccessGroups'
    )
    volume_id = types.IntIdType(required=True, serialized_name='volumeID')
    volume_size = types.IntType(required=True, serialized_name='volumeSize')
    volume_utilization = types.FloatType(required=False,
                                         serialized_name=api_constants.VOLUME_UTILIZATION)
    write_bytes = types.IntType(required=True, serialized_name='writeBytes')
    write_latency_usec = types.IntType(required=False, serialized_name=api_constants.WRITE_LATENCY_USEC)
    write_ops = types.IntType(required=True, serialized_name='writeOps')
    zero_blocks = types.IntType(required=True, serialized_name='zeroBlocks')

    # Version 6.0 Additions
    async_delay = types.IntType(serialized_name='asyncDelay', version_added=sf_version.SfVersion('6.0'))
    desired_metadata_hosts = types.ModelType(
        model_spec=MetadataHosts,
        serialized_name='desiredMetadataHosts',
        version_added=sf_version.SfVersion('6.0')
    )
    metadata_hosts = types.ModelType(
        model_spec=MetadataHosts,
        required=False,
        serialized_name='metadataHosts',
        version_added=sf_version.SfVersion('6.0')
    )

    # Version 9.0 Additions
    read_bytes_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.READ_BYTES_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )
    read_ops_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.READ_OPS_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )
    sample_period_msec = types.IntType(
        required=True,
        serialized_name=api_constants.SAMPLE_PERIOD_MSEC,
        version_added=sf_version.SfVersion('9.0')
    )
    write_bytes_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_BYTES_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )
    write_ops_last_sample = types.IntType(
        required=True,
        serialized_name=api_constants.WRITE_OPS_LAST_SAMPLE,
        version_added=sf_version.SfVersion('9.0')
    )

    # Version 10.0 Additions
    read_latency_usec_total = types.IntType(
        required=False,
        serialized_name=api_constants.READ_LATENCY_USEC_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )
    write_latency_usec_total = types.IntType(
        required=False,
        serialized_name=api_constants.WRITE_LATENCY_USEC_TOTAL,
        version_added=sf_version.SfVersion('10.0')
    )

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_id', self.volume_id,
            'volume_size', self.volume_size,
            'read_ops', self.read_ops,
            'write_ops', self.write_ops
        )


class StandbyServiceInfo(sf_models.Base):
    """Standby Service Info in slice data model."""
    failed_version = types.IntIdType(required=True, serialized_name='failedVersion')
    active_version = types.IntIdType(required=True, serialized_name='activeVersion')
    service_id = types.IntType(required=True, serialized_name='serviceID')
    state = types.EnumType(enums.StandbyStates, required=True, serialized_name='state')

    def __repr__(self):
        return '{}<service_id={}, state={}>'.format(self.__class__.__name__, self.service_id,
                                                    self.state)


class Slice(sf_models.Base):
    """Slice data model.

    Corresponds to the slice object in GetSliceInfo.
    """
    active_branch = types.IntIdType(required=True, serialized_name='activeBranch')
    active_branch_incarnation = types.IntIdType(required=True, serialized_name='activeBranchIncarnation')
    block_size = types.IntType(required=True, serialized_name=api_constants.BLOCK_SIZE)
    bulk_operations = types.ListType(
        field=types.IntIdType(), default=list, required=True, serialized_name='bulkOperations')
    clone_snapshots = types.ListType(
        field=types.IntIdType(), default=list, required=True, serialized_name='cloneSnapshots')
    current_protection_scheme = types.EnumType(
        enums.ProtectionScheme, required=True, serialized_name='currentProtectionScheme',
        version_added=sf_version.SfVersion('11.0'))
    dead_clone_jobs = types.DictType(types.BaseType(), required=True, serialized_name='deadCloneJobs')
    dead_secondaries = types.ListType(
        field=types.IntIdType(), default=list, required=True,
        serialized_name=api_constants.DEAD_SECONDARIES)
    desired_primary = types.IntIdType(required=True, serialized_name='desiredPrimary')
    desired_secondaries = types.ListType(
        field=types.IntIdType(), default=list, required=True, serialized_name='desiredSecondaries')
    incarnations = types.DictType(types.BaseType(), required=True, serialized_name='incarnations')
    is_new = types.BooleanType(required=True, serialized_name='isNew')
    is_snap_mirror_transfer_in_progress = types.BooleanType(
        required=True, serialized_name='isSnapMirrorTransferInProgress')
    live_secondaries = types.ListType(
        field=types.IntIdType(), required=True, serialized_name=api_constants.LIVE_SECONDARIES)
    number_external_snapshots = types.IntType(required=True, serialized_name='numberExternalSnapshots')
    primary = types.IntIdType(required=True, serialized_name=api_constants.PRIMARY)
    reassignment_type = types.StringType(required=True, serialized_name='reassignmentType')
    # TODO: remote rep info type
    remote_replication_info = types.DictType(
        types.BaseType(), required=True, serialized_name='remoteReplicationInfo')
    resize_incarnation = types.IntType(required=True, serialized_name='resizeIncarnation')
    resync_incarnation = types.IntType(required=True, serialized_name='resyncIncarnation')
    slice_id = types.IntIdType(required=True, serialized_name=api_constants.SLICE_ID)
    snapshots = types.DictType(types.BaseType(), required=True, serialized_name='snapshots')
    snapshots_in_progress = types.ListType(
        field=types.IntIdType(), required=True, serialized_name='snapshotsInProgress')
    snapshots_incarnation = types.IntIdType(required=True, serialized_name='snapshotsIncarnation')
    standby_services = types.ListType(
        field=types.ModelType(StandbyServiceInfo), required=True, serialized_name='standbyServices',
        version_added=sf_version.SfVersion('11.1'))
    total_size = types.IntType(required=True, serialized_name=api_constants.TOTAL_SIZE)
    version_bump_reasons = types.ListType(
        field=types.StringType(), required=True, serialized_name='versionBumpReasons')
    volume_id = types.IntIdType(required=True, serialized_name=api_constants.VOLUME_ID)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'volume_id', self.volume_id,
            'slice_id', self.slice_id,
            'primary', self.primary,
            'live_secondaries', self.live_secondaries
        )


class SliceAssignment(sf_models.Base):
    """Slice Assignment data model."""
    primary = types.IntIdType(required=True, serialized_name=api_constants.PRIMARY)
    secondaries = types.ListType(field=types.IntIdType(),
                                 required=True, serialized_name=api_constants.SECONDARIES)
    standby_services = types.ListType(field=types.IntType,
                                      required=True, serialized_name=u'standbyServices')

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'primary', self.primary,
            'secondaries', self.secondaries,
            'standbyServices', self.standby_services,
        )
