"""Tests that AsyncResult's PolyType field works."""
import unittest

from sf_api import (
    api_constants,
    enums,
    models,
)

drive_removal_get = {
    'createTime': '2017-06-27T08:55:32Z',
    'lastUpdateTime': '2017-06-27T08:56:46Z',
    'result': {
        'bsDisks': [],
        'ssDisks': []
    },
    'resultType': 'DriveRemoval',
    'status': 'complete'
}

drive_add_get = {
    'createTime': '2017-06-27T08:56:53Z',
    'lastUpdateTime': '2017-06-27T08:56:55Z',
    'result': {
        'abortedDrives': [],
        'addedDrives': [
            11
        ],
        'pendingDrives': []
    },
    'resultType': 'DriveAdd',
    'status': 'complete'
}


class TestAsyncResult(unittest.TestCase):
    """Test that the PolyType is working with various data."""

    def test_get_details_change(self):
        remove_drive_async_result = models.AsyncResult(drive_removal_get)
        assert isinstance(remove_drive_async_result.details, models.RemoveDrivesDetails)

        add_drive_async_result = models.AsyncResult(drive_add_get)
        assert isinstance(add_drive_async_result.details, models.AddDrivesDetails)

    def test_error_status_override(self):
        """Ensure that status becomes failed when the error key exists."""
        gc_details_dict = models.GarbageCollectionDetailsPreMg.get_mock_data()
        async_result_dict = models.AsyncResult.get_mock_data(
            overrides=dict(status=enums.AsyncResultStatus.complete)
        )
        async_result_dict.pop(api_constants.DETAILS)
        async_result_dict[api_constants.ERROR] = gc_details_dict
        async_result_dm = models.AsyncResult(async_result_dict)
        assert async_result_dm.status is enums.AsyncResultStatus.failed

    def test_clone_volume_details(self):
        mock_data = models.CloneVolumeDetails.get_mock_data()
        async_result = models.AsyncResult({'result': mock_data,
                                           'resultType': enums.AsyncResultType.clone})
        assert isinstance(async_result.details, models.CloneVolumeDetails)

    def test_group_clone_volume_details(self):
        mock_data = models.GroupCloneVolumesDetails.get_mock_data()
        async_result = models.AsyncResult({'result': mock_data,
                                           'resultType': enums.AsyncResultType.group_clone})
        assert isinstance(async_result.details, models.GroupCloneVolumesDetails)

    def test_add_nodes_details(self):
        mock_data = models.AddNodesDetails.get_mock_data()
        async_result = models.AsyncResult({'result': mock_data})
        assert isinstance(async_result.details, models.AddNodesDetails)

    def test_fill_slice_details(self):
        mock_data = models.FillSliceDetails.get_mock_data()
        async_result = models.AsyncResult({'result': mock_data,
                                           'resultType': enums.AsyncResultType.fill_slice})
        assert isinstance(async_result.details, models.FillSliceDetails)

    def test_garbage_collection_details_pre_mg(self):
        mock_data = models.GarbageCollectionDetailsPreMg.get_mock_data()
        async_result = models.AsyncResult({'result': mock_data,
                                           'resultType': enums.AsyncResultType.garbage_collection})
        assert isinstance(async_result.details, models.GarbageCollectionDetailsPreMg)

    def test_garbage_collection_details(self):
        mock_data = models.GarbageCollectionDetails.get_mock_data(include_optional=True)
        async_result = models.AsyncResult({'result': mock_data,
                                           'resultType': enums.AsyncResultType.garbage_collection})
        assert isinstance(async_result.details, models.GarbageCollectionDetails)

    def test_error_response(self):
        mock_data = models.ErrorResponse.get_mock_data(include_optional=True)
        async_result = models.AsyncResult({'result': mock_data, 'Error': 'error'})
        assert isinstance(async_result.details, models.ErrorResponse)

    def test_details_string(self):
        async_result = models.AsyncResult({'result': 'fake'})
        assert isinstance(async_result.details, models.DetailsString)

    def test_async_result_repr(self):
        for result in models.DETAILS_MODELS:
            mock_obj = result.get_mock_object()
            assert repr(mock_obj) is not None

    def test_async_already_dm(self):
        async_result = models.AsyncResult({'result': 'fake'})
        async_result_same = models.AsyncResult(async_result)
        assert isinstance(async_result_same.details, models.DetailsString)

    def test_start_bs_check_details(self):
        mock_data = models.BSCheckDetails.get_mock_data(include_optional=True)
        async_result = models.AsyncResult({'result': mock_data,
                                           'resultType': enums.AsyncResultType.bs_check})
        assert isinstance(async_result.details, models.BSCheckDetails)
