"""Request data models related to accounts."""
from sf_api import (
    api_constants,
    enums,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import _base


class AddAccount(_base.BaseClusterApi):
    """AddAccount request."""
    _version_added = sf_version.SfVersion('5.0')
    username = types.StringType(
        required=True,
        serialized_name=api_constants.USERNAME,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_account_name}
        }
    )
    initiator_secret = types.StringType(
        serialized_name=api_constants.INITIATOR_SECRET,
        min_length=12,
        max_length=16,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    target_secret = types.StringType(
        serialized_name=api_constants.TARGET_SECRET,
        min_length=12,
        max_length=16,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    attributes = types.DictType(
        field=types.StringType(),
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )

    def __init__(self, **kwargs):
        super(AddAccount, self).__init__(result_dm=result_models.AddAccount, **kwargs)
        self.add_teardown_api(RemoveAccount)


class GetAccountByID(_base.BaseClusterApi):
    """GetAccountByID request."""
    _version_added = sf_version.SfVersion('5.0')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_account,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.unknown_account},
        }
    )

    def __init__(self, **kwargs):
        super(GetAccountByID, self).__init__(result_dm=result_models.GetAccountById, **kwargs)
        self.add_setup_api(AddAccount)
        self.add_inspection_api(ListAccounts, list_all=True)


class GetAccountByName(_base.BaseClusterApi):
    """GetAccountByName request."""
    _version_added = sf_version.SfVersion('5.0')
    username = types.StringType(
        required=True,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.USERNAME],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.db_bad_arguments}
        }
    )

    def __init__(self, **kwargs):
        super(GetAccountByName, self).__init__(result_dm=result_models.GetAccountByName, **kwargs)
        self.add_setup_api(AddAccount)
        self.add_inspection_api(ListAccounts, list_all=True)


class GetAccountEfficiency(_base.BaseClusterApi):
    """GetAccountEfficiency request."""
    _version_added = sf_version.SfVersion('5.0')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetAccountEfficiency, self).__init__(
            result_dm=result_models.GetAccountEfficiency, **kwargs
        )
        self.add_setup_api(AddAccount)
        self.add_inspection_api(ListAccounts, list_all=True)


class ListAccounts(_base.BaseClusterApi):
    """ListAccounts request.

    :param bool list_all: Whether to exclude all constraint parameters (default: False).
    """
    _version_added = sf_version.SfVersion('5.0')
    start_account_id = types.IntIdType()
    limit = types.IntType()

    def __init__(self, list_all=False, **kwargs):
        super(ListAccounts, self).__init__(result_dm=result_models.ListAccounts, **kwargs)
        self.add_setup_api(AddAccount)

        if list_all:
            self.fields_to_exclude = [ListAccounts.limit, ListAccounts.start_account_id]


class ModifyAccount(_base.BaseClusterApi):
    """ModifyAccount request."""
    _version_added = sf_version.SfVersion('5.0')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_account,
                enums.ApiExceptionType.param_value_does_not_exist: enums.SfException.unknown_account}
        }
    )
    attributes = types.DictType(
        field=types.StringType(),
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    initiator_secret = types.StringType(
        serialized_name=api_constants.INITIATOR_SECRET,
        min_length=12,
        max_length=16,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    status = types.EnumType(
        enum_class=enums.AccountStatus,
        choices=[enums.AccountStatus.active, enums.AccountStatus.locked]
    )
    target_secret = types.StringType(
        serialized_name=api_constants.TARGET_SECRET,
        min_length=12,
        max_length=16,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    username = types.StringType(
        serialized_name=api_constants.USERNAME,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )

    def __init__(self, **kwargs):
        super(ModifyAccount, self).__init__(result_dm=result_models.ModifyAccount, **kwargs)
        self.add_setup_api(AddAccount)
        self.add_inspection_api(ListAccounts, list_all=True)


class RemoveAccount(_base.BaseClusterApi):
    """RemoveAccount request."""
    _version_added = sf_version.SfVersion('5.0')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(RemoveAccount, self).__init__(**kwargs)
        self.add_setup_api(AddAccount)
        self.add_inspection_api(ListAccounts, list_all=True)
