"""Request data models related to administration."""
from sf_api import (
    api_constants,
    enums,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import _base


class AddClusterAdmin(_base.BaseClusterApi):
    """AddClusterAdmin request."""
    _version_added = sf_version.SfVersion('5.0')
    access = types.ListType(
        field=types.EnumType(enums.ClusterAccessLevel),
        required=True,
        metadata={api_constants.NULL_INVALID: False}
    )
    accept_eula = types.BooleanType(
        required=True,
        choices=[True],
        serialized_name='acceptEula',
        metadata={
            api_constants.INVALID_CHOICES: {False: enums.SfException.param_not_valid},
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.param_not_valid
            }
        }
    )
    attributes = types.DictType(
        field=types.StringType(),
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    password = types.StringType(
        required=True,
        min_length=1,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.missing_param
            }
        }
    )
    username = types.StringType(
        required=True,
        min_length=1,
        max_length=1024,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_account_name
            }
        }
    )

    def __init__(self, **kwargs):
        super(AddClusterAdmin, self).__init__(result_dm=result_models.AddClusterAdmin, **kwargs)
        self.add_teardown_api(RemoveClusterAdmin)


class AddLdapClusterAdmin(_base.BaseClusterApi):
    """AddLdapClusterAdmin request."""
    _version_added = sf_version.SfVersion('8.0')
    access = types.ListType(
        field=types.EnumType(enums.ClusterAccessLevel),
        required=True,
        metadata={api_constants.NULL_INVALID: False}
    )
    accept_eula = types.BooleanType(
        required=True,
        choices=[True],
        serialized_name='acceptEula',
        metadata={
            api_constants.INVALID_CHOICES: {False: enums.SfException.param_not_valid},
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.param_not_valid
            }
        }
    )
    attributes = types.DictType(
        field=types.StringType(),
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    username = types.StringType(
        required=True,
        min_length=1,
        max_length=1024,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_account_name
            }
        }
    )

    def __init__(self, **kwargs):
        super(AddLdapClusterAdmin, self).__init__(result_dm=result_models.AddLdapClusterAdmin, **kwargs)
        self.add_teardown_api(RemoveClusterAdmin)


class GetCurrentClusterAdmin(_base.BaseClusterApi):
    """GetCurrentClusterAdmin request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetCurrentClusterAdmin, self).__init__(
            result_dm=result_models.GetCurrentClusterAdmin, **kwargs
        )
        self.add_setup_api(AddClusterAdmin)
        self.add_inspection_api(ListClusterAdmins)


class ListClusterAdmins(_base.BaseClusterApi):
    """GetCurrentClusterAdmin request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListClusterAdmins, self).__init__(result_dm=result_models.ListClusterAdmins, **kwargs)
        self.add_setup_api(AddClusterAdmin)


# NOTE: Changing 'access' of the primary cluster admin returns xPrimaryClusterAdmin
# NOTE: Changing 'password' of the primary cluster admin is allowed but breaks automation
class ModifyClusterAdmin(_base.BaseClusterApi):
    """ModifyClusterAdmin request."""
    _version_added = sf_version.SfVersion('5.0')
    access = types.ListType(types.EnumType(enums.ClusterAccessLevel))
    attributes = types.DictType(
        field=types.StringType(),
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    cluster_admin_id = types.IntIdType(
        required=True,
        serialized_name='clusterAdminID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['clusterAdminID'],
            api_constants.INVALID_CHOICES: {1: enums.SfException.primary_cluster_admin},
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.db_no_such_path}
        }
    )
    password = types.StringType()

    def __init__(self, **kwargs):
        super(ModifyClusterAdmin, self).__init__(**kwargs)
        self.add_setup_api(AddClusterAdmin)
        self.add_inspection_api(ListClusterAdmins)
        self.failing = True


class RemoveClusterAdmin(_base.BaseClusterApi):
    """RemoveClusterAdmin request."""
    _version_added = sf_version.SfVersion('5.0')
    cluster_admin_id = types.IntIdType(
        required=True,
        serialized_name='clusterAdminID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['clusterAdminID'],
            api_constants.INVALID_CHOICES: {1: enums.SfException.primary_cluster_admin},
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.db_no_such_path}
        }
    )

    def __init__(self, **kwargs):
        super(RemoveClusterAdmin, self).__init__(**kwargs)
        self.add_setup_api(AddClusterAdmin)
        self.add_inspection_api(ListClusterAdmins)
