"""Base request model from which others inherit."""
import abc
import logging
import time

import requests

from sf_api import (
    api_constants,
    api_errors,
    models,
)

_log = logging.getLogger(__name__)


class BaseApiRequest(models.Base):
    """Base API request class.

    :param Any result_dm: Model to use for validating API responses.
    :param str method: SolidFire API Method for API calls.
    :param dict[sf_api.enums.ApiExceptionType,sf_api.enums.SfException] exceptions: Default exception
        overrides for the indicated exception types.
    :param int timeout: Time in seconds before raising an exception while calling the API.
    :param int async_timeout: Time in seconds before raising an exception for asynchronous API.
    :param int async_poll_interval: Interval at which to poll the async handle for asynchronous API.
    :param bool wait_for_async_setup: Whether to wait for any async setup processes to complete before
        calling the API.
    :param dict[sf_api.models.types.BaseType,Any] loop_for_response: Response field conditions on which
        to loop before returning successfully.
    :param dict[sf_api.models.types.BaseType,Any] conditions: Response field conditions for a call to be
        considered successful.
    :param int node_requirement: Minimum number of nodes required for a successful API call (default: 1).
    :param bool skip_setup: Whether to skip calling setup API.
    """
    _version_added = None

    def __init__(self, result_dm=None, method=None, exceptions=None, timeout=None, async_timeout=None,
                 async_poll_interval=None, wait_for_async_setup=True, loop_for_response=None,
                 conditions=None, node_requirement=None, skip_setup=False):
        super(BaseApiRequest, self).__init__()
        self.result_dm = result_dm
        self.method = method or self.__class__.__name__
        self.exceptions = exceptions or {}
        self.timeout = timeout or 10
        self.async_timeout = async_timeout or 30
        self.async_poll_interval = async_poll_interval or 5
        self.wait_for_async_setup = wait_for_async_setup
        self.loop_for_response = loop_for_response
        self.conditions = conditions
        self.node_requirement = node_requirement or 1
        self._skip_setup = skip_setup

        self.fields_to_exclude = []
        self.fields_to_force = []
        self.field_overrides = {}

        self.setup_list = []
        self.inspection_list = []
        self.teardown_list = []

        # Parameters for categorizing API
        self.is_async = False
        self.cant_test = False
        self.failing = False  # TODO: Remove once all API models have been validated

    @property
    def level_2(self):
        """Whether this API is suitable for CICD level-2 testing.

        :rtype: bool
        """
        return (
                self.timeout < 30 and
                not self.is_async and
                not self.cant_test and
                not self.failing and
                all(api.level_2 for api in self.setup_list)
        )

    @property
    def level_3(self):
        """Whether this API is suitable for CICD level-3 testing.

        :rtype: bool
        """
        return not self.level_2 and not self.cant_test and not self.failing

    @abc.abstractmethod
    def call(self, api_server, params, version=None):
        """Call the API with the specified parameters at the provided server address.

        :param sf_api._base.BaseJsonRpcClient api_server: Server IP address.
        :param dict params: Parameters for the API request.
        :param sf_api.SfVersion version: The API endpoint against which to make the call.
        """

    def _call(self, api_server, params):
        """Call the API with the specified parameters.

        :param sf_api._base.BaseJsonRpcClient api_server: Server IP address.
        :param dict params: Parameters for the API request.
        :returns: The API response from the cluster.
        :rtype: dict
        """
        old_timeout = api_server.timeout
        api_server.timeout = (old_timeout[0], self.timeout)

        response = None
        result = None
        error = None
        num_retries = 0
        while num_retries <= api_server.retry_count:
            try:
                response = api_server.call_method_raw(self.method, params)
                result = response.get(api_constants.RESULT)
                error = response.get(api_constants.ERROR)
                break
            except (requests.ConnectionError, requests.HTTPError, requests.Timeout) as e:
                if num_retries == api_server.retry_count:
                    raise
                num_retries += 1
                retry_delay = getattr(e, 'retry_delay', 1)
                time.sleep(retry_delay)

        api_server.timeout = old_timeout
        if result is not None:
            _log.info('Received %s response: %s', self.method, result)
        if error is not None:
            code = error[api_constants.NAME]
            message = error[api_constants.MESSAGE]
            _log.info("Received %s response: %s:\"%s\"", self.method, code, message)
            exception_class = api_errors.get_exception_class_from_string(code)
            raise exception_class(
                msg='{}<{}>'.format(code, message),
                response=response
            )

        return result

    @classmethod
    def api_exists_in_version(cls, version):
        """Determine whether or not the API exists in the specified version.

        :param sf_api.SfVersion version: The SolidFire version to check.
        :returns: Whether the API exists in the specified version
        :rtype: bool
        """
        if cls._version_added is not None:
            if cls._version_added > version:
                return False
        return True

    def add_setup_api(self, request_dm, **kwargs):
        """Adds API required to setup the environment for successful API calls.

        :param Any request_dm: API request model.
        """
        if not self._skip_setup:
            self.setup_list.append(request_dm(**kwargs))

    def add_inspection_api(self, request_dm, **kwargs):
        """Adds API required for retrieving references for different parameters.

        :param Any request_dm: API request model.
        """
        self.inspection_list.append(request_dm(skip_setup=True, **kwargs))

    def add_teardown_api(self, request_dm, **kwargs):
        """Adds API required to teardown the effects of the called API.

        :param Any request_dm: API request model.
        """
        self.teardown_list.append(request_dm(skip_setup=True, **kwargs))


class BaseClusterApi(BaseApiRequest):
    """Base cluster API request class."""

    def call(self, api_server, params, version=None):
        """Call the API with the specified parameters at the specified cluster address.

        :param sf_api.ClusterApi api_server: The server against which to make API calls.
        :param dict params: Parameters for the API request.
        :param sf_api.SfVersion version: The version of the API to use.
        :returns: The API response from the cluster.
        :rtype: dict
        """
        _log.info('Sending %s request to cluster[%s]: %s', self.method, api_server.ip, params)
        return self._call(api_server, params)

    def __hash__(self):
        """Get the hash."""
        return hash('ClusterAPI_{}'.format(self.method))


class BaseNodeApi(BaseApiRequest):
    """Base node API request class."""

    def call(self, api_server, params, version=None):
        """Call the API with the specified parameters at the specified node address.

        :param sf_api.NodeApi api_server: The server against which to make API calls.
        :param dict params: Parameters for the API request.
        :param sf_api.SfVersion version: The version of the API to use.
        :returns: The API response from the cluster.
        :rtype: dict
        """
        _log.info('Sending %s request to node[%s]: %s', self.method, api_server.ip, params)
        return self._call(api_server, params)

    def __hash__(self):
        """Get the hash."""
        return hash('NodeAPI_{}'.format(self.method))
