"""Request data models related to clusters."""
from sf_api import (
    api_constants,
    enums,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import _base


class AddNodes(_base.BaseClusterApi):
    """AddNodes request."""
    _version_added = sf_version.SfVersion('5.0')
    pending_nodes = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name=api_constants.PENDING_NODES,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.NODE_ID]}
    )

    def __init__(self, **kwargs):
        super(AddNodes, self).__init__(result_dm=result_models.AddNodes, node_requirement=4, **kwargs)
        self.add_inspection_api(ListPendingNodes)
        self.failing = True


class CheckProposedNodeAdditions(_base.BaseClusterApi):
    """CheckProposedNodeAdditions request."""
    _version_added = sf_version.SfVersion('11.0')
    nodes = types.ListType(
        field=types.IPAddressType(),
        required=True,
        min_size=1,
        metadata={
            api_constants.ORPHAN_NODES_REQUIRED: True,
            api_constants.EXCEPTION_OVERRIDES: {enums.ApiExceptionType.param_value_invalid: None}
        }
    )

    def __init__(self, **kwargs):
        super(CheckProposedNodeAdditions, self).__init__(
            result_dm=result_models.CheckProposedNodeAdditions, timeout=35, **kwargs
        )


class ClearClusterFaults(_base.BaseClusterApi):
    """ClearClusterFaults request."""
    _version_added = sf_version.SfVersion('5.0')
    fault_types = types.EnumType(
        enum_class=enums.ClusterFaultStatus,
        serialized_name=api_constants.FAULT_TYPES,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.invalid_api_param}
        }
    )


class DemoteClusterMaster(_base.BaseClusterApi):
    """DemoteClusterMaster request."""
    _version_added = sf_version.SfVersion('8.0')
    node_id = types.IntIdType(
        serialized_name=api_constants.NODE_ID,
        version_added=sf_version.SfVersion('10.3'),
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.NODE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param
            }
        }
    )
    service_id = types.IntIdType(
        serialized_name=api_constants.SERVICE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ASSOCIATED_MASTER_SERVICE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param
            }
        }
    )

    # Define interdependent parameter relations
    node_id.metadata[api_constants.CONFLICTING_FIELDS] = [service_id]
    service_id.metadata[api_constants.CONFLICTING_FIELDS] = [node_id]

    def __init__(self, **kwargs):
        super(DemoteClusterMaster, self).__init__(
            result_dm=result_models.DemoteClusterMaster, node_requirement=2, **kwargs
        )
        self.add_inspection_api(ListAllNodes)


class EnableFeature(_base.BaseClusterApi):
    """EnableFeature request.

    :param bool vvols: Whether to override the request to enable the vvols feature.
    """
    _version_added = sf_version.SfVersion('5.0')
    # Don't randomly choose fips_drives because it requires EAR to be enabled first.
    # Don't randomly choose fips because This API depends on specific hardware
    # noinspection PyTypeChecker
    _sf_feature = set(enums.SfFeature)
    feature = types.EnumType(
        enum_class=enums.SfFeature,
        required=True,
        choices=[f for f in _sf_feature if f not in (enums.SfFeature.fips_drives, enums.SfFeature.fips)],
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.unknown_exception,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.unknown_enum_type}
        }
    )

    def __init__(self, vvols=False, **kwargs):
        super(EnableFeature, self).__init__(**kwargs)
        if vvols:
            self.field_overrides[EnableFeature.feature] = enums.SfFeature.vvols.value


class GetClusterCapacity(_base.BaseClusterApi):
    """GetClusterCapacity request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterCapacity, self).__init__(result_dm=result_models.GetClusterCapacity, **kwargs)


class GetClusterFullThreshold(_base.BaseClusterApi):
    """GetClusterFullThreshold request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterFullThreshold, self).__init__(
            result_dm=result_models.GetClusterFullThreshold, **kwargs
        )


class GetClusterHardwareInfo(_base.BaseClusterApi):
    """GetClusterHardwareInfo request."""
    _version_added = sf_version.SfVersion('5.0')
    type = types.EnumType(
        enum_class=enums.ClusterHardwareInfoType,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.invalid_api_param}
        }
    )

    def __init__(self, **kwargs):
        super(GetClusterHardwareInfo, self).__init__(
            result_dm=result_models.GetClusterHardwareInfo, **kwargs
        )
        self.failing = True


class GetClusterInfo(_base.BaseClusterApi):
    """GetClusterInfo request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterInfo, self).__init__(result_dm=result_models.GetClusterInfo, **kwargs)


class GetClusterMasterNodeID(_base.BaseClusterApi):
    """GetClusterMasterNodeID request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterMasterNodeID, self).__init__(
            result_dm=result_models.GetClusterMasterNodeID, **kwargs
        )


class GetClusterStats(_base.BaseClusterApi):
    """GetClusterStats request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterStats, self).__init__(result_dm=result_models.GetClusterStats, **kwargs)


class GetClusterVersionInfo(_base.BaseClusterApi):
    """GetClusterVersionInfo request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterVersionInfo, self).__init__(
            result_dm=result_models.GetClusterVersionInfo, **kwargs
        )


class GetFeatureStatus(_base.BaseClusterApi):
    """GetFeatureStatus request."""
    _version_added = sf_version.SfVersion('5.0')
    feature = types.EnumType(
        enum_class=enums.SfFeature,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.unknown_exception,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.unknown_enum_type}
        }
    )

    def __init__(self, **kwargs):
        super(GetFeatureStatus, self).__init__(result_dm=result_models.GetFeatureStatus, **kwargs)


class GetLoginSessionInfo(_base.BaseClusterApi):
    """GetLoginSessionInfo request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetLoginSessionInfo, self).__init__(
            result_dm=result_models.GetLoginSessionInfo, **kwargs
        )


class GetNodeHardwareInfo(_base.BaseClusterApi):
    """GetNodeHardwareInfo request."""
    _version_added = sf_version.SfVersion('5.0')
    node_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.NODE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.NODE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.node_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetNodeHardwareInfo, self).__init__(
            result_dm=result_models.GetNodeHardwareInfo, **kwargs
        )
        self.add_inspection_api(ListAllNodes)
        self.failing = True


class GetNodeStats(_base.BaseClusterApi):
    """GetNodeStats request."""
    _version_added = sf_version.SfVersion('5.0')
    node_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.NODE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.NODE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.node_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetNodeStats, self).__init__(result_dm=result_models.GetNodeStats, **kwargs)
        self.add_inspection_api(ListAllNodes)


class GetRawStats(_base.BaseClusterApi):
    """GetRawStats request."""
    _version_added = sf_version.SfVersion('10.0')

    def __init__(self, **kwargs):
        super(GetRawStats, self).__init__(result_dm=result_models.GetRawStats, **kwargs)
        self.failing = True


class ListActiveNodes(_base.BaseClusterApi):
    """ListActiveNodes request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListActiveNodes, self).__init__(result_dm=result_models.ListActiveNodes, **kwargs)


class ListAllNodes(_base.BaseClusterApi):
    """ListAllNodes request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListAllNodes, self).__init__(result_dm=result_models.ListAllNodes, **kwargs)


class ListClusterFaults(_base.BaseClusterApi):
    """ListClusterFaults request."""
    _version_added = sf_version.SfVersion('5.0')
    best_practices = types.BooleanType(serialized_name='bestPractices')
    fault_types = types.EnumType(
        enum_class=enums.ClusterFaultStatus,
        serialized_name=api_constants.FAULT_TYPES,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.invalid_api_param}
        }
    )

    def __init__(self, **kwargs):
        super(ListClusterFaults, self).__init__(result_dm=result_models.ListClusterFaults, **kwargs)


class ListEvents(_base.BaseClusterApi):
    """ListEvents request."""
    _version_added = sf_version.SfVersion('5.0')
    start_event_id = types.IntIdType(serialized_name='startEventID')
    end_event_id = types.IntIdType(serialized_name='endEventID')
    max_events = types.IntType(serialized_name='maxEvents')

    start_event_id.metadata[api_constants.MAX_LIMIT_FIELD] = end_event_id
    end_event_id.metadata[api_constants.MIN_LIMIT_FIELD] = start_event_id

    def __init__(self, **kwargs):
        super(ListEvents, self).__init__(result_dm=result_models.ListEvents, **kwargs)


class ListNodeStats(_base.BaseClusterApi):
    """ListNodeStats request."""

    def __init__(self, **kwargs):
        super(ListNodeStats, self).__init__(
            result_dm=result_models.ListNodeStats, timeout=30, **kwargs
        )


class ListISCSISessions(_base.BaseClusterApi):
    """ListISCSISessions request."""

    def __init__(self, **kwargs):
        super(ListISCSISessions, self).__init__(result_dm=result_models.ListISCSISessions, **kwargs)


class ListServices(_base.BaseClusterApi):
    """ListServices request."""

    def __init__(self, **kwargs):
        super(ListServices, self).__init__(result_dm=result_models.ListServices, **kwargs)


class ListPendingNodes(_base.BaseClusterApi):
    """ListPendingNodes request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListPendingNodes, self).__init__(
            result_dm=result_models.ListPendingNodes, node_requirement=4, **kwargs
        )


class ListPendingActiveNodes(_base.BaseClusterApi):
    """ListPendingActiveNodes request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListPendingActiveNodes, self).__init__(
            result_dm=result_models.ListPendingActiveNodes, node_requirement=4, **kwargs
        )


class ModifyClusterFullThreshold(_base.BaseClusterApi):
    """ModifyClusterFullThreshold request."""
    _version_added = sf_version.SfVersion('5.0')
    stage2_aware_threshold = types.IntType(
        required=True,
        serialized_name='stage2AwareThreshold',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_MIN: result_models.GetClusterFullThreshold.stage3_low_threshold,
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param}
        }
    )
    stage3_block_threshold_percent = types.IntType(
        required=True,
        max_value=100,
        serialized_name='stage3BlockThresholdPercent',
        metadata={
            api_constants.NULL_INVALID: False,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param}
        }
    )
    max_metadata_over_provision_factor = types.IntType(
        required=True,
        min_value=1,
        max_value=100,
        serialized_name='maxMetadataOverProvisionFactor',
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value}
        }
    )

    # Define interdependent parameter relations
    stage2_aware_threshold.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [
        stage3_block_threshold_percent,
        max_metadata_over_provision_factor
    ]
    stage3_block_threshold_percent.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [
        stage2_aware_threshold,
        max_metadata_over_provision_factor
    ]
    max_metadata_over_provision_factor.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [
        stage2_aware_threshold,
        stage3_block_threshold_percent
    ]

    def __init__(self, **kwargs):
        super(ModifyClusterFullThreshold, self).__init__(
            result_dm=result_models.ModifyClusterFullThreshold, **kwargs
        )
        self.add_inspection_api(GetClusterFullThreshold)
        self.failing = True


class PromoteClusterMaster(_base.BaseClusterApi):
    """PromoteClusterMaster request."""
    _version_added = sf_version.SfVersion('10.1')
    node_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.NODE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.NODE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param,
                enums.ApiExceptionType.param_null: enums.SfException.invalid_param,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param
            }
        }
    )
    service_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.SERVICE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ASSOCIATED_MASTER_SERVICE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param,
                enums.ApiExceptionType.param_null: enums.SfException.invalid_param,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param
            }
        }
    )

    # Define interdependent parameter relations
    node_id.metadata[api_constants.CONFLICTING_FIELDS] = [service_id]
    service_id.metadata[api_constants.CONFLICTING_FIELDS] = [node_id]

    def __init__(self, **kwargs):
        super(PromoteClusterMaster, self).__init__(result_dm=result_models.PromoteClusterMaster,
                                                   **kwargs)
        self.add_inspection_api(ListAllNodes)
        self.failing = True


class RemoveNodes(_base.BaseClusterApi):
    """RemoveNodes request."""
    _version_added = sf_version.SfVersion('5.0')
    nodes = types.ListType(
        field=types.IntIdType(),
        required=True,
        metadata={
            api_constants.NULL_INVALID: False,
            api_constants.REF_PARAMS: [api_constants.NODE_ID]
        }
    )

    def __init__(self, **kwargs):
        super(RemoveNodes, self).__init__(node_requirement=4, **kwargs)
        self.add_inspection_api(ListActiveNodes)
        self.failing = True


class SetLoginSessionInfo(_base.BaseClusterApi):
    """SetLoginSessionInfo request."""
    _version_added = sf_version.SfVersion('5.0')
    timeout = types.StringType(
        regex=r'^\d\d:\d\d:(\d\d){2,4}$',
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_param_type}
        }
    )
