"""Request data models related to drives."""
from sf_api import (
    api_constants,
    enums,
    models,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import _base


class Drive(models.Base):
    """Drive data model.

    Corresponds to the elements of the drives object in the AddDrives request.
    """
    drive_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.DRIVE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.DRIVE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.drive_id_does_not_exist}
        }
    )
    type = types.EnumType(
        enum_class=enums.DriveType,
        required=True,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.TYPE],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.unknown_enum_type}
        }
    )


class AddDrives(_base.BaseClusterApi):
    """AddDrives request."""
    _version_added = sf_version.SfVersion('5.0')
    drives = types.ListType(
        field=types.ModelType(models.Drive),
        required=True,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_CONDITIONS: {models.Drive.status: enums.DriveStatus.available.value},
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_param}
        }
    )
    force_during_bin_sync = types.BooleanType(
        serialized_name=api_constants.FORCE_DURING_BIN_SYNC,
        metadata={api_constants.CANT_TEST: True}
    )
    force_during_upgrade = types.BooleanType(serialized_name=api_constants.FORCE_DURING_UPGRADE)
    ignore_unbalanced_cluster_check = types.BooleanType(
        serialized_name=api_constants.IGNORE_UNBALANCED_CLUSTER_CHECK
    )
    ignore_usable_capacity_check = types.BooleanType(
        serialized_name=api_constants.IGNORE_USABLE_CAPACITY_CHECK
    )

    def __init__(self, **kwargs):
        super(AddDrives, self).__init__(result_dm=result_models.AddDrives, async_timeout=600, **kwargs)
        self.add_setup_api(RemoveDrives)
        self.add_inspection_api(ListDrives)
        self.field_overrides[AddDrives.force_during_bin_sync] = True
        self.is_async = True


class GetDriveHardwareInfo(_base.BaseClusterApi):
    """GetDriveHardwareInfo request."""
    _version_added = sf_version.SfVersion('5.0')
    drive_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.DRIVE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.DRIVE_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.drive_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetDriveHardwareInfo, self).__init__(
            result_dm=result_models.GetDriveHardwareInfo, timeout=60, **kwargs
        )
        self.add_inspection_api(ListDrives)


class GetDriveStats(_base.BaseClusterApi):
    """GetDriveStats request."""
    _version_added = sf_version.SfVersion('5.0')
    drive_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.DRIVE_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.DRIVE_ID],
            api_constants.REF_CONDITIONS: {models.Drive.status: enums.DriveStatus.active.value},
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.drive_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetDriveStats, self).__init__(result_dm=result_models.GetDriveStats, **kwargs)
        self.add_inspection_api(ListDrives)
        self.failing = True


class ListDrives(_base.BaseClusterApi):
    """ListDrives request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListDrives, self).__init__(result_dm=result_models.ListDrives, **kwargs)


class ListDriveStats(_base.BaseClusterApi):
    """ListDriveStats request."""
    _version_added = sf_version.SfVersion('5.0')
    drives = types.ListType(
        field=types.IntIdType(),
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.DRIVE_ID]
        }
    )

    def __init__(self, **kwargs):
        super(ListDriveStats, self).__init__(result_dm=result_models.ListDriveStats, **kwargs)
        self.add_inspection_api(ListDrives)
        self.failing = True


class RemoveDrives(_base.BaseClusterApi):
    """RemoveDrives request."""
    _version_added = sf_version.SfVersion('5.0')
    drives = types.ListType(
        field=types.IntIdType(),
        required=True,
        max_size=8,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.DRIVE_ID],
            api_constants.REF_CONDITIONS: {
                models.Drive.type: enums.DriveType.block.value,
                models.Drive.status: enums.DriveStatus.active.value,
            },
            api_constants.REF_MAX: 2,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.too_few_drives}
        }
    )
    force_during_bin_sync = types.BooleanType(
        serialized_name=api_constants.FORCE_DURING_BIN_SYNC,
        metadata={api_constants.CANT_TEST: True}
    )
    force_during_upgrade = types.BooleanType(serialized_name=api_constants.FORCE_DURING_UPGRADE)
    ignore_unbalanced_cluster_check = types.BooleanType(
        serialized_name=api_constants.IGNORE_UNBALANCED_CLUSTER_CHECK
    )

    def __init__(self, **kwargs):
        super(RemoveDrives, self).__init__(
            result_dm=result_models.RemoveDrives, async_timeout=3600, **kwargs
        )
        self.add_inspection_api(ListDrives)
        self.add_teardown_api(AddDrives)
        self.field_overrides[RemoveDrives.force_during_bin_sync] = True
        self.is_async = True


class SecureEraseDrives(_base.BaseClusterApi):
    """SecureEraseDrives request."""
    _version_added = sf_version.SfVersion('5.0')
    drives = types.ListType(
        field=types.IntIdType(),
        required=True,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.DRIVE_ID],
            api_constants.REF_CONDITIONS: {models.Drive.status: enums.DriveStatus.available.value}
        }
    )

    def __init__(self, **kwargs):
        super(SecureEraseDrives, self).__init__(result_dm=result_models.SecureEraseDrives, **kwargs)
        self.add_setup_api(RemoveDrives)
        self.add_inspection_api(ListDrives)

        # Can't test without a way to ensure no bin syncing is occurring
        self.cant_test = True
