"""Request data models related to fibre channel."""
from sf_api import (
    api_constants,
    enums,
    models,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import (
    _base,
    _volume,
)


class GetVolumeAccessGroupLunAssignments(_base.BaseClusterApi):
    """GetVolumeAccessGroupLunAssignments request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.volume_access_group_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetVolumeAccessGroupLunAssignments, self).__init__(
            result_dm=result_models.GetVolumeAccessGroupLunAssignments, **kwargs
        )
        self.add_setup_api(_volume.CreateVolumeAccessGroup)
        self.add_inspection_api(_volume.ListVolumeAccessGroups)


class ListFibreChannelPortInfo(_base.BaseClusterApi):
    """ListFibreChannelPortInfo request."""
    _version_added = sf_version.SfVersion('5.0')
    force = types.BooleanType()

    def __init__(self, **kwargs):
        super(ListFibreChannelPortInfo, self).__init__(
            result_dm=result_models.ListFibreChannelPortInfo, **kwargs
        )


class ListFibreChannelSessions(_base.BaseClusterApi):
    """ListFibreChannelSessions request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListFibreChannelSessions, self).__init__(
            result_dm=result_models.ListFibreChannelSessions, **kwargs
        )


class ListNodeFibreChannelPortInfo(_base.BaseNodeApi):
    """ListNodeFibreChannelPortInfo request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListNodeFibreChannelPortInfo, self).__init__(
            result_dm=result_models.ListNodeFibreChannelPortInfo, **kwargs
        )

        # Missing "nodes" field in response
        self.failing = True


class LunAssignment(models.Base):
    """LunAssignment data model.

    Corresponds to the elements of the lun_assignments object in the
    ModifyVolumeAccessGroupLunAssignments request.
    """
    lun = types.IntIdType(
        max_value=16383,
        metadata={api_constants.EXCEPTION_OVERRIDES: {
            enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_lun}
        }
    )
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID, api_constants.VOLUMES],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.not_in_volume_access_group}

        }
    )

    def __repr__(self):
        """Represent the field in the model hierarchy."""
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'lun', self.lun,
            'volume_id', self.volume_id
        )


class ModifyVolumeAccessGroupLunAssignments(_base.BaseClusterApi):
    """ModifyVolumeAccessGroupLunAssignments request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.volume_access_group_id_does_not_exist}
        }
    )
    lun_assignments = types.ListType(
        field=types.ModelType(LunAssignment),
        required=True,
        serialized_name='lunAssignments',
        metadata={api_constants.REF_REQUIRED: True, api_constants.NULL_INVALID: False,
                  api_constants.EXCEPTION_OVERRIDES: {
                      enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
                  }
    )

    def __init__(self, **kwargs):
        super(ModifyVolumeAccessGroupLunAssignments, self).__init__(
            result_dm=result_models.ModifyVolumeAccessGroupLunAssignments, **kwargs
        )
        self.add_setup_api(_volume.CreateVolumeAccessGroup, force_volumes=True, vol_count=1)
        self.add_inspection_api(_volume.ListVolumeAccessGroups, list_all=True)
