"""Request data models related to LDAP."""
from sf_api import (
    api_constants,
    enums,
    models,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import _base


class AddLdapClusterAdmin(_base.BaseClusterApi):
    """AddLdapClusterAdmin request."""
    _version_added = sf_version.SfVersion('5.0')
    username = types.StringType(
        required=True,
        min_length=1,
        max_length=1024,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_account_name,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_account_name}
        }
    )
    accept_eula = types.BooleanType(
        required=True,
        serialized_name='acceptEula',
        metadata={api_constants.CANT_TEST: True}
    )
    access = types.ListType(
        field=types.EnumType(enums.ClusterAccessLevel),
        required=True,
        metadata={api_constants.NULL_INVALID: False}
    )
    attributes = types.DictType(
        field=types.StringType(),
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )

    def __init__(self, **kwargs):
        super(AddLdapClusterAdmin, self).__init__(**kwargs)
        self.field_overrides[AddLdapClusterAdmin.accept_eula] = True


class DisableLdapAuthentication(_base.BaseClusterApi):
    """DisableLdapAuthentication request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(DisableLdapAuthentication, self).__init__(**kwargs)
        self.add_setup_api(EnableLdapAuthenticationSearchAndBind)


class EnableLdapAuthenticationBase(_base.BaseClusterApi):
    """EnableLdapAuthentication request base class."""
    _version_added = sf_version.SfVersion('5.0')
    auth_type = types.EnumType(
        enum_class=enums.LdapAuthType,
        serialized_name='authType',
        metadata={api_constants.CANT_TEST: True}
    )
    group_search_base_dn = types.StringType(serialized_name='groupSearchBaseDN')
    group_search_type = types.EnumType(
        enum_class=enums.LdapSearchType,
        serialized_name='groupSearchType',
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.unknown_exception}
        }
    )
    server_uris = types.ListType(
        field=types.StringType(regex=r'^ldap://\d.\d.\d.\d(:\d+)?$'),
        required=True,
        min_size=1,
        max_size=4,
        serialized_name='serverURIs',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )

    def __init__(self, **kwargs):
        super(EnableLdapAuthenticationBase, self).__init__(method='EnableLdapAuthentication', **kwargs)
        self.add_teardown_api(DisableLdapAuthentication)


class EnableLdapAuthenticationSearchAndBind(EnableLdapAuthenticationBase):
    """EnableLdapAuthentication request with SearchAndBind authType (default)."""
    user_search_base_dn = types.StringType(
        required=True,
        min_length=1,
        serialized_name='userSearchBaseDN',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value}
        }
    )
    search_bind_dn = types.StringType(
        required=True,
        min_length=1,
        serialized_name='searchBindDN',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value}
        }
    )
    search_bind_password = types.StringType(
        required=True,
        min_length=1,
        serialized_name='searchBindPassword',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value}
        }
    )
    user_search_filter = types.StringType(serialized_name='userSearchFilter')

    def __init__(self, **kwargs):
        super(EnableLdapAuthenticationSearchAndBind, self).__init__(**kwargs)
        self.field_overrides[
            EnableLdapAuthenticationSearchAndBind.auth_type
        ] = enums.LdapAuthType.search_and_bind.value


class EnableLdapAuthenticationDirectBind(EnableLdapAuthenticationBase):
    """EnableLdapAuthentication request with DirectBind authType."""
    user_dn_template = types.StringType(
        required=True,
        min_length=1,
        serialized_name='userDNTemplate',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value}
        }
    )

    def __init__(self, **kwargs):
        super(EnableLdapAuthenticationDirectBind, self).__init__(**kwargs)
        self.fields_to_force.append(EnableLdapAuthenticationDirectBind.auth_type)
        self.field_overrides[
            EnableLdapAuthenticationDirectBind.auth_type
        ] = enums.LdapAuthType.direct_bind.value


class GetLdapConfiguration(_base.BaseClusterApi):
    """GetLdapConfiguration request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetLdapConfiguration, self).__init__(
            result_dm=result_models.GetLdapConfiguration, **kwargs
        )


class TestLdapAuthentication(_base.BaseClusterApi):
    """TestLdapAuthentication request."""
    _version_added = sf_version.SfVersion('5.0')
    username = types.StringType(required=True)
    password = types.StringType(required=True)
    ldap_configuration = types.ModelType(models.LdapConfig, serialized_name='ldapConfiguration')

    def __init__(self, **kwargs):
        super(TestLdapAuthentication, self).__init__(
            result_dm=result_models.TestLdapAuthentication, **kwargs
        )
        # Requires LDAP credentials which are not available via the API
        self.cant_test = True
