"""Request data models related to nodes."""
from sf_api import (
    api_constants,
    enums,
    models,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import (
    _base,
    _cluster,
    _drive,
)


class CheckProposedCluster(_base.BaseNodeApi):
    """CheckProposedCluster request."""
    _version_added = sf_version.SfVersion('11.0')
    nodes = types.ListType(
        field=types.IPAddressType(),
        required=True,
        min_size=1,
        metadata={
            api_constants.ORPHAN_NODES_REQUIRED: True,
            api_constants.EXCEPTION_OVERRIDES: {enums.ApiExceptionType.param_value_invalid: None}
        }
    )

    def __init__(self, **kwargs):
        super(CheckProposedCluster, self).__init__(
            result_dm=result_models.CheckProposedCluster, timeout=35, **kwargs
        )


class CreateClusterSupportBundle(_base.BaseNodeApi):
    """CreateClusterSupportBundle request."""
    _version_added = sf_version.SfVersion('5.0')
    allow_incomplete = types.BooleanType(serialized_name='allowIncomplete')
    bundle_name = types.StringType(serialized_name='bundleName')
    extra_args = types.StringType(serialized_name='extraArgs')
    mvip = types.IPv4Type(required=True)
    nodes = types.ListType(types.IPv4Type(), required=True)
    password = types.StringType(required=True)
    username = types.StringType(required=True)

    # Define interdependent parameter relations
    mvip.metadata[api_constants.CONFLICTING_FIELDS] = [nodes]
    nodes.metadata[api_constants.CONFLICTING_FIELDS] = [mvip]

    def __init__(self, **kwargs):
        super(CreateClusterSupportBundle, self).__init__(
            result_dm=result_models.CreateClusterSupportBundle, **kwargs
        )

        # APINotPermitted errors
        self.failing = True


class CreateSupportBundle(_base.BaseNodeApi):
    """CreateSupportBundle request."""
    _version_added = sf_version.SfVersion('5.0')
    bundle_name = types.StringType(serialized_name='bundleName')
    extra_args = types.StringType(serialized_name='extraArgs')
    timeout_sec = types.IntType(serialized_name='timeoutSec', default=1500)

    def __init__(self, **kwargs):
        super(CreateSupportBundle, self).__init__(
            result_dm=result_models.CreateSupportBundle, timeout=600, **kwargs
        )
        self.add_teardown_api(DeleteAllSupportBundles)


class DeleteAllSupportBundles(_base.BaseNodeApi):
    """DeleteAllSupportBundles request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(DeleteAllSupportBundles, self).__init__(**kwargs)
        self.add_setup_api(CreateSupportBundle)


class GetClusterConfig(_base.BaseNodeApi):
    """GetClusterConfig request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterConfig, self).__init__(result_dm=result_models.GetClusterConfig, **kwargs)


class GetClusterState(_base.BaseNodeApi):
    """GetClusterState request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetClusterState, self).__init__(result_dm=result_models.GetClusterState, **kwargs)


class GetConfig(_base.BaseNodeApi):
    """GetConfig request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetConfig, self).__init__(result_dm=result_models.GetConfig, **kwargs)
        self.failing = True


class GetDriveConfig(_base.BaseNodeApi):
    """GetDriveConfig request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetDriveConfig, self).__init__(result_dm=result_models.GetDriveConfig, **kwargs)
        self.failing = True


class GetNodeFipsDrivesReport(_base.BaseNodeApi):
    _version_added = sf_version.SfVersion('11.5')

    def __init__(self, **kwargs):
        super(GetNodeFipsDrivesReport, self).__init__(
                result_dm=result_models.GetNodeFipsDrivesReport, **kwargs
        )


class GetHardwareConfig(_base.BaseNodeApi):
    """GetHardwareConfig request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetHardwareConfig, self).__init__(result_dm=result_models.GetHardwareConfig, **kwargs)
        self.failing = True


class GetHardwareInfo(_base.BaseNodeApi):
    """GetHardwareInfo request."""
    _version_added = sf_version.SfVersion('5.0')
    force = types.BooleanType()

    def __init__(self, **kwargs):
        super(GetHardwareInfo, self).__init__(result_dm=result_models.GetHardwareInfo, **kwargs)
        self.failing = True


class GetIpmiConfig(_base.BaseNodeApi):
    """GetIpmiConfig request."""
    _version_added = sf_version.SfVersion('5.0')
    chassis_type = types.EnumType(enums.ChassisType, serialized_name=api_constants.CHASSIS_TYPE)

    def __init__(self, **kwargs):
        super(GetIpmiConfig, self).__init__(result_dm=result_models.GetIpmiConfig, **kwargs)
        self.failing = True


class GetIpmiInfo(_base.BaseNodeApi):
    """GetIpmiInfo request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetIpmiInfo, self).__init__(result_dm=result_models.GetIpmiInfo, **kwargs)
        self.failing = True


class GetNetworkConfig(_base.BaseNodeApi):
    """GetNetworkConfig request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetNetworkConfig, self).__init__(result_dm=result_models.GetNetworkConfig, **kwargs)
        self.failing = True


class GetPendingOperation(_base.BaseNodeApi):
    """GetPendingOperation request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetPendingOperation, self).__init__(
            result_dm=result_models.GetPendingOperation, **kwargs
        )


class GetNetworkInterface(_base.BaseNodeApi):
    """GetNetworkInterface request."""
    _version_added = sf_version.SfVersion('5.0')
    interface = types.EnumType(
        enums.NetworkInterface,
        required=True,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['name'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.check_failure,
                enums.ApiExceptionType.param_invalid:
                    enums.SfException.invalid_network_interface,
                enums.ApiExceptionType.param_enum_invalid:
                    enums.SfException.invalid_network_interface}
        }
    )

    def __init__(self, **kwargs):
        super(GetNetworkInterface, self).__init__(
            result_dm=result_models.GetNetworkInterface, **kwargs
        )
        self.add_setup_api(ListNetworkInterfaces)


class ListNetworkInterfaces(_base.BaseNodeApi):
    """ListNetworkInterfaces request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListNetworkInterfaces, self).__init__(
            result_dm=result_models.ListNetworkInterfaces, **kwargs
        )


class ListDriveHardware(_base.BaseNodeApi):
    """ListDriveHardware request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListDriveHardware, self).__init__(result_dm=result_models.ListDriveHardware, **kwargs)
        self.failing = True


class ListTests(_base.BaseNodeApi):
    """ListTests request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListTests, self).__init__(result_dm=result_models.ListTests, **kwargs)


class ListUtilities(_base.BaseNodeApi):
    """ListUtilities request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListUtilities, self).__init__(result_dm=result_models.ListUtilities, **kwargs)


class ResetNode(_base.BaseNodeApi):
    """ResetNode request."""
    _version_added = sf_version.SfVersion('5.0')
    build = types.StringType()
    force = types.BooleanType()
    options = types.DictType(types.StringType())

    def __init__(self, **kwargs):
        super(ResetNode, self).__init__(
            result_dm=result_models.ResetNode, node_requirement=4, **kwargs
        )
        self.add_setup_api(_cluster.RemoveNodes)
        self.add_inspection_api(_cluster.ListAllNodes)


class ResetDrives(_base.BaseNodeApi):
    """ResetDrives request."""
    _version_added = sf_version.SfVersion('5.0')
    drives = types.ListType(types.StringType(regex=r'^slot[0-9]$'), required=True)
    force = types.BooleanType(required=True)

    def __init__(self, **kwargs):
        super(ResetDrives, self).__init__(result_dm=result_models.ResetDrives, **kwargs)
        self.add_inspection_api(_drive.ListDrives)
        self.failing = True


class RestartNetworking(_base.BaseNodeApi):
    """RestartNetworking request."""
    _version_added = sf_version.SfVersion('5.0')
    force = types.BooleanType(required=True)

    def __init__(self, **kwargs):
        super(RestartNetworking, self).__init__(**kwargs)
        self.failing = True


class RestartServices(_base.BaseNodeApi):
    """RestartServices request."""
    _version_added = sf_version.SfVersion('5.0')
    force = types.BooleanType(required=True)
    service = types.StringType()
    action = types.EnumType(
        enum_class=enums.RestartServicesAction,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.daemon_unknown_type}
        }
    )

    def __init__(self, **kwargs):
        super(RestartServices, self).__init__(result_dm=result_models.RestartServices, **kwargs)
        self.failing = True


class SetClusterConfig(_base.BaseNodeApi):
    """SetClusterConfig request."""
    _version_added = sf_version.SfVersion('5.0')
    cluster = types.ModelType(models.ClusterConfig)

    def __init__(self, **kwargs):
        super(SetClusterConfig, self).__init__(result_dm=result_models.SetClusterConfig, **kwargs)
        # Kicks off untracked async process that blocks subsequent calls for a significant period of time
        self.cant_test = True


class SetConfig(_base.BaseNodeApi):
    """SetConfig request."""
    _version_added = sf_version.SfVersion('5.0')
    config = types.ModelType(models.ClusterNodeConfig)

    def __init__(self, **kwargs):
        super(SetConfig, self).__init__(result_dm=result_models.SetConfig, **kwargs)
        # Requires extensive nested models as input
        self.cant_test = True


class SetNetworkConfig(_base.BaseNodeApi):
    """SetNetworkConfig request."""
    _version_added = sf_version.SfVersion('5.0')
    network = types.ModelType(models.NetworkInterfaceDetails)

    def __init__(self, **kwargs):
        super(SetNetworkConfig, self).__init__(result_dm=result_models.SetNetworkConfig, **kwargs)
        # Requires extensive nested models as input
        self.cant_test = True


class Shutdown(_base.BaseNodeApi):
    """Shutdown request."""
    _version_added = sf_version.SfVersion('5.0')
    option = types.EnumType(
        enum_class=enums.ShutdownOption,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.unknown_enum_type,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.unknown_exception}
        }
    )

    def __init__(self, **kwargs):
        super(Shutdown, self).__init__(result_dm=result_models.Shutdown, **kwargs)

        # Unknown API in Sodium
        self.failing = True


class TestConnectEnsemble(_base.BaseNodeApi):
    """TestConnectEnsemble request."""
    _version_added = sf_version.SfVersion('5.0')
    ensemble = types.ListType(types.IPv4Type())

    def __init__(self, **kwargs):
        super(TestConnectEnsemble, self).__init__(
            result_dm=result_models.TestConnectEnsemble, **kwargs
        )
        self.failing = True


class TestConnectMvip(_base.BaseNodeApi):
    """TestConnectMvip request."""
    _version_added = sf_version.SfVersion('5.0')
    mvip = types.IPv4Type(
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.MVIP],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.check_failure,
                enums.ApiExceptionType.param_value_does_not_exist: enums.SfException.check_failure}
        }
    )

    def __init__(self, **kwargs):
        super(TestConnectMvip, self).__init__(result_dm=result_models.TestConnectMvip, **kwargs)
        self.add_inspection_api(_cluster.GetClusterInfo)
        self.failing = True


class TestConnectSvip(_base.BaseNodeApi):
    """TestConnectSvip request."""
    _version_added = sf_version.SfVersion('5.0')
    svip = types.IPv4Type(
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['svip'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.check_failure,
                enums.ApiExceptionType.param_value_does_not_exist: enums.SfException.check_failure}
        }
    )

    def __init__(self, **kwargs):
        super(TestConnectSvip, self).__init__(result_dm=result_models.TestConnectSvip, **kwargs)
        self.add_inspection_api(_cluster.GetClusterInfo)
        self.failing = True


class TestDrives(_base.BaseNodeApi):
    """TestDrives request."""
    _version_added = sf_version.SfVersion('5.0')
    force = types.BooleanType()
    minutes = types.IntType()

    def __init__(self, **kwargs):
        super(TestDrives, self).__init__(**kwargs)

        # Must be tested against an orphan node
        self.failing = True


class TestHardwareConfig(_base.BaseNodeApi):
    """TestHardwareConfig request."""
    _version_added = sf_version.SfVersion('5.0')
    clean = types.BooleanType()
    force = types.BooleanType()

    def __init__(self, **kwargs):
        super(TestHardwareConfig, self).__init__(result_dm=result_models.TestHardwareConfig, **kwargs)


class TestLocateCluster(_base.BaseNodeApi):
    """TestLocateCluster request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(TestLocateCluster, self).__init__(result_dm=result_models.TestLocateCluster, **kwargs)


class TestLocalConnectivity(_base.BaseNodeApi):
    """TestLocalConnectivity request."""
    _version_added = sf_version.SfVersion('6.0')

    def __init__(self, **kwargs):
        super(TestLocalConnectivity, self).__init__(
            result_dm=result_models.TestLocalConnectivity, **kwargs
        )


class TestNetworkConfig(_base.BaseNodeApi):
    """TestNetworkConfig request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(TestNetworkConfig, self).__init__(
            result_dm=result_models.TestNetworkConfig, timeout=60, **kwargs
        )


class TestPing(_base.BaseNodeApi):
    """TestPing request."""
    _version_added = sf_version.SfVersion('5.0')
    attempts = types.IntType(default=5)
    hosts = types.ListType(types.IPv4Type())
    total_timeout_sec = types.IntType(serialized_name='totalTimeoutSec')
    packet_size = types.IntType(serialized_name='packetSize')
    ping_timeout_msec = types.IntType(default=500, serialized_name='pingTimeoutMsec')
    prohibit_fragmentation = types.BooleanType(serialized_name='prohibitFragmentation')

    def __init__(self, **kwargs):
        super(TestPing, self).__init__(result_dm=result_models.TestPing, **kwargs)
        self.failing = True


class TestRemoteConnectivity(_base.BaseNodeApi):
    """TestRemoteConnectivity request."""
    _version_added = sf_version.SfVersion('6.0')

    def __init__(self, **kwargs):
        super(TestRemoteConnectivity, self).__init__(
            result_dm=result_models.TestRemoteConnectivity, **kwargs
        )
