"""Private request data models."""
from sf_api import (
    api_constants,
    enums,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import _base


class GetReport(_base.BaseClusterApi):
    """GetReport request.

    :param str report_name: Override for the report_name parameter.
    """
    _version_added = sf_version.SfVersion('5.0')
    report_name = types.EnumType(
        enum_class=enums.JsonReport,
        required=True,
        serialized_name='reportName',
    )

    def __init__(self, report_name=None, **kwargs):
        super(GetReport, self).__init__(result_dm=result_models.GetReport, **kwargs)
        self.cant_test = True

        if report_name:
            self.field_overrides[GetReport.report_name] = report_name


class SetConstants(_base.BaseClusterApi):
    """SetConstants request.

    :param bool skip_gc: Override for the cSkipGC constant.
    :param bool disable_bin_sync: Override for the cDisableBinAssignmentUpdates constant.
    """
    _version_added = sf_version.SfVersion('5.0')
    disable_bin_assignment_updates = types.BooleanType(serialized_name='cDisableBinAssignmentUpdates')
    skip_gc = types.BooleanType(serialized_name='cSkipGC')

    def __init__(self, skip_gc=None, disable_bin_sync=None, **kwargs):
        super(SetConstants, self).__init__(result_dm=result_models.SetConstants, **kwargs)
        self.add_teardown_api(ResetConstants, reset_all=True)

        if skip_gc is not None:
            self.field_overrides[SetConstants.skip_gc] = skip_gc
        if disable_bin_sync is not None:
            self.field_overrides[SetConstants.disable_bin_assignment_updates] = disable_bin_sync


class ResetConstants(_base.BaseClusterApi):
    """ResetConstants request.

    :param bool reset_all: Whether to override the request to reset all constants.
    """
    _version_added = sf_version.SfVersion('5.0')
    constants_old = types.ListType(
        field=types.EnumType(enums.ClusterConstants),
        required=True,
        serialized_name='constants',
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_api_constant,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.unknown_api_constant,
                enums.ApiExceptionType.param_conflict: None
            }
        },
        version_removed=sf_version.SfVersion('11.1')
    )
    reset_all = types.BooleanType(
        required=True,
        choices=[True],
        serialized_name='resetAll',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: None,
                enums.ApiExceptionType.param_value_invalid: enums.SfException.missing_param}
        }
    )

    # NOTE: See ELEM-3706 for details on this change in behavior
    constants_new = types.ListType(
        field=types.EnumType(enums.ClusterConstants),
        required=True,
        serialized_name='constants',
        metadata={
            api_constants.NULL_INVALID: False,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: None
            }
        },
        version_added=sf_version.SfVersion('11.1')
    )

    constants_old.metadata[api_constants.CONFLICTING_FIELDS] = [reset_all]
    constants_new.metadata[api_constants.CONFLICTING_FIELDS] = [reset_all]
    reset_all.metadata[api_constants.CONFLICTING_FIELDS] = [constants_old, constants_new]

    def __init__(self, reset_all=None, **kwargs):
        super(ResetConstants, self).__init__(**kwargs)

        if reset_all:
            self.field_overrides[ResetConstants.reset_all] = True
