"""Request data models related to security."""
from sf_api import (
    api_constants,
    enums,
    models,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import (
    _base,
    _cluster,
)


class AddKeyServerToProviderKmip(_base.BaseClusterApi):
    """AddKeyServerToProviderKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_PROVIDER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_PROVIDER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )
    key_server_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_SERVER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_SERVER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(AddKeyServerToProviderKmip, self).__init__(result_dm=None, **kwargs)
        self.add_setup_api(CreateKeyProviderKmip)
        self.add_setup_api(CreateKeyServerKmip)
        self.add_teardown_api(RemoveKeyServerFromProviderKmip)
        self.failing = True  # Sometimes keyProviderID is null


class CreateKeyProviderKmip(_base.BaseClusterApi):
    """CreateKeyProviderKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_name = types.StringType(
        required=True,
        serialized_name=api_constants.KEY_PROVIDER_NAME,
        metadata={api_constants.NULL_INVALID: False},
    )

    def __init__(self, **kwargs):
        super(CreateKeyProviderKmip, self).__init__(
            result_dm=result_models.CreateKeyProviderKmip, **kwargs)


class CreateKeyServerKmip(_base.BaseClusterApi):
    """CreateKeyServerKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    kmip_ca_certificate = types.StringType(
        required=True,
        serialized_name=api_constants.KMIP_CA_CERTIFICATE,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_client_certificate = types.StringType(
        required=True,
        serialized_name=api_constants.KMIP_CLIENT_CERTIFICATE,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_key_server_hostnames = types.ListType(
        types.StringType,
        required=True,
        serialized_name=api_constants.KMIP_KEY_SERVER_HOSTNAMES,
        metadata={api_constants.NULL_INVALID: True,
                  api_constants.EXCEPTION_OVERRIDES: {
                      enums.ApiExceptionType.param_null: enums.SfException.invalid_param}
                  },
    )
    kmip_key_server_name = types.StringType(
        required=True,
        serialized_name=api_constants.KMIP_KEY_SERVER_NAME,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_key_server_port = types.IntType(serialized_name=api_constants.KMIP_KEY_SERVER_PORT)

    def __init__(self, **kwargs):
        super(CreateKeyServerKmip, self).__init__(
            result_dm=result_models.CreateKeyServerKmip, **kwargs)


class CreatePublicPrivateKeyPair(_base.BaseClusterApi):
    """CreatePublicPrivateKeyPair request."""
    _version_added = sf_version.SfVersion('11.7')
    common_name = types.StringType(serialized_name=api_constants.COMMON_NAME)
    organization = types.StringType(serialized_name=api_constants.ORGANIZATION)
    organizational_unit = types.StringType(serialized_name=api_constants.ORGANIZATIONAL_UNIT)
    locality = types.StringType(serialized_name=api_constants.LOCALITY)
    state = types.StringType(serialized_name=api_constants.STATE)
    country = types.StringType(serialized_name=api_constants.COUNTRY)
    email_address = types.StringType(serialized_name=api_constants.EMAIL_ADDRESS)

    def __init__(self, **kwargs):
        super(CreatePublicPrivateKeyPair, self).__init__(result_dm=None, **kwargs)
        self.failing = True  # API not implemented


class DeleteKeyProviderKmip(_base.BaseClusterApi):
    """DeleteKeyProviderKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_PROVIDER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_PROVIDER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(DeleteKeyProviderKmip, self).__init__(result_dm=None, **kwargs)
        self.add_setup_api(CreateKeyProviderKmip)


class DeleteKeyServerKmip(_base.BaseClusterApi):
    """DeleteKeyServerKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_server_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_SERVER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_SERVER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(DeleteKeyServerKmip, self).__init__(result_dm=None, **kwargs)
        self.add_setup_api(CreateKeyServerKmip)


class GetClientCertificateSignRequest(_base.BaseClusterApi):
    """GetClientCertificateSignRequest request."""
    _version_added = sf_version.SfVersion('11.7')

    def __init__(self, **kwargs):
        super(GetClientCertificateSignRequest, self).__init__(
            result_dm=result_models.GetClientCertificateSignRequest, **kwargs)
        self.failing = True  # API not implemented


class GetFipsReport(_base.BaseClusterApi):
    _version_added = sf_version.SfVersion('10.3')

    def __init__(self, **kwargs):
        super(GetFipsReport, self).__init__(result_dm=result_models.GetFipsReport, **kwargs)
        self.cant_test = True  # This API depends on specific hardware


class GetKeyProviderKmip(_base.BaseClusterApi):
    """GetKeyProviderKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_PROVIDER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_PROVIDER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(GetKeyProviderKmip, self).__init__(
            result_dm=result_models.GetKeyProviderKmip, **kwargs)
        self.add_setup_api(CreateKeyProviderKmip)


class GetKeyServerKmip(_base.BaseClusterApi):
    """GetKeyServerKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_server_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_SERVER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_SERVER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(GetKeyServerKmip, self).__init__(
            result_dm=result_models.GetKeyServerKmip, **kwargs)
        self.add_setup_api(CreateKeyServerKmip)


class ListKeyProvidersKmip(_base.BaseClusterApi):
    """ListKeyProvidersKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_is_active = types.BooleanType(serialized_name=api_constants.KEY_PROVIDER_IS_ACTIVE)
    kmip_key_provider_has_server_assigned = types.BooleanType(
        serialized_name=api_constants.KMIP_KEY_PROVIDER_HAS_SERVER_ASSIGNED
    )

    key_provider_is_active.metadata[api_constants.CONFLICTING_FIELDS] = \
        [kmip_key_provider_has_server_assigned]
    kmip_key_provider_has_server_assigned.metadata[api_constants.CONFLICTING_FIELDS] = \
        [key_provider_is_active]

    def __init__(self, **kwargs):
        super(ListKeyProvidersKmip, self).__init__(
            result_dm=result_models.ListKeyProvidersKmip, **kwargs)
        self.add_setup_api(CreateKeyProviderKmip)
        self.failing = True  # Conflicting fields


class ListKeyServersKmip(_base.BaseClusterApi):
    """ListKeyServersKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_id = types.IntIdType(serialized_name=api_constants.KEY_PROVIDER_ID)
    kmip_assigned_provider_is_active = types.BooleanType(
        serialized_name=api_constants.KMIP_ASSIGNED_PROVIDER_IS_ACTIVE
    )
    kmip_has_provider_assigned = types.BooleanType(
        serialized_name=api_constants.KMIP_HAS_PROVIDER_ASSIGNED
    )

    kmip_assigned_provider_is_active.metadata[api_constants.CONFLICTING_FIELDS] = \
        [kmip_has_provider_assigned]
    kmip_has_provider_assigned.metadata[api_constants.CONFLICTING_FIELDS] = \
        [kmip_assigned_provider_is_active]

    def __init__(self, **kwargs):
        super(ListKeyServersKmip, self).__init__(
            result_dm=result_models.ListKeyServersKmip, **kwargs)
        self.add_setup_api(CreateKeyServerKmip)
        self.failing = True  # Conflicting fields


class ModifyKeyServerKmip(_base.BaseClusterApi):
    """ModifyKeyServerKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_server_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_SERVER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_SERVER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )
    kmip_ca_certificate = types.StringType(
        serialized_name=api_constants.KMIP_CA_CERTIFICATE,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_client_certificate = types.StringType(
        serialized_name=api_constants.KMIP_CLIENT_CERTIFICATE,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_key_server_hostnames = types.ListType(
        types.StringType,
        serialized_name=api_constants.KMIP_KEY_SERVER_HOSTNAMES,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_key_server_name = types.StringType(
        serialized_name=api_constants.KMIP_KEY_SERVER_NAME,
        metadata={api_constants.NULL_INVALID: False},
    )
    kmip_key_server_port = types.IntType(serialized_name=api_constants.KMIP_KEY_SERVER_PORT)

    def __init__(self, **kwargs):
        super(ModifyKeyServerKmip, self).__init__(
            result_dm=result_models.ModifyKeyServerKmip, **kwargs)
        self.add_setup_api(CreatePublicPrivateKeyPair)
        self.add_setup_api(CreateKeyServerKmip)


class RemoveKeyServerFromProviderKmip(_base.BaseClusterApi):
    """RemoveKeyServerFromProviderKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_server_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_SERVER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_SERVER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(RemoveKeyServerFromProviderKmip, self).__init__(result_dm=None, **kwargs)
        self.add_setup_api(CreateKeyProviderKmip)
        self.add_setup_api(CreateKeyServerKmip)
        self.add_setup_api(AddKeyServerToProviderKmip)
        self.failing = True  # Sometimes keyProviderID is null


class TestKeyProviderKmip(_base.BaseClusterApi):
    """TestKeyProviderKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_provider_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_PROVIDER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_PROVIDER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(TestKeyProviderKmip, self).__init__(result_dm=None, **kwargs)
        self.add_setup_api(CreateKeyProviderKmip)
        self.failing = True  # Requires external key server


class TestKeyServerKmip(_base.BaseClusterApi):
    """TestKeyServerKmip request."""
    _version_added = sf_version.SfVersion('11.7')
    key_server_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.KEY_SERVER_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.KEY_SERVER_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.object_does_not_exist,
            }
        },
    )

    def __init__(self, **kwargs):
        super(TestKeyServerKmip, self).__init__(result_dm=None, **kwargs)
        self.add_setup_api(CreateKeyServerKmip)
        self.failing = True  # Requires external key server


class DisableEncryptionAtRest(_base.BaseClusterApi):
    """DisableEncryptionAtRest request."""
    _version_added = sf_version.SfVersion('5.0')
    force_during_bin_sync = types.BooleanType(
        serialized_name=api_constants.FORCE_DURING_BIN_SYNC,
        version_added=sf_version.SfVersion('11.7'),
        metadata={api_constants.CANT_TEST: True}
    )

    def __init__(self, **kwargs):
        super(DisableEncryptionAtRest, self).__init__(**kwargs)
        self.add_setup_api(EnableEncryptionAtRest)
        self.add_inspection_api(
            request_dm=_cluster.GetClusterInfo,
            loop_for_response={
                models.Cluster.encryption_at_rest_state: (enums.EarState.enabled.value,
                                                          enums.EarState.disabled.value)}
        )
        self.add_inspection_api(
            request_dm=_cluster.GetClusterInfo,
            conditions={models.Cluster.encryption_at_rest_state: enums.EarState.enabled.value}
        )
        self.add_teardown_api(
            request_dm=_cluster.GetClusterInfo,
            loop_for_response={
                models.Cluster.encryption_at_rest_state: enums.EarState.disabled.value}
        )
        self.failing = True


class EnableEncryptionAtRest(_base.BaseClusterApi):
    """EnableEncryptionAtRest request."""
    _version_added = sf_version.SfVersion('5.0')
    key_provider_id = types.IntIdType(
        serialized_name=api_constants.KEY_PROVIDER_ID,
        version_added=sf_version.SfVersion('11.7')
    )
    force_during_bin_sync = types.BooleanType(
        serialized_name=api_constants.FORCE_DURING_BIN_SYNC,
        version_added=sf_version.SfVersion('11.7'),
        metadata={api_constants.CANT_TEST: True}
    )

    def __init__(self, **kwargs):
        super(EnableEncryptionAtRest, self).__init__(**kwargs)
        self.add_teardown_api(DisableEncryptionAtRest)
        self.failing = True


class GetEncryptionAtRestInfo(_base.BaseClusterApi):
    """GetEncryptionAtRestInfo request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetEncryptionAtRestInfo, self).__init__(
            result_dm=result_models.GetEncryptionAtRestInfo, **kwargs)
