"""Request data models related to volumes."""
import sf_models

from sf_api import (
    api_constants,
    enums,
    models,
    result_models,
    sf_version,
)
from sf_api.models import types
from sf_api.request_models import (
    _account,
    _base,
    _cluster,
)


class AddInitiatorsToVolumeAccessGroup(_base.BaseClusterApi):
    """AddInitiatorsToVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    initiators = types.ListType(
        field=types.StringType(min_length=1, max_length=128),
        required=True,
        default=list,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.INITIATOR_ID]
        }
    )
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )

    def __init__(self, **kwargs):
        super(AddInitiatorsToVolumeAccessGroup, self).__init__(
            result_dm=result_models.AddInitiatorsToVolumeAccessGroup, **kwargs
        )
        self.add_setup_api(CreateVolumeAccessGroup, no_initiators=True)
        self.add_setup_api(CreateInitiators, skip_setup=True)
        self.failing = True


class AddVolumesToVolumeAccessGroup(_base.BaseClusterApi):
    """AddVolumesToVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )
    volumes = types.ListType(
        field=types.IntIdType(),
        required=True,
        default=list,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(AddVolumesToVolumeAccessGroup, self).__init__(
            result_dm=result_models.AddVolumesToVolumeAccessGroup, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.add_setup_api(CreateVolumeAccessGroup, skip_setup=True)
        self.failing = True


class CancelClone(_base.BaseClusterApi):
    """CancelClone request."""
    _version_added = sf_version.SfVersion('5.0')
    clone_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.CLONE_ID,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.CLONE_ID]}
    )

    def __init__(self, **kwargs):
        super(CancelClone, self).__init__(wait_for_async_setup=False, **kwargs)
        self.add_setup_api(CloneVolume)


class CancelGroupClone(_base.BaseClusterApi):
    """CancelGroupClone request."""
    _version_added = sf_version.SfVersion('5.0')
    group_clone_id = types.IntIdType(
        required=True,
        serialized_name='groupCloneID',
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: ['groupCloneID']}
    )

    def __init__(self, **kwargs):
        super(CancelGroupClone, self).__init__(wait_for_async_setup=False, **kwargs)
        self.add_setup_api(CloneMultipleVolumes)


class CloneMultipleVolumes(_base.BaseClusterApi):
    """CloneMultipleVolumes request."""
    _version_added = sf_version.SfVersion('5.0')
    access = types.EnumType(
        enum_class=enums.VolumeAccess,
        choices=[
            enums.VolumeAccess.read_only,
            enums.VolumeAccess.read_write,
            enums.VolumeAccess.locked,
            enums.VolumeAccess.replication_target
        ],
        metadata={
            api_constants.INVALID_CHOICES: {
                enums.VolumeAccess.snap_mirror_target.value: enums.SfException.invalid_param}
        }
    )
    group_snapshot_id = types.IntIdType(
        serialized_name=api_constants.GROUP_SNAPSHOT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.GROUP_SNAPSHOT_ID]
        }
    )
    new_account_id = types.IntIdType(
        serialized_name='newAccountID',
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID]}
    )
    volumes = types.ListType(
        field=types.ModelType(models.Volume),
        required=True,
        max_size=32,
        default=list,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max:
                    enums.SfException.max_volumes_per_group_snapshot_exceeded
            }
        }
    )

    def __init__(self, **kwargs):
        super(CloneMultipleVolumes, self).__init__(
            result_dm=result_models.CloneMultipleVolumes,
            async_timeout=3600,
            **kwargs
        )
        self.add_setup_api(CreateGroupSnapshot)
        self.add_inspection_api(ListGroupSnapshots)
        self.add_teardown_api(DeleteVolumes)
        self.is_async = True
        self.failing = True


class CloneVolume(_base.BaseClusterApi):
    """CloneVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    access = types.EnumType(
        enum_class=enums.VolumeAccess,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_invalid: enums.SfException.unrecognized_enum_string,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.invalid_api_param,
                enums.SfException.invalid_api_param: enums.ApiExceptionType.param_invalid,
            }
        },
        choices=[enums.VolumeAccess.read_only,
                 enums.VolumeAccess.read_write,
                 enums.VolumeAccess.locked,
                 enums.VolumeAccess.replication_target]
    )
    name = types.StringType(
        required=True,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param}
        }
    )
    new_account_id = types.IntIdType(
        serialized_name='newAccountID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist: enums.SfException.db_no_such_path}
        }
    )
    new_size = types.IntType(
        serialized_name='newSize',
        min_value=1000000000,
        max_value=8000000491520,
        default=1000000000,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(CloneVolume, self).__init__(
            result_dm=result_models.CloneVolume,
            async_timeout=3600,
            **kwargs
        )
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.add_teardown_api(DeleteVolume)
        self.is_async = True


class CopyVolume(_base.BaseClusterApi):
    """CopyVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    dst_volume_id = types.IntIdType(
        required=True,
        serialized_name='dstVolumeID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )
    snapshot_id = types.IntIdType(
        serialized_name=api_constants.SNAPSHOT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.SNAPSHOT_ID]
        }
    )
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(CopyVolume, self).__init__(
            result_dm=result_models.CopyVolume,
            async_timeout=3600,
            **kwargs
        )
        self.add_setup_api(CreateVolume)
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.is_async = True


class CreateBackupTarget(_base.BaseClusterApi):
    """CreateBackupTarget request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType(), required=True)
    name = types.StringType(required=True)

    def __init__(self, **kwargs):
        super(CreateBackupTarget, self).__init__(result_dm=result_models.CreateBackupTarget, **kwargs)
        self.failing = True


class CreateGroupSnapshot(_base.BaseClusterApi):
    """CreateGroupSnapshot request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType)
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION)
    name = types.StringType()
    retention = types.StringType(regex=r'^\d\d:\d\d:\d\d$')
    volumes = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(CreateGroupSnapshot, self).__init__(
            result_dm=result_models.CreateGroupSnapshot, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.add_teardown_api(DeleteGroupSnapshot)
        self.failing = True


class InitiatorNew(sf_models.Base):
    """Initiator data model.

    Corresponds to elements of the initiators object in the CreateInitiators request.
    """
    alias = types.StringType()
    attributes = types.DictType(types.StringType())
    name = types.StringType(required=True, min_length=1)
    volume_access_group_id = types.IntIdType(
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )


class CreateInitiators(_base.BaseClusterApi):
    """CreateInitiators request."""
    _version_added = sf_version.SfVersion('5.0')
    initiators = types.ListType(
        field=types.ModelType(InitiatorNew),
        required=True,
        default=list,
        metadata={api_constants.REF_REQUIRED: True}
    )

    def __init__(self, **kwargs):
        super(CreateInitiators, self).__init__(result_dm=result_models.CreateInitiators, **kwargs)
        self.add_setup_api(CreateVolumeAccessGroup)
        self.add_teardown_api(DeleteInitiators)
        self.failing = True


class QoS(models.Base):
    """Quality of Service data model.

    Corresponds to the qos object in the CreateVolume request.
    """
    min_iops = types.IntType(
        min_value=50,
        max_value=15000,
        default=50,
        serialized_name=api_constants.MIN_IOPS,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )
    max_iops = types.IntType(
        min_value=100,
        max_value=200000,
        default=15000,
        serialized_name=api_constants.MAX_IOPS,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )
    burst_iops = types.IntType(
        min_value=100,
        max_value=200000,
        default=15000,
        serialized_name=api_constants.BURST_IOPS,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )
    burst_time = types.IntType(
        choices=[60],
        serialized_name=api_constants.BURST_TIME,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_type_invalid: enums.SfException.unknown_exception,
                enums.ApiExceptionType.param_value_invalid: enums.SfException.invalid_value}
        }
    )

    # Define interdependent parameter relations
    min_iops.metadata[api_constants.MAX_LIMIT_FIELD] = max_iops
    max_iops.metadata[api_constants.MIN_LIMIT_FIELD] = min_iops
    max_iops.metadata[api_constants.MAX_LIMIT_FIELD] = burst_iops
    burst_iops.metadata[api_constants.MIN_LIMIT_FIELD] = max_iops


class CreateMultipleVolumes(_base.BaseClusterApi):
    """CreateMultipleVolumes request.

    :param bool default_qos: Whether to override the request to use default QOS settings.
    :param int vol_count: Override the request to create this many volumes.
    """
    _version_added = sf_version.SfVersion('5.0')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )
    attributes = types.DictType(
        field=types.StringType,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    enable_512e = types.BooleanType(required=True, serialized_name=api_constants.ENABLE_512E)
    names = types.ListType(
        field=types.StringType(required=True, min_length=1, max_length=64),
        required=True,
        min_size=0,
        max_size=100,
        default=['testVolume1', 'testVolume2'],
        metadata={
            api_constants.NULL_INVALID: False,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param}
        }
    )
    qos = types.ModelType(model_spec=QoS, serialized_name=api_constants.QOS)
    qos_policy_id = types.IntIdType(
        serialized_name=api_constants.QOS_POLICY_ID,
        version_added=sf_version.SfVersion('10.0'),
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.QOS_POLICY_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )
    slice_count = types.IntType(serialized_name=api_constants.SLICE_COUNT)
    total_size = types.IntType(
        required=True,
        serialized_name=api_constants.TOTAL_SIZE,
        min_value=1000000000,
        max_value=8000000491520,
        default=1000000000,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )

    # Define interdependent parameter relations
    qos.metadata[api_constants.CONFLICTING_FIELDS] = [qos_policy_id]
    qos_policy_id.metadata[api_constants.CONFLICTING_FIELDS] = [qos]

    def __init__(self, default_qos=False, vol_count=None, **kwargs):
        super(CreateMultipleVolumes, self).__init__(
            result_dm=result_models.CreateMultipleVolumes,
            exceptions={
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_param_type},
            **kwargs
        )
        self.add_setup_api(_account.AddAccount)
        self.add_setup_api(CreateQoSPolicy)
        self.add_teardown_api(DeleteVolumes)
        self.failing = True

        if default_qos:
            self.field_overrides[CreateMultipleVolumes.qos] = {}
            self.fields_to_exclude.append(CreateMultipleVolumes.qos_policy_id)
        if vol_count:
            self.field_overrides[CreateMultipleVolumes.names] = [
                'mock-volume-{}'.format(i) for i in range(vol_count)
            ]


class CreateQoSPolicy(_base.BaseClusterApi):
    """CreateQoSPolicy request.

    :param bool default: Whether to override the request to use default QOS settings.
    """
    _version_added = sf_version.SfVersion('10.0')
    name = types.StringType(
        required=True,
        serialized_name=api_constants.NAME,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_qos_policy_name,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_qos_policy_name,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_qos_policy_name}
        }
    )
    qos = types.ModelType(
        model_spec=QoS,
        required=True,
        serialized_name=api_constants.QOS,
        metadata={api_constants.NULL_INVALID: False}
    )

    def __init__(self, default=False, **kwargs):
        super(CreateQoSPolicy, self).__init__(result_dm=result_models.CreateQoSPolicy, **kwargs)
        self.add_teardown_api(DeleteQoSPolicy)

        if default:
            self.field_overrides[CreateQoSPolicy.qos] = {}


class ScheduleInfo(models.Base):
    """ScheduleInfo data model.

    Corresponds to the schedule_info object in the CreateSchedule request.
    """
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION
    )
    name = types.StringType()
    retention = types.BaseType()
    volume_id = types.IntType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_snapshot_schedule,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_snapshot_schedule,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )
    volumes = types.ListType(
        field=types.IntIdType(),
        required=True,
        max_size=32,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.NULL_INVALID: False,
            api_constants.EXCEPTION_OVERRIDES: {
                # enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_snapshot_schedule,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_snapshot_schedule,
                enums.ApiExceptionType.param_exceeded_max:
                    enums.SfException.max_volumes_per_group_snapshot_exceeded,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    # Define interdependent parameter relations
    volume_id.metadata[api_constants.CONFLICTING_FIELDS] = [volumes]
    volumes.metadata[api_constants.CONFLICTING_FIELDS] = [volume_id]


class Weekday(sf_models.Base):
    """Weekday data model.

    Corresponds to the elements of the weekdays object in the CreateSchedule request.
    """
    day = types.IntType(
        required=True,
        min_value=0,
        max_value=6,
        metadata={
            api_constants.NULL_INVALID: False,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param_type,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_param_type}
        }
    )
    offset = types.IntType(
        required=True,
        min_value=0,
        max_value=6,
        metadata={
            api_constants.NULL_INVALID: False,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_param_type,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_param_type}
        }
    )


class CreateSchedule(_base.BaseClusterApi):
    """CreateSchedule request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.BaseType()
    hours = types.IntType(
        required=True,
        min_value=0,
        max_value=23,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_schedule_parameters}
        }
    )
    minutes = types.IntType(
        required=True,
        min_value=0,
        max_value=59,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_schedule_parameters}
        }
    )
    monthdays = types.ListType(
        field=types.IntType(min_value=1, max_value=31),
        required=True,
        default=list,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_schedule_parameters}
        }
    )
    paused = types.BooleanType()
    recurring = types.BooleanType()
    schedule_info = types.ModelType(
        model_spec=ScheduleInfo,
        required=True,
        serialized_name=api_constants.SCHEDULE_INFO,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_snapshot_schedule,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_snapshot_schedule}
        }
    )
    schedule_name = types.StringType(
        required=True,
        min_length=1,
        serialized_name=api_constants.SCHEDULE_NAME,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_schedule_parameters}
        }
    )
    schedule_type = types.EnumType(
        enum_class=enums.ScheduleType,
        required=True,
        serialized_name=api_constants.SCHEDULE_TYPE,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_type,
                enums.ApiExceptionType.param_enum_invalid: enums.SfException.invalid_schedule_type}
        }
    )
    starting_date = types.DateTimeType(
        required=True,
        serialized_name=api_constants.STARTING_DATE,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_schedule_parameters}
        }
    )
    weekdays = types.ListType(
        field=types.ModelType(Weekday),
        required=True,
        default=list,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_schedule_parameters,
                enums.ApiExceptionType.param_missing: enums.SfException.invalid_schedule_parameters}
        }
    )

    # Define interdependent parameter relations
    weekdays.metadata[api_constants.CONFLICTING_FIELDS] = [monthdays]
    monthdays.metadata[api_constants.CONFLICTING_FIELDS] = [weekdays]
    hours.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [minutes, monthdays, weekdays, starting_date]
    minutes.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [hours, monthdays, weekdays, starting_date]
    monthdays.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [hours, minutes, weekdays, starting_date]
    weekdays.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [hours, minutes, monthdays, starting_date]
    starting_date.metadata[api_constants.OPTIONAL_WITH_FIELDS] = [hours, minutes, monthdays, weekdays]

    def __init__(self, **kwargs):
        super(CreateSchedule, self).__init__(result_dm=result_models.CreateSchedule, **kwargs)
        self.add_setup_api(CreateMultipleVolumes, vol_count=33)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.add_teardown_api(ModifySchedule, delete=True)

        # - Zero is a valid value for hours/minutes, but also counts as the parameter being excluded,
        # which can result in an error if other OPTIONAL_WITH_FIELDS aren't also present
        # - hours=24/minutes=60 don't consistently raise errors
        self.failing = True


class CreateSnapshot(_base.BaseClusterApi):
    """CreateSnapshot request.

    :param bool new_snapshot: Whether to override the request to create a new snapshot.
    """
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType)
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unknown_enum_type
            }
        }
    )
    name = types.StringType()
    retention = types.StringType(regex=r'^\d\d:\d\d:\d\d$')
    snapshot_id = types.IntIdType(
        serialized_name=api_constants.SNAPSHOT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.SNAPSHOT_ID]
        }
    )
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist},
            }
    )

    def __init__(self, new_snapshot=None, **kwargs):
        super(CreateSnapshot, self).__init__(result_dm=result_models.CreateSnapshot, **kwargs)
        self.add_setup_api(CreateMultipleVolumes)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.add_teardown_api(DeleteSnapshot)
        self.failing = True

        if new_snapshot:
            self.fields_to_exclude.append(CreateSnapshot.snapshot_id)
        else:
            self.add_setup_api(CreateSnapshot, new_snapshot=True)


class CreateVolume(_base.BaseClusterApi):
    """CreateVolume request.

    :param bool default_qos: Whether to override the request to use default QOS settings.
    """
    _version_added = sf_version.SfVersion('5.0')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )
    associate_with_qos_policy = types.BooleanType(
        serialized_name=api_constants.ASSOCIATE_WITH_QOS_POLICY
    )
    attributes = types.DictType(
        field=types.StringType,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    enable_512e = types.BooleanType(
        required=True,
        serialized_name=api_constants.ENABLE_512E
    )
    name = types.StringType(
        required=True,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param}
        }
    )
    qos = types.ModelType(
        model_spec=QoS,
        serialized_name=api_constants.QOS
    )
    qos_policy_id = types.IntIdType(
        serialized_name=api_constants.QOS_POLICY_ID,
        version_added=sf_version.SfVersion('10.0'),
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.QOS_POLICY_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )
    total_size = types.IntType(
        required=True,
        serialized_name=api_constants.TOTAL_SIZE,
        min_value=1000000000,
        max_value=8000000491520,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )

    # Define interdependent parameter relations
    qos.metadata[api_constants.CONFLICTING_FIELDS] = [qos_policy_id]
    qos_policy_id.metadata[api_constants.CONFLICTING_FIELDS] = [qos]

    def __init__(self, default_qos=False, **kwargs):
        super(CreateVolume, self).__init__(result_dm=result_models.CreateVolume, **kwargs)
        self.add_setup_api(_account.AddAccount)
        self.add_setup_api(CreateQoSPolicy)
        self.add_teardown_api(DeleteVolume)
        self.failing = True

        if default_qos:
            self.field_overrides[CreateVolume.qos] = {}
            self.fields_to_exclude.append(CreateVolume.qos_policy_id)


class CreateVolumeAccessGroup(_base.BaseClusterApi):
    """CreateVolumeAccessGroup request.

    :param bool force_volumes: Whether to force the inclusion of the volumes parameter.
    :param bool no_initiators: Whether to force the exclusion of the initiators parameter.
    :param int vol_count: Override the request to create this many volumes.
    """
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(
        field=types.StringType,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    initiators_5 = types.ListType(
        field=types.StringType(min_length=1, max_length=64),
        version_removed=sf_version.SfVersion('11.0'),
        serialized_name='initiators',
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    name = types.StringType(
        required=True,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param}
        }
    )
    volumes = types.ListType(
        field=types.IntIdType,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    # Version 11.0 Additions
    initiators = types.ListType(
        field=types.IntIdType(),
        version_added=sf_version.SfVersion('11.0'),
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.INITIATOR_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )

    def __init__(self, force_volumes=False, no_initiators=False, vol_count=None, **kwargs):
        super(CreateVolumeAccessGroup, self).__init__(
            result_dm=result_models.CreateVolumeAccessGroup,
            **kwargs
        )
        self.add_setup_api(CreateInitiators, skip_setup=True)
        self.add_setup_api(CreateMultipleVolumes, vol_count=vol_count)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.add_teardown_api(DeleteVolumeAccessGroup)
        self.failing = True

        if force_volumes:
            self.fields_to_force.append(CreateVolumeAccessGroup.volumes)
        if no_initiators:
            self.fields_to_exclude.append(CreateVolumeAccessGroup.initiators)
            self.fields_to_exclude.append(CreateVolumeAccessGroup.initiators_5)


class DeleteGroupSnapshot(_base.BaseClusterApi):
    """DeleteGroupSnapshot request."""
    _version_added = sf_version.SfVersion('5.0')
    group_snapshot_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.GROUP_SNAPSHOT_ID,
        metadata={api_constants.REF_PARAMS: [api_constants.GROUP_SNAPSHOT_ID]}
    )
    save_members = types.BooleanType(serialized_name='saveMembers')

    def __init__(self, **kwargs):
        super(DeleteGroupSnapshot, self).__init__(**kwargs)
        self.add_setup_api(CreateGroupSnapshot)
        self.add_inspection_api(ListGroupSnapshots)


class DeleteInitiators(_base.BaseClusterApi):
    """DeleteInitiators request."""
    _version_added = sf_version.SfVersion('5.0')
    initiators = types.ListType(
        field=types.IntIdType(),
        required=True,
        default=list,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.INITIATOR_ID]
        }
    )

    def __init__(self, **kwargs):
        super(DeleteInitiators, self).__init__(**kwargs)
        self.add_setup_api(CreateInitiators)
        self.add_inspection_api(ListInitiators)
        self.failing = True


class DeleteQoSPolicy(_base.BaseClusterApi):
    """DeleteQosPolicy request."""
    _version_added = sf_version.SfVersion('10.0')
    qos_policy_id = types.IntIdType(
        required=True,
        serialized_name='qosPolicyID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['qosPolicyID'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(DeleteQoSPolicy, self).__init__(**kwargs)
        self.add_setup_api(CreateQoSPolicy)
        self.add_inspection_api(ListQoSPolicies)


class DeleteSnapshot(_base.BaseClusterApi):
    """DeleteSnapshot request."""
    _version_added = sf_version.SfVersion('5.0')
    snapshot_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.SNAPSHOT_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.SNAPSHOT_ID]}
    )

    def __init__(self, **kwargs):
        super(DeleteSnapshot, self).__init__(**kwargs)
        self.add_setup_api(CreateSnapshot)
        self.add_inspection_api(ListSnapshots)


class DeleteVolume(_base.BaseClusterApi):
    """DeleteVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID, api_constants.VOLUMES, 'volumeIDs'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(DeleteVolume, self).__init__(result_dm=result_models.DeleteVolume, **kwargs)
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.add_teardown_api(PurgeDeletedVolume)


class DeleteVolumes(_base.BaseClusterApi):
    """DeleteVolumes request."""
    _version_added = sf_version.SfVersion('9.0')
    account_ids = types.ListType(
        field=types.IntIdType,
        required=True,
        serialized_name='accountIDs',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param,
            }
        }
    )
    volume_access_group_ids = types.ListType(
        field=types.IntIdType,
        required=True,
        serialized_name='volumeAccessGroupIDs',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_access_group_id_does_not_exist,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param,
            }
        }
    )
    volume_ids = types.ListType(
        field=types.IntIdType,
        required=True,
        serialized_name=api_constants.VOLUME_IDS,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    # Define interdependent parameter relations
    account_ids.metadata[api_constants.CONFLICTING_FIELDS] = [volume_access_group_ids, volume_ids]
    volume_access_group_ids.metadata[api_constants.CONFLICTING_FIELDS] = [account_ids, volume_ids]
    volume_ids.metadata[api_constants.CONFLICTING_FIELDS] = [account_ids, volume_access_group_ids]

    def __init__(self, **kwargs):
        super(DeleteVolumes, self).__init__(result_dm=result_models.DeleteVolumes, **kwargs)
        self.add_setup_api(CreateVolumeAccessGroup, force_volumes=True)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.add_inspection_api(ListVolumeAccessGroups, list_all=True)
        self.add_teardown_api(PurgeDeletedVolumes)


class DeleteVolumeAccessGroup(_base.BaseClusterApi):
    """DeleteVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.volume_access_group_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_access_group_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(DeleteVolumeAccessGroup, self).__init__(**kwargs)
        self.add_setup_api(CreateVolumeAccessGroup)
        self.add_inspection_api(ListVolumeAccessGroups, list_all=True)


class GetBackupTarget(_base.BaseClusterApi):
    """GetBackupTarget request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType())
    backup_target_id = types.IntIdType(
        required=True,
        serialized_name='backupTargetID',
        metadata={
            api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: ['backupTargetID'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_missing:
                    enums.SfException.backup_target_id_does_not_exist,
            }
        }
    )
    name = types.StringType(required=True)

    def __init__(self, **kwargs):
        super(GetBackupTarget, self).__init__(
            result_dm=result_models.GetBackupTarget,
            **kwargs
        )
        self.add_setup_api(CreateBackupTarget)
        self.add_inspection_api(ListBackupTargets)
        self.failing = True


class GetDefaultQoS(_base.BaseClusterApi):
    """GetDefaultQoS request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetDefaultQoS, self).__init__(result_dm=result_models.GetDefaultQoS, **kwargs)


class GetQoSPolicy(_base.BaseClusterApi):
    """GetQoSPolicy request."""
    _version_added = sf_version.SfVersion('10.0')
    qos_policy_id = types.IntIdType(
        required=True,
        serialized_name='qosPolicyID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['qosPolicyID'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetQoSPolicy, self).__init__(result_dm=result_models.GetQoSPolicy, **kwargs)
        self.add_setup_api(CreateQoSPolicy)
        self.add_inspection_api(ListQoSPolicies)


class GetVolumeAccessGroupEfficiency(_base.BaseClusterApi):
    """GetVolumeAccessGroupEfficiency request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )

    def __init__(self, **kwargs):
        super(GetVolumeAccessGroupEfficiency, self).__init__(
            result_dm=result_models.GetVolumeAccessGroupEfficiency, **kwargs
        )
        self.add_setup_api(CreateVolumeAccessGroup)
        self.add_inspection_api(ListVolumeAccessGroups)
        self.failing = True


class GetVolumeCount(_base.BaseClusterApi):
    """GetVolumeCount request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(GetVolumeCount, self).__init__(result_dm=result_models.GetVolumeCount, **kwargs)
        self.add_setup_api(CreateMultipleVolumes)


class GetVolumeEfficiency(_base.BaseClusterApi):
    """GetVolumeEfficiency request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(GetVolumeEfficiency, self).__init__(
            result_dm=result_models.GetVolumeEfficiency, **kwargs
        )
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListVolumes)
        self.failing = True


class GetVolumeStats(_base.BaseClusterApi):
    """GetVolumeStats request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(GetVolumeStats, self).__init__(result_dm=result_models.GetVolumeStats, **kwargs)
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListActiveVolumes, list_all=True)


class ListActiveVolumes(_base.BaseClusterApi):
    """ListActiveVolumes request.

    :param bool list_all: Whether to exclude all constraint parameters (default: False).
    """
    _version_added = sf_version.SfVersion('5.0')
    limit = types.IntType()
    start_volume_id = types.IntIdType(
        serialized_name='startVolumeID',
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, list_all=False, **kwargs):
        super(ListActiveVolumes, self).__init__(
            result_dm=result_models.ListVolumes,
            exceptions={enums.ApiExceptionType.param_invalid: enums.SfException.invalid_api_param},
            **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)

        if list_all:
            self.fields_to_exclude.append(ListActiveVolumes.limit)
            self.fields_to_exclude.append(ListActiveVolumes.start_volume_id)


class ListBackupTargets(_base.BaseClusterApi):
    """ListBackupTargets request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListBackupTargets, self).__init__(result_dm=result_models.ListBackupTargets, **kwargs)
        self.add_setup_api(CreateBackupTarget)


class ListBulkVolumeJobs(_base.BaseClusterApi):
    """ListBulkVolumeJobs request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListBulkVolumeJobs, self).__init__(result_dm=result_models.ListBulkVolumeJobs, **kwargs)


class ListDeletedVolumes(_base.BaseClusterApi):
    """ListDeletedVolumes request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListDeletedVolumes, self).__init__(
            result_dm=result_models.ListVolumes,
            exceptions={enums.ApiExceptionType.param_invalid: enums.SfException.invalid_api_param},
            **kwargs
        )
        self.add_setup_api(DeleteVolumes)


class ListGroupSnapshots(_base.BaseClusterApi):
    """ListGroupSnapshots request."""
    _version_added = sf_version.SfVersion('5.0')
    group_snapshot_id = types.IntIdType(
        serialized_name=api_constants.GROUP_SNAPSHOT_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.GROUP_SNAPSHOT_ID]}
    )
    volumes = types.ListType(
        field=types.IntIdType(),
        serialized_name=api_constants.VOLUME_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(ListGroupSnapshots, self).__init__(
            result_dm=result_models.ListGroupSnapshots, **kwargs)
        self.add_setup_api(CreateGroupSnapshot)


class ListInitiators(_base.BaseClusterApi):
    """ListInitiators request."""
    _version_added = sf_version.SfVersion('5.0')
    initiators = types.ListType(
        field=types.IntIdType(),
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.INITIATOR_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_invalid: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param
            }
        }
    )
    limit = types.IntType(
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param}
        }
    )
    start_initiator_id = types.IntIdType(serialized_name='startInitiatorID')

    # Define interdependent parameter relations
    initiators.metadata[api_constants.CONFLICTING_FIELDS] = [limit, start_initiator_id]
    limit.metadata[api_constants.CONFLICTING_FIELDS] = [initiators]
    start_initiator_id.metadata[api_constants.CONFLICTING_FIELDS] = [initiators]

    def __init__(self, **kwargs):
        super(ListInitiators, self).__init__(result_dm=result_models.ListInitiators, **kwargs)
        self.add_setup_api(CreateInitiators)


class ListQoSPolicies(_base.BaseClusterApi):
    """ListQoSPolicies request."""
    _version_added = sf_version.SfVersion('10.0')

    def __init__(self, **kwargs):
        super(ListQoSPolicies, self).__init__(result_dm=result_models.ListQoSPolicies, **kwargs)
        self.add_setup_api(CreateQoSPolicy)


class ListSchedules(_base.BaseClusterApi):
    """ListSchedules request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListSchedules, self).__init__(result_dm=result_models.ListSchedules, **kwargs)
        self.add_setup_api(CreateSchedule)


class ListSnapshots(_base.BaseClusterApi):
    """ListSnapshots request."""
    _version_added = sf_version.SfVersion('5.0')
    snapshot_id = types.IntIdType(
        serialized_name=api_constants.SNAPSHOT_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.SNAPSHOT_ID]}
    )
    volume_id = types.IntIdType(
        serialized_name=api_constants.VOLUME_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(ListSnapshots, self).__init__(result_dm=result_models.ListSnapshots, **kwargs)
        self.add_setup_api(CreateSnapshot)


class ListSyncJobs(_base.BaseClusterApi):
    """ListSyncJobs request."""
    _version_added = sf_version.SfVersion('5.0')

    def __init__(self, **kwargs):
        super(ListSyncJobs, self).__init__(result_dm=result_models.ListSyncJobs, **kwargs)


class ListVolumeAccessGroups(_base.BaseClusterApi):
    """ListVolumeAccessGroups request.

    :param bool list_all: Whether to exclude all constraint parameters (default: False).
    """
    _version_added = sf_version.SfVersion('5.0')
    limit = types.IntType()
    start_volume_access_group_id = types.IntIdType(
        serialized_name='startVolumeAccessGroupID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )

    def __init__(self, list_all=False, **kwargs):
        super(ListVolumeAccessGroups, self).__init__(
            result_dm=result_models.ListVolumeAccessGroups,
            exceptions={enums.ApiExceptionType.param_invalid: enums.SfException.invalid_api_param},
            **kwargs
        )
        self.add_setup_api(CreateVolumeAccessGroup)

        if list_all:
            self.fields_to_exclude.append(ListVolumeAccessGroups.limit)
            self.fields_to_exclude.append(ListVolumeAccessGroups.start_volume_access_group_id)


class ListVolumes(_base.BaseClusterApi):
    """ListVolumes request."""
    _version_added = sf_version.SfVersion('8.0')
    accounts = types.ListType(
        field=types.IntIdType,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param}
        }
    )
    is_paired = types.BooleanType(serialized_name='isPaired')
    limit = types.IntType(
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param}
        }
    )
    start_volume_id = types.IntIdType(
        serialized_name='startVolumeID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param}
        }
    )
    volume_name = types.StringType(serialized_name='volumeName')
    volume_status = types.EnumType(
        enum_class=enums.VolumeStatus,
        serialized_name='volumeStatus',
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.unrecognized_enum_string,
                enums.ApiExceptionType.param_value_invalid:
                    enums.SfException.unrecognized_enum_string}
        }
    )
    volume_ids = types.ListType(
        field=types.IntIdType,
        serialized_name=api_constants.VOLUME_IDS,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_conflict: enums.SfException.invalid_param}
        }
    )

    # Define interdependent parameter relations
    accounts.metadata[api_constants.CONFLICTING_FIELDS] = [volume_ids]
    start_volume_id.metadata[api_constants.CONFLICTING_FIELDS] = [volume_ids]
    limit.metadata[api_constants.CONFLICTING_FIELDS] = [volume_ids]
    volume_ids.metadata[api_constants.CONFLICTING_FIELDS] = [accounts, start_volume_id, limit]

    def __init__(self, **kwargs):
        super(ListVolumes, self).__init__(
            result_dm=result_models.ListVolumes,
            exceptions={enums.ApiExceptionType.param_invalid: enums.SfException.invalid_api_param},
            **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.failing = True


class ListVolumesForAccount(_base.BaseClusterApi):
    """ListVolumesForAccount request."""
    _version_added = sf_version.SfVersion('8.0')
    include_virtual_volumes = types.BooleanType(serialized_name='includeVirtualVolumes')
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID]}
    )

    def __init__(self, **kwargs):
        super(ListVolumesForAccount, self).__init__(
            result_dm=result_models.ListVolumesForAccount, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.failing = True


class ListVolumeStats(_base.BaseClusterApi):
    """ListVolumeStats request."""
    _version_added = sf_version.SfVersion('5.0')
    include_virtual_volumes = types.BooleanType(serialized_name='includeVirtualVolumes')
    volume_ids = types.ListType(
        field=types.IntIdType(),
        serialized_name=api_constants.VOLUME_IDS,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(ListVolumeStats, self).__init__(result_dm=result_models.ListVolumeStats, **kwargs)
        self.add_setup_api(CreateMultipleVolumes)
        self.add_inspection_api(ListActiveVolumes, list_all=True)


class ListVolumeStatsByAccount(_base.BaseClusterApi):
    """ListVolumeStatsByAccount request."""
    _version_added = sf_version.SfVersion('5.0')
    include_virtual_volumes = types.BooleanType(serialized_name='includeVirtualVolumes')
    accounts = types.ListType(
        field=types.IntIdType(),
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID]}
    )

    def __init__(self, **kwargs):
        super(ListVolumeStatsByAccount, self).__init__(
            result_dm=result_models.ListVolumeStatsByAccount, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.add_inspection_api(ListActiveVolumes, list_all=True)


class ListVolumeStatsByVirtualVolume(_base.BaseClusterApi):
    """ListVolumeStatsByVirtualVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    virtual_volume_ids = types.ListType(
        field=types.IntIdType(),
        serialized_name='virtualVolumeIDs',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VIRTUAL_VOLUME_ID]
        }
    )

    def __init__(self, **kwargs):
        super(ListVolumeStatsByVirtualVolume, self).__init__(
            result_dm=result_models.ListVolumeStatsByVirtualVolume, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.add_setup_api(_cluster.EnableFeature, vvols=True)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.failing = True


class ListVolumeStatsByVolume(_base.BaseClusterApi):
    """ListVolumeStatsByVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    include_virtual_volumes = types.BooleanType(serialized_name='includeVirtualVolumes')

    def __init__(self, **kwargs):
        super(ListVolumeStatsByVolume, self).__init__(
            result_dm=result_models.ListVolumeStatsByVolume, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.add_inspection_api(ListActiveVolumes, list_all=True)


class ListVolumeStatsByVolumeAccessGroup(_base.BaseClusterApi):
    """ListVolumeStatsByVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    include_virtual_volumes = types.BooleanType(serialized_name='includeVirtualVolumes')
    volume_access_groups = types.ListType(
        field=types.IntIdType(),
        serialized_name=api_constants.VOLUME_ACCESS_GROUPS,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )

    def __init__(self, **kwargs):
        super(ListVolumeStatsByVolumeAccessGroup, self).__init__(
            result_dm=result_models.ListVolumeStatsByVolumeAccessGroup, **kwargs
        )
        self.add_setup_api(CreateMultipleVolumes)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.failing = True


class ModifyBackupTarget(_base.BaseClusterApi):
    """ModifyBackupTarget request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType)
    backup_target_id = types.IntIdType(
        required=True,
        serialized_name='backupTargetID',
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: ['backupTargetID']}
    )
    name = types.StringType(required=True)

    def __init__(self, **kwargs):
        super(ModifyBackupTarget, self).__init__(**kwargs)
        self.add_setup_api(CreateBackupTarget)
        self.add_inspection_api(ListBackupTargets)
        self.failing = True


class ModifyGroupSnapshot(_base.BaseClusterApi):
    """ModifyGroupSnapshot request."""
    _version_added = sf_version.SfVersion('5.0')
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION)
    expiration_time = types.TimestampType(serialized_name=api_constants.EXPIRATION_TIME)
    group_snapshot_id = types.IntIdType(
        serialized_name=api_constants.GROUP_SNAPSHOT_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.GROUP_SNAPSHOT_ID]}
    )

    def __init__(self, **kwargs):
        super(ModifyGroupSnapshot, self).__init__(
            result_dm=result_models.ModifyGroupSnapshot, **kwargs
        )
        self.add_setup_api(CreateGroupSnapshot)


class InitiatorModified(models.Base):
    """Initiator data model.

    Corresponds to elements of the initiators object in the ModifyInitiators request.
    """
    alias = types.StringType()
    attributes = types.DictType(types.StringType())
    initiator_id = types.IntIdType(required=True, serialized_name=api_constants.INITIATOR_ID)
    volume_access_group_id = types.IntIdType(serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID)


class ModifyInitiators(_base.BaseClusterApi):
    """ModifyInitiators request."""
    _version_added = sf_version.SfVersion('5.0')
    initiators = types.ListType(types.ModelType(InitiatorModified), required=True)

    def __init__(self, **kwargs):
        super(ModifyInitiators, self).__init__(result_dm=result_models.ModifyInitiators, **kwargs)
        self.add_setup_api(CreateInitiators)
        self.add_inspection_api(ListInitiators)
        self.failing = True


class ModifyQoSPolicy(_base.BaseClusterApi):
    """ModifyQoSPolicy request."""
    _version_added = sf_version.SfVersion('10.0')
    qos_policy_id = types.IntIdType(
        required=True,
        serialized_name='qosPolicyID',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: ['qosPolicyID'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )
    name = types.StringType(
        min_length=1,
        max_length=64,
        serialized_name=api_constants.NAME,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_qos_policy_name,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_qos_policy_name,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_qos_policy_name}
        }
    )
    qos = types.ModelType(QoS, serialized_name=api_constants.QOS)

    def __init__(self, **kwargs):
        super(ModifyQoSPolicy, self).__init__(result_dm=result_models.ModifyQoSPolicy, **kwargs)
        self.add_setup_api(CreateQoSPolicy, default=True)
        self.add_inspection_api(ListQoSPolicies)


class ModifySchedule(_base.BaseClusterApi):
    """ModifySchedule request.

    :param bool delete: Whether to override the request to delete a schedule.
    """
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType)
    hours = types.IntType(min_value=0, max_value=23)
    minutes = types.IntType(min_value=0, max_value=59)
    monthdays = types.ListType(types.IntType(min_value=1, max_value=31))
    paused = types.BooleanType()
    recurring = types.BooleanType()
    run_next_interval = types.BooleanType(serialized_name='runNextInterval')
    schedule_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.SCHEDULE_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.SCHEDULE_ID]}
    )
    schedule_info = types.ModelType(models.ScheduleInfo, serialized_name=api_constants.SCHEDULE_INFO)
    schedule_name = types.StringType(serialized_name=api_constants.SCHEDULE_NAME)
    schedule_type = types.EnumType(enums.ScheduleType, serialized_name=api_constants.SCHEDULE_TYPE)
    starting_date = types.DateTimeType(serialized_name=api_constants.STARTING_DATE)
    to_be_deleted = types.BooleanType(serialized_name='toBeDeleted')
    weekdays = types.ListType(types.ModelType(models.Weekday), required=True, default=list)

    def __init__(self, delete=False, **kwargs):
        super(ModifySchedule, self).__init__(result_dm=result_models.ModifySchedule, **kwargs)
        self.add_setup_api(CreateSchedule)

        if delete:
            self.field_overrides[ModifySchedule.to_be_deleted] = True


class ModifySnapshot(_base.BaseClusterApi):
    """ModifySnapshot request."""
    _version_added = sf_version.SfVersion('5.0')
    enable_remote_replication = types.BooleanType(
        serialized_name=api_constants.ENABLE_REMOTE_REPLICATION)
    expiration_time = types.TimestampType(serialized_name=api_constants.EXPIRATION_TIME)
    snapshot_id = types.IntIdType(
        serialized_name=api_constants.SNAPSHOT_ID,
        metadata={api_constants.REF_REQUIRED: True,
                  api_constants.REF_PARAMS: [api_constants.SNAPSHOT_ID]}
    )

    def __init__(self, **kwargs):
        super(ModifySnapshot, self).__init__(result_dm=result_models.ModifySnapshot, **kwargs)
        self.add_setup_api(CreateSnapshot)


class ModifyVolume(_base.BaseClusterApi):
    """ModifyVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    access = types.EnumType(enums.VolumeAccess)
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )
    associate_with_qos_policy = types.BooleanType(
        serialized_name=api_constants.ASSOCIATE_WITH_QOS_POLICY
    )
    attributes = types.DictType(
        field=types.StringType,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    create_time = types.DateTimeType(serialized_name=api_constants.CREATE_TIME)
    qos = types.ModelType(
        model_spec=QoS,
        serialized_name=api_constants.QOS,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_invalid: enums.SfException.invalid_api_param,
            }
        }
    )
    qos_policy_id = types.IntIdType(
        serialized_name=api_constants.QOS_POLICY_ID,
        version_added=sf_version.SfVersion('10.0'),
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.QOS_POLICY_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )
    set_create_time = types.BooleanType(serialized_name=api_constants.SET_CREATE_TIME)
    total_size = types.IntType(
        required=True,
        serialized_name=api_constants.TOTAL_SIZE,
        min_value=1000000000,
        max_value=8000000491520,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID, api_constants.VOLUMES, 'volumeIDs'],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    # Define interdependent parameter relations
    qos.metadata[api_constants.CONFLICTING_FIELDS] = [qos_policy_id]
    qos_policy_id.metadata[api_constants.CONFLICTING_FIELDS] = [qos]

    def __init__(self, **kwargs):
        super(ModifyVolume, self).__init__(result_dm=result_models.ModifyVolume, **kwargs)
        self.add_setup_api(CreateVolume, default_qos=True)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.failing = True


class ModifyVolumeAccessGroup(_base.BaseClusterApi):
    """ModifyVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(
        field=types.StringType,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    delete_orphan_initiators = types.BooleanType(serialized_name='deleteOrphanInitiators')
    initiators = types.ListType(
        field=types.IntIdType(),
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.INITIATOR_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    name = types.StringType(
        required=True,
        min_length=1,
        max_length=64,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param}
        }
    )
    volumes = types.ListType(
        field=types.IntIdType,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )

    def __init__(self, **kwargs):
        super(ModifyVolumeAccessGroup, self).__init__(
            result_dm=result_models.ModifyVolumeAccessGroup, **kwargs
        )
        self.add_setup_api(CreateVolumeAccessGroup)
        self.add_inspection_api(ListVolumeAccessGroups)


class ModifyVolumes(_base.BaseClusterApi):
    """ModifyVolumes request."""
    _version_added = sf_version.SfVersion('5.0')
    access = types.EnumType(enums.VolumeAccess)
    account_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.ACCOUNT_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.account_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )
    associate_with_qos_policy = types.BooleanType(
        serialized_name=api_constants.ASSOCIATE_WITH_QOS_POLICY
    )
    attributes = types.DictType(
        field=types.StringType,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    create_time = types.DateTimeType(serialized_name=api_constants.CREATE_TIME)
    qos = types.ModelType(
        model_spec=QoS,
        serialized_name=api_constants.QOS
    )
    qos_policy_id = types.IntIdType(
        serialized_name=api_constants.QOS_POLICY_ID,
        version_added=sf_version.SfVersion('10.0'),
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.QOS_POLICY_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.qos_policy_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.qos_policy_id_does_not_exist}
        }
    )
    set_create_time = types.BooleanType(serialized_name=api_constants.SET_CREATE_TIME)
    total_size = types.IntType(
        required=True,
        serialized_name=api_constants.TOTAL_SIZE,
        min_value=1000000000,
        max_value=8000000491520,
        metadata={
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_api_param,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_api_param}
        }
    )
    volume_ids = types.ListType(
        field=types.IntIdType(),
        required=True,
        serialized_name=api_constants.VOLUME_IDS,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID, api_constants.VOLUMES, 'volumeIDs']
        }
    )

    # Define interdependent parameter relations
    qos.metadata[api_constants.CONFLICTING_FIELDS] = [qos_policy_id]
    qos_policy_id.metadata[api_constants.CONFLICTING_FIELDS] = [qos]

    def __init__(self, **kwargs):
        super(ModifyVolumes, self).__init__(result_dm=result_models.ModifyVolumes, **kwargs)
        self.add_setup_api(CreateMultipleVolumes, default_qos=True)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.failing = True


class PurgeDeletedVolume(_base.BaseClusterApi):
    """PurgeDeletedVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(PurgeDeletedVolume, self).__init__(**kwargs)
        self.add_setup_api(DeleteVolume)
        self.add_inspection_api(ListDeletedVolumes)


class PurgeDeletedVolumes(_base.BaseClusterApi):
    """PurgeDeletedVolumes request."""
    _version_added = sf_version.SfVersion('9.0')
    volume_ids = types.ListType(
        field=types.IntIdType,
        required=True,
        serialized_name=api_constants.VOLUME_IDS,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )
    account_ids = types.ListType(
        field=types.IntIdType,
        required=True,
        serialized_name='accountIDs',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.ACCOUNT_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.account_id_does_not_exist}
        }
    )
    volume_access_group_ids = types.ListType(
        field=types.IntIdType,
        required=True,
        serialized_name='volumeAccessGroupIDs',
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_access_group_id_does_not_exist}
        }
    )

    # Define interdependent parameter relations
    volume_ids.metadata[api_constants.CONFLICTING_FIELDS] = [account_ids, volume_access_group_ids]
    account_ids.metadata[api_constants.CONFLICTING_FIELDS] = [volume_ids, volume_access_group_ids]
    volume_access_group_ids.metadata[api_constants.CONFLICTING_FIELDS] = [volume_ids, account_ids]

    def __init__(self, **kwargs):
        super(PurgeDeletedVolumes, self).__init__(**kwargs)
        self.add_setup_api(DeleteVolumes)
        self.add_inspection_api(ListDeletedVolumes)
        self.add_inspection_api(ListVolumeAccessGroups, list_all=True)
        self.failing = True


class RemoveBackupTarget(_base.BaseClusterApi):
    """RemoveBackupTarget request."""
    _version_added = sf_version.SfVersion('5.0')
    backup_target_id = types.IntIdType(
        required=True,
        serialized_name='backupTargetID',
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: ['backupTargetID']}
    )

    def __init__(self, **kwargs):
        super(RemoveBackupTarget, self).__init__(**kwargs)
        self.add_setup_api(CreateBackupTarget)
        self.add_inspection_api(ListBackupTargets)


class RemoveInitiatorsFromVolumeAccessGroup(_base.BaseClusterApi):
    """RemoveInitiatorsFromVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    delete_orphan_initiators = types.BooleanType(serialized_name='deleteOrphanInitiators')
    initiators = types.ListType(
        field=types.IntIdType(),
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.INITIATOR_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_type_invalid: enums.SfException.invalid_api_param}
        }
    )
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )

    def __init__(self, **kwargs):
        super(RemoveInitiatorsFromVolumeAccessGroup, self).__init__(
            result_dm=result_models.RemoveInitiatorsFromVolumeAccessGroup, **kwargs
        )
        self.add_setup_api(CreateVolumeAccessGroup)
        self.add_inspection_api(ListVolumeAccessGroups)
        self.failing = True


class RemoveVolumesFromVolumeAccessGroup(_base.BaseClusterApi):
    """RemoveVolumesFromVolumeAccessGroup request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_access_group_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ACCESS_GROUP_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ACCESS_GROUP_ID]
        }
    )
    volumes = types.ListType(
        required=True,
        field=types.IntIdType(),
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.VOLUME_ID]}
    )

    def __init__(self, **kwargs):
        super(RemoveVolumesFromVolumeAccessGroup, self).__init__(
            result_dm=result_models.RemoveVolumesFromVolumeAccessGroup, **kwargs
        )
        self.add_setup_api(CreateVolumeAccessGroup)
        self.add_inspection_api(ListVolumeAccessGroups)
        self.failing = True


class RestoreDeletedVolume(_base.BaseClusterApi):
    """RestoreDeletedVolume request."""
    _version_added = sf_version.SfVersion('5.0')
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(RestoreDeletedVolume, self).__init__(**kwargs)
        self.add_setup_api(DeleteVolume)


class SetDefaultQoS(_base.BaseClusterApi):
    """SetDefaultQoS request.

    :param bool restore: Whether to override the request to restore default QOS settings.
    """
    _version_added = sf_version.SfVersion('5.0')
    min_iops = types.IntType(
        min_value=50,
        max_value=15000,
        default=50,
        serialized_name=api_constants.MIN_IOPS,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )
    max_iops = types.IntType(
        min_value=100,
        max_value=200000,
        default=15000,
        serialized_name=api_constants.MAX_IOPS,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )
    burst_iops = types.IntType(
        min_value=100,
        max_value=200000,
        default=15000,
        serialized_name=api_constants.BURST_IOPS,
        metadata={
            api_constants.NULL_INVALID: True,
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_min: enums.SfException.invalid_value,
                enums.ApiExceptionType.param_exceeded_max: enums.SfException.invalid_value}
        }
    )

    # Define interdependent parameter relations
    min_iops.metadata[api_constants.MAX_LIMIT_FIELD] = max_iops
    max_iops.metadata[api_constants.MIN_LIMIT_FIELD] = min_iops
    max_iops.metadata[api_constants.MAX_LIMIT_FIELD] = burst_iops
    burst_iops.metadata[api_constants.MIN_LIMIT_FIELD] = max_iops

    def __init__(self, restore=False, **kwargs):
        super(SetDefaultQoS, self).__init__(result_dm=result_models.SetDefaultQoS, **kwargs)

        if restore:
            self.field_overrides[SetDefaultQoS.min_iops] = 50
            self.field_overrides[SetDefaultQoS.max_iops] = 15000
            self.field_overrides[SetDefaultQoS.burst_iops] = 15000
        else:
            self.add_teardown_api(SetDefaultQoS, restore=True)


class StartBulkVolumeRead(_base.BaseClusterApi):
    """StartBulkVolumeRead request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType())
    format = types.EnumType(enums.IOFormat, required=True)
    script = types.StringType()
    script_parameters = types.DictType(types.StringType(), serialized_name='scriptParameters')
    snapshot_id = types.IntIdType(
        serialized_name=api_constants.SNAPSHOT_ID,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: [api_constants.SNAPSHOT_ID]
                  }
    )
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null:
                    enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(StartBulkVolumeRead, self).__init__(
            result_dm=result_models.StartBulkVolumeRead, async_timeout=300, **kwargs
        )
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.is_async = True


class StartBulkVolumeWrite(_base.BaseClusterApi):
    """StartBulkVolumeWrite request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType())
    format = types.EnumType(enums.IOFormat, required=True)
    script = types.StringType()
    script_parameters = types.DictType(types.StringType(), serialized_name='scriptParameters')
    volume_id = types.IntIdType(
        required=True,
        serialized_name=api_constants.VOLUME_ID,
        metadata={
            api_constants.REF_REQUIRED: True,
            api_constants.REF_PARAMS: [api_constants.VOLUME_ID],
            api_constants.EXCEPTION_OVERRIDES: {
                enums.ApiExceptionType.param_null: enums.SfException.volume_id_does_not_exist,
                enums.ApiExceptionType.param_value_does_not_exist:
                    enums.SfException.volume_id_does_not_exist}
        }
    )

    def __init__(self, **kwargs):
        super(StartBulkVolumeWrite, self).__init__(
            result_dm=result_models.StartBulkVolumeWrite, async_timeout=300, **kwargs
        )
        self.add_setup_api(CreateVolume)
        self.add_inspection_api(ListActiveVolumes, list_all=True)
        self.is_async = True


class UpdateBulkVolumeStatus(_base.BaseClusterApi):
    """UpdateBulkVolumeStatus request."""
    _version_added = sf_version.SfVersion('5.0')
    attributes = types.DictType(types.StringType())
    key = types.StringType(
        required=True,
        metadata={api_constants.REF_REQUIRED: True, api_constants.REF_PARAMS: ['key']}
    )
    message = types.StringType()
    percent_complete = types.StringType(serialized_name='percentComplete')
    status = types.EnumType(enums.AsyncResultStatus, required=True)

    def __init__(self, **kwargs):
        super(UpdateBulkVolumeStatus, self).__init__(
            result_dm=result_models.UpdateBulkVolumeStatus, wait_for_async_setup=False, **kwargs
        )
        self.add_setup_api(StartBulkVolumeRead)
