import unittest

import mock
import pytest
import requests

import sf_api
from sf_api import (
    api_errors,
    request_models,
    sf_version,
)


class TestBaseRequest(unittest.TestCase):

    def test_exists_in_version(self):
        class MockRequest(request_models.BaseApiRequest):
            _version_added = None

        request_dm = MockRequest()
        assert request_dm.api_exists_in_version(sf_version.SfVersion('11.2'))

        MockRequest._version_added = sf_version.SfVersion('11.1')
        assert request_dm.api_exists_in_version(sf_version.SfVersion('11.2'))

        MockRequest._version_added = sf_version.SfVersion('11.3')
        assert not request_dm.api_exists_in_version(sf_version.SfVersion('11.2'))

    def test_add_associated_api(self):
        request_dm = request_models.BaseApiRequest()
        assert not request_dm.setup_list
        assert not request_dm.inspection_list
        assert not request_dm.teardown_list

        request_dm.add_setup_api(request_models.BaseApiRequest, timeout=1)
        assert len(request_dm.setup_list) == 1
        assert request_dm.setup_list[0].timeout == 1
        assert not request_dm.setup_list[0]._skip_setup

        request_dm.add_inspection_api(request_models.BaseApiRequest, timeout=2)
        assert len(request_dm.inspection_list) == 1
        assert request_dm.inspection_list[0].timeout == 2
        assert request_dm.inspection_list[0]._skip_setup

        request_dm.add_teardown_api(request_models.BaseApiRequest, timeout=3)
        assert len(request_dm.teardown_list) == 1
        assert request_dm.teardown_list[0].timeout == 3
        assert request_dm.teardown_list[0]._skip_setup

    def test_level_properties(self):
        # Level 2
        request_dm = request_models.BaseApiRequest(timeout=29)
        assert request_dm.level_2
        assert not request_dm.level_3

        # Level 3
        request_dm = request_models.BaseApiRequest(timeout=31)
        assert not request_dm.level_2
        assert request_dm.level_3

        request_dm = request_models.BaseApiRequest()
        request_dm.is_async = True
        assert not request_dm.level_2
        assert request_dm.level_3

        request_dm = request_models.BaseApiRequest()
        request_dm.add_setup_api(request_models.BaseApiRequest, timeout=31)
        assert not request_dm.level_2
        assert request_dm.level_3

        # Not Tested
        request_dm = request_models.BaseApiRequest()
        request_dm.cant_test = True
        assert not request_dm.level_2
        assert not request_dm.level_3

        request_dm = request_models.BaseApiRequest()
        request_dm.failing = True
        assert not request_dm.level_2
        assert not request_dm.level_3

    @mock.patch.object(sf_api.BaseJsonRpcClient, 'call_method_raw')
    @mock.patch.object(sf_api.BaseJsonRpcClient, 'authenticate', mock.Mock())
    @mock.patch.object(api_errors, 'get_exception_class_from_string',
                       mock.Mock(return_value=api_errors.AssertionFailure))
    @mock.patch('time.sleep', mock.Mock())
    def test_call(self, call_method_raw):
        api_server = sf_api.ClusterApi('10.117.78.10')
        request_dm = request_models.BaseClusterApi()
        mock_params = {'mock_params': True}
        mock_result = {'mock_result': True}
        mock_error = {'name': 'mockError', 'message': 'mockMessage'}

        # Standard result
        call_method_raw.return_value = {'result': mock_result}
        result = request_dm.call(api_server, mock_params)
        assert result == mock_result
        call_method_raw.assert_called_with('BaseClusterApi', mock_params)

        # Error
        call_method_raw.return_value = {'error': mock_error}
        with pytest.raises(api_errors.AssertionFailure):
            request_dm.call(api_server, mock_params)

        # Retries
        call_method_raw.call_count = 0
        call_method_raw.side_effect = [requests.ConnectionError, {'result': mock_result}]
        result = request_dm.call(api_server, mock_params)
        assert result == mock_result
        assert call_method_raw.call_count == 2
