import json
import os
import unittest

import mock
import pytest
import requests

import sf_api
from sf_api import (
    api_errors,
    enums,
    models,
)

MNODE_MIP = '1.2.3.4'
CLUSTER_MVIP = '2.3.4.5'
RS_CLUSTER_DM = models.ResourceCluster.get_mock_object(overrides={'mvip': CLUSTER_MVIP})
RS_MNODE_DM = models.ResourceManagementNode.get_mock_object(overrides={'mip': MNODE_MIP})
# Requests can't prep a request on an ipv6-addressed URL.

CWD, _ = os.path.split(os.path.abspath(os.path.expanduser(__file__)))
GET_RESULTS = os.path.join(CWD, 'data', 'rest_healthcheck_get.json')


class TestOauth2Auth(unittest.TestCase):
    def setUp(self):
        self.mock_cluster_dm = RS_CLUSTER_DM
        self.client_id = 'mnode_client'
        self.base_url = 'https://{}/mnode/'.format('10.117.64.202')

    @mock.patch.object(requests, 'post')
    def test_oauth2_repr(self, mock_post):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {'result': {'access_token':
                                                      'thisisanaccesstoken',
                                                      'expires_in': 600}}
        mock_post.return_value = mock_response
        auth = sf_api.Oauth2Auth(self.mock_cluster_dm, self.client_id, self.base_url)
        assert repr(auth) == 'Oauth2Auth<token=None>'
        _ = auth.bearer_token
        assert repr(auth) == 'Oauth2Auth<token=thisisanaccesstoken>'

    @mock.patch.object(requests, 'post')
    def test_oauth2_get_bearer_token(self, mock_post):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {'result': {'access_token':
                                                      'thisisanaccesstoken',
                                                      'expires_in': 600}}
        mock_post.return_value = mock_response
        auth = sf_api.Oauth2Auth(self.mock_cluster_dm, self.client_id, self.base_url)
        assert auth.bearer_token == 'thisisanaccesstoken'

    @mock.patch.object(requests, 'post')
    def test_oauth2_token_header(self, mock_post):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {'result': {'access_token':
                                                      'thisisanaccesstoken',
                                                      'expires_in': 600}}
        mock_post.return_value = mock_response
        auth = sf_api.Oauth2Auth(self.mock_cluster_dm, self.client_id, self.base_url)
        assert auth(requests.Response()).headers == {'Authorization':
                                                     'Bearer thisisanaccesstoken'}

    @mock.patch.object(requests, 'post')
    def test_oauth2_is_expired(self, mock_post):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {'result': {'access_token':
                                                      'thisisanaccesstoken',
                                                      'expires_in': 0}}
        mock_post.return_value = mock_response
        auth = sf_api.Oauth2Auth(self.mock_cluster_dm, self.client_id, self.base_url)
        _ = auth.bearer_token
        assert auth.is_expired

    @mock.patch.object(requests, 'post')
    def test_oauth2_is_not_expired(self, mock_post):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {'result': {'access_token':
                                                      'thisisanaccesstoken',
                                                      'expires_in': 600}}
        mock_post.return_value = mock_response
        auth = sf_api.Oauth2Auth(self.mock_cluster_dm, self.client_id, self.base_url)
        _ = auth.bearer_token
        assert not auth.is_expired


class TestBaseRestClient(unittest.TestCase):
    def setUp(self):
        self.mock_cluster_dm = RS_CLUSTER_DM
        self.mock_mnode_dm = RS_MNODE_DM
        self.client_id = 'mnode_client'
        self.base_url = MNODE_MIP
        self.client = sf_api.BaseRestClient(cluster_dm=self.mock_cluster_dm,
                                            client_id=self.client_id,
                                            base_url=self.base_url,
                                            authenticate=False,
                                            )

    def test_get_method(self):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {'result': 'success!'}
        mock_response.status_code = 200
        with mock.patch('requests.Session.send', return_value=mock_response) as mock_get:
            response = self.client.get(service='mnode', path='1', params={'one': 'two'})
            url = 'https://{}/{}/{}?{}'.format(self.base_url, 'mnode', str(1), 'one=two')
            assert 'success!' in response.result.values()
            mock_get.assert_called_once()
            assert mock_get.call_args[0][0].method == 'GET'
            assert mock_get.call_args[0][0].url == url

    def test_post_method(self):
        mock_response = mock.Mock(requests.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {'result': 'success!'}
        with mock.patch('requests.Session.send', return_value=mock_response) as mock_post:
            response = self.client.post('mnode', 'fake/path', {'dontstop': 'continue'})
            url = 'https://{}/{}/{}'.format(self.base_url, 'mnode', 'fake/path')
            assert 'success!' in response.result.values()
            mock_post.assert_called_once()
            assert mock_post.call_args[0][0].method == 'POST'
            assert mock_post.call_args[0][0].url == url

    def test_put_method(self):
        mock_response = mock.Mock(requests.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {'result': 'success!'}
        with mock.patch('requests.Session.send', return_value=mock_response) as mock_put:
            response = self.client.put('mnode', path='new/fake/path', data={'one': 'two'})
            url = 'https://{}/{}/{}'.format(self.base_url, 'mnode', 'new/fake/path')
            assert 'success!' in response.result.values()
            mock_put.assert_called_once()
            assert mock_put.call_args[0][0].method == 'PUT'
            assert mock_put.call_args[0][0].url == url

    def test_delete_method(self):
        mock_response = mock.Mock(requests.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {'result': 'success!'}
        with mock.patch('requests.Session.send', return_value=mock_response) as mock_delete:
            response = self.client.delete('mnode', path='/path')
            url = 'https://{}/{}/{}'.format(self.base_url, 'mnode', 'path')
            assert 'success!' in response.result.values()
            mock_delete.assert_called_once()
            assert mock_delete.call_args[0][0].method == 'DELETE'
            assert mock_delete.call_args[0][0].url == url

    def test_process_response_valid(self):
        mock_response = mock.Mock(requests.Response)
        mock_response.text = 'blah'
        mock_response.json.return_value = {'result': 'success!'}
        response = self.client._process_response(mock_response)
        assert response.result is not None
        assert isinstance(response, sf_api.RestResponse)

    def test_process_response_error(self):
        mock_response = mock.Mock(requests.Response)
        mock_response.status_code = 400
        mock_response.raise_for_status.side_effect = requests.exceptions.HTTPError()
        with pytest.raises(requests.HTTPError):
            self.client._process_response(mock_response)

    def test_get_method_result_dm(self):
        with open(GET_RESULTS) as f:
            text = f.read()
        text_json = json.loads(text)
        r = models.HealthcheckResultModel(text_json)

        mock_response = mock.Mock(requests.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = text_json

        with mock.patch('requests.Session.send', return_value=mock_response):
            response = self.client.get(service='mnode', path='path',
                                       result_dm_class=models.HealthcheckResultModel)
            assert isinstance(response.result_dm, models.HealthcheckResultModel)
            assert response.result_dm == r
            assert response.result_dm.state == enums.HealthcheckState.error
            assert response.result_dm.status.status == enums.HealthcheckStatus.running

    def test_get_url(self):
        with open(GET_RESULTS) as f:
            text = f.read()
        text_json = json.loads(text)
        r = models.HealthcheckResultModel(text_json)

        mock_response = mock.Mock(requests.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = text_json

        with mock.patch('requests.Session.send', return_value=mock_response):
            response = self.client.get(url='http://blah.yada',
                                       result_dm_class=models.HealthcheckResultModel)
            assert isinstance(response.result_dm, models.HealthcheckResultModel)
            assert response.result_dm == r
            assert response.result_dm.state == enums.HealthcheckState.error
            assert response.result_dm.status.status == enums.HealthcheckStatus.running

    def test_bad_get(self):
        with pytest.raises(AssertionError):
            self.client.get(url='http://blah', service='mnode', path='1', params={'one': 'two'})

        with pytest.raises(AssertionError):
            self.client.get(params={'one': 'two'})
