import time
import unittest

import aenum
import mock
import pytest
import requests
import urllib3

import sf_api
from sf_api import (
    SfErrorResponse,
    SfResponse,
    api_errors,
    enums,
    get_json_rpc_id,
    models,
    result_models,
)


class BaseJsonRpcClient(sf_api.BaseJsonRpcClient):
    """Define abstract property so class can be instantiated."""

    def __init__(self, ip, username=None, password=None, version=None, authenticate=False,
                 retry_sf_exceptions_override=None, retry_count=2):
        super(BaseJsonRpcClient, self).__init__(ip, username, password, version, authenticate,
                                                retry_sf_exceptions_override, retry_count)


class DummyMethodEnum(aenum.Enum):
    unreal_api_method = 'unrealApiMethod'


class TestBaseJsonRpcClient(unittest.TestCase):
    def setUp(self):
        # Reset auth cache
        self.ip = '106.84.228.182'
        self.ip2 = '78.232.47.98'
        self.valid_response = {'id': 1, 'result': {'param': True}}
        self.invalid_response = {'id': 1}
        self.error_response = {
            'id': 1,
            'error': {
                'code': 1, 'name': 'xUnknownException', 'message': 'Something broke'
            }
        }
        self.valid_version_response = {
            'id': 1,
            'result': {
                'param': True, 'currentVersion': '9.0'
            }
        }

    def test_sf_auth_call(self):
        requests_response = mock.Mock(requests.Response)
        requests_response.headers = {}

        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '1.0')
        sf_auth = client.get_auth()
        assert 'Authorization' not in requests_response.headers
        sf_auth(requests_response)
        assert 'Authorization' in requests_response.headers

    def test_client_root_uri(self):
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '1.0')
        assert client.root_uri == 'https://{}'.format(self.ip)

    def test_auth_root_uri(self):
        auth = sf_api.SfAuth(self.ip, '3333', 'user', 'pass', '1.0')
        assert auth.root_uri == 'https://{}:{}'.format(self.ip, '3333')

    def test_sf_error_response_repr(self):
        sf_error_response = SfErrorResponse(self.error_response)
        assert repr(sf_error_response) is not None

    @mock.patch.object(requests, 'post')
    def test_get_version(self, mock_post):
        requests_response = mock.Mock(requests.Response)
        requests_response.json.return_value = self.valid_version_response
        mock_post.return_value = requests_response
        auth = sf_api.SfAuth(self.ip, '3333', 'user', 'pass', '1.0')
        sf_version = auth.get_version()
        assert sf_version.version_string == '9.0'

    @mock.patch.object(sf_api.SfAuth, 'get_version')
    def test_version_provided(self, mock_get_version):
        version = sf_api.SfVersion('12.123')
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', version=version)

        assert client.version == version
        assert not mock_get_version.called

    def test_process_response_valid(self):
        requests_response = mock.Mock(requests.Response)
        requests_response.json.return_value = self.valid_response
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        response = client._process_response(requests_response, enums.ClusterMethod.list_all_nodes)
        assert response.result is not None
        assert isinstance(response, sf_api.SfResponse)

    def test_process_response_invalid(self):
        requests_response = mock.Mock(requests.Response)
        requests_response.json.return_value = self.invalid_response
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        with pytest.raises(api_errors.MalformedResponseError):
            client._process_response(requests_response, enums.ClusterMethod.list_all_nodes)

    def test_process_response_with_error(self):
        requests_response = mock.Mock(requests.Response)
        requests_response.json.return_value = self.error_response
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        with pytest.raises(api_errors.SfApiException):
            client._process_response(requests_response, enums.ClusterMethod.list_all_nodes)

    def test_response_no_result_dm(self):
        requests_response = mock.Mock(requests.Response)
        requests_response.json.return_value = self.valid_response
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        response = client._process_response(requests_response, DummyMethodEnum.unreal_api_method)
        assert response.result_dm is None
        assert response.method is DummyMethodEnum.unreal_api_method
        assert response.raw_data == self.valid_response
        assert response.result == self.valid_response['result']
        assert response.error is None

    def test_response_with_result_dm(self):
        list_vol_dm = result_models.ListVolumes.get_mock_object()
        self.valid_response['result'] = list_vol_dm.serialize()
        requests_response = mock.Mock(requests.Response)
        requests_response.json.return_value = self.valid_response
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        response = client._process_response(requests_response, enums.ClusterMethod.list_volumes,
                                            result_models.ListVolumes)
        result_dm = response.result_dm
        assert response.result_dm is result_dm
        assert response.result_dm == list_vol_dm
        assert response.method == enums.ClusterMethod.list_volumes
        assert response.raw_data == self.valid_response
        assert response.result == self.valid_response['result']
        assert response.error is None
        assert isinstance(repr(response), str)

    @mock.patch.object(sf_api.BaseJsonRpcClient, 'authenticate')
    def test_authenticate_called_in_constructor(self, mock_authenticate):
        BaseJsonRpcClient(self.ip, 'name', 'name', '1.0', authenticate=True)
        assert mock_authenticate.called

    @mock.patch.object(sf_api.BaseJsonRpcClient, 'authenticate')
    def test_authenticate_called_in_instance(self, mock_authenticate):
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        client.authenticate()
        assert mock_authenticate.called

    @mock.patch('sf_api._base.requests.Session.get')
    def test_authenticate_with_http_success(self, mock_get):
        mock_get.return_value.ok = True
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        response = client.authenticate()
        assert response.ok is True

    @mock.patch('sf_api._base.requests.Session.get')
    def test_authenticate_with_request_exception(self, mock_get):
        mock_get.side_effect = requests.exceptions.RequestException()
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        with pytest.raises(requests.exceptions.RequestException):
            client.authenticate()

    @mock.patch('sf_api._base.SfAuth.get_version')
    @mock.patch('sf_api._base.requests.Session.get')
    def test_authenticate_no_version(self, mock_get, mock_get_version):
        mock_get.return_value.ok = True
        mock_get_version.return_value = sf_api.SfVersion('11.0')
        client = BaseJsonRpcClient(self.ip)
        client.authenticate()
        uri = mock_get.call_args[0][0]
        assert 'None' not in uri
        assert '11.0' in uri

    @mock.patch('sf_api._base.requests.Session.get')
    def test_authenticate_with_http_error(self, mock_get):
        mock_response = mock.Mock()
        mock_response.ok = False
        mock_response.raise_for_status.side_effect = requests.exceptions.HTTPError()
        mock_get.return_value = mock_response
        client = BaseJsonRpcClient(self.ip, 'name', 'name', '1.0')
        with pytest.raises(requests.exceptions.HTTPError):
            client.authenticate()

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch('sf_api._base.SfAuth.get_version')
    def test_call_method_with_no_version(self, mock_get_version, mock_post, mock_process_response):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        mock_process_response.side_effect = [mock_response]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0')

        # Remove version to trigger _get_version() call
        client.version = None
        mock_get_version.return_value = '8.0'
        response = client.call_method(enums.ClusterMethod.get_cluster_info)

        assert isinstance(response, SfResponse)
        assert client.version == '8.0'

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep')
    def test_max_retries(self, mock_sleep, mock_post, mock_process_response):
        """The max number of retries is hit, so an exception is raised."""
        mock_process_response.side_effect = api_errors.DBOperationTimeout('msg', {})
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_sf_exceptions_override=(api_errors.DBOperationTimeout,),
                                   retry_count=1)
        with pytest.raises(api_errors.DBOperationTimeout):
            client.call_method(enums.ClusterMethod.get_cluster_info)
        assert mock_post.call_count == 2

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep')
    def test_retry_success_with_override(self, mock_sleep, mock_post, mock_process_response):
        """The retry happens because it's in the retry fault list."""
        mock_response = result_models.GetClusterInfo.get_mock_response()
        mock_process_response.side_effect = [api_errors.ExceededLimit('msg', {}), mock_response]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_sf_exceptions_override=(api_errors.ExceededLimit,),
                                   retry_count=1)
        response = client.call_method(enums.ClusterMethod.get_cluster_info)
        assert isinstance(response, SfResponse)

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep')
    def test_retry_success_without_override(self, mock_sleep, mock_post, mock_process_response):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        # DBConnectionLoss is a global retry
        mock_process_response.side_effect = [api_errors.DBConnectionLoss('msg', {}), mock_response]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_count=1)
        response = client.call_method(enums.ClusterMethod.get_cluster_info)
        assert isinstance(response, SfResponse)

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep')
    def test_retry_success_with_req_error(self, mock_sleep, mock_post, mock_process_response):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        # DBConnectionLoss is a global retry
        mock_process_response.side_effect = [api_errors.DBConnectionLoss('msg', {}), mock_response]
        mock_post.side_effect = [requests.ConnectionError(), mock.MagicMock(requests.Response),
                                 mock.MagicMock(requests.Response)]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_count=2)
        response = client.call_method(enums.ClusterMethod.get_cluster_info)
        assert isinstance(response, SfResponse)

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep', mock.MagicMock())
    def test_retry_success_with_http_error(self, mock_post, mock_process_response):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        mock_http_error_response = mock.MagicMock(requests.Response)
        mock_http_error_response.raise_for_status.side_effect = requests.HTTPError()
        mock_post.side_effect = [mock_http_error_response, mock_http_error_response,
                                 mock.MagicMock(requests.Response)]
        mock_process_response.side_effect = [mock_response]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_count=2)
        response = client.call_method(enums.ClusterMethod.get_cluster_info)
        assert isinstance(response, SfResponse)
        assert mock_post.call_count == 3

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep')
    def test_retry_success_http_error_min_retry(self, mock_sleep, mock_post, mock_process_response):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        mock_http_error_response = mock.MagicMock(requests.Response)
        mock_http_error_response.raise_for_status.side_effect = requests.HTTPError()
        mock_post.side_effect = [mock_http_error_response, mock.MagicMock(requests.Response)]
        mock_process_response.side_effect = [mock_response]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_count=2)
        client.min_retry_delay = 5
        response = client.call_method(enums.ClusterMethod.get_cluster_info)
        assert isinstance(response, SfResponse)
        assert mock_post.call_count == 2
        mock_sleep.assert_called_once_with(5)

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    @mock.patch.object(time, 'sleep', mock.MagicMock())
    def test_retry_success_with_read_time_out_error(self, mock_post, mock_process_response):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        mock_read_time_out_error = mock.MagicMock(requests.Response)
        mock_read_time_out_error.raise_for_status.side_effect = requests.ReadTimeout()
        mock_post.side_effect = [mock_read_time_out_error, mock_read_time_out_error,
                                 mock.MagicMock(requests.Response)]
        mock_process_response.side_effect = [mock_response]
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_count=2)
        response = client.call_method(enums.ClusterMethod.get_cluster_info)
        assert isinstance(response, SfResponse)
        assert mock_post.call_count == 3

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    def test_num_retries_is_zero(self, mock_post, mock_process_response):
        """Exception is in the fault list, but retries is set to zero."""
        mock_process_response.side_effect = api_errors.DBOperationTimeout('msg', {})
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_sf_exceptions_override=(api_errors.DBOperationTimeout,),
                                   retry_count=0)
        with pytest.raises(api_errors.DBOperationTimeout):
            client.call_method(enums.ClusterMethod.get_cluster_info)
        assert mock_post.call_count == 1

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    def test_retry_exceptions_override_is_empty_tuple(self, mock_post, mock_process_response):
        """This will override it to retry on no exceptions."""
        mock_process_response.side_effect = api_errors.DBOperationTimeout('msg', {})
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_sf_exceptions_override=(),
                                   retry_count=1)
        with pytest.raises(api_errors.DBOperationTimeout):
            client.call_method(enums.ClusterMethod.get_cluster_info)
        assert mock_post.call_count == 1

    @mock.patch('sf_api._base.BaseJsonRpcClient._process_response')
    @mock.patch('sf_api._base.requests.Session.post')
    def test_retry_exception_not_in_ignored_list(self, mock_post, mock_process_response):
        """The retry does not occur since it is not in the retry_fault list."""
        mock_process_response.side_effect = api_errors.DBOperationTimeout('msg', {})
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0',
                                   retry_sf_exceptions_override=(api_errors.DBPathNotEmpty,),
                                   retry_count=2)
        with pytest.raises(api_errors.DBOperationTimeout):
            client.call_method(enums.ClusterMethod.get_cluster_info)
        assert mock_post.call_count == 1

    @mock.patch('sf_api._base.requests.Session.get')
    @mock.patch('sf_api._base.time.sleep')
    def test_retry_on_http_get(self, mock_sleep, mock_get):
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0', retry_count=3)

        valid_response = requests.Response()
        valid_response.status_code = 200

        invalid_response = requests.Response()
        invalid_response.status_code = 404

        mock_get.side_effect = [invalid_response,
                                invalid_response,
                                invalid_response,
                                valid_response]
        response = client.http_get('/some/path', retry_delay=1)
        mock_sleep.assert_has_calls([mock.call(1), mock.call(1), mock.call(1)])
        assert isinstance(response, requests.Response)

    @mock.patch('sf_api._base.requests.Session.get')
    @mock.patch('sf_api._base.time.sleep')
    def test_exceed_retries_on_http_get(self, mock_sleep, mock_get):
        client = BaseJsonRpcClient(self.ip, 'user', 'pass', '9.0', retry_count=2)

        valid_response = requests.Response()
        valid_response.status_code = 200

        invalid_response = requests.Response()
        invalid_response.status_code = 404

        mock_get.side_effect = [invalid_response,
                                invalid_response,
                                invalid_response,
                                valid_response]
        with pytest.raises(requests.exceptions.RequestException):
            client.http_get('/some/path', retry_delay=1)

    def test_construct_params_removes_none_value(self):
        result = BaseJsonRpcClient._construct_params(dict(force=True, set_to_none=None))
        assert result == {'force': True}

    def test_construct_params_converts_single_enum(self):
        result = BaseJsonRpcClient._construct_params(
            dict(force=True, account_type=enums.AccountType.standard))
        assert result == {'force': True, 'account_type': enums.AccountType.standard.value}

    def test_construct_params_converts_list_of_enums(self):
        result = BaseJsonRpcClient._construct_params(dict(force=False,
                                                          account_types=[enums.AccountType.standard,
                                                                         enums.AccountType.vvol]))
        assert result == {'force': False,
                          'account_types': [
                              enums.AccountType.standard.value, enums.AccountType.vvol.value]
                          }


class TestGetMockResponse(unittest.TestCase):
    def test_id_count(self):
        mock_response = result_models.GetClusterInfo.get_mock_response()
        id1 = mock_response.id
        mock_response = result_models.GetClusterInfo.get_mock_response()
        assert id1 == mock_response.id - 1

    def test_mock_sf_response_with_result_dm(self):
        mock_volume = models.Volume.get_mock_object()
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=[mock_volume]))
        assert isinstance(mock_response, SfResponse)
        assert mock_response.result == {'volumes': [mock_volume.serialize()]}


class TestGetJsonRpcID(unittest.TestCase):
    def test_get_json_rpc_id(self):
        rpc_id_list = []
        rpc_id_set = set()
        for x in range(10):
            rpc_id = get_json_rpc_id()
            rpc_id_list.append(rpc_id)
            rpc_id_set.add(rpc_id)
        # test that ids are not repeated
        assert len(rpc_id_set) == 10
        # test that ids iterate lowest to highest
        assert rpc_id_list == sorted(rpc_id_list)
