"""Contains the Base class that all models inherit from."""
import abc
import copy
import random

import six
from schematics import (
    exceptions,
    models,
)

from sf_models import (
    errors,
    sf_version,
    types,
)


def enable_strict_validation():
    """Enable strict model validation.

    Any API response with a 'rogue' attribute in the JSON object will call an exception to be raised
    on ``validate``. A rogue attribute is an attribute not defined on the model but received it when
    deserializing.

    A test should only enable this once at the very beginning. This allows the tester to get more
    implicit API testing. It may not make sense to leave this set in every test, but a test could be
    run occasionally where this flag is enabled.


    .. warning:: This applies to all models!
    """
    Base.STRICT_MODELS = True


def disable_strict_validation():
    """Disable strict model validation.

    All 'rogue' attributes when deserializing data models.
    """
    Base.STRICT_MODELS = False


class Base(models.Model):
    """Base data model that inherits from schematics.Model.

    :param raw_data: The data used to initialize the model
    :type raw_data: dict[str, Any]
    :param resource_dm: Resource data model. This is required for models that have an associated
                        resource model. The resource data model is specified by the static variable
                        _resource_dm_class.
    :type resource_dm: sf_models.Base
    :raises errors.DataValidationError: If the constructor cannot initialize properly due to
                                        improper args.
    """

    STRICT_MODELS = False
    # Models that have an associated resource must set this variable
    _resource_dm_class = None

    # Data models are mutable and therefore are not hashable.
    __hash__ = None

    def __init__(self, raw_data=None, resource_dm=None, **kwargs):
        if 'strict' not in kwargs:
            # Set strict to False, so it ignores unexpected fields. During API testing, this will
            # be set to True.
            kwargs['strict'] = self.STRICT_MODELS
        self._resource_dm = resource_dm
        try:
            super(Base, self).__init__(raw_data, **kwargs)
        except ValueError as err:
            raise errors.DataValidationError({'ValueError': err.args[0]})
        except TypeError as err:
            raise errors.DataValidationError({'TypeError': err.args[0]})
        except exceptions.DataError as err:
            raise errors.DataValidationError(dict(err.errors))

    @staticmethod
    def _field_exists_in_version(field, version):
        """Determine if a field exists for a given ElementOS version.

        :param sf_models.types.BaseType field: The field object
        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: sf_models.SfVersion
        :return: True if the field is exists for the given version, False otherwise
        :rtype: bool
        """
        if field.version_added is not None:
            if field.version_added > version:
                return False
        if field.version_removed is not None:
            if field.version_removed <= version:
                return False
        return True

    @classmethod
    def _get_field_list(cls):
        """Returns field list.

        :return: List of field list tuples (<field name>, <field instance>)
        :rtype: list[tuple[str, sf_models.Base]]
        """
        return list(six.iteritems(cls._schema.fields))  # pylint: disable=no-member

    @classmethod
    def get_mock_object(cls, context=None, overrides=None, version=None,
                        include_optional=None, force_mock=None):
        """Get a mock object.

        :param schematics.datastructures.Context context: A Context object that encapsulates
            configuration options and app_data. Primarily used internally by schematics.
        :param overrides: Overrides for the model. This should be a nested dictionary of field names in
            which the particular fields to override are mapped to the values with which they are to be
            overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}``
        :type overrides: dict[str, Any]
        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: sf_models.SfVersion
        :param bool include_optional: Whether to mock non-required fields. None randomly mocks optional
            fields.
        :param force_mock: Force the mocking of a particular field. This should be a nested dictionary
            of field names in which the particular fields to be mocked are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_mock': True}}}``
        :type force_mock: dict[str, dict|bool]
        :return: The data model instance
        :rtype: Any[sf_models.Base]
        """
        overrides = copy.copy(overrides) or {}
        for name, field in cls._get_field_list():
            # Override all fields either not in overrides or in overrides as a parent (a field in
            # overrides is considered a parent if it is a ModelType or ListType mapped to a dictionary)
            if name not in overrides or (type(overrides.get(name)) is dict and
                                         (isinstance(field, types.ModelType) or
                                          isinstance(field, types.ListType))):

                # Force fields to be mocked which have been explicitly specified
                if isinstance(force_mock, dict):
                    force_mock_value = force_mock.get(name)
                else:
                    force_mock_value = force_mock

                # Determine whether the API version is valid, and whether or not we will force override
                # the field value.
                version_valid, force_override = cls._determine_version_valid_and_force_override(
                    name, field, overrides, version, include_optional, force_mock_value)

                # Set the override value based on the field type.
                cls._set_override_value(name, field, overrides, version_valid, version, include_optional,
                                        force_override, force_mock_value)

        context = context  # type: dict
        obj = super(Base, cls).get_mock_object(context=context, overrides=overrides)
        if cls._resource_dm_class:
            # Mock associated resource data model if one exists
            # pylint: disable=protected-access
            obj._resource_dm = cls._resource_dm_class.get_mock_object(overrides=overrides)

        return obj

    @classmethod
    def _determine_version_valid_and_force_override(cls, name, field, overrides, version,
                                                    include_optional, force_mock_value):
        """Determine if the API version is valid and if we will force override the field value.

        :param str name: Field name
        :param sf_models.types.ModelType field: Field structure
        :param dict[str,Any] overrides: Overrides for the model. This should be a nested dictionary of
            field names in which the particular fields to override are mapped to the values with which
            they are to be overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}``
        :param sf_models.SfVersion version: SolidFire API version to mock.
        :param bool include_optional: Whether to mock non-required fields (randomly mocks optional
            fields by default).
        :param dict|bool force_mock_value: Value from a mock field.

        :return: tuple containing whether the API version is valid, and if we should force override the
            field value.
        :rtype: tuple[bool,bool]
        """

        force_override = include_optional

        # For force mock value, go ahead and override version validity and
        # set the force_override flag.
        if type(force_mock_value) is dict or force_mock_value is True:
            version_valid = True  # Override version validity if forced
            force_override = True
        elif version is not None:
            version_valid = cls._field_exists_in_version(field, version)
        else:
            version_valid = True

        # Force mocking of parameters in overrides
        if name in overrides:
            force_override = True

        # Randomly mock optional parameters if include_optional is None and none of
        # the earlier pre-conditions set the value.
        if force_override is None:
            force_override = random.choice([True, False])

        return version_valid, force_override

    @classmethod
    def _set_override_value(cls, name, field, overrides, version_valid, version, include_optional,
                            force_override, force_mock_value):
        """Set the field override data based on the type and any potential forced data.

        :param str name: Field name
        :param sf_models.types.ModelType field: Field structure
        :param dict[str,Any] overrides: Overrides for the model. This should be a nested dictionary of
            field names in which the particular fields to override are mapped to the values with which
            they are to be overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}``
        :param bool version_valid: Whether the API version is valid to use for overriding.
        :param sf_models.SfVersion version: SolidFire API version to mock.
        :param bool include_optional: Whether to mock non-required fields (randomly mocks optional
            fields by default).
        :param bool force_override: Whether to force override the field value
        :param dict|bool force_mock_value: Value from a mock field.
        """

        # Override version-invalid fields to None
        if not version_valid:
            overrides[name] = None
        # Override non-forced optional fields to Default (or None if no Default exists)
        elif not field.required and not force_override:
            overrides[name] = None
            if field.default is not models.Undefined:
                overrides[name] = field.default
        # Mock model-type fields
        elif isinstance(field, types.ModelType):
            nested_model = field.model_class
            overrides[name] = nested_model.get_mock_object(
                overrides=overrides.get(name),
                version=version,
                include_optional=include_optional,
                force_mock=force_mock_value
            )
        # Mock list-type fields
        elif isinstance(field, types.ListType) or isinstance(field, types.StringListType):
            cls._mock_list_type(name, field, overrides, version, include_optional,
                                force_mock_value)

        elif field.choices is not None:
            overrides[name] = random.choice(field.choices)
        else:
            overrides[name] = field._mock()  # pylint: disable=protected-access

    @classmethod
    def _mock_list_type(cls, name, field, overrides, version, include_optional, force_mock_value):
        """Fill in the override data for a field expecting list values.

        :param str name: Field name
        :param sf_models.types.ModelType field: Field structure
        :param dict[str,Any] overrides: Overrides for the model. This should be a nested dictionary of
            field names in which the particular fields to override are mapped to the values with which
            they are to be overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}``
        :param sf_models.SfVersion version: SolidFire API version to mock.
        :param bool include_optional: Whether to mock non-required fields (randomly mocks optional
            fields by default).
        :param dict|bool force_mock_value: Value from a mock field.

        :return: tuple containing whether the API version is valid, and if we should force override the
            field value.
        """

        min_size = field.min_size or 1
        max_size = field.max_size or random.randint(min_size, min_size + 1)
        if min_size > max_size:
            message = 'Minimum list size is greater than maximum list size.'
            raise exceptions.MockCreationError(message)

        mock_size = random.randint(min_size, max_size)
        # Mock model-type elements of list-type fields
        if isinstance(field.field, types.ModelType):
            overrides[name] = field.field.model_class.get_mock_object_list(
                length=mock_size,
                overrides=overrides.get(name),
                version=version,
                include_optional=include_optional,
                force_mock=force_mock_value
            )
        else:
            overrides[name] = []
            for _ in range(mock_size):
                # pylint: disable=protected-access
                overrides[name].append(field.field._mock())

    @classmethod
    def get_mock_object_list(cls, length, context=None, overrides=None, version=None,
                             include_optional=None, force_mock=None):
        """Get a list of mock objects.

        :param int length: Number of objects to include in output list
        :param schematics.datastructures.Context context: A Context object that encapsulates
            configuration options and app_data. Primarily used internally by schematics.
        :param overrides: Overrides for the model. This should be a nested dictionary of field names in
            which the particular fields to override are mapped to the values with which they are to be
            overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}}``
        :type overrides: dict[str, Any]
        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: sf_models.SfVersion
        :param bool include_optional: Whether to mock non-required fields (randomly mocks optional
            fields by default).
        :param force_mock: Force the mocking of a particular field. This should be a nested dictionary
            of field names in which the particular fields to be mocked are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_mock': True}}}``
        :type force_mock: dict[str, dict|bool]
        :return: List of data model instances
        :rtype: list[sf_models.Base]
        """
        return [cls.get_mock_object(context=context,
                                    overrides=overrides,
                                    version=version,
                                    include_optional=include_optional,
                                    force_mock=force_mock)
                for _ in range(length)]

    @classmethod
    def get_mock_data(cls, overrides=None, version=None, include_optional=None, force_mock=None,
                      exclude_fields=None):
        """Get mock data.

        :param overrides: Overrides for the model. This should be a nested dictionary of field names in
            which the particular fields to override are mapped to the values with which they are to be
            overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}}``
        :type overrides: dict[str, Any]
        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: sf_models.SfVersion
        :param bool include_optional: Whether to mock non-required fields (randomly mocks optional
            fields by default).
        :param force_mock: Force the mocking of a particular field. This should be a nested dictionary
            of field names in which the particular fields to be mocked are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_mock': True}}}``
        :type force_mock: dict[str, dict|bool]
        :param exclude_fields: Exclude the specified fields. This should be a nested dictionary of
            field names in which the particular fields to be excluded are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_exclude': True}}}``
        :type exclude_fields: dict[str, dict|bool]
        :return: Dictionary of mocked data
        :rtype: dict
        """
        mock_object = cls.get_mock_object(overrides=overrides,
                                          version=version,
                                          include_optional=include_optional,
                                          force_mock=force_mock)

        return mock_object.serialize(version=version, exclude_fields=exclude_fields)

    @classmethod
    def get_mock_data_list(cls, length, overrides=None, version=None, include_optional=None,
                           force_mock=None, exclude_fields=None):
        """Get a list of mock data.

        :param int length: Number of serialized objects to include in output
        :param overrides: Overrides for the model. This should be a nested dictionary of field names in
            which the particular fields to override are mapped to the values with which they are to be
            overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}}``
        :type overrides: dict[str, Any]
        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: sf_models.SfVersion
        :param bool include_optional: Whether to mock non-required fields (randomly mocks optional
            fields by default).
        :param force_mock: Force the mocking of a particular field. This should be a nested dictionary
            of field names in which the particular fields to be mocked are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_mock': True}}}``
        :type force_mock: dict[str, dict|bool]
        :param exclude_fields: Exclude the specified fields. This should be a nested dictionary of
            field names in which the particular fields to be excluded are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_exclude': True}}}``
        :type exclude_fields: dict[str, dict|bool]
        :return: List of dictionaries of mocked data
        :rtype: list[dict]
        """
        return [cls.get_mock_data(overrides=overrides,
                                  version=version,
                                  include_optional=include_optional,
                                  force_mock=force_mock,
                                  exclude_fields=exclude_fields)
                for _ in range(length)]

    @classmethod
    def create(cls, **kwargs):
        """Implement this method on objects that users will need to instantiate directly.

        This typically applies to objects that have create calls, such as Account, Volume, etc. For
        example, to use the CreateVolumes component, the user must pass in a list of partial volume
        data models which define the profile of the volumes to create. These partial models will not
        populate fields such as the volume_id which are not known until after the volume is created.

        :return: Data model with fields populated for a create API request.
        :rtype: sf_models.Base
        """
        return cls(kwargs)

    def serialize(self, version=None, exclude_fields=None):  # pylint: disable=too-complex
        """Converts data model to a dictionary representation.

        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: sf_models.SfVersion
        :param exclude_fields: Exclude the specified fields. This should be a nested dictionary of
            field names in which the particular fields to be excluded are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_exclude': True}}}``
        :type exclude_fields: dict[str, dict|bool]
        :return: The serialized data
        :rtype: dict[str, Any]
        """
        exclude_fields = exclude_fields or {}
        # As of schematics-2.0.0, serialize() was changed to be different from to_primitive(). It now
        # calls validate() and does not serialize invalid data. to_primitive() is called instead of
        # serialize() to preserve the old behavior from schematics-2.0.0a1.
        data = super(Base, self).to_primitive()
        valid_serialized_names = set()
        for name, field in self._get_field_list():
            # Get the field's serialized name if it has one
            serialized_name = name
            if field.serialized_name is not None:
                serialized_name = field.serialized_name

            # Exclude fields that are out-of-version or which have been explicitly specified
            field_valid = True
            if version is not None:
                field_valid = self._field_exists_in_version(field, version)
            if exclude_fields.get(name) is True:
                field_valid = False

            if not field_valid and serialized_name in data:
                del data[serialized_name]
            else:
                # Raise an error if multiple valid fields serialize to the same thing
                if serialized_name not in valid_serialized_names:
                    valid_serialized_names.add(serialized_name)
                else:
                    raise exceptions.ConversionError(
                        'Failed to serialize field[{}]. Serialization[{}] already in use by another '
                        'valid field.'.format(name, serialized_name)
                    )

                # Serialize nested models
                if isinstance(field, types.ModelType):
                    nested_model = self.get(name)
                    if nested_model:
                        data[serialized_name] = nested_model.serialize(
                            version=version,
                            exclude_fields=exclude_fields.get(name)
                        )

        return data

    def validate(self, version=None, **kwargs):  # pylint: disable=too-complex
        """Validates fields in data model.

        If a version is not provided, a validation error will occur if any required field is missing. If
        a version is provided, a validation error will occur if a field that is required for that version
        is missing or a field that does not exist in that version is set.

        :param version: SolidFire API version to validate against.
        :type version: sf_models.SfVersion
        :raises errors.DataValidationError: If a required field is missing, a field doesn't
            exist for the given version, or if a field does not meet the specified schema.
        """
        errs = {}
        for name, field in self._get_field_list():
            if isinstance(field, types.ModelType):
                # Nested models must be explicitly validated
                nested_model = self.get(name)
                if nested_model:
                    nested_model.validate(version, **kwargs)
            value = self.get(name)

            if field.default is not models.Undefined and value == field.default:
                pass
            elif version is not None:
                if value is None and field.required and self._field_exists_in_version(field, version):
                    errs[name] = exceptions.ConversionError('This field is required in version {}'
                                                            .format(version))
                if value is not None and name and not self._field_exists_in_version(field, version):
                    errs[name] = exceptions.ConversionError('This field does not exist in version {}'
                                                            .format(version))
            else:
                # If version is None, don't validate fields tied to specific versions
                exists_in_all_versions = self._field_exists_in_version(field,
                                                                       sf_version.SfVersion("4.0"))
                if value is None and field.required and exists_in_all_versions:
                    errs[name] = exceptions.ConversionError('This field is required')
        if errs:
            raise errors.DataValidationError(errs)
        try:
            super(Base, self).validate(partial=True, **kwargs)
        except exceptions.DataError as err:
            raise errors.DataValidationError(dict(err.errors))

    def __deepcopy__(self, memodict=None):
        return self.__class__(self.serialize())


class ResourceInterface(object):
    """All resources should have a static_id.

    Data models that use the resources_dm field should also have a static_id that maps to the same ID
    provided by the resource model.
    """

    @property
    @abc.abstractmethod
    def static_id(self):
        """Returns a unique identifier for the resource.

        :rtype: str
        """

    @static_id.setter
    @abc.abstractmethod
    def static_id(self, id_):
        """Set the unique_id, this should only be used for testing as resources should be immutable.

        :param str id_: new id
        """
