"""Data model for clients."""

from sf_models import (
    _base,
    _resource_network,
    at2_constants,
    at2_enums,
    types,
)


class ResourceClient(_base.Base):
    """Client Resource Model class as per AT2."""
    username = types.StringType(required=True)
    password = types.StringType(required=True)
    mip = types.IPAddressType(required=True, serialized_name=at2_constants.ONE_GIG_IP)
    sip = types.IPAddressType(required=True, serialized_name=at2_constants.TEN_GIG_IP)
    idrac_ip = types.At2OobOrEmptyStringType(required=False, serialized_name=at2_constants.IDRAC_IP)
    resource_id = types.IntIdType(required=True, serialized_name=at2_constants.RESOURCE_ID)
    location_id = types.IntType(required=True, serialized_name=at2_constants.LOCATION_ID)
    os = types.EnumType(at2_enums.ClientOS, required=True, serialized_name=at2_constants.OS)
    type = types.EnumType(at2_enums.ClientType, required=True)
    fibre_channel = types.EnumType(at2_enums.FibreChannel, required=True,
                                   serialized_name=at2_constants.FIBRE_CHANNEL)
    client_name = types.StringType(required=True, min_length=1,
                                   serialized_name=at2_constants.CLIENT_NAME)
    network_profile = types.ModelType(_resource_network.NetworkProfile,
                                      serialized_name=at2_constants.NETWORK_PROFILE, required=True)

    # At2 Stuff
    _location_name = types.StringType(serialized_name=at2_constants.LOCATION_NAME)
    _hardware = types.StringType(serialized_name=at2_constants.HARDWARE)
    _resource_type_id = types.IntType(serialized_name=at2_constants.RESOURCE_TYPE_ID)
    resource_type_name = types.StringType(required=True,
                                          serialized_name=at2_constants.RESOURCE_TYPE_NAME)
    local_attributes = {}

    @classmethod
    def get_mock_object(cls, context=None, overrides=None, version=None, include_optional=None,
                        force_mock=None):
        """Get a mock ResourceClient object.

        Override because mocking type and iDracIP are related. If type is virtual, iDracIP is empty
        string. If type is physical iDracIP is an IP or hostname.

        :param schematics.datastructures.Context context: A Context object that encapsulates
            configuration options and app_data. Primarily used internally by schematics.
        :param overrides: Overrides for the model. This should be a nested dictionary of field names in
            which the particular fields to override are mapped to the values with which they are to be
            overridden. Keys must be equal to the field names, not the serialized names
            (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_override': override_value}}``
        :type overrides: dict[str, Any]
        :param version: SolidFire API version to mock.
            Fields that do not exist in given version will be set to None.
        :type version: str|sf_models.SfVersion
        :param bool include_optional: Whether to mock non-required fields. None randomly mocks optional
            fields.
        :param force_mock: Force the mocking of a particular field. This should be a nested dictionary
            of field names in which the particular fields to be mocked are mapped to True. Keys must be
            equal to the field names, not the serialized names (e.g. use drive_id instead of driveID).
            Example: ``{'parent_field': {'child_field': {... {'field_to_mock': True}}}``
        :type force_mock: dict[str, dict|bool]
        :return: The data model instance
        :rtype: sf_models.ResourceClient
        """
        obj = super(ResourceClient, cls).get_mock_object(context=context,
                                                         overrides=overrides,
                                                         version=version,
                                                         include_optional=include_optional,
                                                         force_mock=force_mock)
        data = obj.serialize()
        if data['type'] == 'virtual':
            data[at2_constants.IDRAC_IP] = ''
            obj = cls(data)
        return obj

    def __repr__(self):
        return '{}<name={}, mip={}, sip={}, idrac_ip={}>'.format(
            self.__class__.__name__, self.client_name, self.mip, self.sip, self.idrac_ip)

    @property
    def name(self):
        """Returns client name.

        :rtype: str
        """
        return self.client_name

    @property
    def at2_type(self):
        """Returns client type.

        :rtype: str
        """
        return self.type.value
