"""External key server data model."""
from sf_models import (
    _base,
    enums,
    at2_constants,
    types,
    errors,
)


class ExternalKeyServer(_base.Base):
    """External key server model.

    The only required fields are hostname, username, and password. Port will default to
    the standard KMIP port 5696. The KMIP and SSL versions will default to the latest and most
    secure versions.
    """
    hostname = types.FqdnType(required=True)
    port = types.IntType(required=False, default=5696)
    username = types.StringType(required=True)
    password = types.StringType(required=True)
    vendor = types.EnumType(enums.ExternalKeyServerVendor, required=True)
    kmip_version = types.EnumType(enums.KmipVersion, required=False,
                                  default=enums.KmipVersion.proto_1_4,
                                  serialized_name=at2_constants.KMIP_VERSION)
    ssl_version = types.EnumType(enums.SslVersion, required=False,
                                 default=enums.SslVersion.ssl_2_3,
                                 serialized_name=at2_constants.SSL_VERSION)

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'hostname', self.hostname,
            'port', self.port,
            'vendor', self.vendor)


class VormetricExternalKeyServer(ExternalKeyServer):
    """Vormetric has some additional credentials that need to be captured in the EKS model."""
    domain_name = types.StringType(required=True)
    domain_username = types.StringType(required=True)
    domain_password = types.StringType(required=True)


class ExternalKeyServerCluster(_base.Base):
    """Cluster of external key servers."""
    nodes = types.ListType(types.ModelType(ExternalKeyServer), required=True, min_size=2)

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'nodes', self.nodes)

    def validate(self, version=None, **kwargs):
        """Validates fields in data model.

        Override the base validate method to also check that the vendor type of all nodes
        in the cluster are the same.

        :param version: SolidFire API version to validate against.
        :type version: sf_models.SfVersion
        :raises errors.DataValidationError: If a required field is missing, a field doesn't
            exist for the given version, if a field does not meet the specified schema, or
            if all EKS vendors do not have the same vendor type.
        """
        vendors = [node.vendor for node in self.nodes]  # pylint: disable=not-an-iterable
        if len(set(vendors)) != 1:
            err_msg = 'All nodes in EKS cluster must have the same vendor type'
            raise errors.DataValidationError({'ValueError': err_msg})

        super(ExternalKeyServerCluster, self).validate(version=version, **kwargs)

    @property
    def primary(self):
        """Return the node that we designate as primary.

        EKS clusters automatically handle replication across all nodes in the cluster.
        It doesn't matter which node we call primary.

        :rtype: sf_models.ExternalKeyServer
        """
        return self.nodes[0]  # pylint: disable=unsubscriptable-object

    @property
    def secondaries(self):
        """Return list of nodes excluding the primary.

        :rtype: list[sf_models.ExternalKeyServer]
        """
        return self.nodes[1:]  # pylint: disable=unsubscriptable-object
