"""Class to encapsulate valid network interface names."""

from sf_models import enums


class NetworkInterfaceName(object):
    """Class to represent possible network interface names.

    Common names such as eth0, Bond1G, ext are still in the NetworkInterface enum. This class
    represents those and adds the bonded interface names with virtual network tags. These tags can be
    from 1 - 4094 for both the 1G and 10G interface.

    On the Cluster Master, bonded network interfaces are named Bond1G:M and Bond10G:S, so if the network
    interface has a VLAN tag and is on the CM node you will get Bond10G.3000:M.

    :param value: The name of the network interface, such as eth0, Bond1G, or Bond1G.3000.
    :type value: str|sf_models.enums.NetworkInterface
    """
    def __init__(self, value):
        # Value is enum
        if isinstance(value, enums.NetworkInterface):
            self.value = value.value

        # Value is string
        elif self._could_be_enum(value):
            self.value = value
        elif self._is_tagged_vlan_interface(value):
            self.value = value
        elif self._is_docker_interface(value):
            self.value = value
        else:
            raise ValueError('Given name, {}, is not a NetworkInterface or VLAN tagged bonded '
                             'interface'.format(value))

    def __repr__(self):
        return '<NetworkInterfaceName.value: %r>' % self.value

    def __eq__(self, other):
        """Compares the NetworkInterfaceName.value.

        :param sf_models.NetworkInterfaceName other: Object to compare against
        :rtype: bool
        """
        if isinstance(other, NetworkInterfaceName) and other.value == self.value:
            return True
        else:
            return False

    @property
    def is_vlan(self):
        return self._is_tagged_vlan_interface(self.value)

    @staticmethod
    def _could_be_enum(value):
        """Helper method that tries to instantiate enum and catches ValueError.

        :param str value: Value to try as enum
        :rtype: bool
        """
        try:
            enums.NetworkInterface(value)
        except ValueError:
            return False
        else:
            return True

    @staticmethod
    def _is_tagged_vlan_interface(name):
        """Returns True if name is a valid VLAN tagged network interface name.

        :param str name: a potential VLAN tagged network interface
        :rtype: bool
        """
        if len(name.split('.')) != 2:
            return False
        _, tag = name.split('.')

        # check if tag has :M or :S attached
        if len(tag.split(':')) == 2:
            tag, bond = tag.split(':')
            if bond != 'M' and bond != 'S':
                return False
        elif len(tag.split(':')) > 2:
            return False

        # check if tag is an int and in range
        try:
            tag_int = int(tag)
        except ValueError:
            return False
        if tag_int < 1 or tag_int > 4094:
            return False

        return True

    @staticmethod
    def _is_docker_interface(value):
        """Determines if value a valid docker interface.

        :param str value: Value to check
        :rtype: bool
        """
        return value.startswith('veth') or value.startswith('docker')
