"""Node Data Models."""
from sf_models import (
    _base,
    _hypervisor,
    _resource_network,
    at2_constants,
    at2_enums,
    enums,
    types,
)


class ResourceNode(_base.Base, _base.ResourceInterface):
    """AT2 Node Model class (immutable)."""
    username = types.StringType(required=True)
    password = types.StringType(required=True)
    oob_ip = types.At2OobType(serialized_name=at2_constants.IDRAC_IP)
    resource_id = types.IntIdType(serialized_name=at2_constants.RESOURCE_ID)
    sip = types.FqdnType(required=True, serialized_name=at2_constants.TEN_GIG_IP)
    sipV6 = types.IPAddressType()
    sipV4 = types.IPAddressType()
    mip = types.FqdnType(required=True, serialized_name=at2_constants.ONE_GIG_IP)
    mipV6 = types.IPAddressType()
    mipV4 = types.IPAddressType()
    hypervisor_host = types.ModelType(_hypervisor.HypervisorHost,
                                      serialized_name=at2_constants.HYPERVISOR_HOST)
    virtual_hardware_profile = types.ModelType(_hypervisor.VirtualHardwareProfile,
                                               serialized_name=at2_constants.VIRTUAL_HARDWARE_PROFILE)
    location_id = types.IntType(serialized_name=at2_constants.LOCATION_ID)
    fibre_channel = types.EnumType(at2_enums.FibreChannel, required=True,
                                   serialized_name=at2_constants.FIBRE_CHANNEL)
    model = types.NodeTypeType(required=True)
    node_name = types.StringType(required=True, serialized_name=at2_constants.NODE_NAME)
    network_profile = types.ModelType(_resource_network.NetworkProfile,
                                      serialized_name=at2_constants.NETWORK_PROFILE)

    # At2 Stuff
    resource_type_name = types.StringType(serialized_name=at2_constants.RESOURCE_TYPE_NAME)
    _lastOOBReset = types.StringType(serialized_name=at2_constants.LAST_OOB_RESET)

    def __repr__(self):
        return '{}<name={}, mip={}, sip={}, oob_ip={}>'.format(
            self.__class__.__name__, self.node_name, self.mip, self.sip, self.oob_ip)

    @property
    def static_id(self):
        """Returns a unique identifier for the node.

        :rtype: str
        """
        return self.mip

    @static_id.setter
    def static_id(self, id_):
        """Set the static_id.

        :param str id_: new id
        """
        self.mip = id_

    def is_witness(self):
        """Returns whether witness node or not.

        :rtype: bool
        """
        return self.model and self.model.is_witness  # pylint: disable=no-member


class ResourceComputeNode(_base.Base):
    """ResourceComputeNode data model."""
    resource_pool_id = types.IntIdType(serialized_name=at2_constants.RESOURCE_POOL_ID)
    node_name = types.StringType(required=True, serialized_name=at2_constants.NODE_NAME)
    network_profile = types.ModelType(_resource_network.NetworkProfile,
                                      serialized_name=at2_constants.NETWORK_PROFILE)
    resource_id = types.IntIdType(required=True, serialized_name=at2_constants.RESOURCE_ID)
    data_type_name = types.StringType(required=True, serialized_name=at2_constants.DATA_TYPE_NAME)
    resource_type_name = types.StringType(serialized_name=at2_constants.RESOURCE_TYPE_NAME)
    location_name = types.StringType(required=True, serialized_name=at2_constants.LOCATION_NAME)
    resource_type_id = types.IntIdType(required=True, serialized_name=at2_constants.RESOURCE_TYPE_ID)
    location_id = types.IntType(required=True, serialized_name=at2_constants.LOCATION_ID)
    serial_no = types.StringType(required=True, serialized_name=at2_constants.SERIAL_NO)
    model = types.EnumType(enums.ComputeNodeModel, required=True)
    oob_ip = types.At2OobType(required=True, serialized_name=at2_constants.OOB_IP)
    mip = types.IPAddressType(required=False)
    username = types.StringType(required=False)
    password = types.StringType(required=False)
    os = types.EnumType(at2_enums.ClientOS, required=False)
