"""Upgrade data model."""

from sf_models import (
    _base,
    enums,
    types,
)


class UpgradeNode(_base.Base):
    """Contains information about the node details in status."""
    node_id = types.IntIdType(serialized_name='nodeID')
    num_attempt = types.IntType(serialized_name='numAttempt')
    message = types.StringType(serialized_name='message')
    step = types.EnumType(enums.NodeUpgradeStep, serialized_name='step')


class StorageServiceStatusBase(_base.Base):
    """Contains information about the status object from storage service endpoints."""

    message = types.StringType(serialized_name='message')
    percent_completed = types.IntType(serialized_name='percent')
    timestamp = types.UTCDateTimeType(serialized_name='timestamp')


class Healthcheck(_base.Base):
    """Contains information about healthcheck as returned by solidfire-healthtools in JSON blob."""
    check_id = types.IntType(required=False, serialized_name='checkID')
    display_name = types.StringType(required=False, serialized_name='displayName')
    name = types.StringType(required=True, serialized_name='name')
    passed = types.BooleanType(required=True, serialized_name='passed')
    kb = types.StringType(required=False, serialized_name='KB')
    description = types.StringType(required=True, serialized_name='description')
    remedy = types.StringType(required=False, serialized_name='remedy', default=None,
                              serialize_when_none=False)
    severity = types.EnumType(enum_class=enums.HealthcheckSeverity,
                              required=False, serialized_name='severity', default=None,
                              serialize_when_none=False)
    data = types.BaseType(required=True, default=None, serialized_name='data')
    node_id = types.IntType(required=False, serialized_name='nodeID', default=None,
                            serialize_when_none=False)

    def __repr__(self):
        return '<{} name:{} passed:{}>'.format(
            self.__class__.__name__, self.name, self.passed)


class StorageHealthcheckStatusModel(StorageServiceStatusBase):
    """Nested status for Storage Health Check REST endpoint."""

    checks = types.ListType(types.ModelType(Healthcheck), default=list)


class StorageHealthcheckResultModel(_base.Base):
    """Total healthcheck result that returns both nodes and cluster checks.

    Corresponds to the healthchecks object in the PreUpgradeHealthCheck API response.
    """
    links = types.DictType(types.StringType(), serialized_name='_links')
    config = types.BaseType(serialized_name='config')
    date_completed = types.UTCDateTimeType(serialized_name='dateCompleted')
    date_created = types.UTCDateTimeType(serialized_name='dateCreated')
    healthcheck_id = types.UUIDType(serialized_name='healthCheckId')
    state = types.EnumType(enum_class=enums.UpgradeState)
    status = types.ModelType(StorageHealthcheckStatusModel)
    storage_id = types.UUIDType(required=False, serialized_name='storageId')
    task_id = types.UUIDType(required=False, serialized_name='taskId')


class UpgradeStatus(StorageServiceStatusBase):
    """Contains information about the status object in sfinstall."""
    step = types.EnumType(enums.UpgradeStep, serialized_name='step')
    available_actions = types.ListType(types.StringType, serialized_name='availableActions',
                                       default=list)
    node_details = types.ListType(types.ModelType(UpgradeNode), default=list,
                                  serialized_name='nodeDetails')
    failed_health_checks = types.ListType(types.ModelType(Healthcheck),
                                          default=list,
                                          serialized_name='failedHealthChecks')


class UpgradeResource(_base.Base):
    """Contains all information required to perform upgrades for management node."""
    storage_id = types.UUIDType(serialized_name='storageId')
    package_id = types.UUIDType(required=True, serialized_name='packageId')
    config = types.BaseType(serialized_name='config')
    state = types.EnumType(enums.UpgradeState, serialized_name='state')
    status = types.ModelType(UpgradeStatus, serialized_name='status')
    task_id = types.UUIDType(serialized_name='taskId')
    upgrade_id = types.UUIDType(serialized_name='upgradeId')
    date_created = types.UTCDateTimeType(serialized_name='dateCreated')
    date_completed = types.UTCDateTimeType(serialized_name='dateCompleted')

    def __repr__(self):
        return '<{} upgrade ID:{} status:{}>'.format(
            self.__class__.__name__, self.upgrade_id, self.status)


class UpgradePackage(_base.Base):
    """Contains information about variables needed for the package manager."""
    url = types.URLType(required=False)
    id = types.UUIDType(required=False)
    version = types.StringType(required=False)

    def __repr__(self):
        return '<{} package: {}>'.format(
            self.__class__.__name__, self.id)


class UpgradeLogs(_base.Base):
    """Contains information about the upgrade logs."""
    mnode_storage = types.DictType(types.ListType(types.StringType()), serialized_name='mnode_storage')


class UpgradeConfig(_base.Base):
    """Contains information about the sfinstall options for upgrade."""

    # Everything here MUST have a default value.  We will only override the delta.

    # Cluster Upgrader Settings
    reboot_nodes_before_upgrading = types.BooleanType(default=False,
                                                      serialized_name='rebootNodesBeforeUpgrading')
    reboot_nodes_on_resume = types.BooleanType(default=False,
                                               serialized_name='rebootNodesOnResume')
    drum = types.BooleanType(default=True, serialized_name='drum')
    set_constants = types.DictType(types.UnionType([types.IntType, types.BooleanType]),
                                   default=None, serialized_name='setConstants')

    # Storage Node Upgrader
    rtfi_flags = types.StringType(default=None, serialized_name='rtfiFlags')
    max_install_attempts = types.IntType(default=3,
                                         serialized_name='maxInstallAttempts',
                                         min_value=1)
    max_cm_move_wait_time = types.IntType(default=300,
                                          serialized_name='maxClusterMasterMoveWaitTime')
    fc_node_install_delay = types.IntType(default=5 * 60,
                                          serialized_name='fibreChannelNodeInstallDelay',
                                          min_value=0)

    # Slice Manager Settings
    move_primaries_away_from_node = types.BooleanType(default=True,
                                                      serialized_name='movePrimariesAwayFromNode')
    move_primaries_wait_time = types.IntType(default=15,
                                             serialized_name='movePrimariesWaitTime',
                                             min_value=0)
    move_primaries_gradual_group_count = types.IntType(default=4,
                                                       serialized_name='movePrimariesGradualGroupCount')
    move_primaries_retry_count = types.IntType(default=6,
                                               serialized_name='movePrimariesRetryCount')

    # ClusterFaultMonitor Settings
    ignore_existing_drive_faults = types.BooleanType(default=False,
                                                     serialized_name='ignoreExistingDriveFaults')
    ignored_cluster_faults = types.ListType(types.StringType, default=list,
                                            serialized_name='ignoredClusterFaults')
    min_wait_for_faults = types.IntType(default=300, serialized_name='minWaitForFaults', min_value=0)

    # ScacheMonitor Settings
    s_cache_wait_time = types.IntType(default=60,
                                      serialized_name='scacheWaitTime',
                                      min_value=0)
    s_cache_max_wait_time = types.IntType(default=43200,
                                          serialized_name='scacheMaxWaitTime')
    s_cache_threshold = types.IntType(default=5 * 1000 * 1000 * 1000,
                                      serialized_name='scacheThreshold')

    # VolumeStatusMonitor Settings
    monitor_bulk_volume = types.BooleanType(default=False,
                                            serialized_name='monitorBulkVolume')
    volume_status_check_period = types.IntType(default=30,
                                               serialized_name='volumeStatusCheckPeriod')
    volume_status_check_exit = types.BooleanType(default=True,
                                                 serialized_name='volumeStatusCheckExit')
    max_volume_health_check_iterations = types.IntType(default=5,
                                                       serialized_name='maxVolumeHealthCheckIterations')

    # Cluster and Node Client Settings
    api_retry_interval = types.IntType(
        default=30, serialized_name='apiRetryInterval', min_value=0)
    api_max_retry_count = types.IntType(
        default=20, serialized_name='apiMaxRetryCount', min_value=1)
    api_connect_timeout = types.IntType(
        default=30, serialized_name='apiConnectTimeout', min_value=1)
    api_read_timeout = types.IntType(
        default=300, serialized_name='apiReadTimeout', min_value=1)

    node_id_upgrade_order = types.ListType(types.IntType, default=list,
                                           serialized_name='nodeIdUpgradeOrder')

    metadata = types.StringType(default=None, serialized_name='metadata')

    # HC Validation Config
    validation_skip_list = types.ListType(types.StringType, default=list)
    var_log_free_percent = types.IntType(default=75, serialized_name='varLogFreePercent')
    min_root_fs_free_space = types.IntType(default=12 * (1000 * 1000 * 1000),
                                           serialized_name='minRootFsFreeSpace')
    async_result_threshold = types.IntType(default=25000, serialized_name='asyncResultThreshold')
    force_setting_bond_down_delay = types.BooleanType(default=False,
                                                      serialized_name='forceSettingBondDownDelay')

    # Package Validator Config
    check_valid_upgrade_version = types.BooleanType(default=True,
                                                    serialized_name='checkValidUpgradeVersion')
    check_valid_upgrade_hardware = types.BooleanType(serialized_name='checkValidUpgradeHardware',
                                                     default=True)
