"""Enums for AT2's dumb dictionaries."""

import aenum

from sf_models import at2_constants


class BondMode(aenum.Enum):
    """Used by bond_mode field in NetworkInterfaceDetails data model."""
    active_passive = 'ActivePassive'
    lacp = 'LACP'
    alb = 'ALB'
    lacp_slow = 'LACP_SLOW'


class ClientOS(aenum.Enum):
    """Used for the os field for a Resource Client."""
    # These are the client OSes in AT2 resources
    windows_2008 = 'Windows Server 2008 R2'
    windows_2012 = 'Windows Server 2012 R2'
    windows_2016 = 'Windows Server 2016'

    ubuntu_12_04 = 'Ubuntu 12.04'
    ubuntu_14_04 = 'Ubuntu 14.04'
    ubuntu_16_04 = 'Ubuntu 16.04'
    ubuntu_18_04 = 'Ubuntu 18.04'

    solaris = 'Solaris 11'

    esx_5_1 = 'ESX 5.1'
    esx_5_5 = 'ESX 5.5'
    esx_6 = 'ESX 6.0'
    esx_6_5 = 'ESX 6.5'
    esx_6_7 = 'ESX 6.7'

    centos = 'CentOS'  # TODO: LABDESK-2900 has been opened to make sure a version is specified
    centos_6 = 'CentOS 6'
    centos_7 = 'CentOS 7.0'
    centos_7_6 = 'CentOS 7.6'

    @property
    def is_ubuntu(self):
        """Check if the client is ubuntu 12.04 or 14.04.

        :rtype: bool
        """
        return self in (self.ubuntu_12_04, self.ubuntu_14_04, self.ubuntu_16_04)

    @property
    def is_windows(self):
        """Check if the client is windows.

        :rtype: bool
        """
        return self in (self.windows_2008, self.windows_2012, self.windows_2016)

    @property
    def is_esx(self):
        """Check if the client is esx 5.1, 5.5, 6, 6.5 or 6.7.

        :rtype: bool
        """
        return self in (self.esx_5_1, self.esx_5_5, self.esx_6, self.esx_6_5, self.esx_6_7)

    @property
    def is_centos(self):
        """Check if the client is centos version unknown, 6 or 7.0.

        :rtype: bool
        """
        return self in (self.centos, self.centos_6, self.centos_7, self.centos_7_6)

    @property
    def is_solaris(self):
        """Check if the client is solaris.

        :rtype: bool
        """
        return self in (self.solaris, )

    @property
    def is_linux(self):
        """Check if the client is ubuntu or centos.

        :rtype: bool
        """
        return self.is_ubuntu or self.is_centos


class ClientType(aenum.Enum):
    """Used for the type field for a Resource Client."""
    physical = 'physical'
    virtual = 'virtual'


class ControlType(aenum.Enum):
    """Used for the controlType field for a Resource Client."""
    libesx = 'libesx'


class FibreChannel(aenum.Enum):
    """Used by FibreChannel field for Resource Node."""
    none = 'None'
    qlogic = 'QLogic'
    qlogic_8g = 'QLogic 8G'
    qlogic_16g = 'QLogic 16G'
    emulex_8g = 'Emulex 8G'
    emulex_16g = 'Emulex 16G'


class NetworkSpeed(aenum.Enum):
    """Used by network field in IP."""
    ten_gig = '10g'
    one_gig = '1g'


class OntapModel(aenum.Enum):
    """Valid ONTAP Models."""
    vsim = 'vsim'
    fas8020 = 'fas8020'
    fas8040 = 'fas8040'
    fas8060 = 'fas8060'
    fas8080 = 'fas8080'
    fas8080ex = 'fas8080ex'
    affa200 = 'affa200'
    affa300 = 'affa300'
    affa700 = 'affa700'
    affa700s = 'affa700s'
    aff8020 = 'aff8020'
    aff8040 = 'aff8040'
    aff8060 = 'aff8060'
    aff8080 = 'aff8080'
    aff8080ex = 'aff8080ex'


class OntapPorts(aenum.Enum):
    """Valid ONTAP Cluster Ports."""
    e0a = 'e0a'
    e0b = 'e0b'
    e0c = 'e0c'
    e0d = 'e0d'
    e0e = 'e0e'
    e0f = 'e0f'
    e0g = 'e0g'
    e0h = 'e0h'
    e0i = 'e0i'
    e0j = 'e0j'
    e0k = 'e0k'
    e0l = 'e0l'
    e0M = 'e0M'
    _0e = '0e'
    _0f = '0f'
    _0g = '0g'
    _0h = '0h'


class TypeEnum(aenum.Enum):
    """Used for HypervisorHost return values."""
    ESX = at2_constants.ESX
    KVM = at2_constants.KVM
    PHYSICAL = at2_constants.PHYSICAL
