"""Class to encapsulate data size conversions."""

from __future__ import division

import aenum


class DataUnit(aenum.Enum):
    """Used when talking about data size.

    {bytes, gigabytes, etc}. Values are tuples of (multiplier, shorthand_unit), where multiplier is what
    you multiply by to convert to bytes.
    """
    bytes = (1., 'B')
    blocks = (4096, 'blocks')
    kilobytes = (1000. ** 1, 'kB')
    megabytes = (1000. ** 2, 'MB')
    gigabytes = (1000. ** 3, 'GB')
    terabytes = (1000. ** 4, 'TB')
    petabytes = (1000. ** 5, 'PB')
    kibibytes = (1024. ** 1, 'kiB')
    mebibytes = (1024. ** 2, 'MiB')
    gibibytes = (1024. ** 3, 'GiB')
    tebibytes = (1024. ** 4, 'TiB')
    pebibytes = (1024. ** 5, 'PiB')


class DataSize(object):
    """Class to represent bytes and multiples of bytes with units. Supports all basic arithmetic.

    :param int|float size: The size of the data in units, default unit is bytes.
    :param sf_models.data_size.DataUnit|tuple unit: The unit of the size parameter.
    """

    def __init__(self, size, unit=DataUnit.bytes):
        self._unit = unit
        if unit is DataUnit.bytes and size % 1 != 0:
            raise ValueError('You cannot use a fraction when the unit is bytes. Given: {}'.format(size))
        self.raw_size = size
        self._bytes = size * self._unit.value[0]

    @property
    def pre_carbon_size(self):
        """Size a volume should be if Create/Clone/Modify volume given this size as the totalSize param.

        Basically round up to the nearest block.

        :rtype: sf_models.data_size.DataSize
        """
        block_size = int(DataUnit.blocks.value[0])
        return DataSize(self.bytes // block_size + bool(self.bytes % block_size), DataUnit.blocks)

    @property
    def carbon_size(self):
        """Size a volume should be if Create/Clone/ModifyVolume given this size as the totalSize param.

        Basically round up to the nearest multiple of 1MiB.

        :rtype: sf_models.data_size.DataSize
        """
        mib = int(DataUnit.mebibytes.value[0])
        return DataSize(self.bytes // mib + bool(self.bytes % mib), DataUnit.mebibytes)

    @property
    def unit(self):
        """Unit associated with this size object.

        :rtype: sf_models.data_size.DataUnit
        """
        return self._unit

    @property
    def unit_size(self):
        """Bytes per specified unit for this size object.

        :rtype: int
        """
        return self._unit.value[0]

    @property
    def unit_name(self):
        """Name of unit for this size object.

        :rtype: str
        """
        return self._unit.value[1]

    @property
    def bytes(self):
        """Number of bytes in this object (same as int(self)).

        :rtype: int
        """
        return int(self._bytes / DataUnit.bytes.value[0])

    @property
    def blocks(self):
        """Number of blocks in this object (must be an int).

        :rtype: int
        """
        return int(self._bytes / DataUnit.blocks.value[0])

    @property
    def kilobytes(self):
        """Number of decimal kilobytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.kilobytes.value[0])

    @property
    def megabytes(self):
        """Number of decimal megabytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.megabytes.value[0])

    @property
    def gigabytes(self):
        """Number of decimal gigabytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.gigabytes.value[0])

    @property
    def terabytes(self):
        """Number of decimal terabytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.terabytes.value[0])

    @property
    def petabytes(self):
        """Number of decimal petabytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.petabytes.value[0])

    @property
    def kibibytes(self):
        """Number of binary kilobytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.kibibytes.value[0])

    @property
    def mebibytes(self):
        """Number of binary mebibytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.mebibytes.value[0])

    @property
    def gibibytes(self):
        """Number of binary gibibytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.gibibytes.value[0])

    @property
    def tebibytes(self):
        """Number of binary tebibytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.tebibytes.value[0])

    @property
    def pebibytes(self):
        """Number of binary pebibytes in this object.

        :rtype: float
        """
        return float(self._bytes / DataUnit.pebibytes.value[0])

    def __repr__(self):
        """This object as a string.

        DataSize(repr(x)) is the same as x if x is a DataSize object.
        """
        return '{} {}'.format(float(self._bytes) / self.unit_size, self.unit_name)

    def __int__(self):
        """Number of bytes in this object (must be an int)."""
        return int(self._bytes)

    def __float__(self):
        """Number of bytes in this object as a float."""
        return float(self._bytes)

    def __eq__(self, x):
        """Equality is taken at the byte level."""
        try:
            return int(self._bytes) == int(x)
        except TypeError:
            return False

    def __ne__(self, x):
        """Inequality is taken at the byte level."""
        return not self.__eq__(x)

    def __lt__(self, x):
        """Compare two DataSize objects."""
        return int(self._bytes) < x

    def __le__(self, x):
        """Compare two DataSize objects."""
        return int(self.bytes) <= x

    def __gt__(self, x):
        """Compare two DataSize objects."""
        return int(self._bytes) > x

    def __ge__(self, x):
        """Compare two DataSize objects."""
        return int(self._bytes) >= x

    def __add__(self, x):
        """Add two DataSize objects."""
        if isinstance(x, DataSize):
            if self.unit_size == x.unit_size:
                return DataSize(int(self.raw_size + x.raw_size), self._unit)
            elif self.unit_size % x.unit_size == 0:
                # noinspection PyTypeChecker
                return DataSize(int(float(self) / x.unit_size + x.raw_size), x.unit)
            elif x.unit_size % self.unit_size == 0:
                # noinspection PyTypeChecker
                return DataSize(int(self.raw_size + float(x) / self.unit_size), self._unit)
        return DataSize(int(self._bytes + int(x)))

    def __radd__(self, x):
        """Add two DataSize objects."""
        return self.__add__(x)

    def __sub__(self, x):
        """Subtract two DataSize objects."""
        return self.__add__(-x)

    def __rsub__(self, x):
        """Subtract two DataSize objects."""
        return (-self._bytes).__add__(x)

    def __neg__(self):
        """Additive inverse of a size."""
        return DataSize(int(-self.raw_size), self._unit)

    def __mul__(self, x):
        """Multiply a DataSize object by a scalar."""
        return DataSize(int(self.raw_size * float(x)), self._unit)

    def __rmul__(self, x):
        """Multiply a DataSize object by a scalar."""
        return self.__mul__(x)

    def __div__(self, x):
        """Divide a DataSize object by a either a DataSize or scalar.

        Note -

            Division of a DataSize with a DataSize will return a float.
            Division of a DataSize with a scalar returns another DataSize, which
            will be rounded down to the nearest integer value of that unit size.
              (e.g. DataSize(5, DataUnit.kibibytes) / 2 == DataSize(2, DataUnit.kibibytes)
                    DataSize(5, DataUnit.kibibytes) / 2.5 == DataSize(2, DataUnit.kibibytes)

        # We import future division in this file, so expect Python 3 behavior.
        assert five_kib / 2 == two_half_kib
        assert five_kib / 2.0 == two_half_kib

        # Test legacy Python 2 behavior, which always rounds down (even with floats)
        assert five_kib.__div__(2) == two_kib
        assert five_kib.__div__(2.5) == two_kib


        :param DataSize|int|float x: divisor
        :rtype: DataSize|float
        :return: Value of the division operation
        """
        if isinstance(x, DataSize):
            return self.__truediv__(x)
        else:
            return self.__floordiv__(x)

    def __floordiv__(self, x):
        """Divide a DataSize object by a scalar using floor division.

        Note -

            Division of a DataSize with a DataSize returns an int.
            Division of a DataSize with a scalar returns another DataSize.

        :param DataSize|int|float x: divisor
        :rtype: DataSize|int
        :return: Value of the floor division operation
        """
        if isinstance(x, DataSize):
            # noinspection PyTypeChecker
            return int(self) // int(x)

        return DataSize(float(self.raw_size) // x, self._unit)

    def __truediv__(self, x):
        """Divide a DataSize object by a scalar using Python 3 true division.

        :param DataSize|int|float x: divisor
        :rtype: DataSize|float
        :return: Value of the division operation
        """
        if isinstance(x, DataSize):
            # noinspection PyTypeChecker
            return float(self) / float(x)

        # noinspection PyTypeChecker
        return DataSize(int(float(self) / x), DataUnit.bytes)

    def __mod__(self, x):
        """Calculate the modulo of two DataSize objects.

        Note -

            Modulo of a DataSize with a DataSize returns a DataSize.
            Modulo of a DataSize with a scalar also returns DataSize.

        :param DataSize|int|float x: divisor
        :rtype: DataSize
        :return: Value of the modulo operation
        """
        if isinstance(x, DataSize):
            if self.unit_size == x.unit_size:
                return DataSize(self.raw_size % x.raw_size, self._unit)
            else:
                # noinspection PyTypeChecker
                return DataSize((float(self) / x.unit_size) % x.raw_size, x.unit)
        else:
            return DataSize(self.raw_size % x, self._unit)
