"""Enums for the sf-models package."""

import aenum


class ApiEnum(aenum.Enum):
    """Provides versioning for enum fields."""

    @classmethod
    def fields_in_version(cls, version):
        """Returns the list of enum fields valid for the specified version.

        By default this method returns every field. Override for version-dependent enums.

        :param sf_models.SfVersion version: SolidFire API version.
        :rtype: set
        """
        # pylint: disable=unused-argument

        # PyTypeChecker doesn't recognize this is an iterable in Python2 (not an issue in Python3).
        # This has something to do with extending the enum class and using a class method.
        # noinspection PyTypeChecker
        return set(cls)

    def exists_in_version(self, version):
        """Returns whether the enum field exists in the specified version.

        :param sf_models.SfVersion version: SolidFire API version.
        :rtype: bool
        """
        return self in self.fields_in_version(version)


class ComputeNodeModel(ApiEnum):
    """Defines the available ComputeNode models used by SolidFire."""
    H500E = 'H500E'
    H700E = 'H700E'
    H300E = 'H300E'
    H410C = 'H410C'
    H410C_14020 = 'H410C14020'
    H410C_15020 = 'H410C15020'
    H410C_17020 = 'H410C17020'
    H410C_25020 = 'H410C25020'
    H410C_27020 = 'H410C27020'
    H410C_28020 = 'H410C28020'
    H410C_57020 = 'H410C57020'
    H410C_58020 = 'H410C58020'
    H410C_35020 = 'H410C35020'
    H410C_37020 = 'H410C37020'
    H610C = 'H610C'
    H610C_45112 = 'H610C45112'
    H615C = 'H615C'
    H615C_64031 = 'H615C64031'
    H615C_65031 = 'H615C65031'
    H615C_75031 = 'H615C75031'
    H615C_77031 = 'H615C77031'
    H615C_78031 = 'H615C78031'
    H615C_7A031 = 'H615C7A031'
    H615C_85031 = 'H615C85031'
    H615C_88031 = 'H615C88031'
    H615C_8A031 = 'H615C8A031'
    H615C_97031 = 'H615C97031'
    H615C_98031 = 'H615C98031'
    H615C_9A031 = 'H615C9A031'
    H615C_A4231 = 'H615CA4231'


class NetworkInterface(ApiEnum):
    """Defines network interfaces used by SolidFire such as Bond10G."""
    bond_1g = 'Bond1G'
    bond_10g = 'Bond10G'
    eth0 = 'eth0'
    eth1 = 'eth1'
    eth2 = 'eth2'
    eth3 = 'eth3'
    lo = 'lo'
    net0 = 'net0'
    net1 = 'net1'

    # Specific to physical cluster-master nodes
    bond_1g_m = 'Bond1G:M'  # MVIP
    bond_10g_s = 'Bond10G:S'  # SVIP

    # Specific to virtual cluster-master nodes
    eth0_m = 'eth0:M'  # MVIP
    eth0_s = 'eth0:S'  # SVIP

    # Specific to fibre-channel nodes
    eth4 = 'eth4'
    eth5 = 'eth5'


class NodeRole(ApiEnum):
    """Used by role field on NodeClusterConfig data model."""
    storage = 'Storage'
    witness = 'Witness'
    fibre_channel = 'FibreChannel'


class ExternalKeyServerVendor(ApiEnum):
    """Used by vendor field in EKS data model."""
    cryptsoft = 'Cryptsoft'
    gemalto_safenet = 'GemaltoSafenet'
    gemalto_safenet_150v = 'GemaltoSafenet150v'
    safenet_at_350v = 'SafenetAT350v'
    hytrust = 'HyTrust'
    vormetric = 'Vormetric'
    ibm_sklm = 'IbmSklm'


class KmipVersion(ApiEnum):
    """Used by kmip_version field in EKS data model."""
    proto_1_0 = '1.0'
    proto_1_1 = '1.1'
    proto_1_2 = '1.2'
    proto_1_3 = '1.3'
    proto_1_4 = '1.4'


class SslVersion(ApiEnum):
    """Used by ssl_version field in EKS data model."""
    tls = 'PROTOCOL_TLS'
    tls_1 = 'PROTOCOL_TLSv1'
    tls_1_1 = 'PROTOCOL_TLSv1_1'
    tls_1_2 = 'PROTOCOL_TLSv1_2'
    ssl_2_3 = 'PROTOCOL_SSLv23'
    ssl_2 = 'PROTOCOL_SSLv2'
    ssl_3 = 'PROTOCOL_SSLv3'


class TaskActions(ApiEnum):
    """Available actions for Tasks."""
    pause = 'pause'
    resume = 'resume'
    abort = 'abort'
    terminate = 'terminate'


class UpgradeState(ApiEnum):
    """Upgrade State enum."""
    initializing = 'initializing'
    preparing = 'preparing'
    running = 'running'
    pausing = 'pausing'
    paused = 'paused'
    paused_error = 'pausedWithError'
    resuming = 'resuming'
    error = 'error'
    finished = 'finished'
    failed = 'failed'


class UpgradeStep(ApiEnum):
    """Upgrade Step enum."""
    pre_start = 'ClusterPreStart'
    validate = 'ClusterValidate'
    stage = 'ClusterStage'
    upgrade = 'ClusterUpgrade'
    post_upgrade = 'ClusterPostUpgrade'
    finished = 'ClusterFinished'


class NodeUpgradeStep(ApiEnum):
    """Node Upgrade step enum."""
    pre_start = 'NodePreStart'
    staging = 'NodeStage'
    moving_sessions = 'NodeMoveSessions'
    rtfi = 'NodeInstallSoftware'
    flushing_scache = 'NodeFlushScache'
    wait_for_cluster_faults = 'NodeWaitForClusterFaults'
    wait_for_volumes_healthy = 'NodeWaitForVolumesHealthy'
    finished = 'NodeFinished'


class HealthcheckSeverity(ApiEnum):
    """Severity of reported healthcheck results."""
    informational = 'informational'
    best_practice = 'bestPractice'
    warning = 'warning'
    critical = 'critical'
    error = 'error'


class BundleReportType(ApiEnum):
    """The type of log collection."""
    storage = 'Storage'
    compute = 'Compute'
    management = 'Management'


class BundleState(ApiEnum):
    """The current state of Support Bundle."""
    not_requested = 'notRequested'
    in_progress = 'inProgress'
    canceled = 'canceled'
    completed = 'completed'
    failed = 'failed'
    timed_out = 'timedOut'
    deleted = 'deleted'
