"""Class representing a generic version."""
import functools

from schematics import undefined

from sf_models import sf_version


@functools.total_ordering
class GenericVersion(object):
    """GenericVersion class.

    This represents a generic tuple version class, not Solidfire version.
    This can be any version in format x[.y]* where x and y are ints.
    Ex: 1.0, 1.2.4, 1.3.4.65.211

    :param str version_string: GenericVersion as represented by a string ('8.0.0.1').
    """

    def __init__(self, version_string):
        self.version_string = version_string
        self.version_tuple, _ = sf_version.get_version_tuple_and_sha(self.version_string)

    def __hash__(self):
        """Custom hash function."""
        return hash(self.version_tuple)

    def __eq__(self, other):
        """Equality operator.

        :param sf_models.GenericVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """
        if other is None or other is undefined.Undefined:
            return False
        elif isinstance(other, GenericVersion):
            return self.version_tuple == other.version_tuple
        else:
            raise ValueError('Incomparable types: {} vs {}'
                             .format(other.__class__.__name__, GenericVersion.__name__))

    def __ne__(self, other):
        """Inequality operator.

        :param sf_models.GenericVersion other: Object to compare against
        """
        return not self.__eq__(other)

    def __lt__(self, other):
        """Less than operator.

        :param sf_models.GenericVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """
        if not isinstance(other, GenericVersion):
            raise ValueError('Incomparable types: {} vs {}'
                             .format(other.__class__.__name__, GenericVersion.__name__))
        return self.version_tuple < other.version_tuple

    def __str__(self):
        """String representation."""
        return '{}'.format(self.version_string)

    def __repr__(self):
        return 'GenericVersion({})'.format(self.version_string)
