"""Class representing a SolidFire version.

This code was copied from sf_platform.utils.helpers, but has been modified.
"""
import re
import functools
import netaddr
from schematics import undefined
import abc
import six


ACCEPTABLE_VERSION_FORMATS = \
    ['<major>.<minor>[.<version>[.<patch>]]-<commits_since_major.minor>-g<commit-sha>',
     '<major>.<minor>.<int>.<int>',
     '<major>.<minor>.<int>.<int>[-<commits_since_major.minor>-g<commit-sha>]']

# Regex to look for supported build versions. Supports dev builds which append
# the git commit sha to version.
GIT_DESCRIBE_REGEX = re.compile(r'((?:\d+\.?){2,4})(-\d+)?(-g[a-f\d]+)?', re.I)


def get_version_tuple(version_string):
    """Make a version tuple out of a string.

    This will take any reasonable version string composed of "."-separated
    integers and convert it into a tuple of values.

    >>> assert get_version_tuple('8.0.0.8') > get_version_tuple('5.5')
    >>> assert get_version_tuple('4') < get_version_tuple('5.5')
    >>> assert get_version_tuple('3.898.23') < get_version_tuple('5.5')

    :param str version_string: The version as a "." separated-string.
    :return: A tuple of integers, suitable for comparison.
    :rtype: tuple[int, list[str]]
    """
    try:
        return tuple(map(int, version_string.split('.')))
    except ValueError:
        raise ValueError('Invalid version: {} - not "."-separated integers'.format(version_string))
    except AttributeError:
        raise ValueError('Invalid version type. Expected: str, Given: {}'.format(type(version_string)))


def get_version_tuple_and_sha(version_str):
    """This will take any reasonable version string and return version tuple and commit sha.

    :param str version_str: The version string in the above mentioned format.
    :rtype: tuple
    :return: A tuple of integers, suitable for comparison, commit-hash

    >>> assert get_version_tuple_and_sha('12.0.152-2-g512dc1b') > get_version_tuple_and_sha('5.5')
    >>> assert get_version_tuple_and_sha('3.898.23')[0] == (3, 898, 23)
    >>> assert get_version_tuple_and_sha('11.7.2.3')[0] == (11, 7, 2, 3)
    >>> assert get_version_tuple_and_sha('11.7.2-3-gabfabf')[1] == 'gabfabf'
    """
    release_versions, commit_sha = None, None
    try:
        match = GIT_DESCRIBE_REGEX.search(version_str)
        release_versions = tuple(map(int, match.group(1).split('.')))
        # Distance and SHA will be missing from tagged releases
        if match.group(2) is not None and match.group(3) is not None:
            commit_distance = int(match.group(2).split('-')[1])
            commit_sha = match.group(3).split('-')[1]

            release_versions = release_versions + (commit_distance,)
    except (ValueError, AttributeError, TypeError):
        msg = ('Got an invalid version: {}. Expecting format: {}'.format(version_str,
                                                                         ACCEPTABLE_VERSION_FORMATS))
        raise ValueError(msg)

    return release_versions, commit_sha


@six.add_metaclass(abc.ABCMeta)
class BaseVersion(object):
    """Base class for representing solidfire build versions.

    :param str version_string: SolidFire version as represented by a string ('8.0.0.1').
    :param str dev_image_repo: This overrides the repo name on the pxe server for development use.
    :raises ValueError: If the specified version is not a recognized version of the SolidFire Element OS.
    """

    ELEMENT_MAP = {
        1: 'hydrogen',
        4: 'beryllium',
        5: 'boron',
        6: 'carbon',
        7: 'nitrogen',
        8: 'oxygen',
        9: 'fluorine',
        10: 'neon',
        11: 'sodium',
        12: 'magnesium',
        13: 'aluminum',
        14: 'silicon',
        15: 'phosphorus',
        16: 'sulfur',
        1000: 'ember'
    }
    STACK_RELEASE_MAP = {
        10: {
            2: '18.1',
            3: '18.2',
            4: 'Triton-pPatch',
        },
        11: {
            0: '18.3',
            1: '18.3.1',
            3: '19.1',
            5: '19.2',
            7: '19.2.1',
        },
        12: {
            0: '20.1'
        }
    }

    def __init__(self, version_string, dev_image_repo=None):
        """Init.

        :param version_string:
        :param dev_image_repo:
        """
        self.version_string = version_string
        _tpl, _sha = get_version_tuple_and_sha(self.version_string)
        self.version_tuple = _tpl
        self.commit_sha = _sha
        self.dev_image_repo = dev_image_repo
        self.stack_release = None
        self.release = None

    @abc.abstractmethod
    def __hash__(self):
        """Custom hash function."""

    @abc.abstractmethod
    def __eq__(self, other):
        """Equality operator. Ember might decide equality on different conditions.

        :param sf_models.SfVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """

    def __ne__(self, other):
        """Inequality operator.

        :param sf_models.SfVersion other: Object to compare against
        """
        return not self.__eq__(other)

    def __str__(self):
        """String representation."""
        return '{}'.format(self.version_string)

    @abc.abstractmethod
    def __lt__(self, other):
        """Less than operator.

        :param sf_models.SfVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """

    @abc.abstractmethod
    def __repr__(self):
        """Repr.

        :rtype: str
        """

    @property
    @abc.abstractmethod
    def is_ember(self):
        """Whether this build is an ember build.

        :rtype: bool
        """

    @property
    @abc.abstractmethod
    def element(self):
        """Backwards compatible property to return the element.

        :rtype: list[str]
        """

    @property
    @abc.abstractmethod
    def element_id(self):
        """The element ID is the major version.

        :rtype: str
        """

    @property
    @abc.abstractmethod
    def rtfi_name(self):
        """Return the "rtfi" name.

        :rtype: str
        """

    @property
    @abc.abstractmethod
    def api_endpoint(self):
        """Return the api endpoint.

        :rtype: str
        """

    @abc.abstractmethod
    def get_pxe_url(self, network_profile):
        """Return the pxe url for reset node call.

        :param sf_models.Base network_profile: network profile object of node.
        :rtype: str
        """

    @property
    def xnode_name(self):
        """Return the "xnode" name.

        :rtype: str
        """
        return 'solidfire-binaries-{!s}'.format(self)

    @property
    def fdva_name(self):
        """Return the "fdva" name.

        :rtype: str
        """
        return 'solidfire-fdva-{!s}'.format(self)

    @property
    def package_name(self):
        """Return the "package" name.

        :rtype: str
        """
        return 'solidfire-san-{!s}'.format(self)

    def set_dev_repo_name(self, repo_name):
        """Set the pxe url for custom builds used in development.

        :param str repo_name: the complete url to the pxe image
        """
        self.dev_image_repo = repo_name

    @classmethod
    def from_version_info(cls, version_info):
        """Create an SfApiVersion object from a version-info dictionary.

        :param dict version_info: A version-info dictionary as returned by GetVersionInfo.
        :raises ValueError: If the version-info dictionary has no 'Version' key.
        :rtype: sf_models.SfVersion
        """
        version_string = version_info.get('Version')
        if version_string is not None:
            return cls(version_string)
        else:
            raise ValueError('Invalid version info: {}'.format(version_info))

    @classmethod
    def from_build_dict(cls, build_dict):
        """Create an SfApiVersion object from an AT2 build dictionary.

        :param dict build_dict: An AT2 build dictionary.
        :raises ValueError: If the build dictionary has no 'version' key.
        :rtype: sf_models.SfVersion
        """
        version_string = build_dict.get('version')
        if version_string is not None:
            return cls(version_string)
        else:
            raise ValueError('Invalid build dictionary: {}'.format(build_dict))


@functools.total_ordering
class SfVersion(BaseVersion):
    """Version of the SolidFire Element OS.

    :param str version_string: SolidFire version as represented by a string ('8.0.0.1').
    :param str dev_image_repo: This overrides the repo name on the pxe server for development use.
    :raises ValueError: If the specified version is not a recognized version of the SolidFire Element OS.
    """

    def __init__(self, version_string, dev_image_repo=None):
        """Init.

        :param version_string:
        :param dev_image_repo:
        """
        super(SfVersion, self).__init__(version_string, dev_image_repo)

        maj_version = self.version_tuple[0]
        if maj_version in self.ELEMENT_MAP:
            self.release = self.element
            if maj_version >= 8 and self.version_tuple[1] > 0:
                self.release += '-patch{}'.format(self.version_tuple[1])
        else:
            raise ValueError('Unrecognized version: {}'.format(version_string))

        # Translate the Provided Element version into its corresponding stack release version.
        self.stack_release = None
        if maj_version in self.STACK_RELEASE_MAP:
            if self.version_tuple[1] in self.STACK_RELEASE_MAP[maj_version]:
                min_version = self.version_tuple[1]
                self.stack_release = self.STACK_RELEASE_MAP[maj_version][min_version]

    def __hash__(self):
        """Custom hash function."""
        return hash((self.release, self.version_tuple))

    def __eq__(self, other):
        """Equality operator.

        :param sf_models.SfVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """
        if other is None or other is undefined.Undefined:
            return False
        elif isinstance(other, SfVersion):
            return self.version_tuple == other.version_tuple
        else:
            raise ValueError('Incomparable types: {} vs {}'
                             .format(other.__class__.__name__, SfVersion.__name__))

    def __lt__(self, other):
        """Less than operator.

        :param sf_models.SfVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """
        if not isinstance(other, SfVersion):
            raise ValueError('Incomparable types: {} vs {}'
                             .format(other.__class__.__name__, SfVersion.__name__))
        return self.version_tuple < other.version_tuple

    def __repr__(self):
        return 'SfVersion({})'.format(self.version_string)

    @property
    def is_ember(self):
        """This does not represent an Ember build.

        :rtype: bool
        """
        return False

    @property
    def element(self):
        """Backwards compatible property to return the element.

        :rtype: list[str]
        """
        return self.ELEMENT_MAP[self.version_tuple[0]]

    @property
    def element_id(self):
        """The element ID is the major version.

        :rtype: str
        """
        return self.version_tuple[0]

    @property
    def rtfi_name(self):
        """Return the "rtfi" name.

        :rtype: str
        """
        if self.dev_image_repo:
            return 'solidfire-rtfi-{}-{!s}'.format(self.dev_image_repo, self)
        else:
            return 'solidfire-rtfi-{}-{!s}'.format(self.release, self)

    @property
    def api_endpoint(self):
        """Return the api endpoint.

        :rtype: str
        """
        patch = '0' if self.version_tuple[0] < 8 else self.version_tuple[1]
        return '{}.{}'.format(self.version_tuple[0], patch)

    def get_pxe_url(self, network_profile):
        """Return the pxe url for reset node call.

        :param sf_models.Base network_profile: network profile object of node.
        :rtype: str
        """
        fetch_v6 = 'ftp://[{image_server}]/{basepath}/filesystem.squashfs'
        fetch_v4 = 'ftp://{image_server}/{basepath}/filesystem.squashfs'
        fetch = fetch_v6 if netaddr.valid_ipv6(network_profile.pxe_server_address) else fetch_v4
        basepath = 'images/rtfi/{version}/casper'.format(version=self.rtfi_name)
        fetch = fetch.format(image_server=network_profile.pxe_server_address,
                             basepath=basepath)
        return fetch


class EmberVersion(BaseVersion):
    """Version of SolidFire Ember layer.

    :param str version_string: SolidFire version as represented by a string ('8.0.0.1').
    :param str dev_image_repo: This overrides the repo name on the pxe server for development use.
    :raises ValueError: If the specified version is not a recognized version of the SolidFire Element OS.
    """

    def __init__(self, version_string, dev_image_repo=None):
        super(EmberVersion, self).__init__(version_string, dev_image_repo)

        # Cannot check for major version in ELEMENT_MAP because ember builds don't start with 1000.x.x.x
        # EmberVersion to hardwire release to 'ember'
        self.release = 'ember'

    def __hash__(self):
        """Custom hash function."""
        return hash((self.release, self.version_tuple))

    def __eq__(self, other):
        """Equality operator.

        :param sf_models.EmberVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """
        if other is None or other is undefined.Undefined:
            return False
        elif isinstance(other, EmberVersion):
            return self.version_tuple == other.version_tuple
        else:
            raise ValueError('Incomparable types: {} vs {}'
                             .format(other.__class__.__name__, EmberVersion.__name__))

    def __ne__(self, other):
        """Inequality operator.

        :param sf_models.EmberVersion other: Object to compare against
        """
        return not self.__eq__(other)

    def __lt__(self, other):
        """Less than operator.

        :param sf_models.EmberVersion other: Object to compare against
        :raises ValueError: If other is not the correct type
        """
        if not isinstance(other, EmberVersion):
            raise ValueError('Incomparable types: {} vs {}'
                             .format(other.__class__.__name__, EmberVersion.__name__))
        return self.version_tuple < other.version_tuple

    def __str__(self):
        """String representation."""
        return '{}'.format(self.version_string)

    def __repr__(self):
        return 'EmberVersion({})'.format(self.version_string)

    @property
    def is_ember(self):
        """Version object represents an ember build.

        :rtype: bool
        """
        return True

    @property
    def element(self):
        """Element is always ember for EmberVersion.

        :rtype: str
        """
        return 'ember'

    @property
    def element_id(self):
        """Ember element ID is always 1000.

        :rtype: int
        """
        return 1000

    @property
    def rtfi_name(self):
        """Return the rtfi-name, no solidfire-rtfi prefix for ember builds.

        :rtype: str
        """
        if self.dev_image_repo:
            return '{}-{!s}'.format(self.dev_image_repo, self)
        else:
            return '{}-{!s}'.format(self.release, self)

    @property
    def api_endpoint(self):
        """Return the api endpoint. None on ember.

        :rtype: str
        """
        return

    def get_pxe_url(self, network_profile):
        """For ember, this logic is embedded in sf-rtfi repo.

        :param sf_models.Base network_profile: network profile object of node.
        :rtype: str
        """
        return
