"""Wraps all the built-in _types in schematics.types and contains custom types.

This module should be imported instead of schematics.types when creating models.
"""
import ast
import copy
import datetime
import random
import re
import string

import rstr
import six
from schematics import (
    exceptions,
    models,
)
from schematics import types as _types

import sf_models
from sf_models import (
    _network_interface_name,
    data_size,
    enums,
    sf_version,
)

API_VERSION_LEN = 2
BUILD_VERSION_LEN = 4


class BaseType(_types.BaseType):
    """Extended BaseType with extra parameters and methods.

    :param sf_models.SfVersion version_added: SolidFire Version in which this attribute became valid
    :param sf_models.SfVersion version_removed: SolidFire Version in which this attribute became invalid
    """
    def __init__(self, *args, **kwargs):
        self.version_added = kwargs.get('version_added')
        self.version_removed = kwargs.get('version_removed')

        if self.version_added is not None:
            kwargs.pop('version_added')
        if self.version_removed is not None:
            kwargs.pop('version_removed')

        super(BaseType, self).__init__(*args, **kwargs)


class BooleanType(BaseType, _types.BooleanType):
    """BooleanType object with extended BaseType parameters and methods."""


class DateTimeType(BaseType, _types.DateTimeType):
    """DateTimeType object with extended BaseType parameters and methods."""
    def __init__(self, *args, **kwargs):
        # Require time-zone data unless otherwise specified
        if kwargs.get('tzd') is None:
            kwargs.update(tzd='require')
        super(DateTimeType, self).__init__(*args, **kwargs)

    def to_primitive(self, value, context=None):
        """Convert to an ISO-8601-valid string.

        If a time-designator is represented as ``[+-]<dddd>``, convert it to ``[+-]<dd>:<dd>``.

        :param datetime.datetime value: Datetime object to convert.
        :param schematics.datastructures.Context context: Context object.
        :rtype: str
        """
        if value is None:
            return ''
        primitive_string = super(DateTimeType, self).to_primitive(value, context)

        # Represent all time-zone designators in a [+-]<dd>:<dd> format
        primitive_split = primitive_string.split(':')
        datetime_string = primitive_split[0]
        if len(primitive_split[-1]) > 2:
            primitive_split[-1] = primitive_split[-1][:-2] + ':' + primitive_split[-1][-2:]
        for split_string in primitive_split[1:]:
            datetime_string += ':' + split_string

        return datetime_string

    def to_native(self, value, context=None):
        """Convert value to native type.

        :param datetime.datetime value: Datetime object to convert.
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.datetime|None
        """

        if value == '':
            return None
        return super(DateTimeType, self).to_native(value, context)


class DateType(BaseType, _types.DateType):
    """DateType object with extended BaseType parameters and methods."""

    def to_primitive(self, value, context=None):
        """Convert to primitive type.

        :param datetime.datetime|None value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """

        if value is None:
            return ''
        return super(DateType, self).to_primitive(value, context)

    def to_native(self, value, context=None):
        """Convert value to native type.

        :param str value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.date|None
        """

        if value == '':
            return None
        return super(DateType, self).to_native(value, context)


class DecimalType(BaseType, _types.DecimalType):
    """DecimalType object with extended BaseType parameters and methods."""


class DictType(_types.DictType, BaseType):
    """DictType object with extended BaseType parameters and methods."""

    def _mock(self, context=None):
        """Create a mock dictionary object.

        :param schematics.datastructures.Context context: Context object.
        :rtype: dict
        """
        return {'mocked_value': self.field._mock(context)}  # pylint: disable=protected-access


class EmailType(BaseType, _types.EmailType):
    """EmailType object with extended BaseType parameters and methods."""


class FloatType(BaseType, _types.FloatType):
    """FloatType object with extended BaseType parameters and methods."""


class GeoPointType(BaseType, _types.GeoPointType):
    """GeoPointType object with extended BaseType parameters and methods."""


class HashType(BaseType, _types.HashType):
    """HashType object with extended BaseType parameters and methods."""


class IPAddressType(BaseType, _types.IPAddressType):
    """IPAddressType object with extended BaseType parameters and methods."""

    def validate(self, value, context=None):
        """Custom validate function that accepts an empty string.

        :param datetime.datetime|str|None value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.datetime|None
        """
        if value is None or value == '':
            return None
        return super(IPAddressType, self).validate(value, context)

    def to_native(self, value, context=None):
        """Convert value to native type.

        :param str value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.datetime|None
        """
        if not value:
            return value
        return super(IPAddressType, self).to_native(value, context)

    def to_primitive(self, value, context=None):
        """Convert to primitive type.

        :param datetime.datetime|None value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        if value is None:
            return ''
        return super(IPAddressType, self).to_primitive(value, context)

    def _mock(self, context=None):
        """Mocks by returning either an ipv4 or ipv6 address.

        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        address_type = random.choice([IPv4Type, IPv6Type])(required=self.required)
        return address_type._mock()  # pylint: disable=protected-access


class IPv4Type(BaseType, _types.IPv4Type):
    """IPv4Type object with extended BaseType parameters and methods."""


class IPv6Type(IPAddressType):
    """A field that stores a valid IPv6 address."""

    VERSION = 'v6'
    REGEX = re.compile(r'^%s$' % _types.net.IPV6, re.I + re.X)

    def _mock(self, context=None):
        """Mock for ipv6 works in master branch of schematics but not here so we are adding it.

        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        mock_func = random.choice([self._mock_fully_expanded, self._mock_last_segment_only,
                                   self._mock_beginning_zeros, self._mock_zeros_in_middle,
                                   self._mock_shortened_segments, self._mock_ending_zeros_abbrev,
                                   self._mock_ending_zeros, self._mock_shortened_segments_with_zeros,
                                   self._mock_shortened_segments_with_varying_zeros])
        return mock_func()

    def _mock_all(self):
        """Returns a list with one mock of each type for testing.

        :rtype: list[str]
        """
        return [self._mock_fully_expanded(), self._mock_last_segment_only(),
                self._mock_beginning_zeros(), self._mock_zeros_in_middle(),
                self._mock_shortened_segments(), self._mock_ending_zeros_abbrev(),
                self._mock_ending_zeros(), self._mock_shortened_segments_with_zeros(),
                self._mock_shortened_segments_with_varying_zeros()]

    @staticmethod
    def _mock_fully_expanded():
        """Mocks a fully expanded ipv6 address.

        ex: 2001:db8:20de:bb39:5f58:b95:f703:dbac
        :rtype: str
        """
        return '2001:db80:' + ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(6)
        )

    @staticmethod
    def _mock_last_segment_only():
        """Mock an address that only contains the last segment.

        ex: ::c4cd
        :rtype: str
        """
        return '::' + ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(1))

    @staticmethod
    def _mock_beginning_zeros():
        """Mock an address that starts with zeros.

        ex: ::df4d:66cd:3140:4d4b
        :rtype: str
        """
        return '::' + ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(4))

    @staticmethod
    def _mock_ending_zeros():
        """Mock an address that ends with zeros.

        ex: a9f4:e915:f5be:b5b:a685:8073:c9d1:0000
        :rtype: str
        """
        return ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(7)) + ':0000'

    @staticmethod
    def _mock_ending_zeros_abbrev():
        """Mock an address that ends with abbreviated zeros.

        ex: fb01:c6f2:b750:48c1::
        :rtype: str
        """
        return ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(4)) + '::'

    @staticmethod
    def _mock_zeros_in_middle():
        """Mock an address with zeros in the middle shortened.

        ex: 8a44:9fd9::a995:7321:d901

        :rtype: str
        """
        hextets = 7
        first = random.randrange(6)+1
        last = random.randrange(hextets-first)+1

        return ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(first)) + '::' + ':'.join(
            '%x' % (random.randrange(1 << 16)) for _ in range(last))

    @staticmethod
    def _mock_shortened_segments():
        """Tests with variations of ipv6 addresses guaranteed to contain shortened semgments.

        In the fully expanded mock, there's only a 1 in 16 chance the number will be less than 1000(hex)
        ex: 2001:db8:95:84:5b:43:42:37

        :rtype: str
        """
        return '2001:db80:' + ':'.join(
            '%x' % (random.randrange(1 << 8)) for _ in range(6)
        )

    @staticmethod
    def _mock_shortened_segments_with_varying_zeros():
        """Tests with variations of ipv6 addresses with shortened segments with leading zeros.

        ex: 2001:db808a:0b6:00ce:37:008f:066
        :rtype: str
        """
        return '2001:db80:' + ':'.join(
             '%s%x' % ('0' * random.randrange(3), random.randrange(1 << 8)) for _ in range(6)
        )

    @staticmethod
    def _mock_shortened_segments_with_zeros():
        """Tests with variations of ipv6 addresses with shortened segments with leading zeros.

        ex: 2001:db8:008a:00b6:00ce:0037:008f:0066
        :rtype: str
        """
        return '2001:db80:' + ':'.join(
            '00%x' % (random.randrange(1 << 8)) for _ in range(6)
        )


class IntType(BaseType, _types.IntType):
    """IntType object with extended BaseType parameters and methods."""


class ListType(_types.ListType, BaseType):
    """ListType object with extended BaseType parameters and methods."""


class LongType(BaseType, _types.LongType):
    """LongType object with extended BaseType parameters and methods."""


class MD5Type(BaseType, _types.MD5Type):
    """MD5Type object with extended BaseType parameters and methods."""


class ModelType(_types.ModelType, BaseType):
    """ModelType object with extended BaseType parameters and methods."""


class MultilingualStringType(BaseType, _types.MultilingualStringType):
    """MultilingualStringType object with extended BaseType parameters and methods."""


class NumberType(BaseType, _types.NumberType):
    """NumberType object with extended BaseType parameters and methods."""


class PolyModelType(_types.PolyModelType, BaseType):
    """PolyModelType object with extended BaseType parameters and methods."""

    def _convert(self, value, context):
        """Converts the value to its native type; in this case, a data model.

        This is overridden since in schematics-2.1.0 PolyModelTypes are now recursively validated. This
        was causing _claim_polymorphic to be called with a data model and also created the following
        bug: https://github.com/schematics/schematics/issues/564.

        This workaround skips the conversion process if the value is already a data model.

        :param Any value: the value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.Base
        """
        # If it's already a data model, skip the conversion.
        if isinstance(value, models.Model):
            return value
        return super(PolyModelType, self)._convert(value, context)


class SHA1Type(BaseType, _types.SHA1Type):
    """SHA1Type object with extended BaseType parameters and methods."""


class StringType(BaseType, _types.StringType):
    """StringType object with extended BaseType parameters and methods."""
    allow_casts = bytes

    def __init__(self, *args, **kwargs):
        super(StringType, self).__init__(*args, **kwargs)
        if self.regex is not None:
            if self.min_length is not None or self.max_length is not None:
                raise TypeError('if regex is specified, min/max-length must be None')

    def _mock(self, context=None):
        """Return a random string with valid length and regex formatting.

        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.data_size.DataSize
        """
        if self.regex is None:
            return super(StringType, self)._mock(context=context)
        else:
            return rstr.xeger(self.regex)


class TimestampType(BaseType, _types.TimestampType):
    """TimestampType object with extended BaseType parameters and methods."""


class UTCDateTimeType(BaseType, _types.UTCDateTimeType):
    """UTCDateTimeType object with extended BaseType parameters and methods."""


class TimedeltaStringType(BaseType, _types.TimedeltaType):
    """TimedeltaStringType object with extended BaseType parameters and methods.

    Deserializes a string from the "HH:MM:SS.uuuuuu" form to a datetime.timedelta object.
    It's unlikely that there will be more than 99 hours. Assumes max of 99, 59, 59, 999999.

    Note - Schematics's _types.Timedelta deserializes a float to a datetime.timedelta. This
    TimedeltaStringType has a specific string format.
    """

    def to_native(self, value, context=None):
        """Convert string to a timedelta object.

        :param str value: Value to convert to native timedelta object
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.timedelta
        """
        if isinstance(value, datetime.timedelta):
            return value

        try:
            # Split off hours because you can't go above 23 with a datetime object
            #  but we want to use the smart parsing of the datetime object to deal with microseconds...
            td_split = value.split(':')
            hours_int = int(td_split[0])
            remainder_str = ':'.join(td_split[1:])
            dt = datetime.datetime.strptime(remainder_str, '%M:%S.%f')
        except ValueError:
            raise exceptions.ConversionError('Invalid timedelta string {}'.format(type(value)))

        return datetime.timedelta(hours=hours_int, minutes=dt.minute, seconds=dt.second,
                                  microseconds=dt.microsecond)

    def to_primitive(self, value, context=None):
        """Converts a timedelta to a string in the format "HH:MM:SS.uuuuuu".

        :param datetime.timedelta value: data size object to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: int
        """
        hours, remainder = divmod(value.total_seconds(), 3600)
        minutes, remainder = divmod(remainder, 60)
        seconds, microseconds = divmod(remainder, 1)
        microseconds = round(microseconds, 6)

        hours = '{:02d}'.format(int(hours))
        output = hours + ':' + str(int(minutes)) + ':' + str(int(seconds))
        return output + format(microseconds, 'f').lstrip('0')

    def validate(self, value, context=None):
        """Verify that value is a timedelta.

        :param datetime.timedelta|str|None value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.timedelta
        """
        if not isinstance(value, datetime.timedelta):
            raise exceptions.ValidationError('{} is not a {}'.format(value, datetime.timedelta))
        return value

    def _mock(self, context=None):
        return datetime.timedelta(hours=random.randint(0, 99), minutes=random.randint(0, 59),
                                  seconds=random.randint(0, 59), microseconds=random.randint(0, 999999))


class URLType(BaseType, _types.URLType):
    """URLType object with extended BaseType parameters and methods."""


class UUIDType(BaseType, _types.UUIDType):
    """UUIDType object with extended BaseType parameters and methods."""


class UnionType(BaseType, _types.UnionType):
    """UnionType object with extended BaseType parameters and methods."""


class DataSizeType(BaseType, _types.NumberType):
    """Deserializes an int into a DataSize object. If units are not specified use bytes.

    :param sf_models.data_size.DataUnit unit: Data size unit.
    """

    def __init__(self, unit=data_size.DataUnit.bytes, **kwargs):
        super(DataSizeType, self).__init__(**kwargs)
        self._unit = unit

    def to_primitive(self, value, context=None):
        """Convert data type to raw size.

        :param sf_models.data_size.DataSize value: data size object to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: int
        """
        return getattr(value, self._unit.name)

    def to_native(self, value, context=None):
        """Convert value to data size object.

        :param int value: Value to convert to native data size object
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.data_size.DataSize
        """
        if isinstance(value, data_size.DataSize):
            return value
        return data_size.DataSize(value, unit=self._unit)

    def _mock(self, context=None):
        """Return a data size object of random size.

        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.data_size.DataSize
        """
        return data_size.DataSize(random.randint(4096, 1024 ** 2))


class EnumType(BaseType):
    """EnumType object with extended BaseType parameters and methods.

    Deserializes into a python 3 style enum.

    :param aenum.Enum enum_class: Reference to enum class
    :param bool lowercase: Whether to lowercase values to cover up silly product behavior.
    """

    def __init__(self, enum_class, lowercase=False, **kwargs):
        super(EnumType, self).__init__(**kwargs)
        self._lowercase = lowercase
        self._enum_class = enum_class

    def _mock(self, context=None):
        """Custom mock function that randomly chooses a value.

        :param schematics.datastructures.Context context: Context object
        :rtype: str|int
        """
        values = list(self._enum_class)
        choice = random.randint(0, len(values) - 1)
        return values[choice]

    def validate(self, value, context=None):
        """Custom validate function that makes sure values is an instance of the enum class.

        :param aenum.EnumMeta value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: aenum.EnumMeta
        """
        if not isinstance(value, self._enum_class):
            raise exceptions.ValidationError('{} is not of enum type {}'.format(value, self._enum_class))
        return value

    def to_native(self, value, context=None):
        """Convert a raw string into native enum type.

        This causes a ValueError to be raised if an invalid value is passed to the
        data model constructor. This is different from the behavior of StringType, where the
        choices are not verified until validate() is called. However, this is similar to what
        happens when you pass a string to an IntType.

        :param str|int value: Value to convert to enum
        :param schematics.datastructures.Context context: Context object
        :rtype: Any
        """

        if self._lowercase and isinstance(value, six.string_types):
            value = value.lower()
        return self._enum_class(value)

    def to_primitive(self, value, context=None):
        """Convert enum to primitive string type.

        :param aenum.EnumMeta value: Enum to convert to primitive value
        :param schematics.datastructures.Context context: Context object
        :rtype: str|int
        """
        return value.value


class IntIdType(IntType):
    """Use for integer IDs that should be unique.

    Same as IntType, but changes mocking behavior to prevent collisions.
    """
    _count = 0

    def _mock(self, context=None):
        """Mocks an int that is between min_value and max_value.

        The static _count variable is incremented each time to prevent collisions.

        :param schematics.datastructures.Context context: Context object
        :rtype: int
        """
        value = IntIdType._count
        min_value = 0 if self.min_value is None else self.min_value
        if self.max_value is not None:
            # If we reach the max value, loop back around to the min
            value = value % (self.max_value - min_value + 1) + min_value
        else:
            value += min_value
        IntIdType._count += 1
        return value


class At2OobType(StringType):
    """Use for iDracIP field for AT2 resources.

    Accepts IPv4, hostnames that is in one of the valid suffixes below, and will strip any spaces after
    the address/hostname.
    """
    VALID_SUFFIXES = ('oob.den.solidfire.net', 'oob.pw.solidfire.net', 'oob.sqd.solidfire.net',
                      'mgmt.openenglab.netapp.com', 'oob.ict.openenglab.netapp.com',
                      'heartgard.den.solidfire.net')

    def to_native(self, value, context=None):
        """Convert a raw string into native At2OobType. This will rstrip any extra spaces.

        :param str value: Value to convert to At2OobType.
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        return super(At2OobType, self).to_native(value=value.rstrip(), context=context)

    def validate(self, value, context=None):
        """Verify that value is an IPv4 address, or a valid hostname.

        :param str value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        try:
            ipv4 = IPv4Type()
            return ipv4.validate(value, context)
        except exceptions.ValidationError:
            for suffix in At2OobType.VALID_SUFFIXES:
                if value.endswith(suffix):
                    return super(At2OobType, self).validate(value, context)
        raise exceptions.ValidationError('Invalid hostname suffix for {}'.format(value))

    def _mock(self, context=None):
        """Randomly mock an IPv4 or At2OoB hostname.

        :param schematics.datastructures.Context context: Context object
        :return: An IPv4 address or a hostname with a valid suffix.
        :rtype: str
        """
        random_str = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        rand_hostname = random_str + random.choice(At2OobType.VALID_SUFFIXES)
        rand_ip = IPv4Type()._mock()  # pylint: disable=protected-access
        return random.choice([rand_hostname, rand_ip])


class FqdnType(StringType):
    """Use for MIP which can be either FQDN or IPAddress type.

    Accepts IPv4, IPv6, and hostnames that is in one of the valid suffixes below
    """
    VALID_SUFFIXES = ('den.solidfire.net', 'pw.solidfire.net')

    def validate(self, value, context=None):
        """Verify that value is an IPv4 address, IPv6 update, or a valid hostname.

        :param str value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        try:
            ip = IPAddressType()
            return ip.validate(value, context)
        except exceptions.ValidationError:
            for suffix in FqdnType.VALID_SUFFIXES:
                if value.endswith(suffix):
                    return super(FqdnType, self).validate(value, context)
        raise exceptions.ValidationError('Invalid hostname suffix for {}'.format(value))

    def _mock(self, context=None):
        """Randomly mock an IPv4 ipv6, or FqdnType hostname.

        :param schematics.datastructures.Context context: Context object
        :return: An IPv4 address or a hostname with a valid suffix.
        :rtype: str
        """
        random_str = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        rand_hostname = random_str + random.choice(FqdnType.VALID_SUFFIXES)
        rand_ip = IPAddressType(required=True)._mock()  # pylint: disable=protected-access
        return random.choice([rand_hostname, rand_ip])


class At2OobOrEmptyStringType(At2OobType):
    """Same as At2OobType, except this will allow the empty string."""
    def _mock(self, context=None):
        """Randomly mock either an empty string or an At2Oob address/hostname.

        :param schematics.datastructures.Context context: Context object
        :return: An empty string, IPv4 address or hostname with a valid suffix.
        :rtype: str
        """
        return random.choice(['', super(At2OobOrEmptyStringType, self)._mock(context)])

    def validate(self, value, context=None):
        """Validate that the value is either an empty string, IPv4 address, or valid hostname.

        :param str value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        if value == '':
            return value
        return super(At2OobOrEmptyStringType, self).validate(value, context)

    def to_primitive(self, value, context=None):
        """Convert a At2OobOrEmptyStringType into a raw string.

        :param str value: Value to convert to raw string.
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        if value == '':
            return value
        return super(At2OobOrEmptyStringType, self).to_primitive(value, context)

    def to_native(self, value, context=None):
        """Convert a raw string into native At2OobOrEmptyStringType.

        :param str value: Value to convert to At2OobOrEmptyStringType.
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.types.At2OobOrEmptyStringType
        """
        if value == '':
            return value
        return super(At2OobOrEmptyStringType, self).to_native(value, context)


class OptionalDateTimeType(DateTimeType):
    """Same as schematics.types.DateTimeType, but also accepts an empty string.

    This is needed because some responses for time fields return empty strings instead of null.
    This only applies to optional DateTime fields, so required must be set False.
    """
    def __init__(self, **kwargs):
        super(OptionalDateTimeType, self).__init__(**kwargs)
        if 'default' not in kwargs:
            self._default = None
        if kwargs.get('required'):
            raise ValueError('OptionalDateTimeType set to required (MUST be optional)')

    def validate(self, value, context=None):
        """Custom validate function that accepts an empty string.

        :param datetime.datetime|str|None value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.datetime|None
        """
        if value is None or value == '':
            return None
        return super(OptionalDateTimeType, self).validate(value, context)

    def to_native(self, value, context=None):
        """Convert value to native type.

        :param str value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: datetime.datetime|None
        """
        if value == '':
            return None
        return super(OptionalDateTimeType, self).to_native(value, context)

    def to_primitive(self, value, context=None):
        """Convert to primitive type.

        :param datetime.datetime|None value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        if value is None:
            return ''
        return super(OptionalDateTimeType, self).to_primitive(value, context)


class OptionalDictType(DictType):
    """Same as schematics.types.DictType, but also accepts an empty string.

    This is needed because some responses for dict fields return empty strings instead of null.
    This only applies to optional Dict fields, so required must be set False.

    :param schematics.types.BaseType field: Dictionary item type definition.
    """
    def __init__(self, field, **kwargs):
        super(OptionalDictType, self).__init__(field, **kwargs)
        if 'default' not in kwargs:
            self._default = None
        if kwargs.get('required'):
            raise ValueError('OptionalDictType set to required (MUST be optional)')

    def validate(self, value, context=None):
        """Custom validate function that accepts an empty string.

        :param dict|str|None value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: dict|None
        """
        if value is None or value == '':
            return None
        return super(OptionalDictType, self).validate(value, context)

    def to_native(self, value, context=None):
        """Convert value to native type.

        :param str value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: dict|None
        """
        if value == '':
            return None
        return super(OptionalDictType, self).to_native(value, context)

    def to_primitive(self, value, context=None):
        """Convert to primitive type.

        :param dict|None value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        if value is None:
            return ''
        return super(OptionalDictType, self).to_primitive(value, context)

    def _convert(self, value, context):
        """Converts the value to its native type; in this case, a data model.

        This workaround skips the conversion process if the value is an empty string.

        :param Any value: the value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.Base
        """

        if value == '':
            return None
        else:
            return super(OptionalDictType, self)._convert(value, context)


class StringListType(ListType):
    """Treat it like a list type, however it's really a comma separated string.

    E.g. '10.1.1.1, 192.1.1.1, 10.3.1.2' will be understood as a list, split at the commas, whitespace
    removed
    """

    def _coerce(self, value):
        """Coerce the value into a list, even if it's a string.

        :param list[str]|str value: value to change to a list
        :return: value if already a list or ',' split and stripped list
        :rtype: list
        :raises schematics.exceptions.ConversionError: When value is not a str or list
        """
        if isinstance(value, six.string_types):
            return [v.strip() for v in value.split(',')]
        elif isinstance(value, list):
            return value
        else:
            raise exceptions.ConversionError('Value is not a string or list; '
                                             'found {}'.format(type(value)))

    # pylint: disable=redefined-builtin
    def _export(self, list_instance, format, context):
        """Export the list as a comma separated string. Calls super to export the field type.

        :param list list_instance: The list to export
        :param str format: NATIVE or PRIMITIVE constants
        :param schematics.datastructures.Context context: Context object
        :return: comma separated string representation of the list
        :rtype: str
        """
        data = super(StringListType, self)._export(list_instance, format, context)
        data = [str(d) for d in data]
        return ', '.join(data)

    def check_length(self, value, context):
        """Verify min and max size. Have to convert to a list then call super.

        :param list|str value: The list or string to check
        :param schematics.datastructures.Context context: Context object
        """

        value = self._coerce(value)
        super(StringListType, self).check_length(value, context)


class LiteralStringListType(ListType):
    """Treat it like a list type, however it's a list that's casted as a string.

    E.g. '[1, 2, 3]' will be treated as a list.
    """

    def _coerce(self, value):
        """Coerce the value into a list, even if it's a string.

        :param list[str]|str value: value to change to a list
        :rtype: list
        :raises schematics.exceptions.ConversionError: When value is not a str or list
        """
        if isinstance(value, six.string_types):
            return ast.literal_eval(value)
        elif isinstance(value, list):
            return value
        else:
            raise exceptions.ConversionError('Value is not a string or list; '
                                             'found {}'.format(type(value)))

    # pylint: disable=redefined-builtin
    def _export(self, list_instance, format, context):
        """Export the list as a comma separated string. Calls super to export the field type.

        :param list list_instance: The list to export
        :param str format: NATIVE or PRIMITIVE constants
        :param schematics.datastructures.Context context: Context object
        :return: comma separated string representation of the list
        :rtype: str
        """
        list_instance = [str(s) for s in list_instance]
        data = super(LiteralStringListType, self)._export(list_instance, format, context)
        data = [str(d) for d in data]
        return str(data)

    def check_length(self, value, context):
        """Verify min and max size. Have to convert to a list then call super.

        :param list|str value: The list or string to check
        :param schematics.datastructures.Context context: Context object
        """

        value = self._coerce(value)
        super(LiteralStringListType, self).check_length(value, context)


class SfApiVersionType(BaseType):
    """SfApiVersion type, deserializes api_endpoint into an SfVersion object."""

    def _mock(self, context=None):
        """Custom mock function that randomly mocks a valid api version string.

        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.SfVersion
        """
        element = six.iterkeys(sf_version.SfVersion.ELEMENT_MAP)
        choice = random.choice(list(element))
        version_string = str(choice) + '.0'
        return sf_version.SfVersion(version_string)

    def to_native(self, value, context=None):
        """Convert value to native type.

        :param str|sf_models.SfVersion value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.SfVersion
        """
        if isinstance(value, sf_version.SfVersion):
            pass
        # The ensuing horrible hack should be removed once PE-1517 is
        # properly fixed. "Allows" empty version field.
        elif value == '0.0':
            value = sf_version.SfVersion('10.3')
            value.version_string = '0.0'
        elif value == '0.0.0.0':
            value = sf_version.SfVersion('10.3.0.0')
            value.version_string = '0.0.0.0'
        else:
            value = sf_version.SfVersion(value)
        return value

    def to_primitive(self, value, context=None):
        """Convert value to primitive type.

        :param sf_models.SfVersion value: Value to convert
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        return value.version_string

    def validate(self, value, context=None):
        """Validates if value is an instance of sf_models.SfVersion.

        :param sf_models.SfVersion|str value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.SfVersion
        """
        if not isinstance(value, sf_version.SfVersion):
            # The ensuing horrible hack should be removed once PE-1517 is
            # properly fixed. "Allows" empty version field.
            if value == '0.0':
                value = sf_version.SfVersion('10.3')
                value.version_string = '0.0'
            else:
                value = sf_version.SfVersion(value)
        if len(value.version_tuple) != API_VERSION_LEN:
            raise exceptions.ValidationError(
                '{} is not a valid API version'.format(value.version_string))
        return value


class SfBuildVersionType(SfApiVersionType):
    """SfBuildVersion type, deserializes build_version into an SfVersion object."""

    def _mock(self, context=None):
        """Custom mock function that randomly mocks a valid build version string.

        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        api_version = super(SfBuildVersionType, self)._mock(context)
        api_version.version_string += '.0.0'
        version_tuple = list(api_version.version_tuple)
        version_tuple.extend([0, 0])
        api_version.version_tuple = tuple(version_tuple)
        return api_version

    def validate(self, value, context=None):
        """Validates if value is an instance of sf_models.SfVersion.

        :param sf_models.SfVersion|str value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.SfVersion
        """
        if not isinstance(value, sf_version.SfVersion):
            # The ensuing horrible hack should be removed once PE-1517 is
            # properly fixed. "Allows" empty version field.
            if value == '0.0.0.0':
                value = sf_version.SfVersion('10.3.0.0')
                value.version_string = '0.0.0.0'
            else:
                value = sf_version.SfVersion(value)
        if len(value.version_tuple) != BUILD_VERSION_LEN:
            raise exceptions.ValidationError(
                '{} is not a valid build version'.format(value.version_tuple))
        return value


class IntIdTypeWithDictKeyConvert(IntIdType):
    """IntId type that handles converting a single key from a dict if necessary.

    In particular, this type is useful for a field that has changed formats during different
    releases to ensure backwards compatibility for a specific field between those different
    releases.  If you need to represent the newer data format in the model, then you should
    add an additional field to track the new format and use the existing field to keep
    compatibility.

    For example, VolumeAccessGroup has a field (volumes) which started as a IntId
    type, but then transitioned to a dictionary input with one of the keys matching that
    original IntId format.
    ::

        # The following represent a field that changed formats between releases:

        #  volumes_6 = _types.ListType(
        #    field=_types.ModelType(VagVolume),  # contains a 'volume' key that matches below.
        #    required=True,
        #    serialized_name='volumes',
        #    version_added=sf_version.SfVersion('6.1'),
        #    version_removed=sf_version.SfVersion('7.0')
        #  )

        #  volumes = _types.ListType(
        #    field=_types.IntIdType(),
        #    required=True,
        #    serialized_name='volumes',
        #    version_added=sf_version.SfVersion('7.0')
        #  )

        # It can now be implemented to support both formats, each returning an IntId type.

        volumes = _types.ListType(
            field=_types.IntIdTypeWithDictKeyConvert(api_constants.VOLUME_ID),
            required=True,
            serialized_name='volumes',
        )


    :param str convert_key: key in input dictionary which can pull the equivalent IntId value.
    """

    def __init__(self, convert_key=None, **kwargs):
        super(IntIdTypeWithDictKeyConvert, self).__init__(**kwargs)

        self.convert_key = convert_key
        if self.convert_key is None:
            raise ValueError('convert_key is required for IntIdTypeWithDictConvert objects.')

    def to_native(self, value, context=None):
        """Convert the native value into a form the IntIdType can consume.

        :param int|dict value: expected field ValueError
        :param schematics.datastructures.Context context: Context object.
        :rtype: int
        :return: field compatible value
        :raises schematics.exceptions.ConversionError: When value is not an int or dict
        """
        if isinstance(value, int):
            return super(IntIdTypeWithDictKeyConvert, self).to_native(value=value, context=context)
        elif isinstance(value, dict):
            if self.convert_key in value:
                return super(IntIdTypeWithDictKeyConvert, self).to_native(value=value[self.convert_key],
                                                                          context=context)
            else:
                raise exceptions.ConversionError(
                    "Value doesn't contain expected conversion key ({}): {}".format(
                        self.convert_key, value))

        raise exceptions.ConversionError('Value is not a int or dict; found {}'.format(type(value)))


class AttributeKeyedModelListType(BaseType):
    """Mutate data in an un-modelable object so it can be modeled.

    Use this type when the raw data has non-determistic top-level-keys, and top-level-values all
    deserialize into the same model type.

    For example, if you had the below structure where the number after Bond10G could be anything:
    .. note::

        {
            'Bond10G.1000':
            {
                '#default': False,
                'address': '10.116.178.154',
                'netmask': '255.255.255.0',
                'network': '10.116.178.0',
                'virtualNetworkTag': '1000'
            },
            'Bond10G.2002':
            {
                '#default': True,
                'address': '10.116.179.154',
                'netmask': '255.255.255.0',
                'network': '10.116.179.0',
                'virtualNetworkTag': '2002'
            }
        }

    After being deserialized, assuming attr_for_key=name, the structure, prior to model instantiation,
    would be:
    .. note::

        [
            {
                '#default': False,
                'address': '10.116.178.154',
                'name': 'Bond10G.1000',
                'netmask': '255.255.255.0',
                'network': '10.116.178.0',
                'virtualNetworkTag': '1000'
            },
            {
                '#default': True,
                'address': '10.116.179.154',
                'name': 'Bond10G.2002',
                'netmask': '255.255.255.0',
                'network': '10.116.179.0',
                'virtualNetworkTag': '2002'
            }
        ]

    Note: This does require you to add an optional attribute to the underlying model class to
    support the key being converted into an attribute (name in this example).

    :param sf_models.types.BaseType|property attr_for_key: Move top-level-key into this attribute.
    :param sf_models.Base model_class: Model class for top-level-values of raw data.
    """
    def __init__(self, attr_for_key, model_class, **kwargs):
        super(AttributeKeyedModelListType, self).__init__(**kwargs)
        self.model_class = model_class
        self.attr_for_key_name = attr_for_key.serialized_name or attr_for_key.name

    def to_primitive(self, value, context=None):
        """Convert AttributeKeyedModelListType into dictionary.

        :param list[sf_models.Base] value: List of models to convert back to dict.
        :param schematics.datastructures.Context context: Context object.
        :rtype: dict
        """
        output = {}
        for model in value:
            data = model.serialize()
            key_name = data[self.attr_for_key_name]
            del data[self.attr_for_key_name]
            output[key_name] = data
        return output

    def to_native(self, value, context=None):
        """Convert dictionary into AttributeKeyedModelListType.

        :param dict value: Dict to convert to AttributeKeyedModelListType.
        :param schematics.datastructures.Context context: Context object.
        :rtype: list[sf_models.Base]
        """
        value_copy = copy.deepcopy(value)
        model_list = []
        for k, v in six.iteritems(value_copy):
            v[self.attr_for_key_name] = k
            model_list.append(self.model_class(v))
        return model_list

    def validate(self, value, context=None):
        """Validate the model type and values of nested models.

        :param list[sf_models.Base] value: List of model_class objects to validate.
        :param schematics.datastructures.Context context: Context object.
        """
        for dm in value:
            assert isinstance(dm, self.model_class)
            dm.validate()


class NetworkInterfaceNameType(StringType):
    """Use for interfaces names such as mvip, svip, cipi, mipi, and sipi.

    Accepts strings with values in the enums.NetworkInterface or VLAN bonded interface names like
    Bond1G.3000 and Bond1G.3001.
    """

    def to_native(self, value, context=None):
        """Convert a raw string into a native NetworkInterface enum or NetworkInterfaceName object.

        :param str value: Value to convert to enum
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.enums.NetworkInterface|sf_models.NetworkInterfaceName
        """
        if (isinstance(value, _network_interface_name.NetworkInterfaceName) or
                isinstance(value, enums.NetworkInterface)):
            return value

        try:
            name = enums.NetworkInterface(value)
            return name
        except ValueError:
            pass

        try:
            name = _network_interface_name.NetworkInterfaceName(value)
        except ValueError:
            raise exceptions.ConversionError('Invalid string {}'.format(value))
        return name

    def to_primitive(self, value, context=None):
        """Convert a NetworkInterfaceName object into a primitive string.

        :param sf_models.NetworkInterfaceName value: Convert to primitive value
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        return value.value

    def validate(self, value, context=None):
        """Verify that value is a NetworkInterface enum or a NetworkInterfaceName object.

        :param str value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.enums.NetworkInterface|sf_models.NetworkInterfaceName
        """
        if (not isinstance(value, enums.NetworkInterface) and
                not isinstance(value, _network_interface_name.NetworkInterfaceName)):
            raise exceptions.ValidationError('{} is not a valid native type'.format(value))
        return value

    def _mock(self, context=None):
        """Randomly mock a NetworkInterface enum value.

        Bonded interface names with VLANs are not commonly used yet, so leaving them out for simplicity.

        :param schematics.datastructures.Context context: Context object
        :return: A native NetworkInterfaceName object
        :rtype: sf_models.enums.NetworkInterface
        """
        # noinspection PyTypeChecker
        values = list(enums.NetworkInterface)
        choice = random.randint(0, len(values) - 1)
        return values[choice]


class NodeTypeType(StringType):
    """Use for node types like SF3010, SFDOCKER, and dynamic customer facing node types.

    Accepts strings with values in the NodeType.STANDARD_TYPES or strings which are valid node types.

    Yes the name is silly, but we thought it best to stick with our convention for custom model types.
    """

    def to_native(self, value, context=None):
        """Convert a raw string into a native NodeType object.

        :param str value: Value to convert to object
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.NodeType
        """
        if isinstance(value, sf_models.NodeType):
            return value

        try:
            name = sf_models.NodeType(value)
            return name
        except ValueError:
            raise exceptions.ConversionError('Invalid string {}'.format(value))

    def to_primitive(self, value, context=None):
        """Convert a NodeType object into a primitive string.

        :param sf_models.NodeType value: Convert to primitive value
        :param schematics.datastructures.Context context: Context object
        :rtype: str
        """
        return value.value

    def validate(self, value, context=None):
        """Verify that value can be a NodeType object.

        :param str|sf_models.NodeType value: Value to validate
        :param schematics.datastructures.Context context: Context object
        :rtype: sf_models.NodeType
        """
        if isinstance(value, sf_models.NodeType):
            return value

        try:
            name = sf_models.NodeType(value)
            return name
        except ValueError:
            raise exceptions.ValidationError('{} is not a valid native type'.format(value))

    def _mock(self, context=None):
        """Randomly mock a NodeType.STANDARD_TYPES value.

        Dynamic customer facing node types are not commonly used yet, so leaving them out for simplicity.

        :param schematics.datastructures.Context context: Context object
        :return: A native NodeType object
        :rtype: sf_models.NodeType
        """
        # noinspection PyTypeChecker
        enum_values = list(sf_models.NodeType.STANDARD_TYPES)
        # witness nodes use a separate code path for many functions that cause UT failures if the test
        # wasn't intending to test that particular code path.
        enum_values.remove(sf_models.NodeType.STANDARD_TYPES.WITNESS)
        choice = random.randint(0, len(enum_values) - 1)
        return sf_models.NodeType(enum_values[choice].value)
