"""Test suite with features unique to testing ElementOS.

All test Element test suites should inherit from this class. It is needed to prevent the base camelot
TestCase from having a bunch of Element specific logic that teams like HCI, Platform, etc. do not need.
"""

import abc
import collections
import json
import re

import autologger
import sf_ssh
import six

import camelot
from camelot import groot
from sf_api import sf_version
from valence import (
    functional_models,
    models,
    monitors,
)

_log = autologger.get_sflogging(__name__)

INV_ELEM_MAP = {v: k for k, v in sf_version.SfVersion.ELEMENT_MAP.items()}
BUILD_INFO_PATTERN = r'"{}": "(.*?)",'
VERSION_PATTERN = r'"Version": "(\d+[\.\d+]*)"'
HCI_BUILD_INFO_PATTERN = r'{}=([^\s]+)'
HCI_VERSION_PATTERN = r'\bVersion=(\d+[\.\d+]*)'


@six.add_metaclass(abc.ABCMeta)
class ElementTestSuite(camelot.TestSuite):
    """Abstract base class for all ElementOS test suites.

    Contains some Element specific logic like storing faults in in the test results database.

    :param valence.models.SuiteResources resources: Resources data model for all test cases
    :param valence.models.SuiteConfig config: Suite configuration data model
    """

    # If True, turn off the backend monitor which looks for unexpected faults/events
    _skip_monitoring = False

    def __init__(self, resources, config):
        super(ElementTestSuite, self).__init__(resources, config)
        if not isinstance(self.resources, models.SuiteResources):
            raise TypeError('ElementTestSuite requires that resources are of type '
                            'valence.models.SuiteResources. Given: {}'.format(type(self.resources)))
        # Initialize global resources
        models.set_node_resources(self.resources.node_dms)
        models.set_cluster_resources(self.resources.cluster_dms)
        self._backend_monitor = None
        if self.resources.cluster_dms:
            self._backend_monitor = monitors.BackendMonitor(resources)
            self._backend_monitor.set_as_global_monitor()
        self._last_fault_map = {}

    @classmethod
    def from_input_files(cls, resources_file, config_file=None):
        """Convenience method for instantiating a suite by providing file paths instead of data models.

        :param str resources_file: Path to a suite resources file
        :param str config_file: Path to a suite config file. If not provided, a default config will
            be used.
        :return: An instantiated suite
        :rtype: valence.camelot.ElementTestSuite
        """
        with open(resources_file) as r:
            resource_data = json.load(r)
            resources = models.SuiteResources(resource_data)
        if config_file is None:
            config = models.SuiteConfig()
        else:
            with open(config_file) as c:
                config_data = json.load(c)
                config = models.SuiteConfig(config_data)
        return cls(resources, config)

    def create_final_report(self):
        """Create the final report after the suite has finished.

        This is overridden so we can change the status if there are monitor failures.
        """
        if self.result.monitor_failures and self.status is not camelot.ResultStatus.error:
            self.result.status = self.status = camelot.ResultStatus.failed
        super(ElementTestSuite, self).create_final_report()
        if self.result.monitor_failures:
            _log.info('Monitor Failures: \n{}'.format(
                '\n'.join(str(f) for f in self.result.monitor_failures)))

    def execute(self):
        """Extend execute with some element specific setup.

        :returns: A list of test results
        :rtype: tuple[valence.camelot.TestRun]
        """
        if not self._skip_remote_reporting:
            self._update_build_info()
        return super(ElementTestSuite, self).execute()

    def _update_build_info(self):
        """Update the node and cluster build info resources objects."""
        if not isinstance(self.resources, models.SuiteResources):
            return
        cluster_build_info = collections.defaultdict(str)
        for node in self.resources.nodes:
            if not node.config.build:
                node.config.build = models.Build.create()
            self._update_node_build_info(node, cluster_build_info)

        for cluster in self.resources.clusters:
            self._update_cluster_build_info(cluster, cluster_build_info)

    def _pre_test_case_execute(self, run):
        """Method that is run before a test case has executed.

        Override this if you need to perform some operation before every test case. Just make sure to
        call the 'super' method so the normal requisite operations are performed.

        :param valence.camelot.TestRun run: TestRun object
        """
        self._update_always_fail_lists(run)

    def _post_test_case_execute(self, run):
        self._post_faults_to_groot(run)
        self._check_for_monitor_failures(run)

    def _post_faults_to_groot(self, run):
        cluster_dms = self.resources.cluster_dms
        if self._skip_remote_reporting is False:
            # I'm not sure if this is worth it, we can't currently tell which which cluster had which
            # fault after the fact.  Should this just capture faults from cluster[0]?
            for cluster_dm in cluster_dms:
                cluster_fm = functional_models.Cluster(cluster_dm)
                try:
                    faults = cluster_fm.list_cluster_faults()
                except Exception as ex:  # pylint: disable=broad-except
                    _log.debug(
                        'Attempted to save faults, is the cluster up? Error observed={}'.format(ex))
                    break
                if self._last_fault_map.get(cluster_dm.unique_id):
                    new_faults = [f.serialize() for f in faults if
                                  f.date > self._last_fault_map[cluster_dm.unique_id]]
                else:
                    new_faults = [f.serialize() for f in faults]
                if new_faults:
                    groot.post_to_groot(
                        groot.FAULTS_ENDPOINT + '/{}'.format(
                            run.case.current_result.instance_id), new_faults)
                    self._last_fault_map[cluster_dm.unique_id] = max([f.date for f in faults])

    def _check_for_monitor_failures(self, run):
        _log.info('Checking for unexpected faults and events in the backend monitor')
        if not self._backend_monitor or self._skip_monitoring:
            return
        try:
            self._backend_monitor.run_iteration()
        except monitors.MonitorCriticalError:
            # If a critical error is observed, stop execution of the test suite immediately
            _log.warn('A critical error occurred. Suite teardown will not run.')
            self.config.run_suite_teardown = False
            raise
        except Exception as e:  # pylint: disable=broad-except
            msg = 'Unexpected error occurred during checking for monitor failures: {}'.format(e)
            _log.debug(msg)
            self._backend_monitor._store_failure(  # pylint: disable=protected-access
                msg, run.result.end_time)
        result = run.result
        failures = self._backend_monitor.get_failures(start_time=result.start_time,
                                                      end_time=result.end_time)
        self.result.monitor_failures.extend(failures)
        # Restore the default always fail lists
        self._backend_monitor.set_always_fail_faults()
        self._backend_monitor.set_always_fail_events()

    @staticmethod
    def _update_node_build_info(node, cluster_build_info):
        """SSH into the nodes and get the latest build info then update the resources object.

        :param valence.models.NodeSuiteResource node: The node resource object
        :param dict cluster_build_info: A dictionary to update the cluster build info with the nodes
            build infos
        """
        build_info_pattern = BUILD_INFO_PATTERN
        version_pattern = VERSION_PATTERN
        try:
            with sf_ssh.Ssh(node.model.mip) as ssh_conn:
                result = ssh_conn.run_command('cat /sf/rtfi/buildinfo.json')

            if not result.stdout:  # On HCI storage nodes there is no buildinfo.json
                with sf_ssh.Ssh(node.model.mip) as ssh_conn:
                    result = ssh_conn.run_command('/sf/bin/sfapp -Version', use_sudo=False,
                                                  raise_on_failure=True)
                build_info_pattern = HCI_BUILD_INFO_PATTERN
                version_pattern = HCI_VERSION_PATTERN

        except Exception as error:  # pylint: disable=broad-except
            _log.warn('Ssh raised an exception trying to gather build info: {}'.format(error))
            return

        c_elem_pat = re.compile(build_info_pattern.format('Element'))
        element = c_elem_pat.search(str(result.stdout))

        if element:
            node.config.build.element_id = INV_ELEM_MAP[element.group(1)]
            cluster_build_info['element_id'] = INV_ELEM_MAP[element.group(1)]

        c_repo_pat = re.compile(build_info_pattern.format('Repository'))
        repo = c_repo_pat.search(str(result.stdout))
        if repo:
            node.config.build.repository_name = repo.group(1)
            cluster_build_info['repository_name'] = repo.group(1)

        c_release_pat = re.compile(build_info_pattern.format('Release'))
        release = c_release_pat.search(str(result.stdout))
        if release:
            node.config.build.release = release.group(1)
            cluster_build_info['release'] = release.group(1)

        c_ver_pat = re.compile(version_pattern)
        version = c_ver_pat.search(str(result.stdout))
        if version:
            node.config.build.version = version.group(1)
            cluster_build_info['version'] = version.group(1)

    @staticmethod
    def _update_cluster_build_info(cluster, cluster_build_info):
        """Use the build info from the nodes to update the cluster resource object.

        :param valence.models.ClusterSuiteResource cluster: The cluster resource object
        :param dict cluster_build_info: A dictionary to update the cluster build info with the nodes
            build infos
        """
        if not cluster.config.build:
            cluster.config.build = models.Build.create()
        if 'element_id' in cluster_build_info:
            cluster.config.build.element_id = cluster_build_info['element_id']
        if 'repository_name' in cluster_build_info:
            cluster.config.build.repository_name = cluster_build_info['repository_name']
        if 'release' in cluster_build_info:
            cluster.config.build.release = cluster_build_info['release']
        if 'version' in cluster_build_info:
            cluster.config.build.version = cluster_build_info['version']

    def _update_always_fail_lists(self, test_run):
        """Updates the global always fail fault and events lists for each test case.

        :param valence.camelot.TestRun test_run: test case object
        """
        test_case = test_run.case
        if self._backend_monitor:
            for event in getattr(test_case, 'monitor_fail_on_events', []):
                self._backend_monitor.add_always_fail_event(event)
            for fault in getattr(test_case, 'monitor_fail_on_faults', []):
                self._backend_monitor.add_always_fail_fault(fault)
            for event in getattr(test_case, 'monitor_ignore_events', []):
                self._backend_monitor.remove_always_fail_event(event)
            for fault in getattr(test_case, 'monitor_ignore_faults', []):
                self._backend_monitor.remove_always_fail_fault(fault)
