import os

import mock
import pytest
import requests
import sf_ssh

from camelot import groot
from valence import (
    camelot,
    functional_models,
    models,
    monitors,
    unittest_utils,
)
from valence.unittest_utils import (
    helpers,
    suites,
)
from valence.utils import time_utils

DATA_DIR = unittest_utils.get_data_dir(__file__)


class TestElementSuite(helpers.MockHistoryUnitTest):
    """Test default test suite parameter values."""

    @classmethod
    def setUpClass(cls):
        cls.config = models.SuiteConfig()
        cls.resources = models.SuiteResources()
        cls.resources_file = os.path.join(DATA_DIR, 'resources.json')
        cls.config_file = os.path.join(DATA_DIR, 'suite_config.json')

    def test_from_input_files_resources_only(self):
        suite = suites.PassingSuite.from_input_files(self.resources_file)
        assert isinstance(suite.resources, models.SuiteResources)
        assert isinstance(suite.config, models.SuiteConfig)

    def test_from_input_files_resources_and_config(self):
        suite = suites.PassingSuite.from_input_files(self.resources_file, self.config_file)
        assert isinstance(suite.resources, models.SuiteResources)
        assert isinstance(suite.config, models.SuiteConfig)

    @mock.patch.object(monitors.BackendMonitor, 'run_iteration', lambda x: None)
    @mock.patch.object(monitors.BackendMonitor, 'get_failures')
    def test_monitor_failures(self, mock_get_failures):
        mock_get_failures.return_value = [monitors.MonitorFailure('fail', time_utils.get_now_utc())]
        suite = suites.PassingSuite.from_input_files(self.resources_file)
        suite._skip_monitoring = False
        suite.execute()
        assert suite.status is camelot.ResultStatus.failed
        report = suite.get_final_report()
        assert report['monitor_failures']

    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    @mock.patch.object(groot, 'post_to_groot')
    def test_groot_post(self, mock_groot, mock_faults):
        """Test method _post_faults_to_groot method."""
        suite = suites.PassingSuite(self.resources, self.config)
        suite._skip_remote_reporting = False
        suite.resources.clusters = models.ClusterSuiteResource.get_mock_object_list(2)
        test_faults = models.ClusterFault.get_mock_object_list(2)
        mock_faults.return_value = test_faults
        suite.create_test_runs()
        suite.test_runs[0].case.execute()
        suite._post_faults_to_groot(suite.test_runs[0])
        suite.last_fault_map = {0: 1}
        suite._post_faults_to_groot(suite.test_runs[0])
        assert mock_groot.called

    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    @mock.patch.object(groot, 'post_to_groot', lambda x: None)
    def test_groot_fail(self, mock_faults):
        """Test method _post_faults_to_groot method."""
        suite = suites.PassingSuite(self.resources, self.config)
        suite._skip_remote_reporting = False
        suite.resources.clusters = models.ClusterSuiteResource.get_mock_object_list(2)
        mock_faults.side_effect = iter([requests.ConnectionError, requests.HTTPError])
        suite.create_test_runs()
        suite.test_runs[0].case.execute()
        suite._post_faults_to_groot(suite.test_runs[0])
        assert mock_faults.called_once()

    @mock.patch.object(camelot.ElementTestSuite, '_post_faults_to_groot', lambda x: None)
    @mock.patch.object(camelot.ElementTestSuite, '_update_node_build_info')
    @mock.patch.object(camelot.ElementTestSuite, '_update_cluster_build_info')
    def test_update_build_info(self, mock_update_cluster_build_info, mock_update_node_build_info):
        self.resources.nodes = models.NodeSuiteResource.get_mock_object_list(1)
        self.resources.nodes[0].config = models.NodeSuiteConfig.get_mock_object()
        self.resources.nodes[0].config.build = None
        self.resources.clusters = models.ClusterSuiteResource.get_mock_object_list(1)
        self.resources.clusters[0].config = models.ClusterSuiteConfig.get_mock_object()
        self.resources.clusters[0].config.build = None
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._skip_remote_reporting = False
        passing_suite.execute()
        mock_update_cluster_build_info.assert_called_once()
        mock_update_node_build_info.assert_called_once()

    def test_update_build_info_bad_resources(self):
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite.resources = mock.Mock()
        passing_suite._skip_remote_reporting = False
        passing_suite.execute()

    @mock.patch.object(camelot.ElementTestSuite, '_post_faults_to_groot', lambda x: None)
    @mock.patch.object(camelot.ElementTestSuite, '_update_build_info', side_effect=ValueError)
    def test_update_build_info_exception(self, mock_update_info):
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._skip_remote_reporting = False
        with pytest.raises(ValueError):
            passing_suite.execute()

    def test_instantiate_suite_bad_resources(self):
        with pytest.raises(TypeError):
            suites.PassingSuite(mock.Mock(), self.config)

    def test_update_cluster_build_info(self):
        self.resources.clusters = models.ClusterSuiteResource.get_mock_object_list(1)
        self.resources.clusters[0].config = models.ClusterSuiteConfig.get_mock_object()
        self.resources.clusters[0].config.build = models.Build.get_mock_object()
        self.resources.clusters[0].config.build.element_id = 11
        self.resources.clusters[0].config.build.repository_name = 'neon'
        self.resources.clusters[0].config.build.version = '10.0.0.284'
        self.resources.clusters[0].config.build.release = 'sodium'

        cluster_build_info = {'element_id': 10,
                              'version': '11.0.0.1',
                              'release': 'neon',
                              'repository_name': 'sodium'}
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._update_cluster_build_info(self.resources.clusters[0], cluster_build_info)

        assert self.resources.clusters[0].config.build.element_id == 10
        assert self.resources.clusters[0].config.build.repository_name == 'sodium'
        assert self.resources.clusters[0].config.build.version == '11.0.0.1'
        assert self.resources.clusters[0].config.build.release == 'neon'

    def test_update_cluster_build_no_build(self):
        self.resources.clusters = models.ClusterSuiteResource.get_mock_object_list(1)
        self.resources.clusters[0].config = models.ClusterSuiteConfig.get_mock_object()
        self.resources.clusters[0].config.build = None

        cluster_build_info = {'element_id': 10,
                              'version': '11.0.0.1',
                              'release': 'neon',
                              'repository_name': 'sodium'}
        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._update_cluster_build_info(self.resources.clusters[0], cluster_build_info)

        assert self.resources.clusters[0].config.build.element_id == 10
        assert self.resources.clusters[0].config.build.repository_name == 'sodium'
        assert self.resources.clusters[0].config.build.version == '11.0.0.1'
        assert self.resources.clusters[0].config.build.release == 'neon'

    @mock.patch('sf_ssh.Ssh')
    def test_update_node_build_info(self, mock_ssh):
        self.resources.nodes = models.NodeSuiteResource.get_mock_object_list(1)
        self.resources.nodes[0].config = models.NodeSuiteConfig.get_mock_object()
        self.resources.nodes[0].config.build = models.Build.get_mock_object()
        self.resources.nodes[0].config.build.element_id = 6
        self.resources.nodes[0].config.build.repository_name = 'oxygen'
        self.resources.nodes[0].config.build.version = '13.0.0.284'
        self.resources.nodes[0].config.build.release = 'aluminium'

        stdout = (b'{\n   "Element": "sodium",\n  "Release": "sodium",\n  "Version": "11.0.0.100",\n '
                  b' "Repository": "sodium",\n}\n')
        mock_ssh.return_value.__enter__.return_value.run_command.return_value = \
            sf_ssh.ssh.RunCommandOutput(stdout=stdout, stderr='', return_code=0)

        cluster_build_info = {'element_id': 10,
                              'version': '11.0.0.1',
                              'release': 'neon',
                              'repository_name': 'sulfur'}

        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._update_node_build_info(self.resources.nodes[0], cluster_build_info)

        assert self.resources.nodes[0].config.build.element_id == 11
        assert self.resources.nodes[0].config.build.repository_name == 'sodium'
        assert self.resources.nodes[0].config.build.version == '11.0.0.100'
        assert self.resources.nodes[0].config.build.release == 'sodium'
        assert cluster_build_info['element_id'] == 11
        assert cluster_build_info['version'] == '11.0.0.100'
        assert cluster_build_info['release'] == 'sodium'
        assert cluster_build_info['repository_name'] == 'sodium'

    @mock.patch('sf_ssh.Ssh')
    def test_update_node_build_info_hci(self, mock_ssh):
        self.resources.nodes = models.NodeSuiteResource.get_mock_object_list(1)
        self.resources.nodes[0].config = models.NodeSuiteConfig.get_mock_object()
        self.resources.nodes[0].config.build = models.Build.get_mock_object()
        self.resources.nodes[0].config.build.element_id = 6
        self.resources.nodes[0].config.build.repository_name = 'oxygen'
        self.resources.nodes[0].config.build.version = '13.0.0.284'
        self.resources.nodes[0].config.build.release = 'aluminium'

        stdout = ('sfapp BuildType=Release Element=sodium Release=sodium '
                  'ReleaseShort=sodium-availability-zones Version=11.0.0.100 EmberVersion=2.0.71 '
                  'Repository=sodium VCS=hg Revision=c64a000f5142 Options=timing,timing '
                  'BuildDate=2018-08-02T18:14:40UTC md5=a68ecae687206937b4f1fda310898022')
        mock_ssh.return_value.__enter__.return_value.run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout=u'', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout=stdout, stderr='', return_code=0)]

        cluster_build_info = {'element_id': 10,
                              'version': '11.0.0.1',
                              'release': 'neon',
                              'repository_name': 'sulfur'}

        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._update_node_build_info(self.resources.nodes[0], cluster_build_info)

        assert self.resources.nodes[0].config.build.element_id == 11
        assert self.resources.nodes[0].config.build.repository_name == 'sodium'
        assert self.resources.nodes[0].config.build.version == '11.0.0.100'
        assert self.resources.nodes[0].config.build.release == 'sodium'
        assert cluster_build_info['element_id'] == 11
        assert cluster_build_info['version'] == '11.0.0.100'
        assert cluster_build_info['release'] == 'sodium'
        assert cluster_build_info['repository_name'] == 'sodium'

    @mock.patch('sf_ssh.Ssh')
    def test_update_node_build_info(self, mock_ssh):
        self.resources.nodes = models.NodeSuiteResource.get_mock_object_list(1)
        self.resources.nodes[0].config = models.NodeSuiteConfig.get_mock_object()
        self.resources.nodes[0].config.build = models.Build.get_mock_object()
        self.resources.nodes[0].config.build.element_id = 6
        self.resources.nodes[0].config.build.repository_name = 'oxygen'
        self.resources.nodes[0].config.build.version = '13.0.0.284'
        self.resources.nodes[0].config.build.release = 'aluminium'

        cluster_build_info = {}
        mock_ssh.return_value.__enter__.return_value.run_command.side_effect = sf_ssh.SshError

        passing_suite = suites.PassingSuite(self.resources, self.config)
        passing_suite._update_node_build_info(self.resources.nodes[0], cluster_build_info)

        assert self.resources.nodes[0].config.build.element_id == 6
        assert self.resources.nodes[0].config.build.repository_name == 'oxygen'
        assert self.resources.nodes[0].config.build.version == '13.0.0.284'
        assert self.resources.nodes[0].config.build.release == 'aluminium'

    @mock.patch.object(monitors.BackendMonitor, 'run_iteration')
    def test_check_for_monitor_failures_fails(self, mock_run_iteration):
        mock_run_iteration.side_effect = Exception
        suite = suites.PassingSuite.from_input_files(self.resources_file)
        suite._skip_monitoring = False
        suite.execute()
        assert suite.status is camelot.ResultStatus.failed
        report = suite.get_final_report()
        assert report['monitor_failures']

    @mock.patch.object(camelot.TestSuite, 'teardown')
    @mock.patch.object(monitors.BackendMonitor, 'run_iteration')
    def test_monitor_critical_error(self, mock_run_iteration, mock_teardown):
        mock_run_iteration.side_effect = monitors.MonitorCriticalError('Ya basic')
        suite = suites.PassingSuite.from_input_files(self.resources_file)
        suite._skip_monitoring = False
        suite.execute()
        mock_teardown.assert_not_called()
        assert suite.status is camelot.ResultStatus.error

    def test_element_suite_fail_ignore_lists(self):
        suite = suites.PassingElementSuite.from_input_files(self.resources_file)
        suite.execute()
        assert suite.status is camelot.ResultStatus.passed
