"""Valence Integration test cases."""
import time

import autologger

from valence import (
    camelot,
    enums,
    functional_models,
    selectors,
    vl_components,
    vl_helpers,
)
from valence.vl_components import vdbench

_log = autologger.get_sflogging(__name__)

IGNORE_DETAILS_REGEXES = [
    # Faults related to old hardware
    'NVRAM device error={health=Low.*',
    'NVRAM device warning={excessiveCapacitorVoltage.*',
    'NVRAM device warning={excessiveCurrent.*',
    'NVRAM device warning={fanInletAmbientTemperature.*',
    'NVRAM device warning={hardwareRevision.*',
    'File System /mnt/pendingDirtyBlocks has *',
    '.*low wear reserves.',
    'MPTSAS_FIRMWARE_VERSION.*',
]


class ValenceIntegration(camelot.ElementTestCase):
    """Test that basic IO and high risk parts of the code work.

    :param valence.models.ResourceCluster rs_cluster_dm: cluster data model
    :param list[valence.models.ResourceNode] rs_node_dms: node data models
    :param list[valence.models.ResourceClient] rs_client_dms: Client data models
    :param sf_artifacts.ArtifactManager artifact_manager: Manager that will be used to store support
        bundles.
    """

    def __init__(self, rs_cluster_dm, rs_node_dms, rs_client_dms, artifact_manager):
        super(ValenceIntegration, self).__init__()
        self.artifact_manager = artifact_manager
        self.rs_client_dms = rs_client_dms
        self.rs_cluster_dm = rs_cluster_dm
        self.rs_node_dms = rs_node_dms
        self.cluster_fm = functional_models.Cluster(rs_cluster_dm, authenticate=False)
        self.client_fms = [functional_models.get_client_from_dm(client) for client in rs_client_dms]
        self.node_fms = [
            functional_models.Node(node, rs_cluster_dm, authenticate=False) for node in rs_node_dms
        ]

        self.vdb_profile = vdbench.StandardVdbProfile.from_clients(
            tag='integration',
            cluster_dm=rs_cluster_dm,
            client_dms=rs_client_dms,
        )
        self.vdb_profile.sanitize_configs(node_dms=self.rs_node_dms)
        self.vdb_env_cmt = self.vdb_profile.get_environment(master_client_dm=rs_client_dms[0])
        self.vdb_run_cmt = vdbench.ManageRunVdbs.from_environment_cmt(
            environments_cmt=self.vdb_env_cmt,
            validation_config=vdbench.ValidationConfig.on_reads,
            journaling_config=vdbench.JournalingConfig.none
        )

        self.rm_drives_cmt = vl_components.RemoveDrives(
            cluster_dm=rs_cluster_dm,
            drive_selector=selectors.DriveSelector(
                cluster_dm=rs_cluster_dm,
                num=1,
                types=[enums.DriveType.block]
            )
        )

    def _preconditions(self):
        """Assert resource requirements."""
        assert len(self.rs_node_dms) >= 3

    @property
    def monitor_ignore_faults(self):
        """If we have vNodes ignore the file_system_capacity_low fault.

        :rtype: list[valence.enums.ClusterFaultCode]
        """
        ignore_faults = []
        if any(node_dm.model.is_virtual for node_dm in self.rs_node_dms):
            ignore_faults.append(enums.ClusterFaultCode.file_system_capacity_low)
        return ignore_faults

    def _setup(self):
        """Authenticate cluster and node."""
        self.cluster_fm.authenticate()
        for node_fm in self.node_fms:
            node_fm.authenticate()

    def _test(self):
        """Make an account and some volumes then run some basic tests."""
        self._log_step('Create accounts/volumes and establish client connections for running IO')
        self.vdb_env_cmt.apply()

        self._log_step('Start Vdbench IO')
        self.vdb_run_cmt.apply_async()

        self._log_step('Remove a random block drive')
        self.rm_drives_cmt.apply()

        self._log_step('Wait 30 seconds then assert Vdbench is running')
        time.sleep(30)
        self.vdb_run_cmt.assert_valid(is_running=True)

        self._log_step('Stop Vdbench IO')
        self.vdb_run_cmt.set_test_identifier(str(self.current_result.instance_id))
        self.vdb_run_cmt.finish(kill=True)
        self.vdb_run_cmt.ensure_ready()

        self._log_step('Add back removed drive')
        self.rm_drives_cmt.reset()

        self._log_step('Wait for cluster faults to clear')
        vl_helpers.wait_for_cluster_fault_resolution(
            cluster_fm=self.cluster_fm,
            ignore_details_regexes=IGNORE_DETAILS_REGEXES,
            timeout=60 * 30
        )

    def _teardown(self):
        """Teardown the test and restore cluster to a clean state."""
        # Get support bundles
        vl_helpers.pull_support_bundles(
            artifact_manager=self.artifact_manager,
            node_fms=self.node_fms,
            client_fms=self.client_fms
        )

        # Get vdbench artifacts
        self.vdb_run_cmt.store_artifacts()

        # Ensure ready for all components
        self.vdb_run_cmt.ensure_ready()
        self.vdb_env_cmt.ensure_ready()
        self.rm_drives_cmt.ensure_ready()
