"""Valence Sanity testing cases."""
import time

import autologger

from valence import (
    camelot,
    components,
    enums,
    functional_models,
    models,
    selectors,
    vl_components,
    vl_helpers,
)
from valence.vl_components import vdbench

_log = autologger.get_sflogging(__name__)

IGNORE_DETAILS_REGEXES = [
    # Faults related to old hardware
    'NVRAM device error={health=Low.*',
    'NVRAM device warning={excessiveCapacitorVoltage.*',
    'NVRAM device warning={excessiveCurrent.*',
    'NVRAM device warning={fanInletAmbientTemperature.*',
    'NVRAM device warning={hardwareRevision.*',
    'File System /mnt/pendingDirtyBlocks has *',
    '.*low wear reserves.',
]


class SanityCommitSmoke(camelot.ElementTestCase):
    """Creates an account and 20x50Gb volumes.

    :param valence.models.ResourceCluster rs_cluster_dm: cluster data model
    :param list[valence.models.ResourceNode] rs_node_dms: node data models
    :param list[valence.models.ResourceClient] rs_client_dms: Client data models
    :param sf_artifacts.ArtifactManager artifact_manager: Manager that will be used to store support
        bundles.
    """

    def __init__(self, rs_cluster_dm, rs_node_dms, rs_client_dms, artifact_manager):
        super(SanityCommitSmoke, self).__init__()
        self.artifact_manager = artifact_manager
        self.rs_client_dms = rs_client_dms
        self.rs_cluster_dm = rs_cluster_dm
        self.rs_node_dms = rs_node_dms
        self.cluster_fm = functional_models.Cluster(rs_cluster_dm, authenticate=False)
        self.client_fms = [functional_models.get_client_from_dm(client) for client in rs_client_dms]
        self.node_fms = [
            functional_models.Node(node, rs_cluster_dm, authenticate=False) for node in rs_node_dms
        ]

        self.vdb_profile = vdbench.StandardVdbProfile.from_clients(
            tag='sanity',
            cluster_dm=rs_cluster_dm,
            client_dms=rs_client_dms[1:]
        )
        self.vdb_profile.sanitize_configs(node_dms=self.rs_node_dms)
        self.vdb_env_cmt = self.vdb_profile.get_environment(master_client_dm=rs_client_dms[0])
        self.vdb_run_cmt = vdbench.ManageRunVdbs.from_environment_cmt(
            environments_cmt=self.vdb_env_cmt,
            validation_config=vdbench.ValidationConfig.on_reads,
            journaling_config=vdbench.JournalingConfig.none
        )

        self.enable_ear_cmt = vl_components.EnableEar(rs_cluster_dm)
        self.rm_drives_cmt = vl_components.RemoveDrives(
            cluster_dm=rs_cluster_dm,
            drive_selector=selectors.DriveSelector(
                cluster_dm=rs_cluster_dm,
                num=3,
                types=[enums.DriveType.block]
            )
        )
        self.down_interface_cm_1g_cmt = vl_components.DownInterfacesOnNodes(
            interfaces=[enums.NetworkInterface.bond_1g],
            cluster_dm=self.rs_cluster_dm,
            node_selector=selectors.ActiveNodeSelector(self.rs_cluster_dm, cluster_master=True)
        )
        self.down_interface_eth_cmt = vl_components.DownInterfacesOnNodes(
            interfaces=[enums.NetworkInterface.eth0],
            cluster_dm=self.cluster_fm.cluster_dm,
            node_selector=selectors.ActiveNodeSelector(self.rs_cluster_dm, num=1)
        )

    def _preconditions(self):
        """Assert resource requirements."""
        assert len(self.rs_node_dms) >= 4

    @property
    def monitor_ignore_faults(self):
        """If we have vNodes ignore the file_system_capacity_low fault.

        :rtype: list[valence.enums.ClusterFaultCode]
        """
        ignore_faults = []
        if any(node_dm.model.is_virtual for node_dm in self.rs_node_dms):
            ignore_faults.append(enums.ClusterFaultCode.file_system_capacity_low)
        return ignore_faults

    def _setup(self):
        """Authenticate cluster and node."""
        self.cluster_fm.authenticate()
        for node_fm in self.node_fms:
            node_fm.authenticate()

    def _test(self):
        """Make an account and some volumes then run some basic tests."""
        self._log_step('Create accounts/volumes and establish client connections for running IO')
        self.vdb_env_cmt.apply()

        self._log_step('Disable OTPW')
        for node_fm in self.node_fms:
            node_fm.disable_otpw()

        self._log_step('Enable encryption at rest if allowed')
        if any(node_dm.model.is_virtual or node_dm.model.is_non_encrypting for
               node_dm in self.rs_node_dms):
            _log.info('Nodes are virtual or non-encrypting, do not Enable EAR')
        else:
            self.enable_ear_cmt.apply()

        self._log_step('Test hardware config')
        vl_helpers.test_hardware_config(self.node_fms)

        self._log_step('Start Vdbench IO')
        self.vdb_run_cmt.apply_async()

        self._log_step('Remove 3 random block drives')
        self.rm_drives_cmt.apply()

        self._log_step('Wait 2 minutes then assert Vdbench is running')
        time.sleep(2 * 60)
        self.vdb_run_cmt.assert_valid(is_running=True)

        self._log_step('Stop Vdbench IO')
        self.vdb_run_cmt.set_test_identifier(str(self.current_result.instance_id))
        self.vdb_run_cmt.finish(kill=True)
        self.vdb_run_cmt.ensure_ready()

        self._log_step('Add back removed drives')
        self.rm_drives_cmt.reset()

        self._log_step('Stop SolidFire on all nodes')
        for node_fm in self.node_fms:
            node_fm.stop_binary(enums.NodeBinary.solidfire)

        self._log_step('Sleep for 10 seconds')
        time.sleep(10)

        self._log_step('Start SolidFire on all nodes')
        for node_fm in self.node_fms:
            node_fm.start_binary(enums.NodeBinary.solidfire)

        self._log_step('Sleep for 180 seconds')
        time.sleep(180)

        self._log_step('Wait for cluster faults to clear')
        vl_helpers.wait_for_cluster_fault_resolution(
            cluster_fm=self.cluster_fm,
            ignore_details_regexes=IGNORE_DETAILS_REGEXES,
            timeout=60 * 30
        )

        self._log_step('Start Vdbench IO')
        self.vdb_run_cmt.apply_async()

        self._log_step('Check for cores on nodes')
        vl_helpers.check_for_core_and_crash_dumps(self.node_fms)

        self._log_step("Set CM's Bond1G down")
        cluster_master_id = self.cluster_fm.get_cluster_info().mvip_node_id
        self.down_interface_cm_1g_cmt.apply()

        self._log_step('Sleep for 2 minutes to allow CM to change')
        time.sleep(2 * 60)

        self._log_step('Assert CM changed')
        new_cluster_master_id = self.cluster_fm.get_cluster_info().mvip_node_id
        assert new_cluster_master_id != cluster_master_id, 'Cluster master did not change'

        self._log_step('Wait 1 minute then assert Vdbench is running')
        time.sleep(1 * 60)
        self.vdb_run_cmt.assert_valid(is_running=True)

        self._log_step('Stop Vdbench IO')
        self.vdb_run_cmt.finish(kill=True)
        self.vdb_run_cmt.ensure_ready()

        self._log_step("Bring CM's Bond1G back up")
        self.down_interface_cm_1g_cmt.reset()

        self._log_step('Wait for cluster faults to clear')
        vl_helpers.wait_for_cluster_fault_resolution(
            cluster_fm=self.cluster_fm,
            ignore_details_regexes=IGNORE_DETAILS_REGEXES,
            timeout=60 * 30
        )

        self._log_step('Disable encryption at rest')
        # self.enable_ear_cmt.ensure_ready()

        self._log_step('Start Vdbench IO')
        self.vdb_run_cmt.apply_async()

        self._log_step('Down eth0 on random node')
        self.down_interface_eth_cmt.apply()

        self._log_step('Up eth0 on node')
        self.down_interface_eth_cmt.reset()

        self._log_step('Wait 2 minutes then assert Vdbench is running')
        time.sleep(2 * 60)
        self.vdb_run_cmt.assert_valid(is_running=True)

        self._log_step('Stop Vdbench IO')
        self.vdb_run_cmt.finish(kill=True)
        self.vdb_run_cmt.ensure_ready()

        self._log_step('Check for cores on nodes')
        vl_helpers.check_for_core_and_crash_dumps(self.node_fms)

    def _teardown(self):
        """Teardown the test and restore cluster to a clean state."""
        # Get support bundles
        vl_helpers.pull_support_bundles(
            artifact_manager=self.artifact_manager,
            node_fms=self.node_fms,
            client_fms=self.client_fms
        )

        # Get vdbench artifacts
        self.vdb_run_cmt.store_artifacts()

        # Ensure ready for all components
        self.vdb_run_cmt.ensure_ready()
        self.vdb_env_cmt.ensure_ready()
        self.rm_drives_cmt.ensure_ready()
        # self.enable_ear_cmt.ensure_ready()


class SanityClones(camelot.ElementTestCase):
    """Clones sanity test case.

    :param valence.models.ResourceCluster rs_cluster_dm: cluster data model
    :param list[valence.models.ResourceNode] rs_node_dms: node data models
    :param list[valence.models.ResourceClient] rs_client_dms: Client data model
    :param sf_artifacts.ArtifactManager artifact_manager: Manager that will be used to store support
        bundles.
    """
    SANITY_MAX_IO_RATE = [
        vdbench.IoPeriodConfig(
            io_rate_config=vdbench.IoRateCurve(curve=(75,)),
            interval=30,
            duration=24 * 60 * 60
        )
    ]

    def __init__(self, rs_cluster_dm, rs_node_dms, rs_client_dms, artifact_manager):
        super(SanityClones, self).__init__()
        self.rs_cluster_dm = rs_cluster_dm
        self.rs_node_dms = rs_node_dms
        self.rs_client_dms = rs_client_dms
        self.artifact_manager = artifact_manager
        self.cluster_fm = functional_models.Cluster(rs_cluster_dm, authenticate=False)
        self.client_fms = [functional_models.get_client_from_dm(c) for c in self.rs_client_dms]
        self.node_fms = [
            functional_models.Node(node, rs_cluster_dm, authenticate=False) for node in rs_node_dms
        ]
        self.create_accounts_cmt = vl_components.CreateAccounts(
            tag='sanity',
            cluster_dm=rs_cluster_dm
        )
        self.create_volumes_cmt_primary = vl_components.CreateMultipleVolumes(
            tag='sanity-primary',
            cluster_dm=rs_cluster_dm,
            vol_count=2 * len(self.node_fms),
            partial_volume_dm=models.Volume.create(
                total_size=models.DataSize(10, unit=enums.DataUnit.gigabytes),
                qos=models.Qos.create(min_iops=100, max_iops=100000, burst_iops=100000)
            ),
            accounts_cmt=self.create_accounts_cmt,
        )
        self.clone_volumes_cmt = vl_components.CloneVolumes(
            tag='sanity-clones',
            cluster_dm=rs_cluster_dm,
            volumes_cmt=self.create_volumes_cmt_primary,
            accounts_cmt=self.create_accounts_cmt
        )

        vdbench.StandardVdbProfile.DEFAULT_PERIOD_CONFIGS = self.SANITY_MAX_IO_RATE
        self.vdb_profile_primary = vdbench.StandardVdbProfile.from_volumes(
            tag='sanity-primary',
            cluster_dm=self.rs_cluster_dm,
            client_dms=self.rs_client_dms[:1],
            volumes_cmts=[self.create_volumes_cmt_primary]
        )
        self.vdb_profile_clones = vdbench.StandardVdbProfile.from_volumes(
            tag='sanity-clones',
            cluster_dm=self.rs_cluster_dm,
            client_dms=self.rs_client_dms[:1],
            volumes_cmts=[self.clone_volumes_cmt]
        )
        self.vdb_profile_bystander = vdbench.StandardVdbProfile.from_clients(
            tag='sanity-bystander',
            cluster_dm=self.rs_cluster_dm,
            client_dms=self.rs_client_dms[1:],
            volume_count=4 * len(self.node_fms),
            connections_per_volume=len(self.rs_client_dms[1:])
        )
        self.vdb_profile_primary.sanitize_configs(node_dms=self.rs_node_dms)
        self.vdb_profile_clones.sanitize_configs(node_dms=self.rs_node_dms)
        self.vdb_profile_bystander.sanitize_configs(node_dms=self.rs_node_dms)

        self.vdb_env_cmt_primary = self.vdb_profile_primary.get_environment()
        self.vdb_env_cmt_clones = self.vdb_profile_clones.get_environment()
        self.vdb_env_cmt_bystander = self.vdb_profile_bystander.get_environment()

        self.vdb_run_cmt_primary = vdbench.ManageRunVdbs.from_environment_cmt(
            environments_cmt=self.vdb_env_cmt_primary,
            validation_config=vdbench.ValidationConfig.on_reads,
            journaling_config=vdbench.JournalingConfig.none
        )
        self.vdb_run_cmt_clones = vdbench.ManageRunVdbs.from_environment_cmt(
            environments_cmt=self.vdb_env_cmt_clones,
            validation_config=vdbench.ValidationConfig.on_reads,
            journaling_config=vdbench.JournalingConfig.none
        )
        self.vdb_run_cmt_bystander = vdbench.ManageRunVdbs.from_environment_cmt(
            environments_cmt=self.vdb_env_cmt_bystander,
            validation_config=vdbench.ValidationConfig.on_reads,
            journaling_config=vdbench.JournalingConfig.none
        )

        self.remove_drives_cmt = vl_components.RemoveDrives(
            rs_cluster_dm,
            selectors.DriveSelector(
                rs_cluster_dm, num=1,
                types=[enums.DriveType.volume]))

    def _preconditions(self):
        """Assert resource requirements."""
        assert len(self.rs_client_dms) >= 2
        assert len(self.rs_node_dms) >= 4

    @property
    def monitor_ignore_faults(self):
        """If we have vNodes ignore the file_system_capacity_low fault.

        :rtype: list[valence.enums.ClusterFaultCode]
        """
        ignore_faults = []
        if any(node_dm.model.is_virtual for node_dm in self.rs_node_dms):
            ignore_faults.append(enums.ClusterFaultCode.file_system_capacity_low)
        return ignore_faults

    def _setup(self):
        """Make sure you can authenticate with nodes and cluster."""
        self.cluster_fm.authenticate()
        for node_fm in self.node_fms:
            node_fm.authenticate()

    def _test(self):
        """Clones sanity test case."""
        self._log_step('Create 3 accounts')
        self.create_accounts_cmt.apply()

        self._log_step('Create 4 "bystander" volumes per node')
        self.vdb_env_cmt_bystander.apply()

        self._log_step('Start Vdbench IO to the "bystander" volumes')
        self.vdb_run_cmt_bystander.apply_async()

        self._log_step('Create 2 "primary" volumes per node')
        self.create_volumes_cmt_primary.apply()

        self._log_step('Start Vdbench IO to the "primary" volumes')
        self.vdb_env_cmt_primary.apply()
        self.vdb_run_cmt_primary.apply_async()

        self._log_step('Sleep 120 seconds while Vdbench runs')
        time.sleep(120)

        self._log_step('Remove random slice drive')
        self.remove_drives_cmt.apply()

        self._log_step('Clone all "primary" volumes')
        self.clone_volumes_cmt.apply_async()

        self._log_step('Stop Vdbench IO to the "primary" volumes')
        self.vdb_run_cmt_primary.set_test_identifier(str(self.current_result.instance_id))
        self.vdb_run_cmt_primary.finish(kill=True)
        self.vdb_run_cmt_primary.reset()
        # Reset the primary connections so we can connect the clone volumes to the same client
        self.vdb_profile_primary.connections_cmt.reset()

        self._log_step('Wait for clones to finish')
        self.clone_volumes_cmt.finish()

        self._log_step('Sleep 20 seconds')
        time.sleep(20)

        self._log_step('Add back removed slice drive')
        self.remove_drives_cmt.reset()

        self._log_step('Wait for slice sync to start')
        vl_helpers.wait_for_slice_sync_to_start(self.cluster_fm)

        self._log_step('Wait for slice sync to finish')
        vl_helpers.wait_for_slice_sync_to_finish(self.cluster_fm)

        self._log_step('Start Vdbench IO to the cloned volumes')
        self.vdb_env_cmt_clones.apply()
        self.vdb_run_cmt_clones.apply_async()

        self._log_step('Wait 120s for all cloned-volume IO to finish')
        time.sleep(120)
        self.vdb_run_cmt_clones.finish(kill=True)
        self.vdb_run_cmt_clones.reset()

        self._log_step('Stop Vdbench IO to the "bystander" volumes')
        self.vdb_run_cmt_bystander.finish(kill=True)
        self.vdb_run_cmt_bystander.reset()

        self._log_step('Check LBAs at rest')
        volume_dms = self.cluster_fm.list_volumes()
        clone_volumes_checksums = vl_helpers.check_lbas_at_rest(self.cluster_fm,
                                                                volume_dms=volume_dms)

        self._log_step('Assert primary/secondary pairs match')
        clone_volumes_checksums.sort()
        assert clone_volumes_checksums[0] > 0

    def _teardown(self):
        """Teardown the test case."""
        # Get support bundles if we failed
        # in Sanity 2.0 we need just have a test case that collects support bundles once to make sure
        # that we can, but since this runs last in vl sanity, we'll just only collect on error or
        # failure or something.
        if self.status not in (camelot.ResultStatus.passed, camelot.ResultStatus.cancelled):

            vl_helpers.pull_support_bundles(
                artifact_manager=self.artifact_manager,
                node_fms=self.node_fms,
                client_fms=self.client_fms
            )

            # Get vdbench artifacts
            if self.vdb_run_cmt_primary.state is components.States.applied:
                self.vdb_run_cmt_primary.store_artifacts()
            if self.vdb_run_cmt_clones.state is components.States.applied:
                self.vdb_run_cmt_clones.store_artifacts()
            if self.vdb_run_cmt_bystander.state is components.States.applied:
                self.vdb_run_cmt_bystander.store_artifacts()

        # Ensure ready for all components
        self.vdb_run_cmt_bystander.ensure_ready()
        self.vdb_env_cmt_bystander.ensure_ready()
        self.vdb_run_cmt_clones.ensure_ready()
        self.vdb_env_cmt_clones.ensure_ready()
        self.vdb_run_cmt_primary.ensure_ready()
        self.vdb_env_cmt_primary.ensure_ready()
        self.remove_drives_cmt.ensure_ready()
        self.clone_volumes_cmt.ensure_ready()
        self.create_volumes_cmt_primary.ensure_ready()
        self.create_accounts_cmt.ensure_ready()
