"""Valence continuous components."""
import abc
import threading
import time

import autologger
import six

from valence import utils
from valence.components import (
    _killable,
    _threaded,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseContinuous(_killable.BlockingApplyKillable):
    """Base continuous component.

    ``apply`` is not supported on continuous components. Users must use ``apply_async``.

    Continously 'apply' something over and over until kill_after_seconds is reached or max iterations
    are reached.

    To timeout on the last iteration of the component, you can call ``finish`` with that timeout.

    :param int kill_after_seconds: How long in seconds the loops will run at a minimum.
    :param int sleep_time: How long in seconds to sleep between loops
    :param int max_iterations: Max number of times to run the loop
    """

    def __init__(self, kill_after_seconds=None, sleep_time=None, max_iterations=None):
        self._in_applied_loop = threading.Event()
        self.kill_after_seconds = kill_after_seconds
        self.sleep_time = sleep_time
        self.max_iterations = max_iterations
        self._iter_lock = threading.Lock()
        self.current_iteration = None
        super(BaseContinuous, self).__init__()

    def _apply_setup(self):
        self._in_applied_loop.clear()
        self.current_iteration = None

    def apply(self, timeout=None):
        """Apply the component synchronously.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        :raises RuntimeError: If a stop condition is not provided.
        """
        if not (self.kill_after_seconds or self.max_iterations):
            raise RuntimeError('Continuous Components cannot be synchronously applied without a stop '
                               'condition.')
        super(BaseContinuous, self).apply(timeout=timeout)

    def finish(self, timeout=None, kill=True):
        """Wait for the last iteration to finish.

        :param int timeout: How long to wait for finish if applied async
        :param bool kill: Terminate the continuous loop at the next iteration
        :raises valence.components.TimeoutError: if kill, timeout, or max iterations are not set.
        """
        self._in_applied_loop.wait()
        if not kill:  # kill is false, so we need to make sure that we will die eventually
            if not self.kill_after_seconds and not self.max_iterations:
                # we will run forever if neither are set, so finish it fast and let it raise
                self._kill = True
                timeout = 0.000001
                kill = True
        super(BaseContinuous, self).finish(timeout=timeout, kill=kill)

    def _apply(self, timeout=None):
        """The public method to thread the loop.

        :param int timeout: Unused.
        """
        if self.kill_after_seconds:
            kill_timer = utils.timeout_timer(self.kill_after_seconds)
            next(kill_timer)  # start the timer
        self._kill = False
        while not self._kill:
            self._in_applied_loop.set()
            self._set_iteration()
            _log.info('Iteration {} starting for {}'.format(self.current_iteration, self))
            self._apply_continuous()
            if self.kill_after_seconds and not next(kill_timer):
                break
            if self.max_iterations and self.current_iteration == self.max_iterations:
                break
            if self.sleep_time:
                time.sleep(self.sleep_time)
            # check again at the end of the sleep
            if self.kill_after_seconds and not next(kill_timer):
                break  # pragma: no cover  # this thing is a PIA to reach unless you want longer tests
            self._in_applied_loop.clear()

    def _set_iteration(self):
        with self._iter_lock:
            if self.current_iteration is None:
                self.current_iteration = 0
            else:
                self.current_iteration += 1

    def _apply_is_running(self):
        return self._worker.is_alive() if self._worker else False

    @abc.abstractmethod
    def _apply_continuous(self):
        """Inner part of the continuous loop.

        Be sure that each iteration can timeout if needed.
        """


@six.add_metaclass(abc.ABCMeta)
class ContinuousTransient(_threaded.Transient, BaseContinuous):

    @abc.abstractmethod
    def _reset(self, timeout=None):
        """Steps to recover when apply/apply_async fails.

        :param int timeout: Maximum time reset can take before raising an error.
        """


@six.add_metaclass(abc.ABCMeta)
class ContinuousReversible(_threaded.Reversible, BaseContinuous):

    @abc.abstractmethod
    def _reset(self, timeout=None):
        """Steps to recover after all iterations have stopped/finished/failed.

        :param int timeout: Maximum time reset can take before raising an error.
        """
