"""Component error definitions."""

from valence import errors


class Error(errors.Error):
    """Base exception for components.

    This exception should not be raised directly.
    """


class ComponentStateError(Error):
    """An operation was attempted while the component was in the incorrect state."""


class ComponentAsyncError(Error):
    """An asynchronous operation entered an error state."""


class PartialStateError(Error):
    """An attempt to load the component found a partial state."""


class ResourceStateError(Error):
    """At least one resource is in an invalid state, the component cannot proceed."""


class ComponentTimeoutError(errors.TimeoutError, Error):
    """Exception that is raised when something times out."""


class UnknownBlockIDError(Error):
    """Error when ClusterBsCheck finds xUnknownBlockIDs.

    ..note:: Unknown blocks must be passed in as dictionaries containing the following keys:
             'volumeID', 'lba', 'blockID'

    :param str message: Error message.
    :param list[dict] unknown_blocks: List of unknown blocks.
    """
    def __init__(self, message, unknown_blocks):
        msg_strings = [message]
        for entry in unknown_blocks:
            msg_strings.append('{}: {}, {}: {}, {}: {}'.format(
                'volumeID', entry['volumeID'],
                'LBA', entry['lba'],
                'blockID', entry['blockID']
            )
            )
        super(UnknownBlockIDError, self).__init__('\n'.join(msg_strings))
        self.args = (message, unknown_blocks,)
