"""Killable base classes."""
import abc

import autologger
import six

from valence.components import (
    _enums,
    _threaded,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseKillableComponent(_threaded.Component):
    """Define a non-blocking implementation that is killable."""

    def __init__(self):
        super(BaseKillableComponent, self).__init__()
        self._kill = False

    def kill(self):
        """Flag the active apply operation to kill itself at the earliest possible opportunity."""
        if self.state is not _enums.States.applying:
            self._raise_state_error('kill')
        self._kill = True

    def finish(self, timeout=None, kill=False):  # pylint: disable=arguments-differ
        """Finish an asynchronous operation.

        :param int timeout: Time in seconds to wait for the finish operation to complete.
        :param bool kill: Whether to kill the apply operation if in progress.
        """
        if kill and self.state is _enums.States.applying:
            self.kill()

        super(BaseKillableComponent, self).finish(timeout=timeout)
        self._kill = False

    def ensure_applied(self, timeout=None, kill=False):  # pylint: disable=arguments-differ
        """Ensure the component is in the "applied" state, applying if necessary.

        :param int timeout: Time in seconds to wait for the component to reach the "applied" state.
        :param bool kill: Whether to kill the apply operation if in progress.
        """
        if kill and self.state is _enums.States.applying:
            self.kill()

        super(BaseKillableComponent, self).ensure_applied(timeout=timeout)
        self._kill = False

    def ensure_ready(self, timeout=None, force=False, kill=True):  # pylint: disable=arguments-differ
        """Transition the component to the "ready" state regardless of its current state.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        :param bool force: Whether to ignore errors that may occur during asynchronous operations.
        :param bool kill: Whether to kill the apply operation if in progress.
        """
        if kill and self.state is _enums.States.applying:
            self.kill()

        super(BaseKillableComponent, self).ensure_ready(timeout=timeout, force=force)
        self._kill = False

    def ensure_ready_async(self, kill=True):  # pylint: disable=arguments-differ
        """Asynchronously transition the component to the "ready" state.

        :param bool kill: Whether to kill the apply operation if in progress.
        """
        if kill and self.state is _enums.States.applying:
            self.kill()

        super(BaseKillableComponent, self).ensure_ready_async()


@six.add_metaclass(abc.ABCMeta)
class BlockingApplyKillable(_threaded.BlockingApply, BaseKillableComponent):
    """Define a blocking implementation that is killable."""

    # there is actually no change from the Base implementation, but we can't use this
    # as the base because `isinstance` wouldn't work correctly


@six.add_metaclass(abc.ABCMeta)
class NonBlockingApplyKillable(_threaded.NonBlockingApply, BaseKillableComponent):
    """Define a non-blocking implementation that is killable."""

    @abc.abstractmethod
    def _kill_apply_async(self):
        """Must make api calls, etc to kill the external process."""

    def kill(self):
        """Kill the underlying component while applying asynchronously."""
        super(NonBlockingApplyKillable, self).kill()
        self._kill_apply_async()
