"""Mock component mixin."""


class MockComponentMixin(object):
    """Mixin to override _apply and _reset with NOOPs.

    This object is used to create mock components. Mock components should only be used for unit testing
    and hacking around with existing datamodels. It is strongly recommended that you do not use them for
    production tests.

    Mock components never make API calls or change any actual state on the system. If you include this
    mixin don't violate that rule.

    NOTE: The exception to the above rule is that they may call `ensure_applied` on any components
    passed to them.
    """
    def __init__(self, *args, **kwargs):
        super(MockComponentMixin, self).__init__(*args, **kwargs)
        self.apply_count = 0

    @property
    def applied(self):
        return bool(self.apply_count)

    def _apply(self, timeout=None):
        """Noop.

        :param float timeout: the timeout.
        """
        self.apply_count += 1

    def _reset(self, timeout=None):
        """Noop.

        :param float timeout: the timeout.
        """

    def _apply_async(self, timeout=None):
        """Noop.

        :param float timeout: the timeout.
        """
        self.apply_count += 1

    def _apply_is_running(self):
        """Noop."""
