"""Process component base classes."""
import abc
import ctypes
import multiprocessing

import autologger
import sf_batch
import six

from valence.components import (
    _enums,
    _errors,
    _killable,
    _threaded,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class ProcessedComponent(_threaded.Component):
    """Process base component class."""

    def __init__(self):
        # noinspection PyTypeChecker
        self._kill_val = multiprocessing.Value(ctypes.c_bool, False)
        super(ProcessedComponent, self).__init__()
        self._lock = multiprocessing.Lock()

        self.worker_result = object()

    @property
    def _kill(self):
        """Get the private kill flag.

        Simulates the _kill flag used by threaded components.

        :rtype: bool
        """
        return self._kill_val.value

    @_kill.setter
    def _kill(self, value):
        """Set the private kill flag.

        Simulates the _kill flag used by threaded components.

        :param bool value: to kill; or not to kill...
        """
        self._kill_val.value = value

    def _finish_ensure_ready(self, timeout=None):
        """Wait for the component to finish ensuring-ready.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        """
        try:
            super(ProcessedComponent, self)._finish_ensure_ready(timeout=timeout)
        except _errors.ComponentStateError:
            raise
        except Exception:
            self.state = _enums.States.failed
            raise
        else:
            self.state = _enums.States.ready

    def _start_worker(self, method, **kwargs):
        """Start a process-based worker object.

        Mixins can override this with process calls.

        :param types.FunctionType method: Method around which to spin thread.
        :param Any kwargs: Any keyword arguments to pass when calling method.
        :rtype: sf_batch.SFProcess
        """
        worker = sf_batch.SFProcess(method, name=self, **kwargs)
        worker.start()
        return worker

    def _finish_worker(self, worker, timeout):
        """Finish the specified worker object.

        Mixins can override this with process calls.

        :param sf_batch.SFProcess worker: Worker to finish.
        :param int timeout: Time in seconds to wait for the object to finish
        :raises valence.components.errors.ComponentAsyncError: If there's an error from the worker
        :raises valence.components.errors.ComponentTimeoutError: If there's a timeout...
        """
        try:
            worker.join(timeout)
            if worker.is_alive():
                raise _errors.ComponentTimeoutError(timeout)
            try:
                # see if we've already gotten something
                # raises any internal exception or # WrappersTimeoutsError
                # returns the value from the func, but we don't really need it.
                self.worker_result = worker.get(0)
            except sf_batch.WrappersTimeoutsError as e:
                exc = worker.get_exception_info()
                if exc is not None:
                    et, ev, tb = exc
                    six.reraise(et, ev, tb.as_traceback())
                raise _errors.ComponentAsyncError(e.args)
        except _errors.ComponentTimeoutError:
            raise
        except Exception:
            worker.kill_get_message()
            raise


@six.add_metaclass(abc.ABCMeta)
class ProcessedReversible(_threaded.Reversible, ProcessedComponent):
    """Reversible Component.

    Can be applied and retracted.
    """


@six.add_metaclass(abc.ABCMeta)
class ProcessedBlockingApply(_threaded.BlockingApply, ProcessedComponent):
    """Define a blocking implementation.

    .. document private functions
    .. automethod:: _apply_setup
    .. automethod:: _apply_is_running
    """


# Killable Components
@six.add_metaclass(abc.ABCMeta)
class ProcessedBaseKillableComponent(ProcessedComponent, _killable.BaseKillableComponent):
    """Define a non-blocking process based implementation that is killable."""


@six.add_metaclass(abc.ABCMeta)
class ProcessedBlockingApplyKillable(ProcessedBlockingApply,
                                     ProcessedBaseKillableComponent):
    """Define a blocking implementation that is killable."""

    # there is actually no change from the Base implementation, but we can't use this
    # as the base because `isinstance` wouldn't work correctly
