"""Component base classes."""
import abc
import threading
import time

import autologger
import sf_batch
import six

from valence.components import (
    _constants,
    _enums,
    _errors,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class Component(object):
    """Component base class.

    .. document private functions
    .. automethod:: _finish_apply
    .. automethod:: _finish_reset
    .. automethod:: _reload
    .. automethod:: _apply_setup
    .. automethod:: _apply_is_running
    .. automethod:: _raise_state_error
    .. automethod:: _log_transition
    """

    def __init__(self):
        super(Component, self).__init__()
        self._state = _enums.States.ready
        self._last_state = None
        self._lock = threading.Lock()
        self._worker = None
        self._ensure_worker = None

    @property
    def state(self):
        """Get the current state of the component.

        :rtype: valence.components.States
        """
        return self._state

    @state.setter
    def state(self, value):
        """Set the state of the component.

        :param valence.components.States value: State to set.
        """
        with self._lock:
            self._last_state = self._state
            self._state = value

    @property
    def component_name(self):
        """Get the component name.

        :rtype: str
        """
        return self.__class__.__name__

    # ==================================================================================================
    # Apply Methods
    # ==================================================================================================
    def apply(self, timeout=None):
        """Apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        if not self._ensure_worker and self.state is _enums.States.ready:
            self._private_apply(timeout=timeout)
        else:
            self._raise_state_error('apply')

    def _private_apply(self, timeout=None):
        """Apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self.state = _enums.States.applying
        self._log_transition(_constants.START_APPLY_MSG)
        try:
            self._apply_setup()
            if not self._reload():
                self._internal_apply(timeout=timeout)
        except _errors.ComponentTimeoutError:
            raise
        except _errors.ComponentStateError:
            self.state = self._last_state
            raise
        except Exception:
            self._log_transition(_constants.FAIL_APPLY_MSG)
            self.state = _enums.States.failed
            raise
        else:
            self._log_transition(_constants.COMPLETE_APPLY_MSG)
            self.state = _enums.States.applied

    def _apply_setup(self):
        """Perform necessary blocking operations at the beginning of applying the component.

        IMPORTANT: These steps are performed before beginning asynchronous operations, so apply_async()
        will not return until these steps have completed.
        """

    def _reload(self):
        """Attempt to reload the component.

        Override this to provide a reload mechanism. If this method returns True, _apply(_async) will
        not be called. This method should update all the same instance variables the _apply method would.
        If an attempt to reload fails, or ends up in a state that can't be applied, a PartialStateError
        exception shall be raised.

        :returns: Whether the underlying component was reloaded.
        :rtype: bool
        """
        return False

    @abc.abstractmethod
    def _internal_apply(self, timeout=None):
        """Apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """

    def apply_async(self, timeout=None):
        """Asynchronously apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        if self._ensure_worker or self.state is not _enums.States.ready:
            self._raise_state_error('apply_async')

        self.state = _enums.States.applying
        self._log_transition(_constants.START_APPLY_ASYNC_MSG)
        try:
            self._apply_setup()
            if not self._reload():
                self._internal_apply_async(timeout=timeout)
        except _errors.ComponentStateError:
            self.state = self._last_state
            raise
        except Exception:
            self.state = _enums.States.failed
            raise

    @abc.abstractmethod
    def _internal_apply_async(self, timeout=None):
        """Asynchronously apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """

    # ==================================================================================================
    # Reset Methods
    # ==================================================================================================
    def reset(self, timeout=None):
        """Reset the component, returning to the state before the component was applied.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """
        if not self._ensure_worker and self.state in {_enums.States.applied, _enums.States.failed}:
            self._private_reset(timeout=timeout)
        else:
            self._raise_state_error('reset')

    def _private_reset(self, timeout=None):
        """Reset the component, returning to the state before the component was applied.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """
        self.state = _enums.States.resetting
        self._log_transition(_constants.START_RESET_MSG)
        try:
            self._internal_reset(timeout=timeout)
        except _errors.ComponentTimeoutError:
            raise
        except _errors.ComponentStateError:
            self.state = self._last_state
            raise
        except Exception:
            self._log_transition(_constants.FAIL_RESET_MSG)
            self.state = _enums.States.failed
            raise
        else:
            self._log_transition(_constants.COMPLETE_RESET_MSG)
            self.state = _enums.States.ready

    @abc.abstractmethod
    def _internal_reset(self, timeout=None):
        """Reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """

    def reset_async(self, timeout=None):
        """Asynchronously reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        if self._ensure_worker or self.state not in {_enums.States.applied, _enums.States.failed}:
            self._raise_state_error('reset_async')

        self.state = _enums.States.resetting
        self._log_transition(_constants.START_RESET_ASYNC_MSG)
        try:
            self._internal_reset_async(timeout=timeout)
        except _errors.ComponentStateError:
            self.state = self._last_state
            raise
        except Exception:
            self.state = _enums.States.failed
            raise

    @abc.abstractmethod
    def _internal_reset_async(self, timeout=None):
        """Asynchronously reset the component if it is in the failed state.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentStateError: If the component is not in the failed state.
        """

    # ==================================================================================================
    # Finish Methods
    # ==================================================================================================
    def finish(self, timeout=None):
        """Finish an asynchronous operation.

        :param int timeout: Time in seconds to wait for the finish operation to complete.
        """
        if self._ensure_worker:
            self._finish_ensure_ready(timeout=timeout)
        elif self.state is _enums.States.applying:
            self._finish_apply(timeout=timeout)
        elif self.state is _enums.States.resetting:
            self._finish_reset(timeout=timeout)
        else:
            self._raise_state_error('finish')

    def _finish_apply(self, timeout=None, raise_on_failure=True):
        """Wait for the component to finish applying.

        :param int timeout: Time in seconds to wait for the component to reach the "applied" state.
        :param bool raise_on_failure: Whether to raise an error on failure.
        """
        try:
            self._internal_finish_apply(timeout=timeout)
            self._worker = None
        except _errors.ComponentTimeoutError:
            raise
        except Exception:  # pylint: disable=broad-except
            self._log_transition(_constants.FAIL_APPLY_MSG)
            self.state = _enums.States.failed
            if not raise_on_failure:
                _log.info('Ignoring error on apply finish:', exc_info=True)
            else:
                raise
        else:
            self._log_transition(_constants.COMPLETE_APPLY_MSG)
            self.state = _enums.States.applied

    @abc.abstractmethod
    def _internal_finish_apply(self, timeout=None):
        """Finish applying the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """

    def _finish_reset(self, timeout=None, raise_on_failure=True):
        """Wait for the component to finish resetting.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        :param bool raise_on_failure: Whether to raise an error on failure.
        """
        try:
            self._internal_finish_reset(timeout=timeout)
            self._worker = None
        except _errors.ComponentTimeoutError:
            raise
        except Exception:  # pylint: disable=broad-except
            self._log_transition(_constants.FAIL_RESET_MSG)
            self.state = _enums.States.failed
            if not raise_on_failure:
                _log.info('Ignoring error on reset finish:', exc_info=True)
            else:
                raise
        else:
            self._log_transition(_constants.COMPLETE_RESET_MSG)
            self.state = _enums.States.ready

    @abc.abstractmethod
    def _internal_finish_reset(self, timeout=None):
        """Finish resetting the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """

    def _finish_ensure_ready(self, timeout=None):
        """Wait for the component to finish ensuring-ready.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        """
        self._finish_worker(self._ensure_worker, timeout)
        self._ensure_worker = None

    # ==================================================================================================
    # Ensure Methods
    # ==================================================================================================
    def ensure_applied(self, timeout=None):
        """Ensure the component is in the "applied" state, applying if necessary.

        :param int timeout: Time in seconds to wait for the component to reach the "applied" state.
        """
        if self._ensure_worker:
            self._raise_state_error('ensure_applied')
        elif self.state is _enums.States.applied:
            return
        elif self.state is _enums.States.ready:
            self._private_apply(timeout=timeout)
        elif self.state is _enums.States.applying:
            self._finish_apply(timeout=timeout)
        else:
            self._raise_state_error('ensure_applied')

    def ensure_ready(self, timeout=None, force=False):
        """Transition the component to the "ready" state regardless of its current state.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        :param bool force: Whether to ignore errors that may occur during asynchronous operations.
        """
        if not self._ensure_worker:
            self._internal_ensure_ready(timeout=timeout, force=force)
        else:
            self._finish_ensure_ready(timeout=timeout)

    def ensure_ready_async(self, timeout=None, force=False):
        """Asynchronously transition the component to the "ready" state.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        :param bool force: Whether to ignore errors that may occur during asynchronous operations.
        """
        if not self._ensure_worker:
            self._ensure_worker = self._start_worker(
                self._internal_ensure_ready,
                timeout=timeout,
                force=force
            )
        else:
            self._raise_state_error('ensure_ready_async')

    def _internal_ensure_ready(self, timeout=None, force=False):
        """Transition the component to the "ready" state regardless of its current state.

        :param int timeout: Time in seconds to wait for the component to reach the "ready" state.
        :param bool force: Whether to ignore errors that may occur during asynchronous operations.
        """
        if self.state is _enums.States.ready:
            return

        time_start = time.time()
        if self.state is _enums.States.applying:
            self._finish_apply(timeout=timeout, raise_on_failure=not force)

        timeout = timeout - int(time.time() - time_start) if timeout is not None else None
        if self.state is _enums.States.resetting:
            self._finish_reset(timeout=timeout, raise_on_failure=not force)

        timeout = timeout - int(time.time() - time_start) if timeout is not None else None
        if self.state in {_enums.States.applied, _enums.States.failed}:
            self._private_reset(timeout=timeout)

    # ==================================================================================================
    # Utility Methods
    # ==================================================================================================
    def is_running(self):
        """Query the underlying process status.

            * This is NOT the same as state==ready or state==applied
            * If is_running() == False then finish() is guaranteed to execute quickly

        :returns: Whether the underlying process is still running.
        :rtype: bool
        """
        if self._ensure_worker:
            return self._ensure_worker.is_alive()
        elif self.state is _enums.States.applying:
            return self._apply_is_running()
        elif self._worker:
            return self._worker.is_alive()
        else:
            return False

    @abc.abstractmethod
    def _apply_is_running(self):
        """Query the apply operation status.

        :returns: Whether the apply operation is still running.
        :rtype: bool
        """

    def _start_worker(self, method, **kwargs):
        """Start a thread-based worker object.

        Mixins can override this with whatever async operator

        :param typing.Callable method: Method around which to spin thread.
        :param Any kwargs: Any keyword arguments to pass when calling method.
        :rtype: sf_batch.SFThread
        """
        worker = sf_batch.SFThread(method, thread_name=self, **kwargs)
        worker.start_thread()
        return worker

    def _finish_worker(self, worker, timeout):
        """Finish the specified worker object.

        Mixins can override this with whatever async operator. However, the same exceptions should be
        raised in the same use cases.

        :param sf_batch.SFThread worker: Worker to finish.
        :param int timeout: Time in seconds to wait for the worker to finish.
        :raises valence.components.errors.ComponentAsyncError: If there's an error from the worker
        :raises valence.components.errors.ComponentTimeoutError: If there's a timeout...
        """
        worker.join(timeout)
        if not worker.is_stopped():
            raise _errors.ComponentTimeoutError(timeout)
        elif worker.is_error():
            exc_info = worker.get_exception_info()
            if exc_info is not None:
                six.reraise(*exc_info)
            else:
                msg = worker.get_error()
                raise _errors.ComponentAsyncError('Asynchronous operation failed: {}'.format(msg))
        else:
            worker.get_result()

    def _raise_state_error(self, requested_action_str):
        """Raise a state error for the requested action.

        :param str requested_action_str: Requested action that generated the state error.
        :raises errors.ComponentStateError: When called.
        """
        if self._ensure_worker:
            state_msg = 'ensuring-ready'
        else:
            state_msg = 'in state {}'.format(self.state)

        raise _errors.ComponentStateError(
            'Cannot {} on {} while {}'.format(requested_action_str, self.component_name, state_msg)
        )

    def _log_transition(self, transition):
        """Log a state transition.

        :param str transition: Transition string
        """
        _log.info('{} {} from state {}'.format(transition, self.component_name, self._last_state))


@six.add_metaclass(abc.ABCMeta)
class Permanent(Component):
    """Permanent Component.

    Resetting from applied is illegal.
    """

    def reset(self, timeout=None):
        """Reset the component if it is in the failed state.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """
        if self.state is _enums.States.failed:
            super(Permanent, self).reset(timeout=timeout)
        else:
            self._raise_state_error('reset')

    def reset_async(self, timeout=None):
        """Asynchronously reset the component if it is in the failed state.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        if self.state is _enums.States.failed:
            super(Permanent, self).reset_async(timeout=timeout)
        else:
            self._raise_state_error('reset_async')

    def _internal_reset(self, timeout=None):
        """Reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentStateError: If the component is not in the failed state.
        """
        self._reset(timeout)

    def _internal_reset_async(self, timeout=None):
        """Asynchronously reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        self._worker = self._start_worker(self._reset, timeout=timeout)

    def _internal_finish_reset(self, timeout=None):
        """Finish resetting the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        self._finish_worker(self._worker, timeout)

    @abc.abstractmethod
    def _reset(self, timeout=None):
        """Reset the component, returning to the state before the component was applied.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """

    def __enter__(self):
        """Enter method for the context manager."""
        self.apply_async()

    def __exit__(self, exception_type, exception_value, traceback):
        """Exit method for the context manager.

        :param Any exception_type: The exception type
        :param Any exception_value: The exception value
        :param Any traceback: The traceback
        """
        self.finish()


@six.add_metaclass(abc.ABCMeta)
class Reversible(Component):
    """Reversible Component.

    Can be applied and retracted.
    """

    def _internal_reset(self, timeout=None):
        """Reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        self._reset(timeout)

    def _internal_reset_async(self, timeout=None):
        """Asynchronously reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        self._worker = self._start_worker(self._reset, timeout=timeout)

    def _internal_finish_reset(self, timeout=None):
        """Finish resetting the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        self._finish_worker(self._worker, timeout)

    @abc.abstractmethod
    def _reset(self, timeout=None):
        """Reset the component, returning to the state before the component was applied.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """

    def __enter__(self):
        """Enter method for the context manager."""
        self.apply()

    def __exit__(self, exception_type, exception_value, traceback):
        """Exit method for the context manager.

        :param Any exception_type: The exception type
        :param Any exception_value: The exception value
        :param Any traceback: The traceback
        """
        self.reset()


@six.add_metaclass(abc.ABCMeta)
class Transient(Component):
    """Transient Component.

    Resetting from applied is trivial.
    """

    def _internal_reset(self, timeout=None):
        """Reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        if self._last_state is _enums.States.failed:
            self._reset(timeout)

    def _internal_reset_async(self, timeout=None):
        """Asynchronously reset the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        if self._last_state is _enums.States.failed:
            self._worker = self._start_worker(self._reset, timeout=timeout)

    def _internal_finish_reset(self, timeout=None):
        """Finish resetting the component.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        """
        if self._last_state is _enums.States.failed:
            self._finish_worker(self._worker, timeout)

    @abc.abstractmethod
    def _reset(self, timeout=None):
        """Reset the component, returning to the state before the component was applied.

        :param int timeout: Time in seconds to wait for the reset operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """

    def __enter__(self):
        """Enter method for the context manager."""
        self.apply_async()

    def __exit__(self, exception_type, exception_value, traceback):
        """Exit method for the context manager.

        :param Any exception_type: The exception type
        :param Any exception_value: The exception value
        :param Any traceback: The traceback
        """
        self.finish()
        self.reset()


@six.add_metaclass(abc.ABCMeta)
class BlockingApply(Component):
    """Define a blocking implementation.

    .. document private functions
    .. automethod:: _apply
    """

    def _internal_apply(self, timeout=None):
        """Apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self._apply(timeout)

    def _internal_apply_async(self, timeout=None):
        """Asynchronously apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self._worker = self._start_worker(self._apply, timeout=timeout)

    def _internal_finish_apply(self, timeout=None):
        """Finish applying the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self._finish_worker(self._worker, timeout)

    def _apply_is_running(self):
        """Query the apply operation status.

        :returns: Whether the apply operation is still running.
        :rtype: bool
        """
        if self._worker:
            return self._worker.is_alive()
        else:
            return False

    @abc.abstractmethod
    def _apply(self, timeout=None):
        """Apply the component, creating the desired state managed by the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """


@six.add_metaclass(abc.ABCMeta)
class NonBlockingApply(Component):
    """Define a non-blocking implementation.

    .. document private functions
    .. automethod:: _apply_async
    .. automethod:: _apply_finish
    """
    # if the default apply_finish is used, this should be overridden with the default timeout
    # either here or in __init__
    TIMEOUT = None

    # How often, in seconds, to poll async process during apply_finish.
    POLL_PERIOD = 5

    def _internal_apply(self, timeout=None):
        """Apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self._apply_async(timeout=timeout)
        self._apply_finish(timeout=timeout)

    def _internal_apply_async(self, timeout=None):
        """Asynchronously apply the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self._apply_async(timeout=timeout)

    def _internal_finish_apply(self, timeout=None):
        """Finish applying the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
        self._apply_finish(timeout=timeout)

    def _apply_finish(self, timeout=None):
        """Finish applying the component.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        :raises errors.ComponentTimeoutError: May be raised if the timeout is reached before finishing.
        """
        timeout = self.TIMEOUT if timeout is None else timeout
        if timeout is None:
            raise AttributeError('{} is incorrectly defined, timeout is None'.format(
                self.component_name))
        start = time.time()
        while self._apply_is_running():
            if time.time() - start > timeout:
                raise _errors.ComponentTimeoutError
            time.sleep(self.POLL_PERIOD)

    @abc.abstractmethod
    def _apply_async(self, timeout=None):
        """Override this to apply your component asynchronously.

        :param int timeout: Time in seconds to wait for the apply operation to complete.
        """
