"""Examples on using basic components.

This file is meant to serve as a quick intro on using processed and killable components. And how data
is communicated to the lower processes.

There are two Manage components you can use. Both call apply_async on their subcomponents,
to start them in a nonblocking way. You can pass in Blocking (BA) subcomponents, which, when applied
asynchronously, will spawn a thread/process to run the ``_apply`` method within. If a NonBlocking (NBA)
is provided, the ``_apply_async`` methods will simple be called as they themselves do not block.

After they are applied, the ManageFinish component will just call ``finish()`` on each sub component.
Finish is blocking here and will make sure each sub component blocks until the timeout is finished in
_apply.

The ManageKill component will call ``kill()`` on each sub component. Depending on the implementation
of the subcomponent, this could block and make it stop to completion or just set a flag/send an API call
that could in fact kill the operation.
"""

from __future__ import print_function

import ctypes  # pylint: disable=unused-import
import multiprocessing  # pylint: disable=unused-import
import time

import autologger
import sf_batch

from valence import components
from valence.components import _processed

_log = autologger.get_sflogging(__name__)


class ManageFinish(components.Transient, components.BlockingApply):
    """This is an example manage component.

    This component takes a list of subcomponents; calls apply_async on them and then
    immediately calls finish.

    Because this a 'Threaded' component, if the sub components are process based and you attempt to
    call apply_async on this component it will fail.

    :param list[valence.components.BaseComponent] sub_cmts: sub components to apply
    """

    def __init__(self, sub_cmts):
        self.sub_cmts = sub_cmts
        super(ManageFinish, self).__init__()

    def _apply(self, timeout=3600):
        _log.info('Start other components')

        for cmt in self.sub_cmts:
            cmt.apply_async()

        _log.info('Finishing other components')
        for cmt in self.sub_cmts:
            cmt.finish()
        _log.info('Components are done.')

    def _reset(self, timeout=None):
        pass


class ManageKill(components.Transient, components.BlockingApply):
    """This is an example manage component that kills the subcomponents 1 sec after apply.

    This component takes a list of subcomponents; calls apply_async on them and then
    immediately calls kill.

    Because this a 'Threaded' component, if the sub components are process based and you attempt to
    call apply_async on this component it will fail.

    :param list[valence.components.BaseComponent] sub_cmts: sub components to apply and kill
    """

    def __init__(self, sub_cmts):
        self.sub_cmts = sub_cmts
        super(ManageKill, self).__init__()

    def _apply(self, timeout=3600):
        _log.info('Start other components')

        for cmt in self.sub_cmts:
            cmt.apply_async()

        time.sleep(1)

        _log.info('Killing other components')
        for cmt in self.sub_cmts:
            cmt.kill()
        _log.info('Components are dead, right?')

    def _reset(self, timeout=None):
        pass


# class BA(components.Transient, components.BlockingApplyKillable):
class BA(components.Transient, _processed.ProcessedBlockingApplyKillable):
    """BlockingApply component example.

    This component, as it is blocking, simply sits in a while loop until it is killed, or finishes on
    its own.

    Swap the class declaration above with the commented out line to see the difference with
    a threaded component. Overall the API is exactly same though.

    Because we call apply_async on the object when instantiated, the _apply method will be running in a
    process.

    :param int timeout: when the apply will be done
    """

    def __init__(self, timeout):
        self.timeout = timeout
        super(BA, self).__init__()

    def _apply(self, timeout=None):
        _log.info('Start me')
        st = time.time()
        # run until you timeout
        while time.time() - st < self.timeout:
            # or you check the kill flag
            if self._kill:
                _log.info('killed it')
                return
            time.sleep(2)
        _log.info('Finish me')

    def _reset(self, timeout=None):
        pass


class NBA(components.Transient, components.NonBlockingApplyKillable):
    """NonBlockingApply component example.

    This component, as it is non-blocking, starts a thread that's running a while loop until it's
    killed, via a flag that would simulate an API call.

    For NonBlockingApply, there's no difference between the threaded/processed one because the
    _apply_async method itself should be not block.

    Experiment with the ``async_operation`` as an SFThread object or an SFProcess. As a result,
    you'll need to modify the ``_fake_api_call`` flag. What happens if you make a method not attached
    to the class asynchronous?

    Experiment with using the _kill flag. When can you use it. When can you not?

    :param int timeout: when the apply will be done
    """

    def __init__(self, timeout):
        self.timeout = timeout
        self.async_operation = None
        self._fake_api_call = False
        # see _apply_async to use this:
        # self._fake_api_call = multiprocessing.Value(ctypes.c_bool, False)
        super(NBA, self).__init__()

    def _haha(self, timeout):
        """The thing to run in a nonblocking way.

        Since self is passed to the thread and process we can check self._kill, but that's cheating

        For a thread, we can simulate the API call with a bool, ``self._fake_api_call``. However,
        if we switch to a process, you'll need to update _fake_api_call to be something that works
        across processes (like ``self._kill``).
        See the comments below and update the flag in ``__init__``.

        """
        _log.info('Start me')
        st = time.time()
        # run until you timeout
        while time.time() - st < timeout:
            # or you get an API call in your thread/process to stop it.
            # if self._kill:  # because 'self._haha' is the process, it has reference to self._kill.
            # if self._fake_api_call.value:
            if self._fake_api_call:
                _log.info('killed it')
                return
            time.sleep(2)
        _log.info('Finish me')

    def _apply_async(self, timeout=None):
        """Swap between a thread and a process for the NonBlocking functionality here.

        If you swap to SFProcess, ``self._fake_api_call`` will not work; change the check in
        ``self._haha`` to use ``self._fake_api_call.value``.
        """

        # self.async_operation = sf_batch.SFProcess(self._haha, self.timeout)
        self.async_operation = sf_batch.SFThread(self._haha, self.timeout)
        self.async_operation.start()

    def _reset(self, timeout=None):
        pass

    def _kill_apply_async(self):
        """Provide a hook to your process.

        If ``_apply_async`` is using a thread, you can just set ``self._fake_api_call`` to ``True``.
        If it's using a process, you'll need to use the shared value.

        Alternatively, you can straight up kill the process dead if you want, using ``terminate``.
        Exceptions will be logged, but will not be raised.
        """
        # uncomment to see how the behavior changes if you're using SFProcess
        # self.async_operation.terminate()
        # if you have a hook into your process, use it...
        # alternatively you can use the
        # self._fake_api_call.value = True
        self._fake_api_call = True
        print('yep')

    def _apply_is_running(self):
        return self.async_operation.is_alive()


if __name__ == '__main__':
    _log.info('START')
    cs = [BA(20), BA(10), BA(15)]
    # cs = [NBA(20), NBA(10), NBA(15)]
    # m = ManageKill(cs)
    m = ManageFinish(cs)
    m.apply()
    _log.info('DONE')
