import time
import unittest

import mock
import pytest

from valence.components import (
    _continuous,
    _enums,
    _errors,
)


class FakeTransientContinuous(_continuous.ContinuousTransient):
    def __init__(self, kill_after_seconds=None, sleep_time=None, max_iterations=None):
        super(FakeTransientContinuous, self).__init__(kill_after_seconds, sleep_time, max_iterations)

    def _apply_continuous(self):
        time.sleep(0.01)

    def _reset(self, timeout=None):
        time.sleep(0.01)


class FakeReversibleContinuous(_continuous.ContinuousReversible):
    def __init__(self, kill_after_seconds=None, sleep_time=None, max_iterations=None):
        super(FakeReversibleContinuous, self).__init__(kill_after_seconds, sleep_time, max_iterations)

    def _apply_continuous(self):
        time.sleep(0.01)

    def _reset(self, timeout=None):
        time.sleep(0.01)


class TestTransientContinuous(unittest.TestCase):
    """Only tests the changes Continuous Components introduced, majority of component unit testing is
       covered in other unit tests."""
    def setUp(self):
        self.cmt_base = FakeTransientContinuous()
        self.cmt_iter = FakeTransientContinuous(max_iterations=2, sleep_time=0.000001)
        self.cmt_kill_iter = FakeTransientContinuous(kill_after_seconds=0.001)

    def test_apply_no_stop_condition(self):
        with pytest.raises(RuntimeError):
            self.cmt_base.apply()
        assert self.cmt_base.state is _enums.States.ready

    def test_apply_max_iterations(self):
        self.cmt_iter.apply()
        assert self.cmt_iter.state is _enums.States.applied

    def test_apply_timeout(self):
        self.cmt_kill_iter.apply()
        assert self.cmt_kill_iter.state is _enums.States.applied

    def test_apply_async(self):
        self.cmt_iter.apply_async()
        self.cmt_iter.finish(kill=False)
        assert self.cmt_iter.state is _enums.States.applied

    def test_no_args(self):
        with pytest.raises(_errors.ComponentTimeoutError):
            self.cmt_base.apply_async()
            self.cmt_base.finish(kill=False)
        assert self.cmt_base.state is _enums.States.applying
        assert self.cmt_base.is_running()

    def test_no_args_finish_kill(self):
        self.cmt_base.apply_async()
        self.cmt_base.finish()
        assert self.cmt_base.state is _enums.States.applied

    def test_apply_max_iter(self):
        self.cmt_iter.apply_async()
        self.cmt_iter.finish(kill=False)
        assert self.cmt_iter.state is _enums.States.applied

    def test_apply_kill_after(self):
        self.cmt_kill_iter.apply_async()
        self.cmt_kill_iter.finish(kill=False)
        assert self.cmt_kill_iter.state is _enums.States.applied

    def test_reset(self):
        self.cmt_kill_iter._reset = mock.MagicMock()
        self.cmt_kill_iter.apply_async()
        with pytest.raises(_errors.ComponentStateError):
            self.cmt_kill_iter.reset()
        self.cmt_kill_iter.finish()
        self.cmt_kill_iter.reset()
        self.cmt_kill_iter._reset.assert_not_called()

    def test_reset_on_failure(self):
        self.cmt_kill_iter._reset = mock.MagicMock()
        self.cmt_kill_iter.apply_async()
        self.cmt_kill_iter.finish()
        self.cmt_kill_iter._state = _enums.States.failed
        self.cmt_kill_iter.reset()
        self.cmt_kill_iter._reset.assert_called()


class TestReversibleContinuous(unittest.TestCase):
    """Mostly everything for the conituous comps is tested above.

    This just needs to make sure the reset behavior is g2g.
    """

    def setUp(self):
        self.cmt_kill_iter = FakeReversibleContinuous(kill_after_seconds=0.2)

    def test_reset(self):
        self.cmt_kill_iter._reset = mock.MagicMock()
        self.cmt_kill_iter.apply_async()
        with pytest.raises(_errors.ComponentStateError):
            self.cmt_kill_iter.reset()
        self.cmt_kill_iter.finish()
        self.cmt_kill_iter.reset()
        self.cmt_kill_iter._reset.assert_called()
