import time
import unittest

import mock
import pytest

from valence.components import (
    _enums,
    _errors,
    _killable,
)


class TestBaseKillableComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_killable.BaseKillableComponent, '__abstractmethods__', new_callable=set):
            cls.cmt = _killable.BaseKillableComponent()

    def setUp(self):
        self.cmt.ensure_ready()

    def test_kill(self):
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.kill()

        self.cmt.apply_async()
        self.cmt.kill()
        assert self.cmt._kill

    @mock.patch.object(_killable.BaseKillableComponent, 'kill')
    def test_finish(self, kill):
        self.cmt.ensure_ready_async()
        self.cmt.finish(kill=True)
        assert not kill.called

        self.cmt.apply_async()
        self.cmt.finish(kill=True)
        assert kill.called
        assert not self.cmt._kill

    @mock.patch.object(_killable.BaseKillableComponent, 'kill')
    def test_ensure_applied(self, kill):
        self.cmt.ensure_applied(kill=True)
        assert not kill.called

        self.cmt.state = _enums.States.ready
        self.cmt.apply_async()
        self.cmt.ensure_applied(kill=True)
        assert kill.called
        assert not self.cmt._kill

    @mock.patch.object(_killable.BaseKillableComponent, 'kill')
    def test_ensure_ready(self, kill):
        self.cmt.ensure_ready(kill=True)
        assert not kill.called

        self.cmt.apply_async()
        self.cmt.ensure_ready(kill=True)
        assert kill.called
        assert not self.cmt._kill

    @mock.patch.object(_killable.BaseKillableComponent, 'kill')
    def test_ensure_ready_async(self, kill):
        self.cmt.ensure_ready_async(kill=True)
        self.cmt.finish()
        assert not kill.called

        self.cmt.apply_async()
        self.cmt.ensure_ready_async(kill=True)
        assert kill.called
        assert not self.cmt._kill


class TestBlockingApplyKillableComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_killable.BlockingApplyKillable, '__abstractmethods__', new_callable=set):
            cls.cmt = _killable.BlockingApplyKillable()

    def setUp(self):
        self.cmt.ensure_ready()

    def mock_apply_killable(self, timeout=None):
        time_start = time.time()
        while not self.cmt._kill:
            if time.time() - time_start > 0.5:
                break

    @mock.patch.object(_killable.BlockingApplyKillable, '_apply')
    def test_ensure_ready(self, apply):
        apply.__name__ = 'mock_apply_killable'
        apply.side_effect = self.mock_apply_killable
        self.cmt.apply_async()
        self.cmt.ensure_ready(kill=True)
        assert self.cmt.state is _enums.States.ready


class TestNonBlockingApplyKillableComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_killable.NonBlockingApplyKillable, '__abstractmethods__',
                               new_callable=set):
            cls.cmt = _killable.NonBlockingApplyKillable()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_killable.NonBlockingApplyKillable, '_kill_apply_async')
    def test_kill(self, kill_apply_async):
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.kill()

        self.cmt.apply_async()
        self.cmt.kill()
        assert self.cmt._kill
        assert kill_apply_async.called
