from __future__ import print_function

import time
import unittest

import mock
import pytest
import sf_batch

from valence.components import (
    _enums,
    _errors,
    _processed,
)


class MockException(Exception):
    """Mock exception."""


class MockTraceback(object):
    """Mock exception."""

    @staticmethod
    def as_traceback():
        return None


class TestBaseProcessedComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_processed.ProcessedComponent, '__abstractmethods__', new_callable=set):
            cls.cmt = _processed.ProcessedComponent()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(sf_batch.SFProcess, 'get')
    @mock.patch.object(sf_batch.SFProcess, 'is_alive')
    def test_ensure_ready_async(self, process_is_alive, process_get):
        process_is_alive.return_value = False
        process_get.return_value = 'mock_return'
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        self.cmt.finish()
        assert process_get.called
        assert self.cmt.worker_result == 'mock_return'
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(sf_batch.SFProcess, 'is_alive')
    def test_ensure_ready_async_timeout(self, process_is_alive):
        process_is_alive.return_value = True
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        with pytest.raises(_errors.ComponentTimeoutError):
            self.cmt.finish()
        assert self.cmt.state is _enums.States.failed

    @mock.patch.object(sf_batch.SFProcess, 'is_alive')
    def test_ensure_ready_async_state_error(self, process_is_alive):
        process_is_alive.side_effect = _errors.ComponentStateError
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.finish()
        assert self.cmt.state is _enums.States.applied

    @mock.patch.object(sf_batch.SFProcess, 'get_exception_info')
    @mock.patch.object(sf_batch.SFProcess, 'get')
    @mock.patch.object(sf_batch.SFProcess, 'is_alive')
    def test_finish_worker(self, process_is_alive, process_get, process_get_exc_info):
        process_is_alive.return_value = True
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        with mock.patch.object(sf_batch.SFProcess, 'join') as process_join:
            with pytest.raises(_errors.ComponentTimeoutError):
                self.cmt.finish(timeout=600)
                assert process_join.called_with(600)

    @mock.patch.object(sf_batch.SFProcess, 'get_exception_info')
    @mock.patch.object(sf_batch.SFProcess, 'get')
    @mock.patch.object(sf_batch.SFProcess, 'is_alive')
    def test_finish_worker_timeout(self, process_is_alive, process_get, process_get_exc_info):
        process_is_alive.return_value = False
        process_get.side_effect = sf_batch.WrappersTimeoutsError('')
        process_get_exc_info.return_value = None
        self.cmt.ensure_ready_async()
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.finish()

    @mock.patch.object(sf_batch.SFProcess, 'get_exception_info')
    @mock.patch.object(sf_batch.SFProcess, 'get')
    @mock.patch.object(sf_batch.SFProcess, 'is_alive')
    def test_finish_worker_reraise(self, process_is_alive, process_get, process_get_exc_info):
        process_is_alive.return_value = False
        process_get.side_effect = sf_batch.WrappersTimeoutsError('')
        process_get_exc_info.return_value = (MockException, None, MockTraceback)
        self.cmt.ensure_ready_async()
        with pytest.raises(MockException):
            self.cmt.finish()

    @mock.patch.object(sf_batch.SFProcess, 'get')
    @mock.patch.object(sf_batch.SFProcess, 'get_exception_info', mock.MagicMock(return_value=None))
    def test_finish_worker_async_error(self, process_get):
        process_get.side_effect = sf_batch.WrappersTimeoutsError('')
        self.cmt.ensure_ready_async()
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.finish()


class TestProcessedBlockingApplyKillableComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_processed.ProcessedBlockingApplyKillable, '__abstractmethods__',
                               new_callable=set):
            cls.cmt = _processed.ProcessedBlockingApplyKillable()

    def setUp(self):
        self.cmt.ensure_ready()

    def mock_apply_killable(self, timeout=None):
        time_start = time.time()
        while not self.cmt._kill:
            if time.time() - time_start > 0.5:
                break

    @mock.patch.object(_processed.ProcessedBlockingApplyKillable, '_apply')
    def test_ensure_ready(self, apply):
        apply.__name__ = 'mock_apply_killable'
        apply.side_effect = self.mock_apply_killable
        self.cmt.apply_async()
        self.cmt.ensure_ready(kill=True)
        assert self.cmt.state is _enums.States.ready
