import unittest

import mock
import pytest
import sf_batch

from valence.components import (
    _enums,
    _errors,
    _threaded,
)


class MockException(Exception):
    """Mock exception."""


class TestBaseComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_threaded.Component, '__abstractmethods__', new_callable=set):
            cls.cmt = _threaded.Component()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_threaded.Component, '_internal_apply')
    def test_apply(self, internal_apply):
        self.cmt.state = _enums.States.applying
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply()

        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply()

        self.cmt.state = _enums.States.resetting
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply()

        self.cmt.state = _enums.States.failed
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply()

        self.cmt.state = _enums.States.ready
        self.cmt.apply(timeout=600)
        assert internal_apply.called_with(timeout=600)
        assert self.cmt.state is _enums.States.applied

    @mock.patch.object(_threaded.Component, '_internal_apply_async')
    def test_apply_async(self, internal_apply_async):
        self.cmt.state = _enums.States.applying
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply_async()

        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply_async()

        self.cmt.state = _enums.States.resetting
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply_async()

        self.cmt.state = _enums.States.failed
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply_async()

        self.cmt.state = _enums.States.ready
        self.cmt.apply_async()
        assert internal_apply_async.called
        assert self.cmt.state is _enums.States.applying

        internal_apply_async.side_effect = _errors.ComponentStateError
        self.cmt.state = _enums.States.ready
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.apply_async()
        assert self.cmt.state is _enums.States.ready

        internal_apply_async.side_effect = _errors.ComponentAsyncError
        self.cmt.state = _enums.States.ready
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.apply_async()
        assert self.cmt.state is _enums.States.failed

    @mock.patch.object(_threaded.Component, '_reload', mock.MagicMock(return_value=True))
    @mock.patch.object(_threaded.Component, '_internal_apply')
    def test_reload(self, internal_apply):
        self.cmt.state = _enums.States.ready
        self.cmt.apply()
        assert not internal_apply.called
        assert self.cmt.state is _enums.States.applied

    @mock.patch.object(_threaded.Component, '_internal_reset')
    def test_reset(self, internal_reset):
        self.cmt.state = _enums.States.resetting
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset()

        self.cmt.state = _enums.States.ready
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset()

        self.cmt.state = _enums.States.applying
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset()

        self.cmt.state = _enums.States.failed
        self.cmt.reset(timeout=600)
        assert internal_reset.called_with(timeout=600)
        assert self.cmt.state is _enums.States.ready
        internal_reset.called = False

        self.cmt.state = _enums.States.applied
        self.cmt.reset()
        assert internal_reset.called
        assert self.cmt.state is _enums.States.ready

        internal_reset.side_effect = _errors.ComponentTimeoutError
        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentTimeoutError):
            self.cmt.reset()
        assert self.cmt.state is _enums.States.resetting

    @mock.patch.object(_threaded.Component, '_internal_reset_async')
    def test_reset_async(self, internal_reset_async):
        self.cmt.state = _enums.States.resetting
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset_async()

        self.cmt.state = _enums.States.ready
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset_async()

        self.cmt.state = _enums.States.applying
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset_async()

        self.cmt.state = _enums.States.failed
        self.cmt.reset_async()
        assert internal_reset_async.called
        assert self.cmt.state is _enums.States.resetting
        internal_reset_async.called = False

        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        assert internal_reset_async.called
        assert self.cmt.state is _enums.States.resetting

        internal_reset_async.side_effect = _errors.ComponentStateError
        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset_async()
        assert self.cmt.state is _enums.States.applied

        internal_reset_async.side_effect = _errors.ComponentAsyncError
        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.reset_async()
        assert self.cmt.state is _enums.States.failed

    @mock.patch.object(_threaded.Component, '_internal_finish_reset')
    @mock.patch.object(_threaded.Component, '_internal_finish_apply')
    def test_finish(self, internal_finish_apply, internal_finish_reset):
        self.cmt.state = _enums.States.ready
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.finish()

        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.finish()

        self.cmt.state = _enums.States.failed
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.finish()

        self.cmt.ensure_ready_async()
        self.cmt.finish()
        assert not internal_finish_apply.called
        assert not internal_finish_reset.called
        assert self.cmt.state is _enums.States.ready

        self.cmt.apply_async()
        self.cmt.finish(timeout=600)
        assert internal_finish_apply.called_with(timeout=600)
        assert not internal_finish_reset.called
        assert self.cmt.state is _enums.States.applied
        internal_finish_apply.called = False

        self.cmt.reset_async()
        self.cmt.finish(timeout=600)
        assert not internal_finish_apply.called
        assert internal_finish_reset.called_with(timeout=600)
        assert self.cmt.state is _enums.States.ready

        internal_finish_reset.side_effect = _errors.ComponentTimeoutError
        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        with pytest.raises(_errors.ComponentTimeoutError):
            self.cmt.finish()
        assert self.cmt.state is _enums.States.resetting

    @mock.patch.object(_threaded.Component, '_internal_finish_apply')
    @mock.patch.object(_threaded.Component, '_internal_apply')
    def test_ensure_applied(self, internal_apply, internal_finish_apply):
        self.cmt.apply_async()
        self.cmt.ensure_applied(timeout=600)
        assert not internal_apply.called
        assert internal_finish_apply.called_with(timeout=600)
        assert self.cmt.state is _enums.States.applied
        internal_finish_apply.called = False

        self.cmt.reset_async()
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.ensure_applied()

        self.cmt.ensure_ready_async()
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.ensure_applied()
        self.cmt.finish()

        self.cmt.state = _enums.States.failed
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.ensure_applied()

        self.cmt.ensure_ready()
        self.cmt.ensure_applied(timeout=600)
        assert internal_apply.called_with(timeout=600)
        assert not internal_finish_apply.called
        assert self.cmt.state is _enums.States.applied
        internal_apply.called = False

        self.cmt.ensure_applied()
        assert not internal_apply.called
        assert not internal_finish_apply.called
        assert self.cmt.state is _enums.States.applied

    @mock.patch('time.time')
    @mock.patch.object(_threaded.Component, '_internal_finish_reset')
    @mock.patch.object(_threaded.Component, '_internal_reset')
    @mock.patch.object(_threaded.Component, '_internal_finish_apply')
    def test_ensure_ready(self, internal_finish_apply, internal_reset, internal_finish_reset, time):
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready()
        assert not internal_finish_apply.called
        assert internal_reset.called
        assert not internal_finish_reset.called
        assert self.cmt.state is _enums.States.ready
        internal_reset.called = False

        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        self.cmt.ensure_ready()
        assert not internal_finish_apply.called
        assert not internal_reset.called
        assert internal_finish_reset.called
        assert self.cmt.state is _enums.States.ready
        internal_finish_reset.called = False

        time.return_value = 0
        self.cmt.ensure_ready()
        assert not internal_finish_apply.called
        assert not internal_reset.called
        assert not internal_finish_reset.called
        assert self.cmt.state is _enums.States.ready

        self.cmt.state = _enums.States.failed
        self.cmt.ensure_ready()
        assert not internal_finish_apply.called
        assert internal_reset.called
        assert not internal_finish_reset.called
        assert self.cmt.state is _enums.States.ready
        internal_reset.called = False

        time.side_effect = [0, 100, 101]
        self.cmt.apply_async()
        self.cmt.ensure_ready(timeout=600)
        assert internal_finish_apply.called_with(timeout=600)
        assert internal_reset.called_with(timeout=500)
        assert not internal_finish_reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(_threaded.Component, '_internal_finish_reset')
    @mock.patch.object(_threaded.Component, '_internal_finish_apply')
    def test_ensure_ready_force(self, internal_finish_apply, internal_finish_reset):
        internal_finish_apply.side_effect = _errors.ComponentAsyncError
        self.cmt.apply_async()
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.ensure_ready()

        self.cmt.state = _enums.States.ready
        self.cmt.apply_async()
        self.cmt.ensure_ready(force=True)
        assert self.cmt.state is _enums.States.ready

        internal_finish_reset.side_effect = _errors.ComponentAsyncError
        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.ensure_ready()

        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        self.cmt.ensure_ready(force=True)
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(sf_batch.SFThread, 'get_exception_info')
    @mock.patch.object(sf_batch.SFThread, 'is_error')
    @mock.patch.object(sf_batch.SFThread, 'is_stopped')
    def test_finish_worker(self, thread_is_stopped, thread_is_error, thread_get_exc_info):
        thread_is_error.return_value = False
        thread_is_stopped.return_value = False
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        with mock.patch.object(sf_batch.SFThread, 'join') as thread_join:
            with pytest.raises(_errors.ComponentTimeoutError):
                self.cmt.finish(timeout=600)
                assert thread_join.called_with(600)

        thread_is_stopped.return_value = True
        self.cmt.finish()
        assert self.cmt.state is _enums.States.ready

        thread_is_error.return_value = True
        thread_get_exc_info.return_value = None
        self.cmt.ensure_ready_async()
        with pytest.raises(_errors.ComponentAsyncError):
            self.cmt.finish()
        thread_is_error.return_value = False
        self.cmt.finish()

        thread_is_error.return_value = True
        thread_get_exc_info.return_value = (MockException, None)
        self.cmt.ensure_ready_async()
        with pytest.raises(MockException):
            self.cmt.finish()


class TestPermanentComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_threaded.Permanent, '__abstractmethods__', new_callable=set):
            cls.cmt = _threaded.Permanent()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_threaded.Permanent, '_reset')
    def test_reset(self, reset):
        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset()

        self.cmt.state = _enums.States.failed
        self.cmt.reset()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(_threaded.Permanent, '_reset')
    def test_reset_async(self, reset):
        self.cmt.state = _enums.States.applied
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.reset_async()

        self.cmt.state = _enums.States.failed
        self.cmt.reset_async()
        self.cmt.finish()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(_threaded.Permanent, '_internal_apply_async')
    def test_context_management(self, _internal_apply_async):
        with self.cmt:
            assert _internal_apply_async.called
            assert self.cmt.state is _enums.States.applying

        assert self.cmt.state is _enums.States.applied


class TestReversibleComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_threaded.Reversible, '__abstractmethods__', new_callable=set):
            cls.cmt = _threaded.Reversible()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_threaded.Reversible, '_reset')
    def test_reset(self, reset):
        self.cmt.state = _enums.States.applied
        self.cmt.reset()
        assert reset.called
        assert self.cmt.state is _enums.States.ready
        reset.called = False

        self.cmt.state = _enums.States.failed
        self.cmt.reset()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(sf_batch.SFThread, 'is_alive', mock.MagicMock(return_value=True))
    @mock.patch.object(_threaded.Reversible, '_reset')
    def test_reset_async(self, reset):
        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        assert self.cmt.is_running()
        self.cmt.finish()
        assert reset.called
        assert self.cmt.state is _enums.States.ready
        reset.called = False

        self.cmt.state = _enums.States.failed
        self.cmt.reset_async()
        self.cmt.finish()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(sf_batch.SFThread, 'is_alive', mock.MagicMock(return_value=True))
    @mock.patch.object(_threaded.Reversible, '_reset')
    def test_ensure_ready_async(self, reset):
        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        assert self.cmt.is_running()
        with pytest.raises(_errors.ComponentStateError):
            self.cmt.ensure_ready_async()
        self.cmt.finish()

        self.cmt.state = _enums.States.applied
        self.cmt.ensure_ready_async()
        self.cmt.finish()
        assert reset.called
        assert self.cmt.state is _enums.States.ready
        reset.called = False

        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        self.cmt.ensure_ready_async()
        self.cmt.finish()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(_threaded.Reversible, '_reset')
    @mock.patch.object(_threaded.Reversible, '_internal_apply')
    def test_context_management(self, _internal_apply, reset):
        with self.cmt:
            assert _internal_apply.called
            assert self.cmt.state is _enums.States.applied

        assert reset.called
        assert self.cmt.state is _enums.States.ready


class TestTransientComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_threaded.Transient, '__abstractmethods__', new_callable=set):
            cls.cmt = _threaded.Transient()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_threaded.Transient, '_reset')
    def test_reset(self, reset):
        self.cmt.state = _enums.States.applied
        self.cmt.reset()
        assert not reset.called
        assert self.cmt.state is _enums.States.ready

        self.cmt.state = _enums.States.failed
        self.cmt.reset()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(_threaded.Transient, '_reset')
    def test_reset_async(self, reset):
        self.cmt.state = _enums.States.applied
        self.cmt.reset_async()
        self.cmt.finish()
        assert not reset.called
        assert self.cmt.state is _enums.States.ready

        self.cmt.state = _enums.States.failed
        self.cmt.reset_async()
        self.cmt.finish()
        assert reset.called
        assert self.cmt.state is _enums.States.ready

    @mock.patch.object(_threaded.Transient, '_internal_apply_async')
    def test_context_management(self, _internal_apply_async):
        with self.cmt:
            assert _internal_apply_async.called
            assert self.cmt.state is _enums.States.applying

        assert self.cmt.state is _enums.States.ready


class TestBlockingApplyComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_threaded.BlockingApply, '__abstractmethods__', new_callable=set):
            cls.cmt = _threaded.BlockingApply()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_threaded.BlockingApply, '_apply')
    def test_apply(self, apply):
        self.cmt.apply(timeout=600)
        assert apply.called_with(timeout=600)
        assert self.cmt.state is _enums.States.applied

    @mock.patch.object(sf_batch.SFThread, 'is_alive', mock.MagicMock(return_value=True))
    @mock.patch.object(_threaded.BlockingApply, '_apply')
    def test_apply_async(self, apply):
        assert not self.cmt.is_running()
        assert not self.cmt._apply_is_running()
        self.cmt.apply_async()
        assert self.cmt.is_running()
        self.cmt.finish()
        assert apply.called
        assert self.cmt.state is _enums.States.applied


class TestNonBlockingApplyComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(_threaded.NonBlockingApply, '__abstractmethods__', new_callable=set):
            cls.cmt = _threaded.NonBlockingApply()

    def setUp(self):
        self.cmt.ensure_ready()

    @mock.patch.object(_threaded.NonBlockingApply, '_apply_finish')
    @mock.patch.object(_threaded.NonBlockingApply, '_apply_async')
    def test_apply(self, apply_async, apply_finish):
        self.cmt.apply(timeout=600)
        assert apply_async.called
        assert apply_finish.called_with(timeout=600)
        assert self.cmt.state is _enums.States.applied

    @mock.patch.object(_threaded.NonBlockingApply, '_apply_finish')
    @mock.patch.object(_threaded.NonBlockingApply, '_apply_async')
    def test_apply_async(self, apply_async, apply_finish):
        self.cmt.apply_async()
        self.cmt.finish()
        assert apply_async.called
        assert apply_finish.called
        assert self.cmt.state is _enums.States.applied

    @mock.patch('time.sleep', mock.MagicMock())
    @mock.patch('time.time')
    @mock.patch.object(_threaded.NonBlockingApply, '_apply_is_running')
    def test_apply_finish(self, apply_is_running, time):
        with pytest.raises(AttributeError):
            self.cmt.apply()

        self.cmt.TIMEOUT = 600
        time.side_effect = [0, 1, 601]
        apply_is_running.side_effect = [True, False]
        self.cmt.state = _enums.States.ready
        self.cmt.apply()
        assert apply_is_running.call_count == 2

        time.side_effect = [0, 1, 601]
        apply_is_running.side_effect = [True, True, False]
        self.cmt.state = _enums.States.ready
        with pytest.raises(_errors.ComponentTimeoutError):
            self.cmt.apply()
