"""Classes used for constructing criteria for filtering data models.

All criteria classes implement a _meets_criteria method that describes how DMs are filtered.

For usage examples, see valence/criteria/tests/test_criteria.py unit tests and
valence/functional_models/_cluster.cluster.py

>>> import operator
>>> from sf_api import models
>>> from sf_api.models import types
>>> from valence import criteria

>>> class Person(models.Base):
   ...:    first_name = types.StringType()
   ...:    last_name = types.StringType()
   ...:    attributes = types.DictType(types.BaseType())
   ...:

>>> people = [Person(dict(first_name='david', last_name='bowie', attributes={'musical': 'yes'})),
    ... Person(dict(first_name='guns', last_name='n roses'))]

A simple filter to find people named "guns":

>>> name_is_guns = criteria.Equals(Person.first_name, 'guns')

>>> guns = name_is_guns.filter(people)

>>> print(guns[0].first_name)
guns

The negate option can be used to reverse the logic of a filter:

>>> name_is_not_guns = criteria.Equals(Person.first_name, 'guns', negate=True)
>>> not_guns = name_is_not_guns.filter(people)
>>> print(not_guns[0].first_name)
david

Compound filter for first name 'guns' reusing old filter and david:

>>> guns_and_david = criteria.Or(name_is_guns, criteria.Equals(Person.first_name, 'david'))
>>> guns_and_david.filter(people)
Out[18]: [<Person instance>, <Person instance>]

Use a custom getter to filter on a nested field:

>>> musical = criteria.Equals(Person.attributes, 'yes', getter=operator.itemgetter('musical'))

See the operator docs for examples of more complex getters:
https://docs.python.org/3/library/operator.html#operator.attrgetter
"""
from valence.criteria._criteria import (
    And,
    BaseCriteria,
    CompoundCriteria,
    ContainedIn,
    Contains,
    Equals,
    EvaluatesToTrue,
    Intersection,
    Is,
    MatchIn,
    Or,
    SingularCriteria,
)
