"""Classes used for constructing criteria for filtering data models.

All criteria classes implement a _meets_criteria method that describes how DMs are filtered.

For usage examples, see valence/utils/tests/test_criteria.py unit tests and
valence/functional_models/_cluster.cluster.py

>>> import operator
>>> from sf_api import models
>>> from sf_api.models import types
>>> from valence import criteria

>>> class Person(models.Base):
   ...:    first_name = types.StringType()
   ...:    last_name = types.StringType()
   ...:    attributes = types.DictType(types.BaseType())
   ...:

>>> people = [Person(dict(first_name='david', last_name='bowie', attributes={'musical': 'yes'})),
    ... Person(dict(first_name='guns', last_name='n roses'))]

A simple filter to find people named "guns":

>>> name_is_guns = criteria.Equals(Person.first_name, 'guns')

>>> guns = name_is_guns.filter(people)

>>> print(guns[0].first_name)
guns

The negate option can be used to reverse the logic of a filter:

>>> name_is_not_guns = criteria.Equals(Person.first_name, 'guns', negate=True)
>>> not_guns = name_is_not_guns.filter(people)
>>> print(not_guns[0].first_name)
david

Compound filter for first name 'guns' reusing old filter and david.
>>> guns_and_david = criteria.Or(name_is_guns, criteria.Equals(Person.first_name, 'david'))
>>> guns_and_david.filter(people)
Out[18]: [<Person instance>, <Person instance>]

Use a custom getter to filter on a nested field
>>> musical = criteria.Equals(Person.attributes, 'yes', getter=operator.itemgetter('musical'))

See the operator docs for examples of more complex getters:
https://docs.python.org/3/library/operator.html#operator.attrgetter
"""
import abc
import re

import aenum
import six

from sf_api import models
from sf_api.models import types


class Undefined(object):
    """Type used to specify when an attribute is undefined."""


@six.add_metaclass(abc.ABCMeta)
class BaseCriteria(object):
    """Abstract base class for filter criteria."""

    @abc.abstractmethod
    def meet_criteria(self, item):
        """Determine if a single item matches the criteria.

        Abstract method that all classes must implement.

        :param Any item: Item to check
        :return: True if item matches criteria, False otherwise
        :rtype: bool
        """

    def filter(self, items):
        """Returns objects where all match criteria.

        :param list[sf_api.models.Base] items: Objects to filter
        :return: Objects that meet criteria
        :rtype: list[sf_api.models.Base]
        """
        if not isinstance(items, list):
            raise TypeError('Items should be a list of data models')

        if (len(items) > 0) and (not isinstance(items[0], models.Base)):
            raise TypeError('Items should be a list of data models, a subclass of base.Base')

        return [item for item in items if self.meet_criteria(item)]


class SingularCriteria(BaseCriteria):
    """Abstract base class for singular filter criteria.

    :param sf_api.models.types.BaseType|property attribute: Attribute to filter on
    :param Any compare_value: Value to compare attribute against
    :param bool negate: If True, invert the logic of the criteria
    :param bool ignore_none: If True, do not filter when compare_value is None
    :param operator.itemgetter|operator.attrgetter getter: A getter function for retrieving nested values
        from the data model. Example usage::

            burst_iops_filter = criteria.Equals(models.Volume.qos, compare_value=7,
                                                getter=operator.itemgetter('burstIOPS'))
    """

    def __init__(self, attribute, compare_value=Undefined, negate=False, ignore_none=True, getter=None):
        self.attribute = attribute
        # compare_value has a default of Undefined so it is possible to explicitly compare against None
        self.compare_value = compare_value
        self.negate = negate
        self.ignore_none = ignore_none
        self.getter = getter

    def _get_value(self, item, attribute):
        """Gets the value of an attribute/property on the item.

        :param sf_api.models.Base item: Data model to get value from
        :param sf_api.models.types.BaseType|property attribute: Attribute
        :return: Value of attribute on item
        :rtype: Any
        :raises TypeError: If you try to filter on an invalid field
        :raises AttributeError: If attribute does not exist on item
        """
        if isinstance(attribute, property):
            value = self._get_value_from_property(attribute, item)
        else:
            value = self._get_value_from_attribute(attribute, item)
        if value is not None and self.getter:
            value = self.getter(value)
        return value

    @staticmethod
    def _get_value_from_property(attribute, item):
        property_name = attribute.fget.__name__
        value = getattr(item, property_name)
        return value

    @staticmethod
    def _get_value_from_attribute(attribute, item):
        if not isinstance(attribute, types.BaseType):
            raise TypeError('Attribute {} should be a types.BaseType object'.format(attribute))
        if not hasattr(item, attribute.name):
            raise TypeError('Object of type {} does not have attribute {}'.format(
                type(item),
                attribute.name))
        else:
            value = getattr(item, attribute.name)
        return value

    @abc.abstractmethod
    def _meet_criteria(self, value):
        """Determine if a single value matches the criteria.

        Abstract method that all classes must implement.

        :param sf_api.models.Base value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """

    def meet_criteria(self, item):
        """Determine if a single item matches the criteria.

        :param sf_api.models.Base item: Item to check
        :return: True if item matches criteria
        :rtype: bool
        """
        if self.ignore_none and self.compare_value is None:
            return True
        value = self._get_value(item, self.attribute)
        result = self._meet_criteria(value)
        if self.negate:
            result = not result
        return result


@six.add_metaclass(abc.ABCMeta)
class CompoundCriteria(BaseCriteria):
    """Abstract base class for compound criteria.

    :param valence.criteria.BaseCriteria|typing.Iterable criteria: Sequence of criteria to combine
        (positional arguments)
    """

    def __init__(self, *criteria):
        self.criteria = list(criteria)  # type: list

    def add_criteria(self, criteria):
        """Add a single criteria.

        :param valence.criteria.SingularCriteria criteria: Criteria to add
        """
        self.criteria.append(criteria)


class EvaluatesToTrue(SingularCriteria):
    """The attribute must be defined and evaluate to True when passed to bool() function.

    This can be used to filter out values such as None, False, [], and ''.

    :param sf_api.models.types.BaseType attribute: Attribute to filter on
    :param bool negate: If True, invert the logic of the criteria
    :param operator.itemgetter|operator.attrgetter getter: A getter function for retrieving nested values
    """
    def __init__(self, attribute, negate=False, getter=None):
        super(EvaluatesToTrue, self).__init__(attribute, negate=negate, getter=getter)

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        return bool(value)


class Equals(SingularCriteria):
    """Criteria where value of attribute must be equal to compare_value."""

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        match = True
        if (value is Undefined) or (value != self.compare_value):
            match = False
        return match


class Is(SingularCriteria):
    """Criteria where value of attribute must be the same object as compare_value."""

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        match = True
        if (value is Undefined) or (value is not self.compare_value):
            match = False
        return match


class ContainedIn(SingularCriteria):
    """Criteria where the value of given attribute must be in given container.

    :param sf_api.models.types.BaseType|property attribute: Attribute to filter on
    :param typing.Container container: Criteria will check that attribute value is in container
    :param bool negate: If True, invert the logic of the criteria
    :param operator.itemgetter|operator.attrgetter getter: A getter function for retrieving nested values
    """
    # pylint: disable=useless-super-delegation
    def __init__(self, attribute, container, negate=False, getter=None):
        super(ContainedIn, self).__init__(attribute, container, negate, getter=getter)

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        return value in self.compare_value


class Contains(SingularCriteria):
    """Criteria where the compare_value is in the value of attribute."""

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        return self.compare_value in value


class Intersection(SingularCriteria):
    """Criteria where at least one item in the given attribute must be in given container."""

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        return bool(set(self.compare_value).intersection(value))


class MatchIn(SingularCriteria):
    """Criteria where the value of given attribute is matched by a regex strings in the container.

    :param sf_api.models.types.BaseType|property attribute: Attribute to filter on
    :param typing.Container container: Criteria will check that attribute value is in container
    :param bool negate: If True, invert the logic of the criteria
    :param operator.itemgetter|operator.attrgetter getter: A getter function for retrieving nested values
    """
    # pylint: disable=useless-super-delegation
    def __init__(self, attribute, container, negate=False, getter=None):
        super(MatchIn, self).__init__(attribute, container, negate, getter=getter)

    def _meet_criteria(self, value):
        """See docstring on SingularCriteria.

        :param Any value: Value to check
        :return: True if value matches criteria
        :rtype: bool
        """
        if value is Undefined or value is None:
            return False
        if isinstance(value, aenum.Enum):
            value = value.value
        for v in self.compare_value:
            if re.search(v, str(value)):
                return True
        return False


class And(CompoundCriteria):
    """And multiple criteria together.

    See CompoundCriteria docstring.
    """
    def meet_criteria(self, item):
        """See docstring on BaseCriteria.

        :param Any item: Item to check
        :return: True if item matches criteria, False otherwise
        :rtype: bool
        """
        for criterion in self.criteria:
            if not criterion.meet_criteria(item):
                return False
        return True


class Or(CompoundCriteria):
    """Or multiple criteria together.

    See CompoundCriteria docstring.
    """
    def meet_criteria(self, item):
        """See docstring on BaseCriteria.

        :param Any item: Item to check
        :return: True if item matches criteria, False otherwise
        :rtype: bool
        """
        for criterion in self.criteria:
            if criterion.meet_criteria(item):
                return True
        return False
