"""Tests valence.criteria."""

import operator

import aenum
import pytest

from sf_api import models
from sf_api.models import types
from valence import criteria


class Job(aenum.Enum):
    doctor = 'DOCTOR'
    demigod = 'DEMIGOD'
    bartender = 'BARTENDER'
    hobo = 'HOBO'


class Fake(models.Base):
    fake = types.StringType()

    @property
    def really_fake(self):
        return self.fake


class Person(models.Base):
    first_name = types.StringType()
    last_name = types.StringType()
    age = types.IntType()
    sex = types.StringType()
    children = types.ListType(types.BaseType, required=False)
    job = types.EnumType(Job)
    hobby = 'motorcycles'
    attributes = types.DictType(types.BaseType())

    @property
    def full_name(self):
        return '{} {}'.format(self.first_name, self.last_name)


@pytest.fixture(scope='module')
def people():
    return [
        Person(dict(first_name='david', last_name='bowie', age=69, sex='male', job=Job.demigod,
                    attributes=dict(living_in_space=True))),
        Person(dict(first_name='bob', last_name='smith', age=20, sex='male', children=[
            Person(dict(first_name='junior', last_name='smith', age=5, sex='male')),
            Person(dict(first_name='kid', last_name='smith', age=3, sex='male'))])),
        Person(dict(first_name='jane', last_name='smith', age=22, sex='female', job=Job.bartender)),
        Person(dict(first_name='anne', last_name='smith', age=20, sex='female', job=Job.doctor))
    ]


def test_filter_on_nested_attributes(people):
    f = criteria.EvaluatesToTrue(Person.attributes, getter=operator.itemgetter('living_in_space'))
    people = f.filter(people)
    assert len(people) == 1


def test_equals_filter_match_one(people):
    f = criteria.Equals(Person.first_name, 'bob')
    people = f.filter(people)
    assert len(people) == 1


def test_equals_filter_on_property(people):
    f = criteria.Equals(Person.full_name, 'bob smith')
    people = f.filter(people)
    assert len(people) == 1


def test_equals_filter_on_wrong_property(people):
    f = criteria.Equals(Fake.really_fake, 'bob smith')
    with pytest.raises(AttributeError):
        f.filter(people)


def test_equals_filter_negate_match_one(people):
    f = criteria.Equals(Person.first_name, 'bob', negate=True)
    filtered_people = f.filter(people)
    assert len(filtered_people) == len(people) - 1


def test_equals_filter_ignore_none(people):
    """Make sure you can explicitly match None."""
    f = criteria.Equals(Person.first_name, None, ignore_none=True)
    people = f.filter(people)
    assert len(people) == len(people)


def test_equals_filter_match_none(people):
    """Make sure you can explicitly match None."""
    f = criteria.Equals(Person.first_name, None, ignore_none=False)
    people[0].first_name = None
    people = f.filter(people)
    assert len(people) == 1


def test_equals_filter_non_existent_attribute(people):
    with pytest.raises(TypeError):
        f = criteria.Equals(Fake.fake, 'value')
        f.filter(people)


def test_is_filter_match_one(people):
    f = criteria.Is(Person.job, Job.demigod)
    people = f.filter(people)
    assert len(people) == 1


def test_negate_is_filter_match_one(people):
    f = criteria.Is(Person.job, Job.demigod, negate=True)
    filtered_people = f.filter(people)
    assert len(filtered_people) == len(people) - 1


def test_is_filter_ignore_none(people):
    """Make sure you can explicitly match None."""
    f = criteria.Is(Person.job, None, ignore_none=True)
    filtered_people = f.filter(people)
    assert len(filtered_people) == len(people)


def test_is_filter_match_none(people):
    """Make sure you can explicitly match None."""
    f = criteria.Is(Person.job, None, ignore_none=False)
    people[0].first_name = None
    people = f.filter(people)
    assert len(people) == 1


def test_is_filter_non_existent_attribute(people):
    with pytest.raises(TypeError):
        f = criteria.Is(Fake.fake, 'value')
        f.filter(people)


def test_and_with_or(people):
    male_or_20 = criteria.Or(
        criteria.Equals(Person.sex, 'male'),
        criteria.Equals(Person.age, 20)
    )
    female_or_20 = criteria.Or(
        criteria.Equals(Person.sex, 'female'),
        criteria.Equals(Person.age, 20)
    )
    f = criteria.And(
        male_or_20,
        female_or_20
    )
    people = f.filter(people)
    assert len(people) == 2
    for person in people:
        assert person.age == 20


def test_or_with_and(people):
    male_and_bob = criteria.And(
        criteria.Equals(Person.sex, 'male'),
        criteria.Equals(Person.first_name, 'bob')
    )
    female_and_smith = criteria.Or(
        criteria.Equals(Person.sex, 'female'),
        criteria.Equals(Person.last_name, 'smith')
    )
    f = criteria.Or(
        male_and_bob,
        female_and_smith
    )
    people = f.filter(people)
    assert len(people) == 3


def test_in_container(people):
    f = criteria.ContainedIn(Person.age, [20, 22])
    people = f.filter(people)
    assert len(people) == 3


def test_in_container_enum(people):
    f = criteria.ContainedIn(Person.job, [Job.demigod, Job.doctor])
    people = f.filter(people)
    assert len(people) == 2


def test_not_in_container(people):
    f = criteria.ContainedIn(Person.age, [99, 200])
    people = f.filter(people)
    assert len(people) == 0


def test_match_in(people):
    f = criteria.MatchIn(Person.sex, ['^male', 'nope'])
    people = f.filter(people)
    assert len(people) == 2


def test_not_match_in(people):
    f = criteria.MatchIn(Person.job, ['HOBO'])
    people = f.filter(people)
    assert len(people) == 0


def test_evaluates_to_true_match(people):
    f = criteria.EvaluatesToTrue(Person.first_name)
    people = f.filter(people)
    assert len(people) == len(people)


def test_contains_filter_dict(people):
    with pytest.raises(TypeError):
        f = criteria.Contains(Person.first_name, 'jack')
        f.filter(dict(something='yada'))


def test_contains_filter_list_of_dict(people):
    with pytest.raises(TypeError):
        f = criteria.Contains(Person.first_name, 'jack')
        f.filter([dict(something='yada')])


def test_contains_filter_wrong_attribute_type(people):
    with pytest.raises(TypeError):
        f = criteria.Contains(Person.hobby, 'knitting')
        f.filter(people)


def test_and_filter_match_one(people):
    f = criteria.And(
        criteria.Equals(Person.last_name, 'smith'),
        criteria.Equals(Person.first_name, 'jane'),
        criteria.Equals(Person.age, 22)
    )
    people = f.filter(people)
    assert len(people) == 1


def test_and_filter_match_two(people):
    f = criteria.And(
        criteria.Equals(Person.sex, 'female'),
        criteria.Equals(Person.first_name, 'jane')
    )
    people = f.filter(people)
    assert len(people) == 1


def test_and_filter_match_none(people):
    f = criteria.And(
        criteria.Equals(Person.last_name, 'smith'),
        criteria.Equals(Person.age, 2)
    )
    people = f.filter(people)
    assert len(people) == 0


def test_or_filter_match_one(people):
    f = criteria.Or(
        criteria.Equals(Person.last_name, 'bowie'),
        criteria.Equals(Person.age, 69)
    )
    people = f.filter(people)
    assert len(people) == 1


def test_or_filter_match_two(people):
    f = criteria.Or(
        criteria.Equals(Person.first_name, 'jane'),
        criteria.Equals(Person.sex, 'female')
    )
    people = f.filter(people)
    assert len(people) == 2
