"""Example test cases to demonstrate how NOT to use inheritance."""
import random

from valence import (
    camelot,
    vl_components,
)
from valence.examples import example_case


class BadOverrideCase(example_case.ExampleCase):
    """Calls super to inherit part of _test.

    Why is this not allowed?
    ------------------------

    It makes it more difficult to easily understand the purpose and steps of a test case
    by reading the _test method. We'd like to avoid having to look at multiple classes to determine what
    the test actually does. tc_scribe will report an error for this test case.

    What to do instead
    ------------------

    Move the shared logic to private methods in the parent test case that can be reused. Or, create
    helper functions for the shared logic.
    """

    def _test(self):
        super(BadOverrideCase, self)._test()
        self._log_step('Do another step')
        # More test case logic goes here


class BadParametrizedCase(camelot.TestCase):
    """Contains dynamic behavior which can make the test case change at runtime.

    This test case is a trivial example and is just meant to demonstrate the concept of parametrized
    test cases.

    Why is this not allowed?
    ------------------------

    We'd like to be able to create an inventory of all test cases without having to run them. This
    allows specific test cases to be tied to the requirement they are testing, and makes it easy to
    statically build an inventory of test cases that can be sent to a database or tool such as QTest.

    If a test case takes arguments that result in dynamic behavior, this effectively makes new test
    cases at runtime based on the arguments passed to the constructor.

    There is a bit of grey area as to what constitutes dynamic behavior, so use your best judgement
    to determine when a separate explicit test case is needed.

    What to do instead
    ------------------

    Use class variables that can be overridden in child classes to create new test cases. In this
    example, you could add a NUM_ACCOUNTS class variable, and might have child classes such as
    CreateSingleAccountCase, CreateMaxAccountsCase, CreateOverMaxAccountsCase, etc.
    """

    def __init__(self, rs_cluster_dm, num_accounts=5):
        super(BadParametrizedCase, self).__init__()
        self.rs_cluster_dm = rs_cluster_dm
        self.num_accounts = num_accounts

        # BAD: the number of accounts can change at runtime
        self.create_accounts_cmt = vl_components.CreateAccounts(tag='vl-example',
                                                                cluster_dm=self.rs_cluster_dm,
                                                                num=self.num_accounts)

    def _test(self):
        """Implementation not necessary for example."""


class BadStepInConditionalCase(camelot.TestCase):
    """Calls _log_step inside of a conditional.

    Why is this not allowed?
    ------------------------

    All the steps in a test cases should be able to be determined without needing to run the test case.
    Putting _log_step inside of a conditional means that the number of steps could change at runtime.

    What to do instead
    ------------------

    You can call _log_step outside of the conditional and have an action such as "Do step if X
    condition is met". Or, you may want to consider having separate test case. tc_scribe will report an
    error for this test case.
    """

    def _test(self):
        """Call random.choice to simulate the output of some API call."""
        result = random.choice([True, False])

        if result:
            # BAD: this step may not be logged at runtime
            self._log_step('Do a step')
            # Step logic goes here
