"""Example test case for Camelot documentation purposes."""
import time

from valence import (
    camelot,
    enums,
    models,
    vl_components,
)
from valence.vl_components import vdbench


class ExampleCase(camelot.TestCase):
    """Example test case that creates an account and some volumes.

    This is meant to demonstrate the structure and coding style of a test case, not to be an example of
    good test case design.

    :param valence.models.ResourceCluster rs_cluster_dm: Resource cluster data model
    :param valence.vl_Components.vdbench.StandardVdbProfile define_vdb_cmt: Define vdb component.
        This will be defined by the suite.
    :param dict artifact_credentials: Artifact credentials to be used by vdbench component
    """

    # Child test cases can override this static/class variable
    NUM_VOLUMES = 3

    def __init__(self, rs_cluster_dm, define_vdb_cmt, artifact_credentials):
        super(ExampleCase, self).__init__()
        self.rs_cluster_dm = rs_cluster_dm
        self.artifact_credentials = artifact_credentials

        # Instantiate components
        self.create_accounts_cmt = vl_components.CreateAccounts(tag='vl-example',
                                                                cluster_dm=self.rs_cluster_dm,
                                                                num=1)
        partial_volume_dms = [models.Volume(total_size=models.DataSize(1, enums.DataUnit.gibibytes))
                              for _ in range(self.NUM_VOLUMES)]
        self.create_volumes_cmt = vl_components.CreateVolumes(tag='vl-example',
                                                              cluster_dm=self.rs_cluster_dm,
                                                              partial_volume_dms=partial_volume_dms,
                                                              accounts_cmt=self.create_accounts_cmt)
        self.define_vdb_cmt = define_vdb_cmt
        self.run_io_cmt = vdbench.RunVdb(self.define_vdb_cmt)

    def _preconditions(self):
        """[Optional] Use this to assert the test case can run.

        NO API CALLS! The preconditions for each test case are run before any test case is executed.
        This is a good place to put inexpensive checks that will make your test case/suite fail early if
        there are any problems.
        """
        assert float(self.rs_cluster_dm.api_endpoint) > 8.0

        # Need at least 1 client
        assert len(self.define_vdb_cmt.client_dms) >= 1

    def _setup(self):
        """Setup stuff if you need for the test case.

        If there is an exception, the suite will exit with error status and not move on
        to next test case.
        """
        # We want to just have IO running the whole test, but it's not a part of the test
        self.run_io_cmt.apply_async()

    def _test(self):
        """Actual test logic.

        Use the assert statement to make assertions, and use _log_step to indicate steps in your test
        case. If there is an exception, teardown will be called and the suite will continue to the next
        test case.
        """
        self._log_step('Create an account')
        self.create_accounts_cmt.apply()

        self._log_step('Create {NUM_VOLUMES} volumes'.format(NUM_VOLUMES=self.NUM_VOLUMES))
        self.create_volumes_cmt.apply()

        self._log_step('Assert volumes were created')
        volume_dms = self.create_volumes_cmt.get_volume_dms()
        # This is just a example of how to use an assert statement. In an actual test case, this specific
        # assertion is not necessary because the CreateVolumes component implicitly checks that the
        # desired volumes were created.
        assert len(volume_dms) == self.NUM_VOLUMES, 'Expected number of volumes were not created'

        self._log_step('Sleep for 5 minutes to allow Vdbench to run')
        time.sleep(60 * 5)

        self._log_step('Assert Vdbench is running')
        self.run_io_cmt.assert_valid(is_running=True)

        self._log_step('Remove account and volumes')
        self.create_volumes_cmt.reset()
        self.create_accounts_cmt.reset()

    def _post_test_assertions(self):
        """Make assertions after _test has finished.

        This is useful if you want to define a base test case with assertions that will be run for every
        inherited test case. Like _test, if there is an exception, teardown will be called and the suite
        can continue.
        """
        self.run_io_cmt.assert_valid(is_running=False)

    def _teardown(self):
        """Teardown the test case and attempt to reset the cluster to the previous state.

        Like _setup, if there is an exception, the suite will fail completely.
        """
        self.create_volumes_cmt.ensure_ready()
        self.create_accounts_cmt.ensure_ready()
        self.run_io_cmt.ensure_ready(force=True)
