"""Example test cases to demonstrate valid use of inheritance."""
import time

from valence import (
    camelot,
    enums,
    functional_models,
    models,
    vl_components,
)
from valence.vl_components import vdbench


class RunIoBaseCase(camelot.TestCase):
    """Creates an account and some volumes while running Vdbench in the background.

    :param valence.models.ResourceCluster rs_cluster_dm: Resource cluster data model
    :param valence.vl_Components.vdbench.SimpleVdbProfile manage_vdb_cmt: Manage vdb component.
    :param dict artifact_credentials: Artifact credentials to be used by vdbench component
    """
    # Number of volumes to create
    NUM_VOLUMES = 3

    # Time in minutes to allow Vdbench to run
    MINUTES_TO_RUN = 5

    def __init__(self, rs_cluster_dm, manage_vdb_cmt, artifact_credentials):
        super(RunIoBaseCase, self).__init__()

        self.rs_cluster_dm = rs_cluster_dm
        self.manage_vdb_cmt = manage_vdb_cmt
        self.artifact_credentials = artifact_credentials

        # Instantiate components
        self.rs_cluster_fm = functional_models.Cluster(rs_cluster_dm, authenticate=False)
        self.create_accounts_cmt = vl_components.CreateAccounts(tag='vl-example',
                                                                cluster_dm=self.rs_cluster_dm,
                                                                num=1)
        partial_volume_dms = [models.Volume(total_size=models.DataSize(1, enums.DataUnit.gibibytes))
                              for _ in range(self.NUM_VOLUMES)]
        self.create_volumes_cmt = vl_components.CreateVolumes(tag='vl-example',
                                                              cluster_dm=self.rs_cluster_dm,
                                                              partial_volume_dms=partial_volume_dms,
                                                              accounts_cmt=self.create_accounts_cmt)
        self.run_io_cmt = vdbench.RunVdb(self.manage_vdb_cmt)

    def _preconditions(self):
        """Make sure the resources meet requirements."""
        assert float(self.rs_cluster_dm.api_endpoint) > 8.0

        # Need at least 1 client
        assert len(self.manage_vdb_cmt.client_dms) >= 1

    def _setup(self):
        """Run IO in the background."""
        # We want to just have IO running the whole test, but it's not a part of the test
        self.run_io_cmt.apply_async()

    def _test(self):
        """Create an account and some volumes while IO is running in the background."""
        self._log_step('Create an account')
        self.create_accounts_cmt.apply()

        self._log_step('Create {NUM_VOLUMES} volumes'.format(NUM_VOLUMES=self.NUM_VOLUMES))
        self.create_volumes_cmt.apply()

        self._log_step('Assert volumes were created')
        volume_dms = self.create_volumes_cmt.get_volume_dms()
        # This is just a example of how to use an assert statement. In an actual test case, this specific
        # assertion is not necessary because the CreateVolumes component implicitly checks that the
        # desired volumes were created.
        assert len(volume_dms) == self.NUM_VOLUMES, 'Expected number of volumes were not created'

        self._log_step('Sleep for {MINUTES_TO_RUN} minutes to allow Vdbench to run'
                       .format(MINUTES_TO_RUN=self.MINUTES_TO_RUN))
        time.sleep(60 * self.MINUTES_TO_RUN)

        self._log_step('Assert Vdbench is running')
        self.run_io_cmt.assert_valid(is_running=True)

        self._log_step('Remove account and volumes')
        self.create_volumes_cmt.reset()
        self.create_accounts_cmt.reset()

    def _post_test_assertions(self):
        """Make assertions after _test has finished.

        This is useful if you want to define a base test case with assertions that will be run for every
        inherited test case. Like _test, if there is an exception, teardown will be called and the suite
        can continue.
        """
        self.run_io_cmt.assert_valid(is_running=False)

    def _teardown(self):
        """Teardown the test case and attempt to reset the cluster to the previous state."""
        self.create_volumes_cmt.ensure_ready()
        self.create_accounts_cmt.ensure_ready()
        self.run_io_cmt.ensure_ready(force=True)


class FiveVolumesIoCase(RunIoBaseCase):
    """Override the number of volumes in the parent test case."""
    NUM_VOLUMES = 5


class TenMinutesIoCase(RunIoBaseCase):
    """Override the number of minutes to run IO."""
    MINUTES_TO_RUN = 10


class RunIoWithEarCase(RunIoBaseCase):
    """A simple IO test with encryption at rest enabled."""

    def _setup(self):
        """Extend _setup on parent case to enable encryption at rest."""
        # Enable EAR before running setup on base class (depending on the test case, you may want to
        # call super first).
        self.enable_encryption = self.rs_cluster_fm.enable_encryption_at_rest()
        super(RunIoWithEarCase, self)._setup()

    def _teardown(self):
        """Teardown test."""
        super(RunIoWithEarCase, self)._teardown()
        # Disable EAR after IO has finished
        self.rs_cluster_fm.disable_encryption_at_rest()


class PrivateMethodCase(camelot.TestCase):
    """A case with private methods that can be used by child classes."""

    def _create_volumes_and_accounts(self):
        """Imagine this does something useful that will be reused in child classes."""

    def _test(self):
        self._log_step('Create volumes and accounts')
        self._create_volumes_and_accounts()


class ChildPrivateMethodCase(PrivateMethodCase):
    """This child test case reuses private methods from a base case."""

    def _test(self):
        """The _test method is fully overridden and does not call super().

        This allows you to glance at the _test method and easily see all the steps in the test case
        without having to look at parent class.
        """
        self._log_step('Create volumes and accounts')
        self._create_volumes_and_accounts()

        self._log_step('Run another step')
        # Logic for step goes here
