"""Valence AsyncHandlers.

These handlers are used for interacting with asynchronous API Calls. See the class
documentation for :py:class:`~valence.functional_models._async_handler.AsyncHandler` for details on
creating a concrete AsyncHandler for a new API call.
"""
import datetime
import time
import typing  # pylint: disable=unused-import

import autologger

import sf_api
from sf_api import enums
from valence import (
    errors,
    models,
)
from valence.functional_models._errors import AsyncHandlerError
from valence.utils import time_utils

_log = autologger.get_sflogging()


class AsyncHandler(object):
    """Async Handler.

    The base class defines a public API for working with asynchronous SolidFire API methods. A subclass
    should be written for each asynchronous API method to implement _get_models and optionally
    _get_fake_async_status. Those implementations will differ depending on which API method the
    AsyncHandler is for.

    _get_models should be overwritten when the asynchronous API method has objects associated with it. It
    will return the updated data models being operated on. Make sure to provide a proper default timeout
    in the subclass's init.

    _get_fake_async_status should be overwritten if at one point there was not an AsyncHandle associated
    with the AsyncHandler's API method. Typically happens when an API method changes from synchronous to
    asynchronous.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: valence.functional_models.Cluster|None
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle|None
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, async_handle_dm=None, timeout=None):
        self.async_handle_dm = async_handle_dm
        self.timeout = timeout
        self._cluster_fm = cluster_fm

        # Attributes for fake async results
        self._fake_create_time = time_utils.get_now_utc()
        # Subclass should provide _fake_async_result_type for each fakeable Async Handler

    @property
    def is_faked(self):
        """True if the AsyncHandle is faked.

        :return: Whether or not the AsyncResult is faked.
        :rtype: bool
        """
        return self.async_handle_dm is None

    def get_async_result(self):
        """Returns the AsyncResult associated with the Valence AsyncHandler.

        :return: AsyncResult associated with the AsyncHandler
        :rtype: valence.models.AsyncResult
        """
        if self.is_faked:
            async_result = self._get_fake_async_result()
        else:
            async_result = self._get_real_async_result()
        return async_result

    def get_status(self):
        """Returns the status of the SolidFire AsyncHandle associated with the Valence AsyncHandler.

        If you call GetAsyncResult using a SolidFire AsyncHandle object and the status is completed or
        failed, the next time GetAsyncResult is called on the same handle, the async results are gone.
        An optional parameter, keepResults, was added to GetAsyncResults in Oxygen to work around
        this behavior. The newer API method ListAsyncResults always behaves the same as GetAsyncResult
        with keepResults set.

        A Valence AsyncHandler will maintain the async results for you.

        :return: Status of the AsyncHandler
        :rtype: valence.enums.AsyncResultStatus
        """
        return self.get_async_result().status

    def is_running(self):
        """Returns true if AsyncResult.status is running, false if completed or failed.

        :rtype: bool
        """
        return self.get_status() is sf_api.AsyncResultStatus.running

    def join(self, poll_period=10, timeout=None):
        """Block and wait for the async operation to complete.

        :param int poll_period: When blocking, number of seconds to poll AsyncHandler's is_running
        :param int timeout: Seconds to timeout on blocking methods, default is set in the subclass

        :raises valence.TimeoutError: if a blocking method times out
        :raises valence.functional_models.AsyncHandlerError: If the async process fails
        """
        # Subclass should provide a reasonable default timeout for each Async Handler
        if not timeout:
            timeout = self.timeout

        start_time = time.time()
        async_status = self.get_status()
        while async_status is sf_api.AsyncResultStatus.running:
            if time.time() - start_time > timeout:
                raise errors.TimeoutError('Timeout of {} hit when waiting for {}'
                                          'to complete async operation'.format(timeout, self))
            time.sleep(poll_period)
            async_status = self.get_status()

        if async_status is sf_api.AsyncResultStatus.failed:
            async_result = self.get_async_result()
            raise AsyncHandlerError('Async process failed: {}'.format(async_result.serialize()),
                                    async_result=async_result)

    def get_models(self, blocking=False, poll_period=10, timeout=None):
        """Return data models associated with the AsyncHandler.

        Returns None if there aren't any. You may also have this method block until the SolidFire async
        operation completes.

        :param bool blocking: Set to True to block until the AsyncHandler is done running
        :param int poll_period: When blocking, number of seconds to poll AsyncHandler's is_running
        :param int timeout: Seconds to timeout on blocking methods, default is set in the subclass

        :return: Any data models associated with the AsyncHandler
        :rtype: valence.models.Base

        :raises valence.TimeoutError: if a blocking method times out
        """
        if blocking:
            self.join(poll_period=poll_period, timeout=timeout)
        return self._get_models()

    def _get_fake_async_result(self):
        """Construct a fake AsyncResult for the Valence AsyncHandler.

        This is needed to fake async processes that don't have AsyncHandles, or had AsyncHandles added
        later in their lifecycle.

        :return: Fake AsyncResult associated with the AsyncHandler.
        :rtype: valence.models.AsyncResult
        """
        fake_last_update_time = time_utils.get_now_utc()
        fake_async_result = models.AsyncResult.create(
            async_result_id=-1,
            create_time=self._fake_create_time,
            details=None,
            last_update_time=fake_last_update_time,
            result_type=sf_api.AsyncResultType.fake,
            status=self._get_fake_async_status())
        return fake_async_result

    def _get_real_async_result(self):
        """Return the AsyncResult associated with the AsyncHandler.

        :return: AsyncResult from the cluster
        :rtype: valence.models.AsyncResult
        """
        try:
            return self._cluster_fm.get_async_result(self.async_handle_dm)
        except sf_api.api_errors.AsyncHandleDoesNotExist as e:
            _log.debug('The async handle does not exist yet: {}. Sleeping 15s.'.format(e))
            time.sleep(15)
            return self._cluster_fm.get_async_result(self.async_handle_dm)

    def _get_fake_async_status(self):
        """Override this if the SolidFire API method has not always been asynchronous.

        The implementation will be asynchronous, but does not return an AsyncHandle.

        :raises valence.functional_models.AsyncHandlerError: if not implemented on subclass
        """
        raise AsyncHandlerError('Async handle data model is None, _get_fake_async_status must '
                                'be subclassed for async operations without an AsyncHandle')

    def _get_models(self):
        """Override this if SolidFire API method that has data models associated with it.

        :rtype: Any
        """
        return None

    def __repr__(self):
        return '{}<async_handle_id={}, timeout={}>'.format(
                self.__class__.__name__,
                self.async_handle_dm.async_handle if self.async_handle_dm else 'fake',
                self.timeout)


class GarbageCollectionAsyncHandler(AsyncHandler):
    """Garbage collection asynchronous handler.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model associated
        with handler.
    :param valence.models.AsyncHandle async_handle_dm: Async Handle data model.
    :param int timeout: Timeout in seconds that is used by get_models.
    :param float start_time: Time after which GC was started.
    """
    def __init__(self, cluster_fm, async_handle_dm=None, timeout=1800, start_time=None):
        super(GarbageCollectionAsyncHandler, self).__init__(
            cluster_fm=cluster_fm,
            async_handle_dm=async_handle_dm,
            timeout=timeout or 1800
        )

        # Get the GCStarted event so we can determine which block services are eligible
        self._eligible_block_services = None
        if cluster_fm.get_current_version() < models.SfVersion('12.0'):
            gc_started_events = []
            while not gc_started_events:
                gc_events = cluster_fm.list_events(
                    event_type=enums.ClusterEventType.gc_event,
                    start_report_time=datetime.datetime.fromtimestamp(start_time) if start_time else None
                )
                gc_started_events = [event for event in gc_events if event.message == 'GCStarted']
                if time.time() - start_time > 60:
                    raise errors.TimeoutError('Timed out after 60s waiting for GCStarted event')

            self._eligible_block_services = gc_started_events[-1].details['eligibleBS']

    def get_status(self):
        """Returns the status of the associated garbage collection AsyncHandle.

        Prior to Magnesium garbage collection async results would report "Working" indefinitely for any
        unhealthy services on the cluster. We work around this by only reporting the status based on
        healthy services.

        :return: Status of the AsyncHandler
        :rtype: valence.enums.AsyncResultStatus
        """
        async_result = self.get_async_result()
        if isinstance(async_result.details, models.GarbageCollectionDetailsPreMg):
            slice_services_status = self._get_services_status(
                services=async_result.details.slice_services,
                service_details=async_result.details.service_details
            )
            block_services_status = self._get_services_status(
                services=async_result.details.block_services,
                service_details=async_result.details.service_details,
                eligible_services=self._eligible_block_services
            )
            if enums.AsyncResultStatus.running in (block_services_status, slice_services_status):
                return enums.AsyncResultStatus.running
            elif enums.AsyncResultStatus.failed in (block_services_status, slice_services_status):
                return enums.AsyncResultStatus.failed
            else:
                return enums.AsyncResultStatus.complete
        else:
            return async_result.status

    @staticmethod
    def _get_services_status(services, service_details=None, eligible_services=None):
        """Get the combined status for the specified services.

        This is a workaround specifically for pre-Magnesium behavior in which the async-result status
        does not appropriately reflect the status of the associated GC event.

        :param dict[str] services: Dictionary of service IDs mapped to their respective status.
        :param dict[str] service_details: Dictionary of service IDs mapped to strings with additional
            information about the status of that service.
        :param list[int] eligible_services: List of service IDs that are eligible for GC.
        :rtype: valence.enums.AsyncResultStatus
        """
        service_details = service_details if type(service_details) is dict else {}
        working_count = 0
        failed_count = 0
        for service_id, status in services.items():
            if service_details.get(service_id) == 'Service decommissioned':
                continue
            elif eligible_services and int(service_id) not in eligible_services:
                continue
            elif status == 'Working':
                working_count += 1
            elif status == 'Failed':
                failed_count += 1

        if working_count:
            return enums.AsyncResultStatus.running
        elif failed_count:
            return enums.AsyncResultStatus.failed
        else:
            return enums.AsyncResultStatus.complete


class EnableEarAsyncHandler(AsyncHandler):
    """Enable encryption at rest async handler.

    :param Any cluster_fm: Cluster functional model associated
        with handler.
    :param valence.models.AsyncHandle async_handle_dm: Async Handle data model.
    :param int timeout: Timeout in seconds that is used by get_models.
    """
    def __init__(self, cluster_fm, async_handle_dm=None, timeout=600):
        super(EnableEarAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                    timeout=timeout)

    def _get_fake_async_status(self):
        """Return a fake async status for EnableEncryptionAtRest.

        GetClusterInfo is used to determine the async status.

        :rtype: valence.enums.AsyncResultStatus
        """
        cluster_dm = self._cluster_fm.get_cluster_info()
        if cluster_dm.encryption_at_rest_state is sf_api.EarState.enabled:
            return sf_api.AsyncResultStatus.complete
        elif cluster_dm.encryption_at_rest_state is sf_api.EarState.enabling:
            return sf_api.AsyncResultStatus.running
        else:
            return sf_api.AsyncResultStatus.failed


class DisableEarAsyncHandler(AsyncHandler):
    """Disable encryption at rest async handler.

    :param Any cluster_fm: Cluster functional model associated
        with handler.
    :param valence.models.AsyncHandle async_handle_dm: Async Handle data model.
    :param int timeout: Timeout in seconds that is used by get_models.
    """
    def __init__(self, cluster_fm, async_handle_dm=None, timeout=600):
        super(DisableEarAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                     timeout=timeout)

    def _get_fake_async_status(self):
        """Return a fake async status for DisableEncryptionAtRest.

        GetClusterInfo is used to determine the async status.

        :rtype: valence.enums.AsyncResultStatus
        """
        cluster_dm = self._cluster_fm.get_cluster_info()
        if cluster_dm.encryption_at_rest_state is sf_api.EarState.disabled:
            return sf_api.AsyncResultStatus.complete
        elif cluster_dm.encryption_at_rest_state is sf_api.EarState.disabling:
            return sf_api.AsyncResultStatus.running
        else:
            return sf_api.AsyncResultStatus.failed


class EnableFeatureAsyncHandler(AsyncHandler):
    """Enable feature async handler.

    :param Any cluster_fm: Cluster functional model associated
        with handler.
    :param valence.enums.SfFeature feature: Feature to enable.
    :param valence.models.AsyncHandle async_handle_dm: Async Handle data model.
    :param int timeout: Timeout in seconds that is used by get_models.
    """

    def __init__(self, cluster_fm, feature, async_handle_dm=None, timeout=600):
        super(EnableFeatureAsyncHandler, self).__init__(cluster_fm=cluster_fm,
                                                        async_handle_dm=async_handle_dm,
                                                        timeout=timeout)
        self.feature = feature

    def _get_fake_async_status(self):
        """Return a fake async status for EnableFeature.

        GetFeatureStatus is used to determine the async status.

        :rtype: valence.enums.AsyncResultStatus
        """
        feature_status = self._cluster_fm.get_feature_status(self.feature)[0]

        if feature_status.enabled:
            return sf_api.AsyncResultStatus.complete
        else:
            return sf_api.AsyncResultStatus.running


class SecureEraseDrivesAsyncHandler(AsyncHandler):
    """Secure erase async handler.

    :param Any cluster_fm: Cluster functional model associated with handler.
    :param list[int] drive_ids: List of drive ids being erased
    :param valence.models.AsyncHandle async_handle_dm: Async Handle data model. None implies
        _get_fake_async_status is overridden
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """

    def __init__(self, cluster_fm, drive_ids, async_handle_dm, timeout=600):
        super(SecureEraseDrivesAsyncHandler, self).__init__(cluster_fm=cluster_fm,
                                                            async_handle_dm=async_handle_dm,
                                                            timeout=timeout)
        self.drive_ids = set(drive_ids)  # type: typing.Set

    def _get_models(self):
        """Return Drive data models for the SecureEraseDrivesAsyncHandler drive_ids list.

        :return: List of drive data models
        :rtype: list[valence.models.Drive]
        """
        all_drive_dms = self._cluster_fm.list_drives()
        return [d for d in all_drive_dms if d.drive_id in self.drive_ids]


class AddDrivesAsyncHandler(AsyncHandler):
    """Add drives async handler.

    :param Any cluster_fm: Cluster functional model associated with handler.
    :param list[int] drive_ids: List of drive ids being added
    :param valence.models.AsyncHandle async_handle_dm: Async Handle data model. None implies
        _get_fake_async_status is overridden
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """

    def __init__(self, cluster_fm, drive_ids, async_handle_dm, timeout=600):
        super(AddDrivesAsyncHandler, self).__init__(cluster_fm=cluster_fm,
                                                    async_handle_dm=async_handle_dm,
                                                    timeout=timeout)
        self.drive_ids = set(drive_ids)  # type: typing.Set

    def _get_fake_async_status(self):
        """Return a fake async status for AddDrives.

        Returns failed if any Drive is in a failed state, running if running or erasing, otherwise
        complete.

        :return: Status of the AsyncHandler
        :rtype: valence.enums.AsyncResultStatus
        """
        drive_dms = self._get_models()
        statuses = {d.status for d in drive_dms}
        running_statuses = {sf_api.DriveStatus.available}
        complete_statuses = {sf_api.DriveStatus.active}
        drive_status = sf_api.DriveStatus
        # noinspection PyTypeChecker
        failed_statuses = set(drive_status) - running_statuses - complete_statuses
        if failed_statuses & statuses:
            # One or more of the drives has a status that indicates failure
            return sf_api.AsyncResultStatus.failed
        elif statuses == complete_statuses:
            return sf_api.AsyncResultStatus.complete
        else:
            return sf_api.AsyncResultStatus.running

    def _get_models(self):
        """Return Drive data models for the AddDrivesAsyncHandler drive_ids list.

        :return: List of drive data models
        :rtype: list[valence.models.Drive]
        """
        all_drive_dms = self._cluster_fm.list_drives()
        return [d for d in all_drive_dms if d.drive_id in self.drive_ids]


class RemoveDrivesAsyncHandler(AsyncHandler):
    """Remove Drives Async Handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: Any
    :param list[int] drive_ids: List of drive ids being removed
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, drive_ids, async_handle_dm=None, timeout=600):
        super(RemoveDrivesAsyncHandler, self).__init__(
            cluster_fm=cluster_fm,
            async_handle_dm=async_handle_dm,
            timeout=timeout or 600
        )
        self.drive_ids = set(drive_ids)  # type: typing.Set

    def _get_models(self):
        """Returns Drive data models for the RemoveDrivesAsyncHandler drive_ids list.

        :return: List of drive data models
        :rtype: list[valence.models.Drive]
        """
        all_drive_dms = self._cluster_fm.list_drives()
        drive_dms = [d for d in all_drive_dms if d.drive_id in self.drive_ids]
        return drive_dms


class AddNodesAsyncHandler(AsyncHandler):
    """Add Nodes Async Handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: Any
    :param str node_static_id: unique id of the node, usually node_dm.static_id
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle|None
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, node_static_id, async_handle_dm, timeout=1200):
        super(AddNodesAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                   timeout=timeout)
        self.node_static_id = node_static_id

    @property
    def is_faked(self):
        """Even when AddNodes has an asyncHandle, it doesn't wait for everything it should.

        :return: True
        :rtype: bool
        """
        return True

    def _get_fake_async_status(self):
        """Return a fake async status for AddNodes.

        Returns running if any AddNode process is in running state, otherwise returns complete.

        :return: Status of the AsyncHandler
        :rtype: valence.enums.AsyncResultStatus
        """
        # If the cluster says the handle is running or failed, return that
        # If the cluster says the handle is complete, do some additional checking for state==active
        if self.async_handle_dm:
            real_async_status = self._get_real_async_result().status
            if real_async_status is not sf_api.AsyncResultStatus.complete:
                return real_async_status

        if self._get_models():
            return sf_api.AsyncResultStatus.complete
        else:
            return sf_api.AsyncResultStatus.running

    def _get_models(self):
        """Return ActiveNode data models for the AddNodesAsyncHandler.

        :return: List of ActiveNode data models
        :rtype: list[valence.models.ActiveNode]
        """
        active_node_dms = self._cluster_fm.list_active_nodes()
        return [n for n in active_node_dms if n.static_id == self.node_static_id]


class RemoveNodesAsyncHandler(AsyncHandler):
    """Remove Nodes Async Handler.

    :param cluster_fm: Cluster functional model associated with the Cluster
    :type cluster_fm: Any
    :param str node_static_id: unique id of the node, usually node_dm.static_id
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle|None
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, node_static_id, async_handle_dm, timeout=600):
        super(RemoveNodesAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                      timeout=timeout)
        self.node_static_id = node_static_id

    @property
    def is_faked(self):
        """Even when RemoveNodes has an asyncHandle, it doesn't wait for everything it should.

        :return: True
        :rtype: bool
        """
        return True

    def _get_fake_async_status(self):
        """Return a fake async status for RemoveNodes.

        Returns running if any add node process is in running state, otherwise returns complete.

        :return: State of the AsyncHandler
        :rtype: list[valence.models.PendingNode]
        """
        # If the cluster says the handle is running or failed, return that
        # If the cluster says the handle is complete, do some additional checking for state==pending
        if self.async_handle_dm:
            real_async_status = self._get_real_async_result().status
            if real_async_status is not sf_api.AsyncResultStatus.complete:
                return real_async_status

        if self._get_models():
            return sf_api.AsyncResultStatus.complete
        else:
            return sf_api.AsyncResultStatus.running

    def _get_models(self):
        """Return PendingNode data models for the RemoveNodesAsyncHandler.

        :rtype: list[valence.models.PendingNode]
        """
        pending_node_dms = self._cluster_fm.list_pending_nodes()
        return [n for n in pending_node_dms if n.static_id == self.node_static_id]


class RebalanceSlicesAsyncHandler(AsyncHandler):
    """Rebalance Slices Async Handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: valence.functional_models.Cluster
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, async_handle_dm=None, timeout=3600):
        super(RebalanceSlicesAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                          timeout=timeout)

    def _get_fake_async_status(self):
        """Returns the status of the RebalanceSlices operation.

        :return: Status of the AsyncHandler
        :rtype: valence.enums.AsyncResultStatus
        """
        are_slices_syncing = self._cluster_fm.is_slice_syncing()

        if are_slices_syncing is True:
            return sf_api.AsyncResultStatus.running
        else:
            return sf_api.AsyncResultStatus.complete


class CloneVolumeAsyncHandler(AsyncHandler):
    """CloneVolume Async Handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: Any
    :param int volume_id: VolumeID of newly created volume.
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, volume_id, async_handle_dm, timeout=600):
        super(CloneVolumeAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                      timeout=timeout)
        self.volume_id = volume_id

    def _get_models(self):
        """Returns volume data model for the volume created by the clone operation.

        :return: Volume data model
        :rtype: list[valence.models.Volume]
        """
        return self._cluster_fm.list_volumes(volume_ids=[self.volume_id])


class SliceReassignmentAsyncHandler(AsyncHandler):
    """Slice reasssignment async handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: Any
    :param int slice_id: VolumeID of newly created volume.
    :param int primary: The new desired primary service ID
    :param list[int] secondaries: List of desired secondary service IDs
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """
    def __init__(self, cluster_fm, slice_id, primary=None, secondaries=None, async_handle_dm=None,
                 timeout=120):
        super(SliceReassignmentAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                            timeout=timeout)
        self.slice_id = slice_id
        self.primary = primary
        self.secondaries = set(secondaries or [])  # type: set

    def _get_fake_async_status(self):
        """Returns the status of the reassign slice operation.

        :return: Status of the AsyncHandler
        :rtype: valence.enums.AsyncResultStatus
        """
        slice_info = self._cluster_fm.get_slice_info(slice_id=self.slice_id)
        if self.primary is not None and slice_info.slice.primary != self.primary:
            return sf_api.AsyncResultStatus.running
        if self.secondaries and set(slice_info.slice.live_secondaries) != self.secondaries:
            extra_secondaries = list(set(slice_info.slice.live_secondaries).difference(self.secondaries))
            if extra_secondaries:
                self._cluster_fm._reassign_slice(  # pylint: disable=protected-access
                    self.slice_id, db_version=slice_info.db_version, action=enums.SliceAction.prune,
                    prune=extra_secondaries)
            return sf_api.AsyncResultStatus.running
        return sf_api.AsyncResultStatus.complete


class FillSliceAsyncHandler(AsyncHandler):
    """FillSlice async handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: Any
    :param int slice_id: VolumeID of newly created volume.
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """

    def __init__(self, cluster_fm, slice_id, async_handle_dm, timeout=600):
        super(FillSliceAsyncHandler, self).__init__(cluster_fm, async_handle_dm=async_handle_dm,
                                                    timeout=timeout)
        self.slice_id = slice_id

    def _get_models(self):
        """Returns volume data model for the volume that was filled.

        :return: Volume data model
        :rtype: list[valence.models.Volume]
        """
        return self._cluster_fm.list_volumes(volume_ids=[self.slice_id])


class RtfiAsyncHandler(AsyncHandler):
    """Rtfi async handler.

    :param sf_rtfi.reactor.HTTPIRTFIReactor reactor: reactor used to monitor rtfi
    :param int timeout: Optional for get_models. Default depends on each async handler subclass
    """

    def __init__(self, reactor, timeout=1800):
        super(RtfiAsyncHandler, self).__init__(cluster_fm=None, async_handle_dm=None,
                                               timeout=timeout)
        self.reactor = reactor

    def _get_fake_async_status(self):
        """Returns status of the ongoing rtfi.

        :return: Status of the rtfi
        :rtype: valence.enums.AsyncResultStatus
        """
        # check if successfully completed first, if not, done indicates a failed state.
        if self.reactor.rtfi_state.complete():
            self.reactor.join()
            return sf_api.AsyncResultStatus.complete
        elif (self.reactor.rtfi_state.failed() or self.reactor.rtfi_state.aborted() or
                self.reactor.rtfi_state.timedout() or self.reactor.rtfi_state.done()):
            self.reactor.join()
            return sf_api.AsyncResultStatus.failed
        else:
            return sf_api.AsyncResultStatus.running


class SetClusterStructureAsyncHandler(AsyncHandler):
    """SetClusterStructure async handler.

    :param cluster_fm: Cluster functional model associated with handler.
    :type cluster_fm: Any
    :param async_handle_dm: Async Handle data model. None implies _get_fake_async_status is overridden
    :type async_handle_dm: valence.models.AsyncHandle
    :param int timeout: Optional for get_models. Default depends on each async handler subclass

    """
    def __init__(self, cluster_fm, async_handle_dm, timeout=300):
        super(SetClusterStructureAsyncHandler, self).__init__(cluster_fm,
                                                              async_handle_dm=async_handle_dm,
                                                              timeout=timeout)
