"""Base class for support bundles functionality."""

import abc

import autologger
import sf_ssh

from valence.functional_models import _errors

_log = autologger.get_sflogging()


class BundleInfo(object):
    """Data object for storing info about a support bundle.

    :param str bundle_path: Path where bundle will be stored
    :param list[str] missing_files: List of files that were expected, but are missing
    :param list[str] empty_files: List of files that are empty
    :param set remote_rm_paths: Remote folders that should be removed by cleanup_remote
    :param list attempts: List of attempts to collect support bundles of the format (method, success)
        where method is the method used (API vs. SSH) and success is a bool that will be set to True
        if the attempt was successful.
    :param str bundle_name: Name of the bundle. This will be set by create_bundle
    :param valence.functional_models._node.BundleMethod method: Method used to create the bundle.
        This will be set by _attempt_making_bundle.
    """

    def __init__(self, bundle_path=None, missing_files=None, empty_files=None,
                 remote_rm_paths=None, attempts=None, bundle_name=None,
                 method=None):
        self.bundle_path = bundle_path
        self.missing_files = missing_files or []
        self.empty_files = empty_files or []
        self.remote_rm_paths = remote_rm_paths or set()
        self.attempts = attempts or []
        self.bundle_name = bundle_name
        self.method = method


class BundleBase(object):
    """Base class for support bundles.

    Defines abstract methods that must be defined by functional models that have the ability
    to pull support bundles. Also defines concrete methods that are used for support bundle collection.
    """

    # Compression type used for bundles
    COMPRESSION_TYPE = 'gz'

    @abc.abstractmethod
    def _get_sf_msg(self, msg):
        """Override to prepend extra info to logging messages such as the object type and IP.

        For example, the prefix for nodes looks like this: "Node(<mip>): ".

        :param str msg: Message
        :rtype: str
        :return: Formatted message
        """

    @abc.abstractmethod
    def pull_support_bundle(self, *args, **kwargs):
        """Create a support bundle and store on given artifact manager."""

    @abc.abstractmethod
    def get_ssh_connection(self):
        """Return an SSH connection to the functional model's associated object.

        :rtype: sf_ssh.Ssh
        """

    def _cleanup_remote(self, remote_rm_paths):
        """Clean up any remote files or folders which are not necessary.

        Description::
            Nodes, Clients and FDVA may have any redundant files or folders(logs, cores etc.)
            These may need to be removed once the Bundle Collection is successful.

        :param set remote_rm_paths: Paths to folders/files to be removed
        :raises valence.functional_models.errors.BundleError: If the files were not able to be removed.
        """
        remote_rm_paths = list(remote_rm_paths)
        _log.debug(self._get_sf_msg('List of paths (on host) to remove: {}'.format(remote_rm_paths)))
        try:
            with self.get_ssh_connection() as ssh_conn:
                err_messages = []
                for path in remote_rm_paths:
                    result = ssh_conn.run_command('rm -r {}'.format(path))
                    if result.return_code == 0:
                        remote_rm_paths.remove(path)
                    else:
                        err_messages.append(result.stderr)
                if remote_rm_paths:
                    raise _errors.BundleError(
                        self._get_sf_msg('Unable to remove (on host): {} and Exceptions: {}'.format(
                            remote_rm_paths, err_messages)))
        except sf_ssh.errors.SshError as exc:
            raise _errors.BundleError(self._get_sf_msg('Unable to remove: {}, Exception: {}'
                                                       .format(remote_rm_paths, exc)))
        _log.debug(self._get_sf_msg('All paths on host Removed Successfully!'))
