"""Base class for client functional model."""

import abc

import autologger
import sf_ssh
import six

from valence import models
from valence.functional_models import _errors

_log = autologger.get_sflogging()


@six.add_metaclass(abc.ABCMeta)
class Client(object):
    """Client functional model base class.

    This abstract class defines methods that are required by the client functional models for every OS.

    :param valence.models.ResourceClient client_dm: client data model
    """

    def __init__(self, client_dm):
        self.client_dm = client_dm

    def __repr__(self):
        return '{}<{}={}>'.format('ClientFm', 'mip', self.client_dm.mip)

    @property
    def mip(self):
        return self.client_dm.mip

    @property
    def username(self):
        return self.client_dm.username

    @property
    def password(self):
        return self.client_dm.password

    def _get_sf_msg(self, msg):
        """Prefix message with client object info.

        :param str msg: Message
        :rtype: str
        """
        return 'Client({}): {}'.format(self.mip, msg)

    def get_ssh_connection(self):
        """Return an SSH connection to the client.

        :rtype: sf_ssh.Ssh
        """
        return sf_ssh.Ssh(self.mip, self.username, self.password)

    def get_interactive_shell(self, use_sudo=True):
        """Return an interactive SSH Shell connection to the client.

        ;param use_sudo: Default to running commands using sudo prefix

        :rtype: sf_ssh.SshShell
        """
        return sf_ssh.SshShell(self.mip, self.username, self.password, use_sudo=use_sudo)

    def run_command(self, command, timeout=600, raise_on_failure=False):
        """Executes a shell command.

        :param str command: shell command to execute
        :param int timeout: Timeout before Command returns output
        :param bool raise_on_failure: If True, raise exception if return code not 0. If False, do not
               raise exception, just return tuple.
        :raises errors.FunctionalClientError: Raise if command on client did not complete successfully
        :rtype: sf_ssh.ssh.RunCommandOutput
        """

        with self.get_ssh_connection() as ssh:
            result = ssh.run_command(command, timeout, use_sudo=True)
            if result.return_code != 0 and raise_on_failure:
                msg = ('Failed to run command [{}] on client {}. Failed with rc={}: {}'
                       .format(command, self.mip, result.return_code, result.stderr))
                raise _errors.FunctionalClientError(msg)
            return result

    @classmethod
    def get_mock_fm(cls):
        """Factory method for instantiating a mock client FM.

        :rtype: valence.functional_models.Client
        """
        rs_client_dm = models.ResourceClient.get_mock_object()
        return cls(rs_client_dm)
