"""Linux client functional model."""
from __future__ import division

import abc
import os
import re
import tempfile
import threading
import time

import autologger
import six

from valence import enums
from valence.functional_models import (
    _bundle_base,
    _constants,
    _errors,
    _fm_utils,
)
from valence.functional_models._client import base

_log = autologger.get_sflogging()

ISCSI_CONFIG_DIR = '/etc/iscsi'
ISCSI_CONNECTION = 'iscsi'


@six.add_metaclass(abc.ABCMeta)
class LinuxClient(base.Client, _bundle_base.BundleBase):
    """Linux Client functional model."""
    def update_iscsi_d_conf(self, auth_user, auth_pass, tsecret=None, double_authentication=False):
        """Update iscsid.conf with Account details.

        :param str auth_user: Username of SF Account
        :param str auth_pass: Initiator Secret of SF Account
        :param str tsecret: Target Secret of SF Account. Only used if double_authentication is True.
        :param bool double_authentication: Enable Double Authentication.
        """
        if not double_authentication:
            tsecret = None
        elif tsecret == auth_pass:
            raise _errors.FunctionalClientError(
                'You should not have same INITIATOR {} and Target Secret {} '
                'when using double authentication.'.format(auth_pass, tsecret))
        iscsid_conf = self._get_client_iscsid_conf()
        iscsid_conf = self._iscsi_set_targets_and_discovery_info(iscsid_conf, auth_user, auth_pass,
                                                                 tsecret)
        self._put_iscsid_conf(iscsid_conf)

    def deactivate_chap(self):
        """Comment out all of the "CHAP Settings" section in iscsid.conf.

        NOTE: This is used for volume-access-group connections in place of update_iscsi_d_conf().
        """
        iscsid_conf = self._get_client_iscsid_conf()
        in_chap_settings = False
        for i, line in enumerate(iscsid_conf):
            if 'CHAP Settings' in line:
                # entering CHAP settings lines, comment EVERYTHING!!
                in_chap_settings = True
            elif in_chap_settings and ('Timeouts' in line):
                # we are now past the CHAP settings lines
                break
            elif in_chap_settings and not line.startswith('#'):
                # we are in chap settings and line is NOT commented
                iscsid_conf[i] = '#CHAP_DISABLED:{}'.format(line)

        self._put_iscsid_conf(iscsid_conf)

    def reactivate_chap(self):
        """Uncomment any lines commented out in iscsid.conf for disabling CHAP."""
        iscsid_conf = self._get_client_iscsid_conf()
        for i, line in enumerate(iscsid_conf):
            if line.startswith('#CHAP_DISABLED:'):
                iscsid_conf[i] = line.replace('#CHAP_DISABLED:', '', 1)

        self._put_iscsid_conf(iscsid_conf)

    def volume_ceremony(self, portal_ip, targets=None, purge=False):
        """Run volume ceremony on linux client via iscsi.

        Note you MUST run update_iscsi_d_conf prior to running volume_ceremony.

        :param str portal_ip: Storage IP.
        :param list[str] targets: List of Solidfire Targets
        :param bool purge: Should we purge restart the iscsi credentials
        :rtype: dict
        """
        if purge:
            # If volumes were deleted, probably want to purge and restart
            self.purge_restart()
        else:
            self.iscsid_start(restart=False)
        if not targets:
            targets = self._discover(portal_ip)
        self.login_all(svip=portal_ip, targets=targets)
        disk_map = self.get_disk_map()
        return disk_map

    @abc.abstractmethod
    def iscsid_start(self, restart=False):
        """Start open-iscsi.

        :param bool restart: Restart iscsid
        """

    @abc.abstractmethod
    def iscsid_stop(self, iscsi_root_volume=False):
        """Stop iscsid.

        :param bool iscsi_root_volume: Is the volume the iscsi root
        """

    def wait_until_iscsid(self, status='up'):
        """Wait until iscsid up or down.

        :param str status: Expected iscsid status, can be 'up' or 'down'
        """
        oper = _constants.OPER_EQ
        if status == 'up':
            oper = _constants.OPER_NEQ
        try:
            b_poll = _fm_utils.BlockPoll(10 * 60,
                                         self.get_pid_of_iscsid,
                                         [],
                                         oper,
                                         poll_period=10)
            b_poll.start_polling()
        except _errors.PollMaxError:
            raise _errors.IscsiError('iscsid services is not {} even after 10 minutes on {}'
                                     .format(status, self))

    def get_pid_of_iscsid(self):
        """Get Process ids of iscsid.

        :rtype: list
        """
        pidof_result = self.run_command('pidof iscsid')
        if pidof_result.return_code:
            return []
        else:
            outlist = pidof_result.stdout.strip().split(' ')
            for index in range(len(outlist)):
                outlist[index] = int(outlist[index])
            return outlist

    def login_all(self, mode=enums.IscsiMode.all,
                  batch_size=_constants.DEFAULT_BATCH_SIZE, svip=None, targets=None):
        """Login to all discovered iscsi targets with [iscsiadm -m node -L all].

        :param int mode: The mode to use for iscsi_login_all.
        :param int batch_size: The size of batches to use for BATCH
        :param str svip: The SVIP to login
        :param list targets: The list of targets to use for login (for BATCH)
        """
        if mode not in enums.IscsiMode:
            raise _errors.IscsiError('Invalid mode used for iscsi_login_all(), mode {}'.format(mode))

        if mode == enums.IscsiMode.all:
            _log.debug('Logging into all iscsi targets for {}...'.format(self))
            cmd = 'iscsiadm -m node -L all'
            self._run_command_check_output(cmd)
        elif mode == enums.IscsiMode.batch:
            try:
                self.login_all_batch_threaded(svip=svip, targets=targets)
            except Exception:  # pylint: disable=broad-except
                self.login_all_batch(svip=svip, batch_size=batch_size, targets=targets)

    def login_all_batch(self, svip=None, batch_size=_constants.DEFAULT_BATCH_SIZE, targets=None):
        """Log in to all iscsi targets in batched chunks.

        :param str svip: The IP address of the portal.
        :param int batch_size: The number of commands to place in a single batch.
        :param list targets: A list of targets to login to.

        :rtype: list[str]
        """

        _log.debug('Logging in to all iscsi targets on '
                   '{0} in batches of {1}'.format(self, batch_size))

        if targets is None:
            targets = self._discover(svip)
        # making sure we have the last space-delimited field in this list
        targets = [x.split(' ')[-1] for x in targets]

        batch_return = []
        batch_count = 1
        while len(targets):
            batch = []
            idx = min(batch_size, len(targets))
            while idx > 0:
                idx -= 1
                batch.append(targets[idx])
                del targets[idx]
            result = self.run_batch_login(batch, svip)
            batch_return.extend(result)
            batch_count += 1
        return batch_return

    def login_all_batch_threaded(self, svip=None,
                                 batch_size=_constants.DEFAULT_THREADED_BATCH_SIZE, targets=None):
        """Log in to all iscsi targets in batched chunks using multiple threads.

        :param str svip: The IP address of the portal.
        :param int batch_size: The number of commands to place in a single batch.
        :param list targets: A list of targets to login to.
        :rtype: tuple
        """
        _log.debug('Logging in to all iscsi targets on '
                   '{0} in batches of {1}'.format(self, batch_size))

        if targets is None:
            targets = self._discover(svip)
        # making sure we have the last space-delimited field in this list
        targets = [x.split(' ')[-1] for x in targets]

        threads = []
        while len(targets):
            batch = []
            idx = min(batch_size, len(targets))
            while idx > 0:
                idx -= 1
                batch.append(targets[idx])
                del targets[idx]
            t = threading.Thread(target=self.run_batch_login, args=(batch, svip,))
            threads.append(t)

        _ = [t.start() for t in threads]
        _ = [t.join() for t in threads]
        return threads

    def logout_all(self, svip=None):
        """Log out of all iscsi targets with [iscsiadm -m node -U all].

        :param str svip: The IP address of the portal.
        :rtype: str
        """

        _log.debug('Logging out of all iscsi targets on {}...'.format(self))
        if svip:
            _log.debug(('SVIP was passed in to logout_all, only logging out '
                        'of volumes with sessions to {}').format(svip))
            cmd = 'iscsiadm -m node -u -p {}'.format(svip)
        else:
            cmd = 'iscsiadm -m node -U all'
        return self._run_command_check_output(cmd)

    def purge_restart(self, iscsi_root_volume=False):
        """Purge and Restart iscsid.

        :param bool iscsi_root_volume: Is the volume the iscsi root
        """
        try:
            self.logout_all()
        except _errors.IscsiError as err:
            _log.debug('No devices found when trying to logout... Thats OK... ERROR {}'.format(err))
        self.iscsid_stop(iscsi_root_volume)
        self.purge_credentials()
        self.iscsid_start()

    def purge_credentials(self, iscsi_root_volume=False):
        """Purge the Credentials by removing folders related to iscsi.

        :param bool iscsi_root_volume: Is the volume the iscsi root
        """
        _log.debug('Purging iscsi credentials on {}'.format(self))
        cmd = 'rm -rf /etc/iscsi/nodes /etc/iscsi/ifaces /etc/iscsi/send_targets'
        self._run_command_check_output(cmd)
        cmd = 'rm -rf /var/lib/iscsi'
        self._run_command_check_output(cmd)
        self._check_iscsi_initramfs(iscsi_root_volume=iscsi_root_volume)

    def run_batch_login(self, batch, svip):
        """Function to run the bash script on clients via ssh.

        This is used inside multiple threads to try and speed up logins.

        :param list batch: Each batch of volumes to login.
        :param str svip: The IP address of the portal.
        :rtype: str
        """
        cmd = """
        pids=()
        for iqn in {}; do
            iscsiadm --mode node --targetname $iqn --portal {}:3260 --login &
            pids[${{#pids[@]}}]=$!
        done
        for pid in "${{pids[@]}}"; do
            wait $pid
        done
        """.format(' '.join(batch), svip)
        result = self._run_command_check_output(cmd)
        return result

    def get_disk_map(self):
        """Get Disk Map after logged into Volumes.

        :rtype: dict
        """
        # Use By ID
        target_map = self._get_target_map()
        disk_map = {}
        counter = 1
        retry_count = 30  # This used to be 72 but that seems excessive, on second thought, not excessive

        # Sometimes the nodes returned include non-active nodes
        # In that case the disk_map will never match the nodes
        # It may be worth trying to delete all nodes not in use, then re-try
        remove_nodes_cmd = 'iscsiadm --mode node --op=delete'

        while counter <= retry_count:
            _log.debug('Counter: {}'.format(counter))
            disk_map = self._list_disks_by_id()
            if set(target_map.keys()) == set(disk_map.keys()):
                break
            time.sleep(10)
            counter += 1
            if counter == retry_count:  # See remove_nodes_cmd comments above
                self.run_command(remove_nodes_cmd)
                target_map = self._get_target_map()
        else:
            raise _errors.IscsiError('Devices {} for all the targets {} are not found'
                                     .format(disk_map, target_map.keys()))

        return disk_map

    def get_initiator_name(self):
        """Get the initiator name from /etc/iscsi/initiatorname.iscsi.

        :rtype: str
        """
        _log.debug('Get Initiator Name from {}'.format(self))
        cmd = 'cat /etc/iscsi/initiatorname.iscsi'
        out_list = self._run_command_check_output(cmd).splitlines()
        init_comp = re.compile('^InitiatorName=(.*)')
        for each_line in out_list:
            init_match = init_comp.search(each_line)
            if init_match:
                return init_match.group(1)
        raise _errors.IscsiError('File /etc/iscsi/initiatorname.iscsi does'
                                 ' not contain Initiator Name on {}'.format(self))

    def _check_iscsi_initramfs(self, iscsi_root_volume=False):
        """Checks the status of /etc/iscsi/iscsi.initramfs.

        This is an indication (though not a guarantee) that we have booted the system with an
        iscsi-mounted root volume.  Stopping the iscsid service will cause the root
        volume to go away under these circumstances, so if we aren't using iscsi
        for the root volume, we should not have this file present.
        Main reason for this is that in open-iscsi rev 873, the existence of this
        file will cause the service to not be stopped, but will exit with code 0.

        :param bool iscsi_root_volume: Is the volume the iscsi root
        """

        if iscsi_root_volume:
            cmd = 'touch /etc/iscsi/iscsi.initramfs'
        else:
            cmd = 'rm -f /etc/iscsi/iscsi.initramfs'
        self._run_command_check_output(cmd)

    def _get_target_map(self):
        """Get the target map of the nodes from the client.

        :rtype: dict
        """
        nodes = self._node_info()
        target_map = {}
        for node in nodes:
            target = node.strip().split()[-1]
            target_map[target.split('.')[-1]] = target
        return target_map

    def _node_info(self):
        """Return the output of [iscsiadm -m node], lists discovered targets.

        :rtype: list[str]
        """
        cmd = 'iscsiadm -m node'
        node_command_out = self._run_command_check_output(cmd) or []
        iqns = node_command_out.splitlines()
        hostname_result = self.run_command('hostname', raise_on_failure=True)
        outlist = hostname_result.stdout.splitlines()
        hostname = outlist[0] if len(hostname_result.stdout) > 0 else 'Unknown'
        hostname = hostname.strip().lower()

        sub_iqns = []
        for iqn in iqns:
            if hostname not in iqn:
                sub_iqns.append(iqn)
            else:
                _log.debug('Skipping boot from san volume {}'.format(iqn))

        filtered_list = list(set(iqns) - set(sub_iqns))
        if filtered_list:
            _log.debug('Boot from san volumes found, ignoring: {}'.format(filtered_list))
        return sub_iqns

    def _list_disks_by_id(self):
        """Lists disks by IDs.

        :rtype: dict
        """
        disk_map = {}
        cmd = 'ls -l /dev/disk/by-path/ 2>/dev/null | grep solidfire'
        run_cmd_output = self.run_command(cmd)
        outlist = run_cmd_output.stdout.splitlines()
        for line in outlist:
            volid = '{}'.format(line.split()[8].split('.')[-1].split('-')[0])
            disk = '/dev/{}'.format(line.split()[10].split('/')[-1])
            disk_map[volid] = disk
        return disk_map

    def _discover(self, portal_ip, portal_port='3260'):
        """Perform iscsi discovery.

        :param str portal_ip: Portal IP address (usually the cluster 10 GB IP)
        :param str portal_port: Portal port to use (default: 3260)
        :rtype: list[str]
        """
        _log.debug('Discovering targets on portal: [{}:{}] on {}...'
                   .format(portal_ip, portal_port, self))
        cmd = 'iscsiadm -m discovery -t sendtargets -p {}:{}'.format(portal_ip, portal_port)
        run_command_output = self._run_command_check_output(cmd, retry=3)
        outlist = run_command_output.splitlines()
        targets = [target.split()[-1] for target in outlist]
        return targets

    def _run_command_check_output(self, cmd, retry=1):
        """Run Command and Check Output of the Command.

        :param str cmd: Command to Run.
        :param int retry: Retry Count

        :rtype: str
        :raises IscsiError: If command was not successful
        """
        _log.debug('running Command {} on {}'.format(cmd, self))
        for _ in range(retry):
            iscsid_return = self.run_command(cmd)
            if iscsid_return.return_code == 0:
                return iscsid_return.stdout
            _log.debug('CMD: {} failed. Retrying.'.format(cmd))
            time.sleep(10)

        # If we got here, the command was not successful
        iscsi_err_msg = _constants.open_iscsi_errors.get(iscsid_return.return_code,
                                                         'Unknown open iscsi error code'),
        err_msg = ('Command [{}] failed! stdout = {}, stderr = {},'
                   ' return_code = {}, iscsi error info: {}'
                   .format(cmd, iscsid_return.stdout, iscsid_return.stderr,
                           iscsid_return.return_code, iscsi_err_msg))

        raise _errors.IscsiError(err_msg)

    @staticmethod
    def _iscsi_set_targets_and_discovery_info(iscsid, auth_user,  # pylint: disable=too-complex
                                              auth_pass, tsecret=None):
        """Set iscsi discover credentials in iscsid.conf.

        :param list[str] iscsid: List of iscsi.conf lines
        :param str auth_user: Username of SF Account
        :param str auth_pass: Initiator Secret of SF Account
        :param str tsecret: Target Secret of SF Account
        :return: The updated iscsid.conf lines
        :rtype: list[str]
        """
        iscsidout = []
        for line in iscsid:
            if 'discovery.sendtargets.auth.username =' in line:
                line = 'discovery.sendtargets.auth.username = {0}\n'.format(auth_user)
            elif 'discovery.sendtargets.auth.password =' in line:
                line = 'discovery.sendtargets.auth.password = {0}\n'.format(auth_pass)
            elif 'node.session.auth.username =' in line:
                line = 'node.session.auth.username = {0}\n'.format(auth_user)
            elif 'node.session.auth.password =' in line:
                line = 'node.session.auth.password = {0}\n'.format(auth_pass)
            if tsecret:
                if 'discovery.sendtargets.auth.username_in =' in line:
                    if line.startswith('#'):
                        _log.debug('Activating Double Authentication from inactive'
                                   ', are you sure you want double authentication?')
                    line = 'discovery.sendtargets.auth.username_in = {0}\n'.format(auth_user)
                elif 'discovery.sendtargets.auth.password_in =' in line:
                    line = 'discovery.sendtargets.auth.password_in = {0}\n'.format(tsecret)
                elif 'node.session.auth.username_in =' in line:
                    if line.startswith('#'):
                        _log.debug('Activating Double Authentication from inactive'
                                   ', are you sure you want double authentication?')
                    line = 'node.session.auth.username_in = {0}\n'.format(auth_user)
                elif 'node.session.auth.password_in =' in line:
                    line = 'node.session.auth.password_in = {0}\n'.format(tsecret)
            else:
                if line.startswith('discovery.sendtargets.auth.username_in ='):
                    _log.debug('Deactivating target Chap username for iscsi discovery session')
                    line = '#' + line
                elif line.startswith('discovery.sendtargets.auth.password_in ='):
                    _log.debug('Deactivating target Chap password for iscsi discovery session')
                    line = '#' + line
                elif line.startswith('node.session.auth.username_in ='):
                    _log.debug('Deactivating target Chap username for iscsi node session')
                    line = '#' + line
                elif line.startswith('node.session.auth.password_in ='):
                    _log.debug('Deactivating target Chap password for iscsi node session')
                    line = '#' + line
            iscsidout.append(line)
        return iscsidout

    def _get_client_iscsid_conf(self):
        """Get the contents of the client's /etc/iscsi/iscsid.conf file.

        :return: A list of lines for the iscsid.conf file from the client with newlines preserved.
        :rtype: list[str]
        """
        _log.debug('Getting iscsid.conf file from client (name: {}, ip: {})'.format(
            self.client_dm.client_name,
            self.mip))

        # get the iscsid.conf file
        iscsid = self.run_command('cat {}/iscsid.conf'.format(ISCSI_CONFIG_DIR),
                                  raise_on_failure=True).stdout
        _log.debug('iscsd.conf file contents: {}'.format(iscsid))

        # Pass True to splitlines to keep line endings
        return iscsid.splitlines(True)

    def _put_iscsid_conf(self, iscsid_lines):
        """Write out a list of strings to /etc/iscsi/iscsid.conf.

        :param list[str] iscsid_lines: New contents of iscsid file as a list of strings.
        """
        with tempfile.NamedTemporaryFile() as tmp_file:
            tmp_file.write(''.join(iscsid_lines).encode())
            tmp_file.flush()
            with self.get_ssh_connection() as ssh:
                ssh.put([tmp_file.name], ISCSI_CONFIG_DIR)
                # Rename the file
                self.run_command('mv {iscsi_dir}/{tmp_name} {iscsi_dir}/iscsid.conf'.format(
                    iscsi_dir=ISCSI_CONFIG_DIR,
                    tmp_name=os.path.basename(tmp_file.name)))

    @property
    def version(self):
        return self.client_dm.os

    def _run_cmd_validate_return(self, command, options='', expected_return_code=0):
        """Calls command with specified options and validates return code.

        :param str command: The command to run
        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        cmd = '{} {}'.format(command, options)
        output = self.run_command(cmd)
        _log.debug(output.stdout)
        if output.stderr:
            _log.debug(output.stderr)
        _log.debug('Return code: {}'.format(output.return_code))

        if output.return_code != expected_return_code:
            msg = ('Cmd [{}] on client {} return code did not match expected {}'
                   .format(cmd, self.mip, expected_return_code))
            raise _errors.FunctionalClientError(msg)

        return output

    def sg_raw(self, options='', expected_return_code=0):
        """Call sg_raw with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('sg_raw', options, expected_return_code)

    def sginfo(self, options='', expected_return_code=0):
        """Calls sginfo with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('sginfo', options, expected_return_code)

    def sg_vpd(self, options='', expected_return_code=0):
        """Calls sg_vpd with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('sg_vpd', options, expected_return_code)

    def sg_persist(self, options='', expected_return_code=0):
        """Calls sg_persist with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('sg_persist', options, expected_return_code)

    def sg_read(self, options='', expected_return_code=0):
        """Calls sg_read with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('sg_read', options, expected_return_code)

    def sg_map(self, options='', expected_return_code=0):
        """Calls sg_map with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('sg_map', options, expected_return_code)

    def multipath(self, options='', expected_return_code=0):
        """Calls multipath with specified options and validates return code.

        :param str options: The options to pass in to the command
        :param int expected_return_code: Expected return code from the command
        :return: The output from the command
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        return self._run_cmd_validate_return('multipath', options, expected_return_code)

    def get_fc_drives(self):
        """Get list of Fibre Channel drives on client.

        :return: List of Fibre Channel drives
        :rtype: list[str]

        :raises valence._errors.Error: If no SolidFire Fibre Channel drives are found
        """
        _log.debug('Getting Fibre Channel hosts.')
        output = self.run_command('ls /sys/class/fc_host')
        hosts = output.stdout.splitlines()

        _log.debug('Get drive mapping.')
        output = self.sg_map('-x -i')
        drive_map = output.stdout.splitlines()

        drives = []
        for host in hosts:
            for drive in drive_map:
                disk = drive.split()
                if 'SolidFir' in drive and host[-1] in disk[1]:
                    drives.append(disk[0])

        if not drives:
            raise _errors.Error('No Fibre Channel drives found on client {}.'.format(
                self.client_dm.mip))

        return drives

    def start_kernel_trace(self, file_output='/tmp/trace_data', event=None, event_type=None):
        """Start client kernel tracing process.

        Sets up and starts background kernel tracing and redirects the output to a file, used
        for tracing kernel function calls for debugging. The file is setup with a 4MB buffer
        size which wraps when full.

        :param str file_output: Name of the output file for redirected trace data
        :param str event: Name of trace event
        :param str event_type: Name of trace event type. If None, all event types will be captured

        :raises valence._errors.Error: If the event or event_type is not found on the client
        """

        trace_dir = '/sys/kernel/debug/tracing'

        output = self.run_command('ls {}/events'.format(trace_dir), raise_on_failure=True)
        client_events = output.stdout.splitlines()
        if event not in client_events:
            raise _errors.Error('{} is not available on client {}.'.format(event,
                                                                           self.client_dm.mip))
        # '*' captures all types for the given field, so this check can be skipped
        # alternatively, a blank type also captures all types
        if event_type and event_type != '*':
            output = self.run_command('ls {}/events/{}'.format(trace_dir, event), raise_on_failure=True)
            client_event_types = output.stdout.splitlines()
            if event_type not in client_event_types:
                raise _errors.Error('{}:{} is not available on client {}.'.format(
                    event, event_type, self.client_dm.mip))

        self.run_command('echo 4096 > {}/buffer_size_kb'.format(trace_dir), raise_on_failure=True)
        self.run_command('echo {}:{} > {}/set_event'.format(event, event_type or '', trace_dir),
                         raise_on_failure=True)
        self.run_command('echo 1 > {}/tracing_on'.format(trace_dir), raise_on_failure=True)

        _log.debug('Starting trace process on client {}.'.format(self.client_dm.mip))
        self.run_command('cat {}/trace_pipe > {} &'.format(trace_dir, file_output),
                         raise_on_failure=True)

    def stop_kernel_trace(self):
        """Stop client tracing process."""

        trace_dir = '/sys/kernel/debug/tracing'
        _log.debug('Stopping trace process on client {}.'.format(self.client_dm.mip))
        self.run_command('echo " " > {}/set_event'.format(trace_dir), raise_on_failure=True)
        # the trace process will automatically stop when 0 is echoed into tracing_on
        self.run_command('echo 0 > {}/tracing_on'.format(trace_dir), raise_on_failure=True)

    #########################################################################################
    # Support bundle methods
    #########################################################################################

    def pull_support_bundle(self, artifact_manager):
        """Pull support bundle for the client and store on artifact manager.

        :param sf_artifacts.ArtifactManager artifact_manager: Artifact manager object that defines
            where the support bundle will be stored.
        """
        bundle_info = self.create_bundle()
        names = artifact_manager.store_artifacts_remote(
            artifacts=[bundle_info.bundle_path], remote_host_dm=self.client_dm
        )
        self._cleanup_remote(bundle_info.remote_rm_paths)
        destination = artifact_manager.artifact_url + list(names)[0]
        _log.debug(self._get_sf_msg('Support bundle collected successfully'
                                    ' and located at: {}'.format(destination)))

    @staticmethod
    def _get_protocol():
        """Get type of Connection.

        :rtype: str
        """
        return ISCSI_CONNECTION

    def _get_support_bundle_command_path(self, bundle_path):
        """Get support bundle command.

        :param str bundle_path: Path to the Bundle
        :rtype: str
        """
        command = 'create_support_bundle.py -o {output_dir} -p {protocol} -c {compression}'
        return command.format(output_dir=bundle_path,
                              protocol=self._get_protocol(),
                              compression=self.COMPRESSION_TYPE)

    def _get_client_bundle_path(self):
        """Get path to client bundle.

        :rtype: str
        """
        timestamp = time.strftime('%m%d%Y_%H%M%S')
        # client bundles are pulled to the local system
        bundle_name = '{}-{}-Client'.format(self.client_dm.client_name, timestamp)
        bundle_path = os.path.join('/tmp', bundle_name)
        return bundle_path

    def create_bundle(self):
        """Create a support bundle for the client.

        :rtype: valence.functional_models.BundleInfo
        """
        bundle_info = _bundle_base.BundleInfo()
        bundle_path = self._get_client_bundle_path()
        new_bundle_path = None
        _log.debug(self._get_sf_msg('Bundle Name to be used: {}'.format(bundle_path)))
        command = self._get_support_bundle_command_path(bundle_path)
        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command(command, timeout=3600)
            if result.return_code != 0:
                if result.return_code == -99:
                    _log.error('Timed out after waiting an hour for support bundle creation. '
                               'You will either need to clean up bundles more often '
                               'or SET will have to increase the timeout.')
                error_msg = self._get_sf_msg('Script failed, STDOUT: {} STDERROR: {}'
                                             .format(result.stdout, result.stderr))
                raise _errors.ClientBundleError(error_msg)
            else:
                new_bundle_path = '{}.tar.{}'.format(bundle_path, self.COMPRESSION_TYPE)

        bundle_info.bundle_path = new_bundle_path
        bundle_info.remote_rm_paths.add(new_bundle_path)
        return bundle_info

    #########################################################################################
    # Certificate related methods
    #########################################################################################
    def create_temp_cert_dir(self):
        """Create a temporary directory to store certs and configuration files.

        :raises: FunctionalClientError if there is any error or rc != 0

        :rtype: str
        """
        run_command_output = self.run_command('mktemp -t test_certs.XXXXXXXX -d', raise_on_failure=True)
        std_out = run_command_output.stdout.splitlines()
        std_err = run_command_output.stderr.splitlines()
        try:
            cert_dir = std_out[0]
        except IndexError:
            raise _errors.FunctionalClientError('Create directory failed on client stdout={} '
                                                'stderr={}'.format(std_out, std_err))
        return cert_dir

    def create_file(self, target_dir, filename, raw):
        """Create a local file on a client.

        :param str target_dir: directory to create the file in
        :param str filename: file name
        :param str raw: file contents, will cast to bytes
        """
        suffix, prefix = filename.split('.')[:2]
        with self.get_ssh_connection() as ssh_con:
            with tempfile.NamedTemporaryFile(suffix=suffix, prefix=prefix) as temp_file:
                temp_file.write(raw.encode('utf-8'))
                temp_file.flush()
                target_path = os.path.join(target_dir, filename)
                ssh_con.put([temp_file.name], target_path)

    def get_file_contents(self, file_directory, filename, timeout):
        """Cat a file on a client, return the response joined by newline.

        Why?  Because cert components need to have keys and certs off the clients.

        :param str file_directory: directory the file is in
        :param str filename: file name to cat
        :param int timeout: timeout for ssh command

        :return: string of the file
        :rtype: str
        """
        command_output = self.run_command('cat {}'.format(os.path.join(file_directory, filename)),
                                          timeout=timeout, raise_on_failure=True)
        return command_output.stdout

    def get_cert_from_port(self, ip, port=442):
        """Get the certificate from an ip and port.  Containerized execution should make this obsolete.

        :param str ip: ip to get certificate off of
        :param int port: port to get certificate off of, default to 442
        :return str: string of the ssl certificate from the node

        :rtype: str
        """
        cmd = ['openssl', 's_client', '-connect', '{}:{}'.format(ip, port), '-showcerts', '-prexit']
        with self.get_ssh_connection() as ssh_inst:
            std_out = ssh_inst.run_command(' '.join(cmd), timeout=600, raise_on_failure=True).stdout
        # + 25 characters to strip the end of the string and return only the certificate
        certificate = std_out[std_out.find('-----BEGIN'):std_out.find('-----END') + 25]
        # Why add a newline?  Cause paramiko appears to strip off a final newline.
        return certificate + '\n'
