"""Cluster FM methods related to accounts."""
from sf_api import (
    api_constants,
    enums,
    result_models,
)
from valence import (
    criteria,
    models,
)
from valence.functional_models._cluster import base


class AccountMixin(base.ClusterBase):
    """Cluster FM Account methods."""

    def create_account(self, username, initiator_secret=None, target_secret=None, attributes=None):
        """Creates an Account on the cluster via AddAccount.

        :param str username: Account username, must be unique.
        :param str initiator_secret: Account iSCSI initiator secret.
        :param str target_secret: Account iSCSI target secret.
        :param dict attributes: Any additional key value pairs to be added to the Account.
        :return: Account data model
        :rtype: valence.models.Account
        """
        # Create the Account on the cluster
        params = {
            api_constants.USERNAME: username,
            api_constants.INITIATOR_SECRET: initiator_secret,
            api_constants.TARGET_SECRET: target_secret,
            api_constants.ATTRIBUTES: attributes
        }
        response = self.api.call_method(enums.ClusterMethod.add_account, params=params,
                                        result_dm_class=result_models.AddAccount)

        # Get the Account's data as created on the cluster. As of Fluorine, this data is in an object
        account_dm = response.result_dm.account
        account_id = response.result_dm.account_id
        if not account_dm:
            account_dm = self.get_account_by_id(account_id)
        return account_dm

    def list_accounts(self, account_name=None, account_status=None, account_ids=None,
                      include_storage_containers=True):
        """Returns account data models for each Account on a cluster via ListAccounts.

        The account_name and account_status parameters are mutually exclusive and cannot be used
        together.

        :param str account_name: Name of the Account to be returned
        :param sf_api.AccountStatus account_status: Return all Accounts with this status
        :param list[int] account_ids: Return all Accounts for these IDs
        :param bool include_storage_containers: Accounts that are storage containers are included by
                                                Option to exclude Accounts which are storage containers,
        :return: List of account data models
        :rtype: list[valence.models.Account]
        """
        if account_name and account_status:
            raise ValueError('account_name and account_status are mutually exclusive')

        params = {
            api_constants.INCLUDE_STORAGE_CONTAINERS: include_storage_containers,
        }
        response = self.api.call_method(enums.ClusterMethod.list_accounts, params=params,
                                        result_dm_class=result_models.ListAccounts)
        all_accounts = response.result_dm.accounts

        # Return all Accounts if no filter parameters were specified
        if not account_name and not account_status and not account_ids and include_storage_containers:
            return all_accounts

        # Put account_ids into a set to improve efficiency
        account_ids = set(account_ids) if account_ids else None

        account_filter = criteria.And(
            criteria.Equals(models.Account.username, account_name),
            criteria.Equals(models.Account.status, account_status if account_status else None),
            criteria.ContainedIn(models.Account.account_id, account_ids),
        )
        if not include_storage_containers:
            account_filter.add_criteria(criteria.Equals(models.Account.storage_container_id,
                                                        '00000000-0000-0000-0000-000000000000'))
        return_accounts = account_filter.filter(all_accounts)
        return return_accounts

    def modify_account(self, account_id, status=None, initiator_secret=None, target_secret=None,
                       attributes=None, username=None):
        """Call ModifyAccount to update account matching provided account_id.

        :param int account_id: Id of Account to modify.
        :param sf_api.enums.AccountStatus status: enum indicating status: active, locked, or removed.
        :param str initiator_secret: 12 to 16 chars, cannot be the same as the other secret.
        :param str target_secret: 12 to 16 chars, cannot be the same as the other secret.
        :param dict attributes: Key/Value pairs.
        :param str username: Must remain unique for each Account.
        """
        params = {
            api_constants.ACCOUNT_ID: account_id,
            api_constants.STATUS: status,
            api_constants.INITIATOR_SECRET: initiator_secret,
            api_constants.TARGET_SECRET: target_secret,
            api_constants.ATTRIBUTES: attributes,
            api_constants.USERNAME: username
        }
        self.api.call_method(enums.ClusterMethod.modify_account, params=params,
                             result_dm_class=result_models.ModifyAccount)

    def delete_accounts(self, account_dms):
        """Deletes each Account provided on the cluster via RemoveAccount.

        Each account in the data model list is deleted individually.

        :param typing.Iterable[valence.models.Account] account_dms: Account dm's to create
        """
        for account_dm in account_dms:
            params = {api_constants.ACCOUNT_ID: account_dm.account_id}
            self.api.call_method(enums.ClusterMethod.remove_account,
                                 params=params, result_dm_class=None)

    def get_account_by_id(self, account_id):
        """Returns Account data model for an account_id.

        :param int account_id: ID of the account to be returned
        :return: Account data model
        :rtype: valence.models.Account
        """
        params = {api_constants.ACCOUNT_ID: account_id}
        response = self.api.call_method(enums.ClusterMethod.get_account_by_id,
                                        params=params,
                                        result_dm_class=result_models.GetAccountById)
        return response.result_dm.account
