"""Cluster FM methods related to async."""
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    models,
)
from valence.functional_models import _errors
from valence.functional_models._cluster import base


class AsyncMixin(base.ClusterBase):
    """Cluster FM Async methods."""

    def get_async_result(self, async_handle_dm, keep_result=True):
        """Get an AsyncResult from an AsyncHandle via GetAsyncResult.

        This returns the Valence AsyncResult which is different from SF's GetAsyncResult
        and ListAsyncResult result_models.

        :param async_handle_dm: Async handle to query
        :type async_handle_dm: valence.models.AsyncHandle
        :param bool keep_result: Prevent result deletion if complete
        :rtype: valence.models.AsyncResult
        """
        async_handle_id = async_handle_dm.async_handle

        params = {
            api_constants.ASYNC_HANDLE: async_handle_dm.async_handle,
            api_constants.KEEP_RESULT: keep_result
        }
        response = self.call_method(enums.ClusterMethod.get_async_result, params=params,
                                    result_dm_class=result_models.GetAsyncResult)

        # The GetAsyncResult response does not have the ID, add it to the AsyncResult dm and validate
        dm = models.AsyncResult(response.result_dm.serialize())
        dm.async_result_id = async_handle_id
        if self._validate_responses:
            dm.validate(self.cluster_dm.api_endpoint)
        return dm

    def list_async_results(self, async_result_types=None):
        """Get a list of AsyncResults via ListAsyncResults.

        This returns a list of Valence AsyncResults which is different from SF's GetAsyncResult
        and ListAsyncResult result_models.

        :param list[str] async_result_types: Optional list of types of results; DriveRemoval, Clone, ext
        :rtype: list[valence.models.AsyncResult]
        """
        response = self.call_method(enums.ClusterMethod.list_async_results,
                                    params={api_constants.ASYNC_RESULT_TYPES: async_result_types},
                                    result_dm_class=result_models.ListAsyncResults)
        dms = []

        # Create the AsyncResult for each async handle in the ListAsyncResults result dm
        for dm in response.result_dm.async_handles:
            async_result_dm = models.AsyncResult(dm.serialize())

            # ListAsyncResults does not have a status...
            if not dm.completed and not dm.success:
                async_result_dm.status = enums.AsyncResultStatus.running
            if not dm.completed and dm.success:
                msg = 'Async result cannot be incomplete and marked as success'
                raise _errors.FunctionalClusterError(msg)
            if dm.completed and not dm.success:
                async_result_dm.status = enums.AsyncResultStatus.failed
            if dm.completed and dm.success:
                async_result_dm.status = enums.AsyncResultStatus.complete

            if self._validate_responses:
                async_result_dm.validate(self.cluster_dm.api_endpoint)

            dms.append(async_result_dm)

        return dms
