"""Cluster FM methods related to authentication sessions."""
import datetime

from sf_api import (
    api_constants,
    enums,
    result_models,
)
from valence import utils
from valence.functional_models._cluster import base


class AuthSessionMixin(base.ClusterBase):
    """Cluster FM Authentication Session methods."""

    def create_auth_session(self, cluster_admin, desired_expiration_duration=None):
        """Creates an Authentication Session on the cluster via CreateAuthSession.

        :param valence.models.ClusterAdmin cluster_admin: Cluster Admin associated with the auth session
        :param datetime.timedelta desired_expiration_duration: Optional time delta from 'now' which will
            defines the desired time at which the auth session will definitely expire.  This value may
            or may not be honored.
        :return: Authentication Session Info data model
        :rtype: valence.models.AuthSessionInfo
        """

        # Convert to seconds since epoch which is what the API expects.
        if desired_expiration_duration:
            epoch = datetime.datetime(1970, 1, 1)
            epoch_value = utils.get_now_utc() - epoch + desired_expiration_duration
            desired_expiration_duration = str(int(epoch_value.total_seconds()))

        # Create the Authentication Session on the cluster
        params = {
            api_constants.CLUSTER_ADMIN_ID: cluster_admin.cluster_admin_id,
            api_constants.DESIRED_EXPIRATION_DURATION: desired_expiration_duration
        }
        response = self.api.call_method(enums.ClusterMethod.create_auth_session, params=params,
                                        result_dm_class=result_models.CreateAuthSessionResult)

        return response.result_dm.session

    def delete_auth_session(self, cluster_admin, session=None):
        """Delete an Authentication Session on the cluster via DeleteAuthSessions.

        .. note:: This method does not validate that an individual session to be deleted matches the
         provided cluster_admin.

        :param valence.models.ClusterAdmin cluster_admin: Cluster Admin associated with the session(s).
        :param valence.models.AuthSessionInfo session: the auth session to be deleted.
        :return: Authentication Session Info data models deleted
        :rtype: list[valence.models.AuthSessionInfo]
        """
        # Delete Authentication Sessions on the cluster
        session_id = str(session.session_id) if session else None

        params = {
            api_constants.CLUSTER_ADMIN_ID: cluster_admin.cluster_admin_id,
            api_constants.SESSION_ID: session_id,
        }
        response = self.api.call_method(enums.ClusterMethod.delete_auth_sessions, params=params,
                                        result_dm_class=result_models.DeleteAuthSessionsResult)

        return response.result_dm.sessions

    def list_auth_sessions(self, cluster_admins=None):
        """List an Authentication Session on the cluster via ListAuthSessions.

        :param list[valence.models.ClusterAdmin] cluster_admins: Filters the result to those auth
            sessions matching any of the cluster admins.
        :return: Authentication Session Info data models
        :rtype: list[valence.models.AuthSessionInfo]
        """
        # List Authentication Sessions on the cluster, the cluster_admin_ids parameter is required, with
        # an empty list being the indicator to list all sessions.
        if cluster_admins:
            cluster_admin_ids = [cluster_admin.cluster_admin_id for cluster_admin in cluster_admins]
        else:
            cluster_admin_ids = []

        params = {
            api_constants.CLUSTER_ADMIN_IDS: cluster_admin_ids,
        }
        response = self.api.call_method(enums.ClusterMethod.list_auth_sessions, params=params,
                                        result_dm_class=result_models.ListAuthSessionsResult)

        return response.result_dm.sessions

    def list_expired_auth_sessions(self):
        """List Expired Authentication Sessions on the cluster via ListExpiredAuthSessions.

        :return: Authentication Session Info data models that have expired
        :rtype: list[valence.models.AuthSessionInfo]
        """
        response = self.api.call_method(enums.ClusterMethod.list_expired_auth_sessions,
                                        result_dm_class=result_models.ListAuthSessionsResult)

        return response.result_dm.sessions

    def update_auth_session(self, session):
        """Update an Authentication Session on the cluster via UpdateAuthSession.

        .. note:: This method does not validate that an individual session to be updated matches the
         provided cluster_admin.

        :param valence.models.AuthSessionInfo session: the auth session to be updated.
        :return: Authentication Session Info data model updated
        :rtype: valence.models.AuthSessionInfo
        """
        # Update Authentication Session on the cluster
        params = {
            api_constants.SESSION_ID: str(session.session_id),
        }
        response = self.api.call_method(enums.ClusterMethod.update_auth_session, params=params,
                                        result_dm_class=result_models.UpdateAuthSessionResult)

        return response.result_dm.session
