"""Base class for Cluster functional model."""
import threading

import autologger
import requests

import sf_api
from sf_api import api_constants
from valence import (
    enums,
    models,
)
from valence.functional_models import _errors

_log = autologger.get_sflogging()
# Lock for setting the upgrading attribute
_upgrade_setter_lock = threading.Lock()


class ClusterBase(object):
    """Cluster functional model base class.

    :param valence.models.ResourceCluster cluster_dm: cluster data model
    :param bool authenticate: If True, check if cluster can be authenticated via HTTP in constructor.
    """

    # A dict to keep track of which clusters are currently upgrading.
    # Has the form: {<static_id>: <bool>}
    _upgrading_status = {}

    def __init__(self, cluster_dm, authenticate=False):
        # The following instance variables are set in __init__ as opposed to __new__ since we want them
        # to be overridden each time the cluster_fm is instantiated.
        self.cluster_dm = cluster_dm
        self.api = sf_api.ClusterApi(self.cluster_dm.mvip, self.cluster_dm.username,
                                     self.cluster_dm.password, self.cluster_dm.api_endpoint,
                                     authenticate=False)
        self._validate_responses = True
        # Flag to log the full JSON response from sf_api
        self.log_responses = True
        if self.upgrading is None:
            self.upgrading = False
        if authenticate:
            self.authenticate()

    def __repr__(self):
        return '{}<{}={}>'.format('ClusterFm', 'mvip', self.cluster_dm.mvip)

    @property
    def mvip(self):
        """Shortcut for mvip attribute on cluster model.

        :rtype: str
        """
        return self.cluster_dm.mvip

    @property
    def svip(self):
        return self.cluster_dm.svip

    @property
    def name(self):
        return self.cluster_dm.name

    @property
    def static_id(self):
        """Shortcut for static_id attribute of the cluster.

        :rtype: str
        """
        return self.cluster_dm.mvip

    @property
    def upgrading(self):
        """Determine if the current cluster is upgrading.

        :rtype: bool
        """
        return ClusterBase._upgrading_status.get(self.static_id)

    @upgrading.setter
    def upgrading(self, value):
        """Set the upgrading status of the cluster.

        :param bool value: Value to set upgrading flag to
        """
        with _upgrade_setter_lock:
            ClusterBase._upgrading_status[self.static_id] = value

    def authenticate(self, cluster_dm=None):
        """Verify that cluster can be authenticated via HTTP.

        This method does not need to be called before making any API calls. It simply verifies that
        the authentication credentials are valid. If a cluster data model is given, self.cluster_dm will
        be overridden and the updated credentials will be used.

        :param valence.models.ResourceCluster cluster_dm: cluster data model that will be used to
            override the existing one. This should only be used when re-authenticating a cluster with
            a different cluster admin. This will set self.cluster_dm and self.api.
        """
        if cluster_dm is not None:
            self.cluster_dm = cluster_dm
            self.api = sf_api.ClusterApi(self.cluster_dm.mvip, self.cluster_dm.username,
                                         self.cluster_dm.password, self.cluster_dm.api_endpoint,
                                         authenticate=False)
            models.set_cluster_resources([cluster_dm])
        self.api.authenticate()

    def call_method(self, method, params=None, result_dm_class=None, log_response=None):
        """Call the API.

        :param sf_api.ClusterMethod method: method to call
        :param dict params: the params for the method. Enums will be converted to strings
        :param Any result_dm_class: The API Response class the result is modeled after. If there is no
            result data model associated with the call, this can be None.
        :param bool log_response: Option to log response output. If None, default to the value of
            self.log_responses.
        :rtype: sf_api.SfResponse
        """
        if log_response is None:
            log_response = self.log_responses
        # For calls to API through FM, always use the cluster_fm API version.
        # This version has to be updated using update_api_version() method after upgrade has finished.
        auth = self.api.get_auth()
        auth.version = self.cluster_dm.api_endpoint
        response = self.api.call_method(method, params=params, result_dm_class=result_dm_class,
                                        log_response=log_response, auth=auth)
        if self._validate_responses and response.result_dm is not None:
            # Version validation is disabled during upgrades, but otherwise we have to make sure that
            # cluster_dm.api_endpoint is updated after upgrade is finished.
            version = None if self.upgrading else self.cluster_dm.api_endpoint
            response.result_dm.validate(version)
        return response

    def update_api_version(self):
        """Update this cluster FM and cluster DM with the current version.

        This can be called after running an upgrade to update the version used to call the API.
        """
        response = self.get_api(show_internal=False)
        current_version = response[api_constants.CURRENT_VERSION]  # type: str
        _log.debug('Updating version to {}'.format(current_version))
        current_version = sf_api.SfVersion(current_version)
        # This update is kind of redundant since all calls to API should be doing through the call_method
        # in the FM, which supplies latest version information. But it's still good to keep everything
        # in sync.
        self.api.version = current_version
        self.cluster_dm.api_endpoint = current_version

    def get_api(self, show_internal=True, log_response=False):
        """Returns a list of all customer facing API calls, and optionally include internal calls.

        :param bool show_internal: Option to show internal API calls.
        :param bool log_response: Option to log response output.
        :rtype: dict[str,Any]
        :returns:
            GetAPIResultObject::

                {
                    "6.0": [
                        "AptUpdate",
                        "GetClusterConfig",
                        .........
                       ]
                 }
        """
        return self.call_method(enums.ClusterMethod.get_api, dict(__includeSecretMethods=show_internal),
                                result_dm_class=None, log_response=log_response).result

    def api_supported(self, method):
        """See if the node supports the given API method.

        :param valence.enums.ClusterMethod method: Method to check for support
        :return: True if supported; False if not supported or not applicable
        :rtype: bool
        :raises valence.functional_models.FunctionalClusterError: If the resource cluster is out of sync
            with the actual version the cluster is RTFI'ed to.
        """
        response = self.get_api(show_internal=True, log_response=False)
        current_version = response[api_constants.CURRENT_VERSION]
        version = str(self.api.version)
        # While upgrade is in progress, we don't care about validating API calls per version.
        if str(version) != current_version and not self.upgrading:
            raise _errors.FunctionalClusterError('Resource cluster is version {}, but cluster is at '
                                                 'version {}'.format(version, current_version))

        available_apis = response.get(version)
        if available_apis is not None:
            return method.value in available_apis
        else:
            raise _errors.FunctionalClusterError(
                'Failed to find available APIs for version: {};\n {}'.format(
                    version, response
                )
            )

    def api_is_responsive(self):
        """See if the mvip is responsive.

        :rtype: bool
        :return: True if API is responsive, False if not.
        """
        try:
            self.get_api()
        except (requests.HTTPError, requests.ConnectionError, requests.Timeout):
            return False
        return True

    @classmethod
    def from_resource_data(cls, rs_cluster_data, **kwargs):
        """Factory method for instantiating a cluster FM from the resource data.

        Takes the same kwargs as __init__.

        :param dict rs_cluster_data: Data that corresponds to ResourceCluster
        :return: Cluster FM instance
        :rtype: valence.functional_models.Cluster
        """
        rs_cluster_dm = models.ResourceCluster(rs_cluster_data)
        rs_cluster_dm.validate()
        return cls(rs_cluster_dm, **kwargs)

    @classmethod
    def get_mock_fm(cls):
        """Factory method for instantiating a mock cluster FM.

        No authentication or response validation will be done. This is purely for unit testing purposes.

        :rtype: valence.functional_models.Cluster
        """
        rs_cluster_dm = models.ResourceCluster.get_mock_object()
        instance = cls(rs_cluster_dm, authenticate=False)
        instance._validate_responses = False  # pylint: disable=protected-access
        return instance
