"""Cluster functional model."""
import time

import autologger

import sf_api
from sf_api import (
    api_constants,
    result_models,
    sf_version,
)
from valence import (
    criteria,
    enums,
    errors,
    models,
)
from valence.functional_models import (
    _async_handler,
    _errors,
)
from valence.functional_models._cluster import (
    account_mixin,
    async_mixin,
    auth_session_mixin,
    drive_mixin,
    drum_mixin,
    hardware_mixin,
    initiator_mixin,
    mfa_mixin,
    node_mixin,
    protection_domain_mixin,
    raw_stats_mixin,
    report_mixin,
    schedule_mixin,
    security_mixin,
    snapmirror_mixin,
    snapshot_mixin,
    time_mixin,
    virtual_network_mixin,
    volume_access_group_mixin,
    volume_mixin,
    zk_mixin,
)

_log = autologger.get_sflogging()


class Cluster(
        account_mixin.AccountMixin,
        async_mixin.AsyncMixin,
        auth_session_mixin.AuthSessionMixin,
        drive_mixin.DriveMixin,
        drum_mixin.DrumMixin,
        hardware_mixin.HardwareMixin,
        initiator_mixin.InitiatorMixin,
        mfa_mixin.MfaMixin,
        node_mixin.NodeMixin,
        protection_domain_mixin.ProtectionDomainMixin,
        raw_stats_mixin.RawStatsMixin,
        report_mixin.ReportMixin,
        schedule_mixin.ScheduleMixin,
        security_mixin.SecurityMixin,
        snapmirror_mixin.SnapmirrorMixin,
        snapshot_mixin.SnapshotMixin,
        time_mixin.TimeMixin,
        virtual_network_mixin.VirtualNetworkMixin,
        volume_access_group_mixin.VolumeAccessGroupMixin,
        volume_mixin.VolumeMixin,
        zk_mixin.ZkMixin):
    """Cluster functional model."""

    def get_current_version(self):
        """Get the current version of the cluster.

        .. note:: This is the version as returned by GetAPI, so it is not the full build version.
                  SfVersion('10.0'), not a build version like SFVersion('10.0.0.284').

        :rtype: valence.models.SfVersion
        """
        result = self.get_api(show_internal=False, log_response=False)
        version_string = result[sf_api.api_constants.CURRENT_VERSION]
        return models.SfVersion(version_string)

    def get_cluster_version_info(self):
        """Get information on the element software for the cluster and each node related to it.

        :rtype: valence.models.ClusterVersionInfo
        """
        response = self.api.call_method(enums.ClusterMethod.get_cluster_version_info,
                                        result_dm_class=result_models.GetClusterVersionInfo)
        return response.result_dm

    def add_cluster_admin(self, username, password, access=None, attributes=None):
        """Add a cluster admin.

        Returns a new resource cluster data model with the updated credentials. In order to
        re-authenticate, instantiate a new cluster functional model with the updated data model
        or call cluster_fm.authenticate(cluster_dm), where "cluster_dm" is the new cluster data model.

        :param str username: Unique username for this Cluster Admin
        :param str password: Password used to authenticate this Cluster Admin
        :param list[valence.enums.ClusterAccessLevel] access: Controls which methods this Cluster Admin
            can use. See the Access Control appendix of the API Reference for more details. If not
            given, default to [ClusterAccessLevel.administrator]
        :param dict attributes: Dictionary of attributes
        :rtype: valence.models.ResourceCluster
        """
        params = {
            api_constants.USERNAME: username,
            api_constants.PASSWORD: password,
            api_constants.ACCEPT_EULA: True,
            api_constants.ACCESS: access or [enums.ClusterAccessLevel.administrator],
            api_constants.ATTRIBUTES: attributes
        }
        self.call_method(enums.ClusterMethod.add_cluster_admin, params=params,
                         result_dm_class=result_models.AddClusterAdmin)

        # Make a copy of the cluster data model with updated credentials
        cluster_data = self.cluster_dm.serialize()
        cluster_data[api_constants.USERNAME] = username
        cluster_data[api_constants.PASSWORD] = password
        return models.ResourceCluster(cluster_data)

    def remove_cluster_admin(self, cluster_admin_dm):
        """Remove a cluster admin.

        :param valence.models.ClusterAdmin cluster_admin_dm: Cluster Admin data model to remove.
        """
        params = {
            api_constants.CLUSTER_ADMIN_ID: cluster_admin_dm.cluster_admin_id,
        }
        self.call_method(enums.ClusterMethod.remove_cluster_admin, params=params,
                         result_dm_class=None)

    def get_cluster_capacity(self):
        """Get the cluster capacity via GetClusterCapacity.

        :rtype: valence.models.ClusterCapacity
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_capacity,
                                    result_dm_class=result_models.GetClusterCapacity)
        return response.result_dm.cluster_capacity

    def get_cluster_info(self):
        """Get info about a cluster via GetClusterInfo.

        :rtype: valence.models.Cluster
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_info,
                                    result_dm_class=result_models.GetClusterInfo)
        cluster_dm = response.result_dm.cluster_info
        result = models.Cluster(cluster_dm, resource_dm=self.cluster_dm)
        return result

    def get_max_volume_size(self):
        """Get the maximum supported volume size for the Cluster.

        :rtype: valence.models.DataSize
        :return: Maximum Volume Data Size
        """
        if self.get_current_version() >= sf_version.SfVersion('11.0'):
            return api_constants.SODIUM_MAX_VOLUME_SIZE
        else:
            return api_constants.DEFAULT_MAX_VOLUME_SIZE

    def list_cluster_admins(self, show_hidden=False):
        """List cluster admins.

        :param bool show_hidden: include the hidden admin accounts
        :rtype: list[valence.models.ClusterAdmin]
        """

        response = self.call_method(enums.ClusterMethod.list_cluster_admins,
                                    params={api_constants.SHOW_HIDDEN: show_hidden},
                                    result_dm_class=result_models.ListClusterAdmins)
        return response.result_dm.cluster_admins

    def create_event(self, event_type, message, data=None, node_id=0):
        """Create a cluster event.

        :param valence.enums.ClusterEventType event_type: type of event
        :param str message: event message
        :param dict data: this corresponse to 'details' in the ListEvents response
        :param int node_id: ID of node, use 0 for cluster events.
        :rtype: sf_api.result_models.CreateEvent
        """

        params = {
            api_constants.EVENT_TYPE: event_type,
            api_constants.MESSAGE: message,
            api_constants.DATA: data or {},
            api_constants.NODE_ID: node_id,
        }

        response = self.call_method(enums.ClusterMethod.create_event, params=params,
                                    result_dm_class=result_models.CreateEvent)

        return response.result_dm

    def list_cluster_faults(self, fault_status=None, include_codes=None, exclude_codes=None,
                            drive_ids=None, node_ids=None, update=False, best_practices=None,
                            include_severities=None, service_ids=None, include_cluster_exceptions=False):
        """List cluster faults via ListClusterFaults.

        :param valence.enums.ClusterFaultStatus fault_status: Determines types of faults returned
            `current` - List active, unresolved faults.
            `resolved` - List faults that were previously detected and resolved.
            `all` - List both current and resolved faults. You can see the fault status in the
            "resolved" field of the Cluster Fault object.
        :param list[valence.enums.ClusterFaultCode] include_codes: Include faults with these codes.
            Mutually exclusive with exclude_codes.
        :param list[valence.enums.ClusterFaultCode] exclude_codes: Exclude faults with these codes.
            Mutually exclusive with include_codes.
        :param list[int] drive_ids: Return cluster faults associated with these drives.
        :param list[int] node_ids: Return cluster faults associated with these nodes.
        :param bool update: Option to force a synchronous fault check.
        :param bool best_practices: Include faults triggered by suboptimal config.
        :param bool include_cluster_exceptions: Include special exception type cluster faults.
        :param list[valence.enums.ClusterFaultSeverity] include_severities: Include faults with these
            severities.
        :param list[int] service_ids: Return cluster faults associated with these services.
        :return: Cluster fault data models
        :rtype: list[valence.models.ClusterFault]
        :raises ValueError: When mutually exclusive parameters are provided.
        """
        # TODO: find a way to extend the normal http timeout for this call when update=True
        params = {
            api_constants.FAULT_TYPES: fault_status if fault_status else sf_api.ClusterFaultStatus.all,
            api_constants.BEST_PRACTICES: best_practices,
            api_constants.UPDATE: update,
            api_constants.EXCEPTIONS: include_cluster_exceptions
        }
        response = self.call_method(enums.ClusterMethod.list_cluster_faults, params=params,
                                    result_dm_class=result_models.ListClusterFaults)
        all_faults = response.result_dm.faults

        if include_codes is not None and exclude_codes is not None:
            raise ValueError('include_codes and exclude_codes are mutually exclusive')

        # Make lists in to sets to improve efficiency
        include_codes = set(include_codes) if include_codes else None
        exclude_codes = set(exclude_codes) if exclude_codes else None
        include_severities = set(include_severities) if include_severities else None
        drive_ids = set(drive_ids) if drive_ids else None
        node_ids = set(node_ids) if node_ids else None

        fault_filter = criteria.And(
                criteria.ContainedIn(models.ClusterFault.code, include_codes),
                criteria.ContainedIn(models.ClusterFault.code, exclude_codes, negate=True),
                criteria.ContainedIn(models.ClusterFault.severity, include_severities),
                criteria.ContainedIn(models.ClusterFault.drive_id, drive_ids),
                criteria.ContainedIn(models.ClusterFault.node_id, node_ids),
                criteria.ContainedIn(models.ClusterFault.service_id, service_ids)
        )
        filtered_faults = fault_filter.filter(all_faults)
        return filtered_faults

    def clear_cluster_faults(self):
        """Clear cluster faults on the system.

        By default we only clear resolved faults, to clear other faults use sf_api directly.
        """
        self.call_method(enums.ClusterMethod.clear_cluster_faults, params={},
                         result_dm_class=None)

    def list_events(self, max_events=None, start_event_id=None, end_event_id=None, event_type=None,
                    queue_type=None, service_id=None, node_id=None, drive_id=None,
                    start_report_time=None, end_report_time=None, start_publish_time=None,
                    end_publish_time=None):
        """Returns events detected on the cluster, sorted from newest to oldest.

        .. Note:: The event log is limited to the last 1500 messages.

        :param int max_events: Maximum number of events to list
        :param int start_event_id: The ID to start listing from
        :param int end_event_id: The ID to end listing from
        :param valence.enums.ClusterEventType event_type: The type of event to filter on.
        :param valence.enums.EventQueueType queue_type: The type of event queue to filter on.
        :param int service_id: The serviceID to filter on
        :param int node_id: The nodeID to filter on
        :param int drive_id: A DriveID to filter on
        :param datetime.datetime start_report_time: The start time to start listing from in UTC
        :param datetime.datetime end_report_time: The end time to end listing from in UTC
        :param datetime.datetime start_publish_time: The publish time to start listing from in UTC
        :param datetime.datetime end_publish_time: The publish time to end listing from in UTC
        :rtype: list[valence.models.ClusterEvent]
        """
        dt_fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
        if start_report_time:
            start_report_time = start_report_time.strftime(dt_fmt)
        if end_report_time:
            end_report_time = end_report_time.strftime(dt_fmt)
        if start_publish_time:
            start_publish_time = start_publish_time.strftime(dt_fmt)
        if end_publish_time:
            end_publish_time = end_publish_time.strftime(dt_fmt)

        params = {'maxEvents': max_events,
                  'startEventID': start_event_id,
                  'endEventID': end_event_id,
                  'eventType': event_type,
                  'eventQueueType': queue_type,
                  api_constants.SERVICE_ID: service_id,
                  api_constants.NODE_ID: node_id,
                  api_constants.DRIVE_ID: drive_id,
                  'startReportTime': start_report_time,
                  'endReportTime': end_report_time,
                  'startPublishTime': start_publish_time,
                  'endPublishTime': end_publish_time,
                  }
        response = self.call_method(enums.ClusterMethod.list_events, params=params,
                                    result_dm_class=result_models.ListEvents)
        return response.result_dm.events

    def list_sync_jobs(self, node_ids=None, slice_ids=None, job_type=None, job_stage=None):
        """List synchronization jobs that are running on a cluster via ListSyncJobs.

        :param list[int] node_ids: Return sync jobs associated with these node_ids.
        :param list[int] slice_ids: Return sync jobs associated with these slice_ids.
        :param valence.enums.SyncJobType job_type: Return sync jobs associated with this type.
        :param valence.enums.SyncJobStage job_stage: Return sync jobs associated with this stage.
        :return: List of SyncJob data models.
        :rtype: list[valence.models.SyncJob]
        """
        response = self.call_method(enums.ClusterMethod.list_sync_jobs,
                                    result_dm_class=result_models.ListSyncJobs)
        all_sync_jobs = response.result_dm.sync_jobs

        # Return all jobs is no filter parameters were specified
        if not node_ids and not slice_ids and not job_type and not job_stage:
            return all_sync_jobs

        # Put node_ids into set to improve efficiency
        node_ids = set(node_ids) if node_ids else None

        sync_jobs_filter = criteria.And(
                criteria.ContainedIn(models.SyncJob.node_id, node_ids),
                criteria.ContainedIn(models.SyncJob.slice_id, slice_ids),
                criteria.Is(models.SyncJob.type, job_type),
                criteria.Is(models.SyncJob.stage, job_stage)
        )

        return_sync_jobs = sync_jobs_filter.filter(all_sync_jobs)
        return return_sync_jobs

    def get_constants(self, changed_only=False, constants_dm=None):
        """Return all constants for cluster via GetConstants.

        :param bool changed_only: If True, will only return values changed from defaults
        :param valence.models.ClusterConstants constants_dm: If not None, which constants to get
        :return: Return all constants for cluster
        :rtype: valence.models.ClusterConstants
        """
        if not constants_dm:
            constants_dm = models.ClusterConstantsVerbose.create()

        params = {'constants': list(constants_dm.serialize().keys())}
        if changed_only:
            params['changedOnly'] = True

        response = self.call_method(enums.ClusterMethod.get_constants, params=params,
                                    result_dm_class=result_models.GetConstants)
        return response.result_dm

    def set_constants(self, constants_dm, force=False):
        """Set cluster constants using SetConstants.

        Strips away all the constants with a value of none before applying it

        :param valence.models.ClusterConstants constants_dm: The cluster constants data model to set.
        :param bool force: If true, sets a cluster constant that may not exist in this version.
        :return: Return all constants for cluster
        :rtype: valence.models.ClusterConstants
        """
        params = constants_dm.serialize()

        if force:
            params['force'] = True

        response = self.call_method(enums.ClusterMethod.set_constants, params=params,
                                    result_dm_class=result_models.SetConstants)
        return response.result_dm

    def reset_constants(self, constants_dm=None, reset_all=False):
        """Reset cluster API constants.

        :param valence.models.ClusterConstants constants_dm: The cluster constants data model to set.
        :param bool reset_all: Reset all changed constants.  Data model provided is ignored.
        :return: Return all constants for cluster
        :rtype: valence.models.ClusterConstants
        """
        if reset_all:
            params = {'resetAll': True}
        else:
            if not constants_dm:
                raise ValueError('Constants to reset must be provided or reset all specified.')
            params = {'constants': list(constants_dm.serialize().keys())}

        response = self.api.call_method(enums.ClusterMethod.reset_constants, params=params,
                                        result_dm_class=result_models.ResetConstants)
        return response.result_dm

    def get_limits(self):
        """Return all cluster limits.

        :rtype: valence.models.ClusterLimits
        """
        response = self.call_method(enums.ClusterMethod.get_limits,
                                    result_dm_class=result_models.GetLimits)
        return response.result_dm

    def get_immutable_values(self):
        """Return all immutable values for the cluster.

        :rtype: sf_api.result_models.GetImmutableValues
        """
        if self.api_supported(enums.ClusterMethod.get_immutable_values):
            response = self.call_method(enums.ClusterMethod.get_immutable_values,
                                        result_dm_class=result_models.GetImmutableValues)
            return response.result_dm
        else:
            return result_models.GetImmutableValues()

    def get_service_status(self):
        """Get the status of services on the cluster.

        :rtype: sf_api.result_models.GetServiceStatus
        """
        response = self.call_method(enums.ClusterMethod.get_service_status,
                                    result_dm_class=result_models.GetServiceStatus)
        return response.result_dm

    def list_services(self, service_types=None, node_dms=None, drive_dms=None):
        """List all services on the cluster.

        :param list[valence.enums.ServiceType] service_types: Service types on which to filter.
        :param list[valence.models.ActiveNode] node_dms: Nodes on which to filter.
        :param list[valence.models.Drive] drive_dms: Drives on which to filter.
        :rtype: list[sf_api.models.ClusterServiceInfo]
        """
        node_ids = None if node_dms is None else [dm.node_id for dm in node_dms]
        drive_ids = None if drive_dms is None else [dm.drive_id for dm in drive_dms]
        response = self.call_method(enums.ClusterMethod.list_services,
                                    result_dm_class=result_models.ListServices)
        cluster_service_info = response.result_dm.services
        all_services = [s.service for s in cluster_service_info]
        services_filter = criteria.And(
            criteria.ContainedIn(models.ClusterService.service_type, service_types),
            criteria.ContainedIn(models.ClusterService.node_id, node_ids),
            criteria.Intersection(models.ClusterService.drive_ids, drive_ids)
        )
        filtered_service_ids = [s.service_id for s in services_filter.filter(all_services)]
        return [s for s in cluster_service_info if s.service.service_id in filtered_service_ids]

    def get_cluster_full_threshold(self):
        """Get the cluster full threshold on the cluster.

        :rtype: valence.models.ClusterFullThreshold
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_full_threshold,
                                    result_dm_class=result_models.GetClusterFullThreshold)
        return response.result_dm

    def modify_cluster_full_threshold(self, stage2_aware_threshold=None,
                                      stage3_block_threshold_percent=None,
                                      stage3_metadata_threshold_percent=None,
                                      max_metadata_overprovision_factor=None):
        """Modify cluster full threshold.

        :param int stage2_aware_threshold: The number of nodes of capacity remaining in the cluster
                                           before the system triggers a capacity notification.

        :param int stage3_block_threshold_percent: The percentage of block storage utilization below the
                                                   'Error' threshold that causes the system to trigger
                                                   a cluster 'Warning' alert.

        :param int stage3_metadata_threshold_percent: The percentage of metadata storage utilization
                                                      below the 'Error' threshold that causes the
                                                      system to trigger a cluster 'Warning' alert.

        :param int max_metadata_overprovision_factor: A value representative of the number of times
                                                      metadata space can be overprovisioned relative
                                                      to the amount of space available. For example,
                                                      if there was enough metadata space to store 100
                                                      TiB of volumes and this number was set to 5,
                                                      then 500 TiB worth of volumes can be created.

        :rtype: sf_api.result_models.ModifyClusterFullThreshold
        """
        params = {'stage2AwareThreshold': stage2_aware_threshold,
                  'stage3BlockThresholdPercent': stage3_block_threshold_percent,
                  'stage3MetadataThresholdPercent': stage3_metadata_threshold_percent,
                  'maxMetadataOverProvisionFactor': max_metadata_overprovision_factor}

        response = self.call_method(enums.ClusterMethod.modify_cluster_full_threshold, params=params,
                                    result_dm_class=result_models.ModifyClusterFullThreshold)
        return response.result_dm

    def demote_cluster_master(self, node_id=None):
        """Demote the cluster master.

        :param int node_id: The node id of the cluster master to demote
        :rtype: sf_api.result_models.DemoteClusterMaster
        """
        params = {'nodeID': str(node_id)} if node_id else {}
        response = self.call_method(enums.ClusterMethod.demote_cluster_master, params=params,
                                    result_dm_class=result_models.DemoteClusterMaster)
        return response.result_dm

    def promote_cluster_master(self, node_id, allow_witness_promotion=False):
        """Promote a node to become the cluster master.

        :param int node_id: The node id of the node to promote to cluster master
        :param bool allow_witness_promotion: If true, allow witness to be promoted to cluster master
        :rtype: sf_api.result_models.PromoteClusterMaster
        """
        params = {'nodeID': str(node_id),
                  'allowWitnessNodePromotion': allow_witness_promotion}
        response = self.call_method(enums.ClusterMethod.promote_cluster_master, params=params,
                                    result_dm_class=result_models.PromoteClusterMaster)
        return response.result_dm

    def start_gc(self, timeout=None):
        """Start garbage collection.

        :param int timeout: Async handler timeout.
        :rtype: valence.functional_models.GarbageCollectionAsyncHandler
        """
        params = {'asyncHandle': True}
        start_time = time.time()
        response = self.call_method(enums.ClusterMethod.start_gc, params=params,
                                    result_dm_class=result_models.StartGC)
        async_handle_dm = models.AsyncHandle(response.result)
        async_handler_fm = _async_handler.GarbageCollectionAsyncHandler(
            cluster_fm=self,
            async_handle_dm=async_handle_dm,
            timeout=timeout,
            start_time=start_time
        )
        return async_handler_fm

    def get_gc_status(self):
        """Get the status of garbage collection in the cluster.

        :rtype: sf_api.result_models.GetGCStatus
        """
        response = self.call_method(enums.ClusterMethod.get_gc_status,
                                    result_dm_class=result_models.GetGCStatus)
        return response.result_dm

    def abort_gc(self):
        """Abort the current garbage collection process on the cluster."""
        if self.api_supported(enums.ClusterMethod.abort_gc):
            self.call_method(enums.ClusterMethod.abort_gc, result_dm_class=None)
        else:
            # If we don't have access to abort_gc, just wait for the current GC to finish.
            start_time = time.time()
            timeout = 30 * 60
            poll_period = 10
            while self.get_gc_status().status is enums.GcStatus.in_progress:
                if time.time() - start_time > timeout:
                    msg = 'Timeout of {} hit when waiting for garbage collection'.format(timeout)
                    raise errors.TimeoutError(msg)
                time.sleep(poll_period)

    def get_cluster_structure(self):
        """Get the structure of the cluster.

        :rtype: valence.models.ClusterStructure
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_structure,
                                    result_dm_class=result_models.GetClusterStructure)
        return response.result_dm

    def set_cluster_structure(self, cluster_structure):
        """Set the saved structure on a cluster.

        :param valence.models.ClusterStructure cluster_structure: The structure to set.
        :rtype: valence.functional_models.SetClusterStructureAsyncHandler
        """
        params = cluster_structure.serialize()
        response = self.call_method(enums.ClusterMethod.set_cluster_structure, params=params,
                                    result_dm_class=result_models.SetClusterStructure)
        async_handle_dm = models.AsyncHandle(response.result)
        return _async_handler.SetClusterStructureAsyncHandler(self, async_handle_dm)

    def get_active_tls_ciphers(self):
        """Get the active TLS ciphers from a cluster.

        :rtype: sf_api.result_models.GetActiveTlsCiphers
        """
        response = self.call_method(enums.ClusterMethod.get_active_tls_ciphers,
                                    result_dm_class=result_models.GetActiveTlsCiphers)
        return response.result_dm

    def get_supported_tls_ciphers(self):
        """Get the supported TLS ciphers from a cluster.

        :rtype: sf_api.result_models.GetSupportedTlsCiphers
        """
        response = self.call_method(enums.ClusterMethod.get_supported_tls_ciphers,
                                    result_dm_class=result_models.GetSupportedTlsCiphers)
        return response.result_dm

    def set_supplemental_tls_ciphers(self, ciphers):
        """Set the elements in the list of Supplemental TLS cipher suites for the cluster.

        :param list[str] ciphers: List of ciphers to create the supplemental cipher list.
        :rtype: sf_api.result_models.SetSupplementalTlsCiphers
        """
        params = {'supplementalCiphers': ciphers}
        response = self.call_method(enums.ClusterMethod.set_supplemental_tls_ciphers,
                                    params=params,
                                    result_dm_class=result_models.SetSupplementalTlsCiphers)
        return response.result_dm

    def reset_supplemental_tls_ciphers(self):
        """Reset the elements in the Supplemental list of TLS cipher suites back to the defaults.

        :rtype: sf_api.result_models.ResetSupplementalTlsCiphers
        """
        response = self.call_method(enums.ClusterMethod.reset_supplemental_tls_ciphers,
                                    result_dm_class=result_models.ResetSupplementalTlsCiphers)
        return response.result_dm

    def start_cluster_pairing(self):
        """Get a cluster pairing key for pairing to another cluster.

        :rtype: str
        """
        response = self.call_method(enums.ClusterMethod.start_cluster_pairing,
                                    result_dm_class=result_models.StartClusterPairing)
        return response.result_dm.cluster_pairing_key

    def complete_cluster_pairing(self, pairing_key):
        """Complete cluster pairing.

        :param str pairing_key: Pairing key
        :rtype: valence.models.ClusterPair
        """
        params = {api_constants.CLUSTER_PAIRING_KEY: pairing_key}
        response = self.call_method(enums.ClusterMethod.complete_cluster_pairing,
                                    params=params,
                                    result_dm_class=result_models.CompleteClusterPairing)
        return response.result_dm

    def list_cluster_pairs(self):
        """List cluster pairs.

        :rtype: list[valence.models.ClusterPair]
        """
        response = self.call_method(enums.ClusterMethod.list_cluster_pairs,
                                    result_dm_class=result_models.ListClusterPairs)
        return response.result_dm.cluster_pairs

    def remove_cluster_pair(self, pair_dm):
        """Remove cluster pair.

        :param valence.models.ClusterPair pair_dm: Cluster pair data model.
        """
        params = {api_constants.CLUSTER_PAIR_ID: pair_dm.cluster_pair_id}
        self.call_method(enums.ClusterMethod.remove_cluster_pair, params=params, result_dm_class=None)

    def start_cluster_bs_check(self, async_handle=True, slice_ids=None, timeout=1800,
                               volume_dms=None):
        """Start cluster BSCheck.

        :param bool async_handle: if true return an async handle
        :param list[int] slice_ids: sliceIDs to check
        :param int timeout: Async handler timeout.
        :param list[valence.models.Volume] volume_dms: Volumes to check
        :rtype: valence.functional_models.AsyncHandler
        """
        params = {'asyncHandle': async_handle}
        if volume_dms:
            if len(volume_dms) == 1:
                params['volumeID'] = volume_dms[0].volume_id
            else:
                volume_id_list = []
                for volume_dm in volume_dms:
                    volume_id_list.append(volume_dm.volume_id)
                params['volumeIDs'] = volume_id_list
        elif slice_ids:
            if len(slice_ids) == 1:
                params['sliceID'] = slice_ids[0]
            else:
                params['sliceIDs'] = slice_ids
        response = self.call_method(enums.ClusterMethod.start_cluster_bs_check, params=params,
                                    result_dm_class=result_models.StartClusterBSCheck)
        async_handle_dm = models.AsyncHandle(response.result)
        async_handler_fm = _async_handler.AsyncHandler(
            cluster_fm=self,
            async_handle_dm=async_handle_dm,
            timeout=timeout
        )
        return async_handler_fm

    def stop_cluster_bs_check(self):
        """Stop the current BSCheck process on the cluster.

        :rtype: sf_api.result_models.BaseApiResult
        """
        response = self.call_method(enums.ClusterMethod.stop_cluster_bs_check)
        return response

    def add_block_to_bs(self, block_hash=None, hash_data=None):
        """Add Block to BS.

        :param str block_hash: block to add data to
        :param str hash_data: data to add to block
        :rtype: sf_api.result_models.AddBlockToBS
        """
        params = {'hash': block_hash,
                  'hashData': hash_data}
        response = self.call_method(enums.ClusterMethod.add_block_to_bs, params=params,
                                    result_dm_class=result_models.AddBlockToBS)
        return response.result_dm

    def remove_block_from_bs(self, block_hash=None):
        """Remove Block from BS.

        :param str block_hash: hash of block to check
        :rtype: sf_api.result_models.RemoveBlockFromBS
        """
        params = {'hash': block_hash}
        response = self.call_method(enums.ClusterMethod.remove_block_from_bs, params=params,
                                    result_dm_class=result_models.RemoveBlockFromBS)
        blocks_removed_successfully = len(response.result_dm.success) \
            if response.result_dm.success is not None else 0
        expected_blocks_removed = self.get_cluster_info().rep_count

        # This will verify blocks were actually removed.
        # Otherwise, the API could succeed without changes being made.
        if blocks_removed_successfully != expected_blocks_removed:
            msg = 'Blocks removed {} does not match expected count of {}'.format(
                blocks_removed_successfully, expected_blocks_removed)
            raise _errors.FunctionalClusterError(msg)
        return response.result_dm

    def check_block(self, block_hash=None):
        """Check Cluster Block.

        :param str block_hash: hash of block to check
        :rtype: sf_api.result_models.CheckBlock
        """
        params = {'hash': block_hash}
        response = self.call_method(enums.ClusterMethod.check_block, params=params,
                                    result_dm_class=result_models.CheckBlock)
        return response.result_dm

    def list_node_stats(self):
        """Return the individual node stats for the cluster.

        :rtype: list[valence.models.NodeStats]
        """
        response = self.call_method(enums.ClusterMethod.list_node_stats, params={},
                                    result_dm_class=result_models.ListNodeStats)
        return response.result_dm.node_stats.nodes

    def set_ssl_certificate(self, certificate, private_key, force=False):
        """Set the cluster ssl certificate.

        :param str certificate: certificate.pem string to set
        :param str private_key: private key for the certificate
        :param bool force: hidden method to force setting the ssl certificate.
                           WARNING:  this will bypass checking to see if the certificate is valid.
                           In theory the cluster will check again when attempting to apply the
                           certificate
        """
        params = {'certificate': certificate,
                  api_constants.PRIVATE_KEY: private_key,
                  api_constants.FORCE: force}
        self.call_method(enums.ClusterMethod.set_ssl_certificate, params=params,
                         result_dm_class=result_models.SetSslCertificate)

    def get_ssl_certificate(self):
        """Get the cluster ssl certificate.

        :rtype: sf_api.result_models.GetSslCertificate
        """
        response = self.call_method(enums.ClusterMethod.get_ssl_certificate, params={},
                                    result_dm_class=result_models.GetSslCertificate)
        return response.result_dm

    def remove_ssl_certificate(self):
        """Remove any customer installed ssl certificates on the cluster.

        If no customer installed certificate is present, this does nothing.
        """
        self.call_method(enums.ClusterMethod.remove_ssl_certificate, params={},
                         result_dm_class=result_models.RemoveSslCertificate)

    def get_feature_status(self, feature=None):
        """Get the cluster feature status via GetFeatureStatus.

        :param valence.enums.SfFeature feature: feature value to check
        :rtype: list[valence.models.FeatureStatus]
        """
        params = {api_constants.FEATURE: feature.value} if feature else None
        response = self.call_method(enums.ClusterMethod.get_feature_status, params=params,
                                    result_dm_class=result_models.GetFeatureStatus)
        return response.result_dm.features

    def enable_feature(self, feature):
        """Enable a SolidFire feature that is disabled by default.

        :param valence.enums.SfFeature feature: feature value to check
        :rtype: valence.functional_models.EnableFeatureAsyncHandler
        """
        params = {api_constants.FEATURE: feature.value}
        self.call_method(enums.ClusterMethod.enable_feature, params=params, result_dm_class=None)
        return _async_handler.EnableFeatureAsyncHandler(cluster_fm=self, feature=feature)

    def get_fips_report(self):
        """Get fips report via GetFipsReport.

        :rtype: sf_api.result_models.GetFipsReport
        """
        response = self.call_method(enums.ClusterMethod.get_fips_report,
                                    result_dm_class=result_models.GetFipsReport)
        return response.result_dm

    def start_upgrade(self, package_name):
        """Sets upgrade status on cluster with given package name.

        :param str package_name: name of package downloaded to a node
        """

        self.call_method(enums.ClusterMethod.start_upgrade,
                         params={api_constants.PACKAGE_NAME: package_name})

    def finish_upgrade(self):
        """Clears upgrade status on cluster."""

        self.call_method(enums.ClusterMethod.finish_upgrade)

    def complete_cluster_upgrade(self):
        """Temporary fix until managementnode fm is in place and does upgrades.

        This method will set the cluster fm upgrading flag to False and update the api endpoint.
        """
        self.upgrading = False
        self.cluster_dm.api_endpoint = self.get_current_version()

    def get_cluster_stats(self):
        """Gets cluster stats.

        :rtype: sf_api.result_models.GetClusterStats
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_stats, params={},
                                    result_dm_class=result_models.GetClusterStats)
        return response.result_dm

    def shutdown(self, nodes, option=enums.ShutdownOption.restart):
        """Shuts down one or more cluster nodes.

        :param list[valence.models.ActiveNode] nodes: ActiveNode data models to restart or halt
        :param valence.enums.ShutdownOption option: One of 'restart' or 'halt' shutdown option type.
        :returns: Result Data Model for Shutdown
        :rtype: sf_api.result_models.Shutdown
        """
        node_ids = [node.node_id for node in nodes]
        params = {api_constants.NODES: node_ids,
                  api_constants.OPTION: option}

        response = self.call_method(enums.ClusterMethod.shutdown, params=params,
                                    result_dm_class=result_models.Shutdown)

        return response.result_dm
