"""Cluster FM methods related to drives."""
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    criteria,
    enums,
    models,
)
from valence.functional_models import (
    _async_handler,
    monitor_decorators,
)
from valence.functional_models._cluster import base


class DriveMixin(base.ClusterBase):
    """Cluster FM Drive methods."""

    @monitor_decorators.notify_monitor
    def add_drives(self, drive_dms, ignore_unbalanced_cluster_check=None, force_during_bin_sync=None,
                   force_non_fips=None):
        """Add specified available drives to a cluster via AddDrives.

        This call is asynchronous, and returns an AddDrives async handler.

        :param typing.Iterable[valence.models.Drive] drive_dms: drive data models
        :param bool ignore_unbalanced_cluster_check: Whether to skip the check for an unbalanced cluster
        :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
        :param bool force_non_fips: Force add non-fips drives to a fips cluster
        :return: AsyncHandler handler
        :rtype: valence.functional_models.AddDrivesAsyncHandler
        """
        drives = []
        drive_ids = []

        for drive_dm in drive_dms:
            # AddDrives expects dictionaries that contain driveID and type
            drives.append({api_constants.DRIVE_ID: drive_dm.drive_id,
                           api_constants.TYPE: drive_dm.type.value})
            drive_ids.append(drive_dm.drive_id)

        params = {
            api_constants.DRIVES: drives,
            api_constants.IGNORE_UNBALANCED_CLUSTER_CHECK: ignore_unbalanced_cluster_check,
            api_constants.FORCE_DURING_BIN_SYNC: force_during_bin_sync,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading,
            api_constants.FORCE_NON_FIPS: force_non_fips
        }
        response = self.call_method(enums.ClusterMethod.add_drives, params=params,
                                    result_dm_class=result_models.AddDrives)

        if response.result_dm.async_handle is not None:
            async_handle_dm = models.AsyncHandle(response.result)
        else:
            async_handle_dm = None
        async_handler_fm = _async_handler.AddDrivesAsyncHandler(self, drive_ids=drive_ids,
                                                                async_handle_dm=async_handle_dm)
        return async_handler_fm

    def list_drives(self, node_ids=None, drive_ids=None, slots=None, statuses=None, drive_types=None,
                    serials=None):
        """List drives on a cluster via ListDrives.

        :param typing.Iterable[int] node_ids: Return drives associated with these node_ids.
        :param typing.Iterable[int] drive_ids: Return drives with these IDs.
        :param typing.Iterable[int] slots: Return drives in these slots.
        :param typing.Iterable[valence.enums.DriveStatus] statuses: Return drives with this status.
        :param typing.Iterable[valence.enums.DriveType] drive_types: Return drives of this type.
        :param typing.Iterable[str] serials: Return drives only if serial numbers match
        :return: List of drive data models.
        :rtype: list[valence.models.Drive]
        """
        response = self.call_method(enums.ClusterMethod.list_drives,
                                    result_dm_class=result_models.ListDrives)
        all_drives = response.result_dm.drives
        # Put into a set to improve efficiency
        drive_ids = set(drive_ids) if drive_ids else None
        node_ids = set(node_ids) if node_ids else None
        slots = set(slots) if slots else None
        statuses = set(statuses) if statuses else None
        drive_types = set(drive_types) if drive_types else None
        serials = set(serials) if serials else None

        drive_filter = criteria.And(
            criteria.ContainedIn(models.Drive.node_id, node_ids),
            criteria.ContainedIn(models.Drive.drive_id, drive_ids),
            criteria.ContainedIn(models.Drive.slot, slots),
            criteria.ContainedIn(models.Drive.status, statuses),
            criteria.ContainedIn(models.Drive.type, drive_types),
            criteria.ContainedIn(models.Drive.serial, serials)
        )
        filtered_drives = drive_filter.filter(all_drives)
        return filtered_drives

    def list_drive_stats(self, drive_ids=None):
        """List the stats on all the drives with ListDriveStats.

        :param typing.Iterable[int] drive_ids: Return drives with these IDs.
        :return: List of drive status data models.
        :rtype: list[valence.models.DriveStats]
        """
        params = {api_constants.DRIVES: drive_ids}
        response = self.call_method(enums.ClusterMethod.list_drive_stats,
                                    params=params,
                                    result_dm_class=result_models.ListDriveStats)
        drive_stats = response.result_dm.drive_stats

        return drive_stats

    def resize_drives(self, drive_dms, new_size):
        """Resize drives to requested size.

        :param typing.Iterable[valence.models.Drive] drive_dms: drive data models
        :param valence.models.DataSize new_size: new size of the drive in bytes.
        :return: List of drive data models.
        :rtype: list[valence.models.Drive]
        """

        drive_ids = [drive_dm.drive_id for drive_dm in drive_dms]
        params = {
            api_constants.DRIVES: drive_ids,
            api_constants.NEW_USABLE_SIZE: new_size.bytes
        }
        response = self.call_method(enums.ClusterMethod.resize_drives, params=params,
                                    result_dm_class=result_models.ResizeDrives)
        all_drives = response.result_dm.drives
        return all_drives

    @monitor_decorators.notify_monitor
    def remove_drives(self, drive_dms, ignore_unbalanced_cluster_check=None, timeout=None,
                      force_during_bin_sync=None):
        """Remove the drive associated with each given data model.

        :param typing.Iterable[valence.models.Drive] drive_dms: drives to remove
        :param bool ignore_unbalanced_cluster_check: Whether to skip the check for an unbalanced cluster
        :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
        :param int timeout: Async handler timeout.
        :return: AsyncHandler handler
        :rtype: valence.functional_models.RemoveDrivesAsyncHandler
        """
        drive_ids = [d.drive_id for d in drive_dms]

        params = {
            api_constants.DRIVES: drive_ids,
            api_constants.IGNORE_UNBALANCED_CLUSTER_CHECK: ignore_unbalanced_cluster_check,
            api_constants.FORCE_DURING_BIN_SYNC: force_during_bin_sync,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }
        response = self.call_method(enums.ClusterMethod.remove_drives, params=params,
                                    result_dm_class=result_models.RemoveDrives)
        # cluster_api.remove_drives returns the value of the asyncHandle from the SF API call,
        # reindex it inside a dict for async_handle_dm to construct the async_handle int properly
        async_handle_dm = models.AsyncHandle(
            {api_constants.ASYNC_HANDLE: response.result_dm.async_handle}
        )

        async_handler_fm = _async_handler.RemoveDrivesAsyncHandler(
            cluster_fm=self,
            drive_ids=drive_ids,
            async_handle_dm=async_handle_dm,
            timeout=timeout
        )

        return async_handler_fm

    @monitor_decorators.notify_monitor
    def secure_erase_drives(self, drive_dms):
        """Secure erases the specified drives via SecureEraseDrives.

        This call is asynchronous, and returns a SecureEraseDrives async handler.

        :param typing.Iterable[valence.models.Drive] drive_dms: drive data models
        :return: AsyncHandler handler
        :rtype: valence.functional_models.SecureEraseDrivesAsyncHandler
        """
        drive_ids = [d.drive_id for d in drive_dms]
        params = {api_constants.DRIVES: drive_ids}
        response = self.call_method(enums.ClusterMethod.secure_erase_drives, params=params,
                                    result_dm_class=result_models.SecureEraseDrives)

        async_handle_dm = models.AsyncHandle(
            {api_constants.ASYNC_HANDLE: response.result_dm.async_handle}
        )

        async_handler_fm = _async_handler.SecureEraseDrivesAsyncHandler(self, drive_ids=drive_ids,
                                                                        async_handle_dm=async_handle_dm)
        return async_handler_fm

    def get_drive_hardware_info(self, drive_dm):
        """Get the hardware info for the specified drive.

        :param valence.models.Drive drive_dm: Drive data model.
        :rtype: valence.models.DriveHardwareInfo
        """
        params = {api_constants.DRIVE_ID: drive_dm.drive_id}
        response = self.call_method(enums.ClusterMethod.get_drive_hardware_info, params=params,
                                    result_dm_class=result_models.GetDriveHardwareInfo)
        return response.result_dm.drive_hardware_info
