"""Cluster functional model methods related to hardware."""

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models import _errors
from valence.functional_models._cluster import base


class HardwareMixin(base.ClusterBase):
    """Cluster FM hardware methods."""

    def get_hardware_info_overrides(self):
        """GetHardwareInfoOverrides.

        :rtype: valence.models.HardwareInfoOverrides
        """
        response = self.call_method(enums.ClusterMethod.get_hardware_info_overrides,
                                    result_dm_class=result_models.GetHardwareInfoOverrides)
        return response.result_dm.hardware_info_overrides

    def set_hardware_info_overrides(self, chassis_serial_overrides):
        """SetHardwareInfoOverrides.

        :param dict[str,dict[str,str]] chassis_serial_overrides: A dictionary with the node ids and the
            new chassis serials to override them
        :rtype: valence.models.HardwareInfoOverrides
        """
        response = self.call_method(enums.ClusterMethod.set_hardware_info_overrides,
                                    params={api_constants.NODES: chassis_serial_overrides},
                                    result_dm_class=result_models.GetHardwareInfoOverrides)
        return response.result_dm.hardware_info_overrides

    def get_cluster_hardware_info(self, hw_type=enums.ClusterHardwareInfoType.all):
        """GetClusterHardwareInfo.

        :param valence.enums.ClusterHardwareInfoType hw_type: The hardware type to get information about
        :rtype: valence.models.ClusterHardwareInfo
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_hardware_info,
                                    params={api_constants.TYPE: hw_type},
                                    result_dm_class=result_models.GetClusterHardwareInfo)
        return response.result_dm.cluster_hardware_info

    def check_proposed_node_additions(self, node_dms, raise_on_failure=True):
        """Run checks to verify that a set of nodes could be added to an existing cluster.

        :param list[valence.models.ResourceNode] node_dms: ReourceNode data models for nodes to add
        :param bool raise_on_failure: raise an exception if any checks fail
        :raises valence.functional_models.FunctionalClusterError: when any of the checks fail and
            flag is set
        :rtype: sf_api.result_models.CheckProposedCluster
        """
        params = {
            api_constants.NODES: [node.mip for node in node_dms]
        }
        check_proposed_node_additions_dm = self.call_method(
            enums.ClusterMethod.check_proposed_node_additions,
            params=params,
            result_dm_class=result_models.CheckProposedNodeAdditions
        ).result_dm
        proposed_cluster_errors = check_proposed_node_additions_dm.proposed_cluster_errors
        failed_checks = [str(error.description) for error in proposed_cluster_errors]
        if raise_on_failure and failed_checks:
            raise _errors.FunctionalClusterError('CheckProposedNodeAdditions API returned a failed '
                                                 'check result: {}'.format('\n'.join(failed_checks)))

        return check_proposed_node_additions_dm
