"""Cluster FM methods related to initiators."""
from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._cluster import base


class InitiatorMixin(base.ClusterBase):
    """Cluster FM Initiator methods."""

    def create_initiators(self, initiators):
        """Create initiators on the cluster.

        :param list[valence.models.Initiator] initiators: Array of initiator objects.
        :return: Initiator data models
        :rtype: list[valence.models.Initiator]
        :raises ValueError: If multiple volume access groups are passed in any of the initiator DMs
        """
        init_dicts = []
        for i in initiators:
            init_dict = {
                api_constants.NAME: i.initiator_name,
                api_constants.ALIAS: i.alias,
                api_constants.ATTRIBUTES: i.attributes,
                api_constants.CHAP_USERNAME: i.chap_username,
                api_constants.INITIATOR_SECRET: i.initiator_secret,
                api_constants.REQUIRE_CHAP: i.require_chap,
                api_constants.TARGET_SECRET: i.target_secret
            }

            if i.volume_access_groups:
                if len(i.volume_access_groups) != 1:
                    raise ValueError('Initiators must be created with only one volume access group')
                else:
                    init_dict[api_constants.VOLUME_ACCESS_GROUP_ID] = i.volume_access_groups[0]

            init_dicts.append(init_dict)

        response = self.call_method(enums.ClusterMethod.create_initiators,
                                    params={api_constants.INITIATORS: init_dicts,
                                            api_constants.FORCE_DURING_UPGRADE: self.upgrading},
                                    result_dm_class=result_models.CreateInitiators)

        return response.result_dm.initiators

    def delete_initiators(self, initiators):
        """Delete initiators on the cluster.

        :param list[valence.models.Initiator] initiators: InitiatorIDs of the initiators to delete.
        """
        initiator_ids = [i.initiator_id for i in initiators]
        self.call_method(enums.ClusterMethod.delete_initiators,
                         params={api_constants.INITIATORS: initiator_ids,
                                 api_constants.FORCE_DURING_UPGRADE: self.upgrading},
                         result_dm_class=None)

    def modify_initiator(self, initiator_dm, alias=None, attributes=None):
        """Modify initiators on the cluster.

        :param valence.models.Initiator initiator_dm: Initiator data model for initiator to modify.
        :param str alias: New alias for the initiator.
        :param dict attributes: New attributes for the initiator.
        :return: Initiator data model
        :rtype: valence.models.Initiator
        """
        params = {
            api_constants.INITIATOR_ID: initiator_dm.initiator_id,
            api_constants.ALIAS: alias,
            api_constants.ATTRIBUTES: attributes,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }
        response = self.call_method(enums.ClusterMethod.modify_initiator, params=params,
                                    result_dm_class=result_models.ModifyInitiator)

        return response.result_dm.initiator

    def modify_initiators(self, initiators):
        """Modify initiators on the cluster.

        NOTE: In order to remove an initiator from its existing volume access group, pass None
        for volume_access_groups in the data model.

        :param list[valence.models.Initiator] initiators: List of Initiator data models to modify.
        :return: Initiator data models.
        :rtype: list[valence.models.Initiator]
        :raises ValueError: If multiple volume access groups are passed in any of the initiator DMs
        """
        init_dicts = []
        for i in initiators:
            volume_access_groups = i.volume_access_groups
            i = i.serialize()
            i = {k: v for k, v in i.items() if v is not None}
            if volume_access_groups is None:
                i[api_constants.VOLUME_ACCESS_GROUP_ID] = None
            elif len(volume_access_groups) == 0:
                pass
            elif len(volume_access_groups) == 1:
                i[api_constants.VOLUME_ACCESS_GROUP_ID] = volume_access_groups[0]
            else:
                raise ValueError('Initiators can only be modified to belong to one volume access group')
            init_dicts.append(i)

        response = self.call_method(enums.ClusterMethod.modify_initiators,
                                    params={api_constants.INITIATORS: init_dicts,
                                            api_constants.FORCE_DURING_UPGRADE: self.upgrading},
                                    result_dm_class=result_models.ModifyInitiators)

        return response.result_dm.initiators

    def list_initiators(self, start_initiator_id=None, limit=None):
        """List the current initiators on the cluster.

        :param int start_initiator_id: InitiatorID to start the list of initiators from.
        :param int limit: The number of initiators to list.
        :return: Initiator data models
        :rtype: list[valence.models.Initiator]
        """
        params = {
            api_constants.START_INITIATOR_ID: start_initiator_id,
            api_constants.LIMIT: limit
        }
        response = self.call_method(enums.ClusterMethod.list_initiators, params=params,
                                    result_dm_class=result_models.ListInitiators)
        return response.result_dm.initiators

    def add_initiators_to_volume_access_group(self, volume_access_group_id, initiator_ids):
        """Add initiators to a volume access group.

        :param int volume_access_group_id: VAG id to add the initiators to.
        :param list[int] initiator_ids: Initiator ids to add to the VAG.
        :return: Volume access group data model
        :rtype: valence.models.VolumeAccessGroup
        """
        params = {
            api_constants.VOLUME_ACCESS_GROUP_ID: volume_access_group_id,
            api_constants.INITIATORS: initiator_ids,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }
        response = self.call_method(enums.ClusterMethod.add_initiators_to_volume_access_group,
                                    params=params,
                                    result_dm_class=result_models.AddInitiatorsToVolumeAccessGroup)

        return response.result_dm.volume_access_group

    def remove_initiators_from_volume_access_group(self, volume_access_group_id, initiator_ids,
                                                   delete_orphan_initiators=True):
        """Remove initiators from a volume access group.

        :param int volume_access_group_id: VAG id to remove initiators from.
        :param list[int] initiator_ids: Initiators ids to remove from the VAG.
        :param bool delete_orphan_initiators: Delete initiator objects after they are removed.
        :return: Volume access group data model
        :rtype: valence.models.VolumeAccessGroup
        """
        params = {
            api_constants.VOLUME_ACCESS_GROUP_ID: volume_access_group_id,
            api_constants.INITIATORS: initiator_ids,
            api_constants.DELETE_ORPHAN_INITIATORS: delete_orphan_initiators,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }
        response = self.call_method(enums.ClusterMethod.remove_initiators_from_volume_access_group,
                                    params=params,
                                    result_dm_class=result_models.RemoveInitiatorsFromVolumeAccessGroup)

        return response.result_dm.volume_access_group
