"""Cluster FM methods related to authentication sessions."""
import datetime

from sf_api import (
    api_constants,
    enums,
    result_models,
)
from valence import utils
from valence.functional_models._cluster import base


class MfaMixin(base.ClusterBase):
    """Cluster FM  MultiFactor Authentication methods."""

    def add_idp_cluster_admin(self, username, access=None, attributes=None):
        """Add an idp cluster admin.

        :param str username: attribute-vlaue mapping to an IdP cluster admin(e.g. email=foo@bar.com)
        :param list[valence.enums.ClusterAccessLevel] access: Controls which methods this Cluster Admin
            can use. See the Access Control appendix of the API Reference for more details. If not
            given, default to [ClusterAccessLevel.administrator]
        :param dict attributes: json formatted name-value pairs
        :rtype: int
        """
        params = {
            api_constants.USERNAME: username,
            api_constants.ACCEPT_EULA: True,
            api_constants.ACCESS: access or [enums.ClusterAccessLevel.administrator],
            api_constants.ATTRIBUTES: attributes
        }
        response = self.call_method(enums.ClusterMethod.add_idp_cluster_admin, params=params,
                                    result_dm_class=result_models.AddClusterAdmin)
        return response.result_dm.cluster_admin_id

    def create_idp_auth_session(self, saml_attributes, desired_expiration_duration=None,
                                username=None):
        """Creates an Authentication Session on the cluster via CreateAuthSession.

        :param list[str] saml_attributes: list of name-value pairs for determining
            IdP Cluster Admin id(s) for this session.
        :param datetime.timedelta desired_expiration_duration: Optional time delta from 'now' which will
            defines the desired time at which the auth session will definitely expire.  This value may
            or may not be honored.
        :param str username: name-value pair used to identify the user for auditing operations
            within the session
        :return: Authentication Session Info data model
        :rtype: valence.models.AuthSessionInfo
        """
        # Convert to seconds since epoch which is what the API expects.
        if desired_expiration_duration:
            epoch = datetime.datetime(1970, 1, 1)
            epoch_value = utils.get_now_utc() - epoch + desired_expiration_duration
            desired_expiration_duration = str(int(epoch_value.total_seconds()))

        # Create the Authentication Session on the cluster
        params = {
            api_constants.SAML_ATTRIBUTE_STATEMENTS: saml_attributes,
            api_constants.USERNAME: username,
            api_constants.DESIRED_EXPIRATION_DURATION: desired_expiration_duration
        }
        response = self.api.call_method(enums.ClusterMethod.create_idp_auth_session, params=params,
                                        result_dm_class=result_models.CreateAuthSessionResult)

        return response.result_dm.session

    def create_idp_configuration(self, idp_url):
        """Create and Idp configuration.

        :param str idp_url: URL for retrieving IdP Metadata for configuration and integration details
            for SAML 2.0 single sign-on.
        :return: Data model for Idp Configuration created.
        :rtype: valence.models.IdpConfigInfo
        """

        params = {
            api_constants.IDP_METADATA_URL: idp_url
        }

        response = self.api.call_method(enums.ClusterMethod.create_idp_configuration, params=params,
                                        result_dm_class=result_models.CreateIdpConfigurationResult)
        return response.result_dm.idp_config_info

    def delete_idp_configuration(self, idp_config):
        """Delete existing Idp Configuration.

        :param valence.models.IdpConfigInfo idp_config: IDP configuration to delete.
        """

        params = {
            api_constants.IDP_CONFIGURATION_ID: str(idp_config.idp_configuration_id)
        }

        self.api.call_method(enums.ClusterMethod.delete_idp_configuration, params=params,
                             result_dm_class=result_models.DeleteIdpConfigurationResult)

    def list_idp_configurations(self, idp_url=None):
        """List idp configurations.  Optionally filter by idp metadata url.

        :param str idp_url: filter on idp_url.
        :rtype: list[valence.models.IdpConfigInfo]
        """

        params = {
            api_constants.IDP_METADATA_URL: idp_url
        }

        response = self.api.call_method(enums.ClusterMethod.list_idp_configurations, params=params,
                                        result_dm_class=result_models.ListIdpConfigurationsResult)
        return response.result_dm.idp_config_infos

    def update_idp_configuration(self, idp_config, idp_url=None, refresh_cert_expiration=None):
        """Update existing idp configuration identified by the config id and metadata url.

        :param valence.models.IdpConfigInfo idp_config:  Idp configuration to update.
        :param str idp_url: URL for retrieving IdP Metadata for configuration and integration details
            for SAML 2.0 single sign-on.
        :param bool refresh_cert_expiration: Update the certificate expiration.
        :rtype: valence.models.IdpConfigInfo
        """

        params = {
            api_constants.IDP_CONFIGURATION_ID: str(idp_config.idp_configuration_id),
            api_constants.IDP_METADATA_URL: idp_url,
            api_constants.REFRESH_CERTIFICATE_EXPIRATION_TIME: refresh_cert_expiration
        }

        response = self.api.call_method(enums.ClusterMethod.update_idp_configuration, params=params,
                                        result_dm_class=result_models.UpdateIdpConfigurationResult)
        return response.result_dm.idp_config_info

    def enable_idp_authentication(self, idp_config=None):
        """Enable authentication using an Idp configuration.

        :param valence.models.IdpConfigInfo idp_config: IDP configuration to enable
            if more than one exists.
        """
        if idp_config is not None:
            idp_config_id = str(idp_config.idp_configuration_id)
        else:
            idp_config_id = None
        params = {
            api_constants.IDP_CONFIGURATION_ID: idp_config_id
        }

        self.api.call_method(enums.ClusterMethod.enable_idp_authentication, params=params,
                             result_dm_class=result_models.EnableIdpAuthenticationResult)

    def disable_idp_authentication(self):
        """Disable Idp authentication."""

        params = {}

        self.api.call_method(enums.ClusterMethod.disable_idp_authentication, params=params,
                             result_dm_class=result_models.DisableIdpAuthenticationResult)

    def get_idp_authentication_state(self):
        """Get whether idp auth is enabled or disabled.

        :rtype: bool
        """
        params = {}
        response = self.api.call_method(enums.ClusterMethod.get_idp_authentication_state, params=params,
                                        result_dm_class=result_models.GetIdpAuthenticationStateResult)

        return response.result_dm.enabled
