"""Cluster FM methods related to nodes."""
import autologger
import requests

from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    criteria,
    enums,
    models,
)
from valence.functional_models import _async_handler
from valence.functional_models._cluster import base

_log = autologger.get_sflogging()


class NodeMixin(base.ClusterBase):
    """Cluster FM Node methods."""

    def add_nodes(self, pending_node_dms, force_non_fips=None, auto_install=True):
        """Add pending nodes to a cluster via AddNodes.

        This call is asynchronous, and returns a list of async handlers. It also spins up a
        reactor to monitor each node, this is done just for logging purposes. If the reactor
        fails all that will be lost is logging and the test will not be effected.

        :param typing.Iterable[valence.models.PendingNode] pending_node_dms: nodes to add
        :param bool force_non_fips: Force add non-fips nodes to a fips cluster
        :param bool auto_install: auto install added node to the cluster's current build
        :rtype: list[valence.functional_models.AddNodesAsyncHandler]
        """
        pending_node_ids = []
        pending_node_static_ids = []
        for pending_node_dm in pending_node_dms:
            pending_node_static_ids.append(pending_node_dm.static_id)
            pending_node_ids.append(pending_node_dm.pending_node_id)

        response = self.call_method(enums.ClusterMethod.add_nodes,
                                    params={api_constants.PENDING_NODES: pending_node_ids,
                                            api_constants.FORCE_DURING_UPGRADE: self.upgrading,
                                            api_constants.FORCE_NON_FIPS: force_non_fips,
                                            api_constants.AUTO_INSTALL: auto_install},
                                    result_dm_class=result_models.AddNodes)

        async_handler_fms = []
        for node in response.result_dm.nodes:
            if node.async_handle is not None and node.async_handle != 0:
                async_handle_dm = models.AsyncHandle({api_constants.ASYNC_HANDLE: node.async_handle})
                async_handler_fms.append(_async_handler.AddNodesAsyncHandler
                                         (self, node.static_id,
                                          async_handle_dm=async_handle_dm))
            else:
                async_handler_fms.append(_async_handler.AddNodesAsyncHandler
                                         (self, pending_node_static_ids.pop(), async_handle_dm=None))
        return async_handler_fms

    def remove_nodes(self, node_dms):
        """Remove nodes from the cluster.

        :param typing.Iterable[valence.models.ActiveNode] node_dms: List of active node data models
        :rtype: list[valence.functional_models.RemoveNodesAsyncHandler]
        """
        node_ids = []
        node_static_ids = []
        for node_dm in node_dms:
            node_static_ids.append(node_dm.static_id)
            node_ids.append(node_dm.node_id)

        cm_targeted = True if self.get_cluster_master_node_id() in node_ids else False
        try:
            self.call_method(enums.ClusterMethod.remove_nodes,
                             params={api_constants.NODES: node_ids,
                                     api_constants.FORCE_DURING_UPGRADE: self.upgrading},
                             result_dm_class=None)
        except requests.exceptions.ReadTimeout:
            if cm_targeted:
                active_nodes = self.list_active_nodes()

                target_nodes_remaining = [active_node.node_id for active_node in active_nodes if
                                          active_node.node_id in node_ids]
                if target_nodes_remaining:
                    raise ValueError('Not all target nodes were removed: {}'.format(
                        target_nodes_remaining))
            else:
                raise

        async_handler_fms = []
        for node_dm in node_dms:
            async_handler_fms.append(_async_handler.RemoveNodesAsyncHandler(
                self, node_dm.static_id, async_handle_dm=None))

        return async_handler_fms

    def get_cluster_master(self):
        """Get the cluster master node data model.

        :rtype: valence.models.ActiveNode
        """
        node_id = self.get_cluster_master_node_id()
        nodes = self.list_active_nodes()
        node_filter = criteria.Equals(models.ActiveNode.node_id, node_id)
        return node_filter.filter(nodes)[0]

    def get_cluster_master_node_id(self):
        """Get the ID of the node currently acting as cluster master.

        :rtype: int
        """
        response = self.call_method(enums.ClusterMethod.get_cluster_master_node_id,
                                    result_dm_class=result_models.GetClusterMasterNodeID)
        return response.result_dm.node_id

    def get_node_fms(self, statuses=(enums.NodeStatus.active,), node_roles=None, authenticate=True):
        """Get node functional models for nodes on the cluster with the given statuses.

        :param typing.Iterable[valence.enums.NodeStatus] statuses: Node statuses on which to filter.
        :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
        :param bool authenticate: If true, attempt to authenticate each node
        :rtype: list[valence.functional_models.Node]
        """
        # Import the node functional model here to avoid a circular import
        from valence.functional_models._node import node

        node_dms = self.list_all_nodes(statuses=statuses, node_roles=node_roles)
        return [node.Node(
            node_dm=dm.resource_dm,
            cluster_dm=self.cluster_dm,
            authenticate=authenticate
        ) for dm in node_dms if dm.resource_dm is not None]

    def get_resource_nodes(self, statuses=(enums.NodeStatus.active,)):
        """Get resource node data models associated with the cluster.

        :param typing.Iterable[valence.enums.NodeStatus] statuses: Node statuses to filter on
        :rtype: list[valence.models.ResourceNode]
        """
        node_dms = self.list_all_nodes(statuses=statuses)
        return [dm.resource_dm for dm in node_dms]

    def get_network_interface_status(self, interfaces=None):
        """For each node, get the status of each network interface it has.

        This method calls the cluster API method GetNetworkInterfaceStatus with force=True. Without the
        force, SolidFire returns an error...

        "This method will be performed on all nodes in the cluster. If this is not
        what you intended, then log into an individual node and call this method directly
        on that node. If this is what you intended please reissue your request with force=true."

        .. note:: There is also a node_fm.get_network_interface_status() method which calls the node API
                  GetNetworkInterfaceStatus method.

        :param interfaces: List of Interfaces to get the status for. If None, return all interfaces.
        :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
        :rtype: list[valence.models.NodeNetworkInterfaceStatus]
        """
        params = {'force': True}
        if interfaces:
            params[api_constants.INTERFACES] = [i.value for i in interfaces]
        response = self.api.call_method(enums.ClusterMethod.get_network_interface_status, params=params,
                                        result_dm_class=result_models.ClusterGetNetworkInterfaceStatus)
        return response.result_dm.nodes

    def list_all_nodes(self, statuses=None, node_roles=None):
        """List all nodes associated with the cluster.

        :param typing.Iterable[valence.enums.NodeStatus] statuses: Node statuses on which to filter.
        :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
        :returns: A list of pending, active, and pending active nodes.
        :rtype: list[valence.models.NodeBase]
        """
        # Put into a set to improve efficiency
        statuses = set(statuses) if statuses else None
        node_roles = set(node_roles) if node_roles else None
        nodes_filter = criteria.And(
            criteria.ContainedIn(models.SfNodeBase.role, node_roles),
        )
        response = self.call_method(
            method=enums.ClusterMethod.list_all_nodes,
            result_dm_class=result_models.ListAllNodes
        )
        node_dms = []
        if not statuses or enums.NodeStatus.active in statuses:
            node_dms.extend(self._instantiate_nodes(
                node_dms=nodes_filter.filter(response.result_dm.nodes),
                node_class=models.ActiveNode
            ))
        if not statuses or enums.NodeStatus.pending in statuses:
            node_dms.extend(self._instantiate_nodes(
                node_dms=nodes_filter.filter(response.result_dm.pending_nodes),
                node_class=models.PendingNode
            ))
        if not statuses or enums.NodeStatus.pending_active in statuses:
            node_dms.extend(self._instantiate_nodes(
                node_dms=nodes_filter.filter(response.result_dm.pending_active_nodes),
                node_class=models.PendingActiveNode
            ))
        return node_dms

    def list_active_nodes(self, static_ids=None, node_roles=None, node_names=None):
        """List active nodes via ListActiveNodes.

        :param typing.Iterable[str] static_ids: Static IDs on which to filter.
        :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
        :param list[str] node_names: Node names on which to filter.
        :rtype: list[valence.models.ActiveNode]
        """
        # Put into a set to improve efficiency
        static_ids = set(static_ids) if static_ids else None
        node_roles = set(node_roles) if node_roles else None

        nodes_filter = criteria.And(
            criteria.ContainedIn(models.SfNodeBase.static_id, static_ids),
            criteria.ContainedIn(models.SfNodeBase.role, node_roles),
            criteria.ContainedIn(models.SfNodeBase.name, node_names)
        )
        response = self.call_method(
            method=enums.ClusterMethod.list_active_nodes,
            result_dm_class=result_models.ListActiveNodes
        )

        if models.get_resource_nodes() is None:
            _log.warn('Resource nodes have not been set with valence.models.set_node_resources(). The '
                      'returned active nodes will not have a reference to the resource node.')

        return self._instantiate_nodes(
            node_dms=nodes_filter.filter(response.result_dm.nodes),
            node_class=models.ActiveNode
        )

    def list_pending_nodes(self, static_ids=None, node_roles=None):
        """List pending nodes via ListPendingNodes.

        :param typing.Iterable[str] static_ids: Static IDs on which to filter.
        :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
        :rtype: list[valence.models.PendingNode]
        """
        # Put into a set to improve efficiency
        static_ids = set(static_ids) if static_ids else None
        node_roles = set(node_roles) if node_roles else None
        nodes_filter = criteria.And(
            criteria.ContainedIn(models.SfNodeBase.static_id, static_ids),
            criteria.ContainedIn(models.SfNodeBase.role, node_roles),
        )
        response = self.call_method(
            method=enums.ClusterMethod.list_pending_nodes,
            result_dm_class=result_models.ListPendingNodes
        )
        return self._instantiate_nodes(
            node_dms=nodes_filter.filter(response.result_dm.pending_nodes),
            node_class=models.PendingNode
        )

    def list_pending_active_nodes(self, static_ids=None, node_roles=None):
        """List pending active nodes via ListPendingActiveNodes.

        :param typing.Iterable[str] static_ids: Static IDs on which to filter.
        :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
        :rtype: list[valence.models.PendingActiveNode]
        """
        # Put into a set to improve efficiency
        static_ids = set(static_ids) if static_ids else None
        node_roles = set(node_roles) if node_roles else None
        nodes_filter = criteria.And(
            criteria.ContainedIn(models.SfNodeBase.static_id, static_ids),
            criteria.ContainedIn(models.SfNodeBase.role, node_roles),
        )
        response = self.call_method(
            method=enums.ClusterMethod.list_pending_active_nodes,
            result_dm_class=result_models.ListPendingActiveNodes
        )
        return self._instantiate_nodes(
            node_dms=nodes_filter.filter(response.result_dm.pending_active_nodes),
            node_class=models.PendingActiveNode
        )

    @staticmethod
    def _instantiate_nodes(node_dms, node_class):
        """Instantiate valence node data models.

        The node data models coming from sf_api are actually different than the Valence ones. The
        valence data models combine the resource node with the sf_api node.

        :param list[valence.models.Base] node_dms: List of sf_api node data models
        :param Any node_class: The class of the valence node to instantiate,
            e.g. valence.models.ActiveNode
        :rtype: list[valence.models.NodeBase]
        """
        result_dms = []

        for node in node_dms:
            rs_node_dm = models.get_resource_node_by_static_id(node.static_id)
            valence_node_dm = node_class(node, resource_dm=rs_node_dm)
            result_dms.append(valence_node_dm)
        return result_dms

    def get_node_stats(self, node_id):
        """Get stats for a specific node.

        :param int node_id:  Node ID to inspect
        :rtype: sf_api.models.NodeStats
        """
        params = {api_constants.NODE_ID: node_id}
        response = self.call_method(enums.ClusterMethod.get_node_stats, params=params,
                                    result_dm_class=result_models.GetNodeStats)
        return response.result_dm.node_stats
