"""Cluster FM methods related to protection domains."""

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._cluster import base


class ProtectionDomainMixin(base.ClusterBase):
    """Cluster FM methods for protection domains."""

    def get_protection_domain_layout(self):
        """List the protection domains on the cluster.

        The following cluster constant must be set for protection domains:

        :rtype: list[valence.models.NodeProtectionDomainLayout]
        """
        response = self.call_method(enums.ClusterMethod.get_protection_domain_layout,
                                    result_dm_class=result_models.GetProtectionDomainLayout)
        return response.result_dm.protection_domain_layout

    def set_protection_domain_layout_chassis_override(self, node_protection_domain_layouts):
        """Set chassis protection domains on the cluster.

        WARNING: This is dangerous, it will replace the "chassis" domain on nodes with custom defined
        domains.  While the public API will error if you are about to do something really bad, this
        will not. Kept private because customers shouldn't be overriding the chassis on nodes.

        When calling list active nodes the real chassis will still be reported.

        Primarily used for simulating different chassis configurations without moving nodes.

        This is a private method, used to be SetProtectionDomains.

        :param list[valence.models.NodeProtectionDomainLayout] node_protection_domain_layouts: A list of
            node protection domain layout data models
        :rtype: list[valence.models.NodeProtectionDomainLayout]
        """
        layouts = []
        for node_layout in node_protection_domain_layouts:
            chassis_layout = [pd.serialize() for pd in node_layout.protection_domains
                              if pd.protection_domain_type is
                              enums.ProtectionDomainType.chassis
                              ]
            layouts.append({
                api_constants.NODE_ID: node_layout.node_id,
                api_constants.PROTECTION_DOMAINS: chassis_layout
            })
        params = {api_constants.PROTECTION_DOMAIN_LAYOUT: layouts}

        response = self.call_method(enums.ClusterMethod.set_protection_domain_layout_chassis_override,
                                    params=params,
                                    result_dm_class=result_models.SetProtectionDomainLayout)
        return response.result_dm.protection_domain_layout

    def set_protection_domain_layout(self, node_protection_domain_layouts):
        """Set custom protection domains on the cluster.

        This is a public method.

        :param list[valence.models.NodeProtectionDomainLayout] node_protection_domain_layouts: A list
            of node protection domain layout data models
        :rtype: list[valence.models.NodeProtectionDomainLayout]
        """
        layouts = []
        for node_layout in node_protection_domain_layouts:
            custom_layout = [pd.serialize() for pd in node_layout.protection_domains
                             if pd.protection_domain_type is
                             enums.ProtectionDomainType.custom
                             ]
            layouts.append({
                api_constants.NODE_ID: node_layout.node_id,
                api_constants.PROTECTION_DOMAINS: custom_layout
            })
        params = {api_constants.PROTECTION_DOMAIN_LAYOUT: layouts}

        response = self.call_method(enums.ClusterMethod.set_protection_domain_layout,
                                    params=params,
                                    result_dm_class=result_models.SetProtectionDomainLayout)
        return response.result_dm.protection_domain_layout

    def list_protection_domain_levels(self):
        """List the protection domain levels on the cluster.

        :rtype: list[valence.models.ProtectionDomainLevel]
        """
        response = self.call_method(enums.ClusterMethod.list_protection_domain_levels,
                                    result_dm_class=result_models.ListProtectionDomainLevels)
        return response.result_dm.protection_domain_levels

    def list_protection_domains(self):
        """List the protection domains on the cluster.

        The following cluster constant must be set for protection domains:
        "cEnableConfigurableProtectionDomains = True"

        :rtype: list[valence.models.PrivateProtectionDomain]
        """
        response = self.call_method(enums.ClusterMethod.list_protection_domains,
                                    result_dm_class=result_models.ListProtectionDomains)
        return response.result_dm.nodes

    def set_protection_domains(self, protection_domains):
        """Set the protection domains on the cluster.

        The following cluster constant must be set for protection domains:
        "cEnableConfigurableProtectionDomains = True"

        :param list[valence.models.PrivateProtectionDomain] protection_domains: A list of protection
            domain data models
        """
        params = {api_constants.NODES: [pd.serialize() for pd in protection_domains]}
        self.call_method(enums.ClusterMethod.set_protection_domains,
                         params=params,
                         result_dm_class=None)
