"""Cluster FM methods related to reports."""
import json
import re
import time

import autologger
import sf_artifacts
from sf_artifacts import errors as sf_errors

from sf_api import api_constants
from sf_api.models import errors
from valence import (
    criteria,
    enums,
    models,
)
from valence.functional_models._cluster import base

_log = autologger.get_sflogging()


class ReportMixin(base.ClusterBase):
    """Cluster FM report methods."""

    def get_report(self, report):
        """Get a JSON report from the cluster.

        :param valence.enums.JsonReport report: Enum value of desired report
        :rtype: dict
        """
        response = self.api.http_get('reports/{}'.format(report.value))
        data = self._json_cpp_load_string_hack(response.text)
        response.raise_for_status()
        return data

    def get_bin_sync_reports(self):
        """Get the Bin Reports for each bin.

        :rtype: list[valence.models.BinReport]
        """
        _log.debug('Deprecated, use get_bin_assingnments instead.')
        return self.get_bin_assignments().bins

    def get_slice_sync_report(self):
        """Get the Slice Report from the cluster.

        :rtype: valence.models.SliceReport
        """
        return models.SliceReport(self.get_report(enums.JsonReport.slices))

    def get_slice_balance_report(self):
        """Get the replica based slice balance report from the cluster.

        :rtype: valence.models.SliceBalanceReport
        """
        slice_bal_report = self.get_report(enums.JsonReport.slice_balance)
        return models.SliceBalanceReport(slice_bal_report)

    def get_report_raw(self, report):
        """Gets raw report.

        :param valence.enums.HtmlReport report: report to return
        :rtype: string
        """
        return self.api.http_get('reports/{}'.format(report.value)).content

    @staticmethod
    def _json_cpp_load_string_hack(data):
        """Attempt to fix json in reports as necessary if they contain inf or -inf.

        Python's standard JSON decoder doesn't accept "inf" or "-inf", but it does
        accept "Infinity" and "-Infinity".

        This was fixed in Sodium build 11.0.0.190. See https://solidfire.fogbugz.com/f/cases/13295.

        :param str data: Raw JSON string
        :rtype: dict
        """
        try:
            return json.loads(data)
        except ValueError:
            _log.warn('JsonCpp might be inserting invalid json into your report. '
                      'Attempting to work around this by replacing "inf" with "Infinity". '
                      'See case 13295.')
            # inf and -inf are values, and are preceded by a comma, bracket, or colon, and followed by
            # a comma, bracket, or curly brace.
            return json.loads(re.sub(r'([,\[:])\s*(-?)inf\s*', r'\1\2Infinity', data))

    def get_bin_assignments(self, validate=True,   # pylint: disable=too-complex
                            include_statuses=None,
                            exclude_statuses=None,
                            services=None
                            ):
        """Get the replica based bin assignments report from the cluster.

        :param bool validate: If true validate the balance and status of bins.
        :param list[valence.enums.BinStatus] include_statuses: Only include bins and service info for
            services matching these statuses.
        :param list[valence.enums.BinStatus] exclude_statuses: Exclude service info for services matching
            these statuses along with any bins with no other services.
        :param list[valence.models.ClusterServiceInfo] services: Only include bins and service info
            associated with these services.
        :rtype: valence.models.BinAssignments
        """
        bin_report = self.get_report(enums.JsonReport.bins)
        # Early format of bins report returns a list - put it in a dict
        if type(bin_report) is list:
            super_bin_report = {'bins': bin_report}
        # Later format of bins report returns a dictionary
        else:
            super_bin_report = bin_report
        bin_assignments = models.BinAssignments(super_bin_report)
        if validate:
            try:
                bin_assignments.validate()
            except errors.DataValidationError as dve:
                # Save the bin report with the errors as a text file.
                try:
                    manager = sf_artifacts.get_global_manager()
                    report_dict = bin_assignments.serialize()
                    report_json = json.dumps(report_dict)
                    prefix = time.strftime('bin_reports_%y%m%d_%H%M%S', time.localtime())
                    manager.store_text(report_json, prefix, '.json')
                except sf_errors.ArtifactsError:
                    _log.debug('Unable to save bin assignments, no artifact manager available')
                raise dve

        # Filter the report
        if include_statuses or exclude_statuses or services:
            service_ids = [s.service.service_id for s in services] if services else None
            bin_reports = []
            for bin_report in bin_assignments.bins:
                bin_service_filter = criteria.And(
                    criteria.ContainedIn(models.BinService.service_id, service_ids),
                    criteria.ContainedIn(models.BinService.status, include_statuses),
                    criteria.ContainedIn(models.BinService.status, exclude_statuses, negate=True)
                )
                bin_services = [s for s in bin_service_filter.filter(bin_report.services)]
                if bin_services:
                    bin_report.services = bin_services
                    bin_reports.append(bin_report)
            bin_assignments.bins = bin_reports

        return bin_assignments

    @property
    def slices_are_balanced(self):
        """Checks to see if a cluster is still balancing slices.

        :rtype: bool
        """
        # get slice report
        slices = self.get_slice_sync_report().slices

        # get slicebalance report
        slice_balance = self.get_slice_balance_report()

        syncing_slices = self.is_slice_syncing()

        # get number of slicebalance decisions
        num_decisions = len(slice_balance.decisions)

        primary_reassignments = []
        secondary_reassignments = []
        if not slices:
            return True

        for slc in slices:  # pylint: disable=not-an-iterable
            if slc.desired_primary == 0:
                pass
            elif slc.desired_primary == slc.primary:
                msg = ('Slice {} was balanced correctly, but the desiredPrimary was '
                       'not properly cleared. Marking this volumes balance as successful to work '
                       'around case 11650.')
                _log.debug(msg.format(slc.slice_id))
            else:
                primary_reassignments.append({slc.slice_id: {
                    api_constants.PRIMARY: slc.primary,
                    api_constants.DESIRED_PRIMARY: slc.desired_primary}})

            if len(slc.desired_secondaries) == 0:
                pass
            elif slc.desired_secondaries == slc.live_secondaries:
                msg = ('Slice {} was balanced correctly, but the '
                       'desiredSecondaries were not properly cleared. '
                       'Marking this volumes balance as successful to work around case 11650.')
                _log.debug(msg.format(slc.slice_id))
            else:
                secondary_reassignments.append({slc.slice_id: {
                    api_constants.LIVE_SECONDARIES: slc.live_secondaries,
                    api_constants.DESIRED_SECONDARIES: slc.desired_secondaries}})

        _log.debug('Desired Primary Reassignments: {}'.format(primary_reassignments))
        _log.debug('Desired Secondary Reassignments: {}'.format(secondary_reassignments))
        _log.debug('Syncing Slices: {}'.format(syncing_slices))

        return any([syncing_slices, num_decisions,
                    len(primary_reassignments), len(secondary_reassignments)])

    def is_slice_syncing(self, volume_dms=None):
        """Scrapes the slices.json report.

        Returns false if all volumes have a live primary, a single live secondary,
        and no dead secondaries. Defaults to all volumes if volume_ids is not specified.


        :param list[valence.models.Volume] volume_dms: Volume to wait for slice sync to finish on
        :rtype: bool
        """
        volume_ids = [volume_dm.volume_id for volume_dm in volume_dms] if volume_dms else None
        report = self.get_slice_sync_report()

        rep_count = self.get_cluster_info().rep_count
        expected_live_secondaries = rep_count - 1

        syncing = False
        syncing_slices = []

        if report:
            slices = report.slices
            for slice_ in slices:  # pylint: disable=not-an-iterable
                if volume_ids is None or slice_.volume_id in volume_ids:
                    if len(slice_.live_secondaries) != expected_live_secondaries:
                        syncing_slices.append(slice_.volume_id)
                        syncing = True
                    elif slice_.dead_secondaries and len(slice_.dead_secondaries) > 0:
                        syncing_slices.append(slice_.volume_id)
                        syncing = True
        else:
            syncing = False

        _log.debug('Syncing slices: {}'.format(set(syncing_slices)))

        return syncing
