"""Cluster functional model methods related to Schedules."""

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._cluster import base


class ScheduleMixin(base.ClusterBase):
    """Cluster FM schedule methods."""

    def create_schedule(self, schedule_name, schedule_info, schedule_type=enums.ScheduleType.snapshot,
                        monthdays=None, weekdays=None, recurring=None, paused=None, attributes=None,
                        starting_date=None, snapmirror_label=None, schedule_time=None):
        """Call CreateSchedule.

        The schedule_info should contain either "volumes" or "volume_id", which is a list of
        volume_ids or a single volume_id, respectively. It may also include a "name" (the name to give
        the snapshot), "enable_remote_replication" (whether the snapshot should be remotely replicated),
        and "retention" (the length of time the snapshot is kept in HH:MM:SS).

        The weekdays are described by "day", an integer value where 0 is Sunday and 6 is Saturday,
        and an "offset", an integer value for the occurrence in the month. If "offset" is 0,
        it will snapshot every week, if 1-4, it will snapshot on the 1st, 2nd, 3rd, or 4th occurrence
        of the day in the month.

        Some combination of hours, minutes, weekdays and monthdays are required. See the API guide for
        more information.

        :param str schedule_name: The name of the schedule
        :param valence.models.ScheduleInfo schedule_info: The schedule info. Describes what the
            snapshot is.
        :param valence.enums.ScheduleType schedule_type: The type of schedule. 'Snapshot' only supported.
        :param list[int] monthdays: The days of the month to schedule a snapshot
        :param list[valence.models.Weekday] weekdays: The days of the week to schedule a snapshot.
        :param bool recurring: If the schedule should be recurring or now
        :param bool paused: If it should be paused immediately.
        :param dict attributes: The attributes dictionary
        :param datetime.datetime starting_date: the ISO 8601 formatted timestamp to start.
        :param str snapmirror_label: Label used by snapmirror to specify snapshot retention on
            snapmirror endpoints.
        :param datetime.time schedule_time: Scheduled time for taking snapshots.

        :rtype: valence.models.Schedule

        """
        params = {
            api_constants.SCHEDULE_INFO: schedule_info.serialize(),
            api_constants.SCHEDULE_NAME: schedule_name,
            api_constants.SCHEDULE_TYPE: schedule_type,
            api_constants.MONTHDAYS: monthdays,
            api_constants.WEEKDAYS: (weekdays if weekdays is None else
                                     [wday.serialize() for wday in weekdays]),
            api_constants.RECURRING: recurring,
            api_constants.PAUSED: paused,
            api_constants.HOURS: schedule_time if schedule_time is None else schedule_time.hour,
            api_constants.MINUTES: schedule_time if schedule_time is None else schedule_time.minute,
            api_constants.ATTRIBUTES: attributes,
            api_constants.STARTING_DATE: starting_date,
            api_constants.SNAPMIRROR_LABEL: snapmirror_label
        }
        response = self.call_method(enums.ClusterMethod.create_schedule, params=params,
                                    result_dm_class=result_models.CreateSchedule)
        return response.result_dm.schedule

    def list_schedules(self):
        """Call ListSchedules.

        :rtype: list[valence.models.Schedule]
        """
        response = self.call_method(enums.ClusterMethod.list_schedules,
                                    result_dm_class=result_models.ListSchedules)
        return response.result_dm.schedules

    def modify_schedule(self, schedule_dm, schedule_name=None, schedule_info=None,
                        schedule_type=enums.ScheduleType.snapshot, monthdays=None, weekdays=None,
                        recurring=None, paused=None, hours=None, minutes=None, attributes=None,
                        starting_date=None, snapmirror_label=None, to_be_deleted=None):
        """Call ModifySchedule.

        :param valence.models.Schedule schedule_dm: Data model of Schedule you wish to modify.
        :param str schedule_name: The new name of the schedule
        :param dict[str,Any] schedule_info: The new schedule info. Describes what the snapshot is.
        :param valence.enums.ScheduleType schedule_type: The type of schedule. 'Snapshot' only supported.
        :param list[int] monthdays: The new days of the month to schedule a snapshot
        :param list[dict[str,int]] weekdays: The new days of the week to schedule a snapshot.
        :param bool recurring: If the schedule should be recurring or now
        :param bool paused: If it should be paused immediately.
        :param int hours: The new scheduled hours between taking snapshots
        :param int minutes: The new scheduled minutes between taking snapshots
        :param dict attributes: The new attributes dictionary
        :param str starting_date: The new ISO 8601 formatted timestamp to start
        :param str snapmirror_label: Label used by snapmirror to specify snapshot retention on
            snapmirror endpoints.
        :param bool to_be_deleted: Whether the schedule is marked for deletion.
        :rtype: valence.models.Schedule
        """

        params = {
            api_constants.SCHEDULE_ID: schedule_dm.schedule_id,
            api_constants.SCHEDULE_TYPE: schedule_type,
            api_constants.SCHEDULE_INFO: schedule_info,
            api_constants.SCHEDULE_NAME: schedule_name,
            api_constants.MONTHDAYS: monthdays,
            api_constants.WEEKDAYS: weekdays,
            api_constants.RECURRING: recurring,
            api_constants.PAUSED: paused,
            api_constants.HOURS: hours,
            api_constants.MINUTES: minutes,
            api_constants.ATTRIBUTES: attributes,
            api_constants.STARTING_DATE: starting_date,
            api_constants.SNAPMIRROR_LABEL: snapmirror_label,
            api_constants.TO_BE_DELETED: to_be_deleted,
        }
        response = self.call_method(enums.ClusterMethod.modify_schedule, params=params,
                                    result_dm_class=result_models.ModifySchedule)
        return response.result_dm.schedule
