"""Cluster FM methods related to security."""
from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models import _async_handler
from valence.functional_models._cluster import base


class SecurityMixin(base.ClusterBase):
    """Cluster FM Security methods."""

    def enable_encryption_at_rest(self, key_provider_dm=None, force_during_bin_sync=None):
        """Enable EAR via EnableEncryptionAtRest.

        :param valence.models.KeyProviderKmip key_provider_dm: Key provider DM
        :param bool force_during_bin_sync: Whether to enable Ear even during bin sync
        :rtype: valence.functional_models.EnableEarAsyncHandler
        """
        params = {api_constants.FORCE_DURING_BIN_SYNC: force_during_bin_sync}
        if key_provider_dm:
            params[api_constants.KEY_PROVIDER_ID] = key_provider_dm.id

        self.call_method(enums.ClusterMethod.enable_encryption_at_rest,
                         params=params, result_dm_class=result_models.EnableEncryptionAtRest)
        return _async_handler.EnableEarAsyncHandler(self)

    def disable_encryption_at_rest(self, force_during_bin_sync=None):
        """Disable EAR via DisableEncryptionAtRest.

        :param bool force_during_bin_sync: Whether to disable Ear even during bin sync
        :rtype: valence.functional_models.DisableEarAsyncHandler
        """
        params = {api_constants.FORCE_DURING_BIN_SYNC: force_during_bin_sync}
        self.call_method(enums.ClusterMethod.disable_encryption_at_rest,
                         params=params, result_dm_class=None)
        return _async_handler.DisableEarAsyncHandler(self)

    def get_encryption_at_rest_info(self):
        """Get encryption at rest information for the cluster.

        :rtype: valence.models.EncryptionAtRestInfo
        """
        response = self.call_method(enums.ClusterMethod.get_encryption_at_rest_info, params={},
                                    result_dm_class=result_models.GetEncryptionAtRestInfo)
        return response.result_dm.encryption_at_rest_info

    def add_key_server_to_provider_kmip(self, key_server_dm, key_provider_dm):
        """Associate the given key server with the key provider specified.

        :param valence.models.KeyServerKmip key_server_dm: Key server DM
        :param valence.models.KeyProviderKmip key_provider_dm: Key provider DM
        """
        params = {
            api_constants.KEY_SERVER_ID: key_server_dm.id,
            api_constants.KEY_PROVIDER_ID: key_provider_dm.id
        }
        self.call_method(enums.ClusterMethod.add_key_server_to_provider_kmip, params=params,
                         result_dm_class=None)

    def create_key_provider_kmip(self, key_provider_name):
        """Create a key provider on the cluster.

        :param str key_provider_name: Name of key provider
        :rtype: valence.models.KeyProviderKmip
        """
        params = {api_constants.KEY_PROVIDER_NAME: key_provider_name}
        response = self.call_method(enums.ClusterMethod.create_key_provider_kmip, params=params,
                                    result_dm_class=result_models.CreateKeyProviderKmip)
        return response.result_dm.kmip_key_provider

    def create_key_server_kmip(self, key_server_name, key_server_hostnames, client_certificate,
                               ca_certificate, key_server_port=5696):
        """Create a key server on the cluster.

        :param str key_server_name: Name of key server
        :param list[str] key_server_hostnames: List of key server hostnames
        :param str client_certificate: Client's certificate returned from key server
        :param str ca_certificate: Certificate authority's certificate
        :param int key_server_port: KMIP server port
        :rtype: valence.models.KeyServerKmip
        """
        params = {
            api_constants.KMIP_KEY_SERVER_NAME: key_server_name,
            api_constants.KMIP_KEY_SERVER_HOSTNAMES: key_server_hostnames,
            api_constants.KMIP_CLIENT_CERTIFICATE: client_certificate,
            api_constants.KMIP_CA_CERTIFICATE: ca_certificate,
            api_constants.KMIP_KEY_SERVER_PORT: key_server_port
        }
        response = self.call_method(enums.ClusterMethod.create_key_server_kmip, params=params,
                                    result_dm_class=result_models.CreateKeyServerKmip)
        return response.result_dm.kmip_key_server

    def create_public_private_key_pair(self, certificate_subject_dm=None):
        """Create a public/private key pair on the cluster to be used when generating a CSR.

        :param valence.models.CertificateSubject certificate_subject_dm: Subject DM
        """
        params = certificate_subject_dm.serialize() if certificate_subject_dm else {}
        self.call_method(enums.ClusterMethod.create_public_private_key_pair, params=params,
                         result_dm_class=None)

    def delete_key_provider_kmip(self, key_provider_dm):
        """Delete the key provider on the cluster.

        :param valence.models.KeyProviderKmip key_provider_dm: Key provider DM
        """
        params = {api_constants.KEY_PROVIDER_ID: key_provider_dm.id}
        self.call_method(enums.ClusterMethod.delete_key_provider_kmip, params=params,
                         result_dm_class=None)

    def delete_key_server_kmip(self, key_server_dm):
        """Delete the key server on the cluster.

        :param valence.models.KeyServerKmip key_server_dm: Key server DM
        """
        params = {api_constants.KEY_SERVER_ID: key_server_dm.id}
        self.call_method(enums.ClusterMethod.delete_key_server_kmip, params=params,
                         result_dm_class=None)

    def get_client_certificate_sign_request(self):
        """Generate a CSR for signing by a Certificate Authority.

        :rtype: sf_api.result_models.GetClientCertificateSignRequest
        """
        response = self.call_method(enums.ClusterMethod.get_client_certificate_sign_request, params={},
                                    result_dm_class=result_models.GetClientCertificateSignRequest)
        return response.result_dm

    def get_key_provider_kmip(self, key_provider_id):
        """Get the key provider with the given ID from the cluster.

        :param int key_provider_id: Key provider ID
        :rtype: valence.models.KeyProviderKmip
        """
        params = {api_constants.KEY_PROVIDER_ID: key_provider_id}
        response = self.call_method(enums.ClusterMethod.get_key_provider_kmip, params=params,
                                    result_dm_class=result_models.GetKeyProviderKmip)
        return response.result_dm.kmip_key_provider

    def get_key_server_kmip(self, key_server_id):
        """Get the key server with the given ID from the cluster.

        :param int key_server_id: Key server ID
        :rtype: valence.models.KeyServerKmip
        """
        params = {api_constants.KEY_SERVER_ID: key_server_id}
        response = self.call_method(enums.ClusterMethod.get_key_server_kmip, params=params,
                                    result_dm_class=result_models.GetKeyServerKmip)
        return response.result_dm.kmip_key_server

    def list_key_providers_kmip(self, key_provider_is_active=None,
                                kmip_key_provider_has_server_assigned=None):
        """List out the key providers on the cluster.

        :param bool key_provider_is_active: Filter on active key providers
        :param bool kmip_key_provider_has_server_assigned: Filter on key providers that have
                                                           associated key servers.
        :rtype: list[valence.models.KeyProviderKmip]
        """
        params = {
            api_constants.KEY_PROVIDER_IS_ACTIVE: key_provider_is_active,
            api_constants.KMIP_KEY_PROVIDER_HAS_SERVER_ASSIGNED: kmip_key_provider_has_server_assigned
        }
        response = self.call_method(enums.ClusterMethod.list_key_providers_kmip, params=params,
                                    result_dm_class=result_models.ListKeyProvidersKmip)
        return response.result_dm.kmip_key_providers

    def list_key_servers_kmip(self, key_provider_id=None, kmip_assigned_provider_is_active=None,
                              kmip_has_provider_assigned=None):
        """List out the key servers on the cluster.

        :param int key_provider_id: Key provider ID
        :param bool kmip_assigned_provider_is_active: Filter on key servers that have an active
                                                      provider.
        :param bool kmip_has_provider_assigned: Filter on key servers that have a provider assigned
                                                to it.
        :rtype: list[valence.models.KeyServerKmip]
        """
        params = {
            api_constants.KEY_PROVIDER_ID: key_provider_id,
            api_constants.KMIP_ASSIGNED_PROVIDER_IS_ACTIVE: kmip_assigned_provider_is_active,
            api_constants.KMIP_HAS_PROVIDER_ASSIGNED: kmip_has_provider_assigned
        }
        response = self.call_method(enums.ClusterMethod.list_key_servers_kmip, params=params,
                                    result_dm_class=result_models.ListKeyServersKmip)
        return response.result_dm.kmip_key_servers

    def modify_key_server_kmip(self, key_server_dm, key_server_name=None, key_server_hostnames=None,
                               client_certificate=None, ca_certificate=None, key_server_port=None):
        """Modify the attributes of the given key server on the cluster.

        :param valence.models.KeyServerKmip key_server_dm: Key server DM
        :param str key_server_name: Key server name
        :param list[str] key_server_hostnames: List of hostnames
        :param str client_certificate: Client's certificate
        :param str ca_certificate: Certificate authority's certificate
        :param int key_server_port: KMIP server port
        :rtype: valence.models.KeyServerKmip
        """
        params = {
            api_constants.KEY_SERVER_ID: key_server_dm.id,
            api_constants.KMIP_KEY_SERVER_NAME: key_server_name,
            api_constants.KMIP_KEY_SERVER_HOSTNAMES: key_server_hostnames,
            api_constants.KMIP_CLIENT_CERTIFICATE: client_certificate,
            api_constants.KMIP_CA_CERTIFICATE: ca_certificate,
            api_constants.KMIP_KEY_SERVER_PORT: key_server_port
        }
        response = self.call_method(enums.ClusterMethod.modify_key_server_kmip, params=params,
                                    result_dm_class=result_models.ModifyKeyServerKmip)
        return response.result_dm.kmip_key_server

    def remove_key_server_from_provider_kmip(self, key_server_dm):
        """Remove the associated key provider from the given key server.

        :param valence.models.KeyServerKmip key_server_dm: Key server DM
        """
        params = {api_constants.KEY_SERVER_ID: key_server_dm.id}
        self.call_method(enums.ClusterMethod.remove_key_server_from_provider_kmip, params=params,
                         result_dm_class=None)

    def test_key_provider(self, key_provider_dm):
        """Run a set of tests against the key provider on the cluster.

        :param valence.models.KeyProviderKmip key_provider_dm: Key provider DM
        """
        params = {api_constants.KEY_PROVIDER_ID: key_provider_dm.id}
        self.call_method(enums.ClusterMethod.test_key_provider_kmip, params=params,
                         result_dm_class=None)

    def test_key_server(self, key_server_dm):
        """Run a set of tests against the key server on the cluster.

        :param valence.models.KeyServerKmip key_server_dm: Key server DM
        """
        params = {api_constants.KEY_SERVER_ID: key_server_dm.id}
        self.call_method(enums.ClusterMethod.test_key_server_kmip, params=params,
                         result_dm_class=None)
