"""Cluster FM methods related to SnapMirror."""

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._cluster import base


class SnapmirrorMixin(base.ClusterBase):
    """Cluster FM SnapMirror methods."""

    def break_snapmirror_volume(self, volume_dm, snapshot_id=None, preserve=None, access=None):
        """Call BreakSnapMirrorVolume.

        Breaks the SnapMirror relationship between Ontap source container and SolidFire target volume.
        In the event that the Ontap cluster is inaccessible this allows the user to rollback the
        SolidFire volume to a snapshot and make it read-write.

        :param valence.models.Volume volume_dm: The volume on which to perform the break operation.
            The volume access mode must snapMirrorTarget.
        :param int snapshot_id: Optionally, rollback the volume to this snapshot. The default behavior
            is to rollback the most recent snapshot.
        :param bool preserve: If preserve is false then any snapshot newer than the snapshotID will be
            deleted. (Default is False)
        :param valence.enums.SnapMirrorAccessMode access: Resulting volume access mode,
            readWrite(default)|readOnly|locked
        """
        params = {
            api_constants.VOLUME_ID: volume_dm.volume_id,
            api_constants.SNAPSHOT_ID: snapshot_id,
            api_constants.PRESERVE: preserve,
            api_constants.ACCESS: access
        }
        self.call_method(enums.ClusterMethod.break_snapmirror_volume, params=params,
                         result_dm_class=None)

    def break_snapmirror_volumes(self, volume_dms, access=None):
        """Call BreakSnapMirrorVolumes.

        Performs BreakSnapMirrorVolume on multiple volumes.

        :param list[valence.models.Volume] volume_dms: List of volume ids on which to perform the break
             operation. The volume access of these volumes must be snapMirrorTarget.
        :param valence.enums.SnapMirrorAccessMode access: Resulting volume access mode,
            readWrite(default)|readOnly|locked
        """
        params = {
            api_constants.VOLUME_IDS: [vol.volume_id for vol in volume_dms],
            api_constants.ACCESS: access
        }
        self.call_method(enums.ClusterMethod.break_snapmirror_volumes, params=params,
                         result_dm_class=None)

    def create_snapmirror_endpoint(self, name, ip_addresses):
        """Call CreateSnapMirrorEndpoint.

        :param str name: Name for the snapmirror endpoint cluster
        :param list[str] ip_addresses: List of IP Addresses for a cluster of SnapMirror compatible
            storage systems that will need to communicate with this SolidFire cluster.
        :rtype: valence.models.SnapMirrorEndpoint
        """
        params = {
            api_constants.NAME: name,
            api_constants.IP_ADDRESSES: ip_addresses
        }
        response = self.call_method(enums.ClusterMethod.create_snapmirror_endpoint, params=params,
                                    result_dm_class=result_models.CreateSnapMirrorEndpoint)
        return response.result_dm.snapmirror_endpoint

    def delete_snapmirror_endpoint(self, snapmirror_endpoint_dm):
        """Call DeleteSnapMirrorEndpoint.

        :param valence.models.SnapMirrorEndpoint snapmirror_endpoint_dm: The SnapMirror Endpoint DM for
             the endpoint you wish to delete.
        """
        params = {api_constants.SNAPMIRROR_ENDPOINT_ID: snapmirror_endpoint_dm.snapmirror_endpoint_id}
        self.call_method(enums.ClusterMethod.delete_snapmirror_endpoint, params=params,
                         result_dm_class=None)

    def list_snapmirror_endpoints(self):
        """Call ListSnapMirrorEndpoints.

        :rtype: list[valence.models.SnapMirrorEndpoint]
        """
        response = self.call_method(enums.ClusterMethod.list_snapmirror_endpoints,
                                    result_dm_class=result_models.ListSnapMirrorEndpoints)
        return response.result_dm.snapmirror_endpoints

    def modify_snapmirror_endpoint(self, snapmirror_endpoint_dm, name=None, ip_addresses=None):
        """Call ModifySnapMirrorEndpoint.

        :param valence.models.SnapMirrorEndpoint snapmirror_endpoint_dm: The SnapMirror Endpoint ID for
             the endpoint you wish to modify.
        :param str name: The new name for the existing SnapMirror Endpoint
        :param list[str] ip_addresses: The new ip_addresses for the snapmirror endpoint
        :rtype: valence.models.SnapMirrorEndpoint
        """
        params = {
            api_constants.SNAPMIRROR_ENDPOINT_ID: snapmirror_endpoint_dm.snapmirror_endpoint_id,
            api_constants.NAME: name,
            api_constants.IP_ADDRESSES: ip_addresses
        }
        response = self.call_method(enums.ClusterMethod.modify_snapmirror_endpoint, params=params,
                                    result_dm_class=result_models.ModifySnapMirrorEndpoint)
        return response.result_dm.snapmirror_endpoint
