"""Cluster FM methods related to snapshots."""
from __future__ import division

from datetime import timedelta

from sf_models import types

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._cluster import base


class SnapshotMixin(base.ClusterBase):
    """Cluster FM Snapshot methods."""

    def create_snapshot(self, volume_dm, snapshot_dm=None, name=None, enable_remote_replication=None,
                        retention=None, attributes=None):
        """Create a snapshot on a cluster via CreateSnapshot.

        :param valence.models.Volume volume_dm: ID of source volume.
        :param valence.models.Snapshot snapshot_dm: The snapshot of the same volume
               of which to make a snapshot
        :param str name: Name of the snapshot.
        :param datetime.timedelta retention: Timedelta object describing lifetime of snapshot; if
               omitted the snapshot will not expire and must be deleted manually.
        :param bool enable_remote_replication: If true snapshot will replicate to remote storage.
        :param dict attributes: Any additional key value pairs to be added to the account.

        :return: Snapshot as returned by API
        :rtype: valence.models.Snapshot
        """
        retention_str = None
        if retention:
            hours = int(retention.total_seconds() / 60 ** 2)
            minutes = int((retention - timedelta(hours=hours)).total_seconds() / 60)
            seconds = int((retention - timedelta(hours=hours, minutes=minutes)).total_seconds())
            retention_str = '{}:{}:{}'.format(str(hours).zfill(2),
                                              str(minutes).zfill(2),
                                              str(seconds).zfill(2))
        params = {
            api_constants.VOLUME_ID: volume_dm.volume_id,
            api_constants.NAME: name,
            api_constants.RETENTION: retention_str,
            api_constants.ENABLE_REMOTE_REPLICATION: enable_remote_replication,
            api_constants.ATTRIBUTES: attributes,
        }

        if snapshot_dm:
            params[api_constants.SNAPSHOT_ID] = snapshot_dm.snapshot_id

        response = self.call_method(enums.ClusterMethod.create_snapshot, params=params,
                                    result_dm_class=result_models.CreateSnapshot)
        return response.result_dm.snapshot

    def create_group_snapshot(self, volume_dms, name=None,
                              enable_remote_replication=None, retention=None, attributes=None):
        """Create a snapshot of a group of volumes using CreateGroupSnapshot.

        :param list[valence.models.Volume] volume_dms: Source volume data models
        :param str name: Name of the snapshot.
        :param datetime.timedelta retention: Timedelta object describing lifetime of snapshot; if
               omitted the snapshot will not expire and must be deleted manually.
        :param bool enable_remote_replication: If true snapshot will replicate to remote storage.
        :param dict attributes: Any additional key value pairs to be added to the account.

        :return: Group snapshot as returned by API
        :rtype: valence.models.GroupSnapshot
        """
        retention_str = None
        if retention:
            hours = int(retention.total_seconds() / 60 ** 2)
            minutes = int((retention - timedelta(hours=hours)).total_seconds() / 60)
            seconds = int((retention - timedelta(hours=hours, minutes=minutes)).total_seconds())
            retention_str = '{}:{}:{}'.format(str(hours).zfill(2),
                                              str(minutes).zfill(2),
                                              str(seconds).zfill(2))
        params = {
            api_constants.VOLUMES: [v.volume_id for v in volume_dms],
            api_constants.NAME: name,
            api_constants.RETENTION: retention_str,
            api_constants.ENABLE_REMOTE_REPLICATION: enable_remote_replication,
            api_constants.ATTRIBUTES: attributes,
        }

        response = self.call_method(enums.ClusterMethod.create_group_snapshot, params=params,
                                    result_dm_class=result_models.CreateGroupSnapshot)
        return response.result_dm.group_snapshot

    def list_group_snapshots(self, volume_dms=None, group_snapshot_id=None):
        """Returns a list of group snapshots via ListGroupSnapshots.

        :param list[valence.models.Volume] volume_dms: Source volume data models
        :param int group_snapshot_id: Group Snapshot ID

        :rtype: list[valence.models.GroupSnapshot]
        """

        params = {api_constants.VOLUMES: [v.volume_id for v in volume_dms] if volume_dms else None,
                  api_constants.GROUP_SNAPSHOT_ID: group_snapshot_id}

        response = self.call_method(enums.ClusterMethod.list_group_snapshots, params=params,
                                    result_dm_class=result_models.ListGroupSnapshots)
        return response.result_dm.group_snapshots

    def list_snapshots(self, volume_dm=None, internal=False):
        """Returns a list of all the snapshots via ListSnapshots.

        This list is (including volume default snapshots),  optionally filtered for a particular volume.

        :param valence.models.Volume volume_dm: filter return data to snapshots from a particular volume
        :param bool internal: whether to show internal snapshots also
        :rtype: list[valence.models.Snapshot]
        """
        params = {
            api_constants.INTERNAL: internal,
            api_constants.VOLUME_ID: volume_dm.volume_id if volume_dm else None
        }

        response = self.call_method(enums.ClusterMethod.list_snapshots, params=params,
                                    result_dm_class=result_models.ListSnapshots)
        return response.result_dm.snapshots

    def delete_group_snapshot(self, group_snapshot_dm, save_members=False):
        """Delete a group snapshot.

        :param valence.models.GroupSnapshot group_snapshot_dm: Group snapshot data model.
        :param bool save_members: Whether to save the constituent snapshots in a group snapshot.
        """
        params = {
            api_constants.GROUP_SNAPSHOT_ID: group_snapshot_dm.group_snapshot_id,
            api_constants.SAVE_MEMBERS: save_members
        }

        self.call_method(enums.ClusterMethod.delete_group_snapshot, params=params,
                         result_dm_class=None)

    def delete_snapshot(self, snapshot_dm):
        """Delete a snapshot.

        :param valence.models.Snapshot snapshot_dm: Snapshot data model.
        """
        params = {api_constants.SNAPSHOT_ID: snapshot_dm.snapshot_id}

        self.call_method(enums.ClusterMethod.delete_snapshot, params=params,
                         result_dm_class=None)

    def rollback_to_group_snapshot(self, group_snapshot_dm, attributes=None, name=None,
                                   save_current_state=None):
        """Rollback all volumes in a group snapshot to each volume's individual snapshot.

        When performing a rollback, a temporary snapshot is created for each
        volume before the rollback begins.

        :param valence.models.GroupSnapshot group_snapshot_dm: Group snapshot data model to perform
            rollback from
        :param dict attributes: Additional key/value pairs to add to the temporary group snapshot
        :param str name: Name for the temporary group snapshot if save_current_state is set to True
        :param bool save_current_state: Whether to save the temporary group snapshot created of the
            current volume state before performing a rollback
        :rtype: sf_api.result_models.RollbackToGroupSnapshot
        """
        params = {api_constants.GROUP_SNAPSHOT_ID: group_snapshot_dm.group_snapshot_id,
                  api_constants.ATTRIBUTES: attributes,
                  api_constants.NAME: name,
                  api_constants.SAVE_CURRENT_STATE: save_current_state}

        response = self.call_method(enums.ClusterMethod.rollback_to_group_snapshot, params=params,
                                    result_dm_class=result_models.RollbackToGroupSnapshot)

        return response.result_dm

    def rollback_to_snapshot(self, snapshot_dm, attributes=None, name=None,
                             save_current_state=None):
        """Rollback a volume from a snapshot.

        When performing a rollback, a temporary snapshot is created from the current volume before
        the rollback begins.

        :param valence.models.Snapshot snapshot_dm: Snapshot data model to perform rollback from
        :param dict attributes: Additional key/value pairs to add to the temporary snapshot
        :param str name: Name for the temporary snapshot if save_current_state is set to True
        :param bool save_current_state: Whether to save the temporary snapshot created of the current
            volume state before performing a rollback
        :rtype: sf_api.result_models.RollbackToSnapshot
        """
        params = {api_constants.VOLUME_ID: snapshot_dm.volume_id,
                  api_constants.SNAPSHOT_ID: snapshot_dm.snapshot_id,
                  api_constants.ATTRIBUTES: attributes,
                  api_constants.NAME: name,
                  api_constants.SAVE_CURRENT_STATE: save_current_state}

        response = self.call_method(enums.ClusterMethod.rollback_to_snapshot, params=params,
                                    result_dm_class=result_models.RollbackToSnapshot)

        return response.result_dm

    def modify_snapshot(self, snapshot_dm, enable_remote_replication=None, expiration_time=None):
        """Modify a Snapshot.

        :param valence.models.Snapshot snapshot_dm: Snapshot to modify
        :param bool enable_remote_replication: whether to enable/disable remote replication
        :param datetime.datetime expiration_time: new expiration time

        :rtype: list[valence.models.Snapshot]
        """

        # TODO (VL-2516): If DateTimeType handles None, this code can be simplified.
        if expiration_time:
            expiration_time_str = types.DateTimeType().to_primitive(expiration_time)
        else:
            expiration_time_str = None

        params = {
            api_constants.SNAPSHOT_ID: snapshot_dm.snapshot_id,
            api_constants.ENABLE_REMOTE_REPLICATION: enable_remote_replication,
            api_constants.EXPIRATION_TIME: expiration_time_str,
        }

        response = self.call_method(enums.ClusterMethod.modify_snapshot, params=params,
                                    result_dm_class=result_models.ModifySnapshot)
        return response.result_dm.snapshot

    def modify_group_snapshot(self, group_snapshot_dm, enable_remote_replication=None,
                              expiration_time=None):
        """Modify a Group Snapshot.

        :param valence.models.GroupSnapshot group_snapshot_dm: Group Snapshot to modify
        :param bool enable_remote_replication: whether to enable/disable remote replication
        :param datetime.datetime expiration_time: new expiration time

        :rtype: list[valence.models.GroupSnapshot]
        """

        # TODO (VL-2516): If DateTimeType handles None, this code can be simplified.
        if expiration_time:
            expiration_time_str = types.DateTimeType().to_primitive(expiration_time)
        else:
            expiration_time_str = None

        params = {
            api_constants.GROUP_SNAPSHOT_ID: group_snapshot_dm.group_snapshot_id,
            api_constants.ENABLE_REMOTE_REPLICATION: enable_remote_replication,
            api_constants.EXPIRATION_TIME: expiration_time_str,
        }

        response = self.call_method(enums.ClusterMethod.modify_group_snapshot, params=params,
                                    result_dm_class=result_models.ModifyGroupSnapshot)
        return response.result_dm.group_snapshot
