import unittest

import mock
import pytest

import sf_api
from sf_api import result_models
from valence import (
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterAccountMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.account_dm = models.Account.get_mock_object()

    def test_create_account_pre_fluorine(self, mock_call_method):
        """cluster_api.add_account only returns the account ID. get_account_by_id is mocked because
        add_account will only return an account_id. FM.create_account needs to call get_account_by_id to
        get the account_object.
        """
        mock_create_response = result_models.AddAccount.get_mock_response(
            version=sf_api.SfVersion('8.0'), overrides=dict(account_id=self.account_dm.account_id))
        mock_get_response = result_models.GetAccountById.get_mock_response(
            overrides=dict(account=self.account_dm))
        mock_call_method.side_effect = [mock_create_response, mock_get_response]
        response = self.cluster_fm.create_account(self.account_dm.username)
        assert response == self.account_dm

    def test_create_account_fluorine(self, mock_call_method):
        """cluster_api.add_account returns the account ID and account object."""
        mock_add_data = result_models.AddAccount.get_mock_data(version=sf_api.SfVersion('9.0'))
        mock_call_method.return_value = result_models.AddAccount.get_mock_response(
            overrides=mock_add_data)
        self.cluster_fm.create_account(self.account_dm.username)

    def test_delete_accounts(self, mock_call_method):
        mock_response = result_models.BaseApiResult.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.delete_accounts([self.account_dm])

    def test_modify_account(self, mock_call_method):
        """cluster_api.add_account returns the account ID and account object."""
        mock_modify_data = result_models.ModifyAccount.get_mock_data(version=sf_api.SfVersion('9.0'))
        mock_call_method.return_value = result_models.ModifyAccount.get_mock_response(
            overrides=mock_modify_data)
        self.cluster_fm.modify_account(self.account_dm.username, initiator_secret='abcdefghijklmn')

    def test_get_account_by_id(self, mock_call_method):
        account_dm = models.Account.get_mock_object(overrides=dict(account_id=2))
        mock_response = result_models.GetAccountById.get_mock_response(
            overrides=dict(account=account_dm))
        mock_call_method.return_value = mock_response
        acc = self.cluster_fm.get_account_by_id(account_id=2)
        assert acc == account_dm


@mock.patch('sf_api.ClusterApi.call_method')
class TestClusterListAccounts(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.rs_cluster_dm = models.ResourceCluster.get_mock_object(
            overrides=dict(api_endpoint=models.SfVersion('9.0')))
        cls.cluster_fm = functional_models.Cluster(cls.rs_cluster_dm, authenticate=False)
        cls.account_dm1 = models.Account.get_mock_object(overrides=dict(account_id=1, username='alex',
                                                                        status='active'))
        cls.account_dm2 = models.Account.get_mock_object(overrides=dict(account_id=2, username='bob',
                                                                        status='active'))
        cls.mock_accounts = [cls.account_dm1, cls.account_dm2]
        cls.mock_response = result_models.ListAccounts.get_mock_response(
            overrides=dict(accounts=cls.mock_accounts))

    def test_mutually_exclusive_params_fail(self, mock_call_method):
        with pytest.raises(ValueError):
            self.cluster_fm.list_accounts(account_name='alex',
                                          account_status=sf_api.AccountStatus.active)

    def test_no_filters(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        assert len(self.cluster_fm.list_accounts()) == 2

    def test_filter_account_name(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        accounts = self.cluster_fm.list_accounts(account_name='alex')
        assert accounts[0] == self.account_dm1

    def test_filter_account_ids(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        accounts = self.cluster_fm.list_accounts(account_ids=[1])
        assert accounts[0] == self.account_dm1

    def test_filter_account_status(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        accounts = self.cluster_fm.list_accounts(account_status=sf_api.AccountStatus.active)
        assert len(accounts) == 2

    def test_filter_account_storage_container(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.list_accounts(include_storage_containers=False)
