import unittest

import mock
import pytest

from sf_api import (
    api_constants,
    result_models,
    sf_version,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterAsyncMixin(unittest.TestCase):
    """Test that the logic for converting the success and completed bools to a status."""

    def setUp(self):
        # The unit tests mutate the cluster_fm, so the cannot be done in setUpClass
        self.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_async_result(self, mock_call_method):
        """Test that list_async_results will still validate."""
        self.cluster_fm._validate_responses = True
        self.cluster_fm.cluster_dm.api_endpoint = sf_version.SfVersion('10.0')
        async_handle = models.AsyncHandle({api_constants.ASYNC_HANDLE: 1})
        mock_call_method.return_value.result_dm = models.AsyncResult.get_mock_object(
            overrides=dict(details=models.AddDrivesDetails.get_mock_object()))
        async_result = self.cluster_fm.get_async_result(async_handle)
        assert isinstance(async_result, models.AsyncResult)

    def test_list_async_results_running(self, mock_call_method):
        async_result_async_handle = models.AsyncResult.get_mock_data(overrides=dict(completed=False,
                                                                                    success=False))
        list_async_results = result_models.ListAsyncResults.get_mock_object(
            overrides=dict(async_handles=[async_result_async_handle]))
        mock_call_method.return_value.result_dm = list_async_results
        list_results = self.cluster_fm.list_async_results()
        assert len(list_results) == 1
        for ar in list_results:
            assert isinstance(ar, models.AsyncResult)
            assert ar.status is enums.AsyncResultStatus.running

    def test_list_async_results_completed(self, mock_call_method):
        """Test that list_async_results will still validate."""
        self.cluster_fm._validate_responses = True
        self.cluster_fm.cluster_dm.api_endpoint = models.SfVersion('10.0')

        async_result_async_handle = models.AsyncResult.get_mock_data(
            overrides=dict(completed=True, success=True,
                           details=models.AddDrivesDetails.get_mock_object()))
        list_async_results = result_models.ListAsyncResults.get_mock_object(
            overrides=dict(async_handles=[async_result_async_handle]))
        mock_call_method.return_value.result_dm = list_async_results

        list_results = self.cluster_fm.list_async_results()
        assert len(list_results) == 1
        for ar in list_results:
            assert isinstance(ar, models.AsyncResult)
            assert ar.status is enums.AsyncResultStatus.complete

    def test_list_async_results_failed(self, mock_call_method):
        async_result_async_handle = models.AsyncResult.get_mock_data(
            overrides=dict(completed=True, success=False))
        list_async_results = result_models.ListAsyncResults.get_mock_object(
            overrides=dict(async_handles=[async_result_async_handle]))
        mock_call_method.return_value.result_dm = list_async_results

        list_results = self.cluster_fm.list_async_results()
        assert len(list_results) == 1
        for ar in list_results:
            assert isinstance(ar, models.AsyncResult)
            assert ar.status is enums.AsyncResultStatus.failed

    def test_list_async_results_impossible(self, mock_call_method):
        """Test that list async result data can never be running and succeeded."""
        async_result_async_handle = models.AsyncResult.get_mock_data(
            overrides=dict(completed=False, success=True))
        list_async_results = result_models.ListAsyncResults.get_mock_object(
            overrides=dict(async_handles=[async_result_async_handle]))
        mock_call_method.return_value.result_dm = list_async_results

        with pytest.raises(functional_models._errors.FunctionalClusterError):
            self.cluster_fm.list_async_results()
