import datetime
import unittest

import mock

from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
    utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterAuthSessionMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.cluster_admin_dm = models.ClusterAdmin.get_mock_object()
        cls.auth_session_dm = models.AuthSessionInfo.get_mock_object()

    @mock.patch.object(utils, 'get_now_utc')
    def test_create_auth_session(self, mock_utc, mock_call_method):
        """cluster_api.create_auth_session returns AuthSessionInfo"""

        # Set the test_time to the epoch, so we get a easy value to calculate
        test_time = datetime.datetime(1970, 1, 1)
        mock_utc.return_value = test_time

        mock_call_method.return_value = result_models.CreateAuthSessionResult.get_mock_response(
            overrides=dict(session=self.auth_session_dm))
        result = self.cluster_fm.create_auth_session(
            self.cluster_admin_dm,
            desired_expiration_duration=datetime.timedelta(hours=1)
        )

        assert result == self.auth_session_dm
        mock_call_method.assert_called_with(
            enums.ClusterMethod.create_auth_session, params={
                'clusterAdminID': self.cluster_admin_dm.cluster_admin_id,
                'desiredExpirationDuration': '3600'
            },
            result_dm_class=result_models.CreateAuthSessionResult
        )

    def test_delete_auth_sessions_single_session(self, mock_call_method):
        """cluster_api.delete_auth_sessions returns list[AuthSessionInfo]"""
        mock_call_method.return_value = result_models.DeleteAuthSessionsResult.get_mock_response(
            overrides=dict(sessions=[self.auth_session_dm]))
        result = self.cluster_fm.delete_auth_session(self.cluster_admin_dm, session=self.auth_session_dm)
        assert result == [self.auth_session_dm]
        mock_call_method.assert_called_with(
            enums.ClusterMethod.delete_auth_sessions, params={
                'clusterAdminID': self.cluster_admin_dm.cluster_admin_id,
                'sessionID': str(self.auth_session_dm.session_id),
            },
            result_dm_class=result_models.DeleteAuthSessionsResult
        )

    def test_delete_auth_sessions_per_user(self, mock_call_method):
        """cluster_api.delete_auth_sessions with a user returns list[AuthSessionInfo]"""
        mock_call_method.return_value = result_models.DeleteAuthSessionsResult.get_mock_response(
            overrides=dict(sessions=[self.auth_session_dm]))
        result = self.cluster_fm.delete_auth_session(self.cluster_admin_dm)
        assert result == [self.auth_session_dm]
        mock_call_method.assert_called_with(
            enums.ClusterMethod.delete_auth_sessions, params={
                'clusterAdminID': self.cluster_admin_dm.cluster_admin_id,
                'sessionID': None,
            },
            result_dm_class=result_models.DeleteAuthSessionsResult
        )

    def test_list_auth_sessions(self, mock_call_method):
        """cluster_api.list_auth_sessions returns list[AuthSessionInfo]"""
        mock_call_method.return_value = result_models.ListAuthSessionsResult.get_mock_response(
            overrides=dict(sessions=[self.auth_session_dm]))
        result = self.cluster_fm.list_auth_sessions()
        assert result == [self.auth_session_dm]
        mock_call_method.assert_called_with(
            enums.ClusterMethod.list_auth_sessions, params={
                'clusterAdminIDs': []
            },
            result_dm_class=result_models.ListAuthSessionsResult
        )

    def test_list_auth_sessions_single_user(self, mock_call_method):
        """cluster_api.list_auth_sessions with a specific cluster admin returns list[AuthSessionInfo]"""
        mock_call_method.return_value = result_models.ListAuthSessionsResult.get_mock_response(
            overrides=dict(sessions=[self.auth_session_dm]))
        result = self.cluster_fm.list_auth_sessions(cluster_admins=[self.cluster_admin_dm])
        assert result == [self.auth_session_dm]
        mock_call_method.assert_called_with(
            enums.ClusterMethod.list_auth_sessions, params={
                'clusterAdminIDs': [self.cluster_admin_dm.cluster_admin_id]
            },
            result_dm_class=result_models.ListAuthSessionsResult
        )

    def test_list_expired_auth_sessions(self, mock_call_method):
        """cluster_api.list_auth_sessions returns list[AuthSessionInfo]"""
        mock_call_method.return_value = result_models.ListAuthSessionsResult.get_mock_response(
            overrides=dict(sessions=[self.auth_session_dm]))
        result = self.cluster_fm.list_expired_auth_sessions()
        assert result == [self.auth_session_dm]
        mock_call_method.assert_called_with(
            enums.ClusterMethod.list_expired_auth_sessions,
            result_dm_class=result_models.ListAuthSessionsResult
        )

    def test_update_auth_session(self, mock_call_method):
        """cluster_api.update_auth_session with a user returns AuthSessionInfo"""
        mock_call_method.return_value = result_models.UpdateAuthSessionResult.get_mock_response(
            overrides=dict(session=self.auth_session_dm))
        result = self.cluster_fm.update_auth_session(self.auth_session_dm)
        assert result == self.auth_session_dm
        mock_call_method.assert_called_with(
            enums.ClusterMethod.update_auth_session, params={
                'sessionID': str(self.auth_session_dm.session_id)
            },
            result_dm_class=result_models.UpdateAuthSessionResult
        )
