import unittest

import mock
import requests

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


class TestClusterBase(unittest.TestCase):
    """Tests base methods on the cluster."""

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_from_resource_data(self):
        data = models.ResourceCluster.get_mock_data()
        cluster_fm = functional_models.Cluster.from_resource_data(data, authenticate=False)
        assert isinstance(cluster_fm, functional_models.Cluster)

    def test_cluster_repr(self):
        assert 'ClusterFm' in repr(self.cluster_fm)
        assert 'mvip=' in repr(self.cluster_fm)
        assert self.cluster_fm.mvip in repr(self.cluster_fm)

    def test_properties(self):
        assert self.cluster_fm.static_id is not None
        assert self.cluster_fm.mvip is not None
        assert self.cluster_fm.name is not None
        assert self.cluster_fm.svip is not None

    def test_upgrading_affects_multiple_objects(self):
        assert self.cluster_fm.upgrading is False
        self.cluster_fm.upgrading = True
        cluster_dm = models.ResourceCluster(self.cluster_fm.cluster_dm.serialize())
        new_cluster_fm = functional_models.Cluster(cluster_dm)
        assert new_cluster_fm is not self.cluster_fm
        assert new_cluster_fm.upgrading is True

    @mock.patch.object(functional_models.Cluster, 'get_api')
    def test_get_and_update_version(self, mock_get_api):
        stale_version = self.cluster_fm.api.version
        new_version_str = '{}.{}'.format(stale_version.version_tuple[0],
                                         stale_version.version_tuple[1] + 1)
        new_version = sf_api.SfVersion(new_version_str)
        mock_get_api.return_value = {
            api_constants.CURRENT_VERSION: str(new_version)
        }
        self.cluster_fm.update_api_version()
        assert self.cluster_fm.api.version == new_version
        assert self.cluster_fm.cluster_dm.api_endpoint == new_version

    @mock.patch.object(functional_models.Cluster, 'get_api')
    def test_api_responsive(self, mock_get_api):
        assert self.cluster_fm.api_is_responsive()
        mock_get_api.side_effect = [requests.HTTPError]
        assert not self.cluster_fm.api_is_responsive()

    @mock.patch.object(models.Base, 'validate')
    @mock.patch.object(sf_api.ClusterApi, 'call_method',
                       side_effect=unittest_utils.helpers.check_params_serializable)
    def test_call_method_while_upgrading(self, mock_call_method, mock_validate):
        self.cluster_fm._validate_responses = True

        # Set upgrading to True and make sure it passes version=None to validate
        self.cluster_fm.upgrading = True
        mock_call_method.return_value = result_models.ListDrives.get_mock_response()
        self.cluster_fm.call_method(enums.ClusterMethod.list_drives)
        mock_validate.assert_called_with(None)

        # Set upgrading back to False and make sure it passes in the version
        self.cluster_fm.upgrading = False
        self.cluster_fm.call_method(enums.ClusterMethod.list_drives)
        mock_validate.assert_called_with(self.cluster_fm.cluster_dm.api_endpoint)

        self.cluster_fm._validate_responses = False


@mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
@mock.patch.object(functional_models.Cluster, 'get_cluster_info')
class TestIsSliceSyncing(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.RS_CLUSTER_DM, authenticate=False)

    def setUp(self):
        self.slice_sync_report = models.SliceReport.get_mock_object(include_optional=True)
        self.slice_sync_report.slices.append(models.Slice.get_mock_object())

    def test_is_slice_syncing_false(self, mock_cluster_info,
                                    mock_slice_sync_report):
        mock_slice_sync_report.return_value = {}
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is False

    def test_is_slice_syncing_true_without_live_secondaries(self, mock_cluster_info,
                                                            mock_slice_sync_report):
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].live_secondaries = []
        mock_slice_sync_report.return_value = self.slice_sync_report
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is True

    def test_is_slice_syncing_true(self, mock_cluster_info,
                                   mock_slice_sync_report):
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].dead_secondaries = []
        mock_slice_sync_report.return_value = self.slice_sync_report
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is True

    def test_is_slice_syncing_true_with_live_secondaries(self, mock_cluster_info,
                                                         mock_slice_sync_report):
        self.cluster_fm.cluster_dm.rep_count = 1
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].live_secondaries = ['stuff']
        self.slice_sync_report.slices[0].dead_secondaries = []
        mock_slice_sync_report.return_value = self.slice_sync_report
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is True

    def test_is_slice_syncing_false_with_live_secondaries(self, mock_cluster_info,
                                                          mock_slice_sync_report):
        self.cluster_fm.cluster_dm.rep_count = 2
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].live_secondaries = ['stuff']
        self.slice_sync_report.slices[0].dead_secondaries = []
        self.slice_sync_report.slices = [self.slice_sync_report.slices[0]]
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        mock_slice_sync_report.return_value = self.slice_sync_report
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is False

    def test_is_slice_syncing_true_with_dead_secondaries(self, mock_cluster_info,
                                                         mock_slice_sync_report):
        self.cluster_fm.cluster_dm.rep_count = 3
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].live_secondaries = ['stuff']
        self.slice_sync_report.slices[0].dead_secondaries = ['things']
        mock_slice_sync_report.return_value = self.slice_sync_report
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is True

    def test_is_slice_syncing_false_with_volumes(self, mock_cluster_info,
                                                 mock_slice_sync_report):
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].live_secondaries = ['stuff']
        self.slice_sync_report.slices[0].dead_secondaries = []
        self.slice_sync_report.slices[1].volume_id = 2
        self.slice_sync_report.slices[1].live_secondaries = ['stuff']
        self.slice_sync_report.slices[1].dead_secondaries = ['more', 'things']
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing()
        assert slice_sync is False

    def test_is_slice_syncing_true_with_volumes(self, mock_cluster_info,
                                                mock_slice_sync_report):
        self.slice_sync_report.slices[0].volume_id = 1
        self.slice_sync_report.slices[0].live_secondaries = ['stuff']
        self.slice_sync_report.slices[0].dead_secondaries = []
        self.slice_sync_report.slices[1].volume_id = 2
        self.slice_sync_report.slices[1].live_secondaries = ['more', 'stuff']
        self.slice_sync_report.slices[1].dead_secondaries = []
        mock_slice_sync_report.return_value = self.slice_sync_report
        mock_cluster_info.return_value = self.RS_CLUSTER_DM
        slice_sync = self.cluster_fm.is_slice_syncing(
            volume_dms=models.Volume.get_mock_object_list(1, overrides=dict(volume_id=2)))
        assert slice_sync is True


@mock.patch.object(sf_api.ClusterApi, 'call_method',
                   side_effect=unittest_utils.helpers.check_params_serializable)
class TestListNodeStats(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.RS_CLUSTER_DM, authenticate=False)

    def test_list_node_stats(self, mock_call_method):
        mock_response = result_models.ListNodeStats.get_mock_response()
        mock_call_method.return_value = mock_response

        result = self.cluster_fm.list_node_stats()
        assert isinstance(result, list)
        assert isinstance(result[0], models.NodeStats)
