import unittest

import mock
import pytest

import sf_api
from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestHardwareMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.node_dm = models.ResourceNode.get_mock_object()

    def setUp(self):
        self.mock_response = result_models.GetHardwareInfoOverrides.get_mock_response()

    def test_get_hardware_info_overrides(self, mock_call_method):
        mock_response = result_models.GetHardwareInfoOverrides.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_hardware_info_overrides()
        assert isinstance(result, models.HardwareInfoOverrides)

    def test_set_info(self, mock_call_method):
        mock_response = result_models.GetHardwareInfoOverrides.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.set_hardware_info_overrides(chassis_serial_overrides={'1': {'a': '1'}})
        assert isinstance(result, models.HardwareInfoOverrides)

    def test_get_hardware_info(self, mock_call_method):
        mock_response = result_models.GetClusterHardwareInfo.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_cluster_hardware_info()
        assert isinstance(result, models.ClusterHardwareInfo)

    def test_get_drive_info(self, mock_call_method):
        mock_response = result_models.GetClusterHardwareInfo.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_cluster_hardware_info(
            hw_type=enums.ClusterHardwareInfoType.drives.value)
        assert isinstance(result, models.ClusterHardwareInfo)

    def test_check_proposed_node_additions(self, mock_call_method):
        mock_response = result_models.CheckProposedNodeAdditions.get_mock_response(
            include_optional=True, overrides={'proposed_cluster_errors': []},
            version=models.SfVersion('11.0'))
        mock_call_method.return_value = mock_response
        check_proposed_node_additions_dm = self.cluster_fm.check_proposed_node_additions(
            [self.node_dm],
            raise_on_failure=True
        )
        assert isinstance(check_proposed_node_additions_dm,
                          sf_api.result_models.CheckProposedNodeAdditions)
        assert mock_call_method.called

    def test_check_proposed_node_additions_raises(self, mock_call_method):
        mock_response = result_models.CheckProposedNodeAdditions.get_mock_response(
            include_optional=False, version=models.SfVersion('11.0'))
        mock_call_method.return_value = mock_response
        with pytest.raises(functional_models._errors.FunctionalClusterError):
            self.cluster_fm.check_proposed_node_additions([self.node_dm], raise_on_failure=True)
        assert mock_call_method.called

    def test_check_proposed_node_additions_doesnt_raise(self, mock_call_method):
        mock_response = result_models.CheckProposedNodeAdditions.get_mock_response(
            include_optional=True, version=models.SfVersion('11.0'))
        mock_call_method.return_value = mock_response
        check_proposed_node_additions_dm = self.cluster_fm.check_proposed_node_additions(
            [self.node_dm],
            raise_on_failure=False
        )
        assert isinstance(check_proposed_node_additions_dm,
                          sf_api.result_models.CheckProposedNodeAdditions)
        assert mock_call_method.called

    def test_check_proposed_node_additions_raise_non_fipsdrives(self, mock_call_method):
        mock_response = result_models.CheckProposedNodeAdditions.get_mock_response(
            include_optional=True,
            overrides={'proposed_cluster_errors': [{
                'errorCode': 'nonFipsDrivesCapable',
                'description': 'Unable to add a node with non-FIPS capable drive(s)',
                'nodeIPs': [self.node_dm.mip]
            }]},
            version=models.SfVersion('11.0'))
        mock_call_method.return_value = mock_response
        with pytest.raises(functional_models._errors.FunctionalClusterError):
            self.cluster_fm.check_proposed_node_additions([self.node_dm], raise_on_failure=True)
        assert mock_call_method.called

    def test_check_proposed_node_additions_doesnt_raise_non_fipsdrives(self, mock_call_method):
        mock_response = result_models.CheckProposedNodeAdditions.get_mock_response(
            include_optional=True,
            overrides={'proposed_cluster_errors': [{
                'errorCode': 'nonFipsDrivesCapable',
                'description': 'Unable to add a node with non-FIPS capable drive(s)',
                'nodeIPs': [self.node_dm.mip]
            }]},
            version=models.SfVersion('11.0'))
        mock_call_method.return_value = mock_response
        check_proposed_node_additions_dm = self.cluster_fm.check_proposed_node_additions(
            [self.node_dm],
            raise_on_failure=False
        )
        assert isinstance(check_proposed_node_additions_dm,
                          sf_api.result_models.CheckProposedNodeAdditions)
        assert mock_call_method.called
