"""This file tests all the methods defined in valence/functional_models/_cluster.cluster.py."""
import datetime
import unittest

import mock
import pytest

import sf_api
from sf_api import (
    api_constants,
    result_models,
    sf_version,
)
from sf_api.models import errors as model_errors
from valence import (
    enums,
    errors,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterMixin(unittest.TestCase):
    """Misc methods that don't need a separate TestCase class with specialized setUp and mocks."""

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_api(self, mock_call_method):
        self.cluster_fm.get_api(show_internal=False)
        mock_call_method.assert_called()

    def test_get_current_version(self, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        response.result = {sf_api.api_constants.CURRENT_VERSION: '8.1'}
        mock_call_method.return_value = response
        current_version = self.cluster_fm.get_current_version()
        assert current_version == sf_api.SfVersion('8.1')

    def test_complete_cluster_upgrade(self, mock_call_method):
        response = result_models.GetClusterVersionInfo.get_mock_response()
        response.result = {sf_api.api_constants.CURRENT_VERSION: '11.1'}
        mock_call_method.return_value = response
        self.cluster_fm.upgrading = True
        self.cluster_fm.complete_cluster_upgrade()
        assert not self.cluster_fm.upgrading
        assert self.cluster_fm.cluster_dm.api_endpoint == models.SfVersion('11.1')

    def test_get_cluster_version_info(self, mock_call_method):
        response = result_models.GetClusterVersionInfo.get_mock_response()
        mock_call_method.return_value = response
        cluster_version_info = self.cluster_fm.get_cluster_version_info()
        assert isinstance(cluster_version_info, models.ClusterVersionInfo)

    def test_api_supported(self, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        response.result = {sf_api.api_constants.CURRENT_VERSION: current_version,
                           current_version: [enums.ClusterMethod.create_volume.value]}
        mock_call_method.return_value = response
        assert self.cluster_fm.api_supported(enums.ClusterMethod.create_volume)
        assert not self.cluster_fm.api_supported(enums.ClusterMethod.delete_volume)

    def test_api_supported_version_mismatch(self, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        current_version = '1' + current_version
        response.result = {sf_api.api_constants.CURRENT_VERSION: current_version,
                           current_version: [enums.ClusterMethod.create_volume.value]}
        mock_call_method.return_value = response
        with pytest.raises(functional_models._errors.FunctionalClusterError):
            self.cluster_fm.api_supported(enums.ClusterMethod.create_volume)

    def test_api_supported_missing_api_version(self, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        response.result = {sf_api.api_constants.CURRENT_VERSION: current_version}
        mock_call_method.return_value = response

        with pytest.raises(functional_models._errors.FunctionalClusterError):
            assert self.cluster_fm.api_supported(enums.ClusterMethod.create_volume)

    def test_enable_encryption_at_rest(self, mock_call_method):
        mock_key_provider = models.KeyProviderKmip.get_mock_object()
        async_handler = self.cluster_fm.enable_encryption_at_rest(key_provider_dm=mock_key_provider)
        assert isinstance(async_handler, functional_models.EnableEarAsyncHandler)

    def test_disable_encryption_at_rest(self, mock_call_method):
        async_handler = self.cluster_fm.disable_encryption_at_rest()
        assert isinstance(async_handler, functional_models.DisableEarAsyncHandler)

    def test_get_cluster_full_threshold(self, mock_call_method):
        mock_response = result_models.GetClusterFullThreshold.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_cluster_full_threshold()
        assert isinstance(result, sf_api.result_models.GetClusterFullThreshold)

    def test_modify_cluster_full_threshold(self, mock_call_method):
        mock_response = result_models.ModifyClusterFullThreshold.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.modify_cluster_full_threshold()
        assert isinstance(result, sf_api.result_models.ModifyClusterFullThreshold)

    def test_get_cluster_structure(self, mock_call_method):
        mock_response = result_models.GetClusterStructure.get_mock_response()
        mock_call_method.return_value = mock_response
        structure = self.cluster_fm.get_cluster_structure()
        assert isinstance(structure, sf_api.result_models.GetClusterStructure)

    def test_set_cluster_structure(self, mock_call_method):
        mock_response = result_models.SetClusterStructure.get_mock_response()
        mock_call_method.return_value = mock_response
        structure = result_models.GetClusterStructure.get_mock_object()
        async_handle = self.cluster_fm.set_cluster_structure(structure)
        assert isinstance(async_handle, functional_models.SetClusterStructureAsyncHandler)

    def test_get_active_tls_ciphers(self, mock_call_method):
        mock_response = result_models.GetActiveTlsCiphers.get_mock_response()
        mock_call_method.return_value = mock_response
        ciphers = self.cluster_fm.get_active_tls_ciphers()
        assert isinstance(ciphers, sf_api.result_models.GetActiveTlsCiphers)

    def test_get_supported_tls_ciphers(self, mock_call_method):
        mock_response = result_models.GetSupportedTlsCiphers.get_mock_response()
        mock_call_method.return_value = mock_response
        ciphers = self.cluster_fm.get_supported_tls_ciphers()
        assert isinstance(ciphers, sf_api.result_models.GetSupportedTlsCiphers)

    def test_set_supplemental_tls_ciphers(self, mock_call_method):
        supp_ciphers = ['TEST_CIPHERS']
        mock_response = result_models.SetSupplementalTlsCiphers.get_mock_response(
            overrides={'supplemental_ciphers': supp_ciphers})
        mock_call_method.return_value = mock_response
        ciphers = self.cluster_fm.set_supplemental_tls_ciphers(supp_ciphers)
        assert isinstance(ciphers, sf_api.result_models.SetSupplementalTlsCiphers)
        assert supp_ciphers == ciphers.supplemental_ciphers

    def test_reset_supplemental_tls_ciphers(self, mock_call_method):
        mock_response = result_models.ResetSupplementalTlsCiphers.get_mock_response()
        mock_call_method.return_value = mock_response
        ciphers = self.cluster_fm.reset_supplemental_tls_ciphers()
        assert isinstance(ciphers, sf_api.result_models.ResetSupplementalTlsCiphers)

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_get_cluster_info(self, mock_call_method, mock_authenticate):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()
        mock_response = result_models.GetClusterInfo.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_cluster_info()
        assert isinstance(result, models.Cluster)
        assert isinstance(result.resource_dm, models.ResourceCluster)

    def test_clear_cluster_fault(self, mock_call_method):
        self.cluster_fm.clear_cluster_faults()

    def test_get_cluster_capacity(self, mock_call_method):
        mock_response = result_models.GetClusterCapacity.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_cluster_capacity()
        assert isinstance(result, models.ClusterCapacity)

    def test_get_constants(self, mock_call_method):
        response = result_models.GetConstants.get_mock_response()
        mock_call_method.return_value = response
        cluster_fm = functional_models.Cluster.get_mock_fm()
        result = cluster_fm.get_constants()
        get_param_dict = {'constants':
                          list(models.ClusterConstantsVerbose.create().serialize().keys())}
        mock_call_method.assert_called_with(enums.ClusterMethod.get_constants,
                                            params=get_param_dict,
                                            result_dm_class=result_models.GetConstants,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.ClusterConstants)

        result = cluster_fm.get_constants(changed_only=True)
        get_param_dict['changedOnly'] = True
        mock_call_method.assert_called_with(enums.ClusterMethod.get_constants,
                                            params=get_param_dict,
                                            result_dm_class=result_models.GetConstants,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.ClusterConstants)

    def test_get_constants_filter(self, mock_call_method):
        response = result_models.GetConstants.get_mock_response()
        mock_call_method.return_value = response
        cluster_fm = functional_models.Cluster.get_mock_fm()
        result = cluster_fm.get_constants(constants_dm=response.result_dm)
        get_param_dict = {'constants': list(response.result_dm.serialize().keys())}
        mock_call_method.assert_called_with(enums.ClusterMethod.get_constants,
                                            params=get_param_dict,
                                            result_dm_class=result_models.GetConstants,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.ClusterConstants)

    def test_set_constants(self, mock_call_method):
        response = result_models.SetConstants.get_mock_response()
        mock_call_method.return_value = response
        result = self.cluster_fm.set_constants(response.result_dm, force=True)
        params_in = response.result_dm.serialize()
        params_in['force'] = True
        assert isinstance(result, models.ClusterConstants)
        mock_call_method.assert_called_with(enums.ClusterMethod.set_constants,
                                            params=params_in,
                                            result_dm_class=result_models.SetConstants,
                                            auth=mock.ANY, log_response=True
                                            )

    def test_reset_constants(self, mock_call_method):
        response = result_models.ResetConstants.get_mock_response()
        mock_call_method.return_value = response

        result = self.cluster_fm.reset_constants(response.result_dm)
        assert isinstance(result, models.ClusterConstants)
        reset_param_dict = {'constants': list(response.result_dm.serialize().keys())}
        mock_call_method.assert_called_with(enums.ClusterMethod.reset_constants,
                                            params=reset_param_dict,
                                            result_dm_class=result_models.ResetConstants)

    def test_reset_constants_bad_input(self, mock_call_method):
        with pytest.raises(ValueError):
            result = self.cluster_fm.reset_constants(None)

    def test_reset_all_constants(self, mock_call_method):
        response = result_models.ResetConstants.get_mock_response()
        mock_call_method.return_value = response

        result = self.cluster_fm.reset_constants(None, reset_all=True)
        assert isinstance(result, models.ClusterConstants)
        reset_param_dict = {'resetAll': True}
        mock_call_method.assert_called_with(enums.ClusterMethod.reset_constants,
                                            params=reset_param_dict,
                                            result_dm_class=result_models.ResetConstants)

    def test_get_limits(self, mock_call_method):
        response = result_models.GetLimits.get_mock_response()
        mock_call_method.return_value = response
        result = self.cluster_fm.get_limits()
        assert isinstance(result, result_models.GetLimits)

    def test_get_immutable_values(self, mock_call_method):
        get_api_response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        get_api_response.result = {
            sf_api.api_constants.CURRENT_VERSION: current_version,
            current_version: [enums.ClusterMethod.get_immutable_values.value]
        }
        response = result_models.GetImmutableValues.get_mock_response()
        mock_call_method.side_effect = [get_api_response, response]
        result = self.cluster_fm.get_immutable_values()
        assert isinstance(result, result_models.GetImmutableValues)

    def test_get_immutable_values_not_implemented(self, mock_call_method):
        get_api_response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        get_api_response.result = {
            sf_api.api_constants.CURRENT_VERSION: current_version,
            current_version: []
        }
        mock_call_method.return_value = get_api_response
        result = self.cluster_fm.get_immutable_values()
        assert isinstance(result, result_models.GetImmutableValues)

    def test_get_service_status(self, mock_call_method):
        service_status = result_models.GetServiceStatus.get_mock_response()
        mock_call_method.return_value = service_status
        result = self.cluster_fm.get_service_status()
        assert isinstance(result, sf_api.result_models.GetServiceStatus)

    def test_add_cluster_admin(self, mock_call_method):
        mock_response = result_models.AddClusterAdmin.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.add_cluster_admin('tswift', 'red')
        assert isinstance(result, models.ResourceCluster)
        assert result.username == 'tswift'
        assert result.password == 'red'

    def test_list_cluster_admins(self, mock_call_method):
        mock_response = result_models.ListClusterAdmins.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.list_cluster_admins(show_hidden=True)
        assert isinstance(result[0], models.ClusterAdmin)
        mock_call_method.assert_called_with(enums.ClusterMethod.list_cluster_admins,
                                            params={api_constants.SHOW_HIDDEN: True},
                                            result_dm_class=result_models.ListClusterAdmins,
                                            auth=mock.ANY, log_response=True)

    def test_create_event(self, mock_call_method):
        mock_response = result_models.CreateEvent.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.create_event(enums.ClusterEventType.upgrade_event,
                                              'this is a message')
        assert isinstance(result, models.ClusterEvent)

    def test_start_upgrade(self, mock_call_method):
        result = self.cluster_fm.start_upgrade('package_name')

        mock_call_method.assert_called_with(enums.ClusterMethod.start_upgrade,
                                            params={api_constants.PACKAGE_NAME: 'package_name'},
                                            auth=mock.ANY, log_response=True,
                                            result_dm_class=None)

    def test_finish_upgrade(self, mock_call_method):
        result = self.cluster_fm.finish_upgrade()

        mock_call_method.assert_called_with(enums.ClusterMethod.finish_upgrade,
                                            params=None,
                                            auth=mock.ANY, log_response=True,
                                            result_dm_class=None)

    def test_remove_cluster_admin(self, mock_call_method):
        mock_admin = models.ClusterAdmin.get_mock_object()

        self.cluster_fm.remove_cluster_admin(mock_admin)
        mock_call_method.assert_called_with(
            enums.ClusterMethod.remove_cluster_admin,
            params={api_constants.CLUSTER_ADMIN_ID: mock_admin.cluster_admin_id},
            result_dm_class=None,
            auth=mock.ANY, log_response=True
        )

    def test_demote_cluster_master(self, mock_call_method):
        mock_response = result_models.DemoteClusterMaster.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.demote_cluster_master()
        assert isinstance(result, sf_api.result_models.DemoteClusterMaster)

    def test_promote_cluster_master(self, mock_call_method):
        mock_response = result_models.PromoteClusterMaster.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.promote_cluster_master(node_id=1)
        assert isinstance(result, sf_api.result_models.PromoteClusterMaster)

    def test_list_events(self, mock_call_method):
        t = datetime.datetime(2017, 1, 15, 12, 30, 3, 555)
        t2 = datetime.datetime(2017, 1, 15, 12, 32)
        t3 = datetime.datetime(2017, 1, 15, 12, 35)
        t4 = datetime.datetime(2017, 1, 15, 12, 37, 3, 555)
        mock_res = result_models.ListEvents.get_mock_response()
        mock_call_method.return_value = mock_res
        # the UT coverage for timstamps sucks
        result = self.cluster_fm.list_events(start_publish_time=t, end_publish_time=t2,
                                             start_report_time=t3, end_report_time=t4)
        result_event = result[0]

        assert isinstance(result, list)
        assert isinstance(result_event, sf_api.models.ClusterEvent)
        assert isinstance(result_event.event_info_type, sf_api.enums.ClusterEventType)

    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_start_gc(self, mock_get_current_version, mock_call_method):
        mock_get_current_version.return_value = sf_version.SfVersion('12.0')
        mock_response = result_models.StartGC.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.start_gc()
        assert isinstance(result, functional_models.GarbageCollectionAsyncHandler)

    def test_get_gc_status(self, mock_call_method):
        mock_response = result_models.GetGCStatus.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_gc_status()
        assert isinstance(result, sf_api.result_models.GetGCStatus)

    def test_abort_gc_exists(self, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        response.result = {sf_api.api_constants.CURRENT_VERSION: current_version,
                           current_version: [enums.ClusterMethod.abort_gc.value]}
        mock_call_method.return_value = response
        self.cluster_fm.abort_gc()

    @mock.patch.object(functional_models.Cluster, 'get_gc_status')
    @mock.patch('time.sleep')
    def test_abort_gc_missing(self, mock_sleep, mock_get_gc_status, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        response.result = {sf_api.api_constants.CURRENT_VERSION: current_version,
                           current_version: []}
        mock_call_method.return_value = response
        mock_get_gc_status.side_effect = [
            result_models.GetGCStatus.get_mock_object(
                overrides={'status': enums.GcStatus.in_progress}),
            result_models.GetGCStatus.get_mock_object(

                overrides={'status': enums.GcStatus.complete})
        ]
        self.cluster_fm.abort_gc()

    @mock.patch.object(functional_models.Cluster, 'get_gc_status')
    @mock.patch('time.time')
    def test_abort_gc_timeout(self, mock_time, mock_get_gc_status, mock_call_method):
        response = result_models.BaseApiResult.get_mock_response()
        current_version = self.cluster_fm.api.version.version_string
        response.result = {sf_api.api_constants.CURRENT_VERSION: current_version,
                           current_version: []}
        mock_call_method.return_value = response
        mock_get_gc_status.return_value = result_models.GetGCStatus.get_mock_object(
            overrides={'status': enums.GcStatus.in_progress})
        mock_time.side_effect = [0, 30 * 60 * 2]
        with pytest.raises(errors.TimeoutError):
            self.cluster_fm.abort_gc()

    def test_get_current_version_8TB(self, mock_call_method):
        with mock.patch.object(functional_models.Cluster, 'get_current_version',
                               return_value=sf_api.SfVersion('8.1')):
            max_vol_size = self.cluster_fm.get_max_volume_size()
            assert max_vol_size == api_constants.DEFAULT_MAX_VOLUME_SIZE

    def test_get_current_version_16TB(self, mock_call_method):
        with mock.patch.object(functional_models.Cluster, 'get_current_version',
                               return_value=sf_api.SfVersion('11.0')):
            max_vol_size = self.cluster_fm.get_max_volume_size()
            assert max_vol_size == api_constants.SODIUM_MAX_VOLUME_SIZE

    def test_start_cluster_pairing(self, mock_call_method):
        mock_response = result_models.StartClusterPairing.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.start_cluster_pairing()
        assert result == mock_response.result_dm.cluster_pairing_key

    def test_complete_cluster_pairing(self, mock_call_method):
        mock_response = result_models.CompleteClusterPairing.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.complete_cluster_pairing('somepairingkey')
        assert result == mock_response.result_dm

    def test_list_cluster_pairs(self, mock_call_method):
        mock_response = result_models.ListClusterPairs.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.list_cluster_pairs()
        assert result == mock_response.result_dm.cluster_pairs

    def test_remove_cluster_pair(self, mock_call_method):
        mock_pair = models.ClusterPair.get_mock_object()
        self.cluster_fm.remove_cluster_pair(mock_pair)
        mock_call_method.assert_called_with(
            enums.ClusterMethod.remove_cluster_pair, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params={api_constants.CLUSTER_PAIR_ID: mock_pair.cluster_pair_id})

    def test_start_cluster_bs_check(self, mock_call_method):
        mock_response = result_models.StartClusterBSCheck.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.start_cluster_bs_check()
        assert isinstance(result, functional_models.AsyncHandler)
        volume_dms = models.Volume.get_mock_object_list(1)
        result = self.cluster_fm.start_cluster_bs_check(volume_dms=volume_dms)
        assert isinstance(result, functional_models.AsyncHandler)
        volume_dms = models.Volume.get_mock_object_list(2)
        result = self.cluster_fm.start_cluster_bs_check(volume_dms=volume_dms)
        assert isinstance(result, functional_models.AsyncHandler)
        slice_ids = [1]
        result = self.cluster_fm.start_cluster_bs_check(slice_ids=slice_ids)
        assert isinstance(result, functional_models.AsyncHandler)
        slice_ids = [1, 2]
        result = self.cluster_fm.start_cluster_bs_check(slice_ids=slice_ids)
        assert isinstance(result, functional_models.AsyncHandler)

    def test_stop_cluster_bs_check(self, mock_call_method):
        mock_response = result_models.StopClusterBSCheck.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.stop_cluster_bs_check()

    def test_add_block_to_bs(self, mock_call_method):
        mock_response = result_models.AddBlockToBS.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.add_block_to_bs()

    @mock.patch('valence.functional_models.Cluster.get_cluster_info')
    def test_remove_block_from_bs(self, mock_get_cluster, mock_call_method):
        cluster_info = models.Cluster.get_mock_object(overrides=dict(rep_count=1))
        mock_get_cluster.return_value = cluster_info

        mock_response = result_models.RemoveBlockFromBS.get_mock_response(
            overrides=dict(success={'11': True}))
        mock_call_method.return_value = mock_response
        self.cluster_fm.remove_block_from_bs()

        cluster_info = models.Cluster.get_mock_object(overrides=dict(rep_count=2))
        mock_get_cluster.return_value = cluster_info

        with pytest.raises(functional_models.FunctionalClusterError):
            self.cluster_fm.remove_block_from_bs()

    def test_check_block(self, mock_call_method):
        mock_response = result_models.CheckBlock.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.check_block()

    def test_set_ssl(self, mock_call_method):
        mock_response = result_models.SetSslCertificate.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.set_ssl_certificate('cert', 'pk')

    def test_get_ssl(self, mock_call_method):
        mock_response = result_models.GetSslCertificate.get_mock_response()
        mock_call_method.return_value = mock_response
        ssl = self.cluster_fm.get_ssl_certificate()
        assert isinstance(ssl, result_models.GetSslCertificate)

    def test_remove_ssl(self, mock_call_method):
        mock_response = result_models.RemoveSslCertificate.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.remove_ssl_certificate()

    def test_get_ear_info(self, mock_call_method):
        mock_response = result_models.GetEncryptionAtRestInfo.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.get_encryption_at_rest_info()
        mock_call_method.assert_called_with(enums.ClusterMethod.get_encryption_at_rest_info,
                                            auth=mock.ANY, log_response=True, params={},
                                            result_dm_class=result_models.GetEncryptionAtRestInfo)

    def test_get_feature_status(self, mock_call_method):
        mock_response = result_models.GetFeatureStatus.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_feature_status()
        assert isinstance(result[0], models.FeatureStatus)
        result = self.cluster_fm.get_feature_status(feature=enums.SfFeature.fips_drives)
        mock_call_method.assert_called_with(enums.ClusterMethod.get_feature_status,
                                            params={api_constants.FEATURE: 'FipsDrives'},
                                            result_dm_class=result_models.GetFeatureStatus,
                                            auth=mock.ANY, log_response=True)

    def test_enable_feature(self, mock_call_method):
        self.cluster_fm.enable_feature(enums.SfFeature.fips_drives)
        params = {api_constants.FEATURE: enums.SfFeature.fips_drives.value}
        mock_call_method.assert_called_with(enums.ClusterMethod.enable_feature, params=params,
                                            result_dm_class=None,
                                            auth=mock.ANY, log_response=True)

    def test_get_fips_report(self, mock_call_method):
        mock_response = result_models.GetFipsReport.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_fips_report()
        assert isinstance(result, sf_api.result_models.GetFipsReport)

    def test_get_cluster_stats(self, mock_call_method):
        mock_response = result_models.GetClusterStats.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_cluster_stats()
        assert isinstance(result, sf_api.result_models.GetClusterStats)

    def test_shutdown(self, mock_call_method):
        mock_response = result_models.Shutdown.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.shutdown([])
        assert isinstance(result, sf_api.result_models.Shutdown)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListClusterFaults(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.fault_fm0 = models.ClusterFault.create(drive_id=0, node_id=1).serialize()
        cls.fault_fm1 = models.ClusterFault.create(drive_id=1, node_id=1).serialize()
        cls.fault_fm2 = models.ClusterFault.create(drive_id=2, node_id=1).serialize()
        cls.fault_fm3 = models.ClusterFault.create(drive_id=3, node_id=2).serialize()
        cls.fault_fm4 = models.ClusterFault.create(drive_id=4, node_id=3).serialize()
        cls.fault_fm5 = models.ClusterFault.create(drive_id=5, node_id=4).serialize()
        cls.mock_faults = [cls.fault_fm0, cls.fault_fm1, cls.fault_fm2, cls.fault_fm3,
                           cls.fault_fm4, cls.fault_fm5]
        cls.mock_response = result_models.ListClusterFaults.get_mock_response(
            overrides=dict(faults=cls.mock_faults))

    def test_list_cluster_faults(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        response = self.cluster_fm.list_cluster_faults(fault_status=sf_api.ClusterFaultStatus.current,
                                                       update=True, include_cluster_exceptions=True)
        assert len(response) == len(self.mock_faults)

        call_params = {
            api_constants.FAULT_TYPES: sf_api.ClusterFaultStatus.current,
            api_constants.BEST_PRACTICES: None,
            api_constants.UPDATE: True,
            api_constants.EXCEPTIONS: True
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.list_cluster_faults, params=call_params,
                                            result_dm_class=result_models.ListClusterFaults,
                                            auth=mock.ANY, log_response=True)

    def test_list_cluster_faults_by_node(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        response = self.cluster_fm.list_cluster_faults(node_ids=[1])
        nodes = [f for f in self.mock_faults if f[api_constants.NODE_ID] == 1]
        assert len(response) == len(nodes)

    def test_list_cluster_faults_by_drive(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        response = self.cluster_fm.list_cluster_faults(drive_ids=[1, 2])
        assert len(response) == 2

    def test_mutually_exclusive(self, mock_call_method):
        with pytest.raises(ValueError):
            self.cluster_fm.list_cluster_faults(include_codes=[sf_api.ClusterFaultCode.cluster_full],
                                                exclude_codes=[sf_api.ClusterFaultCode.drive_failed])


class TestClusterAuthenticate(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_authenticate_in_constructor(self, mock_authenticate_cluster):
        functional_models.Cluster(self.cluster_dm, authenticate=True)
        assert mock_authenticate_cluster.called

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_authenticate(self, mock_authenticate_cluster):
        cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        cluster_fm.authenticate()
        assert mock_authenticate_cluster.called

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_authenticate_with_cluster_dm(self, mock_authenticate_cluster):
        cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        new_cluster_dm = models.ResourceCluster.get_mock_object()
        cluster_fm.authenticate(cluster_dm=new_cluster_dm)
        assert cluster_fm.cluster_dm is new_cluster_dm
        assert cluster_fm.cluster_dm is not self.cluster_dm
        assert mock_authenticate_cluster.called


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterListDrives(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.drive_dm1 = models.Drive.get_mock_object(overrides=dict(node_id=1,
                                                                    drive_id=1,
                                                                    slot=0,
                                                                    status='active',
                                                                    type='block'))
        cls.drive_dm2 = models.Drive.get_mock_object(overrides=dict(node_id=2,
                                                                    drive_id=2,
                                                                    slot=1,
                                                                    status='available',
                                                                    type='volume'))
        cls.mock_response = result_models.ListDrives.get_mock_response(
            overrides=dict(drives=[cls.drive_dm1, cls.drive_dm2]))

    def test_list_drives(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        drives = self.cluster_fm.list_drives()
        assert self.drive_dm1 in drives
        assert self.drive_dm2 in drives

    def test_list_drives_invalid_response(self, mock_call_method):
        bad_response = self.mock_response
        bad_response.result_dm.drives[0].status = 'invalid_status'
        mock_call_method.return_value = self.mock_response
        self.cluster_fm._validate_responses = True
        with pytest.raises(model_errors.DataValidationError):
            self.cluster_fm.list_drives()

    def test_filter_node_id(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        drives = self.cluster_fm.list_drives(node_ids=[1])
        assert len(drives) == 1

    def test_filter_drive_id(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        drives = self.cluster_fm.list_drives(drive_ids=[2])
        assert len(drives) == 1

    def test_filter_slot(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        drives = self.cluster_fm.list_drives(slots=[0])
        assert len(drives) == 1

    def test_filter_status(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        drives = self.cluster_fm.list_drives(statuses=[sf_api.DriveStatus.active])
        assert len(drives) == 1

    def test_filter_drive_type(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        drives = self.cluster_fm.list_drives(drive_types=[sf_api.DriveType.block])
        assert len(drives) == 1


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterListSyncJobs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.rs_cluster_dm = models.ResourceCluster.get_mock_object(
            overrides=dict(api_endpoint=models.SfVersion('9.0')))
        cls.cluster_fm = functional_models.Cluster(cls.rs_cluster_dm, authenticate=False)
        cls.job_dm1 = models.SyncJob.get_mock_object(
            overrides=dict(node_id=1, type=enums.SyncJobType.clone,
                           stage=enums.SyncJobStage.data))
        cls.job_dm2 = models.SyncJob.get_mock_object(
            overrides=dict(slice_id=6, type=enums.SyncJobType.clone,
                           stage=enums.SyncJobStage.data))
        cls.job_dm3 = models.SyncJob.get_mock_object(
            overrides=dict(type=enums.SyncJobType.remote,
                           stage=enums.SyncJobStage.data))
        cls.job_dm4 = models.SyncJob.get_mock_object(
            overrides=dict(type=enums.SyncJobType.clone,
                           stage=enums.SyncJobStage.whole))
        cls.mock_jobs = [cls.job_dm1, cls.job_dm2, cls.job_dm3, cls.job_dm4]
        cls.mock_response = result_models.ListSyncJobs.get_mock_response(
            overrides=dict(sync_jobs=cls.mock_jobs))

    def test_no_filters(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        assert len(self.cluster_fm.list_sync_jobs()) == 4

    def test_filter_node_id(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        jobs = self.cluster_fm.list_sync_jobs(node_ids=[1])
        assert jobs[0] == self.job_dm1

    def test_filter_slice_id(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        jobs = self.cluster_fm.list_sync_jobs(slice_ids=[6])
        assert jobs[0] == self.job_dm2

    def test_filter_sync_job_type(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        jobs = self.cluster_fm.list_sync_jobs(job_type=enums.SyncJobType.remote)
        assert jobs[0] == self.job_dm3

    def test_filter_sync_job_stage(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        jobs = self.cluster_fm.list_sync_jobs(job_stage=enums.SyncJobStage.whole)
        assert jobs[0] == self.job_dm4


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterListServices(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        services = [
            models.ClusterService.create(
                service_id=0, service_type=enums.ServiceType.block, node_id=0, drive_ids=[0]),
            models.ClusterService.create(
                service_id=1, service_type=enums.ServiceType.slice, node_id=0, drive_ids=[1]),
            models.ClusterService.create(
                service_id=2, service_type=enums.ServiceType.block, node_id=1, drive_ids=[0]),
            models.ClusterService.create(
                service_id=3, service_type=enums.ServiceType.slice, node_id=1, drive_ids=[1]),
            models.ClusterService.create(
                service_id=4, service_type=enums.ServiceType.transport, node_id=0, drive_ids=[]),
            models.ClusterService.create(
                service_id=5, service_type=enums.ServiceType.transport, node_id=1, drive_ids=[]),
        ]
        cls.mock_response = result_models.ListServices.get_mock_response(
            overrides={'services': [models.ClusterServiceInfo.create(service=s) for s in services]}
        )

    def test_no_filters(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        result = self.cluster_fm.list_services()
        assert isinstance(result, list)
        assert isinstance(result[0], sf_api.models.ClusterServiceInfo)
        assert isinstance(result[0].service, sf_api.models.ClusterService)
        assert set(s.service.service_id for s in result) == {0, 1, 2, 3, 4, 5}

    def test_filter_service_type(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        result = self.cluster_fm.list_services(service_types=[enums.ServiceType.block])
        assert set(s.service.service_id for s in result) == {0, 2}

    def test_filter_node_id(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        result = self.cluster_fm.list_services(node_dms=[models.ActiveNode.create(node_id=0)])
        assert set(s.service.service_id for s in result) == {0, 1, 4}

    def test_filter_drive_id(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        result = self.cluster_fm.list_services(drive_dms=[models.Drive.create(drive_id=1)])
        assert set(s.service.service_id for s in result) == {1, 3}

    def test_multiple_filters(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        result = self.cluster_fm.list_services(
            service_types=[enums.ServiceType.block],
            node_dms=[models.ActiveNode.create(node_id=0)]
        )
        assert set(s.service.service_id for s in result) == {0}
