"""Tests for node_mixin.py methods on the cluster FM."""
import unittest

import mock
import pytest
import requests
import sf_rtfi

import sf_api
from sf_api import api_constants
from sf_api import models as api_models
from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)
from valence.models import _resource


@mock.patch.object(functional_models.Cluster, 'list_all_nodes')
class TestGetNodeFms(unittest.TestCase):
    """Tests for all the list nodes methods."""
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.rs_node_dm = models.ResourceNode.get_mock_object()

    def test_default(self, list_all_nodes):
        list_all_nodes.return_value = models.ActiveNode.get_mock_object_list(1)
        node_fms = self.cluster_fm.get_node_fms(authenticate=False)
        assert list_all_nodes.called_with(
            statuses=(enums.NodeStatus.active,),
            node_roles=None
        )
        assert all(isinstance(fm, functional_models.Node) for fm in node_fms)

    def test_filters(self, list_all_nodes):
        list_all_nodes.return_value = models.ActiveNode.get_mock_object_list(1)
        node_statuses = [enums.NodeStatus.pending, enums.NodeStatus.pending_active]
        node_roles = [enums.NodeRole.witness]
        node_fms = self.cluster_fm.get_node_fms(
            statuses=node_statuses,
            node_roles=node_roles,
            authenticate=False
        )
        assert list_all_nodes.called_with(
            statuses=node_statuses,
            node_roles=node_roles
        )
        assert all(isinstance(fm, functional_models.Node) for fm in node_fms)


@mock.patch.object(functional_models.Cluster, 'list_all_nodes')
class TestGetResourceNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.rs_node_dm = models.ResourceNode.get_mock_object()
        cls.node_dms = [
            models.ActiveNode.create(),
            models.PendingNode.create(),
            models.PendingActiveNode.create()
        ]
        for node_dm in cls.node_dms:
            node_dm._resource_dm = cls.rs_node_dm

    def test_default(self, list_all_nodes):
        list_all_nodes.return_value = self.node_dms
        rs_node_dms = self.cluster_fm.get_resource_nodes()
        assert list_all_nodes.called_with(
            statuses=(enums.NodeStatus.active,),
            node_roles=None
        )
        assert all(isinstance(dm, models.ResourceNode) for dm in rs_node_dms)

    def test_status_filter(self, list_all_nodes):
        list_all_nodes.return_value = self.node_dms
        node_statuses = [
            enums.NodeStatus.active,
            enums.NodeStatus.pending,
            enums.NodeStatus.pending_active
        ]
        self.cluster_fm.get_resource_nodes(statuses=node_statuses)
        assert list_all_nodes.called_with(
            statuses=node_statuses,
            node_roles=None
        )

        list_all_nodes.called = False
        node_statuses = [
            enums.NodeStatus.pending,
            enums.NodeStatus.pending_active
        ]
        self.cluster_fm.get_resource_nodes(statuses=node_statuses)
        assert list_all_nodes.called_with(
            statuses=node_statuses,
            node_roles=None
        )


@mock.patch.object(sf_api.ClusterApi, 'call_method')
class TestGetNetworkInterfaceStatusInternal(unittest.TestCase):
    """Test the cluster_fm.get_network_interface_status method."""
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_network_interface_status(self, mock_call_method):
        self.cluster_fm.get_network_interface_status(interfaces=[enums.NetworkInterface.eth0])
        mock_call_method.assert_called()


@mock.patch.object(sf_api.ClusterApi, 'call_method')
class TestGetNodeStats(unittest.TestCase):
    """Test the cluster_fm.get_network_interface_status method."""
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_node_stats(self, mock_call_method):
        mock_call_method.return_value = result_models.GetNodeStats.get_mock_response()
        result = self.cluster_fm.get_node_stats(1)

        assert isinstance(result, models.NodeStats)


@mock.patch('valence.functional_models.Cluster._instantiate_nodes')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListAllNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.node_dms = [
            api_models.ActiveNode.create(
                mip='10.117.78.10',
                internal_role=enums.NodeRole.storage
            ),
            api_models.PendingNode.create(
                mip='10.117.78.11',
                internal_role=enums.NodeRole.storage
            ),
            api_models.PendingActiveNode.create(
                mip='10.117.78.12',
                internal_role=enums.NodeRole.witness
            )
        ]
        cls.mock_response = result_models.ListAllNodes.get_mock_response(
            overrides=dict(
                nodes=[cls.node_dms[0]],
                pending_nodes=[cls.node_dms[1]],
                pending_active_nodes=[cls.node_dms[2]]
            )
        )

    def test_default(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_all_nodes()
        assert init_nodes.call_count == 3
        assert init_nodes.call_args_list[0][1] == dict(
            node_dms=[self.node_dms[0]],
            node_class=models.ActiveNode
        )
        assert init_nodes.call_args_list[1][1] == dict(
            node_dms=[self.node_dms[1]],
            node_class=models.PendingNode
        )
        assert init_nodes.call_args_list[2][1] == dict(
            node_dms=[self.node_dms[2]],
            node_class=models.PendingActiveNode
        )

    def test_status_filter(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_all_nodes(statuses=[enums.NodeStatus.active])
        assert init_nodes.call_count == 1
        assert init_nodes.call_args_list[0][1] == dict(
            node_dms=[self.node_dms[0]],
            node_class=models.ActiveNode
        )

        init_nodes.call_count = 0
        init_nodes.call_args_list = []

        self.cluster_fm.list_all_nodes(
            statuses=[enums.NodeStatus.pending, enums.NodeStatus.pending_active]
        )
        init_nodes.assert_called_with(
            node_dms=self.node_dms[2:],
            node_class=models.PendingActiveNode
        )
        assert init_nodes.call_count == 2
        assert init_nodes.call_args_list[0][1] == dict(
            node_dms=[self.node_dms[1]],
            node_class=models.PendingNode
        )
        assert init_nodes.call_args_list[1][1] == dict(
            node_dms=[self.node_dms[2]],
            node_class=models.PendingActiveNode
        )

    def test_node_role_filter(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_all_nodes(node_roles=[enums.NodeRole.storage])
        assert init_nodes.call_count == 3
        assert init_nodes.call_args_list[0][1] == dict(
            node_dms=[self.node_dms[0]],
            node_class=models.ActiveNode
        )
        assert init_nodes.call_args_list[1][1] == dict(
            node_dms=[self.node_dms[1]],
            node_class=models.PendingNode
        )
        assert init_nodes.call_args_list[2][1] == dict(
            node_dms=[],
            node_class=models.PendingActiveNode
        )

        init_nodes.call_count = 0
        init_nodes.call_args_list = []

        self.cluster_fm.list_all_nodes(node_roles=[enums.NodeRole.witness])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[2:],
            node_class=models.PendingActiveNode
        )
        assert init_nodes.call_count == 3
        assert init_nodes.call_args_list[0][1] == dict(
            node_dms=[],
            node_class=models.ActiveNode
        )
        assert init_nodes.call_args_list[1][1] == dict(
            node_dms=[],
            node_class=models.PendingNode
        )
        assert init_nodes.call_args_list[2][1] == dict(
            node_dms=[self.node_dms[2]],
            node_class=models.PendingActiveNode
        )


@mock.patch('valence.functional_models.Cluster._instantiate_nodes')
@mock.patch('valence.models.get_resource_nodes')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListActiveNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.rs_node_dm = models.ResourceNode.get_mock_object()
        cls.node_dms = [
            api_models.ActiveNode.create(mip='10.117.78.10', internal_role=enums.NodeRole.storage),
            api_models.ActiveNode.create(mip='10.117.78.11', internal_role=enums.NodeRole.storage),
            api_models.ActiveNode.create(mip='10.117.78.12', internal_role=enums.NodeRole.witness)
        ]
        cls.mock_response = result_models.ListActiveNodes.get_mock_response(
            overrides=dict(nodes=cls.node_dms)
        )

    def test_default(self, call_method, get_rs_nodes, init_nodes):
        call_method.return_value = self.mock_response
        get_rs_nodes.return_value = self.rs_node_dm
        self.cluster_fm.list_active_nodes()
        init_nodes.assert_called_with(
            node_dms=self.node_dms,
            node_class=models.ActiveNode
        )

    def test_static_id_filter(self, call_method, get_rs_nodes, init_nodes):
        call_method.return_value = self.mock_response
        get_rs_nodes.return_value = self.rs_node_dm
        self.cluster_fm.list_active_nodes(static_ids=[self.node_dms[0].static_id])
        init_nodes.assert_called_with(
            node_dms=[self.node_dms[0]],
            node_class=models.ActiveNode
        )

        self.cluster_fm.list_active_nodes(static_ids=['fake_id'])
        init_nodes.assert_called_with(
            node_dms=[],
            node_class=models.ActiveNode
        )

    def test_node_role_filter(self, call_method, get_rs_nodes, init_nodes):
        call_method.return_value = self.mock_response
        get_rs_nodes.return_value = self.rs_node_dm
        self.cluster_fm.list_active_nodes(node_roles=[enums.NodeRole.storage])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[:2],
            node_class=models.ActiveNode
        )

        self.cluster_fm.list_active_nodes(node_roles=[enums.NodeRole.witness])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[2:],
            node_class=models.ActiveNode
        )

    @mock.patch('autologger.main.ContextualLogger.warn')
    def test_no_resource_dm(self, context_warn, call_method, get_rs_nodes, init_nodes):
        call_method.return_value = self.mock_response
        get_rs_nodes.return_value = None
        self.cluster_fm.list_active_nodes()
        assert context_warn.called


@mock.patch('valence.functional_models.Cluster._instantiate_nodes')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListPendingNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.node_dms = [
            api_models.PendingNode.create(mip='10.117.78.10', internal_role=enums.NodeRole.storage),
            api_models.PendingNode.create(mip='10.117.78.11', internal_role=enums.NodeRole.storage),
            api_models.PendingNode.create(mip='10.117.78.12', internal_role=enums.NodeRole.witness)
        ]
        cls.mock_response = result_models.ListPendingNodes.get_mock_response(
            overrides=dict(pending_nodes=cls.node_dms)
        )

    def test_default(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_pending_nodes()
        init_nodes.assert_called_with(
            node_dms=self.node_dms,
            node_class=models.PendingNode
        )

    def test_static_id_filter(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_pending_nodes(static_ids=[self.node_dms[0].static_id])
        init_nodes.assert_called_with(
            node_dms=[self.node_dms[0]],
            node_class=models.PendingNode
        )

        self.cluster_fm.list_pending_nodes(static_ids=['fake_id'])
        init_nodes.assert_called_with(
            node_dms=[],
            node_class=models.PendingNode
        )

    def test_node_role_filter(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_pending_nodes(node_roles=[enums.NodeRole.storage])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[:2],
            node_class=models.PendingNode
        )

        self.cluster_fm.list_pending_nodes(node_roles=[enums.NodeRole.witness])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[2:],
            node_class=models.PendingNode
        )


@mock.patch('valence.functional_models.Cluster._instantiate_nodes')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListPendingActiveNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.node_dms = [
            api_models.PendingActiveNode.create(
                mip='10.117.78.10',
                internal_role=enums.NodeRole.storage
            ),
            api_models.PendingActiveNode.create(
                mip='10.117.78.11',
                internal_role=enums.NodeRole.storage
            ),
            api_models.PendingActiveNode.create(
                mip='10.117.78.12',
                internal_role=enums.NodeRole.witness
            )
        ]
        cls.mock_response = result_models.ListPendingActiveNodes.get_mock_response(
            overrides=dict(pending_active_nodes=cls.node_dms)
        )

    def test_default(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_pending_active_nodes()
        init_nodes.assert_called_with(
            node_dms=self.node_dms,
            node_class=models.PendingActiveNode
        )

    def test_static_id_filter(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_pending_active_nodes(static_ids=[self.node_dms[0].static_id])
        init_nodes.assert_called_with(
            node_dms=[self.node_dms[0]],
            node_class=models.PendingActiveNode
        )

        self.cluster_fm.list_pending_active_nodes(static_ids=['fake_id'])
        init_nodes.assert_called_with(
            node_dms=[],
            node_class=models.PendingActiveNode
        )

    def test_node_role_filter(self, call_method, init_nodes):
        call_method.return_value = self.mock_response
        self.cluster_fm.list_pending_active_nodes(node_roles=[enums.NodeRole.storage])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[:2],
            node_class=models.PendingActiveNode
        )

        self.cluster_fm.list_pending_active_nodes(node_roles=[enums.NodeRole.witness])
        init_nodes.assert_called_with(
            node_dms=self.node_dms[2:],
            node_class=models.PendingActiveNode
        )


@mock.patch('valence.models.get_resource_node_by_static_id')
class TestInstantiateNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.rs_node_dm = models.ResourceNode.get_mock_object()
        cls.active_node_dm = api_models.ActiveNode.create()
        cls.pending_node_dm = api_models.PendingNode.create()
        cls.pending_active_node_dm = api_models.PendingActiveNode.create()

    def test_default(self, get_rs_node):
        get_rs_node.return_value = self.rs_node_dm
        node_dms = self.cluster_fm._instantiate_nodes(
            node_dms=[self.active_node_dm],
            node_class=models.ActiveNode
        )
        assert len(node_dms) == 1
        assert all(isinstance(dm, models.ActiveNode) for dm in node_dms)
        assert all(dm.resource_dm is self.rs_node_dm for dm in node_dms)

        node_dms = self.cluster_fm._instantiate_nodes(
            node_dms=[self.pending_node_dm, self.pending_node_dm],
            node_class=models.PendingNode
        )
        assert len(node_dms) == 2
        assert all(isinstance(dm, models.PendingNode) for dm in node_dms)
        assert all(dm.resource_dm is self.rs_node_dm for dm in node_dms)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterAddNodes(unittest.TestCase):
    def setUp(self):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()
        pending_node1 = models.PendingActiveNode.get_mock_object(version=models.SfVersion('9.0'))
        pending_node2 = models.PendingActiveNode.get_mock_object(version=models.SfVersion('9.0'))
        pending_node3 = models.PendingActiveNode.get_mock_object(version=models.SfVersion('8.0'))
        pending_node4 = models.PendingActiveNode.get_mock_object(version=models.SfVersion('8.0'))
        self.pending_nodes_with_handle = [pending_node1, pending_node2]
        self.pending_nodes_without_handle = [pending_node3, pending_node4]

    def test_add_nodes_no_handle(self, mock_call_method):
        """Test add_nodes when it does not return asyncHandle."""
        mock_call_method.return_value = mock.MagicMock(wraps=sf_api.SfResponse)
        mock_call_method.return_value.result_dm = mock.MagicMock(
            wraps=result_models.AddNodes)
        mock_call_method.return_value.result_dm.nodes = self.pending_nodes_without_handle
        async_handler = self.cluster_fm.add_nodes(self.pending_nodes_without_handle)

        called_with_nodes = [n.pending_node_id for n in self.pending_nodes_without_handle]
        mock_call_method.assert_called_with(enums.ClusterMethod.add_nodes,
                                            params={api_constants.PENDING_NODES: called_with_nodes,
                                                    api_constants.FORCE_DURING_UPGRADE: False,
                                                    api_constants.FORCE_NON_FIPS: None,
                                                    api_constants.AUTO_INSTALL: True},
                                            result_dm_class=result_models.AddNodes,
                                            auth=mock.ANY, log_response=True)
        for am in async_handler:
            assert am.async_handle_dm is None

    def test_add_nodes_with_handle(self, mock_call_method):
        """Test add_nodes when returns asyncHandle."""
        mock_call_method.return_value = mock.MagicMock(wraps=sf_api.SfResponse)
        mock_call_method.return_value.result_dm = mock.MagicMock(
            wraps=result_models.AddNodes)
        mock_call_method.return_value.result_dm.nodes = self.pending_nodes_with_handle
        async_handler = self.cluster_fm.add_nodes(self.pending_nodes_with_handle)

        called_with_nodes = [n.pending_node_id for n in self.pending_nodes_with_handle]
        mock_call_method.assert_called_with(enums.ClusterMethod.add_nodes,
                                            params={api_constants.PENDING_NODES: called_with_nodes,
                                                    api_constants.FORCE_DURING_UPGRADE: False,
                                                    api_constants.FORCE_NON_FIPS: None,
                                                    api_constants.AUTO_INSTALL: True},
                                            result_dm_class=result_models.AddNodes,
                                            auth=mock.ANY, log_response=True)
        for am in async_handler:
            assert am.async_handle_dm.async_handle in [n.async_handle
                                                       for n in self.pending_nodes_with_handle]


class TestClusterRemoveNodes(unittest.TestCase):
    def setUp(self):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()
        self.node_dms = models.ActiveNode.get_mock_object_list(3)

    @mock.patch('sf_api.ClusterApi.call_method',
                side_effect=unittest_utils.helpers.check_params_serializable)
    def test_remove_nodes(self, mock_call_method):
        node_ids = []
        for node_dm in self.node_dms:
            node_ids.append(node_dm.node_id)
        self.cluster_fm.remove_nodes(self.node_dms)
        mock_call_method.assert_called_with(enums.ClusterMethod.remove_nodes,
                                            params={api_constants.NODES: node_ids,
                                                    api_constants.FORCE_DURING_UPGRADE: False},
                                            result_dm_class=None,
                                            auth=mock.ANY, log_response=True)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_master_node_id')
    @mock.patch('sf_api.ClusterApi.call_method', side_effect=requests.exceptions.ReadTimeout)
    def test_remove_nodes_raise_timeout(self, mock_call_method, mock_cm_id):
        node_ids = []
        node = models.ActiveNode.get_mock_object()
        mock_call_method.side_effect = requests.exceptions.ReadTimeout
        mock_cm_id.return_value = node.node_id

        for node_dm in self.node_dms:
            node_ids.append(node_dm.node_id)

        with pytest.raises(requests.exceptions.ReadTimeout):
            self.cluster_fm.remove_nodes(self.node_dms)
            mock_call_method.assert_called_with(enums.ClusterMethod.remove_nodes,
                                                params={api_constants.NODES: node_ids,
                                                        api_constants.FORCE_DURING_UPGRADE: False},
                                                result_dm_class=None,
                                                auth=mock.ANY, log_response=True)
        assert mock_call_method.called

    @mock.patch.object(functional_models.Cluster, 'get_cluster_master_node_id')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch('sf_api.ClusterApi.call_method', side_effect=requests.exceptions.ReadTimeout)
    def test_remove_nodes_raise_value_error(self, mock_call_method, mock_list_active, mock_cm_id):
        node_ids = []
        mock_call_method.side_effect = requests.exceptions.ReadTimeout
        for node_dm in self.node_dms:
            node_ids.append(node_dm.node_id)

        mock_cm_id.return_value = node_ids[0]
        mock_active_nodes = models.ActiveNode.get_mock_object_list(1)
        mock_active_nodes[0].node_id = node_ids[0]
        mock_list_active.return_value = mock_active_nodes

        with pytest.raises(ValueError):
            self.cluster_fm.remove_nodes(self.node_dms)
            mock_call_method.assert_called_with(enums.ClusterMethod.remove_nodes,
                                                params={api_constants.NODES: node_ids,
                                                        api_constants.FORCE_DURING_UPGRADE: False},
                                                result_dm_class=None, auth=mock.ANY, log_response=True)
        assert mock_call_method.called

    @mock.patch.object(functional_models.Cluster, 'get_cluster_master_node_id')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch('sf_api.ClusterApi.call_method', side_effect=requests.exceptions.ReadTimeout)
    def test_remove_nodes_raise_timeout_pass(self, mock_call_method, mock_list_active, mock_cm_id):
        node_ids = []
        mock_call_method.side_effect = requests.exceptions.ReadTimeout
        mock_list_active.return_value = models.ActiveNode.get_mock_object_list(1)
        for node_dm in self.node_dms:
            node_ids.append(node_dm.node_id)

        mock_cm_id.return_value = node_ids[0]
        self.cluster_fm.remove_nodes(self.node_dms)
        mock_call_method.assert_called_with(enums.ClusterMethod.remove_nodes,
                                            params={api_constants.NODES: node_ids,
                                                    api_constants.FORCE_DURING_UPGRADE: False},
                                            result_dm_class=None,
                                            auth=mock.ANY, log_response=True)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterGetMasterNodeId(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_cluster_master_node_id(self, mock_call_method):
        response = sf_api.result_models.GetClusterMasterNodeID.get_mock_response()
        mock_call_method.return_value = response
        node_id = self.cluster_fm.get_cluster_master_node_id()
        assert isinstance(node_id, int)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_master_node_id')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_get_cluster_master(self, mock_list_active_nodes, mock_cm_id, _):
        node = models.ActiveNode.get_mock_object()
        mock_list_active_nodes.return_value = [node]
        mock_cm_id.return_value = node.node_id
        cm = self.cluster_fm.get_cluster_master()
        assert cm == node
