from __future__ import division

import random
import unittest

import mock

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('valence.functional_models.Cluster.list_drives')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestDrivesMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        active_drive1 = models.Drive.get_mock_object(
            overrides={api_constants.STATUS: sf_api.DriveStatus.active})
        active_drive2 = models.Drive.get_mock_object(
            overrides={api_constants.STATUS: sf_api.DriveStatus.active})
        cls.active_drive_dms = [active_drive1, active_drive2]
        avail_drive1 = models.Drive.get_mock_object(
            overrides={api_constants.STATUS: sf_api.DriveStatus.available})
        avail_drive2 = models.Drive.get_mock_object(
            overrides={api_constants.STATUS: sf_api.DriveStatus.available})
        cls.avail_drive_dms = [avail_drive1, avail_drive2]

    def test_list_drive_stats(self, mock_call_method, _):
        mock_response = result_models.ListDriveStats.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.list_drive_stats()

    def test_remove_drives(self, mock_call_method, _):
        mock_response = result_models.RemoveDrives.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.remove_drives(self.avail_drive_dms)

    def test_resize_drives(self, mock_call_method, _):
        mock_response = result_models.ResizeDrives.get_mock_response(version=models.SfVersion('12.0'))
        mock_call_method.return_value = mock_response
        percent = random.randint(5, 90)
        requested_size = models.DataSize(
            size=(self.avail_drive_dms[0].capacity * percent // 100))
        self.cluster_fm.resize_drives(self.avail_drive_dms, requested_size)

    def test_add_drives_no_handle_active(self, mock_call_method, mock_list_drives):
        """Test add_drives when it does not return asyncHandle and drives are active."""
        mock_response = result_models.AddDrives.get_mock_response(overrides=dict(async_handle=None))
        mock_call_method.return_value = mock_response
        mock_list_drives.return_value = self.active_drive_dms
        async_handler = self.cluster_fm.add_drives(self.active_drive_dms, force_during_bin_sync=False)
        drives = [{api_constants.DRIVE_ID: self.active_drive_dms[0].drive_id,
                   api_constants.TYPE: self.active_drive_dms[0].type.value},
                  {api_constants.DRIVE_ID: self.active_drive_dms[1].drive_id,
                   api_constants.TYPE: self.active_drive_dms[1].type.value}]
        add_drives_params = {
            api_constants.DRIVES: drives,
            api_constants.IGNORE_UNBALANCED_CLUSTER_CHECK: None,
            api_constants.FORCE_DURING_UPGRADE: False,
            api_constants.FORCE_DURING_BIN_SYNC: False,
            api_constants.FORCE_NON_FIPS: None
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.add_drives, params=add_drives_params,
                                            result_dm_class=result_models.AddDrives,
                                            auth=mock.ANY, log_response=True)
        assert async_handler.async_handle_dm is None
        assert async_handler.get_models(blocking=True) == self.active_drive_dms

    def test_add_drives_no_handle_avail(self, mock_call_method, mock_list_drives):
        """Test add_drives when it does not return asyncHandle and drives are available."""
        mock_response = result_models.AddDrives.get_mock_response(overrides=dict(async_handle=None))
        mock_call_method.return_value = mock_response
        mock_list_drives.return_value = self.avail_drive_dms
        async_handler = self.cluster_fm.add_drives(self.avail_drive_dms, force_during_bin_sync=False)
        drives = [{api_constants.DRIVE_ID: self.avail_drive_dms[0].drive_id,
                   api_constants.TYPE: self.avail_drive_dms[0].type.value},
                  {api_constants.DRIVE_ID: self.avail_drive_dms[1].drive_id,
                   api_constants.TYPE: self.avail_drive_dms[1].type.value}]
        params = {
            api_constants.DRIVES: drives,
            api_constants.IGNORE_UNBALANCED_CLUSTER_CHECK: None,
            api_constants.FORCE_DURING_UPGRADE: False,
            api_constants.FORCE_DURING_BIN_SYNC: False,
            api_constants.FORCE_NON_FIPS: None,
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.add_drives, params=params,
                                            result_dm_class=result_models.AddDrives,
                                            auth=mock.ANY, log_response=True)
        assert async_handler.async_handle_dm is None
        assert async_handler.get_models() == self.avail_drive_dms

    @mock.patch('valence.functional_models.Cluster.get_async_result')
    def test_add_drives_with_handle(self, mock_get_async_result, mock_call_method, mock_list_drives):
        """Test add_drives when it does return an asyncHandle."""
        cluster_fm = functional_models.Cluster.get_mock_fm()
        cluster_fm._validate_responses = True
        cluster_fm.cluster_dm.api_endpoint = models.SfVersion('10.0')
        mock_response = result_models.AddDrives.get_mock_response()
        mock_call_method.return_value = mock_response
        async_handler_id = mock_response.result_dm.async_handle
        mock_response = result_models.GetAsyncResult.get_mock_response()
        mock_get_async_result.return_value = mock_response
        mock_list_drives.return_value = self.active_drive_dms
        async_handler = cluster_fm.add_drives(self.active_drive_dms)
        assert async_handler.async_handle_dm.async_handle == async_handler_id
        assert async_handler.get_models() == self.active_drive_dms

    def test_get_drive_hardware_info(self, mock_call_method, mock_list_drives):
        mock_call_method.return_value = result_models.GetDriveHardwareInfo.get_mock_response()
        drive_hardware_info = self.cluster_fm.get_drive_hardware_info(self.active_drive_dms[0])
        assert mock_call_method.called_once()
        assert isinstance(drive_hardware_info, models.DriveHardwareInfo)

    def test_secure_erase_drives(self, mock_call_method, mock_list_drives):
        mock_response = result_models.SecureEraseDrives.get_mock_response()
        mock_call_method.return_value = mock_response
        mock_list_drives.return_value = self.avail_drive_dms
        async_handler = self.cluster_fm.secure_erase_drives(self.avail_drive_dms)
        assert async_handler.get_models(blocking=True) == self.avail_drive_dms
