import unittest

import mock
import pytest

from sf_api import result_models
from valence import (
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterInitiators(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_create_response = result_models.CreateInitiators.get_mock_response()
        self.mock_modify_response = result_models.ModifyInitiator.get_mock_response()
        self.mock_list_response = result_models.ListInitiators.get_mock_response()
        self.mock_modify_initiators_response = result_models.ModifyInitiators.get_mock_response()
        self.mock_add_to_vag_response = result_models.AddInitiatorsToVolumeAccessGroup.get_mock_response(
            version=models.SfVersion('9.0'))
        self.mock_remove_from_vag_response =\
            result_models.RemoveInitiatorsFromVolumeAccessGroup.get_mock_response(
                version=models.SfVersion('9.0'))

    def test_create_initiators(self, mock_call_method):
        initiators = models.Initiator.create(alias='', attributes={},
                                             initiator_name='21:00:00:0e:10:07:56:71',
                                             volume_access_groups=[1],
                                             initiator_id=0)
        mock_call_method.return_value = self.mock_create_response
        initiator_dms = self.cluster_fm.create_initiators([initiators])
        assert isinstance(initiator_dms[0], models.Initiator)

    def test_create_with_too_many_VAGs(self, mock_call_method):
        initiator = models.Initiator.create(alias='', attributes={},
                                            initiator_name='21:00:00:0e:10:07:56:71',
                                            volume_access_groups=[1, 2],
                                            initiator_id=0)
        with pytest.raises(ValueError):
            self.cluster_fm.create_initiators([initiator])

    def test_delete_initiators(self, mock_call_method):
        self.cluster_fm.delete_initiators(self.cluster_fm.list_initiators())

    def test_modify_initiator(self, mock_call_method):
        mock_call_method.return_value = self.mock_list_response
        initiator_dms = self.cluster_fm.list_initiators()
        mock_call_method.return_value = self.mock_modify_response
        initiator_dm = self.cluster_fm.modify_initiator(initiator_dms[0])
        assert isinstance(initiator_dm, models.Initiator)

    def test_modify_initiators(self, mock_call_method):
        mock_call_method.return_value = self.mock_modify_initiators_response
        initiator_dm = models.Initiator.get_mock_object()
        initiator_dm.volume_access_groups = [1]
        initiator_dms = self.cluster_fm.modify_initiators([initiator_dm])
        assert isinstance(initiator_dms[0], models.Initiator)

    def test_modify_initiators_remove_vag(self, mock_call_method):
        mock_call_method.return_value = self.mock_modify_initiators_response
        initiator_dm = models.Initiator.get_mock_object()
        initiator_dm.volume_access_groups = None
        initiator_dms = self.cluster_fm.modify_initiators([initiator_dm])
        assert isinstance(initiator_dms[0], models.Initiator)

    def test_modify_initiators_empty_vags(self, mock_call_method):
        mock_call_method.return_value = self.mock_modify_initiators_response
        initiator_dm = models.Initiator.get_mock_object()
        initiator_dm.volume_access_groups = []
        initiator_dms = self.cluster_fm.modify_initiators([initiator_dm])
        assert isinstance(initiator_dms[0], models.Initiator)

    def test_modify_initiators_too_much_vag(self, mock_call_method):
        mock_call_method.return_value = self.mock_modify_initiators_response
        initiator_dm = models.Initiator.get_mock_object()
        initiator_dm.volume_access_groups = [1, 2]
        with pytest.raises(ValueError):
            self.cluster_fm.modify_initiators([initiator_dm])

    def test_list_initiators(self, mock_call_method):
        mock_call_method.return_value = self.mock_list_response
        initiator_dms = self.cluster_fm.list_initiators()
        assert isinstance(initiator_dms[0], models.Initiator)

    def test_add_initiators_to_volume_access_group(self, mock_call_method):
        mock_call_method.return_value = self.mock_add_to_vag_response
        vag_dm = self.cluster_fm.add_initiators_to_volume_access_group(1, [0])
        assert isinstance(vag_dm, models.VolumeAccessGroup)

    def test_remove_initiators_from_volume_access_group(self, mock_call_method):
        mock_call_method.return_value = self.mock_remove_from_vag_response
        vag_dm = self.cluster_fm.remove_initiators_from_volume_access_group(1, [0, 1])
        assert isinstance(vag_dm, models.VolumeAccessGroup)
