import datetime
import unittest

import mock

from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
    utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterMfaMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.cluster_admin_dm = models.ClusterAdmin.get_mock_object()
        cls.auth_session_dm = models.AuthSessionInfo.get_mock_object()
        cls.idp_config_info = models.IdpConfigInfo.get_mock_object()

    def test_create_idp_cluster_admin(self, mock_call_method):
        mock_call_method.return_value = result_models.AddClusterAdmin.get_mock_response(
            overrides=dict(cluster_admin_id=1))
        username = 'email=foo@bar.com'
        self.cluster_fm.add_idp_cluster_admin(username)

        mock_call_method.assert_called_with(
            enums.ClusterMethod.add_idp_cluster_admin, auth=mock.ANY, log_response=True, params={
                'username': username,
                'acceptEula': True,
                'access':
                    [enums.ClusterAccessLevel.administrator],
                'attributes': None},
            result_dm_class=result_models.AddClusterAdmin
        )

    @mock.patch.object(utils, 'get_now_utc')
    def test_create_idp_auth_session(self, mock_utc, mock_call_method):
        """cluster_api.create_idp_auth_session returns AuthSessionInfo"""

        # Set the test_time to the epoch, so we get a easy value to calculate
        test_time = datetime.datetime(1970, 1, 1)
        mock_utc.return_value = test_time
        saml_attributes = ['email=foo@bar.com']
        username = 'email=foo@bar.com'
        mock_call_method.return_value = result_models.CreateAuthSessionResult.get_mock_response(
            overrides=dict(session=self.auth_session_dm))
        result = self.cluster_fm.create_idp_auth_session(
            saml_attributes=saml_attributes,
            desired_expiration_duration=datetime.timedelta(hours=1),
            username=username
        )

        assert result == self.auth_session_dm
        mock_call_method.assert_called_with(
            enums.ClusterMethod.create_idp_auth_session, params={
                'samlAttributeStatements': saml_attributes,
                'desiredExpirationDuration': '3600',
                'username': username
            },
            result_dm_class=result_models.CreateAuthSessionResult
        )

    def test_create_idp_configuration(self, mock_call_method):
        mock_call_method.return_value = result_models.CreateIdpConfigurationResult.get_mock_response(
            overrides=dict(idp_config_info=self.idp_config_info)
        )
        url = self.idp_config_info.idp_metadata_url
        result = self.cluster_fm.create_idp_configuration(url)
        assert result == self.idp_config_info

        mock_call_method.assert_called_with(
            enums.ClusterMethod.create_idp_configuration, params={
                'idpMetadataUrl': url,
            },
            result_dm_class=result_models.CreateIdpConfigurationResult
        )

    def test_list_idp_configuration(self, mock_call_method):
        mock_call_method.return_value = result_models.ListIdpConfigurationsResult.get_mock_response(
            overrides=dict(idp_config_infos=[self.idp_config_info])
        )
        result = self.cluster_fm.list_idp_configurations()
        assert result == [self.idp_config_info]

        mock_call_method.assert_called_with(
            enums.ClusterMethod.list_idp_configurations, params={
                'idpMetadataUrl': None
            },
            result_dm_class=result_models.ListIdpConfigurationsResult
        )

    def test_update_idp_configuration(self, mock_call_method):
        mock_call_method.return_value = result_models.UpdateIdpConfigurationResult.get_mock_response(
            overrides=dict(idp_config_info=self.idp_config_info)
        )
        url = self.idp_config_info.idp_metadata_url
        config_id = str(self.idp_config_info.idp_configuration_id)
        result = self.cluster_fm.update_idp_configuration(self.idp_config_info, url,
                                                          refresh_cert_expiration=True)
        assert result == self.idp_config_info

        mock_call_method.assert_called_with(
            enums.ClusterMethod.update_idp_configuration, params={
                'idpConfigurationID': config_id,
                'idpMetadataUrl': url,
                'refreshCertificateExpirationTime': True,
            },
            result_dm_class=result_models.UpdateIdpConfigurationResult
        )

    def test_delete_idp_configuration(self, mock_call_method):
        config_id = str(self.idp_config_info.idp_configuration_id)
        self.cluster_fm.delete_idp_configuration(self.idp_config_info)

        mock_call_method.assert_called_with(
            enums.ClusterMethod.delete_idp_configuration, params={
                'idpConfigurationID': config_id
            },
            result_dm_class=result_models.DeleteIdpConfigurationResult
        )

    def test_enable_idp_authentication(self, mock_call_method):
        config_id = str(self.idp_config_info.idp_configuration_id)
        mock_call_method.return_value = result_models.EnableIdpAuthenticationResult.get_mock_response()
        self.cluster_fm.enable_idp_authentication(idp_config=self.idp_config_info)
        mock_call_method.assert_called_with(
            enums.ClusterMethod.enable_idp_authentication, params={
                'idpConfigurationID': config_id
            },
            result_dm_class=result_models.EnableIdpAuthenticationResult
        )

    def test_enable_idp_authentication_no_id_given(self, mock_call_method):
        mock_call_method.return_value = result_models.EnableIdpAuthenticationResult.get_mock_response()
        self.cluster_fm.enable_idp_authentication()

        mock_call_method.assert_called_with(
            enums.ClusterMethod.enable_idp_authentication, params={
                'idpConfigurationID': None
            },
            result_dm_class=result_models.EnableIdpAuthenticationResult
        )

    def test_disable_idp_authentication(self, mock_call_method):
        self.cluster_fm.disable_idp_authentication()
        mock_call_method.assert_called_with(
            enums.ClusterMethod.disable_idp_authentication, params={},
            result_dm_class=result_models.DisableIdpAuthenticationResult)

    def test_get_idp_authentication_state(self, mock_call_method):
        mock_call_method.return_value = result_models.GetIdpAuthenticationStateResult.get_mock_response()
        idp_auth_enabled = self.cluster_fm.get_idp_authentication_state()
        mock_call_method.assert_called_with(
            enums.ClusterMethod.get_idp_authentication_state, params={},
            result_dm_class=result_models.GetIdpAuthenticationStateResult
        )
        assert isinstance(idp_auth_enabled, bool)
