"""This file tests all the methods defined in valence/functional_models/_cluster/report_mixin.py."""
import unittest

import mock
import pytest
import sf_artifacts
from sf_artifacts import errors as sf_errors

from sf_api.models import errors
from valence import (
    enums,
    functional_models,
    models,
)


@mock.patch('sf_api.ClusterApi.http_get')
class TestReportMethods(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_report(self, mock_get):
        response = mock.MagicMock()
        response.text = '{"fake_json": true}'
        mock_get.return_value = response
        result = self.cluster_fm.get_report(enums.JsonReport.bins)
        assert result['fake_json'] is True

    def test_get_report_invalid_json(self, mock_get):
        response = mock.MagicMock()
        response.text = '{fake_json": true}'
        mock_get.return_value = response
        with pytest.raises(ValueError):
            self.cluster_fm.get_report(enums.JsonReport.bins)

    def test_get_report_invalid_json_hack(self, mock_get):
        """This tests that we properly work around the invalid JSON that some reports return."""
        response = mock.MagicMock()
        response.text = '{"fake_json": inf, "another_key": -inf}'
        mock_get.return_value = response
        result = self.cluster_fm.get_report(enums.JsonReport.bins)
        assert result['fake_json'] == float('inf')
        assert result['another_key'] == float('-inf')

    def test_slice_balance_report(self, mock_get):
        """This tests the get_slice_balance_report."""
        response = mock.MagicMock()
        response.text = '{"fake_json": true}'
        mock_get.return_value = response
        report = self.cluster_fm.get_slice_balance_report()
        assert report

    def test_get_raw_report(self, mock_get):
        """This tests the get_raw_report."""
        response = mock.MagicMock()
        response.text = 'string'
        mock_get.return_value = response
        report = self.cluster_fm.get_report_raw(enums.HtmlReport.ss)
        assert report


@mock.patch('valence.models.BinAssignments.validate')
@mock.patch('valence.functional_models.Cluster.get_report')
class TestClusterGetBinAssignments(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        # Define a set of mock services
        active_service = models.BinService.create(service_id=1, status=enums.BinStatus.active)
        updating_service1 = models.BinService.create(service_id=2, status=enums.BinStatus.updating)
        updating_service2 = models.BinService.create(service_id=3, status=enums.BinStatus.updating)
        prd_service = models.BinService.create(
            service_id=4,
            status=enums.BinStatus.pending_removal_dead
        )
        # Define a set of mock bin reports
        bin_report1 = models.BinReport.create(bin_id=1, services=[active_service])
        bin_report2 = models.BinReport.create(bin_id=2, services=[updating_service1])
        bin_report3 = models.BinReport.create(
            bin_id=3,
            services=[updating_service1, prd_service]
        )
        bin_report4 = models.BinReport.create(
            bin_id=4,
            services=[updating_service2, prd_service]
        )
        cls.mock_response = [bin_report1, bin_report2, bin_report3, bin_report4]

    def test_old_format(self, mock_get_report, mock_validate):
        mock_get_report.return_value = [models.BinReport.get_mock_data() for _ in range(3)]
        self.cluster_fm.get_bin_assignments(validate=True)

    def test_new_format(self, mock_get_report, mock_validate):
        mock_get_report.return_value = models.BinAssignments.get_mock_data()
        self.cluster_fm.get_bin_assignments()

    def test_filter_include_statuses(self, mock_get_report, mock_validate):
        mock_get_report.return_value = self.mock_response
        result = self.cluster_fm.get_bin_assignments(
            include_statuses=[enums.BinStatus.active]
        )
        assert set(b.bin_id for b in result.bins) == {1, }

        result = self.cluster_fm.get_bin_assignments(
            include_statuses=[enums.BinStatus.pending_removal_dead]
        )
        assert set(b.bin_id for b in result.bins) == {3, 4}

    def test_filter_exclude_statuses(self, mock_get_report, mock_validate):
        mock_get_report.return_value = self.mock_response
        result = self.cluster_fm.get_bin_assignments(
            exclude_statuses=[enums.BinStatus.active]
        )
        assert set(b.bin_id for b in result.bins) == {2, 3, 4}

    def test_filter_services(self, mock_get_report, mock_validate):
        mock_get_report.return_value = self.mock_response
        service_info = models.ClusterServiceInfo.create(
            service=models.ClusterService.create(service_id=3)
        )
        result = self.cluster_fm.get_bin_assignments(
            include_statuses=[enums.BinStatus.updating],
            services=[service_info]
        )
        assert set(b.bin_id for b in result.bins) == {4, }

    @mock.patch('sf_artifacts.get_global_manager')
    @mock.patch.object(sf_artifacts.ArtifactManager, 'store_text')
    def test_validate_failure(self, mock_artifact_manager, mock_global_manager,
                              mock_get_report, mock_validate):
        mock_validate.side_effect = errors.DataValidationError({'Validation error': 'Mock Test'})
        mock_get_report.return_value = models.BinAssignments.get_mock_data()
        with pytest.raises(errors.DataValidationError):
            self.cluster_fm.get_bin_assignments(validate=True)

    @mock.patch('sf_artifacts.get_global_manager')
    @mock.patch.object(sf_artifacts.ArtifactManager, 'store_text')
    def test_validate_failure_artifact_failure(self, mock_artifact_manager, mock_global_manager,
                                               mock_get_report, mock_validate):
        mock_validate.side_effect = errors.DataValidationError({'Validation error': 'Mock Test'})
        mock_global_manager.side_effect = sf_errors.ArtifactsError('Artifact error')
        mock_get_report.return_value = models.BinAssignments.get_mock_data()
        with pytest.raises(errors.DataValidationError):
            self.cluster_fm.get_bin_assignments(validate=True)


class TestBinSliceReports(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    @mock.patch('valence.functional_models.Cluster.get_bin_assignments')
    def test_bin(self, mock_get_bin_assignments):
        bins = [models.BinReport.get_mock_object()]
        assignments = models.BinAssignments.get_mock_object(overrides=dict(bins=bins))
        mock_get_bin_assignments.return_value = assignments
        result = self.cluster_fm.get_bin_sync_reports()
        assert isinstance(result[0], models.BinReport)

    @mock.patch('valence.functional_models.Cluster.get_report')
    def test_slice(self, mock_report):
        mock_report.return_value = {}
        self.cluster_fm.get_slice_sync_report()


@mock.patch.object(functional_models.Cluster, 'get_slice_balance_report')
@mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
@mock.patch.object(functional_models.Cluster, 'is_slice_syncing')
class TestSlicesAreBalanced(unittest.TestCase):
    def setUp(self):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_get_slice_sync_report(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        self.cluster_fm.slices_are_balanced
        assert mock_sync_report.called

    def test_get_slice_balance_report(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        self.cluster_fm.slices_are_balanced
        assert mock_balance_report.called

    def test_is_slice_syncing_called(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        self.cluster_fm.slices_are_balanced
        assert mock_slice_syncing.called

    def test_is_slice_syncing_true(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        mock_slice_syncing.return_value = True
        return_val = self.cluster_fm.slices_are_balanced
        self.assertTrue(return_val)

    def test_is_slice_syncing_false(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        mock_slice_syncing.return_value = False
        return_val = self.cluster_fm.slices_are_balanced
        self.assertFalse(return_val)

    def test_number_decisions_true(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        mock_slice_syncing.return_value = False
        mock_balance_instance = mock_balance_report.return_value
        mock_balance_instance.decisions = [1, 2, 3]
        return_val = self.cluster_fm.slices_are_balanced
        self.assertTrue(return_val)

    def test_number_decisions_false(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        mock_slice_syncing.return_value = False
        mock_balance_instance = mock_balance_report.return_value
        mock_balance_instance.decisions = []
        return_val = self.cluster_fm.slices_are_balanced
        self.assertFalse(return_val)

    def test_desired_primaries_zero_false(self, mock_slice_syncing, mock_sync_report,
                                          mock_balance_report):
        mock_slice_syncing.return_value = False
        mock_balance_instance = mock_balance_report.return_value
        mock_balance_instance.decisions = []
        slice_1 = models.Slice.create(slice_id=1, primary=0,
                                      desired_primary=0, desired_secondaries=[])
        slice_2 = models.Slice.create(slice_id=2, primary=0,
                                      desired_primary=0, desired_secondaries=[])
        slice_3 = models.Slice.create(slice_id=3, primary=0,
                                      desired_primary=0, desired_secondaries=[])
        p_slices = mock.PropertyMock(return_value=[slice_1, slice_2, slice_3])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertFalse(return_val)

    def test_desired_primaries_zero_true(self, mock_slice_syncing, mock_sync_report,
                                         mock_balance_report):
        mock_slice_syncing.return_value = False
        mock_balance_instance = mock_balance_report.return_value
        mock_balance_instance.decisions = [1, 2, 3]
        slice_1 = models.Slice.create(slice_id=1, primary=0,
                                      desired_primary=0, desired_secondaries=[])
        slice_2 = models.Slice.create(slice_id=2, primary=0,
                                      desired_primary=0, desired_secondaries=[])
        slice_3 = models.Slice.create(slice_id=3, primary=0,
                                      desired_primary=0, desired_secondaries=[])
        p_slices = mock.PropertyMock(return_value=[slice_1, slice_2, slice_3])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertTrue(return_val)

    def test_desired_primaries_not_zero_match(self, mock_slice_syncing, mock_sync_report,
                                              mock_balance_report):
        mock_slice_syncing.return_value = False
        slice_1 = models.Slice.create(slice_id=1, primary=1,
                                      desired_primary=1, desired_secondaries=[])
        slice_2 = models.Slice.create(slice_id=2, primary=1,
                                      desired_primary=1, desired_secondaries=[])
        slice_3 = models.Slice.create(slice_id=3, primary=1,
                                      desired_primary=1, desired_secondaries=[])
        p_slices = mock.PropertyMock(return_value=[slice_1, slice_2, slice_3])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertFalse(return_val)

    def test_desired_primaries_not_zero_no_match(self, mock_slice_syncing, mock_sync_report,
                                                 mock_balance_report):
        mock_slice_syncing.return_value = False
        slice_1 = models.Slice.create(slice_id=1, primary=1,
                                      desired_primary=2, desired_secondaries=[])
        slice_2 = models.Slice.create(slice_id=2, primary=1,
                                      desired_primary=2, desired_secondaries=[])
        slice_3 = models.Slice.create(slice_id=3, primary=1,
                                      desired_primary=2, desired_secondaries=[])
        p_slices = mock.PropertyMock(return_value=[slice_1, slice_2, slice_3])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertTrue(return_val)

    def test_desired_secndaries_not_zero_match(self, mock_slice_syncing, mock_sync_report,
                                               mock_balance_report):
        mock_slice_syncing.return_value = False
        slice_1 = models.Slice.create(slice_id=1, primary=1,
                                      desired_primary=0, desired_secondaries=[],
                                      live_secondaries=[])
        slice_2 = models.Slice.create(slice_id=2, primary=1,
                                      desired_primary=0, desired_secondaries=[],
                                      live_secondaries=[])
        slice_3 = models.Slice.create(slice_id=3, primary=1,
                                      desired_primary=0, desired_secondaries=[1],
                                      live_secondaries=[1])
        p_slices = mock.PropertyMock(return_value=[slice_1, slice_2, slice_3])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertFalse(return_val)

    def test_desired_secndaries_not_zero_no_match(self, mock_slice_syncing, mock_sync_report,
                                                  mock_balance_report):
        mock_slice_syncing.return_value = False
        slice_1 = models.Slice.create(slice_id=1, primary=1,
                                      desired_primary=0, desired_secondaries=[],
                                      live_secondaries=[])
        slice_2 = models.Slice.create(slice_id=2, primary=1,
                                      desired_primary=0, desired_secondaries=[],
                                      live_secondaries=[])
        slice_3 = models.Slice.create(slice_id=3, primary=1,
                                      desired_primary=0, desired_secondaries=[1],
                                      live_secondaries=[2])
        p_slices = mock.PropertyMock(return_value=[slice_1, slice_2, slice_3])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertTrue(return_val)

    def test_no_slices(self, mock_slice_syncing, mock_sync_report, mock_balance_report):
        mock_slice_syncing.return_value = False
        p_slices = mock.PropertyMock(return_value=[])
        type(mock_sync_report.return_value).slices = p_slices
        return_val = self.cluster_fm.slices_are_balanced
        self.assertTrue(return_val)
