import unittest

import mock

from sf_api import (
    api_constants,
    models,
    result_models,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)
from valence.utils import random_string


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestSecurityMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_key_provider = models.KeyProviderKmip.get_mock_object()
        self.mock_key_server = models.KeyServerKmip.get_mock_object()

    def test_add_key_server_to_provider_kmip(self, mock_call_method):
        self.cluster_fm.add_key_server_to_provider_kmip(self.mock_key_server, self.mock_key_provider)

        params = {
            api_constants.KEY_PROVIDER_ID: self.mock_key_provider.id,
            api_constants.KEY_SERVER_ID: self.mock_key_server.id
        }
        mock_call_method.assert_called_with(
            enums.ClusterMethod.add_key_server_to_provider_kmip, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params=params)

    def test_create_key_provider_kmip(self, mock_call_method):
        mock_response = result_models.CreateKeyProviderKmip.get_mock_response(
            overrides=dict(kmip_key_provider=self.mock_key_provider))
        mock_call_method.return_value = mock_response

        name = self.mock_key_provider.name
        result = self.cluster_fm.create_key_provider_kmip(name)
        assert result == mock_response.result_dm.kmip_key_provider

    def test_create_key_server_kmip(self, mock_call_method):
        mock_response = result_models.CreateKeyServerKmip.get_mock_response(
            overrides=dict(kmip_key_server=self.mock_key_server))
        mock_call_method.return_value = mock_response

        name = self.mock_key_server.name
        hostnames = self.mock_key_server.kmip_key_server_hostnames
        client_cert = self.mock_key_server.kmip_client_certificate
        ca_cert = self.mock_key_server.kmip_ca_certificate
        result = self.cluster_fm.create_key_server_kmip(name, hostnames, client_cert, ca_cert)
        assert result == mock_response.result_dm.kmip_key_server

    def test_create_public_private_key_pair(self, mock_call_method):
        mock_subject = models.CertificateSubject.get_mock_object()
        self.cluster_fm.create_public_private_key_pair(mock_subject)
        params = mock_subject.serialize()
        mock_call_method.assert_called_with(
            enums.ClusterMethod.create_public_private_key_pair, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params=params)

    def test_delete_key_provider_kmip(self, mock_call_method):
        self.cluster_fm.delete_key_provider_kmip(self.mock_key_provider)
        params = {api_constants.KEY_PROVIDER_ID: self.mock_key_provider.id}
        mock_call_method.assert_called_with(
            enums.ClusterMethod.delete_key_provider_kmip, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params=params)

    def test_delete_key_server_kmip(self, mock_call_method):
        self.cluster_fm.delete_key_server_kmip(self.mock_key_server)
        params = {api_constants.KEY_SERVER_ID: self.mock_key_server.id}
        mock_call_method.assert_called_with(
            enums.ClusterMethod.delete_key_server_kmip, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params=params)

    def test_get_client_certificate_sign_request(self, mock_call_method):
        mock_response = result_models.GetClientCertificateSignRequest.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_client_certificate_sign_request()
        assert result == mock_response.result_dm

    def test_get_key_provider_kmip(self, mock_call_method):
        mock_response = result_models.GetKeyProviderKmip.get_mock_response(
            overrides=dict(kmip_key_provider=self.mock_key_provider))
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_key_provider_kmip(self.mock_key_provider.id)
        assert result == mock_response.result_dm.kmip_key_provider

    def test_get_key_server_kmip(self, mock_call_method):
        mock_response = result_models.GetKeyServerKmip.get_mock_response(
            overrides=dict(kmip_key_server=self.mock_key_server))
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.get_key_server_kmip(self.mock_key_server.id)
        assert result == mock_response.result_dm.kmip_key_server

    def test_list_key_providers_kmip(self, mock_call_method):
        key_provider_1 = models.KeyProviderKmip.get_mock_object(overrides=dict(key_provider_id=1))
        key_provider_2 = models.KeyProviderKmip.get_mock_object(overrides=dict(key_provider_id=2))

        mock_response = result_models.ListKeyProvidersKmip.get_mock_response(
            overrides=dict(kmip_key_providers=[key_provider_1, key_provider_2]))
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.list_key_providers_kmip()
        for provider in [key_provider_1, key_provider_2]:
            assert provider in result

    def test_list_key_servers_kmip(self, mock_call_method):
        key_server_1 = models.KeyServerKmip.get_mock_object(
            overrides=dict(key_server_id=1, key_provider_id=1))
        key_server_2 = models.KeyServerKmip.get_mock_object(
            overrides=dict(key_server_id=2, key_provider_id=1))

        mock_response = result_models.ListKeyServersKmip.get_mock_response(
            overrides=dict(kmip_key_servers=[key_server_1, key_server_2]))
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.list_key_servers_kmip(1)
        for server in [key_server_1, key_server_2]:
            assert server in result

    def test_modify_key_server_kmip(self, mock_call_method):
        new_mock_key_server = self.mock_key_server
        new_mock_key_server.name = 'new_name'

        mock_response = result_models.ModifyKeyServerKmip.get_mock_response(
            overrides=dict(kmip_key_server=new_mock_key_server))
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.modify_key_server_kmip(self.mock_key_server,
                                                        key_server_name='new_name')
        assert result == mock_response.result_dm.kmip_key_server

    def test_remove_key_server_from_provider_kmip(self, mock_call_method):
        self.cluster_fm.remove_key_server_from_provider_kmip(self.mock_key_server)
        params = {api_constants.KEY_SERVER_ID: self.mock_key_server.id}
        mock_call_method.assert_called_with(
            enums.ClusterMethod.remove_key_server_from_provider_kmip, auth=mock.ANY, log_response=True,
            result_dm_class=None, params=params)

    def test_test_key_provider(self, mock_call_method):
        self.cluster_fm.test_key_provider(self.mock_key_provider)
        params = {api_constants.KEY_PROVIDER_ID: self.mock_key_provider.id}
        mock_call_method.assert_called_with(
            enums.ClusterMethod.test_key_provider_kmip, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params=params)

    def test_test_key_server(self, mock_call_method):
        self.cluster_fm.test_key_server(self.mock_key_server)
        params = {api_constants.KEY_SERVER_ID: self.mock_key_server.id}
        mock_call_method.assert_called_with(
            enums.ClusterMethod.test_key_server_kmip, auth=mock.ANY, log_response=True,
            result_dm_class=None,
            params=params)
