import unittest

import mock

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterSnapMirror(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_break_snapmirror_volume(self, mock_call_method):
        mock_volume = models.Volume.get_mock_object()
        self.cluster_fm.break_snapmirror_volume(mock_volume,
                                                access=sf_api.enums.SnapMirrorAccessMode.read_only)

        params = {
            api_constants.VOLUME_ID: mock_volume.volume_id,
            api_constants.SNAPSHOT_ID: None,
            api_constants.PRESERVE: None,
            api_constants.ACCESS: sf_api.enums.SnapMirrorAccessMode.read_only
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.break_snapmirror_volume, params=params,
                                            result_dm_class=None,
                                            auth=mock.ANY, log_response=True)

    def test_break_snapmirror_volumes(self, mock_call_method):
        mock_volumes = models.Volume.get_mock_object_list(5)
        self.cluster_fm.break_snapmirror_volumes(mock_volumes,
                                                 access=sf_api.enums.SnapMirrorAccessMode.locked)

        params = {
            api_constants.VOLUME_IDS: [v.volume_id for v in mock_volumes],
            api_constants.ACCESS: sf_api.enums.SnapMirrorAccessMode.locked
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.break_snapmirror_volumes, params=params,
                                            result_dm_class=None,
                                            auth=mock.ANY, log_response=True)

    def test_create_snapmirror_endpoint(self, mock_call_method):
        mock_response = result_models.CreateSnapMirrorEndpoint.get_mock_response()
        mock_call_method.return_value = mock_response

        name = 'this is a name'
        ip_addresses = ['1.1.1.1', '2.2.2.2']

        result = self.cluster_fm.create_snapmirror_endpoint(name, ip_addresses)

        params = {
            api_constants.NAME: name,
            api_constants.IP_ADDRESSES: ip_addresses
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.create_snapmirror_endpoint,
                                            params=params,
                                            result_dm_class=result_models.CreateSnapMirrorEndpoint,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.SnapMirrorEndpoint)

    def test_delete_snapmirror_endpoint(self, mock_call_method):
        mock_endpoint = models.SnapMirrorEndpoint.get_mock_object()
        self.cluster_fm.delete_snapmirror_endpoint(mock_endpoint)

        params = {api_constants.SNAPMIRROR_ENDPOINT_ID: mock_endpoint.snapmirror_endpoint_id}
        mock_call_method.assert_called_with(enums.ClusterMethod.delete_snapmirror_endpoint,
                                            params=params, result_dm_class=None,
                                            auth=mock.ANY, log_response=True)

    def test_list_snapmirror_endpoints(self, mock_call_method):
        mock_response = result_models.ListSnapMirrorEndpoints.get_mock_response()
        mock_call_method.return_value = mock_response

        response = self.cluster_fm.list_snapmirror_endpoints()
        mock_call_method.assert_called_with(enums.ClusterMethod.list_snapmirror_endpoints,
                                            params=None,
                                            result_dm_class=result_models.ListSnapMirrorEndpoints,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(response[0], models.SnapMirrorEndpoint)

    def test_modify_snapmirror_endpoint(self, mock_call_method):
        mock_endpoint = models.SnapMirrorEndpoint.get_mock_object()
        mock_response = result_models.CreateSnapMirrorEndpoint.get_mock_response()
        mock_call_method.return_value = mock_response

        name = 'this is a name'
        ip_addresses = ['1.1.1.1', '2.2.2.2']

        result = self.cluster_fm.modify_snapmirror_endpoint(mock_endpoint, name, ip_addresses)

        params = {
            api_constants.SNAPMIRROR_ENDPOINT_ID: mock_endpoint.snapmirror_endpoint_id,
            api_constants.NAME: name,
            api_constants.IP_ADDRESSES: ip_addresses
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.modify_snapmirror_endpoint,
                                            params=params,
                                            result_dm_class=result_models.ModifySnapMirrorEndpoint,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.SnapMirrorEndpoint)
