import unittest
from datetime import timedelta

import mock
from sf_models import types

from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
    utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestCreateSnapshot(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_volumes = models.Volume.get_mock_object_list(length=3)

    def setUp(self):
        self.mock_response = result_models.CreateSnapshot.get_mock_response()

    def test_create_snapshot(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.create_snapshot(self.mock_volumes[0])

    def test_create_snapshot_from_snapshot(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        mock_snapshot_dm = models.Snapshot.get_mock_object()
        self.cluster_fm.create_snapshot(self.mock_volumes[0], snapshot_dm=mock_snapshot_dm)

    def test_create_snapshot_with_retention(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.create_snapshot(self.mock_volumes[0],
                                        retention=timedelta(hours=5, minutes=12, seconds=54))
        assert '05:12:54' == mock_call_method.call_args[1]['params']['retention']

    def test_long_retention(self, mock_call_method):
        """Timedelta gets weird once it's longer than a day."""
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.create_snapshot(self.mock_volumes[1],
                                        retention=timedelta(days=5, minutes=12, seconds=54))
        assert '120:12:54' == mock_call_method.call_args[1]['params']['retention']


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestCreateGroupSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_volumes = models.Volume.get_mock_object_list(length=3)

    def setUp(self):
        self.mock_response = result_models.CreateGroupSnapshot.get_mock_response()

    def test_create_group_snapshot(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.create_group_snapshot(self.mock_volumes)

    def test_create_group_snapshot_with_retention(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.create_group_snapshot(self.mock_volumes,
                                              retention=timedelta(days=5, minutes=12, seconds=54))


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListGroupSnapshots(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_volumes = models.Volume.get_mock_object_list(length=3)

    def setUp(self):
        self.mock_response = result_models.ListGroupSnapshots.get_mock_response()

    def test_list_group_snapshots(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.list_group_snapshots(volume_dms=self.mock_volumes)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListSnapshots(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_response = result_models.ListSnapshots.get_mock_response()

    def test_list_snapshots(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        self.cluster_fm.list_snapshots()


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestDeleteGroupSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_group_snapshot = models.GroupSnapshot.get_mock_object()

    def test_delete_group_snapshots(self, mock_call_method):
        self.cluster_fm.delete_group_snapshot(self.mock_group_snapshot)
        assert self.mock_group_snapshot.group_snapshot_id == mock_call_method.call_args[1]['params'][
            'groupSnapshotID']


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestDeleteSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_snapshot = models.Snapshot.get_mock_object()

    def test_delete_group_snapshots(self, mock_call_method):
        self.cluster_fm.delete_snapshot(self.mock_snapshot)
        assert self.mock_snapshot.snapshot_id == mock_call_method.call_args[1]['params']['snapshotID']


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestRollbackToGroupSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_group_snapshot = models.GroupSnapshot.get_mock_object()

    def test_rollback_to_group_snapshot(self, mock_call_method):
        self.cluster_fm.rollback_to_group_snapshot(self.mock_group_snapshot)
        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.rollback_to_group_snapshot
        assert mock_call_method.call_args[1]['params'][
                   'groupSnapshotID'] == self.mock_group_snapshot.group_snapshot_id


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestRollbackToSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_snapshot = models.Snapshot.get_mock_object()

    def test_rollback_to_snapshot(self, mock_call_method):
        self.cluster_fm.rollback_to_snapshot(self.mock_snapshot)
        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.rollback_to_snapshot
        assert mock_call_method.call_args[1]['params']['volumeID'] == self.mock_snapshot.volume_id
        assert mock_call_method.call_args[1]['params']['snapshotID'] == self.mock_snapshot.snapshot_id


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestModifySnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_snapshot = models.Snapshot.get_mock_object()

    def test_rollback_to_snapshot(self, mock_call_method):
        now = utils.get_now_utc()
        iso_str = types.DateTimeType().to_primitive(now)

        self.cluster_fm.modify_snapshot(self.mock_snapshot, expiration_time=now,
                                        enable_remote_replication=False)

        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.modify_snapshot
        assert mock_call_method.call_args[1]['params']['snapshotID'] == self.mock_snapshot.snapshot_id
        assert mock_call_method.call_args[1]['params']['enableRemoteReplication'] is False
        assert mock_call_method.call_args[1]['params']['expirationTime'] == iso_str

    def test_rollback_to_snapshot_no_params(self, mock_call_method):

        self.cluster_fm.modify_snapshot(self.mock_snapshot)

        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.modify_snapshot
        assert mock_call_method.call_args[1]['params']['snapshotID'] == self.mock_snapshot.snapshot_id
        assert mock_call_method.call_args[1]['params']['enableRemoteReplication'] is None
        assert mock_call_method.call_args[1]['params']['expirationTime'] is None


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestModifyGroupSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.mock_group_snapshot = models.GroupSnapshot.get_mock_object()

    def test_rollback_to_snapshot(self, mock_call_method):
        now = utils.get_now_utc()
        iso_str = types.DateTimeType().to_primitive(now)

        self.cluster_fm.modify_group_snapshot(self.mock_group_snapshot, expiration_time=now,
                                              enable_remote_replication=False)

        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.modify_group_snapshot
        assert mock_call_method.call_args[1]['params'][
                   'groupSnapshotID'] == self.mock_group_snapshot.group_snapshot_id
        assert mock_call_method.call_args[1]['params']['enableRemoteReplication'] is False
        assert mock_call_method.call_args[1]['params']['expirationTime'] == iso_str

    def test_rollback_to_snapshot_no_params(self, mock_call_method):

        self.cluster_fm.modify_group_snapshot(self.mock_group_snapshot)

        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.modify_group_snapshot
        assert mock_call_method.call_args[1]['params'][
                   'groupSnapshotID'] == self.mock_group_snapshot.group_snapshot_id
        assert mock_call_method.call_args[1]['params']['enableRemoteReplication'] is None
        assert mock_call_method.call_args[1]['params']['expirationTime'] is None
