import unittest

import mock

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestVolumeAccessGroupMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.vag_dm = models.VolumeAccessGroup.get_mock_object()

    def test_create_volume_access_group(self, mock_call_method):
        mock_response = result_models.CreateVolumeAccessGroup.get_mock_response(
            version=models.SfVersion('9.0'))
        mock_volumes = models.Volume.get_mock_object_list(5)
        mock_call_method.return_value = mock_response
        name = 'a name'
        result = self.cluster_fm.create_volume_access_group(name, volume_dms=mock_volumes)
        params = {
            api_constants.NAME: name,
            api_constants.INITIATORS: None,
            api_constants.VOLUMES: [vol.volume_id for vol in mock_volumes],
            api_constants.ATTRIBUTES: None,
            api_constants.FORCE_DURING_UPGRADE: False
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.create_volume_access_group,
                                            params=params,
                                            result_dm_class=result_models.CreateVolumeAccessGroup,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.VolumeAccessGroup)

    def test_delete_volume_access_group(self, mock_call_method):
        mock_vag = models.VolumeAccessGroup.get_mock_object()
        self.cluster_fm.delete_volume_access_group(mock_vag)
        params = {
            api_constants.VOLUME_ACCESS_GROUP_ID: mock_vag.volume_access_group_id,
            api_constants.DELETE_ORPHAN_INITIATORS: None,
            api_constants.FORCE: None,
            api_constants.FORCE_DURING_UPGRADE: False
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.delete_volume_access_group,
                                            params=params,
                                            result_dm_class=None,
                                            auth=mock.ANY, log_response=True)

    def test_modify_volume_access_group(self, mock_call_method):
        mock_response = result_models.ModifyVolumeAccessGroup.get_mock_response(
            version=models.SfVersion('9.0'))
        mock_volumes = models.Volume.get_mock_object_list(5)
        mock_call_method.return_value = mock_response
        name = 'new name'
        result = self.cluster_fm.modify_volume_access_group(
            volume_access_group_dm=self.vag_dm,
            name=name,
            volume_dms=mock_volumes,
            delete_orphan_initiators=True
        )
        params = {
            api_constants.VOLUME_ACCESS_GROUP_ID: self.vag_dm.volume_access_group_id,
            api_constants.NAME: name,
            api_constants.INITIATORS: None,
            api_constants.VOLUMES: [vol.volume_id for vol in mock_volumes],
            api_constants.ATTRIBUTES: None,
            api_constants.DELETE_ORPHAN_INITIATORS: True,
            api_constants.FORCE_DURING_UPGRADE: False
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.modify_volume_access_group,
                                            params=params,
                                            result_dm_class=result_models.ModifyVolumeAccessGroup,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result, models.VolumeAccessGroup)

    def test_list_volume_access_groups(self, mock_call_method):
        mock_response = result_models.ListVolumeAccessGroups.get_mock_response(
            version=models.SfVersion('9.0'))
        mock_call_method.return_value = mock_response
        vag_id = 3
        result = self.cluster_fm.list_volume_access_groups(start_volume_access_group_id=vag_id)
        params = {
            api_constants.START_VOLUME_ACCESS_GROUP_ID: vag_id,
            api_constants.LIMIT: None,
            api_constants.VOLUME_ACCESS_GROUPS: None
        }
        mock_call_method.assert_called_with(enums.ClusterMethod.list_volume_access_groups,
                                            params=params,
                                            result_dm_class=result_models.ListVolumeAccessGroups,
                                            auth=mock.ANY, log_response=True)
        assert isinstance(result[0], models.VolumeAccessGroup)
