import unittest
import uuid

import mock
import pytest

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
@mock.patch('valence.functional_models.Cluster.list_volumes')
class TestClusterCreateVolume(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dm = models.Volume.get_mock_object()

    def test_create_volume_pre_fluorine(self, mock_list_volumes, mock_call_method):
        mock_response = result_models.CreateVolume.get_mock_response(
            version=models.SfVersion('8.0'), overrides=dict(volume_id=self.volume_dm.volume_id,
                                                            curve=self.volume_dm.qos.curve))
        mock_list_volumes.return_value = [self.volume_dm]
        mock_call_method.return_value = mock_response
        volume_dm = self.cluster_fm.create_volume(self.volume_dm.name, self.volume_dm.account_id,
                                                  self.volume_dm.total_size,
                                                  min_iops=self.volume_dm.qos.min_iops)
        assert volume_dm == self.volume_dm

    def test_response_missing_volume_id(self, mock_list_volumes, mock_call_method):
        mock_response = result_models.CreateVolume.get_mock_response(
            overrides=dict(volume_id=None, volume=None))
        mock_call_method.return_value = mock_response
        with pytest.raises(functional_models.FunctionalClusterError):
            self.cluster_fm.create_volume(self.volume_dm.name, self.volume_dm.account_id,
                                          self.volume_dm.total_size)

    def test_create_volume_fluorine(self, mock_list_volumes, mock_call_method):
        mock_response = result_models.CreateVolume.get_mock_response(
            overrides=dict(volume=self.volume_dm, curve=self.volume_dm.qos.curve,
                           volume_id=self.volume_dm.volume_id))
        mock_call_method.return_value = mock_response
        volume_dm = self.cluster_fm.create_volume(self.volume_dm.name, self.volume_dm.account_id,
                                                  self.volume_dm.total_size)
        assert volume_dm == self.volume_dm

    def test_create_volume_neon(self, mock_list_volumes, mock_call_method):
        volume_dm_neon = models.Volume.get_mock_object(version=sf_api.SfVersion('10.0'))
        mock_response = result_models.CreateVolume.get_mock_response(
            version=models.SfVersion('10.0'), overrides=dict(volume=volume_dm_neon,
                                                             curve=volume_dm_neon.qos.curve,
                                                             volume_id=volume_dm_neon.volume_id,
                                                             qos_policy_id=volume_dm_neon.qos_policy_id))
        mock_call_method.return_value = mock_response
        volume_dm = self.cluster_fm.create_volume(self.volume_dm.name, self.volume_dm.account_id,
                                                  self.volume_dm.total_size)
        assert volume_dm == volume_dm_neon


@mock.patch.object(functional_models.Cluster, 'get_constants')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
@mock.patch('valence.functional_models.Cluster.list_volumes')
class TestCreateMultipleVolumes(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dm = models.Volume.get_mock_object()
        cls.volume_id = cls.volume_dm.volume_id
        cls.volume_dm_neon = models.Volume.get_mock_object(version=models.SfVersion('10.0'))
        qos_id_neon = cls.volume_dm_neon.qos_policy_id
        cls.mock_response = result_models.CreateMultipleVolumes.get_mock_response(
            version=models.SfVersion('8.0'), overrides=dict(volumes=[cls.volume_dm],
                                                            volume_ids=[cls.volume_id]))
        cls.mock_response_neon = result_models.CreateMultipleVolumes.get_mock_response(
            version=models.SfVersion('10.0'), overrides=dict(volumes=[cls.volume_dm_neon],
                                                             qos_ids=[qos_id_neon]))

    def test_create_multiple_volumes_ids_returned(self, mock_list_volumes, mock_call_method,
                                                  mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_call_method.return_value = self.mock_response
        mock_list_volumes.return_value = [self.volume_dm]
        volume_dms = self.cluster_fm.create_multiple_volumes(
            number=1, name_prefix=self.volume_dm.name, account_id=self.volume_dm.account_id,
            total_size=self.volume_dm.total_size, min_iops=self.volume_dm.qos.min_iops)
        assert volume_dms[0] == self.volume_dm

    def test_create_multiple_volumes_qos_ids_returned(self, mock_list_volumes, mock_call_method,
                                                      mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_call_method.return_value = self.mock_response_neon
        mock_list_volumes.return_value = [self.volume_dm_neon]
        volume_dms = self.cluster_fm.create_multiple_volumes(
            number=1, name_prefix=self.volume_dm_neon.name, account_id=self.volume_dm_neon.account_id,
            total_size=self.volume_dm_neon.total_size, min_iops=self.volume_dm_neon.qos.min_iops)
        assert volume_dms[0] == self.volume_dm_neon

    def test_create_multiple_volumes_batching(self, mock_list_volumes, mock_call_method,
                                              mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_list_volumes.return_value = [self.volume_dm for _ in range(20)]
        self.cluster_fm.create_multiple_volumes(
            number=20, name_prefix=self.volume_dm.name, account_id=self.volume_dm.account_id,
            total_size=self.volume_dm.total_size, min_iops=self.volume_dm.qos.min_iops)
        assert mock_call_method.call_count == 4

    def test_create_multiple_volumes_batching_default(self, mock_list_volumes, mock_call_method,
                                                      mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=None))
        mock_list_volumes.return_value = [self.volume_dm for _ in range(1001)]
        self.cluster_fm.create_multiple_volumes(
            number=1001, name_prefix=self.volume_dm.name, account_id=self.volume_dm.account_id,
            total_size=self.volume_dm.total_size, min_iops=self.volume_dm.qos.min_iops)
        assert mock_call_method.call_count == 3

    def test_response_missing_volume(self, mock_list_volumes, mock_call_method, mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_call_method.return_value = self.mock_response
        mock_list_volumes.return_value = []
        with pytest.raises(functional_models.FunctionalClusterError):
            self.cluster_fm.create_multiple_volumes(
                number=1, name_prefix=self.volume_dm.name, account_id=self.volume_dm.account_id,
                total_size=self.volume_dm.total_size, min_iops=self.volume_dm.qos.min_iops)

    def test_response_missing_volume_id(self, mock_list_volumes, mock_call_method,
                                        mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_response = result_models.CreateMultipleVolumes.get_mock_response(
            overrides=dict(volume_ids=None))
        mock_call_method.return_value = mock_response
        with pytest.raises(functional_models.FunctionalClusterError):
            self.cluster_fm.create_multiple_volumes(
                number=1, name_prefix=self.volume_dm.name, account_id=self.volume_dm.account_id,
                total_size=self.volume_dm.total_size, min_iops=self.volume_dm.qos.min_iops)

    def test_create_multiple_volumes_not_returned(self, mock_list_volumes, mock_call_method,
                                                  mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_response = result_models.CreateMultipleVolumes.get_mock_response(
            overrides=dict(volumes=None, volume_ids=None))
        mock_call_method.return_value = mock_response
        with pytest.raises(functional_models.FunctionalClusterError):
            self.cluster_fm.create_multiple_volumes(
                number=1, name_prefix=self.volume_dm.name, account_id=self.volume_dm.account_id,
                total_size=self.volume_dm.total_size, min_iops=self.volume_dm.qos.min_iops)


@mock.patch('sf_api.ClusterApi.get_api')
@mock.patch.object(functional_models.Cluster, 'get_constants')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterDeletePurgeVolumes(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dm = models.Volume(
            result_models.DeleteVolume.get_mock_data(version=models.SfVersion('8.0'),
                                                     overrides=dict(volume_id=5)))
        cls.mocked_volumes = models.Volume.get_mock_data_list(5)
        cls.mocked_responses = [
            result_models.DeleteVolume.get_mock_response(
                overrides=dict(volume=models.Volume(vol), volume_id=vol.get(api_constants.VOLUME_ID),
                               name=vol.get(api_constants.NAME)))
            for vol in cls.mocked_volumes]

        cls.mocked_responses_plural = result_models.DeleteVolumes.get_mock_response()
        max_requests = 2
        volume_data_chunks = [cls.mocked_volumes[x:x + max_requests]
                              for x in range(0, len(cls.mocked_volumes), max_requests)]

        cls.mocked_responses_chunks = [
            result_models.DeleteVolumes.get_mock_response(overrides={'volumes': volume_data})
            for volume_data in volume_data_chunks]

    def test_delete_volumes_plural(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_get_api.return_value.result = {
            self.cluster_fm.cluster_dm.api_endpoint.version_string: ['DeleteVolumes']}
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_call_method.return_value = self.mocked_responses_plural
        mocked_dms = [models.Volume(vol) for vol in self.mocked_volumes]
        response = self.cluster_fm.delete_volumes(mocked_dms)  # List with same vol 5 times

        assert mock_call_method.called
        assert response == self.mocked_responses_plural.result_dm.volumes

    def test_delete_volumes_iterative(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_get_api.return_value.result = {self.cluster_fm.cluster_dm.api_endpoint.version_string: []}
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        mock_call_method.side_effect = self.mocked_responses
        mocked_dms = [models.Volume(vol) for vol in self.mocked_volumes]
        response = self.cluster_fm.delete_volumes(mocked_dms)  # List with same vol 5 times
        assert len(response) == len(self.mocked_volumes)
        for i in range(5):
            assert mocked_dms[i] == response[i]

    def test_delete_volumes_chunks(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_get_api.return_value.result = {
            self.cluster_fm.cluster_dm.api_endpoint.version_string: ['DeleteVolumes']}
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=2))
        mock_call_method.side_effect = self.mocked_responses_chunks
        mocked_dms = [models.Volume(vol) for vol in self.mocked_volumes]

        response = self.cluster_fm.delete_volumes(mocked_dms)
        assert mock_call_method.call_count == 3
        assert len(response) == 5
        for i in range(5):
            assert mocked_dms[i] == response[i]

    def test_purge_volumes_plural(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_get_api.return_value.result = {
            self.cluster_fm.cluster_dm.api_endpoint.version_string: ['PurgeDeletedVolumes']}
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        self.cluster_fm.purge_deleted_volumes([self.volume_dm]*5)

    def test_purge_volumes_chunks(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_get_api.return_value.result = {
            self.cluster_fm.cluster_dm.api_endpoint.version_string: ['PurgeDeletedVolumes']}
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        self.cluster_fm.purge_deleted_volumes([self.volume_dm] * 15)

    def test_purge_volumes_iterative(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_get_api.return_value.result = {
            self.cluster_fm.cluster_dm.api_endpoint.version_string: []}
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
            overrides=dict(max_volumes_per_multi_api=5))
        self.cluster_fm.purge_deleted_volumes([self.volume_dm] * 5)

    def test_purge_all_deleted_volumes(self, mock_call_method, mock_get_constants, mock_get_api):
        mock_response = result_models.PurgeAllDeletedVolumes.get_mock_response(
            overrides=dict(volume_ids=[self.volume_dm.volume_id]))
        mock_call_method.return_value = mock_response
        self.cluster_fm.purge_all_deleted_volumes()


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestCloneVolume(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dm = models.Volume.get_mock_object()

    def test_clone_volume(self, mock_call_method):
        mock_response = result_models.CloneVolume.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.clone_volume(self.volume_dm)
        assert self.volume_dm.name == mock_call_method.call_args[1].get('params', {}).get('name')

    def test_clone_volume_override(self, mock_clone_volume):
        mock_response = result_models.CloneVolume.get_mock_response()
        mock_clone_volume.return_value = mock_response
        self.cluster_fm.clone_volume(self.volume_dm, name='shit', account_id=9000)
        assert 'shit' == mock_clone_volume.call_args[1].get('params', {}).get('name')
        assert 9000 == mock_clone_volume.call_args[1].get('params', {}).get('newAccountID')

    def test_clone_volume_from_snapshot(self, mock_call_method):
        snapshot_dm = models.Snapshot.get_mock_object()
        mock_response = result_models.CloneVolume.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.clone_volume_from_snapshot(snapshot_dm, 'Glados')


@mock.patch.object(functional_models.Cluster, 'api_supported')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestGetDefaultQoS(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.default_qos = {
            api_constants.MIN_IOPS: 100,
            api_constants.MAX_IOPS: 15000,
            api_constants.BURST_IOPS: 15000,
            api_constants.BURST_TIME: 60,
            api_constants.CURVE: {
                '4096': 100,
                '8192': 160,
                '16384': 270,
                '32768': 500,
                '65536': 1000,
                '131072': 1900,
                '262144': 3000,
                '524288': 7500,
                '1048576': 15000
            }
        }
        cls.mock_response = result_models.GetDefaultQoS.get_mock_response()

    def test_get_default_qos_not_in_api(self, call_method, mock_api_supported):
        mock_api_supported.return_value = False
        qos = self.cluster_fm.get_default_qos()
        assert qos.serialize() == self.default_qos

    def test_get_default_qos_in_api(self, call_method, mock_api_supported):
        mock_api_supported.return_value = True
        call_method.return_value = self.mock_response
        qos = self.cluster_fm.get_default_qos()
        assert qos.serialize() == self.mock_response.result_dm.serialize()


class TestGetVolumeEfficiency(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_response = result_models.GetVolumeEfficiency.get_mock_response()
        self.volume = models.Volume.get_mock_object()

    @mock.patch('sf_api.ClusterApi.call_method',
                side_effect=unittest_utils.helpers.check_params_serializable)
    def test_get_volume_efficiency(self, mock_call_method):
        mock_call_method.return_value = self.mock_response
        volume_efficiency = self.cluster_fm.get_volume_efficiency(self.volume)
        assert mock_call_method.called_once()
        assert isinstance(volume_efficiency, result_models.GetVolumeEfficiency)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestClusterListVolumes(unittest.TestCase):
    def setUp(self):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()
        self.cluster_fm.cluster_dm.api_endpoint = models.SfVersion('8.0')
        self.account_id = 42
        self.volume_dm = models.Volume.get_mock_object(version=models.SfVersion('8.0'))

    def test_api_filter(self, mock_call_method):
        cluster_dm = models.ResourceCluster.get_mock_object(
            overrides=dict(api_endpoint=models.SfVersion('9.0')))
        cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=[self.volume_dm]))
        mock_call_method.return_value = mock_response
        volume_dms = cluster_fm.list_volumes(account_ids=[self.account_id])
        assert len(volume_dms) == 1

    def test_manual_filter_mutually_exclusive(self, mock_call_method):
        with pytest.raises(ValueError):
            self.cluster_fm.list_volumes(account_ids=[self.account_id], volume_ids=[1])

    def test_manual_filter_accounts(self, mock_call_method):
        volume_dm = models.Volume.get_mock_object(overrides=dict(
            account_id=self.account_id+1))
        volume_dm2 = models.Volume.get_mock_object(overrides=dict(
            account_id=self.account_id))
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=[volume_dm, volume_dm2]))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(account_ids=[self.account_id])
        assert len(volume_dms) == 1

    def test_manual_filter_volume_ids(self, mock_call_method):
        volumes = [models.Volume.get_mock_data(overrides=dict(volume_id=1)),
                   models.Volume.get_mock_data(overrides=dict(volume_id=2))]
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=volumes))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(volume_ids=[1, 3])
        assert len(volume_dms) == 1

    def test_manual_filter_single_volume_id(self, mock_call_method):
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=[self.volume_dm]))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(volume_ids=[self.volume_dm.volume_id])
        assert len(volume_dms) == 1

    def test_manual_filter_volume_name(self, mock_call_method):
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=[self.volume_dm]))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(volume_name=self.volume_dm.name)
        assert len(volume_dms) == 1
        volume_dms = self.cluster_fm.list_volumes(volume_name=self.volume_dm.name+'x')
        assert len(volume_dms) == 0

    def test_manual_filter_is_paired(self, mock_call_method):
        volume = models.Volume.get_mock_data(overrides=dict(volume_pairs=None))
        mock_response = result_models.ListVolumes.get_mock_response(
            overrides=dict(volumes=[volume]))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(is_paired=True)
        assert len(volume_dms) == 0
        volume_dms = self.cluster_fm.list_volumes(is_paired=False)
        assert len(volume_dms) == 1

    def test_manual_filter_vvol(self, mock_call_method):
        volume = models.Volume.get_mock_data(overrides=dict(virtual_volume_id=uuid.uuid4()))
        mock_response = result_models.ListVolumes.get_mock_response(overrides=dict(volumes=[volume]))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(include_virtual_volumes=False)
        assert len(volume_dms) == 0
        volume_dms = self.cluster_fm.list_volumes(include_virtual_volumes=True)
        assert len(volume_dms) == 1

    def test_manual_filter_status(self, mock_call_method):
        volume = models.Volume.get_mock_object(overrides=dict(status='active'))
        mock_response = result_models.ListVolumes.get_mock_response(overrides=dict(volumes=[volume]))
        mock_call_method.return_value = mock_response
        volume_dms = self.cluster_fm.list_volumes(volume_status=sf_api.VolumeStatus.deleted)
        assert len(volume_dms) == 0
        volume_dms = self.cluster_fm.list_volumes(volume_status=sf_api.VolumeStatus.active)
        assert len(volume_dms) == 1


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestReadSliceLba(unittest.TestCase):
    def test_read_slice_lba(self, call_method):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        mock_response = result_models.ReadSliceLBA.get_mock_response()
        call_method.return_value = mock_response
        result = cluster_fm.read_slice_lbas(1, 1, 1)
        assert isinstance(result, dict)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestReadSliceLBAChecksum(unittest.TestCase):
    def test_read_slice_lba_checksum(self, call_method):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        mock_response = result_models.ReadSliceLBAChecksum.get_mock_response()
        call_method.return_value = mock_response
        result = cluster_fm.read_slice_lba_checksum(1, 1, 1)
        assert isinstance(result, dict)


@mock.patch('valence.functional_models.SliceReassignmentAsyncHandler')
@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestSlice(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_response = result_models.GetSliceInfo.get_mock_response()
        self.volumes = models.Volume.get_mock_object_list(5)

    def test_get(self, mock_get, mock_wait):
        mock_get.return_value = self.mock_response
        self.cluster_fm.get_slice_info(1)

    def test_reassign(self, mock_reassign, mock_wait):
        mock_reassign.return_value = self.mock_response
        self.cluster_fm.reassign_slice(1, 1, [1])

    def test_dbversion(self, mock_reassign, mock_wait):
        mock_reassign.side_effect = iter([self.mock_response,
                                          sf_api.api_errors.DBVersionMismatch('stuff', {}),
                                          self.mock_response,
                                          self.mock_response])

        self.cluster_fm.reassign_slice(1, 1, [1])

        mock_reassign.side_effect = iter([self.mock_response,
                                          sf_api.api_errors.DBVersionMismatch('stuff', {}),
                                          sf_api.api_errors.DBVersionMismatch('stuff', {})])
        with pytest.raises(sf_api.api_errors.DBVersionMismatch):
            self.cluster_fm.reassign_slice(1, 1, [1])

    def test_fill_slice(self, mock_call_method, _):
        mock_response = result_models.FillSlice.get_mock_response()
        mock_call_method.return_value = mock_response
        self.cluster_fm.fill_slice(123)
        assert 123 == mock_call_method.call_args[1].get('params', {}).get('sliceID')

    def test_rebalance_slices(self, mock_call_method, mock_wait):
        mock_response = None
        mock_call_method.return_value = mock_response
        self.cluster_fm.rebalance_slices()


class TestModifyVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_response = result_models.ModifyVolumes.get_mock_response()
        self.volumes = models.Volume.get_mock_object_list(5)

    @mock.patch('sf_api.ClusterApi.call_method',
                side_effect=unittest_utils.helpers.check_params_serializable)
    def test_modify_volumes(self, mock_mod):
        mock_mod.return_value = self.mock_response
        self.cluster_fm.modify_volumes(self.volumes, min_iops=1, max_iops=1, burst_iops=0)
        assert mock_mod.called_once()
        mvols = self.cluster_fm.modify_volumes(self.volumes, access='locked')
        assert isinstance(mvols, list)
        assert isinstance(mvols[0], models.Volume)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestVolumePairing(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_volume_dms = models.Volume.get_mock_object_list(2)

    def test_start_volume_pairing(self, mock_call_method):
        mock_response = result_models.StartVolumePairing.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.start_volume_pairing(self.mock_volume_dms[0],
                                                      enums.RemoteRepMode.snapshots_only)
        assert result == mock_response.result_dm.volume_pairing_key

    def test_complete_volume_pairing(self, mock_call_method):
        self.cluster_fm.complete_volume_pairing(self.mock_volume_dms[0], 'somepairkey')
        mock_call_method.assert_called_with(
            enums.ClusterMethod.complete_volume_pairing,
            params={api_constants.VOLUME_ID: self.mock_volume_dms[0].volume_id,
                    api_constants.VOLUME_PAIRING_KEY: 'somepairkey'},
            result_dm_class=None,
            auth=mock.ANY, log_response=True)

    def test_list_active_paired_volumes(self, mock_call_method):
        mock_response = result_models.ListActivePairedVolumes.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.list_active_paired_volumes()
        assert result == mock_response.result_dm.volumes

    def test_modify_volume_pair(self, mock_call_method):
        self.cluster_fm.modify_volume_pair(self.mock_volume_dms[0],
                                           False,
                                           enums.RemoteRepMode.snapshots_only)
        mock_call_method.assert_called_with(
            enums.ClusterMethod.modify_volume_pair,
            params={api_constants.VOLUME_ID: self.mock_volume_dms[0].volume_id,
                    api_constants.PAUSED_MANUAL: False,
                    api_constants.MODE: enums.RemoteRepMode.snapshots_only},
            result_dm_class=None,
            auth=mock.ANY, log_response=True)

    def test_remove_volume_pair(self, mock_call_method):
        self.cluster_fm.remove_volume_pair(self.mock_volume_dms[0])
        mock_call_method.assert_called_with(
            enums.ClusterMethod.remove_volume_pair,
            params={api_constants.VOLUME_ID: self.mock_volume_dms[0].volume_id},
            result_dm_class=None,
            auth=mock.ANY, log_response=True)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestRestoreDeletedVolume(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_volume_dm = models.Volume.get_mock_object()

    def test_restore_deleted_volume(self, mock_call_method):
        self.cluster_fm.restore_deleted_volume(self.mock_volume_dm)
        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.restore_deleted_volume
        assert mock_call_method.call_args[1]['params']['volumeID'] == self.mock_volume_dm.volume_id


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestComputeVolumeChecksum(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_volume_dm = models.Volume.get_mock_object()

    def test_compute_volume_checksum_expect_pass(self, mock_call_method):
        mock_response = result_models.ComputeVolumeChecksum.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.cluster_fm.compute_volume_checksum(self.mock_volume_dm)
        assert result == mock_response.result_dm.checksums


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListVolumeQosHistogram(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_volume_dm = models.Volume.get_mock_object()

    def test_list_volume_qos_histogram_all(self, mock_call_method):
        mock_call_method.return_value = result_models.ListVolumeQoSHistograms.get_mock_response()
        result = self.cluster_fm.list_volume_qos_histograms()
        assert isinstance(result[0], models.VolumeQosHistograms)

    def test_list_volume_qos_histogram_by_id(self, mock_call_method):
        self.cluster_fm.list_volume_qos_histograms([self.mock_volume_dm.volume_id])
        assert mock_call_method.call_args[0][0] == enums.ClusterMethod.list_volume_qos_histograms
        assert mock_call_method.call_args[1]['params'][api_constants.VOLUME_IDS] ==\
            [self.mock_volume_dm.volume_id]


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestListVolumeStats(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_list_volume_stats_by_volume(self, mock_call_method):
        mock_call_method.return_value = result_models.ListVolumeStatsByVolume.get_mock_response()

        result = self.cluster_fm.list_volume_stats_by_volume()

        assert isinstance(result[0], models.VolumeStats)


@mock.patch('sf_api.ClusterApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestMovePrimariesAwayFromNode(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_move_primaries_away_from_node(self, mock_call_method):
        result = self.cluster_fm.move_primaries_away_from_node(1)

        mock_call_method.assert_called_with(enums.ClusterMethod.move_primaries_away_from_node,
                                            params={api_constants.NODE_ID: 1},
                                            result_dm_class=result_models.MovePrimariesAwayFromNode,
                                            auth=mock.ANY, log_response=True)
