"""Cluster FM methods related to virtual networks."""
from sf_api import result_models
from valence import enums
from valence.functional_models._cluster import base


class VirtualNetworkMixin(base.ClusterBase):
    """Cluster FM Virtual Network methods."""

    def create_virtual_network(self, address_blocks, name, netmask, svip, virtual_network_tag,
                               attributes=None, gateway=None, namespace=False):
        """Add a virtual network to cluster.

        :param list[valence.models.AddressBlocks] address_blocks: Range of address blocks assigned
            to the VLAN
        :param str name: User defined name for the new VLAN
        :param str netmask: Network mask for VLAN
        :param str svip: Storage IP for VLAN
        :param dict attributes: Any additional key value pairs to add to the VLAN
        :param str gateway: IP address of the VLAN gateway. Only valid when namespace is set to true
        :param bool namespace: Enables routable storage VLANs functionality
        :param int virtual_network_tag: Unique VLAN tag, supported values are 1 through 4094
        :rtype: int
        """
        params = {'addressBlocks': [block.serialize() for block in address_blocks],
                  'attributes': attributes,
                  'gateway': gateway,
                  'name': name,
                  'namespace': namespace,
                  'netmask': netmask,
                  'svip': svip,
                  'virtualNetworkTag': virtual_network_tag,
                  }
        response = self.call_method(enums.ClusterMethod.add_virtual_network, params=params,
                                    result_dm_class=result_models.AddVirtualNetwork)
        return response.result_dm.virtual_network_id

    def modify_virtual_network(self, virtual_network_tag, address_blocks=None, attributes=None,
                               gateway=None, name=None, namespace=None, netmask=None, svip=None,
                               virtual_network_id=None):
        """Modify a virtual network on a cluster.

        :param int virtual_network_tag: The network tag that identifies the VLAN to modify.
        :param list[valence.models.AddressBlocks] address_blocks: Range of address blocks to assign
            to the VLAN
        :param dict attributes: Any key value pairs to modify the VLAN
        :param str gateway: IP address of the VLAN gateway. Only valid when namespace is set to true
        :param str name: User defined name for the VLAN.
        :param bool namespace: Enables routable storage VLANs functionality
        :param str netmask: Network mask for the VLAN.
        :param str svip: Storage virtual IP for the VLAN.
        :param int virtual_network_id: VLAN ID that identifies the VLAN to modify. One of either this
            parameter or virtual_network_tag is required, but not both.
        """
        params = {'addressBlocks': None if address_blocks is None else [block.serialize() for block in
                                                                        address_blocks],
                  'attributes': attributes,
                  'gateway': gateway,
                  'name': name,
                  'namespace': namespace,
                  'netmask': netmask,
                  'svip': svip,
                  'virtualNetworkTag': virtual_network_tag,
                  'virtualNetworkID': virtual_network_id
                  }
        self.call_method(enums.ClusterMethod.modify_virtual_network, params=params,
                         result_dm_class=None)

    def list_virtual_networks(self, virtual_network_ids=None, virtual_network_tags=None):
        """List virtual networks on a cluster.

        :param list[int] virtual_network_ids: List of network IDs to include in the list
        :param list[int] virtual_network_tags: List of network tags to include in the list
        :rtype: list[valence.models.VirtualNetwork]
        """
        params = {'virtualNetworkIDs': virtual_network_ids,
                  'virtualNetworkTags': virtual_network_tags}
        response = self.call_method(enums.ClusterMethod.list_virtual_networks, params=params,
                                    result_dm_class=result_models.ListVirtualNetworks)
        return response.result_dm.virtual_networks

    def delete_virtual_network(self, virtual_network_tag, virtual_network_id=None):
        """Remove virtual network from a cluster.

        :param int virtual_network_tag: Virtual network tag to remove
        :param int virtual_network_id: Virtual network ID to remove. One of either this parameter or
            virtual_network_tag is required, but not both.
        """
        params = {'virtualNetworkID': virtual_network_id,
                  'virtualNetworkTag': virtual_network_tag}
        self.call_method(enums.ClusterMethod.remove_virtual_network, params=params,
                         result_dm_class=None)
