"""Cluster FM methods related to volume access groups."""
from __future__ import division

from sf_api import (
    api_constants,
    enums,
    result_models,
)
from valence.functional_models._cluster import base


class VolumeAccessGroupMixin(base.ClusterBase):
    """Cluster FM Volume Access Group methods."""

    def create_volume_access_group(self, name, initiators=None, volume_dms=None, attributes=None):
        """Call CreateVolumeAccessGroup.

        :param str name: Name for the VAG.
        :param list[str] initiators: Initiators to add to the VAG.
        :param list[valence.models.Volume] volume_dms: Volumes to add to the VAG.
        :param dict attributes: Attributes you want to add.
        :rtype: valence.models.VolumeAccessGroup
        """
        params = {
            api_constants.NAME: name,
            api_constants.INITIATORS: initiators,
            api_constants.VOLUMES: [vol.volume_id for vol in volume_dms],
            api_constants.ATTRIBUTES: attributes,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }

        response = self.call_method(
            method=enums.ClusterMethod.create_volume_access_group,
            params=params,
            result_dm_class=result_models.CreateVolumeAccessGroup
        )
        return response.result_dm.volume_access_group

    def delete_volume_access_group(self, volume_access_group_dm, delete_orphan_initiators=None,
                                   force=None):
        """Call DeleteVolumeAccessGroup.

        :param valence.models.VolumeAccessGroup volume_access_group_dm: VAG to delete.
        :param bool delete_orphan_initiators: Whether to delete initiators when the volume access group
            is deleted (default: True).
        :param bool force: Force the VAG deletion even if it has a VLAN TAG or ID (default: False).
        """
        params = {
            api_constants.VOLUME_ACCESS_GROUP_ID: volume_access_group_dm.volume_access_group_id,
            api_constants.DELETE_ORPHAN_INITIATORS: delete_orphan_initiators,
            api_constants.FORCE: force,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }
        self.call_method(
            method=enums.ClusterMethod.delete_volume_access_group,
            params=params,
            result_dm_class=None
        )

    def modify_volume_access_group(self, volume_access_group_dm, name=None, initiators=None,
                                   volume_dms=None, attributes=None, delete_orphan_initiators=None):
        """Call ModifyVolumeAccessGroup.

        :param valence.models.VolumeAccessGroup volume_access_group_dm: VAG to modify.
        :param str name: New name for the VAG.
        :param list[str] initiators: Initiators to add to the VAG.
        :param list[valence.models.Volume] volume_dms: Volumes to add to the VAG.
        :param dict attributes: Attributes you want to add.
        :param bool delete_orphan_initiators: Whether to delete initiators when they are removed from the
            volume access group (default: True).
        :return: Volume access group as returned by the API.
        :rtype: valence.models.VolumeAccessGroup
        """
        params = {
            api_constants.VOLUME_ACCESS_GROUP_ID: volume_access_group_dm.volume_access_group_id,
            api_constants.NAME: name,
            api_constants.INITIATORS: initiators,
            api_constants.VOLUMES: [vol.volume_id for vol in volume_dms],
            api_constants.ATTRIBUTES: attributes,
            api_constants.DELETE_ORPHAN_INITIATORS: delete_orphan_initiators,
            api_constants.FORCE_DURING_UPGRADE: self.upgrading
        }
        response = self.call_method(
            method=enums.ClusterMethod.modify_volume_access_group,
            params=params,
            result_dm_class=result_models.ModifyVolumeAccessGroup
        )
        return response.result_dm.volume_access_group

    def list_volume_access_groups(self, start_volume_access_group_id=None, limit=None,
                                  volume_access_groups=None):
        """Call ListVolumeAccessGroups.

        :param int start_volume_access_group_id: The volume access group ID at which to begin listing.
        :param int limit: Maximum number of volume access groups to list.
        :param list volume_access_groups: Volume access group IDs for the volume access groups to list
            (mutually exclusive with "start_volume_access_group_id" and "limit").
        :return: Volume access group data models
        :rtype: list[valence.models.VolumeAccessGroup]
        """
        params = {
            api_constants.START_VOLUME_ACCESS_GROUP_ID: start_volume_access_group_id,
            api_constants.LIMIT: limit,
            api_constants.VOLUME_ACCESS_GROUPS: volume_access_groups
        }
        response = self.call_method(
            method=enums.ClusterMethod.list_volume_access_groups,
            params=params,
            result_dm_class=result_models.ListVolumeAccessGroups
        )
        return response.result_dm.volume_access_groups
