"""Cluster FM methods related to volumes."""
import autologger

import sf_api
from sf_api import (
    api_constants,
    enums,
    result_models,
)
from valence import (
    criteria,
    models,
)
from valence.functional_models import (
    _async_handler,
    _errors,
)
from valence.functional_models._cluster import base

_log = autologger.get_sflogging()


class VolumeMixin(base.ClusterBase):
    """Cluster FM Volume methods."""

    def clone_volume(self, volume_dm, name=None, account_id=None, size=None, access=None,
                     attributes=None):
        """Clone a volume on a cluster via CloneVolume. New volume data model will be returned.

        Attributes of the volume DM will be applied to the clone. You can use the optional args
        to change the clone's name, account, size, etc...

        :param valence.models.Volume volume_dm: Data model with the source volume's volumeID.
        :param str name: 1 - 64 chars, not required to be unique.
        :param int account_id: Account for the Volume to belong to.
        :param int size: Size in bytes, gets rounded up to nearest MB, may be larger or smaller.
        :param valence.enums.VolumeAccess access: Access mode enum
        :param dict attributes: Key/Value pairs.
        :return: CloneVolumesAsyncHandler handler
        :rtype: valence.functional_models.CloneVolumeAsyncHandler
        """
        params = {
            api_constants.VOLUME_ID: volume_dm.volume_id,
            api_constants.NAME: name or volume_dm.name,
            api_constants.NEW_ACCOUNT_ID: account_id or volume_dm.account_id,
            api_constants.NEW_SIZE: size or volume_dm.total_size.bytes,
            api_constants.VOLUME_ACCESS: access or volume_dm.access,
            api_constants.ATTRIBUTES: attributes or volume_dm.attributes,
        }
        response = self.call_method(enums.ClusterMethod.clone_volume, params=params,
                                    result_dm_class=result_models.CloneVolume)
        # Extract the asyncHandle and new volume object from the API response
        async_handle_dm = models.AsyncHandle(
            {api_constants.ASYNC_HANDLE: response.result_dm.async_handle}
        )
        clone_volume_dm = models.Volume(response.result_dm.volume)
        # Create an async handler to keep track of the clone's progress
        async_handler_fm = _async_handler.CloneVolumeAsyncHandler(self, clone_volume_dm.volume_id,
                                                                  async_handle_dm=async_handle_dm)
        return async_handler_fm

    def clone_volume_from_snapshot(self, snapshot_dm, name, account_dm=None, total_size=None,
                                   access=None):
        """Clone the provided snapshot on a cluster via CloneVolume.

        If optional inputs are not provided they will default to the attributes of the source volume.

        :param valence.models.Snapshot snapshot_dm: Snapshot to source new volume.
        :param str name: Name of new volume.
        :param valence.models.Account account_dm: Account for new volume.
        :param valence.models.DataSize total_size: Size of new volume.
        :param valence.enums.VolumeAccess access: Read/write access for new volume.
        :return: CloneVolumesAsyncHandler handler
        :rtype: valence.functional_models.CloneVolumeAsyncHandler
        """
        params = {
            api_constants.VOLUME_ID: snapshot_dm.volume_id,
            api_constants.NAME: name,
            api_constants.NEW_ACCOUNT_ID: account_dm.account_id if account_dm else None,
            api_constants.NEW_SIZE: total_size.bytes if total_size else None,
            api_constants.VOLUME_ACCESS: access if access else None,
        }
        response = self.call_method(enums.ClusterMethod.clone_volume, params=params,
                                    result_dm_class=result_models.CloneVolume)

        async_handle_dm = models.AsyncHandle(
            {api_constants.ASYNC_HANDLE: response.result_dm.async_handle}
        )
        async_handler_fm = _async_handler.CloneVolumeAsyncHandler(self, snapshot_dm.volume_id,
                                                                  async_handle_dm=async_handle_dm)
        return async_handler_fm

    @staticmethod
    def _get_qos_data(min_iops=None, max_iops=None, burst_iops=None, burst_time=None):
        """Return qos data based on primitives.

        If all primitives are None, return None.

        :param int min_iops: minIOPs setting for created volume.
        :param int max_iops: maxIOPs setting for created volume.
        :param int burst_iops: burstIOPs setting for created volume.
        :param int burst_time: burstTime setting for created volume.
        :rtype: dict|None
        """
        if any(e is not None for e in (min_iops, max_iops, burst_iops, burst_time)):
            qos_dm = models.Qos.create(min_iops=min_iops,
                                       max_iops=max_iops,
                                       burst_iops=burst_iops,
                                       burst_time=burst_time)
            qos = {k: v for k, v in qos_dm.serialize().items() if v is not None}
        else:
            qos = None
        return qos

    def create_multiple_volumes(self, number, name_prefix, account_id, total_size,
                                enable512e=None, min_iops=None, max_iops=None, burst_iops=None,
                                burst_time=None, qos_policy_id=None, associate_with_qos_policy=None,
                                enable_snapmirror_replication=None, attributes=None,
                                skip_fullness_check=None, access=None):
        """Create n number of identical volumes on the cluster via CreateMultipleVolumes.

        Each volume is based on the arguments provided. volume_name will be used as a prefix for the
        resulting volumes.
        CreateMultipleVolumes returns a dict with the keys 'volumes' and 'volumeIDs' but cluster_api
        returns a list of ints.

        :param int number: Number of volumes to create
        :param str name_prefix: Prefix to use for volume names
        :param int account_id: AccountID volume will belong to.
        :param valence.models.DataSize total_size: Size of volume.
        :param bool enable512e: If true enable 512 byte emulation.
        :param int min_iops: minIOPs setting for created volume.
        :param int max_iops: maxIOPs setting for created volume.
        :param int burst_iops: burstIOPs setting for created volume.
        :param int burst_time: burstTime setting for created volume.
        :param int qos_policy_id: QoSPolicyID from which to define QoS.
        :param bool associate_with_qos_policy: Associates the volume to the specified QoS policy
            when true. Default value is True.
        :param bool enable_snapmirror_replication: Boolean to enable snapmirror replication
        :param dict attributes: Any additional key value pairs to be added to the Account.
        :param bool skip_fullness_check: Whether to skip the cluster fullness check preventing volumes
            from being created in stage 4.
        :param valence.enums.VolumeAccess access: Access allowed for the volume.

        :return: List of volumes as returned by API
        :rtype: list[valence.models.Volume]
        :raises valence.functional_models.FunctionalClusterError: if the created volume is not
                found when calling list_volumes.
        """
        # NOTE: the slice_count argument was intentionally omitted because the system limits it to
        # a value of 1 and it is made obsolete by erasure coding.
        volume_ids = []
        qos = self._get_qos_data(min_iops=min_iops,
                                 max_iops=max_iops,
                                 burst_iops=burst_iops,
                                 burst_time=burst_time)
        qos = qos if qos else self.get_default_qos().serialize()
        # Use name_prefix as prefix for resulting volume names. Zero-pad 3 digits for sorting.
        names = ['{}{}'.format(name_prefix, str(i).rjust(4, '0')) for i in range(number)]
        constants_dm = self.get_constants()
        # This splits up the volume_ids into chunks of size that the API can handle at once
        max_requests = constants_dm.max_volumes_per_multi_api or 500
        volume_name_chunks = [names[x:x + max_requests] for x in range(0, len(names), max_requests)]
        for volume_name_chunk in volume_name_chunks:
            params = {
                api_constants.NAMES: volume_name_chunk,
                api_constants.ACCOUNT_ID: account_id,
                api_constants.TOTAL_SIZE: total_size.bytes,
                api_constants.ENABLE_512E: enable512e,
                api_constants.ATTRIBUTES: attributes,
                api_constants.QOS: qos,
                api_constants.ENABLE_SNAPMIRROR_REPLICATION: enable_snapmirror_replication,
                api_constants.QOS_POLICY_ID: qos_policy_id,
                api_constants.ASSOCIATE_WITH_QOS_POLICY: associate_with_qos_policy,
                api_constants.SKIP_FULLNESS_CHECK: skip_fullness_check,
                api_constants.VOLUME_ACCESS: access.value if access else None
            }
            response = self.call_method(enums.ClusterMethod.create_multiple_volumes, params=params,
                                        result_dm_class=result_models.CreateMultipleVolumes)
            if response.result_dm.volume_ids:
                volume_ids.extend(response.result_dm.volume_ids)
            else:
                raise _errors.FunctionalClusterError(
                    'CreateMultipleVolumes response is missing volumeIDs'
                )
        if volume_ids:
            volume_dms = self.list_volumes(volume_ids=volume_ids)
            if len(volume_dms) != number:
                msg = 'Expected {} volumes, found {}'.format(number, len(volume_dms))
                raise _errors.FunctionalClusterError(msg)
            return volume_dms

    def create_volume(self, name, account_id, total_size, enable512e=None, min_iops=None,
                      max_iops=None, burst_iops=None, burst_time=None, qos_policy_id=None,
                      associate_with_qos_policy=None, enable_snapmirror_replication=None,
                      attributes=None, access=None):
        """Create a volume on a cluster via CreateVolume. New volume data model will be returned.

        :param bool associate_with_qos_policy: Associates the volume to the specified QoS policy
            when true. Default value is True.
        :param str name: Name of volume.
        :param int account_id: AccountID volume will belong to.
        :param valence.models.DataSize total_size: Size of volume.
        :param bool enable512e: If true enable 512 byte emulation.
        :param int min_iops: minIOPs setting for created volume.
        :param int max_iops: maxIOPs setting for created volume.
        :param int burst_iops: burstIOPs setting for created volume.
        :param int burst_time: burstTime setting for created volume.
        :param int qos_policy_id: QoSPolicyID from which to define QoS.
        :param bool associate_with_qos_policy: Associates the volume to the specified QoS policy
            when true. Default value is True.
        :param bool enable_snapmirror_replication: Disabled by default.
        :param dict attributes: Any additional key value pairs to be added to the Volume.
        :param valence.enums.VolumeAccess access: Access allowed for the volume.

        :return: Volume as returned by API
        :rtype: valence.models.Volume
        :raises valence.functional_models.FunctionalClusterError: if the created volume is not
                found when calling list_volumes.
        """
        # NOTE: the slice_count argument was intentionally omitted because the system limits it to
        # a value of 1 and it is made obsolete by erasure coding.
        qos = self._get_qos_data(min_iops=min_iops,
                                 max_iops=max_iops,
                                 burst_iops=burst_iops,
                                 burst_time=burst_time)
        params = {
            api_constants.NAME: name,
            api_constants.ACCOUNT_ID: account_id,
            api_constants.TOTAL_SIZE: total_size.bytes,
            api_constants.ENABLE_512E: enable512e,
            api_constants.ATTRIBUTES: attributes,
            api_constants.QOS: qos,
            api_constants.ENABLE_SNAPMIRROR_REPLICATION: enable_snapmirror_replication,
            api_constants.QOS_POLICY_ID: qos_policy_id,
            api_constants.ASSOCIATE_WITH_QOS_POLICY: associate_with_qos_policy,
            api_constants.VOLUME_ACCESS: access.value if access else None
        }

        response = self.call_method(enums.ClusterMethod.create_volume, params=params,
                                    result_dm_class=result_models.CreateVolume)

        if response.result_dm.volume:
            # The volume object was added to the response in fluorine
            return response.result_dm.volume
        elif response.result_dm.volume_id is not None:
            volume_id = response.result_dm.volume_id
            return self.list_volumes(volume_ids=[volume_id])[0]
        else:
            raise _errors.FunctionalClusterError('CreateVolume response is missing volumeID')

    def delete_volume(self, volume_dm):
        """Call DeleteVolume.

        :param valence.models.Volume volume_dm: Volume to be deleted
        :rtype: valence.models.Volume
        """
        response = self.call_method(enums.ClusterMethod.delete_volume,
                                    params={api_constants.VOLUME_ID: volume_dm.volume_id},
                                    result_dm_class=result_models.DeleteVolume)
        return response.result_dm.volume

    def delete_volumes(self, volume_dms):
        """Delete provided volumes on a cluster.

        Uses DeleteVolumes in chunks of cMaxVolumesPerMultiAPI if available. Volumes that back
        VirtualVolumes are automatically purged. Normal volumes will enter the deleted state.

        :param typing.Iterable[valence.models.Volume] volume_dms: Volumes to be deleted
        :return: Models of volumes that have been deleted
        :rtype: list[valence.models.Volume]
        """
        return_volumes = []
        volume_ids = [volume_dm.volume_id for volume_dm in volume_dms]
        version_string = self.cluster_dm.api_endpoint.version_string
        has_plural_delete_volumes = 'DeleteVolumes' in self.api.get_api().result.get(version_string)
        if has_plural_delete_volumes:
            constants_dm = self.get_constants()
            # This splits up the volume_ids into chunks of size that the API can handle at once
            MAX_REQUESTS = constants_dm.max_volumes_per_multi_api
            volume_id_chunks = [volume_ids[x:x + MAX_REQUESTS]
                                for x in range(0, len(volume_ids), MAX_REQUESTS)]
            for volume_id_chunk in volume_id_chunks:
                params = {api_constants.VOLUME_IDS: volume_id_chunk}
                response = self.call_method(enums.ClusterMethod.delete_volumes, params=params,
                                            result_dm_class=result_models.DeleteVolumes)

                return_volumes.extend(response.result_dm.volumes)
        else:
            for volume in volume_dms:
                return_volumes.append(self.delete_volume(volume))
        return return_volumes

    def get_default_qos(self):
        """Get Default QOS settings for volumes on a cluster via GetDefaultQoS.

        If GetDefaultQoS is not available, returns a data model with the following settings::

            {
                'minIOPS': 100,
                'maxIOPS': 15000,
                'burstIOPS': 15000,
                'burstTime': 60,
                'curve': {
                    '4096': 100,
                    '8192': 160,
                    '16384': 270,
                    '32768': 500,
                    '65536': 1000,
                    '131072': 1900,
                    '262144': 3000,
                    '524288': 7500,
                    '1048576': 15000
                }
            }

        :returns: QoS data model
        :rtype: valence.models.Qos
        """
        if self.api_supported(enums.ClusterMethod.get_default_qos):
            response = self.call_method(enums.ClusterMethod.get_default_qos,
                                        result_dm_class=result_models.GetDefaultQoS)
            qos = response.result_dm.serialize()
        else:
            qos = {
                api_constants.MIN_IOPS: 100,
                api_constants.MAX_IOPS: 15000,
                api_constants.BURST_IOPS: 15000,
                api_constants.BURST_TIME: 60,
                api_constants.CURVE: {
                    '4096': 100,
                    '8192': 160,
                    '16384': 270,
                    '32768': 500,
                    '65536': 1000,
                    '131072': 1900,
                    '262144': 3000,
                    '524288': 7500,
                    '1048576': 15000
                }
            }
        return models.Qos(qos)

    def get_volume_efficiency(self, volume_dm):
        """Returns the VolumeEfficiency for a given volume.

        :param valence.models.Volume volume_dm: Volume data model.
        :rtype: sf_api.result_models.GetVolumeEfficiency
        """
        params = {api_constants.VOLUME_ID: volume_dm.volume_id}
        return self.call_method(enums.ClusterMethod.get_volume_efficiency, params=params,
                                result_dm_class=result_models.GetVolumeEfficiency).result_dm

    def list_volume_qos_histograms(self, volume_ids=None):
        """List QoS histograms for each volume.

        ListVolumeQoSHistograms returns histograms detailing volume performance relative
        to QOS settings. It may take up to 5 seconds for newly created volumes to have accurate
        histogram data available.

        :param list[int] volume_ids: List of volumes to return data for. If no volumes are
            specified then information for all volumes will be returned.
        :return: List of VolumeQosHistograms data models
        :rtype: list[valence.models.VolumeQosHistograms]
        """
        params = None if volume_ids is None else {
            api_constants.VOLUME_IDS: volume_ids
        }

        response = self.call_method(enums.ClusterMethod.list_volume_qos_histograms,
                                    params=params,
                                    result_dm_class=result_models.ListVolumeQoSHistograms)
        return response.result_dm.qos_histograms

    def list_volumes(self, account_ids=None, include_virtual_volumes=None, is_paired=None,
                     volume_ids=None, volume_name=None, volume_status=None):
        """List volumes on a cluster via ListVolumes.

        The volume_ids and account_ids parameters are mutually exclusive and cannot be used together.

        :param list[int] account_ids: Return volumes associated with these account_ids.
                                      Mutually exclusive with volumeIDs.
        :param bool include_virtual_volumes: Virtual volumes included by default, exclude with False.
                                             Returns volumes that back virtual volumes, not virtual
                                             volumes themselves.
        :param bool is_paired: True-Return paired volumes;False-Return unpaired volumes;
                               None-Return both paired and unpaired.
        :param list[int] volume_ids: Return volumes with these IDs. Mutually exclusive with account_dms.
        :param str volume_name: Return volumes with exactly this name.
        :param sf_api.VolumeStatus volume_status: Return volumes with this status.
        :return: List of volume data models
        :rtype: list[valence.models.Volume]
        :raises ValueError: When mutually exclusive parameters are provided.
        """
        # Most of the filtering options were added in Fluorine
        if self.cluster_dm.api_endpoint >= sf_api.SfVersion('9.0'):
            params = {
                api_constants.ACCOUNTS: account_ids,
                api_constants.INCLUDE_VIRTUAL_VOLUMES: include_virtual_volumes,
                api_constants.IS_PAIRED: is_paired,
                api_constants.VOLUME_IDS: volume_ids,
                api_constants.VOLUME_NAME: volume_name,
                api_constants.VOLUME_STATUS: volume_status if volume_status else None
            }

            response = self.call_method(enums.ClusterMethod.list_volumes, params=params,
                                        result_dm_class=result_models.ListVolumes)
            volumes = response.result_dm.volumes
        else:
            if account_ids is not None and volume_ids is not None:
                raise ValueError('account_dms and volume_ids are mutually exclusive')
            # If we only need to get one volume, use limit and startVolumeID
            # This seems to be a PyCharm bug, volume_ids is properly defined.
            # noinspection PyTypeChecker
            if volume_ids and len(volume_ids) == 1:
                params = {
                    api_constants.LIMIT: 1,
                    api_constants.START_VOLUME_ID: volume_ids[0],
                }
                response = self.call_method(enums.ClusterMethod.list_volumes, params=params,
                                            result_dm_class=result_models.ListVolumes)
                volumes = response.result_dm.volumes
            else:
                # Put volume_ids and account_ids into a set to improve efficiency
                volume_ids = set(volume_ids) if volume_ids else None
                account_ids = set(account_ids) if account_ids else None
                response = self.call_method(enums.ClusterMethod.list_volumes,
                                            result_dm_class=result_models.ListVolumes)
                all_volumes = response.result_dm.volumes

                volume_filter = criteria.And(
                    criteria.Equals(models.Volume.name, volume_name),
                    criteria.ContainedIn(models.Volume.account_id, account_ids),
                    criteria.ContainedIn(models.Volume.volume_id, volume_ids),
                    criteria.Equals(models.Volume.status, volume_status if volume_status else None)
                )
                if include_virtual_volumes is False:
                    # Virtual volume ID should be undefined or None
                    volume_filter.add_criteria(
                        criteria.EvaluatesToTrue(models.Volume.virtual_volume_id, negate=True)
                    )
                if is_paired is not None:
                    if is_paired:
                        # Select volumes where volume_pairs is defined and not an empty list
                        paired_volumes_filter = criteria.EvaluatesToTrue(models.Volume.volume_pairs)
                        volume_filter.add_criteria(paired_volumes_filter)
                    else:
                        # Select volumes where volume_pairs is undefined or an empty list
                        unpaired_volumes_filter = criteria.EvaluatesToTrue(
                            models.Volume.volume_pairs,
                            negate=True)
                        volume_filter.add_criteria(unpaired_volumes_filter)
                volumes = volume_filter.filter(all_volumes)
        return volumes

    def list_volume_stats_by_volume(self):
        """Method ListVolumeStatsByVolume returns high-level activity measurements for every volume.

        :rtype: list[valence.models.VolumeStats]
        """
        response = self.call_method(enums.ClusterMethod.list_volume_stats_by_volume,
                                    result_dm_class=result_models.ListVolumeStatsByVolume)

        return response.result_dm.volume_stats

    def purge_all_deleted_volumes(self):
        """Purge all deleted volumes from the cluster via PurgeAllDeletedVolumes.

        This is an internal API request.

        :raises valence.functional_models.FunctionalClusterError: If a volume could not be purged
        """
        self.call_method(enums.ClusterMethod.purge_all_deleted_volumes,
                         result_dm_class=result_models.PurgeAllDeletedVolumes)

    def purge_deleted_volumes(self, volume_dms):
        """Purge volumes that have been deleted via PurgeDeleteVolume and PurgeDeletedVolumes.

        :param list[valence.models.Volume] volume_dms: Volumes to be purged
        """
        volume_ids = [volume_dm.volume_id for volume_dm in volume_dms]
        version_string = self.cluster_dm.api_endpoint.version_string
        has_plural_delete_volumes = ('PurgeDeletedVolumes' in
                                     self.api.get_api().result.get(version_string))
        if has_plural_delete_volumes:
            constants_dm = self.get_constants()
            # This splits up the volume_ids into chunks of size that the API can handle at once
            MAX_REQUESTS = constants_dm.max_volumes_per_multi_api
            volume_id_chunks = [volume_ids[x:x + MAX_REQUESTS]
                                for x in range(0, len(volume_ids), MAX_REQUESTS)]
            for volume_id_chunk in volume_id_chunks:
                if has_plural_delete_volumes:
                    self.call_method(enums.ClusterMethod.purge_deleted_volumes,
                                     params={api_constants.VOLUME_IDS: volume_id_chunk},
                                     result_dm_class=None)
        else:
            for volume_id in volume_ids:
                self.call_method(enums.ClusterMethod.purge_deleted_volume,
                                 params={api_constants.VOLUME_ID: volume_id},
                                 result_dm_class=None)

    def read_slice_lbas(self, slice_id, lba, n_blocks, snapshot_id=None):
        """Read an array of block ids for a range of LBAs.

        :param int slice_id: Slice ID (this is usually the same as the volume ID)
        :param int lba: The first LBA in the range
        :param int n_blocks: The number of LBAs to get blockIDs for
        :param int snapshot_id: the snapshot to read from (default is active branch)
        :rtype: dict
        :returns: A dictionary of slice service IDs mapped to lists of block IDs
        """
        params = {
            api_constants.SLICE_ID: slice_id,
            api_constants.LBA: lba,
            api_constants.N_BLOCKS: n_blocks,
            api_constants.SNAPSHOT_ID: snapshot_id
        }
        response = self.call_method(enums.ClusterMethod.read_slice_lba, params=params,
                                    result_dm_class=result_models.ReadSliceLBA)
        return response.result_dm.block_ids

    def read_slice_lba_checksum(self, slice_id, lba, n_blocks, snapshot_id=None):
        """Returns a map of slice service IDs and checksums of the block ID for an LBA.

        :param int slice_id: Slice ID (this is usually the same as the volume ID)
        :param int lba: The first LBA in the range
        :param int n_blocks: The number of LBAs to get blockIDs for
        :param int snapshot_id: the snapshot to read from (default is active branch)
        :rtype: dict
        :returns: A dictionary of slice service IDs mapped to lists of block IDs
        """
        params = {
            api_constants.SLICE_ID: slice_id,
            api_constants.LBA: lba,
            api_constants.N_BLOCKS: n_blocks,
            api_constants.SNAPSHOT_ID: snapshot_id
        }
        result = self.call_method(enums.ClusterMethod.read_slice_lba_checksum, params=params,
                                  result_dm_class=result_models.ReadSliceLBAChecksum)
        return result.result_dm.block_checksums

    def modify_volumes(self, volume_dms, access=None, attributes=None, min_iops=None, max_iops=None,
                       burst_iops=None, qos_policy_id=None, associate_with_qos_policy=None,
                       account_id=None, total_size=None, enable_snapmirror_replication=None):
        """Modify settings on existing volumes.

        :param list[valence.models.Volume] volume_dms: Volume data models for the volumes to be modified
        :param valence.enums.VolumeAccess access: Access allowed for the volume.
            If not specified, the access value will not be changed.
        :param dict attributes: List of Name / Value pairs in JSON object format
        :param int min_iops: minIOPs setting for created volume.
        :param int max_iops: maxIOPs setting for created volume.
        :param int burst_iops: burstIOPs setting for created volume.
        :param int qos_policy_id: ID of a QOS policy object
            * Used to set QOS (qos must be None)
        :param bool associate_with_qos_policy: Whether to associate the volume with the QoS Policy.
            * Changes to the QoS Policy affect the volume's QoS settings.
        :param int account_id: AccountID to which the volume will be reassigned.
        :param int total_size: New size of volume in bytes
        :param bool enable_snapmirror_replication: Flag that controls whether the volumes can be used for
            replication with SnapMirror endpoints
        :return: Modified volume data models
        :rtype: list[valence.models.Volume]
        """
        qos = self._get_qos_data(min_iops=min_iops, max_iops=max_iops, burst_iops=burst_iops)
        params = {
            api_constants.VOLUME_IDS: [v.volume_id for v in volume_dms],
            api_constants.QOS: qos,
            api_constants.QOS_POLICY_ID: qos_policy_id,
            api_constants.ASSOCIATE_WITH_QOS_POLICY: associate_with_qos_policy,
            api_constants.VOLUME_ACCESS: access,
            api_constants.ATTRIBUTES: attributes,
            api_constants.ACCOUNT_ID: account_id,
            api_constants.TOTAL_SIZE: total_size,
            api_constants.ENABLE_SNAPMIRROR_REPLICATION: enable_snapmirror_replication
        }
        return self.call_method(enums.ClusterMethod.modify_volumes, params=params,
                                result_dm_class=result_models.ModifyVolumes).result_dm.volumes

    def get_slice_info(self, slice_id):
        """Returns the SliceInfo for a given slice_id.

        :param int slice_id: Slice ID (this is usually the same as the volume ID)
        :rtype: sf_api.result_models.GetSliceInfo
        :returns: SliceInfo data model
        """
        params = {api_constants.SLICE_ID: slice_id}
        return self.call_method(enums.ClusterMethod.get_slice_info, params=params,
                                result_dm_class=result_models.GetSliceInfo).result_dm

    def fill_slice(self, slice_id, num_blocks=None, seed=None, bs_generate=True, start_lba=0,
                   comp_ratio=1.0, comp_algorithm=enums.CompressionAlgorithm.lzf):
        """Fill the specified volume with randomly generated blockIDs.

        The data is the blockID repeated 256 times.

        :param bool bs_generate: If False, only generate metadata. All blockIDs will be
                                xUnknownBlockIDs.
        :param int slice_id: sliceID of the volume you want to fill
        :param int num_blocks: Number of blocks to fill the volume with. Defaults to size of volume.
        :param int seed: Seed for the random number generator, defaults to random int
        :param int start_lba: LBA to start filling at.
        :param float comp_ratio: Compression ratio for the fill.
        :param valence.enums.CompressionAlgorithm comp_algorithm: Compression algorithm
        :rtype: valence.functional_models.FillSliceAsyncHandler
        """
        params = {'bsGenerate': bs_generate,
                  'compRatio': comp_ratio,
                  'compAlgorithm': comp_algorithm.value,
                  'password': 'DataDestroyer',
                  'sliceID': slice_id,
                  'startLBA': start_lba,
                  'nBlocks': num_blocks,
                  'seed': seed}
        response = self.call_method(enums.ClusterMethod.fill_slice, params=params,
                                    result_dm_class=result_models.FillSlice)

        async_handle_dm = models.AsyncHandle(
            {api_constants.ASYNC_HANDLE: response.result_dm.async_handle}
        )

        async_handler_fm = _async_handler.FillSliceAsyncHandler(self, slice_id,
                                                                async_handle_dm=async_handle_dm)

        return async_handler_fm

    def reassign_slice(self, slice_id, primary=None, secondaries=None, retries=1,
                       action=enums.SliceAction.desired):
        """Reassign a slice.

        :param int slice_id: Slice ID to modify
        :param int primary: The new desired primary service ID
        :param list[int] secondaries: List of desired secondary service IDs
        :param int retries: number of times to retry on xDBVersionMismatch
        :param valence.enums.SliceAction action: Action for moving the slice.
        :returns: WaitForSliceReassignment handler
        :rtype: valence.functional_models.SliceReassignmentAsyncHandler
        """
        attempts = 0
        while attempts <= retries:
            try:
                db_version = self.get_slice_info(slice_id=slice_id).db_version
                _log.debug('Trying to reassign slice with db_version={}'.format(db_version))
                self._reassign_slice(slice_id=slice_id, db_version=db_version,
                                     action=action, primary=primary,
                                     secondaries=secondaries)
                break
            except sf_api.api_errors.DBVersionMismatch:
                _log.debug('DB version mismatch exception caught')
                if attempts == retries:
                    raise
            attempts += 1
        async_handler_fm = _async_handler.SliceReassignmentAsyncHandler(
            self, slice_id=slice_id, primary=primary, secondaries=secondaries)
        return async_handler_fm

    def _reassign_slice(self, slice_id, db_version, action, primary=None, secondaries=None, prune=None):
        """Call ReassignSlice method.

        :param int slice_id: The SliceID to modify
        :param int db_version: The DB version to modify (fetch the current from get_slice_info)
        :param valence.enums.SliceAction action: The action to perform enum
        :param int primary: If action is 'Required' or 'Desired' reassignment, the new desired primary
        :param list[int] secondaries: If action is 'Required' or 'Desired' reassignment, the new desired
                                      secondaries
        :param list[int] prune: If action is 'Prune', the services to prune
        :rtype: sf_api.result_models.ReassignSlice
        :returns: SliceInfo data model
        """
        params = {
            api_constants.SLICE_ID: slice_id,
            api_constants.DB_VERSION: db_version,
            api_constants.ACTION: action.value,
            api_constants.PRIMARY: primary,
            api_constants.SECONDARIES: secondaries or [],
            api_constants.PRUNE: prune
        }
        return self.call_method(enums.ClusterMethod.reassign_slice, params=params,
                                result_dm_class=result_models.ReassignSlice)

    def start_volume_pairing(self, volume_dm, mode=None):
        """Get a volume pairing key for pairing to a volume on a remote cluster.

        :param valence.models.Volume volume_dm: Volume data model.
        :param valence.enums.RemoteRepMode mode: Remote replication mode for source volume.
        :rtype: str
        :returns: Volume pairing key
        """
        params = {
            api_constants.MODE: mode,
            api_constants.VOLUME_ID: volume_dm.volume_id
        }
        response = self.call_method(enums.ClusterMethod.start_volume_pairing, params=params,
                                    result_dm_class=result_models.StartVolumePairing)
        return response.result_dm.volume_pairing_key

    def complete_volume_pairing(self, volume_dm, volume_pairing_key):
        """Complete volume pairing.

        :param valence.models.Volume volume_dm: Volume data model.
        :param str volume_pairing_key: Volume pairing key returned from start_volume_pairing
        """
        params = {
            api_constants.VOLUME_ID: volume_dm.volume_id,
            api_constants.VOLUME_PAIRING_KEY: volume_pairing_key
        }
        self.call_method(enums.ClusterMethod.complete_volume_pairing, params=params,
                         result_dm_class=None)

    def list_active_paired_volumes(self):
        """List active paired volumes.

        :return: List of active paired volumes.
        :rtype: list[valence.models.Volume]
        """
        response = self.call_method(enums.ClusterMethod.list_active_paired_volumes,
                                    result_dm_class=result_models.ListActivePairedVolumes)
        return response.result_dm.volumes

    def modify_volume_pair(self, volume_dm, paused_manual=None, mode=None):
        """Modify volume pair.

        :param valence.models.Volume volume_dm: Volume data model.
        :param bool paused_manual: Manually pause remote replication.
        :param valence.enums.RemoteRepMode mode: Remote replication mode.
        """
        params = {api_constants.VOLUME_ID: volume_dm.volume_id,
                  api_constants.PAUSED_MANUAL: paused_manual,
                  api_constants.MODE: mode}
        self.call_method(enums.ClusterMethod.modify_volume_pair,
                         params=params,
                         result_dm_class=None)

    def remove_volume_pair(self, volume_dm):
        """Remove volume pair.

        :param valence.models.Volume volume_dm: Volume data model.
        """
        params = {api_constants.VOLUME_ID: volume_dm.volume_id}
        self.call_method(enums.ClusterMethod.remove_volume_pair,
                         params=params,
                         result_dm_class=None)

    def restore_deleted_volume(self, volume_dm):
        """Restore a deleted volume.

        :param valence.models.Volume volume_dm: Volume to be restored
        """
        self.call_method(enums.ClusterMethod.restore_deleted_volume,
                         params={api_constants.VOLUME_ID: volume_dm.volume_id},
                         result_dm_class=None)

    def rebalance_slices(self):
        """Rebalance Slices on cluster."""
        self.call_method(enums.ClusterMethod.rebalance_slices,
                         params={},
                         result_dm_class=None)

    def compute_volume_checksum(self, volume_dm, snapshot_dm=None):
        """Calculate slice/snapshot checksums across the primary and live secondaries.

        :param valence.models.Volume volume_dm: volume dm to checksum
        :param valence.models.Snapshot snapshot_dm: optional snapshot dm on that slice
          (default = active branch)
        :rtype: dict
        :returns: result dict maps service ID to checksum value for the SSs hosting the slice
        """

        params = {api_constants.VOLUME_ID: volume_dm.volume_id,
                  api_constants.SNAPSHOT_ID: snapshot_dm.snapshot_id if snapshot_dm else 0}
        response = self.call_method(enums.ClusterMethod.compute_volume_checksum, params=params,
                                    result_dm_class=result_models.ComputeVolumeChecksum)
        return response.result_dm.checksums

    def move_primaries_away_from_node(self, node_id):
        """Move primary slice copies off of given node.

        :param int node_id: ID of node to transition slice copies away from
        :rtype: sf_api.result_models.MovePrimariesAwayFromNode
        """

        response = self.call_method(enums.ClusterMethod.move_primaries_away_from_node,
                                    params={api_constants.NODE_ID: node_id},
                                    result_dm_class=result_models.MovePrimariesAwayFromNode)
        return response.result_dm
