"""Base class for external key server functional model."""

import abc
import os
import ssl

import autologger
import sf_models
import sf_ssh
import six
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import (
    hashes,
    serialization,
)
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509 import oid
from kmip.pie import client

from valence.functional_models import _errors
from valence.utils import random_string

_log = autologger.get_sflogging()


@six.add_metaclass(abc.ABCMeta)
class ExternalKeyServerBase(object):
    """EKS functional model base class.

    This abstract class defines methods that are required by the EKS functional models for each
    vendor type. The abstract methods defined in the base class must be implemented by the vendor
    subclasses using their respective APIs.

    KMIP is a standard protocol and thus all KMIP related methods should be standard across all
    vendor types. Management operations are vendor specific and must be implemented by the vendor
    subclasses.

    :param sf_models.ExternalKeyServer eks_dm: external key server data model
    """
    def __init__(self, eks_dm):
        self.eks_dm = eks_dm
        self.ca_certificate = None
        self.pykmip_client_certificate = None
        self.pykmip_client_private_key = None
        self.pykmip_client_csr = None
        self.pykmip_username = None
        self.pykmip_password = None

        self._unique_id = random_string.generate_random_string(8)
        self._pykmip_client = None
        self._pykmip_connected = False

    def __del__(self):
        """Delete any pykmip certs that were created."""
        if self._pykmip_connected:
            self._cleanup_pykmip_certificates()

    @property
    def hostname(self):
        """Pass through to hostname in data model.

        :rtype: str
        """
        return self.eks_dm.hostname

    @property
    def port(self):
        """Pass through to port in data model.

        :rtype: int
        """
        return self.eks_dm.port

    @property
    def username(self):
        """Pass through to username in data model.

        :rtype: str
        """
        return self.eks_dm.username

    @property
    def password(self):
        """Pass through to password in data model.

        :rtype: str
        """
        return self.eks_dm.password

    @property
    def vendor(self):
        """Pass through to vendor in data model.

        :rtype: str
        """
        return self.eks_dm.vendor

    @property
    def kmip_version(self):
        """Pass through to kmip_version in data model.

        :rtype: sf_models.enums.KmipVersion
        """
        return self.eks_dm.kmip_version

    @property
    def ssl_version(self):
        """Pass through to ssl_version in data model.

        :rtype: sf_models.enums.SslVersion
        """
        return self.eks_dm.ssl_version

    @abc.abstractmethod
    def sign_csr(self, csr, start_date=None, end_date=None, force=False):
        """Take the given CSR and get it signed by a Trusted CA (usually local to the EKS).

        :param cryptography.x509.CertificateSigningRequest csr: x509 object for the CSR
        :param datetime.datetime start_date: Datetime object representing the certificate start date
        :param datetime.datetime end_date: Datetime object representing the certificate expiration date
        :param bool force: If True, react to failures and force the operation.
        :rtype: tuple[cryptography.x509.Certificate]
        """

    @abc.abstractmethod
    def enable(self):
        """Use to restore the connection between the SF Cluster and EKS."""

    @abc.abstractmethod
    def disable(self):
        """Use to sever the connection between the SF Cluster and EKS."""

    @abc.abstractmethod
    def configure_for_testing(self, kmip_version=None):
        """Confirm EKS is ready for test execution.

        Some Vendors should perform specific checks such as confirming an EKS account exists to use for
        the Organizational Unit, or that a local CA exists.

        :param sf_models.enums.KmipVersion kmip_version: If possible, configure EKS to this
        """

    @abc.abstractmethod
    def is_running(self):
        """Return True if the KMIP service or EKS is up and running.

        :rtype: bool
        """

    def get_kmip_connection(self):
        """Convenience method for returning the KMIP client connection.

        This method will establish trust relationship between localhost and EKS if not
        already established.

        Usage:
            with self.get_kmip_connection() as c:
                key_id = c.register(key)
                c.activate(key_id)

        :rtype: kmip.pie.client.ProxyKmipClient
        """
        if not self._pykmip_connected:
            self.connect()
        return self._pykmip_client

    def get_ssh_connection(self):
        """Convenience method for returning the SSH connection.

        Usage:
            with self.get_ssh_connection() as ssh_conn:
                output = ssh_conn.run_command("ls /")

        :rtype: sf_ssh.Ssh
        """
        return sf_ssh.Ssh(self.hostname, username=self.username, password=self.password)

    def run_command(self, command, timeout=300, raise_on_failure=False, use_sudo=False):
        """Executes a shell command.

        :param str command: shell command to execute
        :param int timeout: Timeout before Command returns output
        :param bool raise_on_failure: If True, raise exception if return code not 0. If False, do not
               raise exception, just return tuple.
        :param bool use_sudo: If True, run command with sudo.
        :raises errors.ExternalKeyServerError: Raise if command on client did not complete successfully
        :rtype: sf_ssh.ssh.RunCommandOutput
        """

        with self.get_ssh_connection() as ssh:
            result = ssh.run_command(command, timeout, use_sudo=use_sudo)
            if result.return_code != 0 and raise_on_failure:
                msg = ('Failed to run command [{}] on EKS {}. Failed with rc={}: {}'
                       .format(command, self.hostname, result.return_code, result.stderr))
                raise _errors.ExternalKeyServerError(msg)
            return result

    def run_multiple_commands(self, commands, timeout=300, raise_on_failure=False, use_sudo=False):
        """Executes a list of shell commands in a single ssh session.

        Does not retain result of each command. Raises if a single command fails.

        :param list[str] commands: shell commands to execute
        :param int timeout: Timeout before Command returns output
        :param bool raise_on_failure: If True, raise exception if return code not 0. If False, do not
               raise exception, just return tuple.
        :param bool use_sudo: If True, run command with sudo.
        :raises errors.FunctionalClientError: Raise if command on client did not complete successfully
        """
        with self.get_ssh_connection() as ssh:
            for cmd in commands:
                result = ssh.run_command(cmd, timeout, use_sudo=use_sudo)
                if result.return_code != 0 and raise_on_failure:
                    msg = ('Failed to run command [{}] on EKS {}. Failed with rc={}: {}'
                           .format(cmd, self.hostname, result.return_code, result.stderr))
                    raise _errors.ExternalKeyServerError(msg)

    def connect(self):
        """Create the required certificates and instantiate the PyKMIP ProxyKmipClient object.

        :raises ssl.SSLError: Raise if ProxyKmipClient connection fails
        :rtype: kmip.pie.client.ProxyKmipClient
        """
        if not self.is_pykmip_trusted():
            self._setup_pykmip_certificates()

        try:
            self._pykmip_client = client.ProxyKmipClient(
                hostname=str(self.hostname),
                port=self.port,
                cert=self.pykmip_client_certificate,
                key=self.pykmip_client_private_key,
                ca=self.ca_certificate,
                ssl_version=self.ssl_version.value,
                username=self.pykmip_username,
                password=self.pykmip_password
            )

            # Test the connection
            with self._pykmip_client:
                self._pykmip_connected = True
        except ssl.SSLError as e:
            _log.error('Error connecting via KMIP to {}:{}'.format(self.hostname, self.port))
            raise e

        return self._pykmip_client

    def is_pykmip_trusted(self):
        """Returns True if the private key, client certificate, and CA certificate are created.

        :rtype: bool
        """
        if None in [self.pykmip_client_certificate, self.pykmip_client_private_key, self.ca_certificate]:
            return False
        else:
            return True

    def _cleanup_pykmip_certificates(self):
        """Delete pykmip certificate files that were created on localhost."""
        if self.ca_certificate and os.path.exists(self.ca_certificate):
            os.remove(self.ca_certificate)
        if self.pykmip_client_certificate and os.path.exists(self.pykmip_client_certificate):
            os.remove(self.pykmip_client_certificate)
        if self.pykmip_client_private_key and os.path.exists(self.pykmip_client_private_key):
            os.remove(self.pykmip_client_private_key)
        if self.pykmip_client_csr and os.path.exists(self.pykmip_client_csr):
            os.remove(self.pykmip_client_csr)

    def _setup_pykmip_certificates(self):
        """Private method for establishing the trust relationship between localhost and EKS.

        This method will generate a private key, create a CSR, and get the CSR signed by the CA on
        the EKS.
        """
        self.pykmip_client_private_key = self._create_pykmip_private_key()
        self.pykmip_client_csr = self._create_pykmip_csr()

        with open(self.pykmip_client_csr, 'rb') as fh:
            csr = x509.load_pem_x509_csr(fh.read(), default_backend())
        client_certificate, ca_certificate = self.sign_csr(csr)

        # Pykmip requires the key, client cert, and ca cert to be written to disk.
        # When this FM is destroyed, these files will be deleted.
        self.pykmip_client_certificate = '/tmp/client_{}.crt'.format(self._unique_id)
        self.ca_certificate = '/tmp/ca_{}.crt'.format(self._unique_id)
        with open(self.pykmip_client_certificate, 'w') as fh:
            fh.write(client_certificate.public_bytes(serialization.Encoding.PEM))
        with open(self.ca_certificate, 'w') as fh:
            fh.write(ca_certificate.public_bytes(serialization.Encoding.PEM))

    def _create_pykmip_private_key(self):
        """Use the cryptography library to generate an RSA private key.

        Writes the PEM formatted key to tmp and returns the path to this file.

        :rtype: str
        """
        key_path = '/tmp/client_{}.key'.format(self._unique_id)
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048,
                                               backend=default_backend())
        pem = private_key.private_bytes(encoding=serialization.Encoding.PEM,
                                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                                        encryption_algorithm=serialization.NoEncryption())
        with open(key_path, 'wb') as fh:
            fh.write(pem)

        return key_path

    def _create_pykmip_csr(self):
        """Use the cryptography library to create a certificate signing request.

        Writes the PEM formatted CSR to tmp and returns the path to this file.

        :rtype: str
        """
        csr_path = '/tmp/client_{}.csr'.format(self._unique_id)

        # Get the private key object
        with open(self.pykmip_client_private_key, 'rb') as fh:
            private_key = serialization.load_pem_private_key(fh.read(), password=None,
                                                             backend=default_backend())

        # Create a CSR and sign it
        csr = x509.CertificateSigningRequestBuilder().subject_name(x509.Name([
            x509.NameAttribute(oid.NameOID.COUNTRY_NAME, u"US"),
            x509.NameAttribute(oid.NameOID.STATE_OR_PROVINCE_NAME, u"Colorado"),
            x509.NameAttribute(oid.NameOID.LOCALITY_NAME, u"Boulder"),
            x509.NameAttribute(oid.NameOID.ORGANIZATIONAL_UNIT_NAME, u"SolidFire"),
            x509.NameAttribute(oid.NameOID.COMMON_NAME, u"{}_{}".format(os.uname()[1], self._unique_id)),
        ])).sign(private_key, hashes.SHA256(), default_backend())

        # Write the CSR to disk and return its path
        with open(csr_path, 'wb') as fh:
            fh.write(csr.public_bytes(serialization.Encoding.PEM))
        return csr_path

    @classmethod
    def get_mock_fm(cls, overrides=None):
        """Factory method for instantiating a mock EKS FM.

        :param dict overrides: Override mock DM values
        :rtype: valence.functional_models._external_key_server.ExternalKeyServerBase
        """
        rs_eks_dm = sf_models.ExternalKeyServer.get_mock_object(overrides=overrides)
        return cls(rs_eks_dm)
