"""Cryptsoft EKS functional model."""
import datetime
import re
import tempfile

import autologger
import sf_artifacts
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from valence.functional_models import _errors
from valence.functional_models._external_key_server import base

_log = autologger.get_sflogging()


class CryptsoftExternalKeyServer(base.ExternalKeyServerBase):
    """EKS functional model for Cryptsoft test server.

    This class will act as an interface to a Cryptsoft external key server. The server is a collection
    of binaries that can be deployed on a standard Linux client. Interaction with the Cryptsoft server
    will mostly be through  OpenSSL calls or executing some of the provided binaries. Test cases
    should utilize the abstract methods whenever possible so that they can be EKS agnostic.

    :param sf_models.ExternalKeyServer eks_dm: EKS data model
    """
    UBUNTU_SERVICE_FILES = 'http://sf-artifactory.solidfire.net/artifactory/generic-binary-local/' \
                           'kmipc_server-1.9.4.tar.gz'
    CENTOS_SERVICE_FILES = 'http://sf-artifactory.solidfire.net/artifactory/generic-binary-local/' \
                           'kmip_server-1.9.5a-centos.tar.gz'

    def __init__(self, eks_dm):
        super(CryptsoftExternalKeyServer, self).__init__(eks_dm)
        self.kmip_dir = '/tmp/kmip_server'
        self.openssl_config = self.kmip_dir + '/certs/ocsp/root/ca/openssl.cnf'
        self.kmip_bin = self.kmip_dir + '/bin'
        self.servers_config = self.kmip_bin + '/servers.cfg'

        # CA key, csr, and cert.
        self.root_key = self.kmip_dir + '/certs/ocsp/root/ca/private/ca.key.pem'
        self.root_csr = self.kmip_dir + '/certs/ocsp/root/ca/csr/ca.csr.pem'
        self.root_cert = self.kmip_dir + '/certs/ocsp/root/ca/certs/ca.cert.pem'

        # Server key, csr, and cert. Not really used beyond being there for the Cryptsoft service to see.
        self.server_key = self.kmip_dir + '/certs/ocsp/root/ca/private/server.key.pem'
        self.server_csr = self.kmip_dir + '/certs/ocsp/root/ca/csr/server.csr.pem'
        self.server_cert = self.kmip_dir + '/certs/ocsp/root/ca/certs/server.cert.pem'

        # Passed to Cryptsoft to sign and return. Will probably receive two sign_csr requests
        # 1 - Test runner CSR/signed CA
        # 2 - A cluster CSR/signed CA from Element's GetClientCertificateSignRequest
        self.user_csr_dir = self.kmip_dir + '/certs/ocsp/root/ca/csr/'
        self.user_cert_dir = self.kmip_dir + '/certs/ocsp/root/ca/certs/'
        self.signed_csrs = []
        self.signed_certs = []

        self.deployed = False
        self._os_release = None

    @property
    def os_release(self):
        """Return os type from os-release file.

        :rtype: str
        """
        if self._os_release is None:
            os_cmd = "awk -F= '/^NAME/{print $2}' /etc/os-release"
            self._os_release = self.run_command(os_cmd).stdout
        return self._os_release

    @property
    def is_ubuntu(self):
        """Return True if client is Ubuntu.

        :rtype: bool
        """
        return True if 'Ubuntu' in self.os_release else False

    @property
    def is_centos(self):
        """Return True if client is CentOS.

        :rtype: bool
        """
        return True if 'CentOS' in self.os_release else False

    def deploy_cryptsoft_server(self, ca_start_date=None, ca_end_date=None, server_start_date=None,
                                server_end_date=None):
        """Creates files, generates certificates, and initializes a client as a Cryptsoft server.

        Does not start the server, use power_on().

        :param datetime.datetime ca_start_date: Datetime object for the ca certificate start date
        :param datetime.datetime ca_end_date: Datetime object for the ca certificate expiration date
        :param datetime.datetime server_start_date: Datetime object for the ca certificate start date
        :param datetime.datetime server_end_date: Datetime object for the ca certificate expiration date
        """
        if self.is_running():
            _log.debug('Skipping deploy, kmip_server already running')
            self.deployed = True
            return

        # create the files and generate the certificates needed to run the Cryptsoft Service
        self._add_cryptsoft_service_files()
        self._generate_server_certs(ca_start_date, ca_end_date, server_start_date, server_end_date)

        # initialize the server
        init_commands = list()
        init_commands.append('{}/init'.format(self.kmip_bin))
        init_commands.append('{}/inituser "{}"'.format(self.kmip_bin, self.server_cert))
        self.run_multiple_commands(init_commands, raise_on_failure=True)
        self.deployed = True

    def teardown_cryptsoft_server(self):
        """Removes all the files created during deploy."""
        self.run_command('rm -rf /tmp/kmip_server*', raise_on_failure=True)
        self.deployed = False

    def _add_cryptsoft_service_files(self):
        """Creates the files on a client needed to run the Cryptsoft service.

        :raises: _errors.ExternalKeyServerError
        """
        if self.is_ubuntu:
            service_files = self.UBUNTU_SERVICE_FILES
        elif self.is_centos:
            service_files = self.CENTOS_SERVICE_FILES
        else:
            raise _errors.ExternalKeyServerError('Unsupported OS type')

        wget_cmd = 'wget {} -O /tmp/kmip_server.tar.gz;'.format(service_files)
        tar_cmd = 'tar -xzf /tmp/kmip_server.tar.gz -C /tmp'
        sed_cmd = 'sed -i "s|YOUR_PATH|{}/certs/ocsp|g" "{}"'.format(self.kmip_dir, self.openssl_config)
        sed_port_cmd = 'sed -i "s|LOCALHOST.port=[[:digit:]]\\+|LOCALHOST.port={}|g" "{}"'.format(
            self.port, self.servers_config)
        self.run_multiple_commands([wget_cmd, tar_cmd, sed_cmd, sed_port_cmd], raise_on_failure=True)

    def _generate_server_certs(self, ca_start_date=None, ca_end_date=None, server_start_date=None,
                               server_end_date=None):
        """Generate a self signed root CA cert, a server cert, and move them to the kmip_bin.

        :param datetime.datetime ca_start_date: Datetime object for the ca certificate start date
        :param datetime.datetime ca_end_date: Datetime object for the ca certificate expiration date
        :param datetime.datetime server_start_date: Datetime object for the ca certificate start date
        :param datetime.datetime server_end_date: Datetime object for the ca certificate expiration date
        """
        # datetime translated to YYMMDDHHMMSSZ format
        ca_start_date, ca_end_date = self._get_start_end_dates(ca_start_date, ca_end_date)
        server_start_date, server_end_date = self._get_start_end_dates(server_start_date,
                                                                       server_end_date)

        commands = list()
        # Generate a self-signed certificate for the CA (root)
        root_genrsa_cmd = 'openssl genrsa -out "{}" 2048'.format(self.root_key)
        root_req_cmd = 'openssl req -new -key "{}" -out "{}" ' \
                       '-subj "/C=US/ST=NC/L=RTP/O=NETAPP/OU=NTAP/CN=NTAP-ROOTCA"'\
            .format(self.root_key, self.root_csr)
        root_ca_cmd = 'openssl ca -batch -config "{}" -selfsign -keyfile "{}" -extensions v3_ca ' \
                      '-notext -startdate {} -enddate {} -in "{}" -out "{}"'\
            .format(self.openssl_config, self.root_key, ca_start_date, ca_end_date, self.root_csr,
                    self.root_cert)
        commands.extend([root_genrsa_cmd, root_req_cmd, root_ca_cmd])

        # Generate a server certificate (local)
        local_genrsa_cmd = 'openssl genrsa -out "{}" 2048'.format(self.server_key)
        # Note - Setting Common Name to the IP address of the client, instead of the client's hostname,
        #        this probably doesn't matter.
        local_req_cmd = 'openssl req -config "{}" -key "{}" -new -sha256 -out "{}" ' \
                        '-subj "/C=US/ST=NC/L=RTP/O=NETAPP/OU=NTAP/CN={}"'\
            .format(self.openssl_config, self.server_key, self.server_csr, self.hostname)

        local_ca_cmd = 'openssl ca -batch -config "{}" -extensions server_cert -notext -md sha256 ' \
                       '-in "{}" -out "{}" -startdate {} -enddate {}'\
            .format(self.openssl_config, self.server_csr, self.server_cert, server_start_date,
                    server_end_date)
        commands.extend([local_genrsa_cmd, local_req_cmd, local_ca_cmd])

        # Move certs into the kmip bin and fill server cert with x509 pem info
        commands.append('sudo rm {0}/CA.pem; sudo rm {0}/server.pem'.format(self.kmip_bin))
        commands.append('(openssl x509 -noout -text -in "{}"; printf "\n") >> {}/server.pem'.format(
            self.server_cert, self.kmip_bin))
        commands.append('cat "{}" >> {}/CA.pem'.format(self.root_cert, self.kmip_bin))
        commands.append('(cat "{}"; printf "\n"; cat "{}") >> {}/server.pem'.format(
            self.server_cert, self.server_key, self.kmip_bin))
        commands.append('cp "{}" "{}/bin"'.format(self.server_cert, self.kmip_dir))

        self.run_multiple_commands(commands, raise_on_failure=True)

    def _generate_user_certificate(self, csr, csr_file_path, cert_file_path, start_date, end_date,
                                   force=False):
        """Generate a user cert using the given CSR and register it with the Cryptsoft server.

        :param cryptography.x509.CertificateSigningRequest csr: x509 object for the CSR
        :param str csr_file_path: File path of the CSR
        :param str cert_file_path: File path used when creating user certificate
        :param datetime.datetime start_date: Datetime object representing the certificate start date
        :param datetime.datetime end_date: Datetime object representing the certificate expiration date
        :param bool force: If True, react to failures and force the operation.
        :raises: _errors.ExternalKeyServerError
        """
        # Generate a user certificate
        ca_cmd = 'openssl ca -batch -config "{}" -extensions usr_cert -notext -md sha256 -in "{}"' \
                 ' -out "{}" -startdate {} -enddate {}'.format(
                  self.openssl_config, csr_file_path, cert_file_path, start_date, end_date)

        # Move the user certificate into bin
        cp_cmd = 'cp "{}" "{}"'.format(csr_file_path, self.kmip_bin)

        power_back_on = False
        if self.is_running():
            self.power_off()
            power_back_on = True

        # Initialize the user cert with Cryptsoft server
        inituser_cmd = '{}/inituser "{}"'.format(self.kmip_bin, cert_file_path)
        self.run_multiple_commands([ca_cmd, cp_cmd], raise_on_failure=True)

        try:
            self.run_command(inituser_cmd, raise_on_failure=True)
        except _errors.ExternalKeyServerError as e:
            if 'username are not unique' in str(e) and force:
                username = csr.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0].value
                _log.info('Username {} already exists in cryptsoft database, removing it.'.format(
                    username))
                del_cmd = 'cd {}; ./kmip_server_usr_01 del username="{}"'.format(self.kmip_bin,
                                                                                 username)
                self.run_multiple_commands([del_cmd, inituser_cmd], raise_on_failure=True)
            else:
                if power_back_on:
                    self.power_on()
                raise e

        if power_back_on:
            self.power_on()

    def _get_start_end_dates(self, start_date=None, end_date=None):
        """Takes given start/end dates, sets default values, and converts to correct format.

        :param datetime.datetime start_date: Datetime object for the start date
        :param datetime.datetime end_date: Datetime object for the end date
        :rtype: tuple[str]
        """
        if start_date is None:
            start_date = datetime.datetime.utcnow()
        if end_date is None:
            end_date = start_date + datetime.timedelta(days=365)
        start_date = start_date.strftime('%Y%m%d%H%M%SZ')
        end_date = end_date.strftime('%Y%m%d%H%M%SZ')
        return start_date, end_date

    def execute_sql(self, sql_query, commit=False):
        """Run the given SQL query against the Cryptsoft database.

        There are two tables currently utilized in the Cryptsoft FM: obj and att.

        obj - Table containing all authentication key values
        Each row in the obj table represent an authentication key created by
        the server. The key value is stored in several columns depending on
        the format that the key needs to be returned in (kmip, json, xml , etc).

        att - Table containing all key attributes
        Each row of the att table is a an attribute of a particular key.
        The are many different key attributes and the name of the attribute can
        be found in the name column. Just like the obj table, this table also contains
        attribute values in various forms. The oid and ouuid columns will tell you
        which key the attribute is associated with. It maps directly to obj.id and
        obj.uuid respectively.

        :param str sql_query: SQL query to execute
        :param bool commit: Commit the database transaction
        :raises: sf_ssh.errors.SshError
        :rtype: str
        """
        kdb_cmd = 'cd {}; ./k_db sql="{}"'.format(self.kmip_bin, sql_query)
        output = self.run_command(kdb_cmd, raise_on_failure=True)

        if commit:
            kdb_commit = 'cd {}; ./k_db commit'.format(self.kmip_bin)
            self.run_command(kdb_commit, raise_on_failure=True)
        return output.stdout

    def get_key(self, key_id):
        """Query the key value of the given key ID from the Cryptsoft database.

        :param str key_id: ID of the key on the EKS
        :raises: _errors.ExternalKeyServerError
        :rtype: str
        """
        # Find the object ID
        oid = self._get_object_id(key_id)

        # The key value is stored in 4 columns in the obj table: value, json, tagtext, and xml.
        # It would be nice to get the key from the value column, but that value is padded with
        # extra data. It's easier to pull it out of one of the other columns.
        tag_query = 'SELECT tagtext FROM obj WHERE id={};'.format(oid)
        tag_output = self.execute_sql(tag_query)

        key_match = re.search(r'KEY_MATERIAL:bstr:(.*)', tag_output)
        if key_match:
            return key_match.group(1)
        else:
            raise _errors.ExternalKeyServerError('Unable to find key ID {}'.format(key_id))

    def set_key(self, key_id, new_key_value, key=None):
        """Set the value of the given key ID.

        :param str key_id: ID of the key on the EKS
        :param str new_key_value: New key value that will replace the current value
        :param str key: Optional parameter if you already know the current key value
        """
        if key is None:
            key = self.get_key(key_id)
        oid = self._get_object_id(key_id)

        # The key value returned by the KMIP operation is stored in the value column,
        # but to be consistent, let's change the key value for all the columns.
        update_sql = "UPDATE obj SET value = REPLACE(value, '{x}', '{y}')," \
                     "json = REPLACE(json, '{x}', '{y}')," \
                     "tagtext = REPLACE(tagtext, '{x}', '{y}')," \
                     "xml = REPLACE(tagtext, '{x}', '{y}') WHERE id={z};"
        self.execute_sql(update_sql.format(x=key, y=new_key_value, z=oid), commit=True)

    def _get_object_id(self, key_id):
        """Returns the object ID that matches the given key ID.

        :param str key_id: ID of the key on the EKS
        :raises: _errors.ExternalKeyServerError
        :rtype: int
        """
        oid_query = "SELECT oid FROM att WHERE name='Name' AND " \
                    "json LIKE '%x-SolidFire-KeyID-{}%';".format(key_id)
        oid_output = self.execute_sql(oid_query)
        oid_match = re.search(r'oid=(\d+)', oid_output)
        if oid_match:
            return int(oid_match.group(1))
        else:
            raise _errors.ExternalKeyServerError('Unable to find key ID {}'.format(key_id))

    def is_running(self):
        """Returns True if a kmip_server process is listening to the configured port.

        kmip_server is the name of the process for the Cryptsoft server.

        :rtype: bool
        """
        pid = self.get_server_pid()
        if pid:
            return True
        else:
            return False

    def get_server_pid(self, port=None):
        """Returns the PID of kmip_server.

        :param int|str port: Port to look for
        :rtype: int|None
        """
        if port is None:
            port = self.port

        cmd = 'netstat -ltnp | grep ":{}"'.format(port)
        output = self.run_command(cmd)
        if 'kmip_server' not in output.stdout:
            return None
        pid_process = [word for word in output.stdout.split() if 'kmip_server' in word]
        pid = pid_process[0].split('/')[0]
        return int(pid)

    def power_on(self):
        """Run the kmip_server, return the pid."""
        pid = self.get_server_pid()
        if pid is None:
            _log.debug('Starting Cryptsoft as kmip_server')
            self.run_command('screen -d -m {}/run kmip_server debug=1'.format(
                self.kmip_bin), raise_on_failure=True, use_sudo=True)
        else:
            _log.debug('kmip_server already running as PID: {}'.format(pid))

    def power_off(self):
        """Kill the kmip_server pid."""
        pid = self.get_server_pid()
        if pid is None:
            _log.debug('Cryptsoft kmip_server process is not running')
        else:
            _log.debug('Killing Cryptsoft kmip_server process: {}'.format(pid))
            self.run_command('kill -9 {}'.format(pid), raise_on_failure=True)

    def enable(self):
        """Wraps power_on to enable the EKS.

        Use to restore the connection between the SF Cluster and the EKS.
        """
        _log.debug('Enable kmip server')
        self.power_on()

    def disable(self):
        """Wraps power_off to disable the EKS.

        Use to sever the connection between the SF Cluster and the EKS.
        """
        _log.debug('Disable kmip server')
        self.power_off()

    def get_logs(self):
        """Pull the traffic log files to local tmp dir.

        Convenience method for test development.
        """
        with self.get_ssh_connection() as ssh_conn:
            ssh_conn.get([self.kmip_bin + '/traffic.log',
                          self.kmip_bin + '/traffic-default_partition.log'],
                         local_path='/tmp')

    def store_logs(self, client_dm):
        """Stores the traffic log files in the artifact manager.

        :param valence.models.ResourceClient client_dm: the Resource Client Cryptsoft EKS is deployed on.
        """
        result = self.run_command('ls {} {}'.format(
            self.kmip_bin + '/traffic.log', self.kmip_bin + '/traffic-default_partition.log'))

        if result.return_code == 0:
            log_files = result.stdout.splitlines()
        else:
            _log.debug('No traffic log files found')
            return

        manager = sf_artifacts.get_global_manager()
        manager.store_artifacts_remote(
            artifacts=log_files,
            remote_host_dm=client_dm
        )

    def configure_for_testing(self, kmip_version=None):
        """Confirm EKS is ready for test execution.

        - Confirm server is running on the expected port number
        - Confirm KMIP is enabled with eks_fm.is_running()

        Note - kmip_server/bin/server.cfg controls the settings for a cryptsoft server, however,
        modifying these values did not seem to have any effect when calling the run kmip_server script.

        :param sf_models.enums.KmipVersion kmip_version: Version to use.
        """
        # Confirm port matches the given value
        cmd = 'grep "LOCALHOST.port={}" "{}"'.format(self.port, self.servers_config)
        output = self.run_command(cmd)
        if not output.stdout:
            # Stop any running servers
            while True:
                pid = self.get_server_pid(port='*')
                if pid is None:
                    break
                self.run_command('kill -9 {}'.format(pid), raise_on_failure=True)

            # Replace port value in config file
            sed_cmd = 'sed -i "s|LOCALHOST.port=[[:digit:]]\\+|LOCALHOST.port={}|g" "{}"'.format(
                self.port, self.servers_config)
            self.run_command(sed_cmd)

        # Confirm KMIP is enabled
        if not self.is_running():
            _log.debug('KMIP was not enabled, enabling the Cryptsoft server.')
            self.enable()

    def sign_csr(self, csr, start_date=None, end_date=None, force=False, name=None):
        """Take the given CSR and get it signed by the CA local to the Cryptsoft service.

        The CSR is on the test runner and copied to the client running Cryptsoft. The client generates
        the CA and initializes it with the Cryptsoft service. Copy this new CA and the server CA back
        to the test runner and return the x509.Certificate objects for them.

        :param cryptography.x509.CertificateSigningRequest csr: x509 object for the CSR
        :param datetime.datetime start_date: Datetime object representing the certificate start date
        :param datetime.datetime end_date: Datetime object representing the certificate expiration date
        :param bool force: If True, react to failures and force the operation.
        :param str name: Optional name for csr and ca on server, for tracking multiple sign_csr requests
        :rtype: tuple[cryptography.x509.Certificate]
        """
        # datetime translated to YYMMDDHHMMSSZ format
        start_date, end_date = self._get_start_end_dates(start_date, end_date)

        # Add new CSR from request
        if name is not None:
            csr_request = self.user_csr_dir + name + '.csr.pem'
            cert_request = self.user_cert_dir + name + '.cert.pem'
        else:
            csr_request = self.user_csr_dir + self._unique_id + '.csr.pem'
            cert_request = self.user_cert_dir + self._unique_id + '.cert.pem'

        # Move CSR text to client
        csr_text = csr.public_bytes(serialization.Encoding.PEM)
        with tempfile.NamedTemporaryFile() as fh:
            fh.write(csr_text)
            fh.flush()
            with self.get_ssh_connection() as ssh_conn:
                ssh_conn.put([fh.name], remote_path=csr_request)

        self._generate_user_certificate(csr, csr_request, cert_request, start_date, end_date, force)

        self.signed_csrs.append(csr_request)
        self.signed_certs.append(cert_request)

        # Pull the client and CA certs back to the test runner from their Cryptsoft EKS paths
        # CryptsoftExternalKeyServer.cert_request was just created and will be sent back as client_cert
        # CryptsoftExternalKeyServer.root_cert was created to spin up a Cryptsoft server
        # todo - use NamedTemporaryFile
        client_cert_path = '/tmp/client_{}.cert.pem'.format(self._unique_id)
        ca_cert_path = '/tmp/ca_{}.cert.pem'.format(self._unique_id)
        with self.get_ssh_connection() as ssh_conn:
            ssh_conn.get([cert_request], local_path=client_cert_path)
            ssh_conn.get([self.root_cert], local_path=ca_cert_path)

        # Grab the client CA and server CA text on the test runner
        with open(client_cert_path, 'rb') as fh:
            client_cert_text = fh.read()
        with open(ca_cert_path, 'rb') as fh:
            ca_cert_text = fh.read()

        # Create the Certificate objects to return
        client_cert = x509.load_pem_x509_certificate(client_cert_text, default_backend())
        ca_cert = x509.load_pem_x509_certificate(ca_cert_text, default_backend())
        return client_cert, ca_cert
