"""HyTrust EKS functional model."""
import datetime
import os
import tempfile
import zipfile

import autologger
import requests
import six
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from valence import enums
from valence.functional_models import _errors
from valence.functional_models._external_key_server import base

_log = autologger.get_sflogging()


class HyTrustExternalKeyServer(base.ExternalKeyServerBase):
    """EKS functional model for HyTrust KeyControl.

    This class will act as an interface to any HyTrust external key server. The HyTrust API
    will be translated into Python functions and used to implement the abstract methods defined
    in ExternalKeyServerBase. Testcases should utilize the abstract methods whenever possible so
    that they can be EKS agnostic.

    For more details about HyTrust's API, see the documentation:
    https://docs.hytrust.com/DataControl/4.3/Online/Content/OLH-Files/Help-content-map-all-books.html

    :param sf_models.ExternalKeyServer eks_dm: EKS data model
    """
    # Mapping kmip version number to HyTrust syntax
    PROTOCOL_MAP = {
        enums.KmipVersion.proto_1_0: '0x1',
        enums.KmipVersion.proto_1_1: '0x200',
        enums.KmipVersion.proto_1_2: '0x10000',
        enums.KmipVersion.proto_1_3: '0x40000',
        enums.KmipVersion.proto_1_4: '0x800000',
    }
    INVERSE_PROTOCOL_MAP = {
        '0x1': enums.KmipVersion.proto_1_0,
        '0x200': enums.KmipVersion.proto_1_1,
        '0x10000': enums.KmipVersion.proto_1_2,
        '0x40000': enums.KmipVersion.proto_1_3,
        '0x800000': enums.KmipVersion.proto_1_4,
    }

    def __init__(self, eks_dm):
        super(HyTrustExternalKeyServer, self).__init__(eks_dm)
        self.root_uri = 'https://{}/'.format(self.hostname)
        self.access_token = None

    def sign_csr(self, csr, start_date=None, end_date=None, force=False):
        """Take the given CSR and get it signed by the CA local to HyTrust.

        This method will create the KMIP certificate bundle on HyTrust, download the bundle,
        and extract the client and CA certificate from the bundle.

        :param cryptography.x509.CertificateSigningRequest csr: x509 object for the CSR
        :param datetime.datetime start_date: Datetime object representing the certificate start date
        :param datetime.datetime end_date: Datetime object representing the certificate expiration date
        :param bool force: N/A
        :rtype: tuple[cryptography.x509.Certificate]
        """

        # HyTrust cannot specify a starting date, only a total duration.
        # So always use now as the start date
        start_date = datetime.datetime.utcnow()
        if end_date is None:
            end_date = start_date + datetime.timedelta(days=365)
        delta = end_date - start_date
        cert_duration = delta.days if delta.days > 0 else 1

        # Extract text from CSR
        csr_text = csr.public_bytes(serialization.Encoding.PEM)

        # For the username, use the OU field of the CSR
        user = csr.subject.get_attributes_for_oid(x509.oid.NameOID.ORGANIZATIONAL_UNIT_NAME)[0].value
        cert_id = user + self._unique_id
        self.create_kmip_certificate_bundle(cert_id=cert_id, expire=cert_duration, username=user,
                                            csr=csr_text)

        client_cert_text, ca_cert_text = self.download_kmip_certificate_bundle(cert_id=cert_id)
        client_cert = x509.load_pem_x509_certificate(client_cert_text, default_backend())
        ca_cert = x509.load_pem_x509_certificate(ca_cert_text, default_backend())

        return client_cert, ca_cert

    # Authentication and Admin Methods
    def login(self):
        """Issue the login API and save the access token.

        :rtype: dict
        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: API returned status
            code other than 200
        """
        url = self.root_uri + 'v5/kc/login/'
        params = dict(
            username=self.username,
            password=self.password,
        )

        resp = requests.post(url, json=params, verify=False)
        if resp.status_code == 200:
            data = resp.json()
            self.access_token = data['access_token']
        else:
            err_msg = 'API call failed with code {}: {}'.format(resp.status_code, resp.content)
            raise _errors.HyTrustExternalKeyServerError(err_msg)
        return data

    def renew_token(self):
        """Issue the renew token API and save the new access token.

        :rtype: dict
        """
        data = self._call_method('v5/kc/renew/', 'POST')
        self.access_token = data['access_token']
        return data

    def logout(self):
        """Issue the logout API and remove the current access token.

        :rtype: dict
        """
        data = self._call_method('v5/kc/logout/', 'POST')
        self.access_token = None
        return data

    def get_admin_key(self, target='./secroot.key'):
        """Retrieves the most recent Admin Key for the user associated with the request.

        :param str target: File path to save the admin key to
        :rtype: str
        """
        resp = self._call_method('v5/admin_key/', 'GET', json=False)
        with open(target, 'wb') as fh:
            fh.write(resp.content)
        return os.path.abspath(target)

    def get_alerts(self):
        """Displays the most recent alerts accessible to the user associated with the request.

        :rtype: dict
        """
        return self._call_method('v5/alerts/', 'GET')

    def delete_alerts(self, alert_ids):
        """Clear the specified alerts.

        :param str alert_ids: Comma separated string of alert IDs
        :rtype: dict
        """
        params = dict(alertids=alert_ids)
        return self._call_method('v5/alerts/', 'PATCH', params=params)

    # Group Management Methods
    def list_groups(self):
        """Returns a list of all the cloud admin groups.

        :rtype: dict
        """
        return self._call_method('v5/groups/', 'GET')

    def get_group_details(self, group_id):
        """Returns details and member list for the specified group.

        :param str group_id: Group ID
        :rtype: dict
        """
        return self._call_method('v5/groups/{}/'.format(group_id), 'GET')

    def create_group(self, name, privilege, desc, user_members):
        """Create a new cloud admin group.

        Privilege must be CLOUD_ADMIN or DOMAIN_ADMIN.
        Member list must be in the form [{'id': 'username'}].

        :param str name: Name of the group
        :param str privilege: Privilege level
        :param str desc: Description of the group
        :param list[dict] user_members: List of members for the group
        :rtype: dict
        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: Invalid privilege given
        """
        if privilege and privilege not in ['CLOUD_ADMIN', 'DOMAIN_ADMIN']:
            err_msg = 'Invalid privilege parameter: {}'.format(privilege)
            raise _errors.HyTrustExternalKeyServerError(err_msg)

        params = dict(
            name=name,
            privilege=privilege,
            desc=desc,
            user_members=user_members,
        )
        return self._call_method('v5/groups/', 'POST', params=params)

    def edit_group(self, group_id, name=None, privilege=None, desc=None, members=None, add_members=None,
                   delete_members=None):
        """Modify the details of the specified group.

        :param str group_id: Group ID
        :param str name: Name of the group
        :param str privilege: Privilege level
        :param str desc: Description of the group
        :param list[str] members: List of members for the group
        :param list[str] add_members: List of members to add to the group
        :param list[str] delete_members: List of members to remove from the group
        :rtype: dict
        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: Invalid privilege given
        """
        if privilege and privilege not in ['CLOUD_ADMIN', 'DOMAIN_ADMIN']:
            err_msg = 'Invalid privilege parameter: {}'.format(privilege)
            raise _errors.HyTrustExternalKeyServerError(err_msg)

        params = dict(
            name=name,
            privilege=privilege,
            desc=desc,
            members=members,
            add_members=add_members,
            delete_members=delete_members,
        )
        return self._call_method('v5/groups/{}/'.format(group_id), 'PATCH', params=params)

    def delete_group(self, group_id):
        """Delete the specified group.

        You can delete a group that contains users but you cannot delete a group that owns resources.
        The resources must be reassigned to a different group before the group can be deleted.

        :param str group_id: Group ID
        :rtype: dict
        """
        return self._call_method('v5/groups/{}/'.format(group_id), 'DELETE')

    # User Management Methods
    def list_users(self):
        """Returns all users created on the HyTrust system.

        :rtype: dict
        """
        return self._call_method('v5/users/', 'GET')

    def get_user_details(self, user_id):
        """Returns details about the specified user.

        :param str user_id: User ID
        :rtype: dict
        """
        return self._call_method('v5/users/{}/'.format(user_id), 'GET')

    def change_default_user_settings(self, password=None, default_type=None):
        """Change the default user creation settings.

        Changes to these settings affect user accounts created after the changes are made.
        Existing user accounts are not affected.

        :param dict password: Default password details
        :param str default_type: Default user type
        :rtype: dict
        """
        params = dict(
            password=password,
            type=default_type,
        )
        return self._call_method('v5/system_settings/user_authentication/', 'PATCH', params=params)

    def create_user(self, email, full_name, groups, login_name, password, privileges,
                    account_expiration=None, password_expiration=None, radius_auth=None,
                    radius_port=None, radius_server=None, account_enable=None, authentication=None):
        """Create a new KeyControl user account.

        :param str email: Email
        :param str full_name: Name
        :param list[str] groups: Groups the user is associated with
        :param str login_name: Username
        :param str password: Password
        :param list[str] privileges: Privileges of this user
        :param str account_expiration: Date on which this user account should expire (dd/mm/yyyy)
        :param str password_expiration: Date on which the password should expire (dd/mm/yyyy)
        :param str radius_auth: The RADIUS server authentication method for this account
        :param str radius_port: The port number for the RADIUS server
        :param str radius_server: The RADIUS IP address or hostname, if specified
        :param str account_enable: Enable the account after creation (yes or no)
        :param str authentication: Authentication type (Local, RADIUS, or LDAP)
        :rtype: dict
        """
        params = dict(
            account_enable=account_enable,
            account_expiration=account_expiration,
            authentication=authentication,
            email=email,
            full_name=full_name,
            groups=groups,
            login_name=login_name,
            password=password,
            password_expiration=password_expiration,
            privileges=privileges,
            radius_auth=radius_auth,
            radius_port=radius_port,
            radius_server=radius_server,
        )
        return self._call_method('v5/users/', 'POST', params=params)

    def edit_user(self, user_id, email, full_name, groups, password, privileges,
                  account_expiration=None, password_expiration=None, account_enable=None):
        """Edit details of the specified user account.

        :param str user_id: User ID
        :param str email: Email
        :param str full_name: Name
        :param list[str] groups: Groups the user is associated with
        :param str password: Password
        :param list[str] privileges: Privileges of this user
        :param str account_expiration: Date on which this user account should expire (dd/mm/yyyy)
        :param str password_expiration: Date on which the password should expire (dd/mm/yyyy)
        :param str account_enable: Enable the account after creation (yes or no)
        :rtype: dict
        """
        params = dict(
            account_enable=account_enable,
            account_expiration=account_expiration,
            email=email,
            full_name=full_name,
            groups=groups,
            password=password,
            password_expiration=password_expiration,
            privileges=privileges,
        )
        return self._call_method('v5/users/{}/'.format(user_id), 'PATCH', params=params)

    def delete_user(self, user_id):
        """Delete the specified user account.

        :param str user_id: User ID
        :rtype: dict
        """
        return self._call_method('v5/users/{}/'.format(user_id), 'DELETE')

    # KMIP Server Management Methods
    def get_kmip_server(self):
        """Return a dict with the result and kmipsrv_info settings.

        :rtype: dict
        """
        return self._call_method('v5/system_settings/kmipsrv_info/', 'GET')

    def enable_kmip_server(self, host, protocol, loglevel='ALL', nbio=False, port=5696, reconnect=True,
                           state='ENABLED', timeout=0, verify=True):
        """Enable the KMIP server with the specified settings.

        :param str host: KMIP Server IP address
        :param valence.enums.KmipVersion protocol: KMIP version
        :param str loglevel: Log level to use for the KMIP server (ALL, CREATE-GET, OFF)
        :param bool nbio: Whether the KMIP server requires non-blocking I/O
        :param int port: The KMIP server port
        :param bool reconnect: Automatic reconnect
        :param str state: Enable or disable the server (ENABLED, DISABLED)
        :param int timeout: Length of time, in seconds, after which a client request will time out
        :param bool verify: Verify KMIP client identity
        :rtype: dict
        """
        params = dict(
            host=host,
            loglevel=loglevel,
            nbio=str(int(nbio)),
            port=str(port),
            protocol=self.PROTOCOL_MAP[protocol],
            reconnect=str(int(reconnect)),
            state=state,
            timeout=timeout,
            verify='yes' if verify else 'no',
        )
        return self._call_method('v5/system_settings/kmipsrv_info/', 'PATCH', params=params)

    def enable(self):
        """Wraps enable_kmip_server to enable the EKS.

        Use to restore the connection between the SF Cluster and HyTrust EKS.

        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: API result != success
        """
        _log.debug('Enable kmip server')
        settings = self.get_kmip_server()['settings']['kmipsrv_info']
        resp = self.enable_kmip_server(self.hostname, self.kmip_version, loglevel=settings['loglevel'],
                                       nbio=settings['nbio'], port=self.port,
                                       reconnect=settings['reconnect'], state='ENABLED',
                                       timeout=settings['timeout'], verify=settings['verify'])
        if resp['result'] != 'success':
            raise _errors.HyTrustExternalKeyServerError('Failed to enable kmip server.')

    def disable(self):
        """Wraps enable_kmip_server to disable the EKS.

        Use to sever the connection between the SF Cluster and HyTrust EKS.

        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: API result != success
        """
        _log.debug('Disable kmip server')
        settings = self.get_kmip_server()['settings']['kmipsrv_info']
        resp = self.enable_kmip_server(self.hostname, self.kmip_version, loglevel=settings['loglevel'],
                                       nbio=settings['nbio'], port=self.port,
                                       reconnect=settings['reconnect'], state='DISABLED',
                                       timeout=settings['timeout'], verify=settings['verify'])

        if resp['result'] != 'success':
            raise _errors.HyTrustExternalKeyServerError('Failed to disable kmip server.')

    def configure_for_testing(self, kmip_version=None):
        """Confirm EKS is ready for test execution.

        - Confirm server is running on the expected port number
        - Confirm KMIP is enabled.
        - Confirm HyTrust is using desired KMIP version

        :param sf_models.enums.KmipVersion kmip_version: Version to use.
        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: if changing the kmip
            versions fails
        """
        kmip_server_dict = self.get_kmip_server()
        kmip_settings = kmip_server_dict['settings']['kmipsrv_info']

        # Confirm desired KMIP version and port
        if ((kmip_version is not None
                and kmip_settings['protocol'] != self.PROTOCOL_MAP[kmip_version])
                or kmip_settings['port'] != str(self.port)):

            if kmip_version is None:
                kmip_version = self.INVERSE_PROTOCOL_MAP[kmip_settings['protocol']]

            _log.debug('Re-enable KMIP server to version {}, port {}'.format(kmip_version.value,
                                                                             self.port))
            resp = self.enable_kmip_server(self.hostname, kmip_version,
                                           loglevel=kmip_settings['loglevel'],
                                           nbio=kmip_settings['nbio'], port=self.port,
                                           reconnect=kmip_settings['reconnect'],
                                           state='ENABLED',
                                           timeout=kmip_settings['timeout'],
                                           verify=kmip_settings['verify'])
            if resp['result'] != 'success':
                raise _errors.HyTrustExternalKeyServerError('Failed to re-enable KMIP server.')
            kmip_server_dict = self.get_kmip_server()
            kmip_settings = kmip_server_dict['settings']['kmipsrv_info']

        # Confirm KMIP is enabled
        if kmip_server_dict['result'] != 'success' or kmip_settings['state'] != 'ENABLED':
            _log.debug('KMIP was not enabled, enabling the HyTrust server.')
            self.enable()

    def is_running(self):
        """Return True if the KMIP service or EKS is up and running.

        :rtype: bool
        """
        kmip_server_dict = self.get_kmip_server()
        kmip_settings = kmip_server_dict['settings']['kmipsrv_info']
        if kmip_server_dict['result'] == 'success' and kmip_settings['state'] == 'ENABLED':
            return True
        else:
            return False

    def create_kmip_certificate_bundle(self, cert_id, expire, username=None, password=None, csr=None):
        """Sign the specified CSR and create a client certificate.

        Client certificate is saved as a "kmip certificate bundle" and must be downloaded
        with the download_kmip_certificate_bundle() API.

        :param str cert_id: Certificate bundle ID
        :param int expire: Number of days for which this certificate should be valid
        :param str username: A user-defined name for the certificate bundle
        :param str password: Optional passphrase used to encrypt the certificates in the bundle
        :param str csr: Certificate Signing Request
        :rtype: dict
        """
        params = dict(
            expire=str(expire),
            username=username,
            passwd=password,
            kmip_csr=csr,
        )
        return self._call_method('v5/kmipsrv_user/{}/'.format(cert_id), 'POST', params=params)

    def download_kmip_certificate_bundle(self, cert_id):
        """Download the certificate bundle as a zip file.

        :param str cert_id: Certificate bundle ID
        :rtype: tuple[str]
        """
        resp = self._call_method('v5/kmipsrv-certs/{}/'.format(cert_id), 'GET', json=False)

        with tempfile.NamedTemporaryFile() as fh:
            fh.write(resp.content)
            fh.flush()
            archive = zipfile.ZipFile(fh.name, 'r')
            client_cert_text = archive.read('{}.pem'.format(cert_id))
            ca_cert_text = archive.read('cacert.pem')

        return client_cert_text, ca_cert_text

    def edit_kmip_certificate(self, cert_id, expire=None, password=None):
        """Edit details about the specified certificate bundle.

        :param str cert_id: Certificate bundle ID
        :param int expire: Number of days for which this certificate should be valid
        :param str password: Optional passphrase used to encrypt the certificates in the bundle
        :rtype: dict
        """
        params = dict(
            expire=str(expire) if expire else None,
            passwd=password,
        )
        return self._call_method('v5/kmipsrv_user/{}/'.format(cert_id), 'PATCH', params=params)

    def delete_kmip_certificate(self, cert_id, username=None):
        """Delete the specified certificate bundle.

        :param str cert_id: Certificate bundle ID
        :param str username: A user-defined name for the certificate bundle
        :rtype: dict
        """
        params = dict(username=username)
        return self._call_method('v5/kmipsrv_user/{}/'.format(cert_id), 'DELETE', params=params)

    def delete_all_kmip_certificates(self):
        """Deletes every certificate bundle on the system.

        :rtype: dict
        """
        return self._call_method('v5/kmipsrv_user/', 'DELETE')

    def list_kmip_objects(self, limit=100, offset=0, attr=None):
        """Returns KMIP objects that exist on the system.

        :param int limit: The total number of objects to retrieve
        :param int offset: The index of the first object you want to retrieve
        :param str attr: Optional parameter that returns details about each KMIP object
        :rtype: dict
        """
        params = dict(
            _limit=limit,
            _offset=offset,
        )
        if attr is not None:
            params['attr'] = attr

        # This is the only API that does not support JSON input params, must be passed in via URL.
        url = 'v5/kmipsrv_object/?'
        param_list = []
        for k, v in six.iteritems(params):
            param_list.append('{}={}'.format(k, v))
        url += '&'.join(param_list)

        return self._call_method(url, 'GET')

    def get_kmip_object(self, obj_uuid):
        """Returns details about the specified KMIP object.

        :param str obj_uuid: KMIP object UUID
        :rtype: dict
        """
        return self._call_method('v5/kmipsrv_object/{}/'.format(obj_uuid), 'GET')

    def activate_kmip_object(self, obj_uuid):
        """Activates the specified KMIP object.

        :param str obj_uuid: KMIP object UUID
        :rtype: dict
        """
        params = dict(action='activate')
        return self._call_method('v5/kmipsrv_object/{}/'.format(obj_uuid), 'PATCH', params=params)

    def archive_kmip_object(self, obj_uuid):
        """Archives the specified KMIP object.

        :param str obj_uuid: KMIP object UUID
        :rtype: dict
        """
        params = dict(action='archive')
        return self._call_method('v5/kmipsrv_object/{}/'.format(obj_uuid), 'PATCH', params=params)

    def destroy_kmip_object(self, obj_uuid):
        """Destroys the specified KMIP object.

        :param str obj_uuid: KMIP object UUID
        :rtype: dict
        """
        params = dict(action='destroy')
        return self._call_method('v5/kmipsrv_object/{}/'.format(obj_uuid), 'PATCH', params=params)

    def recover_kmip_object(self, obj_uuid):
        """Recovers the specified KMIP object.

        :param str obj_uuid: KMIP object UUID
        :rtype: dict
        """
        params = dict(action='recover')
        return self._call_method('v5/kmipsrv_object/{}/'.format(obj_uuid), 'PATCH', params=params)

    def revoke_kmip_object(self, obj_uuid, revcode, revmsg=None):
        """Revokes the specified KMIP object.

        KMIP Revoke Codes:
        1  Unspecified
        2  Key Compromise
        3  CA Compromise
        4  Affiliation Changed
        5  Superseded
        6  Cessation of Operation
        7  Privilege Withdrawn

        :param str obj_uuid: KMIP object UUID
        :param int revcode: Standard KMIP reason codes corresponding to why the object is being revoked
        :param str revmsg: Optional string describing why the object is being revoked
        :rtype: dict
        """
        params = dict(
            action='revoke',
            revcode=revcode,
            revmsg=revmsg,
        )
        return self._call_method('v5/kmipsrv_object/{}/'.format(obj_uuid), 'PATCH', params=params)

    def _call_method(self, uri, method, params=None, json=True):
        """Executes the requests HTTP call.

        :param str uri: API URI to be appended to root URI
        :param str method: HTTP verb
        :param dict params: Parameters to send in the request body
        :param bool json: If True, return json data of the response.
                          If False, return the whole response.
        :rtype: dict|requests.Response
        :raises valence.functional_models._errors.HyTrustExternalKeyServerError: API returned a
            bad status code
        """
        url = self.root_uri + requests.utils.requote_uri(uri)
        if self.access_token is None:
            self.login()

        # Remove any None values from params
        if params:
            params = {k: v for k, v in six.iteritems(params) if v is not None}

        attempted_login = False
        while True:
            headers = {'Auth-Token': self.access_token}
            _log.debug('{} {}, HEADERS: {}, PARAMS: {}'.format(method, url, headers, params))
            resp = requests.request(method, url, json=params, headers=headers, verify=False)

            if resp.status_code == 401:
                if attempted_login:
                    err_msg = 'Invalid access token'
                    raise _errors.HyTrustExternalKeyServerError(err_msg)
                else:
                    _log.info('Invalid access token, attempting to login again.')
                    self.login()
                    attempted_login = True
            elif resp.status_code != 200:
                err_msg = 'API call failed with code {}: {}'.format(resp.status_code, resp.content)
                raise _errors.HyTrustExternalKeyServerError(err_msg)
            else:
                return resp.json() if json else resp
