"""Safenet EKS functional model."""
import datetime
import re
import time

import autologger
import paramiko
import sf_ssh
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from valence import models
from valence.functional_models import _errors
from valence.functional_models._external_key_server import base

_log = autologger.get_sflogging()


class SafenetExternalKeyServer(base.ExternalKeyServerBase):
    """EKS functional model for Safenet K150v and G350v.

    This class will act as an interface to Safenet EKS versions 150v and 350v. Versions 150v and 350v
    are nearly identical, but marketed to different customers. The Safenet API will be translated into
    Python functions and used to implement the abstract methods defined in ExternalKeyServerBase.
    Testcases should utilize the abstract methods whenever possible so that they can be EKS agnostic.
    """

    # Abstract Methods
    def sign_csr(self, csr, start_date=None, end_date=None, force=False):
        """Take the given CSR and get it signed by a Trusted CA (usually local to the EKS).

        :param cryptography.x509.CertificateSigningRequest csr: x509 object for the CSR
        :param datetime.datetime start_date: Datetime object representing the certificate start date
        :param datetime.datetime end_date: Datetime object representing the certificate expiration date
        :param bool force: N/A
        :rtype: tuple[cryptography.x509.Certificate]
        """
        ca_name = 'KeySecure CA'
        cert_purpose = 2  # 2 = client cert

        # Safenet cannot specify a starting date, only a total duration.
        # So always use now as the start date
        start_date = datetime.datetime.utcnow()
        if end_date is None:
            end_date = start_date + datetime.timedelta(days=365)
        delta = end_date - start_date
        cert_duration = delta.days if delta.days > 0 else 1

        # Extract text from CSR
        csr_text = csr.public_bytes(serialization.Encoding.PEM)

        # Sign the CSR
        client_cert = self.sign_request(ca_name, cert_purpose, cert_duration, csr_text,
                                        raise_on_failure=True)

        # Run the show ca cert command
        local_ca = self.show_local_ca(ca_name, raise_on_failure=True)

        # Load the data into x509.Certificate
        x509_client_cert = x509.load_pem_x509_certificate(
            client_cert.text.encode(), default_backend())  # pylint: disable=no-member
        x509_ca_cert = x509.load_pem_x509_certificate(
            local_ca.text.encode(), default_backend())  # pylint: disable=no-member

        return x509_client_cert, x509_ca_cert

    # API Methods
    def sign_request(self, ca_name, cert_purpose, cert_duration, csr, timeout=60,
                     raise_on_failure=False):
        """Executes the interactive 'sign request' CLI command.

        :param str ca_name: Name of the local CA on the EKS
        :param int cert_purpose: Purpose of the certificate. Must be '1', '2', or '3':
                                 1) Server 2) Client 3) Intermediate CA
        :param int cert_duration: Number of days until certificate expires
        :param str csr: String containing the CSR text
        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :rtype: valence.models.SafenetClientCertificate
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        commands = ['sign request', ca_name, str(cert_purpose), str(cert_duration), csr]
        stdout = self.run_interactive_command(commands, timeout)

        purpose_map = {
            '1': 'Server',
            '2': 'Client',
            '3': 'Intermediate CA'
        }

        # Parse the response into a dict
        result = None
        match = re.search(
            r"The signed certificate is shown below:\n(-+BEGIN CERTIFICATE-+.*-+END CERTIFICATE-+)",
            stdout, re.S)
        if match:
            result = models.SafenetClientCertificate.create(
                ca_name=ca_name,
                purpose=purpose_map[str(cert_purpose)],
                duration=cert_duration,
                text=match.group(1).strip()
            )
        else:
            err_msg = 'sign_request command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

        return result

    def show_local_ca(self, ca_name=None, timeout=60, raise_on_failure=False):
        """Executes the 'show local ca' CLI command.

        If ca_name is provided, then details about that CA will be returned as a dict.
        If no ca_name is provided, then the details of all local CAs will be returned as a list.
        # noqa

        :param str ca_name: Name of the local CA on the EKS
        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :rtype: valence.models.SafenetLocalCertificateAuthority|list[valence.models.SafenetLocalCertificateAuthority]
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        if ca_name:
            # If a CA name is given, the output will contain information specific to the given CA.
            return self._show_specific_local_ca(ca_name, timeout=timeout,
                                                raise_on_failure=raise_on_failure)
        else:
            # If there is no CA given, then the command will get details for all local CAs.
            cmd = 'show local ca'
            stdout = self.run_interactive_command([cmd], timeout=timeout)
            lines = stdout.splitlines()

            results = []
            if len(lines) > 2:
                for line in lines[2:]:
                    split_line = re.split(r"\t+", line.strip())
                    specific_result = self._show_specific_local_ca(split_line[0], timeout=timeout,
                                                                   raise_on_failure=raise_on_failure)
                    results.append(specific_result)

            return results

    def _show_specific_local_ca(self, ca_name, timeout=60, raise_on_failure=False):
        """Return details about the given local CA.

        :param str ca_name: Name of the local CA on the EKS
        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :rtype: valence.models.SafenetLocalCertificateAuthority
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        cmd = 'show local ca "{}"'.format(ca_name)
        stdout = self.run_interactive_command([cmd], timeout=timeout)

        # Parse the response into a model
        result = models.SafenetLocalCertificateAuthority.create()
        match = re.search(r"PROPERTIES:(.*)Issuer:(.*)Subject(.*)Purpose:(.*)TEXT:(.*)", stdout, re.S)
        if match:
            # Property
            prop_re_list = [
                r"Certificate Name:\s+(.*)",
                r"Serial Number:\s+(.*)",
                r"Key Size:\s+(.*)",
                r"Start Date:\s+(.*)",
                r"Expiration:\s+(.*)"
            ]
            prop_re = r"\r?\n?".join(prop_re_list)
            prop_match = re.search(prop_re, match.group(1))
            if prop_match:
                result.name = prop_match.group(1)
                result.serial_number = prop_match.group(2)
                result.key_size = prop_match.group(3)
                result.start_date = prop_match.group(4)
                result.expiration = prop_match.group(5)

            # Issuer
            issuer_re_list = [
                r"C:\s+(.*)",
                r"ST:\s+(.*)",
                r"L:\s+(.*)",
                r"O:\s+(.*)",
                r"OU:\s+(.*)",
                r"CN:\s+(.*)",
                r"emailAddress:\s+(.*)",
            ]
            issuer_re = r"\r?\n?".join(issuer_re_list)
            issuer_match = re.search(issuer_re, match.group(2))
            if issuer_match:
                issuer_dm = models.CertificateSubject.create(
                    country=issuer_match.group(1),
                    state=issuer_match.group(2),
                    locality=issuer_match.group(3),
                    organization=issuer_match.group(4),
                    organizational_unit=issuer_match.group(5),
                    common_name=issuer_match.group(6),
                    email_address=issuer_match.group(7)
                )
                result.issuer = issuer_dm

            # Subject (same as issuer)
            subj_re = r"\r?\n?".join(issuer_re_list)
            subj_match = re.search(subj_re, match.group(3))
            if subj_match:
                subj_dm = models.CertificateSubject.create(
                    country=subj_match.group(1),
                    state=subj_match.group(2),
                    locality=subj_match.group(3),
                    organization=subj_match.group(4),
                    organizational_unit=subj_match.group(5),
                    common_name=subj_match.group(6),
                    email_address=subj_match.group(7)
                )
                result.subject = subj_dm

            # Purpose
            purposes = map(str.strip, match.group(4).split('\n'))
            purposes = [x for x in purposes if x != '']
            result.purpose = purposes

            # Certificate Text
            result.text = match.group(5).strip()
        else:
            err_msg = 'show_local_ca command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

        return result

    def show_nae_server_settings(self, timeout=60, raise_on_failure=False):
        """Executes the 'show nae-server settings' CLI command.

        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :rtype: list[valence.models.SafenetNaeServer]
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        cmd = 'show nae-server settings'
        stdout = self.run_interactive_command([cmd], timeout=timeout)

        result = []
        server_re_list = [
            r"Protocol:\s+(.*)",
            r"IP:\s+(.*)",
            r"Port:\s+(.*)",
            r"Use SSL:\s+(.*)",
            r"Server Certificate:\s+(.*)",
        ]
        server_re = r"\r?\n?".join(server_re_list)
        server_matches = re.findall(server_re, stdout)
        if len(server_matches):
            for match in server_matches:
                nae_server_dm = models.SafenetNaeServer.create(
                    protocol=match[0],
                    ip=match[1],
                    port=int(match[2]),
                    use_ssl=True if match[3] == 'yes' else False,
                    server_certificate=match[4],
                )
                result.append(nae_server_dm)
        else:
            err_msg = 'show_nae_server_settings command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

        return result

    def modify_nae_server_port(self, current_ip, current_port, port, timeout=60,
                               raise_on_failure=False):
        """Change the current KMIP port number.

        :param str current_ip: IP of the current nae-server service
        :param int current_port: Port number of the current nae-server service
        :param int port: Port to change to
        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        cmds = ['nae-server settings {} {}'.format(current_ip, current_port),
                '', str(port), '', '', '', '', '']
        stdout = self.run_interactive_command(cmds, timeout=timeout)

        if 'successfully saved' not in stdout:
            err_msg = 'modify_kmip_port command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

    def show_services(self, timeout=60, raise_on_failure=False):
        """Executes the 'show services' CLI command.

        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :rtype: list[valence.models.SafenetService]
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        cmd = 'show services'
        stdout = self.run_interactive_command([cmd], timeout=timeout)
        lines = stdout.splitlines()
        error = False

        # First line is the header
        headers = ['service_group', 'service', 'status', 'startup']

        # The rest are the services
        result = []
        if len(lines) > 1:
            for line in lines[1:]:
                split_line = re.split(r"\s{2,}", line)
                if len(split_line) == 4:
                    service = dict(zip(headers, split_line))
                    service_dm = models.SafenetService.create(**service)
                    result.append(service_dm)
                else:
                    # If there aren't four columns, something went wrong
                    error = True
                    break
        else:
            # If there's only one line, something went wrong
            error = True

        if error:
            err_msg = 'show_services command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

        return result

    def start_service(self, service, timeout=60, raise_on_failure=False):
        """Executes the 'service run' CLI command.

        :param str service: Name of the service
        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        cmds = ['{} run'.format(service), 'y']
        stdout = self.run_interactive_command(cmds, timeout=timeout)

        if 'successfully started' not in stdout:
            err_msg = 'start_service command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

    def stop_service(self, service, timeout=60, raise_on_failure=False):
        """Executes the 'no service run' CLI command.

        :param str service: Name of the service
        :param int timeout: Socket timeout on API command
        :param bool raise_on_failure: Raise exception if API fails
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: Command failed or
            returned unexpected output
        """
        cmds = ['no {} run'.format(service), 'y']
        stdout = self.run_interactive_command(cmds, timeout=timeout)

        if 'successfully stopped' not in stdout:
            err_msg = 'stop_service command failed: {}'.format(stdout)
            if raise_on_failure:
                raise _errors.SafenetExternalKeyServerError(err_msg)
            else:
                _log.error(err_msg)

    def enable(self):
        """Use to restore the connection between the SF Cluster and EKS."""
        _log.debug('Enable Safenets nae-server service')
        self.start_service('nae-server')

    def disable(self):
        """Use to sever the connection between the SF Cluster and EKS."""
        _log.debug('Disable Safenets nae-server service')
        self.stop_service('nae-server')

    def configure_for_testing(self, kmip_version=None):
        """Confirm EKS is ready for test execution.

        - Confirm server is running on the expected port number
        - Verify a local CA named KeySecure CA exists on the EKS.
        - Confirm KMIP is enabled

        We should also check that a user 'SolidFire' exists on the SafeNet EKS, but the CLI commands are
        limited.

        :param sf_models.enums.KmipVersion kmip_version: Changing kmip protocol is not implemented for
            SafeNet
        :raises valence.functional_models._errors.SafenetExternalKeyServerError: if an issue is found
            with the desired SafeNet settings
        """
        # Verify there is the expected local CA that other FM methods require
        self._check_for_local_ca()

        # Confirm KMIP is enabled
        self._confirm_kmip_enabled()

        # Now check that we're running on the correct port
        self._confirm_kmip_port()

        if kmip_version is not None:
            raise _errors.SafenetExternalKeyServerError('Cannot change A SafeNet EKS kmip version')

    def _check_for_local_ca(self):
        """Check for KeySecure CA.

        :raises valence.functional_models._errors.SafenetExternalKeyServerError: if an issue is found
            with the desired SafeNet settings
        """
        _log.debug('Checking for local CA called KeySecure CA')
        local_cas = self.show_local_ca()
        for ca in local_cas:
            if ca.name == 'KeySecure CA':
                break
        else:
            ca_issuers = ', '.join(str(ca.issuer) for ca in local_cas)
            msg = 'No local CA named KeySecure CA. {}'.format(ca_issuers)
            raise _errors.SafenetExternalKeyServerError(msg)

    def _confirm_kmip_enabled(self):
        """Make sure KMIP service is enabled."""
        _log.debug('Verifying KMIP service is running')
        if not self.is_running():
            _log.debug('KMIP was not enabled, enabling the SafeNet server.')
            self.enable()

    def _confirm_kmip_port(self):
        """Make sure KMIP port is expected value."""
        nae_servers = self.show_nae_server_settings()
        for server in nae_servers:
            if server.protocol == 'KMIP':
                if server.port != self.port:
                    _log.debug('Changing KMIP port from {} to {}'.format(server.port, self.port))
                    self.modify_nae_server_port(server.ip, server.port, self.port)

    def is_running(self):
        """Return True if the KMIP service or EKS is up and running.

        :rtype: bool
        """
        exists = False
        started = False

        # First check that KMIP is part of nae server.
        nae_servers = self.show_nae_server_settings()
        for server in nae_servers:
            if server.protocol == 'KMIP':
                exists = True
                break

        # Then check that nae server is started.
        services = self.show_services()
        for service in services:
            if service.service_group == 'nae-server' and service.status == 'Started':
                started = True
                break

        return True if exists and started else False

    # SSH Methods
    def get_ssh_connection(self):
        """Override to return custom ShellHandler.

        :rtype: valence.functional_models._external_key_server.safenet.SafenetShellHandler
        """
        return SafenetShellHandler(self.hostname, self.username, self.password)

    def run_command(self, command, timeout=60, raise_on_failure=False, use_sudo=False):
        """Override to use custom ShellHandler.

        The Safenet shell does not write errors to stderr and it does not return an exit
        status for commands executed. We cannot determine if the command passed or failed until
        we parse the output of stdout. To be compatible with other FMS, we will return a
        RunCommandOutput object, but the stderr and return_code values are useless.

        :param str command: Command to execute
        :param int timeout: Socket timeout
        :param bool raise_on_failure: Unused parameter. Safenet shell doesn't return exit status.
        :param bool use_sudo: Unused parameter. Sudo doesn't exist in Safenet shell.
        :rtype: sf_ssh.ssh.RunCommandOutput
        """
        stdout = self.run_interactive_command([command], timeout=timeout)
        return sf_ssh.RunCommandOutput(stdout=stdout, stderr='', return_code=0)

    def run_interactive_command(self, commands, timeout=60):
        """Execute the list of strings on the shell.

        This method is not for executing multiple shell commands on the CLI. Many of the Safenet CLI
        commands are interactive. They require multiple, sequential input strings. In this case,
        each string should be included as a separate entry in the cmds list.

        :param list[str] commands: List of strings to be sent to the shell
        :param int timeout: Socket timeout
        :rtype: str
        """
        with self.get_ssh_connection() as ssh:
            return ssh.execute(commands, timeout=timeout)


class SafenetShellHandler:
    """Handles all interactions with the Safenet shell over SSH.

    :param str host: Hostname
    :param str user: Username
    :param str psw: Password
    """
    def __init__(self, host, user, psw):
        self.host = host
        self.username = user
        self.password = psw
        self.ssh = None
        self.channel = None
        self.prompt = None
        self._init_client()

    def __enter__(self):
        """Return self."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Call my close method."""
        self.close()

    def execute(self, cmds, timeout=60):
        """Execute the list of strings one after another and return a concatenated stdout.

        :param list[str] cmds: List of commands to execute
        :param int timeout: Socket timeout
        :rtype: str
        """
        self._set_timeout(timeout)

        output = ''
        for cmd in cmds:
            _log.debug('Sending {}'.format(cmd))
            self._send(cmd)
            output += self.channel.recv(4096)

        stdout = []
        if output:
            for line in map(str.strip, output.split('\n')):
                if line.startswith(self.prompt):
                    break
                else:
                    stdout.append(line)

        # First line is always the command being executed, get rid of it.
        return '\n'.join(stdout[1:])

    def close(self):
        """Close down the ssh connection."""
        _log.debug('Closing ssh connection with {}'.format(self))
        self.ssh.close()

    def _init_client(self):
        """Create the paramiko client and and prepare the channel."""
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.connect(self.host, username=self.username, password=self.password, port=22)

        self.channel = self.ssh.invoke_shell()
        while not self.channel.recv_ready():
            time.sleep(1)

        # We're at the prompt now, clear stdout
        self.channel.recv(4096)

        # Put the shell into config mode and save the prompt
        self.prompt = self._prepare()

    def _prepare(self):
        """Send the config command and return the new prompt text.

        The prompt will be used by execute() to determine when a command is finished executing.

        :rtype: str
        """
        self._send('config')
        output = self.channel.recv(4096)
        for line in map(str.strip, output.split('\n')):
            if line and not line.endswith('config'):
                return line

    def _set_timeout(self, timeout):
        """Set the socket timeout value.

        :param int timeout: Socket timeout
        """
        channel_timeout = self.channel.gettimeout()
        if channel_timeout is None or int(channel_timeout) != timeout:
            self.channel.settimeout(float(timeout))

    def _send(self, cmd):
        """Send the command to the ssh channel and wait for it to be ready to read.

        :param str cmd: Command to execute
        """
        self.channel.send(cmd+'\n')
        while not self.channel.recv_ready():
            time.sleep(1)
