"""Unit tests for Cryptsoft Functional Models."""

import sys
import time
import unittest

import mock
import pytest
import sf_ssh
from cryptography import x509

from valence import (
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors

DATA_DIR = unittest_utils.get_data_dir(__file__)
module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestCryptsoft(unittest.TestCase):
    def setUp(self):
        self.eks_fm = functional_models.CryptsoftExternalKeyServer.get_mock_fm()

        if sys.version_info >= (3, 0):
            self.mock_open_str = 'builtins.open'
        else:
            self.mock_open_str = '__builtin__.open'

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'is_running')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, '_add_cryptsoft_service_files')
    def test_deploy_cryptsoft_server_running(self, mock_add_files, mock_is_running):
        mock_is_running.return_value = True
        self.eks_fm.deploy_cryptsoft_server()
        mock_add_files.assert_not_called()
        assert self.eks_fm.deployed is True

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'is_running')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, '_add_cryptsoft_service_files')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, '_generate_server_certs')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    def test_deploy_cryptsoft_server(self, mock_run_multiple_cmds, mock_add_files, mock_generate_certs,
                                     mock_is_running):
        mock_is_running.return_value = False
        self.eks_fm.deploy_cryptsoft_server()
        mock_add_files.assert_called()
        mock_generate_certs.assert_called()
        mock_run_multiple_cmds.assert_called_with(mock.ANY, raise_on_failure=True)
        assert self.eks_fm.deployed is True

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_teardown_cryptsoft_server(self, mock_run_cmd):
        self.eks_fm.teardown_cryptsoft_server()
        mock_run_cmd.assert_called_with(mock.ANY, raise_on_failure=True)
        assert self.eks_fm.deployed is False

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    def test_add_cryptsoft_service_files_ubuntu(self, mock_run_multiple_cmds, mock_run_cmd):
        """Just test that run_multiple_commands is called with raise_on_failure=True.

        Adding cryptsoft files runs a series of linux commands on a client. Since run_multiple_commands
        doesn't check output, there isn't much to test that is appropriate for a UT.
        """
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(stdout='Ubuntu', stderr='', return_code=0)
        self.eks_fm._add_cryptsoft_service_files()
        mock_run_multiple_cmds.assert_called_with(mock.ANY, raise_on_failure=True)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    def test_add_cryptsoft_service_files_centos(self, mock_run_multiple_cmds, mock_run_cmd):
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(stdout='CentOS', stderr='', return_code=0)
        self.eks_fm._add_cryptsoft_service_files()
        mock_run_multiple_cmds.assert_called_with(mock.ANY, raise_on_failure=True)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_add_cryptsoft_service_files_unsupported(self, mock_run_cmd):
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(stdout='Windows', stderr='', return_code=0)
        with pytest.raises(_errors.ExternalKeyServerError):
            self.eks_fm._add_cryptsoft_service_files()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    def test_generate_server_certs(self, mock_run_multiple_cmds):
        """Just test that run_multiple_commands is called with raise_on_failure=True.

        Generating server certs runs a series of rigid linux commands. Since run_multiple_commands
        doesn't check output, there isn't much to test that is appropriate for a UT.
        """
        self.eks_fm._generate_server_certs()
        mock_run_multiple_cmds.assert_called_with(mock.ANY, raise_on_failure=True)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_server_pid')
    def test_is_running(self, mock_get_server_pid):
        mock_get_server_pid.side_effect = [1, None]
        assert self.eks_fm.is_running() is True
        assert self.eks_fm.is_running() is False

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_get_server_pid(self, mock_run_cmd):
        stdout = 'tcp        0      0 0.0.0.0:5696            0.0.0.0:*               LISTEN' \
                 '      9649/kmip_server\n'
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(stdout=stdout, stderr='', return_code=0)
        assert self.eks_fm.get_server_pid() == 9649

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_get_server_pid_none(self, mock_run_cmd):
        stdout = ''
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(stdout=stdout, stderr='', return_code=0)
        assert self.eks_fm.get_server_pid() is None

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_server_pid')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_power_on(self, mock_run_cmd, mock_get_server_pid):
        mock_get_server_pid.return_value = None
        self.eks_fm.power_on()
        mock_run_cmd.assert_called()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_server_pid')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_power_on_fails(self, mock_run_cmd, mock_get_server_pid):
        """Fails meaning that the fm is already on."""
        mock_get_server_pid.return_value = 1
        self.eks_fm.power_on()
        mock_run_cmd.assert_not_called()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_server_pid')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_power_off(self, mock_run_cmd, mock_get_server_pid):
        mock_get_server_pid.return_value = 1
        self.eks_fm.power_off()
        mock_run_cmd.assert_called()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_server_pid')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_power_off_fails(self, mock_run_cmd, mock_get_server_pid):
        """Fails meaning that the fm is already off."""
        mock_get_server_pid.return_value = None
        self.eks_fm.power_off()
        mock_run_cmd.assert_not_called()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_on')
    def test_enable_implemented(self, mock_power_on):
        self.eks_fm.enable()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_off')
    def test_disable_implemented(self, mock_power_off):
        self.eks_fm.disable()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_ssh_connection')
    def test_get_logs(self, mock_ssh):
        self.eks_fm.get_logs()
        mock_ssh.return_value.__enter__.return_value.get.assert_called()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch('sf_artifacts.managers.GlobalManager')
    def test_store_logs(self, mock_manager, mock_run_command):
        output = sf_ssh.RunCommandOutput(stdout='/sf/core1/test.txt', stderr='', return_code=0)
        mock_run_command.return_value = output
        mock_client = models.ResourceClient.get_mock_object()
        self.eks_fm.store_logs(mock_client)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch('sf_artifacts.managers.GlobalManager')
    def test_store_logs_not_found(self, mock_manager, mock_run_command):
        output = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=2)
        mock_run_command.return_value = output
        mock_client = models.ResourceClient.get_mock_object()
        self.eks_fm.store_logs(mock_client)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'enable')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'is_running')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_server_pid')
    def test_configure_for_testing(self, mock_get_pid, mock_run_cmd, mock_is_running, mock_enable):
        mock_is_running.return_value = False
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        mock_get_pid.side_effect = [555, None]
        self.eks_fm.configure_for_testing()
        mock_enable.assert_called()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_ssh_connection')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    @mock.patch.object(x509, 'load_pem_x509_certificate')
    def test_sign_csr(self, mock_x509_cert_load, mock_x509_csr_load, mock_run_cmd,
                      mock_run_multiple_cmds, mock_ssh):
        # Create the mock CSR and put in the mock CSR text
        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        with open('{}/mock.csr'.format(DATA_DIR), 'rb') as csr:
            mock_x509_csr.public_bytes.return_value = csr.read()

        # Create a mock subject and put in the mock OU field
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_x509_csr_load.return_value = mock_x509_csr

        mock_x509_cert = mock.MagicMock(x509.Certificate)
        mock_x509_cert_load.return_value = mock_x509_cert

        with mock.patch(self.mock_open_str, mock.mock_open()):
            client_cert, ca_cert = self.eks_fm.sign_csr(mock_x509_csr)

        mock_ssh.return_value.__enter__.return_value.put.assert_called()
        mock_ssh.return_value.__enter__.return_value.get.assert_called()
        assert isinstance(client_cert, x509.Certificate)
        assert isinstance(ca_cert, x509.Certificate)
        assert len(self.eks_fm.signed_csrs) == 1
        assert len(self.eks_fm.signed_certs) == 1

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'is_running')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_on')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_off')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_ssh_connection')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    @mock.patch.object(x509, 'load_pem_x509_certificate')
    def test_sign_csr_named(self, mock_x509_cert_load, mock_x509_csr_load, mock_run_cmd,
                            mock_run_multiple_cmds, mock_ssh, mock_off, mock_on, mock_running):
        """Test that it can be called multiple times with names and maintains previously signed certs."""
        # Create the mock CSR and put in the mock CSR text
        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        with open('{}/mock.csr'.format(DATA_DIR), 'rb') as csr:
            mock_x509_csr.public_bytes.return_value = csr.read()

        # Create a mock subject and put in the mock OU field
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_x509_csr_load.return_value = mock_x509_csr

        mock_x509_cert = mock.MagicMock(x509.Certificate)
        mock_x509_cert_load.return_value = mock_x509_cert

        mock_running.return_value = True

        with mock.patch(self.mock_open_str, mock.mock_open()):
            self.eks_fm.sign_csr(mock_x509_csr, name='test_runner')
            self.eks_fm.sign_csr(mock_x509_csr, name='element')

        assert len(self.eks_fm.signed_csrs) == 2
        assert len(self.eks_fm.signed_certs) == 2

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'is_running')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_on')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_off')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_ssh_connection')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_multiple_commands')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    @mock.patch.object(x509, 'load_pem_x509_certificate')
    def test_sign_csr_force(self, mock_x509_cert_load, mock_x509_csr_load, mock_run_cmd,
                            mock_run_multiple_cmds, mock_ssh, mock_off, mock_on, mock_running):
        # Create the mock CSR and put in the mock CSR text
        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        with open('{}/mock.csr'.format(DATA_DIR), 'rb') as csr:
            mock_x509_csr.public_bytes.return_value = csr.read()

        # Create a mock subject and put in the mock OU field
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_x509_csr_load.return_value = mock_x509_csr

        mock_x509_cert = mock.MagicMock(x509.Certificate)
        mock_x509_cert_load.return_value = mock_x509_cert
        mock_running.return_value = True

        ex = _errors.ExternalKeyServerError('username are not unique')
        mock_run_cmd.side_effect = [ex]

        with mock.patch(self.mock_open_str, mock.mock_open()):
            client_cert, ca_cert = self.eks_fm.sign_csr(mock_x509_csr, force=True)

        mock_ssh.return_value.__enter__.return_value.put.assert_called()
        mock_ssh.return_value.__enter__.return_value.get.assert_called()
        assert isinstance(client_cert, x509.Certificate)
        assert isinstance(ca_cert, x509.Certificate)
        assert len(self.eks_fm.signed_csrs) == 1
        assert len(self.eks_fm.signed_certs) == 1

        ex = _errors.ExternalKeyServerError('different exception')
        mock_run_cmd.side_effect = [ex]

        with mock.patch(self.mock_open_str, mock.mock_open()):
            with pytest.raises(_errors.ExternalKeyServerError):
                client_cert, ca_cert = self.eks_fm.sign_csr(mock_x509_csr, force=True)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'run_command')
    def test_execute_sql(self, mock_run_cmd):
        mock_uuid = '680c8cec-ff90-4eeb-a890-36a2557176bd'
        mock_run_cmd.return_value = sf_ssh.RunCommandOutput(
            stdout=mock_uuid,
            stderr='',
            return_code=0
        )

        sql_query = 'SELECT json from obj;'
        output = self.eks_fm.execute_sql(sql_query, commit=True)
        assert output == mock_uuid

        mock_cmd = 'cd {}; ./k_db sql="{}"'.format(self.eks_fm.kmip_bin, sql_query)
        mock_run_cmd.assert_any_call(mock_cmd, raise_on_failure=True)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'execute_sql')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, '_get_object_id')
    def test_get_key(self, mock_get_obj_id, mock_execute_sql):
        mock_get_obj_id.return_value = 1
        mock_key_value = 'zuq935ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_execute_sql.return_value = 'KEY_MATERIAL:bstr:{}'.format(mock_key_value)
        key = self.eks_fm.get_key('680c8cec-ff90-4eeb-a890-36a2557176bd')
        assert key == mock_key_value

        mock_cmd = 'SELECT tagtext FROM obj WHERE id=1;'
        mock_execute_sql.assert_any_call(mock_cmd)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'execute_sql')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, '_get_object_id')
    def test_get_key_fail(self, mock_get_obj_id, mock_execute_sql):
        mock_get_obj_id.return_value = 1
        mock_key_value = 'zuq935ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_execute_sql.return_value = 'KEY_MTERIAL:bstr:{}'.format(mock_key_value)
        with pytest.raises(_errors.ExternalKeyServerError):
            self.eks_fm.get_key('680c8cec-ff90-4eeb-a890-36a2557176bd')

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'execute_sql')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_key')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, '_get_object_id')
    def test_set_key(self, mock_get_obj_id, mock_get_key, mock_execute_sql):
        mock_old_key = 'regg35ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_new_key = 'zuq935ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_key_id = '680c8cec-ff90-4eeb-a890-36a2557176bd'
        mock_get_obj_id.return_value = 1
        mock_get_key.return_value = mock_old_key

        self.eks_fm.set_key(mock_key_id, mock_new_key)
        mock_cmd = "UPDATE obj SET value = REPLACE(value, '{x}', '{y}')," \
                   "json = REPLACE(json, '{x}', '{y}')," \
                   "tagtext = REPLACE(tagtext, '{x}', '{y}')," \
                   "xml = REPLACE(tagtext, '{x}', '{y}') WHERE id=1;"
        mock_execute_sql.assert_called_with(mock_cmd.format(x=mock_old_key, y=mock_new_key), commit=True)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'execute_sql')
    def test_get_object_id(self, mock_execute_sql):
        mock_key_id = '680c8cec-ff90-4eeb-a890-36a2557176bd'
        mock_execute_sql.return_value = 'oid=3'
        oid = self.eks_fm._get_object_id(mock_key_id)
        assert oid == 3

        oid_query = "SELECT oid FROM att WHERE name='Name' AND " \
                    "json LIKE '%x-SolidFire-KeyID-{}%';".format(mock_key_id)
        mock_execute_sql.assert_called_with(oid_query)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'execute_sql')
    def test_get_object_id_fail(self, mock_execute_sql):
        mock_execute_sql.return_value = 'od=3'
        with pytest.raises(_errors.ExternalKeyServerError):
            self.eks_fm._get_object_id('680c8cec-ff90-4eeb-a890-36a2557176bd')
