"""Unit tests for EKS Functional Models."""
import os
import ssl
import sys
import time
import unittest

import mock
import pytest
import sf_models
import sf_ssh
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from kmip.pie import client

from valence import (
    enums,
    functional_models,
    unittest_utils,
)
from valence.functional_models import _errors
from valence.functional_models._external_key_server import (
    base,
    hytrust,
)

DATA_DIR = unittest_utils.get_data_dir(__file__)

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestExternalKeyServerBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.eks = hytrust.HyTrustExternalKeyServer.get_mock_fm(overrides=dict(hostname='1.1.1.1'))

        if sys.version_info >= (3, 0):
            cls.mock_open_str = 'builtins.open'
        else:
            cls.mock_open_str = '__builtin__.open'

    @mock.patch.object(os.path, 'exists')
    @mock.patch.object(os, 'remove')
    def test_del(self, mock_remove, mock_exists):
        eks_fm = hytrust.HyTrustExternalKeyServer.get_mock_fm(overrides=dict(hostname='1.1.1.1'))
        eks_fm._pykmip_connected = True
        mock_exists.return_value = True
        eks_fm.ca_certificate = '/path/to/file'
        eks_fm.pykmip_client_certificate = '/path/to/file'
        eks_fm.pykmip_client_private_key = '/path/to/file'
        eks_fm.pykmip_client_csr = '/path/to/file'
        eks_fm.__del__()

    def test_properties(self):
        assert isinstance(self.eks.vendor, enums.ExternalKeyServerVendor)
        assert isinstance(self.eks.kmip_version, enums.KmipVersion)
        assert isinstance(self.eks.ssl_version, enums.SslVersion)
        self.eks.pykmip_username = 'user'
        assert self.eks.pykmip_username == 'user'
        self.eks.pykmip_password = 'pass'
        assert self.eks.pykmip_password == 'pass'
        assert self.eks.port == self.eks.eks_dm.port

    def test_get_ssh_connection(self):
        ssh = self.eks.get_ssh_connection()
        assert isinstance(ssh, sf_ssh.Ssh)

    @mock.patch.object(base.ExternalKeyServerBase, 'connect')
    def test_get_kmip_connection(self, mock_connect):
        self.eks._pykmip_connected = False
        self.eks.get_kmip_connection()

    @mock.patch.object(base.ExternalKeyServerBase, 'is_pykmip_trusted')
    @mock.patch.object(base.ExternalKeyServerBase, '_setup_pykmip_certificates')
    @mock.patch.object(client, 'ProxyKmipClient')
    def test_connect(self, mock_pykmip, mock_setup_certs, mock_trusted):
        mock_trusted.return_value = False
        pykmip = mock.MagicMock()
        mock_pykmip.return_value = pykmip

        with mock.patch(self.mock_open_str, mock.mock_open()) as mock_open:
            mock_open.return_value = pykmip
            pykmip.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
            pykmip.__exit__ = mock.Mock(return_value=None)
            self.eks.connect()

    @mock.patch.object(base.ExternalKeyServerBase, 'is_pykmip_trusted')
    @mock.patch.object(base.ExternalKeyServerBase, '_setup_pykmip_certificates')
    @mock.patch.object(client, 'ProxyKmipClient')
    def test_connect_error(self, mock_pykmip, mock_setup_certs, mock_trusted):
        mock_trusted.return_value = False
        pykmip = mock.MagicMock()
        mock_pykmip.return_value = pykmip

        with mock.patch(self.mock_open_str, mock.mock_open()) as mock_open:
            mock_open.return_value = pykmip
            pykmip.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
            pykmip.__enter__.side_effect = [ssl.SSLError]
            pykmip.__exit__ = mock.Mock(return_value=None)
            with pytest.raises(ssl.SSLError):
                self.eks.connect()

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_run_command(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='hi',
                                                                    stderr='',
                                                                    return_code=0)
        output = self.eks.run_command('echo hi')
        assert output.stdout == 'hi'

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_run_command_failed(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='',
                                                                    stderr='error',
                                                                    return_code=1)
        with pytest.raises(_errors.Error):
            self.eks.run_command('echo hi', raise_on_failure=True)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_run_multiple_commands(self, mock_multiple_commands):
        mock_multiple_commands.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='hi', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='hello', stderr='', return_code=0),
        ]
        self.eks.run_multiple_commands(['echo hi', 'echo hello'])
        calls = [mock.call('echo hi', 300, use_sudo=False), mock.call('echo hello', 300, use_sudo=False)]
        mock_multiple_commands.assert_has_calls(calls)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_run_multiple_commands_failed(self, mock_multiple_commands):
        mock_multiple_commands.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='hi', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='', stderr='error', return_code=1),
        ]
        with pytest.raises(_errors.Error):
            self.eks.run_multiple_commands(['echo hi', 'nope'], raise_on_failure=True)

    def test_is_pykmip_trusted(self):
        self.eks.pykmip_client_certificate = None
        assert self.eks.is_pykmip_trusted() is False

        self.eks.pykmip_client_certificate = 'path/to/cert'
        self.eks.pykmip_client_private_key = 'path/to/key'
        self.eks.ca_certificate = 'path/to/cert'
        assert self.eks.is_pykmip_trusted()

    @mock.patch.object(hytrust.HyTrustExternalKeyServer, 'sign_csr')
    @mock.patch.object(base.ExternalKeyServerBase, '_create_pykmip_private_key')
    @mock.patch.object(base.ExternalKeyServerBase, '_create_pykmip_csr')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    def test__setup_pykmip_certificates(self, mock_x509_csr_load, mock_create_csr,
                                        mock_create_private_key, mock_sign_csr):
        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        mock_x509_csr.public_bytes.return_value = 'fake cert'
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_x509_csr_load.return_value = mock_x509_csr

        mock_client_cert = mock.MagicMock(x509.Certificate)
        mock_ca_cert = mock.MagicMock(x509.Certificate)

        mock_create_private_key.return_value = 'path/to/key'
        mock_create_csr.return_value = 'path/to/cert'
        mock_sign_csr.return_value = (mock_client_cert, mock_ca_cert)
        with mock.patch(self.mock_open_str, mock.mock_open(read_data='fake cert')):
            self.eks._setup_pykmip_certificates()
        assert self.eks.pykmip_client_private_key is not None
        assert self.eks.pykmip_client_certificate is not None
        assert self.eks.ca_certificate is not None

    def test_create_private_key(self):
        self.eks._create_pykmip_private_key()

    @mock.patch.object(x509.CertificateSigningRequestBuilder, 'sign')
    @mock.patch.object(serialization, 'load_pem_private_key')
    def test_create_csr(self, mock_load_pem_private_key, mock_sign):
        with open('{}/mock.key'.format(DATA_DIR), 'rb') as key:
            mock_key_str = key.read()

        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        mock_x509_csr.public_bytes.return_value = 'fake cert'
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_sign.return_value = mock_x509_csr

        with mock.patch(self.mock_open_str, mock.mock_open(read_data=mock_key_str)):
            self.eks._create_pykmip_csr()


class TestExternalKeyServerFactory(unittest.TestCase):
    def test_get_eks_from_dm(self):
        hytrust_dm = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.hytrust))
        hytrust_fm = functional_models.get_eks_from_dm(hytrust_dm)
        assert isinstance(hytrust_fm, functional_models.HyTrustExternalKeyServer)

        safenet_dm1 = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.gemalto_safenet))
        safenet_fm1 = functional_models.get_eks_from_dm(safenet_dm1)
        assert isinstance(safenet_fm1, functional_models.SafenetExternalKeyServer)

        safenet_dm2 = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.gemalto_safenet_150v))
        safenet_fm2 = functional_models.get_eks_from_dm(safenet_dm2)
        assert isinstance(safenet_fm2, functional_models.SafenetExternalKeyServer)

        safenet_dm3 = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.safenet_at_350v))
        safenet_fm3 = functional_models.get_eks_from_dm(safenet_dm3)
        assert isinstance(safenet_fm3, functional_models.SafenetExternalKeyServer)

        cryptsoft_dm = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.cryptsoft))
        cryptsoft_fm = functional_models.get_eks_from_dm(cryptsoft_dm)
        assert isinstance(cryptsoft_fm, functional_models.CryptsoftExternalKeyServer)

        vormetric_dm = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.vormetric))
        vormetric_fm = functional_models.get_eks_from_dm(vormetric_dm)
        assert isinstance(vormetric_fm, functional_models.VormetricExternalKeyServer)
