"""Unit tests for HyTrust Functional Models."""
import os
import sys
import time
import unittest
import zipfile

import mock
import pytest
import requests
from cryptography import x509

from valence import (
    enums,
    functional_models,
    unittest_utils,
)
from valence.utils import random_string

DATA_DIR = unittest_utils.get_data_dir(__file__)
module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


@mock.patch.object(functional_models.HyTrustExternalKeyServer, '_call_method')
class TestHyTrustAPI(unittest.TestCase):
    def setUp(self):
        self.eks_fm = functional_models.HyTrustExternalKeyServer.get_mock_fm()

        if sys.version_info >= (3, 0):
            self.mock_open_str = 'builtins.open'
        else:
            self.mock_open_str = '__builtin__.open'

        # Mock group
        mock_group_name = 'Mock Group Name'
        mock_user_name = 'MockUser'
        self.mock_group = dict(
                ac_policy_max_changes=50,
                adgroup_members=[],
                cvmset_list=[],
                desc='Mock Description',
                id=mock_group_name,
                indestructible='yes',
                name=mock_group_name,
                type='CLOUD_ADMIN',
                uri='{}v5/groups/{}'.format(self.eks_fm.root_uri, mock_group_name.replace(' ', '%20')),
                user_members=[dict(
                    id=mock_user_name,
                    name=mock_user_name,
                    uri='{}v5/users/{}'.format(self.eks_fm.root_uri, mock_user_name)
                )]
            )

        # Mock user
        mock_username = 'root'
        self.mock_user = dict(
            authentication='Local',
            id=mock_username,
            last_login='2019-05-28 16:20:13 UTC',
            login_name=mock_username,
            name='Root Admin',
            privileges=['SEC_ADMIN', 'DOMAIN_ADMIN', 'CLOUD_ADMIN'],
            status='Enabled',
            uri='{}v5/users/{}'.format(self.eks_fm.root_uri, mock_username)
        )
        self.mock_user['2fa_enabled'] = False

        # Mock KMIP object
        self.mock_kmip_obj = dict(
            ATTRS=dict(
                State='Preactive'
            ),
            DESC='Description',
            IDENT=[],
            UUID='5cccd76d-5c67-11e9-8c6e-0050569a80ef'
        )
        self.mock_kmip_server = dict(
            result='success',
            settings=dict(
                kmipsrv_info=dict(
                    host='1.1.1.1',
                    loglevel='ALL',
                    nbio='1',
                    port='5696',
                    protocol='0x800000',
                    reconnect='1',
                    state='ENABLED',
                    timeout='0',
                    uri='https://1.1.1.1/v5/system_settings/kmip_server/',
                    verify='yes'
                )
            )
        )

    @mock.patch.object(requests, 'post')
    def test_login_workflow(self, mock_post, mock_call_method):
        mock_access_token = random_string.generate_random_string(64)
        mock_resp = dict(
            access_token=mock_access_token,
            result='success',
            username=self.eks_fm.username
        )

        # Login
        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 200
        requests_response.json.return_value = mock_resp
        mock_post.return_value = requests_response
        resp = self.eks_fm.login()
        assert self.eks_fm.access_token == mock_access_token
        assert resp == mock_resp

        # Renew
        mock_call_method.return_value = mock_resp
        resp = self.eks_fm.renew_token()
        assert self.eks_fm.access_token == mock_access_token
        assert resp == mock_resp

        # Logout
        mock_resp = {'result': 'success'}
        mock_call_method.return_value = mock_resp
        resp = self.eks_fm.logout()
        assert self.eks_fm.access_token is None
        assert resp == mock_resp

    @mock.patch.object(requests, 'post')
    def test_login_fail(self, mock_post, mock_call_method):
        mock_resp = dict(result='failure')
        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 403
        requests_response.json.return_value = mock_resp
        mock_post.return_value = requests_response

        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm.login()

    def test_get_admin_key(self, mock_call_method):
        with open('{}/mock.key'.format(DATA_DIR), 'rb') as key:
            mock_key = key.read()

        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 200
        requests_response.content.return_value = mock_key
        mock_call_method.return_value = requests_response

        target_path = '/mock.key'
        with mock.patch(self.mock_open_str, mock.mock_open(read_data=mock_key)):
            file_path = self.eks_fm.get_admin_key(target=target_path)
        assert file_path == target_path

    def test_get_alerts(self, mock_call_method):
        mock_alert_id = '9730201c-5c55-11e9-a652-0050569a80ef'
        mock_resp = dict(
            alerts=[dict(
                date='2019-04-11 12:30:29 UTC',
                id=mock_alert_id,
                message='Alert message',
                uri='{}v5/alerts/{}'.format(self.eks_fm.root_uri, mock_alert_id)
            )],
            result='success'
        )
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.get_alerts()
        assert resp == mock_resp

    def test_delete_alerts(self, mock_call_method):
        mock_alert_id = '9730201c-5c55-11e9-a652-0050569a80ef'
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.delete_alerts(alert_ids=mock_alert_id)
        assert resp == mock_resp

    def test_list_groups(self, mock_call_method):
        mock_resp = dict(
            groups=[self.mock_group],
            result='success'
        )
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.list_groups()
        assert resp == mock_resp

    def test_get_group_details(self, mock_call_method):
        mock_resp = self.mock_group
        mock_resp['result'] = 'success'
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.get_group_details(group_id=self.mock_group['id'])
        assert resp == mock_resp

    def test_create_group(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.create_group(name='Test group', privilege='CLOUD_ADMIN',
                                        desc='Test group', user_members=[{'id': 'root'}])
        assert resp == mock_resp

    def test_create_group_privilege(self, mock_call_method):
        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm.create_group(name='Test group', privilege='FAKE',
                                     desc='Test group', user_members=[{'id': 'root'}])

    def test_edit_group(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.edit_group(group_id='Test group', name='New Test Group')
        assert resp == mock_resp

    def test_edit_group_privilege(self, mock_call_method):
        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm.edit_group(group_id='Test group', privilege='FAKE')

    def test_delete_group(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.delete_group(group_id='New Test Group')
        assert resp == mock_resp

    def test_list_users(self, mock_call_method):
        mock_resp = dict(
            users=[self.mock_user],
            result='success'
        )
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.list_users()
        assert resp == mock_resp

    def test_get_user_details(self, mock_call_method):
        mock_resp = self.mock_user
        mock_resp['result'] = 'success'
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.get_user_details(user_id='root')
        assert resp == mock_resp

    def test_change_default_user_settings(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.change_default_user_settings(default_type='LDAP')
        assert resp == mock_resp

    def test_create_user(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.create_user(login_name='user', password='pass', full_name='Username',
                                       email='user@email.com', privileges=['CLOUD_ADMIN'],
                                       groups=['Test Group'])
        assert resp == mock_resp

    def test_edit_user(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.edit_user(user_id='user', password='pass', full_name='Username',
                                     email='user@email.com', privileges=['CLOUD_ADMIN'],
                                     groups=['Test Group'])
        assert resp == mock_resp

    def test_delete_user(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.delete_user(user_id='user')
        assert resp == mock_resp

    def test_get_kmip_server_implemented(self, mock_call_method):
        mock_resp = dict(result='success', settings=dict())
        mock_call_method.return_value = mock_resp
        resp = self.eks_fm.get_kmip_server()
        assert isinstance(resp, dict)

    def test_enable_kmip_server(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.enable_kmip_server(host='1.1.1.1', nbio=True, port=5696,
                                              protocol=enums.KmipVersion.proto_1_4, reconnect=True,
                                              state='ENABLED', timeout=0, verify=True)
        assert resp == mock_resp

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'enable_kmip_server')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'get_kmip_server')
    def test_enable_fails(self, mock_get_kmip_server, mock_enable_mip_server, mock_call_method):
        mock_get_kmip_server.return_value = self.mock_kmip_server
        mock_enable_mip_server.return_value = dict(result='failure')
        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm.enable()

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'enable_kmip_server')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'get_kmip_server')
    def test_disable_fails(self, mock_get_kmip_server, mock_enable_kmip_server, mock_call_method):
        mock_get_kmip_server.return_value = self.mock_kmip_server
        mock_enable_kmip_server.return_value = dict(result='failure')
        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm.disable()

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'enable')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'get_kmip_server')
    def test_configure_for_testing(self, mock_get_kmip_server, mock_enable, mock_call_method):
        mock_get_kmip_server.return_value = dict(
            result='success',
            settings=dict(
                kmipsrv_info=dict(
                    host='1.1.1.1',
                    loglevel='ALL',
                    nbio='1',
                    port=str(self.eks_fm.port),
                    protocol='0x800000',
                    reconnect='1',
                    state='DISABLED',
                    timeout='0',
                    uri='https://1.1.1.1/v5/system_settings/kmip_server/',
                    verify='yes'
                )
            )
        )
        self.eks_fm.configure_for_testing()
        mock_enable.assert_called()

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'enable_kmip_server')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'get_kmip_server')
    def test_configure_for_testing_kmip_change(self, mock_get_kmip_server, mock_enable_kmip_server,
                                               mock_call_method):
        mock_get_kmip_server.return_value = dict(
            result='success',
            settings=dict(
                kmipsrv_info=dict(
                    host='1.1.1.1',
                    loglevel='ALL',
                    nbio='1',
                    port='5691',
                    protocol='0x1',
                    reconnect='1',
                    state='ENABLED',
                    timeout='0',
                    uri='https://1.1.1.1/v5/system_settings/kmip_server/',
                    verify='yes'
                )
            )
        )
        mock_enable_kmip_server.return_value = dict(
            result='success',
            settings=dict(
                kmipsrv_info=dict(
                    host='1.1.1.1',
                    loglevel='ALL',
                    nbio='1',
                    port='5696',
                    protocol='0x800000',
                    reconnect='1',
                    state='ENABLED',
                    timeout='0',
                    uri='https://1.1.1.1/v5/system_settings/kmip_server/',
                    verify='yes'
                )
            )
        )
        self.eks_fm.configure_for_testing()
        self.eks_fm.configure_for_testing(kmip_version=enums.KmipVersion.proto_1_4)
        mock_enable_kmip_server.assert_called()

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'enable_kmip_server')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'get_kmip_server')
    def test_configure_for_testing_kmip_change_and_fail(self, mock_get_kmip_server,
                                                        mock_enable_kmip_server, mock_call_method):
        mock_get_kmip_server.return_value = dict(
            result='success',
            settings=dict(
                kmipsrv_info=dict(
                    host='1.1.1.1',
                    loglevel='ALL',
                    nbio='1',
                    port='5696',
                    protocol='0x1',
                    reconnect='1',
                    state='ENABLED',
                    timeout='0',
                    uri='https://1.1.1.1/v5/system_settings/kmip_server/',
                    verify='yes'
                )
            )
        )
        mock_enable_kmip_server.return_value = dict(result='failed')
        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm.configure_for_testing(kmip_version=enums.KmipVersion.proto_1_4)

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'get_kmip_server')
    def test_is_running(self, mock_get_kmip_server, mock_call_method):
        mock_server = dict(
            result='success',
            settings=dict(
                kmipsrv_info=dict(
                    host='1.1.1.1',
                    loglevel='ALL',
                    nbio='1',
                    port='5696',
                    protocol='0x1',
                    reconnect='1',
                    state='ENABLED',
                    timeout='0',
                    uri='https://1.1.1.1/v5/system_settings/kmip_server/',
                    verify='yes'
                )
            )
        )
        mock_get_kmip_server.return_value = mock_server
        assert self.eks_fm.is_running()

        mock_server['settings']['kmipsrv_info']['state'] = 'DISABLED'
        mock_get_kmip_server.return_value = mock_server
        assert self.eks_fm.is_running() is False

    def test_create_kmip_certificate_bundle(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        with open('{}/mock.csr'.format(DATA_DIR), 'r') as key:
            mock_csr = key.read()

        resp = self.eks_fm.create_kmip_certificate_bundle(cert_id='fakeid', expire=365, csr=mock_csr)
        assert resp == mock_resp

    @mock.patch.object(zipfile, 'ZipFile')
    def test_download_kmip_certificate_bundle(self, mock_zip, mock_call_method):
        mock_cert_id = 'SolidFire' + self.eks_fm._unique_id
        mock_zip.return_value = MockZipFile([mock_cert_id + '.pem', 'cacert.pem'])

        mock_zip = bytearray(os.urandom(1000))
        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 200
        requests_response.content = mock_zip
        mock_call_method.return_value = requests_response

        with mock.patch(self.mock_open_str, mock.mock_open()):
            self.eks_fm.download_kmip_certificate_bundle(cert_id='fakeid')

    def test_edit_kmip_certificate(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.edit_kmip_certificate(cert_id='fakeid', expire=1)
        assert resp == mock_resp

    def test_delete_kmip_certificate(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.delete_kmip_certificate(cert_id='fakeid', username='root')
        assert resp == mock_resp

    def test_delete_all_kmip_certificates(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.delete_all_kmip_certificates()
        assert resp == mock_resp

    def test_list_kmip_objects(self, mock_call_method):
        mock_resp = dict(
            lastref=0,
            objcount=1,
            result='success',
            total_objects=1
        )
        mock_resp['return'] = [self.mock_kmip_obj]
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.list_kmip_objects(attr='State')
        assert resp == mock_resp

    def test_get_kmip_object(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_resp['return'] = self.mock_kmip_obj
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.get_kmip_object(obj_uuid=self.mock_kmip_obj['UUID'])
        assert resp == mock_resp

    def test_activate_kmip_object(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.activate_kmip_object(obj_uuid=self.mock_kmip_obj['UUID'])
        assert resp == mock_resp

    def test_archive_kmip_object(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.archive_kmip_object(obj_uuid=self.mock_kmip_obj['UUID'])
        assert resp == mock_resp

    def test_destroy_kmip_object(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.destroy_kmip_object(obj_uuid=self.mock_kmip_obj['UUID'])
        assert resp == mock_resp

    def test_recover_kmip_object(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.recover_kmip_object(obj_uuid=self.mock_kmip_obj['UUID'])
        assert resp == mock_resp

    def test_revoke_kmip_object(self, mock_call_method):
        mock_resp = dict(result='success')
        mock_call_method.return_value = mock_resp

        resp = self.eks_fm.revoke_kmip_object(obj_uuid=self.mock_kmip_obj['UUID'], revcode=2)
        assert resp == mock_resp


class TestHyTrust(unittest.TestCase):
    def setUp(self):
        self.eks_fm = functional_models.HyTrustExternalKeyServer.get_mock_fm()

        if sys.version_info >= (3, 0):
            self.mock_open_str = 'builtins.open'
        else:
            self.mock_open_str = '__builtin__.open'

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'create_kmip_certificate_bundle')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'download_kmip_certificate_bundle')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    @mock.patch.object(x509, 'load_pem_x509_certificate')
    def test_sign_csr(self, mock_x509_cert_load, mock_x509_csr_load, mock_download, mock_create):
        # Create the mock CSR and put in the mock CSR text
        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        with open('{}/mock.csr'.format(DATA_DIR), 'r') as csr:
            mock_x509_csr.public_bytes.return_value = csr.read()

        # Create a mock subject and put in the mock OU field
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_x509_csr_load.return_value = mock_x509_csr

        mock_x509_cert = mock.MagicMock(x509.Certificate)
        mock_x509_cert_load.return_value = mock_x509_cert
        mock_download.return_value = ('client cert', 'ca cert')

        with mock.patch(self.mock_open_str, mock.mock_open()):
            client_cert, ca_cert = self.eks_fm.sign_csr(mock_x509_csr)

        assert isinstance(client_cert, x509.Certificate)
        assert isinstance(ca_cert, x509.Certificate)

    @mock.patch.object(requests, 'request')
    def test_call_method(self, mock_request):
        self.eks_fm.access_token = 'not_none'

        mock_resp = {'id': 1}
        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 200
        requests_response.json.return_value = mock_resp
        mock_request.return_value = requests_response

        resp = self.eks_fm._call_method(self.eks_fm.root_uri, 'GET')
        assert resp == mock_resp

    @mock.patch.object(requests, 'request')
    def test_call_method_fail(self, mock_request):
        self.eks_fm.access_token = 'not_none'

        mock_resp = {'id': 1}
        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 500
        requests_response.json.return_value = mock_resp
        mock_request.return_value = requests_response

        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm._call_method(self.eks_fm.root_uri, 'GET')

    @mock.patch.object(requests, 'request')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'login')
    def test_call_method_login(self, mock_login, mock_request):
        self.eks_fm.access_token = None
        mock_login.side_effect = self.set_token

        mock_resp = {'id': 1}
        requests_response = mock.Mock(requests.Response)
        requests_response.status_code = 200
        requests_response.json.return_value = mock_resp
        mock_request.return_value = requests_response

        resp = self.eks_fm._call_method(self.eks_fm.root_uri, 'GET')
        assert self.eks_fm.access_token is not None
        assert resp == mock_resp

    @mock.patch.object(requests, 'request')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'login')
    def test_call_method_login_fail(self, mock_login, mock_request):
        self.eks_fm.access_token = 'not_none'

        mock_resp = {'id': 1}
        requests_response_401 = mock.Mock(requests.Response)
        requests_response_401.status_code = 401
        requests_response_401.json.return_value = mock_resp

        requests_response_200 = mock.Mock(requests.Response)
        requests_response_200.status_code = 200
        requests_response_200.json.return_value = mock_resp

        mock_request.side_effect = [requests_response_401, requests_response_200]
        mock_login.side_effect = self.set_token

        resp = self.eks_fm._call_method(self.eks_fm.root_uri, 'GET', params={'fake': 'param'})
        assert resp == mock_resp
        assert self.eks_fm.access_token == 'definitely_not_none'

    @mock.patch.object(requests, 'request')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'login')
    def test_call_method_login_double_fail(self, mock_login, mock_request):
        self.eks_fm.access_token = 'not_none'

        mock_resp = {'id': 1}
        requests_response_401 = mock.Mock(requests.Response)
        requests_response_401.status_code = 401
        requests_response_401.json.return_value = mock_resp

        requests_response_200 = mock.Mock(requests.Response)
        requests_response_200.status_code = 401
        requests_response_200.json.return_value = mock_resp

        mock_request.side_effect = [requests_response_401, requests_response_200]
        mock_login.side_effect = self.set_token

        with pytest.raises(functional_models._errors.HyTrustExternalKeyServerError):
            self.eks_fm._call_method(self.eks_fm.root_uri, 'GET')

    def set_token(self):
        self.eks_fm.access_token = 'definitely_not_none'


class MockZipFile:
    def __init__(self, files=[]):
        self.files = files

    def __iter__(self):
        return iter(self.files)

    def write(self, fname):
        self.files.append(fname)

    def read(self, fname):
        return 'fake output'
