"""Unit tests for Safenet Functional Models."""

import sys
import time
import unittest

import mock
import paramiko
import pytest
from cryptography import x509

from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors
from valence.functional_models._external_key_server import safenet

DATA_DIR = unittest_utils.get_data_dir(__file__)
module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


@mock.patch.object(safenet.SafenetExternalKeyServer, 'run_interactive_command')
class TestSafenetAPI(unittest.TestCase):
    def setUp(self):
        self.eks_fm = functional_models.SafenetExternalKeyServer.get_mock_fm()

        self.mock_local_ca_stdout = """PROPERTIES:
Certificate Name: KeySecure CA
Serial Number:    0x00 (0)
Key Size:         2048
Start Date:       May 12 16:28:23 2019 GMT
Expiration:       May 10 16:28:23 2029 GMT

Issuer:
C:              US
ST:             CO
L:              Boulder
O:              Netapp
OU:             SF
CN:             EKM
emailAddress:

Subject:
C:              US
ST:             CO
L:              Boulder
O:              Netapp
OU:             SF
CN:             EKM
emailAddress:

Purpose:
SSL client & CA
SSL server & CA
Netscape SSL server & CA
S/MIME signing & CA
S/MIME encryption & CA
CRL signing & CA
Time Stamp signing CA

TEXT:
-----BEGIN CERTIFICATE-----
MIIEFDC...CAvygAwIB
-----END CERTIFICATE-----
"""

        mock_subject = models.CertificateSubject.create(
            common_name='EKM',
            country='US',
            email_address='',
            locality='Boulder',
            organization='Netapp',
            organizational_unit='SF',
            state='CO',
        )
        self.mock_local_ca_dm = models.SafenetLocalCertificateAuthority.create(
            name='KeySecure CA',
            expiration='May 10 16:28:23 2029 GMT',
            key_size=2048,
            serial_number='0x00 (0)',
            start_date='May 12 16:28:23 2019 GMT',
            purpose=[
                'SSL client & CA',
                'SSL server & CA',
                'Netscape SSL server & CA',
                'S/MIME signing & CA',
                'S/MIME encryption & CA',
                'CRL signing & CA',
                'Time Stamp signing CA'
            ],
            subject=mock_subject,
            issuer=mock_subject,
            text='-----BEGIN CERTIFICATE-----\nMIIEFDC...CAvygAwIB\n-----END CERTIFICATE-----',
        )

    def test_sign_request(self, mock_run_int_cmd):
        mock_sign_request_stdout = """The signed certificate is shown below:
-----BEGIN CERTIFICATE-----
MIIEFDC...CAvygAwIB
-----END CERTIFICATE-----"""

        mock_cert_dm = models.SafenetClientCertificate.create(
            ca_name='KeySecure CA',
            purpose='Client',
            duration=365,
            text='-----BEGIN CERTIFICATE-----\nMIIEFDC...CAvygAwIB\n-----END CERTIFICATE-----'
        )

        mock_run_int_cmd.return_value = mock_sign_request_stdout
        client_cert = self.eks_fm.sign_request('KeySecure CA', 2, 365, 'fake')
        assert client_cert.validate() is None
        assert client_cert == mock_cert_dm

    def test_sign_request_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'

        self.eks_fm.sign_request('KeySecure CA', 2, 365, 'fake')
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.sign_request('KeySecure CA', 2, 365, 'fake', raise_on_failure=True)

    def test_show_local_ca_one(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = self.mock_local_ca_stdout
        local_ca = self.eks_fm.show_local_ca('KeySecure CA')
        assert local_ca.validate() is None
        assert local_ca == self.mock_local_ca_dm

    def test_show_local_ca_one_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'

        self.eks_fm.show_local_ca('KeySecure CA')
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.show_local_ca('KeySecure CA', raise_on_failure=True)

    @mock.patch.object(safenet.SafenetExternalKeyServer, '_show_specific_local_ca')
    def test_show_local_ca_many(self, mock_specific, mock_run_int_cmd):
        mock_stdout = 'CA Certificates\nName\tStatus\nKeySecure CA\tCA Certificate Active'
        mock_run_int_cmd.return_value = mock_stdout
        mock_specific.return_value = self.mock_local_ca_dm

        local_cas = self.eks_fm.show_local_ca()
        for ca in local_cas:
            print(ca.serialize())
            assert ca.validate() is None
            assert ca == self.mock_local_ca_dm

    def test_show_nae_server_settings(self, mock_run_int_cmd):
        mock_show_nae_server_settings_stdout = """1.\tProtocol:               NAE-XML
IP:                     [All]
Port:                   9000
Use SSL:                yes
Server Certificate:     KeySecure Server

2.\tProtocol:               KMIP
IP:                     [All]
Port:                   5696
Use SSL:                yes
Server Certificate:     KeySecure Server
"""

        mock_nae_server_1 = models.SafenetNaeServer.create(
            ip='[All]',
            port=9000,
            protocol='NAE-XML',
            server_certificate='KeySecure Server',
            use_ssl=True
        )
        mock_nae_server_2 = models.SafenetNaeServer.create(
            ip='[All]',
            port=5696,
            protocol='KMIP',
            server_certificate='KeySecure Server',
            use_ssl=True
        )
        mock_show_nae_server_settings_list = [mock_nae_server_1, mock_nae_server_2]

        mock_run_int_cmd.return_value = mock_show_nae_server_settings_stdout
        settings = self.eks_fm.show_nae_server_settings()
        for server in settings:
            assert server.validate() is None
        assert settings == mock_show_nae_server_settings_list

    def test_show_nae_server_settings_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'

        self.eks_fm.show_nae_server_settings()
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.show_nae_server_settings(raise_on_failure=True)

    def test_modify_nae_server_port(self, mock_run_int_cmd):
        mock_modify_nae_server_port_stdout = 'Cryptographic Key Server settings successfully saved.'
        mock_run_int_cmd.return_value = mock_modify_nae_server_port_stdout
        self.eks_fm.modify_nae_server_port('[All]', 5696, 5111)

    def test_modify_nae_server_port_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'
        self.eks_fm.modify_nae_server_port('[All]', 5696, 5111)
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.modify_nae_server_port('[All]', 5696, 5111, raise_on_failure=True)

    def test_show_services(self, mock_run_int_cmd):
        mock_show_services_stdout = """Service Group  Service                   Status          Startup
nae-server     NAE Server                Started         Enabled
webadmin       Web Administration        Started         Enabled
sshadmin       SSH Administration        Started         Enabled
snmp           SNMP Agent                Stopped         Disabled
"""

        mock_service_1 = models.SafenetService.create(
            service='NAE Server',
            service_group='nae-server',
            startup='Enabled',
            status='Started'
        )
        mock_service_2 = models.SafenetService.create(
            service='Web Administration',
            service_group='webadmin',
            startup='Enabled',
            status='Started'
        )
        mock_service_3 = models.SafenetService.create(
            service='SSH Administration',
            service_group='sshadmin',
            startup='Enabled',
            status='Started'
        )
        mock_service_4 = models.SafenetService.create(
            service='SNMP Agent',
            service_group='snmp',
            startup='Disabled',
            status='Stopped'
        )

        mock_show_services_list = [mock_service_1, mock_service_2, mock_service_3, mock_service_4]

        mock_run_int_cmd.return_value = mock_show_services_stdout
        services = self.eks_fm.show_services()
        for service in services:
            assert service.validate() is None
        assert services == mock_show_services_list

    def test_show_services_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'

        self.eks_fm.show_services()
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.show_services(raise_on_failure=True)

        mock_run_int_cmd.return_value = 'garbage\ngarbage\ngarbage'
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.show_services(raise_on_failure=True)

    def test_start_service(self, mock_run_int_cmd):
        mock_start_service_stdout = 'Web administration successfully started.'
        mock_run_int_cmd.return_value = mock_start_service_stdout
        self.eks_fm.start_service('webadmin')

    def test_start_service_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'

        self.eks_fm.start_service('webadmin')
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.start_service('webadmin', raise_on_failure=True)

    def test_stop_service(self, mock_run_int_cmd):
        mock_stop_service_stdout = 'Web administration successfully stopped.'
        mock_run_int_cmd.return_value = mock_stop_service_stdout
        self.eks_fm.stop_service('webadmin')

    def test_stop_service_failure(self, mock_run_int_cmd):
        mock_run_int_cmd.return_value = 'garbage'

        self.eks_fm.stop_service('webadmin')
        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.stop_service('webadmin', raise_on_failure=True)

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'start_service')
    def test_enable_implemented(self, mock_start_service, mock_run_int_cmd):
        self.eks_fm.enable()

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'stop_service')
    def test_disable_implemented(self, mock_stop_service, mock_run_int_cmd):
        self.eks_fm.disable()

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_local_ca')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_nae_server_settings')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'modify_nae_server_port')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_services')
    def test_configure_for_testing(self, mock_show_services, mock_modify, mock_show_nae_server_settings,
                                   mock_show_local_ca, mock_run_int_cmd):
        mock_local_ca_dm = models.SafenetLocalCertificateAuthority.get_mock_object(
            overrides=dict(name='KeySecure CA'))
        mock_show_local_ca.return_value = [mock_local_ca_dm]

        mock_safenet_nae_server = models.SafenetNaeServer.get_mock_object(
            overrides=dict(protocol='KMIP', port=self.eks_fm.port + 1))
        mock_show_nae_server_settings.return_value = [mock_safenet_nae_server]

        mock_service = models.SafenetService.get_mock_object(
            overrides=dict(service_group='nae-server', status='Started'))
        mock_show_services.return_value = [mock_service]

        self.eks_fm.configure_for_testing()

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_local_ca')
    def test_configure_for_testing_fails_ca(self, mock_show_local_ca, mock_run_int_cmd):
        mock_local_ca_dm = models.SafenetLocalCertificateAuthority.get_mock_object(
            overrides=dict(name='Nope'))
        mock_show_local_ca.return_value = [mock_local_ca_dm]

        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.configure_for_testing()

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_local_ca')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_nae_server_settings')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'enable')
    def test_configure_for_testing_fails_server(self, mock_enable, mock_show_nae_server_settings,
                                                mock_show_local_ca, mock_run_int_cmd):
        mock_local_ca_dm = models.SafenetLocalCertificateAuthority.get_mock_object(
            overrides=dict(name='KeySecure CA'))
        mock_show_local_ca.return_value = [mock_local_ca_dm]

        mock_safenet_nae_server = models.SafenetNaeServer.get_mock_object(
            overrides=dict(protocol='NOPE'))
        mock_show_nae_server_settings.return_value = [mock_safenet_nae_server]

        self.eks_fm.configure_for_testing()
        mock_enable.assert_called()

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_local_ca')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_nae_server_settings')
    def test_configure_for_testing_no_kmip(self, mock_show_nae_server_settings, mock_show_local_ca,
                                           mock_run_int_cmd):
        mock_local_ca_dm = models.SafenetLocalCertificateAuthority.get_mock_object(
            overrides=dict(name='KeySecure CA'))
        mock_show_local_ca.return_value = [mock_local_ca_dm]

        mock_safenet_nae_server = models.SafenetNaeServer.get_mock_object(
            overrides=dict(protocol='KMIP'))
        mock_show_nae_server_settings.return_value = [mock_safenet_nae_server]

        with pytest.raises(_errors.SafenetExternalKeyServerError):
            self.eks_fm.configure_for_testing(kmip_version=enums.KmipVersion.proto_1_0)


@mock.patch.object(safenet.SafenetShellHandler, 'close')
@mock.patch.object(safenet.SafenetShellHandler, '_init_client')
class TestSafenet(unittest.TestCase):
    def setUp(self):
        self.eks_fm = functional_models.SafenetExternalKeyServer.get_mock_fm()

        if sys.version_info >= (3, 0):
            self.mock_open_str = 'builtins.open'
        else:
            self.mock_open_str = '__builtin__.open'

    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'sign_request')
    @mock.patch.object(functional_models.SafenetExternalKeyServer, 'show_local_ca')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    @mock.patch.object(x509, 'load_pem_x509_certificate')
    def test_sign_csr(self, mock_x509_cert_load, mock_x509_csr_load, mock_show_local_ca,
                      mock_sign_request, mock_init, mock_close):
        # Create the mock CSR and put in the mock CSR text
        mock_x509_csr = mock.MagicMock(x509.CertificateSigningRequest)
        with open('{}/mock.csr'.format(DATA_DIR), 'r') as csr:
            mock_x509_csr.public_bytes.return_value = csr.read()

        # Create a mock subject and put in the mock OU field
        mock_x509_name = mock.MagicMock(x509.Name)
        mock_x509_name.get_attributes_for_oid.return_value = ['SolidFire']
        mock_x509_csr.subject.return_value = mock_x509_name
        mock_x509_csr_load.return_value = mock_x509_csr

        mock_x509_cert = mock.MagicMock(x509.Certificate)
        mock_x509_cert_load.return_value = mock_x509_cert

        mock_cert_text = '-----BEGIN CERTIFICATE-----\nMIIEFDC...CAvygAwIB\n-----END CERTIFICATE-----'
        mock_cert_dm = models.SafenetClientCertificate.get_mock_object(
            overrides=dict(text=mock_cert_text))
        mock_local_ca_dm = models.SafenetLocalCertificateAuthority.get_mock_object(
            overrides=dict(text=mock_cert_text))

        mock_sign_request.return_value = mock_cert_dm
        mock_show_local_ca.return_value = mock_local_ca_dm

        with mock.patch(self.mock_open_str, mock.mock_open()):
            client_cert, ca_cert = self.eks_fm.sign_csr(mock_x509_csr)

        assert isinstance(client_cert, x509.Certificate)
        assert isinstance(ca_cert, x509.Certificate)

    def test_get_ssh_connection(self, mock_init, mock_close):
        ssh = self.eks_fm.get_ssh_connection()
        assert isinstance(ssh, safenet.SafenetShellHandler)

    @mock.patch.object(safenet.SafenetShellHandler, 'execute')
    def test_run_command(self, mock_execute, mock_init, mock_close):
        mock_execute.return_value = 'hi'
        output = self.eks_fm.run_command('echo hi')
        assert output.stdout == 'hi'

    @mock.patch.object(safenet.SafenetShellHandler, 'execute')
    def test_run_interactive_command(self, mock_execute, mock_init, mock_close):
        mock_execute.return_value = 'hi'
        output = self.eks_fm.run_interactive_command(['echo hi'])
        assert output == 'hi'


class SafenetShellHandler(unittest.TestCase):
    @mock.patch.object(safenet.SafenetShellHandler, 'close')
    @mock.patch.object(safenet.SafenetShellHandler, '_init_client')
    @mock.patch.object(safenet.SafenetShellHandler, '_set_timeout')
    @mock.patch.object(safenet.SafenetShellHandler, '_send')
    def test_execute(self, mock_send, mock_set_timeout, mock_init_client, mock_close):
        shell = safenet.SafenetShellHandler('1.1.1.1', 'user', 'password')
        shell.prompt = 'safenet#'
        shell.channel = mock.Mock(paramiko.channel.Channel)
        shell.channel.recv.return_value = 'echo hi\nhi\nsafenet#'
        stdout = shell.execute(['echo hi'])
        assert stdout == 'hi'

    @mock.patch.object(safenet.SafenetShellHandler, '_init_client')
    def test_close(self, mock_init_client):
        shell = safenet.SafenetShellHandler('1.1.1.1', 'user', 'password')
        shell.ssh = mock.Mock(paramiko.SSHClient)
        shell.close()

    @mock.patch.object(paramiko, 'SSHClient')
    @mock.patch.object(safenet.SafenetShellHandler, '_prepare')
    def test_init_client(self, mock_prepare, mock_ssh):
        mock_channel = mock.MagicMock(paramiko.channel.Channel)
        mock_channel.recv_ready.side_effect = [False, True]
        mock_ssh().invoke_shell.return_value = mock_channel

        shell = safenet.SafenetShellHandler('1.1.1.1', 'user', 'password')
        assert isinstance(shell, safenet.SafenetShellHandler)

    @mock.patch.object(safenet.SafenetShellHandler, '_init_client')
    @mock.patch.object(safenet.SafenetShellHandler, '_send')
    @mock.patch.object(paramiko.channel.Channel, 'recv')
    def test_prepare(self, mock_recv, mock_send, mock_init_client):
        shell = safenet.SafenetShellHandler('1.1.1.1', 'user', 'password')
        shell.channel = mock.Mock(paramiko.channel.Channel)
        shell.channel.recv.return_value = 'config\nsafenet#'
        prompt = shell._prepare()
        assert prompt == 'safenet#'

    @mock.patch.object(safenet.SafenetShellHandler, '_init_client')
    @mock.patch.object(paramiko.channel.Channel, 'gettimeout')
    @mock.patch.object(paramiko.channel.Channel, 'settimeout')
    def test_set_timeout(self, mock_set, mock_get, mock_init_client):
        shell = safenet.SafenetShellHandler('1.1.1.1', 'user', 'password')
        shell.channel = mock.Mock(paramiko.channel.Channel)
        shell.channel.gettimeout.return_value = 59
        shell._set_timeout(60)

    @mock.patch.object(safenet.SafenetShellHandler, '_init_client')
    @mock.patch.object(paramiko.channel.Channel, 'send')
    @mock.patch.object(paramiko.channel.Channel, 'recv_ready')
    def test_send(self, mock_recv_ready, mock_send, mock_init_client):
        shell = safenet.SafenetShellHandler('1.1.1.1', 'user', 'password')
        shell.channel = mock.Mock(paramiko.channel.Channel)
        shell.channel.recv_ready.side_effect = [False, True]
        shell._send('echo hi')
