"""Vormetric EKS functional model."""
import autologger

from valence.functional_models import _errors
from valence.functional_models._external_key_server import base

_log = autologger.get_sflogging()


class VormetricExternalKeyServer(base.ExternalKeyServerBase):
    """EKS functional model for Vormetric DSM.

    This class will act as an interface to Vormetric DSM. The Vormetric API will be translated into
    Python functions and used to implement the abstract methods defined in ExternalKeyServerBase.
    Testcases should utilize the abstract methods whenever possible so that they can be EKS agnostic.
    """
    def sign_csr(self, csr, start_date=None, end_date=None, force=False):
        """Take the given CSR and get it signed by a Trusted CA (usually local to the EKS).

        This method should return a tuple in the format: (path_to_user_cert, path_to_ca_cert)

        :param str csr: Path to file containing PEM encoded certificate signing request
        :param datetime.datetime start_date: Datetime object representing the certificate start date
        :param datetime.datetime end_date: Datetime object representing the certificate expiration date
        :param bool force: N/A
        :rtype: tuple[str]
        """
        raise _errors.Error('sign_csr method not implemented yet.')

    def enable(self):
        """Use to restore the connection between the SF Cluster and EKS."""
        raise _errors.Error('enable method not implemented yet.')

    def disable(self):
        """Use to sever the connection between the SF Cluster and EKS."""
        raise _errors.Error('disable method not implemented yet.')

    def configure_for_testing(self, kmip_version=None):
        """Confirm EKS is ready for test execution.

        :param sf_models.enums.KmipVersion kmip_version: Version to use.
        """
        raise _errors.Error('configure_for_testing method not implemented yet.')

    def is_running(self):
        """Return True if the KMIP service or EKS is up and running."""
        raise _errors.Error('is_running method not implemented yet.')
