"""Utils for functional models."""

import time

import autologger

from valence.functional_models import (
    _constants,
    _errors,
)

_log = autologger.get_sflogging()


def simple_sf_comparator(a_obj, b_obj, oper=_constants.OPER_EQ):
    """Compares two lists.

    See original sf_comparator in sf_platform/utils/helpers.py for reference.

    :param list a_obj: Object which supports above specified operators
    :param list b_obj: Object which supports above specified operators.
    :param str oper: Type of Operator you are requesting to test, only handles equals or not equals

    :rtype: bool
    :Returns: True, if they meet the required operator condition.
    """
    comp_result = False
    if a_obj == b_obj:
        comp_result = True
    if oper == _constants.OPER_NEQ:
        comp_result = not comp_result
    return comp_result


class BasePoll(object):
    """Base Poll Class responsible for polling on a method.

    User can pass a Function, Arguments, Test Val and TestCondition.

    :param int max_time: Max time in seconds
    :param Any api_fun: Function on which polling need to happen
    :param Any test_val: Test Value any type of object but same as
                            function return type
    :param str oper: Type of Operator
    :param int poll_period: Polling Should happend for how long
    """
    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, max_time, api_fun, test_val, oper,
                 poll_period=60, *args, **argv):
        """Constructor of Base Poll."""
        self.api_fun = api_fun
        self.test_val = test_val
        self.oper = oper
        self.args = args
        self.argv = argv
        self.max_time = max_time
        self.poll_period = poll_period
        self.keep_running = True
        self.total_time = 0
        self.result_dict = {}
        self.iteration = 0
        self.is_polled_max = False
        self.api_result = None

    def poll(self):
        """Run polling on method till it reaches specified max time."""
        start_time = time.time()
        while self.keep_running and self.total_time <= self.max_time:
            self.iteration += 1
            _log.debug('Iteration {} is running, {}/{} seconds'
                       .format(self.iteration, self.total_time, self.max_time))

            api_result = self.api_fun(*self.args, **self.argv)
            _log.debug('Poll returned result [{0}], operator [{1}] against value [{2}]'
                       .format(api_result, self.oper, self.test_val))
            self.api_result = api_result
            _log.debug('self.test_val: {}, api_result: {}, self.oper: {}'.
                       format(self.test_val, api_result, self.oper))
            cond_met = simple_sf_comparator(self.test_val, api_result, self.oper)
            _log.debug("is '{}' {} '{}' ? '{}'".format(self.test_val, self.oper,
                                                       api_result, cond_met))
            if cond_met:
                self.stop_polling()
                break
            else:
                time.sleep(self.poll_period)
            self.total_time = (time.time() - start_time)

        if self.total_time > self.max_time:
            self.is_polled_max = True
            raise _errors.PollMaxError('Polling max_time {} was exceeded! Total time = {},'
                                       ' api_result: {}, is_max_reached: {}'
                                       .format(self.max_time, self.total_time,
                                               self.api_result, self.is_polled_max))

    def stop_polling(self):
        """User can stop the polling."""
        self.keep_running = False


class BlockPoll(BasePoll):
    """Block Poll waits till it the condition met or expires.

    :param int max_time: Max time in seconds
    :param Any api_fun: Function on which polling need to happen
    :param Any test_val: Test Value any type of object but same as
                            function return type
    :param str oper: Type of Operator
    :param int poll_period: Polling Should happend for how long
    :param list args: List of Positional Arguments for API Function
    :param dict argv: Key word arguments.
    """
    def __init__(self, max_time, api_fun, test_val, oper,
                 poll_period, *args, **argv):
        """Constructor of Block Poll."""
        BasePoll.__init__(self, max_time, api_fun, test_val, oper,
                          poll_period, *args, **argv)

    def start_polling(self):
        """Simply start the polling."""
        self.poll()
