"""Base class for Management Node functional model."""

import autologger
import sf_ssh

from valence import models
from valence.functional_models import _errors

_log = autologger.get_sflogging()


class ManagementNodeBase(object):
    """Management Node Functional Model base class.

    :param valence.models.ResourceManagementNode management_node_dm: node data model
    """

    def __init__(self, management_node_dm):
        self._management_node_dm = management_node_dm

    def __repr__(self):
        return '{}<{}={}>'.format('ManagementNodeFm', 'mip', self._management_node_dm.mip)

    @property
    def mip(self):
        """Shortcut for mip attribute on node model.

        :rtype: str
        """
        return self._management_node_dm.mip

    @property
    def username(self):
        """Shortcut for username attribute on node model.

        :rtype: str
        """
        return self._management_node_dm.username

    @property
    def password(self):
        """Shortcut for password attribute on node model.

        :rtype: str
        """
        return self._management_node_dm.password

    @property
    def version(self):
        """Shortcut for version attribute on node model.

        :rtype: valence.models.SfVersion
        """
        return self._management_node_dm.build_version.sf_version

    @property
    def static_id(self):
        """Shortcut for static_id attribute on node model.

        :rtype: str
        """
        return self._management_node_dm.static_id

    def get_ssh_connection(self, quiet=True):
        """Convenience method for returning the SSH connection for the node.

        Usage::

            with self.get_ssh_connection() as ssh_conn:
                output = ssh_conn.run_command("ls /")

        :param bool quiet: If true, don't print out the ssh login/stdout/stderrs.
        :rtype: sf_ssh.Ssh
        """
        return sf_ssh.Ssh(self.mip, username=self.username, password=self.password, quiet=quiet)

    def get_uptime(self, timeout=10):
        """Return the uptime of an mnode in seconds.

        :param int timeout: timeout in seconds
        :returns: seconds of uptime from cat /proc/uptime (no rounding)
        :rtype: int
        """
        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command('cat /proc/uptime', timeout=timeout)
            if result.stderr:
                raise _errors.FunctionalManagementNodeError(
                    'SSH returned an error while getting uptime: {}'.format(result.stderr))
            # format uptime to return seconds
            up_secs = int(result.stdout.split(' ')[0].split('.')[0])
            return up_secs

    @classmethod
    def from_resource_data(cls, rs_mnode_data):
        """Factory method for instantiating a mnode FM from the resource data.

        Takes the same kwargs as __init__.

        :param dict rs_mnode_data: Data that corresponds to ResourceManagementNode
        :return: Mnode FM instance
        :rtype: valence.functional_models.ManagementNodeBase
        """
        rs_mnode_dm = models.ResourceManagementNode(rs_mnode_data)
        rs_mnode_dm.validate()
        return cls(rs_mnode_dm)

    @classmethod
    def get_mock_fm(cls):
        """Factory method for instantiating a mock node FM.

        No response validation will be done by default. This is purely for unit testing purposes.

        :rtype: valence.functional_models.ManagementNodeBase
        """
        rs_node_dm = models.ResourceManagementNode.get_mock_object()
        return cls(rs_node_dm)
