"""Management Node Functional Model for REST-based Mnode2.x."""

import time

import autologger
import sf_models

import sf_api
from sf_api import (
    api_constants,
    api_rest_constants,
)
from valence import (
    models,
    utils,
)
from valence.functional_models import (
    _constants,
    _errors,
)
from valence.functional_models._management_node import base

_log = autologger.get_sflogging()


class ManagementNode2x(base.ManagementNodeBase):
    """Management Node FM for REST-based mnode 2.x.

    :param valence.models.ResourceManagementNode management_node_dm: node data model
    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param str|None client_id: the ID of the auth client
    """

    def __init__(self, management_node_dm, cluster_dm=None, client_id=None):
        super(ManagementNode2x, self).__init__(management_node_dm)
        self.api = sf_api.BaseRestClient(
            cluster_dm=cluster_dm,
            client_id=client_id,
            base_url=self.mip,
            authenticate=False)

    def get_all_assets(self):
        """Returns existing assets.

        :rtype: list[valence.models.AssetResult]
        :returns: list of existing asset models.
        """
        response = self.api.get(service=api_rest_constants.MNODE,
                                path=api_constants.ASSETS_REST_PATH)
        asset_models = [models.AssetResult(a) for a in response.result]
        return asset_models

    def get_storage_asset_id_for_storage_cluster_mvip(self, cluster_dm):
        """Returns the storage ID for a given MVIP.

        :param valence.models.ResourceCluster cluster_dm: Cluster under test.
        :rtype: str|None
        :returns: Storage UUID or None if ID isn't found in assets.
        """
        asset_models = self.get_all_assets()
        storage_id = None
        for a in asset_models:
            cluster_asset = next((x for x in a.storage if x.ip == cluster_dm.mvip), None)
            if cluster_asset:
                storage_id = cluster_asset.asset_id
                break
        return storage_id

    def add_asset_if_needed(self, asset_name=None):
        """Adds a named asset group to mnode2.x services.

        :param str|None asset_name: Asset name to add. If none, a randomly generated name that does
            not collide with existing assets will be added.
        :rtype: str
        :returns: UUID of newly added asset or existing asset.
        :raises: _errors.FunctionalManagementNodeError
        """
        asset_uuid = None
        asset_models = self.get_all_assets()
        asset_names = [n.name for n in asset_models]

        if asset_name and asset_name in asset_names:
            _log.debug('Asset {} already exists on mnode. Done here.'.format(asset_name))
        else:
            # Randomly generate an asset name that doesn't conflict with existing names.
            if not asset_name:
                asset_name = self._generate_unique_asset_name(asset_names)

            self._add_asset(asset_name)

            # Verify that it got added.
            asset_models = self.get_all_assets()
            if asset_name not in [n.name for n in asset_models]:
                raise _errors.FunctionalManagementNodeError("Asset {} didn't get added!".format(
                    asset_name))

        # pull out the uuid b/c that's what we need to return.
        asset_uuid = next((x for x in asset_models if x.name == asset_name), None).asset_id
        return asset_uuid

    def add_storage_id_to_asset(self, asset_id, cluster_dm):
        """Adds storage cluster to assets on mnode.

        :param valence.models.ResourceCluster cluster_dm: Cluster under test.
        :param str asset_id: Which asset to add the storage mvip to.
        :rtype: str|None
        :returns: Storage UUID that was added.
        """
        params = {api_constants.USERNAME: cluster_dm.username,
                  api_constants.PASSWORD: cluster_dm.password,
                  api_constants.IP: cluster_dm.mvip,
                  api_constants.CONFIG: {}}
        response = self.api.post(service=api_rest_constants.MNODE,
                                 path='{}/{}/{}'.format(
                                     api_constants.ASSETS_REST_PATH, asset_id,
                                     api_rest_constants.STORAGE_CLUSTERS),
                                 request_json=params,
                                 result_dm_class=models.Asset)
        storage_id = response.result_dm.asset_id
        return storage_id

    def delete_all_assets(self):
        """Deletes all storage clusters and assets on mnode."""
        asset_models = self.get_all_assets()
        for a in asset_models:
            for storage in a.storage:
                self._delete_storage_from_asset(a.asset_id, storage.asset_id)
            self._delete_asset(a.asset_id)

    def services_upload(self, service_bundle_data, validate_version=False):
        """Send a service bundle to the mnode upload endpoint.

        :param dict service_bundle_data: contains the fat tar service bundle data.
        :param bool validate_version: Validate that the service bundle version is the same
            as the currently installed version.
        :return: checksum: the checksum of the service_bundle uploaded.
        :rtype: str
        """
        params = {'data': service_bundle_data,
                  'validate_version': validate_version}

        response = self.api.put(service=api_rest_constants.MNODE,
                                path='{}/{}'.format(api_constants.SERVICES,
                                                    api_rest_constants.UPLOAD),
                                data=params,
                                result_dm_class=sf_models.ServicesUpload)

        checksum = response.result_dm.checksum
        return checksum

    def execute_preupgrade_health_check(self, cluster_dm):
        """The REST adapter for healthchecks on mnode.

        This adapter will run healthcheck commands on mnode using REST and convert the output into
        data structure that's common between CLI and REST interfaces.

        :param valence.models.ResourceCluster cluster_dm: Cluster under test.
        :rtype: valence.models.HealthcheckResultModel
        :returns: The result of the healthchecks.
        :raises: valence.functional_models._errors.FunctionalManagementNodeError
        """
        storage_id = self.get_storage_asset_id_for_storage_cluster_mvip(cluster_dm)
        if not storage_id:
            raise _errors.FunctionalManagementNodeError(
                'Storage MVIP {} has not been added to the mnode assets yet, please do so and then '
                'rerun healthchecks.'.format(cluster_dm.mvip))

        healthcheck_id, status_url = self._start_healthcheck(storage_id)
        _log.debug('Healthcheck started, healthcheck_id {}'.format(healthcheck_id))
        result_dm = self._wait_for_healthcheck_to_finish(status_url)
        return result_dm

    def _generate_unique_asset_name(self, existing_asset_names):
        """Generate a name that isn't in a list of existing names.

        :param list[str] existing_asset_names: existing names.
        :rtype: str
        :returns: Name that's not in list of existing names.
        """
        asset_name = None
        for _ in range(10):
            random_name = utils.generate_random_string(8)
            if random_name not in existing_asset_names:
                asset_name = random_name
                break
            else:
                _log.debug('Randomly generated name {} collided with existing asset name {}'.
                           format(random_name, ' '.join(existing_asset_names)))
                continue
        assert asset_name, 'asset name should not be empty.'
        return asset_name

    def _add_asset(self, asset_name):
        """Adds a named asset group to mnode2.x services.

        :param str asset_name: Name of the asset.
        :rtype: str
        :returns: UUID of newly added asset
        """
        params = {api_constants.NAME: asset_name, api_constants.CONFIG: {}}
        response = self.api.post(service=api_rest_constants.MNODE,
                                 path=api_constants.ASSETS_REST_PATH,
                                 request_json=params,
                                 result_dm_class=models.AssetResult)
        asset_uuid = response.result_dm.asset_id
        return asset_uuid

    def _delete_asset(self, asset_id):
        """Deletes an asset.

        This doesn't check that asset is empty, you should do that.

        :param str asset_id: asset to delete.
        """
        self.api.delete(service=api_rest_constants.MNODE,
                        path='{}/{}'.format(api_constants.ASSETS_REST_PATH, asset_id))

    def _delete_storage_from_asset(self, asset_id, storage_id):
        """Deletes a storage id from asset.

        :param str asset_id: asset to delete from.
        :param str storage_id: storage id to delete.
        """
        self.api.delete(service=api_rest_constants.MNODE,
                        path='{}/{}/{}/{}'.format(
                            api_constants.ASSETS_REST_PATH,
                            asset_id,
                            api_rest_constants.STORAGE_CLUSTERS,
                            storage_id))

    def _start_healthcheck(self, storage_id):
        """Starts healthchecks for a given asset and storage IDs.

        :param str storage_id: Which storage cluster to run checks on.
        :rtype: tuple[str,str]
        :returns: healthcheck ID and URL for checking the resource
        """
        params = {api_constants.STORAGE_ID: storage_id,
                  api_constants.CONFIG: {}}
        post_response = self.api.post(
            service=api_rest_constants.STORAGE_SERVICE,
            path=api_rest_constants.HEALTHCHECKS_REST_PATH, request_json=params,
            result_dm_class=models.HealthcheckResultModel)

        return post_response.result_dm.healthcheck_id, post_response.result_dm.links['self']

    def _wait_for_healthcheck_to_finish(self, self_healthcheck_url):
        """Waits for healthcheck to finish.

        :param str self_healthcheck_url: URL from POST response to wait on.
        :rtype: valence.models.HealthcheckResultModel
        :returns: The result of the healthchecks.
        :raises: _errors.HealthcheckError
        """
        timer = utils.timeout_timer(_constants.HEALTHCHECK_EXECUTION_TIMEOUT_S)
        while next(timer):
            get_response = self.api.get(url=self_healthcheck_url,
                                        result_dm_class=models.HealthcheckResultModel)
            _log.debug('Healthcheck operation state is {} and status is {}'.format(
                get_response.result_dm.state,
                get_response.result_dm.status.status if get_response.result_dm.status else 'N/A'))
            if get_response.result_dm.state in [sf_api.HealthcheckState.error,
                                                sf_api.HealthcheckState.complete]:
                if get_response.result_dm.status:
                    return get_response.result_dm
            else:
                time.sleep(1)
        raise _errors.HealthcheckError('Healthchecks timed out')

    def start_storage_cluster_upgrade(self, cluster_dm, build_url, sfinstall_options=None):
        """Starts a storage cluster upgrade.

        :param valence.models.ResourceCluster cluster_dm: Cluster data model under test
        :param str build_url: The location to upgrade to
        :param valence.models.UpgradeConfig sfinstall_options: The list of options to pass to sfinstall

        :return: Upgrade ID of the running upgrade process
        :rtype: str
        """
        _log.debug('Starting upgrade with options: {}'.format(sfinstall_options))
        storage_id = self.get_storage_asset_id_for_storage_cluster_mvip(cluster_dm)
        config = sfinstall_options if sfinstall_options else {}
        params = {
            api_constants.STORAGE_ID: storage_id,
            api_constants.PACKAGE: build_url,
            api_constants.CONFIG: config
        }
        post_response = self.api.post(service=api_rest_constants.STORAGE_SERVICE,
                                      path=api_rest_constants.STORAGE_UPGRADE_PATH,
                                      request_json=params,
                                      result_dm_class=models.UpgradeResource)
        return post_response.result_dm.upgrade_id

    def pause_storage_cluster_upgrade(self, upgrade_id):
        """Pauses a storage cluster upgrade.

        :param str upgrade_id: Upgrade ID of the process to be paused.
        """
        _log.debug('Pausing upgrade with upgrade ID: {}'.format(upgrade_id))
        data = {api_constants.ACTION: 'pause'}
        self.api.put(service=api_rest_constants.STORAGE_SERVICE,
                     path=api_rest_constants.STORAGE_UPGRADE_PATH,
                     data=data, result_dm_class=models.UpgradeResource)

    def resume_storage_cluster_upgrade(self, upgrade_id):
        """Pauses a storage cluster upgrade.

        :param str upgrade_id: Upgrade ID of the process to be resumed.
        """
        _log.debug('Resuming upgrade with upgrade ID: {}'.format(upgrade_id))
        data = {api_constants.ACTION: 'resume'}
        self.api.put(service=api_rest_constants.STORAGE_SERVICE,
                     path=api_rest_constants.STORAGE_UPGRADE_PATH,
                     data=data, result_dm_class=models.UpgradeResource)

    def abort_storage_cluster_upgrade(self, upgrade_id):
        """Pauses a storage cluster upgrade.

        :param str upgrade_id: Upgrade ID of the process to be aborted.
        """
        _log.debug('Aborting upgrade with upgrade ID: {}'.format(upgrade_id))
        data = {api_constants.ACTION: 'abort'}
        self.api.put(service=api_rest_constants.STORAGE_SERVICE,
                     path=api_rest_constants.STORAGE_UPGRADE_PATH,
                     data=data, result_dm_class=models.UpgradeResource)

    def get_storage_cluster_upgrade(self, upgrade_id):
        """Returns the upgrade status of the process.

        :param str upgrade_id: Upgrade ID of the process
        :return: Status of the running process
        :rtype: valence.models.UpgradeResource
        """
        response = self.api.get(service=api_rest_constants.STORAGE_SERVICE,
                                path=api_rest_constants.STORAGE_UPGRADE_PATH,
                                params={api_rest_constants.UPGRADE_ID: upgrade_id},
                                result_dm_class=models.UpgradeResource)
        return response.result_dm

    @classmethod
    def from_resource_data(cls, rs_mnode_data):
        """Factory method for instantiating a mnode FM from the resource data.

        Takes the same kwargs as __init__.

        :param dict rs_mnode_data: Data that corresponds to ResourceManagementNode
        :return: Mnode FM instance
        :rtype: valence.functional_models.ManagementNode2x
        """
        rs_mnode_dm = models.ResourceManagementNode(rs_mnode_data)
        rs_mnode_dm.validate()
        return cls(rs_mnode_dm)

    @classmethod
    def get_mock_fm(cls):
        """Factory method for instantiating a mock node FM.

        No response validation will be done by default. This is purely for unit testing purposes.

        :rtype: valence.functional_models.ManagementNode2x
        """
        return cls(models.ResourceManagementNode.get_mock_object())
