"""Management node Functional Model for legacy mnode."""

import os

import autologger
import sf_ssh

from valence import models
from valence.functional_models import _constants
from valence.functional_models._management_node import (
    base,
    processed_healthchecks,
)

_log = autologger.get_sflogging()


class ManagementNodeLegacy(base.ManagementNodeBase):
    """Management node Functional Model for mnode 1.5 and mnode 1.1.

    These mnodes do not support any REST calls and only work over SSH.
    No compute nodes support exists.
    """

    def start_storage_cluster_upgrade(self, cluster_dm, build):
        """Start upgrade of storage cluster.

        :param valence.models.ResourceCluster cluster_dm: Cluster DM to start upgrade on.
        :param valence.models.Build build: Build to upgrade storage cluster to.
        """
        raise NotImplementedError

    def pause_storage_cluster_upgrade(self, cluster_dm):
        """Pause upgrade of storage cluster.

        :param valence.models.ResourceCluster cluster_dm: Pause upgrade of this storage cluster.
        """
        raise NotImplementedError

    def kill_storage_cluster_upgrade(self, cluster_dm):
        """Kill upgrade of storage cluster.

        :param valence.models.ResourceCluster cluster_dm: Pause upgrade of this storage cluster.
        """
        raise NotImplementedError

    def get_storage_cluster_upgrade_status(self, cluster_dm):
        """Get the status of storage cluster upgrade.

        :param valence.models.ResourceCluster cluster_dm: Pause upgrade of this storage cluster.
        """
        raise NotImplementedError

    def execute_preupgrade_health_check(self,
                                        cluster_dm,
                                        show_help=False,
                                        show_version=False,
                                        run_checks_against_asset=True,
                                        skip_version_check=False,
                                        deploy_only_versions=None,
                                        sfhealthtools_url_to_use=None,
                                        raise_on_failure=True,
                                        metadata=None):
        """The CLI adapter for healthchecks on mnode.

        This adapter will run healthcheck commands on mnode using SSH and convert the output into
        data structure that's common between CLI and REST interfaces.

        The rules for sfupgradecheck stdout, stderr, rc are as follows:
        sfupgradecheck -v: output goes to stdout
        sfupgradecheck -u -p mvip: output of checks goes to stdout, "installing..." goes to stderr
        sfupgradecheck -U valid_version: output goes to stdout
        sfupgradecheck -U invalid_version: stderr only.

        sfupgradecheck rc rules are:
        rc != 0 when:
        credentials are wrong.
        Actually running checks (so no -v) and version is outdated.
        Bad syntax.

        :param valence.models.ResourceCluster cluster_dm: Cluster under test.
        :param bool show_help: Display help.
        :param bool show_version: Show versions currently installed
        :param bool run_checks_against_asset: run checks against cluster/mnode/etc...asset under test.
        :param bool skip_version_check: Skip latest version check.
        :param bool raise_on_failure: If rc != 0, raise ssh exception.
        :param typing.Iterable[tuple] deploy_only_versions: comma separated list
            of versions to only deploy to nodes, without running the checks afterwards.
            Requires cluster credentials.
        :param str sfhealthtools_url_to_use: URL to query for latest version, or None for default.
        :param str|None metadata: metadata URI
        :rtype: valence.functional_models.ProcessedHealthchecks
        :returns: Processed healthchecks structure and the stderr from call.
        """
        cmd = self._build_sfupgradecheck_cmd(
            cluster_dm,
            show_help=show_help,
            show_version=show_version,
            run_checks_against_asset=run_checks_against_asset,
            skip_version_check=skip_version_check,
            deploy_only_versions=deploy_only_versions,
            sfhealthtools_url_to_use=sfhealthtools_url_to_use,
            metadata=metadata)

        rco = self._get_ssh_connection().run_command(cmd, raise_on_failure=raise_on_failure,
                                                     use_sudo=False)
        _log.debug('{} stdout: {}'.format(_constants.SFUPGRADECHECK, rco.stdout))
        _log.debug('{} stderr: {}'.format(_constants.SFUPGRADECHECK, rco.stderr))

        if rco.return_code and not raise_on_failure:
            _log.debug('sfupgradecheck failed with rc {}, stderr {}'.format(rco.return_code, rco.stderr))
            raise sf_ssh.SshResponseError(
                host=self.mip,  # pylint: disable=protected-access
                cmd=cmd, out=rco.stdout, err=rco.stderr, rc=rco.return_code)

        if not rco.return_code and rco.stdout:
            processed_checks = processed_healthchecks.ProcessedHealthchecks.from_json(rco.stdout)
            processed_checks.stderr = rco.stderr
            processed_checks.stdout = rco.stdout
        else:
            processed_checks = processed_healthchecks.ProcessedHealthchecks(stdout=rco.stdout,
                                                                            stderr=rco.stderr)
        return processed_checks

    def run_sfupdate_healthtools_cmd(self, show_help=False,
                                     show_installed_version=False, show_latest_version=False,
                                     show_all_previous_versions=False, revert_to_this_version=None,
                                     sfhealthtools_url_to_use=None, metadata=None):
        """Run sfupdate_healthtools command with provided options.

        In order to use install_this_version, the healthtools package must already be uploaded to
        fdva into /tmp/ura-packages/, using ura.put_healthtools_package() fcn.
        Main option that's missing from this method is install_this_version, which is supposed to install
        a new tarball. In order to do that, please use install_new_version() fcn.
        That's due to ELEM-3745.

        :param bool show_help: Show help message
        :param bool show_installed_version: Display the currently installed version.
        :param bool show_latest_version: Display the latest version available on BitBucket
        :param bool show_all_previous_versions: Display all previously installed and available versions.
        :param valence.models.GenericVersion|None revert_to_this_version: Revert to this
            existing version.
        :param str sfhealthtools_url_to_use: URL to query for latest version, or None for default.
        :param str|None metadata: Metadata URI
        :rtype: tuple[str,str]
        :returns: stdout, stderr from ssh command.
        :raises: ssh.SshResponseError if ssh command returns bad rc code.
        """
        cmd = self._build_sfupdate_healthtools_cmd(
            show_help=show_help,
            show_installed_version=show_installed_version,
            show_latest_version=show_latest_version,
            show_all_previous_versions=show_all_previous_versions,
            revert_to_this_version=revert_to_this_version,
            sfhealthtools_url_to_use=sfhealthtools_url_to_use,
            metadata=metadata,
            install_this_version=None)

        rco = self._get_ssh_connection().run_command(cmd, use_sudo=False)

        _log.debug('sfupdate-healthtools stdout: {}'.format(rco.stdout))
        _log.debug('sfupdate-healthtools stderr: {}'.format(rco.stderr))

        if rco.return_code:
            raise sf_ssh.SshResponseError(
                host=self.mip,  # pylint: disable=protected-access
                cmd=cmd, out=rco.stdout, err=rco.stderr, rc=rco.return_code)
        return rco.stdout, rco.stderr

    def get_current_healthtools_version(self):
        """Returns the currently installed version of healthtools on mnode.

        :rtype: valence.models.GenericVersion
        """
        stdout, _ = self.run_sfupdate_healthtools_cmd(show_installed_version=True)
        output_list = stdout.split('\n')
        output_list = [s.strip() for s in output_list]
        index = output_list.index(_constants.CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION)
        version_str = output_list[index + 1].strip()
        return models.GenericVersion(version_str)

    def install_new_healthtools_version(self, install_this_version, original_version=None):
        """Helper method to install new version on mnode.

        :param valence.models.GenericVersion install_this_version: Install this version
            of healthtools package.
        :param valence.models.GenericVersion|None original_version: Original version on mnode
        :rtype: tuple[str,str]
        :returns: stdout, stderr from ssh command.
        :raises: ssh.SshResponseError if ssh command returns bad rc code.
        """
        if install_this_version == original_version:
            _log.debug('Instead of installing version {} reverting to original version {}'.format(
                install_this_version, original_version))
            cmd = self._build_sfupdate_healthtools_cmd(revert_to_this_version=original_version)
        else:
            _log.debug('Installing version {}'.format(install_this_version))
            cmd = self._build_sfupdate_healthtools_cmd(install_this_version=install_this_version)

        rco = self._get_ssh_connection().run_command(cmd, use_sudo=False)
        _log.debug('sfupdate-healthtools stdout: {}'.format(rco.stdout))
        _log.debug('sfupdate-healthtools stderr: {}'.format(rco.stderr))

        if rco.return_code:
            raise sf_ssh.SshResponseError(
                host=self.mip,  # pylint: disable=protected-access
                cmd=cmd, out=rco.stdout, err=rco.stderr, rc=rco.return_code)
        return rco.stdout, rco.stderr

    def _get_ssh_connection(self, quiet=True):
        return sf_ssh.Ssh(self.mip,
                          username=self.username,
                          password=self.password,
                          quiet=quiet)

    # pylint: disable=too-complex
    def _build_sfupdate_healthtools_cmd(self,
                                        show_help=False,
                                        show_installed_version=False,
                                        show_latest_version=False,
                                        show_all_previous_versions=False,
                                        sfhealthtools_url_to_use=None,
                                        metadata=None,
                                        install_this_version=None,
                                        revert_to_this_version=None):
        """Build the sfupdate-healthtools command.

        This fcn exists to ensure that we guard the command args against accidental usage.

        :param bool show_help: Show help message
        :param bool show_installed_version: Display the currently installed version.
        :param bool show_latest_version: Display the latest version available on BitBucket
        :param bool show_all_previous_versions: Display all previously installed and available versions.
        :param valence.models.GenericVersion|None install_this_version: Install this version
            of healthtools package.
        :param valence.models.GenericVersion revert_to_this_version: Revert to this version.
        :param str sfhealthtools_url_to_use: URL to query for latest version, or None for default.
        :param str|None metadata: Metadata URI
        :rtype: str
        :returns: command to run.
        """
        args = list()
        if sfhealthtools_url_to_use:
            _log.debug('Using SFHEALTHTOOLS URL {}'.format(sfhealthtools_url_to_use))
            args.append('export {}="{}" && '.format(_constants.SFHEALTHTOOLS_URL_ENV,
                                                    sfhealthtools_url_to_use))
        args.append('sudo -E ' + _constants.CMD_SFUPDATE_HEALTHTOOLS)
        if show_help:
            args.append('-h')
        if install_this_version:
            args.append(os.path.join(_constants.URA_PACKAGES_DIR, '{}-{}.{}'.format(
                _constants.SOLIDFIRE_HEALTHTOOLS, install_this_version, 'tgz')))
        if show_installed_version:
            args.append('--version')
        if show_latest_version:
            args.append('--latest-version')
        if show_all_previous_versions:
            args.append('--show-versions')
        if revert_to_this_version:
            args.append('-r')
            # GenericVersion.to_str() will produce x.y.z.w. string, but this command takes x.y.z.w
            args.append(str(revert_to_this_version))
        if metadata:
            args.append('--metadata={}'.format(metadata))

        cmd = ' '.join(args)
        return cmd

    @staticmethod
    # The pylint complexity failure is disabled here b/c this function simply builds a command string
    # for a command with many attributes.
    # pylint: disable=too-complex
    def _build_sfupgradecheck_cmd(cluster_dm, show_help=False, show_version=False,
                                  run_checks_against_asset=True, skip_version_check=False,
                                  deploy_only_versions=None, sfhealthtools_url_to_use=None,
                                  metadata=None):
        """Builds sfupgradecheck with any additional args.

        :param valence.models.ResourceCluster cluster_dm: Cluster under test.
        :param bool show_help: Display help.
        :param bool show_version: Show versions currently installed
        :param bool run_checks_against_asset: Authenticate to cluster?
        :param bool skip_version_check: Skip latest version check.
        :param typing.Iterable[tuple] deploy_only_versions: comma separated list
            of versions to only deploy to nodes, without running the checks afterwards.
            Requires cluster credentials.
        :param str sfhealthtools_url_to_use: URL to query for latest version, or None for default.
        :param str|None metadata: metadata URI
        :rtype: str
        :returns: command string
        """
        args = list()
        if sfhealthtools_url_to_use:
            args.append('export {}="{}" && '.format(_constants.SFHEALTHTOOLS_URL_ENV,
                                                    sfhealthtools_url_to_use))
        args.extend(['sudo -E', _constants.SFUPGRADECHECK])
        args.append('--json-output')

        if show_help:
            args.append('--help')
        if show_version:
            args.append('--version')
        if skip_version_check:
            args.append('--skip-version-check')
        if run_checks_against_asset:
            args.append('-u {user} -p {password} {mvip}'.format(
                user=cluster_dm.username, password=cluster_dm.password,
                mvip=cluster_dm.mvip))
        if deploy_only_versions:
            args.append('--deploy-only ' +
                        ','.join('{}:{}'.format(n[0], n[1]) for n in deploy_only_versions))
        if metadata:
            args.append('--metadata={}'.format(metadata))

        return ' '.join(args)

    @classmethod
    def from_resource_data(cls, rs_mnode_data):
        """Factory method for instantiating a mnode FM from the resource data.

        Takes the same kwargs as __init__.

        :param dict rs_mnode_data: Data that corresponds to ResourceManagementNode
        :return: Mnode FM instance
        :rtype: valence.functional_models.ManagementNodeLegacy
        """
        rs_mnode_dm = models.ResourceManagementNode(rs_mnode_data)
        rs_mnode_dm.validate()
        return cls(rs_mnode_dm)

    @classmethod
    def get_mock_fm(cls):
        """Factory method for instantiating a mock node FM.

        No response validation will be done by default. This is purely for unit testing purposes.

        :rtype: valence.functional_models.ManagementNodeLegacy
        """
        return cls(models.ResourceManagementNode.get_mock_object())
