"""Base class for Management Node functional model."""

import json

import autologger

from sf_api import api_constants
from valence import models

_log = autologger.get_sflogging()


class ProcessedHealthchecks(object):
    """Class that holds the parsed checks data.

    This will get rewritten later once we add fully flushed out REST checks connector, which
    will happen after we retrofit all existing healthcheck tests to valence.
    For now we're keeping the same model we've used in prevalence, where this class is replacing
    the ura_check.parsed_checks named tuple.

    :param list[valence.models.Healthcheck]|None cluster_checks: List of cluster checks.
    :param dict[int,list[valence.models.Healthcheck]]|None node_checks: dict of node ID and per-node data
    :param str|None stderr: stderr from the healthcheck command b/c that can contain useful info.
    :param str|None stdout: stdout from the healthcheck command b/c that can contain useful info.
    """

    def __init__(self, cluster_checks=None, node_checks=None, stdout=None, stderr=None):
        self.cluster_checks = cluster_checks
        self.node_checks = node_checks
        self.stderr = stderr
        self.stdout = stdout

    @classmethod
    def from_json(cls, json_text):
        """Class method to instantiate object by parsing json from the stdout of sfupgradecheck.

        We will probably change this mechanism once we retrofit the legacy code to valence and then
        start adding REST implementation. Right now it's a bit hard to see whether this or just a flat
        list will work better.

        :param str json_text: the stdout of sfupgradecheck.
        :returns: cluster_checks, node_checks_by_id
        :rtype: valence.functional_models.ProcessedHealthchecks
        """
        cluster = []
        nodes = {}
        try:
            result = json.loads(json_text)[api_constants.RESULT]
            nodes = {node[api_constants.NODE_ID]: [models.Healthcheck(check) for check in
                                                   node[api_constants.RESULT]]
                     for node in result[api_constants.NODES]}
            cluster = [models.Healthcheck(check) for check in result['clusterChecks']]
        except ValueError:
            # We can't decode JSON so its OK.
            pass
        return cls(cluster_checks=cluster, node_checks=nodes)
