import os
import unittest

import mock
import pytest
import sf_ssh

from valence import (
    functional_models,
    models,
)

CWD, _ = os.path.split(os.path.abspath(os.path.expanduser(__file__)))
DATA = os.path.join(CWD, 'data')
CHECKS = os.path.join(DATA, 'checks.json')


@pytest.yield_fixture(scope='function')
def mock_ssh_run_command(monkeypatch):
    """Returns a mock object of sf_ssh.Ssh.run_command.

    This mocks Ssh.close and Ssh._init_client so it works with the sf_ssh context manager.
    """
    monkeypatch.setattr(sf_ssh.Ssh, 'close', mock.Mock())
    monkeypatch.setattr(sf_ssh.Ssh, '_init_client', mock.Mock())
    patcher = mock.patch('sf_ssh.Ssh.run_command')
    yield patcher.start()
    patcher.stop()


class TestMnodeAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mnode_dm = models.ResourceManagementNode().get_mock_object()
        cls.cluster_dm = models.ResourceCluster.create(mvip='127.0.0.1', username='yoda',
                                                       password='letyouiniwill')
        cls.mnode_fm = functional_models.ManagementNodeLegacy.from_resource_data(cls.mnode_dm)

    def test_cli_adapter_build_cmd(self):
        cmd = self.mnode_fm._build_sfupgradecheck_cmd(self.cluster_dm,
                                                      show_help=True,
                                                      show_version=True,
                                                      run_checks_against_asset=True,
                                                      skip_version_check=True,
                                                      sfhealthtools_url_to_use='http://yada',
                                                      metadata='blah.json')

        assert cmd == "export SFHEALTHTOOLS_URL=\"http://yada\" &&  sudo -E sfupgradecheck " \
                      '--json-output --help --version --skip-version-check ' \
                      '-u yoda -p letyouiniwill 127.0.0.1 --metadata=blah.json'

        deploy = [(0, '1.2.3.4'), (1, '2.3.4.5')]
        cmd = self.mnode_fm._build_sfupgradecheck_cmd(self.cluster_dm,
                                                      show_version=True,
                                                      run_checks_against_asset=True,
                                                      skip_version_check=True,
                                                      sfhealthtools_url_to_use='http://yada',
                                                      metadata='blah.json',
                                                      deploy_only_versions=deploy)

        assert cmd == "export SFHEALTHTOOLS_URL=\"http://yada\" &&  sudo -E sfupgradecheck " \
                      '--json-output --version --skip-version-check ' \
                      '-u yoda -p letyouiniwill 127.0.0.1 --deploy-only 0:1.2.3.4,1:2.3.4.5 ' \
                      '--metadata=blah.json'

    def test_start_storage_cluster_upgrade(self):
        with pytest.raises(NotImplementedError):
            self.mnode_fm.start_storage_cluster_upgrade(
                self.cluster_dm, models.Build.get_mock_object())

    def test_pause_storage_cluster_upgrade(self):
        with pytest.raises(NotImplementedError):
            self.mnode_fm.pause_storage_cluster_upgrade(self.cluster_dm)

    def test_kill_storage_cluster_upgrade(self):
        with pytest.raises(NotImplementedError):
            self.mnode_fm.kill_storage_cluster_upgrade(self.cluster_dm)

    def test_status_storage_cluster_upgrade(self):
        with pytest.raises(NotImplementedError):
            self.mnode_fm.get_storage_cluster_upgrade_status(self.cluster_dm)


@pytest.mark.usefixtures('mock_ssh_run_command')
def test_cli_adapter_ssh_no_error(mock_ssh_run_command):
    with open(CHECKS) as f:
        checks_text = f.read()

    mock_ssh_run_command.return_value = sf_ssh.RunCommandOutput(
        stdout=checks_text, stderr='', return_code=0)

    mnode_dm = models.ResourceManagementNode.create(mip='127.0.0.1')
    cluster_dm = models.ResourceCluster()
    mnode_fm = functional_models.ManagementNodeLegacy(mnode_dm)

    processed_checks = mnode_fm.execute_preupgrade_health_check(cluster_dm)
    assert isinstance(processed_checks, functional_models.ProcessedHealthchecks)
    assert processed_checks.cluster_checks
    assert processed_checks.node_checks
    assert not processed_checks.stderr


@pytest.mark.usefixtures('mock_ssh_run_command')
def test_cli_adapter_ssh_error(mock_ssh_run_command):
    with open(CHECKS) as f:
        checks_text = f.read()

    mock_ssh_run_command.return_value = sf_ssh.RunCommandOutput(
        stdout=checks_text, stderr='', return_code=1)

    mnode_dm = models.ResourceManagementNode.create(mip='127.0.0.1')
    cluster_dm = models.ResourceCluster()
    mnode_fm = functional_models.ManagementNodeLegacy(mnode_dm)

    with pytest.raises(sf_ssh.SshResponseError):
        mnode_fm.execute_preupgrade_health_check(cluster_dm, raise_on_failure=False)
