"""Unit tests for Node Functional Models."""

import unittest

import mock
import pytest
import sf_models

import sf_api
from sf_api import (
    api_constants,
    api_rest_constants,
    enums,
)
from valence import (
    functional_models,
    models,
    utils,
)
from valence.functional_models import _errors

get_asset_result_storage = [
    {
        'compute': [],
        'config': {},
        'controller': [],
        'id': '9bbcc4c7-e8a6-488f-842c-cc27a2c435b5',
        'name': '10.117.78.87',
        'ots': [],
        'storage': [
            {
                '_links': {
                    'collection':
                        'https://10.116.157.240/mnode/1/assets/9bbcc4c7-e8a6-488f-842c-cc27a2c435b5/',
                    'parent':
                        'https://10.116.157.240/mnode/1/assets/9bbcc4c7-e8a6-488f-842c-cc27a2c435b5',
                    'register':
                        'https://10.116.157.240/mnode/1/assets/9bbcc4c7-e8a6-488f-842c-cc27a2c435b5/',
                    'root':
                        'https://10.116.157.240/mnode/1/assets',
                    'self':
                        'https://10.116.157.240/mnode/1/assets/9bbcc4c7-e8a6-488f-842c-cc27a2c435b5/'
                },
                'config': {},
                'credentialid': 'ac99f4dc-8029-43c0-a4be-2d9685b7ebc2',
                'host_name': 'string',
                'id': '917291a4-734e-4bb3-88c7-24bc615852f1',
                'ip': '10.117.78.87',
                'parent': '9bbcc4c7-e8a6-488f-842c-cc27a2c435b5',
                'ssl_certificate': 'SIllyPutty'
            }
        ],
        'telemetry_active': True
    }
]

post_asset = [
    {
        'compute': [],
        'config': {},
        'controller': [],
        'id': '9bbcc4c7-e8a6-488f-842c-cc27a2c435b5',
        'name': 'Blahblah',
        'ots': [],
        'storage': [],
        'telemetry_active': True
    }
]

post_storage_asset = {
    '_links': {
        'collection': 'https://10.116.159.49/mnode/1/assets/39668365/storage-clusters',
        'parent': 'https://10.116.159.49/mnode/1/assets/39668365-e4fa-4eff-ad1b-8ff238be5681',
        'register': 'https://10.116.159.49/mnode/1/assets/39668365/storage-clusters/4054abf9/register',
        'root': 'https://10.116.159.49/mnode/1/assets',
        'self': 'https://10.116.159.49/mnode/1/assets/39668365/storage-clusters/4054abf9'
    },
    'config': {},
    'credentialid': 'e3f3b736-14e3-4f54-ba8f-63b27a2b44fc',
    'host_name': None,
    'id': '4054abf9-0dd1-4bd5-b019-a70b7e38d44b',
    'ip': '10.117.78.87',
    'parent': '39668365-e4fa-4eff-ad1b-8ff238be5681',
    'ssl_certificate': None
}


class TestMnode2xs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        res_mnode = models.ResourceManagementNode().get_mock_object(include_optional=True)
        cls.mnode_fm = functional_models.ManagementNode2x.from_resource_data(res_mnode)
        cls.cluster_dm = models.ResourceCluster.create(mvip='10.117.78.87')
        mock_fm = functional_models.ManagementNode2x.get_mock_fm()

    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_adapter_get_method(self, mock_get):
        mock_rest_response = mock.Mock(sf_api.RestResponse)
        mock_rest_response.result = get_asset_result_storage
        mock_get.return_value = mock_rest_response
        storage_id = self.mnode_fm.get_storage_asset_id_for_storage_cluster_mvip(self.cluster_dm)
        assert storage_id == '917291a4-734e-4bb3-88c7-24bc615852f1'

    @mock.patch.object(sf_api.BaseRestClient, 'post')
    def test_adapter_post_method(self, mock_post):
        mock_rest_response = mock.MagicMock()
        mock_rest_response.result_dm.healthcheck_id = '42'
        mocked_links = {'self': 'http://some_weird_url'}
        mock_rest_response.result_dm.links.__getitem__.side_effect = mocked_links.__getitem__
        mock_post.return_value = mock_rest_response
        healthcheck_id, self_url = self.mnode_fm._start_healthcheck('1')
        assert healthcheck_id == '42'
        assert self_url == 'http://some_weird_url'

    @mock.patch.object(functional_models.ManagementNode2x, '_start_healthcheck')
    @mock.patch.object(functional_models.ManagementNode2x,
                       'get_storage_asset_id_for_storage_cluster_mvip')
    @mock.patch.object(functional_models.ManagementNode2x, '_wait_for_healthcheck_to_finish')
    def test_execute_healthchecks(self, mock_wait, mock_get, mock_start):
        mock_get.return_value = '1'
        mock_start.return_value = '42'
        response = mock.Mock(models.HealthcheckResultModel)
        mock_wait.return_value = response
        checks = self.mnode_fm.execute_preupgrade_health_check(self.cluster_dm)
        assert checks == response

    @mock.patch.object(functional_models.ManagementNode2x,
                       'get_storage_asset_id_for_storage_cluster_mvip')
    def test_execute_healthchecks_no_asset(self, mock_get):
        mock_get.return_value = None
        with pytest.raises(functional_models.FunctionalManagementNodeError):
            self.mnode_fm.execute_preupgrade_health_check(self.cluster_dm)
        assert mock_get.call_count == 1

    @mock.patch('time.sleep')
    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_wait_for_healthcheck_to_finish(self, mock_get, mock_sleep):
        mock_rest_response_1 = mock.Mock(sf_api.RestResponse)
        mock_rest_response_1.result_dm = mock.Mock()
        mock_rest_response_1.result_dm.state = enums.HealthcheckState.initializing

        mock_rest_response_2 = mock.Mock(sf_api.RestResponse)
        mock_rest_response_2.result_dm = mock.Mock()
        mock_rest_response_2.result_dm.state = enums.HealthcheckState.running

        mock_rest_response_3 = mock.Mock(sf_api.RestResponse)
        mock_rest_response_3.result_dm = mock.Mock()
        mock_rest_response_3.result_dm.state = enums.HealthcheckState.complete

        mock_get.side_effect = [mock_rest_response_1, mock_rest_response_2, mock_rest_response_3]
        mock_sleep.return_value = None
        self.mnode_fm._wait_for_healthcheck_to_finish('1')

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_timed_out(self, mock_get, mock_timeout_timer, mock_sleep):
        mock_timeout_timer.return_value = iter([True, False])

        mock_rest_response_1 = mock.Mock(sf_api.RestResponse)
        mock_rest_response_1.result_dm = mock.Mock()
        mock_rest_response_1.result_dm.state = 'initializing'

        mock_rest_response_2 = mock.Mock(sf_api.RestResponse)
        mock_rest_response_2.result_dm = mock.Mock()
        mock_rest_response_2.result_dm.state = 'running'

        mock_get.side_effect = [mock_rest_response_1, mock_rest_response_2]
        mock_sleep.return_value = None

        with pytest.raises(_errors.HealthcheckError):
            self.mnode_fm._wait_for_healthcheck_to_finish('1')

    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_get_storage_id_for_storage_cluster_wrong_mvip(self, mock_get):
        mock_rest_response = mock.Mock(sf_api.RestResponse)
        mock_rest_response.result = get_asset_result_storage
        mock_get.return_value = mock_rest_response
        fake_cluster_dm = models.ResourceCluster.create(mvip='127.0.0.1')
        storage_id = self.mnode_fm.get_storage_asset_id_for_storage_cluster_mvip(fake_cluster_dm)
        assert not storage_id

    @mock.patch.object(utils, 'generate_random_string')
    def test_generate_unique_asset_name(self, mock_gen_rnd):
        mock_gen_rnd.side_effect = ['abcdefgh', '12345678']
        existing_names = ['asdfasd', 'abcdefgh', 'blah']
        new_name = self.mnode_fm._generate_unique_asset_name(existing_names)
        assert new_name == '12345678'
        assert mock_gen_rnd.call_count == 2

    @mock.patch.object(sf_api.BaseRestClient, 'post')
    def test_add_asset(self, mock_post):
        mock_rest_response = mock.Mock(sf_api.RestResponse)
        mock_rest_response.result = post_asset
        mock_rest_response.result_dm.asset_id = '9bbcc4c7-e8a6-488f-842c-cc27a2c435b5'
        mock_post.return_value = mock_rest_response
        asset_id = self.mnode_fm._add_asset('blah')
        assert asset_id == '9bbcc4c7-e8a6-488f-842c-cc27a2c435b5'

    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_get_existing_assets(self, mock_get):
        mock_rest_response = mock.Mock(sf_api.RestResponse)
        mock_rest_response.result = get_asset_result_storage
        mock_get.return_value = mock_rest_response
        existing_assets = self.mnode_fm.get_all_assets()
        assert existing_assets[0].asset_id == '9bbcc4c7-e8a6-488f-842c-cc27a2c435b5'

    @mock.patch.object(sf_api.BaseRestClient, 'post')
    def test_add_storage_to_asset(self, mock_post):
        mock_rest_response = mock.Mock(sf_api.RestResponse)
        mock_rest_response.result = post_storage_asset
        mock_rest_response.result_dm.asset_id = 'some strange id'
        mock_post.return_value = mock_rest_response
        storage_id = self.mnode_fm.add_storage_id_to_asset('somerandomasset', self.cluster_dm)
        assert storage_id == 'some strange id'
        assert 'assets/somerandomasset/storage-clusters' == mock_post.call_args_list[0][1]['path']

    @mock.patch.object(functional_models.ManagementNode2x, 'get_all_assets')
    @mock.patch.object(functional_models.ManagementNode2x, '_generate_unique_asset_name')
    @mock.patch.object(functional_models.ManagementNode2x, '_add_asset')
    def test_add_asset_if_needed_no_name_no_asset(self, mock_add, mock_generate, mock_get):
        get_assets_first = [models.AssetResult.create(name='a', id='blah'),
                            models.AssetResult.create(name='b', id='yada')]
        get_assets_second = [models.AssetResult.create(name='a', id='blah'),
                             models.AssetResult.create(name='b', id='yada'),
                             models.AssetResult.create(name='random_name', id='hey')]
        mock_get.side_effect = [get_assets_first, get_assets_second]
        mock_generate.return_value = 'random_name'
        asset_uuid = self.mnode_fm.add_asset_if_needed()
        assert asset_uuid == 'hey'
        assert mock_add.call_count == 1
        assert mock_generate.call_count == 1
        assert mock_get.call_count == 2

    @mock.patch.object(functional_models.ManagementNode2x, 'get_all_assets')
    @mock.patch.object(functional_models.ManagementNode2x, '_generate_unique_asset_name')
    @mock.patch.object(functional_models.ManagementNode2x, '_add_asset')
    def test_add_asset_if_needed_didnt_add(self, mock_add, mock_generate, mock_get):
        get_assets_first = [models.AssetResult.create(name='a', id='blah'),
                            models.AssetResult.create(name='b', id='yada')]
        mock_get.side_effect = [get_assets_first, get_assets_first]
        mock_generate.return_value = 'random_name'
        with pytest.raises(functional_models._errors.FunctionalManagementNodeError):
            self.mnode_fm.add_asset_if_needed()
        assert mock_add.call_count == 1

    @mock.patch.object(functional_models.ManagementNode2x, 'get_all_assets')
    @mock.patch.object(functional_models.ManagementNode2x, '_generate_unique_asset_name')
    @mock.patch.object(functional_models.ManagementNode2x, '_add_asset')
    def test_add_asset_if_needed_with_name_no_asset(self, mock_add, mock_generate, mock_get):
        get_assets_first = [models.AssetResult.create(name='a', id='blah'),
                            models.AssetResult.create(name='b', id='yada')]
        get_assets_second = [models.AssetResult.create(name='a', id='blah'),
                             models.AssetResult.create(name='b', id='yada'),
                             models.AssetResult.create(name='random_name', id='hey')]
        mock_get.side_effect = [get_assets_first, get_assets_second]
        asset_uuid = self.mnode_fm.add_asset_if_needed('random_name')
        assert asset_uuid == 'hey'
        assert mock_add.call_count == 1
        assert mock_generate.call_count == 0
        assert mock_get.call_count == 2

    @mock.patch.object(functional_models.ManagementNode2x, 'get_all_assets')
    @mock.patch.object(functional_models.ManagementNode2x, '_generate_unique_asset_name')
    @mock.patch.object(functional_models.ManagementNode2x, '_add_asset')
    def test_add_asset_if_needed_with_name_in_asset(self, mock_add, mock_generate, mock_get):
        get_assets_first = [models.AssetResult.create(name='a', id='blah'),
                            models.AssetResult.create(name='b', id='yada'),
                            models.AssetResult.create(name='existing_name', id='hey')]
        get_assets_second = [models.AssetResult.create(name='a', id='blah'),
                             models.AssetResult.create(name='b', id='yada'),
                             models.AssetResult.create(name='existing_name', id='hey')]
        mock_get.side_effect = [get_assets_first, get_assets_second]
        asset_uuid = self.mnode_fm.add_asset_if_needed('existing_name')
        assert asset_uuid == 'hey'
        assert mock_add.call_count == 0
        assert mock_generate.call_count == 0
        assert mock_get.call_count == 1

    @mock.patch.object(sf_api.BaseRestClient, 'delete')
    def test_delete_storage_from_asset(self, mock_delete):
        self.mnode_fm._delete_storage_from_asset('somerandomasset', 'sid')
        assert 'assets/somerandomasset/storage-clusters/sid' == mock_delete.call_args_list[0][1]['path']

    @mock.patch.object(sf_api.BaseRestClient, 'delete')
    def test_delete_asset(self, mock_delete):
        self.mnode_fm._delete_asset('somerandomasset')
        assert 'assets/somerandomasset' == mock_delete.call_args_list[0][1]['path']

    @mock.patch.object(sf_api.BaseRestClient, 'put')
    def test_services_upload(self, mock_put):
        bundle_data = {'foo': 'bar'}
        self.mnode_fm.services_upload(service_bundle_data=bundle_data, validate_version=True)
        mock_put.assert_called_once()
        mock_put.assert_called_with(service=api_rest_constants.MNODE,
                                    path='{}/{}'.format(api_constants.SERVICES,
                                                        api_rest_constants.UPLOAD),
                                    data={'validate_version': True, 'data': bundle_data},
                                    result_dm_class=sf_models.ServicesUpload)

    @mock.patch.object(functional_models.ManagementNode2x, 'get_all_assets')
    @mock.patch.object(sf_api.BaseRestClient, 'delete')
    def test_clear_assets(self, mock_delete, mock_get):
        get_storage_a = [models.Asset.create(id='storage_a_1'),
                         models.Asset.create(id='storage_a_2')]
        get_storage_b = [models.Asset.create(id='storage_b_1'),
                         models.Asset.create(id='storage_b_2')]
        get_assets = [models.AssetResult.create(id='blah', storage=get_storage_a),
                      models.AssetResult.create(id='yada', storage=get_storage_b),
                      models.AssetResult.create(id='hey')]
        mock_get.return_value = get_assets
        self.mnode_fm.delete_all_assets()
        assert mock_get.call_count == 1
        assert mock_delete.call_count == 7
        assert mock_delete.call_args_list[0][1]['path'] == 'assets/blah/storage-clusters/storage_a_1'
        assert mock_delete.call_args_list[1][1]['path'] == 'assets/blah/storage-clusters/storage_a_2'
        assert mock_delete.call_args_list[2][1]['path'] == 'assets/blah'
        assert mock_delete.call_args_list[3][1]['path'] == 'assets/yada/storage-clusters/storage_b_1'
        assert mock_delete.call_args_list[4][1]['path'] == 'assets/yada/storage-clusters/storage_b_2'
        assert mock_delete.call_args_list[5][1]['path'] == 'assets/yada'
        assert mock_delete.call_args_list[6][1]['path'] == 'assets/hey'

    @mock.patch.object(sf_api.BaseRestClient, 'post')
    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_start_cluster_upgrade(self, mock_get, mock_post):
        options = models.UpgradeConfig({'ignoredClusterFaults': ['drive_wear_fault']})
        storage_id = '917291a4-734e-4bb3-88c7-24bc615852f1'
        mock_rest_response = mock.Mock(sf_api.RestResponse)
        mock_rest_response.result = get_asset_result_storage
        mock_get.return_value = mock_rest_response
        self.mnode_fm.start_storage_cluster_upgrade(cluster_dm=self.cluster_dm,
                                                    build_url='www.some-url/package',
                                                    sfinstall_options=options)
        assert mock_post.call_count == 1
        assert mock_get.call_count == 1
        assert mock_post.call_args_list[0][1]['request_json']['storageId'] == storage_id

    @mock.patch.object(sf_api.BaseRestClient, 'put')
    def test_pause_cluster_upgrade(self, mock_put):
        upgrade_id = 'some_random_id'
        self.mnode_fm.pause_storage_cluster_upgrade(upgrade_id)
        assert mock_put.call_count == 1

    @mock.patch.object(sf_api.BaseRestClient, 'put')
    def test_resume_cluster_upgrade(self, mock_put):
        upgrade_id = 'some_random_id'
        self.mnode_fm.resume_storage_cluster_upgrade(upgrade_id)
        assert mock_put.call_count == 1

    @mock.patch.object(sf_api.BaseRestClient, 'put')
    def test_abort_cluster_upgrade(self, mock_put):
        upgrade_id = 'some_random_id'
        self.mnode_fm.abort_storage_cluster_upgrade(upgrade_id)
        assert mock_put.call_count == 1

    @mock.patch.object(sf_api.BaseRestClient, 'get')
    def test_get_upgrade_status(self, mock_get):
        upgrade_id = 'some_random_id'
        self.mnode_fm.get_storage_cluster_upgrade(upgrade_id)
        mock_get.assert_called_with(params={'upgradeID': upgrade_id}, path='upgrades',
                                    result_dm_class=models.UpgradeResource, service='storage')
