"""Tests for sfupdate-healthtools"""

import time
import unittest

import mock
import pytest
import sf_ssh

from valence import (
    functional_models,
    models,
    unittest_utils,
    vl_helpers,
)

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION = 'Currently installed version of HealthTools:'
HEALTHTOOLS_DOWNLOAD_INSTRUCTIONS = 'To download the latest version of the HealthTools, go to:'
MGMT_PLTFRM = 'Management Platform Node: Ember'

STD_ERR = """
Currently installed version of HealthTools:
2019.06.01.105
Management Platform Node: Ember
The latest version of HealthTools can be downloaded from NetApp Support Site
Go to https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""

STD_OUT = """
Currently installed version of HealthTools:
2019.06.01.102
Management Platform Node: Ember
To download the latest version of the HealthTools, go to:
https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""

STD_OUT_SFUPL = """
Latest available version of HealthTools:
2019.06.01.107
Management Platform Node: Ember
The latest version of HealthTools can be downloaded from NetApp Support Site
Go to https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""

STD_OUT_HT_19_2 = """
Currently installed version of HealthTools:
2019.12.01.0
Management Platform Node: Ember
"""

STD_ERR_OUTPUT_LIST_LEN = """
Latest available version of HealthTools:
2019.06.01.107
Currently installed version of HealthTools:
Management Platform Node: Ember
"""


class TestSFUpdateHealthToolsMnodeLegacy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mnode_fm = functional_models.ManagementNodeLegacy.get_mock_fm()
        cls.nodes_dm = [models.ActiveNode.get_mock_object(overrides={'node_id': i})
                        for i in range(1, 5)]

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_run_cmd(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0)
        self.mnode_fm.run_sfupdate_healthtools_cmd()

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_run_cmd_raise(self, mock_run_command):
        version = models.GenericVersion('2019.06.01.102')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=1)
        with pytest.raises(sf_ssh.SshResponseError):
            self.mnode_fm.run_sfupdate_healthtools_cmd(
                show_help=True,
                sfhealthtools_url_to_use='http://jenkins.eng.solidfire.net/healthtools-2019.06.01...',
                show_all_previous_versions=True,
                show_installed_version=True,
                show_latest_version=True,
                revert_to_this_version=version,
                metadata='http://jenkins.eng.solidfire.net/healthtools-metadata.json'
            )

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_install_new_version(self, mock_run_command):
        new_version = models.GenericVersion('2019.06.01.102')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0)
        self.mnode_fm.install_new_healthtools_version(new_version)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_install_new_version_raise(self, mock_run_command):
        new_version = models.GenericVersion('2019.06.01.102')
        og_version = models.GenericVersion('2019.06.01.102')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=1)
        with pytest.raises(sf_ssh.SshResponseError):
            self.mnode_fm.install_new_healthtools_version(new_version, og_version)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_install_new_version_unequal_versions(self, mock_run_command):
        new_version = models.GenericVersion('2019.06.01.102')
        og_version = models.GenericVersion('2019.06.01.105')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=1)
        with pytest.raises(sf_ssh.SshResponseError):
            self.mnode_fm.install_new_healthtools_version(new_version, og_version)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_get_current_version(self, mock_run_command):
        new_version = models.GenericVersion('2019.12.01.0')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout=STD_OUT_HT_19_2, stderr='', return_code=0)
        curr_version = self.mnode_fm.get_current_healthtools_version()
        assert curr_version == new_version
