"""Base class for Node functional model."""

import autologger
import b2sdk
import requests
import sf_mc
import sf_ssh

import sf_api
from sf_api import (
    api_constants,
    enums,
)
from valence import models
from valence.functional_models import (
    _constants,
    _errors,
)
from valence.functional_models._cluster import cluster

_log = autologger.get_sflogging()


class NodeBase(object):
    """Node Functional Model base class.

    :param valence.models.ResourceNode node_dm: node data model
    :param valence.models.ResourceCluster|None cluster_dm: Optional cluster data model. If given,
        the node will authenticate using the cluster's credentials. For an orphaned node, set this
        argument to None. If the node is orphaned, you must call node_fm.authenticate(cluster_dm) to
        authenticate the node once it is added to the cluster.
    :param bool authenticate: If True, check if node can be authenticated via HTTP in constructor.
    :param bool validate_responses: If True, validate responses from SF Api (default=True)
    """

    def __init__(self, node_dm, cluster_dm, authenticate=False, validate_responses=True):
        self.node_dm = node_dm
        self.cluster_dm = cluster_dm
        self.validate_responses = validate_responses
        # Flag to log the full JSON response from sf_api
        self.log_responses = True
        if cluster_dm is not None:
            username = cluster_dm.username
            password = cluster_dm.password
            version = cluster_dm.api_endpoint
        else:
            username = password = version = None
        self.api = sf_api.NodeApi(self.mip, username, password, version=version, authenticate=False)
        if authenticate:
            self.authenticate()

        self._bourbon = None
        self.vsphere = None
        # When vsphere is used as a context manager, allows control over the witness node VM.
        if isinstance(self.node_dm, models.ResourceNode) and self.node_dm.is_witness():
            if self.node_dm.hypervisor_host:
                hypervisor_host = self.node_dm.hypervisor_host
                self.vsphere = sf_mc.esx_control.Vsphere(hypervisor_host.management_ip,
                                                         hypervisor_host.username,
                                                         hypervisor_host.password)
            else:
                _log.info('Hypervisor host not set for this witness node. Not all FM methods will work.')

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format('NodeFm',
                                         'mip', self.node_dm.mip,
                                         'mvip', self.cluster_dm.mvip if self.cluster_dm else None)

    @property
    def mip(self):
        """Shortcut for mip attribute on node model.

        :rtype: str
        """
        return self.node_dm.mip

    @property
    def static_id(self):
        """Shortcut for static_id attribute on node model.

        :rtype: str
        """
        return self.node_dm.static_id

    @property
    def bourbon(self):
        """Returns b2sdk object.

        It reuses a previously created object if present, or creates a new
        b2sdk.B2SDK object if it doesnt already exist

        :rtype: b2sdk.B2SDK
        """
        if not self._bourbon:
            self._bourbon = b2sdk.B2SDK(username=_constants.BOURBON_USR_NAME,
                                        password=_constants.BOURBON_PASSWORD)

        return self._bourbon

    @property
    def ip(self):
        """The IP the call the api over.

        :rtype: str
        """
        return self.api.ip

    @ip.setter
    def ip(self, value):
        """Set IP the call the api over.

        By default, this will be the same as the mip. In some special cases, the IP needs to be
        overridden temporarily. For example, when changing the mip or taking down the Bond1G interface.

        :param str value: The ip to use
        """
        self.api.ip = value
        self.api._auth.ip = value  # pylint: disable=protected-access

    def _call_method(self, method, params=None, result_dm_class=None, log_response=None):
        """Call the API.

        This method isolates the functional model from changes in sf_api.NodeMethod.call_method.

        :param valence.enums.NodeMethod method: method to call
        :param dict params: the params for the method. Enums will be converted to strings
        :param Any result_dm_class: The API Response class the result is modeled after. If there is no
            result data model associated with the call, this can be None.
        :param bool log_response: Option to log response output. If None, default to the value of
            self.log_responses.
        :rtype: sf_api.SfResponse
        """
        if log_response is None:
            log_response = self.log_responses

        auth_object = self.api.get_auth()
        if self.cluster_dm:
            auth_object.version = self.cluster_dm.api_endpoint
        response = self.api.call_method(method, params=params, result_dm_class=result_dm_class,
                                        log_response=log_response, auth=auth_object)
        if self.validate_responses and response.result_dm is not None:
            version = None
            if self.cluster_dm:
                cluster_fm = cluster.Cluster(self.cluster_dm, authenticate=False)
                # Disable version validation if the cluster is currently upgrading
                version = None if cluster_fm.upgrading else self.cluster_dm.api_endpoint
            response.result_dm.validate(version)
        return response

    def api_supported(self, method):
        """See if the node supports the given API method.

        :param valence.enums.NodeMethod method: Method to check for support
        :return: True if supported; False if not supported or not applicable
        :rtype: bool
        :raises valence.functional_models.FunctionalNodeError: If the resource cluster is out of sync
            with the actual version the cluster is RTFI'ed to.
        """
        response = self.get_api(show_internal=True)
        current_version = response[api_constants.CURRENT_VERSION]
        version = str(self.api.version)
        if str(version) != current_version:
            raise _errors.FunctionalNodeError('Resource cluster is version {}, but cluster is at '
                                              'version {}'.format(version, current_version))
        available_apis = response[version]
        return method.value in available_apis

    def authenticate(self, cluster_dm=None):
        """Verify that node can be authenticated via HTTP.

        This method does not need to be called before making any API calls. It simply verifies that
        the authentication credentials are valid. If a cluster data model is given, self.cluster_dm will
        be overridden and the updated credentials will be used.

        :param valence.models.ResourceCluster cluster_dm: cluster data model that will be used to
            override the existing one. This should only be used when re-authenticating a node with
            a different cluster admin. This will set self.cluster_dm and re-instantiate self.api.
        """
        if cluster_dm is not None:
            self.cluster_dm = cluster_dm
            self.api = sf_api.NodeApi(self.mip, self.cluster_dm.username,
                                      self.cluster_dm.password, version=cluster_dm.api_endpoint,
                                      authenticate=False)
        self.api.authenticate()

    def get_ssh_connection(self):
        """Convenience method for returning the SSH connection for the node.

        Usage::

            with self.get_ssh_connection() as ssh_conn:
                output = ssh_conn.run_command("ls /")

        :rtype: sf_ssh.Ssh
        """
        return sf_ssh.Ssh(self.mip, username=self.node_dm.username)

    def get_node_dm(self):
        """Query the cluster to get an up to date node data model.

        This can return multiple node types, including ActiveNode, PendingActiveNode or PendingNode.

        :raises: valence.functional_models.FunctionalNodeError: If there are no cluster credentials.
        :rtype: valence.models.ActiveNode|valence.models.PendingNode|valence.models.PendingActiveNode
        """
        if not self.cluster_dm:
            raise _errors.FunctionalNodeError('No cluster credentials are associated with this node '
                                              'functional model.')
        cluster_fm = cluster.Cluster(self.cluster_dm)
        nodes = cluster_fm.list_all_nodes()
        node = [n for n in nodes if n.static_id == self.node_dm.static_id][0]
        return node

    @classmethod
    def get_mock_fm(cls, validate_responses=False):
        """Factory method for instantiating a mock node FM.

        No response validation will be done by default. This is purely for unit testing purposes.

        :param bool validate_responses: validate the data models from the api
        :rtype: valence.functional_models.Node
        """
        rs_node_dm = models.ResourceNode.get_mock_object()
        rs_cluster_dm = models.ResourceCluster.get_mock_object()
        return cls(rs_node_dm, cluster_dm=rs_cluster_dm, authenticate=False,
                   validate_responses=validate_responses)

    def update_api_version(self):
        """Update this node FM with the current version.

        This can be called after running an upgrade to update the version used to call the API.
        """
        response = self.get_api(show_internal=False)
        current_version = response[api_constants.CURRENT_VERSION]  # type: str
        _log.debug('Updating version to {}'.format(current_version))
        self.api.version = sf_api.SfVersion(current_version)

    def get_api(self, show_internal=True):
        """Returns a list of all customer facing API calls, and optionally include internal calls.

        :param bool show_internal: Option to show internal API calls.
        :rtype: dict[str,list|str]
        :returns:
            GetAPIResultObject::

                {
                    6.0: [
                        "AptUpdate",
                        "GetAptSourceLinesFromPackage",
                        "GetClusterConfig",
                        "GetClusterState",
                        "GetConfig",
                        "GetDriveConfig",
                        "GetEnsembleConfig",
                        "GetFirewallRules",
                        "GetHardwareConfig",
                        "GetNTPStatus",
                        "GetNetworkConfig",
                        "GetNetworkInterfaceStatus",
                        "GetPendingOperation",
                        "GetSystemStatus",
                        "GetTime",
                        "GetTracingConfig",
                        .........
                        ]
                 }
        """
        return self._call_method(enums.NodeMethod.get_api,
                                 dict(__includeSecretMethods=show_internal), result_dm_class=None,
                                 log_response=False).result

    def api_is_responsive(self):
        """See if the mvip is responsive.

        :rtype: bool
        :return: True if API is responsive, False if not.
        """
        try:
            self.get_api()
        except (requests.HTTPError, requests.ConnectionError, requests.Timeout):
            return False
        return True
