"""Node FM method for utilizing bourbon 2 SDK."""

import autologger

from sf_api import result_models
from valence import (
    enums,
    models,
)
from valence.functional_models._node import base

_log = autologger.get_sflogging()


class BourbonMixin(base.NodeBase):
    """Node FM switchport and port channel methods."""

    def get_attached_switchports(self):
        """Use bourbon sdk to find connected switchports for this node FM.

        :return: a list of connected switchports to all interfaces on this node
        :rtype: list[valence.models.SwitchPort]
        """
        response = self._call_method(enums.NodeMethod.get_network_config,
                                     result_dm_class=result_models.GetNetworkConfig)
        network_config = response.result_dm.network
        macs = [network_config.bond_10g.mac_address, network_config.bond_1g.mac_address]

        b2_result = self.bourbon.find_switch_ports(macs)

        switch_ports = [models.SwitchPort(sp) for sp in b2_result['response']]
        return switch_ports

    def _flip_switchports(self, switchports, enabled):
        """Turn switchport off or on.

        :param list[valence.models.SwitchPort] switchports: list of switchports to be taken up/down
        :param bool enabled: True to bring up switches False to take them down
        :raises EnvironmentError: if bourbon credentials are not found in env
        """
        bourbon = self.bourbon
        addresses = [sp.ip for sp in switchports]
        interfaces = [sp.port_slot for sp in switchports]
        bourbon.set_interfaces_config(addresses=addresses,
                                      interface_type=enums.BourbonInterfaceType.switch_port.value,
                                      interfaces=interfaces, enabled=enabled)

    def up_switchports(self, switchports):
        """Bring up switchports.

        :param list[valence.models.SwitchPort] switchports: list of switchports to be brought up
        """
        self._flip_switchports(switchports, True)

    def down_switchports(self, switchports):
        """Take down switchports.

        :param list[valence.models.SwitchPort] switchports: list of switchports to be taken down
        """
        self._flip_switchports(switchports, False)

    def set_bond_mode(self, switchports, bond_mode):
        """Set the bond mode of port channels associated with a switchport.

        :param list[valence.models.SwitchPort] switchports: list of switchports to change bond mode on
        :param list[valence.enums.BourbonBondMode] bond_mode: bond mode to switch to.
        :raises EnvironmentError: if bourbon credentials are not found in env.
        """
        addresses = [sp.ip for sp in switchports]
        inter = self.bourbon.get_interfaces(addresses, enums.BourbonInterfaceType.port_channel.value)
        interfaces = []
        for pc in inter['response']:
            interfaces.extend(pc['interfaces'])

        self.bourbon.set_interfaces_config(addresses=addresses,
                                           interface_type=enums.BourbonInterfaceType.port_channel.value,
                                           interfaces=interfaces,
                                           permutate=True,
                                           bond_mode=bond_mode.value)

    def get_vlans(self):
        """Obtains the vlans attached to each switchport for this node.

        This function requires both the ip and the port_slot to be set to meaningful
        values for a correct response. Otherwise it just returns empty values for each
        vlan field in valence.models.SwitchPort object

        :return: a list of switchports with the vlans that are attached to it
        :rtype: list[valence.models.SwitchPort]
        """
        switchport_dms = self.get_attached_switchports()
        switches = []
        interfaces = []
        for switchport in switchport_dms:
            switches.append(switchport.ip)
            interfaces.append(switchport.port_slot)

        b2_result = self.bourbon.get_interfaces_config(switches=switches,
                                                       interface_type=enums.BourbonInterfaceType.
                                                       switch_port.value,
                                                       interfaces=interfaces)

        for response, dm in zip(b2_result['response'], switchport_dms):
            dm.access_native_vlan = response['access_native_vlan']
            dm.trunk_native_vlan = response['trunk_native_vlan']
            dm.tagged_vlans = response['tagged_vlans']
        return switchport_dms

    def set_untagged_vlans(self, switchports, vlan):
        """Sets untagged vlan for each switchport in the arguments.

        :param list[valence.models.SwitchPort] switchports: switchports on which the vlans are to be set
        :param int vlan: vlan which is to be set on the switchports
        :rtype: list[valence.models.SwitchPort]
        """
        switch_ips = []
        port_slots = []
        for switch in switchports:
            switch_ips.append(switch.ip)
            port_slots.append(switch.port_slot)
            switch.trunk_native_vlan = vlan
            switch.access_native_vlan = vlan

        self.bourbon.set_interfaces_config(switches=switch_ips,
                                           interface_type=enums.BourbonInterfaceType.switch_port.value,
                                           interfaces=port_slots, trunk_native_vlan=vlan,
                                           access_native_vlan=vlan)
        return switchports

    def clear_untagged_vlans(self, switchports):
        """Clears all untagged vlans for each switchport in the arguments.

        :param list[valence.models.SwitchPort] switchports: switchports on which the
            vlans have to be unset
        :rtype: list[valence.models.SwitchPort]
        """
        switch_ips = []
        port_slots = []
        vlan = None
        for switch in switchports:
            switch_ips.append(switch.ip)
            port_slots.append(switch.port_slot)
            switch.access_native_vlan = vlan
            switch.trunk_native_vlan = vlan

        self.bourbon.set_interfaces_config(switches=switch_ips,
                                           interface_type=enums.BourbonInterfaceType.switch_port.value,
                                           interfaces=port_slots, trunk_native_vlan=vlan,
                                           access_native_vlan=vlan)
        return switchports

    def set_tagged_vlans(self, switchports, vlans):
        """Sets tagged vlan for each switchport in the arguments.

        :param list[valence.models.SwitchPort] switchports: switchports on which the vlans are to be set
        :param list[int] vlans: vlans which are to be set on the switchports
        :rtype: list[valence.models.SwitchPort]
        """
        switch_ips = []
        port_slots = []
        for switch in switchports:
            switch_ips.append(switch.ip)
            port_slots.append(switch.port_slot)
            switch.tagged_vlans = vlans

        self.bourbon.set_interfaces_config(switches=switch_ips,
                                           interface_type=enums.BourbonInterfaceType.switch_port.value,
                                           interfaces=port_slots, tagged_vlans=vlans)
        return switchports

    def clear_tagged_vlans(self, switchports):
        """Clears all tagged vlans set on this switchport.

        :param list[valence.models.SwitchPort] switchports: switchports on which the
            vlans are to be cleared
        :rtype: list[valence.models.SwitchPort]
        """
        switch_ips = []
        port_slots = []
        vlans = []
        for switch in switchports:
            switch_ips.append(switch.ip)
            port_slots.append(switch.port_slot)

        self.bourbon.set_interfaces_config(switches=switch_ips,
                                           interface_type=enums.BourbonInterfaceType.switch_port.value,
                                           interfaces=port_slots, tagged_vlans=vlans)
