"""Node FM methods related to support bundles."""

import os
import sys
import time

import aenum
import autologger
import requests
import six

from camelot.models import at2_constants
from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    enums,
    models,
)
from valence.functional_models import (
    _bundle_base,
    _errors,
)
from valence.functional_models._node import base

# SUPPORT BUNDLE Constants
SF_BUNDLE_LOCATION = '/tmp/supportbundles'
SUPPORT_BUNDLE_NAME_LENGTH = 100

CORE_FILE = '/sf/core*'
CRASH_DUMP_FILE = '/var/crash/*'

DISK_INFO = 'disk_info'
HARDWARE_XML = 'hardware.xml'
LOGS_SFMASTER_INFO = 'logs/sf-master.info'
LOGS_SFMASTER_ERROR = 'logs/sf-master.error'
LOGS_SFSLICE_INFO = 'logs/sf-slice.info'
LOGS_SFSLICE_ERROR = 'logs/sf-slice.error'
LOGS_SFBLOCK_INFO = 'logs/sf-block.info'
LOGS_SFBLOCK_ERROR = 'logs/sf-block.error'
LOGS_SFCONFIG_INFO = 'logs/sf-config.info'
LOGS_SFCONFIG_ERROR = 'logs/sf-config.error'
LOGS_SFFIBRE_INFO = 'logs/sf-fibre.info'
LOGS_SFFIBRE_ERROR = 'logs/sf-fibre.error'
LOGS_KERN = 'logs/kern.log'
CONFIG_SOLIDFIRE_JSON = 'config/etc/solidfire.json'
CONFIG_NETWORK_JSON = 'config/etc/network/network.json'
CONFIG_RSYSLOG_CONF = 'config/etc/rsyslog.conf'
CONFIG_RTFI_VERSION = 'config/sf/rtfi/version'

SUPPORTED_PLATFORMS = 'supported_platform'

# TODO: not all node models are accounted for. Use NodeType.STANDARD_TYPES.
FILES_TO_CHECK = {
    DISK_INFO: {SUPPORTED_PLATFORMS: at2_constants.NON_FC_NODE_MODELS},
    HARDWARE_XML: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    LOGS_SFMASTER_INFO: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    LOGS_SFMASTER_ERROR: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    LOGS_SFSLICE_INFO: {SUPPORTED_PLATFORMS: at2_constants.NON_FC_NODE_MODELS},
    LOGS_SFSLICE_ERROR: {SUPPORTED_PLATFORMS: at2_constants.NON_FC_NODE_MODELS},
    LOGS_SFBLOCK_INFO: {SUPPORTED_PLATFORMS: at2_constants.NON_FC_NODE_MODELS},
    LOGS_SFBLOCK_ERROR: {SUPPORTED_PLATFORMS: at2_constants.NON_FC_NODE_MODELS},
    LOGS_SFCONFIG_INFO: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    LOGS_SFCONFIG_ERROR: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    LOGS_SFFIBRE_INFO: {SUPPORTED_PLATFORMS: at2_constants.FC_NODE_MODELS},
    LOGS_SFFIBRE_ERROR: {SUPPORTED_PLATFORMS: at2_constants.FC_NODE_MODELS},
    LOGS_KERN: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    CONFIG_SOLIDFIRE_JSON: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    CONFIG_NETWORK_JSON: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    CONFIG_RSYSLOG_CONF: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS},
    CONFIG_RTFI_VERSION: {SUPPORTED_PLATFORMS: at2_constants.NODE_MODELS}
}


# Support bundle enums
class BundleMethod(aenum.Enum):
    """Method used to gather support bundles.

    api - SF API was used. This is the preferred method.
    ssh - The bundles were retrieved via SSH. This is used as a fallback when the API is not available.
    """
    api = 'api'  # type: BundleMethod
    ssh = 'ssh'  # type: BundleMethod


_log = autologger.get_sflogging()


class BundleMixin(base.NodeBase, _bundle_base.BundleBase):
    """Node FM support bundle methods."""

    #######################################################################
    # Public methods
    #######################################################################

    def create_support_bundle(self, bundle_name, extra_args=None, timeout_sec=1800):
        """Create a support bundle via the API.

        :param str bundle_name: Name of the bundle
        :param str extra_args: Extra parameters to give to /sf/scripts/sf_make_support_bundle script
        :param int timeout_sec: Timeout in seconds for sf_make_support_bundle
        :raises valence.functional_models.NodeBundleError: if command failed
        :rtype: sf_api.result_models.CreateSupportBundle
        """
        params = {
            api_constants.BUNDLE_NAME: bundle_name,
            api_constants.TIMEOUT_SEC: timeout_sec,
            api_constants.EXTRA_ARGS: extra_args
        }
        create_bundle_dm = self._call_method(enums.NodeMethod.create_support_bundle, params=params,
                                             result_dm_class=result_models.CreateSupportBundle).result_dm
        if not create_bundle_dm.result.is_passed():
            raise _errors.NodeBundleError('CreateSupportBundle API responded with error: {}'.format(
                create_bundle_dm.details.output))
        return create_bundle_dm

    def delete_all_support_bundles(self):
        """Delete all support bundles on the node via the API.

        :raises valence.functional_models.NodeBundleError: if command failed
        :rtype: sf_api.result_models.DeleteAllSupportBundles
        """
        delete_bundle_dm = self._call_method(
            enums.NodeMethod.delete_all_support_bundles,
            result_dm_class=result_models.DeleteAllSupportBundles).result_dm
        if not delete_bundle_dm.result.is_passed():
            msg = 'DeleteAllSupportBundles failed - {}'.format(delete_bundle_dm.details.output)
            raise _errors.NodeBundleError(self._get_sf_msg(msg))

    def pull_support_bundle(self, artifact_manager, purge_cores=True, purge_crash_dumps=True,
                            compress=True, collect_backtraces=False, zk_heap_dump=False,
                            verify_logs=False):
        """Pull a support bundle for the node.

        :param sf_artifacts.ArtifactManager artifact_manager: Artifact manager object that defines
            where the support bundle will be stored.
        :param bool purge_cores: If True, purge core files
        :param bool purge_crash_dumps: If True, purge crash dump files
        :param bool compress: If True, pass compress flag to sf_make_support_bundle
        :param bool collect_backtraces: If True, collect backtraces
        :param bool zk_heap_dump: If True, do Zookeeper heap dump
        :param bool verify_logs: Verify support bundle logs
        """
        bundle_info = self.create_bundle(compress=compress, collect_backtraces=collect_backtraces,
                                         zk_heap_dump=zk_heap_dump, verify_logs=verify_logs)
        names = artifact_manager.store_artifacts_remote(
            artifacts=[bundle_info.bundle_path], remote_host_dm=self.node_dm
        )

        self._cleanup_remote_node(bundle_info, purge_cores=purge_cores,
                                  purge_crash_dumps=purge_crash_dumps)
        try:
            destination = artifact_manager.artifact_url + list(names)[0]
            _log.debug(self._get_sf_msg('Support bundle collected successfully'
                                        ' and located at: {}'.format(destination)))
        except (IndexError, TypeError):
            _log.error(self._get_sf_msg('Could not get a destination path on Artifact for the bundle!'))

        # Even though bundle collection succeeded, if API to get bundles failed,
        # we need to be failing the get_support_bundles step as a failed API is a product bug.
        api_call_result = [x[1] for x in bundle_info.attempts if x[0] is BundleMethod.api]
        if False in api_call_result:
            raise _errors.NodeBundleError(self._get_sf_msg(
                'Though bundles were collected successfully, CreateSupportBundle API Failed. '
                'Check the output for more information!'))

    def create_bundle(self, compress=True, collect_backtraces=False,  # pylint: disable=too-complex
                      zk_heap_dump=False, verify_logs=False):
        """Create support bundles for the node.

        Will attempt to create the bundle via the API and use SSH as a fallback.

        :param bool compress: If True, pass compress flag to sf_make_support_bundle
        :param bool collect_backtraces: If True, collect backtraces
        :param bool zk_heap_dump: If True, do Zookeeper heap dump
        :param bool verify_logs: Verify support bundle logs
        :rtype: valence.functional_models.BundleInfo
        :returns: A populated bundle info object
        """
        bundle_info = _bundle_base.BundleInfo()
        sfapp_version = self.cluster_dm.api_endpoint
        timestamp = time.strftime('%m%d%Y_%H%M%S')
        bundle_name = '{}_{!s}_{}_node_bundle'.format(self.node_dm.node_name,
                                                      sfapp_version, timestamp)
        bundle_name = self._truncate_bundle_name_if_necessary(bundle_name)
        bundle_info.bundle_name = bundle_name
        _log.debug(self._get_sf_msg('Bundle name to be used {}'.format(bundle_info.bundle_name)))

        # Nitrogen Patch 3 (and only NP3) has a bug where using this flag can
        # cause an EIO. Earlier builds ignore it, later builds will have the fix.
        if collect_backtraces and sfapp_version == models.SfVersion('7.3'):
            msg = 'Ignoring backtraces due to case 16652 (version is {})'.format(sfapp_version)
            _log.debug(self._get_sf_msg(msg))
            collect_backtraces = False

        if sfapp_version < models.SfVersion('5.5'):
            extra_args = ''
        else:
            if compress:
                # --compress flag support was added to sf_make_support_bundle in 9.0.0.883
                if sfapp_version < models.SfVersion('9.0.0.883'):
                    _log.debug(self._get_sf_msg('Not passing --compress to sf_make_support_bundle as'
                                                ' version {} does not support it'.format(sfapp_version)))
                    compress = False

            extra_args = '{} -b -i all {} {}'.format('--compress {}'.format(self.COMPRESSION_TYPE)
                                                     if compress else '',
                                                     '-i zk_heap_dump' if zk_heap_dump else '',
                                                     '-i backtraces' if collect_backtraces else '')

        if self.api_supported(enums.NodeMethod.create_support_bundle):
            methods = (BundleMethod.api, BundleMethod.ssh)
        else:
            methods = (BundleMethod.ssh,)
        exception_info = None
        for method in methods:
            _log.debug(self._get_sf_msg('Method: {}, BundleName: {}, and extra_args: {}'.format(
                method, bundle_info.bundle_name, extra_args)))
            try:
                self._attempt_making_bundle(method, bundle_info, extra_args=extra_args,
                                            verify_logs=verify_logs)
            except _errors.NodeBundleError:
                exc_info = sys.exc_info()
                _log.error(self._get_sf_msg('API or Script Failed.'), exc_info=True)
                if not exception_info:
                    exception_info = exc_info
                bundle_info.attempts.append((method, False))
            else:
                bundle_info.attempts.append((method, True))
                _log.debug(self._get_sf_msg(
                    'Attempted methods and results are: {}'.format(bundle_info.attempts)))
                return bundle_info

        # No methods were successful, so raise an exception
        _log.error(self._get_sf_msg('Failed to get support bundles via all allowed methods: '
                                    '{}'.format(bundle_info.attempts)))
        six.reraise(*exception_info)

    #######################################################################
    # Private methods
    #######################################################################

    def _verify_bundle_removed(self):
        """Verify if the SF_BUNDLE_LOCATION directory is empty.

        :return: True if the directory is empty; False if not
        :rtype: bool
        """
        with self.get_ssh_connection() as ssh_conn:
            cmd = 'ls -1 {}'.format(SF_BUNDLE_LOCATION)
            response = ssh_conn.run_command(cmd)
            return (response.return_code == 0) and len(response.stdout) == 0

    @staticmethod
    def _verify_file(ssh_conn, bundle_folder, to_verify):
        """Verify that the given file exists on the given connection.

        Passing ssh connections around is not nice, but this is too much of a
        pain otherwise.

        :param sf_ssh.Ssh ssh_conn: An open Ssh connection
        :param str bundle_folder: The folder in which the bundle was unpacked.
        :param str to_verify: The relative file path to verify.
        :return: A pair representing (found, non-empty).
        :rtype: tuple[bool, bool]
        """
        file_found = True
        has_contents = True
        file_path = os.path.join(bundle_folder, to_verify)
        # See if the specified file can be found in the specified location
        response = ssh_conn.run_command('stat -c%s {}'.format(file_path))
        if response.return_code != 0:
            # See if the specified file can be found anywhere
            file_name = file_path.split('/')[-1]
            find_cmd = "find {} -type f -name {} | egrep '.*'"
            find_cmd = find_cmd.format(bundle_folder, file_name)
            response2 = ssh_conn.run_command(find_cmd)
            if response2.return_code != 0 or len(response2.stdout) == 0:
                file_found = False
            else:
                actual_path = response2.stdout[0]
                _log.debug('File {} found at {} but it is supposed to be at {}'.format(
                    file_name, actual_path, file_path))
                response = ssh_conn.run_command('stat -c%s {}'.format(actual_path))
        if file_found and response.stdout[0] == '0':
            has_contents = False
        return file_found, has_contents

    def _get_sf_msg(self, msg):
        """Prepends SF Object Info to the Log message.

        :param str msg: Message to be logged

        :rtype: str
        :returns: Prepended sfObject info message
        """
        return 'Node({}): {}'.format(self.mip, msg)

    def _verify_bundle_contents(self, bundle_info):
        """Verify the bundle at bundle_path.

        :param valence.functional_models.BundleInfo bundle_info: Bundle info object. This object will
            be modified with missing_files, empty files, and a list of folders to remove.
        :raises valence.functional_models.NodeBundleError: If the bundle contents are not able
            to be verified or extracted.
        """
        files_to_verify = [x for x, y in FILES_TO_CHECK.items()
                           if self.node_dm.model.value in y[SUPPORTED_PLATFORMS]]
        _log.debug(self._get_sf_msg('Files to verify if they are in the bundle: {}'.format(
            files_to_verify)))
        bundle_folder = bundle_info.bundle_path.split('.tar.gz')[0]
        with self.get_ssh_connection() as ssh_conn:
            cmd = 'cd {}; sudo tar xzvf {}'.format(os.path.dirname(bundle_info.bundle_path),
                                                   bundle_info.bundle_path)
            response = ssh_conn.run_command(cmd, use_sudo=True)
            bundle_info.remote_rm_paths.add(bundle_folder)
            if response.return_code != 0:
                msg = 'Unable to extract {} Error: {}'.format(bundle_info.bundle_path, response.stderr)
                msg = self._get_sf_msg(msg)
                raise _errors.NodeBundleError(msg)

            for file_to_verify in files_to_verify:
                file_found, has_contents = self._verify_file(
                    ssh_conn=ssh_conn,
                    bundle_folder=bundle_folder,
                    to_verify=file_to_verify
                )
                if not file_found:
                    bundle_info.missing_files.append(file_to_verify)
                elif not has_contents:
                    bundle_info.empty_files.append(file_to_verify)

    def _cleanup_remote_node(self, bundle_info, purge_cores, purge_crash_dumps):
        """Clean up any remote files or folders on the node which are not necessary.

        :param valence.functional_models.BundleInfo bundle_info: Object containing info about
            support bundles
        :param bool purge_cores: If True, purge core files
        :param bool purge_crash_dumps: If True, purge crash dump files
        """
        with self.get_ssh_connection() as ssh_conn:
            if purge_cores:
                response = ssh_conn.run_command('ls {}'.format(CORE_FILE))
                if response.return_code == 0 and len(response.stdout) != 0:
                    bundle_info.remote_rm_paths.add(CORE_FILE)

            if purge_crash_dumps:
                response = ssh_conn.run_command('ls {}'.format(CRASH_DUMP_FILE))
                if response.return_code == 0 and len(response.stdout) != 0:
                    bundle_info.remote_rm_paths.add(CRASH_DUMP_FILE)

        if bundle_info.method is BundleMethod.api:
            self.delete_all_support_bundles()
            if not self._verify_bundle_removed():
                msg = '{} is not empty after using DeleteAllSupportBundles.'.format(
                    SF_BUNDLE_LOCATION)
                raise _errors.NodeBundleError(self._get_sf_msg(msg))
        else:
            bundle_info.remote_rm_paths.add(bundle_info.bundle_path)
        self._cleanup_remote(bundle_info.remote_rm_paths)

    def _create_bundles_api(self, bundle_name, extra_args):
        """Create support bundles on a node via the API.

        :param str bundle_name: The name of the bundle to use.
        :param str extra_args: Any extra arguments to pass to the support bundles script
        :return: The full path to the created bundle.
        :rtype: str
        """
        _log.debug(self._get_sf_msg('Use CreateSupportBundle API to generate support bundle'))

        create_bundle_dm = self.create_support_bundle(bundle_name=bundle_name,
                                                      extra_args=extra_args)

        bundle_path = os.path.join(SF_BUNDLE_LOCATION, create_bundle_dm.details.files[0])

        # Verify that the support bundle exists on the node's http server
        urls = create_bundle_dm.details.url
        for url in urls:
            if self.api.root_uri in url:
                # In a dual stack environment, there will be duplicate entries for the support bundles.
                # One for IPv4 and one for IPv6. In that case, only http GET the support bundles from
                # the url that the NodeApi object is using.

                # Strip off https://<mip>:<port> from the url since the http_get method adds that back in
                path = url.replace(self.api.root_uri, '')

                try:
                    self.api.http_get(path)
                except requests.HTTPError:
                    raise _errors.NodeBundleError(
                        'Unable to get bundle via HTTP at: {}'.format(url))

        return bundle_path

    def _create_bundles_ssh(self, bundle_name, extra_args):
        """Get the support bundles on a node via SSH.

        :param str bundle_name: The name of the bundle to use.
        :param str extra_args: Any extra arguments to pass to the support bundles script
        :rtype: str
        :return: The full path to the created bundle.
        """
        _log.debug(self._get_sf_msg('Use /sf/scripts/sf_make_support_bundle script'
                                    ' to generate support bundle'))
        with self.get_ssh_connection() as ssh_conn:
            bundle_path = os.path.join('/tmp', bundle_name)

            cmd = '/sf/scripts/sf_make_support_bundle {} {}'.format(extra_args, bundle_path)
            # Product bug - 21180
            # Currently stderr redirected to stdout
            # https://solidfire.fogbugz.com/f/cases/21180
            # Timeout on CreateSupportBundle API is 1500 seconds.
            # We need 3000 seconds for the fall back if API fails.
            # https://solidfire.fogbugz.com/f/cases/21316
            result = ssh_conn.run_command(cmd, use_sudo=True, timeout=3000)
            error_msg = 'Unknown Error'
            if result.return_code != 0:
                error_msg = self._get_sf_msg('Script: {} FAILED. STDOUT: {} ERROR: {}'.format(
                    cmd, result.stdout, result.stderr))
                _log.error(error_msg)
                if result.return_code == -99:
                    raise _errors.NodeBundleError(self._get_sf_msg(
                        'Command: {} timed out while trying to Create Bundles'.format(cmd)))

            cmd = 'ls -1dt {}*'.format(bundle_path)
            result2 = ssh_conn.run_command(cmd)
            if result2.return_code != 0:
                if len(result2.stdout) > 0:
                    error_msg = 'Partial support bundle collection. ' + error_msg
                raise _errors.NodeBundleError(error_msg)
            bundle_path = result2.stdout.splitlines()[0]

        return bundle_path

    def _attempt_making_bundle(self, method, bundle_info, extra_args, verify_logs):
        """Attempt to make the support bundle using the given method.

        :param valence.functional_models._node.BundleMethod method: Method to use (SSH vs API)
        :param valence.functional_models.BundleInfo bundle_info: Bundle info object
        :param str extra_args: Extra args to add to support bundles script
        :param bool verify_logs: If True, verify the bundle contents
        """
        if method is BundleMethod.api:
            bundle_info.bundle_path = self._create_bundles_api(bundle_info.bundle_name, extra_args)
        elif method is BundleMethod.ssh:
            bundle_info.bundle_path = self._create_bundles_ssh(bundle_info.bundle_name, extra_args)
        bundle_info.method = method

        # If compress flag was not passed to the product, issue a separate call to compress it
        if '--compress' not in extra_args and not bundle_info.bundle_path.endswith(
                self.COMPRESSION_TYPE):
            with self.get_ssh_connection() as ssh_conn:
                cmd = 'pigz {}'.format(bundle_info.bundle_path)
                result = ssh_conn.run_command(cmd)
                if result.return_code != 0 or len(result.stderr):
                    # We have our bundles, don't raise exception as that will fail the entire step.
                    # Failure to zip will lead to larger bundles to be copied over, but that will be
                    # handled by adaptive scp timeouts based on bundle sizes
                    _log.debug(self._get_sf_msg('Unable to zip {}, ERROR: {}'.format(
                        bundle_info.bundle_name, result.stderr)))
                else:
                    bundle_info.bundle_path += '.gz'

        if verify_logs:
            self._verify_bundle_contents(bundle_info)

    @staticmethod
    def _truncate_bundle_name_if_necessary(bundle_name):
        """Truncate the bundle_name if it is longer than the limit imposed by the API.

        :param str bundle_name: The name of the bundle to use.
        :return: bundle_name that is not longer than the max bundle_name length
        :rtype: str
        """
        if len(bundle_name) > SUPPORT_BUNDLE_NAME_LENGTH:
            original_bundle_name_length = len(bundle_name)
            bundle_name = bundle_name[:SUPPORT_BUNDLE_NAME_LENGTH]
            _log.debug('Bundle name is too long (length:{} max:{}), truncating to {}'
                       .format(original_bundle_name_length, SUPPORT_BUNDLE_NAME_LENGTH, bundle_name))

        return bundle_name
