"""Node FM methods related to hardware."""
import six

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models import _errors
from valence.functional_models._node import base


class HardwareMixin(base.NodeBase):
    """Node FM Hardware methods."""

    def test_hardware_config(self, clean=False, raise_on_failure=True):
        """Test the hardware config via TestHardwareConfig.

        :param bool clean: If the test should start with clean cache.
        :param bool raise_on_failure: raise an exception if any of the tests fail
        :raises valence.functional_models.errors.FunctionalNodeError: when any of the tests fail and
            flag is set
        :rtype: valence.models.NodeHardwareTest
        """
        params = {api_constants.CLEAN: clean}
        response = self._call_method(enums.NodeMethod.test_hardware_config, params=params,
                                     result_dm_class=result_models.TestHardwareConfig)
        hardware_config_data = response.result_dm
        test_details = hardware_config_data.details
        failed_tests = [str(test) for test in six.itervalues(test_details) if not test.passed]
        if raise_on_failure and failed_tests:
            raise _errors.FunctionalNodeError('TestHardwareConfig API returned a failed test result:'
                                              '{}'.format('\n'.join(failed_tests)))
        return hardware_config_data

    def get_hardware_info(self):
        """Get the hardware info from the node.

        :rtype: valence.models.NodeHardwareInfo
        """
        response = self._call_method(enums.NodeMethod.get_hardware_info,
                                     result_dm_class=result_models.GetHardwareInfo)
        return response.result_dm.hardware_info

    def check_proposed_cluster(self, node_dms, raise_on_failure=True):
        """Run checks to verify that a set of nodes could be used to create a cluster.

        :param list[valence.models.ResourceNode] node_dms: ResourceNode data models for nodes to check
        :param bool raise_on_failure: raise an exception if any checks fail
        :raises valence.functional_models.FunctionalNodeError: when any of the checks fail and
            flag is set
        :rtype: sf_api.result_models.CheckProposedCluster
        """
        params = {
            api_constants.NODES: [node.mip for node in node_dms]
        }
        check_proposed_cluster_dm = self._call_method(
            enums.NodeMethod.check_proposed_cluster,
            params=params,
            result_dm_class=result_models.CheckProposedCluster
        ).result_dm
        proposed_cluster_errors = check_proposed_cluster_dm.proposed_cluster_errors
        failed_checks = [str(error.description) for error in proposed_cluster_errors]
        if raise_on_failure and failed_checks:
            raise _errors.FunctionalNodeError('CheckProposedCluster API returned a failed check result: '
                                              '{}'.format('\n'.join(failed_checks)))

        return check_proposed_cluster_dm

    def get_node_fips_drives_report(self):
        """Get node's Fips report.

        :rtype: sf_api.result_models.GetNodeFipsDrivesReport
        """
        return self._call_method(enums.NodeMethod.get_node_fips_drives_report,
                                 result_dm_class=result_models.GetNodeFipsDrivesReport).result_dm
