"""Node FM methods related to networks."""
import autologger
import requests

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models import (
    _errors,
    monitor_decorators,
)
from valence.functional_models._node import base

_log = autologger.get_sflogging()


class NetworkMixin(base.NodeBase):
    """Node FM Network methods."""

    def get_network_config(self):
        """Get the network configuration of a given node.

        :rtype: valence.models.NetworkConfig
        """
        response = self._call_method(enums.NodeMethod.get_network_config,
                                     result_dm_class=result_models.GetNetworkConfig)
        return response.result_dm.network

    def set_network_config(self, network_config_dm, use_the_force_luke=False):
        """Sets the network configuration on a given node and validates that it was set correctly.

        .. note:: If the mip is changed, self.node_dm.mip will be modified.

        :param valence.models.NetworkConfig network_config_dm: Network config data model.
        :param bool use_the_force_luke: Force the command no matter what.
        :rtype: valence.models.NetworkConfig
        """
        params = {
            api_constants.NETWORK: network_config_dm.serialize(),
            api_constants.USE_THE_FORCE_LUKE: use_the_force_luke

        }
        bond_1g = network_config_dm.bond_1g
        new_mip = None
        net0 = network_config_dm.net0
        if bond_1g and bond_1g.address is not None or net0 and net0.address is not None:
            _log.debug('Changing address of Bond1G, so the call will be made against the Bond10G')
            self.ip = self.node_dm.sip
            if bond_1g:
                new_mip = bond_1g.address
            else:
                new_mip = net0.address
        try:
            response = self._call_method(enums.NodeMethod.set_network_config, params=params,
                                         result_dm_class=result_models.SetNetworkConfig)
            if new_mip:
                self.node_dm.mip = new_mip
        finally:
            if new_mip:
                self.ip = self.mip
        network_config_dm = response.result_dm.network
        return network_config_dm

    def set_config(self, network_config_dm=None, config_dm=None, use_the_force_luke=False):
        """Sets the configuration on a given node and validates that it was set correctly.

        .. note:: If the mip is changed, self.node_dm.mip will be modified.

        :param valence.models.NetworkConfig network_config_dm: Network config data model.
        :param valence.models.ClusterConfig config_dm: Cluster config data model.
        :param bool use_the_force_luke: Force the command no matter what.
        :rtype: valence.models.ClusterNodeConfig
        """
        params = {api_constants.USE_THE_FORCE_LUKE: use_the_force_luke}
        if network_config_dm:
            params[api_constants.NETWORK] = network_config_dm.serialize()
        if config_dm:
            params[api_constants.CLUSTER] = config_dm.serialize()

        bond_1g = network_config_dm.bond_1g if hasattr(network_config_dm, 'bond_1g') else None
        new_mip = None
        if bond_1g and bond_1g.address is not None:
            _log.debug('Changing address of Bond1G, so the call will be made against the Bond10G')
            self.ip = self.node_dm.sip
            new_mip = bond_1g.address
        try:
            response = self._call_method(enums.NodeMethod.set_config, params=params,
                                         result_dm_class=result_models.SetConfig)
            if new_mip:
                self.node_dm.mip = new_mip
        finally:
            if new_mip:
                self.ip = self.mip
        config_dm = response.result_dm
        return config_dm.config

    def get_network_interface_status(self, interfaces=None):
        """Get the status of each network interface on the node.

        .. note:: There is also a cluster_fm.get_network_interface_status() method which calls the
                  cluster API GetNetworkInterfaceStatus method.

        :param interfaces: List of Interfaces to get the status for. If None, return all interfaces.
        :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
        :return: dictionary with keys of NetworkInterfaceNames and values are enums.NetworkStatus
        :rtype: dict[str,valence.enums.NetworkStatus]
        """
        response = self._get_network_interface_status(interfaces=interfaces)
        return response.result_dm.interfaces

    def _get_network_interface_status(self, api=None, interfaces=None):
        """Internal get_network_interface_status.

        This method is needed so that the call can be made with a different instance of the node api.

        :param sf_api.NodeApi api: Optional node api instance
        :param interfaces: List of Interfaces to get the status for.
            If None, return status for all interfaces.
        :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
        :rtype: sf_api.SfResponse
        """
        params = {}
        if interfaces:
            params[api_constants.INTERFACES] = [i.value for i in interfaces]
        api = api or self.api
        return api.call_method(enums.NodeMethod.get_network_interface_status, params=params,
                               result_dm_class=result_models.NodeGetNetworkInterfaceStatus)

    @monitor_decorators.notify_monitor
    def set_network_interface_status_and_verify(self, interfaces,  # pylint: disable=too-complex
                                                verify=True, minutes=5,
                                                status=enums.NetworkStatus.up_and_running,
                                                max_try=2, use_the_force_luke=True):
        """Set network interface status and verify it.

        :param interfaces: List of Interfaces to set the status for
        :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
        :param bool verify: Whether to verify the success of set_network_interface
        :param int minutes: Timeout before this API call sets the networking interfaces
        :param valence.enums.NetworkStatus status: The new status to set the NIC(s) to
        :param int max_try: Number of maximum tries
        :param bool use_the_force_luke: Set to True to force the call to happen no matter what.
        """
        api = self.api
        # If the desired status is 'Up', the actual status 'Up' and 'UpAndRunning' are both valid
        if status in [enums.NetworkStatus.up, enums.NetworkStatus.up_and_running]:
            desired_status = [enums.NetworkStatus.up, enums.NetworkStatus.up_and_running]
        else:
            desired_status = [status]

        # by default, we query the NodeApi over the mip (which is typically tied to the Bond1G)
        # if the user is attempting to do anything with Bond1G we should use the 10G (aka sip [most of
        # the time]). This is fragile as the 10G could also be in the list of interfaces making this call
        # hang...
        if enums.NetworkInterface.bond_1g in interfaces:
            api = sf_api.NodeApi(self.node_dm.sip, self.api.username, self.api.username,
                                 version=self.api.version, authenticate=True)

        if verify:
            attempt = 1
            success = False
            while attempt <= max_try and success is False:
                try:
                    self._set_network_interface_status(api=api, interfaces=interfaces, minutes=minutes,
                                                       status=status,
                                                       use_the_force_luke=use_the_force_luke)
                except requests.ConnectionError:
                    _log.warning('Received Connection Error when setting {} to status "{}"'.format(
                        interfaces, status.value))
                finally:
                    response = self._get_network_interface_status(api=api, interfaces=interfaces)
                    interface_status = response.result_dm.interfaces

                    failed_interfaces = {}
                    for interface in interfaces:
                        actual_status = enums.NetworkStatus(interface_status[interface.value])
                        if actual_status not in desired_status:
                            failed_interfaces[interface] = actual_status

                    if failed_interfaces:
                        if attempt >= max_try:
                            msg = ''
                            for interface, actual_status in failed_interfaces.items():
                                msg += '{} has actual status "{}"\n'.format(
                                    interface.value, actual_status.value)
                            raise _errors.FunctionalNodeError(msg)
                        else:
                            attempt += 1
                    else:
                        success = True
        else:
            try:
                self._set_network_interface_status(api=api, interfaces=interfaces, minutes=minutes,
                                                   status=status,
                                                   use_the_force_luke=use_the_force_luke)
            except requests.ConnectionError:
                _log.warning('Received Connection Error when setting {} to status "{}"'.format(
                    interfaces, status))

    def _set_network_interface_status(self, api=None, interfaces=None, minutes=None, status=None,
                                      use_the_force_luke=None):
        """Internal set_network_interface_status.

        This method is needed so that the call can be made with a different instance of the node api.

        :param sf_api.NodeApi api: Optional node api instance
        :param interfaces: List of Interfaces to get the status for. If None, return all interfaces.
        :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
        :param int minutes: Timeout before this API call sets the networking interfaces
        :param valence.enums.NetworkStatus status: The new status to set the NIC(s) to
        :param bool use_the_force_luke: Set to True to force the call to happen no matter what.
        :rtype: sf_api.SfResponse
        """
        params = {
            api_constants.MINUTES: minutes,
            api_constants.STATUS: status,
            api_constants.USE_THE_FORCE_LUKE: use_the_force_luke
        }
        if interfaces:
            params[api_constants.INTERFACES] = [i.value for i in interfaces]

        api = api or self.api
        return api.call_method(enums.NodeMethod.set_network_interface_status, params=params,
                               result_dm_class=result_models.SetNetworkInterfaceStatus)

    def list_network_interfaces(self):
        """List the network interfaces on a node.

        :rtype: list[valence.models.NetworkInterface]
        """
        response = self._call_method(enums.NodeMethod.list_network_interfaces,
                                     result_dm_class=result_models.ListNetworkInterfaces)
        return response.result_dm.interfaces
