"""Node Functional Model."""

import json
import time

import autologger
import requests
import sf_artifacts
import sf_batch
import sf_mc
import sf_ssh
from sf_models import types
from sf_rtfi import (
    node_rtfi,
    reactor,
)

from sf_api import (
    api_constants,
    result_models,
)
from valence import (
    criteria,
    enums,
    models,
    utils,
)
from valence.functional_models import (
    _async_handler,
    _constants,
    _errors,
    _fm_utils,
)
from valence.functional_models._node import (
    bourbon_mixin,
    bundle_mixin,
    cert_mixin,
    download_mixin,
    hardware_mixin,
    network_mixin,
    nvdimm_mixin,
    otpw_mixin,
    package_mixin,
    restart_mixin,
    service_mixin,
    zookeeper_mixin,
)

_log = autologger.get_sflogging()


class Node(otpw_mixin.OtpwMixin, service_mixin.ServiceMixin, network_mixin.NetworkMixin,
           hardware_mixin.HardwareMixin, bundle_mixin.BundleMixin, zookeeper_mixin.ZookeeperMixin,
           restart_mixin.RestartMixin, package_mixin.PackageMixin, bourbon_mixin.BourbonMixin,
           cert_mixin.CertMixin, nvdimm_mixin.NvdimmMixin, download_mixin.DownloadMixin):
    """Node Functional Model."""

    def get_uptime(self, timeout=10):
        """Return the uptime of a node in seconds.

        :param int timeout: timeout in seconds
        :returns: seconds of uptime from cat /proc/uptime (no rounding)
        :rtype: int
        """
        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command('cat /proc/uptime', timeout=timeout)
            if result.stderr:
                raise _errors.FunctionalNodeError('SSH returned an error '
                                                  'while getting uptime: {}'.format(result.stderr))
            # format uptime to return seconds
            up_secs = int(result.stdout.split(' ')[0].split('.')[0])
            return up_secs

    def core_dump_check(self, exclude_regexes=None, timeout=1800):  # pylint: disable=too-complex
        """Check if there are any core files on the node, and raise an exception if any exist.

        :param list[str] exclude_regexes: Do not fail on any core dump that matches one of these
            regular expressions.
        :param int timeout: timeout in seconds
        :raises valence.functional_models.FunctionalNodeError: If any core dumps were found
        """
        start_time = time.time()
        exclude_regexes = exclude_regexes or []
        with self.get_ssh_connection() as ssh_inst:
            while (time.time() - start_time) < timeout:
                breakit = True
                result = ssh_inst.run_command('ls /sf/core*')
                stdout_lines = result.stdout.splitlines()
                for core in stdout_lines:
                    if 'uncompressed' in core or 'progress' in core:
                        _log.info('Cores are still dumping!')
                        breakit = False
                        time.sleep(10)
                if breakit:
                    break
        if result.return_code == 2:
            _log.debug('No cores found on {}'.format(self.mip))
            return
        elif result.return_code == 0:
            filtered_core_dumps = utils.regex_filter(stdout_lines, exclude_regexes, negate=True)
            diff = set(stdout_lines).difference(filtered_core_dumps)
            if diff:
                _log.debug('Filtering out core dumps: {}'.format(diff))
            if filtered_core_dumps:
                msg = 'Core dumps found on {}: {}'.format(self.mip, ', '.join(filtered_core_dumps))
                raise _errors.FunctionalNodeError(msg)
        else:
            raise _errors.FunctionalNodeError('Error looking for cores: {}, {}, {} '
                                              .format(stdout_lines, result.stderr,
                                                      result.return_code))

    def crash_dump_check(self, exclude_regexes=None):
        """Check if there are any crash dump files on the node, and raise an exception if any exist.

        It checks /var/crash for any directory with timestamp as a name,  such as
        /var/crash/201701210037/dump.201701210037. It ignores the kexec_cmd file.

        :param list[str] exclude_regexes: Do not fail on any crash dump that matches one of
             these regular expressions.
        :raises valence.functional_models.FunctionalNodeError: If any crash dumps were found
        """

        exclude_regexes = exclude_regexes or []

        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command('ls /var/crash | grep -v kexec_cmd')
        if result.return_code == 1:
            _log.debug('No crash dumps found on {}'.format(self.mip))
            return
        elif result.return_code == 0:
            stdout_lines = result.stdout.splitlines()
            _log.warn('Found crash dumps on {}: {}'.format(self.mip, ', '.join(stdout_lines)))
            filtered_crash_dumps = utils.regex_filter(stdout_lines, exclude_regexes, negate=True)
            diff = set(stdout_lines).difference(filtered_crash_dumps)
            if diff:
                _log.debug('Filtering out crash dumps {}'.format(diff))
            if filtered_crash_dumps:
                msg = 'Crash dumps found on {}: {}'.format(self.mip, ', '.join(filtered_crash_dumps))
                raise _errors.FunctionalNodeError(msg)
        else:
            raise _errors.FunctionalNodeError('Error looking for crash dumps: {}, {}, {}'
                                              .format(result.stdout, result.stderr, result.return_code))

    def get_core_files(self):
        """Get the core files on the node."""
        with self.get_ssh_connection() as ssh_inst:
            result = ssh_inst.run_command('ls /sf/core*/*')
            if result.return_code == 2:
                _log.debug('No core files found on {}'.format(self.mip))
                return
            elif result.return_code == 0:
                core_files = result.stdout.splitlines()
        manager = sf_artifacts.get_global_manager()
        manager.store_artifacts_remote(
            artifacts=core_files,
            remote_host_dm=self.node_dm
        )

    def api_is_up(self):
        """Attempt to open a socket to port 442 on the node.

        This should return correctly even if ssh is disabled on the node.
        True if the node is pingable, false if it is not pingable.

        :rtype: bool
        """
        try:
            sf_ssh.socket_connect(self.mip, port=api_constants.NODE_API_PORT)
            return True
        except sf_ssh.SshError:
            return False

    def reboot_node(self, seconds=None, force=None):
        """Reboot node using the RebootNode per node API call.

        This should return nothing, the api response is empty.

        :param int seconds: Seconds to keep the node down before restarting
        :param bool force: whether to add the force flag to the parameters
        """
        params = {
            api_constants.SECONDS: seconds,
            api_constants.FORCE: force
        }
        self._call_method(enums.NodeMethod.reboot_node, params=params,
                          result_dm_class=None)

    def get_drive_config(self):
        """Gets configuration information for the drives that are connected to the cluster.

        Enables you to display drive information for expected slice and
        block drive counts as well as the number of slices and block drives that are currently
        connected to the node.
        Note: This method is available only through the per-node API endpoint 5.0 or later.

        :rtype: valence.models.DriveConfig
        """
        response = self._call_method(enums.NodeMethod.get_drive_config,
                                     result_dm_class=result_models.GetDriveConfig)
        return response.result_dm.drive_config

    def get_hardware_config(self):
        """Gets list of hardware information and current settings.

        Enables you to display the hardware configuration information for a node.
        Note: This method is available only through the per-node API endpoint 5.0 or later.

        :rtype: valence.models.NodeHardwareConfig
        """
        response = self._call_method(enums.NodeMethod.get_hardware_config,
                                     result_dm_class=result_models.GetHardwareConfig)
        return response.result_dm.hardware_config

    def get_hardware_config_ssh(self):
        """Gets list of hardware information and current settings via ssh.

        Enables you to display the hardware configuration information for a node.
        This method will work on nodes that do not have the API.

        :rtype: valence.models.NodeHardwareConfig
        """
        get_hw_config = 'cat /sf/etc/hardware/config.json'

        with self.get_ssh_connection() as node:
            result = json.loads(node.run_command(cmd=get_hw_config).stdout)
        return models.NodeHardwareConfig(result)

    def get_startup_flags(self, flags=None, binaries=None):
        """Get the startup flags on the cluster.

        :param list[str] flags: List of the flags you want to get.
        :param list[valence.enums.NodeBinary] binaries: List of enums denoting the binaries you want to
            get the flags for.
        :rtype: dict
        """
        params = {
            api_constants.BINARIES: binaries,
            api_constants.FLAGS: flags
        }
        response = self._call_method(enums.NodeMethod.get_startup_flags, params=params,
                                     result_dm_class=result_models.GetStartupFlags)
        return response.result_dm.startup_flags

    def set_startup_flags(self, flags, binaries=None, reset=False):
        """Set the startup flags on the cluster.

        :param dict[str,Any] flags: Dictionary of the flags and their new values to set.
        :param list[valence.enums.NodeBinary] binaries: The binaries you want to set the flags for.
            The default is sfapp.
        :param bool reset: Reset the startup flags.
        :rtype: dict
        """
        if binaries is None:
            binaries = [enums.NodeBinary.sfapp]
        params = {
            api_constants.BINARIES: binaries,
            api_constants.FLAGS: flags,
            api_constants.RESET: reset
        }
        response = self._call_method(enums.NodeMethod.set_startup_flags, params=params,
                                     result_dm_class=result_models.SetStartupFlags)
        return response.result_dm.startup_flags

    def poll_ping_sshd_up(self, max_time=360, poll_period=15):
        """Ping the sshd being up for up to max time.

        :param int max_time: Stop time to stop the Poll pinging
        :param int poll_period: Polling Period
        """
        sf_ssh.poll_ssh_query(self.node_dm.mip, poll_max=max_time, poll_period=poll_period,
                              pingable=True)

    def poll_ping_sshd_down(self, max_time=360, poll_period=15):
        """Ping the sshd being down for up to max time.

        :param int max_time: Stop time to stop the Poll pinging
        :param int poll_period: Polling Period
        """
        sf_ssh.poll_ssh_query(self.node_dm.mip, poll_max=max_time, poll_period=poll_period,
                              pingable=False)

    def verify_pernode_api(self):
        """Verify if the pernode_api is up or down.

        :rtype: bool
        """
        _log.debug('Attempting to verify if the per-node API is up/down on Node: {}'
                   .format(self.node_dm.mip))
        try:
            self.get_api()
            _log.debug('Per node API up on Node: {}'.format(self.node_dm.mip))
            return True
            # 9.0 > above will give you a HTTPError (502), before that, the connection didn't even occur
        except (requests.ConnectionError, requests.HTTPError):
            _log.debug('Per node API down on Node: {}'.format(self.node_dm.mip))
            return False

    def poll_verify_pernode_api(self, max_time=360, poll_period=15):
        """Poll verify_pernode_api, wait up to 360 seconds for node to accept API calls.

        :param int max_time: Maximum time to poll in seconds
        :param int poll_period: Polling period in seconds
        :raises errors.PollMaxError: Raised if node api is down after poll_max is reached
        """
        try:
            block_poll = _fm_utils.BlockPoll(max_time,
                                             self.verify_pernode_api,
                                             True,
                                             _constants.OPER_EQ,
                                             poll_period)
            block_poll.start_polling()
        except _errors.PollMaxError:
            raise _errors.PollMaxError('Node {} API down, waited for {} sec'.format(self.node_dm.mip,
                                                                                    max_time))

    def fail_drives(self, drive_dms):
        """Fail drives on the node.

        The recommended way to fail drives is to use the cluster helper methods not the node
        methods directly. This is because only the cluster is aware of the node id's, so only it can
        ensure the correct drives are removed from the correct nodes.

        :param list[valence.models.Drive] drive_dms: Drive data models to remove
        """
        slice_drives = [d for d in drive_dms if d.type in
                        (enums.DriveType.volume, enums.DriveType.slice)]
        block_drives = [d for d in drive_dms if d.type is enums.DriveType.block]
        if block_drives:
            self.fail_block_drives(block_drives)
        if slice_drives:
            self.fail_slice_drives(slice_drives)

    def fail_block_drives(self, drive_dms):
        """Fail the given block drives on the node.

        The recommended way to fail drives is to use the cluster helper methods not the node
        methods directly. This is because only the cluster is aware of the node id's, so only it can
        ensure the correct drives are removed from the correct nodes.

        The fstab file being edited has a structure like::

            #slice /dev/slot0                          none      raw
            #block /dev/slot1                          none      raw
            #block /dev/slot2                          none      raw

        The regex below matches to the #block, then the drive's dev path which is /dev/slot#, then the
        none and raw. It will append dev=0 after the raw part which triggers the product to treat it as a
        failed drive.

        :param list[valence.models.Drive] drive_dms: Drive data models to remove
        """
        cmd_list = ''
        drive_configs = self.get_drive_config().drives
        # Fail using Carbon mechanism, Valence does not support pre-Carbon clusters
        for drive_dm in drive_dms:
            # pylint: disable=anomalous-backslash-in-string
            config = [d for d in drive_configs if d.slot == drive_dm.slot][0]
            cmd_list += r" -e 's:\(#block\s\+{}\s\+none\s\+raw\):".format(config.dev_path)
            cmd_list += "\\1    dev=0:'"
        cmd_list = 'sed -i {} /etc/fstab'.format(cmd_list)

        self._fail_drive_command(cmd=cmd_list)

    def fail_slice_drives(self, drive_dms):
        """Fail the given slice drives on the node.

        The recommended way to fail drives is to use the cluster helper methods not the node
        methods directly. This is because only the cluster is aware of the node id's, so only it can
        ensure the correct drives are removed from the correct nodes.

        :param list[valence.models.Drive] drive_dms: Drive data models to remove
        """
        cmd_list = ''
        drive_configs = self.get_drive_config().drives
        for drive_dm in drive_dms:
            # pylint: disable=anomalous-backslash-in-string
            config = [d for d in drive_configs if d.slot == drive_dm.slot][0]
            cmd_list += r" -e 's:\(#slice\s\+{}\s\+none\s\+raw\):".format(config.dev_path)
            cmd_list += "\\1    dev=0:'"
        cmd_list = 'sed -i {} /etc/fstab'.format(cmd_list)

        self._fail_drive_command(cmd=cmd_list)

    def _fail_drive_command(self, cmd):
        """Helper method to run the fail drive command on the node.

        :param str cmd: The command to run on the node.
        :raises errors.FunctionalNodeError: If the ssh command to edit /etc/fstab fails
        """
        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command(cmd)
            if result.return_code != 0 or result.stderr:
                raise _errors.FunctionalNodeError(
                    'Node fail drives command did not work on {}, stderr: {}, stdout: {}'.format(
                     self.node_dm.mip, result.stderr, result.stdout))
            # TODO: Fluorine's Drive refactor prevents failed drives from showing up in ListDrives within
            # a reasonable time period. As a fix for FR and for the LSFSv2 we are introducing a blanket
            # sleep with the intention of moving to ListServices. See case AA-852
            time.sleep(60)

    def restore_failed_drives(self, drive_dms=None):
        """Restore any failed drives on the node.

        If there are no failed drives, this does nothing.

        :param list[valence.models.Drive] drive_dms: drives to restore,
            if None all drives will be restored

        :raises errors.FunctionalNodeError: if ssh reports an error restoring failed drives.
        """
        cmd = 'sudo cat /etc/fstab > /tmp/fstab.to_restore;'
        if drive_dms is not None:
            config = self.get_drive_config()
            drive_slots = set([d.slot for d in drive_dms])
            fstab_entries = [c.dev_path for c in config.drives if c.slot in drive_slots]
            for entry in fstab_entries:
                cmd += (r"sudo sed -e 's:\({}\s\+none\s\+raw\)\s\+\(dev=0\s*\)\+:\1:' "
                        '/etc/fstab -i;').format(entry)
        else:
            cmd += r"sudo sed 's:\(none\s\+raw\)\s\+\(dev=0\s*\)\+:\1:' /etc/fstab -i;"
        cmd += ('sudo diff --unchanged-line-format= --old-line-format= '
                "--new-line-format='%L' /etc/fstab /tmp/fstab.to_restore|"
                r"grep -Ew '\w*slot.\w*|sdimm0p.\w*'")
        with self.get_ssh_connection() as ssh_inst:
            result = ssh_inst.run_command(cmd)
            if result.return_code != 0 or result.stderr:
                raise _errors.FunctionalNodeError(
                    'Error attempting to restore drives: '
                    'stdout: {}, stderr: {}, rc: {}'.format(result.stdout, result.stderr,
                                                            result.return_code))

    def _is_witness_compatible(self):
        """Returns whether this function is compatible with witness nodes.

        :rtype: bool
        """
        return isinstance(self.node_dm, models.ResourceNode) and self.vsphere

    def get_oob_power_status(self):
        """Get node power state via OOB/IPMI/iDRAC/Hypervisor.

        :return: power state of node
        :rtype: valence.enums.OobPowerState
        """
        if self._is_witness_compatible():
            with self.vsphere as v_conn:
                vm_info = v_conn.find_vms_by_name([self.node_dm.node_name])
                if vm_info[0]['state'] == 'poweredOn':
                    return enums.OobPowerState.on
                else:
                    return enums.OobPowerState.off

        machine_control_node = models.LegacyNodeAdapter(self.node_dm)

        if sf_mc.machine_power_status(machine_control_node) == 'OFF':
            return enums.OobPowerState.off
        else:
            return enums.OobPowerState.on

    def reboot_node_via_oob(self, wait_for_ping=False, max_time=600, poll_period=15):
        """Reboot node via OOB/IPMI/iDRAC/Hypervisor.

        :param bool wait_for_ping: Wait for the node to be pingable.
        :param int max_time: Maximum time to poll in seconds
        :param int poll_period: Polling period in seconds
        """
        if self._is_witness_compatible():
            with self.vsphere as v_conn:
                _log.debug('Rebooting witness node {}'.format(self.node_dm.node_name))
                v_conn.reset_vms([self.node_dm.node_name])
        else:
            machine_control_node = models.LegacyNodeAdapter(self.node_dm)
            sf_mc.machine_power_reboot_hard(machine_control_node)

        if wait_for_ping:
            _log.info('Waiting for nodes {} to come online'.format(self.node_dm.mip))
            self.poll_ping_sshd_up(max_time=max_time, poll_period=poll_period)

    def power_off_via_oob(self, wait_for_power_down=True):
        """Power off a node via OOB/IPMI/iDRAC/Hypervisor.

        :param bool wait_for_power_down: Wait for power to register 'off'.
        """
        if self._is_witness_compatible():
            with self.vsphere as v_conn:
                v_conn.power_off_vms([self.node_dm.node_name])
        else:
            machine_control_node = models.LegacyNodeAdapter(self.node_dm)
            if wait_for_power_down:
                sf_mc.MachineControl.machine_power_off_hard_wait(machine_control_node)
            else:
                sf_mc.MachineControl.machine_power_off_hard(machine_control_node)

    def power_on_via_oob(self, wait_for_power_up=True, wait_for_ping=False, wait_for_sfapp=True,
                         max_time=900, poll_period=15):
        """Power on a sec of nodes via OOB/IPMI/iDRAC/Hypervisor.

        wait_for_ping will poll the nodes every 15 seconds for 6 minutes until they're pingable, although
        this does not guarantee sfapp is running which could lead to unexpected behavior (i.e. failure to
        http auth).

        wait_for_sfapp will poll the nodes every 15 seconds for 6 minutes by default until nodes accept
        GetAPI call, indicating sfapp is running.

        :param bool wait_for_power_up: Wait for the power to register as 'on'.
        :param bool wait_for_ping: Wait for the node to be pingable.
        :param bool wait_for_sfapp: Wait for sfapp to run.
        :param int max_time: Maximum time to poll in seconds
        :param int poll_period: Polling period in seconds
        """
        if self._is_witness_compatible():
            with self.vsphere as v_conn:
                v_conn.power_on_vms([self.node_dm.node_name])
        # Add a Static Sleep of 60 Seconds until we figure OUT Marvell, Radian Cards issues
        # while Power On..Ray/Eric/Srid
        # TODO: Remove the Static Sleep Here (JIRA: SET-544)
        else:
            time.sleep(60)
            node_ips = [self.node_dm.mip]

            machine_control_node = models.LegacyNodeAdapter(self.node_dm)
            if wait_for_power_up:
                sf_mc.MachineControl.machine_power_on_wait(machine_control_node)
            else:
                sf_mc.MachineControl.machine_power_on(machine_control_node)

            if wait_for_ping:
                _log.info('Waiting for nodes {} to come online'.format(node_ips))
                self.poll_ping_sshd_up(max_time=max_time, poll_period=poll_period)

            if wait_for_sfapp:
                _log.info('Waiting for nodes {} sfapp to run'.format(node_ips))
                self.poll_verify_pernode_api(max_time=max_time, poll_period=poll_period)

    def set_cluster_config(self, cluster, force=False, use_the_force_luke=False):
        """Sets the Cluster Configuration on a given node.

        If you are wanting to change the mipi/cipi/sipi addresses, you change
        those through SetNetworkConfig

        :param str cluster: Cluster name to set on the node
        :param bool force: used when the node is a part of a cluster
        :param bool use_the_force_luke: Set to True to force the call to happen no matter what.
        :rtype: valence.models.ClusterConfig
        """
        params = {
            api_constants.CLUSTER: {api_constants.CLUSTER: cluster},
            api_constants.FORCE: force,
            api_constants.USE_THE_FORCE_LUKE: use_the_force_luke
        }
        return self.api.call_method(enums.NodeMethod.set_cluster_config, params=params,
                                    result_dm_class=result_models.SetClusterConfig).result_dm.cluster

    def set_constants(self, constants_dm):
        """Set node specific API constants.  Setting these values will override cluster constant setting.

        Strips away all the constants with a value of none before applying it

        :param valence.models.ClusterConstants constants_dm: The cluster constants data model to set.
        :return: Return all constants for cluster
        :rtype: valence.models.ClusterConstants
        """
        params = constants_dm.serialize()
        response = self.api.call_method(enums.NodeMethod.set_node_constants, params=params,
                                        result_dm_class=result_models.SetConstants)
        return response.result_dm

    def get_constants(self, changed_only=False, constants_dm=None):
        """Return all constants for cluster via GetConstants.

        :param bool changed_only: If True, will only return values changed from defaults
        :param valence.models.ClusterConstants constants_dm: If not None, which constants to get
        :return: Return all constants for cluster
        :rtype: valence.models.ClusterConstants
        """
        if not constants_dm:
            constants_dm = models.ClusterConstantsVerbose.create()

        params = {'constants': list(constants_dm.serialize().keys())}
        if changed_only:
            params['changedOnly'] = True

        response = self.api.call_method(enums.NodeMethod.get_node_constants,
                                        result_dm_class=result_models.GetConstants,
                                        params=params)
        return response.result_dm

    def reset_constants(self, constants_dm):
        """Reset node specific API constants.  Must call reset to stop override of cluster constant.

        :param valence.models.ClusterConstants constants_dm: The cluster constants data model to set.
        :return: Return all constants for cluster
        :rtype: valence.models.ClusterConstants
        """
        params = {'constants': list(constants_dm.serialize().keys())}
        response = self.api.call_method(enums.NodeMethod.reset_node_constants, params=params,
                                        result_dm_class=result_models.ResetConstants)
        return response.result_dm

    def get_cluster_state(self):
        """Call GetClusterState on the specific node.

        The GetClusterState API method enables you to indicate if a node is part of a cluster or not.
        The three states are:
        Available: Node has not been configured with a cluster name.
        Pending: Node is pending for a specific named cluster and can be added.
        Active: Node is an active member of a cluster and may not be added to another cluster.
        Note: This method is available only through the per-node API endpoint 5.0 or later.

        :rtype: valence.models.ClusterState
        """
        response = self.api.call_method(enums.NodeMethod.get_cluster_state, params={},
                                        result_dm_class=result_models.GetClusterState)
        return response.result_dm

    def create_cluster(self, mvip, svip, nodes, username='admin', password='admin', rep_count=2,
                       attributes=None, unique_id=None, allow_two_node_ensemble=False,
                       skip_node_type_checks=True):
        """Call CreateCluster / CreateStandaloneCluster from a given node.

        This will call CreateCluster if the node count in nodes is greater than 1. Otherwise it will
        call CreateStandaloneCluster.

        :param str mvip: Floating (virtual) IP address for the cluster on the management network.
        :param str svip: Floating (virtual) IP address for the cluster on the storage (iSCSI) network.
        :param list[valence.models.ResourceNode] nodes: Set of nodes making up the cluster.
        :param str username: User name for the cluster admin.
        :param str password: Initial password for the cluster admin account.
        :param int rep_count: Number of replicas of each piece of data to store in the cluster. Valid
            value is "2".
        :param dict attributes: Dictionary of attributes
        :param str unique_id: You can run either the ListActiveVolumes or GetClusterInfo to find the
            uniqueID.
        :param bool allow_two_node_ensemble: Whether to allow cluster creation with 2 nodes.
        :param bool skip_node_type_checks: Whether to check node types for 2-node cluster.
        """
        node_ips = [node.sip for node in nodes]
        params = {api_constants.MVIP: mvip,
                  api_constants.SVIP: svip,
                  api_constants.USERNAME: username,
                  api_constants.PASSWORD: password,
                  api_constants.ACCEPT_EULA: True,
                  api_constants.NODES: node_ips,
                  api_constants.REP_COUNT: rep_count,
                  api_constants.ATTRIBUTES: attributes,
                  api_constants.UNIQUE_ID: unique_id,
                  api_constants.ALLOW_TWO_NODE_ENSEMBLE: allow_two_node_ensemble,
                  api_constants.SKIP_NODE_TYPE_CHECKS: skip_node_type_checks}
        # Override auth since there is no port when calling create_cluster
        auth = self.api.get_auth()
        auth.port = None
        if len(nodes) == 1:
            # remove the entries that are not used for a standalone cluster
            params.pop(api_constants.NODES)
            params.pop(api_constants.REP_COUNT)
            self.api.call_method(enums.NodeMethod.create_standalone_cluster, params=params,
                                 auth=auth)
        else:
            self.api.call_method(enums.NodeMethod.create_cluster, params=params,
                                 auth=auth)

    def get_bootstrap_config(self):
        """Gets cluster and node bootstrap information.

        Returns cluster and node information from the bootstrap configuration file.
        Use this API method on an individual node before it has been joined with a cluster.
        You can use the information this method returns in the cluster configuration interface
        when you create a cluster. If a cluster has already been created, this can be used
        to obtain the MVIP and SVIP addresses of the cluster.

        :rtype: valence.models.BootstrapConfig
        """
        params = {}
        auth = self.api.get_auth()
        auth.port = None
        response = self.api.call_method(enums.NodeMethod.get_bootstrap_config, params=params,
                                        result_dm_class=result_models.GetBootstrapConfig, auth=auth)
        return response.result_dm

    def start_rtfi(self, sf_rtfi_version, skip_post_rtfi_network_config=False):  # pragma: no cover
        """Start inplace rtfi and return an async handler.

           I had to use some threading here, in violation of the rules.
           Just hope Drew does not notice. #  pragma: no drew
           This will be removed when REST call is ready.

        :param valence.models.SfVersion sf_rtfi_version: version to rtfi to
        :param bool skip_post_rtfi_network_config: skip calling SetNetworkConfig post RTFI if True
        :raises valence.functional_models.FunctionalNodeError: If RTFI fails
        :rtype: valence.functional_models.RtfiAsyncHandler
        """
        node_info = self.node_dm.serialize()
        node_dm = models.RtfiNodeAdapter(node_info)
        inplace_params = {'sf_secure_erase': False,
                          'sf_inplace': True,
                          'sf_keep_cluster_config': False,
                          'sf_keep_network_config': False,
                          'sf_keep_data': False}

        manager = sf_artifacts.get_global_manager()
        rtfi_watcher = node_rtfi.NodeRtfiProvider(
                           node_dm, sf_rtfi_version,
                           inplace_params,
                           manager.artifact_credentials,
                           skip_post_rtfi_network_config=skip_post_rtfi_network_config
        )
        reactor._log.disable()  # pylint: disable=protected-access
        rtfi_watcher.start_reactor()  # all real errors are logged in main thread

        def run_rtfi():
            rtfi_watcher.rtfi()

        rtfi_thread = sf_batch.SFThread(run_rtfi)
        rtfi_thread.start_thread()
        async_handler = _async_handler.RtfiAsyncHandler(rtfi_watcher.reactor)
        return async_handler

    def start_traditional_rtfi(self, sf_rtfi_version):  # pragma: no cover
        """Start traditional rtfi on node.

        :param valence.models.SfVersion sf_rtfi_version: version to rtfi to
        :raises valence.functional_models.FunctionalNodeError: If RTFI fails
        :rtype: valence.functional_models.RtfiAsyncHandler
        """
        node_info = self.node_dm.serialize()
        node_dm = models.RtfiNodeAdapter(node_info)
        params = {'sf_secure_erase': False,
                  'sf_keep_cluster_config': False,
                  'sf_keep_network_config': False,
                  'sf_keep_data': False}

        rtfi_watcher = node_rtfi.NodeRtfiProvider(node_dm, sf_rtfi_version, params, {})
        rtfi_watcher.start_reactor()

        def run_trtfi():
            rtfi_watcher.traditional_rtfi(use_reactor=True)

        reactor._log.disable()  # pylint: disable=protected-access
        rtfi_thread = sf_batch.SFThread(run_trtfi)
        rtfi_thread.start_thread()
        async_handler = _async_handler.RtfiAsyncHandler(rtfi_watcher.reactor)
        return async_handler

    def get_cluster_config(self):
        """Gets cluster configuration information the node uses to communicate with the cluster.

        Enables you to return information about the cluster configuration
        this node uses to communicate with the cluster that it is a part of.

        :rtype: valence.models.ClusterConfig
        """
        return self.api.call_method(enums.NodeMethod.get_cluster_config,
                                    result_dm_class=result_models.GetClusterConfig).result_dm.cluster

    def get_config_extended(self):
        """Return the extended network and cluster configuration of a given node.

        :rtype: valence.models.ClusterNodeConfigExtended
        """
        response = self.api.call_method(enums.NodeMethod.get_config_extended,
                                        result_dm_class=result_models.GetConfigExtended)
        return response.result_dm.config

    def set_config_extended(self, extended_network_dms, cluster_config_dm):
        """Set the network and cluster configuration of a given node with the extended API.

        .. note:: The extended API is used to configure two VLANs on a given bonded interface.
                  If this is not your use case, use the legacy network and cluster methods.

        :param list[valence.models.NetworkInterfaceDetailsExtended] extended_network_dms: List of
            network details extended data models
        :param valence.models.ClusterConfig cluster_config_dm: Cluster config data model
        :rtype: valence.models.ClusterNodeConfigExtended
        """
        convert_extended = types.AttributeKeyedModelListType(
            attr_for_key=models.NetworkInterfaceDetailsExtended.name,
            model_class=models.NetworkInterfaceDetailsExtended
        )
        params = {
            api_constants.NETWORK: convert_extended.to_primitive(extended_network_dms),
            api_constants.CLUSTER: cluster_config_dm.serialize()
        }
        response = self.api.call_method(enums.NodeMethod.set_config_extended, params=params,
                                        result_dm_class=result_models.SetConfigExtended)
        return response.result_dm.config

    def get_version_info(self):
        """API Method used to return version information about binaries on a node.

        :rtype: sf_api.result_models.GetVersionInfo
        """
        response = self.api.call_method(enums.NodeMethod.get_version_info,
                                        result_dm_class=result_models.GetVersionInfo)

        return response.result_dm

    def read_link(self, path):
        """Verifies that a file exists at the given path on the host.

        :param str path: path on host to read
        :rtype: sf_api.result_models.ReadLink
        """

        return self._call_method(enums.NodeMethod.read_link, params={api_constants.PATH: path},
                                 result_dm_class=result_models.ReadLink).result_dm

    def rotate_logs(self, force=False):
        """Calls log rotate on the node.

        :param bool force: forces log rotatation
        """
        self._call_method(enums.NodeMethod.rotate_logs, params={api_constants.FORCE: force})

    def list_available_nodes(self, static_ids=None, node_roles=None):
        """List available nodes via ListAvailableNodes.

        :param typing.Iterable[str] static_ids: Mips on which to filter.
        :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
        :rtype: list[valence.models.AvailableNode]
        """
        # Put into a set to improve efficiency
        static_ids = set(static_ids) if static_ids else None
        node_roles = set(node_roles) if node_roles else None
        nodes_filter = criteria.And(
            criteria.ContainedIn(models.AvailableNode.mip, static_ids),
            criteria.ContainedIn(models.SfNodeBase.role, node_roles),
        )

        # Override auth since there is no port when calling list_available_nodes
        auth = self.api.get_auth()
        auth.port = None
        response = self.api.call_method(
            method=enums.NodeMethod.list_available_nodes,
            result_dm_class=result_models.ListAvailableNodes,
            auth=auth
        )
        return nodes_filter.filter(response.result_dm.nodes)

    def get_fips_drives_config(self):
        """Get the fipsDriveConfiguration value from /etc/solidfire.json.

        :rtype: bool
        """
        config_path = '/etc/solidfire.json'

        with self.get_ssh_connection() as ssh_conn:
            raw_config = ssh_conn.run_command('cat {}'.format(config_path), raise_on_failure=True).stdout
            config = json.loads(raw_config)

        return config['fipsDriveConfiguration']

    def set_fips_drives_config(self, enabled=True):
        """Enable or disable encryption capability for a node.

        :param bool enabled: True=enable, False=disable.
        """
        config_path = '/etc/solidfire.json'

        with self.get_ssh_connection() as ssh_conn:
            raw_config = ssh_conn.run_command('cat {}'.format(config_path), raise_on_failure=True).stdout
            config = json.loads(raw_config)
            config['fipsDriveConfiguration'] = enabled
            # Was getting permission denied without this, even with use_sudo and get_pty
            ssh_conn.run_command('chmod 777 {}'.format(config_path), raise_on_failure=True)
            ssh_conn.run_command("echo '{}' > {}".format(json.dumps(config, indent=4), config_path),
                                 use_sudo=True, raise_on_failure=True)

        self.restart_binary(service=enums.NodeBinary.solidfire)

    def set_fake_drives_encryption_capability(self, encryption_capability, slots=None):
        """Fake encryption capability for specific drives.

        :param valence.enums.DriveEncryptionCapability encryption_capability: Capability to fake.
        :param list[int] slots: Drive slots to modify. Default=all
        :returns: Path to config file.
        :rtype: str
        """
        if slots is None:
            slots = [d.slot for d in self.get_drive_config().drives]

        path = '/var/tmp/overrideDrives.json'
        config = {'slots': []}
        for slot in slots:
            slot_config = {'slotID': 'slot' + str(slot),
                           'driveEncryptionCapability': encryption_capability.value}
            config['slots'].append(slot_config)

        with self.get_ssh_connection() as ssh_conn:
            cmd = "echo '{}' > {}"
            ssh_conn.run_command(cmd.format(json.dumps(config, indent=4), path), raise_on_failure=True)

        self.restart_binary(service=enums.NodeBinary.solidfire)
        return path

    def save_clear_ipmitool_logs(self, artifact_manager=None):
        """Save old ipmitool logs to a file and clear the logs.

        The file will be saved in /root/ on the remote machine

        :param sf_artifacts.ArtifactManager artifact_manager: optional ArtifactManager object that
              defines where a copy of the saved logfile will be stored locally
        :return: list of strings detailing the results of the commands that were run
        :rtype: list
        """
        save_ipmi = (
            'if ((`ipmitool sel elist | wc -l` > 1)); then '
            '    ipmifile=ipmitool_sel_elist-{0}`date +%Y-%m-%d_%H-%M-%S`.log; '
            '    ipmitool sel save $ipmifile | $1>/dev/null; '
            '    echo "saved old ipmitool logs to $(pwd)/$ipmifile"; '
            '    cat $ipmifile; '
            'fi'
        ).format(self.node_dm.node_name)
        clear_ipmi = 'ipmitool sel clear'

        with self.get_ssh_connection() as node:
            results = []

            message = node.run_command(
                cmd=save_ipmi, raise_on_failure=True, use_sudo=False, timeout=10
            ).stdout
            if message:
                filename_message, file_contents = message.split('\n', 1)
                results.append(str(filename_message))
                _ = artifact_manager and artifact_manager.store_text(file_contents)
            else:
                results.append('logs were already empty so no file has been saved')

            node.run_command(clear_ipmi, raise_on_failure=True)
            results.append('successfully cleared ipmitool logs')
        return results
