"""Node Nvdimms Functional Models."""

from __future__ import division  # x/y always returns float

import json
import re
import time

import autologger

from valence import enums
from valence.functional_models import _errors
from valence.functional_models._node import base

_log = autologger.get_sflogging()

# register values for _nvdimm_regread/regwrite
NVDIMM_MGT_CMD0 = '0x40'
ARM_CMD = '0x45'
SET_ES_POLICY_CMD = '0x49'
NVDIMM_READY = '0x60'
NVDIMM_CMD_STATUS0 = '0x61'
ARM_STATUS0 = '0x6A'
PAGE_ZERO = '0'
PAGE_THREE = '3'
FW_SLOT_INFO = '0x42'
USE_NEW_FW = '1'


class NvdimmMixin(base.NodeBase):

    def flash_nvdimm_firmware(self, expected_version, downgrade=False, quick_downgrade=False,
                              nvdimm_diags=None):
        """Flash NVDIMM firmware, currently limited to SMART and Micron NVDIMMs.

        :param str expected_version: Firmware Version number

        ``These two params only have an effect for smart nvdimms```
        :param bool downgrade: if set to True, will downgrade the firmware
        :param bool quick_downgrade: if set to True, will perform a quick downgrade

        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function
        """
        _hardware_config = self.get_hardware_config_ssh()
        vendor = _hardware_config.nvram_vendor
        diags = nvdimm_diags or self.get_nvdimm_diagnostics(hw_cfg=_hardware_config)

        if vendor.is_smart():
            self._flash_smart_firmware(expected_version=expected_version, downgrade=downgrade,
                                       quick_downgrade=quick_downgrade, nvdimm_diags=diags)
        elif vendor.is_micron():
            self._flash_micron_firmware(new_firmware_version=expected_version, vendor=vendor,
                                        nvdimm_diags=diags)

    def _flash_smart_firmware(self, expected_version, downgrade=False, quick_downgrade=False,
                              nvdimm_diags=None):
        """Flash SMART nvdimm firmware.

        :param str expected_version: Firmware Version number
        :param bool downgrade: if set to True, will downgrade the firmware
        :param bool quick_downgrade: if set to True, will perform a quick downgrade
        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function
        """
        FLASH_NVDIMM_CMD = 'smart-nvdimm --nvid={} --update={}'
        UPGRADE_NVDIMM_ARM_CMD = 'smart-nvdimm --arm --all'
        DOWNGRADE_NVDIMM_ARM_CMD = 'smart-nvdimm --all --reloadAndArm'
        QUICK_NVDIMM_DOWNGRADE_CMD = 'smart-nvdimm --all --updateerase'

        diags = nvdimm_diags or self.get_nvdimm_diagnostics()

        self.verify_minimum_number_of_nvdimms_exist(nvdimm_diags=diags)
        fw_file_path = self.get_nvdimm_firmware_filepath(
            expected_version, vendor=enums.NvramVendor.smart, nvdimm_diags=diags)

        with self.get_ssh_connection() as ssh_conn:
            if quick_downgrade:
                ssh_conn.run_command(cmd=QUICK_NVDIMM_DOWNGRADE_CMD, timeout=300, raise_on_failure=True)
            else:
                # The nvid will vary between different node models
                for nvid in diags:
                    # Flashing firmware on node
                    cmd = FLASH_NVDIMM_CMD.format(nvid, fw_file_path)
                    ssh_conn.run_command(cmd=cmd, timeout=300, raise_on_failure=True)

            # Arm the nvdimms
            if downgrade or quick_downgrade:
                ssh_conn.run_command(cmd=DOWNGRADE_NVDIMM_ARM_CMD, raise_on_failure=True, timeout=60)
            else:
                ssh_conn.run_command(cmd=UPGRADE_NVDIMM_ARM_CMD, raise_on_failure=True, timeout=60)

    def _flash_micron_firmware(self, new_firmware_version, vendor=None, nvdimm_diags=None):
        """Flash Micron nvdimm firmware.

        :param str new_firmware_version: Firmware Version number
        :param sf_api.enums.NvramVendor vendor: Optional param to explicitly set the nvdimm vendor
        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function
        """
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()
        fw_file_path = self.get_nvdimm_firmware_filepath(
            new_firmware_version, vendor=vendor, nvdimm_diags=diags)

        self.verify_minimum_number_of_nvdimms_exist(nvdimm_diags=diags)
        for nvid in diags:
            _log.debug(
                'Flashing MICRON NVDIMM nvid="{0}" Firmware from: active_firmware_version="{1}" '
                'to: new_firmware_version="{2}" using fw_file="{3}"'.format(
                    nvid, diags[nvid]['activeFirmwareVersion'], new_firmware_version, fw_file_path)
            )
        with self.get_ssh_connection() as node:
            # this may take a while (up to 17 minutes) so timeout after 20 to be safe
            node.run_command(cmd='nvdimm_upgrade -ar {}'.format(fw_file_path), timeout=20*60)

        for nvid in diags:
            # make sure the new firmware slot is activated
            self.nvdimm_select_fw_slot(nvid, USE_NEW_FW)
            self._firmware_micron_arm_device(nvid)
            _log.debug('NVDIMM nvid="{}" FW upgrade to new_firmware_version="{}" is complete'.format(
                nvid, new_firmware_version))
        self.verify_nvdimm_armed()  # this needs to use fresh diagnostics so don't pass down diags

    def _firmware_micron_arm_device(self, nvid):
        arm_timeout = self._firmware_micron_get_arm_timeout(nvid)

        _log.debug('Arming the nvdimm: nvid="{0}", arm_timeout="{1}"'.format(nvid, arm_timeout))

        self._firmware_micron_check_ready(nvid, NVDIMM_CMD_STATUS0, '0x00')

        # Set the energy save policy to device managed
        self._nvdimm_regwrite(nvid, PAGE_ZERO, SET_ES_POLICY_CMD, '0x01')
        self._firmware_micron_check_ready(nvid, NVDIMM_CMD_STATUS0, '0x00')

        # Clear the Arm Status reg
        self._nvdimm_regwrite(nvid, PAGE_ZERO, NVDIMM_MGT_CMD0, '0x20')

        # Arm the device
        self._nvdimm_regwrite(nvid, PAGE_ZERO, ARM_CMD, '0x01')

        # Wait for Arm in Progress to clear
        self._firmware_micron_check_ready(nvid, NVDIMM_CMD_STATUS0, '0x00', arm_timeout)

        # Check Arm Status is success
        self._firmware_micron_check_ready(nvid, ARM_STATUS0, '0x05')

        _log.debug('Arm complete for nvdimm nvid="{}"'.format(nvid))

    def _firmware_micron_get_arm_timeout(self, nvid):
        """Reads the two registers that represent ARM_TIMEOUT with a bit for sec/ms.

        Outputs a minimum of 1 sec.

        :param str nvid: nvid of the dimm to check. Format of nvid will vary for different nvdimm vendors
        :rtype: int
        """
        byte1 = int(self._nvdimm_regread(nvid, PAGE_ZERO, '0x20'), 16)
        byte2 = int(self._nvdimm_regread(nvid, PAGE_ZERO, '0x21'), 16)

        # fransform the two 8-bit numbers into a 16-bit number; e.g., byte1=0x3A, byte2=0x19 -> 0x193A
        timeout = ((byte2 & 0x7f) << 8) | byte1 if (byte2 & 0x80) == 0 else 1

        return timeout

    def _firmware_micron_check_ready(self, nvid, reg, value, max_delay=80):
        """Check that the device is ready and not busy.

        :param str nvid: nvid of the dimm to check. Format of nvid will vary for different nvdimm vendors
        :param str reg: nvdimm register index
        :param str value: Register value to check for. Must use hex 0xnn form
        :param int max_delay: max wait time in seconds. Default is 80
        """
        delay = 2
        elapsed_seconds = 0

        _log.debug('nvid="{}", reg="{}", value="{}", max_delay="{}"'.format(
            nvid, reg, value, max_delay))

        nvdimm_ready_reg = None
        value_reg = None
        while elapsed_seconds < max_delay:
            nvdimm_ready_reg = self._nvdimm_regread(nvid, PAGE_ZERO, NVDIMM_READY)
            value_reg = self._nvdimm_regread(nvid, PAGE_ZERO, reg)

            if nvdimm_ready_reg != '0xA5' or value_reg != value:
                time.sleep(delay)
                elapsed_seconds += delay
            else:
                break

        if not nvdimm_ready_reg == '0xA5':
            raise _errors.FlashNvdimmFirmwareError(
                'NVDIMM nvid="{0}" Not Ready after {1} seconds.'.format(nvid, max_delay))

        if not value_reg == value:
            raise _errors.FlashNvdimmFirmwareError(
                'NVDIMM nvid="{0}" reg="{1}" not at expected value value="{2}" '
                'after {3} seconds.'.format(nvid, reg, value, max_delay))

    def _nvdimm_regread(self, nvid, page, reg):
        """Read the value of a register on Smart or Micron nvdimm. Inconsistent results on Smart nvdimms.

        :param str nvid: nvid of the nvdimm
        :param str page: page containing the register to read from
        :param str reg:  register to read from. Must use hex 0xnn form

        :rtype: str
        """
        with self.get_ssh_connection() as node:
            result = node.run_command(
                cmd="nvdimm_regread {0} {1} {2} | awk '{{print $4}}'".format(nvid, page, reg))
        return result.stdout.rstrip('\n')

    def _nvdimm_regwrite(self, nvid, page, reg, value):
        """Read the value of a register on Smart or Micron nvdimm.

        :param str nvid: nvid of the nvdimm
        :param str page: page containing the register to write to
        :param str reg:  register to write to. Must use hex 0xnn form
        :param str value: value to write to register. Must use hex 0xnn form
        """
        with self.get_ssh_connection() as node:
            node.run_command(cmd='nvdimm_regwrite {0} {1} {2} {3}'.format(nvid, page, reg, value))

    def nvdimm_select_fw_slot(self, nvid, slot):
        """Select the firmware slot for next boot/re-arm on nvdimm.

        :param str nvid: nvid of the nvdimm to write to
        :param str|int slot: slot to set for next boot/re-arm
        """
        assert str(slot) in ['0', '1']
        self._nvdimm_regwrite(nvid, PAGE_THREE, FW_SLOT_INFO, slot)

    def get_nvdimm_firmware_filepath(self, expected_version, vendor=None, nvdimm_diags=None):
        """Returns the filepath of the matching firmware file on the node.

        :param str expected_version: Firmware Version number
        :param sf_api.enums.NvramVendor vendor: Optional param to explicitly set the nvdimm vendor
        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function.
                                  Setting this param has no effect if vendor is not Micron.

        :raises: valence.functional_models.VerifyNvdimmFailedError if fw file cannot be located
        :return str fw_file_path: the path to the NVDIMM fw file
        :rtype: str
        """
        _hardware_config = None
        if not vendor:
            _hardware_config = self.get_hardware_config_ssh()
            vendor = _hardware_config.nvram_vendor

        if vendor.is_smart():
            return self._get_smart_firmware_filepath(expected_version)
        elif vendor.is_micron():
            diags = nvdimm_diags or self.get_nvdimm_diagnostics(hw_cfg=_hardware_config)
            return self._get_micron_firmware_filepath(expected_version, vendor, nvdimm_diags=diags)

    def _get_smart_firmware_filepath(self, expected_version):
        # allow for version number passed in as '26.2C', '2C', '2.11', etc
        match = re.match(r'^(?:\d+\.)?(\d[A-F])$', expected_version)
        if match is not None:
            expected_version = '.'.join(str(int(char, 16)) for char in match.groups()[0])

        with self.get_ssh_connection() as ssh_conn:
            ssh_conn.run_command(cmd='updatedb', timeout=10)
            result = ssh_conn.run_command(cmd='locate {}.bin'.format(
                expected_version), timeout=10)
        if result.return_code != 0:
            raise _errors.VerifyNvdimmFailedError(
                'SSH returned an error while checking if firmware version {} Exists: {} '.format(
                    expected_version, result.stderr))
        else:
            return result.stdout.rstrip('\n')

    def _get_micron_firmware_filepath(self, expected_version, vendor, nvdimm_diags=None):
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()
        # allow for version number passed in as '26.2C' or '26'
        new_major_version = int(expected_version.split('.')[0])

        if vendor == enums.NvramVendor.micron:
            fw_filename = self._get_micron_firmware_filename_first_gen(new_major_version, diags)
        elif vendor == enums.NvramVendor.micron_gen2:
            fw_filename = self._get_micron_firmware_filename_second_gen(new_major_version, diags)

        with self.get_ssh_connection() as node:
            result = node.run_command('updatedb && locate {}'.format(fw_filename))

        if result.return_code != 0:
            raise _errors.VerifyNvdimmFailedError('SSH returned an error while checking if firmware '
                                                  'version {} Exists: {}'
                                                  .format(expected_version, result.stderr))
        else:
            return result.stdout.rstrip('\n')

    def _get_micron_firmware_filename_first_gen(self, new_major_version, diags):
        for dimm_data in diags.values():
            current_major_version = int(dimm_data['activeFirmwareVersion'].split('.')[0])

            if current_major_version in [22, 23]:
                upgrade_type = 'PSOC' if new_major_version in [22, 23] else 'ALL'
            elif current_major_version in [24, 26]:
                upgrade_type = 'PSOC' if new_major_version in [24, 26] else 'ALL'
            else:
                upgrade_type = 'ALL'

            if upgrade_type == 'ALL':
                break  # if any of the dimms require the ALL upgrade, that's the file we have to use
        return 'P_V{}_{}.img'.format(new_major_version, upgrade_type)

    def _get_micron_firmware_filename_second_gen(self, new_major_version, diags):
        for dimm_data in diags.values():
            current_major_version = int(dimm_data['activeFirmwareVersion'].split('.')[0])

            if current_major_version == 20:
                upgrade_type = 'PSOC' if new_major_version == 20 else 'ALL'
            elif current_major_version in [22, 25]:
                upgrade_type = 'PSOC' if new_major_version in [22, 25] else 'ALL'
            else:
                upgrade_type = 'ALL'

            if upgrade_type == 'ALL':
                break  # if any of the dimms require the ALL upgrade, that's the file we have to use
        return '*AGIGA-SRI-RAM4PA.RIVER16-V{}*-UPGRADE_{}-signed.img'.format(
            new_major_version/10, upgrade_type)

    def get_nvdimm_diagnostics(self, hw_cfg=None):
        """Get the diagnostics data for NVDIMMs - currently supports SMART and Micron NVDIMMs.

        :param sf_api.result_models.GetHardwareConfig hw_cfg: json blob from get_hardware_config_ssh,
            if it has already been called. Use this parameter to prevent repeated calls to that function

        :return: Dict of values obtained from the command "sfnvdimm_diagnostics -a"
        :rtype: dict
        """
        diags_cmd = 'sfnvdimm_diagnostics -a | tr -d " "'

        _hardware_config = hw_cfg or self.get_hardware_config_ssh()
        vendor = _hardware_config.nvram_vendor
        nvids = _hardware_config.nvids
        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command(cmd=diags_cmd, raise_on_failure=False, timeout=60)
        json_diags = json.loads(result.stdout)

        if vendor.is_smart():
            for field in ['slot0FirmwareVersion',
                          'slot1FirmwareVersion',
                          'activeFirmwareVersion']:
                for dimm in json_diags:
                    if re.match(r'^\d+\.\d[A-F]$', dimm[field]):
                        # convert firmware notation from (for example) '26.2C' to '2.12'
                        dimm[field] = '.'.join(str(int(digit, 16)) for digit in dimm[field][-2:])

        diags = dict(zip(nvids, json_diags))

        _log.debug('NVDIMM diagnostics, parsed:\n{}'.format(json.dumps(diags, sort_keys=True)))
        return diags

    def verify_minimum_number_of_nvdimms_exist(self, nvdimm_diags=None):
        """Checks to see if two NVDIMMSs exist.

        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function

        :raises: valence.functional_models.VerifyNvdimmFailedError if there is less than 2 NVDIMMS
        """
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()
        if len(diags) < 2:
            raise _errors.VerifyNvdimmFailedError('Expected at least 2 NVDIMMs, found only {}'.format(
                len(diags)))

    def verify_nvdimm_health(self, nvdimm_diags=None):
        """Verify nvdimms armed, have no ecc errors, and have sufficient energy for csave.

        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function
        """
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()

        self.verify_nvdimm_fw_version(nvdimm_diags=diags)
        self.verify_nvdimm_armed(nvdimm_diags=diags)
        self.verify_nvdimm_sufficient_csave_energy(nvdimm_diags=diags)
        self.verify_nvdimm_no_ecc_errors()

    def verify_nvdimm_armed(self, nvdimm_diags=None):
        """Verify that all of the NVDIMMs are armed.

        Checks the armStatusArmSuccess and armStatusSaveNArmed values from "sfnvdimm_diagnostics".

        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function

        :raises: valence.functional_models.VerifyNvdimmFailedError if the NVDIMMS are not Armed
        """
        errors = []
        should_be_true_keys = ['armStatusArmSuccess', 'armStatusSaveNArmed']
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()

        # Always expect 2 NVDIMMs per node, until we see otherwise.
        self.verify_minimum_number_of_nvdimms_exist(nvdimm_diags=diags)
        for nvid, slot_data in diags.items():
            for key in should_be_true_keys:
                if key not in slot_data:
                    errors.append('Unexpected or incomplete output from "smart-nvdimm --diags --all"'
                                  ', node {}'.format(self))
                elif not slot_data[key]:
                    errors.append('Expected {} to be True nvid {}, found False'
                                  .format(key, nvid))
        if errors:
            raise _errors.VerifyNvdimmFailedError('verify_nvdimm_armed() - found errors:\n * {}'
                                                  .format('\n * '.join(errors)))

    def verify_nvdimm_sufficient_csave_energy(self, nvdimm_diags=None):
        """Check if the diagnositics indicate that there is sufficient energy for catastrophic save.

        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function

        :raises: valence.functional_models.VerifyNvdimmFailedError if there is no sufficient csave energy
        """
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()

        for nvid, slot_data in diags.items():
            if slot_data['notEnoughEnergyForCatastrophicSave']:
                err = 'Not enough energy for catastrophic save'
                raise _errors.VerifyNvdimmFailedError('verify_nvdimm_sufficient_csave_energy() - {}found'
                                                      'error: {}'.format(nvid, err))

    def verify_nvdimm_no_ecc_errors(self):
        """Function to check for ecc errors.

        :raises: valence.functional_models.VerifyNvdimmFailedError if there are ecc errors on the node
        """
        cmd = 'ipmitool sel elist | grep Uncorrectable.*Asserted'

        with self.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command(cmd=cmd, raise_on_failure=False, timeout=30)

        if result.stdout:
            raise _errors.VerifyNvdimmFailedError('There are some ecc errors on the Node {}'.format(
                result.stdout))

    def verify_nvdimm_fw_version(self, expected_version=None, nvdimm_diags=None):
        """Verify that the firmware is the expected version.

        If the version does not match the expected value, an exception is thrown.

        :param str expected_version: Version number, e.g., 2.11, 26.2C, etc
        :param dict nvdimm_diags: json blob from get_nvdimm_diagnostics, if it's already been called.
                                  Use this parameter to prevent repeated calls to that function

        :raises: valence.functional_models.VerifyNvdimmFailedError if verification of firmware
        """
        errors = []
        diags = nvdimm_diags or self.get_nvdimm_diagnostics()

        self.verify_minimum_number_of_nvdimms_exist(nvdimm_diags=diags)
        for nvid, slot_data in diags.items():
            if 'activeFirmwareVersion' not in slot_data:
                errors.append('Unexpected or incomplete output from "smart-nvdimm --diags --all"')
            if expected_version and expected_version not in slot_data['activeFirmwareVersion']:
                errors.append('Expected firmware version "{}", found "{}" - nvid {}'
                              .format(expected_version, slot_data['activeFirmwareVersion'], nvid))
        if errors:
            raise _errors.VerifyNvdimmFailedError('verify_nvdimm_firmware_version() - found errors:\n * '
                                                  '{}'.format('\n * '.join(errors)))

    def umount_nvram(self):
        """Setup for I/O to nvram by stopping/disabling the solidfire service and unmounting nvram."""
        unmount_cmd = 'umount /mnt/pendingDirtyBlocks'

        _log.debug('Stopping and disabling solidfire service, unmounting /mnt/pendingDirtyBlocks.')

        self.stop_binary_ssh(enums.NodeBinary.solidfire)
        self.disable_binary_ssh(enums.NodeBinary.solidfire)
        with self.get_ssh_connection() as node:
            node.run_command(unmount_cmd, raise_on_failure=True)

    def mkfs_nvram(self, device_path=None):
        """Return to normal state by mounting the device and starting/enabling the solidfire service.

        :param str device_path: optional full filepath for device to be remounted
            /dev/pmem0p1 for nvdimm
            /dev/nvme0n1 for nvme
        """
        device_path = device_path or self.get_cache_device_path()
        mkfs_cmd = 'mkfs.ext4 -F {0}'.format(device_path)
        mount_cmd = 'mount {0} /mnt/pendingDirtyBlocks'.format(device_path)

        _log.debug(('Mounting {} at /mnt/pendingDirtyBlocks, '
                    'starting and enabling solidfire service.').format(device_path))

        with self.get_ssh_connection() as node:
            node.run_command(mkfs_cmd, raise_on_failure=True)
            node.run_command(mount_cmd, raise_on_failure=True)
        try:
            self.start_binary_ssh(enums.NodeBinary.solidfire)
        except IndexError:
            # ignore IndexError. start_binary tries to return PID of process via stdout.split()[-1]
            # but 'start solidfire' prints nothing, so [-1] is out of range
            pass
        self.enable_binary_ssh(enums.NodeBinary.solidfire)

    def fio_write_pattern(self, pattern, bs='256k', size='7G', device_path=None):
        """Write data to device with journaled I/O via FIO.

        :param str pattern: Pattern to be written to the device; the less compressible the better
        :param str bs: block size for I/O units
        :param str size: total size of I/O
        :param str device_path: optional full filepath for device to be validated
            /dev/pmem0p1 for nvdimm
            /dev/nvme0n1 for nvme

        """
        device_path = device_path or self.get_cache_device_path()
        write_command = ('fio --bs={0} --direct=1 --ioengine=libaio --rw=write --size={1} '
                         '--verify=pattern --verify_pattern={2} '
                         '--verify_interval=4096 --do_verify=0 --verify_state_save=1 '
                         '--name=write_verify_job --filename={3}'.format(bs, size, pattern, device_path))

        _log.debug('writing data to device via FIO')
        with self.get_ssh_connection() as node:
            node.run_command(write_command, raise_on_failure=True)

    def fio_verify_pattern(self, pattern, bs='256k', size='7G', device_path=None):
        """Validate data written to NVDIMM from journaled I/O via FIO.

        :param str pattern: Pattern that was written to the device when fio_write_pattern was run
        :param str bs: block size for I/O units that was used when fio_write_pattern was run
        :param str size: total size of I/O that was used when fio_write_pattern was run
        :param str device_path: optional full filepath for device to be remounted
            /dev/pmem0p1 for nvdimm
            /dev/nvme0n1 for nvme
        """
        device_path = device_path or self.get_cache_device_path()
        validate_command = ('fio --bs={0} --direct=1 --ioengine=libaio --rw=read --verify=pattern '
                            '--verify_pattern={1} --verify_interval=4096 --size={2} '
                            '--verify_state_load=1 --do_verify=1 --verify_fatal=1 --verify_dump=1 '
                            '--exitall_on_error=1 --name=write_verify_job --filename={3}'.format(
                                bs, pattern, size, device_path))

        _log.debug('validating data written to device via fio_write_pattern')
        with self.get_ssh_connection() as node:
            node.run_command(validate_command, raise_on_failure=True)

    def get_cache_device_path(self):
        """Get the full filepath for the node's cache device.

        :rtype: str
        """
        nvram_vendor = self.get_hardware_config_ssh().nvram_vendor
        return '/dev/nvme0n1' if nvram_vendor.is_radian() else '/dev/pmem0p1'
