"""Node FM methods related to OTPW."""
import time

import autologger
import sf_ssh

from valence.functional_models import _errors
from valence.functional_models._node import base

_log = autologger.get_sflogging()

# Constants related to enabling and disabling one time password.
# This code was initially copied from sf_platform.core.OTPW and slightly modified.
SSHD_CONFIG = '/etc/ssh/sshd_config'
PAM_CONFIG = '/etc/pam.d/sshd'

# This is exactly what the OTPW enabling code does in oxygen. It should be safe
# back to B/C/N...
# see oxygen/install/solidfire-otp/postinst
PAM_SSH_OTPW_ENABLE_CMD = r'''sudo ed {pam_conf} <<EOF
g/^@include common-auth/\\
s/^/#/\\
a\\
# This will skip the next two modules for users in sfmgmt group,\\
#  eg skip OTPW for local admin users\\
auth [success=2 default=ignore] pam_succeed_if.so user ingroup sfmgmt\\
# OTPW authentication for users not in sfmgmt group\\
auth       required     pam_otpw.so nolock\\
session    optional     pam_otpw.so nolock\\
# This will skip next module for users not in sfmgmt group, eg do normal (password)\\
# auth for local admin users but don't allow OTWP failure to then try normal auth\\
auth [success=1 default=ignore] pam_succeed_if.so user notingroup sfmgmt\\
# Standard unix authorization for users in sfmgmt group\\
auth sufficient pam_unix.so debug\\
.
w
q
EOF'''.format(pam_conf=PAM_CONFIG)


SSHD_CONFIG_OTPW_ENABLE_CMD = (
    r'sudo sed -i -e "s/ChallengeResponseAuthentication[ ]\+no/ChallengeResponseAuthentication yes/"'
    r' -e "s/^#[[:space:]]*ChallengeResponseAuthentication/ChallengeResponseAuthentication/"'
    r' -e "s/^#?PasswordAuthentication[ ]\+yes/PasswordAuthentication no/" {sshd_conf}'
).format(sshd_conf=SSHD_CONFIG)


SSHD_CONFIG_GROUP_MANIPULATION_CMD = r"""sudo sed -i -e 's/AllowUsers.*//' {sshd_conf};
if sudo grep -q -s 'AllowGroups' {sshd_conf}; then
    sudo grep -q -s 'AllowGroups.*sfadmin' {sshd_conf} ||\
    sudo sed -i -e 's/AllowGroups[[:space:]]*\(.*\)/AllowGroups \1 sfadmin/' {sshd_conf}
else
    sudo echo 'AllowGroups sfadmin' | sudo tee -a {sshd_conf} > /dev/null
fi""".format(sshd_conf=SSHD_CONFIG)

# if the banner is commented out, re-add it.
SSHD_BANNER_FIX_CMD = (
    r"""sudo bash -c 'grep -q -s "^Banner" {sshd_conf} &> /dev/null || """
    """sudo echo "Banner /etc/issue.net" | sudo tee -a {sshd_conf} > /dev/null'"""
).format(sshd_conf=SSHD_CONFIG)

SSHD_RESTART_CMD = 'sudo restart ssh'

SSHD_CONDITIONAL_PERMIT_ROOT = (
    '[[ -e /var/lib/sf_root_enabled ]] && ('
    "sudo sed -i 's/^PermitRootLogin\\s.*/PermitRootLogin yes/' {sshd_conf};"
    "sudo sed -i -r 's/^AllowGroups\\s(.*)$/AllowGroups \\1 root/' {sshd_conf};"
    "sudo sed -i -r 's/^AllowGroups\\s(.*) root root$/AllowGroups \\1 root/' {sshd_conf};"
    '); true'
).format(sshd_conf=SSHD_CONFIG)


RESTORE_BACKED_UP_CONFIGS = (
    'sudo cp /sf/backup/sshd /etc/pam.d/; '
    'sudo cp /sf/backup/sshd_config /etc/ssh/'
)

OTPW_ENABLE_CMDS = (
    RESTORE_BACKED_UP_CONFIGS,
    PAM_SSH_OTPW_ENABLE_CMD,
    SSHD_CONFIG_OTPW_ENABLE_CMD,
    SSHD_CONFIG_GROUP_MANIPULATION_CMD,
    '[[ -e /var/lib/sf_no_otp ]] && sudo rm /var/lib/sf_no_otp; true',
    SSHD_CONDITIONAL_PERMIT_ROOT,
    SSHD_BANNER_FIX_CMD,
    SSHD_RESTART_CMD
)


OTPW_DISABLE_CMDS = (
    RESTORE_BACKED_UP_CONFIGS,
    'sudo touch /var/lib/sf_no_otp',
    SSHD_CONDITIONAL_PERMIT_ROOT,
    SSHD_BANNER_FIX_CMD,
    SSHD_RESTART_CMD
)


class OtpwMixin(base.NodeBase):
    """Node FM OTPW methods."""

    def disable_otpw(self):
        """Disable one time password."""
        # make sure we leave root login enabled if it already was
        # and add a banner line if we don't have it already.
        with self.get_ssh_connection() as ssh:
            for cmd in OTPW_DISABLE_CMDS:
                _log.debug('command: {}'.format(cmd))
                result = ssh.run_command(cmd, use_sudo=False)
                if result.return_code != 0:
                    msg = ('Failed to disable OTP on node because running [{}]'
                           ' failed with rc={}: {}').format(cmd, result.return_code, result.stderr)
                    raise _errors.FunctionalNodeError(msg)

        # Make sure OTP is really gone
        try:
            self.verify_otpw_is_not_enabled()
        except sf_ssh.errors.SshError:
            # This occurs sometimes when the SSH daemon is still being restarted after disabling otpw
            _log.debug('Got SSHException when verifying that OTPW was disabled. Retrying after sleep.')
            time.sleep(10)
            self.verify_otpw_is_not_enabled()

    def enable_otpw(self):
        """Enable one time password."""
        with self.get_ssh_connection() as ssh:
            for cmd in OTPW_ENABLE_CMDS:
                _log.debug('command: {}'.format(cmd))
                result = ssh.run_command(cmd, use_sudo=False)
                if result.return_code != 0:
                    msg = ('Failed to disable OTP on node because running [{}]'
                           ' failed with rc={}: {}').format(cmd, result.return_code, result.stderr)
                    raise _errors.FunctionalNodeError(msg)

        # Make sure OTP is really enabled
        try:
            self.restore_otpw_list()
        except sf_ssh.errors.SshError:
            # This occurs sometimes when the SSH daemon is still being restarted after enabling otpw
            _log.debug('Got SSHException when restoring OTPW list. Retrying after sleep.')
            time.sleep(10)
            self.restore_otpw_list()
        self.verify_otpw_is_enabled()

    def verify_otpw_is_enabled(self):
        """Verify that one time password is enabled.

        :raises valence.functional_models.FunctionalNodeError: If OTPW prompt was not encountered
        """
        with sf_ssh.Ssh(self.mip, self.node_dm.username, force_otpw=True) as ssh:
            if not isinstance(ssh.auth_method, sf_ssh.auth.AuthOtpw):
                _log.info('auth_method = {!s}'.format(ssh.auth_method))
                raise _errors.FunctionalNodeError("Didn't get OTPW prompt when we should")

    def verify_otpw_is_not_enabled(self):
        """Verify that one time password is not enabled.

        :raises valence.functional_models.FunctionalNodeError: If OTPW prompt was encountered
        """
        with self.get_ssh_connection() as ssh:
            if isinstance(ssh.auth_method, sf_ssh.auth.AuthOtpw):
                raise _errors.FunctionalNodeError("Got OTPW prompt when we shouldn't")

    def restore_otpw_list(self):
        """Restore original OTPW password list by reinstalling package."""
        with self.get_ssh_connection() as ssh:
            cmd = 'apt-get install -y --only-upgrade --reinstall "solidfire-otp-*"'
            result = ssh.run_command(cmd, use_sudo=True)
            if result.return_code == 0:
                _log.info('Restored original OTPW list')
                return
            cmd = 'cp .otpw.orig .otpw'
            result = ssh.run_command(cmd, use_sudo=True)
            if result.return_code == 0:
                _log.info('Restored original OTPW list')
                return
            raise _errors.FunctionalNodeError('Failed to restore OTPW list')
