"""Node FM methods related to packages."""

import autologger

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._node import base

_log = autologger.get_sflogging()


class PackageMixin(base.NodeBase):
    """Node FM package methods."""

    def list_package_info(self, name_regex=None, active=None, factory_installed=None,
                          show_in_ui=None, version=None):
        """The ListPackageInfo API method call.

        :param str name_regex: Package name
        :param bool active: True, If the package version is active
        :param bool factory_installed: True, If the package version is factory_installed
        :param bool show_in_ui: If True, the package shows up in TUI
        :param str version: Package version
        :rtype: list[valence.models.Package]
        """
        params = {}
        params[api_constants.NAME_REGEX] = name_regex
        params[api_constants.ACTIVE] = active
        params[api_constants.FACTORY_INSTALLED] = factory_installed
        params[api_constants.SHOW_IN_UI] = show_in_ui
        params[api_constants.VERSION] = version

        response = self._call_method(enums.NodeMethod.list_package_info,
                                     params=params,
                                     result_dm_class=result_models.ListPackageInfo)

        return response.result_dm.packages

    def set_active_package_version(self, name, version):
        """The SetActivePackageVersion API method call.

        :param str name: Package name
        :param str version: Package version
        :rtype: valence.models.Package
        """
        params = {}
        params[api_constants.NAME] = name
        params[api_constants.VERSION] = version

        response = self._call_method(enums.NodeMethod.set_active_package_version,
                                     params=params,
                                     result_dm_class=result_models.SetActivePackageVersion)

        return response.result_dm.package

    def install_package(self, url, set_active=None, persist=None):
        """The InstallPackage API method call.

        :param str url: URL to the .tgz package to install
        :param bool set_active: Set active flag for package
        :param bool persist: Persist package install files across iRTFI
        :rtype: valence.models.Package
        """
        params = {}
        params[api_constants.URL] = url
        params[api_constants.SET_ACTIVE] = set_active
        params[api_constants.PERSIST] = persist

        response = self._call_method(enums.NodeMethod.install_package,
                                     params=params,
                                     result_dm_class=result_models.InstallPackage)

        return response.result_dm.package

    def uninstall_package(self, name, version):
        """The UninstallPackage API method call.

        :param str name: Package name
        :param str version: Package version
        :rtype: valence.models.Package
        """
        params = {}
        params[api_constants.NAME] = name
        params[api_constants.VERSION] = version

        response = self._call_method(enums.NodeMethod.uninstall_package,
                                     params=params,
                                     result_dm_class=result_models.UninstallPackage)

        return response.result_dm.package
