"""Node FM methods related restarting/resetting."""

import autologger
from sf_rtfi import node_rtfi

from sf_api import (
    api_constants,
    api_errors,
    result_models,
)
from valence import (
    enums,
    models,
)
from valence.functional_models import _async_handler
from valence.functional_models._node import base

_log = autologger.get_sflogging()


class RestartMixin(base.NodeBase):
    """Node FM Restart methods."""

    def reset_node(self, build=None, reboot=None):
        """Reset the Node to Factory Defaults.

        Note: Builds <= Boron 5.1256 will reboot, and Carbon <= 6.447 will reboot the node by default.
        The build string should be a URL to a squashfs mount.

        :param valence.models.SfVersion build: Element software version to which the node will be reset
        :param bool reboot: whether or not to allow node to reboot
        :rtype: valence.functional_models.RtfiAsyncHandler
        """
        build_location = None
        node = self.node_dm
        node_info = node.serialize()
        node_dm = models.RtfiNodeAdapter(node_info)
        if build:
            network = node.network_profile
            build_location = build.get_pxe_url(network)
        params = {
            api_constants.BUILD: build_location,  # send location
            api_constants.REBOOT: reboot,
            api_constants.SF_AUTO: '1',  # Enables SSH on the node
            api_constants.USE_THE_FORCE_LUKE: True,
            api_constants.SF_ALLOW_VIRTNODE: '1'
        }
        rtfi_watcher = node_rtfi.NodeRtfiProvider(node_dm, build, {}, {})
        rtfi_watcher.start_reactor()
        try:
            self._call_method(enums.NodeMethod.reset_node, params=params,
                              result_dm_class=result_models.ResetNode)
        except api_errors.SfApiException:
            rtfi_watcher.reactor.join()
            raise

        rtfi_async_handler = _async_handler.RtfiAsyncHandler(rtfi_watcher.reactor)

        return rtfi_async_handler
