"""Node FM methods related to services."""

from sf_api import (
    api_constants,
    result_models,
)
from valence import enums
from valence.functional_models._node import base


class ServiceMixin(base.NodeBase):
    """Node FM Service methods."""

    def start_binary(self, service):
        """Start a SolidFire service via the API.

        :param valence.enums.NodeBinary service: enum of the service to start
        :rtype: sf_api.result_models.RestartServices
        """
        return self._restart_binary(service, action=enums.RestartServicesAction.start)

    def _restart_binary(self, service=None, action=None):
        """Call RestartServices.

        This is private, since start_binary and stop_binary are the preferred methods to use.

        :param valence.enums.NodeBinary service: Enum of the service to manipulate.
        :param valence.enums.RestartServicesAction action: Enum of action to perform
        :rtype: sf_api.result_models.RestartServices
        """
        params = {
            # Force must always be equal to True for this call
            api_constants.FORCE: True,
            api_constants.SERVICE: service,
            api_constants.ACTION: action
        }
        return self._call_method(enums.NodeMethod.restart_services, params=params,
                                 result_dm_class=result_models.RestartServices).result_dm

    def restart_binary(self, service):
        """Restart a SolidFire service via the API.

        :param valence.enums.NodeBinary service: enum of the service to stop
        :rtype: sf_api.result_models.RestartServices
        """
        return self._restart_binary(service, action=enums.RestartServicesAction.restart)

    def stop_binary(self, service):
        """Stop a SolidFire service via the API.

        :param valence.enums.NodeBinary service: enum of the service to stop
        :rtype: sf_api.result_models.RestartServices
        """
        return self._restart_binary(service, action=enums.RestartServicesAction.stop)

    def start_binary_ssh(self, service):
        """Start a service via SSH.

        :param valence.enums.NodeBinary service: enum of the service to start
        :rtype: int
        :return: PID of running service
        """
        with self.get_ssh_connection() as ssh_conn:
            cmd = 'start {}'.format(service.value)
            out = ssh_conn.run_command(cmd, raise_on_failure=True).stdout
        return int(out.split()[-1])

    def stop_binary_ssh(self, service):
        """Stop a service via SSH.

        :param valence.enums.NodeBinary service: enum of the service to stop
        """
        with self.get_ssh_connection() as ssh_conn:
            cmd = 'stop {}'.format(service.value)
            ssh_conn.run_command(cmd, raise_on_failure=True)

    def _get_pid(self, service, service_id=None):
        """Fetches the pid for the given service and service_id.

        :param valence.enums.NodeService service: enum of the service for which pid is needed
        :param sf_api.models.types.IntIdType service_id: service_id of the service
        :rtype: int
        :return: PID of the service
        """
        if service_id:
            cmd = 'ps aux | grep \"service={}.*-serviceid={}\"'.format(service.value,
                                                                       service_id)
        else:
            cmd = 'ps aux | grep \"service={}\"'.format(service.value)
        with self.get_ssh_connection() as ssh_conn:
            run_command_output = ssh_conn.run_command(cmd, raise_on_failure=True)
            commands_string = run_command_output.stdout.split('\n')
            pid = [i.split()[1] for i in commands_string if 'grep' not in i and i != '']
        return pid[0]

    def kill_master_service_ssh(self, signal):
        """Kill master service via SSH.

        :param valence.enums.UnixSignal signal: enum of the unix signal to be used with kill command
        """
        with self.get_ssh_connection() as ssh_conn:
            pid = self._get_pid(enums.NodeService.master)
            cmd = 'kill -{} {}'.format(signal.value, pid)
            ssh_conn.run_command(cmd, raise_on_failure=True)

    def kill_drive_service_ssh(self, service_dm, signal):
        """Kill drive service via SSH.

        :param valence.models.ClusterServiceInfo service_dm: enum of the service to stop
        :param valence.enums.UnixSignal signal: enum of the unix signal to be used with kill command
        """
        with self.get_ssh_connection() as ssh_conn:
            pid = self._get_pid(service_dm.service.service_type, service_dm.service.service_id)
            cmd = 'kill -{} {}'.format(signal.value, pid)
            ssh_conn.run_command(cmd, raise_on_failure=True)

    def enable_binary_ssh(self, service):
        """Enable a service via SSH.

        :param valence.enums.NodeBinary service: enum of the service to enable
        """
        with self.get_ssh_connection() as ssh_conn:
            cmd = 'service {} enable'.format(service.value)
            ssh_conn.run_command(cmd, raise_on_failure=True)

    def disable_binary_ssh(self, service):
        """Disable a service via SSH.

        :param valence.enums.NodeBinary service: enum of the service to disable
        """
        with self.get_ssh_connection() as ssh_conn:
            cmd = 'service {} disable'.format(service.value)
            ssh_conn.run_command(cmd, raise_on_failure=True)
