import unittest

import mock
from b2sdk import B2SDK

import sf_api
from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)

module_mock = unittest_utils.ModuleLevelMock()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestBourbonMixin(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()
        cls.switchport = models.SwitchPort.get_mock_object()
        cls.bourbon = B2SDK(username='somename', password='somepass')

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch.object(B2SDK, 'find_switch_ports')
    def test_get_switchports(self, mock_find_switch_ports, mock_bourbon, mock_call_method):
        mock_call_method.return_value = result_models.GetNetworkConfig.get_mock_response(
            version=models.SfVersion('10.0'), force_mock=dict(network=dict(bond_1g=True, bond_10g=True)))
        mock_bourbon.return_value = self.bourbon
        switch_ports = {u'response': [{u'address': u'44:a8:42:2c:f8:79', u'interface': u'0/31',
                                       u'interface_type': u'switch_port', u'port_channel': None,
                                       u'switch': u'10.117.18.98'},
                                      {u'address': u'44:a8:42:2c:f8:7d',
                                       u'interface': u'0/31',
                                       u'interface_type': u'switch_port',
                                       u'port_channel': None, u'switch': u'10.117.18.101'}]}

        mock_find_switch_ports.return_value = switch_ports
        sps = self.node_fm.get_attached_switchports()
        assert sps[0] == models.SwitchPort(switch_ports['response'][0])

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch.object(functional_models.Node, 'get_attached_switchports')
    @mock.patch.object(B2SDK, 'get_interfaces_config')
    def test_get_vlans(self, mock_get_vlans, mock_get_switchport, mock_bourbon, mock_call_method):
        api_data = {u'response': [{u'access_native_vlan': None, u'interface': u'0/19',
                                   u'port_channel': None, u'switch': u'10.117.18.140',
                                   u'tagged_vlans': u'', u'trunk_native_vlan': None},
                                  {u'access_native_vlan': None, u'interface': u'1/19',
                                   u'port_channel': None, u'switch': u'10.117.18.143',
                                   u'tagged_vlans': u'', u'trunk_native_vlan': None}]}

        switch_ports = [models.SwitchPort(a) for a in api_data['response']]

        vlans = {u'response': [{u'access_native_vlan': 2065, u'interface': u'0/19',
                                u'port_channel': None, u'switch': u'10.117.18.140',
                                u'tagged_vlans': u'', u'trunk_native_vlan': 2065},
                               {u'access_native_vlan': 1065, u'interface': u'1/19',
                                u'port_channel': None, u'switch': u'10.117.18.143',
                                u'tagged_vlans': u'', u'trunk_native_vlan': 1065}]}
        mock_get_vlans.return_value = vlans
        mock_get_switchport.return_value = switch_ports
        mock_bourbon.return_value = self.bourbon
        result = self.node_fm.get_vlans()
        assert result[0] == models.SwitchPort(vlans['response'][0])

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch('b2sdk.B2SDK.set_interfaces_config')
    def test_up_the_down_switchports(self, mock_int_config, mock_bourbon, mock_call_method):

        switch_port_data = {u'address': u'44:a8:42:2c:f8:79', u'interface': u'0/31',
                                        u'interface_type': u'switch_port', u'port_channel': None,
                                        u'switch': u'10.117.18.98'}

        sp = models.SwitchPort(switch_port_data)
        mock_bourbon.return_value = self.bourbon
        self.node_fm.down_switchports([sp])
        mock_int_config.assert_called_with(addresses=[sp.ip],
                                           interface_type=enums.BourbonInterfaceType.
                                           switch_port.value,
                                           interfaces=[sp.port_slot], enabled=False)
        self.node_fm.up_switchports([sp])
        mock_int_config.assert_called_with(addresses=[sp.ip],
                                           interface_type=enums.BourbonInterfaceType.
                                           switch_port.value,
                                           interfaces=[sp.port_slot], enabled=True)

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch('b2sdk.B2SDK.set_interfaces_config')
    @mock.patch('b2sdk.B2SDK.get_interfaces')
    def test_mode_bond_mode(self, mock_get_int, mock_int_config, mock_bourbon, mock_call_method):

        switch_port_data = {u'address': u'44:a8:42:2c:f8:79', u'interface': u'0/31',
                                        u'interface_type': u'switch_port', u'port_channel': None,
                                        u'switch': u'10.117.18.98'}

        mock_get_int.return_value = {u'response': [{u'interface_type': u'port_channel',
                                     u'interfaces': [1, 7, 30, 31, 32, 33, 99, 100]}]}

        sp = models.SwitchPort(switch_port_data)
        mock_bourbon.return_value = self.bourbon
        self.node_fm.set_bond_mode([sp], enums.BourbonBondMode.active)
        mock_int_config.assert_called_with(addresses=[sp.ip],
                                           interface_type=enums.BourbonInterfaceType.
                                           port_channel.value,
                                           interfaces=[1, 7, 30, 31, 32, 33, 99, 100], permutate=True,
                                           bond_mode=enums.BourbonBondMode.active.value)

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch('b2sdk.B2SDK.set_interfaces_config')
    def test_set_untagged_vlans(self, mock_set_interfaces, mock_bourbon, mock_call_method):
        switchport = self.switchport
        vlan = 1065
        mock_bourbon.return_value = self.bourbon
        self.node_fm.set_untagged_vlans([switchport], vlan)
        mock_set_interfaces.assert_called_with(switches=[switchport.ip],
                                               interface_type=enums.BourbonInterfaceType.
                                               switch_port.value,
                                               interfaces=[switchport.port_slot],
                                               trunk_native_vlan=vlan,
                                               access_native_vlan=vlan)

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch('b2sdk.B2SDK.set_interfaces_config')
    def test_unset_untagged_vlans(self, mock_set_interfaces, mock_bourbon, mock_call_method):
        switchport = self.switchport
        vlan = None
        mock_bourbon.return_value = self.bourbon
        self.node_fm.clear_untagged_vlans([switchport])
        mock_set_interfaces.assert_called_with(switches=[switchport.ip],
                                               interface_type=enums.BourbonInterfaceType.
                                               switch_port.value,
                                               interfaces=[switchport.port_slot],
                                               trunk_native_vlan=vlan,
                                               access_native_vlan=vlan)

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch('b2sdk.B2SDK.set_interfaces_config')
    def test_set_tagged_vlans(self, mock_set_interfaces, mock_bourbon, mock_call_method):
        switchport = self.switchport
        vlan = [1065]
        mock_bourbon.return_value = self.bourbon
        self.node_fm.set_tagged_vlans([switchport], vlan)
        mock_set_interfaces.assert_called_with(switches=[switchport.ip],
                                               interface_type=enums.BourbonInterfaceType.
                                               switch_port.value,
                                               interfaces=[switchport.port_slot],
                                               tagged_vlans=vlan)

    @mock.patch.object(functional_models.Node, 'bourbon', new_callable=mock.PropertyMock)
    @mock.patch('b2sdk.B2SDK.set_interfaces_config')
    def test_unset_tagged_vlans(self, mock_set_interfaces, mock_bourbon, mock_call_method):
        switchport = models.SwitchPort.get_mock_object()
        mock_bourbon.return_value = self.bourbon
        vlan = []
        self.node_fm.clear_tagged_vlans([switchport])
        mock_set_interfaces.assert_called_with(switches=[switchport.ip],
                                               interface_type=enums.BourbonInterfaceType.
                                               switch_port.value,
                                               interfaces=[switchport.port_slot],
                                               tagged_vlans=vlan)
