import unittest

import mock
import pytest
import requests
import sf_models
import sf_ssh

import sf_api
from valence import (
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')


def tearDownModule():
    module_mock.restore()


@mock.patch.object(functional_models.Node, 'create_bundle')
@mock.patch.object(functional_models.Node, '_cleanup_remote_node')
class TestPullSupportBundle(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.artifact_manager = mock.MagicMock()
        self.artifact_manager.store_artifacts_remote.return_value = ['name1']

    def test_pull_support_bundle_success(self, mock_cleanup_remote, mock_create_bundle):
        mock_create_bundle.return_value = functional_models.BundleInfo(
            attempts=[(functional_models._node.BundleMethod.api, True)]
        )
        self.node_fm.pull_support_bundle(self.artifact_manager)

    def test_pull_support_bundle_fail(self, mock_cleanup_remote, mock_create_bundle):
        mock_create_bundle.return_value = functional_models.BundleInfo(
            attempts=[(functional_models._node.BundleMethod.api, False)]
        )
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm.pull_support_bundle(self.artifact_manager)

    def test_pull_support_bundle_bad_destination_path(self, mock_cleanup_remote, mock_create_bundle):
        mock_create_bundle.return_value = functional_models.BundleInfo()
        self.artifact_manager.store_artifacts_remote.return_value = []
        self.node_fm.pull_support_bundle(self.artifact_manager)


@mock.patch.object(functional_models.Node, '_attempt_making_bundle')
class TestCreateBundle(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.node_fm.cluster_dm.api_endpoint = models.SfVersion('10.0')
        self.node_fm.api_supported = mock.MagicMock(return_value=True)

    def test_create_bundle_success(self, mock_make_bundle):
        result = self.node_fm.create_bundle(compress=True, zk_heap_dump=True, collect_backtraces=True)
        assert isinstance(result, functional_models.BundleInfo)
        assert '--compress' in mock_make_bundle.call_args[1]['extra_args']
        assert 'zk_heap_dump' in mock_make_bundle.call_args[1]['extra_args']
        assert 'backtraces' in mock_make_bundle.call_args[1]['extra_args']

        self.node_fm.create_bundle(compress=False, zk_heap_dump=False, collect_backtraces=False)
        assert '--compress' not in mock_make_bundle.call_args[1]['extra_args']
        assert 'zk_heap_dump' not in mock_make_bundle.call_args[1]['extra_args']
        assert 'backtraces' not in mock_make_bundle.call_args[1]['extra_args']

    def test_create_bundle_collect_backtraces_workaround(self, mock_make_bundle):
        self.node_fm.cluster_dm.api_endpoint = models.SfVersion('7.3')
        self.node_fm.create_bundle(collect_backtraces=True)
        # This flag can cause an EIO with 7.3, so it shouldn't be passed to make_bundle
        assert 'backtraces' not in mock_make_bundle.call_args[1]['extra_args']

    def test_create_bundle_pre_extra_args(self, mock_make_bundle):
        self.node_fm.cluster_dm.api_endpoint = models.SfVersion('5.0')
        self.node_fm.create_bundle()
        # This flag can cause an EIO with 7.3, so it shouldn't be passed to make_bundle
        assert mock_make_bundle.call_args[1]['extra_args'] == ''

    def test_create_bundle_api_error(self, mock_make_bundle):
        mock_make_bundle.side_effect = _errors.NodeBundleError
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm.create_bundle()

    def test_create_bundle_generic_error(self, mock_make_bundle):
        mock_make_bundle.side_effect = _errors.NodeBundleError
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm.create_bundle()

    def test_create_bundle_api_not_supported(self, mock_make_bundle):
        self.node_fm.api_supported.return_value = False
        self.node_fm.create_bundle()
        assert mock_make_bundle.call_args[0][0] is functional_models._node.BundleMethod.ssh


@mock.patch('sf_ssh.Ssh.run_command')
@mock.patch.object(functional_models.Node, '_verify_file')
class TestVerifyBundleContents(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.node_fm.node_dm.model = sf_models.NodeType.STANDARD_TYPES.SF3010

    def setUp(self):
        self.bundle_info = functional_models.BundleInfo(bundle_path='somewhere.tar.gz')

    def test_verify_bundle_contents_success(self, mock_verify_file, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='win', stderr='', return_code=0)
        mock_verify_file.return_value = (True, True)
        self.node_fm._verify_bundle_contents(self.bundle_info)

    def test_verify_bundle_contents_fail(self, mock_verify_file, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='lose', return_code=1)
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm._verify_bundle_contents(self.bundle_info)

    def test_verify_bundle_contents_missing_file(self, mock_verify_file, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='win', stderr='', return_code=0)
        mock_verify_file.return_value = (False, True)
        self.node_fm._verify_bundle_contents(self.bundle_info)
        assert len(self.bundle_info.missing_files) > 0

    def test_verify_bundle_contents_empty_file(self, mock_verify_file, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='win', stderr='', return_code=0)
        mock_verify_file.return_value = (True, False)
        self.node_fm._verify_bundle_contents(self.bundle_info)
        assert len(self.bundle_info.empty_files) > 0


@mock.patch.object(functional_models.Node, '_cleanup_remote')
@mock.patch('sf_ssh.Ssh.run_command')
class TestCleanupRemote(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.bundle_info = functional_models.BundleInfo()

    @mock.patch.object(functional_models.Node, '_verify_bundle_removed')
    @mock.patch.object(functional_models.Node, 'delete_all_support_bundles')
    def test_cleanup_remote_success_via_api(self, mock_delete_all_bundles, mock_verify_removed,
                                            mock_run_command, mock_generic_cleanup):
        mock_verify_removed.return_value = True
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='win', stderr='', return_code=0)
        self.bundle_info.method = functional_models._node.BundleMethod.api
        self.node_fm._cleanup_remote_node(self.bundle_info, purge_cores=True, purge_crash_dumps=True)
        assert len(self.bundle_info.remote_rm_paths) > 0

    def test_cleanup_remote_success_not_via_api(self, mock_run_command, mock_generic_cleanup):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='win', stderr='', return_code=0)
        self.bundle_info.method = functional_models._node.BundleMethod.ssh
        self.node_fm._cleanup_remote_node(self.bundle_info, purge_cores=True, purge_crash_dumps=True)
        assert len(self.bundle_info.remote_rm_paths) > 0

    @mock.patch.object(functional_models.Node, '_verify_bundle_removed')
    @mock.patch.object(functional_models.Node, 'delete_all_support_bundles')
    def test_cleanup_remote_fail(self, mock_delete_all_bundles, mock_verify_removed, mock_run_command,
                                 mock_generic_cleanup):
        mock_verify_removed.return_value = False
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='win', stderr='', return_code=0)
        self.bundle_info.method = functional_models._node.BundleMethod.api
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm._cleanup_remote_node(self.bundle_info, purge_cores=False,
                                              purge_crash_dumps=False)


@mock.patch('sf_ssh.Ssh.run_command')
class TestGenericCleanupRemote(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_generic_cleanup_remote_success(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout="it's gone. are you happy?", stderr='', return_code=0)
        self.node_fm._cleanup_remote({'path'})

    def test_generic_cleanup_remote_run_command_fails(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='you messed up', return_code=1)
        with pytest.raises(_errors.BundleError):
            self.node_fm._cleanup_remote({'path'})

    def test_generic_cleanup_remote_ssh_error(self, mock_run_command):
        mock_run_command.side_effect = sf_ssh.errors.SshError
        with pytest.raises(_errors.BundleError):
            self.node_fm._cleanup_remote({'path'})


@mock.patch.object(functional_models.Node, '_create_bundles_api')
class TestAttemptMakingBundle(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.bundle_info = functional_models.BundleInfo()

    @mock.patch.object(functional_models.Node, '_verify_bundle_contents')
    def test_attempt_making_bundle_success(self, mock_create_bundles_api, mock_verify):
        mock_create_bundles_api.return_value = 'har.tar.gz'
        self.node_fm._attempt_making_bundle(functional_models._node.BundleMethod.api, self.bundle_info,
                                            extra_args='--compress', verify_logs=True)
        mock_verify.assert_called()
        assert self.bundle_info.method is functional_models._node.BundleMethod.api

    @mock.patch.object(functional_models.Node, '_create_bundles_ssh')
    def test_attempt_making_bundle_ssh_success(self, mock_create_bundles_ssh, mock_create_bundles_api):
        mock_create_bundles_ssh.return_value = 'masterclass.gz'
        self.node_fm._attempt_making_bundle(functional_models._node.BundleMethod.ssh, self.bundle_info,
                                            extra_args='', verify_logs=False)
        mock_create_bundles_ssh.assert_called()
        assert not mock_create_bundles_api.called
        assert self.bundle_info.method is functional_models._node.BundleMethod.ssh

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_attempt_making_bundle_compress_manually(self, mock_run_command, mock_create_bundles_api):
        mock_create_bundles_api.return_value = 'a_file'
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='pass', stderr='', return_code=0)
        self.node_fm._attempt_making_bundle(functional_models._node.BundleMethod.api, self.bundle_info,
                                            extra_args='', verify_logs=False)
        assert self.bundle_info.bundle_path == 'a_file.gz'

    @mock.patch.object(functional_models.Node, '_get_sf_msg')
    @mock.patch('sf_ssh.Ssh.run_command')
    def test_attempt_making_bundle_compress_manually_error(self, mock_run_command, mock_get_msg,
                                                           mock_create_bundles_api):
        mock_create_bundles_api.return_value = 'a_file'
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='many bad things happened', return_code=1)
        self.node_fm._attempt_making_bundle(functional_models._node.BundleMethod.api, self.bundle_info,
                                            extra_args='', verify_logs=False)
        mock_get_msg.assert_called()


@mock.patch.object(sf_api.NodeApi, 'http_get')
@mock.patch.object(functional_models.Node, 'create_support_bundle')
class TestCreateBundlesApi(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.root_uri = cls.node_fm.api.root_uri
        cls.create_bundle_dm = sf_api.result_models.CreateSupportBundle.get_mock_object(
            overrides={'details': {'url': ['{}/bundle.tar.gz'.format(cls.root_uri),
                                           'http://fake.1.1.1/bundle.tar.gz']}}
        )

    def test_create_bundles_api_success(self, mock_create_bundle, mock_http_get):
        mock_create_bundle.return_value = self.create_bundle_dm
        self.node_fm._create_bundles_api('he who must be named', '')
        mock_http_get.assert_called_once()

    def test_create_bundles_api_http_get_fails(self, mock_create_bundle, mock_http_get):
        mock_create_bundle.return_value = self.create_bundle_dm
        mock_http_get.side_effect = requests.HTTPError
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm._create_bundles_api('mario', '')


@mock.patch('sf_ssh.Ssh.run_command')
class TestCreateBundlesSsh(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.bundle_info = functional_models.BundleInfo()

    def test_create_bundles_ssh_success(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='thepath', stderr='', return_code=0)
        path = self.node_fm._create_bundles_ssh('unnamed name', '')
        assert path == 'thepath'

    def test_create_bundles_ssh_script_fails(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='so much fail', return_code=1)
        with pytest.raises(functional_models.NodeBundleError):
            self.node_fm._create_bundles_ssh('Gordon Freeman', '')

    def test_create_bundles_ssh_script_timeout(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='outtatime', return_code=-99)
        with pytest.raises(functional_models.NodeBundleError):
            self.node_fm._create_bundles_ssh('Marty', '')

    def test_create_bundles_ssh_partial_collection(self, mock_run_command):
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='out', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='words', stderr='planet error', return_code=1),
        ]
        with pytest.raises(functional_models.NodeBundleError):
            self.node_fm._create_bundles_ssh('Mr. Bubbles', '')


class TestNodeBundleHelperMethods(unittest.TestCase):
    """Test all the smaller helper methods."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_verify_bundle_removed(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0)
        assert self.node_fm._verify_bundle_removed() is True

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_verify_bundle_removed_false(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='bad things', stderr='', return_code=1)
        assert self.node_fm._verify_bundle_removed() is False

    def test_get_sf_msg(self):
        msg = self.node_fm._get_sf_msg('Message')
        assert 'Message' in msg
        assert self.node_fm.mip in msg

    def test_truncate_bundle_name_if_necessary(self):
        max_len = functional_models._node.bundle_mixin.SUPPORT_BUNDLE_NAME_LENGTH
        name = self.node_fm._truncate_bundle_name_if_necessary('name' * max_len)
        assert len(name) == max_len

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_verify_file_no_contents(self, mock_run_command):
        ssh_obj = sf_ssh.Ssh('fakeip')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='0', stderr='', return_code=0)
        result = self.node_fm._verify_file(ssh_obj, 'folder', '/rel/path')
        assert result == (True, False)

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_verify_file_has_contents(self, mock_run_command):
        ssh_obj = sf_ssh.Ssh('fakeip')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='1024', stderr='', return_code=0)
        result = self.node_fm._verify_file(ssh_obj, 'folder', '/rel/path')
        assert result == (True, True)

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_verify_file_found_in_other_location(self, mock_run_command):
        ssh_obj = sf_ssh.Ssh('fakeip')
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='', stderr='not found bro', return_code=1),
            sf_ssh.ssh.RunCommandOutput(stdout='/here/it/is', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='201', stderr='', return_code=0),
        ]
        result = self.node_fm._verify_file(ssh_obj, 'folder', '/rel/path')
        assert result == (True, True)

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_verify_file_not_found_anywhere(self, mock_run_command):
        ssh_obj = sf_ssh.Ssh('fakeip')
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='', stderr='itsa nota here', return_code=1),
            sf_ssh.ssh.RunCommandOutput(stdout='', stderr='nope', return_code=1),
        ]
        result = self.node_fm._verify_file(ssh_obj, 'folder', '/rel/path')
        assert result == (False, True)


@mock.patch('sf_api.NodeApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestCreateSupportBundle(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.create_bundle_response = sf_api.result_models.CreateSupportBundle.get_mock_response()

    def test_create_support_bundle_success(self, mock_call_method):
        self.create_bundle_response.result_dm.result = sf_api.enums.ApiTestResult.passed
        mock_call_method.return_value = self.create_bundle_response
        result = self.node_fm.create_support_bundle('lizard')
        assert isinstance(result, sf_api.result_models.CreateSupportBundle)

    def test_create_support_bundle_fail(self, mock_call_method):
        self.create_bundle_response.result_dm.result = sf_api.enums.ApiTestResult.failed
        mock_call_method.return_value = self.create_bundle_response
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm.create_support_bundle('wizard')


@mock.patch('sf_api.NodeApi.call_method',
            side_effect=unittest_utils.helpers.check_params_serializable)
class TestDeleteAllSupportBundles(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def setUp(self):
        self.delete_bundle_response = sf_api.result_models.DeleteAllSupportBundles.get_mock_response()

    def test_delete_bundles_success(self, mock_call_method):
        self.delete_bundle_response.result_dm.result = sf_api.enums.ApiTestResult.passed
        mock_call_method.return_value = self.delete_bundle_response
        self.node_fm.delete_all_support_bundles()

    def test_delete_bundles_fail(self, mock_call_method):
        self.delete_bundle_response.result_dm.result = sf_api.enums.ApiTestResult.failed
        mock_call_method.return_value = self.delete_bundle_response
        with pytest.raises(_errors.NodeBundleError):
            self.node_fm.delete_all_support_bundles()
