import unittest

import mock
import pytest
import requests

import sf_api
from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeGetNetworkConfig(unittest.TestCase):
    """Test Node.get_network_config."""

    def test_get_network_config(self, mock_call_method):
        mock_call_method.return_value = result_models.GetNetworkConfig.get_mock_response(
            version=models.SfVersion('10.0'))
        node = functional_models.Node.get_mock_fm()
        network_config_dm = node.get_network_config()
        assert isinstance(network_config_dm, models.NetworkConfig)
        assert mock_call_method.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeSetNetworkConfig(unittest.TestCase):
    """Test Node.set_network_config."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.set_response = result_models.GetNetworkConfig.get_mock_response(
            version=models.SfVersion('10.0'), force_mock=dict(network=dict(bond_1g=True)))

    def test_set_network_config(self, mock_call_method):
        mock_call_method.return_value = self.set_response
        network_config_dm = self.set_response.result_dm.network
        returned_config = self.node_fm.set_network_config(network_config_dm=network_config_dm)
        assert isinstance(returned_config, models.NetworkConfig)

    def test_set_network_config_sparse(self, mock_call_method):
        sparse_network_config_dm = models.NetworkConfig.create(bond_1g=None)
        mock_call_method.return_value = self.set_response
        new_nc_dm = self.node_fm.set_network_config(network_config_dm=sparse_network_config_dm)
        # Network config data model should be updated
        assert new_nc_dm.bond_1g is not None

    @mock.patch.object(functional_models.Node, 'ip', new_callable=mock.PropertyMock)
    def test_change_mip(self, mock_ip, mock_call_method):
        new_mip = '10.117.64.99'
        mock_call_method.return_value = self.set_response
        config = models.NetworkConfig.create(bond_1g={'address': new_mip})
        self.node_fm.set_network_config(config)
        assert self.node_fm.mip == new_mip
        # Make sure it changed the IP to the sip, then changed it back to the mip
        calls = [mock.call(self.node_fm.node_dm.sip), mock.call(new_mip)]
        mock_ip.assert_has_calls(calls)

    @mock.patch.object(functional_models.Node, 'ip', new_callable=mock.PropertyMock)
    def test_change_sip(self, mock_ip, mock_call_method):
        mock_call_method.return_value = self.set_response
        config = models.NetworkConfig.create(bond_10g={'address': '10.117.64.9'})
        self.node_fm.set_network_config(config)
        mock_ip.assert_not_called()

    @mock.patch.object(functional_models.Node, 'ip', new_callable=mock.PropertyMock)
    def test_change_mip_net0(self, mock_ip, mock_call_method):
        new_mip = '10.117.64.99'
        mock_call_method.return_value = self.set_response
        config = models.NetworkConfig.create(net0={'address': new_mip})
        self.node_fm.set_network_config(config)
        assert self.node_fm.mip == new_mip
        # Make sure it changed the IP to the sip, then changed it back to the mip
        calls = [mock.call(self.node_fm.node_dm.sip), mock.call(new_mip)]
        mock_ip.assert_has_calls(calls)


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeSetConfig(unittest.TestCase):
    """Test Node.set_config."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()
        cls.set_response = result_models.GetConfig.get_mock_response(
            version=models.SfVersion('10.0'))
        cls.set_config_dm = models.ClusterConfig(dict(name='WitnessName'))

    def test_set_config(self, mock_call_method):
        mock_call_method.return_value = self.set_response
        network_config_dm = self.set_response.result_dm.config.network
        returned_config = self.node_fm.set_config(config_dm=self.set_config_dm,
                                                  network_config_dm=network_config_dm['mocked_value'])
        assert isinstance(returned_config, models.ClusterNodeConfig)

    def test_set_config_sparse(self, mock_call_method):
        sparse_network_config_dm = models.NetworkConfig.create(bond_1g=None)
        mock_call_method.return_value = self.set_response
        new_nc_dm = self.node_fm.set_config(network_config_dm=sparse_network_config_dm)
        # Network config data model should be updated
        assert new_nc_dm.network['mocked_value'].serialize() is not None

    @mock.patch.object(functional_models.Node, 'ip', new_callable=mock.PropertyMock)
    def test_change_mip(self, mock_ip, mock_call_method):
        new_mip = '10.117.64.99'
        mock_call_method.return_value = self.set_response
        config = models.NetworkConfig.create(bond_1g={'address': new_mip})
        self.node_fm.set_config(config)
        assert self.node_fm.mip == new_mip
        # Make sure it changed the IP to the sip, then changed it back to the mip
        calls = [mock.call(self.node_fm.node_dm.sip), mock.call(new_mip)]
        mock_ip.assert_has_calls(calls)

    @mock.patch.object(functional_models.Node, 'ip', new_callable=mock.PropertyMock)
    def test_change_sip(self, mock_ip, mock_call_method):
        mock_call_method.return_value = self.set_response
        config = models.NetworkConfig.create(bond_10g={'address': '10.117.64.9'})
        self.node_fm.set_config(config)
        mock_ip.assert_not_called()


@mock.patch.object(requests.Session, 'post')
class TestNodeGetNetworkInterfaceStatus(unittest.TestCase):
    """Test Node.get_network_interface_status."""

    def test_get_network_interface_status(self, mock_post):
        node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        get_response = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'))
        requests_response = mock.MagicMock()
        requests_response.json.return_value = get_response.raw_data
        mock_post.return_value = requests_response
        node_network_interfaces = node_fm.get_network_interface_status()
        for key in node_network_interfaces:
            assert isinstance(node_network_interfaces[key], enums.NetworkStatus)


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetNetworkInterfaceStatusInternal(unittest.TestCase):
    """Test the internal _get_network_interface_status call."""
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()

    def test_get_network_interface_status(self, mock_call_method):
        self.node_fm._get_network_interface_status(interfaces=[enums.NetworkInterface.eth0])
        mock_call_method.assert_called_with(enums.NodeMethod.get_network_interface_status,
                                            params={'interfaces': [enums.NetworkInterface.eth0.value]},
                                            result_dm_class=result_models.NodeGetNetworkInterfaceStatus)


@mock.patch.object(functional_models.Node, '_get_network_interface_status')
@mock.patch.object(functional_models.Node, '_set_network_interface_status')
class TestNodeSetNetworkInterfaceStatus(unittest.TestCase):
    """Test Node.set_network_interface_status_and_verify."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.set_response = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'))

    def test_success(self, set_status, get_status):
        """Test the case where everything works."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'eth0': enums.NetworkStatus.up_and_running.value,
                                      'eth1': enums.NetworkStatus.up_and_running.value}})

        self.node_fm.set_network_interface_status_and_verify(
            interfaces=[enums.NetworkInterface.eth0, enums.NetworkInterface.eth1],
            verify=True, status=enums.NetworkStatus.up_and_running, minutes=2)

        set_status.assert_called_once()
        get_status.assert_called_once()

    def test_network_interface_name(self, set_status, get_status):
        """Test the fm method with a NetworkInterfaceName object."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'Bond10G.3001': enums.NetworkStatus.up_and_running.value,
                                      'Bond10G.3001:S': enums.NetworkStatus.up_and_running.value}})

        self.node_fm.set_network_interface_status_and_verify(
            interfaces=[models.NetworkInterfaceName('Bond10G.3001'),
                        models.NetworkInterfaceName('Bond10G.3001:S')],
            verify=True, status=enums.NetworkStatus.up_and_running, minutes=2)

        set_status.assert_called_once()
        get_status.assert_called_once()

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_success_on_1G(self, mock_auth, set_status, get_status):
        """Test the case where everything works."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'Bond1G': enums.NetworkStatus.up_and_running.value}})

        self.node_fm.set_network_interface_status_and_verify(
            interfaces=[enums.NetworkInterface.bond_1g],
            verify=True, status=enums.NetworkStatus.up_and_running, minutes=2)

    def test_one_failure(self, set_status, get_status):
        """Test the case where an interface doesn't enter the desired state."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'eth0': enums.NetworkStatus.up_and_running.value,
                                      'eth1': enums.NetworkStatus.up_and_running.value}})

        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.set_network_interface_status_and_verify(
                interfaces=[enums.NetworkInterface.eth0, enums.NetworkInterface.eth1],
                verify=True, status=enums.NetworkStatus.down.value, minutes=2)
        assert set_status.call_count == 2
        assert get_status.call_count == 2

    def test_connection_error(self, set_status, get_status):
        """Test the case where we get a ConnectionError."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'eth0': enums.NetworkStatus.up_and_running.value,
                                      'eth1': enums.NetworkStatus.up_and_running.value}})

        set_status.side_effect = requests.ConnectionError
        self.node_fm.set_network_interface_status_and_verify(
            interfaces=[enums.NetworkInterface.eth0, enums.NetworkInterface.eth1],
            verify=True, status=enums.NetworkStatus.up_and_running, minutes=2)
        set_status.assert_called_once()
        get_status.assert_called_once()

    def test_no_verify_success(self, set_status, get_status):
        """Test the case where verify is False and status is success."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'eth0': enums.NetworkStatus.up_and_running.value,
                                      'eth1': enums.NetworkStatus.up_and_running.value}})

        self.node_fm.set_network_interface_status_and_verify(
            interfaces=[enums.NetworkInterface.eth0, enums.NetworkInterface.eth1],
            verify=False, status=enums.NetworkStatus.up_and_running, minutes=2)
        set_status.assert_called_once()

    def test_no_verify_connection_error(self, set_status, get_status):
        """Test the case where verify is False and we get ConnectionError."""
        get_status.return_value = result_models.NodeGetNetworkInterfaceStatus.get_mock_response(
            version=models.SfVersion('10.0'),
            overrides={'interfaces': {'eth0': enums.NetworkStatus.up_and_running.value,
                                      'eth1': enums.NetworkStatus.up_and_running.value}})

        set_status.side_effect = requests.ConnectionError
        self.node_fm.set_network_interface_status_and_verify(
            interfaces=[enums.NetworkInterface.eth0, enums.NetworkInterface.eth1],
            verify=False, status=enums.NetworkStatus.up_and_running, minutes=2)
        set_status.assert_called_once()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestSetNetworkInterfaceStatusInternal(unittest.TestCase):
    """Test the internal _set_network_interface_status call."""

    def test_set_network_interface_status(self, mock_call_method):
        node = functional_models.Node.get_mock_fm()
        node._set_network_interface_status(interfaces=[enums.NetworkInterface.eth0],
                                           status=enums.NetworkStatus.up)
        mock_call_method.assert_called()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestListNetworkInterfaces(unittest.TestCase):
    def test_list_network_interfaces(self, mock_call_method):
        dc = result_models.ListNetworkInterfaces.get_mock_response()
        mock_call_method.return_value = dc
        node = functional_models.Node.get_mock_fm()
        result = node.list_network_interfaces()
        assert result == dc.result_dm.interfaces
