
import unittest

import mock
import pytest
import requests

import sf_api
from sf_api import api_constants
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


class TestNodeBase(unittest.TestCase):
    """Miscellaneous methods from NodeBase class."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()
        cls.version = str(cls.node_fm.api.version)

    def test_node_repr(self):
        assert 'NodeFm' in repr(self.node_fm)
        assert 'mvip=' in repr(self.node_fm)
        assert self.node_fm.cluster_dm.mvip in repr(self.node_fm)
        assert 'mip=' in repr(self.node_fm)
        assert self.node_fm.mip in repr(self.node_fm)

    @mock.patch.object(functional_models.Cluster, 'list_all_nodes')
    def test_get_node_dm(self, mock_list_all_nodes):
        nodes = models.ActiveNode.get_mock_object_list(2)

        # NOTE: this will need to change once the static ID is different.
        nodes[0].static_id = self.node_fm.node_dm.static_id

        mock_list_all_nodes.return_value = nodes
        result = self.node_fm.get_node_dm()
        assert result is nodes[0]

    def test_ip_property(self):
        ip = self.node_fm.ip
        assert ip is not None
        self.node_fm.ip = ip
        assert self.node_fm.api.ip == self.node_fm.ip

    def test_get_node_dm_no_cluster(self):
        node_fm = functional_models.Node(node_dm=self.node_fm.node_dm, cluster_dm=None)
        with pytest.raises(functional_models.FunctionalNodeError):
            node_fm.get_node_dm()

    @mock.patch.object(functional_models.Node, 'get_api')
    def test_api_supported_true(self, mock_get_api):
        result = {
            api_constants.CURRENT_VERSION: self.version,
            api_constants.SUPPORTED_VERSIONS: ['4.0', '5.0', '5.1', '6.0', '7.0', '7.1', '7.2',
                                               '7.3', '7.4', '8.0', '8.1', '8.2', '8.3', '8.4',
                                               '8.5', '8.6', '8.7', '9.0', '9.1', '9.2', '9.3',
                                               '9.4', '9.5', '10.0', '10.1', '10.2', '10.3',
                                               '11.0'],
            self.version: [
                enums.NodeMethod.get_network_config.value
            ]

        }
        mock_get_api.return_value = result
        result = self.node_fm.api_supported(enums.NodeMethod.get_network_config)
        assert result is True

    @mock.patch.object(functional_models.Node, 'get_api')
    def test_api_supported_false(self, mock_get_api):
        result = {
            api_constants.CURRENT_VERSION: self.version,
            self.version: [
                enums.NodeMethod.get_network_interface_status.value
            ]

        }
        mock_get_api.return_value = result
        result = self.node_fm.api_supported(enums.NodeMethod.get_network_config)
        assert result is False

    @mock.patch.object(functional_models.Node, 'get_api')
    def test_api_supported_version_mismatch(self, mock_get_api):
        result = {
            api_constants.CURRENT_VERSION: '0.0',
            self.version: [
                enums.NodeMethod.get_network_interface_status.value
            ]

        }
        mock_get_api.return_value = result
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.api_supported(enums.NodeMethod.get_network_config)

    def test_static_id_property(self):
        assert self.node_fm.static_id is not None

    @mock.patch('b2sdk.B2SDK')
    def test_bourbon_property(self, mock_b2sdk):
        mock_b2sdk.return_value = ''
        bourbon = self.node_fm.bourbon
        assert bourbon == ''

    @mock.patch.object(functional_models.Node, 'get_api')
    def test_get_and_update_version(self, mock_get_api):
        stale_version = self.node_fm.api.version
        new_version_str = '{}.{}'.format(stale_version.version_tuple[0],
                                         stale_version.version_tuple[1] + 1)
        new_version = sf_api.SfVersion(new_version_str)
        mock_get_api.return_value = {
            api_constants.CURRENT_VERSION: str(new_version)
        }
        self.node_fm.update_api_version()
        assert self.node_fm.api.version == new_version

    @mock.patch.object(functional_models.Node, 'get_api')
    def test_api_responsive(self, mock_get_api):
        assert self.node_fm.api_is_responsive()
        mock_get_api.side_effect = [requests.HTTPError]
        assert not self.node_fm.api_is_responsive()

    def test_vsphere_storage(self):
        assert not self.node_fm.vsphere

    def test_vsphere(self):
        witness_dm = models.ResourceNode.get_mock_object(overrides=dict(model='witness'),
                                                         force_mock=dict(hypervisor_host=True))
        witness_fm = functional_models.Node(witness_dm, None)
        assert witness_fm.vsphere

    def test_vsphere_no_hypervisor(self):
        witness_dm = models.ResourceNode.get_mock_object(overrides=dict(model='witness'))
        witness_dm.hypervisor_host = None
        witness_fm = functional_models.Node(witness_dm, None)
        assert not witness_fm.vsphere
