import unittest

import mock
import pytest

import sf_api
from sf_api import result_models
from valence import (
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestTestHardwareConfig(unittest.TestCase):
    """Test Node.test_hardware_config."""

    def test_test_hardware_config(self, mock_call_method):
        hw_config_data = result_models.TestHardwareConfig.get_mock_response(
            include_optional=True, version=models.SfVersion('10.0'))
        mock_call_method.return_value = hw_config_data
        node = functional_models.Node.get_mock_fm()
        hardware_config_dm = node.test_hardware_config(raise_on_failure=False)
        assert isinstance(hardware_config_dm, sf_api.result_models.TestHardwareConfig)
        assert mock_call_method.called

    def test_test_hardware_config_raises(self, mock_call_method):
        hw_config_data = result_models.TestHardwareConfig.get_mock_response(
            include_optional=True, overrides={'details': {'mocked_value': {'Passed': False}}},
            version=models.SfVersion('10.0'))
        mock_call_method.return_value = hw_config_data
        node = functional_models.Node.get_mock_fm()
        with pytest.raises(_errors.FunctionalNodeError):
            node.test_hardware_config(raise_on_failure=True)
        assert mock_call_method.called

    def test_test_hardware_config_doesnt_raise(self, mock_call_method):
        hw_config_data = result_models.TestHardwareConfig.get_mock_response(
            include_optional=True, overrides={'details': {'mocked_value': {'Passed': False}}},
            version=models.SfVersion('10.0'))
        mock_call_method.return_value = hw_config_data
        node = functional_models.Node.get_mock_fm()
        hardware_config_dm = node.test_hardware_config(raise_on_failure=False)
        assert isinstance(hardware_config_dm, sf_api.result_models.TestHardwareConfig)
        assert mock_call_method.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetHardwareInfo(unittest.TestCase):
    def test_get_hardware_info(self, mock_call_method):
        dc = result_models.GetHardwareInfo.get_mock_response()
        mock_call_method.return_value = dc
        node = functional_models.Node.get_mock_fm()
        result = node.get_hardware_info()
        assert result == dc.result_dm.hardware_info


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestCheckProposedCluster(unittest.TestCase):
    """Test Node.check_proposed_cluster."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=False)

    def test_check_proposed_cluster(self, mock_call_method):
        proposed_cluster_data = result_models.CheckProposedCluster.get_mock_response(
            include_optional=True, overrides={'proposed_cluster_errors': []},
            version=models.SfVersion('11.0'))
        mock_call_method.return_value = proposed_cluster_data
        check_proposed_cluster_dm = self.node_fm.check_proposed_cluster([self.node_fm.node_dm],
                                                                        raise_on_failure=True)
        assert isinstance(check_proposed_cluster_dm, sf_api.result_models.CheckProposedCluster)
        assert mock_call_method.called

    def test_check_proposed_cluster_raises(self, mock_call_method):
        proposed_cluster_data = result_models.CheckProposedCluster.get_mock_response(
            include_optional=False, version=models.SfVersion('11.0'))
        mock_call_method.return_value = proposed_cluster_data
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.check_proposed_cluster([self.node_fm.node_dm], raise_on_failure=True)
        assert mock_call_method.called

    def test_check_proposed_cluster_doesnt_raise(self, mock_call_method):
        proposed_cluster_data = result_models.CheckProposedCluster.get_mock_response(
            include_optional=True, version=models.SfVersion('11.0'))
        mock_call_method.return_value = proposed_cluster_data
        check_proposed_cluster_dm = self.node_fm.check_proposed_cluster([self.node_fm.node_dm],
                                                                        raise_on_failure=False)
        assert isinstance(check_proposed_cluster_dm, sf_api.result_models.CheckProposedCluster)
        assert mock_call_method.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetNodeFipsDrivesReport(unittest.TestCase):
    def test_get_node_fips_drives_report(self, mock_call_method):
        node_dm = functional_models.Node.get_mock_fm()
        node_dm.get_node_fips_drives_report()
