import json
import time
import unittest

import mock
import pytest
import requests
import sf_mc
import sf_ssh

import sf_api
from sf_api import api_constants
from sf_api import models as api_models
from sf_api import result_models
from sf_api.models import types
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


@mock.patch.object(sf_ssh.Ssh, 'run_command')
class TestGetUptime(unittest.TestCase):
    def test_get_uptime(self, mock_run_command):
        node = functional_models.Node.get_mock_fm()
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='237320.88 2683148.12\n',
                                                                    stderr='',
                                                                    return_code=0)
        uptime = node.get_uptime(timeout=1)
        assert uptime == 237320
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='237320.88 2683148.12\n',
                                                                    stderr='SOme ERrror',
                                                                    return_code=0)
        with pytest.raises(_errors.FunctionalNodeError):
            _ = node.get_uptime(timeout=1)


@mock.patch('sf_ssh.Ssh.run_command')
class TestNodeCoreDumpCheck(unittest.TestCase):
    """Test Node.core_dump_check."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_no_cores_found(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='no matches found',
                                                                    return_code=2)
        self.node_fm.core_dump_check()

    def test_cores_found(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='/sf/core1', stderr='',
                                                                    return_code=0)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.core_dump_check()

    def test_cores_filtered(self, mock_run_command):
        """Cores are found, but filtered out."""
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='/sf/core1', stderr='',
                                                                    return_code=0)
        self.node_fm.core_dump_check(exclude_regexes=[r'.*1'])

    def test_cores_partially(self, mock_run_command):
        """Some cores are filtered out, but error is still raised."""
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='/sf/core1\n/sf/core2',
                                                                    stderr='', return_code=0)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.core_dump_check(exclude_regexes=[r'.*1'])

    def test_unexpected_return_code(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='error',
                                                                    return_code=1)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.core_dump_check()

    @mock.patch('time.sleep')
    def test_core_dump_in_progress(self, mock_sleep, mock_run_command):
        out1 = sf_ssh.ssh.RunCommandOutput(stdout='/sf/core-uncompressed1', stderr='', return_code=0)
        out2 = sf_ssh.ssh.RunCommandOutput(stdout='/sf/core1', stderr='', return_code=0)
        mock_run_command.side_effect = [out1, out2]
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.core_dump_check()
        assert mock_sleep.called


@mock.patch('sf_ssh.Ssh.run_command')
class TestNodeCrashDumpCheck(unittest.TestCase):
    """Test Node.crash_dump_check."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_no_crashes_found(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='no matches found',
                                                                    return_code=1)
        self.node_fm.crash_dump_check()

    def test_crashes_found(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='/var/crash/201705',
                                                                    stderr='', return_code=0)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.crash_dump_check()

    def test_crashes_filtered(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='201705\n201705',
            stderr='', return_code=0)
        self.node_fm.crash_dump_check(exclude_regexes=['2017.*'])

    def test_unexpected_error(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='error',
                                                                    return_code=2)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.crash_dump_check()


class TestNodeAuthenticate(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_dm = models.ResourceNode.get_mock_object()

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_authenticate_in_constructor(self, mock_authenticate_node):
        functional_models.Node(self.node_dm, cluster_dm=None, authenticate=True)
        assert mock_authenticate_node.called

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_authenticate(self, mock_authenticate_node):
        node_fm = functional_models.Node(self.node_dm, cluster_dm=None, authenticate=False)
        node_fm.authenticate()
        assert mock_authenticate_node.called

    @mock.patch('sf_api._base.BaseJsonRpcClient.authenticate')
    def test_authenticate_with_cluster_dm(self, mock_authenticate_node):
        cluster_dm = models.ResourceCluster.get_mock_object()
        node_fm = functional_models.Node(self.node_dm, cluster_dm=cluster_dm, authenticate=True)
        api = node_fm.api
        assert mock_authenticate_node.called
        node_fm.authenticate(cluster_dm)
        assert api is not node_fm.api


@mock.patch.object(sf_ssh, 'socket_connect', autospec=True)
class TestPerNodeUp(unittest.TestCase):
    def test_up(self, mock_socket):
        node = functional_models.Node.get_mock_fm()
        assert node.api_is_up()

    def test_err(self, mock_socket):
        node = functional_models.Node.get_mock_fm()
        mock_socket.side_effect = sf_ssh.SshError
        assert not node.api_is_up()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestReboot(unittest.TestCase):
    def test_reboot(self, mock_call_method):
        node = functional_models.Node.get_mock_fm()
        node.reboot_node()
        assert mock_call_method.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetDriveConfig(unittest.TestCase):
    def test_get(self, mock_call_method):
        dc = result_models.GetDriveConfig.get_mock_response()
        mock_call_method.return_value = dc
        node = functional_models.Node.get_mock_fm()
        result = node.get_drive_config()
        assert result == dc.result_dm.drive_config


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetHardwareConfig(unittest.TestCase):
    def test_get_hardware_config(self, mock_call_method):
        dc = result_models.GetHardwareConfig.get_mock_response()
        mock_call_method.return_value = dc
        node = functional_models.Node.get_mock_fm()
        result = node.get_hardware_config()
        assert result == dc.result_dm.hardware_config


@mock.patch('sf_ssh.Ssh.run_command')
class TestGetHardwareConfigSsh(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_get_hardware_config_ssh(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='{"nvramVendor": "Smart"}',
                                                                    stderr='', return_code=0)
        self.assertEqual(self.node_fm.get_hardware_config_ssh().nvram_vendor, enums.NvramVendor.smart)


class TestStartupFlags(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('sf_api.NodeApi.call_method')
    def test_get_startup_flags(self, mock_call_method):
        fake_result_dm = {'fog': 'bugz'}
        mock_call_method.return_value.result_dm.startup_flags = fake_result_dm
        result = self.node_fm.get_startup_flags(binaries=[enums.NodeBinary.sfapp], flags=['fakedrives'])
        assert result == fake_result_dm

    @mock.patch('sf_api.NodeApi.call_method')
    def test_set_startup_flags(self, mock_call_method):
        fake_result_dm = {'fog': 'bugz'}
        flags = {'fakedrives': True}
        mock_call_method.return_value.result_dm.startup_flags = fake_result_dm
        result = self.node_fm.set_startup_flags(flags=flags, reset=True)
        assert result == fake_result_dm


@mock.patch('sf_ssh.Ssh.run_command')
class TestNodeGetCoreFiles(unittest.TestCase):
    """Test Node.get_core_files."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    @mock.patch('sf_artifacts.managers.GlobalManager')
    def test_get_core_files(self, mock_artifact_manager, mock_run_command):
        out1 = sf_ssh.ssh.RunCommandOutput(stdout='/sf/core1/test.txt', stderr='', return_code=0)
        mock_run_command.side_effect = [out1]
        self.node_fm.get_core_files()

    def test_get_core_files_not_found(self, mock_run_command):
        out1 = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=2)
        mock_run_command.side_effect = [out1]
        self.node_fm.get_core_files()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeGetApi(unittest.TestCase):
    """Test Node.get_api."""

    def test_get_api(self, mock_call_method):
        mock_call_method.return_value = result_models.BaseApiResult.get_mock_response(
            version=models.SfVersion('10.0'))
        mock_node = functional_models.Node.get_mock_fm()
        mock_node.get_api()
        assert mock_call_method.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestSetClusterConfig(unittest.TestCase):
    """Test the set_cluster_config_call."""

    def test_set_cluster_config(self, mock_call_method):
        node_fm = functional_models.Node.get_mock_fm()
        cluster = 'cluster-string'
        node_fm.set_cluster_config(cluster=cluster)

        mock_call_method.assert_called_with(enums.NodeMethod.set_cluster_config,
                                            params={
                                                    api_constants.CLUSTER: {
                                                        api_constants.CLUSTER: 'cluster-string'},
                                                    api_constants.FORCE: False,
                                                    api_constants.USE_THE_FORCE_LUKE: False
                                                    },
                                            result_dm_class=result_models.SetClusterConfig
                                            )


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeConstants(unittest.TestCase):
    def test_get_constants(self, mock_call_method):
        response = result_models.GetConstants.get_mock_response()
        mock_call_method.return_value = response
        node_fm = functional_models.Node.get_mock_fm()
        result = node_fm.get_constants()
        get_param_dict = {'constants':
                          list(models.ClusterConstantsVerbose.create().serialize().keys())}
        mock_call_method.assert_called_with(enums.NodeMethod.get_node_constants,
                                            params=get_param_dict,
                                            result_dm_class=result_models.GetConstants)
        assert isinstance(result, models.ClusterConstants)

        result = node_fm.get_constants(changed_only=True)
        get_param_dict['changedOnly'] = True
        mock_call_method.assert_called_with(enums.NodeMethod.get_node_constants,
                                            params=get_param_dict,
                                            result_dm_class=result_models.GetConstants)
        assert isinstance(result, models.ClusterConstants)

    def test_get_constants_filter(self, mock_call_method):
        response = result_models.GetConstants.get_mock_response()
        mock_call_method.return_value = response
        node_fm = functional_models.Node.get_mock_fm()
        result = node_fm.get_constants(constants_dm=response.result_dm)
        get_param_dict = {'constants': list(response.result_dm.serialize().keys())}
        mock_call_method.assert_called_with(enums.NodeMethod.get_node_constants,
                                            params=get_param_dict,
                                            result_dm_class=result_models.GetConstants)
        assert isinstance(result, models.ClusterConstants)

    def test_set_constants(self, mock_call_method):
        response = result_models.SetConstants.get_mock_response()
        mock_call_method.return_value = response
        node_fm = functional_models.Node.get_mock_fm()

        result = node_fm.set_constants(response.result_dm)
        assert isinstance(result, models.ClusterConstants)
        mock_call_method.assert_called_with(enums.NodeMethod.set_node_constants,
                                            params=response.result_dm.serialize(),
                                            result_dm_class=result_models.SetConstants)

    def test_reset_constants(self, mock_call_method):
        response = result_models.ResetConstants.get_mock_response()
        mock_call_method.return_value = response
        node_fm = functional_models.Node.get_mock_fm()

        result = node_fm.reset_constants(response.result_dm)
        assert isinstance(result, models.ClusterConstants)
        reset_param_dict = {'constants': list(response.result_dm.serialize().keys())}
        mock_call_method.assert_called_with(enums.NodeMethod.reset_node_constants,
                                            params=reset_param_dict,
                                            result_dm_class=result_models.ResetConstants)


class TestClusterState(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('sf_api.NodeApi.call_method')
    def test_get_cluster_state(self, mock_call_method):
        mock_response = result_models.GetClusterState.get_mock_response()
        mock_call_method.return_value = mock_response
        result = self.node_fm.get_cluster_state()
        assert result == mock_response.result_dm


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestCreateCluster(unittest.TestCase):
    """Test the create_cluster call."""

    def test_create_cluster(self, mock_call_method):
        node_fm = functional_models.Node.get_mock_fm()
        mvip = types.IPAddressType()._mock()
        svip = types.IPAddressType()._mock()
        nodes = [models.ResourceNode.get_mock_object() for _ in range(4)]
        node_ips = [node.sip for node in nodes]
        node_fm.create_cluster(mvip, svip, nodes)
        ip = '[{}]'.format(node_fm.mip) if types.IPv6Type.valid_ip(node_fm.mip) else node_fm.mip
        mock_call_method.assert_called()

    def test_create_cluster_single_node(self, mock_call_method):
        node_fm = functional_models.Node.get_mock_fm()
        mvip = types.IPAddressType()._mock()
        svip = types.IPAddressType()._mock()
        nodes = [models.ResourceNode.get_mock_object() for _ in range(1)]
        node_fm.create_cluster(mvip, svip, nodes)
        ip = '[{}]'.format(node_fm.mip) if types.IPv6Type.valid_ip(node_fm.mip) else node_fm.mip
        call_args = mock_call_method.call_args[1]
        assert call_args['params'] == {
            api_constants.MVIP: mvip,
            api_constants.SVIP: svip,
            api_constants.USERNAME: 'admin',
            api_constants.PASSWORD: 'admin',
            api_constants.ACCEPT_EULA: True,
            api_constants.UNIQUE_ID: None,
            api_constants.ATTRIBUTES: None,
            api_constants.ALLOW_TWO_NODE_ENSEMBLE: False,
            api_constants.SKIP_NODE_TYPE_CHECKS: True
        }
        assert call_args['auth'].uri == 'https://{}/json-rpc/{}'.format(ip, node_fm.api.version)


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetBootstrap(unittest.TestCase):
    """Test the create_cluster call with a single node cluster."""

    def test_create_cluster(self, mock_call_method):
        node_fm = functional_models.Node.get_mock_fm()
        node_fm.get_bootstrap_config()
        ip = '[{}]'.format(node_fm.mip) if types.IPv6Type.valid_ip(node_fm.mip) else node_fm.mip
        mock_call_method.assert_called()


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetClusterConfig(unittest.TestCase):
    def test_get_cluster_config(self, mock_call_method):
        response = result_models.GetClusterConfig.get_mock_response()
        mock_call_method.return_value = response
        node_fm = functional_models.Node.get_mock_fm()
        result = node_fm.get_cluster_config()
        assert isinstance(result, models.ClusterConfig)


class TestOobPowerNode(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('sf_mc.MachineControl.machine_power_off_hard_wait')
    def test_power_off_via_oob(self, mock_machine_power_off_hard_wait):
        self.node_fm.power_off_via_oob(wait_for_power_down=True)

    @mock.patch('sf_mc.MachineControl.machine_power_off_hard')
    def test_power_off_via_oob_no_wait(self, mock_machine_power_off_hard):
        self.node_fm.power_off_via_oob(wait_for_power_down=False)

    @mock.patch('valence.functional_models.Node.verify_pernode_api')
    @mock.patch('sf_mc.MachineControl.machine_power_on_wait')
    def test_power_on_via_oob(self, mock_machine_power_on_wait,
                              mock_verify_pernode_api):
        mock_verify_pernode_api.return_value = True
        self.node_fm.power_on_via_oob(wait_for_power_up=True, wait_for_sfapp=True)

    @mock.patch('valence.functional_models.Node.poll_ping_sshd_up')
    @mock.patch('sf_mc.MachineControl.machine_power_on')
    def test_power_on_via_oob_wait_for_ping(self, mock_machine_power_on,
                                            mock_poll_ping_sshd_up):
        mock_poll_ping_sshd_up.return_value = []
        self.node_fm.power_on_via_oob(wait_for_power_up=False, wait_for_sfapp=False,
                                      wait_for_ping=True)

    @mock.patch('sf_mc.machine_power_status')
    def test_get_power_status(self, mock_get_status):
        mock_get_status.return_value = 'ON'
        assert self.node_fm.get_oob_power_status() is enums.OobPowerState.on

        mock_get_status.return_value = 'OFF'
        assert self.node_fm.get_oob_power_status() is enums.OobPowerState.off


class TestPollingHelpers(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('sf_ssh.poll_ssh_query')
    def test_poll_ping_sshd_up(self, mock_poll_pings):
        self.node_fm.poll_ping_sshd_up()

    @mock.patch('sf_ssh.poll_ssh_query')
    def test_poll_ping_sshd_down(self, mock_poll_pings):
        self.node_fm.poll_ping_sshd_down()

    @mock.patch('valence.functional_models.Node.get_api')
    def test_verify_pernode_api(self, mock_get_api):
        result = self.node_fm.verify_pernode_api()
        assert result is True

    @mock.patch('valence.functional_models.Node.get_api')
    def test_verify_pernode_api_fails(self, mock_get_api):
        mock_get_api.side_effect = requests.HTTPError
        result = self.node_fm.verify_pernode_api()
        assert result is False

    @mock.patch('valence.functional_models.Node.verify_pernode_api')
    @mock.patch('sf_api.ClusterApi.call_method')
    def test_poll_verify_pernode_api(self, mock_call_method, mock_verify_pernode_api):
        mock_verify_pernode_api.return_value = True
        self.node_fm.poll_verify_pernode_api()

    @mock.patch('valence.functional_models.Node.verify_pernode_api')
    @mock.patch('sf_api.ClusterApi.call_method')
    def test_poll_verify_pernode_api_fails(self, mock_call_method, mock_verify_pernode_api):
        mock_verify_pernode_api.return_value = True
        mock_verify_pernode_api.side_effect = functional_models._errors.PollMaxError
        with pytest.raises(functional_models._errors.PollMaxError):
            self.node_fm.poll_verify_pernode_api()


class TestHardReboot(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('sf_mc.machine_power_reboot_hard')
    def test_reboot_node_via_oob(self, mock_machine_power_reboot_hard):
        self.node_fm.reboot_node_via_oob(wait_for_ping=False)
        assert mock_machine_power_reboot_hard.called

    @mock.patch('valence.functional_models.Node.poll_ping_sshd_up')
    @mock.patch('sf_mc.machine_power_reboot_hard')
    def test_reboot_node_via_oob_wait_for_ping(self, mock_machine_power_reboot_hard,
                                               mock_poll_ping_sshd_up):
        mock_poll_ping_sshd_up.return_value = []
        self.node_fm.reboot_node_via_oob(wait_for_ping=True)
        assert mock_machine_power_reboot_hard.called


class TestFailBlockDrives(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('valence.functional_models.Node.get_drive_config')
    @mock.patch('valence.functional_models.Node._fail_drive_command')
    def test_fail_block_drives(self, mock_fail_drive_command, mock_config):
        configs = models.DriveConfig.get_mock_object()
        mock_config.return_value = configs
        drive_dms = models.Drive.get_mock_object_list(3)
        for drive in drive_dms:
            drive.slot = configs.drives[0].slot
        self.node_fm.fail_block_drives(drive_dms)

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_fail_drive_command(self, mock_run_command):
        fail_cmd = mock.Mock()
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        self.node_fm._fail_drive_command(fail_cmd)

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_fail_drive_command_fails(self, mock_run_command):
        fail_cmd = mock.Mock()
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=1)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm._fail_drive_command(fail_cmd)


class TestFailSliceDrive(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('valence.functional_models.Node.get_drive_config')
    @mock.patch('valence.functional_models.Node._fail_drive_command')
    def test_fail_slice(self, mock_fail_command, mock_config):
        configs = models.DriveConfig.get_mock_object()
        mock_config.return_value = configs
        drives = models.Drive.get_mock_object_list(2)
        for drive in drives:
            drive.slot = configs.drives[0].slot
        drive_dms = drives
        self.node_fm.fail_slice_drives(drive_dms)


class TestFailDrive(unittest.TestCase):
    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()

    @mock.patch('valence.functional_models.Node.fail_slice_drives')
    @mock.patch('valence.functional_models.Node.fail_block_drives')
    def test_fail_slice(self, mock_fail_command, mock_config):
        drives = models.Drive.get_mock_object_list(2)
        drives[0].type = enums.DriveType.slice
        drives[1].type = enums.DriveType.block
        self.node_fm.fail_drives(drives)


@mock.patch.object(sf_ssh.Ssh, 'run_command', autospec=True)
@mock.patch.object(functional_models.Node, 'get_drive_config', autospec=True)
class TestRestore(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.config_response = models.DriveConfig.get_mock_object()

    def test_restore_fail(self, mock_get, mock_run):
        mock_run.return_value = sf_ssh.ssh.RunCommandOutput(stdout='',
                                                            stderr='error',
                                                            return_code=0)
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.restore_failed_drives()

    def test_restore_drives(self, mock_get, mock_run):
        mock_drive = models.Drive.get_mock_object()
        mock_drive.slot = self.config_response.drives[0].slot
        mock_get.return_value = self.config_response
        mock_run.return_value = sf_ssh.ssh.RunCommandOutput(stdout='drives restored',
                                                            stderr='',
                                                            return_code=0)
        self.node_fm.restore_failed_drives(drive_dms=[mock_drive])
        assert mock_run.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeGetConfigExtended(unittest.TestCase):
    """Test Node.get_network_config."""

    def test_get_config_extended(self, mock_call_method):
        response = result_models.GetConfigExtended.get_mock_response(version=models.SfVersion('12.0'))
        mock_call_method.return_value = response
        node = functional_models.Node.get_mock_fm()
        config_extended_dm = node.get_config_extended()
        assert isinstance(config_extended_dm, models.ClusterNodeConfigExtended)
        assert mock_call_method.called


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestNodeSetConfigExtended(unittest.TestCase):
    """Test Node.set_network_config."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.response = result_models.SetConfigExtended.get_mock_response(
            version=models.SfVersion('12.0'))

    def test_set_config_extended(self, mock_call_method):
        mock_call_method.return_value = self.response
        network_dms = [models.NetworkInterfaceDetailsExtended.get_mock_object()]
        cluster_dm = models.ClusterConfig.get_mock_object()
        returned_config = self.node_fm.set_config_extended(extended_network_dms=network_dms,
                                                           cluster_config_dm=cluster_dm)
        assert isinstance(returned_config, models.ClusterNodeConfigExtended)


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestGetVersionInfo(unittest.TestCase):
    """Test Node.set_network_config."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_get_version_info(self, mock_call_method):
        mock_call_method.return_value = result_models.GetVersionInfo.get_mock_response()
        res = self.node_fm.get_version_info()

        assert isinstance(res, result_models.GetVersionInfo)


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestReadLink(unittest.TestCase):
    """Test Node.set_network_config."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_read_link(self, mock_call_method):
        mock_call_method.return_value = result_models.ReadLink.get_mock_response()
        res = self.node_fm.read_link('/path')

        assert isinstance(res, result_models.ReadLink)


@mock.patch.object(sf_api.NodeApi, 'call_method')
class TestRotateLogs(unittest.TestCase):
    """Test Node.set_network_config."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_rotate_logs(self, mock_call_method):
        self.node_fm.rotate_logs(force=True)

        mock_call_method.assert_called_with(
            enums.NodeMethod.rotate_logs,
            params={api_constants.FORCE: True},
            auth=mock.ANY, log_response=True,
            result_dm_class=None
        )


@mock.patch('sf_api.NodeApi.call_method')
class TestListAvailableNodes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.node_dms = [
            api_models.AvailableNode.create(mip='10.117.78.10', internal_role=enums.NodeRole.storage),
            api_models.AvailableNode.create(mip='10.117.78.11', internal_role=None),
            api_models.AvailableNode.create(mip='10.117.78.12', internal_role=enums.NodeRole.witness)
        ]
        cls.mock_response = result_models.ListAvailableNodes.get_mock_response(
            overrides=dict(nodes=cls.node_dms)
        )
        cls.rs_node_dm = models.ResourceNode.get_mock_object()

    def test_default(self, call_method):
        call_method.return_value = self.mock_response
        available_nodes = self.node_fm.list_available_nodes()
        assert len(available_nodes) == 3

    def test_static_id_filter(self, call_method):
        call_method.return_value = self.mock_response
        available_nodes = self.node_fm.list_available_nodes(static_ids=[self.node_dms[0].mip])
        assert len(available_nodes) == 1
        available_nodes = self.node_fm.list_available_nodes(static_ids=['fake_id'])
        assert len(available_nodes) == 0

    def test_node_role_filter(self, call_method):
        call_method.return_value = self.mock_response
        available_nodes = self.node_fm.list_available_nodes(node_roles=[enums.NodeRole.storage])
        assert len(available_nodes) == 2
        available_nodes = self.node_fm.list_available_nodes(node_roles=[enums.NodeRole.witness])
        assert len(available_nodes) == 1


class TestGetSetFipsNodeConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    @mock.patch.object(json, 'dumps')
    @mock.patch.object(json, 'loads')
    @mock.patch.object(functional_models.Node, 'restart_binary')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Node, 'get_ssh_connection')
    def test_set_fips(self, mock_get_ssh_conn, mock_run_command, mock_restart_binary, mock_loads,
                      mock_dumps):
        mock_loads.return_value = {}
        self.node_fm.set_fips_drives_config()
        mock_dumps.assert_called_with({'fipsDriveConfiguration': True}, indent=4)

    @mock.patch.object(json, 'loads')
    @mock.patch.object(functional_models.Node, 'get_ssh_connection')
    def test_get_fips(self, mock_get_conn, mock_loads):
        mock_loads.return_value = {'fipsDriveConfiguration': True}
        assert self.node_fm.get_fips_drives_config() is True


class TestFakeEncryptionCapability(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    @mock.patch.object(functional_models.Node, 'restart_binary')
    @mock.patch.object(functional_models.Node, 'get_ssh_connection')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Node, 'get_drive_config')
    def test_fake_drives_encryption_config(self, mock_drive_config, mock_run_command, mock_get_ssh_conn,
                                           mock_restart_binary):
        mock_drives = models.DriveInfo.get_mock_object_list(1)
        mock_return = mock.MagicMock()
        mock_return.drives = mock_drives
        mock_drive_config.return_value = mock_return
        self.node_fm.set_fake_drives_encryption_capability(enums.DriveEncryptionCapability.fips)


@mock.patch('sf_ssh.Ssh.run_command', autospec=True)
class TestSaveClearIpmitoolLogs(unittest.TestCase):
    """Test Node.save_clear_ipmitool_logs."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_save_clear_ipmitool_logs_file_saved(self, mock_run_command):
        mock_run_command.side_effect = [sf_ssh.ssh.RunCommandOutput(
                                            stdout='file.log\nOoooee look at all these file contents',
                                            stderr='', return_code=0),
                                        sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0)]
        self.node_fm.save_clear_ipmitool_logs()
        self.assertEqual(mock_run_command.call_count, 2)

    def test_save_clear_ipmitool_logs_no_save(self, mock_run_command):
        mock_run_command.side_effect = [sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0),
                                        sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0)]
        self.node_fm.save_clear_ipmitool_logs()
        self.assertEqual(mock_run_command.call_count, 2)


class TestWitnessNodeFm(unittest.TestCase):
    """Test witness node functional model methods."""
    def setUp(self):
        self.witness_dm = models.ResourceNode.get_mock_object(overrides=dict(model='witness'),
                                                              force_mock=dict(hypervisor_host=True))
        self.witness_fm = functional_models.Node(self.witness_dm, None)

    def test_is_witness_compatible(self):
        active_node = models.ActiveNode.get_mock_object()
        witness_fm_from_active = functional_models.Node(active_node, None)
        storage_dm = models.ResourceNode.get_mock_object(overrides=dict(model='SF3010'))
        storage_fm = functional_models.Node(storage_dm, None)
        assert not storage_fm._is_witness_compatible()
        assert not witness_fm_from_active._is_witness_compatible()
        assert self.witness_fm._is_witness_compatible()

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(sf_mc.esx_control.Vsphere, 'find_vms_by_name')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    def test_get_oob_power_status_on(self, mock_disconnect, mock_vm, mock_connect):
        mock_vm.return_value = [{'name': self.witness_dm.node_name, 'state': 'poweredOn'}]
        power_state = self.witness_fm.get_oob_power_status()
        assert power_state is enums.OobPowerState.on

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(sf_mc.esx_control.Vsphere, 'find_vms_by_name')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    def test_get_oob_power_status_off(self, mock_disconnect, mock_vm, mock_connect):
        mock_vm.return_value = [{'name': self.witness_dm.node_name, 'state': 'poweredOff'}]
        power_state = self.witness_fm.get_oob_power_status()
        assert power_state is enums.OobPowerState.off

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(sf_mc.esx_control.Vsphere, 'reset_vms')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    def test_reboot_node_via_oob(self, mock_disconnect, mock_reset, mock_connect):
        self.witness_fm.reboot_node_via_oob(wait_for_ping=False)
        mock_reset.assert_called()

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(sf_mc.esx_control.Vsphere, 'reset_vms')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    @mock.patch('valence.functional_models.Node.poll_ping_sshd_up')
    def test_reboot_node_via_oob_wait_for_ping(self, mock_disconnect, mock_reset, mock_connect,
                                               mock_poll_ping_sshd_up):
        mock_poll_ping_sshd_up.return_value = []
        self.witness_fm.reboot_node_via_oob(wait_for_ping=True)
        mock_reset.assert_called()

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(sf_mc.esx_control.Vsphere, 'power_off_vms')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    def test_power_off_via_oob(self, mock_disconnect, mock_power_off, mock_connect):
        self.witness_fm.power_off_via_oob()
        mock_power_off.assert_called()

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(sf_mc.esx_control.Vsphere, 'power_on_vms')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    def test_power_on_via_oob(self, mock_disconnect, mock_power_on, mock_connect):
        self.witness_fm.power_on_via_oob()
        mock_power_on.assert_called()
