import time
import unittest
from itertools import cycle

import mock
import sf_ssh

from sf_api import result_models
from valence import (
    enums,
    functional_models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


@mock.patch('sf_ssh.Ssh.run_command')
@mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
@mock.patch('valence.functional_models.Node.get_hardware_config_ssh')
class TestFlashNvdimmFirmware(unittest.TestCase):
    """Test Node.flash_nvdimm_firmware."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.diags = {
            '0x1000004': {
                'value': True
            },
            '0x1000104': {
                'value': True
            }
        }
        cls.smart_config = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Smart'}
            }
        ).hardware_config
        cls.micron_config = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Micron'}
            }
        ).hardware_config
        cls.ssh_conn = cls.node_fm.get_ssh_connection()

    def test_flash_smart_firmware_upgrade(self, mock_get_hardware_config_ssh,
                                          mock_get_nvdimm_diagnostics, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.smart_config
        mock_get_nvdimm_diagnostics.return_value = self.diags
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0)
        self.assertIsNone(
            self.node_fm.flash_nvdimm_firmware(downgrade=False, expected_version='26.2B',
                                               quick_downgrade=False)
        )

    def test_flash_smart_firmware_downgrade(self, mock_get_hardware_config_ssh,
                                            mock_get_nvdimm_diagnostics, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.smart_config
        mock_get_nvdimm_diagnostics.return_value = self.diags
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0)
        self.assertIsNone(
            self.node_fm.flash_nvdimm_firmware(downgrade=True, expected_version='2.11',
                                               quick_downgrade=False)
        )

    def test_flash_firmware_quick_downgrade_success(self, mock_get_hardware_config_ssh,
                                                    mock_get_nvdimm_diagnostics, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.smart_config
        mock_get_nvdimm_diagnostics.return_value = self.diags
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        self.assertIsNone(
            self.node_fm.flash_nvdimm_firmware(downgrade=True, expected_version='2.11',
                                               quick_downgrade=True)
        )
        mock_run_command.assert_any_call(cmd='smart-nvdimm --all --updateerase',
                                         raise_on_failure=True, timeout=300)
        mock_run_command.assert_any_call(cmd='smart-nvdimm --all --reloadAndArm',
                                         raise_on_failure=True, timeout=60)
        assert mock_run_command.call_count == 4

    def test_flash_firmware_quick_downgrade_success2(self, mock_get_hardware_config_ssh,
                                                     mock_get_nvdimm_diagnostics, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.smart_config
        mock_get_nvdimm_diagnostics.return_value = self.diags
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0)

        self.assertIsNone(
            self.node_fm.flash_nvdimm_firmware(downgrade=False, expected_version='2.11',
                                               quick_downgrade=True)
        )
        mock_run_command.assert_any_call(cmd='smart-nvdimm --all --updateerase',
                                         raise_on_failure=True, timeout=300)
        mock_run_command.assert_any_call(cmd='smart-nvdimm --all --reloadAndArm',
                                         raise_on_failure=True, timeout=60)
        assert mock_run_command.call_count == 4

    @mock.patch('valence.functional_models.Node.get_nvdimm_firmware_filepath')
    @mock.patch('valence.functional_models.Node.nvdimm_select_fw_slot')
    @mock.patch('valence.functional_models.Node._firmware_micron_arm_device')
    @mock.patch('valence.functional_models.Node.verify_nvdimm_armed')
    def test_flash_firmware_micron_success(
        self, mock_verify_nvdimm_armed, mock_micron_firmware_arm_device, mock_nvdimm_select_fw_slot,
        mock_get_nvdimm_firmware_filepath, mock_get_hardware_config_ssh, mock_get_nvdimm_diagnostics,
        mock_run_command
    ):
        mock_verify_nvdimm_armed.return_value = None
        mock_micron_firmware_arm_device.return_value = None
        mock_nvdimm_select_fw_slot.return_value = None
        mock_get_nvdimm_firmware_filepath.return_value = \
            '/sf/packages/nvram-micron-firmware-solidfire-20181114/lib/firmware/P_V26_PSOC.img'
        mock_get_hardware_config_ssh.return_value = self.micron_config
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '24.6C'},
                                                    '1': {'activeFirmwareVersion': '24.6C'}}

        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=0)

        self.assertIsNone(self.node_fm.flash_nvdimm_firmware(expected_version='26.2C'))


@mock.patch('valence.functional_models.Node._firmware_micron_get_arm_timeout', autospec=True)
@mock.patch('valence.functional_models.Node._firmware_micron_check_ready', autospec=True)
@mock.patch('valence.functional_models.Node._nvdimm_regwrite')
class TestFirmwareMicronArmDevice(unittest.TestCase):
    """Test Node._firmware_micron_arm_device."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_firmware_micron_arm_device(self, mock_nvdimm_regwrite, mock_fmcr, mock_fmgat):
        mock_nvdimm_regwrite.return_value = None
        mock_fmcr.return_value = None
        mock_fmgat.return_value = 1

        self.assertIsNone(self.node_fm._firmware_micron_arm_device('0'))


@mock.patch('valence.functional_models.Node._nvdimm_regread')
class TestFirmwareMicronGetArmTimeout(unittest.TestCase):
    """Test Node._firmware_micron_get_arm_timeout."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_firmware_micron_get_arm_timeout(self, mock_nvdimm_regread):
        mock_nvdimm_regread.side_effect = ['0x4b', '0x03']

        self.assertEqual(self.node_fm._firmware_micron_get_arm_timeout('0'), 0x034b)


@mock.patch('valence.functional_models.Node._nvdimm_regread')
class TestFirmwareMicronCheckReady(unittest.TestCase):
    """Test Node._firmware_micron_check_ready."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_firmware_micron_check_ready_success(self, mock_nvdimm_regread):
        mock_nvdimm_regread.side_effect = ['0xA5', '0x00']

        self.assertIsNone(self.node_fm._firmware_micron_check_ready('0', '0x61', '0x00'))

    def test_firmware_micron_check_ready_never_ready(self, mock_nvdimm_regread):
        mock_nvdimm_regread.side_effect = cycle(['0xA4', '0x00'])

        self.assertRaises(
            _errors.FlashNvdimmFirmwareError,
            self.node_fm._firmware_micron_check_ready, nvid='0', reg='0x61', value='0x00'
        )

    def test_firmware_micron_check_ready_bad_val(self, mock_nvdimm_regread):
        mock_nvdimm_regread.side_effect = cycle(['0xA5', '0x01'])

        self.assertRaises(
            _errors.FlashNvdimmFirmwareError,
            self.node_fm._firmware_micron_check_ready, nvid='0', reg='0x61', value='0x00'
        )


@mock.patch('sf_ssh.Ssh.run_command')
class TestNvdimmRegreadRegwrite(unittest.TestCase):
    """Test Node._nvdimm_regread/regwrite."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.ssh_conn = cls.node_fm.get_ssh_connection()

    def test_nvdimm_regread(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='0xA5', stderr='', return_code=0
        )

        self.assertEqual(self.node_fm._nvdimm_regread('0', '0', '0x60'), '0xA5')

    def test_nvdimm_regwrite(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0
        )

        self.assertIsNone(self.node_fm._nvdimm_regwrite('0', '3', '0x42', '1'))
        mock_run_command.assert_called()


@mock.patch('valence.functional_models.Node._nvdimm_regwrite')
class TestNvdimmSelectFwSlot(unittest.TestCase):
    """Test Node.nvdimm_select_fw_slot."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_nvdimm_select_fw_slot(self, mock_nvdimm_regwrite):
        mock_nvdimm_regwrite.return_value = None

        self.assertIsNone(self.node_fm.nvdimm_select_fw_slot('0', 1))
        mock_nvdimm_regwrite.assert_called()


@mock.patch('sf_ssh.Ssh.run_command')
@mock.patch('valence.functional_models.Node.get_hardware_config_ssh')
class TestGetNvdimmFirmwareFilepath(unittest.TestCase):
    """Test Node.get_firmware_filepath."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.ssh_conn = cls.node_fm.get_ssh_connection()
        cls.smart_config = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Smart'}
            }
        ).hardware_config
        cls.micron_config_first_gen = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Micron'}
            }
        ).hardware_config
        cls.micron_config_second_gen = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Micron_Gen2'}
            }
        ).hardware_config

    def test_get_smart_firmware_filepath_true(self, mock_get_hardware_config_ssh, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.smart_config
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='/sf/rtfi/firmware/nvram/smart/2.11.bin', stderr='', return_code=0
        )
        self.assertEqual(
            self.node_fm.get_nvdimm_firmware_filepath(expected_version='2.11'),
            '/sf/rtfi/firmware/nvram/smart/2.11.bin'
        )

    def test_get_smart_firmware_filepath_false(self, mock_get_hardware_config_ssh, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.smart_config
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='No matches found',
                                                                    return_code=1)

        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.get_nvdimm_firmware_filepath,
                          expected_version='2.11')

    @mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
    def test_get_micron_firmware_filepath_first_gen(self, mock_get_nvdimm_diagnostics,
                                                    mock_get_hardware_config_ssh, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.micron_config_first_gen
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='/mock/firmware/file/path.img',
            stderr='', return_code=0
        )

        # v22 -> v23: PSOC
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '22.8C'}}
        self.node_fm.get_nvdimm_firmware_filepath(expected_version='23')
        assert 'P_V23_PSOC.img' in mock_run_command.call_args[0][0]

        # v24 -> v26.2C: PSOC
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '24.6C'}}
        firmware_file_path = self.node_fm.get_nvdimm_firmware_filepath(expected_version='26.2C')
        assert 'P_V26_PSOC.img' in mock_run_command.call_args[0][0]
        assert firmware_file_path == '/mock/firmware/file/path.img'

        # v24 -> v25: ALL
        self.node_fm.get_nvdimm_firmware_filepath(expected_version='25')
        assert 'P_V25_ALL.img' in mock_run_command.call_args[0][0]

        # v22/v24 -> v26: ALL
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '22.8C'},
                                                    '1': {'activeFirmwareVersion': '24.6C'}}
        self.node_fm.get_nvdimm_firmware_filepath(expected_version='26')
        assert 'P_V26_ALL.img' in mock_run_command.call_args[0][0]

    @mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
    def test_get_micron_firmware_filepath_second_gen(self, mock_get_nvdimm_diagnostics,
                                                     mock_get_hardware_config_ssh, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.micron_config_second_gen
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='/mock/firmware/file/path.img',
            stderr='', return_code=0
        )

        # v20 -> v20: PSOC
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '20.2C'}}
        firmware_file_path = self.node_fm.get_nvdimm_firmware_filepath(expected_version='20')
        assert 'V2.0*-UPGRADE_PSOC-signed.img' in mock_run_command.call_args[0][0]
        assert firmware_file_path == '/mock/firmware/file/path.img'

        # v22 -> v25.2C: PSOC
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '22.7C'}}
        self.node_fm.get_nvdimm_firmware_filepath(expected_version='25.2C')
        assert 'V2.5*-UPGRADE_PSOC-signed.img' in mock_run_command.call_args[0][0]

        # v22 -> v24: ALL
        self.node_fm.get_nvdimm_firmware_filepath(expected_version='24')
        assert 'V2.4*-UPGRADE_ALL-signed.img' in mock_run_command.call_args[0][0]

        # v20/v22 -> v25: ALL
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '20.6C'},
                                                    '1': {'activeFirmwareVersion': '22.7C'}}
        self.node_fm.get_nvdimm_firmware_filepath(expected_version='25')
        assert 'V2.5*-UPGRADE_ALL-signed.img' in mock_run_command.call_args[0][0]

    @mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
    def test_get_micron_firmware_filepath_first_gen_false(self, mock_get_nvdimm_diagnostics,
                                                          mock_get_hardware_config_ssh,
                                                          mock_run_command):
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '99'},
                                                    '1': {'activeFirmwareVersion': '99'}}
        mock_get_hardware_config_ssh.return_value = self.micron_config_first_gen
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='No matches found',
                                                                    return_code=1)
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.get_nvdimm_firmware_filepath,
                          expected_version='100')

    @mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
    def test_get_micron_firmware_filepath_second_gen_false(self, mock_get_nvdimm_diagnostics,
                                                           mock_get_hardware_config_ssh,
                                                           mock_run_command):
        mock_get_nvdimm_diagnostics.return_value = {'0': {'activeFirmwareVersion': '99'},
                                                    '1': {'activeFirmwareVersion': '99'}}
        mock_get_hardware_config_ssh.return_value = self.micron_config_second_gen
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='No matches found',
                                                                    return_code=1)
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.get_nvdimm_firmware_filepath,
                          expected_version='100')


@mock.patch('sf_ssh.Ssh.run_command', autospec=True)
@mock.patch('valence.functional_models.Node.get_hardware_config_ssh')
class TestGetNvdimmDiagnostics(unittest.TestCase):
    """Test Node.get_nvdimm_diagnostics."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.diags_expected = {
                              '0x1000004': {
                                  u'activeFirmwareSlot': 1,
                                  u'activeFirmwareVersion': u'2.12',
                                  u'armStatusAbortError': False,
                                  u'armStatusAbortSuccess': False,
                                  u'slot0FirmwareVersion': u'6.12',
                                  u'slot1FirmwareVersion': u'2.12'
                               },
                              '0x1000104': {
                                  u'activeFirmwareSlot': 1,
                                  u'activeFirmwareVersion': u'2.12',
                                  u'armStatusAbortError': False,
                                  u'armStatusAbortSuccess': False,
                                  u'slot0FirmwareVersion': u'6.12',
                                  u'slot1FirmwareVersion': u'2.12'
                               }
                             }
        cls.hardware_config = result_models.GetHardwareConfig({
                                  'hardwareConfig': {
                                      'nvramVendor': 'Smart',
                                      'nvids': [
                                          '0x1000004',
                                          '0x1000104'
                                      ]
                                  }
                              }).hardware_config
        cls.json_stdout = ('[{"activeFirmwareSlot":1,"activeFirmwareVersion":"26.2C",'
                           '"armStatusAbortError":false,"armStatusAbortSuccess":false,'
                           '"slot0FirmwareVersion":"24.6C","slot1FirmwareVersion":"26.2C"},'
                           '{"activeFirmwareSlot":1,"activeFirmwareVersion":"26.2C",'
                           '"armStatusAbortError":false,"armStatusAbortSuccess":false,'
                           '"slot0FirmwareVersion":"24.6C","slot1FirmwareVersion":"26.2C"}]')
        cls.ssh_conn = cls.node_fm.get_ssh_connection()

    def test_get_nvdimm_diagnostics(self, mock_get_hardware_config_ssh, mock_run_command):
        mock_get_hardware_config_ssh.return_value = self.hardware_config
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout=self.json_stdout,
                                                                    stderr='', return_code=0)
        self.assertEqual(self.node_fm.get_nvdimm_diagnostics(), self.diags_expected)


@mock.patch('valence.functional_models.Node.verify_nvdimm_no_ecc_errors', autospec=True)
@mock.patch('valence.functional_models.Node.verify_nvdimm_sufficient_csave_energy', autospec=True)
@mock.patch('valence.functional_models.Node.verify_nvdimm_fw_version', autospec=True)
@mock.patch('valence.functional_models.Node.verify_nvdimm_armed', autospec=True)
class TestVerifyNvdimmHealth(unittest.TestCase):
    """Test Node.verify_nvdimm_health."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.diags = {'0': True, '1': True}

    def test_verify_nvdimm_health_true(self, mock_vna, mock_vnfv, mock_vnsce, mock_vnnee):
        mock_vna.return_value = None
        mock_vnfv.return_value = None
        mock_vnsce.return_value = None
        mock_vnnee.return_value = None
        self.assertIsNone(self.node_fm.verify_nvdimm_health(nvdimm_diags=self.diags))


@mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
class TestVerifyNvdimmArmed(unittest.TestCase):
    """Test Node.verify_nvdimm_armed."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.diags1 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True
            }
        }
        cls.diags2 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True
            }
        }
        cls.diags3 = {
            '0x1000004': {
                'armStatusArmSuccess': True
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True
            }
        }
        cls.diags4 = {
            '0x1000004': {
                'armStatusArmSuccess': False,
                'armStatusSaveNArmed': False

            },
            '0x1000104': {
                'armStatusArmSuccess': False,
                'armStatusSaveNArmed': False
            }
        }

    def test_verify_nvdimm_armed_true(self, mock_get):
        mock_get.return_value = self.diags1
        self.assertIsNone(self.node_fm.verify_nvdimm_armed())

    def test_verify_nvdimm_armed_false_missing_nvdimm(self, mock_get):
        mock_get.return_value = self.diags2
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.verify_nvdimm_armed)

    def test_verify_nvdimm_armed_false_incomplete_diags(self, mock_get):
        mock_get.return_value = self.diags3
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.verify_nvdimm_armed)

    def test_verify_nvdimm_armed_false_not_armed(self, mock_get):
        mock_get.return_value = self.diags4
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.verify_nvdimm_armed)


@mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
class TestVerifyNvdimmSufficientCsaveEnergy(unittest.TestCase):
    """Test Node.verify_nvdimm_no_ecc_errors."""
    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.diags1 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': False
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': False
            }
        }
        cls.diags2 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True
            }
        }

    def test_verify_nvdimm_sufficient_csave_energy_true(self, mock_get):
        mock_get.return_value = self.diags1
        self.assertIsNone(self.node_fm.verify_nvdimm_sufficient_csave_energy())

    def test_verify_nvdimm_sufficient_csave_energy_false(self, mock_get):
        mock_get.return_value = self.diags2
        self.assertRaises(_errors.VerifyNvdimmFailedError,
                          self.node_fm.verify_nvdimm_sufficient_csave_energy)


@mock.patch('sf_ssh.Ssh.run_command')
class TestVerifyNvdimmNoEccErrors(unittest.TestCase):
    """Test Node.verify_nvdimm_no_ecc_errors."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_verify_nvdimm_no_ecc_errors_true(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='',
                                                                    return_code=0)
        self.assertIsNone(self.node_fm.verify_nvdimm_no_ecc_errors())

    def test_verify_nvdimm_no_ecc_errors_false(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='Assert', stderr='',
                                                                    return_code=1)
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.verify_nvdimm_no_ecc_errors)


@mock.patch('valence.functional_models.Node.get_nvdimm_diagnostics')
class TestVerifyNvdimmFwVersion(unittest.TestCase):
    """Test Node.verify_nvdimm_fw_version."""
    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.diags1 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True,
                'activeFirmwareVersion': '2.12'
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True,
                'activeFirmwareVersion': '2.12'
            }
        }
        cls.diags2 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True,
                'activeFirmwareVersion': '2.12'
            }
        }
        cls.diags3 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True,
                'activeFirmwareVersion': '2.11'
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True,
                'activeFirmwareVersion': '2.11'
            }
        }
        cls.diags4 = {
            '0x1000004': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True
            },
            '0x1000104': {
                'armStatusArmSuccess': True,
                'armStatusSaveNArmed': True,
                'notEnoughEnergyForCatastrophicSave': True
            }
        }

    def test_verify_nvdimm_fw_version_true(self, mock_get):
        mock_get.return_value = self.diags1
        self.assertIsNone(self.node_fm.verify_nvdimm_fw_version(expected_version='2.12'))

    def test_verify_nvdimm_fw_version_false_one_nvdimm(self, mock_get):
        mock_get.return_value = self.diags2
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.verify_nvdimm_fw_version,
                          expected_version='2.12')

    def test_verify_nvdimm_fw_version_false_fw_mismatch(self, mock_get):
        mock_get.return_value = self.diags3
        self.assertRaises(_errors.VerifyNvdimmFailedError, self.node_fm.verify_nvdimm_fw_version,
                          expected_version='2.12')

    def test_verify_nvdimm_fw_version_false_active_fw_non_existent(self, mock_get):
        mock_get.return_value = self.diags4
        self.assertRaises(KeyError, self.node_fm.verify_nvdimm_fw_version,
                          expected_version='2.12')


@mock.patch('sf_ssh.Ssh.run_command', autospec=True)
@mock.patch('valence.functional_models.Node.disable_binary_ssh', autospec=True)
@mock.patch('valence.functional_models.Node.stop_binary_ssh', autospec=True)
class TestUmountNvram(unittest.TestCase):
    """Test Node.umount_nvram"""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_umount_nvram(self, mock_stop_binary_ssh, mock_disable_binary_ssh, mock_run_command):
        mock_stop_binary_ssh.return_value = None
        mock_disable_binary_ssh.return_value = None
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='successfully unmounted /mnt/pendingDirtyBlocks', stderr='', return_code=0)
        self.node_fm.umount_nvram()
        mock_run_command.assert_called()


@mock.patch('sf_ssh.Ssh.run_command', autospec=True)
@mock.patch('valence.functional_models.Node.get_cache_device_path', autospec=True)
@mock.patch('valence.functional_models.Node.enable_binary_ssh', autospec=True)
@mock.patch('valence.functional_models.Node.start_binary_ssh', autospec=True)
class TestMkfsNvram(unittest.TestCase):
    """Test Node.mkfs_nvram"""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_mkfs_nvram(self, mock_start_binary_ssh, mock_enable_binary_ssh,
                        mock_get_cache_device_path, mock_run_command):
        mock_start_binary_ssh.side_effect = IndexError
        mock_enable_binary_ssh.return_value = None
        mock_get_cache_device_path.return_value = '/dev/pmem0p1'
        mock_run_command.return_value = [
            sf_ssh.ssh.RunCommandOutput(stdout='successfully created filesystem',
                                        stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='successfully mounted /mnt/pendingDirtyBlocks',
                                        stderr='', return_code=0)
        ]
        self.node_fm.mkfs_nvram()
        mock_get_cache_device_path.assert_called()
        self.assertEqual(mock_run_command.call_count, 2)


@mock.patch('sf_ssh.Ssh.run_command', autospec=True)
@mock.patch('valence.functional_models.Node.get_cache_device_path', autospec=True)
class TestFioWriteVerifyPattern(unittest.TestCase):
    """Test Node.write_validation_data."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.pattern = '123abc'

    def test_fio_write_pattern(self, mock_get_cache_device_path, mock_run_command):
        mock_run_command.return_code = 0
        mock_get_cache_device_path.return_value = '/dev/pmem0p1'
        self.assertIsNone(self.node_fm.fio_write_pattern(pattern=self.pattern))
        mock_get_cache_device_path.assert_called()

    def test_fio_verify_pattern(self, mock_get_cache_device_path, mock_run_command):
        mock_run_command.return_code = 0
        mock_get_cache_device_path.return_value = '/dev/pmem0p1'
        self.assertIsNone(self.node_fm.fio_verify_pattern(pattern=self.pattern))
        mock_get_cache_device_path.assert_called()


@mock.patch('valence.functional_models.Node.get_hardware_config_ssh', autospec=True)
class TestGetCacheDevicePath(unittest.TestCase):
    """Test Node.get_cache_device_path."""

    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_get_cache_device_path_pmem(self, mock_get_hw_cfg):
        mock_get_hw_cfg.return_value = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Smart'}
            }
        ).hardware_config
        result = self.node_fm.get_cache_device_path()
        self.assertEqual(result, '/dev/pmem0p1')
        mock_get_hw_cfg.assert_called()

    def test_get_cache_device_path_nvme(self, mock_get_hw_cfg):
        mock_get_hw_cfg.return_value = result_models.GetHardwareConfig(
            {
                'hardwareConfig': {'nvramVendor': 'Radian'}
            }
        ).hardware_config
        result = self.node_fm.get_cache_device_path()
        self.assertEqual(result, '/dev/nvme0n1')
        mock_get_hw_cfg.assert_called()
