import time
import unittest

import mock
import pytest
import sf_ssh

from valence import (
    functional_models,
    unittest_utils,
)
from valence.functional_models import _errors

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestNodeOtpw(unittest.TestCase):
    """Test methods related to OTPW."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)
        cls.run_command_pass = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        cls.run_command_fail = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='error', return_code=1)

    @mock.patch('sf_ssh.Ssh.run_command')
    @mock.patch('valence.functional_models.Node.verify_otpw_is_enabled')
    @mock.patch('valence.functional_models.Node.restore_otpw_list')
    def test_enable_otpw(self, mock_restore_otpw, mock_verify_otpw, mock_run_command):
        mock_run_command.return_value = self.run_command_pass
        self.node_fm.enable_otpw()
        assert mock_restore_otpw.called
        assert mock_verify_otpw.called

    @mock.patch('sf_ssh.Ssh.run_command')
    @mock.patch('valence.functional_models.Node.verify_otpw_is_enabled')
    @mock.patch('valence.functional_models.Node.restore_otpw_list')
    def test_enable_otpw_failed(self, mock_restore_otpw, mock_verify_otpw, mock_run_command):
        mock_run_command.return_value = self.run_command_fail
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.enable_otpw()

    @mock.patch('sf_ssh.Ssh.run_command')
    @mock.patch('valence.functional_models.Node.verify_otpw_is_enabled')
    @mock.patch('valence.functional_models.Node.restore_otpw_list')
    def test_enable_otpw_retry(self, mock_restore_otpw, mock_verify_otpw, mock_run_command):
        mock_run_command.return_value = self.run_command_pass
        mock_restore_otpw.side_effect = [sf_ssh.errors.SshError, None]
        self.node_fm.enable_otpw()
        assert mock_restore_otpw.call_count > 1

    @mock.patch('sf_ssh.Ssh.run_command')
    @mock.patch('valence.functional_models.Node.verify_otpw_is_not_enabled')
    def test_disable_otpw(self, mock_verify_not_otpw, mock_run_command):
        mock_run_command.return_value = self.run_command_pass
        self.node_fm.disable_otpw()
        assert mock_verify_not_otpw.called

    @mock.patch('sf_ssh.Ssh.run_command')
    @mock.patch('valence.functional_models.Node.verify_otpw_is_not_enabled')
    def test_disable_otpw_retry(self, mock_verify_not_otpw, mock_run_command):
        mock_run_command.return_value = self.run_command_pass
        mock_verify_not_otpw.side_effect = [sf_ssh.errors.SshError, None]
        self.node_fm.disable_otpw()
        assert mock_verify_not_otpw.call_count > 1

    @mock.patch('sf_ssh.Ssh.run_command')
    @mock.patch('valence.functional_models.Node.verify_otpw_is_not_enabled')
    def test_disable_otpw_failed(self, mock_verify_not_otpw, mock_run_command):
        mock_run_command.return_value = self.run_command_fail
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.disable_otpw()

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_restore_otpw_list(self, mock_run_command):
        mock_run_command.return_value = self.run_command_pass
        self.node_fm.restore_otpw_list()

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_restore_otpw_list_first_try_fails(self, mock_run_command):
        mock_run_command.side_effect = [self.run_command_fail, self.run_command_pass]
        self.node_fm.restore_otpw_list()

    @mock.patch('sf_ssh.Ssh.run_command')
    def test_restore_otpw_fail(self, mock_run_command):
        mock_run_command.return_value = self.run_command_fail
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.restore_otpw_list()

    @mock.patch('sf_ssh.Ssh')
    def test_verify_otpw(self, mock_ssh):
        inst = mock_ssh.return_value.__enter__.return_value
        inst.auth_method = sf_ssh.AuthOtpw('filename', 'username')
        self.node_fm.verify_otpw_is_enabled()
        inst.auth_method = sf_ssh.AuthInteractivePassword('username', 'password')
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.verify_otpw_is_enabled()

    @mock.patch('sf_ssh.Ssh')
    def test_verify_not_otpw(self, mock_ssh):
        inst = mock_ssh.return_value.__enter__.return_value
        inst.auth_method = sf_ssh.AuthInteractivePassword('username', 'password')
        self.node_fm.verify_otpw_is_not_enabled()
        inst.auth_method = sf_ssh.AuthOtpw('filename', 'username')
        with pytest.raises(_errors.FunctionalNodeError):
            self.node_fm.verify_otpw_is_not_enabled()
