import time
import unittest

import mock
import sf_ssh

import sf_api
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


@mock.patch('sf_ssh.Ssh.run_command')
class TestStartStopBinarySsh(unittest.TestCase):
    """Test methods that start/stop services with ssh."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_start_binary_ssh(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='sfconfig start/running, process 28233\n', stderr='', return_code=0)
        pid = self.node_fm.start_binary_ssh(enums.NodeBinary.sfconfig)
        assert pid == 28233
        mock_run_command.assert_called()

    def test_stop_binary_ssh(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout=r'\sfconfig start/running, process 28233\n', stderr='', return_code=0)
        self.node_fm.stop_binary_ssh(enums.NodeBinary.solidfire)
        mock_run_command.assert_called()


@mock.patch('sf_ssh.Ssh.run_command')
class TestKillServiceSsh(unittest.TestCase):
    """Test method that kills service over ssh."""
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    @mock.patch.object(functional_models.Node, '_get_pid')
    def test_kill_master_service_ssh(self, mock_get_pid, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0)
        mock_get_pid.return_value = 1
        self.node_fm.kill_master_service_ssh(enums.UnixSignal.SIGKILL)
        mock_get_pid.assert_called()
        cmd = 'kill -{} {}'.format(enums.UnixSignal.SIGKILL.value, 1)
        mock_run_command.assert_called_with(cmd, raise_on_failure=True)

    @mock.patch.object(functional_models.Node, '_get_pid')
    def test_kill_drive_service_ssh(self, mock_get_pid, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0)
        mock_get_pid.return_value = 1
        service = models.ClusterService.get_mock_object(overrides={
            'service_type': enums.ServiceType.block, 'service_id': 1})
        service_dm = models.ClusterServiceInfo.get_mock_object(overrides={'service': service})
        self.node_fm.kill_drive_service_ssh(service_dm, enums.UnixSignal.SIGKILL)
        cmd = 'kill -{} {}'.format(enums.UnixSignal.SIGKILL.value, 1)
        mock_get_pid.assert_called()
        mock_run_command.assert_called_with(cmd, raise_on_failure=True)

    def test_get_pid(self, mock_run_command):
        cmd = 'ps aux | grep \"service={}\"'.format(enums.NodeService.master.value)
        stdout = 'root      3854  1.2  0.1 2697992 78380 ? '
        run_command_output = sf_ssh.ssh.RunCommandOutput(stdout, '', 0)
        mock_run_command.return_value = run_command_output
        self.node_fm._get_pid(enums.NodeService.master)
        mock_run_command.assert_called_with(cmd, raise_on_failure=True)
        self.node_fm._get_pid(enums.NodeService.block, 1)
        cmd = 'ps aux | grep \"service={}.*-serviceid={}\"'.format(enums.NodeService.block.value, 1)
        mock_run_command.assert_called_with(cmd, raise_on_failure=True)


class TestStartStopBinary(unittest.TestCase):
    """Test methods that use RestartServices."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    @mock.patch.object(functional_models.Node, '_restart_binary')
    def test_start_binary(self, mock_restart_binary):
        self.node_fm.start_binary(enums.NodeBinary.sfconfig)
        mock_restart_binary.assert_called_with(enums.NodeBinary.sfconfig,
                                               action=sf_api.RestartServicesAction.start)

    @mock.patch.object(functional_models.Node, '_restart_binary')
    def test_stop_binary(self, mock_restart_binary):
        self.node_fm.stop_binary(enums.NodeBinary.solidfire)
        mock_restart_binary.assert_called_with(enums.NodeBinary.solidfire,
                                               action=sf_api.RestartServicesAction.stop)

    @mock.patch.object(functional_models.Node, '_restart_binary')
    def test_restart_binary(self, mock_restart_binary):
        self.node_fm.restart_binary(enums.NodeBinary.solidfire)
        mock_restart_binary.assert_called_with(enums.NodeBinary.solidfire,
                                               action=sf_api.RestartServicesAction.restart)

    @mock.patch.object(sf_api.NodeApi, 'call_method')
    def test_private_restart_binary(self, mock_call_method):
        self.node_fm._restart_binary(enums.NodeBinary.solidfire, enums.RestartServicesAction.stop)
        mock_call_method.assert_called()


@mock.patch('sf_ssh.Ssh.run_command')
class TestSEnableDisableBinarySsh(unittest.TestCase):
    """Test methods that start/stop services with ssh."""

    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_enable_binary_ssh(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='sfconfig enabled\n', stderr='', return_code=0)
        self.node_fm.enable_binary_ssh(enums.NodeBinary.sfconfig)
        mock_run_command.assert_called()

    def test_disable_binary_ssh(self, mock_run_command):
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='sfconfig disabled\n', stderr='', return_code=0)
        self.node_fm.disable_binary_ssh(enums.NodeBinary.solidfire)
        mock_run_command.assert_called()
