"""Decorators for the backend monitor."""
import datetime
import functools
import sys

import pytz


class GlobalMonitor(object):
    """Class to hold the global backend monitor instance."""
    BACKEND_MONITOR = None


class MethodCall(object):
    """Data class to hold information about a functional model call.

    :param Any method: Pointer to a bound method object
    :param dict[str,Any] arguments: The arguments the method was called with
    :param datetime.datetime timestamp: Datetime object for when the method was called
    """
    def __init__(self, method, arguments, timestamp):

        self.method = self._get_unbound_method(method)
        self.arguments = arguments
        self.timestamp = timestamp
        # This will be set by observers when registering a timeout that is associated with the call
        self.timeout_id = None

    @property
    def static_id(self):
        """Return the static ID of the associated functional model.

        :rtype: str
        """
        return self.arguments['self'].static_id

    @staticmethod
    def _get_unbound_method(method):
        """Given a 'bound' method, return the unbound method.

        This is needed so comparisons like the one below are possible.

        .. code-block:: python

            method_call.method == valence.functional_models.Cluster.remove_nodes

        :param Any method: Bound method
        :rtype: types.MethodType
        """
        # In python 3 there is no such thing as an unbound method, it was removed to just be a function.
        # The class method im_func is also renamed to __func__.
        if sys.version_info >= (3, 0):
            return method.__func__  # pragma: no cover
        else:
            # This gets a reference to the class, and calls getattr to get a reference to the unbound
            # method. For example, if method == MyClass().do_it, method.im_self.__class__ == MyClass,
            # and the unbound method == MyClass.do_it.
            return getattr(method.im_self.__class__, method.im_func.func_name)  # pragma: no cover

    def is_method(self, method):
        """Helper method to compare an unbound method to the MethodCall's method.

        :param types.MethodType method: Method to compare, e.g. functional_models.Cluster.add_drives
        :rtype: bool
        """
        return method == self.method

    def __repr__(self):
        return 'MethodCall<name={}>'.format(self.method.__name__)


def set_backend_monitor(monitor):
    """Function to set the backend monitor for the decorator.

    Since we don't pass the backend monitor to the functional models, this sets a global instance of the
    backend monitor needed for the notify decorator.

    :param valence.monitors.BackendMonitor monitor: The backend monitor instance.
    """
    GlobalMonitor.BACKEND_MONITOR = monitor


def get_backend_monitor():
    """Returns the global monitor.

    :rtype: valence.monitors.BackendMonitor
    """
    return GlobalMonitor.BACKEND_MONITOR


def remove_backend_monitor():
    """Function to remove the global backend monitor.

    For cleanup purposes. Useful for creating a clean state and in unit tests.
    """
    GlobalMonitor.BACKEND_MONITOR = None


def notify_monitor(method):
    """Decorator to notify the backend monitor of method calls.

    The wrapper logic below will extract all the method's parameters as kwargs, regardless of how the
    decorated method was called. This is so each Observer won't need need to have logic to handle
    args vs kwargs.

    :param Any method: The functional model method.
    :rtype: typing.Callable
    """

    # This decorator is needed so that the wrapper looks like the wrapped function. Without this, when
    # you reference functional_models.Cluster.add_drives, it would return the wrapper function rather
    # that add_drives.
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        # If a BackgroundMonitor instance does not exist, skip decorator
        if not GlobalMonitor.BACKEND_MONITOR or not GlobalMonitor.BACKEND_MONITOR.is_running():
            return method(*args, **kwargs)

        # Get the function's arguments
        arg_names = method.__code__.co_varnames[:method.__code__.co_argcount]
        wrapper_args = args[:len(arg_names)]

        # Wrap args with their default values
        defaults = method.__defaults__ or ()
        required_params_num = len(arg_names)-len(defaults)-1
        if required_params_num:
            values_to_override = [None]*required_params_num
            defaults = tuple(values_to_override) + defaults
        wrapper_args = wrapper_args + defaults[len(defaults) -
                                               (method.__code__.co_argcount - len(args)):]

        # Make the params dictionary and update with any kwargs
        params = dict(zip(arg_names, wrapper_args))
        params.update(kwargs)
        extra_args = args[len(arg_names):]

        # If extra undefined positional args were passed (*args) then add them to the params dict
        if extra_args:
            params.update({'args': args})

        # Instantiate a MethodCall
        bound_method = getattr(params['self'], method.__name__)
        method_call = MethodCall(bound_method, params, datetime.datetime.now(pytz.utc))

        # Notify the backend monitor of the method call, then call the method
        GlobalMonitor.BACKEND_MONITOR.notify(method_call)
        return method(*args, **kwargs)
    return wrapper
