import time
import unittest

import mock
import pytest
import sf_rtfi

import sf_api
from sf_api import (
    api_constants,
    result_models,
    sf_version,
)
from valence import (
    enums,
    errors,
    functional_models,
    models,
)


@mock.patch.object(functional_models.Cluster, 'get_cluster_info')
class TestEarHandlers(unittest.TestCase):
    """Test the encryption at rest async handlers."""
    def setUp(self):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()
        self.cluster_dm = models.Cluster.get_mock_object()

    def test_enable_ear_async_handler_running(self, mock_get_cluster_info):
        self.cluster_dm.encryption_at_rest_state = sf_api.EarState.enabling
        mock_get_cluster_info.return_value = self.cluster_dm
        async_handler = functional_models.EnableEarAsyncHandler(self.cluster_fm)
        assert async_handler.get_status() is enums.AsyncResultStatus.running

    def test_enable_ear_async_handler_complete(self, mock_get_cluster_info):
        self.cluster_dm.encryption_at_rest_state = sf_api.EarState.enabled
        mock_get_cluster_info.return_value = self.cluster_dm
        async_handler = functional_models.EnableEarAsyncHandler(self.cluster_fm)
        assert async_handler.get_status() is enums.AsyncResultStatus.complete

    def test_enable_ear_async_handler_failed(self, mock_get_cluster_info):
        self.cluster_dm.encryption_at_rest_state = sf_api.EarState.disabled
        mock_get_cluster_info.return_value = self.cluster_dm
        async_handler = functional_models.EnableEarAsyncHandler(self.cluster_fm)
        assert async_handler.get_status() is enums.AsyncResultStatus.failed

    def test_disable_ear_async_handler_running(self, mock_get_cluster_info):
        self.cluster_dm.encryption_at_rest_state = sf_api.EarState.disabling
        mock_get_cluster_info.return_value = self.cluster_dm
        async_handler = functional_models.DisableEarAsyncHandler(self.cluster_fm)
        assert async_handler.get_status() is enums.AsyncResultStatus.running

    def test_disable_ear_async_handler_complete(self, mock_get_cluster_info):
        self.cluster_dm.encryption_at_rest_state = sf_api.EarState.disabled
        mock_get_cluster_info.return_value = self.cluster_dm
        async_handler = functional_models.DisableEarAsyncHandler(self.cluster_fm)
        assert async_handler.get_status() is enums.AsyncResultStatus.complete

    def test_disable_ear_async_handler_failed(self, mock_get_cluster_info):
        self.cluster_dm.encryption_at_rest_state = sf_api.EarState.enabling
        mock_get_cluster_info.return_value = self.cluster_dm
        async_handler = functional_models.DisableEarAsyncHandler(self.cluster_fm)
        assert async_handler.get_status() is enums.AsyncResultStatus.failed


class TestAsyncHandle(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.async_handle_fm = functional_models.AsyncHandler(cluster_fm, async_handle_dm)
        self.async_handle_does_not_exist_error = sf_api.api_errors.AsyncHandleDoesNotExist(
            'xAsyncHandleDoesNotExist', mock.Mock())

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    def test_get_status_non_async(self, mock_get_async_result):
        self.async_handle_fm.async_handle_dm = None

        with pytest.raises(functional_models._errors.AsyncHandlerError):
            self.async_handle_fm.get_status()

    @mock.patch.object(time, 'sleep')
    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    def test_async_result_does_not_exist(self, mock_get_async_result, mock_time):
        mock_get_async_result.side_effect = self.async_handle_does_not_exist_error
        with pytest.raises(sf_api.api_errors.AsyncHandleDoesNotExist):
            self.async_handle_fm.get_async_result()

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    def test_is_running(self, mock_get_async_result):
        overrides = dict(status=enums.AsyncResultStatus.running)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        assert self.async_handle_fm.is_running()
        overrides = dict(status=enums.AsyncResultStatus.complete)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        assert not self.async_handle_fm.is_running()

    def test_base_get_models_is_none(self):
        assert self.async_handle_fm.get_models() is None

    @mock.patch.object(functional_models.AsyncHandler, 'join')
    def test_get_models_blocks(self, mock_join):
        self.async_handle_fm.get_models(blocking=True, timeout=600)
        assert mock_join.call_count == 1

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(time, 'sleep')
    def test_join_success(self, mock_sleep, mock_get_async_result):
        running = enums.AsyncResultStatus.running
        complete = enums.AsyncResultStatus.complete
        running_result = models.AsyncResult.get_mock_object(overrides=dict(status=running))
        complete_result = models.AsyncResult.get_mock_object(overrides=dict(status=complete))
        mock_get_async_result.side_effect = [running_result, complete_result]
        self.async_handle_fm.join(timeout=600)
        assert mock_get_async_result.call_count == 2

    @mock.patch.object(functional_models.AsyncHandler, 'get_async_result')
    @mock.patch.object(time, 'sleep')
    @mock.patch.object(time, 'time')
    def test_join_timeout(self, mock_time, mock_sleep, mock_get_async_result):
        """Test that a timeout of 10 min works."""
        overrides = dict(status=enums.AsyncResultStatus.running)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        mock_time.side_effect = [0, 601]
        self.async_handle_fm.timeout = 600
        with pytest.raises(errors.TimeoutError):
            self.async_handle_fm.join()

    @mock.patch.object(functional_models.AsyncHandler, 'get_async_result')
    def test_join_failure(self, mock_get_async_result):
        overrides = dict(status=enums.AsyncResultStatus.failed)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        with pytest.raises(functional_models.AsyncHandlerError):
            self.async_handle_fm.join()


class TestGarbageCollectionAsyncHandler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.async_handle_dm = models.AsyncHandle.get_mock_object()
        cls.gc_events = [
            models.ClusterEvent.create(message='GCComplete'),
            models.ClusterEvent.create(message='GCStarted', details={'eligibleBS': [0, 2]})
        ]

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_get_status(self, get_current_version, get_async_result):
        """Test that get_status returns appropriately."""
        get_current_version.return_value = sf_version.SfVersion('12.0')
        gc_handler = functional_models.GarbageCollectionAsyncHandler(
            cluster_fm=self.cluster_fm,
            async_handle_dm=self.async_handle_dm
        )
        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.running,
            details=models.GarbageCollectionDetails.create()
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.running

        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.complete,
            details=models.GarbageCollectionDetails.create()
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.complete

        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.failed,
            details=models.GarbageCollectionDetails.create()
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.failed

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_get_status_pre_mg(self, get_current_version, get_async_result, list_events):
        """Test that get_status returns appropriately for pre-Magnesium async results."""
        list_events.side_effect = [self.gc_events[:-1], self.gc_events]
        get_current_version.return_value = sf_version.SfVersion('11.0')
        with pytest.raises(errors.TimeoutError):
            gc_handler = functional_models.GarbageCollectionAsyncHandler(
                cluster_fm=self.cluster_fm,
                async_handle_dm=self.async_handle_dm,
                start_time=0.0
            )

        gc_handler = functional_models.GarbageCollectionAsyncHandler(
            cluster_fm=self.cluster_fm,
            async_handle_dm=self.async_handle_dm,
            start_time=time.time()
        )
        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.running,
            details=models.GarbageCollectionDetailsPreMg.create(
                slice_services={'0': 'Complete'},
                block_services={'1': 'Working', '2': 'Working'}
            )
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.running

        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.running,
            details=models.GarbageCollectionDetailsPreMg.create(
                slice_services={'0': 'Complete'},
                block_services={'1': 'Complete', '2': 'Complete'}
            )
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.complete

        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.failed,
            details=models.GarbageCollectionDetailsPreMg.create(
                slice_services={'0': 'Complete'},
                block_services={'1': 'Complete', '2': 'Failed'}
            )
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.failed

        # Tests for special cases that arise with GCv1
        # ---------------------------------------------------------------------
        # 1. Only working service is ineligible
        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.running,
            details=models.GarbageCollectionDetailsPreMg.create(
                slice_services={'0': 'Complete'},
                block_services={'1': 'Working', '2': 'Complete'}
            )
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.complete

        # 2. Only failed service is ineligible
        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.failed,
            details=models.GarbageCollectionDetailsPreMg.create(
                slice_services={'0': 'Complete'},
                block_services={'1': 'Failed', '2': 'Complete'}
            )
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.complete

        # 3. Only failed service is decommissioned
        get_async_result.return_value = models.AsyncResult.create(
            status=enums.AsyncResultStatus.failed,
            details=models.GarbageCollectionDetailsPreMg.create(
                slice_services={'0': 'Complete'},
                block_services={'1': 'Complete', '2': 'Failed'},
                service_details={'2': 'Service decommissioned'}
            )
        )
        assert gc_handler.get_status() is enums.AsyncResultStatus.complete


class TestAddDrivesAsyncHandler(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.drive_dm1 = models.Drive.get_mock_object()
        self.drive_dm2 = models.Drive.get_mock_object()
        self.drive_dm3 = models.Drive.get_mock_object()
        self.drives_to_add = [self.drive_dm1, self.drive_dm2]
        self.async_handler = functional_models.AddDrivesAsyncHandler(
            cluster_fm, [d.drive_id for d in self.drives_to_add], async_handle_dm)

    @mock.patch.object(functional_models.AddDrivesAsyncHandler, '_get_models')
    def test_get_fake_async_status_failed(self, mock_get_models):
        """Test that various drive status correspond to failed AsyncResult status."""
        self.drive_dm1.status = sf_api.DriveStatus.active
        self.drive_dm2.status = sf_api.DriveStatus.removed
        mock_get_models.return_value = self.drives_to_add
        assert self.async_handler._get_fake_async_status() is enums.AsyncResultStatus.failed

        self.drive_dm1.status = sf_api.DriveStatus.failed
        self.drive_dm2.status = sf_api.DriveStatus.available
        mock_get_models.return_value = self.drives_to_add
        assert self.async_handler._get_fake_async_status() is enums.AsyncResultStatus.failed

    @mock.patch.object(functional_models.AddDrivesAsyncHandler, '_get_models')
    def test_get_fake_async_status_running(self, mock_get_models):
        """Test that various drive status correspond to running AsyncResult status."""
        self.drive_dm1.status = sf_api.DriveStatus.available
        self.drive_dm2.status = sf_api.DriveStatus.active
        mock_get_models.return_value = self.drives_to_add
        assert self.async_handler._get_fake_async_status() is enums.AsyncResultStatus.running

    @mock.patch.object(functional_models.AddDrivesAsyncHandler, '_get_models')
    def test_get_fake_async_status_complete(self, mock_get_models):
        """Test that various drive status correspond to complete AsyncResult status."""
        self.drive_dm1.status = sf_api.DriveStatus.active
        self.drive_dm2.status = sf_api.DriveStatus.active
        self.drive_dm3.status = sf_api.DriveStatus.failed
        mock_get_models.return_value = self.drives_to_add
        status = self.async_handler._get_fake_async_status()
        assert status is enums.AsyncResultStatus.complete

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_models(self, mock_list_drives):
        """Test that _get_models only returns models associated with the AsyncHandler."""
        mock_list_drives.return_value = self.drives_to_add + [self.drive_dm3]
        assert self.async_handler._get_models() == self.drives_to_add


class TestAddNodesAsyncHandler(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.node_dm1 = models.PendingNode.get_mock_object()
        self.node_dm2 = models.PendingNode.get_mock_object()
        self.node_dm3 = models.PendingNode.get_mock_object()
        self.node_to_add = self.node_dm1
        self.async_handler = functional_models.AddNodesAsyncHandler(
            cluster_fm, self.node_dm1.static_id, async_handle_dm)

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(functional_models.AddNodesAsyncHandler, '_get_models')
    def test_get_fake_async_status_complete(self, mock_get_models, mock_get_async_result):
        """Test that the node status correspond to complete AsyncResult status."""
        mock_get_models.return_value = [self.node_dm1]
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(
            overrides=dict(status=sf_api.AsyncResultStatus.complete)
        )
        assert self.async_handler._get_fake_async_status() is sf_api.AsyncResultStatus.complete

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(functional_models.AddNodesAsyncHandler, '_get_models')
    def test_get_fake_async_status_running(self, mock_get_models, mock_get_async_result):
        """Test that the node status correspond to running AsyncResult status."""
        mock_get_models.return_value = []
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(
            overrides=dict(status=sf_api.AsyncResultStatus.complete)
        )
        assert self.async_handler._get_fake_async_status() is sf_api.AsyncResultStatus.running

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_get_models(self, mock_active_nodes):
        """Test that _get_models only returns models associated with the AsyncHandler."""
        mock_active_nodes.return_value = [self.node_to_add, self.node_dm3, self.node_dm2]
        assert self.async_handler._get_models() == [self.node_to_add]

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    def test_real_async_shortcut(self, mock_get_async_result):
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(
            overrides=dict(status=sf_api.AsyncResultStatus.running)
        )
        assert self.async_handler.get_status() is sf_api.AsyncResultStatus.running


class TestRemoveDrivesAsyncHandler(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        self.drive_dm1 = models.Drive.get_mock_object(overrides=dict(drive_id=1))
        self.drive_dm2 = models.Drive.get_mock_object(overrides=dict(drive_id=2))
        self.remove_drives_handler = functional_models.RemoveDrivesAsyncHandler(
            cluster_fm, [self.drive_dm1.drive_id])

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_models(self, mock_list_drives):
        """Test that _get_models only returns models associated with the AsyncHandler."""
        mock_list_drives.return_value = [self.drive_dm1, self.drive_dm2]
        assert len(self.remove_drives_handler._get_models()) == 1


class TestCloneVolumeAsyncHandler(unittest.TestCase):
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_get_models(self, mock_list_volumes):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        async_handle_dm = models.AsyncHandle.get_mock_object()
        async_handler_fm = functional_models.CloneVolumeAsyncHandler(cluster_fm, 42, async_handle_dm)
        volume_dm = models.Volume.get_mock_object()
        mock_list_volumes.return_value = [volume_dm]
        assert async_handler_fm.get_models() == [volume_dm]


class TestRemoveNodesAsyncHandler(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.node_dm1 = models.PendingNode.get_mock_object()
        self.node_dm2 = models.PendingNode.get_mock_object()
        self.node_to_be_removed = self.node_dm1
        self.async_handler = functional_models.RemoveNodesAsyncHandler(
            cluster_fm, self.node_dm1.static_id, async_handle_dm)

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(functional_models.RemoveNodesAsyncHandler, '_get_models')
    def test_get_fake_async_status_complete(self, mock_get_models, mock_get_async_result):
        """Test that the node status correspond to complete AsyncResult status."""
        mock_get_models.return_value = [self.node_dm1]
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(
            overrides=dict(status=sf_api.AsyncResultStatus.complete)
        )
        assert self.async_handler._get_fake_async_status() is sf_api.AsyncResultStatus.complete

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(functional_models.RemoveNodesAsyncHandler, '_get_models')
    def test_get_fake_async_status_running(self, mock_get_models, mock_get_async_result):
        """Test that the node status correspond to running AsyncResult status."""
        mock_get_models.return_value = []
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(
            overrides=dict(status=sf_api.AsyncResultStatus.complete)
        )
        assert self.async_handler._get_fake_async_status() is sf_api.AsyncResultStatus.running

    @mock.patch.object(functional_models.Cluster, 'list_pending_nodes')
    def test_get_models(self, mock_pending_nodes):
        """Test that _get_models only returns models associated with the AsyncHandler."""
        mock_pending_nodes.return_value = [self.node_to_be_removed, self.node_dm2]
        assert self.async_handler._get_models() == [self.node_to_be_removed]

    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    def test_real_async_shortcut(self, mock_get_async_result):
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(
            overrides=dict(status=sf_api.AsyncResultStatus.running)
        )
        assert self.async_handler.get_status() is sf_api.AsyncResultStatus.running


@mock.patch('valence.functional_models.Cluster.get_slice_info')
class TestWaitForReassign(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def setUp(self):
        self.mock_response = result_models.GetSliceInfo.get_mock_response()
        self.mock_result_dm = self.mock_response.result_dm
        slice_id = self.mock_result_dm.slice.slice_id
        self.ah_fm = functional_models.SliceReassignmentAsyncHandler(self.cluster_fm, slice_id, 1, [2])

    def test_get_status(self, mock_slice):
        self.mock_result_dm.slice.primary = 1
        self.mock_result_dm.slice.live_secondaries = [2]
        mock_slice.return_value = self.mock_result_dm
        status = self.ah_fm._get_fake_async_status()
        assert status is enums.AsyncResultStatus.complete

    def test_primary_mismatch(self, mock_slice):
        self.mock_result_dm.slice.primary = 2
        self.mock_result_dm.slice.live_secondaries = [2]
        mock_slice.return_value = self.mock_result_dm
        status = self.ah_fm._get_fake_async_status()
        assert status is enums.AsyncResultStatus.running

    @mock.patch('valence.functional_models.Cluster._reassign_slice')
    def test_prune(self, mock_reassign, mock_slice):
        self.mock_result_dm.slice.primary = 1
        self.mock_result_dm.slice.live_secondaries = [2, 3]
        mock_slice.return_value = self.mock_result_dm
        status = self.ah_fm._get_fake_async_status()
        mock_slice.assert_called_with(slice_id=self.mock_result_dm.slice.slice_id)
        mock_reassign.assert_called_with(self.mock_result_dm.slice.slice_id,
                                         action=enums.SliceAction.prune,
                                         db_version=self.mock_result_dm.db_version,
                                         prune=[3])
        assert status is enums.AsyncResultStatus.running


class TestFillSliceAsyncHandler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_get_models(self, mock_list_volumes):
        volume_dm = models.Volume.get_mock_object()
        volume_dm.volume_id = 123
        mock_list_volumes.return_value = [volume_dm]
        async_handle_dm = models.AsyncHandle.get_mock_object()
        async_handler = functional_models.FillSliceAsyncHandler(
            self.cluster_fm, 123, async_handle_dm)
        assert async_handler._get_models() == [volume_dm]


class TestRtfiAsyncHandler(unittest.TestCase):
    @mock.patch('sf_rtfi.reactor.HTTPIRTFIReactor')
    def test_get_fake_async_status(self, mock_reactor):
        mock_async_handler = functional_models.RtfiAsyncHandler(mock_reactor)
        mock_reactor.rtfi_state.complete.return_value = False
        mock_reactor.rtfi_state.failed.return_value = True
        mock_reactor.rtfi_state.aborted.return_value = False
        mock_reactor.rtfi_state.timedout.return_value = False
        mock_reactor.rtfi_state.done.return_value = False
        assert mock_async_handler._get_fake_async_status() == sf_api.AsyncResultStatus.failed
        mock_reactor.rtfi_state.failed.return_value = False
        mock_reactor.rtfi_state.complete.return_value = True
        assert mock_async_handler._get_fake_async_status() == sf_api.AsyncResultStatus.complete
        mock_reactor.rtfi_state.complete.return_value = False
        assert mock_async_handler._get_fake_async_status() == sf_api.AsyncResultStatus.running


class TestRebalanceSlicesAsyncHandler(unittest.TestCase):
    def setUp(self):
        self.cluster_fm = functional_models.Cluster.get_mock_fm()

    @mock.patch.object(functional_models.Cluster, 'is_slice_syncing')
    def test_rebalance_slices_async_handler_running(self,
                                                    mock_is_slice_syncing):
        mock_is_slice_syncing.return_value = True
        async_handler = functional_models.RebalanceSlicesAsyncHandler(self.cluster_fm)
        assert async_handler._get_fake_async_status() is enums.AsyncResultStatus.running

    @mock.patch.object(functional_models.Cluster, 'is_slice_syncing')
    def test_rebalance_slices_async_handler_complete(self,
                                                     mock_is_slice_syncing):
        mock_is_slice_syncing.return_value = False
        async_handler = functional_models.RebalanceSlicesAsyncHandler(self.cluster_fm)
        assert async_handler._get_fake_async_status() is enums.AsyncResultStatus.complete

    @mock.patch.object(functional_models._async_handler.RebalanceSlicesAsyncHandler,
                       '_get_fake_async_status')
    def test_rebalance_slices_async_handler_complete2(self, mock_fake_async_status):
        mock_fake_async_status.return_value = True
        async_handler = functional_models.RebalanceSlicesAsyncHandler(self.cluster_fm)

    @mock.patch.object(functional_models._async_handler.RebalanceSlicesAsyncHandler,
                       '_get_fake_async_status')
    def test_rebalance_slices_async_handler_running2(self, mock_fake_async_status):
        mock_fake_async_status.return_value = False
        async_handler = functional_models.RebalanceSlicesAsyncHandler(self.cluster_fm)


class TestSecureEraseDrivesAsyncHandler(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.drive_dm1 = models.Drive.get_mock_object()
        self.drive_dm2 = models.Drive.get_mock_object()
        self.drives_to_erase = [self.drive_dm1, self.drive_dm2]
        self.async_handler = functional_models.SecureEraseDrivesAsyncHandler(
            cluster_fm, [d.drive_id for d in self.drives_to_erase], async_handle_dm)

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_models(self, mock_list_drives):
        """Test that _get_models only returns models associated with the AsyncHandler."""
        mock_list_drives.return_value = self.drives_to_erase
        assert self.async_handler._get_models() == self.drives_to_erase


class TestEnableFeatureAsyncHandler(unittest.TestCase):
    def setUp(self):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        self.feature = enums.SfFeature.fips_drives
        self.enable_feature_ah = functional_models.EnableFeatureAsyncHandler(cluster_fm,
                                                                             enums.SfFeature.fips_drives)

    @mock.patch.object(functional_models.Cluster, 'get_feature_status')
    def test_enable_feature_async_handler(self, mock_get_status):
        disabled_status = [models.FeatureStatus.get_mock_object(
            overrides=dict(enabled=False, feature=self.feature)
        )]
        enabled_status = [models.FeatureStatus.get_mock_object(
            overrides=dict(enabled=True, feature=self.feature)
        )]
        mock_get_status.side_effect = [disabled_status, enabled_status]
        self.enable_feature_ah.join(poll_period=0)
