"""Unittests for client functional model."""
from __future__ import division

import threading
import unittest

import mock
import pytest
import sf_ssh
from sf_ssh import (
    ssh,
    utils,
)

from camelot.models import at2_enums
from valence import (
    enums,
    functional_models,
    models,
    unittest_utils,
)

DATA_DIR = unittest_utils.get_data_dir(__file__)


@pytest.yield_fixture(scope='function')
def mock_ssh_run_command(monkeypatch):
    """Returns a mock object of sf_ssh.Ssh.run_command.

    This mocks Ssh.close and Ssh._init_client so it works with the sf_ssh context manager.
    """
    monkeypatch.setattr(sf_ssh.Ssh, 'close', mock.Mock())
    monkeypatch.setattr(sf_ssh.Ssh, '_init_client', mock.Mock())
    patcher = mock.patch('sf_ssh.Ssh.run_command')
    yield patcher.start()
    patcher.stop()


@pytest.yield_fixture(scope='function')
def mock_invoke_shell(monkeypatch):
    """Returns a mock object of sf_ssh.SshShell.invoke_shell"""
    monkeypatch.setattr(sf_ssh.Ssh, 'close', mock.Mock())
    monkeypatch.setattr(sf_ssh.Ssh, '_init_client', mock.Mock())
    patcher = mock.patch('sf_ssh.SshShell.invoke_shell')
    yield patcher.start()
    patcher.stop()


@pytest.yield_fixture(scope='function')
def mock_ssh_shell_run_command(monkeypatch):
    """Returns a mock object of sf_ssh.SshShell.invoke_shell"""
    monkeypatch.setattr(sf_ssh.Ssh, 'close', mock.Mock())
    monkeypatch.setattr(sf_ssh.Ssh, '_init_client', mock.Mock())
    patcher = mock.patch('sf_ssh.SshShell.run_command')
    yield patcher.start()
    patcher.stop()


@pytest.fixture(scope='module')
def client_fm():
    client_dm = models.ResourceClient.get_mock_object()
    return functional_models.UbuntuClient(client_dm)


#############################
# General client FM methods
#############################

def test_client_repr(client_fm):
    assert 'ClientFm' in repr(client_fm)
    assert 'mip=' in repr(client_fm)
    assert client_fm.mip in repr(client_fm)


def test_run_command_on_client(client_fm, mock_ssh_run_command):
    mock_ssh_run_command.return_value = sf_ssh.RunCommandOutput(
        stdout='hello, world!', stderr='', return_code=0)
    out = client_fm.run_command(command='echo hello, world!')
    assert out.stdout == 'hello, world!'


def test_run_command_on_client_fail(client_fm, mock_ssh_run_command):
    mock_ssh_run_command.return_value = sf_ssh.RunCommandOutput(
        stdout='', stderr='', return_code=1)
    with pytest.raises(functional_models._errors.FunctionalClientError):
        client_fm.run_command(command='echo hello, world!', raise_on_failure=True)


def test_get_interactive_shell(client_fm, mock_invoke_shell, mock_ssh_shell_run_command):
    mock_ssh_shell_run_command.return_value = ('hello', 0)
    shell = client_fm.get_interactive_shell()
    shell.invoke_shell()
    out, rc = shell.run_command('echo hello')
    shell.close()
    assert out == 'hello'
    assert rc == 0


def test_get_client_from_dm():
    client_dm = models.ResourceClient()
    client_dm.os = mock.Mock()
    client_dm.os.is_ubuntu = True
    client_fm = functional_models.get_client_from_dm(client_dm)
    assert isinstance(client_fm, functional_models.UbuntuClient)


def test_get_client_from_windows_dm():
    client_dm = models.ResourceClient()
    client_dm.os = mock.Mock()
    client_dm.os.is_ubuntu = False
    client_dm.os.is_centos = False
    client_dm.os.is_windows = True
    client_fm = functional_models.get_client_from_dm(client_dm)
    assert isinstance(client_fm, functional_models.WindowsClient)


def test_get_client_from_dm_failure():
    client_dm = models.ResourceClient()
    client_dm.os = mock.Mock()
    client_dm.os.is_ubuntu = False
    client_dm.os.is_centos = False
    client_dm.os.is_windows = False
    with pytest.raises(ValueError):
        client_fm = functional_models.get_client_from_dm(client_dm)


################################
# Support bundle tests
################################

def test_pull_support_bundle(client_fm, monkeypatch):
    mock_create_bundle = mock.Mock()
    mock_cleanup = mock.Mock()
    monkeypatch.setattr(functional_models.UbuntuClient, '_cleanup_remote', mock_cleanup)
    monkeypatch.setattr(functional_models.UbuntuClient, 'create_bundle', mock_create_bundle)
    artifact_manager = mock.MagicMock()
    artifact_manager.store_artifacts_remote.return_value = {'name1': '/path/to/artifact'}
    client_fm.pull_support_bundle(artifact_manager)
    mock_create_bundle.assert_called()
    mock_cleanup.assert_called()


def test_get_client_bundle_path(client_fm):
    path = client_fm._get_client_bundle_path()
    assert client_fm.client_dm.client_name in path


def test_create_bundle_success(client_fm, mock_ssh_run_command):
    mock_ssh_run_command.return_value = sf_ssh.RunCommandOutput(stdout='abc', stderr='',
                                                                return_code=0)
    bundle_info = client_fm.create_bundle()
    assert isinstance(bundle_info, functional_models.BundleInfo)


def test_create_bundle_timeout(client_fm, mock_ssh_run_command):
    mock_ssh_run_command.return_value = sf_ssh.RunCommandOutput(stdout='abc', stderr='',
                                                                return_code=-99)
    with pytest.raises(functional_models._errors.ClientBundleError):
        client_fm.create_bundle()


class TestUbuntuClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        client_dm = models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04))
        cls.client_fm = functional_models.UbuntuClient(client_dm)
        cls.portal_ip = next(utils.generate_ips(1))

    @mock.patch.object(functional_models.UbuntuClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.UbuntuClient, 'run_command')
    def test_update_iscsi_d_conf(self, mock_run_command, mock_put):
        with open('{}/iscsid.conf'.format(DATA_DIR)) as iscsid_file:
            iscsid_contents = iscsid_file.read()
        output = ssh.RunCommandOutput(stdout=iscsid_contents, stderr='', return_code=0)
        mock_run_command.return_value = output
        self.client_fm.update_iscsi_d_conf('harry', 'magicdude', 'potter', double_authentication=True)
        modified_iscsid = ''.join(mock_put.call_args_list[0][0][0])
        assert iscsid_contents != modified_iscsid
        assert 'magicdude' in modified_iscsid

    @mock.patch.object(functional_models.UbuntuClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.UbuntuClient, '_get_client_iscsid_conf')
    @mock.patch.object(functional_models.UbuntuClient, '_iscsi_set_targets_and_discovery_info')
    def test_update_iscsi_d_conf_no_double_auth(self, mock_set_targets, mock_get_iscsid_conf, mock_put):
        mock_get_iscsid_conf.return_value = 'dummyconf'
        self.client_fm.update_iscsi_d_conf('harry', 'magicdude', 'potter')
        mock_set_targets.assert_called_with('dummyconf', 'harry', 'magicdude', None)

    @mock.patch.object(functional_models.UbuntuClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.UbuntuClient, '_get_client_iscsid_conf')
    @mock.patch.object(functional_models.UbuntuClient, '_iscsi_set_targets_and_discovery_info')
    def test_update_iscsi_d_conf_double_auth(self, mock_set_targets, mock_get_iscsid_conf, mock_put):
        mock_get_iscsid_conf.return_value = 'dummyconf'
        self.client_fm.update_iscsi_d_conf('harry', 'magicdude', 'potter', double_authentication=True)
        mock_set_targets.assert_called_with('dummyconf', 'harry', 'magicdude', 'potter')

    def test_update_iscsi_d_conf_double_auth_same_password(self):
        with pytest.raises(functional_models._errors.FunctionalClientError):
            self.client_fm.update_iscsi_d_conf('harry', 'potter', 'potter', double_authentication=True)

    def test_deactivate_target_chap(self):
        with open('{}/iscsid.conf'.format(DATA_DIR)) as iscsid_file:
            iscsid_lines = iscsid_file.read().splitlines(True)
        for i, line in enumerate(iscsid_lines):
            if line.startswith('#discovery.sendtargets.auth.username_in'):
                iscsid_lines[i] = line.replace('#', '')
            elif line.startswith('#discovery.sendtargets.auth.password_in'):
                iscsid_lines[i] = line.replace('#', '')
            elif line.startswith('#node.session.auth.username_in'):
                iscsid_lines[i] = line.replace('#', '')
            elif line.startswith('#node.session.auth.password_in'):
                iscsid_lines[i] = line.replace('#', '')
        new_iscsid_lines = self.client_fm._iscsi_set_targets_and_discovery_info(iscsid_lines, 'harry',
                                                                                'potter')
        assert new_iscsid_lines[47].startswith('#node.session.auth.username_in')
        assert new_iscsid_lines[48].startswith('#node.session.auth.password_in')
        assert new_iscsid_lines[61].startswith('#discovery.sendtargets.auth.username_in')
        assert new_iscsid_lines[62].startswith('#discovery.sendtargets.auth.password_in')

    @mock.patch.object(functional_models.UbuntuClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.UbuntuClient, '_get_client_iscsid_conf')
    def test_deactivate_chap(self, mock_get_iscsid_conf, mock_put):
        with open('{}/iscsid.conf'.format(DATA_DIR)) as iscsid_file:
            iscsid_contents = iscsid_file.read()
        iscsid_lines = iscsid_contents.splitlines(True)
        assert not iscsid_lines[38].startswith('#CHAP_DISABLED:')
        mock_get_iscsid_conf.return_value = iscsid_lines
        self.client_fm.deactivate_chap()
        modified_iscsid_lines = mock_put.call_args_list[-1][0][0]
        assert modified_iscsid_lines[38].startswith('#CHAP_DISABLED:')

        mock_get_iscsid_conf.return_value = modified_iscsid_lines
        self.client_fm.reactivate_chap()
        modified_iscsid_lines = mock_put.call_args_list[-1][0][0]
        assert not modified_iscsid_lines[38].startswith('#CHAP_DISABLED:')

    @mock.patch('sf_ssh.Ssh')
    def test_put_iscsi_d_conf(self, mock_ssh):
        self.client_fm._put_iscsid_conf(['line1\n', 'line2\n'])
        mock_ssh.return_value.__enter__.return_value.put.assert_called()

    @mock.patch('valence.functional_models._fm_utils.BlockPoll.start_polling')
    def test_wait_until_iscsid_down(self, mock_start_polling):
        self.client_fm.wait_until_iscsid('down')

    @mock.patch('valence.functional_models._fm_utils.BlockPoll.start_polling')
    def test_wait_until_iscsid_up(self, mock_start_polling):
        self.client_fm.wait_until_iscsid('up')

    @mock.patch('valence.functional_models._fm_utils.BlockPoll.start_polling')
    def test_wait_until_iscsid_error(self, mock_start_polling):
        mock_start_polling.side_effect = functional_models._errors.PollMaxError
        with pytest.raises(functional_models._errors.IscsiError):
            self.client_fm.wait_until_iscsid('up')

    @mock.patch('valence.functional_models.UbuntuClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    @mock.patch('valence.functional_models.UbuntuClient.version', new_callable=mock.PropertyMock)
    def test_iscsid_start_14_04_or_earlier(self, mock_version, mock_run_command_check_output,
                                           mock_wait_until_iscsid):
        mock_version.return_value = at2_enums.ClientOS.ubuntu_14_04
        self.client_fm.iscsid_start(restart=False)
        mock_run_command_check_output.assert_called_with('/etc/init.d/open-iscsi start')
        mock_wait_until_iscsid.assert_called_with('up')

    @mock.patch('valence.functional_models.UbuntuClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    @mock.patch('valence.functional_models.UbuntuClient.version', new_callable=mock.PropertyMock)
    def test_iscsid_restart_14_04_or_earlier(self, mock_version, mock_run_command_check_output,
                                             mock_wait_until_iscsid):
        mock_version.return_value = at2_enums.ClientOS.ubuntu_14_04
        self.client_fm.iscsid_start(restart=True)
        mock_run_command_check_output.assert_called_with('/etc/init.d/open-iscsi restart')
        mock_wait_until_iscsid.assert_called_with('up')

    @mock.patch('valence.functional_models.UbuntuClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    @mock.patch('valence.functional_models.UbuntuClient.version', new_callable=mock.PropertyMock)
    def test_iscsid_start_later_than_14_04(self, mock_version, mock_run_command_check_output,
                                           mock_wait_until_iscsid):
        mock_version.return_value = at2_enums.ClientOS.ubuntu_16_04
        self.client_fm.iscsid_start(restart=False)
        mock_run_command_check_output.assert_called_with('systemctl start iscsid open-iscsi')
        mock_wait_until_iscsid.assert_called_with('up')

    @mock.patch('valence.functional_models.UbuntuClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    @mock.patch('valence.functional_models.UbuntuClient.version', new_callable=mock.PropertyMock)
    def test_iscsid_restart_later_than_14_04(self, mock_version, mock_run_command_check_output,
                                             mock_wait_until_iscsid):
        mock_version.return_value = at2_enums.ClientOS.ubuntu_16_04
        self.client_fm.iscsid_start(restart=True)
        mock_run_command_check_output.assert_called_with('systemctl restart iscsid open-iscsi')
        mock_wait_until_iscsid.assert_called_with('up')

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_get_pid_of_iscsid(self, mock_run_command):
        mock_run_command.return_value = ssh.RunCommandOutput('7172 7171', 'stderr', 0)
        pid_list = self.client_fm.get_pid_of_iscsid()
        assert pid_list == [7172, 7171]

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_get_pid_of_iscsid_empty(self, mock_run_command):
        mock_run_command.return_value = ssh.RunCommandOutput('stdout', 'stderr', 1)
        pid_list = self.client_fm.get_pid_of_iscsid()
        assert pid_list == []

    def test_login_all_invalid_mode(self):
        with pytest.raises(functional_models._errors.IscsiError):
            self.client_fm.login_all(mode='invalid_mode')

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_login_all_mode_all(self, mock_run_command_check_output):
        self.client_fm.login_all(mode=enums.IscsiMode.all)

    @mock.patch('valence.functional_models.UbuntuClient.login_all_batch_threaded')
    @mock.patch('valence.functional_models.UbuntuClient.login_all_batch')
    def test_login_all_mode_batch_with_exception(self, mock_batch_login, mock_threaded_batch_login):
        mock_threaded_batch_login.side_effect = Exception
        self.client_fm.login_all(mode=enums.IscsiMode.batch)

    @mock.patch('valence.functional_models.UbuntuClient.run_batch_login')
    @mock.patch('valence.functional_models.UbuntuClient._discover')
    def test_login_all_batch(self, mock_discover, mock_run_batch_login):
        mock_discover.return_value = ['iqn.2010-01.com.solidfire:p0jq.vol1.1',
                                      'iqn.2010-01.com.solidfire:p0jq.vol2.2',
                                      'iqn.2010-01.com.solidfire:p0jq.vol3.3']
        mock_run_batch_login.return_value = '1'
        batch_return = self.client_fm.login_all_batch(self.portal_ip, 1)
        assert batch_return == ['1', '1', '1']

    @mock.patch('valence.functional_models.UbuntuClient.run_batch_login')
    @mock.patch('valence.functional_models.UbuntuClient._discover')
    def test_login_all_batch_threaded(self, mock_discover, mock_run_batch_login):
        mock_discover.return_value = ['iqn.2010-01.com.solidfire:p0jq.vol1.1',
                                      'iqn.2010-01.com.solidfire:p0jq.vol2.2',
                                      'iqn.2010-01.com.solidfire:p0jq.vol3.3',
                                      'iqn.2010-01.com.solidfire:p0jq.morevols-0199.405']
        mock_run_batch_login.return_value = '1'
        batch_return = self.client_fm.login_all_batch_threaded(self.portal_ip, 2)
        assert len(batch_return) == 2
        assert type(batch_return[0]) == threading.Thread

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_run_batch_login(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = 'login batches'
        result = self.client_fm.run_batch_login(['yeah', 'batches'], self.portal_ip)
        assert result == 'login batches'

    @mock.patch('valence.functional_models.UbuntuClient._list_disks_by_id')
    @mock.patch('valence.functional_models.UbuntuClient._get_target_map')
    def test_get_disk_map(self, mock_get_target_map, mock_list_disks_by_id):
        mock_get_target_map.return_value = {'1': 'iqn.2010-01.com.solidfire:p0jq.vol1.1',
                                            '10': 'iqn.2010-01.com.solidfire:p0jq.extravols-0004.10',
                                            '100': 'iqn.2010-01.com.solidfire:p0jq.extravols-0094.100'}
        mock_list_disks_by_id.return_value = {'1': '/dev/sdht',
                                              '10': '/dev/sdfg',
                                              '100': '/dev/sdkr'}
        disk_map = self.client_fm.get_disk_map()
        assert disk_map == {'1': '/dev/sdht', '10': '/dev/sdfg', '100': '/dev/sdkr'}

    @mock.patch('time.sleep')
    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    @mock.patch('valence.functional_models.UbuntuClient._list_disks_by_id')
    @mock.patch('valence.functional_models.UbuntuClient._get_target_map')
    def test_get_disk_map_fail(self, mock_get_target_map, mock_list_disks_by_id,
                               mock_run_command, mock_sleep):
        mock_get_target_map.return_value = {'1': 'iqn.2010-01.com.solidfire:p0jq.vol1.1'}
        mock_list_disks_by_id.return_value = {'1': '/dev/sdht',
                                              '10': '/dev/sdfg'}
        with pytest.raises(functional_models._errors.IscsiError):
            disk_map = self.client_fm.get_disk_map()

    @mock.patch('valence.functional_models.UbuntuClient._node_info')
    def test_get_target_map(self, mock_node_info):
        mock_node_info.return_value = \
            ['10.117.94.233:3260,1 iqn.2010-01.com.solidfire:p0jq.morevols-0181.387',
             '10.117.94.233:3260,1 iqn.2010-01.com.solidfire:p0jq.extravols-0064.70',
             '10.117.94.233:3260,1 iqn.2010-01.com.solidfire:p0jq.morevols-0124.330']
        target_map = self.client_fm._get_target_map()
        assert target_map == {'70': 'iqn.2010-01.com.solidfire:p0jq.extravols-0064.70',
                              '330': 'iqn.2010-01.com.solidfire:p0jq.morevols-0124.330',
                              '387': 'iqn.2010-01.com.solidfire:p0jq.morevols-0181.387'}

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_node_info(self, mock_run_command_check_output, mock_run_command):
        mock_run_command_check_output.return_value = \
            ('10.117.94.233:3260,1'
             'iqn.2010-01.com.solidfire:p0jq.morevols-0181.387\n'
             '10.117.94.233:3260,1'
             'iqn.2010-01.com.solidfire:p0jq.extravols-0064.70\n')
        mock_run_command.return_value = ssh.RunCommandOutput('clabv0012\n', 'stderr', 0)
        node_info = self.client_fm._node_info()
        assert node_info == ['10.117.94.233:3260,1iqn.2010-01.com.solidfire:p0jq.morevols-0181.387',
                             '10.117.94.233:3260,1iqn.2010-01.com.solidfire:p0jq.extravols-0064.70']

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_node_info_skip_boot(self, mock_run_command_check_output, mock_run_command):
        mock_run_command_check_output.return_value = 'clabv0012'
        mock_run_command.return_value = ssh.RunCommandOutput('clabv0012', 'stderr', 0)
        node_info = self.client_fm._node_info()

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_list_disks_by_id(self, mock_run_command):
        disk_stdout = \
            ('lrwxrwxrwx 1 root root 10 Nov 27 20:18 '
             'ip-10.117.94.233:3260-iscsi-iqn.2010-01.com.solidfire:p0jq.extravols-0000.6-lun-0'
             ' -> ../../sdcm\n'
             'lrwxrwxrwx 1 root root 10 Nov 27 20:18 '
             'ip-10.117.94.233:3260-iscsi-iqn.2010-01.com.solidfire:p0jq.extravols-0001.7-lun-0'
             ' -> ../../sdbc\n')
        mock_run_command.return_value = ssh.RunCommandOutput(disk_stdout, 'stderr', 0)
        disk_map = self.client_fm._list_disks_by_id()
        assert disk_map == {'7': '/dev/sdbc', '6': '/dev/sdcm'}

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_discovery(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = ('10.117.94.233:3260,1 '
                                                      'iqn.2010-01.com.solidfire:p0jq.vol1.1\n'
                                                      '10.117.94.233:3260,1 '
                                                      'iqn.2010-01.com.solidfire:p0jq.vol2.2\n')
        targets = self.client_fm._discover(self.portal_ip)
        assert targets == ['iqn.2010-01.com.solidfire:p0jq.vol1.1',
                           'iqn.2010-01.com.solidfire:p0jq.vol2.2']

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_run_command_check_output(self, mock_run_command):
        mock_run_command.return_value = ssh.RunCommandOutput('command success', 'stderr', 0)
        cmd_stdout = self.client_fm._run_command_check_output('real command')
        assert cmd_stdout == 'command success'

    @mock.patch('time.sleep')
    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_run_command_check_output_FAIL(self, mock_run_command, mock_sleep):
        mock_run_command.return_value = ssh.RunCommandOutput('command success', 'stderr', 1)
        with pytest.raises(functional_models._errors.IscsiError):
            cmd_stdout = self.client_fm._run_command_check_output('real command')

    @mock.patch('valence.functional_models.UbuntuClient.get_disk_map')
    @mock.patch('valence.functional_models.UbuntuClient.login_all')
    @mock.patch('valence.functional_models.UbuntuClient._discover')
    @mock.patch('valence.functional_models.UbuntuClient.iscsid_start')
    def test_volume_ceremony(self, mock_iscsid_start, mock_discover, mock_login_all, mock_get_disk_map):
        mock_get_disk_map.return_value = {'1': '/dev/sdba', '2': '/dev/sdcm'}
        disk_map = self.client_fm.volume_ceremony(self.portal_ip)
        assert disk_map == {'1': '/dev/sdba', '2': '/dev/sdcm'}

    @mock.patch('valence.functional_models.UbuntuClient.get_disk_map')
    @mock.patch('valence.functional_models.UbuntuClient.login_all')
    @mock.patch('valence.functional_models.UbuntuClient._discover')
    @mock.patch('valence.functional_models.UbuntuClient.purge_restart')
    def test_volume_ceremony_purge(self, mock_purge_restart, mock_discover,
                                   mock_login_all, mock_get_disk_map):
        mock_get_disk_map.return_value = {'1': '/dev/sdba', '2': '/dev/sdcm'}
        disk_map = self.client_fm.volume_ceremony(portal_ip=self.portal_ip, purge=True)
        assert disk_map == {'1': '/dev/sdba', '2': '/dev/sdcm'}

    @mock.patch('valence.functional_models.UbuntuClient.iscsid_start')
    @mock.patch('valence.functional_models.UbuntuClient.purge_credentials')
    @mock.patch('valence.functional_models.UbuntuClient.iscsid_stop')
    @mock.patch('valence.functional_models.UbuntuClient.logout_all')
    def test_purge_restart(self, mock_logout_all, mock_iscsid_stop,
                           mock_purge_credentials, mock_iscsid_start):
        self.client_fm.purge_restart()

    @mock.patch('valence.functional_models.UbuntuClient.iscsid_start')
    @mock.patch('valence.functional_models.UbuntuClient.purge_credentials')
    @mock.patch('valence.functional_models.UbuntuClient.iscsid_stop')
    @mock.patch('valence.functional_models.UbuntuClient.logout_all')
    def test_purge_restart_exception(self, mock_logout_all, mock_iscsid_stop,
                                     mock_purge_credentials, mock_iscsid_start):
        mock_logout_all.side_effect = functional_models._errors.IscsiError
        self.client_fm.purge_restart()

    @mock.patch('valence.functional_models.UbuntuClient._check_iscsi_initramfs')
    @mock.patch('valence.functional_models.UbuntuClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    @mock.patch('valence.functional_models.UbuntuClient.version', new_callable=mock.PropertyMock)
    def test_iscsid_stop_14_04_or_earlier(self, mock_version, mock_run_command_check_output,
                                          mock_wait_until_iscsid, mock_check_iscsi_initramfs):
        mock_version.return_value = at2_enums.ClientOS.ubuntu_14_04
        self.client_fm.iscsid_stop()
        mock_run_command_check_output.assert_called_with('/etc/init.d/open-iscsi stop')
        mock_wait_until_iscsid.assert_called_with('down')
        mock_check_iscsi_initramfs.assert_called_once()

    @mock.patch('valence.functional_models.UbuntuClient._check_iscsi_initramfs')
    @mock.patch('valence.functional_models.UbuntuClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    @mock.patch('valence.functional_models.UbuntuClient.version', new_callable=mock.PropertyMock)
    def test_iscsid_stop_later_than_14_04(self, mock_version, mock_run_command_check_output,
                                          mock_wait_until_iscsid, mock_check_iscsi_initramfs):
        mock_version.return_value = at2_enums.ClientOS.ubuntu_16_04
        self.client_fm.iscsid_stop()
        mock_run_command_check_output.assert_called_with('systemctl stop iscsid open-iscsi')
        mock_wait_until_iscsid.assert_called_with('down')
        mock_check_iscsi_initramfs.assert_called_once()

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_logout_all(self, mock_run_command_check_output):
        self.client_fm.logout_all()

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_logout_all_svip(self, mock_run_command_check_output):
        self.client_fm.logout_all(self.portal_ip)

    @mock.patch('valence.functional_models.UbuntuClient._check_iscsi_initramfs')
    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_purge_credentials(self, mock_run_command_check_output, mock_check_iscsi_initramfs):
        self.client_fm.purge_credentials()

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_check_iscsi_initramfs(self, mock_run_command_check_output):
        self.client_fm._check_iscsi_initramfs(iscsi_root_volume=False)

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_check_iscsi_initramfs_root_true(self, mock_run_command_check_output):
        self.client_fm._check_iscsi_initramfs(iscsi_root_volume=True)

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_get_initiator_name(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = \
            'InitiatorName=iqn.2010-01.net.solidfire.eng:clabv0006'
        initiator_name = self.client_fm.get_initiator_name()
        assert initiator_name == 'iqn.2010-01.net.solidfire.eng:clabv0006'

    @mock.patch('valence.functional_models.UbuntuClient._run_command_check_output')
    def test_get_initiator_name_fails(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = \
            'InitiatorNameFAKE=iqn.2010-01.net.solidfire.eng:clabv0006'
        with pytest.raises(functional_models._errors.IscsiError):
            initiator_name = self.client_fm.get_initiator_name()

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_create_tmp_cert(self, mock_run):
        mock_run.return_value = ssh.RunCommandOutput('/tmp/unittest', '', 0)
        self.client_fm.create_temp_cert_dir()

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_create_tmp_no_return(self, mock_run):
        mock_run.return_value = ssh.RunCommandOutput('', '', 0)
        with pytest.raises(functional_models._errors.FunctionalClientError):
            self.client_fm.create_temp_cert_dir()

    @mock.patch('valence.functional_models.UbuntuClient.get_ssh_connection')
    def test_create_file(self, mock_run):
        self.client_fm.create_file(target_dir='/tmp', filename='unit.conf', raw='unittest')

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_cat_file(self, mock_run):
        mock_run.return_value = ssh.RunCommandOutput(stdout='/tmp/unittest', stderr='', return_code=0)
        self.client_fm.get_file_contents(file_directory='/tmp/unittest', filename='unit', timeout=0)

    @mock.patch('valence.functional_models.UbuntuClient.get_ssh_connection')
    def test_get_cert_from_port(self, mock_run):
        """Automatically generated docstring for test_get_cert.

        :param Any mock_ssh: Autogenerated description.
        """
        run_safe = mock_run.return_value.__enter__.return_value.run_command
        run_safe.return_value = sf_ssh.RunCommandOutput(
            stdout=('-----BEGIN CERTIFICATE-----\n'
                    'MIID2DCCAsACAhAAMA0GCSqGSIb3DQEBBAUAMIHDMQswCQYDVQQGEwJVUzERMA8G\n'
                    'A1UECBMIQ29sb3JhZG8xEDAOBgNVBAcTB0JvdWxkZXIxDjAMBgNVBBETBTgwMzAz\n'
                    'MRswGQYDVQQJFBIxMDQ4IFBlYXJsIFN0ICMyNTAxGTAXBgNVBAoTEFNvbGlkZmly\n'
                    'ZSBOZXRhcHAxDDAKBgNVBAsTA1NFVDESMBAGA1UEAxMJY2xhYnYwMDEyMSUwIwYJ\n'
                    'KoZIhvcNAQkBFhZ0ZXN0X2NvbmZpZ0BuZXRhcHAuY29tMCIYDzIwMTYwMzIyMDky\n'
                    'NTIyWhgPMjAxNzAzMTIwOTI1MjJaMIGaMQswCQYDVQQGEwJVUzERMA8GA1UECBMI\n'
                    'Q29sb3JhZG8xEDAOBgNVBAcTB0JvdWxkZXIxGTAXBgNVBAoTEFNvbGlkZmlyZSBO\n'
                    'ZXRhcHAxDDAKBgNVBAsTA1NFVDEWMBQGA1UEAxMNMTAuMTE3Ljc4LjIzMjElMCMG\n'
                    'CSqGSIb3DQEJARYWdGVzdF9jb25maWdAbmV0YXBwLmNvbTCCASIwDQYJKoZIhvcN\n'
                    'AQEBBQADggEPADCCAQoCggEBALLaKuf1WK+m2V82n0NqLzrak9QbsjPKwt//djeR\n'
                    'GuFJtk9qujb9XXEC/UeMubra1d1CcTjrNtvnBq1hxBrnWioziFcsTW/3wyZHa2u/\n'
                    '0P1yizs59TnMYBR0Bui0LxDVUDAxNfjUxgcJX/MzJ+VrHn1l3mhaMSJ5XXOWmtKG\n'
                    'XZQJXcium7zvCg+VzZmeu/sCk4AaWjyLi0aKrAX3X8Y2olSxe0j07xU3KPufzxal\n'
                    'eXOcNeyaF30jyUZGGEINr2j3dAr7Hc+HA2rKtrUoBR4Tsq1QltDUPzjHhcWyOXnS\n'
                    'pTjvHgrdLY7+j+qdPxd98Ya3tUnRL95TdilI/il9E7uiGekCAwEAATANBgkqhkiG\n'
                    '9w0BAQQFAAOCAQEAP3B+UBUaX8KQme22hTZl12m3VbQ4c9oeMlI08/pC5YEP/PAY\n'
                    '9nyHLBRrU9S8c7jGiJD3stGGXco8jSArbfviJvYEFbngSyoYFYIJYlJn4x8mJruu\n'
                    'n1iH7cok9wFR8YUiDTe3xX05CfMoM4He6iUvcOvhVgVRb/4HmnYcqjIq1D7ulHw1\n'
                    'UK7PBCz9wiqzeen0FsYnlFvb8jHR92/I+UviZlw2f95uboxTga0fND2MYjvgzGPK\n'
                    '0w1XqNVGKW+bYIEli9kkDK6mB8RiFw0krJHheQuU3TITf9aWYXfCO/vDrrP5GKAh\n'
                    'HI/GdDS/Cn5KyAraksvdSw7gjMQBmbjAL8IHoQ==\n'
                    '-----END CERTIFICATE-----'),
            stderr='',
            return_code=0)
        self.client_fm.get_cert_from_port('UNITTESTIP')

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_run_command_wrappers(self, mock_run):
        # This list of methods all share the same test
        methods = [
            'sg_raw',
            'sginfo',
            'sg_vpd',
            'sg_persist',
            'sg_read',
            'sg_map',
            'multipath'
        ]
        for method in methods:
            # Test a successful call
            mock_run_output = ssh.RunCommandOutput(
                stdout='mock std out', stderr='mock std err', return_code=0)
            mock_run.return_value = mock_run_output
            output = getattr(self.client_fm, method)('mock options')
            assert output.stdout == 'mock std out'
            assert output.stderr == 'mock std err'
            assert output.return_code == 0

            # Test a failed call
            mock_run_output = ssh.RunCommandOutput(
                stdout='mock std out', stderr='mock std err', return_code=1)
            mock_run.return_value = mock_run_output
            output = getattr(self.client_fm, method)('mock options', 1)
            assert output.stdout == 'mock std out'
            assert output.stderr == 'mock std err'
            assert output.return_code == 1

            # Test an exception for an unexpected return code
            mock_run_output = ssh.RunCommandOutput(
                stdout='mock std out', stderr='mock std err', return_code=1)
            mock_run.return_value = mock_run_output
            with pytest.raises(functional_models._errors.FunctionalClientError):
                getattr(self.client_fm, method)('mock options', 0)

    @mock.patch('valence.functional_models.UbuntuClient.sg_map')
    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_get_fc_drives(self, mock_run, mock_sg_map):
        mock_run_output = ssh.RunCommandOutput(
            stdout='host7', stderr='mock std err', return_code=0)
        mock_run.return_value = mock_run_output
        mock_map_output = ssh.RunCommandOutput(
            stdout='/dev/sg4  7 0 0 0  0  /dev/sdd  SolidFir  SSD SAN           0001',
            stderr='mock std err', return_code=0)
        mock_sg_map.return_value = mock_map_output
        drives = self.client_fm.get_fc_drives()
        assert ' '.join(drives) == '/dev/sg4'

    @mock.patch('valence.functional_models.UbuntuClient.sg_map')
    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_get_fc_drives_fail(self, mock_run, mock_sg_map):
        mock_run_output = ssh.RunCommandOutput(
            stdout='host7', stderr='mock std err', return_code=0)
        mock_run.return_value = mock_run_output
        mock_map_output = ssh.RunCommandOutput(
            stdout='mock std out', stderr='mock std err', return_code=0)
        mock_sg_map.return_value = mock_map_output
        with pytest.raises(functional_models._errors.Error):
            self.client_fm.get_fc_drives()

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_start_kernel_trace(self, mock_run):
        trace_dir = '/sys/kernel/debug/tracing'
        mock_run_output = ssh.RunCommandOutput(
            stdout='scsi\nscsi_eh_wakeup', stderr='mock std err', return_code=0)
        mock_run.return_value = mock_run_output
        self.client_fm.start_kernel_trace(event='scsi', event_type='scsi_eh_wakeup')
        mock_run.assert_any_call('ls {}/events'.format(trace_dir), raise_on_failure=True)
        mock_run.assert_any_call('ls {}/events/scsi'.format(trace_dir), raise_on_failure=True)
        mock_run.assert_any_call('echo 4096 > {}/buffer_size_kb'.format(trace_dir),
                                 raise_on_failure=True)
        mock_run.assert_any_call('echo scsi:scsi_eh_wakeup > {}/set_event'.format(trace_dir),
                                 raise_on_failure=True)
        mock_run.assert_any_call('echo 1 > {}/tracing_on'.format(trace_dir), raise_on_failure=True)
        mock_run.assert_any_call('cat {}/trace_pipe > /tmp/trace_data &'.format(trace_dir),
                                 raise_on_failure=True)

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_start_kernel_trace_fail(self, mock_run):
        with pytest.raises(functional_models._errors.Error):
            mock_run_output = ssh.RunCommandOutput(
                stdout='test', stderr='mock std err', return_code=0)
            mock_run.return_value = mock_run_output
            self.client_fm.start_kernel_trace(event='scsi')

        with pytest.raises(functional_models._errors.Error):
            mock_run_output = ssh.RunCommandOutput(
                stdout='scsi', stderr='mock std err', return_code=0)
            mock_run.return_value = mock_run_output
            self.client_fm.start_kernel_trace(event='scsi', event_type='fieldnotexist')

    @mock.patch('valence.functional_models.UbuntuClient.run_command')
    def test_stop_kernel_trace(self, mock_run):
        trace_dir = '/sys/kernel/debug/tracing'
        mock_run_output = ssh.RunCommandOutput(
            stdout='test', stderr='mock std err', return_code=0)
        mock_run.return_value = mock_run_output
        self.client_fm.stop_kernel_trace()
        mock_run.assert_any_call('echo " " > {}/set_event'.format(trace_dir), raise_on_failure=True)
        mock_run.assert_any_call('echo 0 > {}/tracing_on'.format(trace_dir), raise_on_failure=True)


class TestCentOSClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        client_dm = models.ResourceClient.get_mock_object(overrides={'os': at2_enums.ClientOS.centos_7})
        cls.client_fm = functional_models.CentOSClient(client_dm)
        cls.portal_ip = next(utils.generate_ips(1))

    @mock.patch.object(functional_models.CentOSClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.CentOSClient, 'run_command')
    def test_update_iscsi_d_conf(self, mock_run_command, mock_put):
        with open('{}/iscsid.conf'.format(DATA_DIR)) as iscsid_file:
            iscsid_contents = iscsid_file.read()
        output = ssh.RunCommandOutput(stdout=iscsid_contents, stderr='', return_code=0)
        mock_run_command.return_value = output
        self.client_fm.update_iscsi_d_conf('harry', 'magicdude', 'potter', double_authentication=True)
        modified_iscsid = ''.join(mock_put.call_args_list[0][0][0])
        assert iscsid_contents != modified_iscsid
        assert 'magicdude' in modified_iscsid

    @mock.patch.object(functional_models.CentOSClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.CentOSClient, '_get_client_iscsid_conf')
    @mock.patch.object(functional_models.CentOSClient, '_iscsi_set_targets_and_discovery_info')
    def test_update_iscsi_d_conf_no_double_auth(self, mock_set_targets, mock_get_iscsid_conf, mock_put):
        mock_get_iscsid_conf.return_value = 'dummyconf'
        self.client_fm.update_iscsi_d_conf('harry', 'magicdude', 'potter')
        mock_set_targets.assert_called_with('dummyconf', 'harry', 'magicdude', None)

    @mock.patch.object(functional_models.CentOSClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.CentOSClient, '_get_client_iscsid_conf')
    @mock.patch.object(functional_models.CentOSClient, '_iscsi_set_targets_and_discovery_info')
    def test_update_iscsi_d_conf_double_auth(self, mock_set_targets, mock_get_iscsid_conf, mock_put):
        mock_get_iscsid_conf.return_value = 'dummyconf'
        self.client_fm.update_iscsi_d_conf('harry', 'magicdude', 'potter', double_authentication=True)
        mock_set_targets.assert_called_with('dummyconf', 'harry', 'magicdude', 'potter')

    def test_update_iscsi_d_conf_double_auth_same_password(self):
        with pytest.raises(functional_models._errors.FunctionalClientError):
            self.client_fm.update_iscsi_d_conf('harry', 'potter', 'potter', double_authentication=True)

    def test_deactivate_target_chap(self):
        with open('{}/iscsid.conf'.format(DATA_DIR)) as iscsid_file:
            iscsid_lines = iscsid_file.read().splitlines(True)
        for i, line in enumerate(iscsid_lines):
            if line.startswith('#discovery.sendtargets.auth.username_in'):
                iscsid_lines[i] = line.replace('#', '')
            elif line.startswith('#discovery.sendtargets.auth.password_in'):
                iscsid_lines[i] = line.replace('#', '')
            elif line.startswith('#node.session.auth.username_in'):
                iscsid_lines[i] = line.replace('#', '')
            elif line.startswith('#node.session.auth.password_in'):
                iscsid_lines[i] = line.replace('#', '')
        new_iscsid_lines = self.client_fm._iscsi_set_targets_and_discovery_info(iscsid_lines, 'harry',
                                                                                'potter')
        assert new_iscsid_lines[47].startswith('#node.session.auth.username_in')
        assert new_iscsid_lines[48].startswith('#node.session.auth.password_in')
        assert new_iscsid_lines[61].startswith('#discovery.sendtargets.auth.username_in')
        assert new_iscsid_lines[62].startswith('#discovery.sendtargets.auth.password_in')

    @mock.patch.object(functional_models.CentOSClient, '_put_iscsid_conf')
    @mock.patch.object(functional_models.CentOSClient, '_get_client_iscsid_conf')
    def test_deactivate_chap(self, mock_get_iscsid_conf, mock_put):
        with open('{}/iscsid.conf'.format(DATA_DIR)) as iscsid_file:
            iscsid_contents = iscsid_file.read()
        iscsid_lines = iscsid_contents.splitlines(True)
        assert not iscsid_lines[38].startswith('#CHAP_DISABLED:')
        mock_get_iscsid_conf.return_value = iscsid_lines
        self.client_fm.deactivate_chap()
        modified_iscsid_lines = mock_put.call_args_list[-1][0][0]
        assert modified_iscsid_lines[38].startswith('#CHAP_DISABLED:')

        mock_get_iscsid_conf.return_value = modified_iscsid_lines
        self.client_fm.reactivate_chap()
        modified_iscsid_lines = mock_put.call_args_list[-1][0][0]
        assert not modified_iscsid_lines[38].startswith('#CHAP_DISABLED:')

    @mock.patch('sf_ssh.Ssh')
    def test_put_iscsi_d_conf(self, mock_ssh):
        self.client_fm._put_iscsid_conf(['line1\n', 'line2\n'])
        mock_ssh.return_value.__enter__.return_value.put.assert_called()

    @mock.patch('valence.functional_models._fm_utils.BlockPoll.start_polling')
    def test_wait_until_iscsid_down(self, mock_start_polling):
        self.client_fm.wait_until_iscsid('down')

    @mock.patch('valence.functional_models._fm_utils.BlockPoll.start_polling')
    def test_wait_until_iscsid_up(self, mock_start_polling):
        self.client_fm.wait_until_iscsid('up')

    @mock.patch('valence.functional_models._fm_utils.BlockPoll.start_polling')
    def test_wait_until_iscsid_error(self, mock_start_polling):
        mock_start_polling.side_effect = functional_models._errors.PollMaxError
        with pytest.raises(functional_models._errors.IscsiError):
            self.client_fm.wait_until_iscsid('up')

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_iscsid_start(self, mock_run_command_check_output):
        self.client_fm.iscsid_start(restart=False)
        mock_run_command_check_output.assert_called_with('{};{};{}'.format(
            'systemctl start iscsid.socket',
            'systemctl start iscsi.service',
            'systemctl start iscsid.service'))

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    def test_get_pid_of_iscsid(self, mock_run_command):
        mock_run_command.return_value = ssh.RunCommandOutput('7172 7171', 'stderr', 0)
        pid_list = self.client_fm.get_pid_of_iscsid()
        assert pid_list == [7172, 7171]

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    def test_get_pid_of_iscsid_empty(self, mock_run_command):
        mock_run_command.return_value = ssh.RunCommandOutput('stdout', 'stderr', 1)
        pid_list = self.client_fm.get_pid_of_iscsid()
        assert pid_list == []

    def test_login_all_invalid_mode(self):
        with pytest.raises(functional_models._errors.IscsiError):
            self.client_fm.login_all(mode='invalid_mode')

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_login_all_mode_all(self, mock_run_command_check_output):
        self.client_fm.login_all(mode=enums.IscsiMode.all)

    @mock.patch('valence.functional_models.CentOSClient.login_all_batch_threaded')
    @mock.patch('valence.functional_models.CentOSClient.login_all_batch')
    def test_login_all_mode_batch_with_exception(self, mock_batch_login, mock_threaded_batch_login):
        mock_threaded_batch_login.side_effect = Exception
        self.client_fm.login_all(mode=enums.IscsiMode.batch)

    @mock.patch('valence.functional_models.CentOSClient.run_batch_login')
    @mock.patch('valence.functional_models.CentOSClient._discover')
    def test_login_all_batch(self, mock_discover, mock_run_batch_login):
        mock_discover.return_value = ['iqn.2010-01.com.solidfire:p0jq.vol1.1',
                                      'iqn.2010-01.com.solidfire:p0jq.vol2.2',
                                      'iqn.2010-01.com.solidfire:p0jq.vol3.3']
        mock_run_batch_login.return_value = '1'
        batch_return = self.client_fm.login_all_batch(self.portal_ip, 1)
        assert batch_return == ['1', '1', '1']

    @mock.patch('valence.functional_models.CentOSClient.run_batch_login')
    @mock.patch('valence.functional_models.CentOSClient._discover')
    def test_login_all_batch_threaded(self, mock_discover, mock_run_batch_login):
        mock_discover.return_value = ['iqn.2010-01.com.solidfire:p0jq.vol1.1',
                                      'iqn.2010-01.com.solidfire:p0jq.vol2.2',
                                      'iqn.2010-01.com.solidfire:p0jq.vol3.3',
                                      'iqn.2010-01.com.solidfire:p0jq.morevols-0199.405']
        mock_run_batch_login.return_value = '1'
        batch_return = self.client_fm.login_all_batch_threaded(self.portal_ip, 2)
        assert len(batch_return) == 2
        assert type(batch_return[0]) == threading.Thread

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_run_batch_login(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = 'login batches'
        result = self.client_fm.run_batch_login(['yeah', 'batches'], self.portal_ip)
        assert result == 'login batches'

    @mock.patch('valence.functional_models.CentOSClient._list_disks_by_id')
    @mock.patch('valence.functional_models.CentOSClient._get_target_map')
    def test_get_disk_map(self, mock_get_target_map, mock_list_disks_by_id):
        mock_get_target_map.return_value = {'1': 'iqn.2010-01.com.solidfire:p0jq.vol1.1',
                                            '10': 'iqn.2010-01.com.solidfire:p0jq.extravols-0004.10',
                                            '100': 'iqn.2010-01.com.solidfire:p0jq.extravols-0094.100'}
        mock_list_disks_by_id.return_value = {'1': '/dev/sdht',
                                              '10': '/dev/sdfg',
                                              '100': '/dev/sdkr'}
        disk_map = self.client_fm.get_disk_map()
        assert disk_map == {'1': '/dev/sdht', '10': '/dev/sdfg', '100': '/dev/sdkr'}

    @mock.patch('time.sleep')
    @mock.patch('valence.functional_models.CentOSClient.run_command')
    @mock.patch('valence.functional_models.CentOSClient._list_disks_by_id')
    @mock.patch('valence.functional_models.CentOSClient._get_target_map')
    def test_get_disk_map_fail(self, mock_get_target_map, mock_list_disks_by_id,
                               mock_run_command, mock_sleep):
        mock_get_target_map.return_value = {'1': 'iqn.2010-01.com.solidfire:p0jq.vol1.1'}
        mock_list_disks_by_id.return_value = {'1': '/dev/sdht',
                                              '10': '/dev/sdfg'}
        with pytest.raises(functional_models._errors.IscsiError):
            disk_map = self.client_fm.get_disk_map()

    @mock.patch('valence.functional_models.CentOSClient._node_info')
    def test_get_target_map(self, mock_node_info):
        mock_node_info.return_value = \
            ['10.117.94.233:3260,1 iqn.2010-01.com.solidfire:p0jq.morevols-0181.387',
             '10.117.94.233:3260,1 iqn.2010-01.com.solidfire:p0jq.extravols-0064.70',
             '10.117.94.233:3260,1 iqn.2010-01.com.solidfire:p0jq.morevols-0124.330']
        target_map = self.client_fm._get_target_map()
        assert target_map == {'70': 'iqn.2010-01.com.solidfire:p0jq.extravols-0064.70',
                              '330': 'iqn.2010-01.com.solidfire:p0jq.morevols-0124.330',
                              '387': 'iqn.2010-01.com.solidfire:p0jq.morevols-0181.387'}

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_node_info(self, mock_run_command_check_output, mock_run_command):
        mock_run_command_check_output.return_value = \
            ('10.117.94.233:3260,1'
             'iqn.2010-01.com.solidfire:p0jq.morevols-0181.387\n'
             '10.117.94.233:3260,1'
             'iqn.2010-01.com.solidfire:p0jq.extravols-0064.70\n')
        mock_run_command.return_value = ssh.RunCommandOutput('clabv0012\n', 'stderr', 0)
        node_info = self.client_fm._node_info()
        assert node_info == ['10.117.94.233:3260,1iqn.2010-01.com.solidfire:p0jq.morevols-0181.387',
                             '10.117.94.233:3260,1iqn.2010-01.com.solidfire:p0jq.extravols-0064.70']

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_node_info_skip_boot(self, mock_run_command_check_output, mock_run_command):
        mock_run_command_check_output.return_value = 'clabv0012'
        mock_run_command.return_value = ssh.RunCommandOutput('clabv0012', 'stderr', 0)
        node_info = self.client_fm._node_info()

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    def test_list_disks_by_id(self, mock_run_command):
        disk_stdout = \
            ('lrwxrwxrwx 1 root root 10 Nov 27 20:18 '
             'ip-10.117.94.233:3260-iscsi-iqn.2010-01.com.solidfire:p0jq.extravols-0000.6-lun-0'
             ' -> ../../sdcm\n'
             'lrwxrwxrwx 1 root root 10 Nov 27 20:18 '
             'ip-10.117.94.233:3260-iscsi-iqn.2010-01.com.solidfire:p0jq.extravols-0001.7-lun-0'
             ' -> ../../sdbc\n')
        mock_run_command.return_value = ssh.RunCommandOutput(disk_stdout, 'stderr', 0)
        disk_map = self.client_fm._list_disks_by_id()
        assert disk_map == {'7': '/dev/sdbc', '6': '/dev/sdcm'}

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_discovery(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = ('10.117.94.233:3260,1 '
                                                      'iqn.2010-01.com.solidfire:p0jq.vol1.1\n'
                                                      '10.117.94.233:3260,1 '
                                                      'iqn.2010-01.com.solidfire:p0jq.vol2.2\n')
        targets = self.client_fm._discover(self.portal_ip)
        assert targets == ['iqn.2010-01.com.solidfire:p0jq.vol1.1',
                           'iqn.2010-01.com.solidfire:p0jq.vol2.2']

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    def test_run_command_check_output(self, mock_run_command):
        mock_run_command.return_value = ssh.RunCommandOutput('command success', 'stderr', 0)
        cmd_stdout = self.client_fm._run_command_check_output('real command')
        assert cmd_stdout == 'command success'

    @mock.patch('time.sleep')
    @mock.patch('valence.functional_models.CentOSClient.run_command')
    def test_run_command_check_output_FAIL(self, mock_run_command, mock_sleep):
        mock_run_command.return_value = ssh.RunCommandOutput('command success', 'stderr', 1)
        with pytest.raises(functional_models._errors.IscsiError):
            cmd_stdout = self.client_fm._run_command_check_output('real command')

    @mock.patch('valence.functional_models.CentOSClient.get_disk_map')
    @mock.patch('valence.functional_models.CentOSClient.login_all')
    @mock.patch('valence.functional_models.CentOSClient._discover')
    @mock.patch('valence.functional_models.CentOSClient.iscsid_start')
    def test_volume_ceremony(self, mock_iscsid_start, mock_discover, mock_login_all, mock_get_disk_map):
        mock_get_disk_map.return_value = {'1': '/dev/sdba', '2': '/dev/sdcm'}
        disk_map = self.client_fm.volume_ceremony(self.portal_ip)
        assert disk_map == {'1': '/dev/sdba', '2': '/dev/sdcm'}

    @mock.patch('valence.functional_models.CentOSClient.get_disk_map')
    @mock.patch('valence.functional_models.CentOSClient.login_all')
    @mock.patch('valence.functional_models.CentOSClient._discover')
    @mock.patch('valence.functional_models.CentOSClient.purge_restart')
    def test_volume_ceremony_purge(self, mock_purge_restart, mock_discover,
                                   mock_login_all, mock_get_disk_map):
        mock_get_disk_map.return_value = {'1': '/dev/sdba', '2': '/dev/sdcm'}
        disk_map = self.client_fm.volume_ceremony(portal_ip=self.portal_ip, purge=True)
        assert disk_map == {'1': '/dev/sdba', '2': '/dev/sdcm'}

    @mock.patch('valence.functional_models.CentOSClient.iscsid_start')
    @mock.patch('valence.functional_models.CentOSClient.purge_credentials')
    @mock.patch('valence.functional_models.CentOSClient.iscsid_stop')
    @mock.patch('valence.functional_models.CentOSClient.logout_all')
    def test_purge_restart(self, mock_logout_all, mock_iscsid_stop,
                           mock_purge_credentials, mock_iscsid_start):
        self.client_fm.purge_restart()

    @mock.patch('valence.functional_models.CentOSClient.iscsid_start')
    @mock.patch('valence.functional_models.CentOSClient.purge_credentials')
    @mock.patch('valence.functional_models.CentOSClient.iscsid_stop')
    @mock.patch('valence.functional_models.CentOSClient.logout_all')
    def test_purge_restart_exception(self, mock_logout_all, mock_iscsid_stop,
                                     mock_purge_credentials, mock_iscsid_start):
        mock_logout_all.side_effect = functional_models._errors.IscsiError
        self.client_fm.purge_restart()

    @mock.patch('valence.functional_models.CentOSClient._check_iscsi_initramfs')
    @mock.patch('valence.functional_models.CentOSClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_iscsid_stop(self, mock_run_command_check_output, mock_wait_until_iscsid,
                         mock_check_iscsi_initramfs):
        self.client_fm.iscsid_stop()
        mock_check_iscsi_initramfs.assert_called_with(iscsi_root_volume=False)
        mock_run_command_check_output.assert_called_with('{};{};{}'.format(
            'systemctl stop iscsi.service',
            'systemctl stop iscsid.socket',
            'systemctl stop iscsid.service'))
        mock_wait_until_iscsid.assert_called_with('down')

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_logout_all(self, mock_run_command_check_output):
        self.client_fm.logout_all()

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_logout_all_svip(self, mock_run_command_check_output):
        self.client_fm.logout_all(self.portal_ip)

    @mock.patch('valence.functional_models.CentOSClient._check_iscsi_initramfs')
    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_purge_credentials(self, mock_run_command_check_output, mock_check_iscsi_initramfs):
        self.client_fm.purge_credentials()

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_check_iscsi_initramfs(self, mock_run_command_check_output):
        self.client_fm._check_iscsi_initramfs(iscsi_root_volume=False)

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_check_iscsi_initramfs_root_true(self, mock_run_command_check_output):
        self.client_fm._check_iscsi_initramfs(iscsi_root_volume=True)

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_get_initiator_name(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = \
            'InitiatorName=iqn.2010-01.net.solidfire.eng:clabv0006'
        initiator_name = self.client_fm.get_initiator_name()
        assert initiator_name == 'iqn.2010-01.net.solidfire.eng:clabv0006'

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_get_initiator_name_fails(self, mock_run_command_check_output):
        mock_run_command_check_output.return_value = \
            'InitiatorNameFAKE=iqn.2010-01.net.solidfire.eng:clabv0006'
        with pytest.raises(functional_models._errors.IscsiError):
            initiator_name = self.client_fm.get_initiator_name()

    @mock.patch('valence.functional_models.CentOSClient.run_command')
    def test_run_command_wrappers(self, mock_run):
        # This list of methods all share the same test
        methods = [
            'sg_raw',
            'sginfo',
            'sg_vpd',
            'sg_persist',
            'sg_read',
            'multipath'
        ]
        for method in methods:
            # Test a successful call
            mock_run_output = ssh.RunCommandOutput(
                stdout='mock std out', stderr='mock std err', return_code=0)
            mock_run.return_value = mock_run_output
            output = getattr(self.client_fm, method)('mock options')
            assert output.stdout == 'mock std out'
            assert output.stderr == 'mock std err'
            assert output.return_code == 0

            # Test a failed call
            mock_run_output = ssh.RunCommandOutput(
                stdout='mock std out', stderr='mock std err', return_code=1)
            mock_run.return_value = mock_run_output
            output = getattr(self.client_fm, method)('mock options', 1)
            assert output.stdout == 'mock std out'
            assert output.stderr == 'mock std err'
            assert output.return_code == 1

            # Test an exception for an unexpected return code
            mock_run_output = ssh.RunCommandOutput(
                stdout='mock std out', stderr='mock std err', return_code=1)
            mock_run.return_value = mock_run_output
            with pytest.raises(functional_models._errors.FunctionalClientError):
                getattr(self.client_fm, method)('mock options', 0)


class TestCentOS6Client(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        client_dm = models.ResourceClient.get_mock_object(overrides={'os': at2_enums.ClientOS.centos_6})
        cls.client_fm = functional_models.CentOSClient(client_dm)
        cls.portal_ip = next(utils.generate_ips(1))

    @mock.patch('valence.functional_models.CentOSClient._check_iscsi_initramfs')
    @mock.patch('valence.functional_models.CentOSClient.wait_until_iscsid')
    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_iscsid_stop(self, mock_run_command_check_output, mock_wait_until_iscsid,
                         mock_check_iscsi_initramfs):
        self.client_fm.iscsid_stop()
        mock_check_iscsi_initramfs.assert_called_with(iscsi_root_volume=False)
        mock_run_command_check_output.assert_called_with('/etc/rc.d/init.d/iscsid stop')
        mock_wait_until_iscsid.assert_called_with('down')

    @mock.patch('valence.functional_models.CentOSClient._run_command_check_output')
    def test_iscsid_start(self, mock_run_command_check_output):
        self.client_fm.iscsid_start(restart=False)
        mock_run_command_check_output.assert_called_with('/etc/rc.d/init.d/iscsid start')
