"""Unit tests for the monitor decorators."""
import datetime
import sys
import unittest

import mock
import six

from valence import functional_models


class MockBackendMonitor(object):
    def notify(self, method_call):
        param_str = ''
        if 'self' in method_call.arguments:
            method_call.arguments.pop('self')
        if 'args' in method_call.arguments:
            method_call.arguments['args'] = tuple(list(method_call.arguments['args'][1:]))
        for key in method_call.arguments:
            param_str += '{}: {} '.format(key, method_call.arguments[key])
        print('Name: {}, Params: {}'.format(method_call.method.__name__, param_str))

    def is_running(self):
        return True


class MockMethods(object):
    def __init__(self, value):
        self.mock_value = value

    @property
    def static_id(self):
        return 'fake_id'

    @functional_models.notify_monitor
    def mock_multiply_func(self, arg1, arg2=5):
        return arg1 * arg2 * self.mock_value

    @functional_models.notify_monitor
    def mock_add_func(self, *args):
        return sum([self.mock_value] + list(args))

    @functional_models.notify_monitor
    def mock_kwargs_func(self, **kwargs):
        return_string = ''
        for key in kwargs:
            return_string += (kwargs[key] + ' ')
        return_string += str(self.mock_value)


class TestNoBackendMonitor(unittest.TestCase):
    def setUp(self):
        self.captured_output = six.StringIO()
        sys.stdout = self.captured_output
        self.methods = MockMethods(value=10)

    def test_no_backend_monitor(self):
        return_value = self.methods.mock_multiply_func(2)
        assert return_value == 100
        assert self.captured_output.getvalue() == ''


class TestNotifyMonitor(unittest.TestCase):
    def setUp(self):
        self.captured_output = six.StringIO()
        sys.stdout = self.captured_output
        self.mock_backend_monitor = MockBackendMonitor()
        functional_models.monitor_decorators.set_backend_monitor(self.mock_backend_monitor)
        self.methods = MockMethods(value=5)

    def test_notify_monitor_multiply(self):
        return_value = self.methods.mock_multiply_func(2)
        assert return_value == 50
        assert self.captured_output.getvalue() == 'Name: mock_multiply_func, Params: arg1: 2 arg2: 5 \n'

    def test_notify_monitor_args(self):
        return_value = self.methods.mock_add_func(2, 4, 6, 8)
        assert return_value == 25
        assert self.captured_output.getvalue() == 'Name: mock_add_func, Params: args: (2, 4, 6, 8) \n'

    def test_notify_monitor_kwargs(self):
        return_value = self.methods.mock_kwargs_func(monitor_lizards='breath fire')
        assert self.captured_output.getvalue() == ('Name: mock_kwargs_func,'
                                                   ' Params: monitor_lizards: breath fire \n')

    def tearDown(self):
        functional_models.monitor_decorators.remove_backend_monitor()


class TestMethodCall(unittest.TestCase):
    def setUp(self):
        self.captured_output = six.StringIO()
        sys.stdout = self.captured_output
        self.mock_backend_monitor = MockBackendMonitor()
        functional_models.monitor_decorators.set_backend_monitor(self.mock_backend_monitor)
        self.methods = MockMethods(value=5)

    def test_is_method(self):
        params = {'arg1': 2}
        mc = functional_models.MethodCall(self.methods.mock_multiply_func,
                                          params, datetime.datetime.now())
        assert 'mock_multiply' in str(mc)
        assert mc.is_method(MockMethods.mock_multiply_func)

    def test_static_id_property(self):
        mc = functional_models.MethodCall(self.methods.mock_multiply_func,
                                          {}, datetime.datetime.now())
        mc.arguments['self'] = self.methods
        assert mc.static_id is not None

    def tearDown(self):
        functional_models.monitor_decorators.remove_backend_monitor()
