"""Valence models package.

Overview
=========
Remember to look at the source code for the most details.

All the data model classes are imported into the root namespace of the models package.

Imports should be done like this::

    from valence import models
    node_dm = models.ResourceNode(node_data)


Creation
==============
Data models can be created a few different ways::

    # Using the create method
    volume_dm = models.Volume.create(name="abc")

    # From a dictionary
    volume_data = {"name": "abc"}
    volume_dm = models.Volume(volume_data)

    # Using the mock method (note: this is a somewhat expensive call, so be careful not to get too
    # carried away with mock when writing unit tests. If create() is sufficient, use that instead.)
    volume_dm = models.Volume.get_mock_object()

    # Generate a list of 5 mock data models
    volume_dms = models.Volume.get_mock_object_list(5)


Serialization
==============
Data models can be serialized to a dict like this::

    node_data = node_dm.serialize()

.. note::

    The serialized keys use the camelCase naming convention to be consistent with SolidFire. However, the
    attributes of the data models use snake_case to be consistent with python's naming conventions.

Validation
==============
Data models can be validated like this::

    # Standard validation (against a specific sf version)
    version = sf_api.SfVersion("9.0.0.0")
    node_dm.validate(version)

    # Don't enforce required fields
    node_dm.validate(version, partial=True)

    # Complain about unrecognized keys
    node_dm.validate(version, strict=True)

A `valence.models.DataValidationError` exception is raised if validation fails.

See :ref:`the models guide<models-int>` for more details.
"""
from sf_models import GenericVersion

from camelot.models import (
    ArtifactCredentials,
    BaseSuiteResource,
    CamelotLogFilter,
    ClientSuiteResource,
    ClusterSuiteConfig,
    ClusterSuiteResource,
    ComputeNodeSuiteResource,
    HipChatAPI,
    HypervisorSuiteResource,
    IPRangeSuiteResource,
    IpSuiteResource,
    ManagementNodeSuiteConfig,
    ManagementNodeSuiteResource,
    NetworkSuiteResource,
    NodeSuiteConfig,
    NodeSuiteResource,
    OntapSuiteResource,
    RepoChangeset,
    ResourceOntap,
    SuiteConfig,
    SuiteResources,
    VirtualNetworkSuiteResource,
    WitnessResource,
)
from sf_api.models import *  # pylint: disable=wildcard-import
# Need to explicitly import types to avoid pylint errors
# Import all the models from sf_api since they are also part of the Valence public API
from sf_api.models import types
from sf_api.sf_version import SfVersion
from valence.models._async import AsyncHandle
from valence.models._bourbon import (
    PortChannel,
    SwitchPort,
)
from valence.models._errors import (
    DataModelError,
    DataValidationError,
    Error,
    ResourceError,
)
from valence.models._external_key_server import (
    CertificateSubject,
    SafenetClientCertificate,
    SafenetLocalCertificateAuthority,
    SafenetNaeServer,
    SafenetService,
)
from valence.models._hypervisor import (
    HypervisorHost,
    VirtualHardwareProfile,
)
from valence.models._raw_stats import (
    DiskRawStats,
    RawStats,
    ServiceRawStats,
    VolumeRawStats,
)
from valence.models._reports import (
    BinAssignments,
    BinAssignmentsProperties,
    BinReport,
    BinService,
    ProtectionDomainServiceInfo,
    ProtectionDomainServices,
    SliceBalanceReport,
    SliceReport,
    SliceService,
    StandbyServiceInfo,
)
from valence.models._resource import (
    get_resource_cluster_by_static_id,
    get_resource_clusters,
    get_resource_node_by_static_id,
    get_resource_nodes,
    set_cluster_resources,
    set_node_resources,
)
from valence.models._resource_cluster import Cluster
from valence.models._resource_network import RtfiNetworkProfileAdapter
from valence.models._resource_node import (
    ActiveNode,
    AvailableNode,
    LegacyNodeAdapter,
    NodeBase,
    PendingActiveNode,
    PendingNode,
    RtfiNodeAdapter,
)
