"""GetRawStats API data model.

NetApp engineering uses the GetRawStats API method to troubleshoot new features.
The data returned from GetRawStats is not documented and does not conform to typical
data models.
"""

import autologger
import six

from sf_api import (
    api_constants,
    models,
)
from sf_api.models import (
    errors,
    types,
)

_log = autologger.get_sflogging()


class RawStats(models.Base):
    """Cluster GetRawStats API data model.

    Corresponds to the elements returned by get_raw_stats().
    NetApp engineering uses the GetRawStats API method to troubleshoot new features.
    The data returned from GetRawStats is not documented and does not conform to typical
    data models.

    :param dict raw_data: The raw stats data.
    """
    NAME_ID = 'name_id'
    result = types.DictType(types.BaseType, required=True, serialized_name=api_constants.RESULT)

    def __init__(self, raw_data):
        super(RawStats, self).__init__(raw_data=raw_data)
        self._disk_stats_dms_map = {}
        self._volume_stats_dms_map = {}
        self._service_stats_dms_map = {}
        self._initialize_map_for_each_stats()
        self._volume_hist_dms_map = {}

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            api_constants.RESULT, self.result,
        )

    def _initialize_map_for_each_stats(self):
        for name_id, stats in six.iteritems(self.result):
            stats[self.NAME_ID] = name_id
            int_id = int(name_id.split('-')[1])
            if name_id.startswith('disk'):
                stats['disk_id'] = int_id
                self._disk_stats_dms_map[int_id] = DiskRawStats(stats)
            elif name_id.startswith('service'):
                stats['service_id'] = int_id
                self._service_stats_dms_map[int_id] = ServiceRawStats(stats)
            elif name_id.startswith('volume'):
                stats['volume_id'] = int_id
                self._volume_stats_dms_map[int_id] = VolumeRawStats(stats)

    def get_volume_ids(self):
        """Returns the list of volume ids in the raw stats.

        :return: List of ids
        :rtype: list[int]
        """
        return list(self._volume_stats_dms_map.keys())

    def get_disk_ids(self):
        """Returns the list of disk ids in the raw stats.

        :return: List of ids
        :rtype: list[int]
        """
        return list(self._disk_stats_dms_map.keys())

    def get_service_ids(self):
        """Returns the list of service ids in the raw stats.

        :return: List of ids
        :rtype: list[int]
        """
        return list(self._service_stats_dms_map.keys())

    def get_disk_stats(self, disk_ids=None):
        """Returns the disk stats data model.

        :param list[int] disk_ids: List of ids for disk stats to return. None to return
            for all disks.
        :rtype: list[valence.models.DiskRawStats]
        """
        # filter by key
        filtered_disks = []
        if disk_ids:
            filtered_disks = [self._disk_stats_dms_map[key] for key in disk_ids
                              if key in self._disk_stats_dms_map]

        return filtered_disks if disk_ids else list(self._disk_stats_dms_map.values())

    def get_service_stats(self, service_ids=None):
        """Returns the service stats data model.

        :param list[int] service_ids: List of ids for service stats to return. None to return
            for all services.
        :rtype: list[valence.models.ServiceRawStats]
        """
        # filter by key
        filtered_dms = []
        if service_ids:
            filtered_dms = [self._service_stats_dms_map[key] for key in service_ids
                            if key in self._service_stats_dms_map]

        return filtered_dms if service_ids else list(self._service_stats_dms_map.values())

    def get_volume_stats(self, volume_ids=None):
        """Returns the volume stats data model.

        :param list[int] volume_ids: List of ids for volume stats to return. None to return
            for all volumes.
        :rtype: list[valence.models.VolumeRawStats]
        """
        # filter by key
        filtered_dms = []
        if volume_ids:
            filtered_dms = [self._volume_stats_dms_map[key] for key in volume_ids
                            if key in self._volume_stats_dms_map]

        return filtered_dms if volume_ids else list(self._volume_stats_dms_map.values())

    def generate_ssload_histogram(self):
        """Generates the ssLoad histogram from the service raw stats data.

        :return: QuintileHistogram dm of ssLoad from raw stats.
        :rtype: valence.models.QuintileHistogram
        """
        ssload_histogram = None
        for stats in self.get_service_stats():
            # search for the one service stats containing the ssLoad data
            if stats.ss_load_0 is not None:
                ssload_histogram = self._convert_to_quintile_histogram(stats, 'ss_load')
                break
        return ssload_histogram

    def generate_volume_qos_histograms(self, volume_ids=None):
        """Generates the volume QoS histograms from the raw stats data.

        :param list[int] volume_ids: Selected volume ids or None for all volumes.
        :return: Map of volume_id and VolumeQosHistogram dm
        :rtype: dict[int, valence.models.VolumeQosHistograms]
        """
        if not self._volume_hist_dms_map:
            self._volume_hist_dms_map = {}
            for stats in self.get_volume_stats(volume_ids):
                qos_histograms = models.VolumeQosHistograms()
                qos_histograms.volume_id = stats.volume_id
                histogram = models.Histograms()
                histogram.min_to_max_iops_percentages = self._convert_to_quintile_histogram(
                    stats, 'min_to_max')
                histogram.target_utilization_percentages = self._convert_to_quintile_histogram(
                    stats, 'target_utilization')
                histogram.throttle_percentages = self._convert_to_quintile_histogram(
                    stats, 'throttle')
                histogram.below_min_iops_percentages = self._convert_to_quintile_histogram(
                    stats, 'below_min')
                histogram.write_block_sizes = self._convert_to_block_size_histogram(
                    stats, 'write_block_size')
                histogram.read_block_sizes = self._convert_to_block_size_histogram(
                    stats, 'read_block_size')
                qos_histograms.histograms = histogram
                self._volume_hist_dms_map[stats.volume_id] = qos_histograms

        return self._volume_hist_dms_map

    @staticmethod
    def _convert_to_quintile_histogram(stats_dm, stats_prefix):
        """Converts the raw stats into a QuintileHistogram data model.

        :param dict|valence.models.ServiceRawStats|valence.models.VolumeRawStats stats_dm: Raw
            stats dm (i.e. valence.models.VolumeRawStats or ServiceRawStats).
        :param str stats_prefix: String prefix of selected stats field (i.e. 'throttle').
        :return: QuintileHistogram dm for fields in stats_dm that match the prefix
            (i.e. throttle_1_to_19).
        :rtype: valence.models.QuintileHistogram
        """
        quintile_hist = models.QuintileHistogram()
        quintile_hist.bucket_0 = getattr(stats_dm, '{}_0'.format(stats_prefix))
        quintile_hist.bucket_1_to_19 = getattr(stats_dm, '{}_1_to_19'.format(stats_prefix))
        quintile_hist.bucket_20_to_39 = getattr(stats_dm, '{}_20_to_39'.format(stats_prefix))
        quintile_hist.bucket_40_to_59 = getattr(stats_dm, '{}_40_to_59'.format(stats_prefix))
        quintile_hist.bucket_60_to_79 = getattr(stats_dm, '{}_60_to_79'.format(stats_prefix))
        quintile_hist.bucket_80_to_100 = getattr(stats_dm, '{}_80_to_100'.format(stats_prefix))
        quintile_hist.bucket_101_plus = getattr(stats_dm, '{}_101_plus'.format(stats_prefix))
        return quintile_hist

    @staticmethod
    def _convert_to_block_size_histogram(stats_dm, stats_prefix):
        """Converts the raw stats into a BlockSizeHistogram data model.

        :param valence.models.VolumeRawStats stats_dm: Raw stats dm for a volume.
        :param str stats_prefix: String prefix of selected stats field (i.e. 'read_block_size').
        :return: BlockSizeHistogram dm for fields in VolumeRawStats dm that match the prefix
            (i.e. read_block_size_4k).
        :rtype: valence.models.BlockSizeHistogram
        """
        block_size_hist = models.BlockSizeHistogram()
        block_size_hist.bucket_512_to_4095 = getattr(stats_dm, '{}_512'.format(stats_prefix))
        block_size_hist.bucket_4096_to_8191 = getattr(stats_dm, '{}_4k'.format(stats_prefix))
        block_size_hist.bucket_8192_to_16383 = getattr(stats_dm, '{}_8k'.format(stats_prefix))
        block_size_hist.bucket_16384_to_32767 = getattr(stats_dm, '{}_16k'.format(stats_prefix))
        block_size_hist.bucket_32768_to_65535 = getattr(stats_dm, '{}_32k'.format(stats_prefix))
        block_size_hist.bucket_65536_to_131071 = getattr(stats_dm, '{}_64k'.format(stats_prefix))
        block_size_hist.bucket_131072_plus = getattr(stats_dm, '{}_128k_plus'.format(stats_prefix))
        return block_size_hist

    def validate_ssload_histogram(self, min_ssload, max_ssload):
        """Validates the GetRawStats SSload histogram data against SSLoad data boundaries.

        Compares the GetRawStats SSLoad histogram data to validate that it is between or equal
        to previous(min_ssload) and current(max_ssload) SSLoad histogram data.

        :param valence.models.QuintileHistogram min_ssload: Previous values for SSLoad histogram
            to compare with rawStats.
        :param valence.models.QuintileHistogram max_ssload: Current values for SSLoad histogram
            to compare with rawStats.
        :raises sf_api.models.errors.DataValidationError:
        """
        raw_ssload = self.generate_ssload_histogram()
        error_msgs = []
        for bucket in [key for key in raw_ssload.keys() if raw_ssload[key]]:
            if min_ssload.get(bucket) is None or min_ssload[bucket] > raw_ssload[bucket]:
                error_msgs.append('Previous SSLoadHistogram {bucket}={min_val} should be less or '
                                  'equal to current value '
                                  '{bucket}={raw_val}'.format(bucket=bucket, min_val=min_ssload[bucket],
                                                              raw_val=raw_ssload[bucket]))

            if max_ssload.get(bucket) is None or max_ssload[bucket] < raw_ssload[bucket]:
                error_msgs.append('Current SSLoadHistogram {bucket}={max_val} should be more or '
                                  'equal to previous value '
                                  '{bucket}={raw_val}'.format(bucket=bucket, max_val=max_ssload[bucket],
                                                              raw_val=raw_ssload[bucket]))

        if error_msgs:
            error_msgs.append('Raw SSLoadHistogram:{}'.format(str(raw_ssload)))
            raise errors.DataValidationError({'SSLoadHistogram RawStats Validation': ', '
                                             .join(error_msgs)})

    def validate_volume_histogram(self, min_volume_hist, max_volume_hist):
        """Validates the GetRawStats Volume QoS Histogram data against volume data boundaries.

        Compares the GetRawStats volume qos histogram data to validate that it is between or equal
        to previous(min_volume_hist) and current(max_volume_hist) volume qos histogram data.

        :param valence.models.VolumeQosHistograms min_volume_hist: Previous values for volume qos
            histograms to compare with rawStats.
        :param valence.models.VolumeQosHistograms max_volume_hist: Current values for volume qos
            histograms to compare with rawStats.
        :raises sf_api.models.errors.DataValidationError:
        """
        volume_id = min_volume_hist.volume_id

        error_msgs = []
        raw_volume_hist = self.generate_volume_qos_histograms()[volume_id]
        for hist_name, raw_value in raw_volume_hist.histograms.items():
            for bucket in [key for key in raw_value.keys() if raw_value[key]]:
                # validate lower boundary for bucket in a histogram grouping
                min_hist = min_volume_hist.histograms[hist_name]
                if min_hist.get(bucket) is None or min_hist[bucket] > raw_value[bucket]:
                    error_msgs.append(
                        'Previous VolumeQosHistogram {hist}.{bucket}={min_val} should be less or '
                        'equal to current value '
                        '{bucket}={raw_val}'.format(hist=hist_name, bucket=bucket,
                                                    min_val=min_hist[bucket],
                                                    raw_val=raw_value[bucket]))
                # validate upper boundary for bucket in a histogram grouping
                max_hist = max_volume_hist.histograms[hist_name]
                if max_hist.get(bucket) is None or max_hist[bucket] < raw_value[bucket]:
                    error_msgs.append(
                        'Current VolumeQosHistogram {hist}.{bucket}={max_val} should be more or '
                        'equal to previous value '
                        '{bucket}={raw_val}'.format(hist=hist_name, bucket=bucket,
                                                    max_val=max_hist[bucket],
                                                    raw_val=raw_value[bucket]))

        if error_msgs:
            error_msgs.append('Raw VolumeQosHistogram:{}'.format(str(raw_volume_hist)))
            raise errors.DataValidationError({'VolumeQosHistogram RawStats Validation': ', '
                                             .join(error_msgs)})

    def is_valid_volume_histogram(self, min_volume_hist, max_volume_hist):
        """Validates the GetRawStats Volume QoS Histogram data against data boundaries.

        Compares the GetRawStats volume qos histogram data to validate that it is between or equal
        to previous(min_volume_hist) and current(max_volume_hist) volume qos histogram data.

        :param valence.models.VolumeQosHistograms min_volume_hist: Previous values for volume qos
            histograms to compare with rawStats.
        :param valence.models.VolumeQosHistograms max_volume_hist: Current values for volume qos
            histograms to compare with rawStats.
        :return: True if rawStats histogram data is between boundaries or False otherwise
        :rtype: bool
        """
        is_valid = True
        try:
            self.validate_volume_histogram(min_volume_hist=min_volume_hist,
                                           max_volume_hist=max_volume_hist)
        except errors.DataValidationError as ex:
            is_valid = False
            _log.debug(str(ex))
        return is_valid


class DiskRawStats(models.Base):
    """DiskRawStats data model.

    Corresponds to the elements returned by get_raw_stats().

    :param dict raw_data: The raw disk stats data.
    """
    name_id = types.StringType(required=False, serialized_name=RawStats.NAME_ID)
    disk_id = types.IntIdType(required=False, serialized_name=api_constants.DISK_ID)
    life_remaining_percent = types.IntType(required=True,
                                           serialized_name=api_constants.LIFE_REMAINING_PERCENT)
    lifetime_read_bytes = types.IntType(required=True,
                                        serialized_name=api_constants.LIFETIME_READ_BYTES)
    lifetime_write_bytes = types.IntType(required=True,
                                         serialized_name=api_constants.LIFETIME_WRITE_BYTES)
    power_on_hours = types.IntType(required=True,
                                   serialized_name=api_constants.POWER_ON_HOURS)
    reallocated_sectors = types.IntType(required=True,
                                        serialized_name=api_constants.REALLOCATED_SECTORS)
    reserve_capacity_percent = types.IntType(required=True,
                                             serialized_name=api_constants.RESERVE_CAPACITY_PERCENT)
    uncorrectable_errors = types.IntType(required=True,
                                         serialized_name=api_constants.UNCORRECTABLE_ERRORS)

    def __init__(self, raw_data):
        super(DiskRawStats, self).__init__(raw_data=raw_data)

    def __repr__(self):
        return ('{}<{}={}, {}={}, {}={}, {}={}, {}={}'
                '{}={}, {}={}, {}={}, {}={}>').format(
            self.__class__.__name__,
            RawStats.NAME_ID, self.name_id,
            api_constants.DISK_ID, self.disk_id,
            api_constants.LIFE_REMAINING_PERCENT, self.life_remaining_percent,
            api_constants.LIFETIME_READ_BYTES, self.lifetime_read_bytes,
            api_constants.LIFETIME_WRITE_BYTES, self.lifetime_write_bytes,
            api_constants.POWER_ON_HOURS, self.power_on_hours,
            api_constants.REALLOCATED_SECTORS, self.reallocated_sectors,
            api_constants.RESERVE_CAPACITY_PERCENT, self.reserve_capacity_percent,
            api_constants.UNCORRECTABLE_ERRORS, self.uncorrectable_errors,
        )


class VolumeRawStats(models.Base):
    """VolumeRawStats data model.

    Corresponds to the elements returned by get_raw_stats().

    :param dict raw_data: The raw volume stats data.
    """
    name_id = types.StringType(required=False, serialized_name=RawStats.NAME_ID)
    volume_id = types.IntIdType(required=False, serialized_name=api_constants.VOLUME_ID)
    actual_iops = types.IntType(required=False, serialized_name=api_constants.ACTUAL_IOPS)
    async_delay_usec = types.IntType(required=False, serialized_name='asyncDelayUSec')
    average_iop_size = types.IntType(required=False, serialized_name=api_constants.AVERAGE_IOP_SIZE)
    below_min_0 = types.IntType(required=False,
                                serialized_name=api_constants.BELOW_MIN_0)
    below_min_1_to_19 = types.IntType(required=False,
                                      serialized_name=api_constants.BELOW_MIN_1_TO_19)
    below_min_20_to_39 = types.IntType(required=False,
                                       serialized_name=api_constants.BELOW_MIN_20_TO_39)
    below_min_40_to_59 = types.IntType(required=False,
                                       serialized_name=api_constants.BELOW_MIN_40_TO_59)
    below_min_60_to_79 = types.IntType(required=False,
                                       serialized_name=api_constants.BELOW_MIN_60_TO_79)
    below_min_80_to_100 = types.IntType(required=False,
                                        serialized_name=api_constants.BELOW_MIN_80_TO_100)
    below_min_101_plus = types.IntType(required=False,
                                       serialized_name=api_constants.BELOW_MIN_101_PLUS)
    burst_iops_credit = types.IntType(required=False,
                                      serialized_name=api_constants.BURST_IOPS_CREDIT)
    client_queue_depth = types.IntType(required=False, serialized_name=api_constants.CLIENT_QUEUE_DEPTH)
    latency_usec = types.IntType(required=False, serialized_name=api_constants.LATENCY_USEC)
    local_write_latency_usec = types.IntType(required=False,
                                             serialized_name='localWriteLatencyUSec')
    min_to_max_0 = types.IntType(required=False,
                                 serialized_name=api_constants.MIN_TO_MAX_0)
    min_to_max_1_to_19 = types.IntType(required=False,
                                       serialized_name=api_constants.MIN_TO_MAX_1_TO_19)
    min_to_max_20_to_39 = types.IntType(required=False,
                                        serialized_name=api_constants.MIN_TO_MAX_20_TO_39)
    min_to_max_40_to_59 = types.IntType(required=False,
                                        serialized_name=api_constants.MIN_TO_MAX_40_TO_59)
    min_to_max_60_to_79 = types.IntType(required=False,
                                        serialized_name=api_constants.MIN_TO_MAX_60_TO_79)
    min_to_max_80_to_100 = types.IntType(required=False,
                                         serialized_name=api_constants.MIN_TO_MAX_80_TO_100)
    min_to_max_101_plus = types.IntType(required=False,
                                        serialized_name=api_constants.MIN_TO_MAX_101_PLUS)
    normalized_iops = types.IntType(required=False,
                                    serialized_name=api_constants.NORMALIZED_IOPS)
    read_block_size_128k_plus = types.IntType(required=False,
                                              serialized_name=api_constants.READ_BLOCK_SIZE_128K_PLUS)
    read_block_size_16k = types.IntType(required=False,
                                        serialized_name=api_constants.READ_BLOCK_SIZE_16K)
    read_block_size_32k = types.IntType(required=False,
                                        serialized_name=api_constants.READ_BLOCK_SIZE_32K)
    read_block_size_4k = types.IntType(required=False,
                                       serialized_name=api_constants.READ_BLOCK_SIZE_4K)
    read_block_size_512 = types.IntType(required=False,
                                        serialized_name=api_constants.READ_BLOCK_SIZE_512)
    read_block_size_64k = types.IntType(required=False,
                                        serialized_name=api_constants.READ_BLOCK_SIZE_64K)
    read_block_size_8k = types.IntType(required=False,
                                       serialized_name=api_constants.READ_BLOCK_SIZE_8K)
    read_bytes = types.IntType(required=False, serialized_name=api_constants.READ_BYTES)
    read_bytes_last_sample = types.IntType(required=False,
                                           serialized_name=api_constants.READ_BYTES_LAST_SAMPLE)
    read_latency_usec = types.IntType(required=False,
                                      serialized_name=api_constants.READ_LATENCY_USEC)
    read_latency_usec_total = types.IntType(required=False,
                                            serialized_name=api_constants.READ_LATENCY_USEC_TOTAL)
    read_ops = types.IntType(required=False, serialized_name=api_constants.READ_OPS)
    read_ops_last_sample = types.IntType(required=False,
                                         serialized_name=api_constants.READ_OPS_LAST_SAMPLE)
    remote_write_latency_usec = types.IntType(required=False,
                                              serialized_name='remoteWriteLatencyUSec')
    sample_period_msec = types.IntType(required=False, serialized_name=api_constants.SAMPLE_PERIOD_MSEC)
    service_ssload = types.IntType(required=False, serialized_name='serviceSSLoad')
    target_iops = types.IntType(required=False, serialized_name='targetIOPS')
    target_iops_including_burst = types.IntType(required=False,
                                                serialized_name='targetIOPSIncludingBurst')
    target_read_iops = types.IntType(required=False, serialized_name='targetReadIOPS')
    target_read_iops_including_burst = types.IntType(required=False,
                                                     serialized_name='targetReadIOPSIncludingBurst')
    target_utilization_0 = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_0)
    target_utilization_1_to_19 = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_1_TO_19)
    target_utilization_20_to_39 = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_20_TO_39)
    target_utilization_40_to_59 = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_40_TO_59)
    target_utilization_60_to_79 = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_60_TO_79)
    target_utilization_80_to_100 = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_80_TO_100)
    target_utilization_101_plus = types.IntType(
        required=False,
        serialized_name=api_constants.TARGET_UTILIZATION_101_PLUS)
    target_write_iops = types.IntType(required=False,
                                      serialized_name='targetWriteIOPS')
    target_write_iops_including_burst = types.IntType(
        required=False,
        serialized_name='targetWriteIOPSIncludingBurst')
    throttle = types.IntType(required=False, serialized_name=api_constants.THROTTLE)
    throttle_0 = types.IntType(required=False, serialized_name=api_constants.THROTTLE_0)
    throttle_1_to_19 = types.IntType(required=False,
                                     serialized_name=api_constants.THROTTLE_1_TO_19)
    throttle_20_to_39 = types.IntType(required=False,
                                      serialized_name=api_constants.THROTTLE_20_TO_39)
    throttle_40_to_59 = types.IntType(required=False,
                                      serialized_name=api_constants.THROTTLE_40_TO_59)
    throttle_60_to_79 = types.IntType(required=False,
                                      serialized_name=api_constants.THROTTLE_60_TO_79)
    throttle_80_to_100 = types.IntType(required=False,
                                       serialized_name=api_constants.THROTTLE_80_TO_100)
    throttle_101_plus = types.IntType(required=False,
                                      serialized_name=api_constants.THROTTLE_101_PLUS)
    throttle_read = types.IntType(required=False,
                                  serialized_name=api_constants.THROTTLE_READ)
    throttle_write = types.IntType(required=False,
                                   serialized_name=api_constants.THROTTLE_WRITE)
    unaligned_read_length = types.IntType(required=False,
                                          serialized_name='unalignedReadLength')
    unaligned_read_start = types.IntType(required=False,
                                         serialized_name='unalignedReadStart')
    unaligned_reads = types.IntType(required=False, serialized_name=api_constants.UNALIGNED_READS)
    unaligned_write_length = types.IntType(required=False,
                                           serialized_name='unalignedWriteLength')
    unaligned_write_start = types.IntType(required=False,
                                          serialized_name='unalignedWriteStart')
    unaligned_writes = types.IntType(required=False, serialized_name=api_constants.UNALIGNED_WRITES)
    volume_sswrite_limit = types.IntType(required=False, serialized_name='volumeSSWriteLimit')
    volume_utilization = types.DecimalType(required=False,
                                           serialized_name=api_constants.VOLUME_UTILIZATION)
    write_block_size_128k_plus = types.IntType(
        required=False,
        serialized_name=api_constants.WRITE_BLOCK_SIZE_128K_PLUS)
    write_block_size_16k = types.IntType(required=False,
                                         serialized_name=api_constants.WRITE_BLOCK_SIZE_16K)
    write_block_size_32k = types.IntType(required=False,
                                         serialized_name=api_constants.WRITE_BLOCK_SIZE_32K)
    write_block_size_4k = types.IntType(required=False,
                                        serialized_name=api_constants.WRITE_BLOCK_SIZE_4K)
    write_block_size_512 = types.IntType(required=False,
                                         serialized_name=api_constants.WRITE_BLOCK_SIZE_512)
    write_block_size_64k = types.IntType(required=False,
                                         serialized_name=api_constants.WRITE_BLOCK_SIZE_64K)
    write_block_size_8k = types.IntType(required=False,
                                        serialized_name=api_constants.WRITE_BLOCK_SIZE_8K)
    write_bytes = types.IntType(required=False, serialized_name=api_constants.WRITE_BYTES)
    write_bytes_last_sample = types.IntType(required=False,
                                            serialized_name=api_constants.WRITE_BYTES_LAST_SAMPLE)
    write_latency_usec = types.IntType(required=False,
                                       serialized_name=api_constants.WRITE_LATENCY_USEC)
    write_latency_usec_total = types.IntType(required=False,
                                             serialized_name=api_constants.WRITE_LATENCY_USEC_TOTAL)
    write_ops = types.IntType(required=False, serialized_name=api_constants.WRITE_OPS)
    write_ops_last_sample = types.IntType(required=False,
                                          serialized_name=api_constants.WRITE_OPS_LAST_SAMPLE)

    def __init__(self, raw_data):
        super(VolumeRawStats, self).__init__(raw_data=raw_data)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            RawStats.NAME_ID, self.name_id,
            api_constants.VOLUME_ID, self.volume_id,
        )


class ServiceRawStats(models.Base):
    """ServiceRawStats data model.

    Corresponds to the elements returned by get_raw_stats().

    :param dict raw_data: The raw service stats data.
    """
    name_id = types.StringType(required=False, serialized_name=RawStats.NAME_ID)
    service_id = types.IntIdType(required=False, serialized_name=api_constants.SERVICE_ID)
    active_bytes_after_last_gc = types.IntType(required=False,
                                               serialized_name='activeBytesAfterLastGC')
    active_data_pct = types.DecimalType(required=False, serialized_name='activeDataPct')
    active_disk_bytes = types.IntType(required=False, serialized_name='activeDiskBytes')
    active_sessions = types.IntType(required=False, serialized_name=api_constants.ACTIVE_SESSIONS)
    additional_reads_on_block_reads = types.IntType(required=False,
                                                    serialized_name='additionalReadsOnBlockReads')
    additional_reads_on_block_writes = types.IntType(required=False,
                                                     serialized_name='additionalReadsOnBlockWrites')
    additional_reads_on_dup_block_writes = types.IntType(
        required=False,
        serialized_name='additionalReadsOnDupBlockWrites')
    bin_sync_time_ms = types.IntType(required=False,
                                     serialized_name=api_constants.BIN_SYNC_TIME_MS)
    bins_completed = types.IntType(required=False,
                                   serialized_name=api_constants.BINS_COMPLETED)
    block_bin_count = types.IntType(required=False,
                                    serialized_name=api_constants.BLOCK_BIN_COUNT)
    block_reads = types.IntType(required=False,
                                serialized_name=api_constants.BLOCK_READS)
    block_writes = types.IntType(required=False, serialized_name=api_constants.BLOCK_WRITES)
    bs_read_diff = types.IntType(required=False, serialized_name='bsReadDiff')
    c_bytes_in = types.IntType(required=False, serialized_name=api_constants.C_BYTES_IN)
    c_bytes_out = types.IntType(required=False, serialized_name=api_constants.C_BYTES_OUT)
    compression_est = types.IntType(required=False, serialized_name='compressionEst')
    count = types.IntType(required=False, serialized_name='count')
    cpu = types.IntType(required=False, serialized_name=api_constants.CPU)
    cpu_total = types.IntType(required=False, serialized_name=api_constants.CPU_TOTAL)
    db_runtime_inconsistency_count = types.IntType(required=False,
                                                   serialized_name='dbRuntimeInconsistencyCount')
    discarded_disk_bytes = types.IntType(required=False, serialized_name='discardedDiskBytes')
    dupe_block_writes = types.IntType(required=False, serialized_name='dupeBlockWrites')
    last_gc_blocks_discarded = types.IntType(required=False,
                                             serialized_name='lastGCBlocksDiscarded')
    last_gc_time = types.IntType(required=False, serialized_name='lastGCTime')
    last_recycle_active_pct = types.DecimalType(required=False,
                                                serialized_name='lastRecycleActivePct')
    last_segment_length = types.IntType(required=False, serialized_name='lastSegmentLength')
    master_version = types.IntType(required=False, serialized_name=api_constants.MASTER_VERSION)
    max_unique_blocks_per_bin = types.IntType(
        required=False,
        serialized_name=api_constants.MAX_UNIQUE_BLOCKS_PER_BIN)
    min_unique_blocks_per_bin = types.IntType(
        required=False,
        serialized_name=api_constants.MIN_UNIQUE_BLOCKS_PER_BIN)
    m_bytes_in = types.IntType(required=False, serialized_name=api_constants.M_BYTES_IN)
    m_bytes_out = types.IntType(required=False, serialized_name=api_constants.M_BYTES_OUT)
    network_utilization_cluster = types.IntType(
        required=False,
        serialized_name=api_constants.NETWORK_UTILIZATION_CLUSTER)
    network_utilization_storage = types.IntType(
        required=False,
        serialized_name=api_constants.NETWORK_UTILIZATION_STORAGE)
    no_bss_for_read = types.IntType(required=False, serialized_name='noBSsForRead')
    number_of_updating_bins = types.IntType(
        required=False,
        serialized_name=api_constants.NUMBER_OF_UPDATING_BINS)
    num_pcache_read_attempts = types.IntType(required=False,
                                             serialized_name='numPCacheReadAttempts')
    num_pcache_read_hits = types.IntType(required=False,
                                         serialized_name='numPCacheReadHits')
    paging_false_positive_percentage = types.IntType(
        required=False,
        serialized_name='pagingFalsePositivePercentage')
    paging_filter_false_positives = types.IntType(required=False,
                                                  serialized_name='pagingFilterFalsePositives')
    paging_filter_lookups = types.IntType(required=False, serialized_name='pagingFilterLookups')
    pcache_segments_in_use = types.IntType(required=False, serialized_name='pcacheSegmentsInUse')
    primary_blocks_failed = types.IntType(required=False, serialized_name='PrimaryBlocksFailed')
    primary_blocks_sent = types.IntType(required=False, serialized_name='PrimaryBlocksSent')
    primary_sets_completed = types.IntType(required=False, serialized_name='PrimarySetsCompleted')
    primary_sets_outstanding = types.IntType(required=False,
                                             serialized_name='PrimarySetsOutstanding')
    primary_sets_sent = types.IntType(required=False, serialized_name='PrimarySetsSent')
    read_bytes = types.IntType(required=False, serialized_name=api_constants.READ_BYTES)
    read_cache_hits = types.IntType(required=False, serialized_name='readCacheHits')
    read_ops = types.IntType(required=False, serialized_name=api_constants.READ_OPS)
    scache_bytes_in_use = types.IntType(required=False, serialized_name='scacheBytesInUse')
    secondary_blocks_failed = types.IntType(required=False,
                                            serialized_name='SecondaryBlocksFailed')
    secondary_blocks_sent = types.IntType(required=False,
                                          serialized_name='SecondaryBlocksSent')
    secondary_sets_completed = types.IntType(required=False,
                                             serialized_name='SecondarySetsCompleted')
    secondary_sets_sent = types.IntType(required=False, serialized_name='SecondarySetsSent')
    s_bytes_in = types.IntType(required=False, serialized_name=api_constants.S_BYTES_IN)
    s_bytes_out = types.IntType(required=False, serialized_name=api_constants.S_BYTES_OUT)
    ss_load_0 = types.IntType(required=False, serialized_name=api_constants.SS_LOAD_0)
    ss_load_1_to_19 = types.IntType(required=False,
                                    serialized_name=api_constants.SS_LOAD_1_TO_19)
    ss_load_20_to_39 = types.IntType(required=False,
                                     serialized_name=api_constants.SS_LOAD_20_TO_39)
    ss_load_40_to_59 = types.IntType(required=False,
                                     serialized_name=api_constants.SS_LOAD_40_TO_59)
    ss_load_60_to_79 = types.IntType(required=False,
                                     serialized_name=api_constants.SS_LOAD_60_TO_79)
    ss_load_80_to_100 = types.IntType(required=False,
                                      serialized_name=api_constants.SS_LOAD_80_TO_100)
    ss_load_101_plus = types.IntType(required=False,
                                     serialized_name=api_constants.SS_LOAD_101_PLUS)
    ss_load_primary = types.IntType(required=False,
                                    serialized_name=api_constants.SS_LOAD_PRIMARY)
    ss_load_secondary = types.IntType(required=False,
                                      serialized_name=api_constants.SS_LOAD_SECONDARY)
    total_capacity = types.IntType(required=False,
                                   serialized_name=api_constants.TOTAL_CAPACITY)
    truncated_map_size = types.IntType(required=False, serialized_name='truncatedMapSize')
    unique_blocks = types.IntType(required=False, serialized_name=api_constants.UNIQUE_BLOCKS)
    unique_blocks_after_last_gc = types.IntType(required=False,
                                                serialized_name='uniqueBlocksAfterLastGC')
    unique_blocks_of_complete_bins = types.IntType(required=False,
                                                   serialized_name='uniqueBlocksOfCompleteBins')
    unique_blocks_used_space = types.IntType(required=False,
                                             serialized_name=api_constants.UNIQUE_BLOCK_USED_SPACE)
    unique_blocks_used_space_after_last_gc = types.IntType(
        required=False,
        serialized_name='uniqueBlocksUsedSpaceAfterLastGC')
    unique_blocks_used_space_of_complete_bins = types.IntType(
        required=False,
        serialized_name='uniqueBlocksUsedSpaceOfCompleteBins')
    unique_blocks_used_space_without_overhead = types.IntType(
        required=False,
        serialized_name='uniqueBlocksUsedSpaceWithoutOverhead')
    unresponsive_bss_for_read = types.IntType(required=False,
                                              serialized_name='unresponsiveBSsForRead')
    unresponsive_bss_for_write = types.IntType(required=False,
                                               serialized_name='unresponsiveBSsForWrite')
    unused_slice_file_blocks = types.IntType(required=False,
                                             serialized_name='unusedSliceFileBlocks')
    usable_capacity = types.IntType(required=False,
                                    serialized_name=api_constants.USABLE_CAPACITY)
    used_disk_bytes = types.IntType(required=False, serialized_name='usedDiskBytes')
    used_memory = types.IntType(required=False, serialized_name=api_constants.USED_MEMORY)
    used_slice_file_blocks = types.IntType(required=False, serialized_name='usedSliceFileBlocks')
    used_slice_file_blocks_all_branches = types.IntType(
        required=False,
        serialized_name='usedSliceFileBlocksAllBranches')
    write_bytes = types.IntType(required=False, serialized_name=api_constants.WRITE_BYTES)
    write_ops = types.IntType(required=False, serialized_name=api_constants.WRITE_OPS)

    def __init__(self, raw_data):
        super(ServiceRawStats, self).__init__(raw_data=raw_data)

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            RawStats.NAME_ID, self.name_id,
            api_constants.SERVICE_ID, self.service_id,
        )
