"""Reports from the product.

These models are not included in sf_api as they are not API responses.
These are not part of the official solidfire API, but are available at {mvip}/reports.
These are useful for determining cluster state and debugging.

Available reports:
/reports/binsyncing = BS Bin Syncing
/reports/bsdiskdata  = BS Disk Data
/reports/bsmaps  = BS Maps (WARNING: May Be Taxing on Large Clusters)
/reports/bsqueues  = BS Queues
/reports/bins  = BS Status
/reports/flushing  = Cache Flushing
/reports/clones  = Clones
/reports/timing  = Code Timing
/reports/ensemble  = Ensemble
/reports/events  = Event Report
/reports/iscsi  = ISCSI Sessions
/reports/memory  = Memory Usage
/reports/mutexes  = Mutex Timing
/reports/qos  = QoS Status
/reports/rpc  = RPC connections
/reports/rpcTime  = RPCTime
/reports/rpcTimeReset  = RPCTimeReset
/reports/rrbdt  = RR Block Data Transfer
/reports/scsiCommands  = SCSI Commands
/reports/scsiCommandsAndSessions  = SCSI CommandsAndSessions
/reports/scsiNexuses  = SCSI ITL Nexuses
/reports/scsiReservations  = SCSI Reservations
/reports/scsiVolumes  = SCSI Volumes
/reports/ssdirector  = SSDirector
/reports/slices  = Slice Assignments
/reports/slicebalance  = Slice Balance
/reports/slicefilesystem  = Slice File System
/reports/sliceinfo  = Slice Info
/reports/ss  = Slice Services
/reports/slicesyncing  = Slice Syncing
/reports/snapmirror  = SnapMirror
/reports/socketservice  = Socket Services
/reports/unresponsive  = Unresponsive BS
/reports/vvolBindings  = Vvol Bindings

"""
import collections
import typing  # pylint: disable=unused-import

import autologger
import six

from sf_api import (
    models,
    sf_version,
)
from sf_api.models import (
    errors,
    types,
)
from valence import enums

_log = autologger.get_sflogging()


class ReportModel(models.Base):
    """Base report model.

    This base class is here purely to disable strict validation for reports. JSON reports change often
    and do not go through the API alliance so it is not practical to do strict validation for them.

    :param dict raw_data: the dictionary returned by get_reports
    :param bool strict: Set to True to enable strict validation
    """

    def __init__(self, raw_data, strict=False, **kwargs):
        super(ReportModel, self).__init__(raw_data=raw_data, strict=strict, **kwargs)


class BinService(ReportModel):
    """Service report from bin report."""
    service_id = types.IntIdType(required=True, serialized_name='serviceID')
    status = types.EnumType(enums.BinStatus, required=True)
    replica = types.IntIdType(required=True, version_added=sf_version.SfVersion('12.0'))

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'service_id', self.service_id,
            'status', self.status,
            'replica', self.replica
        )


class BinReport(ReportModel):
    """Product bin report."""
    bin_id = types.IntIdType(required=True, serialized_name='binID')
    services = types.ListType(types.ModelType(BinService), required=True)  # type: typing.List

    def __repr__(self):
        return '{}<bin_id={}>'.format(self.__class__.__name__, self.bin_id)


class ProtectionDomainServiceInfo(models.Base):
    """Protection domain service info."""
    service_id = types.IntIdType(required=True, serialized_name='serviceID')
    budget = types.IntType(required=True)


class ProtectionDomainServices(models.Base):
    """Protection domain services."""
    protection_domain_name = types.StringType(required=True, serialized_name='protectionDomainName')
    services = types.ListType(types.ModelType(ProtectionDomainServiceInfo), required=True)


class ServiceStrandedCapacityInfo(models.Base):
    """Service stranded capacity info."""
    service_id = types.IntIdType(required=True, serialized_name='serviceID')
    stranded_capacity = types.IntType(required=True, serialized_name='strandedCapacity')


class BinAssignmentsProperties(models.Base):
    """Bin assignments properties."""
    algorithm_runtime = types.IntType(required=True, serialized_name='algorithmRuntimeMS')
    replicas_valid = types.BooleanType(required=True, serialized_name='areReplicasValid')
    bin_count = types.IntType(required=True, serialized_name='binCount')
    is_balanced = types.BooleanType(required=True, serialized_name='isBalanced')
    is_stable = types.BooleanType(required=True, serialized_name='isStable')
    layout = types.ListType(types.ModelType(ProtectionDomainServices), required=True)
    num_swaps = types.IntType(required=True, serialized_name='numSwaps')
    num_updating_bins = types.IntType(required=True, serialized_name='numUpdatingBins')
    replication_count = types.IntType(required=True, serialized_name='replicationCount')
    service_stranded_capacities = types.ListType(
        field=types.ModelType(ServiceStrandedCapacityInfo),
        required=True,
        serialized_name='serviceStrandedCapacities'
    )
    time_published = types.DateTimeType(required=True, serialized_name='timePublished')
    valid_schemes = types.ListType(
        field=types.EnumType(enums.ProtectionScheme),
        required=True,
        serialized_name='validSchemes'
    )


class BinAssignments(ReportModel):
    """Cluster bin assignments.

    Corresponds to the elements returned by the bins.json report returned by get_bin_assignments().

    :param dict raw_data: The raw bins.json report data.
    """
    # pylint: disable=no-member
    bins = types.ListType(types.ModelType(BinReport), required=True)  # type: typing.List
    master_version = types.IntType(required=False, serialized_name='masterVersion')
    properties = types.ModelType(
        model_spec=BinAssignmentsProperties,
        required=True,
        version_added=sf_version.SfVersion('12.0')
    )

    def __init__(self, raw_data):
        super(BinAssignments, self).__init__(raw_data=raw_data)
        self._rep_count = self.properties.replication_count if self.properties else 2  # type: int
        self._domain_service_map = None
        self._service_replica_map_not_dead = None
        self._error_invalid_replica = 0
        self._error_invalid_active = 0
        self._error_no_updating = 0
        self._error_no_removing = 0
        self._error_invalid_pra = 0
        self._error_service_unbalanced = 0
        self._error_zone_contains_two_replicas = 0
        self._error_service_in_multiple_zones = 0
        self._error_no_service_budget = 0
        self._error_bins_unbalanced_by_budget = 0
        self._error_zone_multiple_active = 0
        self._is_service_level = None

        self._updating_statuses = {
            enums.BinStatus.updating,
            enums.BinStatus.updating_no_sync,
            enums.BinStatus.updating_from_active,
            enums.BinStatus.pending_active
        }
        self._removing_statuses = {
            enums.BinStatus.pending_removal_active,
            enums.BinStatus.pending_removal_dead,
            enums.BinStatus.pending_removal_dead_from_active
        }
        self._goal_statuses = {
            enums.BinStatus.active,
            enums.BinStatus.updating,
            enums.BinStatus.updating_no_sync,
            enums.BinStatus.updating_from_active,
            enums.BinStatus.pending_active
        }

    @property
    def replica_based(self):
        """Whether the bin assignments are replica-based or not.

        :rtype: bool
        """
        # pylint: disable=unsubscriptable-object
        bin_0 = self.bins[0]
        if not self.properties or bin_0.services is None or bin_0.services[0].replica is None:
            return False
        else:
            return True

    def validate(self, version=None, **kwargs):
        """Validate the bin assignments object returned from get_bin_assignments().

        :param sf_api.SfVersion version: SolidFire API version to validate against.
        """
        # No explicit validation available for non-replica-based bin assignments
        if not self.replica_based:
            return

        # Perform type-validation on the properties
        # pylint: disable=unexpected-keyword-arg
        self.get('properties').validate(version=version, **kwargs)

        # Validate the bin_count matches the number of bins in the report
        if self.properties.bin_count != len(self.bins):
            raise errors.DataValidationError({
                'Validation error': 'Number of bins in report ({}) != bin_count ({})'
                .format(len(self.bins), self.properties.bin_count)
            })

        # Reset error counters before each validation run
        self._error_invalid_replica = 0
        self._error_invalid_active = 0
        self._error_no_updating = 0
        self._error_no_removing = 0
        self._error_invalid_pra = 0
        self._error_service_unbalanced = 0
        self._error_zone_contains_two_replicas = 0
        self._error_service_in_multiple_zones = 0
        self._error_bins_unbalanced_by_budget = 0
        self._error_no_service_budget = 0
        self._error_zone_multiple_active = 0

        # Validate each bin assignment
        self._domain_service_map = self._get_domain_service_map()
        self._service_replica_map_not_dead = collections.defaultdict(
            lambda: {replica: 0 for replica in range(self._rep_count)}
        )
        for bin_assignment_report in self.bins:  # pylint: disable=not-an-iterable
            self._validate_bin_assignment(bin_assignment_report)

        # tests that apply only when isBalanced is set
        if self.properties.is_balanced:
            # Validate each service has an equal number of bin assignments for each replica
            for replica_map in self._service_replica_map_not_dead.values():
                min_replica_count = min(replica_map.values())
                max_replica_count = max(replica_map.values())
                if abs(max_replica_count - min_replica_count) > 2:
                    self._error_service_unbalanced += 1

            self._validate_count_for_each_replica()

            # Validate each service has a balanced count of bin assignments
            self._validate_bin_assignments_balanced_by_budget()

        # verify all the zone mappings.
        self._validate_zones()

        errors_found = any([
            self._error_invalid_replica,
            self._error_invalid_active,
            self._error_no_updating,
            self._error_no_removing,
            self._error_invalid_pra,
            self._error_service_unbalanced,
            self._error_zone_contains_two_replicas,
            self._error_service_in_multiple_zones,
            self._error_no_service_budget,
            self._error_bins_unbalanced_by_budget,
            self._error_zone_multiple_active
        ])
        if errors_found:
            error_msgs = [
                'invalid replicas={}'.format(self._error_invalid_replica),
                'active replicas with multiple services={}'.format(self._error_invalid_active),
                'inactive replicas with no updating service={}'.format(self._error_no_updating),
                'inactive replicas with no pending-removal service={}'.format(self._error_no_removing),
                'replicas with invalid PRA services={}'.format(self._error_invalid_pra),
                'services with unbalanced replicas={}'.format(self._error_service_unbalanced),
                'zone contained two replicas={}'.format(self._error_zone_contains_two_replicas),
                'service exists in more than one zone={}'.format(self._error_service_in_multiple_zones),
                'service missing a bins budget={}'.format(self._error_no_service_budget),
                'service with unbalanced bin assignments={}'.format(
                    self._error_bins_unbalanced_by_budget),
                'multiple bins statuses of A, PA, U or UNS encountered in zone={}'.format(
                    self._error_zone_multiple_active
                )
            ]
            raise errors.DataValidationError({'Validation error': ', '.join(error_msgs)})

    def validate_service_level_pd(self):
        """Verifies the Protection Domain layouts are in Service Level for the Bin Assignments.

        :raises sf_api.models.errors.DataValidationError:
        """
        error_msg = []
        found_errors = {}

        # check layouts for pd
        service_level_errors = {}
        for pd_layout in self.properties.layout:
            if pd_layout.protection_domain_name != str(pd_layout.services[0].service_id):
                error_msg.append("ProtectionDomainName '{}' must match serviceID '{}' for Service "
                                 'level'.format(pd_layout.protection_domain_name,
                                                pd_layout.services[0].service_id))
                service_level_errors[pd_layout.protection_domain_name] = 1
            if len(pd_layout.services) != 1:
                error_msg.append("Expected 1 service instead got '{}' in ProtectionDomainName '{}'"
                                 .format(len(pd_layout.services), pd_layout.protection_domain_name))
                service_level_errors[pd_layout.protection_domain_name] = 1

        found_errors['Invalid service level pd'] = sum(service_level_errors.values())

        # check service_stranded_capacities in service level
        found_errors['Invalid service stranded capacities'] = len(
            self.properties.service_stranded_capacities)
        if found_errors['Invalid service stranded capacities']:
            error_msg.append("Expected no serviceStrandedCapacities in Service Level but got '{}'"
                             .format(found_errors['Invalid service stranded capacities']))

        if error_msg:
            _log.debug('; '.join(error_msg))

        if any(found_errors.values()):
            raise errors.DataValidationError(found_errors)

    def is_service_level(self):
        """Lazy check if the Bin Assignments have Service Level Protection Domain.

        :return: bool: True if using Service Level, False otherwise.
        :rtype: bool
        """
        if self._is_service_level is not None:
            return self._is_service_level

        try:
            self.validate_service_level_pd()
            self._is_service_level = True
        except errors.DataValidationError:
            self._is_service_level = False
        return self._is_service_level

    def _get_domain_service_map(self):
        """Get protection-domain-to-service map.

        Maps service IDs to the list of protection domains in which that service resides.

        Example::
            {0: ['domainA', 'domainB'], 1: ['domainC'], ...}

        :rtype: dict[list[str]]
        """
        domain_service_map = collections.defaultdict(list)
        for protection_domain in self.properties.layout:
            for service in protection_domain.services:
                domain_service_map[service.service_id].append(protection_domain.protection_domain_name)

        return domain_service_map

    def _create_service_id_bin_assignment_cnt_map(self):
        """Create map of ServiceId/BinAssignmentCount from bin assignment report.

        Count only includes bin assignments with a status of Updating(U), Updating No Sync(UNS),
        Updating From Active(UA), Pending Active(PA), and Active(A)
        :rtype: dict
        """

        service_id_bin_assignment_cnt_map = collections.defaultdict(int)
        for bin_assignment in self.bins:  # pylint: disable=not-an-iterable
            for service in bin_assignment.services:
                service_id_bin_assignment_cnt_map[service.service_id] += int(
                    service.status in self._goal_statuses)
        return service_id_bin_assignment_cnt_map

    def _validate_bin_assignments_balanced_by_budget(self):
        """Validate the bin assignments are balanced as measured against the service budget.

        Bin assignments with a status of Updating, Updating No Sync, Updating From Active,
        Pending Active, and Active should have a count that is the same as
        the budget * replicationCount +/- replicationCount.
        """
        service_id_bin_assignment_cnt_map = self._create_service_id_bin_assignment_cnt_map()

        service_id_budget_map = {}
        for layout in self.properties.layout:
            for service in layout.services:
                service_id_budget_map[service.service_id] = service.budget

        msg = ('Service id {} does not have a balanced number of bin assignments. '
               'Bin count={} Expected={} +/-{}')
        tolerance = self._rep_count
        for service_id, count in six.iteritems(service_id_bin_assignment_cnt_map):
            try:
                expected_bins = service_id_budget_map[service_id] * self._rep_count
                if not expected_bins - tolerance <= count <= expected_bins + tolerance:
                    self._error_bins_unbalanced_by_budget += 1
                    _log.debug(msg.format(service_id, count, expected_bins, tolerance))
            except KeyError:
                self._error_no_service_budget += 1
                _log.debug('Service id {} does not have a layout budget'.format(service_id))

    def _validate_bin_assignment(self, bin_assignment_report):
        """Validate the specified bin assignment.

        :param valence.models.BinReport bin_assignment_report: Bin assignment to validate.
        """
        # Validate there are no more than the expected number of replicas
        self._error_invalid_replica += sum(
            int(s.replica > self._rep_count - 1) for s in bin_assignment_report.services
        )

        # Validate no zone has more than one active replica. PRA, PRD, and PRDA are ignored.
        replica_map = collections.defaultdict(set)
        for service in bin_assignment_report.services:
            if service.status not in self._removing_statuses:
                domains_for_service = tuple(set(self._domain_service_map[service.service_id]))
                for domain in domains_for_service:
                    replica_map[domain].add(service.replica)
                    if len(replica_map[domain]) != 1:
                        self._error_zone_contains_two_replicas += 1

        # Validate that each replica has the correct number of services given their status
        for replica in range(self._rep_count):
            services = [s for s in bin_assignment_report.services if s.replica == replica]
            self._validate_replica_services(services)
            # Build a map of bin assignment counts for a given replica there are on each service
            for service in services:
                if service.status not in self._removing_statuses:
                    self._service_replica_map_not_dead[service.service_id][replica] += 1

    def _validate_replica_services(self, services):
        """Validate the specified services form a valid replica for a given bin assignment.

        The services associated with each replica of a bin must put it into one of the following valid
        replica states:

        Abbreviations:
            PRA: Pending-Removal-Active
            PRD: Pending-Removal-Dead

        Example::
            +-----------------------------------------------------------
            | Service-Status (valid number of services with this status)
            +-----------------------------------------------------------

        Valid Replica States::
            +-----------------------------------------------------------
            | Active (1), PRA (0,1)
            +-----------------------------------------------------------
            | Pending-Active (1), PRA/PRD (1), PRD (>=0)
            +-----------------------------------------------------------
            | Updating (1), PRA/PRD (1), PRD (>=0)
            +-----------------------------------------------------------
            | Updating-No-Sync (1), PRD (>=1)
            +-----------------------------------------------------------

        :param list[valence.models.BinService] services: Replica services.
        """
        # pylint: disable=too-complex
        statuses = [service.status for service in services]

        if not services:
            self._error_invalid_replica += 1
        elif enums.BinStatus.active in statuses:
            service_len = 1
            if enums.BinStatus.pending_removal_active in statuses:
                service_len += 1
            if len(services) != service_len:
                self._error_invalid_active += 1
        elif len(self._updating_statuses.intersection(statuses)) != 1:
            self._error_no_updating += 1
        elif len(self._removing_statuses.intersection(statuses)) == 0:
            self._error_no_removing += 1
        elif statuses.count(enums.BinStatus.pending_removal_active) == 1:
            if enums.BinStatus.updating_no_sync in statuses:
                self._error_invalid_pra += 1
        elif statuses.count(enums.BinStatus.pending_removal_active) > 1:
            self._error_invalid_pra += 1

    def _validate_zones(self):
        """Verify all the zone mappings.

        This will verify that a service does not exist in more than one zone.
        It will also verify that the status for a bin in each zone only has one of
        _goal_statuses values.
        Note: protection_domains == zones
        """

        protection_domains = self.properties.layout
        domain_service_id_list = collections.defaultdict(list)
        # verify that each service only appears in one zone

        if protection_domains:
            # service_count is used to verify that the service only exists in one zone.
            service_count = collections.defaultdict(int)
            for domain in protection_domains:
                domain_name = domain.protection_domain_name
                # get the bins for each service in the protection domain.
                for service in domain.services:
                    # generate a list of service ids for a specific protections domain
                    domain_service_id_list[domain_name].append(service.service_id)
                    service_count[service.service_id] += 1

            # Verify that the service only exists in one zone
            for service_id in service_count:
                if service_count[service_id] != 1:
                    self._error_service_in_multiple_zones += 1

            # get the bins for each zone
            zone_bin_service_map = self._get_zone_bin_status_map(domain_service_id_list)
            for bin_status_list in zone_bin_service_map.values():  # for each zone and bin
                self._verify_only_one_active_status(bin_status_list)

    def _verify_only_one_active_status(self, bin_status_list):
        """This will verify that the status(es) for a bin in a zone have only one active status.

        The status_list can only have one of Active, Updating, Updating No Sync, Updating From Active,
        or Pending Active. Additional PRAs, PRDs, or PRDAs are not considered since
        _validate_replica_services already checks for those.
        This updates the class error count for _error_zone_multiple_active

        :param list[valence.enums.BinStatus] bin_status_list: list of status(es) for that bin
                                                              in that zone.
        """
        status_ct = 0
        for status in bin_status_list:
            if status in self._goal_statuses:
                status_ct += 1
        if status_ct > 1:
            self._error_zone_multiple_active += status_ct - 1

    def _get_zone_bin_status_map(self, domain_service_id_list):
        """This will re-order the bin information into map (zoneID, binID):[statuses]].

        :param collections.defaultdict[int,list[int]] domain_service_id_list: list of the service ids
                                                                         for each zone id
        :return: tuple(zoneID, binID):[statuses] A list of statuses for all the service entries
                                                 for the bin in that zone
        :rtype: collections.defaultdict[list[valence.enums.BinStatus]]
        """
        zone_bin_service_map = collections.defaultdict(list)
        for bin_entry in self.bins:  # pylint: disable=not-an-iterable
            bin_id = bin_entry.bin_id
            for service in bin_entry.services:
                service_id = service.service_id
                for zone_id in domain_service_id_list:
                    service_set = domain_service_id_list[zone_id]
                    if service_id in service_set:
                        zone_bin_service_map[(zone_id, bin_id)].append(service.status)
                        break
        return zone_bin_service_map

    def _validate_count_for_each_replica(self):
        """Verify bin count for each replica are +/- 1 in bin assignments."""
        service_replicas = collections.defaultdict(int)
        for b in self.bins:  # pylint: disable=not-an-iterable
            for service in b.services:
                service_replicas[(service.service_id, service.replica)] += int(service.status in
                                                                               self._goal_statuses)
        for domain in self.properties.layout:
            for service in domain.services:
                actual_replicas = [service_replicas[(service.service_id, r)] for r
                                   in range(self._rep_count)]
                unbalanced = sum([not (service.budget - 1 <= actual <=
                                       service.budget + 1) for actual in actual_replicas])
                if unbalanced:
                    _log.debug('service id {} is unbalanced! expected {} +/-1 and found {}'.format(
                        service.service_id, service.budget,
                        ','.join([str(actual) for actual in actual_replicas])))
                self._error_service_unbalanced += unbalanced


class SliceService(ReportModel):
    """Slice service report."""
    health = types.EnumType(enums.SliceHealth, required=True)
    ip = types.StringType(required=True)
    node_id = types.IntIdType(required=True, serialized_name='nodeID')
    service_id = types.IntIdType(required=True, serialized_name='serviceID')

    def __repr__(self):
        return '{}<node={}, health={}, service_id={}>'.format(
            self.__class__.__name__, self.node_id, self.health, self.service_id)


class RepBranchInfo(ReportModel):
    """Replication branch information returned in a slice report."""
    last_synced_time = types.OptionalDateTimeType(serialized_name='lastSyncedTime')
    local_incarn = types.IntType(required=True, serialized_name='localIncarn')
    remote_incarn = types.IntType(required=True, serialized_name='remoteIncarn')
    snapshot_id = types.IntType(required=True, serialized_name='snapshotID')
    source_snapshot_id = types.IntType(required=True, serialized_name='sourceSnapshotID')
    state = types.StringType(required=True)  # VL-1749 enum
    state_details = types.StringType(required=True, serialized_name='stateDetails')
    type = types.StringType(required=True)  # VL-1749 enum


class RemoteReplicationInfo(ReportModel):
    """Information about remote replication returned in a slice report."""
    current_replicating_branch = types.StringType(
        required=True, serialized_name='currentReplicatingBranch')  # VL-1749 enum
    is_configured = types.BooleanType(required=True, serialized_name='isConfigured')
    is_target = types.BooleanType(required=True, serialized_name='isTarget')
    mode = types.StringType(required=True)  # VL-1749 enum
    pause_limit = types.IntType(required=True, serialized_name='pauseLimit')
    remote_service_id = types.IntType(required=True, serialized_name='remoteServiceID')
    rep_branch_info = types.ListType(
        types.ModelType(RepBranchInfo, required=True), serialized_name='repBranchInfos')
    resume_details = types.StringType(required=True, serialized_name='resumeDetails')
    sync_snap_id = types.IntType(required=True, serialized_name='syncSnapID')
    volume_pair_uuid = types.StringType(required=True, serialized_name='volumePairUUID')


class StandbyServiceInfo(ReportModel):
    """Information about Standby Services in slice model."""
    failed_version = types.IntIdType(required=True, serialized_name='failedVersion')
    active_version = types.IntIdType(required=True, serialized_name='activeVersion')
    service_id = types.IntType(required=True, serialized_name='serviceID')
    state = types.EnumType(enums.StandbyStates, required=True, serialized_name='state')

    def __repr__(self):
        return '{}<service_id={}, state={}>'.format(self.__class__.__name__, self.service_id,
                                                    self.state)


class Slice(ReportModel):
    """Slice report."""
    active_branch = types.IntType(required=True, serialized_name='activeBranch')
    active_branch_incarnation = types.IntType(required=True, serialized_name='activeBranchIncarnation')
    block_size = types.IntType(required=True, serialized_name='blockSize')
    bulk_operations = types.ListType(types.IntType, required=True, serialized_name='bulkOperations')
    clone_snapshots = types.ListType(types.IntType, required=True, serialized_name='cloneSnapshots')
    current_protection_scheme = types.EnumType(
        enums.ProtectionScheme, required=True, serialized_name='currentProtectionScheme')
    dead_clone_jobs = types.DictType(types.IntType, required=True, serialized_name='deadCloneJobs')
    dead_secondaries = types.ListType(types.IntIdType, required=True, serialized_name='deadSecondaries')
    desired_primary = types.IntIdType(required=True, serialized_name='desiredPrimary')
    desired_secondaries = types.ListType(
        types.IntIdType, required=True, serialized_name='desiredSecondaries')
    incarnations = types.DictType(types.IntType, required=True)
    is_new = types.BooleanType(required=True, serialized_name='isNew')
    is_snapmirror_transfer_in_progress = types.BooleanType(
        required=True, serialized_name='isSnapMirrorTransferInProgress')
    live_secondaries = types.ListType(types.IntIdType, required=True, serialized_name='liveSecondaries')
    number_external_snapshots = types.IntType(required=True, serialized_name='numberExternalSnapshots')
    primary = types.IntIdType(required=True)
    reassignment_type = types.StringType(
        required=True, serialized_name='reassignmentType')  # VL-1749 enum
    remote_replication_info = types.ModelType(
        RemoteReplicationInfo, required=True, serialized_name='remoteReplicationInfo')
    resize_incarnation = types.IntType(required=True, serialized_name='resizeIncarnation')
    resync_incarnation = types.IntType(required=True, serialized_name='resyncIncarnation')
    slice_id = types.IntIdType(required=True, serialized_name='sliceID')
    snapshots = types.DictType(types.IntType, required=True)
    snapshots_in_progress = types.ListType(
        types.DictType(types.StringType(), required=True),
        required=True, serialized_name='snapshotsInProgress')
    snapshots_incarnation = types.IntType(required=True, serialized_name='snapshotsIncarnation')
    standby_services = types.ListType(types.ModelType(StandbyServiceInfo), required=True,
                                      serialized_name='standbyServices')
    total_size = types.IntType(required=True, serialized_name='totalSize')
    version_bump_reasons = types.ListType(
        types.StringType, required=True, serialized_name='versionBumpReasons')  # VL-1749 enum
    volume_id = types.IntIdType(required=True, serialized_name='volumeID')


# TODO: refactor using enums from VL-1749
class SliceReport(ReportModel):
    """Product slice report."""
    services = types.ListType(types.ModelType(SliceService, required=True))
    slices = types.ListType(types.ModelType(Slice, required=False))

    def __repr__(self):
        return '{}<services={}, slices={}>'.format(self.__class__.__name__, dict(services=self.services),
                                                   dict(slices=self.slices))


class SliceBalanceSettings(ReportModel):
    """SliceBalanceReport Settings."""
    allowed_metadata_over_provision_factor = types.IntType(
        required=True, serialized_name=u'allowedMetadataOverprovisionFactor')


class SliceBalanceLayout(ReportModel):
    """SliceBalanceReport Generic Layout."""
    service_loads = types.DictType(types.BaseType, required=True, serialized_name=u'serviceLoads')
    settings = types.ModelType(SliceBalanceSettings, required=True, serialized_name=u'settings')
    slices = types.DictType(types.BaseType, required=True, serialized_name=u'slices')


class SliceBalanceDecisions(ReportModel):
    """SliceBalanceReport Decisions."""
    to = types.ModelType(models.SliceAssignment, required=True, serialized_name=u'to')
    type = types.StringType(required=True, serialized_name=u'type')
    old_slice_info = types.ModelType(models.Slice, required=True, serialized_name=u'oldSliceInfo')


class SliceBalanceZoneLayouts(ReportModel):
    """SliceBalanceReport ZoneLayouts."""
    zone_layout = types.DictType(types.ModelType(models.ZoneLayout),
                                 required=True, serialized_name=u'zoneLayout')


class SliceBalanceReport(ReportModel):
    """Product Slice Balance Report."""
    thought_process = types.ListType(types.BaseType, required=True,
                                     serialized_name=u'thoughtProcess')
    goal_layout = types.ModelType(SliceBalanceLayout, required=True,
                                  serialized_name=u'goalLayout')
    original_layout = types.ModelType(SliceBalanceLayout, required=True,
                                      serialized_name=u'originalLayout')
    decisions = types.ListType(types.ModelType(SliceBalanceDecisions),
                               required=True, serialized_name=u'decisions')
    zone_layouts = types.ListType(types.BaseType, required=True, serialized_name=u'zoneLayouts')

    def __repr__(self):
        return '{}<thoughtProcess={}, goalLayout={}, originalLayout={}, ' \
               'decisions={}, zoneLayouts={}>'.format(self.__class__.__name__, self.thought_process,
                                                      self.goal_layout, self.original_layout,
                                                      self.decisions, self.zone_layouts)
