"""This module stores global state of resources.

Resource data models represent resources that are instantiated at the beginning of a test suite and
are immutable.
"""
import autologger
import sf_models

_log = autologger.get_sflogging(__name__)

# Global variables that store test resources
_clusters = {}
_nodes = None
_nodes_by_static_id = None


def set_cluster_resources(cluster_dms):
    """Sets global test clusters resource state.

    :param list[valence.models.ResourceCluster] cluster_dms: Cluster data models.
    """
    # pylint: disable=global-statement
    global _clusters

    if not isinstance(cluster_dms, list) or not all(
            isinstance(c, sf_models.ResourceCluster) for c in cluster_dms):
        return

    for cluster_dm in cluster_dms:
        _clusters[cluster_dm.mvip] = cluster_dm


def set_node_resources(node_dms):
    """Sets global test resource state.

    :param list[valence.models.ResourceNode] node_dms: Resource node data models
    """
    # pylint: disable=global-statement
    global _nodes
    global _nodes_by_static_id

    if _nodes is not None:
        _log.warning('Global resource nodes have already been set. Previous nodes will be overridden.')

    # Make sure the node_dms are the right type since the suite_resources could be overridden
    if not isinstance(node_dms, list) or not all(
            isinstance(n, sf_models.ResourceNode) for n in node_dms):
        return

    _nodes = tuple(node_dms)
    _nodes_by_static_id = {n.static_id: n for n in _nodes}


def get_resource_clusters():
    """Return all cluster resources.

    :rtype: tuple[valence.models.ResourceCluster]
    """
    return tuple(_clusters.values())


def get_resource_nodes():
    """Return all node resources.

    :rtype: tuple[valence.models.ResourceNode]
    """
    return _nodes


def get_resource_cluster_by_static_id(static_id):
    """Return cluster resource with given mvip.

    Will check the 'mvip' attribute on the global cluster dms.

    :param str static_id: The static id to filter on
    :return: Cluster resource with given static id
    :rtype: valence.models.ResourceCluster|None
    """
    return _clusters.get(static_id) if _clusters else None


def get_resource_node_by_static_id(static_id):
    """Return node resource with the given static_id.

    Will check the 'static_id' attribute on the global node dms.

    :param str static_id: The static id to filter on
    :return: Node resource with given static id
    :rtype: valence.models.ResourceNode|None
    """
    return _nodes_by_static_id.get(static_id) if _nodes_by_static_id else None


def _clear_resources():
    """Sets all global state to None.

    This is a helper method that should only be used for unit tests
    """
    # pylint: disable=global-statement
    global _clusters
    global _nodes
    global _nodes_by_static_id
    _clusters = {}
    _nodes = None
    _nodes_by_static_id = None
