"""Node Data Models.

Includes resource node data models and versions of SF models with references to a resource node.
"""
import sf_models

from camelot.models import at2_constants
from sf_api import models
from sf_api.models import types
from valence.models import _resource_network


class NodeBase(models.Base):
    """Abstract model that defines properties to connect an SF node with a resource node."""
    _resource_dm_class = sf_models.ResourceNode

    @property
    def resource_dm(self):
        """Make resource dm public to allow users to access this attribute.

        :rtype: valence.models.ResourceNode
        """
        return self._resource_dm

    @property
    def oob_ip(self):
        return self.resource_dm.oob_ip

    @property
    def node_name(self):
        return self.resource_dm.node_name

    @property
    def ssh_username(self):
        return self.resource_dm.username

    @property
    def ssh_password(self):
        return self.resource_dm.password

    @property
    def resource_id(self):
        return self.resource_dm.resource_id

    @property
    def location_id(self):
        return self.resource_dm.location_id

    @property
    def fibre_channel(self):
        return self.resource_dm.fibre_channel

    @property
    def model(self):
        return self.resource_dm.model

    @property
    def network_profile(self):
        return self.resource_dm.network_profile


class ActiveNode(models.ActiveNode, NodeBase):
    """ActiveNode data model.

    Corresponds to SF object returned by ListAvailableNodes and contains reference to a ResourceNode.
    """


class AvailableNode(models.AvailableNode, NodeBase):
    """AvailableNode data model.

    Corresponds to SF object returned by ListActiveNodes and contains reference to a ResourceNode.
    """


class PendingNode(models.PendingNode, NodeBase):
    """PendingNode data model.

    Corresponds to SF object returned by ListPendingNodes and contains reference to a ResourceNode.
    """


class PendingActiveNode(models.PendingActiveNode, NodeBase):
    """PendingActiveNode data model.

    Corresponds to SF object returned by ListPendingActiveNodes and contains reference to a ResourceNode.
    """


class LegacyNodeAdapter(object):
    """Node adapter class to use machine control in a more generic way.

    Since the node data models have changed between legacy and Valence code we need a way to populate
    a temporary node like object for use with machine control. To use provide a Valence node data model
    and use the returned object as a node within machine control code only.

    :param valence.models.ResourceNode node_dm: A valence node data model
    """
    def __init__(self, node_dm):
        self.name = node_dm.node_name
        self.model = node_dm.model.value
        self.control_ip = node_dm.oob_ip
        self.local_attributes = getattr(node_dm, 'local_attribute', {})


class RtfiNodeAdapter(sf_models.ResourceNode):
    """Node adapter class to work with sf-rtfi."""
    network_profile = types.ModelType(_resource_network.RtfiNetworkProfileAdapter,
                                      serialized_name=at2_constants.NETWORK_PROFILE)
    model = types.StringType()

    # These are the things we 'DON"T HAVE'
    def to_dict(self):
        """Converts the data model to a serializable dictionary.

        :rtype: dict
        """
        return self.serialize()

    @property
    def ip_address(self):
        return self.mip

    @property
    def local_attributes(self):
        if not hasattr(self, '_attributes'):
            self._attributes = {}  # pylint: disable=attribute-defined-outside-init
        return self._attributes

    @local_attributes.setter
    def local_attributes(self, obj):
        """Get and set local attributes.

        :param dict obj: replace it
        """
        assert isinstance(obj, dict)
        self._attributes = obj  # pylint: disable=attribute-defined-outside-init

    @property
    def name(self):
        return self.node_name

    @property
    def location(self):
        return self.network_profile.location_name  # pylint: disable=no-member

    @property
    def network(self):
        return self.network_profile

    @property
    def is_virtual(self):
        return sf_models.NodeType(self.model).is_virtual

    @property
    def control_ip(self):
        return self.oob_ip

    @property
    def type(self):
        return self.resource_type_name
