"""Test RawStats."""
from __future__ import print_function

import unittest

import pytest

from sf_api.models import errors
from valence import models


class TestRawStats(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.volume_0_stats = models.VolumeRawStats.create(volume_id=0, throttle_0=90)
        cls.volume_1_stats = models.VolumeRawStats.create(volume_id=1, throttle_0=55,
                                                          write_block_size_16k=222)
        cls.service_2_stats = models.ServiceRawStats.create(ss_load_0=90)
        cls.disk_3_stats = models.DiskRawStats.create()

        cls.result = {'volume-0': cls.volume_0_stats,
                      'volume-1': cls.volume_1_stats,
                      'service-2': cls.service_2_stats,
                      'disk-3': cls.disk_3_stats
                      }

    def setUp(self):
        self.raw_stats = models.RawStats.create(
            result=self.result
        )

    def test_raw_stats_repr(self):
        assert isinstance(repr(self.raw_stats), str)

    def test_get_volume_ids(self):
        ids = self.raw_stats.get_volume_ids()
        assert isinstance(ids, list), 'get_volume_ids should return a list'
        assert 1 in ids, 'Missing volume id=1'
        assert len(ids) == 2, 'Expected 2 volume ids'

    def test_get_disk_ids(self):
        ids = self.raw_stats.get_disk_ids()
        assert isinstance(ids, list), 'get_disk_ids should return a list'
        assert 3 in ids, 'Missing volume id=3'

    def test_get_service_ids(self):
        ids = self.raw_stats.get_service_ids()
        assert isinstance(ids, list), 'get_service_ids should return a list'
        assert 2 in ids, 'Missing volume id=2'
        assert ids[0] not in (0, 1, 3), 'Missing volume id=2'

    def test_get_volume_stats_dms(self):
        dms = self.raw_stats.get_volume_stats()
        assert isinstance(dms, list), 'get_volume_stats should return a list'
        assert isinstance(dms[0], models.VolumeRawStats), 'Expected VolumeRawStats data model.'
        assert len(dms) == 2, 'Expected 2 VolumeStats data models.'

    def test_get_volume_stats_dm_by_id(self):
        dms = self.raw_stats.get_volume_stats([0])
        assert isinstance(dms, list), 'get_volume_stats should return a list'
        assert len(dms) == 1, 'Expected 1 VolumeStats data model.'
        assert dms[0].volume_id == 0, 'Expected VolumeStats id=0'

    def test_get_disk_stats_dms(self):
        dms = self.raw_stats.get_disk_stats()
        assert isinstance(dms, list), 'get_disk_stats should return a list'
        assert isinstance(dms[0], models.DiskRawStats), 'Expected DiskRawStats data model.'
        assert len(dms) == 1, 'Expected 1 DiskStats data model.'

    def test_get_disk_stats_dm_by_id(self):
        dms = self.raw_stats.get_disk_stats([3])
        assert isinstance(dms, list), 'get_disk_stats should return a list'
        assert len(dms) == 1, 'Expected 1 DiskStats data model.'
        assert dms[0].disk_id == 3, 'Expected DiskStats id=3'

    def test_get_service_stats_dms(self):
        dms = self.raw_stats.get_service_stats()
        assert isinstance(dms, list), 'get_service_stats should return a list'
        assert isinstance(dms[0], models.ServiceRawStats), 'Expected ServiceRawStats data model.'
        assert len(dms) == 1, 'Expected 1 ServiceStats data model.'

    def test_get_service_stats_dm_by_id(self):
        dms = self.raw_stats.get_service_stats([2])
        assert isinstance(dms, list), 'get_service_stats should return a list'
        assert len(dms) == 1, 'Expected 1 ServiceStats data model.'
        assert dms[0].service_id == 2, 'Expected ServiceStats id=2'

    def test_generate_ssload_histogram(self):
        dm = self.raw_stats.generate_ssload_histogram()
        assert isinstance(dm, models.QuintileHistogram), ('generate_ssload_histogram should return '
                                                          'a valence.models.QuintileHistogram')
        assert dm.bucket_0 == self.service_2_stats.ss_load_0

    def test_generate_volume_qos_histograms(self):
        dms = self.raw_stats.generate_volume_qos_histograms()
        assert isinstance(dms, dict), 'generate_volume_qos_histograms should return a map.'
        assert len(dms) == 2, 'Expected 2 VolumeQosHistograms.'
        assert isinstance(dms[0], models.VolumeQosHistograms)
        assert dms[0].volume_id == 0, 'Expected VolumeQosHistograms for volume-0.'
        assert dms[0].histograms.throttle_percentages.bucket_0 == self.volume_0_stats.throttle_0

    def test_validate_ssload_histogram_pass(self):
        ssload_dm = self.raw_stats.generate_ssload_histogram()
        self.raw_stats.validate_ssload_histogram(min_ssload=ssload_dm, max_ssload=ssload_dm)

    def test_validate_ssload_histogram_bad_min(self):
        ssload_dm = self.raw_stats.generate_ssload_histogram()
        min_dm = self.raw_stats.generate_ssload_histogram()
        min_dm.bucket_0 = ssload_dm.bucket_0 + 100
        self.raw_stats = models.RawStats.create(result=self.result)
        with pytest.raises(errors.DataValidationError) as dve:
            self.raw_stats.validate_ssload_histogram(min_ssload=min_dm, max_ssload=ssload_dm)
        assert 'Previous' in repr(dve.value)

    def test_validate_ssload_histogram_bad_max(self):
        ssload_dm = self.raw_stats.generate_ssload_histogram()
        max_dm = self.raw_stats.generate_ssload_histogram()
        max_dm.bucket_0 = ssload_dm.bucket_0 - 100
        with pytest.raises(errors.DataValidationError) as dve:
            self.raw_stats.validate_ssload_histogram(min_ssload=ssload_dm, max_ssload=max_dm)
        assert 'Current' in repr(dve.value)

    def test_validate_volume_histogram(self):
        hist_dm = self.raw_stats.generate_volume_qos_histograms()[0]
        self.raw_stats.validate_volume_histogram(min_volume_hist=hist_dm, max_volume_hist=hist_dm)

    def test_validate_volume_histogram_bad_min(self):
        hist_dm = self.raw_stats.generate_volume_qos_histograms()[0]
        min_hist = self.raw_stats.generate_volume_qos_histograms()[1]
        min_hist.histograms.throttle_percentages.bucket_0 += 100
        self.raw_stats = models.RawStats.create(result=self.result)
        with pytest.raises(errors.DataValidationError) as dve:
            self.raw_stats.validate_volume_histogram(min_volume_hist=min_hist,
                                                     max_volume_hist=hist_dm)
        assert 'Previous' in repr(dve.value)

    def test_validate_volume_histogram_bad_max(self):
        hist_dm = self.raw_stats.generate_volume_qos_histograms()[0]
        max_hist = self.raw_stats.generate_volume_qos_histograms()[1]
        max_hist.histograms.throttle_percentages.bucket_0 = \
            hist_dm.histograms.throttle_percentages.bucket_0 - 100
        with pytest.raises(errors.DataValidationError) as dve:
            self.raw_stats.validate_volume_histogram(min_volume_hist=hist_dm,
                                                     max_volume_hist=max_hist)
        assert 'Current' in repr(dve.value)

    def test_is_valid_volume_histogram(self):
        hist_dm = self.raw_stats.generate_volume_qos_histograms()[0]
        is_valid = self.raw_stats.is_valid_volume_histogram(min_volume_hist=hist_dm,
                                                            max_volume_hist=hist_dm)
        assert is_valid, 'Expected the VolumeQosHistogram to be valid=True.'

    def test_is_valid_volume_histogram_false(self):
        hist_dm = self.raw_stats.generate_volume_qos_histograms()[0]
        max_hist = self.raw_stats.generate_volume_qos_histograms()[1]
        max_hist.histograms.throttle_percentages.bucket_0 = \
            hist_dm.histograms.throttle_percentages.bucket_0 - 100
        is_valid = self.raw_stats.is_valid_volume_histogram(min_volume_hist=hist_dm,
                                                            max_volume_hist=max_hist)
        assert not is_valid, 'Expected the VolumeQosHistogram to be valid=False.'
