"""Test reports."""
from __future__ import print_function

import json
import os
import unittest

import mock
import pytest

from sf_api import api_constants
from sf_api.models import errors
from valence import (
    enums,
    models,
    unittest_utils,
)

BIN_STATUS = enums.BinStatus
DATA_PATH = unittest_utils.get_data_dir(__file__)


class TestReports(unittest.TestCase):

    def test_get_bin(self):
        with open(os.path.join(DATA_PATH, 'bins.json')) as bin_file:
            data = json.load(bin_file)[0]
        # TODO: Add this assertion back when VL-2228 is fixed
        # assert models.BinReport(data).serialize(version=sf_version.SfVersion('9.0')) == data
        assert isinstance(repr(models.BinReport(data)), str)
        assert isinstance(repr(models.BinAssignments.get_mock_object()), str)
        assert isinstance(repr(models.BinService.get_mock_object()), str)

    def test_get_slice(self):
        with open(os.path.join(DATA_PATH, 'slices.json')) as slice_file:
            data = json.load(slice_file)
        # Replacing optional times with None.
        for s in data['slices']:
            for r in s['remoteReplicationInfo']['repBranchInfos']:
                r['lastSyncedTime'] = None
        assert models.SliceReport(data).serialize() == data
        assert isinstance(repr(models.SliceReport(data)), str)

    def test_get_slice_balance(self):
        with open(os.path.join(DATA_PATH, 'slicebalance.json')) as slice_balance_file:
            data = json.load(slice_balance_file)
        data_serz = models.SliceBalanceReport(data).serialize()
        assert data == data_serz
        assert isinstance(repr(models.SliceBalanceReport(data)), str)

    def test_standby_service_info_in_slice(self):
        with open(os.path.join(DATA_PATH, 'standbyservices.json')) as standby_service_file:
            data = json.load(standby_service_file)
        serialized_data = models.SliceReport(data).serialize()
        serialized_standbys = [slice_[api_constants.STANDBY_SERVICES] for slice_ in
                               serialized_data[api_constants.SLICES]]
        standbys = [slice_[api_constants.STANDBY_SERVICES] for slice_ in data[api_constants.SLICES]]
        for standby in serialized_standbys:
            assert standby in standbys
            assert isinstance(repr(standby), str)
        for standby in standbys:
            assert standby in serialized_standbys
            assert isinstance(repr(standby), str)

    def test_standby_service_info(self):
        with open(os.path.join(DATA_PATH, 'standbyservices.json')) as standby_service_file:
            data = json.load(standby_service_file)
        sb_data = data[api_constants.SLICES][0][api_constants.STANDBY_SERVICES][0]
        standby_service_model = models.StandbyServiceInfo(sb_data)
        assert isinstance(repr(standby_service_model), str)
        assert sb_data == standby_service_model.serialize()


class TestBinAssignments(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.protection_domain_layout = [
            models.ProtectionDomainServices.create(
                protection_domain_name='0',
                services=[models.ProtectionDomainServiceInfo.create(service_id=0, budget=1)]
            ),
            models.ProtectionDomainServices.create(
                protection_domain_name='1',
                services=[models.ProtectionDomainServiceInfo.create(service_id=1, budget=1)]
            ),
            models.ProtectionDomainServices.create(
                protection_domain_name='2',
                services=[models.ProtectionDomainServiceInfo.create(service_id=2, budget=1)]
            )
        ]
        cls.bin_assignment_properties = models.BinAssignmentsProperties.get_mock_object(
                overrides=dict(
                    bin_count=3,
                    replication_count=2,
                    is_balanced=True,
                    layout=cls.protection_domain_layout
                )
            )
        cls.bin_reports = [
            models.BinReport.create(bin_id=0, services=[
                models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
                models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1)
            ]),
            models.BinReport.create(bin_id=1, services=[
                models.BinService.create(service_id=1, status=BIN_STATUS.active, replica=0),
                models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1)
            ]),
            models.BinReport.create(bin_id=2, services=[
                models.BinService.create(service_id=1, status=BIN_STATUS.active, replica=0),
                models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=1)
            ])
        ]

    def setUp(self):
        self.bin_assignments = models.BinAssignments.create(
            properties=self.bin_assignment_properties,
            bins=[bin_report for bin_report in self.bin_reports]
        )

    def test_properties_invalid(self):
        self.bin_assignments.properties.replication_count = 'invalid_type'
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_bin_count_invalid(self):
        self.bin_assignments.properties.bin_count = len(self.bin_assignments.bins) - 1
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_service_unbalanced(self):
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
                models.BinService.create(service_id=1, status=BIN_STATUS.active, replica=0),
                models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_service_is_balanced_false(self):
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
                models.BinService.create(service_id=1, status=BIN_STATUS.active, replica=0),
                models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1)
        ])
        self.bin_assignments.properties.is_balanced = False
        self.bin_assignments.validate()

    def test_invalid_replica(self):
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=2)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_domain_violation(self):
        self.bin_assignments.bins[0] = models.BinReport.create(services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.pending_removal_active, replica=0)
        ])
        self.bin_assignments.properties.layout.append(
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[
                    models.ProtectionDomainServiceInfo.create(service_id=0, budget=1),
                    models.ProtectionDomainServiceInfo.create(service_id=2, budget=1)
                ]
            )
        )
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_active(self):
        self.bin_assignments.validate()

        # bin assignment with pra
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_active, replica=1),
        ])
        self.bin_assignments.properties.layout.append(
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[models.ProtectionDomainServiceInfo.create(service_id=3, budget=1)]
            ))
        self.bin_assignments.validate()

        # Multiple active assignments for the same replica
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.active, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_no_updating(self):
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.pending_removal_active, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_pending_active(self):

        # No pending-removal assignments
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.pending_active, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        # Multiple PRDs
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.pending_active, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_dead, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        self.bin_assignments.properties.layout.extend([
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[models.ProtectionDomainServiceInfo.create(service_id=3, budget=1)]
            ),
            models.ProtectionDomainServices.create(
                protection_domain_name='4',
                services=[models.ProtectionDomainServiceInfo.create(service_id=4, budget=1)]
            )])
        self.bin_assignments.validate()

        # Single PRA
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.pending_active, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_active, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        self.bin_assignments.validate()

        # Multiple PRAs
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.pending_active, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_active, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_active, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_updating(self):
        # No pending-removal assignments
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        # Multiple PRDs
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_dead, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        self.bin_assignments.properties.layout.extend([
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[models.ProtectionDomainServiceInfo.create(service_id=3, budget=1)]
            ),
            models.ProtectionDomainServices.create(
                protection_domain_name='4',
                services=[models.ProtectionDomainServiceInfo.create(service_id=4, budget=1)]
            )])
        self.bin_assignments.validate()

        # Single PRA
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_active, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        self.bin_assignments.validate()

        # Multiple PRAs
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_active, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_active, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        # Sanity tests for ELEM-7862
        # Validate with a PRDA status
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_dead, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead_from_active,
                                     replica=1)
        ])
        self.bin_assignments.validate()

        # Validate with a UA status
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating_from_active, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_dead, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        self.bin_assignments.validate()

    def test_updating_no_sync(self):
        # No pending-removal assignments
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating_no_sync, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        # Multiple PRDs
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating_no_sync, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_dead, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        self.bin_assignments.properties.layout.extend([
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[models.ProtectionDomainServiceInfo.create(service_id=3, budget=1)]
            ),
            models.ProtectionDomainServices.create(
                protection_domain_name='4',
                services=[models.ProtectionDomainServiceInfo.create(service_id=4, budget=1)]
            )])
        self.bin_assignments.validate()

        # Single PRA
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating_no_sync, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.pending_removal_active, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.pending_removal_dead, replica=1)
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_zone_has_two_replicas(self):
        self.bin_assignments.properties.layout.append(
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[
                    models.ProtectionDomainServiceInfo.create(service_id=3, budget=1),
                    models.ProtectionDomainServiceInfo.create(service_id=4, budget=1)
                ]
            )
        )
        self.bin_assignments.bins[0] = models.BinReport.create(services=[
            models.BinService.create(service_id=3, status=BIN_STATUS.updating, replica=1),
            models.BinService.create(service_id=4, status=BIN_STATUS.updating_no_sync, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.updating_no_sync, replica=1),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=0),

        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    def test_service_in_multiple_zones(self):

        self.bin_assignments.properties.layout.append(models.ProtectionDomainServices.create(
                protection_domain_name='4',
                services=[models.ProtectionDomainServiceInfo.create(service_id=0, budget=1)]
            ))
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()
        self.bin_assignments.properties.layout = self.bin_assignments.properties.layout[:-1]
        self.bin_assignments.validate()

    def test_multiple_bin_status_in_same_zone(self):
        # order is independent, just need all combinations of Active, updating, UNS and pending active.
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.updating, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.updating_no_sync, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.pending_active, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.updating, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.updating, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.updating, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.updating_no_sync, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.updating, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.pending_active, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.updating_no_sync, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.updating_no_sync, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.updating_no_sync, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.pending_active, replica=0),
        ])
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.pending_active, replica=0),
            models.BinService.create(service_id=0, status=BIN_STATUS.pending_active, replica=0),
        ])

    def test_bin_count_plus(self):
        self.bin_assignments.bins[0] = models.BinReport.create(bin_id=0, services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1),
        ])
        self.bin_assignments.validate()
        self.bin_assignments.properties.layout.append(
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[
                    models.ProtectionDomainServiceInfo.create(service_id=0, budget=2),
                    models.ProtectionDomainServiceInfo.create(service_id=2, budget=2)
                ]
            )
        )
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate()

    @mock.patch.object(models.BinAssignments, '_validate_bin_assignment')
    def test_bin_assignments_not_balanced_by_budget(self, validate_bin_assignment):
        self.bin_assignments.properties.bin_count = 8
        self.bin_assignments.bins.extend([
            models.BinReport.create(services=[
                models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
                models.BinService.create(service_id=2, status=BIN_STATUS.active, replica=1),
            ])]*5)
        with pytest.raises(errors.DataValidationError) as dve:
            self.bin_assignments.validate()
        assert 'service with unbalanced bin assignments=2' in str(dve.value)

    @mock.patch.object(models.BinAssignments, '_validate_bin_assignment')
    def test_layout_budget_missing(self, validate_bin_assignment):
        self.bin_assignments.bins[0] = models.BinReport.create(services=[
            models.BinService.create(service_id=0, status=BIN_STATUS.active, replica=0),
            models.BinService.create(service_id=2, status=BIN_STATUS.updating, replica=1),
            models.BinService.create(service_id=3, status=BIN_STATUS.updating, replica=1)
        ])
        with pytest.raises(errors.DataValidationError) as dve:
            self.bin_assignments.validate()
        assert 'service missing a bins budget=1' in str(dve.value)

    @mock.patch.object(models.BinAssignments, '_validate_bin_assignment')
    def test_pre_magnesium(self, validate_bin_assignment):
        assert self.bin_assignments.replica_based

        # Remove all Magnesium additions to bin assignments
        self.bin_assignments.properties = None
        for bin_report in self.bin_assignments.bins:
            for bin_service in bin_report.services:
                bin_service.replica = None

        assert not self.bin_assignments.replica_based
        self.bin_assignments.validate()
        assert not validate_bin_assignment.called

    def test_pd_invalid_service_level(self):
        self.bin_assignments.properties.layout.append(
            models.ProtectionDomainServices.create(
                protection_domain_name='3',
                services=[
                    models.ProtectionDomainServiceInfo.create(service_id=0),
                    models.ProtectionDomainServiceInfo.create(service_id=2)
                ]
            )
        )
        with pytest.raises(errors.DataValidationError):
            self.bin_assignments.validate_service_level_pd()

    def test_pd_valid_service_level(self):
        self.bin_assignments.properties.layout.extend([
            models.ProtectionDomainServices.create(
                protection_domain_name='0',
                services=[models.ProtectionDomainServiceInfo.create(service_id=0, budget=1)]
            ),
            models.ProtectionDomainServices.create(
                protection_domain_name='2',
                services=[models.ProtectionDomainServiceInfo.create(service_id=2, budget=1)]
            )
        ])
        self.bin_assignments.properties.service_stranded_capacities = []
        self.bin_assignments.validate_service_level_pd()

    @mock.patch.object(models.BinAssignments, 'validate_service_level_pd')
    def test_is_service_level(self, validate_service_level_pd):
        assert self.bin_assignments.is_service_level()
        assert validate_service_level_pd.called
        self.bin_assignments.is_service_level()
        assert validate_service_level_pd.call_count == 1

    @mock.patch.object(models.BinAssignments, 'validate_service_level_pd')
    def test_is_service_level_false(self, validate_service_level_pd):
        validate_service_level_pd.side_effect = errors.DataValidationError({})
        assert not self.bin_assignments.is_service_level()
