"""Tests valence.models.resource and the resource data models."""

import os
import unittest

import mock

from valence import models
from valence.models import _resource

DATA_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')


class TestResourceModule(unittest.TestCase):
    """Tests functions used for setting and retrieving resource models."""

    @classmethod
    def setUpClass(cls):
        cls.node_dms = models.ResourceNode.get_mock_object_list(2)
        cls.cluster_dms = models.ResourceCluster.get_mock_object_list(2)

    def setUp(self):
        _resource._clear_resources()

    def test_get_resource_nodes(self):
        models.set_node_resources(self.node_dms)
        nodes = models.get_resource_nodes()
        assert len(nodes) == len(self.node_dms)
        assert nodes[0] is self.node_dms[0]

    def test_get_resource_clusters(self):
        models.set_cluster_resources(self.cluster_dms)
        clusters = models.get_resource_clusters()
        assert len(clusters) == len(self.cluster_dms)
        assert clusters[0] in self.cluster_dms
        assert clusters[1] in self.cluster_dms

    @mock.patch('autologger.main.ContextualLogger.warning')
    def test_set_resources_twice(self, mock_warn):
        """A warning will be logged when attempting to set resources twice."""
        models.set_node_resources(self.node_dms)
        models.set_node_resources(self.node_dms)
        mock_warn.assert_called()
        models.set_cluster_resources(self.cluster_dms)
        models.set_cluster_resources(self.cluster_dms)
        mock_warn.assert_called()

    def test_set_resources_wrong_type(self):
        models.set_node_resources([{'something': 'wrong'}])
        assert models.get_resource_nodes() is None
        models.set_cluster_resources([{'more': 'wrong'}])
        assert bool(models.get_resource_clusters()) is False

    def test_get_resource_node_by_static_id(self):
        models.set_node_resources(self.node_dms)
        node = self.node_dms[0]
        mip = node.mip
        returned_node = models.get_resource_node_by_static_id(mip)
        assert node is returned_node

    def test_get_resource_cluster_by_static_id(self):
        models.set_cluster_resources(self.cluster_dms)
        cluster = self.cluster_dms[0]
        mvip = cluster.mvip
        returned_cluster = models.get_resource_cluster_by_static_id(mvip)
        assert cluster is returned_cluster

    def test_get_resource_node_by_static_id_nodes_not_set(self):
        assert not models.get_resource_node_by_static_id('node_mip_foo')


class TestResourceModels(unittest.TestCase):
    """Tests serialization and deserialization of resource data models."""
    def test_repr(self):
        assert isinstance(repr(models.AsyncHandle()), str)
        assert isinstance(repr(models.AsyncResult()), str)
        assert isinstance(repr(models.AddDrivesDetails()), str)
        assert isinstance(repr(models.CloneVolumeDetails()), str)
        assert isinstance(repr(models.ResourceOntap()), str)
        assert isinstance(repr(models.ArtifactCredentials()), str)
        assert isinstance(repr(models.SuiteConfig()), str)
        assert isinstance(repr(models.SuiteResources()), str)
        assert isinstance(repr(models.SyncJob()), str)
        assert isinstance(repr(models.RepoChangeset()), str)
        assert isinstance(repr(models.CamelotLogFilter()), str)
        assert isinstance(repr(models.HypervisorHost()), str)
        assert isinstance(repr(models.VirtualHardwareProfile()), str)
        assert isinstance(repr(models.WitnessResource()), str)
        assert isinstance(repr(models.HipChatAPI()), str)
        assert isinstance(repr(models.SwitchPort()), str)


class TestSuiteResources(unittest.TestCase):
    def test_properties(self):
        """This provides code coverage on the properties."""
        resources_dm = models.SuiteResources.get_mock_object()
        assert len(resources_dm.clusters) == len(resources_dm.cluster_dms)
        assert len(resources_dm.nodes) == len(resources_dm.node_dms)
        assert len(resources_dm.management_nodes) == len(resources_dm.management_node_dms)
        assert len(resources_dm.clients) == len(resources_dm.client_dms)
        assert len(resources_dm.networks) == len(resources_dm.network_dms)
        assert len(resources_dm.virtual_networks) == len(resources_dm.virtual_network_dms)
        assert len(resources_dm.hypervisors) == len(resources_dm.hypervisor_dms)
        assert len(resources_dm.ip_ranges) == len(resources_dm.ip_range_dms)
        assert len(resources_dm.ontaps) == len(resources_dm.ontap_dms)
        assert len(resources_dm.compute_nodes) == len(resources_dm.compute_node_dms)
        assert len(resources_dm.ips) == len(resources_dm.ip_dms)

        assert isinstance(repr(resources_dm), str)


class TestClientModel(unittest.TestCase):
    """Tests client-specific get-mock-object logic."""
    def test_virt_client(self):
        client = models.ResourceClient.get_mock_object(overrides={'type': 'virtual'})
        assert client.idrac_ip == ''


class TestNetworkRTFIProfileAdapter(unittest.TestCase):
    """Test the RTFINetworkProfileAdapter."""
    def setUp(self):
        self.network_profile = models.NetworkProfile.get_mock_object()
        self.rtfi_adapter_profile = models.RtfiNetworkProfileAdapter(self.network_profile.serialize())

    def test_attributes(self):
        assert self.rtfi_adapter_profile.pxe_address == self.network_profile.pxe_server_address
        assert self.rtfi_adapter_profile.pxe_config_path == self.network_profile.pxe_server_config_path
        assert self.rtfi_adapter_profile.pxe_username == self.network_profile.pxe_server_username
        assert self.rtfi_adapter_profile.pxe_password == self.network_profile.pxe_server_password
