"""Backend monitor and associated classes."""
import collections
import copy
import sys
import threading
import time

import autologger
import logbook
import sf_batch
from six.moves import queue

from valence import (
    criteria,
    enums,
    functional_models,
    models,
    utils,
)
from valence.monitors import (
    _constants,
    _errors,
)


class ObserverTimeout(object):
    """Store a timeout that was registered by an observer.

    :param int seconds_to_wait: Time to wait before calling back method.
    :param int timeout_id: Unique id associated with this timeout.
    """

    def __init__(self, seconds_to_wait, timeout_id):
        self.seconds_to_wait = seconds_to_wait
        self.timeout_id = timeout_id
        self.start_time = time.time()

    def is_ready(self):
        """Determine if the callback is ready to be called.

        :rtype: bool
        """
        now = time.time()
        if (now - self.start_time) > self.seconds_to_wait:
            return True
        else:
            return False


class AtomicCounter(object):
    """An atomic counter that increments an integer."""

    def __init__(self):
        self.value = 0
        self._lock = threading.Lock()

    def get_next(self):
        """Get the next integer value.

        :rtype: int
        """
        with self._lock:
            self.value += 1
            return self.value


class BackendMonitor(object):
    """Backend test case monitor.

    This class spins up a thread that monitors cluster faults and events in the background as a test
    case runs.

    :param valence.models.SuiteResources suite_resources: Data model of suite resources
    :param logbook.handlers.Handler logging_handler: Typically a FileHandler for loggers to report to
    :param int poll_period: Time to wait in between polling cluster, in seconds
    """

    def __init__(self, suite_resources, logging_handler=None, poll_period=40):
        self.suite_resources = suite_resources
        self.using_custom_logger = True
        if not logging_handler:
            self.using_custom_logger = False
            logging_handler = logbook.StreamHandler(sys.stdout)
        self.handler = logging_handler
        self.log = self._create_backend_monitor_logger()
        self.poll_period = poll_period
        self._fault_whitelist = set()
        self._always_fail_faults = set(_constants.DEFAULT_ALWAYS_FAIL_FAULTS)
        self._always_fail_events = set(_constants.DEFAULT_ALWAYS_FAIL_EVENTS)
        self._always_fail_event_filters = None
        self._critical_error_event_filters = _constants.CRITICAL_ERROR_EVENTS
        self._notify_queue = queue.Queue()
        self._backend_monitor_failures = []
        self._claimed_faults = set()

        self._cluster_fm = functional_models.Cluster(self.suite_resources.cluster_dms[0],
                                                     authenticate=False)
        self._cluster_fm.log_responses = False
        self._observers = []
        self._unresolved_fault_ids = set()
        self._resolved_fault_ids = set()
        self._last_event_id = 1

        # Map of timeout_ids to Timeout objects
        self._timeout_map = {}
        self._timeouts_lock = threading.Lock()

        self._monitor_thread = None
        self._kill_flag = False
        self._set_observers([])
        self._counter = AtomicCounter()

    def _create_backend_monitor_logger(self):
        """Creates the backend monitor's Logger.

        :rtype: autologger.main.ContextualLogger
        """
        monitor_logger = autologger.get_sflogging('backend_monitor')
        monitor_logger.handlers.append(self.handler)
        return monitor_logger

    @property
    def failures(self):
        """Returns a list of failures that the monitors and observers have encountered.

        :rtype: list[valence.monitors.MonitorFailure]
        """
        all_failures = []
        for observer in self._observers:
            all_failures.extend(observer.failures)
        all_failures.extend(self._backend_monitor_failures)
        return all_failures

    def _set_observers(self, observers):
        """Instantiates all the observers and sets self._observers.

        :param list observers: A list of classes that inherit from BaseObserver
        """
        self._fault_callback_map = collections.defaultdict(list)
        self._observers = []
        self._claimed_faults = set()
        for observer in observers:
            ob = observer(self.suite_resources, self)
            self._observers.append(ob)
            if ob.claimed_fault is not None:
                self._claimed_faults.add(ob.claimed_fault)

    def set_always_fail_faults(self, always_fail_faults=_constants.DEFAULT_ALWAYS_FAIL_FAULTS):
        """Set the list of faults that will always cause a failure if they occur.

        When using the ElementTestSuite, the list will be reset to the default after every test case.

        :param always_fail_faults: A list/tuple/set of faults that will cause failure. If none is
            given, the default list will be used.
        :type always_fail_faults: typing.Iterable[valence.enums.ClusterFaultCode]
        """
        self._always_fail_faults = set(always_fail_faults)

    def get_always_fail_faults(self):
        """Get a copy of the current faults to always fail on.

        This set can be modified and passed to set_always_fail_faults.

        :rtype: set[valence.enums.ClusterFaultCode]
        """
        return copy.copy(self._always_fail_faults)

    def add_always_fail_fault(self, fault):
        """Add fault to list of faults that will always cause a test failure.

        :param valence.enums.ClusterFaultCode fault: fault to add to always fail faults list.
        """
        self._always_fail_faults.add(fault)

    def remove_always_fail_fault(self, fault):
        """Remove fault from list of faults that will always cause a test failure.

        :param valence.enums.ClusterFaultCode fault: fault to remove from always fail fault list.
        """
        if fault in self._always_fail_faults:
            self._always_fail_faults.remove(fault)

    def set_always_fail_events(self, always_fail_events=_constants.DEFAULT_ALWAYS_FAIL_EVENTS):
        """Set the list of events that will always cause a failure if they occur.

        When using the ElementTestSuite, the list will be reset to the default after every test case.
        # noqa

        :param always_fail_events: A list/tuple/set of events that will cause failure. If none is
            given, the default list will be used.
        :type always_fail_events: typing.Iterable[valence.enums.ClusterEventType|valence.criteria.BaseCriteria]
        """
        self._always_fail_events = set()
        self._always_fail_event_filters = []
        for event in always_fail_events:
            self.add_always_fail_event(event)

    def get_always_fail_events(self):
        """Get a copy of the current events to always fail on.

        This set can be modified and passed to set_always_fail_events.

        :rtype: set[valence.enums.ClusterEventType]
        """
        return copy.copy(self._always_fail_events)

    def add_always_fail_event(self, event):
        """Add event to list of events that will always cause a test failure.

        :param event: event to add to the always fail events list.
        :type event: valence.enums.ClusterEventType|valence.criteria.BaseCriteria
        """
        if isinstance(event, enums.ClusterEventType):
            self._always_fail_events.add(event)
        elif isinstance(event, criteria.BaseCriteria):
            self._always_fail_event_filters.append(event)

    def remove_always_fail_event(self, event):
        """Remove event from list of events that will always cause a test failure.

        :param valence.enums.ClusterEventType event: event to remove from list of always fail events.
        """
        if event in self._always_fail_events:
            self._always_fail_events.remove(event)

    def is_running(self):
        """Return True if the monitor is currently running in a background thread.

        :rtype: bool
        """
        if self._monitor_thread is not None and self._monitor_thread.is_alive():
            return True
        else:
            return False

    def notify(self, method_call):
        """Push the method_call object onto a queue, to be processed later.

        This is called by the notify_monitor decorator that is used on functional model methods.

        :param valence.functional_models.MethodCall method_call: Object that holds info about the fm call
        """
        self._notify_queue.put(method_call)

    def start_monitoring(self):
        """Start the monitoring process.

        :raises valence.monitors.MonitorError: If the monitor is already running
        """
        self.set_as_global_monitor()
        if self._monitor_thread is not None and self._monitor_thread.is_alive():
            raise _errors.MonitorError('Monitor is already running')
        self._kill_flag = False
        self._monitor_thread = sf_batch.SFThread(self._monitor_loop, thread_name='monitor')
        self._monitor_thread.start()

    def set_as_global_monitor(self):
        """Set the global backend monitor to the current object."""
        functional_models.monitor_decorators.set_backend_monitor(self)

    def _monitor_loop(self):
        """Main loop for monitoring that runs in a background thread."""

        while True:
            self.run_iteration()
            # Check the flag here rather than in while conditional to ensure the loop runs at least once
            if self._kill_flag and not self._timeout_map:
                break
            time.sleep(self.poll_period)  # pragma: no cover

    def run_iteration(self):
        """Run a single iteration of the monitor loop.

        This method is useful for testing the monitor by hand, but should not be called directly in
        a test case or suite.
        """
        self._prune_expired_timeouts()
        method_calls = self._get_method_calls()
        now = utils.get_now_utc()
        faults = self._cluster_fm.list_cluster_faults()
        events = self._cluster_fm.list_events(start_event_id=self._last_event_id+1)
        if events:
            self._last_event_id = events[0].event_id
        self._check_for_critical_error_events(events)
        self._update_observers(method_calls, faults, events, timestamp=now)
        self._check_for_unexpected_faults(faults)
        self._check_for_unexpected_events(events)

    def _prune_expired_timeouts(self):
        """Remove timeouts that have expired."""
        with self._timeouts_lock:
            self._timeout_map = {k: v for k, v in self._timeout_map.items() if not v.is_ready()}

    def _get_method_calls(self):
        """Get the method calls that have happened since last iteration.

        :rtype: list[valence.functional_models.MethodCall]
        """
        queue_length = self._notify_queue.qsize()
        method_calls = []
        for _ in range(queue_length):
            try:
                method_calls.append(self._notify_queue.get_nowait())
            except queue.Empty:  # pragma: no cover
                # This exception should only occur if some other thread gets from the queue after
                # we get the qsize. This should not happen since this is the only place where we get
                # from the queue, but it's good to be overly defensive in case that assumption changes
                # in the future.
                break
        return method_calls

    def _update_observers(self, method_calls, faults, events, timestamp):
        """Update all observers with latest method calls and faults.

        :param list[valence.functional_models.MethodCall] method_calls: List of method calls
        :param list[valence.models.ClusterFault] faults: List of faults
        :param list[valence.models.ClusterEvent] events: List of events
        :param datetime.datetime timestamp: Time when list_cluster_faults was called
        """
        fault_map = self._get_fault_map(faults)
        method_call_map = self._get_method_call_map(method_calls)
        event_map = self._get_event_map(events)
        for observer in self._observers:
            observer.update(method_call_map=method_call_map, fault_map=fault_map,
                            event_map=event_map, timestamp=timestamp)

    def _get_fault_map(self, faults):
        """Build a dict of codes mapped to faults.

        :param list[valence.models.ClusterFault] faults: List of faults
        :rtype: dict[valence.enums.ClusterFaultCode, list[valence.models.ClusterFault]
        """
        fault_map = collections.defaultdict(list)
        for fault in faults:
            if self._ignore_fault(fault):
                continue
            fault_map[fault.code].append(fault)
        return fault_map

    def _get_event_map(self, events):
        """Build a dict of event_info_type mapped to events.

        :param list[valence.models.ClusterEvent] events: List of events
        :rtype: dict[valence.enums.ClusterEventType, list[valence.models.ClusterEvent]]
        """
        event_map = collections.defaultdict(list)
        for event in events:
            event_map[event.event_info_type].append(event)
        return event_map

    def _ignore_fault(self, fault):
        """Determine if the fault does not need to be sent to observers.

        This is true for faults that the observers already know about, and have not been updated.

        :param valence.models.ClusterFault fault: Fault object to filter
        :rtype: bool
        :return: True, if fault does not need to be sent to the observers
        """
        fault_id = fault.cluster_fault_id
        if fault_id in self._resolved_fault_ids:
            return True
        if fault.resolved:
            if fault_id in self._unresolved_fault_ids:
                self._unresolved_fault_ids.remove(fault_id)
            self._resolved_fault_ids.add(fault_id)
        else:
            if fault_id in self._unresolved_fault_ids:
                # The observers have already been alerted, but fault is not yet resolved
                return True
            else:
                # Fault is occurring for the first time
                self._unresolved_fault_ids.add(fault_id)
        return False

    @staticmethod
    def _get_method_call_map(method_calls):
        """Build a dict of methods mapped to MethodCall objects.

        :param list[valence.functional_models.MethodCall] method_calls: List of method calls
        :rtype: dict[types.MethodType, list[valence.functional_models.MethodCall]]
        """
        method_map = collections.defaultdict(list)
        for method_call in method_calls:
            method_map[method_call.method].append(method_call)
        return method_map

    def register_timeout(self, seconds_to_wait):
        """Register a timeout that can be used to block a test case while monitoring wraps up.

        This is needed because a test case or suite may finish before the expected faults and events
        have had a chance to occur.

        :param int seconds_to_wait: Seconds to wait from the current time.
        :rtype: int
        :returns: A unique id associated with this timeout
        """
        timeout_id = self._counter.get_next()
        with self._timeouts_lock:
            self._timeout_map[timeout_id] = ObserverTimeout(seconds_to_wait, timeout_id)
        return timeout_id

    def deregister_timeout(self, timeout_id):
        """Allows a timeout to be de-registered.

        Deletes the object with timeout_id from the timeout_map

        :param int timeout_id: unique id associated with timeout
        """
        with self._timeouts_lock:
            if timeout_id in self._timeout_map:
                del self._timeout_map[timeout_id]

    def _check_for_unexpected_faults(self, faults):
        for fault in faults:
            if fault.resolved:
                continue
            if not self._fault_is_expected(fault.code):
                msg = 'Unexpected fault occurred: {}'.format(fault)
                self._store_failure(msg, fault.date)

    def _fault_is_expected(self, fault_code):
        """Determine if the fault is expected according to whitelist and always fail list.

        :param valence.enums.ClusterFaultCode fault_code: Code of the fault to check
        :rtype: bool
        """
        if fault_code in self._claimed_faults:
            # An observer has taken responsibility for this fault, so it will raise any failures
            return True
        elif self._fault_whitelist and fault_code not in self._fault_whitelist:
            return False
        elif fault_code in self._always_fail_faults:
            return False
        else:
            return True

    def _check_for_unexpected_events(self, events):
        event_filter = criteria.ContainedIn(models.ClusterEvent.event_info_type,
                                            self._always_fail_events)
        if self._always_fail_event_filters:
            filters = [event_filter] + self._always_fail_event_filters
            event_filter = criteria.Or(*filters)
        unexpected_events = event_filter.filter(events)
        for event in unexpected_events:
            msg = 'Unexpected event occurred: {}'.format(event)
            self._store_failure(msg, event.time_of_report)

    def _check_for_critical_error_events(self, events):
        event_filter = criteria.Or(*self._critical_error_event_filters)
        unexpected_events = event_filter.filter(events)
        if unexpected_events:
            msg = ('Critical failure detected. Please leave cluster up for debugging purposes '
                   'and contact Ankit.Kumar@netapp.com.\nUnexpected Event(s): {}').format(
                unexpected_events)
            raise _errors.MonitorCriticalError(msg)

    def _store_failure(self, message, timestamp):
        """Log and store a failure.

        :param str message: A string describing the failure
        :param datetime.datetime timestamp: Time when failure occurred
        """
        self._backend_monitor_failures.append(_errors.MonitorFailure(message, timestamp))
        self.log.debug(message)

    def stop_monitoring(self, raise_failures=False):
        """Stop the monitoring process.

        If there are outstanding timeout callbacks, this method will block until all of them are
        completed.

        :param bool raise_failures: If True, raise an exception if any failures occurred during
            monitoring.
        """
        self._kill_flag = True
        self._monitor_thread.join()
        if raise_failures:
            self.get_failures(raise_failures=True)

    def get_failures(self, start_time=None, end_time=None, raise_failures=False):
        """Get failures that have occurred during monitoring.

        :param datetime.datetime start_time: Only select failures that occurred on or after this start
            time
        :param datetime.datetime end_time: Only select failures that occurred on or before this end
            time
        :param bool raise_failures: If True, raise an exception if there are any failures.
        :rtype: list[valence.monitors.MonitorFailure]
        :raises valence.monitors.errors.MonitorError: If there are any accumulated errors and
            raise_on_failure is True
        """
        failures = self.failures
        if start_time:
            failures = [f for f in failures if f.timestamp >= start_time]
        if end_time:
            failures = [f for f in failures if f.timestamp <= end_time]
        if failures and raise_failures:
            raise _errors.MonitorError('Failures occurred when monitoring: {}'.format(
                '\n'.join([f.message for f in failures])
            ))
        return failures

    def wait_for_timeouts(self):
        """Wait for all timeouts that have been registered by the observers.

        Use in a test case to block and wait until all timeouts are depleted before teardown.
        """
        while self._timeout_map:
            time.sleep(self.poll_period // 4)  # pragma: no cover
