"""Monitor exceptions."""

from valence import errors
from valence.utils import time_utils


class Error(errors.Error):
    """Base exception for monitors package.

    This error should not be raised directly.
    """


class MonitorError(Error):
    """Backend monitor error."""


class MonitorCriticalError(Error):
    """Backend monitor critical error.

    This exception indicates a monitor error that should immediately stop the execution of the test
    suite without running teardown. It is used to keep the cluster up for debugging when a rare fault
    or event occurs.
    """


class MonitorFailure(object):
    """Stores info about a monitor failure.

    This class itself is not an exception, but stores info needed to identify a monitor failure.

    :param str message: A message describing the failure
    :param datetime.datetime timestamp: A timestamp of when the failure occurred. If None is given,
        use the current time.
    """
    def __init__(self, message, timestamp=None):
        self.message = message
        self.timestamp = timestamp if timestamp is not None else time_utils.get_now_utc()

    def serialize(self):
        """Convert object to a dictionary.

        :rtype: dict
        """
        return {
            'message': self.message,
            'timestamp': self.timestamp.isoformat(),
        }

    def __repr__(self):
        return 'MonitorFailure<message={}, timestamp={}>'.format(self.message, self.timestamp)
