"""Monitor Query class."""

import bisect
import collections

from valence import (
    criteria,
    models,
    utils,
)


class MonitorQuery(object):
    """Holds and queries state used by the observers.

    :param list[types.MethodType] methods_to_store: List of methods that will be stored by this object.
    :param list[valence.enums.ClusterFaultCode] faults_to_store: Codes of faults that will be stored by
        this object.
    :param list[valence.enums.ClusterEventType] events_to_store: Types of events that will be stored by
        this object
    """

    def __init__(self, methods_to_store, faults_to_store, events_to_store):
        # This timestamp will be updated each time self.update() is called
        self.last_updated = utils.get_now_utc()

        self._events_to_store = set(events_to_store)
        self._methods_to_store = set(methods_to_store)
        self._faults_to_store = set(faults_to_store)
        self._method_call_map = collections.defaultdict(list)
        self._fault_map = collections.defaultdict(list)
        self._event_map = collections.defaultdict(list)

    def update(self, method_call_map, fault_map, event_map, timestamp=None):
        """Update object with relevant data.

        :param method_call_map: A map of unbound methods to lists of MethodCall objects
        :type method_call_map: dict[types.MethodType, list[valence.functional_models.MethodCall]]
        :param fault_map: A map of fault codes to lists of ClusterFault objects
        :type fault_map: dict[valence.enums.ClusterFaultCode, list[valence.models.ClusterFault]
        :param datetime.datetime timestamp: Time when list_cluster_faults was called
        :param event_map: A map of event types to lists of ClusterEvent objects
        :type event_map: dict[valence.enums.ClusterEventType, list[valence.models.ClusterEvent]]
        """
        self.last_updated = timestamp or utils.get_now_utc()
        self._update_method_calls(method_call_map)
        self._update_faults(fault_map)
        self._update_events(event_map)

    def _update_method_calls(self, method_call_map):
        """Update internal method call data structure.

        :param method_call_map: A map of unbound methods to lists of MethodCall objects
        :type method_call_map: dict[types.MethodType, list[valence.functional_models.MethodCall]]
        """
        for method, method_calls in method_call_map.items():
            if method in self._methods_to_store:
                self._method_call_map[method].extend(method_calls)

    def _update_faults(self, fault_map):
        """Update internal fault data structure.

        :param fault_map: A map of fault codes to lists of ClusterFault objects
        :type fault_map: dict[valence.enums.ClusterFaultCode, list[valence.models.ClusterFault]
        """
        for fault_code, faults in fault_map.items():
            if fault_code in self._faults_to_store:
                faults.sort(key=lambda f: f.date)
                self._fault_map[fault_code].extend(faults)

    def _update_events(self, events_map):
        """Update internal event data structure.

        :param events_map: a map of event types to lists of ClusterEvent objects
        :type events_map: dict[valence.enums.ClusterEventType, list[valence.models.ClusterEvent]]
        """
        for event_type, events in events_map.items():
            if event_type in self._events_to_store:
                self._event_map[event_type].extend(events)

    def get_method_calls(self, method):
        """Return a list of MethodCall objects for the given method.

        :param types.MethodType method: An unbound method, e.g. functional_models.Cluster.add_drives
        :rtype: list[valence.functional_models.MethodCall]
        """
        return self._method_call_map.get(method, [])

    def get_faults(self, fault_code, details_regex=None, node_id=None, date_range=None):
        """Return a list of ClusterFault objects, sorted by date.

        :param valence.enums.ClusterFaultCode fault_code: Code to get faults for
        :param str details_regex: A regex to filter by the details field of the fault. This is needed
            since some of the faults are used for too many purposes and need to be differentiated by
            the details field.
        :param int node_id: The node ID that is associated with the fault
        :param tuple[datetime.datetime,datetime.datetime] date_range: Filter on faults where the date
            attribute is within the given date range.
        :rtype: list[valence.models.ClusterFault]
        """
        faults = self._fault_map.get(fault_code, [])
        regexes = [details_regex] if details_regex else None
        fault_filter = criteria.And(
            criteria.MatchIn(models.ClusterFault.details, regexes),
            criteria.Equals(models.ClusterFault.node_id, node_id),
        )
        faults = fault_filter.filter(faults)
        if date_range:
            dates = [f.date for f in faults]
            left_index = bisect.bisect_left(dates, date_range[0])
            right_index = bisect.bisect_right(dates, date_range[1])
            faults = faults[left_index:right_index]
        return faults

    def get_events(self, event_type, time_range=None):
        """Returns a list of ClusterEvent objects.

        :param valence.enums.ClusterEventType event_type: type of event to get events for
        :param tuple[datetime.datetime,datetime.datetime] time_range: Filter on events
            where the time_of_report attribute is within a given time range
        :rtype: list[valence.models.ClusterEvent]
        """
        events = self._event_map.get(event_type, [])
        if time_range:
            times = [e.time_of_report for e in events]
            left_index = bisect.bisect_left(times, time_range[0])
            right_index = bisect.bisect_right(times, time_range[1])
            events = events[left_index:right_index]
        return events
