"""Base observer class.

This class defines the interface that all observers must implement.
"""
import abc

import six

from valence.monitors import _errors


class BaseObserver(six.with_metaclass(abc.ABCMeta, object)):
    """Abstract base class for observers.

    :param valence.models.SuiteResources suite_resources: Suite resources data model
    :param valence.monitors.BackendMonitor monitor: Instance of the backend monitor
    """

    def __init__(self, suite_resources, monitor):
        self.suite_resources = suite_resources
        self.monitor = monitor
        self.query = self._get_monitor_query()
        self._failures = []

    @property
    def claimed_fault(self):
        """Fault that observer has taken responsibility for.

        If the observer is responsible for an event, this does not need to be overridden.

        :rtype: valence.enums.ClusterFaultCode
        """
        return None

    @property
    def log(self):
        """Property to expose the background monitor's logger.

        :rtype: autologger.main.ContextualLogger
        """
        return self.monitor.log

    @property
    def failures(self):
        """Returns a list of failures that the observer has encountered.

        :rtype: list[valence.monitors.MonitorFailure]
        """
        return self._failures

    def _store_failure(self, message, timestamp):
        """Log and store a failure.

        :param str message: A string describing the failure
        :param datetime.datetime timestamp: Time when failure occurred
        """
        self.monitor.log.debug('Observer Failure: {}'.format(message))
        failure = _errors.MonitorFailure(message, timestamp)
        self._failures.append(failure)

    @abc.abstractmethod
    def _get_monitor_query(self):
        """Return the MonitorQuery object.

        :rtype: valence.monitors.MonitorQuery
        """

    def update(self, method_call_map, fault_map, event_map, timestamp=None):
        """Update observer with latest data on method calls and faults.

        The observer will store any failures it detects in response to the update. This is called on
        each iteration of the backend monitor loop.

        :param method_call_map: A map of unbound methods to lists of MethodCall objects
        :type method_call_map: dict[types.MethodType, list[valence.functional_models.MethodCall]]
        :param fault_map: A map of fault codes to lists of ClusterFault objects
        :type fault_map: dict[valence.enums.ClusterFaultCode, list[valence.models.ClusterFault]
        :param datetime.datetime timestamp: Time when list_cluster_faults was called
        :param event_map: A map of event types to lists of ClusterEvent objects
        :type event_map: dict[valence.enums.ClusterEventType, list[valence.models.ClusterEvent]]
        """
        self.query.update(method_call_map=method_call_map, fault_map=fault_map,
                          event_map=event_map, timestamp=timestamp)
        self._check_for_failures()

    @abc.abstractmethod
    def _check_for_failures(self):
        """Check for any failures, and store them to be raised later.

        This is called on each iteration of the backend monitor loop after updating the MonitorQuery
        object.
        """
