"""Code to handle monitored driveAvailable faults."""
# pylint: skip-file
import time

from valence import (
    enums,
    functional_models,
)
from valence.monitors import (
    _errors,
    monitor_query,
)
from valence.monitors.observers import base_observer


class Node(object):
    """The expected node object."""
    def __init__(self):
        self.network_state = enums.NetworkStatus.up
        self.power_state = 'up'
        self.fault_type = ''
        self.fault_resolved = True
        self.timer_id = -1  # cleared
        self.time_of_perturbation = None
        self.recopy_started = False


class DriveAvailableObserver(base_observer.BaseObserver):
    """Observer to look for driveAvailable faults and raise exceptions if a fault is unexpected.

    :param valence.models.SuiteResources suite_resources: The working environment
    :param valence.monitors.BackendMonitor monitor: Instance of the backend monitor
    """

    # maximum time for drive to receive driveAvailable fault after getting blockServiceUnhealthy fault,
    # in seconds.
    DRIVE_FAIL_TO_AVAILABLE_MAX_TIMEOUT = 600  # needs to be investigated
    DRIVE_ADD_MAX_TIMEOUT = 600  # needs to be investigated
    DRIVE_REMOVE_MAX_TIMEOUT = 600  # needs to be investigated
    DID_NOT_GET = ' did not get the expected driveAvailable fault in time'

    def __init__(self, suite_resources, monitor):
        super(DriveAvailableObserver, self).__init__(suite_resources, monitor)
        self.cluster_dm = suite_resources.cluster_dms[0]
        self.cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        self.nodes_by_id = {}

    def _get_monitor_query(self):
        return monitor_query.MonitorQuery(
            methods_to_store=[
                functional_models.Cluster.remove_drives,
                functional_models.Cluster.add_drives,
            ],
            faults_to_store=[enums.ClusterFaultCode.drive_available],
            events_to_store=[]
        )

    def _clear_node_expectation(self, node_id):
        """Clear the node_expectation entry for the given node.

        :param int node_id: The ID of the node to clear
        """
        self.nodes_by_id[node_id].network_state = enums.NetworkStatus.up
        self.nodes_by_id[node_id].power_state = 'up'
        self.nodes_by_id[node_id].fault_type = ''
        self.nodes_by_id[node_id].fault_resolved = True
        self.nodes_by_id[node_id].timer_id = -1  # cleared
        self.nodes_by_id[node_id].time_of_perturbation = None
        self.nodes_by_id[node_id].timeout_value = 0
        self.nodes_by_id[node_id].recopy_started = False

    def _handle_drive_available_fault(self, fault):
        """Handle the driveAvailable fault.

        :param sf_api.models.ClusterFault fault: The driveAvailable fault object.
        """
        if not fault.resolved:
            # Check the fault node and make sure the fault was expected
            if (self.nodes_by_id[fault.node_id].fault_type is
                    enums.ClusterFaultCode.drive_available and
                    not self.nodes_by_id[fault.node_id].fault_resolved):
                # The expected fault has occurred
                # Clear the node expectation.
                self._clear_node_expectation(fault.node_id)
        else:
            # Check the fault node and make sure the fault was expected
            if (self.nodes_by_id[fault.node_id].fault_type is
                    enums.ClusterFaultCode.drive_available and
                    self.nodes_by_id[fault.node_id].fault_resolved):
                # The expected fault has occurred
                # Clear the node expectation.
                self._clear_node_expectation(fault.node_id)

    def _fault_callback(self, fault):
        """Called when ever a registered fault is detected.

        :param sf_api.models.ClusterFault fault: A fault object
        """
        if fault.node_id not in self.nodes_by_id:
            return

        # driveAvailable fault issued when drive(s) change state to Available.
        if fault.code is enums.ClusterFaultCode.drive_available:
            self._handle_drive_available_fault(fault)

    def _timeout_callback(self, timer_id):
        """Called when the timer times out.

        :param int timer_id: An ID to identify which timer is timing out
        :raises valence.monitors.errors.MonitorError: when any expired timer finds an expected fault
            which has not been received.
        """

        # Iterate through all node expectations and find any for the given timer_id that have not been
        # cleared.
        timeout_value = 0
        bad_nodes = []
        for node_id in self.nodes_by_id:
            if self.nodes_by_id[node_id].timer_id == timer_id:
                # Here is a node that did not get the fault in time -- mark for exception
                self._clear_node_expectation(node_id)
                bad_nodes.append(node_id)
                timeout_value = self.nodes_by_id[node_id].timeout_value

        if bad_nodes:
            raise _errors.MonitorError('Node IDs {}{}: {} seconds'.format(bad_nodes, self.DID_NOT_GET,
                                                                          timeout_value))

    def _set_network_perturbation(self, node_id, net_state, perturb_time):
        """Set expected dict for given drive for a perturbation.

        :param int node_id: The ID of the affected node
        :param valence.enums.NetworkStatus net_state: The new state of the network.
        :param float perturb_time: The timestamp when the perturbation arrived.
        """
        # Mark the network_state
        self.nodes_by_id[node_id].network_state = net_state

        # Mark the time the perturbation was received
        self.nodes_by_id[node_id].time_of_perturbation = perturb_time

    def _set_forthcoming_node_fault(self, node_id, fault_type, fault_resolved, timer_id,
                                    perturb_time, timeout_value):
        """Set the given drive to expect the given fault.

        :param int node_id: The ID of the drive to expect the fault on
        :param sf_api.enums.ClusterFaultCode fault_type: The fault type to expect for the drive
        :param bool fault_resolved: The fault is resolved (true of false)
        :param int timer_id: The ID of the timer. Used to drives together by same perturbations.
        :param float perturb_time: The timestamp when the perturbation arrived.
        :param int timeout_value: The timeout value for this timer.
        """
        self.nodes_by_id[node_id].fault_type = fault_type
        self.nodes_by_id[node_id].fault_resolved = fault_resolved
        self.nodes_by_id[node_id].timer_id = timer_id
        self.nodes_by_id[node_id].time_of_perturbation = perturb_time
        self.nodes_by_id[node_id].timeout_value = timeout_value

    def _handle_network_interface_perturbations(self, node_id, net_state, perturb_time):
        """Handle network interface perturbation on the given node.

        :param int node_id: the node ID of the hosting node
        :param valence.enums.NetworkStatus net_state: The state of the perturbation, 'up' or 'down'
        :param float perturb_time: The timestamp when the perturbation arrived.
        """
        if node_id not in self.nodes_by_id:
            return

        # PERTURBATION: network state has changed to 'down'
        if net_state is enums.NetworkStatus.down:
            # mark the time the perturbation started and mark network_state
            self._set_network_perturbation(node_id, net_state, perturb_time)

        # RESOLUTION: network being brought back up
        else:
            timer_id = -1
            self._set_network_perturbation(node_id, net_state, perturb_time)

            # If the network is coming back up, and recopy of drive(s) has already started,
            # expect driveAvailable fault within time limits.
            if self.nodes_by_id[node_id].recopy_started:
                if timer_id == -1:
                    timer_id = self.register_timeout_callback(self.DRIVE_FAIL_TO_AVAILABLE_MAX_TIMEOUT)
                self._clear_node_expectation(node_id)
                self._set_forthcoming_node_fault(node_id, enums.ClusterFaultCode.drive_available, False,
                                                 timer_id, perturb_time,
                                                 self.DRIVE_FAIL_TO_AVAILABLE_MAX_TIMEOUT)

            # If the network is coming back up, and the drive recopy has not already started,
            # the drive will go back to Active, no fault expected.
            else:
                self._clear_node_expectation(node_id)

    def _handle_add_drive_perturbation(self, drive_list, perturb_time):
        """Handle drive perturbation for addition of the given drive.

        :param list[valence.models.Drive] drive_list: A list of Drives to add
        :param float perturb_time: The timestamp when the perturbation arrived.
        """
        node_id_set = set(d.node_id for d in drive_list)  # yields a list of unique node IDs
        timer_id = self.register_timeout_callback(self.DRIVE_ADD_MAX_TIMEOUT)

        for node_id in node_id_set:
            self._clear_node_expectation(node_id)
            self._set_forthcoming_node_fault(node_id, enums.ClusterFaultCode.drive_available, True,
                                             timer_id, perturb_time, self.DRIVE_ADD_MAX_TIMEOUT)

    def _handle_remove_drive_perturbation(self, drive_list, perturb_time):
        """Handle drive perturbation for removal of the given drive.

        :param list[valence.models.Drive] drive_list: A list of drives to remove
        :param float perturb_time: The timestamp when the perturbation arrived.
        """
        node_id_set = set(d.node_id for d in drive_list)  # yields a list of unique node IDs
        timer_id = self.register_timeout_callback(self.DRIVE_REMOVE_MAX_TIMEOUT)

        for node_id in node_id_set:
            self._clear_node_expectation(node_id)
            self._set_forthcoming_node_fault(node_id, enums.ClusterFaultCode.drive_available, False,
                                             timer_id, perturb_time, self.DRIVE_REMOVE_MAX_TIMEOUT)

    def _update_expect(self):
        """Take current node states and update expect object with it."""
        current_nodes = self.cluster_fm.list_active_nodes()
        current_node_ids = [n.node_id for n in current_nodes]
        past_node_ids = self.nodes_by_id.keys()

        # First add any new nodes to nodes_by_id dict
        added_nodes = set(current_node_ids) - set(past_node_ids)
        if added_nodes:
            for new_node in current_nodes:
                if new_node.node_id in added_nodes:
                    self.nodes_by_id[new_node.node_id] = Node()

        # Next remove any non-existent drives from drives_by_id dict
        removed_nodes = set(past_node_ids) - set(current_node_ids)
        if removed_nodes:
            for gone_node in current_nodes:
                if gone_node.node_id in removed_nodes:
                    del self.nodes_by_id[gone_node.node_id]

            # Leave everything else in the object as it is

    def _notify(self, method_call):
        """This method is called when a functional model method is called in the test environment.

        :param valence.functional_models.MethodCall method_call: Object that holds info about the fm call
        """
        # Refresh current drive list. Handle new node_ids or disappeared node_ids
        if not self.nodes_by_id:  # First time? Initialize
            self.nodes_by_id = {n.node_id: Node() for n in self.cluster_fm.list_active_nodes()}
        else:
            self._update_expect()

        perturb_time = time.time()

        if method_call.is_method(functional_models.Cluster.remove_drives):
            self._handle_remove_drive_perturbation(method_call.arguments['drive_dms'], perturb_time)

        elif method_call.is_method(functional_models.Cluster.add_drives):
            self._handle_add_drive_perturbation(method_call.arguments['drive_dms'], perturb_time)

        # Need ID of the node this network interface belongs to
        # elif method.__name__ == 'set_network_interface_status_and_verify':
        #    self._handle_network_interface_perturbations(node_id, args['status'], perturb_time)
