"""Node hardware fault observer class."""
import datetime

import autologger

from valence import (
    enums,
    functional_models,
)
from valence.monitors import monitor_query
from valence.monitors.observers import base_observer

_log = autologger.get_sflogging()


class NodeHardwareFaultObserver(base_observer.BaseObserver):
    """Node Hardware Fault observer.

    This observer only checks for the fault where the Bond1G is down.

    :param valence.models.SuiteResources suite_resources: Suite resources data model
    :param valence.monitors.BackendMonitor monitor: backend monitor
    """

    # NodeHardwareFault is overloaded, so we need to filter them based on the details field
    DETAILS_RE = '.+Bond1G is down.+'

    def __init__(self, suite_resources, monitor):
        super(NodeHardwareFaultObserver, self).__init__(suite_resources, monitor)

        # The fault should occur or be resolved within this amount of time after the api call
        self.max_timeout = 75

        cluster_dm = suite_resources.cluster_dms[0]
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

    def _get_monitor_query(self):
        return monitor_query.MonitorQuery(
            methods_to_store=[functional_models.Node.set_network_interface_status_and_verify],
            faults_to_store=[enums.ClusterFaultCode.node_hardware_fault],
            events_to_store=[]
        )

    @property
    def claimed_fault(self):
        """Fault that observer takes responsibility for.

        :rtype: valence.enums.ClusterFaultCode
        """
        return enums.ClusterFaultCode.node_hardware_fault

    def _check_for_failures(self):
        """This method is called during each iteration of the monitor loop.

        If any failures are detected, they are logged and stored.
        """
        node_ids_to_static_ids = self._get_node_id_to_static_id_map()
        expected_fault_ids = set()

        method_calls = self.query.get_method_calls(
            functional_models.Node.set_network_interface_status_and_verify)
        for method_call in method_calls:
            fault_ids = self._process_method_call(method_call, node_ids_to_static_ids)
            expected_fault_ids.update(fault_ids)

        self._check_for_unexpected_faults(expected_fault_ids)

    def _check_for_unexpected_faults(self, expected_fault_ids):
        faults = self.query.get_faults(self.claimed_fault)
        unexpected_faults = [f for f in faults if f.cluster_fault_id not in expected_fault_ids]
        for fault in unexpected_faults:
            self._store_failure('No associated method call for fault: {}'.format(fault), fault.date)

    def _process_method_call(self, method_call, node_ids_to_static_ids):
        """Process an up/down Bond1G method call.

        :param valence.monitors.MethodCall method_call: The Bond1G method call
        :param dict[int,Any]] node_ids_to_static_ids: Map of node_ids to static_ids
        :rtype: list[int]
        :returns: The fault ids of all the faults that were expected based on the method call
        """
        expected_fault_ids = []
        faults = self._get_faults_associated_with_method_call(method_call, node_ids_to_static_ids)
        if len(faults) > 1:
            self._store_failure('Multiple faults seen for single call: {}'.format(method_call),
                                method_call.timestamp)
        elif self._is_down_bond_1g_call(method_call):
            expected_fault_ids = self._process_down_bond_1g_call(method_call, faults)
        elif self._is_up_bond_1g_call(method_call):
            self._process_up_bond_1g_call(method_call, faults)
        return expected_fault_ids

    def _process_down_bond_1g_call(self, method_call, faults):
        """Process a down Bond1G method call.

        This includes verifying that the expected fault happens within the required time. If the fault
        has not occurred yet, a timeout will be registered with the monitor.

        :param valence.monitors.MethodCall method_call: The down Bond1G method call
        :param list[valence.models.ClusterFault] faults: List of faults
        :rtype: list[int]
        :returns: The fault ids of all the faults that were expected based on the method call
        """
        if not faults:
            if not self._within_timeout(method_call.timestamp, self.query.last_updated):
                self._store_failure(
                    'NodeHardware fault did not occur in time after: {}'.format(method_call),
                    method_call.timestamp)
            elif method_call.timeout_id is None:
                delta = (self.query.last_updated - method_call.timestamp).total_seconds()
                method_call.timeout_id = self.monitor.register_timeout(self.max_timeout - delta)
        elif not self._within_timeout(method_call.timestamp, faults[0].date):
            self._store_failure('NodeHardware fault did not occur in time after: {}'.format(method_call),
                                method_call.timestamp)
        return [f.cluster_fault_id for f in faults]

    def _process_up_bond_1g_call(self, method_call, faults):
        """Process an up Bond1G method call.

        This includes verifying that the expected fault is resolved within the required time. If the
        fault has not been resolved yet, a timeout will be registered with the monitor.

        :param valence.monitors.MethodCall method_call: The down Bond1G method call
        :param list[valence.models.ClusterFault] faults: List of faults
        """
        if not faults:
            if not self._within_timeout(method_call.timestamp, self.query.last_updated):
                self._store_failure(
                    'NodeHardware fault did not resolve in time after: {}'.format(method_call),
                    method_call.timestamp)
            elif method_call.timeout_id is None:
                delta = (self.query.last_updated - method_call.timestamp).total_seconds()
                method_call.timeout_id = self.monitor.register_timeout(self.max_timeout - delta)
        elif not self._within_timeout(method_call.timestamp, faults[0].resolved_date):
            self._store_failure(
                'NodeHardware fault did not resolve in time after: {}'.format(method_call),
                method_call.timestamp)

    def _get_faults_associated_with_method_call(self, method_call, node_ids_to_static_ids):
        date_range = (method_call.timestamp,
                      method_call.timestamp + datetime.timedelta(seconds=self.max_timeout))
        faults = self.query.get_faults(self.claimed_fault, details_regex=self.DETAILS_RE,
                                       date_range=date_range)
        if self._is_up_bond_1g_call(method_call):
            # We only care about resolved faults for calls that bring the Bond1G up
            faults = [f for f in faults if f.resolved is True]
        # Filter by static id
        faults = [f for f in faults if node_ids_to_static_ids[f.node_id] == method_call.static_id]
        return faults

    def _within_timeout(self, start_time, end_time):
        delta = (end_time - start_time).total_seconds()
        return delta <= self.max_timeout

    def _get_node_id_to_static_id_map(self):
        """Get a map of node_ids to static_ids for all active nodes.

        The fault objects only have a node id, so this is needed to correlate a fault to a node.

        TODO: remove this once VL-2091 is implemented

        :rtype: dict[str, str]
        """
        result = {}
        nodes = self.cluster_fm.list_active_nodes()
        for node in nodes:
            result[node.node_id] = node.static_id
        return result

    @staticmethod
    def _is_up_bond_1g_call(method_call):
        args = method_call.arguments
        if (args['interfaces'] == [enums.NetworkInterface.bond_1g] and
                args['status'] == enums.NetworkStatus.up_and_running):
            return True
        else:
            return False

    @staticmethod
    def _is_down_bond_1g_call(method_call):
        args = method_call.arguments
        if (args['interfaces'] == [enums.NetworkInterface.bond_1g] and
                args['status'] == enums.NetworkStatus.down):
            return True
        else:
            return False
