"""Unit tests for drive_available_observer.py."""
import datetime
import unittest

import mock
import pytest
from mock import MagicMock

from valence import (
    enums,
    functional_models,
    models,
)
from valence.monitors import _errors
from valence.monitors.observers.drive_available_observer import (
    DriveAvailableObserver,
    Node,
)


class BackendMonitor(object):
    pass


class MockNodeDataModel(object):
    """A mock node data model class."""
    node_id = '23'


class MockDataModel(object):
    """A mock data model class."""
    node_dm = MockNodeDataModel()


class MockNode(object):
    """Mock node object.

    :param str state: node state
    :param int nid: node ID
    """
    def __init__(self, state, nid):
        self.state = state
        self.node_id = nid


class Drive(object):
    """Mock drive object.

    :param int did: drive ID
    :param str state: drive state
    :param int nid: node ID
    """
    def __init__(self, did, state, nid):
        self.drive_id = did
        self.state = state
        self.node_id = nid


class MockClusterFunctionalModel(object):
    """Mock cluster functional model.

    :param int number_nodes: the number of nodes to simulate.
    """
    def __init__(self, number_nodes):
        self.number_nodes = number_nodes

    def list_active_nodes(self):
        """Mock the list_active_nodes() method of the cluster functional model.

        :rtype: list[Any]
        """
        nodes = []
        for n in range(self.number_nodes):
            nodes.append(MockNode('active', n + 1))
        return nodes


class Model(object):
    """A mock model class."""
    pass


class ClusterObj(object):
    """A mock cluster object class."""
    cluster_dm = {'ip': '10.10.10.10'}
    model = Model()


class SuiteResources(object):
    """A mock suite resource class."""
    my_cluster = ClusterObj()
    cluster_dms = [my_cluster]


class ClusterFault(object):
    """Mock ClusterFault object.

    :param str code: the fault code
    :param bool resolved: whether fault is resolved or not
    :param int node_id: node ID
    :return:
    """
    def __init__(self, code, resolved, node_id):
        self.code = code
        self.resolved = resolved
        self.node_id = node_id


@pytest.mark.skipif(True, reason='Disable until major refactor is complete')
class TestNotify(unittest.TestCase):
    """Test the observer notify method."""
    feb_27_2018_15_30_14 = 1519770614.0

    def setUp(self):
        """Set up a couple of mock models."""
        self.cluster_fm = functional_models.Cluster.get_mock_fm()
        self.cluster_dm = models.Cluster.get_mock_object()

    def unknown_command(self):
        """Mock method for class."""
        pass

    def set_network_interface_status_and_verify(self):
        """Mock method for class."""
        pass

    def remove_drives(self):
        """Mock method for class."""
        pass

    def add_drives(self):
        """Mock method for class."""
        pass

    @mock.patch('valence.functional_models.monitor_decorators.MethodCall.is_method')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_update_expect')
    @mock.patch('valence.functional_models.Cluster.add_drives')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_add_drive_perturbation')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_notify_add_drives_nodes_by_id_is_empty(self, mock_time, mock_hnip, mock_hadp, mock_rfcb,
                                                    mock_fmc, mock_cad, mock_ue, mock_is_method):
        """Unit test notify with add_drives call."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_hadp.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_cad.return_value = None
        mock_ue.return_value = None
        mock_is_method.side_effect = [False, True]  # false for remove_drives, then true for add_drives

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao._update_expect = MagicMock(return_value=None)
        dao.cluster_fm.list_active_nodes = MagicMock(return_value=[MockNode('available', 1),
                                                                   MockNode('active', 2),
                                                                   MockNode('active', 3)])
        args = {'self': MockDataModel(), 'drive_dms': [Drive(1, 'active', 1)]}
        method_call = functional_models.MethodCall(self.add_drives, args, datetime.datetime.now())
        dao._notify(method_call)

        assert (dao.cluster_fm.list_active_nodes.called and not mock_ue.called and
                mock_hadp.called)

    @mock.patch('valence.functional_models.monitor_decorators.MethodCall.is_method')
    @mock.patch('valence.functional_models.Cluster.add_drives')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_add_drive_perturbation')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_notify_add_drives(self, mock_time, mock_hnip, mock_hadp, mock_rfcb, mock_fmc, mock_cad,
                               mock_is_method):
        """Unit test notify with add_drives call."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_hadp.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_cad.return_value = None
        mock_is_method.side_effect = [False, True]  # false for remove_drives, then true for add_drives

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        args = {'self': MockDataModel(), 'drive_dms': [Drive(1, 'active', 1)]}
        method_call = functional_models.MethodCall(self.add_drives, args, datetime.datetime.now())
        dao._notify(method_call)

        assert mock_hadp.called

    @mock.patch('valence.functional_models.monitor_decorators.MethodCall.is_method')
    @mock.patch('valence.functional_models.Cluster.remove_drives')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_remove_drive_perturbation')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_notify_remove_drives(self, mock_time, mock_hnip, mock_hrdp, mock_rfcb, mock_fmc, mock_crd,
                                  mock_is_method):
        """Unit test notify with remove_drives call."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_hrdp.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_crd.return_value = None
        mock_is_method.side_effect = [True, False]  # false for remove_drives, then true for add_drives

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        self.nodes_by_id = {n.node_id: Node()
                            for n in MockClusterFunctionalModel(3).list_active_nodes()}
        args = {'self': MockDataModel(), 'drive_dms': [Drive(1, 'active', 1)]}
        method_call = functional_models.MethodCall(self.remove_drives, args, datetime.datetime.now())
        dao._notify(method_call)

        assert mock_hrdp.called

    @mock.patch('valence.functional_models.Node.set_network_interface_status_and_verify')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_notify_set_network_interface(self, mock_time, mock_hnip, mock_rfcb, mock_fmc, mock_snisav):
        """Unit test notify with set_network_interface."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_snisav.return_value = None

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        args = {'self': MockDataModel(), 'status': 'good'}
        method_call = functional_models.MethodCall(self.set_network_interface_status_and_verify,
                                                   args, datetime.datetime.now())
        dao._notify(method_call)

        assert not mock_hnip.called  # temporary until network perturbations can be supported

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_add_drive_perturbation')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_remove_drive_perturbation')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_notify_unknown_command(self, mock_time, mock_hnip, mock_hrdp, mock_hadp, mock_rfcb,
                                    mock_fmc):
        """Unit test notify with unknown command."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_hadp.return_value = None
        mock_hrdp.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        args = {}
        method_call = functional_models.MethodCall(self.unknown_command, args, datetime.datetime.now())
        dao._notify(method_call)

        assert (not mock_hnip.called and not mock_hadp.called and not mock_hrdp.called)

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_fault_callback_unhandled_fault(self, mock_time, mock_hnip, mock_rfcb, mock_fmc):
        """Unit test fault_callback with unhandled/unregistered fault."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        my_fault = ClusterFault(enums.ClusterFaultCode.drive_wear_fault, False, 1)
        dao.nodes_by_id[1].fault_type = enums.ClusterFaultCode.drive_available
        dao.nodes_by_id[1].fault_resolved = False
        dao._fault_callback(my_fault)

        assert (dao.nodes_by_id[1].fault_type is enums.ClusterFaultCode.drive_available
                and not dao.nodes_by_id[1].fault_resolved)

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_fault_callback_drive_available_fault_initial(self, mock_time, mock_hnip, mock_rfcb,
                                                          mock_fmc):
        """Unit test fault_callback with driveAvailable initial fault."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        my_fault = ClusterFault(enums.ClusterFaultCode.drive_available, False, 1)
        dao.nodes_by_id[1].fault_type = enums.ClusterFaultCode.drive_available
        dao.nodes_by_id[1].fault_resolved = False
        dao._fault_callback(my_fault)

        assert (dao.nodes_by_id[1].fault_type == '' and
                dao.nodes_by_id[1].fault_resolved and
                dao.nodes_by_id[1].timer_id == -1)

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_drive_available_fault')
    @mock.patch('time.time')
    def test_fault_callback_drive_available_fault_resolved(self, mock_time, mock_hdaf, mock_rfcb,
                                                           mock_fmc):
        """Unit test fault_callback with driveAvailable resolved fault."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hdaf.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        my_fault = ClusterFault(enums.ClusterFaultCode.drive_available, True, 1)
        dao.nodes_by_id[1].fault_type = enums.ClusterFaultCode.drive_available
        dao.nodes_by_id[1].fault_resolved = False
        dao._fault_callback(my_fault)

        assert mock_hdaf.called

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_handle_network_interface_perturbations')
    @mock.patch('time.time')
    def test_set_network_perturbation(self, mock_time, mock_hnip, mock_rfcb, mock_fmc):
        """Unit test internal method set_network_perturbation."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_hnip.return_value = None
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        dao._set_network_perturbation(1, enums.NetworkStatus.down,  # pylint: disable=W0212
                                      self.feb_27_2018_15_30_14)

        assert (dao.nodes_by_id[1].network_state is enums.NetworkStatus.down and
                dao.nodes_by_id[1].time_of_perturbation == self.feb_27_2018_15_30_14 and
                dao.nodes_by_id[2].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[2].time_of_perturbation is None and
                dao.nodes_by_id[3].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[3].time_of_perturbation is None)

    @mock.patch('valence.monitors.observers.drive_available_observer.base_observer.BaseObserver.'
                'register_timeout_callback')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_handle_network_interface_perturbations_unknown_node_net_down(
            self, mock_time, mock_rfcb, mock_fmc, mock_rtocb):
        """Unit test internal method _handle_network_interface_perturbations on unknown node ID
        with a network down.
        """
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_rtocb.return_value = 1

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        node_id = 2
        dao._handle_network_interface_perturbations(  # pylint: disable=W0212
            node_id, enums.NetworkStatus.down, self.feb_27_2018_15_30_14)

        assert (dao.nodes_by_id[1].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[1].time_of_perturbation is None and
                dao.nodes_by_id[2].network_state is enums.NetworkStatus.down and
                dao.nodes_by_id[2].time_of_perturbation is self.feb_27_2018_15_30_14 and
                dao.nodes_by_id[3].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[3].time_of_perturbation is None and
                not mock_rtocb.called)

    @mock.patch('valence.monitors.observers.drive_available_observer.base_observer.BaseObserver.'
                'register_timeout_callback')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_handle_network_interface_perturbations_unknown_node_net_up(
            self, mock_time, mock_rfcb, mock_fmc, mock_rtocb):
        """Unit test internal method _handle_network_interface_perturbations on unknown node ID
        with a network up.
        """
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_rtocb.return_value = 1

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        node_id = 2
        dao._handle_network_interface_perturbations(  # pylint: disable=W0212
            node_id, enums.NetworkStatus.up, self.feb_27_2018_15_30_14)

        assert (dao.nodes_by_id[1].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[1].time_of_perturbation is None and
                dao.nodes_by_id[2].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[2].time_of_perturbation is None and
                dao.nodes_by_id[3].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[3].time_of_perturbation is None and
                not mock_rtocb.called)

    @mock.patch('valence.monitors.observers.drive_available_observer.base_observer.BaseObserver.'
                'register_timeout_callback')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_handle_network_interface_perturbations_known_node_net_down(
            self, mock_time, mock_rfcb, mock_fmc, mock_rtocb):
        """Unit test internal method _handle_network_interface_perturbations on known node ID
        with a network down.
        """
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_rtocb.return_value = 1

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        node_id = 1
        dao._handle_network_interface_perturbations(  # pylint: disable=W0212
            node_id, enums.NetworkStatus.down, self.feb_27_2018_15_30_14)

        assert (dao.nodes_by_id[1].network_state is enums.NetworkStatus.down and
                dao.nodes_by_id[1].time_of_perturbation == self.feb_27_2018_15_30_14 and
                dao.nodes_by_id[2].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[2].time_of_perturbation is None and
                dao.nodes_by_id[3].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[3].time_of_perturbation is None and
                not mock_rtocb.called)

    @mock.patch('valence.monitors.observers.drive_available_observer.base_observer.BaseObserver.'
                'register_timeout_callback')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_handle_network_interface_perturbations_known_node_net_up_no_recopy(
            self, mock_time, mock_rfcb, mock_fmc, mock_rtocb):
        """Unit test internal method _handle_network_interface_perturbations on known node ID
        with a network up and no disk recopy started.
        """
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_rtocb.return_value = 1

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        node_id = 1
        dao._handle_network_interface_perturbations(  # pylint: disable=W0212
            node_id, enums.NetworkStatus.up, self.feb_27_2018_15_30_14)

        assert (dao.nodes_by_id[1].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[1].time_of_perturbation is None and
                dao.nodes_by_id[2].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[2].time_of_perturbation is None and
                dao.nodes_by_id[3].network_state is enums.NetworkStatus.up and
                dao.nodes_by_id[3].time_of_perturbation is None and
                not mock_rtocb.called)

    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_set_forthcoming_node_fault')
    @mock.patch('valence.monitors.observers.drive_available_observer.base_observer.BaseObserver.'
                'register_timeout_callback')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_handle_add_drive_perturbation(self, mock_time, mock_rfcb, mock_fmc, mock_rtocb,
                                           mock_sfdfbd):
        """Unit test internal method _handle_add_drive_perturbation, with 'add'."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_rtocb.return_value = 1
        mock_sfdfbd.return_value = None

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        drive_list = [Drive(1, 'active', 1)]
        dao._handle_add_drive_perturbation(drive_list,  # pylint: disable=W0212
                                           self.feb_27_2018_15_30_14)

        mock_sfdfbd.assert_called_with(1, enums.ClusterFaultCode.drive_available, True, 1,
                                       self.feb_27_2018_15_30_14, dao.DRIVE_ADD_MAX_TIMEOUT)

    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                '_set_forthcoming_node_fault')
    @mock.patch('valence.monitors.observers.drive_available_observer.base_observer.BaseObserver.'
                'register_timeout_callback')
    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_handle_remove_drive_perturbation(self, mock_time, mock_rfcb, mock_fmc, mock_rtocb,
                                              mock_sfdfbd):
        """Unit test internal method _handle_remove_drive_perturbation."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)
        mock_rtocb.return_value = 1
        mock_sfdfbd.return_value = None

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        drive_list = [Drive(1, 'active', 1)]
        dao._handle_remove_drive_perturbation(drive_list,  # pylint: disable=W0212
                                              self.feb_27_2018_15_30_14)

        mock_sfdfbd.assert_called_with(1, enums.ClusterFaultCode.drive_available, False, 1,
                                       self.feb_27_2018_15_30_14, dao.DRIVE_REMOVE_MAX_TIMEOUT)

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_set_forthcoming_node_fault_1(self, mock_time, mock_rfcb, mock_fmc):
        """Unit test internal method _set_forthcoming_node_fault for drive 1."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        dao._set_forthcoming_node_fault(  # pylint: disable=W0212
            1, enums.ClusterFaultCode.drive_available, False, 1, self.feb_27_2018_15_30_14, 10)

        assert (dao.nodes_by_id[1].fault_type is enums.ClusterFaultCode.drive_available and
                dao.nodes_by_id[1].fault_resolved is False and
                dao.nodes_by_id[1].timer_id == 1 and
                dao.nodes_by_id[1].time_of_perturbation == self.feb_27_2018_15_30_14 and
                dao.nodes_by_id[2].fault_type == '' and
                dao.nodes_by_id[2].fault_resolved is True and
                dao.nodes_by_id[2].timer_id == -1 and
                dao.nodes_by_id[2].time_of_perturbation is None and
                dao.nodes_by_id[3].fault_type == '' and
                dao.nodes_by_id[3].fault_resolved is True and
                dao.nodes_by_id[3].timer_id == -1 and
                dao.nodes_by_id[3].time_of_perturbation is None)

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_set_forthcoming_node_fault_2(self, mock_time, mock_rfcb, mock_fmc):
        """Unit test internal method _set_forthcoming_node_fault for drive 2."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        dao._set_forthcoming_node_fault(  # pylint: disable=W0212
            2, enums.ClusterFaultCode.drive_available, False, 1, self.feb_27_2018_15_30_14, 10)

        assert (dao.nodes_by_id[2].fault_type is enums.ClusterFaultCode.drive_available and
                dao.nodes_by_id[2].fault_resolved is False and
                dao.nodes_by_id[2].timer_id == 1 and
                dao.nodes_by_id[2].time_of_perturbation == self.feb_27_2018_15_30_14 and
                dao.nodes_by_id[1].fault_type == '' and
                dao.nodes_by_id[1].fault_resolved is True and
                dao.nodes_by_id[1].timer_id == -1 and
                dao.nodes_by_id[1].time_of_perturbation is None and
                dao.nodes_by_id[3].fault_type == '' and
                dao.nodes_by_id[3].fault_resolved is True and
                dao.nodes_by_id[3].timer_id == -1 and
                dao.nodes_by_id[3].time_of_perturbation is None)

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_timeout_callback_expect_no_exception_1(self, mock_time, mock_rfcb, mock_fmc):
        """Unit test internal method _timeout_callback, expect no exception for timer_id 1."""
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        dao._timeout_callback(1)  # expect no exception thrown. Test framework will capture it if so.

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_timeout_callback_expect_no_exception_2(self, mock_time, mock_rfcb, mock_fmc):
        """Internal method _timeout_callback, no exception for timer_id 1 while timer_id 2 still running.
        """
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(3)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(3).list_active_nodes()}
        dao.nodes_by_id[2].timer_id = 2
        dao._timeout_callback(1)  # expect no exception thrown. Test framework will capture it if so.

    @mock.patch('valence.functional_models.Cluster')
    @mock.patch('valence.monitors.observers.drive_available_observer.DriveAvailableObserver.'
                'register_fault_callback')
    @mock.patch('time.time')
    def test_timeout_callback_expect_exceptions(self, mock_time, mock_rfcb, mock_fmc):
        """Unit test internal method _timeout_callback, expect exceptions for 3 timer_ids, none for 4th.

        Mark 3 of 7 drives as having uncleared expected faults for timer_id 1. Expect exception
         containing 3 uncleared drive IDs.
        Mark 2 drives having uncleared expected faults for timer_id 2. Expect exception containing 2
         uncleared drive IDs.
        Mark 1 drive having uncleared expected fault for timer_id 3. Expect exception containing 1
         uncleared drive ID.
        One drive will have a cleared timer_id for timer_id 4.
        """
        mock_time.return_value = self.feb_27_2018_15_30_14
        mock_rfcb.return_value = None
        mock_fmc.return_value = MockClusterFunctionalModel(7)

        dao = DriveAvailableObserver(SuiteResources(), BackendMonitor())
        dao.nodes_by_id = {n.node_id: Node() for n in MockClusterFunctionalModel(7).list_active_nodes()}
        dao.nodes_by_id[1].timer_id = 1
        dao.nodes_by_id[2].timer_id = 1
        dao.nodes_by_id[5].timer_id = 1
        dao.nodes_by_id[3].timer_id = 2
        dao.nodes_by_id[4].timer_id = 2
        dao.nodes_by_id[6].timer_id = 3

        self.assertRaisesRegexp(_errors.MonitorError, r'\[1, 2, 5\]{}'.format(dao.DID_NOT_GET),
                                dao._timeout_callback, 1)
        self.assertRaisesRegexp(_errors.MonitorError, r'\[3, 4\]{}'.format(dao.DID_NOT_GET),
                                dao._timeout_callback, 2)
        self.assertRaisesRegexp(_errors.MonitorError, r'\[6\]{}'.format(dao.DID_NOT_GET),
                                dao._timeout_callback, 3)
        dao._timeout_callback(4)  # expect no exception thrown. Test framework will capture it if so.
