"""Test NodeHardwareFaultObserver."""

import datetime
import unittest

import mock
import pytz

from camelot import driver
from valence import (
    enums,
    functional_models,
    models,
    monitors,
)
from valence.monitors import observers


class TestNodeHardwareFaultObserver(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.suite_resources = driver.get_mock_resources_dm()
        cls.monitor = monitors.BackendMonitor(cls.suite_resources, poll_period=0)
        cls.now = datetime.datetime.now(tz=pytz.utc)

    def setUp(self):
        self.observer = observers.NodeHardwareFaultObserver(self.suite_resources, self.monitor)
        self.node = models.ActiveNode.get_mock_object()
        self.fault = models.ClusterFault.create(
            code=enums.ClusterFaultCode.node_hardware_fault,
            details='Network interface Bond1G is down or cable is unplugged.',
            resolved=False,
            node_id=self.node.node_id,
            cluster_fault_id=1,
            date=self.now + datetime.timedelta(seconds=1),
        )
        self.fault_map = {
            enums.ClusterFaultCode.node_hardware_fault: [self.fault]
        }
        self.node_fm = functional_models.Node(self.node.resource_dm,
                                              cluster_dm=models.ResourceCluster())
        self.down_bond_1g = functional_models.MethodCall(
            self.node_fm.set_network_interface_status_and_verify,
            arguments={
                'self': self.node_fm,
                'interfaces': [enums.NetworkInterface.bond_1g],
                'status': enums.NetworkStatus.down
            },
            timestamp=self.now)
        self.up_bond_1g = functional_models.MethodCall(
            self.node_fm.set_network_interface_status_and_verify,
            arguments={
                'self': self.node_fm,
                'interfaces': [enums.NetworkInterface.bond_1g],
                'status': enums.NetworkStatus.up_and_running
            },
            timestamp=self.now)

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_get_node_id_to_static_id_map(self, mock_list_nodes):
        nodes = [self.node, models.ActiveNode.get_mock_object()]
        mock_list_nodes.return_value = nodes
        result = self.observer._get_node_id_to_static_id_map()
        static_ids = set(n.static_id for n in nodes)
        assert static_ids == set(result.values())

    @mock.patch.object(observers.NodeHardwareFaultObserver, '_get_node_id_to_static_id_map')
    def test_update_no_new_faults(self, mock_get_map):
        mock_get_map.return_value = {}
        self.observer.update({}, {}, {})
        assert not self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_fault_is_resolved_but_not_in_time(self, mock_list_nodes):
        """The observer sees the resolved fault, but it took to long to resolve."""
        mock_list_nodes.return_value = [self.node]
        self.fault.resolved = True
        self.observer.max_timeout = 10
        self.fault.resolved_date = self.now + datetime.timedelta(seconds=20)
        self.up_bond_1g.timestamp = self.now
        method_call_map = self.monitor._get_method_call_map([self.up_bond_1g])
        self.observer.update(method_call_map=method_call_map,
                             fault_map=self.fault_map,
                             event_map={})
        assert self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_fault_is_unresolved_and_timeout_is_reached(self, mock_list_nodes):
        """The observer sees the fault, but it is still not resolved."""
        mock_list_nodes.return_value = [self.node]
        self.fault.resolved = False
        method_call_map = self.monitor._get_method_call_map([self.up_bond_1g])
        self.observer.update(method_call_map=method_call_map,
                             fault_map=self.fault_map,
                             event_map={},
                             timestamp=self.now + datetime.timedelta(
                                 seconds=self.observer.max_timeout+1))
        assert self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_unexpected_fault(self, mock_list_nodes):
        """A fault occurred with no method calls at all."""
        mock_list_nodes.return_value = [self.node]
        self.observer.update(method_call_map={},
                             fault_map=self.fault_map,
                             event_map={})
        assert self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_expected_fault_does_not_occur(self, mock_list_nodes):
        """There was an api call that should have caused a fault, but the fault never happened."""
        method_call_map = self.monitor._get_method_call_map([self.down_bond_1g])

        # Make it so the method call happened far enough in the past to trigger the error
        self.down_bond_1g.timestamp -= datetime.timedelta(seconds=self.observer.max_timeout)

        self.observer.update(
            method_call_map=method_call_map,
            fault_map={},
            event_map={})
        assert self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_fault_does_not_occur_in_time(self, mock_list_nodes):
        """The fault occurred, but it was outside the timeout window."""
        mock_list_nodes.return_value = [self.node]
        method_call_map = self.monitor._get_method_call_map([self.down_bond_1g])

        # Set the fault date outside the timeout window
        self.fault.date = self.down_bond_1g.timestamp + datetime.timedelta(
            seconds=self.observer.max_timeout + 1)

        self.observer.update(
            method_call_map=method_call_map,
            fault_map=self.fault_map,
            event_map={},
            timestamp=self.fault.date)
        assert self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_multiple_faults_for_one_method_call(self, mock_list_nodes):
        mock_list_nodes.return_value = [self.node]
        method_call_map = self.monitor._get_method_call_map([self.down_bond_1g])
        fault2 = models.ClusterFault(self.fault.serialize())
        fault2.cluster_fault_id += 1
        self.fault_map[enums.ClusterFaultCode.node_hardware_fault].append(fault2)
        self.observer.update(
            method_call_map=method_call_map,
            fault_map=self.fault_map,
            event_map={},
            timestamp=self.fault.date)
        assert self.observer.failures and 'Multiple' in self.observer.failures[0].message

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_fault_with_different_node_id(self, mock_list_nodes):
        """There was a down bond 1g method call, but with a different node."""
        node2 = models.ActiveNode.create(node_id=self.node.node_id+1)
        node2.static_id = self.node.static_id + '1'
        mock_list_nodes.return_value = [self.node, node2]
        self.fault.node_id = node2.node_id
        method_call_map = self.monitor._get_method_call_map([self.down_bond_1g])
        self.observer.update(
            method_call_map=method_call_map,
            fault_map=self.fault_map,
            event_map={},
            timestamp=self.down_bond_1g.timestamp + datetime.timedelta(
                seconds=self.observer.max_timeout+1))
        assert self.observer.failures

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_expected_fault(self, mock_list_nodes):
        mock_list_nodes.return_value = [self.node]
        method_call_map = self.monitor._get_method_call_map([self.down_bond_1g])
        self.observer.update(
            method_call_map=method_call_map,
            fault_map=self.fault_map,
            event_map={}
        )
        assert not self.observer.failures
