"""UnresponsiveService observer class.

This class observes and handles all the UnresponsiveService cluster faults, that are encountered by
the Backend Monitor
"""
# pylint: skip-file

import re
from collections import defaultdict

from sf_api import enums
from valence import functional_models
from valence.monitors import monitor_query
from valence.monitors.observers import (
    base_observer,
    errors,
)

# Key constants for a fault record within the unresponsiveService Observer
METHOD = 'method'
NODE_NAME = 'node_name'
FAULTS = 'faults'
EXPECTED_NUM_FAULTS = 'expected_num_faults'
TIMEOUT_CALLBACK_ID = 'timeout_callback_id'
FAULT_STATUS = 'fault_status'
RESOLVING = 'resolving'
RAISING = 'raising'

# Types of services running on a node
MASTER = 'master'
BULK_VOLUME = 'bulk volume'
METADATA = 'metadata'
FIBER = 'fiber'
BLOCK = 'block'


class FaultRecorder(object):
    """This is a helper class to maintain a list of faults that have occured or are expeected."""

    def __init__(self):
        self.fault_records = []

    # TODO: make use of args in fault number calculations. Check for FC node and num drives
    @classmethod
    def create_expected_faults_dict(cls, method_name, node_dm):  # pylint: disable=unused-argument
        """This function creates a map for the number of faults to expect per service.

        The map is created per method, node combination. It is used to verify if all the expected
        number of faults were raised or resolved.

        :param str method_name: The name of the method within NodeFM that was called
        :param valence.models.ResourceNode node_dm: The underlying node data model of the Node FM
        :rtype: dict
        :returns: A unique id associated with this callback dict {<service>: <expected number of faults>}
                    e.g. {'master': 1, 'fiber': 0 ...}
        """
        expected_faults_dict = {MASTER: 1,
                                BULK_VOLUME: 1,
                                FIBER: 0,
                                METADATA: 1,
                                BLOCK: 10}
        return expected_faults_dict

    def start_recording_new_faults(self, method_name, node_dm, timeout_callback_id):
        """This function will create a new fault record dictionary.

        This function is called when the observer is notified for the first time by the monitor of a
        method call within NodeFM. This method call is expected to raise cluster faults.

        :param str method_name: The name of the method within NodeFM that was called
        :param valence.models.ResourceNode node_dm: The underlying node data model of the Node FM
        :param int timeout_callback_id: The timeout callback id returned by the backend monitor
        """
        node_name = node_dm.node_name
        expected_faults = self.create_expected_faults_dict(method_name, node_dm)
        fault_record = {
            METHOD: method_name,
            NODE_NAME: node_name,
            FAULTS: defaultdict(list),
            EXPECTED_NUM_FAULTS: expected_faults,
            TIMEOUT_CALLBACK_ID: timeout_callback_id,
            FAULT_STATUS: RAISING
        }
        self.fault_records.append(fault_record)

    def add_new_fault(self, fault, node_name):
        """Adds fault into the fault_records list for a given node and method call.

        :param valence.models.ClusterFault fault: The fault to be added
        :param str node_name: The name of the node for which the fault occurred
        :raises: valence.monitors.observers.error.ObserverError
        """
        fault_record = [r for r in self.fault_records if r[NODE_NAME] == node_name][0]
        fault_service = re.match(r'A (([a-z]+ ){1,2})service is not responding.',
                                 fault.details).group(1).rstrip()
        if not fault_record[FAULT_STATUS] == RAISING:
            raise errors.ObserverError('New cluster fault observed when the already observed faults '
                                       'should be resolving')
        fault_record[FAULTS][fault_service].append(fault)

    def resolve_fault(self, fault, node_name):
        """This is a function to update a raised fault with a resolved fault within fault_records.

        :param valence.models.ClusterFault fault: The resolved fault object
        :param str node_name: The name of the node for which the fault was resolved
        :raises: valence.monitors.observers.error.ObserverError
        """
        fault_service = re.match(r'A (([a-z]+ ){1,2})service is not responding.',
                                 fault.details).group(1).rstrip()
        fault_id = fault.id
        fault_record = [r for r in self.fault_records if r[NODE_NAME] == node_name][0]
        fault_to_resolve = [f for f in fault_record[FAULTS][fault_service] if f.id ==
                            fault_id][0]
        if not fault_record[FAULT_STATUS] == RESOLVING:
            raise errors.ObserverError('A cluster fault was resolved when the new faults should be '
                                       'raised')
        fault_record[FAULTS][fault_service].remove(fault_to_resolve)
        fault_record[FAULTS][fault_service].append(fault)

    def assert_all_faults_raised_or_resolved(self, timeout_callback_id):
        """This functions asserts if all the expected faults were raised or resolved.

        A timeout_callback_id is associated with a group of cluster faults that would have been raised
        as a side effect of a Node FM method call. Use the timeout_callback_id to determine if all the
        faults related to the that id were resolved or raised when the timeout occurred

        :param int timeout_callback_id: Unique id to identify the timeout callback request with. It is
                                        the id that is identifies the timeout occured for which group of
                                        faults
        """
        fault_record = [r for r in self.fault_records if r[TIMEOUT_CALLBACK_ID] ==
                        timeout_callback_id][0]
        resolved = fault_record[FAULT_STATUS] == RAISING
        for k, v in fault_record[EXPECTED_NUM_FAULTS].items():
            faults = [f for f in fault_record[FAULTS][k] if f.resolved == resolved]
            if not len(faults) == v:
                if not resolved:
                    raise errors.FaultsNotRaisedError(
                        'Expected to raise {} faults for {} service, only observed {}: {}'
                        .format(v, k, len(faults), faults))
                else:
                    raise errors.FaultsNotRaisedError(
                        'Expected to resolve {} faults for {} service, only observed {}:{}'
                        .format(v, k, len(faults), faults))

    def start_resolving_old_faults(self, node_name, method_name, state, timeout_callback_id):
        """This function is called called to change the state of a fault record.

        This function is called when the observer is notified by the monitor of a method call within
        NodeFM. This method call is expected to resolve all the raise cluster faults.

        :param str node_name: The name of the node for recorder state needs to be altered in
                              fault_records
        :param str method_name: The name of the method call for recorder state needs to be altered in
                                fault_records
        :param str state: The state in which the fault_record should be. raising | resolving
        :param int timeout_callback_id: The timeout callback id returned by the backend monitor
        """
        for fault_record in self.fault_records:
            if fault_record[NODE_NAME] == node_name and fault_record[METHOD] == \
                    method_name:
                fault_record[FAULT_STATUS] = state
                fault_record[TIMEOUT_CALLBACK_ID] = timeout_callback_id


class UnresponsiveServiceObserver(base_observer.BaseObserver):
    """Fault Observer for the UnresponsiveService fault.

    UnresponsiveService Fault is seen when when any or all of Master, Bulk Volume, Fibre, Block or
    Metadata service on a node are not responsive.

    :param valence.models.SuiteResources suite_resources: Suite resources data model
    :param valence.monitors.BackendMonitor monitor: Instance of the backend monitor
    """

    def __init__(self, suite_resources, monitor):
        super(UnresponsiveServiceObserver, self).__init__(suite_resources=suite_resources,
                                                          monitor=monitor)
        self.fault_recorder = FaultRecorder()
        self._cluster_fm = functional_models.Cluster(self.suite_resources.cluster_dms[0],
                                                     authenticate=False)

    def _get_monitor_query(self):
        return monitor_query.MonitorQuery(
            methods_to_store=[
                functional_models.Node.start_binary,
                functional_models.Node.stop_binary,
            ],
            faults_to_store=[enums.ClusterFaultCode.unresponsive_service],
            events_to_store=[]
        )

    def _fault_callback(self, fault):
        """Fault callback method for the Unresponsive Service Observer.

        :param valence.models.ClusterFault fault: The fault object
        :raises valence.monitors.MonitorError: If the fault is not valid or expected.
        """
        nodes = self._cluster_fm.list_active_nodes()  # pylint: disable=protected-access
        node = [n for n in nodes if n.node_id == fault.node_id][0]
        node_name = node.name
        if not fault.resolved:
            self.fault_recorder.add_new_fault(fault=fault, node_name=node_name)
        else:
            self.fault_recorder.resolve_fault(fault=fault, node_name=node_name)

    def _timeout_callback(self, callback_id):
        """Timeout callback method for the Unresponsive Service Observer.

        :param int callback_id: A unique id for the callback that was returned during the initial
            register_timeout_callback call.
        """
        self.fault_recorder.assert_all_faults_raised_or_resolved(timeout_callback_id=callback_id)

    def _notify(self, method_call):
        """Notify method which will be called when a Functional method is called.

        :param valence.functional_models.MethodCall method_call: Object that holds info about the fm call
        """
        method_name = method_call.method.__name__
        node_dm = method_call.arguments['self'].node_dm
        node_name = node_dm.node_name
        if (method_call.is_method(functional_models.Node.stop_binary) and
                method_call.arguments['service'] == enums.NodeBinary.solidfire):
            timeout_callback_id = self.register_timeout_callback(seconds_to_wait=180)
            self.fault_recorder.start_recording_new_faults(method_name=method_name,
                                                           node_dm=node_dm,
                                                           timeout_callback_id=timeout_callback_id)
        if (method_call.is_method(functional_models.Node.start_binary) and
                method_call.arguments['service'] == enums.NodeBinary.solidfire):
            timeout_callback_id = self.register_timeout_callback(seconds_to_wait=360)
            self.fault_recorder.start_resolving_old_faults(node_name=node_name,
                                                           method_name=method_name,
                                                           state=RESOLVING,
                                                           timeout_callback_id=timeout_callback_id)
