import datetime
import operator
import unittest

import logbook
import mock
import pytest

from camelot import driver
from valence import (
    enums,
    functional_models,
    models,
    monitors,
)
from valence.monitors import (
    _backend_monitor,
    _constants,
)
from valence.utils import (
    criteria,
    time_utils,
)


class MockObserver(monitors.BaseObserver):
    """Mock observer that can optionally store failures."""

    def __init__(self, suite_resources, monitor):
        super(MockObserver, self).__init__(suite_resources, monitor)
        self.raise_failure = False
        self._failures = []

    def _check_for_failures(self):
        if self.raise_failure:
            self._store_failure('You broke stuff', time_utils.get_now_utc())

    def _notify(self, method_call):
        """Defined for mocking purposes."""

    def _get_monitor_query(self):
        return monitors.MonitorQuery(
            methods_to_store=[],
            faults_to_store=[enums.ClusterFaultCode.cluster_full],
            events_to_store=[]
        )

    @property
    def claimed_fault(self):
        return enums.ClusterFaultCode.cluster_full


class TestBackendMonitor(unittest.TestCase):
    """Test valence.monitors.BackendMonitor class."""

    @classmethod
    def setUpClass(cls):
        cls.suite_resources = driver.get_mock_resources_dm()

    def setUp(self):
        self.new_fault = models.ClusterFault.create(
            resolved=False,
            code=enums.ClusterFaultCode.cluster_full,
        )
        self.resolved_fault = models.ClusterFault.create(
            resolved=True,
            code=enums.ClusterFaultCode.cluster_full,
            cluster_fault_id=self.new_fault.cluster_fault_id,
        )
        null_handler = logbook.NullHandler()
        self.monitor = monitors.BackendMonitor(self.suite_resources, null_handler, poll_period=0)
        self.monitor._fault_whitelist = set()
        # NOTE: if you want to mock a method on the observer, you must call _set_observers again.
        # Otherwise, the mocking will not work since the _fault_callback_map will have a stale reference.
        self.monitor._set_observers([MockObserver])
        method = functional_models.Node.get_mock_fm().stop_binary
        args = {'service': enums.NodeBinary.solidfire}
        self.method_call = functional_models.MethodCall(method, args, datetime.datetime.now())

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_raise_failures_time_filters(self, mock_list_faults, mock_list_events):
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor._observers[0].raise_failure = True
        self.monitor.run_iteration()
        failure_time = self.monitor._observers[0].failures[0].timestamp
        with pytest.raises(monitors.MonitorError):
            self.monitor.get_failures(raise_failures=True)
        self.monitor.get_failures(start_time=failure_time + datetime.timedelta(seconds=10),
                                  raise_failures=True)
        self.monitor.get_failures(end_time=failure_time - datetime.timedelta(seconds=10),
                                  raise_failures=True)

    def test_deregister_timeout(self):
        timeout_id = self.monitor.register_timeout(1)
        assert self.monitor._timeout_map != {}
        self.monitor.deregister_timeout(timeout_id)
        assert self.monitor._timeout_map == {}

    def test_get_and_set_always_fail_events(self):
        original_events = self.monitor.get_always_fail_events()
        self.monitor.set_always_fail_events([enums.ClusterEventType.bin_assignments_event])
        new_events = self.monitor.get_always_fail_events()
        assert original_events != new_events

    def test_get_and_set_always_fail_faults(self):
        original_faults = self.monitor.get_always_fail_faults()
        self.monitor.set_always_fail_faults([enums.ClusterFaultCode.block_cluster_full])
        new_faults = self.monitor.get_always_fail_faults()
        assert original_faults != new_faults

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_claimed_faults_do_not_cause_failures(self, mock_list_faults, mock_list_events):
        """If a fault is claimed, it should not be counted as an unexpected fault."""
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor.run_iteration()
        assert not self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_unexpected_fault_with_whitelist(self, mock_list_faults, mock_list_events):
        """An unclaimed fault that is not whitelisted occurs."""
        self.new_fault.code = enums.ClusterFaultCode.disconnected_cluster_pair
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor._fault_whitelist = {enums.ClusterFaultCode.block_cluster_full}
        self.monitor.run_iteration()
        assert self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_always_fail_faults(self, mock_list_faults, mock_list_events):
        """An unclaimed fault that is always fail list occurs."""
        # Need to set it to an unclaimed fault
        self.new_fault.code = enums.ClusterFaultCode.disconnected_cluster_pair
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor.set_always_fail_faults([self.new_fault.code])
        self.monitor.run_iteration()
        assert self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_empty_whitelist_does_not_cause_failures(self, mock_list_faults, mock_list_events):
        """An unclaimed fault that is not whitelisted occurs."""
        mock_list_faults.return_value = []
        mock_list_events.return_value = []
        self.monitor._fault_whitelist = set([])
        self.monitor.run_iteration()
        assert not self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_unexpected_event(self, mock_list_faults, mock_list_events):
        """A blacklisted event occurs."""
        event_type = enums.ClusterEventType.bin_assignments_event
        self.monitor.set_always_fail_events([event_type])
        mock_list_faults.return_value = []
        mock_list_events.return_value = [models.ClusterEvent.create(event_info_type=event_type)]
        self.monitor.run_iteration()
        assert self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_blacklisted_event_details_match(self, mock_list_faults, mock_list_events):
        """A blacklisted event occurs where the type and details match."""
        event_type = enums.ClusterEventType.data_event
        exc_code = 'xBadExceptionCode'
        exc_code_key = 'exceptionCode'
        event_filter = criteria.And(
            criteria.Equals(models.ClusterEvent.event_info_type, event_type),
            criteria.Equals(models.ClusterEvent.details, exc_code,
                            getter=operator.itemgetter(exc_code_key))
        )
        self.monitor.set_always_fail_events([event_filter])
        mock_list_faults.return_value = []
        mock_list_events.return_value = [
            models.ClusterEvent.create(
                event_info_type=event_type,
                details={exc_code_key: exc_code})
        ]
        self.monitor.run_iteration()
        assert self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_blacklisted_event_details_do_not_match(self, mock_list_faults, mock_list_events):
        """A blacklisted event occurs, but the details do not match."""
        event_type = enums.ClusterEventType.data_event
        exc_code = 'xPhysicalPageChecksumFailed'
        exc_code_key = 'exceptionCode'
        event_filter = criteria.And(
            criteria.Equals(models.ClusterEvent.event_info_type, event_type),
            criteria.Equals(models.ClusterEvent.details, exc_code,
                            getter=operator.itemgetter(exc_code_key))
        )
        self.monitor.set_always_fail_events([event_filter])
        mock_list_faults.return_value = []
        mock_list_events.return_value = [
            models.ClusterEvent.create(
                event_info_type=event_type,
                details={exc_code_key: 'xNotTheBadException'})
        ]
        self.monitor.run_iteration()
        assert not self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_blacklisted_event_with_criteria_and_enum(self, mock_list_faults, mock_list_events):
        """The same event is blacklisted with just the enum and a criteria object."""
        event_type = enums.ClusterEventType.data_event
        exc_code = 'xPhysicalPageChecksumFailed'
        exc_code_key = 'exceptionCode'
        event_filter = criteria.And(
            criteria.Equals(models.ClusterEvent.event_info_type, event_type),
            criteria.Equals(models.ClusterEvent.details, exc_code,
                            getter=operator.itemgetter(exc_code_key))
        )
        # Add the complex filter and just the simple event type. The event type should pick up the
        # failure even though the details on the complex filter does not match.
        self.monitor.set_always_fail_events([event_filter, event_type])
        mock_list_faults.return_value = []
        mock_list_events.return_value = [
            models.ClusterEvent.create(
                event_info_type=event_type,
                details={exc_code_key: 'xNotTheBadException'})
        ]
        self.monitor.run_iteration()
        assert self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_critical_error_event(self, mock_list_faults, mock_list_events):
        """A critical error should raise an exception immediately."""
        event_type = enums.ClusterEventType.ie_event
        exc_code = 'xPhysicalPageChecksumFailed'
        exc_code_key = 'exceptionCode'
        mock_list_faults.return_value = []
        mock_list_events.return_value = [
            models.ClusterEvent.create(
                event_info_type=event_type,
                details={exc_code_key: exc_code})
        ]
        with pytest.raises(monitors.MonitorCriticalError):
            self.monitor.run_iteration()

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_remove_from_always_fail_events(self, mock_list_faults, mock_list_events):
        """If you remove an event from the blacklist, it will not trigger a failure (ELEM-2253)."""
        data_event = enums.ClusterEventType.data_event
        mock_list_faults.return_value = []
        mock_list_events.return_value = [models.ClusterEvent.create(event_info_type=data_event,
                                                                    event_id=1)]

        # First, make sure there is a failure when we see a data event
        self.monitor.run_iteration()
        assert len(self.monitor.failures) == 1

        # Then remove the data event and make sure no new failures happen
        events = self.monitor.get_always_fail_events()
        assert data_event in events
        events.remove(data_event)
        self.monitor.set_always_fail_events(events)
        self.monitor.run_iteration()
        assert len(self.monitor.failures) == 1

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_unexpected_fault_resolved(self, mock_list_faults, mock_list_events):
        self.new_fault.code = enums.ClusterFaultCode.disconnected_cluster_pair
        self.new_fault.resolved = True
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor.run_iteration()
        assert not self.monitor.failures

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    @mock.patch.object(monitors.BackendMonitor, '_update_observers')
    def test_monitor_runs_at_least_once(self, mock_update_observers, mock_list_faults, mock_list_events):
        """Stop monitoring immediately and make sure the monitor loop still runs."""
        mock_list_events.return_value = []
        self.monitor.start_monitoring()
        assert self.monitor.is_running()
        self.monitor.stop_monitoring()
        assert not self.monitor.is_running()
        mock_update_observers.assert_called()

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    @mock.patch.object(monitors.BackendMonitor, '_update_observers')
    def test_monitor_start_twice(self, mock_update_observers, mock_list_faults, mock_list_events):
        mock_list_events.return_value = []
        self.monitor.start_monitoring()
        with pytest.raises(monitors.MonitorError):
            self.monitor.start_monitoring()
        self.monitor.stop_monitoring()

    def test_monitor_fault_becomes_resolved(self):
        """Test that we only handle the fault when it first occurs and when it's resolved."""
        fault_map = self.monitor._get_fault_map([self.new_fault])
        assert fault_map
        fault_map = self.monitor._get_fault_map([self.new_fault])
        assert not fault_map
        fault_map = self.monitor._get_fault_map([self.resolved_fault])
        assert fault_map

    def test_monitor_fault_ignored_after_resolved(self):
        self.monitor._get_fault_map([self.new_fault])
        fault_map = self.monitor._get_fault_map([self.resolved_fault])
        assert fault_map
        fault_map = self.monitor._get_fault_map([self.resolved_fault])
        assert not fault_map

    def test_monitor_fault_resolved_before_first_seen(self):
        fault_map = self.monitor._get_fault_map([self.resolved_fault])
        assert fault_map
        fault_map = self.monitor._get_fault_map([self.resolved_fault])
        assert not fault_map

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_monitor_fail_raise_failures(self, mock_list_faults, mock_list_events):
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor._observers[0].raise_failure = True
        with pytest.raises(monitors.MonitorError):
            self.monitor.start_monitoring()
            self.monitor.stop_monitoring(raise_failures=True)

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_monitor_unregistered_fault(self, mock_list_faults, mock_list_events):
        self.new_fault.code = enums.ClusterFaultCode.unbalanced_mixed_nodes
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor.start_monitoring()
        self.monitor.stop_monitoring(raise_failures=False)
        observer = self.monitor._observers[0]
        assert not observer.query.get_faults(self.new_fault.code)

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_fault_whitelist(self, mock_list_faults, mock_list_events):
        """When a fault is whitelisted, it should not cause a failure."""
        self.monitor._set_observers([MockObserver])
        self.new_fault.code = enums.ClusterFaultCode.unbalanced_mixed_nodes
        mock_list_faults.return_value = [self.new_fault]
        mock_list_events.return_value = []
        self.monitor._fault_whitelist = [enums.ClusterFaultCode.unbalanced_mixed_nodes]
        self.monitor.run_iteration()
        assert not self.monitor.failures

    def test_notify(self):
        self.monitor.notify(self.method_call)
        result = self.monitor._get_method_calls()[0]
        assert result is self.method_call

    def test_get_method_call_map(self):
        calls = [self.method_call]
        result = self.monitor._get_method_call_map(calls)
        assert self.method_call in result[self.method_call.method]

    @mock.patch('time.time')
    def test_timeout_callback_is_ready(self, mock_time):
        mock_time.side_effect = [0, 1, 3]
        timeout_callback = _backend_monitor.ObserverTimeout(1, 99)
        assert not timeout_callback.is_ready()
        assert timeout_callback.is_ready()

    @mock.patch.object(functional_models.Cluster, 'list_events')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    @mock.patch.object(_backend_monitor.ObserverTimeout, 'is_ready')
    # it is necessary to use thread method otherwise on timeout the monitor thread will keep running and
    # try to send commands using the cluster fm in the monitor's run_iteration method
    @pytest.mark.timeout(1, method='thread')
    @mock.patch('time.sleep', mock.Mock())
    def test_wait_for_timeout_callbacks(self, mock_is_ready, mock_list_faults,
                                        mock_list_events):
        mock_list_faults.return_value = []
        mock_list_events.return_value = []
        mock_is_ready.side_effect = [False, True, True]
        self.monitor.poll_period = 0
        self.monitor.start_monitoring()
        self.monitor.register_timeout(0)
        self.monitor.wait_for_timeouts()
        self.monitor.stop_monitoring()
        # Depending on timing, is_ready could be called as part of the monitor loop. Because of that,
        # the call could should be at least 2
        assert mock_is_ready.call_count >= 2

    def test_event_map(self):
        drive_event = enums.ClusterEventType.drive_event
        event = models.ClusterEvent.get_mock_object(
            overrides={'event_info_type': drive_event})
        event_map = self.monitor._get_event_map([event])
        assert event_map == {drive_event: [event]}
