"""Test the MonitorQuery class."""

import datetime
import unittest

from valence import (
    enums,
    functional_models,
    models,
    monitors,
    utils,
)

remove_drives = functional_models.Cluster.remove_drives


class TestMonitorQuery(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_full = enums.ClusterFaultCode.cluster_full
        cls.drive_failed = enums.ClusterFaultCode.drive_failed
        cls.drive_event = enums.ClusterEventType.drive_event
        cls.fault_map = {
            cls.cluster_full: [
                models.ClusterFault.get_mock_object(overrides={'code': cls.cluster_full.value})
            ],
            cls.drive_failed: [
                models.ClusterFault.get_mock_object(overrides={'code': cls.drive_failed.value})
            ]
        }
        cls.event_map = {
            cls.drive_event: [
                models.ClusterEvent.get_mock_object(overrides={'event_info_type': cls.drive_event})
            ]
        }
        cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.method_map = {
            remove_drives: [
                functional_models.MethodCall(cluster_fm.remove_drives, {}, datetime.datetime.now())
            ]
        }

    def test_get_faults(self):
        monitor_query = monitors.MonitorQuery(methods_to_store=[], faults_to_store=[self.cluster_full],
                                              events_to_store=[])
        monitor_query.update(method_call_map={}, fault_map=self.fault_map, event_map={})
        faults = monitor_query.get_faults(self.cluster_full)
        assert len(faults) == 1
        assert faults[0].code is self.cluster_full

    def test_get_faults_date_range(self):
        now = utils.get_now_utc()
        faults = [
            models.ClusterFault.create(date=now, code=self.cluster_full),
            models.ClusterFault.create(date=now+datetime.timedelta(seconds=10),
                                       code=self.cluster_full),
            models.ClusterFault.create(date=now+datetime.timedelta(seconds=20),
                                       code=self.cluster_full),
        ]
        fault_map = {
            self.cluster_full: faults
        }
        monitor_query = monitors.MonitorQuery(methods_to_store=[], faults_to_store=[self.cluster_full],
                                              events_to_store=[])
        monitor_query.update(method_call_map={}, fault_map=fault_map, event_map={})

        date_range = (now, now + datetime.timedelta(seconds=15))
        result = monitor_query.get_faults(self.cluster_full, date_range=date_range)
        assert result == faults[0:2]

        date_range = (now + datetime.timedelta(seconds=30), now + datetime.timedelta(seconds=40))
        result = monitor_query.get_faults(self.cluster_full, date_range=date_range)
        assert len(result) == 0

        date_range = (now + datetime.timedelta(seconds=10), now + datetime.timedelta(seconds=40))
        result = monitor_query.get_faults(self.cluster_full, date_range=date_range)
        assert result == faults[1:]

    def test_get_faults_none_found(self):
        monitor_query = monitors.MonitorQuery(methods_to_store=[], faults_to_store=[self.cluster_full],
                                              events_to_store=[])
        monitor_query.update(method_call_map={}, fault_map=self.fault_map, event_map={})
        faults = monitor_query.get_faults(enums.ClusterFaultCode.disconnected_cluster_pair)
        assert not faults

    def test_update_no_faults_stored(self):
        monitor_query = monitors.MonitorQuery(methods_to_store=[], faults_to_store=[self.drive_failed],
                                              events_to_store=[])
        monitor_query.update(method_call_map={}, fault_map=self.fault_map, event_map={})
        faults = monitor_query.get_faults(self.cluster_full)
        assert not faults

    def test_get_events(self):
        monitor_query = monitors.MonitorQuery(methods_to_store=[], faults_to_store=[],
                                              events_to_store=[self.drive_event])
        monitor_query.update(method_call_map={}, fault_map={}, event_map=self.event_map)
        events = monitor_query.get_events(event_type=self.drive_event)
        assert len(events) == 1
        assert events[0].event_info_type == self.drive_event

    def test_get_events_date_range(self):
        now = utils.get_now_utc()
        events = [
            models.ClusterEvent.create(time_of_report=now, event_info_type=self.drive_event),
            models.ClusterEvent.create(time_of_report=now+datetime.timedelta(seconds=10),
                                       event_info_type=self.drive_event),
            models.ClusterEvent.create(time_of_report=now+datetime.timedelta(seconds=20),
                                       event_info_type=self.drive_event)
        ]
        event_map = {self.drive_event: events}
        monitor_query = monitors.MonitorQuery(methods_to_store=[], faults_to_store=[],
                                              events_to_store=[self.drive_event])
        monitor_query.update(method_call_map={}, fault_map={}, event_map=event_map)
        time_range = (now, now + datetime.timedelta(seconds=15))
        result = monitor_query.get_events(self.drive_event, time_range=time_range)
        assert result == events[0:2]

        time_range = (now + datetime.timedelta(seconds=30), now + datetime.timedelta(seconds=40))
        result = monitor_query.get_events(self.drive_event, time_range=time_range)
        assert len(result) == 0

        time_range = (now + datetime.timedelta(seconds=10), now + datetime.timedelta(seconds=40))
        result = monitor_query.get_events(self.drive_event, time_range=time_range)
        assert result == events[1:]

    def test_get_method_calls(self):
        monitor_query = monitors.MonitorQuery(methods_to_store=[remove_drives], faults_to_store=[],
                                              events_to_store=[])
        monitor_query.update(method_call_map=self.method_map, fault_map={}, event_map={})
        calls = monitor_query.get_method_calls(remove_drives)
        assert len(calls) == 1
        assert calls[0].method == remove_drives

    def test_get_method_calls_none_found(self):
        monitor_query = monitors.MonitorQuery(methods_to_store=[remove_drives], faults_to_store=[],
                                              events_to_store=[])
        monitor_query.update(method_call_map=self.method_map, fault_map={}, event_map={})
        calls = monitor_query.get_method_calls(functional_models.Cluster.create_account)
        assert not calls
