"""Helper module for constructing qTest models."""
# pragma: no cover

import datetime

import autologger
from qtestapi import (
    constants,
    models,
)

from valence.scribe import _qt_constants

_log = autologger.get_sflogging(__name__)


def qtest_case_builder(api, test_case_doc, test_case_dm=None):
    """Builds new qTest.TestCase object if one is not supplied.  Otherwise updates supplied model.

    :param qtestapi.QTestAPI api: API object, required for field/property mapping.
    :param valence.scribe.TestCaseDocumentation test_case_doc: Scribe model to update from
    :param qtestapi.models.TestCase test_case_dm: Existing test case model.  Provided for updating.
    :rtype: qtestapi.models.TestCase
    """
    # TODO : remove, shouldn't be necessary if we prevalidate docstrings while building docs
    if test_case_doc.created_on is None:  # pragma: no cover
        test_case_doc.created_on = datetime.datetime.now()

    tc_name = test_case_doc.name
    tc_steps = [models.TestStep(description=step.action, expected=step.expected_result)
                for step in test_case_doc.steps]

    # If model is passed in, use it and update where necessary
    if test_case_dm:
        for field_name, tc_attr in _qt_constants.QTEST_FIELD_MAP.items():
            try:
                test_case_dm.properties[field_name].update(getattr(test_case_doc, tc_attr))
            except KeyError:
                field = api.get_ns_field_by_label(constants.TEST_CASES, field_name)
                test_case_dm.properties.append(field.to_property(getattr(test_case_doc, tc_attr)))
        test_case_dm.name = tc_name
        test_case_dm.test_steps = tc_steps

    else:
        # Build the properties
        properties = []

        status_field = api.get_ns_field_by_label(constants.TEST_CASES, _qt_constants.STATUS)
        properties.append(status_field.to_property(_qt_constants.QTestCaseStatus.new.value))

        for field_name, tc_attr in _qt_constants.QTEST_FIELD_MAP.items():
            field = api.get_ns_field_by_label(constants.TEST_CASES, field_name)
            properties.append(field.to_property(getattr(test_case_doc, tc_attr)))

        test_case_dm = models.TestCase(name=tc_name, test_steps=tc_steps, properties=properties)

    return test_case_dm


def check_case_requirements(api, test_case_doc, test_case_dm):
    """Attempts to link the requirements specified on the test_case_doc to qTest/JIRA Reequirement.

    :param qtestapi.QTestAPI api: API object, required for field/property mapping.
    :param valence.scribe.TestCaseDocumentation test_case_doc: Scribe model to update from
    :param qtestapi.models.TestCase test_case_dm: Existing test case model.  Provided for updating.
    """

    if test_case_doc.requirements is not None:
        # Parse the docstring line containing requirements
        tc_requirements = [_.strip() for _ in test_case_doc.requirements.split(',')]

        # Get a list of all existing qTest requirements.  These are imported manually from JIRA
        qtest_requirements = api.get_all_requirements()

        # Get a list of already linked requirements for this test_case
        tc_linked_requirements = api.get_test_case_linked_requirements(test_case_dm.id)
        linked_requirement_ids = [req.id for req in tc_linked_requirements]

        for tc_req in tc_requirements:
            qt_req = next((x for x in qtest_requirements if tc_req in x.name), None)

            if qt_req is None:
                _log.debug('Docstring requirement <{}> not found in qTest.  Import may be needed.'
                           .format(tc_req))
            elif qt_req.id in linked_requirement_ids:
                _log.debug('Docstring requirement <{}> already linked in qTest.'.format(tc_req))
                linked_requirement_ids.remove(qt_req.id)
            elif qt_req.id not in linked_requirement_ids:
                _log.debug('Linking qTest requirement.')
                api.link_test_case_to_requirement(test_case_dm.id, qt_req.id)

        # Removed any linked requirements from this list.  What is left is stale
        for old_req_id in linked_requirement_ids:
            _log.debug('Removing unused requirement id {}'.format(old_req_id))
            api.remove_test_case_link_to_requirement(test_case_dm.id, old_req_id)
