"""This module builds test-case and test-suite documentation objects."""
import copy
import importlib
import inspect
import re

import sf_artifacts
import six  # pylint: disable=unused-import

import camelot
from camelot import (
    driver,
    inspection,
)
from valence import models
from valence.scribe import (
    _errors,
    _qt_constants,
)

AUTHOR_RE = re.compile(r':tc_info\s+Author:\s(.+)')
CREATED_RE = re.compile(r':tc_info\sCreatedOn:\s(.+)')
FEATURE_RE = re.compile(r':tc_info\sFeature:\s(.+)')
REQ_RE = re.compile(r':tc_info\sRequirements:\s(.+)')
DESCRIPTION_RE = re.compile(r':tc_info.+\n|:param.+\n|:var.+\n')


class TestCaseDocumentation(object):
    """Class for storing scraped information from testcase documentation for a camelot testcase.

    :param str docstring: Class level docstring
    :param six.class_types case_class: Class of testcase
    :param str test_path: Module name test class is from
    :param str author: Author of class
    :param str created_on: Date string of creation - no format defined
    :param str feature: Solidfire feature the test relates to
    :param str name: Name of the test case
    :param str preconditions: Source code for test case _preconditions
    :param str setup: Source code for test case _setup
    :param str teardown: Source code for test case _teardown
    :param str post_test_assertions: Docstring for post test assertions method.
    :param str requirements: List of requirements for test
    :param list[camelot.inspection.StepInfo] steps: List of steps found from logstep
    :param str description: Description from docstring, everything except params, vars, tc_info
    :param str source_hash: md5 checksum of source code for test case and parent classes
    :param list[camelot.inspection.CaseInspectorProblem] test_case_errors: Scribe errors of the
        test case
    """

    def __init__(self, docstring, case_class, test_path, author, created_on, feature, name,
                 preconditions, setup, teardown, post_test_assertions, requirements, steps, description,
                 source_hash, test_case_errors=None):

        self.docstring = docstring
        self.case_class = case_class
        self.test_path = '.'.join([test_path, case_class.__name__])
        self.author = author
        self.created_on = created_on
        self.feature = feature
        self.name = name
        self.preconditions = preconditions
        self.setup = setup
        self.teardown = teardown
        self.post_test_assertions = post_test_assertions
        self.requirements = requirements
        self.source_hash = source_hash
        self.steps = steps
        self.description = description
        self.test_case_errors = test_case_errors or []

    @classmethod
    def from_test_case(cls, test_case):  # pylint: disable=too-complex
        """Class method for creating a TestCaseDocumentation object from a test class.

        :param six.class_types test_case: Test case to inspect
        :rtype: valence.scribe.TestCaseDocumentation
        """
        inspector = inspection.CaseInspector(test_case)
        docstring = test_case.__doc__

        if not docstring:
            raise _errors.NoDocstringError('Test case class has no docstring')

        case_class = inspector.test_case_class
        test_path = case_class.__module__
        author = cls.find_pattern_in_docstring(AUTHOR_RE, docstring)
        created_on = cls.find_pattern_in_docstring(CREATED_RE, docstring)
        feature = cls.find_pattern_in_docstring(FEATURE_RE, docstring)
        name = case_class.__name__ or 'N/A'

        preconditions = None
        setup = None
        teardown = None
        post_test_assertions = None
        # We don't want inherited code from the base class.
        for attr, value in vars(test_case).items():
            if attr == '_preconditions' and inspect.isfunction(value):
                # pylint: disable=protected-access
                preconditions = test_case._preconditions.__doc__
            if attr == '_setup' and inspect.isfunction(value):
                # pylint: disable=protected-access
                setup = test_case._setup.__doc__
            if attr == '_teardown' and inspect.isfunction(value):
                # pylint: disable=protected-access
                teardown = test_case._teardown.__doc__
            if attr == '_post_test_assertions' and inspect.isfunction(value):
                # pylint: disable=protected-access
                post_test_assertions = test_case._post_test_assertions.__doc__

        requirements = cls.find_pattern_in_docstring(REQ_RE, docstring)
        source_hash = inspector.get_source_hash()

        steps = None
        test_case_errors = []
        try:
            steps = inspector.get_steps()
        except camelot.InspectionError:
            test_case_errors = inspector.test_case_errors

        # Filling in defaults for a step object in case get_steps does not find anything.
        if not steps:
            steps = [inspection.StepInfo(index=0, action='No steps defined',
                                         expected_result='No results defined',
                                         source_lines=[])]
        description = ' '.join(DESCRIPTION_RE.sub('', docstring).split()) or 'No description defined'

        return cls(docstring, case_class, test_path, author, created_on, feature, name,
                   preconditions, setup, teardown, post_test_assertions, requirements, steps,
                   description, source_hash, test_case_errors)

    @staticmethod
    def find_pattern_in_docstring(regex, docstring):
        """Private helper to search self.docstring with provided regular expression.

        :param Any regex: Regular expression do find in docstring
        :param str docstring: Docstring to be searched
        :rtype: str
        """
        match = regex.search(docstring)
        if match:
            return match.group(1).strip()
        else:
            return None

    def get_qtest_errors(self):
        """Checks for qTest required fields on object.  Returns any errors found.

        :rtype: list[str]
        """
        qtest_errors = []
        for attr in _qt_constants.REQD_DOCSTRING_FIELDS:
            if getattr(self, attr) is None:
                qtest_errors.append('{} missing field `{}`'.format(self.name, attr))
        return qtest_errors

    def to_json(self):
        """Returns JSON/dict matching GROOT schema requirements.

        :rtype: dict
        """
        result = copy.deepcopy(self.__dict__)
        result['case_class'] = repr(result['case_class'])

        for i, _ in enumerate(result['steps']):
            result['steps'][i] = result['steps'][i].__dict__

        return result


class TestSuiteDocumentation(object):
    """Class for storing scraped information from test suite documentation for a camelot TestSuite.

    :param str name: The class name of the suite
    :param str full_name: The module + class name of the suite
    :param str docstring: the test suites doc string
    :param str init_code: the python code of the init method of the code as a string
    :param str setup_code: the python code of the set_up method of the suite as a string
    :param list[camelot.TestCase] tests: list of test cases that are attached to the suite
                                                       when instantiated
    :param list[valence.models.RepoChangeset] changesets: list of changesets for repos in environment
    :param list[valence.scribe.TestCaseDocumentation] testcases_docs: list of test case documentation
        objects that are contained in this suite.
    """

    def __init__(self, name, full_name, docstring, init_code, setup_code, tests, changesets,
                 testcases_docs):
        self.name = name
        self.full_name = full_name
        self.docstring = docstring
        self.init_code = init_code
        self.setup_code = setup_code
        self.tests = [test.__module__ + '.' + test.name for test in tests]
        self.changesets = [cs.serialize() for cs in changesets]
        self.testcases_docs = testcases_docs

    @classmethod
    def from_test_suite(cls, test_suite, changesets, config_dm=None):
        """Class method for creating a TestSuiteDocumentation object from a test suite.

        :param six.class_types test_suite: Test suite class to inspect
        :param list[valence.models.RepoChangeset] changesets: list of changesets for repos in environment
        :param valence.models.SuiteConfig config_dm: SuiteConfig needed for this test suite
        :rtype: valence.scribe.TestSuiteDocumentation
        """

        docstring = test_suite.__doc__
        init_code = inspect.getsource(test_suite.__init__)
        setup_code = inspect.getsource(test_suite.setup)
        resources_dm = driver.get_mock_resources_dm()
        # We don't scribe creating artifacts.  Later unit tests may require it, so restore..
        save_skip_artifacts = test_suite._skip_artifact_dir  # pylint: disable=protected-access
        test_suite._skip_artifact_dir = True   # pylint: disable=protected-access
        # Thanks, globovars!
        models._resource._clear_resources()  # pylint: disable=protected-access
        sf_artifacts.get_and_set_global_manager({'local': True})
        suite = test_suite(resources_dm, config_dm)
        test_suite._skip_artifact_dir = save_skip_artifacts  # pylint: disable=protected-access

        tests = suite.tests
        name = suite.name
        full_name = suite.full_name

        testcases_docs = [TestCaseDocumentation.from_test_case(tt.__class__) for tt in tests]

        return cls(name, full_name, docstring, init_code, setup_code, tests, changesets, testcases_docs)

    def to_json(self):
        """Returns JSON/dict matching GROOT schema requirements.

        :rtype: dict
        """
        result = copy.deepcopy(self.__dict__)
        # The testcase_docs attr of the class isn't needed for groot.
        # Yeah, this is a hack to not change the test suite groot structure.
        del result['testcases_docs']
        return result


class ModuleDocumentation(object):
    """Class to store all Documentation objects from a given module.

    :param str module_path: full path to importable module
    :param valence.models.SuiteConfig config_dm: SuiteConfig for documentation process
    """

    def __init__(self, module_path, config_dm=None):
        self.module = importlib.import_module(module_path)
        self.config_dm = models.SuiteConfig() if config_dm is None else config_dm
        self.changesets = driver.get_changesets(self.config_dm.changeset_modules)
        self._test_case_docs = None
        self._test_suite_docs = None
        self.test_case_errors = []

    @property
    def test_case_docs(self):
        if self._test_case_docs is None:
            self._build_test_case_docs()
        return self._test_case_docs

    def _build_test_case_docs(self):
        """Uses inspection module to build the test case docs."""
        test_cases = inspection.get_cases_from_module(self.module)
        self._test_case_docs = []

        for test_case in test_cases:
            try:
                test_case_doc = TestCaseDocumentation.from_test_case(test_case)
                self._test_case_docs.append(test_case_doc)
                self.test_case_errors.extend(test_case_doc.test_case_errors)
            except (_errors.NoDocstringError, camelot.InspectionError) as e:
                self.test_case_errors.append(inspection.CaseInspectorProblem(
                    module_name=self.module.__name__, case_name=test_case.__name__, error_msg=str(e)))

    @property
    def test_suite_docs(self):
        if self._test_suite_docs is None:
            self._build_test_suite_docs()
        return self._test_suite_docs

    def _build_test_suite_docs(self):
        """Uses inspection module to build the test suite docs."""
        test_suites = inspection.get_suites_from_module(self.module)

        # TODO : after refactoring inspection, add error checking
        self._test_suite_docs = [TestSuiteDocumentation.from_test_suite(
            t, self.changesets, self.config_dm) for t in test_suites]
