"""This module is for exporting module documentation information."""
from __future__ import print_function

import csv
import os

import autologger

from camelot import groot
from valence.scribe import _errors

_log = autologger.get_sflogging(__name__)


def post_to_groot(module_docs):
    """Posts serialized objects in (test_case_docs, test_suite_docs) to GROOT.

    :param valence.scribe.ModuleDocumentation module_docs: module documentation object
    """

    if not module_docs.test_case_docs and not module_docs.test_suite_docs:
        raise _errors.EmptyModuleError('No Camelot test cases or suites found in module.')

    if module_docs.test_case_docs:
        post_data = [case.to_json() for case in module_docs.test_case_docs]
        groot.post_to_groot(groot.CASES_ENDPOINT, post_data)

    if module_docs.test_suite_docs:
        post_data = [suite.to_json() for suite in module_docs.test_suite_docs]
        groot.post_to_groot(groot.SUITES_ENDPOINT, post_data)


def export_test_cases_to_csv(module_docs, filename='testcases.csv', append=False, ignore_errors=False):
    """Exports all test case documentation to a formatted csv file.

    :param valence.scribe.ModuleDocumentation module_docs: module documentation object
    :param str filename: filename of csv output
    :param bool append: append an existing file
    :param bool ignore_errors: Whether to ignore Scribe errors
    """
    if module_docs.test_case_docs:
        _export_single_test_case_list(
            module_docs.test_case_docs, module_docs.test_case_errors, filename, append, ignore_errors)

    if module_docs.test_suite_docs:
        _export_suites(module_docs.test_suite_docs, filename, append)
        for suite in module_docs.test_suite_docs:
            _export_single_test_case_list(
                suite.testcases_docs, None, filename, append, ignore_errors, suite.name)


def _export_suites(suites, filename='Suites_overview.csv', append=False):
    """Exports all test suites documentation to a formatted suites overview csv file.

    :param list[valence.scribe.TestSuiteDocumentation] suites: list of suite docs
    :param str filename: filename of csv output
    :param bool append: append an existing file
    """

    print('\nGenerating Suite overview {}...'.format(filename))
    head, tail = os.path.split(filename)
    filename = os.path.join(head, '{}{}'.format('Suites_overview_', tail))
    with open(filename, 'a' if append else 'w') as csv_file:
        case_writer = csv.writer(csv_file)
        case_writer.writerow(['#', 'Suite Name', 'Suite Objective',
                              'HW', 'Config Parameters', 'IO Y/N?',
                              'Source Release', 'Feature/Functionality Under Test', 'Est. Duration',
                              'Owner', 'Cadence', 'Misc Info',
                              'Known Open Bugs'])
        for index, suite in enumerate(suites):
            case_writer.writerow([
                index, suite.name, suite.docstring,
                '', '', 'Y',
                '', '', '00:00',
                '', '', '',
                ''])

    print('\nGenerated {}'.format(filename))


def _export_single_test_case_list(test_case_docs, test_case_errors, filename='testcases.csv',
                                  append=False, ignore_errors=False, suite_name_prefix=''):
    """Exports all test case documentation to a formatted csv file.

    :param list[valence.scribe.TestCaseDocumentation] test_case_docs: module documentation object
    :param list[str]|None test_case_errors: errors for the testcases
    :param str filename: filename of csv output
    :param bool append: append an existing file
    :param bool ignore_errors: Whether to ignore Scribe errors
    :param str suite_name_prefix: Prefix for filename based on suite name.
    """
    suite_name_prefix = '{}_'.format(suite_name_prefix) if suite_name_prefix else ''
    head, tail = os.path.split(filename)
    filename = os.path.join(head, '{}{}'.format(suite_name_prefix, tail))

    if test_case_errors:
        print('\nErrors found:')
        for tc_error in sorted(test_case_errors,
                               key=lambda x: (x.module_name, x.line_number)):
            print(tc_error)

        if not ignore_errors:
            print('\nAbort generating {}'.format(filename))
            return

    blank_prefix = ['', '', '']
    blank_postfix = ['', '']

    print('\nGenerating {}...'.format(filename))

    with open(filename, 'a' if append else 'w') as csv_file:
        case_writer = csv.writer(csv_file)
        case_writer.writerow(['Name', 'Summary', 'Requirements', 'Step', 'Step Details',
                              'Expected Result', 'Comments', 'Expected Runtime'])
        for test in test_case_docs:
            case_writer.writerow(
                [test.name, test.description, test.requirements, '', '', '', '', '00:00'])
            _write_test_case_method(test.preconditions, '_preconditions', case_writer,
                                    blank_prefix, blank_postfix)
            _write_test_case_method(test.setup, '_setup', case_writer, blank_prefix, blank_postfix)
            for step in test.steps:
                real_result = step.expected_result if step.expected_result != 'Pass' else ''
                case_writer.writerow(
                    blank_prefix + [step.index, step.action, real_result] + blank_postfix)
            _write_test_case_method(test.post_test_assertions, '_post_test_assertions',
                                    case_writer, blank_prefix, blank_postfix)
            _write_test_case_method(test.teardown, '_teardown', case_writer, blank_prefix, blank_postfix)

    print('\nGenerated {}'.format(filename))


def _write_test_case_method(method, method_name, case_writer, prefix, postfix):
    """Writes a single row.

    :param str method: Docstring for the method we're writing.
    :param str method_name: Name of the method
    :param Any case_writer: Writer
    :param list[str] prefix: prefix for row
    :param list[str] postfix: postfix for row
    """
    if method and case_writer and method_name:
        case_writer.writerow(prefix + [method_name, method, ''] + postfix)


def post_test_cases_to_qtest(module_docs, username, password, project_id):
    """Posts structured test information to qTest.

    Uses the qTest SDK to do the following:

    - Check if Test Case is already defined in qTest (uses the python path to determine this)
        - If not, create the Test Case model and submit it to qTest
        - If so, update qTest resource if hash of source code does not match current qTest version.
    - Attempts to link Test Cases to Requirements that have been imported to qTest.

    :param valence.scribe.ModuleDocumentation module_docs: module documentation object
    :param str username: qTest project username (email)
    :param str password: qTest project password
    :param int project_id: ID of project to scrape test information to
    """
    from qtestapi import QTestAPI
    from valence.scribe import _qt_constants
    from valence.scribe import _qt_helpers

    if not module_docs.test_case_docs:
        raise _errors.EmptyModuleError('No Camelot test cases found in module.')

    api = QTestAPI(username, password, project_id)

    for test_case_doc in module_docs.test_case_docs:
        qt_errors = test_case_doc.get_qtest_errors()
        if qt_errors:
            _log.warning('Skipping case - {}'.format(qt_errors))
            continue

        # Determine if the test case already exists in qTest
        results = api.search_test_cases_by_name(test_case_doc.name)
        exact_results = [tc for tc in results if tc.name == test_case_doc.name
                         and tc.properties[_qt_constants.PATH].field_value == test_case_doc.test_path]

        if exact_results:
            matched_result = exact_results[0]
            # Try to use a hash of the source lines from inspection module to determine
            #   if an update is required for qTest.
            update_required = False
            matched_hash = matched_result.properties[_qt_constants.SOURCE_HASH].field_value
            if test_case_doc.source_hash != matched_hash:
                update_required = True

            if update_required:
                print('Found matching case, update deemed required...')
                updated_tc = _qt_helpers.qtest_case_builder(api, test_case_doc,
                                                            test_case_dm=matched_result)
                test_case = api.update_resource(updated_tc)
                updated_test_case = api.approve_test_case(test_case)
            else:
                print('Source code hash matches, no need to update...')
                updated_test_case = matched_result

        else:
            # create new test_case
            new_case = _qt_helpers.qtest_case_builder(api, test_case_doc)
            updated_test_case = api.create_resource(new_case)

        # Sync JIRA requirements with those listed in docstring
        _qt_helpers.check_case_requirements(api, test_case_doc, updated_test_case)
