"""Scribe test-case docstring verification."""
from __future__ import print_function

import os

import repo_utils
from valence.scribe import _scribe_in


class ScribeDocstringChecker(object):
    """Validates whether TestCase classes for provided input conform to standards.

    :param str check_path: Path containing files to run check against.
    :param bool modified_only: Indicates whether to check only modified files in repo
    :param repo_utils.Repo repo: Repo under analysis
    """
    def __init__(self, check_path, modified_only=True, repo=None):
        self.check_path = check_path
        self.modified_only = modified_only
        self.repo = repo
        self.module_errors = []

    @staticmethod
    def _get_tree_python_files(tree_root_dir):
        """Recursively searching file system tree for files with '.py' extension.

        :param str tree_root_dir: Path to root of directory to search
        :rtype: list[str]
        :returns: list of paths to python files
        """
        python_files = []
        for root, _, files in os.walk(tree_root_dir):
            for file_ in files:
                if file_.endswith('.py'):
                    python_files.append(os.path.join(root, file_))
        return python_files

    def _gather_module_errors(self, module_path):
        """Gather scribe related errors on a module.  Store them in the module_errors list.

        :param str module_path: path to module for analysis.
        """
        module_docs = _scribe_in.ModuleDocumentation(module_path)
        test_case_docs = module_docs.test_case_docs

        for tc_doc in test_case_docs:
            self.module_errors.extend(['{}:{}'.format(module_path, err)
                                      for err in tc_doc.get_qtest_errors()])
            self.module_errors.extend([str(err) for err in tc_doc.test_case_errors])

    def run_verification(self):
        """Executes scribe docstring analysis checks.  Returns results.

        :rtype: list[str]
        :returns: list of errors found from analysis.
        """
        files_to_check = []
        if not self.modified_only:
            files_to_check.extend(self._get_tree_python_files(
                os.path.join(self.repo.root_path, self.check_path)))
        else:
            files_to_check.extend(self.repo.get_py_changes(os.path.join(self.repo.root_path,
                                                                        self.check_path)))
        for f2c in files_to_check:
            module_path = repo_utils.convert_to_python_path(f2c)
            self._gather_module_errors(module_path)

        return self.module_errors
