import unittest

import mock
import pytest
from qtestapi.tests import utils

from valence import (
    models,
    scribe,
)
from valence.scribe import _qt_helpers

DUMMY_CHANGESET_LIST = [
    models.RepoChangeset({'repo_name': 'valence', 'changeset_id': 'abcd1234'}),
    models.RepoChangeset({'repo_name': 'staging-autoplatform', 'changeset_id': 'sillystring'})
]

MOCK_API = utils.get_mock_project()
MOCK_TEST_CASE = utils.get_ns_mock_model('test-cases')
MOCK_REQ_0 = utils.get_ns_mock_model('requirements')
MOCK_REQ_1 = utils.get_ns_mock_model('requirements')


@pytest.mark.timeout(1)
class TestQtestCaseBuilder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api = MOCK_API
        cls.mocked_test_case = MOCK_TEST_CASE
        with mock.patch('camelot.driver.get_changesets') as mock_changesets:
            mock_changesets.return_value = DUMMY_CHANGESET_LIST
            docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
        cls.tc_doc = docs.test_case_docs[0]

    def test_builder_new_test_case(self):
        """Tests building new testcase model from steps, name."""
        tc = _qt_helpers.qtest_case_builder(self.api, self.tc_doc)

        assert tc.name == 'ExampleFeatureTestCase'
        assert tc.name_space == 'test-cases'
        assert tc.properties['Status'].hr_value == 'New'
        assert tc.properties['Author'].hr_value == 'garrett.berg@solidfire.com'

    def test_builder_existing_case(self):
        """Verify that we update all of the properties on the mocked model."""
        self.mocked_test_case.properties['Author'].update('shaun.adams@netapp.com')

        tc = _qt_helpers.qtest_case_builder(self.api, self.tc_doc, self.mocked_test_case)

        assert tc.name == 'ExampleFeatureTestCase'
        assert tc.name_space == 'test-cases'
        assert tc.properties['Status'].hr_value == 'New'
        assert tc.properties['Author'].hr_value == 'garrett.berg@solidfire.com'


@pytest.mark.timeout(1)
class TestQtestCaseRequirements(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.api = MOCK_API
        cls.mocked_test_case = MOCK_TEST_CASE
        req_0 = MOCK_REQ_0
        req_0.name = 'FR-EX-1 requirement'
        req_1 = MOCK_REQ_1
        req_1.name = 'FR-EX-0 requirement'
        cls.mocked_requirements = [req_0, req_1]

        with mock.patch('camelot.driver.get_changesets') as changeset_mock:
            changeset_mock.return_value = DUMMY_CHANGESET_LIST
            docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
            cls.tc_docs = docs.test_case_docs

    def test_linked_requirement(self):
        """Verify linked requirement is found."""
        self.api.get_all_requirements = mock.MagicMock()
        self.api.get_all_requirements.return_value = self.mocked_requirements

        self.api.get_test_case_linked_requirements = mock.MagicMock()
        self.api.get_test_case_linked_requirements.return_value = [self.mocked_requirements[0]]

        self.api.link_test_case_to_requirement = mock.MagicMock()
        _qt_helpers.check_case_requirements(self.api, self.tc_docs[0], self.mocked_test_case)
        assert not self.api.link_test_case_to_requirement.called

    def test_unlinked_requirement(self):
        """Verify we attempt to link unlinked requirements."""
        self.api.get_all_requirements = mock.MagicMock()
        self.api.get_all_requirements.return_value = self.mocked_requirements

        self.api.get_test_case_linked_requirements = mock.MagicMock()
        self.api.get_test_case_linked_requirements.return_value = []

        self.api.link_test_case_to_requirement = mock.MagicMock()
        _qt_helpers.check_case_requirements(self.api, self.tc_docs[0], self.mocked_test_case)

        self.api.link_test_case_to_requirement.assert_called_once()

    def test_unimported_requirement(self):
        """Check that we do nothing for requirements we can't find in qTest."""
        self.api.get_all_requirements = mock.MagicMock()
        self.api.get_all_requirements.return_value = self.mocked_requirements

        self.api.get_test_case_linked_requirements = mock.MagicMock()
        self.api.get_test_case_linked_requirements.return_value = []
        self.api.link_test_case_to_requirement = mock.MagicMock()

        _qt_helpers.check_case_requirements(self.api, self.tc_docs[1], self.mocked_test_case)

    def test_stale_requirement(self):
        """Check that requirements that are no longer valid get unlinked."""
        self.api.get_all_requirements = mock.MagicMock()
        self.api.get_all_requirements.return_value = self.mocked_requirements

        self.api.get_test_case_linked_requirements = mock.MagicMock()
        self.api.get_test_case_linked_requirements.return_value = self.mocked_requirements
        self.api.link_test_case_to_requirement = mock.MagicMock()
        self.api.remove_test_case_link_to_requirement = mock.MagicMock()

        _qt_helpers.check_case_requirements(self.api, self.tc_docs[0], self.mocked_test_case)

        self.api.remove_test_case_link_to_requirement.assert_called_once()
