"""Test the valence.camelot.scribe module."""

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import json
import os
import sys
import unittest

import mock
import pytest
import requests
from qtestapi.tests import utils

from camelot import (
    driver,
    groot,
    inspection,
)
from camelot.driver import _list_command
from valence import (
    models,
    scribe,
    unittest_utils,
)
from valence.scribe import _qt_constants

DUMMY_CHANGESET_LIST = [
    models.RepoChangeset({'repo_name': 'valence', 'changeset_id': 'abcd1234'}),
    models.RepoChangeset({'repo_name': 'staging-autoplatform', 'changeset_id': 'sillystring'})
]

module_mock = unittest_utils.ModuleLevelMock()
MOCK_API = utils.get_mock_project()
USERNAME = 'username'
PASSWORD = 'password'
PROJECT_ID = 12345

with mock.patch('camelot.driver.get_changesets') as mock_cs:
    ex_path = 'valence.unittest_utils.cases.scribe_examples'
    inheritance_path = 'valence.unittest_utils.cases.scribe_inheritance_examples'
    suite_pass_path = 'valence.unittest_utils.suites.passing_suite'
    fail_ex_path = 'valence.unittest_utils.cases.scribe_failing_examples'
    mock_cs.return_value = DUMMY_CHANGESET_LIST
    SCRIBE_EX_DOCS = scribe.ModuleDocumentation(ex_path)
    SCRIBE_INHERITANCE_DOCS = scribe.ModuleDocumentation(inheritance_path)
    SCRIBE_SUITE_CASES_DOCS = scribe.ModuleDocumentation(suite_pass_path)
    SCRIBE_FAIL_EX_DOCS = scribe.ModuleDocumentation(fail_ex_path)


def setUpModule():
    # Mock get_resource_dm since it is time consuming and can cause timeouts
    module_mock.mock(_list_command, 'get_resources_dm',
                     mock.MagicMock(return_value=models.SuiteResources.get_mock_object()))
    module_mock.mock(driver, 'get_changesets',
                     mock.MagicMock(return_value=DUMMY_CHANGESET_LIST))


def tearDownModule():
    module_mock.restore()


@pytest.mark.timeout(1)
class EmptyModulePathTest(unittest.TestCase):
    def test_invalid_module_path(self):
        module_doc = scribe.ModuleDocumentation('valence.unittest_utils.components')

        with pytest.raises(scribe.EmptyModuleError):
            scribe.post_test_cases_to_qtest(module_doc, USERNAME, PASSWORD, PROJECT_ID)

        with pytest.raises(scribe.EmptyModuleError):
            scribe.post_to_groot(module_doc)


@pytest.mark.timeout(1)
class NoDocstringTestCaseTest(unittest.TestCase):
    """Tests that we properly get error message when the test case has no docstring."""
    def test_no_docstring_test_case(self):
        module_doc = scribe.ModuleDocumentation(
            'valence.scribe.tests.cases.scribe_no_docstring_examples')
        module_doc._build_test_case_docs()
        assert 'Test case class has no docstring' == module_doc.test_case_errors[0].error_msg


@pytest.mark.timeout(1)
class QtestScribeFieldMapTest(unittest.TestCase):
    def test_scribe_attrs(self):
        docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')

        for qt_field, scribe_attr in _qt_constants.QTEST_FIELD_MAP.items():
            assert hasattr(docs.test_case_docs[0], scribe_attr)


@pytest.mark.timeout(1)
class PostQtestCaseModuleTest(unittest.TestCase):
    """Test suite for the export to qTest scribe action."""
    def setUp(self):
        tc_0 = utils.get_ns_mock_model('test-cases', api=MOCK_API)
        tc_0.name = 'ExampleFeatureTestCase'
        tc_0.properties['Source Hash'].update('deadbeef')
        py_path = 'valence.unittest_utils.cases.scribe_examples.ExampleFeatureTestCase'
        tc_0.properties['Path'].update(py_path)

        tc_1 = utils.get_ns_mock_model('test-cases', api=MOCK_API)
        tc_1.name = 'SecondExampleFeatureTestCase'
        tc_1.properties['Source Hash'].update('feedbeef')
        py_path = 'valence.unittest_utils.cases.scribe_examples.SecondExampleFeatureTestCase'
        tc_1.properties['Path'].update(py_path)

        self.duped_test_cases = [tc_0, tc_0]
        self.test_cases = [tc_0, tc_1]

        req_0 = utils.get_ns_mock_model('requirements', api=MOCK_API)
        req_0.name = 'FR-EX-1 requirement 1'

        req_1 = utils.get_ns_mock_model('requirements', api=MOCK_API)
        req_1.name = 'FR-EX-2 requirement 2'
        self.requirements = [req_0, req_1]

    @mock.patch('qtestapi.QTestAPI')
    def test_case_field_validation_fail(self, mock_api):
        mocked_client = mock.Mock()
        mock_api.return_value = mocked_client

        docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
        docs._test_case_docs = [docs.test_case_docs[2]]
        mocked_client.search_test_cases_by_name.return_value = self.test_cases
        scribe.post_test_cases_to_qtest(docs, USERNAME, PASSWORD, PROJECT_ID)

    @mock.patch('valence.scribe._qt_helpers.qtest_case_builder')
    @mock.patch('valence.scribe._qt_helpers.check_case_requirements')
    @mock.patch.object(inspection.CaseInspector, 'get_source_hash')
    @mock.patch('qtestapi.QTestAPI')
    def test_existing_case_update_required(self, mock_api, mock_get_hash, mock_req_check, mock_builder):
        mocked_client = mock.Mock()
        mock_api.return_value = mocked_client

        # hashes shouldn't match
        mock_get_hash.return_value = 'feedbeef'

        docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
        docs._test_case_docs = [docs.test_case_docs[0]]

        mocked_client.search_test_cases_by_name.return_value = self.test_cases
        scribe.post_test_cases_to_qtest(docs, USERNAME, PASSWORD, PROJECT_ID)

        mocked_client.update_resource.assert_called_once()
        mocked_client.approve_test_case.assert_called_once()

    @mock.patch('valence.scribe._qt_helpers.qtest_case_builder')
    @mock.patch('valence.scribe._qt_helpers.check_case_requirements')
    @mock.patch.object(inspection.CaseInspector, 'get_source_hash')
    @mock.patch('qtestapi.QTestAPI')
    def test_existing_case_no_update_required(self, mock_api, mock_get_hash, mock_req_check,
                                              mock_builder):
        mocked_client = mock.Mock()
        mock_api.return_value = mocked_client

        # hashes match
        mock_get_hash.return_value = 'deadbeef'

        docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
        docs._test_case_docs = [docs.test_case_docs[0]]

        mocked_client.search_test_cases_by_name.return_value = self.test_cases

        scribe.post_test_cases_to_qtest(docs, USERNAME, PASSWORD, PROJECT_ID)

        assert mocked_client.update_resource.call_count == 0

    @mock.patch('valence.scribe._qt_helpers.qtest_case_builder')
    @mock.patch('valence.scribe._qt_helpers.check_case_requirements')
    @mock.patch.object(inspection.CaseInspector, 'get_source_hash')
    @mock.patch('qtestapi.QTestAPI')
    def test_new_case(self, mock_api, mock_get_hash, mock_req_check, mock_builder):
        mocked_client = mock.Mock()
        mock_api.return_value = mocked_client

        docs = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
        docs._test_case_docs = [docs.test_case_docs[1]]
        mocked_client.search_test_cases_by_name.return_value = [self.test_cases[0]]

        scribe.post_test_cases_to_qtest(docs, USERNAME, PASSWORD, PROJECT_ID)

        mocked_client.create_resource.assert_called_once()


@pytest.mark.timeout(1)
class PostGrootSuiteModuleTest(unittest.TestCase):
    """Test suite for export JSON method in scribe"""

    @mock.patch('valence.models.Base.validate')
    @mock.patch('camelot.driver.get_mock_resources_dm')
    @mock.patch('requests.post')
    def test_post_to_groot(self, post_mock, mock_res, mock_val):
        """ Tests the export to json method creates a json with filename parameter"""
        documentation = scribe.ModuleDocumentation('valence.unittest_utils.suites.passing_suite')
        mock_res.return_value = models.SuiteResources.get_mock_object()
        scribe.post_to_groot(documentation)
        post_data = {'json_payload': json.dumps([suite.to_json() for suite in
                                                 documentation.test_suite_docs])}
        post_mock.assert_called_with(groot.GROOT_URL + '/suites',
                                     data=post_data)

    @mock.patch('valence.models.Base.validate')
    @mock.patch('camelot.driver.get_mock_resources_dm')
    @mock.patch('requests.post')
    def test_requests_exception(self, post_mock, mock_res, mock_val):
        """Confirm failure path is exercised when requests module throws (any) exception"""
        documentation = scribe.ModuleDocumentation('valence.unittest_utils.suites.passing_suite')
        mock_res.return_value = models.SuiteResources.get_mock_object()
        post_mock.side_effect = requests.RequestException()
        scribe.post_to_groot(documentation)


@pytest.mark.timeout(1)
class PostGrootTestCaseModuleTest(unittest.TestCase):
    """Test suite for export JSON method in scribe"""
    @mock.patch('requests.post')
    def test_post_to_groot(self, post_mock):
        """ Tests the export to json method creates a json with filename parameter"""
        documentation = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')

        scribe.post_to_groot(documentation)
        post_data = {'json_payload': json.dumps([case.to_json() for case in
                                                 documentation._test_case_docs])}
        post_mock.assert_called_with(groot.GROOT_URL + '/cases',
                                     data=post_data)

    @mock.patch('requests.post')
    def test_requests_exception(self, post_mock):
        """Confirm failure path is exercised when requests module throws (any) exception"""
        post_mock.side_effect = requests.RequestException()

        documentation = scribe.ModuleDocumentation('valence.unittest_utils.cases.scribe_examples')
        scribe.post_to_groot(documentation)


@pytest.mark.timeout(1)
class CsvExportTestCaseModuleTest(unittest.TestCase):
    """Test suite for export method in scribe"""

    @classmethod
    def setUpClass(cls):
        cls.documentation = SCRIBE_EX_DOCS
        cls.suite_documentation = SCRIBE_SUITE_CASES_DOCS
        cls.fail_documentation = SCRIBE_FAIL_EX_DOCS

        cls.my_filename = '/tmp/some/tempfile'

        if sys.version_info >= (3, 0):
            cls.mock_open_str = 'builtins.open'
        else:
            cls.mock_open_str = '__builtin__.open'

    def test_export_to_csv(self):
        """ Tests the export to csv method creates a csv with filename parameter"""
        m = mock.mock_open()
        test_cases = self.documentation.test_case_docs
        with mock.patch(self.mock_open_str, m):
            scribe.export_test_cases_to_csv(self.documentation, self.my_filename)
        assert m.mock_calls[0][1][0] == self.my_filename
        num_writes = len([c for c in m.mock_calls if 'write' in c[0]])
        assert num_writes >= 15  # Of steps in the 3 example test cases + header

    @mock.patch('camelot.driver.get_mock_resources_dm')
    def test_export_suite_to_csv(self, mock_res):
        """ Tests the suite export to csv method creates a csv with filename parameter"""
        m = mock.mock_open()
        mock_res.return_value = models.SuiteResources.get_mock_object()
        suites = self.suite_documentation.test_suite_docs
        with mock.patch(self.mock_open_str, m):
            scribe.export_test_cases_to_csv(self.suite_documentation, self.my_filename)

        suite_filenames = []
        head, tail = os.path.split(self.my_filename)
        suite_filenames.append(os.path.join(head, 'Suites_overview_' + tail))

        for s in suites:
            suite_filenames.append(os.path.join(head, '{}_{}'.format(s.name, tail)))

        # Here is the mock call to Open:
        # call('/tmp/some/MultiErrorFailingSuiteArtifacts_tempfile', 'w')
        # Other calls will be .write or enter or whatever. We only care about open calls.
        filenames_from_open_calls = []
        for m in m.mock_calls:
            if m[1] and len(m[1]) > 1 and m[1][1] == 'w':
                filenames_from_open_calls.append(m[1][0])

        assert filenames_from_open_calls == suite_filenames

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_export_to_csv_fail(self, stdout_mock):
        """ Tests the export to csv method using failing example creates a csv with filename parameter"""
        m = mock.mock_open()
        test_cases = self.fail_documentation.test_case_docs
        with mock.patch(self.mock_open_str, m):
            scribe.export_test_cases_to_csv(self.fail_documentation, self.my_filename)
        output = stdout_mock.getvalue()

        assert 'Errors found' in output

    def test_append_csv(self):
        """ Test the export csv method can append an existing file"""
        m = mock.mock_open()
        test_cases = self.documentation.test_case_docs
        with mock.patch(self.mock_open_str, m):
            scribe.export_test_cases_to_csv(self.documentation, self.my_filename, append=True)
        assert m.mock_calls[0][1][0] == self.my_filename
        num_writes = len([c for c in m.mock_calls if 'write' in c[0]])
        assert num_writes >= 15  # Of steps in the 3 example test cases + header


@pytest.mark.timeout(1)
class ScribeTest(unittest.TestCase):
    """Unit test suite for scribe module"""

    @classmethod
    def setUpClass(cls):
        cls.first_case = SCRIBE_EX_DOCS.test_case_docs[0]
        cls.second_case = SCRIBE_EX_DOCS.test_case_docs[1]
        cls.third_case = SCRIBE_EX_DOCS.test_case_docs[2]
        cls.baseclass_case = SCRIBE_INHERITANCE_DOCS.test_case_docs[0]
        cls.sub_without_super_case = SCRIBE_INHERITANCE_DOCS.test_case_docs[1]

    def test_author(self):
        """ Tests that author is found correctly"""
        assert self.first_case.author == 'garrett.berg@solidfire.com'
        assert self.second_case.author == 'sean.turner@solidfire.com'
        assert self.third_case.author is None

    def test_created_on(self):
        """ Tests that created_on is found correctly"""
        assert self.second_case.created_on == '4/25/17'
        assert self.first_case.created_on == '2/8/17'
        assert self.third_case.created_on is None

    def test_feature(self):
        """ Tests that feature is found correctly"""
        assert self.first_case.feature == 'EXP'
        assert self.second_case.feature == 'OTHER'
        assert self.third_case.feature is None

    def test_preconditions(self):
        """ Tests that preconditions are found correctly"""
        assert 'Preconditions.' in self.first_case.preconditions
        assert self.third_case.preconditions is None

    def test_setup(self):
        """ Tests that preconditions are found correctly"""
        assert 'No setup for this test.' in self.first_case.setup
        assert self.third_case.setup is None

    def test_teardown(self):
        """ Tests that preconditions are found correctly"""
        assert self.first_case.teardown is None
        assert '3rd testcase teardown.' in self.third_case.teardown

    def test_posttest_assertions(self):
        """ Tests that post test assertions are found correctly"""
        assert self.first_case.post_test_assertions is None
        assert 'Verifies that everything is fine in post-test.' in self.third_case.post_test_assertions

    def test_requirements(self):
        """ Tests that requirements are found correctly"""
        assert self.first_case.requirements == 'FR-EX-1'
        assert self.second_case.requirements == 'FR-EX-2'
        assert self.third_case.requirements is None

    def test_description(self):
        """ Tests that description is found correctly"""
        description_contains = 'This is an example feature-specific test case.'
        description_excludes = ['tc_info:', ':param', ':var']
        assert description_contains in self.first_case.description
        for keyword in description_excludes:
            self.assertNotIn(keyword, self.first_case.description)
        assert 'Terrible description.' == self.third_case.description

    def test_field_validation(self):
        """Tests that a test-case containing all required fields positively validates"""
        assert not self.first_case.get_qtest_errors()

    def test_field_validation_fail(self):
        """Tests that a test-case missing required fields negatively validates"""
        assert self.third_case.get_qtest_errors()

    def test_scribe_failure(self):
        """Tests that we properly get error object when we can't build documentation for a module."""
        module_doc = SCRIBE_FAIL_EX_DOCS
        module_doc._build_test_case_docs()

        assert hasattr(module_doc.test_case_errors[0], 'module_name')
        assert hasattr(module_doc.test_case_errors[0], 'case_name')
        assert hasattr(module_doc.test_case_errors[0], 'error_msg')
        assert hasattr(module_doc.test_case_errors[0], 'line_number')
        error_messages = [e.error_msg for e in module_doc.test_case_errors]
        assert inspection.RETURN_INSIDE_TEST_ERROR in error_messages
        assert inspection.SUPER_TEST_CALLED_IN_TEST in error_messages

    def test_scribe_step_inheritance(self):
        """Test that subclass with super calls properly lists baseclass steps."""
        assert len(self.baseclass_case.steps) == 6
        assert len(self.sub_without_super_case.steps) == 1
