"""Valence Cluster Capacity Selectors."""
import abc
import math

import autologger
import six

from valence import (
    functional_models,
    models,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class ClusterCapacitySelector(object):
    """Base selector for bytes of cluster capacity.

    NOTE: On a cluster with 40GB of used space:
    100GB = ClusterCapacitySelector(100GB)
    60GB = ClusterCapacitySelector(100GB, relative_to_used_capacity=True)

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param bool relative_to_used_capacity: Whether the selected DataSize should be calculated relative
        to the existing state of cluster fullness.
    """
    def __init__(self, cluster_dm, relative_to_used_capacity=False):
        self.relative = relative_to_used_capacity
        self.target_bytes = None

        self._cluster_fm = functional_models.Cluster(cluster_dm)
        self._added_selectors = []

    def __call__(self):
        """Return a DataSize object representing the desired cluster capacity.

        :rtype: valence.models.DataSize
        """
        cluster_full_threshold = self._cluster_fm.get_cluster_full_threshold()
        self.target_bytes = self.get_bytes(cluster_full_threshold)
        if not self.relative:
            return models.DataSize(self.target_bytes)
        else:
            return models.DataSize(self.target_bytes - self._get_used_bytes(cluster_full_threshold))

    def get_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        Adjust the number of bytes by the bytes associated with any added selectors.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        target_bytes = self._get_target_bytes(cluster_full_threshold)
        for selector in self._added_selectors:
            target_bytes += selector.get_bytes(cluster_full_threshold)
        return target_bytes

    @abc.abstractmethod
    def _get_used_bytes(self, cluster_full_threshold):
        """Get the number of used bytes on the cluster.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """

    @abc.abstractmethod
    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """


@six.add_metaclass(abc.ABCMeta)
class ClusterBlockCapacitySelector(ClusterCapacitySelector):
    """Base selector for bytes of cluster block capacity."""

    def add(self, capacity_selector):
        """Add another selector by which to adjust the selected DataSize.

        :param valence.selectors.ClusterBlockCapacitySelector capacity_selector: Block capacity selector.
        """
        assert not capacity_selector.relative
        self._added_selectors.append(capacity_selector)

    def _get_used_bytes(self, cluster_full_threshold):
        """Get the number of used bytes on the cluster.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        return cluster_full_threshold.sum_used_cluster_bytes


class FixedBlockCapacitySelector(ClusterBlockCapacitySelector):
    """Selector for a fixed number of bytes of cluster block capacity.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int cluster_bytes: Number of bytes as will appear on the cluster.
    :param bool relative_to_used_capacity: Whether the selected DataSize should be calculated relative
        to the existing state of cluster fullness.
    """
    def __init__(self, cluster_dm, cluster_bytes, relative_to_used_capacity=False):
        super(FixedBlockCapacitySelector, self).__init__(cluster_dm, relative_to_used_capacity)
        self.cluster_bytes = cluster_bytes

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        return self.cluster_bytes


class PercentBlockCapacitySelector(ClusterBlockCapacitySelector):
    """Selector for a percentage of cluster block capacity.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int percent_capacity: Percent of total cluster block capacity.
    :param bool relative_to_used_capacity: Whether the selected DataSize should be calculated relative
        to the existing state of cluster fullness.
    """
    def __init__(self, cluster_dm, percent_capacity, relative_to_used_capacity=False):
        super(PercentBlockCapacitySelector, self).__init__(cluster_dm, relative_to_used_capacity)
        self.percent_capacity = percent_capacity
        assert -100 <= self.percent_capacity <= 100

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        total_bytes = cluster_full_threshold.sum_total_cluster_bytes
        return int(math.ceil(float(self.percent_capacity * total_bytes) / 100))


class StageBlockCapacitySelector(ClusterBlockCapacitySelector):
    """Selector for a stage of cluster block fullness.

    Returns a DataSize object representing the number of bytes needed to achieve the desired stage of
    cluster fullness relative to the current state of cluster fullness.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int stage: Stage of cluster block fullness.
    :param int into_stage_percent: Percent into the specified stage.
    """
    def __init__(self, cluster_dm, stage, into_stage_percent=0):
        super(StageBlockCapacitySelector, self).__init__(cluster_dm, True)
        self.stage = stage
        self.into_stage_percent = into_stage_percent

        assert 2 <= self.stage <= 5
        if self.stage < 5:
            assert 0 <= self.into_stage_percent <= 100
        else:
            assert self.into_stage_percent == 0

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        if self.stage == 2:
            stage_bytes = cluster_full_threshold.stage2_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage3_block_threshold_bytes
        elif self.stage == 3:
            stage_bytes = cluster_full_threshold.stage3_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage4_block_threshold_bytes
        elif self.stage == 4:
            stage_bytes = cluster_full_threshold.stage4_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage5_block_threshold_bytes
        else:
            stage_bytes = cluster_full_threshold.stage5_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage5_block_threshold_bytes
        return int(math.ceil(
            stage_bytes + float(self.into_stage_percent * (next_stage_bytes - stage_bytes)) / 100
        ))


class PercentStageBlockCapacitySelector(ClusterBlockCapacitySelector):
    """Selector for a percent of the difference between two stage thresholds of cluster block fullness.

    0% - Represents 0 bytes.
    100% - Represents the number of bytes needed to go from the specified stage to the next stage.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int stage: Stage of cluster block fullness.
    :param int percent_stage: Percent between the threshold of the specified stage and the next stage.
    """
    def __init__(self, cluster_dm, stage, percent_stage):
        super(PercentStageBlockCapacitySelector, self).__init__(cluster_dm, False)
        self.stage = stage
        self.percent_stage = percent_stage

        assert 2 <= self.stage <= 4
        assert -100 <= self.percent_stage <= 100

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        if self.stage == 2:
            stage_bytes = cluster_full_threshold.stage2_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage3_block_threshold_bytes
        elif self.stage == 3:
            stage_bytes = cluster_full_threshold.stage3_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage4_block_threshold_bytes
        else:
            stage_bytes = cluster_full_threshold.stage4_block_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage5_block_threshold_bytes
        return int(math.ceil(float(self.percent_stage * (next_stage_bytes - stage_bytes)) / 100))


@six.add_metaclass(abc.ABCMeta)
class ClusterMetaCapacitySelector(ClusterCapacitySelector):
    """Base selector for bytes of cluster metadata capacity."""

    def add(self, capacity_selector):
        """Add another selector by which to adjust the selected DataSize.

        :param valence.selectors.ClusterMetaCapacitySelector capacity_selector: Metadata capacity
            selector.
        """
        assert not capacity_selector.relative
        self._added_selectors.append(capacity_selector)

    def _get_used_bytes(self, cluster_full_threshold):
        """Get the number of used bytes on the cluster.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        return cluster_full_threshold.sum_used_metadata_cluster_bytes


class FixedMetaCapacitySelector(ClusterMetaCapacitySelector):
    """Selector for a fixed number of bytes of cluster metadata capacity.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int cluster_bytes: Number of bytes as will appear on the cluster.
    :param bool relative_to_used_capacity: Whether the selected DataSize should be calculated relative
        to the existing state of cluster fullness.
    """
    def __init__(self, cluster_dm, cluster_bytes, relative_to_used_capacity=False):
        super(FixedMetaCapacitySelector, self).__init__(cluster_dm, relative_to_used_capacity)
        self.cluster_bytes = cluster_bytes

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        return self.cluster_bytes


class PercentMetaCapacitySelector(ClusterMetaCapacitySelector):
    """Selector for a percentage of cluster metadata capacity.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int percent_capacity: Percent of total cluster metadata capacity.
    :param bool relative_to_used_capacity: Whether the selected DataSize should be calculated relative
        to the existing state of cluster fullness.
    """
    def __init__(self, cluster_dm, percent_capacity, relative_to_used_capacity=False):
        super(PercentMetaCapacitySelector, self).__init__(cluster_dm, relative_to_used_capacity)
        self.percent_capacity = percent_capacity
        assert -100 <= self.percent_capacity <= 100

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        total_bytes = cluster_full_threshold.sum_total_metadata_cluster_bytes
        return int(math.ceil(float(self.percent_capacity * total_bytes) / 100))


class StageMetadataCapacitySelector(ClusterMetaCapacitySelector):
    """Selector for a stage of cluster metadata fullness.

    Returns a DataSize object representing the number of bytes needed to achieve the desired stage of
    cluster fullness relative to the current state of cluster fullness.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int stage: Stage of cluster block fullness.
    :param int into_stage_percent: Percent into the specified stage.
    """
    def __init__(self, cluster_dm, stage, into_stage_percent=0):
        super(StageMetadataCapacitySelector, self).__init__(cluster_dm, True)
        self.stage = stage
        self.into_stage_percent = into_stage_percent

        assert 2 <= self.stage <= 5
        if self.stage < 5:
            assert 0 <= self.into_stage_percent <= 100
        else:
            assert self.into_stage_percent == 0

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        if self.stage == 2:
            stage_bytes = cluster_full_threshold.stage2_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage3_metadata_threshold_bytes
        elif self.stage == 3:
            stage_bytes = cluster_full_threshold.stage3_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage4_metadata_threshold_bytes
        elif self.stage == 4:
            stage_bytes = cluster_full_threshold.stage4_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage5_metadata_threshold_bytes
        else:
            stage_bytes = cluster_full_threshold.stage5_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage5_metadata_threshold_bytes
        return int(math.ceil(
            stage_bytes + float(self.into_stage_percent * (next_stage_bytes - stage_bytes)) / 100
        ))


class PercentStageMetaCapacitySelector(ClusterMetaCapacitySelector):
    """Selector for a percent of the difference between two stage thresholds of cluster slice fullness.

    0% - Represents 0 bytes.
    100% - Represents the number of bytes needed to go from the specified stage to the next stage.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param int stage: Stage of cluster block fullness.
    :param int percent_stage: Percent between the threshold of the specified stage and the next stage.
    """
    def __init__(self, cluster_dm, stage, percent_stage):
        super(PercentStageMetaCapacitySelector, self).__init__(cluster_dm, False)
        self.stage = stage
        self.percent_stage = percent_stage

        assert 2 <= self.stage <= 4
        assert -100 <= self.percent_stage <= 100

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param valence.models.ClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        if self.stage == 2:
            stage_bytes = cluster_full_threshold.stage2_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage3_metadata_threshold_bytes
        elif self.stage == 3:
            stage_bytes = cluster_full_threshold.stage3_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage4_metadata_threshold_bytes
        else:
            stage_bytes = cluster_full_threshold.stage4_metadata_threshold_bytes
            next_stage_bytes = cluster_full_threshold.stage5_metadata_threshold_bytes
        return int(math.ceil(float(self.percent_stage * (next_stage_bytes - stage_bytes)) / 100))
