"""Valence drive selectors."""
import random

from valence import functional_models
from valence.selectors import _errors


class DriveSelector(object):
    """Selects drives from a cluster based on various criteria.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param int num: number of drives you want to randomly select; if None, all drives are returned
    :param list[valence.enums.DriveStatus] statuses: statues you want
    :param list[valence.enums.DriveType] types: types of drives you want
    :param list[int] slots: list of slots you want
    :param valence.selectors.ActiveNodeSelector node_selector: node to get drives from
    :param bool always_unique: If true never return the same drive twice
    :param valence.selectors.ServiceSelector service_selector: Service Selector to filter nodes.

    :raises ValueError: If always_unique is used without specifying num.
    """

    def __init__(self, cluster_dm, num=None, statuses=None, types=None, slots=None, node_selector=None,
                 always_unique=False, service_selector=None):
        # ApiDrive has logic for which slot is default
        self.cluster_dm = cluster_dm
        self._num = num
        self._statuses = set(statuses) if statuses else set()
        self._types = set(types) if types else set()
        self._slots = set(slots) if slots else set()
        self._node_selector = node_selector
        self._always_unique = always_unique
        self._used_drive_ids = set()
        self._service_selector = service_selector
        if self._always_unique and self._num is None:
            raise ValueError('Must provide num when using always_unique=True')

    def __call__(self):
        """Return drives as specified by the selector.

        :rtype: list[valence.models.Drives]
        :raises valence.selectors.NotEnoughItemsError: if insufficient drives meeting all criteria
             are available to return the requested number of drives.
        """
        cluster_fm = functional_models.Cluster(self.cluster_dm)
        if self._node_selector is not None:
            node_ids = [n.node_id for n in self._node_selector()]
        else:
            node_ids = None
        drives = cluster_fm.list_drives(node_ids=node_ids, slots=self._slots, statuses=self._statuses,
                                        drive_types=self._types)
        if self._always_unique:
            drives = [d for d in drives if d.drive_id not in self._used_drive_ids]
        if self._service_selector:
            service_drive_ids = [s.service.drive_id for s in
                                 self._service_selector() if s.service.drive_id]
            drives = [drive for drive in drives if drive.drive_id in service_drive_ids]
        if self._num:
            random.shuffle(drives)
            drives = drives[:self._num]
            if len(drives) < self._num:
                raise _errors.NotEnoughItemsError(
                    'Insufficient number of drives meeting criteria!  Filtered on {} and found '
                    'drives {}'.format(self.__dict__, drives))
        self._used_drive_ids.update([drive.drive_id for drive in drives])
        return drives


class FixedDriveSelector(DriveSelector):
    """Returns the provided drive, replacement for lambdas.

    :param list[valence.models.Drive] drives: List of valence drive data model.
    """

    # pylint: disable=super-init-not-called
    def __init__(self, drives):
        self._drives = drives

    def __call__(self):
        """Returns the drives provided to init.

        :rtype: list[valence.models.Drive]
        """

        return self._drives
