"""Valence node selectors."""
import abc
import random

import autologger

from valence import (
    criteria,
    enums,
    functional_models,
    models,
    vl_helpers,
)
from valence.selectors import _errors

_log = autologger.get_sflogging()


class ClusteredNodeSelector(object):
    """Selects active nodes from the cluster based on various criteria.

    :param valence.models.ResourceCluster cluster_dm: cluster data model
    :param int num: number of nodes you want to randomly select; if None, all nodes
                   fitting the given criteria are returned
    :param list[sf_models.NodeType] node_models: Filter by node model
    :param list[valence.models.SfVersion] versions: Filter by SolidFire software version
    :param bool pingable: Return only nodes with the node api up. If False, return the opposite.
    :param bool always_unique: Return only unique nodes each time instance is called.
    :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
    :param valence.selectors.BaseNodeSelectorFilter custom_filter: Custom filter to be used
    """
    def __init__(self, cluster_dm, num=None, node_models=None, versions=None, pingable=None,
                 always_unique=False, node_roles=None, custom_filter=None):
        self._num = num
        self._models = node_models
        self._versions = set(versions) if versions else None
        self._pingable = pingable
        self._cluster_fm = functional_models.Cluster(cluster_dm)
        self._always_unique = always_unique
        self._used_node_mips = set()
        self._node_roles = node_roles
        self._custom_filter = custom_filter

        if self._always_unique and self._num is None:
            raise ValueError('Must provide num when using always_unique=True')

    @property
    def cluster_fm(self):
        """Return the cluster FM for this selector.

        :rtype: valence.functional_models.Cluster
        """
        return self._cluster_fm

    @property
    def cluster_dm(self):
        """Return the cluster FM for this selector.

        :rtype: valence.models.ResourceCluster
        """
        return self.cluster_fm.cluster_dm

    def _list_eligible_nodes(self):
        """Return all nodes eligible for selection.

        :rtype: list[valence.models.NodeBase]
        """
        return self._cluster_fm.list_all_nodes()

    def _additional_criteria(self):
        """Subclasses should add additional criteria here.

        :rtype: list[valence.criteria.SingularCriteria]
        """
        return []

    def _apply_run_time_filters(self, nodes):
        """Additional filtering that cannot be included in criteria filters.

        :param list[valence.models.NodeBase] nodes: List of nodes to filter.
        :rtype: list[valence.models.NodeBase]
        """
        if self._pingable is not None:
            nodes = self._filter_pingable(rs_node_dms=nodes, negate=not self._pingable)

        if self._custom_filter:
            nodes = self._custom_filter.filter(nodes)
        return nodes

    def __call__(self):
        """Selects nodes that match the given criteria.

        :rtype: list[valence.models.NodeBase]
        :raises valence.selectors.NotEnoughItemsError: if insufficient nodes meeting all criteria
             are available to return the requested number of nodes.
        """
        if self._node_roles is None:
            self._node_roles = [enums.NodeRole.storage]
        eligible_nodes = self._list_eligible_nodes()
        eligible_nodes = self._apply_run_time_filters(eligible_nodes)
        resource_sips = [n.sip for n in eligible_nodes]
        node_filter = criteria.And(
            criteria.ContainedIn(models.NodeBase.model, self._models),
            criteria.ContainedIn(models.SfNodeBase.software_version, self._versions),
            criteria.ContainedIn(models.SfNodeBase.sip, resource_sips),
            criteria.ContainedIn(models.SfNodeBase.role, self._node_roles),
        )
        # Add specific criteria for subclasses
        for c in self._additional_criteria():
            node_filter.add_criteria(c)
        nodes = node_filter.filter(eligible_nodes)

        if self._always_unique:
            nodes = [n for n in nodes if n.mip not in self._used_node_mips]
        if self._num:
            random.shuffle(nodes)
            nodes = nodes[:self._num]
            if len(nodes) < self._num:
                raise _errors.NotEnoughItemsError(
                    'Insufficient number of nodes meeting criteria!  Filtered on {} and found '
                    'nodes {}'.format(self.__dict__, nodes))

        self._used_node_mips.update([node.mip for node in nodes])

        return nodes

    @staticmethod
    def _filter_pingable(rs_node_dms, negate=False):
        """Return only the nodes with their api up.

        :param list[valence.models.NodeBase] rs_node_dms: resource nodes to filter
        :param bool negate: If True, revert the logic and return nodes that are not pingable.

        :rtype: list[valence.models.NodeBase]
        """
        node_fms = [functional_models.Node(n, cluster_dm=None) for n in rs_node_dms]
        pings = vl_helpers.api_up_on_nodes(node_fms)
        return [n for p, n in zip(pings, rs_node_dms) if p != negate]


class PendingNodeSelector(ClusteredNodeSelector):
    """Selects pending nodes based on various criteria.

    :param valence.models.ResourceCluster cluster_dm: cluster data model
    :param int num: number of nodes you want to randomly select; if None, all nodes
                   fitting the given criteria are returned
    :param list[sf_models.NodeType] node_models: Filter by node model
    :param list[valence.models.SfVersion] versions: Filter by SolidFire software version
    :param bool pingable: Return only nodes with the node api up. If False, return the opposite.
    :param bool always_unique: Return only unique nodes each time instance is called.
    :param list[str] chassis: Return only nodes in these chassis
    :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
    :param valence.selectors.BaseNodeSelectorFilter custom_filter: Custom filter to be used
    """
    def __init__(self, cluster_dm, num=None, node_models=None, versions=None, pingable=None,
                 always_unique=False, chassis=None, node_roles=None, custom_filter=None):
        super(PendingNodeSelector, self).__init__(cluster_dm, num=num, node_models=node_models,
                                                  versions=versions, pingable=pingable,
                                                  always_unique=always_unique,
                                                  node_roles=node_roles,
                                                  custom_filter=custom_filter)
        self._chassis = chassis

    def _additional_criteria(self):
        """Additional filter criteria.

        :rtype: list[valence.criteria.SingularCriteria]
        """
        return [criteria.ContainedIn(models.PendingNode.chassis_name, self._chassis)]

    def _list_eligible_nodes(self):
        """Return all nodes eligible for selection.

        :rtype: list[valence.models.PendingNode]
        """
        return self._cluster_fm.list_pending_nodes()

    def __call__(self):
        """Selects Pending nodes that match the given criteria.

        :rtype: list[valence.models.PendingNode]
        """
        # pylint: disable=useless-super-delegation
        return super(PendingNodeSelector, self).__call__()


class ActiveNodeSelector(ClusteredNodeSelector):
    """Selects active nodes from the cluster based on various criteria.

    :param valence.models.ResourceCluster cluster_dm: cluster data model
    :param int num: number of nodes you want to randomly select; if None, all nodes
                   fitting the given criteria are returned
    :param valence.enums.ZookeeperType zk_type: Filter by zookeeper type
    :param bool cluster_master: return only the cluster master if True, if False return only nodes that
             are not the cluster master.
    :param list[sf_models.NodeType] node_models: Filter by node model
    :param list[valence.models.SfVersion] versions: Filter by SolidFire software version
    :param bool pingable: Return only nodes with the node api up. If False, return the opposite.
    :param list[str] domains: Return only nodes in these protection domains (currently chassisSerial)
    :param bool always_unique: Return only unique nodes each time instance is called.
    :param typing.Iterable[valence.enums.NodeRole] node_roles: Node roles on which to filter.
    :param valence.selectors.ServiceSelector service_selector: limit nodes to nodes associated with
        services returned by service_selector.
    :param valence.selectors.BaseNodeSelectorFilter custom_filter: Custom filter to be used
    """

    def __init__(self, cluster_dm, num=None, zk_type=None, cluster_master=None, node_models=None,
                 versions=None, pingable=None, domains=None, always_unique=False, node_roles=None,
                 service_selector=None, custom_filter=None):
        super(ActiveNodeSelector, self).__init__(cluster_dm, num=num, node_models=node_models,
                                                 versions=versions, pingable=pingable,
                                                 always_unique=always_unique,
                                                 node_roles=node_roles,
                                                 custom_filter=custom_filter)
        self._zk_type = zk_type
        self._cluster_master = cluster_master
        self._cluster_fm = functional_models.Cluster(cluster_dm)
        self._domains = domains
        self._service_selector = service_selector

    def _list_eligible_nodes(self):
        """Return all nodes eligible for selection.

        :rtype: list[valence.models.ActiveNode]
        """
        return self._cluster_fm.list_active_nodes()

    def _additional_criteria(self):
        """Additional filter criteria.

        :rtype: list[valence.criteria.SingularCriteria]
        """
        if self._cluster_master is not None:
            master_sip = self._cluster_fm.get_cluster_master().sip
        else:
            master_sip = None
        criteria_list = [
                criteria.Equals(models.ActiveNode.sip, master_sip, negate=not self._cluster_master),
                criteria.Or(
                    criteria.ContainedIn(models.ActiveNode.chassis_name, self._domains),
                    criteria.ContainedIn(models.ActiveNode.custom_protection_domain_name, self._domains),
                    )
            ]
        # if a service selector was given restrict the return to those that match the given services
        if self._service_selector:
            node_ids = [s.node.node_id for s in self._service_selector()]
            criteria_list.append(criteria.And(criteria.ContainedIn(models.ActiveNode.node_id, node_ids)))
        return criteria_list

    def _apply_run_time_filters(self, nodes):
        """Additional filtering that cannot be included in criteria filters.

        :param list[valence.models.ActiveNode] nodes: List of nodes to filter.
        :rtype: list[valence.models.ActiveNode]
        """
        nodes = super(ActiveNodeSelector, self)._apply_run_time_filters(nodes)
        if self._zk_type:
            nodes = self._filter_zk_type(rs_node_dms=nodes)
        return nodes

    def __call__(self):
        """Selects active nodes from the cluster.

        :rtype: list[valence.models.ActiveNode]
        :raises valence.selectors.NotEnoughItemsError: if insufficient nodes meeting all criteria
             are available to return the requested number of nodes.
        """
        # pylint: disable=useless-super-delegation
        return super(ActiveNodeSelector, self).__call__()

    def _filter_zk_type(self, rs_node_dms):
        """Filter nodes by zookeeper type.

        :param list[valence.models.NodeBase] rs_node_dms: resource nodes to filter

        :rtype: list[valence.models.NodeBase]
        """
        if self._zk_type is enums.ZookeeperType.ensemble:
            ensemble = self._cluster_fm.get_cluster_info().ensemble
            return [n for n in rs_node_dms if
                    n.sip in ensemble]  # pylint: disable=unsupported-membership-test
        elif self._zk_type:
            node_fms = [functional_models.Node(n, cluster_dm=self._cluster_fm.cluster_dm)
                        for n in rs_node_dms]
            zk_type_map = vl_helpers.get_ensemble_config_on_nodes(node_fms)
            zk_type_map_with_non_ensemble = [t if t else enums.ZookeeperType.non_ensemble
                                             for t in zk_type_map]
            # Looks like criteria requires an attribute to use.
            # Zookeeper type should not be an attribute since it can change at any time
            return [n for i, n in enumerate(rs_node_dms) if
                    zk_type_map_with_non_ensemble[i] is self._zk_type]


class FixedNodeSelector(ActiveNodeSelector, PendingNodeSelector):
    """Returns the provided node_dms, replacement for lambdas.

    :param list[valence.models.NodeBase] node_dms: List of valence node data models.
    """
    # pylint: disable=super-init-not-called
    def __init__(self, node_dms):
        self._node_dms = node_dms

    def __call__(self):
        """Returns the nodes provided to init.

        :rtype: list[valence.models.NodeBase]
        """
        return self._node_dms


class BaseNodeSelectorFilter(object):
    @abc.abstractmethod
    def filter(self, node_dms):
        """Implement custom filtering logic that doesn't have shared value.

        :param list[valence.models.NodeBase] node_dms: List of valence node data models.
        :rtype: list[valence.models.NodeBase]
        :return: list[valence.models.NodeBase]
        """
