"""Valence service selectors."""
import random
from collections import namedtuple

from valence import functional_models
from valence.selectors import _errors

services_tuple = namedtuple('services_tuple', ('service_object',
                                               'service_type',
                                               'service_status',
                                               'node_ids',
                                               'drive_ids'))


class ServiceSelector(object):
    """Selects services from a cluster based on various criteria.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param list[valence.enums.NodeService] service_types: list of service types to filter from
    :param list[valence.enums.ServiceStatus] service_statuses: Optional, service statuses to filter on
    :param valence.selectors.ActiveNodeSelector node_selector: node to get services from
    :param valence.selectors.DriveSelector drive_selector: drives to filter
    :param int num: number of services you want to select randomly, returns all services if None


    """
    def __init__(self, cluster_dm, service_types=None, service_statuses=None,
                 node_selector=None, drive_selector=None, num=None):
        self._cluster_dm = cluster_dm
        self._node_selector = node_selector
        self._service_types = service_types
        self._drive_selector = drive_selector
        self._num = num
        self._service_statuses = service_statuses

    def __call__(self):
        """Return services as specified by the selector.

        :rtype: list[valence.models.ClusterServiceInfo]
        :raises valence.selectors.NotEnoughItemsError: if insufficient services meeting all criteria
             are available to return the requested number of services.
        """
        cluster_fm = functional_models.Cluster(self._cluster_dm)
        services = cluster_fm.list_services()
        # named_service_list has the service object, type of service,
        # sets of node ids and drive ids of the service object.
        service_tuples = []
        for s in services:
            service_tuples.append(services_tuple(s,
                                                 s.service.service_type.value,
                                                 s.service.status.value,
                                                 {s.node.node_id},
                                                 {d.drive_id for d in s.drives if s.drives != []}))

        if self._service_types is not None:
            service_tuples = (i for i in service_tuples
                              if i.service_type in (t.value for t in self._service_types))
        if self._service_statuses is not None:
            service_tuples = (i for i in service_tuples
                              if i.service_status in (t.value for t in self._service_statuses))
        if self._node_selector is not None:
            node_ids = {n.node_id for n in self._node_selector()}
            service_tuples = (i for i in service_tuples
                              if i.node_ids.issubset(node_ids))
        if self._drive_selector is not None:
            drive_ids = {d.drive_id for d in self._drive_selector()}
            service_tuples = (i for i in service_tuples
                              if i.drive_ids.issubset(drive_ids))

        service_tuples = list(service_tuples)
        if self._num:
            random.shuffle(service_tuples)
            service_tuples = service_tuples[:self._num]
            if len(service_tuples) < self._num:
                raise _errors.NotEnoughItemsError(
                    'Insufficient number of services meeting criteria!  Filtered on {} and found '
                    'services {}'.format(self.__dict__, services)
                )

        return [i.service_object for i in service_tuples]


class FixedServiceSelector(ServiceSelector):
    """Returns the provided service, replacement for lambdas.

    :param list[valence.models.ClusterServiceInfo] services: List of valence ClusterServiceInfo model.
    """

    # pylint: disable=super-init-not-called
    def __init__(self, services):
        self._services = services

    def __call__(self):
        """Returns the service provided to init.

        :rtype: list[valence.models.ClusterServiceInfo]
        """

        return self._services
