import unittest

import mock
import pytest

from sf_api import result_models
from valence import (
    models,
    selectors,
)


@mock.patch(
    'valence.functional_models.Cluster.get_cluster_full_threshold',
    mock.MagicMock(return_value=result_models.GetClusterFullThreshold.create(
        sum_used_cluster_bytes=10,
        sum_total_cluster_bytes=1000,
        stage2_block_threshold_bytes=200,
        stage3_block_threshold_bytes=700,
        stage4_block_threshold_bytes=800,
        stage5_block_threshold_bytes=900,
        stage2_metadata_threshold_bytes=200,
        stage3_metadata_threshold_bytes=700,
        stage4_metadata_threshold_bytes=800,
        stage5_metadata_threshold_bytes=900,
        sum_used_metadata_cluster_bytes=3,
        sum_total_metadata_cluster_bytes=200
    ))
)
class TestBlockCapacitySelectors(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def test_fixed_capacity_selector(self):
        capacity_selector = selectors.FixedBlockCapacitySelector(self.cluster_dm, 200)
        assert not capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 200

        # Test relative capacity
        capacity_selector = selectors.FixedBlockCapacitySelector(
            cluster_dm=self.cluster_dm,
            cluster_bytes=200,
            relative_to_used_capacity=True
        )
        assert capacity_selector.relative

        capacity = capacity_selector()
        assert capacity.bytes == 190

    def test_percent_capacity_selector(self):
        capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 100)
        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 101)

        capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, -100)
        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, -101)

        capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 20)
        assert not capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 200

        # Test relative capacity
        capacity_selector = selectors.PercentBlockCapacitySelector(
            cluster_dm=self.cluster_dm,
            percent_capacity=20,
            relative_to_used_capacity=True
        )
        assert capacity_selector.relative

        capacity = capacity_selector()
        assert capacity.bytes == 190

    def test_stage_capacity_selector(self):
        capacity_selector = selectors.StageBlockCapacitySelector(self.cluster_dm, 2)
        assert capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 190

        # Test into-stage-percent
        capacity_selector = selectors.StageBlockCapacitySelector(
            cluster_dm=self.cluster_dm,
            stage=3,
            into_stage_percent=20
        )
        assert capacity_selector.relative
        assert capacity_selector().bytes == 710

        # Test other stages
        capacity_selector = selectors.StageBlockCapacitySelector(self.cluster_dm, 4)
        assert capacity_selector().bytes == 790
        capacity_selector = selectors.StageBlockCapacitySelector(self.cluster_dm, 5)
        assert capacity_selector().bytes == 890

        # Exception cases
        with pytest.raises(AssertionError):
            capacity_selector = selectors.StageBlockCapacitySelector(
                cluster_dm=self.cluster_dm,
                stage=5,
                into_stage_percent=1
            )

        with pytest.raises(AssertionError):
            capacity_selector = selectors.StageBlockCapacitySelector(
                cluster_dm=self.cluster_dm,
                stage=6
            )

    def test_percent_stage_capacity_selector(self):
        capacity_selector = selectors.PercentStageBlockCapacitySelector(self.cluster_dm, 2, 40)
        assert not capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 200

        # Test other stages
        capacity_selector = selectors.PercentStageBlockCapacitySelector(self.cluster_dm, 3, 10)
        assert capacity_selector().bytes == 10
        capacity_selector = selectors.PercentStageBlockCapacitySelector(self.cluster_dm, 4, 20)
        assert capacity_selector().bytes == 20

        # Exception cases
        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentStageBlockCapacitySelector(self.cluster_dm, 5, 10)

        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentStageBlockCapacitySelector(self.cluster_dm, 4, 101)

    def test_percent_stage_meta_capacity_selector(self):
        capacity_selector = selectors.PercentStageMetaCapacitySelector(self.cluster_dm, 2, 40)
        assert not capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 200

        # Test other stages
        capacity_selector = selectors.PercentStageMetaCapacitySelector(self.cluster_dm, 3, 10)
        assert capacity_selector().bytes == 10
        capacity_selector = selectors.PercentStageMetaCapacitySelector(self.cluster_dm, 4, 20)
        assert capacity_selector().bytes == 20

        # Exception cases
        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentStageMetaCapacitySelector(self.cluster_dm, 5, 10)

        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentStageMetaCapacitySelector(self.cluster_dm, 4, 101)

    def test_add(self):
        capacity_selector = selectors.StageBlockCapacitySelector(self.cluster_dm, 3)
        capacity_selector.add(selectors.PercentStageBlockCapacitySelector(self.cluster_dm, 2, -1))
        capacity = capacity_selector()
        assert capacity.bytes == 685

        capacity_selector.add(selectors.FixedBlockCapacitySelector(self.cluster_dm, 3))
        capacity = capacity_selector()
        assert capacity.bytes == 688


@mock.patch(
    'valence.functional_models.Cluster.get_cluster_full_threshold',
    mock.MagicMock(return_value=result_models.GetClusterFullThreshold.create(
        sum_used_cluster_bytes=10,
        sum_total_cluster_bytes=1000,
        stage2_block_threshold_bytes=200,
        stage3_block_threshold_bytes=700,
        stage4_block_threshold_bytes=800,
        stage5_block_threshold_bytes=900,
        stage2_metadata_threshold_bytes=200,
        stage3_metadata_threshold_bytes=700,
        stage4_metadata_threshold_bytes=800,
        stage5_metadata_threshold_bytes=900,
        sum_used_metadata_cluster_bytes=3,
        sum_total_metadata_cluster_bytes=200
    ))
)
class TestMetaCapacitySelectors(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def test_fixed_capacity_selector(self):
        capacity_selector = selectors.FixedMetaCapacitySelector(self.cluster_dm, 20)
        assert not capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 20

        # Test relative capacity
        capacity_selector = selectors.FixedMetaCapacitySelector(
            cluster_dm=self.cluster_dm,
            cluster_bytes=20,
            relative_to_used_capacity=True
        )
        assert capacity_selector.relative

        capacity = capacity_selector()
        assert capacity.bytes == 17

    def test_percent_capacity_selector(self):
        capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 100)
        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 101)

        capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, -100)
        with pytest.raises(AssertionError):
            capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, -101)

        capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 20)
        assert not capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 40

        # Test relative capacity
        capacity_selector = selectors.PercentMetaCapacitySelector(
            cluster_dm=self.cluster_dm,
            percent_capacity=20,
            relative_to_used_capacity=True
        )
        assert capacity_selector.relative

        capacity = capacity_selector()
        assert capacity.bytes == 37

    def test_add(self):
        capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 20)
        capacity_selector.add(selectors.FixedMetaCapacitySelector(self.cluster_dm, -13))
        capacity = capacity_selector()
        assert capacity.bytes == 27

        with pytest.raises(AssertionError):
            capacity_selector.add(selectors.PercentMetaCapacitySelector(
                cluster_dm=self.cluster_dm,
                percent_capacity=20,
                relative_to_used_capacity=True
            ))

    def test_stage_capacity_selector(self):
        capacity_selector = selectors.StageMetadataCapacitySelector(self.cluster_dm, 2)
        assert capacity_selector.relative

        capacity = capacity_selector()
        assert isinstance(capacity, models.DataSize)
        assert capacity.bytes == 197

        # Test into-stage-percent
        capacity_selector = selectors.StageMetadataCapacitySelector(
            cluster_dm=self.cluster_dm,
            stage=3,
            into_stage_percent=20
        )
        assert capacity_selector.relative
        assert capacity_selector().bytes == 717

        # Test other stages
        capacity_selector = selectors.StageMetadataCapacitySelector(self.cluster_dm, 4)
        assert capacity_selector().bytes == 797
        capacity_selector = selectors.StageMetadataCapacitySelector(self.cluster_dm, 5)
        assert capacity_selector().bytes == 897

        # Exception cases
        with pytest.raises(AssertionError):
            capacity_selector = selectors.StageMetadataCapacitySelector(
                cluster_dm=self.cluster_dm,
                stage=5,
                into_stage_percent=1
            )

        with pytest.raises(AssertionError):
            capacity_selector = selectors.StageMetadataCapacitySelector(
                cluster_dm=self.cluster_dm,
                stage=6
            )
