import unittest

import mock
import pytest

from valence import (
    models,
    selectors,
)


@mock.patch('valence.functional_models.Cluster.list_drives')
class TestDriveSelector(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.drive_dms = models.Drive.get_mock_object_list(5)

    def test_works(self, mock_list_drives):
        mock_list_drives.return_value = self.drive_dms
        selector = selectors.DriveSelector(self.cluster_dm)
        drives = selector()
        assert drives is self.drive_dms

    def test_works_some(self, mock_list_drives):
        mock_list_drives.return_value = self.drive_dms
        selector = selectors.DriveSelector(self.cluster_dm, 3)
        drives = selector()
        assert len(drives) == 3

    def test_node_selector(self, mock_list_drives):
        nodes = models.ActiveNode.get_mock_object_list(1)
        mock_list_drives.return_value = self.drive_dms
        selector = selectors.DriveSelector(self.cluster_dm, node_selector=lambda: nodes)
        drives = selector()
        mock_list_drives.assert_called_with(
            node_ids=[nodes[0].node_id],
            slots=set(),
            statuses=set(),
            drive_types=set()
        )

    def test_too_few(self, mock_list_drives):
        nodes = models.ActiveNode.get_mock_object_list(1)
        mock_list_drives.return_value = self.drive_dms
        selector = selectors.DriveSelector(self.cluster_dm, node_selector=lambda: nodes,
                                           num=len(self.drive_dms) + 1)
        with pytest.raises(selectors.NotEnoughItemsError):
            selector()

    def test_always_unique(self, mock_list_drives):
        mock_list_drives.return_value = self.drive_dms
        selector = selectors.DriveSelector(self.cluster_dm, num=2, always_unique=True)
        first_drive_ids = [d.drive_id for d in selector()]
        second_drive_ids = [d.drive_id for d in selector()]
        assert not set(first_drive_ids).intersection(second_drive_ids)

    def test_always_unique_requires_num(self, mock_list_drives):
        with pytest.raises(ValueError):
            selector = selectors.DriveSelector(self.cluster_dm, always_unique=True)

    def test_fixed_selector_returns_list(self, mock_list_drives):
        mock_list_drives.return_value = self.drive_dms
        selector = selectors.DriveSelector(self.cluster_dm, num=4)
        drives = selector()
        fixed_selector = selectors.FixedDriveSelector(drives)
        fixed_drives = fixed_selector()
        assert drives is fixed_drives

    def test_drive_selector_with_service_selector(self, mock_list_drives):
        mock_list_drives.return_value = self.drive_dms
        selected_drives = self.drive_dms[:3]
        services = [models.ClusterService.get_mock_object(overrides=dict(drive_id=drive.drive_id))
                    for drive in selected_drives]
        service_infos = [models.ClusterServiceInfo.get_mock_object(overrides=dict(service=service))
                         for service in services]
        drive_selector = selectors.DriveSelector(self.cluster_dm, service_selector=lambda: service_infos)
        drive_dms = drive_selector()

        assert len(drive_dms) == len(selected_drives)
        for drive in drive_dms:
            assert drive in selected_drives
