from __future__ import print_function

import unittest

import mock
import pytest
import sf_models

import sf_api
from sf_api import enums
from valence import (
    models,
    selectors,
)


class TestNodeSelector(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.Cluster.get_mock_object()
        overrides = dict(internal_role='Storage')
        cls.node_dms = [models.ActiveNode.get_mock_object(overrides=overrides),
                        models.ActiveNode.get_mock_object(overrides=overrides),
                        models.PendingActiveNode.get_mock_object(overrides=overrides),
                        models.PendingActiveNode.get_mock_object(overrides=overrides),
                        models.PendingNode.get_mock_object(overrides=overrides),
                        models.PendingNode.get_mock_object(overrides=overrides)]
        for node in cls.node_dms:
            node.platform_info.node_type = sf_models.NodeType.STANDARD_TYPES.SF9010

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    def test_simple(self, mock_list_nodes):
        mock_list_nodes.return_value = self.node_dms
        selector = selectors.ActiveNodeSelector(self.cluster_dm)
        nodes = selector()
        assert nodes == self.node_dms

    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_num(self, mock_list_nodes):
        mock_list_nodes.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, num=1)
        nodes = selector()
        assert len(nodes) == 1
        assert selector.cluster_fm == selector._cluster_fm
        assert selector.cluster_dm == self.cluster_dm

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    @mock.patch('valence.functional_models.Cluster.get_cluster_info')
    def test_status(self, mock_get_info, mock_list_nodes):
        mock_list_nodes.return_value = [n for n in self.node_dms if type(n) is models.ActiveNode]
        cluster_info = sf_api.result_models.GetClusterInfo.get_mock_object()
        cluster_info.ensemble = [n['sip'] for n in self.node_dms]
        mock_get_info.return_value = cluster_info
        selector = selectors.ActiveNodeSelector(self.cluster_dm,
                                                zk_type=enums.ZookeeperType.ensemble)
        nodes = selector()
        assert len(nodes) == 2

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    @mock.patch('valence.vl_helpers.get_ensemble_config_on_nodes')
    def test_leader(self, mock_get_ens, mock_list_nodes):
        mock_list_nodes.return_value = [n for n in self.node_dms if type(n) is models.ActiveNode]
        mock_get_ens.return_value = [enums.ZookeeperType.leader, None]
        selector = selectors.ActiveNodeSelector(self.cluster_dm,
                                                zk_type=enums.ZookeeperType.leader)
        assert len(selector()) == 1

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    @mock.patch('valence.functional_models.Cluster.get_cluster_master')
    def test_master(self, mock_get_master, mock_list):
        # It seems like this should be the case by default when mocking. Just a thought.
        active_node_dms = [n for n in self.node_dms if type(n) is models.ActiveNode]
        mock_master = models.ActiveNode.get_mock_object()
        mock_master.sip = active_node_dms[0].sip
        mock_get_master.return_value = mock_master
        mock_list.return_value = active_node_dms
        selector = selectors.ActiveNodeSelector(self.cluster_dm, cluster_master=True)
        nodes = selector()
        assert len(nodes) == 1
        selector = selectors.ActiveNodeSelector(self.cluster_dm, cluster_master=False)
        nodes = selector()
        assert len(nodes) == len(active_node_dms) - 1

    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_models(self, mock_list):
        for n in self.node_dms:
            n.resource_dm.model = sf_models.NodeType.STANDARD_TYPES.SF3010
        mock_list.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm,
                                                   node_models=[
                                                       sf_models.NodeType.STANDARD_TYPES.SF3010])
        nodes = selector()
        assert len(nodes) == len(self.node_dms)
        selector = selectors.ClusteredNodeSelector(self.cluster_dm,
                                                   node_models=[
                                                       sf_models.NodeType.STANDARD_TYPES.SF2405])
        nodes = selector()
        assert len(nodes) == 0

    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_version(self, mock_list):
        test_version = models.SfVersion('10.3.0.114')
        for n in self.node_dms:
            n.software_version = test_version
        mock_list.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, versions=[test_version])
        nodes = selector()
        assert len(nodes) == len(self.node_dms)

    @mock.patch('valence.vl_helpers.api_up_on_nodes')
    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_pingable_is_true(self, mock_list, mock_ping):
        mock_ping.return_value = [True, False, True, False, True, True]
        mock_list.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, pingable=True)
        nodes = selector()
        assert len(nodes) == 4

    @mock.patch('valence.vl_helpers.api_up_on_nodes')
    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_pingable_is_false(self, mock_list, mock_ping):
        mock_ping.return_value = [True, False, True, False, True, True]
        mock_list.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, pingable=False)
        nodes = selector()
        assert len(nodes) == 2

    @mock.patch('valence.vl_helpers.api_up_on_nodes')
    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_too_few(self, mock_list, mock_ping):
        mock_ping.return_value = [True, False, True, False, True, False]
        mock_list.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, num=4, pingable=True)
        with pytest.raises(selectors.NotEnoughItemsError):
            selector()

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    def test_domains(self, mock_list):
        for n in self.node_dms[:2]:
            n.chassis_name = 'Testing'
        for n in self.node_dms[2:]:
            n.chassis_name = 'do not select'
        mock_list.return_value = [n for n in self.node_dms if type(n) is models.ActiveNode]
        selector = selectors.ActiveNodeSelector(self.cluster_dm, domains=['Testing', 'b'])
        nodes = selector()
        assert len(nodes) == 2

    @mock.patch('valence.functional_models.Cluster.list_pending_nodes')
    def test_pending(self, mock_list):
        mock_list.return_value = [n for n in self.node_dms if type(n) is models.PendingNode]
        selector = selectors.PendingNodeSelector(self.cluster_dm, num=1)
        assert len(selector()) == 1

    def test_fixed(self):
        fixed_nodes = [self.node_dms[0], self.node_dms[3]]
        selector = selectors.FixedNodeSelector(fixed_nodes)
        assert selector() == fixed_nodes

    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_always_unique_requires_num(self, mock_list_nodes):
        with pytest.raises(ValueError):
            selector = selectors.ClusteredNodeSelector(self.cluster_dm, always_unique=True)

    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_always_unique(self, mock_list_nodes):
        mock_list_nodes.return_value = self.node_dms
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, num=2, always_unique=True)
        first_node_mips = [n.mip for n in selector()]
        second_node_mips = [n.mip for n in selector()]
        assert not set(first_node_mips).intersection(second_node_mips)

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    def test_include_witness(self, mock_list_nodes):
        mock_witness = models.ActiveNode.get_mock_object(overrides=dict(internal_role='Witness'))
        mock_witness.platform_info.node_type = sf_models.NodeType.STANDARD_TYPES.SF9010
        mock_list_nodes.return_value = [mock_witness]
        selector = selectors.ActiveNodeSelector(self.cluster_dm, node_roles=[enums.NodeRole.witness])
        nodes = selector()
        assert nodes[0] == mock_witness

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    def test_active_node_with_service_selector(self, mock_list_active_nodes):
        active_nodes = self.node_dms[:2]
        mock_list_active_nodes.return_value = active_nodes
        services_node_tuples = [
            (models.ClusterService.get_mock_object(overrides=dict(node_id=node.node_id)), node)
            for node in active_nodes]
        service_infos = [models.ClusterServiceInfo.get_mock_object(overrides=dict(service=service,
                                                                                  node=node))
                         for service, node in services_node_tuples]
        node_selector = selectors.ActiveNodeSelector(self.cluster_dm,
                                                     service_selector=lambda: service_infos)
        node_dms = node_selector()
        assert len(node_dms) == len(active_nodes)
        for drive in node_dms:
            assert drive in active_nodes

    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    def test_active_node_with_service_selector(self, mock_list_active_nodes):
        active_nodes = []
        mock_list_active_nodes.return_value = active_nodes
        services_node_tuples = [
            (models.ClusterService.get_mock_object(overrides=dict(node_id=node.node_id)), node)
            for node in active_nodes]
        service_infos = [models.ClusterServiceInfo.get_mock_object(overrides=dict(service=service,
                                                                                  node=node))
                         for service, node in services_node_tuples]
        node_selector = selectors.ActiveNodeSelector(self.cluster_dm,
                                                     service_selector=lambda: service_infos)
        node_dms = node_selector()
        assert node_dms == []

    @mock.patch('valence.functional_models.Cluster.list_all_nodes')
    def test_custom_filter(self, mock_list_nodes):
        mock_list_nodes.return_value = self.node_dms

        class TritonNvdimmFilterLowerFirmware(selectors.BaseNodeSelectorFilter):
            def filter(self, node_dms):
                return [n for n in node_dms if isinstance(n, models.ActiveNode)]

        custom_filter = TritonNvdimmFilterLowerFirmware()
        selector = selectors.ClusteredNodeSelector(self.cluster_dm, custom_filter=custom_filter)
        nodes_selected = selector()
        assert len(nodes_selected) == 2
