import unittest

import mock
import pytest

from valence import (
    enums,
    functional_models,
    models,
    selectors,
)


class TestServiceSelector(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)
        cls.node_dms = models.ActiveNode.get_mock_object_list(3)
        cls.drive_dms = models.ClusterServiceDrive.get_mock_object_list(3)
        cls.cluster_services = models.ClusterService.get_mock_object_list(3, overrides={
            'service_type': enums.ServiceType.block,
            'status': enums.ServiceStatus.healthy})

    def setUp(self):
        self.service_models = []
        for i in range(0, 3):
            self.service_models.append(models.ClusterServiceInfo.get_mock_object(
                overrides={'node': self.node_dms[i],
                           'drives': [self.drive_dms[i]],
                           'service': self.cluster_services[i]}))

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_basic_functionality(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        selector = selectors.ServiceSelector(self.cluster_dm)
        services = selector()
        assert services == self.service_models

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_num(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        selector = selectors.ServiceSelector(self.cluster_dm, num=2)
        services = selector()
        assert len(services) == 2

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_node_selector(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        node_ids = [n.node_id for n in self.node_dms]
        selector = selectors.ServiceSelector(self.cluster_dm, node_selector=lambda: self.node_dms[:1])
        services = selector()
        for s in services:
            if s.node.node_id not in node_ids:
                assert False
        assert True

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_drive_selector(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        drive_ids = [d.drive_id for d in self.drive_dms]
        selector = selectors.ServiceSelector(self.cluster_dm, drive_selector=lambda: self.drive_dms)
        services = selector()
        for s in services:
            service_drive_ids = [d.drive_id for d in s.drives]
            if not set(service_drive_ids).issubset(set(drive_ids)):
                assert False
        assert True

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_service_types(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        selector = selectors.ServiceSelector(self.cluster_dm, service_types=[enums.NodeService.block])
        services = selector()
        assert self.service_models == services

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_statuses(self, mock_list_services):
        mock_list_services.return_value = self.service_models

        selector = selectors.ServiceSelector(self.cluster_dm,
                                             service_statuses=[enums.ServiceStatus.healthy])
        services = selector()
        assert self.service_models == services

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_statuses_no_match(self, mock_list_services):
        mock_list_services.return_value = self.service_models

        selector = selectors.ServiceSelector(self.cluster_dm,
                                             service_statuses=[enums.ServiceStatus.migrating])
        services = selector()
        assert services == []

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_with_num_fail(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        selector = selectors.ServiceSelector(self.cluster_dm, num=len(self.service_models) + 1)
        with pytest.raises(selectors.NotEnoughItemsError):
            selector()

    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_fixed_service_selector(self, mock_list_services):
        mock_list_services.return_value = self.service_models
        selector = selectors.ServiceSelector(self.cluster_dm, num=3)
        services = selector()
        fixed_selector = selectors.FixedServiceSelector(services)
        fixed_services = fixed_selector()
        assert services is fixed_services
