"""Contains a test case that passes an assert."""
import types  # pylint: disable=unused-import

from valence import (
    camelot,
    enums,
)


class PassingCase(camelot.TestCase):
    """Test case that passes an assert."""

    def _test(self):
        """Test method."""
        self._log_step('Assert {}'.format('True'))
        assert True
        self._log_step('Assert 1 is an int', 'Pass')
        assert isinstance(1, int)


class PassingElementCase(camelot.ElementTestCase):
    """Test case that passes an assert."""

    def _test(self):
        """Test method."""
        self._log_step('Assert {}'.format('True'))
        assert True
        self._log_step('Assert 1 is an int', 'Pass')
        assert isinstance(1, int)


class PassingElementCaseWithAttributes(camelot.ElementTestCase):
    """Test case that passes an assert."""

    def _test(self):
        """Test method."""
        self._log_step('Assert {}'.format('True'))
        assert True
        self._log_step('Assert 1 is an int', 'Pass')
        assert isinstance(1, int)

    @property
    def monitor_fail_on_faults(self):
        return [enums.ClusterFaultCode.block_cluster_full]

    @property
    def monitor_fail_on_events(self):
        return [enums.ClusterEventType.bs_kill_event]

    @property
    def monitor_ignore_faults(self):
        return [enums.ClusterFaultCode.mtu_check_failure, enums.ClusterFaultCode.block_cluster_full]

    @property
    def monitor_ignore_events(self):
        return [enums.ClusterEventType.ie_event, enums.ClusterEventType.bs_kill_event]


class PassingCaseNoSteps(camelot.TestCase):
    """Test case without any steps."""
    def _test(self):
        """Test method."""
        assert True
        assert isinstance(1, int)


class PassingLogStepCase(camelot.TestCase):
    """Tests different ways to call log_step using static variables."""
    NUMBER = 1
    REASON = 'I want'

    def __init__(self):
        super(PassingLogStepCase, self).__init__()
        self.reason = 'I said so'

    def _test(self):
        """Test method."""
        self._log_step('Assert {}'.format('True'))
        assert True
        self._log_step('Assert {} is an int because {}'.format(self.NUMBER, self.REASON), 'Pass')
        assert isinstance(1, int)
        self._log_step('Assert {n} is an int because {r}'.format(n=self.NUMBER, r=self.reason), 'Pass')
        assert isinstance(1, int)
        self._log_step('Assert unknown type is {unknown}'.format(unknown=object()))
        assert object is object


class InheritedCase(PassingCase):
    """A case that inherits from PassingCase to test that inspection finds _test in the parent."""


class MixinCase(object):
    """A mixin class for a test case."""

    def special_method(self):
        """Test method.

        :rtype: str
        """
        return self.__class__.__name__


class InheritedCaseWithMixin(InheritedCase, MixinCase):
    """A case with multiple inheritance."""


class PassingCaseWithAssertions(camelot.TestCase):
    """A case with posttestassertions."""

    def _test(self):
        """Test method."""
        self._log_step('Assert {} is True. Truely'.format('True'))
        assert True

    def _post_test_assertions(self):
        """Test is yo."""
        assert True


class LogStepMultipleInheritance(PassingCase, PassingCaseWithAssertions):
    """A test case that inherits from two passing cases."""


class SkippedCase(camelot.TestCase):
    """Skip this test case based on preconditions or setup."""
    def _skip(self):
        """Oh crap, we don't want to do this."""
        return True

    def _test(self):
        """Fail gracefully."""
        assert False, 'We should have never come here.'

    def _teardown(self):
        """Break it if you call it."""
        assert False, 'What did you break?'
