"""Assorted unit test utility functions."""
import inspect
import json
import os
import unittest

import mock

import sf_api

CWD = os.path.split(os.path.abspath(__file__))[0]
DATA_DIR = os.path.join(CWD, 'data')
VCS_ROOT = os.path.split(os.path.split(CWD)[0])[0]


def patch_sleep():
    """Decorator to patch time.sleep and replace with a noop.

    When using this decorator, no arguments are passed to the unit test method.
    Be careful using this when testing multithreaded code that relies on sleeps for
    context switching.

    Usage::

        @unittest_utils.patch_sleep()
        @mock.patch('foo')
        def test_of_method_with_long_sleep(self, mock_foo):
            method_under_test()
            mock_foo.assert_called()
    """
    def new_sleep(_):
        pass
    return mock.patch('time.sleep', new=new_sleep)


def get_data_dir(file_path):
    """Return the data directory of a unit test.

    Pass __file__ to this from your unit test.

    :param str file_path: The file path of the unit test
    :rtype: str
    """
    return os.path.join(os.path.dirname(os.path.realpath(file_path)), 'data')


def generate_at2_task_instance_dict():
    """Generate an At2 task instance dict.

    :rtype: dict
    """
    with open(os.path.join(DATA_DIR, 'task.json')) as f:
        task = json.load(f)
    return task


def check_params_serializable(method, params=None,  # pylint:disable=unused-argument
                              result_dm_class=None, uri=None,  # pylint:disable=unused-argument
                              log_response=True, auth=None):  # pylint:disable=unused-argument
    """Check call method params.

    This function acts as a side effect for mocking call_method, it checks that params are
    serializable.

    :param sf_api.ClusterMethod|sf_api.NodeMethod method: method to call
    :param dict params: the params for the method. Enums will be converted to strings
    :param Any result_dm_class: The API Response class the result is modeled after. If there is no
        result data model associated with the call, this can be None.
    :param str uri: Override self.uri with a specific uri
    :param sf_api.SfAuth auth: Override auth with custom object.
    :param bool log_response: If true, log the full JSON response from the API
    :rtype: mock.DEFAULT
    """
    if params:
        params = sf_api.BaseJsonRpcClient._construct_params(params)  # pylint:disable=protected-access
        try:
            params = json.dumps(params)
        except TypeError:
            raise TypeError('params passed to call_method are not json serializable, params: {}'.format(
                params))

    return mock.DEFAULT


class ModuleLevelMock(object):
    """Mock multiple functions/classes/etc and restore when finished.

    Used for mocking things in unit tests with setUpModule and tearDownModule.

    Usage
    =====

    module_mock = unittest_utils.ModuleLevelMock()

    def setUpModule():
        module_mock.mock(my_obj, 'my_attribute')
        module_mock.mock(my_other_obj, 'my_attribute', mock.MagicMock(return_value=100))

    def tearDownModule():
        module_mock.restore()
    """
    def __init__(self):
        self.original_objects = []

    def mock(self, obj, attribute_name, mock_obj=None):
        """Mock given object.

        :param Any obj: object where the mock will be done
        :param str attribute_name: Name of the attribute to mock
        :param mock.Mock|mock.MagicMock mock_obj: Optional mock object. Can be used to set a desired
            return value or side effect. If not given, use a standard MagicMock object that returns None.
        """
        orig_attr = getattr(obj, attribute_name)
        self.original_objects.append((obj, attribute_name, orig_attr))
        if mock_obj is None:
            mock_obj = mock.MagicMock()
        setattr(obj, attribute_name, mock_obj)

    def restore(self):
        """Restore original objects."""
        for obj, attr_name, original_attr in self.original_objects:
            setattr(obj, attr_name, original_attr)


class MockHistoryUnitTest(unittest.TestCase):
    """A unit test class that will mock inspect.stack for all test cases.

    This is used to speed up execution for tests that don't need to test the camelot History object.
    """
    _dummy_stack = inspect.stack()

    def setUp(self):
        """Mock inspect.stack."""
        self._original_stack_fn = inspect.stack
        inspect.stack = mock.Mock(return_value=self._dummy_stack)

    def tearDown(self):
        """Reset inspect.stack to original function."""
        inspect.stack = self._original_stack_fn
