"""Valence Account components."""
from __future__ import division

import abc

import autologger
import six

from valence import (
    components,
    functional_models,
    models,
    vl_helpers,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class AccountsInterface(components.Reversible, components.BlockingApply):
    """Interface component for a set of accounts."""

    @property
    @abc.abstractmethod
    def account_dms(self):
        """Get the cached account data models.

        :rtype: list[valence.models.Account]
        """

    @abc.abstractmethod
    def get_account_dms(self):
        """Get the account models created by the component.

        This will make an actual API call. Use `account_dms` for the cached data models.

        :rtype: tuple[valence.models.Account]
        :raises valence.components.ComponentStateError: If component has not been applied
        """

    @abc.abstractmethod
    def get_account_ids(self):
        """Get the accountIDs for the accounts created by the component.

        :rtype: list[int]
        :raises valence.components.ComponentStateError: If component has not been applied
        """

    @abc.abstractmethod
    def __len__(self):
        """The number of accounts."""


class CreateAccounts(AccountsInterface):
    """Component used to create accounts.

    This component attempts to reload by checking for existing accounts with the provided tag.
    If no accounts exist, accounts will be created with the username as "{tag}-TAG-{n}" where {tag} is
    the parameter from the constructor; {n} is the nth account created in this component.
    The tsecret and isecret will be automatically generated.

    :param str tag: the UNIQUE tag to use for these accounts.
    :param valence.models.ResourceCluster cluster_dm: cluster to create accounts on. It is assumed that
        the cluster has been authenticated already.
    :param int num: the number of accounts to create.
    """

    def __init__(self, tag, cluster_dm, num=1):
        super(CreateAccounts, self).__init__()

        if num < 1:
            raise ValueError('cannot create less than 1 accounts')

        self.tag = tag
        self.cluster = cluster_dm
        self.num = num
        self._accounts = None
        self._account_ids = None
        self._cluster_fm = functional_models.Cluster(self.cluster, authenticate=False)

    @staticmethod
    def mock(tag, cluster_dm, num=1):  # pylint: disable=unused-argument
        """Mock the CreateAccounts.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._account.CreateAccounts',
                        vl_components.CreateAccounts.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._account.MockAccounts
        """
        account_dms = [models.Account.create(account_id=i) for i in range(num)]
        return MockAccounts(tag, account_dms)

    @property
    def account_dms(self):
        """Get the account data models.

        :rtype: list[valence.models.Account]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._accounts

    def get_account_dms(self):
        """Get the account models created by the component.

        This will make an actual API. Use `account_dms` for the cached data models.

        :rtype: tuple[valence.models.Account]
        :raises valence.components.ComponentStateError: If component has not been applied
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._get_existing()

    def get_account_ids(self):
        """Get the accountIDs for the accounts created by the component.

        :rtype: list[int]
        :raises valence.components.ComponentStateError: If component has not been applied
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return [a.account_id for a in self._accounts]

    # Component methods
    def _reload(self):
        """Attempt to reload the component.

        Checks to see if the accounts already exist.

        :returns: whether the component was able to reload
        :rtype: bool
        """
        accounts = self._get_existing()
        if accounts:
            _log.info('accounts with tag {} were reloaded'.format(self.tag))
            self._assert_valid(accounts)
            self._accounts = accounts
            return True
        else:
            return False

    def _apply(self, timeout=None):
        """Create the accounts and verify creation.

        :param int timeout: Not used since account creation is synchronous.
        """
        # either load existing accounts or create new ones
        accounts = self._create(self.num)
        self._assert_valid(accounts)
        _log.info('accounts with tag {} were created'.format(self.tag))
        self._accounts = accounts

    def _reset(self, timeout=None):
        """Delete the accounts associated with the component.

        If the accounts were found via reload, they will still be deleted.

        :param int timeout: Not used since account creation is synchronous.
        """
        _log.info('Resetting accounts component: {} accounts with tag {}'.format(
            self.num, repr(self.tag)))
        accounts = self._get_existing()
        self._delete(accounts)
        self._accounts = None

    @property
    def full_tag(self):
        """Actual tag used in account username.

        Using '-TAG-' helps ensure the uniqueness.

        :rtype: str
        """
        return self.tag + '-TAG-'

    def _assert_valid(self, accounts):
        """Assert that the accounts meet the criteria of the component.

        Verifies the number of accounts created are the number requested.
        If the component has been reloaded, the number should be the same as the original component.

        :param tuple[valence.models.Account] accounts: accounts to validate
        :raises valence.components.PartialStateError: if the accounts requested are not created
        """
        if len(accounts) != self.num:
            msg = 'length of accounts is invalid: {} != {}'.format(
                len(accounts), self.num)
            raise components.PartialStateError(msg)

    def _get_existing(self):
        """Get existing accounts.

        :rtype: tuple[valence.models.Account]
        """
        return tuple(a for a in self._list_accounts() if a.username.startswith(self.full_tag))

    def _list_accounts(self):
        """Calls ListAccounts on the cluster and returns the data models.

        :rtype: tuple[valence.models.Account]
        """
        accounts = self._cluster_fm.list_accounts()
        return tuple(accounts)

    def _create(self, num, start=0):
        """Sets up the account names and create the accounts.

        :param int num: Number of accounts to create
        :param int start: the starting index for the account username
        :rtype: tuple[valence.models.Account]
        """
        _log.info('Creating {} accounts with tag {}'.format(num, repr(self.tag)))
        names = tuple(self.full_tag + str(n) for n in range(start, start + num))
        return self._create_helper(names)

    def _create_helper(self, names):
        """Actually create accounts.

        :param tuple[str] names: tuple of account names to create
        :rtype: tuple[valence.models.Account]
        """
        partial_account_dms = []
        for name in names:
            partial_account_dms.append(models.Account.create(username=name))
        account_dms = vl_helpers.create_accounts(self._cluster_fm, account_dms=partial_account_dms)
        return tuple(account_dms)

    def _delete(self, accounts):
        """Delete accounts.

        :param tuple[valence.models.Account] accounts: accounts to delete
        """
        vl_helpers.delete_accounts(self._cluster_fm, account_dms=accounts)

    def __len__(self):
        """The number of accounts."""
        return self.num


class MockAccounts(components.MockComponentMixin, AccountsInterface):
    """A mock accounts component representing a list of account data models.

    See the documentation on MockComponentMixin for more details about correct usage.

    :param str tag: the UNIQUE tag to use for these accounts.
    :param list[valence.models.Account] account_dms: the account data models this
        component pretends to be.
    """
    def __init__(self, tag, account_dms):
        super(MockAccounts, self).__init__()
        self.tag = tag
        self.num = len(account_dms)
        self._account_dms = account_dms

    @property
    def account_dms(self):
        """Get the account data models.

        :rtype: list[valence.models.Account]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._account_dms

    @property
    def full_tag(self):
        """Actual tag used in account username.

        Using '-TAG-' helps ensure the uniqueness.

        :rtype: str
        """
        return self.tag + '-TAG-'

    def get_account_dms(self):
        """Get the account models created by the component.

        Since this is a facade component, no API calls are made.

        :rtype: tuple[valence.models.Account]
        :raises valence.components.ComponentStateError: If component has not been applied
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._account_dms

    def get_account_ids(self):
        """Get the accountIDs for the accounts created by the component.

        :rtype: list[int]
        :raises valence.components.ComponentStateError: If component has not been applied
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return [a.account_id for a in self._account_dms]

    def __len__(self):
        """The number of accounts."""
        return len(self._account_dms)
