"""Valence capacity components."""
import itertools

import autologger

from valence import (
    components,
    enums,
    functional_models,
    selectors,
    vl_helpers,
)
from valence.vl_components import _drive

_log = autologger.get_sflogging(__name__)


class UnbalanceCluster(components.BlockingApply, components.Reversible):
    """Unbalance the cluster by removing drives from all but the largest node.

    This puts the cluster into a state where a single drive failure could strand capacity.

    :param valence.models.ResourceCluster cluster_dm: The cluster to unbalance.
    :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
    """
    def __init__(self, cluster_dm, force_during_bin_sync=False):
        super(UnbalanceCluster, self).__init__()
        self.cluster_dm = cluster_dm
        self._force_during_bin_sync = force_during_bin_sync

        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self.sorted_node_fms = None
        self._remove_drives_cmt = None

    def _apply_setup(self):
        """Sort the nodes and initialize the drive removal components."""
        self.sorted_node_fms = vl_helpers.list_nodes_by_usable_block_capacity(self.cluster_fm)
        if len(self.sorted_node_fms) == 1:
            raise RuntimeError('Cannot unbalance a single node cluster.')
        elif len(self.sorted_node_fms) == 2:
            _log.debug('No drive removals necessary to unbalance a 2-node cluster')
        else:
            self._remove_drives_cmt = self._get_remove_drives_cmt()

    def _get_remove_drives_cmt(self):
        """Get drive removal component that will unbalance the cluster.

        :rtype: valence.vl_components.RemoveDrives
        """
        node_id_map = {n.mip: n.node_id for n in self.cluster_fm.list_active_nodes()}

        node_drives_map = {}
        for node_fm in self.sorted_node_fms:
            node_drives_map[node_fm.mip] = self.cluster_fm.list_drives(
                node_ids=[node_id_map[node_fm.mip]],
                statuses=[enums.DriveStatus.active],
                drive_types=[enums.DriveType.block]
            )

        node_capacities = []
        for node_fm in self.sorted_node_fms:
            node_capacities.append(sum(d.usable_capacity for d in node_drives_map[node_fm.mip]))

        # Remove drives from all but the largest node
        node_fms_loop = itertools.cycle(self.sorted_node_fms[:-1])
        capacity_to_remove = sum(node_capacities) - 2 * node_capacities[-1]
        drives_to_remove = []
        removed_capacity = 0
        for node_fm in node_fms_loop:
            if node_drives_map[node_fm.mip]:
                drive = node_drives_map[node_fm.mip].pop()
                if removed_capacity + drive.usable_capacity > capacity_to_remove:
                    break
                else:
                    removed_capacity += drive.usable_capacity
                    drives_to_remove.append(drive)

        remove_drives_cmt = None
        if drives_to_remove:
            remove_drives_cmt = _drive.RemoveDrives(
                cluster_dm=self.cluster_dm,
                drive_selector=selectors.FixedDriveSelector(drives_to_remove),
                ignore_unbalanced_cluster_check=True,
                force_during_bin_sync=self._force_during_bin_sync
            )

        return remove_drives_cmt

    def _apply(self, timeout=None):
        # TODO: VL-2435 Remove constant manipulation once RemoveDrives passes timeout through properly
        if self._remove_drives_cmt:
            old_timeout = self._remove_drives_cmt.TIMEOUT
            if timeout:
                self._remove_drives_cmt.TIMEOUT = timeout
            self._remove_drives_cmt.apply(timeout)
            self._remove_drives_cmt.TIMEOUT = old_timeout

    def _reset(self, timeout=None):
        if self._remove_drives_cmt:
            self._remove_drives_cmt.ensure_ready(timeout)


class StrandCapacity(components.BlockingApply, components.Reversible):
    """Strand capacity on the cluster by failing a drive on the second-largest node.

    :param valence.vl_components.UnbalanceCluster unbalance_cluster_cmt: Component for unbalancing the
        cluster. This is required as capacity cannot be stranded on a balanced cluster.
    """
    def __init__(self, unbalance_cluster_cmt):
        super(StrandCapacity, self).__init__()
        self.unbalance_cluster_cmt = unbalance_cluster_cmt
        self.cluster_dm = unbalance_cluster_cmt.cluster_dm
        self.cluster_fm = unbalance_cluster_cmt.cluster_fm
        self.second_largest_node_fm = None

        self._fail_drives_cmt = None

    def assert_valid(self):
        """Ensure a current stranded capacity fault exists on the cluster."""
        assert self.cluster_fm.list_cluster_faults(
            fault_status=enums.ClusterFaultStatus.current,
            include_codes=[enums.ClusterFaultCode.stranded_capacity],
            best_practices=True,
            update=True
        )

    def _apply_setup(self):
        """Ensure the cluster is unbalanced and initialize the drive failure component."""
        self.unbalance_cluster_cmt.ensure_applied()
        self.second_largest_node_fm = self.unbalance_cluster_cmt.sorted_node_fms[-2]
        vl_helpers.set_fake_drives_flag_on_node(self.second_largest_node_fm)
        drive_selector = selectors.DriveSelector(
            cluster_dm=self.cluster_dm,
            num=1,
            statuses=[enums.DriveStatus.active],
            types=[enums.DriveType.block],
            node_selector=selectors.FixedNodeSelector([self.second_largest_node_fm.get_node_dm()])
        )
        self._fail_drives_cmt = _drive.FailDrives(self.cluster_dm, drive_selector)

    def _apply(self, timeout=None):
        self._fail_drives_cmt.apply(timeout)
        self.assert_valid()

    def _reset(self, timeout=None):
        self._fail_drives_cmt.reset(timeout)
