"""Components related to customer installable SSL certificates."""
import abc
import datetime
import os
import time

import autologger
import six

from valence import (
    components,
    functional_models,
)

_log = autologger.get_sflogging(__name__)

cert_component_dir = os.path.dirname(os.path.realpath(__file__))

REQ_CONF_FILE = os.path.join(cert_component_dir, 'cert_request.conf')

SSL_CONF_FILE = os.path.join(cert_component_dir, 'ssl.conf')

BEGIN_CERT = '-----BEGIN CERTIFICATE-----'


def make_ss_cert_cmd(target_dir, key_out, cert_out, days, config):
    """Create the command to self sign a certificate.

    :param str target_dir: directory the certificate, key, and output should be in
    :param str key_out: filename for the key
    :param str cert_out: filename for the certificate
    :param int days: number of days this certificate should be valid for
    :param str config: complete path to the ssl configuration to use
    :return str: command to execute to make the certificate

    :rtype: str
    """
    MAKE_SS_CMD = ('openssl req -x509 -newkey rsa:2048 -keyout {}'
                   ' -out {} -days {} -nodes -config {}')
    return MAKE_SS_CMD.format(os.path.join(target_dir, key_out), os.path.join(target_dir, cert_out),
                              days, config)


def create_sign_cmds(target_dir, key, csr, crt, conf, ssl_conf, ca, root_key, start_date, end_date):
    """Create commands to create a key, certificate signing request, and certificate.

    :param str target_dir: directory the certificate, key, and output should be in
    :param str key: filename for the key
    :param str csr: filename for the signing request
    :param str crt: filename for the signed certificate
    :param str conf: complete path to the ssl configuration to use to create the request
    :param str ssl_conf: certificate authority configuration to sign the request
    :param str ca: filename for the certificate authority
    :param str root_key: filename for the root key
    :param str start_date: '%Y%m%d%H%M%SZ' formatted datetime when the cert should start being valid
    :param str end_date: '%Y%m%d%H%M%SZ' formatted datetime when the cert should stop being valid
    :return (str, str, str): gen_key, gen_csr, sign_csr commands to execute

    :rtype: tuple
    """
    GEN_ROOT_KEY = 'openssl genrsa -out {} 2048'
    GEN_CSR = 'openssl req -new -key {} -out {} -config {}'
    SIGN_CA_CSR = ('openssl ca -batch -config {} -in {} -cert {} -keyfile {} -out {} -startdate {} '
                   '-enddate {}')
    gen_key_command = GEN_ROOT_KEY.format(os.path.join(target_dir, key))
    gen_csr_command = GEN_CSR.format(os.path.join(target_dir, key),
                                     os.path.join(target_dir, csr),
                                     os.path.join(target_dir, conf))
    sign_csr_command = SIGN_CA_CSR.format(os.path.join(target_dir, ssl_conf),
                                          os.path.join(target_dir, csr),
                                          os.path.join(target_dir, ca),
                                          os.path.join(target_dir, root_key),
                                          os.path.join(target_dir, crt),
                                          start_date, end_date)
    return gen_key_command, gen_csr_command, sign_csr_command


@six.add_metaclass(abc.ABCMeta)
class BaseClientCertificateComponent(components.BlockingApply, components.Reversible):
    """Base Client Certificate Component.

    Creates certificates, signs them and so on
    Must override the _apply method to make a real test component - the base component doesn't
    know what certificates to send to the cluster.

    :param valence.models.ResourceClient client_dm: client data model to create the certs on
    :param valence.models.Cluster cluster_dm: clusters to set the certificates on
    :param valence.models.ResourceManagementNode management_node_dm: node data model
    """
    CLIENT_CN_CONF = 'ca_req.conf'
    CLIENT_SSL_CONF = 'ca.conf'
    CLIENT_ROOT_KEY = 'rootCA.key'
    CLIENT_CA = 'TESTCA.pem'
    IP_CN_CONF = '{}IP.conf'
    DOMAIN_CN_CONF = '{}domain.conf'
    IP_SS_KEY = '{}IPkey.pem'
    IP_SS = '{}IP_SS.pem'
    SOON_TO_EXPIRE_IP_SS = '{}IP_SSexpiring.pem'
    DOM_SS_KEY = '{}DOMkey.pem'
    DOM_SS = '{}DOM_SS.pem'
    SOON_TO_EXPIRE_DOM_SS = '{}DOM_SSexpiring.pem'
    DEV_IP_KEY = '{}IPdevice.key'
    DEV_IP_CSR = '{}IPdevice.csr'
    DEV_IP_CRT = '{}IPdevice.crt'
    DEV_DOM_KEY = '{}DOMdevice.key'
    DEV_DOM_CSR = '{}DOMdevice.csr'
    DEV_DOM_CRT = '{}DOMdevice.crt'
    TEST_DOMAIN = '*.test.solidfire.net'
    GEN_ROOT_KEY = 'openssl genrsa -out {} 2048'
    SELF_SIGN_CA = ('openssl req -x509 -new -nodes -key {} -sha256 -days 100 -out '
                    '{} -config {}')
    REVOKE_CSR = 'openssl ca -config {} -revoke {}'
    SSH_TIMEOUT = 10

    def __init__(self, client_dm, cluster_dm, management_node_dm):
        """Constructor for BaseClientCertComponent."""
        super(BaseClientCertificateComponent, self).__init__()
        self._client_dm = client_dm
        self._client_fm = functional_models.get_client_from_dm(self._client_dm)
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(self._cluster_dm)
        self._management_node_dm = management_node_dm
        self._mnode_fm = functional_models.ManagementNodeAdapter(
            self._management_node_dm, self._cluster_dm)
        self.temp_dir = None
        self.start_date = datetime.datetime.now()
        self.end_date = datetime.datetime.now() + datetime.timedelta(days=365)
        self.days = 60
        self.certificate = None
        self.private_key = None
        self.fdva_certificate = None
        self.fdva_private_key = None
        self.req_conf = None
        self.ssl_conf = None

    def _put_configs_on_client(self):
        """Create the config files required to make a CA and stuff."""
        with open(REQ_CONF_FILE, 'r') as req_conf:
            self.req_conf = req_conf.read()
        with open(SSL_CONF_FILE, 'r') as ssl_conf:
            self.ssl_conf = ssl_conf.read()
        self._client_fm.create_file(target_dir=self.temp_dir,
                                    filename=self.CLIENT_CN_CONF,
                                    raw=self.req_conf.format(self._client_dm.client_name))
        self._client_fm.create_file(target_dir=self.temp_dir,
                                    filename=self.CLIENT_SSL_CONF,
                                    raw=self.ssl_conf.format(self.temp_dir))

        with self._client_fm.get_ssh_connection() as ssh_inst:
            ssh_inst.run_command('touch {}'.format(os.path.join(self.temp_dir, 'index.txt')))
            ssh_inst.run_command('echo 1000 > {}'.format(os.path.join(self.temp_dir, 'serial')))
        self._client_fm.create_file(
            target_dir=self.temp_dir,
            filename=self.IP_CN_CONF.format(self._cluster_dm.mvip),
            raw=self.req_conf.format(self._cluster_dm.mvip))
        self._client_fm.create_file(
            target_dir=self.temp_dir,
            filename=self.DOMAIN_CN_CONF.format(self._cluster_dm.mvip),
            raw=self.req_conf.format(self.TEST_DOMAIN))
        self._client_fm.create_file(
            target_dir=self.temp_dir,
            filename=self.IP_CN_CONF.format(self._management_node_dm.mip),
            raw=self.req_conf.format(self._management_node_dm.mip))
        self._client_fm.create_file(
            target_dir=self.temp_dir,
            filename=self.DOMAIN_CN_CONF.format(self._management_node_dm.mip),
            raw=self.req_conf.format(self.TEST_DOMAIN))

    def _create_self_signed_certs(self):
        """Create self signed certs for test."""
        self_sign_commands = []
        ip_config = os.path.join(self.temp_dir, self.IP_CN_CONF.format(self._cluster_dm.mvip))
        domain_config = os.path.join(self.temp_dir,
                                     self.DOMAIN_CN_CONF.format(self._cluster_dm.mvip))
        make_ip_ss_cmd = make_ss_cert_cmd(self.temp_dir,
                                          self.IP_SS_KEY.format(self._cluster_dm.mvip),
                                          self.IP_SS.format(self._cluster_dm.mvip),
                                          self.days, ip_config)
        make_dom_ss_cmd = make_ss_cert_cmd(self.temp_dir,
                                           self.DOM_SS_KEY.format(self._cluster_dm.mvip),
                                           self.DOM_SS.format(self._cluster_dm.mvip),
                                           self.days, domain_config)
        self_sign_commands.extend([make_ip_ss_cmd, make_dom_ss_cmd])
        ip_config = os.path.join(self.temp_dir, self.IP_CN_CONF.format(
            self._management_node_dm.mip))
        domain_config = os.path.join(self.temp_dir,
                                     self.DOMAIN_CN_CONF.format(self._management_node_dm.mip))
        make_ip_ss_cmd = make_ss_cert_cmd(
            self.temp_dir, self.IP_SS_KEY.format(self._management_node_dm.mip),
            self.IP_SS.format(self._management_node_dm.mip), self.days, ip_config)
        make_dom_ss_cmd = make_ss_cert_cmd(
            self.temp_dir, self.DOM_SS_KEY.format(self._management_node_dm.mip),
            self.DOM_SS.format(self._management_node_dm.mip), self.days, domain_config)
        self_sign_commands.extend([make_ip_ss_cmd, make_dom_ss_cmd])
        with self._client_fm.get_ssh_connection() as ssh_inst:
            for self_sign in self_sign_commands:
                ssh_inst.run_command(self_sign, timeout=self.SSH_TIMEOUT, raise_on_failure=True)

    def _create_cert_authority(self):
        """Create a new root key and certificate authority."""
        gen_root_key_command = self.GEN_ROOT_KEY.format(
            os.path.join(self.temp_dir, self.CLIENT_ROOT_KEY))
        self_sign_ca_command = self.SELF_SIGN_CA.format(
            os.path.join(self.temp_dir, self.CLIENT_ROOT_KEY),
            os.path.join(self.temp_dir, self.CLIENT_CA),
            os.path.join(self.temp_dir, self.CLIENT_CN_CONF))
        with self._client_fm.get_ssh_connection() as ssh_inst:
            for cmd in [gen_root_key_command, self_sign_ca_command]:
                ssh_inst.run_command(cmd, timeout=self.SSH_TIMEOUT, raise_on_failure=True)

    def _sign_with_cert_authority(self):
        """Using an existing certificate authority self.CLIENT_CA sign certs."""
        cert_auth_sign_commands = []
        start_date = datetime.datetime.strftime(self.start_date, '%Y%m%d%H%M%SZ')
        end_date = datetime.datetime.strftime(self.end_date, '%Y%m%d%H%M%SZ')
        # TODO: ELEM-5292 refactor this
        (gen_ip_key_command, gen_ip_dev_csr_command,
         sign_ip_csr_command) = create_sign_cmds(
            self.temp_dir, self.DEV_IP_KEY.format(self._cluster_dm.mvip),
            self.DEV_IP_CSR.format(self._cluster_dm.mvip),
            self.DEV_IP_CRT.format(self._cluster_dm.mvip),
            self.IP_CN_CONF.format(self._cluster_dm.mvip),
            self.CLIENT_SSL_CONF, self.CLIENT_CA, self.CLIENT_ROOT_KEY,
            start_date, end_date)
        (gen_dom_key_command, gen_dom_dev_csr_command,
         sign_dom_csr_command) = create_sign_cmds(
            self.temp_dir, self.DEV_DOM_KEY.format(self._cluster_dm.mvip),
            self.DEV_DOM_CSR.format(self._cluster_dm.mvip),
            self.DEV_DOM_CRT.format(self._cluster_dm.mvip),
            self.DOMAIN_CN_CONF.format(self._cluster_dm.mvip),
            self.CLIENT_SSL_CONF, self.CLIENT_CA, self.CLIENT_ROOT_KEY,
            start_date, end_date)
        cert_auth_sign_commands.extend([gen_ip_key_command, gen_ip_dev_csr_command,
                                        sign_ip_csr_command,
                                        gen_dom_key_command, gen_dom_dev_csr_command,
                                        sign_dom_csr_command])
        (gen_ip_key_command, gen_ip_dev_csr_command,
         sign_ip_csr_command) = create_sign_cmds(
            self.temp_dir, self.DEV_IP_KEY.format(self._management_node_dm.mip),
            self.DEV_IP_CSR.format(self._management_node_dm.mip),
            self.DEV_IP_CRT.format(self._management_node_dm.mip),
            self.IP_CN_CONF.format(self._management_node_dm.mip),
            self.CLIENT_SSL_CONF, self.CLIENT_CA, self.CLIENT_ROOT_KEY,
            start_date, end_date)
        (gen_dom_key_command, gen_dom_dev_csr_command,
         sign_dom_csr_command) = create_sign_cmds(
            self.temp_dir, self.DEV_DOM_KEY.format(self._management_node_dm.mip),
            self.DEV_DOM_CSR.format(self._management_node_dm.mip),
            self.DEV_DOM_CRT.format(self._management_node_dm.mip),
            self.DOMAIN_CN_CONF.format(self._management_node_dm.mip),
            self.CLIENT_SSL_CONF, self.CLIENT_CA, self.CLIENT_ROOT_KEY,
            start_date, end_date)
        cert_auth_sign_commands.extend([gen_ip_key_command, gen_ip_dev_csr_command,
                                        sign_ip_csr_command,
                                        gen_dom_key_command, gen_dom_dev_csr_command,
                                        sign_dom_csr_command])

        with self._client_fm.get_ssh_connection() as ssh_inst:
            # First make sure we do not have a requirement that each signed cert be unique
            ssh_inst.run_command(
                'echo "unique_subject = no" > {}/index.txt.attr'.format(self.temp_dir),
                timeout=self.SSH_TIMEOUT, raise_on_failure=True)
            for self_sign in cert_auth_sign_commands:
                ssh_inst.run_command(self_sign, timeout=self.SSH_TIMEOUT, raise_on_failure=True)

    def _revoke_ca_certs(self):
        """Attempt to revoke all the signed certs from the certificate authority."""
        # TODO: ELEM-5292 refactor this
        revoke_cmds = []
        revoke_ip_cmd = self.REVOKE_CSR.format(
            os.path.join(self.temp_dir, self.CLIENT_SSL_CONF),
            os.path.join(self.temp_dir, self.DEV_IP_CRT.format(self._cluster_dm.mvip)),)
        revoke_dom_cmd = self.REVOKE_CSR.format(
            os.path.join(self.temp_dir, self.CLIENT_SSL_CONF),
            os.path.join(self.temp_dir, self.DEV_DOM_CRT.format(self._cluster_dm.mvip)),)
        revoke_cmds.extend([revoke_ip_cmd, revoke_dom_cmd])
        revoke_ip_cmd = self.REVOKE_CSR.format(
            os.path.join(self.temp_dir, self.CLIENT_SSL_CONF),
            os.path.join(self.temp_dir,
                         self.DEV_IP_CRT.format(self._management_node_dm.mip)),)
        revoke_dom_cmd = self.REVOKE_CSR.format(
            os.path.join(self.temp_dir, self.CLIENT_SSL_CONF),
            os.path.join(self.temp_dir,
                         self.DEV_DOM_CRT.format(self._management_node_dm.mip)),)
        revoke_cmds.extend([revoke_ip_cmd, revoke_dom_cmd])

        with self._client_fm.get_ssh_connection() as ssh_inst:
            for self_sign in revoke_cmds:
                ssh_inst.run_command(self_sign, timeout=self.SSH_TIMEOUT, raise_on_failure=True)

    def _apply_setup(self):
        """Prepare the client, this includes creating certificates, signing them, etc."""
        # Create temp directory
        self.temp_dir = self._client_fm.create_temp_cert_dir()
        self._put_configs_on_client()
        self._create_self_signed_certs()
        self._create_cert_authority()
        self._sign_with_cert_authority()

    def _reset(self, timeout=None):
        """Remove any certificates, revoke certificates from any CAs if not already revoked.

        :param int|None timeout: timeout
        """
        self._cluster_fm.remove_ssl_certificate()
        self._mnode_fm.authenticate()
        self._mnode_fm.remove_node_ssl_certificate()
        try:
            self._revoke_ca_certs()
        except functional_models.FunctionalClientError as e:
            if 'Already revoked' in str(e):
                _log.warning('Certificates already revoked, no need to reset')
            else:
                raise
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateSelfSignedIPCert(BaseClientCertificateComponent):
    """CreateSelfSignedIPCert, creates self signed certificates using the ip as the commonName.

    Sets the ssl certificate on a cluster to this using cluster_api.set_ssl_certificate
    """

    def _apply(self, timeout=None):
        self.certificate = self._client_fm.get_file_contents(
            self.temp_dir, self.IP_SS.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)

        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.IP_SS_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key)
        self.fdva_certificate = self._client_fm.get_file_contents(
            self.temp_dir, self.IP_SS.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.IP_SS_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateSelfSignedDOMCert(BaseClientCertificateComponent):
    """Self signed domain certificate.

    CreateSelfSignedDOMCert, creates self signed certificates using test domain as the commonName.

    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate
    """

    def _apply(self, timeout=None):
        self.certificate = self._client_fm.get_file_contents(
            self.temp_dir, self.DOM_SS.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DOM_SS_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key)
        self.fdva_certificate = self._client_fm.get_file_contents(
            self.temp_dir, self.DOM_SS.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DOM_SS_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateCertAuthorityIPCert(BaseClientCertificateComponent):
    """Creates certificate authority signed certificates using the ip as the commonName.

    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate
    """

    def _apply(self, timeout=None):
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_CRT.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key)
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_CRT.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateCertAuthorityDOMCert(BaseClientCertificateComponent):
    """Creates certificate authority signed certificates using the test domain.

    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate
    """

    def _apply(self, timeout=None):
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_CRT.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key)
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_CRT.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateCertAuthorityIPSoonToExpire(BaseClientCertificateComponent):
    """Creates certificate authority signed certificates using the IP.

    Certs will expire in 10 days
    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate

    :param valence.models.ResourceClient client_dm: client data model to create the certs on
    :param valence.models.Cluster cluster_dm: clusters to set the certificates on
    :param valence.models.ResourceManagementNode management_node_dm: node data model
    """

    def __init__(self, client_dm, cluster_dm=None, management_node_dm=None):
        """Constructor for CreateCertAuthorityIPSoonToExpire."""
        super(CreateCertAuthorityIPSoonToExpire, self).__init__(client_dm, cluster_dm,
                                                                management_node_dm)
        self.start_date = datetime.datetime.now() - datetime.timedelta(days=365)
        self.end_date = datetime.datetime.now() + datetime.timedelta(days=10)

    def _apply(self, timeout=None):
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_CRT.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key)
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_CRT.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateCertAuthorityDOMSoonToExpire(BaseClientCertificateComponent):
    """Creates certificate authority signed certificates using the test domain.

    Certs will expire in 10 days
    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate

    :param valence.models.ResourceClient client_dm: client data model to create the certs on
    :param valence.models.Cluster cluster_dm: clusters to set the certificates on
    :param valence.models.ResourceManagementNode management_node_dm: node data model
    """

    def __init__(self, client_dm, cluster_dm=None, management_node_dm=None):
        """Constructor for CreateCertAuthorityDOMSoonToExpire."""
        super(CreateCertAuthorityDOMSoonToExpire, self).__init__(client_dm, cluster_dm,
                                                                 management_node_dm)
        self.start_date = datetime.datetime.now() - datetime.timedelta(days=365)
        self.end_date = datetime.datetime.now() + datetime.timedelta(days=10)

    def _apply(self, timeout=None):
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_CRT.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key)
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_CRT.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateCertAuthorityIPExpired(BaseClientCertificateComponent):
    """Creates certificate authority signed certificates using the IP.

    Certs are expired
    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate

    :param valence.models.ResourceClient client_dm: client data model to create the certs on
    :param valence.models.Cluster cluster_dm: clusters to set the certificates on
    :param valence.models.ResourceManagementNode management_node_dm: node data model
    """

    def __init__(self, client_dm, cluster_dm=None, management_node_dm=None):
        """Constructor for CreateCertAuthorityIPExpired."""
        super(CreateCertAuthorityIPExpired, self).__init__(client_dm, cluster_dm, management_node_dm)
        self.start_date = datetime.datetime.now() - datetime.timedelta(days=365)
        self.end_date = datetime.datetime.now() - datetime.timedelta(days=10)

    def _apply(self, timeout=None):
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_CRT.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key, force=True)
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_CRT.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_IP_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key, force=True)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)


class CreateCertAuthorityDOMExpired(BaseClientCertificateComponent):
    """Creates certificate authority signed certificates using the test domain.

    Certs are expired
    Sets the ssl certificate on a cluster to this using self._cluster_fm.set_ssl_certificate

    :param valence.models.ResourceClient client_dm: client data model to create the certs on
    :param valence.models.Cluster cluster_dm: clusters to set the certificates on
    :param valence.models.ResourceManagementNode management_node_dm: node data model
    """

    def __init__(self, client_dm, cluster_dm=None, management_node_dm=None):
        """Constructor for CreateCertAuthorityDOMExpired."""
        super(CreateCertAuthorityDOMExpired, self).__init__(client_dm, cluster_dm, management_node_dm)
        self.start_date = datetime.datetime.now() - datetime.timedelta(days=365)
        self.end_date = datetime.datetime.now() - datetime.timedelta(days=10)

    def _apply(self, timeout=None):
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_CRT.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self.certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_KEY.format(self._cluster_dm.mvip),
            timeout=self.SSH_TIMEOUT)
        self._cluster_fm.set_ssl_certificate(certificate=self.certificate,
                                             private_key=self.private_key, force=True)
        raw_cert = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_CRT.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self.fdva_certificate = raw_cert[raw_cert.find(BEGIN_CERT):]
        self.fdva_private_key = self._client_fm.get_file_contents(
            self.temp_dir, self.DEV_DOM_KEY.format(self._management_node_dm.mip),
            timeout=self.SSH_TIMEOUT)
        self._mnode_fm.set_node_ssl_certificate(certificate=self.fdva_certificate,
                                                private_key=self.fdva_private_key, force=True)
        # cover the minor delay in pointing to a new .crt
        time.sleep(1)
