"""Valence cluster components."""
import collections
import time

import autologger

from valence import (
    components,
    criteria,
    enums,
    errors,
    functional_models,
    models,
    selectors,
    vl_helpers,
)
from valence.vl_components import _constants

_log = autologger.get_sflogging(__name__)


class RemoveNodes(components.Reversible, components.BlockingApply):
    """A reversible component to remove nodes from cluster gracefully and add them back on reset.

        Also, as a part of graceful removal, we must remove drives.
        Reset will add them back to the cluster.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.ActiveNodeSelector node_selector: node selector
    :param int drive_timeout_factor: multiplier for the drive timeout. May want to increase
        during heavy IO or during upgrades.
    :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
    :param bool auto_install: auto install added node to the cluster's current build
    """
    def __init__(self, cluster_dm, node_selector, drive_timeout_factor=1, force_during_bin_sync=None,
                 auto_install=True):
        super(RemoveNodes, self).__init__()
        self._node_selector = node_selector
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._drive_timeout_factor = drive_timeout_factor
        self._nodes = None          # ActiveNodes to be removed during apply
        self._pending_nodes = None  # PendingNodes to be added during reset
        self.force_during_bin_sync = force_during_bin_sync
        self.auto_install = auto_install

    @property
    def pending_node_dms(self):
        """Get the gracefully removed nodes as their PendingNode.

        :rtype: list[valence.models.PendingNode]
        """
        if self.state != components.States.applied:
            raise components.ComponentStateError('Cannot get nodes while not applied')
        return self._pending_nodes

    def _reload(self):
        """Reload the component by checking if any pending node matches the selected node by mip.

            Check whether the node was removed or not.

        :return: whether the component was able to reload
        :rtype: bool
        """
        # Confirm there are pending nodes
        filter_criteria = criteria.ContainedIn(models.PendingNode.static_id,
                                               [n.static_id for n in self._nodes])
        pending_nodes = filter_criteria.filter(self._cluster_fm.list_pending_nodes())

        if len(pending_nodes) == len(self._nodes):
            _log.debug('Node removal for {} successfully reloaded'.format(self._nodes))
            return True
        else:
            return False

    def _apply(self, timeout=None):
        """Remove the selected nodes.

            First remove the drives and wait for the async handler to complete.
            Then remove the nodes.

        :param int timeout: Not used since Node deletion is synchronous.
        """
        drive_dms = self._cluster_fm.list_drives(node_ids=[n.node_id for n in self._nodes])
        drive_timeout = vl_helpers.get_bin_sync_timeout_for_nodes(self._cluster_fm, self._nodes)
        calculated_drive_timeout = drive_timeout*self._drive_timeout_factor
        timeout = calculated_drive_timeout if timeout is None else timeout

        vl_helpers.remove_drives(self._cluster_fm, drive_dms, timeout=timeout,
                                 force_during_bin_sync=self.force_during_bin_sync)
        self._pending_nodes = vl_helpers.remove_nodes(self._cluster_fm, self._nodes, timeout=timeout)
        _log.info('Successfully removed nodes. {}'.format(self._nodes))

    def _apply_setup(self, timeout=None):
        """Force Blocking Apply.

            Check whether we have nodes to remove.

        :param int timeout: Timeout in seconds
        """
        self._nodes = self._node_selector()
        if not self._nodes:
            raise components.ComponentStateError('The node to be removed is not currently '
                                                 'active {}'.format(self._nodes))

    def _reset(self, timeout=None):
        """Reverts back to original component.

            Adds the node associated with the component back to the cluster.
            Check if originally remove nodes are still pending.
            If true, add nodes and wait for their async handles to complete.
            Then add the drives back and wait for their async handles to complete.

        :param int timeout: Not used since Node deletion is synchronous
        """
        # Confirm there are pending nodes
        filter_criteria = criteria.ContainedIn(models.PendingNode.static_id,
                                               [n.static_id for n in self._nodes])
        pending_nodes = filter_criteria.filter(self._cluster_fm.list_pending_nodes())

        # If upgrading, RTFI pending nodes to cluster build version
        _log.debug('Is the cluster upgrading? {}'.format(self._cluster_fm.upgrading))
        if self._cluster_fm.upgrading:
            self._rtfi_nodes_to_cm_version_during_upgrade(pending_nodes)
            # RTFI may change the pendingNodeID, so refresh our pending info:
            pending_nodes = filter_criteria.filter(self._cluster_fm.list_pending_nodes())

        # add drives back to node and add node back to cluster
        _log.debug('Add nodes and drives back to cluster...')
        drive_timeout = vl_helpers.get_bin_sync_timeout_for_nodes(self._cluster_fm, self._nodes)
        calculated_drive_timeout = drive_timeout*self._drive_timeout_factor
        timeout = calculated_drive_timeout if timeout is None else timeout

        if len(pending_nodes) != len(self._nodes):
            raise components.PartialStateError('Not all nodes are pending. Expected {} vs '
                                               'Found {}'.format(self._nodes, pending_nodes))
        _log.debug('Adding pending nodes {}'.format(pending_nodes))
        vl_helpers.add_nodes(self._cluster_fm, pending_nodes, timeout=timeout,
                             auto_install=self.auto_install)
        # It's possible that drives don't display immediately after node has been added.
        _log.debug('Waiting for available drives to appear.')
        vl_helpers.wait_for_any_drive_statuses(self._cluster_fm, [enums.DriveStatus.available])
        add_drive_filtered_nodes = filter_criteria.filter(self._cluster_fm.list_active_nodes())
        drive_dms = self._cluster_fm.list_drives(node_ids=[n.node_id for n in add_drive_filtered_nodes])
        vl_helpers.add_drives(self._cluster_fm, drive_dms, timeout=timeout,
                              force_during_bin_sync=self.force_during_bin_sync)
        self._pending_nodes = None
        _log.info('Successfully added nodes back. {}'.format(self._nodes))

    def _rtfi_nodes_to_cm_version_during_upgrade(self, pending_nodes):
        """RTFI the nodes to cluster build version.

        Note: if the node is RTFI'd, its pendingNodeID may change, so refresh your dms.

        :param list[valence.models.PendingNode] pending_nodes: check version matches cluster
        """
        _log.debug('Upgrade detected, checking pending node versions to cluster build version')
        cluster_version_info_dm = self._cluster_fm.get_cluster_version_info()
        build_version = cluster_version_info_dm.cluster_version  # ie 10.0.0.284, not 10.0

        # confirm each pending_node is cluster build version
        pending_nodes_to_rtfi = []
        for n in pending_nodes:
            if n.software_version != build_version:
                _log.debug('RTFI Pending Node {}'.format(n.name))
                _log.debug('    {}'.format(n.serialize()))
                pending_nodes_to_rtfi.append(n)
            else:
                _log.debug("{} will not be RTFI'ed since it's already at desired version".format(n.name))

        # RTFI pending nodes to cluster build version
        if pending_nodes_to_rtfi:
            node_fms_to_rtfi = [functional_models.Node(n.resource_dm, self._cluster_dm)
                                for n in pending_nodes_to_rtfi]

            # Workaround alert. tRTFI will reset the network config, save this state.
            # save network config before RTFI attempts
            network_configs = {}
            for n in node_fms_to_rtfi:
                network_configs[n.node_dm.node_name] = n.get_network_config()

            vl_helpers.provision_nodes(node_fms_to_rtfi, build_version)

            # restore network config after RTFI attempts
            for n in node_fms_to_rtfi:
                _log.debug('Beginning watch of pernode_api to be up on node {}'.format(n.mip))
                n.poll_verify_pernode_api()
                _log.debug('Restoring the NetworkConfig on node {}'.format(n.mip))
                n.set_network_config(network_configs[n.node_dm.node_name])


class RemoveAndOrphanNodes(components.Reversible, components.BlockingApply):
    """Completely remove nodes.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.ActiveNodeSelector node_selector: node selector
    :param int drive_timeout_factor: multiplier for the drive timeout. May want to increase
        during heavy IO or during upgrades.
    :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
    """
    TIMEOUT = 600  # Default hardcoded timeout that does not adjust dynamically based on drive size.

    def __init__(self, cluster_dm, node_selector, drive_timeout_factor=1, force_during_bin_sync=None):
        super(RemoveAndOrphanNodes, self).__init__()
        self._node_selector = node_selector
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._drive_timeout_factor = drive_timeout_factor
        self._remove_nodes_cmt = RemoveNodes(self._cluster_dm, self._node_selector)
        self._rs_nodes = None  # ResourceNodes to be associated during reset
        self.force_during_bin_sync = force_during_bin_sync

    @property
    def rs_node_dms(self):
        """Get the completely removed nodes (orphaned) as their ResourceNode.

        :rtype: valence.models.ResourceNode
        """
        if self.state != components.States.applied:
            raise components.ComponentStateError('Cannot get nodes while not applied')
        return self._rs_nodes

    def _apply(self, timeout=None):
        timeout = self.TIMEOUT if timeout is None else timeout
        self._remove_nodes_cmt.apply()
        self._rs_nodes = vl_helpers.orphan_pending_nodes(self._cluster_fm,
                                                         self._remove_nodes_cmt.pending_node_dms,
                                                         timeout=timeout)

    def _reset(self, timeout=None):
        timeout = self.TIMEOUT if timeout is None else timeout
        # pylint: disable=protected-access
        rs_node_dms = [n.resource_dm for n in self._remove_nodes_cmt._nodes]
        vl_helpers.associate_orphaned_nodes(self._cluster_fm, rs_node_dms, timeout=timeout)
        self._rs_nodes = None
        self._remove_nodes_cmt.reset()


class _ManipulateMasterService(components.BlockingApply):
    """Base class to restart/stop master service.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm to which the node belongs
    :param valence.models.ResourceNode node_dm: The node dm where the service is to be restarted
    :param valence.enums.UnixSignal signal: The unix signal to be used to kill service
    """

    def __init__(self, cluster_dm, node_dm, signal):
        super(_ManipulateMasterService, self).__init__()
        self._cluster_dm = cluster_dm
        self._node_dm = node_dm
        self._node_fm = functional_models.Node(node_dm=node_dm, cluster_dm=self._cluster_dm)
        self._signal = signal

    @property
    def node_dm(self):
        """Returns the Node DM where the service is being manipulated.

        :rtype: valence.models.ResourceNode
        :returns: the node dm where the service is manipulated.
        """
        return self._node_dm

    def _apply(self, timeout=None):
        self._node_fm.kill_master_service_ssh(signal=self._signal)

    def _reset(self, timeout=None):
        """Brings back the master service.

        :param int timeout: this is not used
        """
        self._node_fm.stop_binary(enums.NodeBinary.solidfire)
        self._node_fm.start_binary(enums.NodeBinary.solidfire)


class _ManipulateDriveService(components.BlockingApply):
    """Base class to restart/stop node services.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm to which the node belongs
    :param valence.models.ResourceNode node_dm: The node dm where the service is to be restarted
    :param valence.models.ClusterServiceInfo service_dm: The service data model of the service,
        which is to be stopped
    :param valence.enums.UnixSignal signal: The unix signal to be used to kill service
    """

    def __init__(self, cluster_dm, node_dm, service_dm, signal):
        super(_ManipulateDriveService, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(self._cluster_dm)
        self._node_dm = node_dm
        self._service_dm = service_dm
        self._node_fm = functional_models.Node(node_dm=node_dm, cluster_dm=self._cluster_dm)
        self._signal = signal

    @property
    def node_dm(self):
        """Returns the Node DM where the service is being manipulated.

        :rtype: valence.models.ResourceNode
        :returns: the node dm where the service is manipulated.
        """
        return self._node_dm

    def _apply(self, timeout=None):
        self._node_fm.kill_drive_service_ssh(service_dm=self._service_dm,
                                             signal=self._signal)

    def _reset(self, timeout=None):
        """Brings back the block/slice service in case it didn't start on its own.

        :param int timeout: this is not used
        """
        drive_dms = self._cluster_fm.list_drives(node_ids=[self._service_dm.drives[0].node_id],
                                                 drive_ids=[self._service_dm.drives[0].drive_id])
        # if drive is already available skip this call to remove_drives
        if not drive_dms[0].drive_status is enums.DriveStatus.available:
            vl_helpers.remove_drives(self._cluster_fm, drive_dms)
        vl_helpers.wait_for_drive_status(self._cluster_fm, drive_dms)
        vl_helpers.add_drives(cluster_fm=self._cluster_fm, drive_dms=drive_dms)


class _RestartMasterService(_ManipulateMasterService, components.Transient):
    """A transient component which restarts the master service on a node.

        The master process is killed on the node with the unix signal '9',
        which is then automatically restarted by sfapp/sfconfig. This should not be called directly
        use the parent component RestartNodeService.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm to which the node belongs
    :param valence.models.ResourceNode node_dm: The node dm where the service is to be restarted
    """

    def __init__(self, cluster_dm, node_dm):
        super(_RestartMasterService, self).__init__(cluster_dm, node_dm, signal=enums.UnixSignal.SIGKILL)


class _RestartDriveService(_ManipulateDriveService, components.Transient):
    """A transient component which restarts the block/slice service on a node.

        The block/slice service is killed on the node with the unix signal '9',
        which is then automatically restarted by sfapp/sfconfig. This should not be called directly
        use the parent component RestartNodeService.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm to which the node belongs
    :param valence.models.ResourceNode node_dm: The node dm where the service is to be restarted
    :param valence.models.ClusterServiceInfo service_dm: The service data model of the service,
        which is to be stopped
    """

    def __init__(self, cluster_dm, node_dm, service_dm):
        super(_RestartDriveService, self).__init__(cluster_dm, node_dm, service_dm,
                                                   signal=enums.UnixSignal.SIGKILL)


class RestartNodeServices(components.Transient, components.BlockingApply):
    """A transient component which restarts the master, block or slice services.

        _RestartMasterService is called for master services and
        _RestartDriveServices is called for slice and block services.
        This is done for all services returned by the service_selector.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which
        the services have to be restarted
    :param valence.selectors.ServiceSelector service_selector: The service selector of the services
        that need to be restarted
    """

    def __init__(self, cluster_dm, service_selector):
        super(RestartNodeServices, self).__init__()
        self._cluster_dm = cluster_dm
        self._service_selector = service_selector
        self._restart_service_cmts = []

    @property
    def node_dms(self):
        """Returns a list of nodes that are being manipulated by the component.

        :returns: List of nodes that are being manipulated by the component.
        :rtype: list[valence.models.ResourceNode]
        """

        node_dms = []
        for cmt in self._restart_service_cmts:
            node_dms.append(cmt.node_dm)
        return node_dms

    def _apply_setup(self):
        """Initializes _RestartDriveService and _RestartMasterService components."""
        self._restart_service_cmts = []
        services = self._service_selector()
        for s in services:
            node_dm = models.get_resource_node_by_static_id(s.node.static_id)

            if s.service.service_type is enums.ServiceType.master:
                restart_master_service_cmt = _RestartMasterService(self._cluster_dm, node_dm)
                self._restart_service_cmts.append(restart_master_service_cmt)

            elif s.service.service_type.is_drive_service:
                restart_drive_service_cmt = _RestartDriveService(self._cluster_dm, node_dm, s)
                self._restart_service_cmts.append(restart_drive_service_cmt)

    def _apply(self, timeout=None):
        for cmt in self._restart_service_cmts:
            cmt.apply()

    def _reset(self, timeout=None):
        for cmt in reversed(self._restart_service_cmts):
            cmt.reset()


class RoundRobinRestartServicesByNode(components.ContinuousTransient):
    """Continuously round robin restart specific service_type(s) one node at a time.

        kill_after_seconds or max_iterations must be specified if you intend to apply synchronously.

    :param valence.models.ResourceCluster cluster_dm: Cluster to operate on.
    :param list[valence.enums.NodeService] service_types: Which service types to restart.
    :param int kill_after_seconds: How long in seconds the loops will run at a minimum.
    :param int sleep_time: How long in seconds to sleep between loops
    :param int max_iterations: Max number of times to run the loop
    """
    def __init__(self, cluster_dm, service_types, kill_after_seconds=None, sleep_time=None,
                 max_iterations=None):
        super(RoundRobinRestartServicesByNode, self).__init__(kill_after_seconds=kill_after_seconds,
                                                              sleep_time=sleep_time,
                                                              max_iterations=max_iterations)
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(self._cluster_dm)
        self._service_types = service_types
        # This selector exists to ensure we hit every node once before starting over.
        self._node_sel = selectors.ClusteredNodeSelector(self._cluster_dm, num=1, always_unique=True)

    def _apply_continuous(self):
        try:
            fixed_node_sel = selectors.FixedNodeSelector(node_dms=self._node_sel())
        except selectors.NotEnoughItemsError:
            self._node_sel = selectors.ClusteredNodeSelector(self._cluster_dm, num=1, always_unique=True)
            fixed_node_sel = selectors.FixedNodeSelector(node_dms=self._node_sel())

        service_sel = selectors.ServiceSelector(self._cluster_dm,
                                                service_types=self._service_types,
                                                node_selector=fixed_node_sel)
        fixed_service_sel = selectors.FixedServiceSelector(services=service_sel())
        service_dms = fixed_service_sel()
        restart_services_cmt = RestartNodeServices(self._cluster_dm, fixed_service_sel)
        restart_services_cmt.apply()

        service_ids = [service_dm.service.service_id for service_dm in service_dms]
        vl_helpers.wait_for_healthy_services(self._cluster_fm, service_ids)

    def _reset(self, timeout=None):
        pass


class _StopMasterService(_ManipulateMasterService, components.Reversible):
    """A reversible component which stops the master service on a node.

        The master service is killed on the node with the unix signal '3',
        This should not be called directly, use the parent component StopNodeServices

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which the master
        service has to be stopped
    :param valence.models.ResourceNode node_dm: The node dm where the master service
        has to be stopped
    """

    def __init__(self, cluster_dm, node_dm):
        super(_StopMasterService, self).__init__(cluster_dm, node_dm, signal=enums.UnixSignal.SIGQUIT)


class _StopDriveService(_ManipulateDriveService, components.Reversible):
    """A reversible component which stops the block/slice service on a node.

        The block/slice service is killed on the node with the unix signal '3'.
        This should not be called directly use the parent component StopNodeService.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm to which the node belongs
    :param valence.models.ResourceNode node_dm: The node dm where the service is to be stopped
    :param valence.models.ClusterServiceInfo service_dm: The service data model of the service,
            which is to be stopped
    """
    def __init__(self, cluster_dm, node_dm, service_dm):
        super(_StopDriveService, self).__init__(cluster_dm, node_dm, service_dm,
                                                signal=enums.UnixSignal.SIGQUIT)


class StopNodeServices(components.Reversible, components.BlockingApply):
    """A reversible component which restarts the master, block or slice services.

        _StopMasterService is called for master services and
        _StopDriveServices is called for slice and block services.
        This is done for all services returned by the service_selector.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which
        the services have to be stopped
    :param valence.selectors.ServiceSelector service_selector: The service selector of the services
        that need to be stopped
    """

    def __init__(self, cluster_dm, service_selector):
        super(StopNodeServices, self).__init__()
        self._cluster_dm = cluster_dm
        self._service_selector = service_selector
        self._stop_service_cmts = []

    @property
    def node_dms(self):
        """Returns a list of nodes that are being manipulated by the component.

        :returns: List of nodes that are being manipulated by the component.
        :rtype: list[valence.models.ResourceNode]
        """

        node_dms = []
        for cmt in self._stop_service_cmts:
            node_dms.append(cmt.node_dm)
        return node_dms

    def _apply_setup(self):
        """Initializes _StopDriveService and _StopMasterService components."""
        self._stop_service_cmts = []
        services = self._service_selector()
        for s in services:
            node_dm = models.get_resource_node_by_static_id(s.node.static_id)

            if s.service.service_type is enums.ServiceType.master:
                stop_master_service_cmt = _StopMasterService(self._cluster_dm, node_dm)
                self._stop_service_cmts.append(stop_master_service_cmt)

            elif s.service.service_type.is_drive_service:
                stop_drive_service_cmt = _StopDriveService(self._cluster_dm, node_dm, s)
                self._stop_service_cmts.append(stop_drive_service_cmt)

    def _apply(self, timeout=None):
        for cmt in self._stop_service_cmts:
            cmt.apply()

    def _reset(self, timeout=None):
        for cmt in reversed(self._stop_service_cmts):
            cmt.reset()


class MockGarbageCollection(components.MockComponentMixin, components.Transient,
                            components.NonBlockingApply):
    """A mock garbage collection component that does nothing.

    See the documentation on MockComponentMixin for more details about correct usage.
    """
    TIMEOUT = 30 * 60


class RunGarbageCollection(components.NonBlockingApply, components.Transient):
    """Component used to start garbage collection on a cluster.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which to run GC.
    """
    # TODO: ELEM-6788 Make GC timeout scale with capacity utilization
    TIMEOUT = 60 * 60

    def __init__(self, cluster_dm):
        super(RunGarbageCollection, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._async_handler_fm = None
        self._disable_automatic_gc = _constants.DisableAutomaticGarbageCollection(self._cluster_dm)

    @staticmethod
    def mock(cluster_dm):  # pylint: disable=unused-argument
        """Mock the RunGarbageCollection.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._cluster.RunGarbageCollection',
                        vl_components.RunGarbageCollection.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._cluster.MockGarbageCollection
        """
        return MockGarbageCollection()

    def get_async_handler(self):
        """Return the internal AsyncHandler object.

        :rtype: valence.functional_models.AsyncHandler
        """
        return self._async_handler_fm

    def get_gc_status(self):
        """Get the current GC status, which is separate from the AsyncResult status.

        :raises valence.ElementVersionError: Version too old to get GC status.
        :raises valence.components.ComponentStateError: Status not available when in ready state.
        :rtype: valence.enums.GcStatus
        """
        if self.state is not components.States.ready:
            async_result = self._async_handler_fm.get_async_result()
            if isinstance(async_result.details, models.GarbageCollectionDetails):
                return async_result.details.status
            else:
                raise errors.ElementVersionError('status only available in 12.0 and newer endpoints.')
        else:
            raise components.ComponentStateError('Gc has not started and thus has no status.')

    def _apply_async(self, timeout=None):
        """Run garbage collection.

        :param int timeout: Timeout in seconds.
        """
        self._disable_automatic_gc.apply()
        self._cluster_fm.abort_gc()
        self._async_handler_fm = self._cluster_fm.start_gc(timeout or self.TIMEOUT)

    def _apply_is_running(self):
        if self._async_handler_fm.is_running():
            # A bug in GCv1 can cause the async handler to get stuck in certain situations
            # (e.g. after drive addition)
            # Since SkipGC is set, we can check GetGCStatus to see if GC is actually still in progress
            if self._cluster_fm.get_gc_status().status is enums.GcStatus.in_progress:
                return True

        return False

    def _apply_finish(self, timeout=None):
        super(RunGarbageCollection, self)._apply_finish(timeout)
        self._disable_automatic_gc.reset()

    def _reset(self, timeout=None):
        """No reset required for GC -- docstring necessary for unittest coverage.

        :param int timeout: Optional timeout in seconds.
        """
        self._disable_automatic_gc.ensure_ready()


class RunClusterBSCheck(components.NonBlockingApplyKillable, components.Transient):
    """Component used to run cluster_bs_check and set appropriate cluster constants.

    :param valence.models.ResourceCluster cluster_dm: Cluster to run bs_check on.
    :param bool full_throttle: Remove Built-in pause between checks. Only use when cluster is
                               otherwise idle.
    :param bool heavy: Resource intensive check. If false check if blockIDs are in BSmap only. If true
                       check if blockIDs are in maps, set blocks in use, and verify blockID matches
                       disk segment.
    :param list[int] slice_ids: list of one or more slice_ids to run BS Check against
    :param list[valence.models.Volume] volume_dms: the volumes to run BS Check against
    """
    TIMEOUT = 1800

    def __init__(self, cluster_dm, full_throttle=False, heavy=False,
                 slice_ids=None, volume_dms=None):
        super(RunClusterBSCheck, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._prior_event_id = None
        self._async_handler_fm = None
        self.slice_ids = slice_ids
        self.volume_dms = volume_dms
        constants_dm = models.ClusterConstants.create(
            bs_check_interval_seconds=0,
            bs_check_pause_msec=0 if full_throttle else None,
            cluster_bs_check_mode=1 if heavy else None
        )
        self._set_constants_cmt = _constants.SetClusterConstants(self.cluster_dm, constants_dm)

    def get_unknown_blocks(self):
        """Returns a unique list of dictionaries where each has a key for volumeID, LBA, and blockID.

        :rtype: list[dict]
        """
        bs_check_events = self.cluster_fm.list_events(
            start_event_id=self._prior_event_id, event_type=enums.ClusterEventType.data_event)
        unknown_events = [event for event in bs_check_events if event.details['source'] == 'bscheck']

        # FillSlice can create false positive xUnknownBlockIDs. Let's filter them out.
        polluted_blocks = []
        unknown_block = collections.namedtuple('unknown_block', 'volumeID lba blockID')
        unknown_blocks = set()
        for event in unknown_events:
            try:
                block_id, volume_id, lba = event.message.split()
            except ValueError:
                # Some special variations of bs_check have an additional flag at the front
                _, block_id, volume_id, lba = event.message.split()
            volume = self.cluster_fm.list_volumes(volume_status=enums.VolumeStatus.active,
                                                  volume_ids=[volume_id])[0]
            fill_slice_status = volume.attributes.get('fillSliceStatus')
            if fill_slice_status in {enums.AsyncResultStatus.running.value,
                                     enums.AsyncResultStatus.failed.value}:
                polluted_blocks.append({'volumeID': volume_id, 'lba': lba, 'blockID': block_id})
            else:
                unknown_blocks.add(unknown_block(volumeID=volume_id, lba=lba, blockID=block_id))
        if polluted_blocks:
            msg = 'The following blocks were corrupted by FillSlice, ignoring: {}'
            _log.debug(msg.format(polluted_blocks))
        return [dict(single_block._asdict()) for single_block in unknown_blocks]

    def _apply_async(self, timeout=None):
        timeout = timeout if timeout else self.TIMEOUT
        self._set_constants_cmt.apply()
        self._prior_event_id = self.cluster_fm.list_events()[0].event_id
        self.cluster_fm.stop_cluster_bs_check()
        self._async_handler_fm = self.cluster_fm.start_cluster_bs_check(timeout=timeout,
                                                                        slice_ids=self.slice_ids,
                                                                        volume_dms=self.volume_dms)

    def _apply_is_running(self):
        return self._async_handler_fm.is_running()

    def _kill_apply_async(self):
        self.cluster_fm.stop_cluster_bs_check()

    def _apply_finish(self, timeout=None):
        super(RunClusterBSCheck, self)._apply_finish(timeout)

        # Revert any modified constants to their initial value.
        self._set_constants_cmt.reset()

        # Check if ClusterBsCheck failed.
        bs_check_failed = False
        async_result = self._async_handler_fm.get_async_result()
        if async_result.status is not enums.AsyncResultStatus.complete:
            bs_check_failed = True
            msg = 'BSCheck failed. Message: {}'.format(
                async_result)
        else:
            msg = 'BSCheck completed successfully.'
        # Raise error if any xUnknownBlockIDs are found, or BsCheck failed.
        unknown_blocks = self.get_unknown_blocks()
        if unknown_blocks:
            raise components.UnknownBlockIDError(msg, unknown_blocks)
        if bs_check_failed:
            raise components.ComponentAsyncError(msg)

    def _reset(self, timeout=None):
        # Revert any modified constants to their initial value.
        self._set_constants_cmt.ensure_ready()


class FlushPcache(components.BlockingApply, components.Transient):
    """Component used to flush pcache.

    :param valence.models.ResourceCluster cluster_dm: Cluster to flush pcache on.
    """

    def __init__(self, cluster_dm):

        super(FlushPcache, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        constants_dm = models.ClusterConstants.create(
            cPCacheIgnoreMinFlushSizeIntervalSeconds=1,
            cDisablePCacheCleanFiles=True
        )
        self._set_constants_cmt = _constants.SetClusterConstants(self.cluster_dm, constants_dm)
        self._gc_cmt = RunGarbageCollection(cluster_dm)

    def _apply(self, timeout=None):
        """Flush the primary cache.

        :param int timeout: Time in seconds after which to raise an error.
        """
        self._set_constants_cmt.apply()
        timeout = timeout if timeout is not None else 600
        time_start = time.time()
        if vl_helpers.is_pcache_empty(self.cluster_fm):
            pass
        else:
            _log.debug('Waiting for cluster to be idle...')
            vl_helpers.wait_for_cluster_idle(self.cluster_fm, timeout=timeout)

            _log.debug('Flushing cluster pcache by rebooting slice services...')
            timeout -= time.time() - time_start
            vl_helpers.reboot_slice_services(self.cluster_fm, timeout=timeout)

            time_start = time.time()
            while not vl_helpers.is_pcache_empty(self.cluster_fm):
                if time.time() - time_start > timeout:
                    raise errors.TimeoutError(
                        'Timed out after {}s waiting for cluster pcache to flush'.format(timeout)
                    )
                else:
                    time.sleep(5)

        _log.debug('Running garbage collection to update cluster info...')
        timeout -= time.time() - time_start
        self._gc_cmt.apply(timeout)
        self._gc_cmt.reset()
        self._set_constants_cmt.reset()

    def _reset(self, timeout=None):
        # Revert any modified constants to their initial value.
        self._set_constants_cmt.ensure_ready()
        self._gc_cmt.ensure_ready()
