"""Valence connections components."""
from __future__ import division

import abc
import collections
import itertools
import string
import time

import autologger
import six

from valence import (
    components,
    enums,
    functional_models,
    models,
)
from valence.vl_components import (
    _account,
    _initiator,
    _volume,
    _volume_access_group,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class ConnectionsInterface(components.Reversible, components.BlockingApply):
    """Interface component for a set of volume-client connections."""

    @property
    @abc.abstractmethod
    def client_disk_map(self):
        """Property to get the resulting client disk map.

        :returns: A dictionary which maps the client to it's disk map.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {client.mip: {1: '/dev/sda2'}}

            The pseudo type could be thought of as:

                dict[client.mip, dict[volume_id, disk_path]]

        :rtype: dict[str, dict[int, str]]
        """

    @property
    @abc.abstractmethod
    def child_cmts(self):
        """Get all connection-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components._connection.BaseCreateConnection]
        """

    @property
    @abc.abstractmethod
    def client_dms(self):
        """Return all clients associated with this component.

        :rtype: list[valence.models.ResourceClient]
        """


@six.add_metaclass(abc.ABCMeta)
class BaseCreateConnection(ConnectionsInterface):
    """Base component for creating volume-client connections for a single client.

    :param valence.models.ResourceClient client_dm: Client data model to connect with.
    :param list[valence.vl_components._volume.VolumesInterface] volumes_cmts: The volumes
        associated with this connection.
    :param int jvm_count: Number of java virtual machines to spin up for driving IO.
    """
    def __init__(self, client_dm, volumes_cmts, jvm_count=None):
        super(BaseCreateConnection, self).__init__()
        self._volumes_cmts = volumes_cmts
        self._client_dm = client_dm
        self.jvm_count = jvm_count

    @property
    @abc.abstractmethod
    def disk_map(self):
        """Property to get the resulting disk map.

        :returns: A dictionary which represents the disk map on the client.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {1: '/dev/sda2'}

            The pseudo type could be thought of as:

                dict[volume_id, disk_path]

        :rtype: dict[int, str]
        """

    @property
    def client_dm(self):
        """Return the client datamodel.

        Elementary Connections are to only a single client by definition.

        :rtype: valence.models.ResourceClient
        """
        return self._client_dm

    @property
    def volumes_cmts(self):
        """Property to get the volumes components associated with this connection.

        :rtype: list[valence.vl_components._volume.VolumesInterface]
        """
        return self._volumes_cmts

    @property
    def client_dms(self):
        """Return clients associates with this component.

        :rtype: list[valence.models.ResourceClient]
        """
        return [self.client_dm]

    @property
    def child_cmts(self):
        """Get all connection-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components._connection.BaseCreateConnection]
        """
        return [self]

    @property
    def client_disk_map(self):
        """Property to get the resulting client disk map.

        :returns: A dictionary which maps the client to it's disk map.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {client.mip: {1: '/dev/sda2'}}

            The pseudo type could be thought of as:

                dict[client.mip, dict[volume_id, disk_path]]

        :rtype: dict[str, dict[int, str]]
        """
        client_ip = self.client_dm.mip
        return {client_ip: self.disk_map}

    def _apply_volumes_and_get_ids(self):
        """Helper method.

        - calls ensure_applied on all self.volumes_cmts
        - returns their associated volume_ids

        :rtype: set[int]
        """
        volume_ids = set()
        for vol_cmt in self.volumes_cmts:
            vol_cmt.ensure_applied()
            for vol in vol_cmt.volume_dms:
                volume_ids.add(vol.volume_id)
        return volume_ids


# TODO: Determine how to purge connections for component reset (VL-2044)
@six.add_metaclass(abc.ABCMeta)
class BaseIscsiConnection(BaseCreateConnection):
    """Base iscsi connection components."""


class CreateIscsiAccountConnection(BaseIscsiConnection):
    """Create ISCSI connections to all volumes attached to a particular account.

    .. note:: All volumes attached to the account will be connected, but only the specified volumes will
        be included in the exported disk map (and receive IO if used in conjunction with the vdbench
        components).

    :param valence.models.ResourceCluster cluster_dm: Cluster data model, needs volumes to connect to.
    :param valence.models.ResourceClient client_dm: Client data model to connect with.
    :param valence.vl_components.AccountsInterface accounts_cmt: Account component to
        use, must be only one account.
    :param list[valence.vl_components.VolumesInterface] volumes_cmts: Volumes
        associated with this connection to include in the disk map.
    :param int account_index: The index of the account to be used in the list returned by
        accounts_cmt.get_account_dms()
    :param int jvm_count: Number of java virtual machines to spin up for driving IO.
    """

    def __init__(self, cluster_dm, client_dm, volumes_cmts, accounts_cmt, account_index=0,
                 jvm_count=None):
        super(CreateIscsiAccountConnection, self).__init__(client_dm, volumes_cmts, jvm_count=jvm_count)
        self.cluster_dm = cluster_dm
        self.accounts_cmt = accounts_cmt
        self._account_index = account_index
        self._disk_map = None
        self._cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        self._client_fm = functional_models.get_client_from_dm(self.client_dm)

    @staticmethod
    def mock(cluster_dm, client_dm, volumes_cmts, accounts_cmt,  # pylint: disable=unused-argument
             jvm_count=None):
        """Mock the CreateIscsiAccountConnection.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._connection.CreateIscsiAccountConnection',
                        vl_components.CreateIscsiAccountConnection.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._connection.MockConnection
        """
        drive_letters = itertools.cycle(string.ascii_lowercase)
        fake_volume_id = 0

        disk_map = {}
        for volumes_cmt in volumes_cmts:
            volume_dms = volumes_cmt.partial_volume_dms
            for volume_dm in volume_dms:
                # Get a volume ID to use
                volume_id = volume_dm.volume_id
                if volume_id is None:
                    while fake_volume_id in disk_map:
                        fake_volume_id += 1
                    volume_id = fake_volume_id

                # Give the volume a fake disk location
                disk_map[volume_id] = '/dev/sd{}'.format(next(drive_letters))

        return MockConnection(disk_map, client_dm, volumes_cmts, jvm_count=jvm_count)

    @property
    def disk_map(self):
        """Property to get the resulting disk map.

        :returns: A dictionary which represents the disk map on the client.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {1: '/dev/sda2'}

            The pseudo type could be thought of as:

                dict[volume_id, disk_path]

        :rtype: dict[int, str]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._disk_map

    def _apply(self, timeout=None):
        """Set the client disk_map.

        :param int timeout: Maximum time to apply.
        """
        # Only for UbuntuClient
        # TODO: A bit hacky, but need to make sure we only auth to one account until we have VAGs
        self.accounts_cmt.ensure_applied()
        volume_ids = self._apply_volumes_and_get_ids()

        account_dm = self.accounts_cmt.get_account_dms()[self._account_index]
        account_name = account_dm.username
        account_initiator_secret = account_dm.initiator_secret
        self._client_fm.update_iscsi_d_conf(
            auth_user=account_name,
            auth_pass=account_initiator_secret,
            double_authentication=False
        )
        disk_map = self._client_fm.volume_ceremony(self.cluster_dm.svip, purge=True)
        disk_map = convert_disk_map(disk_map)
        self._disk_map = filter_disk_map(disk_map, volume_ids)

    def _reset(self, timeout=None):
        """Delete the client disk_map.

        :param int timeout: Maximum time to reset.
        """
        self._disk_map = None


class CreateIscsiVagConnection(BaseIscsiConnection):
    """Create ISCSI connections to all volumes of a volume access group.

    :param valence.models.ResourceCluster cluster_dm: Cluster of the volume access group.
    :param valence.models.ResourceClient client_dm: Client with which to connect.
    :param list[valence.vl_components.VolumeAccessGroupInterface] vag_cmts: Volume access groups for the
        volumes to be connected.
    :param int jvm_count: Number of java virtual machines to spin up for driving IO.
    """

    def __init__(self, cluster_dm, client_dm, vag_cmts, jvm_count=None):
        super(CreateIscsiVagConnection, self).__init__(
            client_dm=client_dm,
            volumes_cmts=[cmt.volumes_cmt for cmt in vag_cmts],
            jvm_count=jvm_count
        )
        self.cluster_dm = cluster_dm
        self.vag_cmts = vag_cmts
        self._disk_map = None
        self._cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        self._client_fm = functional_models.get_client_from_dm(self.client_dm)

        # Ensure each VAG has initiators for the specified client
        assert all([client_dm in cmt.client_dms for cmt in self.vag_cmts])
        assert all([cmt.initiators_cmt is not None for cmt in self.vag_cmts])
        assert all([client_dm in cmt.initiators_cmt.client_dms for cmt in self.vag_cmts])

    @staticmethod
    def mock(cluster_dm, client_dm, vag_cmts, jvm_count):  # pylint: disable=unused-argument
        """Mock the CreateIscsiVagConnection.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._connection.CreateIscsiVagConnection',
                        vl_components.CreateIscsiVagConnection.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._connection.MockConnection
        """
        drive_letters = itertools.cycle(string.ascii_lowercase)
        fake_volume_id = 0

        disk_map = {}
        volumes_cmts = [cmt.volumes_cmt for cmt in vag_cmts]
        for volumes_cmt in volumes_cmts:
            volume_dms = volumes_cmt.partial_volume_dms
            for volume_dm in volume_dms:
                # Get a volume ID to use
                volume_id = volume_dm.volume_id
                if volume_id is None:
                    while fake_volume_id in disk_map:
                        fake_volume_id += 1
                    volume_id = fake_volume_id

                # Give the volume a fake disk location
                disk_map[volume_id] = '/dev/sd{}'.format(next(drive_letters))

        return MockConnection(disk_map, client_dm, volumes_cmts, jvm_count=jvm_count)

    @property
    def disk_map(self):
        """Property to get the resulting disk map.

        :returns: A dictionary which represents the disk map on the client.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {1: '/dev/sda2'}

            The pseudo type could be thought of as:

                dict[volume_id, disk_path]

        :rtype: dict[int, str]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._disk_map

    def _apply(self, timeout=None):
        """Set the client disk_map.

        :param int timeout: N/A
        """
        for cmt in self.vag_cmts:
            cmt.ensure_applied()
        volume_ids = self._apply_volumes_and_get_ids()

        self._client_fm.deactivate_chap()
        disk_map = self._client_fm.volume_ceremony(self.cluster_dm.svip, purge=True)
        disk_map = convert_disk_map(disk_map)
        self._disk_map = filter_disk_map(disk_map, volume_ids)
        self._client_fm.reactivate_chap()

    def _reset(self, timeout=None):
        """Delete the client disk_map.

        :param int timeout: Maximum time to reset.
        """
        self._client_fm.reactivate_chap()
        self._disk_map = None

    def _apply_volumes_and_get_ids(self):
        """Helper method.

        - calls ensure_applied on all self.volumes_cmts
        - returns their associated volume_ids

        :rtype: set[int]
        """
        volume_ids = set()
        for vol_cmt in self.volumes_cmts:
            vol_cmt.ensure_applied()
            for vol in vol_cmt.volume_dms:
                volume_ids.add(vol.volume_id)
        return volume_ids


class ManageConnections(ConnectionsInterface):
    """Manage a set of connection components together.

    Can manage connections across multiple clients, using either ISCSI or FibreChannel.

    :param list[valence.vl_components.ConnectionsInterface] connection_cmts: List of connections to
        manage.
    """

    def __init__(self, connection_cmts):
        super(ManageConnections, self).__init__()
        self.connection_cmts = connection_cmts
        self.accounts_cmts = []
        self.volumes_cmts = []
        self.initiators_cmts = []
        self.vag_cmts = []

        all_client_mips = set()
        for conn_cmt in self.connection_cmts:
            client_mips = {dm.mip for dm in conn_cmt.client_dms}
            overloaded_mips = client_mips.intersection(all_client_mips)
            if overloaded_mips:
                raise ValueError(
                    'Multiple connection components specified for the same client '
                    '(only one connection per client supported): {}'.format(overloaded_mips)
                )
            all_client_mips.update(client_mips)

    @classmethod
    def from_client_dms(cls, tag, cluster_dm, client_dms, volume_count, connections_per_volume=1,
                        jvm_count=None):
        """Get a set of connections for the specified clients.

        .. note:: Creates and manages accounts/volumes/initiators/volume-access-groups for you.

        :param str tag: Unique tag for all the managed components to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which the volumes reside.
        :param list[valence.models.ResourceClient] client_dms: Clients to connect.
        :param int volume_count: Number of volumes to create.
        :param int connections_per_volume: Number of clients to connect to each volume (default: 1).
        :param int jvm_count: Number of java virtual machines to spin up for driving IO.
        :rtype: valence.vl_components.ManageConnections
        """
        accounts_cmt = _account.CreateAccounts(tag=tag, cluster_dm=cluster_dm)
        volume_dm = models.Volume.create(total_size=models.DataSize(8, enums.DataUnit.gibibytes))

        volumes_cmts = []
        volume_fraction = volume_count // len(client_dms)
        volume_remainder = volume_count % len(client_dms)
        for i in range(len(client_dms)):
            cmt_volume_count = volume_fraction
            if volume_remainder:
                cmt_volume_count += 1
                volume_remainder -= 1
            if cmt_volume_count:
                volumes_cmts.append(_volume.CreateMultipleVolumes(
                    tag='{}-{}'.format(tag, i),
                    cluster_dm=cluster_dm,
                    vol_count=cmt_volume_count,
                    partial_volume_dm=volume_dm,
                    accounts_cmt=accounts_cmt
                ))

        connections_manager = cls._create_connections_from_volumes(
            tag=tag,
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            volumes_cmts=volumes_cmts,
            connections_per_volume=connections_per_volume,
            jvm_count=jvm_count
        )
        connections_manager.accounts_cmts = [accounts_cmt]
        connections_manager.volumes_cmts = volumes_cmts
        return connections_manager

    @classmethod
    def from_volumes_cmts(cls, tag, cluster_dm, client_dms, volumes_cmts, connections_per_volume=1,
                          jvm_count=None):
        """Get a set of connections for the specified clients and volumes.

        .. note:: Creates and manages initiators/volume-access-groups for you.
        .. note:: Volumes components passed in must be managed (reset) manually.

        :param str tag: Unique tag for all the managed components to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which the volumes reside.
        :param list[valence.models.ResourceClient] client_dms: Clients to connect.
        :param list[valence.vl_components.VolumesInterface] volumes_cmts: Volumes to connect.
        :param int connections_per_volume: Number of clients to connect to each volume (default: 1).
        :param int jvm_count: Number of java virtual machines to spin up for driving IO.
        :rtype: valence.vl_components.ManageConnections
        """
        return cls._create_connections_from_volumes(
            tag=tag,
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            volumes_cmts=volumes_cmts,
            connections_per_volume=connections_per_volume,
            jvm_count=jvm_count
        )

    @classmethod
    def _create_connections_from_volumes(cls, tag, cluster_dm, client_dms, volumes_cmts,
                                         connections_per_volume=1, jvm_count=None):
        """Get a set of connections for the specified clients and volumes.

        .. note:: Creates and manages initiators/volume-access-groups for you.
        .. note:: Volumes components passed in must be managed (reset) manually.

        :param str tag: Unique tag for all the managed components to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which the volumes reside.
        :param list[valence.models.ResourceClient] client_dms: Clients to connect.
        :param list[valence.vl_components.VolumesInterface] volumes_cmts: Volumes to connect.
        :param int connections_per_volume: Number of clients to connect to each volume (default: 1).
        :param int jvm_count: Number of java virtual machines to spin up for driving IO.
        :rtype: valence.vl_components.ManageConnections
        """
        if cluster_dm.api_endpoint < models.SfVersion('9.0'):
            return cls._create_legacy_connections_from_volumes(
                cluster_dm=cluster_dm,
                client_dms=client_dms,
                volumes_cmts=volumes_cmts,
                jvm_count=jvm_count
            )

        if connections_per_volume > len(client_dms):
            raise ValueError(
                'Connections-per-volume[{}] is too high to form the specified number of '
                'connections with each volumes component given the number of clients: {}'
                .format(connections_per_volume, len(client_dms))
            )

        child_volumes_cmts = []
        for cmt in volumes_cmts:
            child_volumes_cmts.extend(cmt.child_cmts)
        if len(child_volumes_cmts) * connections_per_volume < len(client_dms):
            raise ValueError(
                'Connections-per-volume[{}] is too low to form connections with '
                'every client given the number of volumes components: {}'
                .format(connections_per_volume, len(child_volumes_cmts))
            )

        # Create initiators for unique client sets
        initiators_cmts = cls._create_connection_initiators(
            tag=tag,
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            max_initiator_count=min(len(client_dms), len(child_volumes_cmts)),
            clients_per_initiator=connections_per_volume
        )
        initiators_cmt_map = {dm.mip: [] for dm in client_dms}
        for initiators_cmt in initiators_cmts:
            for client_dm in initiators_cmt.client_dms:
                initiators_cmt_map[client_dm.mip].append(initiators_cmt)

        # Create volume access groups for each connection group
        vag_cmts = cls._create_connection_vags(
            cluster_dm=cluster_dm,
            initiators_cmts=initiators_cmts,
            volumes_cmts=child_volumes_cmts
        )

        # Create connections for each client
        connection_cmts = []
        for client_dm in client_dms:
            conn_init_cmts = initiators_cmt_map[client_dm.mip]
            connection_cmts.append(CreateIscsiVagConnection(
                cluster_dm=cluster_dm,
                client_dm=client_dm,
                vag_cmts=[cmt for cmt in vag_cmts if cmt.initiators_cmt in conn_init_cmts],
                jvm_count=jvm_count
            ))

        connections_manager = cls(connection_cmts)
        connections_manager.initiators_cmts = initiators_cmts
        connections_manager.vag_cmts = vag_cmts
        return connections_manager

    @classmethod
    def _create_legacy_connections_from_volumes(cls, cluster_dm, client_dms, volumes_cmts,
                                                jvm_count=None):
        """Get a set of connections for the specified clients and volumes.

        .. note:: This is for support of pre-Fluorine ONLY. Creates connections based on accounts
            instead of volume access groups because they weren't well supported yet.
        .. note:: This method only supports a limited set of volumes components (CreateVolumes,
            CreateMultipleVolumes, CloneVolumes). No other volumes components are supported.

        :param valence.models.ResourceCluster cluster_dm: The cluster on which the volumes reside.
        :param list[valence.models.ResourceClient] client_dms: Clients to connect.
        :param list[valence.vl_components.VolumesInterface] volumes_cmts: Volumes to connect.
        :param int jvm_count: Number of java virtual machines to spin up for driving IO.
        :rtype: valence.vl_components.ManageConnections
        """
        _log.warning('Using legacy methods to construct client connections')
        child_volumes_cmts = []
        for cmt in volumes_cmts:
            child_volumes_cmts.extend(cmt.child_cmts)

        accounts_cmts = {}
        account_volume_map = collections.defaultdict(list)
        for vol_cmt in child_volumes_cmts:
            accounts_cmts[vol_cmt.accounts_cmt.tag] = vol_cmt.accounts_cmt
            account_key = '{}-{}'.format(
                vol_cmt.accounts_cmt.tag,
                vol_cmt._account_index  # pylint: disable=protected-access
            )
            account_volume_map[account_key].append(vol_cmt)

        connection_cmts = []
        client_dms = itertools.cycle(client_dms)
        for accounts_cmt in accounts_cmts.values():
            for account_index in range(accounts_cmt.num):
                account_key = '{}-{}'.format(accounts_cmt.tag, account_index)
                conn_vol_cmts = account_volume_map.get(account_key)
                if conn_vol_cmts:
                    connection_cmts.append(CreateIscsiAccountConnection(
                        cluster_dm=cluster_dm,
                        client_dm=next(client_dms),
                        volumes_cmts=conn_vol_cmts,
                        accounts_cmt=accounts_cmt,
                        account_index=account_index,
                        jvm_count=jvm_count
                    ))

        connections_manager = cls(connection_cmts)
        return connections_manager

    @staticmethod
    def _create_connection_initiators(tag, cluster_dm, client_dms, max_initiator_count,
                                      clients_per_initiator):
        """Create initiators components grouping the specified clients into unique sets.

        :param str tag: Unique tag to use for these components.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which to create the initiators.
        :param list[valence.models.ResourceClient] client_dms: Clients to use.
        :param int max_initiator_count: Maximum number of initiators to create (also limited by the
            number of unique client sets).
        :param int clients_per_initiator: Number of clients to put on each initiator.
        :rtype: list[valence.vl_components.InitiatorsInterface]
        """
        # Ensure clients_per_initiator does not exceed the number of clients provided
        assert clients_per_initiator <= len(client_dms)

        # Ensure each client can be included in at least one initiator
        assert clients_per_initiator * max_initiator_count >= len(client_dms)

        initiators_cmts = []
        clients_keys = []
        client_iterator = itertools.cycle(client_dms)
        for i in range(max_initiator_count):
            initiator_client_dms = [next(client_iterator) for _ in range(clients_per_initiator)]
            clients_key = '_'.join([dm.client_name for dm in initiator_client_dms])
            if clients_key not in clients_keys:
                clients_keys.append(clients_key)
                initiators_cmts.append(_initiator.CreateIscsiInitiators(
                    tag='{}-{}'.format(tag, i),
                    cluster_dm=cluster_dm,
                    client_dms=initiator_client_dms
                ))
            else:
                break

        return initiators_cmts

    @staticmethod
    def _create_connection_vags(cluster_dm, initiators_cmts, volumes_cmts):
        """Create volume access group components defining a set of connection groups.

        Each connection group represents a set of volumes to be connected to a set of clients as
        defined by the specified initiators.

        :param valence.models.ResourceCluster cluster_dm: The cluster on which to create the VAGs.
        :param list[valence.vl_components.InitiatorsInterface] initiators_cmts: Initiators to use.
        :param list[valence.vl_components.VolumesInterface] volumes_cmts: Volumes to use.
        :rtype: list[valence.vl_components.VolumeAccessGroupInterface]
        """
        # Ensure each initiators component can be assigned volumes without assigning volumes to multiple
        # initiators components (maintains the desired number of connections per volume)
        assert len(volumes_cmts) >= len(initiators_cmts)

        # Map each volumes component to an initiators component
        volumes_cmt_map = {cmt.tag: [] for cmt in initiators_cmts}
        initiators_iterator = itertools.cycle(initiators_cmts)
        for volumes_cmt in volumes_cmts:
            initiators_cmt = next(initiators_iterator)
            volumes_cmt_map[initiators_cmt.tag].append(volumes_cmt)

        # Create volume access groups for each connection group
        vag_cmts = []
        for initiators_cmt in initiators_cmts:
            vag_cmts.append(_volume_access_group.CreateVolumeAccessGroup(
                name=initiators_cmt.tag,
                cluster_dm=cluster_dm,
                initiators_cmt=initiators_cmt,
                volumes_cmt=_volume.ManageVolumes(volumes_cmt_map[initiators_cmt.tag])
            ))

        return vag_cmts

    @property
    def client_dms(self):
        """Return all clients associated with this component.

        :rtype: list[valence.models.ResourceClient]
        """
        out = []
        for c in self.connection_cmts:
            out.extend(c.client_dms)
        return out

    @property
    def child_cmts(self):
        """Get all connection-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components._connection.BaseCreateConnection]
        """
        out = []
        for c in self.connection_cmts:
            out.extend(c.child_cmts)
        return out

    @property
    def client_disk_map(self):
        """Property to get the resulting client disk map.

        :returns: A dictionary which maps the client to it's disk map.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {client.mip: {1: '/dev/sda2'}}

            The pseudo type could be thought of as:

                dict[client.mip, dict[volume_id, disk_path]]

        :rtype: dict[str, dict[int, str]]
        """
        disk_map = {}
        for c in self.connection_cmts:
            for client, cl_disk_map in six.iteritems(c.client_disk_map):
                if client not in disk_map:
                    disk_map[client] = {}
                disk_map[client].update(cl_disk_map)
        return disk_map

    def _apply(self, timeout=None):  # pylint: disable=too-complex
        """Call apply on all connections components.

        :param int timeout: Maximum time to apply.
        """
        for cmt in self.accounts_cmts:
            cmt.ensure_applied()
        for cmt in self.volumes_cmts:
            cmt.ensure_applied()
        for cmt in self.initiators_cmts:
            cmt.ensure_applied()
        for cmt in self.vag_cmts:
            cmt.ensure_applied()

        # Ensure all volumes are setup before attempting to apply connections in parallel
        # NOTE: Prevents multiple connections from calling ensure_applied()
        #       on the same volumes component at the same time
        for cmt in self.child_cmts:
            for volumes_cmt in cmt.volumes_cmts:
                volumes_cmt.ensure_applied()

        # Setup all connections in parallel
        time_start = time.time()
        for cmt in self.child_cmts:
            cmt.apply_async()
        for cmt in self.child_cmts:
            relative_timeout = int(timeout - (time.time() - time_start)) if timeout is not None else None
            cmt.finish(relative_timeout)

    def _reset(self, timeout=None):
        """Reset all connections components.

        :param int timeout: Maximum time to reset.
        """
        time_start = time.time()
        for cmt in self.child_cmts:
            relative_timeout = int(timeout - (time.time() - time_start)) if timeout is not None else None
            cmt.ensure_ready(relative_timeout)
        for cmt in self.vag_cmts:
            cmt.ensure_ready()
        for cmt in self.initiators_cmts:
            cmt.ensure_ready()
        for cmt in self.volumes_cmts:
            cmt.ensure_ready()
        for cmt in self.accounts_cmts:
            cmt.ensure_ready()


class MockConnection(components.MockComponentMixin, BaseCreateConnection):
    """A mock connection component representing a volume disk map.

    See the documentation on MockComponentMixin for more details about correct usage.

    :param dict[int,str] disk_map: the fake disk map that this component pretends to be.
    :param valence.models.ResourceClient client_dm: Client data model the disk_map is
        associated with.
    :param list[valence.vl_components.VolumesInterface] volumes_cmts: The volumes
        associated with this connection.
    :param int jvm_count: Number of java virtual machines to spin up for driving IO.
    """
    def __init__(self, disk_map, client_dm, volumes_cmts, jvm_count=None):
        super(MockConnection, self).__init__(client_dm, volumes_cmts, jvm_count=jvm_count)
        self._original_disk_map = disk_map
        self._disk_map = None

    @property
    def disk_map(self):
        """Property to get the resulting disk map.

        :returns: A dictionary which represents the disk map on the client.
            Each disk map is just a dictionary of the volume id to
            the path on the disk where it is connected. I.e.

                {1: '/dev/sda2'}

            The pseudo type could be thought of as:

                dict[volume_id, disk_path]

        :rtype: dict[int, str]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._disk_map

    def _apply(self, timeout=None):
        """Apply the sub components and filter the disk map.

        :param int timeout: N/A.
        """
        volume_ids = self._apply_volumes_and_get_ids()
        self._disk_map = filter_disk_map(self._original_disk_map, volume_ids)

    def _reset(self, timeout=None):
        """Clear the disk map.

        :param int timeout: N/A.
        """
        self._disk_map = None


def convert_disk_map(disk_map):
    """Convert the volume ids to integers.

    :param dict disk_map: Input disk_map dictionary, dict[str, str].

    :rtype: dict[int, str]
    """
    return {int(k): v for (k, v) in six.iteritems(disk_map)}


def filter_disk_map(disk_map, volume_ids):
    """Filter the disk map based on the volume ids.

    :param dict[int,str] disk_map: Input disk_map dictionary.
    :param set[int] volume_ids: the volume ids to keep.

    :rtype: dict[int, str]
    """
    return {vid: path for vid, path in disk_map.items() if vid in volume_ids}
