"""Valence cluster constant components."""
from valence import (
    components,
    functional_models,
    models,
)


class SetClusterConstants(components.Reversible, components.BlockingApply):
    """Set cluster constants.

    This component stores the value of the constants before it changes
    them. When it is reset the constants are returned to their original
    value.

    If multiple SetClusterConstants are applied, they must be reset the in the opposite order
    to return to the original state.


    :param valence.models.ResourceCluster cluster_dm: the cluster to set constants on
    :param valence.models.ClusterConstants constants_dm: the constants to set,
        None values in the data model are ignored
    """

    def __init__(self, cluster_dm, constants_dm):
        super(SetClusterConstants, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm)
        self._constants_dm = constants_dm
        self._original_constants = None

    def _apply(self, timeout=None):
        """Sets Cluster Constants.

        :param int timeout: Unused
        """
        orig_constants = self._cluster_fm.get_constants()
        self._original_constants = orig_constants
        self._cluster_fm.set_constants(self._constants_dm)

    def _reset(self, timeout=None):
        """Rolls back any changes made to cluster constants to the original values.

        :param int timeout: Unused
        """
        if self._original_constants:
            self._cluster_fm.set_constants(self._original_constants)
        self._original_constants = None


class DisableBinSync(SetClusterConstants):
    """Disables bin syncing.

    :param valence.models.ResourceCluster cluster_dm: The cluster on which to disable bin sync.
    """
    def __init__(self, cluster_dm):
        self.bin_sync_constants = models.ClusterConstants.create(cDisableBinAssignmentUpdates=True)
        super(DisableBinSync, self).__init__(cluster_dm, self.bin_sync_constants)


class DisableAutomaticSliceBalancing(SetClusterConstants):
    """Disables automatic slice balancing.

    :param valence.models.ResourceCluster cluster_dm: The cluster on which to disable automatic slice
     balancing.
    """

    def __init__(self, cluster_dm):
        self.slice_balance_constants = models.ClusterConstants.create(
            cSliceBalanceEnableAutoTimerIopsRebalancing=0)
        super(DisableAutomaticSliceBalancing, self).__init__(cluster_dm, self.slice_balance_constants)


class DisableAutomaticGarbageCollection(SetClusterConstants):
    """Disables automatic garbage collection.

    :param valence.models.ResourceCluster cluster_dm: The cluster on which to disable automatic garbage
     collection.
    """

    def __init__(self, cluster_dm):
        self.gc_constants = models.ClusterConstants.create(cSkipGC=1)
        super(DisableAutomaticGarbageCollection, self).__init__(cluster_dm, self.gc_constants)


class EnableMerkleChecker(SetClusterConstants):
    """Enables slice Merkle Checker.

    Merkle Checker monitors and checks for Slice Service(SS) metadata consistency.
    If inconsistencies are found, a core dump and log event will be generated.

    Find more here:
    https://solidfire.atlassian.net/wiki/spaces/SET/pages/225050794/Merkle+background+SS+metadata+checker

    :param valence.models.ResourceCluster cluster_dm: The cluster on which to disable slice rebalancing.
    """

    def __init__(self, cluster_dm):
        self.merkle_constants = models.ClusterConstants.create(
            merkle_checker_every_time=1,
            merkle_check_time=1800,
            merkle_idle_cycles=1,
            event_slice_lbas_max_mismatches=10,
            panic_on_mismatched_lbas=1,
            start_merkle_check_from_qos=1
        )
        super(EnableMerkleChecker, self).__init__(cluster_dm, self.merkle_constants)


class SetDriveServiceTimeout(SetClusterConstants):
    """Set the cluster timeout for unavailable drive services.

    :param valence.models.ResourceCluster cluster_dm: The cluster on which to set the service timeout.
    :param int timeout: Number of seconds before unavailable drive services will time out. If a service
        times out its data will be replicated elsewhere and the associated drive will be removed from
        the cluster.
    """

    def __init__(self, cluster_dm, timeout=24 * 60 * 60):
        if timeout % 30:
            raise ValueError(
                'Drive service timeout must be divisible by 30 because the slice service timeout is '
                'measured in 30s sample periods.'
            )

        # NOTE: Block-service timeout is measured in milliseconds
        # NOTE: Slice-service timeout is measured in health samples which are collected every 30 seconds
        self.service_timeout_constants = models.ClusterConstants.create(
            time_until_bs_is_killed_msec=timeout * 1000,
            ss_health_monitor_interval_for_permanence=timeout // 30
        )
        super(SetDriveServiceTimeout, self).__init__(cluster_dm, self.service_timeout_constants)
