"""Valence drive components."""
from __future__ import division

import abc
import time

import autologger
import six

from valence import (
    components,
    criteria,
    enums,
    functional_models,
    models,
    selectors,
    vl_helpers,
)
from valence.components import _errors
from valence.vl_components import _constants

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class DrivesInterface(components.Reversible):
    """Interface component for a set of drives."""

    @property
    @abc.abstractmethod
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """


class EnableMdss(components.BlockingApply, components.Reversible):
    """Enable multi drive slice service.

    Removes slot#2 drives, adds them back as volume type drives.
    apply and reset timeouts are passed through to each helper and do not act as an overall timeout, but
    rather as a per drive add / remove timeout - recommend ~ 1800 seconds

    See support documentation on MDSS:
        https://docs.google.com/document/d/1Yo1lWTi3pKmIORzByg0CFZKt6-9HMXTGuecXDsHKwf0/edit#

    Note: This component can only be applied on active cluster with all drives active.

    TODO: Remove sleep related to EOS-10102.  There is a timing issue when gracefully removing
    a drive from the cluster then immediately adding it back.  The new service cannot come up
    because the old one has not terminated.  This doesn't just apply to MDSS but seems to be brought
    about by applying this component on an empty cluster.

    :param valence.models.ResourceCluster cluster_dm: cluster data model
    """

    # TODO (VL-2583): Remove these variables when tests are updated to no longer reference this variable.
    # TIMEOUT cvar is deprecated and no longer used.
    TIMEOUT = 1800

    def __init__(self, cluster_dm):
        super(EnableMdss, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

        # These dicts contains mapping from node -> slice/block drive data models used by component
        #   We need this for non-homogenous clusters whose slot indexing might be different

        self._node_mdss_drive_dms = {}
        self._primary_slice_drives_dms = None
        self._slot_two_drives_dms = None
        self._slot_two_mdss_drives_dms = None
        self._active_nodes_dms = None
        self._removed_drive_dms = None
        self._mdss_constants_dm = models.ClusterConstants.create(
            enable_space_based_rebalancing=False,
            slice_balance_enable_auto_timer_iops_rebalancing=0,
            skip_gc=True
        )

        self._mdss_constants_cmt = _constants.SetClusterConstants(self._cluster_dm,
                                                                  self._mdss_constants_dm)

        self._active_node_selector = selectors.ActiveNodeSelector(self._cluster_dm)

        self._drive_selector = selectors.DriveSelector(self._cluster_dm,
                                                       node_selector=self._active_node_selector)

    def _get_default_slice_drive_slot_for_node_id(self, node_id):
        """Get default slice drive slot for the given node.

        This method is purposely avoiding API calls and using cached data already obtained by the
        component since the method is used in filters.

        :param int node_id: node id to use to look up the node_dms from cached data

        :rtype: int
        """
        active_node_dm = [node for node in self._active_nodes_dms if node.node_id == node_id]
        return vl_helpers.get_default_slice_drive_slot(active_node_dm[0])

    def _apply_setup(self):
        # This feature should only be applied the entire cluster per documentation.

        constants_dm = self._cluster_fm.get_constants()

        slot_num = constants_dm.allow_drive_type_modification_of_slot_num

        # Retrieving Active Nodes
        self._active_nodes_dms = self._active_node_selector()

        # Retrieving Active Drives from the above active nodes
        node_drive_dms = self._drive_selector()

        # Retrieving Primary slice drives of volume type and  default slice drive slot
        self._primary_slice_drives_dms = [drive for drive in node_drive_dms if drive.slot ==
                                          self._get_default_slice_drive_slot_for_node_id(drive.node_id)
                                          and drive.type == enums.DriveType.volume]

        # Retrieving Slot Two Drives from Active Node Drives
        self._slot_two_drives_dms = [drive for drive in node_drive_dms if drive.slot == slot_num]

        # Retrieving the Drives of volume type from the above slot 2 drives
        self._slot_two_mdss_drives_dms = [drive for drive in self._slot_two_drives_dms if drive.type ==
                                          enums.DriveType.volume]

    def _reload(self):
        """Reload component.

        :rtype: bool
        """
        if len(self._slot_two_mdss_drives_dms) > 0 and len(self._slot_two_mdss_drives_dms) == len(
                self._active_nodes_dms):
            return True

        elif len(self._slot_two_mdss_drives_dms) > 0:
            raise components.PartialStateError('Cluster not fully MDSS enabled: {}'.
                                               format(self._slot_two_mdss_drives_dms))
        return False

    def _apply(self, timeout=None):
        """Timeout is passed directly to both remove_drives and add_drives.

        :param int timeout: Time in seconds to apply on the add/remove drives.
        """
        if timeout is None:
            timeout = vl_helpers.get_bin_sync_timeout(
                cluster_fm=self._cluster_fm,
                removed_drive_dms=self._slot_two_drives_dms)

        vl_helpers.wait_for_bin_sync_to_finish_with_timeout(self._cluster_fm, timeout=timeout)
        self._mdss_constants_cmt.apply()

        # Remove all slot#2 block drives at once
        self._removed_drive_dms = vl_helpers.remove_drives(self._cluster_fm,
                                                           drive_dms=self._slot_two_drives_dms,
                                                           timeout=timeout)

        # Modify these drives to be metadata type
        modified_drive_dms = vl_helpers.set_drives_type(self._cluster_fm,
                                                        self._removed_drive_dms,
                                                        enable_mdss=True)

        # Wait for bin syncing to complete
        vl_helpers.wait_for_bin_sync_to_finish_with_timeout(self._cluster_fm, timeout=timeout)

        # Add the drives back one at a time
        for drive_dm in modified_drive_dms:
            mdss_drive_dms = vl_helpers.add_drives(self._cluster_fm,
                                                   [drive_dm],
                                                   timeout=timeout,
                                                   enable_mdss_check=False)

            node_dm = vl_helpers.get_active_node_dm_from_drive_dm(self._cluster_fm, mdss_drive_dms[0])

            self._node_mdss_drive_dms[node_dm.static_id] = mdss_drive_dms[0]

        self._mdss_constants_cmt.reset()

    def _reset_mdss_drives_dms(self, primary_and_mdss_drives, primary_slice_drives, timeout):
        """Remove primary and mdss volume drives and Add the primary slice drive.

        :param list[valence.models.Drive] primary_and_mdss_drives: the list of both the primary and
        slot 2 drives per same node that you want to Remove
        :param list[valence.models.Drive] primary_slice_drives: the list of primary slice drives per
        same node that you want to Add
        :param int timeout: Time in seconds to wait for the drives to reset back.
        """

        # Here we are removing both the primary and slot 2 drives
        vl_helpers.remove_drives(self._cluster_fm, drive_dms=primary_and_mdss_drives, timeout=timeout)

        # TODO: Remove sleep when EOS-10102 is resolved
        time.sleep(60)

        # Add back the slice drive to restart the slice service on this node
        vl_helpers.add_drives(self._cluster_fm,
                              drive_dms=primary_slice_drives,
                              timeout=timeout,
                              enable_mdss_check=False)

        #  Wait for slice sync to happen....
        #  If there are no volumes or no balancing deemed necessary, it may not happen.
        # Waiting for slice sync to start/finish
        try:
            vl_helpers.wait_for_slice_sync_to_start(self._cluster_fm, timeout=120)
            vl_helpers.wait_for_slice_sync_to_finish(self._cluster_fm)
        except vl_helpers.HelperError:
            pass

    def _reset(self, timeout=None):
        """Return to non-MDSS cluster state.

        :param int timeout: Time in seconds to wait for the cluster/drives to reset back.
        """
        if self._node_mdss_drive_dms:
            if timeout is None:
                timeout = vl_helpers.get_bin_sync_timeout(
                    cluster_fm=self._cluster_fm,
                    added_drive_dms=list(six.itervalues(self._node_mdss_drive_dms)))

            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(self._cluster_fm, timeout=timeout)
            self._mdss_constants_cmt.apply()
            for node_dm in self._active_nodes_dms:
                # We MUST remove both slice drives at the same time, cannot be done any other way.

                # This consists of primary slice drive per same node.
                primary_slice_drive_per_same_node = [drive for drive in self._primary_slice_drives_dms if
                                                     drive.node_id == self._node_mdss_drive_dms[
                                                         node_dm.static_id].node_id]

                # to_remove  consists of primary slice drive and slot 2 drives per same node.
                to_remove = primary_slice_drive_per_same_node + [self._node_mdss_drive_dms[
                                                                      node_dm.static_id]]

                self._reset_mdss_drives_dms(primary_slice_drives=primary_slice_drive_per_same_node,
                                            primary_and_mdss_drives=to_remove,
                                            timeout=timeout)

            # Add back the block drives (mdss drives)
            block_drive_dms = vl_helpers.set_drives_type(self._cluster_fm,
                                                         list(self._node_mdss_drive_dms.values()),
                                                         enable_mdss=False)

            vl_helpers.add_drives(self._cluster_fm,
                                  drive_dms=block_drive_dms,
                                  timeout=timeout,
                                  enable_mdss_check=False)

            self._mdss_constants_cmt.reset()

            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(self._cluster_fm, timeout=timeout)

        # Resetting the variables
        self._node_mdss_drive_dms = {}
        self._primary_slice_drives_dms = None
        self._active_nodes_dms = None
        self._removed_drive_dms = None


class FailDrives(DrivesInterface, components.BlockingApply):
    """Fail drives on a cluster.

    Drives are failed by editing the /etc/fstab and appending it with dev=0 after raw.

    Usage::

        def setup(self):
            self.fail_drives_cmt = vl_components.FailDrives(self.cluster_dm, self.selector)

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param valence.selectors.DriveSelector drive_selector: Drive selector.
    """

    def __init__(self, cluster_dm, drive_selector):
        super(FailDrives, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._drive_selector = drive_selector
        self._drive_dms = None
        self._drive_dms_stale = None

    @staticmethod
    def mock(cluster_dm, drive_selector):  # pylint: disable=unused-argument
        """Mock the FailDrives.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._drive.FailDrives', vl_components.FailDrives.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._drive.MockDrives
        """
        return MockDrives(drive_selector)

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        if self._drive_dms_stale:
            return self._drive_dms_stale
        elif not self._drive_dms:
            self._drive_dms = self._drive_selector()
        return self._drive_dms

    def _apply_setup(self):
        """Select drives to fail."""
        self._drive_dms_stale = None
        if not self._drive_dms:
            self._drive_dms = self._drive_selector()

    def _reload(self):
        """Reload component.

        :raises valence.components.PartialStateError: If drives are partially failed.
        :rtype: bool
        """
        filtered_drives = self._cluster_fm.list_drives(
            drive_ids=[d.drive_id for d in self.drive_dms],
            statuses=[enums.DriveStatus.failed]
        )
        if not filtered_drives:
            return False
        elif len(filtered_drives) != len(self.drive_dms):
            raise _errors.PartialStateError(
                'Drives were partially failed: desired {} vs actual {}'
                .format(self.drive_dms, filtered_drives)
            )
        else:
            _log.debug('Drive Failure for {} successfully reloaded'.format(self.drive_dms))
            return True

    def _apply(self, timeout=None):
        """Fail drives and verify they show up in failed drives list.

        :param int timeout: timeout used to wait for drive failure
        """
        vl_helpers.fail_drives_in_cluster(
            cluster_fm=self._cluster_fm,
            drive_dms=self._drive_dms,
            wait_for_failure=True,
            timeout=timeout
        )

    def _reset(self, timeout=None):
        """Restore drives failed through this component instance.

        When failing a drive, it will begin to sync out its data after a certain timeout has been
        exceeded (see vl_components.SetDriveServiceTimeout). Once sync-out has begun, the drive WILL
        transition to the Available state once sync-out has completed. However, if the drive is restored
        during this sync-out period, it will show up as Active again until sync-out completes. This can
        create an undesirable race condition depending on when the component is reset::

                                              Restored
                                                 |
            Drive State:  Failed                 |         Active                       Available
                            |----------------X---|>>>>>>>>>>>|------------------------------|
                                             |                                              |
                                             |>>>>>>>>>>>>>>> drive sync-out >>>>>>>>>>>>>>>|
                                             |
                                      Service Timeout

        In order to avoid this the component checks for PRD bin assignments on each drive which would
        indicate sync-out is occurring. If drives are being synced out then we wait for the drives to
        become available before adding back to the cluster any drive that synced out while the component
        was applied.

        NOTE:: You can prevent drives from syncing out and shorten test duration by adjusting the drive
        service timeout using the SetDriveServiceTimeout component.

        :param int timeout: max time to wait for drives to restore
        """
        timeout = 1800 if timeout is None else timeout
        time_start = time.time()
        vl_helpers.restore_failed_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self._drive_dms,
            wait_for_restore=True,
            timeout=timeout
        )
        vl_helpers.recover_failed_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self._drive_dms,
            timeout=timeout - int(time.time() - time_start),
            wait_for_sync=False
        )
        self._drive_dms_stale = self._drive_dms
        self._drive_dms = None


class CascadeFailDrives(DrivesInterface, components.BlockingApply):
    """Iteratively loop on failing a batch of drives and then wait for bin assignments.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.DriveSelector drive_selector: a specific drive selector to use.
    :param int iterations: number of cascading iterations to perform, defaults to 2.
    :param bool wait_for_drive_failed: If true, wait for drive to show up as failed
    :param bool wait_for_bin_assignments: If true, wait for bin assignments after failure
    :param int failure_interval: Time to sleep (seconds) after drive failure
    """

    def __init__(self, cluster_dm, drive_selector, iterations=2, wait_for_drive_failed=True,
                 wait_for_bin_assignments=True, failure_interval=0):
        super(CascadeFailDrives, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

        self._drive_selector = drive_selector
        self._iterations = iterations
        self._wait_for_drive_failed = wait_for_drive_failed
        self._wait_for_bin_assignments = wait_for_bin_assignments
        self._failure_interval = failure_interval
        self._internal_drive_dms_list = None

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('must be applied to access drive dms.')
        return self._internal_drive_dms

    @property
    def _internal_drive_dms(self):
        """All drives associated with this component.

        :rtype: list[valence.models.Drive]
        """
        all_drive_dms = []
        for drive_dms in self._internal_drive_dms_list:
            all_drive_dms.extend(drive_dms)

        return all_drive_dms

    def _apply_setup(self):
        # If fake_drives isn't set fail early
        for node_fm in self._cluster_fm.get_node_fms():
            vl_helpers.check_fake_drives_flag_on_node(node_fm)

    def _apply(self, timeout=None):
        """Fail drives and verify they show up in failed drives list.

        :param int timeout: timeout used to wait for drive failure
        """
        self._internal_drive_dms_list = []
        for iteration in range(self._iterations):
            self._internal_drive_dms_list.append(self._drive_selector())

            # Get the next event ID, so we have a bookmark for event searching for the next bin
            # assignments and don't accidentally match on an existing one.
            next_event_id = vl_helpers.get_next_event_id(self._cluster_fm)
            vl_helpers.fail_drives_in_cluster(
                cluster_fm=self._cluster_fm,
                drive_dms=self._internal_drive_dms_list[-1],
                wait_for_failure=self._wait_for_drive_failed,
                timeout=timeout
            )

            # Wait for bin assignment, which should occur after the last queried event ID.
            if self._wait_for_bin_assignments:
                vl_helpers.wait_for_bin_assignments(self._cluster_fm, start_event_id=next_event_id)

            # Sleep following drive failure
            if self._failure_interval > 0:
                _log.debug('Sleeping for {} seconds following drive failure {} in {}'
                           .format(self._failure_interval, iteration, type(self)))
                time.sleep(self._failure_interval)

    def _reset(self, timeout=None):
        """Restores drives failed through this component instance.  Waits for recovery.

        Drives which were in the active state during apply are added to the cluster

        :param int timeout: max time to wait for drives to restore
        """
        timeout = 1800 if timeout is None else timeout
        time_start = time.time()
        vl_helpers.restore_failed_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self._internal_drive_dms,
            wait_for_restore=True,
            timeout=timeout
        )
        vl_helpers.recover_failed_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self._internal_drive_dms,
            timeout=timeout - int(time.time() - time_start)
        )
        self._internal_drive_dms_list = None


class RemoveDrives(DrivesInterface, components.NonBlockingApply):
    """Remove Drives at rest component.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.DriveSelector drive_selector: drive selector
    :param bool ignore_unbalanced_cluster_check: Whether to ignore errors raised
        due to drive removal stranding cluster capacity.
    :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
    :param bool wait_for_bin_sync_complete: Whether to wait for bin syncing to complete before
        attempting to remove drives.
    """

    # TODO (VL-2583): Remove these variables when tests are updated to no longer reference this variable.
    # TIMEOUT cvar is deprecated and no longer used.
    TIMEOUT = 600

    def __init__(self, cluster_dm, drive_selector, ignore_unbalanced_cluster_check=False,
                 force_during_bin_sync=False, wait_for_bin_sync_complete=True):
        super(RemoveDrives, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._drive_selector = drive_selector
        self._ignore_unbalanced_cluster_check = ignore_unbalanced_cluster_check
        self._force_during_bin_sync = force_during_bin_sync
        self._wait_for_bin_sync_complete = wait_for_bin_sync_complete
        self._internal_drive_dms = None
        self.rm_drives_async_handler = None
        self.node_dms = None
        self.cluster_fault = None

    @staticmethod
    def mock(cluster_dm, drive_selector,  # pylint: disable=unused-argument
             ignore_unbalanced_cluster_check=False,  # pylint: disable=unused-argument
             force_during_bin_sync=False,  # pylint: disable=unused-argument
             wait_for_bin_sync_complete=True):  # pylint: disable=unused-argument
        """Mock the RemoveDrives.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._drive.RemoveDrives', vl_components.RemoveDrives.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._drive.MockDrives
        """
        return MockDrives(drive_selector)

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        if not self._internal_drive_dms:
            self._internal_drive_dms = self._drive_selector()
        return self._internal_drive_dms

    def _reload(self):
        """Reload component.

        :raises valence.components.PartialStateError: If drives are partially removed.
        :rtype: bool
        """
        filtered_drives = self._cluster_fm.list_drives(
            drive_ids=[d.drive_id for d in self.drive_dms],
            statuses=[enums.DriveStatus.available]
        )
        if not filtered_drives:
            return False
        elif len(filtered_drives) == len(self.drive_dms):
            _log.debug('Drive Removal for {} successfully reloaded'.format(self.drive_dms))
            return True
        else:
            raise _errors.PartialStateError(
                'Drives were partially removed desired: {} vs actual {}'
                .format(self.drive_dms, filtered_drives)
            )

    def _apply_async(self, timeout=None):
        """Remove drives asynchronously.

        :param int timeout: Timeout in seconds.
        """

        if timeout is None:
            timeout = vl_helpers.get_bin_sync_timeout(self._cluster_fm, removed_drive_dms=self.drive_dms)
        if self._wait_for_bin_sync_complete:
            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(self._cluster_fm)
        self.rm_drives_async_handler = self._cluster_fm.remove_drives(
            drive_dms=self.drive_dms,
            ignore_unbalanced_cluster_check=self._ignore_unbalanced_cluster_check,
            force_during_bin_sync=self._force_during_bin_sync,
            timeout=timeout
        )

    def _apply_is_running(self):
        """Use async handler to check if apply is still running.

        :rtype: bool
        """
        return self.rm_drives_async_handler.is_running()

    def _apply_finish(self, timeout=None):
        """Finish applying.

        :param int timeout: Timeout in seconds
        """
        self.rm_drives_async_handler.get_models(blocking=True, timeout=timeout)

        self.node_dms = [vl_helpers.get_active_node_dm_from_drive_dm(
            cluster_fm=self._cluster_fm, drive_dm=drive) for drive in self.drive_dms]

        self.cluster_fault = vl_helpers.wait_for_cluster_fault_occurrence(
            cluster_fm=self._cluster_fm,
            fault_code=enums.ClusterFaultCode.drive_available,
            only_nodes=self.node_dms)

        _log.debug('Cluster fault drive available info {}'.format(self.cluster_fault))

    def _reset(self, timeout=None):
        """Reset by adding drives back.

        :param int timeout: Timeout in seconds
        :raises valence.components.PartialStateError: If drives are partially added already.
        """
        if timeout is None:
            timeout = vl_helpers.get_bin_sync_timeout(self._cluster_fm, added_drive_dms=self.drive_dms)

        drive_dms = self._cluster_fm.list_drives(drive_ids=[d.drive_id for d in self.drive_dms])
        available_drive_dms = [dm for dm in drive_dms if dm.status is enums.DriveStatus.available]
        if self._last_state is components.States.applied and len(available_drive_dms) != len(drive_dms):
            _log.debug('Removed drives in the available state: {}'.format(available_drive_dms))
            invalid_drive_dms = [dm for dm in drive_dms if dm not in available_drive_dms]
            raise _errors.PartialStateError(
                'Drives found in unexpected states (not available): {}'.format(invalid_drive_dms)
            )

        if available_drive_dms:
            time_start = time.time()
            vl_helpers.add_drives(
                cluster_fm=self._cluster_fm,
                drive_dms=available_drive_dms,
                timeout=timeout,
                ignore_unbalanced_cluster_check=True,
                force_during_bin_sync=True
            )
            vl_helpers.wait_for_drive_assignments(
                cluster_fm=self._cluster_fm,
                drive_dms=self.drive_dms,
                timeout=timeout - int(time.time() - time_start)
            )
            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(
                cluster_fm=self._cluster_fm,
                timeout=timeout - int(time.time() - time_start)
            )

            vl_helpers.wait_for_cluster_fault_resolution(
                cluster_fm=self._cluster_fm,
                only_nodes=self.node_dms,
                only_fault_codes=[enums.ClusterFaultCode.drive_available],
                timeout=timeout - int(time.time() - time_start))

        self._internal_drive_dms = None


class ContinuousFailDrives(components.ContinuousTransient):
    """ContinuousFailDrives component.

    Fail drives over and over. A drive selector may be used to generate one random block drive
    every time, or use a lambda to fail the same drive over and over. By default, it will pick any
    drive on any active node.

    :param valence.models.ResourceCluster cluster_dm: Resource cluster data model. This is needed for
        cluster credentials if the nodes are part of a cluster.
    :param valence.selectors.DriveSelector drive_selector: drive selector, if none 1 random drive is
        selected, from any active node
    :param int kill_after_seconds: Time to run the component
    :param int sleep_time: How long in seconds to sleep between loops
    :param int max_iterations: How many iterations to loop
    """

    def __init__(self, cluster_dm=None, drive_selector=None, kill_after_seconds=None, sleep_time=None,
                 max_iterations=None):
        super(ContinuousFailDrives, self).__init__(kill_after_seconds, sleep_time, max_iterations)
        self._drive_selector = drive_selector
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

    def _apply_continuous(self):
        """Fail drives and restore them continuously.

        Will fail drives by editing the fstab file and will wait for sf to report them as failed.
        Once they're failed, restore the fstab file, then wait for the drive to become active or
        available.
        """
        if not self._drive_selector:
            # Fail drives in a random node
            node_selector = selectors.ActiveNodeSelector(self._cluster_dm, num=1)
            drive_selector = selectors.DriveSelector(self._cluster_dm, node_selector=node_selector)
        else:
            drive_selector = self._drive_selector
        vl_helpers.fail_drives_in_cluster(self._cluster_fm, drive_selector())

        # Restore the drives
        vl_helpers.restore_failed_drives(self._cluster_fm, drive_selector())

    def _reset(self, timeout=None):
        """If only some drives failed or we hit an error ensure all failed drives are restored.

        :param int timeout: Time in seconds to wait for reset
        """
        failed_drives = self._cluster_fm.list_drives(statuses=[enums.DriveStatus.failed])
        vl_helpers.restore_failed_drives(self._cluster_fm, failed_drives)


class AddDrives(DrivesInterface, components.NonBlockingApply):
    """Add Drives to Cluster component.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.DriveSelector drive_selector: drive selector
    :param bool ignore_unbalanced_cluster_check: Whether to ignore errors raised
        due to drive removal stranding cluster capacity.
    """
    TIMEOUT = 600

    def __init__(self, cluster_dm, drive_selector, ignore_unbalanced_cluster_check=False):
        super(AddDrives, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._ignore_unbalanced_cluster_check = ignore_unbalanced_cluster_check
        self._drive_selector = drive_selector
        self._internal_drive_dms = None
        self.add_drives_async_handler = None

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        if not self._internal_drive_dms:
            self._internal_drive_dms = self._drive_selector()
        return self._internal_drive_dms

    def _reload(self):
        """Reload component.

        :raises valence.components.PartialStateError: If drives are partially removed.
        :rtype: bool
        """
        filtered_drives = self._cluster_fm.list_drives(
            drive_ids=[d.drive_id for d in self.drive_dms],
            statuses=[enums.DriveStatus.active]
        )
        if not filtered_drives:
            return False
        elif len(filtered_drives) == len(self.drive_dms):
            _log.debug('Drive Addition for {} successfully reloaded'.format(self.drive_dms))
            return True
        else:
            raise _errors.PartialStateError('Drives were partially added: {} vs '
                                            'actual {}'.format(self.drive_dms, filtered_drives))

    def _apply_async(self, timeout=None):
        """Remove drives asynchronously.

        :param int timeout: N/A
        """
        self.add_drives_async_handler = self._cluster_fm.add_drives(
            drive_dms=self.drive_dms,
            ignore_unbalanced_cluster_check=self._ignore_unbalanced_cluster_check
        )

    def _apply_is_running(self):
        """Use async handler to check if apply is still running.

        :rtype: bool
        """
        return self.add_drives_async_handler.is_running()

    def _apply_finish(self, timeout=None):
        """Finish applying.

        :param int timeout: Timeout in seconds
        """
        self.add_drives_async_handler.get_models(blocking=True, timeout=timeout)

    def _reset(self, timeout=None):
        """Reset by removing the previously added drives.

        :param int timeout: Timeout in seconds
        :raises valence.components.PartialStateError: If drives are partially removed already.
        """
        filtered_drives = self._cluster_fm.list_drives(
            drive_ids=[d.drive_id for d in self.drive_dms],
            statuses=[enums.DriveStatus.active]
        )
        if len(filtered_drives) != len(self.drive_dms):
            raise _errors.PartialStateError('Drives were partially removed: {}'.format(filtered_drives))

        vl_helpers.remove_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self.drive_dms,
            timeout=timeout,
            ignore_unbalanced_cluster_check=True,
            force_during_bin_sync=True
        )
        self._internal_drive_dms = None


class CascadeAddDrives(DrivesInterface, components.BlockingApply):
    """Add a batch of drives, wait for bin assignment, then add another batch of drives.

    This component assumes the system has drives in the removed state that can be added
    and uses a drive selector accordingly.  If the iterations (taking into account the
    DriveSelector) consumes more than the number of drives that can be added, you can
    expect NotEnoughItemsError errors.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param int iterations: number of cascading iterations to perform, defaults to 2.
    :param valence.selectors.DriveSelector drive_selector: a specific drive selector to use.
        If specified, this takes precedence over batch_size.
    """
    TIMEOUT = 1200

    def __init__(self, cluster_dm, drive_selector, iterations=2):
        super(CascadeAddDrives, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._drive_selector = drive_selector
        self._iterations = iterations
        self._add_drive_cmts = []

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        total_drive_dms = []
        for add_drive_cmt in self._add_drive_cmts:
            total_drive_dms.append(add_drive_cmt.drive_dms)

        return total_drive_dms

    def _apply_setup(self):
        for _ in range(self._iterations):
            add_cmt = AddDrives(self._cluster_dm, drive_selector=self._drive_selector)
            self._add_drive_cmts.append(add_cmt)

    def _apply(self, timeout=None):
        """Add the drives and verify they show up in added drives list.

        :param int timeout: timeout used to wait for drive failure
        """

        for add_drive_cmt in self._add_drive_cmts:
            # Get the next event ID, so we have a bookmark for event searching for the next bin
            # assignments and don't accidentally match on an existing one.
            next_event_id = vl_helpers.get_next_event_id(self._cluster_fm)

            add_drive_cmt.apply()

            # Wait for bin assignment, which should occur after the last queried event ID.
            vl_helpers.wait_for_bin_assignments(self._cluster_fm, timeout=self.TIMEOUT,
                                                start_event_id=next_event_id)

    def _reset(self, timeout=None):
        """Removes any added drives through this component instance.  Waits for recovery.

        NOTE: These drives will be removed in batches and likely to cause data loss, so you should really
        consider whether or not you need to reset() this component.

        :param int timeout: max time to wait for drives to restore
        """

        for add_drive_cmt in self._add_drive_cmts:
            add_drive_cmt.ensure_ready()

        vl_helpers.wait_for_bin_sync_to_finish_with_timeout(cluster_fm=self._cluster_fm, timeout=timeout)


class MockDrives(components.MockComponentMixin, DrivesInterface, components.BlockingApply):
    """A mock drives component representing a drive selector.

    :param valence.selectors.DriveSelector drive_selector: Drive selector.
    """
    TIMEOUT = 600

    def __init__(self, drive_selector):
        super(MockDrives, self).__init__()
        self._drive_selector = drive_selector
        self._internal_drive_dms = None

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        if not self._internal_drive_dms:
            self._internal_drive_dms = self._drive_selector()
        return self._internal_drive_dms


@six.add_metaclass(abc.ABCMeta)
class FailDriveTypeBase(DrivesInterface, components.NonBlockingApply):
    """Fail drives on a cluster.

    Drives are failed by editing the /etc/fstab and appending it with dev=0 after raw.

    Usage::

        def setup(self):
            self.fail_drives_cmt = vl_components.FailDrives(self.cluster_dm, self.selector)

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param valence.selectors.DriveSelector drive_selector: Drive selector.
    :param bool wait_for_sync: whether to wait for syncing during reset
    :param bool add_drives: whether to add drives back during reset
    """

    def __init__(self, cluster_dm, drive_selector, wait_for_sync=True, add_drives=True):
        super(FailDriveTypeBase, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._drive_selector = drive_selector
        self._drive_dms = None
        self._drive_dms_stale = None
        self.wait_for_sync = wait_for_sync
        self.add_drives = add_drives

    @staticmethod
    def mock(cluster_dm, drive_selector):  # pylint: disable=unused-argument
        """Mock the FailDrives.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._drive.FailDrives', vl_components.FailDrives.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._drive.MockDrives
        """
        return MockDrives(drive_selector)

    @property
    def drive_dms(self):
        """Property to lazily load the drive_dms from the selector.

        :rtype: list[valence.models.Drive]
        """
        if self._drive_dms_stale:
            return self._drive_dms_stale
        elif not self._drive_dms:
            self._drive_dms = self._drive_selector()
        return self._drive_dms

    def _apply_setup(self):
        """Select drives to fail."""
        self._drive_dms_stale = None
        if not self._drive_dms:
            self._drive_dms = self._drive_selector()

    def _reload(self):
        """Reload component.

        :raises valence.components.PartialStateError: If drives are partially failed.
        :rtype: bool
        """
        drive_ids = [d.drive_id for d in self._drive_dms]
        services_filter = criteria.ContainedIn(models.ClusterService.drive_id, drive_ids)
        current_services = [s.service for s in self._cluster_fm.list_services()]
        filtered_services = services_filter.filter(current_services)
        if len(filtered_services) == len(self._drive_dms):
            return False
        elif len(filtered_services) != 0:
            raise _errors.PartialStateError(
                'Drives were partially failed: desired {} services still active {}'.format(
                    self.drive_dms, filtered_services)
            )
        else:
            return True

    def _apply_async(self, timeout=None):
        """Fail drives, return immediately, do not wait for failure status.

        :param int timeout: Not used.
        """
        vl_helpers.fail_drives_in_cluster(
            cluster_fm=self._cluster_fm,
            drive_dms=self._drive_dms,
            wait_for_failure=False,
        )

    def _apply_is_running(self):
        drive_ids = [d.drive_id for d in self._drive_dms]
        services_filter = criteria.ContainedIn(models.ClusterService.drive_id, drive_ids)
        current_services = [s.service for s in self._cluster_fm.list_services()]
        if services_filter.filter(current_services):
            return True
        else:
            return False

    def _apply_finish(self, timeout=None):
        vl_helpers.wait_for_sync_out(self._cluster_fm, self._drive_dms)

    @abc.abstractmethod
    def _reset(self, timeout=None):
        """Must be implemented.

        :param int timeout: not used
        """


class FailSliceDrives(FailDriveTypeBase):
    """Fail slice drives on a cluster.

    Drives are failed by editing the /etc/fstab and appending it with dev=0 after raw.

    Usage::

        def setup(self):
            self.fail_drives_cmt = vl_components.FailSliceDrives(self.cluster_dm, self.selector)

    """
    def _apply_setup(self):
        super(FailSliceDrives, self)._apply_setup()
        block_drives = [d for d in self._drive_dms if d.type is enums.DriveType.block]
        if block_drives:
            raise components.ComponentStateError('Trying to remove block drive as slice. {}'.format(
                block_drives))

    def _reset(self, timeout=None):
        """Restore slice drives failed through this component instance.

        When failing a slice drive, it shouldn't matter if the drive was synced-out or failed.
        Just restore it.

        :param int timeout: timeout for adding drives back if they were synced out
        """
        timeout = 1800 if timeout is None else timeout
        vl_helpers.restore_failed_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self._drive_dms,
            wait_for_restore=True,
        )
        if self.add_drives:
            available_drive_dms = self._cluster_fm.list_drives(
                drive_ids=[dm.drive_id for dm in self._drive_dms],
                statuses=[enums.DriveStatus.available]
            )
            if available_drive_dms:
                vl_helpers.add_drives(
                    cluster_fm=self._cluster_fm,
                    drive_dms=available_drive_dms,
                    timeout=timeout,
                )
        if self.wait_for_sync:
            vl_helpers.wait_for_sync_in(self._cluster_fm, self._drive_dms)
        self._drive_dms_stale = self._drive_dms
        self._drive_dms = None


class FailBlockDrives(FailDriveTypeBase):
    """Fail block drives on a cluster.

    Drives are failed by editing the /etc/fstab and appending it with dev=0 after raw.

    Usage::

        def setup(self):
            self.fail_drives_cmt = vl_components.FailBlockDrives(self.cluster_dm, self.selector)
    """
    def _apply_setup(self):
        super(FailBlockDrives, self)._apply_setup()
        slice_drives = [d for d in self._drive_dms if d.type is enums.DriveType.volume]
        if slice_drives:
            raise components.ComponentStateError('Trying to remove slice drive as block. {}'.format(
                slice_drives))

    def _reset(self, timeout=None):
        """Restore block drives failed through this component instance.

        This will not wait for syncing to complete.

        :param int timeout: not used
        """
        need_to_sync_again = False
        vl_helpers.restore_failed_drives(
            cluster_fm=self._cluster_fm,
            drive_dms=self._drive_dms,
            wait_for_restore=True,
        )
        if self.add_drives:
            available_drive_dms = self._cluster_fm.list_drives(
                drive_ids=[dm.drive_id for dm in self._drive_dms],
                statuses=[enums.DriveStatus.available]
            )
            if available_drive_dms:
                vl_helpers.add_drives(
                    cluster_fm=self._cluster_fm,
                    drive_dms=available_drive_dms,
                    timeout=timeout,
                    ignore_unbalanced_cluster_check=True,
                    force_during_bin_sync=True
                )
                need_to_sync_again = True

        if self.wait_for_sync and need_to_sync_again:
            vl_helpers.wait_for_sync_in(self._cluster_fm, self._drive_dms)
        self._drive_dms_stale = self._drive_dms
        self._drive_dms = None
