"""Valence fill components.

Example Usage::

    capacity_selector = StageBlockCapacitySelector(cluster_dm, 3)
    fill_cmt = FillCluster('fill', cluster_dm, capacity_selector)
    fill_cmt.apply()
    fill_cmt.reset()


Example Usage::

    capacity_selector = PercentBlockCapacitySelector(cluster_dm, percent_capacity=70)
    capacity_selector.add(FixedBlockCapacitySelector(cluster_dm, cluster_bytes=-1000000000))
    fill_cmt = FillCluster('fill', cluster_dm, capacity_selector)
    fill_cmt.apply()
    fill_cmt.reset()


Example Usage::

    # On a cluster with 30GB used space out of 100GB total capacity
    >> capacity_selector = PercentBlockCapacitySelector(cluster_dm, percent_capacity=70)
    >> capacity_selector().gigabytes
    70

    # Use "relative_to_used_capacity" to factor in the existing used space
    >> capacity_selector = PercentBlockCapacitySelector(
            cluster_dm=cluster_dm,
            percent_capacity=70,
            relative_to_used_capacity=True
        )
    >> capacity_selector().gigabytes
    40


Example Usage::

    block_capacity_selector = PercentBlockCapacitySelector(cluster_dm, percent_capacity=1)
    meta_capacity_selector = PercentMetaCapacitySelector(cluster_dm, percent_capacity=40)
    fill_cmt = FillCluster('fill', cluster_dm, block_capacity_selector, meta_capacity_selector)
    fill_cmt.apply()
    fill_cmt.reset()
"""
from __future__ import division

import autologger

from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    utils,
    vl_helpers,
)
from valence.vl_components import (
    _account,
    _cluster,
    _constants,
    _volume,
)

_log = autologger.get_sflogging(__name__)


class FillCluster(components.BlockingApply, components.Reversible):
    """Fills the cluster to the specified block/metadata fullness.

    Creates and manages volumes to reach the desired fullness.

    NOTE: Any IO to these volumes can impact the state of fullness. For best results drive IO with the
    same dedup/compression ratios as specified here. To keep fullness stable while driving IO, use
    dedup/compression ratios of 1.

    :param str tag: The UNIQUE tag to use for associated volumes and accounts.
    :param valence.models.ResourceCluster cluster_dm: Cluster model.
    :param valence.selectors.ClusterBlockCapacitySelector block_capacity_selector: Block capacity
        selector.
    :param valence.selectors.ClusterMetaCapacitySelector meta_capacity_selector: Meta capacity
        selector.
    :param float comp_ratio: Compression ratio to target when filling.
    :param float dedup_ratio: Deduplication ratio to target when filling (cannot be used with
        meta_capacity_selector).
    """
    BLOCK_PERCENT_TOLERANCE = 1
    META_PERCENT_TOLERANCE = 1

    def __init__(self, tag, cluster_dm, block_capacity_selector, meta_capacity_selector=None,
                 comp_ratio=1., dedup_ratio=1.):
        super(FillCluster, self).__init__()
        self.tag = tag
        self.cluster_dm = cluster_dm
        self.block_capacity_selector = block_capacity_selector
        self.meta_capacity_selector = meta_capacity_selector
        self.comp_ratio = comp_ratio
        self._dedup_ratio = dedup_ratio
        self.rep_count = 2
        self.block_size = int(4096 / self.comp_ratio)

        self._cluster_fm = functional_models.Cluster(self.cluster_dm)
        self._extra_volumes_cmt = None
        self._block_data_size = None
        self._meta_data_size = None

        self._gc_cmt = _cluster.RunGarbageCollection(cluster_dm)

        capacity_selector = self.block_capacity_selector
        if self.meta_capacity_selector:
            assert self._dedup_ratio == 1
            capacity_selector = BlockForMetaCapacitySelector(
                cluster_dm=self.cluster_dm,
                meta_capacity_selector=self.meta_capacity_selector,
                comp_ratio=self.comp_ratio
            )

        # Initialize volumes that will be filled 100% when generating block data
        # NOTE: IO may be run against these volumes with minimal impact to cluster fullness
        self.volumes_cmt = _volume.CreateVolumesForCapacity(
            tag='{}-full'.format(self.tag),
            cluster_dm=self.cluster_dm,
            capacity_selector=capacity_selector,
            comp_ratio=self.comp_ratio,
            dedup_ratio=self._dedup_ratio,
            under_size=True
        )

        # In the case we just want to fill Metadata, lets do it faster and with bigger vols
        if self.dedup_ratio > 100:
            self.volumes_cmt.DEFAULT_VOL_COUNT = 1

    @property
    def block_data_size(self):
        """Block data size to be generated on the cluster.

        :rtype: valence.models.DataSize
        """
        if not self._block_data_size:
            self._block_data_size = self.block_capacity_selector()
        return self._block_data_size

    @property
    def meta_data_size(self):
        """Metadata size to be generated on the cluster.

        :rtype: valence.models.DataSize
        """
        if self.meta_capacity_selector and not self._meta_data_size:
            self._meta_data_size = self.meta_capacity_selector()
        return self._meta_data_size

    @property
    def dedup_ratio(self):
        """Deduplication ratio of the data to be generated on the cluster.

        :rtype: float
        """
        if self.meta_capacity_selector and self._dedup_ratio == 1:
            block_metadata_size = vl_helpers.get_block_metadata_size(
                cluster_fm=self._cluster_fm,
                block_data_size=self.block_data_size,
                comp_ratio=self.comp_ratio
            )
            self._dedup_ratio = float(self.meta_data_size.bytes) / block_metadata_size.bytes

        return self._dedup_ratio

    def assert_valid(self):
        """Validate cluster-fullness selectors have achieved the desired fullness state."""
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Must be applied to assert valid')

        if self.meta_capacity_selector:
            if not self.meta_capacity_selector.relative:
                raise ValueError('Only relative metadata capacities can be validated')
        else:
            if not self.block_capacity_selector.relative:
                raise ValueError('Only relative block capacities can be validated')

        assert self._block_capacity_achieved()
        if self.meta_capacity_selector:
            assert self._meta_capacity_achieved()

    def _block_capacity_achieved(self):
        """Check whether the target block capacity has been achieved.

        :rtype: bool
        """
        cluster_full_threshold = self._cluster_fm.get_cluster_full_threshold()
        total_bytes = cluster_full_threshold.sum_total_cluster_bytes
        used_bytes = cluster_full_threshold.sum_used_cluster_bytes
        _log.debug('Cluster block-data fullness: {}%'.format(float(100 * used_bytes) / total_bytes))

        capacity_achieved = True
        if self.block_capacity_selector.relative:
            tolerance_bytes = self.BLOCK_PERCENT_TOLERANCE * total_bytes // 100
            min_bytes = self.block_capacity_selector.target_bytes - tolerance_bytes
            max_bytes = self.block_capacity_selector.target_bytes + tolerance_bytes
            capacity_achieved = min_bytes <= used_bytes <= max_bytes
        return capacity_achieved

    def _meta_capacity_achieved(self):
        """Check whether the target metadata capacity has been achieved.

        :rtype: bool
        """
        cluster_full_threshold = self._cluster_fm.get_cluster_full_threshold()
        total_bytes = cluster_full_threshold.sum_total_metadata_cluster_bytes
        used_bytes = cluster_full_threshold.sum_used_metadata_cluster_bytes
        _log.debug('Cluster meta-data fullness: {}%'.format(float(100 * used_bytes) / total_bytes))

        capacity_achieved = True
        if self.meta_capacity_selector.relative:
            tolerance_bytes = self.META_PERCENT_TOLERANCE * total_bytes // 100
            min_bytes = self.meta_capacity_selector.target_bytes - tolerance_bytes
            max_bytes = self.meta_capacity_selector.target_bytes + tolerance_bytes
            capacity_achieved = min_bytes <= used_bytes <= max_bytes
        return capacity_achieved

    def _apply(self, timeout=None):
        """Generate the configured cluster fullness.

        :param int timeout: No effect.
        """
        _log.debug('Creating volumes for cluster fullness')
        self.volumes_cmt.apply()

        # Create an additional volume to handle remaining block fill capacity
        created_capacity = self.volumes_cmt.block_capacity.bytes / self.dedup_ratio
        remaining_capacity = self.block_data_size.bytes - created_capacity
        if remaining_capacity > 0:
            capacity_selector = selectors.FixedBlockCapacitySelector(
                cluster_dm=self.cluster_dm,
                cluster_bytes=int(remaining_capacity)
            )
            self._extra_volumes_cmt = _volume.CreateVolumesForCapacity(
                tag='{}-extra'.format(self.tag),
                cluster_dm=self.cluster_dm,
                capacity_selector=capacity_selector,
                comp_ratio=self.comp_ratio,
                dedup_ratio=self.dedup_ratio,
                vol_count=1
            )
            self._extra_volumes_cmt.apply()

        _log.debug('Generating block-data: {}'.format(self.block_capacity_selector))
        if self.meta_capacity_selector:
            _log.debug('Generating meta-data: {}'.format(self.meta_capacity_selector))

        vl_helpers.fill_volumes(
            cluster_fm=self._cluster_fm,
            volume_dms=self.volumes_cmt.volume_dms,
            fill_fraction=min(1., self.block_data_size.bytes / created_capacity),
            comp_ratio=self.comp_ratio,
            dedup_ratio=self.dedup_ratio
        )
        self.volumes_cmt.dedup_ratio = self.dedup_ratio

        if self._extra_volumes_cmt:
            # Fudge factor to ensure we're above the desired threshold
            remaining_capacity += 0.5 * self.block_size * self.dedup_ratio

            # Fill the last volume only as full as necessary to achieve the desired fullness
            vl_helpers.fill_volumes(
                cluster_fm=self._cluster_fm,
                volume_dms=self._extra_volumes_cmt.volume_dms,
                fill_fraction=min(1., remaining_capacity / self._extra_volumes_cmt.block_capacity.bytes),
                comp_ratio=self.comp_ratio,
                dedup_ratio=self.dedup_ratio
            )

        # Check the final state of cluster fullness
        vl_helpers.update_cluster_full_threshold(self._cluster_fm, self._gc_cmt)
        assert self._block_capacity_achieved()
        if self.meta_capacity_selector:
            assert self._meta_capacity_achieved()

    def _reset(self, timeout=None):
        """Delete all volumes and run garbage collection.

        :param int timeout: No effect.
        """
        if self._extra_volumes_cmt:
            self._extra_volumes_cmt.ensure_ready()
        self.volumes_cmt.ensure_ready()
        vl_helpers.update_cluster_full_threshold(self._cluster_fm, self._gc_cmt)

        self._extra_volumes_cmt = None
        self._block_data_size = None
        self._meta_data_size = None

        if self.meta_capacity_selector:
            self._dedup_ratio = 1.


class FillClusterMetadata(FillCluster):
    """Fills the cluster to the specified metadata fullness.

    Creates and manages volumes to reach the desired fullness.

    NOTE: Any IO to these volumes can impact the state of fullness. For best results drive IO with the
    same dedup/compression ratios as specified here. To keep fullness stable while driving IO, use
    dedup/compression ratios of 1.

    :param str tag: The UNIQUE tag to use for associated volumes and accounts.
    :param valence.models.ResourceCluster cluster_dm: Cluster model.
    :param valence.selectors.ClusterMetaCapacitySelector meta_capacity_selector: Meta capacity
        selector.
    :param float comp_ratio: Compression ratio to target when filling.
    :param float dedup_ratio: Deduplication ratio to target when filling (cannot be used with
        meta_capacity_selector).
    """
    BLOCK_CAPACITY_PERCENT = 1

    def __init__(self, tag, cluster_dm, meta_capacity_selector, comp_ratio=1., dedup_ratio=1.):

        # Fill to 1% block capacity since we only care about Metadata
        block_capacity_selector = selectors.PercentBlockCapacitySelector(cluster_dm,
                                                                         self.BLOCK_CAPACITY_PERCENT)

        super(FillClusterMetadata, self).__init__(tag, cluster_dm, block_capacity_selector,
                                                  meta_capacity_selector, comp_ratio, dedup_ratio)


class BlockForMetaCapacitySelector(selectors.ClusterBlockCapacitySelector):
    """Selector for cluster block capacity required to achieve target metadata capacity.

    Returns the total block data size required to achieve the target metadata capacity with no dedup.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param valence.selectors.ClusterMetaCapacitySelector meta_capacity_selector: Meta capacity
        selector.
    :param float comp_ratio: Compression ratio to target when filling.
    """
    def __init__(self, cluster_dm, meta_capacity_selector, comp_ratio=1.):
        super(BlockForMetaCapacitySelector, self).__init__(cluster_dm, False)
        self.meta_capacity_selector = meta_capacity_selector
        self.comp_ratio = comp_ratio

    def _get_target_bytes(self, cluster_full_threshold):
        """Get the number of bytes representing the desired cluster capacity.

        :param sf_api.result_models.GetClusterFullThreshold cluster_full_threshold: ClusterFullThreshold
            result data model.
        :rtype: int
        """
        meta_data_size = self.meta_capacity_selector()
        return vl_helpers.get_metadata_block_size(
            cluster_fm=self._cluster_fm,
            metadata_size=meta_data_size,
            comp_ratio=self.comp_ratio,
            under_size=True
        ).bytes


class SetupRecycling(components.BlockingApply, _volume.VolumesInterface):
    """FillsCluster with homogeneous active data.

    Pre 11.7 Behavior:
    Performs several operations where volumes are filled and then some subset of them are purged.
    After garbage collection is performed, this creates holes in the data with a constant
    recycling percentage on the cluster for future IO.

    Post 11.7 Behavior:
    Utilizes the cBSFirstTimeWriteRatio constant in conjunction with FillSlice to reach the target
    fullness and lap the drive at the same time. No volume delete/purge operations necessary.

    This component exposes a volumes interface for use with other components.  Each volume left over
    after the apply will have its own Volumes Component so that volumes can be sliced for
    multiple uses.

    User has the option to specify a minimum number of volumes OR minimum volume size for the
    volumes provided after the apply.  To keep the math simple, the user cannot specify both
    options to this component.  If a variety of volume sizes are needed, one can modify the
    volumes provided to the required size keeping in mind that total block space usage will increase.

    :param str tag: tag to create accounts/volumes under
    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param float fullness_percent: fraction of cluster to fill
    :param int min_num_volumes: minumum volume count available after apply()
    :param valence.models.DataSize min_volume_size: minimum size of volumes available after apply()
    :param float comp_ratio: compression ratio of cluster fill data
    :param valence.enums.ProtectionScheme protection_scheme: protection scheme of created volumes
    """

    # Thou shall not touch
    MAX_FILLSLICE_VOLUMES = 10
    USED_BLOCKSPACE_TOLERANCE = 0.05

    SCHEME_REP_COUNT_MAP = {
        enums.ProtectionScheme.single_helix: 1,
        enums.ProtectionScheme.double_helix: 2,
        enums.ProtectionScheme.triple_helix: 3
    }

    def __init__(self, tag, cluster_dm, fullness_percent, min_num_volumes=None, min_volume_size=None,
                 comp_ratio=1.0, protection_scheme=enums.ProtectionScheme.double_helix):
        super(SetupRecycling, self).__init__()
        self.tag = tag
        self.cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm)
        self._fullness_percent = fullness_percent
        self._min_num_volumes = min_num_volumes
        self._min_volume_size = min_volume_size
        self._rep_count = self.SCHEME_REP_COUNT_MAP[protection_scheme]
        self._comp_ratio = comp_ratio
        self._protection_scheme = protection_scheme

        self._partial_volume_dm = None
        self._account_cmt = None
        self._volume_cmts = []
        self._volume_cmts_subsets = []
        self._volume_size = None
        self._volume_count = None
        self._scalar = None
        self._num_cycles = None
        self._keepers_per_cycle = None
        self._block_size = 4096 // self._comp_ratio  # FillSlice is always 4k
        self._max_used_space = None
        self._cluster_constants = None
        self._cluster_limits = None

        self._first_time_write_ratio_cmt = None

        self._skip_gc_cmt = _constants.SetClusterConstants(
            self.cluster_dm,
            models.ClusterConstants.create(skip_gc=1)
        )

        self._run_gc_cmt = _cluster.RunGarbageCollection(self.cluster_dm)

    def _reset(self, timeout=None):
        """Remove accounts and volumes, perform garbage collection.

        :param int timeout: unused
        """
        for vol_cmt in self._volume_cmts:
            vol_cmt.reset()

        self._account_cmt.reset()
        self._run_gc_cmt.apply()
        self._skip_gc_cmt.ensure_ready()
        if self._first_time_write_ratio_cmt:
            self._first_time_write_ratio_cmt.ensure_ready()

    @property
    def child_cmts(self):
        return self._volume_cmts

    @property
    def account_dms(self):
        return self._account_cmt.account_dms

    @property
    def cluster_constants(self):
        if not self._cluster_constants:
            self._cluster_constants = self._cluster_fm.get_constants()
        return self._cluster_constants

    @property
    def cluster_limits(self):
        if not self._cluster_limits:
            self._cluster_limits = self._cluster_fm.get_limits()
        return self._cluster_limits

    @property
    def max_used_space(self):
        if not self._max_used_space:
            cluster_capacity = self._cluster_fm.get_cluster_capacity()
            self._max_used_space = models.DataSize(cluster_capacity.max_used_space)
        return self._max_used_space

    def get_account_dms(self):
        """Returns the account data model used by this component.

        :rtype: list[valence.models.Account]
        """
        return self._account_cmt.get_account_dms()

    def get_volume_dms(self):
        """Returns a list of volume data models owned by the child volume components.

        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()

        # Instead of making a call for each child component, just do it once here
        return [dm for dm in self._cluster_fm.list_volumes() if self.tag in dm.name]

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        return [self._partial_volume_dm] * len(self._volume_cmts)

    @property
    def volume_dms(self):
        """Returns a list of volume data models owned by the child volume components.

        :rtype: list[valence.models.Volume]
        """
        volume_dms = []
        for vol_cmt in self._volume_cmts:
            volume_dms.extend(vol_cmt.volume_dms)
        return volume_dms

    def assert_valid(self):
        """Verifies the child components have been properly applied."""
        for cmt in self._volume_cmts:
            cmt.assert_valid()

    def _validate_inputs(self):
        """Sanity check all user inputs and ensure valid for this cluster."""
        if self._comp_ratio < 1.0:
            raise ValueError('Compression ratio must be greater than or equal to 1.0')
        if self._fullness_percent % 10 != 0:
            raise ValueError('Currently, fill % must be divisible by 10 (10%, 20%...)')
        if self._min_num_volumes and self._min_volume_size:
            raise ValueError('User can specify end volume count or size, but not both.')

        num_vols_to_create = self._num_cycles * self.MAX_FILLSLICE_VOLUMES

        if self._min_volume_size > self.max_used_space / num_vols_to_create:
            raise ValueError('Minimum volume size too large.')
        if num_vols_to_create > self.cluster_limits.volume_count_max:
            raise ValueError('Invalid volume count specified, max {}'
                             .format(self.cluster_limits.volume_count_max))
        if self._min_volume_size and self._min_volume_size < self.cluster_limits.volume_size_min:
            raise ValueError('Minimum volume size less than cluster minimum {}'
                             .format(self.cluster_limits.volume_size_min))

    def _calculate_apply_parameters(self):
        lap_drive_pct = self._fullness_percent if self._first_time_write_ratio_cmt else 100
        lap_drive_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, lap_drive_pct)

        total_volumes_size = lap_drive_selector() * self._comp_ratio / (
            vl_helpers.get_block_overhead_factor(self._cluster_fm) * self._rep_count)

        base_num_volumes = 100
        scalar = 1
        # If user specifies min # of volumes left over after apply
        if self._min_num_volumes:
            # look for scalar multiple of base_num_volumes to satisfy user input
            for ii in range(1, 100):
                if self._fullness_percent * ii >= self._min_num_volumes:
                    scalar = ii
                    break
            self._volume_count = scalar * base_num_volumes
        else:
            # we've already established min_volume_size to be less than 1% of usable space
            self._volume_count = base_num_volumes

        self._volume_size = total_volumes_size // self._volume_count
        self._num_cycles = self.MAX_FILLSLICE_VOLUMES * scalar
        self._keepers_per_cycle = int(self._fullness_percent * self.MAX_FILLSLICE_VOLUMES /
                                      lap_drive_pct)

    def _create_components(self):
        self._account_cmt = _account.CreateAccounts(self.tag + '-act', self.cluster_dm)
        self._partial_volume_dm = models.Volume.create(total_size=self._volume_size)

        for i in range(self._volume_count):
            vol_cmt = _volume.CreateVolumes(self.tag + '-vol{}'.format(i),
                                            self.cluster_dm,
                                            [self._partial_volume_dm],
                                            self._account_cmt)
            vol_cmt.apply()
            self._volume_cmts.append(vol_cmt)

    def _split_volumes_cmts_subsets(self):
        """Break the volumes components into equal size chunks for each cycle."""
        volumes_per_cycle = len(self._volume_cmts) // self._num_cycles
        for i in range(0, len(self._volume_cmts), volumes_per_cycle):
            self._volume_cmts_subsets.append(self._volume_cmts[i:i + volumes_per_cycle])

    def _apply_setup(self):
        # None means the constant doesn't exist, 1.0 means the feature is off and should not be used
        if self.cluster_constants.bs_first_time_write_ratio not in [None, 1.0]:
            self._first_time_write_ratio_cmt = _constants.SetClusterConstants(
                self.cluster_dm,
                models.ClusterConstants.create(bs_first_time_write_ratio=self._fullness_percent / 100.)
            )
        self._calculate_apply_parameters()
        self._validate_inputs()

        if self._cluster_fm.list_volumes():
            raise components.ComponentStateError('Cluster has pre-existing data.')

    def _fill_purge_cycle(self, vol_cmts_subset):
        """Performs volume fill and purge operations for the subsets of volume components.

        :param list[valence.vl_components.CreateVolumes] vol_cmts_subset: list of volumes for this cycle
        """
        volume_dms = []
        for vol_cmt in vol_cmts_subset:
            volume_dms.extend(vol_cmt.volume_dms)

        vl_helpers.fill_volumes(self._cluster_fm, volume_dms, comp_ratio=self._comp_ratio)

        # pylint: disable=invalid-unary-operand-type
        for vol_cmt in vol_cmts_subset[:-self._keepers_per_cycle]:
            vol_cmt.reset()
            self._volume_cmts.remove(vol_cmt)

    def _gc_cleanup_loop(self):
        """This method performs a sequence of gc operations to clean up unused block data.

        Multiple GC operations are often required because of the false positive rate on GC.
        Once the used capacity of the cluster is within tolerance, we stop.
        """
        timer = utils.timeout_timer(2*60*60)

        # active_block_space is only calculated based on values after last gc, so we gc here first.
        self._run_gc_cmt.apply()
        self._run_gc_cmt.reset()

        while next(timer):
            cluster_capacity = self._cluster_fm.get_cluster_capacity()
            target_space = self._fullness_percent * self.max_used_space // 100
            maximum_delta = self.USED_BLOCKSPACE_TOLERANCE * target_space

            if (cluster_capacity.active_block_space - target_space) < maximum_delta:
                break

            self._run_gc_cmt.apply()
            self._run_gc_cmt.reset()

    def _apply(self, timeout=None):
        self._create_components()
        self._split_volumes_cmts_subsets()

        # Skip garbage collection during the fill operations, we will perform it manually
        self._skip_gc_cmt.apply()

        # Set cBSFirstTimeWriteRatio if it's available which will distribute the data product side
        if self._first_time_write_ratio_cmt:
            self._first_time_write_ratio_cmt.apply()

        for index, vol_cmts_subset in enumerate(self._volume_cmts_subsets):
            _log.debug('Beginning fill/purge loop #{}'.format(index))
            self._fill_purge_cycle(vol_cmts_subset)

            # Perform a GC in the middle to make sure we have some free space to work with
            # We can skip this step if cBSFirstTimeWriteRatio is used because no volumes are deleted
            if index == len(self._volume_cmts_subsets) // 2 and not self._first_time_write_ratio_cmt:
                _log.debug('Starting garbage collection')
                self._run_gc_cmt.apply()
                self._run_gc_cmt.reset()

        self._gc_cleanup_loop()
        self._skip_gc_cmt.reset()

        if self._first_time_write_ratio_cmt:
            self._first_time_write_ratio_cmt.reset()
