"""This component will run fallocate command on a given cluster node OS partition."""
import os
import random

import autologger
import sf_ssh

from valence import (
    components,
    functional_models,
)
from valence.utils import random_string

_log = autologger.get_sflogging(__name__)


class FillPartition(components.Reversible, components.BlockingApply):
    """This component will run fallocate command on a given cluster node OS partition.

    This component doesn't do anything with actual cluster storage or volumes.
    It purely exists to mess with the underlying base OS partitions like /var/log, /tmp, /, etc.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.ActiveNodeSelector node_selector: random selector
    :param str partition: partition to fill
    :param valence.models.DataSize max_free_space_remaining: How much free space
        to leave
    :param valence.models.DataSize fill_data: write this much data
    """

    def __init__(self, cluster_dm, node_selector, partition, max_free_space_remaining=None,
                 fill_data=None):
        super(FillPartition, self).__init__()
        self.cluster_dm = cluster_dm
        self.node_selector = node_selector
        self.partition = partition
        self.max_free_space_remaining = max_free_space_remaining
        self.fill_data = fill_data
        if self.max_free_space_remaining and self.fill_data:
            raise ValueError('Cannot specify both max_free_space_remaining and fill_data')
        if not (self.max_free_space_remaining or self.fill_data):
            raise ValueError('Must specify either max_free_space_remaining and fill_data')
        self.remote_temp_files = {}
        self._filled_nodes = None

    @property
    def filled_nodes(self):
        if self.state is components.States.ready:
            raise components.ComponentStateError("Querying for filled nodes before it's applied")
        return self._filled_nodes

    def _get_fsize(self, available_size_b):
        """Calculates size of data to allocate.

        :param int available_size_b: unused space
        :returns: Amount of space to allocate (in bytes)
        :rtype: int
        """
        if self.fill_data:
            fsize = self.fill_data
        else:
            will_leave = random.randint(0, self.max_free_space_remaining.bytes)
            # This ensures that we will at least create the file in case there's less free
            # space than what was asked for.
            fsize = max(0, available_size_b - will_leave)
        return fsize

    def _run_df(self, conn):
        """Runs df command and returns output for a specific partition.

        root@flabv0004:~# df -B1 /
        Filesystem             1B-blocks        Used   Available Use% Mounted on
        /dev/disk/by-uuid/blah 84418347008 15117012992 64989433856  19% /

        :param sf_ssh.ssh.Ssh conn: Open ssh connection.
        :returns: Amount of space available
        :rtype: int
        """
        rco = conn.run_command('df -B1 ' + self.partition, use_sudo=False)
        stdout = (line.strip() for line in rco.stdout.splitlines())
        stdout = [line for line in stdout if line]
        columns = stdout[1].split()
        return int(columns[3])

    def _apply_setup(self, timeout=None):
        """Valence will try reloading the component and call the selector at that time.

        :param int timeout: Timeout when waiting for completing
        """
        self._filled_nodes = self.node_selector()

    def _apply(self, timeout=None):
        """Fills given partition on all selected nodes.

        :param int timeout: Timeout when waiting for completing
        """
        _log.info('Selecting the following nodes for test: {}'.
                  format(' '.join([str(n.name) for n in self._filled_nodes])))

        for node in self._filled_nodes:
            _log.info('Filling {}'.format(node.name))
            node_fms = functional_models.Node(node.resource_dm, self.cluster_dm)
            with node_fms.get_ssh_connection() as conn:
                available_size_b = self._run_df(conn)
                _log.debug('{} has {} available bytes'.format(self.partition, available_size_b))

                remote_temp = os.path.join(self.partition, random_string.generate_random_string(10))

                fsize = self._get_fsize(available_size_b)

                # We'll try to write directly to tmp file on the partition. This way we don't have to
                # figure out / vs /var/www vs /home/ whatever.
                _log.info('Will fallocate a temp file {} of size {}'.format(remote_temp, fsize))
                rco = conn.run_command('fallocate -l {} {}'.format(fsize, remote_temp))
                self.remote_temp_files[node.name] = remote_temp

                # -rw-r--r-- 1 root root 10000000000 Oct 26 17:41 blob.bin
                rco = conn.run_command('ls -la ' + remote_temp)
                written_size = int(rco.stdout.split()[4])
                _log.debug('Got size {} vs actual size {}'.format(written_size, fsize))

                post_dd_avail = self._run_df(conn)
                _log.debug('After our file creation, available space is {}'.format(post_dd_avail))

    def _reset(self, timeout=None):
        """Cleans up all files we've written to the selected nodes.

        :param int timeout: Timeout when waiting for completing
        :raises valence.components.ComponentStateError: If file not deleted
        """

        cleaned_up = True
        for node in self._filled_nodes:
            node_fm = functional_models.Node(node.resource_dm, self.cluster_dm)
            try:
                if node.name in self.remote_temp_files:
                    file_to_be_deleted = self.remote_temp_files[node.name]
                    with node_fm.get_ssh_connection() as conn:
                        conn.run_command('rm ' + file_to_be_deleted)
                        del self.remote_temp_files[node.name]
            except sf_ssh.errors.SshError:
                _log.debug('Unable to delete file {} from node {}.'.
                           format(file_to_be_deleted, node.name))
                cleaned_up = False
        if not cleaned_up:
            raise components.ComponentStateError()
