"""Valence Initiators components."""
from __future__ import division

import abc
import re

import autologger
import six

from valence import (
    components,
    functional_models,
    models,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class InitiatorsInterface(components.BlockingApply, components.Reversible):
    """Interface component for a set of initiators."""

    @property
    @abc.abstractmethod
    def initiator_dms_map(self):
        """Get the cached initiator data models by client.

        :returns: A map of the clients to their cluster initiator objects.
        :raises components.ComponentStateError: If component has not been applied.
        :rtype: dict[str, list[valence.models.Initiator]]
        """

    @abc.abstractmethod
    def get_initiator_dms_map(self):
        """Make an API call to get the applied initiators.

        :returns: A map of the clients to their cluster initiator objects.
        :raises components.ComponentStateError: If component has not been applied.
        :rtype: dict[str, list[valence.models.Initiator]]
        """

    @abc.abstractmethod
    def assert_valid(self):
        """Assert the existing initiators match the configuration."""


class CreateIscsiInitiators(InitiatorsInterface):
    """Create multiple iscsi initiator objects on the cluster for the given clients.

    :param str tag: The unique tag to use for these initiators.
    :param valence.models.ResourceCluster cluster_dm: The cluster on which to create the initiators.
    :param list[valence.models.ResourceClient] client_dms: The clients to get the initiators.
    """
    INITIATOR_ALIAS_FMT = r'{tag}-TAG-{cl_name}-{index}'

    def __init__(self, tag, cluster_dm, client_dms):
        super(CreateIscsiInitiators, self).__init__()
        if not client_dms:
            raise ValueError('Must pass in at least one client dm.')
        self.tag = tag
        self.cluster_dm = cluster_dm
        self.client_dms = client_dms
        self._client_dm_map = {c.client_name: c for c in client_dms}
        self._cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        self._client_fms = [functional_models.get_client_from_dm(client_dm)
                            for client_dm in client_dms]
        self._initiator_dms_map = None

    @staticmethod
    def mock(tag, cluster_dm, client_dms):  # pylint: disable=unused-argument
        """Mock the CreateIscsiInitiators.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._initiator.CreateIscsiInitiators',
                        vl_components.CreateIscsiInitiators.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._initiator.MockInitiators
        """
        initiator_dms_map = {}
        for client_dm in client_dms:
            initiator_dms_map[client_dm.mip] = [models.Initiator.create(initiator_name='init0',
                                                                        initiator_id=0
                                                                        )]
        return MockInitiators(tag, client_dms, initiator_dms_map)

    @property
    def initiator_dms_map(self):
        """Get the cached initiator data models by client.

        :returns: A map of the client's mips to their cluster initiator objects.
        :raises components.ComponentStateError: If component has not been applied.
        :rtype: dict[str, list[valence.models.Initiator]]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Must be applied to access initiator dms.')
        return self._initiator_dms_map

    def get_initiator_dms_map(self):
        """Get the applied initiators from the cluster by client.

        :returns: A map of the clients to their cluster initiator objects.
        :raises components.ComponentStateError: If component has not been applied.
        :rtype: dict[str, list[valence.models.Initiator]]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Must be applied to get the initiator dms.')
        return self._get_existing()

    def assert_valid(self):
        """Assert the existing initiators match the configuration.

        :raises components.ComponentStateError: If component has not been applied.
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Must be applied to try assert_valid.')
        self._assert_valid()

    @property
    def full_tag(self):
        """Tag used in resource names.

        :returns: A string using the provided unique identifier with -TAG- added.
        :rtype: str
        """
        return self.tag + '-TAG-'

    def _apply(self, timeout=None):
        """Create initiator objects on the cluster.

        :param int timeout: No effect.
        """
        sf_initators = []
        for client_dm in self.client_dms:
            raw_name = get_raw_iscsi_initiator(client_dm)
            alias = self.INITIATOR_ALIAS_FMT.format(
                tag=self.tag, cl_name=client_dm.client_name, index=0)
            sf_initators.append(
                models.Initiator.create(
                    initiator_name=raw_name,
                    alias=alias
                )
            )

        self._cluster_fm.create_initiators(
            initiators=sf_initators
        )
        self._initiator_dms_map = self._get_existing()
        self._assert_valid(self._initiator_dms_map)

    def _reset(self, timeout=None):
        """Remove the initiators from the cluster.

        :param int timeout: No effect.
        """
        initiator_dms = []
        initiator_dms_map = self._get_existing()
        for client_dm in self.client_dms:
            if client_dm.mip in initiator_dms_map:
                initiator_dms.extend(initiator_dms_map[client_dm.mip])

        if initiator_dms:
            self._cluster_fm.delete_initiators(initiator_dms)
        self._initiator_dms_map = None

    def _reload(self):
        """Attempt to reload the component.

        Checks to see if the initiators already exist on the cluster and are valid for their respective
        clients.

        :returns: Whether the component was able to reload.
        :rtype: bool
        """
        # Get all initiators on the cluster
        initiator_dms_map = self._get_existing()
        if not initiator_dms_map:
            return False

        self._assert_valid(initiator_dms_map)
        self._initiator_dms_map = initiator_dms_map
        _log.info('Initiators with tag {} were reloaded'.format(self.tag))
        return True

    def _assert_valid(self, initiator_dms_map=None):  # pylint: disable=too-complex
        """Assert the specified initiators are valid.

        :param dict[str,list[valence.models.Initiator]] initiator_dms_map: Sets of initiators
            for each client (current set of initiators on cluster by default).
        """
        if initiator_dms_map is None:
            initiator_dms_map = self._get_existing()

        err = []
        for client_dm in self.client_dms:
            initiator_dms = initiator_dms_map.get(client_dm.mip)
            if initiator_dms is None:
                err.append('No initiators found for client[{}]'.format(client_dm.client_name))
                continue

            invalid_initiator_names = []
            raw_initiators = get_raw_iscsi_initiator(client_dm)
            for cluster_init_dm in initiator_dms:
                if cluster_init_dm.initiator_name != raw_initiators:
                    invalid_initiator_names.append(cluster_init_dm.initiator_name)
            if invalid_initiator_names:
                err.append('Invalid initiators for client[{}] found on cluster: {}'.format(
                    client_dm.client_name,
                    invalid_initiator_names
                ))
                continue

            expected_initiator_count = 1
            if len(initiator_dms) != expected_initiator_count:
                err.append('Initiators number is invalid on client[{}]. Found {}, expected {}'.format(
                    client_dm.client_name,
                    len(initiator_dms),
                    expected_initiator_count
                ))
                continue

        if err:
            raise components.PartialStateError('\n'.join(err))

    def _get_existing(self):
        """Get existing initiators on cluster.

        :returns: Sets of initiators on the cluster for each client with the unique tag in the name.
        :rtype: dict[str, list[valence.models.Initiator]]
        """
        init_dms = self._get_cluster_initiator_dms()
        init_dms = tuple(init_dm for init_dm in init_dms
                         if init_dm.alias.startswith(self.full_tag))

        initiators = {}
        initiator_alias_pattern = re.compile(self.INITIATOR_ALIAS_FMT.format(
            tag=r'^.+', cl_name=r'(.+)', index=r'\d+$'))
        for init_dm in init_dms:
            client_name = initiator_alias_pattern.search(init_dm.alias).group(1)
            if client_name not in self._client_dm_map:
                raise components.PartialStateError(
                    'Initiator {} found on cluster for unrecognized client {}'
                    .format(init_dm.alias, client_name))
            client_dm = self._client_dm_map[client_name]
            if client_dm.mip not in initiators:
                initiators[client_dm.mip] = [init_dm]
            else:
                initiators[client_dm.mip].append(init_dm)

        return initiators

    def _get_cluster_initiator_dms(self):
        """Get all initiators on the cluster.

        :rtype: list[valence.models.Initiator]
        """
        return self._cluster_fm.list_initiators()


def get_raw_iscsi_initiator(client_dm):
    """Make ssh calls on the client or whatever to get the iscsi initiator name.

    :param valence.models.ResourceClient client_dm: The client to get the name from.
    :rtype: str
    """
    client_fm = functional_models.get_client_from_dm(client_dm)
    return client_fm.get_initiator_name()


class MockInitiators(components.MockComponentMixin, InitiatorsInterface):
    """A mock initiators component representing an initiator data model map.

    See the documentation on MockComponentMixin for more details about correct usage.

    :param str tag: The unique tag to use for these initiators.
    :param list[valence.models.ResourceClient] client_dms: The clients to get the initiators.
    :param dict[str,list[valence.models.Initiator]] initiator_dms_map: The initiators that this
        component emulates.
    """

    def __init__(self, tag, client_dms, initiator_dms_map):
        super(MockInitiators, self).__init__()
        self.tag = tag
        self.client_dms = client_dms
        self._initiator_dms_map = initiator_dms_map

    @property
    def initiator_dms_map(self):
        """Get the cached initiator data models by client.

        :returns: A map of the clients to their cluster initiator objects.
        :raises components.ComponentStateError: If component has not been applied.
        :rtype: dict[str, list[valence.models.Initiator]]
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()
        return self._initiator_dms_map

    @property
    def full_tag(self):
        """Tag used in resource names.

        :returns: A string using the provided unique identifier with -TAG- added.
        :rtype: str
        """
        return self.tag + '-TAG-'

    def get_initiator_dms_map(self):
        """Get the applied initiators.

        Because this is a facade component, this does not make an API call.

        :returns: A map of the clients to their cluster initiator objects.
        :raises components.ComponentStateError: If component has not been applied.
        :rtype: dict[str, list[valence.models.Initiator]]
        """
        return self.initiator_dms_map

    def assert_valid(self):
        """Assert the existing initiators match the configuration.

        Note: this is a noop because this is a facade compoennt.
        """
