"""Valence network components."""

import collections

from valence import (
    components,
    enums,
    functional_models,
    vl_helpers,
)


class DownInterfacesOnNodes(components.Reversible, components.BlockingApply):
    """DownInterfacesOnNodes component.

    This reversible component will bring down the selected interfaces on given nodes.

    :param interfaces: List of interfaces to bring down
    :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
    :param valence.models.ResourceCluster cluster_dm: Resource cluster data model. This is needed for
        cluster credentials if the nodes are part of a cluster.
    :param valence.selectors.ActiveNodeSelector node_selector: node selector

    """

    def __init__(self, interfaces, cluster_dm, node_selector):
        super(DownInterfacesOnNodes, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._node_selector = node_selector
        self._interfaces = interfaces
        self._internal_node_dms = None
        self._node_fms = None

    @property
    def node_dms(self):
        """Property to lazily load the node_dms from the node selector.

        :rtype: list[valence.models.ResourceNode]
        """
        if not self._internal_node_dms:
            self._internal_node_dms = self._node_selector()
        return self._internal_node_dms

    def _reload(self):
        """Reload component.

        :rtype: bool
        """
        for node_fm in self._node_fms:
            node_network_interfaces = node_fm.get_network_interface_status(self._interfaces)
            for interface in self._interfaces:
                status = node_network_interfaces[interface.value]
                if status is not enums.NetworkStatus.down:
                    # One of the interfaces is not down. Reloading is not possible.
                    return False
        # All of the interfaces are already down. Reload the component.
        return True

    def _apply_setup(self, timeout=None):
        """Forced apply for Down Interfaces.

        :param int timeout: Timeout in seconds
        """
        self._node_fms = [
            functional_models.Node(n, authenticate=False,
                                   cluster_dm=self._cluster_dm) for n in self.node_dms]

    def _apply(self, timeout=None):
        """Down interfaces synchronously.

        :param int timeout: Timeout in seconds
        """
        vl_helpers.down_network_interfaces(self._node_fms, self._interfaces)

    def _reset(self, timeout=None):
        """Bring interfaces back up.

        :param int timeout: Timeout in seconds
        """
        vl_helpers.up_network_interfaces(self._node_fms, self._interfaces)


class ContinuousFlapInterfacesOnNodes(components.ContinuousTransient):
    """ContinuousFlapInterfacesOnNodes component.

    :param interfaces: List of interfaces to bring down
    :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
    :param valence.selectors.ActiveNodeSelector node_selector: node selector
    :param valence.models.ResourceCluster cluster_dm: Resource cluster data model. This is needed for
        cluster credentials if the nodes are part of a cluster.
    :param int kill_after_seconds: Time to run the component
    :param int sleep_time: How long in seconds to sleep between loops
    :param int max_iterations: How many iterations to run the loop
    """

    def __init__(self, interfaces, node_selector, cluster_dm=None, kill_after_seconds=None,
                 sleep_time=None, max_iterations=None):
        super(ContinuousFlapInterfacesOnNodes, self).__init__(kill_after_seconds, sleep_time,
                                                              max_iterations)
        self.cmt = DownInterfacesOnNodes(interfaces, cluster_dm, node_selector)

    def _apply_continuous(self):
        """Down and up the interface."""
        self.cmt.apply()
        self.cmt.reset()

    def _reset(self, timeout=None):
        """Ensure any downed interfaces are restored.

        :param int timeout: Maximum time reset should take.
        """
        self.cmt.ensure_ready()


class InterNodeNetemNetworkPerturbation(components.Reversible, components.BlockingApply):
    """Component to apply network pertubations between sets of nodes based on netem.

    .. note::

        If specifying a bidirectional perturbation, the rules applied will likely amplify 2x since
        each node will have the perturbation with each other (e.g. a 200 ms latency on both sides of the
        cluster pairing will see a 400ms total latency when performing a ping).

    :param valence.selectors.ClusteredNodeSelector node_selector_one: Node selector for the
        source of nodes.
    :param valence.selectors.ClusteredNodeSelector node_selector_two: Node selector for the
        target of perturbation.
    :param list[str] rules: list of rules to apply
    :param bool bidirectional: whether to introduce the perturbation on both connection ends.
    :param tuple[valence.enums.NetworkInterface] networks: networks to perturb (e.g. Bond10G, Bond1G...)
    :param bool virtual_ip: Whether or not to include the MVIP and/or SVIP as applicable.
    """

    def __init__(self, node_selector_one, node_selector_two,
                 rules,
                 bidirectional=True,
                 networks=(enums.NetworkInterface.bond_10g, enums.NetworkInterface.bond_1g),
                 virtual_ip=True,
                 ):
        super(InterNodeNetemNetworkPerturbation, self).__init__()
        self._node_selector_one = node_selector_one
        self._cluster_one_dm = node_selector_one.cluster_dm
        self._node_selector_two = node_selector_two
        self._cluster_two_dm = node_selector_two.cluster_dm

        self._rules = rules

        self._bidirectional = bidirectional
        self._networks = networks
        self._network_values = [n.value for n in self._networks]
        self._virtual_ip = virtual_ip

        self._set_one_node_dms = None
        self._set_one_node_fms = None
        self._set_two_node_dms = None
        self._set_two_node_fms = None

    def _apply_setup(self, timeout=None):
        """Ensure there aren't any existing netem rules applied on any of the impacted networks.

        :param int timeout: Timeout in seconds
        """

        self._set_one_node_dms = self._node_selector_one()
        self._set_one_node_fms = [
            functional_models.Node(node.resource_dm, self._cluster_one_dm, authenticate=False) for node
            in self._set_one_node_dms]

        self._set_two_node_dms = self._node_selector_two()
        self._set_two_node_fms = [
            functional_models.Node(node.resource_dm, self._cluster_two_dm, authenticate=False) for node
            in self._set_two_node_dms]

        self._ensure_no_existing_rules(self._set_one_node_fms)

        if self._bidirectional:
            self._ensure_no_existing_rules(self._set_two_node_fms)

    def _apply(self, timeout=None):
        """Apply a set of netem rules to perturb the connection between nodes.

        :param int timeout: Timeout in seconds
        """

        self._set_one_node_dms = self._node_selector_one()
        self._set_one_node_fms = [
            functional_models.Node(node.resource_dm, self._cluster_one_dm, authenticate=False) for node
            in self._set_one_node_dms]

        self._set_two_node_dms = self._node_selector_two()
        self._set_two_node_fms = [
            functional_models.Node(node.resource_dm, self._cluster_two_dm, authenticate=False) for node
            in self._set_two_node_dms]

        # Create a map of cluster interfaces to destination IPs to perturb.
        cluster_one_device_target_map = self._generate_network_map(
            self._cluster_two_dm,
            self._set_two_node_dms)

        vl_helpers.add_netem_rules_inter_nodes(self._set_one_node_fms,
                                               cluster_one_device_target_map,
                                               self._rules)

        # If we're doing a bi-directional perturbation, set the same rules on the other cluster.
        if self._bidirectional:
            cluster_two_device_target_map = self._generate_network_map(
                self._cluster_one_dm,
                self._set_one_node_dms)

            vl_helpers.add_netem_rules_inter_nodes(self._set_two_node_fms,
                                                   cluster_two_device_target_map, self._rules)

    def _ensure_no_existing_rules(self, node_fms):
        """Ensure there aren't any existing netem rules on the nodes.

        :param list[valence.functional_models.Node] node_fms: Target Nodes to ensure have no rules.

        :raises valence.components.ResourceStateError: if device already has rules applied to it.
        """

        list_cmd = 'tc qdisc show dev {dev}'
        error_msg = 'Cannot apply rules to node {node} on device {dev} that has existing netem rules.'

        for node_fm in node_fms:
            with node_fm.get_ssh_connection() as ssh_conn:
                for network in self._networks:
                    response = ssh_conn.run_command(list_cmd.format(dev=network.value),
                                                    raise_on_failure=True).stdout

                    if response.strip() != 'qdisc noqueue 0: root refcnt 2':
                        raise components.ResourceStateError(
                            error_msg.format(node=node_fm.node_dm.node_name, dev=network.value))

    def _generate_network_map(self, target_cluster_dm, target_nodes):
        """Generate Network Map containing the various IPs of the Target nodes to leverage.

        :param valence.models.ResourceCluster target_cluster_dm: Cluster data model
        :param list[valence.models.NodeBase] target_nodes: Target Nodes that we need to collect IPs from

        :rtype: dict[str,list[str]]
        :returns: A mapping of network device name and a list of IPs.
        """

        device_target_map = collections.defaultdict(list)
        for network in self._networks:
            if network is enums.NetworkInterface.bond_1g:

                if self._virtual_ip:
                    # Add the Management Virtual IP for the Cluster
                    device_target_map[network.value] = [
                        target_cluster_dm.mvip
                    ]

                # Append all the individual node Management IPs
                device_target_map[network.value].extend([
                    node.mip for node in target_nodes
                ])
            elif network is enums.NetworkInterface.bond_10g:

                # Add the Storage Virtual IP for the Cluster
                if self._virtual_ip:
                    device_target_map[network.value] = [
                        target_cluster_dm.svip
                    ]

                # Append all the individual node Storage IPs
                device_target_map[network.value].extend([
                    node.sip for node in target_nodes
                ])

        return device_target_map

    def _reset(self, timeout=None):
        """Clear the set of netem rules to restore the network connection between nodes.

        :param int timeout: Timeout in seconds (not used)
        """

        if self._set_one_node_fms:
            vl_helpers.clear_netem_rules(self._set_one_node_fms, self._network_values)

        if self._bidirectional and self._set_two_node_fms:
            vl_helpers.clear_netem_rules(self._set_two_node_fms, self._network_values)


class InterNodeLatencyPerturbation(InterNodeNetemNetworkPerturbation):
    """InterNodeLatencyPerturbation component.

    This reversible component will introduce network latency between Nodes

    :param valence.selectors.ClusteredNodeSelector node_selector_one: Node selector for the
        source of nodes.
    :param valence.selectors.ClusteredNodeSelector node_selector_two: Node selector for the
        target of perturbation.
    :param int latency: time in milliseconds to introduce latency
    :param bool bidirectional: whether to introduce the latency on both connection ends.
    :param tuple[valence.enums.NetworkInterface] networks: networks to perturb (e.g. Bond10G, Bond1G...)
    :param bool virtual_ip: Whether or not to include the MVIP and/or SVIP as applicable.
    :param int jitter: optional jitter time in milliseconds to include in the latency

    :raises ValueError: if you provide a jitter value that isn't an integer between 1..100
    """

    def __init__(self, node_selector_one, node_selector_two,
                 latency,
                 bidirectional=True,
                 networks=(enums.NetworkInterface.bond_10g, enums.NetworkInterface.bond_1g),
                 virtual_ip=True,
                 jitter=None):
        self._latency = latency
        self._jitter = jitter

        assert isinstance(self._latency, int)

        rule = 'delay {}ms'.format(self._latency)
        if self._jitter and isinstance(self._jitter, int) and self._jitter > 0:
            rule += ' {}ms'.format(self._jitter)
        elif self._jitter:
            raise ValueError('jitter must be an integer greater than 0')

        super(InterNodeLatencyPerturbation, self).__init__(node_selector_one,
                                                           node_selector_two,
                                                           [rule],
                                                           bidirectional=bidirectional,
                                                           networks=networks,
                                                           virtual_ip=virtual_ip)


class InterNodePacketLossPerturbation(InterNodeNetemNetworkPerturbation):
    """InterNodePacketLossPerturbation component.

    This reversible component will introduce network packet loss between Nodes

    :param valence.selectors.ClusteredNodeSelector node_selector_one: Node selector for the
        source of nodes.
    :param valence.selectors.ClusteredNodeSelector node_selector_two: Node selector for the
        target of perturbation.
    :param float pct_packet_loss: the percentage of packets to drop
    :param bool bidirectional: whether to introduce the packet loss on both connection ends.
    :param tuple[valence.enums.NetworkInterface] networks: networks to perturb (e.g. Bond10G, Bond1G...)
    :param bool virtual_ip: Whether or not to include the MVIP and/or SVIP as applicable.

    :raises ValueError: If pct_packet_loss isn't a valid value.
    """

    def __init__(self, node_selector_one, node_selector_two,
                 pct_packet_loss,
                 bidirectional=True,
                 networks=(enums.NetworkInterface.bond_10g, enums.NetworkInterface.bond_1g),
                 virtual_ip=True):
        if pct_packet_loss < 0. or pct_packet_loss > 100.:
            raise ValueError('pct_packet_loss must be between 0s and 100')

        super(InterNodePacketLossPerturbation, self).__init__(node_selector_one,
                                                              node_selector_two,
                                                              ['loss {}%'.format(pct_packet_loss)],
                                                              bidirectional=bidirectional,
                                                              networks=networks,
                                                              virtual_ip=virtual_ip)
