"""Valence node components."""
from autologger import get_sflogging

from sf_api import enums
from valence import (
    components,
    functional_models,
    models,
    selectors,
    vl_helpers,
)
from valence.vl_components import _power

_log = get_sflogging(__name__)


class ResetNode(components.Permanent, components.ProcessedBlockingApply):
    """Reset a list of nodes.

    Resetting a node is accomplished by issuing an API call, ResetNode, to the per node api.

    :param valence.models.ResourceCluster cluster_dm: cluster
    :param valence.selectors.ActiveNodeSelector node_selector: selector
    :param valence.models.SfVersion build: Element software build to which the node will be reset
    :param bool reboot: whether or not to allow node to reboot
    """

    def __init__(self, cluster_dm, node_selector, build=None, reboot=None):
        super(ResetNode, self).__init__()
        self._node_selector = node_selector
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._nodes = None
        self.build = build if build else cluster_dm.api_endpoint
        self.reboot = reboot

    def _apply(self, timeout=None):
        node_fms = [functional_models.Node(node.resource_dm, self._cluster_dm) for node in self._nodes
                    if node.resource_dm is not None]
        vl_helpers.reset_nodes(node_fms, self.build, self.reboot)

    def _apply_setup(self):
        self._nodes = self._node_selector()
        if not self._nodes:
            raise components.ComponentStateError('No nodes meet selection criteria for {}'.format(self))

    def _reset(self, timeout=None):
        raise components.ComponentStateError('Cannot recover from a failure to apply reset node.')


class SetNodeConstants(components.Reversible, components.BlockingApply):
    """Set node constants.

    This component stores the value of the constants before it changes
    them. When it is reset the constants are returned to their original
    value.  If this is the first modification to a value, the ResetNodeConstant
    API call will be used to ensure cluster constant is not overridden.
    Note: Not all Cluster constants are valid node constants.  If a node constant
    is set, it will override the cluster constant until the reset method is called.

    If multiple SetNodeConstants are applied, they must be reset the in the opposite order
    to return to the original state.

    :param valence.models.ResourceNode node_dm: the node to set constants on
    :param valence.models.ResourceCluster|None cluster_dm: the cluster to authenticate with.
        For an orphaned node, cluster_dm should be provided as None
    :param valence.models.ClusterConstants constants_dm: the constants to set,
        None values in the data model are ignored
    """

    def __init__(self, node_dm, cluster_dm, constants_dm):
        super(SetNodeConstants, self).__init__()
        self._node_dm = node_dm
        self._cluster_dm = cluster_dm
        self._node_fm = functional_models.Node(node_dm, cluster_dm)
        self._constants_dm = constants_dm
        self._modified_original_constants = None

    def _apply(self, timeout=None):
        """Sets Node Specific API Constants.

        :param int timeout: Unused
        """
        self._modified_original_constants = self._node_fm.get_constants(changed_only=True)
        self._node_fm.set_constants(self._constants_dm)

    def _reset(self, timeout=None):
        """Rolls back any changes made to node constants to the original values.

        :param int timeout: Unused
        """
        applied_constants_dict = self._constants_dm.serialize()
        modified_constants_dict = self._modified_original_constants.serialize()

        if modified_constants_dict:
            mod_constants_dm = models.ClusterConstants(modified_constants_dict)
            self._node_fm.set_constants(mod_constants_dm)

            # Any constants being reverted to original (but non-default) value should not be reset
            for constant in modified_constants_dict:
                applied_constants_dict.pop(constant, None)

        # Any values modified from their default state should be reset
        # This will prevent the node constant from overriding the cluster constant
        if applied_constants_dict:
            applied_constants_dm = models.ClusterConstants(applied_constants_dict)
            self._node_fm.reset_constants(applied_constants_dm)

        self._modified_original_constants = None


class SetFipsNodeConfiguration(components.Reversible, components.BlockingApply):
    """Modify a node's solidfire.json to enable/disable fipsDriveConfiguration.

    :param valence.models.ResourceNode node_dm: Modify encryption capability of drives on this node.
    :param bool enabled: True = enable Fips, False = disable Fips
    :param valence.models.ResourceCluster cluster_dm: Optional cluster that nodes belong to.
    """

    def __init__(self, node_dm, enabled=True, cluster_dm=None):
        super(SetFipsNodeConfiguration, self).__init__()
        self.cluster_fm = functional_models.Cluster(cluster_dm) if cluster_dm else None
        self.node_fm = functional_models.Node(node_dm, cluster_dm=cluster_dm)
        self._enabled = enabled
        self._initial_value = None

    def _apply(self, timeout=None):
        """Sets fipsDriveConfiguration in solidfire.json to the specified value.

        :param int timeout: Unused
        """
        self._initial_value = self.node_fm.get_fips_drives_config()
        self.node_fm.set_fips_drives_config(enabled=self._enabled)
        vl_helpers.wait_for_node_api_up(self.node_fm)
        if self.cluster_fm:
            vl_helpers.wait_for_api_up(cluster_fm=self.cluster_fm)
            vl_helpers.wait_for_healthy_services(cluster_fm=self.cluster_fm)

    def _reset(self, timeout=None):
        """Rolls back any changes made to fipsDriveConfiguration to the original values.

        :param int timeout: Unused
        """
        if self._initial_value is not None:
            self.node_fm.set_fips_drives_config(self._initial_value)
        vl_helpers.wait_for_node_api_up(self.node_fm)
        if self.cluster_fm:
            vl_helpers.wait_for_api_up(cluster_fm=self.cluster_fm)
            vl_helpers.wait_for_healthy_services(cluster_fm=self.cluster_fm)


class SetFipsNodeConfigurationMulti(components.Reversible, components.BlockingApply):
    """Modify nodes solidfire.json to enable/disable fipsDriveConfiguration.

    :param list[valence.models.ResourceNode] node_dms: Modify encryption capability of nodes.
    :param bool enabled: True = enable Fips, False = disable Fips
    :param valence.models.ResourceCluster cluster_dm: Optional cluster that nodes belong to.
    """

    def __init__(self, node_dms, enabled=True, cluster_dm=None):
        super(SetFipsNodeConfigurationMulti, self).__init__()
        self.cluster_fm = functional_models.Cluster(cluster_dm) if cluster_dm else None
        self._singular_cmts = [SetFipsNodeConfiguration(n, enabled, cluster_dm) for n in node_dms]

    def _apply(self, timeout=None):
        """Sets fipsDriveConfiguration in solidfire.json to the specified value.

        :param int timeout: Unused
        """
        for cmt in self._singular_cmts:
            cmt.apply_async()
        for cmt in self._singular_cmts:
            cmt.finish()

    def _reset(self, timeout=None):
        """Rolls back any changes made to fipsDriveConfiguration to the original values.

        :param int timeout: Unused
        """
        for cmt in self._singular_cmts:
            cmt.reset()


class FakeDrivesEncryptionCapability(components.Reversible, components.BlockingApply):
    """Fake drives to present a specific encryption capability.

    :param valence.models.ResourceNode node_dm: Modify encryption capability of drives on this node.
    :param valence.enums.DriveEncryptionCapability encryption_capability: Capability to fake.
    :param valence.models.ResourceCluster cluster_dm: Optional cluster that nodes belong to.
    :param list[int] slots: Drive slots to modify. Default=all
    """

    def __init__(self, node_dm, encryption_capability, cluster_dm=None, slots=None):
        super(FakeDrivesEncryptionCapability, self).__init__()
        self.cluster_fm = functional_models.Cluster(cluster_dm) if cluster_dm else None
        self.node_fm = functional_models.Node(node_dm, cluster_dm=cluster_dm)
        self._encryption_capability = encryption_capability
        self._slots = slots
        self._override_path = None

    def _apply(self, timeout=None):
        """Fake drive encryption capability to the specified value.

        :param int timeout: Unused
        """
        self._override_path = self.node_fm.set_fake_drives_encryption_capability(
            self._encryption_capability,
            self._slots
        )
        vl_helpers.wait_for_node_api_up(self.node_fm)
        if self.cluster_fm:
            vl_helpers.wait_for_api_up(cluster_fm=self.cluster_fm)
            vl_helpers.wait_for_healthy_services(cluster_fm=self.cluster_fm)

    def _reset(self, timeout=None):
        """Returns drives to their physical encryption capability.

        :param int timeout: Unused
        """
        if self._override_path:
            with self.node_fm.get_ssh_connection() as ssh_conn:
                ssh_conn.run_command('rm {}'.format(self._override_path), raise_on_failure=True)
            self.node_fm.restart_binary(service=enums.NodeBinary.solidfire)
        vl_helpers.wait_for_node_api_up(self.node_fm)
        if self.cluster_fm:
            vl_helpers.wait_for_api_up(cluster_fm=self.cluster_fm)
            vl_helpers.wait_for_healthy_services(cluster_fm=self.cluster_fm)


class CascadeFailNodes(components.Reversible, components.BlockingApply):
    """Fail one node, wait for bin assignments to publish, then fail another node.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.ActiveNodeSelector node_selector: node selector, default is None
    """
    TIMEOUT = 1200

    def __init__(self, cluster_dm, node_selector=None):
        super(CascadeFailNodes, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._node_selector = (node_selector if node_selector else
                               selectors.ActiveNodeSelector(self._cluster_dm, always_unique=True, num=1))
        self._first_fail_node_cmt = None
        self._second_fail_node_cmt = None

    def _apply_setup(self):
        self._first_fail_node_cmt = _power.PowerOffNodes(self._node_selector, self._cluster_dm)
        self._second_fail_node_cmt = _power.PowerOffNodes(self._node_selector, self._cluster_dm)

    def _apply(self, timeout=None):
        """Fail nodes and verify they show up in failed drives list.

        :param int timeout: timeout used to wait for drive failure
        """
        first_event_id = self._cluster_fm.list_events()[0].event_id
        first_event_id += 1
        self._first_fail_node_cmt.apply()
        vl_helpers.wait_for_bin_assignments(
            self._cluster_fm, timeout=1200, start_event_id=first_event_id)
        first_event_id = self._cluster_fm.list_events()[0].event_id
        first_event_id += 1
        self._second_fail_node_cmt.apply()
        vl_helpers.wait_for_bin_assignments(
            self._cluster_fm, timeout=1200, start_event_id=first_event_id)

    def _reset(self, timeout=None):
        """Restores nodes failed through this component instance.  Waits for recovery.

        Nodes which were in the active state during apply are added to the cluster

        :param int timeout: max time to wait for nodes to restore. Defaults to 12 hours.
        """
        self._second_fail_node_cmt.ensure_ready()
        self._first_fail_node_cmt.ensure_ready()
        vl_helpers.wait_for_bin_sync_to_finish_with_timeout(cluster_fm=self._cluster_fm, timeout=timeout)


class AddNodes(components.Reversible, components.BlockingApply):
    """Add nodes to cluster.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.PendingNodeSelector node_selector: node selector, default is None which
        uses the PendingNodeSelector.
    :param bool auto_install: auto install added node to the cluster's current build
    """
    def __init__(self, cluster_dm, node_selector=None, auto_install=None):
        super(AddNodes, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._node_selector = (node_selector if node_selector else
                               selectors.PendingNodeSelector(self._cluster_dm))
        self._auto_install = auto_install
        self._add_node_async_handlers = None
        self._pending_node_dms = None
        self._available_drive_selector = None
        self._active_node_dms = None
        self._drives_to_add = None

    def _apply_setup(self):
        self._pending_node_dms = self._node_selector()

    def _apply(self, timeout=None):
        self._add_node_async_handlers = self._cluster_fm.add_nodes(
            pending_node_dms=self._pending_node_dms, auto_install=self._auto_install)
        for handler in self._add_node_async_handlers:
            handler.join()

        self._active_node_dms = vl_helpers.convert_pending_nodes_to_active_nodes(
            self._cluster_fm, self._pending_node_dms)
        node_ids = [node_dm.node_id for node_dm in self._active_node_dms]
        self._drives_to_add = self._cluster_fm.list_drives(
            node_ids=node_ids, statuses=[enums.DriveStatus.available])
        vl_helpers.add_drives(self._cluster_fm, self._drives_to_add)

    def _reset(self, timeout=None):
        vl_helpers.remove_drives(self._cluster_fm, self._drives_to_add)
        self._cluster_fm.remove_nodes(self._active_node_dms)


class CascadeAddNodes(components.Reversible, components.BlockingApply):
    """Add one node, wait for bin assignments to publish, then add another node.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param valence.selectors.PendingNodeSelector node_selector: node selector, default is None which
        uses the PendingNodeSelector.
    """
    TIMEOUT = 800

    def __init__(self, cluster_dm, node_selector=None):
        super(CascadeAddNodes, self).__init__()
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self._node_selector = (node_selector if node_selector else
                               selectors.PendingNodeSelector(self._cluster_dm, num=1,
                                                             always_unique=True))
        self._first_add_node_cmt = None
        self._second_add_node_cmt = None

    def _apply_setup(self):
        self._first_add_node_cmt = AddNodes(self._cluster_dm, self._node_selector)
        self._second_add_node_cmt = AddNodes(self._cluster_dm, self._node_selector)

    def _apply(self, timeout=None):
        """Add pending nodes to cluster.

        :param int timeout: timeout used to wait for node failure
        """
        # Add first pending node to cluster
        first_event_id = self._cluster_fm.list_events()[0].event_id
        self._first_add_node_cmt.apply()
        vl_helpers.wait_for_bin_assignments(
            self._cluster_fm, timeout=self.TIMEOUT, start_event_id=first_event_id)

        # Add second pending node to cluster
        first_event_id = self._cluster_fm.list_events()[0].event_id
        self._second_add_node_cmt.apply()
        vl_helpers.wait_for_bin_assignments(
            self._cluster_fm, timeout=self.TIMEOUT, start_event_id=first_event_id)

    def _reset(self, timeout=None):
        """Removes added nodes.

        :param int timeout: max time to wait for nodes to remove. Defaults to 12 hours.
        """
        self._second_add_node_cmt.ensure_ready()
        self._first_add_node_cmt.ensure_ready()
        vl_helpers.wait_for_bin_sync_to_finish_with_timeout(cluster_fm=self._cluster_fm, timeout=timeout)
