"""Valence Account components."""
import autologger

from valence import (
    components,
    enums,
    functional_models,
    selectors,
    vl_helpers,
)

_log = autologger.get_sflogging(__name__)


class PowerOffNodes(components.ProcessedReversible, components.ProcessedBlockingApply):
    """Component used to power off nodes via OOB.

    .apply() will concurrently power off nodes, use .reset() to power back on.
    If drives any drives are kicked out prior to reset(), reset() will add them back.

    :param valence.selectors.ClusteredNodeSelector node_selector: node selector for
        clustered nodes to power off
    :param valence.models.ResourceCluster cluster_dm: cluster containing the nodes to power off
    """
    def __init__(self, node_selector, cluster_dm):
        super(PowerOffNodes, self).__init__()

        self._node_dms = []
        self._node_fms = []
        self._node_selector = node_selector
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm=self._cluster_dm)
        self._initial_active_drive_serials = None
        self._node_ids = None
        self.power_cycle_node_timeout = None
        self.cluster_fault_timeout = 3600
        self.cluster_fault = None

    def _apply_setup(self):
        """Calls node_selector to get current active nodes."""
        self._node_dms = self._node_selector()
        self._node_fms = [functional_models.Node(node_dm.resource_dm, self._cluster_dm,
                                                 authenticate=False)
                          for node_dm in self._node_dms]
        self._node_ids = [node_dm.node_id for node_dm in self._node_dms]
        initial_active_drive_dms = self._cluster_fm.list_drives(node_ids=self._node_ids,
                                                                statuses=[enums.DriveStatus.active])
        self._initial_active_drive_serials = {drive.serial for drive in initial_active_drive_dms}

        self.power_cycle_node_timeout = vl_helpers.get_bin_sync_timeout_for_nodes(
            cluster_fm=self._cluster_fm,
            removed_node_dms=self._node_dms)

    def _reload(self):
        """Attempts to reload the component.

        Check if all nodes provided are in the `OFF` state.

        :return: whether component was able to reload
        :rtype: bool
        """

        nodes_status = [node_fm.get_oob_power_status() for node_fm in self._node_fms]
        if all([_ is enums.OobPowerState.off for _ in nodes_status]):
            _log.debug('Nodes are already in powered `OFF` state.  Reloading.')
            return True
        elif any([_ is enums.OobPowerState.off for _ in nodes_status]):
            raise components.PartialStateError('Not all nodes are in the `OFF` State')
        else:
            return False

    def _apply(self, timeout=None):
        """Power off nodes via OOB.

        :param int timeout: timeout before killing machine control processes
        """

        vl_helpers.power_off_nodes_via_oob(self._node_fms, wait_for_power_down=True)
        # If you power down a witness node, no bin or slice syncing will be expected.
        if enums.NodeRole.storage in [n.get_node_dm().role for n in self._node_fms]:
            vl_helpers.wait_for_slice_sync_to_finish(
                cluster_fm=self._cluster_fm,
                timeout=self.power_cycle_node_timeout)

            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(
                cluster_fm=self._cluster_fm,
                timeout=self.power_cycle_node_timeout)

        self.cluster_fault = vl_helpers.wait_for_cluster_fault_occurrence(
            cluster_fm=self._cluster_fm,
            fault_code=enums.ClusterFaultCode.node_offline,
            only_nodes=self._node_dms,
            timeout=self.cluster_fault_timeout)

        _log.debug('Cluster fault node offline info {}'.format(self.cluster_fault))

    def _reset(self, timeout=None):
        """Power on nodes brought down by this component.

        :param int timeout: timeout before killing machine control processes
        """

        vl_helpers.power_on_nodes_via_oob(self._node_fms, wait_for_power_up=True, wait_for_ping=False,
                                          wait_for_sfapp=True)

        # If any drives are failed, wait for them to become available
        failed_drive_dms = self._cluster_fm.list_drives(node_ids=self._node_ids,
                                                        serials=self._initial_active_drive_serials,
                                                        statuses=[enums.DriveStatus.failed])

        # TODO: VL-2460
        if failed_drive_dms:
            vl_helpers.wait_for_drive_statuses(self._cluster_fm, failed_drive_dms,
                                               statuses=[enums.DriveStatus.active,
                                                         enums.DriveStatus.available])

        # Add available drives back only if they were active before apply()
        drive_dms_to_add = self._cluster_fm.list_drives(node_ids=self._node_ids,
                                                        serials=self._initial_active_drive_serials,
                                                        statuses=[enums.DriveStatus.available])
        if drive_dms_to_add:
            self._cluster_fm.add_drives(drive_dms_to_add)

            vl_helpers.wait_for_slice_sync_to_finish(
                cluster_fm=self._cluster_fm,
                timeout=self.power_cycle_node_timeout)

            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(
                cluster_fm=self._cluster_fm,
                timeout=self.power_cycle_node_timeout)

        vl_helpers.wait_for_cluster_fault_resolution(
            cluster_fm=self._cluster_fm,
            only_nodes=self._node_dms,
            only_fault_codes=[enums.ClusterFaultCode.node_offline,
                              enums.ClusterFaultCode.unresponsive_service],
            timeout=self.cluster_fault_timeout)


class PowerOffNode(PowerOffNodes):
    """Component used to power off nodes via OOB with appropriate syncs.

    .apply() will concurrently power off nodes and then slice sync, bin sync, wait for cluster fault
    occurrence.
    use .reset() to power back on  .
    If drives any drives are kicked out prior to reset(), reset() will add them back.
    """

    def _apply(self, timeout=None):
        """Power off nodes via OOB with bin, slice syncs and cluster faults occurrences.

        :param int timeout: timeout before killing machine control processes.
        """
        vl_helpers.power_off_nodes_via_oob(self._node_fms, wait_for_power_down=True)
        # If you power down a witness node, no bin or slice syncing will be expected.
        if self._node_fms[0].get_node_dm().role is enums.NodeRole.storage:
            vl_helpers.wait_for_slice_sync_to_finish(
             cluster_fm=self._cluster_fm,
             timeout=self.power_cycle_node_timeout)

            vl_helpers.wait_for_bin_sync_to_start(cluster_fm=self._cluster_fm)

            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(
             cluster_fm=self._cluster_fm,
             timeout=self.power_cycle_node_timeout)

        self.cluster_fault = vl_helpers.wait_for_cluster_fault_occurrence(
         cluster_fm=self._cluster_fm,
         fault_code=enums.ClusterFaultCode.node_offline,
         only_nodes=self._node_dms,
         timeout=self.cluster_fault_timeout)

        _log.debug('Cluster fault node offline info {}'.format(self.cluster_fault))

    def _reset(self, timeout=None):
        """Power on nodes brought down by this component.

        :param int timeout: timeout before killing machine control processes.
        """

        vl_helpers.power_on_nodes_via_oob(self._node_fms, wait_for_power_up=True, wait_for_ping=False,
                                          wait_for_sfapp=True)

        drive_dms = self._cluster_fm.list_drives(serials=self._initial_active_drive_serials)

        vl_helpers.recover_failed_drives(cluster_fm=self._cluster_fm,
                                         drive_dms=drive_dms)


class PowerOffCluster(components.ProcessedReversible, components.ProcessedBlockingApply):
    """Component used to power off ALL nodes in a cluster via OOB.

    .apply() will concurrently power off all cluster nodes, use .reset() to power back on.
    .reset() will wait for API to come back up, and for services to be healthy

    :param valence.models.ResourceCluster cluster_dm: cluster containing the nodes to power off
    """

    def __init__(self, cluster_dm):
        super(PowerOffCluster, self).__init__()
        self._node_fms = []
        self._cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm=self._cluster_dm)

    def _apply_setup(self):
        """Calls node_selector to get current active nodes."""
        _node_dms = selectors.ActiveNodeSelector(self._cluster_dm)()
        self._node_fms = [functional_models.Node(node_dm.resource_dm, self._cluster_dm,
                                                 authenticate=False)
                          for node_dm in _node_dms]

    def _reload(self):
        """Attempts to reload the component.

        Check if all nodes provided are in the `OFF` state.
        :return: whether component was able to reload
        :rtype: bool
        """
        nodes_status = [node_fm.get_oob_power_status() for node_fm in self._node_fms]
        if all([_ is enums.OobPowerState.off for _ in nodes_status]):
            _log.debug('Nodes are already in powered `OFF` state.  Reloading.')
            return True
        elif any([_ is enums.OobPowerState.off for _ in nodes_status]):
            raise components.PartialStateError('Not all nodes are in the `OFF` State')
        else:
            return False

    def _apply(self, timeout=None):
        """Power off nodes via OOB.

        :param int timeout: timeout before killing machine control processes
        """
        vl_helpers.power_off_nodes_via_oob(self._node_fms, wait_for_power_down=True)

    def _reset(self, timeout=None):
        """Power on nodes brought down by this component.

        :param int timeout: timeout before killing machine control processes
        """
        vl_helpers.power_on_nodes_via_oob(self._node_fms, wait_for_power_up=True, wait_for_ping=True,
                                          wait_for_sfapp=True)
        vl_helpers.wait_for_api_up(self._cluster_fm)
        vl_helpers.wait_for_healthy_services(self._cluster_fm)
