"""Valence remote replication components."""

import time

import autologger

from sf_api import api_errors
from valence import (
    components,
    enums,
    errors,
    functional_models,
)
from valence.components import _errors

_log = autologger.get_sflogging(__name__)


class PairClusters(components.BlockingApply, components.Reversible):
    """Set up remote replication on a pair of clusters.

    :param valence.models.ResourceCluster cluster_one_dm: Cluster data model
    :param valence.models.ResourceCluster cluster_two_dm: Cluster data model
    """

    def __init__(self, cluster_one_dm, cluster_two_dm):
        """Constructor for remote replication component."""
        super(PairClusters, self).__init__()
        self.cluster_one_dm = cluster_one_dm
        self.cluster_two_dm = cluster_two_dm
        self._cluster_one_fm = functional_models.Cluster(cluster_one_dm)
        self._cluster_two_fm = functional_models.Cluster(cluster_two_dm)
        self._cluster_one_pair_dm = None
        self._cluster_two_pair_dm = None

    @staticmethod
    def _get_pair_dm(cluster_fm_1, cluster_fm_2):
        """Get a specific cluster pair.

        :param valence.functional_models.Cluster cluster_fm_1: Cluster functional model(typically source)
        :param valence.functional_models.Cluster cluster_fm_2: Cluster functional model(typically target)
        :rtype: None|valence.models.ClusterPair
        """
        cluster_pair_dm = None
        for pair in cluster_fm_1.list_cluster_pairs():
            if pair.mvip == cluster_fm_2.cluster_dm.mvip:
                cluster_pair_dm = pair
                break

        return cluster_pair_dm

    def _pair_clusters(self):
        """Helper method to pair clusters.

        :raises valence.components.ComponentStateError: If unable to verify clusters are paired.
        """

        # There are intermittent DBVersionMismatch with pairing (see CSD-2302), which gets auto-retried
        # and will eventually fail with InvalidPairingKey.  So we'll will retry a second time in that
        # case.  If it fails again, we'll re-raise it.
        retry_count = 0
        while True:
            try:
                start_cluster_pair_key = self._cluster_one_fm.start_cluster_pairing()
                self._cluster_two_pair_dm = self._cluster_two_fm.complete_cluster_pairing(
                    start_cluster_pair_key)
            except api_errors.InvalidPairingKey:
                retry_count += 1
                if retry_count > 1:
                    raise
                _log.debug('Waiting 5 seconds before retrying Cluster Pairing.')
                time.sleep(5)
            else:
                break

        self._cluster_one_pair_dm = self._get_pair_dm(self._cluster_one_fm, self._cluster_two_fm)

        if not self._cluster_one_pair_dm or not self._cluster_two_pair_dm:
            raise _errors.ComponentStateError(
                'Could not find cluster pair for cluster: {}.'.format(self._cluster_one_pair_dm))

    def _unpair_cluster(self):
        """Helper method to unpair clusters."""
        self._cluster_one_fm.remove_cluster_pair(self._cluster_one_pair_dm)
        self._cluster_two_fm.remove_cluster_pair(self._cluster_two_pair_dm)

    def _apply(self, timeout=None):
        """Apply remote rep.

        :param int timeout: ignored required timeout
        """
        self._pair_clusters()

    def _reset(self, timeout=None):
        """Reset remote rep.

        :param int timeout: ignored required timeout

        :raises valence.components.ComponentStateError: If unable to verify clusters are paired.
        """
        if self._cluster_one_pair_dm and self._cluster_two_pair_dm:
            self._unpair_cluster()
        else:
            raise _errors.ComponentStateError('Cluster {} and {} are not paired.'.format(
                self.cluster_one_dm, self.cluster_two_dm))

    def _reload(self):
        """Attempt to reload the component.

        :rtype: bool
        """
        # Check if a cluster pair exists
        self._cluster_one_pair_dm = self._get_pair_dm(self._cluster_one_fm, self._cluster_two_fm)
        self._cluster_two_pair_dm = self._get_pair_dm(self._cluster_two_fm, self._cluster_one_fm)
        if not (self._cluster_one_pair_dm and self._cluster_two_pair_dm):
            return False
        else:
            return True


class PairVolumes(components.Reversible, components.BlockingApply):
    """Pair a volume with a peer volume in another Cluster.

    :param valence.models.ResourceCluster cluster_src_dm: source volume Cluster DM
    :param valence.models.Volume volume_src_dm: source volume DM for pairing
    :param valence.models.ResourceCluster cluster_tgt_dm: target volumes cluster DM
    :param valence.models.Volume volume_tgt_dm: target volume DM for pairing
    :param valence.enums.RemoteRepMode mode: Optional, mode to pair the volume with.
    """

    def __init__(self, cluster_src_dm, volume_src_dm, cluster_tgt_dm, volume_tgt_dm,
                 mode=None):
        super(PairVolumes, self).__init__()

        self._cluster_src_dm = cluster_src_dm
        self._cluster_src_fm = functional_models.Cluster(cluster_src_dm)
        self._volume_src_dm = volume_src_dm
        self._volume_pair_src_dm = None

        self._cluster_tgt_dm = cluster_tgt_dm
        self._cluster_tgt_fm = functional_models.Cluster(cluster_tgt_dm)
        self._volume_tgt_dm = volume_tgt_dm
        self._volume_pair_tgt_dm = None
        self._tgt_volume_original_access_mode = None

        self._mode = mode

    @property
    def volume_source_dm(self):
        """Return the Source Volume DM for the Volume Pair.

        :rtype: valence.models.Volume
        :returns: Source Volume Data Model
        """
        return self._volume_src_dm

    @property
    def volume_target_dm(self):
        """Return the Target Volume DM for the Volume Pair.

        :rtype: valence.models.Volume
        :returns: Target Volume Data Model
        """
        return self._volume_tgt_dm

    def _apply(self, timeout=None):
        """Apply the component, creating the volume pair.

        :param int timeout: Not used for this component.

        :raises: valence.components.ComponentStateError: If unable to verify volume pair specified.
        """

        # Ensure that the target volume is in the appropriate access mode for replication.
        if self._volume_tgt_dm.access != enums.VolumeAccess.replication_target:
            self._tgt_volume_original_access_mode = self._volume_tgt_dm.access
            self._cluster_tgt_fm.modify_volumes([self._volume_tgt_dm],
                                                access=enums.VolumeAccess.replication_target)

        # There are intermittent DBVersionMismatch with pairing (see CSD-2302), which gets auto-retried
        # and will eventually fail with InvalidPairingKey.  So we'll will retry a second time in that
        # case.  If it fails again, we'll re-raise it.
        retry_count = 0
        while True:
            try:
                pairing_key = self._cluster_src_fm.start_volume_pairing(self._volume_src_dm,
                                                                        mode=self._mode)
                self._cluster_tgt_fm.complete_volume_pairing(self._volume_tgt_dm, pairing_key)
            except api_errors.InvalidPairingKey:
                retry_count += 1
                if retry_count > 1:
                    raise
                _log.debug('Waiting 5 seconds before retrying Volume Pairing.')
                time.sleep(5)
            else:
                break

        self._volume_pair_src_dm = self._get_volume_pair(self._cluster_src_fm,
                                                         self._volume_src_dm,
                                                         self._volume_tgt_dm)

        if not self._volume_pair_src_dm:
            raise errors.ElementInvalidStateError(
                'Unable to find volume pair: cluster: {}; local: {}; remote: {}.'.format(
                    self._cluster_src_dm,
                    self._volume_src_dm,
                    self._volume_tgt_dm)
            )

        self._volume_pair_tgt_dm = self._get_volume_pair(self._cluster_tgt_fm,
                                                         self._volume_tgt_dm,
                                                         self._volume_src_dm)
        if not self._volume_pair_tgt_dm:
            raise errors.ElementInvalidStateError(
                'Unable to find volume pair: cluster: {}; local: {}; remote: {}.'.format(
                    self._cluster_tgt_fm,
                    self._volume_tgt_dm,
                    self._volume_src_dm)
            )

    def _reset(self, timeout=None):
        """Reset the component, removing the volume pairs.

        :param int timeout: Not used for this component.
        """
        self._cluster_tgt_fm.remove_volume_pair(self._volume_tgt_dm)
        self._cluster_src_fm.remove_volume_pair(self._volume_src_dm)

        if self._tgt_volume_original_access_mode:
            self._cluster_tgt_fm.modify_volumes([self._volume_tgt_dm],
                                                access=self._tgt_volume_original_access_mode)

    def _reload(self):
        """Attempt to reload the component.

        :rtype: bool
        """
        # There is no specific Volume Pair data model
        self._volume_pair_src_dm = self._get_volume_pair(self._cluster_src_fm,
                                                         self._volume_src_dm,
                                                         self._volume_tgt_dm)

        self._volume_pair_tgt_dm = self._get_volume_pair(self._cluster_src_fm,
                                                         self._volume_src_dm,
                                                         self._volume_tgt_dm)

        if not self._volume_pair_src_dm or not self._volume_pair_tgt_dm:
            return False
        else:
            return True

    @staticmethod
    def _get_volume_pair(local_cluster_fm, local_volume_dm, remote_volume_dm):
        """Verify the Volume Pair relationship given local / remote volumes.

        :param valence.functional_models.Cluster local_cluster_fm: The local cluster functional model
        :param valence.models.Volume local_volume_dm: The source volume of the pair.
        :param valence.models.Volume remote_volume_dm: The source volume of the pair.

        :rtype: None|valence.models.VolumePair
        """
        local_volume_id = local_volume_dm.volume_id
        remote_volume_id = remote_volume_dm.volume_id

        active_volume_pairs = local_cluster_fm.list_active_paired_volumes()

        for volume in active_volume_pairs:
            if volume.volume_id == local_volume_id:
                for volume_pair in volume.volume_pairs:
                    if volume_pair.remote_volume_id == remote_volume_id:
                        return volume_pair
        return None


class CreateGroupSnapshotReplication(components.Reversible, components.BlockingApply):
    """Create group snapshot to replicate to a paired cluster.

    This component assumes the volume data models that are passed in are paired with volumes
    on the target cluster. The component will enable remote replication on the group
    snapshot to be replicated to the target cluster.

    :param valence.models.ResourceCluster cluster_src_dm: Source cluster data model.
    :param valence.models.ResourceCluster cluster_tgt_dm: Target cluster data model.
    :param list[valence.models.Volume] volume_dms: Volume data models.
    :param str tag: Unique tag to use for group snapshot.
    :param datetime.timedelta retention: Timedelta object describing lifetime of snapshot
    :param dict attributes: Any additional key value pairs to be added to the group snapshot
    """
    TIMEOUT = 30 * 60
    SYNCING = [enums.RemoteClusterSnapshotStatus.not_present,
               enums.RemoteClusterSnapshotStatus.syncing]

    def __init__(self, cluster_src_dm, cluster_tgt_dm, volume_dms, tag, retention=None, attributes=None):
        super(CreateGroupSnapshotReplication, self).__init__()
        self.volume_dms = volume_dms
        self.tag = tag
        self.retention = retention
        self.attributes = attributes
        self._cluster_src_fm = functional_models.Cluster(cluster_src_dm)
        self._cluster_tgt_fm = functional_models.Cluster(cluster_tgt_dm)
        self._group_snapshot_dm = None

    def _apply(self, timeout=None):
        """Create group snapshot synchronously.

        param int timeout: Time in seconds to wait for the state to apply (not used).
        """
        self._group_snapshot_dm = self._cluster_src_fm.create_group_snapshot(
            volume_dms=self.volume_dms,
            name=self.full_tag,
            enable_remote_replication=True,
            retention=self.retention,
            attributes=self.attributes)

    def _reset(self, timeout=None):
        try:
            src_group_snapshot_dm = self._get_existing_group_snapshot(self._cluster_src_fm)
        except (components.ComponentStateError, api_errors.SnapshotIDDoesNotExist):
            pass
        else:
            self._cluster_src_fm.delete_group_snapshot(src_group_snapshot_dm)

        try:
            tgt_group_snapshot_dm = self._get_existing_group_snapshot(self._cluster_tgt_fm)
        except (components.ComponentStateError, api_errors.SnapshotIDDoesNotExist):
            pass
        else:
            self._cluster_tgt_fm.delete_group_snapshot(tgt_group_snapshot_dm)

    def _get_existing_group_snapshot(self, cluster_fm):
        """Get existing group snapshots.

        :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
        :raises valence.components.ComponentStateError: Group snapshot does not exist
        :returns: Existing group snapshot data models
        :rtype: valence.models.GroupSnapshot
        """
        existing = None
        group_snapshot_dms = cluster_fm.list_group_snapshots(
            group_snapshot_id=self._group_snapshot_dm.group_snapshot_id)
        for group_snapshot_dm in group_snapshot_dms:
            if group_snapshot_dm.group_snapshot_uuid == self._group_snapshot_dm.group_snapshot_uuid:
                existing = group_snapshot_dm
                break
        if not existing:
            raise components.ComponentStateError('Group snapshot does not exist.')
        else:
            return existing

    @property
    def group_snapshot_dm(self):
        """Return existing group snapshot data model.

        :rtype: valence.models.GroupSnapshot
        """
        return self._get_existing_group_snapshot(self._cluster_src_fm)

    @property
    def full_tag(self):
        """Tag used in resource names.

        :returns: A string using the provided unique identifier with -TAG- added
        :rtype: str
        """
        return self.tag + '-TAG-'
