"""Valence Schedule component."""
import datetime
import time

import autologger

from valence import (
    components,
    functional_models,
    models,
    utils,
)
from valence.vl_components import _constants

_log = autologger.get_sflogging(__name__)


class CreateSchedule(components.Reversible, components.BlockingApply):
    """Component to create a schedule.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which to create schedule.
    :param str schedule_name: The name of the schedule.
    :param valence.models.ScheduleInfo schedule_info: The schedule info. Describes what the snapshot is.
    :param valence.enums.ScheduleType schedule_type: The type of schedule. 'Snapshot' only supported.
    :param datetime.datetime starting_date: the ISO 8601 formatted timestamp to start.
    :param list[int] monthdays: The days of the month to schedule a snapshot.
    :param list[valence.models.Weekday] weekdays: The days of the week to schedule a snapshot.
    :param bool recurring: If the schedule should be recurring or now.
    :param bool paused: If it should be paused immediately.
    :param dict attributes: The attributes dictionary.
    :param datetime.time schedule_time: Scheduled time for taking snapshots.

    :raises ValueError: If both weekdays and monthdays arguments are provided.
    """

    def __init__(self, cluster_dm, schedule_name, schedule_info, schedule_type, starting_date=None,
                 monthdays=None, weekdays=None, recurring=None, paused=None, attributes=None,
                 schedule_time=None):
        super(CreateSchedule, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm=cluster_dm, authenticate=False)
        self.schedule_name = schedule_name
        self.schedule_info = schedule_info
        self.schedule_type = schedule_type
        self.monthdays = monthdays
        self.weekdays = weekdays
        self.recurring = recurring
        self.paused = paused
        self.starting_date = (starting_date if starting_date else
                              datetime.datetime.utcnow().isoformat())
        self.schedule_time = schedule_time
        self.schedule = None
        self.list_schedules = None
        self.scheduler_interval_cmt = None
        self.attributes = attributes if attributes is not None else {}

        msg = ('Must have 1 or 0 inputs, for weekdays, monthdays, .'
               'weekdays = {}, monthdays = {} ').format(weekdays, monthdays)
        # Ensure mutual exclusivity between weekdays and monthdays
        if weekdays and monthdays:
            raise ValueError(msg)

        if 'frequency' not in self.attributes:
            if self.weekdays is not None:
                frequency = 'Days of Week'
            elif self.monthdays is not None:
                frequency = 'Days of Month'
            else:
                frequency = 'Time Interval'
            self.attributes['frequency'] = frequency

    def _apply(self, timeout=None):
        """Creates a schedule for snapshot of a volume.

        :param int timeout: Optional timeout in seconds.
        """
        self.schedule = self.cluster_fm.create_schedule(schedule_name=self.schedule_name,
                                                        schedule_info=self.schedule_info,
                                                        schedule_type=self.schedule_type,
                                                        monthdays=self.monthdays,
                                                        weekdays=self.weekdays,
                                                        recurring=self.recurring,
                                                        paused=self.paused,
                                                        attributes=self.attributes,
                                                        schedule_time=self.schedule_time,
                                                        starting_date=self.starting_date
                                                        )

    def _reset(self, timeout=None):
        """Delete the schedule completely from the cluster.

        :param int timeout: Optional timeout in seconds.
        """
        _log.info('Set the constant for scheduler interval sec')
        self.scheduler_interval_cmt = _constants.SetClusterConstants(
            cluster_dm=self.cluster_dm,
            constants_dm=models.ClusterConstants.create(scheduler_interval_seconds=30))
        self.scheduler_interval_cmt.apply()

        self.cluster_fm.modify_schedule(schedule_dm=self.schedule,
                                        to_be_deleted=True)
        timer = utils.timeout_timer(300)
        _log.debug('Poll until the schedules are deleted')
        while next(timer):
            matching_schedules = [x for x in self.cluster_fm.list_schedules()
                                  if x.schedule_name == self.schedule_name]
            if not matching_schedules:
                break
            time.sleep(10)

        _log.info('Revert back the scheduler interval sec constant back to original value')
        self.scheduler_interval_cmt.reset()
