"""Valence security components."""
import time

import autologger
import sf_batch
import sf_ssh

from valence import (
    components,
    enums,
    errors,
    functional_models,
    models,
    vl_helpers,
)
from valence.utils import random_string

_log = autologger.get_sflogging()


class BreakEar(components.Reversible, components.BlockingApply):
    """Break encryption at rest component.

    If no EKM DMs are provided, this component will assume that internal key management
    is being used. It will manipulate the Shamir keys on each node, preventing Element
    from retrieving the EaR user password.

    If EKM DMs are provided, this component will assume that external key management
    is being used. It will build the EKS FMs and run disable() on each of them, preventing
    Element from being able to contact the EKS.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param list[sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster] eks_dms: List of EKS DMs
    :param bool corrupt: If True, corrupt the shamir key files. If False, move to root. N/A for EKM.
    """
    TIMEOUT = 600

    def __init__(self, cluster_dm, eks_dms=None, corrupt=True):
        super(BreakEar, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self.node_fms = []
        self._node_mips = []
        self.eks_dms = eks_dms
        self.eks_fms = self._get_eks_fms()
        self.corrupt = corrupt
        self._unique_id = random_string.generate_random_string(4)
        self._eks_hostnames = [eks.hostname for eks in self.eks_fms]

    def assert_valid(self, check_for_faults=True):
        """Calls _assert_valid() to check current fault state.

        :param bool check_for_faults: If True, wait for expected cluster faults.
        """
        if self.state is components.States.applied:
            # Assert that fault is injected
            _log.debug('Verifying that EaR fault was injected.')
            self._assert_valid(True, check_for_faults)
        else:
            # Assert that fault is removed
            _log.debug('Verifying that EaR fault was reverted.')
            self._assert_valid(False, check_for_faults)

    def _assert_valid(self, expected_state, check_for_faults=True):
        """Verifies that the fault is either applied or reverted depending on expected state.

        For IKM,
            - Confirms shamir keys are in expected state
        For EKM,
            - Confirms the EKS is in expected state

        :param bool expected_state: If True, we expect the fault to have been injected.
        :param bool check_for_faults: If True, wait for expected cluster faults.
        :raises: AssertionError
        """
        error_msgs = []
        if self.eks_dms is None:
            ear_info = self.cluster_fm.get_encryption_at_rest_info()
            for node_fm in self.node_fms:
                with node_fm.get_ssh_connection() as ssh:
                    gen_file = 'gen-{}'.format(ear_info.generation_number)
                    error_msgs += self._check_shamir_files(gen_file, ssh, expected_state)
        else:
            for eks_fm in self.eks_fms:
                if expected_state and eks_fm.is_running():
                    error_msgs.append('EKS {} is still running'.format(eks_fm.hostname))
                elif not expected_state and not eks_fm.is_running():
                    error_msgs.append('EKS {} is not running'.format(eks_fm.hostname))

            if check_for_faults and self._does_key_server_exist():
                error_msgs += self._check_faults(expected_state)

        if len(error_msgs):
            raise AssertionError('\n'.join(error_msgs))
        else:
            _log.debug('EaR fault verification successful.')

    def _does_key_server_exist(self):
        """Returns True if any of the EKS hostnames are found in active Key Servers.

        :rtype: bool
        """
        eks_hostnames = [vl_helpers.get_eks_hostnames(eks_dm) for eks_dm in self.eks_dms]
        key_servers = self.cluster_fm.list_key_servers_kmip()
        for key_server in key_servers:
            if key_server.is_provider_active:
                if key_server.kmip_key_server_hostnames in eks_hostnames:
                    return True
        return False

    def _check_faults(self, expected_state, timeout=600):
        """Check for kmipServerFault faults.

        :param bool expected_state: If True, we expect the fault to have been injected.
        :param int timeout: Seconds to wait for cluster fault
        :rtype: list[str]
        """
        error_msgs = []
        if expected_state:
            try:
                _log.debug('Waiting for kmipServerFault fault.')
                vl_helpers.wait_for_cluster_fault_occurrence(
                    self.cluster_fm,
                    fault_code=enums.ClusterFaultCode.kmip_server_fault,
                    timeout=timeout,
                )
            except errors.TimeoutError:
                error_msgs.append('Timed out waiting for kmipServerFault fault.')
        else:
            try:
                _log.debug('Waiting for kmipServerFault fault to resolve.')
                vl_helpers.wait_for_cluster_fault_resolution(
                    self.cluster_fm,
                    only_fault_codes=[enums.ClusterFaultCode.kmip_server_fault],
                    timeout=timeout,
                )
            except errors.TimeoutError:
                error_msgs.append('Timed out waiting for kmipServerFault fault to resolve.')

        return error_msgs

    def _check_shamir_files(self, gen_file, ssh, expect_original):
        """Check if the shamir key and it's copy exist.

        :param str gen_file: Name of the key file
        :param sf_ssh.Ssh ssh: Node ssh connection
        :param bool expect_original: If True, the shamir key copy is expected to be in root dir
        :rtype: list[str]
        """
        error_msgs = []
        if self.corrupt:
            # Make sure shamir key exists in segments dir
            cmd = 'ls /sf/data/segments/{}'.format(gen_file)
            output = ssh.run_command(cmd)
            if output.return_code != 0:
                error_msgs.append('{} missing from /sf/data/segments/ directory'.format(gen_file))

        cmd = 'ls /{}.original'.format(gen_file)
        output = ssh.run_command(cmd)
        if expect_original:
            if output.return_code != 0:
                error_msgs.append('{}.original missing from root directory'.format(gen_file))
        else:
            if output.return_code == 0:
                error_msgs.append('{}.original was still found in root directory'.format(gen_file))

        return error_msgs

    def _apply_setup(self):
        """Select active nodes just before calling apply()."""
        self.node_fms = self.cluster_fm.get_node_fms(authenticate=False)
        self._node_mips = [node.static_id for node in self.node_fms]

    def _apply(self, timeout=None):
        """Break EaR by preventing DriveManager from retrieving the drive password.

        :param int timeout: N/A
        """
        if self.eks_dms is None:
            if self.corrupt:
                _log.debug('Corrupting shamir key files on Nodes {}'.format(self._node_mips))
                vl_helpers.corrupt_shamir_secret_keys(self.node_fms)
            else:
                _log.debug('Moving shamir key files on Nodes {}'.format(self._node_mips))
                vl_helpers.move_shamir_secret_keys(self.node_fms)
        else:
            _log.debug('Disabling the following EKSs: {}'.format(self._eks_hostnames))
            for eks_fm in self.eks_fms:
                eks_fm.disable()

    def _reset(self, timeout=None):
        """Restore EaR by undoing the action done by apply().

        :param int timeout: N/A
        """
        if self.eks_dms is None:
            _log.debug('Restoring shamir key files on Nodes {}'.format(self._node_mips))
            vl_helpers.restore_shamir_secret_keys(self.node_fms)
        else:
            _log.debug('Enabling the following EKSs: {}'.format(self._eks_hostnames))
            for eks_fm in self.eks_fms:
                eks_fm.enable()

    def _get_eks_fms(self):
        """Build list of EKS functional models from list of EKS data models.

        :rtype: list[Any]
        """
        all_eks_dms = []
        if self.eks_dms:
            for eks_dm in self.eks_dms:
                if vl_helpers.is_clustered_eks(eks_dm):
                    all_eks_dms.extend(eks_dm.nodes)
                else:
                    all_eks_dms.append(eks_dm)
        return [functional_models.get_eks_from_dm(eks_dm) for eks_dm in all_eks_dms]


class EnableEar(components.Reversible, components.NonBlockingApply):
    """Enable encryption at rest component.

    If no ExternalKeyServer data models are provided, this component will enable EaR
    with the regular internal key method using Shamir secret sharing.

    If a list of ExternalKeyServers is provided, then this component will enable EaR
    using EKM. When first applied, this component will setup the Key Provider and Key Server
    objects before enabling EaR. Any subsequent calls to apply will recreate the Key Provider
    and Key Server with the same details and certificates.

    :param valence.models.ResourceCluster cluster_dm: Cluster data model
    :param list[sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster] eks_dms: List of EKS DMs
    :param datetime.datetime start_date: Datetime object representing the client cert start date
    :param datetime.datetime end_date: Datetime object representing the client cert expiration date
    :param bool force_during_bin_sync: Whether to force Ear state changes even during bin sync
    """
    TIMEOUT = 600

    def __init__(self, cluster_dm, eks_dms=None, start_date=None, end_date=None,
                 force_during_bin_sync=None):
        super(EnableEar, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self.node_fms = []
        self.eks_dms = eks_dms
        self.key_provider = None
        self.key_servers = None
        self.ear_thread = None
        self.start_date = start_date
        self.end_date = end_date
        self._force_during_bin_sync = force_during_bin_sync
        self._unique_id = random_string.generate_random_string(4)

    def assert_valid(self, node_dms=None, drive_dms=None, check_for_faults=True, timeout=600):
        """Calls _assert_valid() to check current EaR state.

        :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
        :param list[valence.models.Drive] drive_dms: list of drive models
        :param bool check_for_faults: If True, look for EaR cluster faults.
        :param int timeout: Timeout in seconds
        :raises: errors.TimeoutError|AssertionError
        """
        if self.state is components.States.applied:
            # Assert that EaR is enabled
            _log.debug('Verifying that EaR was successfully enabled.')
            self._assert_valid(True, node_dms, drive_dms, check_for_faults, timeout)
        else:
            # Assert that EaR is disabled
            _log.debug('Verifying that EaR was successfully disabled.')
            self._assert_valid(False, node_dms, drive_dms, check_for_faults, timeout)

    def _assert_valid(self, expected_state, node_dms=None, drive_dms=None, check_for_faults=True,
                      timeout=600):
        """Verifies that EaR was successfully enabled.

        - Check the encryptionAtRestState field
        - Make sure all active drives have security enabled in expected state
        - Make sure all active drives are unlocked
        - Make sure all active drives have no security failures
        - Check for any EaR related cluster faults

        :param bool expected_state: If True, we expect EaR to be enabled.
        :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
        :param list[valence.models.Drive] drive_dms: list of drive models
        :param bool check_for_faults: If True, look for EaR cluster faults.
        :param int timeout: Timeout in seconds
        :raises: errors.TimeoutError|AssertionError
        """
        start_time = time.time()
        if node_dms is None:
            node_dms = self.cluster_fm.list_active_nodes()

        if drive_dms is None:
            drive_dms = self.cluster_fm.list_drives(statuses=[enums.DriveStatus.active])

        # Verify Ear State
        self._verify_ear_state(expected_state, timeout)

        # Verify Shamir Keys
        if expected_state and self.eks_dms is None:
            err_msg = 'Unable to find expected shamir key files on all nodes'
            assert vl_helpers.verify_shamir_secret_keys(self.cluster_fm), err_msg

        # Verify Security Enabled State
        timeout = self._set_timeout(timeout, start_time)
        self._verify_security_enabled_state(node_dms, drive_dms, expected_state, timeout=timeout)

        # Verify Security Feature Enabled State
        timeout = self._set_timeout(timeout, start_time)
        self._verify_security_enabled_state(
            node_dms, drive_dms, expected_state,
            api_method=enums.ClusterMethod.get_cluster_hardware_info,
            timeout=timeout
        )

        # Verify Security Locked State
        timeout = self._set_timeout(timeout, start_time)
        self._verify_security_locked_state(node_dms, drive_dms, False, timeout=timeout)

        # Verify Drive Security Locked State
        timeout = self._set_timeout(timeout, start_time)
        self._verify_security_locked_state(
            node_dms, drive_dms, False,
            api_method=enums.ClusterMethod.get_cluster_hardware_info,
            timeout=timeout
        )

        # Verify Security Fault Reason
        security_faults = self._verify_security_fault_reasons(drive_dms)
        err_msg = 'Some active drives are failed due to a security reason ({})'.format(security_faults)
        assert len(security_faults) == 0, err_msg

        # Verify Cluster Faults
        if check_for_faults:
            ear_cluster_faults = self._verify_ear_cluster_faults()
            err_msg = 'EaR related cluster faults detected: {}'.format(ear_cluster_faults)
            assert len(ear_cluster_faults) == 0, err_msg

        _log.info('EaR verification successful.')

    def _verify_ear_state(self, expected_state, timeout):
        """Check that encryptionAtRestState matched expected value.

        :param bool expected_state: Expected Ear state
        :param int timeout: Timeout in seconds
        :raises: errors.TimeoutError
        """
        expected_ear_state = enums.EarState.enabled if expected_state else enums.EarState.disabled
        vl_helpers.wait_for_ear_state(self.cluster_fm, expected_ear_state, timeout=timeout)

    def _verify_security_enabled_state(self, active_nodes, active_drives, expected_state,
                                       api_method=enums.NodeMethod.get_hardware_info, timeout=600):
        """Check that securityEnabled matches expected value.

        :param list[valence.models.SfNodeBase] active_nodes: list of active or pending node models
        :param list[valence.models.Drive] active_drives: list of drive models
        :param bool expected_state: Expected Ear state
        :param valence.enums.NodeMethod|valence.enums.ClusterMethod api_method: API method to query
        :param int timeout: Timeout in seconds
        :raises: errors.TimeoutError
        """
        if api_method == enums.ClusterMethod.get_cluster_hardware_info:
            field = 'securityFeatureEnabled'
            _log.info('Checking {} field in GetClusterHardwareInfo'.format(field))
        else:
            field = 'securityEnabled'
            _log.info('Checking {} field in GetHardwareInfo'.format(field))

        vl_helpers.wait_for_drive_security_state(
            self.cluster_fm,
            security_state=expected_state,
            node_dms=active_nodes,
            drive_dms=active_drives,
            api_method=api_method,
            timeout=timeout,
        )

    def _verify_security_locked_state(self, active_nodes, active_drives, expected_state,
                                      api_method=enums.NodeMethod.get_hardware_info, timeout=600):
        """Check that securityLocked state matches expected value.

        :param list[valence.models.SfNodeBase] active_nodes: list of active or pending node models
        :param list[valence.models.Drive] active_drives: list of drive models
        :param bool expected_state: Expected Ear state
        :param valence.enums.NodeMethod|valence.enums.ClusterMethod api_method: API method to query
        :param int timeout: Timeout in seconds
        :raises: errors.TimeoutError
        """
        if api_method == enums.ClusterMethod.get_cluster_hardware_info:
            field = 'driveSecurityLocked'
            _log.info('Checking {} field in GetClusterHardwareInfo'.format(field))
        else:
            field = 'securityLocked'
            _log.info('Checking {} field in GetHardwareInfo'.format(field))

        vl_helpers.wait_for_drive_locked_state(
            self.cluster_fm,
            locked_state=expected_state,
            node_dms=active_nodes,
            drive_dms=active_drives,
            api_method=api_method,
            timeout=timeout,
        )

    def _set_timeout(self, timeout, start_time):
        """Reduce the given timeout values based on the current time minus the start time.

        :param int timeout: Max timeout value in seconds
        :param float start_time: Starting time from time.time()
        :raises: errors.TimeoutError
        :rtype: int
        """
        timeout -= (time.time() - start_time)
        if timeout < 0:
            err_msg = 'Timed out during EnableEar.assert_valid()'
            raise errors.TimeoutError(err_msg)
        else:
            return int(timeout)

    def _verify_security_fault_reasons(self, drive_dms):
        """Return list of security fault reasons in the active drives.

        :param list[valence.models.Drive] drive_dms: list of drive models
        :rtype: list[tuple]
        """
        # Grab the latest drive DMs
        drive_ids = [drive.drive_id for drive in drive_dms]
        drive_dms = self.cluster_fm.list_drives(drive_ids=drive_ids)

        failed_drives = []
        for drive in drive_dms:
            if drive.drive_security_fault_reason and drive.drive_security_fault_reason != 'None':
                failed_drives.append((drive.serial, drive.drive_security_fault_reason))

        return failed_drives

    def _verify_ear_cluster_faults(self):
        """Return list of EaR related cluster faults.

        :rtype: list[valence.models.ClusterFault]
        """
        ear_fault_codes = [enums.ClusterFaultCode.enable_drive_security_failed,
                           enums.ClusterFaultCode.disable_drive_security_failed,
                           enums.ClusterFaultCode.bad_encryption_at_rest_state]
        ear_cluster_faults = self.cluster_fm.list_cluster_faults(
            update=True,
            fault_status=enums.ClusterFaultStatus.current,
            include_codes=ear_fault_codes
        )

        return ear_cluster_faults

    def _reload(self):
        """Reload component.

        :raises valence.components.PartialStateError: If EAR is currently being enabled or disabled.
        :rtype: bool
        """
        cluster_info = self.cluster_fm.get_cluster_info()
        state = cluster_info.encryption_at_rest_state
        if state is enums.EarState.enabled:
            return True
        elif state is enums.EarState.disabled:
            return False
        else:
            # State is enabling or disabling
            raise components.PartialStateError('Unable to enable EAR when state is {}'.format(
                state.value))

    def _apply_setup(self):
        """Select active nodes just before calling apply()."""
        self.node_fms = self.cluster_fm.get_node_fms(authenticate=False)

    def _apply_async(self, timeout=None):
        """Enable Ear asynchronously.

        :param int timeout: N/A
        """
        if self.eks_dms is None:
            self.ear_thread = sf_batch.SFThread(self._enable_ear_with_ikm)
        else:
            self.ear_thread = sf_batch.SFThread(self._enable_ear_with_ekm)
        self.ear_thread.start()

    def _apply_is_running(self):
        """Check if the apply thread is still running.

        :rtype: bool
        """
        return not self.ear_thread.is_stopped()

    def _reset(self, timeout=None):
        """Disable Ear.

        :param int timeout: Timeout in seconds
        """
        if self.eks_dms is None:
            self._disable_ear_with_ikm(timeout=timeout)
        else:
            self._disable_ear_with_ekm(timeout=timeout)

    def _apply_finish(self, timeout=None):
        """Finish applying.

        :param int timeout: Timeout in seconds
        """
        self.ear_thread.join(timeout)
        sf_batch.check_for_error_and_raise_exception([self.ear_thread])

    def _enable_ear_with_ikm(self, timeout=None):
        """Wrapper around vl_helpers.enable_encryption_at_rest.

        :param int timeout: Timeout in seconds
        """
        _log.debug('Enabling EaR with internal key management')
        vl_helpers.enable_encryption_at_rest(
            self.cluster_fm,
            force_during_bin_sync=self._force_during_bin_sync,
            timeout=timeout
        )

    def _disable_ear_with_ikm(self, timeout=None):
        """Wrapper around vl_helpers.disable_encryption_at_rest.

        :param int timeout: Timeout in seconds
        """
        if vl_helpers.is_encryption_at_rest_enabled(self.cluster_fm):
            _log.debug('Disabling EAR')
            vl_helpers.disable_encryption_at_rest(
                self.cluster_fm,
                force_during_bin_sync=self._force_during_bin_sync,
                timeout=timeout
            )

    def _enable_ear_with_ekm(self, timeout=None):
        """Setup EKM and enable EaR with key provider.

        :param int timeout: Timeout in seconds
        """
        if self.key_provider and self.key_servers:
            # If apply() was already called once, recreate the provider and servers with
            # the same details. This lets us reuse the same client certificate.
            _log.debug('Recreating key provider and key servers with saved details')
            self._recreate_provider_servers()
        else:
            _log.debug('Setting up key provider and key servers for EKM')
            provider_name = 'EKM_{}'.format(self._unique_id)
            subj = models.CertificateSubject.create(common_name=provider_name)
            self.key_provider = vl_helpers.setup_external_key_management(
                self.cluster_fm,
                eks_dms=self.eks_dms,
                key_provider_name=provider_name,
                key_server_tag=self._unique_id,
                subject=subj,
                start_date=self.start_date,
                end_date=self.end_date,
            )
            self.key_servers = self.cluster_fm.list_key_servers_kmip(
                key_provider_id=self.key_provider.id
            )

        _log.debug('Enabling EaR with external key management')
        vl_helpers.enable_encryption_at_rest(
            self.cluster_fm,
            key_provider_dm=self.key_provider,
            force_during_bin_sync=self._force_during_bin_sync,
            timeout=timeout,
        )

    def _recreate_provider_servers(self):
        """Recreate the key provider and key servers using same details as previous objects."""
        self.key_provider = self.cluster_fm.create_key_provider_kmip(
            key_provider_name=self.key_provider.name
        )

        for key_server in self.key_servers:
            new_key_server = self.cluster_fm.create_key_server_kmip(
                key_server_name=key_server.name,
                key_server_hostnames=key_server.kmip_key_server_hostnames,
                client_certificate=key_server.kmip_client_certificate,
                ca_certificate=key_server.kmip_ca_certificate,
                key_server_port=key_server.kmip_key_server_port,
            )
            self.cluster_fm.add_key_server_to_provider_kmip(new_key_server, self.key_provider)

        self.key_servers = self.cluster_fm.list_key_servers_kmip(
            key_provider_id=self.key_provider.id
        )

    def _disable_ear_with_ekm(self, timeout=300):
        """Disable EaR and teardown EKM key provider and key servers.

        :param int timeout: Timeout in seconds
        """
        if vl_helpers.is_encryption_at_rest_enabled(self.cluster_fm):
            _log.debug('Disabling EAR')
            vl_helpers.disable_encryption_at_rest(
                self.cluster_fm,
                force_during_bin_sync=self._force_during_bin_sync,
                timeout=timeout
            )

        _log.debug('Removing key provider and key servers')
        vl_helpers.teardown_external_key_management(self.cluster_fm, eks_dms=self.eks_dms,
                                                    key_provider_dm=self.key_provider)


class EnableFipsHttps(components.BlockingApply, components.Permanent):
    """Enable FIPS 140-2 certified encryption for HTTPS communications.

    :param valence.models.ResourceCluster cluster_dm: cluster data model.
    """

    def __init__(self, cluster_dm):
        super(EnableFipsHttps, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

    def _apply(self, timeout=360):
        """Enable the FIPS feature, this will cause the mvip to be unavailable for a short time.

        :param int timeout: seconds to wait for cluster to go down / come back up after enabling
                            FIPS.

        :raises components.ComponentStateError: if FIPS is already enabled.
        :raises components.ComponentTimeoutError: if cluster mvip does not go down or come back up
                                                  within timeout seconds.
        """
        if timeout is None:
            timeout = 360

        self.cluster_fm.enable_feature(feature=enums.SfFeature.fips)

        try:
            _log.debug('Waiting for cluster mvip to become unavailable while all nodes reboot')
            sf_ssh.poll_ssh_query(self.cluster_fm.mvip, poll_max=timeout, poll_period=10,
                                  pingable=False)
        except sf_ssh.errors.HostNotPingable:
            raise components.ComponentTimeoutError('Timed out waiting for cluster mvip to become '
                                                   'unavailable after enabling FIPS, which should '
                                                   'force all nodes to reboot.')

        # Wait for mvip to come back up
        try:
            _log.debug('Waiting for cluster mvip to become available again after all nodes reboot')
            sf_ssh.poll_ssh_query(self.cluster_fm.mvip, poll_max=timeout, poll_period=10,
                                  pingable=True)
        except sf_ssh.errors.HostNotPingable:
            raise components.ComponentTimeoutError('Timed out waiting for cluster mvip to become '
                                                   'available again after enabling FIPS.')

        # Check feature status
        feature_status = self.cluster_fm.get_feature_status(feature=enums.SfFeature.fips)
        if not feature_status[0].enabled:
            raise components.ComponentStateError('FIPS failed to enable on the cluster!')

    def _reload(self):
        """Reload the component if FIPS is already enabled.

        :rtype: bool
        """
        feature_status = self.cluster_fm.get_feature_status(feature=enums.SfFeature.fips)
        return True if feature_status[0].enabled else False

    def _reset(self, timeout=None):
        """Cannot reset FIPS.

        :param int timeout: required timeout parameter is ignored.

        :raises components.ComponentStateError: if called.
        """
        raise components.ComponentStateError('This is a permanent component and cannot be reset!')


class CorruptCryptsoftKey(components.Reversible, components.BlockingApply):
    """Modify the current EaR key on the Cryptsoft EKS to an incorrect value.

    Preconditions:
        1) EaR must be enabled.
        2) The given EKS DM is a Cryptsoft server.

    This component uses the key ID in GetEncryptionAtRestInfo to modify the associated
    key value stored in the Cryptsoft server's database. If Element were to attempt to
    query the key server for the key while this component is applied, it will be given
    a key that does not match the key applied to the drives.

    :param valence.models.ResourceCluster cluster_dm: cluster data model.
    :param sf_models.ExternalKeyServer eks_dm: Cryptsoft EKS DM
    """

    def __init__(self, cluster_dm, eks_dm):
        super(CorruptCryptsoftKey, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self.eks_dm = eks_dm
        self.eks_fm = functional_models.get_eks_from_dm(eks_dm)
        self.key = None
        self.key_id = None

    def _apply_setup(self):
        """Raise exception if preconditions are not met."""
        if self.eks_dm.vendor != enums.ExternalKeyServerVendor.cryptsoft:
            raise components.ResourceStateError('EKS DM must be Cryptsoft')

        if not vl_helpers.is_encryption_at_rest_enabled(self.cluster_fm):
            raise components.ResourceStateError('Cluster must have Encryption at Rest enabled')

    def _apply(self, timeout=None):
        """Find and save the current key value, then modify the value in the Cryptsoft database.

        :param int timeout: NA
        """
        self.key_id = self.cluster_fm.get_encryption_at_rest_info().key_id
        self.key = vl_helpers.corrupt_cryptsoft_key(self.eks_fm, self.key_id)

    def _reset(self, timeout=None):
        """Restore the key value in the Cryptsoft database to the original value.

        :param int timeout: NA
        """
        if self.key_id and self.key:
            self.eks_fm.set_key(self.key_id, self.key)
