"""Node FM methods related to time and ntp."""

import autologger

from valence import (
    components,
    functional_models,
)

_log = autologger.get_sflogging(__name__)


class SetNTPInfo(components.Reversible, components.BlockingApply):
    """Set NTP servers and broadcast client for the cluster.

    This component allows you to configure NTP on cluster nodes. The values
    you set with this component apply to all nodes in the cluster. If an NTP
    broadcast server periodically broadcasts time information on your network,
    you can optionally configure nodes as broadcast clients

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm to work on
    :param list[str] ntp_servers: List of NTP servers to add to each node's
        NTP configuration.
    :param bool broadcast_client: Indicates whether or not the nodes in the
        cluster are listening for broadcastNTP
    """

    def __init__(self, cluster_dm, ntp_servers, broadcast_client=False):
        super(SetNTPInfo, self).__init__()
        self.cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        self.ntp_servers = ntp_servers
        self.broadcast_client = broadcast_client
        self.previous_ntp_info = None

    def _apply_setup(self, timeout=None):
        self.previous_ntp_info = self._cluster_fm.get_ntp_info()

    def _apply(self, timeout=None):
        self._cluster_fm.set_ntp_info(servers=self.ntp_servers,
                                      broadcast_client=self.broadcast_client)

    def _reset(self, timeout=None):
        self._cluster_fm.set_ntp_info(servers=self.previous_ntp_info.servers,
                                      broadcast_client=self.previous_ntp_info.broadcast_client)
