"""valence VLAN cluster components.

See below for examples of how to use the VLAN components.

    .. code-block:: python

        from valence import components
        from valence import functional_models

        #create the cluster vlan component
        cluster1_vlan_cmt = CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        # use the manage vlan component to apply and reset the cluster,
         client, and port components together

        vlan_clusters = [cluster1_vlan_cmt]

        # once the client and switchports are implemented they will also be passed
        manage_vlan_cmt = ManageVlan(vlan_clusters)

        manage_vlan_cmt.apply()

        manage_vlan_cmt.reset()
"""

import autologger

from valence import (
    components,
    functional_models,
)

_log = autologger.get_sflogging(__name__)


class CreateClusterVlan(components.Reversible, components.BlockingApply):
    """Component to create a single VLAN interface on a cluster.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which to create VLAN Interface.
    :param str name: string representing the user-defined name for the VLAN.
    :param list[valence.models.AddressBlocks] address_blocks: Range of address blocks assigned
            to the VLAN.
    :param str netmask: Network mask for VLAN.
    :param str svip: Storage IP for VLAN.
    :param int virtual_network_tag: Unique VLAN tag, supported values are 1 through 4094.
    :param str gateway: IP address of the VLAN gateway. Only valid when
            namespace is set to true.
    :param bool namespace: Enables routable storage VLANs functionality.
    """

    def __init__(self, cluster_dm, name, address_blocks, netmask, svip, virtual_network_tag,
                 gateway=None, namespace=False):
        super(CreateClusterVlan, self).__init__()
        self.name = name
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        self.address_blocks = address_blocks
        self.netmask = netmask
        self.svip = svip
        self.virtual_network_tag = virtual_network_tag
        self.gateway = gateway
        self.namespace = namespace
        self.virtual_network_id = None

    def _apply(self, timeout=None):
        """Apply configured set of VLAN, IP range and gateway to the cluster via the API.

        :param int timeout: Optional timeout in seconds.
        """

        self.virtual_network_id = self.cluster_fm.create_virtual_network(
            address_blocks=self.address_blocks,
            name=self.name,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

    def _reset(self, timeout=None):
        """Remove VLAN completely from the cluster.

        :param int timeout: Optional timeout in seconds.
        """

        self.cluster_fm.delete_virtual_network(virtual_network_tag=self.virtual_network_tag)


class ModifyClusterVlan(components.Reversible, components.BlockingApply):
    """Component to modify a single VLAN on the cluster.

    :param valence.models.ResourceCluster cluster_dm: The cluster_dm on which to create VLAN Interface.
    :param int virtual_network_tag: Unique VLAN tag, supported values are 1 through 4094.
    :param str name: string representing the user-defined name for the VLAN.
    :param list[valence.models.AddressBlocks] address_blocks: Range of address blocks assigned
            to the VLAN.
    :param str netmask: Network mask for VLAN.
    :param str svip: Storage IP for VLAN.
    :param str gateway: IP address of the VLAN gateway. Only valid when namespace is set to true.
    :param bool namespace: Enables routable storage VLANs functionality.
    """

    def __init__(self, cluster_dm,  virtual_network_tag, name=None, address_blocks=None, netmask=None,
                 svip=None, gateway=None, namespace=False):
        super(ModifyClusterVlan, self).__init__()
        self.cluster_dm = cluster_dm
        self.cluster_fm = functional_models.Cluster(self.cluster_dm, authenticate=False)
        self.virtual_network_tag = virtual_network_tag
        self.name = name
        self.address_blocks = address_blocks
        self.netmask = netmask
        self.svip = svip
        self.gateway = gateway
        self.namespace = namespace
        self.original_virtual_networks = None

    def _apply(self, timeout=None):
        """Get the current configuration of the specified virtualNetworkID.

        Apply the provided variables to to the cluster.  This will either set or unset the variable
        if it is provided (None implies ignore the variable)

        :param int timeout: Optional timeout in seconds.
        """

        self.original_virtual_networks = self.cluster_fm.list_virtual_networks(
            virtual_network_tags=[self.virtual_network_tag])

        self.cluster_fm.modify_virtual_network(address_blocks=self.address_blocks,
                                               name=self.name,
                                               netmask=self.netmask,
                                               svip=self.svip,
                                               virtual_network_tag=self.virtual_network_tag,
                                               gateway=self.gateway,
                                               namespace=self.namespace)

    def _reset(self, timeout=None):
        """Return the modified variables to the original value.

        :param int timeout: Optional timeout in seconds.
        """

        self.cluster_fm.modify_virtual_network(
            address_blocks=self.original_virtual_networks[0].address_blocks,
            name=self.original_virtual_networks[0].name,
            netmask=self.original_virtual_networks[0].netmask,
            svip=self.original_virtual_networks[0].svip,
            virtual_network_tag=self.original_virtual_networks[0].virtual_network_tag,
            gateway=self.original_virtual_networks[0].gateway,
            namespace=self.original_virtual_networks[0].namespace)


class ManageVlan(components.Reversible, components.BlockingApply):
    """Component to manage a single VLAN.

    This component will eventually also take the client and port components
    as arguments, and then include them in the apply/reset functions.

    :param list[valence.vl_components._vlan.CreateClusterVlan] vlan_cluster_cmts: list of
            clusters on the vlan
    """

    TIMEOUT = 300

    def __init__(self, vlan_cluster_cmts):
        super(ManageVlan, self).__init__()
        self.vlan_cluster_cmts = vlan_cluster_cmts

    def _apply(self, timeout=None):
        """For every component, call the component's apply function.

        :param int timeout: Optional timeout in seconds.
        """

        for vlan_cluster_cmt in self.vlan_cluster_cmts:
            vlan_cluster_cmt.apply()

    def _reset(self, timeout=None):
        """For every component, call the component's ensure_ready function.

        :param int timeout: Optional timeout in seconds.
        """

        for vlan_cluster_cmt in self.vlan_cluster_cmts:
            vlan_cluster_cmt.reset()


class ManageMultipleVlans(components.Reversible, components.BlockingApply):
    """Component to allow easy apply/reset of the collection of components contained in it.

    :param list[valence.vl_components.ManageVlan] vlan_cmts: List of ManageVlan components to apply
        and reset.
    """

    TIMEOUT = 300

    def __init__(self, vlan_cmts):
        super(ManageMultipleVlans, self).__init__()
        self.vlan_cmts = vlan_cmts

    @property
    def child_cmts(self):
        """Get all ManageVlan components.

        :rtype: list[valence.vl_components.ManageVlan]
        """

        child_cmts = []
        for cmt in self.vlan_cmts:
            child_cmts.append(cmt)
        return child_cmts

    def _apply(self, timeout=None):
        """Apply asynchronously all managed components.

        :param int timeout: Optional timeout in seconds.
        """

        for cmt in self.vlan_cmts:
            cmt.apply()

    def _reset(self, timeout=None):
        """Reset all managed components.

        :param int timeout: Optional timeout in seconds.
        """

        for cmt in self.vlan_cmts:
            cmt.reset()
