"""Valence volume components.

See below for examples of how to use the CreateVolumes and CreateMultipleVolumes components.

.. code-block:: python

    from valence.vl_components import account_components, volume_components
    from valence import models

    # First, an accounts component is needed (assume cluster data model was already instantiated)
    accounts_cmt = vl_components.CreateAccounts('example-account', cluster_dm)

    # Create a partial volume data model (component will set the name and account_id)
    partial_volume_dm = models.Volume({"total_size": models.DataSize(2**30)})

    # Create and apply the volumes component
    volume_cmt = volume_components.CreateVolumes('example-vol', cluster_dm,
        partial_volume_dms=[partial_volume_dm], accounts_cmt=accounts_cmt)

    volume_cmt.apply()

    # CreateMultipleVolumes allows you to create multiple volumes with the same profile in one api call
    multi_volume_cmt = vl_components.CreateMultipleVolumes('example-multi-vol', cluster_dm,
        vol_count=5, partial_volume_dm=partial_volume_dm, accounts_cmt=accounts_cmt)

    multi_volume_cmt.apply()

    # Use the ManageVolumes component to apply and reset multiple volumes components together
    manage_volume_cmt = vl_components.ManageVolumes([volume_cmt, multi_volume_cmt])
    manage_volume_cmt.apply()
    manage_volume_cmt.reset()
"""
from __future__ import division

import abc

import autologger
import six

from valence import (
    components,
    functional_models,
    models,
    vl_helpers,
)
from valence.vl_components import (
    _account,
    _cluster,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class VolumesInterface(components.Reversible):
    """Interface component for a set of volumes.

    This class is used for establishing all necessary abstract methods for a set of volumes
    """

    @property
    @abc.abstractmethod
    def child_cmts(self):
        """Get all volume-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components.volume_components.BaseCreateVolumes]
        """

    @property
    @abc.abstractmethod
    def account_dms(self):
        """Get the cached account data models these volumes use.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """

    @property
    @abc.abstractmethod
    def volume_dms(self):
        """Get the cached volume data models.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """

    @property
    @abc.abstractmethod
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

         Usage::

            self.partial_volume_dms = partial_volume_dms
            self._total_volume_bytes = sum(int(vol.total_size) for vol in self.partial_volume_dms)
            self._total_volume_count = len(self.partial_volume_dms)

        :rtype: list[valence.models.Volume]
        """

    @abc.abstractmethod
    def get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """

    @abc.abstractmethod
    def get_volume_dms(self):
        """Get the applied volumes from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """

    @abc.abstractmethod
    def assert_valid(self):
        """Assert the existing volumes match their configuration."""


@six.add_metaclass(abc.ABCMeta)
class BaseCreateVolumes(VolumesInterface):
    """Base Create-Volumes component.

    This is not intended to define a type and is NOT exported. It exists to avoid code duplication for
    the components that actually create volumes.

    :param str tag: the UNIQUE tag to use for these volumes.
    :param valence.models.ResourceCluster cluster_dm: cluster to create volumes on.
    """

    def __init__(self, tag, cluster_dm):
        super(BaseCreateVolumes, self).__init__()
        self.tag = tag
        self.cluster_dm = cluster_dm
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

        self._account_dms = None
        self._volume_dms = None

    @abc.abstractmethod
    def _assert_valid(self, volume_dms=None):
        """Assert the specified volumes are valid.

        :param typing.Iterable[valence.models.Volume] volume_dms: Volume data models to check.
        :raises valence.components.PartialStateError: If invalid volume configurations are found.
        """

    @abc.abstractmethod
    def _get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """

    @property
    def child_cmts(self):
        """Get all volume-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components._volume.BaseCreateVolumes]
        """
        return [self]

    @property
    def full_tag(self):
        """Tag used in resource names.

        :returns: A string using the provided unique identifier wit -TAG- added
        :rtype: str
        """
        return self.tag + '-TAG-'

    @property
    def account_dms(self):
        """Get the cached account data models these volumes use.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError('must be applied to access account dms.')
        return self._account_dms

    @property
    def volume_dms(self):
        """Get the cached volume data models.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError('must be applied to access volume dms.')
        return list(self._volume_dms)

    def get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError('must be applied to access account dms.')
        return self._get_account_dms()

    def get_volume_dms(self):
        """Get the applied volumes from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()
        return self._get_existing()

    def get_allocated_space(self):
        """Get the total amount of space allocated by this component in bytes.

        :rtype: int
        """
        return sum(int(volume.total_size) for volume in self.partial_volume_dms)

    def assert_valid(self):
        """Assert the existing volumes match their configuration."""
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()
        self._assert_valid()

    def _reset(self, timeout=None):
        """Delete the volumes associated with the component.

        If the volumes were found using reload, they will still be deleted

        :param int timeout: Not used since volume deletion is synchronous
        """
        volumes = self._get_existing()
        vl_helpers.delete_and_purge_volumes(self._cluster_fm, volumes)
        self._account_dms = None
        self._volume_dms = None

    def _reload(self):
        """Attempt to reload the component with existing volumes.

        :rtype: bool
        """
        volume_dms = self._get_existing()
        if not volume_dms:
            _log.info('volumes not found on cluster, continuing with apply')
            return False

        self._account_dms = self._get_account_dms()
        try:
            self._assert_valid(volume_dms)
        except AssertionError as e:
            raise components.PartialStateError(e)

        self._volume_dms = volume_dms
        _log.info('volumes with tag {} were reloaded'.format(self.tag))
        return True

    def _get_existing(self):
        """Get existing volumes.

        :rtype: tuple[valence.models.Volume]
        """
        volume_dms = sorted(self._cluster_fm.list_volumes(), key=lambda v: v.name)
        return tuple(dm for dm in volume_dms if dm.name.startswith(self.full_tag))

    @staticmethod
    def _is_config_match(volume_dm, partial_volume_dm):
        """Check whether the specified volume is a match for the provided configuration.

        :param valence.models.Volume volume_dm: Volume to check.
        :param valence.models.Volume partial_volume_dm: Volume configuration against which to check.
        :rtype: bool
        """
        config_match = True
        if partial_volume_dm.total_size.carbon_size != volume_dm.total_size:
            config_match = False
        elif partial_volume_dm.qos is not None:
            if config_match and partial_volume_dm.qos.min_iops is not None:
                config_match = partial_volume_dm.qos.min_iops == volume_dm.qos.min_iops
            if config_match and partial_volume_dm.qos.max_iops is not None:
                config_match = partial_volume_dm.qos.max_iops == volume_dm.qos.max_iops
            if config_match and partial_volume_dm.qos.burst_iops is not None:
                config_match = partial_volume_dm.qos.burst_iops == volume_dm.qos.burst_iops

        return config_match

    @staticmethod
    def _raise_validation_errors(volumes_invalid_account=None, volumes_invalid_config=None,
                                 configs_not_found=None):
        """Raise appropriate errors for lists of invalid volumes.

        :param list[valence.models.Volume] volumes_invalid_account: Volumes with invalid account IDs.
        :param list[valence.models.Volume] volumes_invalid_config: Volumes with invalid configurations.
        :param list[valence.models.Volume] configs_not_found: Configurations not found.
        :raises AssertionError: If invalid volumes are provided.
        """
        error_msgs = []
        if volumes_invalid_account:
            error_msgs.append(
                'Volumes found with invalid account_ids: {}'.format(volumes_invalid_account)
            )
        if volumes_invalid_config:
            error_msgs.append(
                'Volumes found with invalid configurations: {}'.format(volumes_invalid_config)
            )
        if configs_not_found:
            error_msgs.append('Volume configurations not found: {}'.format(configs_not_found))

        if error_msgs:
            raise AssertionError('\n'.join(error_msgs))


class CreateVolumes(BaseCreateVolumes, components.BlockingApply):
    """Create Multiple Volumes for a single account with different volume profiles.

    :param str tag: the UNIQUE tag to use for these volumes.
    :param valence.models.ResourceCluster cluster_dm: cluster to create volumes on.
    :param list[valence.models.Volume] partial_volume_dms: partial Volume data models that will be
        used for creating new volumes. The name and id will be ignored, but the other settings will
        be used.
    :param valence.vl_components.AccountsInterface accounts_cmt: The account component that contains the
        account to attach the created volumes.
    :param int account_index: The index of the account to be used in the list returned by
        accounts_cmt.get_account_dms()

    :raises valence.components.ResourceStateError: If size is not specified in partial_volume_dm
    """

    def __init__(self, tag, cluster_dm, partial_volume_dms, accounts_cmt, account_index=0):
        super(CreateVolumes, self).__init__(tag, cluster_dm)
        self.accounts_cmt = accounts_cmt
        self._account_index = account_index
        self._partial_volume_dms = partial_volume_dms
        self._vol_count = len(self._partial_volume_dms)

        for volume_dm in self._partial_volume_dms:
            if not isinstance(volume_dm, models.Volume):
                raise TypeError(
                        'All partial_volume_dms must be Volume type instead of {}'.format(
                                type(volume_dm))
                )
            if not volume_dm.total_size:
                raise components.ResourceStateError(
                        'Total size not specified for all partial_volume_dms'
                )

    @staticmethod
    def mock(tag, cluster_dm, partial_volume_dms, accounts_cmt, account_index=0):
        """Mock the CreateVolumes.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._volume.CreateVolumes',
                        vl_components.CreateVolumes.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._volume.MockVolumes
        """
        volume_dms = [models.Volume(dm.serialize()) for dm in partial_volume_dms]
        for i, dm in enumerate(volume_dms):
            dm.name = '{}-TAG-{}'.format(tag, i)
            dm.volume_id = i
        return MockVolumes(tag, cluster_dm, volume_dms, accounts_cmt, account_index=account_index)

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        return self._partial_volume_dms

    def _get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        return [self.accounts_cmt.get_account_dms()[self._account_index]]

    def _apply(self, timeout=None):
        """Create the Volumes and verify they were created on cluster properly.

        :param int timeout: Not used since volume deletion is synchronous
        """
        self.accounts_cmt.ensure_applied()

        volume_dms = []
        for i, partial_volume_dm in enumerate(self._partial_volume_dms):
            # Make a copy so that the existing data models are not modified
            volume_dm = models.Volume(partial_volume_dm.serialize())
            volume_dm.name = self.full_tag + str(i)
            volume_dm.account_id = self.accounts_cmt.account_dms[self._account_index].account_id
            volume_dms.append(volume_dm)
        vl_helpers.create_volumes_individual_profiles(self._cluster_fm, volume_dms=volume_dms)

        self._account_dms = self._get_account_dms()
        self._volume_dms = self._get_existing()
        self._assert_valid(self._volume_dms)

    def _reload(self):
        """Attempt to reload the component with existing volumes.

        :rtype: bool
        """
        self.accounts_cmt.ensure_applied()
        return super(CreateVolumes, self)._reload()

    def _assert_valid(self, volume_dms=None):
        """Assert the specified volumes are valid.

        :param typing.Iterable[valence.models.Volume] volume_dms: Volume data models to check.
        :raises valence.components.PartialStateError: If invalid volume configurations are found.
        """
        account_id = self.accounts_cmt.account_dms[self._account_index].account_id
        if volume_dms is None:
            volume_dms = self.get_volume_dms()

        matched_volume_ids = []
        configs_not_found = []
        for partial_dm in self._partial_volume_dms:
            match_found = False
            for volume_dm in volume_dms:
                if volume_dm.volume_id not in matched_volume_ids:
                    if self._is_config_match(volume_dm, partial_dm):
                        match_found = True
                        matched_volume_ids.append(volume_dm.volume_id)
                        break

            if not match_found:
                configs_not_found.append(partial_dm)

        self._raise_validation_errors(
            volumes_invalid_account=[dm for dm in volume_dms if dm.account_id != account_id],
            volumes_invalid_config=[dm for dm in volume_dms if dm.volume_id not in matched_volume_ids],
            configs_not_found=configs_not_found
        )


class CreateMultipleVolumes(BaseCreateVolumes, components.BlockingApply):
    """Create Multiple Volumes for a single account with all having the same volume profile.

    This component differs from CreateVolumes in that it uses the CreateMultipleVolumes call to create
    multiple volumes with a single API call.

    :param str tag: the UNIQUE tag to use for these volumes.
    :param valence.models.ResourceCluster cluster_dm: cluster to create volumes on.
    :param int vol_count: the number of volumes to create
    :param valence.models.Volume partial_volume_dm: partial Volume data model that will be used
        for creating new volumes. The name and id will be ignored, but the other settings will be used.
    :param valence.vl_components.AccountsInterface accounts_cmt: The account component that contains the
        account to attach the created volumes.
    :param int account_index: The index of the account to be used in the list returned by
        accounts_cmt.get_account_dms()

    :raises components.ResourceStateError: If size is not specified in partial_volume_dm
    """

    def __init__(self, tag, cluster_dm, vol_count, partial_volume_dm, accounts_cmt, account_index=0):
        super(CreateMultipleVolumes, self).__init__(tag, cluster_dm)
        self.accounts_cmt = accounts_cmt
        self._account_index = account_index
        self._partial_volume_dm = partial_volume_dm
        self._vol_count = vol_count
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

        if not isinstance(self._partial_volume_dm, models.Volume):
            raise TypeError('partial_volume_dm must be a Volume type instead of {}'
                            .format(type(self._partial_volume_dm)))
        if not self._partial_volume_dm.total_size:
            raise components.ResourceStateError('Total size not specified for partial_volume_dm')

    @staticmethod
    def mock(tag, cluster_dm, vol_count, partial_volume_dm, accounts_cmt, account_index=0):
        """Mock the CreateMultipleVolumes.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._volume.CreateMultipleVolumes',
                        vl_components.CreateMultipleVolumes.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._volume.MockVolumes
        """
        volume_dms = [models.Volume(partial_volume_dm.serialize()) for _ in range(vol_count)]
        for i in range(vol_count):
            volume_dms[i].name = '{}-TAG-{}'.format(tag, i)
            volume_dms[i].volume_id = i
        return MockVolumes(tag, cluster_dm, volume_dms, accounts_cmt, account_index=account_index)

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        return [self._partial_volume_dm for _ in range(self._vol_count)]

    def _get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        return [self.accounts_cmt.get_account_dms()[self._account_index]]

    def _apply(self, timeout=None):
        """Create the Volumes and verify they were created on cluster properly.

        :param int timeout: Not used since volume deletion is synchronous
        """
        self.accounts_cmt.ensure_applied()

        # Make a copy so that the existing data models are not modified
        volume_dm = models.Volume(self._partial_volume_dm.serialize())
        volume_dm.name = self.full_tag
        volume_dm.account_id = self.accounts_cmt.get_account_dms()[self._account_index].account_id
        vl_helpers.create_volumes_single_profile(
            cluster_fm=self._cluster_fm,
            volume_dm=volume_dm,
            number_of_volumes=self._vol_count
        )

        self._account_dms = self._get_account_dms()
        self._volume_dms = self._get_existing()
        self._assert_valid(self._volume_dms)

    def _reload(self):
        """Attempt to reload the component with existing volumes.

        :rtype: bool
        """
        self.accounts_cmt.ensure_applied()
        return super(CreateMultipleVolumes, self)._reload()

    def _assert_valid(self, volume_dms=None):
        """Assert the specified volumes are valid.

        :param typing.Iterable[valence.models.Volume] volume_dms: Volume data models to check.
        :raises valence.components.PartialStateError: If invalid volume configurations are found.
        """
        account_id = self.accounts_cmt.account_dms[self._account_index].account_id
        if volume_dms is None:
            volume_dms = self.get_volume_dms()

        volumes_invalid_config = []
        for volume_dm in volume_dms:
            if not self._is_config_match(volume_dm, self._partial_volume_dm):
                volumes_invalid_config.append(volume_dm)

        self._raise_validation_errors(
            volumes_invalid_account=[dm for dm in volume_dms if dm.account_id != account_id],
            volumes_invalid_config=volumes_invalid_config
        )
        assert len(volume_dms) == self._vol_count


class CloneVolumes(BaseCreateVolumes, components.NonBlockingApply):
    """Create some clones based on some already existing volumes from the component.

    :param str tag: the UNIQUE tag to use for these volumes.
    :param valence.models.ResourceCluster cluster_dm: cluster to clone volumes on.
    :param valence.vl_components.VolumesInterface volumes_cmt: Volumes to clone.
    :param valence.vl_components.AccountsInterface accounts_cmt: Accounts to use for the cloned volumes.
    :param int account_index: The index of the account to be used in the list returned by
        accounts_cmt.get_account_dms()
    """
    CLONE_VOLUME_ID = 'clonedVolumeID'

    def __init__(self, tag, cluster_dm, volumes_cmt, accounts_cmt, account_index=0):
        super(CloneVolumes, self).__init__(tag, cluster_dm)
        self.accounts_cmt = accounts_cmt
        self._account_index = account_index
        self._vol_cmt = volumes_cmt  # hide the internal volumes!
        self._async_fms = []

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        return self._vol_cmt.partial_volume_dms

    def _get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        return [self.accounts_cmt.get_account_dms()[self._account_index]]

    def _apply_async(self, timeout=None):
        """Attempt to clone some volumes.

        :param int timeout: N/A
        """
        self.accounts_cmt.ensure_applied()
        self._vol_cmt.ensure_applied()

        for i, vol in enumerate(self._vol_cmt.volume_dms):
            self._async_fms.append(self._cluster_fm.clone_volume(
                volume_dm=vol,
                name=self.full_tag + str(i),
                account_id=self.accounts_cmt.account_dms[self._account_index].account_id,
                attributes={self.CLONE_VOLUME_ID: vol.volume_id},
            ))

    def _apply_is_running(self):
        """Use async handler to check if apply is still running.

        :rtype: bool
        """
        return any([async_fm.is_running() for async_fm in self._async_fms])

    def _apply_finish(self, timeout=None):
        """Finish applying by setting the volumes.

        :param int timeout: Timeout in seconds
        """
        vols = []
        # Experimentally it takes about 2 minutes to clone 200 volumes on a sodium cluster. Lets say it
        # takes 1 second per volume so for a timeout scaling number of volumes by 5 seems resonable.
        timeout_by_total_vol_num = len(self._vol_cmt.volume_dms) * 5
        timeout = timeout_by_total_vol_num if timeout is None else timeout
        for async_fm in self._async_fms:
            vols.extend(async_fm.get_models(blocking=True, timeout=timeout))

        self._account_dms = self._get_account_dms()
        self._volume_dms = vols
        self._assert_valid(self._volume_dms)

    def _reload(self):
        """Attempt to reload the component with existing volumes.

        :rtype: bool
        """
        self.accounts_cmt.ensure_applied()
        self._vol_cmt.ensure_applied()
        return super(CloneVolumes, self)._reload()

    def _assert_valid(self, volume_dms=None):
        """Assert the specified volumes are valid clones.

        :param typing.Iterable[valence.models.Volume] volume_dms: Volume data models to check.
        :raises valence.components.PartialStateError: If invalid volume configurations are found.
        """
        account_id = self.accounts_cmt.account_dms[self._account_index].account_id
        if volume_dms is None:
            volume_dms = self.get_volume_dms()

        matched_volumes = []
        clones_invalid_config = []
        for clone_dm in volume_dms:
            config_match = False
            for volume_dm in self._vol_cmt.volume_dms:
                if volume_dm not in matched_volumes:
                    clone_id = clone_dm.attributes.get(self.CLONE_VOLUME_ID)
                    if clone_id == volume_dm.volume_id and self._is_config_match(clone_dm, volume_dm):
                        matched_volumes.append(volume_dm)
                        config_match = True
                        break
            if not config_match:
                clones_invalid_config.append(clone_dm)

        self._raise_validation_errors(
            volumes_invalid_account=[dm for dm in volume_dms if dm.account_id != account_id],
            volumes_invalid_config=clones_invalid_config
        )
        if len(matched_volumes) != len(self._vol_cmt.volume_dms):
            raise AssertionError(
                'No clone found for volumes: {}'
                .format([dm for dm in self._vol_cmt.volume_dms if dm not in matched_volumes])
            )


class CreateVolumesForCapacity(BaseCreateVolumes, components.BlockingApply):
    """Fills the cluster to the specified block/metadata fullness.

    Creates and manages volumes to reach the desired fullness.

    NOTE: Any IO to these volumes (once filled) can impact the state of fullness. For best results
    fill/drive IO with the same dedup/compression ratios as specified here. To keep fullness stable
    while driving IO, use dedup/compression ratios of 1.

    :param str tag: The UNIQUE tag to use for associated volumes and accounts.
    :param valence.models.ResourceCluster cluster_dm: Cluster model.
    :param valence.selectors.ClusterBlockCapacitySelector capacity_selector: Block capacity selector.
    :param float comp_ratio: Compression ratio to target when filling.
    :param float dedup_ratio: Deduplication ratio to target when filling.
    :param int vol_count: Number of volumes to create.
    :param valence.models.DataSize min_size: Minimum volume size.
    :param valence.models.DataSize max_size: Maximum volume size.
    :param bool under_size: Whether to under-allocate volume space for the desired capacity (default:
        over-allocates space for the desired capacity).
    """
    DEFAULT_VOL_COUNT = 60

    def __init__(self, tag, cluster_dm, capacity_selector, comp_ratio=1., dedup_ratio=1., vol_count=None,
                 min_size=None, max_size=None, under_size=False):
        super(CreateVolumesForCapacity, self).__init__(tag, cluster_dm)
        self.capacity_selector = capacity_selector
        self.comp_ratio = comp_ratio
        self.dedup_ratio = dedup_ratio
        self.vol_count = vol_count
        self._vol_size_min = min_size
        self._vol_size_max = max_size
        self._under_size = under_size
        self._cluster_constants = None
        self._immutable_values = None
        self._limits = None
        self._block_data_size = None
        self._total_size = None
        self._vol_count = None
        self._vol_size = None

        self._cluster_fm = functional_models.Cluster(self.cluster_dm)
        self._gc_cmt = _cluster.RunGarbageCollection(cluster_dm)
        self._accounts_cmts = []

        assert self.comp_ratio >= 1
        assert self.dedup_ratio >= 1

    @classmethod
    def mock(cls, tag, cluster_dm, capacity_selector, comp_ratio=1., dedup_ratio=1., vol_count=None,
             min_size=None, max_size=None, under_size=False):
        """Mock the CreateVolumesForCapacity.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._volume.CreateVolumesForCapacity',
                        vl_components.CreateVolumesForCapacity.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._volume.MockVolumes
        """
        cluster_fm = functional_models.Cluster(cluster_dm)
        cluster_limits = cluster_fm.get_limits()

        total_volume_size = vl_helpers.get_block_volume_size(
            cluster_fm=cluster_fm,
            block_data_size=capacity_selector(),
            comp_ratio=comp_ratio,
            dedup_ratio=dedup_ratio
        )
        if not vol_count:
            vol_count = vl_helpers.get_volume_count_for_capacity(
                cluster_fm=cluster_fm,
                block_data_size=total_volume_size,
                default_count=cls.DEFAULT_VOL_COUNT,
                volume_size_min=min_size,
                volume_size_max=max_size,
                cluster_limits=cluster_limits
            )

        partial_volume_dm = models.Volume.create(
            total_size=vl_helpers.get_volume_size_for_capacity(
                cluster_fm=cluster_fm,
                block_data_size=total_volume_size,
                volume_count=vol_count,
                under_size=under_size,
                cluster_limits=cluster_limits
            )
        )
        volume_dms = [models.Volume(partial_volume_dm.serialize()) for _ in range(vol_count)]
        for i in range(vol_count):
            volume_dms[i].name = '{}-TAG-{}'.format(tag, i)
            volume_dms[i].volume_id = i

        return MockVolumes(
            tag=tag,
            cluster_dm=cluster_dm,
            volume_dms=volume_dms,
            accounts_cmt=_account.CreateAccounts.mock(tag, cluster_dm),
            comp_ratio=comp_ratio,
            dedup_ratio=dedup_ratio
        )

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        if self.state is components.States.applied:
            return list(self._volume_dms)
        else:
            raise components.ComponentStateError(
                '{} cannot provide partial_volume_dms if not applied'.format(self.__class__.__name__)
            )

    @property
    def block_capacity(self):
        """Block data capacity of the volumes.

        Size of the block data that will be generated by filling these volumes with the configured ratios
        of data compression and deduplication.

        :rtype: valence.models.DataSize
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Must be applied to determine capacity')

        return vl_helpers.get_volume_block_size(
            cluster_fm=self._cluster_fm,
            volume_size=self._vol_size,
            volume_count=self._vol_count,
            comp_ratio=self.comp_ratio,
            dedup_ratio=self.dedup_ratio,
            cluster_constants=self._cluster_constants,
            immutable_values=self._immutable_values
        )

    @property
    def meta_capacity(self):
        """Metadata capacity of the volumes.

        Size of the metadata that will be generated by filling these volumes with the configured ratio
        of data compression.

        :rtype: valence.models.DataSize
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Must be applied to determine capacity')

        return vl_helpers.get_block_metadata_size(
            cluster_fm=self._cluster_fm,
            block_data_size=self.block_capacity,
            comp_ratio=self.comp_ratio,
            dedup_ratio=self.dedup_ratio,
            cluster_constants=self._cluster_constants,
            immutable_values=self._immutable_values
        )

    def _get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        account_dms = []
        for cmt in self._accounts_cmts:
            account_dms.extend(cmt.get_account_dms())
        return account_dms

    def _apply_setup(self):
        """Determine volume count/size required to achieve the desired capacity."""
        if self.capacity_selector.relative:
            vl_helpers.update_cluster_full_threshold(self._cluster_fm, self._gc_cmt)
        data_size = self.capacity_selector()
        assert data_size.bytes > 0

        # Get cluster constants and other reference values
        self._cluster_constants = self._cluster_fm.get_constants()
        self._immutable_values = self._cluster_fm.get_immutable_values()
        self._limits = self._cluster_fm.get_limits()

        # Get volume size/count
        self._total_size = vl_helpers.get_block_volume_size(
            cluster_fm=self._cluster_fm,
            block_data_size=data_size,
            comp_ratio=self.comp_ratio,
            dedup_ratio=self.dedup_ratio,
            cluster_constants=self._cluster_constants,
            immutable_values=self._immutable_values
        )

        if self.vol_count:
            self._vol_count = self.vol_count
        else:
            self._vol_count = vl_helpers.get_volume_count_for_capacity(
                cluster_fm=self._cluster_fm,
                block_data_size=self._total_size,
                default_count=self.DEFAULT_VOL_COUNT,
                volume_size_min=self._vol_size_min,
                volume_size_max=self._vol_size_max,
                cluster_limits=self._limits
            )

        self._vol_size = vl_helpers.get_volume_size_for_capacity(
            cluster_fm=self._cluster_fm,
            block_data_size=self._total_size,
            volume_count=self._vol_count,
            under_size=self._under_size,
            cluster_limits=self._limits
        )

    def _apply(self, timeout=None):
        """Create the Volumes and verify they were created on cluster properly.

        :param int timeout: No effect.
        """
        tag_index = 0
        vols_remaining = self._vol_count
        while vols_remaining:
            self._accounts_cmts.append(_account.CreateAccounts(
                tag='{}-{}'.format(self.tag, tag_index),
                cluster_dm=self.cluster_dm
            ))
            self._accounts_cmts[-1].apply()
            vol_count = min(vols_remaining, self._limits.volumes_per_account_count_max)
            volume_dm = models.Volume.create(
                name='{}{}-'.format(self.full_tag, tag_index),
                account_id=self._accounts_cmts[-1].account_dms[0].account_id,
                total_size=self._vol_size
            )
            vl_helpers.create_volumes_single_profile(
                cluster_fm=self._cluster_fm,
                volume_dm=volume_dm,
                number_of_volumes=vol_count,
                skip_fullness_check=True
            )
            tag_index += 1
            vols_remaining -= vol_count

        self._account_dms = self._get_account_dms()
        self._volume_dms = self._get_existing()
        self._assert_valid(self._volume_dms)

    def _reset(self, timeout=None):
        """Delete all volumes and run garbage collection.

        :param int timeout: No effect.
        """
        super(CreateVolumesForCapacity, self)._reset(timeout)
        for cmt in self._accounts_cmts:
            cmt.ensure_ready()
        self._accounts_cmts = []
        self._block_data_size = None
        self._vol_count = None
        self._vol_size = None

    def _get_existing(self):
        """Get existing volumes/accounts.

        :returns: Volume data models that exist on cluster with the unique tag in the name.
        :rtype: tuple[valence.models.Volume]
        """
        if not self._accounts_cmts:
            while True:
                accounts_cmt = _account.CreateAccounts(
                    tag='{}-{}'.format(self.tag, len(self._accounts_cmts)),
                    cluster_dm=self.cluster_dm
                )
                if accounts_cmt._reload():  # pylint: disable=protected-access
                    accounts_cmt.apply()
                    self._accounts_cmts.append(accounts_cmt)
                else:
                    break

        return super(CreateVolumesForCapacity, self)._get_existing()

    def _assert_valid(self, volume_dms=None):
        """Assert the specified volumes are valid.

        :param typing.Iterable[valence.models.Volume] volume_dms: Volume data models to check.
        :raises valence.components.PartialStateError: If invalid volume configurations are found.
        """
        if volume_dms is None:
            volume_dms = self.get_volume_dms()

        volume_dms_invalid_size = [
            dm for dm in volume_dms if dm.total_size != self._vol_size.carbon_size
        ]

        error_msgs = []
        if volume_dms_invalid_size:
            error_msgs.append(
                'Volumes found with invalid total_size: {}'.format(volume_dms_invalid_size)
            )
        if len(volume_dms) != self._vol_count:
            error_msgs.append(
                'Incorrect number of volumes found. Expected: {} Actual: {}'
                .format(self._vol_count, len(volume_dms))
            )

        if error_msgs:
            raise AssertionError('\n'.join(error_msgs))


class ManageVolumes(VolumesInterface, components.BlockingApply):
    """Manage a list of volumes components.

    The child components will be applied and reset together. This component exposes an API
    that looks like a single volumes component.

    :param list[valence.vl_components._volume.VolumesInterface] volumes_cmts: list of volumes
        components.
    """

    def __init__(self, volumes_cmts):
        super(ManageVolumes, self).__init__()
        self.volumes_cmts = volumes_cmts

    @property
    def child_cmts(self):
        """Get all volume-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components._volume.BaseCreateVolumes]
        """
        child_cmts = []
        for cmt in self.volumes_cmts:
            child_cmts.extend(cmt.child_cmts)
        return child_cmts

    @property
    def account_dms(self):
        """Get the cached account data models these volumes use.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        accounts_map = {}
        for cmt in self.volumes_cmts:
            cmt_accounts_map = {dm.account_id: dm for dm in cmt.account_dms}
            accounts_map.update(cmt_accounts_map)
        return list(six.itervalues(accounts_map))

    @property
    def volume_dms(self):
        """Get the cached volume data models.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """
        volume_dms = []
        for cmt in self.volumes_cmts:
            volume_dms.extend(cmt.volume_dms)
        return volume_dms

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        partial_volume_dms = []
        for vol in self.volumes_cmts:
            partial_volume_dms.extend(vol.partial_volume_dms)
        return partial_volume_dms

    def get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        accounts_map = {}
        for cmt in self.volumes_cmts:
            cmt_accounts_map = {dm.account_id: dm for dm in cmt.get_account_dms()}
            accounts_map.update(cmt_accounts_map)
        return list(six.itervalues(accounts_map))

    def get_volume_dms(self):
        """Get the applied volumes from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """
        volume_dms = []
        for cmt in self.volumes_cmts:
            volume_dms.extend(cmt.get_volume_dms())
        return volume_dms

    def get_allocated_space(self):
        """Get the total amount of space allocated by this component in bytes.

        :rtype: int
        """
        return sum(int(volume.total_size) for volume in self.partial_volume_dms)

    def assert_valid(self):
        """Assert the existing volumes match their configuration."""
        for cmt in self.volumes_cmts:
            cmt.assert_valid()

    def _apply(self, timeout=None):
        """Apply all sub components.

        :param int timeout: Not used for this component.
        """
        for cmt in self.volumes_cmts:
            cmt.ensure_applied()

    def _reset(self, timeout=None):
        """Reset all sub components.

        :param int timeout: Not used for this component.
        """
        for cmt in self.volumes_cmts:
            cmt.ensure_ready()


class MockVolumes(components.MockComponentMixin, VolumesInterface, components.BlockingApply):
    """A mock volumes component representing a list of volume data models.

    See the documentation on MockComponentMixin for more details about correct use.

    :param str tag: The UNIQUE tag to use for these volumes.
    :param valence.models.ResourceCluster cluster_dm: The cluster on which these volumes reside.
    :param list[valence.models.Volume] volume_dms: The set of volumes to represent.
    :param valence.vl_components.AccountsInterface accounts_cmt: An accounts component that contains the
        account to which these volumes are attached (typically a MockAccounts component).
    :param int account_index: The index of the specific account to which these volumes are attached in
        the list returned by accounts_cmt.get_account_dms().
    :param float comp_ratio: Compression ratio to target when filling (if applicable).
    :param float dedup_ratio: Deduplication ratio to target when filling (if applicable).
    """

    def __init__(self, tag, cluster_dm, volume_dms, accounts_cmt, account_index=0, comp_ratio=1.,
                 dedup_ratio=1.):
        super(MockVolumes, self).__init__()
        assert account_index < len(accounts_cmt)
        self.tag = tag
        self.cluster_dm = cluster_dm
        self._volume_dms = volume_dms
        self.accounts_cmt = accounts_cmt
        self._account_index = account_index
        self.comp_ratio = comp_ratio
        self.dedup_ratio = dedup_ratio

        self._cluster_fm = functional_models.Cluster(cluster_dm)

    @property
    def child_cmts(self):
        """Get all volume-creation components (i.e. no manage components).

        :rtype: list[valence.vl_components._volume.BaseCreateVolumes]
        """
        return [self]

    @property
    def account_dms(self):
        """Get the cached account data models these volumes use.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return [self.accounts_cmt.account_dms[self._account_index]]

    @property
    def volume_dms(self):
        """Get the cached volume data models.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._volume_dms

    @property
    def partial_volume_dms(self):
        """Get the partial volume data models with which this component was instantiated.

        :rtype: list[valence.models.Volume]
        """
        return self._volume_dms

    @property
    def full_tag(self):
        """Tag used in resource names.

        :returns: A string using the provided unique identifier wit -TAG- added
        :rtype: str
        """
        return self.tag + '-TAG-'

    @property
    def block_capacity(self):
        """Block data capacity of the volumes.

        Size of the block data that will be generated by filling these volumes with the configured ratios
        of data compression and deduplication.

        :rtype: valence.models.DataSize
        """
        block_capacity = 0
        for dm in self._volume_dms:
            block_capacity += vl_helpers.get_volume_block_size(
                cluster_fm=self._cluster_fm,
                volume_size=dm.total_size,
                volume_count=1,
                comp_ratio=self.comp_ratio,
                dedup_ratio=self.dedup_ratio
            )
        return block_capacity

    @property
    def meta_capacity(self):
        """Metadata capacity of the volumes.

        Size of the metadata that will be generated by filling these volumes with the configured ratio
        of data compression.

        :rtype: valence.models.DataSize
        """
        return vl_helpers.get_block_metadata_size(
            cluster_fm=self._cluster_fm,
            block_data_size=self.block_capacity,
            comp_ratio=self.comp_ratio,
            dedup_ratio=self.dedup_ratio
        )

    def get_account_dms(self):
        """Get the accounts related to these volumes through API calls.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Account]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return [self.accounts_cmt.get_account_dms()[self._account_index]]

    def get_volume_dms(self):
        """Get the applied volumes from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied
        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._volume_dms

    def get_allocated_space(self):
        """Get the total amount of space allocated by this component in bytes.

        :rtype: int
        """
        return sum(int(volume.total_size) for volume in self.partial_volume_dms)

    def assert_valid(self):
        """Noop for mock component."""

    def _apply(self, timeout=None):
        """Apply the sub components.

        :param int timeout: the timeout.
        """
        self.accounts_cmt.ensure_applied(timeout=timeout)
