"""Valence volume access group components."""
from __future__ import division

import abc

import autologger
import six

from valence import (
    components,
    functional_models,
    models,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class VolumeAccessGroupInterface(components.Reversible):
    """Interface component for a volume access group.

    This class is used to establish all necessary abstract methods for a set of volume access groups.
    """

    @property
    @abc.abstractmethod
    def volume_access_group_dm(self):
        """Get the cached volume access group data model.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: valence.models.VolumeAccessGroup
        """

    @property
    @abc.abstractmethod
    def client_dms(self):
        """Get the client data models associated with the initiators this volume access group uses.

        :rtype: list[valence.models.Client]
        """

    @property
    @abc.abstractmethod
    def initiator_dms_map(self):
        """Get the cached initiator data models by client this volume access group uses.

        :rtype: dict[str,list[valence.models.Initiator]]
        """

    @property
    @abc.abstractmethod
    def volume_dms(self):
        """Get the cached volume data models this volume access group uses.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: list[valence.models.Volume]
        """

    @property
    @abc.abstractmethod
    def assert_valid(self):
        """Assert the existing initiators match the configuration."""

    @abc.abstractmethod
    def get_volume_access_group_dm(self):
        """Get the applied volume access groups from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: valence.models.VolumeAccessGroup
        """


@six.add_metaclass(abc.ABCMeta)
class BaseCreateVolumeAccessGroup(components.Reversible):
    """Base object for a component that creates and manages a volume access group.

    This defines the common API for all "CreateVolumeAccessGroup" components.
    """

    @abc.abstractmethod
    def _get_existing(self):
        """Get existing volume access group.

        :returns: A volume access group model that exists on cluster with the unique tag in the name.
        :rtype: valence.models.VolumeAccessGroup
        """

    @abc.abstractmethod
    def _assert_valid(self, vag_dm=None):
        """Abstract assert_valid method.

        :param valence.models.VolumeAccessGroup vag_dm: Volume access group data model to check.
        """

    @abc.abstractmethod
    def _get_initiator_dms_map(self):
        """Get the cached initiator data models by client this volume access group uses.

        NOTE: The initiators component must have been applied.

        :rtype: dict[str,list[valence.models.Initiator]]
        """

    @abc.abstractmethod
    def _get_volume_dms(self):
        """Get the cached volume data models this volume access group uses.

        NOTE: The volumes component must have been applied.

        :rtype: list[valence.models.Volume]
        """

    def get_volume_access_group_dm(self):
        """Get the applied volume access groups from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: valence.models.VolumeAccessGroup
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()
        return self._get_existing()

    def assert_valid(self):
        """Assert the existing initiators match the configuration."""
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()
        self._assert_valid()

    def _get_initiator_names(self):
        """Get the initiator names for the volume access group.

        :rtype: list[str]
        """
        initiator_dms_map = self._get_initiator_dms_map()
        initiator_names = []
        for client_initiator_dms in initiator_dms_map.values():
            initiator_names.extend([dm.initiator_name for dm in client_initiator_dms])

        return sorted(initiator_names)

    def _get_volume_ids(self):
        """Get the initiator names for the volume access group.

        :rtype: list[str]
        """
        volume_dms = self._get_volume_dms()
        return sorted([dm.volume_id for dm in volume_dms])


class CreateVolumeAccessGroup(BaseCreateVolumeAccessGroup, components.BlockingApply):
    """Create a Volume Access Group.

    This is a somewhat complex component that gives you complete control at every
    layer. It creates a VAG for the specified volumes, using the initiators
    for **only the clients you specify**. So, even if ``initiators_cmt`` contains
    more initiators than the client you specify, it will only use the initiators
    that are associated with the ``client_dms`` you specify.

    If you do not specify ``client_dms`` it uses all available initiators from
    the ``initiators_cmt``.

    :param str name: the full name of the VAG. This must be unique for all
        existing VAGs.
    :param valence.models.ResourceCluster cluster_dm: the cluster to create the VAG on.
    :param valence.vl_components.InitiatorsInterface initiators_cmt: the initiators to use with this VAG.
    :param valence.vl_components.VolumesInterface volumes_cmt: the volumes to use with this VAG.
    :param list[valence.models.ResourceClient] client_dms: the clients to associate with the VAG.
        If not given, will use all initiators from initiators_cmt.
    """

    def __init__(self, name, cluster_dm, initiators_cmt=None, volumes_cmt=None, client_dms=None):
        super(CreateVolumeAccessGroup, self).__init__()
        self.name = name
        self.cluster_dm = cluster_dm
        self.initiators_cmt = initiators_cmt
        self.volumes_cmt = volumes_cmt
        self._client_dms = client_dms
        self._cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)

        self._volume_access_group = None

        _assert_valid_clients_and_initiator(client_dms, initiators_cmt)

    @staticmethod
    def mock(name, cluster_dm,  # pylint: disable=unused-argument
             initiators_cmt=None, volumes_cmt=None, client_dms=None):
        """Mock the CreateVolumeAccessGroup.__init__() for unit testing.

        Example::

            @mock.patch('valence.vl_components._volume_access_group.CreateVolumeAccessGroup',
                        vl_components.CreateVolumeAccessGroup.mock)
            def test_something()
                ...

        .. docstring-checker: disable=missing-param-check

        :rtype: valence.vl_components._volume_access_group.MockVolumeAccessGroup
        """
        vag_dm = models.VolumeAccessGroup.create(
            name=name,
            volume_access_group_id=0
        )
        return MockVolumeAccessGroup(
            volume_access_group_dm=vag_dm,
            initiators_cmt=initiators_cmt,
            volumes_cmt=volumes_cmt,
            client_dms=client_dms
        )

    @property
    def volume_access_group_dm(self):
        """Get the cached volume access group data model.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: valence.models.VolumeAccessGroup
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._volume_access_group

    @property
    def client_dms(self):
        """Get the client data models associated with the initiators this volume access group uses.

        :rtype: list[valence.models.ResourceClient]
        """
        if self._client_dms:
            return self._client_dms
        elif self.initiators_cmt:
            return self.initiators_cmt.client_dms
        else:
            return []

    @property
    def initiator_dms_map(self):
        """Get the cached initiator data models by client this volume access group uses.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: dict[str,list[valence.models.Initiator]]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        else:
            return self._get_initiator_dms_map()

    @property
    def volume_dms(self):
        """Get the cached volume data models this volume access group uses.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        else:
            return self._get_volume_dms()

    def _apply(self, timeout=None):
        """Create and validate the volume access group.

        :param int timeout: Not used since volume deletion is synchronous.
        """
        if self.initiators_cmt:
            self.initiators_cmt.ensure_applied()
        if self.volumes_cmt:
            self.volumes_cmt.ensure_applied()

        self._cluster_fm.create_volume_access_group(
            name=self.name,
            initiators=self._get_initiator_names(),
            volume_dms=self._get_volume_dms()
        )

        vag_dm = self._get_existing()
        self._assert_valid(vag_dm)
        self._volume_access_group = vag_dm

    def _reset(self, timeout=None):
        """Delete the Volume Access Group from the cluster.

        :param int timeout: Not used since volume access group deletion is synchronous.
        """
        vag_dm = self._get_existing()
        if vag_dm:
            self._cluster_fm.delete_volume_access_group(vag_dm)
        self._volume_access_group = None

    def _reload(self):
        """Attempt to reload the component.

        Checks to see if the volume access groups already exist.

        :returns: Whether the component was able to reload.
        :rtype: bool
        """
        if self.initiators_cmt:
            self.initiators_cmt.ensure_applied()
        if self.volumes_cmt:
            self.volumes_cmt.ensure_applied()

        volume_access_group = self._get_existing()

        if not volume_access_group:
            _log.info('volume access group not found on cluster, continuing with apply')
            return False

        self._assert_valid(volume_access_group)
        self._volume_access_group = volume_access_group
        _log.info('volume access group with the specified name({}) was reloaded'.format(self.name))
        return True

    def _get_existing(self):
        """Get existing volume access group.

        :raises valence.components.ComponentStateError: If more than one matching volume access group is
            found.
        :returns: A volume access group that exists on the cluster with the unique tag in the name.
        :rtype: valence.models.VolumeAccessGroup
        """
        volume_access_groups = self._cluster_fm.list_volume_access_groups()
        matching_volume_access_groups = [v for v in volume_access_groups if v.name == self.name]
        if len(matching_volume_access_groups) == 0:
            return None
        elif len(matching_volume_access_groups) == 1:
            return matching_volume_access_groups[0]
        else:
            raise components.ComponentStateError(
                'More than one volume access group found with the specified name({}): {}'
                .format(self.name, volume_access_groups)
            )

    def _assert_valid(self, volume_access_group_dm=None):
        """Assert the specified volume access group is valid.

        :param valence.models.VolumeAccessGroup volume_access_group_dm: Volume access group to validate
            (current volume access group on cluster by default).
        :raises valence.components.ComponentStateError: If no matching volume access group configuration
            is found.
        :raises valence.components.PartialStateError: If an invalid volume access group configuration
            is found.
        """
        if volume_access_group_dm is None:
            volume_access_group_dm = self._get_existing()

        if volume_access_group_dm is None:
            raise components.ComponentStateError(
                'No volume access group found with the specified name({})'.format(self.name)
            )

        err = []

        # Note: you might think that using the ids is better.
        # However, the API call to create initiators uses names, so we use names here.
        initiator_names = set(self._get_initiator_names())
        difference = initiator_names.symmetric_difference(volume_access_group_dm.initiators)
        if difference:
            err.append('Invalid initiator names: {}'.format(difference))

        volume_ids = set(self._get_volume_ids())
        difference = volume_ids.symmetric_difference(volume_access_group_dm.volumes)
        if difference:
            err.append('Invalid volumes: {}'.format(difference))

        if err:
            raise components.PartialStateError(
                'Invalid volume access group configuration found:\n* {}'.format('\n* '.join(err))
            )

    def _get_initiator_dms_map(self):
        """Get the cached initiator data models by client this volume access group uses.

        NOTE: The initiators component must be applied.

        :rtype: dict[str,list[valence.models.Initiator]]
        """
        return _get_initiator_map_from_initiators_cmt(
            self._client_dms, self.initiators_cmt)

    def _get_volume_dms(self):
        """Get the cached volume data models this volume access group uses.

        NOTE: The volumes component must have been applied.

        :rtype: list[valence.models.Volume]
        """
        if self.volumes_cmt is None:
            return []
        else:
            return self.volumes_cmt.volume_dms


class MockVolumeAccessGroup(components.MockComponentMixin, VolumeAccessGroupInterface,
                            components.BlockingApply):
    """A mock volume access group component representing a volume access group data model.

    See the documentation on MockComponentMixin for more details about correct usage.

    :param valence.models.VolumeAccessGroup volume_access_group_dm: The volume access group this
        component emulates.
    :param valence.vl_components.InitiatorsInterface initiators_cmt: The initiators to use with this VAG.
    :param valence.vl_components.VolumesInterface volumes_cmt: The volumes to use with this VAG.
    :param list[valence.models.ResourceClient] client_dms: The clients to associate with the VAG.
        If not given, will use all initiators from initiators_cmt.
    """
    def __init__(self, volume_access_group_dm, initiators_cmt=None, volumes_cmt=None, client_dms=None):
        super(MockVolumeAccessGroup, self).__init__()
        self._volume_access_group_dm = volume_access_group_dm
        self.initiators_cmt = initiators_cmt
        self.volumes_cmt = volumes_cmt
        self._client_dms = client_dms

        _assert_valid_clients_and_initiator(client_dms, initiators_cmt)

    @property
    def volume_access_group_dm(self):
        """Get the cached volume access group data model.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: valence.models.VolumeAccessGroup
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        return self._volume_access_group_dm

    @property
    def volume_dms(self):
        """Get the cached volume data models this volume access group uses.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        elif self.volumes_cmt is None:
            return []
        else:
            return self.volumes_cmt.volume_dms

    @property
    def initiator_dms_map(self):
        """Get the cached initiator data models by client this volume access group uses.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: dict[str,list[valence.models.Initiator]]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError()
        else:
            return _get_initiator_map_from_initiators_cmt(self._client_dms, self.initiators_cmt)

    @property
    def client_dms(self):
        """Get the client data models associated with the initiators this volume access group uses.

        :rtype: list[valence.models.ResourceClient]
        """
        if self._client_dms:
            return self._client_dms
        elif self.initiators_cmt:
            return self.initiators_cmt.client_dms
        else:
            return []

    def get_volume_access_group_dm(self):
        """Get the applied volume access groups from the cluster.

        :raises valence.components.ComponentStateError: If component has not been applied.
        :rtype: valence.models.VolumeAccessGroup
        """
        if self.state is not components.States.applied:  # pylint: disable=no-member
            raise components.ComponentStateError()
        return self._volume_access_group_dm

    def assert_valid(self):
        """Noop for mock component."""

    def _apply(self, timeout=None):
        """Apply all sub components.

        :param int timeout: N/A.
        """
        if self.initiators_cmt:
            self.initiators_cmt.ensure_applied()
        if self.volumes_cmt:
            self.volumes_cmt.ensure_applied()


def _assert_valid_clients_and_initiator(client_dms, initiators_cmt):
    """Helper method for asserting that parameters were correct in init.

    :param list[valence.models.ResourceClient] client_dms: the clients to associate with the VAG.
        If not given, will use all initiators from initiators_cmt.
    :param valence.vl_components.InitiatorsInterface initiators_cmt: the initiators to use with this VAG.
    """

    if client_dms and not initiators_cmt:
        raise ValueError('Cannot specify client_dms without an initiators_cmt')
    elif client_dms:
        client_mips = [dm.mip for dm in client_dms]
        initiator_client_mips = [dm.mip for dm in initiators_cmt.client_dms]
        invalid = set(client_mips).difference(initiator_client_mips)
        if invalid:
            raise ValueError(
                'client_dms contains clients that are not in the initiators: {}'
                .format(invalid)
            )


def _get_initiator_map_from_initiators_cmt(client_dms, initiators_cmt):
    """Get the cached initiator data models by client.

    NOTE: The initiators component must be applied.

    :param list[valence.models.ResourceClient] client_dms: the clients to associate with the VAG.
        If not given, will use all initiators from initiators_cmt.
    :param valence.vl_components.InitiatorsInterface initiators_cmt: the initiators to use with this VAG.
    :rtype: dict[str,list[valence.models.Initiator]]
    """
    if initiators_cmt is None:
        return {}
    elif client_dms is None:
        return initiators_cmt.initiator_dms_map
    else:
        client_mips = [dm.mip for dm in client_dms]
        initiator_dms_map = {
            mip: inits for mip, inits in initiators_cmt.initiator_dms_map.items()
            if mip in client_mips
        }
        return initiator_dms_map
