"""Valence IdP Client Management components."""

from __future__ import division

import abc

import autologger
import six

from valence import (
    components,
    enums,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseIdpDeployInterface(components.Reversible, components.BlockingApply):
    """Base component for managing SSL Certificates for an Identity Provider (IdP).

    :param sf_models.ResourceClient client_dm: client data model
    """

    IDP_TYPE = None  # type: str

    def __init__(self, client_dm):
        super(BaseIdpDeployInterface, self).__init__()

        self.client_dm = client_dm  # type: sf_models.ResourceClient


# TODO ELEM-7193: Refined the param definition for idp_user_dm when that data model exists.
@six.add_metaclass(abc.ABCMeta)
class BaseIdpUserManagementComponent(components.Reversible, components.BlockingApply):
    """Base component for managing IdP users on an Identity Provider (IdP).

    :param sf_models.ResourceClient client_dm: client data model
    :param any idp_user_dm: data model describing the IdP users to apply
    """

    IDP_TYPE = None  # type: str

    def __init__(self, client_dm, idp_user_dm):
        super(BaseIdpUserManagementComponent, self).__init__()

        self.client_dm = client_dm  # type: sf_models.ResourceClient
        self.idp_user_dm = idp_user_dm


class DockerizedShibbolethIdpDeployComponent(BaseIdpDeployInterface):
    """Component to manage SSL Certificate on a docker-based Shibboleth IdP.

    :param sf_models.ResourceClient client_dm: client data model
    :param str docker_compose_source: Docker compose source to use
    :param str shibboleth_version: Version of shibboleth to deploy
    """

    IDP_TYPE = enums.IdpType.shibboleth
    COMPOSE_SRC_DEFAULT = 'https://github.com/UniconLabs/dockerized-idp-testbed.git'
    SHIBBOLETH_VERSION_DEFAULT = '3.4.3'

    def __init__(self, client_dm,
                 docker_compose_source=COMPOSE_SRC_DEFAULT,
                 shibboleth_version=SHIBBOLETH_VERSION_DEFAULT):
        super(DockerizedShibbolethIdpDeployComponent, self).__init__(client_dm)

        self.docker_compose_source = docker_compose_source
        self.shibboleth_version = shibboleth_version

    def _apply(self, timeout=None):
        """Configure an SSL Cert and deploy the shibboleth container.

        :param int timeout: Not Applicable
        """

        # TODO ELEM-7227: Implement apply logic

    def _reset(self, timeout=None):
        """Deconfigure the Shibboleth IDP Container; remove any materials.

        :param int timeout: Not Applicable
        """

        # TODO ELEM-7227: Implement reset logic


class AdfsIdpDeployComponent(BaseIdpDeployInterface):
    """Component for bringing up an ADFS Identity Provider for MFA testing.

    This component may not be applicable for static client testing...
    """

    IDP_TYPE = enums.IdpType.adfs

    def _apply(self, timeout=None):
        """Configure an SSL Cert and deploy the shibboleth container.

        :param int timeout: Not Applicable
        """

        # TODO ELEM-7203: Determine if this is necesary, and implement if so.

    def _reset(self, timeout=None):
        """Deconfigure the Shibboleth IDP Container; remove any materials.

        :param int timeout: Not Applicable
        """

        # TODO ELEM-7203: Determine if this is necesary, and implement if so.


class ShibbolethUserManagementComponent(BaseIdpUserManagementComponent):
    """Component to manage SSL Certificate on a docker-based Shibboleth IdP."""

    IDP_TYPE = enums.IdpType.shibboleth

    def _apply(self, timeout=None):
        """Add users to the IDP based on the specified idp user data models.

        :param int timeout: Not Applicable
        """

    def _reset(self, timeout=None):
        """Remove users from the IDP based on the specified idp user data models.

        :param int timeout: Not Applicable
        """


class AdfsUserManagementComponent(BaseIdpUserManagementComponent):
    """Component to manage SSL Certificate on a staticly managed ADFS IdP."""

    IDP_TYPE = enums.IdpType.adfs

    def _apply(self, timeout=None):
        """Add users to the IDP based on the specified idp user data models.

        :param int timeout: Not Applicable
        """

    def _reset(self, timeout=None):
        """Remove users from the IDP based on the specified idp user data models.

        :param int timeout: Not Applicable
        """
