from __future__ import division

import unittest

import mock
import pytest

from valence import (
    components,
    functional_models,
    models,
    vl_helpers,
)
from valence.vl_components import _account


class TestCreateAccountsComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.one_acct_cmt = _account.CreateAccounts('one', self.cluster_dm)

        self.three_acct_cmt = _account.CreateAccounts('three', self.cluster_dm, num=3)
        self.three_names = tuple(self.three_acct_cmt.full_tag + str(n)
                                 for n in range(0, self.three_acct_cmt.num))
        self.three_accts = tuple(models.Account({'accountID': i,
                                                'username': self.three_names[i],
                                                 'status': 'active'})
                                 for i in range(self.three_acct_cmt.num))
        self.one_names = tuple(self.one_acct_cmt.full_tag + str(n)
                               for n in range(0, self.one_acct_cmt.num))
        self.one_accts = tuple(models.Account({'account_id': i,
                                              'username': self.one_names[i],
                                               'status': 'active'})
                               for i in range(self.one_acct_cmt.num))

    @mock.patch.object(vl_helpers, 'create_accounts')
    def test__create(self, mock_create_accounts):
        mock_create_accounts.return_value = self.three_accts
        created = self.three_acct_cmt._create(self.three_acct_cmt.num, 0)
        assert created == self.three_accts

        mock_create_accounts.return_value = self.one_accts
        created = self.one_acct_cmt._create(self.one_acct_cmt.num, 0)
        assert created == self.one_accts

    @mock.patch.object(vl_helpers, 'delete_accounts')
    def test__delete(self, mock_delete_accounts):
        self.three_acct_cmt._delete(self.three_accts)
        assert mock_delete_accounts.called
        self.one_acct_cmt._delete(self.one_accts)
        assert mock_delete_accounts.called

    @mock.patch.object(_account.CreateAccounts, '_get_existing')
    @mock.patch.object(_account.CreateAccounts, '_apply')
    @mock.patch.object(_account.CreateAccounts, '_reload')
    def test_account_dms(self, mock_reload, mock_apply, get_existing):
        get_existing.return_value = self.three_accts
        with pytest.raises(components.ComponentStateError):
            accounts = self.three_acct_cmt.account_dms
        self.three_acct_cmt.apply()
        self.three_acct_cmt._accounts = tuple(self.three_accts)
        assert self.three_acct_cmt.account_dms == tuple(self.three_accts)

    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test__list_accounts(self, mock_list_accounts):
        mock_list_accounts.return_value = self.three_accts + self.one_accts
        list_accts = self.three_acct_cmt._list_accounts()
        assert list_accts == tuple(self.three_accts + self.one_accts)

    @mock.patch.object(_account.CreateAccounts, '_list_accounts')
    def test__get_existing(self, mock_internal):
        mock_internal.return_value = self.one_accts + self.three_accts
        existing = self.three_acct_cmt._get_existing()
        assert existing == self.three_accts
        existing = self.one_acct_cmt._get_existing()
        assert existing == self.one_accts

    def test__assert_valid(self):
        with pytest.raises(components.PartialStateError):
            self.three_acct_cmt._assert_valid(self.one_accts)
        self.three_acct_cmt._assert_valid(self.three_accts)
        with pytest.raises(components.PartialStateError):
            self.one_acct_cmt._assert_valid(self.three_accts)
        self.one_acct_cmt._assert_valid(self.one_accts)

    def test_resource_tag(self):
        assert self.three_acct_cmt.full_tag == 'three-TAG-'
        assert self.one_acct_cmt.full_tag == 'one-TAG-'

    @mock.patch.object(_account.CreateAccounts, '_get_existing')
    @mock.patch.object(_account.CreateAccounts, '_delete')
    def test__reset(self, mock_del, mock_ge):
        self.three_acct_cmt._accounts = self.three_accts
        self.three_acct_cmt._reset()
        assert self.three_acct_cmt._accounts is None
        self.one_acct_cmt._accounts = self.one_accts
        self.one_acct_cmt._reset()
        assert self.one_acct_cmt._accounts is None

    @mock.patch.object(_account.CreateAccounts, '_create')
    @mock.patch.object(_account.CreateAccounts, '_assert_valid')
    def test__apply(self, mock_valid, mock_create):
        mock_create.return_value = tuple(self.three_accts)
        self.three_acct_cmt._apply()
        assert self.three_acct_cmt._accounts == tuple(self.three_accts)
        mock_create.return_value = tuple(self.one_accts)
        self.one_acct_cmt._apply()
        assert self.one_acct_cmt._accounts == tuple(self.one_accts)

    @mock.patch.object(_account.CreateAccounts, '_get_existing')
    def test__reload(self, mock_ge):
        mock_ge.return_value = None
        reloaded = self.three_acct_cmt._reload()
        assert reloaded is False
        assert self.three_acct_cmt._accounts is None
        mock_ge.return_value = tuple(self.three_accts)
        reloaded = self.three_acct_cmt._reload()
        assert reloaded is True
        assert self.three_acct_cmt._accounts == tuple(self.three_accts)

    @mock.patch.object(_account.CreateAccounts, '_get_existing')
    def test__reload_error(self, mock_ge):
        mock_ge.return_value = tuple(self.three_accts + self.one_accts)
        with pytest.raises(components.PartialStateError):
            self.three_acct_cmt._reload()
        assert self.three_acct_cmt._accounts is None

    @mock.patch.object(_account.CreateAccounts, '_get_existing')
    @mock.patch.object(_account.CreateAccounts, '_apply')
    @mock.patch.object(_account.CreateAccounts, '_reload')
    def test_get_accounts_and_ids(self, mock_reload, mock_apply, get_existing):
        get_existing.return_value = self.three_accts
        with pytest.raises(components.ComponentStateError):
            self.three_acct_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            self.three_acct_cmt.get_account_ids()
        self.three_acct_cmt.apply()
        self.three_acct_cmt._accounts = tuple(self.three_accts)
        assert self.three_acct_cmt.get_account_dms() == tuple(self.three_accts)
        assert self.three_acct_cmt.get_account_ids() == [a.account_id for a in self.three_accts]

    def test_less_than_one(self):
        with pytest.raises(ValueError):
            _account.CreateAccounts('fail', self.cluster_dm, 0)

    def test_length(self):
        assert len(self.three_acct_cmt) == 3


class TestMockAccounts(unittest.TestCase):
    def setUp(self):
        self.account_1 = models.Account.create(
            username='user1', status='active', account_id=1)
        self.account_2 = models.Account.create(
            username='user2', status='active', account_id=2)

    def test_initialize(self):
        accounts_cmt = _account.MockAccounts('mock', [self.account_1])

    def test_full_tag(self):
        accounts_cmt = _account.MockAccounts('mock', [self.account_1])
        assert accounts_cmt.full_tag == 'mock-TAG-'

    def test_apply(self):
        accounts_cmt = _account.MockAccounts('mock', [self.account_1])
        accounts_cmt.apply()
        assert accounts_cmt.state is components.States.applied

    def test_reset(self):
        accounts_cmt = _account.MockAccounts('mock', [self.account_1])
        accounts_cmt.ensure_applied()
        assert accounts_cmt.state is components.States.applied

        accounts_cmt.reset()
        assert accounts_cmt.state is components.States.ready
        accounts_cmt.ensure_ready()
        assert accounts_cmt.state is components.States.ready

    def test_get(self):
        account_dms = [self.account_1]
        accounts_cmt = _account.MockAccounts('mock', account_dms)
        with self.assertRaises(components.ComponentStateError):
            accounts_cmt.account_dms

        with self.assertRaises(components.ComponentStateError):
            accounts_cmt.get_account_dms()

        accounts_cmt.apply()
        assert accounts_cmt.account_dms == account_dms

    def test_multi(self):
        account_dms = [
            self.account_1,
            self.account_2,
        ]
        accounts_cmt = _account.MockAccounts('mock', account_dms)
        accounts_cmt.ensure_applied()
        assert accounts_cmt.account_dms == account_dms

    def test_get_account_ids(self):
        accounts_cmt = _account.MockAccounts('mock', [self.account_1])
        with pytest.raises(components.ComponentStateError):
            account_ids = accounts_cmt.get_account_ids()
        accounts_cmt.apply()
        assert accounts_cmt.get_account_ids() == [self.account_1.account_id]
