"""Test valence.vl_components._capacity."""
import unittest

import mock
import pytest

from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    vl_components,
    vl_helpers,
)
from valence.vl_components import _capacity


class TestUnbalanceClusterComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dms = [models.ActiveNode.get_mock_object() for _ in range(4)]
        cls.node_fms = [functional_models.Node(dm, cls.cluster_dm) for dm in cls.node_dms]

        last_drive_id = 0
        cls.node_drive_dms_list = []
        for _ in cls.node_dms[:-1]:
            node_drive_dms = [models.Drive.create(
                drive_id=last_drive_id + i + 1,
                capacity=100
            ) for i in range(10)]
            last_drive_id = node_drive_dms[-1].drive_id
            cls.node_drive_dms_list.append(node_drive_dms)

        cls.largest_node_drive_dms = [
            models.Drive.create(drive_id=last_drive_id + i + 1, capacity=110) for i in range(10)
        ]
        cls.node_drive_dms_list.append(cls.largest_node_drive_dms)

    def setUp(self):
        self.unbalance_cluster_cmt = _capacity.UnbalanceCluster(self.cluster_dm)

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(vl_helpers, 'list_nodes_by_usable_block_capacity')
    @mock.patch('valence.vl_components._drive.RemoveDrives', vl_components.RemoveDrives.mock)
    def test_apply(self, sort_nodes_by_size, get_node_fms, list_active_nodes, list_drives):
        sort_nodes_by_size.return_value = self.node_fms
        get_node_fms.return_value = self.node_fms
        list_active_nodes.return_value = self.node_dms
        list_drives.side_effect = self.node_drive_dms_list
        self.unbalance_cluster_cmt.apply(timeout=42)
        assert self.unbalance_cluster_cmt._remove_drives_cmt.state == components.States.applied

        # Ensure we're removing drives from each node EXCEPT the largest
        removed_drive_dms = self.unbalance_cluster_cmt._remove_drives_cmt.drive_dms
        removed_drive_ids = set(dm.drive_id for dm in removed_drive_dms)
        assert not removed_drive_ids.intersection(dm.drive_id for dm in self.largest_node_drive_dms)

        # Ensure we're removing just the required number of drives to make the capacity of the
        # largest node equal to half the total cluster capacity
        # Total Capacity: 4100
        # Largest Node: 1100
        # 4100 - 2 * 1100 = 1900
        # 1900 / drive_size(100) = 19
        assert len(removed_drive_dms) == 19

        self.unbalance_cluster_cmt.reset()
        assert self.unbalance_cluster_cmt._remove_drives_cmt.state == components.States.ready

    @mock.patch.object(vl_helpers, 'list_nodes_by_usable_block_capacity')
    def test_small_clusters(self, sort_nodes_by_size):
        # 1-node cluster
        sort_nodes_by_size.return_value = self.node_fms[:1]
        with pytest.raises(RuntimeError):
            self.unbalance_cluster_cmt.apply()
        self.unbalance_cluster_cmt.reset()

        # 2-node cluster
        sort_nodes_by_size.return_value = self.node_fms[:2]
        self.unbalance_cluster_cmt.apply()
        assert self.unbalance_cluster_cmt._remove_drives_cmt is None
        self.unbalance_cluster_cmt.reset()

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(vl_helpers, 'list_nodes_by_usable_block_capacity')
    def test_witness_node(self, sort_nodes_by_size, get_node_fms, list_active_nodes, list_drives):
        sort_nodes_by_size.return_value = self.node_fms[:3]
        get_node_fms.return_value = self.node_fms[:3]
        list_active_nodes.return_value = self.node_dms[:3]

        # No drives on the witness node
        list_drives_response = [x for x in self.node_drive_dms_list[:3]]
        list_drives_response[0] = []
        list_drives.side_effect = list_drives_response

        self.unbalance_cluster_cmt.apply()
        assert self.unbalance_cluster_cmt._remove_drives_cmt is None
        self.unbalance_cluster_cmt.reset()


class TestStrandCapacityComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dms = [models.ActiveNode.get_mock_object() for _ in range(4)]
        cls.node_fms = [functional_models.Node(dm, cls.cluster_dm) for dm in cls.node_dms]

        cls.drive_info_dms = [models.DriveInfo.create(
            drive_type=enums.DriveType.block,
            size=100
        ) for _ in range(10)]
        cls.drive_info_dms.extend([models.DriveInfo.create(
            drive_type=enums.DriveType.slice
        ) for _ in range(2)])

        cls.largest_node_drive_info_dms = [models.DriveInfo.create(
            drive_type=enums.DriveType.block,
            size=110
        ) for _ in range(10)]
        cls.largest_node_drive_info_dms.extend([models.DriveInfo.create(
            drive_type=enums.DriveType.slice
        ) for _ in range(2)])

        cls.drive_configs = [models.DriveConfig.create(drives=cls.drive_info_dms) for _ in range(3)]
        cls.drive_configs.append(models.DriveConfig.create(drives=cls.largest_node_drive_info_dms))

        cls.unbalance_cluster_cmt = _capacity.UnbalanceCluster(cls.cluster_dm)
        cls.unbalance_cluster_cmt.sorted_node_fms = cls.node_fms
        cls.unbalance_cluster_cmt._apply_setup = lambda: None
        cls.unbalance_cluster_cmt._apply = lambda _: None

    def setUp(self):
        self.strand_capacity_cmt = _capacity.StrandCapacity(self.unbalance_cluster_cmt)

    @mock.patch('valence.functional_models.Cluster.list_cluster_faults')
    @mock.patch.object(vl_helpers, 'set_fake_drives_flag_on_node')
    @mock.patch.object(functional_models.Cluster, 'list_all_nodes')
    @mock.patch('valence.vl_components._capacity.UnbalanceCluster')
    @mock.patch('valence.vl_components._drive.FailDrives', vl_components.FailDrives.mock)
    def test_apply(self, unbalance_cluster, list_all_nodes, set_fake_drives_flag_on_node,
                   list_cluster_faults):
        unbalance_cluster.return_value = self.unbalance_cluster_cmt
        list_all_nodes.return_value = self.node_dms
        list_cluster_faults.return_value = [
            models.ClusterFault.create(code=enums.ClusterFaultCode.stranded_capacity)
        ]
        self.strand_capacity_cmt.apply()
        assert set_fake_drives_flag_on_node.called
        assert self.strand_capacity_cmt._fail_drives_cmt.state == components.States.applied

        self.strand_capacity_cmt.reset()
        assert self.strand_capacity_cmt._fail_drives_cmt.state == components.States.ready
