"""Test valence.vl_components._cluster."""
import unittest

import mock
import pytest

import sf_api
from sf_api import (
    result_models,
    sf_version,
)
from valence import (
    components,
    enums,
    errors,
    functional_models,
    models,
    selectors,
    unittest_utils,
    vl_components,
    vl_helpers,
)
from valence.vl_components import (
    _cluster,
    _constants,
)

RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()


class TestRemoveNodesComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.active_nodes_dms = models.ActiveNode.get_mock_object_list(3, overrides={'status': 'active'})
        cls.pending_nodes_dms = models.PendingNode.get_mock_object_list(3)
        cls.selector = lambda x: cls.active_nodes_dms

    def setUp(self):
        self.remove_node_cmt = _cluster.RemoveNodes(self.cluster_dm, self.selector)

    @mock.patch('valence.vl_helpers.remove_nodes')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.vl_helpers.remove_drives')
    @mock.patch.object(_cluster.RemoveNodes, '_reload')
    @mock.patch('valence.vl_helpers.get_bin_sync_timeout_for_nodes')
    def test_apply(self, mock_bs_timeout, mock_reload,
                   mock_remove_drives, mock_list_drives, mock_remove_nodes):
        mock_reload.return_value = False
        self.remove_node_cmt.apply()
        mock_remove_drives.assert_called()
        mock_remove_nodes.assert_called()
        assert self.remove_node_cmt.is_running() is False

    def test_apply_setup_error(self):
        remove_node_cmt = _cluster.RemoveNodes(self.cluster_dm, lambda: [])
        with pytest.raises(components.ComponentStateError):
            remove_node_cmt.apply()

    @mock.patch.object(_cluster.RemoveNodes, '_apply')
    @mock.patch.object(_cluster.RemoveNodes, '_reload')
    def test_rs_node_dms(self, mock_reload, mock_apply):
        with pytest.raises(components.ComponentStateError):
            _ = self.remove_node_cmt.pending_node_dms
        self.remove_node_cmt.apply()
        self.remove_node_cmt._pending_nodes = self.pending_nodes_dms

        assert self.remove_node_cmt.pending_node_dms == self.pending_nodes_dms

    @mock.patch('valence.functional_models.Node.poll_verify_pernode_api')
    @mock.patch('valence.vl_helpers.add_nodes')
    @mock.patch('valence.vl_helpers.add_drives')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.list_pending_nodes')
    @mock.patch('valence.functional_models.Cluster.list_active_nodes')
    @mock.patch('valence.functional_models.Cluster.upgrading')
    @mock.patch.object(_cluster.RemoveNodes, '_reload')
    @mock.patch.object(_cluster.RemoveNodes, '_rtfi_nodes_to_cm_version_during_upgrade')
    @mock.patch('valence.vl_helpers.get_bin_sync_timeout_for_nodes')
    def test_reset(self, mock_bs_timeout, mock_reload, mock_confirm_node_version, mock_upgrading,
                   mock_list_active_nodes, mock_list_pending_nodes, mock_list_drives,
                   mock_add_drives, mock_add_nodes, mock_verify_pernode_api):
        mock_reload.return_value = False
        mock_list_pending_nodes.return_value = self.pending_nodes_dms
        mock_list_active_nodes.return_value = self.active_nodes_dms
        mock_upgrading.return_value = True
        self.remove_node_cmt._apply = mock.Mock()
        self.remove_node_cmt.apply()
        for active, pending in zip(self.active_nodes_dms, self.pending_nodes_dms):
            active.mip = pending.mip
        self.remove_node_cmt.reset()
        mock_add_nodes.assert_called()
        mock_add_drives.assert_called()

    @mock.patch('valence.functional_models.Cluster.get_cluster_version_info')
    def test_confirm_node_version_no_rtfi(self, mock_cluster_version_info):
        mock_cluster_version_info.return_value = models.ClusterVersionInfo.get_mock_object(
            overrides={'cluster_version': models.SfVersion('10.0.0.284')}
        )
        for n in self.pending_nodes_dms:
            n.software_version = models.SfVersion('10.0.0.284')
        self.remove_node_cmt._rtfi_nodes_to_cm_version_during_upgrade(self.pending_nodes_dms)

    @mock.patch('valence.functional_models.Node.poll_verify_pernode_api')
    @mock.patch('valence.functional_models.Cluster.get_cluster_version_info')
    @mock.patch('valence.vl_helpers.provision_nodes')
    @mock.patch('valence.functional_models.Node.get_network_config')
    @mock.patch('valence.functional_models.Node.set_network_config')
    def test_confirm_node_version_rtfi(self, mock_set_network_config, mock_get_network_config,
                                       mock_provision_nodes, mock_cluster_version_info, mock_verify):
        mock_cluster_version_info.return_value = models.ClusterVersionInfo.get_mock_object(
            overrides={'cluster_version': models.SfVersion('10.1.0.666')})
        pending_nodes_dm = models.PendingNode.get_mock_object(
            overrides={'software_version': models.SfVersion('10.0.0.284')})
        mock_get_network_config.return_value = models.NetworkConfig.get_mock_object()
        self.remove_node_cmt._rtfi_nodes_to_cm_version_during_upgrade([pending_nodes_dm])

    @mock.patch('valence.functional_models.Cluster.list_pending_nodes')
    def test_reload_success(self, mock_get_resource_node):
        mock_get_resource_node.return_value = self.active_nodes_dms
        self.remove_node_cmt._apply_setup()
        assert self.remove_node_cmt._reload() is True

    @mock.patch('valence.functional_models.Cluster.list_pending_nodes')
    def test_reload_fail(self, mock_get_resource_node):
        mock_get_resource_node.return_value = models.ActiveNode.get_mock_object_list(3)
        self.remove_node_cmt._apply_setup()
        assert self.remove_node_cmt._reload() is False

    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.list_pending_nodes')
    @mock.patch.object(_cluster.RemoveNodes, '_reload')
    @mock.patch('valence.vl_helpers.get_bin_sync_timeout_for_nodes')
    def test_reset_error(self, mock_bs_timeout, mock_reload, mock_list_pending_nodes, mock_list_drives):
        mock_reload.return_value = False
        mock_list_pending_nodes.return_value = models.PendingNode.get_mock_object_list(3)
        self.remove_node_cmt._apply = mock.Mock()
        self.remove_node_cmt.apply()
        with pytest.raises(components.PartialStateError):
            self.remove_node_cmt.reset()


class TestRemoveAndOrphanNodesComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.active_nodes_dms = models.ActiveNode.get_mock_object_list(3, overrides={'status': 'active'})
        cls.selector = lambda x: cls.active_nodes_dms
        cls.rs_nodes_dms = models.ResourceNode.get_mock_object_list(3)

    def setUp(self):
        self.remove_and_orphan_cmt = _cluster.RemoveAndOrphanNodes(self.cluster_dm, self.selector)

    @mock.patch.object(_cluster.RemoveNodes, '_apply')
    @mock.patch.object(_cluster.RemoveNodes, '_reload')
    @mock.patch('valence.vl_helpers.orphan_pending_nodes')
    def test_apply(self, mock_orphan_pending_nodes, mock_reload, mock_apply):
        self.remove_and_orphan_cmt.apply()
        mock_orphan_pending_nodes.assert_called()
        assert self.remove_and_orphan_cmt.is_running() is False

    @mock.patch.object(_cluster.RemoveNodes, '_apply')
    @mock.patch.object(_cluster.RemoveNodes, '_reload')
    @mock.patch('valence.vl_helpers.orphan_pending_nodes')
    def test_rs_node_dms(self, mock_orphan_pending_nodes, mock_reload, mock_apply):
        with pytest.raises(components.ComponentStateError):
            _ = self.remove_and_orphan_cmt.rs_node_dms
        self.remove_and_orphan_cmt.apply()
        self.remove_and_orphan_cmt._rs_nodes = self.rs_nodes_dms

        assert self.remove_and_orphan_cmt.rs_node_dms == self.rs_nodes_dms

    @mock.patch.object(_cluster.RemoveNodes, 'reset')
    @mock.patch.object(_cluster.RemoveNodes, '_apply')
    @mock.patch.object(_cluster.RemoveNodes, '_reload', return_value=False)
    @mock.patch('valence.vl_helpers.associate_orphaned_nodes')
    @mock.patch('valence.vl_helpers.orphan_pending_nodes')
    def test_reset(self, mock_orphan_pending_nodes, mock_associate_orphaned_nodes,
                   mock_reload, mock_apply, mock_reset):
        self.remove_and_orphan_cmt.apply()
        self.remove_and_orphan_cmt.reset()
        mock_associate_orphaned_nodes.assert_called()


class TestRestartMasterServiceComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dm = models.ResourceNode.get_mock_object()
        cls.restart_master_service_cmt = _cluster._RestartMasterService(cluster_dm=cls.cluster_dm,
                                                                        node_dm=cls.node_dm)

    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    def test_apply(self, mock_kill_service):
        mock_kill_service.return_value = ''
        self.restart_master_service_cmt.apply()
        mock_kill_service.assert_called_with(signal=enums.UnixSignal.SIGKILL)

    @mock.patch.object(functional_models.Node, 'stop_binary')
    @mock.patch.object(functional_models.Node, 'start_binary')
    def test_reset(self, mock_start_binary, mock_stop_binary):
        self.restart_master_service_cmt._state = components.States.failed
        self.restart_master_service_cmt.reset()
        mock_stop_binary.assert_called_with(enums.NodeBinary.solidfire)
        mock_start_binary.assert_called_with(enums.NodeBinary.solidfire)


class TestRestartDriveServiceComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dm = models.ResourceNode.get_mock_object()
        drive_dm = models.ClusterServiceDrive.get_mock_object()
        cls.service_dm = models.ClusterServiceInfo.get_mock_object(overrides={'drives': [drive_dm]})
        cls.restart_drive_service_cmt = _cluster._RestartDriveService(cluster_dm=cls.cluster_dm,
                                                                      node_dm=cls.node_dm,
                                                                      service_dm=cls.service_dm)

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    def test_apply(self, mock_kill_service):
        mock_kill_service.return_value = ''
        self.restart_drive_service_cmt.apply()
        mock_kill_service.assert_called_with(service_dm=self.service_dm,
                                             signal=enums.UnixSignal.SIGKILL)

    @mock.patch.object(vl_helpers, 'wait_for_drive_status')
    @mock.patch.object(vl_helpers, 'remove_drives')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers, 'add_drives')
    def test_reset(self, mock_add_drives, mock_list_drives,
                   mock_remove_drives, mock_wait_drive_status):
        self.restart_drive_service_cmt._state = components.States.failed
        drive_dms = self.service_dm.drives
        mock_list_drives.return_value = drive_dms
        self.restart_drive_service_cmt.reset()
        mock_add_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms)


class TestRestartNodeServiceComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

        cluster_service_master = models.ClusterService.get_mock_object(overrides={
            'service_type': enums.ServiceType.master})
        cluster_service_block = models.ClusterService.get_mock_object(overrides={
            'service_type': enums.ServiceType.block})
        service_objects = []
        service_objects.append(models.ClusterServiceInfo.get_mock_object(overrides={
            'service': cluster_service_master}))
        service_objects.append(models.ClusterServiceInfo.get_mock_object(overrides={
            'service': cluster_service_block}))
        cls.node_dms = models.ResourceNode.get_mock_object_list(2)
        cls.service_selector = lambda x: service_objects

    def setUp(self):
        self.restart_node_service_cmt = _cluster.RestartNodeServices(
            self.cluster_dm, self.service_selector)

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    @mock.patch.object(_cluster._RestartMasterService, '_apply')
    @mock.patch.object(_cluster._RestartDriveService, '_apply')
    def test_apply(self, mock_restart_drive_apply, mock_restart_master_apply,
                   mock_get_resource_node, mock_kill_master_service_ssh,
                   mock_kill_drive_service_ssh):
        mock_get_resource_node.side_effect = [self.node_dms[0], self.node_dms[1]]
        mock_kill_master_service_ssh.return_value = ''
        mock_kill_drive_service_ssh.return_value = ''
        self.restart_node_service_cmt.apply()
        mock_restart_master_apply.assert_called()
        mock_restart_drive_apply.assert_called()

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    @mock.patch.object(_cluster._RestartMasterService, 'reset')
    @mock.patch.object(_cluster._RestartDriveService, 'reset')
    def test_reset(self, mock_restart_drive_reset, mock_restart_master_reset,
                   mock_get_resource_node, mock_kill_master_service_ssh,
                   mock_kill_drive_service):
        mock_get_resource_node.side_effect = [self.node_dms[0], self.node_dms[1]]
        self.restart_node_service_cmt.apply()
        self.restart_node_service_cmt._state = components.States.failed
        self.restart_node_service_cmt.reset()
        mock_restart_drive_reset.assert_called()
        mock_restart_master_reset.assert_called()

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    def test_node_dms(self, mock_get_resource_node, mock_kill_master_service_ssh,
                      mock_kill_drive_service):
        mock_get_resource_node.side_effect = [self.node_dms[0], self.node_dms[1]]
        self.restart_node_service_cmt.apply()
        assert self.restart_node_service_cmt.node_dms == [self.node_dms[0], self.node_dms[1]]


class TestStopMasterServiceComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dm = models.ResourceNode.get_mock_object()
        cls.stop_master_service_cmt = _cluster._StopMasterService(cluster_dm=cls.cluster_dm,
                                                                  node_dm=cls.node_dm)

    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    def test_apply(self, mock_kill_service):
        mock_kill_service.return_value = ''
        self.stop_master_service_cmt.apply()
        mock_kill_service.assert_called_with(signal=enums.UnixSignal.SIGQUIT)

    @mock.patch.object(functional_models.Node, 'stop_binary')
    @mock.patch.object(functional_models.Node, 'start_binary')
    def test_reset(self, mock_start_binary, mock_stop_binary):
        self.stop_master_service_cmt._state = components.States.applied
        self.stop_master_service_cmt.reset()
        mock_stop_binary.assert_called_with(enums.NodeBinary.solidfire)
        mock_start_binary.assert_called_with(enums.NodeBinary.solidfire)


class TestStopDriveServiceComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dm = models.ResourceNode.get_mock_object()
        drive_dm = models.ClusterServiceDrive.get_mock_object()
        cls.service_dm = models.ClusterServiceInfo.get_mock_object(overrides={'drives': [drive_dm]})
        cls.stop_drive_service_cmt = _cluster._StopDriveService(cluster_dm=cls.cluster_dm,
                                                                node_dm=cls.node_dm,
                                                                service_dm=cls.service_dm)
        failed_drive_dm = models.ClusterServiceDrive.get_mock_object(
            overrides={'drive_status': enums.DriveStatus.failed})
        cls.failed_drive_service_dm = models.ClusterServiceInfo.get_mock_object(
            overrides={'drives': [failed_drive_dm]})
        cls.stop_failed_drive_service_cmt = _cluster._StopDriveService(
            cluster_dm=cls.cluster_dm, node_dm=cls.node_dm,
            service_dm=cls.failed_drive_service_dm)

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    def test_apply(self, mock_kill_service):
        mock_kill_service.return_value = ''
        self.stop_drive_service_cmt.apply()
        mock_kill_service.assert_called_with(service_dm=self.service_dm,
                                             signal=enums.UnixSignal.SIGQUIT)

    @mock.patch.object(vl_helpers, 'wait_for_drive_status')
    @mock.patch.object(vl_helpers, 'remove_drives')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers, 'add_drives')
    def test_reset(self, mock_add_drives, mock_list_drives,
                   mock_remove_drives, mock_wait_drive_status):
        self.stop_drive_service_cmt._state = components.States.applied
        drive_dms = self.service_dm.drives
        mock_list_drives.return_value = drive_dms
        self.stop_drive_service_cmt.reset()
        mock_add_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms)

    @mock.patch.object(vl_helpers, 'wait_for_drive_status')
    @mock.patch.object(vl_helpers, 'remove_drives')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers, 'add_drives')
    def test_failed_drive_reset(self, mock_add_drives, mock_list_drives,
                                mock_remove_drives, mock_wait_drive_status):
        self.stop_failed_drive_service_cmt._state = components.States.applied
        drive_dms = self.failed_drive_service_dm.drives
        mock_list_drives.return_value = drive_dms
        self.stop_failed_drive_service_cmt.reset()
        mock_add_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms)

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    def test_node_dms(self, mock_get_resource_node, mock_kill_drive_service):
        _stop_drive_service_cmt = _cluster._StopDriveService(cluster_dm=self.cluster_dm,
                                                             node_dm=self.node_dm,
                                                             service_dm=self.service_dm)

        _stop_drive_service_cmt.apply()
        assert _stop_drive_service_cmt.node_dm == self.node_dm


class TestStopNodeServicesComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

        cluster_service_master = models.ClusterService.get_mock_object(overrides={
            'service_type': enums.ServiceType.master})
        cluster_service_block = models.ClusterService.get_mock_object(overrides={
            'service_type': enums.ServiceType.block})
        service_objects = []
        service_objects.append(models.ClusterServiceInfo.get_mock_object(overrides={
            'service': cluster_service_master}))
        service_objects.append(models.ClusterServiceInfo.get_mock_object(overrides={
            'service': cluster_service_block}))
        cls.node_dms = models.ResourceNode.get_mock_object_list(2)
        cls.service_selector = lambda x: service_objects

    def setUp(self):
        self.stop_node_service_cmt = _cluster.StopNodeServices(
            self.cluster_dm, self.service_selector)

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    @mock.patch.object(_cluster._StopMasterService, '_apply')
    @mock.patch.object(_cluster._StopDriveService, '_apply')
    def test_apply(self, mock_stop_drive_apply, mock_stop_master_apply,
                   mock_get_resource_node, mock_kill_master_service_ssh,
                   mock_kill_drive_service_ssh):
        mock_get_resource_node.side_effect = [self.node_dms[0], self.node_dms[1]]
        mock_kill_master_service_ssh.return_value = ''
        mock_kill_drive_service_ssh.return_value = ''
        self.stop_node_service_cmt.apply()
        mock_stop_master_apply.assert_called()
        mock_stop_drive_apply.assert_called()

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    @mock.patch.object(_cluster._StopMasterService, 'reset')
    @mock.patch.object(_cluster._StopDriveService, 'reset')
    def test_reset(self, mock_stop_drive_reset, mock_stop_master_reset,
                   mock_get_resource_node, mock_kill_master_service_ssh,
                   mock_kill_drive_service):
        mock_get_resource_node.side_effect = [self.node_dms[0], self.node_dms[1]]
        self.stop_node_service_cmt.apply()
        self.stop_node_service_cmt._state = components.States.failed
        self.stop_node_service_cmt.reset()
        mock_stop_drive_reset.assert_called()
        mock_stop_master_reset.assert_called()

    @mock.patch.object(functional_models.Node, 'kill_drive_service_ssh')
    @mock.patch.object(functional_models.Node, 'kill_master_service_ssh')
    @mock.patch.object(models, 'get_resource_node_by_static_id')
    def test_node_dms(self, mock_get_resource_node, mock_kill_master_service_ssh,
                      mock_kill_drive_service):
        _stop_node_service_cmt = _cluster.StopNodeServices(self.cluster_dm, self.service_selector)

        mock_get_resource_node.side_effect = [self.node_dms[0], self.node_dms[1]]
        _stop_node_service_cmt.apply()
        assert _stop_node_service_cmt.node_dms == [self.node_dms[0], self.node_dms[1]]


@mock.patch('time.sleep', mock.MagicMock())
class TestStartGCComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.gc_cmt = _cluster.RunGarbageCollection(self.cluster_dm)
        self.mock_async_handler = mock.MagicMock()
        self.mock_async_handler.is_running.return_value = False

    def test_mock(self):
        mock_gc_cmt = _cluster.RunGarbageCollection.mock(self.cluster_dm)
        mock_gc_cmt.apply()
        mock_gc_cmt.reset()

    @mock.patch.object(_constants.DisableAutomaticGarbageCollection, 'reset')
    @mock.patch.object(_constants.DisableAutomaticGarbageCollection, 'apply')
    @mock.patch.object(functional_models.Cluster, 'start_gc')
    @mock.patch.object(functional_models.Cluster, 'abort_gc')
    def test_apply(self, abort_gc, start_gc, disable_gc_apply, disable_gc_reset):
        start_gc.return_value = self.mock_async_handler
        self.gc_cmt.apply()
        assert disable_gc_apply.called
        assert abort_gc.called
        assert start_gc.called
        assert self.mock_async_handler.is_running.call_count == 1
        assert self.gc_cmt.get_async_handler() is self.mock_async_handler
        assert disable_gc_apply.called
        assert disable_gc_reset.called

    @mock.patch.object(_constants.DisableAutomaticGarbageCollection, 'reset')
    @mock.patch.object(_constants.DisableAutomaticGarbageCollection, 'apply')
    @mock.patch.object(functional_models.Cluster, 'get_gc_status')
    @mock.patch.object(functional_models.Cluster, 'start_gc')
    @mock.patch.object(functional_models.Cluster, 'abort_gc')
    def test_async_hangs(self, abort_gc, start_gc, get_gc_status, disable_gc_apply, disable_gc_reset):
        self.mock_async_handler.is_running.side_effect = [True, True, True]
        start_gc.return_value = self.mock_async_handler
        get_gc_status.side_effect = [
            models.GarbageCollectionDetails.create(status=enums.GcStatus.in_progress),
            models.GarbageCollectionDetails.create(status=enums.GcStatus.completed_partial)
        ]
        self.gc_cmt.apply()
        assert start_gc.called
        assert self.mock_async_handler.is_running.call_count == 2
        assert get_gc_status.call_count == 2

    @mock.patch.object(_constants.DisableAutomaticGarbageCollection, 'ensure_ready')
    def test_reset(self, mock_ensure_ready):
        gc_cmt = _cluster.RunGarbageCollection(self.cluster_dm)
        gc_cmt._state = components.States.failed
        gc_cmt.reset()
        mock_ensure_ready.assert_called()

    def test_status(self):
        status = enums.GcStatus.completed_partial
        mock_details = models.GarbageCollectionDetails.get_mock_object(
            overrides=dict(status=status)
        )
        mock_async_result = models.AsyncResult.get_mock_object(
            overrides=dict(details=mock_details)
        )
        self.mock_async_handler.get_async_result.return_value = mock_async_result
        self.gc_cmt._async_handler_fm = self.mock_async_handler
        self.gc_cmt._state = components.States.applied
        assert self.gc_cmt.get_gc_status() is status

    def test_status_ready(self):
        with pytest.raises(components.ComponentStateError):
            self.gc_cmt.get_gc_status()

    def test_status_old_version(self):
        mock_details = models.GarbageCollectionDetailsPreMg.get_mock_object()
        mock_async_result = models.AsyncResult.get_mock_object(
            overrides=dict(details=mock_details)
        )
        self.mock_async_handler.get_async_result.return_value = mock_async_result
        self.gc_cmt._async_handler_fm = self.mock_async_handler
        self.gc_cmt._state = components.States.applied
        with pytest.raises(errors.ElementVersionError):
            self.gc_cmt.get_gc_status()


class TestRoundRobinRestartServicesByNode(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.rrr_cmt = _cluster.RoundRobinRestartServicesByNode(self.cluster_dm,
                                                                service_types=[enums.NodeService.block],
                                                                max_iterations=2)

    @mock.patch.object(selectors.ClusteredNodeSelector, '__call__')
    @mock.patch.object(selectors.ServiceSelector, '__call__')
    @mock.patch.object(_cluster.RestartNodeServices, 'apply')
    @mock.patch.object(vl_helpers, 'wait_for_healthy_services')
    def test_apply(self, mock_wait_for_healthy, mock_restart_apply, mock_service_call,
                   mock_node_call):
        mock_service_call.return_value = models.ClusterServiceInfo.get_mock_object_list(1)
        node_dms = models.NodeBase.get_mock_object_list(1)
        mock_node_call.side_effect = [node_dms, selectors.NotEnoughItemsError(), node_dms, node_dms]
        self.rrr_cmt.apply()

    def test_reset(self):
        self.rrr_cmt.state = components.States.failed
        self.rrr_cmt.reset()


class TestRunClusterBSCheck(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def setUp(self, mock_get_constants):
        self.rcbc_cmt = _cluster.RunClusterBSCheck(self.cluster_dm, heavy=True, full_throttle=True)
        self.mock_async_handler = mock.MagicMock()

    @mock.patch.object(vl_components.RunClusterBSCheck, 'get_unknown_blocks')
    @mock.patch.object(functional_models.Cluster, 'get_async_result')
    @mock.patch.object(sf_api.ClusterApi, 'call_method')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_apply_async_and_apply_finish(self,
                                          mock_list_events,
                                          mock_call_method,
                                          mock_get_async_result,
                                          mock_get_unknown_blocks):
        mock_response = result_models.ListNodeStats.get_mock_response()
        mock_call_method.return_value = mock_response
        self.rcbc_cmt._apply_async()

        self.rcbc_cmt._apply_is_running = mock.Mock(return_value=False)

        overrides = dict(status=enums.AsyncResultStatus.complete)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        mock_get_unknown_blocks.return_value = None
        self.rcbc_cmt._apply_finish()

        self.rcbc_cmt._apply_async()
        overrides = dict(status=enums.AsyncResultStatus.failed)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        with pytest.raises(components.ComponentAsyncError):
            self.rcbc_cmt._apply_finish()

        self.rcbc_cmt._apply_async()
        overrides = dict(status=enums.AsyncResultStatus.complete)
        mock_get_async_result.return_value = models.AsyncResult.get_mock_object(overrides=overrides)
        mock_get_unknown_blocks.return_value = [{'volumeID': 2,
                                                 'lba': '10',
                                                 'blockID': '3faa8466fffe98119c779b559e84dba3'}]
        with pytest.raises(components.UnknownBlockIDError):
            self.rcbc_cmt._apply_finish()

    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_get_unknown_blocks(self, mock_list_events, mock_list_volumes):
        mock_list_events.return_value = [models.ClusterEvent.get_mock_object(
            overrides=dict(event_info_type=enums.ClusterEventType.data_event,
                           message='01234567890123456789012345678901 99 99',
                           details=dict(source='bscheck')
                           ))]
        self.rcbc_cmt.get_unknown_blocks()

        mock_list_events.return_value = [models.ClusterEvent.get_mock_object(
            overrides=dict(event_info_type=enums.ClusterEventType.data_event,
                           message='flag 01234567890123456789012345678901 99 99',
                           details=dict(source='bscheck')
                           ))]
        self.rcbc_cmt.get_unknown_blocks()

        mock_list_volumes.return_value = [models.Volume.get_mock_object(
            overrides=dict(attributes=(dict(fillSliceStatus='running'))))]
        self.rcbc_cmt.get_unknown_blocks()

    @mock.patch('valence.functional_models.AsyncHandler.is_running')
    def test_apply_is_running(self, mock_is_running):
        cluster_fm = functional_models.Cluster.get_mock_fm()
        self.rcbc_cmt._async_handler_fm = functional_models.AsyncHandler(cluster_fm)
        self.rcbc_cmt._apply_is_running()

    @mock.patch('valence.functional_models.Cluster.stop_cluster_bs_check')
    def test_kill_apply_async(self, mock_stop_bs_check):
        self.rcbc_cmt._kill_apply_async()

    def test_reset(self):
        self.rcbc_cmt._reset()


@mock.patch.object(functional_models.Cluster, 'get_async_result')
@mock.patch('time.time')
@unittest_utils.patch_sleep()
@mock.patch('valence.vl_helpers.reboot_slice_services')
@mock.patch('valence.vl_helpers.wait_for_cluster_idle')
@mock.patch('valence.functional_models.AsyncHandler.join')
@mock.patch('valence.functional_models.Cluster.abort_gc')
@mock.patch('valence.functional_models.Cluster.start_gc')
@mock.patch('valence.vl_helpers.is_pcache_empty')
@mock.patch.object(functional_models.Cluster, 'set_constants')
@mock.patch.object(functional_models.Cluster, 'get_constants')
@mock.patch(
    'valence.functional_models.Cluster.get_current_version',
    mock.MagicMock(return_value=sf_version.SfVersion('12.0'))
)
class TestFlushPcache(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(RS_CLUSTER_DM)
        cls.mock_start_gc_response = result_models.StartGC.get_mock_object()

    def setUp(self):
        self.fpc_cmt = _cluster.FlushPcache(self.cluster_dm)

    def test_apply(self, mock_get_constants, mock_set_constants, mock_is_pcache_empty,
                   mock_start_gc, mock_abort_gc, mock_async_join, mock_wait_for_cluster_idle,
                   mock_reboot_slice_services, mock_time, mock_get_async_result):
        mock_is_pcache_empty.side_effect = [False, False, True]
        mock_time.return_value = 1
        mock_start_gc.return_value = functional_models.GarbageCollectionAsyncHandler(
            self.cluster_fm,
            models.AsyncHandle(self.mock_start_gc_response)
        )
        default_constants = dict(
            cPCacheIgnoreMinFlushSizeIntervalSeconds=120,
            cDisablePCacheCleanFiles=False
        )
        mock_get_constants.return_value = default_constants
        self.fpc_cmt.apply()
        assert mock_is_pcache_empty.call_count == 3
        assert mock_wait_for_cluster_idle.call_count == 1
        assert mock_reboot_slice_services.call_count == 1
        assert mock_set_constants.call_count == 4
        assert mock_set_constants.call_args_list[0][0][0].serialize() == dict(
            cPCacheIgnoreMinFlushSizeIntervalSeconds=1,
            cDisablePCacheCleanFiles=True
        )
        assert mock_set_constants.call_args_list[1][0][0].serialize() == dict(
            cSkipGC=True
        )
        assert mock_set_constants.call_args_list[2][0][0] == default_constants
        assert mock_start_gc.called
        assert mock_abort_gc.call_count == mock_start_gc.call_count
        assert self.fpc_cmt._gc_cmt.state == components.States.ready

        self.fpc_cmt.reset()
        assert self.fpc_cmt._set_constants_cmt.state == components.States.ready
        assert self.fpc_cmt._gc_cmt.state == components.States.ready

        self.fpc_cmt._state = components.States.failed
        self.fpc_cmt.reset()
        assert self.fpc_cmt.state == components.States.ready

    def test_apply_empty(self, mock_get_constants, mock_set_constants, mock_is_pcache_empty,
                         mock_start_gc, mock_abort_gc, mock_async_join, mock_wait_for_cluster_idle,
                         mock_reboot_slice_services, mock_time, mock_get_async_result):
        mock_is_pcache_empty.return_value = True
        mock_start_gc.return_value = functional_models.GarbageCollectionAsyncHandler(
            self.cluster_fm,
            models.AsyncHandle(self.mock_start_gc_response)
        )
        self.fpc_cmt.apply()
        assert mock_set_constants.call_count == 4
        assert mock_is_pcache_empty.call_count == 1
        assert mock_get_constants.call_count == 2
        assert mock_set_constants.call_count == 4
        assert self.fpc_cmt._gc_cmt.state == components.States.ready

    def test_apply_timeout(self, mock_get_constants, mock_set_constants, mock_is_pcache_empty,
                           mock_start_gc, mock_abort_gc, mock_async_join, mock_wait_for_cluster_idle,
                           mock_reboot_slice_services, mock_time, mock_get_async_result):
        mock_is_pcache_empty.return_value = False
        mock_time.side_effect = [0, 1, 1, 601]
        with pytest.raises(errors.TimeoutError):
            self.fpc_cmt.apply()
        assert mock_is_pcache_empty.call_count == 2
        assert mock_wait_for_cluster_idle.call_count == 1
        assert mock_reboot_slice_services.call_count == 1
        assert mock_set_constants.call_count == 1
        assert mock_get_constants.call_count == 1
