from __future__ import division

import random
import string
import unittest

import mock
import pytest

from camelot.models import at2_enums
from valence import (
    components,
    enums,
    models,
)
from valence.vl_components import (
    _account,
    _connection,
    _initiator,
    _volume,
    _volume_access_group,
)


class TestConvertDiskMap(unittest.TestCase):

    def test_convert_disk_map(self):
        disk_map = {'1': 'string', '2': 'stringTwo'}
        int_disk_map = _connection.convert_disk_map(disk_map)
        int_keys = list(int_disk_map.keys())
        assert isinstance(int_keys[0], int)
        assert isinstance(int_keys[1], int)


class BaseTestConnections(object):
    @classmethod
    def setUpClass(cls):
        # define the data models
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_dm.api_endpoint = models.SfVersion('9.0')
        cls.client_dm = models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04))
        cls.account_dm = models.Account.create(
            username='unittest-TAG-0', status='active', account_id=1)

        # full disk map because not all volumes are used.
        cls.full_disk_map = {61: '/dev/sdc', 62: '/dev/sde', 63: '/dev/sdb'}
        cls.disk_map = dict(cls.full_disk_map)
        cls.disk_map.pop(61)
        cls.volume_dms = [
            models.Volume.create(
                name='unittest-TAG-{}'.format(i),
                account_id=1,
                volume_id=i,
                total_size=models.DataSize(i, enums.DataUnit.gibibytes),
                qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
                enable512e=False
            ) for i in cls.disk_map
        ]
        cls.initiator_dm = models.Initiator.get_mock_object(overrides=dict(
            alias=cls.client_dm.client_name + '-rand_init',
            initiator_name='init1',
            initiator_id=1,
        ))
        cls.vag_dm = models.VolumeAccessGroup.get_mock_object(overrides=dict(
            name='test1',
            volume_access_group_id=1,
            volumes=[1, 2],
            initiators=['init1'],
            initiator_ids=[1],
        ))

        # define the component
        cls.accounts_cmt = _account.MockAccounts(
            tag='mock',
            account_dms=[cls.account_dm]
        )
        cls.volumes_cmt = _volume.MockVolumes(
            tag='mock',
            cluster_dm=cls.cluster_dm,
            volume_dms=cls.volume_dms,
            accounts_cmt=cls.accounts_cmt
        )
        cls.initiators_cmt = _initiator.MockInitiators(
            tag='mock',
            client_dms=[cls.client_dm],
            initiator_dms_map={cls.client_dm.mip: [cls.initiator_dm]}
        )
        cls.vag_cmt = _volume_access_group.MockVolumeAccessGroup(
            volume_access_group_dm=cls.vag_dm,
            initiators_cmt=cls.initiators_cmt,
            volumes_cmt=cls.volumes_cmt
        )


class TestCreateIscsiAccountConnections(BaseTestConnections, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCreateIscsiAccountConnections, cls).setUpClass()

        cls.conn_cmt = _connection.CreateIscsiAccountConnection(
            cluster_dm=cls.cluster_dm,
            client_dm=cls.client_dm,
            volumes_cmts=[cls.volumes_cmt],
            accounts_cmt=cls.accounts_cmt,
        )

    @mock.patch('valence.functional_models.LinuxClient.volume_ceremony')
    @mock.patch('valence.functional_models.LinuxClient.update_iscsi_d_conf')
    def test_apply(self, mock_update_iscsi_d_conf, mock_volume_ceremony):
        mock_volume_ceremony.return_value = self.disk_map
        self.conn_cmt.apply()
        assert self.conn_cmt.client_disk_map == {
            self.client_dm.mip: self.disk_map,
        }

    def test__reset(self):
        self.conn_cmt._reset()
        with pytest.raises(components.ComponentStateError):
            self.conn_cmt.client_disk_map()

    def test_client_dms(self):
        client_dms = self.conn_cmt.client_dms
        assert client_dms == [self.client_dm]

    def test_mock(self):
        volumes_cmt = _volume.CreateMultipleVolumes(
            tag='mock',
            cluster_dm=self.cluster_dm,
            vol_count=3,
            partial_volume_dm=models.Volume.create(
                total_size=models.DataSize(20, enums.DataUnit.gibibytes)
            ),
            accounts_cmt=self.accounts_cmt
        )
        mock_conn_cmt = self.conn_cmt.mock(
            cluster_dm=self.cluster_dm,
            client_dm=self.client_dm,
            volumes_cmts=[volumes_cmt, self.volumes_cmt],
            accounts_cmt=self.accounts_cmt
        )
        disk_map = mock_conn_cmt._original_disk_map

        # Ensure the volume IDs generated for the disk map are unique
        assert len(set([volume_id for volume_id in disk_map])) == 3 + len(self.disk_map)
        for volume_id in self.disk_map:
            assert volume_id in disk_map


class TestCreateIscsiVagConnections(BaseTestConnections, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCreateIscsiVagConnections, cls).setUpClass()

        cls.conn_cmt = _connection.CreateIscsiVagConnection(
            cluster_dm=cls.cluster_dm,
            client_dm=cls.client_dm,
            vag_cmts=[cls.vag_cmt]
        )

    @mock.patch('valence.functional_models.LinuxClient.volume_ceremony')
    @mock.patch('valence.functional_models.LinuxClient.reactivate_chap')
    @mock.patch('valence.functional_models.LinuxClient.deactivate_chap')
    def test_apply(self, mock_deactivate_chap, mock_reactivate_chap, mock_volume_ceremony):
        mock_volume_ceremony.return_value = self.disk_map
        self.conn_cmt.apply()
        assert self.conn_cmt.client_disk_map == {
            self.client_dm.mip: self.disk_map,
        }

    @mock.patch('valence.functional_models.LinuxClient.reactivate_chap')
    def test__reset(self, mock_reactivate_chap):
        self.conn_cmt._reset()
        with pytest.raises(components.ComponentStateError):
            self.conn_cmt.client_disk_map()


class TestManageConnections(BaseTestConnections, unittest.TestCase):
    def setUp(self):
        super(TestManageConnections, self).setUp()
        self.conn_cmt_1 = _connection.MockConnection(
            self.disk_map,
            self.client_dm,
            [self.volumes_cmt],
        )

        self.conn_cmt = _connection.ManageConnections(
            [self.conn_cmt_1]
        )
        self.conn_cmt.accounts_cmts = [self.accounts_cmt]
        self.conn_cmt.volumes_cmts = [self.volumes_cmt]
        self.conn_cmt.initiators_cmts = [self.initiators_cmt]
        self.conn_cmt.vag_cmts = [self.vag_cmt]

        self.client_disk_map = {
            self.client_dm.mip: self.disk_map,
        }

    def test_apply(self):
        with pytest.raises(components.ComponentStateError):
            self.conn_cmt.client_disk_map

        self.conn_cmt.apply()
        assert self.accounts_cmt.state == components.States.applied
        assert self.volumes_cmt.state == components.States.applied
        assert self.initiators_cmt.state == components.States.applied
        assert self.vag_cmt.state == components.States.applied
        assert self.conn_cmt.client_disk_map == self.client_disk_map

        self.conn_cmt.reset()
        assert self.accounts_cmt.state == components.States.ready
        assert self.volumes_cmt.state == components.States.ready
        assert self.initiators_cmt.state == components.States.ready
        assert self.vag_cmt.state == components.States.ready
        with pytest.raises(components.ComponentStateError):
            self.conn_cmt.client_disk_map

    def test_child_cmts(self):
        assert self.conn_cmt.child_cmts == [self.conn_cmt_1]

    def test_client_dms(self):
        assert self.conn_cmt.client_dms == [self.client_dm]

    def test_same_clients(self):
        children = [
            _connection.MockConnection(
                {}, self.client_dm, [self.volumes_cmt]
            )
            for _ in range(3)
        ]
        with pytest.raises(ValueError):
            _connection.ManageConnections(children)

    def test_create_initiators(self):
        client_dms = [models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04)
        ) for _ in range(7)]

        with pytest.raises(AssertionError):
            initiators_cmts = _connection.ManageConnections._create_connection_initiators(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=client_dms,
                max_initiator_count=1,
                clients_per_initiator=6
            )

        with pytest.raises(AssertionError):
            initiators_cmts = _connection.ManageConnections._create_connection_initiators(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=client_dms,
                max_initiator_count=7,
                clients_per_initiator=8
            )

        initiators_cmts = _connection.ManageConnections._create_connection_initiators(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=client_dms,
            max_initiator_count=4,
            clients_per_initiator=3
        )
        assert len(initiators_cmts) <= 4
        assert all([len(cmt.client_dms) == 3 for cmt in initiators_cmts])

        # Ensure every client is represented in the initiators
        client_mips = {dm.mip for dm in client_dms}
        init_client_mips = set()
        for initiators_cmt in initiators_cmts:
            init_client_mips.update({dm.mip for dm in initiators_cmt.client_dms})
        assert not init_client_mips.difference(client_mips)

    def test_create_vags(self):
        client_dms = [models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04)
        ) for _ in range(7)]
        initiators_cmts = _connection.ManageConnections._create_connection_initiators(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=client_dms,
            max_initiator_count=4,
            clients_per_initiator=3
        )
        volumes_cmts = [
            _volume.MockVolumes(
                tag='mock-{}'.format(i),
                cluster_dm=self.cluster_dm,
                volume_dms=[models.Volume.get_mock_object()],
                accounts_cmt=self.accounts_cmt
            ) for i in range(10)
        ]

        with pytest.raises(AssertionError):
            vag_cmts = _connection.ManageConnections._create_connection_vags(
                cluster_dm=self.cluster_dm,
                initiators_cmts=initiators_cmts,
                volumes_cmts=volumes_cmts[:len(initiators_cmts) - 1]
            )

        vag_cmts = _connection.ManageConnections._create_connection_vags(
            cluster_dm=self.cluster_dm,
            initiators_cmts=initiators_cmts,
            volumes_cmts=volumes_cmts
        )
        assert len(vag_cmts) == len(initiators_cmts)
        assert all([cmt.name.startswith('unittest') for cmt in vag_cmts])

        # Ensure every volumes component is represented in the vags
        volume_tags = {cmt.tag for cmt in volumes_cmts}
        vag_volume_tags = set()
        for vag_cmt in vag_cmts:
            vag_volume_tags.update({cmt.tag for cmt in vag_cmt.volumes_cmt.child_cmts})
        assert not vag_volume_tags.difference(volume_tags)


class TestConnectionsConstructors(BaseTestConnections, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestConnectionsConstructors, cls).setUpClass()
        ubuntu = {'os': at2_enums.ClientOS.ubuntu_14_04}
        centos = {'os': at2_enums.ClientOS.centos_7}
        cls.client_dms = [models.ResourceClient.get_mock_object(overrides=ubuntu if cl % 2 == 0
                                                                else centos) for cl in range(7)]
        for cl in cls.client_dms:
            cl.client_name = ''.join(random.sample(string.ascii_lowercase, 10))
        cls.volumes_cmts_even = [_volume.MockVolumes(
            tag='mock-{}'.format(i),
            cluster_dm=cls.cluster_dm,
            volume_dms=[models.Volume.create() for _ in range(30)],
            accounts_cmt=cls.accounts_cmt
        ) for i in range(3)]
        cls.volumes_cmts_uneven = [
            _volume.MockVolumes(
                tag='mock-0',
                cluster_dm=cls.cluster_dm,
                volume_dms=[models.Volume.get_mock_object() for _ in range(2)],
                accounts_cmt=cls.accounts_cmt
            ),
            _volume.MockVolumes(
                tag='mock-1',
                cluster_dm=cls.cluster_dm,
                volume_dms=[models.Volume.get_mock_object() for _ in range(1)],
                accounts_cmt=cls.accounts_cmt
            ),
            _volume.MockVolumes(
                tag='mock-2',
                cluster_dm=cls.cluster_dm,
                volume_dms=[models.Volume.get_mock_object() for _ in range(1)],
                accounts_cmt=cls.accounts_cmt
            )
        ]

        cls.initiators_cmts_mixed = _connection.ManageConnections._create_connection_initiators(
            tag='unittest',
            cluster_dm=cls.cluster_dm,
            client_dms=cls.client_dms,
            max_initiator_count=3,
            clients_per_initiator=3
        )
        cls.vag_cmts_mixed = _connection.ManageConnections._create_connection_vags(
            cluster_dm=cls.cluster_dm,
            initiators_cmts=cls.initiators_cmts_mixed,
            volumes_cmts=cls.volumes_cmts_uneven[:3]
        )

        cls.initiators_cmts_omni = _connection.ManageConnections._create_connection_initiators(
            tag='unittest',
            cluster_dm=cls.cluster_dm,
            client_dms=cls.client_dms[:2],
            max_initiator_count=2,
            clients_per_initiator=2
        )
        cls.vag_cmts_omni = _connection.ManageConnections._create_connection_vags(
            cluster_dm=cls.cluster_dm,
            initiators_cmts=cls.initiators_cmts_omni,
            volumes_cmts=cls.volumes_cmts_uneven[:2]
        )

        cls.initiators_cmts_apply = _connection.ManageConnections._create_connection_initiators(
            tag='unittest',
            cluster_dm=cls.cluster_dm,
            client_dms=cls.client_dms[:3],
            max_initiator_count=3,
            clients_per_initiator=2
        )
        cls.vag_cmts_apply = _connection.ManageConnections._create_connection_vags(
            cluster_dm=cls.cluster_dm,
            initiators_cmts=cls.initiators_cmts_apply,
            volumes_cmts=cls.volumes_cmts_even
        )

    @mock.patch('valence.vl_components._connection.CreateIscsiVagConnection',
                _connection.CreateIscsiVagConnection.mock)
    def test_from_client_dms_even(self):
        client_dms = self.client_dms[:3]
        manage_conns_cmt = _connection.ManageConnections.from_client_dms(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=client_dms,
            volume_count=90,
            connections_per_volume=2
        )
        unique_volumes_cmts = set()
        assert len(manage_conns_cmt.child_cmts) == 3
        for conn_cmt in manage_conns_cmt.child_cmts:
            assert len(conn_cmt.client_dms) == 1
            assert conn_cmt.client_dms[0] in client_dms
            assert len(conn_cmt.volumes_cmts) == 2
            for volumes_cmt in conn_cmt.volumes_cmts:
                assert len(volumes_cmt.partial_volume_dms) == 30
                unique_volumes_cmts.add(volumes_cmt)

        assert len(unique_volumes_cmts) == 3

    @mock.patch('valence.vl_components._connection.CreateIscsiVagConnection',
                _connection.CreateIscsiVagConnection.mock)
    def test_from_client_dms_uneven(self):
        client_dms = self.client_dms[:3]
        manage_conns_cmt = _connection.ManageConnections.from_client_dms(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=client_dms,
            volume_count=4
        )
        total_volume_count = 0
        assert len(manage_conns_cmt.child_cmts) == 3
        for conn_cmt in manage_conns_cmt.child_cmts:
            assert len(conn_cmt.client_dms) == 1
            assert conn_cmt.client_dms[0] in client_dms
            assert len(conn_cmt.volumes_cmts) == 1
            volume_count = 0
            for volumes_cmt in conn_cmt.volumes_cmts:
                volume_count += len(volumes_cmt.partial_volume_dms)
            assert volume_count >= 1
            total_volume_count += volume_count

        assert total_volume_count == 4

    @mock.patch('valence.vl_components._volume_access_group.CreateVolumeAccessGroup')
    @mock.patch('valence.vl_components._initiator.CreateIscsiInitiators')
    def test_from_volumes_cmts_mixed_conn(self, create_initiators, create_vag):
        create_initiators.side_effect = self.initiators_cmts_mixed
        create_vag.side_effect = self.vag_cmts_mixed

        # Some volumes connected to some clients
        manage_conns_cmt = _connection.ManageConnections.from_volumes_cmts(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms,
            volumes_cmts=self.volumes_cmts_uneven,
            connections_per_volume=3
        )
        assert manage_conns_cmt.client_dms == self.client_dms
        assert len(manage_conns_cmt.child_cmts) == len(self.client_dms)
        assert not manage_conns_cmt.accounts_cmts
        assert not manage_conns_cmt.volumes_cmts
        assert manage_conns_cmt.initiators_cmts == self.initiators_cmts_mixed
        assert manage_conns_cmt.vag_cmts == self.vag_cmts_mixed
        assert create_initiators.call_args_list[-3][1]['client_dms'] == self.client_dms[0:3]
        assert create_initiators.call_args_list[-2][1]['client_dms'] == self.client_dms[3:6]
        assert create_initiators.call_args_list[-1][1]['client_dms'] == [
            self.client_dms[6], self.client_dms[0], self.client_dms[1]
        ]
        assert create_vag.call_args_list[-3][1]['volumes_cmt'].child_cmts == \
            [self.volumes_cmts_uneven[0]]
        assert create_vag.call_args_list[-2][1]['volumes_cmt'].child_cmts == \
            [self.volumes_cmts_uneven[1]]
        assert create_vag.call_args_list[-1][1]['volumes_cmt'].child_cmts == \
            [self.volumes_cmts_uneven[2]]

    @mock.patch('valence.vl_components._volume_access_group.CreateVolumeAccessGroup')
    @mock.patch('valence.vl_components._initiator.CreateIscsiInitiators')
    def test_from_volumes_cmts_omni_conn(self, create_initiators, create_vag):
        create_initiators.side_effect = self.initiators_cmts_omni
        create_vag.side_effect = self.vag_cmts_omni

        # Every volume connected to every client
        manage_conns_cmt = _connection.ManageConnections.from_volumes_cmts(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            volumes_cmts=self.volumes_cmts_uneven[:2],
            connections_per_volume=2
        )
        assert create_initiators.call_args_list[-1][1]['client_dms'] == self.client_dms[:2]
        assert set(create_vag.call_args_list[-1][1]['volumes_cmt'].child_cmts) == \
            set(self.volumes_cmts_uneven[:2])
        for conn_cmt in manage_conns_cmt.child_cmts:
            conn_cmt_volumes = set()
            for volumes_cmt in conn_cmt.volumes_cmts:
                conn_cmt_volumes.update(volumes_cmt.child_cmts)
            assert conn_cmt_volumes == set(self.volumes_cmts_uneven[:2])

    def test_from_volumes_cmts_legacy(self):
        cluster_dm = models.ResourceCluster.get_mock_object()
        cluster_dm.api_endpoint = models.SfVersion('8.6')
        manage_conns_cmt = _connection.ManageConnections.from_volumes_cmts(
            tag='unittest',
            cluster_dm=cluster_dm,
            client_dms=self.client_dms,
            volumes_cmts=self.volumes_cmts_even
        )
        for conn_cmt in manage_conns_cmt.child_cmts:
            assert isinstance(conn_cmt, _connection.CreateIscsiAccountConnection)

    @mock.patch('valence.functional_models.LinuxClient.volume_ceremony')
    @mock.patch('valence.functional_models.LinuxClient.reactivate_chap')
    @mock.patch('valence.functional_models.LinuxClient.deactivate_chap')
    @mock.patch('valence.vl_components._volume_access_group.CreateVolumeAccessGroup')
    @mock.patch('valence.vl_components._initiator.CreateIscsiInitiators')
    def test_from_volumes_cmts_apply(self, create_initiators, create_vag, deactivate_chap,
                                     reactivate_chap, volume_ceremony):
        create_initiators.side_effect = [
            _initiator.MockInitiators(
                tag=cmt.tag,
                client_dms=cmt.client_dms,
                initiator_dms_map={dm.mip: models.Initiator.get_mock_object() for dm in cmt.client_dms}
            )
            for cmt in self.initiators_cmts_apply
        ]
        create_vag.side_effect = [
            _volume_access_group.MockVolumeAccessGroup(models.VolumeAccessGroup.get_mock_object())
            for _ in self.vag_cmts_apply
        ]

        with pytest.raises(ValueError):
            manage_conns_cmt = _connection.ManageConnections.from_volumes_cmts(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms,
                volumes_cmts=self.volumes_cmts_even,
                connections_per_volume=8
            )

        with pytest.raises(ValueError):
            manage_conns_cmt = _connection.ManageConnections.from_volumes_cmts(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms,
                volumes_cmts=self.volumes_cmts_even,
                connections_per_volume=2
            )

        manage_conns_cmt = _connection.ManageConnections.from_volumes_cmts(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:3],
            volumes_cmts=self.volumes_cmts_even,
            connections_per_volume=2
        )

        # Ensure apply manages all child/sub components
        assert all([cmt.state == components.States.ready for cmt in manage_conns_cmt.child_cmts])
        manage_conns_cmt.apply()
        assert all([cmt.state == components.States.applied for cmt in manage_conns_cmt.child_cmts])
        assert all([cmt.state == components.States.applied for cmt in manage_conns_cmt.initiators_cmts])
        assert all([cmt.state == components.States.applied for cmt in manage_conns_cmt.vag_cmts])

        # ... assign disk maps to each child connections component
        for conn_cmt in manage_conns_cmt.child_cmts:
            conn_cmt._disk_map = {i: '/dev/sd{}'.format(i) for i in range(60)}

        # Ensure the client_disk_map reflects those of the child components
        for client_dm in self.client_dms[:3]:
            assert len(manage_conns_cmt.client_disk_map[client_dm.mip]) == 60

        # Ensure reset manages all child/sub components
        assert all([cmt.state == components.States.applied for cmt in manage_conns_cmt.child_cmts])
        manage_conns_cmt.reset()
        assert all([cmt.state == components.States.ready for cmt in manage_conns_cmt.child_cmts])
        assert all([cmt.state == components.States.ready for cmt in manage_conns_cmt.initiators_cmts])
        assert all([cmt.state == components.States.ready for cmt in manage_conns_cmt.vag_cmts])


class TestMockConnection(BaseTestConnections, unittest.TestCase):
    def setUp(self):
        self.conn_cmt = _connection.MockConnection(
            self.disk_map,
            self.client_dm,
            [self.volumes_cmt],
        )

    def test_apply(self):
        self.conn_cmt.apply()
        assert self.conn_cmt.state is components.States.applied

    def test_reset(self):
        self.conn_cmt.ensure_applied()
        assert self.conn_cmt.state is components.States.applied

        self.conn_cmt.reset()
        assert self.conn_cmt.state is components.States.ready
        self.conn_cmt.ensure_ready()
        assert self.conn_cmt.state is components.States.ready

    def test_get(self):
        with self.assertRaises(components.ComponentStateError):
            disk_map = self.conn_cmt.disk_map

        assert self.conn_cmt.client_dm == self.client_dm
        assert self.conn_cmt.client_dms == [self.client_dm]
        assert self.conn_cmt.volumes_cmts == [self.volumes_cmt]

        self.conn_cmt.apply()
        assert self.conn_cmt.disk_map == self.disk_map
