"""Test valence.vl_components._drive."""
import time
import unittest

import mock
import pytest
import sf_models

from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    vl_components,
    vl_helpers,
)


@mock.patch('valence.vl_helpers.wait_for_cluster_fault_resolution', mock.MagicMock)
@mock.patch('valence.vl_helpers.wait_for_cluster_fault_occurrence', mock.MagicMock)
@mock.patch('valence.vl_helpers.wait_for_drive_assignments', mock.MagicMock())
@mock.patch('valence.vl_helpers.wait_for_bin_sync_to_start', mock.MagicMock())
@mock.patch('valence.vl_helpers.wait_for_bin_sync_to_finish_with_timeout', mock.MagicMock())
class TestRemoveDrivesComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.active_drive_dms = models.Drive.get_mock_object_list(3, overrides={'status': 'active'})

        cls.active_drive_dms = [
            models.Drive.create(drive_id=i, status=enums.DriveStatus.active, capacity=100) for i
            in range(3)
        ]
        cls.available_drive_dms = [
            models.Drive.create(drive_id=i, status=enums.DriveStatus.available) for i in range(3)
        ]
        cls.partially_added_drive_dms = [
            models.Drive.create(drive_id=0, status=enums.DriveStatus.available),
            models.Drive.create(drive_id=1, status=enums.DriveStatus.active),
            models.Drive.create(drive_id=2, status=enums.DriveStatus.available)
        ]

        cls.selector = lambda x: cls.active_drive_dms

    def setUp(self):
        self.remove_drives_cmt = vl_components.RemoveDrives(self.cluster_dm, self.selector)

    @mock.patch.object(vl_helpers, 'get_active_node_dm_from_drive_dm')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.functional_models.Cluster.remove_drives')
    @mock.patch.object(vl_components.RemoveDrives, '_reload')
    def test_apply_async(self, mock_reload, mock_remove_drives, mock_get_constants,
                         mock_list_drives, mock_get_active_nodes):
        mock_constant_value = mock.Mock()
        mock_constant_value.bin_sync_write_mib_per_second = 20
        mock_constant_value.bin_sync_new_service_write_mib_per_second = 60
        mock_get_constants.return_value = mock_constant_value

        mock_reload.return_value = False
        async_handler_mock = mock.MagicMock()
        async_handler_mock.is_running.side_effect = [True, False]
        mock_remove_drives.return_value = async_handler_mock
        self.remove_drives_cmt.apply_async()
        mock_remove_drives.assert_called()

        assert self.remove_drives_cmt.is_running() is True
        assert self.remove_drives_cmt.is_running() is False
        self.remove_drives_cmt.finish()
        async_handler_mock.get_models.assert_called()

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.functional_models.Cluster.remove_drives')
    @mock.patch.object(vl_components.RemoveDrives, '_reload')
    @mock.patch('valence.vl_helpers.add_drives')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reset(self, mock_list_drives, mock_add_drives, mock_reload, mock_remove_drives,
                   mock_get_constants, mock_list_active_nodes):
        mock_constant_value = mock.Mock()
        mock_constant_value.bin_sync_write_mib_per_second = 20
        mock_constant_value.bin_sync_new_service_write_mib_per_second = 60
        mock_get_constants.return_value = mock_constant_value

        mock_reload.return_value = False
        mock_list_drives.return_value = self.available_drive_dms
        self.remove_drives_cmt._apply_is_running = mock.Mock(return_value=False)
        self.remove_drives_cmt.apply()
        self.remove_drives_cmt.reset()
        mock_add_drives.assert_called()
        assert len(mock_add_drives.call_args[1]['drive_dms']) == 3

        # Ensure reset recovers properly on a failed apply if not all drives were removed
        mock_list_drives.return_value = self.partially_added_drive_dms
        self.remove_drives_cmt.apply()
        self.remove_drives_cmt.state = components.States.failed
        self.remove_drives_cmt.reset()
        assert len(mock_add_drives.call_args[1]['drive_dms']) == 2

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.functional_models.Cluster.remove_drives')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reset_error(self, mock_list_drives, mock_remove_drives, mock_get_constants):
        mock_list_drives.side_effect = [self.active_drive_dms, self.partially_added_drive_dms]
        self.remove_drives_cmt.apply()
        with pytest.raises(components.PartialStateError):
            self.remove_drives_cmt.reset()

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload_success(self, mock_list_drives, mock_get_constants):
        mock_list_drives.return_value = self.active_drive_dms
        assert self.remove_drives_cmt._reload() is True

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload_fail(self, mock_list_drives, mock_get_constants):
        mock_list_drives.return_value = []  # if available, it would return nothing.
        assert self.remove_drives_cmt._reload() is False

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload_error(self, mock_list_drives, mock_get_constants):
        mock_list_drives.return_value = [1]  # doesn't really care if it's a drive or not.
        with pytest.raises(components.PartialStateError):
            self.remove_drives_cmt._reload()

    def test_mock(self):
        mock_remove_drives_cmt = vl_components.RemoveDrives.mock(self.cluster_dm, self.selector)
        mock_remove_drives_cmt.apply()
        mock_remove_drives_cmt.reset()
        assert mock_remove_drives_cmt.drive_dms == self.active_drive_dms


@mock.patch('time.sleep', mock.MagicMock())
@mock.patch('valence.vl_helpers.wait_for_bin_sync_to_start', mock.MagicMock())
@mock.patch('valence.vl_helpers.wait_for_bin_sync_to_finish_with_timeout', mock.MagicMock())
class TestFailDrivesComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.drive_dms = [
            models.Drive.create(drive_id=0, type=enums.DriveType.slice, status=enums.DriveStatus.active),
            models.Drive.create(drive_id=3, type=enums.DriveType.block, status=enums.DriveStatus.active),
            models.Drive.create(drive_id=7, type=enums.DriveType.block, status=enums.DriveStatus.active)
        ]
        cls.drive_selector = selectors.DriveSelector(cls.cluster_dm)

    @mock.patch('valence.vl_helpers.recover_failed_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_apply(self, list_drives, fail_drives, restore_drives, recover_drives):
        list_drives.return_value = self.drive_dms
        fail_drives_cmt = vl_components.FailDrives(self.cluster_dm, self.drive_selector)
        assert fail_drives_cmt.drive_dms == self.drive_dms

        list_drives.return_value = []
        fail_drives_cmt.apply(timeout=1)
        assert fail_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert fail_drives.call_args[1].get('wait_for_failure') is True
        assert fail_drives.call_args[1].get('timeout') == 1
        assert fail_drives_cmt.drive_dms == self.drive_dms

        fail_drives_cmt.reset(timeout=2)
        assert restore_drives.called
        assert restore_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert restore_drives.call_args[1].get('wait_for_restore') is True
        assert restore_drives.call_args[1].get('timeout') == 2
        assert recover_drives.called
        assert recover_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert recover_drives.call_args[1].get('wait_for_sync') is False
        assert 0 <= restore_drives.call_args[1].get('timeout') <= 2
        assert fail_drives_cmt.drive_dms == self.drive_dms

        list_drives.return_value = self.drive_dms[:1]
        fail_drives_cmt.apply()
        assert fail_drives_cmt.drive_dms == self.drive_dms[:1]

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload(self, list_drives):
        fail_drives_cmt = vl_components.FailDrives(self.cluster_dm, self.drive_selector)

        list_drives.return_value = self.drive_dms
        assert fail_drives_cmt._reload() is True

        list_drives.return_value = []
        assert fail_drives_cmt._reload() is False

        list_drives.return_value = self.drive_dms[:-1]
        with pytest.raises(components.PartialStateError):
            fail_drives_cmt._reload()

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_mock(self, list_drives):
        list_drives.return_value = self.drive_dms
        mock_fail_drives_cmt = vl_components.FailDrives.mock(self.cluster_dm, self.drive_selector)
        mock_fail_drives_cmt.apply()
        mock_fail_drives_cmt.reset()
        assert mock_fail_drives_cmt.drive_dms == self.drive_dms


class TestContinuousFailDrivesComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.active_drive_dms = models.Drive.get_mock_object_list(3, overrides={'status': 'active'})

    def setUp(self):
        self.fail_drives_cmt = vl_components.ContinuousFailDrives(
            cluster_dm=self.cluster_dm,
            kill_after_seconds=0.2
        )
        self.fail_drives_cmt_w_selector = vl_components.ContinuousFailDrives(
            cluster_dm=self.cluster_dm,
            drive_selector=lambda: self.active_drive_dms,
            kill_after_seconds=0.2
        )

    @mock.patch.object(vl_helpers, 'fail_drives_in_cluster')
    @mock.patch.object(vl_helpers, 'restore_failed_drives')
    @mock.patch.object(vl_components.FailDrives, '_reload')
    @mock.patch.object(selectors, 'ActiveNodeSelector')
    @mock.patch.object(selectors, 'DriveSelector')
    def test_apply_continuous(self, mock_drive, mock_active, mock_reload, mock_restore, mock_fail):
        mock_reload.return_value = False
        self.fail_drives_cmt.apply_async()
        self.fail_drives_cmt.finish()
        assert self.fail_drives_cmt.state is components.States.applied
        mock_restore.assert_called()
        mock_fail.assert_called()

    @mock.patch.object(vl_helpers, 'fail_drives_in_cluster')
    @mock.patch.object(vl_helpers, 'restore_failed_drives')
    @mock.patch.object(vl_components.FailDrives, '_reload')
    @mock.patch.object(selectors, 'ActiveNodeSelector')
    @mock.patch.object(selectors, 'DriveSelector')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reset(self, mock_list_drives, mock_drive, mock_active, mock_reload, mock_restore,
                   mock_fail):
        mock_reload.return_value = False
        mock_list_drives.return_value = []  # if available, it would return nothing.
        self.fail_drives_cmt.apply_async()
        self.fail_drives_cmt.finish()
        self.fail_drives_cmt._state = components.States.failed
        self.fail_drives_cmt.reset()
        assert self.fail_drives_cmt.state is components.States.ready
        mock_restore.assert_called()

    @mock.patch.object(vl_helpers, 'fail_drives_in_cluster')
    @mock.patch.object(vl_helpers, 'restore_failed_drives')
    @mock.patch.object(vl_components.FailDrives, '_reload')
    def test_apply_continuous_with_selector(self, mock_reload, mock_restore, mock_fail):
        mock_reload.return_value = False
        self.fail_drives_cmt_w_selector.apply_async()
        self.fail_drives_cmt_w_selector.finish()
        assert self.fail_drives_cmt_w_selector.state is components.States.applied
        mock_restore.assert_called()
        mock_fail.assert_called()


@mock.patch('time.sleep')
@mock.patch('valence.vl_helpers.wait_for_bin_sync_to_start', mock.MagicMock())
@mock.patch('valence.vl_helpers.wait_for_bin_sync_to_finish_with_timeout', mock.MagicMock())
@mock.patch(
    'valence.functional_models.Cluster.get_constants',
    mock.MagicMock(return_value=models.ClusterConstants.create(
        time_until_bs_is_killed_msec=330000,
        ss_health_monitor_interval_for_permanence=10
    ))
)
class TestCascadeFailDrivesComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_fms = [functional_models.Node.get_mock_fm() for _ in range(2)]
        cls.drive_dms = [
            models.Drive.create(drive_id=0, type=enums.DriveType.slice, status=enums.DriveStatus.active),
            models.Drive.create(drive_id=3, type=enums.DriveType.block, status=enums.DriveStatus.active)
        ]
        cls.drive_selector = selectors.FixedDriveSelector(cls.drive_dms)

    @mock.patch('valence.vl_helpers.recover_failed_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.wait_for_bin_assignments')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.vl_helpers.get_next_event_id')
    @mock.patch('valence.vl_helpers.check_fake_drives_flag_on_node')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.get_node_fms')
    def test_apply_all_default(self, get_node_fms, list_drives, check_fake_drives, get_next_event,
                               fail_drives, wait_for_bin_assignments, restore_drives, recover_drives,
                               mock_sleep):
        get_node_fms.return_value = self.node_fms
        get_next_event.return_value = 10
        list_drives.return_value = self.drive_dms
        fail_drives_cmt = vl_components.CascadeFailDrives(
            cluster_dm=self.cluster_dm,
            drive_selector=self.drive_selector,
        )
        with pytest.raises(components.ComponentStateError):
            drive_dms = fail_drives_cmt.drive_dms

        fail_drives_cmt.apply()

        assert check_fake_drives.call_count == 2
        check_fake_drives.assert_any_call(self.node_fms[0])
        check_fake_drives.assert_any_call(self.node_fms[1])

        assert fail_drives.call_count == 2
        fail_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=self.drive_dms, timeout=None,
                                       wait_for_failure=True)
        wait_for_bin_assignments.assert_called_with(mock.ANY, start_event_id=10)
        mock_sleep.assert_not_called()
        drive_dms = fail_drives_cmt.drive_dms
        assert len(drive_dms) == 4

        fail_drives_cmt.reset()
        restore_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800,
                                          wait_for_restore=True)
        recover_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800)

    @mock.patch('valence.vl_helpers.recover_failed_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.wait_for_bin_assignments')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.vl_helpers.get_next_event_id')
    @mock.patch('valence.vl_helpers.check_fake_drives_flag_on_node')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.get_node_fms')
    def test_apply_dont_wait_for_drive_failure(self, get_node_fms, list_drives, check_fake_drives,
                                               get_next_event, fail_drives, wait_for_bin_assignments,
                                               restore_drives, recover_drives, mock_sleep):
        get_node_fms.return_value = self.node_fms
        get_next_event.return_value = 10
        list_drives.return_value = self.drive_dms
        fail_drives_cmt = vl_components.CascadeFailDrives(
            cluster_dm=self.cluster_dm,
            drive_selector=self.drive_selector,
            wait_for_drive_failed=False
        )
        with pytest.raises(components.ComponentStateError):
            drive_dms = fail_drives_cmt.drive_dms

        fail_drives_cmt.apply()
        assert check_fake_drives.call_count == 2
        check_fake_drives.assert_any_call(self.node_fms[0])
        check_fake_drives.assert_any_call(self.node_fms[1])

        assert fail_drives.call_count == 2
        fail_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=self.drive_dms, timeout=None,
                                       wait_for_failure=False)
        wait_for_bin_assignments.assert_called_with(mock.ANY, start_event_id=10)
        mock_sleep.assert_not_called()
        drive_dms = fail_drives_cmt.drive_dms
        assert len(drive_dms) == 4

        fail_drives_cmt.reset()
        restore_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800,
                                          wait_for_restore=True)
        recover_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800)

    @mock.patch('valence.vl_helpers.recover_failed_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.wait_for_bin_assignments')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.vl_helpers.get_next_event_id')
    @mock.patch('valence.vl_helpers.check_fake_drives_flag_on_node')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.get_node_fms')
    def test_apply_with_failure_interval(self, get_node_fms, list_drives, check_fake_drives,
                                         get_next_event, fail_drives, wait_for_bin_assignments,
                                         restore_drives, recover_drives, mock_sleep):
        get_node_fms.return_value = self.node_fms
        get_next_event.return_value = 10
        list_drives.return_value = self.drive_dms
        fail_drives_cmt = vl_components.CascadeFailDrives(
            cluster_dm=self.cluster_dm,
            drive_selector=self.drive_selector,
            failure_interval=2
        )
        with pytest.raises(components.ComponentStateError):
            drive_dms = fail_drives_cmt.drive_dms

        fail_drives_cmt.apply()
        assert check_fake_drives.call_count == 2
        check_fake_drives.assert_any_call(self.node_fms[0])
        check_fake_drives.assert_any_call(self.node_fms[1])

        assert fail_drives.call_count == 2
        fail_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=self.drive_dms, timeout=None,
                                       wait_for_failure=True)
        wait_for_bin_assignments.assert_called_with(mock.ANY, start_event_id=10)
        mock_sleep.assert_called()
        drive_dms = fail_drives_cmt.drive_dms
        assert len(drive_dms) == 4

        fail_drives_cmt.reset()
        restore_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800,
                                          wait_for_restore=True)
        recover_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800)

    @mock.patch('valence.vl_helpers.recover_failed_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.wait_for_bin_assignments')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.vl_helpers.get_next_event_id')
    @mock.patch('valence.vl_helpers.check_fake_drives_flag_on_node')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.get_node_fms')
    def test_apply_no_wait_for_bin_assignments(self, get_node_fms, list_drives, check_fake_drives,
                                               get_next_event, fail_drives, wait_for_bin_assignments,
                                               restore_drives, recover_drives, mock_sleep):
        get_node_fms.return_value = self.node_fms
        get_next_event.return_value = 10
        list_drives.return_value = self.drive_dms
        fail_drives_cmt = vl_components.CascadeFailDrives(
            cluster_dm=self.cluster_dm,
            drive_selector=self.drive_selector,
            wait_for_bin_assignments=False
        )
        with pytest.raises(components.ComponentStateError):
            drive_dms = fail_drives_cmt.drive_dms

        fail_drives_cmt.apply()
        check_fake_drives.assert_any_call(self.node_fms[0])
        check_fake_drives.assert_any_call(self.node_fms[1])

        assert fail_drives.call_count == 2
        fail_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=self.drive_dms, timeout=None,
                                       wait_for_failure=True)
        wait_for_bin_assignments.assert_not_called()
        mock_sleep.assert_not_called()
        drive_dms = fail_drives_cmt.drive_dms
        assert len(drive_dms) == 4

        fail_drives_cmt.reset()
        restore_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800,
                                          wait_for_restore=True)
        recover_drives.assert_called_with(cluster_fm=mock.ANY, drive_dms=drive_dms, timeout=1800)


class TestAddDrivesComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.available_drive_dms = models.Drive.get_mock_object_list(3, overrides={'status': 'available'})
        cls.selector = selectors.FixedDriveSelector(cls.available_drive_dms)

    def setUp(self):
        self.add_drives_cmt = vl_components.AddDrives(self.cluster_dm, self.selector)

    @mock.patch('valence.functional_models.Cluster.add_drives')
    @mock.patch.object(vl_components.AddDrives, '_reload')
    def test_apply_async(self, mock_reload, mock_add_drives):
        mock_reload.return_value = False
        async_handler_mock = mock.MagicMock()
        async_handler_mock.is_running.side_effect = [True, False]
        mock_add_drives.return_value = async_handler_mock
        self.add_drives_cmt.apply_async()
        mock_add_drives.assert_called()

        assert self.add_drives_cmt.is_running() is True
        assert self.add_drives_cmt.is_running() is False
        self.add_drives_cmt.finish()
        async_handler_mock.get_models.assert_called()

    @mock.patch('valence.functional_models.Cluster.add_drives')
    @mock.patch.object(vl_components.AddDrives, '_reload')
    @mock.patch('valence.functional_models.Cluster.remove_drives')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reset(self, mock_list_drives, mock_remove_drives, mock_reload, mock_add_drives):
        mock_reload.return_value = False
        mock_list_drives.return_value = self.available_drive_dms
        self.add_drives_cmt._apply_is_running = mock.Mock(return_value=False)
        self.add_drives_cmt.apply()
        self.add_drives_cmt.reset()
        mock_reload.assert_called()
        mock_add_drives.assert_called()
        mock_remove_drives.assert_called()

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload_success(self, mock_list_drives):
        mock_list_drives.return_value = self.available_drive_dms
        assert self.add_drives_cmt._reload() is True

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload_fail(self, mock_list_drives):
        mock_list_drives.return_value = []  # if available, it would return nothing.
        assert self.add_drives_cmt._reload() is False

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reload_error(self, mock_list_drives):
        mock_list_drives.return_value = [1]  # doesn't really care if it's a drive or not.
        with pytest.raises(components.PartialStateError):
            self.add_drives_cmt._reload()

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_reset_error(self, mock_list_drives):
        mock_list_drives.side_effect = [[1, 2, 3], [1]]
        self.add_drives_cmt._apply_is_running = mock.Mock(return_value=False)
        self.add_drives_cmt.apply()
        with pytest.raises(components.PartialStateError):
            self.add_drives_cmt.reset()


class TestCascadeAddDrivesComponent(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        self.drive_dms = models.Drive.get_mock_object_list(
            2, overrides=dict(status='active', type='block'))
        self.node_fm = functional_models.Node.get_mock_fm()
        self.node_dm = models.ActiveNode.get_mock_object()

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(vl_helpers, 'wait_for_bin_assignments')
    @mock.patch.object(vl_components.AddDrives, 'apply')
    @mock.patch.object(vl_helpers, 'get_next_event_id')
    def test_apply_with_drive_selector(self, mock_next_event, mock_fail_drives_apply,
                                       mock_wait_for_assignments, mock_get_node_fms, mock_list_drives):
        mock_next_event.return_value = 10
        cascade_add_with_selector_cmt = vl_components.CascadeAddDrives(
            self.cluster_dm,
            drive_selector=selectors.FixedDriveSelector([self.drive_dms[0],
                                                         self.drive_dms[1]])
        )

        mock_get_node_fms.return_value = [self.node_fm]
        mock_list_drives.return_value = self.drive_dms
        cascade_add_with_selector_cmt.apply()
        assert mock_fail_drives_apply.call_count == 2
        assert len(cascade_add_with_selector_cmt.drive_dms) == 2
        mock_wait_for_assignments.assert_called()

    @mock.patch.object(vl_components.AddDrives, 'ensure_ready')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    def test_reset(self, mock_wait_for_sync, mock_ensure_ready):
        cascade_add_with_selector_cmt = vl_components.CascadeAddDrives(
            self.cluster_dm,
            drive_selector=selectors.FixedDriveSelector(self.drive_dms)
        )

        cascade_add_with_selector_cmt._state = components.States.applied
        cascade_add_with_selector_cmt._add_drive_cmts = [
            vl_components.AddDrives(self.cluster_dm, selectors.FixedDriveSelector(self.drive_dms)),
            vl_components.AddDrives(self.cluster_dm, selectors.FixedDriveSelector(self.drive_dms))
        ]
        cascade_add_with_selector_cmt.reset()

        mock_wait_for_sync.assert_called()
        mock_ensure_ready.assert_called()


class TestEnableMdssComponent(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()

        self.node_dm = models.ActiveNode.get_mock_object(overrides=dict(node_id=123,
                                                                        internal_role='Storage'))
        self.node_dm.platform_info.node_type = sf_models.NodeType.STANDARD_TYPES.SF9010
        self.enable_mdss_cmt = vl_components.EnableMdss(self.cluster_dm)

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch.object(vl_helpers._drive, 'get_active_node_dm_from_drive_dm')
    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_apply_setup(self, mock_active_nodes, mock_list_drives, mock_get_default_slice_drive_slot,
                         mock_get_node_dm, mock_get_constants):

        mock_constant_value = mock.Mock()
        mock_constant_value.allow_drive_type_modification_of_slot_num = 2
        mock_constant_value.bin_sync_write_mib_per_second = 20
        mock_constant_value.bin_sync_new_service_write_mib_per_second = 60
        mock_get_constants.return_value = mock_constant_value

        mock_active_nodes.return_value = [self.node_dm]
        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=0, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='volume'))
        drive_3 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=3, type='block'))
        drive_dms = [drive_0, drive_2, drive_3]

        mock_get_default_slice_drive_slot.return_value = 0
        mock_list_drives.return_value = drive_dms

        # Two slice drives should cause fast failure

        self.enable_mdss_cmt._apply_setup()
        self.assertEqual(self.enable_mdss_cmt._primary_slice_drives_dms, [drive_0])
        self.assertEqual(self.enable_mdss_cmt._slot_two_drives_dms, [drive_2])
        self.assertEqual(self.enable_mdss_cmt._slot_two_mdss_drives_dms, [drive_2])
        self.assertEqual(self.enable_mdss_cmt._active_nodes_dms, [self.node_dm])

    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(functional_models.Cluster, 'set_constants')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers, 'add_drives')
    @mock.patch.object(vl_helpers, 'set_drives_type')
    @mock.patch.object(vl_helpers, 'remove_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_apply(self, mock_active_nodes, mock_remove_drives, mock_set_drives, mock_add_drives,
                   mock_list_drives, mock_wait_bin_sync, mock_get_constants, mock_set_constants,
                   mock_get_default_slice_drive_slot):
        mock_active_nodes.return_value = [self.node_dm]
        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=0, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='block'))
        drive_2_mod = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='volume'))

        mock_get_default_slice_drive_slot.return_value = 0

        mock_list_drives.side_effect = [[drive_0], [drive_2]]
        mock_set_drives.return_value = [drive_2_mod]
        mock_add_drives.return_value = [drive_2_mod]
        mock_remove_drives.return_value = [drive_2_mod]

        # Two slice drives should cause fast failure
        self.enable_mdss_cmt.apply()
        self.assertEqual(self.enable_mdss_cmt._removed_drive_dms, [drive_2_mod])

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.Cluster, 'set_constants')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(vl_helpers, 'wait_for_slice_sync_to_start')
    @mock.patch.object(vl_helpers, 'wait_for_slice_sync_to_finish')
    @mock.patch.object(vl_helpers, 'add_drives')
    @mock.patch.object(vl_helpers, 'set_drives_type')
    @mock.patch.object(vl_helpers, 'remove_drives')
    def test_reset(self,  mock_remove_drives, mock_set_drives, mock_add_drives, mock_slice_sync_finish,
                   mock_slice_sync_start, mock_bin_sync_start, mock_get_constants, mock_set_cosntants,
                   mock_sleep, mock_active_nodes, mock_list_drives):
        mock_active_nodes.return_value = [self.node_dm]
        self.enable_mdss_cmt._state = components._enums.States.applied

        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=0, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='block'))
        drive_2_mod = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='volume'))
        mock_list_drives.side_effect = [[drive_0], [drive_2]]

        self.enable_mdss_cmt._node_mdss_drive_dms[self.node_dm.static_id] = drive_2_mod
        mock_set_drives.return_value = [drive_2]
        mock_slice_sync_finish.side_effect = [vl_helpers.HelperError]

        # Two slice drives should cause fast failure
        self.enable_mdss_cmt._apply_setup()
        self.enable_mdss_cmt.reset()

        mock_remove_drives.assert_called_with(self.enable_mdss_cmt._cluster_fm,
                                              drive_dms=[drive_0, drive_2_mod],
                                              timeout=1)

        assert mock_add_drives.call_count == 2

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_reload_success(self, mock_active_nodes, mock_list_drives,
                            mock_get_default_slice_drive_slot, mock_get_constants):
        mock_constant_value = mock.Mock()
        mock_constant_value.allow_drive_type_modification_of_slot_num = 2
        mock_constant_value.bin_sync_write_mib_per_second = 20
        mock_constant_value.bin_sync_new_service_write_mib_per_second = 60
        mock_constant_value.bin_sync_write_mib_per_second = 20
        mock_constant_value.bin_sync_new_service_write_mib_per_second = 60
        mock_get_constants.return_value = mock_constant_value

        mock_active_nodes.return_value = [self.node_dm]
        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=0, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='volume'))
        drive_dms = [drive_0, drive_2]
        mock_get_default_slice_drive_slot.return_value = 0
        mock_list_drives.return_value = drive_dms
        self.enable_mdss_cmt._apply_setup()
        assert self.enable_mdss_cmt._reload() is True

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_reload_fail(self, mock_active_nodes, mock_list_drives, mock_get_default_slice_drive_slot,
                         mock_get_constants):
        mock_constant_value = mock.Mock()
        mock_constant_value.allow_drive_type_modification_of_slot_num = 2
        mock_constant_value.bin_sync_write_mib_per_second = 20
        mock_constant_value.bin_sync_new_service_write_mib_per_second = 60
        mock_get_constants.return_value = mock_constant_value

        mock_active_nodes.return_value = [self.node_dm, self.node_dm]
        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=0, type='volume'))
        drive_1 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=123, slot=2, type='block'))
        drive_dms = [drive_0, drive_1, drive_2]
        mock_get_default_slice_drive_slot.return_value = 2
        mock_list_drives.return_value = drive_dms
        self.enable_mdss_cmt._apply_setup()
        with pytest.raises(components._errors.PartialStateError):
            self.enable_mdss_cmt._reload()


@mock.patch('time.sleep', mock.MagicMock())
class TestFailSliceDriveComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.drive_dms = [
            models.Drive.create(drive_id=0, type=enums.DriveType.slice, status=enums.DriveStatus.failed),
            models.Drive.create(drive_id=3, type=enums.DriveType.slice, status=enums.DriveStatus.active),
            models.Drive.create(drive_id=7, type=enums.DriveType.slice,
                                status=enums.DriveStatus.available)
        ]
        cls.drive_selector = selectors.DriveSelector(cls.cluster_dm)

    @mock.patch('valence.vl_helpers.wait_for_sync_in')
    @mock.patch('valence.functional_models.Cluster.list_services')
    @mock.patch('valence.vl_helpers.wait_for_sync_out')
    @mock.patch('valence.vl_helpers.calculate_slice_sync_timeout')
    @mock.patch('valence.vl_helpers.wait_for_slice_sync_to_finish')
    @mock.patch('valence.vl_helpers.add_drives')
    @mock.patch('valence.vl_helpers.recover_failed_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.vl_helpers.wait_for_drive_statuses')
    def test_apply(self, wait, list_drives, fail_drives, restore_drives, recover_drives, add_drives,
                   wait_for_sync, mock_sync_timeout, mock_sync_out, mock_services, mock_sync_in):
        list_drives.return_value = self.drive_dms
        mock_services.return_value = [
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 0}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 3}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 7}}),
        ]
        fail_drives_cmt = vl_components.FailSliceDrives(self.cluster_dm, self.drive_selector, True)
        assert fail_drives_cmt.drive_dms == self.drive_dms

        list_drives.return_value = []
        fail_drives_cmt.apply(timeout=1)
        assert fail_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert fail_drives.call_args[1].get('wait_for_failure') is False
        assert fail_drives_cmt.drive_dms == self.drive_dms
        assert fail_drives_cmt._apply_is_running()
        list_drives.return_value = self.drive_dms
        fail_drives_cmt.reset(timeout=2)
        assert restore_drives.called
        assert restore_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert restore_drives.call_args[1].get('wait_for_restore') is True
        assert fail_drives_cmt.drive_dms == self.drive_dms
        list_drives.return_value = self.drive_dms[:1]
        mock_services.return_value = [
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 1}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 2}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 5}}),
        ]
        fail_drives_cmt.apply()
        assert fail_drives_cmt.drive_dms == self.drive_dms[:1]
        assert not fail_drives_cmt._apply_is_running()

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_block(self, list_drives):
        drive_dms = [models.Drive.create(drive_id=2, type=enums.DriveType.block)]
        list_drives.return_value = drive_dms
        fail_drives_cmt = vl_components.FailSliceDrives(self.cluster_dm, self.drive_selector, True)
        with pytest.raises(components.ComponentStateError):
            fail_drives_cmt.apply()

    @mock.patch('valence.functional_models.Cluster.list_services')
    def test_reload(self, mock_services):
        fail_drives_cmt = vl_components.FailSliceDrives(self.cluster_dm, self.drive_selector, False)

        mock_services.return_value = [
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 1}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 2}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 6}}),
        ]
        fail_drives_cmt._drive_dms = self.drive_dms
        assert fail_drives_cmt._reload() is True

        mock_services.return_value = [
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 0}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 3}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 7}}),
        ]
        assert fail_drives_cmt._reload() is False

        mock_services.return_value = [
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 1}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 3}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 6}}),
        ]
        with pytest.raises(components.PartialStateError):
            fail_drives_cmt._reload()

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_mock(self, list_drives):
        list_drives.return_value = self.drive_dms
        mock_fail_drives_cmt = vl_components.FailSliceDrives.mock(self.cluster_dm, self.drive_selector)
        mock_fail_drives_cmt.apply()
        mock_fail_drives_cmt.reset()
        assert mock_fail_drives_cmt.drive_dms == self.drive_dms


@mock.patch('time.sleep', mock.MagicMock())
class TestFailBlockDrives(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.drive_dms = [
            models.Drive.create(drive_id=0, type=enums.DriveType.block, status=enums.DriveStatus.failed),
            models.Drive.create(drive_id=3, type=enums.DriveType.block, status=enums.DriveStatus.active),
            models.Drive.create(drive_id=7, type=enums.DriveType.block,
                                status=enums.DriveStatus.available)
        ]
        cls.drive_selector = selectors.DriveSelector(cls.cluster_dm)

    @mock.patch('valence.vl_helpers.wait_for_sync_in')
    @mock.patch('valence.vl_helpers.wait_for_sync_out')
    @mock.patch('valence.functional_models.Cluster.list_services')
    @mock.patch('valence.vl_helpers.add_drives')
    @mock.patch('valence.vl_helpers.restore_failed_drives')
    @mock.patch('valence.vl_helpers.fail_drives_in_cluster')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_apply(self, list_drives, fail_drives, restore_drives, add_drives, mock_services,
                   mock_sync_out, mock_sync_in):
        list_drives.return_value = self.drive_dms
        mock_services.return_value = [
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 0}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 3}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'drive_id': 7}}),
        ]

        fail_drives_cmt = vl_components.FailBlockDrives(self.cluster_dm, self.drive_selector, True)
        assert fail_drives_cmt.drive_dms == self.drive_dms

        list_drives.return_value = []
        fail_drives_cmt.apply(timeout=1)
        assert fail_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert fail_drives.call_args[1].get('wait_for_failure') is False
        assert fail_drives_cmt.drive_dms == self.drive_dms
        assert fail_drives_cmt._apply_is_running()
        assert fail_drives_cmt._drive_dms
        list_drives.return_value = self.drive_dms
        fail_drives_cmt.reset(timeout=2)
        assert restore_drives.called
        assert restore_drives.call_args[1].get('drive_dms') == self.drive_dms
        assert restore_drives.call_args[1].get('wait_for_restore') is True
        assert fail_drives_cmt.drive_dms == self.drive_dms

    @mock.patch('valence.functional_models.Cluster.list_drives')
    def test_slice(self, list_drives):
        drive_dms = [models.Drive.create(drive_id=2, type=enums.DriveType.volume)]
        list_drives.return_value = drive_dms
        fail_drives_cmt = vl_components.FailBlockDrives(self.cluster_dm, self.drive_selector, True)
        with pytest.raises(components.ComponentStateError):
            fail_drives_cmt.apply()
