import unittest

import mock
import pytest

from sf_api import result_models
from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    utils,
    vl_components,
    vl_helpers,
)
from valence.vl_components import (
    _account,
    _constants,
    _volume,
)


@mock.patch(
    'valence.functional_models.Cluster.get_constants',
    mock.MagicMock(return_value=models.ClusterConstants.create(
        bs_unusable_space_percent=7
    ))
)
@mock.patch(
    'valence.functional_models.Cluster.get_immutable_values',
    mock.MagicMock(return_value=result_models.GetImmutableValues.create(
        per_block_overhead_bytes=50,
        per_block_metadata_bytes=16,
        per_page_metadata_blocks=341,
        per_page_metadata_bytes=4096
    ))
)
@mock.patch(
    'valence.functional_models.Cluster.get_limits',
    mock.MagicMock(return_value=result_models.GetLimits.create(
        volume_size_min=1000000000,
        volume_size_max=17592186044416,
        volume_count_max=4000,
        volumes_per_account_count_max=2000
    ))
)
@mock.patch(
    'valence.functional_models.Cluster.get_cluster_full_threshold',
    mock.MagicMock(return_value=result_models.GetClusterFullThreshold.create(
        sum_used_cluster_bytes=4799961490537,
        sum_total_cluster_bytes=11999903744000,
        stage2_block_threshold_bytes=2639978823680,
        stage3_block_threshold_bytes=8279933583360,
        stage4_block_threshold_bytes=8639930695680,
        stage5_block_threshold_bytes=11999903744000,
        stage2_metadata_threshold_bytes=2639978823680,
        stage3_metadata_threshold_bytes=8279933583360,
        stage4_metadatak_threshold_bytes=8639930695680,
        stage5_metadata_threshold_bytes=11999903744000,
        sum_used_metadata_cluster_bytes=8638799872,
        sum_total_metadata_cluster_bytes=539980568984
    ))
)
@mock.patch('valence.functional_models.Cluster.list_volumes', mock.MagicMock(return_value=[]))
@mock.patch('valence.vl_components._cluster.RunGarbageCollection',
            vl_components.RunGarbageCollection.mock)
@mock.patch('valence.vl_components._volume.CreateVolumesForCapacity',
            vl_components.CreateVolumesForCapacity.mock)
class TestFillCluster(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.account_dms = [models.Account.create(username='unittest-0-TAG-0', account_id=0)]

    @mock.patch.object(vl_components.FillCluster, '_meta_capacity_achieved')
    @mock.patch.object(vl_components.FillCluster, '_block_capacity_achieved')
    @mock.patch.object(vl_helpers, 'update_cluster_full_threshold')
    @mock.patch.object(vl_helpers, 'fill_volumes')
    def test_apply_block_fullness(self, fill_volumes, update_full_threshold, block_capacity_achieved,
                                  meta_capacity_achieved):
        block_capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 13)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector,
            comp_ratio=2.2,
            dedup_ratio=3.7
        )
        fill_cmt.apply()
        assert update_full_threshold.call_count == 1
        assert fill_cmt.volumes_cmt.comp_ratio == fill_cmt.comp_ratio
        assert fill_cmt.volumes_cmt.dedup_ratio == fill_cmt.dedup_ratio
        assert fill_cmt.volumes_cmt.state == components.States.applied

        assert fill_cmt._extra_volumes_cmt.comp_ratio == fill_cmt.comp_ratio
        assert fill_cmt._extra_volumes_cmt.dedup_ratio == fill_cmt.dedup_ratio
        assert fill_cmt._extra_volumes_cmt.state == components.States.applied
        assert len(fill_cmt._extra_volumes_cmt.volume_dms) == 1
        extra_volumes_cmt = fill_cmt._extra_volumes_cmt

        # Check that the "full" volumes can be filled 100%
        block_capacity = fill_cmt.volumes_cmt.block_capacity
        assert block_capacity < fill_cmt.block_data_size.bytes
        # Check that the available volume capacity exceeds the target fill level
        block_capacity += fill_cmt._extra_volumes_cmt.block_capacity
        assert block_capacity > fill_cmt.block_data_size.bytes

        assert fill_volumes.call_count == 2
        assert fill_volumes.call_args_list[0][1]['volume_dms'] == fill_cmt.volumes_cmt.volume_dms
        assert fill_volumes.call_args_list[0][1]['fill_fraction'] == 1.0
        assert fill_volumes.call_args_list[0][1]['comp_ratio'] == fill_cmt.comp_ratio
        assert fill_volumes.call_args_list[0][1]['dedup_ratio'] == fill_cmt.dedup_ratio
        assert fill_volumes.call_args_list[1][1]['volume_dms'] == fill_cmt._extra_volumes_cmt.volume_dms
        assert fill_volumes.call_args_list[1][1]['fill_fraction'] <= 1.0
        assert fill_volumes.call_args_list[1][1]['comp_ratio'] == fill_cmt.comp_ratio
        assert fill_volumes.call_args_list[1][1]['dedup_ratio'] == fill_cmt.dedup_ratio

        assert block_capacity_achieved.called
        assert not meta_capacity_achieved.called

        fill_cmt.reset()
        assert update_full_threshold.call_count == 2
        assert fill_cmt.volumes_cmt.state == components.States.ready
        assert extra_volumes_cmt.state == components.States.ready

    @mock.patch.object(vl_components.FillCluster, '_meta_capacity_achieved')
    @mock.patch.object(vl_components.FillCluster, '_block_capacity_achieved')
    @mock.patch.object(vl_helpers, 'update_cluster_full_threshold')
    @mock.patch.object(vl_helpers, 'fill_volumes')
    def test_apply_meta_fullness(self, fill_volumes, update_full_threshold, block_capacity_achieved,
                                 meta_capacity_achieved):
        block_capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 3)
        meta_capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 27)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector,
            meta_capacity_selector=meta_capacity_selector,
            comp_ratio=1.7
        )
        fill_cmt.apply()
        assert update_full_threshold.call_count == 1
        assert fill_cmt.volumes_cmt.comp_ratio == fill_cmt.comp_ratio
        assert fill_cmt.volumes_cmt.dedup_ratio == fill_cmt.dedup_ratio
        assert fill_cmt.volumes_cmt.state == components.States.applied

        assert fill_cmt._extra_volumes_cmt.comp_ratio == fill_cmt.comp_ratio
        assert fill_cmt._extra_volumes_cmt.dedup_ratio == fill_cmt.dedup_ratio
        assert fill_cmt._extra_volumes_cmt.state == components.States.applied
        assert len(fill_cmt._extra_volumes_cmt.volume_dms) == 1
        extra_volumes_cmt = fill_cmt._extra_volumes_cmt

        # Check that the "full" volumes can be filled 100%
        block_capacity = fill_cmt.volumes_cmt.block_capacity
        assert block_capacity < fill_cmt.block_data_size.bytes
        # Check that the available volume capacity exceeds the target fill level
        block_capacity += fill_cmt._extra_volumes_cmt.block_capacity
        assert block_capacity > fill_cmt.block_data_size.bytes

        meta_capacity = fill_cmt.volumes_cmt.meta_capacity
        assert meta_capacity < fill_cmt.meta_data_size.bytes
        meta_capacity += fill_cmt._extra_volumes_cmt.meta_capacity
        assert meta_capacity > fill_cmt.meta_data_size.bytes

        assert fill_volumes.call_count == 2
        assert fill_volumes.call_args_list[0][1]['volume_dms'] == fill_cmt.volumes_cmt.volume_dms
        assert fill_volumes.call_args_list[0][1]['fill_fraction'] == 1.0
        assert fill_volumes.call_args_list[0][1]['comp_ratio'] == fill_cmt.comp_ratio
        assert fill_volumes.call_args_list[0][1]['dedup_ratio'] == fill_cmt.dedup_ratio
        assert fill_volumes.call_args_list[1][1]['volume_dms'] == fill_cmt._extra_volumes_cmt.volume_dms
        assert fill_volumes.call_args_list[1][1]['fill_fraction'] <= 1.0
        assert fill_volumes.call_args_list[1][1]['comp_ratio'] == fill_cmt.comp_ratio
        assert fill_volumes.call_args_list[1][1]['dedup_ratio'] == fill_cmt.dedup_ratio

        assert block_capacity_achieved.called
        assert meta_capacity_achieved.called

        fill_cmt.reset()
        assert update_full_threshold.call_count == 2
        assert fill_cmt.volumes_cmt.state == components.States.ready
        assert extra_volumes_cmt.state == components.States.ready

    @mock.patch.object(vl_components.FillCluster, '_meta_capacity_achieved')
    @mock.patch.object(vl_components.FillCluster, '_block_capacity_achieved')
    @mock.patch.object(vl_helpers, 'update_cluster_full_threshold')
    @mock.patch.object(vl_helpers, 'fill_volumes')
    def test_apply_meta_specific_fullness(self, fill_volumes, update_full_threshold,
                                          block_capacity_achieved,
                                          meta_capacity_achieved):
        meta_capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 27)
        fill_cmt = vl_components.FillClusterMetadata(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            meta_capacity_selector=meta_capacity_selector,
            comp_ratio=1.7
        )
        fill_cmt.apply()
        assert update_full_threshold.call_count == 1
        assert fill_cmt.volumes_cmt.comp_ratio == fill_cmt.comp_ratio
        assert fill_cmt.volumes_cmt.dedup_ratio == fill_cmt.dedup_ratio
        assert fill_cmt.volumes_cmt.state == components.States.applied

        assert fill_cmt._extra_volumes_cmt.comp_ratio == fill_cmt.comp_ratio
        assert fill_cmt._extra_volumes_cmt.dedup_ratio == fill_cmt.dedup_ratio
        assert fill_cmt._extra_volumes_cmt.state == components.States.applied
        assert len(fill_cmt._extra_volumes_cmt.volume_dms) == 1
        extra_volumes_cmt = fill_cmt._extra_volumes_cmt

        # Check that the "full" volumes can be filled 100%
        block_capacity = fill_cmt.volumes_cmt.block_capacity
        assert block_capacity < fill_cmt.block_data_size.bytes
        # Check that the available volume capacity exceeds the target fill level
        block_capacity += fill_cmt._extra_volumes_cmt.block_capacity
        assert block_capacity > fill_cmt.block_data_size.bytes

        meta_capacity = fill_cmt.volumes_cmt.meta_capacity
        assert meta_capacity < fill_cmt.meta_data_size.bytes
        meta_capacity += fill_cmt._extra_volumes_cmt.meta_capacity
        assert meta_capacity > fill_cmt.meta_data_size.bytes

        assert fill_volumes.call_count == 2
        assert fill_volumes.call_args_list[0][1]['volume_dms'] == fill_cmt.volumes_cmt.volume_dms
        assert fill_volumes.call_args_list[0][1]['fill_fraction'] == 1.0
        assert fill_volumes.call_args_list[0][1]['comp_ratio'] == fill_cmt.comp_ratio
        assert fill_volumes.call_args_list[0][1]['dedup_ratio'] == fill_cmt.dedup_ratio
        assert fill_volumes.call_args_list[1][1][
                   'volume_dms'] == fill_cmt._extra_volumes_cmt.volume_dms
        assert fill_volumes.call_args_list[1][1]['fill_fraction'] <= 1.0
        assert fill_volumes.call_args_list[1][1]['comp_ratio'] == fill_cmt.comp_ratio
        assert fill_volumes.call_args_list[1][1]['dedup_ratio'] == fill_cmt.dedup_ratio

        assert block_capacity_achieved.called
        assert meta_capacity_achieved.called

        fill_cmt.reset()
        assert update_full_threshold.call_count == 2
        assert fill_cmt.volumes_cmt.state == components.States.ready
        assert extra_volumes_cmt.state == components.States.ready

    @mock.patch.object(vl_components.FillCluster, '_meta_capacity_achieved')
    @mock.patch.object(vl_components.FillCluster, '_block_capacity_achieved')
    @mock.patch.object(vl_helpers, 'update_cluster_full_threshold')
    @mock.patch.object(vl_helpers, 'fill_volumes')
    def test_apply_block_fullness_small(self, fill_volumes, update_full_threshold,
                                        block_capacity_achieved, meta_capacity_achieved):
        # When using resize-drives to speed up cluster fullness testing, we can end up with very small
        # block fills, so we need to handle that elegantly by filling just a fraction of a single volume
        block_capacity_selector = selectors.FixedBlockCapacitySelector(self.cluster_dm, 1000000000)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector
        )
        fill_cmt.apply()
        assert len(fill_cmt.volumes_cmt.volume_dms) == 1
        assert fill_cmt._extra_volumes_cmt is None
        assert fill_volumes.call_count == 1
        assert fill_volumes.call_args_list[0][1]['fill_fraction'] < 1.0

    @mock.patch.object(vl_components.FillCluster, '_meta_capacity_achieved')
    @mock.patch.object(vl_components.FillCluster, '_block_capacity_achieved')
    @mock.patch.object(vl_helpers, 'update_cluster_full_threshold')
    @mock.patch.object(vl_helpers, 'fill_volumes')
    def test_assert_valid(self, fill_volumes, update_full_threshold, block_capacity_achieved,
                          meta_capacity_achieved):
        block_capacity_selector = selectors.StageBlockCapacitySelector(self.cluster_dm, 3)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector
        )

        with pytest.raises(components.ComponentStateError):
            fill_cmt.assert_valid()

        # Validation of block capacity selector
        fill_cmt.apply()
        fill_cmt.assert_valid()
        assert block_capacity_achieved.called

        # Validation fails on non-relative selectors
        block_capacity_selector.relative = False
        with pytest.raises(ValueError):
            fill_cmt.assert_valid()

        block_capacity_selector.relative = True
        fill_cmt.meta_capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 20)
        with pytest.raises(ValueError):
            fill_cmt.assert_valid()

        # Validation of metadata capacity selector
        fill_cmt.meta_capacity_selector.relative = True
        fill_cmt.assert_valid()
        assert meta_capacity_achieved.called


class TestCapacityValidation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.account_dms = [models.Account.create(username='unittest-0-TAG-0', account_id=0)]

    def setUp(self):
        self.cluster_full_threshold_dm = result_models.GetClusterFullThreshold.create(
            sum_used_cluster_bytes=0,
            sum_total_cluster_bytes=1000,
            stage2_block_threshold_bytes=200,
            stage3_block_threshold_bytes=700,
            stage4_block_threshold_bytes=800,
            stage5_block_threshold_bytes=900,
            stage2_metadata_threshold_bytes=200,
            stage3_metadata_threshold_bytes=700,
            stage4_metadata_threshold_bytes=800,
            stage5_metadata_threshold_bytes=900000,
            sum_used_metadata_cluster_bytes=171000,
            sum_total_metadata_cluster_bytes=181000
        )

    @mock.patch.object(functional_models.Cluster, 'get_cluster_full_threshold')
    def test_block_capacity_achieved(self, get_cluster_full_threshold):
        get_cluster_full_threshold.return_value = self.cluster_full_threshold_dm
        block_capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 20)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector,
            comp_ratio=2.2,
            dedup_ratio=3.7
        )
        fill_cmt._block_data_size = block_capacity_selector()

        # No validation performed on non-relative selectors
        self.cluster_full_threshold_dm.sum_used_cluster_bytes = 189
        assert fill_cmt._block_capacity_achieved()

        block_capacity_selector.relative = True
        assert not fill_cmt._block_capacity_achieved()

        # Validation passes if within tolerance
        assert fill_cmt.BLOCK_PERCENT_TOLERANCE == 1
        self.cluster_full_threshold_dm.sum_used_cluster_bytes = 190
        assert fill_cmt._block_capacity_achieved()

        fill_cmt.BLOCK_PERCENT_TOLERANCE = 0.1
        assert not fill_cmt._block_capacity_achieved()

        self.cluster_full_threshold_dm.sum_used_cluster_bytes = 200
        assert fill_cmt._block_capacity_achieved()

    @mock.patch.object(vl_helpers, 'get_block_metadata_size')
    @mock.patch.object(functional_models.Cluster, 'get_cluster_full_threshold')
    def test_meta_capacity_achieved(self, get_cluster_full_threshold, mock_block_size):
        mock_block_size.return_value = models.DataSize(1, enums.DataUnit.tebibytes)
        get_cluster_full_threshold.return_value = self.cluster_full_threshold_dm
        block_capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 20)
        meta_capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 20)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector,
            meta_capacity_selector=meta_capacity_selector,
            comp_ratio=2.2
        )
        fill_cmt._block_data_size = block_capacity_selector()
        fill_cmt._meta_data_size = meta_capacity_selector()

        # No validation performed on non-relative selectors
        self.cluster_full_threshold_dm.sum_used_metadata_cluster_bytes = 188090
        assert fill_cmt._meta_capacity_achieved()

        meta_capacity_selector.relative = True
        assert not fill_cmt._meta_capacity_achieved()

        # Validation passes if within tolerance
        assert fill_cmt.META_PERCENT_TOLERANCE == 1
        self.cluster_full_threshold_dm.sum_used_metadata_cluster_bytes = 34395
        assert fill_cmt._meta_capacity_achieved()

        fill_cmt.META_PERCENT_TOLERANCE = 0.1
        assert not fill_cmt._meta_capacity_achieved()

        self.cluster_full_threshold_dm.sum_used_metadata_cluster_bytes = 36020
        assert fill_cmt._meta_capacity_achieved()
        print(self.cluster_full_threshold_dm.serialize())
        assert fill_cmt.volumes_cmt.DEFAULT_VOL_COUNT == 60

    @mock.patch.object(vl_helpers, 'get_block_metadata_size')
    @mock.patch.object(functional_models.Cluster, 'get_cluster_full_threshold')
    def test_meta_only_fill(self, get_cluster_full_threshold, mock_block_size):
        mock_block_size.return_value = models.DataSize(1, enums.DataUnit.kilobytes)
        get_cluster_full_threshold.return_value = self.cluster_full_threshold_dm
        block_capacity_selector = selectors.PercentBlockCapacitySelector(self.cluster_dm, 1)
        meta_capacity_selector = selectors.PercentMetaCapacitySelector(self.cluster_dm, 100)
        fill_cmt = vl_components.FillCluster(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            block_capacity_selector=block_capacity_selector,
            meta_capacity_selector=meta_capacity_selector,
            comp_ratio=1
        )
        fill_cmt._block_data_size = block_capacity_selector()
        fill_cmt._meta_data_size = meta_capacity_selector()
        assert fill_cmt.volumes_cmt.DEFAULT_VOL_COUNT == 1


@mock.patch('time.sleep', mock.MagicMock())
class TestSetupRecyclingComponent(unittest.TestCase):
    TAG = 'abc'

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.first_time_write_ratio_cmt = vl_components.SetClusterConstants(
            cluster_dm=cls.cluster_dm,
            constants_dm=models.ClusterConstants.create(bs_first_time_write_ratio=0.5)
        )

    def setUp(self):
        self.sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 30, min_num_volumes=20)

    def test_validate_inputs_bad_comp_ratio(self):
        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 10, comp_ratio=0.5)

        with pytest.raises(ValueError):
            sr_cmt._validate_inputs()

    def test_validate_inputs_bad_fullness(self):
        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 15)

        with pytest.raises(ValueError):
            sr_cmt._validate_inputs()

    def test_validate_inputs_vol_size_and_count(self):
        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 10,
                                              min_num_volumes=10,
                                              min_volume_size=models.DataSize(1000))

        with pytest.raises(ValueError):
            sr_cmt._validate_inputs()

    def test_validate_inputs_vol_size_too_large(self):
        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 10,
                                              min_volume_size=models.DataSize(10001))
        sr_cmt._num_cycles = 10
        sr_cmt._max_used_space = 100000

        with pytest.raises(ValueError):
            sr_cmt._validate_inputs()

    def test_validate_inputs_vol_size_too_small(self):
        limits = result_models.GetLimits.get_mock_object(
            overrides=dict(volume_count_max=101, volume_size_min=1001))

        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 10,
                                              min_volume_size=models.DataSize(1000))
        sr_cmt._num_cycles = 10
        sr_cmt._max_used_space = 100000
        sr_cmt._cluster_limits = limits

        with pytest.raises(ValueError):
            sr_cmt._validate_inputs()

    def test_validate_inputs_too_many_vols(self):
        limits = result_models.GetLimits.get_mock_object(overrides=dict(volume_count_max=99))

        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 10,
                                              min_volume_size=models.DataSize(1000))
        sr_cmt._num_cycles = 10
        sr_cmt._max_used_space = 100000
        sr_cmt._cluster_limits = limits

        with pytest.raises(ValueError):
            sr_cmt._validate_inputs()

    @mock.patch.object(vl_helpers, 'get_block_overhead_factor')
    @mock.patch.object(selectors.PercentBlockCapacitySelector, '__call__')
    def test_calculate_apply_parameters_vol_path(self, mock_cap_selector_call, mock_get_overhead):
        mock_cap_selector_call.return_value = 100000
        mock_get_overhead.return_value = 1

        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 10, min_num_volumes=20)
        sr_cmt._max_used_space = 100000
        sr_cmt._rep_count = 2
        sr_cmt._block_overhead_factor = 1.0

        sr_cmt._calculate_apply_parameters()

        assert sr_cmt._num_cycles == 20
        assert sr_cmt._keepers_per_cycle == 1

    @mock.patch.object(vl_helpers, 'get_block_overhead_factor')
    @mock.patch.object(selectors.PercentBlockCapacitySelector, '__call__')
    def test_calculate_apply_parameters(self, mock_cap_selector_call, mock_get_overhead):
        mock_cap_selector_call.return_value = 100000
        mock_get_overhead.return_value = 1

        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm, 30)
        sr_cmt._max_used_space = 100000
        sr_cmt._rep_count = 2

        sr_cmt._calculate_apply_parameters()

        assert sr_cmt._num_cycles == 10
        assert sr_cmt._keepers_per_cycle == 3

    @mock.patch.object(_account.CreateAccounts, '_reload')
    @mock.patch.object(_account.CreateAccounts, '_apply')
    @mock.patch.object(_volume.CreateVolumes, '_reload')
    @mock.patch.object(_volume.CreateVolumes, '_apply')
    def test_create_components(self, mock_create_volumes, mock_reload_volumes,
                               mock_create_accounts, mock_reload_accounts):
        self.sr_cmt._volume_count = 200
        self.sr_cmt._volume_size = 10000
        self.sr_cmt._create_components()

        assert len(self.sr_cmt._volume_cmts) == 200

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_cluster_constants(self, mock_get_constants):
        mock_get_constants.return_value = models.ClusterConstants.get_mock_object(
                overrides=dict(bs_first_time_write_ratio=0.5)
            )
        assert self.sr_cmt.cluster_constants.bs_first_time_write_ratio == 0.5

    @mock.patch.object(functional_models.Cluster, 'get_limits')
    def test_cluster_limits(self, mock_get_limits):
        mock_get_limits.return_value = models.ClusterLimits.get_mock_object()
        assert isinstance(self.sr_cmt.cluster_limits, models.ClusterLimits)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_capacity')
    def test_max_used_space(self, mock_get_capacity):
        mock_get_capacity.return_value = models.ClusterCapacity.get_mock_object(
            overrides=dict(max_used_space=10000)
        )
        assert self.sr_cmt.max_used_space == 10000

    def test_split_volume_cmts_subsets(self):
        self.sr_cmt._volume_cmts = [object()] * 100
        self.sr_cmt._num_cycles = 10

        self.sr_cmt._split_volumes_cmts_subsets()

        assert len(self.sr_cmt._volume_cmts_subsets) == 10
        assert all(len(subset) == 10 for subset in self.sr_cmt._volume_cmts_subsets)

    @mock.patch.object(vl_components.SetupRecycling, '_validate_inputs')
    @mock.patch.object(vl_components.SetupRecycling, '_calculate_apply_parameters')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_apply_setup(self, mock_list_volumes, mock_calc_parameters,
                         mock_validate_inputs):
        mock_list_volumes.return_value = []
        self.sr_cmt._cluster_constants = models.ClusterConstants.get_mock_object(
            overrides=dict(bs_first_time_write_ratio=0.7)
        )
        self.sr_cmt._apply_setup()
        mock_calc_parameters.assert_called_once()
        mock_validate_inputs.assert_called_once()

    @mock.patch.object(vl_components.SetupRecycling, '_validate_inputs')
    @mock.patch.object(vl_components.SetupRecycling, '_calculate_apply_parameters')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_apply_setup_existing_data(self, mock_list_volumes, mock_calc_parameters,
                                       mock_validate_inputs):
        mock_list_volumes.return_value = models.Volume.get_mock_object_list(2)

        self.sr_cmt._cluster_constants = models.ClusterConstants.get_mock_object(
            overrides=dict(bs_first_time_write_ratio=None)
        )

        with pytest.raises(components.ComponentStateError):
            self.sr_cmt._apply_setup()

        mock_validate_inputs.assert_called_once()

    @mock.patch.object(vl_helpers, 'fill_volumes')
    def test_fill_purge_cycle(self, mock_fill_volumes):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(10):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))
            mock_vol_cmts[i]._state = components.States.applied

        self.sr_cmt._volume_cmts = list(mock_vol_cmts)
        self.sr_cmt._keepers_per_cycle = 1

        self.sr_cmt._fill_purge_cycle(mock_vol_cmts[:2])

        mock_fill_volumes.assert_called_with(self.sr_cmt._cluster_fm, mock_volumes[:2],
                                             comp_ratio=1.0)
        assert mock_vol_cmts[0] not in self.sr_cmt._volume_cmts

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'get_cluster_capacity')
    @mock.patch.object(vl_components.RunGarbageCollection, 'reset')
    @mock.patch.object(vl_components.RunGarbageCollection, 'apply')
    def test_gc_cleanup_loop(self, mock_gc_apply, mock_gc_reset, mock_cluster_capacity, mock_timer):
        mock_timer.return_value = iter([1, 2, 3, 4, 5])

        max_used_space = 1000
        fullness_percent = 10
        mock_capacities = models.ClusterCapacity.get_mock_object_list(
            2, overrides=dict(max_used_space=max_used_space)
        )

        tgt_fullness = int(fullness_percent * max_used_space // 100)
        fullness_delta = tgt_fullness * vl_components.SetupRecycling.USED_BLOCKSPACE_TOLERANCE + 1
        mock_capacities[0].active_block_space = tgt_fullness + fullness_delta
        mock_capacities[1].active_block_space = tgt_fullness - fullness_delta
        mock_cluster_capacity.side_effect = mock_capacities

        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm,
                                              fullness_percent, min_num_volumes=20)
        sr_cmt._max_used_space = max_used_space
        sr_cmt._gc_cleanup_loop()
        assert mock_gc_apply.call_count == 2

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'get_cluster_capacity')
    @mock.patch.object(vl_components.RunGarbageCollection, 'reset')
    @mock.patch.object(vl_components.RunGarbageCollection, 'apply')
    def test_gc_cleanup_loop_timeout(self, mock_gc_apply, mock_gc_reset,
                                     mock_cluster_capacity, mock_timer):
        mock_timer.return_value = iter([1])

        max_used_space = 1000
        fullness_percent = 10
        mock_capacities = models.ClusterCapacity.get_mock_object_list(
            2, overrides=dict(max_used_space=max_used_space)
        )
        tgt_fullness = int(fullness_percent * max_used_space)
        fullness_delta = tgt_fullness * vl_components.SetupRecycling.USED_BLOCKSPACE_TOLERANCE + 1
        mock_capacities[0].active_block_space = tgt_fullness + fullness_delta
        mock_capacities[1].active_block_space = tgt_fullness - fullness_delta
        mock_cluster_capacity.side_effect = mock_capacities

        sr_cmt = vl_components.SetupRecycling(self.TAG, self.cluster_dm,
                                              fullness_percent, min_num_volumes=20)
        sr_cmt._max_used_space = max_used_space
        with pytest.raises(StopIteration):
            sr_cmt._gc_cleanup_loop()

    @mock.patch.object(vl_components.RunGarbageCollection, 'reset')
    @mock.patch.object(vl_components.RunGarbageCollection, 'apply')
    @mock.patch.object(_constants.SetClusterConstants, 'apply')
    @mock.patch.object(_constants.SetClusterConstants, 'reset')
    @mock.patch.object(vl_components.SetupRecycling, '_fill_purge_cycle')
    @mock.patch.object(vl_components.SetupRecycling, '_create_components')
    @mock.patch.object(vl_components.SetupRecycling, '_split_volumes_cmts_subsets')
    @mock.patch.object(vl_components.SetupRecycling, '_gc_cleanup_loop')
    def test_apply(self, mock_cleanup_loop, mock_split_cmts, mock_create_components, mock_fill_purge,
                   mock_reset_constants, mock_apply_constants, mock_gc_apply, mock_gc_reset):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))
            mock_vol_cmts[i]._state = components.States.applied

        self.sr_cmt._volume_cmts = list(mock_vol_cmts)
        self.sr_cmt._volume_cmts_subsets = [[mock_vol_cmts[0]], [mock_vol_cmts[1]]]

        self.sr_cmt._apply()

        assert mock_fill_purge.call_count == 2
        assert mock_gc_apply.call_count == 1

    @mock.patch.object(vl_components.RunGarbageCollection, 'reset')
    @mock.patch.object(vl_components.RunGarbageCollection, 'apply')
    @mock.patch.object(_constants.SetClusterConstants, 'apply')
    @mock.patch.object(_constants.SetClusterConstants, 'reset')
    @mock.patch.object(vl_components.SetupRecycling, '_fill_purge_cycle')
    @mock.patch.object(vl_components.SetupRecycling, '_create_components')
    @mock.patch.object(vl_components.SetupRecycling, '_split_volumes_cmts_subsets')
    @mock.patch.object(vl_components.SetupRecycling, '_gc_cleanup_loop')
    def test_apply_with_first_time_write_constant(self, mock_cleanup_loop,
                                                  mock_split_cmts,
                                                  mock_create_components,
                                                  mock_fill_purge,
                                                  mock_reset_constants,
                                                  mock_apply_constants,
                                                  mock_gc_apply,
                                                  mock_gc_reset):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))
            mock_vol_cmts[i]._state = components.States.applied

        self.sr_cmt._volume_cmts = list(mock_vol_cmts)
        self.sr_cmt._volume_cmts_subsets = [[mock_vol_cmts[0]], [mock_vol_cmts[1]]]
        self.sr_cmt._first_time_write_ratio_cmt = self.first_time_write_ratio_cmt

        self.sr_cmt._apply()

        assert mock_fill_purge.call_count == 2
        assert mock_gc_apply.call_count == 0

    @mock.patch.object(vl_components.RunGarbageCollection, 'ensure_ready')
    @mock.patch.object(vl_components.RunGarbageCollection, 'apply')
    def test_reset(self, mock_gc_apply, mock_const_ensure_ready):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_account_cmt._state = components.States.applied
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))
            mock_vol_cmts[i]._state = components.States.applied

        self.sr_cmt._volume_cmts = list(mock_vol_cmts)
        self.sr_cmt._account_cmt = mock_account_cmt
        self.sr_cmt._first_time_write_ratio_cmt = self.first_time_write_ratio_cmt
        self.sr_cmt._reset()

        assert all(cmt.state is components.States.ready for cmt in mock_vol_cmts)
        assert mock_account_cmt.state.ready is components.States.ready

    def test_child_cmts(self):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))

        self.sr_cmt._volume_cmts = mock_vol_cmts
        assert self.sr_cmt.child_cmts == mock_vol_cmts

    def test_volume_dms(self):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))
            mock_vol_cmts[i]._state = components.States.applied

        self.sr_cmt._volume_cmts = mock_vol_cmts
        assert self.sr_cmt.volume_dms == mock_volumes[:2]

    def test_get_volume_dms_no_apply(self):
        with pytest.raises(components.ComponentStateError):
            assert self.sr_cmt.get_volume_dms()

    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_get_volume_dms(self, mock_list_vols):
        mock_volumes = models.Volume.get_mock_object_list(2)
        for i, mock_vol in enumerate(mock_volumes):
            mock_vol.name = self.TAG + str(i)
        mock_list_vols.return_value = mock_volumes
        self.sr_cmt._state = components.States.applied

        assert self.sr_cmt.get_volume_dms() == mock_volumes

    def test_partial_volume_dms(self):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(10):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))
            mock_vol_cmts[i]._state = components.States.applied

        self.sr_cmt._volume_cmts = mock_vol_cmts
        self.sr_cmt._partial_volume_dm = mock_volumes[0]
        assert self.sr_cmt.partial_volume_dms == [mock_volumes[0]]*10

    def test_assert_valid(self):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account_cmt = _account.MockAccounts(self.TAG, models.Account.get_mock_object())
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))

        self.sr_cmt._volume_cmts = mock_vol_cmts
        self.sr_cmt._account_cmt = mock_account_cmt
        self.sr_cmt.assert_valid()

    def test_account_dms(self):
        mock_volumes = models.Volume.get_mock_object_list(10)
        mock_account = models.Account.get_mock_object()
        mock_account_cmt = _account.MockAccounts(self.TAG, mock_account)
        mock_account_cmt._state = components.States.applied
        mock_vol_cmts = []

        for i in range(2):
            mock_vol_cmts.append(_volume.MockVolumes(self.TAG, self.cluster_dm,
                                                     [mock_volumes[i]], mock_account_cmt))

        self.sr_cmt._volume_cmts = mock_vol_cmts
        self.sr_cmt._account_cmt = mock_account_cmt
        assert self.sr_cmt.account_dms == self.sr_cmt.get_account_dms() == mock_account
