import os
import random
import unittest

import mock
import pytest
import sf_ssh

from valence import (
    components,
    models,
    unittest_utils,
    vl_components,
)
from valence.utils import random_string
from valence.vl_components import _fill_partition

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')


def tearDownModule():
    module_mock.restore()


class TestFillPartition(unittest.TestCase):

    RUN_DF_RETURN = \
        '        Filesystem             1B-blocks        Used   Available Use% Mounted on\n' + \
        '/dev/disk/by-uuid/blah 84418347008 15117012992 64989433856  19% /\n'
    F_ALLOCATE_RETURN = 'fallocate output'
    LS_RETURN = '-rw-r--r-- 1 root root 61977816562 Oct 26 17:41 blob.bin'
    LS_RETURN_WILL_LEAVE = '-rw-r--r-- 1 root root 59364174260 Oct 26 17:41 blob.bin'
    POST_RUN_DF_RETURN = \
        '        Filesystem             1B-blocks        Used   Available Use% Mounted on\n' + \
        '/dev/disk/by-uuid/blah 84418347008 15117012992 12000000000  19% /\n'

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.nodes = models.ActiveNode.get_mock_object_list(4)
        # Mocked strings for simulating returns form ssh commands

    def setUp(self):
        self.node_selector = lambda: self.nodes
        self.fill_partition_cmp = _fill_partition.FillPartition(
            self.cluster_dm,
            self.node_selector,
            '/',
            max_free_space_remaining=models.DataSize(12, models.DataUnit.gigabytes),
            fill_data=None
        )
        # Setting the filled nodes to be the same as the nodes in the cluster makes the test
        # behavior determined.
        self.fill_partition_cmp._filled_nodes = self.nodes
        # Problems have been seen with the names of mock nodes.  Simplify by renaming.
        names = ['a', 'b', 'c', 'd']
        for index, node in enumerate(self.fill_partition_cmp._filled_nodes):
            node.name = names[index]

    def test_invalid_constructor_parameter_combinations_both_are_set(self):
        # Both parameters for fill amount are specified
        with pytest.raises(ValueError):
            vl_components.FillPartition(
                None,
                None,
                '/',
                max_free_space_remaining=models.DataSize(12, models.DataUnit.gigabytes),
                fill_data=models.DataSize(12, models.DataUnit.gigabytes))

    def test_invalid_constructor_parameter_combinations_both_are_None(self):
        # Neither parameter for fill amount specified
        with pytest.raises(ValueError):
            vl_components.FillPartition(
                None,
                None,
                '/',
                max_free_space_remaining=None,
                fill_data=None)

    def test_fill_size_set_value(self):
        # Sunny day test for the helper method named _fill_size
        available_size_b = 10001
        self.fill_partition_cmp.fill_data = available_size_b
        assert self.fill_partition_cmp._get_fsize(available_size_b) == available_size_b

    def test_fill_size_compute_value(self):
        # Sunny day test for computed fill size
        self.fill_partition_cmp.fill_data = None
        available_size_b = 10001
        with mock.patch('random.randint', return_value=42) as mock_random_return:
            mock_random_return.return_value = 42
            mocked_fsize = self.fill_partition_cmp._get_fsize(available_size_b)
            assert mocked_fsize == 9959

    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    def test_run_df(self, mock_run_command):
        # Sunny day test for the _run_df method
        expected_value = self.RUN_DF_RETURN
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout=expected_value, stderr='', return_code=0)
        conn = sf_ssh.ssh.Ssh('12.34.56.78')
        ret = self.fill_partition_cmp._run_df(conn)
        assert ret == 64989433856
        mock_run_command.assert_called()
        assert mock_run_command.call_count == 1

    def test_filled_nodes_before_apply(self):
        # Verify assertion thrown when component state is ready
        with pytest.raises(components.ComponentStateError):
            self.fill_partition_cmp.filled_nodes

    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    def test_apply_setup(self, mock_selector):
        self.fill_partition_cmp.node_selector = mock_selector
        mock_selector.return_value = self.fill_partition_cmp._filled_nodes
        self.fill_partition_cmp._apply_setup()
        assert self.fill_partition_cmp._filled_nodes

    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    def test_apply_fill_data(self, mock_selector, mock_run_command):
        fill_disk_side_effects = [
            sf_ssh.ssh.RunCommandOutput(stdout=self.RUN_DF_RETURN, stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout=self.F_ALLOCATE_RETURN, stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout=self.LS_RETURN, stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout=self.POST_RUN_DF_RETURN, stderr='', return_code=0)
        ] * len(self.nodes)
        mock_run_command.side_effect = fill_disk_side_effects
        self.fill_partition_cmp.node_selector = mock_selector
        mock_selector.return_value = self.fill_partition_cmp._filled_nodes
        self.fill_partition_cmp.apply()
        mock_run_command.assert_called()
        assert mock_run_command.call_count == 4 * len(self.fill_partition_cmp.filled_nodes)

    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    def test_apply_will_leave(self, mock_selector, mock_run_command):
        # Test the "will_leave" option.  A random size is generated for the left amount.
        will_leave_side_effects = [
             sf_ssh.ssh.RunCommandOutput(stdout=self.RUN_DF_RETURN, stderr='', return_code=0),
             sf_ssh.ssh.RunCommandOutput(stdout=self.F_ALLOCATE_RETURN, stderr='', return_code=0),
             sf_ssh.ssh.RunCommandOutput(stdout=self.LS_RETURN_WILL_LEAVE, stderr='', return_code=0),
             sf_ssh.ssh.RunCommandOutput(stdout=self.POST_RUN_DF_RETURN, stderr='', return_code=0)
        ] * len(self.fill_partition_cmp._filled_nodes)
        mock_run_command.side_effect = will_leave_side_effects
        self.fill_partition_cmp.max_free_space_remaining = models.DataSize(12, models.DataUnit.gigabytes)
        self.fill_partition_cmp.fill_data = None
        self.fill_partition_cmp.node_selector = mock_selector
        mock_selector.return_value = self.fill_partition_cmp._filled_nodes
        with mock.patch('random.randint', return_value=123456) as mock_random_return:
            self.fill_partition_cmp.apply()
        mock_run_command.assert_called()
        assert mock_run_command.call_count == 4 * len(self.fill_partition_cmp.filled_nodes)

    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    def test_reset(self, mock_run_command):
        # Sunny day test for the _reset method
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=0)
        # Recreate the structures that were created by the component
        self.fill_partition_cmp.remote_temp_files = {}
        for node in self.fill_partition_cmp._filled_nodes:
            remote_temp = os.path.join(self.fill_partition_cmp.partition,
                                       random_string.generate_random_string(10))
            self.fill_partition_cmp.remote_temp_files[node.name] = remote_temp

        self.fill_partition_cmp._reset()

        mock_run_command.assert_called()
        # Gather evidence to help resolve an intermittent failure of this test
        if not mock_run_command.call_count == len(self.fill_partition_cmp._filled_nodes):
            # Dump the remote_temp_files dictionary
            message = 'node was: {}/n'.format(node.name)
            for h in self.fill_partition_cmp.remote_temp_files:
                message += 'remote Node: {}, {} \n'.format(h, self.fill_partition_cmp._filled_nodes[h])
            assert not mock_run_command.call_count == len(self.fill_partition_cmp._filled_nodes), message
        mock_run_command.assert_called()
        assert mock_run_command.call_count == len(self.fill_partition_cmp._filled_nodes)

    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    def test_reset_Ssh_Errors(self, mock_run_command):
        # Catch sf_ssh.errors.SshError when deleting file, then raise components.ComponentStateError
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(
            stdout='', stderr='', return_code=1)
        mock_run_command.side_effect = sf_ssh.errors.SshError()
        self.fill_partition_cmp.remote_temp_files = {}
        for node in self.fill_partition_cmp._filled_nodes:
            remote_temp = os.path.join(self.fill_partition_cmp.partition,
                                       random_string.generate_random_string(10))
            self.fill_partition_cmp.remote_temp_files[node.name] = remote_temp
        with pytest.raises(components.ComponentStateError):
            self.fill_partition_cmp._reset()
        mock_run_command.assert_called()
        assert mock_run_command.call_count == len(self.fill_partition_cmp._filled_nodes)
