from __future__ import division

import unittest

import mock
import pytest

from camelot.models import at2_enums
from valence import (
    components,
    models,
)
from valence.vl_components import _initiator


class TestCreateIscsiInitiators(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.client_dm = models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04))
        cls.initiator_dm = models.Initiator.get_mock_object()

    def setUp(self):
        self.tag = 'unittest_tag'
        self.init_cmt = _initiator.CreateIscsiInitiators(self.tag,
                                                         self.cluster_dm, [self.client_dm])
        self.initiator_dm.alias = self.init_cmt.full_tag + 'clabv0006-0'

    def test_setup_fail(self):
        with pytest.raises(ValueError):
            self.init_cmt = _initiator.CreateIscsiInitiators(self.tag,
                                                             self.cluster_dm, [])

    @mock.patch('valence.vl_components.CreateIscsiInitiators._assert_valid')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    @mock.patch('valence.functional_models.Cluster.create_initiators')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    def test_apply_and_return_values(self, mock_get_initiator_name, mock_create_initiators,
                                     mock_get_existing, mock_assert_valid):
        mock_get_initiator_name.return_value = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        mock_initiator = self.initiator_dm
        mock_get_existing.return_value = {'192.168.0.100': [mock_initiator]}
        self.init_cmt.apply()
        get_initiator_dms_map = self.init_cmt.get_initiator_dms_map()
        self.init_cmt.assert_valid()
        assert self.init_cmt.full_tag == self.tag + '-TAG-'
        assert self.init_cmt.initiator_dms_map == {'192.168.0.100': [mock_initiator]}
        assert get_initiator_dms_map == {'192.168.0.100': [mock_initiator]}

    @mock.patch('valence.vl_components.CreateIscsiInitiators._assert_valid')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    @mock.patch('valence.functional_models.Cluster.create_initiators')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    def test__apply(self, mock_get_initiator_name, mock_create_initiators,
                    mock_get_existing, mock_assert_valid):
        mock_get_initiator_name.return_value = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        mock_initiator = self.initiator_dm
        mock_get_existing.return_value = {'192.168.0.101': [mock_initiator]}
        self.init_cmt._apply()

    @mock.patch('valence.functional_models.Cluster.delete_initiators')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._assert_valid')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    @mock.patch('valence.functional_models.Cluster.create_initiators')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    def test_reset(self, mock_get_initiator_name, mock_create_initiators,
                   mock_get_existing, mock_assert_valid, mock_delete_initiators):
        mock_get_initiator_name.return_value = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        mock_initiator = self.initiator_dm
        self.client_dm.mip = '192.168.0.200'
        mock_get_existing.return_value = {'192.168.0.200': [mock_initiator]}
        self.init_cmt.apply()
        self.init_cmt.reset()

    @mock.patch('valence.vl_components._initiator.get_raw_iscsi_initiator')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    def test__assert_valid(self, mock_get_existing, mock_get_initiator_name, mock_get_raw_iscsi):
        mock_initiator = self.initiator_dm
        mock_initiator.initiator_name = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        self.client_dm.mip = '192.168.0.300'
        mock_get_existing.return_value = {'192.168.0.300': [mock_initiator]}
        mock_get_raw_iscsi.return_value = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        self.init_cmt._assert_valid()

    @mock.patch('valence.vl_components._initiator.get_raw_iscsi_initiator')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    def test__assert_valid_no_initiators(self, mock_get_existing,
                                         mock_get_initiator_name, mock_get_raw_iscsi):
        self.init_cmt.client_dms[0].mip = '192.168.0.301'
        mock_get_existing.return_value = {'192.168.0.301': None}
        with pytest.raises(components.PartialStateError):
            self.init_cmt._assert_valid()

    @mock.patch('valence.vl_components._initiator.get_raw_iscsi_initiator')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    def test__assert_valid_wrong_initiators(self, mock_get_existing,
                                            mock_get_initiator_name, mock_get_raw_iscsi):
        mock_initiator = self.initiator_dm
        mock_initiator.initiator_name = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        self.client_dm.mip = '192.168.0.400'
        mock_get_existing.return_value = {'192.168.0.400': [mock_initiator]}
        mock_get_raw_iscsi.return_value = 'iqn.2010-01.net.solidfire.eng:clabv0007'
        with pytest.raises(components.PartialStateError):
            self.init_cmt._assert_valid()

    @mock.patch('valence.vl_components._initiator.get_raw_iscsi_initiator')
    @mock.patch('valence.functional_models.LinuxClient.get_initiator_name')
    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    def test__assert_valid_too_many_initiators(self, mock_get_existing,
                                               mock_get_initiator_name, mock_get_raw_iscsi):
        mock_initiator = self.initiator_dm
        mock_initiator.initiator_name = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        self.client_dm.mip = '192.168.0.400'
        mock_get_existing.return_value = {'192.168.0.400': [mock_initiator, mock_initiator]}
        mock_get_raw_iscsi.return_value = 'iqn.2010-01.net.solidfire.eng:clabv0006'
        with pytest.raises(components.PartialStateError):
            self.init_cmt._assert_valid()

    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_cluster_initiator_dms')
    def test_get_existing(self, mock_get_cluster_initiator_dms):
        mock_get_cluster_initiator_dms.return_value = [self.initiator_dm, self.initiator_dm]
        self.client_dm.mip = '192.168.0.500'
        self.init_cmt._client_dm_map = {u'clabv0006': self.client_dm}
        initiator_map = self.init_cmt._get_existing()

    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_cluster_initiator_dms')
    def test_get_existing_error(self, mock_get_cluster_initiator_dms):
        mock_get_cluster_initiator_dms.return_value = [self.initiator_dm]
        self.client_dm.mip = '192.168.0.600'
        with pytest.raises(components.PartialStateError):
            initiator_map = self.init_cmt._get_existing()

    @mock.patch('valence.functional_models.Cluster.list_initiators')
    def test_get_cluster_initiator_dms(self, mock_list_initiator):
        mock_list_initiator.return_value = [self.initiator_dm]
        initiators = self.init_cmt._get_cluster_initiator_dms()
        assert initiators == [self.initiator_dm]

    @mock.patch('valence.vl_components.CreateIscsiInitiators._get_existing')
    def test__reload_no_initiators(self, mock_get_existing):
        mock_get_existing.return_value = None
        assert self.init_cmt._reload() is False

    def test_initiator_dms_map_not_applied(self):
        with pytest.raises(components.ComponentStateError):
            self.init_cmt.initiator_dms_map

    def test_get_initiator_dms_map_not_applied(self):
        with pytest.raises(components.ComponentStateError):
            self.init_cmt.get_initiator_dms_map()

    def test_assert_valid_not_applied(self):
        with pytest.raises(components.ComponentStateError):
            self.init_cmt.assert_valid()


class TestMockInitiators(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        self.client_dm = models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04))
        self.initiator_dm = models.Initiator.get_mock_object()
        self.initiator_map = {'192.168.0.100': [self.initiator_dm]}

        self.init_cmt = _initiator.MockInitiators(
            tag='mock',
            client_dms=[self.client_dm],
            initiator_dms_map=self.initiator_map
        )

    def test_apply(self):
        self.init_cmt.apply()
        assert self.init_cmt.state is components.States.applied
        assert self.init_cmt.client_dms == [self.client_dm]
        assert self.init_cmt.full_tag == 'mock-TAG-'

    def test_reset(self):
        self.init_cmt.ensure_applied()
        assert self.init_cmt.state is components.States.applied

        self.init_cmt.reset()
        assert self.init_cmt.state is components.States.ready
        self.init_cmt.ensure_ready()
        assert self.init_cmt.state is components.States.ready

    def test_get(self):
        with self.assertRaises(components.ComponentStateError):
            self.init_cmt.initiator_dms_map

        with self.assertRaises(components.ComponentStateError):
            self.init_cmt.get_initiator_dms_map()

        self.init_cmt.apply()
        assert self.init_cmt.initiator_dms_map == self.initiator_map
        assert self.init_cmt.get_initiator_dms_map() == self.initiator_map

    def test_assert_valid(self):
        self.init_cmt.apply()
        self.init_cmt.assert_valid()
