import unittest

import mock
import pytest
from sf_ssh import ssh

from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    vl_helpers,
)
from valence.components import _enums
from valence.vl_components import _network


class TestDownInterfacesOnNodes(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.active_nodes_dms = models.ActiveNode.get_mock_object_list(3, overrides={'status': 'active'})
        cls.selector = lambda x: cls.active_nodes_dms
        cls.interfaces = [enums.NetworkInterface.eth0]

    def setUp(self):
        self.down_interface_on_node_cmt = _network.DownInterfacesOnNodes(
            self.interfaces, self.cluster_dm, self.selector)

    @mock.patch.object(_network.DownInterfacesOnNodes, '_reload')
    @mock.patch.object(vl_helpers, 'down_network_interfaces')
    def test_apply(self, mock_down_interfaces, mock_reload):
        mock_reload.return_value = False
        self.down_interface_on_node_cmt.apply()
        mock_down_interfaces.assert_called()

    @mock.patch.object(_network.DownInterfacesOnNodes, '_reload')
    @mock.patch.object(vl_helpers, 'up_network_interfaces')
    @mock.patch.object(vl_helpers, 'down_network_interfaces')
    def test_reset(self, mock_down_interfaces, mock_up_interfaces, mock_reload):
        mock_reload.return_value = False
        self.down_interface_on_node_cmt.apply()
        self.down_interface_on_node_cmt.reset()
        mock_up_interfaces.assert_called()

    @mock.patch.object(functional_models.Node, 'get_network_interface_status')
    def test_reload_success(self, mock_get_status):
        mock_get_status.return_value = {'eth0': enums.NetworkStatus.down}
        self.down_interface_on_node_cmt._apply_setup()
        assert self.down_interface_on_node_cmt._reload() is True

    @mock.patch.object(functional_models.Node, 'get_network_interface_status')
    def test_reload_fail(self, mock_get_status):
        mock_get_status.return_value = {'eth0': enums.NetworkStatus.up}
        self.down_interface_on_node_cmt._apply_setup()
        assert self.down_interface_on_node_cmt._reload() is False


class TestContinuousFlapInterfacesOnNodes(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.node_dms = models.ResourceNode.get_mock_object_list(2)  # type: list[models.ResourceNode]
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.interfaces = [enums.NetworkInterface.eth0]
        cls.selector = lambda x: cls.node_dms

    def setUp(self):
        self.cmt = _network.ContinuousFlapInterfacesOnNodes(
            self.interfaces, self.selector, self.cluster_dm, kill_after_seconds=0.2)

    @mock.patch.object(_network.DownInterfacesOnNodes, '_reload')
    @mock.patch.object(vl_helpers, 'up_network_interfaces')
    @mock.patch.object(vl_helpers, 'down_network_interfaces')
    def test_apply_continuous(self, mock_down_interfaces, mock_up_interfaces, mock_reload):
        mock_reload.return_value = False
        self.cmt.apply_async()
        self.cmt.finish()
        assert self.cmt.state is _enums.States.applied
        mock_down_interfaces.assert_called()
        mock_up_interfaces.assert_called()

    @mock.patch.object(_network.DownInterfacesOnNodes, 'ensure_ready')
    @mock.patch.object(_network.DownInterfacesOnNodes, '_reload')
    @mock.patch.object(vl_helpers, 'up_network_interfaces')
    @mock.patch.object(vl_helpers, 'down_network_interfaces')
    def test_reset(self, mock_down_interfaces, mock_up_interfaces, mock_reload, mock_ensure_ready):
        mock_reload.return_value = False
        self.cmt.apply_async()
        self.cmt.finish()
        self.cmt._state = _enums.States.failed
        self.cmt.reset()
        assert self.cmt.state is _enums.States.ready
        mock_ensure_ready.assert_called()


class BaseInterNodeNetemNetworkPerturbation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_one_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_one_nodes = models.ActiveNode.get_mock_object_list(2)
        cls.cluster_two_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_two_nodes = models.ActiveNode.get_mock_object_list(2)

        cls.cluster_one_node_selector = mock.MagicMock(spec=selectors.ClusteredNodeSelector,
                                                       cluster_dm=cls.cluster_one_dm)
        cls.cluster_one_node_selector.return_value = cls.cluster_one_nodes

        cls.cluster_two_node_selector = mock.MagicMock(spec=selectors.ClusteredNodeSelector,
                                                       cluster_dm=cls.cluster_two_dm)
        cls.cluster_two_node_selector.return_value = cls.cluster_two_nodes

        cls.cluster_one_sips = [n.sip for n in cls.cluster_one_nodes]
        cls.cluster_one_sips.insert(0, cls.cluster_one_dm.svip)

        cls.cluster_one_mips = [n.mip for n in cls.cluster_one_nodes]
        cls.cluster_one_mips.insert(0, cls.cluster_one_dm.mvip)

        cls.cluster_two_sips = [n.sip for n in cls.cluster_two_nodes]
        cls.cluster_two_sips.insert(0, cls.cluster_two_dm.svip)

        cls.cluster_two_mips = [n.mip for n in cls.cluster_two_nodes]
        cls.cluster_two_mips.insert(0, cls.cluster_two_dm.mvip)


class TestInterNodeNetemNetworkPerturbation(BaseInterNodeNetemNetworkPerturbation):
    def setUp(self):
        self.network_perturbation_cmt = _network.InterNodeNetemNetworkPerturbation(
            self.cluster_one_node_selector,
            self.cluster_two_node_selector,
            ['delay 1000ms'],
        )

    def test__init__(self):
        retval = _network.InterNodeNetemNetworkPerturbation(
            self.cluster_one_node_selector,
            self.cluster_two_node_selector,
            ['rule 1', 'rule 2']
        )

        assert retval._rules == ['rule 1', 'rule 2']

    @mock.patch(
        'valence.vl_components.InterNodeNetemNetworkPerturbation._ensure_no_existing_rules')
    def test_apply_setup_bidirectional(self, mock_ensure_no_existing_rules):
        self.network_perturbation_cmt._apply_setup()
        assert mock_ensure_no_existing_rules.call_count == 2
        mock_ensure_no_existing_rules.assert_any_call(self.network_perturbation_cmt._set_one_node_fms)
        mock_ensure_no_existing_rules.assert_any_call(self.network_perturbation_cmt._set_two_node_fms)

    @mock.patch('valence.vl_components.InterNodeNetemNetworkPerturbation._apply_setup')
    @mock.patch('valence.vl_helpers.clear_netem_rules')
    @mock.patch('valence.vl_helpers.add_netem_rules_inter_nodes')
    def test_apply_bidirectional(self, mock_add_netem_rules_inter_cluster, mock_clear_netem_rules,
                                 mock_setup):
        self.network_perturbation_cmt.apply()

        mock_clear_netem_rules.assert_not_called()

        assert mock_add_netem_rules_inter_cluster.call_count == 2

        mock_add_netem_rules_inter_cluster.assert_any_call(
            self.network_perturbation_cmt._set_one_node_fms,
            {
                'Bond10G': self.cluster_two_sips,
                'Bond1G': self.cluster_two_mips
            },
            ['delay 1000ms'])

        mock_add_netem_rules_inter_cluster.assert_any_call(
            self.network_perturbation_cmt._set_two_node_fms,
            {
                'Bond10G': self.cluster_one_sips,
                'Bond1G': self.cluster_one_mips
            },
            ['delay 1000ms'])

        assert mock_setup.call_count == 1

    @mock.patch('valence.vl_components.InterNodeNetemNetworkPerturbation._apply_setup')
    @mock.patch('valence.vl_helpers.clear_netem_rules')
    @mock.patch('valence.vl_helpers.add_netem_rules_inter_nodes')
    def test_apply_single_direction_no_virtual_ip(self, mock_add_netem_rules_inter_cluster,
                                                  mock_clear_netem_rules, mock_setup):
        single_direction_cmt = _network.InterNodeNetemNetworkPerturbation(
            self.cluster_one_node_selector,
            self.cluster_two_node_selector,
            ['delay 500ms 50ms'],
            bidirectional=False,
            virtual_ip=False,
        )

        single_direction_cmt.apply()

        mock_clear_netem_rules.assert_not_called()

        # This test will skip using the virtual IP, so it will be slightly different from the list
        # inherited in the setupClass.
        cluster_two_sips = [n.sip for n in self.cluster_two_nodes]
        cluster_two_mips = [n.mip for n in self.cluster_two_nodes]

        assert mock_add_netem_rules_inter_cluster.call_count == 1
        mock_add_netem_rules_inter_cluster.assert_called_with(
            single_direction_cmt._set_one_node_fms,
            {
                'Bond10G': cluster_two_sips,
                'Bond1G': cluster_two_mips
            },
            ['delay 500ms 50ms'])

        assert mock_setup.call_count == 1

    @mock.patch.object(_network.InterNodeLatencyPerturbation, '_reload')
    @mock.patch('valence.vl_helpers.clear_netem_rules')
    def test_reset(self, mock_clear_netem_rules, mock_reload):
        mock_reload.return_value = False

        self.network_perturbation_cmt._state = components.States.applied

        # Force setting the variables assigned during setup that are needed by reset
        self.network_perturbation_cmt._set_one_node_fms = [functional_models.Node.get_mock_fm()]
        self.network_perturbation_cmt._set_two_node_fms = [functional_models.Node.get_mock_fm()]

        self.network_perturbation_cmt.reset()

        assert mock_clear_netem_rules.call_count == 2
        mock_clear_netem_rules.assert_any_call(self.network_perturbation_cmt._set_one_node_fms,
                                               ['Bond10G', 'Bond1G'])
        mock_clear_netem_rules.assert_any_call(self.network_perturbation_cmt._set_two_node_fms,
                                               ['Bond10G', 'Bond1G'])

    @mock.patch('sf_ssh.Ssh')
    def test_ensure_no_existing_rules(self, mock_ssh):
        node_fm = functional_models.Node.get_mock_fm()

        inst = mock_ssh.return_value.__enter__()
        inst.run_command.return_value = ssh.RunCommandOutput(stderr='',
                                                             stdout='qdisc noqueue 0: root refcnt 2 \n',
                                                             return_code=1)

        self.network_perturbation_cmt._ensure_no_existing_rules([node_fm])

        assert inst.run_command.call_count == 2
        inst.run_command.assert_any_call('tc qdisc show dev Bond10G', raise_on_failure=True)
        inst.run_command.assert_any_call('tc qdisc show dev Bond1G', raise_on_failure=True)

    @mock.patch('sf_ssh.Ssh')
    def test_ensure_no_existing_rules_has_existing_rules(self, mock_ssh):
        existing_rules_msg = ('qdisc prio 1: root refcnt 17 bands 3 priomap  2 2 2 2 2 2 2 2 2 2 2 2 2\n'
                              'qdisc netem 10: parent 1:1 limit 1000 delay 200.0ms  75.0ms loss 5%')

        node_fm = functional_models.Node.get_mock_fm()

        inst = mock_ssh.return_value.__enter__()
        inst.run_command.return_value = ssh.RunCommandOutput(stderr='',
                                                             stdout=existing_rules_msg,
                                                             return_code=1)

        with pytest.raises(components.ResourceStateError):
            self.network_perturbation_cmt._ensure_no_existing_rules([node_fm])

        assert inst.run_command.call_count == 1
        inst.run_command.assert_any_call('tc qdisc show dev Bond10G', raise_on_failure=True)


class TestInterClusterLatencyPerturbation(BaseInterNodeNetemNetworkPerturbation):

    def test__init__(self):
        retval = _network.InterNodeLatencyPerturbation(
            self.cluster_one_node_selector,
            self.cluster_two_node_selector, 56
        )

        assert retval._rules == ['delay 56ms']

    def test__init__with_jitter(self):
        retval = _network.InterNodeLatencyPerturbation(
            self.cluster_one_node_selector,
            self.cluster_two_node_selector, 56,
            jitter=25
        )

        assert retval._rules == ['delay 56ms 25ms']

    def test__init__with_bad_jitter_value(self):
        with pytest.raises(ValueError):
            _network.InterNodeLatencyPerturbation(
                self.cluster_one_node_selector,
                self.cluster_two_node_selector, 56,
                jitter='foo'
            )


class TestInterClusterPacketLossPerturbation(BaseInterNodeNetemNetworkPerturbation):
    def test__init__(self):
        retval = _network.InterNodePacketLossPerturbation(
            self.cluster_one_node_selector,
            self.cluster_two_node_selector,
            56.0
        )

        assert retval._rules == ['loss 56.0%']

    def test__init__with_bad_value(self):
        with pytest.raises(ValueError):
            _network.InterNodePacketLossPerturbation(
                self.cluster_one_node_selector,
                self.cluster_two_node_selector,
                101.0
            )
