"""Test valence.vl_components.vl_components._node components."""

import unittest

import mock
import pytest
import sf_ssh

from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    vl_components,
    vl_helpers,
)
from valence.components import _enums
from valence.vl_components import _node


class TestSetNodeConstantsComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_dm = models.ResourceNode.get_mock_object()
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_constants = models.ClusterConstants.get_mock_object()

    def setUp(self):
        self.set_constants_cmt = vl_components._node.SetNodeConstants(self.node_dm,
                                                                      self.cluster_dm,
                                                                      self.cluster_constants)

    @mock.patch('valence.functional_models.Node.set_constants')
    @mock.patch('valence.functional_models.Node.get_constants')
    def test_apply(self, mock_get_node_const, mock_set_node_const):
        mock_get_node_const.return_value = self.cluster_constants
        self.set_constants_cmt.apply()
        mock_set_node_const.assert_called_with(self.cluster_constants)

    @mock.patch('valence.functional_models.Node.reset_constants')
    @mock.patch('valence.functional_models.Node.set_constants')
    @mock.patch('valence.functional_models.Node.get_constants')
    def test_reset_to_default(self, mock_get_node_const, mock_set_node_const, mock_reset_node_const):
        mock_get_node_const.return_value = models.ClusterConstants({})
        self.set_constants_cmt.apply()
        self.set_constants_cmt.reset()
        reset_const_dm = mock_reset_node_const.call_args_list[-1][0][0]
        assert reset_const_dm.serialize() == self.cluster_constants.serialize()

    @mock.patch('valence.functional_models.Node.reset_constants')
    @mock.patch('valence.functional_models.Node.set_constants')
    @mock.patch('valence.functional_models.Node.get_constants')
    def test_reset_to_original(self, mock_get_node_const, mock_set_node_const, mock_reset_node_const):
        mock_get_node_const.return_value = self.cluster_constants
        self.set_constants_cmt.apply()
        self.set_constants_cmt.reset()
        mock_set_node_const.assert_called_with(self.cluster_constants)
        mock_reset_node_const.assert_not_called()


class TestResetNodesComponent(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        self.active_nodes = models.ActiveNode.get_mock_object_list(1)
        self.selector = selectors.ActiveNodeSelector(self.cluster_dm)
        self.reset_nodes_cmt = vl_components._node.ResetNode(self.cluster_dm, self.selector)

    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    # @mock.patch('valence.vl_helpers.reset_nodes')
    def test_apply(self, mock_selector):
        mock_selector.return_value = []
        with pytest.raises(components.ComponentStateError):
            self.reset_nodes_cmt.apply()

        mock_selector.return_value = self.active_nodes
        self.reset_nodes_cmt.apply()

        with self.assertRaises(components.ComponentStateError):
            self.reset_nodes_cmt.reset()

        self.reset_nodes_cmt.state = _enums.States.failed
        with self.assertRaises(components.ComponentStateError):
            self.reset_nodes_cmt.reset()

    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    @mock.patch('valence.vl_helpers.reset_nodes', mock.MagicMock())
    def test_apply(self, mock_selector):
        mock_selector.return_value = []
        with pytest.raises(components.ComponentStateError):
            self.reset_nodes_cmt.apply()

        mock_selector.return_value = self.active_nodes
        self.reset_nodes_cmt.apply()

    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    def test_no_nodes_from_selector(self, mock_selector):
        mock_selector.return_value = []
        with pytest.raises(components.ComponentStateError):
            self.reset_nodes_cmt.apply()

    @mock.patch('valence.selectors._node.ActiveNodeSelector.__call__')
    def test_reset(self, mock_selector):
        mock_selector.return_value = self.active_nodes
        self.reset_nodes_cmt.state = _enums.States.applied
        with self.assertRaises(components.ComponentStateError):
            self.reset_nodes_cmt.reset()

        self.reset_nodes_cmt.state = _enums.States.failed
        with self.assertRaises(components.ComponentStateError):
            self.reset_nodes_cmt.reset()


class TestCascadeFailNodes(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        node_dm = models.ActiveNode.get_mock_object()
        self.cascade_fail_cmt = _node.CascadeFailNodes(self.cluster_dm,
                                                       node_selector=lambda: [node_dm])

    @mock.patch('valence.functional_models.Cluster.list_events')
    @mock.patch('valence.vl_components.PowerOffNodes.apply')
    @mock.patch('valence.vl_helpers.wait_for_bin_assignments')
    @mock.patch('valence.vl_components.PowerOffNodes.ensure_ready')
    @mock.patch('valence.vl_helpers.wait_for_bin_sync_to_finish_with_timeout')
    def test_apply_reset(self, mock_wait_for_bin_sync, mock_power_off_nodes_ready,
                         mock_wait_for_bin_assignment, mock_power_off_nodes_apply, mock_list_events):
        mock_list_events.return_value = models.ClusterEvent.get_mock_object_list(1)
        self.cascade_fail_cmt.apply()
        self.cascade_fail_cmt.reset()


class TestCascadeAddNodes(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        node_dm1 = models.ActiveNode.get_mock_object()
        node_dm2 = models.ActiveNode.get_mock_object()
        self.cascade_add_cmt = _node.CascadeAddNodes(self.cluster_dm,
                                                     node_selector=lambda: [node_dm1, node_dm2])

    @mock.patch('valence.vl_components.AddNodes.ensure_ready')
    @mock.patch('valence.vl_components.AddNodes.apply')
    @mock.patch('valence.functional_models.Cluster.list_events')
    @mock.patch('valence.vl_helpers.wait_for_bin_assignments')
    @mock.patch('valence.vl_helpers.wait_for_bin_sync_to_finish_with_timeout')
    def test_apply_reset(self, mock_wait_for_bin_sync, mock_wait_for_bin_assignment, mock_list_events,
                         mock_add_nodes_apply, mock_add_nodes_ensure_ready):
        mock_list_events.return_value = models.ClusterEvent.get_mock_object_list(1)
        self.cascade_add_cmt.apply()
        self.cascade_add_cmt.reset()


class TestAddNodes(unittest.TestCase):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        node_dm = models.ActiveNode.get_mock_object()
        self.add_nodes_cmt = _node.AddNodes(self.cluster_dm, node_selector=lambda: [node_dm],
                                            auto_install=False)

    @mock.patch('valence.vl_helpers.remove_drives')
    @mock.patch('valence.vl_helpers.add_drives')
    @mock.patch('valence.vl_helpers.convert_pending_nodes_to_active_nodes')
    @mock.patch('valence.functional_models.Cluster.list_pending_nodes')
    @mock.patch('valence.functional_models.Cluster.list_drives')
    @mock.patch('valence.functional_models.Cluster.add_nodes')
    @mock.patch('valence.functional_models.AsyncHandler.join')
    @mock.patch('valence.functional_models.Cluster.remove_nodes')
    def test_apply_reset(self, mock_remove_nodes, mock_add_nodes_join, mock_add_nodes,
                         mock_list_drives, mock_list_pending_nodes,
                         mock_convert_pending_nodes_to_active_nodes, mock_add_drives,
                         mock_remove_drives):
        mock_list_drives.return_value = models.Drive.get_mock_object_list(2)
        cluster_dm = models.ResourceCluster.get_mock_object()
        cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        async_handle = models.AsyncHandle.get_mock_object()
        async_handler = functional_models.AsyncHandler(cluster_fm, async_handle_dm=async_handle)
        mock_add_nodes.return_value = [async_handler]
        self.add_nodes_cmt.apply()
        self.add_nodes_cmt.reset()


class TestEncryptionCapabilityCmts(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_dm = models.ResourceNode.get_mock_object()
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    @mock.patch.object(vl_helpers, 'wait_for_node_api_up')
    @mock.patch.object(vl_helpers, 'wait_for_healthy_services')
    @mock.patch.object(vl_helpers, 'wait_for_api_up')
    @mock.patch.object(functional_models.Node, 'get_fips_drives_config')
    @mock.patch.object(functional_models.Node, 'set_fips_drives_config')
    def test_enable_fips_node_config(self, mock_set_fips, mock_get_fips, mock_api_wait, mock_svc_wait,
                                     mock_node_api_wait):
        cmt = vl_components.SetFipsNodeConfiguration(self.node_dm, cluster_dm=self.cluster_dm)
        cmt.apply()
        cmt.reset()

    @mock.patch.object(vl_components.SetFipsNodeConfiguration, 'reset')
    @mock.patch.object(vl_components.SetFipsNodeConfiguration, 'finish')
    @mock.patch.object(vl_components.SetFipsNodeConfiguration, 'apply_async')
    def test_enable_fips_node_config_multi(self, mock_apply_async, mock_finish, mock_reset):
        cmt = vl_components.SetFipsNodeConfigurationMulti([self.node_dm], cluster_dm=self.cluster_dm)
        cmt.apply()
        cmt.reset()
        mock_apply_async.assert_called()
        mock_finish.assert_called()
        mock_reset.assert_called()

    @mock.patch.object(vl_helpers, 'wait_for_node_api_up')
    @mock.patch.object(vl_helpers, 'wait_for_healthy_services')
    @mock.patch.object(vl_helpers, 'wait_for_api_up')
    @mock.patch.object(functional_models.Node, 'restart_binary')
    @mock.patch.object(functional_models.Node, 'get_ssh_connection')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Node, 'set_fake_drives_encryption_capability')
    def test_fake_drives_encryption_capability(self, mock_fake_encryption, mock_run_cmd,
                                               mock_get_connection, mock_restart_binary,
                                               mock_api_wait, mock_svc_wait, mock_node_api_wait):
        cmt = vl_components.FakeDrivesEncryptionCapability(self.node_dm,
                                                           enums.DriveEncryptionCapability.fips,
                                                           self.cluster_dm)
        cmt.apply()
        cmt._override_path = 'abc'
        cmt.reset()
