import time
import unittest

import mock
import pytest

from valence import components
from valence import enums as model_enums
from valence import (
    functional_models,
    models,
    selectors,
    vl_helpers,
)
from valence.components import _enums
from valence.vl_components import _power


class TestPowerOffNodesComponent(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.nodes = models.ActiveNode.get_mock_object_list(2)
        self.node_selector = lambda: self.nodes
        self.power_off_cmt = _power.PowerOffNodes(self.node_selector, self.cluster_dm)

    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_start')
    @mock.patch.object(vl_helpers, 'wait_for_slice_sync_to_finish')
    @mock.patch.object(functional_models.Node, 'get_node_dm')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(_power.PowerOffNodes, '_reload')
    @mock.patch.object(vl_helpers, 'power_off_nodes_via_oob')
    @mock.patch.object(vl_helpers, 'get_bin_sync_timeout_for_nodes')
    def test__apply(self, mock_get_sync, mock_power_off, mock_reload,
                    mock_list_drives, mock_get_node_dm,
                    mock_wait_for_slice_sync, mock_wait_for_bin_sync, mock_bin_sync_start,
                    mock_wait_for_fault_occurrence, mock_wait_for_fault_resolution):
        mock_reload.return_value = False
        mock_storage_node = models.ActiveNode.get_mock_object()
        mock_storage_node.role = model_enums.NodeRole.storage
        mock_get_node_dm.return_value = mock_storage_node
        self.power_off_cmt.apply()
        mock_power_off.assert_called_once()
        self.assertEqual(mock_power_off.call_count, 1)

    @mock.patch.object(vl_helpers, 'wait_for_drive_status')
    @mock.patch.object(vl_helpers, 'remove_drives')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'is_bin_syncing')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_start')
    @mock.patch.object(vl_helpers, 'wait_for_slice_sync_to_finish')
    @mock.patch.object(functional_models.Cluster, 'add_drives')
    @mock.patch.object(vl_helpers, 'wait_for_drive_statuses')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(_power.PowerOffNodes, '_reload')
    @mock.patch.object(vl_helpers, 'power_on_nodes_via_oob')
    def test__reset(self, mock_power_on, mock_reload, mock_list_drives, mock_wait,
                    mock_add_drives, mock_wait_for_slice_sync, mock_bin_sync_start,
                    mock_wait_for_bin_sync, mock_is_bin_syncing,
                    mock_wait_for_fault_occurrence, mock_wait_for_fault_resolution,
                    mock_remove_drives, mock_sleep):
        mock_reload.return_value = False

        mock_is_bin_syncing.return_value = True
        failed_drive = models.Drive.get_mock_object(
            overrides=dict(status=model_enums.DriveStatus.failed, drive_id=1))
        avail_drive = models.Drive.get_mock_object(
            overrides=dict(status=model_enums.DriveStatus.available, drive_id=2))
        mock_list_drives.side_effect = [[failed_drive], [avail_drive]]
        self.power_off_cmt._initial_active_drive_ids = [1]
        self.power_off_cmt._state = _enums.States.applied
        self.power_off_cmt.reset()
        mock_power_on.assert_called_once()
        self.assertEqual(mock_power_on.call_count, 1)

    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(vl_helpers, 'wait_for_slice_sync_to_finish')
    @mock.patch.object(functional_models.Node, 'get_node_dm')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers, 'power_off_nodes_via_oob')
    @mock.patch.object(functional_models.Node, 'get_oob_power_status')
    @mock.patch.object(vl_helpers, 'get_bin_sync_timeout_for_nodes')
    def test__reload(self, mock_get_sync, mock_get_status, mock_power_off, mock_list_drives,
                     mock_get_node_dm, mock_wait_for_slice_sync, mock_wait_for_bin_sync,
                     mock_wait_for_fault_occurrence, mock_wait_for_fault_resolution):
        mock_storage_node = models.ActiveNode.get_mock_object()
        mock_storage_node.role = model_enums.NodeRole.storage
        mock_get_node_dm.return_value = mock_storage_node
        self.power_off_cmt.apply()
        # if all nodes are OFF, should be able to reload
        mock_get_status.side_effect = [model_enums.OobPowerState.off, model_enums.OobPowerState.off]
        assert self.power_off_cmt._reload()

        # if all nodes are ON, should not be able to reload
        mock_get_status.side_effect = [model_enums.OobPowerState.on, model_enums.OobPowerState.on]
        assert not self.power_off_cmt._reload()

        # if mixed state, throw partial state error
        mock_get_status.side_effect = [model_enums.OobPowerState.off, model_enums.OobPowerState.on]
        with pytest.raises(components.PartialStateError):
            self.power_off_cmt._reload()


class TestPowerOffNodeComponent(TestPowerOffNodesComponent):

    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.mock_drive_dms = models.Drive.get_mock_object_list(2)
        cls.mock_drive_dms[0].type = model_enums.DriveType.block
        cls.mock_bin_report = models.BinReport.create()

    def setUp(self):
        self.nodes = models.ActiveNode.get_mock_object_list(2)
        self.node_selector = lambda: self.nodes
        self.power_off_node_cmt = _power.PowerOffNode(self.node_selector, self.cluster_dm)

    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(vl_helpers, 'wait_for_bin_sync_to_start')
    @mock.patch.object(vl_helpers, 'wait_for_slice_sync_to_finish')
    @mock.patch.object(functional_models.Node, 'get_node_dm')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(_power.PowerOffNode, '_reload')
    @mock.patch.object(vl_helpers, 'power_off_nodes_via_oob')
    @mock.patch.object(vl_helpers, 'get_bin_sync_timeout_for_nodes')
    def test__apply(self, mock_get_sync, mock_power_off, mock_with_binsync_reload,
                    mock_list_drives, mock_get_node_dm,
                    mock_wait_for_slice_sync, mock_wait_for_bin_sync, mock_bin_sync_start,
                    mock_wait_for_fault_occurrence, mock_wait_for_fault_resolution):
        mock_storage_node = models.ActiveNode.get_mock_object(overrides={
            'platform_info': None,
            'internal_role': model_enums.NodeRole.storage
        })
        mock_get_node_dm.return_value = mock_storage_node
        mock_with_binsync_reload.return_value = False
        self.power_off_node_cmt.apply()
        assert mock_storage_node.role == model_enums.NodeRole.storage
        self.assertEqual(mock_power_off.call_count, 1)

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers, 'recover_failed_drives')
    @mock.patch.object(vl_helpers, 'power_on_nodes_via_oob')
    def test__reset(self, mock_power_on, mock_failed_drives, mock_list_drives):
        failed_drive = models.Drive.get_mock_object(
            overrides=dict(status=model_enums.DriveStatus.failed, drive_id=1, slot=1))
        avail_drive = models.Drive.get_mock_object(
            overrides=dict(status=model_enums.DriveStatus.available, drive_id=2, slot=1))

        self.power_off_node_cmt._state = _enums.States.applied
        self.power_off_node_cmt.reset()
        self.assertEqual(mock_power_on.call_count, 1)

    def test__reload(self):

        super(TestPowerOffNodeComponent, self).setUp()
        super(TestPowerOffNodeComponent, self).test__reload()


class TestPowerOffClusterComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.nodes = models.ActiveNode.get_mock_object_list(4)
        self.node_selector = lambda: self.nodes
        self.power_off_cluster_cmt = _power.PowerOffCluster(cluster_dm=self.cluster_dm)

    @mock.patch.object(selectors, 'ActiveNodeSelector')
    def test__apply_setup(self, mock_node_selectors):
        mock_node_selectors.return_value = self.node_selector
        self.power_off_cluster_cmt._apply_setup()
        self.assertEqual(len(self.nodes), len(self.power_off_cluster_cmt._node_fms))

    @mock.patch.object(vl_helpers, 'power_off_nodes_via_oob')
    def test__apply(self, mock_power_off_oob):
        self.power_off_cluster_cmt._apply()
        mock_power_off_oob.assert_called_once()
        self.assertEqual(mock_power_off_oob.call_count, 1)

    @mock.patch.object(vl_helpers, 'wait_for_api_up')
    @mock.patch.object(vl_helpers, 'wait_for_healthy_services')
    @mock.patch.object(vl_helpers, 'power_on_nodes_via_oob')
    def test__reset(self, mock_power_on, mock_healthy, mock_wait_api):
        self.power_off_cluster_cmt._state = _enums.States.applied
        self.power_off_cluster_cmt.reset()
        self.assertEqual(mock_power_on.call_count, 1)
        self.assertEqual(mock_healthy.call_count, 1)
        self.assertEqual(mock_wait_api.call_count, 1)

    @mock.patch.object(selectors, 'ActiveNodeSelector')
    @mock.patch.object(vl_helpers, 'power_off_nodes_via_oob')
    @mock.patch.object(functional_models.Node, 'get_oob_power_status')
    def test__reload(self, mock_get_status, mock_power_off, mock_selector):
        mock_power_off.return_value = None
        mock_selector.return_value = self.node_selector
        self.power_off_cluster_cmt.apply()
        # if all nodes are OFF, should be able to reload
        mock_get_status.side_effect = [model_enums.OobPowerState.off, model_enums.OobPowerState.off,
                                       model_enums.OobPowerState.off, model_enums.OobPowerState.off]
        assert self.power_off_cluster_cmt._reload()

        # if all nodes are ON, should not be able to reload
        mock_get_status.side_effect = [model_enums.OobPowerState.on, model_enums.OobPowerState.on,
                                       model_enums.OobPowerState.on, model_enums.OobPowerState.on]
        assert not self.power_off_cluster_cmt._reload()

        # if mixed state, throw partial state error
        mock_get_status.side_effect = [model_enums.OobPowerState.off, model_enums.OobPowerState.on,
                                       model_enums.OobPowerState.off, model_enums.OobPowerState.on]
        with pytest.raises(components.PartialStateError):
            self.power_off_cluster_cmt._reload()
