"""Test valence.vl_components._remote_replication."""

import unittest

import mock
import pytest

from sf_api import api_errors
from valence import (
    components,
    enums,
    errors,
    functional_models,
    models,
    vl_components,
)


class TestPairClustersComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm_one = models.ResourceCluster.get_mock_object()
        cls.cluster_dm_two = models.ResourceCluster.get_mock_object()
        cls.cluster_fm_one = functional_models.Cluster(cls.cluster_dm_one)
        cls.cluster_fm_two = functional_models.Cluster(cls.cluster_dm_two)
        cls.cluster_pair_dm_one = models.ClusterPair.get_mock_object(
            overrides={'mvip': cls.cluster_dm_two.mvip})
        cls.cluster_pair_dm_two = models.ClusterPair.get_mock_object(
            overrides={'mvip': cls.cluster_dm_one.mvip})

    def setUp(self):
        self.pair_cmt = vl_components.PairClusters(self.cluster_dm_one, self.cluster_dm_two)

    @mock.patch('valence.functional_models.Cluster.start_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.complete_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.list_cluster_pairs')
    @mock.patch('valence.vl_components.PairClusters._get_pair_dm')
    @mock.patch('valence.vl_components.PairClusters._reload')
    def test_apply(self, mock__reload, mock__get_pair_dm, mock_list_cluster_pairs,
                   mock_complete_cluster_pairing, mock_start_cluster_pairing):
        mock_start_cluster_pairing.return_value = 'somepairingkey'
        mock_complete_cluster_pairing.return_value = 1
        mock__get_pair_dm.return_value = self.cluster_pair_dm_one
        mock_list_cluster_pairs.return_value = [self.cluster_pair_dm_one]
        mock__reload.return_value = False
        self.pair_cmt.apply()
        mock_start_cluster_pairing.assert_called()
        mock_complete_cluster_pairing.assert_called_with('somepairingkey')
        mock__get_pair_dm.assert_called()

    @mock.patch('valence.functional_models.Cluster.start_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.complete_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.list_cluster_pairs')
    @mock.patch('valence.functional_models.Cluster.remove_cluster_pair')
    @mock.patch('valence.vl_components.PairClusters._get_pair_dm')
    def test_reset(self, mock__get_pair_dm, mock_remove_cluster_pair, mock_list_cluster_pairs,
                   mock_complete_cluster_pairing, mock_start_cluster_pairing):
        self.test_apply()
        self.pair_cmt.reset()
        mock_remove_cluster_pair.assert_has_calls([mock.call(self.cluster_pair_dm_one),
                                                   mock.call(1)])

    def test_reset_w_raise(self):
        self.test_apply()
        self.pair_cmt._cluster_two_pair_dm = None
        with pytest.raises(components.ComponentStateError):
            self.pair_cmt.reset()

    @mock.patch('valence.vl_components.PairClusters._get_pair_dm')
    def test_reload(self, mock_filter_pair_id):
        mock_filter_pair_id.side_effect = [1, 2]
        assert self.pair_cmt._reload() is True
        mock_filter_pair_id.side_effect = [1, None]
        assert self.pair_cmt._reload() is False

    @mock.patch('valence.functional_models.Cluster.list_cluster_pairs')
    def test_get_pair_dm(self, mock_list_cluster_pairs):
        mock_list_cluster_pairs.return_value = [self.cluster_pair_dm_one]
        resp = self.pair_cmt._get_pair_dm(self.cluster_fm_one, self.cluster_fm_two)
        assert resp == self.cluster_pair_dm_one

    @mock.patch('valence.functional_models.Cluster.start_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.complete_cluster_pairing')
    @mock.patch('valence.vl_components.PairClusters._get_pair_dm')
    def test_pair_clusters_dm_w_exception(self, mock_get_pair, mock_complete, mock_start):
        mock_get_pair.return_value = None

        with pytest.raises(components.ComponentStateError):
            self.pair_cmt._pair_clusters()

        mock_start.assert_called_with()
        mock_complete.assert_called_with(mock_start.return_value)

    @mock.patch('valence.functional_models.Cluster.start_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.complete_cluster_pairing')
    @mock.patch('valence.vl_components.PairClusters._get_pair_dm')
    def test_pair_clusters_dm_w_exception_complete_none(self, mock_get_pair, mock_complete, mock_start):
        mock_complete.return_value = None

        with pytest.raises(components.ComponentStateError):
            self.pair_cmt._pair_clusters()

        mock_start.assert_called_with()
        mock_get_pair.assert_called_with(self.pair_cmt._cluster_one_fm, self.pair_cmt._cluster_two_fm)

    @mock.patch('valence.functional_models.Cluster.start_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.complete_cluster_pairing')
    @mock.patch('valence.vl_components.PairClusters._get_pair_dm')
    @mock.patch('time.sleep')
    def test_pair_clusters_intermittent_failure(self, mock_sleep, mock_get_pair, mock_complete,
                                                mock_start):
        mock_get_pair.return_value = None
        mock_start.return_value = 'somepairingkey'
        mock_complete.side_effect = [api_errors.InvalidPairingKey('foo', 'bar'),
                                     self.cluster_pair_dm_two]
        mock_get_pair.return_value = self.cluster_pair_dm_one

        self.pair_cmt._pair_clusters()

        assert mock_sleep.call_count == 1
        assert mock_start.call_count == 2
        assert mock_complete.call_count == 2
        mock_complete.assert_called_with(mock_start.return_value)

    @mock.patch('valence.functional_models.Cluster.start_cluster_pairing')
    @mock.patch('valence.functional_models.Cluster.complete_cluster_pairing')
    @mock.patch('time.sleep')
    def test_pair_clusters_multiple_failures(self, mock_sleep, mock_complete, mock_start):
        mock_start.return_value = 'somepairingkey'
        mock_complete.side_effect = api_errors.InvalidPairingKey('foo', 'bar')

        with pytest.raises(api_errors.InvalidPairingKey):
            self.pair_cmt._pair_clusters()

        assert mock_sleep.call_count == 1
        assert mock_start.call_count == 2
        assert mock_complete.call_count == 2
        mock_complete.assert_called_with(mock_start.return_value)


class TestVolumePairs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.src_cluster_dm, cls.tgt_cluster_dm = models.ResourceCluster.get_mock_object_list(2)
        cls.src_volume_dm = models.Volume.get_mock_object(force_mock={'name': True})
        cls.tgt_volume_dm = models.Volume.get_mock_object(force_mock={'name': True}, overrides={
            'access': enums.VolumeAccess.replication_target.value
        })

    def setUp(self):
        self.volume_pair_cmt = vl_components.PairVolumes(
            self.src_cluster_dm, self.src_volume_dm,
            self.tgt_cluster_dm, self.tgt_volume_dm
        )

    def test_init(self):
        test_component = vl_components.PairVolumes(
            self.src_cluster_dm, self.src_volume_dm,
            self.tgt_cluster_dm, self.tgt_volume_dm
        )
        assert test_component._cluster_src_dm == self.src_cluster_dm
        assert test_component._volume_src_dm == self.src_volume_dm
        assert test_component._volume_pair_src_dm is None

        assert test_component._cluster_tgt_dm == self.tgt_cluster_dm
        assert test_component._volume_tgt_dm == self.tgt_volume_dm
        assert test_component._volume_pair_tgt_dm is None

        assert test_component.volume_source_dm == self.src_volume_dm
        assert test_component.volume_target_dm == self.tgt_volume_dm

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.start_volume_pairing')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.complete_volume_pairing')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._reload')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    def test_apply_target_correct_access_mode(self, mock_get_volume_pair, mock_reload,
                                              mock_complete_pairing,
                                              mock_start_volume_pairing):
        mock_start_volume_pairing.return_value = 'foo-pair-key'
        mock_reload.return_value = False

        volume_pairs = models.VolumePair.get_mock_object_list(2)
        mock_get_volume_pair.side_effect = volume_pairs

        self.volume_pair_cmt.apply()

        mock_start_volume_pairing.assert_called_with(
            self.src_volume_dm,
            mode=None
        )

        mock_complete_pairing.assert_called_with(
            self.tgt_volume_dm,
            'foo-pair-key',
        )

        assert mock_get_volume_pair.call_count == 2
        mock_get_volume_pair.assert_any_call(
            self.volume_pair_cmt._cluster_src_fm,
            self.src_volume_dm,
            self.tgt_volume_dm
        )

        mock_get_volume_pair.assert_any_call(
            self.volume_pair_cmt._cluster_tgt_fm,
            self.tgt_volume_dm,
            self.src_volume_dm
        )

        assert self.volume_pair_cmt._volume_pair_src_dm == volume_pairs[0]
        assert self.volume_pair_cmt._volume_pair_tgt_dm == volume_pairs[1]
        assert self.volume_pair_cmt._tgt_volume_original_access_mode is None

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.modify_volumes')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.start_volume_pairing')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.complete_volume_pairing')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._reload')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    def test_apply_target_incorrect_access_mode(self, mock_get_volume_pair, mock_reload,
                                                mock_complete_pairing,
                                                mock_start_volume_pairing,
                                                mock_modify_volume):
        read_only_tgt_volume = models.Volume.get_mock_object(overrides={
            'access': enums.VolumeAccess.read_only.value
        })

        read_only_volume_pair_cmt = vl_components.PairVolumes(
            self.src_cluster_dm, self.src_volume_dm,
            self.tgt_cluster_dm, read_only_tgt_volume
        )

        mock_start_volume_pairing.return_value = 'foo-pair-key'
        mock_reload.return_value = False

        volume_pairs = models.VolumePair.get_mock_object_list(2)
        mock_get_volume_pair.side_effect = volume_pairs

        read_only_volume_pair_cmt.apply()

        mock_modify_volume.assert_called_with(
            [read_only_tgt_volume],
            access=enums.VolumeAccess.replication_target
        )

        mock_start_volume_pairing.assert_called_with(
            self.src_volume_dm,
            mode=None
        )

        mock_complete_pairing.assert_called_with(
            read_only_tgt_volume,
            'foo-pair-key',
        )

        assert mock_get_volume_pair.call_count == 2
        mock_get_volume_pair.assert_any_call(
            read_only_volume_pair_cmt._cluster_src_fm,
            self.src_volume_dm,
            read_only_tgt_volume
        )

        mock_get_volume_pair.assert_any_call(
            read_only_volume_pair_cmt._cluster_tgt_fm,
            read_only_tgt_volume,
            self.src_volume_dm
        )

        assert read_only_volume_pair_cmt._volume_pair_src_dm == volume_pairs[0]
        assert read_only_volume_pair_cmt._volume_pair_tgt_dm == volume_pairs[1]
        assert read_only_volume_pair_cmt._tgt_volume_original_access_mode is read_only_tgt_volume.access

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.start_volume_pairing')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.complete_volume_pairing')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._reload')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    @mock.patch('time.sleep')
    def test_apply_intermittent_failure(self, mock_sleep, mock_get_volume_pair, mock_reload,
                                        mock_complete_pairing,
                                        mock_start_volume_pairing):
        mock_start_volume_pairing.return_value = 'foo-pair-key'
        mock_complete_pairing.side_effect = [api_errors.InvalidPairingKey('foo', 'bar'), None]
        mock_reload.return_value = False

        volume_pairs = models.VolumePair.get_mock_object_list(2)
        mock_get_volume_pair.side_effect = volume_pairs

        self.volume_pair_cmt.apply()

        mock_start_volume_pairing.assert_called_with(
            self.src_volume_dm,
            mode=None
        )

        mock_complete_pairing.assert_called_with(
            self.tgt_volume_dm,
            'foo-pair-key',
        )

        assert mock_get_volume_pair.call_count == 2
        mock_get_volume_pair.assert_any_call(
            self.volume_pair_cmt._cluster_src_fm,
            self.src_volume_dm,
            self.tgt_volume_dm
        )

        mock_get_volume_pair.assert_any_call(
            self.volume_pair_cmt._cluster_tgt_fm,
            self.tgt_volume_dm,
            self.src_volume_dm
        )

        assert mock_sleep.call_count == 1
        assert self.volume_pair_cmt._volume_pair_src_dm == volume_pairs[0]
        assert self.volume_pair_cmt._volume_pair_tgt_dm == volume_pairs[1]
        assert self.volume_pair_cmt._tgt_volume_original_access_mode is None

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.start_volume_pairing')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.complete_volume_pairing')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._reload')
    @mock.patch('time.sleep')
    def test_apply_multiple_failures(self, mock_sleep, mock_reload, mock_complete_pairing,
                                     mock_start_volume_pairing):
        mock_start_volume_pairing.return_value = 'foo-pair-key'
        mock_complete_pairing.side_effect = api_errors.InvalidPairingKey('foo', 'bar')
        mock_reload.return_value = False

        with pytest.raises(api_errors.InvalidPairingKey):
            self.volume_pair_cmt.apply()

        mock_start_volume_pairing.assert_called_with(
            self.src_volume_dm,
            mode=None
        )

        mock_complete_pairing.assert_called_with(
            self.tgt_volume_dm,
            'foo-pair-key',
        )

        assert mock_sleep.call_count == 1

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.remove_volume_pair')
    def test_reset(self, mock_remove_volume_pair):
        self.volume_pair_cmt._state = components.States.applied
        self.volume_pair_cmt.reset()

        assert mock_remove_volume_pair.call_count == 2
        mock_remove_volume_pair.assert_any_call(self.tgt_volume_dm)
        mock_remove_volume_pair.assert_any_call(self.src_volume_dm)

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.modify_volumes')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.remove_volume_pair')
    def test_reset_with_orig_access_mode(self, mock_remove_volume_pair, mock_modify_volumes):
        self.volume_pair_cmt._state = components.States.applied

        self.volume_pair_cmt._tgt_volume_original_access_mode = enums.VolumeAccess.read_write
        self.volume_pair_cmt.reset()

        assert mock_remove_volume_pair.call_count == 2
        mock_remove_volume_pair.assert_any_call(self.tgt_volume_dm)
        mock_remove_volume_pair.assert_any_call(self.src_volume_dm)
        mock_modify_volumes.assert_called_with([self.tgt_volume_dm],
                                               access=enums.VolumeAccess.read_write)

    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    def test_reload(self, mock_get_volume_pair):
        mock_get_volume_pair.side_effect = models.VolumePair.get_mock_object_list(2)

        assert self.volume_pair_cmt._reload() is True

    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    def test_reload_negative(self, mock_get_volume_pair):
        pair = models.VolumePair.get_mock_object()
        mock_get_volume_pair.side_effect = [pair, None]

        assert self.volume_pair_cmt._reload() is False

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.start_volume_pairing')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.complete_volume_pairing')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._reload')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    def test_negative_find_volume_pair_source(self, mock_get_volume_pair, mock_reload,
                                              mock_complete_pairing,
                                              mock_start_volume_pairing):
        mock_reload.return_value = False
        mock_get_volume_pair.side_effect = [
            None,
            models.VolumePair.get_mock_object()
        ]
        with pytest.raises(errors.ElementInvalidStateError):
            self.volume_pair_cmt.apply()

        mock_reload.assert_called()
        mock_start_volume_pairing.assert_called()
        mock_complete_pairing.assert_called()

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.start_volume_pairing')
    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.complete_volume_pairing')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._reload')
    @mock.patch('valence.vl_components._remote_replication.PairVolumes._get_volume_pair')
    def test_negative_find_volume_pair_target(self, mock_get_volume_pair, mock_reload,
                                              mock_complete_pairing,
                                              mock_start_volume_pairing):
        mock_reload.return_value = False
        mock_get_volume_pair.side_effect = [
            models.VolumePair.get_mock_object(),
            None
        ]
        with pytest.raises(errors.ElementInvalidStateError):
            self.volume_pair_cmt.apply()

        mock_reload.assert_called()
        mock_start_volume_pairing.assert_called()
        mock_complete_pairing.assert_called()

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.list_active_paired_volumes')
    def test_get_volume_pair_empty_list_active(self, mock_list_active_paired_volumes):
        target_vol = models.Volume.get_mock_object()
        source_vol = models.Volume.get_mock_object()
        mock_list_active_paired_volumes.return_value = []

        pair = vl_components.PairVolumes._get_volume_pair(functional_models.Cluster.get_mock_fm(),
                                                          source_vol, target_vol)

        assert pair is None

    @mock.patch('valence.functional_models._cluster.volume_mixin.VolumeMixin.list_active_paired_volumes')
    def test_get_volume_pair(self, mock_list_active_paired_volumes):
        target_vol = models.Volume.get_mock_object()
        mock_pair = models.VolumePair.get_mock_object(overrides={
            'remote_volume_id': target_vol.volume_id
        })
        source_vol = models.Volume.get_mock_object(overrides={
            'volume_pairs': [mock_pair]
        })

        mock_list_active_paired_volumes.return_value = [source_vol]

        pair = vl_components.PairVolumes._get_volume_pair(functional_models.Cluster.get_mock_fm(),
                                                          source_vol, target_vol)

        assert pair == mock_pair


class TestCreateGroupSnapshotReplicationComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm_one = models.ResourceCluster.get_mock_object()
        cls.cluster_dm_two = models.ResourceCluster.get_mock_object()
        cls.cluster_fm_one = functional_models.Cluster(cls.cluster_dm_one)
        cls.cluster_fm_two = functional_models.Cluster(cls.cluster_dm_two)
        cls.mock_group_snap_sync_dm = models.GroupSnapshot.get_mock_object(
            overrides={'remote_statuses': [
                {'remote_status': enums.RemoteClusterSnapshotStatus.syncing}]})
        cls.mock_group_snap_delete_dm = models.GroupSnapshot.get_mock_object(
            overrides={'remote_statuses': [
                {'remote_status': enums.RemoteClusterSnapshotStatus.deleted}]})
        cls.mock_group_snap_present_dm = models.GroupSnapshot.get_mock_object(
            overrides={'remote_statuses': [
                {'remote_status': enums.RemoteClusterSnapshotStatus.present}]})
        cls.mock_volume_dms = models.Volume.get_mock_object_list(5)

    def setUp(self):
        self.group_snapshot_cmt = vl_components.CreateGroupSnapshotReplication(
            self.cluster_dm_one,
            self.cluster_dm_two,
            volume_dms=self.mock_volume_dms,
            tag='hamberders',
        )

    @mock.patch('valence.functional_models.Cluster.create_group_snapshot')
    def test_apply(self, mock_create_group_snap):
        mock_create_group_snap.return_value = self.mock_group_snap_sync_dm
        self.group_snapshot_cmt.apply()
        mock_create_group_snap.assert_called_with(volume_dms=self.mock_volume_dms,
                                                  name='hamberders-TAG-',
                                                  enable_remote_replication=True,
                                                  retention=None,
                                                  attributes=None)

    @mock.patch('valence.functional_models.Cluster.create_group_snapshot')
    @mock.patch('valence.vl_components.CreateGroupSnapshotReplication._get_existing_group_snapshot')
    @mock.patch('valence.functional_models.Cluster.delete_group_snapshot')
    def test_reset(self, mock_delete_group, mock_get_existing, mock_create_group_snap):
        mock_create_group_snap.return_value = self.mock_group_snap_present_dm
        mock_get_existing.side_effect = [self.mock_group_snap_present_dm,
                                         self.mock_group_snap_present_dm,
                                         self.mock_group_snap_present_dm]
        self.group_snapshot_cmt.apply()
        self.group_snapshot_cmt.reset()
        mock_delete_group.assert_has_calls(
            [mock.call(self.mock_group_snap_present_dm), mock.call(self.mock_group_snap_present_dm)])

    @mock.patch('valence.functional_models.Cluster.create_group_snapshot')
    @mock.patch('valence.vl_components.CreateGroupSnapshotReplication._get_existing_group_snapshot')
    @mock.patch('valence.functional_models.Cluster.delete_group_snapshot')
    def test_reset_with_exception(self, mock_delete_group, mock_get_existing, mock_create_group_snap):
        mock_create_group_snap.return_value = self.mock_group_snap_present_dm
        mock_get_existing.side_effect = components.ComponentStateError('foo!')
        self.group_snapshot_cmt.apply()
        self.group_snapshot_cmt.reset()

        mock_delete_group.assert_not_called()

    @mock.patch('valence.functional_models.Cluster.list_group_snapshots')
    def test_get_existing_group_snapshot(self, mock_list_group_snaps):
        self.test_apply()
        mock_list_group_snaps.return_value = [self.mock_group_snap_sync_dm]
        group_snap_dm = self.group_snapshot_cmt._get_existing_group_snapshot(self.cluster_fm_one)
        assert group_snap_dm == self.mock_group_snap_sync_dm
        mock_list_group_snaps.assert_called_with(
            group_snapshot_id=self.mock_group_snap_sync_dm.group_snapshot_id)

    @mock.patch('valence.functional_models.Cluster.list_group_snapshots')
    def test_get_existing_group_snapshot_exception(self, mock_list_group_snaps):
        self.group_snapshot_cmt._group_snapshot_dm = self.mock_group_snap_sync_dm
        mock_list_group_snaps.return_value = [self.mock_group_snap_delete_dm]
        with pytest.raises(components.ComponentStateError):
            self.group_snapshot_cmt._get_existing_group_snapshot(self.cluster_fm_one)

    @mock.patch('valence.functional_models.Cluster.list_group_snapshots')
    def test_group_snapshot_dm_property(self, mock_list_group_snaps):
        mock_list_group_snaps.return_value = [self.mock_group_snap_sync_dm]
        self.test_apply()
        assert self.group_snapshot_cmt.group_snapshot_dm == self.mock_group_snap_sync_dm
