"""Test valence.vl_components._schedule."""
import datetime
import time
import unittest

import mock
import pytest
import pytz

from valence import (
    enums,
    functional_models,
    models,
)
from valence.vl_components import (
    _constants,
    _schedule,
)


class TestCreateSchedule(unittest.TestCase):
    """Unit tests for the Create Schedule component."""
    cluster_dm = models.ResourceCluster.get_mock_object()
    schedule_info = models.ScheduleInfo.get_mock_object()
    weekdays = [models.Weekday.get_mock_object()]
    cluster_constants = models.ClusterConstants.get_mock_object()
    schedule_name = 'scheduletest'
    schedule_type = enums.ScheduleType.snapshot
    recurring = False
    paused = True
    attributes = {}
    starting_date = datetime.datetime.utcnow().replace(tzinfo=pytz.utc)

    def setUp(self):
        self.create_schedule_cmt = _schedule.CreateSchedule(cluster_dm=self.cluster_dm,
                                                            schedule_name=self.schedule_name,
                                                            schedule_info=self.schedule_info,
                                                            schedule_type=self.schedule_type,
                                                            weekdays=self.weekdays,
                                                            recurring=self.recurring,
                                                            paused=self.paused,
                                                            attributes=self.attributes,
                                                            starting_date=self.starting_date
                                                            )

    def test_mutual_exclusive_weekdays_monthdays(self):
        schedule_name = 'scheduletest2'
        schedule_type = enums.ScheduleType.snapshot
        schedule_time = datetime.time(23, 59)
        monthdays = [1, 2]
        recurring = False
        paused = True
        attributes = {}
        with pytest.raises(ValueError):
            _schedule.CreateSchedule(cluster_dm=self.cluster_dm,
                                     schedule_name=schedule_name,
                                     schedule_info=self.schedule_info,
                                     schedule_type=schedule_type,
                                     schedule_time=schedule_time,
                                     monthdays=monthdays,
                                     weekdays=self.weekdays,
                                     recurring=recurring,
                                     paused=paused,
                                     attributes=attributes,
                                     starting_date=self.starting_date
                                     )

    def test_frequency_weekdays(self):
        schedule_name = 'scheduletest2'
        schedule_type = enums.ScheduleType.snapshot
        recurring = False
        paused = True
        attributes = {}
        test_schedule = _schedule.CreateSchedule(cluster_dm=self.cluster_dm,
                                                 schedule_name=schedule_name,
                                                 schedule_info=self.schedule_info,
                                                 schedule_type=schedule_type,
                                                 weekdays=self.weekdays,
                                                 recurring=recurring,
                                                 paused=paused,
                                                 attributes=attributes,
                                                 starting_date=self.starting_date)

        assert test_schedule.attributes['frequency'] == 'Days of Week'

    def test_frequency_monthdays(self):
        schedule_name = 'scheduletest2'
        schedule_type = enums.ScheduleType.snapshot
        monthdays = [1, 2]
        recurring = False
        paused = True
        attributes = {}
        test_schedule = _schedule.CreateSchedule(cluster_dm=self.cluster_dm,
                                                 schedule_name=schedule_name,
                                                 schedule_info=self.schedule_info,
                                                 schedule_type=schedule_type,
                                                 monthdays=monthdays,
                                                 recurring=recurring,
                                                 paused=paused,
                                                 attributes=attributes,
                                                 starting_date=self.starting_date)

        assert test_schedule.attributes['frequency'] == 'Days of Month'

    def test_frequency_time_interval(self):
        schedule_name = 'scheduletest2'
        schedule_type = enums.ScheduleType.snapshot
        schedule_time = datetime.time(23, 59)
        recurring = False
        paused = True
        attributes = {}
        test_schedule = _schedule.CreateSchedule(cluster_dm=self.cluster_dm,
                                                 schedule_name=schedule_name,
                                                 schedule_info=self.schedule_info,
                                                 schedule_type=schedule_type,
                                                 schedule_time=schedule_time,
                                                 recurring=recurring,
                                                 paused=paused,
                                                 attributes=attributes,
                                                 starting_date=self.starting_date)

        assert test_schedule.attributes['frequency'] == 'Time Interval'

    @mock.patch.object(functional_models._cluster.Cluster, 'create_schedule')
    def test_apply(self, mock_create_schedule):
        self.create_schedule_cmt.apply()
        mock_create_schedule.assert_called_with(schedule_name=self.schedule_name,
                                                schedule_info=self.schedule_info,
                                                schedule_type=self.schedule_type,
                                                schedule_time=None,
                                                monthdays=None,
                                                weekdays=self.weekdays,
                                                recurring=self.recurring,
                                                paused=self.paused,
                                                attributes=self.attributes,
                                                starting_date=self.starting_date)

    @mock.patch('valence.functional_models.Cluster.get_constants')
    @mock.patch('valence.functional_models.Cluster.set_constants')
    @mock.patch.object(time, 'sleep')
    @mock.patch.object(time, 'time')
    @mock.patch.object(functional_models._cluster.Cluster, 'create_schedule')
    @mock.patch.object(functional_models._cluster.Cluster, 'list_schedules')
    @mock.patch.object(functional_models._cluster.Cluster, 'modify_schedule')
    def test_reset(self, mock_modify_schedule, mock_list_schedules, mock_create_schedule,
                   mock_time, mock_sleep, mock_set_constant, mock_get_constant):
        mock_response = models.Schedule.get_mock_data_list(
            1,
            overrides=dict(schedule_name=self.schedule_name,
                           schedule_info=self.schedule_info,
                           schedule_type=self.schedule_type,
                           weekdays=self.weekdays,
                           recurring=self.recurring,
                           paused=self.paused,
                           attributes=self.attributes,
                           starting_date=self.starting_date
                           ))
        mock_create_schedule.return_value = mock_response
        self.create_schedule_cmt.apply()
        mock_time.side_effect = [300, 400]
        mock_list_schedules.return_value = []
        mock_set_constant.return_value = self.cluster_constants
        cluster_constants = models.ClusterConstants.create(scheduler_interval_seconds=30)
        set_constants_cmt = _constants.SetClusterConstants(self.cluster_dm, cluster_constants)
        set_constants_cmt.apply()
        mock_set_constant.assert_called_with(cluster_constants)
        set_constants_cmt.reset()
        self.create_schedule_cmt.reset()
        mock_modify_schedule.assert_called_with(schedule_dm=mock_response,
                                                to_be_deleted=True)

    @mock.patch('valence.functional_models.Cluster.get_constants')
    @mock.patch('valence.functional_models.Cluster.set_constants')
    @mock.patch.object(time, 'sleep')
    @mock.patch.object(time, 'time')
    @mock.patch.object(functional_models._cluster.Cluster, 'create_schedule')
    @mock.patch.object(functional_models._cluster.Cluster, 'list_schedules')
    @mock.patch.object(functional_models._cluster.Cluster, 'modify_schedule')
    def test_reset_with_schedules(self, mock_modify_schedule, mock_list_schedules, mock_create_schedule,
                                  mock_time, mock_sleep, mock_set_constant, mock_get_constant):
        mock_response = models.Schedule.get_mock_data_list(
            1,
            overrides=dict(schedule_name=self.schedule_name,
                           schedule_info=self.schedule_info,
                           schedule_type=self.schedule_type,
                           weekdays=self.weekdays,
                           recurring=self.recurring,
                           paused=self.paused,
                           attributes=self.attributes,
                           starting_date=self.starting_date
                           ))
        mock_create_schedule.return_value = mock_response
        self.create_schedule_cmt.apply()
        mock_time.side_effect = [300, 400]
        mock_list_schedules.return_value = [models.Schedule(x) for x in mock_response]
        mock_set_constant.return_value = self.cluster_constants
        cluster_constants = models.ClusterConstants.create(scheduler_interval_seconds=30)
        set_constants_cmt = _constants.SetClusterConstants(self.cluster_dm, cluster_constants)
        set_constants_cmt.apply()
        mock_set_constant.assert_called_with(cluster_constants)
        set_constants_cmt.reset()
        with pytest.raises(RuntimeError):
            try:
                self.create_schedule_cmt.reset()
            except StopIteration:  # done for python 2.7 to 3.7 compatibility in accordance with PEP 479
                raise RuntimeError
        mock_modify_schedule.assert_called_with(schedule_dm=mock_response,
                                                to_be_deleted=True)
