"""Test valence.vl_components._security."""
import time
import unittest

import mock
import pytest
import sf_batch
import sf_models
import sf_ssh

import sf_api
from valence import (
    components,
    enums,
    errors,
    functional_models,
    models,
    unittest_utils,
    vl_helpers,
)
from valence.vl_components import _security

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestBreakEarComponentIKM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.break_ear_cmt_corrupt = _security.BreakEar(self.cluster_dm)
        self.break_ear_cmt_move = _security.BreakEar(self.cluster_dm, corrupt=False)

    @mock.patch('valence.vl_helpers.corrupt_shamir_secret_keys')
    @mock.patch('valence.vl_helpers.move_shamir_secret_keys')
    @mock.patch('valence.vl_helpers.restore_shamir_secret_keys')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_apply_reset(self, mock_get_nodes, mock_restore, mock_move, mock_corrupt):
        mock_get_nodes.return_value = [functional_models.Node.get_mock_fm()]

        self.break_ear_cmt_corrupt.apply()
        mock_corrupt.assert_called()
        self.break_ear_cmt_move.apply()
        mock_move.assert_called()

        self.break_ear_cmt_corrupt.reset()
        mock_restore.assert_called()

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    def test_assert_valid(self, mock_get_ear_info, mock_get_ssh):
        mock_get_ear_info.return_value = models.EncryptionAtRestInfo.get_mock_object()
        run_safe = mock_get_ssh.return_value.__enter__.return_value.run_command
        run_safe.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=1)
        self.break_ear_cmt_corrupt.node_fms = [functional_models.Node.get_mock_fm()]
        self.break_ear_cmt_move.node_fms = [functional_models.Node.get_mock_fm()]

        with pytest.raises(AssertionError):
            self.break_ear_cmt_corrupt.assert_valid()

        run_safe.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        with pytest.raises(AssertionError):
            self.break_ear_cmt_move.assert_valid()

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    def test_assert_valid_passing(self, mock_get_ear_info, mock_get_ssh):
        mock_get_ear_info.return_value = models.EncryptionAtRestInfo.get_mock_object()
        run_safe = mock_get_ssh.return_value.__enter__.return_value.run_command
        run_safe.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        self.break_ear_cmt_corrupt.state = components.States.applied
        self.break_ear_cmt_corrupt.node_fms = [functional_models.Node.get_mock_fm()]
        self.break_ear_cmt_corrupt.assert_valid()

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    def test_assert_valid_applied(self, mock_get_ear_info, mock_get_ssh):
        mock_get_ear_info.return_value = models.EncryptionAtRestInfo.get_mock_object()
        run_safe = mock_get_ssh.return_value.__enter__.return_value.run_command
        run_safe.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=1)
        self.break_ear_cmt_corrupt.state = components.States.applied
        self.break_ear_cmt_corrupt.node_fms = [functional_models.Node.get_mock_fm()]
        self.break_ear_cmt_move.node_fms = [functional_models.Node.get_mock_fm()]

        with pytest.raises(AssertionError):
            self.break_ear_cmt_corrupt.assert_valid()

        run_safe.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        with pytest.raises(AssertionError):
            self.break_ear_cmt_move.assert_valid()


class TestBreakEarComponentEKM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        mock_eks = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.hytrust))
        self.eks_dms = [mock_eks]
        self.eks_dms_cluster = [sf_models.ExternalKeyServerCluster.get_mock_object(
            overrides=dict(nodes=[mock_eks, mock_eks]))]

        self.break_ear_cmt = _security.BreakEar(self.cluster_dm, eks_dms=self.eks_dms)
        self.break_ear_cmt_cluster = _security.BreakEar(self.cluster_dm, eks_dms=self.eks_dms_cluster)

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'disable')
    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'enable')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_apply_reset(self, mock_get_nodes, mock_enable, mock_disable):
        mock_get_nodes.return_value = [functional_models.Node.get_mock_fm()]

        self.break_ear_cmt.apply()
        mock_disable.assert_called()
        self.break_ear_cmt.reset()
        mock_enable.assert_called()

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'is_running')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    @mock.patch.object(functional_models.Cluster, 'list_key_servers_kmip')
    def test_assert_valid(self, mock_list_ks, mock_wait_for_fault_resolve, mock_wait_for_fault,
                          mock_is_running):
        mock_ks = []
        for eks in self.eks_dms:
            mock_ks.append(models.KeyServerKmip.get_mock_object(
                overrides=dict(kmip_key_server_hostnames=[eks.hostname],
                               kmip_assigned_provider_is_active=True)))
        mock_list_ks.return_value = mock_ks

        mock_is_running.return_value = False
        with pytest.raises(AssertionError):
            self.break_ear_cmt.assert_valid()

        self.break_ear_cmt.state = components.States.applied
        mock_is_running.return_value = True
        with pytest.raises(AssertionError):
            self.break_ear_cmt.assert_valid()

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'is_running')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    @mock.patch.object(functional_models.Cluster, 'list_key_servers_kmip')
    def test_assert_valid_inactive(self, mock_list_ks, mock_wait_for_fault_resolve, mock_wait_for_fault,
                                   mock_is_running):
        mock_ks = []
        for eks in self.eks_dms:
            mock_ks.append(models.KeyServerKmip.get_mock_object(
                overrides=dict(kmip_key_server_hostnames=[eks.hostname],
                               kmip_assigned_provider_is_active=False)))
        mock_list_ks.return_value = mock_ks

        mock_is_running.return_value = False
        with pytest.raises(AssertionError):
            self.break_ear_cmt.assert_valid()

        self.break_ear_cmt.state = components.States.applied
        mock_is_running.return_value = True
        with pytest.raises(AssertionError):
            self.break_ear_cmt.assert_valid()

    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_occurrence')
    @mock.patch.object(vl_helpers, 'wait_for_cluster_fault_resolution')
    def test_check_faults_timeout(self, mock_wait_for_fault_resolve, mock_wait_for_fault):
        mock_wait_for_fault.side_effect = [errors.TimeoutError]
        mock_wait_for_fault_resolve.side_effect = [errors.TimeoutError]
        errs = self.break_ear_cmt._check_faults(True)
        assert len(errs)
        errs = self.break_ear_cmt._check_faults(False)
        assert len(errs)


class TestEnableEarComponentIKM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.enable_ear_cmt = _security.EnableEar(self.cluster_dm)

    @mock.patch.object(sf_batch, 'SFThread')
    @mock.patch.object(sf_batch.SFThread, 'join')
    @mock.patch.object(sf_batch, 'check_for_error_and_raise_exception')
    @mock.patch.object(_security.EnableEar, '_reload')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_apply_async(self, mock_get_nodes, mock_reload, mock_join, mock_check, mock_sfthread):
        mock_get_nodes.return_value = [functional_models.Node.get_mock_fm()]
        mock_reload.return_value = False
        mock_thread = mock.MagicMock()
        mock_thread.is_stopped.side_effect = [False, True]
        mock_sfthread.return_value = mock_thread
        self.enable_ear_cmt.apply_async()
        mock_sfthread.assert_called()

        assert self.enable_ear_cmt.is_running() is True
        assert self.enable_ear_cmt.is_running() is False
        self.enable_ear_cmt.finish()

    @mock.patch('valence.vl_helpers.is_encryption_at_rest_enabled')
    @mock.patch('valence.vl_helpers.disable_encryption_at_rest')
    @mock.patch.object(_security.EnableEar, '_reload')
    @mock.patch.object(sf_batch, 'SFThread')
    @mock.patch.object(sf_batch, 'check_for_error_and_raise_exception')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_reset(self, mock_get_nodes, mock_check, mock_sfthread, mock_reload, mock_disable,
                   mock_is_enabled):
        mock_is_enabled.return_value = True
        mock_get_nodes.return_value = [functional_models.Node.get_mock_fm()]
        mock_reload.return_value = False
        mock_sfthread.return_value = mock.MagicMock()
        self.enable_ear_cmt._apply_is_running = mock.Mock(return_value=False)
        self.enable_ear_cmt.apply()
        self.enable_ear_cmt.reset()
        mock_sfthread.assert_called()

    @mock.patch('valence.functional_models.Cluster.get_cluster_info')
    def test_reload_success(self, mock_get_cluster_info):
        mock_get_cluster_info.return_value = models.Cluster.create(
            encryption_at_rest_state=enums.EarState.enabled)
        assert self.enable_ear_cmt._reload() is True

    @mock.patch('valence.functional_models.Cluster.get_cluster_info')
    def test_reload_fail(self, mock_get_cluster_info):
        mock_get_cluster_info.return_value = models.Cluster.create(
            encryption_at_rest_state=enums.EarState.disabled)
        assert self.enable_ear_cmt._reload() is False

    @mock.patch('valence.functional_models.Cluster.get_cluster_info')
    def test_reload_error(self, mock_get_cluster_info):
        mock_get_cluster_info.return_value = models.Cluster.create(
            encryption_at_rest_state=enums.EarState.enabling)
        with pytest.raises(components.PartialStateError):
            self.enable_ear_cmt._reload()

    @mock.patch('valence.vl_helpers.enable_encryption_at_rest')
    def test_enable_ear_with_ikm(self, mock_enable):
        self.enable_ear_cmt._enable_ear_with_ikm()
        mock_enable.assert_called()


class TestEnableEarComponentEKM(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.eks_dms = [sf_models.ExternalKeyServer.get_mock_object()]
        self.enable_ear_cmt = _security.EnableEar(self.cluster_dm, eks_dms=self.eks_dms)

    @mock.patch.object(sf_batch, 'SFThread')
    @mock.patch.object(sf_batch.SFThread, 'join')
    @mock.patch.object(sf_batch, 'check_for_error_and_raise_exception')
    @mock.patch.object(_security.EnableEar, '_reload')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_apply_async(self, mock_get_nodes, mock_reload, mock_join, mock_check, mock_sfthread):
        mock_get_nodes.return_value = [functional_models.Node.get_mock_fm()]
        mock_reload.return_value = False
        mock_thread = mock.MagicMock()
        mock_thread.is_stopped.side_effect = [False, True]
        mock_sfthread.return_value = mock_thread
        self.enable_ear_cmt.apply_async()
        mock_sfthread.assert_called()

        assert self.enable_ear_cmt.is_running() is True
        assert self.enable_ear_cmt.is_running() is False
        self.enable_ear_cmt.finish()

    @mock.patch('valence.vl_helpers.is_encryption_at_rest_enabled')
    @mock.patch('valence.vl_helpers.disable_encryption_at_rest')
    @mock.patch('valence.vl_helpers.teardown_external_key_management')
    @mock.patch.object(_security.EnableEar, '_reload')
    @mock.patch.object(sf_batch, 'SFThread')
    @mock.patch.object(sf_batch, 'check_for_error_and_raise_exception')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_reset(self, mock_get_nodes, mock_check, mock_sfthread, mock_reload, mock_teardown,
                   mock_disable, mock_is_enabled):
        mock_is_enabled.return_value = True
        mock_get_nodes.return_value = [functional_models.Node.get_mock_fm()]
        mock_reload.return_value = False
        mock_sfthread.return_value = mock.MagicMock()
        self.enable_ear_cmt._apply_is_running = mock.Mock(return_value=False)
        self.enable_ear_cmt.apply()
        self.enable_ear_cmt.reset()
        mock_sfthread.assert_called()

    @mock.patch.object(_security.EnableEar, '_recreate_provider_servers')
    @mock.patch('valence.vl_helpers.enable_encryption_at_rest')
    def test_enable_ear_with_ekm_recreate(self, mock_enable, mock_recreate):
        self.enable_ear_cmt.key_servers = [models.KeyServerKmip.get_mock_object()]
        self.enable_ear_cmt.key_provider = models.KeyProviderKmip.get_mock_object()
        self.enable_ear_cmt._enable_ear_with_ekm()
        mock_enable.assert_called()

    @mock.patch('valence.vl_helpers.setup_external_key_management')
    @mock.patch.object(functional_models.Cluster, 'list_key_servers_kmip')
    @mock.patch('valence.vl_helpers.enable_encryption_at_rest')
    def test_enable_ear_with_ekm(self, mock_enable, mock_list_ks, mock_setup):
        mock_list_ks.return_value = [models.KeyServerKmip.get_mock_object()]
        self.enable_ear_cmt._enable_ear_with_ekm()
        mock_enable.assert_called()

    @mock.patch.object(functional_models.Cluster, 'create_key_provider_kmip')
    @mock.patch.object(functional_models.Cluster, 'create_key_server_kmip')
    @mock.patch.object(functional_models.Cluster, 'list_key_servers_kmip')
    @mock.patch.object(functional_models.Cluster, 'add_key_server_to_provider_kmip')
    def test_recreate_provider_servers(self, mock_add_to_provider, mock_list, mock_create_ks,
                                       mock_create_kp):
        self.enable_ear_cmt.key_servers = [models.KeyServerKmip.get_mock_object()]
        self.enable_ear_cmt.key_provider = models.KeyProviderKmip.get_mock_object()
        new_mock_ks = models.KeyServerKmip.get_mock_object()
        mock_create_ks.return_value = new_mock_ks
        mock_create_kp.return_value = models.KeyProviderKmip.get_mock_object()
        mock_list.return_value = [new_mock_ks]
        self.enable_ear_cmt._recreate_provider_servers()


@mock.patch.object(functional_models.Cluster, 'list_active_nodes')
@mock.patch.object(functional_models.Cluster, 'list_drives')
@mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
@mock.patch.object(vl_helpers, 'wait_for_ear_state')
@mock.patch.object(vl_helpers, 'verify_shamir_secret_keys')
@mock.patch.object(vl_helpers, 'wait_for_drive_security_state')
@mock.patch.object(vl_helpers, 'wait_for_drive_locked_state')
class TestEnableEarAssertValid(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.eks_dms = [sf_models.ExternalKeyServer.get_mock_object()]
        self.enable_ear_cmt_ekm = _security.EnableEar(self.cluster_dm, eks_dms=self.eks_dms)
        self.enable_ear_cmt_ikm = _security.EnableEar(self.cluster_dm)

    def test_assert_valid_ear_state(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                    mock_wait_ear, mock_list_faults, mock_list_drives, mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )

        mock_wait_ear.side_effect = [errors.TimeoutError]
        with pytest.raises(errors.TimeoutError):
            self.enable_ear_cmt_ekm.assert_valid()

    def test_assert_valid_shamir(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                 mock_wait_ear, mock_list_faults, mock_list_drives, mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )

        mock_verify_shamir.return_value = False
        with pytest.raises(AssertionError):
            self.enable_ear_cmt_ikm.state = components.States.applied
            self.enable_ear_cmt_ikm.assert_valid()

    def test_assert_valid_security_state(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                         mock_wait_ear, mock_list_faults, mock_list_drives,
                                         mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )

        mock_verify_shamir.return_value = True
        mock_wait_security.side_effect = [errors.TimeoutError]

        with pytest.raises(errors.TimeoutError):
            self.enable_ear_cmt_ekm.assert_valid()

    def test_assert_valid_locked_state(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                       mock_wait_ear, mock_list_faults, mock_list_drives,
                                       mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )

        mock_verify_shamir.return_value = True
        mock_wait_locked.side_effect = [errors.TimeoutError]

        with pytest.raises(errors.TimeoutError):
            self.enable_ear_cmt_ekm.assert_valid()

    def test_assert_valid_drive_fault(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                      mock_wait_ear, mock_list_faults, mock_list_drives,
                                      mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason='anything'),
        )
        mock_verify_shamir.return_value = True

        with pytest.raises(AssertionError):
            self.enable_ear_cmt_ekm.assert_valid()

    def test_assert_valid_cluster_faults(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                         mock_wait_ear, mock_list_faults, mock_list_drives,
                                         mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )
        mock_verify_shamir.return_value = True

        mock_list_faults.return_value = models.ClusterFault.get_mock_object_list(
            length=1,
            overrides=dict(fault_code=enums.ClusterFaultCode.disable_drive_security_failed)
        )

        with pytest.raises(AssertionError):
            self.enable_ear_cmt_ekm.assert_valid()

    def test_assert_valid(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                          mock_wait_ear, mock_list_faults, mock_list_drives, mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )
        mock_verify_shamir.return_value = True
        mock_list_faults.return_value = []

        self.enable_ear_cmt_ekm.assert_valid()

    def test_assert_valid_timeout(self, mock_wait_locked, mock_wait_security, mock_verify_shamir,
                                  mock_wait_ear, mock_list_faults, mock_list_drives, mock_list_nodes):
        mock_list_nodes.return_value = models.ActiveNode.get_mock_object_list(length=1)
        mock_list_drives.return_value = models.Drive.get_mock_object_list(
            length=1,
            overrides=dict(status=enums.DriveStatus.active, drive_security_fault_reason=None),
        )
        mock_verify_shamir.return_value = True
        mock_list_faults.return_value = []

        with pytest.raises(errors.TimeoutError):
            self.enable_ear_cmt_ekm.assert_valid(timeout=0)


class TestEnableFipsHttpsComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()

    def setUp(self):
        self.enable_fips_cmt = _security.EnableFipsHttps(self.cluster_dm)

    @mock.patch.object(functional_models.Cluster, 'get_feature_status')
    @mock.patch.object(functional_models.Cluster, 'enable_feature')
    @mock.patch.object(sf_ssh, 'poll_ssh_query')
    def test_apply_status_disabled(self, mock_poll_ssh, mock_enable, mock_get_feature):
        mock_disabled = [sf_api.models.FeatureStatus.get_mock_object(
                overrides=dict(feature=enums.SfFeature.fips.value, enabled=False))
        ]

        mock_get_feature.return_value = mock_disabled
        with pytest.raises(components.ComponentStateError):
            self.enable_fips_cmt.apply()

    @mock.patch.object(functional_models.Cluster, 'get_feature_status')
    @mock.patch.object(functional_models.Cluster, 'enable_feature')
    @mock.patch.object(sf_ssh, 'poll_ssh_query')
    def test_apply_timeouts(self, mock_poll_ssh, mock_enable, mock_get_feature):
        mock_get_feature.return_value = [sf_api.models.FeatureStatus.get_mock_object(
                overrides=dict(feature=enums.SfFeature.fips.value, enabled=False))
        ]

        mock_poll_ssh.side_effect = [sf_ssh.errors.HostNotPingable]
        with pytest.raises(components.ComponentTimeoutError):
            self.enable_fips_cmt.apply()

        self.enable_fips_cmt.state = components.States.ready
        mock_poll_ssh.side_effect = [True, sf_ssh.errors.HostNotPingable]
        with pytest.raises(components.ComponentTimeoutError):
            self.enable_fips_cmt.apply()

    def test_reset(self):
        self.enable_fips_cmt.state = components.States.applied
        with pytest.raises(components.ComponentStateError):
            self.enable_fips_cmt._reset()


class TestCorruptCryptsoftKeyComponent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.eks_dm = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.cryptsoft))

    def setUp(self):
        self.corrupt_key_cmt = _security.CorruptCryptsoftKey(self.cluster_dm, self.eks_dm)

    @mock.patch.object(vl_helpers, 'is_encryption_at_rest_enabled')
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    @mock.patch.object(vl_helpers, 'corrupt_cryptsoft_key')
    def test_apply(self, mock_corrupt, mock_get_ear_info, mock_is_ear_enabled):
        mock_is_ear_enabled.return_value = True
        mock_key = 'zuq935ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_key_id = '680c8cec-ff90-4eeb-a890-36a2557176bd'

        mock_get_ear_info.return_value = models.EncryptionAtRestInfo.get_mock_object(
            overrides=dict(key_id=mock_key_id))
        mock_corrupt.return_value = mock_key
        self.corrupt_key_cmt.apply()
        assert self.corrupt_key_cmt.key == mock_key
        assert self.corrupt_key_cmt.key_id == mock_key_id

    @mock.patch.object(vl_helpers, 'is_encryption_at_rest_enabled')
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    def test_apply_vendor(self, mock_get_ear_info, mock_is_ear_enabled):
        mock_is_ear_enabled.return_value = True
        eks_dm = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.hytrust))
        corrupt_key_cmt = _security.CorruptCryptsoftKey(self.cluster_dm, eks_dm)
        with pytest.raises(components.ResourceStateError):
            corrupt_key_cmt.apply()

    @mock.patch.object(vl_helpers, 'is_encryption_at_rest_enabled')
    def test_apply_ear(self, mock_is_ear_enabled):
        mock_is_ear_enabled.return_value = False
        with pytest.raises(components.ResourceStateError):
            self.corrupt_key_cmt.apply()

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'set_key')
    def test_reset(self, mock_set_key):
        mock_key = 'zuq935ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_key_id = '680c8cec-ff90-4eeb-a890-36a2557176bd'
        self.corrupt_key_cmt.state = components.States.applied
        self.corrupt_key_cmt.key = mock_key
        self.corrupt_key_cmt.key_id = mock_key_id
        self.corrupt_key_cmt.reset()
