"""Test valence.vl_components._vlan."""
import unittest

import mock

from valence import (
    components,
    functional_models,
    models,
)
from valence.vl_components import _vlan


class TestCreateClusterVlan(unittest.TestCase):
    """Unit tests for the Create cluster VLAN Interface component."""
    cluster_dm = models.ResourceCluster.get_mock_object()

    address_blocks = [models.AddressBlocks.create(start='192.86.5.1', size=10),
                      models.AddressBlocks.create(start='192.86.5.50', size=20)]
    svip = '192.86.5.200'
    namespace = False
    netmask = '255.255.192.0'
    gateway = '10.0.1.254'
    virtual_network_tag = 2
    name = 'vlantest'

    def setUp(self):
        self.create_cluster_vlan_cmt = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

    @mock.patch.object(functional_models._cluster.Cluster, 'create_virtual_network')
    def test_apply(self, mock_create_vlan):
        self.create_cluster_vlan_cmt.apply()
        mock_create_vlan.assert_called_with(address_blocks=self.address_blocks,
                                            name=self.name,
                                            netmask=self.netmask,
                                            svip=self.svip,
                                            virtual_network_tag=self.virtual_network_tag,
                                            gateway=self.gateway,
                                            namespace=self.namespace)

    @mock.patch.object(functional_models._cluster.Cluster, 'delete_virtual_network')
    def test_reset(self, mock_delete_vlan):
        self.create_cluster_vlan_cmt._state = components.States.applied
        self.create_cluster_vlan_cmt.reset()
        mock_delete_vlan.assert_called_with(virtual_network_tag=self.virtual_network_tag)


class TestModifyClusterVlan(unittest.TestCase):
    """Unit tests for modify cluster VLAN component."""

    cluster_dm = models.ResourceCluster.get_mock_object()
    address_blocks = [models.AddressBlocks.create(start='192.86.5.1', size=10),
                      models.AddressBlocks.create(start='192.86.5.50', size=20)]
    svip = '192.86.5.200'
    namespace = False
    netmask = '255.255.192.0'
    gateway = '10.0.1.254'
    virtual_network_tag = 2
    name = 'vlantest'

    def setUp(self):
        self.modify_cluster_vlan_cmt = _vlan.ModifyClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

    @mock.patch.object(functional_models._cluster.Cluster, 'list_virtual_networks')
    @mock.patch.object(functional_models._cluster.Cluster, 'modify_virtual_network')
    def test_apply(self, mock_modify_vlan, mock_list_vlans):
        self.modify_cluster_vlan_cmt.apply()
        mock_list_vlans.assert_called_with(virtual_network_tags=[self.virtual_network_tag])
        mock_modify_vlan.assert_called_with(address_blocks=self.address_blocks,
                                            name=self.name,
                                            netmask=self.netmask,
                                            svip=self.svip,
                                            virtual_network_tag=self.virtual_network_tag,
                                            gateway=self.gateway,
                                            namespace=self.namespace)

    @mock.patch.object(functional_models._cluster.Cluster, 'modify_virtual_network')
    def test_reset_without_none(self, mock_modify_vlan):
        address_blocks = [models.AddressBlocks.create(start='192.86.5.1', size=10),
                          models.AddressBlocks.create(start='192.86.5.50', size=20)]
        virtual_networks = [
            VirtualNetwork(name='vlantest', address_blocks=address_blocks, virtual_network_tag=2,
                           namespace=False, gateway='10.0.1.254', netmask='255.255.192.0',
                           svip='192.86.5.200')
        ]

        self.modify_cluster_vlan_cmt.original_virtual_networks = virtual_networks
        self.modify_cluster_vlan_cmt._state = components.States.applied
        self.modify_cluster_vlan_cmt.reset()

        mock_modify_vlan.assert_called_with(address_blocks=self.address_blocks,
                                            name=self.name,
                                            netmask=self.netmask,
                                            svip=self.svip,
                                            virtual_network_tag=self.virtual_network_tag,
                                            gateway=self.gateway,
                                            namespace=self.namespace)


class TestManageVlan(unittest.TestCase):

    cluster_dm = models.ResourceCluster.get_mock_object()

    address_blocks = [models.AddressBlocks.create(start='192.86.5.1', size=10),
                      models.AddressBlocks.create(start='192.86.5.50', size=20)]
    svip = '192.86.5.200'
    namespace = False
    netmask = '255.255.192.0'
    gateway = '10.0.1.254'
    virtual_network_tag = 2
    name = 'vlantest'

    def setUp(self):

        self.create_cluster_vlan_cmt1 = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        self.create_cluster_vlan_cmt2 = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        self.create_cluster_vlan_cmt3 = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        cluster_list = [self.create_cluster_vlan_cmt1,
                        self.create_cluster_vlan_cmt2,
                        self.create_cluster_vlan_cmt3]

        self.manage_vlan_cmt = _vlan.ManageVlan(
            vlan_cluster_cmts=cluster_list
        )

    @mock.patch.object(functional_models._cluster.Cluster, 'create_virtual_network')
    def test_apply(self, mock_create_vlan):

        self.manage_vlan_cmt.apply()
        self.assertEqual(mock_create_vlan.call_count, 3)

    @mock.patch.object(functional_models._cluster.Cluster, 'delete_virtual_network')
    @mock.patch.object(functional_models._cluster.Cluster, 'create_virtual_network')
    def test_reset(self, mock_create_vlan, mock_delete_vlan):

        self.manage_vlan_cmt.apply()
        self.manage_vlan_cmt.reset()
        self.assertEqual(mock_delete_vlan.call_count, 3)


class VirtualNetwork(object):
    def __init__(self, name=None, svip=None, virtual_network_tag=None, address_blocks=None,
                 namespace=None, netmask=None, gateway=None):
        self.name = name
        self.svip = svip
        self.virtual_network_tag = virtual_network_tag
        self.address_blocks = address_blocks
        self.namespace = namespace
        self.netmask = netmask
        self.gateway = gateway


class TestsManageMultipleVlans(unittest.TestCase):
    """Unit tests for Mange Multiple VLANs component."""

    cluster_dm = models.ResourceCluster.get_mock_object()

    address_blocks = [models.AddressBlocks.create(start='192.86.5.1', size=10),
                      models.AddressBlocks.create(start='192.86.5.50', size=20)]
    svip = '192.86.5.200'
    namespace = False
    netmask = '255.255.192.0'
    gateway = '10.0.1.254'
    virtual_network_tag = 2
    name = 'vlantest'

    def setUp(self):

        self.create_cluster_vlan_cmt1 = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        self.create_cluster_vlan_cmt2 = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        self.create_cluster_vlan_cmt3 = _vlan.CreateClusterVlan(
            name=self.name,
            cluster_dm=self.cluster_dm,
            address_blocks=self.address_blocks,
            netmask=self.netmask,
            svip=self.svip,
            virtual_network_tag=self.virtual_network_tag,
            gateway=self.gateway,
            namespace=self.namespace)

        self.create_cluster_vlan_cmts1 = [self.create_cluster_vlan_cmt1]
        self.create_cluster_vlan_cmts2 = [self.create_cluster_vlan_cmt2]
        self.create_cluster_vlan_cmts3 = [self.create_cluster_vlan_cmt3]
        self.manage_vlan_cmt1 = _vlan.ManageVlan(self.create_cluster_vlan_cmts1)
        self.manage_vlan_cmt2 = _vlan.ManageVlan(self.create_cluster_vlan_cmts2)
        self.manage_vlan_cmt3 = _vlan.ManageVlan(self.create_cluster_vlan_cmts3)
        self.vlan_cmts = [self.manage_vlan_cmt1, self.manage_vlan_cmt2, self.manage_vlan_cmt3]
        self.manage_mult_vlans_cmt = _vlan.ManageMultipleVlans(self.vlan_cmts)

    def test_init(self):
        assert self.manage_mult_vlans_cmt.child_cmts == self.vlan_cmts

    @mock.patch.object(functional_models._cluster.Cluster, 'create_virtual_network')
    def test_apply(self, mock_create_vlan):

        self.manage_mult_vlans_cmt.apply()
        self.assertEqual(mock_create_vlan.call_count, 3)

    @mock.patch.object(functional_models._cluster.Cluster, 'delete_virtual_network')
    @mock.patch.object(functional_models._cluster.Cluster, 'create_virtual_network')
    def test_reset(self, mock_create_vlan, mock_delete_vlan):

        self.manage_mult_vlans_cmt.apply()
        self.manage_mult_vlans_cmt.reset()
        self.assertEqual(mock_delete_vlan.call_count, 3)
