from __future__ import division

import unittest

import mock
import pytest

from valence import (
    components,
    enums,
    functional_models,
    models,
)
from valence.vl_components import (
    _account,
    _initiator,
    _volume,
    _volume_access_group,
)


class TestCreateVolumeAccessGroup(unittest.TestCase):
    def setUp(self):
        # define the data models
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        self.initiator_client_dms = [models.ResourceClient.get_mock_object() for _ in range(2)]
        self.extra_client_dms = [models.ResourceClient.get_mock_object() for _ in range(3)]
        self.account_1 = models.Account.create(
            username='user1', status='active', account_id=1)
        self.account_2 = models.Account.create(
            username='user2', status='active', account_id=2)

        self.volume_dms = [
            models.Volume.create(
                account_id=1,
                volume_id=i,
                total_size=models.DataSize(i, enums.DataUnit.gibibytes),
                qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
                enable512e=False
            ) for i in range(1, 3)
        ]

        # We have to know the initiator ids before we create them since they
        # are used in the vag and the vag is used in the initiator...
        self.initiator_1_id = 1
        self.initiator_2_id = 2
        self.initiator_1_name = 'init1'
        self.initiator_2_name = 'init2'

        self.vag_dm = models.VolumeAccessGroup.get_mock_object(overrides=dict(
            name='test1',
            volume_access_group_id=1,
            volumes=[1, 2],
            initiators=[self.initiator_1_name, self.initiator_2_name],
            initiator_ids=[self.initiator_1_id, self.initiator_2_id],
        ))

        self.vag_extra_dm = models.VolumeAccessGroup.get_mock_object(overrides=dict(
            name='test2',
            volume_access_group_id=2,
            volumes=[3],
        ))

        self.vag_dms = [self.vag_dm, self.vag_extra_dm]

        self.initiator_1 = models.Initiator.get_mock_object(overrides=dict(
            alias=self.initiator_client_dms[0].client_name + '-rand_init',
            initiator_name=self.initiator_1_name,
            initiator_id=self.initiator_1_id,
        ))
        self.initiator_2 = models.Initiator.get_mock_object(overrides=dict(
            alias=self.initiator_client_dms[1].client_name + '-rand_init',
            initiator_name=self.initiator_2_name,
            initiator_id=self.initiator_2_id,
        ))
        self.initiator_map = {
            self.initiator_client_dms[0].mip: [self.initiator_1],
            self.initiator_client_dms[1].mip: [self.initiator_2],
        }
        self.initiator_dms = [self.initiator_1, self.initiator_2]

        # define the components
        self.accounts_cmt = _account.MockAccounts(
            tag='mock',
            account_dms=[self.account_1, self.account_2]
        )

        self.volumes_cmt = _volume.MockVolumes(
            tag='mock',
            cluster_dm=self.cluster_dm,
            volume_dms=self.volume_dms,
            accounts_cmt=self.accounts_cmt
        )

        self.initiators_cmt = _initiator.MockInitiators(
            tag='mock',
            client_dms=self.initiator_client_dms,
            initiator_dms_map=self.initiator_map
        )

        self.vag_cmt = _volume_access_group.CreateVolumeAccessGroup(
            name='test1',
            cluster_dm=self.cluster_dm,
            initiators_cmt=self.initiators_cmt,
            volumes_cmt=self.volumes_cmt,
            client_dms=self.initiator_client_dms,
        )

        self.mock_vag_cmt = _volume_access_group.MockVolumeAccessGroup(
            self.vag_dm,
            initiators_cmt=self.initiators_cmt,
            volumes_cmt=self.volumes_cmt,
            client_dms=self.initiator_client_dms,
        )

    @mock.patch.object(functional_models.Cluster, 'create_volume_access_group')
    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_init(self, mock_list, mock_create):
        # Instantiate with clients but no initiators
        with pytest.raises(ValueError):
            vag_cmt = _volume_access_group.CreateVolumeAccessGroup(
                name='test1',
                cluster_dm=self.cluster_dm,
                volumes_cmt=self.volumes_cmt,
                client_dms=self.initiator_client_dms,
            )

        # Instantiate with clients not associated with any initiators
        with pytest.raises(ValueError):
            vag_cmt = _volume_access_group.CreateVolumeAccessGroup(
                name='test2',
                cluster_dm=self.cluster_dm,
                initiators_cmt=self.initiators_cmt,
                volumes_cmt=self.volumes_cmt,
                client_dms=self.extra_client_dms,
            )

    @mock.patch.object(functional_models.Cluster, 'create_volume_access_group')
    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_apply(self, mock_list, mock_create):
        mock_list.side_effect = [[], self.vag_dms, self.vag_dms]
        mock_create.return_value = self.vag_dms

        with pytest.raises(components.ComponentStateError):
            vag = self.vag_cmt.volume_access_group_dm

        self.vag_cmt.apply()
        assert mock_list.call_count == 2
        assert mock_create.call_count == 1
        assert self.vag_cmt.volume_access_group_dm == self.vag_dm

    @mock.patch.object(functional_models.Cluster, 'create_volume_access_group')
    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_reload(self, mock_list, mock_create):
        mock_list.return_value = self.vag_dms
        self.vag_cmt.apply()
        assert mock_list.call_count == 1
        assert mock_create.call_count == 0
        assert self.vag_cmt.volume_access_group_dm == self.vag_dm

    @mock.patch.object(functional_models.Cluster, 'delete_volume_access_group')
    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_reset(self, mock_list, mock_delete):
        mock_list.return_value = self.vag_dms
        self.vag_cmt.apply()

        self.vag_cmt.reset()
        assert mock_delete.call_count == 1
        assert self.vag_cmt._volume_access_group is None

    @mock.patch.object(functional_models.Cluster, 'delete_volume_access_group')
    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_reset_no_action(self, mock_list, mock_delete):
        mock_list.return_value = self.vag_dms
        self.vag_cmt.apply()
        mock_list.return_value = []

        self.vag_cmt.reset()
        assert mock_delete.call_count == 0
        assert self.vag_cmt._volume_access_group is None

    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_assert_valid(self, mock_list):
        mock_list.return_value = self.vag_dms
        self.vag_cmt.apply()

        # Assert valid with valid vag
        self.vag_cmt.assert_valid()
        self.vag_cmt._assert_valid(self.vag_dm)

        # Assert valid with invalid vag passed in
        with pytest.raises(components.PartialStateError):
            self.vag_cmt._assert_valid(self.vag_extra_dm)

        # Assert valid while not applied
        self.vag_cmt._state = components.States.ready
        with pytest.raises(components.ComponentStateError):
            self.vag_cmt.assert_valid()

        # Assert valid with no valid vags found
        mock_list.return_value = [self.vag_extra_dm]
        self.vag_cmt._state = components.States.applied
        with pytest.raises(components.ComponentStateError):
            self.vag_cmt._assert_valid()

    @mock.patch.object(functional_models.Cluster, 'list_volume_access_groups')
    def test_get_vag_dm(self, mock_list):
        mock_list.return_value = self.vag_dms

        # Get vags while not applied
        with pytest.raises(components.ComponentStateError):
            self.vag_cmt.get_volume_access_group_dm()

        # Get vags while applied
        self.vag_cmt._state = components.States.applied
        assert self.vag_cmt.get_volume_access_group_dm() == self.vag_dm

        # Get vags with no valid vags found
        mock_list.return_value = [self.vag_dm, self.vag_dm]
        with pytest.raises(components.ComponentStateError):
            self.vag_cmt.get_volume_access_group_dm()

    def test_volume_dms(self):
        # Access volumes property while not applied
        with pytest.raises(components.ComponentStateError):
            volumes = self.vag_cmt.volume_dms

        # Access volumes property while applied
        self.vag_cmt._state = components.States.applied
        self.volumes_cmt.apply()
        assert self.vag_cmt.volume_dms == self.volume_dms

        # Access volumes property with no volumes component
        self.vag_cmt.volumes_cmt = None
        assert self.vag_cmt.volume_dms == []

    def test_client_dms(self):
        # Access clients property with both clients and initiators specified
        assert self.vag_cmt.client_dms == self.initiator_client_dms

        # Access clients property with just initiators specified
        self.vag_cmt._client_dms = None
        assert self.vag_cmt.client_dms == self.initiator_client_dms

        # Access clients property with neither clients nor initiators specified
        self.vag_cmt.initiators_cmt = None
        assert self.vag_cmt.client_dms == []

    def test_initiator_dms_map(self):
        # Access initiator map while not applied
        with pytest.raises(components.ComponentStateError):
            map = self.vag_cmt.initiator_dms_map

        # Access initiator map with both clients and initiators specified
        self.vag_cmt._state = components.States.applied
        self.volumes_cmt.apply()
        self.initiators_cmt.apply()
        assert self.vag_cmt.initiator_dms_map == self.initiator_map

        # Access initiator map with just initiators specified
        self.vag_cmt._client_dms = None
        assert self.vag_cmt.initiator_dms_map == self.initiator_map

        # Access initiator map with neither clients nor initiators specified
        self.vag_cmt.initiators_cmt = None
        assert self.vag_cmt.initiator_dms_map == {}

    def test_mock_apply(self):
        self.mock_vag_cmt.apply()
        assert self.mock_vag_cmt.state is components.States.applied

    def test_mock_reset(self):
        self.mock_vag_cmt.ensure_applied()
        assert self.mock_vag_cmt.state is components.States.applied

        self.mock_vag_cmt.reset()
        assert self.mock_vag_cmt.state is components.States.ready
        self.mock_vag_cmt.ensure_ready()
        assert self.mock_vag_cmt.state is components.States.ready

    def test_mock_get(self):
        with self.assertRaises(components.ComponentStateError):
            self.mock_vag_cmt.volume_dms

        with self.assertRaises(components.ComponentStateError):
            self.mock_vag_cmt.volume_access_group_dm

        with self.assertRaises(components.ComponentStateError):
            self.mock_vag_cmt.initiator_dms_map

        with self.assertRaises(components.ComponentStateError):
            self.mock_vag_cmt.get_volume_access_group_dm()

        self.mock_vag_cmt.apply()
        self.mock_vag_cmt.assert_valid()

        assert self.mock_vag_cmt.get_volume_access_group_dm() == self.vag_dm
        assert self.mock_vag_cmt.volume_access_group_dm == self.vag_dm
        assert self.mock_vag_cmt.initiator_dms_map == self.initiator_map

        assert self.mock_vag_cmt.volume_dms == self.volume_dms
        self.mock_vag_cmt.volumes_cmt = None
        assert self.mock_vag_cmt.volume_dms == []

        self.mock_vag_cmt._client_dms = [self.initiator_client_dms[0]]
        assert self.mock_vag_cmt.client_dms == [self.initiator_client_dms[0]]
        self.mock_vag_cmt._client_dms = None
        assert self.mock_vag_cmt.client_dms == self.initiator_client_dms
        self.mock_vag_cmt.initiators_cmt = None
        assert self.mock_vag_cmt.client_dms == []
