from __future__ import division

import copy
import unittest

import mock
import pytest

from sf_api import result_models
from valence import (
    components,
    enums,
    functional_models,
    models,
    selectors,
    vl_components,
    vl_helpers,
)


def sort_vols(volume_dms):
    return sorted(volume_dms, key=lambda dm: dm.volume_id)


class TestCreateVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.accounts_cmt = vl_components.CreateAccounts.mock('mock', cls.cluster_dm, 2)

        cls.partial_volume_dms = [models.Volume.create(
            total_size=models.DataSize(10 * i, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=1000 + i, max_iops=12000, burst_iops=20000)
        ) for i in range(3)]
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            volume_id=i,
            account_id=cls.accounts_cmt._account_dms[0].account_id,
            total_size=cls.partial_volume_dms[i].total_size,
            qos=cls.partial_volume_dms[i].qos
        ) for i in range(3)]

    def setUp(self):
        self.accounts_cmt.ensure_ready()

    def test_init(self):
        invalid_partial_volume_dms = [dm for dm in self.partial_volume_dms]
        invalid_partial_volume_dms.append(models.Account.create(username='dummy'))
        with pytest.raises(TypeError):
            volumes_cmt = vl_components.CreateVolumes(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                partial_volume_dms=invalid_partial_volume_dms,
                accounts_cmt=self.accounts_cmt
            )

        invalid_partial_volume_dms = copy.deepcopy(self.partial_volume_dms)
        invalid_partial_volume_dms[-1].total_size = None
        with pytest.raises(components.ResourceStateError):
            volumes_cmt = vl_components.CreateVolumes(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                partial_volume_dms=invalid_partial_volume_dms,
                accounts_cmt=self.accounts_cmt
            )

        volumes_cmt = vl_components.CreateVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            partial_volume_dms=self.partial_volume_dms,
            accounts_cmt=self.accounts_cmt
        )
        assert volumes_cmt.child_cmts == [volumes_cmt]
        assert volumes_cmt.partial_volume_dms == self.partial_volume_dms
        allocated_space = sum(dm.total_size.bytes for dm in self.partial_volume_dms)
        assert volumes_cmt.get_allocated_space() == allocated_space

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_components.CreateVolumes, '_assert_valid')
    @mock.patch.object(vl_helpers, 'create_volumes_individual_profiles')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_apply(self, list_volumes, create_volumes, assert_valid, delete_volumes):
        list_volumes.side_effect = [[], self.volume_dms, self.volume_dms]
        volumes_cmt = vl_components.CreateVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            partial_volume_dms=self.partial_volume_dms,
            accounts_cmt=self.accounts_cmt
        )

        volumes_cmt.apply()
        assert self.accounts_cmt.state == components.States.applied
        assert create_volumes.called
        call_volume_dms = copy.deepcopy(self.volume_dms)
        for dm in call_volume_dms:
            dm.volume_id = None
        assert create_volumes.call_args[1]['volume_dms'] == call_volume_dms
        assert assert_valid.called

        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[0]]
        assert volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[0]]
        assert sort_vols(volumes_cmt.volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(volumes_cmt.get_volume_dms()) == sort_vols(self.volume_dms)

        list_volumes.side_effect = [[]]
        volumes_cmt.reset()
        assert delete_volumes.called
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.account_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_volume_dms()

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_helpers, 'create_volumes_individual_profiles')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_reload(self, list_volumes, create_volumes, delete_volumes):
        volume_dms = copy.deepcopy(self.volume_dms)
        for dm in volume_dms:
            dm.account_id = self.accounts_cmt._account_dms[1].account_id

        self.accounts_cmt.apply()
        invalid_volume_dms = copy.deepcopy(volume_dms)
        invalid_volume_dms[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volume_dms
        volumes_cmt = vl_components.CreateVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            partial_volume_dms=self.partial_volume_dms,
            accounts_cmt=self.accounts_cmt,
            account_index=1
        )

        with pytest.raises(components.PartialStateError):
            volumes_cmt.apply()
        volumes_cmt.reset()
        assert sort_vols(delete_volumes.call_args[0][1]) == sort_vols(invalid_volume_dms)

        list_volumes.return_value = volume_dms
        volumes_cmt.apply()
        assert not create_volumes.called
        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[1]]
        assert volumes_cmt.volume_dms == volume_dms

    @mock.patch.object(vl_helpers, 'create_volumes_individual_profiles')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_assert_valid(self, list_volumes, create_volumes):
        list_volumes.return_value = self.volume_dms
        volumes_cmt = vl_components.CreateVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            partial_volume_dms=self.partial_volume_dms,
            accounts_cmt=self.accounts_cmt
        )
        with pytest.raises(components.ComponentStateError):
            volumes_cmt.assert_valid()

        volumes_cmt.apply()
        volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)
        invalid_volume_dms[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)
        invalid_volume_dms[0].qos.max_iops = 11000
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)
        invalid_volume_dms[-1].qos.burst_iops = 20001
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)
        invalid_volume_dms[0].account_id = 99
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)[:-1]
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

    def test_mock(self):
        volumes_cmt = vl_components.CreateVolumes.mock(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            partial_volume_dms=self.partial_volume_dms,
            accounts_cmt=self.accounts_cmt
        )
        assert volumes_cmt.child_cmts == [volumes_cmt]

        volumes_cmt.apply()
        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[0]]
        assert volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[0]]
        assert len(volumes_cmt.volume_dms) == len(self.volume_dms)
        assert len(set(dm.volume_id for dm in volumes_cmt.volume_dms)) == len(self.volume_dms)
        assert [dm.name.startswith(volumes_cmt.full_tag) for dm in volumes_cmt.volume_dms]
        assert [
            volumes_cmt.volume_dms[i].total_size == self.volume_dms[i].total_size
            for i in range(len(self.volume_dms))
        ]
        assert [
            volumes_cmt.volume_dms[i].qos == self.volume_dms[i].qos
            for i in range(len(self.volume_dms))
        ]

        volumes_cmt.reset()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_volume_dms()


class TestCreateMultipleVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.accounts_cmt = vl_components.CreateAccounts.mock('mock', cls.cluster_dm, 2)

        cls.partial_volume_dm = models.Volume.create(
            total_size=models.DataSize(10, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=1000, max_iops=12000, burst_iops=20000)
        )
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            volume_id=i,
            account_id=cls.accounts_cmt._account_dms[0].account_id,
            total_size=cls.partial_volume_dm.total_size,
            qos=cls.partial_volume_dm.qos,
        ) for i in range(3)]

    def setUp(self):
        self.accounts_cmt.ensure_ready()

    def test_init(self):
        with pytest.raises(TypeError):
            volumes_cmt = vl_components.CreateMultipleVolumes(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                vol_count=3,
                partial_volume_dm=models.Account.create(username='mock'),
                accounts_cmt=self.accounts_cmt
            )

        invalid_partial_volume_dm = copy.deepcopy(self.partial_volume_dm)
        invalid_partial_volume_dm.total_size = None
        with pytest.raises(components.ResourceStateError):
            volumes_cmt = vl_components.CreateMultipleVolumes(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                vol_count=3,
                partial_volume_dm=invalid_partial_volume_dm,
                accounts_cmt=self.accounts_cmt
            )

        volumes_cmt = vl_components.CreateMultipleVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            vol_count=3,
            partial_volume_dm=self.partial_volume_dm,
            accounts_cmt=self.accounts_cmt
        )
        assert volumes_cmt.child_cmts == [volumes_cmt]
        assert volumes_cmt.partial_volume_dms == [self.partial_volume_dm] * 3
        allocated_space = 3 * self.partial_volume_dm.total_size.bytes
        assert volumes_cmt.get_allocated_space() == allocated_space

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_components.CreateMultipleVolumes, '_assert_valid')
    @mock.patch.object(vl_helpers, 'create_volumes_single_profile')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_apply(self, list_volumes, create_volumes, assert_valid, delete_volumes):
        list_volumes.side_effect = [[], self.volume_dms, self.volume_dms]
        volumes_cmt = vl_components.CreateMultipleVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            vol_count=3,
            partial_volume_dm=self.partial_volume_dm,
            accounts_cmt=self.accounts_cmt
        )

        volumes_cmt.apply()
        assert self.accounts_cmt.state == components.States.applied
        assert create_volumes.called
        assert create_volumes.call_args[1]['volume_dm'] == models.Volume.create(
            name='unittest-TAG-',
            account_id=self.accounts_cmt._account_dms[0].account_id,
            total_size=self.partial_volume_dm.total_size,
            qos=self.partial_volume_dm.qos
        )
        assert assert_valid.called

        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[0]]
        assert volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[0]]
        assert sort_vols(volumes_cmt.volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(volumes_cmt.get_volume_dms()) == sort_vols(self.volume_dms)

        list_volumes.side_effect = [[]]
        volumes_cmt.reset()
        assert delete_volumes.called
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.account_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_volume_dms()

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_helpers, 'create_volumes_single_profile')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_reload(self, list_volumes, create_volumes, delete_volumes):
        volume_dms = copy.deepcopy(self.volume_dms)
        for dm in volume_dms:
            dm.account_id = self.accounts_cmt._account_dms[1].account_id

        invalid_volume_dms = copy.deepcopy(volume_dms)
        invalid_volume_dms[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volume_dms
        volumes_cmt = vl_components.CreateMultipleVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            vol_count=3,
            partial_volume_dm=self.partial_volume_dm,
            accounts_cmt=self.accounts_cmt,
            account_index=1
        )

        with pytest.raises(components.PartialStateError):
            volumes_cmt.apply()
        volumes_cmt.reset()
        assert sort_vols(delete_volumes.call_args[0][1]) == sort_vols(invalid_volume_dms)

        list_volumes.return_value = volume_dms
        volumes_cmt.apply()
        assert not create_volumes.called
        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[1]]
        assert volumes_cmt.volume_dms == volume_dms

    @mock.patch.object(vl_helpers, 'create_volumes_single_profile')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_assert_valid(self, list_volumes, create_volumes):
        list_volumes.return_value = self.volume_dms
        volumes_cmt = vl_components.CreateMultipleVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            vol_count=3,
            partial_volume_dm=self.partial_volume_dm,
            accounts_cmt=self.accounts_cmt
        )
        with pytest.raises(components.ComponentStateError):
            volumes_cmt.assert_valid()

        volumes_cmt.apply()
        volumes_cmt.assert_valid()

        invalid_volumes = copy.deepcopy(self.volume_dms)
        invalid_volumes[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volumes
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volumes = copy.deepcopy(self.volume_dms)
        invalid_volumes[0].account_id = 99
        list_volumes.return_value = invalid_volumes
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)[:-1]
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

    def test_mock(self):
        volumes_cmt = vl_components.CreateMultipleVolumes.mock(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            vol_count=3,
            partial_volume_dm=self.partial_volume_dm,
            accounts_cmt=self.accounts_cmt
        )
        assert volumes_cmt.child_cmts == [volumes_cmt]

        volumes_cmt.apply()
        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[0]]
        assert volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[0]]
        assert len(volumes_cmt.volume_dms) == len(self.volume_dms)
        assert len(set(dm.volume_id for dm in volumes_cmt.volume_dms)) == len(self.volume_dms)
        assert [dm.name.startswith(volumes_cmt.full_tag) for dm in volumes_cmt.volume_dms]
        assert [
            volumes_cmt.volume_dms[i].total_size == self.volume_dms[i].total_size
            for i in range(len(self.volume_dms))
        ]
        assert [
            volumes_cmt.volume_dms[i].qos == self.volume_dms[i].qos
            for i in range(len(self.volume_dms))
        ]

        volumes_cmt.reset()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_volume_dms()


class TestCloneVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.accounts_cmt = vl_components.CreateAccounts.mock('mock', cls.cluster_dm, 2)

        cls.partial_volume_dm = models.Volume.create(
            total_size=models.DataSize(10, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=1000, max_iops=12000, burst_iops=20000)
        )
        cls.volumes_cmt = vl_components.CreateMultipleVolumes.mock(
            tag='mock',
            cluster_dm=cls.cluster_dm,
            vol_count=3,
            partial_volume_dm=cls.partial_volume_dm,
            accounts_cmt=cls.accounts_cmt
        )

        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            volume_id=i,
            account_id=cls.accounts_cmt._account_dms[0].account_id,
            total_size=cls.partial_volume_dm.total_size,
            qos=cls.partial_volume_dm.qos,
            attributes={
                vl_components.CloneVolumes.CLONE_VOLUME_ID: cls.volumes_cmt._volume_dms[i].volume_id
            }
        ) for i in range(3)]

        cls.mock_async_handlers = [mock.MagicMock() for _ in range(3)]
        for i in range(3):
            cls.mock_async_handlers[i].is_running.side_effect = [True, True, False]
            cls.mock_async_handlers[i].get_models.return_value = [cls.volume_dms[i]]

    def setUp(self):
        self.accounts_cmt.ensure_ready()
        self.volumes_cmt.ensure_ready()

    def test_init(self):
        volumes_cmt = vl_components.CloneVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            volumes_cmt=self.volumes_cmt,
            accounts_cmt=self.accounts_cmt
        )
        assert volumes_cmt.child_cmts == [volumes_cmt]
        assert sort_vols(volumes_cmt.partial_volume_dms) == \
            sort_vols(self.volumes_cmt.partial_volume_dms)
        assert volumes_cmt.get_allocated_space() == self.volumes_cmt.get_allocated_space()

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_components.CloneVolumes, '_assert_valid')
    @mock.patch.object(functional_models.Cluster, 'clone_volume')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_apply(self, list_volumes, clone_volumes, assert_valid, delete_volumes):
        list_volumes.side_effect = [[], self.volume_dms]
        clone_volumes.side_effect = self.mock_async_handlers
        volumes_cmt = vl_components.CloneVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            volumes_cmt=self.volumes_cmt,
            accounts_cmt=self.accounts_cmt
        )

        volumes_cmt.apply_async()
        assert self.accounts_cmt.state == components.States.applied
        assert self.volumes_cmt.state == components.States.applied
        assert clone_volumes.called
        volume_dms = iter(self.volumes_cmt.volume_dms)
        for call_args in clone_volumes.call_args_list:
            assert call_args[1]['volume_dm'] == next(volume_dms)

        assert volumes_cmt.is_running()
        assert volumes_cmt.is_running()
        assert not volumes_cmt.is_running()
        volumes_cmt.finish()
        assert assert_valid.called

        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[0]]
        assert volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[0]]
        assert sort_vols(volumes_cmt.volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(volumes_cmt.get_volume_dms()) == sort_vols(self.volume_dms)

        list_volumes.side_effect = [[]]
        volumes_cmt.reset()
        assert delete_volumes.called
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.account_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_volume_dms()

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(functional_models.Cluster, 'clone_volume')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_reload(self, list_volumes, clone_volumes, delete_volumes):
        volume_dms = copy.deepcopy(self.volume_dms)
        for dm in volume_dms:
            dm.account_id = self.accounts_cmt._account_dms[1].account_id

        invalid_volume_dms = copy.deepcopy(volume_dms)
        invalid_volume_dms[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volume_dms
        volumes_cmt = vl_components.CloneVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            volumes_cmt=self.volumes_cmt,
            accounts_cmt=self.accounts_cmt,
            account_index=1
        )

        with pytest.raises(components.PartialStateError):
            volumes_cmt.apply()
        volumes_cmt.reset()
        assert sort_vols(delete_volumes.call_args[0][1]) == sort_vols(invalid_volume_dms)

        list_volumes.return_value = volume_dms
        volumes_cmt.apply()
        assert not clone_volumes.called
        assert volumes_cmt.account_dms == [self.accounts_cmt.account_dms[1]]
        assert volumes_cmt.volume_dms == volume_dms

    @mock.patch.object(functional_models.Cluster, 'clone_volume')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_assert_valid(self, list_volumes, clone_volumes):
        list_volumes.return_value = self.volume_dms
        volumes_cmt = vl_components.CloneVolumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            volumes_cmt=self.volumes_cmt,
            accounts_cmt=self.accounts_cmt
        )
        with pytest.raises(components.ComponentStateError):
            volumes_cmt.assert_valid()

        volumes_cmt.apply()
        volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)
        invalid_volume_dms[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)
        invalid_volume_dms[0].account_id = 99
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(self.volume_dms)[:-1]
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()


@mock.patch(
    'valence.functional_models.Cluster.get_constants',
    mock.MagicMock(return_value=models.ClusterConstants.create(
        bs_unusable_space_percent=7
    ))
)
@mock.patch(
    'valence.functional_models.Cluster.get_immutable_values',
    mock.MagicMock(return_value=result_models.GetImmutableValues.create(
        per_block_overhead_bytes=50,
        per_block_metadata_bytes=16,
        per_page_metadata_blocks=341,
        per_page_metadata_bytes=4096
    ))
)
@mock.patch(
    'valence.functional_models.Cluster.get_limits',
    mock.MagicMock(return_value=result_models.GetLimits.create(
        volume_size_min=1000000000,
        volume_size_max=17592186044416,
        volume_count_max=4000,
        volumes_per_account_count_max=2000
    ))
)
@mock.patch('valence.functional_models.Cluster.get_cluster_full_threshold',
            mock.MagicMock(return_value=models.ClusterFullThreshold.create(sum_used_cluster_bytes=0)))
@mock.patch('valence.vl_components._cluster.RunGarbageCollection',
            vl_components.RunGarbageCollection.mock)
class TestCreateVolumesForCapacity(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)
        cls.account_dms = [models.Account.create(username='unittest-0-TAG-0', account_id=0)]
        cls.data_size = models.DataSize(600, enums.DataUnit.gigabytes)
        cls.capacity_selector = selectors.FixedBlockCapacitySelector(
            cluster_dm=cls.cluster_dm,
            cluster_bytes=cls.data_size.bytes
        )
        cls.volume_count = 60
        cls.volume_size = models.DataSize(4.6, enums.DataUnit.gigabytes)
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-0-{}'.format(i),
            volume_id=i,
            account_id=cls.account_dms[0].account_id,
            total_size=cls.volume_size
        ) for i in range(cls.volume_count)]

    def test_init(self):
        with pytest.raises(AssertionError):
            volumes_cmt = vl_components.CreateVolumesForCapacity(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                capacity_selector=self.capacity_selector,
                comp_ratio=0.9,
                dedup_ratio=1.1
            )

        with pytest.raises(AssertionError):
            volumes_cmt = vl_components.CreateVolumesForCapacity(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                capacity_selector=self.capacity_selector,
                comp_ratio=1.1,
                dedup_ratio=0.9
            )

        volumes_cmt = vl_components.CreateVolumesForCapacity(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            capacity_selector=self.capacity_selector,
            comp_ratio=1.1,
            dedup_ratio=1.1
        )
        assert volumes_cmt.comp_ratio == 1.1
        assert volumes_cmt.dedup_ratio == 1.1
        assert volumes_cmt.child_cmts == [volumes_cmt]

    @mock.patch.object(vl_helpers, 'get_volume_size_for_capacity')
    @mock.patch.object(vl_helpers, 'get_volume_count_for_capacity')
    @mock.patch.object(vl_helpers, 'get_block_volume_size')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_apply_setup(self, list_volumes, get_block_vol_size, get_vol_count, get_vol_size):
        get_vol_count.return_value = self.volume_count
        get_vol_size.return_value = self.volume_size
        volumes_cmt = vl_components.CreateVolumesForCapacity(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            capacity_selector=self.capacity_selector
        )
        volumes_cmt._apply_setup()
        assert volumes_cmt._vol_count == self.volume_count
        assert volumes_cmt._vol_size == self.volume_size

        # Test volume-count/compression-ratio/deduplication-ratio/under-size overrides
        volumes_cmt = vl_components.CreateVolumesForCapacity(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            capacity_selector=self.capacity_selector,
            vol_count=42,
            comp_ratio=2.2,
            dedup_ratio=3.7,
            under_size=True
        )
        volumes_cmt._apply_setup()
        assert volumes_cmt._vol_count == 42
        assert get_block_vol_size.call_args[1]['comp_ratio'] == 2.2
        assert get_block_vol_size.call_args[1]['dedup_ratio'] == 3.7
        assert get_vol_size.call_args[1]['under_size']

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_helpers, 'delete_accounts')
    @mock.patch.object(vl_components.CreateVolumesForCapacity, '_assert_valid')
    @mock.patch.object(vl_helpers, 'create_volumes_single_profile')
    @mock.patch.object(vl_helpers, 'create_accounts')
    @mock.patch.object(vl_helpers, 'update_cluster_full_threshold')
    @mock.patch.object(vl_helpers, 'get_volume_size_for_capacity')
    @mock.patch.object(vl_helpers, 'get_volume_count_for_capacity')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_apply(self, list_accounts, list_volumes, get_vol_count, get_vol_size, update_full_threshold,
                   create_accounts, create_volumes, assert_valid, delete_accounts, delete_volumes):
        list_accounts.side_effect = [[], [], self.account_dms, self.account_dms, self.account_dms]
        list_volumes.side_effect = [[], self.volume_dms, self.volume_dms, []]
        get_vol_count.return_value = self.volume_count
        get_vol_size.return_value = self.volume_size
        create_accounts.return_value = self.account_dms
        volumes_cmt = vl_components.CreateVolumesForCapacity(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            capacity_selector=self.capacity_selector
        )

        volumes_cmt.apply()
        assert not update_full_threshold.called
        assert create_accounts.called
        assert create_volumes.called
        assert create_volumes.call_args[1]['volume_dm'] == models.Volume.create(
            name='unittest-TAG-0-',
            account_id=self.account_dms[0].account_id,
            total_size=volumes_cmt._vol_size
        )
        assert assert_valid.called

        assert volumes_cmt.account_dms == self.account_dms
        assert volumes_cmt.get_account_dms() == self.account_dms
        assert sort_vols(volumes_cmt.partial_volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(volumes_cmt.volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(volumes_cmt.get_volume_dms()) == sort_vols(self.volume_dms)
        assert volumes_cmt.get_allocated_space() == volumes_cmt._vol_size.bytes * volumes_cmt._vol_count
        assert volumes_cmt.block_capacity > self.capacity_selector().bytes
        assert volumes_cmt.block_capacity < 1.01 * self.capacity_selector().bytes
        assert volumes_cmt.meta_capacity

        volumes_cmt.reset()
        assert delete_volumes.called
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.account_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.partial_volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_volume_dms()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.get_allocated_space()
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.block_capacity
        with pytest.raises(components.ComponentStateError):
            assert volumes_cmt.meta_capacity

        list_accounts.side_effect = [[], [], self.account_dms, self.account_dms, self.account_dms]
        list_volumes.side_effect = [[], self.volume_dms, self.volume_dms, []]
        self.capacity_selector.relative = True
        volumes_cmt.apply()
        assert update_full_threshold.called
        self.capacity_selector.relative = False

    @mock.patch.object(vl_helpers, 'delete_and_purge_volumes')
    @mock.patch.object(vl_helpers, 'delete_accounts')
    @mock.patch.object(vl_helpers, 'create_volumes_single_profile')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_reload(self, list_accounts, list_volumes, create_volumes, delete_accounts, delete_volumes):
        volumes_cmt = vl_components.CreateVolumesForCapacity(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            capacity_selector=self.capacity_selector
        )

        volumes_cmt._apply_setup()
        volume_dms = [models.Volume.create(
            name='unittest-TAG-0-{}'.format(i),
            volume_id=i,
            total_size=volumes_cmt._vol_size
        ) for i in range(volumes_cmt._vol_count)]
        invalid_volume_dms = copy.deepcopy(volume_dms)
        invalid_volume_dms[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)

        list_accounts.return_value = self.account_dms
        list_volumes.return_value = invalid_volume_dms

        with pytest.raises(components.PartialStateError):
            volumes_cmt.apply()
        volumes_cmt.reset()
        assert sort_vols(delete_volumes.call_args[0][1]) == sort_vols(invalid_volume_dms)

        list_volumes.return_value = volume_dms
        volumes_cmt.apply()
        assert not create_volumes.called
        assert volumes_cmt.account_dms == self.account_dms
        assert sort_vols(volumes_cmt.volume_dms) == sort_vols(volume_dms)

    @mock.patch.object(vl_helpers, 'create_volumes_single_profile')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_assert_valid(self, list_accounts, list_volumes, create_volumes):
        volumes_cmt = vl_components.CreateVolumesForCapacity(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            capacity_selector=self.capacity_selector
        )

        volumes_cmt._apply_setup()
        volume_dms = [models.Volume.create(
            name='unittest-TAG-0-{}'.format(i),
            volume_id=i,
            total_size=volumes_cmt._vol_size
        ) for i in range(volumes_cmt._vol_count)]
        list_accounts.return_value = self.account_dms
        list_volumes.return_value = volume_dms

        with pytest.raises(components.ComponentStateError):
            volumes_cmt.assert_valid()

        volumes_cmt.apply()
        volumes_cmt.assert_valid()

        invalid_volumes = copy.deepcopy(volume_dms)
        invalid_volumes[0].total_size = models.DataSize(1, enums.DataUnit.gibibytes)
        list_volumes.return_value = invalid_volumes
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()

        invalid_volume_dms = copy.deepcopy(volume_dms)[:-1]
        list_volumes.return_value = invalid_volume_dms
        with pytest.raises(AssertionError):
            volumes_cmt.assert_valid()


class TestManageVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.accounts_cmt = vl_components.CreateAccounts.mock('mock', cls.cluster_dm, 2)

        cls.partial_volume_dms = [models.Volume.create(
            total_size=models.DataSize(10 * i, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=1000 + i, max_iops=12000, burst_iops=20000)
        ) for i in range(3)]
        cls.volumes_cmts = [vl_components.CreateVolumes.mock(
            tag='unittest-{}'.format(i),
            cluster_dm=cls.cluster_dm,
            partial_volume_dms=cls.partial_volume_dms,
            accounts_cmt=cls.accounts_cmt
        ) for i in range(3)]
        cls.volume_dms = []
        for cmt in cls.volumes_cmts:
            cls.volume_dms.extend(cmt._volume_dms)
        cls.manage_volumes_cmt = vl_components.ManageVolumes(cls.volumes_cmts)

    def setUp(self):
        self.accounts_cmt.ensure_ready()

    def test_init(self):
        assert self.manage_volumes_cmt.child_cmts == self.volumes_cmts

        partial_volume_dms = []
        for cmt in self.volumes_cmts:
            partial_volume_dms.extend(cmt.partial_volume_dms)
        assert sort_vols(self.manage_volumes_cmt.partial_volume_dms) == sort_vols(partial_volume_dms)

        allocated_space = sum(dm.total_size.bytes for dm in partial_volume_dms)
        assert self.manage_volumes_cmt.get_allocated_space() == allocated_space

    def test_apply(self):
        self.manage_volumes_cmt.apply()
        for cmt in self.volumes_cmts:
            assert cmt.state == components.States.applied

        assert self.manage_volumes_cmt.account_dms == [self.accounts_cmt.account_dms[0]]
        assert self.manage_volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[0]]
        assert sort_vols(self.manage_volumes_cmt.volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(self.manage_volumes_cmt.get_volume_dms()) == sort_vols(self.volume_dms)

        self.manage_volumes_cmt.reset()
        for cmt in self.volumes_cmts:
            assert cmt.state == components.States.ready

        with pytest.raises(components.ComponentStateError):
            assert self.manage_volumes_cmt.account_dms
        with pytest.raises(components.ComponentStateError):
            assert self.manage_volumes_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            assert self.manage_volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert self.manage_volumes_cmt.get_volume_dms()

    @mock.patch('valence.vl_components._volume.MockVolumes.assert_valid')
    def test_assert_valid(self, mock_assert_valid):
        self.manage_volumes_cmt.assert_valid()
        assert mock_assert_valid.call_count == len(self.volumes_cmts)


class TestMockVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.accounts_cmt = vl_components.CreateAccounts.mock('mock', cls.cluster_dm, 2)

        cls.partial_volume_dm = models.Volume.create(
            total_size=models.DataSize(10, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=1000, max_iops=12000, burst_iops=20000)
        )
        cls.mock_volumes_cmt = vl_components.CreateMultipleVolumes.mock(
            tag='unittest',
            cluster_dm=cls.cluster_dm,
            vol_count=3,
            partial_volume_dm=cls.partial_volume_dm,
            accounts_cmt=cls.accounts_cmt,
            account_index=1
        )
        cls.volume_dms = cls.mock_volumes_cmt._volume_dms

    def test_init(self):
        assert self.mock_volumes_cmt.child_cmts == [self.mock_volumes_cmt]
        assert sort_vols(self.mock_volumes_cmt.partial_volume_dms) == sort_vols(self.volume_dms)
        assert self.mock_volumes_cmt.get_allocated_space() == 3 * self.partial_volume_dm.total_size.bytes

    def test_apply(self):
        self.mock_volumes_cmt.apply()
        assert self.mock_volumes_cmt.state is components.States.applied
        assert self.mock_volumes_cmt.account_dms == [self.accounts_cmt.account_dms[1]]
        assert self.mock_volumes_cmt.get_account_dms() == [self.accounts_cmt.account_dms[1]]
        assert sort_vols(self.mock_volumes_cmt.volume_dms) == sort_vols(self.volume_dms)
        assert sort_vols(self.mock_volumes_cmt.get_volume_dms()) == sort_vols(self.volume_dms)

        self.mock_volumes_cmt.reset()
        assert self.mock_volumes_cmt.state is components.States.ready
        with pytest.raises(components.ComponentStateError):
            assert self.mock_volumes_cmt.account_dms
        with pytest.raises(components.ComponentStateError):
            assert self.mock_volumes_cmt.get_account_dms()
        with pytest.raises(components.ComponentStateError):
            assert self.mock_volumes_cmt.volume_dms
        with pytest.raises(components.ComponentStateError):
            assert self.mock_volumes_cmt.get_volume_dms()

    def test_assert_valid(self):
        self.mock_volumes_cmt.assert_valid()
