"""Vdbench enumerations."""
import aenum


# TODO: Workaround for pylint bug (JIRA: VL-2196)
# TODO: Raises "no-value-for-parameter" error when instantiating an enum without a base class
class VdbEnum(aenum.Enum):
    """Vdbench enum class."""


class VdbStatus(VdbEnum):
    """Vdbench Status."""
    not_started = 'not_started'
    starting = 'starting'
    running = 'running'
    finishing = 'finishing'
    completed = 'completed'
    failed = 'failed'

    @property
    def is_running(self):
        return self in (self.running, self.finishing)


class IoDistribution(VdbEnum):
    """IO distribution."""
    exponential = 'exponential'  # Default (recommended)
    uniform = 'uniform'
    deterministic = 'deterministic'  # All IO is evenly spread out using fixed interval times.
    variable = 'variable'
    spike = 'spike'


class ReportingConfig(VdbEnum):
    """Reporting configuration.

    By default Vdbench generates reports with detailed statistics for each storage definition.

    Variants:
    * None: Suppress ALL report generation.
    * Hosts: Include reports for each host.
    * Slaves: Include reports for each slave.
    * All: Include reports for each host and slave.
    """
    none = 'no_sd_detail'
    hosts = 'host_detail'
    slaves = 'slave_detail'
    all = '(host,slave)'


class ValidationConfig(VdbEnum):
    """Validation configuration.

    Vdbench can determine whether the data it has written during that run is valid when it re-reads it.
    It uses an incrementing lookup table or somesuch which has a fairly small memory footprint (the docs
    are really wordy about this).

    NOTE: Only works within a single run. For validation across multiple runs, use journaling.

    Variants:
    * None: Run with NO data validation.
    * On-Reads: Run with data validation (validates data on every read).
    * On-Writes: Run with data validation (re-reads (and validates) on every write).
    * With-Times: Run with data validation and keep track of each timestamp (memory intensive).
    """
    none = 'no'
    on_reads = 'yes'
    on_writes = 'read_after_write'
    with_times = 'time'


class JournalingConfig(VdbEnum):
    """Journaling configuration options.

    Journaling saves the validation tables so that later runs can do validation on the data from
    previous runs.

    NOTE: Data validation will always be performed when running with journaling.
    NOTE: Journal files will automatically be deleted if Vdbench runs with journaling without recovery.

    Variants:
    * None: Run with NO journaling.
    * Full: Run with journaling.
    * AsyncIO: Run with journaling using async I/O on journal files (increases workload performance).
    * Recover: Recover existing journal and validate it, then run with journaling.
    * Validate: Recover existing journal and validate it, then run with NO journaling.
    """
    none = 'no'
    full = 'yes'
    async_io = 'noflush'
    recover = 'recover'
    validate = '(recover,only)'
