"""Vdbench IO Period objects.

These objects define Vdbench configurations for a period of IO.
"""
import copy

from valence.vl_components.vdbench import (
    _base,
    _parm,
)


class IoPeriodConfig(_base.BetterObject):
    """Vdbench configuration for a generic period of IO.

    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration:

        * **Fixed:** Single, fixed IO rate.
        * **Series:** Multiple IO rates run in series.
        * **Curve:** Multiple IO rates as a percentage of max.
        * **Max:** Maximum possible IO (given any requested skew).
        * **Variable:** Multiple IO rates run in series with variable durations.

    :param int interval: Reporting interval in seconds.
    :param int duration: Duration for this IO period in seconds (default: 30s).
    :param int max_data: Number of bytes written at which point to end this IO period (default: None).
    :param int warmup: Warmup period in seconds (default: None).
    :param valence.vl_components.vdbench.ForTransferSizeConfig for_transfer_size_config: For-xfersize
        configuration. **NOTE:** May extend the duration.
    :param valence.vl_components.vdbench.ForReadPercentConfig for_read_perc_config: For-read-perc
        configuration. **NOTE:** May extend the duration.
    :param valence.vl_components.vdbench.ForSeekPercentConfig for_seek_perc_config: For-seek-perc
        configuration. **NOTE:** May extend the duration.
    :param valence.vl_components.vdbench.ForThreadsConfig for_threads_config: For-threads configuration.
        **NOTE:** May extend the duration.
    """

    def __init__(self, io_rate_config, interval=None, duration=None, max_data=None, warmup=None,
                 for_transfer_size_config=None, for_read_perc_config=None, for_seek_perc_config=None,
                 for_threads_config=None):
        self.io_rate_config = io_rate_config
        self.interval = interval if interval is not None else 1
        self.max_data = max_data
        self.warmup = warmup
        self.for_transfer_size_config = for_transfer_size_config
        self.for_read_perc_config = for_read_perc_config
        self.for_seek_perc_config = for_seek_perc_config
        self.for_threads_config = for_threads_config

        # Define period duration (default: 30s per defined IO-rate)
        if duration is not None:
            self.duration = duration
        elif isinstance(self.io_rate_config, _parm.IoRateCurve):
            self.duration = 30 * len(self.io_rate_config.curve)
        elif isinstance(self.io_rate_config, _parm.IoRateSeries):
            self.duration = 30 * len(self.io_rate_config.io_rates)
        elif isinstance(self.io_rate_config, _parm.IoRateVariable):
            self.duration = sum(self.io_rate_config.io_rate_durations)
        else:
            self.duration = 30

        assert self.duration > 0
        if self.interval is not None:
            assert 0 < self.interval <= self.duration // 2
        if self.max_data is not None:
            assert self.max_data > 0
        if self.warmup is not None:
            assert self.warmup >= 0

    def __eq__(self, other):
        """Compare the provided object with this one.

        :param Any other: Object with which to compare.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(zip(self.__dict__.keys(), self.__dict__.values())))

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'io_rate', self.io_rate_config,
            'duration', self.duration
        )


class IoPeriod(_base.BetterObject):
    """Vdbench configuration for a period of IO between particular clients and volumes.

    IO periods are run in series.

    :param valence.vl_components.vdbench.IoPeriodConfig io_period_config: Configuration parameters for
        this period of IO.
    :param list[valence.vl_components.vdbench.IoWorkload] io_workloads: Workloads to run in parallel
        during this IO period.
    """
    MAX_WORKLOADS = 512  # Vdbench enforces a limit of 512 workload definitions per run definition

    def __init__(self, io_period_config, io_workloads):
        self.config = io_period_config
        self.io_workloads = io_workloads
        workload_configs = [io_workload.config for io_workload in io_workloads]
        assert len(set(workload_configs)) <= self.MAX_WORKLOADS

    def copy(self):
        """Deep copy the period configuration without replicating referenced components.

        :rtype: valence.vl_components.vdbench.IoPeriod
        """
        return IoPeriod(
            io_period_config=copy.deepcopy(self.config),
            io_workloads=[workload.copy() for workload in self.io_workloads]
        )

    def __eq__(self, other):
        """Compare the provided object with this one.

        :param Any other: Object with which to compare.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        matched_workloads = []
        unmatched_workloads = copy.copy(self.io_workloads)
        for other_workload in other.io_workloads:
            for workload in unmatched_workloads:
                if other_workload == workload:
                    unmatched_workloads.remove(workload)
                    matched_workloads.append(workload)
                    break
        if unmatched_workloads:
            return False
        if len(matched_workloads) != len(other.io_workloads):
            return False
        return other.config == self.config

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'config', self.config)
