"""Vdbench IO Session objects.

These objects define Vdbench configurations for a session of IO, the IO driven by a single Vdbench
master.
"""
import copy

from valence.vl_components.vdbench import _base


class IoSessionConfig(_base.BetterObject):
    """Vdbench configuration for a session of IO.

    :param float compression_ratio: Ratio of compressed data size to uncompressed data size
        (default: 1, no compression). Vdbench-generated data patterns will try to target the specified
        ratio based on the behavior of the 'LZJB' compression algorithm. Different compression algorithms
        may produce slightly different results however and in general the accuracy of this implementation
        is roughly +/- 5%.
    :param valence.vl_components.vdbench.DedupConfig dedup_config: Data deduplication configuration.
    :param str pattern_path: This is currently not well supported but will be in the future. Currently
        YOU have to manually scp the file and hack this path into your config. This will change in the
        future. The data found in this file is copied as often as possible into the data pattern buffer.
        The first four eight in the data buffer will be replaced with a unique value before each write
        operation, this to prevent this block to become a duplicate, unless of course a dedup_ratio has
        been requested.
    """

    def __init__(self, compression_ratio=None, dedup_config=None, pattern_path=None):
        self.compression_ratio = compression_ratio
        self.dedup_config = dedup_config
        self.pattern_path = pattern_path

        if self.compression_ratio is not None:
            assert self.compression_ratio > 0

    def __eq__(self, other):
        """Compare the provided object with this one.

        :param Any other: Object with which to compare.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(zip(self.__dict__.keys(), self.__dict__.values())))

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'compression_ratio', self.compression_ratio,
            'dedup_config', self.dedup_config
        )


class IoSession(_base.BetterObject):
    """Vdbench configuration for a session of IO for particular clients and volumes.

    Each Vdbench master will handle a single IO session.

    :param valence.vl_components.vdbench.IoSessionConfig io_session_config: Configuration parameters for
        this IO session.
    :param list[valence.vl_components.vdbench.IoPeriod] io_periods: IO periods to run in series during
        this IO session. Each period runs for its defined duration.
    :param bool standalone: Whether to run Vdbench in standalone mode (default: False). By default
        Vdbench runs a slave alongside the master instance. In standalone mode the master performs the
        duties of the slave in addition to its own (only works for single-client instances)
        (reduces memory overhead).
    """

    def __init__(self, io_session_config, io_periods, standalone=False):
        self.config = io_session_config
        self.io_periods = io_periods
        self.standalone = standalone

        if self.config.dedup_config:
            self._validate_periods_for_dedup(self.config.dedup_config, self.io_periods)

        if standalone:
            self._validate_periods_for_standalone(self.io_periods)

    @staticmethod
    def _validate_periods_for_dedup(dedup_config, io_periods):
        """Ensure each IO period uses transfer sizes that are multiples of the dedup-unit.

        Vdbench requires this to properly emulate data deduplication on the cluster.

        :param valence.vl_components.vdbench.DedupConfig dedup_config: Data deduplication configuration.
        :param list[valence.vl_components.vdbench.IoPeriod] io_periods: IO periods.
        """
        dedup_unit = dedup_config.dedup_unit or 128 * 1024
        msg = 'All xfersizes must be divisible by the dedup_unit: {}'.format(dedup_unit)
        for io_period in io_periods:
            if io_period.config.for_transfer_size_config:
                for xfersize in io_period.config.for_transfer_size_config.xfersizes:
                    assert xfersize % dedup_unit == 0, msg
            else:
                for io_workload in io_period.io_workloads:
                    for xfersize in io_workload.config.transfer_size_config.xfersizes:
                        assert xfersize % dedup_unit == 0, msg

    @staticmethod
    def _validate_periods_for_standalone(io_periods):
        """Ensure each IO period uses the same client for standalone operation.

        Running in standalone, Vdbench does not support additional slave clients.

        :param list[valence.vl_components.vdbench.IoPeriod] io_periods: IO periods.
        """
        client_mips = set()
        for period in io_periods:
            client_mips.update([workload.client_dm.mip for workload in period.io_workloads])
        assert len(client_mips) == 1, 'Multiple clients not supported for standalone Vdbench IO'

    def copy(self):
        """Deep copy the session configuration without replicating referenced components.

        :rtype: valence.vl_components.vdbench.IoSession
        """
        return IoSession(
            io_session_config=copy.deepcopy(self.config),
            io_periods=[period.copy() for period in self.io_periods]
        )

    def __eq__(self, other):
        """Compare the provided object with this one.

        :param Any other: Object with which to compare.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        if other.io_periods != self.io_periods:
            return False
        comparisons = [
            other.config == self.config,
            other.standalone == self.standalone
        ]
        return all(comparisons)

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'config', self.config)
