"""Vdbench IO Workload objects.

These objects define Vdbench configurations for single-client to single-volume IO.
"""
import copy

from valence.vl_components.vdbench import _base


class IoWorkloadConfig(_base.BetterObject):
    """Vdbench configuration for a generic IO workload.

    :param int io_rate: Fixed IO rate for this workload (default: The period-defined IO rate).
    :param int read_perc: Read percentage for this workload. If every workload for a given volume has
        read_perc=100, then the volume is opened for input only.
    :param int offset: Offset in bytes (must be a multiple of 512). By default Vdbench always starts at
        the beginning of a LUN. Use offset if a LUN does not start at an exact physical stripe boundary.
        NOTE: Vdbench never accesses block zero on any raw volume. This has been done to make sure that
        it never overwrites a volume label and/or vtoc.
    :param valence.vl_components.vdbench.TransferSizeConfig transfer_size_config: Transfer size
        configuration:

        * **Fixed:** Single, fixed transfer size.
        * **Weighted:** Multiple transfer sizes, each constituting a percentage of the workload IO.
        * **Aligned:** Range of aligned transfer sizes.

    :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration.
    :param valence.vl_components.vdbench.CacheHitConfig hit_config: Cache hit configuration.
    :param int skew_perc: Percentage of "work" for this workload for the entire IO period
        (default: No skew). If skew is defined for any individual workload in an IO period, then all
        other workloads in the same period must have skew defined and they must all add up to 100.
    :param int priority: Workload priority (default: No priority). Priorities go from 1 (highest)
        to 'n' (lowest), and must be in sequence (e.g. 1, 2 NOT 1, 3). If any workload is given a
        priority, then all workloads in a given IO period must have one.
    :param int thread_count: The maximum number of concurrent I/O that can be outstanding for this
        workload. Be aware that depending on the storage subsystem, some of these I/Os may be queued
        inside of the operating system (wait queue).
        WARNING: If you specify a LUN that physically consists of multiple disk drives, the thread count
        determines the maximum concurrency for the LUN, not for the disks. A total of 8 concurrent
        threads for a total of 16 physical disks will not allow for much concurrency. Also, for Solaris,
        make sure that your sd_max_throttle parameter in /etc/system allows the requested amount of
        concurrency.
        NOTE: Be aware that the maximum concurrency will only occur when there is enough demand.
        Requesting 10 IOPS against a device that can handle 1000 IOPS will not give you the concurrency
        that you request with the thread parameter.
    :param valence.vl_components.vdbench.AccessRangeConfig access_range_config: Access range
        configuration.
    :param valence.vl_components.vdbench.StreamsConfig streams_config: Streams configuration.
    :param str open_flags: This parameter allows you to control what parameters are passed to the
        system's open and close functions. By default write operations are handled according to how the
        file system is mounted or for raw devices, how the device normally operates.  This can mean that
        a write operation completes as soon as the data is stored in system cache. This makes for very
        good performance, but does not really exercise the storage (default: 'o_direct' - required for
        volume access from Linux clients).
    """

    def __init__(self, io_rate=None, read_perc=None, offset=None, transfer_size_config=None,
                 seek_config=None, hit_config=None, skew_perc=None, priority=None, thread_count=None,
                 access_range_config=None, streams_config=None, open_flags='o_direct'):
        self.io_rate = io_rate
        self.read_perc = read_perc if read_perc is not None else 80
        self.offset = offset
        self.transfer_size_config = transfer_size_config
        self.seek_config = seek_config
        self.hit_config = hit_config
        self.skew_perc = skew_perc
        self.priority = priority
        self.thread_count = thread_count
        self.access_range_config = access_range_config
        self.streams_config = streams_config
        self.open_flags = open_flags

        assert not (hit_config and streams_config), 'Streams are incompatible with Hit Configuration'
        assert (io_rate is None or priority is not None), 'IO rate specification requires a priority'

        if self.io_rate is not None:
            assert self.io_rate > 0
        if self.read_perc is not None:
            assert 0 <= self.read_perc <= 100
        if self.offset is not None:
            assert self.offset >= 0
            assert self.offset % 512 == 0, 'Offset must be divisible by 512'
        if self.skew_perc is not None:
            assert 0 <= self.skew_perc <= 100
        if self.priority is not None:
            assert self.priority > 0
        if self.thread_count is not None:
            assert self.thread_count > 0

    @property
    def aligned(self):
        if self.transfer_size_config:
            return self.transfer_size_config.aligned
        else:
            return True

    def __eq__(self, other):
        """Compare the provided object with this one.

        :param Any other: Object with which to compare.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(zip(self.__dict__.keys(), self.__dict__.values())))

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'io_rate', self.io_rate,
            'transfer_size', self.transfer_size_config,
            'read_perc', self.read_perc,
            'access_range', self.access_range_config
        )


class IoWorkload(_base.BetterObject):
    """Vdbench configuration for an IO workload between a particular client-volume pair.

    :param valence.vl_components.vdbench.IoWorkloadConfig io_workload_config: Configuration parameters
        for this IO workload.
    :param valence.models.ResourceClient client_dm: The client from which to run IO.
    :param valence.vl_components.VolumesInterface volumes_cmt: Volumes to reference.
    :param int volume_index: The specific volume for which to configure IO.
    """

    def __init__(self, io_workload_config, client_dm, volumes_cmt, volume_index):
        self.config = io_workload_config
        self.client_dm = client_dm
        self.volumes_cmt = volumes_cmt
        self.volume_index = volume_index

        assert len(self.volumes_cmt.child_cmts) == 1
        if not self.config.aligned:
            if not self.volumes_cmt.partial_volume_dms[self.volume_index].enable512e:
                raise ValueError(
                    'Unaligned xfersize[{}] configured for non-512e-enabled volume'
                    .format(self.config.transfer_size_config.get_parm_str())
                )

    def copy(self):
        """Deep copy the workload configuration without replicating referenced components.

        :rtype: valence.vl_components.vdbench.IoWorkload
        """
        return IoWorkload(
            io_workload_config=copy.deepcopy(self.config),
            client_dm=self.client_dm,
            volumes_cmt=self.volumes_cmt,
            volume_index=self.volume_index
        )

    def __eq__(self, other):
        """Compare the provided object with this one.

        :param Any other: Object with which to compare.
        :rtype: bool
        """
        if not isinstance(other, self.__class__):
            return False
        comparisons = [
            other.config == self.config,
            other.client_dm == self.client_dm,
            other.volumes_cmt == self.volumes_cmt,
            other.volume_index == self.volume_index
        ]
        return all(comparisons)

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return '{}<{}={}>'.format(self.__class__.__name__, 'config', self.config)
