"""Vdbench log helpers."""
from valence.vl_components.vdbench._log_parser import (
    constants,
    enums,
    models,
)

LINE_TYPE_MAP = {
    constants.RUN_START_LINE: enums.VdbLogLineType.run_start,
    constants.RUN_DATE_LINE: enums.VdbLogLineType.run_date,
    constants.IO_LINE: enums.VdbLogLineType.io,
    constants.AVG_IO_LINE: enums.VdbLogLineType.avg_io,
    constants.CLIENT_HEARTBEAT_LINE: enums.VdbLogLineType.client_heartbeat,
    constants.MASTER_HEARTBEAT_LINE: enums.VdbLogLineType.master_heartbeat,
    constants.NO_READ_VALIDATIONS_LINE: enums.VdbLogLineType.no_read_validations,
    constants.EIO_LINE: enums.VdbLogLineType.io_error,
    constants.EOF_LINE: enums.VdbLogLineType.end_of_file,
    constants.SOCKET_EXCEPTION_LINE: enums.VdbLogLineType.socket_exception,
    constants.DATA_VALIDATION_ERROR_LINE: enums.VdbLogLineType.data_validation_error,
    constants.JSTACK_LINE: enums.VdbLogLineType.jstack,
    constants.JAVA_POINTER_ERROR_LINE: enums.VdbLogLineType.jnull,
    constants.JAVA_RUNTIME_ERROR_LINE: enums.VdbLogLineType.jruntime,
}


def parse_raw(raw_lines, defined_intervals, last_line=None):
    """Parse raw lines.

    :param list[str] raw_lines: Raw Vdbench log lines to parse.
    :param list[int] defined_intervals: List of intervals (in seconds) indexed by run definition.
    :param valence.vl_components.vdbench._log_parser.VdbLogLine last_line: Last line parsed.
    :rtype: list[valence.vl_components.vdbench._log_parser.VdbLogLine]
    """
    lines = []
    for raw_line in raw_lines:
        line_type = _get_vdb_line_type(raw_line)
        if line_type is enums.VdbLogLineType.io:
            assert last_line is not None, 'Missing run info for IO line in Vdbench log'
            line = models.VdbIoLine(raw_line, defined_intervals, last_line)
        elif line_type is enums.VdbLogLineType.run_start:
            line = models.VdbRunInfoLine(raw_line, last_line)
        else:
            line = models.VdbLogLine(raw_line, line_type, last_line)

        lines.append(line)
        last_line = line

    return lines


def _get_vdb_line_type(raw_line):
    """Get the type for the given line of the Vdbench log.

    :param str raw_line: Raw line string from the Vdbench master log.
    :rtype: valence.vl_components.vdbench._log_parser.VdbLogLineType
    """
    line_type = enums.VdbLogLineType.unknown
    if raw_line == '':
        line_type = enums.VdbLogLineType.new_line

    for regex in LINE_TYPE_MAP:
        if regex.search(raw_line):
            line_type = LINE_TYPE_MAP[regex]
            break

    # Premature termination lines are a subset of Java Runtime Errors
    if line_type == enums.VdbLogLineType.jruntime:
        if constants.PREMATURE_TERMINATION_LINE.search(raw_line):
            line_type = enums.VdbLogLineType.premature_termination

    return line_type
