"""Vdbench log models."""
import math

from valence.vl_components.vdbench._log_parser import (
    constants,
    enums,
)


class VdbRunInfo(object):
    """Vdbench run information.

    :param str name: Name of the Vdbench run.
    :param int index: Index of the Vdbench run.
    """

    def __init__(self, name, index):
        self.name = name
        self.index = index

    def __repr__(self):
        return '{}<{}={}, {}={}>'.format(
            self.__class__.__name__,
            'name', self.name,
            'index', self.index
        )


class VdbLogLine(object):
    """Vdbench log line.

    :param str raw_line: Raw line string from the Vdbench master log.
    :param valence.vl_components.vdbench._log_parser.VdbLogLineType line_type: Type of Vdbench log line.
    :param valence.vl_components.vdbench._log_parser.VdbLogLine last_line: Last line parsed.
    """
    ERROR_TYPES = {
        enums.VdbLogLineType.io_error,
        enums.VdbLogLineType.end_of_file,
        enums.VdbLogLineType.socket_exception,
        enums.VdbLogLineType.data_validation_error,
        enums.VdbLogLineType.jstack,
        enums.VdbLogLineType.jnull,
        enums.VdbLogLineType.jruntime,
        enums.VdbLogLineType.premature_termination,
    }

    def __init__(self, raw_line, line_type, last_line=None):
        self.raw = raw_line
        self.type = line_type

        # Information on the active Vdbench run
        self.run_info = last_line.run_info if last_line else None
        self.run_time = last_line.run_time if last_line else None

    def __str__(self):
        line_str = '{}: {}'.format(self.type, self.raw)
        if self.type in self.ERROR_TYPES:
            line_str = 'ERROR {}'.format(line_str)
        return line_str


class VdbRunInfoLine(VdbLogLine):
    """Vdbench log line representing a new run.

    :param str raw_line: Raw line string from the Vdbench master log.
    :param valence.vl_components.vdbench._log_parser.VdbLogLine last_line: Last line parsed.
    """
    def __init__(self, raw_line, last_line=None):
        super(VdbRunInfoLine, self).__init__(
            raw_line=raw_line,
            line_type=enums.VdbLogLineType.run_start,
            last_line=last_line
        )

        run_index = 0
        if self.run_info and self.run_info.name != 'journal_recovery':
            run_index = self.run_info.index + 1

        self.run_info = VdbRunInfo(
            name=constants.RUN_START_LINE.search(self.raw).group(1),
            index=run_index
        )
        self.run_time = self.run_time + 1 if self.run_time is not None else 0


class VdbIoLine(VdbLogLine):
    """Vdbench log line representing IO (either data or average).

    :param str raw_line: Raw line string from the Vdbench master log.
    :param list[int] defined_intervals: List of intervals (in seconds) indexed by run definition.
    :param valence.vl_components.vdbench._log_parser.VdbLogLine last_line: Last line parsed.
    """
    def __init__(self, raw_line, defined_intervals, last_line):
        super(VdbIoLine, self).__init__(
            raw_line=raw_line,
            line_type=enums.VdbLogLineType.io,
            last_line=last_line
        )

        raw_line_items = raw_line.split()
        self.time_stamp = raw_line_items[0]
        self.interval = int(raw_line_items[1])
        self.iops = self._get_float_or_none(raw_line_items[2])
        self.mbps = self._get_float_or_none(raw_line_items[3])
        self.io_size = int(raw_line_items[4])
        self.read_percentage = self._get_float_or_none(raw_line_items[5])
        self.response_time = self._get_float_or_none(raw_line_items[6])
        self.read_response_time = self._get_float_or_none(raw_line_items[7])
        self.write_response_time = self._get_float_or_none(raw_line_items[8])
        self.max_response_time = self._get_float_or_none(raw_line_items[9])
        self.response_std_dev = self._get_float_or_none(raw_line_items[10])
        self.queue_depth = self._get_float_or_none(raw_line_items[11])

        # IO has occurred, add the relevant interval to the run time
        self.run_time += defined_intervals[self.run_info.index]

    def serialize(self):
        """Convert IO line to dictionary.

        :rtype: dict
        """
        return {
            'time_stamp': self.time_stamp,
            'interval': self.interval,
            'iops': self.iops,
            'mbps': self.mbps,
            'io_size': self.io_size,
            'read_percentage': self.read_percentage,
            'response_time': self.response_time,
            'read_response_time': self.read_response_time,
            'write_response_time': self.write_response_time,
            'max_response_time': self.max_response_time,
            'response_std_dev': self.response_std_dev,
            'queue_depth': self.queue_depth,
        }

    @staticmethod
    def _get_float_or_none(value_str):
        """Get float but NaN -> None.

        :param str value_str: Input value.
        :rtype: float
        """
        value = float(value_str)
        return None if math.isnan(value) else value

    def __str__(self):
        return '{} [{}] {}={} {}={} {}={} {}={}'.format(
            'AVERAGE:' if self.interval is None else 'INTERVAL:{}'.format(self.interval),
            self.time_stamp,
            'max_resp', self.response_time,
            'resp', self.max_response_time,
            'mbps', self.mbps,
            'read%', self.read_percentage
        )
