"""Base methods and classes for Vdbench configuration."""
import abc

import six

from valence.vl_components.vdbench import _base


@six.add_metaclass(abc.ABCMeta)
class ConfigItem(_base.BetterObject):
    """Vdbench configuration item."""

    @abc.abstractmethod
    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """

    def get_parm_str(self):
        """Get the Vdbench parm file representation for this item.

        :rtype: str
        """
        return self._parm_str()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(zip(self.__dict__.keys(), self.__dict__.values())))

    def __repr__(self):
        return '{}<{}>'.format(self.__class__.__name__, self._parm_str())


@six.add_metaclass(abc.ABCMeta)
class ParmLinePartial(ConfigItem):
    """Partial Vdbench configuration line.

    The line is "partial" because it is missing some of the items.
    """


@six.add_metaclass(abc.ABCMeta)
class ParmLine(ConfigItem):
    """Vdbench configuration line.

    :param str comment: A comment to put above the line in the _parm file. Broken up into multiple lines
        based on the presence of '/n' characters.
    """
    def __init__(self, comment=None):
        self.comment = comment

    def get_parm_str(self):
        """Get the Vdbench parm file representation for this item.

        Includes preceding comment lines.

        :rtype: str
        """
        line_str = self._parm_str()
        out = []
        if self.comment:
            out.extend('# ' + l for l in self.comment.split('\n'))
        out.append(line_str)
        return '\n'.join(out)
