"""Vdbench GeneralParm, as defined in section 2.11 of the vdbench docs."""
import itertools
import os

from valence.vl_components.vdbench import _enums
from valence.vl_components.vdbench._parm import (
    base,
    constants,
    utils,
)


class GeneralDefConfig(base.ParmLinePartial):
    """Vdbench general definition configuration.

    NOTE: The following fields are not supported:
    * swat: Some kind of Sun tool.
    * create_anchors: Used for filesystem stuff.
    * formatxfersize: Used for formatting disk files.

    :param float compression_ratio: Configured compression ratio.
    :param valence.vl_components.vdbench.DedupConfig dedup_config: Data deduplication configuration.
    :param str pattern_path: Configured pattern path.
    """

    def __init__(self, compression_ratio=None, dedup_config=None, pattern_path=None):
        self.compression_ratio = compression_ratio
        self.dedup_config = dedup_config
        self.pattern_path = pattern_path

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        config = []
        utils.extend_config(config, self.pattern_path, constants.PATTERN_PATH)
        utils.extend_config(config, self.dedup_config)
        utils.extend_config(config, self.compression_ratio, constants.COMPRESSION_RATIO)
        return '\n'.join(config)


class GeneralDef(base.ParmLine):
    """Vdbench general definition.

    :param valence.vl_components.vdbench._parm.GeneralDefConfig gd_config: Configuration for this general
        definition.
    :param list[valence.vl_components.vdbench._parm.HostDef] host_defs: Host definitions.
    :param list[valence.vl_components.vdbench._parm.RunDef] run_defs: Run definitions.
    :param list[valence.vl_components.vdbench._parm.WorkloadDef] workload_defs: Workload definitions.
    :param list[valence.vl_components.vdbench._parm.StorageDef] storage_defs: Storage definitions.
    :param int port: Override the default port.
    :param int data_errors: Number of data-validation/IO errors after which to terminate.
    :param typing.Sequence[int] histogram: Specify the histogram buckets to use, where each bucket
        size is in microseconds. The maximum length is 64 buckets.
    :param valence.vl_components.vdbench.ErrorInjectionConfig error_injection_config: Error injection
        configuration.
    :param valence.vl_components.vdbench.ReportingConfig reporting_config: Reporting configuration.
    :param valence.vl_components.vdbench.ValidationConfig validation_config: Data validation
        configuration.
    :param str monitor: Path to the file vdbench looks at to determine if you want to kill it.
    :param str comment: A comment to put above the line in the _parm file. Broken up into multiple lines
        based on the presence of '/n' characters.
    """

    def __init__(self, gd_config, host_defs, run_defs, workload_defs, storage_defs, port=None,
                 data_errors=0, histogram=None, error_injection_config=None, reporting_config=None,
                 validation_config=None, monitor=None, comment=None):
        super(GeneralDef, self).__init__(comment)
        self.config = gd_config
        self.host_defs = host_defs
        self.run_defs = run_defs
        self.workload_defs = workload_defs
        self.storage_defs = storage_defs

        self.port = port
        self.data_errors = data_errors
        self.histogram = None if histogram is None else tuple(histogram)
        self.error_injection_config = error_injection_config
        self.reporting_config = reporting_config
        self.validation_config = validation_config
        self.monitor = '/tmp/vdbench_kill' if monitor is None else monitor
        self.journaling_config = _enums.JournalingConfig.none

        if self.port is not None:
            assert self.port >= 0
        assert self.data_errors >= 0
        if self.histogram is not None:
            assert len(self.histogram) <= 64
            assert min(self.histogram) >= 0

    def configure_journaling(self, journal_dir, journaling_config):
        """Configure journaling.

        Updates the journal paths for each storage definition based on the specified journal directory.

        :param str journal_dir: Directory in which to keep journal files.
        :param valence.vl_components.vdbench.JournalingConfig journaling_config: Journaling
            configuration.
        """
        self.journaling_config = journaling_config
        for storage_def in self.storage_defs:
            journal_path = None
            if journaling_config is not _enums.JournalingConfig.none:
                journal_path = os.path.join(journal_dir, storage_def.name)
            storage_def.journal_path = journal_path

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        if self.histogram is None:
            histogram = None
        else:
            # noinspection PyTypeChecker
            histogram = ('default',) + self.histogram

        # Define the validation configuration
        if self.validation_config is _enums.ValidationConfig.none:
            validation = None
        else:
            validation = self.validation_config

        # Define the journaling configuration
        if self.journaling_config is _enums.JournalingConfig.none:
            journaling = None
        else:
            journaling = self.journaling_config

        config = []
        utils.extend_config(config, self.monitor, constants.MONITOR)
        utils.extend_config(config, self.port, constants.PORT)
        utils.extend_config(config, self.data_errors, constants.DATA_ERRORS)
        utils.extend_config(config, self.error_injection_config)
        utils.extend_config(config, self.reporting_config, constants.REPORT)
        utils.extend_config(config, histogram, constants.HISTOGRAM)
        utils.extend_config(config, validation, constants.VALIDATE)
        utils.extend_config(config, journaling, constants.JOURNAL)

        partial = self.config.get_parm_str()
        if partial:
            config.append(partial)
        for d in itertools.chain(self.host_defs, self.storage_defs, self.workload_defs, self.run_defs):
            config.append(d.get_parm_str())
        return '\n'.join(config)
