"""Vdbench parm file helper methods."""

from valence.vl_components.vdbench._parm import (
    gd,
    hd,
    rd,
    sd,
    types,
    wd,
)


def get_storage_def(name, io_workload, host_def, connections_cmt, standalone=False):
    """Get a storage definition based on an IO workload.

    :param str name: Unique name for the storage definition.
    :param valence.vl_components.vdbench.IoWorkload io_workload: IO workload to parse.
    :param valence.vl_components.vdbench._parm.HostDef host_def: Host definition for the IO workload.
    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for the IO workload.
    :param bool standalone: Whether Vdbench is being run in standalone mode (default: False).
        In standalone mode we don't list the host name on each storage definition because everything
        is run on localhost.
    :rtype: valence.vl_components.vdbench._parm.StorageDef
    """
    assert host_def.client_dm is io_workload.client_dm
    assert host_def.client_dm in connections_cmt.client_dms

    disk_map = connections_cmt.client_disk_map[io_workload.client_dm.mip]

    hit_area = None
    if io_workload.config.hit_config is not None:
        hit_area = io_workload.config.hit_config.hit_area
    alignment = None
    if isinstance(io_workload.config.transfer_size_config, types.TransferSizeAligned):
        alignment = io_workload.config.transfer_size_config.alignment

    sd_config = sd.StorageDefConfig(
        hit_area=hit_area,
        alignment=alignment,
        offset=io_workload.config.offset,
        thread_count=io_workload.config.thread_count,
        access_range_config=io_workload.config.access_range_config,
        streams_config=io_workload.config.streams_config,
        open_flags=io_workload.config.open_flags
    )

    # Get the LUN for the volume
    volume_dm = io_workload.volumes_cmt.volume_dms[io_workload.volume_index]
    lun = disk_map[volume_dm.volume_id]

    return sd.StorageDef(
        name=name,
        sd_config=sd_config,
        lun=lun,
        hd_name=host_def.name if not standalone else None
    )


def get_workload_def(name, io_workload_config, storage_defs):
    """Get a workload definition based on an IO workload.

    :param str name: Unique name for the workload definition.
    :param valence.vl_components.vdbench.IoWorkloadConfig io_workload_config: Workload configuration.
    :param list[valence.vl_components.vdbench._parm.StorageDef] storage_defs: Storage definitions for the
        IO workload.
    :rtype: valence.vl_components.vdbench._parm.WorkloadDef
    """
    hit_perc_read = None
    hit_perc_write = None
    if io_workload_config.hit_config is not None:
        hit_perc_read = io_workload_config.hit_config.read_perc
        hit_perc_write = io_workload_config.hit_config.write_perc

    wd_config = wd.WorkloadDefConfig(
        io_rate=io_workload_config.io_rate,
        read_perc=io_workload_config.read_perc,
        transfer_size_config=io_workload_config.transfer_size_config,
        seek_config=io_workload_config.seek_config,
        hit_perc_read=hit_perc_read,
        hit_perc_write=hit_perc_write,
        skew_perc=io_workload_config.skew_perc,
        priority=io_workload_config.priority
    )
    return wd.WorkloadDef(
        name=name,
        wd_config=wd_config,
        sd_names=[sd_def.name for sd_def in storage_defs]
    )


def get_run_def(name, io_period, workload_defs):
    """Get a run definition based on an IO period.

    :param str name: Unique name for the run definition.
    :param valence.vl_components.vdbench.IoPeriod io_period: IO period to parse.
    :param list[valence.vl_components.vdbench._parm.WorkloadDef] workload_defs: Workload definitions for
        the IO period.
    :rtype: valence.vl_components.vdbench._parm.RunDef
    """
    rd_config = rd.RunDefConfig(
        io_rate_config=io_period.config.io_rate_config,
        interval=io_period.config.interval,
        duration=io_period.config.duration,
        max_data=io_period.config.max_data,
        warmup=io_period.config.warmup,
        for_transfer_size_config=io_period.config.for_transfer_size_config,
        for_read_perc_config=io_period.config.for_read_perc_config,
        for_seek_perc_config=io_period.config.for_seek_perc_config,
        for_threads_config=io_period.config.for_threads_config
    )
    return rd.RunDef(
        name=name,
        rd_config=rd_config,
        wd_names=[wd_def.name for wd_def in workload_defs]
    )


def get_general_def(io_session, connections_cmt):
    """Get a general definition based on an IO session.

    :param valence.vl_components.vdbench.IoSession io_session: IO session to parse.
    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for the IO session.
    :rtype: valence.vl_components.vdbench._parm.GeneralDef
    """
    # Get the host definitions for the session
    hd_map = {}
    for i, conn_cmt in enumerate(connections_cmt.child_cmts):
        hd_map[conn_cmt.client_dm.mip] = hd.HostDef(
            name='h{:0>4}'.format(i),
            client_dm=conn_cmt.client_dm,
            jvms=conn_cmt.jvm_count
        )

    # Process each IO period for the session
    run_defs = []
    workload_defs = []
    storage_defs = []
    for period in io_session.io_periods:
        # Get all storage definitions for the period
        workload_sd_defs = {}
        for workload in period.io_workloads:
            storage_defs.append(get_storage_def(
                name='s{:0>4}'.format(len(storage_defs)),
                io_workload=workload,
                host_def=hd_map[workload.client_dm.mip],
                connections_cmt=connections_cmt,
                standalone=io_session.standalone
            ))

            if workload.config not in workload_sd_defs:
                workload_sd_defs[workload.config] = []
            workload_sd_defs[workload.config].append(storage_defs[-1])

        # Get all unique workload definitions for the period
        # NOTE: Combines io_workloads with identical configurations into single workload definitions

        period_wd_defs = []
        for config in workload_sd_defs:
            sd_index = 0
            while sd_index < len(workload_sd_defs[config]):
                period_wd_defs.append(get_workload_def(
                    name='w{:0>4}'.format(len(workload_defs) + len(period_wd_defs)),
                    io_workload_config=config,
                    storage_defs=workload_sd_defs[config]
                    [sd_index:sd_index + period.MAX_WORKLOADS]
                ))
                sd_index += period.MAX_WORKLOADS

        # Define the run definition for the period
        run_defs.append(get_run_def(
            name='r{:0>4}'.format(len(run_defs)),
            io_period=period,
            workload_defs=period_wd_defs
        ))
        workload_defs.extend(period_wd_defs)

    # Define the general definition for the entire IO session
    gd_config = gd.GeneralDefConfig(
        compression_ratio=io_session.config.compression_ratio,
        dedup_config=io_session.config.dedup_config,
        pattern_path=io_session.config.pattern_path
    )

    def sort_by(d):
        """Sort by name."""
        return d.name

    # sort by name that the parmfile is deterministic and easier to read.
    hds = sorted(hd_map.values(), key=sort_by)
    sds = sorted(storage_defs, key=sort_by)
    wds = sorted(workload_defs, key=sort_by)
    rds = sorted(run_defs, key=sort_by)
    return gd.GeneralDef(
        gd_config=gd_config,
        host_defs=hds if not io_session.standalone else [],
        run_defs=rds,
        storage_defs=sds,
        workload_defs=wds,
    )
