"""Vdbench run definition objects.

As defined in section 1.16 of the vdbench docs.

NOTE: We INTENTIONALLY do not support the for* syntax, as it adds complexity for defining rundefs and
can be achieved by simply permuting multiple rundefs (which is supported). In addition, it actually
reduces the flexibility of your workload definitions (as it arbitrarily overrides all of them).
"""
from __future__ import division

from valence.vl_components.vdbench._parm import (
    base,
    constants,
    types,
    utils,
)


class RunDefConfig(base.ParmLinePartial):
    """Vdbench run definition configuration.

    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration:
    :param int duration: Configured duration. Vdbench default is 30 seconds.
    :param int max_data: Configured max data.
    :param int interval: Configured interval. Vdbench default is ``min(elapsed/2, 60)``.
    :param int warmup: Configured warmup.
    :param valence.vl_components.vdbench.ForTransferSizeConfig for_transfer_size_config: For-xfersize
        configuration. **NOTE:** May extend the duration.
    :param valence.vl_components.vdbench.ForReadPercentConfig for_read_perc_config: For-read-perc
        configuration. **NOTE:** May extend the duration.
    :param valence.vl_components.vdbench.ForSeekPercentConfig for_seek_perc_config: For-seek-perc
        configuration. **NOTE:** May extend the duration.
    :param valence.vl_components.vdbench.ForThreadsConfig for_threads_config: For-threads configuration.
        **NOTE:** May extend the duration.
    """

    def __init__(self, io_rate_config=None, interval=None, duration=None, max_data=None, warmup=None,
                 for_transfer_size_config=None, for_read_perc_config=None, for_seek_perc_config=None,
                 for_threads_config=None):
        self.io_rate_config = io_rate_config
        self.interval = interval
        self.duration = duration
        self.max_data = max_data
        self.warmup = warmup
        self.for_transfer_size_config = for_transfer_size_config
        self.for_read_perc_config = for_read_perc_config
        self.for_seek_perc_config = for_seek_perc_config
        self.for_threads_config = for_threads_config

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        if isinstance(self.io_rate_config, types.IoRateCurve):
            elapsed = int(round(float(self.duration) / (1 + len(self.io_rate_config.curve))))
        elif isinstance(self.io_rate_config, types.IoRateSeries):
            elapsed = int(round(float(self.duration) / len(self.io_rate_config.io_rates)))
        else:
            elapsed = self.duration

        config = []
        utils.extend_config(config, self.io_rate_config)
        utils.extend_config(config, elapsed, constants.ELAPSED)
        utils.extend_config(config, self.max_data, constants.MAX_DATA)
        utils.extend_config(config, self.interval, constants.INTERVAL)
        utils.extend_config(config, self.warmup, constants.WARMUP)
        utils.extend_config(config, self.for_threads_config)
        utils.extend_config(config, self.for_transfer_size_config)
        utils.extend_config(config, self.for_read_perc_config)
        utils.extend_config(config, self.for_seek_perc_config)
        return ','.join(config)


class RunDef(base.ParmLine):
    """Vdbench run definition.

    :param str name: Name for this run definition.
    :param valence.vl_components.vdbench._parm.RunDefConfig rd_config: Configuration for this run
        definition.
    :param list[str] wd_names: The names of the workload definitions associated with this run definition.
    :param str comment: A comment to put above the line in the _parm file. Broken up into multiple lines
        based on the presence of '/n' characters.
    """

    def __init__(self, name, rd_config, wd_names, comment=None):
        super(RunDef, self).__init__(comment)
        self.name = name
        self.config = rd_config
        self.wd_names = wd_names

        assert len(self.wd_names) > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        if len(self.wd_names) == 1:
            wd, = self.wd_names
        else:
            wd = utils.tuple_str(self.wd_names)
        return '{}={},{}={},{}'.format(
            constants.RD, self.name,
            constants.WD, wd,
            self.config.get_parm_str()
        )
