"""Vdbench storage definition objects."""
from valence.vl_components.vdbench._parm import (
    base,
    constants,
    utils,
)


class StorageDefConfig(base.ParmLinePartial):
    """Vdbench storage definition configuration.

    :param int hit_area: Configured hit area.
    :param int alignment: Configured alignment. Serializes to `align`.
    :param int offset: Configured offset.
    :param int thread_count: Configured thread count. Serializes to `threads`.
    :param valence.vl_components.vdbench.AccessRangeConfig access_range_config: Configured
        access range. Serializes to `range`.
    :param valence.vl_components.vdbench.StreamsConfig streams_config: Streams configuration.
    :param str open_flags: Configured open flags.
    """

    def __init__(self, hit_area=None, alignment=None, offset=None, thread_count=None,
                 access_range_config=None, streams_config=None, open_flags=None):
        self.hit_area = hit_area
        self.alignment = alignment
        self.offset = offset
        self.thread_count = thread_count
        self.access_range_config = access_range_config
        self.streams_config = streams_config
        self.open_flags = open_flags

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        hit_area = None if self.hit_area is None else '{}k'.format(self.hit_area)

        config = []
        utils.extend_config(config, self.access_range_config)
        utils.extend_config(config, self.thread_count, constants.THREADS)
        utils.extend_config(config, hit_area, constants.HIT_AREA)
        utils.extend_config(config, self.offset, constants.OFFSET)
        utils.extend_config(config, self.alignment, constants.ALIGN)
        utils.extend_config(config, self.open_flags, constants.OPEN_FLAGS)
        utils.extend_config(config, self.streams_config)
        return ','.join(config)


class StorageDef(base.ParmLine):
    """Vdbench storage definition.

    :param str name: Name for this storage definition.
    :param valence.vl_components.vdbench._parm.StorageDefConfig sd_config: Configuration for this storage
        definition.
    :param str lun: The path to the storage lun.
    :param str hd_name: The name of the host definition to use.
    :param int size: Describes the size of the raw disk or file in kB (default: automatically determined
        from the raw disk or file).
    :param str journal_path: Directory name for the journal file. Used with journaling only. Journal
        files are needed for each Storage Definition and are created by default in the current directory.
        The file names are 'sdname.jnl' and 'sdname.map', where 'sdname' is the name of the SD.
    :param str comment: A comment to put above the line in the _parm file. Broken up into multiple lines
        based on the presence of '/n' characters.
    """

    def __init__(self, name, sd_config, lun, hd_name, size=None, journal_path=None, comment=None):
        super(StorageDef, self).__init__(comment)
        self.name = name
        self.config = sd_config
        self.lun = lun
        self.hd_name = hd_name
        self.size = size
        self.journal_path = journal_path

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        config = []
        utils.extend_config(config, self.name, constants.SD)
        if self.hd_name:
            utils.extend_config(config, self.hd_name, constants.HOST)
        utils.extend_config(config, self.lun, constants.LUN)
        utils.extend_config(config, self.journal_path, constants.JOURNAL)
        utils.extend_config(config, self.size, constants.SIZE)
        end = self.config.get_parm_str()
        if end:
            config.append(end)
        return ','.join(config)
