"""Vdbench data types."""
import abc

import six

from valence.vl_components.vdbench import (
    _base,
    _enums,
)
from valence.vl_components.vdbench._parm import (
    base,
    constants,
    utils,
)


@six.add_metaclass(abc.ABCMeta)
class AccessRangeConfig(base.ConfigItem):
    """IO rate configuration for an IO period."""


class AccessRangeBytes(AccessRangeConfig):
    """Byte range defining the portion of a volume on which to operate.

    Example: (bytes_start=1000, bytes_end=2000) will limit IO activity starting at 1000 bytes into the SD
    and ending at 2000 bytes into the SD.

    :param int bytes_start: Beginning of the byte range on a volume.
    :param int bytes_end: End of the byte range on a volume.
    """

    def __init__(self, bytes_start, bytes_end):
        self.bytes_start = bytes_start
        self.bytes_end = bytes_end

        assert self.bytes_start > 100
        assert self.bytes_end > self.bytes_start

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({},{})'.format(constants.RANGE, self.bytes_start, self.bytes_end)


class AccessRangePercentages(AccessRangeConfig):
    """Percentage range defining the portion of a volume on which to operate.

    Example: (percent_start=40, percent_end=60) will limit IO activity starting at 40% into the SD
    and ending at 60% into the SD.

    :param int percent_start: Beginning of the percent range on a volume.
    :param int percent_end: End of the percent range on a volume.
    """

    def __init__(self, percent_start, percent_end):
        self.percent_start = percent_start
        self.percent_end = percent_end

        assert 0 <= self.percent_start < self.percent_end <= 100

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({},{})'.format(constants.RANGE, self.percent_start, self.percent_end)


class CacheHitConfig(_base.BetterObject):
    """Cache hit configuration for an IO workload.

    Used when accessing raw devices or file systems mounted with 'forcedirectio' (or using open_flags).
    Each volume is divided into two parts:
    * Hit Area: The first megabyte of storage accessed for cache hits.
    * Miss Area: The remaining space on the SD will be accessed for cache misses.

    When Vdbench needs to generate a cache hit, it generates an I/O to the Hit Area, assuming that the
    data accessed is, or soon will be, residing in cache. Cache misses will be targeted toward the Miss
    Area, assuming that the miss area is large enough to ensure that most random accesses are cache
    misses. As you can see, this will only be useful when Vdbench has control over which physical volume
    and physical blocks will be ultimately read or written (so no file system cache). The 'hitarea=nn'
    and 'forhitarea=' parameters have been created to allow control over a volume's cache working set
    size. Some cached storage subsystems have different performance characteristics; if too small, a
    subset of the available cache is used. The total size of the LUN/SD must be at least 4 times the
    Hit Area size.

    :param int hit_area: Hit area in kilobytes (default: 1000). Must be < 1/4 of the total storage size.
        Controls the volume's cache working set size.
    :param int read_perc: Read hit percentage.
    :param int write_perc: Write hit percentage.
    """

    def __init__(self, hit_area=None, read_perc=None, write_perc=None):
        self.hit_area = hit_area
        self.read_perc = read_perc
        self.write_perc = write_perc

        assert not (self.read_perc is None and self.write_perc is None)
        if self.hit_area is not None:
            assert self.hit_area >= 1000
        if self.read_perc is not None:
            assert 0 <= self.read_perc <= 100
        if self.write_perc is not None:
            assert 0 <= self.write_perc <= 100

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(zip(self.__dict__.keys(), self.__dict__.values())))

    def __repr__(self):
        return '{}<{}={}, {}={}, {}={}>'.format(
            self.__class__.__name__,
            'hit_area', self.hit_area,
            'read_perc', self.read_perc,
            'write_perc', self.write_perc
        )


class DedupConfig(base.ConfigItem):
    """Data deduplication configuration for an IO session.

    Data Deduplication is built into Vdbench with the understanding that the dedup logic included in the
    target storage device looks at each n-byte data block to see if a block with identical content
    already exists. When there is a match the block no longer needs to be written to storage and a
    pointer to the already existing block is stored instead. Since it is possible for dedup and data
    compression algorithms to be used at the same time, dedup by default generates data patterns that do
    not compress.

    :param float dedup_ratio: Ratio between the original data and the actually written data
        (e.g. dedupratio=2 for a 2:1 ratio) (default: 1, no deduplication).
    :param int dedup_unit: The size in bytes of a data block that dedup tries to match with already
        existing data (default: 128k).
    :param int dedup_src_perc: Percent of all blocks from which to generate duplicates (default: 5%).
    """

    def __init__(self, dedup_ratio, dedup_unit=None, dedup_src_perc=None):
        self.dedup_ratio = dedup_ratio
        self.dedup_unit = dedup_unit
        if dedup_src_perc is None:
            self.dedup_src_perc = '5%'  # type: str
        elif 0 < dedup_src_perc <= 100:
            self.dedup_src_perc = '{}%'.format(dedup_src_perc)  # type: str
        else:
            raise ValueError('Vdbench requires a dedup sets percentage between 0 and 100!')

        # NOTE: By default Vdbench is supposed to source duplicates from 5% of blocks, but not setting
        # this results in inconsistent dedup efficiency on the cluster

        assert self.dedup_ratio >= 0
        if self.dedup_unit:
            assert self.dedup_unit > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        config = []
        utils.extend_config(config, self.dedup_ratio, constants.DEDUP_RATIO)
        utils.extend_config(config, utils.bytes_str(self.dedup_unit), constants.DEDUP_UNIT)
        utils.extend_config(config, self.dedup_src_perc, constants.DEDUP_SETS)
        return ','.join(config)


class ErrorInjectionConfig(base.ConfigItem):
    """Inject data validation errors into an IO session.

    :param int num_read: Inject errors after this many successful reads.
    :param int num_errors: The number of errors to inject.
    """

    def __init__(self, num_read, num_errors=1):
        self.num_read = num_read
        self.num_errors = num_errors

        assert self.num_read > 0
        assert self.num_errors > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({},{})'.format(constants.FORCE_ERROR_AFTER, self.num_read, self.num_errors)


@six.add_metaclass(abc.ABCMeta)
class ForReadPercentConfig(base.ConfigItem):
    """For-read-percent configuration for an IO period (overrides 'read_perc' for all workloads)."""


class ForReadPercentFixed(ForReadPercentConfig):
    """A single, fixed read percentage.

    :param int read_perc: Read percentage.
    """

    def __init__(self, read_perc):
        self.read_perc = read_perc

        assert 0 <= self.read_perc <= 100

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={}'.format(constants.FOR_READ_PERC, self.read_perc)


class ForReadPercentRange(ForReadPercentConfig):
    """A range of read percentages to be run in series.

    Automatically runs the IO period multiple times with each read percent, extending the duration of
    the IO period such that:
    total_duration = period.duration * len(read_percs).

    :param int read_perc_start: Starting read percent for the range.
    :param int read_perc_stop: Limiting read percent for the range.
    :param int increment: Increment by which to adjust the read percent on each subsequent run.
    """

    def __init__(self, read_perc_start, read_perc_stop, increment):
        self.read_perc_start = read_perc_start
        self.read_perc_stop = read_perc_stop
        self.increment = increment

        assert 0 <= self.read_perc_start <= 100
        assert 0 <= self.read_perc_stop <= 100

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({}-{},{})'.format(
            constants.FOR_READ_PERC, self.read_perc_start, self.read_perc_stop, self.increment
        )


@six.add_metaclass(abc.ABCMeta)
class ForSeekPercentConfig(base.ConfigItem):
    """For-seek-percent configuration for an IO period (overrides 'seek_perc' for all workloads)."""


class ForSeekPercentFixed(ForSeekPercentConfig):
    """A single, fixed seek percentage.

    :param int seek_perc: Seek percentage.
    """

    def __init__(self, seek_perc):
        self.seek_perc = seek_perc

        assert -1 <= self.seek_perc <= 100

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={}'.format(constants.FOR_SEEK_PERC, self.seek_perc)


class ForSeekPercentRange(ForSeekPercentConfig):
    """A range of seek percentages to be run in series.

    Automatically runs the IO period multiple times with each seek percent, extending the duration of
    the IO period such that:
    total_duration = period.duration * len(seek_percs).

    :param int seek_perc_start: Starting seek percent for the range.
    :param int seek_perc_stop: Limiting seek percent for the range.
    :param int increment: Increment by which to adjust the read percent on each subsequent run.
    """

    def __init__(self, seek_perc_start, seek_perc_stop, increment):
        self.seek_perc_start = seek_perc_start
        self.seek_perc_stop = seek_perc_stop
        self.increment = increment

        assert -1 <= self.seek_perc_start <= 100
        assert -1 <= self.seek_perc_stop <= 100

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({}-{},{})'.format(
            constants.FOR_SEEK_PERC, self.seek_perc_start, self.seek_perc_stop, self.increment
        )


@six.add_metaclass(abc.ABCMeta)
class ForThreadsConfig(base.ConfigItem):
    """For-threads configuration for an IO period (overrides 'threads' for all workloads)."""


class ForThreadsFixed(ForThreadsConfig):
    """A single, fixed thread count.

    :param int thread_count: Number of concurrent IO threads to use.
    """

    def __init__(self, thread_count):
        self.thread_count = thread_count

        assert self.thread_count > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={}'.format(constants.FOR_THREADS, self.thread_count)


class ForThreadsSeries(ForThreadsConfig):
    """A set of thread counts to be run in series.

    Automatically runs the IO period multiple times with each thread count, extending the duration of
    the IO period such that:
    total_duration = period.duration * len(thread_counts).

    :param typing.Iterable[int] thread_counts: Set of thread counts to use.
    """

    def __init__(self, thread_counts):
        self.thread_counts = tuple(thread_counts)  # type: tuple

        assert min(self.thread_counts) > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={}'.format(constants.FOR_THREADS, utils.tuple_str(self.thread_counts))


class ForThreadsRange(ForThreadsConfig):
    """A range of thread counts to be run in series.

    Automatically runs the IO period multiple times with each thread count, extending the duration of
    the IO period such that:
    total_duration = period.duration * len(thread_counts).

    :param int thread_count_start: Starting thread count for the range.
    :param int thread_count_stop: Limiting thread count for the range.
    :param int increment: Increment by which to adjust the thread count on each subsequent run.
    """

    def __init__(self, thread_count_start, thread_count_stop, increment):
        self.thread_count_start = thread_count_start
        self.thread_count_stop = thread_count_stop
        self.increment = increment

        assert self.thread_count_start > 0
        assert self.thread_count_stop > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({}-{},{})'.format(
            constants.FOR_THREADS, self.thread_count_start, self.thread_count_stop, self.increment
        )


@six.add_metaclass(abc.ABCMeta)
class TransferSizeConfig(base.ConfigItem):
    """Data transfer size configuration for an IO workload."""

    @property
    @abc.abstractmethod
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """

    @property
    def aligned(self):
        """Aligned transfer size configurations do not require 512 emulation.

        :rtype: bool
        """
        return all(xfersize % 4096 == 0 for xfersize in self.xfersizes)


@six.add_metaclass(abc.ABCMeta)
class ForTransferSizeConfig(TransferSizeConfig):
    """For-transfer-size configuration for an IO period (overrides 'xfersize' for all workloads)."""


class ForTransferSizeFixed(ForTransferSizeConfig):
    """A single, fixed data transfer size.

    :param int xfersize: Transfer size in bytes.
    """

    def __init__(self, xfersize=4096):
        self.xfersize = xfersize

        assert self.xfersize >= 0

    @property
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """
        return [self.xfersize]

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={}'.format(constants.FOR_XFERSIZE, utils.bytes_str(self.xfersize))


class ForTransferSizeSeries(ForTransferSizeConfig):
    """A set of transfer sizes to be run in series.

    Automatically runs the IO period multiple times with each transfer size, extending the duration of
    the IO period such that:
    total_duration = period.duration * len(xfersizes).

    :param typing.Iterable[int] xfersizes: Set of transfer sizes to use.
    """

    def __init__(self, xfersizes):
        self._xfersizes = tuple(xfersizes)  # type: tuple

        assert min(self._xfersizes) >= 0

    @property
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """
        return list(self._xfersizes)

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        xfersize_strings = [utils.bytes_str(xfersize) for xfersize in self._xfersizes]
        return '{}={}'.format(constants.FOR_XFERSIZE, utils.tuple_str(xfersize_strings))


class ForTransferSizeRange(ForTransferSizeConfig):
    """A range of transfer sizes to be run in series.

    Automatically runs the IO period multiple times with each transfer size, extending the duration of
    the IO period such that:
    total_duration = period.duration * (xfersize_stop - xfersize_start) / increment

    :param int xfersize_start: Starting transfer size for the range.
    :param int xfersize_stop: Limiting transfer size for the range.
    :param int increment: Increment by which to adjust the transfer size on each subsequent run.
    """

    def __init__(self, xfersize_start, xfersize_stop, increment):
        self.xfersize_start = xfersize_start
        self.xfersize_stop = xfersize_stop
        self.increment = increment

        assert self.xfersize_start >= 0
        assert self.xfersize_stop >= 0

    @property
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """
        return list(range(self.xfersize_start, self.xfersize_stop, self.increment))

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({}-{},{})'.format(
            constants.FOR_XFERSIZE,
            utils.bytes_str(self.xfersize_start),
            utils.bytes_str(self.xfersize_stop),
            utils.bytes_str(self.increment)
        )


@six.add_metaclass(abc.ABCMeta)
class IoRateConfig(base.ConfigItem):
    """IO rate configuration for an IO period."""


class IoRateCurve(IoRateConfig):
    """A set of IO rates to be run in series based on a percentage of maximum possible IO.

    Each workload will run as fast as possible at first in order to determine the maximum possible IO.
    This value is then used to benchmark the curve with the maximum possible IO being equivalent
    to curve=100.

    Once maximum possible IO has been determined, the IO rate will follow the curve as percentages of
    the maximum possible IO. The initial IO to determine max IO, along with each individual portion of
    the curve, will run for an equal portion of the total duration of the IO period (rounded up to
    the nearest second).

    NOTE: When determining maximum possible IO, if any skew is requested, all workloads will run only as
    fast as possible within the requested skew.

    :param typing.Iterable[int] curve: Series of percentages of max IO from which to configure the
        set of IO rates to use.
    :param valence.vl_components.vdbench.IoDistribution distribution: The distribution to use.
    """

    def __init__(self, curve=(10, 50, 70, 80, 90, 100), distribution=_enums.IoDistribution.exponential):
        self.curve = tuple(curve)  # type: tuple
        self.distribution = distribution

        assert min(self.curve) >= 0
        assert max(self.curve) <= 100

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={},{}={},{}={}'.format(
            constants.IO_RATE, constants.IO_RATE_CURVE,
            constants.IO_RATE_CURVE, utils.tuple_str(self.curve),
            constants.IO_DISTRIBUTION, self.distribution.value
        )


class IoRateFixed(IoRateConfig):
    """A single, fixed IO rate.

    :param int io_rate: IO rate.
    :param valence.vl_components.vdbench.IoDistribution distribution: The distribution to use.
    """

    def __init__(self, io_rate, distribution=_enums.IoDistribution.exponential):
        self.io_rate = io_rate
        self.distribution = distribution

        assert self.io_rate >= 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={},{}={}'.format(
            constants.IO_RATE, self.io_rate,
            constants.IO_DISTRIBUTION, self.distribution.value
        )


class IoRateMax(IoRateConfig):
    """A single IO rate based on the maximum rate possible.

    Allows each workload to run as fast as possible. This is called an uncontrolled max workload.

    NOTE: If any skew is requested, each workload will run only as fast as possible within the
    requested skew.

    :param valence.vl_components.vdbench.IoDistribution distribution: The distribution to use.
    """

    def __init__(self, distribution=_enums.IoDistribution.exponential):
        self.distribution = distribution

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={},{}={}'.format(
            constants.IO_RATE, constants.IO_RATE_MAX,
            constants.IO_DISTRIBUTION, self.distribution.value
        )


class IoRateSeries(IoRateConfig):
    """A set of IO rates to be run in series.

    Each IO rate will run for an equal portion of the total duration of the IO period (rounded up to
    the nearest second).

    :param typing.Iterable[int] io_rates: Set of IO rates.
    :param valence.vl_components.vdbench.IoDistribution distribution: The distribution to use.
    """

    def __init__(self, io_rates, distribution=_enums.IoDistribution.exponential):
        self.io_rates = tuple(io_rates)  # type: tuple
        self.distribution = distribution

        assert min(self.io_rates) >= 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        config = []
        utils.extend_config(config, self.io_rates, constants.IO_RATE)
        utils.extend_config(config, self.distribution, constants.IO_DISTRIBUTION)
        return ','.join(config)


class IoRateVariable(IoRateConfig):
    """A set of IO rates to be run in series with variable durations.

    :param list[tuple[int]] io_rates: A list of tuples for each IO rate and the duration for which it
        is to be used in seconds (io_rate, duration). The combined duration for all IO rates must be
        less than 3600s.
    :param valence.vl_components.vdbench.IoDistribution distribution: The distribution to use
        (must be 'variable' or 'spike').
    """

    def __init__(self, io_rates, distribution=_enums.IoDistribution.variable):
        self.io_rates = tuple(io_rate[0] for io_rate in io_rates)  # type: tuple
        self.io_rate_durations = tuple(io_rate[1] for io_rate in io_rates)
        self.distribution = distribution

        assert min(self.io_rates) >= 0
        assert min(self.io_rate_durations) >= 0

        assert distribution is not _enums.IoDistribution.exponential
        assert distribution is not _enums.IoDistribution.uniform
        assert distribution is not _enums.IoDistribution.deterministic

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        io_rates = []
        for i in range(len(self.io_rates)):
            io_rates.append(self.io_rates[i])
            io_rates.append(self.io_rate_durations[i])

        config = []
        utils.extend_config(config, io_rates, constants.IO_RATE)
        utils.extend_config(config, self.distribution, constants.IO_DISTRIBUTION)
        return ','.join(config)


class SeekConfig(base.ConfigItem):
    """Seek configuration for an IO workload.

    :param int seek_perc: Specifies how often a seek to a random lba will be generated. There are a
        couple "special" values::
        * 100: Every I/O will go to a different random seek address.
        * 0: There will be NO random seeks and the run will therefore be purely sequential. When the
        end of the volume is reached, I/O will continue at the beginning.
        * -1: The sequential workload will terminate as soon as the end-of-file is reached. vdbench will
        continue all workloads until all seekpct=-1 have reached EOF.
    :param tuple[int] stride: (min, max) of the stride. Stride changes the behavior when a new lba has
        to be generated because of the use of the seekpct= parameter. Instead of generating a brand new
        random lba, Vdbench skips 'n' blocks, 'n' being a random value between min and max, with a
        multiple of the current selected data transfer size. When end of volume is reached we start
        again at the beginning.
    """

    def __init__(self, seek_perc, stride=None):
        self.seek_perc = seek_perc
        self.stride = stride

        assert -1 <= self.seek_perc <= 100
        if stride is not None:
            assert len(self.stride) == 2
            assert min(self.stride) >= 0
            assert max(self.stride) <= 100
            assert self.stride[0] <= self.stride[1]

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        config = []
        utils.extend_config(config, self.seek_perc, constants.SEEK_PERC)
        utils.extend_config(config, self.stride, constants.STRIDE)
        return ','.join(config)


class StreamsConfig(base.ConfigItem):
    """Data stream configuration for an IO workload.

    Overrides the default processing done when reading or writing sequential data. By default Vdbench
    just reads or writes the whole lun or file, or only a portion of it when the seek_range is used.
    Streams allow multiple concurrent sequential IO.

    This parameter works together with the threads= parameter. The file or lun is split into 'n' pieces,
    where 'n' equals threads/threads_per_stream.

    EXAMPLE: size=100g, threads=8, streams=2

    The SD is split into (threads=8 / streams=2) = 4 equal streams of 25g each.
    Each stream uses two threads; the streams starting at 0%, 25, 50, and 75%
    within the SD.

    NOTE: Though the starting lba of each stream will be the same, the size of the streams will be set
    to 'size'.

    NOTE: CacheHitConfig is ignored when streams are requested. However, if you really want to run cache
    hits with streams, just code size='small' to force all Vdbench to cache.

    :param int streams: Number of streams to use PER THREAD.
    :param int size: Size of each stream in kilobytes.
    """

    def __init__(self, streams, size=None):
        self.streams = streams
        self.size = size

        assert streams > 0
        if self.size is not None:
            assert self.size > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        if self.size is None:
            return '{}={}'.format(constants.STREAMS, self.streams)
        else:
            return '{}=({},{}k)'.format(constants.STREAMS, self.streams, self.size)


class TransferSizeAligned(TransferSizeConfig):
    """A set of aligned data transfer sizes.

    * Dedup ratio must NOT be specified.

    :param int min_xfersize: Minimum transfer size in bytes.
    :param int max_xfersize: Maximum transfer size in bytes.
    :param int alignment: Alignment in bytes.
    """

    def __init__(self, min_xfersize, max_xfersize, alignment):
        self.min_xfersize = min_xfersize
        self.max_xfersize = max_xfersize
        self.alignment = alignment

        assert self.min_xfersize >= 0
        assert self.max_xfersize >= 0
        assert self.alignment >= 0

    @property
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """
        return list(range(self.min_xfersize, self.max_xfersize, self.alignment))

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}=({},{},{})'.format(
            constants.XFERSIZE,
            utils.bytes_str(self.min_xfersize),
            utils.bytes_str(self.max_xfersize),
            self.alignment
        )


class TransferSizeFixed(TransferSizeConfig):
    """A single, fixed data transfer size.

    :param int xfersize: Transfer size in bytes.
    """

    def __init__(self, xfersize=4096):
        self.xfersize = xfersize

        assert self.xfersize >= 0

    @property
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """
        return [self.xfersize]

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        return '{}={}'.format(constants.XFERSIZE, utils.bytes_str(self.xfersize))


class TransferSizeVariable(TransferSizeConfig):
    """A set of weighted data transfer sizes.

    :param list[tuple[int]] weighted_xfersizes: A list of tuples for each transfer size in bytes and its
        associated percentage for the overall workload IO (xfersize, perc).
    :raises ValueError: If the sum of the percentages is not equal to 100.
    """

    def __init__(self, weighted_xfersizes):
        self.weighted_xfersizes = tuple(weighted_xfersizes)  # type: tuple

        assert all(len(xfersize_tuple) == 2 for xfersize_tuple in self.weighted_xfersizes)
        assert all(xfersize >= 0 for xfersize, _ in self.weighted_xfersizes)
        assert all(0 <= perc <= 100 for _, perc in self.weighted_xfersizes)
        assert sum(perc for _, perc in self.weighted_xfersizes) == 100

    @property
    def xfersizes(self):
        """All transfer sizes specified for this configuration.

        :rtype: list[int]
        """
        return [xfersize for xfersize, _ in self.weighted_xfersizes]

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        xfersize_strings = [
            '{},{}'.format(utils.bytes_str(xfersize), str(perc))
            for xfersize, perc in self.weighted_xfersizes
        ]
        return '{}={}'.format(
            constants.XFERSIZE, utils.tuple_str(xfersize_strings)
        )
