"""Vdbench workload definition objects.

As defined in section 1.15 of the vdbench docs.
"""
from valence.vl_components.vdbench._parm import (
    base,
    constants,
    utils,
)


class WorkloadDefConfig(base.ParmLinePartial):
    """Vdbench workload definition configuration.

    :param int io_rate: Configured IO rate.
    :param int read_perc: Configured read percentage.
    :param valence.vl_components.vdbench.TransferSizeConfig transfer_size_config: Transfer size
        configuration.
    :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration.
    :param int|None hit_perc_read: Configured read hit percentage.
    :param int|None hit_perc_write: Configured write hit percentage.
    :param int skew_perc: Configured skew percentage.
    :param int priority: Configured priority.
    """

    def __init__(
            self, io_rate=None, read_perc=None, transfer_size_config=None, seek_config=None,
            hit_perc_read=None, hit_perc_write=None, skew_perc=None, priority=None):
        self.io_rate = io_rate
        self.read_perc = read_perc
        self.transfer_size_config = transfer_size_config
        self.seek_config = seek_config
        self.hit_perc_read = hit_perc_read
        self.hit_perc_write = hit_perc_write
        self.skew_perc = skew_perc
        self.priority = priority

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        config = []
        utils.extend_config(config, self.read_perc, constants.READ_PERC)
        utils.extend_config(config, self.io_rate, constants.IO_RATE)
        utils.extend_config(config, self.priority, constants.PRIORITY)
        utils.extend_config(config, self.seek_config)
        utils.extend_config(config, self.hit_perc_read, constants.READ_HIT_PERC)
        utils.extend_config(config, self.hit_perc_write, constants.WRITE_HIT_PERC)
        utils.extend_config(config, self.skew_perc, constants.SKEW)
        utils.extend_config(config, self.transfer_size_config)
        return ','.join(config)


class WorkloadDef(base.ParmLine):
    """Vdbench workload definition.

    :param str name: Name for this workload definition.
    :param valence.vl_components.vdbench._parm.WorkloadDefConfig wd_config: Configuration for this
        workload definition.
    :param list[str] sd_names: The SD names associated with this workload definition.
    :param str comment: A comment to put above the line in the _parm file. Broken up into multiple lines
        based on the presence of '/n' characters.
    """

    def __init__(self, name, wd_config, sd_names, comment=None):
        super(WorkloadDef, self).__init__(comment)
        self.name = name
        self.config = wd_config
        self.sd_names = sd_names

        assert len(sd_names) > 0

    def _parm_str(self):
        """The specific representation of this item as it may appear in the Vdbench parm file.

        :rtype: str
        """
        if len(self.sd_names) == 1:
            sd = self.sd_names[0]
        else:
            sd = self.sd_names
        config = []
        utils.extend_config(config, self.name, constants.WD)
        utils.extend_config(config, sd, constants.SD)

        config_str = self.config.get_parm_str()
        if config_str:
            config.append(config_str)

        return ','.join(config)
