"""Vdbench run components."""
import abc
import os
import time
from collections import OrderedDict

import autologger
import icechai_sdk
import sf_artifacts
import sf_ssh
import six

from valence import (
    components,
    errors,
    functional_models,
)
from valence.vl_components.vdbench import (
    _enums,
    _errors,
    _helpers,
    _log_parser,
    _parm,
    _stream_handlers,
)

_log = autologger.get_sflogging(__name__)

VDBENCH_ID = 'vdbench_id'
VDBENCH_DIR = 'output_dir'
VDBENCH_LOG_FILE = 'log_file_path'
VDBENCH_STATUS = 'status'


@six.add_metaclass(abc.ABCMeta)
class RunVdbsInterface(components.NonBlockingApplyKillable, components.Reversible):
    """Interface component for running Vdbench."""
    TIMEOUT = 2 * 24 * 60 * 60

    @abc.abstractmethod
    def assert_valid(self, is_running=True):
        """Assert that io is valid.

        :param bool is_running: Is IO running.
        """

    @abc.abstractmethod
    def update_stream(self):
        """Get new data and send it to all stream handlers."""

    @property
    @abc.abstractmethod
    def child_cmts(self):
        """Get all vdb-run components (i.e. no manage components).

        :rtype: list[valence.vl_components.vdbench.RunVdb]
        """


class RunVdb(RunVdbsInterface):
    """Run Vdbench IO from a single master instance.

    :param valence.vl_components.vdbench.SetupVdbEnvironment environment_cmt: Vdbench environment for a
        single master instance.
    :param int timeout: Seconds after which this component will timeout (default: 2 days).
    :param int port: Override the default port.
    :param int data_errors: Terminate after this number of data validation or IO errors.
    :param valence.vl_components.vdbench.ErrorInjectionConfig error_injection_config: Error injection
        configuration.
    :param valence.vl_components.vdbench.ReportingConfig reporting_config: Reporting configuration.
    :param valence.vl_components.vdbench.ValidationConfig validation_config: Data validation
        configuration (default: On-Read validation).
    :param valence.vl_components.vdbench.JournalingConfig journaling_config: Journaling configuration
        (default: None).
    :param bool wait_for_io_start: Whether to wait for IO to start running during apply (default: True).
    :param float wait_for_io_timeout: Maximum amount of time to wait until IO starts (default: 1 day for
        journaling, 10 minutes without).
    """
    MAX_LINES_SSH = 2000  # Maximum number of lines to transmit over ssh

    def __init__(self, environment_cmt, timeout=None, port=None, data_errors=0,
                 error_injection_config=None, reporting_config=None, validation_config=None,
                 journaling_config=None, wait_for_io_start=True, wait_for_io_timeout=None):
        super(RunVdb, self).__init__()
        self.environment_cmt = environment_cmt
        self.port = port
        self.data_errors = data_errors
        self.error_injection_config = error_injection_config
        self.reporting_config = reporting_config
        self.validation_config = validation_config or _enums.ValidationConfig.on_reads
        self.journaling_config = journaling_config or _enums.JournalingConfig.none

        self._master_sdk = None
        self._slave_sdks = None
        self._vdb_parm = None
        self._vdbench_id = None
        self._logfile = None
        self._vdbench_dir = None
        self._last_line = None
        self._last_line_read = 0

        if timeout is not None:
            self.TIMEOUT = timeout

        self._wait_for_io_start = wait_for_io_start
        self._wait_for_io_timeout = wait_for_io_timeout
        if self._wait_for_io_timeout is None:
            if journaling_config in {_enums.JournalingConfig.recover, _enums.JournalingConfig.validate}:
                # Wait for up to a day if the journal is being recovered
                self._wait_for_io_timeout = 24 * 60 * 60
            else:
                # Default: 10 minutes
                self._wait_for_io_timeout = 10 * 60

        # State that is not changed on reset
        self._stream_handlers = OrderedDict()
        self._error_handler = _stream_handlers.ErrorStreamHandler(self)
        self._run_info_handler = _stream_handlers.RunInfoHandler(self)

        # Check for unsupported validation configurations
        if self.validation_config is not _enums.ValidationConfig.none:
            self._validate_periods_for_data_validation(self.environment_cmt.io_session.io_periods)

        # TODO: Add support for Windows clients (VL-2123)
        windows_clients = [dm for dm in environment_cmt.client_dms if dm.os.is_windows]
        if windows_clients:
            raise NotImplementedError(
                'Windows clients are not yet supported: {}'.format(windows_clients)
            )

        self._journal_validate_vdb_cmt = None
        self._setup_journal_validation()

    @staticmethod
    def _validate_periods_for_data_validation(io_periods):
        """Ensure each IO period supports data validation.

        :param list[valence.vl_components.vdbench.IoPeriod] io_periods: IO periods.
        """
        for period in io_periods:
            for workload in period.io_workloads:
                if isinstance(workload.config.transfer_size_config, _parm.TransferSizeAligned):
                    raise ValueError('Vdbench does not support validation of aligned-xfersize IO')
                elif workload.config.offset is not None:
                    raise ValueError('Vdbench does not support validation of offset IO')

    @property
    def child_cmts(self):
        """Get all vdb-run components (i.e. no manage components).

        :rtype: list[valence.vl_components.vdbench.RunVdb]
        """
        return [self]

    @property
    def client_dms(self):
        """Returns the client datamodels associated with this component.

        :rtype: list[valence.models.ResourceClient]
        """
        return self.environment_cmt.client_dms

    @property
    def master_client_dm(self):
        """Get the master client for this vdbench config.

        :rtype: valence.models.ResourceClient
        """
        return self.environment_cmt.master_client_dm

    @property
    def current_io_line_item(self):
        """Get the current IO line from the Vdbench logs.

        :rtype: valence.vl_components.vdbench._log_parser.VdbLogLine
        """
        self.update_stream()
        return self._run_info_handler.current_io_line_item

    def get_vdbench_log(self, max_lines, start_line=0):
        """Return the vdbench logfile starting at start_line.

        :param int max_lines: Returns the first N lines up to N=max_lines.
        :param int start_line: Line to start log at.
        :rtype: list[str]
        """
        return _helpers.tail_vdb_log(self.master_client_dm, self._logfile, start_line, max_lines)

    def update_stream(self):
        """Get new data and send it to all stream handlers.

        This should rarely (if ever) need to be called in a test case. It is most frequently called by
        the stream handlers themselves, which depend on the vdbench data stream in order to make their
        own assertions.
        """
        if self.state is components.States.ready:
            raise components.ComponentStateError('Cannot update_stream in ready state.')
        while True:
            raw_log = self.get_vdbench_log(self.MAX_LINES_SSH, self._last_line_read)
            lines = _log_parser.parse_raw(
                raw_lines=raw_log,
                last_line=self._last_line,
                defined_intervals=self.environment_cmt.defined_intervals
            )
            if lines:
                self._last_line = lines[-1]
            self._last_line_read += len(lines)

            _log.debug('Vdbench updating streams with {} new lines'.format(len(lines)))
            for handler in six.itervalues(self._stream_handlers):
                handler.handle_items(lines)

            if len(raw_log) < self.MAX_LINES_SSH:
                # we did not receive the maximum number of lines
                # so we must have received everything it had
                break

    def assert_valid(self, is_running=True):
        """Assert that the Vdbench instance is valid.

        :param bool is_running: if True then also assert that the component is still running
            (typical use case).
        """
        self._error_handler.assert_valid(self.data_errors)
        if is_running and not self.is_running():
            raise _errors.VdbenchRuntimeError('Vdbench is not running when it should be.')
        elif not is_running and self.is_running():
            raise _errors.VdbenchRuntimeError("Vdbench is running when it shouldn't be.")

    def wait_for_io_running(self, start_time):
        """Wait for Vdbench IO to start.

        :param float start_time: The time (in seconds since the epoch) when Vdbench was initiated.
        """
        _log.debug('Waiting up to {}s for IO to start'.format(self._wait_for_io_timeout))
        while True:
            if not self.is_running():
                raise _errors.VdbenchRuntimeError(
                    'Vdbench instance is no longer running:\n{}'
                    .format(self._error_handler.get_errors_string())
                )
            if self.current_io_line_item:
                break
            if time.time() - start_time > self._wait_for_io_timeout:
                status = self._master_sdk.get_vdbench_master_status(self._vdbench_id)
                raise errors.TimeoutError(
                    'IO failed to start after {}s with master status: {}'
                    .format(self._wait_for_io_timeout, status)
                )

            time.sleep(20)

    def store_artifacts(self, get_bundle=True, get_vdb_dir=False):
        """Collect vdbench artifacts.

        :param bool get_bundle: Whether to collect the full master client bundle.
        :param bool get_vdb_dir: Whether to collect the Vdbench output directory contents (includes the
            flatfile and other supplementary logs).
        """
        client_fm = functional_models.get_client_from_dm(self.master_client_dm)
        manager = sf_artifacts.get_global_manager()

        _log.info('Storing Vdbench artifacts')
        if self._vdb_parm:
            manager.store_text(self._vdb_parm, suffix='_vdbench.config')
        # Why?  Because clientserver will make a copy of vdbench_master.log as
        # TIMESTAMP_vdbench_master.log, but we don't know the timestamp.
        # TODO: when AA-4302 is completed just make a call to ice chai here
        with client_fm.get_ssh_connection() as ssh_conn:
            logfiles = ssh_conn.run_command('ls {}/*_master.log'.format(self._vdbench_dir))
            logfiles = logfiles.stdout.split('\n')[:-1]
        if logfiles:
            manager.store_artifacts_remote(logfiles, self.master_client_dm)
        if get_bundle:
            client_fm.pull_support_bundle(manager)
        if get_vdb_dir:
            try:
                filename = '{}-{}.tar.bz2'.format(
                    self.master_client_dm.client_name,
                    os.path.basename(self._vdbench_dir)
                )
                # Create a tar file of the Vdbench output directory contents
                tar_file = manager.compress_directory(
                    directory=self._vdbench_dir,
                    excludes=['*.mmap', '*.map', '*.jnl', '*swat_mon.txt'],
                    remote_host_dm=self.master_client_dm,
                    destination=filename
                )
                manager.store_artifacts_remote([tar_file], self.master_client_dm)
            except sf_artifacts.ArtifactsError as exc:
                raise _errors.VdbenchError('Failed getting vdbench directory artifacts: {}'.format(exc))
            else:
                # Remove the tar file after storing it
                with client_fm.get_ssh_connection() as ssh_conn:
                    ssh_conn.run_command('rm {}'.format(tar_file))

    def set_test_identifier(self, test_instance_id):
        """Set the unique identifier for a vdbench execution instance.

        Call this after the vdbench thread has been started on the master and before
        any artifacts have been captured.

        :param str|uuid test_instance_id: globally unique test instance id str(uuid)
        """
        self._master_sdk.set_test_identifier(test_identifier=str(test_instance_id))

    def validate_journal(self):
        """Validate the vdbench journal file if the component is not running and has been applied.

        :raise valence.components.ComponentStateError: if component is not stopped
        :raise valence.components.ComponentStateError: if not configured for validation
        :return:
        """
        if self.state not in [components.States.applied, components.States.failed]:
            raise components.ComponentStateError(
                'Component must have been applied and stopped before journal validation')
        if self._journal_validate_vdb_cmt is None:
            raise components.ComponentStateError('Journal validation not configured')
        # Apply our child component which was with our environment cmt and configured to validate only
        self._journal_validate_vdb_cmt.apply()
        self._journal_validate_vdb_cmt.assert_valid(is_running=False)
        self._journal_validate_vdb_cmt.reset()

    def _apply_async(self, timeout=None):
        """Apply the component asynchronously.

        :param int timeout: timeout to wait for vdbench to start running
        """
        timeout = timeout or 60
        slave_client_dms = [dm for dm in self.client_dms if dm is not self.master_client_dm]
        self._slave_sdks = [icechai_sdk.SDK(c.mip) for c in slave_client_dms]
        self._master_sdk = icechai_sdk.SDK(self.master_client_dm.mip)
        if not self.environment_cmt.io_session.standalone and self.master_client_dm in self.client_dms:
            slave_client_dms.append(self.master_client_dm)

        _log.debug('Terminating any active Vdbench sessions on clients: {}'.format(self.client_dms))
        self._stop_all_vdbench()
        self.environment_cmt.ensure_applied()

        _log.info('Starting Vdbench slaves on clients: {}'.format(slave_client_dms))
        if self.master_client_dm in slave_client_dms:
            self._master_sdk.run_vdbench_slave()
        for slave_sdk in self._slave_sdks:
            slave_sdk.run_vdbench_slave()

        _log.info('Starting Vdbench master on client: {}'.format(self.master_client_dm))
        self._vdb_parm = self.environment_cmt.get_vdb_parm(
            port=self.port,
            data_errors=self.data_errors,
            error_injection_config=self.error_injection_config,
            reporting_config=self.reporting_config,
            validation_config=self.validation_config,
            journaling_config=self.journaling_config
        )
        self._setup_journaling()
        vdb_result = self._master_sdk.run_vdbench_master(self._vdb_parm)
        self._vdbench_id = vdb_result[VDBENCH_ID]
        self._vdbench_dir = vdb_result[VDBENCH_DIR]
        self._logfile = vdb_result[VDBENCH_LOG_FILE]

        time_start = time.time()
        while not self.is_running():
            if time.time() - time_start > timeout:
                raise errors.TimeoutError('Timed out waiting for Vdbench master to start')
            else:
                time.sleep(1)

        _log.debug('Vdbench run initiated: vdbench_id={}\n** VDBENCH PARM **\n    {}'.format(
            self._vdbench_id, '\n    '.join(self._vdb_parm.split('\n'))
        ))
        if self._wait_for_io_start:
            self.wait_for_io_running(time.time())

    def _apply_is_running(self):
        """Check whether Vdbench is still running.

        NOTE: this does not currently handle if the client was rebooted.

        :rtype: bool
        """
        if self._vdbench_id is None:
            return False
        return self._master_is_running(self._master_sdk, self._vdbench_id)

    def _kill_apply_async(self):
        """Stop vdbench."""
        if self._vdbench_id:
            self._stop_all_vdbench(self._vdbench_id)
            self._vdbench_id = None  # calling multiple times is a NOOP

    def _reset(self, timeout=None):
        """Reset the component.

        :param int timeout: Maximum time to reset.
        """
        # Store log and clean up output directory if an artifact manager is defined
        if self._vdbench_dir and sf_artifacts.GlobalManager.MANAGER:
            self.store_artifacts(get_bundle=False)
            with sf_ssh.Ssh(self.master_client_dm.mip) as conn:
                conn.run_command('rm -r "{}"'.format(self._vdbench_dir))

        for handler in six.itervalues(self._stream_handlers):
            handler.clear()

        self._vdb_parm = None
        self._vdbench_id = None
        self._logfile = None
        self._vdbench_dir = None
        self._last_line = None
        self._last_line_read = 0

        if self._journal_validate_vdb_cmt:
            self._journal_validate_vdb_cmt.ensure_ready(timeout)

    @staticmethod
    def _master_is_running(sdk, master_id):
        """Check whether the Vdbench master is still running.

        :param icechai_sdk.SDK sdk: SDK for the Vdbench run.
        :param int master_id: ID of the Vdbench master instance.
        :rtype: bool
        """
        status = sdk.get_vdbench_master_status(str(master_id))
        return _enums.VdbStatus(status[VDBENCH_STATUS]).is_running

    @staticmethod
    def _slave_is_running(sdk):
        """Check whether the Vdbench slaves are still running.

        :param icechai_sdk.SDK sdk: SDK for the Vdbench run.
        :rtype: bool
        """
        status = sdk.get_vdbench_slave_status()
        return _enums.VdbStatus(status[VDBENCH_STATUS]).is_running

    def _stop_all_vdbench(self, master_id=None):  # pylint: disable=too-complex
        """Stop Vdbench master and slaves.

        :param int master_id: ID of the Vdbench master instance (default: Gets the active instance).
        """
        if not master_id:
            # If master_id is set to None that means we don't know about any previous vdbench masters
            # (but they might exist)
            master_id = self._master_sdk.get_vdbench_id()[VDBENCH_ID]
        if master_id:
            # If master_id is _still_ None, that means that the master simply isn't running.
            self._master_sdk.stop_vdbench_master(master_id, kill=True)
            start = time.time()
            while self._master_is_running(self._master_sdk, master_id):
                if time.time() - start > 10 * 60:
                    raise errors.TimeoutError(
                        'Timed out waiting for Vdbench master to die: {}'
                        .format(self.master_client_dm)
                    )
                else:
                    time.sleep(10)

        # We always need to kill slaves.
        # NOTE: If slaves are already killed this NEVER returns an error.
        self._master_sdk.stop_vdbench_slave(kill=True)
        for slave_sdk in self._slave_sdks:
            slave_sdk.stop_vdbench_slave(kill=True)

        alive_slaves = [sdk for sdk in self._slave_sdks if self._slave_is_running(sdk)]
        if self._slave_is_running(self._master_sdk):
            alive_slaves.append(self._master_sdk)

        start = time.time()
        while alive_slaves:
            if time.time() - start > 10 * 60:
                raise errors.TimeoutError(
                    'Timed out waiting for Vdbench slaves to die: {}'
                    .format([sdk.ip_address for sdk in alive_slaves])
                )
            else:
                time.sleep(10)
                alive_slaves = [sdk for sdk in alive_slaves if self._slave_is_running(sdk)]

    def _setup_journaling(self):
        """Setup the journaling directories on the master client."""
        if self.journaling_config is not _enums.JournalingConfig.none:
            self._master_sdk.make_temp_directory(self.environment_cmt.JOURNAL_DIR)
            self._master_sdk.make_temp_directory(os.path.join(
                self.environment_cmt.JOURNAL_DIR,
                self.master_client_dm.client_name
            ))
            journal_paths = self.environment_cmt.journal_paths
            for journal_path in journal_paths:
                self._master_sdk.make_temp_directory(
                    directory_path=journal_path,
                    remove_existing=self.journaling_config is _enums.JournalingConfig.full
                )

    def _setup_journal_validation(self):
        """Enable validation of data after the component is stopped.

        This is useful for validating data after failover/recovery scenarios.
        """
        if self.journaling_config in [_enums.JournalingConfig.full,
                                      _enums.JournalingConfig.async_io]:
            # Create a child cmt which will use the journal file from this cmt to validate data.
            vdb = RunVdb(
                environment_cmt=self.environment_cmt,
                journaling_config=_enums.JournalingConfig.validate)
            self._journal_validate_vdb_cmt = vdb

    def _add_stream_handler(self, handler):
        """Add a stream handler to the Vdbench run.

        Stream handlers are objects which have methods that will be called EVERY time a log-line is
        analyzed from vdbench_master.log. They are useful for tasks such as:
        * Checking for errors in the log.
        * Latency checking.
        * Monitoring and statistics gathering.

        :param valence.vl_components.vdbench._stream_handlers.BaseVdbStreamHandler handler: Vdbench
            stream handler to add.
        """
        if not isinstance(handler, _stream_handlers.BaseVdbStreamHandler):
            raise TypeError(handler)
        if not handler.CAN_REGISTER:
            raise TypeError('{} cannot be registered'.format(handler.__class__.__name__))
        self._stream_handlers[id(handler)] = handler

    def _remove_stream_handler(self, handler):
        """Remove a stream handler from the Vdbench run.

        :param valence.vl_components.vdbench._stream_handlers.BaseVdbStreamHandler handler: Vdbench
            stream handler to remove.
        :rtype: collections.OrderedDict
        """
        if not isinstance(handler, _stream_handlers.BaseVdbStreamHandler):
            raise TypeError(handler)
        if not handler.CAN_REGISTER:
            raise TypeError('{} cannot be registered'.format(handler.__class__.__name__))
        return self._stream_handlers.pop(id(handler))

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'master_client_name', self.environment_cmt.master_client_dm.client_name
        )


class ManageRunVdbs(RunVdbsInterface):
    """Manage Vdbench IO from multiple master instances.

    Also handles constructing multiple RunVdb components from a single BaseSetupVdbs
    component (such as ManageSetupVdbs).

    :param list[valence.vl_components.vdbench.RunVdb] run_vdb_cmts: List of RunVdb components.
    :param int timeout: the timeout of this component in seconds. Default is 2 days.
    """

    def __init__(self, run_vdb_cmts, timeout=None):
        super(ManageRunVdbs, self).__init__()
        self._run_vdb_cmts = []
        for cmt in run_vdb_cmts:
            self._run_vdb_cmts.extend(cmt.child_cmts)

        if timeout is not None:
            self.TIMEOUT = timeout

        self._validate_init()
        self.environments_cmt = None

    @classmethod
    def from_environment_cmt(cls, environments_cmt, timeout=None, port=None, data_errors=0,
                             error_injection_config=None, reporting_config=None, validation_config=None,
                             journaling_config=None, wait_for_io_running_time=None):
        """Construct a Vdbench IO manager from a set of Vdbench environments.

        This applies the given kwargs (i.e. timeout, port, data_errors) to each created
        RunVdb component created.

        :param valence.vl_components.vdbench._setup.VdbEnvironmentsInterface environments_cmt: A set
            of Vdbench environments.
        :param int timeout: The timeout of the child run components in seconds. Default is 2 days.
        :param int port: Override the default port.
        :param int data_errors: Terminate after this number of data validation or IO errors.
        :param valence.vl_components.vdbench.ErrorInjectionConfig error_injection_config: Error injection
            configuration.
        :param valence.vl_components.vdbench.ReportingConfig reporting_config: Reporting configuration.
        :param valence.vl_components.vdbench.ValidationConfig validation_config: Data validation
            configuration (default: On-Read validation).
        :param valence.vl_components.vdbench.JournalingConfig journaling_config: Journaling
            configuration (default: None).
        :param float wait_for_io_running_time: Maximum amount of time to wait until IO starts (default:
            Uses a sensibly 'long' value based on the journaling configuration).
        :rtype: valence.vl_components.vdbench.ManageRunVdbs
        """
        run_vdbs = []
        for cmt in environments_cmt.child_cmts:
            run_vdbs.append(RunVdb(
                environment_cmt=cmt,
                timeout=timeout,
                port=port,
                data_errors=data_errors,
                error_injection_config=error_injection_config,
                reporting_config=reporting_config,
                validation_config=validation_config,
                journaling_config=journaling_config,
                wait_for_io_timeout=wait_for_io_running_time
            ))
        run_vdb_manager = cls(run_vdbs, timeout=timeout)
        run_vdb_manager.environments_cmt = environments_cmt

        return run_vdb_manager

    @property
    def child_cmts(self):
        """Get all vdb-run components (i.e. no manage components).

        :rtype: list[valence.vl_components.vdbench.RunVdb]
        """
        return self._run_vdb_cmts

    @property
    def client_dms(self):
        """Get every client associated with these Vdbench run definitions.

        :rtype: list[valence.models.ResourceClient]
        """
        client_dms = []
        for cmt in self._run_vdb_cmts:
            client_dms.extend(cmt.client_dms)
        return client_dms

    def validate_journal(self):
        """Validate data against the vdbench journal file."""

        for cmt in self.child_cmts:
            cmt.validate_journal()

    def update_stream(self):
        """Get new data and send it to all stream handlers."""
        for run_vdb in self._run_vdb_cmts:
            run_vdb.update_stream()

    def assert_valid(self, is_running=True):
        """Call assert_valid for each run_vdb component.

        :param bool is_running: if True then also assert that they
            are still running (typical use case).
        """
        for run_vdb in self._run_vdb_cmts:
            run_vdb.assert_valid(is_running)

    def set_test_identifier(self, test_instance_id):
        """Set the unique identifier for a vdbench execution instance.

        Call this after the vdbench thread has been started on the master and before
        any artifacts have been captured.

        :param str test_instance_id: globally unique test instance id str(uuid)
        """
        for run_vdb in self._run_vdb_cmts:
            run_vdb.set_test_identifier(test_instance_id=str(test_instance_id))

    def store_artifacts(self, get_bundle=True, get_vdb_dir=False):
        """Collect vdbench artifacts.

        :param bool get_bundle: Whether to collect the full master client bundle.
        :param bool get_vdb_dir: Whether to collect the Vdbench output directory contents (includes the
            flatfile and other supplementary logs).
        """
        for run_vdb in self._run_vdb_cmts:
            run_vdb.store_artifacts(get_bundle=get_bundle, get_vdb_dir=get_vdb_dir)

    def _validate_init(self):
        """Validate the init arguments."""
        instances = {}
        clients = set()
        for run_vdb_cmt in self._run_vdb_cmts:
            if id(run_vdb_cmt) in instances:
                raise ValueError('RunVdb exists more than once: {}'.format(run_vdb_cmt))
            instances[id(run_vdb_cmt)] = run_vdb_cmt

            invalid_clients = clients.intersection([dm.mip for dm in run_vdb_cmt.client_dms])
            if invalid_clients:
                raise ValueError('Clients exist in more than one RunVdb: {}'.format(invalid_clients))
            clients.update([dm.mip for dm in run_vdb_cmt.client_dms])

    def _apply_async(self, timeout=None):
        """Apply the component asynchronously.

        :param int timeout: timeout for apply_async on managed run components
        """
        if self.environments_cmt:
            self.environments_cmt.ensure_applied()
        # Wait for IO to start in parallel
        for cmt in self._run_vdb_cmts:
            cmt._wait_for_io_start = False  # pylint: disable=protected-access
        try:
            for cmt in self._run_vdb_cmts:
                cmt.apply_async(timeout=timeout)
            start_time = time.time()
            for cmt in self._run_vdb_cmts:
                cmt.wait_for_io_running(start_time)
        finally:
            # Reset parameters
            for cmt in self._run_vdb_cmts:
                cmt._wait_for_io_start = True  # pylint: disable=protected-access

    def _apply_is_running(self):
        """Returns True if at least one run_vdb is running.

        This is internally used to check in _apply_finish().

        :rtype: bool
        """
        # pylint: disable=protected-access
        active_vdb_runs = [cmt for cmt in self._run_vdb_cmts if cmt._apply_is_running()]
        if active_vdb_runs:
            _log.debug('Active Vdbench instances: {}'.format(active_vdb_runs))
            return True
        else:
            return False

    def _apply_finish(self, timeout=None, kill=False):
        """Wait for the child RunVdb components to finish applying.

        :param int timeout: Time in seconds to wait for the component to reach the "applied" state.
        :param bool kill: Whether to kill the apply operation in progress or not.
        """
        for run_vdb in self._run_vdb_cmts:
            run_vdb.finish(timeout=timeout, kill=kill)

    def _kill_apply_async(self):
        """Kill the underlying components while applying asynchronously."""
        for run_vdb in self._run_vdb_cmts:
            run_vdb.kill()

    def _reset(self, timeout=None):
        """Ensure components are in the ready state, reset if necessary.

        :param int timeout: The maximum time to reset.
        """
        for run_vdb in self._run_vdb_cmts:
            run_vdb.ensure_ready(force=True)

    def __repr__(self):
        return '{}<{}={}>'.format(
            self.__class__.__name__,
            'master_client_names', [cmt.master_client_dm.client_name for cmt in self._run_vdb_cmts]
        )
