"""Vdbench environment setup components."""
import abc
import os

import six

from valence import components
from valence.vl_components.vdbench import _parm


@six.add_metaclass(abc.ABCMeta)
class VdbEnvironmentsInterface(components.BlockingApply):
    """Interface component for a Vdbench IO environment."""

    @property
    @abc.abstractmethod
    def child_cmts(self):
        """Get all environment-setup components (i.e. no manage components).

        :rtype: list[valence.vl_components.vdbench._setup.SetupVdbEnvironment]
        """

    @property
    @abc.abstractmethod
    def client_dms(self):
        """Clients associated with this environment.

        :rtype: list[valence.models.ResourceClient]
        """

    @property
    @abc.abstractmethod
    def volume_dms(self):
        """Volumes associated with this environment.

        :rtype: list[valence.models.Volume]
        """


class SetupVdbEnvironment(VdbEnvironmentsInterface, components.Reversible):
    """Setup a Vdbench IO environment for an individual master.

    :param valence.vl_components.vdbench.IoSession io_session: IO Session to setup.
    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for the IO session.
    :param valence.models.ResourceClient master_client_dm: Client to use as the vdbench master (default:
        first client listed in the connections). If the specified client is not being used in the
        connections, then it will operate as a dedicated master.
    """
    JOURNAL_DIR = '/tmp/vdbench_journal'

    def __init__(self, io_session, connections_cmt, master_client_dm=None):
        super(SetupVdbEnvironment, self).__init__()
        self.io_session = io_session
        self.connections_cmt = connections_cmt
        self.master_client_dm = master_client_dm or self.connections_cmt.client_dms[0]

        self._general_def = None

        if io_session.standalone and master_client_dm is not None:
            raise ValueError('Master clients cannot be assigned to standalone IO sessions')

    @property
    def child_cmts(self):
        """Get all environment-setup components (i.e. no manage components).

        :rtype: list[valence.vl_components.vdbench._setup.SetupVdbEnvironment]
        """
        return [self]

    @property
    def client_dms(self):
        """Clients associated with this environment.

        :rtype: list[valence.models.ResourceClient]
        """
        return self.connections_cmt.client_dms

    @property
    def volume_dms(self):
        """Volumes associated with this environment.

        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('must be applied to access volume dms.')

        volumes_cmts = []
        for io_period in self.io_session.io_periods:
            for io_workload in io_period.io_workloads:
                if io_workload.volumes_cmt not in volumes_cmts:
                    volumes_cmts.append(io_workload.volumes_cmt)

        volume_dms = []
        for volumes_cmt in volumes_cmts:
            volume_dms.extend(volumes_cmt.volume_dms)

        return volume_dms

    @property
    def defined_intervals(self):
        """The list of intervals for each IO period.

        This is used when parsing the vdbench log to determine runtime based on the interval number.

        :rtype: list[int]
        """
        return [period.config.interval for period in self.io_session.io_periods]

    @property
    def journal_paths(self):
        """Journal paths associated with each storage definition.

        :rtype: list[str]
        """
        journal_paths = []
        if self._general_def:
            journal_paths = [sd.journal_path for sd in self._general_def.storage_defs if sd.journal_path]

        return journal_paths

    def get_vdb_parm(self, port=None, data_errors=None, error_injection_config=None,
                     reporting_config=None, validation_config=None, journaling_config=None):
        """Get the Vdbench parm string for the environment.

        :param int port: Override the default port.
        :param int data_errors: Number of data-validation/IO errors after which to terminate.
        :param valence.vl_components.vdbench.ErrorInjectionConfig error_injection_config: Error injection
            configuration.
        :param valence.vl_components.vdbench.ReportingConfig reporting_config: Reporting configuration.
        :param valence.vl_components.vdbench.ValidationConfig validation_config: Data validation
            configuration.
        :param valence.vl_components.vdbench.JournalingConfig journaling_config: Journaling
            configuration.
        :rtype: str
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('Cannot get parm if not applied')

        if port is not None:
            self._general_def.port = port
        if data_errors is not None:
            self._general_def.data_errors = data_errors
        if error_injection_config is not None:
            self._general_def.error_injection_config = error_injection_config
        if reporting_config is not None:
            self._general_def.reporting_config = reporting_config
        if validation_config is not None:
            self._general_def.validation_config = validation_config
        if journaling_config is not None:
            journal_dir = os.path.join(self.JOURNAL_DIR, self.master_client_dm.client_name)
            self._general_def.configure_journaling(journal_dir, journaling_config)

        return self._general_def.get_parm_str()

    def _apply(self, timeout=None):
        """Ensure connections have been established.

        :param int timeout: the timeout.
        """
        self.connections_cmt.ensure_applied(timeout=timeout)
        self._general_def = _parm.get_general_def(self.io_session, self.connections_cmt)

    def _reset(self, timeout=None):
        """Reset the component.

        :param int timeout: N/A.
        """
        self._general_def = None


class ManageVdbEnvironments(VdbEnvironmentsInterface, components.Reversible):
    """Class for managing multiple SetupVdbEnvironment components.

    :param list[valence.vl_components.vdbench.VdbEnvironmentsInterface] vdb_environment_cmts: Vdbench
        environments to manage.
    :param list[valence.vl_components.ConnectionsInterface] connection_cmts: Connections to manage.
    """
    def __init__(self, vdb_environment_cmts, connection_cmts=None):
        super(ManageVdbEnvironments, self).__init__()
        self.vdb_environment_cmts = vdb_environment_cmts
        self._connection_cmts = connection_cmts or []

    @property
    def child_cmts(self):
        """Get all environment-setup components (i.e. no manage components).

        :rtype: list[valence.vl_components.vdbench._setup.SetupVdbEnvironment]
        """
        child_cmts = []
        for cmt in self.vdb_environment_cmts:
            child_cmts.extend(cmt.child_cmts)
        return child_cmts

    @property
    def client_dms(self):
        """Clients associated with this environment.

        :rtype: list[valence.models.ResourceClient]
        """
        client_dms = []
        for child_cmt in self.child_cmts:
            client_dms.extend(dm for dm in child_cmt.client_dms if dm not in client_dms)
        return client_dms

    @property
    def volume_dms(self):
        """Volumes associated with this environment.

        :rtype: list[valence.models.Volume]
        """
        if self.state is not components.States.applied:
            raise components.ComponentStateError('must be applied to access volume dms.')

        volumes_cmts = []
        for child_cmt in self.child_cmts:
            for io_period in child_cmt.io_session.io_periods:
                for io_workload in io_period.io_workloads:
                    if io_workload.volumes_cmt not in volumes_cmts:
                        volumes_cmts.append(io_workload.volumes_cmt)

        volume_dms = []
        for volumes_cmt in volumes_cmts:
            volume_dms.extend(volumes_cmt.volume_dms)

        return volume_dms

    def _apply(self, timeout=None):
        """Apply the component.

        :param int timeout: maximum time for apply.
        """
        for cmt in self._connection_cmts:
            cmt.ensure_applied()
        for cmt in self.vdb_environment_cmts:
            cmt.ensure_applied()

    def _reset(self, timeout=None):
        """Reset the component.

        :param int timeout: maximum time for reset.
        """
        for cmt in self.vdb_environment_cmts:
            cmt.ensure_ready()
        for cmt in self._connection_cmts:
            cmt.ensure_ready()
