"""Vdbench stream handlers."""
import abc

import autologger
import six

from valence.vl_components.vdbench import (
    _errors,
    _log_parser,
)

_log = autologger.get_sflogging(__name__)


@six.add_metaclass(abc.ABCMeta)
class BaseVdbStreamHandler(object):
    """Base object for handling a Vdbench Stream of data."""
    # This is used for StreamHandlers that manage sub-handlers. If False, it is an error to add this as
    # a stream handler directly.
    CAN_REGISTER = True

    @abc.abstractmethod
    def handle_items(self, items):
        """Handle a list of items passed to you.

        :param list[valence.vl_components.vdbench._log_parser.VdbLogLine] items: A list of
            items from the vdbench log.
        """

    @abc.abstractmethod
    def clear(self):
        """This is called whenever the run_vdb component is reset.

        The handler should always clear all of its internal state.
        """

    @abc.abstractmethod
    def destroy(self):
        """Destroy the handler by detaching the handler from its initialized stream.

        It will be an error to use it after this has been called.
        """


class ErrorStreamHandler(BaseVdbStreamHandler):
    """Stream handler from vdbench for handling errors.

    This is automatically created and added to every vdbench instance. This
    stores state that is raised by RunVdb.assert_valid().

    :param valence.vl_components.vdbench.RunVdb run_vdb_cmt: The run_vdb_cmt that has an update_stream
        method attached to it.
    """

    def __init__(self, run_vdb_cmt):
        self.run_vdb_cmt = run_vdb_cmt
        self.eio_count = 0
        self.critical_errors = []
        self.no_read_validations = set()
        run_vdb_cmt._add_stream_handler(self)  # pylint: disable=protected-access

    def destroy(self):
        """Detach the handler from its initialized stream.

        It will be an error to use it after this has been called.
        """
        self.run_vdb_cmt._remove_stream_handler(self)  # pylint: disable=protected-access
        # noinspection PyTypeChecker
        self.run_vdb_cmt = None

    def clear(self):
        """Clear the internal state."""
        self.eio_count = 0
        self.critical_errors = []
        self.no_read_validations = set()

    def handle_items(self, items):  # pylint: disable=too-complex
        """Keep track of items with errors.

        This is called by RunVdb.update_stream() every time there is an update.

        :param list[valence.vl_components.vdbench._log_parser.VdbLogLine] items: A list of items from the
            vdbench log.
        """
        critical_errors = []
        eios = 0
        for item in items:
            if item.type == _log_parser.VdbLogLineType.no_read_validations:
                self.no_read_validations.add(
                    _log_parser.NO_READ_VALIDATIONS_LINE.match(item.raw).group(1)
                )
                continue
            elif item.type not in _log_parser.VdbLogLine.ERROR_TYPES:
                continue
            elif item.type is _log_parser.VdbLogLineType.io_error:
                eios += 1
            elif item.type is _log_parser.VdbLogLineType.premature_termination:
                if _log_parser.PREMATURE_TERMINATION_LINE.match(item.raw).group(1) \
                        in self.no_read_validations:
                    # Ignore it, it was just prematurely terminated because validation wasn't done
                    continue
                else:
                    critical_errors.append(item)
            else:
                critical_errors.append(item)

        if critical_errors:
            _log.debug(
                'Critical errors found in Vdbench run: {}\n{}'
                .format(self.run_vdb_cmt, self.get_errors_string(critical_errors))
            )
        if eios:
            _log.debug('{} IO Errors found in Vdbench run: {}'.format(eios, self.run_vdb_cmt))

        self.eio_count += eios
        self.critical_errors.extend(critical_errors)

    def assert_valid(self, max_eios):
        """Convenience method used by RunVdb.

        :param int max_eios: The maximum EIOs.
        """
        if not self.run_vdb_cmt:
            raise _errors.StreamHandlerDestroyed()
        self.run_vdb_cmt.update_stream()
        if self.critical_errors:
            raise _errors.VdbenchRuntimeError(
                'Critical errors found in Vdbench run: {}\n{}'
                .format(self.run_vdb_cmt, self.get_errors_string())
            )
        if self.eio_count > max_eios:
            raise _errors.VdbenchRuntimeError(
                '{} IO Errors found in Vdbench run (max={}): {}'
                .format(self.eio_count, max_eios, self.run_vdb_cmt)
            )

    def get_errors_string(self, errors=None):
        """A string representing all of the critical errors detected.

        :param list[valence.vl_components.vdbench._log_parser.VdbLogLine] errors: Errors to include.
        :rtype: str
        """
        errors = errors or self.critical_errors
        return '** CRITICAL ERRORS **\n    {}'.format('\n    {}'.join([str(e) for e in errors]))


class RunInfoHandler(BaseVdbStreamHandler):
    """Used to track RunInfo on the vdbench run.

    Use self.current_io_line_item to get the current VdbLogLine object which always has
    line_type == VdbLogLineType.io (or is None).

    :param valence.vl_components.vdbench.RunVdb run_vdb_cmt: a **single** master run vdb component.
    """
    def __init__(self, run_vdb_cmt):
        self.run_vdb_cmt = run_vdb_cmt
        self.current_io_line_item = None
        run_vdb_cmt._add_stream_handler(self)  # pylint: disable=protected-access

    def destroy(self):
        """Detach the handler from its initialized stream.

        It will be an error to use it after this has been called.
        """
        self.run_vdb_cmt._remove_stream_handler(self)  # pylint: disable=protected-access
        # noinspection PyTypeChecker
        self.run_vdb_cmt = None

    def clear(self):
        """Clear the internal state."""
        self.current_io_line_item = None

    def handle_items(self, items):
        """Keep track of items with errors.

        This is called by RunVdb.update_stream() every time there is an update.

        :param list[valence.vl_components.vdbench._log_parser.VdbLogLine] items: a list of items from the
            Vdbench log.
        """
        for item in items:
            if item.type == _log_parser.VdbLogLineType.io:
                self.current_io_line_item = item
