"""Base methods and classes for Vdbench profiles."""
import abc

import six

from valence.vl_components.vdbench import (
    _io_config,
    _setup,
)


@six.add_metaclass(abc.ABCMeta)
class BaseVdbProfile(object):
    """Vdbench profile.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for the IO session.
    :param valence.vl_components.vdbench.IoSessionConfig session_config: IO session configuration.
    :param bool manage_connections: Whether environments generated by this profile should manage
        the provided connections for you (i.e. connections will be reset when the environment is).
    """

    def __init__(self, connections_cmt, session_config, manage_connections):
        self.connections_cmt = connections_cmt
        self.session_config = session_config
        self.manage_connections = manage_connections

    @abc.abstractmethod
    def get_workloads(self):
        """Get a set of workloads for the profile.

        :rtype: list[valence.vl_components.vdbench.IoWorkload]
        """

    @abc.abstractmethod
    def get_periods(self):
        """Get a set of IO periods for the profile.

        :rtype: list[valence.vl_components.vdbench.IoPeriod]
        """

    def get_session(self, standalone=False):
        """Get an IO session for the profile.

        :param bool standalone: Whether to run Vdbench in standalone mode (default: False). By default
            Vdbench runs a slave alongside the master instance. In standalone mode the master performs
            the duties of the slave in addition to its own (only works for single-client instances)
            (reduces memory overhead).
        :rtype: valence.vl_components.vdbench.IoSession
        """
        periods = self.get_periods()
        return _io_config.IoSession(self.session_config, periods, standalone=standalone)

    def get_environment(self, master_client_dm=None):
        """Get a Vdbench environment for the profile.

        :param valence.models.ResourceClient master_client_dm: Client to use as the vdbench master
            (default: first client listed in the connections). If the specified client is not being used
            in the connections, then it will operate as a dedicated master.
        :rtype: valence.vl_components.vdbench.SetupVdbEnvironment
        """
        vdb_env_cmt = _setup.SetupVdbEnvironment(
            io_session=self.get_session(),
            connections_cmt=self.connections_cmt,
            master_client_dm=master_client_dm
        )
        if not self.manage_connections:
            return vdb_env_cmt
        else:
            return _setup.ManageVdbEnvironments(
                vdb_environment_cmts=[vdb_env_cmt],
                connection_cmts=[self.connections_cmt]
            )


@six.add_metaclass(abc.ABCMeta)
class BaseVdbExtendedProfile(object):
    """Vdbench profile.

    :param list[valence.vl_components.vdbench.profiles._base.BaseVdbProfile] profiles: Individual
        profiles for each Vdbench master.
    :param bool standalone: Whether to run Vdbench in standalone mode. By default Vdbench runs a slave
        alongside the master instance. In standalone mode the master performs the duties of the slave in
        addition to its own (only works for single-client instances) (reduces memory overhead).
    """
    # Represent the base profile being extended
    BASE_PROFILE = None

    # Recommended maximum number of slaves per Vdbench master
    # Vdbench tends to become unstable beyond this number (may be affected by master client type)
    # TODO: Dynamically determine this based on client types (VL-2099)
    MAX_SLAVE_COUNT = 11

    def __init__(self, profiles, standalone):
        self._profiles = profiles
        self._standalone = standalone
        assert self.BASE_PROFILE is not None

    def get_sessions(self):
        """Get a set of IO sessions for the profile.

        :rtype: list[valence.vl_components.vdbench.IoSession]
        """
        return [profile.get_session(self._standalone) for profile in self._profiles]

    def get_environment(self):
        """Get a Vdbench environment for the profile.

        :rtype: valence.vl_components.vdbench.ManageVdbEnvironments
        """
        environment_cmts = [profile.get_environment() for profile in self._profiles]
        return _setup.ManageVdbEnvironments(environment_cmts)
