"""Vdbench CSV-based profile object.

Example Usage:
    Setup::

        vdb_profile = CsvVdbProfile(
            tag='example',
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            csv_file_path='/path/to/example.csv'
        )
        setup_vdb_cmt = vdb_profile.get_environment()
        run_vdb_cmt = RunVdb(setup_vdb_cmt)
        setup_vdb_cmt.apply()

    Run::

        run_vdb_cmt.apply_async()
        ...
        run_vdb_cmt.finish(kill=True)

    Teardown::

        run_vdb_cmt.ensure_ready()
        setup_vdb_cmt.ensure_ready()


Example Usage:
    Setup::

        vdb_profile = CsvVdbExtendedProfile(
            tag='example',
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            csv_folder_path='/path/to/csv_folder'
        )
        setup_vdb_cmt = vdb_profile.get_environment()
        run_vdb_cmt = ManageRunVdbs.from_environment_cmt(
            environments_cmt=setup_vdb_cmt,
            journaling_config=JournalingConfig.full
        )

    Run::

        setup_vdb_cmt.apply()
        run_vdb_cmt.apply_async()
        ...
        run_vdb_cmt.finish(kill=True)

    Teardown::

        run_vdb_cmt.ensure_ready()
        setup_vdb_cmt.ensure_ready()
"""
import copy
import os
from collections import defaultdict

import six

from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _io_config,
    _parm,
)
from valence.vl_components.vdbench.profiles import (
    _base,
    csv_profiles,
)


class CsvVdbProfile(_base.BaseVdbProfile):
    """CSV-based Vdbench profile.

    :param str tag: Unique tag for the accounts/volumes to be created.
    :param valence.models.ResourceCluster cluster_dm: The cluster on which to create volumes.
    :param list[valence.models.ResourceClient] client_dms: The clients use.
    :param str csv_file_path: The path to the CSV file from which to generate the profile.
    """
    # Standard configuration of IO rate
    DEFAULT_PERIOD_CONFIGS = [
        _io_config.IoPeriodConfig(
            io_rate_config=_parm.IoRateMax(),
            interval=30,
            duration=24 * 60 * 60
        )
    ]

    # Standard configurations of compression/dedup ratios
    DEDUP_2X = _parm.DedupConfig(dedup_ratio=2, dedup_unit=4096)
    DEFAULT_SESSION_CONFIG = _io_config.IoSessionConfig(compression_ratio=3)

    # Standard CSV file names
    ALIGNED_CSV = 'aligned_io_profile.csv'
    UNALIGNED_CSV = 'unaligned_io_profile.csv'

    def __init__(self, tag, cluster_dm, client_dms, csv_file_path):
        conn_defs = csv_profiles.get_connection_definitions(csv_file_path)
        volume_client_map = self._get_volume_client_map(client_dms, conn_defs)
        self.accounts_cmt = _account.CreateAccounts(tag, cluster_dm)
        self.volumes_cmt_map = self._create_volumes(tag, cluster_dm, conn_defs)
        super(CsvVdbProfile, self).__init__(
            connections_cmt=self._create_connections(cluster_dm, client_dms, volume_client_map),
            session_config=copy.deepcopy(self.DEFAULT_SESSION_CONFIG),
            manage_connections=True
        )
        self.workload_configs = self._get_workload_configs(client_dms, conn_defs)
        self.period_configs = copy.deepcopy(self.DEFAULT_PERIOD_CONFIGS)

        # Only set dedup if all workloads are aligned
        # NOTE: Unaligned workloads require dedup_unit=512, but SolidFire only supports dedup_unit=4k
        if all(config[0].aligned for config in self.workload_configs):
            self.session_config.dedup_config = self.DEDUP_2X

    @staticmethod
    def _get_volume_client_map(client_dms, conn_defs):
        """Get a map of clients to volumes representing the specified connections.

        :param list[valence.models.ResourceClient] client_dms: The clients use.
        :param list[dict] conn_defs: The connection definitions to represent.
        :raises ValueError: If an unexpected number of clients is provided.
        :returns: Map of client MIPs to volume tags.
        :rtype: dict[str,str]
        """
        max_client_index = 0
        volume_client_map = defaultdict(list)
        for conn_def in conn_defs:
            client_index = int(conn_def[csv_profiles.CLIENT_INDEX] or 0)
            max_client_index = max(max_client_index, client_index)
            if client_index < len(client_dms):
                client_dm = client_dms[client_index]
                volume_client_map[client_dm.mip].append(conn_def[csv_profiles.VOLUME_NAME])

        if max_client_index != len(client_dms) - 1:
            raise ValueError(
                'Unexpected number of clients for specified connection definitions: {}/{}'
                .format(len(client_dms), max_client_index)
            )
        else:
            return volume_client_map

    @staticmethod
    def _get_workload_configs(client_dms, conn_defs):
        """Get a set of workload configurations and their associated clients and volumes.

        :param list[valence.models.ResourceClient] client_dms: The clients use.
        :param list[dict] conn_defs: The connection definitions to represent.
        :returns: Workload configurations in the form: (workload_config, client_dm, volume_tag).
        :rtype: tuple
        """
        workload_configs = []
        for conn_def in conn_defs:
            client_dm = client_dms[int(conn_def[csv_profiles.CLIENT_INDEX] or 0)]
            volume_tag = conn_def[csv_profiles.VOLUME_NAME]
            workload_config = csv_profiles.get_workload_config(conn_def)
            workload_configs.append((workload_config, client_dm, volume_tag))

        return workload_configs

    def _create_volumes(self, tag, cluster_dm, conn_defs):
        """Create volumes components for the specified connections.

        :param str tag: Unique tag for the volumes to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which to create volumes.
        :param list[dict] conn_defs: The connection definitions to represent.
        :returns: Map of volumes components to their respective volume tag.
        :rtype: dict[str,valence.vl_components.CreateVolumes]
        """
        volume_dm_map = {}
        for conn_def in conn_defs:
            volume_tag = conn_def[csv_profiles.VOLUME_NAME]
            if volume_tag not in volume_dm_map:
                volume_dm_map[volume_tag] = csv_profiles.generate_volume_dm(conn_def)

        volumes_cmt_map = {}
        for volume_tag, volume_dm in six.iteritems(volume_dm_map):
            volumes_cmt_map[volume_tag] = _volume.CreateVolumes(
                tag='{}-{}'.format(tag, volume_tag),
                cluster_dm=cluster_dm,
                partial_volume_dms=[volume_dm],
                accounts_cmt=self.accounts_cmt
            )

        return volumes_cmt_map

    def _create_connections(self, cluster_dm, client_dms, volume_client_map):
        """Create connections components for the profile.

        :param valence.models.ResourceCluster cluster_dm: The cluster on which to create volumes.
        :param list[valence.models.ResourceClient] client_dms: The clients use.
        :param dict[str,str] volume_client_map: Map of client MIPs to volume tags.
        :rtype: valence.vl_components.ManageConnections
        """
        connection_cmts = []
        for client_dm in client_dms:
            connection_cmts.append(_connection.CreateIscsiAccountConnection(
                cluster_dm=cluster_dm,
                client_dm=client_dm,
                accounts_cmt=self.accounts_cmt,
                volumes_cmts=[self.volumes_cmt_map[tag] for tag in volume_client_map[client_dm.mip]]
            ))
        connections_manager = _connection.ManageConnections(connection_cmts)
        connections_manager.accounts_cmts = [self.accounts_cmt]
        connections_manager.volumes_cmts = list(self.volumes_cmt_map.values())
        return connections_manager

    @classmethod
    def from_csv_folder(cls, tag, cluster_dm, client_dms, csv_folder_path, aligned=True):
        """Construct a csv-based profile from a csv-folder.

        :param str tag: Unique tag for the accounts/volumes to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which to create volumes.
        :param list[valence.models.ResourceClient] client_dms: The clients use.
        :param str csv_folder_path: The path to the CSV folder from which to generate the profile.
        :param bool aligned: Whether to run aligned/unaligned IO (default: aligned).
        :rtype: valence.vl_components.vdbench.CsvVdbProfile
        """
        csv_client_dirs = csv_profiles.get_client_dir_map(csv_folder_path)
        if len(client_dms) in csv_client_dirs:
            file_path = os.path.join(csv_folder_path, csv_client_dirs[len(client_dms)])
        else:
            raise ValueError(
                'Specified CSV folder[{}] does not support the provided number of clients: {}'
                .format(csv_folder_path, len(client_dms))
            )

        if aligned:
            file_path = os.path.join(file_path, cls.ALIGNED_CSV)
            assert os.path.exists(file_path)
        else:
            file_path = os.path.join(file_path, cls.UNALIGNED_CSV)
            assert os.path.exists(file_path)

        return cls(tag, cluster_dm, client_dms, file_path)

    def get_workloads(self):
        """Get a set of csv-based workloads.

        :rtype: list[valence.vl_components.vdbench.IoWorkload]
        """
        workloads = []
        for workload_config, client_dm, volume_tag in self.workload_configs:
            workloads.append(_io_config.IoWorkload(
                io_workload_config=workload_config,
                client_dm=client_dm,
                volumes_cmt=self.volumes_cmt_map[volume_tag],
                volume_index=0
            ))
        return workloads

    def get_periods(self):
        """Get a set of csv-based IO periods.

        :rtype: list[valence.vl_components.vdbench.IoPeriod]
        """
        workloads = self.get_workloads()
        periods = [_io_config.IoPeriod(config, workloads) for config in self.period_configs]
        return periods


class CsvVdbExtendedProfile(_base.BaseVdbExtendedProfile):
    """CSV-based Vdbench extended profile (supports multiple master instances).

    :param str tag: Unique tag for the accounts/volumes to be created.
    :param valence.models.ResourceCluster cluster_dm: The cluster on which to create volumes.
    :param list[valence.models.ResourceClient] client_dms: The clients use.
    :param str csv_folder_path: The path to the CSV folder from which to generate the profile.
    """
    BASE_PROFILE = CsvVdbProfile

    def __init__(self, tag, cluster_dm, client_dms, csv_folder_path):
        # Determine the maximum number of supported clients per CSV-based profile
        csv_client_dirs = csv_profiles.get_client_dir_map(csv_folder_path)
        max_slave_count = min(max(csv_client_dirs.keys()), self.MAX_SLAVE_COUNT)

        client_index = 0
        profiles = []
        # If there is more than one client, setup an unaligned IO profile
        if len(client_dms) > 1:
            client_index += 1
            profiles.append(self.BASE_PROFILE.from_csv_folder(
                tag='{}-{}'.format(tag, len(profiles)),
                cluster_dm=cluster_dm,
                client_dms=client_dms[0:1],
                csv_folder_path=csv_folder_path,
                aligned=False
            ))
        # Use all remaining clients for aligned IO profiles
        # NOTE: Clients will be split up if the max-slave-per-master limit is exceeded
        while client_index < len(client_dms):
            next_index = min(client_index + max_slave_count, len(client_dms))
            profiles.append(self.BASE_PROFILE.from_csv_folder(
                tag='{}-{}'.format(tag, len(profiles)),
                cluster_dm=cluster_dm,
                client_dms=client_dms[client_index:next_index],
                csv_folder_path=csv_folder_path
            ))
            client_index = next_index

        super(CsvVdbExtendedProfile, self).__init__(profiles, False)
