"""Vdbench profile helper methods."""
from __future__ import division

import copy
import itertools

import six

from valence.vl_components.vdbench import (
    _io_config,
    _parm,
)


def get_varied_workload_configs(connections_cmt, base_workload_configs, transfer_size_configs):
    """Get a set of workload configurations with varied transfer sizes.

    Generates multiple workload configurations based on each of the base configurations, but with
    varying transfer size configurations.

    NOTE: The number of workload configurations generated is limited by the maximum number of workloads
    per IO period.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO period.
    :param list[valence.vl_components.vdbench.IoWorkloadConfig] base_workload_configs: Base workload
        configurations from which to generate the varied configurations.
    :param list[valence.vl_components.vdbench.TransferSizeConfig] transfer_size_configs: Transfer size
        configurations from which to generate the varied configurations.
    :rtype: list[valence.vl_components.vdbench.IoWorkloadConfig]
    """
    connection_count = 0
    for conn_cmt in connections_cmt.child_cmts:
        for volumes_cmt in conn_cmt.volumes_cmts:
            connection_count += len(volumes_cmt.partial_volume_dms)
    config_count = min(connection_count, _io_config.IoPeriod.MAX_WORKLOADS)

    workload_configs = []
    base_config_index = 0
    for config_index in range(config_count):
        # Rotate through the base configs
        if config_index * len(base_workload_configs) >= (1 + base_config_index) * config_count:
            base_config_index += 1

        # Set a different transfer size for each configuration
        config = copy.deepcopy(base_workload_configs[base_config_index])
        xfersize_index = len(workload_configs) % len(transfer_size_configs)
        xfersize_config = transfer_size_configs[xfersize_index]

        # Limit xfersize based on volume size limitations
        largest_xfersize = max(xfersize_config.xfersizes)
        max_xfersize = get_max_transfer_size(connections_cmt, config.thread_count)
        if largest_xfersize > max_xfersize:
            xfersize_config = _parm.TransferSizeFixed(4096 * max(1, max_xfersize // 4096))

        config.transfer_size_config = xfersize_config
        workload_configs.append(config)

    return workload_configs


def get_max_transfer_size(connections_cmt, thread_count=None, io_rate_config=None):
    """Get the maximum transfer size for the specified volumes.

    max_xfersize = volume_size / (vdb_queue_depth + user_queue_depth + 256)

    NOTE:
    * vdb_queue_depth = 4000 (2000 if io_rate=max)
    * user_queue_depth = thread_count

    See SET-633 for details.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO period.
    :param int thread_count: Number of concurrent IO streams.
    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration.
    :rtype: int
    """
    thread_count = thread_count or 8
    smallest_vol_size = 0
    for conn_cmt in connections_cmt.child_cmts:
        for vol_cmt in conn_cmt.volumes_cmts:
            for vol_dm in vol_cmt.partial_volume_dms:
                if not smallest_vol_size:
                    smallest_vol_size = vol_dm.total_size.bytes
                elif vol_dm.total_size.bytes < smallest_vol_size:
                    smallest_vol_size = vol_dm.total_size.bytes

    vdb_queue_depth = 4000
    if io_rate_config and isinstance(io_rate_config, _parm.IoRateMax):
        vdb_queue_depth = 2000

    return smallest_vol_size // (vdb_queue_depth + thread_count + 256)


def get_workloads(connections_cmt, workload_configs):
    """Get a set of IO workloads for the provided connections.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO period.
    :param list[valence.vl_components.vdbench.IoWorkloadConfig] workload_configs: Workload configurations
        for a given Vdbench profile.
    :rtype: list[valence.vl_components.vdbench.IoWorkload]
    :raises ValueError: If no aligned workload configurations are provided for connections with
        non-512e-enabled volumes.
    """
    workloads = []
    config_iterator = itertools.cycle(workload_configs)
    for conn_cmt in connections_cmt.child_cmts:
        for volumes_cmt in conn_cmt.volumes_cmts:
            for child_volumes_cmt in volumes_cmt.child_cmts:
                for volume_index in range(len(child_volumes_cmt.partial_volume_dms)):
                    workloads.append(_io_config.IoWorkload(
                        io_workload_config=next(config_iterator),
                        client_dm=conn_cmt.client_dm,
                        volumes_cmt=child_volumes_cmt,
                        volume_index=volume_index
                    ))

    return workloads


def isolate_access_ranges(io_workloads):
    """Isolate the access ranges of multiple workloads attached to the same volume.

    :param list[valence.vl_components.vdbench.IoWorkload] io_workloads: IO Workloads.
    """
    volume_workload_map = {}
    for workload in io_workloads:
        volumes_key = (id(workload.volumes_cmt), workload.volume_index)
        if volumes_key not in volume_workload_map:
            volume_workload_map[volumes_key] = {}
        volume_workload_map[volumes_key][id(workload)] = workload

    for workloads in six.itervalues(volume_workload_map):
        range_size = 100 // len(workloads)
        for i, workload in enumerate(six.itervalues(workloads)):
            workload.config = copy.copy(workload.config)
            workload.config.access_range_config = _parm.AccessRangePercentages(
                percent_start=range_size * i,
                percent_end=range_size * (i + 1)
            )
