"""Vdbench Profile object for QoS Histogram testing."""

from valence.vl_components.vdbench import (
    _enums,
    _io_config,
    _parm,
)
from valence.vl_components.vdbench.profiles import (
    _base,
    _helpers,
    _simple,
)


class QosHistIopsProfile(_simple.SimpleVdbProfile):
    """This profile will setup the fields for IOps testing.

    IoRateCurve should properly hit all the IOps buckets from the Qos Histogram.
    It defines one workload config with rdpct=50a, xfersize=4k. The interval is half of elapsed.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO
        session.
    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration to use
        (default IoRateCurve((1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 22, 31, 40, 49, 58, 67, 76, 85, 94),
        _enums.IoDistribution.uniform)).
    :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration to use.
    :param int elapsed: Equal IO period per curve in seconds (i.e. total io duration is the
        sum of all elapsed periods plus 1).
    """
    #  The test curve to use in vdbench. There are two 1% at the beginning to allow for settling
    CURVE = (1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 22, 31, 40, 49, 58, 67, 76, 85, 94)
    TRANSFER_SIZE_4K = 4096

    def __init__(self, connections_cmt, io_rate_config=None, seek_config=None, elapsed=60):
        io_rate_config = io_rate_config or _parm.IoRateCurve(
            self.CURVE, _enums.IoDistribution.uniform)
        duration = elapsed * (len(io_rate_config.curve) + 1)
        super(QosHistIopsProfile, self).__init__(
            connections_cmt,
            read_perc=50,
            io_rate_config=io_rate_config,
            transfer_size_config=_parm.TransferSizeFixed(self.TRANSFER_SIZE_4K),
            seek_config=seek_config,
            duration=duration
        )
        self.period_config.interval = elapsed // 2


class QosHistStandardBlockSizeProfile(_base.BaseVdbProfile):
    """This profile will make workloads with 0% and 100% reads across standard block sizes.

    Standard block sizes = 2**11 (w/ 512e), 2**12, 2**13, 2**14, 2**15, 2**16, 2**17
    volume component should enable 512e on every 8th volume to support 2k transfers

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO session.
    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration to use.
    :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration to use.
    :param int duration: duration to pass to IoPeriodConfig or None.
    """
    def __init__(self, connections_cmt, io_rate_config=None, seek_config=None, duration=None):
        super(QosHistStandardBlockSizeProfile, self).__init__(
            connections_cmt=connections_cmt,
            session_config=_io_config.IoSessionConfig(),
            manage_connections=False
        )
        self.io_rate_config = io_rate_config or _parm.IoRateMax()
        self.seek_config = seek_config

        self.workload_configs = []
        xfer_sizes = [2**11, 2**12, 2**13, 2**14, 2**15, 2**16, 2**17]
        read_percs = [0, 100]
        for read_perc in read_percs:
            for xfer_size in xfer_sizes:
                self.workload_configs.append(_io_config.IoWorkloadConfig(
                    read_perc=read_perc,
                    transfer_size_config=_parm.TransferSizeFixed(xfer_size),
                    seek_config=seek_config
                ))

        self.period_config = _io_config.IoPeriodConfig(self.io_rate_config, duration=duration)

    def get_workloads(self):
        """Get a set of simple workloads.

        :rtype: list[valence.vl_components.vdbench.IoWorkload]
        """
        workloads = _helpers.get_workloads(
            connections_cmt=self.connections_cmt,
            workload_configs=self.workload_configs
        )
        _helpers.isolate_access_ranges(workloads)
        return workloads

    def get_periods(self):
        """Get a set of simple IO periods.

        :rtype: list[valence.vl_components.vdbench.IoPeriod]
        """
        workloads = self.get_workloads()
        io_period = _io_config.IoPeriod(
            io_period_config=self.period_config,
            io_workloads=workloads
        )
        return [io_period]


class QosHistThrottlingProfile(_simple.SimpleVdbProfile):
    """This profile will produce entries in each of the throttling buckets.

    This profile defines one workload config with rdpct=20 and xfersize=4k.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an
        IO session.
    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration
        to use (default IoRateMax).
    :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration to use.
    :param int thread_count: The thread count to use in IoWorkloadConfig.
    :param int duration: Duration for this IO period in seconds.
    """
    TRANSFER_SIZE_4K = 4096

    def __init__(self, connections_cmt, io_rate_config=None, seek_config=None,
                 thread_count=None, duration=None):
        super(QosHistThrottlingProfile, self).__init__(
            connections_cmt,
            read_perc=20,
            io_rate_config=io_rate_config,
            transfer_size_config=_parm.TransferSizeFixed(self.TRANSFER_SIZE_4K),
            seek_config=seek_config,
            duration=duration
        )
        self.workload_config.thread_count = thread_count
        self.period_config.interval = 1
