"""Simple Vdbench profile object.

Example Usage:
    ::Setup::
    vdb_profile = SimpleVdbProfile(connections_cmt)
    setup_vdb_cmt = vdb_profile.get_environment()
    run_vdb_cmt = RunVdb(setup_vdb_cmt)
    setup_vdb_cmt.apply()

    ::Run::
    run_vdb_cmt.apply_async()
    ...
    run_vdb_cmt.finish(kill=True)

    ::Teardown::
    run_vdb_cmt.ensure_ready()
    setup_vdb_cmt.ensure_ready()
"""
from valence.vl_components.vdbench import (
    _io_config,
    _parm,
)
from valence.vl_components.vdbench.profiles import (
    _base,
    _helpers,
)


class SimpleVdbProfile(_base.BaseVdbProfile):
    """Simple Vdbench profile.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO session.
    :param int read_perc: Read percentage to use.
    :param valence.vl_components.vdbench.IoRateConfig io_rate_config: IO rate configuration to use.
    :param valence.vl_components.vdbench.TransferSizeConfig transfer_size_config: Transfer size
        configuration to use.
    :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration to use.
    :param int duration: duration to pass to IoPeriodConfig or None.
    """

    def __init__(self, connections_cmt, read_perc=None, io_rate_config=None, transfer_size_config=None,
                 seek_config=None, duration=None):
        super(SimpleVdbProfile, self).__init__(
            connections_cmt=connections_cmt,
            session_config=_io_config.IoSessionConfig(),
            manage_connections=False
        )
        self.read_perc = read_perc
        self.io_rate_config = io_rate_config or _parm.IoRateMax()
        self.transfer_size_config = transfer_size_config
        self.seek_config = seek_config

        self.workload_config = _io_config.IoWorkloadConfig(
            read_perc=self.read_perc,
            transfer_size_config=self.transfer_size_config,
            seek_config=self.seek_config
        )
        self.period_config = _io_config.IoPeriodConfig(self.io_rate_config, duration=duration)

    def get_workloads(self):
        """Get a set of simple workloads.

        :rtype: list[valence.vl_components.vdbench.IoWorkload]
        """
        workloads = _helpers.get_workloads(
            connections_cmt=self.connections_cmt,
            workload_configs=[self.workload_config]
        )
        _helpers.isolate_access_ranges(workloads)
        return workloads

    def get_periods(self):
        """Get a set of simple IO periods.

        :rtype: list[valence.vl_components.vdbench.IoPeriod]
        """
        workloads = self.get_workloads()
        io_period = _io_config.IoPeriod(
            io_period_config=self.period_config,
            io_workloads=workloads
        )
        return [io_period]
