"""Standard Vdbench profile object.

Example Usage:
    Setup::

        vdb_profile = StandardVdbProfile(connections_cmt)
        setup_vdb_cmt = vdb_profile.get_environment()
        run_vdb_cmt = RunVdb(setup_vdb_cmt)
        setup_vdb_cmt.apply()

    Run::

        run_vdb_cmt.apply_async()
        ...
        run_vdb_cmt.finish(kill=True)

    Teardown::

        run_vdb_cmt.ensure_ready()
        setup_vdb_cmt.ensure_ready()


Example Usage:
    Setup::

        vdb_profile = StandardVdbExtendedProfile(
            tag='example',
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            connections_per_volume=3
        )
        vdb_profile.override_configs(read_perc=100)
        setup_vdb_cmt = vdb_profile.get_environment()
        run_vdb_cmt = ManageRunVdbs.from_environment_cmt(
            environments_cmt=setup_vdb_cmt,
            journaling_config=JournalingConfig.full
        )

    Run::

        setup_vdb_cmt.apply()
        run_vdb_cmt.apply_async()
        ...
        run_vdb_cmt.finish(kill=True)

    Teardown::

        run_vdb_cmt.ensure_ready()
        setup_vdb_cmt.ensure_ready()
"""
import copy

import autologger

from valence import (
    enums,
    models,
)
from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _io_config,
    _parm,
)
from valence.vl_components.vdbench.profiles import (
    _base,
    _helpers,
)

_log = autologger.get_sflogging()


class StandardVdbProfile(_base.BaseVdbProfile):
    """Standard Vdbench profile.

    :param valence.vl_components.ConnectionsInterface connections_cmt: Connections for an IO session.
    :param bool manage_connections: Whether environments generated by this profile should manage
        the provided connections for you (i.e. connections will be reset when the environment is).
    :param bool aligned: Whether to run aligned/unaligned IO (default: aligned).
    """
    # Standard QOS configurations
    REDUCED_QOS = models.Qos.create(max_iops=1000, burst_iops=1500)

    # Transfer sizes have been ordered to provide maximum benefit for any number of workloads
    # Aligned transfer sizes must be multiples of 1024 bytes
    DEFAULT_XFER_SIZE_CONFIGS_ALIGNED = [
        _parm.TransferSizeFixed(4 * 1024),
        _parm.TransferSizeFixed(16 * 1024),
        _parm.TransferSizeFixed(128 * 1024),
        _parm.TransferSizeFixed(1024 * 1024),
        _parm.TransferSizeFixed(64 * 1024),
        _parm.TransferSizeFixed(512 * 1024),
        _parm.TransferSizeFixed(2048 * 1024),
        _parm.TransferSizeVariable([
            (4 * 1024, 10),
            (32 * 1024, 10),
            (128 * 1024, 20),
            (256 * 1024, 30),
            (512 * 1024, 30)
        ]),
        _parm.TransferSizeFixed(8 * 1024),
        _parm.TransferSizeFixed(32 * 1024),
        _parm.TransferSizeFixed(256 * 1024),
        _parm.TransferSizeVariable([
            (4 * 1024, 30),
            (32 * 1024, 20),
            (128 * 1024, 20),
            (256 * 1024, 20),
            (512 * 1024, 10)
        ])
    ]

    # Unaligned transfer sizes should be odd-multiples of 512 bytes
    DEFAULT_XFER_SIZE_CONFIGS_UNALIGNED = [
        _parm.TransferSizeFixed(1 * 512),
        _parm.TransferSizeFixed(31 * 512),
        _parm.TransferSizeFixed(257 * 512),
        _parm.TransferSizeFixed(2049 * 512),
        _parm.TransferSizeFixed(131 * 512),
        _parm.TransferSizeFixed(511 * 512),
        _parm.TransferSizeVariable([
            (1 * 512, 10),
            (2 * 512, 10),
            (3 * 512, 10),
            (4 * 1024, 10),
            (32 * 1024, 10),
            (128 * 1024, 20),
            (256 * 1024, 20),
            (512 * 1024, 10)
        ])
    ]

    # Standard configurations of seek/read percentages
    # Ordered to provide maximum benefit for any number of workloads
    SEEK_SEQUENTIAL_IO = _parm.SeekConfig(0)
    SEEK_RANDOM_IO = _parm.SeekConfig(100)
    DEFAULT_WORKLOAD_CONFIGS = [
        _io_config.IoWorkloadConfig(seek_config=SEEK_RANDOM_IO, read_perc=40),
        _io_config.IoWorkloadConfig(seek_config=SEEK_SEQUENTIAL_IO, read_perc=5),
        _io_config.IoWorkloadConfig(seek_config=SEEK_SEQUENTIAL_IO, read_perc=95),
        _io_config.IoWorkloadConfig(seek_config=SEEK_RANDOM_IO, read_perc=15),
        _io_config.IoWorkloadConfig(seek_config=SEEK_RANDOM_IO, read_perc=85),
    ]

    # Standard configuration of IO rate
    DEFAULT_PERIOD_CONFIGS = [
        _io_config.IoPeriodConfig(
            io_rate_config=_parm.IoRateMax(),
            interval=30,
            duration=24*60*60
        )
    ]

    # Standard configurations of compression/dedup ratios
    DEDUP_2X = _parm.DedupConfig(dedup_ratio=2, dedup_unit=4096)
    DEFAULT_SESSION_CONFIG = _io_config.IoSessionConfig(compression_ratio=3)

    def __init__(self, connections_cmt, manage_connections=False, aligned=True):
        super(StandardVdbProfile, self).__init__(
            connections_cmt=connections_cmt,
            session_config=copy.deepcopy(self.DEFAULT_SESSION_CONFIG),
            manage_connections=manage_connections
        )
        self.workload_configs = copy.deepcopy(self.DEFAULT_WORKLOAD_CONFIGS)
        self.period_configs = copy.deepcopy(self.DEFAULT_PERIOD_CONFIGS)

        self.aligned = aligned
        if aligned:
            self.xfer_size_configs = copy.deepcopy(self.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED)
        else:
            self.xfer_size_configs = copy.deepcopy(self.DEFAULT_XFER_SIZE_CONFIGS_UNALIGNED)

        # Only set dedup if all workloads are aligned
        # NOTE: Unaligned workloads require dedup_unit=512, but SolidFire only supports dedup_unit=4k
        if aligned:
            self.session_config.dedup_config = self.DEDUP_2X

    @classmethod
    def from_volumes(cls, tag, cluster_dm, client_dms, volumes_cmts, connections_per_volume=1,
                     aligned=True):
        """Construct a standard profile from a set of volumes.

        NOTE: Creates a connections component which will apply/reset the following for you:
            * Initiators/volume-access-groups needed to form the connections

        :param str tag: Unique tag for all the managed components to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which the volumes reside.
        :param list[valence.models.ResourceClient] client_dms: Clients to connect.
        :param list[valence.vl_components.VolumesInterface] volumes_cmts: Volumes to
            connect.
        :param int connections_per_volume: Number of clients to connect to each volume (default: 1).
        :param bool aligned: Whether to run aligned/unaligned IO (default: aligned).
        :rtype: valence.vl_components.vdbench.StandardVdbProfile
        """
        connections_cmt = _connection.ManageConnections.from_volumes_cmts(
            tag=tag,
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            volumes_cmts=volumes_cmts,
            connections_per_volume=connections_per_volume
        )
        return cls(connections_cmt, manage_connections=True, aligned=aligned)

    @classmethod
    def from_clients(cls, tag, cluster_dm, client_dms, volume_count=60, connections_per_volume=1,
                     aligned=True, volume_size=None):
        """Construct a standard profile from a set of clients.

        NOTE: Creates a connections component which will apply/reset the following for you:
            * Volumes/accounts needed for the requested connections
            * Initiators/volume-access-groups needed to form the connections

        :param str tag: Unique tag for all the managed components to be created.
        :param valence.models.ResourceCluster cluster_dm: The cluster on which the volumes reside.
        :param list[valence.models.ResourceClient] client_dms: Clients to connect.
        :param int volume_count: Number of volumes to create (default: 60 = 5 workload configs *
            12 aligned transfer-size configs).
        :param int connections_per_volume: Number of clients to connect to each volume (default: 1).
        :param bool aligned: Whether to run aligned/unaligned IO (default: aligned).
        :param valence.models.DataSize volume_size: Size to use for all volumes (default: 128GiB).
        :rtype: valence.vl_components.vdbench.StandardVdbProfile
        """
        # Create the accounts
        max_vols_per_account = 500  # NOTE: Hard-coded to avoid making API calls in the init
        min_accounts = 1 + (volume_count - 1) // max_vols_per_account
        cmt_count = max(min_accounts, len(client_dms), len(cls.DEFAULT_WORKLOAD_CONFIGS))
        accounts_cmt = _account.CreateAccounts(tag=tag, cluster_dm=cluster_dm, num=cmt_count)

        # Define the volume configurations
        if volume_size is None:
            volume_size = models.DataSize(128, enums.DataUnit.gibibytes)
        volume_dm_standard = models.Volume.create(total_size=volume_size, enable512e=not aligned)
        volume_dm_512e = models.Volume.create(total_size=volume_size, enable512e=True)
        volume_dm_reduced = models.Volume.create(
            total_size=volume_size,
            enable512e=not aligned,
            qos=cls.REDUCED_QOS
        )
        volume_dm_512e_reduced = models.Volume.create(
            total_size=volume_size,
            enable512e=True,
            qos=cls.REDUCED_QOS
        )

        # Create the volumes
        volumes_cmts = []
        cmt_volume_count = volume_count // cmt_count
        volumes_remainder = volume_count % cmt_count
        for i in range(cmt_count):
            # Divide the volumes evenly into components
            volume_count = cmt_volume_count
            if volumes_remainder:
                volume_count += 1
                volumes_remainder -= 1
            elif volume_count == 0:
                break

            # Vary the QOS and 512e settings
            if bool((i + 1) % 2):
                # For half of the volumes, use 512e if unaligned
                if bool((i + 6) % 10):
                    volume_dm = volume_dm_standard
                else:
                    # For up to 20% of the volumes, use reduced QOS
                    volume_dm = volume_dm_reduced
            else:
                # For half of the volumes, use 512e regardless of alignment
                if bool((i + 9) % 10):
                    volume_dm = volume_dm_512e
                else:
                    # For up to 20% of the volumes, use reduced QOS
                    volume_dm = volume_dm_512e_reduced

            volumes_cmts.append(_volume.CreateMultipleVolumes(
                tag='{}-{}'.format(tag, i),
                cluster_dm=cluster_dm,
                vol_count=volume_count,
                partial_volume_dm=volume_dm,
                accounts_cmt=accounts_cmt,
                account_index=i
            ))

        connections_cmt = _connection.ManageConnections.from_volumes_cmts(
            tag=tag,
            cluster_dm=cluster_dm,
            client_dms=client_dms,
            volumes_cmts=volumes_cmts,
            connections_per_volume=connections_per_volume
        )
        connections_cmt.accounts_cmts = [accounts_cmt]
        connections_cmt.volumes_cmts = volumes_cmts
        return cls(connections_cmt, manage_connections=True, aligned=aligned)

    def override_configs(self, io_rate=None, read_perc=None,  # pylint: disable=too-complex
                         transfer_size_config=None, seek_config=None, thread_count=None):
        """Override the configuration parameters for all workloads.

        :param int io_rate: IO rate override.
        :param int read_perc: Read percentage override.
        :param valence.vl_components.vdbench.TransferSizeConfig transfer_size_config: Transfer size
            configuration override.
        :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration override.
        :param int thread_count: Thread count override.
        """
        if io_rate is not None:
            for config in self.period_configs:
                config.io_rate_config = _parm.IoRateFixed(io_rate)
        if read_perc is not None:
            for config in self.workload_configs:
                config.read_perc = read_perc
        if transfer_size_config is not None:
            self.xfer_size_configs = [transfer_size_config]
        if seek_config is not None:
            for config in self.workload_configs:
                config.seek_config = seek_config
        if thread_count is not None:
            for config in self.workload_configs:
                config.thread_count = thread_count

    def sanitize_configs(self, node_dms=None, raise_on_failure=False):
        """Sanitize the configuration parameters for any known Vdbench limitations.

        Automatically adjusts configuration parameters as needed to ensure Vdbench can run properly.

        :param list[valence.models.ResourceNode] node_dms: Nodes against which IO will be run.
        :param bool raise_on_failure: Whether to raise an exception if an invalid configuration is found
            rather than automatically sanitizing it.
        """
        if node_dms:
            virtual_nodes = [dm for dm in node_dms if dm.model.is_virtual]
            virtual_node_invalid_configs = [
                config for config in self.period_configs
                if not isinstance(config.io_rate_config, _parm.IoRateFixed)
                or config.io_rate_config.io_rate <= 100
            ]
            if virtual_nodes and raise_on_failure:
                raise ValueError(
                    'Configured IO rates[{}] exceed the limitations of virtual nodes: {}'
                    .format([c.io_rate_config for c in virtual_node_invalid_configs], virtual_nodes)
                )
            elif virtual_nodes:
                for config in virtual_node_invalid_configs:
                    _log.debug('Reconfiguring IO rate[{}] for virtual node limitations'.format(
                        config.io_rate_config
                    ))
                    config.io_rate_config = _parm.IoRateFixed(100)

    def get_workloads(self):
        """Get a set of standard workloads.

        Example Usage:
            vdb_profile = StandardVdbProfile(connections_cmt)
            io_workloads = vdb_profile.get_workloads()

            io_rate_config = IoRateMax()
            io_period_config = IoPeriodConfig(io_rate_config)
            io_period = IoPeriod(io_period_config, io_workloads)

            io_session_config = IoSessionConfig()
            io_session = IoSession(io_session_config, [io_period])

            setup_vdb_cmt = SetupVdbEnvironment(io_session, vdb_profile.connections_cmt)

        :rtype: list[valence.vl_components.vdbench.IoWorkload]
        """
        varied_configs = _helpers.get_varied_workload_configs(
            connections_cmt=self.connections_cmt,
            base_workload_configs=self.workload_configs,
            transfer_size_configs=self.xfer_size_configs
        )
        workloads = _helpers.get_workloads(
            connections_cmt=self.connections_cmt,
            workload_configs=varied_configs
        )
        _helpers.isolate_access_ranges(workloads)
        return workloads

    def get_periods(self):
        """Get a set of standard IO periods.

        Example Usage:
            vdb_profile = StandardVdbProfile(connections_cmt)
            io_periods = vdb_profile.get_periods()

            io_session_config = IoSessionConfig()
            io_session = IoSession(io_session_config, io_periods)

            setup_vdb_cmt = SetupVdbEnvironment(io_session, vdb_profile.connections_cmt)

        :rtype: list[valence.vl_components.vdbench.IoPeriod]
        """
        workloads = self.get_workloads()
        periods = [_io_config.IoPeriod(config, workloads) for config in self.period_configs]
        return periods


class StandardVdbExtendedProfile(_base.BaseVdbExtendedProfile):
    """Standard Vdbench extended profile (supports multiple master instances).

    :param str tag: Unique tag for the accounts/volumes to be created.
    :param valence.models.ResourceCluster cluster_dm: The cluster on which to create volumes.
    :param list[valence.models.ResourceClient] client_dms: The clients use.
    :param int volume_count: Number of volumes to create (default: 70).
    :param int connections_per_volume: Number of clients to connect to each volume (default: 1).
    :param float percent_aligned: Float between 0 and 1 representing the percentage of IO to make
        aligned vs. unaligned (default: 0.85).
    :param int slaves_per_master: The number of clients to use as slaves to each vdbench master
        (default: Max supported). By default the master runs alongside one of the slaves, so this number
        represents the total number of clients per Vdbench session.
    :param valence.models.DataSize volume_size: Size to use for all volumes (default: 128GiB).
    """
    BASE_PROFILE = StandardVdbProfile

    def __init__(self, tag, cluster_dm, client_dms, volume_count=70, connections_per_volume=1,
                 percent_aligned=0.85, slaves_per_master=None, volume_size=None):
        assert 0.0 <= percent_aligned <= 1.0
        slaves_per_master = self.MAX_SLAVE_COUNT if slaves_per_master is None else slaves_per_master
        assert slaves_per_master <= self.MAX_SLAVE_COUNT

        max_connections_per_volume = max(1, min(len(client_dms) - 1, slaves_per_master))
        assert connections_per_volume <= max_connections_per_volume

        # NOTE: For single-client vdbench the slave may run masterless for reduced memory overhead
        standalone = bool(slaves_per_master == 0)

        client_index = 0
        profiles = []
        # If there is more than one client, setup an unaligned IO profile
        if len(client_dms) > 1 and percent_aligned < 1.0:
            client_index += 1
            unaligned_volume_count = int(volume_count * (1.0 - percent_aligned))
            profiles.append(self.BASE_PROFILE.from_clients(
                tag='{}-{}'.format(tag, len(profiles)),
                cluster_dm=cluster_dm,
                client_dms=client_dms[0:1],
                volume_count=unaligned_volume_count,
                connections_per_volume=1,
                aligned=False,
                volume_size=volume_size
            ))
            volume_count -= unaligned_volume_count
        # Use all remaining clients for aligned IO profiles
        # NOTE: Clients will be split up if the max-slave-per-master limit is exceeded
        while client_index < len(client_dms):
            next_index = min(client_index + max(slaves_per_master, 1), len(client_dms))
            profile_client_dms = client_dms[client_index:next_index]
            profiles.append(self.BASE_PROFILE.from_clients(
                tag='{}-{}'.format(tag, len(profiles)),
                cluster_dm=cluster_dm,
                client_dms=profile_client_dms,
                volume_count=volume_count,
                connections_per_volume=min(connections_per_volume, len(profile_client_dms)),
                volume_size=volume_size
            ))
            client_index = next_index

        super(StandardVdbExtendedProfile, self).__init__(profiles, standalone)

    def override_configs(self, io_rate=None, read_perc=None, transfer_size_config=None,
                         seek_config=None, thread_count=None):
        """Override the configuration parameters for all workloads.

        :param int io_rate: IO rate override.
        :param int read_perc: Read percentage override.
        :param valence.vl_components.vdbench.TransferSizeConfig transfer_size_config: Transfer size
            configuration override.
        :param valence.vl_components.vdbench.SeekConfig seek_config: Seek configuration override.
        :param int thread_count: Thread count override.
        """
        for profile in self._profiles:
            profile.override_configs(
                io_rate=io_rate,
                read_perc=read_perc,
                transfer_size_config=transfer_size_config,
                seek_config=seek_config,
                thread_count=thread_count
            )
