"""Vdbench CSV-based profile helper methods."""
from __future__ import division

import ast
import csv
import os
import re

from autologger import get_sflogging

from sf_api import data_size
from valence import models
from valence.vl_components.vdbench import (
    _io_config,
    _parm,
)
from valence.vl_components.vdbench.profiles.csv_profiles import _constants

_log = get_sflogging(__name__)

DIR_PATH = os.path.dirname(os.path.realpath(os.path.expanduser(__file__)))
CSV_PROFILES_PATH = os.path.join(DIR_PATH, 'csv_profiles')


def get_client_dir_map(csv_dir):
    """Get a map of CSV directories to their respective numbers of supported clients.

    :param str csv_dir: Path to a directory of CSV-defined vdbench profiles. These profiles must be
        organized into sub-directories based on client count (i.e. 1_client, 2_clients, etc)).
    :rtype: dict[int,str]
    """
    client_dir_map = {}
    for sub_dir in next(os.walk(csv_dir))[1]:
        if '_client' in sub_dir:
            client_dir_map[int(sub_dir.split('_client')[0])] = sub_dir

    return client_dir_map


def get_connection_definitions(csv_path):
    """Get a list of connection definitions for the specified CSV.

    :param str csv_path: File path for a CSV-defined vdbench profile.
    :rtype: list[dict]
    """
    _log.debug('Loading vdbench connection definitions: {}'.format(csv_path))
    with open(csv_path) as csv_file:
        return list(csv.DictReader(csv_file))


def generate_volume_dm(conn_def):
    """Generates a volume data model based on the provided profile.

    :param dict conn_def: A vdbench profile for a client-to-volume-component connection.
    :rtype: valence.models.Volume
    """
    volume_size_gb = conn_def.get(_constants.VOLUME_SIZE) or 128
    return models.Volume.create(
        total_size=data_size.DataSize(volume_size_gb, data_size.DataUnit.gibibytes),
        enable512e=conn_def.get(_constants.BLOCK_SIZE, '4k').lower() != '4k'.lower(),
        qos=models.Qos.create(
            min_iops=int(conn_def[_constants.MIN_IOPS] or 100),
            max_iops=int(conn_def[_constants.MAX_IOPS] or 15000),
            burst_iops=int(conn_def[_constants.BURST_IOPS] or 15000)
        )
    )


def get_workload_config(conn_def):
    """Get a set of workload configurations representing the set of provided connection definitions.

    :param dict conn_def: CSV-based connection definition.
    :rtype: valence.vl_components.vdbench.IoWorkloadConfig
    """
    return _io_config.IoWorkloadConfig(
        read_perc=int(conn_def[_constants.READPCT] or 100),
        transfer_size_config=get_transfer_size_config(conn_def[_constants.XFERSIZE]),
        seek_config=get_seek_config(conn_def[_constants.SEEKPCT]),
        thread_count=int(conn_def[_constants.THREADS] or 8),
        access_range_config=get_access_range_config(conn_def[_constants.RANGE]),
    )


def get_transfer_size_config(xfersize):
    """Convert string xfersize into wd parm TransferSizeBase.

    :param str xfersize: The xfersize as provided in csv. Can convert from type '4k' or '(4k,10,8k,90)'.
    :rtype: valence.vl_components.vdbench.TransferSizeConfig
    """
    if ',' in xfersize:
        # Handle variable xfersize configurations
        xfersize_pairs = re.sub('[()]', '', xfersize).split(',')
        xfersize_percents = [int(perc) for perc in xfersize_pairs[1::2]]
        xfersize_strings = xfersize_pairs[::2]
        xfersizes = []
        for xfersize_str in xfersize_strings:
            if 'k' in xfersize_str:
                xfersizes.append(int(1024 * float(xfersize_str.replace('k', ''))))
            else:
                xfersizes.append(int(xfersize_str))

        return _parm.TransferSizeVariable(list(zip(xfersizes, xfersize_percents)))
    elif 'k' in xfersize:
        return _parm.TransferSizeFixed(int(1024 * float(xfersize.replace('k', ''))))
    else:
        return _parm.TransferSizeFixed(int(xfersize))


def get_access_range_config(range_str):
    """Convert string xfersize into wd parm TransferSizeBase.

    :param str range_str: The range as provided in csv.
    :rtype: valence.vl_components.vdbench.AccessRangePercentages|None
    """
    if range_str:
        percent_range = ast.literal_eval(range_str)
        return _parm.AccessRangePercentages(percent_range[0], percent_range[1])
    else:
        return None


def get_seek_config(seekpct):
    """Converts the seekpct string into wd parm SeekPercentage.

    Defaults to seq if provided with an empty string.

    :param str seekpct: The seekpct as provided in csv. Converts seq, rand, eof or any percentage.
    :rtype: valence.vl_components.vdbench.SeekConfig
    """
    if seekpct == 'seq':
        seek_perc = 0
    elif seekpct == 'rand' or seekpct == 'random':
        seek_perc = 100
    elif seekpct == 'eof':
        seek_perc = -1
    else:
        seek_perc = int(seekpct) or 0

    return _parm.SeekConfig(seek_perc)
