from __future__ import (
    absolute_import,
    division,
)

import unittest

from valence import models
from valence.vl_components import (
    _account,
    _volume,
)
from valence.vl_components.vdbench import (
    _io_config,
    _parm,
)


class TestIoWorkloadConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.transfer_size_config = _parm.TransferSizeFixed(4096)
        cls.workload_config = _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=cls.transfer_size_config,
            skew_perc=42,
            priority=1,
            thread_count=10
        )

    def test_offset(self):
        with self.assertRaises(AssertionError):
            workload_config = _io_config.IoWorkloadConfig(offset=1023)
        workload_config = _io_config.IoWorkloadConfig(offset=1024)

    def test_eq(self):
        assert self.workload_config != 'mock_string'
        assert self.workload_config != _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=self.transfer_size_config,
            skew_perc=42,
            priority=1,
            thread_count=9
        )
        assert self.workload_config == _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=self.transfer_size_config,
            skew_perc=42,
            priority=1,
            thread_count=10
        )

    def test_hash(self):
        identical_workload_config = _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=self.transfer_size_config,
            skew_perc=42,
            priority=1,
            thread_count=10
        )
        other_workload_config = _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=self.transfer_size_config,
            skew_perc=42,
            priority=1,
            thread_count=9
        )
        workload_config_dict = {self.workload_config: 1, other_workload_config: 2}
        workload_config_dict[identical_workload_config] = 3
        assert workload_config_dict[self.workload_config] == 3

    def test_repr(self):
        assert str(self.workload_config) == (
            'IoWorkloadConfig<io_rate={}, transfer_size={}, read_perc={}, access_range={}>'.format(
                self.workload_config.io_rate,
                self.workload_config.transfer_size_config,
                self.workload_config.read_perc,
                self.workload_config.access_range_config
            )
        )


class TestIoWorkload(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.client_dm = models.ResourceClient.create(mip='1.0.0.{}')
        cls.mock_accounts = _account.MockAccounts(
            tag='mock',
            account_dms=models.Account.get_mock_object()
        )
        cls.mock_volumes = _volume.MockVolumes(
            tag='mock',
            cluster_dm=cls.cluster_dm,
            volume_dms=[models.Volume.get_mock_object()],
            accounts_cmt=cls.mock_accounts
        )

        cls.transfer_size_config = _parm.TransferSizeFixed(4096)
        cls.workload_config = _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=cls.transfer_size_config,
            priority=1,
            thread_count=10
        )

        cls.io_workload = _io_config.IoWorkload(
            io_workload_config=cls.workload_config,
            client_dm=cls.client_dm,
            volumes_cmt=cls.mock_volumes,
            volume_index=0
        )

    def test_one_volume(self):
        more_volumes = _volume.MockVolumes(
            tag='more',
            cluster_dm=self.cluster_dm,
            volume_dms=[models.Volume.get_mock_object()],
            accounts_cmt=self.mock_accounts
        )
        with self.assertRaises(AssertionError):
            workload = _io_config.IoWorkload(
                io_workload_config=self.workload_config,
                client_dm=self.client_dm,
                volumes_cmt=_volume.ManageVolumes([self.mock_volumes, more_volumes]),
                volume_index=0
            )

    def test_aligned(self):
        workload_config = _io_config.IoWorkloadConfig()
        assert workload_config.aligned

        workload_config = _io_config.IoWorkloadConfig(
            transfer_size_config=_parm.TransferSizeFixed(4096)
        )
        assert workload_config.aligned

        workload_config = _io_config.IoWorkloadConfig(
            transfer_size_config=_parm.TransferSizeFixed(3 * 512)
        )
        assert not workload_config.aligned

    def test_copy(self):
        workload_copy = self.io_workload.copy()
        assert workload_copy == self.io_workload

    def test_eq(self):
        assert self.io_workload != 'mock_string'
        identical_workload_config = _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=self.transfer_size_config,
            priority=1,
            thread_count=10
        )
        identical_workload = _io_config.IoWorkload(
            io_workload_config=identical_workload_config,
            client_dm=self.client_dm,
            volumes_cmt=self.mock_volumes,
            volume_index=0
        )
        assert self.io_workload == identical_workload
        other_workload_config = _io_config.IoWorkloadConfig(
            io_rate=100,
            transfer_size_config=self.transfer_size_config,
            priority=1,
            thread_count=9
        )
        other_workload = _io_config.IoWorkload(
            io_workload_config=other_workload_config,
            client_dm=self.client_dm,
            volumes_cmt=self.mock_volumes,
            volume_index=0
        )
        assert self.io_workload != other_workload

        other_workload = _io_config.IoWorkload(
            io_workload_config=self.workload_config,
            client_dm=self.client_dm,
            volumes_cmt=self.mock_volumes,
            volume_index=1
        )
        assert self.io_workload != other_workload

    def test_hash(self):
        workload_copy = self.io_workload.copy()
        assert hash(workload_copy) != hash(self.io_workload)


class TestIoPeriodConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.io_rate_config = _parm.IoRateMax()
        cls.period_config = _io_config.IoPeriodConfig(
            io_rate_config=cls.io_rate_config,
            interval=3,
            max_data=1000,
            warmup=60
        )

    def test_durations(self):
        default_duration = 30

        period_config = _io_config.IoPeriodConfig(self.io_rate_config)
        assert period_config.duration == default_duration

        period_config = _io_config.IoPeriodConfig(self.io_rate_config, duration=5)
        assert period_config.duration == 5

        curve = (10, 20, 30)
        period_config = _io_config.IoPeriodConfig(_parm.IoRateCurve(curve=curve))
        assert period_config.duration == default_duration * len(curve)

        series = (100, 400, 700, 1000)
        period_config = _io_config.IoPeriodConfig(_parm.IoRateSeries(io_rates=series))
        assert period_config.duration == default_duration * len(series)

        variable_rates = [(100, 10), (400, 7)]
        period_config = _io_config.IoPeriodConfig(_parm.IoRateVariable(io_rates=variable_rates))
        assert period_config.duration == 17

    def test_eq(self):
        assert self.period_config != 'mock_string'
        assert self.period_config != _io_config.IoPeriodConfig(
            io_rate_config=self.io_rate_config,
            interval=3,
            max_data=1001,
            warmup=60
        )
        assert self.period_config == _io_config.IoPeriodConfig(
            io_rate_config=self.io_rate_config,
            interval=3,
            max_data=1000,
            warmup=60
        )

    def test_hash(self):
        identical_period_config = _io_config.IoPeriodConfig(
            io_rate_config=self.io_rate_config,
            interval=3,
            max_data=1000,
            warmup=60
        )
        other_period_config = _io_config.IoPeriodConfig(
            io_rate_config=self.io_rate_config,
            interval=3,
            max_data=1001,
            warmup=60
        )
        period_config_dict = {self.period_config: 1, other_period_config: 2}
        period_config_dict[identical_period_config] = 3
        assert period_config_dict[self.period_config] == 3

    def test_repr(self):
        assert str(self.period_config) == 'IoPeriodConfig<io_rate={}, duration={}>'.format(
            self.period_config.io_rate_config,
            self.period_config.duration
        )


class TestIoPeriod(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.client_dm = models.ResourceClient.create(mip='1.0.0.{}')
        cls.mock_accounts = _account.MockAccounts(
            tag='mock',
            account_dms=models.Account.get_mock_object()
        )
        cls.mock_volumes = _volume.MockVolumes(
            tag='mock',
            cluster_dm=cls.cluster_dm,
            volume_dms=[models.Volume.get_mock_object()],
            accounts_cmt=cls.mock_accounts
        )

        cls.transfer_size_config = _parm.TransferSizeFixed(4096)

        cls.workload_configs = [_io_config.IoWorkloadConfig(
            io_rate=index + 1,
            transfer_size_config=cls.transfer_size_config,
            priority=1,
            thread_count=10
        ) for index in range(_io_config.IoPeriod.MAX_WORKLOADS + 1)]

        cls.io_workloads = [_io_config.IoWorkload(
            io_workload_config=workload_config,
            client_dm=cls.client_dm,
            volumes_cmt=cls.mock_volumes,
            volume_index=index
        ) for index, workload_config in enumerate(cls.workload_configs[:-1])]

        cls.io_rate_config = _parm.IoRateMax()
        cls.period_config = _io_config.IoPeriodConfig(
            io_rate_config=cls.io_rate_config,
            interval=3,
            max_data=1000,
            warmup=60
        )

        cls.io_period = _io_config.IoPeriod(cls.period_config, cls.io_workloads[:2])

    def test_max_workloads(self):
        max_io_workloads = self.io_workloads + [_io_config.IoWorkload(
            io_workload_config=self.workload_configs[-1],
            client_dm=self.client_dm,
            volumes_cmt=self.mock_volumes,
            volume_index=len(self.io_workloads)
        )]

        assert len(max_io_workloads) == _io_config.IoPeriod.MAX_WORKLOADS + 1
        with self.assertRaises(AssertionError):
            period = _io_config.IoPeriod(self.period_config, max_io_workloads)

        # Test: len(io_workloads) > MAX_WORKLOADS, len(<unique configs for io_workloads>) ==
        # MAX_WORKLOADS

        max_io_workloads[-1].config = self.workload_configs[0]
        period = _io_config.IoPeriod(self.period_config, max_io_workloads)
        assert len(period.io_workloads) == len(max_io_workloads)

    def test_copy(self):
        period_copy = self.io_period.copy()
        assert period_copy == self.io_period

    def test_eq(self):
        assert self.io_period != 'mock_string'
        identical_period_config = _io_config.IoPeriodConfig(
            io_rate_config=self.io_rate_config,
            interval=3,
            max_data=1000,
            warmup=60
        )
        identical_period = _io_config.IoPeriod(identical_period_config, self.io_workloads[:2])
        assert self.io_period == identical_period
        other_period_config = _io_config.IoPeriodConfig(
            io_rate_config=self.io_rate_config,
            interval=3,
            max_data=1001,
            warmup=60
        )
        other_period = _io_config.IoPeriod(other_period_config, self.io_workloads[:2])
        assert self.io_period != other_period

        other_period.config = self.io_period.config
        other_period.io_workloads = self.io_period.io_workloads[::-1]
        assert other_period == self.io_period

        other_period.io_workloads.append(self.io_workloads[2])
        assert other_period != self.io_period

        other_period.io_workloads.remove(self.io_workloads[1])
        assert len(other_period.io_workloads) == len(self.io_period.io_workloads)
        assert other_period != self.io_period

    def test_hash(self):
        period_copy = self.io_period.copy()
        assert hash(period_copy) != hash(self.io_period)


class TestIoSessionConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dedup_config = _parm.DedupConfig(dedup_ratio=5)
        cls.session_config = _io_config.IoSessionConfig(
            compression_ratio=2.1,
            dedup_config=cls.dedup_config
        )

    def test_eq(self):
        assert self.session_config != 'mock_string'
        assert self.session_config != _io_config.IoSessionConfig(
            compression_ratio=2.0,
            dedup_config=self.dedup_config
        )
        assert self.session_config == _io_config.IoSessionConfig(
            compression_ratio=2.1,
            dedup_config=self.dedup_config
        )

    def test_hash(self):
        identical_session_config = _io_config.IoSessionConfig(
            compression_ratio=2.1,
            dedup_config=self.dedup_config
        )
        other_session_config = _io_config.IoSessionConfig(
            compression_ratio=2.0,
            dedup_config=self.dedup_config
        )
        session_config_dict = {self.session_config: 1, other_session_config: 2}
        session_config_dict[identical_session_config] = 3
        assert session_config_dict[self.session_config] == 3

    def test_repr(self):
        assert str(self.session_config) == (
            'IoSessionConfig<compression_ratio={}, dedup_config={}>'.format(
                self.session_config.compression_ratio,
                self.session_config.dedup_config
            )
        )


class TestIoSession(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.client_dms = [models.ResourceClient.create(mip='1.0.0.{}'.format(i)) for i in range(2)]
        cls.mock_accounts = _account.MockAccounts(
            tag='mock',
            account_dms=models.Account.get_mock_object()
        )
        cls.mock_volumes = _volume.MockVolumes(
            tag='mock',
            cluster_dm=cls.cluster_dm,
            volume_dms=[models.Volume.get_mock_object()],
            accounts_cmt=cls.mock_accounts
        )

        cls.workload_configs = [
            _io_config.IoWorkloadConfig(
                io_rate=100,
                transfer_size_config=_parm.TransferSizeFixed(4096),
                priority=1,
                thread_count=10
            ),
            _io_config.IoWorkloadConfig(
                io_rate=100,
                transfer_size_config=_parm.TransferSizeVariable([
                    (4 * 1024, 30),
                    (32 * 1024, 20),
                    (128 * 1024, 20),
                    (256 * 1024, 20),
                    (512 * 1024, 10)
                ]),
                priority=1,
                thread_count=10
            )
        ]
        cls.io_workloads = [
            _io_config.IoWorkload(
                io_workload_config=config,
                client_dm=cls.client_dms[0],
                volumes_cmt=cls.mock_volumes,
                volume_index=0
            ) for config in cls.workload_configs
        ]

        cls.io_periods = []
        period_config = _io_config.IoPeriodConfig(
            io_rate_config=_parm.IoRateMax(),
            interval=3,
            max_data=1000,
            warmup=60
        )
        cls.io_periods.extend([_io_config.IoPeriod(period_config, cls.io_workloads) for _ in range(3)])
        period_config = _io_config.IoPeriodConfig(
            io_rate_config=_parm.IoRateFixed(1000),
            interval=3,
            max_data=1000,
            warmup=60,
            for_transfer_size_config=_parm.ForTransferSizeRange(4096, 8 * 4096, 4096)
        )
        cls.io_periods.append(_io_config.IoPeriod(period_config, cls.io_workloads))

        cls.dedup_config = _parm.DedupConfig(dedup_ratio=5, dedup_unit=4 * 1024)
        cls.session_config = _io_config.IoSessionConfig(
            compression_ratio=2.1,
            dedup_config=cls.dedup_config
        )

        cls.io_session = _io_config.IoSession(cls.session_config, cls.io_periods)

    def test_standalone(self):
        io_periods = [period.copy() for period in self.io_periods]
        session = _io_config.IoSession(self.session_config, io_periods, standalone=True)

        other_workload = self.io_workloads[0].copy()
        other_workload.client_dm = self.client_dms[1]
        io_periods[0].io_workloads.append(other_workload)
        with self.assertRaises(AssertionError):
            session = _io_config.IoSession(self.session_config, io_periods, standalone=True)

    def test_copy(self):
        session_copy = self.io_session.copy()
        assert session_copy.config == self.io_session.config
        for i in range(len(session_copy.io_periods)):
            assert session_copy.io_periods[i] == self.io_session.io_periods[i]

    def test_eq(self):
        assert self.io_session != 'mock_string'
        identical_session_config = _io_config.IoSessionConfig(
            compression_ratio=2.1,
            dedup_config=self.dedup_config
        )
        identical_session = _io_config.IoSession(identical_session_config, self.io_periods)
        assert self.io_session == identical_session
        other_session_config = _io_config.IoSessionConfig(
            compression_ratio=2.0,
            dedup_config=self.dedup_config
        )
        other_session = _io_config.IoSession(other_session_config, self.io_periods)
        assert self.io_session != other_session

        other_session = _io_config.IoSession(self.io_session.config, self.io_periods[:-1])
        assert self.io_session != other_session

        other_session = _io_config.IoSession(self.io_session.config, self.io_periods, standalone=True)
        assert self.io_session != other_session

    def test_hash(self):
        session_copy = self.io_session.copy()
        assert hash(session_copy) != hash(self.io_session)
