import os
import unittest

import mock
import pytest

from valence.vl_components.vdbench import (
    _log_parser,
    _run,
    _stream_handlers,
)
from valence.vl_components.vdbench._log_parser import helpers

UNIT_TEST_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_DIR = os.path.join(UNIT_TEST_DIR, 'data')


class TestModels(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.run_name = 'r0000'
        cls.raw_run_line = '03:23:01.003      Starting RD={};'.format(cls.run_name)
        cls.mock_run_line = _log_parser.VdbRunInfoLine(cls.raw_run_line)
        cls.mock_run_line.run_info.index = 1
        cls.mock_run_line.run_time = 42

        cls.raw_io_line = '23:28:18.067    1  692.00  2.70  4096  NaN NaN NaN NaN NaN NaN NaN NaN NaN'
        cls.mock_io_line = _log_parser.VdbLogLine(
            raw_line=cls.raw_io_line,
            line_type=_log_parser.VdbLogLineType.io,
            last_line=cls.mock_run_line
        )

    def test_vdb_log_line(self):
        line = _log_parser.VdbLogLine(self.raw_io_line, _log_parser.VdbLogLineType.io)
        assert line.raw == self.raw_io_line
        assert line.type == _log_parser.VdbLogLineType.io
        assert line.run_info is None
        assert line.run_time is None
        assert str(line) == '{}: {}'.format(_log_parser.VdbLogLineType.io, self.raw_io_line)

        line = _log_parser.VdbLogLine(
            raw_line=self.raw_io_line,
            line_type=_log_parser.VdbLogLineType.io,
            last_line=self.mock_io_line
        )
        assert line.raw == self.raw_io_line
        assert line.type == _log_parser.VdbLogLineType.io
        assert line.run_info is self.mock_io_line.run_info
        assert line.run_time is self.mock_io_line.run_time

    def test_vdb_error_line(self):
        line = _log_parser.VdbLogLine(self.raw_io_line, _log_parser.VdbLogLineType.io_error)
        assert str(line) == 'ERROR {}: {}'.format(_log_parser.VdbLogLineType.io_error, self.raw_io_line)

    def test_vdb_run_info_line(self):
        line = _log_parser.VdbRunInfoLine(self.raw_run_line)
        assert line.run_info.name == self.run_name
        assert line.run_info.index == 0
        assert line.run_time == 0
        assert str(line) == '{}: {}'.format(_log_parser.VdbLogLineType.run_start, self.raw_run_line)

        line = _log_parser.VdbRunInfoLine(self.raw_run_line, last_line=self.mock_io_line)
        assert line.run_info.name == self.run_name
        assert line.run_info.index == self.mock_io_line.run_info.index + 1
        assert line.run_time == self.mock_io_line.run_time + 1

    def test_get_float_or_none(self):
        value = _log_parser.VdbIoLine._get_float_or_none('50')
        assert value == 50

        value = _log_parser.VdbIoLine._get_float_or_none('10.007')
        assert value == 10.007

        value = _log_parser.VdbIoLine._get_float_or_none('NaN')
        assert value is None

    def test_vdb_io_line(self):
        defined_intervals = [3, 2, 4]
        line = _log_parser.VdbIoLine(
            raw_line=self.raw_io_line,
            defined_intervals=defined_intervals,
            last_line=self.mock_run_line
        )
        assert line.time_stamp == '23:28:18.067'
        assert line.interval == 1
        assert line.iops == 692
        assert line.response_time is None
        assert line.queue_depth is None

        active_interval = defined_intervals[self.mock_run_line.run_info.index]
        assert line.run_time == self.mock_run_line.run_time + active_interval
        assert str(line) == 'INTERVAL:1 [23:28:18.067] max_resp=None resp=None mbps=2.7 read%=None'

        line_dict = line.serialize()
        assert line_dict['time_stamp'] == '23:28:18.067'
        assert line_dict['interval'] == 1
        assert line_dict['iops'] == 692
        assert line_dict['response_time'] is None
        assert line_dict['queue_depth'] is None

        line.interval = None
        assert str(line) == 'AVERAGE: [23:28:18.067] max_resp=None resp=None mbps=2.7 read%=None'


class TestHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.run_name = 'r0000'
        cls.raw_run_line = '03:23:01.003      Starting RD={};'.format(cls.run_name)
        cls.mock_run_line = _log_parser.VdbRunInfoLine(cls.raw_run_line)
        cls.mock_run_line.run_info.index = 1
        cls.mock_run_line.run_time = 42

        cls.raw_io_line = (
            '23:28:18.067         1     692.00     2.70    4096  20.23    0.487    0.287    0.537    '
            '2.349    0.148   0.3  18.3   2.8'
        )
        cls.mock_io_line = _log_parser.VdbLogLine(
            raw_line=cls.raw_io_line,
            line_type=_log_parser.VdbLogLineType.io,
            last_line=cls.mock_run_line
        )

    def test_parse_raw_fast(self):
        defined_intervals = [1, 1]
        with open(os.path.join(DATA_DIR, 'master_log_fast')) as master_log:
            raw_lines = master_log.read().split('\n')
        lines = _log_parser.parse_raw(raw_lines, defined_intervals)

        run_line = lines[18]
        run_info = run_line.run_info
        assert run_line.type is _log_parser.VdbLogLineType.run_start
        assert isinstance(run_info, _log_parser.VdbRunInfo)
        assert str(run_info) == 'VdbRunInfo<name=r0000, index=0>'
        assert str(run_line) == (
            'VdbLogLineType.run_start: 03:23:01.003 Starting RD=r0000; '
            'I/O rate: 2000; elapsed=20; For loops: None'
        )

        io_line = lines[22]
        assert io_line.type is _log_parser.VdbLogLineType.io
        assert io_line.run_info == run_info
        assert io_line.interval == 1
        assert io_line.mbps == 3.92
        assert io_line.response_time == 0.513
        assert io_line.run_time == 1

    def test_parse_raw_journal(self):
        defined_intervals = [5, 5]
        with open(os.path.join(DATA_DIR, 'master_log_journal')) as master_log:
            raw_lines = master_log.read().split('\n')
        lines = _log_parser.parse_raw(raw_lines, defined_intervals)

        run_line = lines[134]
        run_info = run_line.run_info
        assert run_line.type is _log_parser.VdbLogLineType.run_start
        assert isinstance(run_info, _log_parser.VdbRunInfo)
        assert run_info.name == 'journal_recovery'
        assert run_info.index == 0
        io_line = lines[188]
        assert io_line.run_info == run_info

        run_line = lines[739]
        run_info = run_line.run_info
        assert run_line.type is _log_parser.VdbLogLineType.run_start
        assert isinstance(run_info, _log_parser.VdbRunInfo)
        assert run_info.name == 'r0000'
        assert run_info.index == 0
        io = lines[743]
        assert io.run_info == run_info

    @mock.patch('valence.vl_components.vdbench.RunVdb')
    def test_parse_raw_validation(self, run_vdb):
        defined_intervals = [5]
        with open(os.path.join(DATA_DIR, 'master_log_validation')) as master_log:
            raw_lines = master_log.read().split('\n')
        lines = _log_parser.parse_raw(raw_lines, defined_intervals)

        stream = _stream_handlers.ErrorStreamHandler(run_vdb)
        stream.handle_items(lines[:-12])
        assert not stream.critical_errors
        stream.handle_items(lines[-12:])
        assert not stream.critical_errors

    @mock.patch('valence.vl_components.vdbench.RunVdb')
    def test_parse_raw_terminated(self, run_vdb):
        defined_intervals = [1]
        with open(os.path.join(DATA_DIR, 'master_log_terminated')) as master_log:
            raw_lines = master_log.read().split('\n')
        lines = _log_parser.parse_raw(raw_lines, defined_intervals)

        assert str(lines[-2]).startswith('ERROR')
        stream = _stream_handlers.ErrorStreamHandler(run_vdb)
        stream.handle_items(lines)
        assert [] == stream.critical_errors

    @mock.patch('valence.vl_components.vdbench.RunVdb')
    def test_parse_raw_eio(self, run_vdb):
        defined_intervals = [1]
        with open(os.path.join(DATA_DIR, 'master_log_eio')) as master_log:
            raw_lines = master_log.read().split('\n')
        lines = _log_parser.parse_raw(raw_lines, defined_intervals)

        eio_line = lines[-10]
        assert eio_line.type == _log_parser.VdbLogLineType.io_error
        stream = _stream_handlers.ErrorStreamHandler(run_vdb)
        stream.handle_items(lines[-10:])
        assert stream.eio_count == 1

    def test_get_vdb_line_type(self):
        line_type = helpers._get_vdb_line_type(self.raw_run_line)
        assert line_type == _log_parser.VdbLogLineType.run_start
        line_type = helpers._get_vdb_line_type('mock   93,  42  interval   i/o   MB/sec')
        assert line_type == _log_parser.VdbLogLineType.run_date
        line_type = helpers._get_vdb_line_type(self.raw_io_line)
        assert line_type == _log_parser.VdbLogLineType.io
        line_type = helpers._get_vdb_line_type(
            '23:28:18.067  avg_50-14  5.27 39.1   92  NaN  NaN  6.1  9.33 0.1  NaN  72.333 NaN  NaN'
        )
        assert line_type == _log_parser.VdbLogLineType.avg_io
        line_type = helpers._get_vdb_line_type(
            '23:28:18.067   HeartBeat.checkHeartBeat(): slave ..has not responded for 125 seconds'
        )
        assert line_type == _log_parser.VdbLogLineType.client_heartbeat
        line_type = helpers._get_vdb_line_type('23:28:18.067 Slave hd11-3 aborting: Heartbeat monitor:')
        assert line_type == _log_parser.VdbLogLineType.master_heartbeat
        line_type = helpers._get_vdb_line_type(
            '23:28:18.067 *   Slave h0000-7 aborting: '
            'No read validations done during a Data Validation run. *'
        )
        assert line_type == _log_parser.VdbLogLineType.no_read_validations

        # Error lines
        line_type = helpers._get_vdb_line_type('EIO I/O error')
        assert line_type == _log_parser.VdbLogLineType.io_error
        line_type = helpers._get_vdb_line_type(
            '23:28:18.067    Receiving  unexpected    EOFException   from   slave:'
        )
        assert line_type == _log_parser.VdbLogLineType.end_of_file
        line_type = helpers._get_vdb_line_type('23:28:18.067    SocketException   from   slave:')
        assert line_type == _log_parser.VdbLogLineType.socket_exception
        line_type = helpers._get_vdb_line_type('At least one Data Validation error detected.')
        assert line_type == _log_parser.VdbLogLineType.data_validation_error

        # Java lines
        line_type = helpers._get_vdb_line_type('23:28:18.067   execute():   ... jstack')
        assert line_type == _log_parser.VdbLogLineType.jstack
        line_type = helpers._get_vdb_line_type('java.lang.NullPointerException')
        assert line_type == _log_parser.VdbLogLineType.jnull
        line_type = helpers._get_vdb_line_type('java.lang.RuntimeException')
        assert line_type == _log_parser.VdbLogLineType.jruntime
        line_type = helpers._get_vdb_line_type(
            'java.lang.RuntimeException: Slave mock-run prematurely terminated.'
        )
        assert line_type == _log_parser.VdbLogLineType.premature_termination

        # Miscellaneous
        line_type = helpers._get_vdb_line_type('')
        assert line_type == _log_parser.VdbLogLineType.new_line
        line_type = helpers._get_vdb_line_type("For documentation, see 'vdbench.pdf'.")
        assert line_type == _log_parser.VdbLogLineType.unknown
