from __future__ import absolute_import

import math
import unittest

import pytest

from camelot.models import at2_enums
from valence import models
from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _enums,
    _io_config,
    _parm,
)
from valence.vl_components.vdbench._parm import utils


class TestHelpersAndIoConfig(unittest.TestCase):
    def setUp(self):
        client_overrides = dict(
            type=at2_enums.ClientType.physical,
        )
        self.client_dm = models.ResourceClient.get_mock_object(overrides=client_overrides)
        self.client_dm2 = models.ResourceClient.get_mock_object(overrides=client_overrides)
        assert self.client_dm.type is at2_enums.ClientType.physical
        assert self.client_dm2.type is at2_enums.ClientType.physical

        cluster_dm = models.ResourceCluster.get_mock_object()
        account_dm = models.Account.get_mock_object()
        volumes_dms = [
            models.Volume.get_mock_object(overrides=dict(account_id=account_dm.account_id))
        ]
        volumes_dms2 = [
            models.Volume.get_mock_object(overrides=dict(account_id=account_dm.account_id))
        ]

        self.accounts_cmt = _account.MockAccounts('mock', [account_dm])
        self.volumes_cmt = _volume.MockVolumes(
            tag='mock-0',
            cluster_dm=cluster_dm,
            volume_dms=volumes_dms,
            accounts_cmt=self.accounts_cmt
        )
        self.volumes_cmt2 = _volume.MockVolumes(
            tag='mock-1',
            cluster_dm=cluster_dm,
            volume_dms=volumes_dms2,
            accounts_cmt=self.accounts_cmt
        )
        self.disk_map = {
            volumes_dms[0].volume_id: '/dev/sda42',
        }
        self.disk_map2 = {
            volumes_dms2[0].volume_id: '/dev/sdb42',
        }
        self.connection_cmt = _connection.MockConnection(
            self.disk_map,
            self.client_dm,
            [self.volumes_cmt],
        )
        self.connection_cmt2 = _connection.MockConnection(
            self.disk_map2,
            self.client_dm2,
            [self.volumes_cmt2],
        )
        self.manage_connection_cmt = _connection.ManageConnections(
            [self.connection_cmt, self.connection_cmt2])

        self.manage_connection_cmt.apply()

        io_workload_config = _io_config.IoWorkloadConfig(thread_count=5)

        self.io_workload = _io_config.IoWorkload(
            io_workload_config=io_workload_config,
            client_dm=self.client_dm,
            volumes_cmt=self.volumes_cmt,
            volume_index=0,
        )
        self.io_workload2 = self.io_workload.copy()
        self.io_workload2.client_dm = self.client_dm2
        self.io_workload2.volumes_cmt = self.volumes_cmt2

        self.io_workload_hit = self.io_workload.copy()
        self.io_workload_hit.config.hit_config = _parm.CacheHitConfig(
            4098, read_perc=50, write_perc=20)

        self.io_workload_xfer = self.io_workload.copy()
        self.io_workload_xfer.config.transfer_size_config = _parm.TransferSizeAligned(
            512, 4098, 512)

        self.workload_configs = [_io_config.IoWorkload(
            io_workload_config=io_workload_config,
            client_dm=self.client_dm,
            volumes_cmt=self.volumes_cmt,
            volume_index=0
        ) for _ in range(2000)]

        io_period_config = _io_config.IoPeriodConfig(
            io_rate_config=_parm.IoRateMax(), interval=5)

        self.io_period = _io_config.IoPeriod(io_period_config, [self.io_workload])
        self.io_period2 = self.io_period.copy()
        self.io_period2.io_workloads = [self.io_workload2]
        self.io_period3 = self.io_period.copy()
        self.io_period3.io_workloads = self.workload_configs

        io_session_config = _io_config.IoSessionConfig()
        self.io_session = _io_config.IoSession(io_session_config, [self.io_period])
        self.io_session2 = self.io_session.copy()
        self.io_session2.io_periods = [self.io_period, self.io_period2]
        self.io_session3 = self.io_session.copy()
        self.io_session3.io_periods = [self.io_period3]

        self.hd_def = _parm.HostDef('h0', self.client_dm)
        self.hd_def2 = _parm.HostDef('h1', self.client_dm2)

        self.sd_def = _parm.get_storage_def(
            's0', self.io_workload, self.hd_def, self.connection_cmt)
        self.sd_def2 = _parm.get_storage_def(
            's1', self.io_workload2, self.hd_def2, self.connection_cmt2)
        self.sd_def_hit = _parm.get_storage_def(
            's0', self.io_workload_hit, self.hd_def, self.connection_cmt)
        self.sd_def_xfer = _parm.get_storage_def(
            's0', self.io_workload_xfer, self.hd_def, self.connection_cmt)

        self.wd_def = _parm.get_workload_def(
            'w0', self.io_workload.config, [self.sd_def])
        self.wd_def2 = _parm.get_workload_def(
            'w1', self.io_workload2.config, [self.sd_def2])
        self.wd_def_hit = _parm.get_workload_def(
            'w0', self.io_workload_hit.config, [self.sd_def_hit])

        self.rd_def = _parm.get_run_def('r0', self.io_period, [self.wd_def])

        self.gd_def = _parm.get_general_def(self.io_session, self.connection_cmt)
        self.gd_def2 = _parm.get_general_def(
            self.io_session2, self.manage_connection_cmt)

        self.gd_def3 = _parm.get_general_def(self.io_session3, self.connection_cmt)

    def test_get_storage_def(self):
        assert (
            repr(self.sd_def) ==
            'StorageDef<sd=s0,host=h0,lun=/dev/sda42,threads=5,openflags=o_direct>'
        )
        assert (
            repr(self.sd_def_hit) ==
            'StorageDef<sd=s0,host=h0,lun=/dev/sda42,threads=5,'
            'hitarea=4098k,openflags=o_direct>'
        )
        assert (
            repr(self.sd_def_xfer) ==
            'StorageDef<sd=s0,host=h0,lun=/dev/sda42,threads=5,'
            'align=512,openflags=o_direct>'
        )

    def test_get_workload_def(self):
        assert repr(self.wd_def) == 'WorkloadDef<wd=w0,sd=s0,rdpct=80>'
        assert repr(self.wd_def_hit) == 'WorkloadDef<wd=w0,sd=s0,rdpct=80,rhpct=50,whpct=20>'

    def test_get_run_def(self):
        assert (
            repr(self.rd_def) ==
            'RunDef<rd=r0,wd=w0,'
            'iorate=max,distribution=exponential,'
            'elapsed=30,interval=5>'
        )

    def test_get_general_def(self):
        expected = (
            'GeneralDef<monitor=/tmp/vdbench_kill\n'
            'data_errors=0\n'
            'hd=h0000,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root\n'
            'sd=s0000,host=h0000,lun=/dev/sda42,threads=5,openflags=o_direct\n'
            'wd=w0000,sd=s0000,rdpct=80\n'
            'rd=r0000,wd=w0000,iorate=max,distribution=exponential'
            ',elapsed=30,interval=5>').format(self.client_dm.mip)
        assert repr(self.gd_def) == expected

        # Test with # workloads w/ same config) < MAX_WORKLOADS  and check
        # that we have only one workload definition in the returned GeneralDef object.

        assert len(self.gd_def.workload_defs) == 1

        expected = (
            'GeneralDef<monitor=/tmp/vdbench_kill\n'
            'data_errors=0\n'
            'hd=h0000,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root\n'
            'hd=h0001,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root\n'
            'sd=s0000,host=h0000,lun=/dev/sda42,threads=5,openflags=o_direct\n'
            'sd=s0001,host=h0001,lun=/dev/sdb42,threads=5,openflags=o_direct\n'
            'wd=w0000,sd=s0000,rdpct=80\n'
            'wd=w0001,sd=s0001,rdpct=80\n'
            'rd=r0000,wd=w0000,iorate=max,distribution=exponential'
            ',elapsed=30,interval=5\n'
            'rd=r0001,wd=w0001,iorate=max,distribution=exponential'
            ',elapsed=30,interval=5>'
        ).format(self.client_dm.mip, self.client_dm2.mip)

        assert repr(self.gd_def2) == expected

        # Add a perturbation that has (# of workloads w/ same config) > MAX_WORKLOADS and
        # check to make sure we end up with more than one workload definition in the
        # returned GeneralDef object

        num_expected_wds = int(math.ceil((len(self.io_period3.io_workloads)/float(
            _io_config.IoPeriod.MAX_WORKLOADS))))

        assert len(self.gd_def3.workload_defs) == num_expected_wds

    def test_io_workload_misc(self):
        with pytest.raises(AssertionError):
            _io_config.IoWorkloadConfig(offset=12)

        with pytest.raises(AssertionError):
            _io_config.IoWorkloadConfig(
                hit_config=_parm.CacheHitConfig(
                    4098, read_perc=50, write_perc=20),
                streams_config=_parm.StreamsConfig(10),
            )

        with pytest.raises(AssertionError):
            _io_config.IoWorkloadConfig(
                io_rate=1000,
                priority=None,
            )

        assert self.io_workload != 42
        assert self.io_workload.config != 42
        config_repr = (
            'IoWorkloadConfig<io_rate=None, transfer_size=None'
            ', read_perc=80, access_range=None>'
        )
        assert repr(self.io_workload.config) == config_repr
        assert (
            repr(self.io_workload) ==
            'IoWorkload<config=' + config_repr + '>'
        )

    def test_io_session_misc(self):
        assert self.io_session != 42
        assert self.io_session.config != 42

        assert self.io_session.config == _io_config.IoSessionConfig()
        assert self.io_session.config != _io_config.IoSessionConfig(compression_ratio=2)

        config_repr = 'IoSessionConfig<compression_ratio=None, dedup_config=None>'
        assert repr(self.io_session.config) == config_repr
        assert repr(self.io_session) == 'IoSession<config=' + config_repr + '>'

        mo_periods = self.io_session.copy()
        mo_periods.io_periods.append(self.io_session.io_periods[0])
        assert self.io_session != mo_periods

        dif_period = self.io_session.copy()
        dif_period.io_periods[0].config.interval = 10
        assert self.io_session != dif_period

        standalone_io_session = _io_config.IoSession(
            io_session_config=self.io_session.config,
            io_periods=[self.io_period],
            standalone=True
        )
        assert standalone_io_session != self.io_session
        with pytest.raises(AssertionError):
            _io_config.IoSession(
                io_session_config=self.io_session.config,
                io_periods=[self.io_period, self.io_period2],
                standalone=True
            )

    def test_io_period_misc(self):
        assert self.io_period != 42
        assert self.io_period.config != 42

        config_repr = (
            'IoPeriodConfig<io_rate=IoRateMax<iorate=max'
            ',distribution=exponential>, duration=30>'
        )
        assert repr(self.io_period.config) == config_repr
        assert repr(self.io_period) == 'IoPeriod<config=' + config_repr + '>'

        mo_workloads = self.io_period.copy()
        mo_workloads.io_workloads.append(mo_workloads.io_workloads[0])
        assert self.io_period != mo_workloads

        change_workload = self.io_period.copy()
        change_workload.io_workloads[0].config.offset = 512
        assert self.io_period != change_workload


class TestEq(unittest.TestCase):
    """These are basic tests to get coverage to 100%"""

    def test_eq_access_range(self):
        left = _parm.AccessRangeBytes(1024, 400000)
        right = _parm.AccessRangeBytes(1024, 400000)
        assert left != 42
        assert left == right
        right.bytes_start = 42
        assert left != right
        assert left != 42

        left = _parm.AccessRangePercentages(10, 12)
        assert left != right
        assert left != 42
        right = _parm.AccessRangePercentages(10, 12)
        assert left == right
        right.percent_start = 1
        assert left != right

    def test_eq_cache_hit_config(self):
        left = _parm.CacheHitConfig(1024, 80, 30)
        right = _parm.CacheHitConfig(1024, 80, 30)
        assert left == right
        assert left != 42
        left.hit_area = 100000
        assert left != right

    def test_eq_io_rate(self):
        left = _parm.IoRateCurve()
        right = _parm.IoRateCurve()
        assert left != 42
        assert left == right
        left.curve = left.curve[1:]
        assert left != right

        left = _parm.IoRateMax()
        right = _parm.IoRateMax()
        assert left == right
        assert left != 42
        left.distribution = _enums.IoDistribution.uniform
        assert left != right

        left = _parm.IoRateFixed(4242)
        assert left != right
        right = _parm.IoRateFixed(4242)
        assert left == right
        right.io_rate = 42
        assert left != right

        left = _parm.IoRateSeries([42, 4242, 424242])
        assert left != right
        right = _parm.IoRateSeries([42, 4242, 424242])
        assert left == right
        right.io_rates = right.io_rates[1:]
        assert left != right

        left = _parm.IoRateVariable([(42, 1), (4242, 9)])
        assert left != right
        right = _parm.IoRateVariable([(42, 1), (4242, 9)])
        assert left == right
        right.io_rates = right.io_rates[1:]
        assert left != right

    def test_eq_dedup_config(self):
        left = _parm.DedupConfig(2, 256, 10)
        right = _parm.DedupConfig(2, 256, 10)
        assert left == right
        assert left != 42
        left.dedup_ratio = 42
        assert left != right

    def test_repr_force_error(self):
        left = _parm.ErrorInjectionConfig(42)
        assert left != 42
        right = _parm.ErrorInjectionConfig(42)
        assert left == right
        left.num_read = 3
        assert left != right

    def test_eq_seek_config(self):
        left = _parm.SeekConfig(10, stride=(10, 20))
        right = _parm.SeekConfig(10, stride=(10, 20))
        assert left != 42
        assert left == right
        left.seek_perc = 20
        assert left != right

    def test_eq_streams_config(self):
        left = _parm.StreamsConfig(10, 112)
        right = _parm.StreamsConfig(10, 112)
        assert left != 42
        assert left == right
        left.streams = 42
        assert left != right

    def test_eq_transfer_size(self):
        left = _parm.TransferSizeAligned(10, 112, 42)
        right = _parm.TransferSizeAligned(10, 112, 42)
        assert hash(left) == hash(right)
        assert left != 42
        assert left != 42
        assert left == right
        left.min_xfersize = 11
        assert left != right

        left = _parm.TransferSizeFixed()
        right = _parm.TransferSizeFixed()
        assert left != 42
        assert left == right
        left.xfersize = 42
        assert left != right

        left = _parm.TransferSizeVariable([(1024, 10), (4096, 90)])
        right = _parm.TransferSizeVariable([(1024, 10), (4096, 90)])
        assert left != 42
        assert left == right
        left.weighted_xfersizes = left.weighted_xfersizes[1:]
        assert left != right


class TestRepr(unittest.TestCase):
    """These are basic tests to get coverage to 100%"""
    def test_repr_access_range_config(self):
        assert repr(_parm.AccessRangeBytes(1024, 400000)) \
            == 'AccessRangeBytes<range=(1024,400000)>'

        assert repr(_parm.AccessRangePercentages(10, 12)) \
            == 'AccessRangePercentages<range=(10,12)>'

    def test_repr_cache_hit_config(self):
        assert repr(_parm.CacheHitConfig(1024, 80, 30)) \
            == 'CacheHitConfig<hit_area=1024, read_perc=80, write_perc=30>'

    def test_repr_dedup_config(self):
        assert repr(_parm.DedupConfig(2, 2048, 10)) \
            == 'DedupConfig<dedupratio=2,dedupunit=2k,dedupsets=10%>'

    def test_repr_transfer_size_config(self):
        assert repr(_parm.TransferSizeFixed()) \
            == 'TransferSizeFixed<xfersize=4k>'
        assert repr(_parm.TransferSizeAligned(1024, 4096, 512)) \
            == 'TransferSizeAligned<xfersize=(1k,4k,512)>'
        assert repr(_parm.TransferSizeVariable([(4096, 10), (512, 90)])) \
            == 'TransferSizeVariable<xfersize=(4k,10,0.5k,90)>'

    def test_repr_force_error(self):
        force = _parm.ErrorInjectionConfig(42)
        assert repr(force) \
            == 'ErrorInjectionConfig<force_error_after=(42,1)>'

    def test_repr_io_rate_config(self):
        curve = _parm.IoRateCurve()
        assert repr(curve) \
            == 'IoRateCurve<iorate=curve,curve=(10,50,70,80,90,100),distribution=exponential>'
        assert repr(_parm.IoRateFixed(4242)) \
            == 'IoRateFixed<iorate=4242,distribution=exponential>'
        assert repr(_parm.IoRateMax()) \
            == 'IoRateMax<iorate=max,distribution=exponential>'
        assert repr(_parm.IoRateSeries([42, 4242, 424242])) \
            == 'IoRateSeries<iorate=(42,4242,424242),distribution=exponential>'
        assert repr(_parm.IoRateSeries([42, 4242, 424242])) \
            == 'IoRateSeries<iorate=(42,4242,424242),distribution=exponential>'
        assert repr(_parm.IoRateVariable([(42, 1), (4242, 9)])) \
            == 'IoRateVariable<iorate=(42,1,4242,9),distribution=variable>'

    def test_repr_seek_config(self):
        assert repr(_parm.SeekConfig(10, stride=(10, 20))) \
            == 'SeekConfig<seekpct=10,stride=(10,20)>'

    def test_repr_streams_config(self):
        streams = _parm.StreamsConfig(10)
        assert repr(streams) == 'StreamsConfig<streams=10>'
        streams.size = 112
        assert repr(streams) \
            == 'StreamsConfig<streams=(10,112k)>'


class TestHash(unittest.TestCase):
    """These are basic tests to get coverage to 100%"""
    def test_hash_access_range_config(self):
        access_range_bytes = _parm.AccessRangeBytes(1024, 400000)
        access_range_percs = _parm.AccessRangePercentages(10, 12)
        assert {access_range_bytes, access_range_percs}

    def test_hash_seek_config(self):
        seek_config = _parm.SeekConfig(10, stride=(10, 20))
        assert {seek_config}


class TestInit(unittest.TestCase):
    def test_init_cache_hit_config(self):
        with pytest.raises(AssertionError):
            _parm.CacheHitConfig()

    def test_init_transfer_size_variable(self):
        with pytest.raises(AssertionError):
            _parm.TransferSizeVariable([(512, 99)])

        with pytest.raises(AssertionError):
            _parm.TransferSizeVariable([(512, 90), (1024, 3)])


class TestHdDef(unittest.TestCase):
    """Test the HdDef class."""

    def setUp(self):
        self.virt_client = models.ResourceClient.get_mock_object(
            overrides=dict(
                type=at2_enums.ClientType.virtual,
            )
        )

        self.client = models.ResourceClient.get_mock_object()

    def test_minimal(self):
        """Test the minimal allowed config."""
        hdef = _parm.HostDef('hd0', self.virt_client)
        expected_parm = (
            'hd=hd0,system={},jvms=1,vdbench=/opt/vdbench,shell=vdbench,user=root'
        ).format(self.virt_client.mip)
        assert repr(hdef) == 'HostDef<{}>'.format(expected_parm)
        assert hdef.get_parm_str() == expected_parm

    def test_complex(self):
        """Test every parameter."""
        hdef = _parm.HostDef(
            name='hd0',
            client_dm=self.client,
            jvms=6,
            vdbench='/opt/vdbench/vdbench.sh',
            shell='/bin/zsh',
            user='jake',
            mount='/mnt/whatever'
        )
        assert hdef.get_parm_str() == (
            'hd=hd0,system={},jvms=6,vdbench=/opt/vdbench/vdbench.sh,'
            'shell=/bin/zsh,user=jake,mount=/mnt/whatever'
        ).format(self.client.mip)

    def test_eq(self):
        left = _parm.HostDef('h0', client_dm=self.client)
        right = _parm.HostDef('h0', client_dm=self.client)
        assert left == right
        assert left != 42
        right.user = 'garrett'
        assert left != right

    def test_comment(self):
        hdef = _parm.HostDef('hd0', client_dm=self.virt_client, comment='sayin stuff')
        expected_parm = (
            'hd=hd0,system={},jvms=1,vdbench=/opt/vdbench,shell=vdbench,user=root'
        ).format(self.virt_client.mip)
        assert repr(hdef) == 'HostDef<{}>'.format(expected_parm)
        assert hdef.get_parm_str() == '# sayin stuff\n{}'.format(expected_parm)


class TestSdDef(unittest.TestCase):
    """Test the SdDef class."""

    def test_minimal(self):
        """Test the minimal allowed config."""
        config = _parm.StorageDefConfig()
        sdef = _parm.StorageDef(
            'sd0', config, '/dev/sdxx', 'hd0', journal_path=None)
        assert sdef.get_parm_str() == 'sd=sd0,host=hd0,lun=/dev/sdxx'

    def test_complex(self):
        """Test every parameter."""
        config = _parm.StorageDefConfig(
            access_range_config=_parm.AccessRangePercentages(10, 80),
            thread_count=8,
            hit_area=4000,
            offset=1024,
            alignment=4096,
            open_flags='o_direct',
        )

        sdef = _parm.StorageDef(
            name='sd0',
            sd_config=config,
            size=10000,
            lun='/dev/sdxx',
            hd_name='hd0',
            journal_path='/tmp/journal_me',
        )
        assert sdef.get_parm_str() == (
            'sd=sd0,host=hd0,lun=/dev/sdxx,journal=/tmp/journal_me,'
            'size=10000,range=(10,80),threads=8,hitarea=4000k'
            ',offset=1024,align=4096,openflags=o_direct'
        )


class TestWdDef(unittest.TestCase):
    """Test the WdConfig class."""

    def test_minimal(self):
        """Test the minimal allowed config."""
        config = _parm.WorkloadDefConfig()
        wdef = _parm.WorkloadDef('wd0', config, ['sd0'], None)
        assert wdef.get_parm_str() == 'wd=wd0,sd=sd0'

    def test_complex(self):
        """Test every parameter."""
        config = _parm.WorkloadDefConfig(
            io_rate=100,
            priority=1,
            read_perc=80,
            hit_perc_read=90,
            hit_perc_write=40,
            seek_config=_parm.SeekConfig(10, (1, 10)),
            skew_perc=60,
            transfer_size_config=_parm.TransferSizeFixed(2048),
        )
        wdef = _parm.WorkloadDef(
            name='wd0',
            wd_config=config,
            sd_names=['sd0'],
            comment=None
        )
        assert wdef.get_parm_str() == (
            'wd=wd0,sd=sd0,rdpct=80,iorate=100,priority=1,'
            'seekpct=10,stride=(1,10),rhpct=90,whpct=40,skew=60,'
            'xfersize=2k'
        )

    def test_multi(self):
        config = _parm.WorkloadDefConfig()
        wdef = _parm.WorkloadDef('wd0', config, ['sd0', 'sd1'], None)
        assert wdef.get_parm_str() == 'wd=wd0,sd=(sd0,sd1)'

    def test_zero(self):
        config = _parm.WorkloadDefConfig()
        with pytest.raises(AssertionError):
            _parm.WorkloadDef('wd0', config, [])


class TestRdDef(unittest.TestCase):
    """Test the RunDefinition class."""

    def test_minimal(self):
        """Test the minimal allowed config."""
        config = _parm.RunDefConfig(io_rate_config=_parm.IoRateFixed(100))
        rdef = _parm.RunDef('rd0', config, ['wd0', 'wd1'], None)
        expected = 'rd=rd0,wd=(wd0,wd1),iorate=100,distribution=exponential'
        assert rdef.get_parm_str() == expected

    def test_complex(self):
        """Test every parameter."""
        config = _parm.RunDefConfig(
            io_rate_config=_parm.IoRateCurve(
                curve=(10, 30, 50),
                distribution=_enums.IoDistribution.exponential,
            ),
            duration=600,
            max_data=999999,
            interval=2,
            warmup=60,
        )

        rdef = _parm.RunDef(
            name='rd0',
            rd_config=config,
            wd_names=['wd0'],
            comment=None
        )

        assert rdef.get_parm_str() == (
            'rd=rd0,wd=wd0,iorate=curve,curve=(10,30,50),distribution=exponential,'
            'elapsed=150,maxdata=999999,interval=2,warmup=60'
        )

    def test_zero(self):
        config = _parm.RunDefConfig(io_rate_config=_parm.IoRateFixed(100))
        with pytest.raises(AssertionError):
            _parm.RunDef('rd0', config, [], None)


class TestGdDef(unittest.TestCase):
    def setUp(self):
        self.clients = [
            models.ResourceClient.get_mock_object(overrides=dict(
                type=at2_enums.ClientType.physical
            ))
            for _ in range(2)
        ]

        self.hdefs = [
            _parm.HostDef('hd0', self.clients[0]),
            _parm.HostDef('hd1', self.clients[1])
        ]

        sconfig = _parm.StorageDefConfig(
            open_flags='o_direct',
        )

        self.sdefs = [
            _parm.StorageDef(
                'sd0', sconfig,
                '/dev/sdxx', 'hd0', None
            ),
            _parm.StorageDef(
                'sd1', sconfig,
                '/dev/sdyy', 'hd1', None
            ),
        ]

        self.wdefs = [
            _parm.WorkloadDef(
                'wd0', _parm.WorkloadDefConfig(read_perc=80), ['sd0'], None
            ),
            _parm.WorkloadDef(
                'wd1', _parm.WorkloadDefConfig(read_perc=100), ['sd1'], None
            ),
        ]

        self.rdefs = [
            _parm.RunDef(
                'rd0',
                _parm.RunDefConfig(
                    duration=240,
                    interval=1,
                    io_rate_config=_parm.IoRateMax()
                ),
                ['wd0'], None
            ),
            _parm.RunDef(
                'rd1',
                _parm.RunDefConfig(
                    io_rate_config=_parm.IoRateCurve((1, 2, 3, 100)),
                    duration=30,
                    interval=1,
                ),
                ['wd1'], None
            ),
            _parm.RunDef(
                'rd2',
                _parm.RunDefConfig(
                    io_rate_config=_parm.IoRateSeries((1000, 15000, 64000)),
                    duration=30,
                    interval=1,
                ),
                ['wd1'], None
            )
        ]

        self.config = _parm.GeneralDef(
            _parm.GeneralDefConfig(compression_ratio=3),
            reporting_config=_enums.ReportingConfig.none,
            host_defs=self.hdefs,
            run_defs=self.rdefs,
            workload_defs=self.wdefs,
            storage_defs=self.sdefs,
            port=12,
            data_errors=0
        )

    def test_complex(self):
        expect = '\n'.join((
            'monitor=/tmp/vdbench_kill',
            'port=12',
            'data_errors=0',
            'report=no_sd_detail',
            'compratio=3',
            'hd=hd0,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root'
            .format(self.clients[0].mip),
            'hd=hd1,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root'
            .format(self.clients[1].mip),
            'sd=sd0,host=hd0,lun=/dev/sdxx,openflags=o_direct',
            'sd=sd1,host=hd1,lun=/dev/sdyy,openflags=o_direct',
            'wd=wd0,sd=sd0,rdpct=80',
            'wd=wd1,sd=sd1,rdpct=100',
            'rd=rd0,wd=wd0,iorate=max,distribution=exponential,elapsed=240,interval=1',
            ('rd=rd1,wd=wd1,iorate=curve,curve=(1,2,3,100),distribution=exponential'
                ',elapsed=6,interval=1'),
            ('rd=rd2,wd=wd1,iorate=(1000,15000,64000),distribution=exponential'
             ',elapsed=10,interval=1')
        ))

        assert self.config.get_parm_str() == expect

    def test_special(self):
        config = _parm.GeneralDef(
            gd_config=_parm.GeneralDefConfig(compression_ratio=3),
            histogram=list(range(200, 1000, 200)),
            validation_config=_enums.ValidationConfig.none,
            reporting_config=_enums.ReportingConfig.none,
            host_defs=self.hdefs,
            storage_defs=self.sdefs,
            workload_defs=self.wdefs,
            run_defs=self.rdefs,
            data_errors=0
        )

        expect = '\n'.join((
            'monitor=/tmp/vdbench_kill',
            'data_errors=0',
            'report=no_sd_detail',
            'histogram=(default,200,400,600,800)',
            'compratio=3',
            'hd=hd0,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root'
            .format(self.clients[0].mip),
            'hd=hd1,system={},jvms=8,vdbench=/opt/vdbench,shell=vdbench,user=root'
            .format(self.clients[1].mip),
            'sd=sd0,host=hd0,lun=/dev/sdxx,openflags=o_direct',
            'sd=sd1,host=hd1,lun=/dev/sdyy,openflags=o_direct',
            'wd=wd0,sd=sd0,rdpct=80',
            'wd=wd1,sd=sd1,rdpct=100',
            'rd=rd0,wd=wd0,iorate=max,distribution=exponential,elapsed=240,interval=1',
            ('rd=rd1,wd=wd1,iorate=curve,curve=(1,2,3,100),distribution=exponential'
                ',elapsed=6,interval=1'),
            ('rd=rd2,wd=wd1,iorate=(1000,15000,64000),distribution=exponential'
             ',elapsed=10,interval=1')
        ))
        assert config.get_parm_str() == expect

    def test_histogram_bad(self):
        with pytest.raises(AssertionError):
            _parm.GeneralDef(
                _parm.GeneralDefConfig(compression_ratio=3),

                # Note: bad historgram
                histogram=list(range(100)),

                reporting_config=_enums.ReportingConfig.none,
                host_defs=self.hdefs,
                storage_defs=self.sdefs,
                workload_defs=self.wdefs,
                run_defs=self.rdefs,
                data_errors=0,
            )


class TestUtils(unittest.TestCase):
    def test_bytes_str(self):
        assert utils.bytes_str(None) is None
        assert utils.bytes_str(4096) == '4k'
        assert utils.bytes_str(512) == '0.5k'
        assert utils.bytes_str(500) == '500'
