from __future__ import absolute_import

import copy
import os
import random
import unittest

import mock
import pytest
from sf_models import _node_type

from camelot.models import at2_enums
from sf_api import data_size
from valence import (
    components,
    enums,
    models,
    vl_components,
)
from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _enums,
    _io_config,
    _parm,
    profiles,
)
from valence.vl_components.vdbench.profiles import csv_profiles


class BaseTestProfile(object):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_dm.api_endpoint = models.SfVersion('9.0')
        cls.client_dms = [models.ResourceClient.create(
            client_name='c{}'.format(i),
            mip='1.0.0.{}'.format(i),
            os=at2_enums.ClientOS.ubuntu_14_04 if i % 2 == 0 else at2_enums.ClientOS.centos_7
        ) for i in range(6)]
        cls.account_dm = models.Account.create(username='unittest-TAG-0', status='active', account_id=1)
        cls.disk_map = {
            61: '/dev/sdc',
            62: '/dev/sde',
            63: '/dev/sdb',
            64: '/dev/sda',
            65: '/dev/sdg',
            66: '/dev/sdf',
            67: '/dev/sdh',
            68: '/dev/sdi',
            69: '/dev/sdj',
            70: '/dev/sdk'
        }
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            account_id=1,
            volume_id=i,
            total_size=models.DataSize(i, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
            enable512e=False
        ) for i in cls.disk_map]
        random.shuffle(cls.volume_dms)
        cls.initiator_dms = [models.Initiator.get_mock_object() for _ in range(2)]

        # define the components
        cls.accounts_cmt = _account.MockAccounts('mock', [cls.account_dm])
        cls.volumes_cmts = [
            _volume.MockVolumes('mock-0', cls.cluster_dm, cls.volume_dms[0:1], cls.accounts_cmt),
            _volume.MockVolumes('mock-1', cls.cluster_dm, cls.volume_dms[1:3], cls.accounts_cmt),
            _volume.MockVolumes('mock-2', cls.cluster_dm, cls.volume_dms[3:6], cls.accounts_cmt),
            _volume.MockVolumes('mock-3', cls.cluster_dm, cls.volume_dms[6:], cls.accounts_cmt),
        ]
        cls.conn_cmts = [
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[0],
                volumes_cmts=cls.volumes_cmts[0:2],
            ),
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[1],
                volumes_cmts=cls.volumes_cmts[1:],
            )
        ]
        cls.conn_manager = vl_components.ManageConnections(cls.conn_cmts)

    @staticmethod
    def _get_workload_map(client_dms, workloads):
        workload_map = {dm.mip: [] for dm in client_dms}
        for workload in workloads:
            workload_map[workload.client_dm.mip].append(workload)

        return workload_map

    @staticmethod
    def _get_volume_dms(profile):
        volume_dms = []
        for conn_cmt in profile.connections_cmt.child_cmts:
            for volumes_cmt in conn_cmt.volumes_cmts:
                volume_dms.extend(volumes_cmt.partial_volume_dms)

        return volume_dms

    @staticmethod
    def _get_volume_map(extended_profile):
        volume_map = {}
        for profile in extended_profile._profiles:
            volume_map[profile] = []
            for conn_cmt in profile.connections_cmt.child_cmts:
                for volumes_cmt in conn_cmt.volumes_cmts:
                    volume_map[profile].extend(volumes_cmt.partial_volume_dms)

        return volume_map


class TestProfileHelpers(BaseTestProfile, unittest.TestCase):
    def setUp(self):
        super(TestProfileHelpers, self).setUp()
        self.base_workload_configs = [
            _io_config.IoWorkloadConfig(thread_count=5),
            _io_config.IoWorkloadConfig(thread_count=10),
            _io_config.IoWorkloadConfig(thread_count=20)
        ]
        self.xfersize_configs = [
            _parm.TransferSizeFixed(1024),
            _parm.TransferSizeFixed(4 * 1024),
            _parm.TransferSizeFixed(64 * 1024)
        ]

    def test_get_varied_workload_configs(self):
        workload_configs = profiles.get_varied_workload_configs(
            connections_cmt=self.conn_manager,
            base_workload_configs=self.base_workload_configs,
            transfer_size_configs=self.xfersize_configs
        )
        assert len(workload_configs) == 12
        assert all([config.thread_count == 5 for config in workload_configs[:4]])
        assert all([config.thread_count == 10 for config in workload_configs[4:8]])
        assert all([config.thread_count == 20 for config in workload_configs[8:]])
        for i, config in enumerate(workload_configs):
            assert config.transfer_size_config == self.xfersize_configs[
                i % len(self.xfersize_configs)]

    @mock.patch('valence.vl_components.vdbench.IoPeriod.MAX_WORKLOADS', 10)
    def test_get_varied_workload_configs_max(self):
        workload_configs = profiles.get_varied_workload_configs(
            connections_cmt=self.conn_manager,
            base_workload_configs=self.base_workload_configs,
            transfer_size_configs=self.xfersize_configs
        )
        assert len(workload_configs) == 10
        assert all([config.thread_count == 5 for config in workload_configs[:4]])
        assert all([config.thread_count == 10 for config in workload_configs[4:7]])
        assert all([config.thread_count == 20 for config in workload_configs[7:]])
        for i, config in enumerate(workload_configs):
            assert config.transfer_size_config == self.xfersize_configs[
                i % len(self.xfersize_configs)]

    def test_get_varied_workload_configs_xfersize_limit(self):
        volume_dm = models.Volume.create(
            volume_id=1,
            total_size=models.DataSize(2, enums.DataUnit.gibibytes)
        )
        volumes_cmt = _volume.MockVolumes('mock-small', self.cluster_dm, [volume_dm], self.accounts_cmt)
        conn_cmt = _connection.MockConnection(
            disk_map={1: '/dev/sdc'},
            client_dm=self.client_dms[0],
            volumes_cmts=[volumes_cmt],
        )
        workload_configs = profiles.get_varied_workload_configs(
            connections_cmt=conn_cmt,
            base_workload_configs=self.base_workload_configs,
            transfer_size_configs=[_parm.TransferSizeFixed(512 * 1024)]
        )
        assert max(workload_configs[0].transfer_size_config.xfersizes) < 512 * 1024

    def test_get_max_transfer_size(self):
        smallest_volume_size = 61 * enums.DataUnit.gibibytes.value[0]

        max_xfersize = profiles.get_max_transfer_size(self.conn_manager)
        assert max_xfersize == smallest_volume_size // (4000 + 8 + 256)

        max_xfersize = profiles.get_max_transfer_size(self.conn_manager, thread_count=3)
        assert max_xfersize == smallest_volume_size // (4000 + 3 + 256)

        max_xfersize = profiles.get_max_transfer_size(
            connections_cmt=self.conn_manager,
            io_rate_config=_parm.IoRateMax()
        )
        assert max_xfersize == smallest_volume_size // (2000 + 8 + 256)

    def test_get_workloads(self):
        workloads = profiles.get_workloads(self.conn_manager, self.base_workload_configs)
        workload_map = self._get_workload_map(self.client_dms[:2], workloads)
        assert len(workload_map) == 2

        # Ensure workloads reflect the specified connections
        client_workloads = workload_map[self.client_dms[0].mip]
        assert len(client_workloads) == 3
        assert set([w.volumes_cmt for w in client_workloads]) == set(self.conn_cmts[0].volumes_cmts)
        for workload in client_workloads:
            if workload.volumes_cmt is self.volumes_cmts[0]:
                assert workload.volume_index in range(1)
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert workload.volume_index in range(2)

        client_workloads = workload_map[self.client_dms[1].mip]
        assert len(client_workloads) == 9
        assert set([w.volumes_cmt for w in client_workloads]) == set(self.conn_cmts[1].volumes_cmts)
        for workload in client_workloads:
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert workload.volume_index in range(2)
            if workload.volumes_cmt is self.volumes_cmts[2]:
                assert workload.volume_index in range(3)
            if workload.volumes_cmt is self.volumes_cmts[3]:
                assert workload.volume_index in range(4)

        # Ensure workloads rotate through the provided configurations
        for i, workload in enumerate(workloads):
            assert workload.config == self.base_workload_configs[i % len(self.base_workload_configs)]

    def test_isolate_access_ranges(self):
        workloads = profiles.get_workloads(self.conn_manager, self.base_workload_configs)
        profiles.isolate_access_ranges(workloads)
        for workload in workloads:
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert (workload.config.access_range_config.percent_end
                        - workload.config.access_range_config.percent_start) == 50
            else:
                assert (workload.config.access_range_config.percent_end
                        - workload.config.access_range_config.percent_start) == 100


class TestSimpleVdbProfile(BaseTestProfile, unittest.TestCase):

    def test_get_workloads(self):
        profile = profiles.SimpleVdbProfile(self.conn_manager)
        workloads = profile.get_workloads()
        workload_map = self._get_workload_map(self.client_dms[:2], workloads)
        assert len(workload_map) == 2

        # Ensure workloads reflect the specified connections
        client_workloads = workload_map[self.client_dms[0].mip]
        assert len(client_workloads) == 3
        assert set([w.volumes_cmt for w in client_workloads]) == set(self.conn_cmts[0].volumes_cmts)
        for workload in client_workloads:
            if workload.volumes_cmt is self.volumes_cmts[0]:
                assert workload.volume_index in range(1)
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert workload.volume_index in range(2)

        client_workloads = workload_map[self.client_dms[1].mip]
        assert len(client_workloads) == 9
        assert set([w.volumes_cmt for w in client_workloads]) == set(self.conn_cmts[1].volumes_cmts)
        for workload in client_workloads:
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert workload.volume_index in range(2)
            if workload.volumes_cmt is self.volumes_cmts[2]:
                assert workload.volume_index in range(3)
            if workload.volumes_cmt is self.volumes_cmts[3]:
                assert workload.volume_index in range(4)

        # Ensure workload configs use the defaults
        for workload in workloads:
            assert workload.config.read_perc == 80
            assert workload.config.seek_config is None
            assert workload.config.transfer_size_config is None

        # Ensure workload access ranges are isolated
        for workload in workloads:
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert (workload.config.access_range_config.percent_end
                        - workload.config.access_range_config.percent_start) == 50
            else:
                assert (workload.config.access_range_config.percent_end
                        - workload.config.access_range_config.percent_start) == 100

        # Ensure workload configs match the initialized configuration
        seek_config = _parm.SeekConfig(seek_perc=23)
        transfer_size_config = _parm.TransferSizeAligned(
            min_xfersize=4096,
            max_xfersize=2 * 4096,
            alignment=4096
        )
        profile = profiles.SimpleVdbProfile(
            connections_cmt=self.conn_manager,
            read_perc=42,
            seek_config=seek_config,
            transfer_size_config=transfer_size_config
        )
        workloads = profile.get_workloads()
        for workload in workloads:
            assert workload.config.read_perc == 42
            assert workload.config.seek_config == seek_config
            assert workload.config.transfer_size_config == transfer_size_config

    def test_get_workloads_custom(self):
        profile = profiles.SimpleVdbProfile(self.conn_manager)
        custom_config = profiles.StandardVdbProfile.DEFAULT_WORKLOAD_CONFIGS[-1]
        profile.workload_config = custom_config
        workloads = profile.get_workloads()
        for workload in workloads:
            assert workload.config.read_perc == custom_config.read_perc
            assert workload.config.seek_config == custom_config.seek_config

    def test_get_periods(self):
        profile = profiles.SimpleVdbProfile(self.conn_manager)
        periods = profile.get_periods()
        for period in periods:
            assert period.config.io_rate_config == _parm.IoRateMax()
            assert period.config.interval == 1
            assert period.config.duration == 30

        # Ensure period configs match the initialized configuration
        io_rate_config = _parm.IoRateCurve(curve=(20, 30, 60))
        profile = profiles.SimpleVdbProfile(self.conn_manager, io_rate_config=io_rate_config)
        periods = profile.get_periods()
        for period in periods:
            assert period.config.io_rate_config == io_rate_config

    def test_get_periods_custom(self):
        profile = profiles.SimpleVdbProfile(self.conn_manager)
        custom_config = _io_config.IoPeriodConfig(
            io_rate_config=_parm.IoRateMax(),
            interval=30,
            duration=24 * 60 * 60
        )
        profile.period_config = custom_config
        periods = profile.get_periods()
        for period in periods:
            assert period.config == custom_config

    @mock.patch('valence.vl_components.vdbench.SimpleVdbProfile.get_periods')
    def test_get_session(self, get_periods):
        profile = profiles.SimpleVdbProfile(self.conn_manager)

        # ... get the periods
        workloads = profile.get_workloads()
        periods = [_io_config.IoPeriod(
            io_period_config=profile.period_config,
            io_workloads=workloads
        )]
        get_periods.return_value = periods

        session = profile.get_session()
        assert session.config.compression_ratio is None
        assert session.config.dedup_config is None
        assert session.config.pattern_path is None
        assert not session.standalone
        assert len(session.io_periods) == len(periods)
        expected_periods = copy.copy(periods)
        for period in session.io_periods:
            expected_periods.remove(period)

        # Ensure standalone sessions cannot be generated by profiles for multiple clients
        with pytest.raises(AssertionError):
            session = profile.get_session(standalone=True)

        # Ensure standalone session can be generated by profiles for single clients
        profile = profiles.SimpleVdbProfile(self.conn_cmts[0])

        # ... get the periods
        workloads = profile.get_workloads()
        periods = [_io_config.IoPeriod(
            io_period_config=profile.period_config,
            io_workloads=workloads
        )]
        get_periods.return_value = periods

        session = profile.get_session(standalone=True)
        assert session.standalone

    def test_get_session_custom(self):
        profile = profiles.SimpleVdbProfile(self.conn_manager)
        profile.session_config.compression_ratio = 10
        session = profile.get_session()
        assert session.config.compression_ratio == 10

    @mock.patch('valence.vl_components.vdbench.SimpleVdbProfile.get_session')
    def test_get_environment(self, get_session):
        profile = profiles.SimpleVdbProfile(self.conn_manager)

        # ... get the session
        periods = profile.get_periods()
        session = _io_config.IoSession(profile.session_config, periods)
        get_session.return_value = session

        environment = profile.get_environment()
        assert environment.connections_cmt is profile.connections_cmt
        assert environment.io_session == session
        assert environment.master_client_dm is profile.connections_cmt.client_dms[0]

        master_client_dm = models.ResourceClient.get_mock_object()
        environment = profile.get_environment(master_client_dm)
        assert environment.master_client_dm is master_client_dm

        assert profile.connections_cmt.state == components.States.ready
        environment.apply()
        assert profile.connections_cmt.state == components.States.applied


class TestStandardVdbProfile(BaseTestProfile, unittest.TestCase):

    def test_get_workloads(self):
        profile = profiles.StandardVdbProfile(self.conn_manager)
        workloads = profile.get_workloads()
        workload_map = self._get_workload_map(self.client_dms[:2], workloads)
        assert len(workload_map) == 2

        # Ensure workloads reflect the specified connections
        client_workloads = workload_map[self.client_dms[0].mip]
        assert len(client_workloads) == 3
        assert set([w.volumes_cmt for w in client_workloads]) == set(self.conn_cmts[0].volumes_cmts)
        for workload in client_workloads:
            if workload.volumes_cmt is self.volumes_cmts[0]:
                assert workload.volume_index in range(1)
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert workload.volume_index in range(2)

        client_workloads = workload_map[self.client_dms[1].mip]
        assert len(client_workloads) == 9
        assert set([w.volumes_cmt for w in client_workloads]) == set(self.conn_cmts[1].volumes_cmts)
        for workload in client_workloads:
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert workload.volume_index in range(2)
            if workload.volumes_cmt is self.volumes_cmts[2]:
                assert workload.volume_index in range(3)
            if workload.volumes_cmt is self.volumes_cmts[3]:
                assert workload.volume_index in range(4)

        # Ensure workload configs use the standard defaults
        read_percs = [c.read_perc for c in profiles.StandardVdbProfile.DEFAULT_WORKLOAD_CONFIGS]
        seek_configs = [c.seek_config for c in profiles.StandardVdbProfile.DEFAULT_WORKLOAD_CONFIGS]
        for workload in workloads:
            assert workload.config.read_perc in read_percs
            assert workload.config.seek_config in seek_configs
            assert workload.config.transfer_size_config in \
                profiles.StandardVdbProfile.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED

        # Ensure workload access ranges are isolated
        for workload in workloads:
            if workload.volumes_cmt is self.volumes_cmts[1]:
                assert (workload.config.access_range_config.percent_end
                        - workload.config.access_range_config.percent_start) == 50
            else:
                assert (workload.config.access_range_config.percent_end
                        - workload.config.access_range_config.percent_start) == 100

    def test_get_workloads_custom(self):
        profile = profiles.StandardVdbProfile(self.conn_manager)
        custom_config = profiles.StandardVdbProfile.DEFAULT_WORKLOAD_CONFIGS[-1]
        profile.workload_configs = [custom_config]
        workloads = profile.get_workloads()
        for workload in workloads:
            assert workload.config.read_perc == custom_config.read_perc
            assert workload.config.seek_config == custom_config.seek_config

    @mock.patch('valence.vl_components.vdbench.StandardVdbProfile.get_workloads')
    def test_get_periods(self, get_workloads):
        profile = profiles.StandardVdbProfile(self.conn_manager)

        # ... get the workloads
        varied_configs = profiles.get_varied_workload_configs(
            connections_cmt=profile.connections_cmt,
            base_workload_configs=profile.workload_configs,
            transfer_size_configs=profile.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED
        )
        workloads = profiles.get_workloads(
            connections_cmt=profile.connections_cmt,
            workload_configs=varied_configs
        )
        profiles.isolate_access_ranges(workloads)
        get_workloads.return_value = workloads

        periods = profile.get_periods()
        for i, period in enumerate(periods):
            assert period.config == profiles.StandardVdbProfile.DEFAULT_PERIOD_CONFIGS[i]
            assert len(period.io_workloads) == len(workloads)
            expected_workloads = copy.copy(workloads)
            for workload in period.io_workloads:
                expected_workloads.remove(workload)

    def test_get_periods_custom(self):
        profile = profiles.StandardVdbProfile(self.conn_manager)
        custom_config = profiles.StandardVdbProfile.DEFAULT_PERIOD_CONFIGS[-1]
        profile.period_configs = [custom_config]
        periods = profile.get_periods()
        for period in periods:
            assert period.config == custom_config

    @mock.patch('valence.vl_components.vdbench.StandardVdbProfile.get_periods')
    def test_get_session(self, get_periods):
        profile = profiles.StandardVdbProfile(self.conn_manager)

        # ... get the periods
        workloads = profile.get_workloads()
        periods = [_io_config.IoPeriod(config, workloads) for config in profile.period_configs]
        get_periods.return_value = periods

        session = profile.get_session()
        assert session.config.compression_ratio == \
            profiles.StandardVdbProfile.DEFAULT_SESSION_CONFIG.compression_ratio
        assert session.config.pattern_path == \
            profiles.StandardVdbProfile.DEFAULT_SESSION_CONFIG.pattern_path
        assert session.config.dedup_config == profiles.StandardVdbProfile.DEDUP_2X
        assert not session.standalone
        assert len(session.io_periods) == len(periods)
        expected_periods = copy.copy(periods)
        for period in session.io_periods:
            expected_periods.remove(period)

        # Ensure standalone sessions cannot be generated by profiles for multiple clients
        with pytest.raises(AssertionError):
            session = profile.get_session(standalone=True)

        # Ensure standalone session can be generated by profiles for single clients
        profile = profiles.StandardVdbProfile(self.conn_cmts[0])

        # ... get the periods
        workloads = profile.get_workloads()
        periods = [_io_config.IoPeriod(config, workloads) for config in profile.period_configs]
        get_periods.return_value = periods

        session = profile.get_session(standalone=True)
        assert session.standalone

    def test_get_session_custom(self):
        profile = profiles.StandardVdbProfile(self.conn_manager)
        profile.session_config.compression_ratio = 10
        session = profile.get_session()
        assert session.config.compression_ratio == 10

    @mock.patch('valence.vl_components.vdbench.StandardVdbProfile.get_session')
    def test_get_environment(self, get_session):
        profile = profiles.StandardVdbProfile(self.conn_manager)

        # ... get the session
        periods = profile.get_periods()
        session = _io_config.IoSession(profile.session_config, periods)
        get_session.return_value = session

        environment = profile.get_environment()
        assert environment.connections_cmt is profile.connections_cmt
        assert environment.io_session == session
        assert environment.master_client_dm is profile.connections_cmt.client_dms[0]

        master_client_dm = models.ResourceClient.get_mock_object()
        environment = profile.get_environment(master_client_dm)
        assert environment.master_client_dm is master_client_dm

        assert profile.connections_cmt.state == components.States.ready
        environment.apply()
        assert profile.connections_cmt.state == components.States.applied

    @mock.patch('valence.vl_components._account.CreateAccounts', vl_components.CreateAccounts.mock)
    @mock.patch('valence.vl_components._initiator.CreateIscsiInitiators',
                vl_components.CreateIscsiInitiators.mock)
    @mock.patch('valence.vl_components._volume_access_group.CreateVolumeAccessGroup',
                vl_components.CreateVolumeAccessGroup.mock)
    @mock.patch('valence.vl_components._connection.CreateIscsiVagConnection',
                vl_components.CreateIscsiVagConnection.mock)
    def test_from_volumes(self):
        profile = profiles.StandardVdbProfile.from_volumes(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            volumes_cmts=self.volumes_cmts
        )

        # Ensure all volumes are connected
        volume_dms = self._get_volume_dms(profile)
        assert len(volume_dms) == len(self.volume_dms)
        assert set(dm.volume_id for dm in volume_dms) == set(dm.volume_id for dm in self.volume_dms)

        # Ensure each volume is being connected to only one client
        volume_ids = set()
        for conn_cmt in profile.connections_cmt.child_cmts:
            conn_volume_ids = set()
            for volumes_cmt in conn_cmt.volumes_cmts:
                conn_volume_ids.update(dm.volume_id for dm in volumes_cmt.partial_volume_dms)
            assert not conn_volume_ids.intersection(volume_ids)
            volume_ids.update(conn_volume_ids)

        # Ensure the generated connections component manages the provided volumes components
        assert all([cmt.state == components.States.ready for cmt in self.volumes_cmts])
        profile.connections_cmt.apply()
        assert all([cmt.state == components.States.applied for cmt in self.volumes_cmts])

        # Ensure an error is raised if connections_per_volume is greater than the number of clients
        with pytest.raises(ValueError):
            profile = profiles.StandardVdbProfile.from_volumes(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms[:2],
                volumes_cmts=self.volumes_cmts,
                connections_per_volume=len(self.client_dms[:2]) + 1
            )

        # Ensure an error is raised if connections_per_volume is too low for the provided volumes to
        # form a connection with each client
        with pytest.raises(ValueError):
            profile = profiles.StandardVdbProfile.from_volumes(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms[:2],
                volumes_cmts=[self.volumes_cmts[0]],
                connections_per_volume=1
            )

    def test_from_clients(self):
        # Aligned profile
        profile = profiles.StandardVdbProfile.from_clients(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            volume_count=100
        )
        volume_dms = self._get_volume_dms(profile)
        assert len(volume_dms) == 100
        percent_512e_enabled = float(sum([dm.enable512e for dm in volume_dms])) / len(volume_dms)
        assert 0.4 <= percent_512e_enabled <= 0.5
        percent_qos_reduced = float(sum(
            [dm.qos == profiles.StandardVdbProfile.REDUCED_QOS for dm in volume_dms]
        )) / len(volume_dms)
        assert 0 < percent_qos_reduced <= 0.4

        # Unaligned profile
        profile = profiles.StandardVdbProfile.from_clients(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            volume_count=100,
            aligned=False
        )
        volume_dms = self._get_volume_dms(profile)
        assert len(volume_dms) == 100
        assert all(dm.enable512e for dm in volume_dms)
        percent_qos_reduced = float(sum(
            [dm.qos == profiles.StandardVdbProfile.REDUCED_QOS for dm in volume_dms]
        )) / len(volume_dms)
        assert 0 < percent_qos_reduced <= 0.4

    def test_override_configs(self):
        profile = profiles.StandardVdbProfile(self.conn_manager)
        profile.override_configs(
            transfer_size_config=profiles.StandardVdbProfile.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED[-1]
        )
        workloads = profile.get_workloads()
        for workload in workloads:
            assert workload.config.io_rate is None
            assert workload.config.transfer_size_config == \
                profiles.StandardVdbProfile.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED[-1]

        profile = profiles.StandardVdbProfile(self.conn_manager)
        profile.override_configs(
            io_rate=42,
            read_perc=3,
            seek_config=profile.SEEK_SEQUENTIAL_IO,
            thread_count=7
        )
        workloads = profile.get_workloads()
        for workload in workloads:
            assert workload.config.read_perc == 3
            assert workload.config.seek_config == profile.SEEK_SEQUENTIAL_IO
        periods = profile.get_periods()
        for period in periods:
            assert isinstance(period.config.io_rate_config, _parm.IoRateFixed)
            assert period.config.io_rate_config.io_rate == 42

    def test_sanitize_configs(self):
        profile = profiles.StandardVdbProfile(self.conn_manager)
        for period in profile.get_periods():
            assert isinstance(period.config.io_rate_config, _parm.IoRateMax)

        # Ensure nothing happens if no virtual nodes are present
        node_dms = [
            models.ResourceNode.create(model=_node_type.NodeType(_node_type.NodeTypeEnum.SF3010.value)),
            models.ResourceNode.create(model=_node_type.NodeType(_node_type.NodeTypeEnum.SF3010.value)),
            models.ResourceNode.create(model=_node_type.NodeType(_node_type.NodeTypeEnum.SF3010.value))
        ]
        profile.sanitize_configs(node_dms=node_dms, raise_on_failure=True)
        for period in profile.get_periods():
            assert isinstance(period.config.io_rate_config, _parm.IoRateMax)

        # Ensure error is raised if configuration is invalid for virtual nodes
        node_dms = [
            models.ResourceNode.create(model=_node_type.NodeType(_node_type.NodeTypeEnum.SF3010.value)),
            models.ResourceNode.create(model=_node_type.NodeType(_node_type.NodeTypeEnum.SFVIRT.value)),
            models.ResourceNode.create(model=_node_type.NodeType(_node_type.NodeTypeEnum.SF3010.value))
        ]
        with pytest.raises(ValueError):
            profile.sanitize_configs(node_dms=node_dms, raise_on_failure=True)

        # Ensure io-rate is fixed for virtual nodes
        profile.sanitize_configs(node_dms=node_dms)
        for period in profile.get_periods():
            assert isinstance(period.config.io_rate_config, _parm.IoRateFixed)
            assert period.config.io_rate_config.io_rate == 100


class TestStandardVdbExtendedProfile(BaseTestProfile, unittest.TestCase):

    def test_percent_aligned(self):
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2]
        )

        # Ensure all volumes are connected
        volume_dms_aligned = []
        volume_dms_unaligned = []
        volume_map = self._get_volume_map(extended_profile)
        for profile, volume_dms in volume_map.items():
            if profile.aligned:
                volume_dms_aligned.extend(volume_dms)
            else:
                volume_dms_unaligned.extend(volume_dms)

        percent_aligned = (
            float(len(volume_dms_aligned)) / (len(volume_dms_aligned) + len(volume_dms_unaligned))
        )
        assert 0.85 - 0.02 <= percent_aligned <= 0.85 + 0.02

    def test_max_slave_count(self):
        client_dms = [
            models.ResourceClient.create(client_name='c{}'.format(i), mip='1.0.0.{}'.format(i),
                                         os='{}'.format('Ubuntu 14.04' if i % 2 == 0 else 'CentOS 7.0'))
            for i in
            range(profiles.StandardVdbExtendedProfile.MAX_SLAVE_COUNT + 1)
        ]

        # Ensure the default slaves-per-master is based on the max-slave-count
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=client_dms
        )
        assert len(extended_profile._profiles) == 2

        client_dms.append(models.ResourceClient.create(client_name='ca', mip='1.0.1.0',
                                                       os='{}'.format('Ubuntu 14.04')))
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=client_dms
        )
        assert len(extended_profile._profiles) == 3

    def test_slaves_per_master_configuration(self):
        # Ensure slaves-per-master is configurable
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms,
            slaves_per_master=1
        )
        assert len(extended_profile._profiles) == len(self.client_dms)

        # Ensure slaves-per-master=0 generates one profile for every client
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms,
            slaves_per_master=0
        )
        assert len(extended_profile._profiles) == len(self.client_dms)

        # Ensure slaves-per-master=0 generates standalone sessions
        sessions = extended_profile.get_sessions()
        assert all([session.standalone for session in sessions])

    @mock.patch('valence.vl_components._account.CreateAccounts', vl_components.CreateAccounts.mock)
    @mock.patch('valence.vl_components._volume.CreateMultipleVolumes',
                vl_components.CreateMultipleVolumes.mock)
    @mock.patch('valence.vl_components._initiator.CreateIscsiInitiators',
                vl_components.CreateIscsiInitiators.mock)
    @mock.patch('valence.vl_components._volume_access_group.CreateVolumeAccessGroup',
                vl_components.CreateVolumeAccessGroup.mock)
    @mock.patch('valence.vl_components._connection.CreateIscsiVagConnection',
                vl_components.CreateIscsiVagConnection.mock)
    def test_get_environment(self):
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            volume_count=len(self.volume_dms)
        )

        environment = extended_profile.get_environment()
        connection_cmts = [profile.connections_cmt for profile in extended_profile._profiles]
        assert all([cmt.state == components.States.ready for cmt in connection_cmts])
        environment.apply()
        assert all([cmt.state == components.States.applied for cmt in connection_cmts])

    def test_override_configs(self):
        extended_profile = profiles.StandardVdbExtendedProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2]
        )

        extended_profile.override_configs(
            transfer_size_config=profiles.StandardVdbProfile.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED[-1]
        )
        for profile in extended_profile._profiles:
            workloads = profile.get_workloads()
            for workload in workloads:
                assert workload.config.transfer_size_config == \
                    profiles.StandardVdbProfile.DEFAULT_XFER_SIZE_CONFIGS_ALIGNED[-1]

        # Ensure unaligned transfer size overrides don't work on aligned profiles
        extended_profile.override_configs(
            transfer_size_config=profiles.StandardVdbProfile.DEFAULT_XFER_SIZE_CONFIGS_UNALIGNED[0]
        )
        for profile in extended_profile._profiles:
            if profile.connections_cmt.child_cmts[0].volumes_cmts[0].partial_volume_dms[0].enable512e:
                workloads = profile.get_workloads()
            else:
                with pytest.raises(ValueError):
                    workloads = profile.get_workloads()


class BaseTestCsvProfile(BaseTestProfile):
    @classmethod
    def setUpClass(cls):
        super(BaseTestCsvProfile, cls).setUpClass()
        cls.cwd = os.path.dirname(os.path.realpath(os.path.expanduser(__file__)))
        cls.csv_dir = os.path.join(cls.cwd, '../profiles/csv_profiles')
        cls.csv_1_client = os.path.join(cls.csv_dir, '1_client/aligned_io_profile.csv')
        cls.csv_2_clients = os.path.join(cls.csv_dir, '2_clients/aligned_io_profile.csv')

        cls.mock_volume_dm = models.Volume.create(
            total_size=data_size.DataSize(1, data_size.DataUnit.gibibytes)
        )

        # Save generate_volume_dm method for overriding global mock when necessary
        cls.generate_volume_dm = profiles.csv_profiles.generate_volume_dm.__call__


class TestCsvProfileHelpers(BaseTestCsvProfile, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCsvProfileHelpers, cls).setUpClass()
        cls.conn_def = {
            csv_profiles.SD_TAG: '1',
            csv_profiles.GROUP: 'Sequential workloads with small and large IO sizes',
            csv_profiles.VOLUME_NAME: 'v1',
            csv_profiles.VOLUME_SIZE: '',
            csv_profiles.BLOCK_SIZE: '4k',
            csv_profiles.MIN_IOPS: '',
            csv_profiles.MAX_IOPS: '',
            csv_profiles.BURST_IOPS: '',
            csv_profiles.XFERSIZE: '4k',
            csv_profiles.SEEKPCT: 'seq',
            csv_profiles.READPCT: '0',
            csv_profiles.RANGE: '',
            csv_profiles.ALIGN: '',
            csv_profiles.THREADS: '8',
            csv_profiles.CLIENT_INDEX: '0'
        }

    def test_get_client_dir_map(self):
        client_dir_map = csv_profiles.get_client_dir_map(self.csv_dir)
        assert set(client_dir_map.keys()) == {1, 2, 3, 4}
        for csv_client_dir in client_dir_map.values():
            assert os.path.exists(os.path.join(self.csv_dir, csv_client_dir))

    def test_get_connection_definitions(self):
        conn_defs = csv_profiles.get_connection_definitions(self.csv_1_client)
        assert len(conn_defs) == 52
        for conn_def in conn_defs:
            assert csv_profiles.SD_TAG in conn_def
            assert csv_profiles.GROUP in conn_def
            assert csv_profiles.VOLUME_NAME in conn_def
            assert csv_profiles.VOLUME_SIZE in conn_def
            assert csv_profiles.BLOCK_SIZE in conn_def
            assert csv_profiles.MIN_IOPS in conn_def
            assert csv_profiles.MAX_IOPS in conn_def
            assert csv_profiles.BURST_IOPS in conn_def
            assert csv_profiles.XFERSIZE in conn_def
            assert csv_profiles.SEEKPCT in conn_def
            assert csv_profiles.READPCT in conn_def
            assert csv_profiles.RANGE in conn_def
            assert csv_profiles.ALIGN in conn_def
            assert csv_profiles.THREADS in conn_def
            assert csv_profiles.CLIENT_INDEX in conn_def

    def test_generate_volume_dm(self):
        conn_def = copy.copy(self.conn_def)
        volume_dm = csv_profiles.generate_volume_dm(conn_def)
        assert volume_dm.total_size == data_size.DataSize(128, data_size.DataUnit.gibibytes)
        assert not volume_dm.enable512e
        assert volume_dm.qos.min_iops == 100
        assert volume_dm.qos.max_iops == 15000
        assert volume_dm.qos.burst_iops == 15000

        conn_def[csv_profiles.VOLUME_SIZE] = 42
        conn_def[csv_profiles.BLOCK_SIZE] = '512e'
        conn_def[csv_profiles.MIN_IOPS] = 500
        conn_def[csv_profiles.MAX_IOPS] = 1000
        conn_def[csv_profiles.BURST_IOPS] = 1500
        volume_dm = csv_profiles.generate_volume_dm(conn_def)
        assert volume_dm.total_size == data_size.DataSize(
            size=conn_def[csv_profiles.VOLUME_SIZE],
            unit=data_size.DataUnit.gibibytes
        )
        assert volume_dm.enable512e
        assert volume_dm.qos.min_iops == conn_def[csv_profiles.MIN_IOPS]
        assert volume_dm.qos.max_iops == conn_def[csv_profiles.MAX_IOPS]
        assert volume_dm.qos.burst_iops == conn_def[csv_profiles.BURST_IOPS]

    def test_get_workload_config(self):
        conn_def = copy.copy(self.conn_def)
        workload_config = csv_profiles.get_workload_config(conn_def)
        assert workload_config.read_perc == 0
        assert workload_config.seek_config.seek_perc == 0
        assert workload_config.transfer_size_config.xfersize == 4 * 1024

        conn_def[csv_profiles.READPCT] = 60
        conn_def[csv_profiles.SEEKPCT] = 'random'
        conn_def[csv_profiles.XFERSIZE] = '(4k,10,32k,10,128k,20,256k,30,512k,30)'
        workload_config = csv_profiles.get_workload_config(conn_def)
        assert workload_config.read_perc == 60
        assert workload_config.seek_config.seek_perc == 100
        assert workload_config.transfer_size_config.weighted_xfersizes == (
            (4 * 1024, 10),
            (32 * 1024, 10),
            (128 * 1024, 20),
            (256 * 1024, 30),
            (512 * 1024, 30)
        )

    def test_get_transfer_size_config(self):
        transfer_size_config = csv_profiles.get_transfer_size_config('32k')
        assert isinstance(transfer_size_config, _parm.TransferSizeFixed)
        assert transfer_size_config.xfersize == 32 * 1024

        transfer_size_config = csv_profiles.get_transfer_size_config('512')
        assert isinstance(transfer_size_config, _parm.TransferSizeFixed)
        assert transfer_size_config.xfersize == 512

        transfer_size_config = csv_profiles.get_transfer_size_config('(0.5k,10,32k,20,128k,70)')
        assert isinstance(transfer_size_config, _parm.TransferSizeVariable)
        assert transfer_size_config.weighted_xfersizes == (
            (512, 10),
            (32 * 1024, 20),
            (128 * 1024, 70)
        )

    def test_get_access_range_config(self):
        access_range_config = csv_profiles.get_access_range_config('(1,40)')
        assert access_range_config.percent_start == 1
        assert access_range_config.percent_end == 40

    def test_get_seek_config(self):
        seek_config = csv_profiles.get_seek_config('seq')
        assert seek_config.seek_perc == 0

        seek_config = csv_profiles.get_seek_config('random')
        assert seek_config.seek_perc == 100

        seek_config = csv_profiles.get_seek_config('eof')
        assert seek_config.seek_perc == -1

        seek_config = csv_profiles.get_seek_config('42')
        assert seek_config.seek_perc == 42


class TestCsvVdbProfile(BaseTestCsvProfile, unittest.TestCase):
    def test_init(self):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:1],
            csv_file_path=self.csv_1_client
        )

        # Ensure an exception is raised if more than the expected number of clients are provided
        with pytest.raises(ValueError):
            profile = profiles.CsvVdbProfile(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms[:2],
                csv_file_path=self.csv_1_client
            )

        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )

    def test_get_workloads(self):
        # Override generate_volume_dm() mock
        profiles.csv_profiles.generate_volume_dm = self.generate_volume_dm

        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )

        workloads = profile.get_workloads()
        workload_map = self._get_workload_map(self.client_dms[:2], workloads)
        assert len(workload_map) == len(self.client_dms[:2])

        # Ensure workloads reflect the CSV file
        assert len(workload_map[self.client_dms[0].mip]) == 26
        assert len(workload_map[self.client_dms[1].mip]) == 26

        workloads = profile.get_workloads()
        assert len(workloads) == 52
        assert len([w for w in workloads if w.config.seek_config.seek_perc == 100]) == 36
        assert len([w for w in workloads if w.config.read_perc == 0]) == 16
        assert len([w for w in workloads if w.config.read_perc == 35]) == 8
        assert len([w for w in workloads if w.config.read_perc == 40]) == 12
        assert len([w for w in workloads if w.config.read_perc == 60]) == 8
        assert len([w for w in workloads if w.config.read_perc == 100]) == 8

        fixed_transfer_sizes = [
            w.config.transfer_size_config for w in workloads
            if isinstance(w.config.transfer_size_config, _parm.TransferSizeFixed)
        ]
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 4 * 1024]) == 7
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 16 * 1024]) == 7
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 64 * 1024]) == 7
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 128 * 1024]) == 5
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 256 * 1024]) == 5
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 512 * 1024]) == 5
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 1024 * 1024]) == 5
        assert len([config for config in fixed_transfer_sizes if config.xfersize == 2048 * 1024]) == 5

        varied_transfer_sizes = [
            w.config.transfer_size_config for w in workloads
            if isinstance(w.config.transfer_size_config, _parm.TransferSizeVariable)
        ]
        assert len(varied_transfer_sizes) == 2

        volume_dms = self._get_volume_dms(profile)
        assert len(volume_dms) == 52
        percent_512e_enabled = float(sum([dm.enable512e for dm in volume_dms])) / len(volume_dms)
        assert percent_512e_enabled == 0.5

    def test_get_workloads_custom(self):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )
        custom_config = profiles.StandardVdbProfile.DEFAULT_WORKLOAD_CONFIGS[-1]
        custom_config.transfer_size_config = _parm.TransferSizeFixed()
        custom_workload_configs = [(custom_config, c, v) for w, c, v in profile.workload_configs]
        profile.workload_configs = custom_workload_configs
        workloads = profile.get_workloads()
        for workload in workloads:
            assert workload.config.read_perc == custom_config.read_perc
            assert workload.config.seek_config == custom_config.seek_config
            assert workload.config.transfer_size_config == custom_config.transfer_size_config

    @mock.patch('valence.vl_components.vdbench.CsvVdbProfile.get_workloads')
    def test_get_periods(self, get_workloads):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )

        # ... get the workloads
        workloads = []
        for workload_config, client_dm, volume_tag in profile.workload_configs:
            workloads.append(_io_config.IoWorkload(
                io_workload_config=workload_config,
                client_dm=client_dm,
                volumes_cmt=profile.volumes_cmt_map[volume_tag],
                volume_index=0
            ))
        get_workloads.return_value = workloads

        periods = profile.get_periods()
        for i, period in enumerate(periods):
            assert period.config == profiles.StandardVdbProfile.DEFAULT_PERIOD_CONFIGS[i]
            assert len(period.io_workloads) == len(workloads)
            expected_workloads = copy.copy(workloads)
            for workload in period.io_workloads:
                expected_workloads.remove(workload)

    def test_get_periods_custom(self):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )
        custom_config = profiles.StandardVdbProfile.DEFAULT_PERIOD_CONFIGS[-1]
        profile.period_configs = [custom_config]
        periods = profile.get_periods()
        for period in periods:
            assert period.config == custom_config

    @mock.patch('valence.vl_components.vdbench.CsvVdbProfile.get_periods')
    def test_get_session(self, get_periods):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )

        # ... get the periods
        workloads = profile.get_workloads()
        periods = [_io_config.IoPeriod(
            io_period_config=profile.DEFAULT_PERIOD_CONFIGS[0],
            io_workloads=workloads
        )]
        get_periods.return_value = periods

        session = profile.get_session()
        assert session.config.compression_ratio == \
            profiles.StandardVdbProfile.DEFAULT_SESSION_CONFIG.compression_ratio
        assert session.config.pattern_path == \
            profiles.StandardVdbProfile.DEFAULT_SESSION_CONFIG.pattern_path
        assert session.config.dedup_config == profiles.StandardVdbProfile.DEDUP_2X
        assert not session.standalone
        assert len(session.io_periods) == len(periods)
        expected_periods = copy.copy(periods)
        for period in session.io_periods:
            expected_periods.remove(period)

        # Ensure standalone sessions cannot be generated by profiles for multiple clients
        with pytest.raises(AssertionError):
            session = profile.get_session(standalone=True)

    @mock.patch('valence.vl_components.vdbench.CsvVdbProfile.get_periods')
    def test_get_session_standalone(self, get_periods):
        # Ensure a standalone session can be generated by profiles for single clients
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:1],
            csv_file_path=self.csv_1_client
        )

        # ... get the periods
        workloads = profile.get_workloads()
        periods = [_io_config.IoPeriod(
            io_period_config=profile.DEFAULT_PERIOD_CONFIGS[0],
            io_workloads=workloads
        )]
        get_periods.return_value = periods

        session = profile.get_session(standalone=True)
        assert session.standalone

    def test_get_session_custom(self):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )
        profile.session_config.compression_ratio = 10
        session = profile.get_session()
        assert session.config.compression_ratio == 10

    @mock.patch('valence.vl_components._account.CreateAccounts', vl_components.CreateAccounts.mock)
    @mock.patch('valence.vl_components._volume.CreateVolumes', vl_components.CreateVolumes.mock)
    @mock.patch('valence.vl_components._connection.CreateIscsiAccountConnection',
                vl_components.CreateIscsiAccountConnection.mock)
    @mock.patch('valence.vl_components.vdbench.CsvVdbProfile.get_session')
    def test_get_environment(self, get_session):
        profile = profiles.CsvVdbProfile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:2],
            csv_file_path=self.csv_2_clients
        )

        # ... get the session
        periods = profile.get_periods()
        session = _io_config.IoSession(profile.session_config, periods)
        get_session.return_value = session

        environment = profile.get_environment()
        assert len(environment.vdb_environment_cmts) == 1
        environment_cmt = environment.vdb_environment_cmts[0]
        assert environment_cmt.connections_cmt is profile.connections_cmt
        assert environment_cmt.io_session == session
        assert environment_cmt.master_client_dm is profile.connections_cmt.client_dms[0]

        master_client_dm = models.ResourceClient.get_mock_object()
        environment = profile.get_environment(master_client_dm)
        environment_cmt = environment.vdb_environment_cmts[0]
        assert environment_cmt.master_client_dm is master_client_dm

        assert profile.connections_cmt.state == components.States.ready
        environment.apply()
        assert profile.connections_cmt.state == components.States.applied

    def test_from_csv_folder_one_client_aligned(self):
        profile = profiles.CsvVdbProfile.from_csv_folder(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:1],
            csv_folder_path=self.csv_dir
        )
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 52

    def test_from_csv_folder_one_client_unaligned(self):
        profile = profiles.CsvVdbProfile.from_csv_folder(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:1],
            csv_folder_path=self.csv_dir,
            aligned=False
        )
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 10

    def test_from_csv_folder_multiple_clients(self):
        # Aligned 3-client profile
        profile = profiles.CsvVdbProfile.from_csv_folder(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:3],
            csv_folder_path=self.csv_dir
        )
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 3
        assert len(workloads) == 56

        # Ensure an exception is raised if the specified alignment is unsupported
        with pytest.raises(AssertionError):
            profile = profiles.CsvVdbProfile.from_csv_folder(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms[:3],
                csv_folder_path=self.csv_dir,
                aligned=False
            )

        # Ensure an exception is raised if an unsupported number of clients is provided
        with pytest.raises(ValueError):
            profile = profiles.CsvVdbProfile.from_csv_folder(
                tag='unittest',
                cluster_dm=self.cluster_dm,
                client_dms=self.client_dms[0:5],
                csv_folder_path=self.csv_dir
            )


class TestCsvVdbExtendedProfile(BaseTestCsvProfile, unittest.TestCase):

    @classmethod
    @mock.patch('valence.vl_components._volume.CreateVolumes', vl_components.CreateVolumes.mock)
    @mock.patch('valence.vl_components._account.CreateAccounts', vl_components.CreateAccounts.mock)
    @mock.patch('valence.vl_components._connection.CreateIscsiAccountConnection',
                vl_components.CreateIscsiAccountConnection.mock)
    def setUpClass(cls):
        super(TestCsvVdbExtendedProfile, cls).setUpClass()
        cls.csv_extended_profile = profiles.CsvVdbExtendedProfile
        cls.csv_extended_profile.MAX_SLAVE_COUNT = 2

        # Go ahead and setup a two client extended profile to share between tests.
        cls.two_client_csv_extended_profile = cls.csv_extended_profile(
            tag='unittest',
            cluster_dm=cls.cluster_dm,
            client_dms=cls.client_dms[:2],
            csv_folder_path=cls.csv_dir
        )

    def test_one_client(self):
        # Ensure an aligned profile is created if only one client is provided
        extended_profile = self.csv_extended_profile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:1],
            csv_folder_path=self.csv_dir
        )
        assert len(extended_profile._profiles) == 1
        profile = extended_profile._profiles[0]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 52

    def test_two_clients(self):
        # Ensure an unaligned profile is created if more than one client is provided
        extended_profile = self.two_client_csv_extended_profile
        assert len(extended_profile._profiles) == 2
        profile = extended_profile._profiles[0]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 10  # Unaligned profile
        profile = extended_profile._profiles[1]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 52  # Aligned profile

    def test_many_clients(self):
        # Ensure only one client is assigned to the unaligned profile
        extended_profile = self.csv_extended_profile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:3],
            csv_folder_path=self.csv_dir
        )
        assert len(extended_profile._profiles) == 2
        profile = extended_profile._profiles[0]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 10  # Unaligned profile
        profile = extended_profile._profiles[1]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 2
        assert len(workloads) == 52  # Aligned profile

    def test_max_clients(self):
        # Exceed the maximum number of clients such that multiple aligned profiles are created
        extended_profile = self.csv_extended_profile(
            tag='unittest',
            cluster_dm=self.cluster_dm,
            client_dms=self.client_dms[:4],
            csv_folder_path=self.csv_dir
        )
        assert len(extended_profile._profiles) == 3
        profile = extended_profile._profiles[0]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 10  # Unaligned profile
        profile = extended_profile._profiles[1]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 2
        assert len(workloads) == 52  # Aligned profile
        profile = extended_profile._profiles[2]
        workloads = profile.get_workloads()
        assert len(profile.connections_cmt.client_dms) == 1
        assert len(workloads) == 52  # Aligned profile

    def test_get_environment(self):
        extended_profile = self.two_client_csv_extended_profile

        environment = extended_profile.get_environment()
        connection_cmts = [profile.connections_cmt for profile in extended_profile._profiles]
        assert all([cmt.state == components.States.ready for cmt in connection_cmts])
        environment.apply()
        assert all([cmt.state == components.States.applied for cmt in connection_cmts])


class TestQosProfile(BaseTestProfile, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_dm.api_endpoint = models.SfVersion('9.0')
        cls.client_dms = [models.ResourceClient.create(
            client_name='c{}'.format(i),
            mip='1.0.0.{}'.format(i),
            os=at2_enums.ClientOS.ubuntu_14_04 if i % 2 == 0 else at2_enums.ClientOS.centos_7
        ) for i in range(6)]
        cls.account_dm = models.Account.create(username='unittest-TAG-0', status='active', account_id=1)
        cls.disk_map = {
            61: '/dev/sdc',
            62: '/dev/sde',
            63: '/dev/sdb',
            64: '/dev/sda',
            65: '/dev/sdg',
            66: '/dev/sdf',
            67: '/dev/sdh',
            68: '/dev/sdi',
            69: '/dev/sdj',
            70: '/dev/sdk',
            71: '/dev/sdl',
            72: '/dev/sdm',
            73: '/dev/sdn',
            74: '/dev/sdo',
        }
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            account_id=1,
            volume_id=i,
            total_size=models.DataSize(i, enums.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
            enable512e=True
        ) for i in cls.disk_map]
        cls.initiator_dms = [models.Initiator.get_mock_object() for _ in range(2)]

        # define the components
        cls.accounts_cmt = _account.MockAccounts('mock', [cls.account_dm])
        cls.volumes_cmts = [
            _volume.MockVolumes('mock-0', cls.cluster_dm, cls.volume_dms[0:1], cls.accounts_cmt),
            _volume.MockVolumes('mock-1', cls.cluster_dm, cls.volume_dms[1:3], cls.accounts_cmt),
            _volume.MockVolumes('mock-2', cls.cluster_dm, cls.volume_dms[3:6], cls.accounts_cmt),
            _volume.MockVolumes('mock-3', cls.cluster_dm, cls.volume_dms[6:], cls.accounts_cmt),
        ]
        cls.conn_cmts = [
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[0],
                volumes_cmts=cls.volumes_cmts[0:2],
            ),
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[1],
                volumes_cmts=cls.volumes_cmts[1:],
            )
        ]
        cls.conn_manager = vl_components.ManageConnections(cls.conn_cmts)

    def test_qos_block_size(self):
        profile = profiles.QosHistStandardBlockSizeProfile(self.conn_manager)
        workloads = profile.get_workloads()
        assert workloads[0].config.read_perc == 0
        assert workloads[0].config.transfer_size_config == _parm.TransferSizeFixed(2**11)
        assert workloads[6].config.read_perc == 0
        assert workloads[13].config.read_perc == 100
        assert workloads[13].config.transfer_size_config == _parm.TransferSizeFixed(2**17)
        profile.get_periods()

    def test_qos_throttling(self):
        thread_cnt = len(self.volume_dms)
        profile = profiles.QosHistThrottlingProfile(self.conn_manager, thread_count=thread_cnt)
        workloads = profile.get_workloads()
        periods = profile.get_periods()
        assert all([w.config.read_perc == 20 for w in workloads])
        assert all([w.config.transfer_size_config == _parm.TransferSizeFixed(2**12) for w in workloads])
        assert all([w.config.thread_count == thread_cnt for w in workloads])
        assert all([p.config.interval == 1 for p in periods])
        assert all([isinstance(p.config.io_rate_config, _parm.IoRateMax) for p in periods])

    def test_qos_iops(self):
        profile = profiles.QosHistIopsProfile(self.conn_manager)
        duration = (len(profile.CURVE)+1) * 60  # using default elapsed=60
        workloads = profile.get_workloads()
        periods = profile.get_periods()
        assert all([w.config.read_perc == 50 for w in workloads])
        assert all([w.config.transfer_size_config == _parm.TransferSizeFixed(2**12) for w in workloads])
        assert all([isinstance(p.config.io_rate_config, _parm.IoRateCurve) for p in periods])
        assert all([p.config.interval == 30 for p in periods])
        assert all([p.config.duration == duration for p in periods])
        assert periods[0].config.io_rate_config.curve == profile.CURVE
        assert periods[0].config.io_rate_config.distribution == _enums.IoDistribution.uniform
