import unittest

import icechai_sdk
import mock
import pytest
import rstr
import sf_artifacts

from camelot.models import at2_enums
from sf_api import data_size
from valence import (
    components,
    errors,
    models,
)
from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _enums,
    _errors,
    _io_config,
    _log_parser,
    _parm,
    _run,
    _setup,
    profiles,
)


class BaseTestRun(object):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.create(mvip='0.0.0.0')
        cls.client_dms = [models.ResourceClient.create(
            client_name='c{}'.format(i),
            mip='1.0.0.{}'.format(i),
            os=at2_enums.ClientOS.ubuntu_14_04
        ) for i in range(3)]
        cls.account_dm = models.Account.create(username='unittest-TAG-0', status='active', account_id=1)
        cls.disk_map = {
            61: '/dev/sdc',
            62: '/dev/sde',
            63: '/dev/sdb',
            64: '/dev/sda',
            65: '/dev/sdg',
            66: '/dev/sdf',
            67: '/dev/sdh',
            68: '/dev/sdi',
            69: '/dev/sdj',
            70: '/dev/sdk'
        }
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            account_id=1,
            volume_id=i,
            total_size=data_size.DataSize(i, data_size.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
            enable512e=False
        ) for i in cls.disk_map]

        # define global artifact manager
        sf_artifacts.set_global_manager(sf_artifacts.LocalArtifactManager())

        # define the components
        cls.accounts_cmt = _account.MockAccounts('mock', [cls.account_dm])
        cls.volumes_cmts = [
            _volume.MockVolumes('mock-0', cls.cluster_dm, cls.volume_dms[0:1], cls.accounts_cmt),
            _volume.MockVolumes('mock-1', cls.cluster_dm, cls.volume_dms[1:3], cls.accounts_cmt),
            _volume.MockVolumes('mock-2', cls.cluster_dm, cls.volume_dms[3:6], cls.accounts_cmt),
            _volume.MockVolumes('mock-3', cls.cluster_dm, cls.volume_dms[6:], cls.accounts_cmt),
        ]
        cls.conn_cmts = [
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[0],
                volumes_cmts=cls.volumes_cmts[0:2],
            ),
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[1],
                volumes_cmts=cls.volumes_cmts[1:],
            )
        ]
        cls.conn_manager = _connection.ManageConnections(cls.conn_cmts)
        cls.standard_profile = profiles.StandardVdbProfile(cls.conn_manager)
        cls.io_session = cls.standard_profile.get_session()

        cls.run_start_line = _log_parser.VdbRunInfoLine(rstr.xeger(_log_parser.RUN_START_LINE))
        cls.io_line = _log_parser.VdbIoLine(
            raw_line=rstr.xeger(_log_parser.IO_LINE),
            defined_intervals=[1],
            last_line=cls.run_start_line
        )

        cls.vdb_id = {'vdbench_id': 17}
        cls.vdb_master_statuses = [
            {'status': 'running'},
            {'status': 'completed'},
            {'status': 'completed'},
            {'status': 'running'},
            {'status': 'running'},
            {'status': 'completed'},
            {'status': 'completed'}
        ]
        cls.vdb_slave_statuses = [
            {'status': 'running'},
            {'status': 'completed'},
            {'status': 'completed'},
            {'status': 'completed'}
        ]
        cls.vdb_run_result = {
            'vdbench_id': 17,
            'output_dir': '/vdb/path',
            'log_file_path': '/log/path'
        }

        cls.mock_vdb_log = [
            rstr.xeger(_log_parser.RUN_START_LINE),
            rstr.xeger(_log_parser.IO_LINE),
            rstr.xeger(_log_parser.IO_LINE),
            rstr.xeger(_log_parser.IO_LINE),
            rstr.xeger(_log_parser.IO_LINE)
        ]

    def setUp(self):
        self.setup_vdb_env_cmt = _setup.SetupVdbEnvironment(self.io_session, self.conn_manager)


class TestRunVdb(BaseTestRun, unittest.TestCase):

    def test_init(self):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        assert run_cmt.child_cmts == [run_cmt]
        assert run_cmt.TIMEOUT == 2 * 24 * 60 * 60
        assert run_cmt._wait_for_io_timeout == 10 * 60

        journaling_config = _enums.JournalingConfig.validate
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, journaling_config=journaling_config)
        assert run_cmt.TIMEOUT == 2 * 24 * 60 * 60
        assert run_cmt._wait_for_io_timeout == 24 * 60 * 60
        assert run_cmt._journal_validate_vdb_cmt is None

        run_cmt = _run.RunVdb(
            environment_cmt=self.setup_vdb_env_cmt,
            timeout=42,
            journaling_config=journaling_config,
            wait_for_io_timeout=7
        )
        assert run_cmt.TIMEOUT == 42
        assert run_cmt._wait_for_io_timeout == 7

        self.conn_cmts[0]._client_dm = models.ResourceClient.get_mock_object(
            overrides=dict(os=at2_enums.ClientOS.windows_2012)
        )
        with pytest.raises(NotImplementedError):
            run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        self.conn_cmts[0]._client_dm = self.client_dms[0]

        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, journaling_config=_enums.JournalingConfig.async_io)

        assert isinstance(run_cmt._journal_validate_vdb_cmt, _run.RunVdb)
        assert run_cmt._journal_validate_vdb_cmt.journaling_config == _enums.JournalingConfig.validate
        assert run_cmt._journal_validate_vdb_cmt.environment_cmt is run_cmt.environment_cmt

    @staticmethod
    def _mock_state():
        return components.States.applied

    @mock.patch('valence.vl_components.vdbench.RunVdb.assert_valid')
    @mock.patch('valence.vl_components.vdbench.RunVdb.apply')
    def test_validate_journal(self, apply_validate_cmt, assert_validate_cmt_valid):

        # test validate_journal while cmt is not applied
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, port=42,
                              journaling_config=_enums.JournalingConfig.async_io)
        with pytest.raises(components.ComponentStateError):
            run_cmt.validate_journal()

        # test validate_journal when not configured to validate journal
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, port=42,
                              journaling_config=_enums.JournalingConfig.none)
        run_cmt.state = components.States.applied
        with pytest.raises(components.ComponentStateError):
            run_cmt.validate_journal()

        # test journal validate component is applied during validate_journal()
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, port=42,
                              journaling_config=_enums.JournalingConfig.async_io)
        run_cmt.state = components.States.applied

        with mock.patch('valence.vl_components.vdbench.RunVdb.state',
                        new_callable=TestRunVdb._mock_state):
            run_cmt.validate_journal()
            assert apply_validate_cmt.call_count == 1
            assert assert_validate_cmt_valid.call_count == 1
            assert_validate_cmt_valid.assert_called_with(is_running=False)
            assert run_cmt._journal_validate_vdb_cmt.state == components.States.ready

        # assert resetting parent also resets children
        run_cmt.reset()
        assert run_cmt._journal_validate_vdb_cmt.state == components.States.ready
        # assert we cannot verify once parent component has been reset
        with pytest.raises(components.ComponentStateError):
            run_cmt.validate_journal()

    def test_validate_periods_for_data_validation(self):
        workload_config = _io_config.IoWorkloadConfig(
            transfer_size_config=_parm.TransferSizeFixed(4096)
        )
        io_workloads = [_io_config.IoWorkload(
            io_workload_config=workload_config,
            client_dm=self.client_dms[0],
            volumes_cmt=self.volumes_cmts[0],
            volume_index=0
        )]
        period_config = _io_config.IoPeriodConfig(_parm.IoRateFixed(1000))
        io_periods = [_io_config.IoPeriod(period_config, io_workloads)]
        _run.RunVdb._validate_periods_for_data_validation(io_periods)

        workload_config = _io_config.IoWorkloadConfig(
            transfer_size_config=_parm.TransferSizeAligned(4096, 8 * 4096, 4096)
        )
        io_workloads = [_io_config.IoWorkload(
            io_workload_config=workload_config,
            client_dm=self.client_dms[0],
            volumes_cmt=self.volumes_cmts[1],
            volume_index=0
        )]
        io_periods.append(_io_config.IoPeriod(period_config, io_workloads))
        with pytest.raises(ValueError):
            _run.RunVdb._validate_periods_for_data_validation(io_periods)

        io_periods.pop()
        io_periods[0].io_workloads[0].config.offset = 4096
        with pytest.raises(ValueError):
            _run.RunVdb._validate_periods_for_data_validation(io_periods)

    def test_client_dms(self):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        assert run_cmt.client_dms == self.setup_vdb_env_cmt.client_dms
        assert run_cmt.master_client_dm == self.setup_vdb_env_cmt.master_client_dm

    @mock.patch('valence.vl_components.vdbench.RunVdb.update_stream')
    def test_current_io_line_item(self, update_stream):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        run_cmt._run_info_handler.current_io_line_item = self.io_line
        assert run_cmt.current_io_line_item == self.io_line

    @mock.patch('valence.vl_components.vdbench._helpers.tail_vdb_log')
    def test_get_vdbench_log(self, tail_vdb_log):
        tail_vdb_log.return_value = self.mock_vdb_log

        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        assert run_cmt.get_vdbench_log(max_lines=42) == self.mock_vdb_log
        tail_vdb_log.assert_called_with(run_cmt.master_client_dm, run_cmt._logfile, 0, 42)
        assert run_cmt.get_vdbench_log(max_lines=42, start_line=5) == self.mock_vdb_log
        tail_vdb_log.assert_called_with(run_cmt.master_client_dm, run_cmt._logfile, 5, 42)

    @mock.patch('valence.vl_components.vdbench._helpers.tail_vdb_log')
    def test_update_stream(self, tail_vdb_log):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        with pytest.raises(components.ComponentStateError):
            run_cmt.update_stream()

        # Ensure the run info handler is updated
        tail_vdb_log.return_value = self.mock_vdb_log
        run_cmt._state = components.States.applying
        run_cmt.update_stream()
        assert run_cmt._last_line_read == len(self.mock_vdb_log)
        assert run_cmt._run_info_handler.current_io_line_item.raw == self.mock_vdb_log[-1]
        assert run_cmt._error_handler.eio_count == 0

        # Ensure the error stream handler is updated
        eio_line = rstr.xeger(_log_parser.EIO_LINE)
        io_line = rstr.xeger(_log_parser.IO_LINE)
        tail_vdb_log.return_value = [eio_line, io_line]
        run_cmt.update_stream()
        assert run_cmt._run_info_handler.current_io_line_item.raw == io_line
        assert run_cmt._error_handler.eio_count == 1

        # Ensure log is read until the max # of lines isn't returned
        run_cmt.MAX_LINES_SSH = 5
        io_lines_max = [rstr.xeger(_log_parser.IO_LINE) for _ in range(run_cmt.MAX_LINES_SSH)]
        io_lines_less = [rstr.xeger(_log_parser.IO_LINE) for _ in range(run_cmt.MAX_LINES_SSH - 1)]
        tail_vdb_log.side_effect = [io_lines_max, io_lines_less, [io_line]]
        run_cmt.update_stream()
        assert run_cmt._last_line_read == sum([
            len(self.mock_vdb_log),
            2,
            len(io_lines_max) + len(io_lines_less)
        ])

    @mock.patch('valence.vl_components.vdbench.RunVdb.update_stream')
    @mock.patch('valence.vl_components.vdbench.RunVdb._master_is_running')
    def test_assert_valid(self, master_is_running, update_stream):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        run_cmt._state = components.States.applying
        run_cmt._vdbench_id = 17

        # Vdbench is running
        master_is_running.return_value = True
        run_cmt.assert_valid()
        with pytest.raises(_errors.VdbenchRuntimeError):
            run_cmt.assert_valid(is_running=False)

        # Vdbench is NOT running
        master_is_running.return_value = False
        run_cmt.assert_valid(is_running=False)
        with pytest.raises(_errors.VdbenchRuntimeError):
            run_cmt.assert_valid()

        # Error recorded (data_errors = 0)
        run_cmt._error_handler.eio_count = 1
        with pytest.raises(_errors.VdbenchRuntimeError):
            run_cmt.assert_valid()
        master_is_running.return_value = True
        with pytest.raises(_errors.VdbenchRuntimeError):
            run_cmt.assert_valid()

        # Error recorded (data_errors = 1)
        run_cmt.data_errors = 1
        run_cmt.assert_valid()

    @mock.patch('os.path.basename')
    @mock.patch('sf_ssh.ssh.Ssh.close')
    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    @mock.patch('sf_ssh.ssh.Ssh._init_client')
    @mock.patch('valence.functional_models.UbuntuClient.pull_support_bundle')
    @mock.patch('sf_artifacts.LocalArtifactManager.store_text')
    @mock.patch('sf_artifacts.LocalArtifactManager.compress_directory')
    @mock.patch('sf_artifacts.LocalArtifactManager.store_artifacts_remote')
    def test_store_artifacts(self, store_artifacts_remote, compress_artifact_directory, store_text,
                             pull_client_bundle, init_client, run_ssh_cmd, close_client, get_path_name):
        compress_artifact_directory.return_value = '/example/path'
        get_path_name.return_value = 'example'
        artifact_manager = sf_artifacts.LocalArtifactManager()
        sf_artifacts.set_global_manager(artifact_manager)
        self.setup_vdb_env_cmt.apply()
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        run_cmt._logfile = '/log/path'
        run_cmt._vdb_parm = 'mock_parm'

        run_cmt.store_artifacts(get_bundle=False)
        assert store_text.called
        assert store_artifacts_remote.call_count == 1
        assert run_ssh_cmd.call_count == 1
        assert not pull_client_bundle.called

        store_artifacts_remote.call_count = 0
        run_cmt.store_artifacts(get_vdb_dir=True)
        compress_artifact_directory.assert_called_with(
            directory=run_cmt._vdbench_dir,
            excludes=['*.mmap', '*.map', '*.jnl', '*swat_mon.txt'],
            remote_host_dm=run_cmt.master_client_dm,
            destination='{}-example.tar.bz2'.format(run_cmt.master_client_dm.client_name)
        )
        assert store_artifacts_remote.call_count == 2
        assert run_ssh_cmd.called
        assert pull_client_bundle.called

        compress_artifact_directory.side_effect = sf_artifacts.ArtifactsError()
        with pytest.raises(_errors.VdbenchError):
            run_cmt.store_artifacts(get_vdb_dir=True)

    @mock.patch('time.sleep')
    @mock.patch('time.time')
    @mock.patch('valence.vl_components.vdbench._helpers.tail_vdb_log')
    @mock.patch('icechai_sdk.icechai_api.SDK.stop_vdbench_slave')
    @mock.patch('icechai_sdk.icechai_api.SDK.stop_vdbench_master')
    @mock.patch('icechai_sdk.icechai_api.SDK.run_vdbench_slave')
    @mock.patch('icechai_sdk.icechai_api.SDK.run_vdbench_master')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_slave_status')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_master_status')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_id')
    @mock.patch('valence.vl_components.vdbench.RunVdb.wait_for_io_running')
    def test_apply(self, wait_for_io_running, get_vdb_id, get_vdb_master_status, get_vdb_slave_status,
                   run_vdb_master, run_vdb_slave, stop_vdb_master, stop_vdb_slave, tail_vdb_log, time,
                   sleep):
        time.side_effect = range(100)
        get_vdb_id.return_value = self.vdb_id
        get_vdb_master_status.side_effect = self.vdb_master_statuses
        get_vdb_slave_status.side_effect = self.vdb_slave_statuses
        run_vdb_master.return_value = self.vdb_run_result
        tail_vdb_log.side_effect = [self.mock_vdb_log, []]

        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, port=42)
        run_cmt.apply(timeout=61)
        assert self.setup_vdb_env_cmt.state == components.States.applied
        assert wait_for_io_running.called
        assert stop_vdb_master.call_count == 1
        assert stop_vdb_slave.call_count == len(self.conn_manager.client_dms)
        assert run_vdb_slave.call_count == len(self.conn_manager.client_dms)
        run_vdb_master.assert_called_with(self.setup_vdb_env_cmt.get_vdb_parm(port=42))
        assert run_cmt._vdbench_id == 17
        assert run_cmt._vdbench_dir == '/vdb/path'
        assert run_cmt._logfile == '/log/path'

        time.side_effect = [0, 1, 2, 3, 4, 65]
        get_vdb_master_status.side_effect = self.vdb_master_statuses
        get_vdb_slave_status.side_effect = self.vdb_slave_statuses
        tail_vdb_log.side_effect = [self.mock_vdb_log, []]

        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, port=42)
        with pytest.raises(errors.TimeoutError):
            run_cmt.apply(timeout=1)

    @mock.patch('time.sleep')
    @mock.patch('valence.vl_components.vdbench.RunVdb.store_artifacts')
    @mock.patch('valence.vl_components.vdbench._helpers.tail_vdb_log')
    @mock.patch('sf_ssh.ssh.Ssh.close')
    @mock.patch('sf_ssh.ssh.Ssh.run_command')
    @mock.patch('sf_ssh.ssh.Ssh._init_client')
    @mock.patch('icechai_sdk.icechai_api.SDK.stop_vdbench_slave')
    @mock.patch('icechai_sdk.icechai_api.SDK.stop_vdbench_master')
    @mock.patch('icechai_sdk.icechai_api.SDK.run_vdbench_slave')
    @mock.patch('icechai_sdk.icechai_api.SDK.run_vdbench_master')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_slave_status')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_master_status')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_id')
    def test_finish(self, get_vdb_id, get_vdb_master_status, get_vdb_slave_status, run_vdb_master,
                    run_vdb_slave, stop_vdb_master, stop_vdb_slave, init_client, run_ssh_cmd,
                    close_client, tail_vdb_log, store_artifacts, sleep):
        get_vdb_id.return_value = self.vdb_id
        run_vdb_master.return_value = self.vdb_run_result

        # Apply
        get_vdb_master_status.side_effect = self.vdb_master_statuses
        get_vdb_slave_status.side_effect = self.vdb_slave_statuses
        tail_vdb_log.side_effect = [self.mock_vdb_log, []]
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        run_cmt.apply_async()

        # Apply
        get_vdb_master_status.side_effect = self.vdb_master_statuses
        get_vdb_slave_status.side_effect = self.vdb_slave_statuses
        tail_vdb_log.side_effect = [self.mock_vdb_log, []]
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        run_cmt.apply_async(timeout=60)
        stop_vdb_master.call_count = 0
        stop_vdb_slave.call_count = 0

        # Finish (kill=True)
        get_vdb_master_status.side_effect = None
        get_vdb_slave_status.side_effect = None
        get_vdb_master_status.return_value = self.vdb_master_statuses[-1]
        get_vdb_slave_status.return_value = self.vdb_slave_statuses[-1]
        run_cmt.data_errors = 1
        run_cmt.finish(kill=True)
        assert stop_vdb_master.call_count == 1
        assert stop_vdb_slave.call_count == len(self.conn_manager.client_dms)
        assert run_cmt._vdbench_id is None
        vdb_dir = run_cmt._vdbench_dir

        # Reset
        run_cmt._error_handler.eio_count = 1
        run_cmt._run_info_handler.current_io_line_item = self.io_line
        run_cmt.reset()
        assert store_artifacts.call_count == 1
        run_ssh_cmd.assert_called_with('rm -r "{}"'.format(vdb_dir))
        assert run_cmt._error_handler.eio_count == 0
        assert run_cmt._run_info_handler.current_io_line_item is None
        assert run_cmt._vdbench_id is None
        assert run_cmt._vdbench_dir is None
        assert run_cmt._logfile is None
        assert run_cmt._last_line is None
        assert run_cmt._last_line_read == 0

    @mock.patch('time.time')
    @mock.patch('time.sleep')
    @mock.patch('valence.vl_components.vdbench.RunVdb.update_stream')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_master_status')
    def test_wait_for_io_running(self, get_master_status, update_stream, sleep, time_now):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, wait_for_io_timeout=1)
        run_cmt._master_sdk = icechai_sdk.SDK(run_cmt.master_client_dm.mip)
        run_cmt._state = components.States.applying
        run_cmt._vdbench_id = 17
        run_cmt._error_handler.critical_errors = ['test_error']

        get_master_status.return_value = {
            _run.VDBENCH_STATUS: _enums.VdbStatus.completed.value
        }
        with pytest.raises(_errors.VdbenchRuntimeError):
            run_cmt.wait_for_io_running(0)
        assert sleep.call_count == 0

        time_now.side_effect = [0, 1, 2]
        get_master_status.return_value = {
            _run.VDBENCH_STATUS: _enums.VdbStatus.running.value
        }
        with pytest.raises(errors.TimeoutError):
            run_cmt.wait_for_io_running(0)
        assert sleep.call_count == 2

    @mock.patch('valence.vl_components.vdbench.RunVdb._master_is_running')
    def test_apply_is_running(self, master_is_running):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, wait_for_io_timeout=1)

        master_is_running.return_value = True
        assert not run_cmt._apply_is_running()

        run_cmt._vdbench_id = 17
        assert run_cmt._apply_is_running()

        master_is_running.return_value = False
        assert not run_cmt._apply_is_running()

    @mock.patch('time.time')
    @mock.patch('time.sleep')
    @mock.patch('icechai_sdk.icechai_api.SDK.stop_vdbench_slave')
    @mock.patch('icechai_sdk.icechai_api.SDK.stop_vdbench_master')
    @mock.patch('valence.vl_components.vdbench.RunVdb._slave_is_running')
    @mock.patch('valence.vl_components.vdbench.RunVdb._master_is_running')
    @mock.patch('icechai_sdk.icechai_api.SDK.get_vdbench_id')
    def test_stop_all_vdbench(self, get_vdb_id, master_is_running, slave_is_running, stop_vdb_master,
                              stop_vdb_slave, sleep, time_now):
        get_vdb_id.return_value = self.vdb_id
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt, wait_for_io_timeout=1)
        run_cmt._master_sdk = icechai_sdk.SDK(run_cmt.master_client_dm.mip)
        run_cmt._slave_sdks = [icechai_sdk.SDK(c.mip) for c in run_cmt.client_dms[1:]]

        time_now.side_effect = [0, 10 * 60, 10 * 60 + 1]
        master_is_running.side_effect = [True for _ in range(2)]
        with pytest.raises(errors.TimeoutError):
            run_cmt._stop_all_vdbench()
        assert sleep.call_count == 1
        assert master_is_running.call_count == 2
        assert slave_is_running.call_count == 0

        time_now.side_effect = [0, 0, 10 * 60, 10 * 60 + 1]
        master_is_running.side_effect = None
        master_is_running.return_value = False
        slave_is_running.side_effect = [True for _ in range(2 * len(self.conn_manager.client_dms))]
        with pytest.raises(errors.TimeoutError):
            run_cmt._stop_all_vdbench()
        assert sleep.call_count == 2
        assert master_is_running.call_count == 3
        assert slave_is_running.call_count == 2 * len(self.conn_manager.client_dms)

    @mock.patch('icechai_sdk.icechai_api.SDK.make_temp_directory')
    def test_setup_journaling(self, make_temp_dir):
        self.setup_vdb_env_cmt.apply()
        self.setup_vdb_env_cmt.get_vdb_parm(journaling_config=_enums.JournalingConfig.recover)
        run_cmt = _run.RunVdb(
            environment_cmt=self.setup_vdb_env_cmt,
            journaling_config=_enums.JournalingConfig.none
        )
        run_cmt._master_sdk = icechai_sdk.SDK(run_cmt.master_client_dm.mip)

        run_cmt._setup_journaling()
        assert not make_temp_dir.called

        run_cmt.journaling_config = _enums.JournalingConfig.full
        run_cmt._setup_journaling()
        assert make_temp_dir.called

    def test_stream_handlers(self):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        run_info_handler = run_cmt._run_info_handler

        with pytest.raises(TypeError):
            run_cmt._remove_stream_handler(1)

        run_info_handler.CAN_REGISTER = False
        with pytest.raises(TypeError):
            run_cmt._remove_stream_handler(run_info_handler)

        run_info_handler.CAN_REGISTER = True
        run_cmt._remove_stream_handler(run_info_handler)
        assert run_info_handler not in run_cmt._stream_handlers.values()

        with pytest.raises(TypeError):
            run_cmt._add_stream_handler(1)

        run_info_handler.CAN_REGISTER = False
        with pytest.raises(TypeError):
            run_cmt._add_stream_handler(run_info_handler)

        run_info_handler.CAN_REGISTER = True
        run_cmt._add_stream_handler(run_info_handler)
        assert run_info_handler in run_cmt._stream_handlers.values()

    def test_set_identifier(self):
        run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)
        master_sdk = mock.MagicMock()
        run_cmt._master_sdk = master_sdk
        run_cmt.set_test_identifier('test')
        master_sdk.set_test_identifier.assert_called_with(test_identifier='test')


class TestManageRunVdbs(BaseTestRun, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestManageRunVdbs, cls).setUpClass()
        cls.conn_cmt_extra = _connection.MockConnection(
            disk_map=cls.disk_map,
            client_dm=cls.client_dms[2],
            volumes_cmts=cls.volumes_cmts[2:],
        )
        standard_profile = profiles.StandardVdbProfile(cls.conn_cmt_extra)
        io_session = standard_profile.get_session()
        cls.setup_env_cmts = [
            _setup.SetupVdbEnvironment(cls.io_session, cls.conn_manager),
            _setup.SetupVdbEnvironment(io_session, cls.conn_cmt_extra)
        ]

    def setUp(self):
        self.run_cmts = [
            _run.RunVdb(self.setup_env_cmts[0], journaling_config=_enums.JournalingConfig.async_io),
            _run.RunVdb(self.setup_env_cmts[1], journaling_config=_enums.JournalingConfig.async_io)
        ]

    def test_init(self):
        manage_runs_cmt = _run.ManageRunVdbs(self.run_cmts)
        assert manage_runs_cmt.TIMEOUT == 2 * 24 * 60 * 60
        assert manage_runs_cmt.child_cmts == self.run_cmts
        assert manage_runs_cmt.client_dms == self.client_dms
        master_client_names = [cmt.master_client_dm.client_name for cmt in self.run_cmts]
        assert str(manage_runs_cmt) == \
            'ManageRunVdbs<master_client_names={}>'.format(master_client_names)

        manage_runs_cmt = _run.ManageRunVdbs(self.run_cmts, timeout=42)
        assert manage_runs_cmt.TIMEOUT == 42

        run_cmts = [self.run_cmts[0]]
        run_cmts.extend(self.run_cmts)
        with pytest.raises(ValueError):
            manage_runs_cmt = _run.ManageRunVdbs(run_cmts)

        run_cmts = [_run.RunVdb(self.setup_env_cmts[0])]
        run_cmts.extend(self.run_cmts)
        with pytest.raises(ValueError):
            manage_runs_cmt = _run.ManageRunVdbs(run_cmts)

    def test_from_environment_cmt(self):
        journaling_config = _enums.JournalingConfig.recover
        manage_env_cmt = _setup.ManageVdbEnvironments(self.setup_env_cmts)
        manage_runs_cmt = _run.ManageRunVdbs.from_environment_cmt(
            environments_cmt=manage_env_cmt,
            timeout=42,
            port=17,
            journaling_config=journaling_config,
        )

        assert len(manage_runs_cmt.child_cmts) == len(self.setup_env_cmts)
        for run_cmt in manage_runs_cmt.child_cmts:
            assert run_cmt.TIMEOUT == 42
            assert run_cmt.port == 17
            assert run_cmt.journaling_config == journaling_config
            assert run_cmt._journal_validate_vdb_cmt is None

        # Ensure the environment manager gets applied correctly
        manage_runs_cmt._run_vdb_cmts = []
        manage_runs_cmt.apply_async()
        assert manage_env_cmt.state == components.States.applied

    @mock.patch('valence.vl_components.vdbench.RunVdb.set_test_identifier')
    @mock.patch('valence.vl_components.vdbench.RunVdb.store_artifacts')
    @mock.patch('valence.vl_components.vdbench.RunVdb.ensure_ready')
    @mock.patch('valence.vl_components.vdbench.RunVdb.validate_journal')
    @mock.patch('valence.vl_components.vdbench.RunVdb.finish')
    @mock.patch('valence.vl_components.vdbench.RunVdb.kill')
    @mock.patch('valence.vl_components.vdbench.RunVdb.assert_valid')
    @mock.patch('valence.vl_components.vdbench.RunVdb.update_stream')
    @mock.patch('valence.vl_components.vdbench.RunVdb.wait_for_io_running')
    @mock.patch('valence.vl_components.vdbench.RunVdb.apply_async')
    def test_cmt_management(self, apply_run_cmt, wait_for_run_cmt_io, update_run_cmt_stream,
                            assert_run_cmt_valid, kill_run_cmt, finish_run_cmt, validate_journal_cmt,
                            ensure_run_cmt_ready, store_run_cmt_artifacts, set_test_id):
        manage_runs_cmt = _run.ManageRunVdbs(self.run_cmts)
        manage_runs_cmt.apply_async()

        assert apply_run_cmt.call_count == len(self.run_cmts)
        assert wait_for_run_cmt_io.call_count == len(self.run_cmts)

        manage_runs_cmt.update_stream()
        assert update_run_cmt_stream.call_count == len(self.run_cmts)

        manage_runs_cmt.assert_valid()
        assert assert_run_cmt_valid.call_count == len(self.run_cmts)
        assert_run_cmt_valid.assert_called_with(True)

        manage_runs_cmt.finish(kill=True)
        assert kill_run_cmt.call_count == len(self.run_cmts)
        assert finish_run_cmt.call_count == len(self.run_cmts)

        manage_runs_cmt.validate_journal()
        assert validate_journal_cmt.call_count == len(manage_runs_cmt.child_cmts)

        manage_runs_cmt.assert_valid(False)
        assert assert_run_cmt_valid.call_count == 2 * len(self.run_cmts)
        assert_run_cmt_valid.assert_called_with(False)

        manage_runs_cmt.reset()
        assert ensure_run_cmt_ready.call_count == len(self.run_cmts)

        manage_runs_cmt.store_artifacts()
        assert store_run_cmt_artifacts.call_count == len(self.run_cmts)
        store_run_cmt_artifacts.assert_called_with(get_bundle=True, get_vdb_dir=False)

        manage_runs_cmt.store_artifacts(False)
        assert store_run_cmt_artifacts.call_count == 2 * len(self.run_cmts)
        store_run_cmt_artifacts.assert_called_with(get_bundle=False, get_vdb_dir=False)

        manage_runs_cmt.set_test_identifier('test')
        assert set_test_id.call_count == len(self.run_cmts)

    @mock.patch('valence.vl_components.vdbench.RunVdb._apply_is_running')
    def test_apply_is_running(self, run_cmt_applying):
        run_cmt_applying.return_value = True
        manage_runs_cmt = _run.ManageRunVdbs(self.run_cmts)
        assert manage_runs_cmt._apply_is_running()
        run_cmt_applying.return_value = False
        manage_runs_cmt = _run.ManageRunVdbs(self.run_cmts)
        assert not manage_runs_cmt._apply_is_running()
