from __future__ import absolute_import

import unittest

import pytest

from sf_api import data_size
from valence import (
    components,
    models,
)
from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _enums,
    _parm,
    _setup,
    profiles,
)


class BaseTestSetup(object):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.client_dms = [models.ResourceClient.get_mock_object() for _ in range(3)]
        cls.account_dm = models.Account.create(username='unittest-TAG-0', status='active', account_id=1)
        cls.disk_map = {
            61: '/dev/sdc',
            62: '/dev/sde',
            63: '/dev/sdb',
            64: '/dev/sda',
            65: '/dev/sdg',
            66: '/dev/sdf',
            67: '/dev/sdh',
            68: '/dev/sdi',
            69: '/dev/sdj',
            70: '/dev/sdk'
        }
        cls.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            account_id=1,
            volume_id=i,
            total_size=data_size.DataSize(i, data_size.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
            enable512e=False
        ) for i in cls.disk_map]

        # define the components
        cls.accounts_cmt = _account.MockAccounts('mock', [cls.account_dm])
        cls.volumes_cmts = [
            _volume.MockVolumes('mock-0', cls.cluster_dm, cls.volume_dms[0:1], cls.accounts_cmt),
            _volume.MockVolumes('mock-1', cls.cluster_dm, cls.volume_dms[1:3], cls.accounts_cmt),
            _volume.MockVolumes('mock-2', cls.cluster_dm, cls.volume_dms[3:6], cls.accounts_cmt),
            _volume.MockVolumes('mock-3', cls.cluster_dm, cls.volume_dms[6:], cls.accounts_cmt),
        ]
        cls.conn_cmts = [
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[0],
                volumes_cmts=cls.volumes_cmts[0:2],
            ),
            _connection.MockConnection(
                disk_map=cls.disk_map,
                client_dm=cls.client_dms[1],
                volumes_cmts=cls.volumes_cmts[1:],
            )
        ]
        cls.conn_manager = _connection.ManageConnections(cls.conn_cmts)
        cls.standard_profile = profiles.StandardVdbProfile(cls.conn_manager)


class TestSetupVdbEnvironment(BaseTestSetup, unittest.TestCase):

    def test_master_client_dm(self):
        session = self.standard_profile.get_session()
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager)
        assert setup_cmt.master_client_dm == self.conn_manager.client_dms[0]

        # Ensure a passed-in client supercedes the first connection client as the master client
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager, self.client_dms[2])
        assert setup_cmt.master_client_dm == self.client_dms[2]

        # Ensure masterless sessions cannot be assigned masters intentionally
        standard_profile = profiles.StandardVdbProfile(self.conn_cmts[0])
        session = standard_profile.get_session(standalone=True)
        with pytest.raises(ValueError):
            setup_cmt = _setup.SetupVdbEnvironment(
                io_session=session,
                connections_cmt=self.conn_manager,
                master_client_dm=self.client_dms[2]
            )

    def test_client_dms(self):
        session = self.standard_profile.get_session()
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager)
        assert setup_cmt.client_dms == self.conn_manager.client_dms

        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager, self.client_dms[2])
        assert setup_cmt.client_dms == self.conn_manager.client_dms

    def test_defined_intervals(self):
        session = self.standard_profile.get_session()
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager)
        assert setup_cmt.defined_intervals == [period.config.interval for period in session.io_periods]

    def test_journal_paths(self):
        session = self.standard_profile.get_session()
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager)
        assert not setup_cmt.journal_paths

        # Ensure journal paths aren't generated until vdb_parm is generated
        setup_cmt.apply()
        assert not setup_cmt.journal_paths

        # Ensure journal paths aren't generated if journaling is none
        setup_cmt.get_vdb_parm(journaling_config=_enums.JournalingConfig.none)
        assert not setup_cmt.journal_paths

        # Ensure journal paths are generated for each storage definition
        setup_cmt.get_vdb_parm(journaling_config=_enums.JournalingConfig.full)
        assert len(setup_cmt.journal_paths) == 12
        path_client_names = [path.split('/')[-2] for path in setup_cmt.journal_paths]
        assert all([name == setup_cmt.master_client_dm.client_name for name in path_client_names])

        # Ensure journaling can be reset
        setup_cmt.get_vdb_parm(journaling_config=_enums.JournalingConfig.none)
        assert not setup_cmt.journal_paths

    def test_get_vdb_parm(self):
        session = self.standard_profile.get_session()
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager)
        with pytest.raises(components.ComponentStateError):
            vdb_parm = setup_cmt.get_vdb_parm()

        setup_cmt.apply()
        vdb_parm = setup_cmt.get_vdb_parm()
        assert vdb_parm == _parm.get_general_def(session, self.conn_manager).get_parm_str()

        error_injection_config = _parm.ErrorInjectionConfig(2)
        reporting_config = _enums.ReportingConfig.slaves
        validation_config = _enums.ValidationConfig.on_writes
        journaling_config = _enums.JournalingConfig.recover
        vdb_parm = setup_cmt.get_vdb_parm(
            port=42,
            data_errors=3,
            error_injection_config=error_injection_config,
            reporting_config=reporting_config,
            validation_config=validation_config,
            journaling_config=journaling_config
        )
        assert '{}=42'.format(_parm.PORT) in vdb_parm
        assert '{}=3'.format(_parm.DATA_ERRORS) in vdb_parm
        assert error_injection_config.get_parm_str() in vdb_parm
        assert reporting_config.value in vdb_parm
        assert validation_config.value in vdb_parm
        assert journaling_config.value in vdb_parm

    def test_apply(self):
        session = self.standard_profile.get_session()
        setup_cmt = _setup.SetupVdbEnvironment(session, self.conn_manager)
        with pytest.raises(components.ComponentStateError):
            volume_dms = setup_cmt.volume_dms

        # Ensure the provided connections component is applied alongside the setup component
        assert self.conn_manager.state == components.States.ready
        setup_cmt.apply()
        assert setup_cmt._general_def
        assert self.conn_manager.state == components.States.applied
        assert len(setup_cmt.volume_dms) == len(self.volume_dms)
        assert all(dm in self.volume_dms for dm in setup_cmt.volume_dms)

        # Ensure the provided connections component is NOT reset alongside the setup component
        setup_cmt.reset()
        assert setup_cmt._general_def is None
        assert self.conn_manager.state == components.States.applied


class TestManageVdbEnvironments(BaseTestSetup, unittest.TestCase):
    def setUp(self):
        super(TestManageVdbEnvironments, self).setUp()
        session = self.standard_profile.get_session()
        self.vdb_env_cmts = [
            _setup.SetupVdbEnvironment(session, self.conn_manager),
            _setup.SetupVdbEnvironment(session, self.conn_manager),
            _setup.SetupVdbEnvironment(session, self.conn_manager)
        ]

    def test_child_cmts(self):
        manage_vdb_env_cmt = _setup.ManageVdbEnvironments(
            vdb_environment_cmts=self.vdb_env_cmts,
            connection_cmts=[self.conn_manager]
        )
        assert manage_vdb_env_cmt.child_cmts == self.vdb_env_cmts

    def test_client_dms(self):
        manage_vdb_env_cmt = _setup.ManageVdbEnvironments(
            vdb_environment_cmts=self.vdb_env_cmts,
            connection_cmts=[self.conn_manager]
        )
        assert len(manage_vdb_env_cmt.client_dms) == len(self.conn_manager.client_dms)
        assert all(dm in self.conn_manager.client_dms for dm in manage_vdb_env_cmt.client_dms)

    def test_apply(self):
        manage_vdb_env_cmt = _setup.ManageVdbEnvironments(
            vdb_environment_cmts=self.vdb_env_cmts,
            connection_cmts=[self.conn_manager]
        )
        with pytest.raises(components.ComponentStateError):
            volume_dms = manage_vdb_env_cmt.volume_dms

        # Ensure the provided connections component is applied alongside the setup component
        for cmt in self.conn_cmts:
            assert cmt.state == components.States.ready
        for cmt in self.vdb_env_cmts:
            assert cmt.state == components.States.ready
        manage_vdb_env_cmt.apply()
        assert len(manage_vdb_env_cmt.volume_dms) == len(self.volume_dms)
        for cmt in self.conn_cmts:
            assert cmt.state == components.States.applied
        for cmt in self.vdb_env_cmts:
            assert cmt.state == components.States.applied

        # Ensure the provided connections component is reset alongside the setup component
        manage_vdb_env_cmt.reset()
        for cmt in self.conn_cmts:
            assert cmt.state == components.States.ready
        for cmt in self.vdb_env_cmts:
            assert cmt.state == components.States.ready
