import copy
import unittest

import mock
import pytest
import rstr

from camelot.models import at2_enums
from sf_api import data_size
from valence import models
from valence.vl_components import (
    _account,
    _connection,
    _volume,
)
from valence.vl_components.vdbench import (
    _errors,
    _log_parser,
    _run,
    _setup,
    _stream_handlers,
    profiles,
)


class BaseTestHandler(object):
    def setUp(self):
        self.cluster_dm = models.ResourceCluster.get_mock_object()
        self.client_dms = [
            models.ResourceClient.get_mock_object(overrides=dict(os=at2_enums.ClientOS.ubuntu_14_04))
            for _ in range(3)
        ]
        self.account_dm = models.Account.create(username='unittest-TAG-0', status='active', account_id=1)
        self.disk_map = {
            61: '/dev/sdc',
            62: '/dev/sde',
            63: '/dev/sdb',
            64: '/dev/sda',
            65: '/dev/sdg',
            66: '/dev/sdf',
            67: '/dev/sdh',
            68: '/dev/sdi',
            69: '/dev/sdj',
            70: '/dev/sdk'
        }
        self.volume_dms = [models.Volume.create(
            name='unittest-TAG-{}'.format(i),
            account_id=1,
            volume_id=i,
            total_size=data_size.DataSize(i, data_size.DataUnit.gibibytes),
            qos=models.Qos.create(min_iops=50, max_iops=15000, burst_iops=15000),
            enable512e=False
        ) for i in self.disk_map]

        # define the components
        self.accounts_cmt = _account.MockAccounts('mock', [self.account_dm])
        self.volumes_cmts = [
            _volume.MockVolumes('mock-0', self.cluster_dm, self.volume_dms[0:1], self.accounts_cmt),
            _volume.MockVolumes('mock-1', self.cluster_dm, self.volume_dms[1:3], self.accounts_cmt),
            _volume.MockVolumes('mock-2', self.cluster_dm, self.volume_dms[3:6], self.accounts_cmt),
            _volume.MockVolumes('mock-3', self.cluster_dm, self.volume_dms[6:], self.accounts_cmt),
        ]
        self.conn_cmts = [
            _connection.MockConnection(
                disk_map=self.disk_map,
                client_dm=self.client_dms[0],
                volumes_cmts=self.volumes_cmts[0:2],
            ),
            _connection.MockConnection(
                disk_map=self.disk_map,
                client_dm=self.client_dms[1],
                volumes_cmts=self.volumes_cmts[1:],
            )
        ]
        self.conn_manager = _connection.ManageConnections(self.conn_cmts)
        self.standard_profile = profiles.StandardVdbProfile(self.conn_manager)
        self.io_session = self.standard_profile.get_session()
        self.setup_vdb_env_cmt = _setup.SetupVdbEnvironment(self.io_session, self.conn_manager)
        self.run_cmt = _run.RunVdb(self.setup_vdb_env_cmt)

        self.mock_vdb_log = [
            rstr.xeger(_log_parser.RUN_START_LINE),
            rstr.xeger(_log_parser.IO_LINE),
            rstr.xeger(_log_parser.IO_LINE),
            rstr.xeger(_log_parser.IO_LINE),
            rstr.xeger(_log_parser.IO_LINE)
        ]


class TestErrorStreamHandler(BaseTestHandler, unittest.TestCase):

    def test_init(self):
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        assert error_handler.eio_count == 0
        assert len(error_handler.critical_errors) == 0
        assert error_handler in self.run_cmt._stream_handlers.values()

    def test_handle_items(self):
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        mock_vdb_log = copy.copy(self.mock_vdb_log)
        mock_vdb_log.append(rstr.xeger(_log_parser.EIO_LINE))
        mock_log_lines = _log_parser.parse_raw(mock_vdb_log, [1])
        error_handler.handle_items(mock_log_lines)
        assert error_handler.eio_count == 1

        # Java runtime error
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        mock_vdb_log = copy.copy(self.mock_vdb_log)
        mock_vdb_log.append(rstr.xeger(_log_parser.JAVA_RUNTIME_ERROR_LINE))
        mock_log_lines = _log_parser.parse_raw(mock_vdb_log, [1])
        error_handler.handle_items(mock_log_lines)
        assert len(error_handler.critical_errors) == 1

        # Premature termination
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        mock_vdb_log = copy.copy(self.mock_vdb_log)
        mock_vdb_log.append(rstr.xeger(_log_parser.PREMATURE_TERMINATION_LINE))
        mock_log_lines = _log_parser.parse_raw(mock_vdb_log, [1])
        error_handler.handle_items(mock_log_lines)
        assert len(error_handler.critical_errors) == 1

        # Premature termination (no read validation)
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        error_handler.no_read_validations.add(
            _log_parser.PREMATURE_TERMINATION_LINE.match(mock_log_lines[-1].raw).group(1)
        )
        error_handler.handle_items(mock_log_lines)
        assert len(error_handler.critical_errors) == 0

    @mock.patch('valence.vl_components.vdbench.RunVdb.update_stream')
    def test_assert_valid(self, update_stream):
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        error_handler.eio_count = 1
        error_handler.assert_valid(1)
        assert update_stream.call_count == 1

        error_handler.eio_count = 2
        with pytest.raises(_errors.VdbenchRuntimeError):
            error_handler.assert_valid(1)

        error_handler.eio_count = 0
        error_handler.critical_errors = ['error']
        with pytest.raises(_errors.VdbenchRuntimeError):
            error_handler.assert_valid(0)

        error_handler.destroy()
        with pytest.raises(_errors.StreamHandlerDestroyed):
            error_handler.assert_valid(0)

    def test_clear(self):
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        error_handler.eio_count = 1
        error_handler.critical_errors = ['error']

        error_handler.clear()
        assert error_handler.eio_count == 0
        assert len(error_handler.critical_errors) == 0

    def test_destroy(self):
        error_handler = _stream_handlers.ErrorStreamHandler(self.run_cmt)
        error_handler.destroy()
        assert error_handler not in self.run_cmt._stream_handlers.values()
        assert error_handler.run_vdb_cmt is None


class TestRunInfoHandler(BaseTestHandler, unittest.TestCase):

    def test_init(self):
        run_info_handler = _stream_handlers.RunInfoHandler(self.run_cmt)
        assert run_info_handler.current_io_line_item is None
        assert run_info_handler in self.run_cmt._stream_handlers.values()

    def test_handle_items(self):
        run_info_handler = _stream_handlers.RunInfoHandler(self.run_cmt)
        mock_vdb_log = copy.copy(self.mock_vdb_log)
        mock_vdb_log.append(rstr.xeger(_log_parser.EIO_LINE))
        mock_log_lines = _log_parser.parse_raw(mock_vdb_log, [1])
        run_info_handler.handle_items(mock_log_lines)
        assert run_info_handler.current_io_line_item is mock_log_lines[-2]

    def test_clear(self):
        run_info_handler = _stream_handlers.RunInfoHandler(self.run_cmt)
        mock_log_lines = _log_parser.parse_raw(self.mock_vdb_log, [1])
        run_info_handler.current_io_line_item = mock_log_lines[-1]

        run_info_handler.clear()
        assert run_info_handler.current_io_line_item is None

    def test_destroy(self):
        run_info_handler = _stream_handlers.RunInfoHandler(self.run_cmt)
        run_info_handler.destroy()
        assert run_info_handler not in self.run_cmt._stream_handlers.values()
        assert run_info_handler.run_vdb_cmt is None
