from __future__ import (
    absolute_import,
    division,
)

import unittest

import pytest

from valence.vl_components import vdbench


class TestAccessRangeConfig(unittest.TestCase):
    def test_bytes(self):
        access_range_config = vdbench.AccessRangeBytes(101, 325)
        assert access_range_config.get_parm_str() == 'range=(101,325)'

        with pytest.raises(AssertionError):
            access_range_config = vdbench.AccessRangeBytes(100, 325)

        with pytest.raises(AssertionError):
            access_range_config = vdbench.AccessRangeBytes(101, 101)

    def test_percentages(self):
        access_range_config = vdbench.AccessRangePercentages(76, 100)
        assert access_range_config.get_parm_str() == 'range=(76,100)'

        with pytest.raises(AssertionError):
            access_range_config = vdbench.AccessRangePercentages(76, 101)

        with pytest.raises(AssertionError):
            access_range_config = vdbench.AccessRangePercentages(76, 75)


class TestCacheHitConfig(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hit_config = vdbench.CacheHitConfig(hit_area=1200, write_perc=80)

    def test_min_spec(self):
        with pytest.raises(AssertionError):
            hit_config = vdbench.CacheHitConfig()

    def test_eq(self):
        other_hit_config = vdbench.CacheHitConfig(hit_area=1200, write_perc=81)
        assert other_hit_config != self.hit_config
        other_hit_config = vdbench.CacheHitConfig(hit_area=1200, write_perc=80)
        assert other_hit_config == self.hit_config

    def test_hash(self):
        identical_hit_config = vdbench.CacheHitConfig(hit_area=1200, write_perc=80)
        other_hit_config = vdbench.CacheHitConfig(hit_area=1200, write_perc=81)
        workload_config_dict = {self.hit_config: 1, other_hit_config: 2}
        workload_config_dict[identical_hit_config] = 3
        assert workload_config_dict[self.hit_config] == 3

    def test_repr(self):
        assert str(self.hit_config) == 'CacheHitConfig<hit_area=1200, read_perc=None, write_perc=80>'


class TestDedupConfig(unittest.TestCase):
    def test_parm(self):
        dedup_config = vdbench.DedupConfig(1.5, dedup_unit=64 * 1024)
        assert dedup_config.get_parm_str() == 'dedupratio=1.5,dedupunit=64k,dedupsets=5%'

    def test_bad_parm(self):
        with self.assertRaises(ValueError) as context:
            dedup_config = vdbench.DedupConfig(1.5, dedup_unit=64 * 1024, dedup_src_perc=-5)
            msg = 'Vdbench requires a dedup sets percentage between 0 and 100!'
            self.assertTrue(msg in context.exception)


class TestErrorInjectionConfig(unittest.TestCase):
    def test_parm(self):
        error_injection_config = vdbench.ErrorInjectionConfig(42)
        assert error_injection_config.get_parm_str() == 'force_error_after=(42,1)'


class TestForReadPercentConfig(unittest.TestCase):
    def test_fixed(self):
        for_read_perc_config = vdbench.ForReadPercentFixed(25)
        assert for_read_perc_config.get_parm_str() == 'forrdpct=25'

    def test_range(self):
        for_read_perc_config = vdbench.ForReadPercentRange(25, 75, 7)
        assert for_read_perc_config.get_parm_str() == 'forrdpct=(25-75,7)'


class TestForSeekPercentConfig(unittest.TestCase):
    def test_fixed(self):
        for_seek_perc_config = vdbench.ForSeekPercentFixed(25)
        assert for_seek_perc_config.get_parm_str() == 'forseekpct=25'

    def test_range(self):
        for_seek_perc_config = vdbench.ForSeekPercentRange(25, 75, 7)
        assert for_seek_perc_config.get_parm_str() == 'forseekpct=(25-75,7)'


class TestForThreadsConfig(unittest.TestCase):
    def test_fixed(self):
        for_threads_config = vdbench.ForThreadsFixed(6)
        assert for_threads_config.get_parm_str() == 'forthreads=6'

    def test_series(self):
        for_threads_config = vdbench.ForThreadsSeries([6, 7, 8])
        assert for_threads_config.get_parm_str() == 'forthreads=(6,7,8)'

    def test_range(self):
        for_threads_config = vdbench.ForThreadsRange(6, 11, 2)
        assert for_threads_config.get_parm_str() == 'forthreads=(6-11,2)'


class TestForTransferSizeConfig(unittest.TestCase):
    def test_fixed(self):
        for_xfersize_config = vdbench.ForTransferSizeFixed(512)
        assert for_xfersize_config.xfersizes == [512]
        assert for_xfersize_config.get_parm_str() == 'forxfersize=0.5k'

    def test_series(self):
        for_xfersize_config = vdbench.ForTransferSizeSeries([512, 1024, 4096])
        assert for_xfersize_config.xfersizes == [512, 1024, 4096]
        assert for_xfersize_config.get_parm_str() == 'forxfersize=(0.5k,1k,4k)'

    def test_range(self):
        for_xfersize_config = vdbench.ForTransferSizeRange(4096, 8192, 1024)
        assert for_xfersize_config.xfersizes == [i * 1024 for i in range(4, 8)]
        assert for_xfersize_config.get_parm_str() == 'forxfersize=(4k-8k,1k)'


class TestIoRateConfig(unittest.TestCase):
    def test_curve(self):
        io_rate_config = vdbench.IoRateCurve((55, 70, 100))
        assert io_rate_config.get_parm_str() == 'iorate=curve,curve=(55,70,100),distribution=exponential'

        with pytest.raises(AssertionError):
            io_rate_config = vdbench.IoRateCurve((55, 70, 101))

    def test_fixed(self):
        io_rate_config = vdbench.IoRateFixed(1000, distribution=vdbench.IoDistribution.uniform)
        assert io_rate_config.get_parm_str() == 'iorate=1000,distribution=uniform'

    def test_max(self):
        io_rate_config = vdbench.IoRateMax(distribution=vdbench.IoDistribution.deterministic)
        assert io_rate_config.get_parm_str() == 'iorate=max,distribution=deterministic'

    def test_series(self):
        io_rate_config = vdbench.IoRateSeries([1000, 1200, 1500])
        assert io_rate_config.get_parm_str() == 'iorate=(1000,1200,1500),distribution=exponential'

    def test_variable(self):
        io_rate_config = vdbench.IoRateVariable([(1000, 5), (1200, 12)])
        assert io_rate_config.get_parm_str() == 'iorate=(1000,5,1200,12),distribution=variable'


class TestSeekConfig(unittest.TestCase):
    def test_parm(self):
        seek_config = vdbench.SeekConfig(100, stride=(77, 100))
        assert seek_config.get_parm_str() == 'seekpct=100,stride=(77,100)'

        with pytest.raises(AssertionError):
            seek_config = vdbench.SeekConfig(101, stride=(77, 100))

        with pytest.raises(AssertionError):
            seek_config = vdbench.SeekConfig(100, stride=(77, 101))


class TestStreamsConfig(unittest.TestCase):
    def test_parm(self):
        streams_config = vdbench.StreamsConfig(3)
        assert streams_config.get_parm_str() == 'streams=3'

        streams_config = vdbench.StreamsConfig(3, size=8)
        assert streams_config.get_parm_str() == 'streams=(3,8k)'


class TestTransferSizeConfig(unittest.TestCase):
    def test_aligned(self):
        xfersize_config = vdbench.TransferSizeAligned(512, 4096, 512)
        assert xfersize_config.xfersizes == [i * 512 for i in range(1, 8)]
        assert xfersize_config.get_parm_str() == 'xfersize=(0.5k,4k,512)'

    def test_fixed(self):
        xfersize_config = vdbench.TransferSizeFixed(2048)
        assert xfersize_config.xfersizes == [2048]
        assert xfersize_config.get_parm_str() == 'xfersize=2k'

    def test_variable(self):
        xfersize_config = vdbench.TransferSizeVariable([(512, 80), (4096, 20)])
        assert xfersize_config.xfersizes == [512, 4096]
        assert xfersize_config.get_parm_str() == 'xfersize=(0.5k,80,4k,20)'

        with pytest.raises(AssertionError):
            xfersize_config = vdbench.TransferSizeVariable([(512, 80), (4096, 21)])

        with pytest.raises(AssertionError):
            xfersize_config = vdbench.TransferSizeVariable([(512, 80), (4096, 19)])
