"""Example continuous components."""

import time

import autologger

from valence import (
    camelot,
    components,
)

_log = autologger.get_sflogging(__name__)


class MockContinuous(components.ContinuousTransient):
    """Log stuff as if we were powering off/on a node.

    :param str name: Name of mock operation as it should appear in the log.
    :param int operation_time: Time the mock operation should take.
    :param int kill_after_seconds: Number of seconds before continuous operation is killed.
    :param int sleep_time: Seconds to sleep between each iteration.
    """
    def __init__(self, name, operation_time=0, kill_after_seconds=None, sleep_time=None):
        super(MockContinuous, self).__init__(kill_after_seconds=kill_after_seconds,
                                             sleep_time=sleep_time)
        self.name = name
        self.operation_time = operation_time
        _log.info('Initialized component {}'.format(name))

    def _apply_continuous(self):
        """Power off a node then power it on."""
        _log.info('Applying {}'.format(self.name))
        time.sleep(self.operation_time)
        _log.info('Resetting {}'.format(self.name))

    def _reset(self, timeout=None):
        """Reset the node to its initial state.

        :param int timeout: Not needed for this cmt.
        """


class TcContinusousCmts(camelot.TestCase):
    """Example test case that makes use of continuous components."""
    def __init__(self):
        super(TcContinusousCmts, self).__init__()
        self.cont_clone_volumes = MockContinuous('clone volumes',
                                                 operation_time=10,
                                                 sleep_time=5)
        self.cont_snapshot_volumes = MockContinuous('snapshot volumes',
                                                    operation_time=5,
                                                    sleep_time=1)
        self.cont_bs_check = MockContinuous('run bscheck',
                                            operation_time=30)
        self.cont_reboot_cm = MockContinuous('reboot cluster master',
                                             operation_time=15,
                                             sleep_time=3)

    def _feature_specific_stuff(self):
        """Do feature specific stuff."""
        time.sleep(10)

    def _test(self):
        """Test Continuous components."""
        self._log_step('Begin continuously running cluster bscheck')
        self.cont_bs_check.apply_async()

        self._log_step('Begin continuously cloning volumes')
        self.cont_clone_volumes.apply_async()

        self._log_step('Begin continuously snapshotting volumes')
        self.cont_snapshot_volumes.apply_async()

        self._log_step('Begin continuously rebooting cluster master')
        self.cont_reboot_cm.apply_async()

        self._log_step('Exercise my feature')
        self._feature_specific_stuff()

        self._log_step('Stop continuously rebooting cluster master')
        self.cont_reboot_cm.finish()
        self.cont_reboot_cm.reset()

        self._log_step('Stop continuously snapshotting volumes')
        self.cont_snapshot_volumes.finish()
        self.cont_snapshot_volumes.reset()

        self._log_step('Stop continuously cloning volumes')
        self.cont_clone_volumes.finish()
        self.cont_clone_volumes.reset()

        self._log_step('Stop continuously running cluster bscheck')
        self.cont_bs_check.finish()
        self.cont_bs_check.reset()


class TcContinuousCmtsContextMgr(TcContinusousCmts):
    """Example that uses continuous components as context managers."""
    def _test(self):
        """Test Continuous component context manager."""
        self._log_step('Exercise feature while bscheck, cloning, snapshots, and cm_reboot run '
                       'continuously')
        with self.cont_bs_check:
            # Do some stuff
            with self.cont_clone_volumes, self.cont_snapshot_volumes:
                # Do other stuff
                with self.cont_reboot_cm:
                    self._feature_specific_stuff()


class TsContinuousCmts(camelot.TestSuite):
    """Test suite that exercises continuous components."""
    def __init__(self, resources, config):
        super(TsContinuousCmts, self).__init__(resources=resources, config=config)
        self.add_tests([TcContinusousCmts()])

    def setup(self):
        """Setup the test suite."""
