"""Valence helpers, or just "helpers", are functions that are used by test components and test cases.

Helpers are typically used to combine the functionality of multiple functional models. Helpers often
parallelize calls across multiple functional models using threads.

This is the preferred way to import helpers::

    from valence import vl_helpers

All public helpers should be imported in this file.
"""
from valence.vl_helpers._account import (
    create_account,
    create_accounts,
    delete_accounts,
)
from valence.vl_helpers._bin_sync import (
    get_bin_node_map,
    get_bin_sync_timeout,
    get_bin_sync_timeout_for_nodes,
    get_bins_assigned_to_service,
    is_bin_syncing,
    wait_for_bin_assignments,
    wait_for_bin_sync_to_finish_with_timeout,
    wait_for_bin_sync_to_start,
)
from valence.vl_helpers._bundle import pull_support_bundles
from valence.vl_helpers._capacity import (
    get_volume_count_for_capacity,
    get_volume_size_for_capacity,
    update_cluster_full_threshold,
)
from valence.vl_helpers._cluster import (
    add_nodes,
    assert_cluster_master_moved,
    associate_orphaned_nodes,
    check_storage_network,
    enable_feature,
    is_cluster_idle,
    is_pcache_empty,
    orphan_active_nodes,
    orphan_pending_nodes,
    recreate_cluster,
    remove_nodes,
    remove_nodes_and_drives,
    store_reports,
    validate_fips_drives_report,
    wait_for_api_up,
    wait_for_bin_and_slice_sync_complete,
    wait_for_cluster_idle,
    wait_for_garbage_collection_to_complete,
    wait_for_healthy_services,
)
from valence.vl_helpers._constants import DEFAULT_HARDWARE_FAULTS_TO_IGNORE
from valence.vl_helpers._data_size import (
    get_block_metadata_size,
    get_block_overhead_factor,
    get_block_volume_size,
    get_metadata_block_size,
    get_unusable_space_factor,
    get_volume_block_size,
)
from valence.vl_helpers._drive import (
    add_available_drives,
    add_drives,
    check_fake_drives_flag_on_node,
    fail_drives_in_cluster,
    get_active_node_dm_from_drive_dm,
    get_default_slice_drive_slot,
    get_drives_without_bins,
    get_unexpected_drives_without_slice_services,
    is_mdss_enabled,
    recover_failed_drives,
    remove_drives,
    restore_failed_drives,
    set_drives_type,
    set_fake_drives_flag_on_node,
    wait_for_any_drive_statuses,
    wait_for_drive_assignments,
    wait_for_drive_services_healthy,
    wait_for_drive_status,
    wait_for_drive_statuses,
    wait_for_services_to_decommission,
    wait_for_sync_out,
    wait_for_sync_in,
)
from valence.vl_helpers._errors import (
    Error,
    HealthtoolsVersionNotFound,
    HelperError,
)
from valence.vl_helpers._events import (
    get_next_event_id,
    wait_for_event,
)
from valence.vl_helpers._faults import (
    wait_for_cluster_fault_occurrence,
    wait_for_cluster_fault_resolution,
)
from valence.vl_helpers._healthtools_versions_for_upgrade_suites import HealthtoolsVersionsForTest
from valence.vl_helpers._mnode import (
    configure_mnode2_0_for_afa,
    setup_mnode_storage_service_for_upgrades,
    get_published_mnode_services_bundle_url,
    get_service_bundle_data_from_url,
)
from valence.vl_helpers._node import (
    add_routes_to_nodes,
    api_up_on_nodes,
    check_for_core_and_crash_dumps,
    convert_pending_nodes_to_active_nodes,
    create_cluster,
    create_witness_model,
    create_witness_nodes,
    delete_witness_vms,
    deploy_witness_vm,
    down_network_interfaces,
    flash_and_verify_nvdimm_fw,
    get_ensemble_config_on_nodes,
    get_file_contents_from_node,
    get_next_available_ip,
    get_ova_url,
    list_nodes_by_usable_block_capacity,
    power_off_nodes_via_oob,
    power_on_nodes_via_oob,
    provision_nodes,
    reboot_nodes_via_oob,
    reload_nginx,
    remove_routes_from_nodes,
    remove_user_certs,
    reset_nodes,
    revert_cert_via_cli,
    sort_nodes_by_size,
    test_hardware_config,
    up_network_interfaces,
    wait_for_bootstrap,
    wait_for_node_api_up,
)
from valence.vl_helpers._preupgrade_healthcheck import run_preupgrade_healthchecks
from valence.vl_helpers._remote_replication import (
    add_netem_rules_inter_nodes,
    clear_netem_rules,
    get_group_snapshots_remote_statuses,
    get_peer_group_snapshot,
    get_peer_snapshot,
    get_snapshots_remote_statuses,
    wait_for_group_snapshots_replication,
    wait_for_snapshots_replication,
)
from valence.vl_helpers._security import (
    corrupt_cryptsoft_key,
    corrupt_shamir_secret_keys,
    create_key_server,
    deploy_cryptsoft_eks,
    disable_encryption_at_rest,
    enable_encryption_at_rest,
    get_eks_hostnames,
    is_clustered_eks,
    is_encryption_at_rest_enabled,
    move_shamir_secret_keys,
    restore_shamir_secret_keys,
    setup_external_key_management,
    teardown_external_key_management,
    verify_shamir_secret_keys,
    wait_for_drive_locked_state,
    wait_for_drive_security_state,
    wait_for_ear_state,
)
from valence.vl_helpers._sfupdate_healthtools import (
    get_expected_platform_string_for_healthtools,
    match_sfupdate_healthtools_output_to_regex,
)
from valence.vl_helpers._sfupgradecheck import (
    deploy_different_versions_to_nodes,
    deploy_same_version_to_nodes,
    parse_sfupgradecheck_v_output,
    prettyprint_sfupgradecheck_json_to_text,
    validate_sfupgradecheck_download,
    validate_sfupgradecheck_v_output,
)
from valence.vl_helpers._sfupgradecheck_validation import (
    CheckCriteria,
    validate_sfupgradecheck_cluster_results,
    validate_sfupgradecheck_node_check_results,
    validate_sfupgradecheck_results,
)
from valence.vl_helpers._slice import (
    calculate_slice_sync_timeout,
    create_slice_node_map,
    fill_volumes,
    get_slices_assigned_to_services,
    is_slice_syncing,
    reboot_slice_services,
    wait_for_slice_sync_to_finish,
    wait_for_slice_sync_to_start,
)
from valence.vl_helpers._slice_assignments import (
    check_slice_health_for_volume,
    check_slice_report_health,
    check_standby_service_state,
    check_standby_state_for_volume,
    check_volume_standbys_assigned,
    check_volume_standbys_unassigned,
    get_service_ids,
    get_slice_dm_from_volume,
    get_slice_services_for_volume,
    get_standby_service_id_for_volume,
    wait_for_desired_primary_to_resolve,
    wait_for_slice_health,
    wait_for_standby_state,
    wait_for_volume_standbys_assigned,
    wait_for_volume_standbys_unassigned,
)
from valence.vl_helpers._volume import (
    check_lbas_at_rest,
    create_volumes_individual_profiles,
    create_volumes_single_profile,
    delete_and_purge_volumes,
    purge_deleted_volumes,
    verify_volume_checksums,
)
