"""Helper functions for account methods."""
import time

import autologger

import sf_api.api_errors
from valence.vl_helpers import _errors

_log = autologger.get_sflogging()


def create_account(cluster_fm, username, ignore_duplicate_username=False,
                   initiator_secret=None, target_secret=None, attributes=None):
    """A helper that creates an Account with the given primitives.

    .. note::

        There is an inconsistency between valid Account username characters and Volume name characters.
        Valid Account characters are 'abcdefghijklmnopqrstuvwxyz1234567890!@#$%^&*:;,.-' but Volume name
        characters can only be alphanumeric and a dash.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param str username: Account username, must be unique
    :param bool ignore_duplicate_username: If true function returns existing account
    :param str initiator_secret: Account iSCSI initiator secret
    :param str target_secret: Account iSCSI target secret
    :param dict attributes: Any additional key value pairs to be added to the Account

    :return: Account data model
    :rtype: valence.models.Account
    :raises sf_api.errors.DuplicateUsername: If an account with given username already exists
    """
    try:
        result_account_dm = cluster_fm.create_account(username,
                                                      initiator_secret=initiator_secret,
                                                      target_secret=target_secret,
                                                      attributes=attributes)
    except sf_api.api_errors.DuplicateUsername:
        if ignore_duplicate_username:
            _log.debug('Ignoring duplicate username: {}'.format(username))
            result_account_dm = cluster_fm.list_accounts(account_name=username)[0]
        else:
            raise

    return result_account_dm


def create_accounts(cluster_fm, account_dms, ignore_duplicate_username=False):
    """A helper that creates Accounts for each given account dm.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Account] account_dms: Account dm's to create
    :param bool ignore_duplicate_username: use the id of the existing account if it exists

    :return: Account data model
    :rtype: list[valence.models.Account]
    """
    return_account_rs_dms = []

    for account_dm in account_dms:
        return_account_rs_dms.append(
            create_account(cluster_fm, account_dm.username,
                           ignore_duplicate_username=ignore_duplicate_username,
                           initiator_secret=account_dm.initiator_secret,
                           target_secret=account_dm.target_secret,
                           attributes=account_dm.attributes))

    return return_account_rs_dms


def delete_accounts(cluster_fm, account_dms, wait_for_accounts_purge=False):
    """A helper that deletes the Account for each account dm given.

    .. note::

        In most cases Accounts disappear immediately after being deleted, but it could potentially
        take a few minutes to purge. When this happens the Account will still appear via
        ListAccounts, but its status will be 'removed.' If you really care about this, set
        wait_for_accounts_purge to True and this helper will block until the Account is purged. There
        is a five minute timeout until it fails.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param typing.Iterable[valence.models.Account] account_dms: Account dm's to create
    :param bool wait_for_accounts_purge: Wait for "removed" accounts to be purged, should take seconds

    :raises valence.vl_helpers.errors.HelperError: if wait_for_accounts_purge is True and
        the Account takes longer than 5 minutes to purge.
    """
    cluster_fm.delete_accounts(account_dms)
    account_ids = set(a.account_id for a in account_dms)

    if wait_for_accounts_purge:
        _log.debug('Waiting five minutes for Accounts to purge')

        purge_account_dms = [a for a in cluster_fm.list_accounts() if a.account_id in account_ids]
        start_time = time.time()

        while purge_account_dms:
            remaining_account_ids = [a.account_id for a in purge_account_dms]
            _log.debug('Account IDs that still need to purge: [{}]'.format(remaining_account_ids))
            _log.debug('Sleeping for 10 seconds...')
            time.sleep(10)

            purge_account_dms = [a for a in cluster_fm.list_accounts() if a.account_id in account_ids]
            if time.time() - start_time > 300:
                msg = 'Timed out waiting for Accounts to purge: [{}]'.format(
                    [a.account_id for a in purge_account_dms])
                raise _errors.HelperError(msg)

        _log.debug('All Accounts were purged')
