"""Helper functions for bin_sync methods."""
from __future__ import division

import time

import autologger

from valence import (
    enums,
    models,
    utils,
)
from valence.vl_helpers import (
    _errors,
    _events,
)

_log = autologger.get_sflogging(__name__)

DEFAULT_TIMEOUT_IN_SEC = 600


def wait_for_bin_assignments(cluster_fm, timeout=DEFAULT_TIMEOUT_IN_SEC, poll_period=20,
                             start_event_id=None):
    """Wait for bin assignments event.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: Time to wait in seconds before raising a timeout exception, Defaults to 600 sec.
    :param int poll_period: Time to wait in seconds between each iteration, Defaults to 20 sec.
    :param int start_event_id: Event ID to start from, in case of timing issues you may need to keep
        track of the last event_id prior to perturbing the system.
    :rtype: list[valence.models.ClusterEvent]
    :returns: list of bin assignments cluster events that have occurred.
    :raises valence.vl_helpers.HelperError: If bin assignments does not start within the given time
    """

    timeout = timeout if timeout is not None else DEFAULT_TIMEOUT_IN_SEC

    return _events.wait_for_event(
        cluster_fm,
        enums.ClusterEventType.service_event,
        message_regexes=[r'^Block data assignment change detected.\s*$'],
        timeout=timeout,
        poll_period=poll_period,
        start_event_id=start_event_id
    )


def is_bin_syncing(cluster_fm):
    """Check if bin syncing is in progress.

    If is_stable is false bin syncing is occurring or about to occur

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :rtype: bool
    """
    bin_assignments = cluster_fm.get_bin_assignments()

    # if the new is_stable flag is available use it.
    if bin_assignments.replica_based:
        return not bin_assignments.properties.is_stable

    # older method for versions not supporting the is_stable flag.
    for _bin in bin_assignments.bins:
        for service in _bin.services:
            if service.status in (enums.BinStatus.updating, enums.BinStatus.pending_active):
                return True

    return False


def wait_for_bin_sync_to_start(cluster_fm, timeout=DEFAULT_TIMEOUT_IN_SEC, poll_period=20):
    """Wait for bin sync to start.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: Time to wait in seconds before raising a timeout exception
    :param int poll_period: Time to wait in seconds between each iteration
    :raises valence.vl_helpers.HelperError: If bin syncing does not start within the given time
    """
    timeout = DEFAULT_TIMEOUT_IN_SEC if timeout is None else timeout
    timer = utils.timeout_timer(timeout)
    while next(timer):
        if is_bin_syncing(cluster_fm):
            return
        _log.debug('Waiting for bin sync to start. Sleeping {} seconds.'.format(poll_period))
        time.sleep(poll_period)

    raise _errors.HelperError(
        'Timed out waiting for bin sync to initialize. All block services are still active after {} '
        'seconds.'.format(timeout))


def wait_for_bin_sync_to_finish_with_timeout(cluster_fm, timeout=None, poll_period=60):
    """Wait for bin sync to finish with a provided timeout.

    This helper has no special logic to determine if bin syncing is still making progress. It will only
    fail if it reaches the huge timeout value.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: Time to wait in seconds before raising a timeout exception
    :param int poll_period: Time to wait in seconds between each iteration
    """
    timeout = 60 * 60 * 12 if timeout is None else timeout
    timer = utils.timeout_timer(timeout)
    start_time = time.time()
    while next(timer):
        if not is_bin_syncing(cluster_fm):
            _log.debug('Bin sync finished after {} seconds.'.format(int(time.time() - start_time)))
            return
        _log.debug('Waiting for bin sync to finish. Sleeping {} seconds.'.format(poll_period))
        time.sleep(poll_period)

    raise _errors.HelperError(
        'Timed out waiting for bin sync to finish. Some block services are not active after {} '
        'seconds.'.format(timeout))


def get_bin_sync_timeout(cluster_fm, added_drive_dms=None, removed_drive_dms=None, tolerance_perc=25):
    """Get the appropriate bin-sync timeout for a given set of drive additions/removals.

    Calculates the maximum amount of time bin syncing should take if the drives in question are at
    100% capacity.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] added_drive_dms: Drives that were added that triggered bin syncing.
    :param list[valence.models.Drive] removed_drive_dms: Drives that were removed (or failed) that
        triggered bin syncing.
    :param int tolerance_perc: Percent tolerance by which to pad the timeout.
    :returns: Maximum number of seconds bin syncing should take for the specified drives.
    :rtype: int
    """
    cluster_constants = cluster_fm.get_constants()
    tolerance_factor = float(100 + tolerance_perc) / 100
    timeout = 0
    if added_drive_dms:
        write_speed = cluster_constants.bin_sync_new_service_write_mib_per_second or 60
        total_size = models.DataSize(sum(dm.usable_capacity for dm in added_drive_dms))
        timeout += int(total_size.mebibytes * tolerance_factor / write_speed)
    if removed_drive_dms:
        write_speed = cluster_constants.bin_sync_write_mib_per_second or 20
        total_size = models.DataSize(sum(dm.usable_capacity for dm in removed_drive_dms))
        timeout += int(total_size.mebibytes * tolerance_factor / write_speed)
    return timeout or 1800


def get_bin_node_map(cluster_fm):
    """Make a map of bin ids to the list of nodeIDs which host block services for the respective bin.

    If a bin with binID 1, has primary block service ID 3 and secondary block service ID 13,
    which live on nodesIDs 1 and 2 resp., the method will return {1: [1, 2]}

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model to work on
    :rtype: dict[int, list[int]]
    """
    _bins = cluster_fm.get_bin_assignments().bins
    all_services = cluster_fm.list_services()
    block_services = [service_info.service for service_info in all_services if
                      service_info.service.service_type is enums.ServiceType.block]
    block_service_node_map = {block_service.service_id: block_service.node_id for
                              block_service in block_services}
    bin_node_map = {b.bin_id: [block_service_node_map[bs.service_id] for bs in b.services] for b in
                    _bins}
    return bin_node_map


def get_bin_sync_timeout_for_nodes(cluster_fm, removed_node_dms=None, added_node_dms=None):
    """Get the appropriate bin-sync timeout for a general case where we expect a bin sync to occur.

    If no node_dms are passed in we will calculate the maximum amount of time bin syncing should take
    if the all the block drives in largest node of the nodes_dms passed in are at 100% capacity.

    If node_dms are passed in we will calculate the maximum amount of time bin syncing should take
    if all the block drives in the nodes are at 100% capacity.

    The difference between the calculation for added nodes versus removing nodes is that there are
    separate cluster constants that determine how fast we bin sync data away
    (cBinSyncNewServiceWriteMiBPerSecond for adding nodes and cBinSyncWriteMiBPerSecond for removing
    nodes).

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.ActiveNode] added_node_dms: Data models of nodes added that
        affect the timeout.
    :param list[valence.models.ActiveNode] removed_node_dms: Data models of nodes removed that
        affect the timeout.
    :rtype: int
    """
    removed_sync_drives = []
    if not added_node_dms and not removed_node_dms:
        msg = ("We didn't pass in node_dms (added or removed) so we assume it's a single node failure "
               '(removal) with the largest node used for the calculation.')
        _log.debug(msg)
        drive_dms = cluster_fm.list_drives(drive_types=[enums.DriveType.block])
        largest_drive_dm = drive_dms[0]

        for d in drive_dms:
            # update the node_id and capacity variables for the largest drive in the list only
            # if it's bigger
            if d.capacity > largest_drive_dm.capacity:
                largest_drive_dm = d

        removed_node_ids = [largest_drive_dm.node_id]
        removed_sync_drives = cluster_fm.list_drives(drive_types=[enums.DriveType.block],
                                                     node_ids=removed_node_ids)

    added_sync_drives = []
    if added_node_dms:
        added_node_ids = [n.node_id for n in added_node_dms]
        added_sync_drives = cluster_fm.list_drives(drive_types=[enums.DriveType.block],
                                                   node_ids=added_node_ids)

    if removed_node_dms:
        removed_node_ids = [n.node_id for n in removed_node_dms]
        if not removed_sync_drives:
            removed_sync_drives = cluster_fm.list_drives(drive_types=[enums.DriveType.block],
                                                         node_ids=removed_node_ids)

    if not added_sync_drives and not removed_sync_drives:
        _log.debug('No drives to be found.  This could be an error.')

    timeout = get_bin_sync_timeout(cluster_fm=cluster_fm, added_drive_dms=added_sync_drives,
                                   removed_drive_dms=removed_sync_drives)
    return timeout


def get_bins_assigned_to_service(bin_assignments, service_ids):
    """Get any bins that are assigned to services.

    :param valence.models.BinAssignments bin_assignments: bin assignments
    :param set[int] service_ids: service ids which should not be in the slice report
    :return: list of invalid bin reports
    :rtype: list[valence.models.BinReport]
    """
    bins_for_service = []
    for b in bin_assignments.bins:
        bins_services = set([s.service_id for s in b.services])
        if any([s in service_ids for s in bins_services]):
            bins_for_service.append(b)
    return bins_for_service
