"""Helper functions for grabbing support bundles."""

import sf_batch


def pull_support_bundles(artifact_manager, node_fms=None, client_fms=None, purge_cores=True,
                         purge_crash_dumps=True, compress=True, collect_backtraces=False,
                         zk_heap_dump=False, verify_logs=False):
    """Grabs support bundles for supplied nodes and/or clients.

    This method will spawn a thread for each functional model provided.
    Assertions raised in a thread will be re-raised via sf_batch.

    :param sf_artifacts.ArtifactManager artifact_manager: Artifact manager object that defines
        where the support bundle will be stored.
    :param typing.Iterable[valence.functional_models.Node] node_fms: node functional models
    :param typing.Iterable[valence.functional_models.LinuxClient] client_fms: client fms
    :param bool purge_cores: If True, purge core files
    :param bool purge_crash_dumps: If True, purge crash dump files
    :param bool compress: If True, pass compress flag to sf_make_support_bundle
    :param bool collect_backtraces: If True, collect backtraces
    :param bool zk_heap_dump: If True, do Zookeeper heap dump
    :param bool verify_logs: Verify support bundle logs
    :raises TypeError: If called without node_fms or client_fms
    """

    node_fms = node_fms or []
    client_fms = client_fms or []

    if not node_fms and not client_fms:
        raise TypeError('Must provide either nodes or clients as input.')

    threads = []

    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.pull_support_bundle, artifact_manager,
                                         purge_cores=purge_cores, purge_crash_dumps=purge_crash_dumps,
                                         compress=compress, collect_backtraces=collect_backtraces,
                                         zk_heap_dump=zk_heap_dump, verify_logs=verify_logs,
                                         thread_name=str(node_fm)))
    for client_fm in client_fms:
        threads.append(sf_batch.SFThread(client_fm.pull_support_bundle, artifact_manager,
                                         thread_name=str(client_fm)))

    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    sf_batch.check_for_error_and_raise_exception(threads)
