"""Helper functions for allocating and filling cluster capacity."""
from __future__ import division

import math

from valence import (
    enums,
    models,
)
from valence.vl_helpers import _bin_sync


def get_volume_count_for_capacity(cluster_fm, block_data_size, default_count=60, volume_size_min=None,
                                  volume_size_max=None, cluster_limits=None):
    """Get a valid number of volumes that can be used to meet the desired capacity.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.DataSize block_data_size: Block data size.
    :param int default_count: Default volume count to use.
    :param valence.models.DataSize volume_size_min: Minimum volume size.
    :param valence.models.DataSize volume_size_max: Maximum volume size.
    :param valence.models.ClusterLimits cluster_limits: Cluster limits model.
    :rtype: int
    """
    if not cluster_limits:
        cluster_limits = cluster_fm.get_limits()

    # Determine min/max volume size
    cluster_vol_size_min = models.DataSize(cluster_limits.volume_size_min)
    if volume_size_min:
        volume_size_min = max(volume_size_min, cluster_vol_size_min)
    else:
        volume_size_min = cluster_vol_size_min

    cluster_vol_size_max = models.DataSize(cluster_limits.volume_size_max)
    if volume_size_max:
        volume_size_max = min(volume_size_max, cluster_vol_size_max)
    else:
        volume_size_max = cluster_vol_size_max

    # Determine min/max volume count based on size constraints
    vol_count_min = int(math.ceil(float(block_data_size.bytes) / volume_size_max.bytes))
    vol_count_max = max(1, block_data_size.bytes // volume_size_min.bytes)

    # Factor in cluster volume count limit
    cluster_vol_count_max = cluster_limits.volume_count_max - len(cluster_fm.list_volumes())
    vol_count_max = min(vol_count_max, cluster_vol_count_max)

    assert vol_count_min <= vol_count_max
    return min(max(default_count, vol_count_min), vol_count_max)


def get_volume_size_for_capacity(cluster_fm, block_data_size, volume_count, under_size=False,
                                 cluster_limits=None):
    """Get the size of volumes required to meet the desired capacity.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.DataSize block_data_size: Block data size.
    :param int volume_count: Number of volumes to use.
    :param bool under_size: Whether to round down (instead of up) to the nearest acceptable size.
    :param valence.models.ClusterLimits cluster_limits: Cluster limits model.
    :rtype: valence.models.DataSize
    """
    if not cluster_limits:
        cluster_limits = cluster_fm.get_limits()

    # Determine min volume size
    min_vol_bytes = int(math.ceil(float(block_data_size.bytes) / volume_count))

    # Factor in cluster volume size limit
    min_vol_bytes = max(min_vol_bytes, cluster_limits.volume_size_min)

    # Round up/down to the nearest mebibyte
    if not under_size or min_vol_bytes == cluster_limits.volume_size_min:
        vol_size = models.DataSize(min_vol_bytes).carbon_size
    else:
        vol_size = models.DataSize(
            size=min_vol_bytes // enums.DataUnit.mebibytes.value[0],
            unit=enums.DataUnit.mebibytes
        )

    assert cluster_limits.volume_size_min <= vol_size.bytes <= cluster_limits.volume_size_max
    return vol_size


def update_cluster_full_threshold(cluster_fm, gc_cmt):
    """Update the cluster-full-threshold values.

    In order to get an accurate measure of cluster capacity, the stats returned by
    GetClusterFullThreshold must be up-to-date.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.vl_components.RunGarbageCollection gc_cmt: Garbage collection component.
    """
    # Update cluster-fullness stats by running garbage collection
    gc_cmt.apply()
    gc_cmt.reset()

    # If necessary, wait for bin syncing to complete as this can prevent cluster-full-threshold
    # from updating
    cannot_sync_faults = cluster_fm.list_cluster_faults(
        fault_status=enums.ClusterFaultStatus.current,
        include_codes=[enums.ClusterFaultCode.cluster_cannot_sync],
        update=True
    )
    # Cannot sync if block-service-unhealthy is critical (e.g. drives failed on different nodes)
    bs_unhealthy_faults = cluster_fm.list_cluster_faults(
        fault_status=enums.ClusterFaultStatus.current,
        include_codes=[enums.ClusterFaultCode.block_service_unhealthy],
        include_severities=[enums.ClusterFaultSeverity.critical],
        update=True
    )
    if not cannot_sync_faults and not bs_unhealthy_faults:
        _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm)

    # Update cluster-full-threshold by updating the cluster faults (post-bin-sync)
    cluster_fm.list_cluster_faults(update=True)
