"""Helper functions for Cluster Functional Models."""
import shutil
import tempfile
import time
from datetime import datetime

import autologger
import pytz
from bs4 import BeautifulSoup

from valence import (
    enums,
    errors,
    functional_models,
    utils,
)
from valence.vl_helpers import (
    _bin_sync,
    _drive,
    _errors,
    _node,
    _slice,
)

_log = autologger.get_sflogging()

NO_CLUSTER = ''
TIMEOUT_TOLERANCE = 1.25


def wait_for_api_up(cluster_fm, timeout=120):
    """Wait up to timeout seconds for mvip to become responsive.

    :param valence.functional_models.Cluster cluster_fm: Cluster to wait for mvip responsiveness on.
    :param int timeout: Seconds to wait before raising.
    :raises: valence.errors.TimeoutError: If the mvip isn't responsive before the timeout expires.
    """
    start_time = time.time()
    while not cluster_fm.api_is_responsive():
        if time.time() > start_time + timeout:
            raise errors.TimeoutError('Timed out waiting for API to become responsive.')
        time.sleep(10)


def add_nodes(cluster_fm, pending_node_dms, timeout=None, force_non_fips=None, auto_install=True):
    """Add nodes.

    This function will block until nodes are added.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.PendingNode] pending_node_dms: Nodes to be added
    :param int timeout: timeout when waiting for completion
    :param bool force_non_fips: Force add non-fips nodes to a fips cluster
    :param bool auto_install: auto install added node to the cluster's current build
    :rtype: list[valence.models.ActiveNode]
    """
    timeout = 600 if timeout is None else timeout
    async_handler = cluster_fm.add_nodes(pending_node_dms, force_non_fips=force_non_fips,
                                         auto_install=auto_install)
    active_node_dms = []
    for handler in async_handler:
        active_node_dms.extend(handler.get_models(blocking=True, timeout=timeout))
    return active_node_dms


def remove_nodes(cluster_fm, node_dms, timeout=None):
    """Remove nodes.

    This function will block until nodes are removed and it assumes that all drives that are associated
    with the nodes have already been removed. The remove_nodes_and_drives helper will remove the
    associated drives for you first.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.ActiveNode] node_dms: Nodes to be removed
    :param int timeout: timeout when waiting for completion
    :rtype: list[valence.models.PendingNode]
    """
    timeout = 600 if timeout is None else timeout
    async_handler = cluster_fm.remove_nodes(node_dms)
    pending_node_dms = []
    for handler in async_handler:
        pending_node_dms.extend(handler.get_models(blocking=True, timeout=timeout))
    return pending_node_dms


def remove_nodes_and_drives(cluster_fm, node_dms, timeout=None):
    """Remove all drives associated with the given nodes, then remove all nodes.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.ActiveNode] node_dms: Nodes to be removed
    :param int timeout: timeout when waiting for completion
    :rtype: list[valence.models.PendingNode]
    """
    node_ids = [n.node_id for n in node_dms]
    drives = cluster_fm.list_drives(node_ids=node_ids, statuses=[enums.DriveStatus.active])
    if drives:
        _drive.remove_drives(cluster_fm, drive_dms=drives, timeout=timeout)
    return remove_nodes(cluster_fm, node_dms=node_dms, timeout=timeout)


def orphan_pending_nodes(cluster_fm, node_dms, timeout=None):
    """Orphan the pending nodes from the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.PendingNode] node_dms: Pending node data models to be orphaned
    :param int timeout: timeout when waiting for completion
    :rtype: list[valence.models.ResourceNode]
    """
    timeout = 600 if timeout is None else timeout
    static_ids = [n.static_id for n in node_dms]
    node_fms = [functional_models.Node(n.resource_dm, cluster_fm.cluster_dm) for n in node_dms]

    for node_fm in node_fms:
        node_fm.set_cluster_config(NO_CLUSTER)

    # Wait for the list of nodes to no longer be pending
    timer = utils.timeout_timer(timeout)
    while next(timer):
        pending = cluster_fm.list_pending_nodes(static_ids=static_ids)
        if not pending:
            return [n.resource_dm for n in node_dms]
        time.sleep(30)

    raise errors.TimeoutError(
        'Timed out waiting for nodes to be orphaned in {} seconds'.format(timeout))


def orphan_active_nodes(cluster_fm, node_dms, timeout=None):
    """Orphan active nodes from the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.ActiveNode] node_dms: active node data models to be orphaned
    :param int timeout: timeout when waiting for completion
    :rtype: list[valence.models.ResourceNode]
    """
    pending_node_dms = remove_nodes_and_drives(cluster_fm, node_dms, timeout)
    return orphan_pending_nodes(cluster_fm, pending_node_dms, timeout)


def associate_orphaned_nodes(cluster_fm, node_dms, timeout=None):
    """Associate the orphaned nodes back to the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.ResourceNode] node_dms: node data models to be associated with cluster
    :param int timeout: timeout when waiting for completion
    :rtype: list[valence.models.PendingNode]
    """
    timeout = 600 if timeout is None else timeout
    static_ids = [n.static_id for n in node_dms]
    cluster_name = cluster_fm.get_cluster_info().name
    node_fms = [functional_models.Node(n, cluster_fm.cluster_dm) for n in node_dms]

    for node_fm in node_fms:
        node_fm.set_cluster_config(cluster_name)

    # Wait for the list of nodes to be pending
    timer = utils.timeout_timer(timeout)
    while next(timer):
        pending_nodes = cluster_fm.list_pending_nodes(static_ids=static_ids)
        if len(pending_nodes) == len(node_fms):
            return pending_nodes
        time.sleep(30)

    raise errors.TimeoutError(
        'Timed out waiting for nodes to be pending in {} seconds.'.format(timeout))


def store_reports(cluster_fm, artifact_manager, include_html_reports=True, include_json_reports=True):
    """Store a cluster's HTML and JSON reports as artifacts.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param sf_artifacts.ArtifactManager artifact_manager: Artifact credentials object
    :param bool include_html_reports: If we want to store the html reports.
    :param bool include_json_reports: If we want to store the json reports.
    """
    temp_dir = tempfile.mkdtemp()
    try:
        tmp_reports = []
        timestamp = datetime.utcnow()
        timestamp = timestamp.replace(tzinfo=pytz.utc).isoformat()
        reports_to_store = []
        if include_html_reports:
            # noinspection PyTypeChecker
            reports_to_store.extend(enums.HtmlReport)
        if include_json_reports:
            # noinspection PyTypeChecker
            reports_to_store.extend(enums.JsonReport)
        for report in reports_to_store:
            output = cluster_fm.api.http_get('reports/{}'.format(report.value)).text
            file_loc = '{}/{}_{}.html'.format(temp_dir, timestamp, report.value)
            with open(file_loc, 'w') as text_file:
                text_file.write(output)
            tmp_reports.append(file_loc)

        artifact_manager.store_artifacts_local(tmp_reports)
    finally:
        shutil.rmtree(temp_dir)


def check_storage_network(cluster_fm, client_fms):
    """On each client, ping and arp the SIP for each active node in a cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.functional_models.Client] client_fms: Clients to check storage connection
    """
    active_nodes = cluster_fm.list_active_nodes()
    sips = [active_node.sip for active_node in active_nodes]
    timeout = 5 * len(active_nodes)

    for client in client_fms:
        client.run_command('for i in {}; do ping -i 0.1 -c 10 $i; done'.format(sips),
                           timeout=timeout, raise_on_failure=True)
        client.run_command('for i in {}; do arp $i; done'.format(sips), timeout=timeout,
                           raise_on_failure=True)
        client.run_command('netstat -re', timeout=10, raise_on_failure=True)


def wait_for_garbage_collection_to_complete(cluster_fm, timeout=30 * 60, poll_period=10):
    """Wait for garbage collection to complete.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param int timeout: The number of seconds for gc to complete before timing out.
    :param int poll_period: The number of seconds between polling intervals.
    """
    start_time = time.time()
    while cluster_fm.get_gc_status().status == enums.GcStatus.in_progress:
        if time.time() - start_time > timeout:
            raise errors.TimeoutError('Timeout of {} hit when waiting for garbage collection'.
                                      format(timeout))
        time.sleep(poll_period)


def is_cluster_idle(cluster_fm):
    """Check whether there is any active IO on the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :rtype: bool
    """
    for node in cluster_fm.list_active_nodes():
        node_fm = functional_models.Node(node.resource_dm, cluster_fm.cluster_dm, authenticate=False)
        drives = cluster_fm.list_drives(node_ids=[node.node_id])

        logical_names = []
        for drive in drives:
            drive_info = cluster_fm.get_drive_hardware_info(drive)
            logical_names.append(drive_info.logicalname)

        cmd = 'iostat -d 5 2 {} | tail -{} | {}'.format(
            ' '.join(logical_names),
            len(logical_names) + 1,
            "awk '{sum+=$5+$6} END {print sum}'"
        )
        with node_fm.get_ssh_connection() as ssh_conn:
            result = ssh_conn.run_command(cmd)
            if int(result.stdout) != 0:
                return False

    return True


def assert_cluster_master_moved(cluster_fm, starting_cm_dm, perturbed_node_dms=None):
    """Assert that the cluster master has moved off of the given node.

    .. note::
        This method will skip the assertion if the starting cluster master is not in a list of perturbed
        nodes.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.ActiveNode starting_cm_dm: the node model for the cluster master before
        the pertrubation
    :param list[valence.models.ActiveNode] perturbed_node_dms: the list of nodes that were perturbed.
        if the starting_cm_dm is not in this list, then this method will not expect
        the cluster master to have moved (since the CM wasn't perturbed)
    :raises: valence.vl_helpers.HelperError: if the CM hasn't moved when expected.
    """

    start_cm_static_id = starting_cm_dm.static_id
    current_cm_static_id = cluster_fm.get_cluster_master().static_id

    perturbed_ids = None
    if perturbed_node_dms:
        perturbed_ids = {n.static_id for n in perturbed_node_dms}

    if (perturbed_ids is not None and
            (start_cm_static_id in perturbed_ids) and
            (current_cm_static_id == start_cm_static_id)):
        raise _errors.HelperError(
            'cluster master did not change, still on node id: {}'.format(start_cm_static_id))


def wait_for_healthy_services(cluster_fm, service_ids=None, timeout=None):
    """Wait for all services (default), or a subset of services to report healthy.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[int] service_ids: ServiceIDs to poll for health.
    :param int timeout: Number of seconds to wait before raising an error.
    """
    timeout = timeout if timeout else (60 * 60)
    service_ids = set(service_ids) if service_ids else set()

    timer = utils.timeout_timer(timeout)
    while next(timer):
        service_status = cluster_fm.get_service_status()
        if not service_status.failed:
            return
        elif service_ids and not service_ids.intersection(set(service_status.failed.keys())):
            return
        else:
            time.sleep(15)
    raise errors.TimeoutError('Timed out waiting for healthy services')


def wait_for_bin_and_slice_sync_complete(cluster_fm, skip_sync_start=False, skip_sync_finish=False):
    """Wait for either Bin or Slice sync to start and both to finish.

    Call wait_for_slice_sync or wait_for_bin_sync if you only want one.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param bool skip_sync_start: Don't wait for bin/slice sync to start. Default is to wait.
    :param bool skip_sync_finish: Don't wait for bin/slice sync to finish. Default is to wait.
    """
    if not skip_sync_start:
        _bin_sync.wait_for_bin_sync_to_start(cluster_fm)
        _slice.wait_for_slice_sync_to_start(cluster_fm)

    if skip_sync_finish:
        return

    bin_sync_in_progress = _bin_sync.is_bin_syncing(cluster_fm)
    slice_sync_in_progress = _slice.is_slice_syncing(cluster_fm)

    if bin_sync_in_progress or slice_sync_in_progress:
        _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm)
        _slice.wait_for_slice_sync_to_finish(cluster_fm)


def recreate_cluster(cluster_fm, dev_image_repo=None):
    """Destroy the existing cluster, then recreate it with the same cluster name and ip addresses.

    Obviously your volumes, accounts, data, etc will not survive.

    :param valence.functional_models.Cluster cluster_fm: Cluster to recreate.
    :param str dev_image_repo: Custom dev_image_repo to use with SfVersion
    """
    node_dms = cluster_fm.list_active_nodes()
    node_fms = [functional_models.Node(n.resource_dm, cluster_fm.cluster_dm) for n in node_dms]
    build_version = node_fms[0].get_version_info().version_info.sfapp.version

    if dev_image_repo:
        build_version.set_dev_repo_name(dev_image_repo)

    rtfi_async_handles = []
    for node_fm in node_fms:
        rtfi_async_handles.append(node_fm.start_rtfi(build_version, skip_post_rtfi_network_config=True))
    for handle in rtfi_async_handles:
        handle.join()

    rs_node_dms = [n.resource_dm for n in node_dms]
    _node.create_cluster(node_fms[0], cluster_fm.mvip, cluster_fm.svip, rs_node_dms, cluster_fm.name)


def enable_feature(cluster_fm, feature):
    """Enable a SolidFire feature that is disabled by default.

    :param valence.functional_models.Cluster cluster_fm: Cluster to operate on.
    :param valence.enums.SfFeature feature: Feature to enable.
    """
    async_handler = cluster_fm.enable_feature(feature=feature)
    async_handler.join()


def validate_fips_drives_report(cluster_fm, node_dms, status):
    """Verify fipsDrives feature report status via GetFipsReport API call.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param bool status: Expected fipsDrives report status
    :param list[valence.models.SfNodeBase] node_dms: List of node models
    :raises: ValueError if any node id is not found or found duplicated ones in fipsDrive report
    :rtype: bool
    """
    reports = cluster_fm.get_fips_report()
    for node_dm in node_dms:
        report_status = [n for n in reports.nodes if n.node_id == node_dm.node_id]
        if len(report_status) == 1:
            if report_status[0].fips_drives != status:
                return False
        else:
            if len(report_status) > 1:
                raise ValueError('Found {} duplicated node ids: {}' .
                                 format(len(report_status), node_dm.node_id))
            else:
                raise ValueError('Not found node id: {}' . format(node_dm.node_id))
    return True


def wait_for_cluster_idle(cluster_fm, timeout):
    """Check whether there is any active IO on the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param int timeout: Time in seconds after which to raise an error.
    """
    cluster_idle = False
    last_iops_count = 0
    time_start = time.time()
    while not cluster_idle:
        if time.time() - time_start > timeout:
            raise errors.TimeoutError(
                'Timed out after {}s waiting for cluster to be idle'.format(timeout)
            )
        else:
            time.sleep(5)
            stats = cluster_fm.get_cluster_stats()
            iops_count = stats.cluster_stats.read_ops + stats.cluster_stats.write_ops
            cluster_idle = (iops_count == last_iops_count)
            last_iops_count = iops_count


def is_pcache_empty(cluster_fm):
    """Check whether the primary cache is empty.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :rtype: bool
    """
    # TODO: ELEM-7279 find a better way to do this once Pcache ActiveSegments is added to a json report

    pcache_empty = True
    ss_report = cluster_fm.get_report_raw(enums.HtmlReport.ss)
    soup = BeautifulSoup(ss_report, 'html.parser')
    # Iterate through all the tables.
    for table in soup.find_all('table'):

        # Look for table headers that don't have colspan in their header
        cache_headers = [header for header in table.find_all('th') if 'colspan' not in header.attrs]

        # Skip the table if it doesn't have correct headers.
        if not cache_headers:
            continue

        # Find the column of the ActiveSegments header
        active_segment_index = [i for i, e in enumerate(cache_headers) if e.text == 'ActiveSegments']

        # If the column list is empty, it doesn't have an ActiveSegment in the table; skip it.
        if not active_segment_index:
            continue
        else:
            # If it does have a list, the first item in the list is the column number we want.
            assert len(active_segment_index) == 1
            active_segment_index = active_segment_index[0]

        # Now get the table data information, and grab the value at the column index point we want.
        data_row = table.find_all('td')

        active_segments = int(data_row[active_segment_index].text)
        if active_segments != 0:
            pcache_empty = False
            break

    return pcache_empty
