"""Helper functions for data size conversions."""
import math

from valence import models


def get_block_overhead_factor(cluster_fm, block_size=4096, immutable_values=None):
    """Get the factor of actual block data to block data representing volume data.

    Every block contains a certain amount of overhead data which does not represent actual volume data.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param int block_size: The size of blocks on the cluster (typically 4k, but impacted by compression).
    :param sf_api.result_models.GetImmutableValues immutable_values: Immutable cluster values.
    :rtype: float
    """
    if not immutable_values:
        immutable_values = cluster_fm.get_immutable_values()

    overhead_bytes = immutable_values.per_block_overhead_bytes
    return 1. + float(overhead_bytes) / block_size


def get_unusable_space_factor(cluster_fm, cluster_constants=None):
    """Get the factor of actual block data to reported block data.

    The reported values include an unusable-space factor to leave room for garbage collection and other
    processes to proceed unhindered.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.ClusterConstants cluster_constants: Cluster constants.
    :rtype: float
    """
    if not cluster_constants:
        cluster_constants = cluster_fm.get_constants()

    return 100. / (100 - cluster_constants.bs_unusable_space_percent)


def get_volume_block_size(cluster_fm, volume_size, volume_count=1, comp_ratio=1., dedup_ratio=1.,
                          cluster_constants=None, immutable_values=None):
    """Get the size of block data generated by the specified volumes.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.DataSize volume_size: Volume size.
    :param int volume_count: The number of volumes of the specified size.
    :param float comp_ratio: Target block-data compression ratio.
    :param float dedup_ratio: Target block-data deduplication ratio.
    :param valence.models.ClusterConstants cluster_constants: Cluster constants.
    :param sf_api.result_models.GetImmutableValues immutable_values: Immutable cluster values.
    :rtype: valence.models.DataSize
    """
    rep_count = 2
    block_size = int(4096. / comp_ratio)
    block_overhead_factor = get_block_overhead_factor(cluster_fm, block_size, immutable_values)
    unusable_space_factor = get_unusable_space_factor(cluster_fm, cluster_constants)

    volume_bytes = volume_size.bytes * volume_count
    actual_bytes = float(volume_bytes * rep_count) / (comp_ratio * dedup_ratio)
    return models.DataSize(int(actual_bytes * block_overhead_factor * unusable_space_factor))


def get_block_volume_size(cluster_fm, block_data_size, comp_ratio=1., dedup_ratio=1.,
                          cluster_constants=None, immutable_values=None):
    """Get the volume size required to generate the specified amount of block data.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.DataSize block_data_size: Block data size.
    :param float comp_ratio: Compression ratio of the specified block data.
    :param float dedup_ratio: Deduplication ratio of the specified block data.
    :param valence.models.ClusterConstants cluster_constants: Cluster constants.
    :param sf_api.result_models.GetImmutableValues immutable_values: Immutable cluster values.
    :rtype: valence.models.DataSize
    """
    rep_count = 2
    block_size = int(4096. / comp_ratio)
    block_overhead_factor = get_block_overhead_factor(cluster_fm, block_size, immutable_values)
    unusable_space_factor = get_unusable_space_factor(cluster_fm, cluster_constants)

    actual_bytes = float(block_data_size.bytes) / (block_overhead_factor * unusable_space_factor)
    return models.DataSize(int(math.ceil(comp_ratio * dedup_ratio * actual_bytes / rep_count)))


def get_block_metadata_size(cluster_fm, block_data_size, comp_ratio=1., dedup_ratio=1.,
                            cluster_constants=None, immutable_values=None):
    """Get the size of metadata generated by the specified amount of block data.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.DataSize block_data_size: Block data size.
    :param float comp_ratio: Compression ratio of the specified block data.
    :param float dedup_ratio: Deduplication ratio of the specified block data.
    :param valence.models.ClusterConstants cluster_constants: Cluster constants.
    :param sf_api.result_models.GetImmutableValues immutable_values: Immutable cluster values.
    :rtype: valence.models.DataSize
    """
    if not immutable_values:
        immutable_values = cluster_fm.get_immutable_values()

    block_size = int(4096. / comp_ratio)
    block_overhead_factor = get_block_overhead_factor(cluster_fm, block_size, immutable_values)
    unusable_space_factor = get_unusable_space_factor(cluster_fm, cluster_constants)

    per_page_meta_blocks = immutable_values.per_page_metadata_blocks
    per_page_meta_bytes = immutable_values.per_page_metadata_bytes
    meta_overhead_factor = 1. + float(per_page_meta_bytes) / (4096 * per_page_meta_blocks)
    per_block_meta_bytes = immutable_values.per_block_metadata_bytes * dedup_ratio

    actual_bytes = float(block_data_size.bytes) / (block_overhead_factor * unusable_space_factor)
    block_count = int(math.ceil(actual_bytes / block_size))
    return models.DataSize(int(block_count * per_block_meta_bytes * meta_overhead_factor))


def get_metadata_block_size(cluster_fm, metadata_size, comp_ratio=1., dedup_ratio=1.,
                            cluster_constants=None, immutable_values=None, under_size=False):
    """Get the size of block data required to generate the specified amount of metadata.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.DataSize metadata_size: Metadata size.
    :param float comp_ratio: Target block-data compression ratio.
    :param float dedup_ratio: Target block-data deduplication ratio.
    :param valence.models.ClusterConstants cluster_constants: Cluster constants.
    :param sf_api.result_models.GetImmutableValues immutable_values: Immutable cluster values.
    :param bool under_size: Whether to round down instead of up to the nearest block count.
    :rtype: valence.models.DataSize
    """
    if not immutable_values:
        immutable_values = cluster_fm.get_immutable_values()

    block_size = int(4096. / comp_ratio)
    block_overhead_factor = get_block_overhead_factor(cluster_fm, block_size, immutable_values)
    unusable_space_factor = get_unusable_space_factor(cluster_fm, cluster_constants)

    per_page_meta_blocks = immutable_values.per_page_metadata_blocks
    per_page_meta_bytes = immutable_values.per_page_metadata_bytes
    meta_overhead_factor = 1. + float(per_page_meta_bytes) / (4096 * per_page_meta_blocks)
    per_block_meta_bytes = immutable_values.per_block_metadata_bytes * dedup_ratio

    block_count = int(math.ceil(
        float(metadata_size.bytes) / (per_block_meta_bytes * meta_overhead_factor)
    ))
    if under_size:
        block_count -= 1
    actual_bytes = block_count * block_size
    return models.DataSize(int(actual_bytes * block_overhead_factor * unusable_space_factor))
