"""Helper functions for drive methods."""
import collections
import time

import autologger
import sf_batch
import sf_models

from valence import (
    criteria,
    enums,
    errors,
    functional_models,
    models,
    utils,
)
from valence.vl_helpers import (
    _bin_sync,
    _errors,
    _faults,
    _slice,
)

_log = autologger.get_sflogging()


def add_drives(cluster_fm, drive_dms, timeout=None, ignore_unbalanced_cluster_check=None,
               force_during_bin_sync=None, enable_mdss_check=True, force_non_fips=None):
    """Add drives with optional MDSS check.

    This function will block until the drives are added.
    If the cluster has MDSS enabled, MDSS drives will be added as slice
    unless the enable_mdss_check flag is false.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: the list of drives you want to add
    :param int timeout: timeout when waiting for completion
    :param bool ignore_unbalanced_cluster_check: Whether to skip the check for an unbalanced cluster
    :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.
    :param bool enable_mdss_check: Automatically match the cluster's MDSS settings.
    :param bool force_non_fips: Force add non-fips drives to a fips cluster

    :return: Drive data models
    :rtype: list[valence.models.Drive]
    """

    if enable_mdss_check and is_mdss_enabled(cluster_fm):
        # get MDSS slot drives that are not on the drive_dms list
        drive_dms = _get_all_mdss_drive_dms(cluster_fm, drive_dms)
        # set_drives_type will change any drives in both slice drive slots to slice
        drive_dms = set_drives_type(cluster_fm=cluster_fm,
                                    drive_dms=drive_dms,
                                    enable_mdss=True)

    timeout = 600 if timeout is None else timeout
    async_handler = cluster_fm.add_drives(
        drive_dms=drive_dms,
        ignore_unbalanced_cluster_check=ignore_unbalanced_cluster_check,
        force_during_bin_sync=force_during_bin_sync,
        force_non_fips=force_non_fips
    )
    # This is the interval that was used in ApiDrive.add_drives
    poll_period = 15
    return async_handler.get_models(blocking=True, timeout=timeout, poll_period=poll_period)


def _get_all_mdss_drive_dms(cluster_fm, drive_dms):
    """Get additional MDSS drives.

    This function makes a list of all the nodes that have a primary slice drive in the node_dms list.
    Then all the available slot 2 drives for those nodes, that are not in the drive_dms list are added
    to the drive_dms list.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: the original list of drives that are being added

    :return: Drive Data Models
    :rtype: list[valence.models.Drive]
    """

    constants_dm = cluster_fm.get_constants()
    mdss_slot = constants_dm.allow_drive_type_modification_of_slot_num

    slice_drive_node_ids = []

    for drive in drive_dms:
        if drive.type in (enums.DriveType.slice, enums.DriveType.volume):
            slice_drive_node_dm = get_active_node_dm_from_drive_dm(cluster_fm, drive)
            if drive.slot == get_default_slice_drive_slot(slice_drive_node_dm):
                slice_drive_node_ids.append(slice_drive_node_dm.node_id)

    slot_two_available_drives = cluster_fm.list_drives(node_ids=slice_drive_node_ids,
                                                       slots=[mdss_slot],
                                                       statuses=[enums.DriveStatus.available])

    drive_ids = list({drive.drive_id for drive in drive_dms})

    for drive in slot_two_available_drives:
        if drive.drive_id not in drive_ids:
            drive_ids.append(drive.drive_id)
            drive_dms.append(drive)

    return drive_dms


def is_mdss_enabled(cluster_fm):
    """Check for cluster MDSS state.

    Determines if MDSS is enabled on at least one node in the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model

    :rtype: bool
    """
    constants_dm = cluster_fm.get_constants()
    mdss_slot = constants_dm.allow_drive_type_modification_of_slot_num
    mdss_drives = cluster_fm.list_drives(slots=[mdss_slot],
                                         drive_types=[enums.DriveType.slice,
                                                      enums.DriveType.volume])
    return bool(mdss_drives)


def remove_drives(cluster_fm, drive_dms, timeout=None, ignore_unbalanced_cluster_check=None,
                  force_during_bin_sync=None):
    """Remove drives.

    This function will block until the drives are removed.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: the list of drives you want to remove
    :param int timeout: timeout when waiting for completion
    :param bool ignore_unbalanced_cluster_check: Whether to skip the check for an unbalanced cluster
    :param bool force_during_bin_sync: Whether to force drive removal even during bin sync.

    :return: Drive data models
    :rtype: list[valence.models.Drive]
    """
    timeout = 600 if timeout is None else timeout
    async_handler = cluster_fm.remove_drives(
        drive_dms=drive_dms,
        ignore_unbalanced_cluster_check=ignore_unbalanced_cluster_check,
        force_during_bin_sync=force_during_bin_sync
    )
    poll_period = 15
    return async_handler.get_models(blocking=True, timeout=timeout, poll_period=poll_period)


def set_drives_type(cluster_fm, drive_dms, enable_mdss=False):
    """Set the type of each drive based on slot and the enable_mdss flag.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: Drive data models. These objects will not be modified.
    :param bool enable_mdss: Enable mdss flag
    :return: New drive data models with drive types set
    :rtype: list[valence.models.Drive]
    """
    # retrieve the default slice drives for each node
    result_drive_dms = []
    for drive_dm in drive_dms:
        new_drive_dm = models.Drive(drive_dm.serialize())
        drive_node_dm = get_active_node_dm_from_drive_dm(cluster_fm, new_drive_dm)
        if enable_mdss:
            # If we want to add a second slice drive, mdss drive is always slot 2
            if new_drive_dm.slot == 2:
                new_drive_dm.type = enums.DriveType.slice
            # If this drive is in the default slice slot, it should be marked slice
            elif get_default_slice_drive_slot(drive_node_dm) == new_drive_dm.slot:
                new_drive_dm.type = enums.DriveType.slice
            # Otherwise, it is a block drive
            else:
                new_drive_dm.type = enums.DriveType.block
        else:
            # If we don't want two slice drives...
            if get_default_slice_drive_slot(drive_node_dm) == new_drive_dm.slot:
                new_drive_dm.type = enums.DriveType.slice
            else:
                new_drive_dm.type = enums.DriveType.block
        result_drive_dms.append(new_drive_dm)
    return result_drive_dms


def add_available_drives(cluster_fm, timeout=None, ignore_unbalanced_cluster_check=None):
    """Add all available drives.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: timeout when waiting for completion
    :param bool ignore_unbalanced_cluster_check: Whether to skip the check for an unbalanced cluster
    """
    timeout = 600 if timeout is None else timeout
    available_drives = cluster_fm.list_drives(statuses=[enums.DriveStatus.available])
    add_drives(cluster_fm=cluster_fm, drive_dms=available_drives, timeout=timeout,
               ignore_unbalanced_cluster_check=ignore_unbalanced_cluster_check)


def wait_for_drive_status(cluster_fm, drive_dms, timeout=None, status=enums.DriveStatus.available,
                          poll_period=None, check_drive_serial=True):
    """Wait for a certain drive status by calling ListDrives.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: the list of drives to be queried
    :param int timeout: timeout when waiting for completion
    :param valence.enums.DriveStatus status: status to wait for
    :param int poll_period: time to wait between list_drives calls
    :param bool check_drive_serial: list_drives call using a list of drive serial option
    """
    wait_for_drive_statuses(cluster_fm, drive_dms, timeout=timeout, statuses=[status],
                            poll_period=poll_period, check_drive_serial=check_drive_serial)


def wait_for_drive_statuses(cluster_fm, drive_dms, timeout=None, statuses=None, poll_period=None,
                            check_drive_serial=True):
    """Wait for drives to be any of the specified statuses by calling ListDrives.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: the list of drives to be queried
    :param int timeout: timeout when waiting for completion
    :param list[valence.enums.DriveStatus] statuses: list of statuses to wait for
    :param int poll_period: time to wait between list_drives calls
    :param bool check_drive_serial: list_drives call using a list of drive serial option
    :raises errors.TimeoutError: If not all drives are in any of the statuses within the timeout
    """
    assert drive_dms, 'Error: drive_dms must contain at least one entry'
    timeout = 600 if timeout is None else timeout
    timer = utils.timeout_timer(timeout)
    poll_period = 15 if poll_period is None else poll_period
    serials = None
    if check_drive_serial:
        serials = [drive.serial for drive in drive_dms]
    final_drives = []
    while next(timer):
        if check_drive_serial:
            final_drives = cluster_fm.list_drives(serials=serials, statuses=statuses)
        else:
            final_drives = cluster_fm.list_drives(statuses=statuses)
        if len(drive_dms) == len(final_drives):
            return
        time.sleep(poll_period)
    raise errors.TimeoutError(
        'Timed out waiting for drive(s) to be in {} after {} seconds: {}'.format(
            [status.value for status in statuses],
            timeout,
            [dm for dm in drive_dms if dm not in final_drives])
    )


def wait_for_any_drive_statuses(cluster_fm, statuses, timeout=None, poll_period=None):
    """Wait for any drives to be in specified status by calling ListDrives.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: timeout when waiting for completion
    :param list[valence.enums.DriveStatus] statuses: Statuses to wait for.
    :param int poll_period: time to wait between list_drives calls
    :raises errors.TimeoutError: If no drive is in any of the statuses within the timeout
    """
    timeout = 600 if timeout is None else timeout
    timer = utils.timeout_timer(timeout)
    poll_period = 15 if poll_period is None else poll_period
    while next(timer):
        if cluster_fm.list_drives(statuses=statuses):
            return
        time.sleep(poll_period)
    raise errors.TimeoutError(
        'Timed out waiting for any drive to be in {} after {} seconds.'.format(
            [status.value for status in statuses], timeout)
    )


def get_default_slice_drive_slot(node_dm):
    """Function to return default slice drive slot based on the node model passed in.

    :param sf_api.models.SfNodeBase node_dm: Node to inspect and return appropriate
     default slice drive slot.
    :rtype: int
    """

    if node_dm.platform_info.node_type in [sf_models.NodeType.STANDARD_TYPES.SF3010]:
        return -1
    elif node_dm.platform_info.node_type in [sf_models.NodeType.STANDARD_TYPES.SF9608]:
        return 1
    else:
        return 0


def get_active_node_dm_from_drive_dm(cluster_fm, drive_dm):
    """Given a drive data model, query for the active node it belongs to.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.models.Drive drive_dm: the drive you are search on
    :rtype: valence.models.ActiveNode
    """

    for node_dm in cluster_fm.list_active_nodes():
        if node_dm.node_id == drive_dm.node_id:
            return node_dm


def _create_node_to_drive_map(active_nodes, drive_dms):
    """Method to map the drives to fail to the node they're on.

    :param list[valence.models.ActiveNode] active_nodes: The active node objects
    :param list[valence.models.Drive] drive_dms: Drives to remove
    :rtype: dict[str,list[valence.models.Drive]]
    """
    node_id_to_drive_map = collections.defaultdict(list)
    node_mip_to_drive_map = {}

    for drive_dm in drive_dms:
        node_id_to_drive_map[drive_dm.node_id].append(drive_dm)

    for active_node in active_nodes:
        if active_node.node_id in node_id_to_drive_map:
            node_mip_to_drive_map[active_node.mip] = node_id_to_drive_map[active_node.node_id]

    return node_mip_to_drive_map


def check_fake_drives_flag_on_node(node_fm):
    """Helper to check if the fakedrives flag has been set on the node.

    :param valence.functional_models.Node node_fm: Node functional model
    """
    fake_drives_flag = node_fm.get_startup_flags(flags=['fakedrives'])
    if not fake_drives_flag[enums.NodeBinary.sfapp.value][0]['current']:
        raise _errors.HelperError('The node {} does not have fakedrives set! You will not '
                                  'be able to fail any drives! Please set the flag in '
                                  'setup cluster.'.format(node_fm.mip))


def fail_drives_in_cluster(cluster_fm, drive_dms, wait_for_failure=True, timeout=None):
    """Fail the given block drives.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Drive] drive_dms: Drives to remove
    :param bool wait_for_failure: Wait until all the drives have failed
    :param int timeout: The max time to wait for the drive to fail
    :raises errors.HelperError: If the drives do not fail within the timeout or if fakedrives is not set
    """
    timeout = 180 if timeout is None else timeout
    node_dms = cluster_fm.list_active_nodes()
    node_fms = [functional_models.Node(dm.resource_dm, cluster_fm.cluster_dm) for dm in node_dms]
    node_mip_to_drive_map = _create_node_to_drive_map(node_dms, drive_dms)
    fail_drives_threads = []
    get_flags_threads = []
    for node_fm in node_fms:
        if node_fm.mip in node_mip_to_drive_map:
            get_flags_threads.append(sf_batch.SFThread(check_fake_drives_flag_on_node, node_fm,
                                                       thread_name='_check_fake_drives_flag-{}'.format(
                                                               node_fm)))
            fail_drives_threads.append(sf_batch.SFThread(node_fm.fail_drives,
                                                         node_mip_to_drive_map[node_fm.mip],
                                                         thread_name='fail_drives-{}'.format(node_fm)))

    sf_batch.start_all_threads(get_flags_threads)
    sf_batch.join_all_threads(get_flags_threads)
    sf_batch.check_for_error_and_raise_exception(get_flags_threads)

    sf_batch.start_all_threads(fail_drives_threads)
    sf_batch.join_all_threads(fail_drives_threads)
    sf_batch.check_for_error_and_raise_exception(fail_drives_threads)

    failed_drives = []
    if wait_for_failure:
        for drive_dm in drive_dms:
            failed_drives.append(sf_batch.SFThread(_wait_for_drive_to_fail, cluster_fm, drive_dm,
                                                   timeout=timeout,
                                                   thread_name='wait_for_drive_to_fail-{}'.format(
                                                           drive_dm)))
        sf_batch.start_all_threads(failed_drives)
        sf_batch.join_all_threads(failed_drives)
        sf_batch.check_for_error_and_raise_exception(failed_drives)


def set_fake_drives_flag_on_node(node_fm):
    """Helper to set the fakedrives flag on a node so it can fail drives.

    Do NOT use this unless you know what you're doing. This will restart sfapp potentially in the
    middle of your test. The recommendation is to set the fakedrives flag in setup cluster. This is why
    the fail_block_drives_in_cluster just performs a check on the flag and will not set the flag for you.

    :param valence.functional_models.Node node_fm: Node functional model
    """
    node_fm.set_startup_flags(flags={'fakedrives': True})
    node_fm.stop_binary(enums.NodeBinary.solidfire)
    node_fm.start_binary(enums.NodeBinary.solidfire)


def wait_for_drive_assignments(cluster_fm, drive_dms, timeout=None):
    """Wait for bin assignments on the specified drives.

    .. note:: Bin assignments need not be active.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drive models.
    :param int timeout: Time in seconds to wait for bin assignments.
    :raises valence.errors.TimeoutError: If bin assignments are not present within the timeout.
    """
    block_drive_dms = [dm for dm in drive_dms if dm.type is enums.DriveType.block]
    if block_drive_dms:
        timer = utils.timeout_timer(600 if timeout is None else timeout)
        drive_services = None
        while next(timer):
            drive_services = cluster_fm.list_services(drive_dms=block_drive_dms)
            if drive_services:
                break

        if not drive_services:
            raise errors.TimeoutError(
                'No services found after {}s for drives: {}'.format(timeout, block_drive_dms)
            )

        drive_assignments = None
        while next(timer):
            drive_assignments = cluster_fm.get_bin_assignments(services=drive_services).bins
            if drive_assignments:
                break

        if not drive_assignments:
            raise errors.TimeoutError(
                'No bin assignments found after {}s for drives: {}'.format(timeout, block_drive_dms)
            )


def _wait_for_drive_to_fail(cluster_fm, fail_drive_dm, timeout):
    """Helper method to wait for a drive to fail.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.models.Drive fail_drive_dm: Drive we're waiting for to fail
    :param int timeout: The max time to wait for the drive to fail
    :rtype: valence.models.Drive
    :raises errors.HelperError: If the drive does not fail within the timeout
    """
    timer = utils.timeout_timer(timeout)

    while next(timer):
        drive_dms = cluster_fm.list_drives()
        for drive_dm in drive_dms:
            if (drive_dm.drive_id == fail_drive_dm.drive_id
                    and drive_dm.status is enums.DriveStatus.failed):
                _log.debug('Drive id {} has failed.'.format(drive_dm.drive_id))
                return drive_dm
        time.sleep(10)

    error_msg = 'Drive id {} has not failed after {} seconds!'.format(fail_drive_dm.drive_id,
                                                                      timeout)
    raise _errors.HelperError(error_msg)


def _wait_for_drive_restore(cluster_fm, restore_drive_dm, timeout):
    """Wait for drive to become active or available.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.models.Drive restore_drive_dm: Drive we're waiting to restore
    :param int timeout: The max time to wait for the drive to restore
    :raises errors.HelperError: if drives are not restored within timeout seconds
    """
    timer = utils.timeout_timer(timeout)
    while next(timer):
        drive_dms = cluster_fm.list_drives()
        for drive_dm in drive_dms:
            if (drive_dm.drive_id == restore_drive_dm.drive_id and drive_dm.status in
                    (enums.DriveStatus.active, enums.DriveStatus.available)):
                _log.debug('Drive id {} restored.'.format(drive_dm.drive_id))
                return
        time.sleep(10)
    error_msg = 'Drive id {} was not restored after {} seconds'.format(restore_drive_dm.drive_id,
                                                                       timeout)
    raise _errors.HelperError(error_msg)


def restore_failed_drives(cluster_fm, drive_dms, wait_for_restore=True, timeout=None):
    """Restore the specified drives to working order.

    Restores IO functionality to the drives but can leave them in an un-deterministic state on the
    cluster (see recover_failed_drives for details).

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drives to restore.
    :param bool wait_for_restore: Whether to wait until all the drives have been restored.
    :param int timeout: The max time to wait for the drives to restore.
    """
    timeout = 300 if timeout is None else timeout
    node_dms = cluster_fm.list_active_nodes()
    node_fms = [functional_models.Node(dm.resource_dm, cluster_fm.cluster_dm) for dm in node_dms]
    node_mip_to_drive_map = _create_node_to_drive_map(node_dms, drive_dms)
    restore_drive_threads = []
    for node_fm in node_fms:
        if node_fm.mip in node_mip_to_drive_map:
            restore_drive_threads.append(sf_batch.SFThread(
                api_fun_ref=node_fm.restore_failed_drives,
                drive_dms=node_mip_to_drive_map[node_fm.mip],
                thread_name='restore_failed_drives-{}'.format(node_fm)
            ))
    sf_batch.start_all_threads(restore_drive_threads)
    sf_batch.join_all_threads(restore_drive_threads)
    sf_batch.check_for_error_and_raise_exception(restore_drive_threads)
    if wait_for_restore:
        restored_drives = []
        for drive_dm in drive_dms:
            restored_drives.append(sf_batch.SFThread(
                api_fun_ref=_wait_for_drive_restore,
                cluster_fm=cluster_fm,
                restore_drive_dm=drive_dm,
                timeout=timeout,
                thread_name='wait_for_restore-{}'.format(drive_dm)
            ))
        sf_batch.start_all_threads(restored_drives)
        sf_batch.join_all_threads(restored_drives)
        sf_batch.check_for_error_and_raise_exception(restored_drives)


def recover_failed_drives(cluster_fm, drive_dms, timeout=None, wait_for_sync=True):
    """Recover the given drives from a recent failure.

    NOTE: Drives must be restored before they can be recovered (see vl_helpers.restore_failed_drives).

    The following diagram represents the most common recovery paths handled by this helper depending
    on when a given drive was restored (vertical lines represent drive restoration paths)::

                                                                           |>>>>>>> sync-out >>>>>>|
                 Failed                                             Service Timeout            Available
                   |---------------------------------------------|---------X---------|-------------|
                                                                 |                   |             |
        Recovered  <---recover_failed_drives---------------------|                   |             |
                                                               Active              Active          |
                                                                                     |             |
        Recovered  <---recover_failed_drives-----------------------------------------|             |
                                                                                 Available         |
                                                                                                   |
        Recovered  <---recover_failed_drives-------------------------------------------------------|

    NOTE: Recovery can be impeded if the cluster is in a state where it cannot sync. You can ignore
    cannot-sync faults by setting wait_for_sync=False, but then recovery will have to be performed
    again once syncing can resume in order to return the cluster to a known state. If the cluster cannot
    sync, the above diagram will be affected as follows::

                 Failed                                             Service Timeout
                   |---------------------------------------------|---------X---------|
                                                                 |                   |
        Recovered  <---recover_failed_drives---------------------|                   |
                                                               Active              Active
                                                                                     |
                                       Syncing Resumes                               |
        Not Recovered (Available)  <----------X---------|<---recover_failed_drives---|
                                                                                  (Active)

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drives to recover.
    :param int timeout: The max time to wait for the drives to recover.
    :param bool wait_for_sync: Whether to wait for the syncing to complete during recovery.
    """
    timeout = 1800 if timeout is None else timeout
    time_start = time.time()
    wait_for_drive_statuses(
        cluster_fm=cluster_fm,
        drive_dms=drive_dms,
        timeout=timeout,
        statuses=[enums.DriveStatus.active, enums.DriveStatus.available]
    )

    # If desired, wait for cluster to be capable of syncing
    if wait_for_sync:
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=cluster_fm,
            only_fault_codes=[enums.ClusterFaultCode.cluster_cannot_sync]
        )

    # If necessary, wait for drives to sync out
    cannot_sync_faults = cluster_fm.list_cluster_faults(
        fault_status=enums.ClusterFaultStatus.current,
        include_codes=[enums.ClusterFaultCode.cluster_cannot_sync],
        update=True
    )
    if not cannot_sync_faults:
        for drive_dm in drive_dms:
            # If there are PRD bin assignments on a recently restored drive then it is being synced out
            services = cluster_fm.list_services(drive_dms=[drive_dm])
            non_prd_bins = cluster_fm.get_bin_assignments(
                exclude_statuses=[enums.BinStatus.pending_removal_dead],
                services=services
            ).bins
            if not non_prd_bins and drive_dm.type is enums.DriveType.block:
                wait_for_drive_status(
                    cluster_fm=cluster_fm,
                    drive_dms=[drive_dm],
                    timeout=timeout - int(time.time() - time_start),
                    status=enums.DriveStatus.available
                )

    # Add back any drives that synced out as a result of the failure
    available_drive_dms = cluster_fm.list_drives(
        drive_ids=[dm.drive_id for dm in drive_dms],
        statuses=[enums.DriveStatus.available]
    )
    if available_drive_dms:
        add_drives(
            cluster_fm=cluster_fm,
            drive_dms=available_drive_dms,
            timeout=timeout - int(time.time() - time_start),
            ignore_unbalanced_cluster_check=True,
            force_during_bin_sync=True
        )

    # If desired, wait for bin syncing to complete
    if wait_for_sync:
        _bin_sync.wait_for_bin_sync_to_finish_with_timeout(
            cluster_fm=cluster_fm,
            timeout=timeout - int(time.time() - time_start)
        )


def wait_for_services_to_decommission(cluster_fm, drive_dms, service_ids):
    """Wait for services not to be associated with drives and to remove themselves.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drives to wait for.
    :param set[int] service_ids: service ids to wait for
    """
    block = [d for d in drive_dms if d.type is enums.DriveType.block]
    slice_drives = [d for d in drive_dms if d.type is enums.DriveType.volume]
    slice_sync_timeout = bin_sync_timeout = 0
    # TODO: ELEM-7810 maximum sync timeout method
    if slice_drives:
        slice_sync_timeout = _slice.calculate_slice_sync_timeout(cluster_fm)
    if block:
        bin_sync_timeout = _bin_sync.get_bin_sync_timeout(cluster_fm, removed_drive_dms=block)
    # Slice minimum of 60 seconds is not sufficient for services to decommission
    timeout = max(slice_sync_timeout, bin_sync_timeout, 600)
    drive_ids = [d.drive_id for d in drive_dms]
    services_filter = criteria.Or(
        criteria.ContainedIn(models.ClusterService.service_id, service_ids),
        criteria.ContainedIn(models.ClusterService.drive_id, drive_ids)
    )
    timer = utils.timeout_timer(timeout)
    while next(timer):
        current_services = [s.service for s in cluster_fm.list_services()]
        if not services_filter.filter(current_services):
            break
        time.sleep(10)
    else:
        raise errors.TimeoutError('Services have not been removed after timeout {}'
                                  .format(timeout))


def wait_for_sync_out(cluster_fm, drive_dms):
    """Wait for drives to sync out.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drives to wait for.
    """
    block = [d for d in drive_dms if d.type is enums.DriveType.block]
    slice_drives = [d for d in drive_dms if d.type is enums.DriveType.volume]
    services = cluster_fm.list_services(drive_dms=drive_dms)
    service_ids = set([s.service.service_id for s in services])
    wait_for_services_to_decommission(cluster_fm, drive_dms, service_ids)
    slices_in_error = None
    blocks_in_error = None
    # TODO: ELEM-7810 maximum sync timeout method

    start_time = time.time()
    slice_sync_timeout = bin_sync_timeout = 0
    if slice_drives:
        slice_sync_timeout = _slice.calculate_slice_sync_timeout(cluster_fm)
    if block:
        bin_sync_timeout = _bin_sync.get_bin_sync_timeout(cluster_fm, removed_drive_dms=block)
    total_timeout = max(slice_sync_timeout, bin_sync_timeout)
    # I put this first cause it **should** take less time than block, right?
    if slice_drives:
        _slice.wait_for_slice_sync_to_finish(cluster_fm, timeout=slice_sync_timeout)
        slice_sync_report = cluster_fm.get_slice_sync_report()
        slices_in_error = _slice.get_slices_assigned_to_services(slice_sync_report, service_ids)
        slices_in_error.extend([s for s in slice_sync_report.slices if s.dead_secondaries])

    if block:
        bin_time = time.time()
        _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm, timeout=total_timeout-(
            int(bin_time-start_time)))
        bin_assignments = cluster_fm.get_bin_assignments()
        blocks_in_error = _bin_sync.get_bins_assigned_to_service(bin_assignments, service_ids)

    # I'm raising here, cause it seems like the right thing to do
    if slices_in_error:
        raise errors.ElementInvalidStateError(
            'Slices are assigned to services that should have been decommissioned {}'.format(
                ', '.join(str([s.serialize() for s in slices_in_error]))))

    # Ditto raise comment from above
    if blocks_in_error:
        raise errors.ElementInvalidStateError(
            'Bins are assigned to services that should have been decommissioned {}'.format(
                ','.join(str([b.serialize() for b in blocks_in_error]))))


def wait_for_drive_services_healthy(cluster_fm, drive_dms):
    """Wait for services to be associated with drives and healthy.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drives to wait for.
    :raises errors.TimeoutError: If services aren't using the drives and healthy after timeout
    """
    # TODO: ELEM-7810 maximum sync timeout method
    block = [d for d in drive_dms if d.type is enums.DriveType.block]
    slice_drives = [d for d in drive_dms if d.type is enums.DriveType.volume]
    slice_sync_timeout = bin_sync_timeout = 0
    if slice_drives:
        slice_sync_timeout = _slice.calculate_slice_sync_timeout(cluster_fm)
    if block:
        bin_sync_timeout = _bin_sync.get_bin_sync_timeout(cluster_fm, removed_drive_dms=block)
    timeout = max(slice_sync_timeout, bin_sync_timeout)
    timer = utils.timeout_timer(timeout)
    drive_ids = set([d.drive_id for d in drive_dms])
    while next(timer):
        current_services = cluster_fm.list_services(drive_dms=drive_dms)
        all_drives = drive_ids.difference(set([s.service.drive_id for s in current_services]))
        if (all([s.service.status is enums.ServiceStatus.healthy for s in current_services]) and
                not all_drives):
            break
    current_services = cluster_fm.list_services(drive_dms=drive_dms)
    all_drives = drive_ids.difference(set([s.service.drive_id for s in current_services]))
    if all_drives or any([s.service.status is not enums.ServiceStatus.healthy for s in
                          current_services]):
        raise errors.TimeoutError('Timeout waiting for for services to become healthy on drives_ids {}'
                                  .format(','.join([str(d) for d in drive_ids])))


def get_unexpected_drives_without_slice_services(cluster_fm, slice_drives):
    """Return any drives that do not have an associated slice service when expected.

    If there aren't enough volumes to justify the slice service count, return an emtpy list

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] slice_drives: Drives to wait for.
    :rtype: list[valence.models.Drive]
    """
    drives_without_slice = []
    volume_count = len(cluster_fm.list_volumes())
    services = cluster_fm.list_services()
    slice_services = [s for s in services if s.service.service_type is enums.ServiceType.slice]
    if volume_count > len(slice_services) * 2:
        for drive in slice_drives:
            associated_service = [s for s in slice_services if s.service.drive_id == drive.drive_id]
            if not associated_service:
                drives_without_slice.append(drive)
    return drives_without_slice


def get_drives_without_bins(cluster_fm, block_drives):
    """Any drives without bins assigned will be returned.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] block_drives: Drives to wait for.
    :rtype: list[valence.models.Drive]
    """
    block_drives_without_bins = []
    bin_assignments = cluster_fm.get_bin_assignments()
    services_with_bins = set()
    for b in bin_assignments.bins:
        services_with_bins.update(set([s.service_id for s in b.services]))
    services = cluster_fm.list_services()

    for drive in block_drives:
        block_services = set([s.service.service_id for s in services if
                              s.service.drive_id == drive.drive_id])
        if not bool(block_services & services_with_bins):
            block_drives_without_bins.append(drive)
    return block_drives_without_bins


def wait_for_sync_in(cluster_fm, drive_dms, timeout=None):
    """Wait for drives to sync in.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.Drive] drive_dms: Drives to wait for.
    :param int timeout: timeout for services to show up and get healthy
    :raises errors.ElementInvalidStateError: If bins/slices aren't assigned after syncing
    """
    block = [d for d in drive_dms if d.type is enums.DriveType.block]
    slice_drives = [d for d in drive_dms if d.type is enums.DriveType.volume]
    wait_for_drive_services_healthy(cluster_fm, drive_dms)
    start_time = time.time()
    slice_sync_timeout = bin_sync_timeout = 0
    slice_drives_in_error = None
    blocks_drives_in_error = None
    # TODO: ELEM-7810 maximum sync timeout method
    if slice_drives:
        slice_sync_timeout = _slice.calculate_slice_sync_timeout(cluster_fm)
    if block:
        bin_sync_timeout = _bin_sync.get_bin_sync_timeout(cluster_fm, removed_drive_dms=block)
    total_timeout = max(slice_sync_timeout, bin_sync_timeout)
    if slice_drives:
        _slice.wait_for_slice_sync_to_finish(cluster_fm, timeout=slice_sync_timeout)
        slice_drives_in_error = get_unexpected_drives_without_slice_services(cluster_fm, slice_drives)
    if block:
        bin_time = time.time()
        _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm, timeout=total_timeout-(
            int(bin_time-start_time)))
        blocks_drives_in_error = get_drives_without_bins(cluster_fm, block)

    if slice_drives_in_error:
        raise errors.ElementInvalidStateError(
            'Drives: {} should be associated with services, but were not'.format(
                ', '.join([str(d.serialize()) for d in slice_drives_in_error])))

    if blocks_drives_in_error:
        raise errors.ElementInvalidStateError(
            'Block drive: {} should have bins assigned, but did not'.format(
                ', '.join([str(d.serialize()) for d in blocks_drives_in_error])))
