"""Helper functions for Cluster Events."""

from __future__ import division

import time

import autologger

from valence import (
    criteria,
    errors,
    models,
    utils,
)

_log = autologger.get_sflogging()

DEFAULT_WAIT_EVENT_TIMEOUT_IN_SEC = 600


def get_next_event_id(cluster_fm):
    """Get the next event ID for a given cluster.

    NOTE: This can be useful when waiting on an event, so you can use a start_event_id value that will
    catch 'new' events and not accidentally use existing events.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model

    :rtype: int
    :returns: the most recent event ID + 1
    """

    events = cluster_fm.list_events(max_events=1)
    if events:
        return events[0].event_id + 1
    else:
        return 0


def wait_for_event(cluster_fm, event_type, message_regexes=None,
                   timeout=DEFAULT_WAIT_EVENT_TIMEOUT_IN_SEC, poll_period=20,
                   start_event_id=None):
    """Wait for event.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.enums.ClusterEventType event_type: Event type to wait for.
    :param list[str] message_regexes: messages to filter on by regexes (e.g. ["^expected message$",]
    :param int timeout: Time to wait in seconds before raising a timeout exception, Defaults to 600 sec.
    :param int poll_period: Time to wait in seconds between each iteration, Defaults to 20 sec.
    :param int start_event_id: Event ID to start from.  If not specified, default to poll based on
        current UTC.
    :rtype: list[valence.models.ClusterEvent]
    :returns: list of cluster events that have occurred matching the expected criteria
    :raises valence.errors.TimeoutError: If event does not start within the given time
    """

    timeout = timeout if timeout is not None else DEFAULT_WAIT_EVENT_TIMEOUT_IN_SEC

    timer = utils.timeout_timer(timeout)

    # How we'll look for events depends on whether the start_event_id was specified.  If it wasn't
    # specified, search based on the current time in UTC.
    start_publish_time = None
    if start_event_id is None:
        start_publish_time = utils.get_now_utc()

    msg = '' if not message_regexes else " with message '{}' ".format(message_regexes)

    while next(timer):
        event_list = cluster_fm.list_events(
            event_type=event_type,
            start_event_id=start_event_id,
            start_publish_time=start_publish_time
        )

        # If we're not filtering on a message, we can skip any additional filtering and return any
        # events that made it through the list_events operation.
        if event_list and not message_regexes:
            return event_list

        # If we are filtering on a message, perform that filter and return if we have any events once
        # we get through that filtering operation.
        elif event_list and message_regexes:
            message_filter = criteria.MatchIn(models.ClusterEvent.message, message_regexes)

            filtered_event_list = message_filter.filter(event_list)
            if filtered_event_list:
                return filtered_event_list

        _log.debug('Waiting for {}{}. Sleeping {} seconds.'.format(event_type, msg, poll_period))
        time.sleep(poll_period)

    raise errors.TimeoutError('Timed out waiting for {}{}after {} seconds.'.format(
        event_type, msg, timeout))
