"""Helper functions for cluster faults."""
import time

import autologger
import requests

from valence import (
    criteria,
    enums,
    errors,
    models,
    utils,
)
from valence.vl_helpers import _errors

_log = autologger.get_sflogging()


def wait_for_cluster_fault_occurrence(cluster_fm, fault_code, only_nodes=None, timeout=1800,
                                      poll_period=20, include_severities=None):
    """Wait for cluster faults to occur.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.enums.ClusterFaultCode fault_code: fault code to wait for
    :param list[valence.enums.ClusterFaultSeverity] include_severities: cluster fault severities
        to wait for
    :param list[valence.models.ActiveNode] only_nodes: only wait for cluster faults associated with these
        nodes.
    :param int timeout: Seconds after which to fail if faults have not occurred.
    :param int poll_period: Poll ListClusterFaults every poll_period seconds.
    :return: Cluster fault data models
    :rtype: list[valence.models.ClusterFault]
    :raises valence.errors.TimeoutError: if faults not found after timeout.
    """
    timeout = timeout if timeout is not None else 1800
    # setup the filter
    fault_filter = criteria.Is(models.ClusterFault.code, fault_code)

    node_ids = None
    if only_nodes:
        node_ids = [node.node_id for node in only_nodes]

    timer = utils.timeout_timer(timeout)
    while next(timer):
        try:
            faults = cluster_fm.list_cluster_faults(update=True,
                                                    node_ids=node_ids,
                                                    fault_status=enums.ClusterFaultStatus.current,
                                                    include_severities=include_severities)
        except requests.ConnectionError:
            _log.debug('Connection error when listing cluster faults')
            time.sleep(poll_period)
            continue
        if faults:
            filtered = fault_filter.filter(faults)
            if filtered:
                return filtered
        time.sleep(poll_period)
    raise errors.TimeoutError('Fault codes not found: {}'.format(fault_code))


def wait_for_cluster_fault_resolution(cluster_fm,  # pylint: disable=too-complex
                                      ignore_code_regexes=None, ignore_details_regexes=None,
                                      ignore_severities=None, only_nodes=None, only_fault_codes=None,
                                      ignore_fault_codes=None, timeout=None, poll_period=20,
                                      crc_error_threshold=5,
                                      include_cluster_exceptions=False):
    """Wait for cluster faults to resolve.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[str] ignore_code_regexes: fault codes to ignore by regexes ie, ["^node",]
    :param list[str] ignore_details_regexes: fault details to ignore by regexes ie, ["^node",]
    :param list[valence.models.ActiveNode] only_nodes: wait for fault resolution on specified nodes
    :param list[valence.enums.ClusterFaultCode] only_fault_codes: wait for resolution on specific codes
    :param list[valence.enums.ClusterFaultCode] ignore_fault_codes: ignore resolution on specific codes
    :param list[valence.enums.ClusterFaultSeverity] ignore_severities: list of severity levels to ignore
    :param int timeout: Seconds after which to fail if faults are not resolved.
    :param int poll_period: Poll ListClusterFaults every poll_period seconds.
    :param int crc_error_threshold: Allowable number of CRC errors before throwing an error.
    :param bool include_cluster_exceptions: Include "exceptions" code in the API call.
    :raises valence.vl_helpers.errors.HelperError: if faults still found after timeout.
    """
    timeout = 1800 if timeout is None else timeout
    # setup the filter
    criteria_list = [criteria.MatchIn(models.ClusterFault.code, ignore_code_regexes, negate=True),
                     criteria.MatchIn(models.ClusterFault.details, ignore_details_regexes, negate=True)]
    crc_criteria = criteria.Contains(models.ClusterFault.details, 'CRC errors')
    if ignore_severities:
        for s in ignore_severities:
            is_compare = criteria.Is(models.ClusterFault.severity,
                                     s, negate=True)
            criteria_list.append(is_compare)
    fault_filter = criteria.And(*criteria_list)
    timer = utils.timeout_timer(timeout)
    filtered = None

    node_ids = None
    if only_nodes:
        node_ids = [n.node_id for n in only_nodes]

    while next(timer):
        try:
            faults = cluster_fm.list_cluster_faults(
                update=True,
                fault_status=enums.ClusterFaultStatus.current,
                node_ids=node_ids,
                include_codes=only_fault_codes,
                include_cluster_exceptions=include_cluster_exceptions,
                exclude_codes=ignore_fault_codes)
        except requests.ConnectionError:
            _log.debug('Connection error when listing cluster faults')
            time.sleep(poll_period)
            continue
        if not faults:
            return
        filtered = fault_filter.filter(faults)
        if not filtered:
            return
        crc_faults = crc_criteria.filter(faults)
        if crc_error_threshold >= len(crc_faults) == len(filtered):
            return
        time.sleep(poll_period)
    raise _errors.HelperError('Faults still found {}'.format(filtered))
