"""Helper container for healthtool builds on jenkins server."""
from __future__ import absolute_import

import random
import re

import autologger
import requests
from bs4 import BeautifulSoup

from valence import (
    camelot,
    models,
)
from valence.functional_models import _constants
from valence.vl_helpers import _errors

_log = autologger.get_sflogging(__name__)

NETAPP_SUPPORT_URL = 'https://library.netapp.com/ecm/ecm_get_file/ECMLP2840740'
JENKINS_METADATA = 'http://bdr-jenkins.eng.solidfire.net/upgradetools/release_candidate/' \
                   'healthtools_metadata.json'
BDR_JENKINS_HOST = 'bdr-jenkins.eng.solidfire.net'
BDR_JENKINS_PATH = 'upgradetools'
RC = 'release_candidate/'
DEV = 'dev/'
STABLE = 'stable/'
OLD = 'previous_releases/'
RQD = 'rqd/'

VERSIONS_TO_AVOID = [
    _constants.HEALTHTOOLS_10_3_SHIPPING_VERSION,
    _constants.HEALTHTOOLS_18_3_VERSION,
]


class HealthtoolsVersionsForTest(object):
    """Object for aggregating healthtool versions for healthtool test suites.

    The testcases that use this object need to be explicit about their scenarios and this allows us to
    differentiate between all the different scenarios by presenting:
    version_under_test: Whatever we're testing. Tests should use this unless.
    latest_version: Something we can set as a latest version so it's returned in JSON from sfupgradecheck
    original_Version: Version that came on the mnode install (or at start of the test)
    lower_version: Something older than latest_version, not equal to original_Version
    current_version: Runtime value of the version on the mnode.
    jenkins_rqd_versions: The healthtools located in /rqd on jenkins. This MAY be different from
    latest versions on jenkisn or release_candidate/ contents b/c if we have an out of band release
    going through testing while RQD is doing RI, we can't inject new version into testing.
    At the same time RQD can't hardcode old version in at2 b/c those versions will go away after test
    cycle is over. So this basically lets us control what HT gets picked up by what suites at what times.

    Note, this object does not automatically pick what version_under_test and latest_version is.
    Those have to be set by the suite by calling this.set_test_versions()

    :param valence.functional_models.ManagementNodeLegacy mnode_fm: Management node functional model.
    """

    def __init__(self, mnode_fm):
        self._mnode_fm = mnode_fm
        self._version_under_test = None
        self._latest_version = None
        self._lower_version = None
        self._original_version = None
        self._all_versions_for_test = []
        self._healthtools_latest_version_public = None
        self._jenkins_rc_versions = []
        self._all_versions_on_jenkins = []
        self._healthtool_urls = {}
        self._jenkins_rqd_versions = []

        assert self._mnode_fm, "Mnode FM can't be None"

    @property
    def version_under_test(self):
        return self._version_under_test

    @property
    def lower_version(self):
        return self._lower_version

    @property
    def latest_version(self):
        return self._latest_version

    @property
    def original_version(self):
        return self._original_version

    @property
    def current_version(self):
        return self._mnode_fm.get_current_healthtools_version()

    @property
    def all_versions_for_test(self):
        return self._all_versions_for_test

    @property
    def all_versions_on_jenkins(self):
        return self._all_versions_on_jenkins

    @property
    def jenkins_rc_versions(self):
        return self._jenkins_rc_versions

    @property
    def jenkins_rqd_versions(self):
        return self._jenkins_rqd_versions

    @property
    def healthtools_latest_version_public(self):
        return self._healthtools_latest_version_public

    def initialize_versions(self, paths_to_healthtools_from_suite_config,
                            error_if_not_on_jenkins=True):
        """Loads the healthtool versions from jenkins/custom URLs and filters out useless crud.

        :param list[str]|None paths_to_healthtools_from_suite_config: Paths to jenkins passed in from
            suite.
        :param bool error_if_not_on_jenkins: Throw error if files are not on jenkins.
        """
        self._initialize_jenkins_versions()
        self._initialize_nss_version()
        self._initialize_original_version(error_if_not_on_jenkins)
        self._initialize_test_versions(paths_to_healthtools_from_suite_config)
        self._filter_test_versions()

        # Do we have any versions for test at this point?
        assert self._all_versions_for_test

        _log.info(camelot.banner('Found all versions on jenkins: {}'.format(
            ' '.join(map(str, self._all_versions_on_jenkins)))))
        _log.info(camelot.banner('All versions used for this test: {}'.format(
            ' '.join(map(str, self._all_versions_for_test)))))

    def _initialize_jenkins_versions(self):
        """Load the versions from jenkins."""

        jenkins = 'http://{}/{}'.format(BDR_JENKINS_HOST, BDR_JENKINS_PATH)

        # Stable versions pretty much exists always. Error if it doesn't.
        _jenkins_stable_versions_tuples = sorted(self.get_healthtools_at_url(
            '{}/{}'.format(jenkins, STABLE)))

        _jenkins_rc_versions_tuples = sorted(self.get_healthtools_at_url(
            '{}/{}'.format(jenkins, RC)))

        _jenkins_rqd_versions_tuples = sorted(self.get_healthtools_at_url(
            '{}/{}'.format(jenkins, RQD)))

        _all_versions_on_jenkins_tuples = (_jenkins_rc_versions_tuples +
                                           _jenkins_stable_versions_tuples +
                                           _jenkins_rqd_versions_tuples)

        # Remove the URL info and store just the versions in all_versions_on_jenkins.
        self._all_versions_on_jenkins = [n[0] for n in _all_versions_on_jenkins_tuples]
        self._jenkins_rc_versions = sorted([n[0] for n in _jenkins_rc_versions_tuples])
        self._jenkins_rqd_versions = sorted([n[0] for n in _jenkins_rqd_versions_tuples])

        # Store the URLs separately in a dict.
        for n in _all_versions_on_jenkins_tuples:
            self._healthtool_urls[n[0]] = n[1]

    def _initialize_nss_version(self):
        """Get the latest NSS version on public server."""
        json_blob = requests.get(NETAPP_SUPPORT_URL).json()
        self._healthtools_latest_version_public = models.GenericVersion(json_blob['latest_version'])

    def _initialize_original_version(self, error_if_not_on_jenkins=False):
        """Load the original version and verify that we know where it is.

        :param bool error_if_not_on_jenkins: Should we error if this version not on Jenkins?
        :raises: valence.vl_helpers._errors.HealthtoolsVersionNotFound
        """
        # Get the current (aka original) version. This usually should exist on jenkins,
        # but some tests don't care about it. So we will honor the error flag,
        # and let the suite fail later if it chooses to
        self._original_version = self.current_version
        orig = [n for n in self._all_versions_on_jenkins if n == self._original_version]
        if not orig:
            if error_if_not_on_jenkins:
                raise _errors.HealthtoolsVersionNotFound(
                    'Unable to find original mnode healthtools version {} on the jenkins server!'.format(
                        self._original_version))
            else:
                _log.warning("We couldn't find original version of HT {} on jenkins but b/c the"
                             "suite is ignoring this error I'm gonna let it slide. You may encounter"
                             'failures later down the road.'.format(self._original_version))

    def _initialize_test_versions(self, paths_to_healthtools_from_suite_config):
        """Load any custom versions that were passed from resources.

        :param list[str]|None paths_to_healthtools_from_suite_config: Paths to jenkins passed in from
            suite.
        """
        # If suite config has passed in paths to healthtools, use those.
        if paths_to_healthtools_from_suite_config:
            # Any path passed in should follow our convention. I'll let individual downloads fail
            # if URL is incorrect.
            _all_versions_for_test_tuples = [self.get_version_from_url(n) for n
                                             in paths_to_healthtools_from_suite_config]

            # Add these versions to the URLs dict.
            for n in _all_versions_for_test_tuples:
                self._healthtool_urls[n[0]] = n[1]

            self._all_versions_for_test = [n[0] for n in _all_versions_for_test_tuples]
        else:
            # For automatically picking versions for test, we only want to look into stable, rc
            # and rqd. Disregard dev. If you want to use dev/ versions, pass them
            # through suite config.
            self._all_versions_for_test = self.all_versions_on_jenkins

    def _filter_test_versions(self):
        """Filter out healthtool versions that aren't supported by mnode or we don't support."""

        # Filter out versions we want to avoid.
        self._all_versions_for_test = list(set(self.all_versions_for_test) - set(VERSIONS_TO_AVOID))

        # In Sodium we broke compatibility of mnode upgrades and therefore healthtools upgrades
        # between Neon and Sodium. As such, we should limit the testing for Sodium healthtools to
        # only selection of healthtools for Sodium.
        # Neon, however, we're OK testing all, b/c new sodium healthtools support both neon and
        # sodium mnodes.
        if self._mnode_fm.version >= _constants.SODIUM:
            self._all_versions_for_test = [h for h in self._all_versions_for_test if
                                           h >= _constants.HEALTHTOOLS_18_3_VERSION]

    def set_healthtool_versions_for_test_suite(self,
                                               healthtools_version_under_test,
                                               previous_version_of_healthtools=None):
        """Sets the test versions of healthtools to be used by the suites.

        :param valence.models.GenericVersion healthtools_version_under_test: Version to test.
        :param valence.models.GenericVersion|None previous_version_of_healthtools: older version of
            healthtools to test. If None then random version will be picked.
        """
        self._version_under_test = healthtools_version_under_test
        self._latest_version = healthtools_version_under_test

        # Get version lower than latest. If none exist, use latest. Let the test assert if needed.
        if previous_version_of_healthtools:
            self._lower_version = previous_version_of_healthtools
        else:
            previous_versions = [n for n in self._all_versions_for_test if
                                 n < self._version_under_test and n != self._original_version]
            self._lower_version = (random.choice(previous_versions)
                                   if previous_versions else self._version_under_test)

        _log.info(camelot.banner('Version of healthtools under test: {}'.format(
            self._version_under_test)))
        _log.info(camelot.banner(
            'Original version {}, lower version {} and version under test {}'.format(
                self._original_version,
                self._lower_version,
                self._version_under_test)))

    def download_healthtools_package_to_mnode(self, healthtools_version, url=None):
        """Downloads a version of solidfire-healthtools package from jenkins or passed in URL to mnode.

        This fcn doesn't check the validity of URL, you give it a URL, we'll pass it to wget and
        let you deal with consequences.

        :param valence.models.GenericVersion healthtools_version: version to download.
        :param str|None url: Url where the healthtools is at. If none, we'll look it up.
        """
        with self._mnode_fm.get_ssh_connection(quiet=False) as conn:
            if not url:
                msg = 'URL for {} not found in the list of known versions'.format(healthtools_version)
                if healthtools_version not in self._healthtool_urls:
                    raise _errors.HealthtoolsVersionNotFound(msg)
                url = self._healthtool_urls[healthtools_version]

            tar_fname = '{}-{}.tgz'.format(_constants.SOLIDFIRE_HEALTHTOOLS, healthtools_version)
            tar_path = '{}/{}'.format(_constants.URA_PACKAGES_DIR, tar_fname)

            conn.run_command('mkdir -p {}'.format(_constants.URA_PACKAGES_DIR))
            conn.run_command('wget -O {} {}'.format(tar_path, url))

    @staticmethod
    def get_healthtools_at_url(url):
        """Return all available versions of healthtools package at a URL.

        This function will check a URL for all available packages matching solidfire-healthools-x.y.z.w
        notation.

        :param str url: URL to scan for healthtool packages.
        :rtype: list[tuple[valence.models.GenericVersion,str]]
        :returns: List of all available versions and their URLs
        """
        soup = BeautifulSoup(requests.get(url).content, features='html.parser')

        # Extract only <a href> nodes. We'll match them to filename nodes later.
        tgz_files_matches = [node.get('href') for node in soup.findAll('a')]

        # Record all of the files that match the pattern above, and look in subfolders.
        all_versions = list()
        for tgz in tgz_files_matches:
            version_url = '{}/{}'.format(url, tgz)
            version, version_url = HealthtoolsVersionsForTest.get_version_from_url(version_url)
            if version:
                all_versions.append((version, version_url))
        return all_versions

    @staticmethod
    def get_version_from_url(url):
        """This performs a regex match of a URL to the known healthtools version string.

        The function will return the matched GenericVersion string and the URL if match has been found.

        :param str url: URL to check for healthtools version filename
        :rtype: tuple[valence.models.GenericVersion,str]
        :returns: GenericVersion and URL
        """
        if url:
            pattern = re.compile(r'.*?solidfire-healthtools-(\d{4}\.\d{2}\.\d{2}\.\d+)\.tgz$')
            matches = pattern.match(url)
            if matches:
                return models.GenericVersion(matches.group(1)), url
        return None, url
