"""Helper functions for Management Node."""

import re

import autologger
import requests
import sf_ssh

from valence import functional_models
from valence.functional_models import _constants
from valence.vl_helpers import _errors

_log = autologger.get_sflogging()

MNODE_ADMIN_USER_NAME = 'admin'
MNODE_ADMIN_PWD = 'admin'


def _build_cmd(cluster_dm, mnode_username, mnode_password):
    """Configure the mnode2.0 for AFA.

    This function will create the command for the mnode2.0 setup-mnode script as outlined on
    https://confluence.ngage.netapp.com/display/KR/Deploy+and+Configure+MNode+for+AFA

    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param str mnode_username: mnode admin username
    :param str mnode_password: mnode admin pwd
    :rtype: str
    :returns: The command to execute.
    """
    cmd = '{} --mnode_admin_user {} --storage_username {} --storage_mvip {} ' \
          '--storage_svip {} --mnode_admin_password {} --storage_password {} ' \
          '--use_persistent_volumes false --testing true'.format(_constants.SETUP_MNODE_CMD,
                                                                 mnode_username,
                                                                 cluster_dm.username, cluster_dm.mvip,
                                                                 cluster_dm.svip, mnode_password,
                                                                 cluster_dm.password)
    return cmd


def _deploy_config_using_credentials(mnode_fm, cluster_dm, username=MNODE_ADMIN_USER_NAME,
                                     password=MNODE_ADMIN_PWD):
    """Execute the setup command with a given set of credentials.

    This function will execute the mnode2.0 setup-mnode script as outlined on
    https://confluence.ngage.netapp.com/display/KR/Deploy+and+Configure+MNode+for+AFA

    Good thing to know is that if this script is executed on an mnode that's already been configured
    then the script will print a bunch of errors but rc will be 0, so it'll just pass quietly.

    :param valence.functional_models.ManagementNodeBase mnode_fm: management node being configured.
    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param str username: Username to use for mnode admin acct.
    :param str password: Password to use for mnode admin acct.
    :raises: sf_ssh.SshError
    """
    with mnode_fm.get_ssh_connection(quiet=False) as conn:
        cmd = _build_cmd(cluster_dm, username, password)
        rco = conn.run_command(cmd, raise_on_failure=True)
        _log.debug('setup-mnode stdout: {}'.format(rco.stdout))
        if rco.stderr:
            _log.debug('setup-mnode stderr: {}'.format(rco.stderr))
            _log.debug("It is normal to see stderr b/c at2 credentials won't work.")


def configure_mnode2_0_for_afa(mnode_fm, cluster_dm, fail_if_already_configured=False,
                               cleanup_if_already_configured=False):
    """Configure the mnode2.0 for AFA.

    This function will execute the mnode2.0 setup-mnode script as outlined on
    https://confluence.ngage.netapp.com/display/KR/Deploy+and+Configure+MNode+for+AFA

    :param valence.functional_models.ManagementNode2x mnode_fm: management node being configured.
    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param bool fail_if_already_configured: Fail if mnode is already setup.
    :param bool cleanup_if_already_configured: Cleanup if already configured.
    """
    mnode_version = mnode_fm.version
    cluster_fm = functional_models.Cluster(cluster_dm)
    cluster_version = cluster_fm.get_current_version()

    if mnode_version >= _constants.SODIUM_PATCH3 and cluster_version >= _constants.SODIUM_PATCH3:
        # Check if mnode is already configured.
        with mnode_fm.get_ssh_connection(quiet=False) as conn:
            # if setup-mnode hasnt been run then this command will return an error.
            rco = conn.run_command('docker service ls', raise_on_failure=False)
            # If rc=0 then command succeeded and docker has services.
            if not rco.return_code:
                # If mnode has been configured, fail if it's been asked for.
                if fail_if_already_configured:
                    raise _errors.ManagementNodeServicesAlreadyConfigured()
                if cleanup_if_already_configured:
                    raise NotImplementedError("Cleaning up mnode isn't supported yet.")
            else:
                # Unfortunately, all mnodes reserved in at2 have their data models populated with root
                # solidfire credentials. That comes from at2 as input into the tasks, so we can't
                # change it. The setup script fails to run with those credentials and needs admin/admin
                # or some other administrative user.
                # So...first, run the setup script with whatever credentials are in the DM.
                # If that fails, run it with admin/admin. If that fails, so be it, pass the exception on.
                try:
                    _deploy_config_using_credentials(mnode_fm, cluster_dm, mnode_fm.username,
                                                     mnode_fm.password)
                except sf_ssh.SshError:
                    _deploy_config_using_credentials(mnode_fm, cluster_dm)
    else:
        _log.debug('Not executing mnode2.0 configuration script since both mnode and cluster have to be'
                   ' installed at version newer than 11.3')


def setup_mnode_storage_service_for_upgrades(mnode_fm, cluster_dm, asset_name=None):
    """Ensure that mnode2.0 assets has storage MVIP added into it.

    :param valence.functional_models.ManagementNode2x mnode_fm: management node being configured.
    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param str|None asset_name: Which asset to add the storage cluster to. If None, a new asset will
        be added.
    """
    storage_id = mnode_fm.get_storage_asset_id_for_storage_cluster_mvip(cluster_dm)
    if storage_id:
        _log.debug('Storage MVIP {} already exists in mnode assets'.format(cluster_dm.mvip))
    else:
        asset_id = mnode_fm.add_asset_if_needed(asset_name=asset_name)
        mnode_fm.add_storage_id_to_asset(asset_id, cluster_dm)


def get_published_mnode_services_bundle_url(jenkins_url):
    """Interrogate mnode_dev2.x_main Jenkins job for the latest published "fat tar".

    :param str jenkins_url: link to jenkins job from which to grab the artifactory url
    :return: artifactory_url: link to latest service bundle in artifactory, empty string if that fails.
    :rtype: str
    :raises: vl_helpers._errors.HelperError
    """
    try:
        resp = requests.get('{}'.format(jenkins_url)).json()

        artifactory_url = re.findall(r'>(.*?)<', resp.get('description'))
        return artifactory_url[0]

    except (requests.ConnectionError, requests.ConnectTimeout) as err:
        raise _errors.HelperError('unable to reach {} due to {}'.format(jenkins_url, err))


def get_service_bundle_data_from_url(service_bundle_url):
    """Pull service_bundle fat tar from a provided URL.

    :param str service_bundle_url: url pointing to the service_bundle.
    :return: service_bundle_data: ~1.5 GB service bundle data file.
    :rtype: dict
    :raises: vl_helpers._errors.HelperError
    """
    try:
        response = requests.get(url=service_bundle_url, timeout=60)

        service_bundle_data = {'fat_tar': response.content}
        return service_bundle_data

    except (requests.ConnectionError, requests.ConnectTimeout) as err:
        raise _errors.HelperError('Unable to pull the services from {} bundle due to {}'.format(
            service_bundle_url, err))
