"""Helper functions for Node Functional Models."""
import functools
import os
import re
import time
import uuid

import autologger
import pyVim
import requests
import sf_batch
import sf_mc
import sf_models
import sf_ssh
from sf_batch import utils

import sf_api
from camelot.models import at2_enums
from sf_api import api_constants
from valence import (
    enums,
    errors,
    functional_models,
    models,
)
from valence import utils as valence_utils
from valence.vl_helpers import (
    _drive,
    _errors,
)

_log = autologger.get_sflogging(__name__)


def wait_for_node_api_up(node_fm, timeout=120):
    """Wait up to timeout seconds for mvip to become responsive.

    :param valence.functional_models.Node node_fm: Cluster to wait for mvip responsiveness on.
    :param int timeout: Seconds to wait before raising.
    :raises: valence.errors.TimeoutError: If the mvip isn't responsive before the timeout expires.
    """
    start_time = time.time()
    while not node_fm.api_is_responsive():
        if time.time() > start_time + timeout:
            raise errors.TimeoutError('Timed out waiting for API to become responsive.')
        time.sleep(10)


def validate_interface(interface):
    """Raise an error if the interface isn't valid.

    :param sf_api.NetworkInterface interface: Network interface enum
    :raises ValueError: if interface is not instance of sf_api.NetworkInterface
    """
    if interface not in sf_api.NetworkInterface:
        raise ValueError('Expected type: sf_api.NetworkInterface, Given: {}'.format(type(interface)))


def down_network_interfaces(node_fms, interfaces):
    """Take down specified interfaces on each node.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param interfaces: List of Interfaces to take down
    :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
    """
    threads = []
    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.set_network_interface_status_and_verify, interfaces,
                                         status=enums.NetworkStatus.down,
                                         thread_name='set_network_interface_status_down-{}'.format(
                                             node_fm)))
    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    sf_batch.check_for_error_and_raise_exception(threads)


def up_network_interfaces(node_fms, interfaces):
    """Bring up specified interfaces on each node.

    .. note::

        if you try bringing up the 1G, the 10G interface must be available. So, threading a call to
        bring both NICs up could result in a race condition.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param interfaces: List of Interfaces bring up
    :type interfaces: list[valence.enums.NetworkInterface|valence.models.NetworkInterfaceName]
    """
    threads = []
    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.set_network_interface_status_and_verify, interfaces,
                                         status=enums.NetworkStatus.up_and_running,
                                         thread_name='set_network_interface_status_up-{}'.format(
                                             node_fm)))
    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    sf_batch.check_for_error_and_raise_exception(threads)


def add_routes_to_nodes(node_fms, route_dms, interface):
    """Add routes to each node functional model given.

    :param list[valence.functional_models.Node] node_fms: list of node functional models
    :param list[valence.models.Route] route_dms: list of routes data models to add
    :param sf_api.NetworkInterface interface: Network interface enum
    :raises ValueError: if any of the routes have already been added
    """
    validate_interface(interface=interface)

    for node_fm in node_fms:
        # Build a sparse network config to pass to set_network_config
        partial_network_config_dm = models.NetworkConfig()
        partial_network_interface_dm = models.NetworkInterfaceDetails()
        existing_network_config = node_fm.get_network_config()
        existing_routes = []
        if interface is sf_api.NetworkInterface.bond_1g:
            # Make a copy of the existing routes to extend
            if existing_network_config.bond_1g.routes:
                existing_routes = existing_network_config.bond_1g.routes[:]
            partial_network_config_dm.bond_1g = partial_network_interface_dm
        elif interface is sf_api.NetworkInterface.bond_10g:
            if existing_network_config.bond_10g.routes:
                existing_routes = existing_network_config.bond_10g.routes[:]
            partial_network_config_dm.bond_10g = partial_network_interface_dm
        if any(route_dm in existing_routes for route_dm in route_dms):
            raise ValueError('This route has already been added!')
        existing_routes.extend(route_dms)
        partial_network_interface_dm.routes = existing_routes
        node_fm.set_network_config(network_config_dm=partial_network_config_dm)


def remove_routes_from_nodes(node_fms, route_dms, interface):
    """Remove routes from nodes this will remove all routes like those in route_dms from all nodes.

    :param list[valence.functional_models.Node] node_fms: list of node functional models
    :param list[valence.models.Route] route_dms: list of routes data models to add
    :param sf_api.NetworkInterface interface: Network interface enum
    :raises ValueError: when attempting to remove a route that does not exist
    """
    validate_interface(interface=interface)
    for node_fm in node_fms:
        # Build a sparse network config to pass to set_network_config
        new_net_config_dm = models.NetworkConfig()
        existing_network_config_dm = node_fm.get_network_config()
        if interface is sf_api.NetworkInterface.bond_1g:
            existing_network_interface_dm = existing_network_config_dm.bond_1g
        elif interface is sf_api.NetworkInterface.bond_10g:
            existing_network_interface_dm = existing_network_config_dm.bond_10g
        new_net_int_data = existing_network_interface_dm.serialize()
        new_net_interface_dm = models.NetworkInterfaceDetails(new_net_int_data)

        for route in route_dms:
            if (not existing_network_interface_dm.routes or
                    route not in existing_network_interface_dm.routes):
                raise ValueError('You have attempted to remove a route that did not exist')
            new_net_interface_dm.routes.remove(route)
        if interface is sf_api.NetworkInterface.bond_1g:
            new_net_config_dm.bond_1g = new_net_interface_dm
        elif interface is sf_api.NetworkInterface.bond_10g:
            new_net_config_dm.bond_10g = new_net_interface_dm
        node_fm.set_network_config(network_config_dm=new_net_config_dm)


def check_for_core_and_crash_dumps(node_fms, exclude_regexes=None):
    """Check for core dump and crash dump files on each node.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param list[str] exclude_regexes: Do not fail on any core dump that matches one of these
        regular expressions.
    :raises valence.functional_models.FunctionalNodeError: If any cores are found.
    """
    threads = []
    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.core_dump_check, exclude_regexes=exclude_regexes,
                                         thread_name='core_dump_check-{}'.format(node_fm)))
        threads.append(sf_batch.SFThread(node_fm.crash_dump_check, exclude_regexes=exclude_regexes,
                                         thread_name='crash_dump_check-{}'.format(node_fm)))
    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    sf_batch.check_for_error_and_raise_exception(threads)


def api_up_on_nodes(node_fms):
    """Check to see if nodes are pingable.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :rtype: list[bool]
    """
    threads = []
    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.api_is_up, thread_name='api_is_up-{}'.format(node_fm)))
    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    return [t.get_result() for t in threads]


def reset_nodes(node_fms, build=None, reboot=None):
    """Call reset nodes.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param valence.models.SfVersion build: version that the node will be reset to
    :param bool reboot: reboot the node or not
    """
    rtfi_handlers = [node.reset_node(build=build, reboot=reboot) for node in node_fms]
    for handler in rtfi_handlers:
        handler.join()


def get_ensemble_config_on_nodes(node_fms):
    """Return the ensemble configuration for each node.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :rtype: list[valence.enums.ZookeeperType|None]
    """
    threads = []
    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.get_ensemble_config,
                                         thread_name='get_ensemble_config-{}'.format(node_fm)))
    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    sf_batch.check_for_error_and_raise_exception(threads)
    result_stats = [t.get_result().server_stats for t in threads]
    return [r.mode if r else None for r in result_stats]


def reboot_nodes_via_oob(node_fms, wait_for_ping=True, timeout=None):
    """Reboot nodes via OOB/IPMI/iDRAC/Hypervisor.

    This helper utilizes multiprocessing as thread around machine control calls will not work.
    This method will not complete until node is rebooted or timeout is reached.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param bool wait_for_ping: Wait for the node to be pingable
    :param int timeout: time to wait before killing worker processes
    """
    timeout = 600 if timeout is None else timeout
    start_time = time.time()
    workers = []

    for node_fm in node_fms:
        worker = sf_batch.SFProcess(node_fm.reboot_node_via_oob, wait_for_ping=wait_for_ping)
        workers.append(worker)
        worker.start()

    for worker in workers:
        worker.get(timeout=max((0, timeout - (time.time() - start_time))))


def power_off_nodes_via_oob(node_fms, wait_for_power_down=True, timeout=None):
    """Power off nodes via OOB/IPMI/iDRAC/Hypervisor.

    This helper utilizes multiprocessing as thread around machine control calls will not work.
    This method will not complete until OOB status of OFF is returned or timeout is reached.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param bool wait_for_power_down: Wait for power to register 'off'
    :param int timeout: time to wait before killing worker processes
    """
    timeout = 600 if timeout is None else timeout
    start_time = time.time()
    workers = []

    for node_fm in node_fms:
        worker = sf_batch.SFProcess(node_fm.power_off_via_oob, wait_for_power_down)
        workers.append(worker)
        worker.start()

    for worker in workers:
        worker.get(timeout=max((0, timeout - (time.time() - start_time))))


def power_on_nodes_via_oob(node_fms, wait_for_power_up=True, wait_for_ping=True,
                           wait_for_sfapp=True, timeout=None):
    """Power on nodes via OOB/IPMI/iDRAC/Hypervisor.

    This helper utilizes multiprocessing as thread around machine control calls will not work.
    This method will not complete until each node API becomes available or the timeout is reached.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param bool wait_for_power_up: Wait for the power to register as 'on'
    :param bool wait_for_ping: Wait for the node to be pingable
    :param bool wait_for_sfapp: Wait for sfapp to run
    :param int timeout: time to wait before killing worker processes
    """
    timeout = 600 if timeout is None else timeout
    start_time = time.time()
    workers = []
    for node_fm in node_fms:
        wrapped_fnc = utils.one_arg(node_fm.power_on_via_oob, wait_for_ping=wait_for_ping,
                                    wait_for_sfapp=wait_for_sfapp)
        worker = sf_batch.SFProcess(wrapped_fnc, wait_for_power_up)
        workers.append(worker)
        worker.start()

    for worker in workers:
        worker.get(timeout=max((0, timeout - (time.time() - start_time))))


def perform_pdu_power_on_node_operation(mcpdus, timeout=None):
    """Perform a power ON operation on a node.

    Can be applied to nodes with a single PSU or dual PSUs.

    :param list[sf_mc.MachineControl_pdu.McControlPdu] mcpdus: PDUs to perform node operation on
    :param int timeout: time to wait before killing worker processes
    """
    timeout = 10 if timeout is None else timeout
    workers = []
    for host in mcpdus:
        op = host.pdu_immediate_power_on_outlets
        worker = sf_batch.SFProcess(op)
        workers.append(worker)
        worker.start()

    for worker in workers:
        worker.get(timeout=timeout)


def perform_pdu_power_off_node_operation(mcpdus, timeout=None):
    """Perform a power OFF operation on a node.

    Can be applied to nodes with a single PSU or dual PSUs.

    :param list[sf_mc.MachineControl_pdu.McControlPdu] mcpdus: PDUs to perform node operation on
    :param int timeout: time to wait before killing worker processes
    """
    timeout = 10 if timeout is None else timeout
    workers = []
    for host in mcpdus:
        op = host.pdu_immediate_power_off_outlets
        worker = sf_batch.SFProcess(op)
        workers.append(worker)
        worker.start()

    for worker in workers:
        worker.get(timeout=timeout)


def perform_pdu_reboot_node_operation(mcpdus, timeout=None):
    """Perform a Reboot operation on a node.

    Can be applied to nodes with a single PSU or dual PSUs.

    :param list[sf_mc.MachineControl_pdu.McControlPdu] mcpdus: PDUs to perform node operation on
    :param int timeout: time to wait before killing worker processes
    """
    timeout = 10 if timeout is None else timeout
    workers = []
    for host in mcpdus:
        op = host.pdu_immediate_reboot_outlets
        worker = sf_batch.SFProcess(op)
        workers.append(worker)
        worker.start()

    for worker in workers:
        worker.get(timeout=timeout)


def create_cluster(node_fm, mvip, svip, nodes, cluster_name=None, username='admin', password='admin',
                   rep_count=2, attributes=None, unique_id=None, add_drives=True,
                   ignore_unbalanced_cluster_check=None, allow_two_node_ensemble=False,
                   skip_node_type_checks=True):
    """Create a cluster and add drives.

    :param valence.functional_models.Node node_fm: Node to use for cluster creation
    :param str mvip: Floating (virtual) IP address for the cluster on the management network.
    :param str svip: Floating (virtual) IP address for the cluster on the storage (iSCSI) network.
    :param str cluster_name: Name of cluster to set on the nodes / create.
    :param list[valence.models.ResourceNode] nodes: Set of nodes making up the cluster.
    :param str username: User name for the cluster admin.
    :param str password: Initial password for the cluster admin account.
    :param int rep_count: Number of replicas of each piece of data to store in the cluster. Valid
        value is "2".
    :param dict attributes: Dictionary of attributes
    :param str unique_id: You can run either the ListActiveVolumes or GetClusterInfo to find the
        uniqueID.
    :param bool add_drives: Add available drives to newly created cluster.
    :param bool ignore_unbalanced_cluster_check: Whether to skip the check for an unbalanced cluster
    :param bool allow_two_node_ensemble: Whether to allow cluster creation with 2 nodes.
    :param bool skip_node_type_checks: Whether to check node types for 2-node cluster.

    :rtype: valence.functional_models.Cluster
    """
    if not cluster_name:
        cluster_name = 'Cluster-{}'.format(str(uuid.uuid4())[:4])
    api_endpoint = node_fm.get_api().get(api_constants.CURRENT_VERSION)

    for node in nodes:
        new_node_fm = functional_models.Node(node, None)
        new_node_fm.set_cluster_config(cluster_name)

    # Now actually create the cluster
    wait_for_bootstrap(node_fm, nodes)

    node_fm.create_cluster(mvip, svip, nodes, username=username, password=password,
                           rep_count=rep_count,
                           attributes=attributes, unique_id=unique_id,
                           allow_two_node_ensemble=allow_two_node_ensemble,
                           skip_node_type_checks=skip_node_type_checks)
    cluster_dm = models.ResourceCluster({api_constants.NAME: cluster_name,
                                         api_constants.MVIP: mvip,
                                         api_constants.SVIP: svip,
                                         api_constants.USERNAME: username,
                                         api_constants.PASSWORD: password,
                                         api_constants.REP_COUNT: rep_count,
                                         api_constants.UNIQUE_ID: unique_id,
                                         'api_endpoint': api_endpoint})

    cluster_fm = functional_models.Cluster(cluster_dm, authenticate=True)
    if add_drives:
        drive_dms = []
        for node in nodes:
            if not node.is_witness():
                node_fm = functional_models.Node(node, cluster_dm, authenticate=False)
                drive_dms.extend(node_fm.get_hardware_info().drive_hardware)
        _drive.wait_for_drive_status(cluster_fm, drive_dms, check_drive_serial=False)

        _drive.add_available_drives(cluster_fm,
                                    ignore_unbalanced_cluster_check=ignore_unbalanced_cluster_check)

    timer = valence_utils.timeout_timer(120)
    while next(timer):
        bin_assignments = cluster_fm.get_bin_assignments(validate=False)
        if bin_assignments and bin_assignments.master_version:
            return cluster_fm
        time.sleep(10)
    raise errors.TimeoutError('Timed out waiting for initial bin assignments.')


def create_witness_nodes(witness_nodes, build, witness_resources=None):
    """Pass in witness node models to be deployed.

    :param list[valence.models.ResourceNode] witness_nodes: witness node data models to deploy.
    :param sf_models.Build build: Ova build to deploy
    :param valence.models.WitnessResource witness_resources: Custom resource object to use for witness
        deployment
    """
    if not witness_resources:
        witness_resources = models.WitnessResource()
    vsphere = sf_mc.esx_control.Vsphere(witness_resources.vcenter_ip,
                                        witness_resources.vc_user,
                                        witness_resources.vc_pass)
    # TODO: AA-4287. If you get SSL error, set PYTHONHTTPSVERIFY environment variable to 0
    vsphere.connection.vsphere_connect()
    try:
        node = witness_nodes[0]
        target_network_name = node.network_profile.network_name.replace('_', ' ')
        network_name_1g = '{}-MGMT-{}'.format(target_network_name, node.network_profile.one_gig_vlan)
        network_name_10g = '{}-STRG-{}'.format(target_network_name, node.network_profile.ten_gig_vlan)
        ova_url = get_ova_url(build)

        #  Threaded Witness Node deploy
        deploy_partial = functools.partial(
            deploy_witness_vm,
            witness_resources,
            network_name_1g,
            network_name_10g,
            ova_url,
            ovf_params=None)

        deploy_node_threads = []
        for node in witness_nodes:
            deploy_node_threads.append(sf_batch.SFThread(deploy_partial, node.node_name,
                                                         thread_name='_deploy_witness-{}'.
                                                         format(node.node_name)))

        sf_batch.start_all_threads(deploy_node_threads)
        sf_batch.join_all_threads(deploy_node_threads)
        sf_batch.check_for_error_and_raise_exception(deploy_node_threads)

        _log.debug('Witness nodes deployed. Sleeping for 60s to power them up.')
        time.sleep(60)  # Sleep to wait for VM to be powered up
        # TODO: Do this dynamically VL-2380

        configure_witness_node_models(witness_nodes, vsphere)
    finally:
        vsphere.connection.vsphere_disconnect()


def get_next_available_ip(ip_range):
    """Increments IP address to get the next available IP address that is not pingable.

    :param list[str] ip_range: IP range to iterate over
    :return: Incremented IP address that's not in use
    :rtype: str
    """
    for ip in ip_range:
        if os.system('ping -c 1 -W 1 ' + ip) != 0:
            return ip
    raise _errors.HelperError('No available IPs in IP pool provided')


def get_ova_url(build, jenkins_url='http://bdr-jenkins.eng.solidfire.net/OVA/',
                ova_prefix='solidfire-witness'):
    """This function gets a list of all valid witness OVAs then returns the passed in build.

    The OVA returned will be the tip of whatever repository the provided build is from if the
    exact version does not exist

    :param sf_models.Build build: Build to try and find an ova for
    :param str jenkins_url: URL to query for witness node ovas
    :param str ova_prefix: Prefix for witness node ova
    :rtype: str
    :return: Given or latest OVA build
    """
    r = requests.get(jenkins_url, timeout=60)
    ova_prefix = '{}-{}'.format(ova_prefix, build.release)
    ova_url = ova_prefix + '-' + build.version + '.ova'
    # Return all OVAs that match the witness node prefix
    ovas = re.findall(r'">({}-\d{{2}}\.\d\.\d{{1,2}}\.\d{{1,4}}\.ova)</a>'.format(ova_prefix), r.text)
    assert ovas, 'Failed to find any valid witness OVAs on jenkins server.'
    # This sorts the ovas by build number
    if ova_url in ovas:
        return jenkins_url + ova_url
    ova = sorted(ovas, key=lambda ova: ova[ova.find(build.version[:3])])[-1]
    return jenkins_url + ova


def wait_for_bootstrap(node_fm, nodes):
    """Call Get BootStrap Config Nodes and make sure each node can see the other.

    :param list[valence.models.ResourceNode] nodes: List of Node objects which have same Cluster Name.
    :param valence.functional_models.Node node_fm: Target Node Object on which you are
                                                         going to run GetBootStrapConfig API Call.

    :rtype: list[str]
    :return: List of Node's Storage IPs. Value returned from GetBootStrapConfig API Call
    """
    node_sips = [node.sip for node in nodes]
    try_count = 0
    max_tries = 60
    result_node_cips = None
    while try_count < max_tries:
        try:
            result = node_fm.get_bootstrap_config()
            result_node_mips = [node.mip for node in result.nodes]
            result_node_mips = result_node_mips or [node.mipV6 for node in result.nodes]
            result_node_cips = [node.cip for node in result.nodes]
            missing_nodes = [node.mip for node in nodes if node.mip not in result_node_mips]
            if not missing_nodes:
                break
            _log.debug('Currently waiting on missing nodes: {}'.format(missing_nodes))

        except Exception as e:  # pylint: disable=broad-except
            _log.debug('GetBootstrapConfig failed due to: {}'.format(e))
            _log.debug(('Retrying GetBootstrapConfig up to '
                        '{} more times').format(max_tries - try_count))
        try_count += 1
        time.sleep(10)
    else:
        msg = 'Check logs prior for failure.'
        if result_node_cips is not None:
            msg = ('Expected [{}] nodes from bootstrap '
                   'config, received [{}] ').format(len(node_sips), len(result_node_cips))
            msg += 'Expecting: {}, Received: {} '.format(node_sips, result_node_cips)
            # this part of message will isolate the missing node(s) out of above list (set diff operator)
            msg += 'Possible issue with node(s): {}'.format(max(set(node_sips), set(result_node_cips))
                                                            - min(set(node_sips),
                                                                  set(result_node_cips)))
        raise _errors.HelperError('Wait for bootstrap Failed: ' + msg)
    _log.debug('Bootstrapping is successful... Found all the nodes')


def deploy_witness_vm(vc_obj, network_name_1g, network_name_10g, ova_url, witness_name,
                      ovf_params=None):
    """This function creates a witness node virtual machine.

    If you want to use node in your test code, be sure to append it to your resources /
    nodes list. Be sure to sleep after this call before retrieving the VM by name in order to
    get the ips populated.

    :param valence.models.WitnessResource vc_obj: vcenter object with fields to deploy
    :param str network_name_1g: Network name / VLAN to set VM on
    :param str network_name_10g: Network name / VLAN to set VM on
    :param str ova_url: NLocation of OVA to deploy
    :param str witness_name: Name of Witness VM
    :param dict ovf_params: Dictionary with additional ovf parameters

    :rtype: None

    kwargs pass in extra parameters for OVFtool such as specifying a proxy server or chunksize
    """
    def format_command():
        """Returns a command to create VM based on args of parent func.

        :return: ovftool command to create virtual machine
        :rtype: str
        """
        cmd_template = ('ovftool --acceptAllEulas --skipManifestCheck --disableVerification '
                        '--noSSLVerify --powerOn -dm=thin -ds="{}" -n="{}" --net:Management="{}" '
                        '--net:Storage="{}" --vmFolder="{}" {} "{}"')
        cmd = cmd_template.format(vc_obj.data_store, witness_name,
                                  network_name_1g, network_name_10g,
                                  vc_obj.vm_folder, ova_url, vc_obj.resource_folder)
        if ovf_params:
            for key, value in ovf_params.items():
                cmd += ' --{} {}'.format(key, str(value))
        return cmd

    command = format_command()
    _log.debug('Creating VM with name {}. This will take a while. Check vCenter'
               'for progress'.format(witness_name))

    with sf_ssh.Ssh(vc_obj.deploy_vm, username=vc_obj.deploy_username,
                    password=vc_obj.deploy_password) as ssh_conn:
        ssh_conn.run_command(command, use_sudo=False, raise_on_failure=True)


def delete_witness_vms(nodes, witness_resource=None):
    """Helper function to delete witness nodes.

    :param list[valence.models.ResourceNode] nodes: List of node objects to delete witnesses on.
    :param valence.models.WitnessResource witness_resource: Vcenter object
    """
    witness_nodes = [node for node in nodes if node.is_witness()]
    if not witness_resource:
        witness_resource = models.WitnessResource()
    if not witness_nodes:
        return
    service_instance = pyVim.connect.SmartConnect(host=witness_resource.vcenter_ip,
                                                  user=witness_resource.vc_user,
                                                  pwd=witness_resource.vc_pass,
                                                  )
    for node in witness_nodes:
        vm = service_instance.content.searchIndex.FindByIp(None, node.mip, True)
        _log.debug('Found: {0}'.format(vm.name))
        _log.debug('The current powerState is: {0}'.format(vm.runtime.powerState))
        _log.debug('Attempting to power off {0}'.format(vm.name))
        vm.PowerOffVM_Task()
        time.sleep(60)
        _log.debug('Destroying VM from vSphere.')
        vm.Destroy_Task()


def get_vm_network_config(vms):
    """Helper function return a mapping of witness node names to AvailableNodes.

    :param list[dict] vms: Vms to map available nodes to.
    :rtype: dict[str, valence.models.NetworkConfig]

    :raises: valence._errors.HelperError if no IP is able to be connected to.
    """
    vm_network_config_map = {}
    for vm in vms:
        for ip in vm['ips']:
            node = functional_models.Node(models.ResourceNode.create(mip=ip), None)
            if node.api_is_up():
                vm_network_config_map[vm['name']] = node.get_network_config()
                break
        else:
            raise _errors.HelperError('No IP was able to serve the API, did the node deploy correctly?')
    return vm_network_config_map


def configure_witness_node_models(witness_nodes, vsphere):
    """Helper function to change fields on witness node models.

    :param list[valence.models.ResourceNode] witness_nodes: List of witness node objects to set fields on
    :param sf_mc.hypervisor_control.Vsphere vsphere: Vsphere Object that has already set up a connection.
    """

    vms = vsphere.find_vms_by_name([witness_node.node_name for witness_node in witness_nodes])
    nw_config_vm_map = get_vm_network_config(vms)

    for witness_node in witness_nodes:
        witness_network = nw_config_vm_map[witness_node.node_name]
        witness_dm = models.ResourceNode.create(
            mip=witness_network.net0.address,
            sip=witness_network.net1.address
        )
        witness_node_fm = functional_models.Node(witness_dm, None)

        sip_nw_dm = models.NetworkConfig(dict(
            net1=dict(address=witness_node.sip, gateway=witness_node.network_profile.gateway_10g)))
        current_config = witness_node_fm.set_network_config(sip_nw_dm)
        witness_node_fm.node_dm.sip = witness_node.sip
        if current_config[api_constants.NET0].address_v6:
            witness_node.mipV6 = current_config[api_constants.NET0].address_v6
            witness_node.gateway_v6 = current_config[api_constants.NET0].gateway_v6
        mip_nw_dm = models.NetworkConfig(dict(net0=dict(address=witness_node.mip)))
        witness_node_fm.set_network_config(mip_nw_dm)
        config = models.ClusterConfig(dict(name=witness_node.node_name))
        witness_node_fm.set_config(config_dm=config)


def convert_pending_nodes_to_active_nodes(cluster_fm, pending_nodes):
    """Convert the list of pending nodes to active nodes.

    :param valence.functional_models.Cluster cluster_fm: The cluster functional model
    :param list[valence.models.PendingNode] pending_nodes: list of pending nodes to convert
    :return: list of active nodes that used to be pending.
    :rtype: list[valence.models.ActiveNode]
    """
    all_active_nodes = cluster_fm.list_active_nodes()
    active_nodes = [anode for anode in all_active_nodes for pnode in pending_nodes
                    if pnode.name == anode.name]
    assert len(active_nodes) == len(pending_nodes)
    return active_nodes


def provision_nodes(node_fms, build):
    """Call inplace rtfi on nodes, and then call traditional rtfi on nodes if inplace fails.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param valence.models.SfVersion build: version that the node will rtfi to
    :raises HelperError: if any node fails both rtfi attempts
    """
    failed_nodes = _try_itrfi(build, node_fms)

    ers = _retry_trtfi(build, failed_nodes)

    if ers:
        raise _errors.HelperError('Node provisioning finished with the following errors {}'.format(ers))


def _retry_trtfi(build, failed_nodes):
    """Call traditional rtfi on nodes if inplace fails.

    :param list[valence.functional_models.Node] failed_nodes: nodes to retry trtfi on
    :param valence.models.SfVersion build: version that the node will rtfi to

    :return: list of errors of any if any nodes failed trtfi
    :rtype: list[valence.functional_models.AsyncHandlerError|valence.errors.TimeoutError]
    """
    retry_handlers = []
    for node in failed_nodes:
        _log.debug('Inplace rtfi failed for node {} trying traditional.'.format(node.node_dm))
        retry_handlers.append(node.start_traditional_rtfi(build))
    ers = []
    for i, handler in enumerate(retry_handlers):
        try:
            handler.join()
        except (functional_models.AsyncHandlerError, errors.TimeoutError) as er:
            _log.error('tRTFI Provisioning failed on node {}'.format(failed_nodes[i].node_dm))
            ers.append(er)
    _log.debug('Nodes have completed the tRTFI.')
    return ers


def _try_itrfi(build, node_fms):
    """Call inplace rtfi on nodes.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :param valence.models.SfVersion build: version that the node will rtfi to

    :return: nodes that failed irtfi
    :rtype: list[valence.functional_models.Node]
    """
    rtfi_handlers = [node.start_rtfi(build) for node in node_fms]
    failed_nodes = []
    for i, handler in enumerate(rtfi_handlers):
        try:
            handler.join()
        except (functional_models.AsyncHandlerError, errors.TimeoutError):
            _log.error('iRTFI Provisioning failed on node {}'.format(node_fms[i].mip))
            failed_nodes.append(node_fms[i])
    _log.debug('iRTFI has been completed.')
    return failed_nodes


def sort_nodes_by_size(node_fms):
    """Get a list of nodes sorted by the node's block capacity from smallest to largest.

    :param list[valence.functional_models.Node] node_fms: Node functional models.
    :rtype: list[valence.functional_models.Node]
    """
    mip_size_map = {}
    for node_fm in node_fms:
        drive_config = node_fm.get_drive_config()
        node_capacity = sum(d.size for d in drive_config.drives if d.drive_type == enums.DriveType.block)
        mip_size_map[node_fm.mip] = node_capacity

    sorted_mips = sorted(mip_size_map.keys(), key=mip_size_map.get)
    mip_node_map = {node_fm.mip: node_fm for node_fm in node_fms}
    return [mip_node_map[mip] for mip in sorted_mips]


def list_nodes_by_usable_block_capacity(cluster_fm):
    """List the nodes on the cluster sorted by their usable block capacity from smallest to largest.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional models.
    :rtype: list[valence.functional_models.Node]
    """
    drive_dms = cluster_fm.list_drives(
        statuses=[enums.DriveStatus.active],
        drive_types=[enums.DriveType.block]
    )
    node_dms = cluster_fm.list_active_nodes(node_roles=[enums.NodeRole.storage])
    node_fms = [functional_models.Node(dm.resource_dm, cluster_fm.cluster_dm) for dm in node_dms]
    node_id_map = {dm.mip: dm.node_id for dm in node_dms}
    mip_size_map = {}
    mip_node_map = {}
    for node_fm in node_fms:
        node_capacity = sum(
            dm.usable_capacity for dm in drive_dms if dm.node_id == node_id_map[node_fm.mip]
        )
        mip_size_map[node_fm.mip] = node_capacity
        mip_node_map[node_fm.mip] = node_fm

    sorted_mips = sorted(mip_size_map.keys(), key=mip_size_map.get)
    return [mip_node_map[mip] for mip in sorted_mips]


def create_witness_model(one_gig, ten_gig, node_name=None, witness_resource=None):
    """Create witness ResourceNode model.

    :param valence.models.ResourceIp one_gig: IP for the mip.
    :param valence.models.ResourceIp ten_gig: IP for the sip.
    :param str node_name: Name for witness node. Default is WitnessNode-<> if not passed in.
    :param valence.models.WitnessResource witness_resource: Vsphere to deploy witness on.
    :rtype: list[valence.models.ResourceNode]
    """
    if not witness_resource:
        witness_resource = models.WitnessResource()
    new_node = models.ResourceNode.create(
            model=sf_models.NodeType(sf_models.NodeType.STANDARD_TYPES.WITNESS),
            node_name=node_name or 'WitnessNode-{}'.format(str(uuid.uuid4())[:4]),
            username='root',
            password='solidfire',
            network_profile=one_gig.network_profile,
            mip=one_gig.address,
            sip=ten_gig.address,
            hypervisor_host=dict(management_ip=witness_resource.vcenter_ip,
                                 username=witness_resource.vc_user,
                                 password=witness_resource.vc_pass,
                                 type=witness_resource.type,
                                 name=witness_resource.name),
            resource_type_name='node',
            fibre_channel=at2_enums.FibreChannel.none)
    return new_node


def get_file_contents_from_node(node_fm, target_dir, file_name, timeout):
    """Cat a file on a node, return the response joined by newline.

    Why?  Because cert components need to have keys and certs off the nodes.

    :param valence.functional_models.Node node_fm: node to cat file on
    :param str target_dir: directory the file is in
    :param str file_name: file name to cat
    :param int timeout: timeout for ssh command

    :rtype: str
    """
    with node_fm.get_ssh_connection() as ssh_inst:
        run_command_output = ssh_inst.run_command(
            'cat {}'.format(os.path.join(target_dir, file_name)), timeout=timeout, raise_on_failure=True)
        return run_command_output.stdout


def reload_nginx(node_fm):
    """Force reloading of nginx on a node.

    :param valence.functional_models.Node node_fm: node functional model
    :raises HelperError: if the return code != 0
    """
    version = node_fm.api.version
    if version >= models.SfVersion('10.2'):
        nginx_service_name = 'nginx'
    else:
        nginx_service_name = 'sfnginx'
    with node_fm.get_ssh_connection() as ssh_inst:
        run_command_output = ssh_inst.run_command('service {} reload'
                                                  .format(nginx_service_name))
        std_out = run_command_output.stdout.splitlines()
        std_err = run_command_output.stderr.splitlines()
        rc = run_command_output.return_code
        if rc != 0:
            raise _errors.HelperError(
                ('Reloading nginx service failed! std_out={}, std_err={}, rc={}'
                 ' Please note, this probably killed the web interface.  Good luck'
                 ).format(std_out, std_err, rc))


def remove_user_certs(node_fm):
    """Remove any user.key and user.crt certificates from a node.

    :param valence.functional_models.Node node_fm: node functional model
    """

    with node_fm.get_ssh_connection() as ssh_inst:
        ssh_inst.run_command('rm /sf/etc/ssl/user.crt', raise_on_failure=True)
        ssh_inst.run_command('rm /sf/etc/ssl/user.key', raise_on_failure=True)


def revert_cert_via_cli(node_fm):
    """Revert a node to the cluster certificate using the cli.

    :param valence.functional_models.Node node_fm: node functional model
    """
    active_certs = []
    with node_fm.get_ssh_connection() as ssh_inst:
        run_command_output = ssh_inst.run_command('ls -al /sf/etc/ssl', raise_on_failure=True)
        certs = run_command_output.stdout.splitlines()
        for cert in certs:
            if cert.find('-> ') > 0:
                active_certs.append(cert[cert.find('-> '):].replace('-> /sf/etc/ssl/', ''))
        if 'cluster.crt' in active_certs and 'cluster.key' in active_certs:
            raise _errors.HelperError(
                'Cannot revert certs via API, already using the cluster certs'
            )
        else:
            ssh_inst.run_command(
                'ln -s -f /sf/etc/ssl/cluster.crt /sf/etc/ssl/active.crt', raise_on_failure=True)
            ssh_inst.run_command(
                'ln -s -f /sf/etc/ssl/cluster.key /sf/etc/ssl/active.key', raise_on_failure=True)
    remove_user_certs(node_fm)
    reload_nginx(node_fm)


def test_hardware_config(node_fms):
    """Test the hardware config.

    :param list[valence.functional_models.Node] node_fms: node functional models
    :raises valence.functional_models.FunctionalNodeError: when any of the tests fail and
        flag is set
    """

    threads = []
    for node_fm in node_fms:
        threads.append(sf_batch.SFThread(node_fm.test_hardware_config, raise_on_failure=True,
                                         thread_name='test_hardware_config'))
    sf_batch.start_all_threads(threads)
    sf_batch.join_all_threads(threads)
    sf_batch.check_for_error_and_raise_exception(threads)


def flash_and_verify_nvdimm_fw(node_fm, downgrade=False, expected_version=None):
    """Perform an NVDIMM firmware flash operation and verify_nvdimm_health.

    :param bool downgrade: if set to True, will downgrade the firmware
    :param str expected_version: Firmware Version number
    :param valence.functional_models.Node node_fm: node functional model
    """
    node_fm.flash_nvdimm_firmware(downgrade=downgrade, expected_version=expected_version)
    node_fm.verify_nvdimm_health()
