"""Helper that runs preupgrade healthchecks and returns a boolean result for pass/fail."""
from __future__ import absolute_import

import autologger
import sf_ssh

from valence import (
    camelot,
    functional_models,
)
from valence.functional_models import _constants
from valence.vl_helpers import (
    _errors,
    _healthtools_versions_for_upgrade_suites,
    _sfupgradecheck,
    _sfupgradecheck_validation,
)

_log = autologger.get_sflogging(__name__)


def _setup_healthtools_on_mnode(mnode_fm,
                                healthtools_helper,
                                url_to_healthtools_to_deploy=None,
                                install_rqd_healthtools=False):
    """Setup healthtools on mnode depending on few parameters.

    :param valence.functional_models.ManagementNodeLegacy mnode_fm: mnode_fm
    :param valence.vl_helpers.HealthtoolsVersionsForTest healthtools_helper: helper monkey.
    :param str url_to_healthtools_to_deploy: which healthtools to deploy
    :param bool install_rqd_healthtools: Install latest healthtools from rqd/ on jenkins.
    :raises: AssertionError
    """
    if url_to_healthtools_to_deploy:
        version, url = healthtools_helper.get_version_from_url(url_to_healthtools_to_deploy)
        healthtools_helper.download_healthtools_package_to_mnode(version, url)
        mnode_fm.install_new_healthtools_version(version, None)
    else:
        healthtools_helper.initialize_versions(None, error_if_not_on_jenkins=False)
        if install_rqd_healthtools:
            assert healthtools_helper.jenkins_rqd_versions
            rqd_version = healthtools_helper.jenkins_rqd_versions[-1]
            healthtools_helper.download_healthtools_package_to_mnode(rqd_version)
            mnode_fm.install_new_healthtools_version(rqd_version, None)


def _validate_preupgrade_checks(mnode_fm, cluster_fm, healthtools_helper, processed_checks):
    """Runs preupgrade checks and returns the result.

    :param valence.functional_models.ManagementNodeLegacy mnode_fm: mnode_fm
    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.vl_helpers.HealthtoolsVersionsForTest healthtools_helper: helper monkey.
    :param valence.functional_models.ProcessedHealthchecks processed_checks: parsed checks data.
    :rtype: bool
    :returns: Pass or Fail result of the checks.
    """
    result = False
    try:
        _log.info(camelot.banner('Verifying output of preupgrade healthchecks: \n{}'.format(
            processed_checks)))
        active_nodes = cluster_fm.list_active_nodes()
        current_ht_version = healthtools_helper.current_version
        mnode_version = mnode_fm.version
        _sfupgradecheck_validation.validate_sfupgradecheck_results(
            mnode_version=mnode_version,
            processed_checks=processed_checks,
            active_nodes=active_nodes,
            healthtools_version=current_ht_version)
        result = True
    except (sf_ssh.SshResponseError, _errors.HelperError) as e:
        _log.debug('Received error from sfupgradecheck call. '
                   'Healthchecks may not have run correctly: {}'.format(e))
    return result


def run_preupgrade_healthchecks(mnode_fm, cluster_dm, url_to_healthtools_to_deploy=None,
                                install_rqd_healthtools=False):
    """Runs healthchecks and validates the that cluster is ready for upgrade.

    This helper will return True if checks pass, and False if they do not or failed to run.
    True will be also returned if healthchecks aren't supported by this version of mnode/cluster.
    For RQD suites, this helper will install a latest version of healthtools that should be used
    by RQD testing.

    :param valence.functional_models.ManagementNodeLegacy mnode_fm: mnode_fm
    :param valence.models.ResourceCluster cluster_dm: Cluster data model.
    :param str url_to_healthtools_to_deploy: which healthtools to deploy
    :param bool install_rqd_healthtools: Install latest healthtools from rqd/ on jenkins.
    :rtype: bool
    :return: True if checks ran and passed, else False.
    """
    did_checks_pass = True
    mnode_version = mnode_fm.version
    if mnode_version >= _constants.FIRST_VERSION_WITH_SFUPDATE_HEALTHTOOLS:
        cluster_fm = functional_models.Cluster(cluster_dm)
        cluster_fm.authenticate()
        healthtool_helper = _healthtools_versions_for_upgrade_suites.HealthtoolsVersionsForTest(mnode_fm)
        _setup_healthtools_on_mnode(mnode_fm, healthtool_helper, url_to_healthtools_to_deploy,
                                    install_rqd_healthtools)

        output = mnode_fm.execute_preupgrade_health_check(
            cluster_dm, skip_version_check=True, raise_on_failure=True)
        _sfupgradecheck.prettyprint_sfupgradecheck_json_to_text(output)
        did_checks_pass = _validate_preupgrade_checks(mnode_fm, cluster_fm, healthtool_helper, output)
    else:
        _log.debug('{} mnode does not support executing preupgrade healthchecks.'.format(mnode_version))

    return did_checks_pass
