"""Helper functions for replication related operations."""

import time

import autologger
import six

from valence import (
    criteria,
    enums,
    models,
    utils,
)
from valence.vl_helpers import _errors

_log = autologger.get_sflogging()


def add_netem_rules_inter_nodes(src_node_fms, device_target_ip_map, rules):
    """Add a netem rule to perturb network traffic between nodes.

    :param list[valence.functional_models.Node] src_node_fms: list of nodes to apply
        the rules against
    :param dict[str,list[str]] device_target_ip_map: a map of network device names to IPs in which to
        apply the rule
    :param list[str] rules: netem rule to use for perturbation (e.g. delay 50ms)
    """

    # This operation re-routes all traffic to a separate un-filtered band.
    mv_cmd = 'tc qdisc add dev {dev} root handle 1: prio priomap 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2'

    # This operation applies the rule to the original band, which shouldn't have any traffic yet
    delay_cmd = 'tc qdisc add dev {dev} parent 1:1 handle 10: netem {rule}'

    # This applies a filter to move only the desired IP to the filtered band
    dest_filter_cmd = ('tc filter add dev {dev} protocol ip parent 1:0 prio 1 u32 '
                       'match ip dst {dst} flowid 1:1')

    for node_fm in src_node_fms:

        with node_fm.get_ssh_connection() as ssh_conn:

            for dev_name, target_ips in six.iteritems(device_target_ip_map):
                _log.debug('Setting traffic shaping rules on {} adapter: {};\n   rule(s): {}'.format(
                    node_fm.node_dm.node_name, dev_name, rules
                ))

                # Move all traffic to a different band
                ssh_conn.run_command(mv_cmd.format(dev=dev_name), raise_on_failure=True)

                # Add the specified rules to the original band
                rule_string = ' '.join(rules)
                ssh_conn.run_command(delay_cmd.format(dev=dev_name, rule=rule_string),
                                     raise_on_failure=True)

                for target_ip in target_ips:
                    # Add a filter to move the target IP to the original band with the new rule
                    ssh_conn.run_command(dest_filter_cmd.format(dev=dev_name, dst=target_ip),
                                         raise_on_failure=True)


def clear_netem_rules(src_node_fms, devices):
    """Clear any netem rules for the specified list of network devices.

    :param list[valence.functional_models.Node] src_node_fms: list of nodes to apply the rules against
    :param list[str] devices: list of network device names
    """
    clear_cmd = 'tc qdisc del root dev {dev}'

    for node_fm in src_node_fms:

        with node_fm.get_ssh_connection() as ssh_conn:

            for dev_name in devices:
                _log.debug('Clearing traffic shaping rules on {} adapter: {}'.format(
                    node_fm.node_dm.node_name, dev_name
                ))
                # Delete any existing rules
                ssh_conn.run_command(clear_cmd.format(dev=dev_name))


def get_group_snapshots_remote_statuses(cluster_fm, filter_group_snapshot_dms=None):
    """Get the remote statuses of group snapshots on a source cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.GroupSnapshot] filter_group_snapshot_dms: Group snapshot data models to
        filter on
    :returns: List of group snapshot statuses
    :rtype: list[valence.enums.RemoteClusterSnapshotStatus]
    """
    remote_statuses = []
    group_snapshot_dms = _get_and_filter_group_snap_dms(
        cluster_fm, filter_group_snapshot_dms=filter_group_snapshot_dms)

    for group_snapshot_dm in group_snapshot_dms:
        if group_snapshot_dm.remote_statuses:
            for status in group_snapshot_dm.remote_statuses:
                remote_statuses.append(status.remote_status)

    return remote_statuses


def get_snapshots_remote_statuses(cluster_fm, filter_snapshot_dms=None):
    """Get the remote statuses of snapshots on a source cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Snapshot] filter_snapshot_dms: Snapshot data models to filter on
    :returns: List of snapshot statuses
    :rtype: list[valence.enums.RemoteClusterSnapshotStatus]
    """
    remote_statuses = []
    snapshot_dms = cluster_fm.list_snapshots()

    if filter_snapshot_dms:
        filter_uuids = [snapshot_dm.snapshot_uuid for snapshot_dm in filter_snapshot_dms]
        snapshot_filter = criteria.ContainedIn(models.Snapshot.snapshot_uuid, filter_uuids)
        snapshot_dms = snapshot_filter.filter(snapshot_dms)

    for snapshot_dm in snapshot_dms:
        if snapshot_dm.remote_statuses:
            for status in snapshot_dm.remote_statuses:
                remote_statuses.append(status.remote_status)

    return remote_statuses


def wait_for_group_snapshots_replication(cluster_fm, filter_group_snapshot_dms=None, timeout=None,
                                         poll_period=60):
    """Wait for group snapshots on a source cluster to complete replication.

    MUST be passed the source cluster because the target cluster does not show remote statuses of
    group snapshots. Default to 90 minute timeout if timeout is not provided.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.GroupSnapshot] filter_group_snapshot_dms: Group snapshot data models to
        filter on
    :param int timeout: Time to wait in seconds before raising an exception
    :param int poll_period: Time to wait in seconds between each poll
    :raises valence.vl_helpers.HelperError: If group snapshots replication does not finish
    """
    timeout = 90 * 60 if timeout is None else timeout

    group_snapshot_dms = _get_and_filter_group_snap_dms(
        cluster_fm, filter_group_snapshot_dms=filter_group_snapshot_dms)

    member_snap_dms = []
    for grp in group_snapshot_dms:
        member_snap_dms.extend(grp.members)

    timer = utils.timeout_timer(timeout)

    start_time = time.time()
    while next(timer):
        remote_statuses = get_group_snapshots_remote_statuses(cluster_fm, filter_group_snapshot_dms)

        if not remote_statuses:
            raise _errors.HelperError(
                'Remote statuses for group snapshots could not be found on {}'.format(
                    cluster_fm.cluster_dm))

        status_set = set(remote_statuses)
        present_status = enums.RemoteClusterSnapshotStatus.present

        # First check to see if the groups themselves report being present on the target.
        if status_set == {present_status}:
            member_remote_statuses = get_snapshots_remote_statuses(cluster_fm,
                                                                   filter_snapshot_dms=member_snap_dms)

            member_status_set = set(member_remote_statuses)

            # Then check the individual member snapshots (as in some edge use cases the group can be
            # in transition and the member snapshots are being manipulated on the target).
            if member_status_set == {present_status}:
                _log.debug('Group snapshots replication finished after {} seconds.'.format(
                    int(time.time() - start_time)))
                return

        _log.debug('Waiting for group snapshots replication to finish. Sleeping {} seconds.'.format(
            poll_period))
        time.sleep(poll_period)

    raise _errors.HelperError(
        'Timed out waiting for group snapshots replication to finish after {} seconds'.format(timeout))


def wait_for_snapshots_replication(cluster_fm, filter_snapshot_dms=None, timeout=None, poll_period=60):
    """Wait for snapshots replication on a source cluster to complete replication.

    MUST be passed the source cluster because the target cluster does not show remote statuses of
    snapshots. Default to 30 minute timeout if timeout is not provided.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Snapshot] filter_snapshot_dms: Snapshot data models to
        filter on
    :param int timeout: Time to wait in seconds before raising an exception
    :param int poll_period: Time to wait in seconds between each poll
    :raises valence.vl_helpers.HelperError: If snapshots replication does not finish
    """
    timeout = 30 * 60 if timeout is None else timeout
    timer = utils.timeout_timer(timeout)

    start_time = time.time()
    while next(timer):
        remote_statuses = get_snapshots_remote_statuses(cluster_fm, filter_snapshot_dms)
        if not remote_statuses:
            raise _errors.HelperError(
                'Remote statuses for snapshots could not be found on {}'.format(
                    cluster_fm.cluster_dm))

        status_set = set(remote_statuses)
        if len(status_set) == 1 and enums.RemoteClusterSnapshotStatus.present in status_set:
            _log.debug('Snapshots replication finished after {} seconds.'.format(
                int(time.time() - start_time)))
            return

        _log.debug('Waiting for snapshots replication to finish. Sleeping {} seconds.'.format(
            poll_period))
        time.sleep(poll_period)

    msg = 'Timed out waiting for snapshots replication to finish after {} seconds'.format(timeout)
    if filter_snapshot_dms:
        msg += ':\nSnapshots: {}'.format(filter_snapshot_dms)

    raise _errors.HelperError(msg)


def get_peer_snapshot(cluster_peer_fm, snapshot_dm):
    """Get the snapshot data model on the peer side for a given source snapshot.

    :param valence.functional_models.Cluster cluster_peer_fm: Peer cluster functional model.
    :param valence.models.Snapshot snapshot_dm: Snapshot data model.
    :returns: Paired target snapshot data model
    :rtype: valence.models.Snapshot

    :raises valence.vl_helpers.errors.HelperError: When matching snapshot not located on peer or
        multiple are found
    """
    peer_snapshot_dms = cluster_peer_fm.list_snapshots()
    paired_filter = criteria.Equals(models.Snapshot.snapshot_uuid,
                                    snapshot_dm.snapshot_uuid)
    filtered_peer_snapshot_dms = paired_filter.filter(peer_snapshot_dms)

    if filtered_peer_snapshot_dms and len(filtered_peer_snapshot_dms) == 1:
        return filtered_peer_snapshot_dms[0]
    elif filtered_peer_snapshot_dms and len(filtered_peer_snapshot_dms) > 1:
        msg = 'Multiple Snapshots with UUID {} found on peer cluster {}'
    else:
        msg = 'Unable to locate Snapshot {} on peer cluster {}'

    raise _errors.HelperError(msg.format(snapshot_dm.snapshot_uuid, cluster_peer_fm.cluster_dm.name))


def get_peer_group_snapshot(cluster_peer_fm, group_snapshot_dm):
    """Get the group snapshot data model on the peer side for a given source group snapshot.

    :param valence.functional_models.Cluster cluster_peer_fm: Peer cluster functional model.
    :param valence.models.GroupSnapshot group_snapshot_dm: Group snapshot data model.
    :returns: Paired target group snapshot data model
    :rtype: valence.models.GroupSnapshot

    :raises valence.vl_helpers.errors.HelperError: When matching group snapshot not located on peer or
        multiple are found
    """
    peer_group_snapshot_dms = cluster_peer_fm.list_group_snapshots()
    paired_group_filter = criteria.Equals(models.GroupSnapshot.group_snapshot_uuid,
                                          group_snapshot_dm.group_snapshot_uuid)
    filtered_peer_group_snapshot_dms = paired_group_filter.filter(peer_group_snapshot_dms)

    if filtered_peer_group_snapshot_dms and len(filtered_peer_group_snapshot_dms) == 1:
        return filtered_peer_group_snapshot_dms[0]
    elif filtered_peer_group_snapshot_dms and len(filtered_peer_group_snapshot_dms) > 1:
        msg = 'Multiple Group Snapshots with UUID {} found on peer cluster {}'
    else:
        msg = 'Unable to locate Group Snapshot {} on peer cluster {}'

    raise _errors.HelperError(msg.format(group_snapshot_dm.group_snapshot_uuid,
                                         cluster_peer_fm.cluster_dm.name))


def _get_and_filter_group_snap_dms(cluster_fm, filter_group_snapshot_dms=None):
    """Get the current list of group snaps and optionally filter given a specified list.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.GroupSnapshot] filter_group_snapshot_dms: Group Snapshot data models to
        filter on
    :returns: Current list of Group Snapshots based on the filter (all if no filter provided)
    :rtype: list[valence.models.GroupSnapshot]
    """
    group_snapshot_dms = cluster_fm.list_group_snapshots()

    if filter_group_snapshot_dms:
        filter_uuids = [group_dm.group_snapshot_uuid for group_dm in filter_group_snapshot_dms]
        group_snapshot_filter = criteria.ContainedIn(models.GroupSnapshot.group_snapshot_uuid,
                                                     filter_uuids)

        group_snapshot_dms = group_snapshot_filter.filter(group_snapshot_dms)
    return group_snapshot_dms
