"""Helper functions for Security Features."""
import time

import autologger
import sf_models
import sf_ssh
import six
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from sf_api import api_errors
from valence import (
    enums,
    errors,
    functional_models,
    utils,
)
from valence.utils import random_string
from valence.vl_helpers import (
    _bin_sync,
    _errors,
)

_log = autologger.get_sflogging()


def setup_external_key_management(cluster_fm, eks_dms, key_provider_name='EKM',
                                  key_server_tag=None, subject=None, start_date=None, end_date=None,
                                  reuse_keys=True, force=False):
    """All the steps to configure EKM in Element.

    1. Create public/private key pair
    2. Create Key Provider
    3. Generate CSR
    4. Sign the CSR with each single node EKS or clustered EKS
    5. Create a Key Server for each single node EKS or clustered EKS
    6. Add those Key Servers to the Key Provider

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param list[sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster] eks_dms: List of EKS DMs
    :param str key_provider_name: Optional name of Key Provider
    :param str key_server_tag: Tag that will be appended to the Key Server name
    :param valence.models.CertificateSubject subject: CertificateSubject DM
    :param datetime.datetime start_date: Datetime object representing the client cert start date
    :param datetime.datetime end_date: Datetime object representing the client cert expiration date
    :param bool reuse_keys: If True and key pair exists, do not try to create new ones.
    :param bool force: If True, tear down existing EKM configuration.
    :rtype: valence.models.KeyProviderKmip
    :raises: valence.vl_helpers._errors.HelperError
    """

    # At the moment, EKM can only support one EKS or one EKS cluster.
    # ELEM-6456 will remove this exception once multiple key servers are supported.
    if len(eks_dms) > 1:
        err_msg = 'EKM can only support one EKS or one EKS cluster, but {} were passed in'.format(
            len(eks_dms))
        raise _errors.HelperError(err_msg)

    _create_key_pair(cluster_fm, subject, reuse_keys, force)
    key_provider_dm = cluster_fm.create_key_provider_kmip(key_provider_name)
    csr = cluster_fm.get_client_certificate_sign_request().client_certificate_sign_request

    for eks_dm in eks_dms:
        key_server_dm = create_key_server(cluster_fm, eks_dm, csr, start_date, end_date,
                                          key_server_tag)
        cluster_fm.add_key_server_to_provider_kmip(key_server_dm, key_provider_dm)

    return cluster_fm.get_key_provider_kmip(key_provider_id=key_provider_dm.id)


def _create_key_pair(cluster_fm, subject, reuse_keys, force):
    """Create the public/private key pair.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param valence.models.CertificateSubject subject: CertificateSubject DM
    :param bool reuse_keys: If True and key pair exists, do not try to create new ones.
    :param bool force: If True, tear down existing EKM configuration.
    """
    try:
        cluster_fm.create_public_private_key_pair(subject)
    except api_errors.KeyPairInUse as e:
        # If KeyPairInUse and force=True, teardown EKM and try again.
        # If reuse_keys=True, continue on.
        # Else rethrow exception.
        if force:
            _log.debug('Key pair in use. Tearing down existing EKM providers and trying again.')
            teardown_external_key_management(cluster_fm, disable_ear=True)
            cluster_fm.create_public_private_key_pair(subject)
        elif reuse_keys:
            _log.debug('Key pair in use. Reusing the existing key pair.')
        else:
            raise e


def create_key_server(cluster_fm, eks_dm, csr, start_date=None, end_date=None, key_server_tag=None,
                      force=False):
    """Sign the csr and create the Key Server in Element.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster eks_dm: EKS DM
    :param str csr: Certificate signing request text
    :param datetime.datetime start_date: Datetime object representing the client cert start date
    :param datetime.datetime end_date: Datetime object representing the client cert expiration date
    :param str key_server_tag: Tag that will be appended to the Key Server name
    :param bool force: Force the certificate signing to go through regardless of existing certs
    :rtype: valence.models.KeyServerKmip
    """
    # Create the EKS FM
    primary_eks_dm = eks_dm.primary if is_clustered_eks(eks_dm) else eks_dm
    eks_fm = functional_models.get_eks_from_dm(primary_eks_dm)

    # Sign the CSR
    x509_csr = x509.load_pem_x509_csr(csr.encode(), default_backend())
    client_cert, ca_cert = eks_fm.sign_csr(x509_csr, start_date=start_date, end_date=end_date,
                                           force=force)
    client_cert_text = client_cert.public_bytes(serialization.Encoding.PEM)
    ca_cert_text = ca_cert.public_bytes(serialization.Encoding.PEM)

    # Generate a name for Key Server
    name = primary_eks_dm.vendor.value.upper()
    if key_server_tag is None:
        key_server_tag = random_string.generate_random_string(4)
    name += '_{}'.format(key_server_tag)

    # Create the Key Server
    nodes = eks_dm.nodes if is_clustered_eks(eks_dm) else [eks_dm]
    hostnames = [eks.hostname for eks in nodes]
    return cluster_fm.create_key_server_kmip(name, hostnames, client_cert_text, ca_cert_text,
                                             key_server_port=primary_eks_dm.port)


def is_clustered_eks(eks_dm):
    """Return True if given EKS DM is a ExternalKeyServerCluster.

    :param sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster eks_dm: EKS DM
    :rtype: bool
    """
    return isinstance(eks_dm, sf_models.ExternalKeyServerCluster)


def get_eks_hostnames(eks_dm):
    """Get hostnames for either single node or clustered EKS.

    :param sf_models.ExternalKeyServer eks_dm: EKS data model
    :rtype: list[str]
    """
    eks_hostnames = []
    if is_clustered_eks(eks_dm):
        for node in eks_dm.nodes:
            eks_hostnames.append(node.hostname)
    else:
        eks_hostnames.append(eks_dm.hostname)
    return eks_hostnames


def teardown_external_key_management(cluster_fm, eks_dms=None, key_provider_dm=None, disable_ear=False):
    """Teardown the existing EKM setup on the cluster.

    If a list of EKS DMs is passed in, this method will delete all Key Server objects
    that contain that EKS. If it's the last Key Server in the Provider, the Provider will
    also be deleted.

    If a Key Provider is passed in, this method will delete all Key Servers associated with
    that Provider and then delete the Provider.

    If neither are passed in, this method will delete all Key Servers and Providers.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param list[sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster] eks_dms: List of EKS DMs
    :param valence.models.KeyProviderKmip key_provider_dm: Key Provider DM
    :param bool disable_ear: If True, disable EaR before tearing down EKM.
    :raises: _errors.HelperError, errors.TimeoutError
    """
    # EaR must be disabled before removing Key Servers and Key Providers.
    # If disable_ear=True, then disable EaR before starting the teardown.
    if is_encryption_at_rest_enabled(cluster_fm):
        if disable_ear:
            _log.debug('Disabling EaR before tearing down EKM.')
            disable_encryption_at_rest(cluster_fm)
        else:
            raise _errors.HelperError('Cannot teardown EKM while EaR is still enabled.')

    kp_dms = [key_provider_dm] if key_provider_dm else cluster_fm.list_key_providers_kmip()
    for kp_dm in kp_dms:
        # Wait for key provider to no longer be active.
        timer = utils.timeout_timer(60)
        while next(timer):
            key_server_dms = cluster_fm.list_key_servers_kmip(key_provider_id=kp_dm.id)
            filtered_key_server_dms = _filter_key_servers(eks_dms, key_server_dms)

            # We only care about active providers if we are removing every key server from the provider.
            if len(key_server_dms) != len(filtered_key_server_dms) or \
                    not _is_key_provider_active(filtered_key_server_dms):
                break
            time.sleep(10)
        else:
            raise errors.TimeoutError('Key Provider {} is still active.'.format(kp_dm))

        for key_server_dm in filtered_key_server_dms:
            cluster_fm.remove_key_server_from_provider_kmip(key_server_dm)
            cluster_fm.delete_key_server_kmip(key_server_dm)

        # If all servers were removed from the provider, delete it.
        if len(key_server_dms) == len(filtered_key_server_dms):
            _log.debug('All Key Servers for Key Provider {} deleted, deleting provider as well'.format(
                kp_dm.id))
            cluster_fm.delete_key_provider_kmip(kp_dm)
        else:
            _log.debug('Key Provider {} not deleted because it contains unspecified Key Servers'.format(
                kp_dm.id))


def _is_key_provider_active(key_server_dms):
    """Return True if any of the given key servers have an active provider.

    :param list[sf_api.models.KeyServerKmip] key_server_dms: List of key server DMs
    :rtype: bool
    """
    return any(dm.is_provider_active for dm in key_server_dms)


def _filter_key_servers(eks_dms, key_server_dms):
    """Return a list of Key Server DMs that match the given list of EKS DMs.

    :param list[sf_models.ExternalKeyServer|sf_models.ExternalKeyServerCluster] eks_dms: List of EKS DMs
    :param list[valence.models.KeyServerKmip] key_server_dms: List of Key Server DMs
    :rtype: list[valence.models.KeyServerKmip]
    """
    # If eks_dms is given, walk through all key_server_dms looking for matching hostnames.
    # If eks_dms is not given, return all the key_server_dms.
    if eks_dms:
        eks_hostnames = []
        for eks_dm in eks_dms:
            if is_clustered_eks(eks_dm):
                for node in eks_dm.nodes:
                    eks_hostnames.append(node.hostname)
            else:
                eks_hostnames.append(eks_dm.hostname)

        matching_key_servers = []
        for key_server_dm in key_server_dms:
            if any(name in key_server_dm.kmip_key_server_hostnames for name in eks_hostnames):
                matching_key_servers.append(key_server_dm)
    else:
        matching_key_servers = key_server_dms

    return matching_key_servers


def is_encryption_at_rest_enabled(cluster_fm):
    """Return True if EaR is enabled on the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :rtype: bool
    """
    cluster_info = cluster_fm.get_cluster_info()
    return cluster_info.encryption_at_rest_state == enums.EarState.enabled


def enable_encryption_at_rest(cluster_fm, key_provider_dm=None, blocking=True,
                              timeout=300, force_during_bin_sync=None):
    """Wrapper around cluster fm enable_encryption_at_rest() that blocks until done.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param valence.models.KeyProviderKmip key_provider_dm: Key Provider DM
    :param bool blocking: If True, wait for EaR to be enabled
    :param int timeout: Seconds to wait for EaR to be enabled
    :param bool force_during_bin_sync: Whether to enable Ear even during bin sync
    """
    # EaR enable fails if bin sync is in progress
    if not force_during_bin_sync:
        if _bin_sync.is_bin_syncing(cluster_fm):
            bin_sync_timeout = _bin_sync.get_bin_sync_timeout(cluster_fm)
            _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm, timeout=bin_sync_timeout)

    enable_ear_async_handler = cluster_fm.enable_encryption_at_rest(
        key_provider_dm=key_provider_dm, force_during_bin_sync=force_during_bin_sync)
    enable_ear_async_handler.get_models(blocking=blocking, timeout=timeout)


def disable_encryption_at_rest(cluster_fm, blocking=True, timeout=300, force_during_bin_sync=None):
    """Wrapper around cluster fm disable_encryption_at_rest() that blocks until done.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param bool blocking: If True, wait for EaR to be disabled
    :param int timeout: Seconds to wait for EaR to be disabled
    :param bool force_during_bin_sync: Whether to disable Ear even during bin sync
    """
    # EaR disable fails if bin sync is in progress
    if not force_during_bin_sync:
        if _bin_sync.is_bin_syncing(cluster_fm):
            bin_sync_timeout = _bin_sync.get_bin_sync_timeout(cluster_fm)
            _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm, timeout=bin_sync_timeout)

    disable_ear_async_handler = cluster_fm.disable_encryption_at_rest(
        force_during_bin_sync=force_during_bin_sync)
    disable_ear_async_handler.get_models(blocking=blocking, timeout=timeout)


def corrupt_shamir_secret_keys(node_fms):
    """Corrupt the latest shamir key on each node.

    A copy of the key will be saved to root and can be restored with the
    restore_shamir_secret_keys() method.

    :param list[valence.functional_models.Node] node_fms: Node functional models
    :raises: valence.vl_helpers._errors.HelperError
    """
    segments_dir = '/sf/data/segments/'
    for node_fm in node_fms:
        with node_fm.get_ssh_connection() as ssh:
            # Get the latest gen file in segments dir
            cmd = 'ls -tr {} | grep "gen-" | tail -1'.format(segments_dir)
            latest_file = ssh.run_command(cmd, raise_on_failure=True).stdout.strip()

            if latest_file:
                gen_file = segments_dir + latest_file
                saved_gen_file = '/{}.original'.format(latest_file)

                # Check if already copied. rc=0 file exists, rc=2 file does not exist
                cmd = 'ls {}'.format(saved_gen_file)
                file_not_exist = ssh.run_command(cmd).return_code

                if file_not_exist:
                    # Copy the original file
                    cmd = 'cp {} {}'.format(gen_file, saved_gen_file)
                    ssh.run_command(cmd, raise_on_failure=True)

                    # Corrupt the gen file
                    cmd = 'dd if=/dev/zero of={} bs=2 count=2 seek=2 conv=notrunc'.format(gen_file)
                    ssh.run_command(cmd, raise_on_failure=True)
                else:
                    msg = 'Shamir key copy already exists at {}'.format(saved_gen_file)
                    raise _errors.HelperError(msg)
            else:
                msg = 'No shamir key found in {}'.format(segments_dir)
                raise _errors.HelperError(msg)


def move_shamir_secret_keys(node_fms):
    """Move the latest shamir key on each node to root dir.

    The key can be restored with the restore_shamir_secret_keys() method.

    :param list[valence.functional_models.Node] node_fms: Node functional models
    :raises: valence.vl_helpers._errors.HelperError
    """
    segments_dir = '/sf/data/segments/'
    for node_fm in node_fms:
        with node_fm.get_ssh_connection() as ssh:
            # Get the latest gen file in segments dir
            cmd = 'ls -tr {} | grep "gen-" | tail -1'.format(segments_dir)
            latest_file = ssh.run_command(cmd, raise_on_failure=True).stdout.strip()

            if latest_file:
                gen_file = segments_dir + latest_file
                saved_gen_file = '/{}.original'.format(latest_file)

                # Check if already copied. rc=0 file exists, rc=2 file does not exist
                cmd = 'ls {}'.format(saved_gen_file)
                file_not_exist = ssh.run_command(cmd).return_code

                if file_not_exist:
                    # Move the original file
                    cmd = 'mv {} {}'.format(gen_file, saved_gen_file)
                    ssh.run_command(cmd, raise_on_failure=True)
                else:
                    msg = 'Shamir key copy already exists at {}'.format(saved_gen_file)
                    raise _errors.HelperError(msg)
            else:
                msg = 'No shamir key found in {}'.format(segments_dir)
                raise _errors.HelperError(msg)


def restore_shamir_secret_keys(node_fms):
    """Moves the original shamir key back into the segments dir.

    Use this after running corrupt_shamir_secret_keys() or move_shamir_secret_keys()
    to restore the shamir key to a working state.

    :param list[valence.functional_models.Node] node_fms: Node functional models
    :raises: valence.vl_helpers._errors.HelperError
    """
    segments_dir = '/sf/data/segments/'
    for node_fm in node_fms:
        with node_fm.get_ssh_connection() as ssh:
            # Get the latest gen file saved in /
            cmd = 'ls -tr / | egrep "gen-.*\\.original" | tail -1'
            latest_file = ssh.run_command(cmd, raise_on_failure=True).stdout.strip()

            if latest_file:
                # Replace the corrupted file with the original
                gen_file = segments_dir + latest_file.replace('.original', '')
                cmd = 'mv /{} {}'.format(latest_file, gen_file)
                ssh.run_command(cmd, raise_on_failure=True)
            else:
                _log.debug('No shamir key copy found in /')


def deploy_cryptsoft_eks(eks_fm, ca_start_date=None, ca_end_date=None, server_start_date=None,
                         server_end_date=None):
    """Deploy a Cryptsoft EKS using the given functional model.

    If a server is already deployed on the client, tear it down and deploy a new server.

    :param valence.functional_models.CryptsoftExternalKeyServer eks_fm: Cryptsoft EKS FM
    :param datetime.datetime ca_start_date: Datetime object for the ca certificate start date
    :param datetime.datetime ca_end_date: Datetime object for the ca certificate expiration date
    :param datetime.datetime server_start_date: Datetime object for the ca certificate start date
    :param datetime.datetime server_end_date: Datetime object for the ca certificate expiration date
    :return:
    """
    _log.info('Setting up Cryptsoft server {}'.format(eks_fm.hostname))
    if eks_fm.is_running():
        _log.debug('Cryptsoft server already running, powering it off.')
        eks_fm.power_off()

    try:
        _log.debug('Tearing down any existing Cryptsoft server')
        eks_fm.teardown_cryptsoft_server()
    except sf_ssh.errors.SshError:
        _log.debug('Cryptsoft server already torn down.')

    _log.debug('Deploying and powering on a fresh Cryptsoft server')
    eks_fm.deploy_cryptsoft_server(ca_start_date=ca_start_date, ca_end_date=ca_end_date,
                                   server_start_date=server_start_date, server_end_date=server_end_date)
    eks_fm.power_on()


def corrupt_cryptsoft_key(eks_fm, key_id):
    """Corrupt the given key ID within the Cryptsoft database.

    :param valence.functional_models._external_key_server.ExternalKeyServerBase eks_fm: Cryptsoft EKS FM
    :param str key_id: ID of the key on the EKS
    :raises: valence.vl_helpers._errors.HelperError
    :rtype: str
    """
    if eks_fm.vendor != enums.ExternalKeyServerVendor.cryptsoft:
        msg = 'vl_helpers.corrupt_cryptsoft_key() can only be run against a Cryptsoft EKS'
        raise _errors.HelperError(msg)

    # Make sure the random string doesn't match the characters at the start of the key
    key = eks_fm.get_key(key_id)
    random_str = ''
    while key.startswith(random_str):
        random_str = utils.random_string.generate_random_string(4).lower()

    corrupted_key = key.replace(key[0:len(random_str)], random_str)
    eks_fm.set_key(key_id, corrupted_key, key=key)
    return key


def verify_shamir_secret_keys(cluster_fm, node_dms=None):
    """Return True is expected shamir key files exists, else False.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param list[valence.models.SfNodeBase] node_dms: List of active node models
    :rtype: bool
    """
    status = True
    ear_info = cluster_fm.get_encryption_at_rest_info()

    if ear_info.generation_number > 0:
        for node_fm in _filter_nodes(cluster_fm, node_dms=node_dms):
            with node_fm.get_ssh_connection() as ssh:
                gen_file = 'gen-{}'.format(ear_info.generation_number)
                cmd = 'ls /sf/data/segments/{}'.format(gen_file)
                output = ssh.run_command(cmd)
                if output.return_code != 0:
                    _log.info('{} missing from /sf/data/segments/ directory'.format(gen_file))
                    status = False
    else:
        _log.debug('EncryptionAtRest has never been enabled on this cluster')
    return status


def wait_for_ear_state(cluster_fm, ear_state, timeout=600):
    """Wait for the Encryption at Rest state to match the given value.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param valence.enums.EarState ear_state: expected EaR state
    :param int timeout: seconds to wait for given state

    :return: True if cluster info matches expected EaR state
    :raises: TimeoutError if cluster does not reach expected state
    :rtype: bool
    """
    _log.info('Waiting {} seconds for EaR state {}'.format(timeout, ear_state.value))
    timer = utils.timeout_timer(timeout)
    while next(timer):
        cluster_info = cluster_fm.get_cluster_info()
        if cluster_info.encryption_at_rest_state == ear_state:
            return True
        time.sleep(10)

    raise errors.TimeoutError('Timed out waiting for EaR state {}'.format(ear_state.value))


def wait_for_drive_security_state(cluster_fm, security_state, node_dms=None, drive_dms=None,
                                  api_method=enums.NodeMethod.get_hardware_info, timeout=600):
    """Wait for all of the given nodes/drives to reach the given security state.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param bool security_state: expected security state
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
    :param list[valence.models.Drive] drive_dms: list of drive models
    :param valence.enums.NodeMethod|valence.enums.ClusterMethod api_method: API method to query
    :param int timeout: seconds to wait for given state

    :return: True if all given nodes/drives match expected security state
    :raises: TimeoutError if objects do not reach expected state
    :rtype: bool
    """
    _log.info('Waiting {} seconds for Security Enabled state {}'.format(timeout, security_state))
    timer = utils.timeout_timer(timeout)
    while next(timer):
        drive_security_states = _get_security_state(cluster_fm, node_dms, drive_dms, api_method)
        if all(state == security_state for state in drive_security_states):
            return True
        time.sleep(10)

    raise errors.TimeoutError('Timed out waiting for Security Enabled state {}'.format(security_state))


def wait_for_drive_locked_state(cluster_fm, locked_state, node_dms=None, drive_dms=None,
                                api_method=enums.NodeMethod.get_hardware_info, timeout=600):
    """Wait for all of the given nodes/drives to reach the given locked state.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param bool locked_state: expected locked state
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
    :param list[valence.models.Drive] drive_dms: list of drive models
    :param valence.enums.NodeMethod|valence.enums.ClusterMethod api_method: API method to query
    :param int timeout: seconds to wait for given state

    :return: True if all given nodes/drives match expected locked state
    :raises: TimeoutError if objects do not reach expected state
    :rtype: bool
    """
    _log.info('Waiting {} seconds for Security Locked state {}'.format(timeout, locked_state))
    timer = utils.timeout_timer(timeout)
    while next(timer):
        drive_locked_states = _get_locked_state(cluster_fm, node_dms, drive_dms, api_method)
        if all(state == locked_state for state in drive_locked_states):
            return True
        time.sleep(10)

    raise errors.TimeoutError('Timed out waiting for Security Locked state {}'.format(locked_state))


def _get_security_state(cluster_fm, node_dms=None, drive_dms=None,
                        api_method=enums.NodeMethod.get_hardware_info):
    """Checks the appropriate security enabled field in all drives depending on provided API method.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
    :param list[valence.models.Drive] drive_dms: list of drive models
    :param valence.enums.NodeMethod|valence.enums.ClusterMethod api_method: API method to query

    :return: list of values from the 'security_enabled' field in drive_hardware
    :rtype: list[bool]
    """
    if api_method == enums.ClusterMethod.get_cluster_hardware_info:
        return _get_cluster_hardware_state(cluster_fm, 'security_feature_enabled', node_dms, drive_dms)
    elif api_method == enums.NodeMethod.get_hardware_info:
        return _get_node_hardware_state(cluster_fm, 'security_enabled', node_dms, drive_dms)
    else:
        raise _errors.HelperError('Unsupported API method {}'.format(api_method))


def _get_locked_state(cluster_fm, node_dms=None, drive_dms=None,
                      api_method=enums.NodeMethod.get_hardware_info):
    """Checks the appropriate security locked field in all drives depending on provided API method.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
    :param list[valence.models.Drive] drive_dms: list of drive models
    :param valence.enums.NodeMethod|valence.enums.ClusterMethod api_method: API method to query

    :return: list of values from the 'security_locked' field in drive_hardware
    :rtype: list[bool]
    """
    if api_method == enums.ClusterMethod.get_cluster_hardware_info:
        return _get_cluster_hardware_state(cluster_fm, 'drive_security_locked', node_dms, drive_dms)
    elif api_method == enums.NodeMethod.get_hardware_info:
        return _get_node_hardware_state(cluster_fm, 'security_locked', node_dms, drive_dms)
    else:
        raise _errors.HelperError('Unsupported API method {}'.format(api_method))


def _get_cluster_hardware_state(cluster_fm, field, node_dms=None, drive_dms=None):
    """Calls GetClusterHardwareInfo API and returns a list of the field filtered by the provided dms.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param str field: field in DriveHardwareInfo that we're looking for
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
    :param list[valence.models.Drive] drive_dms: list of drive models

    :return: list of values from the given field in DriveHardwareInfo
    :rtype: list[Any]
    """
    if node_dms and not drive_dms:
        node_ids = [n.node_id for n in node_dms]
        drive_dms = cluster_fm.list_drives(node_ids=node_ids, statuses=[enums.DriveStatus.active])
    elif drive_dms is None:
        drive_dms = cluster_fm.list_drives(statuses=[enums.DriveStatus.active])

    if field == 'security_feature_enabled' or field == 'drive_security_locked':
        # Filter out drives in slot -1 (typically in SF3010s)
        # slot (-1) is considered an Internal slot so SoldFire does not secure enable or lock them
        invalid_drive_dms = [d for d in drive_dms if d.slot == -1]
        if invalid_drive_dms:
            _log.debug('Filtering out drives in slot (-1): {}'.format(invalid_drive_dms))
            drive_dms = [d for d in drive_dms if d.slot != -1]

    drives_dict = cluster_fm.get_cluster_hardware_info().drives
    cluster_drive_hw_info = list(six.itervalues(drives_dict))
    drive_hw_dms = _filter_drive_hardware_info(cluster_drive_hw_info, drive_dms=drive_dms)

    results = []
    for drive_hw_dm in drive_hw_dms:
        debug_data = '<Drive {} not found>'.format(drive_hw_dm.serial)

        # Find the drive ID and slot by comparing the serial in DriveHardwareInfo and Drive DMs
        for drive_dm in drive_dms:
            if drive_hw_dm.serial in drive_dm.serial:
                debug_data = 'Drive ID {}, Drive Slot {}, {}={}'.format(
                    drive_dm.drive_id,
                    drive_dm.slot,
                    field,
                    drive_hw_dm[field]
                )
                break

        _log.info(debug_data)
        results.append(drive_hw_dm[field])

    return results


def _get_node_hardware_state(cluster_fm, field, node_dms=None, drive_dms=None):
    """Calls GetHardwareInfo API and returns a list of the given fields filtered by the provided dms.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param str field: field in DriveInfo that we're looking for
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models
    :param list[valence.models.Drive] drive_dms: list of drive models

    :return: list of values from the given field in DriveInfo
    :rtype: list[Any]
    """
    node_fms = _filter_nodes(cluster_fm, node_dms=node_dms)
    drive_hw_dms = []
    for node_fm in node_fms:
        drive_hw_dms += node_fm.get_hardware_info().drive_hardware
    drive_hw_dms = _filter_drive_info(drive_hw_dms, drive_dms=drive_dms)

    if drive_dms is None:
        drive_dms = cluster_fm.list_drives(statuses=[enums.DriveStatus.active])

    if field == 'security_enabled' or field == 'security_locked':
        # Filter out drives in slot -1 (typically in SF3010s)
        # slot (-1) is considered an Internal slot so SoldFire does not secure enable or lock them
        invalid_drive_dms = [d for d in drive_dms if d.slot == -1]
        if invalid_drive_dms:
            _log.debug('Filtering out drives in slot (-1): {}'.format(invalid_drive_dms))
            drive_dms = [d for d in drive_dms if d.slot != -1]

    results = []
    for drive_hw_dm in drive_hw_dms:
        debug_data = '<Drive {} not found>'.format(drive_hw_dm.serial)

        # Find the drive ID by comparing the serial in DriveInfo and Drive DMs
        for drive_dm in drive_dms:
            if drive_hw_dm.serial in drive_dm.serial:
                debug_data = 'Drive ID {}, Drive Slot {}, {}={}'.format(
                    drive_dm.drive_id,
                    drive_dm.slot,
                    field,
                    drive_hw_dm[field]
                )
                break

        _log.info(debug_data)
        results.append(drive_hw_dm[field])

    return results


def _filter_nodes(cluster_fm, node_dms=None):
    """Filter node_fms based on the given node models.

    :param valence.functional_models.Cluster cluster_fm: Cluster FM
    :param list[valence.models.SfNodeBase] node_dms: list of active or pending node models

    :return: filtered list of node functional models
    :rtype: list[valence.functional_models.Node]
    """
    node_fms = cluster_fm.get_node_fms()
    if node_dms is not None:
        node_names = []
        for node_dm in node_dms:
            if 'node_name' in node_dm:
                node_names.append(node_dm.node_name)
            elif 'name' in node_dm:
                node_names.append(node_dm.name)
        node_fms = [n for n in node_fms if n.node_dm.node_name in node_names]
    return node_fms


def _filter_drive_hardware_info(drive_hw_info_dms, drive_dms=None):
    """Filter the drive hw models based on the given drive models.

    :param list[valence.models.DriveHardwareInfo] drive_hw_info_dms: drive hw for all nodes
    :param list[valence.models.Drive] drive_dms: drive models that we're searching for

    :returns: filtered list of drive hw models
    :rtype: list[valence.models.DriveInfo]
    """
    result = []
    if drive_dms is not None:
        drive_serials = [d.serial for d in drive_dms]
        # DriveHardwareInfo.serial in Drive.serial
        for serial in drive_serials:
            for drive_hw in drive_hw_info_dms:
                if drive_hw.serial in serial:
                    result.append(drive_hw)
    return result


def _filter_drive_info(drive_info_dms, drive_dms=None):
    """Filter the drive info models based on the given drive models.

    :param list[valence.models.DriveInfo] drive_info_dms: drive hw for all nodes
    :param list[valence.models.Drive] drive_dms: drive models that we're searching for

    :returns: filtered list of drive info models
    :rtype: list[valence.models.DriveInfo]
    """
    if drive_dms is not None:
        drive_serials = [d.serial for d in drive_dms]
        # DriveInfo.name == Drive.serial
        drive_info_dms = [d for d in drive_info_dms if d.name in drive_serials]
    return drive_info_dms
