"""Upgrades sfupdate healthtools helper object."""
from __future__ import absolute_import

import autologger

from valence import models
from valence.functional_models import _constants
from valence.vl_helpers import _errors

_log = autologger.get_sflogging(__name__)


# pylint: disable=too-complex
def match_sfupdate_healthtools_output_to_regex(
        output_to_parse, token_str, expected_versions_list, download_instructions_str=None,
        management_platform_string=None):
    """Extracts versions out of output and matches against expected version.

    :param str output_to_parse: string to parse for output.
    :param str token_str: Token string to search for that indicates versions to follow.
    :param list[valence.models.GenericVersion] expected_versions_list: list of versions
        to follow token_str.
    :param str|None download_instructions_str: instructions to download new version, if any.
    :param str|None management_platform_string: string to validate for mgmt platform, if exists.
    """
    output_list = (line.strip() for line in output_to_parse.split('\n'))
    output_list = [line for line in output_list if line]
    try:
        index = output_list.index(token_str)
        if expected_versions_list:
            versions = output_list[index + 1:index + 1 + len(expected_versions_list)]

            actual_versions = [models.GenericVersion(n.strip()) for n in versions]
            if set(actual_versions) != set(expected_versions_list):
                raise _errors.HelperError('Versions do not match. Expected: {}, Actual:{}'
                                          .format(expected_versions_list, actual_versions))
            len_of_extra_text_at_end = 0
            if download_instructions_str:
                line_after_versions = output_list[index + 1 + len(expected_versions_list)]
                if download_instructions_str not in line_after_versions:
                    raise _errors.HelperError('Download instruction string and '
                                              'line after versions do not match.')

            sodium_line = 1 if management_platform_string else 0
            expected_output = index + 1 + len(actual_versions) + len_of_extra_text_at_end + sodium_line
            if expected_output != len(output_list):
                raise _errors.HelperError('Unexpected output after token and versions!')

            if management_platform_string:
                if output_list[index+2] != management_platform_string:
                    raise _errors.HelperError('Output list and management output string do not match.')
        else:
            if index + 1 != len(output_list):
                raise _errors.HelperError('Unexpected output after token!')
    except ValueError:
        raise _errors.HelperError('Did not find the string {} in output!'.format(token_str))


def get_expected_platform_string_for_healthtools(mnode_version, healthtools_version):
    """Returns the expected Management Node Platform string.

    :param valence.models.SfVersion mnode_version: mnode version
    :param valence.models.GenericVersion healthtools_version: Version of healthtools.
    :rtype: str|None
    :returns: Expected string or None in case this healthtools doesn't have it at all.
    """
    mgmt_platform_str = None
    if healthtools_version > _constants.HEALTHTOOLS_10_3_POST_RELEASE_VERSION:
        mgmt_platform_str = _constants.MGMT_PLATFORM_LEGACY
        if mnode_version >= _constants.SODIUM:
            if healthtools_version >= _constants.HEALTHTOOLS_19_2_VERSION:
                mgmt_platform_str = _constants.MGMT_PLATFORM_11x.format(mnode_version)
            else:
                mgmt_platform_str = _constants.MGMT_PLATFORM_EMBER

    return mgmt_platform_str
