"""Sfupgradecheck helper object."""

import collections
import re

import autologger

from valence import (
    camelot,
    models,
)
from valence.functional_models import _constants
from valence.vl_helpers import _errors

_log = autologger.get_sflogging(__name__)

output_versions = collections.namedtuple(
    'sfupgradecheck_versions', ['latest_version', 'installed_ht_version', 'node_id_versions'])


def parse_sfupgradecheck_v_output(output_to_parse):
    """Parses the version checks given the stderr from ``sfupgradecheck -v``.

    :param str output_to_parse: the stdout/stderr from ``sfupgradecheck -v``
    :rtype: valence.vl_helpers._sfupgradecheck.output_versions
    :returns: actual latest version, actual installed HT version from output, actual node versions list.
    """
    latest_re = re.compile(r'Latest version of upgrade tools available: ([\d.]+)')
    split_output = output_to_parse.split('\n')

    # Grab latest version string.
    actual_latest_version = [models.GenericVersion(m.group(1)) for m in
                             (latest_re.search(l) for l in split_output) if m]

    ht_re = re.compile(r'Installed version of upgrade tools on management node: ([\d.]+)')
    node_re = re.compile(r'Node (\d+) is on version ([\d.]+)')

    actual_installed_ht_version = [models.GenericVersion(m.group(1)) for m in
                                   (ht_re.search(l) for l in split_output) if m]
    actual_node_id_versions = {int(m.group(1)): models.GenericVersion(m.group(2)) for m in
                               (node_re.search(l) for l in split_output) if m}

    return output_versions(actual_latest_version,
                           actual_installed_ht_version[0],
                           actual_node_id_versions)


def validate_sfupgradecheck_v_output(
        output_to_parse,
        expected_installed_ht_version,
        expected_latest_version=None,
        expected_node_versions=None):
    """Validate the version checks given the stderr from ``sfupgradecheck -v``.

    :param str output_to_parse: the stdout/stderr from ``sfupgradecheck -v``
    :param valence.models.GenericVersion expected_installed_ht_version: the expected version
        of the fdva.
    :param valence.models.GenericVersion|None expected_latest_version: warning line
        for the latest version. If None, expect the latest version warning to not exist.
    :param dict[int,valence.models.GenericVersion]|None expected_node_versions: a map of
        node ids with values corresponding to their expected versions. If None, expect no nodes
        to be in output.
    """
    actual_latest_version, actual_installed_ht_version, actual_node_versions = \
        parse_sfupgradecheck_v_output(output_to_parse=output_to_parse)

    if expected_latest_version:
        if actual_latest_version != [expected_latest_version]:
            raise _errors.HelperError('Actual latest Version does not match the Expected. Actual: {}'
                                      'Expected: {}'.format(actual_latest_version,
                                                            expected_latest_version))

    if actual_installed_ht_version != expected_installed_ht_version:
        raise _errors.HelperError('Actual healthtools version does not match the Expected. Actual: {}'
                                  'Expected: {}'.format(actual_installed_ht_version,
                                                        expected_installed_ht_version))

    if expected_node_versions:
        if actual_node_versions != expected_node_versions:
            raise _errors.HelperError('Actual node version does not match the expected node version. '
                                      'Actual: {} Expected: {}'
                                      .format(actual_node_versions, expected_node_versions))


def validate_sfupgradecheck_download(output_to_check, expected_version, active_nodes, fdva_version):
    """Validates that sfupgradecheck downloaded new software to nodes.

    Validates that the expected nodes and fdva updated successfully based on the
    stderr of the call. Returns the new version.

    :param str output_to_check: the output from our ssh calls to parse.
    :param valence.models.GenericVersion expected_version: version that we should see.
    :param list[valence.models.ActiveNode] active_nodes: list of nodes in this cluster.
    :param valence.models.SfVersion fdva_version: version of the fdva.
    """
    node_ids_from_dm = set([n.node_id for n in active_nodes])
    expected_versions = [expected_version] * len(node_ids_from_dm)
    split_output = output_to_check.split('\n')

    # At some point we've stopped downloading to all nodes and started downloading to only those nodes
    # that need it. Can't tell where that happened...sometime in 2019.12.01. releases.
    check_equality = expected_version <= _constants.HEALTHTOOLS_19_2_VERSION

    if expected_version > _constants.HEALTHTOOLS_18_3_1_OOB_VERSION \
            and fdva_version >= _constants.SODIUM:
        uploading_re = re.compile(r'Uploading version ([\d.]+) to node (\w+)')
        uploading_mat = [uploading_re.search(m) for m in split_output if m]

        node_and_version = [(m.group(2), models.GenericVersion(m.group(1)))
                            for m in uploading_mat if m]
        nodes_from_output = set([n[0] for n in node_and_version])
        ids_from_output = [n.node_id for n in active_nodes if n.name in nodes_from_output]
        nodes_from_output = set(ids_from_output)
    else:
        uploading_re = re.compile(r'Uploading solidfire-sfinstalld-([\d.]+).tgz to nodeID (\d+)')
        uploading_mat = [uploading_re.search(m) for m in split_output if m]

        node_and_version = [(int(m.group(2)), models.GenericVersion(m.group(1)))
                            for m in uploading_mat if m]
        nodes_from_output = set([n[0] for n in node_and_version])

    versions_from_output = [n[1] for n in node_and_version]

    if check_equality:
        if expected_versions != versions_from_output:
            raise _errors.HelperError('The expected versions {} do not match versions from output {}'
                                      .format(expected_version, versions_from_output))
        if nodes_from_output != node_ids_from_dm:
            raise _errors.HelperError('Expected node IDs {} do not match actual {}'
                                      .format(nodes_from_output, node_ids_from_dm))
    else:
        if not all(n for n in nodes_from_output if n in node_ids_from_dm):
            raise _errors.HelperError('Expected node IDs {} do not match actual {}'
                                      .format(nodes_from_output, node_ids_from_dm))


def deploy_same_version_to_nodes(mnode_fm, cluster_dm, version, active_nodes):
    """Install the sfupgradecheck validation module on a set of nodes.

    This only deploys the validation checks to nodes, does not run checks.

    This is done using the sfupgradecheck cmd which uses staging files on fdva.

    :param valence.functional_models.ManagementNodeLegacy mnode_fm: mnode_fm
    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param valence.models.GenericVersion|None version: the ura version to install on all nodes
    :param list[valence.models.ActiveNode] active_nodes: list of nodes in this cluster.
    """
    _log.info('Deploying sfupgradecheck version {} to nodes'.format(version))
    if not version:
        raise _errors.HelperError('Need to know which version to deploy!')

    node_ids = set([n.node_id for n in active_nodes if n])
    versions_list = [version] * len(active_nodes)
    deploy_only = zip(node_ids, versions_list)
    mnode_fm.execute_preupgrade_health_check(cluster_dm,
                                             run_checks_against_asset=True,
                                             deploy_only_versions=deploy_only,
                                             skip_version_check=True)


def deploy_different_versions_to_nodes(mnode_fm, cluster_dm, versions, active_nodes):
    """Install the sfupgradecheck validation module on a set of nodes.

    This only deploys the validation checks to nodes, does not run checks.

    This is done using the sfupgradecheck cmd which uses staging files on fdva.

    :param valence.functional_models.ManagementNodeLegacy mnode_fm: mnode_fm
    :param valence.models.ResourceCluster cluster_dm: Cluster under test.
    :param list[valence.models.GenericVersion]|None versions: the ura version to installs,
        per node.
    :param list[valence.models.ActiveNode] active_nodes: list of nodes in this cluster.
    """
    _log.info('Deploying sfupgradecheck versions to nodes')
    if not versions:
        raise _errors.HelperError('Need to know which version to deploy!')

    node_ids = set([n.node_id for n in active_nodes if n])
    deploy_only = zip(node_ids, versions)
    if len(versions) != len(node_ids):
        raise _errors.HelperError('Lengths of lists must be same!')
    mnode_fm.execute_preupgrade_health_check(cluster_dm,
                                             run_checks_against_asset=True,
                                             deploy_only_versions=deploy_only,
                                             skip_version_check=True)


def prettyprint_sfupgradecheck_json_to_text(processed_checks):
    """Pretty print the processed healthchecks.

    :param valence.functional_models.ProcessedHealthchecks processed_checks: Checks data
    """

    msg = '\n'.join([str(n) for n in processed_checks.cluster_checks])
    for n, c in processed_checks.node_checks.items():
        msg += '\nnodeID {}: {}'.format(n, c)
    _log.info(camelot.banner(msg))
