"""Upgrade Readiness Analyzer checks helper object."""

from valence import (
    enums,
    models,
)
from valence.functional_models import _constants
from valence.vl_helpers import _errors


class CheckCriteria(object):
    """Class that checks which healthcheck is supported by which mnode.

    :param list[tuple[valence.models.GenericVersion]] healthtools_version_map: Healthtools
        version map
    :param valence.models.SfVersion mnode_min_version: Min Mnode version
    """

    def __init__(self, healthtools_version_map, mnode_min_version):
        self.healthtools_version_map = healthtools_version_map
        self.mnode_min_version = mnode_min_version

    @classmethod
    def cluster_checks_by_release(cls):
        """Class method to return all supported cluster checks by release.

        :return: cluster_checks_by_release
        :rtype: dict[str,valence.vl_helpers._sfupgradecheck_validation.CheckCriteria]
        """
        check_pending_nodes = cls([(_constants.HEALTHTOOLS_10_2_VERSION, None)],
                                  _constants.NEON_PATCH2)
        check_cluster_faults = cls([(_constants.HEALTHTOOLS_10_3_VERSION, None)],
                                   _constants.NEON_PATCH3)
        check_async_results = cls([(_constants.HEALTHTOOLS_18_3_1_OOB_VERSION, None)],
                                  _constants.NEON_PATCH2)
        return {enums.HealthcheckName.check_pending_nodes: check_pending_nodes,
                enums.HealthcheckName.check_cluster_faults: check_cluster_faults,
                enums.HealthcheckName.check_async_results: check_async_results}

    @classmethod
    def node_checks_by_release(cls):
        """Class method to return all supported node checks by release.

        :return: node_checks_by_release
        :rtype: dict[str,valence.vl_helpers._sfupgradecheck_validation.CheckCriteria]
        """
        check_mnode_connectivity = cls([(_constants.HEALTHTOOLS_10_2_VERSION, None)],
                                       _constants.NEON_PATCH2)
        check_root_partition_space = cls([(_constants.HEALTHTOOLS_10_2_VERSION, None)],
                                         _constants.NEON_PATCH2)
        check_mnode_connection_speed = cls([(_constants.HEALTHTOOLS_10_3_VERSION, None)],
                                           _constants.NEON_PATCH2)
        check_options_files = cls([(_constants.HEALTHTOOLS_10_3_VERSION, None)],
                                  _constants.NEON_PATCH2)
        check_cores = cls([(_constants.HEALTHTOOLS_10_3_SHIPPING_VERSION,
                            _constants.HEALTHTOOLS_10_3_SHIPPING_VERSION),
                           (_constants.HEALTHTOOLS_10_3_BUGFIX_VERSION, None)], _constants.SODIUM)
        check_paging = cls([(_constants.HEALTHTOOLS_18_3_POST_RELEASE_VERSION,
                             _constants.HEALTHTOOLS_18_3_POST_RELEASE_VERSION),
                            (_constants.HEALTHTOOLS_18_3_BUGFIX_VERSION,
                            _constants.HEALTHTOOLS_18_3_BUGFIX_VERSION),
                            (_constants.HEALTHTOOLS_20_1_VERSION, None)], _constants.SODIUM)
        return {enums.HealthcheckName.check_mnode_connectivity: check_mnode_connectivity,
                enums.HealthcheckName.check_root_partition_space: check_root_partition_space,
                enums.HealthcheckName.check_mnode_connection_speed: check_mnode_connection_speed,
                enums.HealthcheckName.check_options_files: check_options_files,
                enums.HealthcheckName.check_cores: check_cores,
                enums.HealthcheckName.check_paging: check_paging}


def validate_sfupgradecheck_results(mnode_version,
                                    processed_checks, active_nodes, healthtools_version,
                                    validate_check_data_fcns=None,
                                    expected_failed_cluster_checks=None,
                                    expected_failed_node_checks=None):
    """Validate the checks that sfupgradecheck performed on the cluster.

    :param valence.models.SfVersion mnode_version: Version of the mnode.
    :param valence.functional_models.ProcessedHealthchecks processed_checks: sfupgradecheck output
    :param list[valence.models.ActiveNode] active_nodes: list of nodes in cluster
    :param valence.models.GenericVersion healthtools_version: Version of HT to check
        for supported checks against.
    :param dict[str,Any] validate_check_data_fcns: Dict of check names and their validation fcns
        to verify per-node discretionary data dict.
    :param list[str]|None expected_failed_cluster_checks: expected failed cluster checks.
    :param dict[int,list[str]]|None expected_failed_node_checks: expected checks to fail, per node ID.
    """
    validate_sfupgradecheck_cluster_results(
        mnode_version,
        processed_checks,
        healthtools_version,
        expected_failed_cluster_checks=expected_failed_cluster_checks,
        validate_check_data_fcns=validate_check_data_fcns)

    validate_sfupgradecheck_node_check_results(
        mnode_version,
        processed_checks,
        active_nodes,
        healthtools_version,
        validate_check_data_fcns=validate_check_data_fcns,
        expected_failed_checks=expected_failed_node_checks)


def validate_sfupgradecheck_cluster_results(mnode_version, parsed_checks_data, healthtools_version,
                                            expected_failed_cluster_checks=None,
                                            validate_check_data_fcns=None):
    """Simplified fcn to verify URA cluster check passed.

    This will validate that all cluster checks present in this version of healthtools pass.

    :param valence.models.SfVersion mnode_version: Version of the mnode.
    :param valence.functional_models.ProcessedHealthchecks parsed_checks_data: sfupgradecheck output
    :param valence.models.GenericVersion healthtools_version: Version of HT to check
        for supported checks against.
    :param list[str]|None expected_failed_cluster_checks: expected checks to fail
    :param dict[str,Any] validate_check_data_fcns: Dict of check names and their validation fcns
        to verify per-node discretionary data dict.
    :raises: _errors.HelperError
    """
    expected_failed_cluster_checks = expected_failed_cluster_checks or []
    validate_check_data_fcns = validate_check_data_fcns if validate_check_data_fcns else {}

    # Make sure all checks that are supposed to be in this version's output are in there.
    _validate_all_checks_present(
        mnode_version, parsed_checks_data.cluster_checks,
        CheckCriteria.cluster_checks_by_release(), healthtools_version)

    for checks in parsed_checks_data.cluster_checks:
        expected_passed_check = checks.name not in expected_failed_cluster_checks
        if checks.passed != expected_passed_check:
            raise _errors.HelperError('Cluster check {} did not have expected result {}'
                                      .format(checks.name, expected_passed_check))
        if checks.name in validate_check_data_fcns:
            if not validate_check_data_fcns[checks.name](checks):
                raise _errors.HelperError('Cluster check results did not match! '
                                          'Expected {}, actual {}'.format(
                                            validate_check_data_fcns[checks.name],
                                            checks.name))


def _validate_all_checks_present(mnode_version, check_list, expected_checks, healthtools_version):
    """Validates that all checks that are supposed to be in this version are in the output.

    :param valence.models.SfVersion mnode_version: Version of the mnode.
    :param list[valence.models.Healthcheck] check_list: list of cluster checks.
    :param dict[str,valence.vl_helpers._sfupgradecheck_validation.CheckCriteria] expected_checks: list
        of expected checks and versions they are in.
    :param valence.models.GenericVersion healthtools_version: Version of HT to check
        for supported checks against.
    """
    checks_in_this_output = [c.name for c in check_list]
    checks_in_this_version = []
    for k, v in expected_checks.items():
        if mnode_version >= v.mnode_min_version:
            for ev in v.healthtools_version_map:
                earlier_version = ev[0] if ev[0] else models.GenericVersion('2000.0.0.0')
                later_version = ev[1] if ev[1] else models.GenericVersion('3000.0.0.0')
                ht_matches = earlier_version <= healthtools_version <= later_version

                if ht_matches:
                    checks_in_this_version.append(k)
                    break

    if set(checks_in_this_output) != set(checks_in_this_version):
        raise _errors.HelperError('Expecting {} checks, got these checks {}'
                                  .format(checks_in_this_version, checks_in_this_output))


def validate_sfupgradecheck_node_check_results(
        mnode_version, parsed_checks_data, active_nodes, healthtools_version,
        validate_check_data_fcns=None,
        expected_failed_checks=None):
    """Simplified fcn to verify all nodes ran URA checks and all checks passed.

    :param valence.models.SfVersion mnode_version: Version of the mnode.
    :param valence.functional_models.ProcessedHealthchecks parsed_checks_data: sfupgradecheck output
    :param list[valence.models.ActiveNode] active_nodes: list of nodes in cluster
    :param valence.models.GenericVersion healthtools_version: Version of HT to check
        for supported checks against.
    :param dict[str,Any] validate_check_data_fcns: Dict of check names and their validation fcns
        to verify per-node discretionary data dict.
    :param dict[int,list[str]]|None expected_failed_checks: expected checks to fail, per node ID.
    """
    # Assert all nodes actually ran
    node_ids = set(parsed_checks_data.node_checks.keys())
    active_node_ids = set(n.node_id for n in active_nodes)
    assert node_ids == active_node_ids

    expected_failed_checks = expected_failed_checks if expected_failed_checks else {}
    validate_check_data_fcns = validate_check_data_fcns if validate_check_data_fcns else {}

    # Assert all nodes passed their checks
    for node_id, results in parsed_checks_data.node_checks.items():
        # Make sure all checks that are supposed to be in this version's output are in there.
        _validate_all_checks_present(
            mnode_version, results, CheckCriteria.node_checks_by_release(), healthtools_version)

        for check_result in results:
            expect_check_to_pass = not (node_id in expected_failed_checks and
                                        check_result.name in expected_failed_checks[node_id])
            if check_result.passed != expect_check_to_pass:
                raise _errors.HelperError('Node check results did not match! '
                                          'Expected {}, actual {}'.format(
                                           expect_check_to_pass, check_result.passed))

            if check_result.name in validate_check_data_fcns:
                if not validate_check_data_fcns[check_result.name](check_result):
                    raise _errors.HelperError('Node check results did not match! '
                                              'Expected {}, actual {}'.format(
                                               validate_check_data_fcns[check_result.name],
                                               check_result.name))
