"""Helper functions for Cluster Functional Models."""

from __future__ import division

import random
import time

import autologger

from valence import (
    enums,
    errors,
    functional_models,
    models,
)
from valence.vl_helpers import (
    _bin_sync,
    _errors,
)

_log = autologger.get_sflogging()


def _calculate_fill_slice_timeout(volume_dms):
    """Calculate the fill timeout with 20min per TB plus a constant 20min.

    :param list[valence.models.Volume] volume_dms: Volumes being filled.
    :rtype: int
    """
    total_size = sum([vol.total_size for vol in volume_dms])
    total_size_in_tb = float(total_size) / (1000 * 1000 * 1000 * 1000)
    return (total_size_in_tb * 1200.0) + 1200.0


def _update_fill_slice_status(cluster_fm, volume_dms, status):
    """Update the fillSliceStatus attribute on a list of volumes.

    Used by cluster_bs_check_and_verify to skip volumes in the middle of fill slice.

    :param valence.functional_models.Cluster cluster_fm: fm of cluster being filled
    :param list[valence.models.Volume] volume_dms: volumes to fill
    :param valence.enums.AsyncResultStatus status: Status enum to insert in fillSliceStatus attribute.
    """
    attr_dict = {'fillSliceStatus': status.value}
    cluster_fm.modify_volumes(volume_dms, attributes=attr_dict)


def _wait_for_fill_slice_handles(cluster_fm, volume_dms, handle_dict,  # pylint: disable=too-complex
                                 start_time, timeout, fail_on_out_of_bs_space,
                                 completion_check_frequency=60):
    """Polls fillSlice async handles for completion.

    Once per the specified frequency, scrape the event log for fillSlice failed events,
    check for any successful/failed asyncHandles, and raise errors where appropriate.

    :param valence.functional_models.Cluster cluster_fm: fm of cluster being filled
    :param list[valence.models.Volume] volume_dms: volumes to fill
    :param dict handle_dict: Keys are volumeIDs, values are lists of FillSliceAsyncHandler.
    :param int start_time: Time when FillSlice was initiated.
    :param int timeout: Maximum time to wait for FillSlice to complete.
    :param bool fail_on_out_of_bs_space: If we expect to hit stage 5 set to False.
    :param int completion_check_frequency: How often to check the handles.
    """
    timeout = timeout if timeout else _calculate_fill_slice_timeout(volume_dms)
    while handle_dict:
        completed_volume_ids = []
        time.sleep(completion_check_frequency)

        for volume_id, async_fms in handle_dict.items():
            running_handles = list()
            for async_fm in async_fms:
                async_status = async_fm.get_status()
                if async_status is enums.AsyncResultStatus.running:
                    if time.time() > start_time + timeout:
                        raise _errors.HelperError('Fill Volumes Timed out')
                    running_handles.append(async_fm)
                elif async_status is enums.AsyncResultStatus.failed:
                    # Running out of bs space causes problems with all volumes - Mark all failed
                    _log.warn('get async result for {} returned {}'.format(async_fm, async_status))
                    # noinspection PyTypeChecker
                    _update_fill_slice_status(cluster_fm, volume_dms, enums.AsyncResultStatus.failed)

                    # TODO : should we actually check that we are out of bs space??
                    if fail_on_out_of_bs_space:
                        raise _errors.HelperError(
                            'Failures Occurred during fillSlice (out of bs space)')
                    else:
                        return
            handle_dict[volume_id] = running_handles
            if len(handle_dict[volume_id]) == 0:
                completed_volume_ids.append(volume_id)
        for volume_id in completed_volume_ids:
            del handle_dict[volume_id]
    # We're done! Mark all volumes as complete
    _update_fill_slice_status(cluster_fm, volume_dms, enums.AsyncResultStatus.complete)


def _validate_fill_slice_inputs(fill_fraction, start_fraction, compression_ratio,
                                compression_algorithm, dedup_ratio):
    """Sanity check inputs to FillSlice.

    :param float fill_fraction: Fraction of the volumes to fill
    :param float start_fraction: Fractional point at which to start filling the volume.
    :param float compression_ratio: Compression ratio of written data, ie: 1x, 2.5x, 3x, etc.
    :param valence.enums.CompressionAlgorithm compression_algorithm: data compression alg. to use
    :param float dedup_ratio: Dedup ratio per-volume.
    """
    # Make sure we aren't asking for the impossible
    if compression_ratio < 1.0:
        raise ValueError('comp_ratio must be greater than or equal to 1.0')
    if compression_algorithm == enums.CompressionAlgorithm.none and compression_ratio != 1.0:
        raise ValueError('Compression ratio must equal 1.0 if compression algorithm is set to None')
    if fill_fraction + start_fraction > 1:
        raise ValueError('Fill fraction + Start Fraction will go past the size of your volume')
    if fill_fraction > 1 or fill_fraction < 0:
        raise ValueError('Invalid fill fraction: {}'.format(fill_fraction))
    if start_fraction > 1 or start_fraction < 0:
        raise ValueError('Invalid start fraction: {}'.format(start_fraction))
    if dedup_ratio < 1.0:
        raise ValueError('Dedup ratio must be greater than or equal to 1.0')


def _generate_fill_slice_seeds(desired_seed_count, provided_seeds=None):
    """Generate list of seeds for fillSlice data generation.

    If the user provides seeds, use them in a round robin fashion.
    Otherwise generate a unique key for each volume.

    :param int desired_seed_count: The number of seeds to be returned.
    :param list[int] provided_seeds: Seeds provided by the user.
    :rtype: list[int]
    """
    if provided_seeds:
        return [provided_seeds[seed % len(provided_seeds)] for seed in range(desired_seed_count)]
    else:
        return [random.randint(0, (2 ** 31) - 1) for _ in range(desired_seed_count)]


def is_slice_syncing(cluster_fm, volume_dms=None):
    """Calls is_slice_syncing fm.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Volume] volume_dms: Volume to wait for slice sync to finish on
    :rtype: bool
    """
    return cluster_fm.is_slice_syncing(volume_dms)


# Private because you should only be using wait_for_slice_sync_to_start and wait_for_slice_sync_to_finish
def _wait_for_slice_sync(cluster_fm, volume_dms=None, timeout=None, finish=True):
    """Wait until slice sync finishes or starts.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Volume] volume_dms: Volume to wait for slice sync to start/finish
    :param int timeout: Duration to wait for slice sync to start/finish
    :param bool finish: If true, wait for slice sync to finish, else wait for slice sync to start
    :raises valence.vl_helpers.errors.HelperError: If timeout is reached before slice syncing
        starts/finishes
    """
    timeout = 1800 if timeout is None else timeout

    syncing = finish
    start_time = time.time()
    while syncing is finish:
        syncing = is_slice_syncing(cluster_fm, volume_dms=volume_dms)
        if syncing is finish:
            _log.debug('Waiting for slice sync... sleeping 15 seconds')
            time.sleep(15)
        end_time = time.time()
        if end_time - start_time > timeout:
            _log.info(cluster_fm.get_report(enums.JsonReport.slices))
            if finish:
                msg = 'Timed out waiting for slice sync to finish after {} seconds'.format(timeout)
            else:
                msg = 'Timed out waiting for slice sync to start after {} seconds'.format(timeout)
            raise _errors.HelperError(msg)


def calculate_slice_sync_timeout(cluster_fm):
    """Calculate worst case slice sync timeout.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :returns: the estimated worst case slice sync timeout
    :rtype: int
    """
    rep_map = {
        enums.ProtectionScheme.single_helix: 1,
        enums.ProtectionScheme.double_helix: 2,
        enums.ProtectionScheme.triple_helix: 3
    }
    meta_fullness = cluster_fm.get_cluster_full_threshold().sum_used_metadata_cluster_bytes
    rep_count = rep_map[cluster_fm.get_cluster_info().default_protection_scheme]
    # Estimated worst case is ~ 5 MB/s, all slices, at least a minute though
    return max([(meta_fullness * (rep_count - 1 / rep_count)) / (5 * 2 ** 20), 60])


def wait_for_slice_sync_to_finish(cluster_fm, volume_dms=None, timeout=None):
    """Wait until each of the specified volumes have a live secondary and no dead secondaries.

    Omitting volumes results in all volumes being checked. If an invalid volume is passed, it will
    be skipped.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Volume] volume_dms: Volume to wait for slice sync to finish on
    :param int timeout: Duration to wait for stable slices
    """
    timeout = 1800 if timeout is None else timeout
    _wait_for_slice_sync(cluster_fm, volume_dms=volume_dms, timeout=timeout, finish=True)


def wait_for_slice_sync_to_start(cluster_fm, volume_dms=None, timeout=None):
    """Wait until slice sync start starts.

    Omitting volumes results in all volumes being checked. If an invalid volume is passed, it will
    be skipped.

    :param valence.functional_models.Cluster cluster_fm: Resource Cluster functional model
    :param list[valence.models.Volume] volume_dms: Volume to wait for slice sync to start on
    :param int timeout: Duration to wait for slice sync to start
    """
    timeout = 1800 if timeout is None else timeout
    _wait_for_slice_sync(cluster_fm, volume_dms=volume_dms, timeout=timeout, finish=False)


def _slice_service_to_node_map(slice_report):
    """Create a map of slice service ids to node ids.

    :param valence.models.SliceReport slice_report: The slice report from the cluster.
    :rtype: dict
    """
    service_to_node_id = {}
    for service in slice_report.services:
        if service.service_id not in service_to_node_id:
            service_to_node_id[service.service_id] = service.node_id
    return service_to_node_id


def _get_slice_primary_and_secondaries(slice_, service_to_node_id):
    """Combine the node_ids of the primary and all secondaries for the given slice into a list.

    :param valence.models._reports.Slice slice_: The slice to get node_ids from
    :param dict service_to_node_id: The map of slice service id to node id
    :rtype: list
    """
    primary_secondary_list = [service_to_node_id[slice_.primary]]
    for secondaries in slice_.live_secondaries:
        primary_secondary_list.append(service_to_node_id[secondaries])
    return primary_secondary_list


def create_slice_node_map(cluster_fm):
    """Make a map of slice ids to the nodes where the primary and secondaries live.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :rtype: dict[int,list[int]]
    """
    slice_report = cluster_fm.get_slice_sync_report()
    service_to_node_id = _slice_service_to_node_map(slice_report)
    slice_to_node_map = {}
    for slice_ in slice_report.slices:
        if slice_.slice_id not in slice_to_node_map:
            primary_secondary_list = _get_slice_primary_and_secondaries(slice_,
                                                                        service_to_node_id)
            slice_to_node_map[slice_.slice_id] = primary_secondary_list
    return slice_to_node_map


def reboot_slice_services(cluster_fm, timeout=600):
    """Reboot slice services.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param int timeout: Time in seconds after which to raise an error.
    """
    services = cluster_fm.list_services()
    slice_services = [s for s in services if s.service.service_type is enums.ServiceType.slice]
    slice_service_ids = [s.service.service_id for s in slice_services]
    _log.debug('Rebooting slice services: {}'.format(slice_service_ids))
    for node in cluster_fm.list_active_nodes():
        node_fm = functional_models.Node(node.resource_dm, cluster_fm.cluster_dm, authenticate=False)
        node_slice_services = [s for s in slice_services if s.node.node_id == node.node_id]
        for slice_service in node_slice_services:
            node_fm.kill_drive_service_ssh(slice_service, enums.UnixSignal.SIGKILL)

    _log.debug('Waiting for services to recover...')
    services_healthy = False
    time_start = time.time()
    while not services_healthy:
        if time.time() - time_start > timeout:
            raise errors.TimeoutError(
                'Timed out after {}s waiting for services to be healthy'.format(timeout)
            )
        else:
            time.sleep(15)
            unhealthy_service_ids = []
            service_status = cluster_fm.get_service_status()
            if service_status.failed:
                unhealthy_service_ids = [int(sid) for sid in service_status.failed.keys()]
            services_healthy = all([ssid not in unhealthy_service_ids for ssid in slice_service_ids])

    # Bin/slice syncing will occur if the slice services are down for more than 5.5 minutes
    if time.time() - time_start > 5.5 * 60:
        wait_for_slice_sync_to_start(cluster_fm, timeout=60)
        wait_for_slice_sync_to_finish(cluster_fm)
        _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm)


def fill_volumes(cluster_fm, volume_dms, fill_fraction=1.0, seeds=None, start_fraction=0.0,
                 fail_on_out_of_bs_space=True, completion_check_frequency=60,
                 fill_timeout=None, generate_bs_data=True, comp_ratio=1.0,
                 comp_algorithm=enums.CompressionAlgorithm.lzf, dedup_ratio=1.0):
    """Calls FillSlice on a list of volumes with the specified Block count and seed.

    If fill_fraction is not specified it will default to the size of the slice. Seed will default
    to a random int for each volume. If seed is specified we loop over volumes and seeds in a circular
    fashion. For example if you have 10 volumes and 2 seeds, each seed will be used with 5 volumes.
    Similarly if you have 10 volumes and 10 seeds, each seed will be used only once.

    :param valence.functional_models.Cluster cluster_fm: MVIP of a cluster
    :param list[valence.models.Volume] volume_dms: volumes to fill
    :param float fill_fraction: Fraction of the volumes to fill
    :param float start_fraction: Fractional point at which to start filling the volume.
    :param list[int] seeds: Seeds for random number generator. If omitted auto-generate seeds.
    :param bool fail_on_out_of_bs_space: Option to fail on stage 5
    :param int completion_check_frequency: How often to poll the async handle for completion
    :param int fill_timeout: Timeout on filling volumes, default is to give 20 mins per TB + 20 mins
    :param bool generate_bs_data: If false, only generate metadata.
    :param float comp_ratio: Compression ratio of written data, ie: 1x, 2.5x, 3x, etc.
    :param valence.enums.CompressionAlgorithm comp_algorithm: data compression alg. to use
    :param float dedup_ratio: Dedup ratio per-volume. Using this in conjunction with seeds can lead
        to additional total dedup.
    """

    # Sanity check all inputs
    _validate_fill_slice_inputs(fill_fraction, start_fraction, comp_ratio, comp_algorithm,
                                dedup_ratio)
    # Populate the seed array
    seeds = _generate_fill_slice_seeds(len(volume_dms), provided_seeds=seeds)

    # Sort the volumes by size (largest first) so we don't use bsGenerate=false in a bad way
    volume_dms = sorted(volume_dms, key=lambda v: v.total_size, reverse=True)

    # Make sure all slices are frozen for the duration of this step.
    orig_constants_dm = cluster_fm.get_constants()
    fill_constants_dm = models.ClusterConstants.create(
        slice_balanece_enable_auto_timer_iops_rebalancing=0,
        bs_check_interval_seconds=0
    )
    wait_for_slice_sync_to_finish(cluster_fm)
    _bin_sync.wait_for_bin_sync_to_finish_with_timeout(cluster_fm, timeout=60*60)
    slice_report = cluster_fm.get_slice_sync_report()
    cluster_fm.set_constants(fill_constants_dm)

    try:
        handle_dict = {}
        start_time = int(time.time())
        used_seeds = set()
        for volume_dm, seed in zip(volume_dms, seeds):
            # Get the volume slice ID
            try:
                slice_id = [sl.slice_id for sl in slice_report.slices
                            if sl.volume_id == volume_dm.volume_id][0]
            except IndexError:
                raise _errors.HelperError('No slice service found for volume[{}]: {}'
                                          .format(volume_dm.volume_id, slice_report.slices))

            # 4096 is always correct, FillSlice skips the volume layer where 512e would apply.
            blocks_in_volume = volume_dm.total_size // 4096
            start_lba = max(int(start_fraction * blocks_in_volume), 1)
            end_lba = int((start_fraction + fill_fraction) * blocks_in_volume)
            blocks_to_fill = end_lba - start_lba + 1
            handles_for_volume = []

            # Break volume into dedup-able chunks
            chunk_size = max(1, int(blocks_to_fill / dedup_ratio))

            while start_lba < end_lba:
                # Don't generate data we have already generated.
                bs_generate = generate_bs_data if seed not in used_seeds else False
                num_blocks = min(chunk_size, end_lba - start_lba)
                async_handle_fm = cluster_fm.fill_slice(
                    slice_id,
                    num_blocks=num_blocks,
                    seed=seed,
                    bs_generate=bs_generate,
                    start_lba=start_lba,
                    comp_ratio=comp_ratio,
                    comp_algorithm=comp_algorithm,
                )
                handles_for_volume.append(async_handle_fm)
                start_lba += chunk_size
                used_seeds.add(seed)
            handle_dict[volume_dm.volume_id] = handles_for_volume

            # Add fillSliceStatus to attributes (so we can filter these volumes out of bscheck)
            _update_fill_slice_status(cluster_fm, [volume_dm], enums.AsyncResultStatus.running)

        _wait_for_fill_slice_handles(cluster_fm, volume_dms, handle_dict, start_time, fill_timeout,
                                     fail_on_out_of_bs_space, completion_check_frequency)
    finally:
        cluster_fm.set_constants(orig_constants_dm)


def get_slices_assigned_to_services(slice_sync_report, service_ids):
    """Return slices associated with services.

    :param valence.models.SliceReport slice_sync_report: slice sync report
    :param set[int] service_ids: service ids which should not be in the slice report
    :return: list of invalid slice services
    :rtype: list[valence.models.SliceService]
    """
    slices_assigned_to_services = []
    for s in slice_sync_report.slices:
        if (s.primary in service_ids or
                any([i for i in s.live_secondaries if i in service_ids])):
            slices_assigned_to_services.append(s)
    return slices_assigned_to_services
