"""DRuM slice assignments helpers."""
import collections
import time

import autologger

from valence import (
    enums,
    utils,
)
from valence.vl_helpers import _errors

_log = autologger.get_sflogging(__name__)


def get_slice_dm_from_volume(cluster_fm, volume_dm):
    """Factory to return a slice data model from a volume data model and a cluster.

    :param valence.models.Volume volume_dm: Volume data model.
    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :raise valence.vl_helpers._errors.HelperError: If slice is not found
    :rtype: valence.models.Slice
    """

    slice_report_dm = cluster_fm.get_slice_sync_report()
    for slice_dm in slice_report_dm.slices:
        if slice_dm.volume_id == volume_dm.volume_id:
            return slice_dm

    raise _errors.HelperError('Volume id : {} not found in slices.json : {}'.format(
        volume_dm.volume_id, [s.serialize() for s in slice_report_dm.slices]))


def get_service_ids(slice_dm, role=enums.SliceRole.primary):
    """Get service id based on service role.

    :param valence.models._reports.Slice slice_dm:  Slice Model
    :param valence.enums.SliceRole role: slice role
    :rtype: list[int]
    """

    if role == enums.SliceRole.primary:
        return [slice_dm.primary]
    if role == enums.SliceRole.live_secondary:
        return slice_dm.live_secondaries
    if role == enums.SliceRole.dead_secondary:
        return slice_dm.dead_secondaries
    if role == enums.SliceRole.standby:
        return [service.service_id for service in slice_dm.standby_services]
    raise ValueError('Invalid slice role: {}'.format(role))


def check_slice_report_health(report_dm,
                              check_distribution=True,
                              check_degraded=True,
                              check_stable_lives=False,
                              check_duplicates=True,
                              check_standby_state=True,
                              check_standby_duplicates=False):
    """Parse a slice report model for health. Return a description of the errors. stop after error.

    :param valence.models.SliceReport report_dm: Slice Report Model
    :param bool check_distribution: Check distribution of services.
    :param bool check_degraded: Check number of live secondaries.
    :param bool check_stable_lives: Check for migrations in process.
    :param bool check_duplicates: Check for duplicate services.
    :param bool check_standby_state: Check standby service presence and state.
    :param bool check_standby_duplicates: Check that standby services are not on primary or secondary.
    :rtype: bool|str
    :return: None if healthy, else string indicating the first slice with health issues
    """

    for slice_ in report_dm.slices:
        health_issue = _get_slice_health_issues(slice_, check_degraded, check_stable_lives,
                                                check_duplicates, check_standby_state,
                                                check_standby_duplicates)
        if health_issue is not None:
            raise _errors.HelperError(
                'slice health summary for slice id - {}: {}'.format(slice_.slice_id, health_issue))
    if check_distribution:
        distribution_issue = _get_slice_distribution_issues(report_dm)
        if distribution_issue is not None:
            raise _errors.HelperError('slice distribution issue: {}'.format(distribution_issue))


def _get_slice_distribution_issues(report_dm):
    """Count primaries, lives, and secondaries and make sure they are distributed across services.

    :param valence.models.SliceReport report_dm: Slice Report Model
    :rtype: str|None
    :return: None if slices distributed, else string indicating slice distribution issues
    """
    p_counter = collections.Counter()
    l_counter = collections.Counter()
    sb_counter = collections.Counter()

    for slice_dm in report_dm.slices:
        # update individual counters
        primary = slice_dm.primary
        p_counter[primary] += 1
        l_counter += collections.Counter({live: 1 for live in slice_dm.live_secondaries})

        # pylint: disable=no-member
        sb_counter += collections.Counter({sb.service_id: 1 for sb in slice_dm.standby_services})

    service_ids = set(list(p_counter) + list(l_counter) + list(sb_counter))
    # TODO: Revisit this logic as part of ELEM-5259
    for counter, type_str in [(p_counter, 'primaries'), (l_counter, 'live_secondaries'),
                              (sb_counter, 'standbys')]:
        for service_id in service_ids:
            count = counter[service_id]
            if not (sum(counter.values()) / 2.0 - 1 <= count <= sum(counter.values()) * 2.0 + 1):
                return 'service {} has too many or too few {}'.format(service_id, type_str)
    return None


def check_standby_service_state(cluster_fm, volume_dm, standby_state,
                                check_degraded=None,
                                check_stable_lives=None,
                                check_standby_duplicates=None
                                ):
    """Check slice health and standby state.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.Volume volume_dm: Volume data model.
    :param valence.enums.StandbyStates standby_state: Expected standby state.
    :param bool check_degraded: Check if slice is degraded.
    :param bool check_stable_lives: Check if slice is stable.
    :param bool check_standby_duplicates: Check if standby duplicates exist.
    """
    expect_ha = (standby_state == enums.StandbyStates.inactive)
    check_degraded = check_degraded if check_degraded is not None else expect_ha
    check_stable_lives = check_stable_lives if check_stable_lives is not None else expect_ha
    check_standby_duplicates = check_standby_duplicates if check_standby_duplicates is not None else True

    check_slice_health_for_volume(cluster_fm, volume_dm,
                                  standby_state=standby_state,
                                  check_degraded=check_degraded,
                                  check_stable_lives=check_stable_lives,
                                  check_standby_duplicates=check_standby_duplicates)


def _get_slice_health_issues(slice_dm,
                             check_degraded=True,
                             check_stable_lives=False,
                             check_duplicates=True,
                             check_standby_state=True,
                             check_standby_duplicates=False,
                             slice_rep_count=2):
    """Check for slice health.

    :param valence.models.Slice slice_dm: Slice model to check.
    :param bool check_degraded: Check if slice is degraded.
    :param bool check_stable_lives: Check if slice is stable.
    :param bool check_duplicates: Check if duplicates exist.
    :param bool check_standby_state: Check standby state.
    :param bool check_standby_duplicates: Check if standby duplicates exist.
    :param int slice_rep_count: Expected slice replication count
    :rtype: str|None
    :return: string describing health issues or None.
    """

    primary = slice_dm.primary
    lives = slice_dm.live_secondaries
    deads = slice_dm.dead_secondaries
    standbys = slice_dm.standby_services
    health_issues = []
    debug_slid = slice_dm.slice_id
    lives_for_ha = slice_rep_count - 1  # num live secondaries needed for high availability
    highly_available = len(lives) >= lives_for_ha
    # check the number of live secondaries ensures HA
    if check_degraded:
        if not highly_available:
            health_issues.append('insufficient lives : {}'.format(lives))
        # check that there are no migrations in process (extra lives, deads, etc.)
        # this flag should be ignored if check_degraded is not true
        if check_stable_lives and (deads or len(lives) > lives_for_ha):
            health_issues.append('unstable lives: lives={} deads={}'.format(lives, deads))
    # check duplicate secondaries or primary on single service
    all_services = lives + deads + [primary]
    if check_duplicates and _duplicates(all_services):
        health_issues.append('duplicates : p={} ls={} ds={}'.format(primary, lives, deads))
    # check standby services for presence and state
    if check_standby_state:
        health_issues.extend(_get_standby_service_issues(slice_dm, highly_available))
    # check that standby services are not on primary or secondary
    sb_ids = [s.service_id for s in standbys]
    all_services.extend(sb_ids)
    if check_standby_duplicates and _duplicates(all_services):
        health_issues.append('duplicates for slice {}: p={} ls={} ds={} sb={}'
                             .format(debug_slid, primary, lives, deads, sb_ids))

    if health_issues:
        _log.debug('Slice json object at the time of failure: {}'.format(slice_dm.serialize()))
        return 'Health issues for slice {} : '.format(debug_slid) + ' ; '.join(health_issues)
    # fell through finding no errors: no health issues = negative test result.
    return None


def _duplicates(list_):
    """Check list for duplicates and return true or false.

    :param list list_: list to check for duplicates
    :rtype: bool
    """
    return len(list_) != len(set(list_))


def _get_standby_service_issues(slice_dm, highly_available):
    """Check standby services in slice for issues related to presence and state.

    :param valence.models.Slice slice_dm: Slice model to check.
    :param bool highly_available: expected highly available
    :rtype: list[str]
    :return: list of standby_service issues
    """
    issues = []
    if len(slice_dm.standby_services) != 1:
        issues.append('wrong number of standbys={}'.format(len(slice_dm.standby_services)))
    elif highly_available and slice_dm.standby_services[0].state != enums.StandbyStates.inactive:
        issues.append(
            'wrong standby state {}: lives={}, standbys={}'.format(slice_dm.standby_services[0].state,
                                                                   slice_dm.live_secondaries,
                                                                   slice_dm.standby_services))
    return issues


def get_slice_services_for_volume(cluster_fm, volume_dm, slice_service_role):
    """Return the slice service serving the desired role for the slice associated with a volume.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.Volume volume_dm: Volume data model.
    :param valence.enums.SliceRole slice_service_role: service role in slice to select on
    :raise HelperError: service of given role could not be found
    :rtype: list[valence.models.ClusterServiceInfo]
    """
    if slice_service_role == enums.SliceRole.primary:
        wait_for_desired_primary_to_resolve(cluster_fm, volume_dm)

    services = cluster_fm.list_services()

    # find the slice object associated with the volume
    slice_dm = get_slice_dm_from_volume(cluster_fm, volume_dm)

    # find the service object associated with the slice
    # first find the service ID
    service_ids = get_service_ids(slice_dm, slice_service_role)

    service_infos = [service_info for service_info in services
                     if service_info.service.service_id in service_ids]

    if not service_infos:
        raise _errors.HelperError('Unable to find service id for service role :{} for volume : '
                                  '{}'.format(slice_service_role, volume_dm.volume_id))
    return service_infos


def wait_for_desired_primary_to_resolve(cluster_fm, volume_dm, timeout=300):
    """This function waits for the desired primary to take over as current primary.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.Volume volume_dm: Volume data model.
    :param int timeout: timeout to stop checking for desired primary.
    """
    timer = utils.timeout_timer(timeout)
    while next(timer):
        _log.debug('waiting for desired primary to take over as primary')
        slice_dm = get_slice_dm_from_volume(cluster_fm, volume_dm)
        if slice_dm.desired_primary == 0:
            return
        if slice_dm.desired_primary and slice_dm.desired_primary == slice_dm.primary:
            return
        time.sleep(5)

    raise _errors.HelperError('Timed out waiting for desired primary to take over')


def wait_for_slice_health(cluster_fm, volume_dm, timeout=300, poll_interval=10,
                          check_degraded=True,
                          check_stable_lives=False,
                          check_duplicates=True,
                          check_standby_state=True,
                          check_standby_duplicates=False):
    """Wait for slice health.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param sf_api.models.Volume volume_dm: Volume data model.
    :param bool check_degraded: Check if slice is degraded.
    :param bool check_stable_lives: Check if slice is stable.
    :param bool check_duplicates: Check if duplicates exist.
    :param bool check_standby_state: Check standby state.
    :param bool check_standby_duplicates: Check if standby duplicates exist.

    :param int timeout: Time in seconds to wait for slice health.
    :param int poll_interval: Time in seconds to poll for slice health.
    """
    timer = utils.timeout_timer(timeout)

    while next(timer):
        slice_dm = get_slice_dm_from_volume(cluster_fm, volume_dm)
        health_issue = _get_slice_health_issues(slice_dm,
                                                check_degraded=check_degraded,
                                                check_stable_lives=check_stable_lives,
                                                check_duplicates=check_duplicates,
                                                check_standby_state=check_standby_state,
                                                check_standby_duplicates=check_standby_duplicates)
        if health_issue is not None:
            _log.debug('Still waiting slice health: {}'.format(health_issue))
            time.sleep(poll_interval)
        else:
            _log.debug('Slice services seem to be healthy, moving on')
            return

    raise _errors.HelperError('Slice is still unhealthy after {}s'.format(timeout))


def check_slice_health_for_volume(cluster_fm, volume_dm, standby_state,
                                  check_degraded=None,
                                  check_stable_lives=None,
                                  check_standby_duplicates=None,
                                  ):
    """Check slice health and standby state.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.Volume volume_dm: Volume data model.
    :param valence.enums.StandbyStates standby_state: Expected standby state
    :param bool check_degraded: Check if slice is degraded.
    :param bool check_stable_lives: Check if slice is stable.
    :param bool check_standby_duplicates: Check if standby duplicates exist.
    """
    expect_ha = (standby_state == enums.StandbyStates.inactive)

    check_degraded = check_degraded if check_degraded is not None else expect_ha
    check_stable_lives = check_stable_lives if check_stable_lives is not None else expect_ha
    check_standby_duplicates = check_standby_duplicates if check_standby_duplicates is not None else True

    slice_dm = get_slice_dm_from_volume(cluster_fm, volume_dm)

    health_issue = _get_slice_health_issues(slice_dm,
                                            check_degraded=check_degraded,
                                            check_stable_lives=check_stable_lives,
                                            check_duplicates=True,
                                            check_standby_state=True,
                                            check_standby_duplicates=check_standby_duplicates)
    if health_issue is not None:
        raise _errors.HelperError(health_issue)
    if standby_state:
        if slice_dm.standby_services[0].state != standby_state:
            raise _errors.HelperError('expected state {} for slice {}'.format(standby_state,
                                                                              slice_dm))


def check_standby_state_for_volume(cluster_fm, volume_dm, standby_state):
    """Check if standby service is active/inactive/failed for a given volume.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model for the cluster.
    :param sf_api.models.Volume volume_dm: Volume data model for volume to be checked
    :param valence.enums.StandbyStates standby_state: Expected standby state
    :rtype: bool
    :return: True if standby state is the expected state, else False.
    """
    slice_dm = get_slice_dm_from_volume(cluster_fm, volume_dm)
    try:
        if slice_dm.standby_services[0].state != standby_state:
            _log.debug('Standby state not in expected state: {}, state found: {}'.format(
                standby_state, slice_dm.standby_services[0].state))
            return False
        else:
            return True
    except IndexError as err:
        raise _errors.HelperError('Standby service not enabled on this cluster/volume - '
                                  '{}'.format(err))


def wait_for_standby_state(cluster_fm, volume_dm, standby_state, timeout=None):
    """Poll for standby state specified.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model for the cluster.
    :param valence.models.Volume volume_dm: Volume data model for volume to be checked
    :param valence.enums.StandbyStates standby_state: Expected standby state
    :param int timeout: Time in seconds to wait for standby state.
    :raises _errors.HelperError: on Timeout
    """

    timeout = timeout if timeout is not None else 300
    timer = utils.timeout_timer(timeout)

    while next(timer):
        found_state = check_standby_state_for_volume(cluster_fm, volume_dm, standby_state)
        if found_state:
            return

    raise _errors.HelperError('Timed out waiting for standby state {}.'.format(standby_state))


def get_standby_service_id_for_volume(cluster_fm, volume_dm):
    """Get standby service id for a given volume.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param valence.models.Volume volume_dm: Volume data model for volume to be checked
    :rtype: bool
    :raises IndexError: if standby service is not enabled
    """
    slice_dm = get_slice_dm_from_volume(cluster_fm, volume_dm)
    try:
        standby_service_id = slice_dm.standby_services[0].service_id
    except IndexError:
        _log.error('Standby service not enabled on this cluster/volume')
        raise
    return standby_service_id


def check_volume_standbys_assigned(cluster_fm, node_dms=None):
    """Check if volumes have assigned standbys.

    If nodes are not provided, check for all active nodes in the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.ActiveNode] node_dms: node data models to check for volume standbys.
    :return: List of unassigned volume ids
    :rtype: list[int]
    """
    if not node_dms:
        node_dms = cluster_fm.list_active_nodes()

    unassigned_volume_ids = []  # Volumes that have not been assigned a standby

    for node in node_dms:
        result = cluster_fm.check_volume_standbys_assigned(node.node_id)
        if result.volumes_ids_with_unassigned_standbys:
            unassigned_volume_ids.extend(result.volumes_ids_with_unassigned_standbys)

    if unassigned_volume_ids:
        _log.debug('The following volumes were not assigned a standby: {}'.format(unassigned_volume_ids))

    return unassigned_volume_ids


def check_volume_standbys_unassigned(cluster_fm, node_dms=None):
    """Check if all volumes have unassigned standbys.

    If nodes are not provided, check for all active nodes in the cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model.
    :param list[valence.models.ActiveNode] node_dms: Node data models to check for volume standbys.
    :return: List of assigned volume ids
    :rtype: list[int]
    """
    if not node_dms:
        node_dms = cluster_fm.list_active_nodes()
    # Get the cluster volume_ids, place into a set to do a difference
    volume_ids = set([volume.volume_id for volume in cluster_fm.list_volumes()])
    unassigned_volume_ids = []

    for node in node_dms:
        result = cluster_fm.check_volume_standbys_assigned(node.node_id)
        if result.volumes_ids_with_unassigned_standbys:
            unassigned_volume_ids.extend(result.volumes_ids_with_unassigned_standbys)

    assigned_volume_ids = volume_ids - set(unassigned_volume_ids)
    if assigned_volume_ids:
        _log.debug('The following volumes still have a standby assigned: {}'.format(assigned_volume_ids))

    return assigned_volume_ids


def wait_for_volume_standbys_assigned(cluster_fm, timeout=None):
    """Wait for the volume standbys to be assigned.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: optional timeout to wait for unassignment

    :raises: _errors.HelperError: If volumes do not have standbys within a specified time period.
    """

    timeout = timeout if timeout is not None else 300
    timer = utils.timeout_timer(timeout)
    while next(timer):
        sb_unassigned_volumes = check_volume_standbys_assigned(cluster_fm)
        if not sb_unassigned_volumes:
            return

    raise _errors.HelperError('Timed out waiting for standby to be assigned to all volumes')


def wait_for_volume_standbys_unassigned(cluster_fm, timeout=None):
    """Wait for the volume standbys to be unassigned.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int timeout: optional timeout to wait for unassignment

    :raises: _errors.HelperError: If volumes standbys have not been unassigned within the specified time.
    """

    timeout = timeout if timeout is not None else 300
    timer = utils.timeout_timer(timeout)
    while next(timer):
        sb_assigned_volumes = check_volume_standbys_unassigned(cluster_fm)
        if not sb_assigned_volumes:
            return

    raise _errors.HelperError('Timed out waiting for standby to be unassigned from all volumes')
