"""Helper functions for Slice methods."""
import autologger
import sf_batch
import six

from valence import (
    enums,
    models,
)
from valence.vl_helpers import (
    _errors,
    _slice,
)

_log = autologger.get_sflogging(__name__)


def _check_slice_lba_chunk(cluster_fm, volume_id, snapshot_id, slice_id, checksum_dict, start_lba,
                           n_blocks, ignored_services, min_rep_count):
    """Thread worker jobs.  Calls ReadSliceLBAChecksum and verifies consistency between slice files.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param int volume_id: id of volume to read slice lbas from
    :param int snapshot_id: optional snapshot id of volume under test
    :param int slice_id: id of slice corresponding to input volume/snapshot
    :param dict checksum_dict: thread shared dictionary for collecting volume chunk slice checksums
    :param int start_lba: lba where to start the check
    :param int n_blocks: number of lbas to check; default is the end of the cluster
    :param list[int] ignored_services: ignore checksum for these services
    :param int min_rep_count: must be at least this many copies (default=cluster rep count,
                          decrement for ignored services)
    :raises HelperError: If ReadSliceLBAChecksum returns different values for primary/secondary
    """

    volume_info = 'volume {}'.format(volume_id)
    if snapshot_id is not None:
        volume_info += ', snapshot {}'.format(snapshot_id)
    worker_info = '{}, LBAs {}-{}'.format(volume_info, start_lba, start_lba + n_blocks - 1)
    _log.debug('Ignored services for {}: {}'.format(volume_info, ignored_services))

    min_slice_files = cluster_fm.get_cluster_info().rep_count if min_rep_count is None else min_rep_count
    min_slice_files = max(1, min_slice_files - len(ignored_services))

    sums = cluster_fm.read_slice_lba_checksum(slice_id, start_lba, n_blocks, snapshot_id)
    _log.debug('Done with worker for {}: {}'.format(worker_info, sums))

    # remove ignored keys from the dictionary
    for service in ignored_services:
        sums.pop(service, None)

    worker_ok = False
    if sums is None:
        _log.warning('ReadSliceLBAChecksum returned None!')
    elif len(sums) < min_slice_files:
        _log.warning('ReadSliceLBAChecksum did not return multiple checksums!')
    elif len(set(sums.values())) > 1:
        _log.warning('ReadSliceLBAChecksum returned inconsistent checksums: {}'.format(sums.values()))

        # TODO: hash this out in VL-213.  n_lba will need to be < 20480 - found in my testing.
        # if not _find_mismatched_lbas(cluster_fm, volume_id, slice_id, start_lba, n_blocks, snapshot_id)
        #    _log.warning("False positive for {} -- did new writes come in?".format(worker_info))
    else:
        checksum_dict[(start_lba, n_blocks)] = list(sums.values())[0]
        worker_ok = True
    if not worker_ok:
        raise _errors.HelperError('ReadSliceLBAChecksum mismatch for {}: {}'
                                  .format(worker_info, sums))


def _check_slice_lbas_at_rest(cluster_fm, volume, chunk_size, start_lba,  # pylint: disable=too-complex
                              n_lbas, ignore_secondaries, min_rep_count):
    """Compare checksum for the volumes. Incoming writes may cause failure.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.models.Volume|valence.models.Snapshot volume: volume to read slice lbas from
    :param int chunk_size: number of lbas to checksum at once.
    :param int start_lba: lba where to start the check
    :param int|None n_lbas: number of lbas to check; default is the end of the cluster
    :param bool ignore_secondaries: ignore secondary slice files if true
    :param int min_rep_count: must be at least this many copies (default=cluster rep count, decrement for
                          ignored services)
    :raises ValueError: If slice service IDs cannot be found or provided LBA range is not valid
    :raises HelperError: Raised if one of the spawned workers raises an exception.
    :returns: 0 if there are mismatches, else positive integer hash of block ids
    :rtype: int
    """

    slice_id = None
    volume_id = volume.volume_id
    snapshot_id = volume.snapshot_id if isinstance(volume, models.Snapshot) else None
    slice_report = cluster_fm.get_report(enums.JsonReport.slices)

    ignored_services = []
    for slice_info in slice_report['slices']:
        if slice_info['volumeID'] == volume.volume_id:
            slice_id = slice_info['sliceID']

            if ignore_secondaries:
                ignored_services.extend(slice_info.get('liveSecondaries', []))
                ignored_services.extend(slice_info.get('deadSecondaries', []))

    if slice_id is None:
        raise ValueError('Could not find slice ID from slices.json report for volume {}'
                         .format(volume_id))

    # Sanity check inputs
    n_lbas = volume.total_size.bytes // 4096 - start_lba if n_lbas is None else n_lbas
    if start_lba >= volume.total_size.bytes // 4096:
        raise ValueError('Start LBA {} provided is larger than max LBA for slice file on volume {}'
                         .format(start_lba, volume_id))
    if start_lba + n_lbas > volume.total_size.bytes // 4096:
        raise ValueError('LBA range provided for slice file is outside the valid range for volume {}'
                         .format(volume_id))
    if n_lbas < 1:
        raise ValueError('n_lbas ({}) for volume {} must be positive'.format(n_lbas, volume_id))

    # Spin up a thread pool to collect checksums for each lba chunk
    checksum_dict = {}
    pool = sf_batch.ThreadPool(stop_on_error=True)
    for lba in range(start_lba, start_lba + n_lbas, chunk_size):
        blocks = min(lba + chunk_size, start_lba + n_lbas) - lba
        job_args = (cluster_fm, volume_id, snapshot_id, slice_id, checksum_dict,
                    lba, blocks, ignored_services, min_rep_count)
        pool.add_job(_check_slice_lba_chunk, args=job_args)
    _log.debug('Done with submitting to the pool')

    pool.run(raise_exception=True)
    _log.debug('Slice files for volume {} are consistent.'.format(volume.volume_id))
    checksums = tuple((k, checksum_dict[k]) for k in sorted(checksum_dict.keys()))
    return max(1, abs(hash(checksums)))


def check_lbas_at_rest(cluster_fm, volume_dms=None, snapshot_dms=None, chunk_size=100000, start_lba=0,
                       n_lbas=None, ignore_secondaries=False, min_rep_count=2):
    """Compares primary/secondary slice file lbas.  Confirms consistency between slice services.

    Returned list is a hash of the block_ids for given input lba range on the volume/snapshots.
    If the slice file data doesn't match an exception will be raised.  Inconsistent slice file
    data is usually caused by a volume not being 'at rest'.  i.e. there are incoming writes
    while slice file data consistency is being checked.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Volume] volume_dms: the volumes to be checked.
    :param list[valence.models.Snapshot] snapshot_dms: the volumes to be checked.
    :param int chunk_size: number of lbas to checksum at once.
    :param int start_lba: the lba where to start the comparison (default 0)
    :param int n_lbas: the number of lbas to check (default is all lbas in the volume)
    :param bool ignore_secondaries: false by default; if true, read only the primary slice file.
    :param int min_rep_count: expected replication level - primaries plus optionally ignored secondaries

    :returns: checksum/hash list of the block_id data corresponding to input lists
    :rtype: list[int]
    """
    if not volume_dms:
        volume_dms = []
    if not snapshot_dms:
        snapshot_dms = []
    if not volume_dms and not snapshot_dms:
        raise ValueError('Invalid input - No volume or snapshot data models provided')

    # Secondaries should be ignored for single-node clusters
    if not ignore_secondaries and len(cluster_fm.list_active_nodes()) == 1:
        ignore_secondaries = True

    checksums = []
    for vol in volume_dms + snapshot_dms:
        _slice.wait_for_slice_sync_to_finish(cluster_fm, [vol])
        volume_checksum = _check_slice_lbas_at_rest(cluster_fm, vol, chunk_size, start_lba, n_lbas,
                                                    ignore_secondaries, min_rep_count)
        checksums.append(volume_checksum)
    return checksums


def create_volumes_single_profile(cluster_fm, volume_dm, number_of_volumes, skip_fullness_check=None):
    """Create number volumes with the profile described in volume_dm.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param valence.models.Volume volume_dm: Profile that all resulting volumes will share.
    :param int number_of_volumes: Number of volumes to create from the profile.
    :param bool skip_fullness_check: Whether to skip the cluster fullness check preventing volumes
        from being created in stage 4.

    :return: List of volumes
    :rtype: list[valence.models.Volume]
    """
    qos = volume_dm.qos
    return cluster_fm.create_multiple_volumes(
        number=number_of_volumes, name_prefix=volume_dm.name, account_id=volume_dm.account_id,
        total_size=volume_dm.total_size,
        enable512e=volume_dm.enable512e,
        min_iops=qos.min_iops if qos else None,
        max_iops=qos.max_iops if qos else None,
        burst_iops=qos.burst_iops if qos else None,
        burst_time=qos.burst_time if qos else None,
        qos_policy_id=volume_dm.qos_policy_id,
        enable_snapmirror_replication=volume_dm.enable_snapmirror_replication,
        attributes=volume_dm.attributes,
        skip_fullness_check=skip_fullness_check,
        access=volume_dm.access
    )


def create_volumes_individual_profiles(cluster_fm, volume_dms):
    """Create a volume for each volume_dm provided.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Volume] volume_dms: Profile that all resulting volumes will share.

    :return: List of volumes
    :rtype: list[valence.models.Volume]
    """
    return_volume_dms = []
    for volume_dm in volume_dms:
        qos = volume_dm.qos
        return_volume_dms.append(cluster_fm.create_volume(
            name=volume_dm.name, account_id=volume_dm.account_id, total_size=volume_dm.total_size,
            enable512e=volume_dm.enable512e,
            min_iops=qos.min_iops if qos else None,
            max_iops=qos.max_iops if qos else None,
            burst_iops=qos.burst_iops if qos else None,
            burst_time=qos.burst_time if qos else None,
            qos_policy_id=volume_dm.qos_policy_id,
            enable_snapmirror_replication=volume_dm.enable_snapmirror_replication,
            attributes=volume_dm.attributes,
            access=volume_dm.access
        ))
    return return_volume_dms


def delete_and_purge_volumes(cluster_fm, volume_dms):
    """Delete and purge provided volumes on a cluster.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param typing.Iterable[valence.models.Volume] volume_dms: Volumes you wish to delete and purge
    """
    deleted_volumes = cluster_fm.delete_volumes(volume_dms)
    purge_deleted_volumes(cluster_fm, deleted_volumes)


def purge_deleted_volumes(cluster_fm, volume_dms=None):
    """Purge provided deleted volumes from a cluster, or all deleted volumes if not specified.

    :param valence.functional_models.Cluster cluster_fm: Cluster functional model
    :param list[valence.models.Volume] volume_dms: Volumes you wish to purge
    """
    if not volume_dms:
        cluster_fm.purge_all_deleted_volumes()
    else:
        cluster_fm.purge_deleted_volumes(volume_dms)


def verify_volume_checksums(cluster_fm, volume_dms):
    """Verify volumes by checksums across primary/live secondaries.

    Throw a dict of bad volumes or return all the checksums.
    This method does not assume any particular protection scheme or number of checksums returned for
    each volume. For example, it does not expect exactly two checksums per volume, as for double helix.
    It only expects that all of them match.

    :param valence.functional_models.Cluster cluster_fm: Cluster fm
    :param list[valence.models.Volume] volume_dms: a list of volumes to verify
    :raise Exception: If the checksums for one or more volumes do not verify
    :return: Dict of all volume checksums if all volume checksums verify
    :rtype: dict
    """
    bad_volumes = {}
    result = {}
    for volume_dm in volume_dms:
        checksums = cluster_fm.compute_volume_checksum(volume_dm=volume_dm)
        result[volume_dm.volume_id] = checksums
        values = set(six.itervalues(checksums))  # set of unique checksum values
        if len(values) != 1:  # if all match, we have exactly one
            bad_volumes[volume_dm.volume_id] = checksums  # otherwise something bad happened

    if len(bad_volumes) > 0:
        raise _errors.HelperError('Volumes {} had errors'.format(bad_volumes))

    return result
