import itertools
import unittest

import mock
import pytest

from sf_api import api_constants
from valence import (
    enums,
    functional_models,
    models,
    utils,
    vl_helpers,
)

# noinspection PyUnusedLocal

# Mock the cluster DM and FM once to avoid extra work
RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
CLUSTER_FM = functional_models.Cluster(RS_CLUSTER_DM, authenticate=False)


class TestBinAssignmentsHelper(unittest.TestCase):

    @mock.patch.object(vl_helpers._events, 'wait_for_event')
    def test_wait_for_bin_assignments_returns_events(self, mock_wait_events):
        events = models.ClusterEvent.get_mock_object_list(2, overrides={
            'event_info_type': enums.ClusterEventType.service_event,
            'message': 'Block data assignment change detected.'
        })
        mock_wait_events.return_value = events

        event_list = vl_helpers.wait_for_bin_assignments(CLUSTER_FM, timeout=543, poll_period=0,
                                                         start_event_id=5)
        assert event_list == events
        mock_wait_events.assert_called_with(
            CLUSTER_FM,
            enums.ClusterEventType.service_event,
            message_regexes=[r'^Block data assignment change detected.\s*$'],
            poll_period=0,
            start_event_id=5,
            timeout=543
        )


class TestGetBinSyncTimeoutForNodes(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.active_node_dms = models.ActiveNode.get_mock_object_list(3)
        cls.active_node_dms[0].node_id = 1
        cls.active_node_dms[1].node_id = 2
        cls.active_node_dms[2].node_id = 3
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

        cls.drive_dms = [
            models.Drive.create(capacity=110 * models.DataUnit.mebibytes.value[0], node_id=1,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=100 * models.DataUnit.mebibytes.value[0], node_id=1,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=100 * models.DataUnit.mebibytes.value[0], node_id=1,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=200 * models.DataUnit.mebibytes.value[0], node_id=2,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=200 * models.DataUnit.mebibytes.value[0], node_id=2,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=200 * models.DataUnit.mebibytes.value[0], node_id=2,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=300 * models.DataUnit.mebibytes.value[0], node_id=3,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=300 * models.DataUnit.mebibytes.value[0], node_id=3,
                                type=enums.DriveType.block),
            models.Drive.create(capacity=300 * models.DataUnit.mebibytes.value[0], node_id=3,
                                type=enums.DriveType.block)
        ]

    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_timeout_for_nodes_removed(self, mock_list_drives, mock_get_bin_sync_timeout):
        mock_get_bin_sync_timeout.side_effect = [100]
        mock_list_drives.side_effect = [self.drive_dms, self.drive_dms]
        # call get_bin_sync_timeout_for_nodes using all 3 active nodes removed
        timeout = vl_helpers.get_bin_sync_timeout_for_nodes(CLUSTER_FM,
                                                            removed_node_dms=self.active_node_dms)
        # assert that there are 9 drives removed in the call to get_bin_sync_timeout
        assert len(mock_get_bin_sync_timeout.call_args_list[0][1]['removed_drive_dms']) == 9

    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_timeout_for_nodes_added(self, mock_list_drives, mock_get_bin_sync_timeout):
        mock_get_bin_sync_timeout.side_effect = [100]
        mock_list_drives.side_effect = [self.drive_dms, self.drive_dms]
        # call get_bin_sync_timeout_for_nodes using all 3 active nodes added
        timeout = vl_helpers.get_bin_sync_timeout_for_nodes(CLUSTER_FM,
                                                            added_node_dms=self.active_node_dms)
        # assert that there are 9 drives added in the call to get_bin_sync_timeout
        assert len(mock_get_bin_sync_timeout.call_args_list[0][1]['added_drive_dms']) == 9

    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_timeout_for_nodes_both(self, mock_list_drives, mock_get_bin_sync_timeout):
        mock_get_bin_sync_timeout.side_effect = [100]
        mock_list_drives.side_effect = [self.drive_dms[0:3], self.drive_dms[3:]]
        # call get_bin_sync_timeout_for_nodes using all 3 active nodes (1 added, 2 removed)
        timeout = vl_helpers.get_bin_sync_timeout_for_nodes(CLUSTER_FM,
                                                            added_node_dms=self.active_node_dms[0:1],
                                                            removed_node_dms=self.active_node_dms[1:])
        # assert that there are 3 drives removed in the call to get_bin_sync_timeout
        assert len(mock_get_bin_sync_timeout.call_args_list[0][1]['added_drive_dms']) == 3
        # assert that there are 6 drives removed in the call to get_bin_sync_timeout
        assert len(mock_get_bin_sync_timeout.call_args_list[0][1]['removed_drive_dms']) == 6

    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_timeout_for_nodes_find_largest(self, mock_list_drives, mock_get_bin_sync_timeout):
        mock_get_bin_sync_timeout.side_effect = [100]
        mock_list_drives.side_effect = [self.drive_dms, self.drive_dms[6:len(self.drive_dms)]]

        # call get_bin_sync_timeout_for_nodes using no removed or added nodes
        timeout = vl_helpers.get_bin_sync_timeout_for_nodes(CLUSTER_FM)
        # assert that there are 9 drives removed in the call to get_bin_sync_timeout
        assert len(mock_get_bin_sync_timeout.call_args_list[0][1]['removed_drive_dms']) == 3
        # assert that the drives we found are the largest drives
        exp_cap = self.drive_dms[-1].capacity
        assert mock_get_bin_sync_timeout.call_args_list[0][1]['removed_drive_dms'][0].capacity == exp_cap

    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_get_timeout_for_nodes_find_largest(self, mock_list_drives, mock_get_bin_sync_timeout):
        mock_get_bin_sync_timeout.side_effect = [100]
        mock_list_drives.side_effect = [self.drive_dms, []]

        # call get_bin_sync_timeout_for_nodes using no removed or added nodes
        timeout = vl_helpers.get_bin_sync_timeout_for_nodes(CLUSTER_FM)
        # assert that there are 9 drives removed in the call to get_bin_sync_timeout
        assert len(mock_get_bin_sync_timeout.call_args_list[0][1]['removed_drive_dms']) == 0


class TestBinSyncHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.properties = models.BinAssignmentsProperties.get_mock_object()

    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    def test_old_is_bin_syncing_returns_true(self, mock_get_assignments):
        bins = models.BinReport.get_mock_object_list(2)
        bins[0].services[0].status = enums.BinStatus.pending_active
        mock_get_assignments.return_value = models.BinAssignments.get_mock_object(
            overrides=dict(bins=bins, properties=None)
        )
        assert vl_helpers.is_bin_syncing(CLUSTER_FM) is True

    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    def test_old_is_bin_syncing_returns_false(self, mock_get_assignments):
        bins = models.BinReport.get_mock_object_list(1)
        for service in bins[0].services:
            service.status = enums.BinStatus.active
        mock_get_assignments.return_value = models.BinAssignments.get_mock_object(
            overrides=dict(bins=bins, properties=None)
        )
        assert vl_helpers.is_bin_syncing(CLUSTER_FM) is False

    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    def test_new_is_bin_syncing_returns_true(self, mock_get_assignments):
        self.properties.is_stable = False
        mock_get_assignments.return_value = models.BinAssignments.get_mock_object(
            overrides=dict(properties=self.properties)
        )
        assert vl_helpers.is_bin_syncing(CLUSTER_FM) is True

    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    def test_new_is_bin_syncing_returns_false(self, mock_get_assignments):
        self.properties.is_stable = True
        mock_get_assignments.return_value = models.BinAssignments.get_mock_object(
            overrides=dict(properties=self.properties)
        )
        assert vl_helpers.is_bin_syncing(CLUSTER_FM) is False

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(vl_helpers._bin_sync, 'is_bin_syncing')
    def test_wait_for_bin_sync_to_start(self, mock_is_bin_syncing, mock_timer):
        mock_is_bin_syncing.side_effect = [False, True]
        mock_timer.return_value = iter([True, True])
        vl_helpers.wait_for_bin_sync_to_start(CLUSTER_FM, timeout=123, poll_period=0)
        assert mock_is_bin_syncing.call_count == 2
        mock_timer.assert_called_with(123)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(vl_helpers._bin_sync, 'is_bin_syncing')
    def test_wait_for_bin_sync_to_start_timeout(self, mock_is_bin_syncing, mock_timer):
        mock_is_bin_syncing.side_effect = [False, True]
        mock_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_bin_sync_to_start(CLUSTER_FM, poll_period=0)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(vl_helpers._bin_sync, 'is_bin_syncing')
    def test_wait_for_bin_sync_to_finish_with_timeout_error(self, mock_is_bin_syncing, mock_timer):
        mock_is_bin_syncing.side_effect = [True, True]
        mock_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_bin_sync_to_finish_with_timeout(CLUSTER_FM, poll_period=0)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(vl_helpers._bin_sync, 'is_bin_syncing')
    def test_wait_for_bin_sync_to_finish_with_timeout_success(self, mock_is_bin_syncing, mock_timer):
        mock_is_bin_syncing.side_effect = [True, False]
        mock_timer.return_value = iter([True, True])
        vl_helpers.wait_for_bin_sync_to_finish_with_timeout(CLUSTER_FM, poll_period=0)
        mock_is_bin_syncing.assert_called()


@mock.patch.object(functional_models.Cluster, 'list_services')
@mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
class TestGetBinNodeMap(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.node_dms = models.ActiveNode.get_mock_object_list(6)
        cls.cluster_service_info_dms = models.ClusterServiceInfo.get_mock_object_list(6)

    def test_bin_node_map_with_correct_values(self, mock_get_bin_assignments, mock_list_services):
        service_info1 = models.ClusterService.get_mock_object(
            overrides={'service_type': 'block', 'service_id': 15, 'node_id': self.node_dms[0].node_id})
        service_info2 = models.ClusterService.get_mock_object(
            overrides={'service_type': 'block', 'service_id': 24, 'node_id': self.node_dms[1].node_id})
        self.cluster_service_info_dms[0].service = service_info1
        self.cluster_service_info_dms[1].service = service_info2
        mock_list_services.return_value = self.cluster_service_info_dms[:2]

        bin_service1 = models.BinService.get_mock_object(overrides={'service_id': 15})
        bin_service2 = models.BinService.get_mock_object(overrides={'service_id': 24})
        bin_report = models.BinReport.get_mock_object(overrides={'services': [bin_service1,
                                                                              bin_service2]})
        mock_get_bin_assignments.return_value = models.BinAssignments.get_mock_object(
            overrides=dict(bins=[bin_report])
        )

        bin_node_map = vl_helpers.get_bin_node_map(CLUSTER_FM)
        assert bin_node_map == {bin_report.bin_id: [n.node_id for n in self.node_dms[:2]]}

    def test_bin_node_map(self, mock_get_bin_assignments, mock_list_services):
        service_infos = models.ClusterService.get_mock_object_list(6, overrides={
            'service_type': enums.ServiceType.block})
        for cl_info_dm, srvc_info in zip(self.cluster_service_info_dms, service_infos):
            cl_info_dm.service = srvc_info

        mock_list_services.return_value = self.cluster_service_info_dms

        bin_services = models.BinService.get_mock_object_list(6)
        bin_reports = models.BinReport.get_mock_object_list(3)
        for b_srvc, srvc_info in zip(bin_services, service_infos):
            b_srvc.service_id = srvc_info.service_id
        for b_report, service_pair in zip(bin_reports, itertools.combinations(bin_services, 2)):
            b_report.services = list(service_pair)
        mock_get_bin_assignments.return_value = models.BinAssignments.get_mock_object(
            overrides=(dict(bins=bin_reports))
        )

        bin_node_map = vl_helpers.get_bin_node_map(CLUSTER_FM)
        assert len(bin_node_map.keys()) == 3
        assert sum([True for v in bin_node_map.values() if len(v) == 2]) == 3


@mock.patch(
    'valence.functional_models.Cluster.get_constants',
    mock.MagicMock(return_value=models.ClusterConstants.create(
        bin_sync_write_mib_per_second=20,
        bin_sync_new_service_write_mib_per_second=60
    ))
)
class TestGetBinSyncTimeout(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.drive_dms = [
            models.Drive.create(capacity=100 * models.DataUnit.mebibytes.value[0]),
            models.Drive.create(capacity=110 * models.DataUnit.mebibytes.value[0]),
            models.Drive.create(capacity=120 * models.DataUnit.mebibytes.value[0])
        ]

    def test_default(self):
        timeout = vl_helpers.get_bin_sync_timeout(self.cluster_fm)
        assert timeout == 1800

    def test_drives_added(self):
        timeout = vl_helpers.get_bin_sync_timeout(
            cluster_fm=self.cluster_fm,
            added_drive_dms=self.drive_dms
        )
        assert timeout == 6

    def test_drives_removed(self):
        timeout = vl_helpers.get_bin_sync_timeout(
            cluster_fm=self.cluster_fm,
            removed_drive_dms=self.drive_dms
        )
        assert timeout == 20

    def test_compound_operations(self):
        timeout = vl_helpers.get_bin_sync_timeout(
            cluster_fm=self.cluster_fm,
            added_drive_dms=self.drive_dms[:1],
            removed_drive_dms=self.drive_dms[1:]
        )
        assert timeout == 16

    def test_tolerance_override(self):
        timeout = vl_helpers.get_bin_sync_timeout(
            cluster_fm=self.cluster_fm,
            added_drive_dms=self.drive_dms[:1],
            removed_drive_dms=self.drive_dms[1:],
            tolerance_perc=50
        )
        assert timeout == 19


class TestBinsInError(unittest.TestCase):
    def setUp(self):
        self.bins = models.BinAssignments.get_mock_object(overrides={
            'bins': {'services': {'service_id': 3}}
        })

    def test_good_bins(self):
        assert len(vl_helpers.get_bins_assigned_to_service(self.bins, {2})) == 0

    def test_bad_bins(self):
        assert len(vl_helpers.get_bins_assigned_to_service(self.bins, {3})) == len(self.bins.bins)
