import unittest

import mock

from sf_api import result_models
from valence import (
    enums,
    functional_models,
    models,
    vl_components,
    vl_helpers,
)


@mock.patch(
    'valence.functional_models.Cluster.get_limits',
    mock.MagicMock(return_value=result_models.GetLimits.create(
        volume_size_min=1000000000,
        volume_size_max=17592186044416,
        volume_count_max=4000,
        volumes_per_account_count_max=2000
    ))
)
class TestGetVolumeCountForCapacity(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)
        cls.volume_dm = models.Volume.create()

    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_default(self, list_volumes):
        list_volumes.return_value = []

        # Test default volume count
        data_size = models.DataSize(100, enums.DataUnit.gigabytes)
        assert vl_helpers.get_volume_count_for_capacity(self.cluster_fm, data_size) == 60

        # Test less than the default volume count
        data_size = models.DataSize(42, enums.DataUnit.gigabytes)
        assert vl_helpers.get_volume_count_for_capacity(self.cluster_fm, data_size) == 42

        # Test more than the default volume count
        data_size = models.DataSize(1000, enums.DataUnit.gibibytes)
        assert vl_helpers.get_volume_count_for_capacity(
            cluster_fm=self.cluster_fm,
            block_data_size=data_size,
            volume_size_max=models.DataSize(10, enums.DataUnit.gibibytes)
        ) == 100

    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_w_existing_volumes(self, list_volumes):
        # Test taking existing volumes into account
        list_volumes.return_value = 3950 * [self.volume_dm]
        data_size = models.DataSize(100, enums.DataUnit.gigabytes)
        assert vl_helpers.get_volume_count_for_capacity(self.cluster_fm, data_size) == 50

    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    def test_size_restrictions(self, list_volumes):
        list_volumes.return_value = []

        # Test minimum volume size
        data_size = models.DataSize(1, enums.DataUnit.gigabytes)
        assert vl_helpers.get_volume_count_for_capacity(
            cluster_fm=self.cluster_fm,
            block_data_size=data_size,
            volume_size_min=models.DataSize(100, enums.DataUnit.mebibytes)
        ) == 1

        # Test maximum volume size
        data_size = models.DataSize(16385, enums.DataUnit.gibibytes)
        assert vl_helpers.get_volume_count_for_capacity(
            cluster_fm=self.cluster_fm,
            block_data_size=data_size,
            default_count=1,
            volume_size_max=models.DataSize(16390, enums.DataUnit.gibibytes)
        ) == 2


@mock.patch(
    'valence.functional_models.Cluster.get_limits',
    mock.MagicMock(return_value=result_models.GetLimits.create(
        volume_size_min=1000000000,
        volume_size_max=17592186044416,
        volume_count_max=4000,
        volumes_per_account_count_max=2000
    ))
)
class TestGetVolumeSizeForCapacity(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)

    def test_default(self):
        data_size = models.DataSize(103, enums.DataUnit.gigabytes)
        vol_count = 9

        # Test standard-sized volumes
        vol_size = vl_helpers.get_volume_size_for_capacity(
            cluster_fm=self.cluster_fm,
            block_data_size=data_size,
            volume_count=vol_count
        )
        target_vol_size = models.DataSize(data_size.bytes // vol_count)
        assert vol_size >= target_vol_size
        target_vol_size += models.DataSize(1, enums.DataUnit.mebibytes)
        assert vol_size <= target_vol_size

    def test_under_size(self):
        data_size = models.DataSize(103, enums.DataUnit.gigabytes)
        vol_count = 9

        # Test under-sized volumes
        vol_size = vl_helpers.get_volume_size_for_capacity(
            cluster_fm=self.cluster_fm,
            block_data_size=data_size,
            volume_count=vol_count,
            under_size=True
        )
        target_vol_size = models.DataSize(data_size.bytes // vol_count)
        assert vol_size <= target_vol_size
        target_vol_size -= models.DataSize(1, enums.DataUnit.mebibytes)
        assert vol_size >= target_vol_size

    def test_min_size(self):
        data_size = models.DataSize(8, enums.DataUnit.gigabytes)
        vol_count = 9

        # Test standard-sized volumes
        vol_size = vl_helpers.get_volume_size_for_capacity(
            cluster_fm=self.cluster_fm,
            block_data_size=data_size,
            volume_count=vol_count
        )
        assert vol_size == models.DataSize(1000000000).carbon_size


class TestUpdateClusterFullThreshold(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)
        cls.cannot_sync_fault = models.ClusterFault.create(
            code=enums.ClusterFaultCode.cluster_cannot_sync,
            severity=enums.ClusterFaultSeverity.error
        )
        cls.bs_unhealthy_fault = models.ClusterFault.create(
            code=enums.ClusterFaultCode.block_service_unhealthy,
            severity=enums.ClusterFaultSeverity.critical
        )

    @mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_default(self, list_faults, wait_for_bin_sync):
        list_faults.return_value = []
        gc_cmt = vl_components.RunGarbageCollection.mock(self.cluster_dm)
        vl_helpers.update_cluster_full_threshold(
            cluster_fm=self.cluster_fm,
            gc_cmt=gc_cmt
        )
        assert gc_cmt.applied
        assert list_faults.call_count == 3
        assert wait_for_bin_sync.called

    @mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_cannot_sync(self, list_faults, wait_for_bin_sync):
        list_faults.side_effect = [[self.cannot_sync_fault], [], []]
        gc_cmt = vl_components.RunGarbageCollection.mock(self.cluster_dm)
        vl_helpers.update_cluster_full_threshold(
            cluster_fm=self.cluster_fm,
            gc_cmt=gc_cmt
        )
        assert gc_cmt.applied
        assert list_faults.call_count == 3
        assert not wait_for_bin_sync.called

    @mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
    def test_bs_unhealthy(self, list_faults, wait_for_bin_sync):
        list_faults.side_effect = [[], [self.bs_unhealthy_fault], []]
        gc_cmt = vl_components.RunGarbageCollection.mock(self.cluster_dm)
        vl_helpers.update_cluster_full_threshold(
            cluster_fm=self.cluster_fm,
            gc_cmt=gc_cmt
        )
        assert gc_cmt.applied
        assert list_faults.call_count == 3
        assert not wait_for_bin_sync.called
