import unittest

import mock

from sf_api import result_models
from valence import (
    functional_models,
    models,
    vl_helpers,
)


class TestGetVolumeCountForCapacity(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)

    @mock.patch.object(functional_models.Cluster, 'get_immutable_values')
    def test_get_block_overhead_factor(self, get_immutable_values):
        immutable_values = result_models.GetImmutableValues.create(
            per_block_overhead_bytes=50
        )
        get_immutable_values.return_value = immutable_values

        block_overhead_factor = vl_helpers.get_block_overhead_factor(
            cluster_fm=self.cluster_fm,
            immutable_values=immutable_values
        )
        assert not get_immutable_values.called
        assert block_overhead_factor == 1. + float(immutable_values.per_block_overhead_bytes) / 4096

        block_overhead_factor = vl_helpers.get_block_overhead_factor(self.cluster_fm, block_size=2048)
        assert get_immutable_values.called
        assert block_overhead_factor == 1. + float(immutable_values.per_block_overhead_bytes) / 2048

    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_get_unusable_space_factor(self, get_constants):
        cluster_constants = models.ClusterConstants.create(bs_unusable_space_percent=7)
        get_constants.return_value = cluster_constants

        unusable_space_factor = vl_helpers.get_unusable_space_factor(
            cluster_fm=self.cluster_fm,
            cluster_constants=cluster_constants
        )
        assert not get_constants.called
        assert unusable_space_factor == 100. / (100 - cluster_constants.bs_unusable_space_percent)

        unusable_space_factor = vl_helpers.get_unusable_space_factor(self.cluster_fm)
        assert get_constants.called
        assert unusable_space_factor == 100. / (100 - cluster_constants.bs_unusable_space_percent)

    @mock.patch.object(functional_models.Cluster, 'get_immutable_values')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_volume_block_size_conversions(self, get_constants, get_immutable_values):
        get_constants.return_value = models.ClusterConstants.create(bs_unusable_space_percent=0)
        get_immutable_values.return_value = result_models.GetImmutableValues.create(
            per_block_overhead_bytes=0
        )

        volume_size = models.DataSize(10)
        volume_block_size = vl_helpers.get_volume_block_size(self.cluster_fm, volume_size)
        assert volume_block_size.bytes == 20
        assert volume_size == vl_helpers.get_block_volume_size(self.cluster_fm, volume_block_size)

        volume_block_size = vl_helpers.get_volume_block_size(
            cluster_fm=self.cluster_fm,
            volume_size=volume_size,
            volume_count=2
        )
        assert volume_block_size.bytes == 40
        assert volume_size == vl_helpers.get_block_volume_size(self.cluster_fm, volume_block_size) / 2

        volume_block_size = vl_helpers.get_volume_block_size(
            cluster_fm=self.cluster_fm,
            volume_size=volume_size,
            comp_ratio=2.
        )
        assert volume_block_size.bytes == 10
        assert volume_size == vl_helpers.get_block_volume_size(
            cluster_fm=self.cluster_fm,
            block_data_size=volume_block_size,
            comp_ratio=2.
        )

        volume_block_size = vl_helpers.get_volume_block_size(
            cluster_fm=self.cluster_fm,
            volume_size=volume_size,
            dedup_ratio=2.
        )
        assert volume_block_size.bytes == 10
        assert volume_size == vl_helpers.get_block_volume_size(
            cluster_fm=self.cluster_fm,
            block_data_size=volume_block_size,
            dedup_ratio=2.
        )

        get_immutable_values.return_value.per_block_overhead_bytes = 4096
        volume_block_size = vl_helpers.get_volume_block_size(self.cluster_fm, models.DataSize(10))
        assert volume_block_size.bytes == 40
        assert volume_size == vl_helpers.get_block_volume_size(self.cluster_fm, volume_block_size)

        get_constants.return_value.bs_unusable_space_percent = 50
        volume_block_size = vl_helpers.get_volume_block_size(self.cluster_fm, models.DataSize(10))
        assert volume_block_size.bytes == 80
        assert volume_size == vl_helpers.get_block_volume_size(self.cluster_fm, volume_block_size)

        # Ensure block-overhead factor is affected by compression ratio
        volume_block_size = vl_helpers.get_volume_block_size(
            cluster_fm=self.cluster_fm,
            volume_size=volume_size,
            comp_ratio=2.
        )
        assert volume_block_size.bytes == 60
        assert volume_size == vl_helpers.get_block_volume_size(
            cluster_fm=self.cluster_fm,
            block_data_size=volume_block_size,
            comp_ratio=2.
        )

    @mock.patch.object(functional_models.Cluster, 'get_immutable_values')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_block_meta_size_conversions(self, get_constants, get_immutable_values):
        get_constants.return_value = models.ClusterConstants.create(bs_unusable_space_percent=0)
        get_immutable_values.return_value = result_models.GetImmutableValues.create(
            per_block_overhead_bytes=0,
            per_block_metadata_bytes=16,
            per_page_metadata_blocks=341,
            per_page_metadata_bytes=4096
        )

        block_data_size = models.DataSize(20 * 4096)
        meta_data_size = vl_helpers.get_block_metadata_size(self.cluster_fm, block_data_size)
        assert meta_data_size.bytes == 320
        assert block_data_size == vl_helpers.get_metadata_block_size(self.cluster_fm, meta_data_size)

        meta_data_size = vl_helpers.get_block_metadata_size(
            cluster_fm=self.cluster_fm,
            block_data_size=block_data_size,
            comp_ratio=2.
        )
        assert meta_data_size.bytes == 641
        assert block_data_size == vl_helpers.get_metadata_block_size(
            cluster_fm=self.cluster_fm,
            metadata_size=meta_data_size,
            comp_ratio=2.
        )

        meta_data_size = vl_helpers.get_block_metadata_size(
            cluster_fm=self.cluster_fm,
            block_data_size=block_data_size,
            dedup_ratio=2.
        )
        assert meta_data_size.bytes == 641
        assert block_data_size == vl_helpers.get_metadata_block_size(
            cluster_fm=self.cluster_fm,
            metadata_size=meta_data_size,
            dedup_ratio=2.
        )

        get_immutable_values.return_value.per_block_overhead_bytes = 4096
        meta_data_size = vl_helpers.get_block_metadata_size(self.cluster_fm, block_data_size)
        assert meta_data_size.bytes == 160
        assert block_data_size == vl_helpers.get_metadata_block_size(self.cluster_fm, meta_data_size)

        get_constants.return_value.bs_unusable_space_percent = 50
        meta_data_size = vl_helpers.get_block_metadata_size(self.cluster_fm, block_data_size)
        assert meta_data_size.bytes == 80
        assert block_data_size == vl_helpers.get_metadata_block_size(self.cluster_fm, meta_data_size)

        # Ensure block-overhead factor is affected by compression ratio
        meta_data_size = vl_helpers.get_block_metadata_size(
            cluster_fm=self.cluster_fm,
            block_data_size=block_data_size,
            comp_ratio=2.
        )
        assert meta_data_size.bytes == 112
        meta_block_size = vl_helpers.get_metadata_block_size(
            cluster_fm=self.cluster_fm,
            metadata_size=meta_data_size,
            comp_ratio=2.
        )
        assert meta_block_size - 4096 <= block_data_size <= meta_block_size
