import time
import unittest

import mock
import pytest
import sf_models

from valence import (
    enums,
    errors,
    functional_models,
    models,
    utils,
    vl_helpers,
)

RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
CLUSTER_FM = functional_models.Cluster(RS_CLUSTER_DM, authenticate=False)


class TestGetDefaultSliceDriveSlot(unittest.TestCase):
    """Tests proper drive_dm -> active_node_dm lookup."""

    def test_helper_mapping(self):
        node_dm = models.ActiveNode.get_mock_object()

        node_dm.platform_info.node_type = sf_models.NodeType.STANDARD_TYPES.SF3010
        assert -1 == vl_helpers._drive.get_default_slice_drive_slot(node_dm)

        node_dm.platform_info.node_type = sf_models.NodeType.STANDARD_TYPES.SF9608
        assert 1 == vl_helpers._drive.get_default_slice_drive_slot(node_dm)

        node_dm.platform_info.node_type = sf_models.NodeType.STANDARD_TYPES.SF9010
        assert 0 == vl_helpers._drive.get_default_slice_drive_slot(node_dm)


class TestGetActiveNodeDmFromDriveDm(unittest.TestCase):
    """Tests proper drive_dm -> active_node_dm lookup."""

    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_helper_mapping(self, mock_list_active_nodes):
        node_0 = models.ActiveNode.get_mock_object(overrides=dict(node_id=0))
        node_1 = models.ActiveNode.get_mock_object(overrides=dict(node_id=1))
        node_dms = [node_0, node_1]
        drive_dm = models.Drive.get_mock_object(overrides=dict(node_id=0))

        mock_list_active_nodes.return_value = node_dms

        ret = vl_helpers._drive.get_active_node_dm_from_drive_dm(CLUSTER_FM, drive_dm)

        assert ret == node_0


@mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
@mock.patch.object(vl_helpers._drive, 'get_active_node_dm_from_drive_dm')
class TestSetDrivesType(unittest.TestCase):

    def test_enable_mdss_false(self, mock_get_node_dm, mock_get_default_slice_drive_slot):
        # Create one primary slice, one mdss slice
        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=0, slot=0, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=0, slot=2, type='volume'))
        drive_3 = models.Drive.get_mock_object(overrides=dict(node_id=0, slot=3, type='volume'))
        drive_dms = [drive_0, drive_2, drive_3]

        node_dm = models.ActiveNode.get_mock_object(overrides=dict(node_id=0))

        mock_get_node_dm.return_value = node_dm
        mock_get_default_slice_drive_slot.return_value = 0

        result = vl_helpers.set_drives_type(CLUSTER_FM, drive_dms)
        assert result[0].type is enums.DriveType.slice
        assert result[1].type is enums.DriveType.block

    def test_enable_mdss_true(self, mock_get_node_dm, mock_get_default_slice_drive_slot):
        drive_0 = models.Drive.get_mock_object(overrides=dict(node_id=0, slot=0, type='volume'))
        drive_2 = models.Drive.get_mock_object(overrides=dict(node_id=0, slot=2, type='volume'))
        drive_3 = models.Drive.get_mock_object(overrides=dict(node_id=0, slot=3, type='volume'))
        drive_dms = [drive_0, drive_2, drive_3]

        node_dm = models.ActiveNode.get_mock_object(overrides=dict(node_id=0))

        mock_get_node_dm.return_value = node_dm
        mock_get_default_slice_drive_slot.return_value = 0

        result = vl_helpers.set_drives_type(CLUSTER_FM, drive_dms, enable_mdss=True)
        assert result[0].type is enums.DriveType.slice
        assert result[1].type is enums.DriveType.slice


@mock.patch.object(functional_models.AsyncHandler, 'get_models')
@mock.patch.object(functional_models.Cluster, 'add_drives')
class TestAddDrivesHelpers(unittest.TestCase):
    """Tests cluster_helpers.add_drives."""

    def setUp(self):
        self.drive_dms = models.Drive.get_mock_object_list(2)
        drive_ids = [d.drive_id for d in self.drive_dms]
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.async_handler = functional_models.AddDrivesAsyncHandler(CLUSTER_FM,
                                                                     drive_ids,
                                                                     async_handle_dm)

        self.slice_drive_node_dm = models.ActiveNode.get_mock_object()

        self.single_drive_dms = [models.Drive.create(drive_id=0,
                                                     node_id=1,
                                                     slot=0,
                                                     status=enums.DriveStatus.active,
                                                     type=enums.DriveType.slice)]

        self.slot_two_mdss_drive_dms = [models.Drive.create(drive_id=1,
                                                            node_id=1,
                                                            slot=2,
                                                            status=enums.DriveStatus.active,
                                                            type=enums.DriveType.slice)]

        self.slot_two_non_mdss_drive_dms = [models.Drive.create(drive_id=1,
                                                                node_id=1,
                                                                slot=2,
                                                                status=enums.DriveStatus.active,
                                                                type=enums.DriveType.block)]

        self.slot_two_drives = [self.slot_two_mdss_drive_dms[0], self.slot_two_non_mdss_drive_dms[0]]

        mock_constant_value = mock.Mock()
        mock_constant_value.allow_drive_type_modification_of_slot_num = 2

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_add_drives_with_no_mdss(self, mock_get_constants, mock_list_drives,
                                     mock_add_drives, mock_get_models):

        mock_list_drives.return_value = []
        mock_add_drives.return_value = self.async_handler
        mock_get_models.return_value = self.drive_dms
        response = vl_helpers.add_drives(CLUSTER_FM, self.drive_dms)
        assert response == self.drive_dms

    @mock.patch.object(vl_helpers._drive, 'set_drives_type')
    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(vl_helpers._drive, 'get_active_node_dm_from_drive_dm')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_add_single_drive_with_mdss_already_slice(self, mock_get_constants, mock_list_drives,
                                                      mock_list_active_nodes, mock_get_node,
                                                      mock_get_slice_slot, mock_set_drive_type,
                                                      mock_add_drives, mock_get_models):

        mock_get_slice_slot.return_value = 0
        mock_get_node.return_value = self.slice_drive_node_dm
        mock_list_drives.return_value = self.slot_two_mdss_drive_dms
        vl_helpers.add_drives(CLUSTER_FM, self.single_drive_dms)
        self.assertEqual(mock_list_drives.call_count, 2)

    @mock.patch.object(vl_helpers._drive, 'set_drives_type')
    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(vl_helpers._drive, 'get_active_node_dm_from_drive_dm')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_add_single_drive_with_mdss_not_already_slice(self, mock_get_constants, mock_list_drives,
                                                          mock_list_active_nodes, mock_get_node,
                                                          mock_get_slice_slot, mock_set_drive_type,
                                                          mock_add_drives, mock_get_models):

        mock_get_slice_slot.return_value = 0
        mock_get_node.return_value = self.slice_drive_node_dm
        mock_list_drives.side_effect = [self.slot_two_mdss_drive_dms, self.slot_two_non_mdss_drive_dms]
        vl_helpers.add_drives(CLUSTER_FM, self.single_drive_dms)
        mock_set_drive_type.assert_called_once()

    @mock.patch.object(vl_helpers._drive, 'set_drives_type')
    @mock.patch.object(vl_helpers._drive, 'get_default_slice_drive_slot')
    @mock.patch.object(vl_helpers._drive, 'get_active_node_dm_from_drive_dm')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    def test_add_multiple_drives_with_mdss(self, mock_get_constants, mock_list_drives,
                                           mock_list_active_nodes, mock_get_node,
                                           mock_get_slice_slot, mock_set_drive_type,
                                           mock_add_drives, mock_get_models):

        mock_get_slice_slot.return_value = 2
        mock_get_node.return_value = self.slice_drive_node_dm
        mock_list_drives.return_value = self.slot_two_mdss_drive_dms
        vl_helpers.add_drives(CLUSTER_FM, self.slot_two_drives)
        mock_set_drive_type.assert_called_once()


@mock.patch.object(functional_models.Cluster, 'list_drives')
@mock.patch.object(functional_models.Cluster, 'get_constants')
class TestIsMdssEnabled(unittest.TestCase):
    """Tests cluster_helpers.is_mdss_enabled."""

    def setUp(self):
        self.mdss_slot_enabled_drive_dms = [
            models.Drive.create(drive_id=1, node_id=1, slot=2, status=enums.DriveStatus.active,
                                type=enums.DriveType.slice)]
        self.mdss_slot_disabled_drive_dms = [
            models.Drive.create(drive_id=2, node_id=3, slot=2, status=enums.DriveStatus.available,
                                type=enums.DriveType.block)]
        mock_constant_value = mock.Mock()
        mock_constant_value.allow_drive_type_modification_of_slot_num = 2
        self.mock_const_val = mock_constant_value

    def test_mdss_enabled(self, mock_get_constants, mock_list_drives):
        mock_get_constants.return_value = self.mock_const_val
        mock_list_drives.return_value = self.mdss_slot_enabled_drive_dms
        self.assertTrue(vl_helpers.is_mdss_enabled(CLUSTER_FM))

    def test_mdss_disabled(self, mock_get_constants, mock_list_drives):
        mock_get_constants.return_value = self.mock_const_val
        mock_list_drives.return_value = []
        self.assertFalse(vl_helpers.is_mdss_enabled(CLUSTER_FM))


class TestRemoveDrivesHelpers(unittest.TestCase):
    """Tests cluster_helpers.remove_drives."""

    def setUp(self):
        self.drive_dms = models.Drive.get_mock_object_list(2)
        drive_ids = [d.drive_id for d in self.drive_dms]
        async_handle_dm = models.AsyncHandle.get_mock_object()
        self.async_handler = functional_models.RemoveDrivesAsyncHandler(CLUSTER_FM,
                                                                        drive_ids,
                                                                        async_handle_dm)

    @mock.patch.object(functional_models.AsyncHandler, 'get_models')
    @mock.patch.object(functional_models.Cluster, 'remove_drives')
    def test_remove_drives(self, mock_remove_drives, mock_get_models):
        mock_remove_drives.return_value = self.async_handler
        mock_get_models.return_value = self.drive_dms
        response = vl_helpers.remove_drives(CLUSTER_FM, self.drive_dms)
        assert response == self.drive_dms


class TestAddAvailableDrivesHelpers(unittest.TestCase):

    @mock.patch.object(vl_helpers._drive, 'add_drives')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_add_available_drives(self, mock_list_drives, mock_add_drives):
        vl_helpers.add_available_drives(CLUSTER_FM)


class TestWaitForDriveStatusHelper(unittest.TestCase):
    def setUp(self):
        self.num_drives = 2
        self.drive_dms = models.Drive.get_mock_object_list(self.num_drives)
        self.cluster_dm = models.Cluster.get_mock_object()
        self.cluster_fm = functional_models.Cluster(self.cluster_dm)

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_status(self, mock_list_drives, mock_timeout_timer, mock_sleep):
        mock_timeout_timer.return_value = iter([True] * self.num_drives)
        mock_list_drives.side_effect = [[self.drive_dms[0]], self.drive_dms]
        vl_helpers.wait_for_drive_status(self.cluster_fm, self.drive_dms)
        mock_list_drives.assert_called()

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_statuses(self, mock_list_drives, mock_timeout_timer, mock_sleep):
        mock_timeout_timer.return_value = iter([True] * self.num_drives)
        mock_list_drives.side_effect = [[self.drive_dms[0]], self.drive_dms]
        vl_helpers.wait_for_drive_statuses(self.cluster_fm, self.drive_dms, None,
                                           [enums.DriveStatus.available])
        mock_list_drives.assert_called()

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_status_no_serial_check(self, mock_list_drives, mock_timeout_timer,
                                                   mock_sleep):
        mock_timeout_timer.return_value = iter([True] * self.num_drives)
        mock_list_drives.side_effect = [[self.drive_dms[0]], self.drive_dms]
        vl_helpers.wait_for_drive_status(self.cluster_fm, self.drive_dms, check_drive_serial=False)
        mock_list_drives.assert_called()

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_statuses_no_serial_check(self, mock_list_drives, mock_timeout_timer,
                                                     mock_sleep):
        mock_timeout_timer.return_value = iter([True] * self.num_drives)
        mock_list_drives.side_effect = [[self.drive_dms[0]], self.drive_dms]
        vl_helpers.wait_for_drive_statuses(self.cluster_fm, self.drive_dms, None,
                                           [enums.DriveStatus.available], check_drive_serial=False)
        mock_list_drives.assert_called()

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_any_drive_status(self, mock_list_drives, mock_timeout_timer, mock_sleep):
        mock_timeout_timer.return_value = iter([True] * self.num_drives)
        mock_list_drives.side_effect = [[self.drive_dms[0]], self.drive_dms]
        vl_helpers.wait_for_any_drive_statuses(self.cluster_fm, [enums.DriveStatus.available])
        mock_list_drives.assert_called()

    @mock.patch('time.sleep')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_any_drive_status_sleeps(self, mock_list_drives, mock_timeout_timer, mock_sleep):
        mock_timeout_timer.return_value = iter([True] * self.num_drives)
        mock_list_drives.side_effect = [None, self.drive_dms]
        vl_helpers.wait_for_any_drive_statuses(self.cluster_fm, [enums.DriveStatus.available])
        mock_list_drives.assert_called()

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_statuses_raises_exception(self, mock_list_drives, mock_timeout_timer):
        mock_list_drives.return_value = []
        mock_timeout_timer.return_value = iter([False])
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_statuses(self.cluster_fm, self.drive_dms, None,
                                               [enums.DriveStatus.available])

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_status_raises_exception(self, mock_list_drives, mock_timeout_timer):
        mock_list_drives.return_value = []
        mock_timeout_timer.return_value = iter([False])
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_status(self.cluster_fm, self.drive_dms)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_any_drive_status_raises_exception(self, mock_list_drives, mock_timeout_timer):
        mock_list_drives.return_value = []
        mock_timeout_timer.return_value = iter([False])
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_any_drive_statuses(self.cluster_fm, [enums.DriveStatus.available])


class TestWaitForDriveAssignments(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mock_services = models.ClusterServiceInfo.get_mock_object_list(1)
        cls.mock_assignments_none = models.BinAssignments.create(bins=[])
        cls.mock_assignments_some = models.BinAssignments.create(
            bins=models.BinReport.get_mock_object_list(1)
        )
        cls.mock_drive_dms = [
            models.Drive.create(type=enums.DriveType.block),
            models.Drive.create(type=enums.DriveType.block),
            models.Drive.create(type=enums.DriveType.slice)
        ]

    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_default(self, list_services, get_bin_assignments):
        list_services.return_value = self.mock_services
        get_bin_assignments.side_effect = [self.mock_assignments_none, self.mock_assignments_some]

        vl_helpers.wait_for_drive_assignments(CLUSTER_FM, self.mock_drive_dms)
        list_services.assert_called_with(drive_dms=self.mock_drive_dms[:-1])
        get_bin_assignments.assert_called_with(services=self.mock_services)
        assert get_bin_assignments.call_count == 2

    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_no_block_drives(self, list_services, get_bin_assignments):
        vl_helpers.wait_for_drive_assignments(CLUSTER_FM, [self.mock_drive_dms[-1]])
        assert not list_services.called
        assert not get_bin_assignments.called

    @mock.patch('time.time', mock.Mock(side_effect=[0, 1, 2, 3, 4]))
    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_services_timeout(self, list_services, get_bin_assignments):
        list_services.return_value = []

        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_assignments(CLUSTER_FM, self.mock_drive_dms, timeout=3)
        assert list_services.call_count == 3

    @mock.patch('time.time', mock.Mock(side_effect=[0, 1, 2, 3, 4]))
    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    def test_assignments_timeout(self, list_services, get_bin_assignments):
        get_bin_assignments.return_value = self.mock_assignments_none

        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_assignments(CLUSTER_FM, self.mock_drive_dms, timeout=3)
        assert get_bin_assignments.call_count == 2


class TestFailBlockDrives(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()
        cls.active_nodes = models.ActiveNode.get_mock_object_list(4)
        cls.mock_drive_dms = models.Drive.get_mock_object_list(2)

    @mock.patch.object(functional_models.Node, 'fail_drives', autospec=True)
    @mock.patch.object(functional_models.Node, 'authenticate')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(vl_helpers._drive, 'check_fake_drives_flag_on_node', autospec=True)
    @mock.patch.object(vl_helpers._drive, '_create_node_to_drive_map', autospec=True)
    @mock.patch.object(vl_helpers._drive, '_wait_for_drive_to_fail', autospec=True)
    def test_fail_drives_in_cluster(
            self, mock_wait_for_drive_to_fail, mock_create_node_to_drive_map,
            mock_check_fake_drives_flag_on_node, mock_list_active_nodes,
            mock_authenticate, mock_fail_block_drives):
        mock_active_nodes = models.ActiveNode.get_mock_object_list(2)
        mock_active_nodes[0].resource_dm.mip = '10.117.1.1'

        mock_list_active_nodes.return_value = mock_active_nodes
        mock_create_node_to_drive_map.return_value = {'10.117.1.1': [1, 2, 3]}
        vl_helpers.fail_drives_in_cluster(CLUSTER_FM, self.mock_drive_dms)

    def test_create_node_to_drive_map(self):
        self.mock_drive_dms[0].node_id = 1
        self.mock_drive_dms[1].node_id = 1
        self.active_nodes[1].node_id = 1
        vl_helpers._drive._create_node_to_drive_map(self.active_nodes, self.mock_drive_dms)

    @mock.patch.object(functional_models.Node, 'set_startup_flags')
    @mock.patch.object(functional_models.Node, 'stop_binary')
    @mock.patch.object(functional_models.Node, 'start_binary')
    def test_set_fake_drives_flag_on_node(self, mock_start_binary, mock_stop_binary,
                                          mock_set_startup_flags):
        vl_helpers.set_fake_drives_flag_on_node(self.node_fm)

    @mock.patch.object(functional_models.Node, 'get_startup_flags')
    def test_check_fake_drives_flag_on_node(self, mock_get_startup_flags):
        mock_get_startup_flags.return_value = {'sfapp': [{'current': True}]}
        vl_helpers.check_fake_drives_flag_on_node(self.node_fm)

    @mock.patch.object(functional_models.Node, 'get_startup_flags')
    def test_check_fake_drives_flag_on_node_fails(self, mock_get_startup_flags):
        mock_get_startup_flags.return_value = {'sfapp': [{'current': False}]}
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.check_fake_drives_flag_on_node(self.node_fm)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(time, 'sleep', return_value=None)
    def test_wait_for_drive_to_fail(self, mock_sleep, mock_list_drives, mock_timeout_timer):
        mock_failed_drive = models.Drive.get_mock_object()
        self.mock_drive_dms[0].drive_id = 1
        self.mock_drive_dms[0].status = enums.DriveStatus.failed
        mock_timeout_timer.return_value = iter([True, True])
        mock_failed_drive.drive_id = 1
        mock_list_drives.return_value = self.mock_drive_dms
        vl_helpers._drive._wait_for_drive_to_fail(CLUSTER_FM, mock_failed_drive, 0)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(time, 'sleep', return_value=None)
    def test_wait_for_drive_to_fail_to_timeout(self, mock_sleep, mock_list_drives, mock_timeout_timer):
        mock_timeout_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers._drive._wait_for_drive_to_fail(CLUSTER_FM, mock.Mock(), 0)


class TestRestoreDrives(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()
        cls.active_nodes = models.ActiveNode.get_mock_object_list(4)
        cls.mock_drive_dms = models.Drive.get_mock_object_list(2)

    @mock.patch.object(functional_models.Node, 'restore_failed_drives', autospec=True)
    @mock.patch.object(functional_models.Node, 'authenticate')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    @mock.patch.object(vl_helpers._drive, '_create_node_to_drive_map', autospec=True)
    @mock.patch.object(vl_helpers._drive, '_wait_for_drive_restore', autospec=True)
    def test_fail_block_drives_in_cluster(
            self, mock_wait_for_drive_to_restore, mock_create_node_to_drive_map,
            mock_list_active_nodes, mock_authenticate, mock_restore_block_drives):
        mock_active_nodes = models.ActiveNode.get_mock_object_list(2)
        mock_active_nodes[0].resource_dm.mip = '10.117.1.1'
        mock_list_active_nodes.return_value = mock_active_nodes
        mock_create_node_to_drive_map.return_value = {'10.117.1.1': [1, 2, 3]}
        vl_helpers.restore_failed_drives(CLUSTER_FM, self.mock_drive_dms)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(time, 'sleep', return_value=None)
    def test_wait_for_restore(self, mock_sleep, mock_list_drives, mock_timer):
        mock_restored_drives = models.Drive.get_mock_object()
        self.mock_drive_dms[0].drive_id = 1
        self.mock_drive_dms[0].status = enums.DriveStatus.available
        mock_timer.return_value = iter([True, True])
        mock_restored_drives.drive_id = 1
        mock_list_drives.return_value = self.mock_drive_dms
        vl_helpers._drive._wait_for_drive_restore(CLUSTER_FM, mock_restored_drives, 0)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(time, 'sleep', return_value=None)
    def test_wait_for_drive_to_restore_timeout(self, mock_sleep, mock_list_drives, mock_timeout_timer):
        mock_timeout_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers._drive._wait_for_drive_restore(CLUSTER_FM, mock.Mock(), 0)


@mock.patch.object(functional_models.Cluster, 'list_services', mock.MagicMock())
@mock.patch.object(functional_models.Cluster, 'list_drives')
@mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
@mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
@mock.patch.object(vl_helpers._bin_sync, 'wait_for_bin_sync_to_finish_with_timeout')
@mock.patch.object(vl_helpers._faults, 'wait_for_cluster_fault_resolution')
@mock.patch.object(vl_helpers._drive, 'add_drives')
@mock.patch.object(vl_helpers._drive, 'wait_for_drive_status')
@mock.patch.object(vl_helpers._drive, 'wait_for_drive_statuses')
class TestRecoverDrives(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mock_drive_dms = models.Drive.get_mock_object_list(2)
        cls.mock_drive_dms[0].type = enums.DriveType.block
        cls.mock_bin_report = models.BinReport.create()

    def test_recover_before_timeout(self, wait_for_statuses, wait_for_status, add_drives,
                                    wait_for_fault_resolution, wait_for_sync_finish, get_bin_assignments,
                                    list_faults, list_drives):
        list_faults.return_value = []
        get_bin_assignments.return_value = models.BinAssignments.create(bins=[self.mock_bin_report])
        list_drives.return_value = []
        vl_helpers.recover_failed_drives(CLUSTER_FM, self.mock_drive_dms)
        assert wait_for_fault_resolution.called
        assert wait_for_statuses.called
        assert wait_for_statuses.call_args[1]['drive_dms'] == self.mock_drive_dms
        assert wait_for_statuses.call_args[1]['statuses'] == [enums.DriveStatus.active,
                                                              enums.DriveStatus.available]
        assert not wait_for_status.called
        assert not add_drives.called
        assert wait_for_sync_finish.called

    def test_recover_after_timeout(self, wait_for_statuses, wait_for_status, add_drives,
                                   wait_for_fault_resolution, wait_for_sync_finish, get_bin_assignments,
                                   list_faults, list_drives):
        list_faults.return_value = []
        get_bin_assignments.side_effect = [
            models.BinAssignments.create(bins=[]),
            models.BinAssignments.create(bins=[self.mock_bin_report])
        ]
        list_drives.return_value = [self.mock_drive_dms[0]]
        vl_helpers.recover_failed_drives(CLUSTER_FM, self.mock_drive_dms)
        assert wait_for_fault_resolution.called
        assert wait_for_statuses.called
        assert wait_for_status.call_count == 1
        assert wait_for_status.call_args[1]['drive_dms'] == [self.mock_drive_dms[0]]
        assert wait_for_status.call_args[1]['status'] == enums.DriveStatus.available
        assert add_drives.called
        assert add_drives.call_args[1]['drive_dms'] == [self.mock_drive_dms[0]]
        assert wait_for_sync_finish.called

    def test_recover_after_timeout_no_sync(self, wait_for_statuses, wait_for_status, add_drives,
                                           wait_for_fault_resolution, wait_for_sync_finish,
                                           get_bin_assignments, list_faults, list_drives):
        list_faults.return_value = [1]
        get_bin_assignments.side_effect = [
            models.BinAssignments.create(bins=[]),
            models.BinAssignments.create(bins=[self.mock_bin_report])
        ]
        list_drives.return_value = []
        vl_helpers.recover_failed_drives(CLUSTER_FM, self.mock_drive_dms, wait_for_sync=False)
        assert not wait_for_fault_resolution.called
        assert wait_for_statuses.called
        assert not wait_for_status.called
        assert not add_drives.called
        assert not wait_for_sync_finish.called

    def test_recover_after_timeout_no_prds(self, wait_for_statuses, wait_for_status, add_drives,
                                           wait_for_fault_resolution, wait_for_sync_finish,
                                           get_bin_assignments, list_faults, list_drives):
        list_faults.return_value = []
        get_bin_assignments.return_value = models.BinAssignments.create(bins=[self.mock_bin_report])
        list_drives.return_value = []
        vl_helpers.recover_failed_drives(CLUSTER_FM, self.mock_drive_dms)
        assert wait_for_fault_resolution.called
        assert wait_for_statuses.called
        assert not wait_for_status.called
        assert not add_drives.called
        assert wait_for_sync_finish.called

    def test_recover_after_sync_out(self, wait_for_statuses, wait_for_status, add_drives,
                                    wait_for_fault_resolution, wait_for_sync_finish,
                                    get_bin_assignments, list_faults, list_drives):
        list_faults.return_value = []
        get_bin_assignments.return_value = models.BinAssignments.create(bins=[self.mock_bin_report])
        list_drives.return_value = [self.mock_drive_dms[0]]
        vl_helpers.recover_failed_drives(CLUSTER_FM, self.mock_drive_dms)
        assert wait_for_fault_resolution.called
        assert wait_for_statuses.called
        assert not wait_for_status.called
        assert add_drives.called
        assert add_drives.call_args[1]['drive_dms'] == [self.mock_drive_dms[0]]
        assert wait_for_sync_finish.called


@mock.patch.object(vl_helpers._drive, 'wait_for_services_to_decommission')
@mock.patch.object(vl_helpers._bin_sync, 'get_bins_assigned_to_service')
@mock.patch.object(vl_helpers._slice, 'get_slices_assigned_to_services')
@mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
@mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
@mock.patch.object(vl_helpers._bin_sync, 'wait_for_bin_sync_to_finish_with_timeout')
@mock.patch.object(vl_helpers._slice, 'wait_for_slice_sync_to_finish')
@mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
@mock.patch.object(vl_helpers._slice, 'calculate_slice_sync_timeout')
@mock.patch.object(functional_models.Cluster, 'list_services')
class TestForSyncOut(unittest.TestCase):

    def setUp(self):
        self.drives = [models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.block}),
                       models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.volume}),
                       ]
        self.services = [models.ClusterServiceInfo.get_mock_object(
            overrides={'service': {'service_id': 3}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'service_id': 4}})]
        self.slices = models.SliceReport.get_mock_object(overrides={
            'slices': {'primary': 0,
                       'live_secondaries': [1, 2],
                       'dead_secondaries': []}})

    def test_sync(self, mock_services, mock_bin_timeout, mock_slice_timeout, mock_slice_wait,
                  mock_bin_wait, mock_slice_report, mock_bins, mock_error_slices, mock_error_bins,
                  mock_remove):
        mock_services.side_effect = [[self.services[0]], [], [self.services[1]]]
        mock_bin_timeout.return_value = 1
        mock_slice_timeout.return_value = 1
        mock_slice_report.return_value = self.slices
        mock_error_bins.return_value = []
        mock_error_slices.return_value = []
        vl_helpers.wait_for_sync_out(CLUSTER_FM, self.drives)
        assert mock_bin_wait.called
        assert mock_slice_wait.called

    def test_bad_slice_behavior(self, mock_services, mock_bin_timeout, mock_slice_timeout,
                                mock_slice_wait, mock_bin_wait, mock_slice_report, mock_bin,
                                mock_error_slices, mock_error_bins,
                                mock_remove):
        mock_services.side_effect = [[self.services[0]], [], [self.services[1]]]
        mock_bin_timeout.return_value = 1
        mock_slice_timeout.return_value = 1
        slices = models.SliceReport.get_mock_object(overrides={
            'slices': {'primary': 3,
                       'live_secondaries': [4, 2],
                       'dead_secondaries': [0]}})
        mock_error_slices.return_value = slices.slices
        with pytest.raises(errors.ElementInvalidStateError):
            vl_helpers.wait_for_sync_out(CLUSTER_FM, self.drives)

    def test_bad_block_behavior(self, mock_services, mock_bin_timeout, mock_slice_timeout,
                                mock_slice_wait, mock_bin_wait, mock_slice_report, mock_bins,
                                mock_error_slices, mock_error_bins, mock_remove):
        mock_services.side_effect = [[self.services[0]], [], [self.services[1]]]
        mock_bin_timeout.return_value = 1
        mock_slice_timeout.return_value = 1
        mock_error_slices.return_value = []
        mock_error_bins.return_value = models.BinAssignments.get_mock_object(overrides={
            'bins': {'services': {'service_id': 3}}
        }).bins
        with pytest.raises(errors.ElementInvalidStateError):
            vl_helpers.wait_for_sync_out(CLUSTER_FM, self.drives)


@mock.patch('time.sleep')
@mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
@mock.patch.object(vl_helpers._slice, 'calculate_slice_sync_timeout')
@mock.patch.object(utils, 'timeout_timer')
@mock.patch.object(functional_models.Cluster, 'list_services')
class TestWaitForDecommission(unittest.TestCase):
    def setUp(self):
        self.drives = [models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.block,
                                                               'driveID': 4}),
                       models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.volume,
                                                               'driveID': 5}),
                       ]
        self.services = [models.ClusterServiceInfo.get_mock_object(
            overrides={'service': {'service_id': 3}}),
            models.ClusterServiceInfo.get_mock_object(overrides={'service': {'service_id': 4}})]

    def test_wait_for_removal(self, mock_services, mock_timer, mock_slice_timeout, mock_bin_timeout,
                              mock_sleep):
        mock_bin_timeout.return_value = 1
        mock_slice_timeout.return_value = 1
        mock_services.side_effect = [[], [], [self.services[1]], [self.services[0]] * 10]
        mock_timer.side_effect = [iter([True] * 10), iter([True] * 10)]
        vl_helpers.wait_for_services_to_decommission(CLUSTER_FM, self.drives, {6, 7})

    def test_drive_timeout(self, mock_services, mock_timer, mock_slice_timeout, mock_bin_timeout,
                           mock_sleep):
        mock_bin_timeout.return_value = 1
        mock_slice_timeout.return_value = 1
        mock_services.return_value = self.services
        mock_timer.return_value = iter([True, False])
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_services_to_decommission(CLUSTER_FM, self.drives, {4})


@mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
@mock.patch.object(vl_helpers._slice, 'calculate_slice_sync_timeout')
@mock.patch.object(utils, 'timeout_timer')
@mock.patch.object(functional_models.Cluster, 'list_services')
class TestWaitForHealthy(unittest.TestCase):
    def setUp(self):
        self.drives = [models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.block,
                                                               'drive_id': 3}),
                       models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.volume,
                                                               'drive_id': 4})

                       ]
        self.services = [models.ClusterServiceInfo.get_mock_object(
            overrides={'service': {'service_id': 3, 'status': enums.ServiceStatus.healthy}}),
            models.ClusterServiceInfo.get_mock_object(
                overrides={'service': {'service_id': 4, 'status': enums.ServiceStatus.healthy,
                                       'drive_id': 1}})]

    def test_good(self, mock_services, mock_timer, mock_s, mock_b):
        mock_s.return_value = 1
        mock_b.return_value = 1
        mock_services.return_value = self.services
        mock_timer.return_value = iter([True] * 10)
        vl_helpers.wait_for_drive_services_healthy(CLUSTER_FM, [])

    def test_no_drives(self, mock_services, mock_timer, mock_s, mock_b):
        mock_s.return_value = 1
        mock_b.return_value = 1
        mock_services.return_value = self.services
        mock_timer.return_value = iter([True, False])
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_services_healthy(CLUSTER_FM, self.drives)

    def test_unhealthy(self, mock_services, mock_timer, mock_s, mock_b):
        mock_s.return_value = 1
        mock_b.return_value = 1
        services = [models.ClusterServiceInfo.get_mock_object(
            overrides={'service': {'service_id': 3, 'status': enums.ServiceStatus.migrating}}),
            models.ClusterServiceInfo.get_mock_object(
                overrides={'service': {'service_id': 4, 'status': enums.ServiceStatus.healthy,
                                       'drive_id': 1}})]
        mock_services.return_value = services
        mock_timer.return_value = iter([True, False])
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_services_healthy(CLUSTER_FM, [])


@mock.patch.object(functional_models.Cluster, 'list_services')
@mock.patch.object(functional_models.Cluster, 'list_volumes')
class TestGetDrivesWoSlice(unittest.TestCase):
    def setUp(self):
        self.drives = [models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.volume,
                                                               'drive_id': 0})]

    def test_drv_wo_slice(self, mock_vols, mock_services):
        mock_vols.return_value = models.Volume.get_mock_object_list(3)
        mock_services.return_value = [models.ClusterServiceInfo.get_mock_object(overrides={
            'service': {'service_type': enums.ServiceType.slice, 'drive_id': 2}
        })]
        drvs_in_err = vl_helpers.get_unexpected_drives_without_slice_services(CLUSTER_FM, self.drives)
        assert len(drvs_in_err) == 1
        assert drvs_in_err[0].drive_id == 0


@mock.patch.object(functional_models.Cluster, 'list_services')
@mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
class TestGetBsWoBins(unittest.TestCase):
    def test_get_bs(self, mock_ba, mock_services):
        self.bins = models.BinAssignments.get_mock_object(overrides={
            'bins': {'services': {'service_id': 3}}
        })
        self.drives = [models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.block,
                                                               'drive_id': 3})]
        mock_services.return_value = [models.ClusterServiceInfo.get_mock_object(overrides={
            'service': {'service_type': enums.ServiceType.slice, 'drive_id': 2}
        })]
        mock_ba.return_value = self.bins
        bad_bs = vl_helpers.get_drives_without_bins(CLUSTER_FM, self.drives)
        assert bad_bs[0].drive_id == 3


@mock.patch.object(vl_helpers._drive, 'get_drives_without_bins')
@mock.patch.object(vl_helpers._drive, 'get_unexpected_drives_without_slice_services')
@mock.patch.object(vl_helpers._bin_sync, 'wait_for_bin_sync_to_finish_with_timeout')
@mock.patch.object(vl_helpers._slice, 'wait_for_slice_sync_to_finish')
@mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
@mock.patch.object(vl_helpers._slice, 'calculate_slice_sync_timeout')
@mock.patch.object(vl_helpers._drive, 'wait_for_drive_services_healthy')
class TestWaitForSyncIn(unittest.TestCase):
    def setUp(self):
        self.drives = [models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.block}),
                       models.Drive.get_mock_object(overrides={'serial': '1',
                                                               'type': enums.DriveType.volume,
                                                               'drive_id': 0}),
                       ]

    def test_wait_for_sync_in(self, mock_service, mock_slice_to, mock_bin_to, mock_slice_sync,
                              mock_bin_sync, mock_bad_slice, mock_bad_bs):
        mock_slice_to.return_value = 1
        mock_bin_to.return_value = 1
        mock_bad_slice.return_value = []
        mock_bad_bs.return_value = []
        vl_helpers.wait_for_sync_in(CLUSTER_FM, self.drives)

    def test_slice_bad(self, mock_service, mock_slice_to, mock_bin_to, mock_slice_sync,
                       mock_bin_sync, mock_bad_slice, mock_bad_bs):
        mock_slice_to.return_value = 1
        mock_bin_to.return_value = 1
        mock_bad_slice.return_value = [models.Drive.get_mock_object()]
        mock_bad_bs.return_value = []
        with pytest.raises(errors.ElementInvalidStateError):
            vl_helpers.wait_for_sync_in(CLUSTER_FM, self.drives)

    def test_block_bad(self, mock_service, mock_slice_to, mock_bin_to, mock_slice_sync,
                       mock_bin_sync, mock_bad_slice, mock_bad_bs):
        mock_slice_to.return_value = 1
        mock_bin_to.return_value = 1
        mock_bad_slice.return_value = []
        mock_bad_bs.return_value = [models.Drive.get_mock_object()]
        with pytest.raises(errors.ElementInvalidStateError):
            vl_helpers.wait_for_sync_in(CLUSTER_FM, self.drives)
