import unittest

import mock
import pytest

from sf_api import enums
from valence import (
    errors,
    functional_models,
    models,
    utils,
    vl_helpers,
)

# noinspection PyUnusedLocal

# Mock the cluster DM and FM once to avoid extra work
RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
CLUSTER_FM = functional_models.Cluster(RS_CLUSTER_DM, authenticate=False)


class TestGetNextEventHelper(unittest.TestCase):

    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_get_next_event_none(self, mock_list_events):
        mock_list_events.return_value = None

        retval = vl_helpers.get_next_event_id(CLUSTER_FM)
        assert retval == 0

    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_get_next_event_with_event(self, mock_list_events):
        mock_event = models.ClusterEvent.get_mock_object_list(1, overrides={
            'event_id': 22
        })
        mock_list_events.return_value = mock_event

        retval = vl_helpers.get_next_event_id(CLUSTER_FM)
        assert retval == mock_event[0].event_id + 1


class TestWaitEventHelper(unittest.TestCase):

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_wait_for_event_timer_expires(self, mock_list_events, mock_timer):
        mock_timer.return_value = iter([True, False])
        mock_list_events.return_value = None

        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_event(CLUSTER_FM, enums.ClusterEventType.bin_assignments_event,
                                      timeout=1, poll_period=0)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_wait_for_event_returns_events(self, mock_list_events, mock_timer):
        mock_timer.return_value = iter([True, True])
        events = models.ClusterEvent.get_mock_object_list(2, overrides={
            'event_info_type': enums.ClusterEventType.bin_assignments_event
        })
        mock_list_events.return_value = events

        event_list = vl_helpers.wait_for_event(CLUSTER_FM, enums.ClusterEventType.bin_assignments_event,
                                               timeout=543, poll_period=0, start_event_id=1)
        assert event_list == events
        mock_list_events.assert_called_with(
            start_event_id=1,
            start_publish_time=None,
            event_type=enums.ClusterEventType.bin_assignments_event
        )
        mock_timer.assert_called_with(543)

    @mock.patch.object(utils, 'get_now_utc')
    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_wait_for_event_returns_events_no_start_event(self, mock_list_events, mock_timer, mock_utc):
        current_utc = utils.get_now_utc()
        mock_utc.return_value = current_utc
        mock_timer.return_value = iter([True, True])
        events = models.ClusterEvent.get_mock_object_list(2, overrides={
            'event_info_type': enums.ClusterEventType.bin_assignments_event
        })
        mock_list_events.return_value = events

        event_list = vl_helpers.wait_for_event(CLUSTER_FM, enums.ClusterEventType.bin_assignments_event,
                                               timeout=543, poll_period=0)
        assert event_list == events
        mock_list_events.assert_called_with(
            event_type=enums.ClusterEventType.bin_assignments_event,
            start_event_id=None,
            start_publish_time=current_utc
        )
        mock_timer.assert_called_with(543)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_wait_for_event_with_match_message_no_match(self, mock_list_events, mock_timer):
        mock_timer.return_value = iter([True, False])
        events = models.ClusterEvent.get_mock_object_list(2, overrides={
            'event_info_type': enums.ClusterEventType.bin_assignments_event,
            'message': 'Not Going to Match'
        })
        mock_list_events.return_value = events

        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_event(CLUSTER_FM,
                                      enums.ClusterEventType.bin_assignments_event,
                                      message_regexes=['Please Match...'],
                                      timeout=544, poll_period=0, start_event_id=1)

        mock_timer.assert_called_with(544)

    @mock.patch.object(utils, 'timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'list_events')
    def test_wait_for_event_with_match_message_returns_events(self, mock_list_events, mock_timer):
        mock_timer.return_value = iter([True, True])
        events = models.ClusterEvent.get_mock_object_list(2, overrides={
            'event_info_type': enums.ClusterEventType.bin_assignments_event,
            'message': 'Match this...'
        })
        mock_list_events.return_value = events

        event_list = vl_helpers.wait_for_event(CLUSTER_FM, enums.ClusterEventType.bin_assignments_event,
                                               message_regexes=['Match this...'],
                                               timeout=545, poll_period=0, start_event_id=1)
        assert event_list == events
        mock_list_events.assert_called_with(
            start_event_id=1,
            start_publish_time=None,
            event_type=enums.ClusterEventType.bin_assignments_event
        )
        mock_timer.assert_called_with(545)
