import time
import unittest

import mock
import pytest
import requests

import sf_api
import sf_api.api_constants
from sf_api import enums
from valence import (
    criteria,
    errors,
    functional_models,
    models,
    utils,
    vl_helpers,
)
from valence.vl_helpers import _faults

# noinspection PyUnusedLocal

# Mock the cluster DM and FM once to avoid extra work
RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
CLUSTER_FM = functional_models.Cluster(RS_CLUSTER_DM, authenticate=False)


@mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
@mock.patch.object(time, 'sleep')
class TestWaitForClusterFaultOccurrence(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cf1 = models.ClusterFault.get_mock_object(
            overrides={'code': sf_api.ClusterFaultCode.cluster_full,
                       'severity': sf_api.ClusterFaultSeverity.warning})
        cls.cf2 = models.ClusterFault.get_mock_object(
            overrides={'code': sf_api.ClusterFaultCode.block_cluster_full,
                       'details': 'THEDONALD',
                       'severity': sf_api.ClusterFaultSeverity.critical})

    @mock.patch.object(criteria.BaseCriteria, 'filter')
    def test_fault_occurs(self, mock_filter, mock_sleep, mock_list_faults):
        mock_list_faults.return_value = [self.cf1, self.cf2]
        mock_filter.return_value = [self.cf1]
        mock_sleep.reset_mock()
        retval = _faults.wait_for_cluster_fault_occurrence(
            cluster_fm=CLUSTER_FM,
            fault_code=sf_api.ClusterFaultCode.cluster_full
        )
        assert self.cf1 in retval
        mock_list_faults.assert_called()
        mock_filter.assert_called()

    @mock.patch.object(criteria.BaseCriteria, 'filter')
    @mock.patch.object(utils, 'timeout_timer')
    def test_fault_doesnt_occur(self, mock_timer, mock_filter, mock_sleep, mock_list_faults):
        mock_sleep.reset_mock()
        mock_timer.return_value = iter([True, True, False])
        mock_filter.return_value = []
        mock_list_faults.return_value = [self.cf1, self.cf2]

        with pytest.raises(errors.TimeoutError):
            _faults.wait_for_cluster_fault_occurrence(
                cluster_fm=CLUSTER_FM,
                fault_code=sf_api.ClusterFaultCode.time_since_last_full_gc_exceeded_limit
            )

        mock_timer.assert_called()
        mock_list_faults.assert_called()
        mock_filter.assert_called()

    @mock.patch.object(criteria.BaseCriteria, 'filter')
    @mock.patch.object(utils, 'timeout_timer')
    def test_connection_error(self, mock_timer, mock_filter, mock_sleep, mock_list):
        mock_timer.return_value = iter([True, True, False])
        mock_list.side_effect = [requests.ConnectionError, [self.cf1, self.cf2]]
        mock_filter.return_value = [self.cf2]
        retval = _faults.wait_for_cluster_fault_occurrence(
            cluster_fm=CLUSTER_FM,
            fault_code=sf_api.ClusterFaultCode.block_cluster_full
        )

        assert self.cf2 in retval

    @mock.patch.object(criteria.BaseCriteria, 'filter')
    def test_fault_with_only_nodes(self, mock_filter, mock_sleep, mock_list_faults):
        mock_node = models.ActiveNode.get_mock_object()

        mock_list_faults.return_value = [self.cf1, self.cf2]
        mock_filter.return_value = [self.cf1]
        mock_sleep.reset_mock()
        retval = _faults.wait_for_cluster_fault_occurrence(
            cluster_fm=CLUSTER_FM,
            fault_code=sf_api.ClusterFaultCode.cluster_full,
            only_nodes=[mock_node]
        )
        assert self.cf1 in retval
        mock_list_faults.assert_called_with(
            fault_status=enums.ClusterFaultStatus.current,
            update=True,
            node_ids=[mock_node.node_id],
            include_severities=None
        )
        mock_filter.assert_called()

    @mock.patch.object(criteria.BaseCriteria, 'filter')
    def test_clusterseverity_with_only_nodes(self, mock_filter, mock_sleep, mock_list_faults):
        mock_node = models.ActiveNode.get_mock_object()

        mock_list_faults.return_value = [self.cf1, self.cf2]
        mock_filter.return_value = [self.cf1]
        mock_sleep.reset_mock()
        retval = _faults.wait_for_cluster_fault_occurrence(
            cluster_fm=CLUSTER_FM,
            fault_code=sf_api.ClusterFaultCode.cluster_full,
            only_nodes=[mock_node],
            include_severities=[enums.ClusterFaultSeverity.warning]
        )
        assert self.cf1 in retval
        mock_list_faults.assert_called_with(
            fault_status=enums.ClusterFaultStatus.current,
            update=True,
            node_ids=[mock_node.node_id],
            include_severities=[enums.ClusterFaultSeverity.warning]
        )
        mock_filter.assert_called()


@mock.patch.object(functional_models.Cluster, 'list_cluster_faults')
@mock.patch.object(time, 'sleep')
class TestWaitForClusterFaultResolution(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cf1 = models.ClusterFault.get_mock_object(
            overrides={'code': sf_api.ClusterFaultCode.cluster_full,
                       'severity': sf_api.ClusterFaultSeverity.warning})
        cls.cf2 = models.ClusterFault.get_mock_object(
            overrides={'code': sf_api.ClusterFaultCode.cluster_iops_are_over_provisioned,
                       'details': 'I have lots of CRC errors, cause I am best!',
                       'severity': sf_api.ClusterFaultSeverity.critical})
        cls.cf3 = models.ClusterFault.get_mock_object(
            overrides={'code': sf_api.ClusterFaultCode.block_cluster_full,
                       'details': 'THEDONALD',
                       'severity': sf_api.ClusterFaultSeverity.critical})

    @mock.patch.object(criteria.BaseCriteria, 'filter')
    def test_no_faults(self, mock_filter, mock_sleep, mock_list_faults):
        mock_list_faults.return_value = []
        _faults.wait_for_cluster_fault_resolution(CLUSTER_FM)
        assert not mock_filter.called

    @mock.patch.object(utils, 'timeout_timer')
    def test_raise(self, mock_timer, mock_sleep, mock_list):
        mock_timer.return_value = iter([False])
        with pytest.raises(vl_helpers.HelperError):
            _faults.wait_for_cluster_fault_resolution(CLUSTER_FM, timeout=1)
        mock_timer.assert_called_with(1)

    @mock.patch.object(utils, 'timeout_timer')
    def test_connection_error(self, mock_timer, mock_sleep, mock_list):
        mock_timer.return_value = iter([True, True, False])
        mock_list.side_effect = [requests.ConnectionError, []]
        mock_list.return_value = []
        _faults.wait_for_cluster_fault_resolution(CLUSTER_FM)

    @mock.patch.object(utils, 'timeout_timer')
    def test_code_regexes(self, mock_timer, mock_sleep, mock_list_faults):
        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf1, self.cf2, self.cf3], [self.cf1, self.cf2]]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            ignore_code_regexes=['^cluster', 'donald'],
            timeout=-1
        )

    @mock.patch.object(utils, 'timeout_timer')
    def test_details_regex(self, mock_timer, mock_sleep, mock_list_faults):
        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf2, self.cf3], [self.cf3]]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            ignore_details_regexes=['DONALD'],
            timeout=-1
        )

    @mock.patch.object(utils, 'timeout_timer')
    def test_severity(self, mock_timer, mock_sleep, mock_list_faults):
        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf1, self.cf2, self.cf3], [self.cf1]]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            ignore_severities=[self.cf1.severity],
            timeout=-1
        )

    @mock.patch.object(utils, 'timeout_timer')
    def test_crc_errors(self, mock_timer, mock_sleep, mock_list_faults):
        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf1, self.cf2, self.cf3], [self.cf2]]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            crc_error_threshold=1,
            timeout=-1
        )

    @mock.patch.object(utils, 'timeout_timer')
    def test_only_nodes(self, mock_timer, mock_sleep, mock_list_faults):
        node_dms = models.ActiveNode.get_mock_object_list(2)  # type: list[valence.models.ActiveNode]
        node_ids = [n.node_id for n in node_dms]
        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf1, self.cf2, self.cf3], []]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            timeout=-1,
            only_nodes=node_dms
        )

        mock_list_faults.assert_called_with(
            update=True,
            fault_status=enums.ClusterFaultStatus.current,
            node_ids=node_ids,
            include_codes=None,
            include_cluster_exceptions=False,
            exclude_codes=None
        )

    @mock.patch.object(utils, 'timeout_timer')
    def test_only_codes(self, mock_timer, mock_sleep, mock_list_faults):
        only_codes = [
            enums.ClusterFaultCode.node_offline,
            enums.ClusterFaultCode.cluster_full
        ]

        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf1, self.cf2, self.cf3], []]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            timeout=-1,
            only_fault_codes=only_codes)

        mock_list_faults.assert_called_with(
            update=True,
            fault_status=enums.ClusterFaultStatus.current,
            node_ids=None,
            include_codes=only_codes,
            include_cluster_exceptions=False,
            exclude_codes=None
        )

    @mock.patch.object(utils, 'timeout_timer')
    def test_ignore_fault_codes(self, mock_timer, mock_sleep, mock_list_faults):
        ignore_fault_codes = [
            enums.ClusterFaultCode.ensemble_degraded
        ]

        mock_timer.return_value = iter([True, True, False])
        mock_list_faults.side_effect = [[self.cf1, self.cf2, self.cf3], []]
        _faults.wait_for_cluster_fault_resolution(
            cluster_fm=CLUSTER_FM,
            timeout=-1,
            ignore_fault_codes=ignore_fault_codes)

        mock_list_faults.assert_called_with(
            update=True,
            fault_status=enums.ClusterFaultStatus.current,
            node_ids=None,
            include_codes=None,
            exclude_codes=ignore_fault_codes,
            include_cluster_exceptions=False
        )
