import os
import time
import unittest

import mock
import pytest
import requests
import sf_ssh

import valence
from valence import (
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _constants
from valence.vl_helpers import (
    _errors,
    _sfupdate_healthtools,
)
from valence.vl_helpers._healthtools_versions_for_upgrade_suites import HealthtoolsVersionsForTest

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


def mocked_requests_get_flat(*args, **kwargs):
    class MockResponse:
        def __init__(self):
            self.content = """'<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN"><html> <head>
            <title>Index of /upgradetools/rqd</title> </head> <body><h1>Index of /upgradetools/rqd</h1>
            <table><tr><th><img src="/icons/blank.gif" alt="[ICO]"></th><th><a href="?C=N;O=D">Name</a>
            </th><th><a href="?C=M;O=A">Last modified</a></th><th><a href="?C=S;O=A">Size</a></th><th>
            <a href="?C=D;O=A">Description</a></th></tr><tr><th colspan="5"><hr></th></tr><tr>
            <td valign="top"><img src="/icons/back.gif" alt="[DIR]"></td><td><a href="/upgradetools/">
            Parent Directory</a></td><td>&nbsp;</td><td align="right">  - </td><td>&nbsp;</td></tr><tr>
            <td valign="top"><img src="/icons/compressed.gif" alt="[   ]"></td><td>
            <a href="solidfire-healthtools-2019.06.01.102.tgz">solidfire-healthtools-2019.06.01.102.tgz
            </a></td><td align="right">17-May-2019 07:27  </td><td align="right">3.9M</td><td>&nbsp;
            </td></tr><tr><td valign="top"><img src="/icons/compressed.gif" alt="[   ]"></td><td>
            <a href="solidfire-healthtools-2019.06.01.104.tgz">solidfire-healthtools-2019.06.01.104.tgz
            </a></td><a href="10.2/">10.2/</a><td align="right">17-May-2019 07:18  </td>
            <td align="right">3.9M</td><td>&nbsp;</td></tr><tr><th colspan="5"><hr></th></tr></table>
            <address>Apache/2.2.22 (Ubuntu) Server at bdr-jenkins.eng.solidfire.net Port 80</address>
            </body></html>'"""

    return MockResponse()


class TestHealthtoolsHelper(unittest.TestCase):
    """Tests the helper metnods for healthtools helper."""

    @classmethod
    def setUpClass(cls):
        cls.mnode_fm = functional_models.ManagementNodeLegacy.get_mock_fm()

        cls.stable = [(models.GenericVersion('2019.01.01.01'),
                       'http://stable/solidfire-healthtools-2019.01.01.01.tgz'),
                      (models.GenericVersion('2019.01.01.02'),
                       'http://stable/solidfire-healthtools-2019.01.01.02.tgz'),
                      (models.GenericVersion('2019.01.01.03'),
                       'http://stable/solidfire-healthtools-2019.01.01.03.tgz'),
                      (models.GenericVersion('2018.05.01.500'),
                       'http://stable/solidfire-healthtools-2018.05.01.500.tgz'),
                      (models.GenericVersion('2018.05.01.600'),
                       'http://stable/solidfire-healthtools-2018.05.01.600.tgz'),
                      (models.GenericVersion('2018.09.01.131'),
                       'http://stable/solidfire-healthtools-2018.09.01.131.tgz')
                      ]
        cls.rc = [(models.GenericVersion('2019.01.01.04'),
                   'http://rc/solidfire-healthtools-2019.01.01.04.tgz'),
                  (models.GenericVersion('2019.01.01.05'),
                   'http://rc/solidfire-healthtools-2019.01.01.05.tgz'),
                  (models.GenericVersion('2019.01.01.06'),
                   'http://rc/solidfire-healthtools-2019.01.01.06.tgz')]
        cls.rqd = [(models.GenericVersion('2019.01.01.07'),
                    'http://rqd/solidfire-healthtools-2019.01.01.07.tgz')]

    def test_get_version_from_url(self):
        test_url = 'http://blah/solidfire-healthtools-2019.06.01.108.tgz'
        version, url = HealthtoolsVersionsForTest.get_version_from_url(test_url)
        assert version == models.GenericVersion('2019.06.01.108')
        assert url == test_url

        test_bad_version_url = 'http://blah/solidfire-healthtools-1.2.3.4.tgz'
        version, url = HealthtoolsVersionsForTest.get_version_from_url(test_bad_version_url)
        assert not version
        assert url == test_bad_version_url

        bad_url = 'http://blah/yada/never_gonna_find_me.tgz'
        version, url = HealthtoolsVersionsForTest.get_version_from_url(bad_url)
        assert not version
        assert bad_url == url

    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, 'get_healthtools_at_url')
    def test_initialize_jenkins_versions(self, mock_get_healthtools_at_url):
        mock_get_healthtools_at_url.side_effect = [self.stable, self.rc, self.rqd]

        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        ht._initialize_jenkins_versions()

        assert not ht.all_versions_for_test
        assert not ht.healthtools_latest_version_public
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        assert sorted(ht.all_versions_on_jenkins) == sorted(versions)
        assert sorted(ht.jenkins_rqd_versions) == sorted([n[0] for n in self.rqd])
        assert sorted(ht.jenkins_rc_versions) == sorted([n[0] for n in self.rc])

    @mock.patch.object(requests, 'get')
    def test_initialize_versions_original_on_jenkins(self, mock_requests_get):
        mock_requests_get.return_value.json.return_value = {'latest_version': '2019.09.11.23'}

        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        ht._initialize_nss_version()
        assert not ht.all_versions_for_test
        assert not ht.all_versions_on_jenkins
        assert ht.healthtools_latest_version_public == models.GenericVersion('2019.09.11.23')

    @mock.patch.object(valence.functional_models.ManagementNodeLegacy, 'get_current_healthtools_version')
    def test_initialize_original_version_not_on_jenkins_error(self, mock_current):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        mock_current.return_value = models.GenericVersion('1.2.3.4')
        with pytest.raises(_errors.HealthtoolsVersionNotFound):
            ht._initialize_original_version(error_if_not_on_jenkins=True)

    @mock.patch.object(valence.functional_models.ManagementNodeLegacy, 'get_current_healthtools_version')
    def test_initialize_original_version(self, mock_current):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        mock_current.return_value = models.GenericVersion('1.2.3.4')
        ht._initialize_original_version(error_if_not_on_jenkins=False)
        assert ht.original_version == models.GenericVersion('1.2.3.4')

    def test_initialize_custom_versions(self):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        custom_urls = ['http://meh/solidfire-healthtools-2015.01.02.03.tgz',
                       'http://meh/solidfire-healthtools-2015.01.02.04.tgz',
                       'http://doh/solidfire-healthtools-2015.01.02.05.tgz']
        ht._initialize_test_versions(custom_urls)
        assert models.GenericVersion('2015.01.02.03') in ht.all_versions_for_test
        assert models.GenericVersion('2015.01.02.04') in ht.all_versions_for_test
        assert models.GenericVersion('2015.01.02.05') in ht.all_versions_for_test
        assert 'http://meh/solidfire-healthtools-2015.01.02.03.tgz' in ht._healthtool_urls.values()
        assert not ht.jenkins_rqd_versions
        assert not ht.jenkins_rc_versions
        assert not ht.all_versions_on_jenkins

    def test_initialize_test_versions_no_custom(self):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        ht._all_versions_for_test = versions
        ht._initialize_test_versions(paths_to_healthtools_from_suite_config=None)
        assert ht.all_versions_for_test == ht.all_versions_on_jenkins

    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    def test_filter_sodium(self, mock_mnode_version):
        mock_mnode_version.return_value = _constants.SODIUM

        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        ht._all_versions_for_test = versions
        ht._filter_test_versions()

        # Versiosn we hate still get filtered out, and 2018.05.01.600 is filtered b/c it's Neon only.
        assert models.GenericVersion('2018.05.01.500') not in ht.all_versions_for_test
        assert models.GenericVersion('2018.09.01.131') not in ht.all_versions_for_test
        assert models.GenericVersion('2018.05.01.600') not in ht.all_versions_for_test
        rc_set = set([n[0] for n in self.rc])
        all_set = set(ht.all_versions_for_test)
        assert rc_set.issubset(all_set)

    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    def test_filter_neon(self, mock_mnode_version):
        mock_mnode_version.return_value = _constants.NEON_PATCH3

        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        ht._all_versions_for_test = versions
        ht._filter_test_versions()

        # Versiosn we hate still get filtered out
        assert models.GenericVersion('2018.05.01.500') not in ht.all_versions_for_test
        assert models.GenericVersion('2018.09.01.131') not in ht.all_versions_for_test

        # 2018.05.01.600 is neon only but should not get filtered out.
        assert models.GenericVersion('2018.05.01.600') in ht.all_versions_for_test
        rc_set = set([n[0] for n in self.rc])
        all_set = set(ht.all_versions_for_test)
        assert rc_set.issubset(all_set)

    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, '_initialize_jenkins_versions')
    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, '_initialize_nss_version')
    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, '_initialize_original_version')
    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, '_initialize_test_versions')
    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, '_filter_test_versions')
    def test_initialize(self, mock_filter, mock_test, mock_orig, mock_nss, mock_jen):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        ht._all_versions_for_test = versions
        ht.initialize_versions(None)
        assert mock_filter.called
        assert mock_test.called
        assert mock_orig.called
        assert mock_nss.called
        assert mock_jen.called
        assert ht.all_versions_for_test == versions

    def test_set_versions(self):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        version_under_test = models.GenericVersion('2020.01.02.03')
        lower_version = models.GenericVersion('2019.01.01.01')
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        ht._all_versions_for_test = versions
        ht.set_healthtool_versions_for_test_suite(
            version_under_test,
            previous_version_of_healthtools=lower_version)
        assert ht.version_under_test == version_under_test
        assert ht.lower_version == lower_version
        assert ht.latest_version == version_under_test

    def test_set_versions_no_lower(self):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        versions = [n[0] for n in (self.stable + self.rc + self.rqd)]
        ht._all_versions_for_test = versions

        version_under_test = models.GenericVersion('2019.01.01.04')
        ht.set_healthtool_versions_for_test_suite(version_under_test)
        assert ht.version_under_test == version_under_test
        assert ht.lower_version in ht.all_versions_for_test
        assert ht.lower_version != ht.original_version
        assert ht.lower_version < ht.version_under_test

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_download_to_mnode_with_url(self, mock_run_command):
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='mkdir worked', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='wget worked', stderr='', return_code=0)]

        ht = HealthtoolsVersionsForTest(self.mnode_fm)

        # This version has a URL.
        version_to_download = models.GenericVersion('2019.01.01.01')
        ht._healthtool_urls[version_to_download] = 'http://y.com'
        ht.download_healthtools_package_to_mnode(version_to_download)
        expected_url = 'wget -O /tmp/ura-packages/solidfire-healthtools-2019.01.01.01.tgz http://y.com'
        assert mock_run_command.call_count == 2
        assert mock_run_command.call_args_list[0].args[0] == 'mkdir -p /tmp/ura-packages'
        assert mock_run_command.call_args_list[1].args[0] == expected_url

        # Version that has no URL.
        version_to_download = models.GenericVersion('2017.01.01.01')
        with pytest.raises(_errors.HealthtoolsVersionNotFound):
            ht.download_healthtools_package_to_mnode(version_to_download)

        # Good version to download, in the list of known versions and we pass in a URL.
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='mkdir worked', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='wget worked', stderr='', return_code=0)]

        version_to_download = models.GenericVersion('2017.01.01.01')
        ht.download_healthtools_package_to_mnode(version_to_download, 'http://hey.url')
        expected_url = 'wget -O /tmp/ura-packages/solidfire-healthtools-2017.01.01.01.tgz http://hey.url'

        assert mock_run_command.call_args_list[-1].args[0] == expected_url

    @mock.patch('requests.get', side_effect=mocked_requests_get_flat)
    def test_get_healthtools_at_url(self, mock_get):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        healthtools = ht.get_healthtools_at_url('http://blah.yada')
        assert healthtools == [
            (models.GenericVersion('2019.06.01.102'),
             'http://blah.yada/solidfire-healthtools-2019.06.01.102.tgz'),
            (models.GenericVersion('2019.06.01.104'),
             'http://blah.yada/solidfire-healthtools-2019.06.01.104.tgz')]

    @mock.patch.object(valence.functional_models.ManagementNodeLegacy, 'get_current_healthtools_version')
    def test_current_version(self, mock_get_current):
        ht = HealthtoolsVersionsForTest(self.mnode_fm)
        mock_get_current.return_value = models.GenericVersion('1.2.3.5')
        assert ht.current_version == models.GenericVersion('1.2.3.5')
