import time
import unittest

import mock
import pytest
import requests
import sf_ssh

from valence import (
    functional_models,
    models,
    unittest_utils,
    vl_helpers,
)
from valence.functional_models import _constants
from valence.vl_helpers import (
    _errors,
    _mnode,
)

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestMnodeHelpers(unittest.TestCase):
    """Tests the helper metnods for mnode fm."""

    @classmethod
    def setUpClass(cls):
        res_mnode = models.ResourceManagementNode().get_mock_object(include_optional=True)
        build_dm = models.Build.create(build_id=1, release='sodium', element_id=11,
                                       version=_constants.SODIUM_PATCH3.version_string,
                                       repository_name='sodium-patch3')
        res_mnode['build_version'] = build_dm
        res_mnode['username'] = 'Valentino'
        res_mnode['password'] = 'Rossi'
        cls.mnode_fm = functional_models.ManagementNode2x.from_resource_data(res_mnode)

        cls.res_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.res_cluster_dm['username'] = 'Marc'
        cls.res_cluster_dm['password'] = 'Marquez'
        cls.res_cluster_dm['mvip'] = '127.0.0.1'
        cls.res_cluster_dm['svip'] = '127.0.0.2'
        cls.cluster_fm = functional_models.Cluster.from_resource_data(cls.res_cluster_dm)

    def test_mnode_setup_afa_cmd(self):
        cmd = _mnode._build_cmd(self.cluster_fm.cluster_dm, self.mnode_fm.username,
                                self.mnode_fm.password)
        assert '/sf/packages/mnode/setup-mnode --mnode_admin_user Valentino --storage_username Marc ' \
               '--storage_mvip 127.0.0.1 --storage_svip 127.0.0.2 --mnode_admin_password Rossi ' \
               '--storage_password Marquez --use_persistent_volumes false --testing true' == cmd

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_mnode_setup_afa_ssh_use_default_credentials(self, mock_version, mock_run_command):
        mock_version.return_value = models.SfVersion('11.5.0.9000')
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='docker service ls', stderr='', return_code=1),
            sf_ssh.SshError(),
            sf_ssh.ssh.RunCommandOutput(stdout='Pass', stderr='', return_code=0)]
        _mnode.configure_mnode2_0_for_afa(self.mnode_fm, self.cluster_fm.cluster_dm)
        assert mock_run_command.called
        assert mock_run_command.call_count == 3
        # Verify that the credentials in the commands change between 2 calls.
        assert mock_run_command.call_args_list[0].args[0] == 'docker service ls'
        assert mock_run_command.call_args_list[1].args[0] == \
            '/sf/packages/mnode/setup-mnode --mnode_admin_user Valentino --storage_username Marc ' \
            '--storage_mvip 127.0.0.1 --storage_svip 127.0.0.2 --mnode_admin_password Rossi ' \
            '--storage_password Marquez --use_persistent_volumes false --testing true'
        assert mock_run_command.call_args_list[2].args[0] == \
            '/sf/packages/mnode/setup-mnode --mnode_admin_user admin --storage_username Marc ' \
            '--storage_mvip 127.0.0.1 --storage_svip 127.0.0.2 --mnode_admin_password admin ' \
            '--storage_password Marquez --use_persistent_volumes false --testing true'

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_mnode_setup_afa_ssh_use_at2_credentials(self, mock_version, mock_run_command):
        mock_version.return_value = models.SfVersion('11.5.0.9000')
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='docker service ls', stderr='', return_code=1),
            sf_ssh.ssh.RunCommandOutput(stdout='Pass', stderr='Hell', return_code=0)]
        _mnode.configure_mnode2_0_for_afa(self.mnode_fm, self.cluster_fm.cluster_dm)
        assert mock_run_command.called

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_mnode_already_setup_fail_if_configured(self, mock_version, mock_run_command):
        mock_version.return_value = models.SfVersion('11.5.0.9000')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        with pytest.raises(_errors.ManagementNodeServicesAlreadyConfigured):
            _mnode.configure_mnode2_0_for_afa(self.mnode_fm, self.cluster_fm.cluster_dm,
                                              fail_if_already_configured=True)
        assert mock_run_command.called

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_mnode_already_setup_not_implemented_cleanup(self, mock_version, mock_run_command):
        mock_version.return_value = models.SfVersion('11.5.0.9000')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        with pytest.raises(NotImplementedError):
            _mnode.configure_mnode2_0_for_afa(self.mnode_fm, self.cluster_fm.cluster_dm,
                                              cleanup_if_already_configured=True)
        assert mock_run_command.called

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_mnode_setup_afa__nap1_no_run(self, mock_version, mock_run_command):
        mock_version.return_value = models.SfVersion('11.5.0.9000')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='',
                                                                    stderr='',
                                                                    return_code=0)

        res_mnode = models.ResourceManagementNode().get_mock_object(include_optional=True)
        build_dm = models.Build.create(build_id=1, release='sodium', element_id=11, version='11.1.0.0',
                                       repository_name='sodium-patch1')
        res_mnode['build_version'] = build_dm
        mnode_fm_nap1 = functional_models.ManagementNode2x.from_resource_data(res_mnode)

        _mnode.configure_mnode2_0_for_afa(mnode_fm_nap1, self.cluster_fm.cluster_dm)
        assert not mock_run_command.called

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_mnode_setup_afa_cluster_nap1_no_run(self, mock_version, mock_run_command):
        mock_version.return_value = models.SfVersion('11.1.0.1')
        mock_run_command.return_value = sf_ssh.ssh.RunCommandOutput(stdout='',
                                                                    stderr='',
                                                                    return_code=0)

        _mnode.configure_mnode2_0_for_afa(self.mnode_fm, self.cluster_fm.cluster_dm)
        assert not mock_run_command.called

    @mock.patch.object(functional_models.ManagementNode2x,
                       'get_storage_asset_id_for_storage_cluster_mvip')
    @mock.patch.object(functional_models.ManagementNode2x, 'add_asset_if_needed')
    @mock.patch.object(functional_models.ManagementNode2x, 'add_storage_id_to_asset')
    def test_mnode_setup_storage_exists(self, mock_add_storage, mock_add_asset, mock_get):
        mnode_fm_nap3 = functional_models.ManagementNode2x.get_mock_fm()
        mock_get.return_value = 'blah'
        vl_helpers.setup_mnode_storage_service_for_upgrades(mnode_fm_nap3, self.res_cluster_dm)
        assert not mock_add_asset.called
        assert not mock_add_storage.called
        assert mock_get.called

    @mock.patch.object(functional_models.ManagementNode2x,
                       'get_storage_asset_id_for_storage_cluster_mvip')
    @mock.patch.object(functional_models.ManagementNode2x, 'add_asset_if_needed')
    @mock.patch.object(functional_models.ManagementNode2x, 'add_storage_id_to_asset')
    def test_mnode_setup_full_flow(self, mock_add_storage, mock_add_asset, mock_get):
        mnode_fm_nap3 = functional_models.ManagementNode2x.get_mock_fm()
        mock_get.return_value = None
        mock_add_asset.return_value = 'asset_uuid'

        vl_helpers.setup_mnode_storage_service_for_upgrades(mnode_fm_nap3, self.res_cluster_dm,
                                                            'my_asset')
        assert mock_add_asset.called
        assert mock_add_storage.called
        assert mock_get.called

        assert mock_add_asset.call_args[1]['asset_name'] == 'my_asset'
        assert mock_add_storage.call_args[0][0] == 'asset_uuid'

    @mock.patch.object(requests, 'get')
    def test_get_published_mnode_services_bundle_url(self, mock_get):
        mock_response = mock.Mock(requests.Response)
        mock_response.json.return_value = {
            'return_code': 200,
            'description': "<a href=\"http://sf-artifactory.eng.solidfire.net/artifactory/"
                           "generic-binary-local/mnode2_2.x.194.tar.gz\">http://sf-artifactory."
                           'eng.solidfire.net/artifactory/generic-binary-local/mnode2_2.x.194.tar.gz'
                           '</a>'
        }
        mock_get.return_value = mock_response
        url = _mnode.get_published_mnode_services_bundle_url(jenkins_url=_constants.MNODE_MAIN_URL)
        assert url == 'http://sf-artifactory.eng.solidfire.net/artifactory/generic-binary-local/' \
                      'mnode2_2.x.194.tar.gz'
        mock_get.side_effect = requests.ConnectionError
        with pytest.raises(_errors.HelperError):
            url = _mnode.get_published_mnode_services_bundle_url(jenkins_url=_constants.MNODE_MAIN_URL)
            assert url == ''

    @mock.patch.object(requests, 'get')
    def test_get_service_bundle_data_from_url(self, mock_get):
        mock_response = mock.Mock(requests.Response)
        mock_response.content = 'foo'
        mock_get.return_value = mock_response
        bundle_data = _mnode.get_service_bundle_data_from_url('http://localhost/bar')
        assert bundle_data == {'fat_tar': 'foo'}
        mock_get.side_effect = requests.ConnectTimeout
        with pytest.raises(_errors.HelperError):
            _mnode.get_service_bundle_data_from_url('http://localhost/baz')
