"""Unit tests for Node helper."""
import random
import time
import unittest

import mock
import pytest
import sf_batch
import sf_mc
import sf_models
import sf_ssh

import sf_api
from sf_api import (
    api_constants,
    result_models,
)
from sf_api.models import types
from valence import (
    enums,
    errors,
    functional_models,
    models,
    unittest_utils,
    utils,
    vl_helpers,
)
from valence.vl_helpers import _errors

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # TODO: figure out a way to make get_version dynamic for uts
    module_mock.mock(sf_api.SfAuth, 'get_version', mock.MagicMock(return_value='5.0'))


def tearDownModule():
    module_mock.restore()


# Mock the node FM once to avoid extra work
NODE_FM = functional_models.Node.get_mock_fm()


@mock.patch.object(functional_models.Node, 'set_network_interface_status_and_verify', autospec=True)
class TestNetworkInterfaceStatusHelpers(unittest.TestCase):
    """Test node helpers related to setting the network interface status."""

    def test_down_network_interfaces(self, mock_set_network_interface):
        nodes = [NODE_FM]
        interfaces = [enums.NetworkInterface.eth0]
        vl_helpers.down_network_interfaces(nodes, interfaces)
        mock_set_network_interface.assert_called_with(NODE_FM, interfaces,
                                                      status=enums.NetworkStatus.down)

    def test_up_network_interfaces(self, mock_set_network_interface):
        nodes = [NODE_FM]
        interfaces = [enums.NetworkInterface.eth0]
        vl_helpers.up_network_interfaces(nodes, interfaces)
        mock_set_network_interface.assert_called_with(NODE_FM, interfaces,
                                                      status=enums.NetworkStatus.up_and_running)


@mock.patch.object(functional_models.Node, 'core_dump_check', autospec=True)
@mock.patch.object(functional_models.Node, 'crash_dump_check', autospec=True)
class TestCheckForCores(unittest.TestCase):

    def test_check_for_cores_none_found(self, mock_crash_check, mock_core_check):
        vl_helpers.check_for_core_and_crash_dumps([NODE_FM])

    def test_check_for_cores_error(self, mock_crash_check, mock_core_check):
        mock_core_check.side_effect = functional_models.FunctionalNodeError
        with pytest.raises(BaseException):
            vl_helpers.check_for_core_and_crash_dumps([NODE_FM])


@mock.patch.object(functional_models.Node, 'api_is_up', autospec=True)
class TestApiUp(unittest.TestCase):

    def test_api_up(self, mock_ping):
        vl_helpers.api_up_on_nodes([NODE_FM])
        mock_ping.assert_called_once_with(NODE_FM)


@mock.patch.object(functional_models.Node, 'get_ensemble_config', autospec=True)
class TestEnsemble(unittest.TestCase):

    def test_not_ensemble(self, mock_get):
        mock_ens = sf_api.result_models.GetEnsembleConfig.get_mock_object(
            overrides={'server_stats': None})
        mock_get.return_value = mock_ens
        t = vl_helpers.get_ensemble_config_on_nodes([NODE_FM])
        mock_get.assert_called_once_with(NODE_FM)
        assert not t[0]

    def test_ensemble(self, mock_get):
        mock_ens = sf_api.result_models.GetEnsembleConfig.get_mock_object(
            overrides={'server_stats': {'mode': enums.ZookeeperType.leader}}
        )
        mock_get.return_value = mock_ens
        t = vl_helpers.get_ensemble_config_on_nodes([NODE_FM])
        mock_get.assert_called_once_with(NODE_FM)
        assert isinstance(t[0], sf_api.enums.ZookeeperType)


@mock.patch('valence.functional_models.Node.set_network_config')
@mock.patch('valence.functional_models.Node.get_network_config')
class TestNodeNetworkConfigHelper(unittest.TestCase):
    """Test Node Helper Network Config implementation."""

    @classmethod
    def setUpClass(cls):
        cls.node_fms = [NODE_FM]
        b1g = models.NetworkInterfaceDetails.get_mock_object()
        b10g = models.NetworkInterfaceDetails.get_mock_object()
        cls.network_config_dm = models.NetworkConfig.get_mock_object(overrides=dict(
            bond_1g=b1g, bond_10g=b10g))
        network_config_no_routes_data = cls.network_config_dm.serialize()
        network_config_no_routes_data[api_constants.BOND_10G][api_constants.ROUTES] = []
        network_config_no_routes_data[api_constants.BOND_1G][api_constants.ROUTES] = []
        cls.network_config_no_routes_dm = models.NetworkConfig(network_config_no_routes_data)
        cls.route_dm = models.Route.get_mock_object()
        cls.route_dm_two = models.Route.get_mock_object()
        cls.bond_1g = sf_api.NetworkInterface.bond_1g
        cls.bond_10g = sf_api.NetworkInterface.bond_10g

    def test_add_routes(self, mock_get, mock_set):
        route_dms = [self.route_dm]
        mock_set.return_value = self.network_config_dm
        vl_helpers.add_routes_to_nodes(node_fms=self.node_fms,
                                       route_dms=route_dms, interface=self.bond_10g)
        assert mock_set.called
        mock_set.reset_mock()
        mock_set.return_value = self.network_config_dm
        vl_helpers.add_routes_to_nodes(node_fms=self.node_fms,
                                       route_dms=route_dms, interface=self.bond_1g)
        assert mock_set.called

    def test_add_existing_route(self, mock_get, mock_set):
        """An exception is raised when attempting to add a route that already exists."""
        route_dms = [self.route_dm]
        self.network_config_dm.bond_10g.routes = route_dms
        mock_set.return_value = self.network_config_dm
        mock_get.return_value = self.network_config_dm
        with pytest.raises(ValueError):
            vl_helpers.add_routes_to_nodes(node_fms=self.node_fms, route_dms=route_dms,
                                           interface=self.bond_10g)
        self.network_config_dm.bond_1g.routes = route_dms
        mock_set.return_value = self.network_config_dm
        mock_get.return_value = self.network_config_dm
        with pytest.raises(ValueError):
            vl_helpers.add_routes_to_nodes(node_fms=self.node_fms, route_dms=route_dms,
                                           interface=self.bond_1g)

    def test_remove_routes(self, mock_get, mock_set):
        route_dms = [self.route_dm, self.route_dm_two]
        self.network_config_dm.bond_10g.routes = route_dms
        self.network_config_dm.bond_1g.routes = route_dms
        mock_get.side_effect = [self.network_config_dm, self.network_config_no_routes_dm]
        mock_set.return_value = self.network_config_dm
        vl_helpers.remove_routes_from_nodes(node_fms=self.node_fms, route_dms=route_dms,
                                            interface=self.bond_10g)
        mock_get.side_effect = [self.network_config_dm, self.network_config_no_routes_dm]
        vl_helpers.remove_routes_from_nodes(node_fms=self.node_fms, route_dms=route_dms,
                                            interface=self.bond_1g)

    def test_remove_not_existing_route(self, mock_get, mock_set):
        route_dms = [self.route_dm, self.route_dm_two]
        mock_get.return_value = self.network_config_no_routes_dm
        with pytest.raises(ValueError):
            vl_helpers.remove_routes_from_nodes(node_fms=self.node_fms, route_dms=route_dms,
                                                interface=self.bond_10g)
        with pytest.raises(ValueError):
            vl_helpers.remove_routes_from_nodes(node_fms=self.node_fms, route_dms=route_dms,
                                                interface=self.bond_1g)

    def test_add_routes_to_nodes_bad_interface(self, mock_set, mock_get):
        with pytest.raises(ValueError):
            vl_helpers.add_routes_to_nodes(node_fms=[], route_dms=[], interface='BAD')


class TestNodePowerHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fms = [functional_models.Node.get_mock_fm(),
                        functional_models.Node.get_mock_fm()]

    @mock.patch('sf_batch.SFProcess')
    def test_reboot_nodes(self, mock_process):
        """Test reboot_nodes_via_oob with node_fms"""
        wait_for_ping = random.choice([True, False])
        vl_helpers.reboot_nodes_via_oob(self.node_fms, wait_for_ping=wait_for_ping)

        assert mock_process.call_count == 2

    @mock.patch('sf_batch.SFProcess')
    def test_power_off_nodes(self, mock_process):
        """Test power_off_nodes_via_oob with node_fms."""
        wait_for_power_down = random.choice([True, False])
        vl_helpers.power_off_nodes_via_oob(self.node_fms[:1], wait_for_power_down=wait_for_power_down)
        mock_process.assert_called_once()

    @mock.patch('sf_batch.SFProcess')
    def test_power_on_nodes(self, mock_process):
        """Test power_on_nodes_via_oob with node_fms."""
        wait_for_power_up = random.choice([True, False])
        wait_for_ping = random.choice([True, False])
        wait_for_sfapp = random.choice([True, False])
        vl_helpers.power_on_nodes_via_oob(self.node_fms, wait_for_power_up=wait_for_power_up,
                                          wait_for_ping=wait_for_ping, wait_for_sfapp=wait_for_sfapp)

        assert mock_process.call_count == 2


@unittest_utils.patch_sleep()
class TestCreateCluster(unittest.TestCase):

    def setUp(self):
        self.node_fm = functional_models.Node.get_mock_fm()
        self.mvip = types.IPAddressType()._mock()
        self.svip = types.IPAddressType()._mock()
        self.nodes = [models.ResourceNode.get_mock_object() for _ in range(4)]
        x = models.NetworkProfile.get_mock_object()  # Need short var name for line length
        self.network_profile = models.NetworkConfig(dict(Bond1G=dict(address='192.168.0.1',
                                                                     gateway='192.168.0.1')))
        # force first node to not be witness node or it will be the target node and the test will fail
        self.nodes[0].model = sf_models.NodeType(sf_models.NodeType.STANDARD_TYPES.H410S_11110_NE)
        self.nodes[0].network_profile = x
        self.cluster_state = result_models.GetClusterState.get_mock_object()
        self.api = {api_constants.CURRENT_VERSION: '11.0'}
        self.assignments_none = models.BinAssignments.create(bins=[], master_version=0)
        self.assignments_one = models.BinAssignments.create(
                    bins=models.BinReport.get_mock_object_list(1),
                    master_version=1)

    @mock.patch('sf_ssh.Ssh')
    @mock.patch('valence.vl_helpers._node.wait_for_bootstrap')
    @mock.patch.object(functional_models.Node, 'get_network_config')
    @mock.patch.object(functional_models.Node, 'set_config')
    @mock.patch.object(functional_models.Node, 'set_network_config')
    @mock.patch.object(functional_models.Node, 'set_cluster_config')
    @mock.patch.object(functional_models.Node, 'create_cluster')
    @mock.patch('valence.functional_models._cluster.base.ClusterBase.authenticate')
    @mock.patch('requests.Session.post')
    @mock.patch.object(functional_models.Node, 'get_hardware_info')
    @mock.patch.object(vl_helpers._drive, 'wait_for_drive_status')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers._drive, 'add_drives')
    @mock.patch.object(functional_models.Node, 'get_api')
    @mock.patch.object(functional_models.Node, 'get_cluster_state')
    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    def test_create_cluster_ipv4(self, mock_get_bin, mock_state, mock_api, mock_add, mock_list,
                                 mock_wait, mock_get_hardware, mock_version, mock_auth, mock_create,
                                 mock_set_config, mock_snc, mock_sc, mock_get_config,
                                 mock_bootstrap, mock_ssh):
        mock_bootstrap.return_value = {'nodes': [{'mip': node.sip, 'cip': node.mip} for
                                                 node in self.nodes]}
        mock_state.return_value = self.cluster_state
        mock_api.return_value = self.api
        mock_version.return_value = types.SfApiVersionType()._mock()
        mock_drive_hardware = [models.DriveInfo.create(drive_type=enums.DriveType.slice)
                               for _ in range(2)]
        mock_get_hardware.return_value = models.NodeHardwareInfo.get_mock_object(
             overrides=dict(drive_hardware=mock_drive_hardware))
        mock_get_bin.side_effect = [self.assignments_none, self.assignments_one]
        cluster_fm = vl_helpers.create_cluster(self.node_fm, self.mvip, self.svip, self.nodes,
                                               add_drives=True)
        cluster_fm.cluster_dm.validate()
        assert cluster_fm.mvip == self.mvip
        mock_wait.assert_called()
        mock_get_bin.return_value = models.BinAssignments.master_version
        mock_get_bin.assert_called_with(validate=False)

    @mock.patch('sf_ssh.Ssh')
    @mock.patch('valence.vl_helpers._node.wait_for_bootstrap')
    @mock.patch.object(functional_models.Node, 'get_network_config')
    @mock.patch.object(functional_models.Node, 'set_config')
    @mock.patch.object(functional_models.Node, 'set_network_config')
    @mock.patch.object(functional_models.Node, 'set_cluster_config')
    @mock.patch.object(functional_models.Node, 'create_cluster')
    @mock.patch('valence.functional_models._cluster.base.ClusterBase.authenticate')
    @mock.patch('requests.Session.post')
    @mock.patch.object(functional_models.Node, 'get_hardware_info')
    @mock.patch.object(vl_helpers._drive, 'wait_for_drive_status')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(vl_helpers._drive, 'add_drives')
    @mock.patch.object(functional_models.Node, 'get_api')
    @mock.patch.object(functional_models.Node, 'get_cluster_state')
    @mock.patch.object(functional_models.Cluster, 'get_bin_assignments')
    @mock.patch.object(utils, 'timeout_timer')
    def test_create_cluster_bin_assign_timeout(self, mock_timer, mock_get_bin, mock_state, mock_api,
                                               mock_add, mock_list, mock_wait, mock_get_hardware,
                                               mock_version, mock_auth, mock_create, mock_set_config,
                                               mock_snc, mock_sc, mock_get_config, mock_bootstrap,
                                               mock_ssh):
        mock_bootstrap.return_value = {'nodes': [{'mip': node.sip, 'cip': node.mip} for
                                                 node in self.nodes]}
        mock_state.return_value = self.cluster_state
        mock_api.return_value = self.api
        mock_version.return_value = types.SfApiVersionType()._mock()
        mock_drive_hardware = [models.DriveInfo.create(drive_type=enums.DriveType.slice)
                               for _ in range(2)]
        mock_get_hardware.return_value = models.NodeHardwareInfo.get_mock_object(
             overrides=dict(drive_hardware=mock_drive_hardware))
        mock_get_bin.return_value = self.assignments_none.master_version
        mock_timer.return_value = iter([False])
        with pytest.raises(errors.TimeoutError):
            cluster_fm = vl_helpers.create_cluster(self.node_fm, self.mvip, self.svip, self.nodes,
                                                   add_drives=True)


class TestWaitForBootstrap(unittest.TestCase):
    def setUp(self):
        self.nodes = [models.ResourceNode.get_mock_object() for _ in range(4)]
        self.node_fm = functional_models.Node(self.nodes[0], None)

    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.Node, 'get_bootstrap_config')
    def test_bootstrap_valid(self, mock_bootstrap, mock_sleep):
        bootstrap_nodes = [models.BootstrapInfo.get_mock_object(overrides={'mip': node.mip, 'cip':
                           node.sip}) for node in self.nodes]

        bootstrap_result = result_models.node.GetBootstrapConfig.get_mock_object(
            overrides={'nodes': bootstrap_nodes})

        mock_bootstrap.side_effect = [Exception, bootstrap_result]
        vl_helpers.wait_for_bootstrap(self.node_fm, self.nodes)

    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.Node, 'get_bootstrap_config')
    def test_bootstrap_invalid(self, mock_bootstrap, mock_sleep):
        mock_bootstrap.side_effect = [Exception,
                                      result_models.node.GetBootstrapConfig().get_mock_object()
                                      ]
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_bootstrap(self.node_fm, self.nodes)


class TestDeployDeleteWitnessNode(unittest.TestCase):
    def setUp(self):
        self.witness_name = 'rando_name'
        self.nodes = [models.ResourceNode.get_mock_object(overrides=dict(
            model=sf_models.NodeType(sf_models.NodeType.STANDARD_TYPES.WITNESS),
            node_name=self.witness_name,
            network_profile=models.NetworkProfile.get_mock_object()
        ))]
        self.vms = [{'name': self.witness_name, 'ips': ['1.1.1.1, 2.2.2.2']}]
        self.network_config = models.NetworkConfig.get_mock_object(force_mock=dict(net0=True, net1=True))
        self.one_gig = models.ResourceIp.get_mock_object()
        self.ten_gig = models.ResourceIp.get_mock_object()

    @mock.patch('sf_ssh.Ssh')
    def test_deploy_witness(self, mock_check_output):
        witness_resources = models.WitnessResource()
        vl_helpers.deploy_witness_vm(witness_resources, '1g', '10g', 'url', 'name', {'key': 'blah'})

    @mock.patch('time.sleep')
    @mock.patch('pyVim.connect.SmartConnect')
    def test_delete_witness_nodes(self, mock_smart_connect, mock_sleep):
        witness = models.ResourceNode.get_mock_object()
        witness.model = sf_models.NodeType(sf_models.NodeType.STANDARD_TYPES.WITNESS)
        nodes = [witness]
        vl_helpers.delete_witness_vms(nodes)
        nodes = []
        vl_helpers.delete_witness_vms(nodes)

    @mock.patch('os.system')
    def test_get_next_ip(self, mock_ping):
        mock_ping.side_effect = [0, 0, 1]
        ip = '192.168.0.1'
        vl_helpers.get_next_available_ip(ip)

    @mock.patch('requests.get')
    def test_get_ova_url(self, mock_get):
        obj = type('returnobject', (object,), {})
        obj = obj()
        build = sf_models.Build.create(version='12.0.0.1', respositoryName='magnesium-dev',
                                       release='magnesium-dev')
        release = build.release
        version = build.version
        obj.text = 'a href="solidfire-witness-{}-{}.ova">' \
                   'solidfire-witness-{}-{}.ova</a>'.format(release, version, release, version)
        mock_get.return_value = obj
        retval = vl_helpers.get_ova_url(build)
        assert retval == 'http://bdr-jenkins.eng.solidfire.net/OVA/solidfire-' \
                         'witness-{}-{}.ova'.format(release, version)

    @mock.patch('requests.get')
    def test_get_ova_url_latest(self, mock_get):
        obj = type('returnobject', (object,), {})
        obj = obj()
        build = sf_models.Build.create(version='12.0.0.1', respositoryName='magnesium',
                                       release='magnesium')
        release = build.release
        version = build.version
        obj.text = '<a href="solidfire-witness-{}-12.1.0.12.ova">' \
                   'solidfire-witness-{}-12.1.0.12.ova</a><a href="' \
                   'solidfire-witness-{}-12.1.0.13.ova' \
                   'a">solidfire-witness-{}-12.1.0.13.ova</a>'.format(release, release, release,
                                                                      release, release, release)
        mock_get.return_value = obj
        retval = vl_helpers.get_ova_url(build)
        assert retval == 'http://bdr-jenkins.eng.solidfire.net/OVA/solidfire-' \
                         'witness-{}-12.1.0.13.ova'.format(release)

    def test_get_next_available_ip_invalid(self):
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.get_next_available_ip([])

    def test_create_witness_node_model(self):
        witness_model = vl_helpers.create_witness_model(self.one_gig,
                                                        self.ten_gig,
                                                        node_name=self.witness_name)
        assert isinstance(witness_model, models.ResourceNode)
        witness_model.validate()
        assert witness_model.node_name == self.witness_name

    def test_create_witness_node_model_no_name(self):
        witness_model = vl_helpers.create_witness_model(self.one_gig, self.ten_gig)
        assert isinstance(witness_model, models.ResourceNode)
        assert witness_model.node_name.startswith('WitnessNode-')

    @mock.patch.object(functional_models.Node, 'api_is_up')
    @mock.patch.object(functional_models.Node, 'get_network_config')
    def test_get_node_map(self, mock_network, mock_api):
        mock_network.return_value = self.network_config
        vm_network_map = vl_helpers._node.get_vm_network_config(self.vms)
        assert vm_network_map[self.witness_name] == self.network_config

    @mock.patch.object(functional_models.Node, 'api_is_up')
    @mock.patch.object(functional_models.Node, 'get_network_config')
    def test_get_node_map_fail(self, mock_network, mock_api):
        mock_api.return_value = False
        with pytest.raises(_errors.HelperError):
            vl_helpers._node.get_vm_network_config(self.vms)

    @mock.patch.object(sf_mc.esx_control.Vsphere, 'find_vms_by_name')
    @mock.patch.object(functional_models.Node, 'api_is_up')
    @mock.patch.object(functional_models.Node, 'get_network_config')
    @mock.patch.object(functional_models.Node, 'set_network_config')
    @mock.patch.object(functional_models.Node, 'set_config')
    def test_configure_witness_models(self, mock_set_config, mock_set_network, mock_get_network,
                                      mock_api, mock_vms_by_name):
        mock_get_network.return_value = self.network_config
        vsphere = mock.MagicMock(sf_mc.esx_control.Vsphere(server='1.1.1.1', username='test',
                                                           password='supersecret'))
        vsphere.find_vms_by_name.return_value = self.vms
        vl_helpers._node.configure_witness_node_models(self.nodes, vsphere)

    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_connect')
    @mock.patch.object(vl_helpers._node, 'get_ova_url')
    @mock.patch('sf_ssh.Ssh')
    @mock.patch.object(sf_batch, 'SFThread')
    @mock.patch.object(sf_batch, 'check_for_error_and_raise_exception')
    @mock.patch.object(time, 'sleep')
    @mock.patch.object(vl_helpers._node, 'configure_witness_node_models')
    @mock.patch.object(sf_mc.esx_connection.VsphereConnection, 'vsphere_disconnect')
    def test_create_witness_node(self, mock_disconnect, mock_configure, mock_raise, mock_sleep,
                                 mock_thread, mock_ssh, mock_get_ova, mock_connect):
        mock_get_ova.return_value = 'solidfire-witness.ova'
        build = sf_models.Build.get_mock_object()
        vl_helpers.create_witness_nodes(self.nodes, build=build)
        mock_get_ova.assert_called_with(build)
        mock_configure.assert_called_with(self.nodes, mock.ANY)


class TestResetNodes(unittest.TestCase):
    """Test Node Helper ResetNodes."""
    @mock.patch.object(functional_models.Node, 'reset_node', autospec=True)
    @mock.patch('sf_rtfi.reactor.HTTPIRTFIReactor')
    def test_reset_nodes(self, mock_reactor,  mock_reset):
        mock_async_handler = functional_models.RtfiAsyncHandler(mock_reactor)
        mock_reactor.rtfi_state.failed.return_value = False
        mock_reactor.rtfi_state.aborted.return_value = False
        mock_reactor.rtfi_state.timedout.return_value = False
        mock_reactor.rtfi_state.done.return_value = False
        mock_reactor.rtfi_state.complete.return_value = True
        mock_reset.return_value = mock_async_handler
        vl_helpers.reset_nodes([NODE_FM])


@mock.patch('valence.functional_models.Cluster.list_active_nodes')
class TestConvertPendingNodesToActiveNodes(unittest.TestCase):
    def test_convert_pending_nodes_to_active_nodes_happy_path(self, mock_list_active_nodes):
        mock_list_active_nodes.return_value = models.ActiveNode.get_mock_object_list(1)
        cluster_dm = models.ResourceCluster.get_mock_object()
        cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        pending_nodes = models.PendingNode.get_mock_object_list(1)
        mock_list_active_nodes.return_value[0].name = pending_nodes[0].name
        vl_helpers.convert_pending_nodes_to_active_nodes(cluster_fm, pending_nodes)

    def test_convert_pending_nodes_to_active_nodes_no_active_node_match(self, mock_list_active_nodes):
        mock_list_active_nodes.return_value = models.ActiveNode.get_mock_object_list(1)
        cluster_dm = models.ResourceCluster.get_mock_object()
        cluster_fm = functional_models.Cluster(cluster_dm, authenticate=False)
        pending_nodes = models.PendingNode.get_mock_object_list(1)
        mock_list_active_nodes.return_value[0].name = pending_nodes[0].name + 'something'
        with self.assertRaises(AssertionError):
            vl_helpers.convert_pending_nodes_to_active_nodes(cluster_fm, pending_nodes)


class TestProvisionNodes(unittest.TestCase):
    """Test provision node helper. """
    @mock.patch.object(functional_models.Node, 'start_rtfi', autospec=True)
    @mock.patch.object(functional_models.Node, 'start_traditional_rtfi', autospec=True)
    @mock.patch('sf_rtfi.reactor.HTTPIRTFIReactor')
    def test_provision_nodes(self, mock_reactor, mock_trtfi, mock_irtfi):
        mock_version = sf_api.SfVersion('10.0.0.284')
        mock_async_handler = functional_models.RtfiAsyncHandler(mock_reactor)
        mock_reactor.rtfi_state.failed.return_value = False
        mock_reactor.rtfi_state.aborted.return_value = False
        mock_reactor.rtfi_state.timedout.return_value = False
        mock_reactor.rtfi_state.done.return_value = False
        mock_reactor.rtfi_state.complete.return_value = True
        mock_irtfi.return_value = mock_async_handler
        vl_helpers.provision_nodes([NODE_FM], mock_version)

    @mock.patch.object(functional_models.Node, 'start_rtfi', autospec=True)
    @mock.patch.object(functional_models.Node, 'start_traditional_rtfi', autospec=True)
    @mock.patch('sf_rtfi.reactor.HTTPIRTFIReactor')
    def test_provision_nodes_fail(self, mock_reactor, mock_trtfi, mock_irtfi):
        mock_version = sf_api.SfVersion('10.0.0.284')
        mock_async_handler = functional_models.RtfiAsyncHandler(mock_reactor)
        mock_reactor.rtfi_state.failed.return_value = True
        mock_reactor.rtfi_state.aborted.return_value = False
        mock_reactor.rtfi_state.timedout.return_value = False
        mock_reactor.rtfi_state.done.return_value = False
        mock_reactor.rtfi_state.complete.return_value = False
        mock_irtfi.return_value = mock_async_handler
        mock_trtfi.return_value = mock_async_handler
        with self.assertRaises(_errors.HelperError):
            vl_helpers.provision_nodes([NODE_FM], mock_version)


class TestSortNodesBySize(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.node_fms = [functional_models.Node.get_mock_fm() for _ in range(4)]

        cls.small_drive_info_dms = [models.DriveInfo.create(
            drive_type=enums.DriveType.block,
            size=100
        ) for _ in range(10)]
        cls.small_drive_info_dms.extend([models.DriveInfo.create(
            drive_type=enums.DriveType.slice
        ) for _ in range(2)])

        cls.medium_drive_info_dms = [models.DriveInfo.create(
            drive_type=enums.DriveType.block,
            size=110
        ) for _ in range(10)]
        cls.medium_drive_info_dms.extend([models.DriveInfo.create(
            drive_type=enums.DriveType.slice
        ) for _ in range(2)])

        cls.large_drive_info_dms = [models.DriveInfo.create(
            drive_type=enums.DriveType.block,
            size=120
        ) for _ in range(10)]
        cls.large_drive_info_dms.extend([models.DriveInfo.create(
            drive_type=enums.DriveType.slice
        ) for _ in range(2)])

        cls.drive_configs = [
            models.DriveConfig.create(drives=cls.medium_drive_info_dms) for _ in range(2)
        ]
        cls.drive_configs.append(models.DriveConfig.create(drives=cls.large_drive_info_dms))
        cls.drive_configs.append(models.DriveConfig.create(drives=cls.small_drive_info_dms))

    @mock.patch.object(functional_models.Node, 'get_drive_config')
    def test_sorting(self, get_drive_config):
        get_drive_config.side_effect = self.drive_configs
        sorted_node_fms = vl_helpers.sort_nodes_by_size(self.node_fms)
        assert sorted_node_fms[0] == self.node_fms[-1]
        assert sorted_node_fms[-1] == self.node_fms[-2]


@mock.patch.object(functional_models.Node, 'authenticate', mock.MagicMock())
class TestListNodesByUsableBlockCapacity(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)
        cls.node_dms = [
            models.ActiveNode.create(node_id=i, mip='10.117.42.{}'.format(i)) for i in range(4)
        ]
        for node_dm in cls.node_dms:
            node_dm._resource_dm = models.ResourceNode.create(mip=node_dm.mip)

        node_drive_dms = [
            models.Drive.create(node_id=cls.node_dms[0].node_id, capacity=100),
            models.Drive.create(node_id=cls.node_dms[1].node_id, capacity=160),
            models.Drive.create(node_id=cls.node_dms[2].node_id, capacity=80),
            models.Drive.create(node_id=cls.node_dms[3].node_id, capacity=100)
        ]
        cls.drive_dms = []
        cls.drive_dms.extend(node_drive_dms[0] for _ in range(10))
        cls.drive_dms.extend(node_drive_dms[1] for _ in range(6))
        cls.drive_dms.extend(node_drive_dms[2] for _ in range(11))
        cls.drive_dms.extend(node_drive_dms[3] for _ in range(11))

    @mock.patch.object(functional_models.Cluster, 'list_drives')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_sorting(self, list_active_nodes, list_drives):
        list_active_nodes.return_value = self.node_dms
        list_drives.return_value = self.drive_dms
        sorted_node_fms = vl_helpers.list_nodes_by_usable_block_capacity(self.cluster_fm)
        assert sorted_node_fms[0].node_dm.mip == self.node_dms[2].mip
        assert sorted_node_fms[1].node_dm.mip == self.node_dms[1].mip
        assert sorted_node_fms[2].node_dm.mip == self.node_dms[0].mip
        assert sorted_node_fms[3].node_dm.mip == self.node_dms[3].mip


@mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
class TestCat(unittest.TestCase):
    """."""
    def test_cat(self, mock_ssh):
        """Automatically generated docstring for test_cat.

        :param Any mock_ssh: Autogenerated description.
        """
        cluster_dm = models.ResourceCluster.get_mock_object()
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.return_value = sf_ssh.RunCommandOutput(stdout='/tmp/unittest', stderr='', return_code=0)
        vl_helpers.get_file_contents_from_node(node_fm, target_dir='/tmp/unit', file_name='unit',
                                               timeout=0)


@mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
class TestRevertViaCLI(unittest.TestCase):
    def setUp(self):
        self.node_dm = models.ActiveNode.get_mock_object()
        cluster_dm = models.Cluster.get_mock_object()
        self.node_fm = functional_models.Node(self.node_dm, cluster_dm)
        self.node_fm.api.version = models.SfVersion('10.1')

    def test_good(self, mock_ssh):
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = iter([
            sf_ssh.RunCommandOutput(
                stdout=('total 28\n'
                        'drwxr-xr-x  2 root root 4096 Mar 22 21:42 .\n'
                        'drwxr-xr-x 19 root root 4096 Mar 22 18:22 ..\n'
                        'lrwxrwxrwx  1 root root   23 Mar 22 18:56 active.crt -> /sf/etc/ssl/user.crt\n'
                        'lrwxrwxrwx  1 root root   23 Mar 22 18:57 active.key -> /sf/etc/ssl/user.key\n'
                        '-rw-r--r--  1 root root 1330 Mar 22 18:21 cluster.crt\n'
                        '-rw-r--r--  1 root root 1704 Mar 22 18:21 cluster.key\n'
                        '-rw-r--r--  1 root root 1330 Mar 22 18:19 node.crt\n'
                        '-rw-r--r--  1 root root 1704 Mar 22 18:19 node.key\n'
                        '-rw-r--r--  1 root root 3034 Mar 22 18:19 node.ssl\n'),
                stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)])
        vl_helpers.revert_cert_via_cli(self.node_fm)

    def test_ember(self, mock_ssh):
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = iter([
            sf_ssh.RunCommandOutput(
                stdout=('total 28\n'
                        'drwxr-xr-x  2 root root 4096 Mar 22 21:42 .\n'
                        'drwxr-xr-x 19 root root 4096 Mar 22 18:22 ..\n'
                        'lrwxrwxrwx  1 root root   23 Mar 22 18:56 active.crt -> /sf/etc/ssl/user.crt\n'
                        'lrwxrwxrwx  1 root root   23 Mar 22 18:57 active.key -> /sf/etc/ssl/user.key\n'
                        '-rw-r--r--  1 root root 1330 Mar 22 18:21 cluster.crt\n'
                        '-rw-r--r--  1 root root 1704 Mar 22 18:21 cluster.key\n'
                        '-rw-r--r--  1 root root 1330 Mar 22 18:19 node.crt\n'
                        '-rw-r--r--  1 root root 1704 Mar 22 18:19 node.key\n'
                        '-rw-r--r--  1 root root 3034 Mar 22 18:19 node.ssl\n'),
                stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)])
        self.node_fm.api.version = models.SfVersion('11.0')
        vl_helpers.revert_cert_via_cli(self.node_fm)

    def test_already_cluster(self, mock_ssh):
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = iter([
            sf_ssh.RunCommandOutput(
                stdout=(
                    'total 28\n'
                    'drwxr-xr-x  2 root root 4096 Mar 22 21:42 .\n'
                    'drwxr-xr-x 19 root root 4096 Mar 22 18:22 ..\n'
                    'lrwxrwxrwx  1 root root   23 Mar 22 18:56 active.crt -> /sf/etc/ssl/cluster.crt\n'
                    'lrwxrwxrwx  1 root root   23 Mar 22 18:57 active.key -> /sf/etc/ssl/cluster.key\n'
                    '-rw-r--r--  1 root root 1330 Mar 22 18:21 cluster.crt\n'
                    '-rw-r--r--  1 root root 1704 Mar 22 18:21 cluster.key\n'
                    '-rw-r--r--  1 root root 1330 Mar 22 18:19 node.crt\n'
                    '-rw-r--r--  1 root root 1704 Mar 22 18:19 node.key\n'
                    '-rw-r--r--  1 root root 3034 Mar 22 18:19 node.ssl\n'),
                stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)])
        with pytest.raises(_errors.HelperError):
            vl_helpers.revert_cert_via_cli(self.node_fm)

    def test_no_reload(self, mock_ssh):
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = iter([
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=1),
        ])
        with pytest.raises(_errors.HelperError):
            vl_helpers.reload_nginx(self.node_fm)


@mock.patch.object(functional_models.Node, 'test_hardware_config')
class TestTestHardwareConfig(unittest.TestCase):
    """Test node helpers related to testing the hardware config."""

    def test_test_hardware_config(self, mock_test_hardware_config):
        nodes = [NODE_FM]
        mock_test_hardware_config.side_effect = functional_models.FunctionalNodeError
        with pytest.raises(BaseException):
            vl_helpers.test_hardware_config(nodes)


@mock.patch.object(functional_models.Node, 'flash_nvdimm_firmware')
@mock.patch.object(functional_models.Node, 'verify_nvdimm_health')
class TestFlashAndVerifyNvdimmFw(unittest.TestCase):
    """Test flash_and_verify_nvdimm_fw()."""
    @classmethod
    def setUp(cls):
        cls.node_fm = functional_models.Node.get_mock_fm(validate_responses=True)

    def test_flash_and_verify_nvdimm_fw(self, mock_flash_nvdimm_firmware, mock_verify_nvdimm_heatlh):
        mock_flash_nvdimm_firmware.return_value = True
        mock_verify_nvdimm_heatlh.return_value = None
        self.assertIsNone(
            vl_helpers.flash_and_verify_nvdimm_fw(node_fm=self.node_fm, expected_version='2.11',
                                                  downgrade=False)
        )


@mock.patch.object(time, 'time')
@mock.patch.object(time, 'sleep')
@mock.patch.object(functional_models.Node, 'api_is_responsive')
class TestWaitForNodeApiUp(unittest.TestCase):
    """Tests for wait_for_node_api_up vl_helper."""
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()

    def test_wait_for_api_up_success(self, mock_responsive, mock_sleep, mock_time):
        mock_time.return_value = 0
        mock_responsive.side_effect = [False, True]
        vl_helpers.wait_for_node_api_up(self.node_fm)

    def test_wait_for_api_up_fail(self, mock_responsive, mock_sleep, mock_time):
        mock_time.side_effect = [0, 2]
        mock_responsive.return_value = False
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_node_api_up(self.node_fm, timeout=1)


@mock.patch.object(sf_batch, 'SFProcess', autospec=True)
class TestPerformPduNodeOperation(unittest.TestCase):
    """Testing performing Reboot/ON/OFF node operation on PDUs."""
    @classmethod
    def setUpClass(cls):
        cls.node_fm = functional_models.Node.get_mock_fm()
        cls.mcpdu1 = mock.MagicMock(sf_mc.McControlPdu(host='10.194.1.1', pdu_model='apc',
                                                       outlets=['4', '21']))
        cls.mcpdu2 = mock.MagicMock(sf_mc.McControlPdu(host='10.194.2.1', pdu_model='apc',
                                                       outlets=['4', '21']))
        cls.mcpdus = [cls.mcpdu1, cls.mcpdu2]

    def test_single_pdu_reboot(self, mock_sfprocess):
        vl_helpers._node.perform_pdu_reboot_node_operation(mcpdus=[self.mcpdu1], timeout=10)
        mock_sfprocess.assert_called_once()

    def test_single_pdu_power_on(self, mock_sfprocess):
        vl_helpers._node.perform_pdu_power_on_node_operation(mcpdus=[self.mcpdu1], timeout=10)
        mock_sfprocess.assert_called_once()

    def test_single_pdu_power_off(self, mock_sfprocess):
        vl_helpers._node.perform_pdu_power_off_node_operation(mcpdus=[self.mcpdu1], timeout=10)
        mock_sfprocess.assert_called_once()

    def test_dual_pdu_reboot(self, mock_sfprocess):
        vl_helpers._node.perform_pdu_reboot_node_operation(mcpdus=self.mcpdus)
        mock_sfprocess.assert_called()

    def test_dual_pdu_power_on(self, mock_sfprocess):
        vl_helpers._node.perform_pdu_power_on_node_operation(mcpdus=self.mcpdus)
        mock_sfprocess.assert_called()

    def test_dual_pdu_power_off(self, mock_sfprocess):
        vl_helpers._node.perform_pdu_power_off_node_operation(mcpdus=self.mcpdus)
        mock_sfprocess.assert_called()
