import os
import time
import unittest

import mock
import sf_ssh

import valence
from valence import (
    functional_models,
    models,
    unittest_utils,
)
from valence.functional_models import _constants
from valence.vl_helpers import (
    _errors,
    _preupgrade_healthcheck,
)
from valence.vl_helpers._healthtools_versions_for_upgrade_suites import HealthtoolsVersionsForTest

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestPreupgradeHealthcheck(unittest.TestCase):
    """Tests the preupgrade healthchecks."""

    @classmethod
    def setUpClass(cls):
        mnode_dm = models.ResourceManagementNode.get_mock_object()
        cls.mnode_fm = functional_models.ManagementNodeLegacy(mnode_dm)
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.cluster_dm)
        cls.node_dms = [
            models.ActiveNode.create(node_id=4),
            models.ActiveNode.create(node_id=1),
            models.ActiveNode.create(node_id=2),
            models.ActiveNode.create(node_id=3)
        ]
        cls.healthtools_helper = HealthtoolsVersionsForTest(cls.mnode_fm)

    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    def test_mnode_old(self, mock_mnode_version):
        mock_mnode_version.return_value = _constants.FLUORINE_PATCH5_1
        result = _preupgrade_healthcheck.run_preupgrade_healthchecks(self.mnode_fm, self.cluster_dm)
        assert result

    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_setup_healthtools_on_mnode_url(self, mock_run_command, mock_mnode_version):
        url_to_deploy = 'http://yada/blah/solidfire-healthtools-2020.01.01.01.tgz'
        mock_mnode_version.return_value = _constants.SODIUM
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='mkdir worked', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='wget worked', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='sfupdate-healthtools worked', stderr='', return_code=0)]

        _preupgrade_healthcheck._setup_healthtools_on_mnode(
            self.mnode_fm, self.healthtools_helper, url_to_healthtools_to_deploy=url_to_deploy,
            install_rqd_healthtools=False)

        assert mock_run_command.call_count == 3
        assert mock_run_command.call_args_list[0].args[0] == 'mkdir -p /tmp/ura-packages'
        assert url_to_deploy in mock_run_command.call_args_list[1].args[0]
        assert 'sudo -E /sf/bin/sfupdate-healthtools /tmp/ura-packages/solidfire-healthtools-' \
               '2020.01.01.01.tgz' == mock_run_command.call_args_list[2].args[0]

    @mock.patch('valence.vl_helpers.HealthtoolsVersionsForTest.jenkins_rqd_versions',
                new_callable=mock.PropertyMock)
    @mock.patch.object(valence.vl_helpers.HealthtoolsVersionsForTest, 'initialize_versions')
    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_setup_healthtools_on_mnode_no_url(self, mock_run_command, mock_mnode_version, mock_init,
                                               mock_rqd):
        mock_mnode_version.return_value = _constants.SODIUM
        mock_run_command.side_effect = [
            sf_ssh.ssh.RunCommandOutput(stdout='mkdir worked', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='wget worked', stderr='', return_code=0),
            sf_ssh.ssh.RunCommandOutput(stdout='sfupdate-healthtools worked', stderr='', return_code=0)]
        mock_rqd.return_value = [
            _constants.HEALTHTOOLS_18_3_1_OOB_VERSION,
            _constants.HEALTHTOOLS_19_1_VERSION,
            _constants.HEALTHTOOLS_19_2_VERSION
        ]

        # fake the healthtool helper initialization.
        self.healthtools_helper._healthtool_urls[_constants.HEALTHTOOLS_19_2_VERSION] = 'http://19.2'
        self.healthtools_helper._healthtool_urls[_constants.HEALTHTOOLS_19_1_VERSION] = 'http://19.1'
        self.healthtools_helper._healthtool_urls[_constants.HEALTHTOOLS_18_3_1_OOB_VERSION] = '//18.3'

        _preupgrade_healthcheck._setup_healthtools_on_mnode(self.mnode_fm, self.healthtools_helper,
                                                            install_rqd_healthtools=True)

        assert mock_run_command.call_count == 3
        assert mock_run_command.call_args_list[0].args[0] == 'mkdir -p /tmp/ura-packages'
        assert str(_constants.HEALTHTOOLS_19_2_VERSION) in mock_run_command.call_args_list[1].args[0]
        assert 'http://19.2' in mock_run_command.call_args_list[1].args[0]
        assert 'sudo -E /sf/bin/sfupdate-healthtools /tmp/ura-packages/solidfire-healthtools-' \
               '2019.12.01.0.tgz' == mock_run_command.call_args_list[2].args[0]

    @mock.patch.object(valence.vl_helpers._sfupgradecheck_validation, 'validate_sfupgradecheck_results')
    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    @mock.patch('valence.vl_helpers.HealthtoolsVersionsForTest.current_version',
                new_callable=mock.PropertyMock)
    @mock.patch.object(valence.functional_models.Cluster, 'list_active_nodes')
    def test_verify_sfupgradecheck_pass(self, mock_nodes, mock_ht, mock_mnode_version, mock_sfupgrade):
        mock_mnode_version.return_value = _constants.SODIUM
        mock_ht.return_value = _constants.HEALTHTOOLS_18_3_VERSION
        mock_nodes.return_value = self.node_dms

        result = _preupgrade_healthcheck._validate_preupgrade_checks(
            self.mnode_fm, self.cluster_fm, self.healthtools_helper, 'blah')

        assert result
        assert mock_sfupgrade.called
        assert mock_mnode_version.called
        assert mock_ht.called
        assert mock_nodes.called

    @mock.patch.object(valence.vl_helpers._sfupgradecheck_validation, 'validate_sfupgradecheck_results')
    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    @mock.patch('valence.vl_helpers.HealthtoolsVersionsForTest.current_version',
                new_callable=mock.PropertyMock)
    @mock.patch.object(valence.functional_models.Cluster, 'list_active_nodes')
    def test_verify_sfupgradecheck_fail(self, mock_nodes, mock_ht, mock_mnode_version, mock_sfupgrade):
        mock_mnode_version.return_value = _constants.SODIUM
        mock_ht.return_value = _constants.HEALTHTOOLS_18_3_VERSION
        mock_nodes.return_value = self.node_dms
        mock_sfupgrade.side_effect = _errors.HelperError()
        result = _preupgrade_healthcheck._validate_preupgrade_checks(
            self.mnode_fm, self.cluster_fm, self.healthtools_helper, 'blah')

        assert not result
        assert mock_sfupgrade.called
        assert mock_mnode_version.called
        assert mock_ht.called
        assert mock_nodes.called

    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    @mock.patch.object(valence.functional_models.Cluster, 'authenticate')
    @mock.patch.object(valence.vl_helpers._preupgrade_healthcheck, '_setup_healthtools_on_mnode')
    @mock.patch.object(valence.functional_models.ManagementNodeLegacy,
                       'execute_preupgrade_health_check')
    @mock.patch.object(valence.vl_helpers._preupgrade_healthcheck, '_validate_preupgrade_checks')
    def test_mnode_run_pass(self, mock_validate, mock_run, mock_setup, mock_auth, mock_mnode_version):
        mock_mnode_version.return_value = _constants.SODIUM
        mock_validate.return_value = True
        fake_thingie = valence.functional_models.ProcessedHealthchecks(
            cluster_checks=[models.Healthcheck.create(name='a', passed=False),
                            models.Healthcheck.create(name='b', passed=True)],
            node_checks={1: [models.Healthcheck.create(name='c', passed=False)]})
        mock_run.side_effect = [fake_thingie, fake_thingie]
        result = _preupgrade_healthcheck.run_preupgrade_healthchecks(self.mnode_fm, self.cluster_dm)
        assert result
        assert mock_auth.called
        assert mock_setup.called
        assert mock_mnode_version.called
        assert mock_run.called
        assert mock_run.call_count == 1
        assert mock_validate.called

    @mock.patch('valence.functional_models.ManagementNodeLegacy.version', new_callable=mock.PropertyMock)
    @mock.patch.object(valence.functional_models.Cluster, 'authenticate')
    @mock.patch.object(valence.vl_helpers._preupgrade_healthcheck, '_setup_healthtools_on_mnode')
    @mock.patch.object(valence.functional_models.ManagementNodeLegacy,
                       'execute_preupgrade_health_check')
    @mock.patch.object(valence.vl_helpers._preupgrade_healthcheck, '_validate_preupgrade_checks')
    def test_mnode_run_fail(self, mock_validate, mock_run, mock_setup, mock_auth, mock_mnode_version):
        mock_mnode_version.return_value = _constants.SODIUM
        mock_validate.return_value = False
        fake_thingie = valence.functional_models.ProcessedHealthchecks(
            cluster_checks=[models.Healthcheck.create(name='a', passed=False),
                            models.Healthcheck.create(name='b', passed=True)],
            node_checks={1: [models.Healthcheck.create(name='c', passed=False)]})
        mock_run.side_effect = [fake_thingie, fake_thingie]
        result = _preupgrade_healthcheck.run_preupgrade_healthchecks(self.mnode_fm, self.cluster_dm)
        assert not result
        assert mock_auth.called
        assert mock_setup.called
        assert mock_mnode_version.called
        assert mock_run.called
        assert mock_run.call_count == 1
        assert mock_validate.called
