import unittest

import mock
import pytest
from sf_ssh import ssh

from valence import (
    enums,
    functional_models,
    models,
    utils,
    vl_helpers,
)
from valence.vl_helpers import _remote_replication

# Mock the cluster DM and FM once to avoid extra work
RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
CLUSTER_FM = functional_models.Cluster(RS_CLUSTER_DM, authenticate=False)

SYNCING_STATUSES = [enums.RemoteClusterSnapshotStatus.present,
                    enums.RemoteClusterSnapshotStatus.present,
                    enums.RemoteClusterSnapshotStatus.not_present,
                    enums.RemoteClusterSnapshotStatus.not_present]

SYNCED_STATUSES = [enums.RemoteClusterSnapshotStatus.present,
                   enums.RemoteClusterSnapshotStatus.present,
                   enums.RemoteClusterSnapshotStatus.present,
                   enums.RemoteClusterSnapshotStatus.present]


@mock.patch('sf_ssh.Ssh')
class TestNetemRules(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """Setup model objects for the individual test cases."""

        cls.rs_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.rs_cluster_dm, authenticate=False)

        cls.node_dms = [models.ActiveNode.get_mock_object(overrides={'username': 'foo'}) for _ in
                        range(2)]
        cls.node_fms = [functional_models.Node(node.resource_dm, cls.rs_cluster_dm, authenticate=False)
                        for node in cls.node_dms]

    def test_add_netem_rules_inter_nodes(self, mock_ssh):
        dest_filter_cmd = ('tc filter add dev {dev} protocol ip parent 1:0 prio 1 u32 '
                           'match ip dst {dst} flowid 1:1')

        inst = mock_ssh.return_value.__enter__()
        inst.run_command.return_value = ssh.RunCommandOutput(stderr='', stdout='', return_code=1)

        dev_target_map = {
            'Bond10G': ['1.2.3.4'],
            'Bond1G': ['1.2.3.5', '1.2.3.6']
        }

        _remote_replication.add_netem_rules_inter_nodes(self.node_fms, dev_target_map,
                                                        [
                                                              'delay 50ms',
                                                              'loss 5%'
                                                          ])

        inst.run_command.assert_any_call(
            'tc qdisc add dev Bond10G root handle 1: prio priomap 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2',
            raise_on_failure=True)
        inst.run_command.assert_any_call(
            'tc qdisc add dev Bond10G root handle 1: prio priomap 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2',
            raise_on_failure=True)

        inst.run_command.assert_any_call(
            'tc qdisc add dev Bond10G parent 1:1 handle 10: netem delay 50ms loss 5%',
            raise_on_failure=True)
        inst.run_command.assert_any_call(
            'tc qdisc add dev Bond1G parent 1:1 handle 10: netem delay 50ms loss 5%',
            raise_on_failure=True)

        inst.run_command.assert_any_call(
            dest_filter_cmd.format(dev='Bond10G', dst='1.2.3.4'),
            raise_on_failure=True
        )
        inst.run_command.assert_any_call(
            dest_filter_cmd.format(dev='Bond1G', dst='1.2.3.5'),
            raise_on_failure=True
        )
        inst.run_command.assert_any_call(
            dest_filter_cmd.format(dev='Bond1G', dst='1.2.3.6'),
            raise_on_failure=True
        )

    def test_clear_netem_rules(self, mock_ssh):
        inst = mock_ssh.return_value.__enter__()
        inst.run_command.return_value = ssh.RunCommandOutput(stderr='', stdout='', return_code=1)

        _remote_replication.clear_netem_rules(self.node_fms, ['Bond10G'])

        inst.run_command.assert_called_with('tc qdisc del root dev Bond10G')


@mock.patch('valence.functional_models.Cluster.list_group_snapshots')
class TestGetGroupSnapshotsRemoteStatuses(unittest.TestCase):

    def setUp(self):
        self.group_snapshot_dms = []
        self.group_snapshot_dms.extend(models.GroupSnapshot.get_mock_object_list(
            2, overrides={'remote_statuses': [{'remote_status': 'Present'}]}))
        self.group_snapshot_dms.extend(models.GroupSnapshot.get_mock_object_list(
            2, overrides={'remote_statuses': [{'remote_status': 'NotPresent'}]}))

    def test_get_group_snapshots_remote_statuses(self, mock_list_group_snapshots):
        mock_list_group_snapshots.return_value = self.group_snapshot_dms
        statuses = _remote_replication.get_group_snapshots_remote_statuses(CLUSTER_FM)
        assert statuses == SYNCING_STATUSES

    def test_get_group_snapshots_remote_statuses_filter(self, mock_list_group_snapshots):
        filter_dms = self.group_snapshot_dms[1:3]
        mock_list_group_snapshots.return_value = self.group_snapshot_dms
        statuses = _remote_replication.get_group_snapshots_remote_statuses(CLUSTER_FM, filter_dms)
        assert statuses == SYNCING_STATUSES[1:3]


@mock.patch('valence.functional_models.Cluster.list_snapshots')
class TestGetSnapshotsRemoteStatuses(unittest.TestCase):

    def setUp(self):
        self.snapshot_dms = []
        self.snapshot_dms.extend(models.Snapshot.get_mock_object_list(
            2, overrides={'remote_statuses': [{'remote_status': 'Present'}]}))
        self.snapshot_dms.extend(models.Snapshot.get_mock_object_list(
            2, overrides={'remote_statuses': [{'remote_status': 'NotPresent'}]}))

    def test_get_group_snapshots_remote_statuses(self, mock_list_snapshots):
        mock_list_snapshots.return_value = self.snapshot_dms
        statuses = _remote_replication.get_snapshots_remote_statuses(CLUSTER_FM)
        assert statuses == SYNCING_STATUSES

    def test_get_group_snapshots_remote_statuses_filter(self, mock_list_snapshots):
        filter_dms = self.snapshot_dms[1:3]
        mock_list_snapshots.return_value = self.snapshot_dms
        statuses = _remote_replication.get_snapshots_remote_statuses(CLUSTER_FM, filter_dms)
        assert statuses == SYNCING_STATUSES[1:3]


@mock.patch('time.sleep')
@mock.patch.object(utils, 'timeout_timer')
@mock.patch.object(vl_helpers._remote_replication, 'get_snapshots_remote_statuses')
@mock.patch.object(vl_helpers._remote_replication, 'get_group_snapshots_remote_statuses')
@mock.patch.object(functional_models.Cluster, 'list_group_snapshots')
class TestWaitForGroupSnapshotReplication(unittest.TestCase):

    def test_wait_for_group_snapshot_replication(self, mock_list_group_snapshots,
                                                 mock_get_group_statuses, mock_get_member_statuses,
                                                 mock_timeout_timer, mock_sleep):
        mock_grp_snaps = models.GroupSnapshot.get_mock_object_list(2)
        mock_members = []
        for grp_snaps in mock_grp_snaps:
            mock_members.extend(grp_snaps.members)

        mock_list_group_snapshots.return_value = mock_grp_snaps
        mock_get_group_statuses.side_effect = [SYNCING_STATUSES, SYNCED_STATUSES]
        mock_get_member_statuses.side_effect = [SYNCED_STATUSES]
        mock_timeout_timer.return_value = iter([True, True])
        vl_helpers.wait_for_group_snapshots_replication(CLUSTER_FM)

        mock_list_group_snapshots.assert_called()
        mock_get_member_statuses.assert_called_with(CLUSTER_FM, filter_snapshot_dms=mock_members)
        mock_get_group_statuses.assert_called_with(CLUSTER_FM, None)

    def test_wait_for_grp_snap_rep_member_syncing(self, mock_list_group_snapshots,
                                                  mock_get_group_statuses, mock_get_member_statuses,
                                                  mock_timeout_timer, mock_sleep):
        mock_get_group_statuses.side_effect = [SYNCED_STATUSES, SYNCED_STATUSES]
        mock_get_member_statuses.side_effect = [SYNCING_STATUSES, SYNCED_STATUSES]
        mock_timeout_timer.return_value = iter([True, True])
        vl_helpers.wait_for_group_snapshots_replication(CLUSTER_FM)

        mock_list_group_snapshots.assert_called()
        mock_get_group_statuses.assert_called_with(CLUSTER_FM, None)

    def test_wait_for_group_snapshot_replication_w_timeout(self, mock_list_group_snapshots,
                                                           mock_get_group_statuses,
                                                           mock_get_member_statuses,
                                                           mock_timeout_timer, mock_sleep):
        mock_get_group_statuses.side_effect = [SYNCING_STATUSES, SYNCING_STATUSES]
        mock_get_member_statuses.side_effect = []
        mock_timeout_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_group_snapshots_replication(CLUSTER_FM)

        mock_list_group_snapshots.assert_called()
        mock_get_group_statuses.assert_called_with(CLUSTER_FM, None)

    def test_wait_for_group_snapshot_member_syncing_w_timeout(self, mock_list_group_snapshots,
                                                              mock_get_group_statuses,
                                                              mock_get_member_statuses,
                                                              mock_timeout_timer, mock_sleep):
        mock_get_group_statuses.side_effect = [SYNCED_STATUSES, SYNCED_STATUSES]
        mock_get_member_statuses.side_effect = [SYNCING_STATUSES, SYNCING_STATUSES]
        mock_timeout_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_group_snapshots_replication(CLUSTER_FM)

        mock_list_group_snapshots.assert_called()
        mock_get_group_statuses.assert_called_with(CLUSTER_FM, None)

    def test_wait_for_group_snapshot_replication_w_o_remote_status(self, mock_list_group_snapshots,
                                                                   mock_get_group_statuses,
                                                                   mock_get_member_statuses,
                                                                   mock_timeout_timer, mock_sleep):
        mock_get_group_statuses.return_value = []
        mock_get_member_statuses.side_effect = []
        mock_timeout_timer.return_value = iter([True])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_group_snapshots_replication(CLUSTER_FM)

        mock_list_group_snapshots.assert_called()
        mock_get_group_statuses.assert_called_with(CLUSTER_FM, None)


@mock.patch('time.sleep')
@mock.patch.object(utils, 'timeout_timer')
@mock.patch.object(vl_helpers._remote_replication, 'get_snapshots_remote_statuses')
class TestWaitForSnapshotReplication(unittest.TestCase):

    def test_wait_for_snapshot_replication(self, mock_get_snap_statuses, mock_timeout_timer,
                                           mock_sleep):
        mock_get_snap_statuses.side_effect = [SYNCING_STATUSES, SYNCED_STATUSES]
        mock_timeout_timer.return_value = iter([True, True])
        vl_helpers.wait_for_snapshots_replication(CLUSTER_FM)
        mock_get_snap_statuses.assert_called_with(CLUSTER_FM, None)

    def test_wait_for_snapshot_replication_w_timeout(self, mock_get_snap_statuses, mock_timeout_timer,
                                                     mock_sleep):
        mock_get_snap_statuses.side_effect = [SYNCING_STATUSES, SYNCING_STATUSES]
        mock_timeout_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_snapshots_replication(CLUSTER_FM)
        mock_get_snap_statuses.assert_called_with(CLUSTER_FM, None)

    def test_wait_for_snapshot_replication_w_timeout_and_filter(self, mock_get_snap_statuses,
                                                                mock_timeout_timer, mock_sleep):
        mock_snaps = models.Snapshot.get_mock_object_list(2)

        mock_get_snap_statuses.side_effect = [SYNCING_STATUSES, SYNCING_STATUSES]
        mock_timeout_timer.return_value = iter([True, False])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_snapshots_replication(CLUSTER_FM, filter_snapshot_dms=mock_snaps)
        mock_get_snap_statuses.assert_called_with(CLUSTER_FM, mock_snaps)

    def test_wait_for_snapshot_replication_w_o_remote_status(self, mock_get_snap_statuses,
                                                             mock_timeout_timer, mock_sleep):
        mock_get_snap_statuses.return_value = []
        mock_timeout_timer.return_value = iter([True])
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.wait_for_snapshots_replication(CLUSTER_FM)
        mock_get_snap_statuses.assert_called_with(CLUSTER_FM, None)


@mock.patch.object(functional_models.Cluster, 'list_snapshots')
class TestGetPeerSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.rs_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.rs_cluster_dm, authenticate=False)

    def setUp(self):
        self.snapshot_dms = models.Snapshot.get_mock_object_list(
            2, overrides={'remote_statuses': [{'remote_status': 'Present'}]}
        )
        self.source_snap = models.Snapshot.get_mock_object(overrides={
            'snapshot_uuid': self.snapshot_dms[0].snapshot_uuid
        })

    def test_get_peer_snapshot(self, mock_get_snaps):
        mock_get_snaps.return_value = self.snapshot_dms

        test_value = vl_helpers.get_peer_snapshot(self.cluster_fm, self.source_snap)
        assert test_value == self.snapshot_dms[0]

    def test_get_peer_snapshot_missing(self, mock_get_snaps):
        mock_get_snaps.return_value = []

        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.get_peer_snapshot(self.cluster_fm, self.source_snap)

    def test_get_peer_snapshot_multiple_found(self, mock_get_snaps):
        mock_get_snaps.return_value = [self.snapshot_dms[0], self.snapshot_dms[0]]

        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.get_peer_snapshot(self.cluster_fm, self.source_snap)


@mock.patch.object(functional_models.Cluster, 'list_group_snapshots')
class TestGetPeerGroupSnapshot(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.rs_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.rs_cluster_dm, authenticate=False)

    def setUp(self):
        self.grp_snapshot_dms = models.GroupSnapshot.get_mock_object_list(
            2, overrides={'remote_statuses': [{'remote_status': 'Present'}]}
        )
        self.source_grp_snap = models.GroupSnapshot.get_mock_object(overrides={
            'group_snapshot_uuid': self.grp_snapshot_dms[0].group_snapshot_uuid
        })

    def test_get_peer_group_snapshot(self, mock_get_grp_snaps):
        mock_get_grp_snaps.return_value = self.grp_snapshot_dms

        test_value = vl_helpers.get_peer_group_snapshot(self.cluster_fm, self.source_grp_snap)
        assert test_value == self.grp_snapshot_dms[0]

    def test_get_peer_group_snapshot_missing(self, mock_get_grp_snaps):
        mock_get_grp_snaps.return_value = []

        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.get_peer_group_snapshot(self.cluster_fm, self.source_grp_snap)

    def test_get_peer_group_snapshot_multiple_found(self, mock_get_grp_snaps):
        mock_get_grp_snaps.return_value = [self.grp_snapshot_dms[0], self.grp_snapshot_dms[0]]

        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.get_peer_group_snapshot(self.cluster_fm, self.source_grp_snap)


@mock.patch.object(functional_models.Cluster, 'list_group_snapshots')
class TestGetAndFilterGroupSnaps(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.rs_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.cluster_fm = functional_models.Cluster(cls.rs_cluster_dm, authenticate=False)

    def setUp(self):
        self.grp_snapshot_dms = models.GroupSnapshot.get_mock_object_list(3)

    def test_get_and_filter_group_snap_dms_no_filter(self, mock_list_group_snaps):
        """Test the routine where a filter isn't provided."""
        mock_list_group_snaps.return_value = self.grp_snapshot_dms
        retval = vl_helpers._remote_replication._get_and_filter_group_snap_dms(self.cluster_fm)
        assert retval == self.grp_snapshot_dms

    def test_get_and_filter_group_snap_dms_subset_filter(self, mock_list_group_snaps):
        """Test the routine where the filter value is a subset of the list_group_snaps values."""
        mock_list_group_snaps.return_value = self.grp_snapshot_dms
        retval = vl_helpers._remote_replication._get_and_filter_group_snap_dms(
            self.cluster_fm, filter_group_snapshot_dms=self.grp_snapshot_dms[1:])

        assert retval == self.grp_snapshot_dms[1:]

    def test_get_and_filter_group_snap_dms_total_filter(self, mock_list_group_snaps):
        """Test the routine where the filter value doesn't match any of the list_group_snaps values."""
        mock_list_group_snaps.return_value = self.grp_snapshot_dms[:1]
        retval = vl_helpers._remote_replication._get_and_filter_group_snap_dms(
            self.cluster_fm, filter_group_snapshot_dms=self.grp_snapshot_dms[1:])

        assert retval == []
