import time
import unittest

import mock
import pytest
import sf_models
import sf_ssh
from cryptography import x509

import sf_api.api_constants
from sf_api import (
    api_errors,
    enums,
    models,
    result_models,
)
from valence import (
    errors,
    functional_models,
    models,
    unittest_utils,
    utils,
    vl_helpers,
)
from valence.vl_helpers import _errors

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


class TestEKMHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        mock_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.mock_cluster_fm = functional_models.Cluster(mock_cluster_dm, authenticate=False)

    def test_setup_external_key_management_big(self):
        mock_eks_dms = sf_models.ExternalKeyServer.get_mock_object_list(length=2)
        with pytest.raises(_errors.HelperError):
            vl_helpers.setup_external_key_management(self.mock_cluster_fm, mock_eks_dms)

    @mock.patch('valence.vl_helpers._security._create_key_pair')
    @mock.patch.object(functional_models.Cluster, 'create_key_provider_kmip')
    @mock.patch.object(functional_models.Cluster, 'get_client_certificate_sign_request')
    @mock.patch.object(vl_helpers._security, 'create_key_server')
    @mock.patch.object(functional_models.Cluster, 'add_key_server_to_provider_kmip')
    @mock.patch.object(functional_models.Cluster, 'get_key_provider_kmip')
    def test_setup_external_key_management(self, mock_get_kp, mock_add, mock_create_key_server,
                                           mock_create_csr, mock_create_kp, mock_create_key_pair):
        # EKS DMs
        mock_eks_dms = [sf_models.ExternalKeyServer.get_mock_object()]

        # Key Provider
        mock_key_provider = models.KeyProviderKmip.get_mock_object()
        mock_create_kp.return_value = mock_key_provider
        mock_get_kp.return_value = mock_key_provider

        # CSR
        mock_csr_result = sf_api.result_models.GetClientCertificateSignRequest.get_mock_object()
        mock_create_csr.return_value = mock_csr_result

        # Key Server
        mock_create_key_server.return_value = models.KeyServerKmip.get_mock_object()

        key_provider_dm = vl_helpers.setup_external_key_management(self.mock_cluster_fm, mock_eks_dms)
        assert key_provider_dm == mock_key_provider

    @mock.patch.object(functional_models.Cluster, 'create_public_private_key_pair')
    def test_create_key_pair_good(self, mock_create_key):
        vl_helpers._security._create_key_pair(self.mock_cluster_fm, None, True, False)

    @mock.patch.object(functional_models.Cluster, 'create_public_private_key_pair')
    def test_create_key_pair_reuse(self, mock_create_key):
        mock_error = api_errors.KeyPairInUse('xKeyPairInUse', mock.Mock())
        mock_create_key.side_effect = mock_error
        vl_helpers._security._create_key_pair(self.mock_cluster_fm, None, True, False)

    @mock.patch.object(functional_models.Cluster, 'create_public_private_key_pair')
    @mock.patch.object(vl_helpers._security, 'teardown_external_key_management')
    def test_create_key_pair_force(self, mock_teardown, mock_create_key):
        mock_error = api_errors.KeyPairInUse('xKeyPairInUse', mock.Mock())
        mock_create_key.side_effect = [mock_error, 1]
        vl_helpers._security._create_key_pair(self.mock_cluster_fm, None, False, True)

    @mock.patch.object(functional_models.Cluster, 'create_public_private_key_pair')
    def test_create_key_pair_raise(self, mock_create_key):
        mock_error = api_errors.KeyPairInUse('xKeyPairInUse', mock.Mock())
        mock_create_key.side_effect = mock_error

        with pytest.raises(api_errors.KeyPairInUse):
            vl_helpers._security._create_key_pair(self.mock_cluster_fm, None, False, False)

    @mock.patch.object(functional_models.HyTrustExternalKeyServer, 'sign_csr')
    @mock.patch.object(functional_models.Cluster, 'create_key_server_kmip')
    @mock.patch.object(x509, 'load_pem_x509_csr')
    def test_create_key_server(self, mock_load, mock_create, mock_sign_csr):
        dm1 = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.hytrust))
        dm2 = sf_models.ExternalKeyServer.get_mock_object(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.hytrust))
        dm_cluster = sf_models.ExternalKeyServerCluster.get_mock_object(overrides=dict(nodes=[dm1, dm2]))

        # Sign CSR
        csr = mock.MagicMock(x509.CertificateSigningRequest)
        mock_load.return_value = csr
        cert1 = mock.MagicMock(x509.Certificate)
        cert1.public_bytes.return_value = 'xxx'
        cert2 = mock.MagicMock(x509.Certificate)
        cert2.public_bytes.return_value = 'yyy'
        mock_sign_csr.return_value = (cert1, cert2)

        vl_helpers.create_key_server(self.mock_cluster_fm, dm1, 'csr', 'tag')
        vl_helpers.create_key_server(self.mock_cluster_fm, dm_cluster, 'csr', None)

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'is_running')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_off')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'teardown_cryptsoft_server')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'deploy_cryptsoft_server')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'power_on')
    def test_deploy_cryptsoft_eks(self, mock_power_on, mock_deploy, mock_teardown, mock_power_off,
                                  mock_is_running):
        eks_fm = functional_models.CryptsoftExternalKeyServer.get_mock_fm()
        mock_is_running.return_value = True
        mock_teardown.side_effect = [sf_ssh.SshError]
        vl_helpers.deploy_cryptsoft_eks(eks_fm)
        mock_deploy.assert_called()

    def test_get_eks_hostnames(self):
        eks_dm_single = sf_models.ExternalKeyServer.get_mock_object()
        assert vl_helpers.get_eks_hostnames(eks_dm_single) == [eks_dm_single.hostname]

        eks_dm_cluster = sf_models.ExternalKeyServerCluster.get_mock_object()
        clustered_hostnames = [node.hostname for node in eks_dm_cluster.nodes]
        assert vl_helpers.get_eks_hostnames(eks_dm_cluster) == clustered_hostnames


@mock.patch('valence.vl_helpers._security.is_encryption_at_rest_enabled')
@mock.patch.object(vl_helpers._security, 'disable_encryption_at_rest')
@mock.patch.object(functional_models.Cluster, 'list_key_providers_kmip')
@mock.patch.object(functional_models.Cluster, 'list_key_servers_kmip')
@mock.patch.object(functional_models.Cluster, 'remove_key_server_from_provider_kmip')
@mock.patch.object(functional_models.Cluster, 'delete_key_server_kmip')
@mock.patch.object(functional_models.Cluster, 'delete_key_provider_kmip')
@mock.patch.object(utils, 'timeout_timer')
class TestEKMTeardown(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        mock_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.mock_cluster_fm = functional_models.Cluster(mock_cluster_dm, authenticate=False)

    def test_teardown_external_key_management(self, mock_timer, mock_delete_kp, mock_delete_ks,
                                              mock_remove, mock_list_ks, mock_list_kp, mock_disable,
                                              mock_is_enabled):
        # EKS
        eks_dms = [sf_models.ExternalKeyServerCluster.get_mock_object()]
        mock_is_enabled.return_value = True

        # Key Provider
        mock_kp = models.KeyProviderKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_ids=[3]))
        mock_list_kp.return_value = [mock_kp]

        # Key Server
        mock_ks = models.KeyServerKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_id=3,
                           kmip_key_server_hostnames=[n.hostname for n in eks_dms[0].nodes],
                           kmip_assigned_provider_is_active=False))
        mock_list_ks.return_value = [mock_ks]
        mock_timer.return_value = iter([True, True])

        # Teardown
        vl_helpers.teardown_external_key_management(self.mock_cluster_fm, eks_dms=eks_dms,
                                                    disable_ear=True)
        vl_helpers.teardown_external_key_management(self.mock_cluster_fm, eks_dms=None,
                                                    disable_ear=True)

    def test_teardown_external_key_management_leftover(self, mock_timer, mock_delete_kp,
                                                       mock_delete_ks, mock_remove, mock_list_ks,
                                                       mock_list_kp, mock_disable, mock_is_enabled):
        # EKS
        eks_dms = [sf_models.ExternalKeyServer.get_mock_object()]
        mock_is_enabled.return_value = False

        # Key Provider
        mock_kp = models.KeyProviderKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_ids=[3]))
        mock_list_kp.return_value = [mock_kp]

        # Key Servers
        mock_ks1 = models.KeyServerKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_id=3,
                           kmip_key_server_hostnames=[eks_dms[0].hostname],
                           kmip_assigned_provider_is_active=False))
        mock_ks2 = models.KeyServerKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_id=3, kmip_assigned_provider_is_active=False))
        mock_list_ks.return_value = [mock_ks1, mock_ks2]
        mock_timer.return_value = iter([True, False])

        # Teardown
        vl_helpers.teardown_external_key_management(self.mock_cluster_fm, eks_dms=eks_dms)

    def test_teardown_external_key_management_raise(self, mock_timer, mock_delete_kp, mock_delete_ks,
                                                    mock_remove, mock_list_ks, mock_list_kp,
                                                    mock_disable, mock_is_enabled):
        # EKS
        eks_dms = [sf_models.ExternalKeyServer.get_mock_object()]
        mock_is_enabled.return_value = True
        mock_timer.return_value = iter([True])

        # Teardown
        with pytest.raises(_errors.HelperError):
            vl_helpers.teardown_external_key_management(self.mock_cluster_fm, eks_dms=eks_dms)

    def test_teardown_external_key_management_raise_timeout(self, mock_timer, mock_delete_kp,
                                                            mock_delete_ks, mock_remove, mock_list_ks,
                                                            mock_list_kp, mock_disable,
                                                            mock_is_enabled):
        # EKS
        eks_dms = [sf_models.ExternalKeyServer.get_mock_object()]
        mock_is_enabled.return_value = False

        # Key Provider
        mock_kp = models.KeyProviderKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_ids=[3]))
        mock_list_kp.return_value = [mock_kp]

        # Key Servers
        mock_ks1 = models.KeyServerKmip.get_mock_object(
            overrides=dict(key_provider_id=2, key_server_id=3,
                           kmip_key_server_hostnames=[eks_dms[0].hostname],
                           kmip_assigned_provider_is_active=True))
        mock_list_ks.return_value = [mock_ks1]
        mock_timer.return_value = iter([True, False])

        with pytest.raises(errors.TimeoutError):
            vl_helpers.teardown_external_key_management(self.mock_cluster_fm, eks_dms=eks_dms)


class TestEARHelpers(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mock_cluster_dm = models.ResourceCluster.get_mock_object()
        cls.mock_cluster_fm = functional_models.Cluster(cls.mock_cluster_dm, authenticate=False)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_info')
    @mock.patch.object(functional_models.Cluster, 'get_current_version')
    def test_is_encryption_at_rest_enabled(self, mock_current_version, mock_cluster_info):
        mock_current_version.return_value = models.SfVersion('11.0')
        mock_cluster_info.return_value = models.Cluster.get_mock_object(
            overrides=dict(encryption_at_rest_state='enabled'))
        assert vl_helpers.is_encryption_at_rest_enabled(self.mock_cluster_fm)

    @mock.patch.object(vl_helpers._bin_sync, 'is_bin_syncing')
    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(vl_helpers._bin_sync, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'enable_encryption_at_rest')
    @mock.patch.object(functional_models._async_handler.EnableEarAsyncHandler, 'get_models')
    def test_enable_encryption_at_rest(self, mock_get_models, mock_enable, mock_wait,
                                       mock_get_sync_timeout, mock_is_sync):
        mock_wait.return_value = True
        vl_helpers.enable_encryption_at_rest(self.mock_cluster_fm)

    @mock.patch.object(vl_helpers._bin_sync, 'is_bin_syncing')
    @mock.patch.object(vl_helpers._bin_sync, 'get_bin_sync_timeout')
    @mock.patch.object(vl_helpers._bin_sync, 'wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'disable_encryption_at_rest')
    @mock.patch.object(functional_models._async_handler.DisableEarAsyncHandler, 'get_models')
    def test_disable_encryption_at_rest(self, mock_get_models, mock_disable, mock_wait,
                                        mock_get_sync_timeout, mock_is_sync):
        mock_wait.return_value = True
        vl_helpers.disable_encryption_at_rest(self.mock_cluster_fm)

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_corrupt_shamir_secret_keys(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='gen-1', stderr='', return_code=0),
            # Check for copy
            sf_ssh.RunCommandOutput(stdout='ls: cannot access \'gen-1\': No such file or directory',
                                    stderr='', return_code=2),
            # Copy file
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
            # Corrupt file
            sf_ssh.RunCommandOutput(stdout='2+0 records in\n2+0 records out\n4 bytes copied, 16.9 kB/s',
                                    stderr='', return_code=0),
        ]
        vl_helpers.corrupt_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_corrupt_shamir_copy_exists(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
        ]
        with pytest.raises(_errors.HelperError):
            vl_helpers.corrupt_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_corrupt_shamir_no_gen_file(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='gen-1', stderr='', return_code=0),
            # Check for copy
            sf_ssh.RunCommandOutput(stdout='gen-1.original',
                                    stderr='', return_code=0),
        ]
        with pytest.raises(_errors.HelperError):
            vl_helpers.corrupt_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_move_shamir_secret_keys(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='gen-1', stderr='', return_code=0),
            # Check for copy
            sf_ssh.RunCommandOutput(stdout='ls: cannot access \'gen-1\': No such file or directory',
                                    stderr='', return_code=2),
            # Move file
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
        ]
        vl_helpers.move_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_move_shamir_copy_exists(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
        ]
        with pytest.raises(_errors.HelperError):
            vl_helpers.move_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_move_shamir_no_gen_file(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='gen-1', stderr='', return_code=0),
            # Check for copy
            sf_ssh.RunCommandOutput(stdout='gen-1.original',
                                    stderr='', return_code=0),
        ]
        with pytest.raises(_errors.HelperError):
            vl_helpers.move_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_restore_shamir_secret_keys(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='gen-1.original', stderr='', return_code=0),
            # Move file
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
        ]
        vl_helpers.restore_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    def test_restore_shamir_no_original(self, mock_ssh):
        node_dm = models.ActiveNode.get_mock_object()
        node_fm = functional_models.Node(node_dm, self.mock_cluster_dm)
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            # Get latest
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0),
        ]
        vl_helpers.restore_shamir_secret_keys([node_fm])

    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'get_key')
    @mock.patch.object(functional_models.CryptsoftExternalKeyServer, 'set_key')
    def test_corrupt_cryptsoft_key(self, mock_set, mock_get):
        mock_key = 'zuq935ff25e41c92dbffedf333ea16d22df88f5a5c4e99c6bddedeac792404cb'
        mock_get.return_value = mock_key
        eks_fm = functional_models.CryptsoftExternalKeyServer.get_mock_fm(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.cryptsoft))
        key = vl_helpers.corrupt_cryptsoft_key(eks_fm, '680c8cec-ff90-4eeb-a890-36a2557176bd')
        assert key == mock_key

    def test_corrupt_cryptsoft_key_exception(self):
        eks_fm = functional_models.CryptsoftExternalKeyServer.get_mock_fm(
            overrides=dict(vendor=enums.ExternalKeyServerVendor.hytrust))
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.corrupt_cryptsoft_key(eks_fm, '00000000-00000-00000000')

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_verify_shamir_secret_keys(self, mock_get_fms, mock_ear_info, mock_ssh):
        mock_ear_info.return_value = models.EncryptionAtRestInfo.get_mock_object(
            overrides=dict(generation_number=1))
        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=1),
        ]
        assert not vl_helpers.verify_shamir_secret_keys(self.mock_cluster_fm)

    @mock.patch.object(functional_models.Node, 'get_ssh_connection', autospec=True)
    @mock.patch.object(functional_models.Cluster, 'get_encryption_at_rest_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_verify_shamir_secret_keys_no_ear(self, mock_get_fms, mock_ear_info, mock_ssh):
        mock_ear_info.return_value = models.EncryptionAtRestInfo.get_mock_object(
            overrides=dict(generation_number=0))
        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]
        run_safe = mock_ssh.return_value.__enter__.return_value.run_command
        run_safe.side_effect = [
            sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=1),
        ]
        assert vl_helpers.verify_shamir_secret_keys(self.mock_cluster_fm)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_info')
    def test_wait_for_ear_state(self, mock_info):
        mock_clus_info_result = result_models.GetClusterInfo.get_mock_object()
        mock_clus_info_result.cluster_info.encryption_at_rest_state = enums.EarState.enabled

        mock_clus_info_result_disabled = result_models.GetClusterInfo.get_mock_object()
        mock_clus_info_result_disabled.cluster_info.encryption_at_rest_state = enums.EarState.disabled

        mock_info.side_effect = [
            mock_clus_info_result_disabled.cluster_info,
            mock_clus_info_result.cluster_info
        ]
        assert vl_helpers.wait_for_ear_state(self.mock_cluster_fm, enums.EarState.enabled)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_info')
    def test_wait_for_ear_state_timeout(self, mock_info):
        mock_clus_info_result = result_models.GetClusterInfo.get_mock_object()
        mock_clus_info_result.cluster_info.encryption_at_rest_state = enums.EarState.disabled
        mock_info.return_value = mock_clus_info_result.cluster_info
        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_ear_state(self.mock_cluster_fm, enums.EarState.enabled, timeout=0)

    @mock.patch.object(functional_models.Node, 'get_hardware_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_security_state(self, mock_list_drives, mock_get_fms, mock_hw_info):
        mock_drive_info = models.DriveInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_enabled=True,
                           security_locked=True))

        mock_drive_info_secure_disabled = models.DriveInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_enabled=False,
                           security_locked=True))

        mock_hw_info.side_effect = [
            models.NodeHardwareInfo.get_mock_object(
                overrides=dict(drive_hardware=[mock_drive_info_secure_disabled])),
            models.NodeHardwareInfo.get_mock_object(
                overrides=dict(drive_hardware=[mock_drive_info])),
        ]

        mock_list_drives.return_value = [
            models.Drive.get_mock_object(overrides=dict(
                serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey', slot=1)),
            models.Drive.get_mock_object(overrides=dict(
                serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey', slot=-1)),
        ]

        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]

        assert vl_helpers.wait_for_drive_security_state(self.mock_cluster_fm, True)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_hardware_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_cluster_drive_security_state(self, mock_list_drives, mock_get_fms, mock_hw_info):
        mock_drive_info = models.DriveHardwareInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_feature_enabled=True,
                           drive_security_locked=True))

        mock_drive_info_secure_disabled = models.DriveHardwareInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_feature_enabled=False,
                           drive_security_locked=True))

        mock_hw_info.side_effect = [
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info_secure_disabled})
            ),
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info})
            ),
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info_secure_disabled})
            ),
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info})
            ),
        ]

        mock_list_drives.return_value = [
            models.Drive.get_mock_object(overrides=dict(
                serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey', slot=1)),
            models.Drive.get_mock_object(overrides=dict(
                serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey', slot=-1)),
        ]

        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]

        assert vl_helpers.wait_for_drive_security_state(
            self.mock_cluster_fm, True, api_method=enums.ClusterMethod.get_cluster_hardware_info)
        assert vl_helpers.wait_for_drive_security_state(
            self.mock_cluster_fm,
            security_state=True,
            node_dms=[models.ActiveNode.get_mock_object()],
            api_method=enums.ClusterMethod.get_cluster_hardware_info
        )

    @mock.patch.object(functional_models.Node, 'get_hardware_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_security_state_timeout(self, mock_list_drives, mock_get_fms, mock_hw_info):
        mock_drive_info = models.DriveInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_enabled=False,
                           security_locked=True))
        mock_hw_info.return_value = models.NodeHardwareInfo.get_mock_object(
            overrides=dict(drive_hardware=[mock_drive_info]))

        mock_list_drives.return_value = [models.Drive.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey'))]

        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]

        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_security_state(self.mock_cluster_fm, True, timeout=0)

    @mock.patch.object(functional_models.Node, 'get_hardware_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_locked_state(self, mock_list_drives, mock_get_fms, mock_hw_info):
        mock_drive_info = models.DriveInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_enabled=False,
                           security_locked=True))

        mock_drive_info_unlocked = models.DriveInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_enabled=False,
                           security_locked=False))

        mock_hw_info.side_effect = [
            models.NodeHardwareInfo.get_mock_object(
                overrides=dict(drive_hardware=[mock_drive_info_unlocked])),
            models.NodeHardwareInfo.get_mock_object(
                overrides=dict(drive_hardware=[mock_drive_info]))
        ]

        mock_list_drives.return_value = [
            models.Drive.get_mock_object(overrides=dict(
                serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey', slot=1)),
            models.Drive.get_mock_object(overrides=dict(
                serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey', slot=-1)),
        ]

        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]

        assert vl_helpers.wait_for_drive_locked_state(self.mock_cluster_fm, True)

    @mock.patch.object(functional_models.Cluster, 'get_cluster_hardware_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_cluster_drive_locked_state(self, mock_list_drives, mock_get_fms, mock_hw_info):
        mock_drive_info = models.DriveHardwareInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_feature_enabled=False,
                           drive_security_locked=True))

        mock_drive_info_unlocked = models.DriveHardwareInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_feature_enabled=False,
                           drive_security_locked=False))

        mock_hw_info.side_effect = [
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info_unlocked})
            ),
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info})
            ),
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info_unlocked})
            ),
            models.ClusterHardwareInfo.get_mock_object(
                overrides=dict(drives={'1': mock_drive_info})
            ),
        ]

        mock_list_drives.return_value = [models.Drive.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey'))]

        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]

        assert vl_helpers.wait_for_drive_locked_state(
            self.mock_cluster_fm, True, api_method=enums.ClusterMethod.get_cluster_hardware_info)
        assert vl_helpers.wait_for_drive_locked_state(
            self.mock_cluster_fm,
            locked_state=True,
            node_dms=[models.ActiveNode.get_mock_object()],
            api_method=enums.ClusterMethod.get_cluster_hardware_info
        )

    @mock.patch.object(functional_models.Node, 'get_hardware_info')
    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    @mock.patch.object(functional_models.Cluster, 'list_drives')
    def test_wait_for_drive_locked_state_timeout(self, mock_list_drives, mock_get_fms, mock_hw_info):
        mock_drive_info = models.DriveInfo.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', security_enabled=False,
                           security_locked=False))
        mock_hw_info.return_value = models.NodeHardwareInfo.get_mock_object(
            overrides=dict(drive_hardware=[mock_drive_info]))

        mock_list_drives.return_value = [models.Drive.get_mock_object(
            overrides=dict(serial='S3T1NX0K702742', drive_security_fault_reason='SecureEnableNoKey'))]

        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]

        with pytest.raises(errors.TimeoutError):
            vl_helpers.wait_for_drive_locked_state(self.mock_cluster_fm, True, timeout=0)

    def test_unsupported_api_method(self):
        with pytest.raises(_errors.HelperError):
            vl_helpers._security._get_security_state(self.mock_cluster_fm,
                                                     api_method=enums.NodeMethod.get_hardware_config)
        with pytest.raises(_errors.HelperError):
            vl_helpers._security._get_locked_state(self.mock_cluster_fm,
                                                   api_method=enums.NodeMethod.get_hardware_config)

    @mock.patch.object(functional_models.Cluster, 'get_node_fms')
    def test_filter_nodes(self, mock_get_fms):
        mock_node_fm = functional_models.Node.get_mock_fm()
        mock_get_fms.return_value = [mock_node_fm]
        mock_nodes = models.ActiveNode.get_mock_object_list(length=1)
        vl_helpers._security._filter_nodes(self.mock_cluster_fm, node_dms=mock_nodes)

        mock_resource_nodes = models.ResourceNode.get_mock_object_list(length=1)
        vl_helpers._security._filter_nodes(self.mock_cluster_fm, node_dms=mock_resource_nodes)

    def test_filter_drive_info(self):
        mock_hw_info = models.DriveInfo.get_mock_object_list(length=1)
        mock_drive_dms = models.Drive.get_mock_object_list(length=1)
        vl_helpers._security._filter_drive_info(mock_hw_info, mock_drive_dms)
