"""Tests for sfupdate-healthtools"""

import time
import unittest

import mock
import pytest
import sf_ssh

from valence import (
    functional_models,
    models,
    unittest_utils,
    vl_helpers,
)

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION = 'Currently installed version of HealthTools:'
HEALTHTOOLS_DOWNLOAD_INSTRUCTIONS = 'To download the latest version of the HealthTools, go to:'
MGMT_PLTFRM = 'Management Platform Node: Ember'

STD_ERR = """
Currently installed version of HealthTools:
2019.06.01.105
Management Platform Node: Ember
The latest version of HealthTools can be downloaded from NetApp Support Site
Go to https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""

STD_OUT = """
Currently installed version of HealthTools:
2019.06.01.102
Management Platform Node: Ember
To download the latest version of the HealthTools, go to:
https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""

STD_OUT_SFUPL = """
Latest available version of HealthTools:
2019.06.01.107
Management Platform Node: Ember
The latest version of HealthTools can be downloaded from NetApp Support Site
Go to https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""

STD_OUT_HT_19_2 = """
Currently installed version of HealthTools:
2019.12.01.0
Management Platform Node: Ember
"""

STD_ERR_OUTPUT_LIST_LEN = """
Latest available version of HealthTools:
2019.06.01.107
Currently installed version of HealthTools:
Management Platform Node: Ember
"""


class TestSFUpdateHealthTools(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        res_mnode = models.ResourceManagementNode().get_mock_object(include_optional=True)
        build_dm = models.Build.create(build_id=1, release='sodium', element_id=11, version='11.3.0.0',
                                       repository_name='sodium-patch3')
        res_mnode['build_version'] = build_dm
        res_mnode['username'] = 'Valentino'
        res_mnode['password'] = 'Rossi'
        cls.fdva_fm = functional_models.ManagementNodeLegacy.from_resource_data(res_mnode)
        cls.nodes_dm = [models.ActiveNode.get_mock_object(overrides={'node_id': i})
                        for i in range(1, 5)]

    def test_match_sfupdate_healthtools_output_to_regex(self):
        version = models.GenericVersion('2019.06.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.match_sfupdate_healthtools_output_to_regex(
                output_to_parse=STD_OUT,
                token_str=CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION,
                expected_versions_list=[version],
                management_platform_string=vl_helpers.get_expected_platform_string_for_healthtools(
                    self.fdva_fm.version, version))

    def test_match_sfupdate_healthtools_output_to_regex_download_instr_present(self):
        version = models.GenericVersion('2019.06.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.match_sfupdate_healthtools_output_to_regex(
                output_to_parse=STD_OUT,
                token_str=CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION,
                expected_versions_list=[version],
                download_instructions_str=HEALTHTOOLS_DOWNLOAD_INSTRUCTIONS
            )

    def test_match_sfupdate_healthtools_output_to_regex_no_version_found(self):
        version = models.GenericVersion('2019.06.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.match_sfupdate_healthtools_output_to_regex(
                output_to_parse=STD_ERR,
                token_str=CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION,
                expected_versions_list=[version])

    def test_match_sfupdate_healthtools_output_to_regex_except(self):
        version = models.GenericVersion('2019.06.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.match_sfupdate_healthtools_output_to_regex(
                output_to_parse=STD_OUT_SFUPL,
                token_str=CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION,
                expected_versions_list=[version])

    def test_match_sfupdate_healthtools_output_to_regex_give_mgmt_string(self):
        version = models.GenericVersion('2019.12.01.0')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.match_sfupdate_healthtools_output_to_regex(
                output_to_parse=STD_OUT_HT_19_2,
                token_str=CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION,
                expected_versions_list=[version],
                management_platform_string=vl_helpers.get_expected_platform_string_for_healthtools(
                    self.fdva_fm.version, version)
            )

    def test_match_sfupdate_healthtools_output_to_regex_output_list_len(self):
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.match_sfupdate_healthtools_output_to_regex(
                output_to_parse=STD_ERR_OUTPUT_LIST_LEN,
                token_str=CURRENTLY_INSTALLED_HEALTHTOOLS_VERSION,
                expected_versions_list=[])
