import time
import unittest

import mock
import pytest
import sf_ssh

from valence import (
    functional_models,
    models,
    unittest_utils,
    vl_helpers,
)

module_mock = unittest_utils.ModuleLevelMock()


def setUpModule():
    # Mock Ssh.close and Ssh._init_client globally for the module so we don't need to mock the
    # Ssh context manager and can mock the Ssh methods directly.
    module_mock.mock(sf_ssh.Ssh, 'close')
    module_mock.mock(sf_ssh.Ssh, '_init_client')

    # Sleep no more
    module_mock.mock(time, 'sleep')


def tearDownModule():
    module_mock.restore()


STD_ERR = """
Uploading version 2019.06.01.102 to node nlabp2814
Uploading version 2019.06.01.102 to node nlabp2815
Uploading version 2019.06.01.102 to node nlabp2816
Uploading version 2019.06.01.102 to node nlabp2813
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 1
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 2
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 3
Uploading solidfire-sfinstalld-2017.10.10.103.tgz to nodeID 4
"""

STD_OUT = """
Uploading version 2019.01.01.131 to node nlabp2814
Uploading version 2019.01.01.131 to node nlabp2815
Uploading version 2019.01.01.131 to node nlabp2816
Uploading version 2019.01.01.131 to node nlabp2813
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 1
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 2
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 3
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 4
"""

STD_OUT_WEIRD_NODE = """
Uploading version 2019.01.01.131 to node nlabp2817
Uploading version 2019.01.01.131 to node nlabp2815
Uploading version 2019.01.01.131 to node nlabp2816
Uploading version 2019.01.01.131 to node nlabp2813
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 8
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 2
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 3
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 4
"""

STD_OUT_EXP = """
Uploading version 2019.12.03.101 to node nlabp2814
Uploading version 2019.12.03.101 to node nlabp2815
Uploading version 2019.12.03.101 to node nlabp2816
Uploading version 0 to node nlabp2813
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 1
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 2
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 3
Uploading solidfire-sfinstalld-2017.10.10.102.tgz to nodeID 0
"""

SFUP_V_STD_OUT = """
Latest version of upgrade tools available: 2018.12.01.107
Installed version of upgrade tools on management node: 2018.12.01.102
Management Node Platform: Ember
Node 1 is on version 2018.12.01.102
HealthTools is out of date: The latest version of HealthTools can be downloaded from NetApp Support Site
Go to https://mysupport.netapp.com/NOW/dgi-bin/software/,
and select HealthTools to download the latest version
"""


class TestOutputVersions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        res_mnode = models.ResourceManagementNode().get_mock_object(include_optional=True)
        build_dm = models.Build.create(build_id=1, release='sodium', element_id=11, version='11.3.0.0',
                                       repository_name='sodium-patch3')
        res_mnode['build_version'] = build_dm
        res_mnode['username'] = 'Valentino'
        res_mnode['password'] = 'Rossi'
        cls.mnode_fm = functional_models.ManagementNodeLegacy(res_mnode)
        cls.cluster_dm = models.ResourceCluster.get_mock_object()
        cls.node_dms = [models.ActiveNode.get_mock_object(overrides={'node_id': 0, 'name': 'nlabp2813'}),
                        models.ActiveNode.get_mock_object(overrides={'node_id': 1, 'name': 'nlabp2814'}),
                        models.ActiveNode.get_mock_object(overrides={'node_id': 2, 'name': 'nlabp2815'}),
                        models.ActiveNode.get_mock_object(overrides={'node_id': 3, 'name': 'nlabp2816'})]

    def test_validate_sfupgradecheck_download_expected_version_less_than_sodium(self):
        """Validates sfupgradecheck download."""
        old_version = models.GenericVersion('2017.10.10.102')
        old_fdva_version = models.SfVersion('10.3.0.0')

        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_download(STD_ERR, old_version,
                                                        self.node_dms, old_fdva_version)

    def test_validate_sfupgradecheck_download_expected_version_greater_than_sodium(self):
        new_version = models.GenericVersion('2019.01.01.131')
        new_fdva_version = models.SfVersion('11.3.0.0')
        vl_helpers.validate_sfupgradecheck_download(STD_OUT, new_version,
                                                    self.node_dms, new_fdva_version)

    def test_validate_sfupgradecheck_download_expected_version_not_all_nodes(self):
        new_version = models.GenericVersion('2019.01.01.131')
        new_fdva_version = models.SfVersion('11.3.0.0')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_download(STD_OUT_WEIRD_NODE, new_version,
                                                        self.node_dms, new_fdva_version)

    def test_validate_sfupgradecheck_download_raise_check_equality(self):
        new_version = models.GenericVersion('2019.12.03.101')
        new_fdva_version = models.SfVersion('11.3.0.0')

        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_download(STD_OUT_EXP, new_version,
                                                        self.node_dms, new_fdva_version)

    @mock.patch.object(functional_models.ManagementNodeLegacy, 'get_ssh_connection')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_validate_sfupgradecheck_v_output(self, mock_run_command, mock_ssh_conn):
        installed_ht_version = models.GenericVersion('2018.12.01.105')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_v_output(
                SFUP_V_STD_OUT, expected_installed_ht_version=installed_ht_version)

    @mock.patch.object(functional_models.ManagementNodeLegacy, 'get_ssh_connection')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_validate_sfupgradecheck_v_output_check_latest_v(self, mock_run_command,
                                                             mock_ssh_conn):
        installed_ht_version = models.GenericVersion('2018.12.01.102')
        latest_version = models.GenericVersion('2017.12.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_v_output(
                SFUP_V_STD_OUT, expected_latest_version=latest_version,
                expected_installed_ht_version=installed_ht_version)

    @mock.patch.object(functional_models.ManagementNodeLegacy, 'get_ssh_connection')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_validate_sfupgradecheck_v_output_check_node_v(self, mock_run_command,
                                                           mock_ssh_conn):
        installed_ht_version = models.GenericVersion('2018.12.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_v_output(
                SFUP_V_STD_OUT, expected_installed_ht_version=installed_ht_version,
                expected_node_versions={2: models.GenericVersion('2018.12.01.105')})

    @mock.patch.object(functional_models.ManagementNodeLegacy, 'get_ssh_connection')
    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_validate_sfupgradecheck_v_node_output(self, mock_run_command,
                                                   mock_ssh_conn):
        installed_ht_version = models.GenericVersion('2018.12.01.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.validate_sfupgradecheck_v_output(
                SFUP_V_STD_OUT, expected_installed_ht_version=installed_ht_version,
                expected_node_versions={2: models.GenericVersion('2018.12.01.105')})

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_deploy_same_version_to_node(self, mock_run):
        version = models.GenericVersion('2017.10.10.102')
        mock_run.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        vl_helpers.deploy_same_version_to_nodes(self.mnode_fm, self.cluster_dm, version, self.node_dms)
        assert '--deploy-only 0:2017.10.10.102,1:2017.10.10.102,2:2017.10.10.102,3:2017.10.10.102' in \
               mock_run.call_args_list[0][0][0]

    def test_deploy_same_version_to_node_version_is_none(self):
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.deploy_same_version_to_nodes(self.mnode_fm, self.cluster_dm, None, self.node_dms)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_deploy_different_version_to_node(self, mock_run):
        version_list = [models.GenericVersion('2017.10.10.102'), models.GenericVersion('2019.01.01.131'),
                        models.GenericVersion('2018.12.01.107'), models.GenericVersion('2018.12.01.108')]
        mock_run.return_value = sf_ssh.RunCommandOutput(stdout='', stderr='', return_code=0)
        vl_helpers.deploy_different_versions_to_nodes(self.mnode_fm, self.cluster_dm, version_list,
                                                      self.node_dms)
        assert mock_run.called
        assert '--deploy-only 0:2017.10.10.102,1:2019.01.01.131,2:2018.12.01.107,3:2018.12.01.108' in \
               mock_run.call_args_list[0][0][0]

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_deploy_different_version_to_node_unequal_versions(self, mock_run):
        version = models.GenericVersion('2017.10.10.102')
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.deploy_different_versions_to_nodes(self.mnode_fm, self.cluster_dm, [version],
                                                          self.node_dms)

    @mock.patch.object(sf_ssh.Ssh, 'run_command')
    def test_deploy_different_version_to_node_version_is_none(self, mock_run):
        with pytest.raises(vl_helpers.HelperError):
            vl_helpers.deploy_different_versions_to_nodes(self.mnode_fm, self.cluster_dm, None,
                                                          self.node_dms)
