import os
import unittest

import pytest

from valence import (
    enums,
    models,
)
from valence.functional_models import (
    ProcessedHealthchecks,
    _constants,
)
from valence.vl_helpers import (
    _errors,
    _sfupgradecheck_validation,
)

CWD, _ = os.path.split(os.path.abspath(os.path.expanduser(__file__)))
DATA = os.path.join(CWD, 'data')
CHECKS = os.path.join(DATA, 'checks.json')
CHECKS_10_3_NO_CONN = os.path.join(DATA, 'checks_10_3_no_conn_speed.json')
CHECKS_10_3_ALL_CHECKS = os.path.join(DATA, 'checks_10_3_all_checks.json')
CHECKS_ALL_PASS = os.path.join(DATA, 'checks_all_pass.json')
CHECKS_CLUSTER_FAIL = os.path.join(DATA, 'checks_cluster_fail.json')
CHECKS_10_3_DATA_VALIDATION = os.path.join(DATA, 'checks_10_3_missing_data_validation.json')


class TestHealthcheckCriteria(unittest.TestCase):
    def setUp(self):
        """Setup test."""
        with open(CHECKS) as f:
            self.checks_text = f.read()

        with open(CHECKS_10_3_NO_CONN) as f:
            self.checks_10_3_text_no_conn = f.read()

        with open(CHECKS_ALL_PASS) as f:
            self.checks_all_pass_text = f.read()

        with open(CHECKS_CLUSTER_FAIL) as f:
            self.checks_cluster_fail_text = f.read()

        with open(CHECKS_10_3_ALL_CHECKS) as f:
            self.checks_10_3_all_pass_text = f.read()

        with open(CHECKS_10_3_DATA_VALIDATION) as f:
            self.checks_10_3_data_validation_text = f.read()

    @classmethod
    def setUpClass(cls):
        cls.nodes_dm = [models.ActiveNode.get_mock_object(overrides={'node_id': i})
                        for i in range(1, 5)]

    def test_all_pass(self):
        """Tests JSON where all checks pass."""
        checks = ProcessedHealthchecks.from_json(self.checks_all_pass_text)
        _sfupgradecheck_validation.validate_sfupgradecheck_results(
            _constants.NEON_PATCH2, checks, self.nodes_dm,
            _constants.HEALTHTOOLS_10_2_VERSION)

    def test_cluster_fail(self):
        """Tests JSON with failed cluster check."""
        checks = ProcessedHealthchecks.from_json(self.checks_cluster_fail_text)
        _sfupgradecheck_validation.validate_sfupgradecheck_cluster_results(
            _constants.NEON_PATCH2, checks,
            _constants.HEALTHTOOLS_10_2_VERSION,
            expected_failed_cluster_checks=[enums.HealthcheckName.check_pending_nodes])

    def test_cluster_fail_detected(self):
        """Test that we can detect cluster check failures."""
        with pytest.raises(_errors.HelperError):
            checks = ProcessedHealthchecks.from_json(self.checks_cluster_fail_text)
            _sfupgradecheck_validation.validate_sfupgradecheck_results(
                _constants.NEON_PATCH2, checks, self.nodes_dm,
                _constants.HEALTHTOOLS_10_2_VERSION)

    def test_not_all_cluster_checks_found(self):
        """Tests that 10.3 code accurately detects missing checks when run against 10.2."""
        with pytest.raises(_errors.HelperError):
            checks = ProcessedHealthchecks.from_json(self.checks_all_pass_text)
            _sfupgradecheck_validation.validate_sfupgradecheck_results(
                _constants.NEON_PATCH3, checks, self.nodes_dm,
                _constants.HEALTHTOOLS_10_3_VERSION)

        # Check that data["really"] field does not exists in the CHECK_CLUSTER_FAULTS check data dict.
        with pytest.raises(_errors.HelperError):
            checks = ProcessedHealthchecks.from_json(self.checks_10_3_all_pass_text)
            _sfupgradecheck_validation.validate_sfupgradecheck_results(
                _constants.NEON_PATCH3, checks, self.nodes_dm,
                _constants.HEALTHTOOLS_10_3_VERSION,
                {enums.HealthcheckName.check_cluster_faults: lambda check: 'really' in check.data})

    def test_10_3_no_connectivity_speed_check(self):
        """Tests that we can correctly detect a 10.3 JSON output with missing 10.3 tests."""
        with pytest.raises(_errors.HelperError):
            checks = ProcessedHealthchecks.from_json(self.checks_10_3_text_no_conn)
            _sfupgradecheck_validation.validate_sfupgradecheck_results(
                _constants.NEON_PATCH3, checks, self.nodes_dm,
                _constants.HEALTHTOOLS_10_3_VERSION)

    def test_all_10_3_checks(self):
        """Tests that all checks in 10.3 are present in 10.3 JSON blob"""
        # Check that data["available"] field exists in the CHECK_OPTIONS_FILE check data dict.
        with pytest.raises(_errors.HelperError):
            checks = ProcessedHealthchecks.from_json(self.checks_10_3_data_validation_text)
            _sfupgradecheck_validation.validate_sfupgradecheck_results(
                _constants.NEON_PATCH3, checks, self.nodes_dm,
                _constants.HEALTHTOOLS_10_3_VERSION,
                {enums.HealthcheckName.check_options_files: lambda check: 'available' in check.data})

    def test_all_pass_data_validation(self):
        # Check that data["really"] field does not exists in the CHECK_OPTIONS_FILE check data dict.
        with pytest.raises(_errors.HelperError):
            checks = ProcessedHealthchecks.from_json(self.checks_10_3_all_pass_text)
            _sfupgradecheck_validation.validate_sfupgradecheck_results(
                _constants.NEON_PATCH3, checks, self.nodes_dm,
                _constants.HEALTHTOOLS_10_3_VERSION,
                {enums.HealthcheckName.check_options_files: lambda check: 'really' in check.data})

    def test_print_helper(self):
        from valence.vl_helpers import _sfupgradecheck
        checks = ProcessedHealthchecks.from_json(self.checks_10_3_all_pass_text)
        _sfupgradecheck.prettyprint_sfupgradecheck_json_to_text(checks)
