import unittest

import mock
import pytest

import sf_api
from valence import enums  # SliceRole, enums.StandbyStates, enums.enums.ServiceType
from valence import (
    functional_models,
    models,
    vl_components,
    vl_helpers,
)
from valence.vl_helpers import _errors


class TestGetSliceDmFromVolume(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        slices = [{'volume_id': 1}, {'volume_id': 2}]
        cls.slice_report = models.SliceReport.get_mock_object(include_optional=True,
                                                              overrides=dict(slices=slices))

    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_slice_dm_from_volume(self, mock_get_slice_sync_report):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))
        mock_get_slice_sync_report.return_value = self.slice_report
        slice_dm = vl_helpers.get_slice_dm_from_volume(self.cluster_fm, volume_dm)
        assert slice_dm.volume_id == volume_dm.volume_id

    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_slice_dm_from_volume_last(self, mock_get_slice_sync_report):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=2))
        mock_get_slice_sync_report.return_value = self.slice_report
        slice_dm = vl_helpers.get_slice_dm_from_volume(self.cluster_fm, volume_dm)
        assert slice_dm.volume_id == volume_dm.volume_id

    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_slice_dm_from_volume_not_found(self, mock_get_slice_sync_report):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=3))
        mock_get_slice_sync_report.return_value = self.slice_report
        with pytest.raises(_errors.HelperError):
            vl_helpers.get_slice_dm_from_volume(self.cluster_fm, volume_dm)


class TestGetServiceId(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls._slice = {'volume_id': 1,
                      'primary': 2,
                      'live_secondaries': [3],
                      'dead_secondaries': [4, 7, 8],
                      'standby_services': [{'service_id': 5}, {'service_id': 6}]}
        cls.slice_dm = models.Slice.get_mock_object(include_optional=True,
                                                    overrides=cls._slice)

    def test_get_primary_service_id(self):
        sid = vl_helpers.get_service_ids(self.slice_dm, enums.SliceRole.primary)

        assert sid == [self._slice['primary']]

        sid = vl_helpers.get_service_ids(self.slice_dm, enums.SliceRole.live_secondary)
        assert sid == self._slice['live_secondaries']

        sid = vl_helpers.get_service_ids(self.slice_dm, enums.SliceRole.dead_secondary)
        assert sid == self._slice['dead_secondaries']

        sid = vl_helpers.get_service_ids(self.slice_dm, enums.SliceRole.standby)
        assert sid == [self._slice['standby_services'][0]['service_id'],
                       self._slice['standby_services'][1]['service_id']]

        with pytest.raises(ValueError):
            vl_helpers.get_service_ids(self.slice_dm, 'foo')


class TestCheckSliceHealth(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.default_slice = {'volume_id': 1,
                             'primary': 2,
                             'live_secondaries': [3],
                             'dead_secondaries': [],
                             'standby_services': [{'service_id': 5, 'state': enums.StandbyStates.active}]
                             }
        cls.default_slice_dm = models.Slice.get_mock_object(include_optional=True,
                                                            overrides=cls.default_slice)

    def test_check_slice_health_all_false(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4, 5],
                  'dead_secondaries': [2],
                  'standby_services': [{'service_id': 5}, {'service_id': 2}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue is None

    def test_check_slice_health_enable_all_checks(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5,
                                        'state': enums.StandbyStates.inactive}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=True,
            check_duplicates=True,
            check_standby_state=True,
            check_standby_duplicates=True)
        assert health_issue is None

    def test_check_degraded_healthy(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue is None

    def test_check_degraded_healthy_no_stable_live_secondaries_check(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4, 5],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue is None

    def test_check_degraded_health_issue(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_stable_lives_healthy(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=True,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue is None

    def test_check_stable_lives_health_issue_too_many_lives(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4, 5],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=True,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_stable_lives_health_issue_due_to_deads(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [5],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=True,
            check_stable_lives=True,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_stable_lives_degraded_check(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [5],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=True,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue is None

    def test_check_duplicates_primary_in_secondaries(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [2, 4],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)

        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=True,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_duplicates_secondary_in_deads(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [3, 5],
                  'standby_services': [{'service_id': 5}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=True,
            check_standby_state=False,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_standby_duplicates_healthy(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5}]}

        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=True)
        assert health_issue is None

    def test_check_standby_duplicates_primary_in_standbys(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 2}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=True)
        assert health_issue

    def test_check_standby_duplicates_secondary_in_standbys(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3, 4],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 4}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=True)
        assert health_issue

    def test_check_standby_duplicates_standby_in_deads(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [4],
                  'standby_services': [{'service_id': 4}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=False,
            check_standby_duplicates=True)
        assert health_issue

    def test_check_standby_state_ha_healthy(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 4, 'state': enums.StandbyStates.inactive}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=True,
            check_standby_duplicates=False)
        assert health_issue is None

    def test_check_standby_state_ha_and_failed_standby(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 4, 'state': enums.StandbyStates.failed}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=True,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_standby_state_too_many_standbys(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [],
                  'standby_services': [{'service_id': 5, 'state': enums.StandbyStates.active},
                                       {'service_id': 6, 'state': enums.StandbyStates.inactive}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=True,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_standby_state_no_standbys(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [3],
                  'dead_secondaries': [],
                  'standby_services': []}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=True,
            check_standby_duplicates=False)
        assert health_issue

    def test_check_standby_state_no_ha_ignore_standby_state_check(self):
        slice_ = {'volume_id': 1,
                  'primary': 2,
                  'live_secondaries': [],
                  'dead_secondaries': [4],
                  'standby_services': [{'service_id': 5, 'state': enums.StandbyStates.failed}]}
        slice_dm = models.Slice.get_mock_object(include_optional=True, overrides=slice_)
        health_issue = vl_helpers._slice_assignments._get_slice_health_issues(
            slice_dm,
            check_degraded=False,
            check_stable_lives=False,
            check_duplicates=False,
            check_standby_state=True,
            check_standby_duplicates=False)
        assert health_issue is None


class TestCheckSliceDistribution(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    def test_slice_distribution(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 50, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 60, 'state': enums.StandbyStates.inactive}]}]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},
                    {'service_id': 40},
                    {'service_id': 50},
                    {'service_id': 60}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        dist_issues = vl_helpers._slice_assignments._get_slice_distribution_issues(report_dm)
        assert dist_issues is None

    def test_slice_distribution_too_many_primaries(self):
        # TODO: impossible condition to even fake as written
        pass

    def test_slice_distribution_too_few_primaries(self):
        slices = [{'volume_id': 1,
                   'primary': 20,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 30, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},  # not primary anywhere
                    {'service_id': 20},
                    {'service_id': 30},
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        dist_issues = vl_helpers._slice_assignments._get_slice_distribution_issues(report_dm)
        assert dist_issues

    def test_slice_distribution_too_many_live_secondaries(self):
        # TODO: impossible condition to even fake as written
        pass

    def test_slice_distribution_too_few_live_secondaries(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 30, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},
                    {'service_id': 40}]  # not secondary anywhere
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        dist_issues = vl_helpers._slice_assignments._get_slice_distribution_issues(report_dm)
        assert dist_issues

    def test_slice_distribution_too_many_standbys(self):
        # TODO: impossible condition to even fake as written
        pass

    def test_slice_distribution_too_few_standbys(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        dist_issues = vl_helpers._slice_assignments._get_slice_distribution_issues(report_dm)
        assert dist_issues


class TestCheckSliceReportHealth(unittest.TestCase):

    def test_fail_one_slice_health_check(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 30, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20, 30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        with pytest.raises(_errors.HelperError):
            vl_helpers.check_slice_report_health(report_dm,
                                                 check_distribution=True,
                                                 check_degraded=True,
                                                 check_stable_lives=True,
                                                 check_duplicates=True,
                                                 check_standby_state=True,
                                                 check_standby_duplicates=True)

    def test_healthy_slices_and_distribution_check(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 30, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        vl_helpers.check_slice_report_health(report_dm,
                                             check_distribution=True,
                                             check_degraded=True,
                                             check_stable_lives=True,
                                             check_duplicates=True,
                                             check_standby_state=True,
                                             check_standby_duplicates=True)

    def test_fail_distribution_check(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        with pytest.raises(_errors.HelperError):
            health_issues = vl_helpers.check_slice_report_health(report_dm,
                                                                 check_distribution=True,
                                                                 check_degraded=True,
                                                                 check_stable_lives=True,
                                                                 check_duplicates=True,
                                                                 check_standby_state=True,
                                                                 check_standby_duplicates=True)

    def test_ignore_distribution_issues(self):
        slices = [{'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        vl_helpers.check_slice_report_health(report_dm,
                                             check_distribution=False,
                                             check_degraded=True,
                                             check_stable_lives=True,
                                             check_duplicates=True,
                                             check_standby_state=True,
                                             check_standby_duplicates=True)


class TestCheckSliceHealthForVolume(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_volume_slice_health_issue(self, mock_get_slice_sync_report):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=4))
        slices = [{'slice_id': 1,
                   'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 2,
                   'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 3,
                   'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 4,
                   'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [40, 30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)

        mock_get_slice_sync_report.return_value = report_dm
        with pytest.raises(_errors.HelperError):
            vl_helpers.check_slice_health_for_volume(self.cluster_fm, volume_dm,
                                                     standby_state=enums.StandbyStates.inactive,
                                                     check_degraded=True,
                                                     check_stable_lives=False,
                                                     check_standby_duplicates=False
                                                     )

    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_volume_standby_state_issue(self, mock_get_slice_sync_report):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=4))
        slices = [{'slice_id': 1,
                   'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 2,
                   'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 3,
                   'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 4,
                   'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)

        mock_get_slice_sync_report.return_value = report_dm
        with pytest.raises(_errors.HelperError):
            vl_helpers.check_slice_health_for_volume(self.cluster_fm, volume_dm,
                                                     standby_state=enums.StandbyStates.active,
                                                     check_degraded=True,
                                                     check_stable_lives=False,
                                                     check_standby_duplicates=False
                                                     )

    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_volume_no_standby_state_issue(self, mock_get_slice_sync_report):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=4))
        slices = [{'slice_id': 1,
                   'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 2,
                   'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 3,
                   'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 4,
                   'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)

        mock_get_slice_sync_report.return_value = report_dm
        vl_helpers.check_slice_health_for_volume(self.cluster_fm, volume_dm,
                                                 standby_state=enums.StandbyStates.inactive,
                                                 check_degraded=True,
                                                 check_stable_lives=False,
                                                 check_standby_duplicates=False
                                                 )


class CheckStandbyState(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dm = models.Volume.get_mock_object()

    @mock.patch('valence.vl_helpers._slice_assignments.check_slice_health_for_volume')
    def test_check_standby_state_in_HA_state_and_all_checks_undef(self,
                                                                  mock_check_slice_health_for_volume):
        mock_check_slice_health_for_volume.return_value = True
        vl_helpers.check_standby_service_state(self.cluster_fm, self.volume_dm,
                                               standby_state=enums.StandbyStates.inactive,
                                               check_degraded=None,
                                               check_stable_lives=None,
                                               check_standby_duplicates=None)
        mock_check_slice_health_for_volume.assert_called_with(self.cluster_fm, self.volume_dm,
                                                              standby_state=enums.StandbyStates.inactive,
                                                              check_degraded=True,
                                                              check_stable_lives=True,
                                                              check_standby_duplicates=True)

    @mock.patch('valence.vl_helpers._slice_assignments.check_slice_health_for_volume')
    def test_check_standby_state_no_HA_state_and_all_checks_undef(self,
                                                                  mock_check_slice_health_for_volume):
        mock_check_slice_health_for_volume.return_value = True
        vl_helpers.check_standby_service_state(self.cluster_fm, self.volume_dm,
                                               standby_state=enums.StandbyStates.active,
                                               check_degraded=None,
                                               check_stable_lives=None,
                                               check_standby_duplicates=None)
        # set since standby_state indicates no HA, degraded and stable lives checked are False
        mock_check_slice_health_for_volume.assert_called_with(self.cluster_fm, self.volume_dm,
                                                              standby_state=enums.StandbyStates.active,
                                                              check_degraded=False,
                                                              check_stable_lives=False,
                                                              check_standby_duplicates=True)

    @mock.patch('valence.vl_helpers._slice_assignments.check_slice_health_for_volume')
    def test_indicate_HA_state_and_set_all_checks_false(self,
                                                        mock_check_slice_health_for_volume):
        mock_check_slice_health_for_volume.return_value = True
        vl_helpers.check_standby_service_state(self.cluster_fm, self.volume_dm,
                                               standby_state=enums.StandbyStates.active,
                                               check_degraded=False,
                                               check_stable_lives=False,
                                               check_standby_duplicates=False)
        # set since standby_state indicates no HA, degraded and stable lives checked are False
        mock_check_slice_health_for_volume.assert_called_with(self.cluster_fm, self.volume_dm,
                                                              standby_state=enums.StandbyStates.active,
                                                              check_degraded=False,
                                                              check_stable_lives=False,
                                                              check_standby_duplicates=False)


class TestGetServiceDmForVolume(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.slices = [{'slice_id': 1,
                       'volume_id': 1,
                       'primary': 10,
                       'live_secondaries': [40],
                       'dead_secondaries': [50],
                       'standby_services': [{'service_id': 30, 'state': enums.StandbyStates.inactive}]},
                      {'slice_id': 2,
                       'volume_id': 2,
                       'primary': 20,
                       'live_secondaries': [],
                       'dead_secondaries': [],
                       'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                      {'slice_id': 3,
                       'volume_id': 3,
                       'primary': 30,
                       'live_secondaries': [20],
                       'dead_secondaries': [],
                       'standby_services': []},
                      {'slice_id': 4,
                       'volume_id': 4,
                       'primary': 40,
                       'live_secondaries': [],
                       'dead_secondaries': [30],
                       'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]}
                      ]
        cls.services = [{'service_id': 10},
                        {'service_id': 20},
                        {'service_id': 30},  # not standby anywhere
                        {'service_id': 40},
                        {'service_id': 50}]
        cls.slice_dms = [models.Slice.get_mock_object(overrides=sl) for sl in cls.slices]
        cls.slice_services = [models.Slice.get_mock_object(overrides=s) for s in cls.services]
        cls.report = dict(slices=cls.slices, services=cls.services)
        cls.report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=cls.report)
        cls.cluster_service_dms = [
            models.ClusterServiceInfo.get_mock_object(
                overrides=dict(
                    service={'service_type': enums.ServiceType.slice, 'service_id': s['service_id']}))
            for s in cls.services]

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_primary_service_dm_for_volume(self, mock_get_slice_sync_report,
                                               mock_get_slice_dm_from_volume,
                                               mock_list_services,
                                               mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))
        slice_dm = self.slice_dms[0]
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm  # we'll work with the first slice
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        service_infos = vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm,
                                                                 enums.SliceRole.live_secondary)
        assert service_infos[0].service.service_id == slice_dm.live_secondaries[0]

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_live_sec_service_dm_for_volume(self, mock_get_slice_sync_report,
                                                mock_get_slice_dm_from_volume,
                                                mock_list_services,
                                                mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))
        slice_dm = self.slice_dms[0]
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        service_infos = vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm,
                                                                 enums.SliceRole.primary)
        assert service_infos[0].service.service_id == slice_dm.primary

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_dead_sec_service_dm_for_volume(self, mock_get_slice_sync_report,
                                                mock_get_slice_dm_from_volume,
                                                mock_list_services,
                                                mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))
        slice_dm = self.slice_dms[0]
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        service_infos = vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm,
                                                                 enums.SliceRole.dead_secondary)
        assert service_infos[0].service.service_id == slice_dm.dead_secondaries[0]

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_get_standby_service_dm_for_volume(self, mock_get_slice_sync_report,
                                               mock_get_slice_dm_from_volume,
                                               mock_list_services,
                                               mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))
        slice_dm = self.slice_dms[0]
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        service_infos = vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm,
                                                                 enums.SliceRole.standby)
        assert service_infos[0].service.service_id == slice_dm.standby_services[0].service_id

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_service_for_volume_not_in_cluster_service_list(self, mock_get_slice_sync_report,
                                                            mock_get_slice_dm_from_volume,
                                                            mock_list_services,
                                                            mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))
        slice_dm = models.Slice.get_mock_object(overrides=dict(primary=100))
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        with pytest.raises(_errors.HelperError):
            vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm, enums.SliceRole.primary)

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_no_live_sec_service_dm_for_volume(self, mock_get_slice_sync_report,
                                               mock_get_slice_dm_from_volume,
                                               mock_list_services,
                                               mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=2))
        slice_dm = self.slice_dms[1]
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        with pytest.raises(_errors.HelperError):
            vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm,
                                                     enums.SliceRole.live_secondary)

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_no_dead_sec_service_dm_for_volume(self, mock_get_slice_sync_report,
                                               mock_get_slice_dm_from_volume,
                                               mock_list_services,
                                               mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=2))
        slice_dm = self.slice_dms[1]
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_wait_for_desired_primary_to_resolve.return_value = True

        mock_list_services.return_value = self.cluster_service_dms
        with pytest.raises(_errors.HelperError):
            vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm,
                                                     enums.SliceRole.dead_secondary)

    @mock.patch('valence.vl_helpers._slice_assignments.wait_for_desired_primary_to_resolve')
    @mock.patch.object(functional_models.Cluster, 'list_services')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_no_standby_services(self, mock_get_slice_sync_report,
                                 mock_get_slice_dm_from_volume,
                                 mock_list_services,
                                 mock_wait_for_desired_primary_to_resolve):
        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=3))
        mock_get_slice_sync_report.return_value = self.report_dm
        slice_dm = self.slice_dms[2]
        mock_wait_for_desired_primary_to_resolve.return_value = True
        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_list_services.return_value = self.cluster_service_dms
        with pytest.raises(_errors.HelperError):
            vl_helpers.get_slice_services_for_volume(self.cluster_fm, volume_dm, enums.SliceRole.standby)


class WaitForDesiredPrimaryToResolve(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.unresolved_slice = {'slice_id': 1,
                                'volume_id': 1,
                                'primary': 10,
                                'desired_primary': 40,
                                'live_secondaries': [40],
                                'dead_secondaries': [50],
                                'standby_services': [
                                    {'service_id': 30, 'state': enums.StandbyStates.inactive}]}
        cls.resolved_slice = {'slice_id': 1,
                              'volume_id': 1,
                              'primary': 40,
                              'desired_primary': 40,
                              'live_secondaries': [10],
                              'dead_secondaries': [50],
                              'standby_services': [
                                  {'service_id': 30, 'state': enums.StandbyStates.inactive}]}

        cls.no_desired_primary_slice = {'slice_id': 1,
                                        'volume_id': 1,
                                        'primary': 40,
                                        'desired_primary': 0,
                                        'live_secondaries': [10],
                                        'dead_secondaries': [50],
                                        'standby_services': [
                                            {'service_id': 30, 'state': enums.StandbyStates.inactive}]}

        cls.unresolved_slice_dm = models.Slice.get_mock_object(include_optional=True,
                                                               overrides=cls.unresolved_slice)
        cls.resolved_slice_dm = models.Slice.get_mock_object(include_optional=True,
                                                             overrides=cls.resolved_slice)

        cls.no_desired_primary_slice_dm = models.Slice.get_mock_object(
            include_optional=True,
            overrides=cls.no_desired_primary_slice)

        cls.volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=1))

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    def test_resolved_in_time(self, mock_get_slice_dm_from_volume, mock_time, mock_sleep):
        mock_time.return_value = iter([True, True, True, True, False])
        mock_get_slice_dm_from_volume.side_effect = [self.unresolved_slice_dm, self.resolved_slice_dm]
        vl_helpers.wait_for_desired_primary_to_resolve(self.cluster_fm, self.volume_dm)

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    def test_resolution_timeout(self, mock_get_slice_dm_from_volume, mock_time, mock_sleep):
        mock_time.return_value = iter([True, True, True, True, False])
        mock_get_slice_dm_from_volume.return_value = self.unresolved_slice_dm
        with pytest.raises(_errors.HelperError):
            vl_helpers.wait_for_desired_primary_to_resolve(self.cluster_fm, self.volume_dm)

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    def test_resolved_not_in_time(self, mock_get_slice_dm_from_volume, mock_time, mock_sleep):
        mock_time.return_value = iter([True, False, False])
        mock_get_slice_dm_from_volume.side_effect = [self.unresolved_slice_dm,
                                                     self.unresolved_slice_dm,
                                                     self.unresolved_slice_dm,
                                                     self.resolved_slice_dm]
        with pytest.raises(_errors.HelperError):
            vl_helpers.wait_for_desired_primary_to_resolve(self.cluster_fm, self.volume_dm)

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    def test_no_desired_primary(self, mock_get_slice_dm_from_volume, mock_time, mock_sleep):
        mock_time.return_value = iter([True, True, True, True, False])
        mock_get_slice_dm_from_volume.return_value = self.no_desired_primary_slice_dm
        vl_helpers.wait_for_desired_primary_to_resolve(self.cluster_fm, self.volume_dm)


class TestWaitForSliceHealth(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

        slices = [{'slice_id': 1,
                   'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 2,
                   'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 3,
                   'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 4,
                   'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 30, 'state': enums.StandbyStates.inactive}]}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        cls.report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)
        cls.volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=4))

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments._get_slice_health_issues')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_slice_healthy_after_time(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume,
                                      mock_check_slice_health,
                                      mock_timer, mock_sleep):
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = self.report_dm.slices[-1]
        mock_check_slice_health.side_effect = ['error1', 'error2', 'error3', 'error4', None, 'error5']

        mock_timer.return_value = iter([True, True, True, True, True, True, True, False])
        vl_helpers.wait_for_slice_health(self.cluster_fm, self.volume_dm, timeout=300, poll_interval=10,
                                         check_degraded=True,
                                         check_stable_lives=False, check_duplicates=True,
                                         check_standby_state=True,
                                         check_standby_duplicates=False)

        assert mock_check_slice_health.call_count == 5

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments._get_slice_health_issues')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_timeout_not_healthy(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume,
                                 mock_check_slice_health,
                                 mock_timer, mock_sleep):
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = self.report_dm.slices[-1]
        mock_check_slice_health.side_effect = ['error1', 'error2', 'error3', 'error4', None, 'error5']
        mock_timer.return_value = iter([True, True, True, False, False])
        with pytest.raises(_errors.HelperError):
            vl_helpers.wait_for_slice_health(self.cluster_fm, self.volume_dm, timeout=300,
                                             poll_interval=10,
                                             check_degraded=True,
                                             check_stable_lives=False, check_duplicates=True,
                                             check_standby_state=True,
                                             check_standby_duplicates=False)

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments._get_slice_health_issues')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_timeout_but_healthy(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume,
                                 mock_check_slice_health,
                                 mock_timer, mock_sleep):
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = self.report_dm.slices[-1]
        mock_check_slice_health.return_value = None
        mock_timer.return_value = iter([True, False])

        vl_helpers.wait_for_slice_health(self.cluster_fm, self.volume_dm, timeout=300, poll_interval=10,
                                         check_degraded=True,
                                         check_stable_lives=False, check_duplicates=True,
                                         check_standby_state=True,
                                         check_standby_duplicates=False)

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments._get_slice_health_issues')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_timeout_but_healthy_2(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume,
                                   mock_check_slice_health,
                                   mock_time, mock_sleep):
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = self.report_dm.slices[-1]
        mock_check_slice_health.side_effect = ['error1', None]
        mock_time.return_value = iter([True, True, False])

        vl_helpers.wait_for_slice_health(self.cluster_fm, self.volume_dm, timeout=300, poll_interval=10,
                                         check_degraded=True,
                                         check_stable_lives=False, check_duplicates=True,
                                         check_standby_state=True,
                                         check_standby_duplicates=False)

    @mock.patch('time.sleep')
    @mock.patch('valence.utils.timeout_timer')
    @mock.patch('valence.vl_helpers._slice_assignments._get_slice_health_issues')
    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_slice_not_timeout_not_healthy(self, mock_get_slice_sync_report,
                                           mock_get_slice_dm_from_volume,
                                           mock_check_slice_health,
                                           mock_timer, mock_sleep):
        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = self.report_dm.slices[-1]
        mock_check_slice_health.side_effect = [None, 'error2', 'error3', 'error4', None, 'error5']
        mock_timer.return_value = iter([True, True, True, True, True, True, True, True, True, False])

        vl_helpers.wait_for_slice_health(self.cluster_fm, self.volume_dm, timeout=300, poll_interval=10,
                                         check_degraded=True,
                                         check_stable_lives=False, check_duplicates=True,
                                         check_standby_state=True,
                                         check_standby_duplicates=False)

        assert mock_check_slice_health.call_count == 1


class TestCheckStandbyStateForVolume(unittest.TestCase):
    @classmethod
    def setUpClass(cls):

        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

        slices = [{'slice_id': 1,
                   'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 2,
                   'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.failed}]},
                  {'slice_id': 3,
                   'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.active}]},
                  {'slice_id': 4,
                   'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': []}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        cls.report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)

    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_check_inactive(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume):
        volume_id = 1
        slice_dm = None
        for s in self.report_dm.slices:
            if s.volume_id == volume_id:
                slice_dm = s
        assert slice_dm

        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_get_slice_sync_report.return_value = self.report_dm

        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=volume_id))
        check = vl_helpers.check_standby_state_for_volume(self.cluster_fm, volume_dm,
                                                          enums.StandbyStates.inactive)
        assert check

    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_check_inactive_failed(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume):
        volume_id = 2
        slice_dm = None
        for s in self.report_dm.slices:
            if s.volume_id == volume_id:
                slice_dm = s
        assert slice_dm

        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_get_slice_sync_report.return_value = self.report_dm

        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=volume_id))
        check = vl_helpers.check_standby_state_for_volume(self.cluster_fm, volume_dm,
                                                          enums.StandbyStates.inactive)
        assert not check

    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_check_inactive_state_with_standby_disabled(self, mock_get_slice_sync_report,
                                                        mock_get_slice_dm_from_volume):
        volume_id = 4
        slice_dm = None
        for s in self.report_dm.slices:
            if s.volume_id == volume_id:
                slice_dm = s
        assert slice_dm

        mock_get_slice_dm_from_volume.return_value = slice_dm
        mock_get_slice_sync_report.return_value = self.report_dm

        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=volume_id))
        with pytest.raises(_errors.HelperError):
            vl_helpers.check_standby_state_for_volume(self.cluster_fm, volume_dm,
                                                      enums.StandbyStates.inactive)


class TestWaitForStandbyState(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dm = models.Volume.get_mock_object()

    @mock.patch('valence.vl_helpers._slice_assignments.check_standby_state_for_volume')
    @mock.patch('valence.utils.timeout_timer')
    def test_state_found(self, mock_timer, mock_check_standby_state_for_volume):
        mock_timer.return_value = iter([True, True, True, True, False])
        mock_check_standby_state_for_volume.side_effect = [False, False, False, True]
        vl_helpers.wait_for_standby_state(cluster_fm=self.cluster_fm, volume_dm=self.volume_dm,
                                          standby_state=enums.StandbyStates.active)

    @mock.patch('valence.vl_helpers._slice_assignments.check_standby_state_for_volume')
    @mock.patch('valence.utils.timeout_timer')
    def test_timeout_on_check(self, mock_timer, mock_check_standby_state_for_volume):
        mock_timer.return_value = iter([True, True, True, True, False])
        mock_check_standby_state_for_volume.side_effect = [False, False, False, False, False]
        with pytest.raises(_errors.HelperError):
            vl_helpers.wait_for_standby_state(cluster_fm=self.cluster_fm, volume_dm=self.volume_dm,
                                              standby_state=enums.StandbyStates.active)


class TestGetStandbyServiceIdForVolume(unittest.TestCase):
    @classmethod
    def setUpClass(cls):

        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

        slices = [{'slice_id': 1,
                   'volume_id': 1,
                   'primary': 10,
                   'live_secondaries': [40],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 20, 'state': enums.StandbyStates.inactive}]},
                  {'slice_id': 2,
                   'volume_id': 2,
                   'primary': 20,
                   'live_secondaries': [10],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 40, 'state': enums.StandbyStates.failed}]},
                  {'slice_id': 3,
                   'volume_id': 3,
                   'primary': 30,
                   'live_secondaries': [20],
                   'dead_secondaries': [],
                   'standby_services': [{'service_id': 10, 'state': enums.StandbyStates.active}]},
                  {'slice_id': 4,
                   'volume_id': 4,
                   'primary': 40,
                   'live_secondaries': [30],
                   'dead_secondaries': [],
                   'standby_services': []}
                  ]
        services = [{'service_id': 10},
                    {'service_id': 20},
                    {'service_id': 30},  # not standby anywhere
                    {'service_id': 40}]
        report = dict(slices=slices, services=services)
        cls.report_dm = models.SliceReport.get_mock_object(include_optional=True, overrides=report)

    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_standby_service_found(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume):

        volume_id = 2
        slice_dm = None
        for s in self.report_dm.slices:
            if s.volume_id == volume_id:
                slice_dm = s
        assert slice_dm

        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=volume_id))

        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm
        service_id = slice_dm.standby_services[0].service_id
        s_id = vl_helpers.get_standby_service_id_for_volume(self.cluster_fm, volume_dm)

        assert s_id == service_id

    @mock.patch('valence.vl_helpers._slice_assignments.get_slice_dm_from_volume')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_standby_service_not_found(self, mock_get_slice_sync_report, mock_get_slice_dm_from_volume):

        volume_id = 4
        slice_dm = None
        for s in self.report_dm.slices:
            if s.volume_id == volume_id:
                slice_dm = s
        assert slice_dm

        volume_dm = models.Volume.get_mock_object(overrides=dict(volume_id=volume_id))

        mock_get_slice_sync_report.return_value = self.report_dm
        mock_get_slice_dm_from_volume.return_value = slice_dm

        with pytest.raises(IndexError):
            vl_helpers.get_standby_service_id_for_volume(self.cluster_fm, volume_dm)


class TestWaitForVolumeStandbysUnassigned(unittest.TestCase):
    sb_resp = [{'sb_unassigned_vols': [],
                'sb_assigned_vols': []},
               {'sb_unassigned_vols': [],
                'sb_assigned_vols': []}]

    @classmethod
    def setUpClass(cls):
        cls.nodes = [models.ActiveNode.get_mock_object(overrides=dict(node_id=0)),
                     models.ActiveNode.get_mock_object(overrides=dict(node_id=1))]
        cls.vols = [
            models.Volume.get_mock_object(overrides=dict(volume_id=v_id)) for v_id in range(1, 6)]
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    @staticmethod
    def _mock_sf_api_sb_assign_to_unassigned(node_id, log_response=None):
        if TestWaitForVolumeStandbysUnassigned.sb_resp[node_id]['sb_assigned_vols']:
            TestWaitForVolumeStandbysUnassigned.sb_resp[node_id]['sb_unassigned_vols'].append(
                TestWaitForVolumeStandbysUnassigned.sb_resp[node_id]['sb_assigned_vols'].pop())

        vols_wout_standbys = TestWaitForVolumeStandbysUnassigned.sb_resp[node_id]['sb_unassigned_vols']
        all_assigned = False if vols_wout_standbys else True
        raw_data = {sf_api.api_constants.RESULT: dict(allStandbysAssigned=all_assigned,
                                                      volumeIDsWithUnassignedStandbys=vols_wout_standbys
                                                      )}
        resp = sf_api.SfResponse(method=enums.ClusterMethod.check_volume_standbys_assigned,
                                 raw_data=raw_data,
                                 result_dm_class=sf_api.result_models.CheckVolumeStandbysAssigned)
        result = resp.result_dm
        return result

    def setUp(self):
        TestWaitForVolumeStandbysUnassigned.sb_resp[0]['sb_unassigned_vols'] = []
        TestWaitForVolumeStandbysUnassigned.sb_resp[1]['sb_unassigned_vols'] = []
        TestWaitForVolumeStandbysUnassigned.sb_resp[0]['sb_assigned_vols'] = [1, 3, 5]
        TestWaitForVolumeStandbysUnassigned.sb_resp[1]['sb_assigned_vols'] = [2, 4]

    @mock.patch('valence.utils.timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'check_volume_standbys_assigned')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_wait_for_all_vols_to_be_unassigned_before_timeout(self, mock_list_nodes, mock_list_volumes,
                                                               mock_api_call, mock_timeout):
        mock_list_nodes.return_value = self.nodes
        mock_list_volumes.return_value = self.vols
        mock_api_call.side_effect = \
            TestWaitForVolumeStandbysUnassigned._mock_sf_api_sb_assign_to_unassigned
        mock_timeout.return_value = iter([True] * 10)
        vl_helpers.wait_for_volume_standbys_unassigned(self.cluster_fm)
        # validate test data to make sure mocked call was called properly,
        # thus wait_for_volumes_standys_unassigned functioned properly
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[0]['sb_assigned_vols'] == []
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[1]['sb_assigned_vols'] == []
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[0]['sb_unassigned_vols'] == [5, 3, 1]
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[1]['sb_unassigned_vols'] == [4, 2]

    @mock.patch('valence.utils.timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'check_volume_standbys_assigned')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_wait_for_all_vols_to_be_unassigned_timeout(self, mock_list_nodes, mock_list_volumes,
                                                        mock_api_call, mock_timeout):
        mock_list_nodes.return_value = self.nodes
        mock_list_volumes.return_value = self.vols
        mock_api_call.side_effect = \
            TestWaitForVolumeStandbysUnassigned._mock_sf_api_sb_assign_to_unassigned
        mock_timeout.return_value = iter([True, False])
        with pytest.raises(_errors.HelperError):
            vl_helpers.wait_for_volume_standbys_unassigned(self.cluster_fm)
        # validate test data to make sure mocked call was called properly,
        # thus wait_for_volumes_standys_unassigned functioned properly
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[0]['sb_assigned_vols'] == [1, 3]
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[1]['sb_assigned_vols'] == [2]
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[0]['sb_unassigned_vols'] == [5]
        assert TestWaitForVolumeStandbysUnassigned.sb_resp[1]['sb_unassigned_vols'] == [4]


class TestWaitForVolumeStandbysAssigned(unittest.TestCase):
    sb_resp = [{'sb_unassigned_vols': [],
                'sb_assigned_vols': []},
               {'sb_unassigned_vols': [],
                'sb_assigned_vols': []}]

    @classmethod
    def setUpClass(cls):
        cls.nodes = [models.ActiveNode.get_mock_object(overrides=dict(node_id=0)),
                     models.ActiveNode.get_mock_object(overrides=dict(node_id=1))]
        cls.vols = [
            models.Volume.get_mock_object(overrides=dict(volume_id=v_id)) for v_id in range(1, 6)]
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()

    @staticmethod
    def _mock_sf_api_sb_unassign_to_assigned(node_id, log_response=None):
        if TestWaitForVolumeStandbysAssigned.sb_resp[node_id]['sb_unassigned_vols']:
            TestWaitForVolumeStandbysAssigned.sb_resp[node_id]['sb_assigned_vols'].append(
                TestWaitForVolumeStandbysAssigned.sb_resp[node_id]['sb_unassigned_vols'].pop())

        vols_wout_standbys = TestWaitForVolumeStandbysAssigned.sb_resp[node_id]['sb_unassigned_vols']
        all_assigned = False if vols_wout_standbys else True
        raw_data = {sf_api.api_constants.RESULT: dict(allStandbysAssigned=all_assigned,
                                                      volumeIDsWithUnassignedStandbys=vols_wout_standbys
                                                      )}
        resp = sf_api.SfResponse(method=enums.ClusterMethod.check_volume_standbys_assigned,
                                 raw_data=raw_data,
                                 result_dm_class=sf_api.result_models.CheckVolumeStandbysAssigned)

        return resp.result_dm

    def setUp(self):
        TestWaitForVolumeStandbysAssigned.sb_resp[0]['sb_unassigned_vols'] = [1, 3, 5]
        TestWaitForVolumeStandbysAssigned.sb_resp[1]['sb_unassigned_vols'] = [2, 4]
        TestWaitForVolumeStandbysAssigned.sb_resp[0]['sb_assigned_vols'] = []
        TestWaitForVolumeStandbysAssigned.sb_resp[1]['sb_assigned_vols'] = []

    @mock.patch('valence.utils.timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'check_volume_standbys_assigned')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_wait_for_all_vols_to_be_assigned_before_timeout(self, mock_list_nodes, mock_list_volumes,
                                                             mock_api_call, mock_timeout):
        mock_list_nodes.return_value = self.nodes
        mock_list_volumes.return_value = self.vols
        mock_api_call.side_effect = \
            TestWaitForVolumeStandbysAssigned._mock_sf_api_sb_unassign_to_assigned
        mock_timeout.return_value = iter([True] * 10)
        vl_helpers.wait_for_volume_standbys_assigned(self.cluster_fm)

        # validate test data to make sure mocked call was called properly,
        # thus wait_for_volumes_standys_assigned functioned properly
        assert TestWaitForVolumeStandbysAssigned.sb_resp[0]['sb_assigned_vols'] == [5, 3, 1]
        assert TestWaitForVolumeStandbysAssigned.sb_resp[1]['sb_assigned_vols'] == [4, 2]
        assert TestWaitForVolumeStandbysAssigned.sb_resp[0]['sb_unassigned_vols'] == []
        assert TestWaitForVolumeStandbysAssigned.sb_resp[1]['sb_unassigned_vols'] == []

    @mock.patch('valence.utils.timeout_timer')
    @mock.patch.object(functional_models.Cluster, 'check_volume_standbys_assigned')
    @mock.patch.object(functional_models.Cluster, 'list_volumes')
    @mock.patch.object(functional_models.Cluster, 'list_active_nodes')
    def test_wait_for_all_vols_to_be_assigned_timeout(self, mock_list_nodes, mock_list_volumes,
                                                      mock_api_call, mock_timeout):
        mock_list_nodes.return_value = self.nodes
        mock_list_volumes.return_value = self.vols
        mock_api_call.side_effect = \
            TestWaitForVolumeStandbysAssigned._mock_sf_api_sb_unassign_to_assigned
        mock_timeout.return_value = iter([True, False])
        with pytest.raises(_errors.HelperError):
            vl_helpers.wait_for_volume_standbys_assigned(self.cluster_fm)

        # validate test data to make sure mocked call was called properly,
        # thus wait_for_volumes_standys_assigned functioned properly
        assert TestWaitForVolumeStandbysAssigned.sb_resp[0]['sb_assigned_vols'] == [5, ]
        assert TestWaitForVolumeStandbysAssigned.sb_resp[1]['sb_assigned_vols'] == [4, ]
        assert TestWaitForVolumeStandbysAssigned.sb_resp[0]['sb_unassigned_vols'] == [1, 3]
        assert TestWaitForVolumeStandbysAssigned.sb_resp[1]['sb_unassigned_vols'] == [2, ]
