
import time
import unittest

import mock
import pytest
import six

from sf_api import result_models
from valence import (
    enums,
    errors,
    functional_models,
    models,
    vl_helpers,
)

# Mock the cluster DM and FM once to avoid extra work
RS_CLUSTER_DM = models.ResourceCluster.get_mock_object()
CLUSTER_FM = functional_models.Cluster(RS_CLUSTER_DM, authenticate=False)


class TestFillVolumes(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster.get_mock_fm()
        cls.volume_dms = [models.Volume.get_mock_object()]

    def setUp(self):
        self.slice_report = models.SliceReport.get_mock_object(include_optional=True)

    @mock.patch('valence.vl_helpers._slice._validate_fill_slice_inputs')
    @mock.patch('valence.vl_helpers._slice._generate_fill_slice_seeds')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.vl_helpers._slice.wait_for_slice_sync_to_finish')
    @mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'set_constants')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    @mock.patch.object(functional_models.Cluster, 'fill_slice')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    @mock.patch('valence.vl_helpers._slice._wait_for_fill_slice_handles')
    def test_fill_volumes_no_dedup(self, mock_wait_handles, mock_update_status, mock_fill_slice,
                                   mock_get_slice_report, mock_set_constants, mock_wait_bin_sync,
                                   mock_wait_slice_sync, mock_get_constants, mock_generate_seeds,
                                   mock_validate_inputs):
        mock_generate_seeds.return_value = list(range(len(self.volume_dms)))
        self.slice_report.slices[0].volume_id = self.volume_dms[0].volume_id
        mock_get_slice_report.return_value = self.slice_report
        vl_helpers.fill_volumes(self.cluster_fm, self.volume_dms)
        assert mock_set_constants.call_count == 2
        mock_fill_slice.assert_called_once()

    @mock.patch('valence.vl_helpers._slice._validate_fill_slice_inputs')
    @mock.patch('valence.vl_helpers._slice._generate_fill_slice_seeds')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.vl_helpers._slice.wait_for_slice_sync_to_finish')
    @mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'set_constants')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    @mock.patch.object(functional_models.Cluster, 'fill_slice')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    @mock.patch('valence.vl_helpers._slice._wait_for_fill_slice_handles')
    def test_fill_volumes_dedup(self, mock_wait_handles, mock_update_status, mock_fill_slice,
                                mock_get_slice_report, mock_set_constants, mock_wait_bin_sync,
                                mock_wait_slice_sync, mock_get_constants, mock_generate_seeds,
                                mock_validate_inputs):
        dedup_ratio = 6.0
        volume_dm = models.Volume.get_mock_object()
        # Volume size needs to be sufficiently large, otherwise dedup loop isn't consistent.
        # The calculated lba 'chunk_size' needs to be > dedup_ratio.  Otherwise call_count is
        # hard to predict.  This really only affects unit tests as real volumes aren't so small.
        volume_dm.total_size += 4096*4096
        mock_generate_seeds.return_value = [12345]
        self.slice_report.slices[0].volume_id = volume_dm.volume_id
        mock_get_slice_report.return_value = self.slice_report
        vl_helpers.fill_volumes(self.cluster_fm, [volume_dm], dedup_ratio=dedup_ratio)
        assert mock_set_constants.call_count == 2
        assert mock_fill_slice.call_count in (6, 7)

    @mock.patch('valence.vl_helpers._slice._validate_fill_slice_inputs')
    @mock.patch('valence.vl_helpers._slice._generate_fill_slice_seeds')
    @mock.patch.object(functional_models.Cluster, 'get_constants')
    @mock.patch('valence.vl_helpers._slice.wait_for_slice_sync_to_finish')
    @mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
    @mock.patch.object(functional_models.Cluster, 'set_constants')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    @mock.patch.object(functional_models.Cluster, 'fill_slice')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    @mock.patch('valence.vl_helpers._slice._wait_for_fill_slice_handles')
    def test_fill_volumes_no_slice_id(self, mock_wait_handles, mock_update_status, mock_fill_slice,
                                      mock_get_slice_report, mock_set_constants, mock_wait_bin_sync,
                                      mock_wait_slice_sync, mock_get_constants, mock_generate_seeds,
                                      mock_validate_inputs):
        mock_generate_seeds.return_value = list(range(len(self.volume_dms)))
        self.slice_report.slices[0].volume_id = self.volume_dms[0].volume_id + 1
        mock_get_slice_report.return_value = self.slice_report

        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.fill_volumes(self.cluster_fm, self.volume_dms)

    def test_validate_fill_slice_inputs(self):
        with pytest.raises(ValueError):
            # compression ratio < 1.0
            vl_helpers._slice._validate_fill_slice_inputs(
                fill_fraction=1.0,
                start_fraction=0.0,
                compression_ratio=0.8,
                compression_algorithm=enums.CompressionAlgorithm.none,
                dedup_ratio=1.0
            )

        with pytest.raises(ValueError):
            # compression alg is none and compression ratio != 1.0
            vl_helpers._slice._validate_fill_slice_inputs(
                fill_fraction=1.0,
                start_fraction=0.0,
                compression_ratio=1.1,
                compression_algorithm=enums.CompressionAlgorithm.none,
                dedup_ratio=1.0
            )

        with pytest.raises(ValueError):
            # fill + start > 1.0
            vl_helpers._slice._validate_fill_slice_inputs(
                fill_fraction=0.55,
                start_fraction=0.55,
                compression_ratio=1.0,
                compression_algorithm=enums.CompressionAlgorithm.none,
                dedup_ratio=1.0
            )

        with pytest.raises(ValueError):
            # fill < 0
            vl_helpers._slice._validate_fill_slice_inputs(
                fill_fraction=-0.55,
                start_fraction=0.55,
                compression_ratio=1.0,
                compression_algorithm=enums.CompressionAlgorithm.none,
                dedup_ratio=1.0
            )

        with pytest.raises(ValueError):
            # start < 0
            vl_helpers._slice._validate_fill_slice_inputs(
                fill_fraction=1.0,
                start_fraction=-1.0,
                compression_ratio=1.0,
                compression_algorithm=enums.CompressionAlgorithm.none,
                dedup_ratio=1.0
            )

        with pytest.raises(ValueError):
            # dedup < 1.0
            vl_helpers._slice._validate_fill_slice_inputs(
                fill_fraction=1.0,
                start_fraction=0.0,
                compression_ratio=1.0,
                compression_algorithm=enums.CompressionAlgorithm.none,
                dedup_ratio=0.9
            )

    def test_generate_fill_slice_seeds(self):
        gen_seeds = vl_helpers._slice._generate_fill_slice_seeds(6, provided_seeds=[1, 2, 3])
        assert gen_seeds.count(1) == 2
        assert gen_seeds.count(2) == 2
        assert gen_seeds.count(3) == 2
        gen_seeds = vl_helpers._slice._generate_fill_slice_seeds(6, None)
        assert len(gen_seeds) == 6

    def test_calculate_fill_slice_timeout(self):
        volume_dms = models.Volume.get_mock_object_list(20)
        vl_helpers._slice._calculate_fill_slice_timeout(volume_dms)

    @mock.patch.object(functional_models.Cluster, 'modify_volumes')
    def test_update_fill_slice_timeout(self, mock_modify_volumes):
        vl_helpers._slice._update_fill_slice_status(self.cluster_fm, self.volume_dms,
                                                    enums.AsyncResultStatus.running)
        attributes = {'fillSliceStatus': enums.AsyncResultStatus.running.value}
        mock_modify_volumes.assert_called_with(self.volume_dms, attributes=attributes)

    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.FillSliceAsyncHandler, 'get_status')
    @mock.patch('time.time')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    def test_wait_for_fill_slice_handles_complete(self, mock_update_status, mock_time,
                                                  mock_get_status, mock_sleep):
        mock_get_status.return_value = enums.AsyncResultStatus.complete
        slice_id = self.slice_report.slices[0].slice_id
        async_handle_dm = models.AsyncHandle.get_mock_object()
        mock_async_fm = functional_models.FillSliceAsyncHandler(self.cluster_fm, slice_id,
                                                                async_handle_dm)
        handle_dict = {self.volume_dms[0].volume_id: [mock_async_fm]}
        vl_helpers._slice._wait_for_fill_slice_handles(self.cluster_fm, self.volume_dms, handle_dict,
                                                       0, 12345, False)

    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.FillSliceAsyncHandler, 'get_status')
    @mock.patch('time.time')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    def test_wait_for_fill_slice_handles_timeout(self, mock_update_status, mock_time,
                                                 mock_get_status, mock_sleep):
        mock_get_status.return_value = enums.AsyncResultStatus.running
        mock_time.side_effect = [12344, 12345, 12346]
        slice_id = self.slice_report.slices[0].slice_id
        async_handle_dm = models.AsyncHandle.get_mock_object()
        mock_async_fm = functional_models.FillSliceAsyncHandler(self.cluster_fm, slice_id,
                                                                async_handle_dm)
        handle_dict = {self.volume_dms[0].volume_id: [mock_async_fm]}
        with pytest.raises(vl_helpers._slice._errors.HelperError):
            vl_helpers._slice._wait_for_fill_slice_handles(self.cluster_fm, self.volume_dms,
                                                           handle_dict,
                                                           0, 12345, False)

    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.FillSliceAsyncHandler, 'get_status')
    @mock.patch('time.time')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    def test_wait_for_fill_slice_handles_fail_no_raise(self, mock_update_status, mock_time,
                                                       mock_get_status, mock_sleep):
        mock_get_status.return_value = enums.AsyncResultStatus.failed
        slice_id = self.slice_report.slices[0].slice_id
        async_handle_dm = models.AsyncHandle.get_mock_object()
        mock_async_fm = functional_models.FillSliceAsyncHandler(self.cluster_fm, slice_id,
                                                                async_handle_dm)
        handle_dict = {self.volume_dms[0].volume_id: [mock_async_fm]}
        vl_helpers._slice._wait_for_fill_slice_handles(self.cluster_fm, self.volume_dms,
                                                       handle_dict,
                                                       0, 12345, False)

    @mock.patch('time.sleep')
    @mock.patch.object(functional_models.FillSliceAsyncHandler, 'get_status')
    @mock.patch('time.time')
    @mock.patch('valence.vl_helpers._slice._update_fill_slice_status')
    def test_wait_for_fill_slice_handles_fail_raise(self, mock_update_status, mock_time,
                                                    mock_get_status, mock_sleep):
        mock_get_status.return_value = enums.AsyncResultStatus.failed
        slice_id = self.slice_report.slices[0].slice_id
        async_handle_dm = models.AsyncHandle.get_mock_object()
        mock_async_fm = functional_models.FillSliceAsyncHandler(self.cluster_fm, slice_id,
                                                                async_handle_dm)
        handle_dict = {self.volume_dms[0].volume_id: [mock_async_fm]}

        with pytest.raises(vl_helpers._slice._errors.HelperError):
            vl_helpers._slice._wait_for_fill_slice_handles(self.cluster_fm, self.volume_dms,
                                                           handle_dict,
                                                           0, 12345, True)


@mock.patch.object(functional_models.Cluster, 'is_slice_syncing')
class TestIsSliceSyncing(unittest.TestCase):
    def test_is_slice_syncing_false(self, mock_is_slice_syncing):
        mock_is_slice_syncing.return_value = False
        slice_sync = vl_helpers.is_slice_syncing(CLUSTER_FM)
        assert slice_sync is False

    def test_is_slice_syncing_true(self, mock_is_slice_syncing):
        mock_is_slice_syncing.return_value = True
        slice_sync = vl_helpers.is_slice_syncing(CLUSTER_FM)
        assert slice_sync is True


@mock.patch.object(vl_helpers._slice, 'is_slice_syncing')
@mock.patch.object(time, 'sleep')
class TestWaitForSliceSync(unittest.TestCase):
    def test_wait_for_slice_sync(self, mock_sleep, mock_is_slice_syncing):
        mock_is_slice_syncing.side_effect = [True, False]
        vl_helpers._slice._wait_for_slice_sync(CLUSTER_FM)

    def test_wait_for_slice_sync_volumes(self, mock_sleep, mock_is_slice_syncing):
        volume_dms = models.Volume.get_mock_object_list(1, overrides=dict(volume_id=42))
        vl_helpers._slice._wait_for_slice_sync(CLUSTER_FM, volume_dms=volume_dms)
        mock_is_slice_syncing.assert_called_with(CLUSTER_FM, volume_dms=volume_dms)

    @mock.patch.object(time, 'time')
    @mock.patch.object(functional_models.Cluster, 'get_report')
    def test_wait_for_slice_sync_timeout(self, mock_report, mock_time, mock_sleep,
                                         mock_is_slice_syncing):
        mock_is_slice_syncing.side_effect = [False]
        mock_time.side_effect = [10, 12]
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers._slice._wait_for_slice_sync(CLUSTER_FM, timeout=1)

    @mock.patch.object(vl_helpers._slice, '_wait_for_slice_sync')
    def test_wait_for_slice_sync_to_start(self, mock__wait_for_slice_sync, mock_sleep,
                                          mock_is_slice_syncing):
        vl_helpers.wait_for_slice_sync_to_start(CLUSTER_FM)
        mock__wait_for_slice_sync.assert_called_with(CLUSTER_FM, volume_dms=None, timeout=1800,
                                                     finish=False)

    @mock.patch.object(time, 'time')
    @mock.patch.object(functional_models.Cluster, 'get_report')
    def test_wait_for_slice_sync_to_start_timeout(self, mock_report, mock_time, mock_sleep,
                                                  mock_is_slice_syncing):
        mock_is_slice_syncing.side_effect = [False]
        mock_time.side_effect = [10, 12]
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.wait_for_slice_sync_to_start(CLUSTER_FM, timeout=1)

    @mock.patch.object(vl_helpers._slice, '_wait_for_slice_sync')
    def test_wait_for_slice_sync_to_finish(self, mock__wait_for_slice_sync, mock_sleep,
                                           mock_is_slice_syncing):
        vl_helpers.wait_for_slice_sync_to_finish(CLUSTER_FM)
        mock__wait_for_slice_sync.assert_called_with(CLUSTER_FM, volume_dms=None, timeout=1800,
                                                     finish=True)

    @mock.patch.object(time, 'time')
    @mock.patch.object(functional_models.Cluster, 'get_report')
    def test_wait_for_slice_sync_to_finish_timeout(self, mock_report, mock_time, mock_sleep,
                                                   mock_is_slice_syncing):
        mock_is_slice_syncing.side_effect = [True]
        mock_time.side_effect = [10, 12]
        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.wait_for_slice_sync_to_finish(CLUSTER_FM, timeout=1)


class TestCalculateSliceSyncTimeout(unittest.TestCase):

    @mock.patch('valence.functional_models.Cluster.get_cluster_full_threshold')
    @mock.patch('valence.functional_models.Cluster.get_cluster_info')
    def test_calculate_slice_timeout(self, mock_info, mock_get_thresholds):
        the_thresholds = result_models.GetClusterFullThreshold.get_mock_object(
            overrides={'sum_used_metadata_cluster_bytes': 20 * 2**20})
        info = result_models.GetClusterInfo.get_mock_object(
            overrides={'default_protection_scheme': 'singleHelix'}).cluster_info
        mock_get_thresholds.return_value = the_thresholds
        mock_info.return_value = info
        calculated_timeout = vl_helpers.calculate_slice_sync_timeout(CLUSTER_FM)
        assert calculated_timeout == 60


class TestSliceFailureDomainHelpers(unittest.TestCase):

    @mock.patch('valence.vl_helpers._slice._get_slice_primary_and_secondaries')
    @mock.patch.object(functional_models.Cluster, 'get_slice_sync_report')
    def test_create_slice_node_map(self, mock_get_slice_sync_report,
                                   mock_get_slice_primary_and_secondaries):
        mock_slice_report = models.SliceReport.get_mock_object(include_optional=True)
        mock_get_slice_sync_report.return_value = mock_slice_report
        mock_get_slice_primary_and_secondaries.return_value = [1, 2]
        slice_to_node_map = vl_helpers.create_slice_node_map(CLUSTER_FM)
        for slice_id, node_id_list in six.iteritems(slice_to_node_map):
            assert node_id_list == [1, 2]

    def test_get_slice_primary_and_secondaries(self):
        mock_slice = mock.Mock()
        mock_slice.primary = 1
        mock_slice.live_secondaries = [2]
        slice_to_node_map = {1: 10, 2: 20}
        primary_secondary_list = vl_helpers._slice._get_slice_primary_and_secondaries(mock_slice,
                                                                                      slice_to_node_map)
        assert primary_secondary_list == [10, 20]


@mock.patch('time.sleep')
@mock.patch('time.time')
@mock.patch('valence.vl_helpers._bin_sync.wait_for_bin_sync_to_finish_with_timeout')
@mock.patch('valence.vl_helpers._slice.wait_for_slice_sync_to_finish')
@mock.patch('valence.vl_helpers._slice.wait_for_slice_sync_to_start')
@mock.patch('valence.functional_models.Node.kill_drive_service_ssh')
@mock.patch('valence.functional_models.Cluster.get_service_status')
@mock.patch('valence.functional_models.Cluster.list_active_nodes')
@mock.patch('valence.functional_models.Cluster.list_services')
class TestRebootSliceServices(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cluster_fm = functional_models.Cluster(RS_CLUSTER_DM)
        cls.node_dms = [models.ActiveNode.get_mock_object(overrides={'node_id': i}) for i in range(2)]
        cls.service_dms = [
            models.ClusterServiceInfo.get_mock_object(overrides=dict(
                service={'service_type': enums.ServiceType.slice},
                node=cls.node_dms[0]
            )),
            models.ClusterServiceInfo.get_mock_object(overrides=dict(
                service={'service_type': enums.ServiceType.block},
                node=cls.node_dms[0]
            ))
        ]
        cls.mock_service_statuses = [
            result_models.GetServiceStatus.get_mock_object(overrides=dict(
                success={str(cls.service_dms[1].service.service_id): True},
                failed={str(cls.service_dms[0].service.service_id): 'error'}
            )),
            result_models.GetServiceStatus.get_mock_object(overrides=dict(
                success={
                    str(cls.service_dms[0].service.service_id): True,
                    str(cls.service_dms[1].service.service_id): True
                },
                failed={}
            ))
        ]

    def test_success(
            self, mock_list_services, mock_list_nodes, mock_get_service_status, mock_kill_drive_service,
            mock_wait_for_slice_sync_start, mock_wait_for_slice_sync_finish,
            mock_wait_for_bin_sync_finish, mock_time, mock_sleep
    ):
        mock_list_services.return_value = self.service_dms
        mock_list_nodes.return_value = self.node_dms
        mock_get_service_status.side_effect = self.mock_service_statuses
        mock_time.side_effect = [0, 1, 600, 330]

        vl_helpers.reboot_slice_services(self.cluster_fm)
        assert mock_kill_drive_service.call_count == 1
        assert mock_get_service_status.call_count == 2
        assert not mock_wait_for_slice_sync_start.called
        assert not mock_wait_for_slice_sync_finish.called
        assert not mock_wait_for_bin_sync_finish.called

    def test_timeout(
            self, mock_list_services, mock_list_nodes, mock_get_service_status, mock_kill_drive_service,
            mock_wait_for_slice_sync_start, mock_wait_for_slice_sync_finish,
            mock_wait_for_bin_sync_finish, mock_time, mock_sleep
    ):
        mock_list_services.return_value = self.service_dms
        mock_list_nodes.return_value = self.node_dms
        mock_get_service_status.side_effect = self.mock_service_statuses
        mock_time.side_effect = [0, 1, 601, 330]

        with pytest.raises(errors.TimeoutError):
            vl_helpers.reboot_slice_services(self.cluster_fm)

    def test_sync(
            self, mock_list_services, mock_list_nodes, mock_get_service_status, mock_kill_drive_service,
            mock_wait_for_slice_sync_start, mock_wait_for_slice_sync_finish,
            mock_wait_for_bin_sync_finish, mock_time, mock_sleep
    ):
        mock_list_services.return_value = self.service_dms
        mock_list_nodes.return_value = self.node_dms
        mock_get_service_status.side_effect = self.mock_service_statuses
        mock_time.side_effect = [0, 1, 600, 331]

        vl_helpers.reboot_slice_services(self.cluster_fm)
        assert mock_kill_drive_service.call_count == 1
        assert mock_get_service_status.call_count == 2
        assert mock_wait_for_slice_sync_start.call_count == 1
        assert mock_wait_for_slice_sync_finish.call_count == 1
        assert mock_wait_for_bin_sync_finish.call_count == 1


class TestGetSlicesAssignedToService(unittest.TestCase):
    def setUp(self):
        self.slices = models.SliceReport.get_mock_object(overrides={
            'slices': {'primary': 0,
                       'live_secondaries': [1, 2],
                       'dead_secondaries': []}})

    def test_slice(self):
        assert len(vl_helpers.get_slices_assigned_to_services(self.slices, [0])) == len(
            self.slices.slices)

    def test_secondary(self):
        assert len(vl_helpers.get_slices_assigned_to_services(self.slices, [1])) == len(
            self.slices.slices)

    def test_good_slice(self):
        assert len(vl_helpers.get_slices_assigned_to_services(self.slices, [3])) == 0
