import time
import unittest

import mock
import pytest
import sf_batch

import sf_api
from sf_api import result_models
from valence import (
    functional_models,
    models,
    vl_helpers,
)
from valence.vl_helpers import (
    _errors,
    _volume,
)

# Mock the node FM once to avoid extra work
CLUSTER_FM = functional_models.Cluster.get_mock_fm()


@mock.patch.object(functional_models.Cluster, 'list_active_nodes')
@mock.patch.object(_volume, '_check_slice_lbas_at_rest')
@mock.patch.object(vl_helpers._slice, 'wait_for_slice_sync_to_finish')
class TestCheckLbasAtRest(unittest.TestCase):
    def test_volumes_provided(self, mock_wait_slice, mock_slice_lbas, mock_get_active_nodes):
        mock_get_active_nodes.return_value = models.ActiveNode.get_mock_object_list(2)
        slice_checksums = [1234578, 87654321]
        mock_slice_lbas.side_effect = slice_checksums
        mock_volumes = models.Volume.get_mock_object_list(2)

        result = _volume.check_lbas_at_rest(CLUSTER_FM, volume_dms=mock_volumes)
        assert set(result) == set(slice_checksums)

    def test_snapshots_provided(self, mock_wait_slice, mock_slice_lbas, mock_get_active_nodes):
        mock_get_active_nodes.return_value = models.ActiveNode.get_mock_object_list(2)
        slice_checksums = [1234578, 87654321]
        mock_slice_lbas.side_effect = slice_checksums
        mock_snapshots = models.Snapshot.get_mock_object_list(2)

        result = _volume.check_lbas_at_rest(CLUSTER_FM, snapshot_dms=mock_snapshots)
        assert set(result) == set(slice_checksums)

    def test_volume_and_snapshots_provided(self, mock_wait_slice, mock_slice_lbas,
                                           mock_get_active_nodes):
        mock_get_active_nodes.return_value = models.ActiveNode.get_mock_object_list(2)
        slice_checksums = [1234, 5678, 8765, 4321]
        mock_slice_lbas.side_effect = slice_checksums

        mock_volumes = models.Volume.get_mock_object_list(2)
        mock_snapshots = models.Snapshot.get_mock_object_list(2)

        result = _volume.check_lbas_at_rest(CLUSTER_FM, volume_dms=mock_volumes,
                                            snapshot_dms=mock_snapshots)
        assert set(result) == set(slice_checksums)

    def test_empty_input(self, mock_wait_slice, mock_slice_lbas, mock_get_active_nodes):
        with pytest.raises(ValueError):
            _volume.check_lbas_at_rest(CLUSTER_FM, volume_dms=None, snapshot_dms=None)

    def test_ignore_secondaries_explicit(self, mock_wait_slice, mock_slice_lbas, mock_list_active_nodes):
        mock_list_active_nodes.return_value = models.ActiveNode.get_mock_object_list(2)
        mock_volumes = models.Volume.get_mock_object_list(2)

        result = _volume.check_lbas_at_rest(CLUSTER_FM, volume_dms=mock_volumes,
                                            ignore_secondaries=True)

        for call in mock_slice_lbas.call_args_list:
            # assert the ignore_secondaries argument is true
            assert call[0][-2] is True

    def test_ignore_secondaries_single_node(self, mock_wait_slice, mock_slice_lbas,
                                            mock_list_active_nodes):
        mock_list_active_nodes.return_value = models.ActiveNode.get_mock_object_list(1)
        mock_volumes = models.Volume.get_mock_object_list(2)

        result = _volume.check_lbas_at_rest(CLUSTER_FM, volume_dms=mock_volumes)

        for call in mock_slice_lbas.call_args_list:
            # assert the ignore_secondaries argument is true
            assert call[0][-2] is True


@mock.patch.object(functional_models.Cluster, 'get_report')
class TestPrivCheckSliceLbasAtRest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mock_volume = models.Volume.get_mock_object()
        cls.mock_snapshot = models.Snapshot.get_mock_object()
        cls.mock_snapshot.volume_id = cls.mock_volume.volume_id
        cls.mock_snapshot.total_size = models.DataSize(4096*10)

        cls.slice_report = {
            'slices': [
                {
                    'volumeID': cls.mock_volume.volume_id,
                    'sliceID': cls.mock_volume.volume_id + 1,
                    'deadSecondaries': [1, 2, 3]
                }
            ]
        }

        cls.slice_report_missing_id = {
            'slices': [
                {
                    'volumeID': cls.mock_volume.volume_id,
                    'sliceID': None,
                    'deadSecondaries': [1, 2, 3]
                }
            ]
        }

    def test_missing_slice_id(self, mock_get_report):
        mock_get_report.return_value = self.slice_report_missing_id

        with pytest.raises(ValueError):
            _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_volume, 100000, 0,
                                              None, False, 2)

    def test_invalid_start_lba(self, mock_get_report):
        mock_get_report.return_value = self.slice_report

        start_lba = self.mock_volume.total_size.bytes / 4096

        with pytest.raises(ValueError):
            _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_volume, 100000, start_lba,
                                              None, False, 2)

    def test_invalid_lba_range(self, mock_get_report):
        mock_get_report.return_value = self.slice_report

        start_lba = self.mock_volume.total_size.bytes / 4096 - 1

        with pytest.raises(ValueError):
            _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_volume, 100000, start_lba,
                                              2, False, 2)

    def test_invalid_n_lba(self, mock_get_report):
        mock_get_report.return_value = self.slice_report
        start_lba = self.mock_volume.total_size.bytes / 4096 - 1

        with pytest.raises(ValueError):
            _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_volume, 100000, start_lba,
                                              0, False, 2)

    @mock.patch.object(sf_batch.ThreadPool, 'add_job')
    @mock.patch.object(sf_batch.ThreadPool, 'run')
    @mock.patch.object(sf_batch.ThreadPool, 'success')
    def test_ignore_secondaries(self, mock_pool_success, mock_pool_run, mock_add_job, mock_get_report):
        mock_get_report.return_value = self.slice_report

        _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_snapshot, 100000, 0,
                                          None, True, 2)
        job_args = (CLUSTER_FM, self.mock_snapshot.volume_id, self.mock_snapshot.snapshot_id,
                    self.mock_snapshot.volume_id + 1, {}, 0, 10, [1, 2, 3], 2)
        mock_add_job.assert_called_once_with(_volume._check_slice_lba_chunk, args=job_args)

    @mock.patch.object(sf_batch.ThreadPool, 'add_job')
    @mock.patch.object(sf_batch.ThreadPool, 'run')
    @mock.patch.object(sf_batch.ThreadPool, 'success')
    def test_thread_one_chunk(self, mock_pool_success, mock_pool_run, mock_add_job, mock_get_report):
        mock_get_report.return_value = self.slice_report
        mock_pool_success.return_value = True
        _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_snapshot, 100000, 0,
                                          None, False, 2)

        # 10 LBAs to check.  Make sure calculation to determine partial chunk < chunk size works.
        job_args = (CLUSTER_FM, self.mock_snapshot.volume_id, self.mock_snapshot.snapshot_id,
                    self.mock_snapshot.volume_id + 1, {}, 0, 10, [], 2)
        mock_add_job.assert_called_once_with(_volume._check_slice_lba_chunk, args=job_args)

    @mock.patch.object(sf_batch.ThreadPool, 'add_job')
    @mock.patch.object(sf_batch.ThreadPool, 'run')
    @mock.patch.object(sf_batch.ThreadPool, 'success')
    def test_thread_chunking(self, mock_pool_success, mock_pool_run, mock_add_job, mock_get_report):
        mock_get_report.return_value = self.slice_report
        mock_pool_success.return_value = True

        # Change chunk size to 1 LBA.  Snapshot should have 10 LBAs to check..
        _volume._check_slice_lbas_at_rest(CLUSTER_FM, self.mock_snapshot, 1, 0,
                                          None, False, 2)

        assert mock_add_job.call_count == 10


class TestPrivCheckSliceLbaChunk(unittest.TestCase):
    @mock.patch.object(functional_models.Cluster, 'read_slice_lba_checksum')
    def test_inconsistent_checksum(self, mock_slice_checksum):
        mock_slice_checksum.return_value = {
            '1': 12345,
            '2': 123456
        }

        with pytest.raises(vl_helpers.HelperError):
            _volume._check_slice_lba_chunk(CLUSTER_FM, 1, 2, 3, {}, 0, 1, [], 2)

    @mock.patch.object(functional_models.Cluster, 'read_slice_lba_checksum')
    def test_api_returns_none(self, mock_slice_checksum):
        mock_slice_checksum.return_value = None

        with pytest.raises(vl_helpers.HelperError):
            _volume._check_slice_lba_chunk(CLUSTER_FM, 1, 2, 3, {}, 0, 1, [], 2)

    @mock.patch.object(functional_models.Cluster, 'read_slice_lba_checksum')
    def test_api_returns_insufficient_reps(self, mock_slice_checksum):
        mock_slice_checksum.return_value = {
            '1': 12345,
        }

        with pytest.raises(vl_helpers.HelperError):
            _volume._check_slice_lba_chunk(CLUSTER_FM, 1, 2, 3, {}, 0, 1, [], 2)

    @mock.patch.object(functional_models.Cluster, 'read_slice_lba_checksum')
    def test_ignored_services(self, mock_slice_checksum):
        mock_slice_checksum.return_value = {
            '1': 12345,
            '2': 123456
        }

        mydict = {}
        _volume._check_slice_lba_chunk(CLUSTER_FM, 1, 2, 3, mydict, 0, 1, ['1'], 2)

        assert mydict == {(0, 1): 123456}


@mock.patch.object(functional_models.Cluster, 'create_account')
class TestCreateAccount(unittest.TestCase):
    """Tests both create_account and create_accounts."""

    def setUp(self):
        self.account_dm = models.Account.get_mock_object()
        self.duplicate_username_error = sf_api.api_errors.DuplicateUsername(
            'xDuplicateUsername', mock.Mock())

    def test_create_account(self, mock_create_account):
        mock_create_account.return_value = []
        vl_helpers.create_account(CLUSTER_FM, self.account_dm.username)

    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_create_account_duplicate_username(self, mock_list_accounts, mock_create_account):
        mock_create_account.side_effect = self.duplicate_username_error
        vl_helpers.create_account(CLUSTER_FM, self.account_dm.username,
                                  ignore_duplicate_username=True)

    def test_create_account_no_duplicate_username(self, mock_create_account):
        mock_create_account.side_effect = self.duplicate_username_error
        with pytest.raises(sf_api.api_errors.DuplicateUsername):
            vl_helpers.create_account(CLUSTER_FM, self.account_dm.username,
                                      ignore_duplicate_username=False)

    def test_create_accounts(self, mock_create_account):
        mock_create_account.side_effect = [
            result_models.AddAccount.get_mock_data(version=models.SfVersion('9.0')),
            result_models.AddAccount.get_mock_data(version=models.SfVersion('9.0'))
        ]

        vl_helpers.create_accounts(CLUSTER_FM, [self.account_dm, self.account_dm])

    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_create_accounts_duplicate_username(self, mock_list_accounts, mock_create_account):
        mock_create_account.side_effect = self.duplicate_username_error
        vl_helpers.create_accounts(CLUSTER_FM, [self.account_dm, self.account_dm],
                                   ignore_duplicate_username=True)

    def test_create_accounts_no_duplicate_username(self, mock_create_account):
        mock_create_account.side_effect = self.duplicate_username_error
        with pytest.raises(sf_api.api_errors.DuplicateUsername):
            vl_helpers.create_accounts(CLUSTER_FM, [self.account_dm, self.account_dm],
                                       ignore_duplicate_username=False)


@mock.patch.object(functional_models.Cluster, 'delete_accounts')
class TestDeleteAccounts(unittest.TestCase):
    def setUp(self):
        self.account_dm = models.Account.get_mock_object()

    def test_delete_accounts(self, mock_delete_accounts):
        assert vl_helpers.delete_accounts(CLUSTER_FM, [self.account_dm]) is None

    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_delete_accounts_purge_instant(self, mock_list_accounts, mock_delete_accounts):
        """Test when wait for purge is True but all Accounts are immediately removed."""
        assert vl_helpers.delete_accounts(CLUSTER_FM, [self.account_dm],
                                          wait_for_accounts_purge=True) is None

    @mock.patch.object(time, 'sleep')
    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_delete_accounts_purge_wait(self, mock_list_accounts, mock_sleep, mock_delete_accounts):
        """Test when wait for purge is True and an Account takes several seconds to disappear from
        ListAccounts."""
        self.account_dm.status = sf_api.AccountStatus.active
        account_dm2 = models.Account(self.account_dm.serialize())
        account_dm2.status = sf_api.AccountStatus.removed
        mock_list_accounts.side_effect = [[account_dm2], []]

        assert vl_helpers.delete_accounts(CLUSTER_FM, [self.account_dm],
                                          wait_for_accounts_purge=True) is None
        assert mock_list_accounts.call_count == 2

    @mock.patch.object(time, 'time')
    @mock.patch.object(time, 'sleep')
    @mock.patch.object(functional_models.Cluster, 'list_accounts')
    def test_delete_accounts_purge_timeout(self, mock_list_accounts, mock_sleep, mock_time,
                                           mock_delete_accounts):
        """Test when wait for purge is True and an Account does not disappear from ListAccounts before
        the timeout."""
        mock_time.side_effect = [0, 301]
        mock_list_accounts.return_value = [self.account_dm]

        with pytest.raises(vl_helpers._errors.HelperError):
            vl_helpers.delete_accounts(CLUSTER_FM, [self.account_dm],
                                       wait_for_accounts_purge=True)


@mock.patch('valence.functional_models.Cluster.compute_volume_checksum')
class TestVerifyVolumeChecksums(unittest.TestCase):

    def setUp(self):
        self.mock_cluster_fm = functional_models.Cluster.get_mock_fm()
        self.mock_volumes = models.Volume.get_mock_object_list(1)

    def test_verify_volume_checksums(self, mock_compute_volume_checksum):
        mock_compute_volume_checksum.return_value = {8: '0xf238e7ae3dc20514', 20: '0xf238e7ae3dc20514'}

        vl_helpers.verify_volume_checksums(self.mock_cluster_fm, self.mock_volumes)

    def test_verify_volume_checksums_bad_checksums(self, mock_compute_volume_checksum):
        mock_compute_volume_checksum.return_value = {8: '0xf238e7af3dc20514', 20: '0xf238e7ae3dc20514'}

        with pytest.raises(_errors.HelperError):
            vl_helpers.verify_volume_checksums(self.mock_cluster_fm, self.mock_volumes)
